// massdecayvevPQ.cc is part of the CHIRON ChPT at two loops program collection
// Copyright (C) 2015 Johan Bijnens, v1.0
// CHIRON is licenced under the GNU GPL version 2 or later,
// see COPYING for details.
// Please respect the Guidelines, see GUIDELINES for details.

// results for partially quenched masses and decay constants
// derived in
//J.~Bijnens, N.~Danielsson and T.~A.~Lahde,
//  ``Three-flavor partially quenched chiral perturbation theory at NNLO
// for meson masses and decay constants,''
//  Phys.\ Rev.\ D {\bf 73} (2006) 074509
//  [hep-lat/0602003].
// J.~Bijnens and T.~A.~Lahde,
//  ``Decay constants of pseudoscalar mesons to two loops in three-flavor
// partially quenched (chi)PT,''
//  Phys.\ Rev.\ D {\bf 71} (2005) 094502
// [hep-lat/0501014].
//  J.~Bijnens, N.~Danielsson and T.~A.~Lahde,
//  ``The Pseudoscalar meson mass to two loops in three-flavor partially
//  quenched chiPT,''
//  Phys.\ Rev.\ D {\bf 70} (2004) 111503
//  [hep-lat/0406017].
//
// contains the partially quenched  expressions for masses and decay
// constants for three sea-quark flavours 
// in terms of the lowest order masses and the chiral limit
// decay constant

#include <iostream>
#include <iomanip>
#include <string>

#include "inputsnf.h"
#include "Linf.h"
#include "Ki.h"
#include "oneloopintegrals.h"
#include "quenchedsunsetintegrals.h"
#include "massdecayvevPQ.h"

const double pi = M_PI;
const double pi2 = M_PI*M_PI;
const double pi16 = 1./(16.*pi*pi);
const double pi162 = pow(pi16,2);

// auxiliary internal functions
inline double RN(const double m11,const double m22){
  return 1./(m11-m22);
}
inline double RS(const double m11,const double m22,const double m33){
  return (m11-m22)/(m11-m33);
}
inline double RS(const double m11,const double m22,const double m33,
		 const double m44,const double m55){
  return (m11-m22)*(m11-m33)/((m11-m44)*(m11-m55));
}
inline double RS(const double m11,const double m22,const double m33,
		 const double m44,const double m55,const double m66,
		 const double m77){
  return (m11-m22)*(m11-m33)*(m11-m44)/((m11-m55)*(m11-m66)*(m11-m77));
}
inline double RD(const double m11,const double m22,const double m33,
		 const double m44,const double m55,const double m66){
  return (m11-m22)*RS(m11,m33,m44,m55,m66);
}
inline double RD(const double m11,const double m22,const double m33,
		 const double m44){
  return (m11-m22)*RS(m11,m33,m44);
}
inline double RC(const double m11,const double m44,const double m66,
		 const double me2){
  return RS(m11,m44,me2)+RS(m11,m66,me2)-RS(m11,m44,m66,me2,me2);
}
inline double RC(const double m11,const double m44,const double m55,
		 const double m66,const double mp2,const double me2){
  return RS(m11,m44,m55,mp2,me2)+RS(m11,m44,m66,mp2,me2)
    +RS(m11,m55,m66,mp2,me2)-RS(m11,m44,m55,m66,mp2,me2,me2)
    -RS(m11,m44,m55,m66,mp2,mp2,me2);
}
//*********************************************************************
// masses
// 1 valence mass 1 sea mass *****************************************

double mv1s1nf3p4(const quarkmassnf mass, const Linf Liin){
  return mv1s1nf3p4L(mass,Liin)+mv1s1nf3p4R(mass);
}

double mv1s1nf3p4L(const quarkmassnf mass, const Linf Liin){
  double L0r,L1r,L2r,L3r,L4r,L5r,L6r,L7r,L8r,L9r,L10r,L11r,H1r,H2r,mu;
  int nfLi;
  Liin.out(L0r,L1r,L2r,L3r,L4r,L5r,L6r,L7r,L8r,L9r,L10r,L11r,H1r,H2r,mu,nfLi);
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if (mu0 != mu){
    cout << "ERROR: differing mu's in mv1s1nf3p4L\n";}
  if ( (nq != 2) || (B0mq.size() != 2)){
    cout << "ERROR: mv1s1nf3p4L probably called with wrong number"
	 << "of quark masses\n";}
  // two masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m44 = 2.*B0mq[1];
  double M12p4x11L =
       + 48*m11*m44*L6r - 24*m11*m44*L4r + 16*pow(m11,2)*L8r - 8*pow(
         m11,2)*L5r;

  return M12p4x11L/pow(f0,2);
}


double mv1s1nf3p4R(const quarkmassnf mass){
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if ( (nq != 2) || (B0mq.size() != 2)){
    cout << "ERROR: mv2s1nf3p4R probably called with wrong number"
	 << "of quark masses\n";}
  // two masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m44 = 2.*B0mq[1];
  double mu2 = mu0*mu0;
  double M12p4x11R =
       + Ab(m11,mu2) * (  - 1./3.*m11 );

      M12p4x11R +=  + Bb(m11,mu2) * ( 1./3.*m11*m44 - 1./3.*pow(m11,2)
          );

  return M12p4x11R/pow(f0,2);
}

double mv1s1nf3p6(const quarkmassnf mass, const Linf Liin, const Ki Kiin){
  return mv1s1nf3p6L(mass,Liin)+mv1s1nf3p6K(mass,Kiin)+mv1s1nf3p6R(mass);
}

double mv1s1nf3p6L(const quarkmassnf mass, const Linf Liin){
  double L0r,L1r,L2r,L3r,L4r,L5r,L6r,L7r,L8r,L9r,L10r,L11r,H1r,H2r,mu;
  int nfLi;
  Liin.out(L0r,L1r,L2r,L3r,L4r,L5r,L6r,L7r,L8r,L9r,L10r,L11r,H1r,H2r,mu,nfLi);
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if (mu0 != mu){
    cout << "ERROR: differing mu's in mv1s1nf3p6L\n";}
  if ( (nq != 2) || (B0mq.size() != 2)){
    cout << "ERROR: mv1s1nf3p6L probably called with wrong number"
	 << "of quark masses\n";}
  // two masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m44 = 2.*B0mq[1];
  double m14 = (m11+m44)/2.;
  double mu2 = mu0*mu0;
  double M12p6x11L =
       + pi16 * ( 3./2.*m11*pow(m44,2)*L3r + 16*m11*pow(m44,2)*L2r + 3*
         m11*pow(m44,2)*L0r + 17./3.*pow(m11,2)*m44*L3r + 26./3.*pow(
         m11,2)*m44*L0r - 5./2.*pow(m11,3)*L3r + 2*pow(m11,3)*L2r + 4*
         pow(m11,3)*L1r - pow(m11,3)*L0r );

      M12p6x11L +=  + Ab(m11,mu2) * (  - 32*m11*m44*L7r - 16*m11*m44*
         L6r - 16./3.*m11*m44*L5r + 16*m11*m44*L4r + 8*m11*m44*L3r + 8*
         m11*m44*L0r - 64./3.*pow(m11,2)*L8r + 32*pow(m11,2)*L7r + 32*
         pow(m11,2)*L6r + 16*pow(m11,2)*L5r - 16*pow(m11,2)*L3r + 20*
         pow(m11,2)*L2r + 8*pow(m11,2)*L1r - 16*pow(m11,2)*L0r );

      M12p6x11L +=  + Ab(m14,mu2) * ( 48*m11*m44*L8r - 24*m11*m44*L5r
          + 30*m11*m44*L3r + 12*m11*m44*L0r + 48*pow(m11,2)*L8r - 24*
         pow(m11,2)*L5r + 30*pow(m11,2)*L3r + 12*pow(m11,2)*L0r );

      M12p6x11L +=  + Ab(m44,mu2) * ( 128*m11*m44*L6r - 128*m11*m44*L4r
          + 32*m11*m44*L2r + 128*m11*m44*L1r );

      M12p6x11L +=  + Bb(m11,mu2) * ( 16./3.*m11*pow(m44,2)*L8r + 16*
         m11*pow(m44,2)*L7r + 32*m11*pow(m44,2)*L6r - 24*m11*pow(m44,2)
         *L4r + 64./3.*pow(m11,2)*m44*L8r - 32*pow(m11,2)*m44*L7r - 48*
         pow(m11,2)*m44*L6r - 16*pow(m11,2)*m44*L5r + 32*pow(m11,2)*m44
         *L4r + 8*pow(m11,2)*m44*L3r + 8*pow(m11,2)*m44*L0r - 32*pow(
         m11,3)*L8r + 16*pow(m11,3)*L7r + 56./3.*pow(m11,3)*L5r - 8*
         pow(m11,3)*L3r - 8*pow(m11,3)*L0r );

      M12p6x11L +=  + Cb(m11,mu2) * ( 32*pow(m11,2)*pow(m44,2)*L6r - 16
         *pow(m11,2)*pow(m44,2)*L4r + 32./3.*pow(m11,3)*m44*L8r - 32*
         pow(m11,3)*m44*L6r - 16./3.*pow(m11,3)*m44*L5r + 16*pow(m11,3)
         *m44*L4r - 32./3.*pow(m11,4)*L8r + 16./3.*pow(m11,4)*L5r );

      M12p6x11L +=  - 1152*m11*pow(m44,2)*L4r*L6r + 576*m11*pow(m44,2)*
         pow(L4r,2) - 384*pow(m11,2)*m44*L5r*L6r - 384*pow(m11,2)*m44*
         L4r*L8r + 384*pow(m11,2)*m44*L4r*L5r - 128*pow(m11,3)*L5r*L8r
          + 64*pow(m11,3)*pow(L5r,2);

  return M12p6x11L/pow(f0,4);
}

double mv1s1nf3p6K(const quarkmassnf mass, const Ki Kiin){
  double KK[116],mu;
  Kiin.out(KK,mu);
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if (mu0 != mu){
    cout << "ERROR: differing mu's in mv1s1nf3p6L\n";}
  if ( (nq != 2) || (B0mq.size() != 2)){
    cout << "ERROR: mv1s1nf3p6K probably called with wrong number"
	 << "of quark masses\n";}
  // two masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m44 = 2.*B0mq[1];
  double M12p6x11K =
       - 32*KK[17]*pow(m11,3) - 96*KK[18]*pow(m11,2)*m44 - 16*KK[19]*
         pow(m11,3) - 48*KK[20]*pow(m11,2)*m44 - 48*KK[21]*m11*pow(
         m44,2) - 144*KK[22]*m11*pow(m44,2) - 16*KK[23]*pow(m11,3) + 48
         *KK[25]*pow(m11,3) + 48*KK[26]*m11*pow(m44,2) + 96*KK[26]*pow(
         m11,2)*m44 + 432*KK[27]*m11*pow(m44,2) + 32*KK[39]*pow(m11,3)
          + 96*KK[40]*pow(m11,2)*m44;

  return M12p6x11K/pow(f0,4);
}

double mv1s1nf3p6R(const quarkmassnf mass){
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if ( (nq != 2) || (B0mq.size() != 2)){
    cout << "ERROR: mv1s1nf3p6R probably called with wrong number"
	 << "of quark masses\n";}
  // two masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m44 = 2.*B0mq[1];
  double m14 = (m11+m44)/2.;
  double mu2 = mu0*mu0;

  double    M12p6x11R =
       + pi162 * (  - 63./64.*m11*pow(m44,2) - 521./288.*pow(m11,2)*m44
          - 187./288.*pow(m11,3) - 17./48.*pi2*m11*pow(m44,2) - 41./108.
         *pi2*pow(m11,2)*m44 - 5./54.*pi2*pow(m11,3) );

      M12p6x11R +=  + pow(Ab(m11,mu2),2) * ( 5./18.*m44 - 13./72.*m11 )
         ;

      M12p6x11R +=  + Ab(m11,mu2)*Bb(m11,mu2) * (  - 11./12.*m11*m44 + 
         77./36.*pow(m11,2) );

      M12p6x11R +=  + Ab(m11,mu2)*Cb(m11,mu2) * (  - 2./9.*pow(m11,2)*
         m44 + 2./9.*pow(m11,3) );

      M12p6x11R +=  + Ab(m14,mu2)*pi16 * ( 3*m11*m44 );

      M12p6x11R +=  + pow(Ab(m14,mu2),2) * (  - 27/(4*m44 + 4*m11)*m11*
         m44 - 19/(4*m44 + 4*m11)*pow(m11,2) );

      M12p6x11R +=  + Ab(m14,mu2)*Bb(m11,mu2) * (  - 2*m11*m44 );

      M12p6x11R +=  + pow(Ab(m44,mu2),2) * (  - m11 );

      M12p6x11R +=  + Ab(m44,mu2)*Bb(m11,mu2) * ( 8./9.*m11*m44 );

      M12p6x11R +=  + pow(Bb(m11,mu2),2) * ( 1./18.*m11*pow(m44,2) - 1./
         3.*pow(m11,2)*m44 + 5./18.*pow(m11,3) );

      M12p6x11R +=  + Bb(m11,mu2)*Cb(m11,mu2) * ( 2./9.*pow(m11,2)*pow(
         m44,2) - 4./9.*pow(m11,3)*m44 + 2./9.*pow(m11,4) );

      M12p6x11R +=  + hh(1,m11,m11,m11,m11,mu2) * ( 5./9.*pow(m11,2) );

      M12p6x11R +=  + hh(1,m11,m14,m14,m11,mu2) * ( 1./4.*m11*m44 - 
         pow(m11,2) );

      M12p6x11R +=  + hh(1,m14,m14,m44,m11,mu2) * ( 2*m11*m44 );

      M12p6x11R +=  + hh(2,m11,m11,m11,m11,mu2) * (  - 4./9.*pow(m11,2)
         *m44 + 4./9.*pow(m11,3) );

      M12p6x11R +=  + hh(2,m11,m14,m14,m11,mu2) * (  - 3./4.*pow(m11,2)
         *m44 + 3./4.*pow(m11,3) );

      M12p6x11R +=  + hh(5,m11,m11,m11,m11,mu2) * ( 2./9.*pow(m11,2)*
         pow(m44,2) - 4./9.*pow(m11,3)*m44 + 2./9.*pow(m11,4) );

      M12p6x11R +=  + hh1(3,m14,m11,m14,m11,mu2) * ( 4*pow(m11,2)*m44
          - 4*pow(m11,3) );

      M12p6x11R +=  + hh21(1,m11,m14,m14,m11,mu2) * ( 3./4.*pow(m11,2)
          );

      M12p6x11R +=  + hh21(1,m44,m14,m14,m11,mu2) * ( 6*pow(m11,2) );

      M12p6x11R +=  + hh21(2,m11,m14,m14,m11,mu2) * ( 3./4.*pow(m11,2)*
         m44 - 3./4.*pow(m11,3) );

  return M12p6x11R/pow(f0,4);
}
// 2 valence mass 1 sea mass *****************************************

double mv2s1nf3p4(const quarkmassnf mass, const Linf Liin){
  return mv2s1nf3p4L(mass,Liin)+mv2s1nf3p4R(mass);
}

double mv2s1nf3p4L(const quarkmassnf mass, const Linf Liin){
  double L0r,L1r,L2r,L3r,L4r,L5r,L6r,L7r,L8r,L9r,L10r,L11r,H1r,H2r,mu;
  int nfLi;
  Liin.out(L0r,L1r,L2r,L3r,L4r,L5r,L6r,L7r,L8r,L9r,L10r,L11r,H1r,H2r,mu,nfLi);
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if (mu0 != mu){
    cout << "ERROR: differing mu's in mv1s1nf3p4L\n";}
  if ( (nq != 3) || (B0mq.size() != 3)){
    cout << "ERROR: mv1s1nf3p4L probably called with wrong number"
	 << "of quark masses\n";}
  // three masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m33 = 2.*B0mq[1];
  double m44 = 2.*B0mq[2];
  double M13p4x21L =
       + 24*m33*m44*L6r - 12*m33*m44*L4r + 4*pow(m33,2)*L8r - 2*pow(
         m33,2)*L5r + 24*m11*m44*L6r - 12*m11*m44*L4r + 8*m11*m33*L8r
          - 4*m11*m33*L5r + 4*pow(m11,2)*L8r - 2*pow(m11,2)*L5r;

  return M13p4x21L/pow(f0,2);
}

double mv2s1nf3p4R(const quarkmassnf mass){
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if ( (nq != 3) || (B0mq.size() != 3)){
    cout << "ERROR: mv1s1nf3p4R probably called with wrong number"
	 << "of quark masses\n";}
  // three masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m33 = 2.*B0mq[1];
  double m44 = 2.*B0mq[2];
  double RN13 = RN(m11,m33);
  double mu2 = mu0*mu0;
  double M13p4x21R =
       + Ab(m11,mu2) * ( 1./6.*m44 - 1./3.*m33 + 1./3.*m33*m44*RN13 - 1.
         /3.*pow(m33,2)*RN13 - 1./6.*m11 );

      M13p4x21R +=  + Ab(m33,mu2) * (  - 1./6.*m44 + 1./6.*m33 - 1./3.*
         m33*m44*RN13 + 1./3.*pow(m33,2)*RN13 );

  return M13p4x21R/pow(f0,2);
}

double mv2s1nf3p6(const quarkmassnf mass, const Linf Liin, const Ki Kiin){
  return mv2s1nf3p6L(mass,Liin)+mv2s1nf3p6K(mass,Kiin)+mv2s1nf3p6R(mass);
}

double mv2s1nf3p6L(const quarkmassnf mass, const Linf Liin){
  double L0r,L1r,L2r,L3r,L4r,L5r,L6r,L7r,L8r,L9r,L10r,L11r,H1r,H2r,mu;
  int nfLi;
  Liin.out(L0r,L1r,L2r,L3r,L4r,L5r,L6r,L7r,L8r,L9r,L10r,L11r,H1r,H2r,mu,nfLi);
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if (mu0 != mu){
    cout << "ERROR: differing mu's in mv2s1nf3p6L\n";}
  if ( (nq != 3) || (B0mq.size() != 3)){
    cout << "ERROR: mv2s1nf3p6L probably called with wrong number"
	 << "of quark masses\n";}
  // three masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m33 = 2.*B0mq[1];
  double m44 = 2.*B0mq[2];
  double m13 = (m11+m33)/2.;
  double m14 = (m11+m44)/2.;
  double m34 = (m33+m44)/2.;
  double RN13 = RN(m11,m33);
  double mu2 = mu0*mu0;
  double M13p6x21L =
       + pi16 * ( 3./4.*m33*pow(m44,2)*L3r + 8*m33*pow(m44,2)*L2r + 3./
         2.*m33*pow(m44,2)*L0r + 17./12.*pow(m33,2)*m44*L3r + 13./6.*
         pow(m33,2)*m44*L0r - 11./24.*pow(m33,3)*L3r + 1./4.*pow(m33,3)
         *L2r + 1./2.*pow(m33,3)*L1r - 1./4.*pow(m33,3)*L0r + 3./4.*m11
         *pow(m44,2)*L3r + 8*m11*pow(m44,2)*L2r + 3./2.*m11*pow(m44,2)*
         L0r + 17./6.*m11*m33*m44*L3r + 13./3.*m11*m33*m44*L0r - 19./24.
         *m11*pow(m33,2)*L3r + 3./4.*m11*pow(m33,2)*L2r + 3./2.*m11*
         pow(m33,2)*L1r - 1./4.*m11*pow(m33,2)*L0r + 17./12.*pow(m11,2)
         *m44*L3r + 13./6.*pow(m11,2)*m44*L0r - 19./24.*pow(m11,2)*m33*
         L3r + 3./4.*pow(m11,2)*m33*L2r + 3./2.*pow(m11,2)*m33*L1r - 1./
         4.*pow(m11,2)*m33*L0r - 11./24.*pow(m11,3)*L3r + 1./4.*pow(
         m11,3)*L2r + 1./2.*pow(m11,3)*L1r - 1./4.*pow(m11,3)*L0r );

      M13p6x21L +=  + Ab(m11,mu2) * ( 8./3.*pow(m44,2)*L8r + 8*pow(
         m44,2)*L7r + 8*pow(m44,2)*L6r - 8*pow(m44,2)*L4r - 32*m33*m44*
         L7r - 16*m33*m44*L6r - 16./3.*m33*m44*L5r + 16*m33*m44*L4r + 3
         *m33*m44*L3r + 6*m33*m44*L0r + 16./3.*m33*pow(m44,2)*L8r*RN13
          + 16*m33*pow(m44,2)*L7r*RN13 + 16*m33*pow(m44,2)*L6r*RN13 - 
         16*m33*pow(m44,2)*L4r*RN13 - 16./3.*pow(m33,2)*L8r + 16*pow(
         m33,2)*L7r + 16./3.*pow(m33,2)*L5r - 4./3.*pow(m33,2)*L3r - 16.
         /3.*pow(m33,2)*L0r - 32*pow(m33,2)*m44*L7r*RN13 - 16*pow(
         m33,2)*m44*L6r*RN13 - 16./3.*pow(m33,2)*m44*L5r*RN13 + 16*pow(
         m33,2)*m44*L4r*RN13 + 4./3.*pow(m33,2)*m44*L3r*RN13 + 16./3.*
         pow(m33,2)*m44*L0r*RN13 - 16./3.*pow(m33,3)*L8r*RN13 + 16*pow(
         m33,3)*L7r*RN13 + 16./3.*pow(m33,3)*L5r*RN13 - 4./3.*pow(
         m33,3)*L3r*RN13 - 16./3.*pow(m33,3)*L0r*RN13 - 16*m11*m44*L7r
          - 8*m11*m44*L6r - 8./3.*m11*m44*L5r + 8*m11*m44*L4r + 7./3.*
         m11*m44*L3r + 10./3.*m11*m44*L0r - 8*m11*m33*L8r + 16*m11*m33*
         L7r );

      M13p6x21L +=  + Ab(m11,mu2) * ( 20./3.*m11*m33*L5r - 14./3.*m11*
         m33*L3r - 20./3.*m11*m33*L0r - 16./3.*pow(m11,2)*L8r + 8*pow(
         m11,2)*L7r + 4*pow(m11,2)*L5r - 4*pow(m11,2)*L3r - 4*pow(
         m11,2)*L0r );

      M13p6x21L +=  + Ab(m13,mu2) * ( 8*pow(m33,2)*L6r + 5*pow(m33,2)*
         L2r + 2*pow(m33,2)*L1r + 16*m11*m33*L6r + 10*m11*m33*L2r + 4*
         m11*m33*L1r + 8*pow(m11,2)*L6r + 5*pow(m11,2)*L2r + 2*pow(
         m11,2)*L1r );

      M13p6x21L +=  + Ab(m14,mu2) * ( 12*m33*m44*L8r - 6*m33*m44*L5r + 
         15./2.*m33*m44*L3r + 3*m33*m44*L0r + 12*m11*m44*L8r - 6*m11*
         m44*L5r + 15./2.*m11*m44*L3r + 3*m11*m44*L0r + 12*m11*m33*L8r
          - 6*m11*m33*L5r + 15./2.*m11*m33*L3r + 3*m11*m33*L0r + 12*
         pow(m11,2)*L8r - 6*pow(m11,2)*L5r + 15./2.*pow(m11,2)*L3r + 3*
         pow(m11,2)*L0r );

      M13p6x21L +=  + Ab(m33,mu2) * (  - 8./3.*pow(m44,2)*L8r - 8*pow(
         m44,2)*L7r - 8*pow(m44,2)*L6r + 8*pow(m44,2)*L4r + 16*m33*m44*
         L7r + 8*m33*m44*L6r + 8./3.*m33*m44*L5r - 8*m33*m44*L4r + m33*
         m44*L3r - 2*m33*m44*L0r - 16./3.*m33*pow(m44,2)*L8r*RN13 - 16*
         m33*pow(m44,2)*L7r*RN13 - 16*m33*pow(m44,2)*L6r*RN13 + 16*m33*
         pow(m44,2)*L4r*RN13 - 8*pow(m33,2)*L7r - 4./3.*pow(m33,2)*L5r
          - 8./3.*pow(m33,2)*L3r + 4./3.*pow(m33,2)*L0r + 32*pow(m33,2)
         *m44*L7r*RN13 + 16*pow(m33,2)*m44*L6r*RN13 + 16./3.*pow(m33,2)
         *m44*L5r*RN13 - 16*pow(m33,2)*m44*L4r*RN13 - 4./3.*pow(m33,2)*
         m44*L3r*RN13 - 16./3.*pow(m33,2)*m44*L0r*RN13 + 16./3.*pow(
         m33,3)*L8r*RN13 - 16*pow(m33,3)*L7r*RN13 - 16./3.*pow(m33,3)*
         L5r*RN13 + 4./3.*pow(m33,3)*L3r*RN13 + 16./3.*pow(m33,3)*L0r*
         RN13 + 5./3.*m11*m44*L3r + 2./3.*m11*m44*L0r - 8./3.*m11*m33*
         L8r + 4./3.*m11*m33*L5r - 10./3.*m11*m33*L3r - 4./3.*m11*m33*
         L0r );

      M13p6x21L +=  + Ab(m34,mu2) * ( 12*m33*m44*L8r - 6*m33*m44*L5r + 
         15./2.*m33*m44*L3r + 3*m33*m44*L0r + 12*pow(m33,2)*L8r - 6*
         pow(m33,2)*L5r + 15./2.*pow(m33,2)*L3r + 3*pow(m33,2)*L0r + 12
         *m11*m44*L8r - 6*m11*m44*L5r + 15./2.*m11*m44*L3r + 3*m11*m44*
         L0r + 12*m11*m33*L8r - 6*m11*m33*L5r + 15./2.*m11*m33*L3r + 3*
         m11*m33*L0r );

      M13p6x21L +=  + Ab(m44,mu2) * ( 64*m33*m44*L6r - 64*m33*m44*L4r
          + 16*m33*m44*L2r + 64*m33*m44*L1r + 64*m11*m44*L6r - 64*m11*
         m44*L4r + 16*m11*m44*L2r + 64*m11*m44*L1r );

      M13p6x21L +=  + Bb(m11,mu2) * ( 16*m33*pow(m44,2)*L6r - 8*m33*
         pow(m44,2)*L4r + 16./3.*pow(m33,2)*m44*L8r - 16*pow(m33,2)*m44
         *L6r - 8./3.*pow(m33,2)*m44*L5r + 8*pow(m33,2)*m44*L4r + 16*
         pow(m33,2)*pow(m44,2)*L6r*RN13 - 8*pow(m33,2)*pow(m44,2)*L4r*
         RN13 - 16./3.*pow(m33,3)*L8r + 8./3.*pow(m33,3)*L5r + 16./3.*
         pow(m33,3)*m44*L8r*RN13 - 16*pow(m33,3)*m44*L6r*RN13 - 8./3.*
         pow(m33,3)*m44*L5r*RN13 + 8*pow(m33,3)*m44*L4r*RN13 - 16./3.*
         pow(m33,4)*L8r*RN13 + 8./3.*pow(m33,4)*L5r*RN13 + 8*m11*pow(
         m44,2)*L6r - 4*m11*pow(m44,2)*L4r + 8*m11*m33*m44*L8r - 16*m11
         *m33*m44*L6r - 4*m11*m33*m44*L5r + 8*m11*m33*m44*L4r + 5./3.*
         m11*m33*m44*L3r + 2./3.*m11*m33*m44*L0r - 16./3.*m11*pow(
         m33,2)*L8r + 8./3.*m11*pow(m33,2)*L5r + 16./3.*pow(m11,2)*m44*
         L8r - 8*pow(m11,2)*m44*L6r - 8./3.*pow(m11,2)*m44*L5r + 4*pow(
         m11,2)*m44*L4r + 5./3.*pow(m11,2)*m44*L3r + 2./3.*pow(m11,2)*
         m44*L0r - 8*pow(m11,2)*m33*L8r + 4*pow(m11,2)*m33*L5r - 5./3.*
         pow(m11,2)*m33*L3r );

      M13p6x21L +=  + Bb(m11,mu2) * (  - 2./3.*pow(m11,2)*m33*L0r - 16./
         3.*pow(m11,3)*L8r + 8./3.*pow(m11,3)*L5r - 5./3.*pow(m11,3)*
         L3r - 2./3.*pow(m11,3)*L0r );

      M13p6x21L +=  + Bb(m33,mu2) * (  - 8*m33*pow(m44,2)*L6r + 4*m33*
         pow(m44,2)*L4r + 8*pow(m33,2)*m44*L6r - 4*pow(m33,2)*m44*L4r
          + 5./3.*pow(m33,2)*m44*L3r + 2./3.*pow(m33,2)*m44*L0r - 16*
         pow(m33,2)*pow(m44,2)*L6r*RN13 + 8*pow(m33,2)*pow(m44,2)*L4r*
         RN13 - 5./3.*pow(m33,3)*L3r - 2./3.*pow(m33,3)*L0r - 16./3.*
         pow(m33,3)*m44*L8r*RN13 + 16*pow(m33,3)*m44*L6r*RN13 + 8./3.*
         pow(m33,3)*m44*L5r*RN13 - 8*pow(m33,3)*m44*L4r*RN13 + 16./3.*
         pow(m33,4)*L8r*RN13 - 8./3.*pow(m33,4)*L5r*RN13 + 8./3.*m11*
         m33*m44*L8r - 4./3.*m11*m33*m44*L5r + 5./3.*m11*m33*m44*L3r + 
         2./3.*m11*m33*m44*L0r - 8./3.*m11*pow(m33,2)*L8r + 4./3.*m11*
         pow(m33,2)*L5r - 5./3.*m11*pow(m33,2)*L3r - 2./3.*m11*pow(
         m33,2)*L0r );

      M13p6x21L +=  - 576*m33*pow(m44,2)*L4r*L6r + 288*m33*pow(m44,2)*
         pow(L4r,2) - 96*pow(m33,2)*m44*L5r*L6r - 96*pow(m33,2)*m44*L4r
         *L8r + 96*pow(m33,2)*m44*L4r*L5r - 16*pow(m33,3)*L5r*L8r + 8*
         pow(m33,3)*pow(L5r,2) - 576*m11*pow(m44,2)*L4r*L6r + 288*m11*
         pow(m44,2)*pow(L4r,2) - 192*m11*m33*m44*L5r*L6r - 192*m11*m33*
         m44*L4r*L8r + 192*m11*m33*m44*L4r*L5r - 48*m11*pow(m33,2)*L5r*
         L8r + 24*m11*pow(m33,2)*pow(L5r,2) - 96*pow(m11,2)*m44*L5r*L6r
          - 96*pow(m11,2)*m44*L4r*L8r + 96*pow(m11,2)*m44*L4r*L5r - 48*
         pow(m11,2)*m33*L5r*L8r + 24*pow(m11,2)*m33*pow(L5r,2) - 16*
         pow(m11,3)*L5r*L8r + 8*pow(m11,3)*pow(L5r,2);

  return M13p6x21L/pow(f0,4);
}

double mv2s1nf3p6K(const quarkmassnf mass, const Ki Kiin){
  double KK[116],mu;
  Kiin.out(KK,mu);
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if (mu0 != mu){
    cout << "ERROR: differing mu's in mv2s1nf3p6L\n";}
  if ( (nq != 3) || (B0mq.size() != 3)){
    cout << "ERROR: mv2s1nf3p6K probably called with wrong number"
	 << "of quark masses\n";}
  // two masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m33 = 2.*B0mq[1];
  double m44 = 2.*B0mq[2];
  double M13p6x21K =
       - 4*KK[17]*pow(m33,3) - 12*KK[17]*m11*pow(m33,2) - 12*KK[17]*
         pow(m11,2)*m33 - 4*KK[17]*pow(m11,3) - 24*KK[18]*pow(m33,2)*
         m44 - 48*KK[18]*m11*m33*m44 - 24*KK[18]*pow(m11,2)*m44 - 4*KK[
         19]*pow(m33,3) - 4*KK[19]*m11*pow(m33,2) - 4*KK[19]*pow(m11,2)
         *m33 - 4*KK[19]*pow(m11,3) - 12*KK[20]*pow(m33,2)*m44 - 24*KK[
         20]*m11*m33*m44 - 12*KK[20]*pow(m11,2)*m44 - 24*KK[21]*m33*
         pow(m44,2) - 24*KK[21]*m11*pow(m44,2) - 72*KK[22]*m33*pow(
         m44,2) - 72*KK[22]*m11*pow(m44,2) - 8*KK[23]*m11*pow(m33,2) - 
         8*KK[23]*pow(m11,2)*m33 + 12*KK[25]*pow(m33,3) + 12*KK[25]*m11
         *pow(m33,2) + 12*KK[25]*pow(m11,2)*m33 + 12*KK[25]*pow(m11,3)
          + 24*KK[26]*m33*pow(m44,2) + 24*KK[26]*pow(m33,2)*m44 + 24*
         KK[26]*m11*pow(m44,2) + 48*KK[26]*m11*m33*m44 + 24*KK[26]*pow(
         m11,2)*m44 + 216*KK[27]*m33*pow(m44,2) + 216*KK[27]*m11*pow(
         m44,2) + 4*KK[39]*pow(m33,3) + 12*KK[39]*m11*pow(m33,2) + 12*
         KK[39]*pow(m11,2)*m33;
      M13p6x21K +=  + 4*KK[39]*pow(m11,3) + 24*KK[40]*pow(m33,2)*m44 + 
         48*KK[40]*m11*m33*m44 + 24*KK[40]*pow(m11,2)*m44;

  return M13p6x21K/pow(f0,4);
}

double mv2s1nf3p6R(const quarkmassnf mass){
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if ( (nq != 3) || (B0mq.size() != 3)){
    cout << "ERROR: mv2s1nf3p6R probably called with wrong number"
	 << "of quark masses\n";}
  // three masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m33 = 2.*B0mq[1];
  double m44 = 2.*B0mq[2];
  double m13 = (m11+m33)/2.;
  double m14 = (m11+m44)/2.;
  double m34 = (m33+m44)/2.;
  double RN13 = RN(m11,m33);
  double mu2 = mu0*mu0;

  double M13p6x21R =
       + pi162 * ( 9./128.*m33*pow(m44,2) - 9./16.*m33*m34*m44 - 17./
         1152.*pow(m33,2)*m44 - 5./32.*pow(m33,2)*m34 + 521./4608.*pow(
         m33,3) - 9./16.*m14*m33*m44 - 5./32.*m14*pow(m33,2) - 31./144.
         *m13*pow(m33,2) + 9./128.*m11*pow(m44,2) - 9./16.*m11*m34*m44
          - 17./576.*m11*m33*m44 - 5./16.*m11*m33*m34 + 469./1536.*m11*
         pow(m33,2) - 9./16.*m11*m14*m44 - 5./16.*m11*m14*m33 - 31./72.
         *m11*m13*m33 - 17./1152.*pow(m11,2)*m44 - 5./32.*pow(m11,2)*
         m34 + 469./1536.*pow(m11,2)*m33 - 5./32.*pow(m11,2)*m14 - 31./
         144.*pow(m11,2)*m13 + 521./4608.*pow(m11,3) - 1./12.*pi2*m33*
         pow(m44,2) - 3./32.*pi2*m33*m34*m44 - 19./864.*pi2*pow(m33,2)*
         m44 - 5./192.*pi2*pow(m33,2)*m34 + 47./1728.*pi2*pow(m33,3) - 
         3./32.*pi2*m14*m33*m44 - 5./192.*pi2*m14*pow(m33,2) - 31./864.
         *pi2*m13*pow(m33,2) - 1./12.*pi2*m11*pow(m44,2) - 3./32.*pi2*
         m11*m34*m44 - 19./432.*pi2*m11*m33*m44 - 5./96.*pi2*m11*m33*
         m34 + 29./576.*pi2*m11*pow(m33,2) - 3./32.*pi2*m11*m14*m44 - 5.
         /96.*pi2*m11*m14*m33 );

      M13p6x21R +=  + pi162 * (  - 31./432.*pi2*m11*m13*m33 - 19./864.*
         pi2*pow(m11,2)*m44 - 5./192.*pi2*pow(m11,2)*m34 + 29./576.*pi2
         *pow(m11,2)*m33 - 5./192.*pi2*pow(m11,2)*m14 - 31./864.*pi2*
         pow(m11,2)*m13 + 47./1728.*pi2*pow(m11,3) );

      M13p6x21R +=  + Ab(m11,mu2)*pi16 * (  - 1./8.*pow(m44,2) + 1./8.*
         m33*m44 - 1./4.*m33*pow(m44,2)*RN13 + 1./16.*pow(m33,2) + 1./6.
         *pow(m33,2)*m44*RN13 + 1./12.*pow(m33,3)*RN13 + 1./24.*m11*m44
          + 1./24.*m11*m33 + 1./48.*pow(m11,2) );

      M13p6x21R +=  + pow(Ab(m11,mu2),2) * ( 151./4320.*pow(m11,-2)*m33
         *pow(m44,2) + 13./1440.*pow(m11,-2)*pow(m33,2)*m44 + 1./270.*
         pow(m11,-2)*pow(m33,2)*pow(m44,2)*RN13 - 1./120.*pow(m11,-2)*
         pow(m33,3)*m44*RN13 + 223./4320.*pow(m11,-1)*pow(m44,2) + 89./
         4320.*pow(m11,-1)*m33*m44 + 131./2160.*pow(m11,-1)*m33*pow(
         m44,2)*RN13 - 89./2592.*pow(m11,-1)*pow(m33,2)*m44*RN13 - 1./
         54.*pow(m11,-1)*pow(m33,2)*pow(m44,2)*pow(RN13,2) + 1./162.*
         pow(m11,-1)*pow(m33,3)*m44*pow(RN13,2) - 23./288.*m44 + 151./
         2160.*pow(m44,2)*RN13 + 67./288.*m33 - 929./4320.*m33*m44*RN13
          + 2./27.*m33*pow(m44,2)*pow(RN13,2) + 1./9.*pow(m33,2)*RN13
          - 1./162.*pow(m33,2)*m44*pow(RN13,2) - 1./18.*pow(m33,3)*pow(
         RN13,2) + 47./288.*m11 );

      M13p6x21R +=  + Ab(m11,mu2)*Ab(m13,mu2) * ( 7./8.*m33 + pow(
         m33,2)*RN13 + 3./8.*m11 );

      M13p6x21R +=  + Ab(m11,mu2)*Ab(m14,mu2) * (  - 3./8.*m44 - 3./8.*
         m33 - 3./4.*m33*m44*RN13 - 1./4.*pow(m33,2)*RN13 - 1./4.*m11 )
         ;

      M13p6x21R +=  + Ab(m11,mu2)*Ab(m33,mu2) * (  - 1./18.*pow(m44,2)*
         RN13 + 1./24.*m33 - 1./9.*m33*pow(m44,2)*pow(RN13,2) + 1./6.*
         pow(m33,2)*RN13 + 1./9.*pow(m33,3)*pow(RN13,2) - 1./72.*m11 );

      M13p6x21R +=  + Ab(m11,mu2)*Ab(m34,mu2) * (  - 3./8.*m44 - 3./4.*
         m33*m44*RN13 - 1./4.*pow(m33,2)*RN13 + 1./8.*m11 );

      M13p6x21R +=  + Ab(m11,mu2)*Ab(m44,mu2) * ( 4./9.*m44 + 8./9.*m33
         *m44*RN13 );

      M13p6x21R +=  + Ab(m11,mu2)*Bb(m11,mu2) * (  - 151./2160.*pow(
         m11,-1)*m33*pow(m44,2) - 13./720.*pow(m11,-1)*pow(m33,2)*m44
          - 1./135.*pow(m11,-1)*pow(m33,2)*pow(m44,2)*RN13 + 1./60.*
         pow(m11,-1)*pow(m33,3)*m44*RN13 - 17./144.*pow(m44,2) + 1./36.
         *m33*m44 - 43./270.*m33*pow(m44,2)*RN13 - 1./48.*pow(m33,2) + 
         19./60.*pow(m33,2)*m44*RN13 - 1./9.*pow(m33,2)*pow(m44,2)*pow(
         RN13,2) - 1./6.*pow(m33,3)*RN13 + 2./9.*pow(m33,3)*m44*pow(
         RN13,2) - 1./9.*pow(m33,4)*pow(RN13,2) - 7./144.*m11*m44 + 29./
         144.*m11*m33 + 1./6.*pow(m11,2) );

      M13p6x21R +=  + Ab(m11,mu2)*Bb(m33,mu2) * (  - 1./24.*m33*m44 + 1.
         /18.*m33*pow(m44,2)*RN13 + 1./24.*pow(m33,2) - 2./9.*pow(
         m33,2)*m44*RN13 + 1./9.*pow(m33,2)*pow(m44,2)*pow(RN13,2) + 1./
         6.*pow(m33,3)*RN13 - 2./9.*pow(m33,3)*m44*pow(RN13,2) + 1./9.*
         pow(m33,4)*pow(RN13,2) + 1./72.*m11*m44 - 1./72.*m11*m33 );

      M13p6x21R +=  + pow(Ab(m13,mu2),2) * (  - 31./144.*pow(m13,-1)*
         pow(m33,2) - 1./8.*m33 - 31./72.*m11*pow(m13,-1)*m33 - 1./8.*
         m11 - 31./144.*pow(m11,2)*pow(m13,-1) );

      M13p6x21R +=  + Ab(m13,mu2)*Ab(m33,mu2) * (  - 5./8.*m33 - pow(
         m33,2)*RN13 - 1./8.*m11 );

      M13p6x21R +=  + Ab(m14,mu2)*pi16 * ( 3./4.*m33*m44 + 3./16.*pow(
         m33,2) + 3./4.*m11*m44 - 3./16.*pow(m11,2) );

      M13p6x21R +=  + pow(Ab(m14,mu2),2) * (  - 9./16.*pow(m14,-1)*m33*
         m44 - 5./32.*pow(m14,-1)*pow(m33,2) - 9./16.*m11*pow(m14,-1)*
         m44 - 5./16.*m11*pow(m14,-1)*m33 - 5./32.*pow(m11,2)*pow(
         m14,-1) );

      M13p6x21R +=  + Ab(m14,mu2)*Ab(m33,mu2) * ( 3./8.*m44 + 3./8.*m33
          + 3./4.*m33*m44*RN13 + 1./4.*pow(m33,2)*RN13 + 1./4.*m11 );

      M13p6x21R +=  + Ab(m14,mu2)*Ab(m34,mu2) * (  - 9./8.*m33 - 9./8.*
         m11 );

      M13p6x21R +=  + Ab(m14,mu2)*Bb(m33,mu2) * (  - 1./8.*m33*m44 + 1./
         8.*pow(m33,2) - 1./8.*m11*m44 + 1./8.*m11*m33 );

      M13p6x21R +=  + Ab(m33,mu2)*pi16 * ( 1./8.*pow(m44,2) - 1./8.*m33
         *m44 + 1./4.*m33*pow(m44,2)*RN13 - 1./16.*pow(m33,2) - 1./6.*
         pow(m33,2)*m44*RN13 - 1./12.*pow(m33,3)*RN13 - 1./24.*m11*m44
          - 1./24.*m11*m33 - 1./48.*pow(m11,2) );

      M13p6x21R +=  + pow(Ab(m33,mu2),2) * (  - 1./32.*pow(m33,-1)*pow(
         m44,2) + 53./288.*m44 - 7./72.*pow(m44,2)*RN13 - 17./288.*m33
          + 19./72.*m33*m44*RN13 + 1./18.*m33*pow(m44,2)*pow(RN13,2) - 
         5./18.*pow(m33,2)*RN13 - 1./18.*pow(m33,3)*pow(RN13,2) + 1./32.
         *m11*pow(m33,-2)*pow(m44,2) + 5./72.*m11*pow(m33,-1)*m44 + 19./
         288.*m11 + 5./288.*pow(m11,2)*pow(m33,-2)*m44 );

      M13p6x21R +=  + Ab(m33,mu2)*Ab(m34,mu2) * ( 3./8.*m44 + 3./4.*m33
         *m44*RN13 + 1./4.*pow(m33,2)*RN13 - 1./8.*m11 );

      M13p6x21R +=  + Ab(m33,mu2)*Ab(m44,mu2) * (  - 4./9.*m44 - 8./9.*
         m33*m44*RN13 );

      M13p6x21R +=  + Ab(m33,mu2)*Bb(m11,mu2) * ( 1./18.*pow(m44,2) - 5.
         /24.*m33*m44 + 1./6.*m33*pow(m44,2)*RN13 + 1./6.*pow(m33,2) - 
         4./9.*pow(m33,2)*m44*RN13 + 1./9.*pow(m33,2)*pow(m44,2)*pow(
         RN13,2) + 5./18.*pow(m33,3)*RN13 - 2./9.*pow(m33,3)*m44*pow(
         RN13,2) + 1./9.*pow(m33,4)*pow(RN13,2) - 1./24.*m11*m44 + 1./
         24.*m11*m33 - 1./72.*pow(m11,2) );

      M13p6x21R +=  + Ab(m33,mu2)*Bb(m33,mu2) * (  - 1./16.*pow(m44,2)
          + 1./16.*m33*m44 - 1./18.*m33*pow(m44,2)*RN13 + 1./3.*pow(
         m33,2)*m44*RN13 - 1./9.*pow(m33,2)*pow(m44,2)*pow(RN13,2) - 5./
         18.*pow(m33,3)*RN13 + 2./9.*pow(m33,3)*m44*pow(RN13,2) - 1./9.
         *pow(m33,4)*pow(RN13,2) - 1./16.*m11*pow(m33,-1)*pow(m44,2) - 
         1./12.*m11*m44 + 7./48.*m11*m33 - 5./144.*pow(m11,2)*pow(
         m33,-1)*m44 + 5./144.*pow(m11,2) );

      M13p6x21R +=  + Ab(m34,mu2)*pi16 * ( 3./4.*m33*m44 - 3./16.*pow(
         m33,2) + 3./4.*m11*m44 + 3./16.*pow(m11,2) );

      M13p6x21R +=  + pow(Ab(m34,mu2),2) * (  - 9./16.*m33*pow(m34,-1)*
         m44 - 5./32.*pow(m33,2)*pow(m34,-1) - 9./16.*m11*pow(m34,-1)*
         m44 - 5./16.*m11*m33*pow(m34,-1) - 5./32.*pow(m11,2)*pow(
         m34,-1) );

      M13p6x21R +=  + Ab(m34,mu2)*Bb(m11,mu2) * (  - 1./8.*m33*m44 - 1./
         8.*m11*m44 + 1./8.*m11*m33 + 1./8.*pow(m11,2) );

      M13p6x21R +=  + pow(Ab(m44,mu2),2) * (  - 1./2.*m33 - 1./2.*m11 )
         ;

      M13p6x21R +=  + Bb(m11,mu2)*pi16 * ( 1./16.*m33*pow(m44,2) + 1./
         48.*pow(m33,2)*m44 + 1./16.*m11*pow(m44,2) - 1./24.*m11*m33*
         m44 - 1./48.*m11*pow(m33,2) - 1./16.*pow(m11,2)*m44 - 1./48.*
         pow(m11,2)*m33 );

      M13p6x21R +=  + pow(Bb(m11,mu2),2) * ( 17./144.*m33*pow(m44,2) - 
         2./9.*pow(m33,2)*m44 + 1./9.*pow(m33,2)*pow(m44,2)*RN13 + 1./9.
         *pow(m33,3) - 2./9.*pow(m33,3)*m44*RN13 + 1./9.*pow(m33,4)*
         RN13 + 1./16.*m11*pow(m44,2) - 17./72.*m11*m33*m44 + 1./9.*m11
         *pow(m33,2) - 1./8.*pow(m11,2)*m44 + 17./144.*pow(m11,2)*m33
          + 1./16.*pow(m11,3) );

      M13p6x21R +=  + Bb(m11,mu2)*Bb(m33,mu2) * (  - 1./72.*m33*pow(
         m44,2) + 1./72.*pow(m33,2)*m44 - 1./72.*m11*pow(m44,2) + 1./36.
         *m11*m33*m44 - 1./72.*m11*pow(m33,2) + 1./72.*pow(m11,2)*m44
          - 1./72.*pow(m11,2)*m33 );

      M13p6x21R +=  + Bb(m33,mu2)*pi16 * ( 1./16.*m33*pow(m44,2) - 1./
         16.*pow(m33,2)*m44 + 1./16.*m11*pow(m44,2) - 1./24.*m11*m33*
         m44 - 1./48.*m11*pow(m33,2) + 1./48.*pow(m11,2)*m44 - 1./48.*
         pow(m11,2)*m33 );

      M13p6x21R +=  + pow(Bb(m33,mu2),2) * (  - 7./144.*m33*pow(m44,2)
          + 7./72.*pow(m33,2)*m44 - 1./9.*pow(m33,2)*pow(m44,2)*RN13 - 
         7./144.*pow(m33,3) + 2./9.*pow(m33,3)*m44*RN13 - 1./9.*pow(
         m33,4)*RN13 + 1./144.*m11*pow(m44,2) - 1./72.*m11*m33*m44 + 1./
         144.*m11*pow(m33,2) );

      M13p6x21R +=  + hh(1,m11,m11,m13,m13,mu2) * ( 1./24.*pow(m44,2)
          - 7./24.*m33*m44 + 1./6.*m33*pow(m44,2)*RN13 + 65./144.*pow(
         m33,2) - 11./18.*pow(m33,2)*m44*RN13 + 1./6.*pow(m33,2)*pow(
         m44,2)*pow(RN13,2) + 4./9.*pow(m33,3)*RN13 - 1./3.*pow(m33,3)*
         m44*pow(RN13,2) + 1./6.*pow(m33,4)*pow(RN13,2) - 5./72.*m11*
         m44 + 25./72.*m11*m33 + 11./144.*pow(m11,2) );

      M13p6x21R +=  + hh(1,m11,m13,m33,m13,mu2) * (  - 1./12.*pow(
         m44,2) + 5./12.*m33*m44 - 1./3.*m33*pow(m44,2)*RN13 - 49./144.
         *pow(m33,2) + pow(m33,2)*m44*RN13 - 1./3.*pow(m33,2)*pow(
         m44,2)*pow(RN13,2) - 2./3.*pow(m33,3)*RN13 + 2./3.*pow(m33,3)*
         m44*pow(RN13,2) - 1./3.*pow(m33,4)*pow(RN13,2) + 1./12.*m11*
         m44 - 7./72.*m11*m33 - 1./144.*pow(m11,2) );

      M13p6x21R +=  + hh(1,m11,m14,m34,m13,mu2) * (  - 1./8.*pow(m44,2)
          + m33*m44 - 1./4.*m33*pow(m44,2)*RN13 - 13./16.*pow(m33,2) + 
         5./4.*pow(m33,2)*m44*RN13 - pow(m33,3)*RN13 + 3./8.*m11*m44 - 
         1./4.*m11*m33 + 1./16.*pow(m11,2) );

      M13p6x21R +=  + hh(1,m13,m13,m13,m13,mu2) * ( 1./16.*pow(m33,2)
          + 1./8.*m11*m33 + 1./16.*pow(m11,2) );

      M13p6x21R +=  + hh(1,m13,m33,m33,m13,mu2) * ( 1./24.*pow(m44,2)
          - 1./8.*m33*m44 + 1./6.*m33*pow(m44,2)*RN13 + 19./144.*pow(
         m33,2) - 7./18.*pow(m33,2)*m44*RN13 + 1./6.*pow(m33,2)*pow(
         m44,2)*pow(RN13,2) + 2./9.*pow(m33,3)*RN13 - 1./3.*pow(m33,3)*
         m44*pow(RN13,2) + 1./6.*pow(m33,4)*pow(RN13,2) - 1./72.*m11*
         m44 + 17./72.*m11*m33 + 25./144.*pow(m11,2) );

      M13p6x21R +=  + hh(1,m14,m33,m34,m13,mu2) * ( 1./8.*pow(m44,2) + 
         5./8.*m33*m44 + 1./4.*m33*pow(m44,2)*RN13 - 15./16.*pow(m33,2)
          + 3./4.*pow(m33,2)*m44*RN13 - pow(m33,3)*RN13 + 1./4.*m11*m44
          - 3./4.*m11*m33 - 5./16.*pow(m11,2) );

      M13p6x21R +=  + hh(1,m14,m34,m44,m13,mu2) * ( m33*m44 + m11*m44 )
         ;

      M13p6x21R +=  + hh(2,m11,m11,m13,m13,mu2) * (  - 1./24.*m33*pow(
         m44,2) + 11./144.*pow(m33,2)*m44 - 1./18.*pow(m33,2)*pow(
         m44,2)*RN13 - 1./18.*pow(m33,3) + 1./9.*pow(m33,3)*m44*RN13 - 
         1./18.*pow(m33,4)*RN13 - 1./72.*m11*pow(m44,2) + 1./72.*m11*
         m33*m44 - 1./48.*m11*pow(m33,2) - 1./144.*pow(m11,2)*m44 + 1./
         36.*pow(m11,2)*m33 + 1./48.*pow(m11,3) );

      M13p6x21R +=  + hh(2,m11,m13,m33,m13,mu2) * ( 1./24.*m33*pow(
         m44,2) - 5./48.*pow(m33,2)*m44 + 1./18.*pow(m33,2)*pow(m44,2)*
         RN13 + 1./18.*pow(m33,3) - 1./9.*pow(m33,3)*m44*RN13 + 1./18.*
         pow(m33,4)*RN13 + 1./72.*m11*pow(m44,2) - 5./72.*m11*m33*m44
          + 7./144.*m11*pow(m33,2) - 1./48.*pow(m11,2)*m44 + 1./36.*
         pow(m11,2)*m33 + 1./144.*pow(m11,3) );

      M13p6x21R +=  + hh(2,m11,m14,m34,m13,mu2) * ( 1./16.*m33*pow(
         m44,2) + 1./16.*m11*pow(m44,2) - 1./16.*pow(m11,2)*m33 - 1./16.
         *pow(m11,3) );

      M13p6x21R +=  + hh(2,m33,m11,m13,m13,mu2) * (  - 1./24.*m33*pow(
         m44,2) + 13./144.*pow(m33,2)*m44 - 1./18.*pow(m33,2)*pow(
         m44,2)*RN13 - 7./144.*pow(m33,3) + 1./9.*pow(m33,3)*m44*RN13
          - 1./18.*pow(m33,4)*RN13 - 1./72.*m11*pow(m44,2) + 1./24.*m11
         *m33*m44 - 1./36.*m11*pow(m33,2) + 1./144.*pow(m11,2)*m44 - 1./
         144.*pow(m11,2)*m33 );

      M13p6x21R +=  + hh(2,m33,m13,m33,m13,mu2) * ( 1./24.*m33*pow(
         m44,2) - 17./144.*pow(m33,2)*m44 + 1./18.*pow(m33,2)*pow(
         m44,2)*RN13 + 11./144.*pow(m33,3) - 1./9.*pow(m33,3)*m44*RN13
          + 1./18.*pow(m33,4)*RN13 + 1./72.*m11*pow(m44,2) - 7./72.*m11
         *m33*m44 + 1./12.*m11*pow(m33,2) - 5./144.*pow(m11,2)*m44 + 5./
         144.*pow(m11,2)*m33 );

      M13p6x21R +=  + hh(2,m33,m14,m34,m13,mu2) * ( 1./16.*m33*pow(
         m44,2) - 1./16.*pow(m33,3) + 1./16.*m11*pow(m44,2) - 1./16.*
         m11*pow(m33,2) );

      M13p6x21R +=  + hh(5,m11,m11,m13,m13,mu2) * ( 5./288.*pow(m33,2)*
         pow(m44,2) + 5./144.*m11*m33*pow(m44,2) - 5./144.*m11*pow(
         m33,2)*m44 + 5./288.*pow(m11,2)*pow(m44,2) - 5./72.*pow(m11,2)
         *m33*m44 + 5./288.*pow(m11,2)*pow(m33,2) - 5./144.*pow(m11,3)*
         m44 + 5./144.*pow(m11,3)*m33 + 5./288.*pow(m11,4) );

      M13p6x21R +=  + hh(5,m11,m33,m13,m13,mu2) * ( 1./144.*pow(m33,2)*
         pow(m44,2) - 1./144.*pow(m33,3)*m44 + 1./72.*m11*m33*pow(
         m44,2) - 1./48.*m11*pow(m33,2)*m44 + 1./144.*m11*pow(m33,3) + 
         1./144.*pow(m11,2)*pow(m44,2) - 1./48.*pow(m11,2)*m33*m44 + 1./
         72.*pow(m11,2)*pow(m33,2) - 1./144.*pow(m11,3)*m44 + 1./144.*
         pow(m11,3)*m33 );

      M13p6x21R +=  + hh(5,m33,m33,m13,m13,mu2) * ( 5./288.*pow(m33,2)*
         pow(m44,2) - 5./144.*pow(m33,3)*m44 + 5./288.*pow(m33,4) + 5./
         144.*m11*m33*pow(m44,2) - 5./72.*m11*pow(m33,2)*m44 + 5./144.*
         m11*pow(m33,3) + 5./288.*pow(m11,2)*pow(m44,2) - 5./144.*pow(
         m11,2)*m33*m44 + 5./288.*pow(m11,2)*pow(m33,2) );

      M13p6x21R +=  + hh1(1,m11,m14,m34,m13,mu2) * (  - 3./2.*m33*m44
          + 3./2.*pow(m33,2) - 2*pow(m33,2)*m44*RN13 + 2*pow(m33,3)*
         RN13 - 1./2.*m11*m44 + 1./2.*m11*m33 );

      M13p6x21R +=  + hh1(1,m13,m11,m11,m13,mu2) * (  - 1./18.*pow(
         m44,2) + 5./18.*m33*m44 - 2./9.*m33*pow(m44,2)*RN13 - 1./2.*
         pow(m33,2) + 2./3.*pow(m33,2)*m44*RN13 - 2./9.*pow(m33,2)*pow(
         m44,2)*pow(RN13,2) - 4./9.*pow(m33,3)*RN13 + 4./9.*pow(m33,3)*
         m44*pow(RN13,2) - 2./9.*pow(m33,4)*pow(RN13,2) + 1./18.*m11*
         m44 - 11./18.*m11*m33 - 5./18.*pow(m11,2) );

      M13p6x21R +=  + hh1(1,m13,m11,m33,m13,mu2) * ( 1./9.*pow(m44,2)
          - 5./9.*m33*m44 + 4./9.*m33*pow(m44,2)*RN13 + 1./2.*pow(
         m33,2) - 4./3.*pow(m33,2)*m44*RN13 + 4./9.*pow(m33,2)*pow(
         m44,2)*pow(RN13,2) + 8./9.*pow(m33,3)*RN13 - 8./9.*pow(m33,3)*
         m44*pow(RN13,2) + 4./9.*pow(m33,4)*pow(RN13,2) - 1./9.*m11*m44
          + 2./9.*m11*m33 + 1./18.*pow(m11,2) );

      M13p6x21R +=  + hh1(1,m13,m33,m33,m13,mu2) * (  - 1./18.*pow(
         m44,2) + 5./18.*m33*m44 - 2./9.*m33*pow(m44,2)*RN13 - 1./2.*
         pow(m33,2) + 2./3.*pow(m33,2)*m44*RN13 - 2./9.*pow(m33,2)*pow(
         m44,2)*pow(RN13,2) - 4./9.*pow(m33,3)*RN13 + 4./9.*pow(m33,3)*
         m44*pow(RN13,2) - 2./9.*pow(m33,4)*pow(RN13,2) + 1./18.*m11*
         m44 - 11./18.*m11*m33 - 5./18.*pow(m11,2) );

      M13p6x21R +=  + hh1(1,m14,m11,m34,m13,mu2) * (  - 1./2.*pow(
         m33,2) - m11*m33 - 1./2.*pow(m11,2) );

      M13p6x21R +=  + hh1(1,m14,m33,m34,m13,mu2) * (  - 3./2.*m33*m44
          + 2*pow(m33,2) - 2*pow(m33,2)*m44*RN13 + 2*pow(m33,3)*RN13 - 
         1./2.*m11*m44 + 3./2.*m11*m33 + 1./2.*pow(m11,2) );

      M13p6x21R +=  + hh1(1,m34,m14,m33,m13,mu2) * (  - 3./2.*m33*m44
          + 3./2.*pow(m33,2) - 2*pow(m33,2)*m44*RN13 + 2*pow(m33,3)*
         RN13 - 1./2.*m11*m44 + 1./2.*m11*m33 );

      M13p6x21R +=  + hh1(3,m13,m11,m11,m13,mu2) * ( 1./6.*m33*pow(
         m44,2) - 7./18.*pow(m33,2)*m44 + 2./9.*pow(m33,2)*pow(m44,2)*
         RN13 + 2./9.*pow(m33,3) - 4./9.*pow(m33,3)*m44*RN13 + 2./9.*
         pow(m33,4)*RN13 + 1./18.*m11*pow(m44,2) - 2./9.*m11*m33*m44 + 
         1./6.*m11*pow(m33,2) - 1./18.*pow(m11,2)*m44 + 1./18.*pow(
         m11,2)*m33 );

      M13p6x21R +=  + hh1(3,m13,m11,m33,m13,mu2) * (  - 1./6.*m33*pow(
         m44,2) + 7./18.*pow(m33,2)*m44 - 2./9.*pow(m33,2)*pow(m44,2)*
         RN13 - 2./9.*pow(m33,3) + 4./9.*pow(m33,3)*m44*RN13 - 2./9.*
         pow(m33,4)*RN13 - 1./18.*m11*pow(m44,2) + 2./9.*m11*m33*m44 - 
         1./6.*m11*pow(m33,2) + 1./18.*pow(m11,2)*m44 - 1./18.*pow(
         m11,2)*m33 );

      M13p6x21R +=  + hh1(3,m13,m33,m11,m13,mu2) * ( 1./6.*m33*pow(
         m44,2) - 7./18.*pow(m33,2)*m44 + 2./9.*pow(m33,2)*pow(m44,2)*
         RN13 + 2./9.*pow(m33,3) - 4./9.*pow(m33,3)*m44*RN13 + 2./9.*
         pow(m33,4)*RN13 + 1./18.*m11*pow(m44,2) - 2./9.*m11*m33*m44 + 
         1./6.*m11*pow(m33,2) - 1./18.*pow(m11,2)*m44 + 1./18.*pow(
         m11,2)*m33 );

      M13p6x21R +=  + hh1(3,m13,m33,m33,m13,mu2) * (  - 1./6.*m33*pow(
         m44,2) + 7./18.*pow(m33,2)*m44 - 2./9.*pow(m33,2)*pow(m44,2)*
         RN13 - 2./9.*pow(m33,3) + 4./9.*pow(m33,3)*m44*RN13 - 2./9.*
         pow(m33,4)*RN13 - 1./18.*m11*pow(m44,2) + 2./9.*m11*m33*m44 - 
         1./6.*m11*pow(m33,2) + 1./18.*pow(m11,2)*m44 - 1./18.*pow(
         m11,2)*m33 );

      M13p6x21R +=  + hh1(7,m13,m11,m11,m13,mu2) * (  - 1./36.*pow(
         m33,2)*pow(m44,2) - 1./18.*m11*m33*pow(m44,2) + 1./18.*m11*
         pow(m33,2)*m44 - 1./36.*pow(m11,2)*pow(m44,2) + 1./9.*pow(
         m11,2)*m33*m44 - 1./36.*pow(m11,2)*pow(m33,2) + 1./18.*pow(
         m11,3)*m44 - 1./18.*pow(m11,3)*m33 - 1./36.*pow(m11,4) );

      M13p6x21R +=  + hh1(7,m13,m33,m33,m13,mu2) * (  - 1./36.*pow(
         m33,2)*pow(m44,2) + 1./18.*pow(m33,3)*m44 - 1./36.*pow(m33,4)
          - 1./18.*m11*m33*pow(m44,2) + 1./9.*m11*pow(m33,2)*m44 - 1./
         18.*m11*pow(m33,3) - 1./36.*pow(m11,2)*pow(m44,2) + 1./18.*
         pow(m11,2)*m33*m44 - 1./36.*pow(m11,2)*pow(m33,2) );

      M13p6x21R +=  + hh21(1,m11,m11,m13,m13,mu2) * (  - 3./16.*pow(
         m33,2) - 3./8.*m11*m33 - 3./16.*pow(m11,2) );

      M13p6x21R +=  + hh21(1,m11,m14,m34,m13,mu2) * ( 9./16.*m33*m44 - 
         9./16.*pow(m33,2) + 3./4.*pow(m33,2)*m44*RN13 - 3./4.*pow(
         m33,3)*RN13 + 3./16.*m11*m44 - 3./16.*m11*m33 );

      M13p6x21R +=  + hh21(1,m13,m11,m11,m13,mu2) * ( 1./24.*pow(m44,2)
          - 5./24.*m33*m44 + 1./6.*m33*pow(m44,2)*RN13 + 13./48.*pow(
         m33,2) - 1./2.*pow(m33,2)*m44*RN13 + 1./6.*pow(m33,2)*pow(
         m44,2)*pow(RN13,2) + 1./3.*pow(m33,3)*RN13 - 1./3.*pow(m33,3)*
         m44*pow(RN13,2) + 1./6.*pow(m33,4)*pow(RN13,2) - 1./24.*m11*
         m44 + 1./4.*m11*m33 + 5./48.*pow(m11,2) );

      M13p6x21R +=  + hh21(1,m13,m11,m33,m13,mu2) * (  - 1./12.*pow(
         m44,2) + 5./12.*m33*m44 - 1./3.*m33*pow(m44,2)*RN13 - 17./48.*
         pow(m33,2) + pow(m33,2)*m44*RN13 - 1./3.*pow(m33,2)*pow(m44,2)
         *pow(RN13,2) - 2./3.*pow(m33,3)*RN13 + 2./3.*pow(m33,3)*m44*
         pow(RN13,2) - 1./3.*pow(m33,4)*pow(RN13,2) + 1./12.*m11*m44 - 
         1./8.*m11*m33 - 1./48.*pow(m11,2) );

      M13p6x21R +=  + hh21(1,m13,m13,m13,m13,mu2) * ( 3./16.*pow(m33,2)
          + 3./8.*m11*m33 + 3./16.*pow(m11,2) );

      M13p6x21R +=  + hh21(1,m13,m33,m33,m13,mu2) * ( 1./24.*pow(m44,2)
          - 5./24.*m33*m44 + 1./6.*m33*pow(m44,2)*RN13 + 13./48.*pow(
         m33,2) - 1./2.*pow(m33,2)*m44*RN13 + 1./6.*pow(m33,2)*pow(
         m44,2)*pow(RN13,2) + 1./3.*pow(m33,3)*RN13 - 1./3.*pow(m33,3)*
         m44*pow(RN13,2) + 1./6.*pow(m33,4)*pow(RN13,2) - 1./24.*m11*
         m44 + 1./4.*m11*m33 + 5./48.*pow(m11,2) );

      M13p6x21R +=  + hh21(1,m14,m11,m34,m13,mu2) * (  - 9./16.*m33*m44
          + 3./4.*pow(m33,2) - 3./4.*pow(m33,2)*m44*RN13 + 3./4.*pow(
         m33,3)*RN13 - 3./16.*m11*m44 + 9./16.*m11*m33 + 3./16.*pow(
         m11,2) );

      M13p6x21R +=  + hh21(1,m14,m33,m34,m13,mu2) * ( 9./16.*m33*m44 - 
         3./4.*pow(m33,2) + 3./4.*pow(m33,2)*m44*RN13 - 3./4.*pow(
         m33,3)*RN13 + 3./16.*m11*m44 - 9./16.*m11*m33 - 3./16.*pow(
         m11,2) );

      M13p6x21R +=  + hh21(1,m33,m13,m33,m13,mu2) * (  - 3./16.*pow(
         m33,2) - 3./8.*m11*m33 - 3./16.*pow(m11,2) );

      M13p6x21R +=  + hh21(1,m33,m14,m34,m13,mu2) * (  - 9./16.*m33*m44
          + 3./4.*pow(m33,2) - 3./4.*pow(m33,2)*m44*RN13 + 3./4.*pow(
         m33,3)*RN13 - 3./16.*m11*m44 + 9./16.*m11*m33 + 3./16.*pow(
         m11,2) );

      M13p6x21R +=  + hh21(1,m34,m11,m14,m13,mu2) * (  - 9./16.*m33*m44
          + 9./16.*pow(m33,2) - 3./4.*pow(m33,2)*m44*RN13 + 3./4.*pow(
         m33,3)*RN13 - 3./16.*m11*m44 + 3./16.*m11*m33 );

      M13p6x21R +=  + hh21(1,m34,m14,m33,m13,mu2) * ( 9./16.*m33*m44 - 
         9./16.*pow(m33,2) + 3./4.*pow(m33,2)*m44*RN13 - 3./4.*pow(
         m33,3)*RN13 + 3./16.*m11*m44 - 3./16.*m11*m33 );

      M13p6x21R +=  + hh21(1,m44,m14,m34,m13,mu2) * ( 3./2.*pow(m33,2)
          + 3*m11*m33 + 3./2.*pow(m11,2) );

      M13p6x21R +=  + hh21(3,m13,m11,m11,m13,mu2) * (  - 1./8.*m33*pow(
         m44,2) + 5./16.*pow(m33,2)*m44 - 1./6.*pow(m33,2)*pow(m44,2)*
         RN13 - 1./6.*pow(m33,3) + 1./3.*pow(m33,3)*m44*RN13 - 1./6.*
         pow(m33,4)*RN13 - 1./24.*m11*pow(m44,2) + 5./24.*m11*m33*m44
          - 7./48.*m11*pow(m33,2) + 1./16.*pow(m11,2)*m44 - 1./12.*pow(
         m11,2)*m33 - 1./48.*pow(m11,3) );

      M13p6x21R +=  + hh21(3,m13,m11,m33,m13,mu2) * ( 1./8.*m33*pow(
         m44,2) - 5./16.*pow(m33,2)*m44 + 1./6.*pow(m33,2)*pow(m44,2)*
         RN13 + 1./6.*pow(m33,3) - 1./3.*pow(m33,3)*m44*RN13 + 1./6.*
         pow(m33,4)*RN13 + 1./24.*m11*pow(m44,2) - 5./24.*m11*m33*m44
          + 7./48.*m11*pow(m33,2) - 1./16.*pow(m11,2)*m44 + 1./12.*pow(
         m11,2)*m33 + 1./48.*pow(m11,3) );

      M13p6x21R +=  + hh21(3,m13,m33,m11,m13,mu2) * (  - 1./8.*m33*pow(
         m44,2) + 13./48.*pow(m33,2)*m44 - 1./6.*pow(m33,2)*pow(m44,2)*
         RN13 - 7./48.*pow(m33,3) + 1./3.*pow(m33,3)*m44*RN13 - 1./6.*
         pow(m33,4)*RN13 - 1./24.*m11*pow(m44,2) + 1./8.*m11*m33*m44 - 
         1./12.*m11*pow(m33,2) + 1./48.*pow(m11,2)*m44 - 1./48.*pow(
         m11,2)*m33 );

      M13p6x21R +=  + hh21(3,m13,m33,m33,m13,mu2) * ( 1./8.*m33*pow(
         m44,2) - 13./48.*pow(m33,2)*m44 + 1./6.*pow(m33,2)*pow(m44,2)*
         RN13 + 7./48.*pow(m33,3) - 1./3.*pow(m33,3)*m44*RN13 + 1./6.*
         pow(m33,4)*RN13 + 1./24.*m11*pow(m44,2) - 1./8.*m11*m33*m44 + 
         1./12.*m11*pow(m33,2) - 1./48.*pow(m11,2)*m44 + 1./48.*pow(
         m11,2)*m33 );

      M13p6x21R +=  + hh21(3,m14,m11,m34,m13,mu2) * ( 3./16.*pow(m33,2)
         *m44 + 3./8.*m11*m33*m44 - 3./16.*m11*pow(m33,2) + 3./16.*pow(
         m11,2)*m44 - 3./8.*pow(m11,2)*m33 - 3./16.*pow(m11,3) );

      M13p6x21R +=  + hh21(3,m34,m33,m14,m13,mu2) * ( 3./16.*pow(m33,2)
         *m44 - 3./16.*pow(m33,3) + 3./8.*m11*m33*m44 - 3./8.*m11*pow(
         m33,2) + 3./16.*pow(m11,2)*m44 - 3./16.*pow(m11,2)*m33 );

      M13p6x21R +=  + hh21(7,m13,m11,m11,m13,mu2) * ( 1./96.*pow(m33,2)
         *pow(m44,2) + 1./48.*m11*m33*pow(m44,2) - 1./48.*m11*pow(
         m33,2)*m44 + 1./96.*pow(m11,2)*pow(m44,2) - 1./24.*pow(m11,2)*
         m33*m44 + 1./96.*pow(m11,2)*pow(m33,2) - 1./48.*pow(m11,3)*m44
          + 1./48.*pow(m11,3)*m33 + 1./96.*pow(m11,4) );

      M13p6x21R +=  + hh21(7,m13,m11,m33,m13,mu2) * ( 1./48.*pow(m33,2)
         *pow(m44,2) - 1./48.*pow(m33,3)*m44 + 1./24.*m11*m33*pow(
         m44,2) - 1./16.*m11*pow(m33,2)*m44 + 1./48.*m11*pow(m33,3) + 1.
         /48.*pow(m11,2)*pow(m44,2) - 1./16.*pow(m11,2)*m33*m44 + 1./24.
         *pow(m11,2)*pow(m33,2) - 1./48.*pow(m11,3)*m44 + 1./48.*pow(
         m11,3)*m33 );

      M13p6x21R +=  + hh21(7,m13,m33,m33,m13,mu2) * ( 1./96.*pow(m33,2)
         *pow(m44,2) - 1./48.*pow(m33,3)*m44 + 1./96.*pow(m33,4) + 1./
         48.*m11*m33*pow(m44,2) - 1./24.*m11*pow(m33,2)*m44 + 1./48.*
         m11*pow(m33,3) + 1./96.*pow(m11,2)*pow(m44,2) - 1./48.*pow(
         m11,2)*m33*m44 + 1./96.*pow(m11,2)*pow(m33,2) );

  return M13p6x21R/pow(f0,4);
}

// 1 valence mass 2 sea mass *****************************************

double mv1s2nf3p4(const quarkmassnf mass, const Linf Liin){
  return mv1s2nf3p4L(mass,Liin)+mv1s2nf3p4R(mass);
}

double mv1s2nf3p4L(const quarkmassnf mass, const Linf Liin){
  double L0r,L1r,L2r,L3r,L4r,L5r,L6r,L7r,L8r,L9r,L10r,L11r,H1r,H2r,mu;
  int nfLi;
  Liin.out(L0r,L1r,L2r,L3r,L4r,L5r,L6r,L7r,L8r,L9r,L10r,L11r,H1r,H2r,mu,nfLi);
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if (mu0 != mu){
    cout << "ERROR: differing mu's in mv1s2nf3p4L\n";}
  if ( (nq != 3) || (B0mq.size() != 3)){
    cout << "ERROR: mv1s2nf3p4L probably called with wrong number"
	 << "of quark masses\n";}
  // three masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m44 = 2.*B0mq[1];
  double m66 = 2.*B0mq[2];
  double M12p4x12L =
       + 16*m11*m66*L6r - 8*m11*m66*L4r + 32*m11*m44*L6r - 16*m11*m44*
         L4r + 16*pow(m11,2)*L8r - 8*pow(m11,2)*L5r;

  return M12p4x12L/pow(f0,2);
}

double mv1s2nf3p4R(const quarkmassnf mass){
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if ( (nq != 3) || (B0mq.size() != 3)){
    cout << "ERROR: mv1s2nf3p4R probably called with wrong number"
	 << "of quark masses\n";}
  // three masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m44 = 2.*B0mq[1];
  double m66 = 2.*B0mq[2];
  double me2 = (m44+2.*m66)/3.;
  double mu2 = mu0*mu0;
  double RC146e = RC(m11,m44,m66,me2);
  double RSe6611 = pow(RS(me2,m66,m11),2);
  double RD146e = RD(m11,m44,m66,me2);
  double M12p4x12R =
       + Ab(m11,mu2) * (  - 1./3.*m11*RC146e );

      M12p4x12R +=  + Ab(me2,mu2) * ( 2./3.*m11*RSe6611 );

      M12p4x12R +=  + Bb(m11,mu2) * (  - 1./3.*m11*RD146e );

  return M12p4x12R/pow(f0,2);
}

double mv1s2nf3p6(const quarkmassnf mass, const Linf Liin, const Ki Kiin){
  return mv1s2nf3p6L(mass,Liin)+mv1s2nf3p6K(mass,Kiin)+mv1s2nf3p6R(mass);
}

double mv1s2nf3p6L(const quarkmassnf mass, const Linf Liin){
  double L0r,L1r,L2r,L3r,L4r,L5r,L6r,L7r,L8r,L9r,L10r,L11r,H1r,H2r,mu;
  int nfLi;
  Liin.out(L0r,L1r,L2r,L3r,L4r,L5r,L6r,L7r,L8r,L9r,L10r,L11r,H1r,H2r,mu,nfLi);
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if (mu0 != mu){
    cout << "ERROR: differing mu's in mv1s2nf3p6L\n";}
  if ( (nq != 3) || (B0mq.size() != 3)){
    cout << "ERROR: mv1s2nf3p6L probably called with wrong number"
	 << "of quark masses\n";}
  // three masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m44 = 2.*B0mq[1];
  double m66 = 2.*B0mq[2];
  double m14 = (m11+m44)/2.;
  double m16 = (m11+m66)/2.;
  double m46 = (m44+m66)/2.;
  double me2 = (m44+2.*m66)/3.;
  double mu2 = mu0*mu0;
  double RC146e = RC(m11,m44,m66,me2);
  double RSe61 = RS(me2,m66,m11);
  double RSe6611 = pow(RSe61,2);
  double RS14e = RS(m11,m44,me2);
  double RS16e = RS(m11,m66,me2);
  double RS146ee = RS14e*RS16e;
  double RD146e = RD(m11,m44,m66,me2);
  double RN1e = RN(m11,me2);

  double    M12p6x12L =
       + pi16 * ( 8./3.*m11*pow(me2,2)*L3r*RSe6611 + 2*m11*pow(me2,2)*
         L2r + 8./3.*m11*pow(me2,2)*L0r*RSe6611 + 1./2.*m11*pow(m66,2)*
         L3r + 2*m11*pow(m66,2)*L2r + m11*pow(m66,2)*L0r + 4*m11*m44*
         m66*L2r + m11*pow(m44,2)*L3r + 8*m11*pow(m44,2)*L2r + 2*m11*
         pow(m44,2)*L0r - 8./3.*pow(m11,2)*L3r*RD146e - 8./3.*pow(
         m11,2)*L0r*RD146e + pow(m11,2)*m66*L3r + 2*pow(m11,2)*m66*L0r
          + 2*pow(m11,2)*m44*L3r + 4*pow(m11,2)*m44*L0r + 3./2.*pow(
         m11,3)*L3r - 4./3.*pow(m11,3)*L3r*RC146e + 2*pow(m11,3)*L2r + 
         4*pow(m11,3)*L1r + 3*pow(m11,3)*L0r - 4./3.*pow(m11,3)*L0r*
         RC146e );

      M12p6x12L +=  + Ab(m11,mu2) * ( 8./3.*m11*L5r*RD146e - 8*m11*L3r*
         RD146e - 8*m11*L0r*RD146e - 32./3.*m11*m66*L7r*RS14e - 16./3.*
         m11*m66*L6r*RC146e + 32./9.*m11*m66*L5r*RS14e*RSe61 - 8./9.*
         m11*m66*L5r*pow(RS14e,2) - 32./3.*m11*m66*L4r*RSe6611 - 8./3.*
         m11*m66*L4r*RS146ee + 8./3.*m11*m66*L4r*RS16e - 32./9.*m11*m66
         *L4r*RS16e*RSe61 - 16./9.*m11*m66*L4r*pow(RS16e,2) + 8./3.*m11
         *m66*L4r*RS14e + 32./9.*m11*m66*L4r*RS14e*RSe61 - 8./9.*m11*
         m66*L4r*pow(RS14e,2) + 16./3.*m11*m66*L4r*RC146e + 32./9.*m11*
         m66*me2*L5r*RN1e*RS14e*RSe61 - 32./3.*m11*m66*me2*L4r*RN1e*
         RSe6611 - 32./9.*m11*m66*me2*L4r*RN1e*RS16e*RSe61 + 32./9.*m11
         *m66*me2*L4r*RN1e*RS14e*RSe61 - 64./9.*m11*pow(m66,2)*L8r*RN1e
         *RS14e*RSe61 - 64./9.*m11*pow(m66,2)*L7r*RN1e*RS14e*RSe61 - 64.
         /9.*m11*pow(m66,2)*L6r*RN1e*RS14e*RSe61 - 64./3.*m11*m44*L7r*
         RS16e - 32./3.*m11*m44*L6r*RC146e - 32./9.*m11*m44*L5r*RS16e*
         RSe61 - 16./9.*m11*m44*L5r*pow(RS16e,2) - 64./3.*m11*m44*L4r*
         RSe6611 );

      M12p6x12L +=  + Ab(m11,mu2) * (  - 16./3.*m11*m44*L4r*RS146ee + 
         16./3.*m11*m44*L4r*RS16e - 64./9.*m11*m44*L4r*RS16e*RSe61 - 32.
         /9.*m11*m44*L4r*pow(RS16e,2) + 16./3.*m11*m44*L4r*RS14e + 64./
         9.*m11*m44*L4r*RS14e*RSe61 - 16./9.*m11*m44*L4r*pow(RS14e,2)
          + 32./3.*m11*m44*L4r*RC146e - 32./9.*m11*m44*me2*L5r*RN1e*
         RS16e*RSe61 - 64./3.*m11*m44*me2*L4r*RN1e*RSe6611 - 64./9.*m11
         *m44*me2*L4r*RN1e*RS16e*RSe61 + 64./9.*m11*m44*me2*L4r*RN1e*
         RS14e*RSe61 - 128./9.*m11*m44*m66*L7r*RN1e*RS16e*RSe61 + 64./9.
         *m11*m44*m66*L7r*RN1e*RS14e*RSe61 + 64./9.*m11*m44*m66*L6r*
         RN1e*RS16e*RSe61 - 128./9.*m11*m44*m66*L6r*RN1e*RS14e*RSe61 + 
         64./9.*m11*pow(m44,2)*L8r*RN1e*RS16e*RSe61 + 128./9.*m11*pow(
         m44,2)*L7r*RN1e*RS16e*RSe61 + 128./9.*m11*pow(m44,2)*L6r*RN1e*
         RS16e*RSe61 - 64./3.*pow(m11,2)*L8r*RC146e + 32*pow(m11,2)*L7r
          + 32*pow(m11,2)*L6r - 32./3.*pow(m11,2)*L5r*RSe6611 - 8./3.*
         pow(m11,2)*L5r*RS146ee + 8./3.*pow(m11,2)*L5r*RS16e + 8./3.*
         pow(m11,2)*L5r*RS14e );

      M12p6x12L +=  + Ab(m11,mu2) * ( 32./3.*pow(m11,2)*L5r*RC146e - 8*
         pow(m11,2)*L3r*RC146e + 20*pow(m11,2)*L2r + 8*pow(m11,2)*L1r
          - 8*pow(m11,2)*L0r*RC146e - 32./3.*pow(m11,2)*me2*L5r*RN1e*
         RSe6611 + 64./3.*pow(m11,2)*m66*L7r*RN1e*RSe61 + 64./3.*pow(
         m11,2)*m66*L6r*RN1e*RSe6611 - 64./3.*pow(m11,2)*m44*L7r*RN1e*
         RSe61 + 128./3.*pow(m11,2)*m44*L6r*RN1e*RSe6611 + 64./3.*pow(
         m11,3)*L8r*RN1e*RSe6611 );

      M12p6x12L +=  + Ab(m14,mu2) * ( 32*m11*m44*L8r - 16*m11*m44*L5r
          + 20*m11*m44*L3r + 8*m11*m44*L0r + 32*pow(m11,2)*L8r - 16*
         pow(m11,2)*L5r + 20*pow(m11,2)*L3r + 8*pow(m11,2)*L0r );

      M12p6x12L +=  + Ab(m16,mu2) * ( 16*m11*m66*L8r - 8*m11*m66*L5r + 
         10*m11*m66*L3r + 4*m11*m66*L0r + 16*pow(m11,2)*L8r - 8*pow(
         m11,2)*L5r + 10*pow(m11,2)*L3r + 4*pow(m11,2)*L0r );

      M12p6x12L +=  + Ab(m44,mu2) * ( 48*m11*m44*L6r - 48*m11*m44*L4r
          + 12*m11*m44*L2r + 48*m11*m44*L1r );

      M12p6x12L +=  + Ab(m46,mu2) * ( 32*m11*m66*L6r - 32*m11*m66*L4r
          + 8*m11*m66*L2r + 32*m11*m66*L1r + 32*m11*m44*L6r - 32*m11*
         m44*L4r + 8*m11*m44*L2r + 32*m11*m44*L1r );

      M12p6x12L +=  + Ab(me2,mu2) * (  - 16./3.*m11*me2*L5r*RSe6611 - 8
         *m11*me2*L4r + 16*m11*me2*L3r*RSe6611 + 4*m11*me2*L2r + 16*m11
         *me2*L1r + 16*m11*me2*L0r*RSe6611 + 64./3.*m11*m66*L7r*RSe61
          + 32./3.*m11*m66*L6r + 32./3.*m11*m66*L6r*RSe6611 - 32./9.*
         m11*m66*L5r*pow(RSe61,2) - 16./3.*m11*m66*L4r - 16./3.*m11*m66
         *L4r*pow(RSe61,2) - 16./3.*m11*m66*L4r*RSe6611 - 32./9.*m11*
         m66*me2*L5r*RN1e*RS14e*RSe61 + 32./3.*m11*m66*me2*L4r*RN1e*
         RSe6611 + 32./9.*m11*m66*me2*L4r*RN1e*RS16e*RSe61 - 32./9.*m11
         *m66*me2*L4r*RN1e*RS14e*RSe61 + 64./9.*m11*pow(m66,2)*L8r*RN1e
         *RS14e*RSe61 + 64./9.*m11*pow(m66,2)*L7r*RN1e*RS14e*RSe61 + 64.
         /9.*m11*pow(m66,2)*L6r*RN1e*RS14e*RSe61 - 64./3.*m11*m44*L7r*
         RSe61 + 16./3.*m11*m44*L6r + 64./3.*m11*m44*L6r*RSe6611 - 16./
         9.*m11*m44*L5r*pow(RSe61,2) - 8./3.*m11*m44*L4r - 32./3.*m11*
         m44*L4r*pow(RSe61,2) - 32./3.*m11*m44*L4r*RSe6611 + 32./9.*m11
         *m44*me2*L5r*RN1e*RS16e*RSe61 + 64./3.*m11*m44*me2*L4r*RN1e*
         RSe6611 );

      M12p6x12L +=  + Ab(me2,mu2) * ( 64./9.*m11*m44*me2*L4r*RN1e*RS16e
         *RSe61 - 64./9.*m11*m44*me2*L4r*RN1e*RS14e*RSe61 + 128./9.*m11
         *m44*m66*L7r*RN1e*RS16e*RSe61 - 64./9.*m11*m44*m66*L7r*RN1e*
         RS14e*RSe61 - 64./9.*m11*m44*m66*L6r*RN1e*RS16e*RSe61 + 128./9.
         *m11*m44*m66*L6r*RN1e*RS14e*RSe61 - 64./9.*m11*pow(m44,2)*L8r*
         RN1e*RS16e*RSe61 - 128./9.*m11*pow(m44,2)*L7r*RN1e*RS16e*RSe61
          - 128./9.*m11*pow(m44,2)*L6r*RN1e*RS16e*RSe61 + 128./3.*pow(
         m11,2)*L8r*RSe6611 - 32./3.*pow(m11,2)*L5r*RSe6611 + 32./3.*
         pow(m11,2)*me2*L5r*RN1e*RSe6611 - 64./3.*pow(m11,2)*m66*L7r*
         RN1e*RSe61 - 64./3.*pow(m11,2)*m66*L6r*RN1e*RSe6611 + 64./3.*
         pow(m11,2)*m44*L7r*RN1e*RSe61 - 128./3.*pow(m11,2)*m44*L6r*
         RN1e*RSe6611 - 64./3.*pow(m11,3)*L8r*RN1e*RSe6611 );

      M12p6x12L +=  + Bb(m11,mu2) * (  - 16./3.*m11*m66*L6r*RD146e + 8*
         m11*m66*L4r*RD146e + 16./9.*m11*pow(m66,2)*L8r*pow(RS14e,2) + 
         16./9.*m11*pow(m66,2)*L7r*pow(RS14e,2) + 16./9.*m11*pow(m66,2)
         *L6r*pow(RS14e,2) - 32./3.*m11*m44*L6r*RD146e + 16*m11*m44*L4r
         *RD146e + 64./9.*m11*m44*m66*L7r*RS14e*RS16e + 32./9.*m11*m44*
         m66*L6r*pow(RS16e,2) + 32./9.*m11*m44*m66*L6r*pow(RS14e,2) + 
         32./9.*m11*pow(m44,2)*L8r*pow(RS16e,2) + 64./9.*m11*pow(m44,2)
         *L7r*pow(RS16e,2) + 64./9.*m11*pow(m44,2)*L6r*pow(RS16e,2) - 
         64./3.*pow(m11,2)*L8r*RD146e + 40./3.*pow(m11,2)*L5r*RD146e - 
         8*pow(m11,2)*L3r*RD146e - 8*pow(m11,2)*L0r*RD146e - 32./3.*
         pow(m11,2)*m66*L7r*RS14e - 32./3.*pow(m11,2)*m66*L6r*RC146e - 
         8./9.*pow(m11,2)*m66*L5r*pow(RS14e,2) - 16./9.*pow(m11,2)*m66*
         L4r*pow(RS16e,2) - 8./9.*pow(m11,2)*m66*L4r*pow(RS14e,2) + 16./
         3.*pow(m11,2)*m66*L4r*RC146e - 64./3.*pow(m11,2)*m44*L7r*RS16e
          - 64./3.*pow(m11,2)*m44*L6r*RC146e - 16./9.*pow(m11,2)*m44*
         L5r*pow(RS16e,2) );

      M12p6x12L +=  + Bb(m11,mu2) * (  - 32./9.*pow(m11,2)*m44*L4r*pow(
         RS16e,2) - 16./9.*pow(m11,2)*m44*L4r*pow(RS14e,2) + 32./3.*
         pow(m11,2)*m44*L4r*RC146e - 32./3.*pow(m11,3)*L8r*RC146e + 16*
         pow(m11,3)*L7r + 16./3.*pow(m11,3)*L5r*RC146e );

      M12p6x12L +=  + Bb(me2,mu2) * (  - 32./9.*m11*m66*me2*L5r*pow(
         RSe61,2) - 16./3.*m11*m66*me2*L4r*pow(RSe61,2) + 64./9.*m11*
         pow(m66,2)*L8r*pow(RSe61,2) + 64./9.*m11*pow(m66,2)*L7r*pow(
         RSe61,2) + 64./9.*m11*pow(m66,2)*L6r*pow(RSe61,2) - 16./9.*m11
         *m44*me2*L5r*pow(RSe61,2) - 32./3.*m11*m44*me2*L4r*pow(
         RSe61,2) - 128./9.*m11*m44*m66*L7r*pow(RSe61,2) + 160./9.*m11*
         m44*m66*L6r*pow(RSe61,2) + 32./9.*m11*pow(m44,2)*L8r*pow(
         RSe61,2) + 64./9.*m11*pow(m44,2)*L7r*pow(RSe61,2) + 64./9.*m11
         *pow(m44,2)*L6r*pow(RSe61,2) );

      M12p6x12L +=  + Cb(m11,mu2) * (  - 32./3.*pow(m11,2)*m66*L6r*
         RD146e + 16./3.*pow(m11,2)*m66*L4r*RD146e - 64./3.*pow(m11,2)*
         m44*L6r*RD146e + 32./3.*pow(m11,2)*m44*L4r*RD146e - 32./3.*
         pow(m11,3)*L8r*RD146e + 16./3.*pow(m11,3)*L5r*RD146e );

      M12p6x12L +=  - 128*m11*pow(m66,2)*L4r*L6r + 64*m11*pow(m66,2)*
         pow(L4r,2) - 512*m11*m44*m66*L4r*L6r + 256*m11*m44*m66*pow(
         L4r,2) - 512*m11*pow(m44,2)*L4r*L6r + 256*m11*pow(m44,2)*pow(
         L4r,2) - 128*pow(m11,2)*m66*L5r*L6r - 128*pow(m11,2)*m66*L4r*
         L8r + 128*pow(m11,2)*m66*L4r*L5r - 256*pow(m11,2)*m44*L5r*L6r
          - 256*pow(m11,2)*m44*L4r*L8r + 256*pow(m11,2)*m44*L4r*L5r - 
         128*pow(m11,3)*L5r*L8r + 64*pow(m11,3)*pow(L5r,2);

  return M12p6x12L/pow(f0,4);
}

double mv1s2nf3p6K(const quarkmassnf mass, const Ki Kiin){
  double KK[116],mu;
  Kiin.out(KK,mu);
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if (mu0 != mu){
    cout << "ERROR: differing mu's in mv1s2nf3p6K\n";}
  if ( (nq != 3) || (B0mq.size() != 3)){
    cout << "ERROR: mv1s2nf3p6K probably called with wrong number"
	 << "of quark masses\n";}
  // three masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m44 = 2.*B0mq[1];
  double m66 = 2.*B0mq[2];
  double M12p6x12K =
       - 32*KK[17]*pow(m11,3) - 32*KK[18]*pow(m11,2)*m66 - 64*KK[18]*
         pow(m11,2)*m44 - 16*KK[19]*pow(m11,3) - 16*KK[20]*pow(m11,2)*
         m66 - 32*KK[20]*pow(m11,2)*m44 - 16*KK[21]*m11*pow(m66,2) - 32
         *KK[21]*m11*pow(m44,2) - 16*KK[22]*m11*pow(m66,2) - 64*KK[22]*
         m11*m44*m66 - 64*KK[22]*m11*pow(m44,2) - 16*KK[23]*pow(m11,3)
          + 48*KK[25]*pow(m11,3) + 16*KK[26]*m11*pow(m66,2) + 32*KK[26]
         *m11*pow(m44,2) + 32*KK[26]*pow(m11,2)*m66 + 64*KK[26]*pow(
         m11,2)*m44 + 48*KK[27]*m11*pow(m66,2) + 192*KK[27]*m11*m44*m66
          + 192*KK[27]*m11*pow(m44,2) + 32*KK[39]*pow(m11,3) + 32*KK[40]
         *pow(m11,2)*m66 + 64*KK[40]*pow(m11,2)*m44;

  return M12p6x12K/pow(f0,4);
}

double mv1s2nf3p6R(const quarkmassnf mass){
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if ( (nq != 3) || (B0mq.size() != 3)){
    cout << "ERROR: mv1s2nf3p6R probably called with wrong number"
	 << "of quark masses\n";}
  // three masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m44 = 2.*B0mq[1];
  double m66 = 2.*B0mq[2];
  double m14 = (m11+m44)/2.;
  double m16 = (m11+m66)/2.;
  double m46 = (m44+m66)/2.;
  double me2 = (m44+2.*m66)/3.;
  double mu2 = mu0*mu0;
  double RC146e = RC(m11,m44,m66,me2);
  double RSe61 = RS(me2,m66,m11);
  double RSe6611 = pow(RSe61,2);
  double RS14e = RS(m11,m44,me2);
  double RS16e = RS(m11,m66,me2);
  double RS146ee = RS14e*RS16e;
  double RD146e = RD(m11,m44,m66,me2);
  double RN1e = RN(m11,me2);


  double  M12p6x12R =
       + pi162 * (  - 1./45.*m66*pow(me2,2) + 2./45.*m66*pow(me2,2)*
         RSe61 - 1./45.*m66*pow(me2,2)*RSe6611 + 1./45.*pow(m66,2)*me2
          - 1./45.*pow(m66,2)*me2*RSe61 - 1./90.*m44*pow(me2,2) - 2./45.
         *m44*pow(me2,2)*RSe61 - 2./45.*m44*pow(me2,2)*RSe6611 + 1./90.
         *pow(m44,2)*me2 + 1./45.*pow(m44,2)*me2*RSe61 - 2./45.*m16*
         pow(me2,2) + 4./45.*m16*pow(me2,2)*RSe61 - 2./45.*m16*pow(
         me2,2)*RSe6611 - 1./54.*m16*m66*RD146e + 11./135.*m16*m66*me2
          - 16./135.*m16*m66*me2*RSe61 + 1./27.*m16*m66*me2*RSe6611 - 1.
         /27.*m16*pow(m66,2) + 1./27.*m16*pow(m66,2)*RSe6611 - 1./54.*
         m16*pow(m66,2)*RC146e - 1./45.*m14*pow(me2,2) - 4./45.*m14*
         pow(me2,2)*RSe61 - 4./45.*m14*pow(me2,2)*RSe6611 - 1./27.*m14*
         m44*RD146e + 11./270.*m14*m44*me2 + 16./135.*m14*m44*me2*RSe61
          + 2./27.*m14*m44*me2*RSe6611 - 1./54.*m14*pow(m44,2) + 2./27.
         *m14*pow(m44,2)*RSe6611 - 1./27.*m14*pow(m44,2)*RC146e - 1./9.
         *m11*me2*RD146e*RSe6611 - 61./480.*m11*pow(me2,2) - 61./240.*
         m11*pow(me2,2)*RSe6611 );

      M12p6x12R +=  + pi162 * ( 2./9.*m11*pow(me2,2)*pow(RSe6611,2) - 
         29./1440.*m11*m66*RD146e + 103./2160.*m11*m66*me2 - 7./120.*
         m11*m66*me2*RSe61 + 13./27.*m11*m66*me2*pow(RSe61,2) - 137./
         2160.*m11*m66*me2*RSe6611 - 4./27.*m11*m66*me2*RS14e*RSe61 + 1.
         /27.*m11*m66*me2*pow(RS14e,2) + 5./48.*m11*pow(m66,2) - 1./48.
         *m11*pow(m66,2)*RSe61 + 1./96.*m11*pow(m66,2)*RSe6611 + 31./
         1440.*m11*pow(m66,2)*RS14e - 1./192.*m11*pow(m66,2)*RC146e - 2.
         /3.*m11*m46*me2*pow(RSe61,2) - 2./9.*m11*m46*me2*RS16e*RSe61
          + 2./9.*m11*m46*me2*RS14e*RSe61 - 13./36.*m11*m46*m66 - 4./27.
         *m11*m46*m66*RS16e*RSe61 - 2./27.*m11*m46*m66*RS14e*RSe61 + 2./
         27.*m11*m46*m66*RS14e*RS16e + 1./27.*m11*m46*m66*pow(RS14e,2)
          - 29./720.*m11*m44*RD146e + 103./4320.*m11*m44*me2 + 7./120.*
         m11*m44*me2*RSe61 + 19./54.*m11*m44*me2*pow(RSe61,2) - 137./
         1080.*m11*m44*me2*RSe6611 + 1./27.*m11*m44*me2*RS16e*RSe61 + 1.
         /54.*m11*m44*me2*pow(RS16e,2) + 5./16.*m11*m44*m66 - 13./36.*
         m11*m44*m46 );

      M12p6x12R +=  + pi162 * (  - 1./9.*m11*m44*m46*pow(RSe61,2) - 2./
         27.*m11*m44*m46*RS16e*RSe61 + 1./27.*m11*m44*m46*pow(RS16e,2)
          + 2./27.*m11*m44*m46*RS14e*RSe61 + 2./27.*m11*m44*m46*RS14e*
         RS16e - 5./96.*m11*pow(m44,2) + 1./48.*m11*pow(m44,2)*RSe61 + 
         1./6.*m11*pow(m44,2)*pow(RSe61,2) + 1./48.*m11*pow(m44,2)*
         RSe6611 + 31./720.*m11*pow(m44,2)*RS16e + 1./3.*m11*pow(m44,2)
         *RS16e*RSe61 + 1./6.*m11*pow(m44,2)*pow(RS16e,2) - 1./96.*m11*
         pow(m44,2)*RC146e + 23./135.*m11*m16*RD146e + 1./27.*m11*m16*
         me2 + 44./135.*m11*m16*me2*RSe61 - 34./135.*m11*m16*me2*
         RSe6611 - 41./216.*m11*m16*m66 + 11./54.*m11*m16*m66*RSe61 + 
         13./108.*m11*m16*m66*RSe6611 - 23./540.*m11*m16*m66*RS14e - 17.
         /216.*m11*m16*m66*RC146e - 31./108.*m11*m16*m44 + 46./135.*m11
         *m14*RD146e + 1./54.*m11*m14*me2 - 44./135.*m11*m14*me2*RSe61
          - 68./135.*m11*m14*me2*RSe6611 - 31./108.*m11*m14*m66 - 11./
         18.*m11*m14*m44 - 11./54.*m11*m14*m44*RSe61 + 13./54.*m11*m14*
         m44*RSe6611 );

      M12p6x12R +=  + pi162 * (  - 23./270.*m11*m14*m44*RS16e - 17./108.
         *m11*m14*m44*RC146e + 7./160.*pow(m11,2)*RD146e + 1./18.*pow(
         m11,2)*RD146e*RC146e + 5./96.*pow(m11,2)*me2 - 197./720.*pow(
         m11,2)*me2*RSe6611 + 4./9.*pow(m11,2)*me2*pow(RSe6611,2) - 4./
         9.*pow(m11,2)*me2*RC146e*RSe6611 + 17./216.*pow(m11,2)*m66 + 5.
         /54.*pow(m11,2)*m66*RSe61 + 25./108.*pow(m11,2)*m66*RSe6611 + 
         1./18.*pow(m11,2)*m66*RS146ee - 1./18.*pow(m11,2)*m66*RS16e + 
         1./27.*pow(m11,2)*m66*pow(RS16e,2) - 41./432.*pow(m11,2)*m66*
         RS14e - 1./3.*pow(m11,2)*m66*RS14e*RSe61 + 1./27.*pow(m11,2)*
         m66*RS14e*RS16e + 5./54.*pow(m11,2)*m66*pow(RS14e,2) - 5./288.
         *pow(m11,2)*m66*RC146e - 2./9.*pow(m11,2)*m46*RS16e*RSe61 - 2./
         27.*pow(m11,2)*m46*pow(RS16e,2) + 2./9.*pow(m11,2)*m46*RS14e*
         RSe61 + 4./27.*pow(m11,2)*m46*RS14e*RS16e - 2./27.*pow(m11,2)*
         m46*pow(RS14e,2) + 181./864.*pow(m11,2)*m44 - 5./54.*pow(
         m11,2)*m44*RSe61 + 25./54.*pow(m11,2)*m44*RSe6611 + 1./9.*pow(
         m11,2)*m44*RS146ee );

      M12p6x12R +=  + pi162 * (  - 41./216.*pow(m11,2)*m44*RS16e + 4./9.
         *pow(m11,2)*m44*RS16e*RSe61 + 7./27.*pow(m11,2)*m44*pow(
         RS16e,2) - 1./9.*pow(m11,2)*m44*RS14e - 1./9.*pow(m11,2)*m44*
         RS14e*RSe61 + 1./27.*pow(m11,2)*m44*RS14e*RS16e + 1./27.*pow(
         m11,2)*m44*pow(RS14e,2) - 5./144.*pow(m11,2)*m44*RC146e + 29./
         162.*pow(m11,2)*m44*pow(m66,3)*pow(RN1e,3) + 29./162.*pow(
         m11,2)*pow(m44,2)*pow(m66,2)*pow(RN1e,3) + 29./648.*pow(m11,2)
         *pow(m44,3)*m66*pow(RN1e,3) - 59./216.*pow(m11,2)*m16 + 23./54.
         *pow(m11,2)*m16*RSe61 + 1./12.*pow(m11,2)*m16*RSe6611 - 227./
         540.*pow(m11,2)*m16*RS14e + 23./216.*pow(m11,2)*m16*RC146e - 
         59./108.*pow(m11,2)*m14 - 23./54.*pow(m11,2)*m14*RSe61 + 1./6.
         *pow(m11,2)*m14*RSe6611 - 227./270.*pow(m11,2)*m14*RS16e + 23./
         108.*pow(m11,2)*m14*RC146e - 53./96.*pow(m11,3) + 97./144.*
         pow(m11,3)*RSe6611 + 1./6.*pow(m11,3)*RS146ee - 229./1080.*
         pow(m11,3)*RS16e - 409./2160.*pow(m11,3)*RS14e + 3./32.*pow(
         m11,3)*RC146e );

      M12p6x12R +=  + pi162 * (  - 2./9.*pow(m11,3)*RC146e*RSe6611 + 1./
         6.*pow(m11,3)*pow(RC146e,2) - 29./162.*pow(m11,3)*pow(m66,3)*
         pow(RN1e,3) - 145./162.*pow(m11,3)*m44*pow(m66,2)*pow(RN1e,3)
          - 319./648.*pow(m11,3)*pow(m44,2)*m66*pow(RN1e,3) - 29./648.*
         pow(m11,3)*pow(m44,3)*pow(RN1e,3) + 58./81.*pow(m11,4)*pow(
         m66,2)*pow(RN1e,3) + 899./648.*pow(m11,4)*m44*m66*pow(RN1e,3)
          + 203./648.*pow(m11,4)*pow(m44,2)*pow(RN1e,3) - 203./216.*
         pow(m11,5)*m66*pow(RN1e,3) - 145./216.*pow(m11,5)*m44*pow(
         RN1e,3) + 29./72.*pow(m11,6)*pow(RN1e,3) - 1./270.*pi2*m66*
         pow(me2,2) + 1./135.*pi2*m66*pow(me2,2)*RSe61 - 1./270.*pi2*
         m66*pow(me2,2)*RSe6611 + 1./270.*pi2*pow(m66,2)*me2 - 1./270.*
         pi2*pow(m66,2)*me2*RSe61 - 1./540.*pi2*m44*pow(me2,2) - 1./135.
         *pi2*m44*pow(me2,2)*RSe61 - 1./135.*pi2*m44*pow(me2,2)*RSe6611
          + 1./540.*pi2*pow(m44,2)*me2 + 1./270.*pi2*pow(m44,2)*me2*
         RSe61 - 1./135.*pi2*m16*pow(me2,2) + 2./135.*pi2*m16*pow(
         me2,2)*RSe61 );

      M12p6x12R +=  + pi162 * (  - 1./135.*pi2*m16*pow(me2,2)*RSe6611
          - 1./324.*pi2*m16*m66*RD146e + 11./810.*pi2*m16*m66*me2 - 8./
         405.*pi2*m16*m66*me2*RSe61 + 1./162.*pi2*m16*m66*me2*RSe6611
          - 1./162.*pi2*m16*pow(m66,2) + 1./162.*pi2*m16*pow(m66,2)*
         RSe6611 - 1./324.*pi2*m16*pow(m66,2)*RC146e - 1./270.*pi2*m14*
         pow(me2,2) - 2./135.*pi2*m14*pow(me2,2)*RSe61 - 2./135.*pi2*
         m14*pow(me2,2)*RSe6611 - 1./162.*pi2*m14*m44*RD146e + 11./1620.
         *pi2*m14*m44*me2 + 8./405.*pi2*m14*m44*me2*RSe61 + 1./81.*pi2*
         m14*m44*me2*RSe6611 - 1./324.*pi2*m14*pow(m44,2) + 1./81.*pi2*
         m14*pow(m44,2)*RSe6611 - 1./162.*pi2*m14*pow(m44,2)*RC146e - 1.
         /54.*pi2*m11*me2*RD146e*RSe6611 - 19./720.*pi2*m11*pow(me2,2)
          - 19./360.*pi2*m11*pow(me2,2)*RSe6611 + 1./27.*pi2*m11*pow(
         me2,2)*pow(RSe6611,2) + 1./540.*pi2*m11*m66*RD146e - 1./405.*
         pi2*m11*m66*me2 + 1./90.*pi2*m11*m66*me2*RSe61 + 13./162.*pi2*
         m11*m66*me2*pow(RSe61,2) - 17./810.*pi2*m11*m66*me2*RSe6611 - 
         2./81.*pi2*m11*m66*me2*RS14e*RSe61 );

      M12p6x12R +=  + pi162 * ( 1./162.*pi2*m11*m66*me2*pow(RS14e,2) + 
         1./540.*pi2*m11*pow(m66,2)*RS14e - 1./9.*pi2*m11*m46*me2*pow(
         RSe61,2) - 1./27.*pi2*m11*m46*me2*RS16e*RSe61 + 1./27.*pi2*m11
         *m46*me2*RS14e*RSe61 - 13./216.*pi2*m11*m46*m66 - 2./81.*pi2*
         m11*m46*m66*RS16e*RSe61 - 1./81.*pi2*m11*m46*m66*RS14e*RSe61
          + 1./81.*pi2*m11*m46*m66*RS14e*RS16e + 1./162.*pi2*m11*m46*
         m66*pow(RS14e,2) + 1./270.*pi2*m11*m44*RD146e - 1./810.*pi2*
         m11*m44*me2 - 1./90.*pi2*m11*m44*me2*RSe61 + 19./324.*pi2*m11*
         m44*me2*pow(RSe61,2) - 17./405.*pi2*m11*m44*me2*RSe6611 + 1./
         162.*pi2*m11*m44*me2*RS16e*RSe61 + 1./324.*pi2*m11*m44*me2*
         pow(RS16e,2) - 13./216.*pi2*m11*m44*m46 - 1./54.*pi2*m11*m44*
         m46*pow(RSe61,2) - 1./81.*pi2*m11*m44*m46*RS16e*RSe61 + 1./162.
         *pi2*m11*m44*m46*pow(RS16e,2) + 1./81.*pi2*m11*m44*m46*RS14e*
         RSe61 + 1./81.*pi2*m11*m44*m46*RS14e*RS16e - 13./144.*pi2*m11*
         pow(m44,2) + 1./36.*pi2*m11*pow(m44,2)*pow(RSe61,2) + 1./270.*
         pi2*m11*pow(m44,2)*RS16e );

      M12p6x12R +=  + pi162 * ( 1./18.*pi2*m11*pow(m44,2)*RS16e*RSe61
          + 1./36.*pi2*m11*pow(m44,2)*pow(RS16e,2) + 23./810.*pi2*m11*
         m16*RD146e + 1./162.*pi2*m11*m16*me2 + 22./405.*pi2*m11*m16*
         me2*RSe61 - 17./405.*pi2*m11*m16*me2*RSe6611 - 41./1296.*pi2*
         m11*m16*m66 + 11./324.*pi2*m11*m16*m66*RSe61 + 13./648.*pi2*
         m11*m16*m66*RSe6611 - 23./3240.*pi2*m11*m16*m66*RS14e - 17./
         1296.*pi2*m11*m16*m66*RC146e - 31./648.*pi2*m11*m16*m44 + 23./
         405.*pi2*m11*m14*RD146e + 1./324.*pi2*m11*m14*me2 - 22./405.*
         pi2*m11*m14*me2*RSe61 - 34./405.*pi2*m11*m14*me2*RSe6611 - 31./
         648.*pi2*m11*m14*m66 - 11./108.*pi2*m11*m14*m44 - 11./324.*pi2
         *m11*m14*m44*RSe61 + 13./324.*pi2*m11*m14*m44*RSe6611 - 23./
         1620.*pi2*m11*m14*m44*RS16e - 17./648.*pi2*m11*m14*m44*RC146e
          + 19./720.*pi2*pow(m11,2)*RD146e + 1./108.*pi2*pow(m11,2)*
         RD146e*RC146e - 17./270.*pi2*pow(m11,2)*me2*RSe6611 + 2./27.*
         pi2*pow(m11,2)*me2*pow(RSe6611,2) - 2./27.*pi2*pow(m11,2)*me2*
         RC146e*RSe6611 );

      M12p6x12R +=  + pi162 * ( 1./54.*pi2*pow(m11,2)*m66*RSe61 + 1./27.
         *pi2*pow(m11,2)*m66*RSe6611 + 1./108.*pi2*pow(m11,2)*m66*
         RS146ee - 1./108.*pi2*pow(m11,2)*m66*RS16e + 1./162.*pi2*pow(
         m11,2)*m66*pow(RS16e,2) - 1./36.*pi2*pow(m11,2)*m66*RS14e - 1./
         18.*pi2*pow(m11,2)*m66*RS14e*RSe61 + 1./162.*pi2*pow(m11,2)*
         m66*RS14e*RS16e + 5./324.*pi2*pow(m11,2)*m66*pow(RS14e,2) + 1./
         324.*pi2*pow(m11,2)*m66*RC146e - 1./27.*pi2*pow(m11,2)*m46*
         RS16e*RSe61 - 1./81.*pi2*pow(m11,2)*m46*pow(RS16e,2) + 1./27.*
         pi2*pow(m11,2)*m46*RS14e*RSe61 + 2./81.*pi2*pow(m11,2)*m46*
         RS14e*RS16e - 1./81.*pi2*pow(m11,2)*m46*pow(RS14e,2) - 1./54.*
         pi2*pow(m11,2)*m44*RSe61 + 2./27.*pi2*pow(m11,2)*m44*RSe6611
          + 1./54.*pi2*pow(m11,2)*m44*RS146ee - 1./18.*pi2*pow(m11,2)*
         m44*RS16e + 2./27.*pi2*pow(m11,2)*m44*RS16e*RSe61 + 7./162.*
         pi2*pow(m11,2)*m44*pow(RS16e,2) - 1./54.*pi2*pow(m11,2)*m44*
         RS14e - 1./54.*pi2*pow(m11,2)*m44*RS14e*RSe61 + 1./162.*pi2*
         pow(m11,2)*m44*RS14e*RS16e );

      M12p6x12R +=  + pi162 * ( 1./162.*pi2*pow(m11,2)*m44*pow(RS14e,2)
          + 1./162.*pi2*pow(m11,2)*m44*RC146e + 29./972.*pi2*pow(m11,2)
         *m44*pow(m66,3)*pow(RN1e,3) + 29./972.*pi2*pow(m11,2)*pow(
         m44,2)*pow(m66,2)*pow(RN1e,3) + 29./3888.*pi2*pow(m11,2)*pow(
         m44,3)*m66*pow(RN1e,3) - 59./1296.*pi2*pow(m11,2)*m16 + 23./
         324.*pi2*pow(m11,2)*m16*RSe61 + 1./72.*pi2*pow(m11,2)*m16*
         RSe6611 - 227./3240.*pi2*pow(m11,2)*m16*RS14e + 23./1296.*pi2*
         pow(m11,2)*m16*RC146e - 59./648.*pi2*pow(m11,2)*m14 - 23./324.
         *pi2*pow(m11,2)*m14*RSe61 + 1./36.*pi2*pow(m11,2)*m14*RSe6611
          - 227./1620.*pi2*pow(m11,2)*m14*RS16e + 23./648.*pi2*pow(
         m11,2)*m14*RC146e - 13./216.*pi2*pow(m11,3) + 1./9.*pi2*pow(
         m11,3)*RSe6611 + 1./36.*pi2*pow(m11,3)*RS146ee - 37./540.*pi2*
         pow(m11,3)*RS16e - 13./270.*pi2*pow(m11,3)*RS14e + 13./432.*
         pi2*pow(m11,3)*RC146e - 1./27.*pi2*pow(m11,3)*RC146e*RSe6611
          + 1./36.*pi2*pow(m11,3)*pow(RC146e,2) - 29./972.*pi2*pow(
         m11,3)*pow(m66,3)*pow(RN1e,3) );

      M12p6x12R +=  + pi162 * (  - 145./972.*pi2*pow(m11,3)*m44*pow(
         m66,2)*pow(RN1e,3) - 319./3888.*pi2*pow(m11,3)*pow(m44,2)*m66*
         pow(RN1e,3) - 29./3888.*pi2*pow(m11,3)*pow(m44,3)*pow(RN1e,3)
          + 29./243.*pi2*pow(m11,4)*pow(m66,2)*pow(RN1e,3) + 899./3888.
         *pi2*pow(m11,4)*m44*m66*pow(RN1e,3) + 203./3888.*pi2*pow(
         m11,4)*pow(m44,2)*pow(RN1e,3) - 203./1296.*pi2*pow(m11,5)*m66*
         pow(RN1e,3) - 145./1296.*pi2*pow(m11,5)*m44*pow(RN1e,3) + 29./
         432.*pi2*pow(m11,6)*pow(RN1e,3) );

      M12p6x12R +=  + Ab(m11,mu2)*pi16 * (  - 1./4.*pow(m11,2) + 1./6.*
         pow(m11,2)*RS16e + 1./12.*pow(m11,2)*RS14e );

      M12p6x12R +=  + pow(Ab(m11,mu2),2) * ( 1./90.*pow(m11,-1)*m66*
         RD146e + 1./90.*pow(m11,-1)*pow(m66,2)*RS14e + 1./45.*pow(
         m11,-1)*m44*RD146e + 1./45.*pow(m11,-1)*pow(m44,2)*RS16e + 19./
         120.*RD146e + 1./18.*RD146e*RC146e - 1./9.*me2*RC146e*RSe6611
          + 1./9.*m66*RSe61 + 2./9.*m66*RSe6611 + 1./18.*m66*RS146ee - 
         1./18.*m66*RS16e + 1./27.*m66*pow(RS16e,2) - 1./6.*m66*RS14e
          - 1./3.*m66*RS14e*RSe61 + 1./27.*m66*RS14e*RS16e + 5./54.*m66
         *pow(RS14e,2) + 1./54.*m66*RC146e - 1./9.*m44*RSe61 + 4./9.*
         m44*RSe6611 + 1./9.*m44*RS146ee - 1./3.*m44*RS16e + 4./9.*m44*
         RS16e*RSe61 + 7./27.*m44*pow(RS16e,2) - 1./9.*m44*RS14e - 1./9.
         *m44*RS14e*RSe61 + 1./27.*m44*RS14e*RS16e + 1./27.*m44*pow(
         RS14e,2) + 1./27.*m44*RC146e - 29./162.*m44*pow(m66,3)*pow(
         RN1e,3) - 29./162.*pow(m44,2)*pow(m66,2)*pow(RN1e,3) - 29./648.
         *pow(m44,3)*m66*pow(RN1e,3) + 5./36.*m11 + 2./3.*m11*RSe6611
          + 1./6.*m11*RS146ee - 37./90.*m11*RS16e - 13./45.*m11*RS14e
          + 13./72.*m11*RC146e );

      M12p6x12R +=  + pow(Ab(m11,mu2),2) * (  - 2./9.*m11*RC146e*
         RSe6611 + 2./9.*m11*pow(RC146e,2) + 29./162.*m11*pow(m66,3)*
         pow(RN1e,3) + 145./162.*m11*m44*pow(m66,2)*pow(RN1e,3) + 319./
         648.*m11*pow(m44,2)*m66*pow(RN1e,3) + 29./648.*m11*pow(m44,3)*
         pow(RN1e,3) - 4./9.*pow(m11,2)*RC146e*RN1e*RSe6611 - 58./81.*
         pow(m11,2)*pow(m66,2)*pow(RN1e,3) - 899./648.*pow(m11,2)*m44*
         m66*pow(RN1e,3) - 203./648.*pow(m11,2)*pow(m44,2)*pow(RN1e,3)
          + 203./216.*pow(m11,3)*m66*pow(RN1e,3) + 145./216.*pow(m11,3)
         *m44*pow(RN1e,3) - 29./72.*pow(m11,4)*pow(RN1e,3) );

      M12p6x12R +=  + Ab(m11,mu2)*Ab(m14,mu2) * ( 4./45.*RD146e + 4./45.
         *m44*RS16e - 4./9.*m11*RSe61 - 8./9.*m11*RSe6611 - 2./9.*m11*
         RS146ee - 34./45.*m11*RS16e + 2./9.*m11*RS14e + 2./3.*m11*
         RC146e - 4./9.*m11*me2*RN1e*RSe61 - 8./9.*m11*me2*RN1e*RSe6611
          - 4./9.*m11*m44*RN1e*RSe61 - 4./9.*pow(m11,2)*RN1e*RSe61 + 8./
         9.*pow(m11,2)*RN1e*RSe6611 );

      M12p6x12R +=  + Ab(m11,mu2)*Ab(m16,mu2) * ( 2./45.*RD146e + 2./45.
         *m66*RS14e + 4./9.*m11*RSe61 - 4./9.*m11*RSe6611 - 1./9.*m11*
         RS146ee + 1./9.*m11*RS16e - 17./45.*m11*RS14e + 1./3.*m11*
         RC146e + 4./9.*m11*me2*RN1e*RSe61 - 4./9.*m11*me2*RN1e*RSe6611
          + 4./9.*m11*m66*RN1e*RSe61 + 4./9.*pow(m11,2)*RN1e*RSe61 + 4./
         9.*pow(m11,2)*RN1e*RSe6611 );

      M12p6x12R +=  + Ab(m11,mu2)*Ab(m44,mu2) * ( 2./3.*m11*m44*RN1e*
         RS16e*RSe61 );

      M12p6x12R +=  + Ab(m11,mu2)*Ab(m46,mu2) * (  - 4./9.*m11*RS16e*
         RSe61 - 2./27.*m11*pow(RS16e,2) + 4./9.*m11*RS14e*RSe61 + 4./
         27.*m11*RS14e*RS16e - 2./27.*m11*pow(RS14e,2) - 4./9.*m11*me2*
         RN1e*RS16e*RSe61 + 4./9.*m11*me2*RN1e*RS14e*RSe61 - 8./27.*m11
         *m66*RN1e*RS16e*RSe61 - 4./27.*m11*m66*RN1e*RS14e*RSe61 - 4./
         27.*m11*m44*RN1e*RS16e*RSe61 + 4./27.*m11*m44*RN1e*RS14e*RSe61
          );

      M12p6x12R +=  + Ab(m11,mu2)*Ab(me2,mu2) * (  - 2./9.*m11*RC146e*
         RSe6611 - 8./27.*m11*m66*RN1e*RS14e*RSe61 + 2./27.*m11*m44*
         RN1e*RS16e*RSe61 + 8./9.*pow(m11,2)*RN1e*pow(RSe6611,2) + 4./9.
         *pow(m11,2)*RC146e*RN1e*RSe6611 );

      M12p6x12R +=  + Ab(m11,mu2)*Bb(m11,mu2) * ( 1./9.*m11*RD146e*
         RC146e + 29./81.*m11*m44*pow(m66,3)*pow(RN1e,3) + 29./81.*m11*
         pow(m44,2)*pow(m66,2)*pow(RN1e,3) + 29./324.*m11*pow(m44,3)*
         m66*pow(RN1e,3) + pow(m11,2) + 2./9.*pow(m11,2)*pow(RC146e,2)
          - 4./9.*pow(m11,2)*RD146e*RN1e*RSe6611 - 29./81.*pow(m11,2)*
         pow(m66,3)*pow(RN1e,3) - 145./81.*pow(m11,2)*m44*pow(m66,2)*
         pow(RN1e,3) - 319./324.*pow(m11,2)*pow(m44,2)*m66*pow(RN1e,3)
          - 29./324.*pow(m11,2)*pow(m44,3)*pow(RN1e,3) + 116./81.*pow(
         m11,3)*pow(m66,2)*pow(RN1e,3) + 899./324.*pow(m11,3)*m44*m66*
         pow(RN1e,3) + 203./324.*pow(m11,3)*pow(m44,2)*pow(RN1e,3) - 
         203./108.*pow(m11,4)*m66*pow(RN1e,3) - 145./108.*pow(m11,4)*
         m44*pow(RN1e,3) + 29./36.*pow(m11,5)*pow(RN1e,3) );

      M12p6x12R +=  + Ab(m11,mu2)*Cb(m11,mu2) * ( 2./9.*pow(m11,2)*
         RD146e*RC146e );

      M12p6x12R +=  + Ab(m14,mu2)*pi16 * (  - 1./6.*m11*RD146e + 1./12.
         *m11*me2 + 1./3.*m11*me2*RSe61 + 1./3.*m11*me2*RSe6611 + 1./2.
         *m11*m66 + 7./6.*m11*m44 + 1./6.*m11*m44*RSe61 + 1./6.*m11*m44
         *RSe6611 + 1./6.*m11*m44*RS16e - 1./12.*m11*m44*RC146e - 1./36.
         *pow(m11,2) + 1./18.*pow(m11,2)*RSe61 + 1./18.*pow(m11,2)*
         RSe6611 + 7./18.*pow(m11,2)*RS16e - 7./36.*pow(m11,2)*RC146e )
         ;

      M12p6x12R +=  + pow(Ab(m14,mu2),2) * (  - 1./45.*pow(m14,-1)*pow(
         me2,2) - 4./45.*pow(m14,-1)*pow(me2,2)*RSe61 - 4./45.*pow(
         m14,-1)*pow(me2,2)*RSe6611 - 1./27.*pow(m14,-1)*m44*RD146e + 
         11./270.*pow(m14,-1)*m44*me2 + 16./135.*pow(m14,-1)*m44*me2*
         RSe61 + 2./27.*pow(m14,-1)*m44*me2*RSe6611 - 1./54.*pow(
         m14,-1)*pow(m44,2) + 2./27.*pow(m14,-1)*pow(m44,2)*RSe6611 - 1.
         /27.*pow(m14,-1)*pow(m44,2)*RC146e - 2./27.*RD146e + 1./27.*
         me2 + 4./27.*me2*RSe61 + 4./27.*me2*RSe6611 - 1./27.*m44 + 4./
         27.*m44*RSe6611 - 2./27.*m44*RC146e + 46./135.*m11*pow(m14,-1)
         *RD146e + 1./54.*m11*pow(m14,-1)*me2 - 44./135.*m11*pow(
         m14,-1)*me2*RSe61 - 68./135.*m11*pow(m14,-1)*me2*RSe6611 - 31./
         108.*m11*pow(m14,-1)*m66 - 11./18.*m11*pow(m14,-1)*m44 - 11./
         54.*m11*pow(m14,-1)*m44*RSe61 + 13./54.*m11*pow(m14,-1)*m44*
         RSe6611 - 23./270.*m11*pow(m14,-1)*m44*RS16e - 17./108.*m11*
         pow(m14,-1)*m44*RC146e - 47./54.*m11 + 1./27.*m11*RSe61 + 1./3.
         *m11*RSe6611 );

      M12p6x12R +=  + pow(Ab(m14,mu2),2) * ( 5./27.*m11*RS16e - 13./54.
         *m11*RC146e - 59./108.*pow(m11,2)*pow(m14,-1) - 23./54.*pow(
         m11,2)*pow(m14,-1)*RSe61 + 1./6.*pow(m11,2)*pow(m14,-1)*
         RSe6611 - 227./270.*pow(m11,2)*pow(m14,-1)*RS16e + 23./108.*
         pow(m11,2)*pow(m14,-1)*RC146e );

      M12p6x12R +=  + Ab(m14,mu2)*Ab(m16,mu2) * (  - m11 );

      M12p6x12R +=  + Ab(m14,mu2)*Ab(me2,mu2) * (  - 2./45.*me2 - 8./45.
         *me2*RSe61 - 8./45.*me2*RSe6611 + 2./45.*m44 + 4./45.*m44*
         RSe61 - 16./45.*m11*RSe61 - 32./45.*m11*RSe6611 + 4./9.*m11*
         me2*RN1e*RSe61 + 8./9.*m11*me2*RN1e*RSe6611 + 4./9.*m11*m44*
         RN1e*RSe61 + 4./9.*pow(m11,2)*RN1e*RSe61 - 8./9.*pow(m11,2)*
         RN1e*RSe6611 );

      M12p6x12R +=  + Ab(m14,mu2)*Bb(m11,mu2) * ( 8./9.*m11*RD146e - 4./
         9.*m11*m44*RS16e - 8./9.*pow(m11,2)*RS16e );

      M12p6x12R +=  + Ab(m16,mu2)*pi16 * (  - 1./12.*m11*RD146e + 1./6.
         *m11*me2 - 1./3.*m11*me2*RSe61 + 1./6.*m11*me2*RSe6611 + 5./24.
         *m11*m66 - 1./6.*m11*m66*RSe61 + 1./12.*m11*m66*RSe6611 + 1./
         12.*m11*m66*RS14e - 1./24.*m11*m66*RC146e + 1./2.*m11*m44 - 1./
         72.*pow(m11,2) - 1./18.*pow(m11,2)*RSe61 + 1./36.*pow(m11,2)*
         RSe6611 + 7./36.*pow(m11,2)*RS14e - 7./72.*pow(m11,2)*RC146e )
         ;

      M12p6x12R +=  + pow(Ab(m16,mu2),2) * (  - 2./45.*pow(m16,-1)*pow(
         me2,2) + 4./45.*pow(m16,-1)*pow(me2,2)*RSe61 - 2./45.*pow(
         m16,-1)*pow(me2,2)*RSe6611 - 1./54.*pow(m16,-1)*m66*RD146e + 
         11./135.*pow(m16,-1)*m66*me2 - 16./135.*pow(m16,-1)*m66*me2*
         RSe61 + 1./27.*pow(m16,-1)*m66*me2*RSe6611 - 1./27.*pow(
         m16,-1)*pow(m66,2) + 1./27.*pow(m16,-1)*pow(m66,2)*RSe6611 - 1.
         /54.*pow(m16,-1)*pow(m66,2)*RC146e - 1./27.*RD146e + 2./27.*
         me2 - 4./27.*me2*RSe61 + 2./27.*me2*RSe6611 - 2./27.*m66 + 2./
         27.*m66*RSe6611 - 1./27.*m66*RC146e + 23./135.*m11*pow(m16,-1)
         *RD146e + 1./27.*m11*pow(m16,-1)*me2 + 44./135.*m11*pow(
         m16,-1)*me2*RSe61 - 34./135.*m11*pow(m16,-1)*me2*RSe6611 - 41./
         216.*m11*pow(m16,-1)*m66 + 11./54.*m11*pow(m16,-1)*m66*RSe61
          + 13./108.*m11*pow(m16,-1)*m66*RSe6611 - 23./540.*m11*pow(
         m16,-1)*m66*RS14e - 17./216.*m11*pow(m16,-1)*m66*RC146e - 31./
         108.*m11*pow(m16,-1)*m44 - 5./27.*m11 - 1./27.*m11*RSe61 + 1./
         6.*m11*RSe6611 );

      M12p6x12R +=  + pow(Ab(m16,mu2),2) * ( 5./54.*m11*RS14e - 13./108.
         *m11*RC146e - 59./216.*pow(m11,2)*pow(m16,-1) + 23./54.*pow(
         m11,2)*pow(m16,-1)*RSe61 + 1./12.*pow(m11,2)*pow(m16,-1)*
         RSe6611 - 227./540.*pow(m11,2)*pow(m16,-1)*RS14e + 23./216.*
         pow(m11,2)*pow(m16,-1)*RC146e );

      M12p6x12R +=  + Ab(m16,mu2)*Ab(me2,mu2) * (  - 4./45.*me2 + 8./45.
         *me2*RSe61 - 4./45.*me2*RSe6611 + 4./45.*m66 - 4./45.*m66*
         RSe61 + 16./45.*m11*RSe61 - 16./45.*m11*RSe6611 - 4./9.*m11*
         me2*RN1e*RSe61 + 4./9.*m11*me2*RN1e*RSe6611 - 4./9.*m11*m66*
         RN1e*RSe61 - 4./9.*pow(m11,2)*RN1e*RSe61 - 4./9.*pow(m11,2)*
         RN1e*RSe6611 );

      M12p6x12R +=  + Ab(m16,mu2)*Bb(m11,mu2) * ( 4./9.*m11*RD146e - 2./
         9.*m11*m66*RS14e - 4./9.*pow(m11,2)*RS14e );

      M12p6x12R +=  + pow(Ab(m44,mu2),2) * (  - 13./24.*m11 + 1./6.*m11
         *pow(RSe61,2) + 1./3.*m11*RS16e*RSe61 + 1./6.*m11*pow(RS16e,2)
          );

      M12p6x12R +=  + Ab(m44,mu2)*Ab(me2,mu2) * (  - 2./3.*m11*m44*RN1e
         *RS16e*RSe61 );

      M12p6x12R +=  + Ab(m44,mu2)*Bb(m11,mu2) * ( 1./3.*m11*m44*pow(
         RS16e,2) );

      M12p6x12R +=  + Ab(m44,mu2)*Bb(me2,mu2) * ( 1./3.*m11*m44*pow(
         RSe61,2) );

      M12p6x12R +=  + pow(Ab(m46,mu2),2) * (  - 2./3.*m11*pow(m46,-1)*
         me2*pow(RSe61,2) - 2./9.*m11*pow(m46,-1)*me2*RS16e*RSe61 + 2./
         9.*m11*pow(m46,-1)*me2*RS14e*RSe61 - 13./36.*m11*pow(m46,-1)*
         m66 - 4./27.*m11*pow(m46,-1)*m66*RS16e*RSe61 - 2./27.*m11*pow(
         m46,-1)*m66*RS14e*RSe61 + 2./27.*m11*pow(m46,-1)*m66*RS14e*
         RS16e + 1./27.*m11*pow(m46,-1)*m66*pow(RS14e,2) - 13./36.*m11*
         m44*pow(m46,-1) - 1./9.*m11*m44*pow(m46,-1)*pow(RSe61,2) - 2./
         27.*m11*m44*pow(m46,-1)*RS16e*RSe61 + 1./27.*m11*m44*pow(
         m46,-1)*pow(RS16e,2) + 2./27.*m11*m44*pow(m46,-1)*RS14e*RSe61
          + 2./27.*m11*m44*pow(m46,-1)*RS14e*RS16e - 2./9.*pow(m11,2)*
         pow(m46,-1)*RS16e*RSe61 - 2./27.*pow(m11,2)*pow(m46,-1)*pow(
         RS16e,2) + 2./9.*pow(m11,2)*pow(m46,-1)*RS14e*RSe61 + 4./27.*
         pow(m11,2)*pow(m46,-1)*RS14e*RS16e - 2./27.*pow(m11,2)*pow(
         m46,-1)*pow(RS14e,2) );

      M12p6x12R +=  + Ab(m46,mu2)*Ab(me2,mu2) * (  - 2./3.*m11*pow(
         RSe61,2) + 4./9.*m11*me2*RN1e*RS16e*RSe61 - 4./9.*m11*me2*RN1e
         *RS14e*RSe61 + 8./27.*m11*m66*RN1e*RS16e*RSe61 + 4./27.*m11*
         m66*RN1e*RS14e*RSe61 + 4./27.*m11*m44*RN1e*RS16e*RSe61 - 4./27.
         *m11*m44*RN1e*RS14e*RSe61 );

      M12p6x12R +=  + Ab(m46,mu2)*Bb(m11,mu2) * ( 4./27.*m11*m66*RS14e*
         RS16e + 2./27.*m11*m66*pow(RS14e,2) + 2./27.*m11*m44*pow(
         RS16e,2) + 4./27.*m11*m44*RS14e*RS16e - 2./27.*pow(m11,2)*pow(
         RS16e,2) + 4./27.*pow(m11,2)*RS14e*RS16e - 2./27.*pow(m11,2)*
         pow(RS14e,2) );

      M12p6x12R +=  + Ab(m46,mu2)*Bb(me2,mu2) * (  - 2./3.*m11*me2*pow(
         RSe61,2) - 2./9.*m11*m44*pow(RSe61,2) );

      M12p6x12R +=  + pow(Ab(me2,mu2),2) * (  - 1./45.*m66 + 2./45.*m66
         *RSe61 - 1./45.*m66*RSe6611 + 1./45.*pow(m66,2)*pow(me2,-1) - 
         1./45.*pow(m66,2)*pow(me2,-1)*RSe61 - 1./90.*m44 - 2./45.*m44*
         RSe61 - 2./45.*m44*RSe6611 + 1./90.*pow(m44,2)*pow(me2,-1) + 1.
         /45.*pow(m44,2)*pow(me2,-1)*RSe61 - 1./9.*m11*pow(me2,-1)*
         RD146e*RSe6611 - 19./120.*m11 - 19./60.*m11*RSe6611 + 4./9.*
         m11*pow(RSe6611,2) - 2./135.*m11*m66*pow(me2,-1) + 1./15.*m11*
         m66*pow(me2,-1)*RSe61 + 13./27.*m11*m66*pow(me2,-1)*pow(
         RSe61,2) - 17./135.*m11*m66*pow(me2,-1)*RSe6611 - 4./27.*m11*
         m66*pow(me2,-1)*RS14e*RSe61 + 1./27.*m11*m66*pow(me2,-1)*pow(
         RS14e,2) + 8./27.*m11*m66*RN1e*RS14e*RSe61 - 1./135.*m11*m44*
         pow(me2,-1) - 1./15.*m11*m44*pow(me2,-1)*RSe61 + 19./54.*m11*
         m44*pow(me2,-1)*pow(RSe61,2) - 34./135.*m11*m44*pow(me2,-1)*
         RSe6611 + 1./27.*m11*m44*pow(me2,-1)*RS16e*RSe61 + 1./54.*m11*
         m44*pow(me2,-1)*pow(RS16e,2) - 2./27.*m11*m44*RN1e*RS16e*RSe61
          - 17./45.*pow(m11,2)*pow(me2,-1)*RSe6611 );

      M12p6x12R +=  + pow(Ab(me2,mu2),2) * ( 4./9.*pow(m11,2)*pow(
         me2,-1)*pow(RSe6611,2) - 1./3.*pow(m11,2)*pow(me2,-1)*RC146e*
         RSe6611 - 8./9.*pow(m11,2)*RN1e*pow(RSe6611,2) );

      M12p6x12R +=  + Ab(me2,mu2)*Bb(m11,mu2) * (  - 2./9.*m11*RD146e*
         RSe6611 + 2./27.*m11*m66*pow(RS14e,2) + 1./27.*m11*m44*pow(
         RS16e,2) - 4./9.*pow(m11,2)*RC146e*RSe6611 + 4./9.*pow(m11,2)*
         RD146e*RN1e*RSe6611 );

      M12p6x12R +=  + Ab(me2,mu2)*Bb(me2,mu2) * ( 8./27.*m11*m66*pow(
         RSe61,2) + 1./27.*m11*m44*pow(RSe61,2) );

      M12p6x12R +=  + Ab(me2,mu2)*Cb(m11,mu2) * (  - 4./9.*pow(m11,2)*
         RD146e*RSe6611 );

      M12p6x12R +=  + pow(Bb(m11,mu2),2) * ( 1./18.*m11*pow(RD146e,2)
          + 2./9.*pow(m11,2)*RD146e*RC146e );

      M12p6x12R +=  + Bb(m11,mu2)*Cb(m11,mu2) * ( 2./9.*pow(m11,2)*pow(
         RD146e,2) );

      M12p6x12R +=  + hh(1,m11,m11,m11,m11,mu2) * ( 1./3.*pow(m11,2) + 
         2./9.*pow(m11,2)*pow(RC146e,2) );

      M12p6x12R +=  + hh(1,m11,m11,me2,m11,mu2) * (  - 8./9.*pow(m11,2)
         *RC146e*RSe6611 );

      M12p6x12R +=  + hh(1,m11,m14,m14,m11,mu2) * (  - 11./18.*m11*
         RD146e - 4./9.*m11*m44*RS16e + 7./9.*pow(m11,2)*RS16e - 5./6.*
         pow(m11,2)*RC146e );

      M12p6x12R +=  + hh(1,m11,m16,m16,m11,mu2) * (  - 11./36.*m11*
         RD146e - 2./9.*m11*m66*RS14e + 7./18.*pow(m11,2)*RS14e - 5./12.
         *pow(m11,2)*RC146e );

      M12p6x12R +=  + hh(1,m11,me2,me2,m11,mu2) * ( 8./9.*pow(m11,2)*
         pow(RSe6611,2) );

      M12p6x12R +=  + hh(1,m14,m14,m44,m11,mu2) * ( 3./4.*m11*m44 );

      M12p6x12R +=  + hh(1,m14,m14,me2,m11,mu2) * ( 1./27.*pow(me2,2)
          + 4./27.*pow(me2,2)*RSe61 + 4./27.*pow(me2,2)*RSe6611 - 2./27.
         *m44*me2 - 4./27.*m44*me2*RSe61 + 1./27.*pow(m44,2) + 1./12.*
         m11*me2 + 17./27.*m11*me2*RSe61 + 25./27.*m11*me2*RSe6611 - 8./
         27.*m11*m44*RSe61 + 16./27.*pow(m11,2)*RSe6611 );

      M12p6x12R +=  + hh(1,m14,m16,m46,m11,mu2) * ( 1./2.*m11*m66 + 1./
         2.*m11*m44 );

      M12p6x12R +=  + hh(1,m16,m16,me2,m11,mu2) * ( 2./27.*pow(me2,2)
          - 4./27.*pow(me2,2)*RSe61 + 2./27.*pow(me2,2)*RSe6611 - 4./27.
         *m66*me2 + 4./27.*m66*me2*RSe61 + 2./27.*pow(m66,2) + 1./6.*
         m11*me2 - 17./27.*m11*me2*RSe61 + 25./54.*m11*me2*RSe6611 + 8./
         27.*m11*m66*RSe61 + 8./27.*pow(m11,2)*RSe6611 );

      M12p6x12R +=  + hh(2,m11,m11,m11,m11,mu2) * ( 4./9.*pow(m11,2)*
         RD146e*RC146e );

      M12p6x12R +=  + hh(2,m11,m14,m14,m11,mu2) * ( 1./2.*pow(m11,2)*
         RD146e );

      M12p6x12R +=  + hh(2,m11,m16,m16,m11,mu2) * ( 1./4.*pow(m11,2)*
         RD146e );

      M12p6x12R +=  + hh(2,m11,me2,m11,m11,mu2) * (  - 8./9.*pow(m11,2)
         *RD146e*RSe6611 );

      M12p6x12R +=  + hh(5,m11,m11,m11,m11,mu2) * ( 2./9.*pow(m11,2)*
         pow(RD146e,2) );

      M12p6x12R +=  + hh1(1,m11,m14,m14,m11,mu2) * ( 4./9.*m11*RD146e
          + 4./9.*m11*m44*RS16e - 16./9.*pow(m11,2)*RS16e + 4./3.*pow(
         m11,2)*RC146e );

      M12p6x12R +=  + hh1(1,m11,m16,m16,m11,mu2) * ( 2./9.*m11*RD146e
          + 2./9.*m11*m66*RS14e - 8./9.*pow(m11,2)*RS14e + 2./3.*pow(
         m11,2)*RC146e );

      M12p6x12R +=  + hh1(1,me2,m14,m14,m11,mu2) * (  - 2./9.*m11*me2
          - 8./9.*m11*me2*RSe61 - 8./9.*m11*me2*RSe6611 + 2./9.*m11*m44
          + 4./9.*m11*m44*RSe61 - 8./9.*pow(m11,2)*RSe61 - 16./9.*pow(
         m11,2)*RSe6611 );

      M12p6x12R +=  + hh1(1,me2,m16,m16,m11,mu2) * (  - 4./9.*m11*me2
          + 8./9.*m11*me2*RSe61 - 4./9.*m11*me2*RSe6611 + 4./9.*m11*m66
          - 4./9.*m11*m66*RSe61 + 8./9.*pow(m11,2)*RSe61 - 8./9.*pow(
         m11,2)*RSe6611 );

      M12p6x12R +=  + hh1(3,m14,m11,m14,m11,mu2) * (  - 8./3.*pow(
         m11,2)*RD146e );

      M12p6x12R +=  + hh1(3,m16,m11,m16,m11,mu2) * (  - 4./3.*pow(
         m11,2)*RD146e );

      M12p6x12R +=  + hh21(1,m11,m14,m14,m11,mu2) * ( pow(m11,2)*RS16e
          - 1./2.*pow(m11,2)*RC146e );

      M12p6x12R +=  + hh21(1,m11,m16,m16,m11,mu2) * ( 1./2.*pow(m11,2)*
         RS14e - 1./4.*pow(m11,2)*RC146e );

      M12p6x12R +=  + hh21(1,m44,m14,m14,m11,mu2) * ( 9./4.*pow(m11,2)
          );

      M12p6x12R +=  + hh21(1,m46,m14,m16,m11,mu2) * ( 3*pow(m11,2) );

      M12p6x12R +=  + hh21(1,me2,m14,m14,m11,mu2) * ( 1./4.*pow(m11,2)
          + pow(m11,2)*RSe61 + pow(m11,2)*RSe6611 );

      M12p6x12R +=  + hh21(1,me2,m16,m16,m11,mu2) * ( 1./2.*pow(m11,2)
          - pow(m11,2)*RSe61 + 1./2.*pow(m11,2)*RSe6611 );

      M12p6x12R +=  + hh21(2,m11,m14,m14,m11,mu2) * (  - 1./2.*pow(
         m11,2)*RD146e );

      M12p6x12R +=  + hh21(2,m11,m16,m16,m11,mu2) * (  - 1./4.*pow(
         m11,2)*RD146e );

  return M12p6x12R/pow(f0,4);
}

// 2 valence mass 2 sea mass *****************************************

double mv2s2nf3p4(const quarkmassnf mass, const Linf Liin){
  return mv2s2nf3p4L(mass,Liin)+mv2s2nf3p4R(mass);
}

double mv2s2nf3p4L(const quarkmassnf mass, const Linf Liin){
  double L0r,L1r,L2r,L3r,L4r,L5r,L6r,L7r,L8r,L9r,L10r,L11r,H1r,H2r,mu;
  int nfLi;
  Liin.out(L0r,L1r,L2r,L3r,L4r,L5r,L6r,L7r,L8r,L9r,L10r,L11r,H1r,H2r,mu,nfLi);
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if (mu0 != mu){
    cout << "ERROR: differing mu's in mv2s2nf3p4L\n";}
  if ( (nq != 4) || (B0mq.size() != 4)){
    cout << "ERROR: mv2s2nf3p4L probably called with wrong number"
	 << "of quark masses\n";}
  // four masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m33 = 2.*B0mq[1];
  double m44 = 2.*B0mq[2];
  double m66 = 2.*B0mq[3];
  double M13p4x22L =
       + 8*m33*m66*L6r - 4*m33*m66*L4r + 16*m33*m44*L6r - 8*m33*m44*L4r
          + 4*pow(m33,2)*L8r - 2*pow(m33,2)*L5r + 8*m11*m66*L6r - 4*m11
         *m66*L4r + 16*m11*m44*L6r - 8*m11*m44*L4r + 8*m11*m33*L8r - 4*
         m11*m33*L5r + 4*pow(m11,2)*L8r - 2*pow(m11,2)*L5r;

  return M13p4x22L/pow(f0,2);
}


double mv2s2nf3p4R(const quarkmassnf mass){
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if ( (nq != 4) || (B0mq.size() != 4)){
    cout << "ERROR: mv2s1nf3p4R probably called with wrong number"
	 << "of quark masses\n";}
  // four masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m33 = 2.*B0mq[1];
  double m44 = 2.*B0mq[2];
  double m66 = 2.*B0mq[3];
  double me2 = (m44+2.*m66)/3.;
  double RD146e = RD(m11,m44,m66,me2);
  double RD346e = RD(m33,m44,m66,me2);
  double RS1463e = RS(m11,m44,m66,m33,me2);
  double RS3461e = RS(m33,m44,m66,m11,me2);
  double RSe6611 = RS(me2,m66,m66,m11,m11);
  double RSe6613 = RS(me2,m66,m66,m11,m33);
  double RSe6633 = RS(me2,m66,m66,m33,m33);
  double mu2 = mu0*mu0;
  double M13p4x22R =
       + Ab(m11,mu2) * ( 1./18.*RD146e - 1./9.*m33*RS1463e - 2./9.*m11*
         RS1463e );

      M13p4x22R +=  + Ab(m33,mu2) * ( 1./18.*RD346e - 2./9.*m33*RS3461e
          - 1./9.*m11*RS3461e );

      M13p4x22R +=  + Ab(me2,mu2) * ( 2./9.*me2*RSe6613 - 1./9.*me2*
         RSe6633 - 1./9.*me2*RSe6611 + 2./9.*m33*RSe6613 + 1./9.*m33*
         RSe6633 + 2./9.*m11*RSe6613 + 1./9.*m11*RSe6611 );

  return M13p4x22R/pow(f0,2);
}

double mv2s2nf3p6(const quarkmassnf mass, const Linf Liin, const Ki Kiin){
  return mv2s2nf3p6L(mass,Liin)+mv2s2nf3p6K(mass,Kiin)+mv2s2nf3p6R(mass);
}

double mv2s2nf3p6L(const quarkmassnf mass, const Linf Liin){
  double L0r,L1r,L2r,L3r,L4r,L5r,L6r,L7r,L8r,L9r,L10r,L11r,H1r,H2r,mu;
  int nfLi;
  Liin.out(L0r,L1r,L2r,L3r,L4r,L5r,L6r,L7r,L8r,L9r,L10r,L11r,H1r,H2r,mu,nfLi);
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if (mu0 != mu){
    cout << "ERROR: differing mu's in mv2s2nf3p6L\n";}
  if ( (nq != 4) || (B0mq.size() != 4)){
    cout << "ERROR: mv2s2nf3p6L probably called with wrong number"
	 << "of quark masses\n";}
  // four masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m33 = 2.*B0mq[1];
  double m44 = 2.*B0mq[2];
  double m66 = 2.*B0mq[3];
  double m13 = (m11+m33)/2;
  double m14 = (m11+m44)/2;
  double m16 = (m11+m66)/2;
  double m34 = (m33+m44)/2;
  double m36 = (m33+m66)/2;
  double m46 = (m44+m66)/2;
  double me2 = (m44+2.*m66)/3.;

  double RN13 = RN(m11,m33);
  double RN1e = RN(m11,me2);
  double RN3e = RN(m33,me2);

  double RD146e = RD(m11,m44,m66,me2);
  double RD346e = RD(m33,m44,m66,me2);

  double RC146e = RC(m11,m44,m66,me2);
  double RC346e = RC(m33,m44,m66,me2);

  double RS14e = RS(m11,m44,me2);
  double RS16e = RS(m11,m66,me2);
  double RS34e = RS(m33,m44,me2);
  double RS36e = RS(m33,m66,me2);
  double RSe61 = RS(me2,m66,m11);
  double RSe63 = RS(me2,m66,m33);

  double RS1463e = RS(m11,m44,m66,m33,me2);
  double RS3461e = RS(m33,m44,m66,m11,me2);
  double RSe6611 = RS(me2,m66,m66,m11,m11);
  double RSe6613 = RS(me2,m66,m66,m11,m33);
  double RSe6633 = RS(me2,m66,m66,m33,m33);

  double mu2 = mu0*mu0;

  double M13p6x22L =
       + pi16 * ( 2./3.*m33*pow(me2,2)*L3r*RSe6613 + 1./3.*m33*pow(
         me2,2)*L3r*RSe6633 + 1./3.*m33*pow(me2,2)*L3r*RSe6611 + m33*
         pow(me2,2)*L2r + 2./3.*m33*pow(me2,2)*L0r*RSe6633 + 2./3.*m33*
         pow(me2,2)*L0r*RSe6611 + 1./4.*m33*pow(m66,2)*L3r + m33*pow(
         m66,2)*L2r + 1./2.*m33*pow(m66,2)*L0r + 2*m33*m44*m66*L2r + 1./
         2.*m33*pow(m44,2)*L3r + 4*m33*pow(m44,2)*L2r + m33*pow(m44,2)*
         L0r - 1./3.*pow(m33,2)*L3r*RD346e - 2./3.*pow(m33,2)*L0r*
         RD346e + 1./4.*pow(m33,2)*m66*L3r + 1./2.*pow(m33,2)*m66*L0r
          + 1./2.*pow(m33,2)*m44*L3r + pow(m33,2)*m44*L0r + 3./8.*pow(
         m33,3)*L3r - 1./3.*pow(m33,3)*L3r*RS3461e - 1./6.*pow(m33,3)*
         L3r*RC346e + 1./4.*pow(m33,3)*L2r + 1./2.*pow(m33,3)*L1r + 3./
         4.*pow(m33,3)*L0r - 1./3.*pow(m33,3)*L0r*RC346e + 2./3.*m11*
         pow(me2,2)*L3r*RSe6613 + 1./3.*m11*pow(me2,2)*L3r*RSe6633 + 1./
         3.*m11*pow(me2,2)*L3r*RSe6611 + m11*pow(me2,2)*L2r + 2./3.*m11
         *pow(me2,2)*L0r*RSe6633 + 2./3.*m11*pow(me2,2)*L0r*RSe6611 + 1.
         /4.*m11*pow(m66,2)*L3r );

      M13p6x22L +=  + pi16 * ( m11*pow(m66,2)*L2r + 1./2.*m11*pow(
         m66,2)*L0r + 2*m11*m44*m66*L2r + 1./2.*m11*pow(m44,2)*L3r + 4*
         m11*pow(m44,2)*L2r + m11*pow(m44,2)*L0r - 1./3.*m11*m33*L3r*
         RD346e - 1./3.*m11*m33*L3r*RD146e - 2./3.*m11*m33*L0r*RD346e
          - 2./3.*m11*m33*L0r*RD146e + 1./2.*m11*m33*m66*L3r + m11*m33*
         m66*L0r + m11*m33*m44*L3r + 2*m11*m33*m44*L0r + 3./8.*m11*pow(
         m33,2)*L3r - 1./3.*m11*pow(m33,2)*L3r*RS3461e - 1./6.*m11*pow(
         m33,2)*L3r*RC346e + 3./4.*m11*pow(m33,2)*L2r + 3./2.*m11*pow(
         m33,2)*L1r + 3./4.*m11*pow(m33,2)*L0r - 1./3.*m11*pow(m33,2)*
         L0r*RC346e - 1./3.*pow(m11,2)*L3r*RD146e - 2./3.*pow(m11,2)*
         L0r*RD146e + 1./4.*pow(m11,2)*m66*L3r + 1./2.*pow(m11,2)*m66*
         L0r + 1./2.*pow(m11,2)*m44*L3r + pow(m11,2)*m44*L0r + 3./8.*
         pow(m11,2)*m33*L3r - 1./3.*pow(m11,2)*m33*L3r*RS1463e - 1./6.*
         pow(m11,2)*m33*L3r*RC146e + 3./4.*pow(m11,2)*m33*L2r + 3./2.*
         pow(m11,2)*m33*L1r + 3./4.*pow(m11,2)*m33*L0r - 1./3.*pow(
         m11,2)*m33*L0r*RC146e );

      M13p6x22L +=  + pi16 * ( 3./8.*pow(m11,3)*L3r - 1./3.*pow(m11,3)*
         L3r*RS1463e - 1./6.*pow(m11,3)*L3r*RC146e + 1./4.*pow(m11,3)*
         L2r + 1./2.*pow(m11,3)*L1r + 3./4.*pow(m11,3)*L0r - 1./3.*pow(
         m11,3)*L0r*RC146e );

      M13p6x22L +=  + Ab(m11,mu2) * (  - 8./9.*m66*L4r*RD146e - 32./27.
         *pow(m66,2)*L8r*RS14e*RSe63 + 32./27.*pow(m66,2)*L8r*RS14e*
         RSe61 - 8./27.*pow(m66,2)*L8r*pow(RS14e,2) - 32./27.*pow(
         m66,2)*L7r*RS14e*RSe63 + 32./27.*pow(m66,2)*L7r*RS14e*RSe61 - 
         8./27.*pow(m66,2)*L7r*pow(RS14e,2) - 32./27.*pow(m66,2)*L6r*
         RS14e*RSe63 + 32./27.*pow(m66,2)*L6r*RS14e*RSe61 - 8./27.*pow(
         m66,2)*L6r*pow(RS14e,2) - 32./27.*pow(m66,2)*me2*L8r*RN1e*
         RS14e*RSe63 + 32./27.*pow(m66,2)*me2*L8r*RN1e*RS14e*RSe61 - 32.
         /27.*pow(m66,2)*me2*L7r*RN1e*RS14e*RSe63 + 32./27.*pow(m66,2)*
         me2*L7r*RN1e*RS14e*RSe61 - 32./27.*pow(m66,2)*me2*L6r*RN1e*
         RS14e*RSe63 + 32./27.*pow(m66,2)*me2*L6r*RN1e*RS14e*RSe61 - 16.
         /9.*m44*L4r*RD146e - 64./27.*m44*m66*L7r*RS16e*RSe63 + 64./27.
         *m44*m66*L7r*RS16e*RSe61 + 32./27.*m44*m66*L7r*RS14e*RSe63 - 
         32./27.*m44*m66*L7r*RS14e*RSe61 - 32./27.*m44*m66*L7r*RS14e*
         RS16e + 32./27.*m44*m66*L6r*RS16e*RSe63 - 32./27.*m44*m66*L6r*
         RS16e*RSe61 );

      M13p6x22L +=  + Ab(m11,mu2) * (  - 16./27.*m44*m66*L6r*pow(
         RS16e,2) - 64./27.*m44*m66*L6r*RS14e*RSe63 + 64./27.*m44*m66*
         L6r*RS14e*RSe61 - 16./27.*m44*m66*L6r*pow(RS14e,2) - 64./27.*
         m44*m66*me2*L7r*RN1e*RS16e*RSe63 + 64./27.*m44*m66*me2*L7r*
         RN1e*RS16e*RSe61 + 32./27.*m44*m66*me2*L7r*RN1e*RS14e*RSe63 - 
         32./27.*m44*m66*me2*L7r*RN1e*RS14e*RSe61 + 32./27.*m44*m66*me2
         *L6r*RN1e*RS16e*RSe63 - 32./27.*m44*m66*me2*L6r*RN1e*RS16e*
         RSe61 - 64./27.*m44*m66*me2*L6r*RN1e*RS14e*RSe63 + 64./27.*m44
         *m66*me2*L6r*RN1e*RS14e*RSe61 + 32./27.*pow(m44,2)*L8r*RS16e*
         RSe63 - 32./27.*pow(m44,2)*L8r*RS16e*RSe61 - 16./27.*pow(
         m44,2)*L8r*pow(RS16e,2) + 64./27.*pow(m44,2)*L7r*RS16e*RSe63
          - 64./27.*pow(m44,2)*L7r*RS16e*RSe61 - 32./27.*pow(m44,2)*L7r
         *pow(RS16e,2) + 64./27.*pow(m44,2)*L6r*RS16e*RSe63 - 64./27.*
         pow(m44,2)*L6r*RS16e*RSe61 - 32./27.*pow(m44,2)*L6r*pow(
         RS16e,2) + 32./27.*pow(m44,2)*me2*L8r*RN1e*RS16e*RSe63 - 32./
         27.*pow(m44,2)*me2*L8r*RN1e*RS16e*RSe61 );

      M13p6x22L +=  + Ab(m11,mu2) * ( 64./27.*pow(m44,2)*me2*L7r*RN1e*
         RS16e*RSe63 - 64./27.*pow(m44,2)*me2*L7r*RN1e*RS16e*RSe61 + 64.
         /27.*pow(m44,2)*me2*L6r*RN1e*RS16e*RSe63 - 64./27.*pow(m44,2)*
         me2*L6r*RN1e*RS16e*RSe61 + 4./9.*m33*L5r*RD146e - 5./3.*m33*
         L3r*RD146e - 2./3.*m33*L0r*RD146e - 16./9.*m33*m66*L7r*RS14e
          - 16./9.*m33*m66*L6r*RS1463e + 16./27.*m33*m66*L5r*RS14e*
         RSe63 + 16./9.*m33*m66*L4r*RS1463e - 16./9.*m33*m66*L4r*
         RSe6613 - 16./27.*m33*m66*L4r*RS16e*RSe63 + 16./27.*m33*m66*
         L4r*RS14e*RSe63 + 16./27.*m33*m66*me2*L5r*RN1e*RS14e*RSe63 - 
         16./9.*m33*m66*me2*L4r*RN1e*RSe6613 - 16./27.*m33*m66*me2*L4r*
         RN1e*RS16e*RSe63 + 16./27.*m33*m66*me2*L4r*RN1e*RS14e*RSe63 - 
         32./27.*m33*pow(m66,2)*L8r*RN1e*RS14e*RSe63 + 16./27.*m33*pow(
         m66,2)*L8r*RN13*RS14e*RS34e - 32./27.*m33*pow(m66,2)*L7r*RN1e*
         RS14e*RSe63 + 16./27.*m33*pow(m66,2)*L7r*RN13*RS14e*RS34e - 32.
         /27.*m33*pow(m66,2)*L6r*RN1e*RS14e*RSe63 + 16./27.*m33*pow(
         m66,2)*L6r*RN13*RS14e*RS34e );

      M13p6x22L +=  + Ab(m11,mu2) * (  - 32./9.*m33*m44*L7r*RS16e - 32./
         9.*m33*m44*L6r*RS1463e - 16./27.*m33*m44*L5r*RS16e*RSe63 + 32./
         9.*m33*m44*L4r*RS1463e - 32./9.*m33*m44*L4r*RSe6613 - 32./27.*
         m33*m44*L4r*RS16e*RSe63 + 32./27.*m33*m44*L4r*RS14e*RSe63 - 16.
         /27.*m33*m44*me2*L5r*RN1e*RS16e*RSe63 - 32./9.*m33*m44*me2*L4r
         *RN1e*RSe6613 - 32./27.*m33*m44*me2*L4r*RN1e*RS16e*RSe63 + 32./
         27.*m33*m44*me2*L4r*RN1e*RS14e*RSe63 - 64./27.*m33*m44*m66*L7r
         *RN1e*RS16e*RSe63 + 32./27.*m33*m44*m66*L7r*RN1e*RS14e*RSe63
          + 32./27.*m33*m44*m66*L7r*RN13*RS16e*RS34e + 32./27.*m33*m44*
         m66*L7r*RN13*RS14e*RS36e + 32./27.*m33*m44*m66*L6r*RN1e*RS16e*
         RSe63 - 64./27.*m33*m44*m66*L6r*RN1e*RS14e*RSe63 + 32./27.*m33
         *m44*m66*L6r*RN13*RS16e*RS36e + 32./27.*m33*m44*m66*L6r*RN13*
         RS14e*RS34e + 32./27.*m33*pow(m44,2)*L8r*RN1e*RS16e*RSe63 + 32.
         /27.*m33*pow(m44,2)*L8r*RN13*RS16e*RS36e + 64./27.*m33*pow(
         m44,2)*L7r*RN1e*RS16e*RSe63 + 64./27.*m33*pow(m44,2)*L7r*RN13*
         RS16e*RS36e );

      M13p6x22L +=  + Ab(m11,mu2) * ( 64./27.*m33*pow(m44,2)*L6r*RN1e*
         RS16e*RSe63 + 64./27.*m33*pow(m44,2)*L6r*RN13*RS16e*RS36e - 16.
         /9.*pow(m33,2)*L8r*RS1463e + 4./9.*pow(m33,2)*L5r*RS1463e - 16.
         /9.*pow(m33,2)*m66*L7r*RN13*RS14e - 16./9.*pow(m33,2)*m66*L6r*
         RN13*RS1463e - 32./9.*pow(m33,2)*m44*L7r*RN13*RS16e - 32./9.*
         pow(m33,2)*m44*L6r*RN13*RS1463e - 16./9.*pow(m33,3)*L8r*RN13*
         RS1463e - 5./3.*m11*L3r*RD146e - 2./3.*m11*L0r*RD146e + 32./9.
         *m11*m66*L7r*RSe63 - 32./9.*m11*m66*L7r*RSe61 - 16./9.*m11*m66
         *L7r*RS14e - 32./9.*m11*m66*L6r*RS1463e + 32./9.*m11*m66*L6r*
         RSe6613 - 32./9.*m11*m66*L6r*RSe6611 + 8./9.*m11*m66*L6r*
         RC146e + 32./27.*m11*m66*L5r*RS14e*RSe63 - 8./27.*m11*m66*L5r*
         RS14e*RS34e + 4./27.*m11*m66*L5r*pow(RS14e,2) + 8./9.*m11*m66*
         L4r*RS3461e + 40./9.*m11*m66*L4r*RS1463e - 32./9.*m11*m66*L4r*
         RSe6613 - 32./27.*m11*m66*L4r*RS16e*RSe63 - 16./27.*m11*m66*
         L4r*RS16e*RS36e + 8./27.*m11*m66*L4r*pow(RS16e,2) + 32./27.*
         m11*m66*L4r*RS14e*RSe63 );

      M13p6x22L +=  + Ab(m11,mu2) * (  - 8./27.*m11*m66*L4r*RS14e*RS34e
          + 4./27.*m11*m66*L4r*pow(RS14e,2) - 4./9.*m11*m66*L4r*RC146e
          + 32./9.*m11*m66*me2*L7r*RN1e*RSe63 - 32./9.*m11*m66*me2*L7r*
         RN1e*RSe61 + 32./9.*m11*m66*me2*L6r*RN1e*RSe6613 - 32./9.*m11*
         m66*me2*L6r*RN1e*RSe6611 + 32./27.*m11*m66*me2*L5r*RN1e*RS14e*
         RSe63 - 32./9.*m11*m66*me2*L4r*RN1e*RSe6613 - 32./27.*m11*m66*
         me2*L4r*RN1e*RS16e*RSe63 + 32./27.*m11*m66*me2*L4r*RN1e*RS14e*
         RSe63 - 32./27.*m11*pow(m66,2)*L8r*RN1e*RS14e*RSe63 - 32./27.*
         m11*pow(m66,2)*L8r*RN1e*RS14e*RSe61 + 32./27.*m11*pow(m66,2)*
         L8r*RN13*RS14e*RS34e - 32./27.*m11*pow(m66,2)*L7r*RN1e*RS14e*
         RSe63 - 32./27.*m11*pow(m66,2)*L7r*RN1e*RS14e*RSe61 + 32./27.*
         m11*pow(m66,2)*L7r*RN13*RS14e*RS34e - 32./27.*m11*pow(m66,2)*
         L6r*RN1e*RS14e*RSe63 - 32./27.*m11*pow(m66,2)*L6r*RN1e*RS14e*
         RSe61 + 32./27.*m11*pow(m66,2)*L6r*RN13*RS14e*RS34e - 32./9.*
         m11*m44*L7r*RSe63 + 32./9.*m11*m44*L7r*RSe61 - 32./9.*m11*m44*
         L7r*RS16e );

      M13p6x22L +=  + Ab(m11,mu2) * (  - 64./9.*m11*m44*L6r*RS1463e + 
         64./9.*m11*m44*L6r*RSe6613 - 64./9.*m11*m44*L6r*RSe6611 + 16./
         9.*m11*m44*L6r*RC146e - 32./27.*m11*m44*L5r*RS16e*RSe63 - 16./
         27.*m11*m44*L5r*RS16e*RS36e + 8./27.*m11*m44*L5r*pow(RS16e,2)
          + 16./9.*m11*m44*L4r*RS3461e + 80./9.*m11*m44*L4r*RS1463e - 
         64./9.*m11*m44*L4r*RSe6613 - 64./27.*m11*m44*L4r*RS16e*RSe63
          - 32./27.*m11*m44*L4r*RS16e*RS36e + 16./27.*m11*m44*L4r*pow(
         RS16e,2) + 64./27.*m11*m44*L4r*RS14e*RSe63 - 16./27.*m11*m44*
         L4r*RS14e*RS34e + 8./27.*m11*m44*L4r*pow(RS14e,2) - 8./9.*m11*
         m44*L4r*RC146e - 32./9.*m11*m44*me2*L7r*RN1e*RSe63 + 32./9.*
         m11*m44*me2*L7r*RN1e*RSe61 + 64./9.*m11*m44*me2*L6r*RN1e*
         RSe6613 - 64./9.*m11*m44*me2*L6r*RN1e*RSe6611 - 32./27.*m11*
         m44*me2*L5r*RN1e*RS16e*RSe63 - 64./9.*m11*m44*me2*L4r*RN1e*
         RSe6613 - 64./27.*m11*m44*me2*L4r*RN1e*RS16e*RSe63 + 64./27.*
         m11*m44*me2*L4r*RN1e*RS14e*RSe63 - 64./27.*m11*m44*m66*L7r*
         RN1e*RS16e*RSe63 );

      M13p6x22L +=  + Ab(m11,mu2) * (  - 64./27.*m11*m44*m66*L7r*RN1e*
         RS16e*RSe61 + 32./27.*m11*m44*m66*L7r*RN1e*RS14e*RSe63 + 32./
         27.*m11*m44*m66*L7r*RN1e*RS14e*RSe61 + 64./27.*m11*m44*m66*L7r
         *RN13*RS16e*RS34e + 64./27.*m11*m44*m66*L7r*RN13*RS14e*RS36e
          + 32./27.*m11*m44*m66*L6r*RN1e*RS16e*RSe63 + 32./27.*m11*m44*
         m66*L6r*RN1e*RS16e*RSe61 - 64./27.*m11*m44*m66*L6r*RN1e*RS14e*
         RSe63 - 64./27.*m11*m44*m66*L6r*RN1e*RS14e*RSe61 + 64./27.*m11
         *m44*m66*L6r*RN13*RS16e*RS36e + 64./27.*m11*m44*m66*L6r*RN13*
         RS14e*RS34e + 32./27.*m11*pow(m44,2)*L8r*RN1e*RS16e*RSe63 + 32.
         /27.*m11*pow(m44,2)*L8r*RN1e*RS16e*RSe61 + 64./27.*m11*pow(
         m44,2)*L8r*RN13*RS16e*RS36e + 64./27.*m11*pow(m44,2)*L7r*RN1e*
         RS16e*RSe63 + 64./27.*m11*pow(m44,2)*L7r*RN1e*RS16e*RSe61 + 
         128./27.*m11*pow(m44,2)*L7r*RN13*RS16e*RS36e + 64./27.*m11*
         pow(m44,2)*L6r*RN1e*RS16e*RSe63 + 64./27.*m11*pow(m44,2)*L6r*
         RN1e*RS16e*RSe61 + 128./27.*m11*pow(m44,2)*L6r*RN13*RS16e*
         RS36e );

      M13p6x22L +=  + Ab(m11,mu2) * (  - 16./3.*m11*m33*L8r*RS1463e - 8.
         /3.*m11*m33*L8r*RC146e + 16./3.*m11*m33*L7r + 8./3.*m11*m33*
         L5r*RS1463e - 16./9.*m11*m33*L5r*RSe6613 + 4./3.*m11*m33*L5r*
         RC146e - 2./3.*m11*m33*L3r*RS1463e - 5./3.*m11*m33*L3r*RC146e
          - 8./3.*m11*m33*L0r*RS1463e - 2./3.*m11*m33*L0r*RC146e - 16./
         9.*m11*m33*me2*L5r*RN1e*RSe6613 + 32./9.*m11*m33*m66*L7r*RN1e*
         RSe63 - 16./9.*m11*m33*m66*L7r*RN13*RS34e - 32./9.*m11*m33*m66
         *L7r*RN13*RS14e + 32./9.*m11*m33*m66*L6r*RN1e*RSe6613 - 16./9.
         *m11*m33*m66*L6r*RN13*RS3461e - 32./9.*m11*m33*m66*L6r*RN13*
         RS1463e - 16./27.*m11*m33*m66*L5r*RN13*RS14e*RS34e + 16./9.*
         m11*m33*m66*L4r*RN13*RS3461e + 16./9.*m11*m33*m66*L4r*RN13*
         RS1463e - 32./27.*m11*m33*m66*L4r*RN13*RS16e*RS36e - 16./27.*
         m11*m33*m66*L4r*RN13*RS14e*RS34e - 32./9.*m11*m33*m44*L7r*RN1e
         *RSe63 - 32./9.*m11*m33*m44*L7r*RN13*RS36e - 64./9.*m11*m33*
         m44*L7r*RN13*RS16e + 64./9.*m11*m33*m44*L6r*RN1e*RSe6613 - 32./
         9.*m11*m33*m44*L6r*RN13*RS3461e );

      M13p6x22L +=  + Ab(m11,mu2) * (  - 64./9.*m11*m33*m44*L6r*RN13*
         RS1463e - 32./27.*m11*m33*m44*L5r*RN13*RS16e*RS36e + 32./9.*
         m11*m33*m44*L4r*RN13*RS3461e + 32./9.*m11*m33*m44*L4r*RN13*
         RS1463e - 64./27.*m11*m33*m44*L4r*RN13*RS16e*RS36e - 32./27.*
         m11*m33*m44*L4r*RN13*RS14e*RS34e - 32./9.*m11*pow(m33,2)*L8r*
         RN13*RS1463e + 16./3.*m11*pow(m33,2)*L7r*RN13 + 16./9.*m11*
         pow(m33,2)*L5r*RN13*RS1463e - 32./9.*pow(m11,2)*L8r*RS1463e + 
         32./9.*pow(m11,2)*L8r*RSe6613 - 32./9.*pow(m11,2)*L8r*RSe6611
          - 16./9.*pow(m11,2)*L8r*RC146e + 8*pow(m11,2)*L7r + 8./9.*
         pow(m11,2)*L5r*RS3461e + 28./9.*pow(m11,2)*L5r*RS1463e - 32./9.
         *pow(m11,2)*L5r*RSe6613 + 8./9.*pow(m11,2)*L5r*RC146e - 2./3.*
         pow(m11,2)*L3r*RS1463e - 5./3.*pow(m11,2)*L3r*RC146e - 8./3.*
         pow(m11,2)*L0r*RS1463e - 2./3.*pow(m11,2)*L0r*RC146e + 32./9.*
         pow(m11,2)*me2*L8r*RN1e*RSe6613 - 32./9.*pow(m11,2)*me2*L8r*
         RN1e*RSe6611 - 32./9.*pow(m11,2)*me2*L5r*RN1e*RSe6613 + 32./9.
         *pow(m11,2)*m66*L7r*RN1e*RSe63 );

      M13p6x22L +=  + Ab(m11,mu2) * ( 32./9.*pow(m11,2)*m66*L7r*RN1e*
         RSe61 - 32./9.*pow(m11,2)*m66*L7r*RN13*RS34e + 32./9.*pow(
         m11,2)*m66*L6r*RN1e*RSe6613 + 32./9.*pow(m11,2)*m66*L6r*RN1e*
         RSe6611 - 32./9.*pow(m11,2)*m66*L6r*RN13*RS3461e - 8./27.*pow(
         m11,2)*m66*L5r*RN13*RS14e*RS34e + 8./9.*pow(m11,2)*m66*L4r*
         RN13*RS3461e + 8./9.*pow(m11,2)*m66*L4r*RN13*RS1463e - 16./27.
         *pow(m11,2)*m66*L4r*RN13*RS16e*RS36e - 8./27.*pow(m11,2)*m66*
         L4r*RN13*RS14e*RS34e - 32./9.*pow(m11,2)*m44*L7r*RN1e*RSe63 - 
         32./9.*pow(m11,2)*m44*L7r*RN1e*RSe61 - 64./9.*pow(m11,2)*m44*
         L7r*RN13*RS36e + 64./9.*pow(m11,2)*m44*L6r*RN1e*RSe6613 + 64./
         9.*pow(m11,2)*m44*L6r*RN1e*RSe6611 - 64./9.*pow(m11,2)*m44*L6r
         *RN13*RS3461e - 16./27.*pow(m11,2)*m44*L5r*RN13*RS16e*RS36e + 
         16./9.*pow(m11,2)*m44*L4r*RN13*RS3461e + 16./9.*pow(m11,2)*m44
         *L4r*RN13*RS1463e - 32./27.*pow(m11,2)*m44*L4r*RN13*RS16e*
         RS36e - 16./27.*pow(m11,2)*m44*L4r*RN13*RS14e*RS34e + 32./9.*
         pow(m11,2)*m33*L8r*RN1e*RSe6613 );

      M13p6x22L +=  + Ab(m11,mu2) * (  - 16./9.*pow(m11,2)*m33*L8r*RN13
         *RS3461e + 32./3.*pow(m11,2)*m33*L7r*RN13 + 16./9.*pow(m11,2)*
         m33*L5r*RN13*RS3461e + 8./9.*pow(m11,2)*m33*L5r*RN13*RS1463e
          + 32./9.*pow(m11,3)*L8r*RN1e*RSe6613 + 32./9.*pow(m11,3)*L8r*
         RN1e*RSe6611 - 32./9.*pow(m11,3)*L8r*RN13*RS3461e + 8./9.*pow(
         m11,3)*L5r*RN13*RS3461e );

      M13p6x22L +=  + Ab(m13,mu2) * ( 8*pow(m33,2)*L6r + 5*pow(m33,2)*
         L2r + 2*pow(m33,2)*L1r + 16*m11*m33*L6r + 10*m11*m33*L2r + 4*
         m11*m33*L1r + 8*pow(m11,2)*L6r + 5*pow(m11,2)*L2r + 2*pow(
         m11,2)*L1r );

      M13p6x22L +=  + Ab(m14,mu2) * ( 8*m33*m44*L8r - 4*m33*m44*L5r + 5
         *m33*m44*L3r + 2*m33*m44*L0r + 8*m11*m44*L8r - 4*m11*m44*L5r
          + 5*m11*m44*L3r + 2*m11*m44*L0r + 8*m11*m33*L8r - 4*m11*m33*
         L5r + 5*m11*m33*L3r + 2*m11*m33*L0r + 8*pow(m11,2)*L8r - 4*
         pow(m11,2)*L5r + 5*pow(m11,2)*L3r + 2*pow(m11,2)*L0r );

      M13p6x22L +=  + Ab(m16,mu2) * ( 4*m33*m66*L8r - 2*m33*m66*L5r + 5.
         /2.*m33*m66*L3r + m33*m66*L0r + 4*m11*m66*L8r - 2*m11*m66*L5r
          + 5./2.*m11*m66*L3r + m11*m66*L0r + 4*m11*m33*L8r - 2*m11*m33
         *L5r + 5./2.*m11*m33*L3r + m11*m33*L0r + 4*pow(m11,2)*L8r - 2*
         pow(m11,2)*L5r + 5./2.*pow(m11,2)*L3r + pow(m11,2)*L0r );

      M13p6x22L +=  + Ab(m33,mu2) * (  - 8./9.*m66*L4r*RD346e + 32./27.
         *pow(m66,2)*L8r*RS34e*RSe63 - 8./27.*pow(m66,2)*L8r*pow(
         RS34e,2) - 32./27.*pow(m66,2)*L8r*RSe61*RS34e + 16./27.*pow(
         m66,2)*L8r*RS14e*RS34e + 32./27.*pow(m66,2)*L7r*RS34e*RSe63 - 
         8./27.*pow(m66,2)*L7r*pow(RS34e,2) - 32./27.*pow(m66,2)*L7r*
         RSe61*RS34e + 16./27.*pow(m66,2)*L7r*RS14e*RS34e + 32./27.*
         pow(m66,2)*L6r*RS34e*RSe63 - 8./27.*pow(m66,2)*L6r*pow(
         RS34e,2) - 32./27.*pow(m66,2)*L6r*RSe61*RS34e + 16./27.*pow(
         m66,2)*L6r*RS14e*RS34e + 32./27.*pow(m66,2)*me2*L8r*RN3e*RS34e
         *RSe63 - 32./27.*pow(m66,2)*me2*L8r*RSe61*RN3e*RS34e + 32./27.
         *pow(m66,2)*me2*L7r*RN3e*RS34e*RSe63 - 32./27.*pow(m66,2)*me2*
         L7r*RSe61*RN3e*RS34e + 32./27.*pow(m66,2)*me2*L6r*RN3e*RS34e*
         RSe63 - 32./27.*pow(m66,2)*me2*L6r*RSe61*RN3e*RS34e - 16./9.*
         m44*L4r*RD346e + 64./27.*m44*m66*L7r*RS36e*RSe63 - 32./27.*m44
         *m66*L7r*RS34e*RSe63 - 32./27.*m44*m66*L7r*RS34e*RS36e - 64./
         27.*m44*m66*L7r*RSe61*RS36e );

      M13p6x22L +=  + Ab(m33,mu2) * ( 32./27.*m44*m66*L7r*RSe61*RS34e
          + 32./27.*m44*m66*L7r*RS16e*RS34e + 32./27.*m44*m66*L7r*RS14e
         *RS36e - 32./27.*m44*m66*L6r*RS36e*RSe63 - 16./27.*m44*m66*L6r
         *pow(RS36e,2) + 64./27.*m44*m66*L6r*RS34e*RSe63 - 16./27.*m44*
         m66*L6r*pow(RS34e,2) + 32./27.*m44*m66*L6r*RSe61*RS36e - 64./
         27.*m44*m66*L6r*RSe61*RS34e + 32./27.*m44*m66*L6r*RS16e*RS36e
          + 32./27.*m44*m66*L6r*RS14e*RS34e + 64./27.*m44*m66*me2*L7r*
         RN3e*RS36e*RSe63 - 32./27.*m44*m66*me2*L7r*RN3e*RS34e*RSe63 - 
         64./27.*m44*m66*me2*L7r*RSe61*RN3e*RS36e + 32./27.*m44*m66*me2
         *L7r*RSe61*RN3e*RS34e - 32./27.*m44*m66*me2*L6r*RN3e*RS36e*
         RSe63 + 64./27.*m44*m66*me2*L6r*RN3e*RS34e*RSe63 + 32./27.*m44
         *m66*me2*L6r*RSe61*RN3e*RS36e - 64./27.*m44*m66*me2*L6r*RSe61*
         RN3e*RS34e - 32./27.*pow(m44,2)*L8r*RS36e*RSe63 - 16./27.*pow(
         m44,2)*L8r*pow(RS36e,2) + 32./27.*pow(m44,2)*L8r*RSe61*RS36e
          + 32./27.*pow(m44,2)*L8r*RS16e*RS36e - 64./27.*pow(m44,2)*L7r
         *RS36e*RSe63 );

      M13p6x22L +=  + Ab(m33,mu2) * (  - 32./27.*pow(m44,2)*L7r*pow(
         RS36e,2) + 64./27.*pow(m44,2)*L7r*RSe61*RS36e + 64./27.*pow(
         m44,2)*L7r*RS16e*RS36e - 64./27.*pow(m44,2)*L6r*RS36e*RSe63 - 
         32./27.*pow(m44,2)*L6r*pow(RS36e,2) + 64./27.*pow(m44,2)*L6r*
         RSe61*RS36e + 64./27.*pow(m44,2)*L6r*RS16e*RS36e - 32./27.*
         pow(m44,2)*me2*L8r*RN3e*RS36e*RSe63 + 32./27.*pow(m44,2)*me2*
         L8r*RSe61*RN3e*RS36e - 64./27.*pow(m44,2)*me2*L7r*RN3e*RS36e*
         RSe63 + 64./27.*pow(m44,2)*me2*L7r*RSe61*RN3e*RS36e - 64./27.*
         pow(m44,2)*me2*L6r*RN3e*RS36e*RSe63 + 64./27.*pow(m44,2)*me2*
         L6r*RSe61*RN3e*RS36e - 5./3.*m33*L3r*RD346e - 2./3.*m33*L0r*
         RD346e - 32./9.*m33*m66*L7r*RSe63 - 16./9.*m33*m66*L7r*RS34e
          + 32./9.*m33*m66*L7r*RSe61 - 16./9.*m33*m66*L7r*RS14e - 32./9.
         *m33*m66*L6r*RS3461e - 16./9.*m33*m66*L6r*RS1463e + 32./9.*m33
         *m66*L6r*RSe6613 - 32./9.*m33*m66*L6r*RSe6633 + 8./9.*m33*m66*
         L6r*RC346e + 4./27.*m33*m66*L5r*pow(RS34e,2) + 32./27.*m33*m66
         *L5r*RSe61*RS34e );

      M13p6x22L +=  + Ab(m33,mu2) * (  - 16./27.*m33*m66*L5r*RS14e*
         RS34e + 16./3.*m33*m66*L4r*RS3461e + 16./9.*m33*m66*L4r*
         RS1463e - 32./9.*m33*m66*L4r*RSe6613 + 8./27.*m33*m66*L4r*pow(
         RS36e,2) + 4./27.*m33*m66*L4r*pow(RS34e,2) - 4./9.*m33*m66*L4r
         *RC346e - 32./27.*m33*m66*L4r*RSe61*RS36e + 32./27.*m33*m66*
         L4r*RSe61*RS34e - 32./27.*m33*m66*L4r*RS16e*RS36e - 16./27.*
         m33*m66*L4r*RS14e*RS34e - 32./9.*m33*m66*me2*L7r*RN3e*RSe63 + 
         32./9.*m33*m66*me2*L7r*RSe61*RN3e + 32./9.*m33*m66*me2*L6r*
         RN3e*RSe6613 - 32./9.*m33*m66*me2*L6r*RN3e*RSe6633 + 32./27.*
         m33*m66*me2*L5r*RSe61*RN3e*RS34e - 32./9.*m33*m66*me2*L4r*RN3e
         *RSe6613 - 32./27.*m33*m66*me2*L4r*RSe61*RN3e*RS36e + 32./27.*
         m33*m66*me2*L4r*RSe61*RN3e*RS34e - 32./27.*m33*pow(m66,2)*L8r*
         RN3e*RS34e*RSe63 - 32./27.*m33*pow(m66,2)*L8r*RSe61*RN3e*RS34e
          - 16./27.*m33*pow(m66,2)*L8r*RN13*RS14e*RS34e - 32./27.*m33*
         pow(m66,2)*L7r*RN3e*RS34e*RSe63 - 32./27.*m33*pow(m66,2)*L7r*
         RSe61*RN3e*RS34e );

      M13p6x22L +=  + Ab(m33,mu2) * (  - 16./27.*m33*pow(m66,2)*L7r*
         RN13*RS14e*RS34e - 32./27.*m33*pow(m66,2)*L6r*RN3e*RS34e*RSe63
          - 32./27.*m33*pow(m66,2)*L6r*RSe61*RN3e*RS34e - 16./27.*m33*
         pow(m66,2)*L6r*RN13*RS14e*RS34e + 32./9.*m33*m44*L7r*RSe63 - 
         32./9.*m33*m44*L7r*RS36e - 32./9.*m33*m44*L7r*RSe61 - 32./9.*
         m33*m44*L7r*RS16e - 64./9.*m33*m44*L6r*RS3461e - 32./9.*m33*
         m44*L6r*RS1463e + 64./9.*m33*m44*L6r*RSe6613 - 64./9.*m33*m44*
         L6r*RSe6633 + 16./9.*m33*m44*L6r*RC346e + 8./27.*m33*m44*L5r*
         pow(RS36e,2) - 32./27.*m33*m44*L5r*RSe61*RS36e - 32./27.*m33*
         m44*L5r*RS16e*RS36e + 32./3.*m33*m44*L4r*RS3461e + 32./9.*m33*
         m44*L4r*RS1463e - 64./9.*m33*m44*L4r*RSe6613 + 16./27.*m33*m44
         *L4r*pow(RS36e,2) + 8./27.*m33*m44*L4r*pow(RS34e,2) - 8./9.*
         m33*m44*L4r*RC346e - 64./27.*m33*m44*L4r*RSe61*RS36e + 64./27.
         *m33*m44*L4r*RSe61*RS34e - 64./27.*m33*m44*L4r*RS16e*RS36e - 
         32./27.*m33*m44*L4r*RS14e*RS34e + 32./9.*m33*m44*me2*L7r*RN3e*
         RSe63 );

      M13p6x22L +=  + Ab(m33,mu2) * (  - 32./9.*m33*m44*me2*L7r*RSe61*
         RN3e + 64./9.*m33*m44*me2*L6r*RN3e*RSe6613 - 64./9.*m33*m44*
         me2*L6r*RN3e*RSe6633 - 32./27.*m33*m44*me2*L5r*RSe61*RN3e*
         RS36e - 64./9.*m33*m44*me2*L4r*RN3e*RSe6613 - 64./27.*m33*m44*
         me2*L4r*RSe61*RN3e*RS36e + 64./27.*m33*m44*me2*L4r*RSe61*RN3e*
         RS34e - 64./27.*m33*m44*m66*L7r*RN3e*RS36e*RSe63 + 32./27.*m33
         *m44*m66*L7r*RN3e*RS34e*RSe63 - 64./27.*m33*m44*m66*L7r*RSe61*
         RN3e*RS36e + 32./27.*m33*m44*m66*L7r*RSe61*RN3e*RS34e - 32./27.
         *m33*m44*m66*L7r*RN13*RS16e*RS34e - 32./27.*m33*m44*m66*L7r*
         RN13*RS14e*RS36e + 32./27.*m33*m44*m66*L6r*RN3e*RS36e*RSe63 - 
         64./27.*m33*m44*m66*L6r*RN3e*RS34e*RSe63 + 32./27.*m33*m44*m66
         *L6r*RSe61*RN3e*RS36e - 64./27.*m33*m44*m66*L6r*RSe61*RN3e*
         RS34e - 32./27.*m33*m44*m66*L6r*RN13*RS16e*RS36e - 32./27.*m33
         *m44*m66*L6r*RN13*RS14e*RS34e + 32./27.*m33*pow(m44,2)*L8r*
         RN3e*RS36e*RSe63 + 32./27.*m33*pow(m44,2)*L8r*RSe61*RN3e*RS36e
          - 32./27.*m33*pow(m44,2)*L8r*RN13*RS16e*RS36e );

      M13p6x22L +=  + Ab(m33,mu2) * ( 64./27.*m33*pow(m44,2)*L7r*RN3e*
         RS36e*RSe63 + 64./27.*m33*pow(m44,2)*L7r*RSe61*RN3e*RS36e - 64.
         /27.*m33*pow(m44,2)*L7r*RN13*RS16e*RS36e + 64./27.*m33*pow(
         m44,2)*L6r*RN3e*RS36e*RSe63 + 64./27.*m33*pow(m44,2)*L6r*RSe61
         *RN3e*RS36e - 64./27.*m33*pow(m44,2)*L6r*RN13*RS16e*RS36e - 32.
         /9.*pow(m33,2)*L8r*RS3461e - 16./9.*pow(m33,2)*L8r*RS1463e + 
         32./9.*pow(m33,2)*L8r*RSe6613 - 32./9.*pow(m33,2)*L8r*RSe6633
          - 16./9.*pow(m33,2)*L8r*RC346e + 8*pow(m33,2)*L7r + 28./9.*
         pow(m33,2)*L5r*RS3461e + 16./9.*pow(m33,2)*L5r*RS1463e - 32./9.
         *pow(m33,2)*L5r*RSe6613 + 8./9.*pow(m33,2)*L5r*RC346e - 2./3.*
         pow(m33,2)*L3r*RS3461e - 5./3.*pow(m33,2)*L3r*RC346e - 8./3.*
         pow(m33,2)*L0r*RS3461e - 2./3.*pow(m33,2)*L0r*RC346e + 32./9.*
         pow(m33,2)*me2*L8r*RN3e*RSe6613 - 32./9.*pow(m33,2)*me2*L8r*
         RN3e*RSe6633 - 32./9.*pow(m33,2)*me2*L5r*RN3e*RSe6613 + 32./9.
         *pow(m33,2)*m66*L7r*RN3e*RSe63 + 32./9.*pow(m33,2)*m66*L7r*
         RSe61*RN3e );

      M13p6x22L +=  + Ab(m33,mu2) * ( 16./9.*pow(m33,2)*m66*L7r*RN13*
         RS14e + 32./9.*pow(m33,2)*m66*L6r*RN3e*RSe6613 + 32./9.*pow(
         m33,2)*m66*L6r*RN3e*RSe6633 + 16./9.*pow(m33,2)*m66*L6r*RN13*
         RS1463e - 32./9.*pow(m33,2)*m44*L7r*RN3e*RSe63 - 32./9.*pow(
         m33,2)*m44*L7r*RSe61*RN3e + 32./9.*pow(m33,2)*m44*L7r*RN13*
         RS16e + 64./9.*pow(m33,2)*m44*L6r*RN3e*RSe6613 + 64./9.*pow(
         m33,2)*m44*L6r*RN3e*RSe6633 + 32./9.*pow(m33,2)*m44*L6r*RN13*
         RS1463e + 32./9.*pow(m33,3)*L8r*RN3e*RSe6613 + 32./9.*pow(
         m33,3)*L8r*RN3e*RSe6633 + 16./9.*pow(m33,3)*L8r*RN13*RS1463e
          + 4./9.*m11*L5r*RD346e - 5./3.*m11*L3r*RD346e - 2./3.*m11*L0r
         *RD346e - 32./9.*m11*m66*L7r*RS34e - 32./9.*m11*m66*L6r*
         RS3461e + 16./27.*m11*m66*L5r*RSe61*RS34e - 8./27.*m11*m66*L5r
         *RS14e*RS34e + 8./3.*m11*m66*L4r*RS3461e + 8./9.*m11*m66*L4r*
         RS1463e - 16./9.*m11*m66*L4r*RSe6613 - 16./27.*m11*m66*L4r*
         RSe61*RS36e + 16./27.*m11*m66*L4r*RSe61*RS34e - 16./27.*m11*
         m66*L4r*RS16e*RS36e );

      M13p6x22L +=  + Ab(m33,mu2) * (  - 8./27.*m11*m66*L4r*RS14e*RS34e
          + 16./27.*m11*m66*me2*L5r*RSe61*RN3e*RS34e - 16./9.*m11*m66*
         me2*L4r*RN3e*RSe6613 - 16./27.*m11*m66*me2*L4r*RSe61*RN3e*
         RS36e + 16./27.*m11*m66*me2*L4r*RSe61*RN3e*RS34e - 32./27.*m11
         *pow(m66,2)*L8r*RSe61*RN3e*RS34e - 32./27.*m11*pow(m66,2)*L8r*
         RN13*RS14e*RS34e - 32./27.*m11*pow(m66,2)*L7r*RSe61*RN3e*RS34e
          - 32./27.*m11*pow(m66,2)*L7r*RN13*RS14e*RS34e - 32./27.*m11*
         pow(m66,2)*L6r*RSe61*RN3e*RS34e - 32./27.*m11*pow(m66,2)*L6r*
         RN13*RS14e*RS34e - 64./9.*m11*m44*L7r*RS36e - 64./9.*m11*m44*
         L6r*RS3461e - 16./27.*m11*m44*L5r*RSe61*RS36e - 16./27.*m11*
         m44*L5r*RS16e*RS36e + 16./3.*m11*m44*L4r*RS3461e + 16./9.*m11*
         m44*L4r*RS1463e - 32./9.*m11*m44*L4r*RSe6613 - 32./27.*m11*m44
         *L4r*RSe61*RS36e + 32./27.*m11*m44*L4r*RSe61*RS34e - 32./27.*
         m11*m44*L4r*RS16e*RS36e - 16./27.*m11*m44*L4r*RS14e*RS34e - 16.
         /27.*m11*m44*me2*L5r*RSe61*RN3e*RS36e - 32./9.*m11*m44*me2*L4r
         *RN3e*RSe6613 );

      M13p6x22L +=  + Ab(m33,mu2) * (  - 32./27.*m11*m44*me2*L4r*RSe61*
         RN3e*RS36e + 32./27.*m11*m44*me2*L4r*RSe61*RN3e*RS34e - 64./27.
         *m11*m44*m66*L7r*RSe61*RN3e*RS36e + 32./27.*m11*m44*m66*L7r*
         RSe61*RN3e*RS34e - 64./27.*m11*m44*m66*L7r*RN13*RS16e*RS34e - 
         64./27.*m11*m44*m66*L7r*RN13*RS14e*RS36e + 32./27.*m11*m44*m66
         *L6r*RSe61*RN3e*RS36e - 64./27.*m11*m44*m66*L6r*RSe61*RN3e*
         RS34e - 64./27.*m11*m44*m66*L6r*RN13*RS16e*RS36e - 64./27.*m11
         *m44*m66*L6r*RN13*RS14e*RS34e + 32./27.*m11*pow(m44,2)*L8r*
         RSe61*RN3e*RS36e - 64./27.*m11*pow(m44,2)*L8r*RN13*RS16e*RS36e
          + 64./27.*m11*pow(m44,2)*L7r*RSe61*RN3e*RS36e - 128./27.*m11*
         pow(m44,2)*L7r*RN13*RS16e*RS36e + 64./27.*m11*pow(m44,2)*L6r*
         RSe61*RN3e*RS36e - 128./27.*m11*pow(m44,2)*L6r*RN13*RS16e*
         RS36e - 16./3.*m11*m33*L8r*RS3461e - 8./3.*m11*m33*L8r*RC346e
          + 32./3.*m11*m33*L7r + 32./9.*m11*m33*L5r*RS3461e + 8./9.*m11
         *m33*L5r*RS1463e - 16./9.*m11*m33*L5r*RSe6613 + 4./3.*m11*m33*
         L5r*RC346e );

      M13p6x22L +=  + Ab(m33,mu2) * (  - 2./3.*m11*m33*L3r*RS3461e - 5./
         3.*m11*m33*L3r*RC346e - 8./3.*m11*m33*L0r*RS3461e - 2./3.*m11*
         m33*L0r*RC346e - 16./9.*m11*m33*me2*L5r*RN3e*RSe6613 + 32./9.*
         m11*m33*m66*L7r*RSe61*RN3e + 16./9.*m11*m33*m66*L7r*RN13*RS34e
          + 32./9.*m11*m33*m66*L7r*RN13*RS14e + 32./9.*m11*m33*m66*L6r*
         RN3e*RSe6613 + 16./9.*m11*m33*m66*L6r*RN13*RS3461e + 32./9.*
         m11*m33*m66*L6r*RN13*RS1463e + 16./27.*m11*m33*m66*L5r*RN13*
         RS14e*RS34e - 16./9.*m11*m33*m66*L4r*RN13*RS3461e - 16./9.*m11
         *m33*m66*L4r*RN13*RS1463e + 32./27.*m11*m33*m66*L4r*RN13*RS16e
         *RS36e + 16./27.*m11*m33*m66*L4r*RN13*RS14e*RS34e - 32./9.*m11
         *m33*m44*L7r*RSe61*RN3e + 32./9.*m11*m33*m44*L7r*RN13*RS36e + 
         64./9.*m11*m33*m44*L7r*RN13*RS16e + 64./9.*m11*m33*m44*L6r*
         RN3e*RSe6613 + 32./9.*m11*m33*m44*L6r*RN13*RS3461e + 64./9.*
         m11*m33*m44*L6r*RN13*RS1463e + 32./27.*m11*m33*m44*L5r*RN13*
         RS16e*RS36e - 32./9.*m11*m33*m44*L4r*RN13*RS3461e - 32./9.*m11
         *m33*m44*L4r*RN13*RS1463e );

      M13p6x22L +=  + Ab(m33,mu2) * ( 64./27.*m11*m33*m44*L4r*RN13*
         RS16e*RS36e + 32./27.*m11*m33*m44*L4r*RN13*RS14e*RS34e + 32./9.
         *m11*pow(m33,2)*L8r*RN3e*RSe6613 + 32./9.*m11*pow(m33,2)*L8r*
         RN13*RS1463e - 16./3.*m11*pow(m33,2)*L7r*RN13 - 16./9.*m11*
         pow(m33,2)*L5r*RN13*RS1463e - 32./9.*pow(m11,2)*L8r*RS3461e + 
         4./3.*pow(m11,2)*L5r*RS3461e + 32./9.*pow(m11,2)*m66*L7r*RN13*
         RS34e + 32./9.*pow(m11,2)*m66*L6r*RN13*RS3461e + 8./27.*pow(
         m11,2)*m66*L5r*RN13*RS14e*RS34e - 8./9.*pow(m11,2)*m66*L4r*
         RN13*RS3461e - 8./9.*pow(m11,2)*m66*L4r*RN13*RS1463e + 16./27.
         *pow(m11,2)*m66*L4r*RN13*RS16e*RS36e + 8./27.*pow(m11,2)*m66*
         L4r*RN13*RS14e*RS34e + 64./9.*pow(m11,2)*m44*L7r*RN13*RS36e + 
         64./9.*pow(m11,2)*m44*L6r*RN13*RS3461e + 16./27.*pow(m11,2)*
         m44*L5r*RN13*RS16e*RS36e - 16./9.*pow(m11,2)*m44*L4r*RN13*
         RS3461e - 16./9.*pow(m11,2)*m44*L4r*RN13*RS1463e + 32./27.*
         pow(m11,2)*m44*L4r*RN13*RS16e*RS36e + 16./27.*pow(m11,2)*m44*
         L4r*RN13*RS14e*RS34e );

      M13p6x22L +=  + Ab(m33,mu2) * ( 16./9.*pow(m11,2)*m33*L8r*RN13*
         RS3461e - 32./3.*pow(m11,2)*m33*L7r*RN13 - 16./9.*pow(m11,2)*
         m33*L5r*RN13*RS3461e - 8./9.*pow(m11,2)*m33*L5r*RN13*RS1463e
          + 32./9.*pow(m11,3)*L8r*RN13*RS3461e - 8./9.*pow(m11,3)*L5r*
         RN13*RS3461e );

      M13p6x22L +=  + Ab(m34,mu2) * ( 8*m33*m44*L8r - 4*m33*m44*L5r + 5
         *m33*m44*L3r + 2*m33*m44*L0r + 8*pow(m33,2)*L8r - 4*pow(m33,2)
         *L5r + 5*pow(m33,2)*L3r + 2*pow(m33,2)*L0r + 8*m11*m44*L8r - 4
         *m11*m44*L5r + 5*m11*m44*L3r + 2*m11*m44*L0r + 8*m11*m33*L8r
          - 4*m11*m33*L5r + 5*m11*m33*L3r + 2*m11*m33*L0r );

      M13p6x22L +=  + Ab(m36,mu2) * ( 4*m33*m66*L8r - 2*m33*m66*L5r + 5.
         /2.*m33*m66*L3r + m33*m66*L0r + 4*pow(m33,2)*L8r - 2*pow(
         m33,2)*L5r + 5./2.*pow(m33,2)*L3r + pow(m33,2)*L0r + 4*m11*m66
         *L8r - 2*m11*m66*L5r + 5./2.*m11*m66*L3r + m11*m66*L0r + 4*m11
         *m33*L8r - 2*m11*m33*L5r + 5./2.*m11*m33*L3r + m11*m33*L0r );

      M13p6x22L +=  + Ab(m44,mu2) * ( 24*m33*m44*L6r - 24*m33*m44*L4r
          + 6*m33*m44*L2r + 24*m33*m44*L1r + 24*m11*m44*L6r - 24*m11*
         m44*L4r + 6*m11*m44*L2r + 24*m11*m44*L1r );

      M13p6x22L +=  + Ab(m46,mu2) * ( 16*m33*m66*L6r - 16*m33*m66*L4r
          + 4*m33*m66*L2r + 16*m33*m66*L1r + 16*m33*m44*L6r - 16*m33*
         m44*L4r + 4*m33*m44*L2r + 16*m33*m44*L1r + 16*m11*m66*L6r - 16
         *m11*m66*L4r + 4*m11*m66*L2r + 16*m11*m66*L1r + 16*m11*m44*L6r
          - 16*m11*m44*L4r + 4*m11*m44*L2r + 16*m11*m44*L1r );

      M13p6x22L +=  + Ab(me2,mu2) * ( 32./27.*m66*me2*L5r*pow(RSe63,2)
          - 16./27.*m66*me2*L5r*RS34e*RSe63 - 64./27.*m66*me2*L5r*RSe61
         *RSe63 + 16./27.*m66*me2*L5r*RSe61*RS34e + 32./27.*m66*me2*L5r
         *pow(RSe61,2) + 16./27.*m66*me2*L5r*RS14e*RSe63 - 16./27.*m66*
         me2*L5r*RS14e*RSe61 - 32./9.*m66*me2*L4r*RSe6613 + 16./9.*m66*
         me2*L4r*pow(RSe63,2) + 16./9.*m66*me2*L4r*RSe6633 + 16./27.*
         m66*me2*L4r*RS36e*RSe63 - 16./27.*m66*me2*L4r*RS34e*RSe63 - 32.
         /9.*m66*me2*L4r*RSe61*RSe63 - 16./27.*m66*me2*L4r*RSe61*RS36e
          + 16./27.*m66*me2*L4r*RSe61*RS34e + 16./9.*m66*me2*L4r*pow(
         RSe61,2) + 16./9.*m66*me2*L4r*RSe6611 - 16./27.*m66*me2*L4r*
         RS16e*RSe63 + 16./27.*m66*me2*L4r*RS16e*RSe61 + 16./27.*m66*
         me2*L4r*RS14e*RSe63 - 16./27.*m66*me2*L4r*RS14e*RSe61 - 32./27.
         *pow(m66,2)*L8r*pow(RSe63,2) + 64./27.*pow(m66,2)*L8r*RSe61*
         RSe63 - 32./27.*pow(m66,2)*L8r*pow(RSe61,2) - 32./27.*pow(
         m66,2)*L7r*pow(RSe63,2) + 64./27.*pow(m66,2)*L7r*RSe61*RSe63
          - 32./27.*pow(m66,2)*L7r*pow(RSe61,2) );

      M13p6x22L +=  + Ab(me2,mu2) * (  - 32./27.*pow(m66,2)*L6r*pow(
         RSe63,2) + 64./27.*pow(m66,2)*L6r*RSe61*RSe63 - 32./27.*pow(
         m66,2)*L6r*pow(RSe61,2) - 32./27.*pow(m66,2)*me2*L8r*RN3e*
         RS34e*RSe63 + 32./27.*pow(m66,2)*me2*L8r*RSe61*RN3e*RS34e + 32.
         /27.*pow(m66,2)*me2*L8r*RN1e*RS14e*RSe63 - 32./27.*pow(m66,2)*
         me2*L8r*RN1e*RS14e*RSe61 - 32./27.*pow(m66,2)*me2*L7r*RN3e*
         RS34e*RSe63 + 32./27.*pow(m66,2)*me2*L7r*RSe61*RN3e*RS34e + 32.
         /27.*pow(m66,2)*me2*L7r*RN1e*RS14e*RSe63 - 32./27.*pow(m66,2)*
         me2*L7r*RN1e*RS14e*RSe61 - 32./27.*pow(m66,2)*me2*L6r*RN3e*
         RS34e*RSe63 + 32./27.*pow(m66,2)*me2*L6r*RSe61*RN3e*RS34e + 32.
         /27.*pow(m66,2)*me2*L6r*RN1e*RS14e*RSe63 - 32./27.*pow(m66,2)*
         me2*L6r*RN1e*RS14e*RSe61 + 16./27.*m44*me2*L5r*pow(RSe63,2) + 
         16./27.*m44*me2*L5r*RS36e*RSe63 - 32./27.*m44*me2*L5r*RSe61*
         RSe63 - 16./27.*m44*me2*L5r*RSe61*RS36e + 16./27.*m44*me2*L5r*
         pow(RSe61,2) - 16./27.*m44*me2*L5r*RS16e*RSe63 + 16./27.*m44*
         me2*L5r*RS16e*RSe61 );

      M13p6x22L +=  + Ab(me2,mu2) * (  - 64./9.*m44*me2*L4r*RSe6613 + 
         32./9.*m44*me2*L4r*pow(RSe63,2) + 32./9.*m44*me2*L4r*RSe6633
          + 32./27.*m44*me2*L4r*RS36e*RSe63 - 32./27.*m44*me2*L4r*RS34e
         *RSe63 - 64./9.*m44*me2*L4r*RSe61*RSe63 - 32./27.*m44*me2*L4r*
         RSe61*RS36e + 32./27.*m44*me2*L4r*RSe61*RS34e + 32./9.*m44*me2
         *L4r*pow(RSe61,2) + 32./9.*m44*me2*L4r*RSe6611 - 32./27.*m44*
         me2*L4r*RS16e*RSe63 + 32./27.*m44*me2*L4r*RS16e*RSe61 + 32./27.
         *m44*me2*L4r*RS14e*RSe63 - 32./27.*m44*me2*L4r*RS14e*RSe61 + 
         64./27.*m44*m66*L7r*pow(RSe63,2) - 128./27.*m44*m66*L7r*RSe61*
         RSe63 + 64./27.*m44*m66*L7r*pow(RSe61,2) - 80./27.*m44*m66*L6r
         *pow(RSe63,2) + 160./27.*m44*m66*L6r*RSe61*RSe63 - 80./27.*m44
         *m66*L6r*pow(RSe61,2) - 64./27.*m44*m66*me2*L7r*RN3e*RS36e*
         RSe63 + 32./27.*m44*m66*me2*L7r*RN3e*RS34e*RSe63 + 64./27.*m44
         *m66*me2*L7r*RSe61*RN3e*RS36e - 32./27.*m44*m66*me2*L7r*RSe61*
         RN3e*RS34e + 64./27.*m44*m66*me2*L7r*RN1e*RS16e*RSe63 - 64./27.
         *m44*m66*me2*L7r*RN1e*RS16e*RSe61 );

      M13p6x22L +=  + Ab(me2,mu2) * (  - 32./27.*m44*m66*me2*L7r*RN1e*
         RS14e*RSe63 + 32./27.*m44*m66*me2*L7r*RN1e*RS14e*RSe61 + 32./
         27.*m44*m66*me2*L6r*RN3e*RS36e*RSe63 - 64./27.*m44*m66*me2*L6r
         *RN3e*RS34e*RSe63 - 32./27.*m44*m66*me2*L6r*RSe61*RN3e*RS36e
          + 64./27.*m44*m66*me2*L6r*RSe61*RN3e*RS34e - 32./27.*m44*m66*
         me2*L6r*RN1e*RS16e*RSe63 + 32./27.*m44*m66*me2*L6r*RN1e*RS16e*
         RSe61 + 64./27.*m44*m66*me2*L6r*RN1e*RS14e*RSe63 - 64./27.*m44
         *m66*me2*L6r*RN1e*RS14e*RSe61 - 16./27.*pow(m44,2)*L8r*pow(
         RSe63,2) + 32./27.*pow(m44,2)*L8r*RSe61*RSe63 - 16./27.*pow(
         m44,2)*L8r*pow(RSe61,2) - 32./27.*pow(m44,2)*L7r*pow(RSe63,2)
          + 64./27.*pow(m44,2)*L7r*RSe61*RSe63 - 32./27.*pow(m44,2)*L7r
         *pow(RSe61,2) - 32./27.*pow(m44,2)*L6r*pow(RSe63,2) + 64./27.*
         pow(m44,2)*L6r*RSe61*RSe63 - 32./27.*pow(m44,2)*L6r*pow(
         RSe61,2) + 32./27.*pow(m44,2)*me2*L8r*RN3e*RS36e*RSe63 - 32./
         27.*pow(m44,2)*me2*L8r*RSe61*RN3e*RS36e - 32./27.*pow(m44,2)*
         me2*L8r*RN1e*RS16e*RSe63 );

      M13p6x22L +=  + Ab(me2,mu2) * ( 32./27.*pow(m44,2)*me2*L8r*RN1e*
         RS16e*RSe61 + 64./27.*pow(m44,2)*me2*L7r*RN3e*RS36e*RSe63 - 64.
         /27.*pow(m44,2)*me2*L7r*RSe61*RN3e*RS36e - 64./27.*pow(m44,2)*
         me2*L7r*RN1e*RS16e*RSe63 + 64./27.*pow(m44,2)*me2*L7r*RN1e*
         RS16e*RSe61 + 64./27.*pow(m44,2)*me2*L6r*RN3e*RS36e*RSe63 - 64.
         /27.*pow(m44,2)*me2*L6r*RSe61*RN3e*RS36e - 64./27.*pow(m44,2)*
         me2*L6r*RN1e*RS16e*RSe63 + 64./27.*pow(m44,2)*me2*L6r*RN1e*
         RS16e*RSe61 - 16./9.*m33*me2*L5r*RSe6613 - 8./9.*m33*me2*L5r*
         RSe6611 - 4*m33*me2*L4r + 4./3.*m33*me2*L3r*RSe6613 + 10./3.*
         m33*me2*L3r*RSe6633 + 10./3.*m33*me2*L3r*RSe6611 + 2*m33*me2*
         L2r + 8*m33*me2*L1r + 16./3.*m33*me2*L0r*RSe6613 + 4./3.*m33*
         me2*L0r*RSe6633 + 4./3.*m33*me2*L0r*RSe6611 + 64./9.*m33*m66*
         L7r*RSe63 + 32./9.*m33*m66*L7r*RSe61 + 16./3.*m33*m66*L6r + 32.
         /9.*m33*m66*L6r*RSe6613 + 16./9.*m33*m66*L6r*RSe6633 - 16./27.
         *m33*m66*L5r*pow(RSe63,2) - 32./27.*m33*m66*L5r*RSe61*RSe63 - 
         8./3.*m33*m66*L4r );

      M13p6x22L +=  + Ab(me2,mu2) * (  - 16./9.*m33*m66*L4r*RSe6613 - 8.
         /9.*m33*m66*L4r*pow(RSe63,2) - 8./9.*m33*m66*L4r*RSe6633 - 16./
         9.*m33*m66*L4r*RSe61*RSe63 + 32./9.*m33*m66*me2*L7r*RN3e*RSe63
          - 32./9.*m33*m66*me2*L7r*RSe61*RN3e - 32./9.*m33*m66*me2*L6r*
         RN3e*RSe6613 + 32./9.*m33*m66*me2*L6r*RN3e*RSe6633 - 32./27.*
         m33*m66*me2*L5r*RSe61*RN3e*RS34e - 16./27.*m33*m66*me2*L5r*
         RN1e*RS14e*RSe63 + 32./9.*m33*m66*me2*L4r*RN3e*RSe6613 + 32./
         27.*m33*m66*me2*L4r*RSe61*RN3e*RS36e - 32./27.*m33*m66*me2*L4r
         *RSe61*RN3e*RS34e + 16./9.*m33*m66*me2*L4r*RN1e*RSe6613 + 16./
         27.*m33*m66*me2*L4r*RN1e*RS16e*RSe63 - 16./27.*m33*m66*me2*L4r
         *RN1e*RS14e*RSe63 + 32./27.*m33*pow(m66,2)*L8r*RN3e*RS34e*
         RSe63 + 32./27.*m33*pow(m66,2)*L8r*RSe61*RN3e*RS34e + 32./27.*
         m33*pow(m66,2)*L8r*RN1e*RS14e*RSe63 + 32./27.*m33*pow(m66,2)*
         L7r*RN3e*RS34e*RSe63 + 32./27.*m33*pow(m66,2)*L7r*RSe61*RN3e*
         RS34e + 32./27.*m33*pow(m66,2)*L7r*RN1e*RS14e*RSe63 + 32./27.*
         m33*pow(m66,2)*L6r*RN3e*RS34e*RSe63 );

      M13p6x22L +=  + Ab(me2,mu2) * ( 32./27.*m33*pow(m66,2)*L6r*RSe61*
         RN3e*RS34e + 32./27.*m33*pow(m66,2)*L6r*RN1e*RS14e*RSe63 - 64./
         9.*m33*m44*L7r*RSe63 - 32./9.*m33*m44*L7r*RSe61 + 8./3.*m33*
         m44*L6r + 64./9.*m33*m44*L6r*RSe6613 + 32./9.*m33*m44*L6r*
         RSe6633 - 8./27.*m33*m44*L5r*pow(RSe63,2) - 16./27.*m33*m44*
         L5r*RSe61*RSe63 - 4./3.*m33*m44*L4r - 32./9.*m33*m44*L4r*
         RSe6613 - 16./9.*m33*m44*L4r*pow(RSe63,2) - 16./9.*m33*m44*L4r
         *RSe6633 - 32./9.*m33*m44*L4r*RSe61*RSe63 - 32./9.*m33*m44*me2
         *L7r*RN3e*RSe63 + 32./9.*m33*m44*me2*L7r*RSe61*RN3e - 64./9.*
         m33*m44*me2*L6r*RN3e*RSe6613 + 64./9.*m33*m44*me2*L6r*RN3e*
         RSe6633 + 32./27.*m33*m44*me2*L5r*RSe61*RN3e*RS36e + 16./27.*
         m33*m44*me2*L5r*RN1e*RS16e*RSe63 + 64./9.*m33*m44*me2*L4r*RN3e
         *RSe6613 + 64./27.*m33*m44*me2*L4r*RSe61*RN3e*RS36e - 64./27.*
         m33*m44*me2*L4r*RSe61*RN3e*RS34e + 32./9.*m33*m44*me2*L4r*RN1e
         *RSe6613 + 32./27.*m33*m44*me2*L4r*RN1e*RS16e*RSe63 - 32./27.*
         m33*m44*me2*L4r*RN1e*RS14e*RSe63 );

      M13p6x22L +=  + Ab(me2,mu2) * ( 64./27.*m33*m44*m66*L7r*RN3e*
         RS36e*RSe63 - 32./27.*m33*m44*m66*L7r*RN3e*RS34e*RSe63 + 64./
         27.*m33*m44*m66*L7r*RSe61*RN3e*RS36e - 32./27.*m33*m44*m66*L7r
         *RSe61*RN3e*RS34e + 64./27.*m33*m44*m66*L7r*RN1e*RS16e*RSe63
          - 32./27.*m33*m44*m66*L7r*RN1e*RS14e*RSe63 - 32./27.*m33*m44*
         m66*L6r*RN3e*RS36e*RSe63 + 64./27.*m33*m44*m66*L6r*RN3e*RS34e*
         RSe63 - 32./27.*m33*m44*m66*L6r*RSe61*RN3e*RS36e + 64./27.*m33
         *m44*m66*L6r*RSe61*RN3e*RS34e - 32./27.*m33*m44*m66*L6r*RN1e*
         RS16e*RSe63 + 64./27.*m33*m44*m66*L6r*RN1e*RS14e*RSe63 - 32./
         27.*m33*pow(m44,2)*L8r*RN3e*RS36e*RSe63 - 32./27.*m33*pow(
         m44,2)*L8r*RSe61*RN3e*RS36e - 32./27.*m33*pow(m44,2)*L8r*RN1e*
         RS16e*RSe63 - 64./27.*m33*pow(m44,2)*L7r*RN3e*RS36e*RSe63 - 64.
         /27.*m33*pow(m44,2)*L7r*RSe61*RN3e*RS36e - 64./27.*m33*pow(
         m44,2)*L7r*RN1e*RS16e*RSe63 - 64./27.*m33*pow(m44,2)*L6r*RN3e*
         RS36e*RSe63 - 64./27.*m33*pow(m44,2)*L6r*RSe61*RN3e*RS36e - 64.
         /27.*m33*pow(m44,2)*L6r*RN1e*RS16e*RSe63 );

      M13p6x22L +=  + Ab(me2,mu2) * ( 32./9.*pow(m33,2)*L8r*RSe6613 + 
         64./9.*pow(m33,2)*L8r*RSe6633 - 8./9.*pow(m33,2)*L5r*RSe6613
          - 16./9.*pow(m33,2)*L5r*RSe6633 - 32./9.*pow(m33,2)*me2*L8r*
         RN3e*RSe6613 + 32./9.*pow(m33,2)*me2*L8r*RN3e*RSe6633 + 32./9.
         *pow(m33,2)*me2*L5r*RN3e*RSe6613 - 32./9.*pow(m33,2)*m66*L7r*
         RN3e*RSe63 - 32./9.*pow(m33,2)*m66*L7r*RSe61*RN3e - 32./9.*
         pow(m33,2)*m66*L6r*RN3e*RSe6613 - 32./9.*pow(m33,2)*m66*L6r*
         RN3e*RSe6633 + 32./9.*pow(m33,2)*m44*L7r*RN3e*RSe63 + 32./9.*
         pow(m33,2)*m44*L7r*RSe61*RN3e - 64./9.*pow(m33,2)*m44*L6r*RN3e
         *RSe6613 - 64./9.*pow(m33,2)*m44*L6r*RN3e*RSe6633 - 32./9.*
         pow(m33,3)*L8r*RN3e*RSe6613 - 32./9.*pow(m33,3)*L8r*RN3e*
         RSe6633 - 16./9.*m11*me2*L5r*RSe6613 - 8./9.*m11*me2*L5r*
         RSe6633 - 4*m11*me2*L4r + 4./3.*m11*me2*L3r*RSe6613 + 10./3.*
         m11*me2*L3r*RSe6633 + 10./3.*m11*me2*L3r*RSe6611 + 2*m11*me2*
         L2r + 8*m11*me2*L1r + 16./3.*m11*me2*L0r*RSe6613 + 4./3.*m11*
         me2*L0r*RSe6633 );

      M13p6x22L +=  + Ab(me2,mu2) * ( 4./3.*m11*me2*L0r*RSe6611 + 32./9.
         *m11*m66*L7r*RSe63 + 64./9.*m11*m66*L7r*RSe61 + 16./3.*m11*m66
         *L6r + 32./9.*m11*m66*L6r*RSe6613 + 16./9.*m11*m66*L6r*RSe6611
          - 32./27.*m11*m66*L5r*RSe61*RSe63 - 16./27.*m11*m66*L5r*pow(
         RSe61,2) - 8./3.*m11*m66*L4r - 16./9.*m11*m66*L4r*RSe6613 - 16.
         /9.*m11*m66*L4r*RSe61*RSe63 - 8./9.*m11*m66*L4r*pow(RSe61,2)
          - 8./9.*m11*m66*L4r*RSe6611 - 32./9.*m11*m66*me2*L7r*RN1e*
         RSe63 + 32./9.*m11*m66*me2*L7r*RN1e*RSe61 - 32./9.*m11*m66*me2
         *L6r*RN1e*RSe6613 + 32./9.*m11*m66*me2*L6r*RN1e*RSe6611 - 16./
         27.*m11*m66*me2*L5r*RSe61*RN3e*RS34e - 32./27.*m11*m66*me2*L5r
         *RN1e*RS14e*RSe63 + 16./9.*m11*m66*me2*L4r*RN3e*RSe6613 + 16./
         27.*m11*m66*me2*L4r*RSe61*RN3e*RS36e - 16./27.*m11*m66*me2*L4r
         *RSe61*RN3e*RS34e + 32./9.*m11*m66*me2*L4r*RN1e*RSe6613 + 32./
         27.*m11*m66*me2*L4r*RN1e*RS16e*RSe63 - 32./27.*m11*m66*me2*L4r
         *RN1e*RS14e*RSe63 + 32./27.*m11*pow(m66,2)*L8r*RSe61*RN3e*
         RS34e );

      M13p6x22L +=  + Ab(me2,mu2) * ( 32./27.*m11*pow(m66,2)*L8r*RN1e*
         RS14e*RSe63 + 32./27.*m11*pow(m66,2)*L8r*RN1e*RS14e*RSe61 + 32.
         /27.*m11*pow(m66,2)*L7r*RSe61*RN3e*RS34e + 32./27.*m11*pow(
         m66,2)*L7r*RN1e*RS14e*RSe63 + 32./27.*m11*pow(m66,2)*L7r*RN1e*
         RS14e*RSe61 + 32./27.*m11*pow(m66,2)*L6r*RSe61*RN3e*RS34e + 32.
         /27.*m11*pow(m66,2)*L6r*RN1e*RS14e*RSe63 + 32./27.*m11*pow(
         m66,2)*L6r*RN1e*RS14e*RSe61 - 32./9.*m11*m44*L7r*RSe63 - 64./9.
         *m11*m44*L7r*RSe61 + 8./3.*m11*m44*L6r + 64./9.*m11*m44*L6r*
         RSe6613 + 32./9.*m11*m44*L6r*RSe6611 - 16./27.*m11*m44*L5r*
         RSe61*RSe63 - 8./27.*m11*m44*L5r*pow(RSe61,2) - 4./3.*m11*m44*
         L4r - 32./9.*m11*m44*L4r*RSe6613 - 32./9.*m11*m44*L4r*RSe61*
         RSe63 - 16./9.*m11*m44*L4r*pow(RSe61,2) - 16./9.*m11*m44*L4r*
         RSe6611 + 32./9.*m11*m44*me2*L7r*RN1e*RSe63 - 32./9.*m11*m44*
         me2*L7r*RN1e*RSe61 - 64./9.*m11*m44*me2*L6r*RN1e*RSe6613 + 64./
         9.*m11*m44*me2*L6r*RN1e*RSe6611 + 16./27.*m11*m44*me2*L5r*
         RSe61*RN3e*RS36e );

      M13p6x22L +=  + Ab(me2,mu2) * ( 32./27.*m11*m44*me2*L5r*RN1e*
         RS16e*RSe63 + 32./9.*m11*m44*me2*L4r*RN3e*RSe6613 + 32./27.*
         m11*m44*me2*L4r*RSe61*RN3e*RS36e - 32./27.*m11*m44*me2*L4r*
         RSe61*RN3e*RS34e + 64./9.*m11*m44*me2*L4r*RN1e*RSe6613 + 64./
         27.*m11*m44*me2*L4r*RN1e*RS16e*RSe63 - 64./27.*m11*m44*me2*L4r
         *RN1e*RS14e*RSe63 + 64./27.*m11*m44*m66*L7r*RSe61*RN3e*RS36e
          - 32./27.*m11*m44*m66*L7r*RSe61*RN3e*RS34e + 64./27.*m11*m44*
         m66*L7r*RN1e*RS16e*RSe63 + 64./27.*m11*m44*m66*L7r*RN1e*RS16e*
         RSe61 - 32./27.*m11*m44*m66*L7r*RN1e*RS14e*RSe63 - 32./27.*m11
         *m44*m66*L7r*RN1e*RS14e*RSe61 - 32./27.*m11*m44*m66*L6r*RSe61*
         RN3e*RS36e + 64./27.*m11*m44*m66*L6r*RSe61*RN3e*RS34e - 32./27.
         *m11*m44*m66*L6r*RN1e*RS16e*RSe63 - 32./27.*m11*m44*m66*L6r*
         RN1e*RS16e*RSe61 + 64./27.*m11*m44*m66*L6r*RN1e*RS14e*RSe63 + 
         64./27.*m11*m44*m66*L6r*RN1e*RS14e*RSe61 - 32./27.*m11*pow(
         m44,2)*L8r*RSe61*RN3e*RS36e - 32./27.*m11*pow(m44,2)*L8r*RN1e*
         RS16e*RSe63 );

      M13p6x22L +=  + Ab(me2,mu2) * (  - 32./27.*m11*pow(m44,2)*L8r*
         RN1e*RS16e*RSe61 - 64./27.*m11*pow(m44,2)*L7r*RSe61*RN3e*RS36e
          - 64./27.*m11*pow(m44,2)*L7r*RN1e*RS16e*RSe63 - 64./27.*m11*
         pow(m44,2)*L7r*RN1e*RS16e*RSe61 - 64./27.*m11*pow(m44,2)*L6r*
         RSe61*RN3e*RS36e - 64./27.*m11*pow(m44,2)*L6r*RN1e*RS16e*RSe63
          - 64./27.*m11*pow(m44,2)*L6r*RN1e*RS16e*RSe61 + 32./3.*m11*
         m33*L8r*RSe6613 + 16./3.*m11*m33*L8r*RSe6633 + 16./3.*m11*m33*
         L8r*RSe6611 - 16./9.*m11*m33*L5r*RSe6613 - 16./9.*m11*m33*L5r*
         RSe6633 - 16./9.*m11*m33*L5r*RSe6611 + 16./9.*m11*m33*me2*L5r*
         RN3e*RSe6613 + 16./9.*m11*m33*me2*L5r*RN1e*RSe6613 - 32./9.*
         m11*m33*m66*L7r*RSe61*RN3e - 32./9.*m11*m33*m66*L7r*RN1e*RSe63
          - 32./9.*m11*m33*m66*L6r*RN3e*RSe6613 - 32./9.*m11*m33*m66*
         L6r*RN1e*RSe6613 + 32./9.*m11*m33*m44*L7r*RSe61*RN3e + 32./9.*
         m11*m33*m44*L7r*RN1e*RSe63 - 64./9.*m11*m33*m44*L6r*RN3e*
         RSe6613 - 64./9.*m11*m33*m44*L6r*RN1e*RSe6613 - 32./9.*m11*
         pow(m33,2)*L8r*RN3e*RSe6613 );

      M13p6x22L +=  + Ab(me2,mu2) * ( 32./9.*pow(m11,2)*L8r*RSe6613 + 
         64./9.*pow(m11,2)*L8r*RSe6611 - 8./9.*pow(m11,2)*L5r*RSe6613
          - 16./9.*pow(m11,2)*L5r*RSe6611 - 32./9.*pow(m11,2)*me2*L8r*
         RN1e*RSe6613 + 32./9.*pow(m11,2)*me2*L8r*RN1e*RSe6611 + 32./9.
         *pow(m11,2)*me2*L5r*RN1e*RSe6613 - 32./9.*pow(m11,2)*m66*L7r*
         RN1e*RSe63 - 32./9.*pow(m11,2)*m66*L7r*RN1e*RSe61 - 32./9.*
         pow(m11,2)*m66*L6r*RN1e*RSe6613 - 32./9.*pow(m11,2)*m66*L6r*
         RN1e*RSe6611 + 32./9.*pow(m11,2)*m44*L7r*RN1e*RSe63 + 32./9.*
         pow(m11,2)*m44*L7r*RN1e*RSe61 - 64./9.*pow(m11,2)*m44*L6r*RN1e
         *RSe6613 - 64./9.*pow(m11,2)*m44*L6r*RN1e*RSe6611 - 32./9.*
         pow(m11,2)*m33*L8r*RN1e*RSe6613 - 32./9.*pow(m11,3)*L8r*RN1e*
         RSe6613 - 32./9.*pow(m11,3)*L8r*RN1e*RSe6611 );

      M13p6x22L +=  + Bb(m11,mu2) * ( 8./9.*m11*m66*L6r*RD146e - 4./9.*
         m11*m66*L4r*RD146e + 16./9.*m11*m44*L6r*RD146e - 8./9.*m11*m44
         *L4r*RD146e - 8./3.*m11*m33*L8r*RD146e + 4./3.*m11*m33*L5r*
         RD146e - 5./3.*m11*m33*L3r*RD146e - 2./3.*m11*m33*L0r*RD146e
          - 16./9.*m11*m33*m66*L6r*RS1463e + 8./9.*m11*m33*m66*L4r*
         RS1463e - 32./9.*m11*m33*m44*L6r*RS1463e + 16./9.*m11*m33*m44*
         L4r*RS1463e - 16./9.*pow(m11,2)*L8r*RD146e + 8./9.*pow(m11,2)*
         L5r*RD146e - 5./3.*pow(m11,2)*L3r*RD146e - 2./3.*pow(m11,2)*
         L0r*RD146e - 32./9.*pow(m11,2)*m66*L6r*RS1463e + 16./9.*pow(
         m11,2)*m66*L4r*RS1463e - 64./9.*pow(m11,2)*m44*L6r*RS1463e + 
         32./9.*pow(m11,2)*m44*L4r*RS1463e - 16./9.*pow(m11,2)*m33*L8r*
         RS1463e + 8./9.*pow(m11,2)*m33*L5r*RS1463e - 32./9.*pow(m11,3)
         *L8r*RS1463e + 16./9.*pow(m11,3)*L5r*RS1463e );

      M13p6x22L +=  + Bb(m33,mu2) * ( 8./9.*m33*m66*L6r*RD346e - 4./9.*
         m33*m66*L4r*RD346e + 16./9.*m33*m44*L6r*RD346e - 8./9.*m33*m44
         *L4r*RD346e - 16./9.*pow(m33,2)*L8r*RD346e + 8./9.*pow(m33,2)*
         L5r*RD346e - 5./3.*pow(m33,2)*L3r*RD346e - 2./3.*pow(m33,2)*
         L0r*RD346e - 32./9.*pow(m33,2)*m66*L6r*RS3461e + 16./9.*pow(
         m33,2)*m66*L4r*RS3461e - 64./9.*pow(m33,2)*m44*L6r*RS3461e + 
         32./9.*pow(m33,2)*m44*L4r*RS3461e - 32./9.*pow(m33,3)*L8r*
         RS3461e + 16./9.*pow(m33,3)*L5r*RS3461e - 8./3.*m11*m33*L8r*
         RD346e + 4./3.*m11*m33*L5r*RD346e - 5./3.*m11*m33*L3r*RD346e
          - 2./3.*m11*m33*L0r*RD346e - 16./9.*m11*m33*m66*L6r*RS3461e
          + 8./9.*m11*m33*m66*L4r*RS3461e - 32./9.*m11*m33*m44*L6r*
         RS3461e + 16./9.*m11*m33*m44*L4r*RS3461e - 16./9.*m11*pow(
         m33,2)*L8r*RS3461e + 8./9.*m11*pow(m33,2)*L5r*RS3461e );

      M13p6x22L +=  + Bb(me2,mu2) * ( 16./27.*m66*pow(me2,2)*L5r*pow(
         RSe63,2) - 32./27.*m66*pow(me2,2)*L5r*RSe61*RSe63 + 16./27.*
         m66*pow(me2,2)*L5r*pow(RSe61,2) + 8./9.*m66*pow(me2,2)*L4r*
         pow(RSe63,2) - 16./9.*m66*pow(me2,2)*L4r*RSe61*RSe63 + 8./9.*
         m66*pow(me2,2)*L4r*pow(RSe61,2) - 32./27.*pow(m66,2)*me2*L8r*
         pow(RSe63,2) + 64./27.*pow(m66,2)*me2*L8r*RSe61*RSe63 - 32./27.
         *pow(m66,2)*me2*L8r*pow(RSe61,2) - 32./27.*pow(m66,2)*me2*L7r*
         pow(RSe63,2) + 64./27.*pow(m66,2)*me2*L7r*RSe61*RSe63 - 32./27.
         *pow(m66,2)*me2*L7r*pow(RSe61,2) - 32./27.*pow(m66,2)*me2*L6r*
         pow(RSe63,2) + 64./27.*pow(m66,2)*me2*L6r*RSe61*RSe63 - 32./27.
         *pow(m66,2)*me2*L6r*pow(RSe61,2) + 8./27.*m44*pow(me2,2)*L5r*
         pow(RSe63,2) - 16./27.*m44*pow(me2,2)*L5r*RSe61*RSe63 + 8./27.
         *m44*pow(me2,2)*L5r*pow(RSe61,2) + 16./9.*m44*pow(me2,2)*L4r*
         pow(RSe63,2) - 32./9.*m44*pow(me2,2)*L4r*RSe61*RSe63 + 16./9.*
         m44*pow(me2,2)*L4r*pow(RSe61,2) + 64./27.*m44*m66*me2*L7r*pow(
         RSe63,2) );

      M13p6x22L +=  + Bb(me2,mu2) * (  - 128./27.*m44*m66*me2*L7r*RSe61
         *RSe63 + 64./27.*m44*m66*me2*L7r*pow(RSe61,2) - 80./27.*m44*
         m66*me2*L6r*pow(RSe63,2) + 160./27.*m44*m66*me2*L6r*RSe61*
         RSe63 - 80./27.*m44*m66*me2*L6r*pow(RSe61,2) - 16./27.*pow(
         m44,2)*me2*L8r*pow(RSe63,2) + 32./27.*pow(m44,2)*me2*L8r*RSe61
         *RSe63 - 16./27.*pow(m44,2)*me2*L8r*pow(RSe61,2) - 32./27.*
         pow(m44,2)*me2*L7r*pow(RSe63,2) + 64./27.*pow(m44,2)*me2*L7r*
         RSe61*RSe63 - 32./27.*pow(m44,2)*me2*L7r*pow(RSe61,2) - 32./27.
         *pow(m44,2)*me2*L6r*pow(RSe63,2) + 64./27.*pow(m44,2)*me2*L6r*
         RSe61*RSe63 - 32./27.*pow(m44,2)*me2*L6r*pow(RSe61,2) - 16./27.
         *m33*m66*me2*L5r*pow(RSe63,2) - 32./27.*m33*m66*me2*L5r*RSe61*
         RSe63 - 8./9.*m33*m66*me2*L4r*pow(RSe63,2) - 16./9.*m33*m66*
         me2*L4r*RSe61*RSe63 + 32./27.*m33*pow(m66,2)*L8r*pow(RSe63,2)
          + 64./27.*m33*pow(m66,2)*L8r*RSe61*RSe63 + 32./27.*m33*pow(
         m66,2)*L7r*pow(RSe63,2) + 64./27.*m33*pow(m66,2)*L7r*RSe61*
         RSe63 );

      M13p6x22L +=  + Bb(me2,mu2) * ( 32./27.*m33*pow(m66,2)*L6r*pow(
         RSe63,2) + 64./27.*m33*pow(m66,2)*L6r*RSe61*RSe63 - 8./27.*m33
         *m44*me2*L5r*pow(RSe63,2) - 16./27.*m33*m44*me2*L5r*RSe61*
         RSe63 - 16./9.*m33*m44*me2*L4r*pow(RSe63,2) - 32./9.*m33*m44*
         me2*L4r*RSe61*RSe63 - 64./27.*m33*m44*m66*L7r*pow(RSe63,2) - 
         128./27.*m33*m44*m66*L7r*RSe61*RSe63 + 80./27.*m33*m44*m66*L6r
         *pow(RSe63,2) + 160./27.*m33*m44*m66*L6r*RSe61*RSe63 + 16./27.
         *m33*pow(m44,2)*L8r*pow(RSe63,2) + 32./27.*m33*pow(m44,2)*L8r*
         RSe61*RSe63 + 32./27.*m33*pow(m44,2)*L7r*pow(RSe63,2) + 64./27.
         *m33*pow(m44,2)*L7r*RSe61*RSe63 + 32./27.*m33*pow(m44,2)*L6r*
         pow(RSe63,2) + 64./27.*m33*pow(m44,2)*L6r*RSe61*RSe63 - 32./27.
         *m11*m66*me2*L5r*RSe61*RSe63 - 16./27.*m11*m66*me2*L5r*pow(
         RSe61,2) - 16./9.*m11*m66*me2*L4r*RSe61*RSe63 - 8./9.*m11*m66*
         me2*L4r*pow(RSe61,2) + 64./27.*m11*pow(m66,2)*L8r*RSe61*RSe63
          + 32./27.*m11*pow(m66,2)*L8r*pow(RSe61,2) + 64./27.*m11*pow(
         m66,2)*L7r*RSe61*RSe63 );

      M13p6x22L +=  + Bb(me2,mu2) * ( 32./27.*m11*pow(m66,2)*L7r*pow(
         RSe61,2) + 64./27.*m11*pow(m66,2)*L6r*RSe61*RSe63 + 32./27.*
         m11*pow(m66,2)*L6r*pow(RSe61,2) - 16./27.*m11*m44*me2*L5r*
         RSe61*RSe63 - 8./27.*m11*m44*me2*L5r*pow(RSe61,2) - 32./9.*m11
         *m44*me2*L4r*RSe61*RSe63 - 16./9.*m11*m44*me2*L4r*pow(RSe61,2)
          - 128./27.*m11*m44*m66*L7r*RSe61*RSe63 - 64./27.*m11*m44*m66*
         L7r*pow(RSe61,2) + 160./27.*m11*m44*m66*L6r*RSe61*RSe63 + 80./
         27.*m11*m44*m66*L6r*pow(RSe61,2) + 32./27.*m11*pow(m44,2)*L8r*
         RSe61*RSe63 + 16./27.*m11*pow(m44,2)*L8r*pow(RSe61,2) + 64./27.
         *m11*pow(m44,2)*L7r*RSe61*RSe63 + 32./27.*m11*pow(m44,2)*L7r*
         pow(RSe61,2) + 64./27.*m11*pow(m44,2)*L6r*RSe61*RSe63 + 32./27.
         *m11*pow(m44,2)*L6r*pow(RSe61,2) );

      M13p6x22L +=  - 64*m33*pow(m66,2)*L4r*L6r + 32*m33*pow(m66,2)*
         pow(L4r,2) - 256*m33*m44*m66*L4r*L6r + 128*m33*m44*m66*pow(
         L4r,2) - 256*m33*pow(m44,2)*L4r*L6r + 128*m33*pow(m44,2)*pow(
         L4r,2) - 32*pow(m33,2)*m66*L5r*L6r - 32*pow(m33,2)*m66*L4r*L8r
          + 32*pow(m33,2)*m66*L4r*L5r - 64*pow(m33,2)*m44*L5r*L6r - 64*
         pow(m33,2)*m44*L4r*L8r + 64*pow(m33,2)*m44*L4r*L5r - 16*pow(
         m33,3)*L5r*L8r + 8*pow(m33,3)*pow(L5r,2) - 64*m11*pow(m66,2)*
         L4r*L6r + 32*m11*pow(m66,2)*pow(L4r,2) - 256*m11*m44*m66*L4r*
         L6r + 128*m11*m44*m66*pow(L4r,2) - 256*m11*pow(m44,2)*L4r*L6r
          + 128*m11*pow(m44,2)*pow(L4r,2) - 64*m11*m33*m66*L5r*L6r - 64
         *m11*m33*m66*L4r*L8r + 64*m11*m33*m66*L4r*L5r - 128*m11*m33*
         m44*L5r*L6r - 128*m11*m33*m44*L4r*L8r + 128*m11*m33*m44*L4r*
         L5r - 48*m11*pow(m33,2)*L5r*L8r + 24*m11*pow(m33,2)*pow(L5r,2)
          - 32*pow(m11,2)*m66*L5r*L6r - 32*pow(m11,2)*m66*L4r*L8r + 32*
         pow(m11,2)*m66*L4r*L5r - 64*pow(m11,2)*m44*L5r*L6r - 64*pow(
         m11,2)*m44*L4r*L8r;
      M13p6x22L +=  + 64*pow(m11,2)*m44*L4r*L5r - 48*pow(m11,2)*m33*L5r
         *L8r + 24*pow(m11,2)*m33*pow(L5r,2) - 16*pow(m11,3)*L5r*L8r + 
         8*pow(m11,3)*pow(L5r,2);

  return M13p6x22L/pow(f0,4);
}

double mv2s2nf3p6K(const quarkmassnf mass, const Ki Kiin){
  double KK[116],mu;
  Kiin.out(KK,mu);
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if (mu0 != mu){
    cout << "ERROR: differing mu's in mv2s2nf3p6L\n";}
  if ( (nq != 4) || (B0mq.size() != 4)){
    cout << "ERROR: mv2s2nf3p6K probably called with wrong number"
	 << "of quark masses\n";}
  // four masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m33 = 2.*B0mq[1];
  double m44 = 2.*B0mq[2];
  double m66 = 2.*B0mq[3];
  double M13p6x22K =
       - 4*KK[17]*pow(m33,3) - 12*KK[17]*m11*pow(m33,2) - 12*KK[17]*
         pow(m11,2)*m33 - 4*KK[17]*pow(m11,3) - 8*KK[18]*pow(m33,2)*m66
          - 16*KK[18]*pow(m33,2)*m44 - 16*KK[18]*m11*m33*m66 - 32*KK[18
         ]*m11*m33*m44 - 8*KK[18]*pow(m11,2)*m66 - 16*KK[18]*pow(m11,2)
         *m44 - 4*KK[19]*pow(m33,3) - 4*KK[19]*m11*pow(m33,2) - 4*KK[19
         ]*pow(m11,2)*m33 - 4*KK[19]*pow(m11,3) - 4*KK[20]*pow(m33,2)*
         m66 - 8*KK[20]*pow(m33,2)*m44 - 8*KK[20]*m11*m33*m66 - 16*KK[
         20]*m11*m33*m44 - 4*KK[20]*pow(m11,2)*m66 - 8*KK[20]*pow(
         m11,2)*m44 - 8*KK[21]*m33*pow(m66,2) - 16*KK[21]*m33*pow(
         m44,2) - 8*KK[21]*m11*pow(m66,2) - 16*KK[21]*m11*pow(m44,2) - 
         8*KK[22]*m33*pow(m66,2) - 32*KK[22]*m33*m44*m66 - 32*KK[22]*
         m33*pow(m44,2) - 8*KK[22]*m11*pow(m66,2) - 32*KK[22]*m11*m44*
         m66 - 32*KK[22]*m11*pow(m44,2) - 8*KK[23]*m11*pow(m33,2) - 8*
         KK[23]*pow(m11,2)*m33 + 12*KK[25]*pow(m33,3) + 12*KK[25]*m11*
         pow(m33,2);
      M13p6x22K +=  + 12*KK[25]*pow(m11,2)*m33 + 12*KK[25]*pow(m11,3)
          + 8*KK[26]*m33*pow(m66,2) + 16*KK[26]*m33*pow(m44,2) + 8*KK[
         26]*pow(m33,2)*m66 + 16*KK[26]*pow(m33,2)*m44 + 8*KK[26]*m11*
         pow(m66,2) + 16*KK[26]*m11*pow(m44,2) + 16*KK[26]*m11*m33*m66
          + 32*KK[26]*m11*m33*m44 + 8*KK[26]*pow(m11,2)*m66 + 16*KK[26]
         *pow(m11,2)*m44 + 24*KK[27]*m33*pow(m66,2) + 96*KK[27]*m33*m44
         *m66 + 96*KK[27]*m33*pow(m44,2) + 24*KK[27]*m11*pow(m66,2) + 
         96*KK[27]*m11*m44*m66 + 96*KK[27]*m11*pow(m44,2) + 4*KK[39]*
         pow(m33,3) + 12*KK[39]*m11*pow(m33,2) + 12*KK[39]*pow(m11,2)*
         m33 + 4*KK[39]*pow(m11,3) + 8*KK[40]*pow(m33,2)*m66 + 16*KK[40
         ]*pow(m33,2)*m44 + 16*KK[40]*m11*m33*m66 + 32*KK[40]*m11*m33*
         m44 + 8*KK[40]*pow(m11,2)*m66 + 16*KK[40]*pow(m11,2)*m44;

  return M13p6x22K/pow(f0,4);
}

double mv2s2nf3p6R(const quarkmassnf mass){
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if ( (nq != 4) || (B0mq.size() != 4)){
    cout << "ERROR: mv2s2nf3p6R probably called with wrong number"
	 << "of quark masses\n";}
  // four masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m33 = 2.*B0mq[1];
  double m44 = 2.*B0mq[2];
  double m66 = 2.*B0mq[3];
  double m13 = (m11+m33)/2;
  double m14 = (m11+m44)/2;
  double m16 = (m11+m66)/2;
  double m34 = (m33+m44)/2;
  double m36 = (m33+m66)/2;
  double m46 = (m44+m66)/2;
  double me2 = (m44+2.*m66)/3.;

  double RN13 = RN(m11,m33);
  double RN1e = RN(m11,me2);
  double RN3e = RN(m33,me2);

  double RD146e = RD(m11,m44,m66,me2);
  double RD346e = RD(m33,m44,m66,me2);

  double RC146e = RC(m11,m44,m66,me2);
  double RC346e = RC(m33,m44,m66,me2);

  double RS14e = RS(m11,m44,me2);
  double RS16e = RS(m11,m66,me2);
  double RS34e = RS(m33,m44,me2);
  double RS36e = RS(m33,m66,me2);
  double RSe61 = RS(me2,m66,m11);
  double RSe63 = RS(me2,m66,m33);

  double RS1463e = RS(m11,m44,m66,m33,me2);
  double RS3461e = RS(m33,m44,m66,m11,me2);
  double RSe6611 = RS(me2,m66,m66,m11,m11);
  double RSe6613 = RS(me2,m66,m66,m11,m33);
  double RSe6633 = RS(me2,m66,m66,m33,m33);

  double mu2 = mu0*mu0;

  double M13p6x22R =
       + pi162 * ( 1./270.*pow(me2,2)*RD346e*RSe6613 - 1./540.*pow(
         me2,2)*RD346e*RSe6633 - 1./540.*pow(me2,2)*RSe6611*RD346e + 1./
         270.*pow(me2,2)*RD146e*RSe6613 - 1./540.*pow(me2,2)*RD146e*
         RSe6633 - 1./540.*pow(me2,2)*RD146e*RSe6611 + 2./405.*pow(
         me2,3)*pow(RSe6613,2) - 2./405.*pow(me2,3)*RSe6633*RSe6613 + 1.
         /810.*pow(me2,3)*pow(RSe6633,2) - 2./405.*pow(me2,3)*RSe6611*
         RSe6613 + 1./405.*pow(me2,3)*RSe6611*RSe6633 + 1./810.*pow(
         me2,3)*pow(RSe6611,2) - 1./45.*m66*pow(me2,2) + 4./135.*m66*
         pow(me2,2)*RSe6613 + 1./45.*m66*pow(me2,2)*RSe63 - 5./27.*m66*
         pow(me2,2)*pow(RSe63,2) - 7./270.*m66*pow(me2,2)*RSe6633 + 13./
         162.*m66*pow(me2,2)*RS34e*RSe63 + 1./45.*m66*pow(me2,2)*RSe61
          + 10./27.*m66*pow(me2,2)*RSe61*RSe63 - 13./162.*m66*pow(
         me2,2)*RSe61*RS34e - 5./27.*m66*pow(me2,2)*pow(RSe61,2) - 7./
         270.*m66*pow(me2,2)*RSe6611 - 13./162.*m66*pow(me2,2)*RS14e*
         RSe63 + 13./162.*m66*pow(me2,2)*RS14e*RSe61 + 1./45.*pow(
         m66,2)*me2 );

      M13p6x22R +=  + pi162 * (  - 1./90.*pow(m66,2)*me2*RSe63 + 13./
         243.*pow(m66,2)*me2*pow(RSe63,2) - 1./90.*pow(m66,2)*me2*RSe61
          - 26./243.*pow(m66,2)*me2*RSe61*RSe63 + 13./243.*pow(m66,2)*
         me2*pow(RSe61,2) + 1./6.*m46*pow(me2,2)*pow(RSe63,2) + 1./27.*
         m46*pow(me2,2)*RS36e*RSe63 - 1./27.*m46*pow(me2,2)*RS34e*RSe63
          - 1./3.*m46*pow(me2,2)*RSe61*RSe63 - 1./27.*m46*pow(me2,2)*
         RSe61*RS36e + 1./27.*m46*pow(me2,2)*RSe61*RS34e + 1./6.*m46*
         pow(me2,2)*pow(RSe61,2) - 1./27.*m46*pow(me2,2)*RS16e*RSe63 + 
         1./27.*m46*pow(me2,2)*RS16e*RSe61 + 1./27.*m46*pow(me2,2)*
         RS14e*RSe63 - 1./27.*m46*pow(me2,2)*RS14e*RSe61 + 2./81.*m46*
         m66*me2*RS36e*RSe63 + 1./81.*m46*m66*me2*RS34e*RSe63 - 2./81.*
         m46*m66*me2*RSe61*RS36e - 1./81.*m46*m66*me2*RSe61*RS34e - 2./
         81.*m46*m66*me2*RS16e*RSe63 + 2./81.*m46*m66*me2*RS16e*RSe61
          - 1./81.*m46*m66*me2*RS14e*RSe63 + 1./81.*m46*m66*me2*RS14e*
         RSe61 - 1./90.*m44*pow(me2,2) + 8./135.*m44*pow(me2,2)*RSe6613
          - 1./45.*m44*pow(me2,2)*RSe63 );

      M13p6x22R +=  + pi162 * (  - 13./108.*m44*pow(me2,2)*pow(RSe63,2)
          - 7./135.*m44*pow(me2,2)*RSe6633 - 13./162.*m44*pow(me2,2)*
         RS36e*RSe63 + 1./54.*m44*pow(me2,2)*RS34e*RSe63 - 1./45.*m44*
         pow(me2,2)*RSe61 + 13./54.*m44*pow(me2,2)*RSe61*RSe63 + 13./
         162.*m44*pow(me2,2)*RSe61*RS36e - 1./54.*m44*pow(me2,2)*RSe61*
         RS34e - 13./108.*m44*pow(me2,2)*pow(RSe61,2) - 7./135.*m44*
         pow(me2,2)*RSe6611 + 13./162.*m44*pow(me2,2)*RS16e*RSe63 - 13./
         162.*m44*pow(me2,2)*RS16e*RSe61 - 1./54.*m44*pow(me2,2)*RS14e*
         RSe63 + 1./54.*m44*pow(me2,2)*RS14e*RSe61 + 16./243.*m44*m66*
         me2*pow(RSe63,2) - 32./243.*m44*m66*me2*RSe61*RSe63 + 16./243.
         *m44*m66*me2*pow(RSe61,2) + 1./27.*m44*m46*me2*pow(RSe63,2) + 
         1./81.*m44*m46*me2*RS36e*RSe63 - 1./81.*m44*m46*me2*RS34e*
         RSe63 - 2./27.*m44*m46*me2*RSe61*RSe63 - 1./81.*m44*m46*me2*
         RSe61*RS36e + 1./81.*m44*m46*me2*RSe61*RS34e + 1./27.*m44*m46*
         me2*pow(RSe61,2) - 1./81.*m44*m46*me2*RS16e*RSe63 + 1./81.*m44
         *m46*me2*RS16e*RSe61 );

      M13p6x22R +=  + pi162 * ( 1./81.*m44*m46*me2*RS14e*RSe63 - 1./81.
         *m44*m46*me2*RS14e*RSe61 + 1./90.*pow(m44,2)*me2 + 1./90.*pow(
         m44,2)*me2*RSe63 - 35./972.*pow(m44,2)*me2*pow(RSe63,2) - 1./
         18.*pow(m44,2)*me2*RS36e*RSe63 + 1./90.*pow(m44,2)*me2*RSe61
          + 35./486.*pow(m44,2)*me2*RSe61*RSe63 + 1./18.*pow(m44,2)*me2
         *RSe61*RS36e - 35./972.*pow(m44,2)*me2*pow(RSe61,2) + 1./18.*
         pow(m44,2)*me2*RS16e*RSe63 - 1./18.*pow(m44,2)*me2*RS16e*RSe61
          - 1./45.*m36*pow(me2,2) - 17./270.*m36*pow(me2,2)*RSe6613 - 7.
         /270.*m36*pow(me2,2)*RSe63 + 2./45.*m36*pow(me2,2)*RSe6633 + 
         19./270.*m36*pow(me2,2)*RSe61 - 1./270.*m36*pow(me2,2)*RSe6611
          - 1./432.*m36*m66*RD346e - 1./432.*m36*m66*RD146e + 11./270.*
         m36*m66*me2 + 1./108.*m36*m66*me2*RSe6613 - 2./135.*m36*m66*
         me2*RSe63 + 1./216.*m36*m66*me2*RSe6633 - 2./45.*m36*m66*me2*
         RSe61 + 1./216.*m36*m66*me2*RSe6611 - 1./54.*m36*pow(m66,2) - 
         1./216.*m36*pow(m66,2)*RS3461e - 1./216.*m36*pow(m66,2)*
         RS1463e );

      M13p6x22R +=  + pi162 * ( 1./108.*m36*pow(m66,2)*RSe6613 + 1./216.
         *m36*pow(m66,2)*RSe6633 - 1./432.*m36*pow(m66,2)*RC346e + 1./
         216.*m36*pow(m66,2)*RSe6611 - 1./432.*m36*pow(m66,2)*RC146e - 
         1./90.*m34*pow(me2,2) - 17./135.*m34*pow(me2,2)*RSe6613 + 7./
         270.*m34*pow(me2,2)*RSe63 + 4./45.*m34*pow(me2,2)*RSe6633 - 19.
         /270.*m34*pow(me2,2)*RSe61 - 1./135.*m34*pow(me2,2)*RSe6611 - 
         1./216.*m34*m44*RD346e - 1./216.*m34*m44*RD146e + 11./540.*m34
         *m44*me2 + 1./54.*m34*m44*me2*RSe6613 + 2./135.*m34*m44*me2*
         RSe63 + 1./108.*m34*m44*me2*RSe6633 + 2./45.*m34*m44*me2*RSe61
          + 1./108.*m34*m44*me2*RSe6611 - 1./108.*m34*pow(m44,2) - 1./
         108.*m34*pow(m44,2)*RS3461e - 1./108.*m34*pow(m44,2)*RS1463e
          + 1./54.*m34*pow(m44,2)*RSe6613 + 1./108.*m34*pow(m44,2)*
         RSe6633 - 1./216.*m34*pow(m44,2)*RC346e + 1./108.*m34*pow(
         m44,2)*RSe6611 - 1./216.*m34*pow(m44,2)*RC146e + 23./8640.*m33
         *pow(RD346e,2) + 19./4320.*m33*RD146e*RD346e + 1./576.*m33*
         pow(RD146e,2) );

      M13p6x22R +=  + pi162 * ( 113./3240.*m33*me2*RD346e*RSe6613 - 137.
         /6480.*m33*me2*RD346e*RSe6633 + 31./6480.*m33*me2*RSe6611*
         RD346e + 49./3240.*m33*me2*RD146e*RSe6613 + 47./6480.*m33*me2*
         RD146e*RSe6633 - 29./1296.*m33*me2*RD146e*RSe6611 - 61./960.*
         m33*pow(me2,2) - 11./1440.*m33*pow(me2,2)*RSe6613 - 17./810.*
         m33*pow(me2,2)*RSe6613*RS3461e - 1./54.*m33*pow(me2,2)*RSe6613
         *RS1463e + 241./1620.*m33*pow(me2,2)*pow(RSe6613,2) - 47./720.
         *m33*pow(me2,2)*RSe6633 + 17./1620.*m33*pow(me2,2)*RSe6633*
         RS3461e + 1./108.*m33*pow(me2,2)*RSe6633*RS1463e - 89./1620.*
         m33*pow(me2,2)*RSe6633*RSe6613 + 11./240.*m33*pow(me2,2)*pow(
         RSe6633,2) + 17./1620.*m33*pow(me2,2)*RC346e*RSe6613 - 17./
         3240.*m33*pow(me2,2)*RC346e*RSe6633 - 13./240.*m33*pow(me2,2)*
         RSe6611 + 17./1620.*m33*pow(me2,2)*RSe6611*RS3461e + 1./108.*
         m33*pow(me2,2)*RSe6611*RS1463e - 31./540.*m33*pow(me2,2)*
         RSe6611*RSe6613 - 59./3240.*m33*pow(me2,2)*RSe6611*RSe6633 - 
         17./3240.*m33*pow(me2,2)*RSe6611*RC346e );

      M13p6x22R +=  + pi162 * ( 61./1296.*m33*pow(me2,2)*pow(RSe6611,2)
          + 1./108.*m33*pow(me2,2)*RC146e*RSe6613 - 1./216.*m33*pow(
         me2,2)*RC146e*RSe6633 - 1./216.*m33*pow(me2,2)*RC146e*RSe6611
          - 1./17280.*m33*m66*RD346e - 5./384.*m33*m66*RD146e + 103./
         4320.*m33*m66*me2 - 121./6480.*m33*m66*me2*RSe6613 + 13./864.*
         m33*m66*me2*RSe63 + 13./162.*m33*m66*me2*pow(RSe63,2) - 131./
         8640.*m33*m66*me2*RSe6633 + 2./81.*m33*m66*me2*RS34e*RSe63 - 1.
         /81.*m33*m66*me2*pow(RS34e,2) - 191./4320.*m33*m66*me2*RSe61
          + 13./81.*m33*m66*me2*RSe61*RSe63 - 2./27.*m33*m66*me2*RSe61*
         RS34e + 11./5184.*m33*m66*me2*RSe6611 - 2./81.*m33*m66*me2*
         RS14e*RSe63 + 1./27.*m33*m66*me2*RS14e*RS34e + 5./96.*m33*pow(
         m66,2) + 5./384.*m33*pow(m66,2)*RS3461e + 5./384.*m33*pow(
         m66,2)*RS1463e - 5./192.*m33*pow(m66,2)*RSe6613 - 41./1728.*
         m33*pow(m66,2)*RSe63 + 1./64.*m33*pow(m66,2)*RSe6633 + 1./81.*
         m33*pow(m66,2)*RS36e*RSe63 + 47./5760.*m33*pow(m66,2)*RS34e - 
         23./486.*m33*pow(m66,2)*RS34e*RSe63 );

      M13p6x22R +=  + pi162 * (  - 1./162.*m33*pow(m66,2)*RS34e*RS36e
          + 5./486.*m33*pow(m66,2)*pow(RS34e,2) - 1./128.*m33*pow(
         m66,2)*RC346e + 23./1728.*m33*pow(m66,2)*RSe61 - 1./81.*m33*
         pow(m66,2)*RSe61*RS36e + 23./486.*m33*pow(m66,2)*RSe61*RS34e
          + 1./64.*m33*pow(m66,2)*RSe6611 + 1./162.*m33*pow(m66,2)*
         RS16e*RS34e - 23./3456.*m33*pow(m66,2)*RS14e + 1./162.*m33*
         pow(m66,2)*RS14e*RS36e - 5./243.*m33*pow(m66,2)*RS14e*RS34e - 
         1./128.*m33*pow(m66,2)*RC146e - 1./9.*m33*m46*me2*pow(RSe63,2)
          - 2./9.*m33*m46*me2*RSe61*RSe63 - 2./27.*m33*m46*me2*RSe61*
         RS36e + 2./27.*m33*m46*me2*RSe61*RS34e - 1./27.*m33*m46*me2*
         RS16e*RSe63 + 1./27.*m33*m46*me2*RS14e*RSe63 - 13./72.*m33*m46
         *m66 - 1./81.*m33*m46*m66*RS34e*RS36e - 1./162.*m33*m46*m66*
         pow(RS34e,2) - 4./81.*m33*m46*m66*RSe61*RS36e - 2./81.*m33*m46
         *m66*RSe61*RS34e - 2./81.*m33*m46*m66*RS16e*RSe63 + 2./81.*m33
         *m46*m66*RS16e*RS34e - 1./81.*m33*m46*m66*RS14e*RSe63 + 2./81.
         *m33*m46*m66*RS14e*RS36e );

      M13p6x22R +=  + pi162 * ( 2./81.*m33*m46*m66*RS14e*RS34e - 1./
         8640.*m33*m44*RD346e - 5./192.*m33*m44*RD146e + 103./8640.*m33
         *m44*me2 - 121./3240.*m33*m44*me2*RSe6613 - 13./864.*m33*m44*
         me2*RSe63 + 19./324.*m33*m44*me2*pow(RSe63,2) - 131./4320.*m33
         *m44*me2*RSe6633 - 1./162.*m33*m44*me2*RS36e*RSe63 - 1./162.*
         m33*m44*me2*pow(RS36e,2) + 191./4320.*m33*m44*me2*RSe61 + 19./
         162.*m33*m44*me2*RSe61*RSe63 + 1./54.*m33*m44*me2*RSe61*RS36e
          + 11./2592.*m33*m44*me2*RSe6611 + 1./162.*m33*m44*me2*RS16e*
         RSe63 + 1./54.*m33*m44*me2*RS16e*RS36e + 5./32.*m33*m44*m66 + 
         10./243.*m33*m44*m66*RS36e*RSe63 + 2./243.*m33*m44*m66*pow(
         RS36e,2) - 11./243.*m33*m44*m66*RS34e*RSe63 - 1./81.*m33*m44*
         m66*RS34e*RS36e + 2./243.*m33*m44*m66*pow(RS34e,2) - 10./243.*
         m33*m44*m66*RSe61*RS36e + 11./243.*m33*m44*m66*RSe61*RS34e - 4.
         /243.*m33*m44*m66*RS16e*RS36e + 1./81.*m33*m44*m66*RS16e*RS34e
          + 1./81.*m33*m44*m66*RS14e*RS36e - 4./243.*m33*m44*m66*RS14e*
         RS34e );

      M13p6x22R +=  + pi162 * (  - 2./729.*m33*m44*pow(m66,7)*pow(
         RN1e,3)*pow(RN3e,3) - 13./72.*m33*m44*m46 - 1./54.*m33*m44*m46
         *pow(RSe63,2) - 1./162.*m33*m44*m46*pow(RS36e,2) - 1./81.*m33*
         m44*m46*RS34e*RS36e - 1./27.*m33*m44*m46*RSe61*RSe63 - 2./81.*
         m33*m44*m46*RSe61*RS36e + 2./81.*m33*m44*m46*RSe61*RS34e - 1./
         81.*m33*m44*m46*RS16e*RSe63 + 2./81.*m33*m44*m46*RS16e*RS36e
          + 2./81.*m33*m44*m46*RS16e*RS34e + 1./81.*m33*m44*m46*RS14e*
         RSe63 + 2./81.*m33*m44*m46*RS14e*RS36e - 5./192.*m33*pow(
         m44,2) + 5./192.*m33*pow(m44,2)*RS3461e + 5./192.*m33*pow(
         m44,2)*RS1463e - 5./96.*m33*pow(m44,2)*RSe6613 + 41./1728.*m33
         *pow(m44,2)*RSe63 + 1./36.*m33*pow(m44,2)*pow(RSe63,2) + 1./32.
         *m33*pow(m44,2)*RSe6633 + 47./2880.*m33*pow(m44,2)*RS36e + 13./
         486.*m33*pow(m44,2)*RS36e*RSe63 - 5./243.*m33*pow(m44,2)*pow(
         RS36e,2) - 1./162.*m33*pow(m44,2)*RS34e*RSe63 - 1./162.*m33*
         pow(m44,2)*RS34e*RS36e - 1./64.*m33*pow(m44,2)*RC346e - 23./
         1728.*m33*pow(m44,2)*RSe61 );

      M13p6x22R +=  + pi162 * ( 1./18.*m33*pow(m44,2)*RSe61*RSe63 + 41./
         486.*m33*pow(m44,2)*RSe61*RS36e + 1./162.*m33*pow(m44,2)*RSe61
         *RS34e + 1./32.*m33*pow(m44,2)*RSe6611 - 23./1728.*m33*pow(
         m44,2)*RS16e + 1./18.*m33*pow(m44,2)*RS16e*RSe63 + 47./486.*
         m33*pow(m44,2)*RS16e*RS36e + 1./162.*m33*pow(m44,2)*RS16e*
         RS34e + 1./162.*m33*pow(m44,2)*RS14e*RS36e - 1./64.*m33*pow(
         m44,2)*RC146e - 1./81.*m33*pow(m44,2)*pow(m66,6)*pow(RN1e,3)*
         pow(RN3e,3) - 5./243.*m33*pow(m44,3)*pow(m66,5)*pow(RN1e,3)*
         pow(RN3e,3) - 25./1458.*m33*pow(m44,4)*pow(m66,4)*pow(RN1e,3)*
         pow(RN3e,3) - 5./648.*m33*pow(m44,5)*pow(m66,3)*pow(RN1e,3)*
         pow(RN3e,3) - 7./3888.*m33*pow(m44,6)*pow(m66,2)*pow(RN1e,3)*
         pow(RN3e,3) - 1./5832.*m33*pow(m44,7)*m66*pow(RN1e,3)*pow(
         RN3e,3) - 1./45.*m33*m36*RD346e + 1./48.*m33*m36*RD146e - 1./
         1080.*m33*m36*me2*RSe6613 + 1./120.*m33*m36*me2*RSe63 - 2./45.
         *m33*m36*me2*RSe6633 + 23./216.*m33*m36*me2*RSe61 - 1./24.*m33
         *m36*me2*RSe6611 );

      M13p6x22R +=  + pi162 * (  - 11./288.*m33*m36*m66 - 13./864.*m33*
         m36*m66*RS3461e - 1./96.*m33*m36*m66*RS1463e + 1./48.*m33*m36*
         m66*RSe6613 + 1./144.*m33*m36*m66*RSe63 - 1./216.*m33*m36*m66*
         RSe6633 + 17./4320.*m33*m36*m66*RS34e + 1./16.*m33*m36*m66*
         RSe61 + 1./216.*m33*m36*m66*RSe6611 - 1./32.*m33*m36*m66*RS14e
          - 1./432.*m33*m36*m66*RC146e - 31./432.*m33*m36*m44 - 2./45.*
         m33*m34*RD346e + 1./24.*m33*m34*RD146e - 1./540.*m33*m34*me2*
         RSe6613 - 1./120.*m33*m34*me2*RSe63 - 4./45.*m33*m34*me2*
         RSe6633 - 23./216.*m33*m34*me2*RSe61 - 1./12.*m33*m34*me2*
         RSe6611 - 31./432.*m33*m34*m66 - 4./27.*m33*m34*m44 - 13./432.
         *m33*m34*m44*RS3461e - 1./48.*m33*m34*m44*RS1463e + 1./24.*m33
         *m34*m44*RSe6613 - 1./144.*m33*m34*m44*RSe63 - 1./108.*m33*m34
         *m44*RSe6633 + 17./2160.*m33*m34*m44*RS36e - 1./16.*m33*m34*
         m44*RSe61 + 1./108.*m33*m34*m44*RSe6611 - 1./16.*m33*m34*m44*
         RS16e - 1./216.*m33*m34*m44*RC146e + 23./960.*pow(m33,2)*
         RD346e );

      M13p6x22R +=  + pi162 * (  - 25./2592.*pow(m33,2)*RD346e*RS3461e
          + 25./2592.*pow(m33,2)*RD346e*RS1463e - 25./1296.*pow(m33,2)*
         RD346e*RSe6613 + 73./2592.*pow(m33,2)*RD346e*RSe6633 - 67./
         25920.*pow(m33,2)*RD346e*RC346e - 23./2592.*pow(m33,2)*RSe6611
         *RD346e + 23./5184.*pow(m33,2)*RC146e*RD346e - 13./1152.*pow(
         m33,2)*RD146e - 59./4320.*pow(m33,2)*RD146e*RS3461e - 11./2592.
         *pow(m33,2)*RD146e*RS1463e + 11./1296.*pow(m33,2)*RD146e*
         RSe6613 - 11./2592.*pow(m33,2)*RD146e*RSe6633 - 1./1728.*pow(
         m33,2)*RD146e*RC346e - 11./2592.*pow(m33,2)*RD146e*RSe6611 + 
         11./5184.*pow(m33,2)*RD146e*RC146e + 5./384.*pow(m33,2)*me2 - 
         13./360.*pow(m33,2)*me2*RSe6613 - 23./216.*pow(m33,2)*me2*
         RSe6613*RS3461e + 5./648.*pow(m33,2)*me2*RSe6613*RS1463e - 5./
         324.*pow(m33,2)*me2*pow(RSe6613,2) + 17./4320.*pow(m33,2)*me2*
         RSe6633 - 233./2160.*pow(m33,2)*me2*RSe6633*RS3461e - 17./144.
         *pow(m33,2)*me2*RSe6633*RS1463e + 1./4.*pow(m33,2)*me2*RSe6633
         *RSe6613 );

      M13p6x22R +=  + pi162 * (  - 121./1296.*pow(m33,2)*me2*pow(
         RSe6633,2) + 7./2160.*pow(m33,2)*me2*RC346e*RSe6613 + 41./4320.
         *pow(m33,2)*me2*RC346e*RSe6633 - 53./1728.*pow(m33,2)*me2*
         RSe6611 + 23./2160.*pow(m33,2)*me2*RSe6611*RS3461e - 1./1296.*
         pow(m33,2)*me2*RSe6611*RS1463e + 1./324.*pow(m33,2)*me2*
         RSe6611*RSe6613 + 7./648.*pow(m33,2)*me2*RSe6611*RSe6633 + 13./
         864.*pow(m33,2)*me2*RSe6611*RC346e - 1./144.*pow(m33,2)*me2*
         pow(RSe6611,2) - 1./1296.*pow(m33,2)*me2*RC146e*RSe6613 - 7./
         2592.*pow(m33,2)*me2*RC146e*RSe6633 + 1./288.*pow(m33,2)*me2*
         RC146e*RSe6611 + 17./864.*pow(m33,2)*m66 - 463./17280.*pow(
         m33,2)*m66*RS3461e - 29./10368.*pow(m33,2)*m66*RS1463e + 29./
         5184.*pow(m33,2)*m66*RSe6613 - 95./1728.*pow(m33,2)*m66*RSe63
          + 11./162.*pow(m33,2)*m66*RSe6633 - 1./162.*pow(m33,2)*m66*
         pow(RS36e,2) - 7./17280.*pow(m33,2)*m66*RS34e - 1./162.*pow(
         m33,2)*m66*RS34e*RS36e - 5./324.*pow(m33,2)*m66*pow(RS34e,2)
          - 139./10368.*pow(m33,2)*m66*RC346e );

      M13p6x22R +=  + pi162 * ( 151./1728.*pow(m33,2)*m66*RSe61 - 1./9.
         *pow(m33,2)*m66*RSe61*RS34e + 5./1728.*pow(m33,2)*m66*RSe6611
          + 2./81.*pow(m33,2)*m66*RS16e*RS36e + 1./81.*pow(m33,2)*m66*
         RS16e*RS34e - 151./3456.*pow(m33,2)*m66*RS14e + 1./81.*pow(
         m33,2)*m66*RS14e*RS36e + 5./81.*pow(m33,2)*m66*RS14e*RS34e - 5.
         /3456.*pow(m33,2)*m66*RC146e + 1./729.*pow(m33,2)*pow(m66,7)*
         pow(RN1e,3)*pow(RN3e,3) + 1./162.*pow(m33,2)*m46*pow(RS36e,2)
          - 1./81.*pow(m33,2)*m46*RS34e*RS36e + 1./162.*pow(m33,2)*m46*
         pow(RS34e,2) - 2./27.*pow(m33,2)*m46*RSe61*RS36e + 2./27.*pow(
         m33,2)*m46*RSe61*RS34e - 2./81.*pow(m33,2)*m46*RS16e*RS36e + 2.
         /81.*pow(m33,2)*m46*RS16e*RS34e + 2./81.*pow(m33,2)*m46*RS14e*
         RS36e - 2./81.*pow(m33,2)*m46*RS14e*RS34e + 181./3456.*pow(
         m33,2)*m44 - 463./8640.*pow(m33,2)*m44*RS3461e - 29./5184.*
         pow(m33,2)*m44*RS1463e + 29./2592.*pow(m33,2)*m44*RSe6613 + 95.
         /1728.*pow(m33,2)*m44*RSe63 + 11./81.*pow(m33,2)*m44*RSe6633
          - 7./8640.*pow(m33,2)*m44*RS36e );

      M13p6x22R +=  + pi162 * (  - 7./162.*pow(m33,2)*m44*pow(RS36e,2)
          - 1./162.*pow(m33,2)*m44*RS34e*RS36e - 1./162.*pow(m33,2)*m44
         *pow(RS34e,2) - 139./5184.*pow(m33,2)*m44*RC346e - 151./1728.*
         pow(m33,2)*m44*RSe61 + 4./27.*pow(m33,2)*m44*RSe61*RS36e - 1./
         27.*pow(m33,2)*m44*RSe61*RS34e + 5./864.*pow(m33,2)*m44*
         RSe6611 - 151./1728.*pow(m33,2)*m44*RS16e + 14./81.*pow(m33,2)
         *m44*RS16e*RS36e + 1./81.*pow(m33,2)*m44*RS16e*RS34e + 1./81.*
         pow(m33,2)*m44*RS14e*RS36e + 2./81.*pow(m33,2)*m44*RS14e*RS34e
          - 5./1728.*pow(m33,2)*m44*RC146e + 20./2187.*pow(m33,2)*m44*
         pow(m66,6)*pow(RN1e,3)*pow(RN3e,3) + 157./4374.*pow(m33,2)*
         pow(m44,2)*pow(m66,5)*pow(RN1e,3)*pow(RN3e,3) + 235./4374.*
         pow(m33,2)*pow(m44,3)*pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) + 
         1315./34992.*pow(m33,2)*pow(m44,4)*pow(m66,3)*pow(RN1e,3)*pow(
         RN3e,3) + 227./17496.*pow(m33,2)*pow(m44,5)*pow(m66,2)*pow(
         RN1e,3)*pow(RN3e,3) + 35./17496.*pow(m33,2)*pow(m44,6)*m66*
         pow(RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + pi162 * ( 1./11664.*pow(m33,2)*pow(m44,7)*pow(
         RN1e,3)*pow(RN3e,3) + 1./288.*pow(m33,2)*m36 + 23./720.*pow(
         m33,2)*m36*RS3461e + 19./2160.*pow(m33,2)*m36*RS34e + 4./27.*
         pow(m33,2)*m36*RSe61 - 2./27.*pow(m33,2)*m36*RS14e + 1./144.*
         pow(m33,2)*m34 + 23./360.*pow(m33,2)*m34*RS3461e + 19./1080.*
         pow(m33,2)*m34*RS36e - 4./27.*pow(m33,2)*m34*RSe61 - 4./27.*
         pow(m33,2)*m34*RS16e + 11./288.*pow(m33,3) - 4393./69120.*pow(
         m33,3)*RS3461e + 3749./103680.*pow(m33,3)*pow(RS3461e,2) - 
         1181./13824.*pow(m33,3)*RS1463e + 89./10368.*pow(m33,3)*
         RS1463e*RS3461e + 11./6912.*pow(m33,3)*pow(RS1463e,2) + 1181./
         6912.*pow(m33,3)*RSe6613 - 89./5184.*pow(m33,3)*RSe6613*
         RS3461e - 11./1728.*pow(m33,3)*RSe6613*RS1463e + 11./1728.*
         pow(m33,3)*pow(RSe6613,2) + 355./6912.*pow(m33,3)*RSe6633 + 19.
         /3456.*pow(m33,3)*RSe6633*RS3461e + 11./3456.*pow(m33,3)*
         RSe6633*RS1463e - 11./1728.*pow(m33,3)*RSe6633*RSe6613 + 11./
         6912.*pow(m33,3)*pow(RSe6633,2) );

      M13p6x22R +=  + pi162 * ( 2029./34560.*pow(m33,3)*RS36e + 2029./
         69120.*pow(m33,3)*RS34e + 23./4608.*pow(m33,3)*RC346e + 4763./
         103680.*pow(m33,3)*RC346e*RS3461e + 365./6912.*pow(m33,3)*
         RC346e*RS1463e - 365./3456.*pow(m33,3)*RC346e*RSe6613 + 839./
         20736.*pow(m33,3)*RC346e*RSe6633 - 103./9216.*pow(m33,3)*pow(
         RC346e,2) - 7./2304.*pow(m33,3)*RSe6611 + 121./10368.*pow(
         m33,3)*RSe6611*RS3461e + 11./3456.*pow(m33,3)*RSe6611*RS1463e
          - 11./1728.*pow(m33,3)*RSe6611*RSe6613 + 11./3456.*pow(m33,3)
         *RSe6611*RSe6633 - 185./20736.*pow(m33,3)*RSe6611*RC346e + 11./
         6912.*pow(m33,3)*pow(RSe6611,2) - 431./6912.*pow(m33,3)*RS16e
          - 431./13824.*pow(m33,3)*RS14e + 7./4608.*pow(m33,3)*RC146e
          - 121./20736.*pow(m33,3)*RC146e*RS3461e - 11./6912.*pow(
         m33,3)*RC146e*RS1463e + 11./3456.*pow(m33,3)*RC146e*RSe6613 - 
         11./6912.*pow(m33,3)*RC146e*RSe6633 + 185./41472.*pow(m33,3)*
         RC146e*RC346e - 11./6912.*pow(m33,3)*RC146e*RSe6611 + 11./
         27648.*pow(m33,3)*pow(RC146e,2) );

      M13p6x22R +=  + pi162 * ( 2./2187.*pow(m33,3)*pow(m66,6)*pow(
         RN1e,3)*pow(RN3e,3) + 28./2187.*pow(m33,3)*m44*pow(m66,5)*pow(
         RN1e,3)*pow(RN3e,3) - 7./4374.*pow(m33,3)*pow(m44,2)*pow(
         m66,4)*pow(RN1e,3)*pow(RN3e,3) - 23./972.*pow(m33,3)*pow(
         m44,3)*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) - 157./8748.*pow(
         m33,3)*pow(m44,4)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) - 169./
         34992.*pow(m33,3)*pow(m44,5)*m66*pow(RN1e,3)*pow(RN3e,3) - 7./
         17496.*pow(m33,3)*pow(m44,6)*pow(RN1e,3)*pow(RN3e,3) - 161./
         8748.*pow(m33,4)*pow(m66,5)*pow(RN1e,3)*pow(RN3e,3) - 599./
         8748.*pow(m33,4)*m44*pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) - 2417.
         /34992.*pow(m33,4)*pow(m44,2)*pow(m66,3)*pow(RN1e,3)*pow(
         RN3e,3) - 479./17496.*pow(m33,4)*pow(m44,3)*pow(m66,2)*pow(
         RN1e,3)*pow(RN3e,3) - 71./17496.*pow(m33,4)*pow(m44,4)*m66*
         pow(RN1e,3)*pow(RN3e,3) - 1./8748.*pow(m33,4)*pow(m44,5)*pow(
         RN1e,3)*pow(RN3e,3) + 89./2916.*pow(m33,5)*pow(m66,4)*pow(
         RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + pi162 * ( 52./729.*pow(m33,5)*m44*pow(m66,3)*pow(
         RN1e,3)*pow(RN3e,3) + 223./3888.*pow(m33,5)*pow(m44,2)*pow(
         m66,2)*pow(RN1e,3)*pow(RN3e,3) + 223./11664.*pow(m33,5)*pow(
         m44,3)*m66*pow(RN1e,3)*pow(RN3e,3) + 13./5832.*pow(m33,5)*pow(
         m44,4)*pow(RN1e,3)*pow(RN3e,3) - 7./486.*pow(m33,6)*pow(m66,3)
         *pow(RN1e,3)*pow(RN3e,3) - 7./324.*pow(m33,6)*m44*pow(m66,2)*
         pow(RN1e,3)*pow(RN3e,3) - 7./648.*pow(m33,6)*pow(m44,2)*m66*
         pow(RN1e,3)*pow(RN3e,3) - 7./3888.*pow(m33,6)*pow(m44,3)*pow(
         RN1e,3)*pow(RN3e,3) - 1./45.*m16*pow(me2,2) - 17./270.*m16*
         pow(me2,2)*RSe6613 + 19./270.*m16*pow(me2,2)*RSe63 - 1./270.*
         m16*pow(me2,2)*RSe6633 - 7./270.*m16*pow(me2,2)*RSe61 + 2./45.
         *m16*pow(me2,2)*RSe6611 - 1./432.*m16*m66*RD346e - 1./432.*m16
         *m66*RD146e + 11./270.*m16*m66*me2 + 1./108.*m16*m66*me2*
         RSe6613 - 2./45.*m16*m66*me2*RSe63 + 1./216.*m16*m66*me2*
         RSe6633 - 2./135.*m16*m66*me2*RSe61 + 1./216.*m16*m66*me2*
         RSe6611 );

      M13p6x22R +=  + pi162 * (  - 1./54.*m16*pow(m66,2) - 1./216.*m16*
         pow(m66,2)*RS3461e - 1./216.*m16*pow(m66,2)*RS1463e + 1./108.*
         m16*pow(m66,2)*RSe6613 + 1./216.*m16*pow(m66,2)*RSe6633 - 1./
         432.*m16*pow(m66,2)*RC346e + 1./216.*m16*pow(m66,2)*RSe6611 - 
         1./432.*m16*pow(m66,2)*RC146e + 11./540.*m16*m33*RD346e - 1./
         432.*m16*m33*RD146e + 1./54.*m16*m33*me2 - 11./1080.*m16*m33*
         me2*RSe6613 + 1./40.*m16*m33*me2*RSe63 - 1./30.*m16*m33*me2*
         RSe6633 + 5./216.*m16*m33*me2*RSe61 + 1./216.*m16*m33*me2*
         RSe6611 - 49./864.*m16*m33*m66 - 17./864.*m16*m33*m66*RS3461e
          - 13./864.*m16*m33*m66*RS1463e + 13./432.*m16*m33*m66*RSe6613
          + 19./432.*m16*m33*m66*RSe63 + 1./4320.*m16*m33*m66*RS34e - 1.
         /432.*m16*m33*m66*RC346e - 5./432.*m16*m33*m66*RSe61 + 1./108.
         *m16*m33*m66*RSe6611 + 5./864.*m16*m33*m66*RS14e - 1./216.*m16
         *m33*m66*RC146e - 31./432.*m16*m33*m44 - 31./432.*m16*pow(
         m33,2) + 113./4320.*m16*pow(m33,2)*RS3461e - 1./96.*m16*pow(
         m33,2)*RS1463e );

      M13p6x22R +=  + pi162 * ( 1./48.*m16*pow(m33,2)*RSe6613 + 1./144.
         *m16*pow(m33,2)*RSe63 - 1./216.*m16*pow(m33,2)*RSe6633 - 221./
         4320.*m16*pow(m33,2)*RS34e + 1./48.*m16*pow(m33,2)*RC346e - 5./
         432.*m16*pow(m33,2)*RSe61 + 1./216.*m16*pow(m33,2)*RSe6611 + 5.
         /864.*m16*pow(m33,2)*RS14e - 1./432.*m16*pow(m33,2)*RC146e - 1.
         /90.*m14*pow(me2,2) - 17./135.*m14*pow(me2,2)*RSe6613 - 19./
         270.*m14*pow(me2,2)*RSe63 - 1./135.*m14*pow(me2,2)*RSe6633 + 7.
         /270.*m14*pow(me2,2)*RSe61 + 4./45.*m14*pow(me2,2)*RSe6611 - 1.
         /216.*m14*m44*RD346e - 1./216.*m14*m44*RD146e + 11./540.*m14*
         m44*me2 + 1./54.*m14*m44*me2*RSe6613 + 2./45.*m14*m44*me2*
         RSe63 + 1./108.*m14*m44*me2*RSe6633 + 2./135.*m14*m44*me2*
         RSe61 + 1./108.*m14*m44*me2*RSe6611 - 1./108.*m14*pow(m44,2)
          - 1./108.*m14*pow(m44,2)*RS3461e - 1./108.*m14*pow(m44,2)*
         RS1463e + 1./54.*m14*pow(m44,2)*RSe6613 + 1./108.*m14*pow(
         m44,2)*RSe6633 - 1./216.*m14*pow(m44,2)*RC346e + 1./108.*m14*
         pow(m44,2)*RSe6611 );

      M13p6x22R +=  + pi162 * (  - 1./216.*m14*pow(m44,2)*RC146e + 11./
         270.*m14*m33*RD346e - 1./216.*m14*m33*RD146e + 1./108.*m14*m33
         *me2 - 11./540.*m14*m33*me2*RSe6613 - 1./40.*m14*m33*me2*RSe63
          - 1./15.*m14*m33*me2*RSe6633 - 5./216.*m14*m33*me2*RSe61 + 1./
         108.*m14*m33*me2*RSe6611 - 31./432.*m14*m33*m66 - 17./108.*m14
         *m33*m44 - 17./432.*m14*m33*m44*RS3461e - 13./432.*m14*m33*m44
         *RS1463e + 13./216.*m14*m33*m44*RSe6613 - 19./432.*m14*m33*m44
         *RSe63 + 1./2160.*m14*m33*m44*RS36e - 1./216.*m14*m33*m44*
         RC346e + 5./432.*m14*m33*m44*RSe61 + 1./54.*m14*m33*m44*
         RSe6611 + 5./432.*m14*m33*m44*RS16e - 1./108.*m14*m33*m44*
         RC146e - 31./216.*m14*pow(m33,2) + 113./2160.*m14*pow(m33,2)*
         RS3461e - 1./48.*m14*pow(m33,2)*RS1463e + 1./24.*m14*pow(
         m33,2)*RSe6613 - 1./144.*m14*pow(m33,2)*RSe63 - 1./108.*m14*
         pow(m33,2)*RSe6633 - 221./2160.*m14*pow(m33,2)*RS36e + 1./24.*
         m14*pow(m33,2)*RC346e + 5./432.*m14*pow(m33,2)*RSe61 + 1./108.
         *m14*pow(m33,2)*RSe6611 );

      M13p6x22R +=  + pi162 * ( 5./432.*m14*pow(m33,2)*RS16e - 1./216.*
         m14*pow(m33,2)*RC146e + 1./648.*m13*pow(RD346e,2) + 1./324.*
         m13*RD146e*RD346e + 1./648.*m13*pow(RD146e,2) + 1./81.*m13*me2
         *RD346e*RSe6613 - 1./162.*m13*me2*RD346e*RSe6633 - 1./162.*m13
         *me2*RSe6611*RD346e + 1./81.*m13*me2*RD146e*RSe6613 - 1./162.*
         m13*me2*RD146e*RSe6633 - 1./162.*m13*me2*RD146e*RSe6611 - 1./
         27.*m13*pow(me2,2)*RSe6613*RS3461e - 1./27.*m13*pow(me2,2)*
         RSe6613*RS1463e + 8./81.*m13*pow(me2,2)*pow(RSe6613,2) + 1./54.
         *m13*pow(me2,2)*RSe6633*RS3461e + 1./54.*m13*pow(me2,2)*
         RSe6633*RS1463e - 8./81.*m13*pow(me2,2)*RSe6633*RSe6613 + 2./
         81.*m13*pow(me2,2)*pow(RSe6633,2) + 1./54.*m13*pow(me2,2)*
         RC346e*RSe6613 - 1./108.*m13*pow(me2,2)*RC346e*RSe6633 + 1./54.
         *m13*pow(me2,2)*RSe6611*RS3461e + 1./54.*m13*pow(me2,2)*
         RSe6611*RS1463e - 8./81.*m13*pow(me2,2)*RSe6611*RSe6613 + 4./
         81.*m13*pow(me2,2)*RSe6611*RSe6633 - 1./108.*m13*pow(me2,2)*
         RSe6611*RC346e );

      M13p6x22R +=  + pi162 * ( 2./81.*m13*pow(me2,2)*pow(RSe6611,2) + 
         1./54.*m13*pow(me2,2)*RC146e*RSe6613 - 1./108.*m13*pow(me2,2)*
         RC146e*RSe6633 - 1./108.*m13*pow(me2,2)*RC146e*RSe6611 - 1./
         108.*m13*m33*RD346e*RS3461e - 1./324.*m13*m33*RD346e*RS1463e
          + 1./162.*m13*m33*RD346e*RSe6613 + 1./108.*m13*m33*RD346e*
         RSe6633 - 1./648.*m13*m33*RD346e*RC346e - 5./324.*m13*m33*
         RSe6611*RD346e + 5./648.*m13*m33*RC146e*RD346e + 1./324.*m13*
         m33*RD146e*RS3461e + 1./108.*m13*m33*RD146e*RS1463e - 1./54.*
         m13*m33*RD146e*RSe6613 + 1./324.*m13*m33*RD146e*RSe6633 + 1./
         648.*m13*m33*RD146e*RC346e + 5./324.*m13*m33*RD146e*RSe6611 - 
         5./648.*m13*m33*RD146e*RC146e + 2./81.*m13*m33*me2*RSe6613*
         RS3461e + 4./81.*m13*m33*me2*RSe6613*RS1463e - 8./81.*m13*m33*
         me2*pow(RSe6613,2) - 1./54.*m13*m33*me2*RSe6633*RS3461e - 5./
         162.*m13*m33*me2*RSe6633*RS1463e + 10./81.*m13*m33*me2*RSe6633
         *RSe6613 - 1./18.*m13*m33*me2*pow(RSe6633,2) - 1./54.*m13*m33*
         me2*RC346e*RSe6613 );

      M13p6x22R +=  + pi162 * ( 7./324.*m13*m33*me2*RC346e*RSe6633 - 1./
         162.*m13*m33*me2*RSe6611*RS3461e - 1./54.*m13*m33*me2*RSe6611*
         RS1463e + 2./27.*m13*m33*me2*RSe6611*RSe6613 - 1./81.*m13*m33*
         me2*RSe6611*RSe6633 - 1./324.*m13*m33*me2*RSe6611*RC346e - 5./
         162.*m13*m33*me2*pow(RSe6611,2) - 1./54.*m13*m33*me2*RC146e*
         RSe6613 + 1./324.*m13*m33*me2*RC146e*RSe6633 + 5./324.*m13*m33
         *me2*RC146e*RSe6611 - 31./144.*m13*pow(m33,2) - 1./162.*m13*
         pow(m33,2)*RS1463e*RS3461e + 1./81.*m13*pow(m33,2)*RSe6613*
         RS3461e - 1./81.*m13*pow(m33,2)*RSe6633*RS3461e + 1./162.*m13*
         pow(m33,2)*RC346e*RS3461e + 1./162.*m13*pow(m33,2)*RC346e*
         RS1463e - 1./81.*m13*pow(m33,2)*RC346e*RSe6613 + 1./54.*m13*
         pow(m33,2)*RC346e*RSe6633 - 5./648.*m13*pow(m33,2)*pow(
         RC346e,2) - 1./162.*m13*pow(m33,2)*RSe6611*RC346e + 1./324.*
         m13*pow(m33,2)*RC146e*RC346e + 1./576.*m11*pow(RD346e,2) + 19./
         4320.*m11*RD146e*RD346e + 23./8640.*m11*pow(RD146e,2) + 49./
         3240.*m11*me2*RD346e*RSe6613 );

      M13p6x22R +=  + pi162 * (  - 29./1296.*m11*me2*RD346e*RSe6633 + 
         47./6480.*m11*me2*RSe6611*RD346e + 113./3240.*m11*me2*RD146e*
         RSe6613 + 31./6480.*m11*me2*RD146e*RSe6633 - 137./6480.*m11*
         me2*RD146e*RSe6611 - 61./960.*m11*pow(me2,2) - 11./1440.*m11*
         pow(me2,2)*RSe6613 - 1./54.*m11*pow(me2,2)*RSe6613*RS3461e - 
         17./810.*m11*pow(me2,2)*RSe6613*RS1463e + 241./1620.*m11*pow(
         me2,2)*pow(RSe6613,2) - 13./240.*m11*pow(me2,2)*RSe6633 + 1./
         108.*m11*pow(me2,2)*RSe6633*RS3461e + 17./1620.*m11*pow(me2,2)
         *RSe6633*RS1463e - 31./540.*m11*pow(me2,2)*RSe6633*RSe6613 + 
         61./1296.*m11*pow(me2,2)*pow(RSe6633,2) + 1./108.*m11*pow(
         me2,2)*RC346e*RSe6613 - 1./216.*m11*pow(me2,2)*RC346e*RSe6633
          - 47./720.*m11*pow(me2,2)*RSe6611 + 1./108.*m11*pow(me2,2)*
         RSe6611*RS3461e + 17./1620.*m11*pow(me2,2)*RSe6611*RS1463e - 
         89./1620.*m11*pow(me2,2)*RSe6611*RSe6613 - 59./3240.*m11*pow(
         me2,2)*RSe6611*RSe6633 - 1./216.*m11*pow(me2,2)*RSe6611*RC346e
          + 11./240.*m11*pow(me2,2)*pow(RSe6611,2) );

      M13p6x22R +=  + pi162 * ( 17./1620.*m11*pow(me2,2)*RC146e*RSe6613
          - 17./3240.*m11*pow(me2,2)*RC146e*RSe6633 - 17./3240.*m11*
         pow(me2,2)*RC146e*RSe6611 - 5./384.*m11*m66*RD346e - 1./17280.
         *m11*m66*RD146e + 103./4320.*m11*m66*me2 - 121./6480.*m11*m66*
         me2*RSe6613 - 191./4320.*m11*m66*me2*RSe63 + 11./5184.*m11*m66
         *me2*RSe6633 + 13./864.*m11*m66*me2*RSe61 + 13./81.*m11*m66*
         me2*RSe61*RSe63 - 2./81.*m11*m66*me2*RSe61*RS34e + 13./162.*
         m11*m66*me2*pow(RSe61,2) - 131./8640.*m11*m66*me2*RSe6611 - 2./
         27.*m11*m66*me2*RS14e*RSe63 + 2./81.*m11*m66*me2*RS14e*RS34e
          + 2./81.*m11*m66*me2*RS14e*RSe61 - 1./81.*m11*m66*me2*pow(
         RS14e,2) + 5./96.*m11*pow(m66,2) + 5./384.*m11*pow(m66,2)*
         RS3461e + 5./384.*m11*pow(m66,2)*RS1463e - 5./192.*m11*pow(
         m66,2)*RSe6613 + 23./1728.*m11*pow(m66,2)*RSe63 + 1./64.*m11*
         pow(m66,2)*RSe6633 + 1./384.*m11*pow(m66,2)*RS34e - 1./128.*
         m11*pow(m66,2)*RC346e - 41./1728.*m11*pow(m66,2)*RSe61 + 1./64.
         *m11*pow(m66,2)*RSe6611 );

      M13p6x22R +=  + pi162 * (  - 1./81.*m11*pow(m66,2)*RS16e*RSe63 + 
         1./81.*m11*pow(m66,2)*RS16e*RSe61 + 301./17280.*m11*pow(m66,2)
         *RS14e + 23./486.*m11*pow(m66,2)*RS14e*RSe63 - 23./486.*m11*
         pow(m66,2)*RS14e*RSe61 - 1./162.*m11*pow(m66,2)*RS14e*RS16e + 
         5./486.*m11*pow(m66,2)*pow(RS14e,2) - 1./128.*m11*pow(m66,2)*
         RC146e - 2./9.*m11*m46*me2*RSe61*RSe63 - 1./27.*m11*m46*me2*
         RSe61*RS36e + 1./27.*m11*m46*me2*RSe61*RS34e - 1./9.*m11*m46*
         me2*pow(RSe61,2) - 2./27.*m11*m46*me2*RS16e*RSe63 + 2./27.*m11
         *m46*me2*RS14e*RSe63 - 13./72.*m11*m46*m66 - 2./81.*m11*m46*
         m66*RSe61*RS36e - 1./81.*m11*m46*m66*RSe61*RS34e - 4./81.*m11*
         m46*m66*RS16e*RSe63 + 2./81.*m11*m46*m66*RS16e*RS34e - 2./81.*
         m11*m46*m66*RS14e*RSe63 + 2./81.*m11*m46*m66*RS14e*RS36e + 2./
         81.*m11*m46*m66*RS14e*RS34e - 1./81.*m11*m46*m66*RS14e*RS16e
          - 1./162.*m11*m46*m66*pow(RS14e,2) - 5./192.*m11*m44*RD346e
          - 1./8640.*m11*m44*RD146e + 103./8640.*m11*m44*me2 - 121./
         3240.*m11*m44*me2*RSe6613 );

      M13p6x22R +=  + pi162 * ( 191./4320.*m11*m44*me2*RSe63 + 11./2592.
         *m11*m44*me2*RSe6633 - 13./864.*m11*m44*me2*RSe61 + 19./162.*
         m11*m44*me2*RSe61*RSe63 + 1./162.*m11*m44*me2*RSe61*RS36e + 19.
         /324.*m11*m44*me2*pow(RSe61,2) - 131./4320.*m11*m44*me2*
         RSe6611 + 1./54.*m11*m44*me2*RS16e*RSe63 + 1./81.*m11*m44*me2*
         RS16e*RS36e - 1./162.*m11*m44*me2*RS16e*RSe61 - 1./162.*m11*
         m44*me2*pow(RS16e,2) + 5./32.*m11*m44*m66 - 10./243.*m11*m44*
         m66*RS16e*RSe63 + 10./243.*m11*m44*m66*RS16e*RSe61 + 2./243.*
         m11*m44*m66*pow(RS16e,2) + 11./243.*m11*m44*m66*RS14e*RSe63 - 
         11./243.*m11*m44*m66*RS14e*RSe61 - 1./81.*m11*m44*m66*RS14e*
         RS16e + 2./243.*m11*m44*m66*pow(RS14e,2) - 2./729.*m11*m44*
         pow(m66,7)*pow(RN1e,3)*pow(RN3e,3) - 13./72.*m11*m44*m46 - 1./
         27.*m11*m44*m46*RSe61*RSe63 - 1./81.*m11*m44*m46*RSe61*RS36e
          + 1./81.*m11*m44*m46*RSe61*RS34e - 1./54.*m11*m44*m46*pow(
         RSe61,2) - 2./81.*m11*m44*m46*RS16e*RSe63 + 2./81.*m11*m44*m46
         *RS16e*RS36e );

      M13p6x22R +=  + pi162 * ( 2./81.*m11*m44*m46*RS16e*RS34e - 1./162.
         *m11*m44*m46*pow(RS16e,2) + 2./81.*m11*m44*m46*RS14e*RSe63 + 2.
         /81.*m11*m44*m46*RS14e*RS36e - 1./81.*m11*m44*m46*RS14e*RS16e
          - 5./192.*m11*pow(m44,2) + 5./192.*m11*pow(m44,2)*RS3461e + 5.
         /192.*m11*pow(m44,2)*RS1463e - 5./96.*m11*pow(m44,2)*RSe6613
          - 23./1728.*m11*pow(m44,2)*RSe63 + 1./32.*m11*pow(m44,2)*
         RSe6633 + 1./192.*m11*pow(m44,2)*RS36e - 1./64.*m11*pow(m44,2)
         *RC346e + 41./1728.*m11*pow(m44,2)*RSe61 + 1./18.*m11*pow(
         m44,2)*RSe61*RSe63 + 1./18.*m11*pow(m44,2)*RSe61*RS36e + 1./36.
         *m11*pow(m44,2)*pow(RSe61,2) + 1./32.*m11*pow(m44,2)*RSe6611
          + 301./8640.*m11*pow(m44,2)*RS16e + 41./486.*m11*pow(m44,2)*
         RS16e*RSe63 + 1./9.*m11*pow(m44,2)*RS16e*RS36e + 13./486.*m11*
         pow(m44,2)*RS16e*RSe61 - 5./243.*m11*pow(m44,2)*pow(RS16e,2)
          + 1./162.*m11*pow(m44,2)*RS14e*RSe63 - 1./162.*m11*pow(m44,2)
         *RS14e*RSe61 - 1./162.*m11*pow(m44,2)*RS14e*RS16e - 1./64.*m11
         *pow(m44,2)*RC146e );

      M13p6x22R +=  + pi162 * (  - 1./81.*m11*pow(m44,2)*pow(m66,6)*
         pow(RN1e,3)*pow(RN3e,3) - 5./243.*m11*pow(m44,3)*pow(m66,5)*
         pow(RN1e,3)*pow(RN3e,3) - 25./1458.*m11*pow(m44,4)*pow(m66,4)*
         pow(RN1e,3)*pow(RN3e,3) - 5./648.*m11*pow(m44,5)*pow(m66,3)*
         pow(RN1e,3)*pow(RN3e,3) - 7./3888.*m11*pow(m44,6)*pow(m66,2)*
         pow(RN1e,3)*pow(RN3e,3) - 1./5832.*m11*pow(m44,7)*m66*pow(
         RN1e,3)*pow(RN3e,3) - 1./432.*m11*m36*RD346e + 11./540.*m11*
         m36*RD146e + 1./54.*m11*m36*me2 - 11./1080.*m11*m36*me2*
         RSe6613 + 5./216.*m11*m36*me2*RSe63 + 1./216.*m11*m36*me2*
         RSe6633 + 1./40.*m11*m36*me2*RSe61 - 1./30.*m11*m36*me2*
         RSe6611 - 49./864.*m11*m36*m66 - 13./864.*m11*m36*m66*RS3461e
          - 17./864.*m11*m36*m66*RS1463e + 13./432.*m11*m36*m66*RSe6613
          - 5./432.*m11*m36*m66*RSe63 + 1./108.*m11*m36*m66*RSe6633 + 5.
         /864.*m11*m36*m66*RS34e - 1./216.*m11*m36*m66*RC346e + 19./432.
         *m11*m36*m66*RSe61 + 1./4320.*m11*m36*m66*RS14e - 1./432.*m11*
         m36*m66*RC146e );

      M13p6x22R +=  + pi162 * (  - 31./432.*m11*m36*m44 - 1./216.*m11*
         m34*RD346e + 11./270.*m11*m34*RD146e + 1./108.*m11*m34*me2 - 
         11./540.*m11*m34*me2*RSe6613 - 5./216.*m11*m34*me2*RSe63 + 1./
         108.*m11*m34*me2*RSe6633 - 1./40.*m11*m34*me2*RSe61 - 1./15.*
         m11*m34*me2*RSe6611 - 31./432.*m11*m34*m66 - 17./108.*m11*m34*
         m44 - 13./432.*m11*m34*m44*RS3461e - 17./432.*m11*m34*m44*
         RS1463e + 13./216.*m11*m34*m44*RSe6613 + 5./432.*m11*m34*m44*
         RSe63 + 1./54.*m11*m34*m44*RSe6633 + 5./432.*m11*m34*m44*RS36e
          - 1./108.*m11*m34*m44*RC346e - 19./432.*m11*m34*m44*RSe61 + 1.
         /2160.*m11*m34*m44*RS16e - 1./216.*m11*m34*m44*RC146e + 41./
         5760.*m11*m33*RD346e - 67./3240.*m11*m33*RD346e*RS3461e - 11./
         360.*m11*m33*RD346e*RS1463e + 17./648.*m11*m33*RD346e*RSe6613
          - 17./1296.*m11*m33*RD346e*RSe6633 + 23./1296.*m11*m33*RD346e
         *RC346e - 17./1296.*m11*m33*RSe6611*RD346e + 1./240.*m11*m33*
         RC146e*RD346e + 41./5760.*m11*m33*RD146e - 13./1080.*m11*m33*
         RD146e*RS3461e );

      M13p6x22R +=  + pi162 * (  - 67./3240.*m11*m33*RD146e*RS1463e + 
         17./648.*m11*m33*RD146e*RSe6613 - 17./1296.*m11*m33*RD146e*
         RSe6633 + 1./240.*m11*m33*RD146e*RC346e - 17./1296.*m11*m33*
         RD146e*RSe6611 + 23./1296.*m11*m33*RD146e*RC146e + 5./192.*m11
         *m33*me2 - 119./540.*m11*m33*me2*RSe6613 - 13./135.*m11*m33*
         me2*RSe6613*RS3461e - 13./135.*m11*m33*me2*RSe6613*RS1463e - 5.
         /162.*m11*m33*me2*pow(RSe6613,2) + 313./8640.*m11*m33*me2*
         RSe6633 - 14./405.*m11*m33*me2*RSe6633*RS3461e - 29./810.*m11*
         m33*me2*RSe6633*RS1463e + 17./162.*m11*m33*me2*RSe6633*RSe6613
          - 17./648.*m11*m33*me2*pow(RSe6633,2) + 1./405.*m11*m33*me2*
         RC346e*RSe6613 - 2./81.*m11*m33*me2*RC346e*RSe6633 + 313./8640.
         *m11*m33*me2*RSe6611 - 59./810.*m11*m33*me2*RSe6611*RS3461e - 
         14./405.*m11*m33*me2*RSe6611*RS1463e + 17./162.*m11*m33*me2*
         RSe6611*RSe6613 + 7./324.*m11*m33*me2*RSe6611*RSe6633 + 7./540.
         *m11*m33*me2*RSe6611*RC346e - 17./648.*m11*m33*me2*pow(
         RSe6611,2) );

      M13p6x22R +=  + pi162 * ( 1./405.*m11*m33*me2*RC146e*RSe6613 + 7./
         540.*m11*m33*me2*RC146e*RSe6633 - 2./81.*m11*m33*me2*RC146e*
         RSe6611 + 17./432.*m11*m33*m66 + 139./8640.*m11*m33*m66*
         RS3461e - 223./25920.*m11*m33*m66*RS1463e + 31./864.*m11*m33*
         m66*RSe6613 + 1./72.*m11*m33*m66*RSe63 + 37./1728.*m11*m33*m66
         *RSe6633 - 29./8640.*m11*m33*m66*RS34e - 61./5184.*m11*m33*m66
         *RC346e + 1./72.*m11*m33*m66*RSe61 - 1./18.*m11*m33*m66*RSe61*
         RS34e + 37./1728.*m11*m33*m66*RSe6611 + 1./81.*m11*m33*m66*
         RS16e*RS36e + 1./162.*m11*m33*m66*RS16e*RS34e + 131./8640.*m11
         *m33*m66*RS14e - 1./18.*m11*m33*m66*RS14e*RSe63 + 1./162.*m11*
         m33*m66*RS14e*RS36e + 5./162.*m11*m33*m66*RS14e*RS34e - 61./
         5184.*m11*m33*m66*RC146e + 2./729.*m11*m33*pow(m66,7)*pow(
         RN1e,3)*pow(RN3e,3) - 1./27.*m11*m33*m46*RSe61*RS36e + 1./27.*
         m11*m33*m46*RSe61*RS34e - 1./27.*m11*m33*m46*RS16e*RSe63 - 1./
         27.*m11*m33*m46*RS16e*RS36e + 1./27.*m11*m33*m46*RS16e*RS34e
          + 1./27.*m11*m33*m46*RS14e*RSe63 );

      M13p6x22R +=  + pi162 * ( 1./27.*m11*m33*m46*RS14e*RS36e - 1./27.
         *m11*m33*m46*RS14e*RS34e + 181./1728.*m11*m33*m44 + 139./4320.
         *m11*m33*m44*RS3461e - 223./12960.*m11*m33*m44*RS1463e + 31./
         432.*m11*m33*m44*RSe6613 - 1./72.*m11*m33*m44*RSe63 + 37./864.
         *m11*m33*m44*RSe6633 - 29./4320.*m11*m33*m44*RS36e - 61./2592.
         *m11*m33*m44*RC346e - 1./72.*m11*m33*m44*RSe61 + 2./27.*m11*
         m33*m44*RSe61*RS36e - 1./54.*m11*m33*m44*RSe61*RS34e + 37./864.
         *m11*m33*m44*RSe6611 + 131./4320.*m11*m33*m44*RS16e + 2./27.*
         m11*m33*m44*RS16e*RSe63 + 7./81.*m11*m33*m44*RS16e*RS36e + 1./
         162.*m11*m33*m44*RS16e*RS34e - 1./54.*m11*m33*m44*RS14e*RSe63
          + 1./162.*m11*m33*m44*RS14e*RS36e + 1./81.*m11*m33*m44*RS14e*
         RS34e - 61./2592.*m11*m33*m44*RC146e + 40./2187.*m11*m33*m44*
         pow(m66,6)*pow(RN1e,3)*pow(RN3e,3) + 157./2187.*m11*m33*pow(
         m44,2)*pow(m66,5)*pow(RN1e,3)*pow(RN3e,3) + 235./2187.*m11*m33
         *pow(m44,3)*pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) + 1315./17496.*
         m11*m33*pow(m44,4)*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + pi162 * ( 227./8748.*m11*m33*pow(m44,5)*pow(
         m66,2)*pow(RN1e,3)*pow(RN3e,3) + 35./8748.*m11*m33*pow(m44,6)*
         m66*pow(RN1e,3)*pow(RN3e,3) + 1./5832.*m11*m33*pow(m44,7)*pow(
         RN1e,3)*pow(RN3e,3) - 59./864.*m11*m33*m36 - 23./4320.*m11*m33
         *m36*RS3461e - 43./4320.*m11*m33*m36*RS1463e + 1./48.*m11*m33*
         m36*RSe6613 + 1./144.*m11*m33*m36*RSe63 - 1./216.*m11*m33*m36*
         RSe6633 - 13./864.*m11*m33*m36*RS34e + 1./16.*m11*m33*m36*
         RSe61 + 1./216.*m11*m33*m36*RSe6611 - 73./864.*m11*m33*m36*
         RS14e + 1./54.*m11*m33*m36*RC146e - 59./432.*m11*m33*m34 - 23./
         2160.*m11*m33*m34*RS3461e - 43./2160.*m11*m33*m34*RS1463e + 1./
         24.*m11*m33*m34*RSe6613 - 1./144.*m11*m33*m34*RSe63 - 1./108.*
         m11*m33*m34*RSe6633 - 13./432.*m11*m33*m34*RS36e - 1./16.*m11*
         m33*m34*RSe61 + 1./108.*m11*m33*m34*RSe6611 - 73./432.*m11*m33
         *m34*RS16e + 1./27.*m11*m33*m34*RC146e + 11./64.*m11*pow(
         m33,2) + 229./13824.*m11*pow(m33,2)*RS3461e + 3551./103680.*
         m11*pow(m33,2)*pow(RS3461e,2) );

      M13p6x22R +=  + pi162 * (  - 61./2560.*m11*pow(m33,2)*RS1463e + 
         667./10368.*m11*pow(m33,2)*RS1463e*RS3461e + 19./20736.*m11*
         pow(m33,2)*pow(RS1463e,2) + 193./2304.*m11*pow(m33,2)*RSe6613
          - 115./5184.*m11*pow(m33,2)*RSe6613*RS3461e - 59./5184.*m11*
         pow(m33,2)*RSe6613*RS1463e + 11./576.*m11*pow(m33,2)*pow(
         RSe6613,2) + 59./2304.*m11*pow(m33,2)*RSe6633 + 115./10368.*
         m11*pow(m33,2)*RSe6633*RS3461e + 1./384.*m11*pow(m33,2)*
         RSe6633*RS1463e - 11./576.*m11*pow(m33,2)*RSe6633*RSe6613 + 11.
         /2304.*m11*pow(m33,2)*pow(RSe6633,2) - 733./34560.*m11*pow(
         m33,2)*RS36e - 733./69120.*m11*pow(m33,2)*RS34e - 53./13824.*
         m11*pow(m33,2)*RC346e + 737./103680.*m11*pow(m33,2)*RC346e*
         RS3461e + 1177./103680.*m11*pow(m33,2)*RC346e*RS1463e - 269./
         10368.*m11*pow(m33,2)*RC346e*RSe6613 + 77./20736.*m11*pow(
         m33,2)*RC346e*RSe6633 + 59./9216.*m11*pow(m33,2)*pow(RC346e,2)
          + 19./2304.*m11*pow(m33,2)*RSe6611 + 115./10368.*m11*pow(
         m33,2)*RSe6611*RS3461e );

      M13p6x22R +=  + pi162 * ( 91./10368.*m11*pow(m33,2)*RSe6611*
         RS1463e - 11./576.*m11*pow(m33,2)*RSe6611*RSe6613 + 11./1152.*
         m11*pow(m33,2)*RSe6611*RSe6633 - 307./20736.*m11*pow(m33,2)*
         RSe6611*RC346e + 11./2304.*m11*pow(m33,2)*pow(RSe6611,2) - 113.
         /6912.*m11*pow(m33,2)*RS16e - 113./13824.*m11*pow(m33,2)*RS14e
          + 155./13824.*m11*pow(m33,2)*RC146e - 1127./103680.*m11*pow(
         m33,2)*RC146e*RS3461e - 83./20736.*m11*pow(m33,2)*RC146e*
         RS1463e + 91./10368.*m11*pow(m33,2)*RC146e*RSe6613 - 155./
         20736.*m11*pow(m33,2)*RC146e*RSe6633 - 5./41472.*m11*pow(
         m33,2)*RC146e*RC346e - 1./768.*m11*pow(m33,2)*RC146e*RSe6611
          - 5./9216.*m11*pow(m33,2)*pow(RC146e,2) - 5./729.*m11*pow(
         m33,2)*pow(m66,6)*pow(RN1e,3)*pow(RN3e,3) + 1./729.*m11*pow(
         m33,2)*m44*pow(m66,5)*pow(RN1e,3)*pow(RN3e,3) - 121./1458.*m11
         *pow(m33,2)*pow(m44,2)*pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) - 
         109./729.*m11*pow(m33,2)*pow(m44,3)*pow(m66,3)*pow(RN1e,3)*
         pow(RN3e,3) );

      M13p6x22R +=  + pi162 * (  - 1069./11664.*m11*pow(m33,2)*pow(
         m44,4)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) - 265./11664.*m11*
         pow(m33,2)*pow(m44,5)*m66*pow(RN1e,3)*pow(RN3e,3) - 7./3888.*
         m11*pow(m33,2)*pow(m44,6)*pow(RN1e,3)*pow(RN3e,3) - 143./4374.
         *m11*pow(m33,3)*pow(m66,5)*pow(RN1e,3)*pow(RN3e,3) - 815./4374.
         *m11*pow(m33,3)*m44*pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) - 2165./
         17496.*m11*pow(m33,3)*pow(m44,2)*pow(m66,3)*pow(RN1e,3)*pow(
         RN3e,3) + 40./2187.*m11*pow(m33,3)*pow(m44,3)*pow(m66,2)*pow(
         RN1e,3)*pow(RN3e,3) + 985./34992.*m11*pow(m33,3)*pow(m44,4)*
         m66*pow(RN1e,3)*pow(RN3e,3) + 163./34992.*m11*pow(m33,3)*pow(
         m44,5)*pow(RN1e,3)*pow(RN3e,3) + 397./2916.*m11*pow(m33,4)*
         pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) + 2275./5832.*m11*pow(
         m33,4)*m44*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) + 275./972.*m11*
         pow(m33,4)*pow(m44,2)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) + 803.
         /11664.*m11*pow(m33,4)*pow(m44,3)*m66*pow(RN1e,3)*pow(RN3e,3)
          + 23./5832.*m11*pow(m33,4)*pow(m44,4)*pow(RN1e,3)*pow(RN3e,3)
          );

      M13p6x22R +=  + pi162 * (  - 40./243.*m11*pow(m33,5)*pow(m66,3)*
         pow(RN1e,3)*pow(RN3e,3) - 185./648.*m11*pow(m33,5)*m44*pow(
         m66,2)*pow(RN1e,3)*pow(RN3e,3) - 65./432.*m11*pow(m33,5)*pow(
         m44,2)*m66*pow(RN1e,3)*pow(RN3e,3) - 95./3888.*m11*pow(m33,5)*
         pow(m44,3)*pow(RN1e,3)*pow(RN3e,3) + 7./108.*m11*pow(m33,6)*
         pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) + 7./108.*m11*pow(m33,6)*
         m44*m66*pow(RN1e,3)*pow(RN3e,3) + 7./432.*m11*pow(m33,6)*pow(
         m44,2)*pow(RN1e,3)*pow(RN3e,3) + 1./48.*m11*m16*RD346e - 1./45.
         *m11*m16*RD146e - 1./1080.*m11*m16*me2*RSe6613 + 23./216.*m11*
         m16*me2*RSe63 - 1./24.*m11*m16*me2*RSe6633 + 1./120.*m11*m16*
         me2*RSe61 - 2./45.*m11*m16*me2*RSe6611 - 11./288.*m11*m16*m66
          - 1./96.*m11*m16*m66*RS3461e - 13./864.*m11*m16*m66*RS1463e
          + 1./48.*m11*m16*m66*RSe6613 + 1./16.*m11*m16*m66*RSe63 + 1./
         216.*m11*m16*m66*RSe6633 - 1./32.*m11*m16*m66*RS34e - 1./432.*
         m11*m16*m66*RC346e + 1./144.*m11*m16*m66*RSe61 - 1./216.*m11*
         m16*m66*RSe6611 );

      M13p6x22R +=  + pi162 * ( 17./4320.*m11*m16*m66*RS14e - 31./432.*
         m11*m16*m44 - 59./864.*m11*m16*m33 - 43./4320.*m11*m16*m33*
         RS3461e - 23./4320.*m11*m16*m33*RS1463e + 1./48.*m11*m16*m33*
         RSe6613 + 1./16.*m11*m16*m33*RSe63 + 1./216.*m11*m16*m33*
         RSe6633 - 73./864.*m11*m16*m33*RS34e + 1./54.*m11*m16*m33*
         RC346e + 1./144.*m11*m16*m33*RSe61 - 1./216.*m11*m16*m33*
         RSe6611 - 13./864.*m11*m16*m33*RS14e + 1./24.*m11*m14*RD346e
          - 2./45.*m11*m14*RD146e - 1./540.*m11*m14*me2*RSe6613 - 23./
         216.*m11*m14*me2*RSe63 - 1./12.*m11*m14*me2*RSe6633 - 1./120.*
         m11*m14*me2*RSe61 - 4./45.*m11*m14*me2*RSe6611 - 31./432.*m11*
         m14*m66 - 4./27.*m11*m14*m44 - 1./48.*m11*m14*m44*RS3461e - 13.
         /432.*m11*m14*m44*RS1463e + 1./24.*m11*m14*m44*RSe6613 - 1./16.
         *m11*m14*m44*RSe63 + 1./108.*m11*m14*m44*RSe6633 - 1./16.*m11*
         m14*m44*RS36e - 1./216.*m11*m14*m44*RC346e - 1./144.*m11*m14*
         m44*RSe61 - 1./108.*m11*m14*m44*RSe6611 + 17./2160.*m11*m14*
         m44*RS16e );

      M13p6x22R +=  + pi162 * (  - 59./432.*m11*m14*m33 - 43./2160.*m11
         *m14*m33*RS3461e - 23./2160.*m11*m14*m33*RS1463e + 1./24.*m11*
         m14*m33*RSe6613 - 1./16.*m11*m14*m33*RSe63 + 1./108.*m11*m14*
         m33*RSe6633 - 73./432.*m11*m14*m33*RS36e + 1./27.*m11*m14*m33*
         RC346e - 1./144.*m11*m14*m33*RSe61 - 1./108.*m11*m14*m33*
         RSe6611 - 13./432.*m11*m14*m33*RS16e + 1./108.*m11*m13*RD346e*
         RS3461e + 1./324.*m11*m13*RD346e*RS1463e - 1./54.*m11*m13*
         RD346e*RSe6613 + 5./324.*m11*m13*RD346e*RSe6633 - 5./648.*m11*
         m13*RD346e*RC346e + 1./324.*m11*m13*RSe6611*RD346e + 1./648.*
         m11*m13*RC146e*RD346e - 1./324.*m11*m13*RD146e*RS3461e - 1./
         108.*m11*m13*RD146e*RS1463e + 1./162.*m11*m13*RD146e*RSe6613
          - 5./324.*m11*m13*RD146e*RSe6633 + 5./648.*m11*m13*RD146e*
         RC346e + 1./108.*m11*m13*RD146e*RSe6611 - 1./648.*m11*m13*
         RD146e*RC146e + 4./81.*m11*m13*me2*RSe6613*RS3461e + 2./81.*
         m11*m13*me2*RSe6613*RS1463e - 8./81.*m11*m13*me2*pow(
         RSe6613,2) );

      M13p6x22R +=  + pi162 * (  - 1./54.*m11*m13*me2*RSe6633*RS3461e
          - 1./162.*m11*m13*me2*RSe6633*RS1463e + 2./27.*m11*m13*me2*
         RSe6633*RSe6613 - 5./162.*m11*m13*me2*pow(RSe6633,2) - 1./54.*
         m11*m13*me2*RC346e*RSe6613 + 5./324.*m11*m13*me2*RC346e*
         RSe6633 - 5./162.*m11*m13*me2*RSe6611*RS3461e - 1./54.*m11*m13
         *me2*RSe6611*RS1463e + 10./81.*m11*m13*me2*RSe6611*RSe6613 - 1.
         /81.*m11*m13*me2*RSe6611*RSe6633 + 1./324.*m11*m13*me2*RSe6611
         *RC346e - 1./18.*m11*m13*me2*pow(RSe6611,2) - 1./54.*m11*m13*
         me2*RC146e*RSe6613 - 1./324.*m11*m13*me2*RC146e*RSe6633 + 7./
         324.*m11*m13*me2*RC146e*RSe6611 - 31./72.*m11*m13*m33 - 2./81.
         *m11*m13*m33*pow(RS3461e,2) - 1./27.*m11*m13*m33*RS1463e*
         RS3461e - 2./81.*m11*m13*m33*pow(RS1463e,2) + 4./81.*m11*m13*
         m33*RSe6613*RS3461e + 4./81.*m11*m13*m33*RSe6613*RS1463e - 1./
         54.*m11*m13*m33*RSe6633*RS3461e - 5./162.*m11*m13*m33*RSe6633*
         RS1463e + 1./54.*m11*m13*m33*RC346e*RS3461e + 1./54.*m11*m13*
         m33*RC346e*RS1463e );

      M13p6x22R +=  + pi162 * (  - 1./54.*m11*m13*m33*RC346e*RSe6613 + 
         5./324.*m11*m13*m33*RC346e*RSe6633 - 5./648.*m11*m13*m33*pow(
         RC346e,2) - 5./162.*m11*m13*m33*RSe6611*RS3461e - 1./54.*m11*
         m13*m33*RSe6611*RS1463e + 1./324.*m11*m13*m33*RSe6611*RC346e
          + 1./54.*m11*m13*m33*RC146e*RS3461e + 1./54.*m11*m13*m33*
         RC146e*RS1463e - 1./54.*m11*m13*m33*RC146e*RSe6613 + 1./324.*
         m11*m13*m33*RC146e*RSe6633 + 5./324.*m11*m13*m33*RC146e*
         RSe6611 - 5./648.*m11*m13*m33*pow(RC146e,2) - 13./1152.*pow(
         m11,2)*RD346e - 11./2592.*pow(m11,2)*RD346e*RS3461e - 59./4320.
         *pow(m11,2)*RD346e*RS1463e + 11./1296.*pow(m11,2)*RD346e*
         RSe6613 - 11./2592.*pow(m11,2)*RD346e*RSe6633 + 11./5184.*pow(
         m11,2)*RD346e*RC346e - 11./2592.*pow(m11,2)*RSe6611*RD346e - 1.
         /1728.*pow(m11,2)*RC146e*RD346e + 23./960.*pow(m11,2)*RD146e
          - 23./2592.*pow(m11,2)*RD146e*RS3461e - 25./2592.*pow(m11,2)*
         RD146e*RS1463e - 25./1296.*pow(m11,2)*RD146e*RSe6613 - 23./
         2592.*pow(m11,2)*RD146e*RSe6633 );

      M13p6x22R +=  + pi162 * ( 23./5184.*pow(m11,2)*RD146e*RC346e + 73.
         /2592.*pow(m11,2)*RD146e*RSe6611 - 67./25920.*pow(m11,2)*
         RD146e*RC146e + 5./384.*pow(m11,2)*me2 - 13./360.*pow(m11,2)*
         me2*RSe6613 + 5./648.*pow(m11,2)*me2*RSe6613*RS3461e - 23./216.
         *pow(m11,2)*me2*RSe6613*RS1463e - 5./324.*pow(m11,2)*me2*pow(
         RSe6613,2) - 53./1728.*pow(m11,2)*me2*RSe6633 - 1./1296.*pow(
         m11,2)*me2*RSe6633*RS3461e + 23./2160.*pow(m11,2)*me2*RSe6633*
         RS1463e + 1./324.*pow(m11,2)*me2*RSe6633*RSe6613 - 1./144.*
         pow(m11,2)*me2*pow(RSe6633,2) - 1./1296.*pow(m11,2)*me2*RC346e
         *RSe6613 + 1./288.*pow(m11,2)*me2*RC346e*RSe6633 + 17./4320.*
         pow(m11,2)*me2*RSe6611 - 35./432.*pow(m11,2)*me2*RSe6611*
         RS3461e - 233./2160.*pow(m11,2)*me2*RSe6611*RS1463e + 1./4.*
         pow(m11,2)*me2*RSe6611*RSe6613 + 7./648.*pow(m11,2)*me2*
         RSe6611*RSe6633 - 7./2592.*pow(m11,2)*me2*RSe6611*RC346e - 121.
         /1296.*pow(m11,2)*me2*pow(RSe6611,2) + 7./2160.*pow(m11,2)*me2
         *RC146e*RSe6613 );

      M13p6x22R +=  + pi162 * ( 13./864.*pow(m11,2)*me2*RC146e*RSe6633
          + 41./4320.*pow(m11,2)*me2*RC146e*RSe6611 + 17./864.*pow(
         m11,2)*m66 - 31./3456.*pow(m11,2)*m66*RS3461e - 143./17280.*
         pow(m11,2)*m66*RS1463e + 29./5184.*pow(m11,2)*m66*RSe6613 + 
         151./1728.*pow(m11,2)*m66*RSe63 + 5./1728.*pow(m11,2)*m66*
         RSe6633 - 55./3456.*pow(m11,2)*m66*RS34e - 5./3456.*pow(m11,2)
         *m66*RC346e - 95./1728.*pow(m11,2)*m66*RSe61 + 11./162.*pow(
         m11,2)*m66*RSe6611 + 1./81.*pow(m11,2)*m66*RS16e*RS36e + 1./
         162.*pow(m11,2)*m66*RS16e*RS34e - 1./162.*pow(m11,2)*m66*pow(
         RS16e,2) + 17./1920.*pow(m11,2)*m66*RS14e - 1./9.*pow(m11,2)*
         m66*RS14e*RSe63 + 1./162.*pow(m11,2)*m66*RS14e*RS36e + 5./162.
         *pow(m11,2)*m66*RS14e*RS34e - 1./162.*pow(m11,2)*m66*RS14e*
         RS16e - 5./324.*pow(m11,2)*m66*pow(RS14e,2) - 139./10368.*pow(
         m11,2)*m66*RC146e + 1./729.*pow(m11,2)*pow(m66,7)*pow(RN1e,3)*
         pow(RN3e,3) - 2./27.*pow(m11,2)*m46*RS16e*RSe63 - 2./81.*pow(
         m11,2)*m46*RS16e*RS36e );

      M13p6x22R +=  + pi162 * ( 2./81.*pow(m11,2)*m46*RS16e*RS34e + 1./
         162.*pow(m11,2)*m46*pow(RS16e,2) + 2./27.*pow(m11,2)*m46*RS14e
         *RSe63 + 2./81.*pow(m11,2)*m46*RS14e*RS36e - 2./81.*pow(m11,2)
         *m46*RS14e*RS34e - 1./81.*pow(m11,2)*m46*RS14e*RS16e + 1./162.
         *pow(m11,2)*m46*pow(RS14e,2) + 181./3456.*pow(m11,2)*m44 - 31./
         1728.*pow(m11,2)*m44*RS3461e - 143./8640.*pow(m11,2)*m44*
         RS1463e + 29./2592.*pow(m11,2)*m44*RSe6613 - 151./1728.*pow(
         m11,2)*m44*RSe63 + 5./864.*pow(m11,2)*m44*RSe6633 - 55./1728.*
         pow(m11,2)*m44*RS36e - 5./1728.*pow(m11,2)*m44*RC346e + 95./
         1728.*pow(m11,2)*m44*RSe61 + 11./81.*pow(m11,2)*m44*RSe6611 + 
         17./960.*pow(m11,2)*m44*RS16e + 4./27.*pow(m11,2)*m44*RS16e*
         RSe63 + 7./81.*pow(m11,2)*m44*RS16e*RS36e + 1./162.*pow(m11,2)
         *m44*RS16e*RS34e - 7./162.*pow(m11,2)*m44*pow(RS16e,2) - 1./27.
         *pow(m11,2)*m44*RS14e*RSe63 + 1./162.*pow(m11,2)*m44*RS14e*
         RS36e + 1./81.*pow(m11,2)*m44*RS14e*RS34e - 1./162.*pow(m11,2)
         *m44*RS14e*RS16e );

      M13p6x22R +=  + pi162 * (  - 1./162.*pow(m11,2)*m44*pow(RS14e,2)
          - 139./5184.*pow(m11,2)*m44*RC146e + 20./2187.*pow(m11,2)*m44
         *pow(m66,6)*pow(RN1e,3)*pow(RN3e,3) + 157./4374.*pow(m11,2)*
         pow(m44,2)*pow(m66,5)*pow(RN1e,3)*pow(RN3e,3) + 235./4374.*
         pow(m11,2)*pow(m44,3)*pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) + 
         1315./34992.*pow(m11,2)*pow(m44,4)*pow(m66,3)*pow(RN1e,3)*pow(
         RN3e,3) + 227./17496.*pow(m11,2)*pow(m44,5)*pow(m66,2)*pow(
         RN1e,3)*pow(RN3e,3) + 35./17496.*pow(m11,2)*pow(m44,6)*m66*
         pow(RN1e,3)*pow(RN3e,3) + 1./11664.*pow(m11,2)*pow(m44,7)*pow(
         RN1e,3)*pow(RN3e,3) - 31./432.*pow(m11,2)*m36 - 1./96.*pow(
         m11,2)*m36*RS3461e + 113./4320.*pow(m11,2)*m36*RS1463e + 1./48.
         *pow(m11,2)*m36*RSe6613 - 5./432.*pow(m11,2)*m36*RSe63 + 1./
         216.*pow(m11,2)*m36*RSe6633 + 5./864.*pow(m11,2)*m36*RS34e - 1.
         /432.*pow(m11,2)*m36*RC346e + 1./144.*pow(m11,2)*m36*RSe61 - 1.
         /216.*pow(m11,2)*m36*RSe6611 - 221./4320.*pow(m11,2)*m36*RS14e
          + 1./48.*pow(m11,2)*m36*RC146e );

      M13p6x22R +=  + pi162 * (  - 31./216.*pow(m11,2)*m34 - 1./48.*
         pow(m11,2)*m34*RS3461e + 113./2160.*pow(m11,2)*m34*RS1463e + 1.
         /24.*pow(m11,2)*m34*RSe6613 + 5./432.*pow(m11,2)*m34*RSe63 + 1.
         /108.*pow(m11,2)*m34*RSe6633 + 5./432.*pow(m11,2)*m34*RS36e - 
         1./216.*pow(m11,2)*m34*RC346e - 1./144.*pow(m11,2)*m34*RSe61
          - 1./108.*pow(m11,2)*m34*RSe6611 - 221./2160.*pow(m11,2)*m34*
         RS16e + 1./24.*pow(m11,2)*m34*RC146e + 67./576.*pow(m11,2)*m33
          - 367./69120.*pow(m11,2)*m33*RS3461e + 19./20736.*pow(m11,2)*
         m33*pow(RS3461e,2) + 997./13824.*pow(m11,2)*m33*RS1463e + 667./
         10368.*pow(m11,2)*m33*RS1463e*RS3461e + 3551./103680.*pow(
         m11,2)*m33*pow(RS1463e,2) + 193./2304.*pow(m11,2)*m33*RSe6613
          - 59./5184.*pow(m11,2)*m33*RSe6613*RS3461e - 115./5184.*pow(
         m11,2)*m33*RSe6613*RS1463e + 11./576.*pow(m11,2)*m33*pow(
         RSe6613,2) + 19./2304.*pow(m11,2)*m33*RSe6633 + 91./10368.*
         pow(m11,2)*m33*RSe6633*RS3461e + 115./10368.*pow(m11,2)*m33*
         RSe6633*RS1463e );

      M13p6x22R +=  + pi162 * (  - 11./576.*pow(m11,2)*m33*RSe6633*
         RSe6613 + 11./2304.*pow(m11,2)*m33*pow(RSe6633,2) - 241./6912.
         *pow(m11,2)*m33*RS36e - 241./13824.*pow(m11,2)*m33*RS34e + 155.
         /13824.*pow(m11,2)*m33*RC346e - 83./20736.*pow(m11,2)*m33*
         RC346e*RS3461e - 1127./103680.*pow(m11,2)*m33*RC346e*RS1463e
          + 91./10368.*pow(m11,2)*m33*RC346e*RSe6613 - 1./768.*pow(
         m11,2)*m33*RC346e*RSe6633 - 5./9216.*pow(m11,2)*m33*pow(
         RC346e,2) + 59./2304.*pow(m11,2)*m33*RSe6611 + 1./384.*pow(
         m11,2)*m33*RSe6611*RS3461e + 115./10368.*pow(m11,2)*m33*
         RSe6611*RS1463e - 11./576.*pow(m11,2)*m33*RSe6611*RSe6613 + 11.
         /1152.*pow(m11,2)*m33*RSe6611*RSe6633 - 155./20736.*pow(m11,2)
         *m33*RSe6611*RC346e + 11./2304.*pow(m11,2)*m33*pow(RSe6611,2)
          - 31./11520.*pow(m11,2)*m33*RS16e - 31./23040.*pow(m11,2)*m33
         *RS14e - 53./13824.*pow(m11,2)*m33*RC146e + 3097./103680.*pow(
         m11,2)*m33*RC146e*RS3461e + 737./103680.*pow(m11,2)*m33*RC146e
         *RS1463e );

      M13p6x22R +=  + pi162 * (  - 269./10368.*pow(m11,2)*m33*RC146e*
         RSe6613 - 307./20736.*pow(m11,2)*m33*RC146e*RSe6633 - 5./41472.
         *pow(m11,2)*m33*RC146e*RC346e + 77./20736.*pow(m11,2)*m33*
         RC146e*RSe6611 + 59./9216.*pow(m11,2)*m33*pow(RC146e,2) - 5./
         729.*pow(m11,2)*m33*pow(m66,6)*pow(RN1e,3)*pow(RN3e,3) + 1./
         729.*pow(m11,2)*m33*m44*pow(m66,5)*pow(RN1e,3)*pow(RN3e,3) - 
         121./1458.*pow(m11,2)*m33*pow(m44,2)*pow(m66,4)*pow(RN1e,3)*
         pow(RN3e,3) - 109./729.*pow(m11,2)*m33*pow(m44,3)*pow(m66,3)*
         pow(RN1e,3)*pow(RN3e,3) - 1069./11664.*pow(m11,2)*m33*pow(
         m44,4)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) - 265./11664.*pow(
         m11,2)*m33*pow(m44,5)*m66*pow(RN1e,3)*pow(RN3e,3) - 7./3888.*
         pow(m11,2)*m33*pow(m44,6)*pow(RN1e,3)*pow(RN3e,3) - 125./4374.
         *pow(m11,2)*pow(m33,2)*pow(m66,5)*pow(RN1e,3)*pow(RN3e,3) - 
         1031./4374.*pow(m11,2)*pow(m33,2)*m44*pow(m66,4)*pow(RN1e,3)*
         pow(RN3e,3) - 1913./17496.*pow(m11,2)*pow(m33,2)*pow(m44,2)*
         pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + pi162 * ( 799./8748.*pow(m11,2)*pow(m33,2)*pow(
         m44,3)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) + 1127./17496.*pow(
         m11,2)*pow(m33,2)*pow(m44,4)*m66*pow(RN1e,3)*pow(RN3e,3) + 167.
         /17496.*pow(m11,2)*pow(m33,2)*pow(m44,5)*pow(RN1e,3)*pow(
         RN3e,3) + 146./729.*pow(m11,2)*pow(m33,3)*pow(m66,4)*pow(
         RN1e,3)*pow(RN3e,3) + 1031./1458.*pow(m11,2)*pow(m33,3)*m44*
         pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) + 209./486.*pow(m11,2)*pow(
         m33,3)*pow(m44,2)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) + 121./
         2916.*pow(m11,2)*pow(m33,3)*pow(m44,3)*m66*pow(RN1e,3)*pow(
         RN3e,3) - 127./11664.*pow(m11,2)*pow(m33,3)*pow(m44,4)*pow(
         RN1e,3)*pow(RN3e,3) - 1529./3888.*pow(m11,2)*pow(m33,4)*pow(
         m66,3)*pow(RN1e,3)*pow(RN3e,3) - 523./648.*pow(m11,2)*pow(
         m33,4)*m44*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) - 79./216.*pow(
         m11,2)*pow(m33,4)*pow(m44,2)*m66*pow(RN1e,3)*pow(RN3e,3) - 37./
         972.*pow(m11,2)*pow(m33,4)*pow(m44,3)*pow(RN1e,3)*pow(RN3e,3)
          + 421./1296.*pow(m11,2)*pow(m33,5)*pow(m66,2)*pow(RN1e,3)*
         pow(RN3e,3) );

      M13p6x22R +=  + pi162 * ( 481./1296.*pow(m11,2)*pow(m33,5)*m44*
         m66*pow(RN1e,3)*pow(RN3e,3) + 31./324.*pow(m11,2)*pow(m33,5)*
         pow(m44,2)*pow(RN1e,3)*pow(RN3e,3) - 7./72.*pow(m11,2)*pow(
         m33,6)*m66*pow(RN1e,3)*pow(RN3e,3) - 7./144.*pow(m11,2)*pow(
         m33,6)*m44*pow(RN1e,3)*pow(RN3e,3) + 1./288.*pow(m11,2)*m16 + 
         23./720.*pow(m11,2)*m16*RS1463e + 4./27.*pow(m11,2)*m16*RSe63
          - 2./27.*pow(m11,2)*m16*RS34e + 19./2160.*pow(m11,2)*m16*
         RS14e + 1./144.*pow(m11,2)*m14 + 23./360.*pow(m11,2)*m14*
         RS1463e - 4./27.*pow(m11,2)*m14*RSe63 - 4./27.*pow(m11,2)*m14*
         RS36e + 19./1080.*pow(m11,2)*m14*RS16e - 31./144.*pow(m11,2)*
         m13 - 1./162.*pow(m11,2)*m13*RS1463e*RS3461e + 1./81.*pow(
         m11,2)*m13*RSe6613*RS1463e - 1./81.*pow(m11,2)*m13*RSe6611*
         RS1463e + 1./162.*pow(m11,2)*m13*RC146e*RS3461e + 1./162.*pow(
         m11,2)*m13*RC146e*RS1463e - 1./81.*pow(m11,2)*m13*RC146e*
         RSe6613 - 1./162.*pow(m11,2)*m13*RC146e*RSe6633 + 1./324.*pow(
         m11,2)*m13*RC146e*RC346e );

      M13p6x22R +=  + pi162 * ( 1./54.*pow(m11,2)*m13*RC146e*RSe6611 - 
         5./648.*pow(m11,2)*m13*pow(RC146e,2) - 5./288.*pow(m11,3) - 
         223./4608.*pow(m11,3)*RS3461e + 11./6912.*pow(m11,3)*pow(
         RS3461e,2) - 4393./69120.*pow(m11,3)*RS1463e + 89./10368.*pow(
         m11,3)*RS1463e*RS3461e + 3749./103680.*pow(m11,3)*pow(
         RS1463e,2) + 1181./6912.*pow(m11,3)*RSe6613 - 11./1728.*pow(
         m11,3)*RSe6613*RS3461e - 89./5184.*pow(m11,3)*RSe6613*RS1463e
          + 11./1728.*pow(m11,3)*pow(RSe6613,2) - 7./2304.*pow(m11,3)*
         RSe6633 + 11./3456.*pow(m11,3)*RSe6633*RS3461e + 121./10368.*
         pow(m11,3)*RSe6633*RS1463e - 11./1728.*pow(m11,3)*RSe6633*
         RSe6613 + 11./6912.*pow(m11,3)*pow(RSe6633,2) - 175./6912.*
         pow(m11,3)*RS36e - 175./13824.*pow(m11,3)*RS34e + 7./4608.*
         pow(m11,3)*RC346e - 11./6912.*pow(m11,3)*RC346e*RS3461e - 121./
         20736.*pow(m11,3)*RC346e*RS1463e + 11./3456.*pow(m11,3)*RC346e
         *RSe6613 - 11./6912.*pow(m11,3)*RC346e*RSe6633 + 11./27648.*
         pow(m11,3)*pow(RC346e,2) );

      M13p6x22R +=  + pi162 * ( 355./6912.*pow(m11,3)*RSe6611 + 11./
         3456.*pow(m11,3)*RSe6611*RS3461e + 19./3456.*pow(m11,3)*
         RSe6611*RS1463e - 11./1728.*pow(m11,3)*RSe6611*RSe6613 + 11./
         3456.*pow(m11,3)*RSe6611*RSe6633 - 11./6912.*pow(m11,3)*
         RSe6611*RC346e + 11./6912.*pow(m11,3)*pow(RSe6611,2) + 2029./
         34560.*pow(m11,3)*RS16e + 2029./69120.*pow(m11,3)*RS14e + 23./
         4608.*pow(m11,3)*RC146e + 79./2304.*pow(m11,3)*RC146e*RS3461e
          + 4763./103680.*pow(m11,3)*RC146e*RS1463e - 365./3456.*pow(
         m11,3)*RC146e*RSe6613 - 185./20736.*pow(m11,3)*RC146e*RSe6633
          + 185./41472.*pow(m11,3)*RC146e*RC346e + 839./20736.*pow(
         m11,3)*RC146e*RSe6611 - 103./9216.*pow(m11,3)*pow(RC146e,2) + 
         2./2187.*pow(m11,3)*pow(m66,6)*pow(RN1e,3)*pow(RN3e,3) + 28./
         2187.*pow(m11,3)*m44*pow(m66,5)*pow(RN1e,3)*pow(RN3e,3) - 7./
         4374.*pow(m11,3)*pow(m44,2)*pow(m66,4)*pow(RN1e,3)*pow(RN3e,3)
          - 23./972.*pow(m11,3)*pow(m44,3)*pow(m66,3)*pow(RN1e,3)*pow(
         RN3e,3) );

      M13p6x22R +=  + pi162 * (  - 157./8748.*pow(m11,3)*pow(m44,4)*
         pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) - 169./34992.*pow(m11,3)*
         pow(m44,5)*m66*pow(RN1e,3)*pow(RN3e,3) - 7./17496.*pow(m11,3)*
         pow(m44,6)*pow(RN1e,3)*pow(RN3e,3) - 143./4374.*pow(m11,3)*m33
         *pow(m66,5)*pow(RN1e,3)*pow(RN3e,3) - 815./4374.*pow(m11,3)*
         m33*m44*pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) - 2165./17496.*pow(
         m11,3)*m33*pow(m44,2)*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) + 40./
         2187.*pow(m11,3)*m33*pow(m44,3)*pow(m66,2)*pow(RN1e,3)*pow(
         RN3e,3) + 985./34992.*pow(m11,3)*m33*pow(m44,4)*m66*pow(
         RN1e,3)*pow(RN3e,3) + 163./34992.*pow(m11,3)*m33*pow(m44,5)*
         pow(RN1e,3)*pow(RN3e,3) + 146./729.*pow(m11,3)*pow(m33,2)*pow(
         m66,4)*pow(RN1e,3)*pow(RN3e,3) + 1031./1458.*pow(m11,3)*pow(
         m33,2)*m44*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) + 209./486.*pow(
         m11,3)*pow(m33,2)*pow(m44,2)*pow(m66,2)*pow(RN1e,3)*pow(
         RN3e,3) + 121./2916.*pow(m11,3)*pow(m33,2)*pow(m44,3)*m66*pow(
         RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + pi162 * (  - 127./11664.*pow(m11,3)*pow(m33,2)*
         pow(m44,4)*pow(RN1e,3)*pow(RN3e,3) - 35./72.*pow(m11,3)*pow(
         m33,3)*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) - 88./81.*pow(m11,3)
         *pow(m33,3)*m44*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) - 293./648.
         *pow(m11,3)*pow(m33,3)*pow(m44,2)*m66*pow(RN1e,3)*pow(RN3e,3)
          - 5./162.*pow(m11,3)*pow(m33,3)*pow(m44,3)*pow(RN1e,3)*pow(
         RN3e,3) + 59./108.*pow(m11,3)*pow(m33,4)*pow(m66,2)*pow(
         RN1e,3)*pow(RN3e,3) + 305./432.*pow(m11,3)*pow(m33,4)*m44*m66*
         pow(RN1e,3)*pow(RN3e,3) + 31./216.*pow(m11,3)*pow(m33,4)*pow(
         m44,2)*pow(RN1e,3)*pow(RN3e,3) - 119./432.*pow(m11,3)*pow(
         m33,5)*m66*pow(RN1e,3)*pow(RN3e,3) - 67./432.*pow(m11,3)*pow(
         m33,5)*m44*pow(RN1e,3)*pow(RN3e,3) + 7./144.*pow(m11,3)*pow(
         m33,6)*pow(RN1e,3)*pow(RN3e,3) - 161./8748.*pow(m11,4)*pow(
         m66,5)*pow(RN1e,3)*pow(RN3e,3) - 599./8748.*pow(m11,4)*m44*
         pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) - 2417./34992.*pow(m11,4)*
         pow(m44,2)*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + pi162 * (  - 479./17496.*pow(m11,4)*pow(m44,3)*
         pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) - 71./17496.*pow(m11,4)*
         pow(m44,4)*m66*pow(RN1e,3)*pow(RN3e,3) - 1./8748.*pow(m11,4)*
         pow(m44,5)*pow(RN1e,3)*pow(RN3e,3) + 397./2916.*pow(m11,4)*m33
         *pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) + 2275./5832.*pow(m11,4)*
         m33*m44*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) + 275./972.*pow(
         m11,4)*m33*pow(m44,2)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) + 803.
         /11664.*pow(m11,4)*m33*pow(m44,3)*m66*pow(RN1e,3)*pow(RN3e,3)
          + 23./5832.*pow(m11,4)*m33*pow(m44,4)*pow(RN1e,3)*pow(RN3e,3)
          - 1529./3888.*pow(m11,4)*pow(m33,2)*pow(m66,3)*pow(RN1e,3)*
         pow(RN3e,3) - 523./648.*pow(m11,4)*pow(m33,2)*m44*pow(m66,2)*
         pow(RN1e,3)*pow(RN3e,3) - 79./216.*pow(m11,4)*pow(m33,2)*pow(
         m44,2)*m66*pow(RN1e,3)*pow(RN3e,3) - 37./972.*pow(m11,4)*pow(
         m33,2)*pow(m44,3)*pow(RN1e,3)*pow(RN3e,3) + 59./108.*pow(
         m11,4)*pow(m33,3)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) + 305./
         432.*pow(m11,4)*pow(m33,3)*m44*m66*pow(RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + pi162 * ( 31./216.*pow(m11,4)*pow(m33,3)*pow(
         m44,2)*pow(RN1e,3)*pow(RN3e,3) - 77./216.*pow(m11,4)*pow(
         m33,4)*m66*pow(RN1e,3)*pow(RN3e,3) - 23./108.*pow(m11,4)*pow(
         m33,4)*m44*pow(RN1e,3)*pow(RN3e,3) + 1./12.*pow(m11,4)*pow(
         m33,5)*pow(RN1e,3)*pow(RN3e,3) + 89./2916.*pow(m11,5)*pow(
         m66,4)*pow(RN1e,3)*pow(RN3e,3) + 52./729.*pow(m11,5)*m44*pow(
         m66,3)*pow(RN1e,3)*pow(RN3e,3) + 223./3888.*pow(m11,5)*pow(
         m44,2)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) + 223./11664.*pow(
         m11,5)*pow(m44,3)*m66*pow(RN1e,3)*pow(RN3e,3) + 13./5832.*pow(
         m11,5)*pow(m44,4)*pow(RN1e,3)*pow(RN3e,3) - 40./243.*pow(
         m11,5)*m33*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) - 185./648.*pow(
         m11,5)*m33*m44*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) - 65./432.*
         pow(m11,5)*m33*pow(m44,2)*m66*pow(RN1e,3)*pow(RN3e,3) - 95./
         3888.*pow(m11,5)*m33*pow(m44,3)*pow(RN1e,3)*pow(RN3e,3) + 421./
         1296.*pow(m11,5)*pow(m33,2)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3)
          + 481./1296.*pow(m11,5)*pow(m33,2)*m44*m66*pow(RN1e,3)*pow(
         RN3e,3) );

      M13p6x22R +=  + pi162 * ( 31./324.*pow(m11,5)*pow(m33,2)*pow(
         m44,2)*pow(RN1e,3)*pow(RN3e,3) - 119./432.*pow(m11,5)*pow(
         m33,3)*m66*pow(RN1e,3)*pow(RN3e,3) - 67./432.*pow(m11,5)*pow(
         m33,3)*m44*pow(RN1e,3)*pow(RN3e,3) + 1./12.*pow(m11,5)*pow(
         m33,4)*pow(RN1e,3)*pow(RN3e,3) - 7./486.*pow(m11,6)*pow(m66,3)
         *pow(RN1e,3)*pow(RN3e,3) - 7./324.*pow(m11,6)*m44*pow(m66,2)*
         pow(RN1e,3)*pow(RN3e,3) - 7./648.*pow(m11,6)*pow(m44,2)*m66*
         pow(RN1e,3)*pow(RN3e,3) - 7./3888.*pow(m11,6)*pow(m44,3)*pow(
         RN1e,3)*pow(RN3e,3) + 7./108.*pow(m11,6)*m33*pow(m66,2)*pow(
         RN1e,3)*pow(RN3e,3) + 7./108.*pow(m11,6)*m33*m44*m66*pow(
         RN1e,3)*pow(RN3e,3) + 7./432.*pow(m11,6)*m33*pow(m44,2)*pow(
         RN1e,3)*pow(RN3e,3) - 7./72.*pow(m11,6)*pow(m33,2)*m66*pow(
         RN1e,3)*pow(RN3e,3) - 7./144.*pow(m11,6)*pow(m33,2)*m44*pow(
         RN1e,3)*pow(RN3e,3) + 7./144.*pow(m11,6)*pow(m33,3)*pow(
         RN1e,3)*pow(RN3e,3) + 1./1620.*pi2*pow(me2,2)*RD346e*RSe6613
          - 1./3240.*pi2*pow(me2,2)*RD346e*RSe6633 );

      M13p6x22R +=  + pi162 * (  - 1./3240.*pi2*pow(me2,2)*RSe6611*
         RD346e + 1./1620.*pi2*pow(me2,2)*RD146e*RSe6613 - 1./3240.*pi2
         *pow(me2,2)*RD146e*RSe6633 - 1./3240.*pi2*pow(me2,2)*RD146e*
         RSe6611 + 1./1215.*pi2*pow(me2,3)*pow(RSe6613,2) - 1./1215.*
         pi2*pow(me2,3)*RSe6633*RSe6613 + 1./4860.*pi2*pow(me2,3)*pow(
         RSe6633,2) - 1./1215.*pi2*pow(me2,3)*RSe6611*RSe6613 + 1./2430.
         *pi2*pow(me2,3)*RSe6611*RSe6633 + 1./4860.*pi2*pow(me2,3)*pow(
         RSe6611,2) - 1./270.*pi2*m66*pow(me2,2) + 2./405.*pi2*m66*pow(
         me2,2)*RSe6613 + 1./270.*pi2*m66*pow(me2,2)*RSe63 - 5./162.*
         pi2*m66*pow(me2,2)*pow(RSe63,2) - 7./1620.*pi2*m66*pow(me2,2)*
         RSe6633 + 13./972.*pi2*m66*pow(me2,2)*RS34e*RSe63 + 1./270.*
         pi2*m66*pow(me2,2)*RSe61 + 5./81.*pi2*m66*pow(me2,2)*RSe61*
         RSe63 - 13./972.*pi2*m66*pow(me2,2)*RSe61*RS34e - 5./162.*pi2*
         m66*pow(me2,2)*pow(RSe61,2) - 7./1620.*pi2*m66*pow(me2,2)*
         RSe6611 - 13./972.*pi2*m66*pow(me2,2)*RS14e*RSe63 + 13./972.*
         pi2*m66*pow(me2,2)*RS14e*RSe61 );

      M13p6x22R +=  + pi162 * ( 1./270.*pi2*pow(m66,2)*me2 - 1./540.*
         pi2*pow(m66,2)*me2*RSe63 + 13./1458.*pi2*pow(m66,2)*me2*pow(
         RSe63,2) - 1./540.*pi2*pow(m66,2)*me2*RSe61 - 13./729.*pi2*
         pow(m66,2)*me2*RSe61*RSe63 + 13./1458.*pi2*pow(m66,2)*me2*pow(
         RSe61,2) + 1./36.*pi2*m46*pow(me2,2)*pow(RSe63,2) + 1./162.*
         pi2*m46*pow(me2,2)*RS36e*RSe63 - 1./162.*pi2*m46*pow(me2,2)*
         RS34e*RSe63 - 1./18.*pi2*m46*pow(me2,2)*RSe61*RSe63 - 1./162.*
         pi2*m46*pow(me2,2)*RSe61*RS36e + 1./162.*pi2*m46*pow(me2,2)*
         RSe61*RS34e + 1./36.*pi2*m46*pow(me2,2)*pow(RSe61,2) - 1./162.
         *pi2*m46*pow(me2,2)*RS16e*RSe63 + 1./162.*pi2*m46*pow(me2,2)*
         RS16e*RSe61 + 1./162.*pi2*m46*pow(me2,2)*RS14e*RSe63 - 1./162.
         *pi2*m46*pow(me2,2)*RS14e*RSe61 + 1./243.*pi2*m46*m66*me2*
         RS36e*RSe63 + 1./486.*pi2*m46*m66*me2*RS34e*RSe63 - 1./243.*
         pi2*m46*m66*me2*RSe61*RS36e - 1./486.*pi2*m46*m66*me2*RSe61*
         RS34e - 1./243.*pi2*m46*m66*me2*RS16e*RSe63 + 1./243.*pi2*m46*
         m66*me2*RS16e*RSe61 );

      M13p6x22R +=  + pi162 * (  - 1./486.*pi2*m46*m66*me2*RS14e*RSe63
          + 1./486.*pi2*m46*m66*me2*RS14e*RSe61 - 1./540.*pi2*m44*pow(
         me2,2) + 4./405.*pi2*m44*pow(me2,2)*RSe6613 - 1./270.*pi2*m44*
         pow(me2,2)*RSe63 - 13./648.*pi2*m44*pow(me2,2)*pow(RSe63,2) - 
         7./810.*pi2*m44*pow(me2,2)*RSe6633 - 13./972.*pi2*m44*pow(
         me2,2)*RS36e*RSe63 + 1./324.*pi2*m44*pow(me2,2)*RS34e*RSe63 - 
         1./270.*pi2*m44*pow(me2,2)*RSe61 + 13./324.*pi2*m44*pow(me2,2)
         *RSe61*RSe63 + 13./972.*pi2*m44*pow(me2,2)*RSe61*RS36e - 1./
         324.*pi2*m44*pow(me2,2)*RSe61*RS34e - 13./648.*pi2*m44*pow(
         me2,2)*pow(RSe61,2) - 7./810.*pi2*m44*pow(me2,2)*RSe6611 + 13./
         972.*pi2*m44*pow(me2,2)*RS16e*RSe63 - 13./972.*pi2*m44*pow(
         me2,2)*RS16e*RSe61 - 1./324.*pi2*m44*pow(me2,2)*RS14e*RSe63 + 
         1./324.*pi2*m44*pow(me2,2)*RS14e*RSe61 + 8./729.*pi2*m44*m66*
         me2*pow(RSe63,2) - 16./729.*pi2*m44*m66*me2*RSe61*RSe63 + 8./
         729.*pi2*m44*m66*me2*pow(RSe61,2) + 1./162.*pi2*m44*m46*me2*
         pow(RSe63,2) );

      M13p6x22R +=  + pi162 * ( 1./486.*pi2*m44*m46*me2*RS36e*RSe63 - 1.
         /486.*pi2*m44*m46*me2*RS34e*RSe63 - 1./81.*pi2*m44*m46*me2*
         RSe61*RSe63 - 1./486.*pi2*m44*m46*me2*RSe61*RS36e + 1./486.*
         pi2*m44*m46*me2*RSe61*RS34e + 1./162.*pi2*m44*m46*me2*pow(
         RSe61,2) - 1./486.*pi2*m44*m46*me2*RS16e*RSe63 + 1./486.*pi2*
         m44*m46*me2*RS16e*RSe61 + 1./486.*pi2*m44*m46*me2*RS14e*RSe63
          - 1./486.*pi2*m44*m46*me2*RS14e*RSe61 + 1./540.*pi2*pow(
         m44,2)*me2 + 1./540.*pi2*pow(m44,2)*me2*RSe63 - 35./5832.*pi2*
         pow(m44,2)*me2*pow(RSe63,2) - 1./108.*pi2*pow(m44,2)*me2*RS36e
         *RSe63 + 1./540.*pi2*pow(m44,2)*me2*RSe61 + 35./2916.*pi2*pow(
         m44,2)*me2*RSe61*RSe63 + 1./108.*pi2*pow(m44,2)*me2*RSe61*
         RS36e - 35./5832.*pi2*pow(m44,2)*me2*pow(RSe61,2) + 1./108.*
         pi2*pow(m44,2)*me2*RS16e*RSe63 - 1./108.*pi2*pow(m44,2)*me2*
         RS16e*RSe61 - 1./270.*pi2*m36*pow(me2,2) - 17./1620.*pi2*m36*
         pow(me2,2)*RSe6613 - 7./1620.*pi2*m36*pow(me2,2)*RSe63 + 1./
         135.*pi2*m36*pow(me2,2)*RSe6633 );

      M13p6x22R +=  + pi162 * ( 19./1620.*pi2*m36*pow(me2,2)*RSe61 - 1./
         1620.*pi2*m36*pow(me2,2)*RSe6611 - 1./2592.*pi2*m36*m66*RD346e
          - 1./2592.*pi2*m36*m66*RD146e + 11./1620.*pi2*m36*m66*me2 + 1.
         /648.*pi2*m36*m66*me2*RSe6613 - 1./405.*pi2*m36*m66*me2*RSe63
          + 1./1296.*pi2*m36*m66*me2*RSe6633 - 1./135.*pi2*m36*m66*me2*
         RSe61 + 1./1296.*pi2*m36*m66*me2*RSe6611 - 1./324.*pi2*m36*
         pow(m66,2) - 1./1296.*pi2*m36*pow(m66,2)*RS3461e - 1./1296.*
         pi2*m36*pow(m66,2)*RS1463e + 1./648.*pi2*m36*pow(m66,2)*
         RSe6613 + 1./1296.*pi2*m36*pow(m66,2)*RSe6633 - 1./2592.*pi2*
         m36*pow(m66,2)*RC346e + 1./1296.*pi2*m36*pow(m66,2)*RSe6611 - 
         1./2592.*pi2*m36*pow(m66,2)*RC146e - 1./540.*pi2*m34*pow(
         me2,2) - 17./810.*pi2*m34*pow(me2,2)*RSe6613 + 7./1620.*pi2*
         m34*pow(me2,2)*RSe63 + 2./135.*pi2*m34*pow(me2,2)*RSe6633 - 19.
         /1620.*pi2*m34*pow(me2,2)*RSe61 - 1./810.*pi2*m34*pow(me2,2)*
         RSe6611 - 1./1296.*pi2*m34*m44*RD346e - 1./1296.*pi2*m34*m44*
         RD146e );

      M13p6x22R +=  + pi162 * ( 11./3240.*pi2*m34*m44*me2 + 1./324.*pi2
         *m34*m44*me2*RSe6613 + 1./405.*pi2*m34*m44*me2*RSe63 + 1./648.
         *pi2*m34*m44*me2*RSe6633 + 1./135.*pi2*m34*m44*me2*RSe61 + 1./
         648.*pi2*m34*m44*me2*RSe6611 - 1./648.*pi2*m34*pow(m44,2) - 1./
         648.*pi2*m34*pow(m44,2)*RS3461e - 1./648.*pi2*m34*pow(m44,2)*
         RS1463e + 1./324.*pi2*m34*pow(m44,2)*RSe6613 + 1./648.*pi2*m34
         *pow(m44,2)*RSe6633 - 1./1296.*pi2*m34*pow(m44,2)*RC346e + 1./
         648.*pi2*m34*pow(m44,2)*RSe6611 - 1./1296.*pi2*m34*pow(m44,2)*
         RC146e + 1./6480.*pi2*m33*pow(RD346e,2) + 1./6480.*pi2*m33*
         RD146e*RD346e + 17./4860.*pi2*m33*me2*RD346e*RSe6613 - 23./
         9720.*pi2*m33*me2*RD346e*RSe6633 + 19./9720.*pi2*m33*me2*
         RSe6611*RD346e + 1./4860.*pi2*m33*me2*RD146e*RSe6613 + 23./
         9720.*pi2*m33*me2*RD146e*RSe6633 - 5./1944.*pi2*m33*me2*RD146e
         *RSe6611 - 19./1440.*pi2*m33*pow(me2,2) - 29./2160.*pi2*m33*
         pow(me2,2)*RSe6613 - 17./4860.*pi2*m33*pow(me2,2)*RSe6613*
         RS3461e );

      M13p6x22R +=  + pi162 * (  - 1./324.*pi2*m33*pow(me2,2)*RSe6613*
         RS1463e + 49./2430.*pi2*m33*pow(me2,2)*pow(RSe6613,2) - 1./135.
         *pi2*m33*pow(me2,2)*RSe6633 + 17./9720.*pi2*m33*pow(me2,2)*
         RSe6633*RS3461e + 1./648.*pi2*m33*pow(me2,2)*RSe6633*RS1463e
          - 11./2430.*pi2*m33*pow(me2,2)*RSe6633*RSe6613 + 7./1080.*pi2
         *m33*pow(me2,2)*pow(RSe6633,2) + 17./9720.*pi2*m33*pow(me2,2)*
         RC346e*RSe6613 - 17./19440.*pi2*m33*pow(me2,2)*RC346e*RSe6633
          - 1./180.*pi2*m33*pow(me2,2)*RSe6611 + 17./9720.*pi2*m33*pow(
         me2,2)*RSe6611*RS3461e + 1./648.*pi2*m33*pow(me2,2)*RSe6611*
         RS1463e - 2./405.*pi2*m33*pow(me2,2)*RSe6611*RSe6613 - 13./
         2430.*pi2*m33*pow(me2,2)*RSe6611*RSe6633 - 17./19440.*pi2*m33*
         pow(me2,2)*RSe6611*RC346e + 13./1944.*pi2*m33*pow(me2,2)*pow(
         RSe6611,2) + 1./648.*pi2*m33*pow(me2,2)*RC146e*RSe6613 - 1./
         1296.*pi2*m33*pow(me2,2)*RC146e*RSe6633 - 1./1296.*pi2*m33*
         pow(me2,2)*RC146e*RSe6611 + 7./3240.*pi2*m33*m66*RD346e - 1./
         810.*pi2*m33*m66*me2 );

      M13p6x22R +=  + pi162 * ( 7./19440.*pi2*m33*m66*me2*RSe6613 + 5./
         648.*pi2*m33*m66*me2*RSe63 + 13./972.*pi2*m33*m66*me2*pow(
         RSe63,2) - 89./12960.*pi2*m33*m66*me2*RSe6633 + 1./243.*pi2*
         m33*m66*me2*RS34e*RSe63 - 1./486.*pi2*m33*m66*me2*pow(RS34e,2)
          - 7./3240.*pi2*m33*m66*me2*RSe61 + 13./486.*pi2*m33*m66*me2*
         RSe61*RSe63 - 1./81.*pi2*m33*m66*me2*RSe61*RS34e - 31./7776.*
         pi2*m33*m66*me2*RSe6611 - 1./243.*pi2*m33*m66*me2*RS14e*RSe63
          + 1./162.*pi2*m33*m66*me2*RS14e*RS34e - 1./324.*pi2*m33*pow(
         m66,2)*RSe63 + 1./486.*pi2*m33*pow(m66,2)*RS36e*RSe63 + 1./
         1080.*pi2*m33*pow(m66,2)*RS34e - 23./2916.*pi2*m33*pow(m66,2)*
         RS34e*RSe63 - 1./972.*pi2*m33*pow(m66,2)*RS34e*RS36e + 5./2916.
         *pi2*m33*pow(m66,2)*pow(RS34e,2) + 1./324.*pi2*m33*pow(m66,2)*
         RSe61 - 1./486.*pi2*m33*pow(m66,2)*RSe61*RS36e + 23./2916.*pi2
         *m33*pow(m66,2)*RSe61*RS34e + 1./972.*pi2*m33*pow(m66,2)*RS16e
         *RS34e - 1./648.*pi2*m33*pow(m66,2)*RS14e + 1./972.*pi2*m33*
         pow(m66,2)*RS14e*RS36e );

      M13p6x22R +=  + pi162 * (  - 5./1458.*pi2*m33*pow(m66,2)*RS14e*
         RS34e - 1./54.*pi2*m33*m46*me2*pow(RSe63,2) - 1./27.*pi2*m33*
         m46*me2*RSe61*RSe63 - 1./81.*pi2*m33*m46*me2*RSe61*RS36e + 1./
         81.*pi2*m33*m46*me2*RSe61*RS34e - 1./162.*pi2*m33*m46*me2*
         RS16e*RSe63 + 1./162.*pi2*m33*m46*me2*RS14e*RSe63 - 13./432.*
         pi2*m33*m46*m66 - 1./486.*pi2*m33*m46*m66*RS34e*RS36e - 1./972.
         *pi2*m33*m46*m66*pow(RS34e,2) - 2./243.*pi2*m33*m46*m66*RSe61*
         RS36e - 1./243.*pi2*m33*m46*m66*RSe61*RS34e - 1./243.*pi2*m33*
         m46*m66*RS16e*RSe63 + 1./243.*pi2*m33*m46*m66*RS16e*RS34e - 1./
         486.*pi2*m33*m46*m66*RS14e*RSe63 + 1./243.*pi2*m33*m46*m66*
         RS14e*RS36e + 1./243.*pi2*m33*m46*m66*RS14e*RS34e + 7./1620.*
         pi2*m33*m44*RD346e - 1./1620.*pi2*m33*m44*me2 + 7./9720.*pi2*
         m33*m44*me2*RSe6613 - 5./648.*pi2*m33*m44*me2*RSe63 + 19./1944.
         *pi2*m33*m44*me2*pow(RSe63,2) - 89./6480.*pi2*m33*m44*me2*
         RSe6633 - 1./972.*pi2*m33*m44*me2*RS36e*RSe63 - 1./972.*pi2*
         m33*m44*me2*pow(RS36e,2) );

      M13p6x22R +=  + pi162 * ( 7./3240.*pi2*m33*m44*me2*RSe61 + 19./
         972.*pi2*m33*m44*me2*RSe61*RSe63 + 1./324.*pi2*m33*m44*me2*
         RSe61*RS36e - 31./3888.*pi2*m33*m44*me2*RSe6611 + 1./972.*pi2*
         m33*m44*me2*RS16e*RSe63 + 1./324.*pi2*m33*m44*me2*RS16e*RS36e
          + 5./729.*pi2*m33*m44*m66*RS36e*RSe63 + 1./729.*pi2*m33*m44*
         m66*pow(RS36e,2) - 11./1458.*pi2*m33*m44*m66*RS34e*RSe63 - 1./
         486.*pi2*m33*m44*m66*RS34e*RS36e + 1./729.*pi2*m33*m44*m66*
         pow(RS34e,2) - 5./729.*pi2*m33*m44*m66*RSe61*RS36e + 11./1458.
         *pi2*m33*m44*m66*RSe61*RS34e - 2./729.*pi2*m33*m44*m66*RS16e*
         RS36e + 1./486.*pi2*m33*m44*m66*RS16e*RS34e + 1./486.*pi2*m33*
         m44*m66*RS14e*RS36e - 2./729.*pi2*m33*m44*m66*RS14e*RS34e - 1./
         2187.*pi2*m33*m44*pow(m66,7)*pow(RN1e,3)*pow(RN3e,3) - 13./432.
         *pi2*m33*m44*m46 - 1./324.*pi2*m33*m44*m46*pow(RSe63,2) - 1./
         972.*pi2*m33*m44*m46*pow(RS36e,2) - 1./486.*pi2*m33*m44*m46*
         RS34e*RS36e - 1./162.*pi2*m33*m44*m46*RSe61*RSe63 - 1./243.*
         pi2*m33*m44*m46*RSe61*RS36e );

      M13p6x22R +=  + pi162 * ( 1./243.*pi2*m33*m44*m46*RSe61*RS34e - 1.
         /486.*pi2*m33*m44*m46*RS16e*RSe63 + 1./243.*pi2*m33*m44*m46*
         RS16e*RS36e + 1./243.*pi2*m33*m44*m46*RS16e*RS34e + 1./486.*
         pi2*m33*m44*m46*RS14e*RSe63 + 1./243.*pi2*m33*m44*m46*RS14e*
         RS36e - 13./288.*pi2*m33*pow(m44,2) + 1./324.*pi2*m33*pow(
         m44,2)*RSe63 + 1./216.*pi2*m33*pow(m44,2)*pow(RSe63,2) + 1./
         540.*pi2*m33*pow(m44,2)*RS36e + 13./2916.*pi2*m33*pow(m44,2)*
         RS36e*RSe63 - 5./1458.*pi2*m33*pow(m44,2)*pow(RS36e,2) - 1./
         972.*pi2*m33*pow(m44,2)*RS34e*RSe63 - 1./972.*pi2*m33*pow(
         m44,2)*RS34e*RS36e - 1./324.*pi2*m33*pow(m44,2)*RSe61 + 1./108.
         *pi2*m33*pow(m44,2)*RSe61*RSe63 + 41./2916.*pi2*m33*pow(m44,2)
         *RSe61*RS36e + 1./972.*pi2*m33*pow(m44,2)*RSe61*RS34e - 1./324.
         *pi2*m33*pow(m44,2)*RS16e + 1./108.*pi2*m33*pow(m44,2)*RS16e*
         RSe63 + 47./2916.*pi2*m33*pow(m44,2)*RS16e*RS36e + 1./972.*pi2
         *m33*pow(m44,2)*RS16e*RS34e + 1./972.*pi2*m33*pow(m44,2)*RS14e
         *RS36e );

      M13p6x22R +=  + pi162 * (  - 1./486.*pi2*m33*pow(m44,2)*pow(
         m66,6)*pow(RN1e,3)*pow(RN3e,3) - 5./1458.*pi2*m33*pow(m44,3)*
         pow(m66,5)*pow(RN1e,3)*pow(RN3e,3) - 25./8748.*pi2*m33*pow(
         m44,4)*pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) - 5./3888.*pi2*m33*
         pow(m44,5)*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) - 7./23328.*pi2*
         m33*pow(m44,6)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) - 1./34992.*
         pi2*m33*pow(m44,7)*m66*pow(RN1e,3)*pow(RN3e,3) - 1./270.*pi2*
         m33*m36*RD346e + 1./288.*pi2*m33*m36*RD146e - 1./6480.*pi2*m33
         *m36*me2*RSe6613 + 1./720.*pi2*m33*m36*me2*RSe63 - 1./135.*pi2
         *m33*m36*me2*RSe6633 + 23./1296.*pi2*m33*m36*me2*RSe61 - 1./
         144.*pi2*m33*m36*me2*RSe6611 - 11./1728.*pi2*m33*m36*m66 - 13./
         5184.*pi2*m33*m36*m66*RS3461e - 1./576.*pi2*m33*m36*m66*
         RS1463e + 1./288.*pi2*m33*m36*m66*RSe6613 + 1./864.*pi2*m33*
         m36*m66*RSe63 - 1./1296.*pi2*m33*m36*m66*RSe6633 + 17./25920.*
         pi2*m33*m36*m66*RS34e + 1./96.*pi2*m33*m36*m66*RSe61 + 1./1296.
         *pi2*m33*m36*m66*RSe6611 );

      M13p6x22R +=  + pi162 * (  - 1./192.*pi2*m33*m36*m66*RS14e - 1./
         2592.*pi2*m33*m36*m66*RC146e - 31./2592.*pi2*m33*m36*m44 - 1./
         135.*pi2*m33*m34*RD346e + 1./144.*pi2*m33*m34*RD146e - 1./3240.
         *pi2*m33*m34*me2*RSe6613 - 1./720.*pi2*m33*m34*me2*RSe63 - 2./
         135.*pi2*m33*m34*me2*RSe6633 - 23./1296.*pi2*m33*m34*me2*RSe61
          - 1./72.*pi2*m33*m34*me2*RSe6611 - 31./2592.*pi2*m33*m34*m66
          - 2./81.*pi2*m33*m34*m44 - 13./2592.*pi2*m33*m34*m44*RS3461e
          - 1./288.*pi2*m33*m34*m44*RS1463e + 1./144.*pi2*m33*m34*m44*
         RSe6613 - 1./864.*pi2*m33*m34*m44*RSe63 - 1./648.*pi2*m33*m34*
         m44*RSe6633 + 17./12960.*pi2*m33*m34*m44*RS36e - 1./96.*pi2*
         m33*m34*m44*RSe61 + 1./648.*pi2*m33*m34*m44*RSe6611 - 1./96.*
         pi2*m33*m34*m44*RS16e - 1./1296.*pi2*m33*m34*m44*RC146e + 1./
         270.*pi2*pow(m33,2)*RD346e + 1./3888.*pi2*pow(m33,2)*RD346e*
         RS3461e + 1./432.*pi2*pow(m33,2)*RD346e*RS1463e - 1./216.*pi2*
         pow(m33,2)*RD346e*RSe6613 + 7./1296.*pi2*pow(m33,2)*RD346e*
         RSe6633 );

      M13p6x22R +=  + pi162 * (  - 53./38880.*pi2*pow(m33,2)*RD346e*
         RC346e - 1./1296.*pi2*pow(m33,2)*RSe6611*RD346e + 1./2592.*pi2
         *pow(m33,2)*RC146e*RD346e - 1./2430.*pi2*pow(m33,2)*RD146e*
         RS3461e - 1./972.*pi2*pow(m33,2)*RD146e*RC346e - 1./4320.*pi2*
         pow(m33,2)*me2*RSe6613 - 5./486.*pi2*pow(m33,2)*me2*RSe6613*
         RS3461e + 1./243.*pi2*pow(m33,2)*me2*RSe6613*RS1463e - 2./243.
         *pi2*pow(m33,2)*me2*pow(RSe6613,2) - 37./6480.*pi2*pow(m33,2)*
         me2*RSe6633 - 211./9720.*pi2*pow(m33,2)*me2*RSe6633*RS3461e - 
         41./1944.*pi2*pow(m33,2)*me2*RSe6633*RS1463e + 23./486.*pi2*
         pow(m33,2)*me2*RSe6633*RSe6613 - 11./648.*pi2*pow(m33,2)*me2*
         pow(RSe6633,2) - 31./9720.*pi2*pow(m33,2)*me2*RC346e*RSe6613
          + 67./19440.*pi2*pow(m33,2)*me2*RC346e*RSe6633 - 23./2592.*
         pi2*pow(m33,2)*me2*RSe6611 - 19./9720.*pi2*pow(m33,2)*me2*
         RSe6611*RS3461e - 1./648.*pi2*pow(m33,2)*me2*RSe6611*RS1463e
          + 1./162.*pi2*pow(m33,2)*me2*RSe6611*RSe6613 - 1./972.*pi2*
         pow(m33,2)*me2*RSe6611*RSe6633 );

      M13p6x22R +=  + pi162 * ( 17./3888.*pi2*pow(m33,2)*me2*RSe6611*
         RC346e - 5./1944.*pi2*pow(m33,2)*me2*pow(RSe6611,2) - 1./648.*
         pi2*pow(m33,2)*me2*RC146e*RSe6613 + 1./3888.*pi2*pow(m33,2)*
         me2*RC146e*RSe6633 + 5./3888.*pi2*pow(m33,2)*me2*RC146e*
         RSe6611 - 101./12960.*pi2*pow(m33,2)*m66*RS3461e - 1./486.*pi2
         *pow(m33,2)*m66*RS1463e + 1./243.*pi2*pow(m33,2)*m66*RSe6613
          - 1./162.*pi2*pow(m33,2)*m66*RSe63 + 2./243.*pi2*pow(m33,2)*
         m66*RSe6633 - 1./972.*pi2*pow(m33,2)*m66*pow(RS36e,2) - 1./240.
         *pi2*pow(m33,2)*m66*RS34e - 1./972.*pi2*pow(m33,2)*m66*RS34e*
         RS36e - 5./1944.*pi2*pow(m33,2)*m66*pow(RS34e,2) + 23./15552.*
         pi2*pow(m33,2)*m66*RC346e + 1./81.*pi2*pow(m33,2)*m66*RSe61 - 
         1./54.*pi2*pow(m33,2)*m66*RSe61*RS34e + 1./243.*pi2*pow(m33,2)
         *m66*RS16e*RS36e + 1./486.*pi2*pow(m33,2)*m66*RS16e*RS34e - 1./
         162.*pi2*pow(m33,2)*m66*RS14e + 1./486.*pi2*pow(m33,2)*m66*
         RS14e*RS36e + 5./486.*pi2*pow(m33,2)*m66*RS14e*RS34e + 1./4374.
         *pi2*pow(m33,2)*pow(m66,7)*pow(RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + pi162 * ( 1./972.*pi2*pow(m33,2)*m46*pow(RS36e,2)
          - 1./486.*pi2*pow(m33,2)*m46*RS34e*RS36e + 1./972.*pi2*pow(
         m33,2)*m46*pow(RS34e,2) - 1./81.*pi2*pow(m33,2)*m46*RSe61*
         RS36e + 1./81.*pi2*pow(m33,2)*m46*RSe61*RS34e - 1./243.*pi2*
         pow(m33,2)*m46*RS16e*RS36e + 1./243.*pi2*pow(m33,2)*m46*RS16e*
         RS34e + 1./243.*pi2*pow(m33,2)*m46*RS14e*RS36e - 1./243.*pi2*
         pow(m33,2)*m46*RS14e*RS34e - 101./6480.*pi2*pow(m33,2)*m44*
         RS3461e - 1./243.*pi2*pow(m33,2)*m44*RS1463e + 2./243.*pi2*
         pow(m33,2)*m44*RSe6613 + 1./162.*pi2*pow(m33,2)*m44*RSe63 + 4./
         243.*pi2*pow(m33,2)*m44*RSe6633 - 1./120.*pi2*pow(m33,2)*m44*
         RS36e - 7./972.*pi2*pow(m33,2)*m44*pow(RS36e,2) - 1./972.*pi2*
         pow(m33,2)*m44*RS34e*RS36e - 1./972.*pi2*pow(m33,2)*m44*pow(
         RS34e,2) + 23./7776.*pi2*pow(m33,2)*m44*RC346e - 1./81.*pi2*
         pow(m33,2)*m44*RSe61 + 2./81.*pi2*pow(m33,2)*m44*RSe61*RS36e
          - 1./162.*pi2*pow(m33,2)*m44*RSe61*RS34e - 1./81.*pi2*pow(
         m33,2)*m44*RS16e );

      M13p6x22R +=  + pi162 * ( 7./243.*pi2*pow(m33,2)*m44*RS16e*RS36e
          + 1./486.*pi2*pow(m33,2)*m44*RS16e*RS34e + 1./486.*pi2*pow(
         m33,2)*m44*RS14e*RS36e + 1./243.*pi2*pow(m33,2)*m44*RS14e*
         RS34e + 10./6561.*pi2*pow(m33,2)*m44*pow(m66,6)*pow(RN1e,3)*
         pow(RN3e,3) + 157./26244.*pi2*pow(m33,2)*pow(m44,2)*pow(m66,5)
         *pow(RN1e,3)*pow(RN3e,3) + 235./26244.*pi2*pow(m33,2)*pow(
         m44,3)*pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) + 1315./209952.*pi2*
         pow(m33,2)*pow(m44,4)*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) + 227.
         /104976.*pi2*pow(m33,2)*pow(m44,5)*pow(m66,2)*pow(RN1e,3)*pow(
         RN3e,3) + 35./104976.*pi2*pow(m33,2)*pow(m44,6)*m66*pow(
         RN1e,3)*pow(RN3e,3) + 1./69984.*pi2*pow(m33,2)*pow(m44,7)*pow(
         RN1e,3)*pow(RN3e,3) + 1./1728.*pi2*pow(m33,2)*m36 + 23./4320.*
         pi2*pow(m33,2)*m36*RS3461e + 19./12960.*pi2*pow(m33,2)*m36*
         RS34e + 2./81.*pi2*pow(m33,2)*m36*RSe61 - 1./81.*pi2*pow(
         m33,2)*m36*RS14e + 1./864.*pi2*pow(m33,2)*m34 + 23./2160.*pi2*
         pow(m33,2)*m34*RS3461e );

      M13p6x22R +=  + pi162 * ( 19./6480.*pi2*pow(m33,2)*m34*RS36e - 2./
         81.*pi2*pow(m33,2)*m34*RSe61 - 2./81.*pi2*pow(m33,2)*m34*RS16e
          + 37./1728.*pi2*pow(m33,3) - 223./25920.*pi2*pow(m33,3)*
         RS3461e + 31./9720.*pi2*pow(m33,3)*pow(RS3461e,2) - 5./324.*
         pi2*pow(m33,3)*RS1463e - 1./1944.*pi2*pow(m33,3)*RS1463e*
         RS3461e + 5./162.*pi2*pow(m33,3)*RSe6613 + 1./972.*pi2*pow(
         m33,3)*RSe6613*RS3461e + 1./162.*pi2*pow(m33,3)*RSe6633 - 1./
         972.*pi2*pow(m33,3)*RSe6633*RS3461e + 19./12960.*pi2*pow(
         m33,3)*RS36e + 19./25920.*pi2*pow(m33,3)*RS34e + 1./288.*pi2*
         pow(m33,3)*RC346e + 17./1620.*pi2*pow(m33,3)*RC346e*RS3461e + 
         19./1944.*pi2*pow(m33,3)*RC346e*RS1463e - 19./972.*pi2*pow(
         m33,3)*RC346e*RSe6613 + 5./648.*pi2*pow(m33,3)*RC346e*RSe6633
          - 5./1944.*pi2*pow(m33,3)*pow(RC346e,2) - 1./1944.*pi2*pow(
         m33,3)*RSe6611*RC346e - 1./108.*pi2*pow(m33,3)*RS16e - 1./216.
         *pi2*pow(m33,3)*RS14e + 1./3888.*pi2*pow(m33,3)*RC146e*RC346e
          + 1./6561.*pi2*pow(m33,3)*pow(m66,6)*pow(RN1e,3)*pow(RN3e,3)
          );

      M13p6x22R +=  + pi162 * ( 14./6561.*pi2*pow(m33,3)*m44*pow(m66,5)
         *pow(RN1e,3)*pow(RN3e,3) - 7./26244.*pi2*pow(m33,3)*pow(m44,2)
         *pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) - 23./5832.*pi2*pow(m33,3)
         *pow(m44,3)*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) - 157./52488.*
         pi2*pow(m33,3)*pow(m44,4)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3)
          - 169./209952.*pi2*pow(m33,3)*pow(m44,5)*m66*pow(RN1e,3)*pow(
         RN3e,3) - 7./104976.*pi2*pow(m33,3)*pow(m44,6)*pow(RN1e,3)*
         pow(RN3e,3) - 161./52488.*pi2*pow(m33,4)*pow(m66,5)*pow(
         RN1e,3)*pow(RN3e,3) - 599./52488.*pi2*pow(m33,4)*m44*pow(
         m66,4)*pow(RN1e,3)*pow(RN3e,3) - 2417./209952.*pi2*pow(m33,4)*
         pow(m44,2)*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) - 479./104976.*
         pi2*pow(m33,4)*pow(m44,3)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3)
          - 71./104976.*pi2*pow(m33,4)*pow(m44,4)*m66*pow(RN1e,3)*pow(
         RN3e,3) - 1./52488.*pi2*pow(m33,4)*pow(m44,5)*pow(RN1e,3)*pow(
         RN3e,3) + 89./17496.*pi2*pow(m33,5)*pow(m66,4)*pow(RN1e,3)*
         pow(RN3e,3) );

      M13p6x22R +=  + pi162 * ( 26./2187.*pi2*pow(m33,5)*m44*pow(m66,3)
         *pow(RN1e,3)*pow(RN3e,3) + 223./23328.*pi2*pow(m33,5)*pow(
         m44,2)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) + 223./69984.*pi2*
         pow(m33,5)*pow(m44,3)*m66*pow(RN1e,3)*pow(RN3e,3) + 13./34992.
         *pi2*pow(m33,5)*pow(m44,4)*pow(RN1e,3)*pow(RN3e,3) - 7./2916.*
         pi2*pow(m33,6)*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) - 7./1944.*
         pi2*pow(m33,6)*m44*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) - 7./
         3888.*pi2*pow(m33,6)*pow(m44,2)*m66*pow(RN1e,3)*pow(RN3e,3) - 
         7./23328.*pi2*pow(m33,6)*pow(m44,3)*pow(RN1e,3)*pow(RN3e,3) - 
         1./270.*pi2*m16*pow(me2,2) - 17./1620.*pi2*m16*pow(me2,2)*
         RSe6613 + 19./1620.*pi2*m16*pow(me2,2)*RSe63 - 1./1620.*pi2*
         m16*pow(me2,2)*RSe6633 - 7./1620.*pi2*m16*pow(me2,2)*RSe61 + 1.
         /135.*pi2*m16*pow(me2,2)*RSe6611 - 1./2592.*pi2*m16*m66*RD346e
          - 1./2592.*pi2*m16*m66*RD146e + 11./1620.*pi2*m16*m66*me2 + 1.
         /648.*pi2*m16*m66*me2*RSe6613 - 1./135.*pi2*m16*m66*me2*RSe63
          + 1./1296.*pi2*m16*m66*me2*RSe6633 );

      M13p6x22R +=  + pi162 * (  - 1./405.*pi2*m16*m66*me2*RSe61 + 1./
         1296.*pi2*m16*m66*me2*RSe6611 - 1./324.*pi2*m16*pow(m66,2) - 1.
         /1296.*pi2*m16*pow(m66,2)*RS3461e - 1./1296.*pi2*m16*pow(
         m66,2)*RS1463e + 1./648.*pi2*m16*pow(m66,2)*RSe6613 + 1./1296.
         *pi2*m16*pow(m66,2)*RSe6633 - 1./2592.*pi2*m16*pow(m66,2)*
         RC346e + 1./1296.*pi2*m16*pow(m66,2)*RSe6611 - 1./2592.*pi2*
         m16*pow(m66,2)*RC146e + 11./3240.*pi2*m16*m33*RD346e - 1./2592.
         *pi2*m16*m33*RD146e + 1./324.*pi2*m16*m33*me2 - 11./6480.*pi2*
         m16*m33*me2*RSe6613 + 1./240.*pi2*m16*m33*me2*RSe63 - 1./180.*
         pi2*m16*m33*me2*RSe6633 + 5./1296.*pi2*m16*m33*me2*RSe61 + 1./
         1296.*pi2*m16*m33*me2*RSe6611 - 49./5184.*pi2*m16*m33*m66 - 17.
         /5184.*pi2*m16*m33*m66*RS3461e - 13./5184.*pi2*m16*m33*m66*
         RS1463e + 13./2592.*pi2*m16*m33*m66*RSe6613 + 19./2592.*pi2*
         m16*m33*m66*RSe63 + 1./25920.*pi2*m16*m33*m66*RS34e - 1./2592.
         *pi2*m16*m33*m66*RC346e - 5./2592.*pi2*m16*m33*m66*RSe61 + 1./
         648.*pi2*m16*m33*m66*RSe6611 );

      M13p6x22R +=  + pi162 * ( 5./5184.*pi2*m16*m33*m66*RS14e - 1./
         1296.*pi2*m16*m33*m66*RC146e - 31./2592.*pi2*m16*m33*m44 - 31./
         2592.*pi2*m16*pow(m33,2) + 113./25920.*pi2*m16*pow(m33,2)*
         RS3461e - 1./576.*pi2*m16*pow(m33,2)*RS1463e + 1./288.*pi2*m16
         *pow(m33,2)*RSe6613 + 1./864.*pi2*m16*pow(m33,2)*RSe63 - 1./
         1296.*pi2*m16*pow(m33,2)*RSe6633 - 221./25920.*pi2*m16*pow(
         m33,2)*RS34e + 1./288.*pi2*m16*pow(m33,2)*RC346e - 5./2592.*
         pi2*m16*pow(m33,2)*RSe61 + 1./1296.*pi2*m16*pow(m33,2)*RSe6611
          + 5./5184.*pi2*m16*pow(m33,2)*RS14e - 1./2592.*pi2*m16*pow(
         m33,2)*RC146e - 1./540.*pi2*m14*pow(me2,2) - 17./810.*pi2*m14*
         pow(me2,2)*RSe6613 - 19./1620.*pi2*m14*pow(me2,2)*RSe63 - 1./
         810.*pi2*m14*pow(me2,2)*RSe6633 + 7./1620.*pi2*m14*pow(me2,2)*
         RSe61 + 2./135.*pi2*m14*pow(me2,2)*RSe6611 - 1./1296.*pi2*m14*
         m44*RD346e - 1./1296.*pi2*m14*m44*RD146e + 11./3240.*pi2*m14*
         m44*me2 + 1./324.*pi2*m14*m44*me2*RSe6613 + 1./135.*pi2*m14*
         m44*me2*RSe63 );

      M13p6x22R +=  + pi162 * ( 1./648.*pi2*m14*m44*me2*RSe6633 + 1./
         405.*pi2*m14*m44*me2*RSe61 + 1./648.*pi2*m14*m44*me2*RSe6611
          - 1./648.*pi2*m14*pow(m44,2) - 1./648.*pi2*m14*pow(m44,2)*
         RS3461e - 1./648.*pi2*m14*pow(m44,2)*RS1463e + 1./324.*pi2*m14
         *pow(m44,2)*RSe6613 + 1./648.*pi2*m14*pow(m44,2)*RSe6633 - 1./
         1296.*pi2*m14*pow(m44,2)*RC346e + 1./648.*pi2*m14*pow(m44,2)*
         RSe6611 - 1./1296.*pi2*m14*pow(m44,2)*RC146e + 11./1620.*pi2*
         m14*m33*RD346e - 1./1296.*pi2*m14*m33*RD146e + 1./648.*pi2*m14
         *m33*me2 - 11./3240.*pi2*m14*m33*me2*RSe6613 - 1./240.*pi2*m14
         *m33*me2*RSe63 - 1./90.*pi2*m14*m33*me2*RSe6633 - 5./1296.*pi2
         *m14*m33*me2*RSe61 + 1./648.*pi2*m14*m33*me2*RSe6611 - 31./
         2592.*pi2*m14*m33*m66 - 17./648.*pi2*m14*m33*m44 - 17./2592.*
         pi2*m14*m33*m44*RS3461e - 13./2592.*pi2*m14*m33*m44*RS1463e + 
         13./1296.*pi2*m14*m33*m44*RSe6613 - 19./2592.*pi2*m14*m33*m44*
         RSe63 + 1./12960.*pi2*m14*m33*m44*RS36e - 1./1296.*pi2*m14*m33
         *m44*RC346e );

      M13p6x22R +=  + pi162 * ( 5./2592.*pi2*m14*m33*m44*RSe61 + 1./324.
         *pi2*m14*m33*m44*RSe6611 + 5./2592.*pi2*m14*m33*m44*RS16e - 1./
         648.*pi2*m14*m33*m44*RC146e - 31./1296.*pi2*m14*pow(m33,2) + 
         113./12960.*pi2*m14*pow(m33,2)*RS3461e - 1./288.*pi2*m14*pow(
         m33,2)*RS1463e + 1./144.*pi2*m14*pow(m33,2)*RSe6613 - 1./864.*
         pi2*m14*pow(m33,2)*RSe63 - 1./648.*pi2*m14*pow(m33,2)*RSe6633
          - 221./12960.*pi2*m14*pow(m33,2)*RS36e + 1./144.*pi2*m14*pow(
         m33,2)*RC346e + 5./2592.*pi2*m14*pow(m33,2)*RSe61 + 1./648.*
         pi2*m14*pow(m33,2)*RSe6611 + 5./2592.*pi2*m14*pow(m33,2)*RS16e
          - 1./1296.*pi2*m14*pow(m33,2)*RC146e + 1./3888.*pi2*m13*pow(
         RD346e,2) + 1./1944.*pi2*m13*RD146e*RD346e + 1./3888.*pi2*m13*
         pow(RD146e,2) + 1./486.*pi2*m13*me2*RD346e*RSe6613 - 1./972.*
         pi2*m13*me2*RD346e*RSe6633 - 1./972.*pi2*m13*me2*RSe6611*
         RD346e + 1./486.*pi2*m13*me2*RD146e*RSe6613 - 1./972.*pi2*m13*
         me2*RD146e*RSe6633 - 1./972.*pi2*m13*me2*RD146e*RSe6611 - 1./
         162.*pi2*m13*pow(me2,2)*RSe6613*RS3461e );

      M13p6x22R +=  + pi162 * (  - 1./162.*pi2*m13*pow(me2,2)*RSe6613*
         RS1463e + 4./243.*pi2*m13*pow(me2,2)*pow(RSe6613,2) + 1./324.*
         pi2*m13*pow(me2,2)*RSe6633*RS3461e + 1./324.*pi2*m13*pow(
         me2,2)*RSe6633*RS1463e - 4./243.*pi2*m13*pow(me2,2)*RSe6633*
         RSe6613 + 1./243.*pi2*m13*pow(me2,2)*pow(RSe6633,2) + 1./324.*
         pi2*m13*pow(me2,2)*RC346e*RSe6613 - 1./648.*pi2*m13*pow(me2,2)
         *RC346e*RSe6633 + 1./324.*pi2*m13*pow(me2,2)*RSe6611*RS3461e
          + 1./324.*pi2*m13*pow(me2,2)*RSe6611*RS1463e - 4./243.*pi2*
         m13*pow(me2,2)*RSe6611*RSe6613 + 2./243.*pi2*m13*pow(me2,2)*
         RSe6611*RSe6633 - 1./648.*pi2*m13*pow(me2,2)*RSe6611*RC346e + 
         1./243.*pi2*m13*pow(me2,2)*pow(RSe6611,2) + 1./324.*pi2*m13*
         pow(me2,2)*RC146e*RSe6613 - 1./648.*pi2*m13*pow(me2,2)*RC146e*
         RSe6633 - 1./648.*pi2*m13*pow(me2,2)*RC146e*RSe6611 - 1./648.*
         pi2*m13*m33*RD346e*RS3461e - 1./1944.*pi2*m13*m33*RD346e*
         RS1463e + 1./972.*pi2*m13*m33*RD346e*RSe6613 + 1./648.*pi2*m13
         *m33*RD346e*RSe6633 );

      M13p6x22R +=  + pi162 * (  - 1./3888.*pi2*m13*m33*RD346e*RC346e
          - 5./1944.*pi2*m13*m33*RSe6611*RD346e + 5./3888.*pi2*m13*m33*
         RC146e*RD346e + 1./1944.*pi2*m13*m33*RD146e*RS3461e + 1./648.*
         pi2*m13*m33*RD146e*RS1463e - 1./324.*pi2*m13*m33*RD146e*
         RSe6613 + 1./1944.*pi2*m13*m33*RD146e*RSe6633 + 1./3888.*pi2*
         m13*m33*RD146e*RC346e + 5./1944.*pi2*m13*m33*RD146e*RSe6611 - 
         5./3888.*pi2*m13*m33*RD146e*RC146e + 1./243.*pi2*m13*m33*me2*
         RSe6613*RS3461e + 2./243.*pi2*m13*m33*me2*RSe6613*RS1463e - 4./
         243.*pi2*m13*m33*me2*pow(RSe6613,2) - 1./324.*pi2*m13*m33*me2*
         RSe6633*RS3461e - 5./972.*pi2*m13*m33*me2*RSe6633*RS1463e + 5./
         243.*pi2*m13*m33*me2*RSe6633*RSe6613 - 1./108.*pi2*m13*m33*me2
         *pow(RSe6633,2) - 1./324.*pi2*m13*m33*me2*RC346e*RSe6613 + 7./
         1944.*pi2*m13*m33*me2*RC346e*RSe6633 - 1./972.*pi2*m13*m33*me2
         *RSe6611*RS3461e - 1./324.*pi2*m13*m33*me2*RSe6611*RS1463e + 1.
         /81.*pi2*m13*m33*me2*RSe6611*RSe6613 - 1./486.*pi2*m13*m33*me2
         *RSe6611*RSe6633 );

      M13p6x22R +=  + pi162 * (  - 1./1944.*pi2*m13*m33*me2*RSe6611*
         RC346e - 5./972.*pi2*m13*m33*me2*pow(RSe6611,2) - 1./324.*pi2*
         m13*m33*me2*RC146e*RSe6613 + 1./1944.*pi2*m13*m33*me2*RC146e*
         RSe6633 + 5./1944.*pi2*m13*m33*me2*RC146e*RSe6611 - 31./864.*
         pi2*m13*pow(m33,2) - 1./972.*pi2*m13*pow(m33,2)*RS1463e*
         RS3461e + 1./486.*pi2*m13*pow(m33,2)*RSe6613*RS3461e - 1./486.
         *pi2*m13*pow(m33,2)*RSe6633*RS3461e + 1./972.*pi2*m13*pow(
         m33,2)*RC346e*RS3461e + 1./972.*pi2*m13*pow(m33,2)*RC346e*
         RS1463e - 1./486.*pi2*m13*pow(m33,2)*RC346e*RSe6613 + 1./324.*
         pi2*m13*pow(m33,2)*RC346e*RSe6633 - 5./3888.*pi2*m13*pow(
         m33,2)*pow(RC346e,2) - 1./972.*pi2*m13*pow(m33,2)*RSe6611*
         RC346e + 1./1944.*pi2*m13*pow(m33,2)*RC146e*RC346e + 1./6480.*
         pi2*m11*RD146e*RD346e + 1./6480.*pi2*m11*pow(RD146e,2) + 1./
         4860.*pi2*m11*me2*RD346e*RSe6613 - 5./1944.*pi2*m11*me2*RD346e
         *RSe6633 + 23./9720.*pi2*m11*me2*RSe6611*RD346e + 17./4860.*
         pi2*m11*me2*RD146e*RSe6613 );

      M13p6x22R +=  + pi162 * ( 19./9720.*pi2*m11*me2*RD146e*RSe6633 - 
         23./9720.*pi2*m11*me2*RD146e*RSe6611 - 19./1440.*pi2*m11*pow(
         me2,2) - 29./2160.*pi2*m11*pow(me2,2)*RSe6613 - 1./324.*pi2*
         m11*pow(me2,2)*RSe6613*RS3461e - 17./4860.*pi2*m11*pow(me2,2)*
         RSe6613*RS1463e + 49./2430.*pi2*m11*pow(me2,2)*pow(RSe6613,2)
          - 1./180.*pi2*m11*pow(me2,2)*RSe6633 + 1./648.*pi2*m11*pow(
         me2,2)*RSe6633*RS3461e + 17./9720.*pi2*m11*pow(me2,2)*RSe6633*
         RS1463e - 2./405.*pi2*m11*pow(me2,2)*RSe6633*RSe6613 + 13./
         1944.*pi2*m11*pow(me2,2)*pow(RSe6633,2) + 1./648.*pi2*m11*pow(
         me2,2)*RC346e*RSe6613 - 1./1296.*pi2*m11*pow(me2,2)*RC346e*
         RSe6633 - 1./135.*pi2*m11*pow(me2,2)*RSe6611 + 1./648.*pi2*m11
         *pow(me2,2)*RSe6611*RS3461e + 17./9720.*pi2*m11*pow(me2,2)*
         RSe6611*RS1463e - 11./2430.*pi2*m11*pow(me2,2)*RSe6611*RSe6613
          - 13./2430.*pi2*m11*pow(me2,2)*RSe6611*RSe6633 - 1./1296.*pi2
         *m11*pow(me2,2)*RSe6611*RC346e + 7./1080.*pi2*m11*pow(me2,2)*
         pow(RSe6611,2) );

      M13p6x22R +=  + pi162 * ( 17./9720.*pi2*m11*pow(me2,2)*RC146e*
         RSe6613 - 17./19440.*pi2*m11*pow(me2,2)*RC146e*RSe6633 - 17./
         19440.*pi2*m11*pow(me2,2)*RC146e*RSe6611 + 7./3240.*pi2*m11*
         m66*RD146e - 1./810.*pi2*m11*m66*me2 + 7./19440.*pi2*m11*m66*
         me2*RSe6613 - 7./3240.*pi2*m11*m66*me2*RSe63 - 31./7776.*pi2*
         m11*m66*me2*RSe6633 + 5./648.*pi2*m11*m66*me2*RSe61 + 13./486.
         *pi2*m11*m66*me2*RSe61*RSe63 - 1./243.*pi2*m11*m66*me2*RSe61*
         RS34e + 13./972.*pi2*m11*m66*me2*pow(RSe61,2) - 89./12960.*pi2
         *m11*m66*me2*RSe6611 - 1./81.*pi2*m11*m66*me2*RS14e*RSe63 + 1./
         243.*pi2*m11*m66*me2*RS14e*RS34e + 1./243.*pi2*m11*m66*me2*
         RS14e*RSe61 - 1./486.*pi2*m11*m66*me2*pow(RS14e,2) + 1./324.*
         pi2*m11*pow(m66,2)*RSe63 - 1./324.*pi2*m11*pow(m66,2)*RSe61 - 
         1./486.*pi2*m11*pow(m66,2)*RS16e*RSe63 + 1./486.*pi2*m11*pow(
         m66,2)*RS16e*RSe61 + 1./405.*pi2*m11*pow(m66,2)*RS14e + 23./
         2916.*pi2*m11*pow(m66,2)*RS14e*RSe63 - 23./2916.*pi2*m11*pow(
         m66,2)*RS14e*RSe61 );

      M13p6x22R +=  + pi162 * (  - 1./972.*pi2*m11*pow(m66,2)*RS14e*
         RS16e + 5./2916.*pi2*m11*pow(m66,2)*pow(RS14e,2) - 1./27.*pi2*
         m11*m46*me2*RSe61*RSe63 - 1./162.*pi2*m11*m46*me2*RSe61*RS36e
          + 1./162.*pi2*m11*m46*me2*RSe61*RS34e - 1./54.*pi2*m11*m46*
         me2*pow(RSe61,2) - 1./81.*pi2*m11*m46*me2*RS16e*RSe63 + 1./81.
         *pi2*m11*m46*me2*RS14e*RSe63 - 13./432.*pi2*m11*m46*m66 - 1./
         243.*pi2*m11*m46*m66*RSe61*RS36e - 1./486.*pi2*m11*m46*m66*
         RSe61*RS34e - 2./243.*pi2*m11*m46*m66*RS16e*RSe63 + 1./243.*
         pi2*m11*m46*m66*RS16e*RS34e - 1./243.*pi2*m11*m46*m66*RS14e*
         RSe63 + 1./243.*pi2*m11*m46*m66*RS14e*RS36e + 1./243.*pi2*m11*
         m46*m66*RS14e*RS34e - 1./486.*pi2*m11*m46*m66*RS14e*RS16e - 1./
         972.*pi2*m11*m46*m66*pow(RS14e,2) + 7./1620.*pi2*m11*m44*
         RD146e - 1./1620.*pi2*m11*m44*me2 + 7./9720.*pi2*m11*m44*me2*
         RSe6613 + 7./3240.*pi2*m11*m44*me2*RSe63 - 31./3888.*pi2*m11*
         m44*me2*RSe6633 - 5./648.*pi2*m11*m44*me2*RSe61 + 19./972.*pi2
         *m11*m44*me2*RSe61*RSe63 );

      M13p6x22R +=  + pi162 * ( 1./972.*pi2*m11*m44*me2*RSe61*RS36e + 
         19./1944.*pi2*m11*m44*me2*pow(RSe61,2) - 89./6480.*pi2*m11*m44
         *me2*RSe6611 + 1./324.*pi2*m11*m44*me2*RS16e*RSe63 + 1./486.*
         pi2*m11*m44*me2*RS16e*RS36e - 1./972.*pi2*m11*m44*me2*RS16e*
         RSe61 - 1./972.*pi2*m11*m44*me2*pow(RS16e,2) - 5./729.*pi2*m11
         *m44*m66*RS16e*RSe63 + 5./729.*pi2*m11*m44*m66*RS16e*RSe61 + 1.
         /729.*pi2*m11*m44*m66*pow(RS16e,2) + 11./1458.*pi2*m11*m44*m66
         *RS14e*RSe63 - 11./1458.*pi2*m11*m44*m66*RS14e*RSe61 - 1./486.
         *pi2*m11*m44*m66*RS14e*RS16e + 1./729.*pi2*m11*m44*m66*pow(
         RS14e,2) - 1./2187.*pi2*m11*m44*pow(m66,7)*pow(RN1e,3)*pow(
         RN3e,3) - 13./432.*pi2*m11*m44*m46 - 1./162.*pi2*m11*m44*m46*
         RSe61*RSe63 - 1./486.*pi2*m11*m44*m46*RSe61*RS36e + 1./486.*
         pi2*m11*m44*m46*RSe61*RS34e - 1./324.*pi2*m11*m44*m46*pow(
         RSe61,2) - 1./243.*pi2*m11*m44*m46*RS16e*RSe63 + 1./243.*pi2*
         m11*m44*m46*RS16e*RS36e + 1./243.*pi2*m11*m44*m46*RS16e*RS34e
          - 1./972.*pi2*m11*m44*m46*pow(RS16e,2) );

      M13p6x22R +=  + pi162 * ( 1./243.*pi2*m11*m44*m46*RS14e*RSe63 + 1.
         /243.*pi2*m11*m44*m46*RS14e*RS36e - 1./486.*pi2*m11*m44*m46*
         RS14e*RS16e - 13./288.*pi2*m11*pow(m44,2) - 1./324.*pi2*m11*
         pow(m44,2)*RSe63 + 1./324.*pi2*m11*pow(m44,2)*RSe61 + 1./108.*
         pi2*m11*pow(m44,2)*RSe61*RSe63 + 1./108.*pi2*m11*pow(m44,2)*
         RSe61*RS36e + 1./216.*pi2*m11*pow(m44,2)*pow(RSe61,2) + 2./405.
         *pi2*m11*pow(m44,2)*RS16e + 41./2916.*pi2*m11*pow(m44,2)*RS16e
         *RSe63 + 1./54.*pi2*m11*pow(m44,2)*RS16e*RS36e + 13./2916.*pi2
         *m11*pow(m44,2)*RS16e*RSe61 - 5./1458.*pi2*m11*pow(m44,2)*pow(
         RS16e,2) + 1./972.*pi2*m11*pow(m44,2)*RS14e*RSe63 - 1./972.*
         pi2*m11*pow(m44,2)*RS14e*RSe61 - 1./972.*pi2*m11*pow(m44,2)*
         RS14e*RS16e - 1./486.*pi2*m11*pow(m44,2)*pow(m66,6)*pow(
         RN1e,3)*pow(RN3e,3) - 5./1458.*pi2*m11*pow(m44,3)*pow(m66,5)*
         pow(RN1e,3)*pow(RN3e,3) - 25./8748.*pi2*m11*pow(m44,4)*pow(
         m66,4)*pow(RN1e,3)*pow(RN3e,3) - 5./3888.*pi2*m11*pow(m44,5)*
         pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + pi162 * (  - 7./23328.*pi2*m11*pow(m44,6)*pow(
         m66,2)*pow(RN1e,3)*pow(RN3e,3) - 1./34992.*pi2*m11*pow(m44,7)*
         m66*pow(RN1e,3)*pow(RN3e,3) - 1./2592.*pi2*m11*m36*RD346e + 11.
         /3240.*pi2*m11*m36*RD146e + 1./324.*pi2*m11*m36*me2 - 11./6480.
         *pi2*m11*m36*me2*RSe6613 + 5./1296.*pi2*m11*m36*me2*RSe63 + 1./
         1296.*pi2*m11*m36*me2*RSe6633 + 1./240.*pi2*m11*m36*me2*RSe61
          - 1./180.*pi2*m11*m36*me2*RSe6611 - 49./5184.*pi2*m11*m36*m66
          - 13./5184.*pi2*m11*m36*m66*RS3461e - 17./5184.*pi2*m11*m36*
         m66*RS1463e + 13./2592.*pi2*m11*m36*m66*RSe6613 - 5./2592.*pi2
         *m11*m36*m66*RSe63 + 1./648.*pi2*m11*m36*m66*RSe6633 + 5./5184.
         *pi2*m11*m36*m66*RS34e - 1./1296.*pi2*m11*m36*m66*RC346e + 19./
         2592.*pi2*m11*m36*m66*RSe61 + 1./25920.*pi2*m11*m36*m66*RS14e
          - 1./2592.*pi2*m11*m36*m66*RC146e - 31./2592.*pi2*m11*m36*m44
          - 1./1296.*pi2*m11*m34*RD346e + 11./1620.*pi2*m11*m34*RD146e
          + 1./648.*pi2*m11*m34*me2 - 11./3240.*pi2*m11*m34*me2*RSe6613
          - 5./1296.*pi2*m11*m34*me2*RSe63 );

      M13p6x22R +=  + pi162 * ( 1./648.*pi2*m11*m34*me2*RSe6633 - 1./
         240.*pi2*m11*m34*me2*RSe61 - 1./90.*pi2*m11*m34*me2*RSe6611 - 
         31./2592.*pi2*m11*m34*m66 - 17./648.*pi2*m11*m34*m44 - 13./
         2592.*pi2*m11*m34*m44*RS3461e - 17./2592.*pi2*m11*m34*m44*
         RS1463e + 13./1296.*pi2*m11*m34*m44*RSe6613 + 5./2592.*pi2*m11
         *m34*m44*RSe63 + 1./324.*pi2*m11*m34*m44*RSe6633 + 5./2592.*
         pi2*m11*m34*m44*RS36e - 1./648.*pi2*m11*m34*m44*RC346e - 19./
         2592.*pi2*m11*m34*m44*RSe61 + 1./12960.*pi2*m11*m34*m44*RS16e
          - 1./1296.*pi2*m11*m34*m44*RC146e + 1./360.*pi2*m11*m33*
         RD346e - 17./19440.*pi2*m11*m33*RD346e*RS3461e - 49./19440.*
         pi2*m11*m33*RD346e*RS1463e + 1./648.*pi2*m11*m33*RD346e*
         RSe6613 - 1./1296.*pi2*m11*m33*RD346e*RSe6633 + 13./7776.*pi2*
         m11*m33*RD346e*RC346e - 1./1296.*pi2*m11*m33*RSe6611*RD346e - 
         23./38880.*pi2*m11*m33*RC146e*RD346e + 1./360.*pi2*m11*m33*
         RD146e + 11./19440.*pi2*m11*m33*RD146e*RS3461e - 17./19440.*
         pi2*m11*m33*RD146e*RS1463e );

      M13p6x22R +=  + pi162 * ( 1./648.*pi2*m11*m33*RD146e*RSe6613 - 1./
         1296.*pi2*m11*m33*RD146e*RSe6633 - 23./38880.*pi2*m11*m33*
         RD146e*RC346e - 1./1296.*pi2*m11*m33*RD146e*RSe6611 + 13./7776.
         *pi2*m11*m33*RD146e*RC146e - 163./6480.*pi2*m11*m33*me2*
         RSe6613 - 7./1215.*pi2*m11*m33*me2*RSe6613*RS3461e - 7./1215.*
         pi2*m11*m33*me2*RSe6613*RS1463e - 4./243.*pi2*m11*m33*me2*pow(
         RSe6613,2) - 53./12960.*pi2*m11*m33*me2*RSe6633 - 53./4860.*
         pi2*m11*m33*me2*RSe6633*RS3461e - 1./90.*pi2*m11*m33*me2*
         RSe6633*RS1463e + 7./243.*pi2*m11*m33*me2*RSe6633*RSe6613 - 7./
         972.*pi2*m11*m33*me2*pow(RSe6633,2) - 23./4860.*pi2*m11*m33*
         me2*RC346e*RSe6613 - 1./648.*pi2*m11*m33*me2*RC346e*RSe6633 - 
         53./12960.*pi2*m11*m33*me2*RSe6611 - 7./405.*pi2*m11*m33*me2*
         RSe6611*RS3461e - 53./4860.*pi2*m11*m33*me2*RSe6611*RS1463e + 
         7./243.*pi2*m11*m33*me2*RSe6611*RSe6613 - 1./486.*pi2*m11*m33*
         me2*RSe6611*RSe6633 + 23./4860.*pi2*m11*m33*me2*RSe6611*RC346e
          - 7./972.*pi2*m11*m33*me2*pow(RSe6611,2) );

      M13p6x22R +=  + pi162 * (  - 23./4860.*pi2*m11*m33*me2*RC146e*
         RSe6613 + 23./4860.*pi2*m11*m33*me2*RC146e*RSe6633 - 1./648.*
         pi2*m11*m33*me2*RC146e*RSe6611 - 29./12960.*pi2*m11*m33*m66*
         RS3461e - 247./38880.*pi2*m11*m33*m66*RS1463e + 1./81.*pi2*m11
         *m33*m66*RSe6613 + 1./324.*pi2*m11*m33*m66*RSe63 - 23./6480.*
         pi2*m11*m33*m66*RS34e + 31./15552.*pi2*m11*m33*m66*RC346e + 1./
         324.*pi2*m11*m33*m66*RSe61 - 1./108.*pi2*m11*m33*m66*RSe61*
         RS34e + 1./486.*pi2*m11*m33*m66*RS16e*RS36e + 1./972.*pi2*m11*
         m33*m66*RS16e*RS34e - 1./2160.*pi2*m11*m33*m66*RS14e - 1./108.
         *pi2*m11*m33*m66*RS14e*RSe63 + 1./972.*pi2*m11*m33*m66*RS14e*
         RS36e + 5./972.*pi2*m11*m33*m66*RS14e*RS34e + 31./15552.*pi2*
         m11*m33*m66*RC146e + 1./2187.*pi2*m11*m33*pow(m66,7)*pow(
         RN1e,3)*pow(RN3e,3) - 1./162.*pi2*m11*m33*m46*RSe61*RS36e + 1./
         162.*pi2*m11*m33*m46*RSe61*RS34e - 1./162.*pi2*m11*m33*m46*
         RS16e*RSe63 - 1./162.*pi2*m11*m33*m46*RS16e*RS36e + 1./162.*
         pi2*m11*m33*m46*RS16e*RS34e );

      M13p6x22R +=  + pi162 * ( 1./162.*pi2*m11*m33*m46*RS14e*RSe63 + 1.
         /162.*pi2*m11*m33*m46*RS14e*RS36e - 1./162.*pi2*m11*m33*m46*
         RS14e*RS34e - 29./6480.*pi2*m11*m33*m44*RS3461e - 247./19440.*
         pi2*m11*m33*m44*RS1463e + 2./81.*pi2*m11*m33*m44*RSe6613 - 1./
         324.*pi2*m11*m33*m44*RSe63 - 23./3240.*pi2*m11*m33*m44*RS36e
          + 31./7776.*pi2*m11*m33*m44*RC346e - 1./324.*pi2*m11*m33*m44*
         RSe61 + 1./81.*pi2*m11*m33*m44*RSe61*RS36e - 1./324.*pi2*m11*
         m33*m44*RSe61*RS34e - 1./1080.*pi2*m11*m33*m44*RS16e + 1./81.*
         pi2*m11*m33*m44*RS16e*RSe63 + 7./486.*pi2*m11*m33*m44*RS16e*
         RS36e + 1./972.*pi2*m11*m33*m44*RS16e*RS34e - 1./324.*pi2*m11*
         m33*m44*RS14e*RSe63 + 1./972.*pi2*m11*m33*m44*RS14e*RS36e + 1./
         486.*pi2*m11*m33*m44*RS14e*RS34e + 31./7776.*pi2*m11*m33*m44*
         RC146e + 20./6561.*pi2*m11*m33*m44*pow(m66,6)*pow(RN1e,3)*pow(
         RN3e,3) + 157./13122.*pi2*m11*m33*pow(m44,2)*pow(m66,5)*pow(
         RN1e,3)*pow(RN3e,3) + 235./13122.*pi2*m11*m33*pow(m44,3)*pow(
         m66,4)*pow(RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + pi162 * ( 1315./104976.*pi2*m11*m33*pow(m44,4)*
         pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) + 227./52488.*pi2*m11*m33*
         pow(m44,5)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) + 35./52488.*pi2
         *m11*m33*pow(m44,6)*m66*pow(RN1e,3)*pow(RN3e,3) + 1./34992.*
         pi2*m11*m33*pow(m44,7)*pow(RN1e,3)*pow(RN3e,3) - 59./5184.*pi2
         *m11*m33*m36 - 23./25920.*pi2*m11*m33*m36*RS3461e - 43./25920.
         *pi2*m11*m33*m36*RS1463e + 1./288.*pi2*m11*m33*m36*RSe6613 + 1.
         /864.*pi2*m11*m33*m36*RSe63 - 1./1296.*pi2*m11*m33*m36*RSe6633
          - 13./5184.*pi2*m11*m33*m36*RS34e + 1./96.*pi2*m11*m33*m36*
         RSe61 + 1./1296.*pi2*m11*m33*m36*RSe6611 - 73./5184.*pi2*m11*
         m33*m36*RS14e + 1./324.*pi2*m11*m33*m36*RC146e - 59./2592.*pi2
         *m11*m33*m34 - 23./12960.*pi2*m11*m33*m34*RS3461e - 43./12960.
         *pi2*m11*m33*m34*RS1463e + 1./144.*pi2*m11*m33*m34*RSe6613 - 1.
         /864.*pi2*m11*m33*m34*RSe63 - 1./648.*pi2*m11*m33*m34*RSe6633
          - 13./2592.*pi2*m11*m33*m34*RS36e - 1./96.*pi2*m11*m33*m34*
         RSe61 );

      M13p6x22R +=  + pi162 * ( 1./648.*pi2*m11*m33*m34*RSe6611 - 73./
         2592.*pi2*m11*m33*m34*RS16e + 1./162.*pi2*m11*m33*m34*RC146e
          + 17./576.*pi2*m11*pow(m33,2) + 1./1296.*pi2*m11*pow(m33,2)*
         RS3461e + 1./1080.*pi2*m11*pow(m33,2)*pow(RS3461e,2) - 79./
         8640.*pi2*m11*pow(m33,2)*RS1463e + 5./1944.*pi2*m11*pow(m33,2)
         *RS1463e*RS3461e - 1./486.*pi2*m11*pow(m33,2)*pow(RS1463e,2)
          + 1./54.*pi2*m11*pow(m33,2)*RSe6613 + 5./972.*pi2*m11*pow(
         m33,2)*RSe6613*RS3461e + 1./243.*pi2*m11*pow(m33,2)*RSe6613*
         RS1463e - 5./1944.*pi2*m11*pow(m33,2)*RSe6633*RS3461e - 5./
         1944.*pi2*m11*pow(m33,2)*RSe6633*RS1463e - 7./540.*pi2*m11*
         pow(m33,2)*RS36e - 7./1080.*pi2*m11*pow(m33,2)*RS34e + 25./
         5184.*pi2*m11*pow(m33,2)*RC346e + 29./4860.*pi2*m11*pow(m33,2)
         *RC346e*RS3461e + 29./4860.*pi2*m11*pow(m33,2)*RC346e*RS1463e
          - 17./1944.*pi2*m11*pow(m33,2)*RC346e*RSe6613 + 11./3888.*pi2
         *m11*pow(m33,2)*RC346e*RSe6633 - 1./7776.*pi2*m11*pow(m33,2)*
         pow(RC346e,2) );

      M13p6x22R +=  + pi162 * (  - 5./1944.*pi2*m11*pow(m33,2)*RSe6611*
         RS3461e - 1./648.*pi2*m11*pow(m33,2)*RSe6611*RS1463e - 1./3888.
         *pi2*m11*pow(m33,2)*RSe6611*RC346e - 7./2592.*pi2*m11*pow(
         m33,2)*RS16e - 7./5184.*pi2*m11*pow(m33,2)*RS14e + 23./5184.*
         pi2*m11*pow(m33,2)*RC146e + 11./4860.*pi2*m11*pow(m33,2)*
         RC146e*RS3461e + 1./648.*pi2*m11*pow(m33,2)*RC146e*RS1463e - 1.
         /648.*pi2*m11*pow(m33,2)*RC146e*RSe6613 + 1./3888.*pi2*m11*
         pow(m33,2)*RC146e*RSe6633 - 1./486.*pi2*m11*pow(m33,2)*RC146e*
         RC346e + 5./3888.*pi2*m11*pow(m33,2)*RC146e*RSe6611 - 5./7776.
         *pi2*m11*pow(m33,2)*pow(RC146e,2) - 5./4374.*pi2*m11*pow(
         m33,2)*pow(m66,6)*pow(RN1e,3)*pow(RN3e,3) + 1./4374.*pi2*m11*
         pow(m33,2)*m44*pow(m66,5)*pow(RN1e,3)*pow(RN3e,3) - 121./8748.
         *pi2*m11*pow(m33,2)*pow(m44,2)*pow(m66,4)*pow(RN1e,3)*pow(
         RN3e,3) - 109./4374.*pi2*m11*pow(m33,2)*pow(m44,3)*pow(m66,3)*
         pow(RN1e,3)*pow(RN3e,3) - 1069./69984.*pi2*m11*pow(m33,2)*pow(
         m44,4)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + pi162 * (  - 265./69984.*pi2*m11*pow(m33,2)*pow(
         m44,5)*m66*pow(RN1e,3)*pow(RN3e,3) - 7./23328.*pi2*m11*pow(
         m33,2)*pow(m44,6)*pow(RN1e,3)*pow(RN3e,3) - 143./26244.*pi2*
         m11*pow(m33,3)*pow(m66,5)*pow(RN1e,3)*pow(RN3e,3) - 815./26244.
         *pi2*m11*pow(m33,3)*m44*pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) - 
         2165./104976.*pi2*m11*pow(m33,3)*pow(m44,2)*pow(m66,3)*pow(
         RN1e,3)*pow(RN3e,3) + 20./6561.*pi2*m11*pow(m33,3)*pow(m44,3)*
         pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) + 985./209952.*pi2*m11*pow(
         m33,3)*pow(m44,4)*m66*pow(RN1e,3)*pow(RN3e,3) + 163./209952.*
         pi2*m11*pow(m33,3)*pow(m44,5)*pow(RN1e,3)*pow(RN3e,3) + 397./
         17496.*pi2*m11*pow(m33,4)*pow(m66,4)*pow(RN1e,3)*pow(RN3e,3)
          + 2275./34992.*pi2*m11*pow(m33,4)*m44*pow(m66,3)*pow(RN1e,3)*
         pow(RN3e,3) + 275./5832.*pi2*m11*pow(m33,4)*pow(m44,2)*pow(
         m66,2)*pow(RN1e,3)*pow(RN3e,3) + 803./69984.*pi2*m11*pow(
         m33,4)*pow(m44,3)*m66*pow(RN1e,3)*pow(RN3e,3) + 23./34992.*pi2
         *m11*pow(m33,4)*pow(m44,4)*pow(RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + pi162 * (  - 20./729.*pi2*m11*pow(m33,5)*pow(
         m66,3)*pow(RN1e,3)*pow(RN3e,3) - 185./3888.*pi2*m11*pow(m33,5)
         *m44*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) - 65./2592.*pi2*m11*
         pow(m33,5)*pow(m44,2)*m66*pow(RN1e,3)*pow(RN3e,3) - 95./23328.
         *pi2*m11*pow(m33,5)*pow(m44,3)*pow(RN1e,3)*pow(RN3e,3) + 7./
         648.*pi2*m11*pow(m33,6)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) + 7.
         /648.*pi2*m11*pow(m33,6)*m44*m66*pow(RN1e,3)*pow(RN3e,3) + 7./
         2592.*pi2*m11*pow(m33,6)*pow(m44,2)*pow(RN1e,3)*pow(RN3e,3) + 
         1./288.*pi2*m11*m16*RD346e - 1./270.*pi2*m11*m16*RD146e - 1./
         6480.*pi2*m11*m16*me2*RSe6613 + 23./1296.*pi2*m11*m16*me2*
         RSe63 - 1./144.*pi2*m11*m16*me2*RSe6633 + 1./720.*pi2*m11*m16*
         me2*RSe61 - 1./135.*pi2*m11*m16*me2*RSe6611 - 11./1728.*pi2*
         m11*m16*m66 - 1./576.*pi2*m11*m16*m66*RS3461e - 13./5184.*pi2*
         m11*m16*m66*RS1463e + 1./288.*pi2*m11*m16*m66*RSe6613 + 1./96.
         *pi2*m11*m16*m66*RSe63 + 1./1296.*pi2*m11*m16*m66*RSe6633 - 1./
         192.*pi2*m11*m16*m66*RS34e );

      M13p6x22R +=  + pi162 * (  - 1./2592.*pi2*m11*m16*m66*RC346e + 1./
         864.*pi2*m11*m16*m66*RSe61 - 1./1296.*pi2*m11*m16*m66*RSe6611
          + 17./25920.*pi2*m11*m16*m66*RS14e - 31./2592.*pi2*m11*m16*
         m44 - 59./5184.*pi2*m11*m16*m33 - 43./25920.*pi2*m11*m16*m33*
         RS3461e - 23./25920.*pi2*m11*m16*m33*RS1463e + 1./288.*pi2*m11
         *m16*m33*RSe6613 + 1./96.*pi2*m11*m16*m33*RSe63 + 1./1296.*pi2
         *m11*m16*m33*RSe6633 - 73./5184.*pi2*m11*m16*m33*RS34e + 1./
         324.*pi2*m11*m16*m33*RC346e + 1./864.*pi2*m11*m16*m33*RSe61 - 
         1./1296.*pi2*m11*m16*m33*RSe6611 - 13./5184.*pi2*m11*m16*m33*
         RS14e + 1./144.*pi2*m11*m14*RD346e - 1./135.*pi2*m11*m14*
         RD146e - 1./3240.*pi2*m11*m14*me2*RSe6613 - 23./1296.*pi2*m11*
         m14*me2*RSe63 - 1./72.*pi2*m11*m14*me2*RSe6633 - 1./720.*pi2*
         m11*m14*me2*RSe61 - 2./135.*pi2*m11*m14*me2*RSe6611 - 31./2592.
         *pi2*m11*m14*m66 - 2./81.*pi2*m11*m14*m44 - 1./288.*pi2*m11*
         m14*m44*RS3461e - 13./2592.*pi2*m11*m14*m44*RS1463e + 1./144.*
         pi2*m11*m14*m44*RSe6613 );

      M13p6x22R +=  + pi162 * (  - 1./96.*pi2*m11*m14*m44*RSe63 + 1./
         648.*pi2*m11*m14*m44*RSe6633 - 1./96.*pi2*m11*m14*m44*RS36e - 
         1./1296.*pi2*m11*m14*m44*RC346e - 1./864.*pi2*m11*m14*m44*
         RSe61 - 1./648.*pi2*m11*m14*m44*RSe6611 + 17./12960.*pi2*m11*
         m14*m44*RS16e - 59./2592.*pi2*m11*m14*m33 - 43./12960.*pi2*m11
         *m14*m33*RS3461e - 23./12960.*pi2*m11*m14*m33*RS1463e + 1./144.
         *pi2*m11*m14*m33*RSe6613 - 1./96.*pi2*m11*m14*m33*RSe63 + 1./
         648.*pi2*m11*m14*m33*RSe6633 - 73./2592.*pi2*m11*m14*m33*RS36e
          + 1./162.*pi2*m11*m14*m33*RC346e - 1./864.*pi2*m11*m14*m33*
         RSe61 - 1./648.*pi2*m11*m14*m33*RSe6611 - 13./2592.*pi2*m11*
         m14*m33*RS16e + 1./648.*pi2*m11*m13*RD346e*RS3461e + 1./1944.*
         pi2*m11*m13*RD346e*RS1463e - 1./324.*pi2*m11*m13*RD346e*
         RSe6613 + 5./1944.*pi2*m11*m13*RD346e*RSe6633 - 5./3888.*pi2*
         m11*m13*RD346e*RC346e + 1./1944.*pi2*m11*m13*RSe6611*RD346e + 
         1./3888.*pi2*m11*m13*RC146e*RD346e - 1./1944.*pi2*m11*m13*
         RD146e*RS3461e );

      M13p6x22R +=  + pi162 * (  - 1./648.*pi2*m11*m13*RD146e*RS1463e
          + 1./972.*pi2*m11*m13*RD146e*RSe6613 - 5./1944.*pi2*m11*m13*
         RD146e*RSe6633 + 5./3888.*pi2*m11*m13*RD146e*RC346e + 1./648.*
         pi2*m11*m13*RD146e*RSe6611 - 1./3888.*pi2*m11*m13*RD146e*
         RC146e + 2./243.*pi2*m11*m13*me2*RSe6613*RS3461e + 1./243.*pi2
         *m11*m13*me2*RSe6613*RS1463e - 4./243.*pi2*m11*m13*me2*pow(
         RSe6613,2) - 1./324.*pi2*m11*m13*me2*RSe6633*RS3461e - 1./972.
         *pi2*m11*m13*me2*RSe6633*RS1463e + 1./81.*pi2*m11*m13*me2*
         RSe6633*RSe6613 - 5./972.*pi2*m11*m13*me2*pow(RSe6633,2) - 1./
         324.*pi2*m11*m13*me2*RC346e*RSe6613 + 5./1944.*pi2*m11*m13*me2
         *RC346e*RSe6633 - 5./972.*pi2*m11*m13*me2*RSe6611*RS3461e - 1./
         324.*pi2*m11*m13*me2*RSe6611*RS1463e + 5./243.*pi2*m11*m13*me2
         *RSe6611*RSe6613 - 1./486.*pi2*m11*m13*me2*RSe6611*RSe6633 + 1.
         /1944.*pi2*m11*m13*me2*RSe6611*RC346e - 1./108.*pi2*m11*m13*
         me2*pow(RSe6611,2) - 1./324.*pi2*m11*m13*me2*RC146e*RSe6613 - 
         1./1944.*pi2*m11*m13*me2*RC146e*RSe6633 );

      M13p6x22R +=  + pi162 * ( 7./1944.*pi2*m11*m13*me2*RC146e*RSe6611
          - 31./432.*pi2*m11*m13*m33 - 1./243.*pi2*m11*m13*m33*pow(
         RS3461e,2) - 1./162.*pi2*m11*m13*m33*RS1463e*RS3461e - 1./243.
         *pi2*m11*m13*m33*pow(RS1463e,2) + 2./243.*pi2*m11*m13*m33*
         RSe6613*RS3461e + 2./243.*pi2*m11*m13*m33*RSe6613*RS1463e - 1./
         324.*pi2*m11*m13*m33*RSe6633*RS3461e - 5./972.*pi2*m11*m13*m33
         *RSe6633*RS1463e + 1./324.*pi2*m11*m13*m33*RC346e*RS3461e + 1./
         324.*pi2*m11*m13*m33*RC346e*RS1463e - 1./324.*pi2*m11*m13*m33*
         RC346e*RSe6613 + 5./1944.*pi2*m11*m13*m33*RC346e*RSe6633 - 5./
         3888.*pi2*m11*m13*m33*pow(RC346e,2) - 5./972.*pi2*m11*m13*m33*
         RSe6611*RS3461e - 1./324.*pi2*m11*m13*m33*RSe6611*RS1463e + 1./
         1944.*pi2*m11*m13*m33*RSe6611*RC346e + 1./324.*pi2*m11*m13*m33
         *RC146e*RS3461e + 1./324.*pi2*m11*m13*m33*RC146e*RS1463e - 1./
         324.*pi2*m11*m13*m33*RC146e*RSe6613 + 1./1944.*pi2*m11*m13*m33
         *RC146e*RSe6633 + 5./1944.*pi2*m11*m13*m33*RC146e*RSe6611 - 5./
         3888.*pi2*m11*m13*m33*pow(RC146e,2) );

      M13p6x22R +=  + pi162 * (  - 1./2430.*pi2*pow(m11,2)*RD346e*
         RS1463e - 1./972.*pi2*pow(m11,2)*RC146e*RD346e + 1./270.*pi2*
         pow(m11,2)*RD146e - 1./1296.*pi2*pow(m11,2)*RD146e*RS3461e + 1.
         /3888.*pi2*pow(m11,2)*RD146e*RS1463e - 1./216.*pi2*pow(m11,2)*
         RD146e*RSe6613 - 1./1296.*pi2*pow(m11,2)*RD146e*RSe6633 + 1./
         2592.*pi2*pow(m11,2)*RD146e*RC346e + 7./1296.*pi2*pow(m11,2)*
         RD146e*RSe6611 - 53./38880.*pi2*pow(m11,2)*RD146e*RC146e - 1./
         4320.*pi2*pow(m11,2)*me2*RSe6613 + 1./243.*pi2*pow(m11,2)*me2*
         RSe6613*RS3461e - 5./486.*pi2*pow(m11,2)*me2*RSe6613*RS1463e
          - 2./243.*pi2*pow(m11,2)*me2*pow(RSe6613,2) - 23./2592.*pi2*
         pow(m11,2)*me2*RSe6633 - 1./648.*pi2*pow(m11,2)*me2*RSe6633*
         RS3461e - 19./9720.*pi2*pow(m11,2)*me2*RSe6633*RS1463e + 1./
         162.*pi2*pow(m11,2)*me2*RSe6633*RSe6613 - 5./1944.*pi2*pow(
         m11,2)*me2*pow(RSe6633,2) - 1./648.*pi2*pow(m11,2)*me2*RC346e*
         RSe6613 + 5./3888.*pi2*pow(m11,2)*me2*RC346e*RSe6633 - 37./
         6480.*pi2*pow(m11,2)*me2*RSe6611 );

      M13p6x22R +=  + pi162 * (  - 29./1944.*pi2*pow(m11,2)*me2*RSe6611
         *RS3461e - 211./9720.*pi2*pow(m11,2)*me2*RSe6611*RS1463e + 23./
         486.*pi2*pow(m11,2)*me2*RSe6611*RSe6613 - 1./972.*pi2*pow(
         m11,2)*me2*RSe6611*RSe6633 + 1./3888.*pi2*pow(m11,2)*me2*
         RSe6611*RC346e - 11./648.*pi2*pow(m11,2)*me2*pow(RSe6611,2) - 
         31./9720.*pi2*pow(m11,2)*me2*RC146e*RSe6613 + 17./3888.*pi2*
         pow(m11,2)*me2*RC146e*RSe6633 + 67./19440.*pi2*pow(m11,2)*me2*
         RC146e*RSe6611 - 1./324.*pi2*pow(m11,2)*m66*RS3461e - 61./
         12960.*pi2*pow(m11,2)*m66*RS1463e + 1./243.*pi2*pow(m11,2)*m66
         *RSe6613 + 1./81.*pi2*pow(m11,2)*m66*RSe63 - 1./648.*pi2*pow(
         m11,2)*m66*RS34e - 1./162.*pi2*pow(m11,2)*m66*RSe61 + 2./243.*
         pi2*pow(m11,2)*m66*RSe6611 + 1./486.*pi2*pow(m11,2)*m66*RS16e*
         RS36e + 1./972.*pi2*pow(m11,2)*m66*RS16e*RS34e - 1./972.*pi2*
         pow(m11,2)*m66*pow(RS16e,2) - 17./6480.*pi2*pow(m11,2)*m66*
         RS14e - 1./54.*pi2*pow(m11,2)*m66*RS14e*RSe63 + 1./972.*pi2*
         pow(m11,2)*m66*RS14e*RS36e );

      M13p6x22R +=  + pi162 * ( 5./972.*pi2*pow(m11,2)*m66*RS14e*RS34e
          - 1./972.*pi2*pow(m11,2)*m66*RS14e*RS16e - 5./1944.*pi2*pow(
         m11,2)*m66*pow(RS14e,2) + 23./15552.*pi2*pow(m11,2)*m66*RC146e
          + 1./4374.*pi2*pow(m11,2)*pow(m66,7)*pow(RN1e,3)*pow(RN3e,3)
          - 1./81.*pi2*pow(m11,2)*m46*RS16e*RSe63 - 1./243.*pi2*pow(
         m11,2)*m46*RS16e*RS36e + 1./243.*pi2*pow(m11,2)*m46*RS16e*
         RS34e + 1./972.*pi2*pow(m11,2)*m46*pow(RS16e,2) + 1./81.*pi2*
         pow(m11,2)*m46*RS14e*RSe63 + 1./243.*pi2*pow(m11,2)*m46*RS14e*
         RS36e - 1./243.*pi2*pow(m11,2)*m46*RS14e*RS34e - 1./486.*pi2*
         pow(m11,2)*m46*RS14e*RS16e + 1./972.*pi2*pow(m11,2)*m46*pow(
         RS14e,2) - 1./162.*pi2*pow(m11,2)*m44*RS3461e - 61./6480.*pi2*
         pow(m11,2)*m44*RS1463e + 2./243.*pi2*pow(m11,2)*m44*RSe6613 - 
         1./81.*pi2*pow(m11,2)*m44*RSe63 - 1./324.*pi2*pow(m11,2)*m44*
         RS36e + 1./162.*pi2*pow(m11,2)*m44*RSe61 + 4./243.*pi2*pow(
         m11,2)*m44*RSe6611 - 17./3240.*pi2*pow(m11,2)*m44*RS16e + 2./
         81.*pi2*pow(m11,2)*m44*RS16e*RSe63 );

      M13p6x22R +=  + pi162 * ( 7./486.*pi2*pow(m11,2)*m44*RS16e*RS36e
          + 1./972.*pi2*pow(m11,2)*m44*RS16e*RS34e - 7./972.*pi2*pow(
         m11,2)*m44*pow(RS16e,2) - 1./162.*pi2*pow(m11,2)*m44*RS14e*
         RSe63 + 1./972.*pi2*pow(m11,2)*m44*RS14e*RS36e + 1./486.*pi2*
         pow(m11,2)*m44*RS14e*RS34e - 1./972.*pi2*pow(m11,2)*m44*RS14e*
         RS16e - 1./972.*pi2*pow(m11,2)*m44*pow(RS14e,2) + 23./7776.*
         pi2*pow(m11,2)*m44*RC146e + 10./6561.*pi2*pow(m11,2)*m44*pow(
         m66,6)*pow(RN1e,3)*pow(RN3e,3) + 157./26244.*pi2*pow(m11,2)*
         pow(m44,2)*pow(m66,5)*pow(RN1e,3)*pow(RN3e,3) + 235./26244.*
         pi2*pow(m11,2)*pow(m44,3)*pow(m66,4)*pow(RN1e,3)*pow(RN3e,3)
          + 1315./209952.*pi2*pow(m11,2)*pow(m44,4)*pow(m66,3)*pow(
         RN1e,3)*pow(RN3e,3) + 227./104976.*pi2*pow(m11,2)*pow(m44,5)*
         pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) + 35./104976.*pi2*pow(
         m11,2)*pow(m44,6)*m66*pow(RN1e,3)*pow(RN3e,3) + 1./69984.*pi2*
         pow(m11,2)*pow(m44,7)*pow(RN1e,3)*pow(RN3e,3) - 31./2592.*pi2*
         pow(m11,2)*m36 );

      M13p6x22R +=  + pi162 * (  - 1./576.*pi2*pow(m11,2)*m36*RS3461e
          + 113./25920.*pi2*pow(m11,2)*m36*RS1463e + 1./288.*pi2*pow(
         m11,2)*m36*RSe6613 - 5./2592.*pi2*pow(m11,2)*m36*RSe63 + 1./
         1296.*pi2*pow(m11,2)*m36*RSe6633 + 5./5184.*pi2*pow(m11,2)*m36
         *RS34e - 1./2592.*pi2*pow(m11,2)*m36*RC346e + 1./864.*pi2*pow(
         m11,2)*m36*RSe61 - 1./1296.*pi2*pow(m11,2)*m36*RSe6611 - 221./
         25920.*pi2*pow(m11,2)*m36*RS14e + 1./288.*pi2*pow(m11,2)*m36*
         RC146e - 31./1296.*pi2*pow(m11,2)*m34 - 1./288.*pi2*pow(m11,2)
         *m34*RS3461e + 113./12960.*pi2*pow(m11,2)*m34*RS1463e + 1./144.
         *pi2*pow(m11,2)*m34*RSe6613 + 5./2592.*pi2*pow(m11,2)*m34*
         RSe63 + 1./648.*pi2*pow(m11,2)*m34*RSe6633 + 5./2592.*pi2*pow(
         m11,2)*m34*RS36e - 1./1296.*pi2*pow(m11,2)*m34*RC346e - 1./864.
         *pi2*pow(m11,2)*m34*RSe61 - 1./648.*pi2*pow(m11,2)*m34*RSe6611
          - 221./12960.*pi2*pow(m11,2)*m34*RS16e + 1./144.*pi2*pow(
         m11,2)*m34*RC146e + 35./1728.*pi2*pow(m11,2)*m33 - 157./25920.
         *pi2*pow(m11,2)*m33*RS3461e );

      M13p6x22R +=  + pi162 * (  - 1./486.*pi2*pow(m11,2)*m33*pow(
         RS3461e,2) + 13./1296.*pi2*pow(m11,2)*m33*RS1463e + 5./1944.*
         pi2*pow(m11,2)*m33*RS1463e*RS3461e + 1./1080.*pi2*pow(m11,2)*
         m33*pow(RS1463e,2) + 1./54.*pi2*pow(m11,2)*m33*RSe6613 + 1./
         243.*pi2*pow(m11,2)*m33*RSe6613*RS3461e + 5./972.*pi2*pow(
         m11,2)*m33*RSe6613*RS1463e - 1./648.*pi2*pow(m11,2)*m33*
         RSe6633*RS3461e - 5./1944.*pi2*pow(m11,2)*m33*RSe6633*RS1463e
          - 5./864.*pi2*pow(m11,2)*m33*RS36e - 5./1728.*pi2*pow(m11,2)*
         m33*RS34e + 23./5184.*pi2*pow(m11,2)*m33*RC346e + 1./648.*pi2*
         pow(m11,2)*m33*RC346e*RS3461e + 11./4860.*pi2*pow(m11,2)*m33*
         RC346e*RS1463e - 1./648.*pi2*pow(m11,2)*m33*RC346e*RSe6613 + 5.
         /3888.*pi2*pow(m11,2)*m33*RC346e*RSe6633 - 5./7776.*pi2*pow(
         m11,2)*m33*pow(RC346e,2) - 5./1944.*pi2*pow(m11,2)*m33*RSe6611
         *RS3461e - 5./1944.*pi2*pow(m11,2)*m33*RSe6611*RS1463e + 1./
         3888.*pi2*pow(m11,2)*m33*RSe6611*RC346e - 4./405.*pi2*pow(
         m11,2)*m33*RS16e );

      M13p6x22R +=  + pi162 * (  - 2./405.*pi2*pow(m11,2)*m33*RS14e + 
         25./5184.*pi2*pow(m11,2)*m33*RC146e + 11./1215.*pi2*pow(m11,2)
         *m33*RC146e*RS3461e + 29./4860.*pi2*pow(m11,2)*m33*RC146e*
         RS1463e - 17./1944.*pi2*pow(m11,2)*m33*RC146e*RSe6613 - 1./
         3888.*pi2*pow(m11,2)*m33*RC146e*RSe6633 - 1./486.*pi2*pow(
         m11,2)*m33*RC146e*RC346e + 11./3888.*pi2*pow(m11,2)*m33*RC146e
         *RSe6611 - 1./7776.*pi2*pow(m11,2)*m33*pow(RC146e,2) - 5./4374.
         *pi2*pow(m11,2)*m33*pow(m66,6)*pow(RN1e,3)*pow(RN3e,3) + 1./
         4374.*pi2*pow(m11,2)*m33*m44*pow(m66,5)*pow(RN1e,3)*pow(
         RN3e,3) - 121./8748.*pi2*pow(m11,2)*m33*pow(m44,2)*pow(m66,4)*
         pow(RN1e,3)*pow(RN3e,3) - 109./4374.*pi2*pow(m11,2)*m33*pow(
         m44,3)*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) - 1069./69984.*pi2*
         pow(m11,2)*m33*pow(m44,4)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3)
          - 265./69984.*pi2*pow(m11,2)*m33*pow(m44,5)*m66*pow(RN1e,3)*
         pow(RN3e,3) - 7./23328.*pi2*pow(m11,2)*m33*pow(m44,6)*pow(
         RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + pi162 * (  - 125./26244.*pi2*pow(m11,2)*pow(
         m33,2)*pow(m66,5)*pow(RN1e,3)*pow(RN3e,3) - 1031./26244.*pi2*
         pow(m11,2)*pow(m33,2)*m44*pow(m66,4)*pow(RN1e,3)*pow(RN3e,3)
          - 1913./104976.*pi2*pow(m11,2)*pow(m33,2)*pow(m44,2)*pow(
         m66,3)*pow(RN1e,3)*pow(RN3e,3) + 799./52488.*pi2*pow(m11,2)*
         pow(m33,2)*pow(m44,3)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) + 
         1127./104976.*pi2*pow(m11,2)*pow(m33,2)*pow(m44,4)*m66*pow(
         RN1e,3)*pow(RN3e,3) + 167./104976.*pi2*pow(m11,2)*pow(m33,2)*
         pow(m44,5)*pow(RN1e,3)*pow(RN3e,3) + 73./2187.*pi2*pow(m11,2)*
         pow(m33,3)*pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) + 1031./8748.*
         pi2*pow(m11,2)*pow(m33,3)*m44*pow(m66,3)*pow(RN1e,3)*pow(
         RN3e,3) + 209./2916.*pi2*pow(m11,2)*pow(m33,3)*pow(m44,2)*pow(
         m66,2)*pow(RN1e,3)*pow(RN3e,3) + 121./17496.*pi2*pow(m11,2)*
         pow(m33,3)*pow(m44,3)*m66*pow(RN1e,3)*pow(RN3e,3) - 127./69984.
         *pi2*pow(m11,2)*pow(m33,3)*pow(m44,4)*pow(RN1e,3)*pow(RN3e,3)
          - 1529./23328.*pi2*pow(m11,2)*pow(m33,4)*pow(m66,3)*pow(
         RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + pi162 * (  - 523./3888.*pi2*pow(m11,2)*pow(m33,4)
         *m44*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) - 79./1296.*pi2*pow(
         m11,2)*pow(m33,4)*pow(m44,2)*m66*pow(RN1e,3)*pow(RN3e,3) - 37./
         5832.*pi2*pow(m11,2)*pow(m33,4)*pow(m44,3)*pow(RN1e,3)*pow(
         RN3e,3) + 421./7776.*pi2*pow(m11,2)*pow(m33,5)*pow(m66,2)*pow(
         RN1e,3)*pow(RN3e,3) + 481./7776.*pi2*pow(m11,2)*pow(m33,5)*m44
         *m66*pow(RN1e,3)*pow(RN3e,3) + 31./1944.*pi2*pow(m11,2)*pow(
         m33,5)*pow(m44,2)*pow(RN1e,3)*pow(RN3e,3) - 7./432.*pi2*pow(
         m11,2)*pow(m33,6)*m66*pow(RN1e,3)*pow(RN3e,3) - 7./864.*pi2*
         pow(m11,2)*pow(m33,6)*m44*pow(RN1e,3)*pow(RN3e,3) + 1./1728.*
         pi2*pow(m11,2)*m16 + 23./4320.*pi2*pow(m11,2)*m16*RS1463e + 2./
         81.*pi2*pow(m11,2)*m16*RSe63 - 1./81.*pi2*pow(m11,2)*m16*RS34e
          + 19./12960.*pi2*pow(m11,2)*m16*RS14e + 1./864.*pi2*pow(
         m11,2)*m14 + 23./2160.*pi2*pow(m11,2)*m14*RS1463e - 2./81.*pi2
         *pow(m11,2)*m14*RSe63 - 2./81.*pi2*pow(m11,2)*m14*RS36e + 19./
         6480.*pi2*pow(m11,2)*m14*RS16e );

      M13p6x22R +=  + pi162 * (  - 31./864.*pi2*pow(m11,2)*m13 - 1./972.
         *pi2*pow(m11,2)*m13*RS1463e*RS3461e + 1./486.*pi2*pow(m11,2)*
         m13*RSe6613*RS1463e - 1./486.*pi2*pow(m11,2)*m13*RSe6611*
         RS1463e + 1./972.*pi2*pow(m11,2)*m13*RC146e*RS3461e + 1./972.*
         pi2*pow(m11,2)*m13*RC146e*RS1463e - 1./486.*pi2*pow(m11,2)*m13
         *RC146e*RSe6613 - 1./972.*pi2*pow(m11,2)*m13*RC146e*RSe6633 + 
         1./1944.*pi2*pow(m11,2)*m13*RC146e*RC346e + 1./324.*pi2*pow(
         m11,2)*m13*RC146e*RSe6611 - 5./3888.*pi2*pow(m11,2)*m13*pow(
         RC146e,2) + 7./576.*pi2*pow(m11,3) - 1./108.*pi2*pow(m11,3)*
         RS3461e - 223./25920.*pi2*pow(m11,3)*RS1463e - 1./1944.*pi2*
         pow(m11,3)*RS1463e*RS3461e + 31./9720.*pi2*pow(m11,3)*pow(
         RS1463e,2) + 5./162.*pi2*pow(m11,3)*RSe6613 + 1./972.*pi2*pow(
         m11,3)*RSe6613*RS1463e - 1./324.*pi2*pow(m11,3)*RS36e - 1./648.
         *pi2*pow(m11,3)*RS34e + 1./162.*pi2*pow(m11,3)*RSe6611 - 1./
         972.*pi2*pow(m11,3)*RSe6611*RS1463e + 19./12960.*pi2*pow(
         m11,3)*RS16e );

      M13p6x22R +=  + pi162 * ( 19./25920.*pi2*pow(m11,3)*RS14e + 1./
         288.*pi2*pow(m11,3)*RC146e + 13./1944.*pi2*pow(m11,3)*RC146e*
         RS3461e + 17./1620.*pi2*pow(m11,3)*RC146e*RS1463e - 19./972.*
         pi2*pow(m11,3)*RC146e*RSe6613 - 1./1944.*pi2*pow(m11,3)*RC146e
         *RSe6633 + 1./3888.*pi2*pow(m11,3)*RC146e*RC346e + 5./648.*pi2
         *pow(m11,3)*RC146e*RSe6611 - 5./1944.*pi2*pow(m11,3)*pow(
         RC146e,2) + 1./6561.*pi2*pow(m11,3)*pow(m66,6)*pow(RN1e,3)*
         pow(RN3e,3) + 14./6561.*pi2*pow(m11,3)*m44*pow(m66,5)*pow(
         RN1e,3)*pow(RN3e,3) - 7./26244.*pi2*pow(m11,3)*pow(m44,2)*pow(
         m66,4)*pow(RN1e,3)*pow(RN3e,3) - 23./5832.*pi2*pow(m11,3)*pow(
         m44,3)*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) - 157./52488.*pi2*
         pow(m11,3)*pow(m44,4)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) - 169.
         /209952.*pi2*pow(m11,3)*pow(m44,5)*m66*pow(RN1e,3)*pow(RN3e,3)
          - 7./104976.*pi2*pow(m11,3)*pow(m44,6)*pow(RN1e,3)*pow(
         RN3e,3) - 143./26244.*pi2*pow(m11,3)*m33*pow(m66,5)*pow(
         RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + pi162 * (  - 815./26244.*pi2*pow(m11,3)*m33*m44*
         pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) - 2165./104976.*pi2*pow(
         m11,3)*m33*pow(m44,2)*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) + 20./
         6561.*pi2*pow(m11,3)*m33*pow(m44,3)*pow(m66,2)*pow(RN1e,3)*
         pow(RN3e,3) + 985./209952.*pi2*pow(m11,3)*m33*pow(m44,4)*m66*
         pow(RN1e,3)*pow(RN3e,3) + 163./209952.*pi2*pow(m11,3)*m33*pow(
         m44,5)*pow(RN1e,3)*pow(RN3e,3) + 73./2187.*pi2*pow(m11,3)*pow(
         m33,2)*pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) + 1031./8748.*pi2*
         pow(m11,3)*pow(m33,2)*m44*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3)
          + 209./2916.*pi2*pow(m11,3)*pow(m33,2)*pow(m44,2)*pow(m66,2)*
         pow(RN1e,3)*pow(RN3e,3) + 121./17496.*pi2*pow(m11,3)*pow(
         m33,2)*pow(m44,3)*m66*pow(RN1e,3)*pow(RN3e,3) - 127./69984.*
         pi2*pow(m11,3)*pow(m33,2)*pow(m44,4)*pow(RN1e,3)*pow(RN3e,3)
          - 35./432.*pi2*pow(m11,3)*pow(m33,3)*pow(m66,3)*pow(RN1e,3)*
         pow(RN3e,3) - 44./243.*pi2*pow(m11,3)*pow(m33,3)*m44*pow(
         m66,2)*pow(RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + pi162 * (  - 293./3888.*pi2*pow(m11,3)*pow(m33,3)
         *pow(m44,2)*m66*pow(RN1e,3)*pow(RN3e,3) - 5./972.*pi2*pow(
         m11,3)*pow(m33,3)*pow(m44,3)*pow(RN1e,3)*pow(RN3e,3) + 59./648.
         *pi2*pow(m11,3)*pow(m33,4)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3)
          + 305./2592.*pi2*pow(m11,3)*pow(m33,4)*m44*m66*pow(RN1e,3)*
         pow(RN3e,3) + 31./1296.*pi2*pow(m11,3)*pow(m33,4)*pow(m44,2)*
         pow(RN1e,3)*pow(RN3e,3) - 119./2592.*pi2*pow(m11,3)*pow(m33,5)
         *m66*pow(RN1e,3)*pow(RN3e,3) - 67./2592.*pi2*pow(m11,3)*pow(
         m33,5)*m44*pow(RN1e,3)*pow(RN3e,3) + 7./864.*pi2*pow(m11,3)*
         pow(m33,6)*pow(RN1e,3)*pow(RN3e,3) - 161./52488.*pi2*pow(
         m11,4)*pow(m66,5)*pow(RN1e,3)*pow(RN3e,3) - 599./52488.*pi2*
         pow(m11,4)*m44*pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) - 2417./
         209952.*pi2*pow(m11,4)*pow(m44,2)*pow(m66,3)*pow(RN1e,3)*pow(
         RN3e,3) - 479./104976.*pi2*pow(m11,4)*pow(m44,3)*pow(m66,2)*
         pow(RN1e,3)*pow(RN3e,3) - 71./104976.*pi2*pow(m11,4)*pow(
         m44,4)*m66*pow(RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + pi162 * (  - 1./52488.*pi2*pow(m11,4)*pow(m44,5)*
         pow(RN1e,3)*pow(RN3e,3) + 397./17496.*pi2*pow(m11,4)*m33*pow(
         m66,4)*pow(RN1e,3)*pow(RN3e,3) + 2275./34992.*pi2*pow(m11,4)*
         m33*m44*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) + 275./5832.*pi2*
         pow(m11,4)*m33*pow(m44,2)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3)
          + 803./69984.*pi2*pow(m11,4)*m33*pow(m44,3)*m66*pow(RN1e,3)*
         pow(RN3e,3) + 23./34992.*pi2*pow(m11,4)*m33*pow(m44,4)*pow(
         RN1e,3)*pow(RN3e,3) - 1529./23328.*pi2*pow(m11,4)*pow(m33,2)*
         pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) - 523./3888.*pi2*pow(m11,4)
         *pow(m33,2)*m44*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) - 79./1296.
         *pi2*pow(m11,4)*pow(m33,2)*pow(m44,2)*m66*pow(RN1e,3)*pow(
         RN3e,3) - 37./5832.*pi2*pow(m11,4)*pow(m33,2)*pow(m44,3)*pow(
         RN1e,3)*pow(RN3e,3) + 59./648.*pi2*pow(m11,4)*pow(m33,3)*pow(
         m66,2)*pow(RN1e,3)*pow(RN3e,3) + 305./2592.*pi2*pow(m11,4)*
         pow(m33,3)*m44*m66*pow(RN1e,3)*pow(RN3e,3) + 31./1296.*pi2*
         pow(m11,4)*pow(m33,3)*pow(m44,2)*pow(RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + pi162 * (  - 77./1296.*pi2*pow(m11,4)*pow(m33,4)*
         m66*pow(RN1e,3)*pow(RN3e,3) - 23./648.*pi2*pow(m11,4)*pow(
         m33,4)*m44*pow(RN1e,3)*pow(RN3e,3) + 1./72.*pi2*pow(m11,4)*
         pow(m33,5)*pow(RN1e,3)*pow(RN3e,3) + 89./17496.*pi2*pow(m11,5)
         *pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) + 26./2187.*pi2*pow(m11,5)
         *m44*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) + 223./23328.*pi2*pow(
         m11,5)*pow(m44,2)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) + 223./
         69984.*pi2*pow(m11,5)*pow(m44,3)*m66*pow(RN1e,3)*pow(RN3e,3)
          + 13./34992.*pi2*pow(m11,5)*pow(m44,4)*pow(RN1e,3)*pow(
         RN3e,3) - 20./729.*pi2*pow(m11,5)*m33*pow(m66,3)*pow(RN1e,3)*
         pow(RN3e,3) - 185./3888.*pi2*pow(m11,5)*m33*m44*pow(m66,2)*
         pow(RN1e,3)*pow(RN3e,3) - 65./2592.*pi2*pow(m11,5)*m33*pow(
         m44,2)*m66*pow(RN1e,3)*pow(RN3e,3) - 95./23328.*pi2*pow(m11,5)
         *m33*pow(m44,3)*pow(RN1e,3)*pow(RN3e,3) + 421./7776.*pi2*pow(
         m11,5)*pow(m33,2)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) + 481./
         7776.*pi2*pow(m11,5)*pow(m33,2)*m44*m66*pow(RN1e,3)*pow(
         RN3e,3) );

      M13p6x22R +=  + pi162 * ( 31./1944.*pi2*pow(m11,5)*pow(m33,2)*
         pow(m44,2)*pow(RN1e,3)*pow(RN3e,3) - 119./2592.*pi2*pow(m11,5)
         *pow(m33,3)*m66*pow(RN1e,3)*pow(RN3e,3) - 67./2592.*pi2*pow(
         m11,5)*pow(m33,3)*m44*pow(RN1e,3)*pow(RN3e,3) + 1./72.*pi2*
         pow(m11,5)*pow(m33,4)*pow(RN1e,3)*pow(RN3e,3) - 7./2916.*pi2*
         pow(m11,6)*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) - 7./1944.*pi2*
         pow(m11,6)*m44*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) - 7./3888.*
         pi2*pow(m11,6)*pow(m44,2)*m66*pow(RN1e,3)*pow(RN3e,3) - 7./
         23328.*pi2*pow(m11,6)*pow(m44,3)*pow(RN1e,3)*pow(RN3e,3) + 7./
         648.*pi2*pow(m11,6)*m33*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) + 7.
         /648.*pi2*pow(m11,6)*m33*m44*m66*pow(RN1e,3)*pow(RN3e,3) + 7./
         2592.*pi2*pow(m11,6)*m33*pow(m44,2)*pow(RN1e,3)*pow(RN3e,3) - 
         7./432.*pi2*pow(m11,6)*pow(m33,2)*m66*pow(RN1e,3)*pow(RN3e,3)
          - 7./864.*pi2*pow(m11,6)*pow(m33,2)*m44*pow(RN1e,3)*pow(
         RN3e,3) + 7./864.*pi2*pow(m11,6)*pow(m33,3)*pow(RN1e,3)*pow(
         RN3e,3) );

      M13p6x22R +=  + Ab(m11,mu2)*pi16 * (  - 1./72.*m33*RD346e*RS1463e
          + 1./144.*m33*RC146e*RD346e + 1./48.*m33*RD146e - 1./72.*m33*
         RD146e*RS1463e + 1./144.*m33*RD146e*RC146e - 1./18.*m33*me2*
         RSe6613*RS1463e + 1./36.*m33*me2*RSe6633*RS1463e + 1./36.*m33*
         me2*RSe6611*RS1463e + 1./36.*m33*me2*RC146e*RSe6613 - 1./72.*
         m33*me2*RC146e*RSe6633 - 1./72.*m33*me2*RC146e*RSe6611 + 1./24.
         *m33*m66*RS1463e - 1./48.*m33*m66*RC146e + 1./12.*m33*m44*
         RS1463e - 1./24.*m33*m44*RC146e + 1./24.*pow(m33,2)*RS1463e + 
         1./27.*pow(m33,2)*RS1463e*RS3461e + 1./108.*pow(m33,2)*pow(
         RS1463e,2) - 1./54.*pow(m33,2)*RSe6613*RS1463e + 1./108.*pow(
         m33,2)*RSe6633*RS1463e - 1./54.*pow(m33,2)*RC346e*RS1463e + 1./
         108.*pow(m33,2)*RSe6611*RS1463e - 1./48.*pow(m33,2)*RC146e - 1.
         /54.*pow(m33,2)*RC146e*RS3461e - 1./108.*pow(m33,2)*RC146e*
         RS1463e + 1./108.*pow(m33,2)*RC146e*RSe6613 - 1./216.*pow(
         m33,2)*RC146e*RSe6633 + 1./108.*pow(m33,2)*RC146e*RC346e - 1./
         216.*pow(m33,2)*RC146e*RSe6611 );

      M13p6x22R +=  + Ab(m11,mu2)*pi16 * ( 1./432.*pow(m33,2)*pow(
         RC146e,2) - 1./72.*m11*RD346e*RS1463e + 1./144.*m11*RC146e*
         RD346e + 1./48.*m11*RD146e - 1./72.*m11*RD146e*RS1463e + 1./
         144.*m11*RD146e*RC146e - 1./18.*m11*me2*RSe6613*RS1463e + 1./
         36.*m11*me2*RSe6633*RS1463e + 1./36.*m11*me2*RSe6611*RS1463e
          + 1./36.*m11*me2*RC146e*RSe6613 - 1./72.*m11*me2*RC146e*
         RSe6633 - 1./72.*m11*me2*RC146e*RSe6611 + 1./24.*m11*m66*
         RS1463e - 1./48.*m11*m66*RC146e + 1./12.*m11*m44*RS1463e - 1./
         24.*m11*m44*RC146e - 1./16.*m11*m33 + 1./24.*m11*m33*RS1463e
          + 5./108.*m11*m33*RS1463e*RS3461e + 5./108.*m11*m33*pow(
         RS1463e,2) - 1./27.*m11*m33*RSe6613*RS1463e + 1./54.*m11*m33*
         RSe6633*RS1463e - 5./216.*m11*m33*RC346e*RS1463e + 1./54.*m11*
         m33*RSe6611*RS1463e + 1./24.*m11*m33*RS16e + 1./48.*m11*m33*
         RS14e - 1./48.*m11*m33*RC146e - 5./216.*m11*m33*RC146e*RS3461e
          - 5./108.*m11*m33*RC146e*RS1463e + 1./54.*m11*m33*RC146e*
         RSe6613 );

      M13p6x22R +=  + Ab(m11,mu2)*pi16 * (  - 1./108.*m11*m33*RC146e*
         RSe6633 + 5./432.*m11*m33*RC146e*RC346e - 1./108.*m11*m33*
         RC146e*RSe6611 + 5./432.*m11*m33*pow(RC146e,2) - 1./16.*pow(
         m11,2) + 1./108.*pow(m11,2)*RS1463e*RS3461e + 1./27.*pow(
         m11,2)*pow(RS1463e,2) - 1./54.*pow(m11,2)*RSe6613*RS1463e + 1./
         108.*pow(m11,2)*RSe6633*RS1463e - 1./216.*pow(m11,2)*RC346e*
         RS1463e + 1./108.*pow(m11,2)*RSe6611*RS1463e + 1./24.*pow(
         m11,2)*RS16e + 1./48.*pow(m11,2)*RS14e - 1./216.*pow(m11,2)*
         RC146e*RS3461e - 1./27.*pow(m11,2)*RC146e*RS1463e + 1./108.*
         pow(m11,2)*RC146e*RSe6613 - 1./216.*pow(m11,2)*RC146e*RSe6633
          + 1./432.*pow(m11,2)*RC146e*RC346e - 1./216.*pow(m11,2)*
         RC146e*RSe6611 + 1./108.*pow(m11,2)*pow(RC146e,2) );

      M13p6x22R +=  + pow(Ab(m11,mu2),2) * ( 1./729.*pow(m11,-2)*m33*
         m44*pow(m66,7)*pow(RN1e,3)*pow(RN3e,3) + 1./162.*pow(m11,-2)*
         m33*pow(m44,2)*pow(m66,6)*pow(RN1e,3)*pow(RN3e,3) + 5./486.*
         pow(m11,-2)*m33*pow(m44,3)*pow(m66,5)*pow(RN1e,3)*pow(RN3e,3)
          + 25./2916.*pow(m11,-2)*m33*pow(m44,4)*pow(m66,4)*pow(RN1e,3)
         *pow(RN3e,3) + 5./1296.*pow(m11,-2)*m33*pow(m44,5)*pow(m66,3)*
         pow(RN1e,3)*pow(RN3e,3) + 7./7776.*pow(m11,-2)*m33*pow(m44,6)*
         pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) + 1./11664.*pow(m11,-2)*m33
         *pow(m44,7)*m66*pow(RN1e,3)*pow(RN3e,3) - 17./4374.*pow(
         m11,-2)*pow(m33,2)*m44*pow(m66,6)*pow(RN1e,3)*pow(RN3e,3) - 83.
         /4374.*pow(m11,-2)*pow(m33,2)*pow(m44,2)*pow(m66,5)*pow(
         RN1e,3)*pow(RN3e,3) - 247./8748.*pow(m11,-2)*pow(m33,2)*pow(
         m44,3)*pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) - 41./2187.*pow(
         m11,-2)*pow(m33,2)*pow(m44,4)*pow(m66,3)*pow(RN1e,3)*pow(
         RN3e,3) - 409./69984.*pow(m11,-2)*pow(m33,2)*pow(m44,5)*pow(
         m66,2)*pow(RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + pow(Ab(m11,mu2),2) * (  - 49./69984.*pow(m11,-2)*
         pow(m33,2)*pow(m44,6)*m66*pow(RN1e,3)*pow(RN3e,3) - 1./972.*
         pow(m11,-2)*pow(m33,3)*m44*pow(m66,5)*pow(RN1e,3)*pow(RN3e,3)
          + 23./1944.*pow(m11,-2)*pow(m33,3)*pow(m44,2)*pow(m66,4)*pow(
         RN1e,3)*pow(RN3e,3) + 25./1296.*pow(m11,-2)*pow(m33,3)*pow(
         m44,3)*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) + 77./7776.*pow(
         m11,-2)*pow(m33,3)*pow(m44,4)*pow(m66,2)*pow(RN1e,3)*pow(
         RN3e,3) + 13./7776.*pow(m11,-2)*pow(m33,3)*pow(m44,5)*m66*pow(
         RN1e,3)*pow(RN3e,3) + 7./648.*pow(m11,-2)*pow(m33,4)*m44*pow(
         m66,4)*pow(RN1e,3)*pow(RN3e,3) + 1./108.*pow(m11,-2)*pow(
         m33,4)*pow(m44,2)*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) + 1./864.
         *pow(m11,-2)*pow(m33,4)*pow(m44,3)*pow(m66,2)*pow(RN1e,3)*pow(
         RN3e,3) - 1./2592.*pow(m11,-2)*pow(m33,4)*pow(m44,4)*m66*pow(
         RN1e,3)*pow(RN3e,3) - 5./648.*pow(m11,-2)*pow(m33,5)*m44*pow(
         m66,3)*pow(RN1e,3)*pow(RN3e,3) - 5./648.*pow(m11,-2)*pow(
         m33,5)*pow(m44,2)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + pow(Ab(m11,mu2),2) * (  - 5./2592.*pow(m11,-2)*
         pow(m33,5)*pow(m44,3)*m66*pow(RN1e,3)*pow(RN3e,3) + 1./1080.*
         pow(m11,-1)*RD146e*RD346e + 1./1080.*pow(m11,-1)*pow(RD146e,2)
          + 1./270.*pow(m11,-1)*me2*RD146e*RSe6613 - 1./540.*pow(
         m11,-1)*me2*RD146e*RSe6633 - 1./540.*pow(m11,-1)*me2*RD146e*
         RSe6611 + 2./405.*pow(m11,-1)*pow(me2,2)*RSe6613*RS1463e - 1./
         405.*pow(m11,-1)*pow(me2,2)*RSe6633*RS1463e - 1./405.*pow(
         m11,-1)*pow(me2,2)*RSe6611*RS1463e - 1./405.*pow(m11,-1)*pow(
         me2,2)*RC146e*RSe6613 + 1./810.*pow(m11,-1)*pow(me2,2)*RC146e*
         RSe6633 + 1./810.*pow(m11,-1)*pow(me2,2)*RC146e*RSe6611 + 7./
         540.*pow(m11,-1)*m66*RD146e - 2./81.*pow(m11,-1)*m66*me2*RS14e
         *RSe63 + 2./81.*pow(m11,-1)*m66*me2*RS14e*RSe61 - 1./162.*pow(
         m11,-1)*m66*me2*pow(RS14e,2) + 1./54.*pow(m11,-1)*pow(m66,2)*
         RSe63 - 1./54.*pow(m11,-1)*pow(m66,2)*RSe61 - 1./81.*pow(
         m11,-1)*pow(m66,2)*RS16e*RSe63 + 1./81.*pow(m11,-1)*pow(m66,2)
         *RS16e*RSe61 );

      M13p6x22R +=  + pow(Ab(m11,mu2),2) * ( 2./135.*pow(m11,-1)*pow(
         m66,2)*RS14e + 23./486.*pow(m11,-1)*pow(m66,2)*RS14e*RSe63 - 
         23./486.*pow(m11,-1)*pow(m66,2)*RS14e*RSe61 - 1./162.*pow(
         m11,-1)*pow(m66,2)*RS14e*RS16e + 5./486.*pow(m11,-1)*pow(
         m66,2)*pow(RS14e,2) + 7./270.*pow(m11,-1)*m44*RD146e + 1./162.
         *pow(m11,-1)*m44*me2*RS16e*RSe63 - 1./162.*pow(m11,-1)*m44*me2
         *RS16e*RSe61 - 1./324.*pow(m11,-1)*m44*me2*pow(RS16e,2) - 10./
         243.*pow(m11,-1)*m44*m66*RS16e*RSe63 + 10./243.*pow(m11,-1)*
         m44*m66*RS16e*RSe61 + 2./243.*pow(m11,-1)*m44*m66*pow(RS16e,2)
          + 11./243.*pow(m11,-1)*m44*m66*RS14e*RSe63 - 11./243.*pow(
         m11,-1)*m44*m66*RS14e*RSe61 - 1./81.*pow(m11,-1)*m44*m66*RS14e
         *RS16e + 2./243.*pow(m11,-1)*m44*m66*pow(RS14e,2) + 1./729.*
         pow(m11,-1)*m44*pow(m66,7)*pow(RN1e,3)*pow(RN3e,3) - 1./54.*
         pow(m11,-1)*pow(m44,2)*RSe63 + 1./54.*pow(m11,-1)*pow(m44,2)*
         RSe61 + 4./135.*pow(m11,-1)*pow(m44,2)*RS16e - 13./486.*pow(
         m11,-1)*pow(m44,2)*RS16e*RSe63 );

      M13p6x22R +=  + pow(Ab(m11,mu2),2) * ( 13./486.*pow(m11,-1)*pow(
         m44,2)*RS16e*RSe61 + 7./972.*pow(m11,-1)*pow(m44,2)*pow(
         RS16e,2) + 1./162.*pow(m11,-1)*pow(m44,2)*RS14e*RSe63 - 1./162.
         *pow(m11,-1)*pow(m44,2)*RS14e*RSe61 - 1./162.*pow(m11,-1)*pow(
         m44,2)*RS14e*RS16e + 1./162.*pow(m11,-1)*pow(m44,2)*pow(m66,6)
         *pow(RN1e,3)*pow(RN3e,3) + 5./486.*pow(m11,-1)*pow(m44,3)*pow(
         m66,5)*pow(RN1e,3)*pow(RN3e,3) + 25./2916.*pow(m11,-1)*pow(
         m44,4)*pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) + 5./1296.*pow(
         m11,-1)*pow(m44,5)*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) + 7./
         7776.*pow(m11,-1)*pow(m44,6)*pow(m66,2)*pow(RN1e,3)*pow(
         RN3e,3) + 1./11664.*pow(m11,-1)*pow(m44,7)*m66*pow(RN1e,3)*
         pow(RN3e,3) - 7./810.*pow(m11,-1)*m33*RD346e*RS1463e - 11./
         1620.*pow(m11,-1)*m33*RC146e*RD346e + 1./60.*pow(m11,-1)*m33*
         RD146e - 7./1080.*pow(m11,-1)*m33*RD146e*RS3461e - 17./3240.*
         pow(m11,-1)*m33*RD146e*RS1463e + 1./108.*pow(m11,-1)*m33*
         RD146e*RSe6613 );

      M13p6x22R +=  + pow(Ab(m11,mu2),2) * (  - 1./216.*pow(m11,-1)*m33
         *RD146e*RSe6633 + 7./2160.*pow(m11,-1)*m33*RD146e*RC346e - 1./
         216.*pow(m11,-1)*m33*RD146e*RSe6611 + 13./1296.*pow(m11,-1)*
         m33*RD146e*RC146e - 17./405.*pow(m11,-1)*m33*me2*RSe6613*
         RS1463e + 1./45.*pow(m11,-1)*m33*me2*RSe6633*RS1463e + 1./810.
         *pow(m11,-1)*m33*me2*RSe6611*RS1463e - 2./405.*pow(m11,-1)*m33
         *me2*RC146e*RSe6613 + 1./90.*pow(m11,-1)*m33*me2*RC146e*
         RSe6633 - 5./324.*pow(m11,-1)*m33*me2*RC146e*RSe6611 - 127./
         6480.*pow(m11,-1)*m33*m66*RS1463e + 1./27.*pow(m11,-1)*m33*m66
         *RSe6613 + 1./54.*pow(m11,-1)*m33*m66*RSe63 + 7./1080.*pow(
         m11,-1)*m33*m66*RS14e - 1./18.*pow(m11,-1)*m33*m66*RS14e*RSe63
          + 31./2592.*pow(m11,-1)*m33*m66*RC146e - 1./729.*pow(m11,-1)*
         m33*pow(m66,7)*pow(RN1e,3)*pow(RN3e,3) - 127./3240.*pow(
         m11,-1)*m33*m44*RS1463e + 2./27.*pow(m11,-1)*m33*m44*RSe6613
          - 1./54.*pow(m11,-1)*m33*m44*RSe63 + 7./540.*pow(m11,-1)*m33*
         m44*RS16e );

      M13p6x22R +=  + pow(Ab(m11,mu2),2) * ( 2./27.*pow(m11,-1)*m33*m44
         *RS16e*RSe63 - 1./54.*pow(m11,-1)*m33*m44*RS14e*RSe63 + 31./
         1296.*pow(m11,-1)*m33*m44*RC146e - 20./2187.*pow(m11,-1)*m33*
         m44*pow(m66,6)*pow(RN1e,3)*pow(RN3e,3) - 157./4374.*pow(
         m11,-1)*m33*pow(m44,2)*pow(m66,5)*pow(RN1e,3)*pow(RN3e,3) - 
         235./4374.*pow(m11,-1)*m33*pow(m44,3)*pow(m66,4)*pow(RN1e,3)*
         pow(RN3e,3) - 1315./34992.*pow(m11,-1)*m33*pow(m44,4)*pow(
         m66,3)*pow(RN1e,3)*pow(RN3e,3) - 227./17496.*pow(m11,-1)*m33*
         pow(m44,5)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) - 35./17496.*
         pow(m11,-1)*m33*pow(m44,6)*m66*pow(RN1e,3)*pow(RN3e,3) - 1./
         11664.*pow(m11,-1)*m33*pow(m44,7)*pow(RN1e,3)*pow(RN3e,3) - 1./
         288.*pow(m11,-1)*pow(m33,2) + 1./1440.*pow(m11,-1)*pow(m33,2)*
         RS1463e + 1./162.*pow(m11,-1)*pow(m33,2)*RS1463e*RS3461e - 1./
         81.*pow(m11,-1)*pow(m33,2)*pow(RS1463e,2) + 2./81.*pow(m11,-1)
         *pow(m33,2)*RSe6613*RS1463e - 5./324.*pow(m11,-1)*pow(m33,2)*
         RSe6633*RS1463e );

      M13p6x22R +=  + pow(Ab(m11,mu2),2) * (  - 7./3240.*pow(m11,-1)*
         pow(m33,2)*RC346e*RS1463e - 1./108.*pow(m11,-1)*pow(m33,2)*
         RSe6611*RS1463e + 1./432.*pow(m11,-1)*pow(m33,2)*RS16e + 1./
         864.*pow(m11,-1)*pow(m33,2)*RS14e + 23./864.*pow(m11,-1)*pow(
         m33,2)*RC146e + 1./120.*pow(m11,-1)*pow(m33,2)*RC146e*RS3461e
          + 1./108.*pow(m11,-1)*pow(m33,2)*RC146e*RS1463e - 1./108.*
         pow(m11,-1)*pow(m33,2)*RC146e*RSe6613 + 1./648.*pow(m11,-1)*
         pow(m33,2)*RC146e*RSe6633 - 1./144.*pow(m11,-1)*pow(m33,2)*
         RC146e*RC346e + 5./648.*pow(m11,-1)*pow(m33,2)*RC146e*RSe6611
          - 5./1296.*pow(m11,-1)*pow(m33,2)*pow(RC146e,2) + 17./4374.*
         pow(m11,-1)*pow(m33,2)*pow(m66,6)*pow(RN1e,3)*pow(RN3e,3) + 41.
         /8748.*pow(m11,-1)*pow(m33,2)*m44*pow(m66,5)*pow(RN1e,3)*pow(
         RN3e,3) + 919./17496.*pow(m11,-1)*pow(m33,2)*pow(m44,2)*pow(
         m66,4)*pow(RN1e,3)*pow(RN3e,3) + 959./11664.*pow(m11,-1)*pow(
         m33,2)*pow(m44,3)*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) + 409./
         8748.*pow(m11,-1)*pow(m33,2)*pow(m44,4)*pow(m66,2)*pow(RN1e,3)
         *pow(RN3e,3) );

      M13p6x22R +=  + pow(Ab(m11,mu2),2) * ( 743./69984.*pow(m11,-1)*
         pow(m33,2)*pow(m44,5)*m66*pow(RN1e,3)*pow(RN3e,3) + 49./69984.
         *pow(m11,-1)*pow(m33,2)*pow(m44,6)*pow(RN1e,3)*pow(RN3e,3) + 1.
         /972.*pow(m11,-1)*pow(m33,3)*pow(m66,5)*pow(RN1e,3)*pow(
         RN3e,3) + 5./108.*pow(m11,-1)*pow(m33,3)*m44*pow(m66,4)*pow(
         RN1e,3)*pow(RN3e,3) + 23./3888.*pow(m11,-1)*pow(m33,3)*pow(
         m44,2)*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) - 17./486.*pow(
         m11,-1)*pow(m33,3)*pow(m44,3)*pow(m66,2)*pow(RN1e,3)*pow(
         RN3e,3) - 43./2592.*pow(m11,-1)*pow(m33,3)*pow(m44,4)*m66*pow(
         RN1e,3)*pow(RN3e,3) - 13./7776.*pow(m11,-1)*pow(m33,3)*pow(
         m44,5)*pow(RN1e,3)*pow(RN3e,3) - 7./648.*pow(m11,-1)*pow(
         m33,4)*pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) - 53./648.*pow(
         m11,-1)*pow(m33,4)*m44*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) - 
         133./2592.*pow(m11,-1)*pow(m33,4)*pow(m44,2)*pow(m66,2)*pow(
         RN1e,3)*pow(RN3e,3) - 5./864.*pow(m11,-1)*pow(m33,4)*pow(
         m44,3)*m66*pow(RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + pow(Ab(m11,mu2),2) * ( 1./2592.*pow(m11,-1)*pow(
         m33,4)*pow(m44,4)*pow(RN1e,3)*pow(RN3e,3) + 5./648.*pow(
         m11,-1)*pow(m33,5)*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) + 25./
         648.*pow(m11,-1)*pow(m33,5)*m44*pow(m66,2)*pow(RN1e,3)*pow(
         RN3e,3) + 55./2592.*pow(m11,-1)*pow(m33,5)*pow(m44,2)*m66*pow(
         RN1e,3)*pow(RN3e,3) + 5./2592.*pow(m11,-1)*pow(m33,5)*pow(
         m44,3)*pow(RN1e,3)*pow(RN3e,3) - 1./405.*RD346e*RS1463e - 1./
         162.*RC146e*RD346e + 1./45.*RD146e - 1./216.*RD146e*RS3461e - 
         7./3240.*RD146e*RS1463e - 1./36.*RD146e*RSe6613 - 1./216.*
         RD146e*RSe6633 + 1./432.*RD146e*RC346e + 7./216.*RD146e*
         RSe6611 - 41./6480.*RD146e*RC146e - 4./81.*me2*RSe6613*RS1463e
          + 2./405.*me2*RSe6633*RS1463e - 1./90.*me2*RSe6611*RS1463e + 
         2./27.*me2*RSe6611*RSe6613 - 2./27.*me2*pow(RSe6611,2) - 19./
         810.*me2*RC146e*RSe6613 + 1./81.*me2*RC146e*RSe6633 + 11./540.
         *me2*RC146e*RSe6611 - 1./54.*m66*RS3461e - 61./2160.*m66*
         RS1463e );

      M13p6x22R +=  + pow(Ab(m11,mu2),2) * ( 2./81.*m66*RSe6613 + 2./27.
         *m66*RSe63 - 1./108.*m66*RS34e - 1./27.*m66*RSe61 + 4./81.*m66
         *RSe6611 + 1./81.*m66*RS16e*RS36e + 1./162.*m66*RS16e*RS34e - 
         1./162.*m66*pow(RS16e,2) - 17./1080.*m66*RS14e - 1./9.*m66*
         RS14e*RSe63 + 1./162.*m66*RS14e*RS36e + 5./162.*m66*RS14e*
         RS34e - 1./162.*m66*RS14e*RS16e - 5./324.*m66*pow(RS14e,2) + 
         23./2592.*m66*RC146e - 1./729.*pow(m66,7)*pow(RN1e,3)*pow(
         RN3e,3) - 1./27.*m44*RS3461e - 61./1080.*m44*RS1463e + 4./81.*
         m44*RSe6613 - 2./27.*m44*RSe63 - 1./54.*m44*RS36e + 1./27.*m44
         *RSe61 + 8./81.*m44*RSe6611 - 17./540.*m44*RS16e + 4./27.*m44*
         RS16e*RSe63 + 7./81.*m44*RS16e*RS36e + 1./162.*m44*RS16e*RS34e
          - 7./162.*m44*pow(RS16e,2) - 1./27.*m44*RS14e*RSe63 + 1./162.
         *m44*RS14e*RS36e + 1./81.*m44*RS14e*RS34e - 1./162.*m44*RS14e*
         RS16e - 1./162.*m44*pow(RS14e,2) + 23./1296.*m44*RC146e - 23./
         4374.*m44*pow(m66,6)*pow(RN1e,3)*pow(RN3e,3) - 37./2187.*pow(
         m44,2)*pow(m66,5)*pow(RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + pow(Ab(m11,mu2),2) * (  - 223./8748.*pow(m44,3)*
         pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) - 659./34992.*pow(m44,4)*
         pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) - 499./69984.*pow(m44,5)*
         pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) - 91./69984.*pow(m44,6)*m66
         *pow(RN1e,3)*pow(RN3e,3) - 1./11664.*pow(m44,7)*pow(RN1e,3)*
         pow(RN3e,3) + 19./144.*m33 + 13./216.*m33*RS1463e + 1./108.*
         m33*RS1463e*RS3461e + 4./135.*m33*pow(RS1463e,2) + 1./9.*m33*
         RSe6613 + 5./162.*m33*RSe6613*RS1463e - 5./324.*m33*RSe6633*
         RS1463e + 17./3240.*m33*RC346e*RS1463e - 5./324.*m33*RSe6611*
         RS1463e - 8./135.*m33*RS16e - 4./135.*m33*RS14e + 25./864.*m33
         *RC146e + 41./1080.*m33*RC146e*RS3461e + 61./1620.*m33*RC146e*
         RS1463e - 17./324.*m33*RC146e*RSe6613 - 1./648.*m33*RC146e*
         RSe6633 - 7./1296.*m33*RC146e*RC346e + 11./648.*m33*RC146e*
         RSe6611 + 1./162.*m33*pow(RC146e,2) + 13./4374.*m33*pow(m66,6)
         *pow(RN1e,3)*pow(RN3e,3) - 53./8748.*m33*m44*pow(m66,5)*pow(
         RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + pow(Ab(m11,mu2),2) * ( 533./17496.*m33*pow(m44,2)
         *pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) + 785./11664.*m33*pow(
         m44,3)*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) + 1571./34992.*m33*
         pow(m44,4)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) + 847./69984.*
         m33*pow(m44,5)*m66*pow(RN1e,3)*pow(RN3e,3) + 77./69984.*m33*
         pow(m44,6)*pow(RN1e,3)*pow(RN3e,3) + 125./8748.*pow(m33,2)*
         pow(m66,5)*pow(RN1e,3)*pow(RN3e,3) + 1031./8748.*pow(m33,2)*
         m44*pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) + 1913./34992.*pow(
         m33,2)*pow(m44,2)*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) - 799./
         17496.*pow(m33,2)*pow(m44,3)*pow(m66,2)*pow(RN1e,3)*pow(
         RN3e,3) - 1127./34992.*pow(m33,2)*pow(m44,4)*m66*pow(RN1e,3)*
         pow(RN3e,3) - 167./34992.*pow(m33,2)*pow(m44,5)*pow(RN1e,3)*
         pow(RN3e,3) - 113./1944.*pow(m33,3)*pow(m66,4)*pow(RN1e,3)*
         pow(RN3e,3) - 1043./3888.*pow(m33,3)*m44*pow(m66,3)*pow(
         RN1e,3)*pow(RN3e,3) - 7./48.*pow(m33,3)*pow(m44,2)*pow(m66,2)*
         pow(RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + pow(Ab(m11,mu2),2) * ( 1./3888.*pow(m33,3)*pow(
         m44,3)*m66*pow(RN1e,3)*pow(RN3e,3) + 13./1944.*pow(m33,3)*pow(
         m44,4)*pow(RN1e,3)*pow(RN3e,3) + 47./648.*pow(m33,4)*pow(
         m66,3)*pow(RN1e,3)*pow(RN3e,3) + 581./2592.*pow(m33,4)*m44*
         pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) + 227./2592.*pow(m33,4)*
         pow(m44,2)*m66*pow(RN1e,3)*pow(RN3e,3) + 1./216.*pow(m33,4)*
         pow(m44,3)*pow(RN1e,3)*pow(RN3e,3) - 5./162.*pow(m33,5)*pow(
         m66,2)*pow(RN1e,3)*pow(RN3e,3) - 155./2592.*pow(m33,5)*m44*m66
         *pow(RN1e,3)*pow(RN3e,3) - 35./2592.*pow(m33,5)*pow(m44,2)*
         pow(RN1e,3)*pow(RN3e,3) + 13./96.*m11 - 1./18.*m11*RS3461e - 
         223./4320.*m11*RS1463e - 1./324.*m11*RS1463e*RS3461e + 41./810.
         *m11*pow(RS1463e,2) + 5./27.*m11*RSe6613 + 1./162.*m11*RSe6613
         *RS1463e - 1./54.*m11*RS36e - 1./108.*m11*RS34e + 1./27.*m11*
         RSe6611 - 1./162.*m11*RSe6611*RS1463e + 19./2160.*m11*RS16e + 
         19./4320.*m11*RS14e + 1./48.*m11*RC146e + 13./324.*m11*RC146e*
         RS3461e );

      M13p6x22R +=  + pow(Ab(m11,mu2),2) * ( 43./540.*m11*RC146e*
         RS1463e - 31./162.*m11*RC146e*RSe6613 - 1./324.*m11*RC146e*
         RSe6633 + 1./648.*m11*RC146e*RC346e + 13./108.*m11*RC146e*
         RSe6611 - 35./1296.*m11*pow(RC146e,2) - 2./27.*m11*me2*RC146e*
         RN1e*RSe6613 + 2./27.*m11*me2*RC146e*RN1e*RSe6611 - 2./2187.*
         m11*pow(m66,6)*pow(RN1e,3)*pow(RN3e,3) - 103./8748.*m11*m44*
         pow(m66,5)*pow(RN1e,3)*pow(RN3e,3) - 179./17496.*m11*pow(
         m44,2)*pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) + 17./3888.*m11*pow(
         m44,3)*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) + 563./69984.*m11*
         pow(m44,4)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) + 221./69984.*
         m11*pow(m44,5)*m66*pow(RN1e,3)*pow(RN3e,3) + 7./17496.*m11*
         pow(m44,6)*pow(RN1e,3)*pow(RN3e,3) - 2./27.*m11*m33*RC146e*
         RN1e*RSe6613 + 1./27.*m11*m33*RN13*RC146e*RS3461e + 277./8748.
         *m11*m33*pow(m66,5)*pow(RN1e,3)*pow(RN3e,3) + 1225./8748.*m11*
         m33*m44*pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) + 4123./34992.*m11*
         m33*pow(m44,2)*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + pow(Ab(m11,mu2),2) * ( 73./4374.*m11*m33*pow(
         m44,3)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) - 809./69984.*m11*
         m33*pow(m44,4)*m66*pow(RN1e,3)*pow(RN3e,3) - 209./69984.*m11*
         m33*pow(m44,5)*pow(RN1e,3)*pow(RN3e,3) - 829./5832.*m11*pow(
         m33,2)*pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) - 5119./11664.*m11*
         pow(m33,2)*m44*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) - 1105./3888.
         *m11*pow(m33,2)*pow(m44,2)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3)
          - 487./11664.*m11*pow(m33,2)*pow(m44,3)*m66*pow(RN1e,3)*pow(
         RN3e,3) + 49./11664.*m11*pow(m33,2)*pow(m44,4)*pow(RN1e,3)*
         pow(RN3e,3) + 35./144.*m11*pow(m33,3)*pow(m66,3)*pow(RN1e,3)*
         pow(RN3e,3) + 44./81.*m11*pow(m33,3)*m44*pow(m66,2)*pow(
         RN1e,3)*pow(RN3e,3) + 293./1296.*m11*pow(m33,3)*pow(m44,2)*m66
         *pow(RN1e,3)*pow(RN3e,3) + 5./324.*m11*pow(m33,3)*pow(m44,3)*
         pow(RN1e,3)*pow(RN3e,3) - 451./2592.*m11*pow(m33,4)*pow(m66,2)
         *pow(RN1e,3)*pow(RN3e,3) - 673./2592.*m11*pow(m33,4)*m44*m66*
         pow(RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + pow(Ab(m11,mu2),2) * (  - 59./1296.*m11*pow(
         m33,4)*pow(m44,2)*pow(RN1e,3)*pow(RN3e,3) + 35./864.*m11*pow(
         m33,5)*m66*pow(RN1e,3)*pow(RN3e,3) + 25./864.*m11*pow(m33,5)*
         m44*pow(RN1e,3)*pow(RN3e,3) - 2./27.*pow(m11,2)*RC146e*RN1e*
         RSe6613 - 2./27.*pow(m11,2)*RC146e*RN1e*RSe6611 + 2./27.*pow(
         m11,2)*RN13*RC146e*RS3461e + 161./8748.*pow(m11,2)*pow(m66,5)*
         pow(RN1e,3)*pow(RN3e,3) + 1009./17496.*pow(m11,2)*m44*pow(
         m66,4)*pow(RN1e,3)*pow(RN3e,3) + 2093./34992.*pow(m11,2)*pow(
         m44,2)*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) + 1835./69984.*pow(
         m11,2)*pow(m44,3)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) + 311./
         69984.*pow(m11,2)*pow(m44,4)*m66*pow(RN1e,3)*pow(RN3e,3) + 1./
         8748.*pow(m11,2)*pow(m44,5)*pow(RN1e,3)*pow(RN3e,3) - 731./
         5832.*pow(m11,2)*m33*pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) - 899./
         2916.*pow(m11,2)*m33*m44*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) - 
         1801./7776.*pow(m11,2)*m33*pow(m44,2)*pow(m66,2)*pow(RN1e,3)*
         pow(RN3e,3) );

      M13p6x22R +=  + pow(Ab(m11,mu2),2) * (  - 1471./23328.*pow(m11,2)
         *m33*pow(m44,3)*m66*pow(RN1e,3)*pow(RN3e,3) - 101./23328.*pow(
         m11,2)*m33*pow(m44,4)*pow(RN1e,3)*pow(RN3e,3) + 1247./3888.*
         pow(m11,2)*pow(m33,2)*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) + 
         1511./2592.*pow(m11,2)*pow(m33,2)*m44*pow(m66,2)*pow(RN1e,3)*
         pow(RN3e,3) + 721./2592.*pow(m11,2)*pow(m33,2)*pow(m44,2)*m66*
         pow(RN1e,3)*pow(RN3e,3) + 65./1944.*pow(m11,2)*pow(m33,2)*pow(
         m44,3)*pow(RN1e,3)*pow(RN3e,3) - 965./2592.*pow(m11,2)*pow(
         m33,3)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) - 1157./2592.*pow(
         m11,2)*pow(m33,3)*m44*m66*pow(RN1e,3)*pow(RN3e,3) - 127./1296.
         *pow(m11,2)*pow(m33,3)*pow(m44,2)*pow(RN1e,3)*pow(RN3e,3) + 77.
         /432.*pow(m11,2)*pow(m33,4)*m66*pow(RN1e,3)*pow(RN3e,3) + 23./
         216.*pow(m11,2)*pow(m33,4)*m44*pow(RN1e,3)*pow(RN3e,3) - 5./
         288.*pow(m11,2)*pow(m33,5)*pow(RN1e,3)*pow(RN3e,3) - 89./2916.
         *pow(m11,3)*pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) - 371./5832.*
         pow(m11,3)*m44*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + pow(Ab(m11,mu2),2) * (  - 193./3888.*pow(m11,3)*
         pow(m44,2)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) - 401./23328.*
         pow(m11,3)*pow(m44,3)*m66*pow(RN1e,3)*pow(RN3e,3) - 13./5832.*
         pow(m11,3)*pow(m44,4)*pow(RN1e,3)*pow(RN3e,3) + 305./1944.*
         pow(m11,3)*m33*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) + 20./81.*
         pow(m11,3)*m33*m44*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) + 335./
         2592.*pow(m11,3)*m33*pow(m44,2)*m66*pow(RN1e,3)*pow(RN3e,3) + 
         175./7776.*pow(m11,3)*m33*pow(m44,3)*pow(RN1e,3)*pow(RN3e,3)
          - 127./432.*pow(m11,3)*pow(m33,2)*pow(m66,2)*pow(RN1e,3)*pow(
         RN3e,3) - 269./864.*pow(m11,3)*pow(m33,2)*m44*m66*pow(RN1e,3)*
         pow(RN3e,3) - 71./864.*pow(m11,3)*pow(m33,2)*pow(m44,2)*pow(
         RN1e,3)*pow(RN3e,3) + 203./864.*pow(m11,3)*pow(m33,3)*m66*pow(
         RN1e,3)*pow(RN3e,3) + 109./864.*pow(m11,3)*pow(m33,3)*m44*pow(
         RN1e,3)*pow(RN3e,3) - 19./288.*pow(m11,3)*pow(m33,4)*pow(
         RN1e,3)*pow(RN3e,3) + 7./486.*pow(m11,4)*pow(m66,3)*pow(
         RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + pow(Ab(m11,mu2),2) * ( 7./324.*pow(m11,4)*m44*
         pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) + 7./648.*pow(m11,4)*pow(
         m44,2)*m66*pow(RN1e,3)*pow(RN3e,3) + 7./3888.*pow(m11,4)*pow(
         m44,3)*pow(RN1e,3)*pow(RN3e,3) - 7./108.*pow(m11,4)*m33*pow(
         m66,2)*pow(RN1e,3)*pow(RN3e,3) - 7./108.*pow(m11,4)*m33*m44*
         m66*pow(RN1e,3)*pow(RN3e,3) - 7./432.*pow(m11,4)*m33*pow(
         m44,2)*pow(RN1e,3)*pow(RN3e,3) + 7./72.*pow(m11,4)*pow(m33,2)*
         m66*pow(RN1e,3)*pow(RN3e,3) + 7./144.*pow(m11,4)*pow(m33,2)*
         m44*pow(RN1e,3)*pow(RN3e,3) - 7./144.*pow(m11,4)*pow(m33,3)*
         pow(RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + Ab(m11,mu2)*Ab(m13,mu2) * (  - 1./162.*RD346e*
         RS1463e + 1./324.*RC146e*RD346e - 1./54.*RD146e*RS3461e - 2./
         81.*RD146e*RS1463e + 1./27.*RD146e*RSe6613 - 1./54.*RD146e*
         RSe6633 + 1./108.*RD146e*RC346e - 1./54.*RD146e*RSe6611 + 1./
         81.*RD146e*RC146e - 2./81.*me2*RSe6613*RS1463e + 1./81.*me2*
         RSe6633*RS1463e + 1./81.*me2*RSe6611*RS1463e + 1./81.*me2*
         RC146e*RSe6613 - 1./162.*me2*RC146e*RSe6633 - 1./162.*me2*
         RC146e*RSe6611 - 1./72.*m33 - 1./27.*m33*RS1463e*RS3461e - 4./
         81.*m33*pow(RS1463e,2) + 8./81.*m33*RSe6613*RS1463e - 5./81.*
         m33*RSe6633*RS1463e + 2./81.*m33*RC346e*RS1463e - 1./27.*m33*
         RSe6611*RS1463e + 1./81.*m33*RC146e*RS3461e + 1./27.*m33*
         RC146e*RS1463e - 1./27.*m33*RC146e*RSe6613 + 1./162.*m33*
         RC146e*RSe6633 + 5./162.*m33*RC146e*RSe6611 - 5./324.*m33*pow(
         RC146e,2) + 1./9.*pow(m33,2)*RN13 + 1./24.*m11 - 1./81.*m11*
         RS1463e*RS3461e + 2./81.*m11*RSe6613*RS1463e - 2./81.*m11*
         RSe6611*RS1463e );

      M13p6x22R +=  + Ab(m11,mu2)*Ab(m13,mu2) * ( 1./81.*m11*RC146e*
         RS3461e + 1./81.*m11*RC146e*RS1463e - 2./81.*m11*RC146e*
         RSe6613 - 1./81.*m11*RC146e*RSe6633 + 1./162.*m11*RC146e*
         RC346e + 1./27.*m11*RC146e*RSe6611 - 5./324.*m11*pow(RC146e,2)
          + 5./9.*m11*m33*RN13 + 1./3.*pow(m11,2)*RN13 );

      M13p6x22R +=  + Ab(m11,mu2)*Ab(m14,mu2) * (  - 29./270.*RD146e - 
         2./27.*m44*RSe63 + 2./27.*m44*RSe61 - 1./135.*m44*RS16e - 2./
         27.*m44*me2*RN1e*RSe63 + 2./27.*m44*me2*RN1e*RSe61 + 41./540.*
         m33*RS1463e - 4./27.*m33*RSe6613 - 2./27.*m33*RSe63 - 1./12.*
         m33*RS16e - 4./27.*m33*me2*RN1e*RSe6613 - 2./27.*m33*me2*RN1e*
         RSe63 - 2./27.*m33*m44*RN1e*RSe63 - 2./27.*m33*m44*RN13*RS36e
          + 2./27.*m11*RS3461e + 109./540.*m11*RS1463e - 4./27.*m11*
         RSe6613 - 2./9.*m11*RSe63 - 2./27.*m11*RS36e + 2./27.*m11*
         RSe61 - 4./27.*m11*RSe6611 + 19./540.*m11*RS16e - 4./27.*m11*
         me2*RN1e*RSe6613 - 2./9.*m11*me2*RN1e*RSe63 + 2./27.*m11*me2*
         RN1e*RSe61 - 4./27.*m11*me2*RN1e*RSe6611 - 2./27.*m11*m44*RN1e
         *RSe63 - 2./27.*m11*m44*RN1e*RSe61 - 4./27.*m11*m44*RN13*RS36e
          + 4./27.*m11*m33*RN1e*RSe6613 - 2./27.*m11*m33*RN1e*RSe63 + 2.
         /27.*m11*m33*RN13*RS3461e - 2./9.*m11*m33*RN13*RS36e + 4./27.*
         pow(m11,2)*RN1e*RSe6613 - 2./27.*pow(m11,2)*RN1e*RSe63 - 2./27.
         *pow(m11,2)*RN1e*RSe61 );

      M13p6x22R +=  + Ab(m11,mu2)*Ab(m14,mu2) * ( 4./27.*pow(m11,2)*
         RN1e*RSe6611 - 2./27.*pow(m11,2)*RN13*RS3461e - 2./9.*pow(
         m11,2)*RN13*RS36e );

      M13p6x22R +=  + Ab(m11,mu2)*Ab(m16,mu2) * (  - 29./540.*RD146e + 
         2./27.*m66*RSe63 - 2./27.*m66*RSe61 - 1./270.*m66*RS14e + 2./
         27.*m66*me2*RN1e*RSe63 - 2./27.*m66*me2*RN1e*RSe61 + 41./1080.
         *m33*RS1463e - 2./27.*m33*RSe6613 + 2./27.*m33*RSe63 - 1./24.*
         m33*RS14e - 2./27.*m33*me2*RN1e*RSe6613 + 2./27.*m33*me2*RN1e*
         RSe63 + 2./27.*m33*m66*RN1e*RSe63 - 1./27.*m33*m66*RN13*RS34e
          + 1./27.*m11*RS3461e + 109./1080.*m11*RS1463e - 2./27.*m11*
         RSe6613 + 2./9.*m11*RSe63 - 1./27.*m11*RS34e - 2./27.*m11*
         RSe61 - 2./27.*m11*RSe6611 + 19./1080.*m11*RS14e - 2./27.*m11*
         me2*RN1e*RSe6613 + 2./9.*m11*me2*RN1e*RSe63 - 2./27.*m11*me2*
         RN1e*RSe61 - 2./27.*m11*me2*RN1e*RSe6611 + 2./27.*m11*m66*RN1e
         *RSe63 + 2./27.*m11*m66*RN1e*RSe61 - 2./27.*m11*m66*RN13*RS34e
          + 2./27.*m11*m33*RN1e*RSe6613 + 2./27.*m11*m33*RN1e*RSe63 + 1.
         /27.*m11*m33*RN13*RS3461e - 1./9.*m11*m33*RN13*RS34e + 2./27.*
         pow(m11,2)*RN1e*RSe6613 + 2./27.*pow(m11,2)*RN1e*RSe63 + 2./27.
         *pow(m11,2)*RN1e*RSe61 );

      M13p6x22R +=  + Ab(m11,mu2)*Ab(m16,mu2) * ( 2./27.*pow(m11,2)*
         RN1e*RSe6611 - 1./27.*pow(m11,2)*RN13*RS3461e - 1./9.*pow(
         m11,2)*RN13*RS34e );

      M13p6x22R +=  + Ab(m11,mu2)*Ab(m33,mu2) * (  - 1./270.*RD346e*
         RS1463e + 1./540.*RC146e*RD346e - 1./270.*RD146e*RS3461e + 1./
         540.*RD146e*RC346e + 1./18.*m33*RS1463e*RS3461e - 1./60.*m33*
         RC346e*RS1463e - 1./540.*m33*RC146e*RS3461e - 1./72.*m33*
         RC146e*RC346e + 1./27.*pow(m33,2)*RN13*RC346e*RS1463e + 1./18.
         *m11*RS1463e*RS3461e - 1./540.*m11*RC346e*RS1463e + 11./540.*
         m11*RC146e*RS3461e - 1./72.*m11*RC146e*RC346e + 2./27.*m11*m33
         *RN13*RC346e*RS1463e - 1./27.*m11*m33*RN13*RC146e*RS3461e - 2./
         27.*pow(m11,2)*RN13*RC146e*RS3461e );

      M13p6x22R +=  + Ab(m11,mu2)*Ab(m34,mu2) * (  - 1./90.*RD146e + 17.
         /135.*m44*RS16e - 19./540.*m33*RS1463e + 1./108.*m33*RS16e + 1.
         /12.*m33*RC146e - 2./27.*m33*m44*RN13*RS16e - 2./27.*pow(
         m33,2)*RN13*RS1463e - 2./27.*pow(m33,2)*RN13*RS16e + 89./540.*
         m11*RS1463e - 83./540.*m11*RS16e + 1./12.*m11*RC146e - 4./27.*
         m11*m44*RN13*RS16e - 8./27.*m11*m33*RN13*RS16e + 2./27.*pow(
         m11,2)*RN13*RS1463e - 2./27.*pow(m11,2)*RN13*RS16e );

      M13p6x22R +=  + Ab(m11,mu2)*Ab(m36,mu2) * (  - 1./180.*RD146e + 
         17./270.*m66*RS14e - 19./1080.*m33*RS1463e + 1./216.*m33*RS14e
          + 1./24.*m33*RC146e - 1./27.*m33*m66*RN13*RS14e - 1./27.*pow(
         m33,2)*RN13*RS1463e - 1./27.*pow(m33,2)*RN13*RS14e + 89./1080.
         *m11*RS1463e - 83./1080.*m11*RS14e + 1./24.*m11*RC146e - 2./27.
         *m11*m66*RN13*RS14e - 4./27.*m11*m33*RN13*RS14e + 1./27.*pow(
         m11,2)*RN13*RS1463e - 1./27.*pow(m11,2)*RN13*RS14e );

      M13p6x22R +=  + Ab(m11,mu2)*Ab(m44,mu2) * ( 1./9.*m44*RS16e*RSe63
          - 1./9.*m44*RS16e*RSe61 - 1./18.*m44*pow(RS16e,2) + 1./9.*m44
         *me2*RN1e*RS16e*RSe63 - 1./9.*m44*me2*RN1e*RS16e*RSe61 + 1./9.
         *m33*m44*RN1e*RS16e*RSe63 + 1./9.*m33*m44*RN13*RS16e*RS36e + 1.
         /9.*m11*m44*RN1e*RS16e*RSe63 + 1./9.*m11*m44*RN1e*RS16e*RSe61
          + 2./9.*m11*m44*RN13*RS16e*RS36e );

      M13p6x22R +=  + Ab(m11,mu2)*Ab(m46,mu2) * (  - 4./81.*m66*RS16e*
         RSe63 + 4./81.*m66*RS16e*RSe61 - 2./81.*m66*RS14e*RSe63 + 2./
         81.*m66*RS14e*RSe61 - 2./81.*m66*RS14e*RS16e - 1./81.*m66*pow(
         RS14e,2) - 4./81.*m66*me2*RN1e*RS16e*RSe63 + 4./81.*m66*me2*
         RN1e*RS16e*RSe61 - 2./81.*m66*me2*RN1e*RS14e*RSe63 + 2./81.*
         m66*me2*RN1e*RS14e*RSe61 - 2./81.*m44*RS16e*RSe63 + 2./81.*m44
         *RS16e*RSe61 - 1./81.*m44*pow(RS16e,2) + 2./81.*m44*RS14e*
         RSe63 - 2./81.*m44*RS14e*RSe61 - 2./81.*m44*RS14e*RS16e - 2./
         81.*m44*me2*RN1e*RS16e*RSe63 + 2./81.*m44*me2*RN1e*RS16e*RSe61
          + 2./81.*m44*me2*RN1e*RS14e*RSe63 - 2./81.*m44*me2*RN1e*RS14e
         *RSe61 - 2./27.*m33*RS16e*RSe63 + 2./27.*m33*RS14e*RSe63 - 2./
         27.*m33*me2*RN1e*RS16e*RSe63 + 2./27.*m33*me2*RN1e*RS14e*RSe63
          - 4./81.*m33*m66*RN1e*RS16e*RSe63 - 2./81.*m33*m66*RN1e*RS14e
         *RSe63 + 2./81.*m33*m66*RN13*RS16e*RS34e + 2./81.*m33*m66*RN13
         *RS14e*RS36e + 2./81.*m33*m66*RN13*RS14e*RS34e - 2./81.*m33*
         m44*RN1e*RS16e*RSe63 );

      M13p6x22R +=  + Ab(m11,mu2)*Ab(m46,mu2) * ( 2./81.*m33*m44*RN1e*
         RS14e*RSe63 + 2./81.*m33*m44*RN13*RS16e*RS36e + 2./81.*m33*m44
         *RN13*RS16e*RS34e + 2./81.*m33*m44*RN13*RS14e*RS36e - 4./27.*
         m11*RS16e*RSe63 - 2./81.*m11*RS16e*RS36e + 2./81.*m11*RS16e*
         RS34e + 1./81.*m11*pow(RS16e,2) + 4./27.*m11*RS14e*RSe63 + 2./
         81.*m11*RS14e*RS36e - 2./81.*m11*RS14e*RS34e - 2./81.*m11*
         RS14e*RS16e + 1./81.*m11*pow(RS14e,2) - 4./27.*m11*me2*RN1e*
         RS16e*RSe63 + 4./27.*m11*me2*RN1e*RS14e*RSe63 - 4./81.*m11*m66
         *RN1e*RS16e*RSe63 - 4./81.*m11*m66*RN1e*RS16e*RSe61 - 2./81.*
         m11*m66*RN1e*RS14e*RSe63 - 2./81.*m11*m66*RN1e*RS14e*RSe61 + 4.
         /81.*m11*m66*RN13*RS16e*RS34e + 4./81.*m11*m66*RN13*RS14e*
         RS36e + 4./81.*m11*m66*RN13*RS14e*RS34e - 2./81.*m11*m44*RN1e*
         RS16e*RSe63 - 2./81.*m11*m44*RN1e*RS16e*RSe61 + 2./81.*m11*m44
         *RN1e*RS14e*RSe63 + 2./81.*m11*m44*RN1e*RS14e*RSe61 + 4./81.*
         m11*m44*RN13*RS16e*RS36e + 4./81.*m11*m44*RN13*RS16e*RS34e + 4.
         /81.*m11*m44*RN13*RS14e*RS36e );

      M13p6x22R +=  + Ab(m11,mu2)*Ab(m46,mu2) * (  - 4./81.*m11*m33*
         RN13*RS16e*RS36e + 4./81.*m11*m33*RN13*RS16e*RS34e + 4./81.*
         m11*m33*RN13*RS14e*RS36e - 4./81.*m11*m33*RN13*RS14e*RS34e - 2.
         /81.*pow(m11,2)*RN13*RS16e*RS36e + 2./81.*pow(m11,2)*RN13*
         RS16e*RS34e + 2./81.*pow(m11,2)*RN13*RS14e*RS36e - 2./81.*pow(
         m11,2)*RN13*RS14e*RS34e );

      M13p6x22R +=  + Ab(m11,mu2)*Ab(me2,mu2) * ( 1./135.*RD146e*
         RSe6613 - 1./270.*RD146e*RSe6633 - 1./270.*RD146e*RSe6611 - 2./
         135.*me2*RSe6613*RS1463e + 1./135.*me2*RSe6633*RS1463e + 1./
         135.*me2*RSe6611*RS1463e + 1./135.*me2*RC146e*RSe6613 - 1./270.
         *me2*RC146e*RSe6633 - 1./270.*me2*RC146e*RSe6611 - 4./81.*m66*
         RS14e*RSe63 + 4./81.*m66*RS14e*RSe61 - 1./81.*m66*pow(RS14e,2)
          - 4./81.*m66*me2*RN1e*RS14e*RSe63 + 4./81.*m66*me2*RN1e*RS14e
         *RSe61 + 1./81.*m44*RS16e*RSe63 - 1./81.*m44*RS16e*RSe61 - 1./
         162.*m44*pow(RS16e,2) + 1./81.*m44*me2*RN1e*RS16e*RSe63 - 1./
         81.*m44*me2*RN1e*RS16e*RSe61 - 13./135.*m33*RSe6613*RS1463e + 
         7./270.*m33*RSe6633*RS1463e - 1./270.*m33*RSe6611*RS1463e - 1./
         270.*m33*RC146e*RSe6613 + 17./540.*m33*RC146e*RSe6633 - 1./36.
         *m33*RC146e*RSe6611 - 4./81.*m33*m66*RN1e*RS14e*RSe63 + 2./81.
         *m33*m66*RN13*RS14e*RS34e + 1./81.*m33*m44*RN1e*RS16e*RSe63 + 
         1./81.*m33*m44*RN13*RS16e*RS36e - 2./27.*pow(m33,2)*RN13*
         RSe6633*RS1463e );

      M13p6x22R +=  + Ab(m11,mu2)*Ab(me2,mu2) * (  - 1./9.*m11*RSe6613*
         RS1463e + 1./270.*m11*RSe6633*RS1463e - 11./270.*m11*RSe6611*
         RS1463e + 4./27.*m11*RSe6611*RSe6613 - 4./27.*m11*pow(
         RSe6611,2) + 1./30.*m11*RC146e*RSe6613 + 1./36.*m11*RC146e*
         RSe6633 - 13./540.*m11*RC146e*RSe6611 + 4./27.*m11*me2*RN1e*
         RSe6611*RSe6613 - 4./27.*m11*me2*RN1e*pow(RSe6611,2) + 2./27.*
         m11*me2*RC146e*RN1e*RSe6613 - 2./27.*m11*me2*RC146e*RN1e*
         RSe6611 - 4./81.*m11*m66*RN1e*RS14e*RSe63 - 4./81.*m11*m66*
         RN1e*RS14e*RSe61 + 4./81.*m11*m66*RN13*RS14e*RS34e + 1./81.*
         m11*m44*RN1e*RS16e*RSe63 + 1./81.*m11*m44*RN1e*RS16e*RSe61 + 2.
         /81.*m11*m44*RN13*RS16e*RS36e + 4./27.*m11*m33*RN1e*RSe6611*
         RSe6613 + 2./27.*m11*m33*RC146e*RN1e*RSe6613 - 4./27.*m11*m33*
         RN13*RSe6633*RS1463e - 2./27.*m11*m33*RN13*RSe6611*RS3461e + 4.
         /27.*pow(m11,2)*RN1e*RSe6611*RSe6613 + 4./27.*pow(m11,2)*RN1e*
         pow(RSe6611,2) + 2./27.*pow(m11,2)*RC146e*RN1e*RSe6613 + 2./27.
         *pow(m11,2)*RC146e*RN1e*RSe6611 );

      M13p6x22R +=  + Ab(m11,mu2)*Ab(me2,mu2) * (  - 4./27.*pow(m11,2)*
         RN13*RSe6611*RS3461e );

      M13p6x22R +=  + Ab(m11,mu2)*Bb(m11,mu2) * (  - 2./729.*pow(
         m11,-1)*m33*m44*pow(m66,7)*pow(RN1e,3)*pow(RN3e,3) - 1./81.*
         pow(m11,-1)*m33*pow(m44,2)*pow(m66,6)*pow(RN1e,3)*pow(RN3e,3)
          - 5./243.*pow(m11,-1)*m33*pow(m44,3)*pow(m66,5)*pow(RN1e,3)*
         pow(RN3e,3) - 25./1458.*pow(m11,-1)*m33*pow(m44,4)*pow(m66,4)*
         pow(RN1e,3)*pow(RN3e,3) - 5./648.*pow(m11,-1)*m33*pow(m44,5)*
         pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) - 7./3888.*pow(m11,-1)*m33*
         pow(m44,6)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) - 1./5832.*pow(
         m11,-1)*m33*pow(m44,7)*m66*pow(RN1e,3)*pow(RN3e,3) + 17./2187.
         *pow(m11,-1)*pow(m33,2)*m44*pow(m66,6)*pow(RN1e,3)*pow(RN3e,3)
          + 83./2187.*pow(m11,-1)*pow(m33,2)*pow(m44,2)*pow(m66,5)*pow(
         RN1e,3)*pow(RN3e,3) + 247./4374.*pow(m11,-1)*pow(m33,2)*pow(
         m44,3)*pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) + 82./2187.*pow(
         m11,-1)*pow(m33,2)*pow(m44,4)*pow(m66,3)*pow(RN1e,3)*pow(
         RN3e,3) + 409./34992.*pow(m11,-1)*pow(m33,2)*pow(m44,5)*pow(
         m66,2)*pow(RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + Ab(m11,mu2)*Bb(m11,mu2) * ( 49./34992.*pow(
         m11,-1)*pow(m33,2)*pow(m44,6)*m66*pow(RN1e,3)*pow(RN3e,3) + 1./
         486.*pow(m11,-1)*pow(m33,3)*m44*pow(m66,5)*pow(RN1e,3)*pow(
         RN3e,3) - 23./972.*pow(m11,-1)*pow(m33,3)*pow(m44,2)*pow(
         m66,4)*pow(RN1e,3)*pow(RN3e,3) - 25./648.*pow(m11,-1)*pow(
         m33,3)*pow(m44,3)*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) - 77./
         3888.*pow(m11,-1)*pow(m33,3)*pow(m44,4)*pow(m66,2)*pow(RN1e,3)
         *pow(RN3e,3) - 13./3888.*pow(m11,-1)*pow(m33,3)*pow(m44,5)*m66
         *pow(RN1e,3)*pow(RN3e,3) - 7./324.*pow(m11,-1)*pow(m33,4)*m44*
         pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) - 1./54.*pow(m11,-1)*pow(
         m33,4)*pow(m44,2)*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) - 1./432.
         *pow(m11,-1)*pow(m33,4)*pow(m44,3)*pow(m66,2)*pow(RN1e,3)*pow(
         RN3e,3) + 1./1296.*pow(m11,-1)*pow(m33,4)*pow(m44,4)*m66*pow(
         RN1e,3)*pow(RN3e,3) + 5./324.*pow(m11,-1)*pow(m33,5)*m44*pow(
         m66,3)*pow(RN1e,3)*pow(RN3e,3) + 5./324.*pow(m11,-1)*pow(
         m33,5)*pow(m44,2)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + Ab(m11,mu2)*Bb(m11,mu2) * ( 5./1296.*pow(m11,-1)*
         pow(m33,5)*pow(m44,3)*m66*pow(RN1e,3)*pow(RN3e,3) + 1./540.*
         pow(RD146e,2) - 2./729.*m44*pow(m66,7)*pow(RN1e,3)*pow(RN3e,3)
          - 1./81.*pow(m44,2)*pow(m66,6)*pow(RN1e,3)*pow(RN3e,3) - 5./
         243.*pow(m44,3)*pow(m66,5)*pow(RN1e,3)*pow(RN3e,3) - 25./1458.
         *pow(m44,4)*pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) - 5./648.*pow(
         m44,5)*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) - 7./3888.*pow(
         m44,6)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) - 1./5832.*pow(
         m44,7)*m66*pow(RN1e,3)*pow(RN3e,3) + 1./540.*m33*RD146e*
         RS1463e + 1./72.*m33*RD146e*RC146e + 2./729.*m33*pow(m66,7)*
         pow(RN1e,3)*pow(RN3e,3) + 40./2187.*m33*m44*pow(m66,6)*pow(
         RN1e,3)*pow(RN3e,3) + 157./2187.*m33*pow(m44,2)*pow(m66,5)*
         pow(RN1e,3)*pow(RN3e,3) + 235./2187.*m33*pow(m44,3)*pow(m66,4)
         *pow(RN1e,3)*pow(RN3e,3) + 1315./17496.*m33*pow(m44,4)*pow(
         m66,3)*pow(RN1e,3)*pow(RN3e,3) + 227./8748.*m33*pow(m44,5)*
         pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + Ab(m11,mu2)*Bb(m11,mu2) * ( 35./8748.*m33*pow(
         m44,6)*m66*pow(RN1e,3)*pow(RN3e,3) + 1./5832.*m33*pow(m44,7)*
         pow(RN1e,3)*pow(RN3e,3) - 17./2187.*pow(m33,2)*pow(m66,6)*pow(
         RN1e,3)*pow(RN3e,3) - 41./4374.*pow(m33,2)*m44*pow(m66,5)*pow(
         RN1e,3)*pow(RN3e,3) - 919./8748.*pow(m33,2)*pow(m44,2)*pow(
         m66,4)*pow(RN1e,3)*pow(RN3e,3) - 959./5832.*pow(m33,2)*pow(
         m44,3)*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) - 409./4374.*pow(
         m33,2)*pow(m44,4)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) - 743./
         34992.*pow(m33,2)*pow(m44,5)*m66*pow(RN1e,3)*pow(RN3e,3) - 49./
         34992.*pow(m33,2)*pow(m44,6)*pow(RN1e,3)*pow(RN3e,3) - 1./486.
         *pow(m33,3)*pow(m66,5)*pow(RN1e,3)*pow(RN3e,3) - 5./54.*pow(
         m33,3)*m44*pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) - 23./1944.*pow(
         m33,3)*pow(m44,2)*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) + 17./243.
         *pow(m33,3)*pow(m44,3)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) + 43.
         /1296.*pow(m33,3)*pow(m44,4)*m66*pow(RN1e,3)*pow(RN3e,3) + 13./
         3888.*pow(m33,3)*pow(m44,5)*pow(RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + Ab(m11,mu2)*Bb(m11,mu2) * ( 7./324.*pow(m33,4)*
         pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) + 53./324.*pow(m33,4)*m44*
         pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) + 133./1296.*pow(m33,4)*
         pow(m44,2)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) + 5./432.*pow(
         m33,4)*pow(m44,3)*m66*pow(RN1e,3)*pow(RN3e,3) - 1./1296.*pow(
         m33,4)*pow(m44,4)*pow(RN1e,3)*pow(RN3e,3) - 5./324.*pow(m33,5)
         *pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) - 25./324.*pow(m33,5)*m44*
         pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) - 55./1296.*pow(m33,5)*pow(
         m44,2)*m66*pow(RN1e,3)*pow(RN3e,3) - 5./1296.*pow(m33,5)*pow(
         m44,3)*pow(RN1e,3)*pow(RN3e,3) + 1./60.*m11*RD146e*RS1463e - 2.
         /27.*m11*RD146e*RSe6613 + 2./27.*m11*RD146e*RSe6611 - 5./216.*
         m11*RD146e*RC146e - 2./27.*m11*me2*RD146e*RN1e*RSe6613 + 2./27.
         *m11*me2*RD146e*RN1e*RSe6611 + 2./729.*m11*pow(m66,7)*pow(
         RN1e,3)*pow(RN3e,3) + 23./2187.*m11*m44*pow(m66,6)*pow(RN1e,3)
         *pow(RN3e,3) + 74./2187.*m11*pow(m44,2)*pow(m66,5)*pow(RN1e,3)
         *pow(RN3e,3) );

      M13p6x22R +=  + Ab(m11,mu2)*Bb(m11,mu2) * ( 223./4374.*m11*pow(
         m44,3)*pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) + 659./17496.*m11*
         pow(m44,4)*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) + 499./34992.*
         m11*pow(m44,5)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) + 91./34992.
         *m11*pow(m44,6)*m66*pow(RN1e,3)*pow(RN3e,3) + 1./5832.*m11*
         pow(m44,7)*pow(RN1e,3)*pow(RN3e,3) + 1./27.*m11*m33*RC146e*
         RS1463e - 2./27.*m11*m33*RD146e*RN1e*RSe6613 + 1./27.*m11*m33*
         RN13*RD146e*RS3461e - 13./2187.*m11*m33*pow(m66,6)*pow(RN1e,3)
         *pow(RN3e,3) + 53./4374.*m11*m33*m44*pow(m66,5)*pow(RN1e,3)*
         pow(RN3e,3) - 533./8748.*m11*m33*pow(m44,2)*pow(m66,4)*pow(
         RN1e,3)*pow(RN3e,3) - 785./5832.*m11*m33*pow(m44,3)*pow(m66,3)
         *pow(RN1e,3)*pow(RN3e,3) - 1571./17496.*m11*m33*pow(m44,4)*
         pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) - 847./34992.*m11*m33*pow(
         m44,5)*m66*pow(RN1e,3)*pow(RN3e,3) - 77./34992.*m11*m33*pow(
         m44,6)*pow(RN1e,3)*pow(RN3e,3) - 125./4374.*m11*pow(m33,2)*
         pow(m66,5)*pow(RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + Ab(m11,mu2)*Bb(m11,mu2) * (  - 1031./4374.*m11*
         pow(m33,2)*m44*pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) - 1913./
         17496.*m11*pow(m33,2)*pow(m44,2)*pow(m66,3)*pow(RN1e,3)*pow(
         RN3e,3) + 799./8748.*m11*pow(m33,2)*pow(m44,3)*pow(m66,2)*pow(
         RN1e,3)*pow(RN3e,3) + 1127./17496.*m11*pow(m33,2)*pow(m44,4)*
         m66*pow(RN1e,3)*pow(RN3e,3) + 167./17496.*m11*pow(m33,2)*pow(
         m44,5)*pow(RN1e,3)*pow(RN3e,3) + 113./972.*m11*pow(m33,3)*pow(
         m66,4)*pow(RN1e,3)*pow(RN3e,3) + 1043./1944.*m11*pow(m33,3)*
         m44*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) + 7./24.*m11*pow(m33,3)
         *pow(m44,2)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) - 1./1944.*m11*
         pow(m33,3)*pow(m44,3)*m66*pow(RN1e,3)*pow(RN3e,3) - 13./972.*
         m11*pow(m33,3)*pow(m44,4)*pow(RN1e,3)*pow(RN3e,3) - 47./324.*
         m11*pow(m33,4)*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) - 581./1296.
         *m11*pow(m33,4)*m44*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) - 227./
         1296.*m11*pow(m33,4)*pow(m44,2)*m66*pow(RN1e,3)*pow(RN3e,3) - 
         1./108.*m11*pow(m33,4)*pow(m44,3)*pow(RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + Ab(m11,mu2)*Bb(m11,mu2) * ( 5./81.*m11*pow(m33,5)
         *pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) + 155./1296.*m11*pow(
         m33,5)*m44*m66*pow(RN1e,3)*pow(RN3e,3) + 35./1296.*m11*pow(
         m33,5)*pow(m44,2)*pow(RN1e,3)*pow(RN3e,3) + 2./27.*pow(m11,2)*
         RC146e*RS1463e - 2./27.*pow(m11,2)*RD146e*RN1e*RSe6613 - 2./27.
         *pow(m11,2)*RD146e*RN1e*RSe6611 + 2./27.*pow(m11,2)*RN13*
         RD146e*RS3461e + 4./2187.*pow(m11,2)*pow(m66,6)*pow(RN1e,3)*
         pow(RN3e,3) + 103./4374.*pow(m11,2)*m44*pow(m66,5)*pow(RN1e,3)
         *pow(RN3e,3) + 179./8748.*pow(m11,2)*pow(m44,2)*pow(m66,4)*
         pow(RN1e,3)*pow(RN3e,3) - 17./1944.*pow(m11,2)*pow(m44,3)*pow(
         m66,3)*pow(RN1e,3)*pow(RN3e,3) - 563./34992.*pow(m11,2)*pow(
         m44,4)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) - 221./34992.*pow(
         m11,2)*pow(m44,5)*m66*pow(RN1e,3)*pow(RN3e,3) - 7./8748.*pow(
         m11,2)*pow(m44,6)*pow(RN1e,3)*pow(RN3e,3) - 277./4374.*pow(
         m11,2)*m33*pow(m66,5)*pow(RN1e,3)*pow(RN3e,3) - 1225./4374.*
         pow(m11,2)*m33*m44*pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + Ab(m11,mu2)*Bb(m11,mu2) * (  - 4123./17496.*pow(
         m11,2)*m33*pow(m44,2)*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) - 73./
         2187.*pow(m11,2)*m33*pow(m44,3)*pow(m66,2)*pow(RN1e,3)*pow(
         RN3e,3) + 809./34992.*pow(m11,2)*m33*pow(m44,4)*m66*pow(
         RN1e,3)*pow(RN3e,3) + 209./34992.*pow(m11,2)*m33*pow(m44,5)*
         pow(RN1e,3)*pow(RN3e,3) + 829./2916.*pow(m11,2)*pow(m33,2)*
         pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) + 5119./5832.*pow(m11,2)*
         pow(m33,2)*m44*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) + 1105./1944.
         *pow(m11,2)*pow(m33,2)*pow(m44,2)*pow(m66,2)*pow(RN1e,3)*pow(
         RN3e,3) + 487./5832.*pow(m11,2)*pow(m33,2)*pow(m44,3)*m66*pow(
         RN1e,3)*pow(RN3e,3) - 49./5832.*pow(m11,2)*pow(m33,2)*pow(
         m44,4)*pow(RN1e,3)*pow(RN3e,3) - 35./72.*pow(m11,2)*pow(m33,3)
         *pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) - 88./81.*pow(m11,2)*pow(
         m33,3)*m44*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) - 293./648.*pow(
         m11,2)*pow(m33,3)*pow(m44,2)*m66*pow(RN1e,3)*pow(RN3e,3) - 5./
         162.*pow(m11,2)*pow(m33,3)*pow(m44,3)*pow(RN1e,3)*pow(RN3e,3)
          );

      M13p6x22R +=  + Ab(m11,mu2)*Bb(m11,mu2) * ( 451./1296.*pow(m11,2)
         *pow(m33,4)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) + 673./1296.*
         pow(m11,2)*pow(m33,4)*m44*m66*pow(RN1e,3)*pow(RN3e,3) + 59./
         648.*pow(m11,2)*pow(m33,4)*pow(m44,2)*pow(RN1e,3)*pow(RN3e,3)
          - 35./432.*pow(m11,2)*pow(m33,5)*m66*pow(RN1e,3)*pow(RN3e,3)
          - 25./432.*pow(m11,2)*pow(m33,5)*m44*pow(RN1e,3)*pow(RN3e,3)
          - 161./4374.*pow(m11,3)*pow(m66,5)*pow(RN1e,3)*pow(RN3e,3) - 
         1009./8748.*pow(m11,3)*m44*pow(m66,4)*pow(RN1e,3)*pow(RN3e,3)
          - 2093./17496.*pow(m11,3)*pow(m44,2)*pow(m66,3)*pow(RN1e,3)*
         pow(RN3e,3) - 1835./34992.*pow(m11,3)*pow(m44,3)*pow(m66,2)*
         pow(RN1e,3)*pow(RN3e,3) - 311./34992.*pow(m11,3)*pow(m44,4)*
         m66*pow(RN1e,3)*pow(RN3e,3) - 1./4374.*pow(m11,3)*pow(m44,5)*
         pow(RN1e,3)*pow(RN3e,3) + 731./2916.*pow(m11,3)*m33*pow(m66,4)
         *pow(RN1e,3)*pow(RN3e,3) + 899./1458.*pow(m11,3)*m33*m44*pow(
         m66,3)*pow(RN1e,3)*pow(RN3e,3) + 1801./3888.*pow(m11,3)*m33*
         pow(m44,2)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + Ab(m11,mu2)*Bb(m11,mu2) * ( 1471./11664.*pow(
         m11,3)*m33*pow(m44,3)*m66*pow(RN1e,3)*pow(RN3e,3) + 101./11664.
         *pow(m11,3)*m33*pow(m44,4)*pow(RN1e,3)*pow(RN3e,3) - 1247./
         1944.*pow(m11,3)*pow(m33,2)*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3)
          - 1511./1296.*pow(m11,3)*pow(m33,2)*m44*pow(m66,2)*pow(
         RN1e,3)*pow(RN3e,3) - 721./1296.*pow(m11,3)*pow(m33,2)*pow(
         m44,2)*m66*pow(RN1e,3)*pow(RN3e,3) - 65./972.*pow(m11,3)*pow(
         m33,2)*pow(m44,3)*pow(RN1e,3)*pow(RN3e,3) + 965./1296.*pow(
         m11,3)*pow(m33,3)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) + 1157./
         1296.*pow(m11,3)*pow(m33,3)*m44*m66*pow(RN1e,3)*pow(RN3e,3) + 
         127./648.*pow(m11,3)*pow(m33,3)*pow(m44,2)*pow(RN1e,3)*pow(
         RN3e,3) - 77./216.*pow(m11,3)*pow(m33,4)*m66*pow(RN1e,3)*pow(
         RN3e,3) - 23./108.*pow(m11,3)*pow(m33,4)*m44*pow(RN1e,3)*pow(
         RN3e,3) + 5./144.*pow(m11,3)*pow(m33,5)*pow(RN1e,3)*pow(
         RN3e,3) + 89./1458.*pow(m11,4)*pow(m66,4)*pow(RN1e,3)*pow(
         RN3e,3) );

      M13p6x22R +=  + Ab(m11,mu2)*Bb(m11,mu2) * ( 371./2916.*pow(m11,4)
         *m44*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) + 193./1944.*pow(
         m11,4)*pow(m44,2)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) + 401./
         11664.*pow(m11,4)*pow(m44,3)*m66*pow(RN1e,3)*pow(RN3e,3) + 13./
         2916.*pow(m11,4)*pow(m44,4)*pow(RN1e,3)*pow(RN3e,3) - 305./972.
         *pow(m11,4)*m33*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) - 40./81.*
         pow(m11,4)*m33*m44*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) - 335./
         1296.*pow(m11,4)*m33*pow(m44,2)*m66*pow(RN1e,3)*pow(RN3e,3) - 
         175./3888.*pow(m11,4)*m33*pow(m44,3)*pow(RN1e,3)*pow(RN3e,3)
          + 127./216.*pow(m11,4)*pow(m33,2)*pow(m66,2)*pow(RN1e,3)*pow(
         RN3e,3) + 269./432.*pow(m11,4)*pow(m33,2)*m44*m66*pow(RN1e,3)*
         pow(RN3e,3) + 71./432.*pow(m11,4)*pow(m33,2)*pow(m44,2)*pow(
         RN1e,3)*pow(RN3e,3) - 203./432.*pow(m11,4)*pow(m33,3)*m66*pow(
         RN1e,3)*pow(RN3e,3) - 109./432.*pow(m11,4)*pow(m33,3)*m44*pow(
         RN1e,3)*pow(RN3e,3) + 19./144.*pow(m11,4)*pow(m33,4)*pow(
         RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + Ab(m11,mu2)*Bb(m11,mu2) * (  - 7./243.*pow(m11,5)
         *pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) - 7./162.*pow(m11,5)*m44*
         pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) - 7./324.*pow(m11,5)*pow(
         m44,2)*m66*pow(RN1e,3)*pow(RN3e,3) - 7./1944.*pow(m11,5)*pow(
         m44,3)*pow(RN1e,3)*pow(RN3e,3) + 7./54.*pow(m11,5)*m33*pow(
         m66,2)*pow(RN1e,3)*pow(RN3e,3) + 7./54.*pow(m11,5)*m33*m44*m66
         *pow(RN1e,3)*pow(RN3e,3) + 7./216.*pow(m11,5)*m33*pow(m44,2)*
         pow(RN1e,3)*pow(RN3e,3) - 7./36.*pow(m11,5)*pow(m33,2)*m66*
         pow(RN1e,3)*pow(RN3e,3) - 7./72.*pow(m11,5)*pow(m33,2)*m44*
         pow(RN1e,3)*pow(RN3e,3) + 7./72.*pow(m11,5)*pow(m33,3)*pow(
         RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + Ab(m11,mu2)*Bb(m33,mu2) * ( 1./540.*RD146e*RD346e
          - 1./60.*m33*RD346e*RS1463e - 1./72.*m33*RC146e*RD346e + 1./
         27.*pow(m33,2)*RN13*RD346e*RS1463e - 1./540.*m11*RD346e*
         RS1463e - 1./72.*m11*RC146e*RD346e + 2./27.*m11*m33*RN13*
         RD346e*RS1463e );

      M13p6x22R +=  + Ab(m13,mu2)*pi16 * ( 1./144.*pow(m33,2)*pow(
         RS3461e,2) + 1./72.*pow(m33,2)*RS1463e*RS3461e + 1./144.*pow(
         m33,2)*pow(RS1463e,2) - 1./36.*pow(m33,2)*RSe6613*RS3461e - 1./
         36.*pow(m33,2)*RSe6613*RS1463e + 1./36.*pow(m33,2)*pow(
         RSe6613,2) + 1./72.*pow(m33,2)*RSe6633*RS3461e + 1./72.*pow(
         m33,2)*RSe6633*RS1463e - 1./36.*pow(m33,2)*RSe6633*RSe6613 + 1.
         /144.*pow(m33,2)*pow(RSe6633,2) - 1./144.*pow(m33,2)*RC346e*
         RS3461e - 1./144.*pow(m33,2)*RC346e*RS1463e + 1./72.*pow(
         m33,2)*RC346e*RSe6613 - 1./144.*pow(m33,2)*RC346e*RSe6633 + 1./
         576.*pow(m33,2)*pow(RC346e,2) + 1./72.*pow(m33,2)*RSe6611*
         RS3461e + 1./72.*pow(m33,2)*RSe6611*RS1463e - 1./36.*pow(
         m33,2)*RSe6611*RSe6613 + 1./72.*pow(m33,2)*RSe6611*RSe6633 - 1.
         /144.*pow(m33,2)*RSe6611*RC346e + 1./144.*pow(m33,2)*pow(
         RSe6611,2) - 1./144.*pow(m33,2)*RC146e*RS3461e - 1./144.*pow(
         m33,2)*RC146e*RS1463e + 1./72.*pow(m33,2)*RC146e*RSe6613 - 1./
         144.*pow(m33,2)*RC146e*RSe6633 );

      M13p6x22R +=  + Ab(m13,mu2)*pi16 * ( 1./288.*pow(m33,2)*RC146e*
         RC346e - 1./144.*pow(m33,2)*RC146e*RSe6611 + 1./576.*pow(
         m33,2)*pow(RC146e,2) + 1./72.*m11*m33*pow(RS3461e,2) + 1./36.*
         m11*m33*RS1463e*RS3461e + 1./72.*m11*m33*pow(RS1463e,2) - 1./
         18.*m11*m33*RSe6613*RS3461e - 1./18.*m11*m33*RSe6613*RS1463e
          + 1./18.*m11*m33*pow(RSe6613,2) + 1./36.*m11*m33*RSe6633*
         RS3461e + 1./36.*m11*m33*RSe6633*RS1463e - 1./18.*m11*m33*
         RSe6633*RSe6613 + 1./72.*m11*m33*pow(RSe6633,2) - 1./72.*m11*
         m33*RC346e*RS3461e - 1./72.*m11*m33*RC346e*RS1463e + 1./36.*
         m11*m33*RC346e*RSe6613 - 1./72.*m11*m33*RC346e*RSe6633 + 1./
         288.*m11*m33*pow(RC346e,2) + 1./36.*m11*m33*RSe6611*RS3461e + 
         1./36.*m11*m33*RSe6611*RS1463e - 1./18.*m11*m33*RSe6611*
         RSe6613 + 1./36.*m11*m33*RSe6611*RSe6633 - 1./72.*m11*m33*
         RSe6611*RC346e + 1./72.*m11*m33*pow(RSe6611,2) - 1./72.*m11*
         m33*RC146e*RS3461e - 1./72.*m11*m33*RC146e*RS1463e + 1./36.*
         m11*m33*RC146e*RSe6613 );

      M13p6x22R +=  + Ab(m13,mu2)*pi16 * (  - 1./72.*m11*m33*RC146e*
         RSe6633 + 1./144.*m11*m33*RC146e*RC346e - 1./72.*m11*m33*
         RC146e*RSe6611 + 1./288.*m11*m33*pow(RC146e,2) + 1./144.*pow(
         m11,2)*pow(RS3461e,2) + 1./72.*pow(m11,2)*RS1463e*RS3461e + 1./
         144.*pow(m11,2)*pow(RS1463e,2) - 1./36.*pow(m11,2)*RSe6613*
         RS3461e - 1./36.*pow(m11,2)*RSe6613*RS1463e + 1./36.*pow(
         m11,2)*pow(RSe6613,2) + 1./72.*pow(m11,2)*RSe6633*RS3461e + 1./
         72.*pow(m11,2)*RSe6633*RS1463e - 1./36.*pow(m11,2)*RSe6633*
         RSe6613 + 1./144.*pow(m11,2)*pow(RSe6633,2) - 1./144.*pow(
         m11,2)*RC346e*RS3461e - 1./144.*pow(m11,2)*RC346e*RS1463e + 1./
         72.*pow(m11,2)*RC346e*RSe6613 - 1./144.*pow(m11,2)*RC346e*
         RSe6633 + 1./576.*pow(m11,2)*pow(RC346e,2) + 1./72.*pow(m11,2)
         *RSe6611*RS3461e + 1./72.*pow(m11,2)*RSe6611*RS1463e - 1./36.*
         pow(m11,2)*RSe6611*RSe6613 + 1./72.*pow(m11,2)*RSe6611*RSe6633
          - 1./144.*pow(m11,2)*RSe6611*RC346e + 1./144.*pow(m11,2)*pow(
         RSe6611,2) );

      M13p6x22R +=  + Ab(m13,mu2)*pi16 * (  - 1./144.*pow(m11,2)*RC146e
         *RS3461e - 1./144.*pow(m11,2)*RC146e*RS1463e + 1./72.*pow(
         m11,2)*RC146e*RSe6613 - 1./144.*pow(m11,2)*RC146e*RSe6633 + 1./
         288.*pow(m11,2)*RC146e*RC346e - 1./144.*pow(m11,2)*RC146e*
         RSe6611 + 1./576.*pow(m11,2)*pow(RC146e,2) );

      M13p6x22R +=  + pow(Ab(m13,mu2),2) * ( 1./648.*pow(m13,-1)*pow(
         RD346e,2) + 1./324.*pow(m13,-1)*RD146e*RD346e + 1./648.*pow(
         m13,-1)*pow(RD146e,2) + 1./81.*pow(m13,-1)*me2*RD346e*RSe6613
          - 1./162.*pow(m13,-1)*me2*RD346e*RSe6633 - 1./162.*pow(
         m13,-1)*me2*RSe6611*RD346e + 1./81.*pow(m13,-1)*me2*RD146e*
         RSe6613 - 1./162.*pow(m13,-1)*me2*RD146e*RSe6633 - 1./162.*
         pow(m13,-1)*me2*RD146e*RSe6611 - 1./27.*pow(m13,-1)*pow(me2,2)
         *RSe6613*RS3461e - 1./27.*pow(m13,-1)*pow(me2,2)*RSe6613*
         RS1463e + 8./81.*pow(m13,-1)*pow(me2,2)*pow(RSe6613,2) + 1./54.
         *pow(m13,-1)*pow(me2,2)*RSe6633*RS3461e + 1./54.*pow(m13,-1)*
         pow(me2,2)*RSe6633*RS1463e - 8./81.*pow(m13,-1)*pow(me2,2)*
         RSe6633*RSe6613 + 2./81.*pow(m13,-1)*pow(me2,2)*pow(RSe6633,2)
          + 1./54.*pow(m13,-1)*pow(me2,2)*RC346e*RSe6613 - 1./108.*pow(
         m13,-1)*pow(me2,2)*RC346e*RSe6633 + 1./54.*pow(m13,-1)*pow(
         me2,2)*RSe6611*RS3461e + 1./54.*pow(m13,-1)*pow(me2,2)*RSe6611
         *RS1463e );

      M13p6x22R +=  + pow(Ab(m13,mu2),2) * (  - 8./81.*pow(m13,-1)*pow(
         me2,2)*RSe6611*RSe6613 + 4./81.*pow(m13,-1)*pow(me2,2)*RSe6611
         *RSe6633 - 1./108.*pow(m13,-1)*pow(me2,2)*RSe6611*RC346e + 2./
         81.*pow(m13,-1)*pow(me2,2)*pow(RSe6611,2) + 1./54.*pow(m13,-1)
         *pow(me2,2)*RC146e*RSe6613 - 1./108.*pow(m13,-1)*pow(me2,2)*
         RC146e*RSe6633 - 1./108.*pow(m13,-1)*pow(me2,2)*RC146e*RSe6611
          - 1./108.*pow(m13,-1)*m33*RD346e*RS3461e - 1./324.*pow(
         m13,-1)*m33*RD346e*RS1463e + 1./162.*pow(m13,-1)*m33*RD346e*
         RSe6613 + 1./108.*pow(m13,-1)*m33*RD346e*RSe6633 - 1./648.*
         pow(m13,-1)*m33*RD346e*RC346e - 5./324.*pow(m13,-1)*m33*
         RSe6611*RD346e + 5./648.*pow(m13,-1)*m33*RC146e*RD346e + 1./
         324.*pow(m13,-1)*m33*RD146e*RS3461e + 1./108.*pow(m13,-1)*m33*
         RD146e*RS1463e - 1./54.*pow(m13,-1)*m33*RD146e*RSe6613 + 1./
         324.*pow(m13,-1)*m33*RD146e*RSe6633 + 1./648.*pow(m13,-1)*m33*
         RD146e*RC346e + 5./324.*pow(m13,-1)*m33*RD146e*RSe6611 - 5./
         648.*pow(m13,-1)*m33*RD146e*RC146e );

      M13p6x22R +=  + pow(Ab(m13,mu2),2) * ( 2./81.*pow(m13,-1)*m33*me2
         *RSe6613*RS3461e + 4./81.*pow(m13,-1)*m33*me2*RSe6613*RS1463e
          - 8./81.*pow(m13,-1)*m33*me2*pow(RSe6613,2) - 1./54.*pow(
         m13,-1)*m33*me2*RSe6633*RS3461e - 5./162.*pow(m13,-1)*m33*me2*
         RSe6633*RS1463e + 10./81.*pow(m13,-1)*m33*me2*RSe6633*RSe6613
          - 1./18.*pow(m13,-1)*m33*me2*pow(RSe6633,2) - 1./54.*pow(
         m13,-1)*m33*me2*RC346e*RSe6613 + 7./324.*pow(m13,-1)*m33*me2*
         RC346e*RSe6633 - 1./162.*pow(m13,-1)*m33*me2*RSe6611*RS3461e
          - 1./54.*pow(m13,-1)*m33*me2*RSe6611*RS1463e + 2./27.*pow(
         m13,-1)*m33*me2*RSe6611*RSe6613 - 1./81.*pow(m13,-1)*m33*me2*
         RSe6611*RSe6633 - 1./324.*pow(m13,-1)*m33*me2*RSe6611*RC346e
          - 5./162.*pow(m13,-1)*m33*me2*pow(RSe6611,2) - 1./54.*pow(
         m13,-1)*m33*me2*RC146e*RSe6613 + 1./324.*pow(m13,-1)*m33*me2*
         RC146e*RSe6633 + 5./324.*pow(m13,-1)*m33*me2*RC146e*RSe6611 - 
         31./144.*pow(m13,-1)*pow(m33,2) - 1./162.*pow(m13,-1)*pow(
         m33,2)*RS1463e*RS3461e );

      M13p6x22R +=  + pow(Ab(m13,mu2),2) * ( 1./81.*pow(m13,-1)*pow(
         m33,2)*RSe6613*RS3461e - 1./81.*pow(m13,-1)*pow(m33,2)*RSe6633
         *RS3461e + 1./162.*pow(m13,-1)*pow(m33,2)*RC346e*RS3461e + 1./
         162.*pow(m13,-1)*pow(m33,2)*RC346e*RS1463e - 1./81.*pow(
         m13,-1)*pow(m33,2)*RC346e*RSe6613 + 1./54.*pow(m13,-1)*pow(
         m33,2)*RC346e*RSe6633 - 5./648.*pow(m13,-1)*pow(m33,2)*pow(
         RC346e,2) - 1./162.*pow(m13,-1)*pow(m33,2)*RSe6611*RC346e + 1./
         324.*pow(m13,-1)*pow(m33,2)*RC146e*RC346e - 1./8.*m33 + 1./108.
         *m11*pow(m13,-1)*RD346e*RS3461e + 1./324.*m11*pow(m13,-1)*
         RD346e*RS1463e - 1./54.*m11*pow(m13,-1)*RD346e*RSe6613 + 5./
         324.*m11*pow(m13,-1)*RD346e*RSe6633 - 5./648.*m11*pow(m13,-1)*
         RD346e*RC346e + 1./324.*m11*pow(m13,-1)*RSe6611*RD346e + 1./
         648.*m11*pow(m13,-1)*RC146e*RD346e - 1./324.*m11*pow(m13,-1)*
         RD146e*RS3461e - 1./108.*m11*pow(m13,-1)*RD146e*RS1463e + 1./
         162.*m11*pow(m13,-1)*RD146e*RSe6613 - 5./324.*m11*pow(m13,-1)*
         RD146e*RSe6633 );

      M13p6x22R +=  + pow(Ab(m13,mu2),2) * ( 5./648.*m11*pow(m13,-1)*
         RD146e*RC346e + 1./108.*m11*pow(m13,-1)*RD146e*RSe6611 - 1./
         648.*m11*pow(m13,-1)*RD146e*RC146e + 4./81.*m11*pow(m13,-1)*
         me2*RSe6613*RS3461e + 2./81.*m11*pow(m13,-1)*me2*RSe6613*
         RS1463e - 8./81.*m11*pow(m13,-1)*me2*pow(RSe6613,2) - 1./54.*
         m11*pow(m13,-1)*me2*RSe6633*RS3461e - 1./162.*m11*pow(m13,-1)*
         me2*RSe6633*RS1463e + 2./27.*m11*pow(m13,-1)*me2*RSe6633*
         RSe6613 - 5./162.*m11*pow(m13,-1)*me2*pow(RSe6633,2) - 1./54.*
         m11*pow(m13,-1)*me2*RC346e*RSe6613 + 5./324.*m11*pow(m13,-1)*
         me2*RC346e*RSe6633 - 5./162.*m11*pow(m13,-1)*me2*RSe6611*
         RS3461e - 1./54.*m11*pow(m13,-1)*me2*RSe6611*RS1463e + 10./81.
         *m11*pow(m13,-1)*me2*RSe6611*RSe6613 - 1./81.*m11*pow(m13,-1)*
         me2*RSe6611*RSe6633 + 1./324.*m11*pow(m13,-1)*me2*RSe6611*
         RC346e - 1./18.*m11*pow(m13,-1)*me2*pow(RSe6611,2) - 1./54.*
         m11*pow(m13,-1)*me2*RC146e*RSe6613 - 1./324.*m11*pow(m13,-1)*
         me2*RC146e*RSe6633 );

      M13p6x22R +=  + pow(Ab(m13,mu2),2) * ( 7./324.*m11*pow(m13,-1)*
         me2*RC146e*RSe6611 - 31./72.*m11*pow(m13,-1)*m33 - 2./81.*m11*
         pow(m13,-1)*m33*pow(RS3461e,2) - 1./27.*m11*pow(m13,-1)*m33*
         RS1463e*RS3461e - 2./81.*m11*pow(m13,-1)*m33*pow(RS1463e,2) + 
         4./81.*m11*pow(m13,-1)*m33*RSe6613*RS3461e + 4./81.*m11*pow(
         m13,-1)*m33*RSe6613*RS1463e - 1./54.*m11*pow(m13,-1)*m33*
         RSe6633*RS3461e - 5./162.*m11*pow(m13,-1)*m33*RSe6633*RS1463e
          + 1./54.*m11*pow(m13,-1)*m33*RC346e*RS3461e + 1./54.*m11*pow(
         m13,-1)*m33*RC346e*RS1463e - 1./54.*m11*pow(m13,-1)*m33*RC346e
         *RSe6613 + 5./324.*m11*pow(m13,-1)*m33*RC346e*RSe6633 - 5./648.
         *m11*pow(m13,-1)*m33*pow(RC346e,2) - 5./162.*m11*pow(m13,-1)*
         m33*RSe6611*RS3461e - 1./54.*m11*pow(m13,-1)*m33*RSe6611*
         RS1463e + 1./324.*m11*pow(m13,-1)*m33*RSe6611*RC346e + 1./54.*
         m11*pow(m13,-1)*m33*RC146e*RS3461e + 1./54.*m11*pow(m13,-1)*
         m33*RC146e*RS1463e - 1./54.*m11*pow(m13,-1)*m33*RC146e*RSe6613
          + 1./324.*m11*pow(m13,-1)*m33*RC146e*RSe6633 );

      M13p6x22R +=  + pow(Ab(m13,mu2),2) * ( 5./324.*m11*pow(m13,-1)*
         m33*RC146e*RSe6611 - 5./648.*m11*pow(m13,-1)*m33*pow(RC146e,2)
          - 1./8.*m11 - 31./144.*pow(m11,2)*pow(m13,-1) - 1./162.*pow(
         m11,2)*pow(m13,-1)*RS1463e*RS3461e + 1./81.*pow(m11,2)*pow(
         m13,-1)*RSe6613*RS1463e - 1./81.*pow(m11,2)*pow(m13,-1)*
         RSe6611*RS1463e + 1./162.*pow(m11,2)*pow(m13,-1)*RC146e*
         RS3461e + 1./162.*pow(m11,2)*pow(m13,-1)*RC146e*RS1463e - 1./
         81.*pow(m11,2)*pow(m13,-1)*RC146e*RSe6613 - 1./162.*pow(m11,2)
         *pow(m13,-1)*RC146e*RSe6633 + 1./324.*pow(m11,2)*pow(m13,-1)*
         RC146e*RC346e + 1./54.*pow(m11,2)*pow(m13,-1)*RC146e*RSe6611
          - 5./648.*pow(m11,2)*pow(m13,-1)*pow(RC146e,2) );

      M13p6x22R +=  + Ab(m13,mu2)*Ab(m33,mu2) * (  - 2./81.*RD346e*
         RS3461e - 1./54.*RD346e*RS1463e + 1./27.*RD346e*RSe6613 - 1./
         54.*RD346e*RSe6633 + 1./81.*RD346e*RC346e - 1./54.*RSe6611*
         RD346e + 1./108.*RC146e*RD346e - 1./162.*RD146e*RS3461e + 1./
         324.*RD146e*RC346e - 2./81.*me2*RSe6613*RS3461e + 1./81.*me2*
         RSe6633*RS3461e + 1./81.*me2*RC346e*RSe6613 - 1./162.*me2*
         RC346e*RSe6633 + 1./81.*me2*RSe6611*RS3461e - 1./162.*me2*
         RSe6611*RC346e + 19./72.*m33 - 1./81.*m33*RS1463e*RS3461e + 2./
         81.*m33*RSe6613*RS3461e - 2./81.*m33*RSe6633*RS3461e + 1./81.*
         m33*RC346e*RS3461e + 1./81.*m33*RC346e*RS1463e - 2./81.*m33*
         RC346e*RSe6613 + 1./27.*m33*RC346e*RSe6633 - 5./324.*m33*pow(
         RC346e,2) - 1./81.*m33*RSe6611*RC346e + 1./162.*m33*RC146e*
         RC346e - 1./9.*pow(m33,2)*RN13 + 5./24.*m11 - 4./81.*m11*pow(
         RS3461e,2) - 1./27.*m11*RS1463e*RS3461e + 8./81.*m11*RSe6613*
         RS3461e - 1./27.*m11*RSe6633*RS3461e + 1./27.*m11*RC346e*
         RS3461e );

      M13p6x22R +=  + Ab(m13,mu2)*Ab(m33,mu2) * ( 1./81.*m11*RC346e*
         RS1463e - 1./27.*m11*RC346e*RSe6613 + 5./162.*m11*RC346e*
         RSe6633 - 5./324.*m11*pow(RC346e,2) - 5./81.*m11*RSe6611*
         RS3461e + 1./162.*m11*RSe6611*RC346e + 2./81.*m11*RC146e*
         RS3461e - 5./9.*m11*m33*RN13 - 1./3.*pow(m11,2)*RN13 );

      M13p6x22R +=  + Ab(m13,mu2)*Ab(me2,mu2) * ( 1./81.*RD346e*RSe6613
          - 1./162.*RD346e*RSe6633 - 1./162.*RSe6611*RD346e + 1./81.*
         RD146e*RSe6613 - 1./162.*RD146e*RSe6633 - 1./162.*RD146e*
         RSe6611 - 2./27.*me2*RSe6613*RS3461e - 2./27.*me2*RSe6613*
         RS1463e + 16./81.*me2*pow(RSe6613,2) + 1./27.*me2*RSe6633*
         RS3461e + 1./27.*me2*RSe6633*RS1463e - 16./81.*me2*RSe6633*
         RSe6613 + 4./81.*me2*pow(RSe6633,2) + 1./27.*me2*RC346e*
         RSe6613 - 1./54.*me2*RC346e*RSe6633 + 1./27.*me2*RSe6611*
         RS3461e + 1./27.*me2*RSe6611*RS1463e - 16./81.*me2*RSe6611*
         RSe6613 + 8./81.*me2*RSe6611*RSe6633 - 1./54.*me2*RSe6611*
         RC346e + 4./81.*me2*pow(RSe6611,2) + 1./27.*me2*RC146e*RSe6613
          - 1./54.*me2*RC146e*RSe6633 - 1./54.*me2*RC146e*RSe6611 + 2./
         27.*m33*RSe6613*RS3461e + 8./81.*m33*RSe6613*RS1463e - 16./81.
         *m33*pow(RSe6613,2) - 4./81.*m33*RSe6633*RS3461e - 5./81.*m33*
         RSe6633*RS1463e + 20./81.*m33*RSe6633*RSe6613 - 1./9.*m33*pow(
         RSe6633,2) );

      M13p6x22R +=  + Ab(m13,mu2)*Ab(me2,mu2) * (  - 4./81.*m33*RC346e*
         RSe6613 + 4./81.*m33*RC346e*RSe6633 - 2./81.*m33*RSe6611*
         RS3461e - 1./27.*m33*RSe6611*RS1463e + 4./27.*m33*RSe6611*
         RSe6613 - 2./81.*m33*RSe6611*RSe6633 - 5./81.*m33*pow(
         RSe6611,2) - 1./27.*m33*RC146e*RSe6613 + 1./162.*m33*RC146e*
         RSe6633 + 5./162.*m33*RC146e*RSe6611 + 8./81.*m11*RSe6613*
         RS3461e + 2./27.*m11*RSe6613*RS1463e - 16./81.*m11*pow(
         RSe6613,2) - 1./27.*m11*RSe6633*RS3461e - 2./81.*m11*RSe6633*
         RS1463e + 4./27.*m11*RSe6633*RSe6613 - 5./81.*m11*pow(
         RSe6633,2) - 1./27.*m11*RC346e*RSe6613 + 5./162.*m11*RC346e*
         RSe6633 - 5./81.*m11*RSe6611*RS3461e - 4./81.*m11*RSe6611*
         RS1463e + 20./81.*m11*RSe6611*RSe6613 - 2./81.*m11*RSe6611*
         RSe6633 + 1./162.*m11*RSe6611*RC346e - 1./9.*m11*pow(
         RSe6611,2) - 4./81.*m11*RC146e*RSe6613 + 4./81.*m11*RC146e*
         RSe6611 );

      M13p6x22R +=  + Ab(m13,mu2)*Bb(m11,mu2) * ( 1./324.*RD146e*RD346e
          + 1./324.*pow(RD146e,2) + 1./81.*me2*RD146e*RSe6613 - 1./162.
         *me2*RD146e*RSe6633 - 1./162.*me2*RD146e*RSe6611 + 1./81.*m33*
         RD146e*RS3461e + 1./54.*m33*RD146e*RS1463e - 1./27.*m33*RD146e
         *RSe6613 + 1./162.*m33*RD146e*RSe6633 + 5./162.*m33*RD146e*
         RSe6611 - 5./324.*m33*RD146e*RC146e + 1./81.*m11*RD146e*
         RS3461e + 1./162.*m11*RD146e*RS1463e - 2./81.*m11*RD146e*
         RSe6613 - 1./81.*m11*RD146e*RSe6633 + 1./162.*m11*RD146e*
         RC346e + 1./27.*m11*RD146e*RSe6611 - 5./324.*m11*RD146e*RC146e
          );

      M13p6x22R +=  + Ab(m13,mu2)*Bb(m33,mu2) * ( 1./324.*pow(RD346e,2)
          + 1./324.*RD146e*RD346e + 1./81.*me2*RD346e*RSe6613 - 1./162.
         *me2*RD346e*RSe6633 - 1./162.*me2*RSe6611*RD346e + 1./162.*m33
         *RD346e*RS3461e + 1./81.*m33*RD346e*RS1463e - 2./81.*m33*
         RD346e*RSe6613 + 1./27.*m33*RD346e*RSe6633 - 5./324.*m33*
         RD346e*RC346e - 1./81.*m33*RSe6611*RD346e + 1./162.*m33*RC146e
         *RD346e + 1./54.*m11*RD346e*RS3461e + 1./81.*m11*RD346e*
         RS1463e - 1./27.*m11*RD346e*RSe6613 + 5./162.*m11*RD346e*
         RSe6633 - 5./324.*m11*RD346e*RC346e + 1./162.*m11*RSe6611*
         RD346e );

      M13p6x22R +=  + Ab(m14,mu2)*pi16 * (  - 1./24.*m33*RD346e + 1./48.
         *m33*me2 + 1./12.*m33*me2*RSe63 + 1./12.*m33*me2*RSe6633 + 1./
         8.*m33*m66 + 7./24.*m33*m44 + 1./48.*m33*m44*RS3461e + 1./48.*
         m33*m44*RS1463e - 1./24.*m33*m44*RSe6613 + 1./48.*m33*m44*
         RSe63 + 1./24.*m33*m44*RSe6633 + 1./48.*m33*m44*RS36e - 1./48.
         *m33*m44*RC346e + 1./48.*m33*m44*RSe61 + 1./24.*m33*m44*
         RSe6611 + 1./48.*m33*m44*RS16e - 1./48.*m33*m44*RC146e + 1./9.
         *pow(m33,2) + 1./36.*pow(m33,2)*RSe63 + 1./36.*pow(m33,2)*
         RSe6633 + 1./9.*pow(m33,2)*RS36e - 1./18.*pow(m33,2)*RC346e - 
         1./24.*m11*RD346e + 1./48.*m11*me2 + 1./12.*m11*me2*RSe63 + 1./
         12.*m11*me2*RSe6633 + 1./8.*m11*m66 + 7./24.*m11*m44 + 1./48.*
         m11*m44*RS3461e + 1./48.*m11*m44*RS1463e - 1./24.*m11*m44*
         RSe6613 + 1./48.*m11*m44*RSe63 + 1./24.*m11*m44*RSe6633 + 1./
         48.*m11*m44*RS36e - 1./48.*m11*m44*RC346e + 1./48.*m11*m44*
         RSe61 + 1./24.*m11*m44*RSe6611 + 1./48.*m11*m44*RS16e - 1./48.
         *m11*m44*RC146e );

      M13p6x22R +=  + Ab(m14,mu2)*pi16 * (  - 1./144.*m11*m33 + 1./48.*
         m11*m33*RS3461e + 1./48.*m11*m33*RS1463e - 1./24.*m11*m33*
         RSe6613 - 1./144.*m11*m33*RSe63 + 1./72.*m11*m33*RSe6633 + 11./
         144.*m11*m33*RS36e - 7./144.*m11*m33*RC346e + 1./48.*m11*m33*
         RSe61 + 1./24.*m11*m33*RSe6611 + 1./48.*m11*m33*RS16e - 1./48.
         *m11*m33*RC146e - 17./144.*pow(m11,2) + 1./48.*pow(m11,2)*
         RS3461e + 1./48.*pow(m11,2)*RS1463e - 1./24.*pow(m11,2)*
         RSe6613 - 5./144.*pow(m11,2)*RSe63 - 1./72.*pow(m11,2)*RSe6633
          - 5./144.*pow(m11,2)*RS36e + 1./144.*pow(m11,2)*RC346e + 1./
         48.*pow(m11,2)*RSe61 + 1./24.*pow(m11,2)*RSe6611 + 1./48.*pow(
         m11,2)*RS16e - 1./48.*pow(m11,2)*RC146e );

      M13p6x22R +=  + pow(Ab(m14,mu2),2) * (  - 1./90.*pow(m14,-1)*pow(
         me2,2) - 17./135.*pow(m14,-1)*pow(me2,2)*RSe6613 - 19./270.*
         pow(m14,-1)*pow(me2,2)*RSe63 - 1./135.*pow(m14,-1)*pow(me2,2)*
         RSe6633 + 7./270.*pow(m14,-1)*pow(me2,2)*RSe61 + 4./45.*pow(
         m14,-1)*pow(me2,2)*RSe6611 - 1./216.*pow(m14,-1)*m44*RD346e - 
         1./216.*pow(m14,-1)*m44*RD146e + 11./540.*pow(m14,-1)*m44*me2
          + 1./54.*pow(m14,-1)*m44*me2*RSe6613 + 2./45.*pow(m14,-1)*m44
         *me2*RSe63 + 1./108.*pow(m14,-1)*m44*me2*RSe6633 + 2./135.*
         pow(m14,-1)*m44*me2*RSe61 + 1./108.*pow(m14,-1)*m44*me2*
         RSe6611 - 1./108.*pow(m14,-1)*pow(m44,2) - 1./108.*pow(m14,-1)
         *pow(m44,2)*RS3461e - 1./108.*pow(m14,-1)*pow(m44,2)*RS1463e
          + 1./54.*pow(m14,-1)*pow(m44,2)*RSe6613 + 1./108.*pow(m14,-1)
         *pow(m44,2)*RSe6633 - 1./216.*pow(m14,-1)*pow(m44,2)*RC346e + 
         1./108.*pow(m14,-1)*pow(m44,2)*RSe6611 - 1./216.*pow(m14,-1)*
         pow(m44,2)*RC146e + 11./270.*pow(m14,-1)*m33*RD346e - 1./216.*
         pow(m14,-1)*m33*RD146e );

      M13p6x22R +=  + pow(Ab(m14,mu2),2) * ( 1./108.*pow(m14,-1)*m33*
         me2 - 11./540.*pow(m14,-1)*m33*me2*RSe6613 - 1./40.*pow(
         m14,-1)*m33*me2*RSe63 - 1./15.*pow(m14,-1)*m33*me2*RSe6633 - 5.
         /216.*pow(m14,-1)*m33*me2*RSe61 + 1./108.*pow(m14,-1)*m33*me2*
         RSe6611 - 31./432.*pow(m14,-1)*m33*m66 - 17./108.*pow(m14,-1)*
         m33*m44 - 17./432.*pow(m14,-1)*m33*m44*RS3461e - 13./432.*pow(
         m14,-1)*m33*m44*RS1463e + 13./216.*pow(m14,-1)*m33*m44*RSe6613
          - 19./432.*pow(m14,-1)*m33*m44*RSe63 + 1./2160.*pow(m14,-1)*
         m33*m44*RS36e - 1./216.*pow(m14,-1)*m33*m44*RC346e + 5./432.*
         pow(m14,-1)*m33*m44*RSe61 + 1./54.*pow(m14,-1)*m33*m44*RSe6611
          + 5./432.*pow(m14,-1)*m33*m44*RS16e - 1./108.*pow(m14,-1)*m33
         *m44*RC146e - 31./216.*pow(m14,-1)*pow(m33,2) + 113./2160.*
         pow(m14,-1)*pow(m33,2)*RS3461e - 1./48.*pow(m14,-1)*pow(m33,2)
         *RS1463e + 1./24.*pow(m14,-1)*pow(m33,2)*RSe6613 - 1./144.*
         pow(m14,-1)*pow(m33,2)*RSe63 - 1./108.*pow(m14,-1)*pow(m33,2)*
         RSe6633 );

      M13p6x22R +=  + pow(Ab(m14,mu2),2) * (  - 221./2160.*pow(m14,-1)*
         pow(m33,2)*RS36e + 1./24.*pow(m14,-1)*pow(m33,2)*RC346e + 5./
         432.*pow(m14,-1)*pow(m33,2)*RSe61 + 1./108.*pow(m14,-1)*pow(
         m33,2)*RSe6611 + 5./432.*pow(m14,-1)*pow(m33,2)*RS16e - 1./216.
         *pow(m14,-1)*pow(m33,2)*RC146e + 1./24.*m11*pow(m14,-1)*RD346e
          - 2./45.*m11*pow(m14,-1)*RD146e - 1./540.*m11*pow(m14,-1)*me2
         *RSe6613 - 23./216.*m11*pow(m14,-1)*me2*RSe63 - 1./12.*m11*
         pow(m14,-1)*me2*RSe6633 - 1./120.*m11*pow(m14,-1)*me2*RSe61 - 
         4./45.*m11*pow(m14,-1)*me2*RSe6611 - 31./432.*m11*pow(m14,-1)*
         m66 - 4./27.*m11*pow(m14,-1)*m44 - 1./48.*m11*pow(m14,-1)*m44*
         RS3461e - 13./432.*m11*pow(m14,-1)*m44*RS1463e + 1./24.*m11*
         pow(m14,-1)*m44*RSe6613 - 1./16.*m11*pow(m14,-1)*m44*RSe63 + 1.
         /108.*m11*pow(m14,-1)*m44*RSe6633 - 1./16.*m11*pow(m14,-1)*m44
         *RS36e - 1./216.*m11*pow(m14,-1)*m44*RC346e - 1./144.*m11*pow(
         m14,-1)*m44*RSe61 - 1./108.*m11*pow(m14,-1)*m44*RSe6611 + 17./
         2160.*m11*pow(m14,-1)*m44*RS16e );

      M13p6x22R +=  + pow(Ab(m14,mu2),2) * (  - 59./432.*m11*pow(
         m14,-1)*m33 - 43./2160.*m11*pow(m14,-1)*m33*RS3461e - 23./2160.
         *m11*pow(m14,-1)*m33*RS1463e + 1./24.*m11*pow(m14,-1)*m33*
         RSe6613 - 1./16.*m11*pow(m14,-1)*m33*RSe63 + 1./108.*m11*pow(
         m14,-1)*m33*RSe6633 - 73./432.*m11*pow(m14,-1)*m33*RS36e + 1./
         27.*m11*pow(m14,-1)*m33*RC346e - 1./144.*m11*pow(m14,-1)*m33*
         RSe61 - 1./108.*m11*pow(m14,-1)*m33*RSe6611 - 13./432.*m11*
         pow(m14,-1)*m33*RS16e + 1./144.*pow(m11,2)*pow(m14,-1) + 23./
         360.*pow(m11,2)*pow(m14,-1)*RS1463e - 4./27.*pow(m11,2)*pow(
         m14,-1)*RSe63 - 4./27.*pow(m11,2)*pow(m14,-1)*RS36e + 19./1080.
         *pow(m11,2)*pow(m14,-1)*RS16e );

      M13p6x22R +=  + Ab(m14,mu2)*Ab(m33,mu2) * (  - 1./90.*RD346e + 7./
         135.*m44*RS36e + 43./180.*m33*RS3461e - 41./180.*m33*RS36e + 1.
         /12.*m33*RC346e + 2./27.*m33*m44*RN13*RS36e - 19./540.*m11*
         RS3461e - 5./36.*m11*RS36e + 1./12.*m11*RC346e + 4./27.*m11*
         m44*RN13*RS36e - 2./27.*m11*m33*RN13*RS3461e + 2./9.*m11*m33*
         RN13*RS36e + 2./27.*pow(m11,2)*RN13*RS3461e + 2./9.*pow(m11,2)
         *RN13*RS36e );

      M13p6x22R +=  + Ab(m14,mu2)*Ab(m34,mu2) * (  - 1./54.*RD346e - 1./
         54.*RD146e + 1./27.*me2 + 2./27.*me2*RSe6613 + 2./27.*me2*
         RSe63 + 1./27.*me2*RSe6633 + 2./27.*me2*RSe61 + 1./27.*me2*
         RSe6611 - 1./27.*m44 - 1./27.*m44*RS3461e - 1./27.*m44*RS1463e
          + 2./27.*m44*RSe6613 + 1./27.*m44*RSe6633 - 1./54.*m44*RC346e
          + 1./27.*m44*RSe6611 - 1./54.*m44*RC146e - 47./108.*m33 - 13./
         108.*m33*RS3461e - 1./12.*m33*RS1463e + 1./6.*m33*RSe6613 - 1./
         36.*m33*RSe63 - 1./27.*m33*RSe6633 + 5./108.*m33*RS36e + 5./
         108.*m33*RSe61 + 1./27.*m33*RSe6611 + 5./108.*m33*RS16e - 1./
         54.*m33*RC146e - 47./108.*m11 - 1./12.*m11*RS3461e - 13./108.*
         m11*RS1463e + 1./6.*m11*RSe6613 + 5./108.*m11*RSe63 + 1./27.*
         m11*RSe6633 + 5./108.*m11*RS36e - 1./54.*m11*RC346e - 1./36.*
         m11*RSe61 - 1./27.*m11*RSe6611 + 5./108.*m11*RS16e );

      M13p6x22R +=  + Ab(m14,mu2)*Ab(m36,mu2) * (  - 1./4.*m33 - 1./4.*
         m11 );

      M13p6x22R +=  + Ab(m14,mu2)*Ab(me2,mu2) * (  - 1./45.*me2 - 44./
         135.*me2*RSe6613 - 19./135.*me2*RSe63 + 1./45.*me2*RSe6633 + 7.
         /135.*me2*RSe61 + 29./135.*me2*RSe6611 + 1./45.*m44 + 17./135.
         *m44*RSe63 - 11./135.*m44*RSe61 + 2./27.*m44*me2*RN1e*RSe63 - 
         2./27.*m44*me2*RN1e*RSe61 - 1./270.*m33*RSe6613 - 7./540.*m33*
         RSe63 - 17./90.*m33*RSe6633 - 1./12.*m33*RSe61 + 4./27.*m33*
         me2*RN1e*RSe6613 + 2./27.*m33*me2*RN1e*RSe63 + 2./27.*m33*m44*
         RN1e*RSe63 + 19./270.*m11*RSe6613 + 1./108.*m11*RSe63 - 1./6.*
         m11*RSe6633 - 49./540.*m11*RSe61 - 1./15.*m11*RSe6611 + 4./27.
         *m11*me2*RN1e*RSe6613 + 2./9.*m11*me2*RN1e*RSe63 - 2./27.*m11*
         me2*RN1e*RSe61 + 4./27.*m11*me2*RN1e*RSe6611 + 2./27.*m11*m44*
         RN1e*RSe63 + 2./27.*m11*m44*RN1e*RSe61 - 4./27.*m11*m33*RN1e*
         RSe6613 + 2./27.*m11*m33*RN1e*RSe63 - 4./27.*pow(m11,2)*RN1e*
         RSe6613 + 2./27.*pow(m11,2)*RN1e*RSe63 + 2./27.*pow(m11,2)*
         RN1e*RSe61 - 4./27.*pow(m11,2)*RN1e*RSe6611 );

      M13p6x22R +=  + Ab(m14,mu2)*Bb(m33,mu2) * ( 1./12.*m33*RD346e + 1.
         /12.*m11*RD346e );

      M13p6x22R +=  + Ab(m16,mu2)*pi16 * (  - 1./48.*m33*RD346e + 1./24.
         *m33*me2 - 1./12.*m33*me2*RSe63 + 1./24.*m33*me2*RSe6633 + 5./
         96.*m33*m66 + 1./96.*m33*m66*RS3461e + 1./96.*m33*m66*RS1463e
          - 1./48.*m33*m66*RSe6613 - 1./48.*m33*m66*RSe63 + 1./48.*m33*
         m66*RSe6633 + 1./96.*m33*m66*RS34e - 1./96.*m33*m66*RC346e - 1.
         /48.*m33*m66*RSe61 + 1./48.*m33*m66*RSe6611 + 1./96.*m33*m66*
         RS14e - 1./96.*m33*m66*RC146e + 1./8.*m33*m44 + 1./18.*pow(
         m33,2) - 1./36.*pow(m33,2)*RSe63 + 1./72.*pow(m33,2)*RSe6633
          + 1./18.*pow(m33,2)*RS34e - 1./36.*pow(m33,2)*RC346e - 1./48.
         *m11*RD346e + 1./24.*m11*me2 - 1./12.*m11*me2*RSe63 + 1./24.*
         m11*me2*RSe6633 + 5./96.*m11*m66 + 1./96.*m11*m66*RS3461e + 1./
         96.*m11*m66*RS1463e - 1./48.*m11*m66*RSe6613 - 1./48.*m11*m66*
         RSe63 + 1./48.*m11*m66*RSe6633 + 1./96.*m11*m66*RS34e - 1./96.
         *m11*m66*RC346e - 1./48.*m11*m66*RSe61 + 1./48.*m11*m66*
         RSe6611 + 1./96.*m11*m66*RS14e - 1./96.*m11*m66*RC146e + 1./8.
         *m11*m44 );

      M13p6x22R +=  + Ab(m16,mu2)*pi16 * (  - 1./288.*m11*m33 + 1./96.*
         m11*m33*RS3461e + 1./96.*m11*m33*RS1463e - 1./48.*m11*m33*
         RSe6613 + 1./144.*m11*m33*RSe63 + 1./144.*m11*m33*RSe6633 + 11.
         /288.*m11*m33*RS34e - 7./288.*m11*m33*RC346e - 1./48.*m11*m33*
         RSe61 + 1./48.*m11*m33*RSe6611 + 1./96.*m11*m33*RS14e - 1./96.
         *m11*m33*RC146e - 17./288.*pow(m11,2) + 1./96.*pow(m11,2)*
         RS3461e + 1./96.*pow(m11,2)*RS1463e - 1./48.*pow(m11,2)*
         RSe6613 + 5./144.*pow(m11,2)*RSe63 - 1./144.*pow(m11,2)*
         RSe6633 - 5./288.*pow(m11,2)*RS34e + 1./288.*pow(m11,2)*RC346e
          - 1./48.*pow(m11,2)*RSe61 + 1./48.*pow(m11,2)*RSe6611 + 1./96.
         *pow(m11,2)*RS14e - 1./96.*pow(m11,2)*RC146e );

      M13p6x22R +=  + pow(Ab(m16,mu2),2) * (  - 1./45.*pow(m16,-1)*pow(
         me2,2) - 17./270.*pow(m16,-1)*pow(me2,2)*RSe6613 + 19./270.*
         pow(m16,-1)*pow(me2,2)*RSe63 - 1./270.*pow(m16,-1)*pow(me2,2)*
         RSe6633 - 7./270.*pow(m16,-1)*pow(me2,2)*RSe61 + 2./45.*pow(
         m16,-1)*pow(me2,2)*RSe6611 - 1./432.*pow(m16,-1)*m66*RD346e - 
         1./432.*pow(m16,-1)*m66*RD146e + 11./270.*pow(m16,-1)*m66*me2
          + 1./108.*pow(m16,-1)*m66*me2*RSe6613 - 2./45.*pow(m16,-1)*
         m66*me2*RSe63 + 1./216.*pow(m16,-1)*m66*me2*RSe6633 - 2./135.*
         pow(m16,-1)*m66*me2*RSe61 + 1./216.*pow(m16,-1)*m66*me2*
         RSe6611 - 1./54.*pow(m16,-1)*pow(m66,2) - 1./216.*pow(m16,-1)*
         pow(m66,2)*RS3461e - 1./216.*pow(m16,-1)*pow(m66,2)*RS1463e + 
         1./108.*pow(m16,-1)*pow(m66,2)*RSe6613 + 1./216.*pow(m16,-1)*
         pow(m66,2)*RSe6633 - 1./432.*pow(m16,-1)*pow(m66,2)*RC346e + 1.
         /216.*pow(m16,-1)*pow(m66,2)*RSe6611 - 1./432.*pow(m16,-1)*
         pow(m66,2)*RC146e + 11./540.*pow(m16,-1)*m33*RD346e - 1./432.*
         pow(m16,-1)*m33*RD146e );

      M13p6x22R +=  + pow(Ab(m16,mu2),2) * ( 1./54.*pow(m16,-1)*m33*me2
          - 11./1080.*pow(m16,-1)*m33*me2*RSe6613 + 1./40.*pow(m16,-1)*
         m33*me2*RSe63 - 1./30.*pow(m16,-1)*m33*me2*RSe6633 + 5./216.*
         pow(m16,-1)*m33*me2*RSe61 + 1./216.*pow(m16,-1)*m33*me2*
         RSe6611 - 49./864.*pow(m16,-1)*m33*m66 - 17./864.*pow(m16,-1)*
         m33*m66*RS3461e - 13./864.*pow(m16,-1)*m33*m66*RS1463e + 13./
         432.*pow(m16,-1)*m33*m66*RSe6613 + 19./432.*pow(m16,-1)*m33*
         m66*RSe63 + 1./4320.*pow(m16,-1)*m33*m66*RS34e - 1./432.*pow(
         m16,-1)*m33*m66*RC346e - 5./432.*pow(m16,-1)*m33*m66*RSe61 + 1.
         /108.*pow(m16,-1)*m33*m66*RSe6611 + 5./864.*pow(m16,-1)*m33*
         m66*RS14e - 1./216.*pow(m16,-1)*m33*m66*RC146e - 31./432.*pow(
         m16,-1)*m33*m44 - 31./432.*pow(m16,-1)*pow(m33,2) + 113./4320.
         *pow(m16,-1)*pow(m33,2)*RS3461e - 1./96.*pow(m16,-1)*pow(
         m33,2)*RS1463e + 1./48.*pow(m16,-1)*pow(m33,2)*RSe6613 + 1./
         144.*pow(m16,-1)*pow(m33,2)*RSe63 - 1./216.*pow(m16,-1)*pow(
         m33,2)*RSe6633 );

      M13p6x22R +=  + pow(Ab(m16,mu2),2) * (  - 221./4320.*pow(m16,-1)*
         pow(m33,2)*RS34e + 1./48.*pow(m16,-1)*pow(m33,2)*RC346e - 5./
         432.*pow(m16,-1)*pow(m33,2)*RSe61 + 1./216.*pow(m16,-1)*pow(
         m33,2)*RSe6611 + 5./864.*pow(m16,-1)*pow(m33,2)*RS14e - 1./432.
         *pow(m16,-1)*pow(m33,2)*RC146e + 1./48.*m11*pow(m16,-1)*RD346e
          - 1./45.*m11*pow(m16,-1)*RD146e - 1./1080.*m11*pow(m16,-1)*
         me2*RSe6613 + 23./216.*m11*pow(m16,-1)*me2*RSe63 - 1./24.*m11*
         pow(m16,-1)*me2*RSe6633 + 1./120.*m11*pow(m16,-1)*me2*RSe61 - 
         2./45.*m11*pow(m16,-1)*me2*RSe6611 - 11./288.*m11*pow(m16,-1)*
         m66 - 1./96.*m11*pow(m16,-1)*m66*RS3461e - 13./864.*m11*pow(
         m16,-1)*m66*RS1463e + 1./48.*m11*pow(m16,-1)*m66*RSe6613 + 1./
         16.*m11*pow(m16,-1)*m66*RSe63 + 1./216.*m11*pow(m16,-1)*m66*
         RSe6633 - 1./32.*m11*pow(m16,-1)*m66*RS34e - 1./432.*m11*pow(
         m16,-1)*m66*RC346e + 1./144.*m11*pow(m16,-1)*m66*RSe61 - 1./
         216.*m11*pow(m16,-1)*m66*RSe6611 + 17./4320.*m11*pow(m16,-1)*
         m66*RS14e );

      M13p6x22R +=  + pow(Ab(m16,mu2),2) * (  - 31./432.*m11*pow(
         m16,-1)*m44 - 59./864.*m11*pow(m16,-1)*m33 - 43./4320.*m11*
         pow(m16,-1)*m33*RS3461e - 23./4320.*m11*pow(m16,-1)*m33*
         RS1463e + 1./48.*m11*pow(m16,-1)*m33*RSe6613 + 1./16.*m11*pow(
         m16,-1)*m33*RSe63 + 1./216.*m11*pow(m16,-1)*m33*RSe6633 - 73./
         864.*m11*pow(m16,-1)*m33*RS34e + 1./54.*m11*pow(m16,-1)*m33*
         RC346e + 1./144.*m11*pow(m16,-1)*m33*RSe61 - 1./216.*m11*pow(
         m16,-1)*m33*RSe6611 - 13./864.*m11*pow(m16,-1)*m33*RS14e + 1./
         288.*pow(m11,2)*pow(m16,-1) + 23./720.*pow(m11,2)*pow(m16,-1)*
         RS1463e + 4./27.*pow(m11,2)*pow(m16,-1)*RSe63 - 2./27.*pow(
         m11,2)*pow(m16,-1)*RS34e + 19./2160.*pow(m11,2)*pow(m16,-1)*
         RS14e );

      M13p6x22R +=  + Ab(m16,mu2)*Ab(m33,mu2) * (  - 1./180.*RD346e + 7.
         /270.*m66*RS34e + 43./360.*m33*RS3461e - 41./360.*m33*RS34e + 
         1./24.*m33*RC346e + 1./27.*m33*m66*RN13*RS34e - 19./1080.*m11*
         RS3461e - 5./72.*m11*RS34e + 1./24.*m11*RC346e + 2./27.*m11*
         m66*RN13*RS34e - 1./27.*m11*m33*RN13*RS3461e + 1./9.*m11*m33*
         RN13*RS34e + 1./27.*pow(m11,2)*RN13*RS3461e + 1./9.*pow(m11,2)
         *RN13*RS34e );

      M13p6x22R +=  + Ab(m16,mu2)*Ab(m34,mu2) * (  - 1./4.*m33 - 1./4.*
         m11 );

      M13p6x22R +=  + Ab(m16,mu2)*Ab(m36,mu2) * (  - 1./108.*RD346e - 1.
         /108.*RD146e + 2./27.*me2 + 1./27.*me2*RSe6613 - 2./27.*me2*
         RSe63 + 1./54.*me2*RSe6633 - 2./27.*me2*RSe61 + 1./54.*me2*
         RSe6611 - 2./27.*m66 - 1./54.*m66*RS3461e - 1./54.*m66*RS1463e
          + 1./27.*m66*RSe6613 + 1./54.*m66*RSe6633 - 1./108.*m66*
         RC346e + 1./54.*m66*RSe6611 - 1./108.*m66*RC146e - 5./54.*m33
          - 13./216.*m33*RS3461e - 1./24.*m33*RS1463e + 1./12.*m33*
         RSe6613 + 1./36.*m33*RSe63 - 1./54.*m33*RSe6633 + 5./216.*m33*
         RS34e - 5./108.*m33*RSe61 + 1./54.*m33*RSe6611 + 5./216.*m33*
         RS14e - 1./108.*m33*RC146e - 5./54.*m11 - 1./24.*m11*RS3461e
          - 13./216.*m11*RS1463e + 1./12.*m11*RSe6613 - 5./108.*m11*
         RSe63 + 1./54.*m11*RSe6633 + 5./216.*m11*RS34e - 1./108.*m11*
         RC346e + 1./36.*m11*RSe61 - 1./54.*m11*RSe6611 + 5./216.*m11*
         RS14e );

      M13p6x22R +=  + Ab(m16,mu2)*Ab(me2,mu2) * (  - 2./45.*me2 - 22./
         135.*me2*RSe6613 + 19./135.*me2*RSe63 + 1./90.*me2*RSe6633 - 7.
         /135.*me2*RSe61 + 29./270.*me2*RSe6611 + 2./45.*m66 - 17./135.
         *m66*RSe63 + 11./135.*m66*RSe61 - 2./27.*m66*me2*RN1e*RSe63 + 
         2./27.*m66*me2*RN1e*RSe61 - 1./540.*m33*RSe6613 + 7./540.*m33*
         RSe63 - 17./180.*m33*RSe6633 + 1./12.*m33*RSe61 + 2./27.*m33*
         me2*RN1e*RSe6613 - 2./27.*m33*me2*RN1e*RSe63 - 2./27.*m33*m66*
         RN1e*RSe63 + 19./540.*m11*RSe6613 - 1./108.*m11*RSe63 - 1./12.
         *m11*RSe6633 + 49./540.*m11*RSe61 - 1./30.*m11*RSe6611 + 2./27.
         *m11*me2*RN1e*RSe6613 - 2./9.*m11*me2*RN1e*RSe63 + 2./27.*m11*
         me2*RN1e*RSe61 + 2./27.*m11*me2*RN1e*RSe6611 - 2./27.*m11*m66*
         RN1e*RSe63 - 2./27.*m11*m66*RN1e*RSe61 - 2./27.*m11*m33*RN1e*
         RSe6613 - 2./27.*m11*m33*RN1e*RSe63 - 2./27.*pow(m11,2)*RN1e*
         RSe6613 - 2./27.*pow(m11,2)*RN1e*RSe63 - 2./27.*pow(m11,2)*
         RN1e*RSe61 - 2./27.*pow(m11,2)*RN1e*RSe6611 );

      M13p6x22R +=  + Ab(m16,mu2)*Bb(m33,mu2) * ( 1./24.*m33*RD346e + 1.
         /24.*m11*RD346e );

      M13p6x22R +=  + Ab(m33,mu2)*pi16 * ( 1./48.*m33*RD346e - 1./72.*
         m33*RD346e*RS3461e + 1./144.*m33*RD346e*RC346e - 1./72.*m33*
         RD146e*RS3461e + 1./144.*m33*RD146e*RC346e - 1./18.*m33*me2*
         RSe6613*RS3461e + 1./36.*m33*me2*RSe6633*RS3461e + 1./36.*m33*
         me2*RC346e*RSe6613 - 1./72.*m33*me2*RC346e*RSe6633 + 1./36.*
         m33*me2*RSe6611*RS3461e - 1./72.*m33*me2*RSe6611*RC346e + 1./
         24.*m33*m66*RS3461e - 1./48.*m33*m66*RC346e + 1./12.*m33*m44*
         RS3461e - 1./24.*m33*m44*RC346e - 1./16.*pow(m33,2) + 1./27.*
         pow(m33,2)*pow(RS3461e,2) + 1./108.*pow(m33,2)*RS1463e*RS3461e
          - 1./54.*pow(m33,2)*RSe6613*RS3461e + 1./108.*pow(m33,2)*
         RSe6633*RS3461e + 1./24.*pow(m33,2)*RS36e + 1./48.*pow(m33,2)*
         RS34e - 1./27.*pow(m33,2)*RC346e*RS3461e - 1./216.*pow(m33,2)*
         RC346e*RS1463e + 1./108.*pow(m33,2)*RC346e*RSe6613 - 1./216.*
         pow(m33,2)*RC346e*RSe6633 + 1./108.*pow(m33,2)*pow(RC346e,2)
          + 1./108.*pow(m33,2)*RSe6611*RS3461e - 1./216.*pow(m33,2)*
         RSe6611*RC346e );

      M13p6x22R +=  + Ab(m33,mu2)*pi16 * (  - 1./216.*pow(m33,2)*RC146e
         *RS3461e + 1./432.*pow(m33,2)*RC146e*RC346e + 1./48.*m11*
         RD346e - 1./72.*m11*RD346e*RS3461e + 1./144.*m11*RD346e*RC346e
          - 1./72.*m11*RD146e*RS3461e + 1./144.*m11*RD146e*RC346e - 1./
         18.*m11*me2*RSe6613*RS3461e + 1./36.*m11*me2*RSe6633*RS3461e
          + 1./36.*m11*me2*RC346e*RSe6613 - 1./72.*m11*me2*RC346e*
         RSe6633 + 1./36.*m11*me2*RSe6611*RS3461e - 1./72.*m11*me2*
         RSe6611*RC346e + 1./24.*m11*m66*RS3461e - 1./48.*m11*m66*
         RC346e + 1./12.*m11*m44*RS3461e - 1./24.*m11*m44*RC346e - 1./
         16.*m11*m33 + 1./24.*m11*m33*RS3461e + 5./108.*m11*m33*pow(
         RS3461e,2) + 5./108.*m11*m33*RS1463e*RS3461e - 1./27.*m11*m33*
         RSe6613*RS3461e + 1./54.*m11*m33*RSe6633*RS3461e + 1./24.*m11*
         m33*RS36e + 1./48.*m11*m33*RS34e - 1./48.*m11*m33*RC346e - 5./
         108.*m11*m33*RC346e*RS3461e - 5./216.*m11*m33*RC346e*RS1463e
          + 1./54.*m11*m33*RC346e*RSe6613 - 1./108.*m11*m33*RC346e*
         RSe6633 );

      M13p6x22R +=  + Ab(m33,mu2)*pi16 * ( 5./432.*m11*m33*pow(
         RC346e,2) + 1./54.*m11*m33*RSe6611*RS3461e - 1./108.*m11*m33*
         RSe6611*RC346e - 5./216.*m11*m33*RC146e*RS3461e + 5./432.*m11*
         m33*RC146e*RC346e + 1./24.*pow(m11,2)*RS3461e + 1./108.*pow(
         m11,2)*pow(RS3461e,2) + 1./27.*pow(m11,2)*RS1463e*RS3461e - 1./
         54.*pow(m11,2)*RSe6613*RS3461e + 1./108.*pow(m11,2)*RSe6633*
         RS3461e - 1./48.*pow(m11,2)*RC346e - 1./108.*pow(m11,2)*RC346e
         *RS3461e - 1./54.*pow(m11,2)*RC346e*RS1463e + 1./108.*pow(
         m11,2)*RC346e*RSe6613 - 1./216.*pow(m11,2)*RC346e*RSe6633 + 1./
         432.*pow(m11,2)*pow(RC346e,2) + 1./108.*pow(m11,2)*RSe6611*
         RS3461e - 1./216.*pow(m11,2)*RSe6611*RC346e - 1./54.*pow(
         m11,2)*RC146e*RS3461e + 1./108.*pow(m11,2)*RC146e*RC346e );

      M13p6x22R +=  + pow(Ab(m33,mu2),2) * ( 1./1080.*pow(m33,-1)*pow(
         RD346e,2) + 1./1080.*pow(m33,-1)*RD146e*RD346e + 1./270.*pow(
         m33,-1)*me2*RD346e*RSe6613 - 1./540.*pow(m33,-1)*me2*RD346e*
         RSe6633 - 1./540.*pow(m33,-1)*me2*RSe6611*RD346e + 2./405.*
         pow(m33,-1)*pow(me2,2)*RSe6613*RS3461e - 1./405.*pow(m33,-1)*
         pow(me2,2)*RSe6633*RS3461e - 1./405.*pow(m33,-1)*pow(me2,2)*
         RC346e*RSe6613 + 1./810.*pow(m33,-1)*pow(me2,2)*RC346e*RSe6633
          - 1./405.*pow(m33,-1)*pow(me2,2)*RSe6611*RS3461e + 1./810.*
         pow(m33,-1)*pow(me2,2)*RSe6611*RC346e + 7./540.*pow(m33,-1)*
         m66*RD346e + 2./81.*pow(m33,-1)*m66*me2*RS34e*RSe63 - 1./162.*
         pow(m33,-1)*m66*me2*pow(RS34e,2) - 2./81.*pow(m33,-1)*m66*me2*
         RSe61*RS34e + 1./81.*pow(m33,-1)*m66*me2*RS14e*RS34e - 1./54.*
         pow(m33,-1)*pow(m66,2)*RSe63 + 1./81.*pow(m33,-1)*pow(m66,2)*
         RS36e*RSe63 + 1./180.*pow(m33,-1)*pow(m66,2)*RS34e - 23./486.*
         pow(m33,-1)*pow(m66,2)*RS34e*RSe63 - 1./162.*pow(m33,-1)*pow(
         m66,2)*RS34e*RS36e );

      M13p6x22R +=  + pow(Ab(m33,mu2),2) * ( 5./486.*pow(m33,-1)*pow(
         m66,2)*pow(RS34e,2) + 1./54.*pow(m33,-1)*pow(m66,2)*RSe61 - 1./
         81.*pow(m33,-1)*pow(m66,2)*RSe61*RS36e + 23./486.*pow(m33,-1)*
         pow(m66,2)*RSe61*RS34e + 1./162.*pow(m33,-1)*pow(m66,2)*RS16e*
         RS34e - 1./108.*pow(m33,-1)*pow(m66,2)*RS14e + 1./162.*pow(
         m33,-1)*pow(m66,2)*RS14e*RS36e - 5./243.*pow(m33,-1)*pow(
         m66,2)*RS14e*RS34e + 7./270.*pow(m33,-1)*m44*RD346e - 1./162.*
         pow(m33,-1)*m44*me2*RS36e*RSe63 - 1./324.*pow(m33,-1)*m44*me2*
         pow(RS36e,2) + 1./162.*pow(m33,-1)*m44*me2*RSe61*RS36e + 1./
         162.*pow(m33,-1)*m44*me2*RS16e*RS36e + 10./243.*pow(m33,-1)*
         m44*m66*RS36e*RSe63 + 2./243.*pow(m33,-1)*m44*m66*pow(RS36e,2)
          - 11./243.*pow(m33,-1)*m44*m66*RS34e*RSe63 - 1./81.*pow(
         m33,-1)*m44*m66*RS34e*RS36e + 2./243.*pow(m33,-1)*m44*m66*pow(
         RS34e,2) - 10./243.*pow(m33,-1)*m44*m66*RSe61*RS36e + 11./243.
         *pow(m33,-1)*m44*m66*RSe61*RS34e - 4./243.*pow(m33,-1)*m44*m66
         *RS16e*RS36e );

      M13p6x22R +=  + pow(Ab(m33,mu2),2) * ( 1./81.*pow(m33,-1)*m44*m66
         *RS16e*RS34e + 1./81.*pow(m33,-1)*m44*m66*RS14e*RS36e - 4./243.
         *pow(m33,-1)*m44*m66*RS14e*RS34e + 1./729.*pow(m33,-1)*m44*
         pow(m66,7)*pow(RN1e,3)*pow(RN3e,3) + 1./54.*pow(m33,-1)*pow(
         m44,2)*RSe63 + 1./90.*pow(m33,-1)*pow(m44,2)*RS36e + 13./486.*
         pow(m33,-1)*pow(m44,2)*RS36e*RSe63 + 7./972.*pow(m33,-1)*pow(
         m44,2)*pow(RS36e,2) - 1./162.*pow(m33,-1)*pow(m44,2)*RS34e*
         RSe63 - 1./162.*pow(m33,-1)*pow(m44,2)*RS34e*RS36e - 1./54.*
         pow(m33,-1)*pow(m44,2)*RSe61 - 13./486.*pow(m33,-1)*pow(m44,2)
         *RSe61*RS36e + 1./162.*pow(m33,-1)*pow(m44,2)*RSe61*RS34e - 1./
         54.*pow(m33,-1)*pow(m44,2)*RS16e - 7./486.*pow(m33,-1)*pow(
         m44,2)*RS16e*RS36e + 1./162.*pow(m33,-1)*pow(m44,2)*RS16e*
         RS34e + 1./162.*pow(m33,-1)*pow(m44,2)*RS14e*RS36e + 1./162.*
         pow(m33,-1)*pow(m44,2)*pow(m66,6)*pow(RN1e,3)*pow(RN3e,3) + 5./
         486.*pow(m33,-1)*pow(m44,3)*pow(m66,5)*pow(RN1e,3)*pow(RN3e,3)
          + 25./2916.*pow(m33,-1)*pow(m44,4)*pow(m66,4)*pow(RN1e,3)*
         pow(RN3e,3) );

      M13p6x22R +=  + pow(Ab(m33,mu2),2) * ( 5./1296.*pow(m33,-1)*pow(
         m44,5)*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) + 7./7776.*pow(
         m33,-1)*pow(m44,6)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) + 1./
         11664.*pow(m33,-1)*pow(m44,7)*m66*pow(RN1e,3)*pow(RN3e,3) + 1./
         45.*RD346e - 7./3240.*RD346e*RS3461e + 1./72.*RD346e*RS1463e
          - 1./36.*RD346e*RSe6613 + 7./216.*RD346e*RSe6633 - 41./6480.*
         RD346e*RC346e - 1./216.*RSe6611*RD346e + 1./432.*RC146e*RD346e
          - 1./405.*RD146e*RS3461e - 1./162.*RD146e*RC346e - 4./81.*me2
         *RSe6613*RS3461e - 1./90.*me2*RSe6633*RS3461e - 1./27.*me2*
         RSe6633*RS1463e + 2./27.*me2*RSe6633*RSe6613 - 2./27.*me2*pow(
         RSe6633,2) - 19./810.*me2*RC346e*RSe6613 + 11./540.*me2*RC346e
         *RSe6633 + 2./405.*me2*RSe6611*RS3461e + 1./81.*me2*RSe6611*
         RC346e - 101./2160.*m66*RS3461e - 1./81.*m66*RS1463e + 2./81.*
         m66*RSe6613 - 1./27.*m66*RSe63 + 4./81.*m66*RSe6633 - 1./162.*
         m66*pow(RS36e,2) - 1./40.*m66*RS34e - 1./162.*m66*RS34e*RS36e
          - 5./324.*m66*pow(RS34e,2) );

      M13p6x22R +=  + pow(Ab(m33,mu2),2) * ( 23./2592.*m66*RC346e + 2./
         27.*m66*RSe61 - 1./9.*m66*RSe61*RS34e + 2./81.*m66*RS16e*RS36e
          + 1./81.*m66*RS16e*RS34e - 1./27.*m66*RS14e + 1./81.*m66*
         RS14e*RS36e + 5./81.*m66*RS14e*RS34e - 1./729.*pow(m66,7)*pow(
         RN1e,3)*pow(RN3e,3) - 101./1080.*m44*RS3461e - 2./81.*m44*
         RS1463e + 4./81.*m44*RSe6613 + 1./27.*m44*RSe63 + 8./81.*m44*
         RSe6633 - 1./20.*m44*RS36e - 7./162.*m44*pow(RS36e,2) - 1./162.
         *m44*RS34e*RS36e - 1./162.*m44*pow(RS34e,2) + 23./1296.*m44*
         RC346e - 2./27.*m44*RSe61 + 4./27.*m44*RSe61*RS36e - 1./27.*
         m44*RSe61*RS34e - 2./27.*m44*RS16e + 14./81.*m44*RS16e*RS36e
          + 1./81.*m44*RS16e*RS34e + 1./81.*m44*RS14e*RS36e + 2./81.*
         m44*RS14e*RS34e - 23./4374.*m44*pow(m66,6)*pow(RN1e,3)*pow(
         RN3e,3) - 37./2187.*pow(m44,2)*pow(m66,5)*pow(RN1e,3)*pow(
         RN3e,3) - 223./8748.*pow(m44,3)*pow(m66,4)*pow(RN1e,3)*pow(
         RN3e,3) - 659./34992.*pow(m44,4)*pow(m66,3)*pow(RN1e,3)*pow(
         RN3e,3) );

      M13p6x22R +=  + pow(Ab(m33,mu2),2) * (  - 499./69984.*pow(m44,5)*
         pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) - 91./69984.*pow(m44,6)*m66
         *pow(RN1e,3)*pow(RN3e,3) - 1./11664.*pow(m44,7)*pow(RN1e,3)*
         pow(RN3e,3) + 55./288.*m33 - 223./4320.*m33*RS3461e + 41./810.
         *m33*pow(RS3461e,2) - 5./54.*m33*RS1463e - 1./324.*m33*RS1463e
         *RS3461e + 5./27.*m33*RSe6613 + 1./162.*m33*RSe6613*RS3461e + 
         1./27.*m33*RSe6633 - 1./162.*m33*RSe6633*RS3461e + 19./2160.*
         m33*RS36e + 19./4320.*m33*RS34e + 1./48.*m33*RC346e + 43./540.
         *m33*RC346e*RS3461e + 31./324.*m33*RC346e*RS1463e - 31./162.*
         m33*RC346e*RSe6613 + 13./108.*m33*RC346e*RSe6633 - 35./1296.*
         m33*pow(RC346e,2) - 1./324.*m33*RSe6611*RC346e - 1./18.*m33*
         RS16e - 1./36.*m33*RS14e + 1./648.*m33*RC146e*RC346e - 2./27.*
         m33*me2*RC346e*RN3e*RSe6613 + 2./27.*m33*me2*RC346e*RN3e*
         RSe6633 - 2./2187.*m33*pow(m66,6)*pow(RN1e,3)*pow(RN3e,3) - 
         103./8748.*m33*m44*pow(m66,5)*pow(RN1e,3)*pow(RN3e,3) - 179./
         17496.*m33*pow(m44,2)*pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + pow(Ab(m33,mu2),2) * ( 17./3888.*m33*pow(m44,3)*
         pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) + 563./69984.*m33*pow(
         m44,4)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) + 221./69984.*m33*
         pow(m44,5)*m66*pow(RN1e,3)*pow(RN3e,3) + 7./17496.*m33*pow(
         m44,6)*pow(RN1e,3)*pow(RN3e,3) - 2./27.*pow(m33,2)*RC346e*RN3e
         *RSe6613 - 2./27.*pow(m33,2)*RC346e*RN3e*RSe6633 - 1./27.*pow(
         m33,2)*RN13*RC346e*RS1463e + 161./8748.*pow(m33,2)*pow(m66,5)*
         pow(RN1e,3)*pow(RN3e,3) + 1009./17496.*pow(m33,2)*m44*pow(
         m66,4)*pow(RN1e,3)*pow(RN3e,3) + 2093./34992.*pow(m33,2)*pow(
         m44,2)*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) + 1835./69984.*pow(
         m33,2)*pow(m44,3)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) + 311./
         69984.*pow(m33,2)*pow(m44,4)*m66*pow(RN1e,3)*pow(RN3e,3) + 1./
         8748.*pow(m33,2)*pow(m44,5)*pow(RN1e,3)*pow(RN3e,3) - 89./2916.
         *pow(m33,3)*pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) - 371./5832.*
         pow(m33,3)*m44*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) - 193./3888.
         *pow(m33,3)*pow(m44,2)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + pow(Ab(m33,mu2),2) * (  - 401./23328.*pow(m33,3)*
         pow(m44,3)*m66*pow(RN1e,3)*pow(RN3e,3) - 13./5832.*pow(m33,3)*
         pow(m44,4)*pow(RN1e,3)*pow(RN3e,3) + 7./486.*pow(m33,4)*pow(
         m66,3)*pow(RN1e,3)*pow(RN3e,3) + 7./324.*pow(m33,4)*m44*pow(
         m66,2)*pow(RN1e,3)*pow(RN3e,3) + 7./648.*pow(m33,4)*pow(m44,2)
         *m66*pow(RN1e,3)*pow(RN3e,3) + 7./3888.*pow(m33,4)*pow(m44,3)*
         pow(RN1e,3)*pow(RN3e,3) + 1./729.*m11*pow(m33,-2)*m44*pow(
         m66,7)*pow(RN1e,3)*pow(RN3e,3) + 1./162.*m11*pow(m33,-2)*pow(
         m44,2)*pow(m66,6)*pow(RN1e,3)*pow(RN3e,3) + 5./486.*m11*pow(
         m33,-2)*pow(m44,3)*pow(m66,5)*pow(RN1e,3)*pow(RN3e,3) + 25./
         2916.*m11*pow(m33,-2)*pow(m44,4)*pow(m66,4)*pow(RN1e,3)*pow(
         RN3e,3) + 5./1296.*m11*pow(m33,-2)*pow(m44,5)*pow(m66,3)*pow(
         RN1e,3)*pow(RN3e,3) + 7./7776.*m11*pow(m33,-2)*pow(m44,6)*pow(
         m66,2)*pow(RN1e,3)*pow(RN3e,3) + 1./11664.*m11*pow(m33,-2)*
         pow(m44,7)*m66*pow(RN1e,3)*pow(RN3e,3) + 1./60.*m11*pow(
         m33,-1)*RD346e );

      M13p6x22R +=  + pow(Ab(m33,mu2),2) * (  - 17./3240.*m11*pow(
         m33,-1)*RD346e*RS3461e - 7./1080.*m11*pow(m33,-1)*RD346e*
         RS1463e + 1./108.*m11*pow(m33,-1)*RD346e*RSe6613 - 1./216.*m11
         *pow(m33,-1)*RD346e*RSe6633 + 13./1296.*m11*pow(m33,-1)*RD346e
         *RC346e - 1./216.*m11*pow(m33,-1)*RSe6611*RD346e + 7./2160.*
         m11*pow(m33,-1)*RC146e*RD346e + 4./405.*m11*pow(m33,-1)*RD146e
         *RS3461e - 11./1620.*m11*pow(m33,-1)*RD146e*RC346e - 17./405.*
         m11*pow(m33,-1)*me2*RSe6613*RS3461e + 1./810.*m11*pow(m33,-1)*
         me2*RSe6633*RS3461e - 2./405.*m11*pow(m33,-1)*me2*RC346e*
         RSe6613 - 5./324.*m11*pow(m33,-1)*me2*RC346e*RSe6633 - 2./135.
         *m11*pow(m33,-1)*me2*RSe6611*RS3461e + 1./90.*m11*pow(m33,-1)*
         me2*RSe6611*RC346e - 29./2160.*m11*pow(m33,-1)*m66*RS3461e - 1.
         /54.*m11*pow(m33,-1)*m66*RS1463e + 1./27.*m11*pow(m33,-1)*m66*
         RSe6613 - 23./1080.*m11*pow(m33,-1)*m66*RS34e + 31./2592.*m11*
         pow(m33,-1)*m66*RC346e + 1./54.*m11*pow(m33,-1)*m66*RSe61 - 1./
         18.*m11*pow(m33,-1)*m66*RSe61*RS34e );

      M13p6x22R +=  + pow(Ab(m33,mu2),2) * ( 1./81.*m11*pow(m33,-1)*m66
         *RS16e*RS36e + 1./162.*m11*pow(m33,-1)*m66*RS16e*RS34e - 1./
         108.*m11*pow(m33,-1)*m66*RS14e + 1./162.*m11*pow(m33,-1)*m66*
         RS14e*RS36e + 5./162.*m11*pow(m33,-1)*m66*RS14e*RS34e - 1./729.
         *m11*pow(m33,-1)*pow(m66,7)*pow(RN1e,3)*pow(RN3e,3) - 29./1080.
         *m11*pow(m33,-1)*m44*RS3461e - 1./27.*m11*pow(m33,-1)*m44*
         RS1463e + 2./27.*m11*pow(m33,-1)*m44*RSe6613 - 23./540.*m11*
         pow(m33,-1)*m44*RS36e + 31./1296.*m11*pow(m33,-1)*m44*RC346e
          - 1./54.*m11*pow(m33,-1)*m44*RSe61 + 2./27.*m11*pow(m33,-1)*
         m44*RSe61*RS36e - 1./54.*m11*pow(m33,-1)*m44*RSe61*RS34e - 1./
         54.*m11*pow(m33,-1)*m44*RS16e + 7./81.*m11*pow(m33,-1)*m44*
         RS16e*RS36e + 1./162.*m11*pow(m33,-1)*m44*RS16e*RS34e + 1./162.
         *m11*pow(m33,-1)*m44*RS14e*RS36e + 1./81.*m11*pow(m33,-1)*m44*
         RS14e*RS34e - 20./2187.*m11*pow(m33,-1)*m44*pow(m66,6)*pow(
         RN1e,3)*pow(RN3e,3) - 157./4374.*m11*pow(m33,-1)*pow(m44,2)*
         pow(m66,5)*pow(RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + pow(Ab(m33,mu2),2) * (  - 235./4374.*m11*pow(
         m33,-1)*pow(m44,3)*pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) - 1315./
         34992.*m11*pow(m33,-1)*pow(m44,4)*pow(m66,3)*pow(RN1e,3)*pow(
         RN3e,3) - 227./17496.*m11*pow(m33,-1)*pow(m44,5)*pow(m66,2)*
         pow(RN1e,3)*pow(RN3e,3) - 35./17496.*m11*pow(m33,-1)*pow(
         m44,6)*m66*pow(RN1e,3)*pow(RN3e,3) - 1./11664.*m11*pow(m33,-1)
         *pow(m44,7)*pow(RN1e,3)*pow(RN3e,3) + 35./144.*m11 + 1./216.*
         m11*RS3461e + 4./135.*m11*pow(RS3461e,2) - 1./18.*m11*RS1463e
          + 1./108.*m11*RS1463e*RS3461e + 1./9.*m11*RSe6613 + 5./162.*
         m11*RSe6613*RS3461e - 5./324.*m11*RSe6633*RS3461e - 7./90.*m11
         *RS36e - 7./180.*m11*RS34e + 25./864.*m11*RC346e + 61./1620.*
         m11*RC346e*RS3461e + 41./1080.*m11*RC346e*RS1463e - 17./324.*
         m11*RC346e*RSe6613 + 11./648.*m11*RC346e*RSe6633 + 1./162.*m11
         *pow(RC346e,2) - 5./324.*m11*RSe6611*RS3461e - 1./648.*m11*
         RSe6611*RC346e - 1./54.*m11*RS16e - 1./108.*m11*RS14e + 17./
         3240.*m11*RC146e*RS3461e );

      M13p6x22R +=  + pow(Ab(m33,mu2),2) * (  - 7./1296.*m11*RC146e*
         RC346e + 13./4374.*m11*pow(m66,6)*pow(RN1e,3)*pow(RN3e,3) - 53.
         /8748.*m11*m44*pow(m66,5)*pow(RN1e,3)*pow(RN3e,3) + 533./17496.
         *m11*pow(m44,2)*pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) + 785./
         11664.*m11*pow(m44,3)*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) + 
         1571./34992.*m11*pow(m44,4)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3)
          + 847./69984.*m11*pow(m44,5)*m66*pow(RN1e,3)*pow(RN3e,3) + 77.
         /69984.*m11*pow(m44,6)*pow(RN1e,3)*pow(RN3e,3) - 2./27.*m11*
         m33*RC346e*RN3e*RSe6613 - 2./27.*m11*m33*RN13*RC346e*RS1463e
          + 277./8748.*m11*m33*pow(m66,5)*pow(RN1e,3)*pow(RN3e,3) + 
         1225./8748.*m11*m33*m44*pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) + 
         4123./34992.*m11*m33*pow(m44,2)*pow(m66,3)*pow(RN1e,3)*pow(
         RN3e,3) + 73./4374.*m11*m33*pow(m44,3)*pow(m66,2)*pow(RN1e,3)*
         pow(RN3e,3) - 809./69984.*m11*m33*pow(m44,4)*m66*pow(RN1e,3)*
         pow(RN3e,3) - 209./69984.*m11*m33*pow(m44,5)*pow(RN1e,3)*pow(
         RN3e,3) );

      M13p6x22R +=  + pow(Ab(m33,mu2),2) * (  - 731./5832.*m11*pow(
         m33,2)*pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) - 899./2916.*m11*
         pow(m33,2)*m44*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) - 1801./7776.
         *m11*pow(m33,2)*pow(m44,2)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3)
          - 1471./23328.*m11*pow(m33,2)*pow(m44,3)*m66*pow(RN1e,3)*pow(
         RN3e,3) - 101./23328.*m11*pow(m33,2)*pow(m44,4)*pow(RN1e,3)*
         pow(RN3e,3) + 305./1944.*m11*pow(m33,3)*pow(m66,3)*pow(RN1e,3)
         *pow(RN3e,3) + 20./81.*m11*pow(m33,3)*m44*pow(m66,2)*pow(
         RN1e,3)*pow(RN3e,3) + 335./2592.*m11*pow(m33,3)*pow(m44,2)*m66
         *pow(RN1e,3)*pow(RN3e,3) + 175./7776.*m11*pow(m33,3)*pow(
         m44,3)*pow(RN1e,3)*pow(RN3e,3) - 7./108.*m11*pow(m33,4)*pow(
         m66,2)*pow(RN1e,3)*pow(RN3e,3) - 7./108.*m11*pow(m33,4)*m44*
         m66*pow(RN1e,3)*pow(RN3e,3) - 7./432.*m11*pow(m33,4)*pow(
         m44,2)*pow(RN1e,3)*pow(RN3e,3) - 17./4374.*pow(m11,2)*pow(
         m33,-2)*m44*pow(m66,6)*pow(RN1e,3)*pow(RN3e,3) - 83./4374.*
         pow(m11,2)*pow(m33,-2)*pow(m44,2)*pow(m66,5)*pow(RN1e,3)*pow(
         RN3e,3) );

      M13p6x22R +=  + pow(Ab(m33,mu2),2) * (  - 247./8748.*pow(m11,2)*
         pow(m33,-2)*pow(m44,3)*pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) - 41.
         /2187.*pow(m11,2)*pow(m33,-2)*pow(m44,4)*pow(m66,3)*pow(
         RN1e,3)*pow(RN3e,3) - 409./69984.*pow(m11,2)*pow(m33,-2)*pow(
         m44,5)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) - 49./69984.*pow(
         m11,2)*pow(m33,-2)*pow(m44,6)*m66*pow(RN1e,3)*pow(RN3e,3) + 5./
         96.*pow(m11,2)*pow(m33,-1) - 157./4320.*pow(m11,2)*pow(m33,-1)
         *RS3461e - 1./81.*pow(m11,2)*pow(m33,-1)*pow(RS3461e,2) + 1./
         162.*pow(m11,2)*pow(m33,-1)*RS1463e*RS3461e + 2./81.*pow(
         m11,2)*pow(m33,-1)*RSe6613*RS3461e - 1./108.*pow(m11,2)*pow(
         m33,-1)*RSe6633*RS3461e - 5./144.*pow(m11,2)*pow(m33,-1)*RS36e
          - 5./288.*pow(m11,2)*pow(m33,-1)*RS34e + 23./864.*pow(m11,2)*
         pow(m33,-1)*RC346e + 1./108.*pow(m11,2)*pow(m33,-1)*RC346e*
         RS3461e + 1./120.*pow(m11,2)*pow(m33,-1)*RC346e*RS1463e - 1./
         108.*pow(m11,2)*pow(m33,-1)*RC346e*RSe6613 + 5./648.*pow(
         m11,2)*pow(m33,-1)*RC346e*RSe6633 );

      M13p6x22R +=  + pow(Ab(m33,mu2),2) * (  - 5./1296.*pow(m11,2)*
         pow(m33,-1)*pow(RC346e,2) - 5./324.*pow(m11,2)*pow(m33,-1)*
         RSe6611*RS3461e + 1./648.*pow(m11,2)*pow(m33,-1)*RSe6611*
         RC346e + 53./3240.*pow(m11,2)*pow(m33,-1)*RC146e*RS3461e - 1./
         144.*pow(m11,2)*pow(m33,-1)*RC146e*RC346e + 17./4374.*pow(
         m11,2)*pow(m33,-1)*pow(m66,6)*pow(RN1e,3)*pow(RN3e,3) + 41./
         8748.*pow(m11,2)*pow(m33,-1)*m44*pow(m66,5)*pow(RN1e,3)*pow(
         RN3e,3) + 919./17496.*pow(m11,2)*pow(m33,-1)*pow(m44,2)*pow(
         m66,4)*pow(RN1e,3)*pow(RN3e,3) + 959./11664.*pow(m11,2)*pow(
         m33,-1)*pow(m44,3)*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) + 409./
         8748.*pow(m11,2)*pow(m33,-1)*pow(m44,4)*pow(m66,2)*pow(RN1e,3)
         *pow(RN3e,3) + 743./69984.*pow(m11,2)*pow(m33,-1)*pow(m44,5)*
         m66*pow(RN1e,3)*pow(RN3e,3) + 49./69984.*pow(m11,2)*pow(
         m33,-1)*pow(m44,6)*pow(RN1e,3)*pow(RN3e,3) + 125./8748.*pow(
         m11,2)*pow(m66,5)*pow(RN1e,3)*pow(RN3e,3) + 1031./8748.*pow(
         m11,2)*m44*pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + pow(Ab(m33,mu2),2) * ( 1913./34992.*pow(m11,2)*
         pow(m44,2)*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) - 799./17496.*
         pow(m11,2)*pow(m44,3)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) - 
         1127./34992.*pow(m11,2)*pow(m44,4)*m66*pow(RN1e,3)*pow(RN3e,3)
          - 167./34992.*pow(m11,2)*pow(m44,5)*pow(RN1e,3)*pow(RN3e,3)
          - 829./5832.*pow(m11,2)*m33*pow(m66,4)*pow(RN1e,3)*pow(
         RN3e,3) - 5119./11664.*pow(m11,2)*m33*m44*pow(m66,3)*pow(
         RN1e,3)*pow(RN3e,3) - 1105./3888.*pow(m11,2)*m33*pow(m44,2)*
         pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) - 487./11664.*pow(m11,2)*
         m33*pow(m44,3)*m66*pow(RN1e,3)*pow(RN3e,3) + 49./11664.*pow(
         m11,2)*m33*pow(m44,4)*pow(RN1e,3)*pow(RN3e,3) + 1247./3888.*
         pow(m11,2)*pow(m33,2)*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) + 
         1511./2592.*pow(m11,2)*pow(m33,2)*m44*pow(m66,2)*pow(RN1e,3)*
         pow(RN3e,3) + 721./2592.*pow(m11,2)*pow(m33,2)*pow(m44,2)*m66*
         pow(RN1e,3)*pow(RN3e,3) + 65./1944.*pow(m11,2)*pow(m33,2)*pow(
         m44,3)*pow(RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + pow(Ab(m33,mu2),2) * (  - 127./432.*pow(m11,2)*
         pow(m33,3)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) - 269./864.*pow(
         m11,2)*pow(m33,3)*m44*m66*pow(RN1e,3)*pow(RN3e,3) - 71./864.*
         pow(m11,2)*pow(m33,3)*pow(m44,2)*pow(RN1e,3)*pow(RN3e,3) + 7./
         72.*pow(m11,2)*pow(m33,4)*m66*pow(RN1e,3)*pow(RN3e,3) + 7./144.
         *pow(m11,2)*pow(m33,4)*m44*pow(RN1e,3)*pow(RN3e,3) - 1./972.*
         pow(m11,3)*pow(m33,-2)*m44*pow(m66,5)*pow(RN1e,3)*pow(RN3e,3)
          + 23./1944.*pow(m11,3)*pow(m33,-2)*pow(m44,2)*pow(m66,4)*pow(
         RN1e,3)*pow(RN3e,3) + 25./1296.*pow(m11,3)*pow(m33,-2)*pow(
         m44,3)*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) + 77./7776.*pow(
         m11,3)*pow(m33,-2)*pow(m44,4)*pow(m66,2)*pow(RN1e,3)*pow(
         RN3e,3) + 13./7776.*pow(m11,3)*pow(m33,-2)*pow(m44,5)*m66*pow(
         RN1e,3)*pow(RN3e,3) + 1./972.*pow(m11,3)*pow(m33,-1)*pow(
         m66,5)*pow(RN1e,3)*pow(RN3e,3) + 5./108.*pow(m11,3)*pow(
         m33,-1)*m44*pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) + 23./3888.*
         pow(m11,3)*pow(m33,-1)*pow(m44,2)*pow(m66,3)*pow(RN1e,3)*pow(
         RN3e,3) );

      M13p6x22R +=  + pow(Ab(m33,mu2),2) * (  - 17./486.*pow(m11,3)*
         pow(m33,-1)*pow(m44,3)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) - 43.
         /2592.*pow(m11,3)*pow(m33,-1)*pow(m44,4)*m66*pow(RN1e,3)*pow(
         RN3e,3) - 13./7776.*pow(m11,3)*pow(m33,-1)*pow(m44,5)*pow(
         RN1e,3)*pow(RN3e,3) - 113./1944.*pow(m11,3)*pow(m66,4)*pow(
         RN1e,3)*pow(RN3e,3) - 1043./3888.*pow(m11,3)*m44*pow(m66,3)*
         pow(RN1e,3)*pow(RN3e,3) - 7./48.*pow(m11,3)*pow(m44,2)*pow(
         m66,2)*pow(RN1e,3)*pow(RN3e,3) + 1./3888.*pow(m11,3)*pow(
         m44,3)*m66*pow(RN1e,3)*pow(RN3e,3) + 13./1944.*pow(m11,3)*pow(
         m44,4)*pow(RN1e,3)*pow(RN3e,3) + 35./144.*pow(m11,3)*m33*pow(
         m66,3)*pow(RN1e,3)*pow(RN3e,3) + 44./81.*pow(m11,3)*m33*m44*
         pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) + 293./1296.*pow(m11,3)*m33
         *pow(m44,2)*m66*pow(RN1e,3)*pow(RN3e,3) + 5./324.*pow(m11,3)*
         m33*pow(m44,3)*pow(RN1e,3)*pow(RN3e,3) - 965./2592.*pow(m11,3)
         *pow(m33,2)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) - 1157./2592.*
         pow(m11,3)*pow(m33,2)*m44*m66*pow(RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + pow(Ab(m33,mu2),2) * (  - 127./1296.*pow(m11,3)*
         pow(m33,2)*pow(m44,2)*pow(RN1e,3)*pow(RN3e,3) + 203./864.*pow(
         m11,3)*pow(m33,3)*m66*pow(RN1e,3)*pow(RN3e,3) + 109./864.*pow(
         m11,3)*pow(m33,3)*m44*pow(RN1e,3)*pow(RN3e,3) - 7./144.*pow(
         m11,3)*pow(m33,4)*pow(RN1e,3)*pow(RN3e,3) + 7./648.*pow(m11,4)
         *pow(m33,-2)*m44*pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) + 1./108.*
         pow(m11,4)*pow(m33,-2)*pow(m44,2)*pow(m66,3)*pow(RN1e,3)*pow(
         RN3e,3) + 1./864.*pow(m11,4)*pow(m33,-2)*pow(m44,3)*pow(m66,2)
         *pow(RN1e,3)*pow(RN3e,3) - 1./2592.*pow(m11,4)*pow(m33,-2)*
         pow(m44,4)*m66*pow(RN1e,3)*pow(RN3e,3) - 7./648.*pow(m11,4)*
         pow(m33,-1)*pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) - 53./648.*pow(
         m11,4)*pow(m33,-1)*m44*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) - 
         133./2592.*pow(m11,4)*pow(m33,-1)*pow(m44,2)*pow(m66,2)*pow(
         RN1e,3)*pow(RN3e,3) - 5./864.*pow(m11,4)*pow(m33,-1)*pow(
         m44,3)*m66*pow(RN1e,3)*pow(RN3e,3) + 1./2592.*pow(m11,4)*pow(
         m33,-1)*pow(m44,4)*pow(RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + pow(Ab(m33,mu2),2) * ( 47./648.*pow(m11,4)*pow(
         m66,3)*pow(RN1e,3)*pow(RN3e,3) + 581./2592.*pow(m11,4)*m44*
         pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) + 227./2592.*pow(m11,4)*
         pow(m44,2)*m66*pow(RN1e,3)*pow(RN3e,3) + 1./216.*pow(m11,4)*
         pow(m44,3)*pow(RN1e,3)*pow(RN3e,3) - 451./2592.*pow(m11,4)*m33
         *pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) - 673./2592.*pow(m11,4)*
         m33*m44*m66*pow(RN1e,3)*pow(RN3e,3) - 59./1296.*pow(m11,4)*m33
         *pow(m44,2)*pow(RN1e,3)*pow(RN3e,3) + 77./432.*pow(m11,4)*pow(
         m33,2)*m66*pow(RN1e,3)*pow(RN3e,3) + 23./216.*pow(m11,4)*pow(
         m33,2)*m44*pow(RN1e,3)*pow(RN3e,3) - 19./288.*pow(m11,4)*pow(
         m33,3)*pow(RN1e,3)*pow(RN3e,3) - 5./648.*pow(m11,5)*pow(
         m33,-2)*m44*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) - 5./648.*pow(
         m11,5)*pow(m33,-2)*pow(m44,2)*pow(m66,2)*pow(RN1e,3)*pow(
         RN3e,3) - 5./2592.*pow(m11,5)*pow(m33,-2)*pow(m44,3)*m66*pow(
         RN1e,3)*pow(RN3e,3) + 5./648.*pow(m11,5)*pow(m33,-1)*pow(
         m66,3)*pow(RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + pow(Ab(m33,mu2),2) * ( 25./648.*pow(m11,5)*pow(
         m33,-1)*m44*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) + 55./2592.*
         pow(m11,5)*pow(m33,-1)*pow(m44,2)*m66*pow(RN1e,3)*pow(RN3e,3)
          + 5./2592.*pow(m11,5)*pow(m33,-1)*pow(m44,3)*pow(RN1e,3)*pow(
         RN3e,3) - 5./162.*pow(m11,5)*pow(m66,2)*pow(RN1e,3)*pow(
         RN3e,3) - 155./2592.*pow(m11,5)*m44*m66*pow(RN1e,3)*pow(
         RN3e,3) - 35./2592.*pow(m11,5)*pow(m44,2)*pow(RN1e,3)*pow(
         RN3e,3) + 35./864.*pow(m11,5)*m33*m66*pow(RN1e,3)*pow(RN3e,3)
          + 25./864.*pow(m11,5)*m33*m44*pow(RN1e,3)*pow(RN3e,3) - 5./
         288.*pow(m11,5)*pow(m33,2)*pow(RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + Ab(m33,mu2)*Ab(m34,mu2) * (  - 29./270.*RD346e + 
         2./27.*m44*RSe63 - 1./135.*m44*RS36e - 2./27.*m44*RSe61 - 2./
         27.*m44*RS16e + 2./27.*m44*me2*RN3e*RSe63 - 2./27.*m44*me2*
         RSe61*RN3e + 109./540.*m33*RS3461e + 2./27.*m33*RS1463e - 4./
         27.*m33*RSe6613 + 2./27.*m33*RSe63 - 4./27.*m33*RSe6633 + 19./
         540.*m33*RS36e - 2./9.*m33*RSe61 - 2./9.*m33*RS16e - 4./27.*
         m33*me2*RN3e*RSe6613 + 2./27.*m33*me2*RN3e*RSe63 - 4./27.*m33*
         me2*RN3e*RSe6633 - 2./9.*m33*me2*RSe61*RN3e - 2./27.*m33*m44*
         RN3e*RSe63 - 2./27.*m33*m44*RSe61*RN3e + 2./27.*m33*m44*RN13*
         RS16e + 4./27.*pow(m33,2)*RN3e*RSe6613 - 2./27.*pow(m33,2)*
         RN3e*RSe63 + 4./27.*pow(m33,2)*RN3e*RSe6633 - 2./27.*pow(
         m33,2)*RSe61*RN3e + 2./27.*pow(m33,2)*RN13*RS1463e + 2./27.*
         pow(m33,2)*RN13*RS16e + 41./540.*m11*RS3461e + 2./27.*m11*
         RS1463e - 4./27.*m11*RSe6613 - 1./12.*m11*RS36e - 2./27.*m11*
         RSe61 - 2./27.*m11*RS16e - 4./27.*m11*me2*RN3e*RSe6613 - 2./27.
         *m11*me2*RSe61*RN3e );

      M13p6x22R +=  + Ab(m33,mu2)*Ab(m34,mu2) * (  - 2./27.*m11*m44*
         RSe61*RN3e + 4./27.*m11*m44*RN13*RS16e + 4./27.*m11*m33*RN3e*
         RSe6613 - 2./27.*m11*m33*RSe61*RN3e + 8./27.*m11*m33*RN13*
         RS16e - 2./27.*pow(m11,2)*RN13*RS1463e + 2./27.*pow(m11,2)*
         RN13*RS16e );

      M13p6x22R +=  + Ab(m33,mu2)*Ab(m36,mu2) * (  - 29./540.*RD346e - 
         2./27.*m66*RSe63 - 1./270.*m66*RS34e + 2./27.*m66*RSe61 - 1./
         27.*m66*RS14e - 2./27.*m66*me2*RN3e*RSe63 + 2./27.*m66*me2*
         RSe61*RN3e + 109./1080.*m33*RS3461e + 1./27.*m33*RS1463e - 2./
         27.*m33*RSe6613 - 2./27.*m33*RSe63 - 2./27.*m33*RSe6633 + 19./
         1080.*m33*RS34e + 2./9.*m33*RSe61 - 1./9.*m33*RS14e - 2./27.*
         m33*me2*RN3e*RSe6613 - 2./27.*m33*me2*RN3e*RSe63 - 2./27.*m33*
         me2*RN3e*RSe6633 + 2./9.*m33*me2*RSe61*RN3e + 2./27.*m33*m66*
         RN3e*RSe63 + 2./27.*m33*m66*RSe61*RN3e + 1./27.*m33*m66*RN13*
         RS14e + 2./27.*pow(m33,2)*RN3e*RSe6613 + 2./27.*pow(m33,2)*
         RN3e*RSe63 + 2./27.*pow(m33,2)*RN3e*RSe6633 + 2./27.*pow(
         m33,2)*RSe61*RN3e + 1./27.*pow(m33,2)*RN13*RS1463e + 1./27.*
         pow(m33,2)*RN13*RS14e + 41./1080.*m11*RS3461e + 1./27.*m11*
         RS1463e - 2./27.*m11*RSe6613 - 1./24.*m11*RS34e + 2./27.*m11*
         RSe61 - 1./27.*m11*RS14e - 2./27.*m11*me2*RN3e*RSe6613 + 2./27.
         *m11*me2*RSe61*RN3e );

      M13p6x22R +=  + Ab(m33,mu2)*Ab(m36,mu2) * ( 2./27.*m11*m66*RSe61*
         RN3e + 2./27.*m11*m66*RN13*RS14e + 2./27.*m11*m33*RN3e*RSe6613
          + 2./27.*m11*m33*RSe61*RN3e + 4./27.*m11*m33*RN13*RS14e - 1./
         27.*pow(m11,2)*RN13*RS1463e + 1./27.*pow(m11,2)*RN13*RS14e );

      M13p6x22R +=  + Ab(m33,mu2)*Ab(m44,mu2) * (  - 1./9.*m44*RS36e*
         RSe63 - 1./18.*m44*pow(RS36e,2) + 1./9.*m44*RSe61*RS36e + 1./9.
         *m44*RS16e*RS36e - 1./9.*m44*me2*RN3e*RS36e*RSe63 + 1./9.*m44*
         me2*RSe61*RN3e*RS36e + 1./9.*m33*m44*RN3e*RS36e*RSe63 + 1./9.*
         m33*m44*RSe61*RN3e*RS36e - 1./9.*m33*m44*RN13*RS16e*RS36e + 1./
         9.*m11*m44*RSe61*RN3e*RS36e - 2./9.*m11*m44*RN13*RS16e*RS36e )
         ;

      M13p6x22R +=  + Ab(m33,mu2)*Ab(m46,mu2) * ( 4./81.*m66*RS36e*
         RSe63 + 2./81.*m66*RS34e*RSe63 - 2./81.*m66*RS34e*RS36e - 1./
         81.*m66*pow(RS34e,2) - 4./81.*m66*RSe61*RS36e - 2./81.*m66*
         RSe61*RS34e + 2./81.*m66*RS16e*RS34e + 2./81.*m66*RS14e*RS36e
          + 2./81.*m66*RS14e*RS34e + 4./81.*m66*me2*RN3e*RS36e*RSe63 + 
         2./81.*m66*me2*RN3e*RS34e*RSe63 - 4./81.*m66*me2*RSe61*RN3e*
         RS36e - 2./81.*m66*me2*RSe61*RN3e*RS34e + 2./81.*m44*RS36e*
         RSe63 - 1./81.*m44*pow(RS36e,2) - 2./81.*m44*RS34e*RSe63 - 2./
         81.*m44*RS34e*RS36e - 2./81.*m44*RSe61*RS36e + 2./81.*m44*
         RSe61*RS34e + 2./81.*m44*RS16e*RS36e + 2./81.*m44*RS16e*RS34e
          + 2./81.*m44*RS14e*RS36e + 2./81.*m44*me2*RN3e*RS36e*RSe63 - 
         2./81.*m44*me2*RN3e*RS34e*RSe63 - 2./81.*m44*me2*RSe61*RN3e*
         RS36e + 2./81.*m44*me2*RSe61*RN3e*RS34e + 1./81.*m33*pow(
         RS36e,2) - 2./81.*m33*RS34e*RS36e + 1./81.*m33*pow(RS34e,2) - 
         4./27.*m33*RSe61*RS36e + 4./27.*m33*RSe61*RS34e - 4./81.*m33*
         RS16e*RS36e );

      M13p6x22R +=  + Ab(m33,mu2)*Ab(m46,mu2) * ( 4./81.*m33*RS16e*
         RS34e + 4./81.*m33*RS14e*RS36e - 4./81.*m33*RS14e*RS34e - 4./
         27.*m33*me2*RSe61*RN3e*RS36e + 4./27.*m33*me2*RSe61*RN3e*RS34e
          - 4./81.*m33*m66*RN3e*RS36e*RSe63 - 2./81.*m33*m66*RN3e*RS34e
         *RSe63 - 4./81.*m33*m66*RSe61*RN3e*RS36e - 2./81.*m33*m66*
         RSe61*RN3e*RS34e - 2./81.*m33*m66*RN13*RS16e*RS34e - 2./81.*
         m33*m66*RN13*RS14e*RS36e - 2./81.*m33*m66*RN13*RS14e*RS34e - 2.
         /81.*m33*m44*RN3e*RS36e*RSe63 + 2./81.*m33*m44*RN3e*RS34e*
         RSe63 - 2./81.*m33*m44*RSe61*RN3e*RS36e + 2./81.*m33*m44*RSe61
         *RN3e*RS34e - 2./81.*m33*m44*RN13*RS16e*RS36e - 2./81.*m33*m44
         *RN13*RS16e*RS34e - 2./81.*m33*m44*RN13*RS14e*RS36e - 2./27.*
         m11*RSe61*RS36e + 2./27.*m11*RSe61*RS34e - 2./81.*m11*RS16e*
         RS36e + 2./81.*m11*RS16e*RS34e + 2./81.*m11*RS14e*RS36e - 2./
         81.*m11*RS14e*RS34e - 2./27.*m11*me2*RSe61*RN3e*RS36e + 2./27.
         *m11*me2*RSe61*RN3e*RS34e - 4./81.*m11*m66*RSe61*RN3e*RS36e - 
         2./81.*m11*m66*RSe61*RN3e*RS34e );

      M13p6x22R +=  + Ab(m33,mu2)*Ab(m46,mu2) * (  - 4./81.*m11*m66*
         RN13*RS16e*RS34e - 4./81.*m11*m66*RN13*RS14e*RS36e - 4./81.*
         m11*m66*RN13*RS14e*RS34e - 2./81.*m11*m44*RSe61*RN3e*RS36e + 2.
         /81.*m11*m44*RSe61*RN3e*RS34e - 4./81.*m11*m44*RN13*RS16e*
         RS36e - 4./81.*m11*m44*RN13*RS16e*RS34e - 4./81.*m11*m44*RN13*
         RS14e*RS36e + 4./81.*m11*m33*RN13*RS16e*RS36e - 4./81.*m11*m33
         *RN13*RS16e*RS34e - 4./81.*m11*m33*RN13*RS14e*RS36e + 4./81.*
         m11*m33*RN13*RS14e*RS34e + 2./81.*pow(m11,2)*RN13*RS16e*RS36e
          - 2./81.*pow(m11,2)*RN13*RS16e*RS34e - 2./81.*pow(m11,2)*RN13
         *RS14e*RS36e + 2./81.*pow(m11,2)*RN13*RS14e*RS34e );

      M13p6x22R +=  + Ab(m33,mu2)*Ab(me2,mu2) * ( 1./135.*RD346e*
         RSe6613 - 1./270.*RD346e*RSe6633 - 1./270.*RSe6611*RD346e - 2./
         135.*me2*RSe6613*RS3461e + 1./135.*me2*RSe6633*RS3461e + 1./
         135.*me2*RC346e*RSe6613 - 1./270.*me2*RC346e*RSe6633 + 1./135.
         *me2*RSe6611*RS3461e - 1./270.*me2*RSe6611*RC346e + 4./81.*m66
         *RS34e*RSe63 - 1./81.*m66*pow(RS34e,2) - 4./81.*m66*RSe61*
         RS34e + 2./81.*m66*RS14e*RS34e + 4./81.*m66*me2*RN3e*RS34e*
         RSe63 - 4./81.*m66*me2*RSe61*RN3e*RS34e - 1./81.*m44*RS36e*
         RSe63 - 1./162.*m44*pow(RS36e,2) + 1./81.*m44*RSe61*RS36e + 1./
         81.*m44*RS16e*RS36e - 1./81.*m44*me2*RN3e*RS36e*RSe63 + 1./81.
         *m44*me2*RSe61*RN3e*RS36e - 1./9.*m33*RSe6613*RS3461e - 11./
         270.*m33*RSe6633*RS3461e - 2./27.*m33*RSe6633*RS1463e + 4./27.
         *m33*RSe6633*RSe6613 - 4./27.*m33*pow(RSe6633,2) + 1./30.*m33*
         RC346e*RSe6613 - 13./540.*m33*RC346e*RSe6633 + 1./270.*m33*
         RSe6611*RS3461e + 1./36.*m33*RSe6611*RC346e + 4./27.*m33*me2*
         RN3e*RSe6633*RSe6613 );

      M13p6x22R +=  + Ab(m33,mu2)*Ab(me2,mu2) * (  - 4./27.*m33*me2*
         RN3e*pow(RSe6633,2) + 2./27.*m33*me2*RC346e*RN3e*RSe6613 - 2./
         27.*m33*me2*RC346e*RN3e*RSe6633 - 4./81.*m33*m66*RN3e*RS34e*
         RSe63 - 4./81.*m33*m66*RSe61*RN3e*RS34e - 2./81.*m33*m66*RN13*
         RS14e*RS34e + 1./81.*m33*m44*RN3e*RS36e*RSe63 + 1./81.*m33*m44
         *RSe61*RN3e*RS36e - 1./81.*m33*m44*RN13*RS16e*RS36e + 4./27.*
         pow(m33,2)*RN3e*RSe6633*RSe6613 + 4./27.*pow(m33,2)*RN3e*pow(
         RSe6633,2) + 2./27.*pow(m33,2)*RC346e*RN3e*RSe6613 + 2./27.*
         pow(m33,2)*RC346e*RN3e*RSe6633 + 2./27.*pow(m33,2)*RN13*
         RSe6633*RS1463e - 13./135.*m11*RSe6613*RS3461e - 1./270.*m11*
         RSe6633*RS3461e - 1./270.*m11*RC346e*RSe6613 - 1./36.*m11*
         RC346e*RSe6633 - 13./270.*m11*RSe6611*RS3461e + 17./540.*m11*
         RSe6611*RC346e - 4./81.*m11*m66*RSe61*RN3e*RS34e - 4./81.*m11*
         m66*RN13*RS14e*RS34e + 1./81.*m11*m44*RSe61*RN3e*RS36e - 2./81.
         *m11*m44*RN13*RS16e*RS36e + 4./27.*m11*m33*RN3e*RSe6633*
         RSe6613 );

      M13p6x22R +=  + Ab(m33,mu2)*Ab(me2,mu2) * ( 2./27.*m11*m33*RC346e
         *RN3e*RSe6613 + 4./27.*m11*m33*RN13*RSe6633*RS1463e + 2./27.*
         m11*m33*RN13*RSe6611*RS3461e + 4./27.*pow(m11,2)*RN13*RSe6611*
         RS3461e );

      M13p6x22R +=  + Ab(m33,mu2)*Bb(m11,mu2) * ( 1./540.*RD146e*RD346e
          - 1./540.*m33*RD146e*RS3461e - 1./72.*m33*RD146e*RC346e + 11./
         540.*m11*RD146e*RS3461e - 1./72.*m11*RD146e*RC346e - 1./27.*
         m11*m33*RN13*RD146e*RS3461e - 2./27.*pow(m11,2)*RN13*RD146e*
         RS3461e );

      M13p6x22R +=  + Ab(m33,mu2)*Bb(m33,mu2) * ( 1./540.*pow(RD346e,2)
          - 2./729.*m44*pow(m66,7)*pow(RN1e,3)*pow(RN3e,3) - 1./81.*
         pow(m44,2)*pow(m66,6)*pow(RN1e,3)*pow(RN3e,3) - 5./243.*pow(
         m44,3)*pow(m66,5)*pow(RN1e,3)*pow(RN3e,3) - 25./1458.*pow(
         m44,4)*pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) - 5./648.*pow(m44,5)
         *pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) - 7./3888.*pow(m44,6)*pow(
         m66,2)*pow(RN1e,3)*pow(RN3e,3) - 1./5832.*pow(m44,7)*m66*pow(
         RN1e,3)*pow(RN3e,3) + 1./60.*m33*RD346e*RS3461e + 1./27.*m33*
         RD346e*RS1463e - 2./27.*m33*RD346e*RSe6613 + 2./27.*m33*RD346e
         *RSe6633 - 5./216.*m33*RD346e*RC346e - 2./27.*m33*me2*RD346e*
         RN3e*RSe6613 + 2./27.*m33*me2*RD346e*RN3e*RSe6633 + 2./729.*
         m33*pow(m66,7)*pow(RN1e,3)*pow(RN3e,3) + 23./2187.*m33*m44*
         pow(m66,6)*pow(RN1e,3)*pow(RN3e,3) + 74./2187.*m33*pow(m44,2)*
         pow(m66,5)*pow(RN1e,3)*pow(RN3e,3) + 223./4374.*m33*pow(m44,3)
         *pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) + 659./17496.*m33*pow(
         m44,4)*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + Ab(m33,mu2)*Bb(m33,mu2) * ( 499./34992.*m33*pow(
         m44,5)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) + 91./34992.*m33*
         pow(m44,6)*m66*pow(RN1e,3)*pow(RN3e,3) + 1./5832.*m33*pow(
         m44,7)*pow(RN1e,3)*pow(RN3e,3) + 2./27.*pow(m33,2)*RC346e*
         RS3461e - 2./27.*pow(m33,2)*RD346e*RN3e*RSe6613 - 2./27.*pow(
         m33,2)*RD346e*RN3e*RSe6633 - 1./27.*pow(m33,2)*RN13*RD346e*
         RS1463e + 4./2187.*pow(m33,2)*pow(m66,6)*pow(RN1e,3)*pow(
         RN3e,3) + 103./4374.*pow(m33,2)*m44*pow(m66,5)*pow(RN1e,3)*
         pow(RN3e,3) + 179./8748.*pow(m33,2)*pow(m44,2)*pow(m66,4)*pow(
         RN1e,3)*pow(RN3e,3) - 17./1944.*pow(m33,2)*pow(m44,3)*pow(
         m66,3)*pow(RN1e,3)*pow(RN3e,3) - 563./34992.*pow(m33,2)*pow(
         m44,4)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) - 221./34992.*pow(
         m33,2)*pow(m44,5)*m66*pow(RN1e,3)*pow(RN3e,3) - 7./8748.*pow(
         m33,2)*pow(m44,6)*pow(RN1e,3)*pow(RN3e,3) - 161./4374.*pow(
         m33,3)*pow(m66,5)*pow(RN1e,3)*pow(RN3e,3) - 1009./8748.*pow(
         m33,3)*m44*pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + Ab(m33,mu2)*Bb(m33,mu2) * (  - 2093./17496.*pow(
         m33,3)*pow(m44,2)*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) - 1835./
         34992.*pow(m33,3)*pow(m44,3)*pow(m66,2)*pow(RN1e,3)*pow(
         RN3e,3) - 311./34992.*pow(m33,3)*pow(m44,4)*m66*pow(RN1e,3)*
         pow(RN3e,3) - 1./4374.*pow(m33,3)*pow(m44,5)*pow(RN1e,3)*pow(
         RN3e,3) + 89./1458.*pow(m33,4)*pow(m66,4)*pow(RN1e,3)*pow(
         RN3e,3) + 371./2916.*pow(m33,4)*m44*pow(m66,3)*pow(RN1e,3)*
         pow(RN3e,3) + 193./1944.*pow(m33,4)*pow(m44,2)*pow(m66,2)*pow(
         RN1e,3)*pow(RN3e,3) + 401./11664.*pow(m33,4)*pow(m44,3)*m66*
         pow(RN1e,3)*pow(RN3e,3) + 13./2916.*pow(m33,4)*pow(m44,4)*pow(
         RN1e,3)*pow(RN3e,3) - 7./243.*pow(m33,5)*pow(m66,3)*pow(
         RN1e,3)*pow(RN3e,3) - 7./162.*pow(m33,5)*m44*pow(m66,2)*pow(
         RN1e,3)*pow(RN3e,3) - 7./324.*pow(m33,5)*pow(m44,2)*m66*pow(
         RN1e,3)*pow(RN3e,3) - 7./1944.*pow(m33,5)*pow(m44,3)*pow(
         RN1e,3)*pow(RN3e,3) - 2./729.*m11*pow(m33,-1)*m44*pow(m66,7)*
         pow(RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + Ab(m33,mu2)*Bb(m33,mu2) * (  - 1./81.*m11*pow(
         m33,-1)*pow(m44,2)*pow(m66,6)*pow(RN1e,3)*pow(RN3e,3) - 5./243.
         *m11*pow(m33,-1)*pow(m44,3)*pow(m66,5)*pow(RN1e,3)*pow(RN3e,3)
          - 25./1458.*m11*pow(m33,-1)*pow(m44,4)*pow(m66,4)*pow(RN1e,3)
         *pow(RN3e,3) - 5./648.*m11*pow(m33,-1)*pow(m44,5)*pow(m66,3)*
         pow(RN1e,3)*pow(RN3e,3) - 7./3888.*m11*pow(m33,-1)*pow(m44,6)*
         pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) - 1./5832.*m11*pow(m33,-1)*
         pow(m44,7)*m66*pow(RN1e,3)*pow(RN3e,3) + 1./540.*m11*RD346e*
         RS3461e + 1./72.*m11*RD346e*RC346e + 2./729.*m11*pow(m66,7)*
         pow(RN1e,3)*pow(RN3e,3) + 40./2187.*m11*m44*pow(m66,6)*pow(
         RN1e,3)*pow(RN3e,3) + 157./2187.*m11*pow(m44,2)*pow(m66,5)*
         pow(RN1e,3)*pow(RN3e,3) + 235./2187.*m11*pow(m44,3)*pow(m66,4)
         *pow(RN1e,3)*pow(RN3e,3) + 1315./17496.*m11*pow(m44,4)*pow(
         m66,3)*pow(RN1e,3)*pow(RN3e,3) + 227./8748.*m11*pow(m44,5)*
         pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) + 35./8748.*m11*pow(m44,6)*
         m66*pow(RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + Ab(m33,mu2)*Bb(m33,mu2) * ( 1./5832.*m11*pow(
         m44,7)*pow(RN1e,3)*pow(RN3e,3) + 1./27.*m11*m33*RC346e*RS3461e
          - 2./27.*m11*m33*RD346e*RN3e*RSe6613 - 2./27.*m11*m33*RN13*
         RD346e*RS1463e - 13./2187.*m11*m33*pow(m66,6)*pow(RN1e,3)*pow(
         RN3e,3) + 53./4374.*m11*m33*m44*pow(m66,5)*pow(RN1e,3)*pow(
         RN3e,3) - 533./8748.*m11*m33*pow(m44,2)*pow(m66,4)*pow(RN1e,3)
         *pow(RN3e,3) - 785./5832.*m11*m33*pow(m44,3)*pow(m66,3)*pow(
         RN1e,3)*pow(RN3e,3) - 1571./17496.*m11*m33*pow(m44,4)*pow(
         m66,2)*pow(RN1e,3)*pow(RN3e,3) - 847./34992.*m11*m33*pow(
         m44,5)*m66*pow(RN1e,3)*pow(RN3e,3) - 77./34992.*m11*m33*pow(
         m44,6)*pow(RN1e,3)*pow(RN3e,3) - 277./4374.*m11*pow(m33,2)*
         pow(m66,5)*pow(RN1e,3)*pow(RN3e,3) - 1225./4374.*m11*pow(
         m33,2)*m44*pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) - 4123./17496.*
         m11*pow(m33,2)*pow(m44,2)*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3)
          - 73./2187.*m11*pow(m33,2)*pow(m44,3)*pow(m66,2)*pow(RN1e,3)*
         pow(RN3e,3) );

      M13p6x22R +=  + Ab(m33,mu2)*Bb(m33,mu2) * ( 809./34992.*m11*pow(
         m33,2)*pow(m44,4)*m66*pow(RN1e,3)*pow(RN3e,3) + 209./34992.*
         m11*pow(m33,2)*pow(m44,5)*pow(RN1e,3)*pow(RN3e,3) + 731./2916.
         *m11*pow(m33,3)*pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) + 899./1458.
         *m11*pow(m33,3)*m44*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) + 1801./
         3888.*m11*pow(m33,3)*pow(m44,2)*pow(m66,2)*pow(RN1e,3)*pow(
         RN3e,3) + 1471./11664.*m11*pow(m33,3)*pow(m44,3)*m66*pow(
         RN1e,3)*pow(RN3e,3) + 101./11664.*m11*pow(m33,3)*pow(m44,4)*
         pow(RN1e,3)*pow(RN3e,3) - 305./972.*m11*pow(m33,4)*pow(m66,3)*
         pow(RN1e,3)*pow(RN3e,3) - 40./81.*m11*pow(m33,4)*m44*pow(
         m66,2)*pow(RN1e,3)*pow(RN3e,3) - 335./1296.*m11*pow(m33,4)*
         pow(m44,2)*m66*pow(RN1e,3)*pow(RN3e,3) - 175./3888.*m11*pow(
         m33,4)*pow(m44,3)*pow(RN1e,3)*pow(RN3e,3) + 7./54.*m11*pow(
         m33,5)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) + 7./54.*m11*pow(
         m33,5)*m44*m66*pow(RN1e,3)*pow(RN3e,3) + 7./216.*m11*pow(
         m33,5)*pow(m44,2)*pow(RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + Ab(m33,mu2)*Bb(m33,mu2) * ( 17./2187.*pow(m11,2)*
         pow(m33,-1)*m44*pow(m66,6)*pow(RN1e,3)*pow(RN3e,3) + 83./2187.
         *pow(m11,2)*pow(m33,-1)*pow(m44,2)*pow(m66,5)*pow(RN1e,3)*pow(
         RN3e,3) + 247./4374.*pow(m11,2)*pow(m33,-1)*pow(m44,3)*pow(
         m66,4)*pow(RN1e,3)*pow(RN3e,3) + 82./2187.*pow(m11,2)*pow(
         m33,-1)*pow(m44,4)*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) + 409./
         34992.*pow(m11,2)*pow(m33,-1)*pow(m44,5)*pow(m66,2)*pow(
         RN1e,3)*pow(RN3e,3) + 49./34992.*pow(m11,2)*pow(m33,-1)*pow(
         m44,6)*m66*pow(RN1e,3)*pow(RN3e,3) - 17./2187.*pow(m11,2)*pow(
         m66,6)*pow(RN1e,3)*pow(RN3e,3) - 41./4374.*pow(m11,2)*m44*pow(
         m66,5)*pow(RN1e,3)*pow(RN3e,3) - 919./8748.*pow(m11,2)*pow(
         m44,2)*pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) - 959./5832.*pow(
         m11,2)*pow(m44,3)*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) - 409./
         4374.*pow(m11,2)*pow(m44,4)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3)
          - 743./34992.*pow(m11,2)*pow(m44,5)*m66*pow(RN1e,3)*pow(
         RN3e,3) );

      M13p6x22R +=  + Ab(m33,mu2)*Bb(m33,mu2) * (  - 49./34992.*pow(
         m11,2)*pow(m44,6)*pow(RN1e,3)*pow(RN3e,3) - 125./4374.*pow(
         m11,2)*m33*pow(m66,5)*pow(RN1e,3)*pow(RN3e,3) - 1031./4374.*
         pow(m11,2)*m33*m44*pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) - 1913./
         17496.*pow(m11,2)*m33*pow(m44,2)*pow(m66,3)*pow(RN1e,3)*pow(
         RN3e,3) + 799./8748.*pow(m11,2)*m33*pow(m44,3)*pow(m66,2)*pow(
         RN1e,3)*pow(RN3e,3) + 1127./17496.*pow(m11,2)*m33*pow(m44,4)*
         m66*pow(RN1e,3)*pow(RN3e,3) + 167./17496.*pow(m11,2)*m33*pow(
         m44,5)*pow(RN1e,3)*pow(RN3e,3) + 829./2916.*pow(m11,2)*pow(
         m33,2)*pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) + 5119./5832.*pow(
         m11,2)*pow(m33,2)*m44*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) + 
         1105./1944.*pow(m11,2)*pow(m33,2)*pow(m44,2)*pow(m66,2)*pow(
         RN1e,3)*pow(RN3e,3) + 487./5832.*pow(m11,2)*pow(m33,2)*pow(
         m44,3)*m66*pow(RN1e,3)*pow(RN3e,3) - 49./5832.*pow(m11,2)*pow(
         m33,2)*pow(m44,4)*pow(RN1e,3)*pow(RN3e,3) - 1247./1944.*pow(
         m11,2)*pow(m33,3)*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + Ab(m33,mu2)*Bb(m33,mu2) * (  - 1511./1296.*pow(
         m11,2)*pow(m33,3)*m44*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) - 721.
         /1296.*pow(m11,2)*pow(m33,3)*pow(m44,2)*m66*pow(RN1e,3)*pow(
         RN3e,3) - 65./972.*pow(m11,2)*pow(m33,3)*pow(m44,3)*pow(
         RN1e,3)*pow(RN3e,3) + 127./216.*pow(m11,2)*pow(m33,4)*pow(
         m66,2)*pow(RN1e,3)*pow(RN3e,3) + 269./432.*pow(m11,2)*pow(
         m33,4)*m44*m66*pow(RN1e,3)*pow(RN3e,3) + 71./432.*pow(m11,2)*
         pow(m33,4)*pow(m44,2)*pow(RN1e,3)*pow(RN3e,3) - 7./36.*pow(
         m11,2)*pow(m33,5)*m66*pow(RN1e,3)*pow(RN3e,3) - 7./72.*pow(
         m11,2)*pow(m33,5)*m44*pow(RN1e,3)*pow(RN3e,3) + 1./486.*pow(
         m11,3)*pow(m33,-1)*m44*pow(m66,5)*pow(RN1e,3)*pow(RN3e,3) - 23.
         /972.*pow(m11,3)*pow(m33,-1)*pow(m44,2)*pow(m66,4)*pow(RN1e,3)
         *pow(RN3e,3) - 25./648.*pow(m11,3)*pow(m33,-1)*pow(m44,3)*pow(
         m66,3)*pow(RN1e,3)*pow(RN3e,3) - 77./3888.*pow(m11,3)*pow(
         m33,-1)*pow(m44,4)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) - 13./
         3888.*pow(m11,3)*pow(m33,-1)*pow(m44,5)*m66*pow(RN1e,3)*pow(
         RN3e,3) );

      M13p6x22R +=  + Ab(m33,mu2)*Bb(m33,mu2) * (  - 1./486.*pow(m11,3)
         *pow(m66,5)*pow(RN1e,3)*pow(RN3e,3) - 5./54.*pow(m11,3)*m44*
         pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) - 23./1944.*pow(m11,3)*pow(
         m44,2)*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) + 17./243.*pow(
         m11,3)*pow(m44,3)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) + 43./
         1296.*pow(m11,3)*pow(m44,4)*m66*pow(RN1e,3)*pow(RN3e,3) + 13./
         3888.*pow(m11,3)*pow(m44,5)*pow(RN1e,3)*pow(RN3e,3) + 113./972.
         *pow(m11,3)*m33*pow(m66,4)*pow(RN1e,3)*pow(RN3e,3) + 1043./
         1944.*pow(m11,3)*m33*m44*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) + 
         7./24.*pow(m11,3)*m33*pow(m44,2)*pow(m66,2)*pow(RN1e,3)*pow(
         RN3e,3) - 1./1944.*pow(m11,3)*m33*pow(m44,3)*m66*pow(RN1e,3)*
         pow(RN3e,3) - 13./972.*pow(m11,3)*m33*pow(m44,4)*pow(RN1e,3)*
         pow(RN3e,3) - 35./72.*pow(m11,3)*pow(m33,2)*pow(m66,3)*pow(
         RN1e,3)*pow(RN3e,3) - 88./81.*pow(m11,3)*pow(m33,2)*m44*pow(
         m66,2)*pow(RN1e,3)*pow(RN3e,3) - 293./648.*pow(m11,3)*pow(
         m33,2)*pow(m44,2)*m66*pow(RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + Ab(m33,mu2)*Bb(m33,mu2) * (  - 5./162.*pow(m11,3)
         *pow(m33,2)*pow(m44,3)*pow(RN1e,3)*pow(RN3e,3) + 965./1296.*
         pow(m11,3)*pow(m33,3)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) + 
         1157./1296.*pow(m11,3)*pow(m33,3)*m44*m66*pow(RN1e,3)*pow(
         RN3e,3) + 127./648.*pow(m11,3)*pow(m33,3)*pow(m44,2)*pow(
         RN1e,3)*pow(RN3e,3) - 203./432.*pow(m11,3)*pow(m33,4)*m66*pow(
         RN1e,3)*pow(RN3e,3) - 109./432.*pow(m11,3)*pow(m33,4)*m44*pow(
         RN1e,3)*pow(RN3e,3) + 7./72.*pow(m11,3)*pow(m33,5)*pow(RN1e,3)
         *pow(RN3e,3) - 7./324.*pow(m11,4)*pow(m33,-1)*m44*pow(m66,4)*
         pow(RN1e,3)*pow(RN3e,3) - 1./54.*pow(m11,4)*pow(m33,-1)*pow(
         m44,2)*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) - 1./432.*pow(m11,4)
         *pow(m33,-1)*pow(m44,3)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) + 1.
         /1296.*pow(m11,4)*pow(m33,-1)*pow(m44,4)*m66*pow(RN1e,3)*pow(
         RN3e,3) + 7./324.*pow(m11,4)*pow(m66,4)*pow(RN1e,3)*pow(
         RN3e,3) + 53./324.*pow(m11,4)*m44*pow(m66,3)*pow(RN1e,3)*pow(
         RN3e,3) );

      M13p6x22R +=  + Ab(m33,mu2)*Bb(m33,mu2) * ( 133./1296.*pow(m11,4)
         *pow(m44,2)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) + 5./432.*pow(
         m11,4)*pow(m44,3)*m66*pow(RN1e,3)*pow(RN3e,3) - 1./1296.*pow(
         m11,4)*pow(m44,4)*pow(RN1e,3)*pow(RN3e,3) - 47./324.*pow(
         m11,4)*m33*pow(m66,3)*pow(RN1e,3)*pow(RN3e,3) - 581./1296.*
         pow(m11,4)*m33*m44*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) - 227./
         1296.*pow(m11,4)*m33*pow(m44,2)*m66*pow(RN1e,3)*pow(RN3e,3) - 
         1./108.*pow(m11,4)*m33*pow(m44,3)*pow(RN1e,3)*pow(RN3e,3) + 
         451./1296.*pow(m11,4)*pow(m33,2)*pow(m66,2)*pow(RN1e,3)*pow(
         RN3e,3) + 673./1296.*pow(m11,4)*pow(m33,2)*m44*m66*pow(RN1e,3)
         *pow(RN3e,3) + 59./648.*pow(m11,4)*pow(m33,2)*pow(m44,2)*pow(
         RN1e,3)*pow(RN3e,3) - 77./216.*pow(m11,4)*pow(m33,3)*m66*pow(
         RN1e,3)*pow(RN3e,3) - 23./108.*pow(m11,4)*pow(m33,3)*m44*pow(
         RN1e,3)*pow(RN3e,3) + 19./144.*pow(m11,4)*pow(m33,4)*pow(
         RN1e,3)*pow(RN3e,3) + 5./324.*pow(m11,5)*pow(m33,-1)*m44*pow(
         m66,3)*pow(RN1e,3)*pow(RN3e,3) );

      M13p6x22R +=  + Ab(m33,mu2)*Bb(m33,mu2) * ( 5./324.*pow(m11,5)*
         pow(m33,-1)*pow(m44,2)*pow(m66,2)*pow(RN1e,3)*pow(RN3e,3) + 5./
         1296.*pow(m11,5)*pow(m33,-1)*pow(m44,3)*m66*pow(RN1e,3)*pow(
         RN3e,3) - 5./324.*pow(m11,5)*pow(m66,3)*pow(RN1e,3)*pow(
         RN3e,3) - 25./324.*pow(m11,5)*m44*pow(m66,2)*pow(RN1e,3)*pow(
         RN3e,3) - 55./1296.*pow(m11,5)*pow(m44,2)*m66*pow(RN1e,3)*pow(
         RN3e,3) - 5./1296.*pow(m11,5)*pow(m44,3)*pow(RN1e,3)*pow(
         RN3e,3) + 5./81.*pow(m11,5)*m33*pow(m66,2)*pow(RN1e,3)*pow(
         RN3e,3) + 155./1296.*pow(m11,5)*m33*m44*m66*pow(RN1e,3)*pow(
         RN3e,3) + 35./1296.*pow(m11,5)*m33*pow(m44,2)*pow(RN1e,3)*pow(
         RN3e,3) - 35./432.*pow(m11,5)*pow(m33,2)*m66*pow(RN1e,3)*pow(
         RN3e,3) - 25./432.*pow(m11,5)*pow(m33,2)*m44*pow(RN1e,3)*pow(
         RN3e,3) + 5./144.*pow(m11,5)*pow(m33,3)*pow(RN1e,3)*pow(
         RN3e,3) );

      M13p6x22R +=  + Ab(m34,mu2)*pi16 * (  - 1./24.*m33*RD146e + 1./48.
         *m33*me2 + 1./12.*m33*me2*RSe61 + 1./12.*m33*me2*RSe6611 + 1./
         8.*m33*m66 + 7./24.*m33*m44 + 1./48.*m33*m44*RS3461e + 1./48.*
         m33*m44*RS1463e - 1./24.*m33*m44*RSe6613 + 1./48.*m33*m44*
         RSe63 + 1./24.*m33*m44*RSe6633 + 1./48.*m33*m44*RS36e - 1./48.
         *m33*m44*RC346e + 1./48.*m33*m44*RSe61 + 1./24.*m33*m44*
         RSe6611 + 1./48.*m33*m44*RS16e - 1./48.*m33*m44*RC146e - 17./
         144.*pow(m33,2) + 1./48.*pow(m33,2)*RS3461e + 1./48.*pow(
         m33,2)*RS1463e - 1./24.*pow(m33,2)*RSe6613 + 1./48.*pow(m33,2)
         *RSe63 + 1./24.*pow(m33,2)*RSe6633 + 1./48.*pow(m33,2)*RS36e
          - 1./48.*pow(m33,2)*RC346e - 5./144.*pow(m33,2)*RSe61 - 1./72.
         *pow(m33,2)*RSe6611 - 5./144.*pow(m33,2)*RS16e + 1./144.*pow(
         m33,2)*RC146e - 1./24.*m11*RD146e + 1./48.*m11*me2 + 1./12.*
         m11*me2*RSe61 + 1./12.*m11*me2*RSe6611 + 1./8.*m11*m66 + 7./24.
         *m11*m44 + 1./48.*m11*m44*RS3461e + 1./48.*m11*m44*RS1463e - 1.
         /24.*m11*m44*RSe6613 );

      M13p6x22R +=  + Ab(m34,mu2)*pi16 * ( 1./48.*m11*m44*RSe63 + 1./24.
         *m11*m44*RSe6633 + 1./48.*m11*m44*RS36e - 1./48.*m11*m44*
         RC346e + 1./48.*m11*m44*RSe61 + 1./24.*m11*m44*RSe6611 + 1./48.
         *m11*m44*RS16e - 1./48.*m11*m44*RC146e - 1./144.*m11*m33 + 1./
         48.*m11*m33*RS3461e + 1./48.*m11*m33*RS1463e - 1./24.*m11*m33*
         RSe6613 + 1./48.*m11*m33*RSe63 + 1./24.*m11*m33*RSe6633 + 1./
         48.*m11*m33*RS36e - 1./48.*m11*m33*RC346e - 1./144.*m11*m33*
         RSe61 + 1./72.*m11*m33*RSe6611 + 11./144.*m11*m33*RS16e - 7./
         144.*m11*m33*RC146e + 1./9.*pow(m11,2) + 1./36.*pow(m11,2)*
         RSe61 + 1./36.*pow(m11,2)*RSe6611 + 1./9.*pow(m11,2)*RS16e - 1.
         /18.*pow(m11,2)*RC146e );

      M13p6x22R +=  + pow(Ab(m34,mu2),2) * (  - 1./90.*pow(m34,-1)*pow(
         me2,2) - 17./135.*pow(m34,-1)*pow(me2,2)*RSe6613 + 7./270.*
         pow(m34,-1)*pow(me2,2)*RSe63 + 4./45.*pow(m34,-1)*pow(me2,2)*
         RSe6633 - 19./270.*pow(m34,-1)*pow(me2,2)*RSe61 - 1./135.*pow(
         m34,-1)*pow(me2,2)*RSe6611 - 1./216.*pow(m34,-1)*m44*RD346e - 
         1./216.*pow(m34,-1)*m44*RD146e + 11./540.*pow(m34,-1)*m44*me2
          + 1./54.*pow(m34,-1)*m44*me2*RSe6613 + 2./135.*pow(m34,-1)*
         m44*me2*RSe63 + 1./108.*pow(m34,-1)*m44*me2*RSe6633 + 2./45.*
         pow(m34,-1)*m44*me2*RSe61 + 1./108.*pow(m34,-1)*m44*me2*
         RSe6611 - 1./108.*pow(m34,-1)*pow(m44,2) - 1./108.*pow(m34,-1)
         *pow(m44,2)*RS3461e - 1./108.*pow(m34,-1)*pow(m44,2)*RS1463e
          + 1./54.*pow(m34,-1)*pow(m44,2)*RSe6613 + 1./108.*pow(m34,-1)
         *pow(m44,2)*RSe6633 - 1./216.*pow(m34,-1)*pow(m44,2)*RC346e + 
         1./108.*pow(m34,-1)*pow(m44,2)*RSe6611 - 1./216.*pow(m34,-1)*
         pow(m44,2)*RC146e - 2./45.*m33*pow(m34,-1)*RD346e + 1./24.*m33
         *pow(m34,-1)*RD146e );

      M13p6x22R +=  + pow(Ab(m34,mu2),2) * (  - 1./540.*m33*pow(m34,-1)
         *me2*RSe6613 - 1./120.*m33*pow(m34,-1)*me2*RSe63 - 4./45.*m33*
         pow(m34,-1)*me2*RSe6633 - 23./216.*m33*pow(m34,-1)*me2*RSe61
          - 1./12.*m33*pow(m34,-1)*me2*RSe6611 - 31./432.*m33*pow(
         m34,-1)*m66 - 4./27.*m33*pow(m34,-1)*m44 - 13./432.*m33*pow(
         m34,-1)*m44*RS3461e - 1./48.*m33*pow(m34,-1)*m44*RS1463e + 1./
         24.*m33*pow(m34,-1)*m44*RSe6613 - 1./144.*m33*pow(m34,-1)*m44*
         RSe63 - 1./108.*m33*pow(m34,-1)*m44*RSe6633 + 17./2160.*m33*
         pow(m34,-1)*m44*RS36e - 1./16.*m33*pow(m34,-1)*m44*RSe61 + 1./
         108.*m33*pow(m34,-1)*m44*RSe6611 - 1./16.*m33*pow(m34,-1)*m44*
         RS16e - 1./216.*m33*pow(m34,-1)*m44*RC146e + 1./144.*pow(
         m33,2)*pow(m34,-1) + 23./360.*pow(m33,2)*pow(m34,-1)*RS3461e
          + 19./1080.*pow(m33,2)*pow(m34,-1)*RS36e - 4./27.*pow(m33,2)*
         pow(m34,-1)*RSe61 - 4./27.*pow(m33,2)*pow(m34,-1)*RS16e - 1./
         216.*m11*pow(m34,-1)*RD346e + 11./270.*m11*pow(m34,-1)*RD146e
          + 1./108.*m11*pow(m34,-1)*me2 );

      M13p6x22R +=  + pow(Ab(m34,mu2),2) * (  - 11./540.*m11*pow(
         m34,-1)*me2*RSe6613 - 5./216.*m11*pow(m34,-1)*me2*RSe63 + 1./
         108.*m11*pow(m34,-1)*me2*RSe6633 - 1./40.*m11*pow(m34,-1)*me2*
         RSe61 - 1./15.*m11*pow(m34,-1)*me2*RSe6611 - 31./432.*m11*pow(
         m34,-1)*m66 - 17./108.*m11*pow(m34,-1)*m44 - 13./432.*m11*pow(
         m34,-1)*m44*RS3461e - 17./432.*m11*pow(m34,-1)*m44*RS1463e + 
         13./216.*m11*pow(m34,-1)*m44*RSe6613 + 5./432.*m11*pow(m34,-1)
         *m44*RSe63 + 1./54.*m11*pow(m34,-1)*m44*RSe6633 + 5./432.*m11*
         pow(m34,-1)*m44*RS36e - 1./108.*m11*pow(m34,-1)*m44*RC346e - 
         19./432.*m11*pow(m34,-1)*m44*RSe61 + 1./2160.*m11*pow(m34,-1)*
         m44*RS16e - 1./216.*m11*pow(m34,-1)*m44*RC146e - 59./432.*m11*
         m33*pow(m34,-1) - 23./2160.*m11*m33*pow(m34,-1)*RS3461e - 43./
         2160.*m11*m33*pow(m34,-1)*RS1463e + 1./24.*m11*m33*pow(m34,-1)
         *RSe6613 - 1./144.*m11*m33*pow(m34,-1)*RSe63 - 1./108.*m11*m33
         *pow(m34,-1)*RSe6633 - 13./432.*m11*m33*pow(m34,-1)*RS36e - 1./
         16.*m11*m33*pow(m34,-1)*RSe61 );

      M13p6x22R +=  + pow(Ab(m34,mu2),2) * ( 1./108.*m11*m33*pow(
         m34,-1)*RSe6611 - 73./432.*m11*m33*pow(m34,-1)*RS16e + 1./27.*
         m11*m33*pow(m34,-1)*RC146e - 31./216.*pow(m11,2)*pow(m34,-1)
          - 1./48.*pow(m11,2)*pow(m34,-1)*RS3461e + 113./2160.*pow(
         m11,2)*pow(m34,-1)*RS1463e + 1./24.*pow(m11,2)*pow(m34,-1)*
         RSe6613 + 5./432.*pow(m11,2)*pow(m34,-1)*RSe63 + 1./108.*pow(
         m11,2)*pow(m34,-1)*RSe6633 + 5./432.*pow(m11,2)*pow(m34,-1)*
         RS36e - 1./216.*pow(m11,2)*pow(m34,-1)*RC346e - 1./144.*pow(
         m11,2)*pow(m34,-1)*RSe61 - 1./108.*pow(m11,2)*pow(m34,-1)*
         RSe6611 - 221./2160.*pow(m11,2)*pow(m34,-1)*RS16e + 1./24.*
         pow(m11,2)*pow(m34,-1)*RC146e );

      M13p6x22R +=  + Ab(m34,mu2)*Ab(me2,mu2) * (  - 1./45.*me2 - 44./
         135.*me2*RSe6613 + 7./135.*me2*RSe63 + 29./135.*me2*RSe6633 - 
         19./135.*me2*RSe61 + 1./45.*me2*RSe6611 + 1./45.*m44 - 11./135.
         *m44*RSe63 + 17./135.*m44*RSe61 - 2./27.*m44*me2*RN3e*RSe63 + 
         2./27.*m44*me2*RSe61*RN3e + 19./270.*m33*RSe6613 - 49./540.*
         m33*RSe63 - 1./15.*m33*RSe6633 + 1./108.*m33*RSe61 - 1./6.*m33
         *RSe6611 + 4./27.*m33*me2*RN3e*RSe6613 - 2./27.*m33*me2*RN3e*
         RSe63 + 4./27.*m33*me2*RN3e*RSe6633 + 2./9.*m33*me2*RSe61*RN3e
          + 2./27.*m33*m44*RN3e*RSe63 + 2./27.*m33*m44*RSe61*RN3e - 4./
         27.*pow(m33,2)*RN3e*RSe6613 + 2./27.*pow(m33,2)*RN3e*RSe63 - 4.
         /27.*pow(m33,2)*RN3e*RSe6633 + 2./27.*pow(m33,2)*RSe61*RN3e - 
         1./270.*m11*RSe6613 - 1./12.*m11*RSe63 - 7./540.*m11*RSe61 - 
         17./90.*m11*RSe6611 + 4./27.*m11*me2*RN3e*RSe6613 + 2./27.*m11
         *me2*RSe61*RN3e + 2./27.*m11*m44*RSe61*RN3e - 4./27.*m11*m33*
         RN3e*RSe6613 + 2./27.*m11*m33*RSe61*RN3e );

      M13p6x22R +=  + Ab(m34,mu2)*Bb(m11,mu2) * ( 1./12.*m33*RD146e + 1.
         /12.*m11*RD146e );

      M13p6x22R +=  + Ab(m36,mu2)*pi16 * (  - 1./48.*m33*RD146e + 1./24.
         *m33*me2 - 1./12.*m33*me2*RSe61 + 1./24.*m33*me2*RSe6611 + 5./
         96.*m33*m66 + 1./96.*m33*m66*RS3461e + 1./96.*m33*m66*RS1463e
          - 1./48.*m33*m66*RSe6613 - 1./48.*m33*m66*RSe63 + 1./48.*m33*
         m66*RSe6633 + 1./96.*m33*m66*RS34e - 1./96.*m33*m66*RC346e - 1.
         /48.*m33*m66*RSe61 + 1./48.*m33*m66*RSe6611 + 1./96.*m33*m66*
         RS14e - 1./96.*m33*m66*RC146e + 1./8.*m33*m44 - 17./288.*pow(
         m33,2) + 1./96.*pow(m33,2)*RS3461e + 1./96.*pow(m33,2)*RS1463e
          - 1./48.*pow(m33,2)*RSe6613 - 1./48.*pow(m33,2)*RSe63 + 1./48.
         *pow(m33,2)*RSe6633 + 1./96.*pow(m33,2)*RS34e - 1./96.*pow(
         m33,2)*RC346e + 5./144.*pow(m33,2)*RSe61 - 1./144.*pow(m33,2)*
         RSe6611 - 5./288.*pow(m33,2)*RS14e + 1./288.*pow(m33,2)*RC146e
          - 1./48.*m11*RD146e + 1./24.*m11*me2 - 1./12.*m11*me2*RSe61
          + 1./24.*m11*me2*RSe6611 + 5./96.*m11*m66 + 1./96.*m11*m66*
         RS3461e + 1./96.*m11*m66*RS1463e - 1./48.*m11*m66*RSe6613 - 1./
         48.*m11*m66*RSe63 );

      M13p6x22R +=  + Ab(m36,mu2)*pi16 * ( 1./48.*m11*m66*RSe6633 + 1./
         96.*m11*m66*RS34e - 1./96.*m11*m66*RC346e - 1./48.*m11*m66*
         RSe61 + 1./48.*m11*m66*RSe6611 + 1./96.*m11*m66*RS14e - 1./96.
         *m11*m66*RC146e + 1./8.*m11*m44 - 1./288.*m11*m33 + 1./96.*m11
         *m33*RS3461e + 1./96.*m11*m33*RS1463e - 1./48.*m11*m33*RSe6613
          - 1./48.*m11*m33*RSe63 + 1./48.*m11*m33*RSe6633 + 1./96.*m11*
         m33*RS34e - 1./96.*m11*m33*RC346e + 1./144.*m11*m33*RSe61 + 1./
         144.*m11*m33*RSe6611 + 11./288.*m11*m33*RS14e - 7./288.*m11*
         m33*RC146e + 1./18.*pow(m11,2) - 1./36.*pow(m11,2)*RSe61 + 1./
         72.*pow(m11,2)*RSe6611 + 1./18.*pow(m11,2)*RS14e - 1./36.*pow(
         m11,2)*RC146e );

      M13p6x22R +=  + pow(Ab(m36,mu2),2) * (  - 1./45.*pow(m36,-1)*pow(
         me2,2) - 17./270.*pow(m36,-1)*pow(me2,2)*RSe6613 - 7./270.*
         pow(m36,-1)*pow(me2,2)*RSe63 + 2./45.*pow(m36,-1)*pow(me2,2)*
         RSe6633 + 19./270.*pow(m36,-1)*pow(me2,2)*RSe61 - 1./270.*pow(
         m36,-1)*pow(me2,2)*RSe6611 - 1./432.*pow(m36,-1)*m66*RD346e - 
         1./432.*pow(m36,-1)*m66*RD146e + 11./270.*pow(m36,-1)*m66*me2
          + 1./108.*pow(m36,-1)*m66*me2*RSe6613 - 2./135.*pow(m36,-1)*
         m66*me2*RSe63 + 1./216.*pow(m36,-1)*m66*me2*RSe6633 - 2./45.*
         pow(m36,-1)*m66*me2*RSe61 + 1./216.*pow(m36,-1)*m66*me2*
         RSe6611 - 1./54.*pow(m36,-1)*pow(m66,2) - 1./216.*pow(m36,-1)*
         pow(m66,2)*RS3461e - 1./216.*pow(m36,-1)*pow(m66,2)*RS1463e + 
         1./108.*pow(m36,-1)*pow(m66,2)*RSe6613 + 1./216.*pow(m36,-1)*
         pow(m66,2)*RSe6633 - 1./432.*pow(m36,-1)*pow(m66,2)*RC346e + 1.
         /216.*pow(m36,-1)*pow(m66,2)*RSe6611 - 1./432.*pow(m36,-1)*
         pow(m66,2)*RC146e - 1./45.*m33*pow(m36,-1)*RD346e + 1./48.*m33
         *pow(m36,-1)*RD146e );

      M13p6x22R +=  + pow(Ab(m36,mu2),2) * (  - 1./1080.*m33*pow(
         m36,-1)*me2*RSe6613 + 1./120.*m33*pow(m36,-1)*me2*RSe63 - 2./
         45.*m33*pow(m36,-1)*me2*RSe6633 + 23./216.*m33*pow(m36,-1)*me2
         *RSe61 - 1./24.*m33*pow(m36,-1)*me2*RSe6611 - 11./288.*m33*
         pow(m36,-1)*m66 - 13./864.*m33*pow(m36,-1)*m66*RS3461e - 1./96.
         *m33*pow(m36,-1)*m66*RS1463e + 1./48.*m33*pow(m36,-1)*m66*
         RSe6613 + 1./144.*m33*pow(m36,-1)*m66*RSe63 - 1./216.*m33*pow(
         m36,-1)*m66*RSe6633 + 17./4320.*m33*pow(m36,-1)*m66*RS34e + 1./
         16.*m33*pow(m36,-1)*m66*RSe61 + 1./216.*m33*pow(m36,-1)*m66*
         RSe6611 - 1./32.*m33*pow(m36,-1)*m66*RS14e - 1./432.*m33*pow(
         m36,-1)*m66*RC146e - 31./432.*m33*pow(m36,-1)*m44 + 1./288.*
         pow(m33,2)*pow(m36,-1) + 23./720.*pow(m33,2)*pow(m36,-1)*
         RS3461e + 19./2160.*pow(m33,2)*pow(m36,-1)*RS34e + 4./27.*pow(
         m33,2)*pow(m36,-1)*RSe61 - 2./27.*pow(m33,2)*pow(m36,-1)*RS14e
          - 1./432.*m11*pow(m36,-1)*RD346e + 11./540.*m11*pow(m36,-1)*
         RD146e );

      M13p6x22R +=  + pow(Ab(m36,mu2),2) * ( 1./54.*m11*pow(m36,-1)*me2
          - 11./1080.*m11*pow(m36,-1)*me2*RSe6613 + 5./216.*m11*pow(
         m36,-1)*me2*RSe63 + 1./216.*m11*pow(m36,-1)*me2*RSe6633 + 1./
         40.*m11*pow(m36,-1)*me2*RSe61 - 1./30.*m11*pow(m36,-1)*me2*
         RSe6611 - 49./864.*m11*pow(m36,-1)*m66 - 13./864.*m11*pow(
         m36,-1)*m66*RS3461e - 17./864.*m11*pow(m36,-1)*m66*RS1463e + 
         13./432.*m11*pow(m36,-1)*m66*RSe6613 - 5./432.*m11*pow(m36,-1)
         *m66*RSe63 + 1./108.*m11*pow(m36,-1)*m66*RSe6633 + 5./864.*m11
         *pow(m36,-1)*m66*RS34e - 1./216.*m11*pow(m36,-1)*m66*RC346e + 
         19./432.*m11*pow(m36,-1)*m66*RSe61 + 1./4320.*m11*pow(m36,-1)*
         m66*RS14e - 1./432.*m11*pow(m36,-1)*m66*RC146e - 31./432.*m11*
         pow(m36,-1)*m44 - 59./864.*m11*m33*pow(m36,-1) - 23./4320.*m11
         *m33*pow(m36,-1)*RS3461e - 43./4320.*m11*m33*pow(m36,-1)*
         RS1463e + 1./48.*m11*m33*pow(m36,-1)*RSe6613 + 1./144.*m11*m33
         *pow(m36,-1)*RSe63 - 1./216.*m11*m33*pow(m36,-1)*RSe6633 - 13./
         864.*m11*m33*pow(m36,-1)*RS34e );

      M13p6x22R +=  + pow(Ab(m36,mu2),2) * ( 1./16.*m11*m33*pow(m36,-1)
         *RSe61 + 1./216.*m11*m33*pow(m36,-1)*RSe6611 - 73./864.*m11*
         m33*pow(m36,-1)*RS14e + 1./54.*m11*m33*pow(m36,-1)*RC146e - 31.
         /432.*pow(m11,2)*pow(m36,-1) - 1./96.*pow(m11,2)*pow(m36,-1)*
         RS3461e + 113./4320.*pow(m11,2)*pow(m36,-1)*RS1463e + 1./48.*
         pow(m11,2)*pow(m36,-1)*RSe6613 - 5./432.*pow(m11,2)*pow(
         m36,-1)*RSe63 + 1./216.*pow(m11,2)*pow(m36,-1)*RSe6633 + 5./
         864.*pow(m11,2)*pow(m36,-1)*RS34e - 1./432.*pow(m11,2)*pow(
         m36,-1)*RC346e + 1./144.*pow(m11,2)*pow(m36,-1)*RSe61 - 1./216.
         *pow(m11,2)*pow(m36,-1)*RSe6611 - 221./4320.*pow(m11,2)*pow(
         m36,-1)*RS14e + 1./48.*pow(m11,2)*pow(m36,-1)*RC146e );

      M13p6x22R +=  + Ab(m36,mu2)*Ab(me2,mu2) * (  - 2./45.*me2 - 22./
         135.*me2*RSe6613 - 7./135.*me2*RSe63 + 29./270.*me2*RSe6633 + 
         19./135.*me2*RSe61 + 1./90.*me2*RSe6611 + 2./45.*m66 + 11./135.
         *m66*RSe63 - 17./135.*m66*RSe61 + 2./27.*m66*me2*RN3e*RSe63 - 
         2./27.*m66*me2*RSe61*RN3e + 19./540.*m33*RSe6613 + 49./540.*
         m33*RSe63 - 1./30.*m33*RSe6633 - 1./108.*m33*RSe61 - 1./12.*
         m33*RSe6611 + 2./27.*m33*me2*RN3e*RSe6613 + 2./27.*m33*me2*
         RN3e*RSe63 + 2./27.*m33*me2*RN3e*RSe6633 - 2./9.*m33*me2*RSe61
         *RN3e - 2./27.*m33*m66*RN3e*RSe63 - 2./27.*m33*m66*RSe61*RN3e
          - 2./27.*pow(m33,2)*RN3e*RSe6613 - 2./27.*pow(m33,2)*RN3e*
         RSe63 - 2./27.*pow(m33,2)*RN3e*RSe6633 - 2./27.*pow(m33,2)*
         RSe61*RN3e - 1./540.*m11*RSe6613 + 1./12.*m11*RSe63 + 7./540.*
         m11*RSe61 - 17./180.*m11*RSe6611 + 2./27.*m11*me2*RN3e*RSe6613
          - 2./27.*m11*me2*RSe61*RN3e - 2./27.*m11*m66*RSe61*RN3e - 2./
         27.*m11*m33*RN3e*RSe6613 - 2./27.*m11*m33*RSe61*RN3e );

      M13p6x22R +=  + Ab(m36,mu2)*Bb(m11,mu2) * ( 1./24.*m33*RD146e + 1.
         /24.*m11*RD146e );

      M13p6x22R +=  + pow(Ab(m44,mu2),2) * (  - 1./18.*me2*pow(RSe63,2)
          - 1./18.*me2*RS36e*RSe63 + 1./9.*me2*RSe61*RSe63 + 1./18.*me2
         *RSe61*RS36e - 1./18.*me2*pow(RSe61,2) + 1./18.*me2*RS16e*
         RSe63 - 1./18.*me2*RS16e*RSe61 - 13./48.*m33 + 1./36.*m33*pow(
         RSe63,2) - 1./36.*m33*pow(RS36e,2) + 1./18.*m33*RSe61*RSe63 + 
         1./9.*m33*RSe61*RS36e + 1./18.*m33*RS16e*RSe63 + 1./9.*m33*
         RS16e*RS36e - 13./48.*m11 + 1./18.*m11*RSe61*RSe63 + 1./18.*
         m11*RSe61*RS36e + 1./36.*m11*pow(RSe61,2) + 1./9.*m11*RS16e*
         RSe63 + 1./9.*m11*RS16e*RS36e - 1./36.*m11*pow(RS16e,2) );

      M13p6x22R +=  + Ab(m44,mu2)*Ab(me2,mu2) * (  - 1./18.*m44*pow(
         RSe63,2) + 1./9.*m44*RSe61*RSe63 - 1./18.*m44*pow(RSe61,2) + 1.
         /9.*m44*me2*RN3e*RS36e*RSe63 - 1./9.*m44*me2*RSe61*RN3e*RS36e
          - 1./9.*m44*me2*RN1e*RS16e*RSe63 + 1./9.*m44*me2*RN1e*RS16e*
         RSe61 - 1./9.*m33*m44*RN3e*RS36e*RSe63 - 1./9.*m33*m44*RSe61*
         RN3e*RS36e - 1./9.*m33*m44*RN1e*RS16e*RSe63 - 1./9.*m11*m44*
         RSe61*RN3e*RS36e - 1./9.*m11*m44*RN1e*RS16e*RSe63 - 1./9.*m11*
         m44*RN1e*RS16e*RSe61 );

      M13p6x22R +=  + Ab(m44,mu2)*Bb(me2,mu2) * (  - 1./18.*m44*me2*
         pow(RSe63,2) + 1./9.*m44*me2*RSe61*RSe63 - 1./18.*m44*me2*pow(
         RSe61,2) + 1./18.*m33*m44*pow(RSe63,2) + 1./9.*m33*m44*RSe61*
         RSe63 + 1./9.*m11*m44*RSe61*RSe63 + 1./18.*m11*m44*pow(
         RSe61,2) );

      M13p6x22R +=  + pow(Ab(m46,mu2),2) * ( 1./6.*pow(m46,-1)*pow(
         me2,2)*pow(RSe63,2) + 1./27.*pow(m46,-1)*pow(me2,2)*RS36e*
         RSe63 - 1./27.*pow(m46,-1)*pow(me2,2)*RS34e*RSe63 - 1./3.*pow(
         m46,-1)*pow(me2,2)*RSe61*RSe63 - 1./27.*pow(m46,-1)*pow(me2,2)
         *RSe61*RS36e + 1./27.*pow(m46,-1)*pow(me2,2)*RSe61*RS34e + 1./
         6.*pow(m46,-1)*pow(me2,2)*pow(RSe61,2) - 1./27.*pow(m46,-1)*
         pow(me2,2)*RS16e*RSe63 + 1./27.*pow(m46,-1)*pow(me2,2)*RS16e*
         RSe61 + 1./27.*pow(m46,-1)*pow(me2,2)*RS14e*RSe63 - 1./27.*
         pow(m46,-1)*pow(me2,2)*RS14e*RSe61 + 2./81.*pow(m46,-1)*m66*
         me2*RS36e*RSe63 + 1./81.*pow(m46,-1)*m66*me2*RS34e*RSe63 - 2./
         81.*pow(m46,-1)*m66*me2*RSe61*RS36e - 1./81.*pow(m46,-1)*m66*
         me2*RSe61*RS34e - 2./81.*pow(m46,-1)*m66*me2*RS16e*RSe63 + 2./
         81.*pow(m46,-1)*m66*me2*RS16e*RSe61 - 1./81.*pow(m46,-1)*m66*
         me2*RS14e*RSe63 + 1./81.*pow(m46,-1)*m66*me2*RS14e*RSe61 + 1./
         27.*m44*pow(m46,-1)*me2*pow(RSe63,2) + 1./81.*m44*pow(m46,-1)*
         me2*RS36e*RSe63 );

      M13p6x22R +=  + pow(Ab(m46,mu2),2) * (  - 1./81.*m44*pow(m46,-1)*
         me2*RS34e*RSe63 - 2./27.*m44*pow(m46,-1)*me2*RSe61*RSe63 - 1./
         81.*m44*pow(m46,-1)*me2*RSe61*RS36e + 1./81.*m44*pow(m46,-1)*
         me2*RSe61*RS34e + 1./27.*m44*pow(m46,-1)*me2*pow(RSe61,2) - 1./
         81.*m44*pow(m46,-1)*me2*RS16e*RSe63 + 1./81.*m44*pow(m46,-1)*
         me2*RS16e*RSe61 + 1./81.*m44*pow(m46,-1)*me2*RS14e*RSe63 - 1./
         81.*m44*pow(m46,-1)*me2*RS14e*RSe61 - 1./9.*m33*pow(m46,-1)*
         me2*pow(RSe63,2) - 2./9.*m33*pow(m46,-1)*me2*RSe61*RSe63 - 2./
         27.*m33*pow(m46,-1)*me2*RSe61*RS36e + 2./27.*m33*pow(m46,-1)*
         me2*RSe61*RS34e - 1./27.*m33*pow(m46,-1)*me2*RS16e*RSe63 + 1./
         27.*m33*pow(m46,-1)*me2*RS14e*RSe63 - 13./72.*m33*pow(m46,-1)*
         m66 - 1./81.*m33*pow(m46,-1)*m66*RS34e*RS36e - 1./162.*m33*
         pow(m46,-1)*m66*pow(RS34e,2) - 4./81.*m33*pow(m46,-1)*m66*
         RSe61*RS36e - 2./81.*m33*pow(m46,-1)*m66*RSe61*RS34e - 2./81.*
         m33*pow(m46,-1)*m66*RS16e*RSe63 + 2./81.*m33*pow(m46,-1)*m66*
         RS16e*RS34e );

      M13p6x22R +=  + pow(Ab(m46,mu2),2) * (  - 1./81.*m33*pow(m46,-1)*
         m66*RS14e*RSe63 + 2./81.*m33*pow(m46,-1)*m66*RS14e*RS36e + 2./
         81.*m33*pow(m46,-1)*m66*RS14e*RS34e - 13./72.*m33*m44*pow(
         m46,-1) - 1./54.*m33*m44*pow(m46,-1)*pow(RSe63,2) - 1./162.*
         m33*m44*pow(m46,-1)*pow(RS36e,2) - 1./81.*m33*m44*pow(m46,-1)*
         RS34e*RS36e - 1./27.*m33*m44*pow(m46,-1)*RSe61*RSe63 - 2./81.*
         m33*m44*pow(m46,-1)*RSe61*RS36e + 2./81.*m33*m44*pow(m46,-1)*
         RSe61*RS34e - 1./81.*m33*m44*pow(m46,-1)*RS16e*RSe63 + 2./81.*
         m33*m44*pow(m46,-1)*RS16e*RS36e + 2./81.*m33*m44*pow(m46,-1)*
         RS16e*RS34e + 1./81.*m33*m44*pow(m46,-1)*RS14e*RSe63 + 2./81.*
         m33*m44*pow(m46,-1)*RS14e*RS36e + 1./162.*pow(m33,2)*pow(
         m46,-1)*pow(RS36e,2) - 1./81.*pow(m33,2)*pow(m46,-1)*RS34e*
         RS36e + 1./162.*pow(m33,2)*pow(m46,-1)*pow(RS34e,2) - 2./27.*
         pow(m33,2)*pow(m46,-1)*RSe61*RS36e + 2./27.*pow(m33,2)*pow(
         m46,-1)*RSe61*RS34e - 2./81.*pow(m33,2)*pow(m46,-1)*RS16e*
         RS36e );

      M13p6x22R +=  + pow(Ab(m46,mu2),2) * ( 2./81.*pow(m33,2)*pow(
         m46,-1)*RS16e*RS34e + 2./81.*pow(m33,2)*pow(m46,-1)*RS14e*
         RS36e - 2./81.*pow(m33,2)*pow(m46,-1)*RS14e*RS34e - 2./9.*m11*
         pow(m46,-1)*me2*RSe61*RSe63 - 1./27.*m11*pow(m46,-1)*me2*RSe61
         *RS36e + 1./27.*m11*pow(m46,-1)*me2*RSe61*RS34e - 1./9.*m11*
         pow(m46,-1)*me2*pow(RSe61,2) - 2./27.*m11*pow(m46,-1)*me2*
         RS16e*RSe63 + 2./27.*m11*pow(m46,-1)*me2*RS14e*RSe63 - 13./72.
         *m11*pow(m46,-1)*m66 - 2./81.*m11*pow(m46,-1)*m66*RSe61*RS36e
          - 1./81.*m11*pow(m46,-1)*m66*RSe61*RS34e - 4./81.*m11*pow(
         m46,-1)*m66*RS16e*RSe63 + 2./81.*m11*pow(m46,-1)*m66*RS16e*
         RS34e - 2./81.*m11*pow(m46,-1)*m66*RS14e*RSe63 + 2./81.*m11*
         pow(m46,-1)*m66*RS14e*RS36e + 2./81.*m11*pow(m46,-1)*m66*RS14e
         *RS34e - 1./81.*m11*pow(m46,-1)*m66*RS14e*RS16e - 1./162.*m11*
         pow(m46,-1)*m66*pow(RS14e,2) - 13./72.*m11*m44*pow(m46,-1) - 1.
         /27.*m11*m44*pow(m46,-1)*RSe61*RSe63 - 1./81.*m11*m44*pow(
         m46,-1)*RSe61*RS36e );

      M13p6x22R +=  + pow(Ab(m46,mu2),2) * ( 1./81.*m11*m44*pow(m46,-1)
         *RSe61*RS34e - 1./54.*m11*m44*pow(m46,-1)*pow(RSe61,2) - 2./81.
         *m11*m44*pow(m46,-1)*RS16e*RSe63 + 2./81.*m11*m44*pow(m46,-1)*
         RS16e*RS36e + 2./81.*m11*m44*pow(m46,-1)*RS16e*RS34e - 1./162.
         *m11*m44*pow(m46,-1)*pow(RS16e,2) + 2./81.*m11*m44*pow(m46,-1)
         *RS14e*RSe63 + 2./81.*m11*m44*pow(m46,-1)*RS14e*RS36e - 1./81.
         *m11*m44*pow(m46,-1)*RS14e*RS16e - 1./27.*m11*m33*pow(m46,-1)*
         RSe61*RS36e + 1./27.*m11*m33*pow(m46,-1)*RSe61*RS34e - 1./27.*
         m11*m33*pow(m46,-1)*RS16e*RSe63 - 1./27.*m11*m33*pow(m46,-1)*
         RS16e*RS36e + 1./27.*m11*m33*pow(m46,-1)*RS16e*RS34e + 1./27.*
         m11*m33*pow(m46,-1)*RS14e*RSe63 + 1./27.*m11*m33*pow(m46,-1)*
         RS14e*RS36e - 1./27.*m11*m33*pow(m46,-1)*RS14e*RS34e - 2./27.*
         pow(m11,2)*pow(m46,-1)*RS16e*RSe63 - 2./81.*pow(m11,2)*pow(
         m46,-1)*RS16e*RS36e + 2./81.*pow(m11,2)*pow(m46,-1)*RS16e*
         RS34e + 1./162.*pow(m11,2)*pow(m46,-1)*pow(RS16e,2) + 2./27.*
         pow(m11,2)*pow(m46,-1)*RS14e*RSe63 );

      M13p6x22R +=  + pow(Ab(m46,mu2),2) * ( 2./81.*pow(m11,2)*pow(
         m46,-1)*RS14e*RS36e - 2./81.*pow(m11,2)*pow(m46,-1)*RS14e*
         RS34e - 1./81.*pow(m11,2)*pow(m46,-1)*RS14e*RS16e + 1./162.*
         pow(m11,2)*pow(m46,-1)*pow(RS14e,2) );

      M13p6x22R +=  + Ab(m46,mu2)*Ab(me2,mu2) * ( 2./9.*me2*pow(
         RSe63,2) + 2./27.*me2*RS36e*RSe63 - 2./27.*me2*RS34e*RSe63 - 4.
         /9.*me2*RSe61*RSe63 - 2./27.*me2*RSe61*RS36e + 2./27.*me2*
         RSe61*RS34e + 2./9.*me2*pow(RSe61,2) - 2./27.*me2*RS16e*RSe63
          + 2./27.*me2*RS16e*RSe61 + 2./27.*me2*RS14e*RSe63 - 2./27.*
         me2*RS14e*RSe61 - 4./81.*m66*me2*RN3e*RS36e*RSe63 - 2./81.*m66
         *me2*RN3e*RS34e*RSe63 + 4./81.*m66*me2*RSe61*RN3e*RS36e + 2./
         81.*m66*me2*RSe61*RN3e*RS34e + 4./81.*m66*me2*RN1e*RS16e*RSe63
          - 4./81.*m66*me2*RN1e*RS16e*RSe61 + 2./81.*m66*me2*RN1e*RS14e
         *RSe63 - 2./81.*m66*me2*RN1e*RS14e*RSe61 + 1./27.*m44*pow(
         RSe63,2) - 2./27.*m44*RSe61*RSe63 + 1./27.*m44*pow(RSe61,2) - 
         2./81.*m44*me2*RN3e*RS36e*RSe63 + 2./81.*m44*me2*RN3e*RS34e*
         RSe63 + 2./81.*m44*me2*RSe61*RN3e*RS36e - 2./81.*m44*me2*RSe61
         *RN3e*RS34e + 2./81.*m44*me2*RN1e*RS16e*RSe63 - 2./81.*m44*me2
         *RN1e*RS16e*RSe61 - 2./81.*m44*me2*RN1e*RS14e*RSe63 + 2./81.*
         m44*me2*RN1e*RS14e*RSe61 );

      M13p6x22R +=  + Ab(m46,mu2)*Ab(me2,mu2) * (  - 1./9.*m33*pow(
         RSe63,2) - 2./9.*m33*RSe61*RSe63 + 4./27.*m33*me2*RSe61*RN3e*
         RS36e - 4./27.*m33*me2*RSe61*RN3e*RS34e + 2./27.*m33*me2*RN1e*
         RS16e*RSe63 - 2./27.*m33*me2*RN1e*RS14e*RSe63 + 4./81.*m33*m66
         *RN3e*RS36e*RSe63 + 2./81.*m33*m66*RN3e*RS34e*RSe63 + 4./81.*
         m33*m66*RSe61*RN3e*RS36e + 2./81.*m33*m66*RSe61*RN3e*RS34e + 4.
         /81.*m33*m66*RN1e*RS16e*RSe63 + 2./81.*m33*m66*RN1e*RS14e*
         RSe63 + 2./81.*m33*m44*RN3e*RS36e*RSe63 - 2./81.*m33*m44*RN3e*
         RS34e*RSe63 + 2./81.*m33*m44*RSe61*RN3e*RS36e - 2./81.*m33*m44
         *RSe61*RN3e*RS34e + 2./81.*m33*m44*RN1e*RS16e*RSe63 - 2./81.*
         m33*m44*RN1e*RS14e*RSe63 - 2./9.*m11*RSe61*RSe63 - 1./9.*m11*
         pow(RSe61,2) + 2./27.*m11*me2*RSe61*RN3e*RS36e - 2./27.*m11*
         me2*RSe61*RN3e*RS34e + 4./27.*m11*me2*RN1e*RS16e*RSe63 - 4./27.
         *m11*me2*RN1e*RS14e*RSe63 + 4./81.*m11*m66*RSe61*RN3e*RS36e + 
         2./81.*m11*m66*RSe61*RN3e*RS34e + 4./81.*m11*m66*RN1e*RS16e*
         RSe63 );

      M13p6x22R +=  + Ab(m46,mu2)*Ab(me2,mu2) * ( 4./81.*m11*m66*RN1e*
         RS16e*RSe61 + 2./81.*m11*m66*RN1e*RS14e*RSe63 + 2./81.*m11*m66
         *RN1e*RS14e*RSe61 + 2./81.*m11*m44*RSe61*RN3e*RS36e - 2./81.*
         m11*m44*RSe61*RN3e*RS34e + 2./81.*m11*m44*RN1e*RS16e*RSe63 + 2.
         /81.*m11*m44*RN1e*RS16e*RSe61 - 2./81.*m11*m44*RN1e*RS14e*
         RSe63 - 2./81.*m11*m44*RN1e*RS14e*RSe61 );

      M13p6x22R +=  + Ab(m46,mu2)*Bb(me2,mu2) * ( 1./9.*pow(me2,2)*pow(
         RSe63,2) - 2./9.*pow(me2,2)*RSe61*RSe63 + 1./9.*pow(me2,2)*
         pow(RSe61,2) + 1./27.*m44*me2*pow(RSe63,2) - 2./27.*m44*me2*
         RSe61*RSe63 + 1./27.*m44*me2*pow(RSe61,2) - 1./9.*m33*me2*pow(
         RSe63,2) - 2./9.*m33*me2*RSe61*RSe63 - 1./27.*m33*m44*pow(
         RSe63,2) - 2./27.*m33*m44*RSe61*RSe63 - 2./9.*m11*me2*RSe61*
         RSe63 - 1./9.*m11*me2*pow(RSe61,2) - 2./27.*m11*m44*RSe61*
         RSe63 - 1./27.*m11*m44*pow(RSe61,2) );

      M13p6x22R +=  + Ab(me2,mu2)*pi16 * ( 1./36.*m33*RD346e*RSe6613 - 
         1./72.*m33*RD346e*RSe6633 - 1./72.*m33*RSe6611*RD346e + 1./36.
         *m33*RD146e*RSe6613 - 1./72.*m33*RD146e*RSe6633 - 1./72.*m33*
         RD146e*RSe6611 + 1./12.*m33*me2*RSe6613 + 1./9.*m33*me2*pow(
         RSe6613,2) - 1./24.*m33*me2*RSe6633 - 1./9.*m33*me2*RSe6633*
         RSe6613 + 1./36.*m33*me2*pow(RSe6633,2) - 1./24.*m33*me2*
         RSe6611 - 1./9.*m33*me2*RSe6611*RSe6613 + 1./18.*m33*me2*
         RSe6611*RSe6633 + 1./36.*m33*me2*pow(RSe6611,2) - 1./12.*m33*
         m66*RSe6613 + 1./24.*m33*m66*RSe6633 + 1./24.*m33*m66*RSe6611
          - 1./6.*m33*m44*RSe6613 + 1./12.*m33*m44*RSe6633 + 1./12.*m33
         *m44*RSe6611 - 1./12.*pow(m33,2)*RSe6613 - 2./27.*pow(m33,2)*
         RSe6613*RS3461e - 1./54.*pow(m33,2)*RSe6613*RS1463e + 1./27.*
         pow(m33,2)*pow(RSe6613,2) + 1./24.*pow(m33,2)*RSe6633 + 1./27.
         *pow(m33,2)*RSe6633*RS3461e + 1./108.*pow(m33,2)*RSe6633*
         RS1463e - 1./27.*pow(m33,2)*RSe6633*RSe6613 + 1./108.*pow(
         m33,2)*pow(RSe6633,2) );

      M13p6x22R +=  + Ab(me2,mu2)*pi16 * ( 1./27.*pow(m33,2)*RC346e*
         RSe6613 - 1./54.*pow(m33,2)*RC346e*RSe6633 + 1./24.*pow(m33,2)
         *RSe6611 + 1./27.*pow(m33,2)*RSe6611*RS3461e + 1./108.*pow(
         m33,2)*RSe6611*RS1463e - 1./27.*pow(m33,2)*RSe6611*RSe6613 + 1.
         /54.*pow(m33,2)*RSe6611*RSe6633 - 1./54.*pow(m33,2)*RSe6611*
         RC346e + 1./108.*pow(m33,2)*pow(RSe6611,2) + 1./108.*pow(
         m33,2)*RC146e*RSe6613 - 1./216.*pow(m33,2)*RC146e*RSe6633 - 1./
         216.*pow(m33,2)*RC146e*RSe6611 + 1./36.*m11*RD346e*RSe6613 - 1.
         /72.*m11*RD346e*RSe6633 - 1./72.*m11*RSe6611*RD346e + 1./36.*
         m11*RD146e*RSe6613 - 1./72.*m11*RD146e*RSe6633 - 1./72.*m11*
         RD146e*RSe6611 + 1./12.*m11*me2*RSe6613 + 1./9.*m11*me2*pow(
         RSe6613,2) - 1./24.*m11*me2*RSe6633 - 1./9.*m11*me2*RSe6633*
         RSe6613 + 1./36.*m11*me2*pow(RSe6633,2) - 1./24.*m11*me2*
         RSe6611 - 1./9.*m11*me2*RSe6611*RSe6613 + 1./18.*m11*me2*
         RSe6611*RSe6633 + 1./36.*m11*me2*pow(RSe6611,2) - 1./12.*m11*
         m66*RSe6613 );

      M13p6x22R +=  + Ab(me2,mu2)*pi16 * ( 1./24.*m11*m66*RSe6633 + 1./
         24.*m11*m66*RSe6611 - 1./6.*m11*m44*RSe6613 + 1./12.*m11*m44*
         RSe6633 + 1./12.*m11*m44*RSe6611 - 1./6.*m11*m33*RSe6613 - 5./
         54.*m11*m33*RSe6613*RS3461e - 5./54.*m11*m33*RSe6613*RS1463e
          + 2./27.*m11*m33*pow(RSe6613,2) + 1./12.*m11*m33*RSe6633 + 5./
         108.*m11*m33*RSe6633*RS3461e + 5./108.*m11*m33*RSe6633*RS1463e
          - 2./27.*m11*m33*RSe6633*RSe6613 + 1./54.*m11*m33*pow(
         RSe6633,2) + 5./108.*m11*m33*RC346e*RSe6613 - 5./216.*m11*m33*
         RC346e*RSe6633 + 1./12.*m11*m33*RSe6611 + 5./108.*m11*m33*
         RSe6611*RS3461e + 5./108.*m11*m33*RSe6611*RS1463e - 2./27.*m11
         *m33*RSe6611*RSe6613 + 1./27.*m11*m33*RSe6611*RSe6633 - 5./216.
         *m11*m33*RSe6611*RC346e + 1./54.*m11*m33*pow(RSe6611,2) + 5./
         108.*m11*m33*RC146e*RSe6613 - 5./216.*m11*m33*RC146e*RSe6633
          - 5./216.*m11*m33*RC146e*RSe6611 - 1./12.*pow(m11,2)*RSe6613
          - 1./54.*pow(m11,2)*RSe6613*RS3461e - 2./27.*pow(m11,2)*
         RSe6613*RS1463e );

      M13p6x22R +=  + Ab(me2,mu2)*pi16 * ( 1./27.*pow(m11,2)*pow(
         RSe6613,2) + 1./24.*pow(m11,2)*RSe6633 + 1./108.*pow(m11,2)*
         RSe6633*RS3461e + 1./27.*pow(m11,2)*RSe6633*RS1463e - 1./27.*
         pow(m11,2)*RSe6633*RSe6613 + 1./108.*pow(m11,2)*pow(RSe6633,2)
          + 1./108.*pow(m11,2)*RC346e*RSe6613 - 1./216.*pow(m11,2)*
         RC346e*RSe6633 + 1./24.*pow(m11,2)*RSe6611 + 1./108.*pow(
         m11,2)*RSe6611*RS3461e + 1./27.*pow(m11,2)*RSe6611*RS1463e - 1.
         /27.*pow(m11,2)*RSe6611*RSe6613 + 1./54.*pow(m11,2)*RSe6611*
         RSe6633 - 1./216.*pow(m11,2)*RSe6611*RC346e + 1./108.*pow(
         m11,2)*pow(RSe6611,2) + 1./27.*pow(m11,2)*RC146e*RSe6613 - 1./
         54.*pow(m11,2)*RC146e*RSe6633 - 1./54.*pow(m11,2)*RC146e*
         RSe6611 );

      M13p6x22R +=  + pow(Ab(me2,mu2),2) * ( 1./270.*RD346e*RSe6613 - 1.
         /540.*RD346e*RSe6633 - 1./540.*RSe6611*RD346e + 1./270.*RD146e
         *RSe6613 - 1./540.*RD146e*RSe6633 - 1./540.*RD146e*RSe6611 + 
         14./405.*me2*pow(RSe6613,2) - 14./405.*me2*RSe6633*RSe6613 + 7.
         /810.*me2*pow(RSe6633,2) - 14./405.*me2*RSe6611*RSe6613 + 7./
         405.*me2*RSe6611*RSe6633 + 7./810.*me2*pow(RSe6611,2) - 1./45.
         *m66 + 4./135.*m66*RSe6613 + 1./45.*m66*RSe63 - 19./81.*m66*
         pow(RSe63,2) - 7./270.*m66*RSe6633 + 13./162.*m66*RS34e*RSe63
          + 1./45.*m66*RSe61 + 38./81.*m66*RSe61*RSe63 - 13./162.*m66*
         RSe61*RS34e - 19./81.*m66*pow(RSe61,2) - 7./270.*m66*RSe6611
          - 13./162.*m66*RS14e*RSe63 + 13./162.*m66*RS14e*RSe61 - 4./81.
         *m66*me2*RN3e*RS34e*RSe63 + 4./81.*m66*me2*RSe61*RN3e*RS34e + 
         4./81.*m66*me2*RN1e*RS14e*RSe63 - 4./81.*m66*me2*RN1e*RS14e*
         RSe61 + 1./45.*pow(m66,2)*pow(me2,-1) - 1./90.*pow(m66,2)*pow(
         me2,-1)*RSe63 + 13./243.*pow(m66,2)*pow(me2,-1)*pow(RSe63,2)
          - 1./90.*pow(m66,2)*pow(me2,-1)*RSe61 );

      M13p6x22R +=  + pow(Ab(me2,mu2),2) * (  - 26./243.*pow(m66,2)*
         pow(me2,-1)*RSe61*RSe63 + 13./243.*pow(m66,2)*pow(me2,-1)*pow(
         RSe61,2) - 1./90.*m44 + 8./135.*m44*RSe6613 - 1./45.*m44*RSe63
          - 41./324.*m44*pow(RSe63,2) - 7./135.*m44*RSe6633 - 13./162.*
         m44*RS36e*RSe63 + 1./54.*m44*RS34e*RSe63 - 1./45.*m44*RSe61 + 
         41./162.*m44*RSe61*RSe63 + 13./162.*m44*RSe61*RS36e - 1./54.*
         m44*RSe61*RS34e - 41./324.*m44*pow(RSe61,2) - 7./135.*m44*
         RSe6611 + 13./162.*m44*RS16e*RSe63 - 13./162.*m44*RS16e*RSe61
          - 1./54.*m44*RS14e*RSe63 + 1./54.*m44*RS14e*RSe61 + 1./81.*
         m44*me2*RN3e*RS36e*RSe63 - 1./81.*m44*me2*RSe61*RN3e*RS36e - 1.
         /81.*m44*me2*RN1e*RS16e*RSe63 + 1./81.*m44*me2*RN1e*RS16e*
         RSe61 + 16./243.*m44*m66*pow(me2,-1)*pow(RSe63,2) - 32./243.*
         m44*m66*pow(me2,-1)*RSe61*RSe63 + 16./243.*m44*m66*pow(me2,-1)
         *pow(RSe61,2) + 1./90.*pow(m44,2)*pow(me2,-1) + 1./90.*pow(
         m44,2)*pow(me2,-1)*RSe63 + 19./972.*pow(m44,2)*pow(me2,-1)*
         pow(RSe63,2) );

      M13p6x22R +=  + pow(Ab(me2,mu2),2) * ( 1./90.*pow(m44,2)*pow(
         me2,-1)*RSe61 - 19./486.*pow(m44,2)*pow(me2,-1)*RSe61*RSe63 + 
         19./972.*pow(m44,2)*pow(me2,-1)*pow(RSe61,2) + 7./405.*m33*
         pow(me2,-1)*RD346e*RSe6613 - 1./81.*m33*pow(me2,-1)*RD346e*
         RSe6633 + 11./810.*m33*pow(me2,-1)*RSe6611*RD346e + 1./810.*
         m33*pow(me2,-1)*RD146e*RSe6613 + 23./1620.*m33*pow(me2,-1)*
         RD146e*RSe6633 - 5./324.*m33*pow(me2,-1)*RD146e*RSe6611 - 19./
         240.*m33 - 29./360.*m33*RSe6613 - 7./270.*m33*RSe6613*RS3461e
          - 1./54.*m33*RSe6613*RS1463e + 88./405.*m33*pow(RSe6613,2) - 
         2./45.*m33*RSe6633 + 7./540.*m33*RSe6633*RS3461e + 1./108.*m33
         *RSe6633*RS1463e - 32./405.*m33*RSe6633*RSe6613 + 2./15.*m33*
         pow(RSe6633,2) + 7./540.*m33*RC346e*RSe6613 - 7./1080.*m33*
         RC346e*RSe6633 - 1./30.*m33*RSe6611 + 7./540.*m33*RSe6611*
         RS3461e + 1./108.*m33*RSe6611*RS1463e - 1./45.*m33*RSe6611*
         RSe6613 - 77./810.*m33*RSe6611*RSe6633 - 7./1080.*m33*RSe6611*
         RC346e );

      M13p6x22R +=  + pow(Ab(me2,mu2),2) * ( 11./162.*m33*pow(
         RSe6611,2) + 1./108.*m33*RC146e*RSe6613 - 1./216.*m33*RC146e*
         RSe6633 - 1./216.*m33*RC146e*RSe6611 - 4./27.*m33*me2*RN3e*
         RSe6633*RSe6613 + 4./27.*m33*me2*RN3e*pow(RSe6633,2) - 1./135.
         *m33*m66*pow(me2,-1) + 7./3240.*m33*m66*pow(me2,-1)*RSe6613 + 
         5./108.*m33*m66*pow(me2,-1)*RSe63 + 13./162.*m33*m66*pow(
         me2,-1)*pow(RSe63,2) - 89./2160.*m33*m66*pow(me2,-1)*RSe6633
          - 1./162.*m33*m66*pow(me2,-1)*pow(RS34e,2) - 7./540.*m33*m66*
         pow(me2,-1)*RSe61 + 13./81.*m33*m66*pow(me2,-1)*RSe61*RSe63 - 
         4./81.*m33*m66*pow(me2,-1)*RSe61*RS34e - 31./1296.*m33*m66*
         pow(me2,-1)*RSe6611 - 2./81.*m33*m66*pow(me2,-1)*RS14e*RSe63
          + 2./81.*m33*m66*pow(me2,-1)*RS14e*RS34e + 4./81.*m33*m66*
         RN3e*RS34e*RSe63 + 4./81.*m33*m66*RSe61*RN3e*RS34e + 4./81.*
         m33*m66*RN1e*RS14e*RSe63 - 1./270.*m33*m44*pow(me2,-1) + 7./
         1620.*m33*m44*pow(me2,-1)*RSe6613 - 5./108.*m33*m44*pow(
         me2,-1)*RSe63 );

      M13p6x22R +=  + pow(Ab(me2,mu2),2) * ( 19./324.*m33*m44*pow(
         me2,-1)*pow(RSe63,2) - 89./1080.*m33*m44*pow(me2,-1)*RSe6633
          - 1./324.*m33*m44*pow(me2,-1)*pow(RS36e,2) + 7./540.*m33*m44*
         pow(me2,-1)*RSe61 + 19./162.*m33*m44*pow(me2,-1)*RSe61*RSe63
          + 1./81.*m33*m44*pow(me2,-1)*RSe61*RS36e - 31./648.*m33*m44*
         pow(me2,-1)*RSe6611 + 1./162.*m33*m44*pow(me2,-1)*RS16e*RSe63
          + 1./81.*m33*m44*pow(me2,-1)*RS16e*RS36e - 1./81.*m33*m44*
         RN3e*RS36e*RSe63 - 1./81.*m33*m44*RSe61*RN3e*RS36e - 1./81.*
         m33*m44*RN1e*RS16e*RSe63 - 1./720.*pow(m33,2)*pow(me2,-1)*
         RSe6613 - 1./81.*pow(m33,2)*pow(me2,-1)*RSe6613*RS3461e + 2./
         81.*pow(m33,2)*pow(me2,-1)*RSe6613*RS1463e - 4./81.*pow(m33,2)
         *pow(me2,-1)*pow(RSe6613,2) - 37./1080.*pow(m33,2)*pow(me2,-1)
         *RSe6633 - 193./1620.*pow(m33,2)*pow(me2,-1)*RSe6633*RS3461e
          - 29./324.*pow(m33,2)*pow(me2,-1)*RSe6633*RS1463e + 17./81.*
         pow(m33,2)*pow(me2,-1)*RSe6633*RSe6613 - 1./36.*pow(m33,2)*
         pow(me2,-1)*pow(RSe6633,2) );

      M13p6x22R +=  + pow(Ab(me2,mu2),2) * ( 7./1620.*pow(m33,2)*pow(
         me2,-1)*RC346e*RSe6613 + 1./3240.*pow(m33,2)*pow(me2,-1)*
         RC346e*RSe6633 - 23./432.*pow(m33,2)*pow(me2,-1)*RSe6611 - 1./
         60.*pow(m33,2)*pow(me2,-1)*RSe6611*RS3461e - 1./108.*pow(
         m33,2)*pow(me2,-1)*RSe6611*RS1463e + 1./27.*pow(m33,2)*pow(
         me2,-1)*RSe6611*RSe6613 - 1./162.*pow(m33,2)*pow(me2,-1)*
         RSe6611*RSe6633 + 1./72.*pow(m33,2)*pow(me2,-1)*RSe6611*RC346e
          - 5./324.*pow(m33,2)*pow(me2,-1)*pow(RSe6611,2) - 1./108.*
         pow(m33,2)*pow(me2,-1)*RC146e*RSe6613 + 1./648.*pow(m33,2)*
         pow(me2,-1)*RC146e*RSe6633 + 5./648.*pow(m33,2)*pow(me2,-1)*
         RC146e*RSe6611 - 4./27.*pow(m33,2)*RN3e*RSe6633*RSe6613 - 4./
         27.*pow(m33,2)*RN3e*pow(RSe6633,2) + 1./810.*m11*pow(me2,-1)*
         RD346e*RSe6613 - 5./324.*m11*pow(me2,-1)*RD346e*RSe6633 + 23./
         1620.*m11*pow(me2,-1)*RSe6611*RD346e + 7./405.*m11*pow(me2,-1)
         *RD146e*RSe6613 + 11./810.*m11*pow(me2,-1)*RD146e*RSe6633 - 1./
         81.*m11*pow(me2,-1)*RD146e*RSe6611 );

      M13p6x22R +=  + pow(Ab(me2,mu2),2) * (  - 19./240.*m11 - 29./360.
         *m11*RSe6613 - 1./54.*m11*RSe6613*RS3461e - 7./270.*m11*
         RSe6613*RS1463e + 88./405.*m11*pow(RSe6613,2) - 1./30.*m11*
         RSe6633 + 1./108.*m11*RSe6633*RS3461e + 7./540.*m11*RSe6633*
         RS1463e - 1./45.*m11*RSe6633*RSe6613 + 11./162.*m11*pow(
         RSe6633,2) + 1./108.*m11*RC346e*RSe6613 - 1./216.*m11*RC346e*
         RSe6633 - 2./45.*m11*RSe6611 + 1./108.*m11*RSe6611*RS3461e + 7.
         /540.*m11*RSe6611*RS1463e - 32./405.*m11*RSe6611*RSe6613 - 77./
         810.*m11*RSe6611*RSe6633 - 1./216.*m11*RSe6611*RC346e + 2./15.
         *m11*pow(RSe6611,2) + 7./540.*m11*RC146e*RSe6613 - 7./1080.*
         m11*RC146e*RSe6633 - 7./1080.*m11*RC146e*RSe6611 - 4./27.*m11*
         me2*RN1e*RSe6611*RSe6613 + 4./27.*m11*me2*RN1e*pow(RSe6611,2)
          - 1./135.*m11*m66*pow(me2,-1) + 7./3240.*m11*m66*pow(me2,-1)*
         RSe6613 - 7./540.*m11*m66*pow(me2,-1)*RSe63 - 31./1296.*m11*
         m66*pow(me2,-1)*RSe6633 + 5./108.*m11*m66*pow(me2,-1)*RSe61 + 
         13./81.*m11*m66*pow(me2,-1)*RSe61*RSe63 );

      M13p6x22R +=  + pow(Ab(me2,mu2),2) * (  - 2./81.*m11*m66*pow(
         me2,-1)*RSe61*RS34e + 13./162.*m11*m66*pow(me2,-1)*pow(
         RSe61,2) - 89./2160.*m11*m66*pow(me2,-1)*RSe6611 - 4./81.*m11*
         m66*pow(me2,-1)*RS14e*RSe63 + 2./81.*m11*m66*pow(me2,-1)*RS14e
         *RS34e - 1./162.*m11*m66*pow(me2,-1)*pow(RS14e,2) + 4./81.*m11
         *m66*RSe61*RN3e*RS34e + 4./81.*m11*m66*RN1e*RS14e*RSe63 + 4./
         81.*m11*m66*RN1e*RS14e*RSe61 - 1./270.*m11*m44*pow(me2,-1) + 7.
         /1620.*m11*m44*pow(me2,-1)*RSe6613 + 7./540.*m11*m44*pow(
         me2,-1)*RSe63 - 31./648.*m11*m44*pow(me2,-1)*RSe6633 - 5./108.
         *m11*m44*pow(me2,-1)*RSe61 + 19./162.*m11*m44*pow(me2,-1)*
         RSe61*RSe63 + 1./162.*m11*m44*pow(me2,-1)*RSe61*RS36e + 19./
         324.*m11*m44*pow(me2,-1)*pow(RSe61,2) - 89./1080.*m11*m44*pow(
         me2,-1)*RSe6611 + 1./81.*m11*m44*pow(me2,-1)*RS16e*RSe63 + 1./
         81.*m11*m44*pow(me2,-1)*RS16e*RS36e - 1./324.*m11*m44*pow(
         me2,-1)*pow(RS16e,2) - 1./81.*m11*m44*RSe61*RN3e*RS36e - 1./81.
         *m11*m44*RN1e*RS16e*RSe63 );

      M13p6x22R +=  + pow(Ab(me2,mu2),2) * (  - 1./81.*m11*m44*RN1e*
         RS16e*RSe61 - 163./1080.*m11*m33*pow(me2,-1)*RSe6613 + 1./135.
         *m11*m33*pow(me2,-1)*RSe6613*RS3461e + 1./135.*m11*m33*pow(
         me2,-1)*RSe6613*RS1463e - 8./81.*m11*m33*pow(me2,-1)*pow(
         RSe6613,2) - 53./2160.*m11*m33*pow(me2,-1)*RSe6633 - 1./15.*
         m11*m33*pow(me2,-1)*RSe6633*RS3461e - 4./45.*m11*m33*pow(
         me2,-1)*RSe6633*RS1463e + 14./81.*m11*m33*pow(me2,-1)*RSe6633*
         RSe6613 - 7./162.*m11*m33*pow(me2,-1)*pow(RSe6633,2) - 19./810.
         *m11*m33*pow(me2,-1)*RC346e*RSe6613 + 1./162.*m11*m33*pow(
         me2,-1)*RC346e*RSe6633 - 53./2160.*m11*m33*pow(me2,-1)*RSe6611
          - 4./45.*m11*m33*pow(me2,-1)*RSe6611*RS3461e - 1./15.*m11*m33
         *pow(me2,-1)*RSe6611*RS1463e + 14./81.*m11*m33*pow(me2,-1)*
         RSe6611*RSe6613 - 1./81.*m11*m33*pow(me2,-1)*RSe6611*RSe6633
          + 7./405.*m11*m33*pow(me2,-1)*RSe6611*RC346e - 7./162.*m11*
         m33*pow(me2,-1)*pow(RSe6611,2) - 19./810.*m11*m33*pow(me2,-1)*
         RC146e*RSe6613 );

      M13p6x22R +=  + pow(Ab(me2,mu2),2) * ( 7./405.*m11*m33*pow(
         me2,-1)*RC146e*RSe6633 + 1./162.*m11*m33*pow(me2,-1)*RC146e*
         RSe6611 - 4./27.*m11*m33*RN3e*RSe6633*RSe6613 - 4./27.*m11*m33
         *RN1e*RSe6611*RSe6613 - 1./720.*pow(m11,2)*pow(me2,-1)*RSe6613
          + 2./81.*pow(m11,2)*pow(me2,-1)*RSe6613*RS3461e - 1./81.*pow(
         m11,2)*pow(me2,-1)*RSe6613*RS1463e - 4./81.*pow(m11,2)*pow(
         me2,-1)*pow(RSe6613,2) - 23./432.*pow(m11,2)*pow(me2,-1)*
         RSe6633 - 1./108.*pow(m11,2)*pow(me2,-1)*RSe6633*RS3461e - 1./
         60.*pow(m11,2)*pow(me2,-1)*RSe6633*RS1463e + 1./27.*pow(m11,2)
         *pow(me2,-1)*RSe6633*RSe6613 - 5./324.*pow(m11,2)*pow(me2,-1)*
         pow(RSe6633,2) - 1./108.*pow(m11,2)*pow(me2,-1)*RC346e*RSe6613
          + 5./648.*pow(m11,2)*pow(me2,-1)*RC346e*RSe6633 - 37./1080.*
         pow(m11,2)*pow(me2,-1)*RSe6611 - 29./324.*pow(m11,2)*pow(
         me2,-1)*RSe6611*RS3461e - 193./1620.*pow(m11,2)*pow(me2,-1)*
         RSe6611*RS1463e + 17./81.*pow(m11,2)*pow(me2,-1)*RSe6611*
         RSe6613 );

      M13p6x22R +=  + pow(Ab(me2,mu2),2) * (  - 1./162.*pow(m11,2)*pow(
         me2,-1)*RSe6611*RSe6633 + 1./648.*pow(m11,2)*pow(me2,-1)*
         RSe6611*RC346e - 1./36.*pow(m11,2)*pow(me2,-1)*pow(RSe6611,2)
          + 7./1620.*pow(m11,2)*pow(me2,-1)*RC146e*RSe6613 + 1./72.*
         pow(m11,2)*pow(me2,-1)*RC146e*RSe6633 + 1./3240.*pow(m11,2)*
         pow(me2,-1)*RC146e*RSe6611 - 4./27.*pow(m11,2)*RN1e*RSe6611*
         RSe6613 - 4./27.*pow(m11,2)*RN1e*pow(RSe6611,2) );

      M13p6x22R +=  + Ab(me2,mu2)*Bb(m11,mu2) * ( 1./135.*me2*RD146e*
         RSe6613 - 1./270.*me2*RD146e*RSe6633 - 1./270.*me2*RD146e*
         RSe6611 - 1./270.*m33*RD146e*RSe6613 + 17./540.*m33*RD146e*
         RSe6633 - 1./36.*m33*RD146e*RSe6611 + 1./30.*m11*RD146e*
         RSe6613 + 1./36.*m11*RD146e*RSe6633 - 13./540.*m11*RD146e*
         RSe6611 + 2./27.*m11*me2*RD146e*RN1e*RSe6613 - 2./27.*m11*me2*
         RD146e*RN1e*RSe6611 - 2./27.*m11*m33*RSe6611*RS1463e + 2./27.*
         m11*m33*RD146e*RN1e*RSe6613 - 4./27.*pow(m11,2)*RSe6611*
         RS1463e + 2./27.*pow(m11,2)*RD146e*RN1e*RSe6613 + 2./27.*pow(
         m11,2)*RD146e*RN1e*RSe6611 );

      M13p6x22R +=  + Ab(me2,mu2)*Bb(m33,mu2) * ( 1./135.*me2*RD346e*
         RSe6613 - 1./270.*me2*RD346e*RSe6633 - 1./270.*me2*RSe6611*
         RD346e + 1./30.*m33*RD346e*RSe6613 - 13./540.*m33*RD346e*
         RSe6633 + 1./36.*m33*RSe6611*RD346e + 2./27.*m33*me2*RD346e*
         RN3e*RSe6613 - 2./27.*m33*me2*RD346e*RN3e*RSe6633 - 4./27.*
         pow(m33,2)*RSe6633*RS3461e + 2./27.*pow(m33,2)*RD346e*RN3e*
         RSe6613 + 2./27.*pow(m33,2)*RD346e*RN3e*RSe6633 - 1./270.*m11*
         RD346e*RSe6613 - 1./36.*m11*RD346e*RSe6633 + 17./540.*m11*
         RSe6611*RD346e - 2./27.*m11*m33*RSe6633*RS3461e + 2./27.*m11*
         m33*RD346e*RN3e*RSe6613 );

      M13p6x22R +=  + Ab(me2,mu2)*Bb(me2,mu2) * (  - 4./81.*m66*me2*
         pow(RSe63,2) + 8./81.*m66*me2*RSe61*RSe63 - 4./81.*m66*me2*
         pow(RSe61,2) - 1./162.*m44*me2*pow(RSe63,2) + 1./81.*m44*me2*
         RSe61*RSe63 - 1./162.*m44*me2*pow(RSe61,2) + 4./81.*m33*m66*
         pow(RSe63,2) + 8./81.*m33*m66*RSe61*RSe63 + 1./162.*m33*m44*
         pow(RSe63,2) + 1./81.*m33*m44*RSe61*RSe63 + 8./81.*m11*m66*
         RSe61*RSe63 + 4./81.*m11*m66*pow(RSe61,2) + 1./81.*m11*m44*
         RSe61*RSe63 + 1./162.*m11*m44*pow(RSe61,2) );

      M13p6x22R +=  + Bb(m11,mu2)*pi16 * ( 1./144.*m33*RD146e*RD346e + 
         1./144.*m33*pow(RD146e,2) + 1./36.*m33*me2*RD146e*RSe6613 - 1./
         72.*m33*me2*RD146e*RSe6633 - 1./72.*m33*me2*RD146e*RSe6611 - 1.
         /48.*m33*m66*RD146e - 1./24.*m33*m44*RD146e - 1./48.*pow(
         m33,2)*RD146e - 1./54.*pow(m33,2)*RD146e*RS3461e - 1./216.*
         pow(m33,2)*RD146e*RS1463e + 1./108.*pow(m33,2)*RD146e*RSe6613
          - 1./216.*pow(m33,2)*RD146e*RSe6633 + 1./108.*pow(m33,2)*
         RD146e*RC346e - 1./216.*pow(m33,2)*RD146e*RSe6611 + 1./432.*
         pow(m33,2)*RD146e*RC146e + 1./144.*m11*RD146e*RD346e + 1./144.
         *m11*pow(RD146e,2) + 1./36.*m11*me2*RD146e*RSe6613 - 1./72.*
         m11*me2*RD146e*RSe6633 - 1./72.*m11*me2*RD146e*RSe6611 - 1./48.
         *m11*m66*RD146e - 1./24.*m11*m44*RD146e - 1./48.*m11*m33*
         RD146e - 5./216.*m11*m33*RD146e*RS3461e - 5./216.*m11*m33*
         RD146e*RS1463e + 1./54.*m11*m33*RD146e*RSe6613 - 1./108.*m11*
         m33*RD146e*RSe6633 + 5./432.*m11*m33*RD146e*RC346e - 1./108.*
         m11*m33*RD146e*RSe6611 );

      M13p6x22R +=  + Bb(m11,mu2)*pi16 * ( 5./432.*m11*m33*RD146e*
         RC146e - 1./216.*pow(m11,2)*RD146e*RS3461e - 1./54.*pow(m11,2)
         *RD146e*RS1463e + 1./108.*pow(m11,2)*RD146e*RSe6613 - 1./216.*
         pow(m11,2)*RD146e*RSe6633 + 1./432.*pow(m11,2)*RD146e*RC346e
          - 1./216.*pow(m11,2)*RD146e*RSe6611 + 1./108.*pow(m11,2)*
         RD146e*RC146e );

      M13p6x22R +=  + pow(Bb(m11,mu2),2) * ( 1./144.*m33*pow(RD146e,2)
          - 5./432.*m11*pow(RD146e,2) + 1./27.*m11*m33*RD146e*RS1463e
          + 2./27.*pow(m11,2)*RD146e*RS1463e );

      M13p6x22R +=  + Bb(m11,mu2)*Bb(m33,mu2) * (  - 1./72.*m33*RD146e*
         RD346e - 1./72.*m11*RD146e*RD346e );

      M13p6x22R +=  + Bb(m33,mu2)*pi16 * ( 1./144.*m33*pow(RD346e,2) + 
         1./144.*m33*RD146e*RD346e + 1./36.*m33*me2*RD346e*RSe6613 - 1./
         72.*m33*me2*RD346e*RSe6633 - 1./72.*m33*me2*RSe6611*RD346e - 1.
         /48.*m33*m66*RD346e - 1./24.*m33*m44*RD346e - 1./54.*pow(
         m33,2)*RD346e*RS3461e - 1./216.*pow(m33,2)*RD346e*RS1463e + 1./
         108.*pow(m33,2)*RD346e*RSe6613 - 1./216.*pow(m33,2)*RD346e*
         RSe6633 + 1./108.*pow(m33,2)*RD346e*RC346e - 1./216.*pow(
         m33,2)*RSe6611*RD346e + 1./432.*pow(m33,2)*RC146e*RD346e + 1./
         144.*m11*pow(RD346e,2) + 1./144.*m11*RD146e*RD346e + 1./36.*
         m11*me2*RD346e*RSe6613 - 1./72.*m11*me2*RD346e*RSe6633 - 1./72.
         *m11*me2*RSe6611*RD346e - 1./48.*m11*m66*RD346e - 1./24.*m11*
         m44*RD346e - 1./48.*m11*m33*RD346e - 5./216.*m11*m33*RD346e*
         RS3461e - 5./216.*m11*m33*RD346e*RS1463e + 1./54.*m11*m33*
         RD346e*RSe6613 - 1./108.*m11*m33*RD346e*RSe6633 + 5./432.*m11*
         m33*RD346e*RC346e - 1./108.*m11*m33*RSe6611*RD346e + 5./432.*
         m11*m33*RC146e*RD346e );

      M13p6x22R +=  + Bb(m33,mu2)*pi16 * (  - 1./48.*pow(m11,2)*RD346e
          - 1./216.*pow(m11,2)*RD346e*RS3461e - 1./54.*pow(m11,2)*
         RD346e*RS1463e + 1./108.*pow(m11,2)*RD346e*RSe6613 - 1./216.*
         pow(m11,2)*RD346e*RSe6633 + 1./432.*pow(m11,2)*RD346e*RC346e
          - 1./216.*pow(m11,2)*RSe6611*RD346e + 1./108.*pow(m11,2)*
         RC146e*RD346e );

      M13p6x22R +=  + pow(Bb(m33,mu2),2) * (  - 5./432.*m33*pow(
         RD346e,2) + 2./27.*pow(m33,2)*RD346e*RS3461e + 1./144.*m11*
         pow(RD346e,2) + 1./27.*m11*m33*RD346e*RS3461e );

      M13p6x22R +=  + hh(1,m11,m11,m13,m13,mu2) * ( 1./324.*pow(
         RD146e,2) + 2./81.*m33*RD146e*RS1463e - 1./54.*m33*RD146e*
         RC146e + 5./32.*pow(m33,2) + 41./648.*pow(m33,2)*pow(
         RS1463e,2) - 7./216.*pow(m33,2)*RC146e*RS1463e + 5./288.*pow(
         m33,2)*pow(RC146e,2) + 1./81.*m11*RD146e*RS1463e - 1./54.*m11*
         RD146e*RC146e + 3./16.*m11*m33 + 25./324.*m11*m33*pow(
         RS1463e,2) - 1./36.*m11*m33*RC146e*RS1463e + 5./144.*m11*m33*
         pow(RC146e,2) + 1./32.*pow(m11,2) + 17./648.*pow(m11,2)*pow(
         RS1463e,2) + 1./216.*pow(m11,2)*RC146e*RS1463e + 5./288.*pow(
         m11,2)*pow(RC146e,2) );

      M13p6x22R +=  + hh(1,m11,m13,m33,m13,mu2) * ( 1./162.*RD146e*
         RD346e + 2./81.*m33*RD346e*RS1463e + 1./81.*m33*RD146e*RS3461e
          + 25./324.*pow(m33,2)*RS1463e*RS3461e - 1./72.*pow(m33,2)*
         RC346e*RS1463e - 1./72.*pow(m33,2)*RC146e*RS3461e + 1./144.*
         pow(m33,2)*RC146e*RC346e + 1./81.*m11*RD346e*RS1463e + 2./81.*
         m11*RD146e*RS3461e + 29./162.*m11*m33*RS1463e*RS3461e - 1./36.
         *m11*m33*RC346e*RS1463e - 1./36.*m11*m33*RC146e*RS3461e + 1./
         72.*m11*m33*RC146e*RC346e + 25./324.*pow(m11,2)*RS1463e*
         RS3461e - 1./72.*pow(m11,2)*RC346e*RS1463e - 1./72.*pow(m11,2)
         *RC146e*RS3461e + 1./144.*pow(m11,2)*RC146e*RC346e );

      M13p6x22R +=  + hh(1,m11,m13,me2,m13,mu2) * ( 2./81.*me2*RD146e*
         RSe6613 - 1./81.*me2*RD146e*RSe6633 - 1./81.*me2*RD146e*
         RSe6611 - 2./81.*pow(me2,2)*RSe6613*RS1463e + 1./81.*pow(
         me2,2)*RSe6633*RS1463e + 1./81.*pow(me2,2)*RSe6611*RS1463e + 1.
         /81.*pow(me2,2)*RC146e*RSe6613 - 1./162.*pow(me2,2)*RC146e*
         RSe6633 - 1./162.*pow(me2,2)*RC146e*RSe6611 - 4./81.*m33*
         RD146e*RSe6613 + 1./81.*m33*RD146e*RSe6633 + 1./27.*m33*RD146e
         *RSe6611 + 10./81.*m33*me2*RSe6613*RS1463e - 2./27.*m33*me2*
         RSe6633*RS1463e - 4./81.*m33*me2*RSe6611*RS1463e - 4./81.*m33*
         me2*RC146e*RSe6613 + 1./81.*m33*me2*RC146e*RSe6633 + 1./27.*
         m33*me2*RC146e*RSe6611 - 41./162.*pow(m33,2)*RSe6613*RS1463e
          + 29./324.*pow(m33,2)*RSe6633*RS1463e + 7./108.*pow(m33,2)*
         RSe6611*RS1463e + 7./108.*pow(m33,2)*RC146e*RSe6613 - 13./648.
         *pow(m33,2)*RC146e*RSe6633 - 5./72.*pow(m33,2)*RC146e*RSe6611
          - 4./81.*m11*RD146e*RSe6613 + 4./81.*m11*RD146e*RSe6611 + 2./
         27.*m11*me2*RSe6613*RS1463e );

      M13p6x22R +=  + hh(1,m11,m13,me2,m13,mu2) * (  - 2./81.*m11*me2*
         RSe6633*RS1463e - 4./81.*m11*me2*RSe6611*RS1463e - 4./81.*m11*
         me2*RC146e*RSe6613 + 4./81.*m11*me2*RC146e*RSe6611 - 35./81.*
         m11*m33*RSe6613*RS1463e + 13./162.*m11*m33*RSe6633*RS1463e + 
         17./162.*m11*m33*RSe6611*RS1463e + 17./162.*m11*m33*RC146e*
         RSe6613 - 1./36.*m11*m33*RC146e*RSe6633 - 19./108.*m11*m33*
         RC146e*RSe6611 - 25./162.*pow(m11,2)*RSe6613*RS1463e + 1./36.*
         pow(m11,2)*RSe6633*RS1463e + 1./36.*pow(m11,2)*RSe6611*RS1463e
          + 1./36.*pow(m11,2)*RC146e*RSe6613 - 1./72.*pow(m11,2)*RC146e
         *RSe6633 - 73./648.*pow(m11,2)*RC146e*RSe6611 );

      M13p6x22R +=  + hh(1,m11,m14,m34,m13,mu2) * ( 1./12.*m33*m44*
         RS1463e - 1./24.*m33*m44*RC146e - 1./8.*pow(m33,2)*RS1463e + 1.
         /8.*pow(m33,2)*RS16e + 1./12.*m11*m44*RS1463e - 1./24.*m11*m44
         *RC146e - 1./3.*m11*m33*RS1463e + 5./12.*m11*m33*RS16e - 1./24.
         *m11*m33*RC146e - 5./24.*pow(m11,2)*RS1463e + 7./24.*pow(
         m11,2)*RS16e - 1./24.*pow(m11,2)*RC146e );

      M13p6x22R +=  + hh(1,m11,m16,m36,m13,mu2) * ( 1./24.*m33*m66*
         RS1463e - 1./48.*m33*m66*RC146e - 1./16.*pow(m33,2)*RS1463e + 
         1./16.*pow(m33,2)*RS14e + 1./24.*m11*m66*RS1463e - 1./48.*m11*
         m66*RC146e - 1./6.*m11*m33*RS1463e + 5./24.*m11*m33*RS14e - 1./
         48.*m11*m33*RC146e - 5./48.*pow(m11,2)*RS1463e + 7./48.*pow(
         m11,2)*RS14e - 1./48.*pow(m11,2)*RC146e );

      M13p6x22R +=  + hh(1,m13,m13,m13,m13,mu2) * ( 1./16.*pow(m33,2)
          + 1./8.*m11*m33 + 1./16.*pow(m11,2) );

      M13p6x22R +=  + hh(1,m13,m33,m33,m13,mu2) * ( 1./324.*pow(
         RD346e,2) + 1./81.*m33*RD346e*RS3461e - 1./54.*m33*RD346e*
         RC346e + 1./32.*pow(m33,2) + 17./648.*pow(m33,2)*pow(
         RS3461e,2) + 1./216.*pow(m33,2)*RC346e*RS3461e + 5./288.*pow(
         m33,2)*pow(RC346e,2) + 2./81.*m11*RD346e*RS3461e - 1./54.*m11*
         RD346e*RC346e + 3./16.*m11*m33 + 25./324.*m11*m33*pow(
         RS3461e,2) - 1./36.*m11*m33*RC346e*RS3461e + 5./144.*m11*m33*
         pow(RC346e,2) + 5./32.*pow(m11,2) + 41./648.*pow(m11,2)*pow(
         RS3461e,2) - 7./216.*pow(m11,2)*RC346e*RS3461e + 5./288.*pow(
         m11,2)*pow(RC346e,2) );

      M13p6x22R +=  + hh(1,m13,m33,me2,m13,mu2) * ( 2./81.*me2*RD346e*
         RSe6613 - 1./81.*me2*RD346e*RSe6633 - 1./81.*me2*RSe6611*
         RD346e - 2./81.*pow(me2,2)*RSe6613*RS3461e + 1./81.*pow(me2,2)
         *RSe6633*RS3461e + 1./81.*pow(me2,2)*RC346e*RSe6613 - 1./162.*
         pow(me2,2)*RC346e*RSe6633 + 1./81.*pow(me2,2)*RSe6611*RS3461e
          - 1./162.*pow(me2,2)*RSe6611*RC346e - 4./81.*m33*RD346e*
         RSe6613 + 4./81.*m33*RD346e*RSe6633 + 2./27.*m33*me2*RSe6613*
         RS3461e - 4./81.*m33*me2*RSe6633*RS3461e - 4./81.*m33*me2*
         RC346e*RSe6613 + 4./81.*m33*me2*RC346e*RSe6633 - 2./81.*m33*
         me2*RSe6611*RS3461e - 25./162.*pow(m33,2)*RSe6613*RS3461e + 1./
         36.*pow(m33,2)*RSe6633*RS3461e + 1./36.*pow(m33,2)*RC346e*
         RSe6613 - 73./648.*pow(m33,2)*RC346e*RSe6633 + 1./36.*pow(
         m33,2)*RSe6611*RS3461e - 1./72.*pow(m33,2)*RSe6611*RC346e - 4./
         81.*m11*RD346e*RSe6613 + 1./27.*m11*RD346e*RSe6633 + 1./81.*
         m11*RSe6611*RD346e + 10./81.*m11*me2*RSe6613*RS3461e - 4./81.*
         m11*me2*RSe6633*RS3461e );

      M13p6x22R +=  + hh(1,m13,m33,me2,m13,mu2) * (  - 4./81.*m11*me2*
         RC346e*RSe6613 + 1./27.*m11*me2*RC346e*RSe6633 - 2./27.*m11*
         me2*RSe6611*RS3461e + 1./81.*m11*me2*RSe6611*RC346e - 35./81.*
         m11*m33*RSe6613*RS3461e + 17./162.*m11*m33*RSe6633*RS3461e + 
         17./162.*m11*m33*RC346e*RSe6613 - 19./108.*m11*m33*RC346e*
         RSe6633 + 13./162.*m11*m33*RSe6611*RS3461e - 1./36.*m11*m33*
         RSe6611*RC346e - 41./162.*pow(m11,2)*RSe6613*RS3461e + 7./108.
         *pow(m11,2)*RSe6633*RS3461e + 7./108.*pow(m11,2)*RC346e*
         RSe6613 - 5./72.*pow(m11,2)*RC346e*RSe6633 + 29./324.*pow(
         m11,2)*RSe6611*RS3461e - 13./648.*pow(m11,2)*RSe6611*RC346e );

      M13p6x22R +=  + hh(1,m13,me2,me2,m13,mu2) * ( 8./81.*pow(me2,2)*
         pow(RSe6613,2) - 8./81.*pow(me2,2)*RSe6633*RSe6613 + 2./81.*
         pow(me2,2)*pow(RSe6633,2) - 8./81.*pow(me2,2)*RSe6611*RSe6613
          + 4./81.*pow(me2,2)*RSe6611*RSe6633 + 2./81.*pow(me2,2)*pow(
         RSe6611,2) - 20./81.*m33*me2*pow(RSe6613,2) + 8./27.*m33*me2*
         RSe6633*RSe6613 - 10./81.*m33*me2*pow(RSe6633,2) + 16./81.*m33
         *me2*RSe6611*RSe6613 - 4./81.*m33*me2*RSe6611*RSe6633 - 2./27.
         *m33*me2*pow(RSe6611,2) + 41./162.*pow(m33,2)*pow(RSe6613,2)
          - 29./162.*pow(m33,2)*RSe6633*RSe6613 + 109./648.*pow(m33,2)*
         pow(RSe6633,2) - 7./54.*pow(m33,2)*RSe6611*RSe6613 + 13./324.*
         pow(m33,2)*RSe6611*RSe6633 + 5./72.*pow(m33,2)*pow(RSe6611,2)
          - 20./81.*m11*me2*pow(RSe6613,2) + 16./81.*m11*me2*RSe6633*
         RSe6613 - 2./27.*m11*me2*pow(RSe6633,2) + 8./27.*m11*me2*
         RSe6611*RSe6613 - 4./81.*m11*me2*RSe6611*RSe6633 - 10./81.*m11
         *me2*pow(RSe6611,2) + 5./9.*m11*m33*pow(RSe6613,2) - 25./81.*
         m11*m33*RSe6633*RSe6613 );

      M13p6x22R +=  + hh(1,m13,me2,me2,m13,mu2) * ( 23./108.*m11*m33*
         pow(RSe6633,2) - 25./81.*m11*m33*RSe6611*RSe6613 + 13./162.*
         m11*m33*RSe6611*RSe6633 + 23./108.*m11*m33*pow(RSe6611,2) + 41.
         /162.*pow(m11,2)*pow(RSe6613,2) - 7./54.*pow(m11,2)*RSe6633*
         RSe6613 + 5./72.*pow(m11,2)*pow(RSe6633,2) - 29./162.*pow(
         m11,2)*RSe6611*RSe6613 + 13./324.*pow(m11,2)*RSe6611*RSe6633
          + 109./648.*pow(m11,2)*pow(RSe6611,2) );

      M13p6x22R +=  + hh(1,m14,m33,m34,m13,mu2) * ( 1./12.*m33*m44*
         RS3461e - 1./24.*m33*m44*RC346e + 1./8.*pow(m33,2)*RS3461e - 1.
         /24.*pow(m33,2)*RS36e - 1./24.*pow(m33,2)*RC346e + 1./12.*m11*
         m44*RS3461e - 1./24.*m11*m44*RC346e + 1./3.*m11*m33*RS3461e - 
         1./4.*m11*m33*RS36e - 1./24.*m11*m33*RC346e + 5./24.*pow(
         m11,2)*RS3461e - 5./24.*pow(m11,2)*RS36e );

      M13p6x22R +=  + hh(1,m14,m34,m44,m13,mu2) * ( 3./8.*m33*m44 + 3./
         8.*m11*m44 );

      M13p6x22R +=  + hh(1,m14,m34,me2,m13,mu2) * ( 1./27.*pow(me2,2)
          + 2./27.*pow(me2,2)*RSe6613 + 2./27.*pow(me2,2)*RSe63 + 1./27.
         *pow(me2,2)*RSe6633 + 2./27.*pow(me2,2)*RSe61 + 1./27.*pow(
         me2,2)*RSe6611 - 2./27.*m44*me2 - 2./27.*m44*me2*RSe63 - 2./27.
         *m44*me2*RSe61 + 1./27.*pow(m44,2) + 1./24.*m33*me2 + 17./54.*
         m33*me2*RSe6613 + 25./108.*m33*me2*RSe63 + 4./27.*m33*me2*
         RSe6633 + 1./12.*m33*me2*RSe61 - 1./6.*m33*m44*RSe6613 - 4./27.
         *m33*m44*RSe63 + 1./12.*m33*m44*RSe6633 + 1./12.*m33*m44*
         RSe6611 + 1./4.*pow(m33,2)*RSe6613 - 1./8.*pow(m33,2)*RSe63 - 
         11./108.*pow(m33,2)*RSe6633 + 1./8.*pow(m33,2)*RSe61 + 1./24.*
         m11*me2 + 17./54.*m11*me2*RSe6613 + 11./36.*m11*me2*RSe63 + 2./
         9.*m11*me2*RSe6633 + 1./108.*m11*me2*RSe61 - 2./27.*m11*me2*
         RSe6611 - 1./6.*m11*m44*RSe6613 - 2./9.*m11*m44*RSe63 + 1./12.
         *m11*m44*RSe6633 + 2./27.*m11*m44*RSe61 + 1./12.*m11*m44*
         RSe6611 + 19./54.*m11*m33*RSe6613 - 1./3.*m11*m33*RSe63 - 5./
         36.*m11*m33*RSe6633 );

      M13p6x22R +=  + hh(1,m14,m34,me2,m13,mu2) * ( 1./3.*m11*m33*RSe61
          + 1./12.*m11*m33*RSe6611 + 1./36.*pow(m11,2)*RSe6613 - 5./24.
         *pow(m11,2)*RSe63 + 5./24.*pow(m11,2)*RSe61 + 13./108.*pow(
         m11,2)*RSe6611 );

      M13p6x22R +=  + hh(1,m14,m36,m46,m13,mu2) * ( 1./8.*m33*m66 + 1./
         8.*m33*m44 + 1./8.*m11*m66 + 1./8.*m11*m44 );

      M13p6x22R +=  + hh(1,m16,m33,m36,m13,mu2) * ( 1./24.*m33*m66*
         RS3461e - 1./48.*m33*m66*RC346e + 1./16.*pow(m33,2)*RS3461e - 
         1./48.*pow(m33,2)*RS34e - 1./48.*pow(m33,2)*RC346e + 1./24.*
         m11*m66*RS3461e - 1./48.*m11*m66*RC346e + 1./6.*m11*m33*
         RS3461e - 1./8.*m11*m33*RS34e - 1./48.*m11*m33*RC346e + 5./48.
         *pow(m11,2)*RS3461e - 5./48.*pow(m11,2)*RS34e );

      M13p6x22R +=  + hh(1,m16,m34,m46,m13,mu2) * ( 1./8.*m33*m66 + 1./
         8.*m33*m44 + 1./8.*m11*m66 + 1./8.*m11*m44 );

      M13p6x22R +=  + hh(1,m16,m36,me2,m13,mu2) * ( 2./27.*pow(me2,2)
          + 1./27.*pow(me2,2)*RSe6613 - 2./27.*pow(me2,2)*RSe63 + 1./54.
         *pow(me2,2)*RSe6633 - 2./27.*pow(me2,2)*RSe61 + 1./54.*pow(
         me2,2)*RSe6611 - 4./27.*m66*me2 + 2./27.*m66*me2*RSe63 + 2./27.
         *m66*me2*RSe61 + 2./27.*pow(m66,2) + 1./12.*m33*me2 + 17./108.
         *m33*me2*RSe6613 - 25./108.*m33*me2*RSe63 + 2./27.*m33*me2*
         RSe6633 - 1./12.*m33*me2*RSe61 - 1./12.*m33*m66*RSe6613 + 4./
         27.*m33*m66*RSe63 + 1./24.*m33*m66*RSe6633 + 1./24.*m33*m66*
         RSe6611 + 1./8.*pow(m33,2)*RSe6613 + 1./8.*pow(m33,2)*RSe63 - 
         11./216.*pow(m33,2)*RSe6633 - 1./8.*pow(m33,2)*RSe61 + 1./12.*
         m11*me2 + 17./108.*m11*me2*RSe6613 - 11./36.*m11*me2*RSe63 + 1.
         /9.*m11*me2*RSe6633 - 1./108.*m11*me2*RSe61 - 1./27.*m11*me2*
         RSe6611 - 1./12.*m11*m66*RSe6613 + 2./9.*m11*m66*RSe63 + 1./24.
         *m11*m66*RSe6633 - 2./27.*m11*m66*RSe61 + 1./24.*m11*m66*
         RSe6611 + 19./108.*m11*m33*RSe6613 + 1./3.*m11*m33*RSe63 - 5./
         72.*m11*m33*RSe6633 );

      M13p6x22R +=  + hh(1,m16,m36,me2,m13,mu2) * (  - 1./3.*m11*m33*
         RSe61 + 1./24.*m11*m33*RSe6611 + 1./72.*pow(m11,2)*RSe6613 + 5.
         /24.*pow(m11,2)*RSe63 - 5./24.*pow(m11,2)*RSe61 + 13./216.*
         pow(m11,2)*RSe6611 );

      M13p6x22R +=  + hh(2,m11,m11,m13,m13,mu2) * (  - 1./54.*m33*pow(
         RD146e,2) - 7./216.*pow(m33,2)*RD146e*RS1463e + 5./144.*pow(
         m33,2)*RD146e*RC146e - 1./54.*m11*pow(RD146e,2) - 1./36.*m11*
         m33*RD146e*RS1463e + 5./72.*m11*m33*RD146e*RC146e + 1./216.*
         pow(m11,2)*RD146e*RS1463e + 5./144.*pow(m11,2)*RD146e*RC146e )
         ;

      M13p6x22R +=  + hh(2,m11,m13,m33,m13,mu2) * (  - 1./72.*pow(
         m33,2)*RD146e*RS3461e + 1./144.*pow(m33,2)*RD146e*RC346e - 1./
         36.*m11*m33*RD146e*RS3461e + 1./72.*m11*m33*RD146e*RC346e - 1./
         72.*pow(m11,2)*RD146e*RS3461e + 1./144.*pow(m11,2)*RD146e*
         RC346e );

      M13p6x22R +=  + hh(2,m11,m13,me2,m13,mu2) * ( 1./81.*pow(me2,2)*
         RD146e*RSe6613 - 1./162.*pow(me2,2)*RD146e*RSe6633 - 1./162.*
         pow(me2,2)*RD146e*RSe6611 - 4./81.*m33*me2*RD146e*RSe6613 + 1./
         81.*m33*me2*RD146e*RSe6633 + 1./27.*m33*me2*RD146e*RSe6611 + 7.
         /108.*pow(m33,2)*RD146e*RSe6613 - 13./648.*pow(m33,2)*RD146e*
         RSe6633 - 5./72.*pow(m33,2)*RD146e*RSe6611 - 4./81.*m11*me2*
         RD146e*RSe6613 + 4./81.*m11*me2*RD146e*RSe6611 + 17./162.*m11*
         m33*RD146e*RSe6613 - 1./36.*m11*m33*RD146e*RSe6633 - 19./108.*
         m11*m33*RD146e*RSe6611 + 1./36.*pow(m11,2)*RD146e*RSe6613 - 1./
         72.*pow(m11,2)*RD146e*RSe6633 - 73./648.*pow(m11,2)*RD146e*
         RSe6611 );

      M13p6x22R +=  + hh(2,m11,m14,m34,m13,mu2) * (  - 1./24.*m33*m44*
         RD146e - 1./24.*m11*m44*RD146e - 1./24.*m11*m33*RD146e - 1./24.
         *pow(m11,2)*RD146e );

      M13p6x22R +=  + hh(2,m11,m16,m36,m13,mu2) * (  - 1./48.*m33*m66*
         RD146e - 1./48.*m11*m66*RD146e - 1./48.*m11*m33*RD146e - 1./48.
         *pow(m11,2)*RD146e );

      M13p6x22R +=  + hh(2,m33,m11,m13,m13,mu2) * (  - 1./72.*pow(
         m33,2)*RD346e*RS1463e + 1./144.*pow(m33,2)*RC146e*RD346e - 1./
         36.*m11*m33*RD346e*RS1463e + 1./72.*m11*m33*RC146e*RD346e - 1./
         72.*pow(m11,2)*RD346e*RS1463e + 1./144.*pow(m11,2)*RC146e*
         RD346e );

      M13p6x22R +=  + hh(2,m33,m13,m33,m13,mu2) * (  - 1./54.*m33*pow(
         RD346e,2) + 1./216.*pow(m33,2)*RD346e*RS3461e + 5./144.*pow(
         m33,2)*RD346e*RC346e - 1./54.*m11*pow(RD346e,2) - 1./36.*m11*
         m33*RD346e*RS3461e + 5./72.*m11*m33*RD346e*RC346e - 7./216.*
         pow(m11,2)*RD346e*RS3461e + 5./144.*pow(m11,2)*RD346e*RC346e )
         ;

      M13p6x22R +=  + hh(2,m33,m13,me2,m13,mu2) * ( 1./81.*pow(me2,2)*
         RD346e*RSe6613 - 1./162.*pow(me2,2)*RD346e*RSe6633 - 1./162.*
         pow(me2,2)*RSe6611*RD346e - 4./81.*m33*me2*RD346e*RSe6613 + 4./
         81.*m33*me2*RD346e*RSe6633 + 1./36.*pow(m33,2)*RD346e*RSe6613
          - 73./648.*pow(m33,2)*RD346e*RSe6633 - 1./72.*pow(m33,2)*
         RSe6611*RD346e - 4./81.*m11*me2*RD346e*RSe6613 + 1./27.*m11*
         me2*RD346e*RSe6633 + 1./81.*m11*me2*RSe6611*RD346e + 17./162.*
         m11*m33*RD346e*RSe6613 - 19./108.*m11*m33*RD346e*RSe6633 - 1./
         36.*m11*m33*RSe6611*RD346e + 7./108.*pow(m11,2)*RD346e*RSe6613
          - 5./72.*pow(m11,2)*RD346e*RSe6633 - 13./648.*pow(m11,2)*
         RSe6611*RD346e );

      M13p6x22R +=  + hh(2,m33,m14,m34,m13,mu2) * (  - 1./24.*m33*m44*
         RD346e - 1./24.*pow(m33,2)*RD346e - 1./24.*m11*m44*RD346e - 1./
         24.*m11*m33*RD346e );

      M13p6x22R +=  + hh(2,m33,m16,m36,m13,mu2) * (  - 1./48.*m33*m66*
         RD346e - 1./48.*pow(m33,2)*RD346e - 1./48.*m11*m66*RD346e - 1./
         48.*m11*m33*RD346e );

      M13p6x22R +=  + hh(5,m11,m11,m13,m13,mu2) * ( 5./288.*pow(m33,2)*
         pow(RD146e,2) + 5./144.*m11*m33*pow(RD146e,2) + 5./288.*pow(
         m11,2)*pow(RD146e,2) );

      M13p6x22R +=  + hh(5,m11,m33,m13,m13,mu2) * ( 1./144.*pow(m33,2)*
         RD146e*RD346e + 1./72.*m11*m33*RD146e*RD346e + 1./144.*pow(
         m11,2)*RD146e*RD346e );

      M13p6x22R +=  + hh(5,m33,m33,m13,m13,mu2) * ( 5./288.*pow(m33,2)*
         pow(RD346e,2) + 5./144.*m11*m33*pow(RD346e,2) + 5./288.*pow(
         m11,2)*pow(RD346e,2) );

      M13p6x22R +=  + hh1(1,m11,m14,m34,m13,mu2) * ( 1./3.*pow(m33,2)*
         RS1463e - 1./3.*pow(m33,2)*RS16e + 2./3.*m11*m33*RS1463e - 2./
         3.*m11*m33*RS16e + 1./3.*pow(m11,2)*RS1463e - 1./3.*pow(m11,2)
         *RS16e );

      M13p6x22R +=  + hh1(1,m11,m16,m36,m13,mu2) * ( 1./6.*pow(m33,2)*
         RS1463e - 1./6.*pow(m33,2)*RS14e + 1./3.*m11*m33*RS1463e - 1./
         3.*m11*m33*RS14e + 1./6.*pow(m11,2)*RS1463e - 1./6.*pow(m11,2)
         *RS14e );

      M13p6x22R +=  + hh1(1,m13,m11,m11,m13,mu2) * (  - 1./27.*m33*
         RD146e*RS1463e + 1./54.*m33*RD146e*RC146e - 1./4.*pow(m33,2)
          - 5./54.*pow(m33,2)*pow(RS1463e,2) + 2./27.*pow(m33,2)*RC146e
         *RS1463e - 1./36.*pow(m33,2)*pow(RC146e,2) - 1./27.*m11*RD146e
         *RS1463e + 1./54.*m11*RD146e*RC146e - 1./2.*m11*m33 - 1./9.*
         m11*m33*pow(RS1463e,2) + 1./9.*m11*m33*RC146e*RS1463e - 1./18.
         *m11*m33*pow(RC146e,2) - 1./4.*pow(m11,2) - 1./54.*pow(m11,2)*
         pow(RS1463e,2) + 1./27.*pow(m11,2)*RC146e*RS1463e - 1./36.*
         pow(m11,2)*pow(RC146e,2) );

      M13p6x22R +=  + hh1(1,m13,m11,m33,m13,mu2) * (  - 1./27.*m33*
         RD346e*RS1463e + 1./54.*m33*RC146e*RD346e - 1./27.*m33*RD146e*
         RS3461e + 1./54.*m33*RD146e*RC346e - 1./9.*pow(m33,2)*RS1463e*
         RS3461e + 2./27.*pow(m33,2)*RC346e*RS1463e + 1./27.*pow(m33,2)
         *RC146e*RS3461e - 1./27.*m11*RD346e*RS1463e + 1./54.*m11*
         RC146e*RD346e - 1./27.*m11*RD146e*RS3461e + 1./54.*m11*RD146e*
         RC346e - 2./9.*m11*m33*RS1463e*RS3461e + 1./9.*m11*m33*RC346e*
         RS1463e + 1./9.*m11*m33*RC146e*RS3461e - 1./9.*pow(m11,2)*
         RS1463e*RS3461e + 1./27.*pow(m11,2)*RC346e*RS1463e + 2./27.*
         pow(m11,2)*RC146e*RS3461e );

      M13p6x22R +=  + hh1(1,m13,m11,me2,m13,mu2) * ( 2./27.*m33*RD146e*
         RSe6613 - 1./27.*m33*RD146e*RSe6633 - 1./27.*m33*RD146e*
         RSe6611 - 4./27.*m33*me2*RSe6613*RS1463e + 2./27.*m33*me2*
         RSe6633*RS1463e + 2./27.*m33*me2*RSe6611*RS1463e + 2./27.*m33*
         me2*RC146e*RSe6613 - 1./27.*m33*me2*RC146e*RSe6633 - 1./27.*
         m33*me2*RC146e*RSe6611 + 10./27.*pow(m33,2)*RSe6613*RS1463e - 
         2./9.*pow(m33,2)*RSe6633*RS1463e - 4./27.*pow(m33,2)*RSe6611*
         RS1463e - 4./27.*pow(m33,2)*RC146e*RSe6613 + 1./27.*pow(m33,2)
         *RC146e*RSe6633 + 1./9.*pow(m33,2)*RC146e*RSe6611 + 2./27.*m11
         *RD146e*RSe6613 - 1./27.*m11*RD146e*RSe6633 - 1./27.*m11*
         RD146e*RSe6611 - 4./27.*m11*me2*RSe6613*RS1463e + 2./27.*m11*
         me2*RSe6633*RS1463e + 2./27.*m11*me2*RSe6611*RS1463e + 2./27.*
         m11*me2*RC146e*RSe6613 - 1./27.*m11*me2*RC146e*RSe6633 - 1./27.
         *m11*me2*RC146e*RSe6611 + 16./27.*m11*m33*RSe6613*RS1463e - 8./
         27.*m11*m33*RSe6633*RS1463e - 8./27.*m11*m33*RSe6611*RS1463e
          - 8./27.*m11*m33*RC146e*RSe6613 );

      M13p6x22R +=  + hh1(1,m13,m11,me2,m13,mu2) * ( 1./27.*m11*m33*
         RC146e*RSe6633 + 7./27.*m11*m33*RC146e*RSe6611 + 2./9.*pow(
         m11,2)*RSe6613*RS1463e - 2./27.*pow(m11,2)*RSe6633*RS1463e - 4.
         /27.*pow(m11,2)*RSe6611*RS1463e - 4./27.*pow(m11,2)*RC146e*
         RSe6613 + 4./27.*pow(m11,2)*RC146e*RSe6611 );

      M13p6x22R +=  + hh1(1,m13,m33,m33,m13,mu2) * (  - 1./27.*m33*
         RD346e*RS3461e + 1./54.*m33*RD346e*RC346e - 1./4.*pow(m33,2)
          - 1./54.*pow(m33,2)*pow(RS3461e,2) + 1./27.*pow(m33,2)*RC346e
         *RS3461e - 1./36.*pow(m33,2)*pow(RC346e,2) - 1./27.*m11*RD346e
         *RS3461e + 1./54.*m11*RD346e*RC346e - 1./2.*m11*m33 - 1./9.*
         m11*m33*pow(RS3461e,2) + 1./9.*m11*m33*RC346e*RS3461e - 1./18.
         *m11*m33*pow(RC346e,2) - 1./4.*pow(m11,2) - 5./54.*pow(m11,2)*
         pow(RS3461e,2) + 2./27.*pow(m11,2)*RC346e*RS3461e - 1./36.*
         pow(m11,2)*pow(RC346e,2) );

      M13p6x22R +=  + hh1(1,m13,m33,me2,m13,mu2) * ( 2./27.*m33*RD346e*
         RSe6613 - 1./27.*m33*RD346e*RSe6633 - 1./27.*m33*RSe6611*
         RD346e - 4./27.*m33*me2*RSe6613*RS3461e + 2./27.*m33*me2*
         RSe6633*RS3461e + 2./27.*m33*me2*RC346e*RSe6613 - 1./27.*m33*
         me2*RC346e*RSe6633 + 2./27.*m33*me2*RSe6611*RS3461e - 1./27.*
         m33*me2*RSe6611*RC346e + 2./9.*pow(m33,2)*RSe6613*RS3461e - 4./
         27.*pow(m33,2)*RSe6633*RS3461e - 4./27.*pow(m33,2)*RC346e*
         RSe6613 + 4./27.*pow(m33,2)*RC346e*RSe6633 - 2./27.*pow(m33,2)
         *RSe6611*RS3461e + 2./27.*m11*RD346e*RSe6613 - 1./27.*m11*
         RD346e*RSe6633 - 1./27.*m11*RSe6611*RD346e - 4./27.*m11*me2*
         RSe6613*RS3461e + 2./27.*m11*me2*RSe6633*RS3461e + 2./27.*m11*
         me2*RC346e*RSe6613 - 1./27.*m11*me2*RC346e*RSe6633 + 2./27.*
         m11*me2*RSe6611*RS3461e - 1./27.*m11*me2*RSe6611*RC346e + 16./
         27.*m11*m33*RSe6613*RS3461e - 8./27.*m11*m33*RSe6633*RS3461e
          - 8./27.*m11*m33*RC346e*RSe6613 + 7./27.*m11*m33*RC346e*
         RSe6633 );

      M13p6x22R +=  + hh1(1,m13,m33,me2,m13,mu2) * (  - 8./27.*m11*m33*
         RSe6611*RS3461e + 1./27.*m11*m33*RSe6611*RC346e + 10./27.*pow(
         m11,2)*RSe6613*RS3461e - 4./27.*pow(m11,2)*RSe6633*RS3461e - 4.
         /27.*pow(m11,2)*RC346e*RSe6613 + 1./9.*pow(m11,2)*RC346e*
         RSe6633 - 2./9.*pow(m11,2)*RSe6611*RS3461e + 1./27.*pow(m11,2)
         *RSe6611*RC346e );

      M13p6x22R +=  + hh1(1,m13,me2,me2,m13,mu2) * ( 8./27.*m33*me2*
         pow(RSe6613,2) - 8./27.*m33*me2*RSe6633*RSe6613 + 2./27.*m33*
         me2*pow(RSe6633,2) - 8./27.*m33*me2*RSe6611*RSe6613 + 4./27.*
         m33*me2*RSe6611*RSe6633 + 2./27.*m33*me2*pow(RSe6611,2) - 10./
         27.*pow(m33,2)*pow(RSe6613,2) + 4./9.*pow(m33,2)*RSe6633*
         RSe6613 - 5./27.*pow(m33,2)*pow(RSe6633,2) + 8./27.*pow(m33,2)
         *RSe6611*RSe6613 - 2./27.*pow(m33,2)*RSe6611*RSe6633 - 1./9.*
         pow(m33,2)*pow(RSe6611,2) + 8./27.*m11*me2*pow(RSe6613,2) - 8./
         27.*m11*me2*RSe6633*RSe6613 + 2./27.*m11*me2*pow(RSe6633,2) - 
         8./27.*m11*me2*RSe6611*RSe6613 + 4./27.*m11*me2*RSe6611*
         RSe6633 + 2./27.*m11*me2*pow(RSe6611,2) - 20./27.*m11*m33*pow(
         RSe6613,2) + 20./27.*m11*m33*RSe6633*RSe6613 - 8./27.*m11*m33*
         pow(RSe6633,2) + 20./27.*m11*m33*RSe6611*RSe6613 - 4./27.*m11*
         m33*RSe6611*RSe6633 - 8./27.*m11*m33*pow(RSe6611,2) - 10./27.*
         pow(m11,2)*pow(RSe6613,2) + 8./27.*pow(m11,2)*RSe6633*RSe6613
          - 1./9.*pow(m11,2)*pow(RSe6633,2) );

      M13p6x22R +=  + hh1(1,m13,me2,me2,m13,mu2) * ( 4./9.*pow(m11,2)*
         RSe6611*RSe6613 - 2./27.*pow(m11,2)*RSe6611*RSe6633 - 5./27.*
         pow(m11,2)*pow(RSe6611,2) );

      M13p6x22R +=  + hh1(1,m14,m11,m34,m13,mu2) * (  - 1./9.*m33*
         RD146e - 2./9.*m33*m44*RS16e + 1./9.*pow(m33,2)*RS1463e - 1./3.
         *pow(m33,2)*RS16e - 1./9.*m11*RD146e - 2./9.*m11*m44*RS16e - 4.
         /9.*m11*m33*RS16e - 1./9.*pow(m11,2)*RS1463e - 1./9.*pow(
         m11,2)*RS16e );

      M13p6x22R +=  + hh1(1,m14,m33,m34,m13,mu2) * (  - 1./3.*pow(
         m33,2)*RS3461e + 1./3.*pow(m33,2)*RS36e - 2./3.*m11*m33*
         RS3461e + 2./3.*m11*m33*RS36e - 1./3.*pow(m11,2)*RS3461e + 1./
         3.*pow(m11,2)*RS36e );

      M13p6x22R +=  + hh1(1,m14,m34,me2,m13,mu2) * (  - 2./9.*m33*me2*
         RSe63 - 2./9.*m33*me2*RSe6633 + 2./9.*m33*me2*RSe61 + 2./9.*
         m33*me2*RSe6611 + 2./9.*m33*m44*RSe63 - 2./9.*m33*m44*RSe61 - 
         2./9.*pow(m33,2)*RSe6613 + 1./3.*pow(m33,2)*RSe63 + 2./9.*pow(
         m33,2)*RSe6633 - 1./3.*pow(m33,2)*RSe61 - 2./9.*m11*me2*RSe63
          - 2./9.*m11*me2*RSe6633 + 2./9.*m11*me2*RSe61 + 2./9.*m11*me2
         *RSe6611 + 2./9.*m11*m44*RSe63 - 2./9.*m11*m44*RSe61 + 2./3.*
         m11*m33*RSe63 + 2./9.*m11*m33*RSe6633 - 2./3.*m11*m33*RSe61 - 
         2./9.*m11*m33*RSe6611 + 2./9.*pow(m11,2)*RSe6613 + 1./3.*pow(
         m11,2)*RSe63 - 1./3.*pow(m11,2)*RSe61 - 2./9.*pow(m11,2)*
         RSe6611 );

      M13p6x22R +=  + hh1(1,m16,m11,m36,m13,mu2) * (  - 1./18.*m33*
         RD146e - 1./9.*m33*m66*RS14e + 1./18.*pow(m33,2)*RS1463e - 1./
         6.*pow(m33,2)*RS14e - 1./18.*m11*RD146e - 1./9.*m11*m66*RS14e
          - 2./9.*m11*m33*RS14e - 1./18.*pow(m11,2)*RS1463e - 1./18.*
         pow(m11,2)*RS14e );

      M13p6x22R +=  + hh1(1,m16,m33,m36,m13,mu2) * (  - 1./6.*pow(
         m33,2)*RS3461e + 1./6.*pow(m33,2)*RS34e - 1./3.*m11*m33*
         RS3461e + 1./3.*m11*m33*RS34e - 1./6.*pow(m11,2)*RS3461e + 1./
         6.*pow(m11,2)*RS34e );

      M13p6x22R +=  + hh1(1,m16,m36,me2,m13,mu2) * ( 2./9.*m33*me2*
         RSe63 - 1./9.*m33*me2*RSe6633 - 2./9.*m33*me2*RSe61 + 1./9.*
         m33*me2*RSe6611 - 2./9.*m33*m66*RSe63 + 2./9.*m33*m66*RSe61 - 
         1./9.*pow(m33,2)*RSe6613 - 1./3.*pow(m33,2)*RSe63 + 1./9.*pow(
         m33,2)*RSe6633 + 1./3.*pow(m33,2)*RSe61 + 2./9.*m11*me2*RSe63
          - 1./9.*m11*me2*RSe6633 - 2./9.*m11*me2*RSe61 + 1./9.*m11*me2
         *RSe6611 - 2./9.*m11*m66*RSe63 + 2./9.*m11*m66*RSe61 - 2./3.*
         m11*m33*RSe63 + 1./9.*m11*m33*RSe6633 + 2./3.*m11*m33*RSe61 - 
         1./9.*m11*m33*RSe6611 + 1./9.*pow(m11,2)*RSe6613 - 1./3.*pow(
         m11,2)*RSe63 + 1./3.*pow(m11,2)*RSe61 - 1./9.*pow(m11,2)*
         RSe6611 );

      M13p6x22R +=  + hh1(1,m34,m14,m33,m13,mu2) * (  - 1./9.*m33*
         RD346e - 2./9.*m33*m44*RS36e - 4./9.*pow(m33,2)*RS3461e + 2./9.
         *pow(m33,2)*RS36e - 1./9.*m11*RD346e - 2./9.*m11*m44*RS36e - 2.
         /3.*m11*m33*RS3461e + 2./9.*m11*m33*RS36e - 2./9.*pow(m11,2)*
         RS3461e );

      M13p6x22R +=  + hh1(1,m36,m16,m33,m13,mu2) * (  - 1./18.*m33*
         RD346e - 1./9.*m33*m66*RS34e - 2./9.*pow(m33,2)*RS3461e + 1./9.
         *pow(m33,2)*RS34e - 1./18.*m11*RD346e - 1./9.*m11*m66*RS34e - 
         1./3.*m11*m33*RS3461e + 1./9.*m11*m33*RS34e - 1./9.*pow(m11,2)
         *RS3461e );

      M13p6x22R +=  + hh1(1,me2,m14,m34,m13,mu2) * (  - 1./9.*m33*me2
          - 2./9.*m33*me2*RSe6613 - 1./3.*m33*me2*RSe63 - 2./9.*m33*me2
         *RSe6633 - 1./9.*m33*me2*RSe61 + 1./9.*m33*m44 + 2./9.*m33*m44
         *RSe63 - 2./3.*pow(m33,2)*RSe6613 + 1./9.*pow(m33,2)*RSe63 + 2.
         /9.*pow(m33,2)*RSe6633 - 1./3.*pow(m33,2)*RSe61 - 1./9.*m11*
         me2 - 2./9.*m11*me2*RSe6613 - 1./3.*m11*me2*RSe63 - 2./9.*m11*
         me2*RSe6633 - 1./9.*m11*me2*RSe61 + 1./9.*m11*m44 + 2./9.*m11*
         m44*RSe63 - 10./9.*m11*m33*RSe6613 + 1./9.*m11*m33*RSe63 + 2./
         9.*m11*m33*RSe6633 - 5./9.*m11*m33*RSe61 - 4./9.*pow(m11,2)*
         RSe6613 - 2./9.*pow(m11,2)*RSe61 );

      M13p6x22R +=  + hh1(1,me2,m16,m36,m13,mu2) * (  - 2./9.*m33*me2
          - 1./9.*m33*me2*RSe6613 + 1./3.*m33*me2*RSe63 - 1./9.*m33*me2
         *RSe6633 + 1./9.*m33*me2*RSe61 + 2./9.*m33*m66 - 2./9.*m33*m66
         *RSe63 - 1./3.*pow(m33,2)*RSe6613 - 1./9.*pow(m33,2)*RSe63 + 1.
         /9.*pow(m33,2)*RSe6633 + 1./3.*pow(m33,2)*RSe61 - 2./9.*m11*
         me2 - 1./9.*m11*me2*RSe6613 + 1./3.*m11*me2*RSe63 - 1./9.*m11*
         me2*RSe6633 + 1./9.*m11*me2*RSe61 + 2./9.*m11*m66 - 2./9.*m11*
         m66*RSe63 - 5./9.*m11*m33*RSe6613 - 1./9.*m11*m33*RSe63 + 1./9.
         *m11*m33*RSe6633 + 5./9.*m11*m33*RSe61 - 2./9.*pow(m11,2)*
         RSe6613 + 2./9.*pow(m11,2)*RSe61 );

      M13p6x22R +=  + hh1(3,m13,m11,m11,m13,mu2) * ( 1./54.*m33*pow(
         RD146e,2) + 2./27.*pow(m33,2)*RD146e*RS1463e - 1./18.*pow(
         m33,2)*RD146e*RC146e + 1./54.*m11*pow(RD146e,2) + 1./9.*m11*
         m33*RD146e*RS1463e - 1./9.*m11*m33*RD146e*RC146e + 1./27.*pow(
         m11,2)*RD146e*RS1463e - 1./18.*pow(m11,2)*RD146e*RC146e );

      M13p6x22R +=  + hh1(3,m13,m11,m33,m13,mu2) * ( 1./54.*m33*RD146e*
         RD346e + 1./27.*pow(m33,2)*RD146e*RS3461e + 1./54.*m11*RD146e*
         RD346e + 1./9.*m11*m33*RD146e*RS3461e + 2./27.*pow(m11,2)*
         RD146e*RS3461e );

      M13p6x22R +=  + hh1(3,m13,m11,me2,m13,mu2) * ( 2./27.*m33*me2*
         RD146e*RSe6613 - 1./27.*m33*me2*RD146e*RSe6633 - 1./27.*m33*
         me2*RD146e*RSe6611 - 4./27.*pow(m33,2)*RD146e*RSe6613 + 1./27.
         *pow(m33,2)*RD146e*RSe6633 + 1./9.*pow(m33,2)*RD146e*RSe6611
          + 2./27.*m11*me2*RD146e*RSe6613 - 1./27.*m11*me2*RD146e*
         RSe6633 - 1./27.*m11*me2*RD146e*RSe6611 - 8./27.*m11*m33*
         RD146e*RSe6613 + 1./27.*m11*m33*RD146e*RSe6633 + 7./27.*m11*
         m33*RD146e*RSe6611 - 4./27.*pow(m11,2)*RD146e*RSe6613 + 4./27.
         *pow(m11,2)*RD146e*RSe6611 );

      M13p6x22R +=  + hh1(3,m13,m33,m11,m13,mu2) * ( 1./54.*m33*RD146e*
         RD346e + 2./27.*pow(m33,2)*RD346e*RS1463e + 1./54.*m11*RD146e*
         RD346e + 1./9.*m11*m33*RD346e*RS1463e + 1./27.*pow(m11,2)*
         RD346e*RS1463e );

      M13p6x22R +=  + hh1(3,m13,m33,m33,m13,mu2) * ( 1./54.*m33*pow(
         RD346e,2) + 1./27.*pow(m33,2)*RD346e*RS3461e - 1./18.*pow(
         m33,2)*RD346e*RC346e + 1./54.*m11*pow(RD346e,2) + 1./9.*m11*
         m33*RD346e*RS3461e - 1./9.*m11*m33*RD346e*RC346e + 2./27.*pow(
         m11,2)*RD346e*RS3461e - 1./18.*pow(m11,2)*RD346e*RC346e );

      M13p6x22R +=  + hh1(3,m13,m33,me2,m13,mu2) * ( 2./27.*m33*me2*
         RD346e*RSe6613 - 1./27.*m33*me2*RD346e*RSe6633 - 1./27.*m33*
         me2*RSe6611*RD346e - 4./27.*pow(m33,2)*RD346e*RSe6613 + 4./27.
         *pow(m33,2)*RD346e*RSe6633 + 2./27.*m11*me2*RD346e*RSe6613 - 1.
         /27.*m11*me2*RD346e*RSe6633 - 1./27.*m11*me2*RSe6611*RD346e - 
         8./27.*m11*m33*RD346e*RSe6613 + 7./27.*m11*m33*RD346e*RSe6633
          + 1./27.*m11*m33*RSe6611*RD346e - 4./27.*pow(m11,2)*RD346e*
         RSe6613 + 1./9.*pow(m11,2)*RD346e*RSe6633 + 1./27.*pow(m11,2)*
         RSe6611*RD346e );

      M13p6x22R +=  + hh1(7,m13,m11,m11,m13,mu2) * (  - 1./36.*pow(
         m33,2)*pow(RD146e,2) - 1./18.*m11*m33*pow(RD146e,2) - 1./36.*
         pow(m11,2)*pow(RD146e,2) );

      M13p6x22R +=  + hh1(7,m13,m33,m33,m13,mu2) * (  - 1./36.*pow(
         m33,2)*pow(RD346e,2) - 1./18.*m11*m33*pow(RD346e,2) - 1./36.*
         pow(m11,2)*pow(RD346e,2) );

      M13p6x22R +=  + hh21(1,m11,m11,m13,m13,mu2) * (  - 3./16.*pow(
         m33,2) - 3./8.*m11*m33 - 3./16.*pow(m11,2) );

      M13p6x22R +=  + hh21(1,m11,m14,m34,m13,mu2) * (  - 1./8.*pow(
         m33,2)*RS1463e + 1./8.*pow(m33,2)*RS16e - 1./4.*m11*m33*
         RS1463e + 1./4.*m11*m33*RS16e - 1./8.*pow(m11,2)*RS1463e + 1./
         8.*pow(m11,2)*RS16e );

      M13p6x22R +=  + hh21(1,m11,m16,m36,m13,mu2) * (  - 1./16.*pow(
         m33,2)*RS1463e + 1./16.*pow(m33,2)*RS14e - 1./8.*m11*m33*
         RS1463e + 1./8.*m11*m33*RS14e - 1./16.*pow(m11,2)*RS1463e + 1./
         16.*pow(m11,2)*RS14e );

      M13p6x22R +=  + hh21(1,m13,m11,m11,m13,mu2) * ( 3./32.*pow(m33,2)
          + 1./24.*pow(m33,2)*pow(RS1463e,2) - 1./24.*pow(m33,2)*RC146e
         *RS1463e + 1./96.*pow(m33,2)*pow(RC146e,2) + 3./16.*m11*m33 + 
         1./12.*m11*m33*pow(RS1463e,2) - 1./12.*m11*m33*RC146e*RS1463e
          + 1./48.*m11*m33*pow(RC146e,2) + 3./32.*pow(m11,2) + 1./24.*
         pow(m11,2)*pow(RS1463e,2) - 1./24.*pow(m11,2)*RC146e*RS1463e
          + 1./96.*pow(m11,2)*pow(RC146e,2) );

      M13p6x22R +=  + hh21(1,m13,m11,m33,m13,mu2) * ( 1./12.*pow(m33,2)
         *RS1463e*RS3461e - 1./24.*pow(m33,2)*RC346e*RS1463e - 1./24.*
         pow(m33,2)*RC146e*RS3461e + 1./48.*pow(m33,2)*RC146e*RC346e + 
         1./6.*m11*m33*RS1463e*RS3461e - 1./12.*m11*m33*RC346e*RS1463e
          - 1./12.*m11*m33*RC146e*RS3461e + 1./24.*m11*m33*RC146e*
         RC346e + 1./12.*pow(m11,2)*RS1463e*RS3461e - 1./24.*pow(m11,2)
         *RC346e*RS1463e - 1./24.*pow(m11,2)*RC146e*RS3461e + 1./48.*
         pow(m11,2)*RC146e*RC346e );

      M13p6x22R +=  + hh21(1,m13,m11,me2,m13,mu2) * (  - 1./6.*pow(
         m33,2)*RSe6613*RS1463e + 1./12.*pow(m33,2)*RSe6633*RS1463e + 1.
         /12.*pow(m33,2)*RSe6611*RS1463e + 1./12.*pow(m33,2)*RC146e*
         RSe6613 - 1./24.*pow(m33,2)*RC146e*RSe6633 - 1./24.*pow(m33,2)
         *RC146e*RSe6611 - 1./3.*m11*m33*RSe6613*RS1463e + 1./6.*m11*
         m33*RSe6633*RS1463e + 1./6.*m11*m33*RSe6611*RS1463e + 1./6.*
         m11*m33*RC146e*RSe6613 - 1./12.*m11*m33*RC146e*RSe6633 - 1./12.
         *m11*m33*RC146e*RSe6611 - 1./6.*pow(m11,2)*RSe6613*RS1463e + 1.
         /12.*pow(m11,2)*RSe6633*RS1463e + 1./12.*pow(m11,2)*RSe6611*
         RS1463e + 1./12.*pow(m11,2)*RC146e*RSe6613 - 1./24.*pow(m11,2)
         *RC146e*RSe6633 - 1./24.*pow(m11,2)*RC146e*RSe6611 );

      M13p6x22R +=  + hh21(1,m13,m13,m13,m13,mu2) * ( 3./16.*pow(m33,2)
          + 3./8.*m11*m33 + 3./16.*pow(m11,2) );

      M13p6x22R +=  + hh21(1,m13,m33,m33,m13,mu2) * ( 3./32.*pow(m33,2)
          + 1./24.*pow(m33,2)*pow(RS3461e,2) - 1./24.*pow(m33,2)*RC346e
         *RS3461e + 1./96.*pow(m33,2)*pow(RC346e,2) + 3./16.*m11*m33 + 
         1./12.*m11*m33*pow(RS3461e,2) - 1./12.*m11*m33*RC346e*RS3461e
          + 1./48.*m11*m33*pow(RC346e,2) + 3./32.*pow(m11,2) + 1./24.*
         pow(m11,2)*pow(RS3461e,2) - 1./24.*pow(m11,2)*RC346e*RS3461e
          + 1./96.*pow(m11,2)*pow(RC346e,2) );

      M13p6x22R +=  + hh21(1,m13,m33,me2,m13,mu2) * (  - 1./6.*pow(
         m33,2)*RSe6613*RS3461e + 1./12.*pow(m33,2)*RSe6633*RS3461e + 1.
         /12.*pow(m33,2)*RC346e*RSe6613 - 1./24.*pow(m33,2)*RC346e*
         RSe6633 + 1./12.*pow(m33,2)*RSe6611*RS3461e - 1./24.*pow(
         m33,2)*RSe6611*RC346e - 1./3.*m11*m33*RSe6613*RS3461e + 1./6.*
         m11*m33*RSe6633*RS3461e + 1./6.*m11*m33*RC346e*RSe6613 - 1./12.
         *m11*m33*RC346e*RSe6633 + 1./6.*m11*m33*RSe6611*RS3461e - 1./
         12.*m11*m33*RSe6611*RC346e - 1./6.*pow(m11,2)*RSe6613*RS3461e
          + 1./12.*pow(m11,2)*RSe6633*RS3461e + 1./12.*pow(m11,2)*
         RC346e*RSe6613 - 1./24.*pow(m11,2)*RC346e*RSe6633 + 1./12.*
         pow(m11,2)*RSe6611*RS3461e - 1./24.*pow(m11,2)*RSe6611*RC346e
          );

      M13p6x22R +=  + hh21(1,m13,me2,me2,m13,mu2) * ( 1./6.*pow(m33,2)*
         pow(RSe6613,2) - 1./6.*pow(m33,2)*RSe6633*RSe6613 + 1./24.*
         pow(m33,2)*pow(RSe6633,2) - 1./6.*pow(m33,2)*RSe6611*RSe6613
          + 1./12.*pow(m33,2)*RSe6611*RSe6633 + 1./24.*pow(m33,2)*pow(
         RSe6611,2) + 1./3.*m11*m33*pow(RSe6613,2) - 1./3.*m11*m33*
         RSe6633*RSe6613 + 1./12.*m11*m33*pow(RSe6633,2) - 1./3.*m11*
         m33*RSe6611*RSe6613 + 1./6.*m11*m33*RSe6611*RSe6633 + 1./12.*
         m11*m33*pow(RSe6611,2) + 1./6.*pow(m11,2)*pow(RSe6613,2) - 1./
         6.*pow(m11,2)*RSe6633*RSe6613 + 1./24.*pow(m11,2)*pow(
         RSe6633,2) - 1./6.*pow(m11,2)*RSe6611*RSe6613 + 1./12.*pow(
         m11,2)*RSe6611*RSe6633 + 1./24.*pow(m11,2)*pow(RSe6611,2) );

      M13p6x22R +=  + hh21(1,m14,m11,m34,m13,mu2) * ( 1./8.*pow(m33,2)*
         RS1463e + 1./8.*pow(m33,2)*RS16e - 1./8.*pow(m33,2)*RC146e + 1.
         /4.*m11*m33*RS1463e + 1./4.*m11*m33*RS16e - 1./4.*m11*m33*
         RC146e + 1./8.*pow(m11,2)*RS1463e + 1./8.*pow(m11,2)*RS16e - 1.
         /8.*pow(m11,2)*RC146e );

      M13p6x22R +=  + hh21(1,m14,m33,m34,m13,mu2) * ( 1./8.*pow(m33,2)*
         RS3461e - 1./8.*pow(m33,2)*RS36e + 1./4.*m11*m33*RS3461e - 1./
         4.*m11*m33*RS36e + 1./8.*pow(m11,2)*RS3461e - 1./8.*pow(m11,2)
         *RS36e );

      M13p6x22R +=  + hh21(1,m14,m34,me2,m13,mu2) * (  - 1./4.*pow(
         m33,2)*RSe6613 - 1./8.*pow(m33,2)*RSe63 + 1./8.*pow(m33,2)*
         RSe61 + 1./4.*pow(m33,2)*RSe6611 - 1./2.*m11*m33*RSe6613 - 1./
         4.*m11*m33*RSe63 + 1./4.*m11*m33*RSe61 + 1./2.*m11*m33*RSe6611
          - 1./4.*pow(m11,2)*RSe6613 - 1./8.*pow(m11,2)*RSe63 + 1./8.*
         pow(m11,2)*RSe61 + 1./4.*pow(m11,2)*RSe6611 );

      M13p6x22R +=  + hh21(1,m16,m11,m36,m13,mu2) * ( 1./16.*pow(m33,2)
         *RS1463e + 1./16.*pow(m33,2)*RS14e - 1./16.*pow(m33,2)*RC146e
          + 1./8.*m11*m33*RS1463e + 1./8.*m11*m33*RS14e - 1./8.*m11*m33
         *RC146e + 1./16.*pow(m11,2)*RS1463e + 1./16.*pow(m11,2)*RS14e
          - 1./16.*pow(m11,2)*RC146e );

      M13p6x22R +=  + hh21(1,m16,m33,m36,m13,mu2) * ( 1./16.*pow(m33,2)
         *RS3461e - 1./16.*pow(m33,2)*RS34e + 1./8.*m11*m33*RS3461e - 1.
         /8.*m11*m33*RS34e + 1./16.*pow(m11,2)*RS3461e - 1./16.*pow(
         m11,2)*RS34e );

      M13p6x22R +=  + hh21(1,m16,m36,me2,m13,mu2) * (  - 1./8.*pow(
         m33,2)*RSe6613 + 1./8.*pow(m33,2)*RSe63 - 1./8.*pow(m33,2)*
         RSe61 + 1./8.*pow(m33,2)*RSe6611 - 1./4.*m11*m33*RSe6613 + 1./
         4.*m11*m33*RSe63 - 1./4.*m11*m33*RSe61 + 1./4.*m11*m33*RSe6611
          - 1./8.*pow(m11,2)*RSe6613 + 1./8.*pow(m11,2)*RSe63 - 1./8.*
         pow(m11,2)*RSe61 + 1./8.*pow(m11,2)*RSe6611 );

      M13p6x22R +=  + hh21(1,m33,m13,m33,m13,mu2) * (  - 3./16.*pow(
         m33,2) - 3./8.*m11*m33 - 3./16.*pow(m11,2) );

      M13p6x22R +=  + hh21(1,m33,m14,m34,m13,mu2) * (  - 1./8.*pow(
         m33,2)*RS3461e + 1./8.*pow(m33,2)*RS36e - 1./4.*m11*m33*
         RS3461e + 1./4.*m11*m33*RS36e - 1./8.*pow(m11,2)*RS3461e + 1./
         8.*pow(m11,2)*RS36e );

      M13p6x22R +=  + hh21(1,m33,m16,m36,m13,mu2) * (  - 1./16.*pow(
         m33,2)*RS3461e + 1./16.*pow(m33,2)*RS34e - 1./8.*m11*m33*
         RS3461e + 1./8.*m11*m33*RS34e - 1./16.*pow(m11,2)*RS3461e + 1./
         16.*pow(m11,2)*RS34e );

      M13p6x22R +=  + hh21(1,m34,m11,m14,m13,mu2) * ( 1./8.*pow(m33,2)*
         RS1463e - 1./8.*pow(m33,2)*RS16e + 1./4.*m11*m33*RS1463e - 1./
         4.*m11*m33*RS16e + 1./8.*pow(m11,2)*RS1463e - 1./8.*pow(m11,2)
         *RS16e );

      M13p6x22R +=  + hh21(1,m34,m14,m33,m13,mu2) * ( 1./8.*pow(m33,2)*
         RS3461e + 1./8.*pow(m33,2)*RS36e - 1./8.*pow(m33,2)*RC346e + 1.
         /4.*m11*m33*RS3461e + 1./4.*m11*m33*RS36e - 1./4.*m11*m33*
         RC346e + 1./8.*pow(m11,2)*RS3461e + 1./8.*pow(m11,2)*RS36e - 1.
         /8.*pow(m11,2)*RC346e );

      M13p6x22R +=  + hh21(1,m34,m14,me2,m13,mu2) * (  - 1./4.*pow(
         m33,2)*RSe6613 + 1./8.*pow(m33,2)*RSe63 + 1./4.*pow(m33,2)*
         RSe6633 - 1./8.*pow(m33,2)*RSe61 - 1./2.*m11*m33*RSe6613 + 1./
         4.*m11*m33*RSe63 + 1./2.*m11*m33*RSe6633 - 1./4.*m11*m33*RSe61
          - 1./4.*pow(m11,2)*RSe6613 + 1./8.*pow(m11,2)*RSe63 + 1./4.*
         pow(m11,2)*RSe6633 - 1./8.*pow(m11,2)*RSe61 );

      M13p6x22R +=  + hh21(1,m36,m11,m16,m13,mu2) * ( 1./16.*pow(m33,2)
         *RS1463e - 1./16.*pow(m33,2)*RS14e + 1./8.*m11*m33*RS1463e - 1.
         /8.*m11*m33*RS14e + 1./16.*pow(m11,2)*RS1463e - 1./16.*pow(
         m11,2)*RS14e );

      M13p6x22R +=  + hh21(1,m36,m16,m33,m13,mu2) * ( 1./16.*pow(m33,2)
         *RS3461e + 1./16.*pow(m33,2)*RS34e - 1./16.*pow(m33,2)*RC346e
          + 1./8.*m11*m33*RS3461e + 1./8.*m11*m33*RS34e - 1./8.*m11*m33
         *RC346e + 1./16.*pow(m11,2)*RS3461e + 1./16.*pow(m11,2)*RS34e
          - 1./16.*pow(m11,2)*RC346e );

      M13p6x22R +=  + hh21(1,m36,m16,me2,m13,mu2) * (  - 1./8.*pow(
         m33,2)*RSe6613 - 1./8.*pow(m33,2)*RSe63 + 1./8.*pow(m33,2)*
         RSe6633 + 1./8.*pow(m33,2)*RSe61 - 1./4.*m11*m33*RSe6613 - 1./
         4.*m11*m33*RSe63 + 1./4.*m11*m33*RSe6633 + 1./4.*m11*m33*RSe61
          - 1./8.*pow(m11,2)*RSe6613 - 1./8.*pow(m11,2)*RSe63 + 1./8.*
         pow(m11,2)*RSe6633 + 1./8.*pow(m11,2)*RSe61 );

      M13p6x22R +=  + hh21(1,m44,m14,m34,m13,mu2) * ( 9./16.*pow(m33,2)
          + 9./8.*m11*m33 + 9./16.*pow(m11,2) );

      M13p6x22R +=  + hh21(1,m46,m14,m36,m13,mu2) * ( 3./8.*pow(m33,2)
          + 3./4.*m11*m33 + 3./8.*pow(m11,2) );

      M13p6x22R +=  + hh21(1,m46,m16,m34,m13,mu2) * ( 3./8.*pow(m33,2)
          + 3./4.*m11*m33 + 3./8.*pow(m11,2) );

      M13p6x22R +=  + hh21(1,me2,m14,m34,m13,mu2) * ( 1./16.*pow(m33,2)
          + 1./4.*pow(m33,2)*RSe6613 + 1./8.*pow(m33,2)*RSe63 + 1./8.*
         pow(m33,2)*RSe61 + 1./8.*m11*m33 + 1./2.*m11*m33*RSe6613 + 1./
         4.*m11*m33*RSe63 + 1./4.*m11*m33*RSe61 + 1./16.*pow(m11,2) + 1.
         /4.*pow(m11,2)*RSe6613 + 1./8.*pow(m11,2)*RSe63 + 1./8.*pow(
         m11,2)*RSe61 );

      M13p6x22R +=  + hh21(1,me2,m16,m36,m13,mu2) * ( 1./8.*pow(m33,2)
          + 1./8.*pow(m33,2)*RSe6613 - 1./8.*pow(m33,2)*RSe63 - 1./8.*
         pow(m33,2)*RSe61 + 1./4.*m11*m33 + 1./4.*m11*m33*RSe6613 - 1./
         4.*m11*m33*RSe63 - 1./4.*m11*m33*RSe61 + 1./8.*pow(m11,2) + 1./
         8.*pow(m11,2)*RSe6613 - 1./8.*pow(m11,2)*RSe63 - 1./8.*pow(
         m11,2)*RSe61 );

      M13p6x22R +=  + hh21(3,m13,m11,m11,m13,mu2) * (  - 1./24.*pow(
         m33,2)*RD146e*RS1463e + 1./48.*pow(m33,2)*RD146e*RC146e - 1./
         12.*m11*m33*RD146e*RS1463e + 1./24.*m11*m33*RD146e*RC146e - 1./
         24.*pow(m11,2)*RD146e*RS1463e + 1./48.*pow(m11,2)*RD146e*
         RC146e );

      M13p6x22R +=  + hh21(3,m13,m11,m33,m13,mu2) * (  - 1./24.*pow(
         m33,2)*RD146e*RS3461e + 1./48.*pow(m33,2)*RD146e*RC346e - 1./
         12.*m11*m33*RD146e*RS3461e + 1./24.*m11*m33*RD146e*RC346e - 1./
         24.*pow(m11,2)*RD146e*RS3461e + 1./48.*pow(m11,2)*RD146e*
         RC346e );

      M13p6x22R +=  + hh21(3,m13,m11,me2,m13,mu2) * ( 1./12.*pow(m33,2)
         *RD146e*RSe6613 - 1./24.*pow(m33,2)*RD146e*RSe6633 - 1./24.*
         pow(m33,2)*RD146e*RSe6611 + 1./6.*m11*m33*RD146e*RSe6613 - 1./
         12.*m11*m33*RD146e*RSe6633 - 1./12.*m11*m33*RD146e*RSe6611 + 1.
         /12.*pow(m11,2)*RD146e*RSe6613 - 1./24.*pow(m11,2)*RD146e*
         RSe6633 - 1./24.*pow(m11,2)*RD146e*RSe6611 );

      M13p6x22R +=  + hh21(3,m13,m33,m11,m13,mu2) * (  - 1./24.*pow(
         m33,2)*RD346e*RS1463e + 1./48.*pow(m33,2)*RC146e*RD346e - 1./
         12.*m11*m33*RD346e*RS1463e + 1./24.*m11*m33*RC146e*RD346e - 1./
         24.*pow(m11,2)*RD346e*RS1463e + 1./48.*pow(m11,2)*RC146e*
         RD346e );

      M13p6x22R +=  + hh21(3,m13,m33,m33,m13,mu2) * (  - 1./24.*pow(
         m33,2)*RD346e*RS3461e + 1./48.*pow(m33,2)*RD346e*RC346e - 1./
         12.*m11*m33*RD346e*RS3461e + 1./24.*m11*m33*RD346e*RC346e - 1./
         24.*pow(m11,2)*RD346e*RS3461e + 1./48.*pow(m11,2)*RD346e*
         RC346e );

      M13p6x22R +=  + hh21(3,m13,m33,me2,m13,mu2) * ( 1./12.*pow(m33,2)
         *RD346e*RSe6613 - 1./24.*pow(m33,2)*RD346e*RSe6633 - 1./24.*
         pow(m33,2)*RSe6611*RD346e + 1./6.*m11*m33*RD346e*RSe6613 - 1./
         12.*m11*m33*RD346e*RSe6633 - 1./12.*m11*m33*RSe6611*RD346e + 1.
         /12.*pow(m11,2)*RD346e*RSe6613 - 1./24.*pow(m11,2)*RD346e*
         RSe6633 - 1./24.*pow(m11,2)*RSe6611*RD346e );

      M13p6x22R +=  + hh21(3,m14,m11,m34,m13,mu2) * (  - 1./8.*pow(
         m33,2)*RD146e - 1./4.*m11*m33*RD146e - 1./8.*pow(m11,2)*RD146e
          );

      M13p6x22R +=  + hh21(3,m16,m11,m36,m13,mu2) * (  - 1./16.*pow(
         m33,2)*RD146e - 1./8.*m11*m33*RD146e - 1./16.*pow(m11,2)*
         RD146e );

      M13p6x22R +=  + hh21(3,m34,m33,m14,m13,mu2) * (  - 1./8.*pow(
         m33,2)*RD346e - 1./4.*m11*m33*RD346e - 1./8.*pow(m11,2)*RD346e
          );

      M13p6x22R +=  + hh21(3,m36,m33,m16,m13,mu2) * (  - 1./16.*pow(
         m33,2)*RD346e - 1./8.*m11*m33*RD346e - 1./16.*pow(m11,2)*
         RD346e );

      M13p6x22R +=  + hh21(7,m13,m11,m11,m13,mu2) * ( 1./96.*pow(m33,2)
         *pow(RD146e,2) + 1./48.*m11*m33*pow(RD146e,2) + 1./96.*pow(
         m11,2)*pow(RD146e,2) );

      M13p6x22R +=  + hh21(7,m13,m11,m33,m13,mu2) * ( 1./48.*pow(m33,2)
         *RD146e*RD346e + 1./24.*m11*m33*RD146e*RD346e + 1./48.*pow(
         m11,2)*RD146e*RD346e );

      M13p6x22R +=  + hh21(7,m13,m33,m33,m13,mu2) * ( 1./96.*pow(m33,2)
         *pow(RD346e,2) + 1./48.*m11*m33*pow(RD346e,2) + 1./96.*pow(
         m11,2)*pow(RD346e,2) );

  return M13p6x22R/pow(f0,4);
}

// 1 valence mass 3 sea mass *****************************************
double mv1s3nf3p4(const quarkmassnf mass, const Linf Liin){
  return mv1s3nf3p4L(mass,Liin)+mv1s3nf3p4R(mass);
}

double mv1s3nf3p4L(const quarkmassnf mass, const Linf Liin){
  double L0r,L1r,L2r,L3r,L4r,L5r,L6r,L7r,L8r,L9r,L10r,L11r,H1r,H2r,mu;
  int nfLi;
  Liin.out(L0r,L1r,L2r,L3r,L4r,L5r,L6r,L7r,L8r,L9r,L10r,L11r,H1r,H2r,mu,nfLi);
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if (mu0 != mu){
    cout << "ERROR: differing mu's in mv1s3nf3p4L\n";}
  if ( (nq != 4) || (B0mq.size() != 4)){
    cout << "ERROR: mv1s3nf3p4L probably called with wrong number"
	 << "of quark masses\n";}
  // four masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m44 = 2.*B0mq[1];
  double m55 = 2.*B0mq[2];
  double m66 = 2.*B0mq[3];
  double M12p4x13L =
       + 16*m11*m66*L6r - 8*m11*m66*L4r + 16*m11*m55*L6r - 8*m11*m55*
         L4r + 16*m11*m44*L6r - 8*m11*m44*L4r + 16*pow(m11,2)*L8r - 8*
         pow(m11,2)*L5r;

  return M12p4x13L/pow(f0,2);
}


double mv1s3nf3p4R(const quarkmassnf mass){
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if ( (nq != 4) || (B0mq.size() != 4)){
    cout << "ERROR: mv2s3nf3p4R probably called with wrong number"
	 << "of quark masses\n";}
  // four masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m44 = 2.*B0mq[1];
  double m55 = 2.*B0mq[2];
  double m66 = 2.*B0mq[3];
  double xb1 = (m44+m55+m66)/3.;
  double msqr = sqrt(xb1*xb1-(m44*m55+m44*m66+m55*m66)/3.);
  double mp2 = xb1-msqr; 
  double me2 = xb1+msqr;

  double mu2 = mu0*mu0;
  double RD1456ep = RD(m11,m44,m55,m66,me2,mp2);
  double RC1456pe = RC(m11,m44,m55,m66,mp2,me2);  
  double RSe45611p = RS(me2,m44,m55,m66,m11,m11,mp2);
  double RSp45611e = RS(mp2,m44,m55,m66,m11,m11,me2);

  double M12p4x13R =
       + Ab(m11,mu2) * (  - 1./3.*m11*RC1456pe );

      M12p4x13R +=  + Ab(me2,mu2) * (  - 1./3.*m11*RSe45611p );

      M12p4x13R +=  + Ab(mp2,mu2) * (  - 1./3.*m11*RSp45611e );

      M12p4x13R +=  + Bb(m11,mu2) * (  - 1./3.*m11*RD1456ep );

  return M12p4x13R/pow(f0,2);
}

double mv1s3nf3p6(const quarkmassnf mass, const Linf Liin, const Ki Kiin){
  return mv1s3nf3p6L(mass,Liin)+mv1s3nf3p6K(mass,Kiin)+mv1s3nf3p6R(mass);
}

double mv1s3nf3p6L(const quarkmassnf mass, const Linf Liin){
  double L0r,L1r,L2r,L3r,L4r,L5r,L6r,L7r,L8r,L9r,L10r,L11r,H1r,H2r,mu;
  int nfLi;
  Liin.out(L0r,L1r,L2r,L3r,L4r,L5r,L6r,L7r,L8r,L9r,L10r,L11r,H1r,H2r,mu,nfLi);
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if (mu0 != mu){
    cout << "ERROR: differing mu's in mv1s3nf3p6L\n";}
  if ( (nq != 4) || (B0mq.size() != 4)){
    cout << "ERROR: mv1s3nf3p6L probably called with wrong number"
	 << "of quark masses\n";}
  // four masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m44 = 2.*B0mq[1];
  double m55 = 2.*B0mq[2];
  double m66 = 2.*B0mq[3];
  double m14 = (m11+m44)/2;
  double m15 = (m11+m55)/2;
  double m16 = (m11+m66)/2;
  double m45 = (m44+m55)/2;
  double m46 = (m44+m66)/2;
  double m56 = (m55+m66)/2;

  double xb1 = (m44+m55+m66)/3.;
  double msqr = sqrt(xb1*xb1-(m44*m55+m44*m66+m55*m66)/3.);
  double mp2 = xb1-msqr; 
  double me2 = xb1+msqr;
  double mu2 = mu0*mu0;

  double RN1p = RN(m11,mp2);
  double RN1e = RN(m11,me2);
  double RNep = RN(me2,mp2);

  double RSp456e = RS(mp2,m44,m55,m66,me2);
  double RSp465e = RS(mp2,m44,m66,m55,me2);
  double RSp564e = RS(mp2,m55,m66,m44,me2);
  double RSe456p = RS(me2,m44,m55,m66,mp2);
  double RSe465p = RS(me2,m44,m66,m55,mp2);
  double RSe564p = RS(me2,m55,m66,m44,mp2);
  double RSp451e = RS(mp2,m44,m55,m11,me2);
  double RSp461e = RS(mp2,m44,m66,m11,me2);
  double RSp561e = RS(mp2,m55,m66,m11,me2);
  double RSe451p = RS(me2,m44,m55,m11,mp2);
  double RSe461p = RS(me2,m44,m66,m11,mp2);
  double RSe561p = RS(me2,m55,m66,m11,mp2);
  double RS456pe = RS(m44,m55,m66,mp2,me2);
  double RS546pe = RS(m55,m44,m66,mp2,me2);
  double RS645pe = RS(m66,m44,m55,mp2,me2);
  double RS156pe = RS(m11,m55,m66,mp2,me2);
  double RS146pe = RS(m11,m44,m66,mp2,me2);
  double RS145pe = RS(m11,m44,m55,mp2,me2);

  double RD1456ep = RD(m11,m44,m55,m66,me2,mp2);
  double RC1456pe = RC(m11,m44,m55,m66,mp2,me2);  
  double RSe45611p = RS(me2,m44,m55,m66,m11,m11,mp2);
  double RSp45611e = RS(mp2,m44,m55,m66,m11,m11,me2);

  double M12p6x13L =
       + pi16 * (  - 4./3.*m11*pow(mp2,2)*L3r*RSp45611e - 2./3.*m11*
         pow(mp2,2)*L2r*RSp564e - 2./3.*m11*pow(mp2,2)*L2r*RSp465e - 2./
         3.*m11*pow(mp2,2)*L2r*RSp456e - 4./3.*m11*pow(mp2,2)*L0r*
         RSp45611e - 4./3.*m11*pow(me2,2)*L3r*RSe45611p - 2./3.*m11*
         pow(me2,2)*L2r*RSe564p - 2./3.*m11*pow(me2,2)*L2r*RSe465p - 2./
         3.*m11*pow(me2,2)*L2r*RSe456p - 4./3.*m11*pow(me2,2)*L0r*
         RSe45611p + 1./2.*m11*pow(m66,2)*L3r + 4*m11*pow(m66,2)*L2r - 
         2./3.*m11*pow(m66,2)*L2r*RS645pe + m11*pow(m66,2)*L0r + 2*m11*
         m55*m66*L2r + 1./2.*m11*pow(m55,2)*L3r + 4*m11*pow(m55,2)*L2r
          - 2./3.*m11*pow(m55,2)*L2r*RS546pe + m11*pow(m55,2)*L0r + 2*
         m11*m44*m66*L2r + 2*m11*m44*m55*L2r + 1./2.*m11*pow(m44,2)*L3r
          + 4*m11*pow(m44,2)*L2r - 2./3.*m11*pow(m44,2)*L2r*RS456pe + 
         m11*pow(m44,2)*L0r - 8./3.*pow(m11,2)*L3r*RD1456ep - 8./3.*
         pow(m11,2)*L0r*RD1456ep + pow(m11,2)*m66*L3r + 2*pow(m11,2)*
         m66*L0r + pow(m11,2)*m55*L3r + 2*pow(m11,2)*m55*L0r + pow(
         m11,2)*m44*L3r );

      M12p6x13L +=  + pi16 * ( 2*pow(m11,2)*m44*L0r + 3./2.*pow(m11,3)*
         L3r - 4./3.*pow(m11,3)*L3r*RC1456pe + 2*pow(m11,3)*L2r + 4*
         pow(m11,3)*L1r + 3*pow(m11,3)*L0r - 4./3.*pow(m11,3)*L0r*
         RC1456pe );

      M12p6x13L +=  + Ab(m11,mu2) * ( 8./3.*m11*L5r*RD1456ep - 8*m11*
         L3r*RD1456ep - 8*m11*L0r*RD1456ep - 32./3.*m11*m66*L7r*RS145pe
          - 16./3.*m11*m66*L6r*RC1456pe - 8./9.*m11*m66*L5r*pow(
         RS145pe,2) - 16./9.*m11*m66*L5r*RSe451p*RS145pe - 16./9.*m11*
         m66*L5r*RSp451e*RS145pe + 8*m11*m66*L4r*RC1456pe + 16./3.*m11*
         m66*L4r*RSe45611p + 16./3.*m11*m66*L4r*RSp45611e - 8./9.*m11*
         m66*L4r*pow(RS145pe,2) - 8./9.*m11*m66*L4r*pow(RS146pe,2) - 8./
         9.*m11*m66*L4r*pow(RS156pe,2) - 16./9.*m11*m66*L4r*RSe561p*
         RS156pe - 16./9.*m11*m66*L4r*RSe461p*RS146pe - 16./9.*m11*m66*
         L4r*RSe451p*RS145pe - 16./9.*m11*m66*L4r*RSp561e*RS156pe - 16./
         9.*m11*m66*L4r*RSp461e*RS146pe - 16./9.*m11*m66*L4r*RSp451e*
         RS145pe - 16./9.*m11*m66*mp2*L5r*RN1p*RSp451e*RS145pe + 16./3.
         *m11*m66*mp2*L4r*RN1p*RSp45611e - 16./9.*m11*m66*mp2*L4r*RN1p*
         RSp561e*RS156pe - 16./9.*m11*m66*mp2*L4r*RN1p*RSp461e*RS146pe
          - 16./9.*m11*m66*mp2*L4r*RN1p*RSp451e*RS145pe - 16./9.*m11*
         m66*me2*L5r*RN1e*RSe451p*RS145pe );

      M12p6x13L +=  + Ab(m11,mu2) * ( 16./3.*m11*m66*me2*L4r*RN1e*
         RSe45611p - 16./9.*m11*m66*me2*L4r*RN1e*RSe561p*RS156pe - 16./
         9.*m11*m66*me2*L4r*RN1e*RSe461p*RS146pe - 16./9.*m11*m66*me2*
         L4r*RN1e*RSe451p*RS145pe + 32./9.*m11*pow(m66,2)*L8r*RN1p*
         RSp451e*RS145pe + 32./9.*m11*pow(m66,2)*L8r*RN1e*RSe451p*
         RS145pe + 32./9.*m11*pow(m66,2)*L7r*RN1p*RSp451e*RS145pe + 32./
         9.*m11*pow(m66,2)*L7r*RN1e*RSe451p*RS145pe + 32./9.*m11*pow(
         m66,2)*L6r*RN1p*RSp451e*RS145pe + 32./9.*m11*pow(m66,2)*L6r*
         RN1e*RSe451p*RS145pe - 32./3.*m11*m55*L7r*RS146pe - 16./3.*m11
         *m55*L6r*RC1456pe - 8./9.*m11*m55*L5r*pow(RS146pe,2) - 16./9.*
         m11*m55*L5r*RSe461p*RS146pe - 16./9.*m11*m55*L5r*RSp461e*
         RS146pe + 8*m11*m55*L4r*RC1456pe + 16./3.*m11*m55*L4r*
         RSe45611p + 16./3.*m11*m55*L4r*RSp45611e - 8./9.*m11*m55*L4r*
         pow(RS145pe,2) - 8./9.*m11*m55*L4r*pow(RS146pe,2) - 8./9.*m11*
         m55*L4r*pow(RS156pe,2) - 16./9.*m11*m55*L4r*RSe561p*RS156pe - 
         16./9.*m11*m55*L4r*RSe461p*RS146pe );

      M12p6x13L +=  + Ab(m11,mu2) * (  - 16./9.*m11*m55*L4r*RSe451p*
         RS145pe - 16./9.*m11*m55*L4r*RSp561e*RS156pe - 16./9.*m11*m55*
         L4r*RSp461e*RS146pe - 16./9.*m11*m55*L4r*RSp451e*RS145pe - 16./
         9.*m11*m55*mp2*L5r*RN1p*RSp461e*RS146pe + 16./3.*m11*m55*mp2*
         L4r*RN1p*RSp45611e - 16./9.*m11*m55*mp2*L4r*RN1p*RSp561e*
         RS156pe - 16./9.*m11*m55*mp2*L4r*RN1p*RSp461e*RS146pe - 16./9.
         *m11*m55*mp2*L4r*RN1p*RSp451e*RS145pe - 16./9.*m11*m55*me2*L5r
         *RN1e*RSe461p*RS146pe + 16./3.*m11*m55*me2*L4r*RN1e*RSe45611p
          - 16./9.*m11*m55*me2*L4r*RN1e*RSe561p*RS156pe - 16./9.*m11*
         m55*me2*L4r*RN1e*RSe461p*RS146pe - 16./9.*m11*m55*me2*L4r*RN1e
         *RSe451p*RS145pe + 32./9.*m11*m55*m66*L7r*RN1p*RSp461e*RS145pe
          + 32./9.*m11*m55*m66*L7r*RN1p*RSp451e*RS146pe + 32./9.*m11*
         m55*m66*L7r*RN1e*RSe461p*RS145pe + 32./9.*m11*m55*m66*L7r*RN1e
         *RSe451p*RS146pe + 32./9.*m11*m55*m66*L6r*RN1p*RSp461e*RS146pe
          + 32./9.*m11*m55*m66*L6r*RN1p*RSp451e*RS145pe + 32./9.*m11*
         m55*m66*L6r*RN1e*RSe461p*RS146pe );

      M12p6x13L +=  + Ab(m11,mu2) * ( 32./9.*m11*m55*m66*L6r*RN1e*
         RSe451p*RS145pe + 32./9.*m11*pow(m55,2)*L8r*RN1p*RSp461e*
         RS146pe + 32./9.*m11*pow(m55,2)*L8r*RN1e*RSe461p*RS146pe + 32./
         9.*m11*pow(m55,2)*L7r*RN1p*RSp461e*RS146pe + 32./9.*m11*pow(
         m55,2)*L7r*RN1e*RSe461p*RS146pe + 32./9.*m11*pow(m55,2)*L6r*
         RN1p*RSp461e*RS146pe + 32./9.*m11*pow(m55,2)*L6r*RN1e*RSe461p*
         RS146pe - 32./3.*m11*m44*L7r*RS156pe - 16./3.*m11*m44*L6r*
         RC1456pe - 8./9.*m11*m44*L5r*pow(RS156pe,2) - 16./9.*m11*m44*
         L5r*RSe561p*RS156pe - 16./9.*m11*m44*L5r*RSp561e*RS156pe + 8*
         m11*m44*L4r*RC1456pe + 16./3.*m11*m44*L4r*RSe45611p + 16./3.*
         m11*m44*L4r*RSp45611e - 8./9.*m11*m44*L4r*pow(RS145pe,2) - 8./
         9.*m11*m44*L4r*pow(RS146pe,2) - 8./9.*m11*m44*L4r*pow(
         RS156pe,2) - 16./9.*m11*m44*L4r*RSe561p*RS156pe - 16./9.*m11*
         m44*L4r*RSe461p*RS146pe - 16./9.*m11*m44*L4r*RSe451p*RS145pe
          - 16./9.*m11*m44*L4r*RSp561e*RS156pe - 16./9.*m11*m44*L4r*
         RSp461e*RS146pe );

      M12p6x13L +=  + Ab(m11,mu2) * (  - 16./9.*m11*m44*L4r*RSp451e*
         RS145pe - 16./9.*m11*m44*mp2*L5r*RN1p*RSp561e*RS156pe + 16./3.
         *m11*m44*mp2*L4r*RN1p*RSp45611e - 16./9.*m11*m44*mp2*L4r*RN1p*
         RSp561e*RS156pe - 16./9.*m11*m44*mp2*L4r*RN1p*RSp461e*RS146pe
          - 16./9.*m11*m44*mp2*L4r*RN1p*RSp451e*RS145pe - 16./9.*m11*
         m44*me2*L5r*RN1e*RSe561p*RS156pe + 16./3.*m11*m44*me2*L4r*RN1e
         *RSe45611p - 16./9.*m11*m44*me2*L4r*RN1e*RSe561p*RS156pe - 16./
         9.*m11*m44*me2*L4r*RN1e*RSe461p*RS146pe - 16./9.*m11*m44*me2*
         L4r*RN1e*RSe451p*RS145pe + 32./9.*m11*m44*m66*L7r*RN1p*RSp561e
         *RS145pe + 32./9.*m11*m44*m66*L7r*RN1p*RSp451e*RS156pe + 32./9.
         *m11*m44*m66*L7r*RN1e*RSe561p*RS145pe + 32./9.*m11*m44*m66*L7r
         *RN1e*RSe451p*RS156pe + 32./9.*m11*m44*m66*L6r*RN1p*RSp561e*
         RS156pe + 32./9.*m11*m44*m66*L6r*RN1p*RSp451e*RS145pe + 32./9.
         *m11*m44*m66*L6r*RN1e*RSe561p*RS156pe + 32./9.*m11*m44*m66*L6r
         *RN1e*RSe451p*RS145pe + 32./9.*m11*m44*m55*L7r*RN1p*RSp561e*
         RS146pe );

      M12p6x13L +=  + Ab(m11,mu2) * ( 32./9.*m11*m44*m55*L7r*RN1p*
         RSp461e*RS156pe + 32./9.*m11*m44*m55*L7r*RN1e*RSe561p*RS146pe
          + 32./9.*m11*m44*m55*L7r*RN1e*RSe461p*RS156pe + 32./9.*m11*
         m44*m55*L6r*RN1p*RSp561e*RS156pe + 32./9.*m11*m44*m55*L6r*RN1p
         *RSp461e*RS146pe + 32./9.*m11*m44*m55*L6r*RN1e*RSe561p*RS156pe
          + 32./9.*m11*m44*m55*L6r*RN1e*RSe461p*RS146pe + 32./9.*m11*
         pow(m44,2)*L8r*RN1p*RSp561e*RS156pe + 32./9.*m11*pow(m44,2)*
         L8r*RN1e*RSe561p*RS156pe + 32./9.*m11*pow(m44,2)*L7r*RN1p*
         RSp561e*RS156pe + 32./9.*m11*pow(m44,2)*L7r*RN1e*RSe561p*
         RS156pe + 32./9.*m11*pow(m44,2)*L6r*RN1p*RSp561e*RS156pe + 32./
         9.*m11*pow(m44,2)*L6r*RN1e*RSe561p*RS156pe - 64./3.*pow(m11,2)
         *L8r*RC1456pe + 32*pow(m11,2)*L7r + 32*pow(m11,2)*L6r + 40./3.
         *pow(m11,2)*L5r*RC1456pe + 16./3.*pow(m11,2)*L5r*RSe45611p + 
         16./3.*pow(m11,2)*L5r*RSp45611e - 8*pow(m11,2)*L3r*RC1456pe + 
         20*pow(m11,2)*L2r + 8*pow(m11,2)*L1r - 8*pow(m11,2)*L0r*
         RC1456pe );

      M12p6x13L +=  + Ab(m11,mu2) * ( 16./3.*pow(m11,2)*mp2*L5r*RN1p*
         RSp45611e + 16./3.*pow(m11,2)*me2*L5r*RN1e*RSe45611p - 32./3.*
         pow(m11,2)*m66*L7r*RN1p*RSp451e - 32./3.*pow(m11,2)*m66*L7r*
         RN1e*RSe451p - 32./3.*pow(m11,2)*m66*L6r*RN1p*RSp45611e - 32./
         3.*pow(m11,2)*m66*L6r*RN1e*RSe45611p - 32./3.*pow(m11,2)*m55*
         L7r*RN1p*RSp461e - 32./3.*pow(m11,2)*m55*L7r*RN1e*RSe461p - 32.
         /3.*pow(m11,2)*m55*L6r*RN1p*RSp45611e - 32./3.*pow(m11,2)*m55*
         L6r*RN1e*RSe45611p - 32./3.*pow(m11,2)*m44*L7r*RN1p*RSp561e - 
         32./3.*pow(m11,2)*m44*L7r*RN1e*RSe561p - 32./3.*pow(m11,2)*m44
         *L6r*RN1p*RSp45611e - 32./3.*pow(m11,2)*m44*L6r*RN1e*RSe45611p
          - 32./3.*pow(m11,3)*L8r*RN1p*RSp45611e - 32./3.*pow(m11,3)*
         L8r*RN1e*RSe45611p );

      M12p6x13L +=  + Ab(m14,mu2) * ( 16*m11*m44*L8r - 8*m11*m44*L5r + 
         10*m11*m44*L3r + 4*m11*m44*L0r + 16*pow(m11,2)*L8r - 8*pow(
         m11,2)*L5r + 10*pow(m11,2)*L3r + 4*pow(m11,2)*L0r );

      M12p6x13L +=  + Ab(m15,mu2) * ( 16*m11*m55*L8r - 8*m11*m55*L5r + 
         10*m11*m55*L3r + 4*m11*m55*L0r + 16*pow(m11,2)*L8r - 8*pow(
         m11,2)*L5r + 10*pow(m11,2)*L3r + 4*pow(m11,2)*L0r );

      M12p6x13L +=  + Ab(m16,mu2) * ( 16*m11*m66*L8r - 8*m11*m66*L5r + 
         10*m11*m66*L3r + 4*m11*m66*L0r + 16*pow(m11,2)*L8r - 8*pow(
         m11,2)*L5r + 10*pow(m11,2)*L3r + 4*pow(m11,2)*L0r );

      M12p6x13L +=  + Ab(m44,mu2) * ( 16*m11*m44*L6r - 16./3.*m11*m44*
         L6r*RS456pe - 16*m11*m44*L4r + 16./3.*m11*m44*L4r*RS456pe + 4*
         m11*m44*L2r - 4./3.*m11*m44*L2r*RS456pe + 16*m11*m44*L1r - 16./
         3.*m11*m44*L1r*RS456pe );

      M12p6x13L +=  + Ab(m45,mu2) * ( 16*m11*m55*L6r - 16*m11*m55*L4r
          + 4*m11*m55*L2r + 16*m11*m55*L1r + 16*m11*m44*L6r - 16*m11*
         m44*L4r + 4*m11*m44*L2r + 16*m11*m44*L1r );

      M12p6x13L +=  + Ab(m46,mu2) * ( 16*m11*m66*L6r - 16*m11*m66*L4r
          + 4*m11*m66*L2r + 16*m11*m66*L1r + 16*m11*m44*L6r - 16*m11*
         m44*L4r + 4*m11*m44*L2r + 16*m11*m44*L1r );

      M12p6x13L +=  + Ab(m55,mu2) * ( 16*m11*m55*L6r - 16./3.*m11*m55*
         L6r*RS546pe - 16*m11*m55*L4r + 16./3.*m11*m55*L4r*RS546pe + 4*
         m11*m55*L2r - 4./3.*m11*m55*L2r*RS546pe + 16*m11*m55*L1r - 16./
         3.*m11*m55*L1r*RS546pe );

      M12p6x13L +=  + Ab(m56,mu2) * ( 16*m11*m66*L6r - 16*m11*m66*L4r
          + 4*m11*m66*L2r + 16*m11*m66*L1r + 16*m11*m55*L6r - 16*m11*
         m55*L4r + 4*m11*m55*L2r + 16*m11*m55*L1r );

      M12p6x13L +=  + Ab(m66,mu2) * ( 16*m11*m66*L6r - 16./3.*m11*m66*
         L6r*RS645pe - 16*m11*m66*L4r + 16./3.*m11*m66*L4r*RS645pe + 4*
         m11*m66*L2r - 4./3.*m11*m66*L2r*RS645pe + 16*m11*m66*L1r - 16./
         3.*m11*m66*L1r*RS645pe );

      M12p6x13L +=  + Ab(me2,mu2) * ( 8./3.*m11*me2*L5r*RSe45611p + 8./
         3.*m11*me2*L4r*RSe564p + 8./3.*m11*me2*L4r*RSe465p + 8./3.*m11
         *me2*L4r*RSe456p - 8*m11*me2*L3r*RSe45611p - 4./3.*m11*me2*L2r
         *RSe564p - 4./3.*m11*me2*L2r*RSe465p - 4./3.*m11*me2*L2r*
         RSe456p - 16./3.*m11*me2*L1r*RSe564p - 16./3.*m11*me2*L1r*
         RSe465p - 16./3.*m11*me2*L1r*RSe456p - 8*m11*me2*L0r*RSe45611p
          - 32./3.*m11*m66*L7r*RSe451p - 16./3.*m11*m66*L6r*RSe45611p
          - 16./3.*m11*m66*L6r*RSe456p - 8./9.*m11*m66*L5r*pow(
         RSe451p,2) - 16./9.*m11*m66*L5r*RSp451e*RSe451p + 8./3.*m11*
         m66*L4r*RSe45611p - 8./9.*m11*m66*L4r*pow(RSe561p,2) - 8./9.*
         m11*m66*L4r*pow(RSe461p,2) - 8./9.*m11*m66*L4r*pow(RSe451p,2)
          - 16./9.*m11*m66*L4r*RSp561e*RSe561p - 16./9.*m11*m66*L4r*
         RSp461e*RSe461p - 16./9.*m11*m66*L4r*RSp451e*RSe451p + 8./3.*
         m11*m66*L4r*RSe456p - 16./9.*m11*m66*mp2*L5r*RNep*RSp451e*
         RSe451p - 16./9.*m11*m66*mp2*L4r*RNep*RSp561e*RSe561p - 16./9.
         *m11*m66*mp2*L4r*RNep*RSp461e*RSe461p );

      M12p6x13L +=  + Ab(me2,mu2) * (  - 16./9.*m11*m66*mp2*L4r*RNep*
         RSp451e*RSe451p + 16./9.*m11*m66*me2*L5r*RN1e*RSe451p*RS145pe
          - 16./3.*m11*m66*me2*L4r*RN1e*RSe45611p + 16./9.*m11*m66*me2*
         L4r*RN1e*RSe561p*RS156pe + 16./9.*m11*m66*me2*L4r*RN1e*RSe461p
         *RS146pe + 16./9.*m11*m66*me2*L4r*RN1e*RSe451p*RS145pe + 32./9.
         *m11*pow(m66,2)*L8r*RNep*RSp451e*RSe451p - 32./9.*m11*pow(
         m66,2)*L8r*RN1e*RSe451p*RS145pe + 32./9.*m11*pow(m66,2)*L7r*
         RNep*RSp451e*RSe451p - 32./9.*m11*pow(m66,2)*L7r*RN1e*RSe451p*
         RS145pe + 32./9.*m11*pow(m66,2)*L6r*RNep*RSp451e*RSe451p - 32./
         9.*m11*pow(m66,2)*L6r*RN1e*RSe451p*RS145pe - 32./3.*m11*m55*
         L7r*RSe461p - 16./3.*m11*m55*L6r*RSe45611p - 16./3.*m11*m55*
         L6r*RSe465p - 8./9.*m11*m55*L5r*pow(RSe461p,2) - 16./9.*m11*
         m55*L5r*RSp461e*RSe461p + 8./3.*m11*m55*L4r*RSe45611p - 8./9.*
         m11*m55*L4r*pow(RSe561p,2) - 8./9.*m11*m55*L4r*pow(RSe461p,2)
          - 8./9.*m11*m55*L4r*pow(RSe451p,2) - 16./9.*m11*m55*L4r*
         RSp561e*RSe561p );

      M12p6x13L +=  + Ab(me2,mu2) * (  - 16./9.*m11*m55*L4r*RSp461e*
         RSe461p - 16./9.*m11*m55*L4r*RSp451e*RSe451p + 8./3.*m11*m55*
         L4r*RSe465p - 16./9.*m11*m55*mp2*L5r*RNep*RSp461e*RSe461p - 16.
         /9.*m11*m55*mp2*L4r*RNep*RSp561e*RSe561p - 16./9.*m11*m55*mp2*
         L4r*RNep*RSp461e*RSe461p - 16./9.*m11*m55*mp2*L4r*RNep*RSp451e
         *RSe451p + 16./9.*m11*m55*me2*L5r*RN1e*RSe461p*RS146pe - 16./3.
         *m11*m55*me2*L4r*RN1e*RSe45611p + 16./9.*m11*m55*me2*L4r*RN1e*
         RSe561p*RS156pe + 16./9.*m11*m55*me2*L4r*RN1e*RSe461p*RS146pe
          + 16./9.*m11*m55*me2*L4r*RN1e*RSe451p*RS145pe + 32./9.*m11*
         m55*m66*L7r*RNep*RSp461e*RSe451p + 32./9.*m11*m55*m66*L7r*RNep
         *RSp451e*RSe461p - 32./9.*m11*m55*m66*L7r*RN1e*RSe461p*RS145pe
          - 32./9.*m11*m55*m66*L7r*RN1e*RSe451p*RS146pe + 32./9.*m11*
         m55*m66*L6r*RNep*RSp461e*RSe461p + 32./9.*m11*m55*m66*L6r*RNep
         *RSp451e*RSe451p - 32./9.*m11*m55*m66*L6r*RN1e*RSe461p*RS146pe
          - 32./9.*m11*m55*m66*L6r*RN1e*RSe451p*RS145pe + 32./9.*m11*
         pow(m55,2)*L8r*RNep*RSp461e*RSe461p );

      M12p6x13L +=  + Ab(me2,mu2) * (  - 32./9.*m11*pow(m55,2)*L8r*RN1e
         *RSe461p*RS146pe + 32./9.*m11*pow(m55,2)*L7r*RNep*RSp461e*
         RSe461p - 32./9.*m11*pow(m55,2)*L7r*RN1e*RSe461p*RS146pe + 32./
         9.*m11*pow(m55,2)*L6r*RNep*RSp461e*RSe461p - 32./9.*m11*pow(
         m55,2)*L6r*RN1e*RSe461p*RS146pe - 32./3.*m11*m44*L7r*RSe561p
          - 16./3.*m11*m44*L6r*RSe45611p - 16./3.*m11*m44*L6r*RSe564p
          - 8./9.*m11*m44*L5r*pow(RSe561p,2) - 16./9.*m11*m44*L5r*
         RSp561e*RSe561p + 8./3.*m11*m44*L4r*RSe45611p - 8./9.*m11*m44*
         L4r*pow(RSe561p,2) - 8./9.*m11*m44*L4r*pow(RSe461p,2) - 8./9.*
         m11*m44*L4r*pow(RSe451p,2) - 16./9.*m11*m44*L4r*RSp561e*
         RSe561p - 16./9.*m11*m44*L4r*RSp461e*RSe461p - 16./9.*m11*m44*
         L4r*RSp451e*RSe451p + 8./3.*m11*m44*L4r*RSe564p - 16./9.*m11*
         m44*mp2*L5r*RNep*RSp561e*RSe561p - 16./9.*m11*m44*mp2*L4r*RNep
         *RSp561e*RSe561p - 16./9.*m11*m44*mp2*L4r*RNep*RSp461e*RSe461p
          - 16./9.*m11*m44*mp2*L4r*RNep*RSp451e*RSe451p + 16./9.*m11*
         m44*me2*L5r*RN1e*RSe561p*RS156pe );

      M12p6x13L +=  + Ab(me2,mu2) * (  - 16./3.*m11*m44*me2*L4r*RN1e*
         RSe45611p + 16./9.*m11*m44*me2*L4r*RN1e*RSe561p*RS156pe + 16./
         9.*m11*m44*me2*L4r*RN1e*RSe461p*RS146pe + 16./9.*m11*m44*me2*
         L4r*RN1e*RSe451p*RS145pe + 32./9.*m11*m44*m66*L7r*RNep*RSp561e
         *RSe451p + 32./9.*m11*m44*m66*L7r*RNep*RSp451e*RSe561p - 32./9.
         *m11*m44*m66*L7r*RN1e*RSe561p*RS145pe - 32./9.*m11*m44*m66*L7r
         *RN1e*RSe451p*RS156pe + 32./9.*m11*m44*m66*L6r*RNep*RSp561e*
         RSe561p + 32./9.*m11*m44*m66*L6r*RNep*RSp451e*RSe451p - 32./9.
         *m11*m44*m66*L6r*RN1e*RSe561p*RS156pe - 32./9.*m11*m44*m66*L6r
         *RN1e*RSe451p*RS145pe + 32./9.*m11*m44*m55*L7r*RNep*RSp561e*
         RSe461p + 32./9.*m11*m44*m55*L7r*RNep*RSp461e*RSe561p - 32./9.
         *m11*m44*m55*L7r*RN1e*RSe561p*RS146pe - 32./9.*m11*m44*m55*L7r
         *RN1e*RSe461p*RS156pe + 32./9.*m11*m44*m55*L6r*RNep*RSp561e*
         RSe561p + 32./9.*m11*m44*m55*L6r*RNep*RSp461e*RSe461p - 32./9.
         *m11*m44*m55*L6r*RN1e*RSe561p*RS156pe - 32./9.*m11*m44*m55*L6r
         *RN1e*RSe461p*RS146pe );

      M12p6x13L +=  + Ab(me2,mu2) * ( 32./9.*m11*pow(m44,2)*L8r*RNep*
         RSp561e*RSe561p - 32./9.*m11*pow(m44,2)*L8r*RN1e*RSe561p*
         RS156pe + 32./9.*m11*pow(m44,2)*L7r*RNep*RSp561e*RSe561p - 32./
         9.*m11*pow(m44,2)*L7r*RN1e*RSe561p*RS156pe + 32./9.*m11*pow(
         m44,2)*L6r*RNep*RSp561e*RSe561p - 32./9.*m11*pow(m44,2)*L6r*
         RN1e*RSe561p*RS156pe - 64./3.*pow(m11,2)*L8r*RSe45611p + 16./3.
         *pow(m11,2)*L5r*RSe45611p - 16./3.*pow(m11,2)*me2*L5r*RN1e*
         RSe45611p + 32./3.*pow(m11,2)*m66*L7r*RN1e*RSe451p + 32./3.*
         pow(m11,2)*m66*L6r*RN1e*RSe45611p + 32./3.*pow(m11,2)*m55*L7r*
         RN1e*RSe461p + 32./3.*pow(m11,2)*m55*L6r*RN1e*RSe45611p + 32./
         3.*pow(m11,2)*m44*L7r*RN1e*RSe561p + 32./3.*pow(m11,2)*m44*L6r
         *RN1e*RSe45611p + 32./3.*pow(m11,3)*L8r*RN1e*RSe45611p );

      M12p6x13L +=  + Ab(mp2,mu2) * ( 8./3.*m11*mp2*L5r*RSp45611e + 8./
         3.*m11*mp2*L4r*RSp564e + 8./3.*m11*mp2*L4r*RSp465e + 8./3.*m11
         *mp2*L4r*RSp456e - 8*m11*mp2*L3r*RSp45611e - 4./3.*m11*mp2*L2r
         *RSp564e - 4./3.*m11*mp2*L2r*RSp465e - 4./3.*m11*mp2*L2r*
         RSp456e - 16./3.*m11*mp2*L1r*RSp564e - 16./3.*m11*mp2*L1r*
         RSp465e - 16./3.*m11*mp2*L1r*RSp456e - 8*m11*mp2*L0r*RSp45611e
          - 32./3.*m11*m66*L7r*RSp451e - 16./3.*m11*m66*L6r*RSp45611e
          - 16./3.*m11*m66*L6r*RSp456e - 8./9.*m11*m66*L5r*pow(
         RSp451e,2) + 8./3.*m11*m66*L4r*RSp45611e - 8./9.*m11*m66*L4r*
         pow(RSp561e,2) - 8./9.*m11*m66*L4r*pow(RSp461e,2) - 8./9.*m11*
         m66*L4r*pow(RSp451e,2) + 8./3.*m11*m66*L4r*RSp456e + 16./9.*
         m11*m66*mp2*L5r*RNep*RSp451e*RSe451p + 16./9.*m11*m66*mp2*L5r*
         RN1p*RSp451e*RS145pe + 16./9.*m11*m66*mp2*L4r*RNep*RSp561e*
         RSe561p + 16./9.*m11*m66*mp2*L4r*RNep*RSp461e*RSe461p + 16./9.
         *m11*m66*mp2*L4r*RNep*RSp451e*RSe451p - 16./3.*m11*m66*mp2*L4r
         *RN1p*RSp45611e );

      M12p6x13L +=  + Ab(mp2,mu2) * ( 16./9.*m11*m66*mp2*L4r*RN1p*
         RSp561e*RS156pe + 16./9.*m11*m66*mp2*L4r*RN1p*RSp461e*RS146pe
          + 16./9.*m11*m66*mp2*L4r*RN1p*RSp451e*RS145pe - 32./9.*m11*
         pow(m66,2)*L8r*RNep*RSp451e*RSe451p - 32./9.*m11*pow(m66,2)*
         L8r*RN1p*RSp451e*RS145pe - 32./9.*m11*pow(m66,2)*L7r*RNep*
         RSp451e*RSe451p - 32./9.*m11*pow(m66,2)*L7r*RN1p*RSp451e*
         RS145pe - 32./9.*m11*pow(m66,2)*L6r*RNep*RSp451e*RSe451p - 32./
         9.*m11*pow(m66,2)*L6r*RN1p*RSp451e*RS145pe - 32./3.*m11*m55*
         L7r*RSp461e - 16./3.*m11*m55*L6r*RSp45611e - 16./3.*m11*m55*
         L6r*RSp465e - 8./9.*m11*m55*L5r*pow(RSp461e,2) + 8./3.*m11*m55
         *L4r*RSp45611e - 8./9.*m11*m55*L4r*pow(RSp561e,2) - 8./9.*m11*
         m55*L4r*pow(RSp461e,2) - 8./9.*m11*m55*L4r*pow(RSp451e,2) + 8./
         3.*m11*m55*L4r*RSp465e + 16./9.*m11*m55*mp2*L5r*RNep*RSp461e*
         RSe461p + 16./9.*m11*m55*mp2*L5r*RN1p*RSp461e*RS146pe + 16./9.
         *m11*m55*mp2*L4r*RNep*RSp561e*RSe561p + 16./9.*m11*m55*mp2*L4r
         *RNep*RSp461e*RSe461p );

      M12p6x13L +=  + Ab(mp2,mu2) * ( 16./9.*m11*m55*mp2*L4r*RNep*
         RSp451e*RSe451p - 16./3.*m11*m55*mp2*L4r*RN1p*RSp45611e + 16./
         9.*m11*m55*mp2*L4r*RN1p*RSp561e*RS156pe + 16./9.*m11*m55*mp2*
         L4r*RN1p*RSp461e*RS146pe + 16./9.*m11*m55*mp2*L4r*RN1p*RSp451e
         *RS145pe - 32./9.*m11*m55*m66*L7r*RNep*RSp461e*RSe451p - 32./9.
         *m11*m55*m66*L7r*RNep*RSp451e*RSe461p - 32./9.*m11*m55*m66*L7r
         *RN1p*RSp461e*RS145pe - 32./9.*m11*m55*m66*L7r*RN1p*RSp451e*
         RS146pe - 32./9.*m11*m55*m66*L6r*RNep*RSp461e*RSe461p - 32./9.
         *m11*m55*m66*L6r*RNep*RSp451e*RSe451p - 32./9.*m11*m55*m66*L6r
         *RN1p*RSp461e*RS146pe - 32./9.*m11*m55*m66*L6r*RN1p*RSp451e*
         RS145pe - 32./9.*m11*pow(m55,2)*L8r*RNep*RSp461e*RSe461p - 32./
         9.*m11*pow(m55,2)*L8r*RN1p*RSp461e*RS146pe - 32./9.*m11*pow(
         m55,2)*L7r*RNep*RSp461e*RSe461p - 32./9.*m11*pow(m55,2)*L7r*
         RN1p*RSp461e*RS146pe - 32./9.*m11*pow(m55,2)*L6r*RNep*RSp461e*
         RSe461p - 32./9.*m11*pow(m55,2)*L6r*RN1p*RSp461e*RS146pe - 32./
         3.*m11*m44*L7r*RSp561e );

      M12p6x13L +=  + Ab(mp2,mu2) * (  - 16./3.*m11*m44*L6r*RSp45611e
          - 16./3.*m11*m44*L6r*RSp564e - 8./9.*m11*m44*L5r*pow(
         RSp561e,2) + 8./3.*m11*m44*L4r*RSp45611e - 8./9.*m11*m44*L4r*
         pow(RSp561e,2) - 8./9.*m11*m44*L4r*pow(RSp461e,2) - 8./9.*m11*
         m44*L4r*pow(RSp451e,2) + 8./3.*m11*m44*L4r*RSp564e + 16./9.*
         m11*m44*mp2*L5r*RNep*RSp561e*RSe561p + 16./9.*m11*m44*mp2*L5r*
         RN1p*RSp561e*RS156pe + 16./9.*m11*m44*mp2*L4r*RNep*RSp561e*
         RSe561p + 16./9.*m11*m44*mp2*L4r*RNep*RSp461e*RSe461p + 16./9.
         *m11*m44*mp2*L4r*RNep*RSp451e*RSe451p - 16./3.*m11*m44*mp2*L4r
         *RN1p*RSp45611e + 16./9.*m11*m44*mp2*L4r*RN1p*RSp561e*RS156pe
          + 16./9.*m11*m44*mp2*L4r*RN1p*RSp461e*RS146pe + 16./9.*m11*
         m44*mp2*L4r*RN1p*RSp451e*RS145pe - 32./9.*m11*m44*m66*L7r*RNep
         *RSp561e*RSe451p - 32./9.*m11*m44*m66*L7r*RNep*RSp451e*RSe561p
          - 32./9.*m11*m44*m66*L7r*RN1p*RSp561e*RS145pe - 32./9.*m11*
         m44*m66*L7r*RN1p*RSp451e*RS156pe - 32./9.*m11*m44*m66*L6r*RNep
         *RSp561e*RSe561p );

      M12p6x13L +=  + Ab(mp2,mu2) * (  - 32./9.*m11*m44*m66*L6r*RNep*
         RSp451e*RSe451p - 32./9.*m11*m44*m66*L6r*RN1p*RSp561e*RS156pe
          - 32./9.*m11*m44*m66*L6r*RN1p*RSp451e*RS145pe - 32./9.*m11*
         m44*m55*L7r*RNep*RSp561e*RSe461p - 32./9.*m11*m44*m55*L7r*RNep
         *RSp461e*RSe561p - 32./9.*m11*m44*m55*L7r*RN1p*RSp561e*RS146pe
          - 32./9.*m11*m44*m55*L7r*RN1p*RSp461e*RS156pe - 32./9.*m11*
         m44*m55*L6r*RNep*RSp561e*RSe561p - 32./9.*m11*m44*m55*L6r*RNep
         *RSp461e*RSe461p - 32./9.*m11*m44*m55*L6r*RN1p*RSp561e*RS156pe
          - 32./9.*m11*m44*m55*L6r*RN1p*RSp461e*RS146pe - 32./9.*m11*
         pow(m44,2)*L8r*RNep*RSp561e*RSe561p - 32./9.*m11*pow(m44,2)*
         L8r*RN1p*RSp561e*RS156pe - 32./9.*m11*pow(m44,2)*L7r*RNep*
         RSp561e*RSe561p - 32./9.*m11*pow(m44,2)*L7r*RN1p*RSp561e*
         RS156pe - 32./9.*m11*pow(m44,2)*L6r*RNep*RSp561e*RSe561p - 32./
         9.*m11*pow(m44,2)*L6r*RN1p*RSp561e*RS156pe - 64./3.*pow(m11,2)
         *L8r*RSp45611e + 16./3.*pow(m11,2)*L5r*RSp45611e - 16./3.*pow(
         m11,2)*mp2*L5r*RN1p*RSp45611e );

      M12p6x13L +=  + Ab(mp2,mu2) * ( 32./3.*pow(m11,2)*m66*L7r*RN1p*
         RSp451e + 32./3.*pow(m11,2)*m66*L6r*RN1p*RSp45611e + 32./3.*
         pow(m11,2)*m55*L7r*RN1p*RSp461e + 32./3.*pow(m11,2)*m55*L6r*
         RN1p*RSp45611e + 32./3.*pow(m11,2)*m44*L7r*RN1p*RSp561e + 32./
         3.*pow(m11,2)*m44*L6r*RN1p*RSp45611e + 32./3.*pow(m11,3)*L8r*
         RN1p*RSp45611e );

      M12p6x13L +=  + Bb(m11,mu2) * (  - 16./3.*m11*m66*L6r*RD1456ep + 
         8*m11*m66*L4r*RD1456ep + 16./9.*m11*pow(m66,2)*L8r*pow(
         RS145pe,2) + 16./9.*m11*pow(m66,2)*L7r*pow(RS145pe,2) + 16./9.
         *m11*pow(m66,2)*L6r*pow(RS145pe,2) - 16./3.*m11*m55*L6r*
         RD1456ep + 8*m11*m55*L4r*RD1456ep + 32./9.*m11*m55*m66*L7r*
         RS146pe*RS145pe + 16./9.*m11*m55*m66*L6r*pow(RS145pe,2) + 16./
         9.*m11*m55*m66*L6r*pow(RS146pe,2) + 16./9.*m11*pow(m55,2)*L8r*
         pow(RS146pe,2) + 16./9.*m11*pow(m55,2)*L7r*pow(RS146pe,2) + 16.
         /9.*m11*pow(m55,2)*L6r*pow(RS146pe,2) - 16./3.*m11*m44*L6r*
         RD1456ep + 8*m11*m44*L4r*RD1456ep + 32./9.*m11*m44*m66*L7r*
         RS156pe*RS145pe + 16./9.*m11*m44*m66*L6r*pow(RS145pe,2) + 16./
         9.*m11*m44*m66*L6r*pow(RS156pe,2) + 32./9.*m11*m44*m55*L7r*
         RS156pe*RS146pe + 16./9.*m11*m44*m55*L6r*pow(RS146pe,2) + 16./
         9.*m11*m44*m55*L6r*pow(RS156pe,2) + 16./9.*m11*pow(m44,2)*L8r*
         pow(RS156pe,2) + 16./9.*m11*pow(m44,2)*L7r*pow(RS156pe,2) + 16.
         /9.*m11*pow(m44,2)*L6r*pow(RS156pe,2) );

      M12p6x13L +=  + Bb(m11,mu2) * (  - 64./3.*pow(m11,2)*L8r*RD1456ep
          + 40./3.*pow(m11,2)*L5r*RD1456ep - 8*pow(m11,2)*L3r*RD1456ep
          - 8*pow(m11,2)*L0r*RD1456ep - 32./3.*pow(m11,2)*m66*L7r*
         RS145pe - 32./3.*pow(m11,2)*m66*L6r*RC1456pe - 8./9.*pow(
         m11,2)*m66*L5r*pow(RS145pe,2) + 16./3.*pow(m11,2)*m66*L4r*
         RC1456pe - 8./9.*pow(m11,2)*m66*L4r*pow(RS145pe,2) - 8./9.*
         pow(m11,2)*m66*L4r*pow(RS146pe,2) - 8./9.*pow(m11,2)*m66*L4r*
         pow(RS156pe,2) - 32./3.*pow(m11,2)*m55*L7r*RS146pe - 32./3.*
         pow(m11,2)*m55*L6r*RC1456pe - 8./9.*pow(m11,2)*m55*L5r*pow(
         RS146pe,2) + 16./3.*pow(m11,2)*m55*L4r*RC1456pe - 8./9.*pow(
         m11,2)*m55*L4r*pow(RS145pe,2) - 8./9.*pow(m11,2)*m55*L4r*pow(
         RS146pe,2) - 8./9.*pow(m11,2)*m55*L4r*pow(RS156pe,2) - 32./3.*
         pow(m11,2)*m44*L7r*RS156pe - 32./3.*pow(m11,2)*m44*L6r*
         RC1456pe - 8./9.*pow(m11,2)*m44*L5r*pow(RS156pe,2) + 16./3.*
         pow(m11,2)*m44*L4r*RC1456pe - 8./9.*pow(m11,2)*m44*L4r*pow(
         RS145pe,2) );

      M12p6x13L +=  + Bb(m11,mu2) * (  - 8./9.*pow(m11,2)*m44*L4r*pow(
         RS146pe,2) - 8./9.*pow(m11,2)*m44*L4r*pow(RS156pe,2) - 32./3.*
         pow(m11,3)*L8r*RC1456pe + 16*pow(m11,3)*L7r + 16./3.*pow(
         m11,3)*L5r*RC1456pe );

      M12p6x13L +=  + Bb(me2,mu2) * (  - 8./9.*m11*m66*me2*L5r*pow(
         RSe451p,2) - 8./9.*m11*m66*me2*L4r*pow(RSe561p,2) - 8./9.*m11*
         m66*me2*L4r*pow(RSe461p,2) - 8./9.*m11*m66*me2*L4r*pow(
         RSe451p,2) + 16./9.*m11*pow(m66,2)*L8r*pow(RSe451p,2) + 16./9.
         *m11*pow(m66,2)*L7r*pow(RSe451p,2) + 16./9.*m11*pow(m66,2)*L6r
         *pow(RSe451p,2) - 8./9.*m11*m55*me2*L5r*pow(RSe461p,2) - 8./9.
         *m11*m55*me2*L4r*pow(RSe561p,2) - 8./9.*m11*m55*me2*L4r*pow(
         RSe461p,2) - 8./9.*m11*m55*me2*L4r*pow(RSe451p,2) + 32./9.*m11
         *m55*m66*L7r*RSe451p*RSe461p + 16./9.*m11*m55*m66*L6r*pow(
         RSe461p,2) + 16./9.*m11*m55*m66*L6r*pow(RSe451p,2) + 16./9.*
         m11*pow(m55,2)*L8r*pow(RSe461p,2) + 16./9.*m11*pow(m55,2)*L7r*
         pow(RSe461p,2) + 16./9.*m11*pow(m55,2)*L6r*pow(RSe461p,2) - 8./
         9.*m11*m44*me2*L5r*pow(RSe561p,2) - 8./9.*m11*m44*me2*L4r*pow(
         RSe561p,2) - 8./9.*m11*m44*me2*L4r*pow(RSe461p,2) - 8./9.*m11*
         m44*me2*L4r*pow(RSe451p,2) + 32./9.*m11*m44*m66*L7r*RSe451p*
         RSe561p );

      M12p6x13L +=  + Bb(me2,mu2) * ( 16./9.*m11*m44*m66*L6r*pow(
         RSe561p,2) + 16./9.*m11*m44*m66*L6r*pow(RSe451p,2) + 32./9.*
         m11*m44*m55*L7r*RSe461p*RSe561p + 16./9.*m11*m44*m55*L6r*pow(
         RSe561p,2) + 16./9.*m11*m44*m55*L6r*pow(RSe461p,2) + 16./9.*
         m11*pow(m44,2)*L8r*pow(RSe561p,2) + 16./9.*m11*pow(m44,2)*L7r*
         pow(RSe561p,2) + 16./9.*m11*pow(m44,2)*L6r*pow(RSe561p,2) );

      M12p6x13L +=  + Bb(mp2,mu2) * (  - 8./9.*m11*m66*mp2*L5r*pow(
         RSp451e,2) - 8./9.*m11*m66*mp2*L4r*pow(RSp561e,2) - 8./9.*m11*
         m66*mp2*L4r*pow(RSp461e,2) - 8./9.*m11*m66*mp2*L4r*pow(
         RSp451e,2) + 16./9.*m11*pow(m66,2)*L8r*pow(RSp451e,2) + 16./9.
         *m11*pow(m66,2)*L7r*pow(RSp451e,2) + 16./9.*m11*pow(m66,2)*L6r
         *pow(RSp451e,2) - 8./9.*m11*m55*mp2*L5r*pow(RSp461e,2) - 8./9.
         *m11*m55*mp2*L4r*pow(RSp561e,2) - 8./9.*m11*m55*mp2*L4r*pow(
         RSp461e,2) - 8./9.*m11*m55*mp2*L4r*pow(RSp451e,2) + 32./9.*m11
         *m55*m66*L7r*RSp451e*RSp461e + 16./9.*m11*m55*m66*L6r*pow(
         RSp461e,2) + 16./9.*m11*m55*m66*L6r*pow(RSp451e,2) + 16./9.*
         m11*pow(m55,2)*L8r*pow(RSp461e,2) + 16./9.*m11*pow(m55,2)*L7r*
         pow(RSp461e,2) + 16./9.*m11*pow(m55,2)*L6r*pow(RSp461e,2) - 8./
         9.*m11*m44*mp2*L5r*pow(RSp561e,2) - 8./9.*m11*m44*mp2*L4r*pow(
         RSp561e,2) - 8./9.*m11*m44*mp2*L4r*pow(RSp461e,2) - 8./9.*m11*
         m44*mp2*L4r*pow(RSp451e,2) + 32./9.*m11*m44*m66*L7r*RSp451e*
         RSp561e );

      M12p6x13L +=  + Bb(mp2,mu2) * ( 16./9.*m11*m44*m66*L6r*pow(
         RSp561e,2) + 16./9.*m11*m44*m66*L6r*pow(RSp451e,2) + 32./9.*
         m11*m44*m55*L7r*RSp461e*RSp561e + 16./9.*m11*m44*m55*L6r*pow(
         RSp561e,2) + 16./9.*m11*m44*m55*L6r*pow(RSp461e,2) + 16./9.*
         m11*pow(m44,2)*L8r*pow(RSp561e,2) + 16./9.*m11*pow(m44,2)*L7r*
         pow(RSp561e,2) + 16./9.*m11*pow(m44,2)*L6r*pow(RSp561e,2) );

      M12p6x13L +=  + Cb(m11,mu2) * (  - 32./3.*pow(m11,2)*m66*L6r*
         RD1456ep + 16./3.*pow(m11,2)*m66*L4r*RD1456ep - 32./3.*pow(
         m11,2)*m55*L6r*RD1456ep + 16./3.*pow(m11,2)*m55*L4r*RD1456ep
          - 32./3.*pow(m11,2)*m44*L6r*RD1456ep + 16./3.*pow(m11,2)*m44*
         L4r*RD1456ep - 32./3.*pow(m11,3)*L8r*RD1456ep + 16./3.*pow(
         m11,3)*L5r*RD1456ep );

      M12p6x13L +=  - 128*m11*pow(m66,2)*L4r*L6r + 64*m11*pow(m66,2)*
         pow(L4r,2) - 256*m11*m55*m66*L4r*L6r + 128*m11*m55*m66*pow(
         L4r,2) - 128*m11*pow(m55,2)*L4r*L6r + 64*m11*pow(m55,2)*pow(
         L4r,2) - 256*m11*m44*m66*L4r*L6r + 128*m11*m44*m66*pow(L4r,2)
          - 256*m11*m44*m55*L4r*L6r + 128*m11*m44*m55*pow(L4r,2) - 128*
         m11*pow(m44,2)*L4r*L6r + 64*m11*pow(m44,2)*pow(L4r,2) - 128*
         pow(m11,2)*m66*L5r*L6r - 128*pow(m11,2)*m66*L4r*L8r + 128*pow(
         m11,2)*m66*L4r*L5r - 128*pow(m11,2)*m55*L5r*L6r - 128*pow(
         m11,2)*m55*L4r*L8r + 128*pow(m11,2)*m55*L4r*L5r - 128*pow(
         m11,2)*m44*L5r*L6r - 128*pow(m11,2)*m44*L4r*L8r + 128*pow(
         m11,2)*m44*L4r*L5r - 128*pow(m11,3)*L5r*L8r + 64*pow(m11,3)*
         pow(L5r,2);

  return M12p6x13L/pow(f0,4);
}

double mv1s3nf3p6K(const quarkmassnf mass, const Ki Kiin){
  double KK[116],mu;
  Kiin.out(KK,mu);
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if (mu0 != mu){
    cout << "ERROR: differing mu's in mv1s3nf3p6L\n";}
  if ( (nq != 4) || (B0mq.size() != 4)){
    cout << "ERROR: mv1s3nf3p6K probably called with wrong number"
	 << "of quark masses\n";}
  // four masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m44 = 2.*B0mq[1];
  double m55 = 2.*B0mq[2];
  double m66 = 2.*B0mq[3];

  double M12p6x13K =
       - 32*KK[17]*pow(m11,3) - 32*KK[18]*pow(m11,2)*m66 - 32*KK[18]*
         pow(m11,2)*m55 - 32*KK[18]*pow(m11,2)*m44 - 16*KK[19]*pow(
         m11,3) - 16*KK[20]*pow(m11,2)*m66 - 16*KK[20]*pow(m11,2)*m55
          - 16*KK[20]*pow(m11,2)*m44 - 16*KK[21]*m11*pow(m66,2) - 16*
         KK[21]*m11*pow(m55,2) - 16*KK[21]*m11*pow(m44,2) - 16*KK[22]*
         m11*pow(m66,2) - 32*KK[22]*m11*m55*m66 - 16*KK[22]*m11*pow(
         m55,2) - 32*KK[22]*m11*m44*m66 - 32*KK[22]*m11*m44*m55 - 16*
         KK[22]*m11*pow(m44,2) - 16*KK[23]*pow(m11,3) + 48*KK[25]*pow(
         m11,3) + 16*KK[26]*m11*pow(m66,2) + 16*KK[26]*m11*pow(m55,2)
          + 16*KK[26]*m11*pow(m44,2) + 32*KK[26]*pow(m11,2)*m66 + 32*
         KK[26]*pow(m11,2)*m55 + 32*KK[26]*pow(m11,2)*m44 + 48*KK[27]*
         m11*pow(m66,2) + 96*KK[27]*m11*m55*m66 + 48*KK[27]*m11*pow(
         m55,2) + 96*KK[27]*m11*m44*m66 + 96*KK[27]*m11*m44*m55 + 48*
         KK[27]*m11*pow(m44,2) + 32*KK[39]*pow(m11,3) + 32*KK[40]*pow(
         m11,2)*m66;
      M12p6x13K +=  + 32*KK[40]*pow(m11,2)*m55 + 32*KK[40]*pow(m11,2)*
         m44;


  return M12p6x13K/pow(f0,4);
}

double mv1s3nf3p6R(const quarkmassnf mass){
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if ( (nq != 4) || (B0mq.size() != 4)){
    cout << "ERROR: mv1s3nf3p6R probably called with wrong number"
	 << "of quark masses\n";}
  // four masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m44 = 2.*B0mq[1];
  double m55 = 2.*B0mq[2];
  double m66 = 2.*B0mq[3];
  double m14 = (m11+m44)/2;
  double m15 = (m11+m55)/2;
  double m16 = (m11+m66)/2;
  double m45 = (m44+m55)/2;
  double m46 = (m44+m66)/2;
  double m56 = (m55+m66)/2;

  double xb1 = (m44+m55+m66)/3.;
  double msqr = sqrt(xb1*xb1-(m44*m55+m44*m66+m55*m66)/3.);
  double mp2 = xb1-msqr; 
  double me2 = xb1+msqr;
  double mu2 = mu0*mu0;

  double RN1p = RN(m11,mp2);
  double RN1e = RN(m11,me2);
  double RNep = RN(me2,mp2);

  double RSp456e = RS(mp2,m44,m55,m66,me2);
  double RSp465e = RS(mp2,m44,m66,m55,me2);
  double RSp564e = RS(mp2,m55,m66,m44,me2);
  double RSe456p = RS(me2,m44,m55,m66,mp2);
  double RSe465p = RS(me2,m44,m66,m55,mp2);
  double RSe564p = RS(me2,m55,m66,m44,mp2);
  double RSp451e = RS(mp2,m44,m55,m11,me2);
  double RSp461e = RS(mp2,m44,m66,m11,me2);
  double RSp561e = RS(mp2,m55,m66,m11,me2);
  double RSe451p = RS(me2,m44,m55,m11,mp2);
  double RSe461p = RS(me2,m44,m66,m11,mp2);
  double RSe561p = RS(me2,m55,m66,m11,mp2);
  double RS456pe = RS(m44,m55,m66,mp2,me2);
  double RS546pe = RS(m55,m44,m66,mp2,me2);
  double RS645pe = RS(m66,m44,m55,mp2,me2);
  double RS156pe = RS(m11,m55,m66,mp2,me2);
  double RS146pe = RS(m11,m44,m66,mp2,me2);
  double RS145pe = RS(m11,m44,m55,mp2,me2);

  double RD1456ep = RD(m11,m44,m55,m66,me2,mp2);
  double RC1456pe = RC(m11,m44,m55,m66,mp2,me2);  
  double RSe45611p = RS(me2,m44,m55,m66,m11,m11,mp2);
  double RSp45611e = RS(mp2,m44,m55,m66,m11,m11,me2);
  double RS1456ppe = RS(m11,m44,m55,m66,mp2,mp2,me2);
  double RS1456pee = RS(m11,m44,m55,m66,mp2,me2,me2);

  double M12p6x13R =
       + pi162 * (  - 1./32.*m11*pow(mp2,2)*RSp45611e + 1./16.*m11*pow(
         mp2,2)*RSp561e + 1./16.*m11*pow(mp2,2)*RSp461e + 1./16.*m11*
         pow(mp2,2)*RSp451e - 1./96.*m11*pow(mp2,2)*RSp564e - 1./96.*
         m11*pow(mp2,2)*RSp465e - 1./96.*m11*pow(mp2,2)*RSp456e - 1./32.
         *m11*pow(me2,2)*RSe45611p + 1./16.*m11*pow(me2,2)*RSe561p + 1./
         16.*m11*pow(me2,2)*RSe461p + 1./16.*m11*pow(me2,2)*RSe451p - 1.
         /96.*m11*pow(me2,2)*RSe564p - 1./96.*m11*pow(me2,2)*RSe465p - 
         1./96.*m11*pow(me2,2)*RSe456p - 1./32.*m11*m66*RD1456ep - 1./
         32.*m11*m66*mp2*RSp45611e + 1./16.*m11*m66*mp2*RSp451e - 1./32.
         *m11*m66*mp2*RSp456e - 1./32.*m11*m66*me2*RSe45611p + 1./16.*
         m11*m66*me2*RSe451p - 1./32.*m11*m66*me2*RSe456p + 15./64.*m11
         *pow(m66,2) - 1./192.*m11*pow(m66,2)*RC1456pe - 1./192.*m11*
         pow(m66,2)*RSe45611p - 1./192.*m11*pow(m66,2)*RSp45611e + 1./
         96.*m11*pow(m66,2)*RS145pe - 3./64.*m11*pow(m66,2)*RS645pe + 1.
         /96.*m11*pow(m66,2)*RSe451p + 1./96.*m11*pow(m66,2)*RSp451e - 
         1./192.*m11*pow(m66,2)*RSe456p );

      M12p6x13R +=  + pi162 * (  - 1./192.*m11*pow(m66,2)*RSp456e - 1./
         32.*m11*m55*RD1456ep - 1./32.*m11*m55*mp2*RSp45611e + 1./16.*
         m11*m55*mp2*RSp461e - 1./32.*m11*m55*mp2*RSp465e - 1./32.*m11*
         m55*me2*RSe45611p + 1./16.*m11*m55*me2*RSe461p - 1./32.*m11*
         m55*me2*RSe465p + 5./32.*m11*m55*m66 + 15./64.*m11*pow(m55,2)
          - 1./192.*m11*pow(m55,2)*RC1456pe - 1./192.*m11*pow(m55,2)*
         RSe45611p - 1./192.*m11*pow(m55,2)*RSp45611e + 1./96.*m11*pow(
         m55,2)*RS146pe - 3./64.*m11*pow(m55,2)*RS546pe + 1./96.*m11*
         pow(m55,2)*RSe461p + 1./96.*m11*pow(m55,2)*RSp461e - 1./192.*
         m11*pow(m55,2)*RSe465p - 1./192.*m11*pow(m55,2)*RSp465e - 1./
         32.*m11*m44*RD1456ep - 1./32.*m11*m44*mp2*RSp45611e + 1./16.*
         m11*m44*mp2*RSp561e - 1./32.*m11*m44*mp2*RSp564e - 1./32.*m11*
         m44*me2*RSe45611p + 1./16.*m11*m44*me2*RSe561p - 1./32.*m11*
         m44*me2*RSe564p + 5./32.*m11*m44*m66 + 5./32.*m11*m44*m55 + 15.
         /64.*m11*pow(m44,2) - 1./192.*m11*pow(m44,2)*RC1456pe - 1./192.
         *m11*pow(m44,2)*RSe45611p );

      M12p6x13R +=  + pi162 * (  - 1./192.*m11*pow(m44,2)*RSp45611e + 1.
         /96.*m11*pow(m44,2)*RS156pe - 3./64.*m11*pow(m44,2)*RS456pe + 
         1./96.*m11*pow(m44,2)*RSe561p + 1./96.*m11*pow(m44,2)*RSp561e
          - 1./192.*m11*pow(m44,2)*RSe564p - 1./192.*m11*pow(m44,2)*
         RSp564e - 11./96.*pow(m11,2)*RD1456ep - 5./96.*pow(m11,2)*mp2*
         RSp45611e + 5./144.*pow(m11,2)*mp2*RSp561e + 5./144.*pow(
         m11,2)*mp2*RSp461e + 5./144.*pow(m11,2)*mp2*RSp451e - 5./288.*
         pow(m11,2)*mp2*RSp564e - 5./288.*pow(m11,2)*mp2*RSp465e - 5./
         288.*pow(m11,2)*mp2*RSp456e - 5./96.*pow(m11,2)*me2*RSe45611p
          + 5./144.*pow(m11,2)*me2*RSe561p + 5./144.*pow(m11,2)*me2*
         RSe461p + 5./144.*pow(m11,2)*me2*RSe451p - 5./288.*pow(m11,2)*
         me2*RSe564p - 5./288.*pow(m11,2)*me2*RSe465p - 5./288.*pow(
         m11,2)*me2*RSe456p + 13./96.*pow(m11,2)*m66 - 31./864.*pow(
         m11,2)*m66*RC1456pe - 1./216.*pow(m11,2)*m66*RSe45611p - 1./
         216.*pow(m11,2)*m66*RSp45611e + 31./432.*pow(m11,2)*m66*
         RS145pe );

      M12p6x13R +=  + pi162 * (  - 19./864.*pow(m11,2)*m66*RS645pe + 1./
         108.*pow(m11,2)*m66*RSe451p + 1./108.*pow(m11,2)*m66*RSp451e
          - 1./216.*pow(m11,2)*m66*RSe456p - 1./216.*pow(m11,2)*m66*
         RSp456e + 13./96.*pow(m11,2)*m55 - 31./864.*pow(m11,2)*m55*
         RC1456pe - 1./216.*pow(m11,2)*m55*RSe45611p - 1./216.*pow(
         m11,2)*m55*RSp45611e + 31./432.*pow(m11,2)*m55*RS146pe - 19./
         864.*pow(m11,2)*m55*RS546pe + 1./108.*pow(m11,2)*m55*RSe461p
          + 1./108.*pow(m11,2)*m55*RSp461e - 1./216.*pow(m11,2)*m55*
         RSe465p - 1./216.*pow(m11,2)*m55*RSp465e + 13./96.*pow(m11,2)*
         m44 - 31./864.*pow(m11,2)*m44*RC1456pe - 1./216.*pow(m11,2)*
         m44*RSe45611p - 1./216.*pow(m11,2)*m44*RSp45611e + 31./432.*
         pow(m11,2)*m44*RS156pe - 19./864.*pow(m11,2)*m44*RS456pe + 1./
         108.*pow(m11,2)*m44*RSe561p + 1./108.*pow(m11,2)*m44*RSp561e
          - 1./216.*pow(m11,2)*m44*RSe564p - 1./216.*pow(m11,2)*m44*
         RSp564e - 3./16.*pow(m11,3) - 25./288.*pow(m11,3)*RC1456pe - 1.
         /288.*pow(m11,3)*RSe45611p );

      M12p6x13R +=  + pi162 * (  - 1./288.*pow(m11,3)*RSp45611e + 43./
         432.*pow(m11,3)*RS145pe + 43./432.*pow(m11,3)*RS146pe + 43./
         432.*pow(m11,3)*RS156pe - 1./864.*pow(m11,3)*RS645pe - 1./864.
         *pow(m11,3)*RS546pe - 1./864.*pow(m11,3)*RS456pe + 1./432.*
         pow(m11,3)*RSe561p + 1./432.*pow(m11,3)*RSe461p + 1./432.*pow(
         m11,3)*RSe451p + 1./432.*pow(m11,3)*RSp561e + 1./432.*pow(
         m11,3)*RSp461e + 1./432.*pow(m11,3)*RSp451e - 1./864.*pow(
         m11,3)*RSe564p - 1./864.*pow(m11,3)*RSe465p - 1./864.*pow(
         m11,3)*RSe456p - 1./864.*pow(m11,3)*RSp564e - 1./864.*pow(
         m11,3)*RSp465e - 1./864.*pow(m11,3)*RSp456e );

      M12p6x13R +=  + Ab(m11,mu2)*pi16 * (  - 1./4.*pow(m11,2) + 1./12.
         *pow(m11,2)*RS145pe + 1./12.*pow(m11,2)*RS146pe + 1./12.*pow(
         m11,2)*RS156pe );

      M12p6x13R +=  + pow(Ab(m11,mu2),2) * ( 1./2.*m11 + 1./18.*m11*
         pow(RC1456pe,2) + 2./9.*pow(m11,2)*RN1p*RSp45611e*RC1456pe + 2.
         /9.*pow(m11,2)*RN1e*RSe45611p*RC1456pe );

      M12p6x13R +=  + Ab(m11,mu2)*Ab(m14,mu2) * ( 2./45.*RD1456ep + 2./
         45.*m44*RS156pe + 4./9.*m11*RC1456pe + 2./9.*m11*RSe45611p + 2.
         /9.*m11*RSp45611e - 22./45.*m11*RS156pe - 2./9.*m11*RSe561p - 
         2./9.*m11*RSp561e + 2./9.*m11*mp2*RN1p*RSp45611e - 2./9.*m11*
         mp2*RN1p*RSp561e + 2./9.*m11*me2*RN1e*RSe45611p - 2./9.*m11*
         me2*RN1e*RSe561p - 2./9.*m11*m44*RN1p*RSp561e - 2./9.*m11*m44*
         RN1e*RSe561p - 2./9.*pow(m11,2)*RN1p*RSp45611e - 2./9.*pow(
         m11,2)*RN1p*RSp561e - 2./9.*pow(m11,2)*RN1e*RSe45611p - 2./9.*
         pow(m11,2)*RN1e*RSe561p );

      M12p6x13R +=  + Ab(m11,mu2)*Ab(m15,mu2) * ( 2./45.*RD1456ep + 2./
         45.*m55*RS146pe + 4./9.*m11*RC1456pe + 2./9.*m11*RSe45611p + 2.
         /9.*m11*RSp45611e - 22./45.*m11*RS146pe - 2./9.*m11*RSe461p - 
         2./9.*m11*RSp461e + 2./9.*m11*mp2*RN1p*RSp45611e - 2./9.*m11*
         mp2*RN1p*RSp461e + 2./9.*m11*me2*RN1e*RSe45611p - 2./9.*m11*
         me2*RN1e*RSe461p - 2./9.*m11*m55*RN1p*RSp461e - 2./9.*m11*m55*
         RN1e*RSe461p - 2./9.*pow(m11,2)*RN1p*RSp45611e - 2./9.*pow(
         m11,2)*RN1p*RSp461e - 2./9.*pow(m11,2)*RN1e*RSe45611p - 2./9.*
         pow(m11,2)*RN1e*RSe461p );

      M12p6x13R +=  + Ab(m11,mu2)*Ab(m16,mu2) * ( 2./45.*RD1456ep + 2./
         45.*m66*RS145pe + 4./9.*m11*RC1456pe + 2./9.*m11*RSe45611p + 2.
         /9.*m11*RSp45611e - 22./45.*m11*RS145pe - 2./9.*m11*RSe451p - 
         2./9.*m11*RSp451e + 2./9.*m11*mp2*RN1p*RSp45611e - 2./9.*m11*
         mp2*RN1p*RSp451e + 2./9.*m11*me2*RN1e*RSe45611p - 2./9.*m11*
         me2*RN1e*RSe451p - 2./9.*m11*m66*RN1p*RSp451e - 2./9.*m11*m66*
         RN1e*RSe451p - 2./9.*pow(m11,2)*RN1p*RSp45611e - 2./9.*pow(
         m11,2)*RN1p*RSp451e - 2./9.*pow(m11,2)*RN1e*RSe45611p - 2./9.*
         pow(m11,2)*RN1e*RSe451p );

      M12p6x13R +=  + Ab(m11,mu2)*Ab(m44,mu2) * ( 2./9.*m11*m44*RN1p*
         RSp561e*RS156pe - 2./27.*m11*m44*RN1p*RSp561e*RS456pe*RS156pe
          + 2./9.*m11*m44*RN1e*RSe561p*RS156pe - 2./27.*m11*m44*RN1e*
         RSe561p*RS456pe*RS156pe );

      M12p6x13R +=  + Ab(m11,mu2)*Ab(m45,mu2) * (  - 1./27.*m11*pow(
         RS146pe,2) + 2./27.*m11*RS156pe*RS146pe - 1./27.*m11*pow(
         RS156pe,2) + 2./27.*m11*RSe561p*RS146pe - 2./27.*m11*RSe561p*
         RS156pe - 2./27.*m11*RSe461p*RS146pe + 2./27.*m11*RSe461p*
         RS156pe + 2./27.*m11*RSp561e*RS146pe - 2./27.*m11*RSp561e*
         RS156pe - 2./27.*m11*RSp461e*RS146pe + 2./27.*m11*RSp461e*
         RS156pe + 2./27.*m11*mp2*RN1p*RSp561e*RS146pe - 2./27.*m11*mp2
         *RN1p*RSp561e*RS156pe - 2./27.*m11*mp2*RN1p*RSp461e*RS146pe + 
         2./27.*m11*mp2*RN1p*RSp461e*RS156pe + 2./27.*m11*me2*RN1e*
         RSe561p*RS146pe - 2./27.*m11*me2*RN1e*RSe561p*RS156pe - 2./27.
         *m11*me2*RN1e*RSe461p*RS146pe + 2./27.*m11*me2*RN1e*RSe461p*
         RS156pe + 2./27.*m11*m55*RN1p*RSp561e*RS146pe + 2./27.*m11*m55
         *RN1p*RSp461e*RS146pe + 2./27.*m11*m55*RN1p*RSp461e*RS156pe + 
         2./27.*m11*m55*RN1e*RSe561p*RS146pe + 2./27.*m11*m55*RN1e*
         RSe461p*RS146pe + 2./27.*m11*m55*RN1e*RSe461p*RS156pe + 2./27.
         *m11*m44*RN1p*RSp561e*RS146pe );

      M12p6x13R +=  + Ab(m11,mu2)*Ab(m45,mu2) * ( 2./27.*m11*m44*RN1p*
         RSp561e*RS156pe + 2./27.*m11*m44*RN1p*RSp461e*RS156pe + 2./27.
         *m11*m44*RN1e*RSe561p*RS146pe + 2./27.*m11*m44*RN1e*RSe561p*
         RS156pe + 2./27.*m11*m44*RN1e*RSe461p*RS156pe );

      M12p6x13R +=  + Ab(m11,mu2)*Ab(m46,mu2) * (  - 1./27.*m11*pow(
         RS145pe,2) + 2./27.*m11*RS156pe*RS145pe - 1./27.*m11*pow(
         RS156pe,2) + 2./27.*m11*RSe561p*RS145pe - 2./27.*m11*RSe561p*
         RS156pe - 2./27.*m11*RSe451p*RS145pe + 2./27.*m11*RSe451p*
         RS156pe + 2./27.*m11*RSp561e*RS145pe - 2./27.*m11*RSp561e*
         RS156pe - 2./27.*m11*RSp451e*RS145pe + 2./27.*m11*RSp451e*
         RS156pe + 2./27.*m11*mp2*RN1p*RSp561e*RS145pe - 2./27.*m11*mp2
         *RN1p*RSp561e*RS156pe - 2./27.*m11*mp2*RN1p*RSp451e*RS145pe + 
         2./27.*m11*mp2*RN1p*RSp451e*RS156pe + 2./27.*m11*me2*RN1e*
         RSe561p*RS145pe - 2./27.*m11*me2*RN1e*RSe561p*RS156pe - 2./27.
         *m11*me2*RN1e*RSe451p*RS145pe + 2./27.*m11*me2*RN1e*RSe451p*
         RS156pe + 2./27.*m11*m66*RN1p*RSp561e*RS145pe + 2./27.*m11*m66
         *RN1p*RSp451e*RS145pe + 2./27.*m11*m66*RN1p*RSp451e*RS156pe + 
         2./27.*m11*m66*RN1e*RSe561p*RS145pe + 2./27.*m11*m66*RN1e*
         RSe451p*RS145pe + 2./27.*m11*m66*RN1e*RSe451p*RS156pe + 2./27.
         *m11*m44*RN1p*RSp561e*RS145pe );

      M12p6x13R +=  + Ab(m11,mu2)*Ab(m46,mu2) * ( 2./27.*m11*m44*RN1p*
         RSp561e*RS156pe + 2./27.*m11*m44*RN1p*RSp451e*RS156pe + 2./27.
         *m11*m44*RN1e*RSe561p*RS145pe + 2./27.*m11*m44*RN1e*RSe561p*
         RS156pe + 2./27.*m11*m44*RN1e*RSe451p*RS156pe );

      M12p6x13R +=  + Ab(m11,mu2)*Ab(m55,mu2) * ( 2./9.*m11*m55*RN1p*
         RSp461e*RS146pe - 2./27.*m11*m55*RN1p*RSp461e*RS546pe*RS146pe
          + 2./9.*m11*m55*RN1e*RSe461p*RS146pe - 2./27.*m11*m55*RN1e*
         RSe461p*RS546pe*RS146pe );

      M12p6x13R +=  + Ab(m11,mu2)*Ab(m56,mu2) * (  - 1./27.*m11*pow(
         RS145pe,2) + 2./27.*m11*RS146pe*RS145pe - 1./27.*m11*pow(
         RS146pe,2) + 2./27.*m11*RSe461p*RS145pe - 2./27.*m11*RSe461p*
         RS146pe - 2./27.*m11*RSe451p*RS145pe + 2./27.*m11*RSe451p*
         RS146pe + 2./27.*m11*RSp461e*RS145pe - 2./27.*m11*RSp461e*
         RS146pe - 2./27.*m11*RSp451e*RS145pe + 2./27.*m11*RSp451e*
         RS146pe + 2./27.*m11*mp2*RN1p*RSp461e*RS145pe - 2./27.*m11*mp2
         *RN1p*RSp461e*RS146pe - 2./27.*m11*mp2*RN1p*RSp451e*RS145pe + 
         2./27.*m11*mp2*RN1p*RSp451e*RS146pe + 2./27.*m11*me2*RN1e*
         RSe461p*RS145pe - 2./27.*m11*me2*RN1e*RSe461p*RS146pe - 2./27.
         *m11*me2*RN1e*RSe451p*RS145pe + 2./27.*m11*me2*RN1e*RSe451p*
         RS146pe + 2./27.*m11*m66*RN1p*RSp461e*RS145pe + 2./27.*m11*m66
         *RN1p*RSp451e*RS145pe + 2./27.*m11*m66*RN1p*RSp451e*RS146pe + 
         2./27.*m11*m66*RN1e*RSe461p*RS145pe + 2./27.*m11*m66*RN1e*
         RSe451p*RS145pe + 2./27.*m11*m66*RN1e*RSe451p*RS146pe + 2./27.
         *m11*m55*RN1p*RSp461e*RS145pe );

      M12p6x13R +=  + Ab(m11,mu2)*Ab(m56,mu2) * ( 2./27.*m11*m55*RN1p*
         RSp461e*RS146pe + 2./27.*m11*m55*RN1p*RSp451e*RS146pe + 2./27.
         *m11*m55*RN1e*RSe461p*RS145pe + 2./27.*m11*m55*RN1e*RSe461p*
         RS146pe + 2./27.*m11*m55*RN1e*RSe451p*RS146pe );

      M12p6x13R +=  + Ab(m11,mu2)*Ab(m66,mu2) * ( 2./9.*m11*m66*RN1p*
         RSp451e*RS145pe - 2./27.*m11*m66*RN1p*RSp451e*RS645pe*RS145pe
          + 2./9.*m11*m66*RN1e*RSe451p*RS145pe - 2./27.*m11*m66*RN1e*
         RSe451p*RS645pe*RS145pe );

      M12p6x13R +=  + Ab(m11,mu2)*Ab(me2,mu2) * ( 1./9.*m11*RSe45611p*
         RC1456pe - 2./27.*m11*m66*RN1p*RSe456p*RSp451e*RS145pe - 2./27.
         *m11*m66*RN1e*RSe456p*RSe451p*RS145pe - 2./27.*m11*m55*RN1p*
         RSe465p*RSp461e*RS146pe - 2./27.*m11*m55*RN1e*RSe465p*RSe461p*
         RS146pe - 2./27.*m11*m44*RN1p*RSe564p*RSp561e*RS156pe - 2./27.
         *m11*m44*RN1e*RSe564p*RSe561p*RS156pe + 2./9.*pow(m11,2)*RN1p*
         RSp45611e*RSe45611p - 2./9.*pow(m11,2)*RN1e*RSe45611p*RC1456pe
          + 2./9.*pow(m11,2)*RN1e*pow(RSe45611p,2) );

      M12p6x13R +=  + Ab(m11,mu2)*Ab(mp2,mu2) * ( 1./9.*m11*RSp45611e*
         RC1456pe - 2./27.*m11*m66*RN1p*RSp456e*RSp451e*RS145pe - 2./27.
         *m11*m66*RN1e*RSp456e*RSe451p*RS145pe - 2./27.*m11*m55*RN1p*
         RSp465e*RSp461e*RS146pe - 2./27.*m11*m55*RN1e*RSp465e*RSe461p*
         RS146pe - 2./27.*m11*m44*RN1p*RSp564e*RSp561e*RS156pe - 2./27.
         *m11*m44*RN1e*RSp564e*RSe561p*RS156pe - 2./9.*pow(m11,2)*RN1p*
         RSp45611e*RC1456pe + 2./9.*pow(m11,2)*RN1p*pow(RSp45611e,2) + 
         2./9.*pow(m11,2)*RN1e*RSp45611e*RSe45611p );

      M12p6x13R +=  + Ab(m11,mu2)*Bb(m11,mu2) * ( 1./9.*m11*RD1456ep*
         RC1456pe + pow(m11,2) + 2./9.*pow(m11,2)*pow(RC1456pe,2) + 2./
         9.*pow(m11,2)*RN1p*RSp45611e*RD1456ep + 2./9.*pow(m11,2)*RN1e*
         RSe45611p*RD1456ep );

      M12p6x13R +=  + Ab(m11,mu2)*Cb(m11,mu2) * ( 2./9.*pow(m11,2)*
         RD1456ep*RC1456pe );

      M12p6x13R +=  + Ab(m14,mu2)*pi16 * (  - 1./12.*m11*RD1456ep - 1./
         12.*m11*mp2*RSp45611e + 1./6.*m11*mp2*RSp561e - 1./12.*m11*mp2
         *RSp564e - 1./12.*m11*me2*RSe45611p + 1./6.*m11*me2*RSe561p - 
         1./12.*m11*me2*RSe564p + 1./4.*m11*m66 + 1./4.*m11*m55 + 1./2.
         *m11*m44 - 1./24.*m11*m44*RC1456pe - 1./24.*m11*m44*RSe45611p
          - 1./24.*m11*m44*RSp45611e + 1./12.*m11*m44*RS156pe - 1./8.*
         m11*m44*RS456pe + 1./12.*m11*m44*RSe561p + 1./12.*m11*m44*
         RSp561e - 1./24.*m11*m44*RSe564p - 1./24.*m11*m44*RSp564e - 7./
         72.*pow(m11,2)*RC1456pe - 1./72.*pow(m11,2)*RSe45611p - 1./72.
         *pow(m11,2)*RSp45611e + 7./36.*pow(m11,2)*RS156pe - 1./72.*
         pow(m11,2)*RS456pe + 1./36.*pow(m11,2)*RSe561p + 1./36.*pow(
         m11,2)*RSp561e - 1./72.*pow(m11,2)*RSe564p - 1./72.*pow(m11,2)
         *RSp564e );

      M12p6x13R +=  + pow(Ab(m14,mu2),2) * (  - 1./27.*RD1456ep - 1./27.
         *mp2*RSp45611e + 2./27.*mp2*RSp561e - 1./27.*mp2*RSp564e - 1./
         27.*me2*RSe45611p + 2./27.*me2*RSe561p - 1./27.*me2*RSe564p - 
         1./27.*m44*RC1456pe - 1./27.*m44*RSe45611p - 1./27.*m44*
         RSp45611e + 1./27.*m44*RSe564p + 1./27.*m44*RSp564e - 1./4.*
         m11 - 13./108.*m11*RC1456pe - 1./12.*m11*RSe45611p - 1./12.*
         m11*RSp45611e + 5./54.*m11*RS156pe + 7./108.*m11*RS456pe + 1./
         54.*m11*RSe561p + 1./54.*m11*RSp561e + 7./108.*m11*RSe564p + 7.
         /108.*m11*RSp564e );

      M12p6x13R +=  + Ab(m14,mu2)*Ab(m15,mu2) * (  - 1./2.*m11 );

      M12p6x13R +=  + Ab(m14,mu2)*Ab(m16,mu2) * (  - 1./2.*m11 );

      M12p6x13R +=  + Ab(m14,mu2)*Ab(me2,mu2) * ( 2./45.*me2*RSe45611p
          - 4./45.*me2*RSe561p + 2./45.*me2*RSe564p + 2./45.*m44*
         RSe561p - 2./45.*m44*RSe564p + 8./45.*m11*RSe45611p - 8./45.*
         m11*RSe561p - 2./9.*m11*me2*RN1e*RSe45611p + 2./9.*m11*me2*
         RN1e*RSe561p + 2./9.*m11*m44*RN1e*RSe561p + 2./9.*pow(m11,2)*
         RN1e*RSe45611p + 2./9.*pow(m11,2)*RN1e*RSe561p );

      M12p6x13R +=  + Ab(m14,mu2)*Ab(mp2,mu2) * ( 2./45.*mp2*RSp45611e
          - 4./45.*mp2*RSp561e + 2./45.*mp2*RSp564e + 2./45.*m44*
         RSp561e - 2./45.*m44*RSp564e + 8./45.*m11*RSp45611e - 8./45.*
         m11*RSp561e - 2./9.*m11*mp2*RN1p*RSp45611e + 2./9.*m11*mp2*
         RN1p*RSp561e + 2./9.*m11*m44*RN1p*RSp561e + 2./9.*pow(m11,2)*
         RN1p*RSp45611e + 2./9.*pow(m11,2)*RN1p*RSp561e );

      M12p6x13R +=  + Ab(m14,mu2)*Bb(m11,mu2) * ( 4./9.*m11*RD1456ep - 
         2./9.*m11*m44*RS156pe - 4./9.*pow(m11,2)*RS156pe );

      M12p6x13R +=  + Ab(m15,mu2)*pi16 * (  - 1./12.*m11*RD1456ep - 1./
         12.*m11*mp2*RSp45611e + 1./6.*m11*mp2*RSp461e - 1./12.*m11*mp2
         *RSp465e - 1./12.*m11*me2*RSe45611p + 1./6.*m11*me2*RSe461p - 
         1./12.*m11*me2*RSe465p + 1./4.*m11*m66 + 1./2.*m11*m55 - 1./24.
         *m11*m55*RC1456pe - 1./24.*m11*m55*RSe45611p - 1./24.*m11*m55*
         RSp45611e + 1./12.*m11*m55*RS146pe - 1./8.*m11*m55*RS546pe + 1.
         /12.*m11*m55*RSe461p + 1./12.*m11*m55*RSp461e - 1./24.*m11*m55
         *RSe465p - 1./24.*m11*m55*RSp465e + 1./4.*m11*m44 - 7./72.*
         pow(m11,2)*RC1456pe - 1./72.*pow(m11,2)*RSe45611p - 1./72.*
         pow(m11,2)*RSp45611e + 7./36.*pow(m11,2)*RS146pe - 1./72.*pow(
         m11,2)*RS546pe + 1./36.*pow(m11,2)*RSe461p + 1./36.*pow(m11,2)
         *RSp461e - 1./72.*pow(m11,2)*RSe465p - 1./72.*pow(m11,2)*
         RSp465e );

      M12p6x13R +=  + pow(Ab(m15,mu2),2) * (  - 1./27.*RD1456ep - 1./27.
         *mp2*RSp45611e + 2./27.*mp2*RSp461e - 1./27.*mp2*RSp465e - 1./
         27.*me2*RSe45611p + 2./27.*me2*RSe461p - 1./27.*me2*RSe465p - 
         1./27.*m55*RC1456pe - 1./27.*m55*RSe45611p - 1./27.*m55*
         RSp45611e + 1./27.*m55*RSe465p + 1./27.*m55*RSp465e - 1./4.*
         m11 - 13./108.*m11*RC1456pe - 1./12.*m11*RSe45611p - 1./12.*
         m11*RSp45611e + 5./54.*m11*RS146pe + 7./108.*m11*RS546pe + 1./
         54.*m11*RSe461p + 1./54.*m11*RSp461e + 7./108.*m11*RSe465p + 7.
         /108.*m11*RSp465e );

      M12p6x13R +=  + Ab(m15,mu2)*Ab(m16,mu2) * (  - 1./2.*m11 );

      M12p6x13R +=  + Ab(m15,mu2)*Ab(me2,mu2) * ( 2./45.*me2*RSe45611p
          - 4./45.*me2*RSe461p + 2./45.*me2*RSe465p + 2./45.*m55*
         RSe461p - 2./45.*m55*RSe465p + 8./45.*m11*RSe45611p - 8./45.*
         m11*RSe461p - 2./9.*m11*me2*RN1e*RSe45611p + 2./9.*m11*me2*
         RN1e*RSe461p + 2./9.*m11*m55*RN1e*RSe461p + 2./9.*pow(m11,2)*
         RN1e*RSe45611p + 2./9.*pow(m11,2)*RN1e*RSe461p );

      M12p6x13R +=  + Ab(m15,mu2)*Ab(mp2,mu2) * ( 2./45.*mp2*RSp45611e
          - 4./45.*mp2*RSp461e + 2./45.*mp2*RSp465e + 2./45.*m55*
         RSp461e - 2./45.*m55*RSp465e + 8./45.*m11*RSp45611e - 8./45.*
         m11*RSp461e - 2./9.*m11*mp2*RN1p*RSp45611e + 2./9.*m11*mp2*
         RN1p*RSp461e + 2./9.*m11*m55*RN1p*RSp461e + 2./9.*pow(m11,2)*
         RN1p*RSp45611e + 2./9.*pow(m11,2)*RN1p*RSp461e );

      M12p6x13R +=  + Ab(m15,mu2)*Bb(m11,mu2) * ( 4./9.*m11*RD1456ep - 
         2./9.*m11*m55*RS146pe - 4./9.*pow(m11,2)*RS146pe );

      M12p6x13R +=  + Ab(m16,mu2)*pi16 * (  - 1./12.*m11*RD1456ep - 1./
         12.*m11*mp2*RSp45611e + 1./6.*m11*mp2*RSp451e - 1./12.*m11*mp2
         *RSp456e - 1./12.*m11*me2*RSe45611p + 1./6.*m11*me2*RSe451p - 
         1./12.*m11*me2*RSe456p + 1./2.*m11*m66 - 1./24.*m11*m66*
         RC1456pe - 1./24.*m11*m66*RSe45611p - 1./24.*m11*m66*RSp45611e
          + 1./12.*m11*m66*RS145pe - 1./8.*m11*m66*RS645pe + 1./12.*m11
         *m66*RSe451p + 1./12.*m11*m66*RSp451e - 1./24.*m11*m66*RSe456p
          - 1./24.*m11*m66*RSp456e + 1./4.*m11*m55 + 1./4.*m11*m44 - 7./
         72.*pow(m11,2)*RC1456pe - 1./72.*pow(m11,2)*RSe45611p - 1./72.
         *pow(m11,2)*RSp45611e + 7./36.*pow(m11,2)*RS145pe - 1./72.*
         pow(m11,2)*RS645pe + 1./36.*pow(m11,2)*RSe451p + 1./36.*pow(
         m11,2)*RSp451e - 1./72.*pow(m11,2)*RSe456p - 1./72.*pow(m11,2)
         *RSp456e );

      M12p6x13R +=  + pow(Ab(m16,mu2),2) * (  - 1./27.*RD1456ep - 1./27.
         *mp2*RSp45611e + 2./27.*mp2*RSp451e - 1./27.*mp2*RSp456e - 1./
         27.*me2*RSe45611p + 2./27.*me2*RSe451p - 1./27.*me2*RSe456p - 
         1./27.*m66*RC1456pe - 1./27.*m66*RSe45611p - 1./27.*m66*
         RSp45611e + 1./27.*m66*RSe456p + 1./27.*m66*RSp456e - 1./4.*
         m11 - 13./108.*m11*RC1456pe - 1./12.*m11*RSe45611p - 1./12.*
         m11*RSp45611e + 5./54.*m11*RS145pe + 7./108.*m11*RS645pe + 1./
         54.*m11*RSe451p + 1./54.*m11*RSp451e + 7./108.*m11*RSe456p + 7.
         /108.*m11*RSp456e );

      M12p6x13R +=  + Ab(m16,mu2)*Ab(me2,mu2) * ( 2./45.*me2*RSe45611p
          - 4./45.*me2*RSe451p + 2./45.*me2*RSe456p + 2./45.*m66*
         RSe451p - 2./45.*m66*RSe456p + 8./45.*m11*RSe45611p - 8./45.*
         m11*RSe451p - 2./9.*m11*me2*RN1e*RSe45611p + 2./9.*m11*me2*
         RN1e*RSe451p + 2./9.*m11*m66*RN1e*RSe451p + 2./9.*pow(m11,2)*
         RN1e*RSe45611p + 2./9.*pow(m11,2)*RN1e*RSe451p );

      M12p6x13R +=  + Ab(m16,mu2)*Ab(mp2,mu2) * ( 2./45.*mp2*RSp45611e
          - 4./45.*mp2*RSp451e + 2./45.*mp2*RSp456e + 2./45.*m66*
         RSp451e - 2./45.*m66*RSp456e + 8./45.*m11*RSp45611e - 8./45.*
         m11*RSp451e - 2./9.*m11*mp2*RN1p*RSp45611e + 2./9.*m11*mp2*
         RN1p*RSp451e + 2./9.*m11*m66*RN1p*RSp451e + 2./9.*pow(m11,2)*
         RN1p*RSp45611e + 2./9.*pow(m11,2)*RN1p*RSp451e );

      M12p6x13R +=  + Ab(m16,mu2)*Bb(m11,mu2) * ( 4./9.*m11*RD1456ep - 
         2./9.*m11*m66*RS145pe - 4./9.*pow(m11,2)*RS145pe );

      M12p6x13R +=  + Ab(m44,mu2)*Ab(me2,mu2) * ( 2./9.*m11*m44*RNep*
         RSp561e*RSe561p - 2./27.*m11*m44*RNep*RSp561e*RSe561p*RS456pe
          - 2./9.*m11*m44*RN1e*RSe561p*RS156pe + 2./27.*m11*m44*RN1e*
         RSe561p*RS456pe*RS156pe );

      M12p6x13R +=  + Ab(m44,mu2)*Ab(mp2,mu2) * (  - 2./9.*m11*m44*RNep
         *RSp561e*RSe561p + 2./27.*m11*m44*RNep*RSp561e*RSe561p*RS456pe
          - 2./9.*m11*m44*RN1p*RSp561e*RS156pe + 2./27.*m11*m44*RN1p*
         RSp561e*RS456pe*RS156pe );

      M12p6x13R +=  + Ab(m44,mu2)*Bb(m11,mu2) * ( 1./9.*m11*m44*pow(
         RS156pe,2) - 1./27.*m11*m44*RS456pe*pow(RS156pe,2) );

      M12p6x13R +=  + Ab(m44,mu2)*Bb(me2,mu2) * ( 1./9.*m11*m44*pow(
         RSe561p,2) - 1./27.*m11*m44*pow(RSe561p,2)*RS456pe );

      M12p6x13R +=  + Ab(m44,mu2)*Bb(mp2,mu2) * ( 1./9.*m11*m44*pow(
         RSp561e,2) - 1./27.*m11*m44*pow(RSp561e,2)*RS456pe );

      M12p6x13R +=  + Ab(m45,mu2)*Ab(me2,mu2) * (  - 1./27.*m11*pow(
         RSe561p,2) + 2./27.*m11*RSe461p*RSe561p - 1./27.*m11*pow(
         RSe461p,2) - 2./27.*m11*RSp561e*RSe561p + 2./27.*m11*RSp561e*
         RSe461p + 2./27.*m11*RSp461e*RSe561p - 2./27.*m11*RSp461e*
         RSe461p - 2./27.*m11*mp2*RNep*RSp561e*RSe561p + 2./27.*m11*mp2
         *RNep*RSp561e*RSe461p + 2./27.*m11*mp2*RNep*RSp461e*RSe561p - 
         2./27.*m11*mp2*RNep*RSp461e*RSe461p - 2./27.*m11*me2*RN1e*
         RSe561p*RS146pe + 2./27.*m11*me2*RN1e*RSe561p*RS156pe + 2./27.
         *m11*me2*RN1e*RSe461p*RS146pe - 2./27.*m11*me2*RN1e*RSe461p*
         RS156pe + 2./27.*m11*m55*RNep*RSp561e*RSe461p + 2./27.*m11*m55
         *RNep*RSp461e*RSe561p + 2./27.*m11*m55*RNep*RSp461e*RSe461p - 
         2./27.*m11*m55*RN1e*RSe561p*RS146pe - 2./27.*m11*m55*RN1e*
         RSe461p*RS146pe - 2./27.*m11*m55*RN1e*RSe461p*RS156pe + 2./27.
         *m11*m44*RNep*RSp561e*RSe561p + 2./27.*m11*m44*RNep*RSp561e*
         RSe461p + 2./27.*m11*m44*RNep*RSp461e*RSe561p - 2./27.*m11*m44
         *RN1e*RSe561p*RS146pe );

      M12p6x13R +=  + Ab(m45,mu2)*Ab(me2,mu2) * (  - 2./27.*m11*m44*
         RN1e*RSe561p*RS156pe - 2./27.*m11*m44*RN1e*RSe461p*RS156pe );

      M12p6x13R +=  + Ab(m45,mu2)*Ab(mp2,mu2) * (  - 1./27.*m11*pow(
         RSp561e,2) + 2./27.*m11*RSp461e*RSp561e - 1./27.*m11*pow(
         RSp461e,2) + 2./27.*m11*mp2*RNep*RSp561e*RSe561p - 2./27.*m11*
         mp2*RNep*RSp561e*RSe461p - 2./27.*m11*mp2*RNep*RSp461e*RSe561p
          + 2./27.*m11*mp2*RNep*RSp461e*RSe461p - 2./27.*m11*mp2*RN1p*
         RSp561e*RS146pe + 2./27.*m11*mp2*RN1p*RSp561e*RS156pe + 2./27.
         *m11*mp2*RN1p*RSp461e*RS146pe - 2./27.*m11*mp2*RN1p*RSp461e*
         RS156pe - 2./27.*m11*m55*RNep*RSp561e*RSe461p - 2./27.*m11*m55
         *RNep*RSp461e*RSe561p - 2./27.*m11*m55*RNep*RSp461e*RSe461p - 
         2./27.*m11*m55*RN1p*RSp561e*RS146pe - 2./27.*m11*m55*RN1p*
         RSp461e*RS146pe - 2./27.*m11*m55*RN1p*RSp461e*RS156pe - 2./27.
         *m11*m44*RNep*RSp561e*RSe561p - 2./27.*m11*m44*RNep*RSp561e*
         RSe461p - 2./27.*m11*m44*RNep*RSp461e*RSe561p - 2./27.*m11*m44
         *RN1p*RSp561e*RS146pe - 2./27.*m11*m44*RN1p*RSp561e*RS156pe - 
         2./27.*m11*m44*RN1p*RSp461e*RS156pe );

      M12p6x13R +=  + Ab(m45,mu2)*Bb(m11,mu2) * ( 1./27.*m11*m55*pow(
         RS146pe,2) + 2./27.*m11*m55*RS156pe*RS146pe + 2./27.*m11*m44*
         RS156pe*RS146pe + 1./27.*m11*m44*pow(RS156pe,2) - 1./27.*pow(
         m11,2)*pow(RS146pe,2) + 2./27.*pow(m11,2)*RS156pe*RS146pe - 1./
         27.*pow(m11,2)*pow(RS156pe,2) );

      M12p6x13R +=  + Ab(m45,mu2)*Bb(me2,mu2) * (  - 1./27.*m11*me2*
         pow(RSe561p,2) + 2./27.*m11*me2*RSe461p*RSe561p - 1./27.*m11*
         me2*pow(RSe461p,2) + 2./27.*m11*m55*RSe461p*RSe561p + 1./27.*
         m11*m55*pow(RSe461p,2) + 1./27.*m11*m44*pow(RSe561p,2) + 2./27.
         *m11*m44*RSe461p*RSe561p );

      M12p6x13R +=  + Ab(m45,mu2)*Bb(mp2,mu2) * (  - 1./27.*m11*mp2*
         pow(RSp561e,2) + 2./27.*m11*mp2*RSp461e*RSp561e - 1./27.*m11*
         mp2*pow(RSp461e,2) + 2./27.*m11*m55*RSp461e*RSp561e + 1./27.*
         m11*m55*pow(RSp461e,2) + 1./27.*m11*m44*pow(RSp561e,2) + 2./27.
         *m11*m44*RSp461e*RSp561e );

      M12p6x13R +=  + Ab(m46,mu2)*Ab(me2,mu2) * (  - 1./27.*m11*pow(
         RSe561p,2) + 2./27.*m11*RSe451p*RSe561p - 1./27.*m11*pow(
         RSe451p,2) - 2./27.*m11*RSp561e*RSe561p + 2./27.*m11*RSp561e*
         RSe451p + 2./27.*m11*RSp451e*RSe561p - 2./27.*m11*RSp451e*
         RSe451p - 2./27.*m11*mp2*RNep*RSp561e*RSe561p + 2./27.*m11*mp2
         *RNep*RSp561e*RSe451p + 2./27.*m11*mp2*RNep*RSp451e*RSe561p - 
         2./27.*m11*mp2*RNep*RSp451e*RSe451p - 2./27.*m11*me2*RN1e*
         RSe561p*RS145pe + 2./27.*m11*me2*RN1e*RSe561p*RS156pe + 2./27.
         *m11*me2*RN1e*RSe451p*RS145pe - 2./27.*m11*me2*RN1e*RSe451p*
         RS156pe + 2./27.*m11*m66*RNep*RSp561e*RSe451p + 2./27.*m11*m66
         *RNep*RSp451e*RSe561p + 2./27.*m11*m66*RNep*RSp451e*RSe451p - 
         2./27.*m11*m66*RN1e*RSe561p*RS145pe - 2./27.*m11*m66*RN1e*
         RSe451p*RS145pe - 2./27.*m11*m66*RN1e*RSe451p*RS156pe + 2./27.
         *m11*m44*RNep*RSp561e*RSe561p + 2./27.*m11*m44*RNep*RSp561e*
         RSe451p + 2./27.*m11*m44*RNep*RSp451e*RSe561p - 2./27.*m11*m44
         *RN1e*RSe561p*RS145pe );

      M12p6x13R +=  + Ab(m46,mu2)*Ab(me2,mu2) * (  - 2./27.*m11*m44*
         RN1e*RSe561p*RS156pe - 2./27.*m11*m44*RN1e*RSe451p*RS156pe );

      M12p6x13R +=  + Ab(m46,mu2)*Ab(mp2,mu2) * (  - 1./27.*m11*pow(
         RSp561e,2) + 2./27.*m11*RSp451e*RSp561e - 1./27.*m11*pow(
         RSp451e,2) + 2./27.*m11*mp2*RNep*RSp561e*RSe561p - 2./27.*m11*
         mp2*RNep*RSp561e*RSe451p - 2./27.*m11*mp2*RNep*RSp451e*RSe561p
          + 2./27.*m11*mp2*RNep*RSp451e*RSe451p - 2./27.*m11*mp2*RN1p*
         RSp561e*RS145pe + 2./27.*m11*mp2*RN1p*RSp561e*RS156pe + 2./27.
         *m11*mp2*RN1p*RSp451e*RS145pe - 2./27.*m11*mp2*RN1p*RSp451e*
         RS156pe - 2./27.*m11*m66*RNep*RSp561e*RSe451p - 2./27.*m11*m66
         *RNep*RSp451e*RSe561p - 2./27.*m11*m66*RNep*RSp451e*RSe451p - 
         2./27.*m11*m66*RN1p*RSp561e*RS145pe - 2./27.*m11*m66*RN1p*
         RSp451e*RS145pe - 2./27.*m11*m66*RN1p*RSp451e*RS156pe - 2./27.
         *m11*m44*RNep*RSp561e*RSe561p - 2./27.*m11*m44*RNep*RSp561e*
         RSe451p - 2./27.*m11*m44*RNep*RSp451e*RSe561p - 2./27.*m11*m44
         *RN1p*RSp561e*RS145pe - 2./27.*m11*m44*RN1p*RSp561e*RS156pe - 
         2./27.*m11*m44*RN1p*RSp451e*RS156pe );

      M12p6x13R +=  + Ab(m46,mu2)*Bb(m11,mu2) * ( 1./27.*m11*m66*pow(
         RS145pe,2) + 2./27.*m11*m66*RS156pe*RS145pe + 2./27.*m11*m44*
         RS156pe*RS145pe + 1./27.*m11*m44*pow(RS156pe,2) - 1./27.*pow(
         m11,2)*pow(RS145pe,2) + 2./27.*pow(m11,2)*RS156pe*RS145pe - 1./
         27.*pow(m11,2)*pow(RS156pe,2) );

      M12p6x13R +=  + Ab(m46,mu2)*Bb(me2,mu2) * (  - 1./27.*m11*me2*
         pow(RSe561p,2) + 2./27.*m11*me2*RSe451p*RSe561p - 1./27.*m11*
         me2*pow(RSe451p,2) + 2./27.*m11*m66*RSe451p*RSe561p + 1./27.*
         m11*m66*pow(RSe451p,2) + 1./27.*m11*m44*pow(RSe561p,2) + 2./27.
         *m11*m44*RSe451p*RSe561p );

      M12p6x13R +=  + Ab(m46,mu2)*Bb(mp2,mu2) * (  - 1./27.*m11*mp2*
         pow(RSp561e,2) + 2./27.*m11*mp2*RSp451e*RSp561e - 1./27.*m11*
         mp2*pow(RSp451e,2) + 2./27.*m11*m66*RSp451e*RSp561e + 1./27.*
         m11*m66*pow(RSp451e,2) + 1./27.*m11*m44*pow(RSp561e,2) + 2./27.
         *m11*m44*RSp451e*RSp561e );

      M12p6x13R +=  + Ab(m55,mu2)*Ab(me2,mu2) * ( 2./9.*m11*m55*RNep*
         RSp461e*RSe461p - 2./27.*m11*m55*RNep*RSp461e*RSe461p*RS546pe
          - 2./9.*m11*m55*RN1e*RSe461p*RS146pe + 2./27.*m11*m55*RN1e*
         RSe461p*RS546pe*RS146pe );

      M12p6x13R +=  + Ab(m55,mu2)*Ab(mp2,mu2) * (  - 2./9.*m11*m55*RNep
         *RSp461e*RSe461p + 2./27.*m11*m55*RNep*RSp461e*RSe461p*RS546pe
          - 2./9.*m11*m55*RN1p*RSp461e*RS146pe + 2./27.*m11*m55*RN1p*
         RSp461e*RS546pe*RS146pe );

      M12p6x13R +=  + Ab(m55,mu2)*Bb(m11,mu2) * ( 1./9.*m11*m55*pow(
         RS146pe,2) - 1./27.*m11*m55*RS546pe*pow(RS146pe,2) );

      M12p6x13R +=  + Ab(m55,mu2)*Bb(me2,mu2) * ( 1./9.*m11*m55*pow(
         RSe461p,2) - 1./27.*m11*m55*pow(RSe461p,2)*RS546pe );

      M12p6x13R +=  + Ab(m55,mu2)*Bb(mp2,mu2) * ( 1./9.*m11*m55*pow(
         RSp461e,2) - 1./27.*m11*m55*pow(RSp461e,2)*RS546pe );

      M12p6x13R +=  + Ab(m56,mu2)*Ab(me2,mu2) * (  - 1./27.*m11*pow(
         RSe461p,2) + 2./27.*m11*RSe451p*RSe461p - 1./27.*m11*pow(
         RSe451p,2) - 2./27.*m11*RSp461e*RSe461p + 2./27.*m11*RSp461e*
         RSe451p + 2./27.*m11*RSp451e*RSe461p - 2./27.*m11*RSp451e*
         RSe451p - 2./27.*m11*mp2*RNep*RSp461e*RSe461p + 2./27.*m11*mp2
         *RNep*RSp461e*RSe451p + 2./27.*m11*mp2*RNep*RSp451e*RSe461p - 
         2./27.*m11*mp2*RNep*RSp451e*RSe451p - 2./27.*m11*me2*RN1e*
         RSe461p*RS145pe + 2./27.*m11*me2*RN1e*RSe461p*RS146pe + 2./27.
         *m11*me2*RN1e*RSe451p*RS145pe - 2./27.*m11*me2*RN1e*RSe451p*
         RS146pe + 2./27.*m11*m66*RNep*RSp461e*RSe451p + 2./27.*m11*m66
         *RNep*RSp451e*RSe461p + 2./27.*m11*m66*RNep*RSp451e*RSe451p - 
         2./27.*m11*m66*RN1e*RSe461p*RS145pe - 2./27.*m11*m66*RN1e*
         RSe451p*RS145pe - 2./27.*m11*m66*RN1e*RSe451p*RS146pe + 2./27.
         *m11*m55*RNep*RSp461e*RSe461p + 2./27.*m11*m55*RNep*RSp461e*
         RSe451p + 2./27.*m11*m55*RNep*RSp451e*RSe461p - 2./27.*m11*m55
         *RN1e*RSe461p*RS145pe );

      M12p6x13R +=  + Ab(m56,mu2)*Ab(me2,mu2) * (  - 2./27.*m11*m55*
         RN1e*RSe461p*RS146pe - 2./27.*m11*m55*RN1e*RSe451p*RS146pe );

      M12p6x13R +=  + Ab(m56,mu2)*Ab(mp2,mu2) * (  - 1./27.*m11*pow(
         RSp461e,2) + 2./27.*m11*RSp451e*RSp461e - 1./27.*m11*pow(
         RSp451e,2) + 2./27.*m11*mp2*RNep*RSp461e*RSe461p - 2./27.*m11*
         mp2*RNep*RSp461e*RSe451p - 2./27.*m11*mp2*RNep*RSp451e*RSe461p
          + 2./27.*m11*mp2*RNep*RSp451e*RSe451p - 2./27.*m11*mp2*RN1p*
         RSp461e*RS145pe + 2./27.*m11*mp2*RN1p*RSp461e*RS146pe + 2./27.
         *m11*mp2*RN1p*RSp451e*RS145pe - 2./27.*m11*mp2*RN1p*RSp451e*
         RS146pe - 2./27.*m11*m66*RNep*RSp461e*RSe451p - 2./27.*m11*m66
         *RNep*RSp451e*RSe461p - 2./27.*m11*m66*RNep*RSp451e*RSe451p - 
         2./27.*m11*m66*RN1p*RSp461e*RS145pe - 2./27.*m11*m66*RN1p*
         RSp451e*RS145pe - 2./27.*m11*m66*RN1p*RSp451e*RS146pe - 2./27.
         *m11*m55*RNep*RSp461e*RSe461p - 2./27.*m11*m55*RNep*RSp461e*
         RSe451p - 2./27.*m11*m55*RNep*RSp451e*RSe461p - 2./27.*m11*m55
         *RN1p*RSp461e*RS145pe - 2./27.*m11*m55*RN1p*RSp461e*RS146pe - 
         2./27.*m11*m55*RN1p*RSp451e*RS146pe );

      M12p6x13R +=  + Ab(m56,mu2)*Bb(m11,mu2) * ( 1./27.*m11*m66*pow(
         RS145pe,2) + 2./27.*m11*m66*RS146pe*RS145pe + 2./27.*m11*m55*
         RS146pe*RS145pe + 1./27.*m11*m55*pow(RS146pe,2) - 1./27.*pow(
         m11,2)*pow(RS145pe,2) + 2./27.*pow(m11,2)*RS146pe*RS145pe - 1./
         27.*pow(m11,2)*pow(RS146pe,2) );

      M12p6x13R +=  + Ab(m56,mu2)*Bb(me2,mu2) * (  - 1./27.*m11*me2*
         pow(RSe461p,2) + 2./27.*m11*me2*RSe451p*RSe461p - 1./27.*m11*
         me2*pow(RSe451p,2) + 2./27.*m11*m66*RSe451p*RSe461p + 1./27.*
         m11*m66*pow(RSe451p,2) + 1./27.*m11*m55*pow(RSe461p,2) + 2./27.
         *m11*m55*RSe451p*RSe461p );

      M12p6x13R +=  + Ab(m56,mu2)*Bb(mp2,mu2) * (  - 1./27.*m11*mp2*
         pow(RSp461e,2) + 2./27.*m11*mp2*RSp451e*RSp461e - 1./27.*m11*
         mp2*pow(RSp451e,2) + 2./27.*m11*m66*RSp451e*RSp461e + 1./27.*
         m11*m66*pow(RSp451e,2) + 1./27.*m11*m55*pow(RSp461e,2) + 2./27.
         *m11*m55*RSp451e*RSp461e );

      M12p6x13R +=  + Ab(m66,mu2)*Ab(me2,mu2) * ( 2./9.*m11*m66*RNep*
         RSp451e*RSe451p - 2./27.*m11*m66*RNep*RSp451e*RSe451p*RS645pe
          - 2./9.*m11*m66*RN1e*RSe451p*RS145pe + 2./27.*m11*m66*RN1e*
         RSe451p*RS645pe*RS145pe );

      M12p6x13R +=  + Ab(m66,mu2)*Ab(mp2,mu2) * (  - 2./9.*m11*m66*RNep
         *RSp451e*RSe451p + 2./27.*m11*m66*RNep*RSp451e*RSe451p*RS645pe
          - 2./9.*m11*m66*RN1p*RSp451e*RS145pe + 2./27.*m11*m66*RN1p*
         RSp451e*RS645pe*RS145pe );

      M12p6x13R +=  + Ab(m66,mu2)*Bb(m11,mu2) * ( 1./9.*m11*m66*pow(
         RS145pe,2) - 1./27.*m11*m66*RS645pe*pow(RS145pe,2) );

      M12p6x13R +=  + Ab(m66,mu2)*Bb(me2,mu2) * ( 1./9.*m11*m66*pow(
         RSe451p,2) - 1./27.*m11*m66*pow(RSe451p,2)*RS645pe );

      M12p6x13R +=  + Ab(m66,mu2)*Bb(mp2,mu2) * ( 1./9.*m11*m66*pow(
         RSp451e,2) - 1./27.*m11*m66*pow(RSp451e,2)*RS645pe );

      M12p6x13R +=  + Ab(me2,mu2)*pi16 * ( 1./12.*m11*me2*RSe561p + 1./
         12.*m11*me2*RSe461p + 1./12.*m11*me2*RSe451p );

      M12p6x13R +=  + pow(Ab(me2,mu2),2) * ( 1./18.*m11*pow(
         RSe45611p,2) - 2./27.*m11*m66*RNep*RSe456p*RSp451e*RSe451p + 2.
         /27.*m11*m66*RN1e*RSe456p*RSe451p*RS145pe - 2./27.*m11*m55*
         RNep*RSe465p*RSp461e*RSe461p + 2./27.*m11*m55*RN1e*RSe465p*
         RSe461p*RS146pe - 2./27.*m11*m44*RNep*RSe564p*RSp561e*RSe561p
          + 2./27.*m11*m44*RN1e*RSe564p*RSe561p*RS156pe - 2./9.*pow(
         m11,2)*RN1e*pow(RSe45611p,2) );

      M12p6x13R +=  + Ab(me2,mu2)*Ab(mp2,mu2) * ( 1./9.*m11*RSp45611e*
         RSe45611p + 2./27.*m11*m66*RNep*RSe456p*RSp451e*RSe451p - 2./
         27.*m11*m66*RNep*RSp456e*RSp451e*RSe451p + 2./27.*m11*m66*RN1p
         *RSe456p*RSp451e*RS145pe + 2./27.*m11*m66*RN1e*RSp456e*RSe451p
         *RS145pe + 2./27.*m11*m55*RNep*RSe465p*RSp461e*RSe461p - 2./27.
         *m11*m55*RNep*RSp465e*RSp461e*RSe461p + 2./27.*m11*m55*RN1p*
         RSe465p*RSp461e*RS146pe + 2./27.*m11*m55*RN1e*RSp465e*RSe461p*
         RS146pe + 2./27.*m11*m44*RNep*RSe564p*RSp561e*RSe561p - 2./27.
         *m11*m44*RNep*RSp564e*RSp561e*RSe561p + 2./27.*m11*m44*RN1p*
         RSe564p*RSp561e*RS156pe + 2./27.*m11*m44*RN1e*RSp564e*RSe561p*
         RS156pe - 2./9.*pow(m11,2)*RN1p*RSp45611e*RSe45611p - 2./9.*
         pow(m11,2)*RN1e*RSp45611e*RSe45611p );

      M12p6x13R +=  + Ab(me2,mu2)*Bb(m11,mu2) * ( 1./9.*m11*RSe45611p*
         RD1456ep - 1./27.*m11*m66*RSe456p*pow(RS145pe,2) - 1./27.*m11*
         m55*RSe465p*pow(RS146pe,2) - 1./27.*m11*m44*RSe564p*pow(
         RS156pe,2) + 2./9.*pow(m11,2)*RSe45611p*RC1456pe - 2./9.*pow(
         m11,2)*RN1e*RSe45611p*RD1456ep );

      M12p6x13R +=  + Ab(me2,mu2)*Bb(me2,mu2) * (  - 1./27.*m11*m66*
         RSe456p*pow(RSe451p,2) - 1./27.*m11*m55*RSe465p*pow(RSe461p,2)
          - 1./27.*m11*m44*RSe564p*pow(RSe561p,2) );

      M12p6x13R +=  + Ab(me2,mu2)*Bb(mp2,mu2) * (  - 1./27.*m11*m66*
         RSe456p*pow(RSp451e,2) - 1./27.*m11*m55*RSe465p*pow(RSp461e,2)
          - 1./27.*m11*m44*RSe564p*pow(RSp561e,2) );

      M12p6x13R +=  + Ab(me2,mu2)*Cb(m11,mu2) * ( 2./9.*pow(m11,2)*
         RSe45611p*RD1456ep );

      M12p6x13R +=  + Ab(mp2,mu2)*pi16 * ( 1./12.*m11*mp2*RSp561e + 1./
         12.*m11*mp2*RSp461e + 1./12.*m11*mp2*RSp451e );

      M12p6x13R +=  + pow(Ab(mp2,mu2),2) * ( 1./18.*m11*pow(
         RSp45611e,2) + 2./27.*m11*m66*RNep*RSp456e*RSp451e*RSe451p + 2.
         /27.*m11*m66*RN1p*RSp456e*RSp451e*RS145pe + 2./27.*m11*m55*
         RNep*RSp465e*RSp461e*RSe461p + 2./27.*m11*m55*RN1p*RSp465e*
         RSp461e*RS146pe + 2./27.*m11*m44*RNep*RSp564e*RSp561e*RSe561p
          + 2./27.*m11*m44*RN1p*RSp564e*RSp561e*RS156pe - 2./9.*pow(
         m11,2)*RN1p*pow(RSp45611e,2) );

      M12p6x13R +=  + Ab(mp2,mu2)*Bb(m11,mu2) * ( 1./9.*m11*RSp45611e*
         RD1456ep - 1./27.*m11*m66*RSp456e*pow(RS145pe,2) - 1./27.*m11*
         m55*RSp465e*pow(RS146pe,2) - 1./27.*m11*m44*RSp564e*pow(
         RS156pe,2) + 2./9.*pow(m11,2)*RSp45611e*RC1456pe - 2./9.*pow(
         m11,2)*RN1p*RSp45611e*RD1456ep );

      M12p6x13R +=  + Ab(mp2,mu2)*Bb(me2,mu2) * (  - 1./27.*m11*m66*
         RSp456e*pow(RSe451p,2) - 1./27.*m11*m55*RSp465e*pow(RSe461p,2)
          - 1./27.*m11*m44*RSp564e*pow(RSe561p,2) );

      M12p6x13R +=  + Ab(mp2,mu2)*Bb(mp2,mu2) * (  - 1./27.*m11*m66*
         RSp456e*pow(RSp451e,2) - 1./27.*m11*m55*RSp465e*pow(RSp461e,2)
          - 1./27.*m11*m44*RSp564e*pow(RSp561e,2) );

      M12p6x13R +=  + Ab(mp2,mu2)*Cb(m11,mu2) * ( 2./9.*pow(m11,2)*
         RSp45611e*RD1456ep );

      M12p6x13R +=  + pow(Bb(m11,mu2),2) * ( 1./18.*m11*pow(RD1456ep,2)
          + 2./9.*pow(m11,2)*RD1456ep*RC1456pe );

      M12p6x13R +=  + Bb(m11,mu2)*Cb(m11,mu2) * ( 2./9.*pow(m11,2)*pow(
         RD1456ep,2) );

      M12p6x13R +=  + Cbeps(m11,mu2)*pi16 * ( 2./9.*pow(
         m11,2)*pow(RD1456ep,2) + 2./9.*pow(m11,2)*mp2*RSp45611e*
         RD1456ep + 2./9.*pow(m11,2)*me2*RSe45611p*RD1456ep + 2./27.*
         pow(m11,2)*m66*RD1456ep + 2./27.*pow(m11,2)*m55*RD1456ep + 2./
         27.*pow(m11,2)*m44*RD1456ep - 4./9.*pow(m11,3)*RD1456ep - 2./9.
         *pow(m11,3)*RS1456pee*RD1456ep - 2./9.*pow(m11,3)*RS1456ppe*
         RD1456ep + 2./9.*pow(m11,3)*RS145pe*RD1456ep + 2./9.*pow(
         m11,3)*RS146pe*RD1456ep + 2./9.*pow(m11,3)*RS156pe*RD1456ep );

      M12p6x13R +=  + Abeps(m11,mu2)*pi16 * ( 1./45.*m66*RD1456ep + 1./
         45.*pow(m66,2)*RS145pe + 1./45.*m55*RD1456ep + 1./45.*pow(
         m55,2)*RS146pe + 1./45.*m44*RD1456ep + 1./45.*pow(m44,2)*
         RS156pe + 19./60.*m11*RD1456ep + 1./9.*m11*RD1456ep*RC1456pe
          + 1./9.*m11*mp2*RSp45611e*RC1456pe + 1./9.*m11*me2*RSe45611p*
         RC1456pe - 2./27.*m11*m66*RC1456pe - 2./9.*m11*m66*RSe45611p
          - 2./9.*m11*m66*RSp45611e - 2./9.*m11*m66*RS145pe + 5./27.*
         m11*m66*pow(RS145pe,2) + 1./27.*m11*m66*RS146pe*RS145pe + 1./
         27.*m11*m66*pow(RS146pe,2) + 1./27.*m11*m66*RS156pe*RS145pe + 
         1./27.*m11*m66*pow(RS156pe,2) + 1./27.*m11*m66*RSe561p*RS145pe
          + 2./27.*m11*m66*RSe561p*RS156pe + 1./27.*m11*m66*RSe461p*
         RS145pe + 2./27.*m11*m66*RSe461p*RS146pe - 1./9.*m11*m66*
         RSe451p + 10./27.*m11*m66*RSe451p*RS145pe + 1./27.*m11*m66*
         RSe451p*RS146pe + 1./27.*m11*m66*RSe451p*RS156pe + 1./27.*m11*
         m66*RSp561e*RS145pe + 2./27.*m11*m66*RSp561e*RS156pe + 1./27.*
         m11*m66*RSp461e*RS145pe );

      M12p6x13R +=  + Abeps(m11,mu2)*pi16 * ( 2./27.*m11*m66*RSp461e*
         RS146pe - 1./9.*m11*m66*RSp451e + 10./27.*m11*m66*RSp451e*
         RS145pe + 1./27.*m11*m66*RSp451e*RS146pe + 1./27.*m11*m66*
         RSp451e*RS156pe - 2./9.*m11*m66*mp2*RN1p*RSp45611e + 1./27.*
         m11*m66*mp2*RN1p*RSp561e*RS145pe + 2./27.*m11*m66*mp2*RN1p*
         RSp561e*RS156pe + 1./27.*m11*m66*mp2*RN1p*RSp461e*RS145pe + 2./
         27.*m11*m66*mp2*RN1p*RSp461e*RS146pe - 1./9.*m11*m66*mp2*RN1p*
         RSp451e + 10./27.*m11*m66*mp2*RN1p*RSp451e*RS145pe + 1./27.*
         m11*m66*mp2*RN1p*RSp451e*RS146pe + 1./27.*m11*m66*mp2*RN1p*
         RSp451e*RS156pe - 2./27.*m11*m66*mp2*RN1p*RSp456e*RSp451e*
         RS145pe - 2./27.*m11*m66*mp2*RN1e*RSp456e*RSe451p*RS145pe - 2./
         27.*m11*m66*me2*RN1p*RSe456p*RSp451e*RS145pe - 2./9.*m11*m66*
         me2*RN1e*RSe45611p + 1./27.*m11*m66*me2*RN1e*RSe561p*RS145pe
          + 2./27.*m11*m66*me2*RN1e*RSe561p*RS156pe + 1./27.*m11*m66*
         me2*RN1e*RSe461p*RS145pe + 2./27.*m11*m66*me2*RN1e*RSe461p*
         RS146pe );

      M12p6x13R +=  + Abeps(m11,mu2)*pi16 * (  - 1./9.*m11*m66*me2*RN1e
         *RSe451p + 10./27.*m11*m66*me2*RN1e*RSe451p*RS145pe + 1./27.*
         m11*m66*me2*RN1e*RSe451p*RS146pe + 1./27.*m11*m66*me2*RN1e*
         RSe451p*RS156pe - 2./27.*m11*m66*me2*RN1e*RSe456p*RSe451p*
         RS145pe + 1./27.*m11*pow(m66,2)*RN1p*RSp561e*RS145pe + 1./27.*
         m11*pow(m66,2)*RN1p*RSp461e*RS145pe - 1./9.*m11*pow(m66,2)*
         RN1p*RSp451e - 2./81.*m11*pow(m66,2)*RN1p*RSp451e*RS145pe + 1./
         27.*m11*pow(m66,2)*RN1p*RSp451e*RS146pe + 1./27.*m11*pow(
         m66,2)*RN1p*RSp451e*RS156pe - 2./27.*m11*pow(m66,2)*RN1p*
         RSp451e*RS645pe*RS145pe + 1./27.*m11*pow(m66,2)*RN1e*RSe561p*
         RS145pe + 1./27.*m11*pow(m66,2)*RN1e*RSe461p*RS145pe - 1./9.*
         m11*pow(m66,2)*RN1e*RSe451p - 2./81.*m11*pow(m66,2)*RN1e*
         RSe451p*RS145pe + 1./27.*m11*pow(m66,2)*RN1e*RSe451p*RS146pe
          + 1./27.*m11*pow(m66,2)*RN1e*RSe451p*RS156pe - 2./27.*m11*
         pow(m66,2)*RN1e*RSe451p*RS645pe*RS145pe - 2./27.*m11*m55*
         RC1456pe );

      M12p6x13R +=  + Abeps(m11,mu2)*pi16 * (  - 2./9.*m11*m55*
         RSe45611p - 2./9.*m11*m55*RSp45611e + 1./27.*m11*m55*pow(
         RS145pe,2) - 2./9.*m11*m55*RS146pe + 1./27.*m11*m55*RS146pe*
         RS145pe + 5./27.*m11*m55*pow(RS146pe,2) + 1./27.*m11*m55*
         RS156pe*RS146pe + 1./27.*m11*m55*pow(RS156pe,2) + 1./27.*m11*
         m55*RSe561p*RS146pe + 2./27.*m11*m55*RSe561p*RS156pe - 1./9.*
         m11*m55*RSe461p + 1./27.*m11*m55*RSe461p*RS145pe + 10./27.*m11
         *m55*RSe461p*RS146pe + 1./27.*m11*m55*RSe461p*RS156pe + 2./27.
         *m11*m55*RSe451p*RS145pe + 1./27.*m11*m55*RSe451p*RS146pe + 1./
         27.*m11*m55*RSp561e*RS146pe + 2./27.*m11*m55*RSp561e*RS156pe
          - 1./9.*m11*m55*RSp461e + 1./27.*m11*m55*RSp461e*RS145pe + 10.
         /27.*m11*m55*RSp461e*RS146pe + 1./27.*m11*m55*RSp461e*RS156pe
          + 2./27.*m11*m55*RSp451e*RS145pe + 1./27.*m11*m55*RSp451e*
         RS146pe - 2./9.*m11*m55*mp2*RN1p*RSp45611e + 1./27.*m11*m55*
         mp2*RN1p*RSp561e*RS146pe + 2./27.*m11*m55*mp2*RN1p*RSp561e*
         RS156pe );

      M12p6x13R +=  + Abeps(m11,mu2)*pi16 * (  - 1./9.*m11*m55*mp2*RN1p
         *RSp461e + 1./27.*m11*m55*mp2*RN1p*RSp461e*RS145pe + 10./27.*
         m11*m55*mp2*RN1p*RSp461e*RS146pe + 1./27.*m11*m55*mp2*RN1p*
         RSp461e*RS156pe + 2./27.*m11*m55*mp2*RN1p*RSp451e*RS145pe + 1./
         27.*m11*m55*mp2*RN1p*RSp451e*RS146pe - 2./27.*m11*m55*mp2*RN1p
         *RSp465e*RSp461e*RS146pe - 2./27.*m11*m55*mp2*RN1e*RSp465e*
         RSe461p*RS146pe - 2./27.*m11*m55*me2*RN1p*RSe465p*RSp461e*
         RS146pe - 2./9.*m11*m55*me2*RN1e*RSe45611p + 1./27.*m11*m55*
         me2*RN1e*RSe561p*RS146pe + 2./27.*m11*m55*me2*RN1e*RSe561p*
         RS156pe - 1./9.*m11*m55*me2*RN1e*RSe461p + 1./27.*m11*m55*me2*
         RN1e*RSe461p*RS145pe + 10./27.*m11*m55*me2*RN1e*RSe461p*
         RS146pe + 1./27.*m11*m55*me2*RN1e*RSe461p*RS156pe + 2./27.*m11
         *m55*me2*RN1e*RSe451p*RS145pe + 1./27.*m11*m55*me2*RN1e*
         RSe451p*RS146pe - 2./27.*m11*m55*me2*RN1e*RSe465p*RSe461p*
         RS146pe + 2./27.*m11*m55*m66*RN1p*RSp461e*RS145pe - 8./81.*m11
         *m55*m66*RN1p*RSp461e*RS146pe );

      M12p6x13R +=  + Abeps(m11,mu2)*pi16 * (  - 8./81.*m11*m55*m66*
         RN1p*RSp451e*RS145pe + 2./27.*m11*m55*m66*RN1p*RSp451e*RS146pe
          + 2./27.*m11*m55*m66*RN1e*RSe461p*RS145pe - 8./81.*m11*m55*
         m66*RN1e*RSe461p*RS146pe - 8./81.*m11*m55*m66*RN1e*RSe451p*
         RS145pe + 2./27.*m11*m55*m66*RN1e*RSe451p*RS146pe + 1./27.*m11
         *pow(m55,2)*RN1p*RSp561e*RS146pe - 1./9.*m11*pow(m55,2)*RN1p*
         RSp461e + 1./27.*m11*pow(m55,2)*RN1p*RSp461e*RS145pe - 2./81.*
         m11*pow(m55,2)*RN1p*RSp461e*RS146pe + 1./27.*m11*pow(m55,2)*
         RN1p*RSp461e*RS156pe - 2./27.*m11*pow(m55,2)*RN1p*RSp461e*
         RS546pe*RS146pe + 1./27.*m11*pow(m55,2)*RN1p*RSp451e*RS146pe
          + 1./27.*m11*pow(m55,2)*RN1e*RSe561p*RS146pe - 1./9.*m11*pow(
         m55,2)*RN1e*RSe461p + 1./27.*m11*pow(m55,2)*RN1e*RSe461p*
         RS145pe - 2./81.*m11*pow(m55,2)*RN1e*RSe461p*RS146pe + 1./27.*
         m11*pow(m55,2)*RN1e*RSe461p*RS156pe - 2./27.*m11*pow(m55,2)*
         RN1e*RSe461p*RS546pe*RS146pe + 1./27.*m11*pow(m55,2)*RN1e*
         RSe451p*RS146pe );

      M12p6x13R +=  + Abeps(m11,mu2)*pi16 * (  - 2./27.*m11*m44*
         RC1456pe - 2./9.*m11*m44*RSe45611p - 2./9.*m11*m44*RSp45611e
          + 1./27.*m11*m44*pow(RS145pe,2) + 1./27.*m11*m44*pow(
         RS146pe,2) - 2./9.*m11*m44*RS156pe + 1./27.*m11*m44*RS156pe*
         RS145pe + 1./27.*m11*m44*RS156pe*RS146pe + 5./27.*m11*m44*pow(
         RS156pe,2) - 1./9.*m11*m44*RSe561p + 1./27.*m11*m44*RSe561p*
         RS145pe + 1./27.*m11*m44*RSe561p*RS146pe + 10./27.*m11*m44*
         RSe561p*RS156pe + 2./27.*m11*m44*RSe461p*RS146pe + 1./27.*m11*
         m44*RSe461p*RS156pe + 2./27.*m11*m44*RSe451p*RS145pe + 1./27.*
         m11*m44*RSe451p*RS156pe - 1./9.*m11*m44*RSp561e + 1./27.*m11*
         m44*RSp561e*RS145pe + 1./27.*m11*m44*RSp561e*RS146pe + 10./27.
         *m11*m44*RSp561e*RS156pe + 2./27.*m11*m44*RSp461e*RS146pe + 1./
         27.*m11*m44*RSp461e*RS156pe + 2./27.*m11*m44*RSp451e*RS145pe
          + 1./27.*m11*m44*RSp451e*RS156pe - 2./9.*m11*m44*mp2*RN1p*
         RSp45611e - 1./9.*m11*m44*mp2*RN1p*RSp561e + 1./27.*m11*m44*
         mp2*RN1p*RSp561e*RS145pe );

      M12p6x13R +=  + Abeps(m11,mu2)*pi16 * ( 1./27.*m11*m44*mp2*RN1p*
         RSp561e*RS146pe + 10./27.*m11*m44*mp2*RN1p*RSp561e*RS156pe + 2.
         /27.*m11*m44*mp2*RN1p*RSp461e*RS146pe + 1./27.*m11*m44*mp2*
         RN1p*RSp461e*RS156pe + 2./27.*m11*m44*mp2*RN1p*RSp451e*RS145pe
          + 1./27.*m11*m44*mp2*RN1p*RSp451e*RS156pe - 2./27.*m11*m44*
         mp2*RN1p*RSp564e*RSp561e*RS156pe - 2./27.*m11*m44*mp2*RN1e*
         RSp564e*RSe561p*RS156pe - 2./27.*m11*m44*me2*RN1p*RSe564p*
         RSp561e*RS156pe - 2./9.*m11*m44*me2*RN1e*RSe45611p - 1./9.*m11
         *m44*me2*RN1e*RSe561p + 1./27.*m11*m44*me2*RN1e*RSe561p*
         RS145pe + 1./27.*m11*m44*me2*RN1e*RSe561p*RS146pe + 10./27.*
         m11*m44*me2*RN1e*RSe561p*RS156pe + 2./27.*m11*m44*me2*RN1e*
         RSe461p*RS146pe + 1./27.*m11*m44*me2*RN1e*RSe461p*RS156pe + 2./
         27.*m11*m44*me2*RN1e*RSe451p*RS145pe + 1./27.*m11*m44*me2*RN1e
         *RSe451p*RS156pe - 2./27.*m11*m44*me2*RN1e*RSe564p*RSe561p*
         RS156pe + 2./27.*m11*m44*m66*RN1p*RSp561e*RS145pe - 8./81.*m11
         *m44*m66*RN1p*RSp561e*RS156pe );

      M12p6x13R +=  + Abeps(m11,mu2)*pi16 * (  - 8./81.*m11*m44*m66*
         RN1p*RSp451e*RS145pe + 2./27.*m11*m44*m66*RN1p*RSp451e*RS156pe
          + 2./27.*m11*m44*m66*RN1e*RSe561p*RS145pe - 8./81.*m11*m44*
         m66*RN1e*RSe561p*RS156pe - 8./81.*m11*m44*m66*RN1e*RSe451p*
         RS145pe + 2./27.*m11*m44*m66*RN1e*RSe451p*RS156pe + 2./27.*m11
         *m44*m55*RN1p*RSp561e*RS146pe - 8./81.*m11*m44*m55*RN1p*
         RSp561e*RS156pe - 8./81.*m11*m44*m55*RN1p*RSp461e*RS146pe + 2./
         27.*m11*m44*m55*RN1p*RSp461e*RS156pe + 2./27.*m11*m44*m55*RN1e
         *RSe561p*RS146pe - 8./81.*m11*m44*m55*RN1e*RSe561p*RS156pe - 8.
         /81.*m11*m44*m55*RN1e*RSe461p*RS146pe + 2./27.*m11*m44*m55*
         RN1e*RSe461p*RS156pe - 1./9.*m11*pow(m44,2)*RN1p*RSp561e + 1./
         27.*m11*pow(m44,2)*RN1p*RSp561e*RS145pe + 1./27.*m11*pow(
         m44,2)*RN1p*RSp561e*RS146pe - 2./81.*m11*pow(m44,2)*RN1p*
         RSp561e*RS156pe - 2./27.*m11*pow(m44,2)*RN1p*RSp561e*RS456pe*
         RS156pe + 1./27.*m11*pow(m44,2)*RN1p*RSp461e*RS156pe + 1./27.*
         m11*pow(m44,2)*RN1p*RSp451e*RS156pe );

      M12p6x13R +=  + Abeps(m11,mu2)*pi16 * (  - 1./9.*m11*pow(m44,2)*
         RN1e*RSe561p + 1./27.*m11*pow(m44,2)*RN1e*RSe561p*RS145pe + 1./
         27.*m11*pow(m44,2)*RN1e*RSe561p*RS146pe - 2./81.*m11*pow(
         m44,2)*RN1e*RSe561p*RS156pe - 2./27.*m11*pow(m44,2)*RN1e*
         RSe561p*RS456pe*RS156pe + 1./27.*m11*pow(m44,2)*RN1e*RSe461p*
         RS156pe + 1./27.*m11*pow(m44,2)*RN1e*RSe451p*RS156pe - 13./18.
         *pow(m11,2) + 1./36.*pow(m11,2)*RC1456pe + 1./3.*pow(m11,2)*
         pow(RC1456pe,2) - 2./3.*pow(m11,2)*RSe45611p + 2./9.*pow(
         m11,2)*RSe45611p*RC1456pe - 2./3.*pow(m11,2)*RSp45611e + 2./9.
         *pow(m11,2)*RSp45611e*RC1456pe - 11./45.*pow(m11,2)*RS145pe - 
         11./45.*pow(m11,2)*RS146pe - 11./45.*pow(m11,2)*RS156pe - 1./9.
         *pow(m11,2)*RSe561p - 1./9.*pow(m11,2)*RSe461p - 1./9.*pow(
         m11,2)*RSe451p - 1./9.*pow(m11,2)*RSp561e - 1./9.*pow(m11,2)*
         RSp461e - 1./9.*pow(m11,2)*RSp451e + 2./9.*pow(m11,2)*RN1p*
         RSp45611e*RD1456ep + 2./9.*pow(m11,2)*RN1e*RSe45611p*RD1456ep
          - 2./3.*pow(m11,2)*mp2*RN1p*RSp45611e );

      M12p6x13R +=  + Abeps(m11,mu2)*pi16 * ( 2./9.*pow(m11,2)*mp2*RN1p
         *pow(RSp45611e,2) - 1./9.*pow(m11,2)*mp2*RN1p*RSp561e - 1./9.*
         pow(m11,2)*mp2*RN1p*RSp461e - 1./9.*pow(m11,2)*mp2*RN1p*
         RSp451e + 2./9.*pow(m11,2)*mp2*RN1e*RSp45611e*RSe45611p + 2./9.
         *pow(m11,2)*me2*RN1p*RSp45611e*RSe45611p - 2./3.*pow(m11,2)*
         me2*RN1e*RSe45611p + 2./9.*pow(m11,2)*me2*RN1e*pow(
         RSe45611p,2) - 1./9.*pow(m11,2)*me2*RN1e*RSe561p - 1./9.*pow(
         m11,2)*me2*RN1e*RSe461p - 1./9.*pow(m11,2)*me2*RN1e*RSe451p + 
         8./27.*pow(m11,2)*m66*RN1p*RSp45611e - 2./9.*pow(m11,2)*m66*
         RN1p*RSp451e + 8./27.*pow(m11,2)*m66*RN1e*RSe45611p - 2./9.*
         pow(m11,2)*m66*RN1e*RSe451p + 8./27.*pow(m11,2)*m55*RN1p*
         RSp45611e - 2./9.*pow(m11,2)*m55*RN1p*RSp461e + 8./27.*pow(
         m11,2)*m55*RN1e*RSe45611p - 2./9.*pow(m11,2)*m55*RN1e*RSe461p
          + 8./27.*pow(m11,2)*m44*RN1p*RSp45611e - 2./9.*pow(m11,2)*m44
         *RN1p*RSp561e + 8./27.*pow(m11,2)*m44*RN1e*RSe45611p - 2./9.*
         pow(m11,2)*m44*RN1e*RSe561p );

      M12p6x13R +=  + Abeps(m11,mu2)*pi16 * ( 2./9.*pow(m11,3)*RN1p*
         RSp45611e - 2./9.*pow(m11,3)*RN1p*RSp45611e*RS1456pee - 2./9.*
         pow(m11,3)*RN1p*RSp45611e*RS1456ppe + 2./9.*pow(m11,3)*RN1p*
         RS145pe*RSp45611e + 2./9.*pow(m11,3)*RN1p*RS146pe*RSp45611e + 
         2./9.*pow(m11,3)*RN1p*RS156pe*RSp45611e - 1./9.*pow(m11,3)*
         RN1p*RSp561e - 1./9.*pow(m11,3)*RN1p*RSp461e - 1./9.*pow(
         m11,3)*RN1p*RSp451e + 2./9.*pow(m11,3)*RN1e*RSe45611p - 2./9.*
         pow(m11,3)*RN1e*RSe45611p*RS1456pee - 2./9.*pow(m11,3)*RN1e*
         RSe45611p*RS1456ppe + 2./9.*pow(m11,3)*RN1e*RS145pe*RSe45611p
          + 2./9.*pow(m11,3)*RN1e*RS146pe*RSe45611p + 2./9.*pow(m11,3)*
         RN1e*RS156pe*RSe45611p - 1./9.*pow(m11,3)*RN1e*RSe561p - 1./9.
         *pow(m11,3)*RN1e*RSe461p - 1./9.*pow(m11,3)*RN1e*RSe451p );

      M12p6x13R +=  + Abeps(m14,mu2)*pi16 * ( 2./45.*pow(mp2,2)*
         RSp45611e - 4./45.*pow(mp2,2)*RSp561e + 2./45.*pow(mp2,2)*
         RSp564e + 2./45.*pow(me2,2)*RSe45611p - 4./45.*pow(me2,2)*
         RSe561p + 2./45.*pow(me2,2)*RSe564p - 1./27.*m44*RD1456ep - 1./
         27.*m44*mp2*RSp45611e + 16./135.*m44*mp2*RSp561e - 11./135.*
         m44*mp2*RSp564e - 1./27.*m44*me2*RSe45611p + 16./135.*m44*me2*
         RSe561p - 11./135.*m44*me2*RSe564p - 1./27.*pow(m44,2)*
         RC1456pe - 1./27.*pow(m44,2)*RSe45611p - 1./27.*pow(m44,2)*
         RSp45611e + 1./27.*pow(m44,2)*RSe564p + 1./27.*pow(m44,2)*
         RSp564e + 46./135.*m11*RD1456ep + 34./135.*m11*mp2*RSp45611e
          - 44./135.*m11*mp2*RSp561e - 1./27.*m11*mp2*RSp564e + 34./135.
         *m11*me2*RSe45611p - 44./135.*m11*me2*RSe561p - 1./27.*m11*me2
         *RSe564p - 31./108.*m11*m66 - 31./108.*m11*m55 - 10./27.*m11*
         m44 - 17./108.*m11*m44*RC1456pe - 13./108.*m11*m44*RSe45611p
          - 13./108.*m11*m44*RSp45611e - 23./270.*m11*m44*RS156pe + 7./
         108.*m11*m44*RS456pe );

      M12p6x13R +=  + Abeps(m14,mu2)*pi16 * (  - 11./54.*m11*m44*
         RSe561p - 11./54.*m11*m44*RSp561e + 11./108.*m11*m44*RSe564p
          + 11./108.*m11*m44*RSp564e - 11./18.*pow(m11,2) + 23./108.*
         pow(m11,2)*RC1456pe - 1./12.*pow(m11,2)*RSe45611p - 1./12.*
         pow(m11,2)*RSp45611e - 227./270.*pow(m11,2)*RS156pe + 7./108.*
         pow(m11,2)*RS456pe - 23./54.*pow(m11,2)*RSe561p - 23./54.*pow(
         m11,2)*RSp561e + 7./108.*pow(m11,2)*RSe564p + 7./108.*pow(
         m11,2)*RSp564e );

      M12p6x13R +=  + Abeps(m15,mu2)*pi16 * ( 2./45.*pow(mp2,2)*
         RSp45611e - 4./45.*pow(mp2,2)*RSp461e + 2./45.*pow(mp2,2)*
         RSp465e + 2./45.*pow(me2,2)*RSe45611p - 4./45.*pow(me2,2)*
         RSe461p + 2./45.*pow(me2,2)*RSe465p - 1./27.*m55*RD1456ep - 1./
         27.*m55*mp2*RSp45611e + 16./135.*m55*mp2*RSp461e - 11./135.*
         m55*mp2*RSp465e - 1./27.*m55*me2*RSe45611p + 16./135.*m55*me2*
         RSe461p - 11./135.*m55*me2*RSe465p - 1./27.*pow(m55,2)*
         RC1456pe - 1./27.*pow(m55,2)*RSe45611p - 1./27.*pow(m55,2)*
         RSp45611e + 1./27.*pow(m55,2)*RSe465p + 1./27.*pow(m55,2)*
         RSp465e + 46./135.*m11*RD1456ep + 34./135.*m11*mp2*RSp45611e
          - 44./135.*m11*mp2*RSp461e - 1./27.*m11*mp2*RSp465e + 34./135.
         *m11*me2*RSe45611p - 44./135.*m11*me2*RSe461p - 1./27.*m11*me2
         *RSe465p - 31./108.*m11*m66 - 10./27.*m11*m55 - 17./108.*m11*
         m55*RC1456pe - 13./108.*m11*m55*RSe45611p - 13./108.*m11*m55*
         RSp45611e - 23./270.*m11*m55*RS146pe + 7./108.*m11*m55*RS546pe
          - 11./54.*m11*m55*RSe461p );

      M12p6x13R +=  + Abeps(m15,mu2)*pi16 * (  - 11./54.*m11*m55*
         RSp461e + 11./108.*m11*m55*RSe465p + 11./108.*m11*m55*RSp465e
          - 31./108.*m11*m44 - 11./18.*pow(m11,2) + 23./108.*pow(m11,2)
         *RC1456pe - 1./12.*pow(m11,2)*RSe45611p - 1./12.*pow(m11,2)*
         RSp45611e - 227./270.*pow(m11,2)*RS146pe + 7./108.*pow(m11,2)*
         RS546pe - 23./54.*pow(m11,2)*RSe461p - 23./54.*pow(m11,2)*
         RSp461e + 7./108.*pow(m11,2)*RSe465p + 7./108.*pow(m11,2)*
         RSp465e );

      M12p6x13R +=  + Abeps(m16,mu2)*pi16 * ( 2./45.*pow(mp2,2)*
         RSp45611e - 4./45.*pow(mp2,2)*RSp451e + 2./45.*pow(mp2,2)*
         RSp456e + 2./45.*pow(me2,2)*RSe45611p - 4./45.*pow(me2,2)*
         RSe451p + 2./45.*pow(me2,2)*RSe456p - 1./27.*m66*RD1456ep - 1./
         27.*m66*mp2*RSp45611e + 16./135.*m66*mp2*RSp451e - 11./135.*
         m66*mp2*RSp456e - 1./27.*m66*me2*RSe45611p + 16./135.*m66*me2*
         RSe451p - 11./135.*m66*me2*RSe456p - 1./27.*pow(m66,2)*
         RC1456pe - 1./27.*pow(m66,2)*RSe45611p - 1./27.*pow(m66,2)*
         RSp45611e + 1./27.*pow(m66,2)*RSe456p + 1./27.*pow(m66,2)*
         RSp456e + 46./135.*m11*RD1456ep + 34./135.*m11*mp2*RSp45611e
          - 44./135.*m11*mp2*RSp451e - 1./27.*m11*mp2*RSp456e + 34./135.
         *m11*me2*RSe45611p - 44./135.*m11*me2*RSe451p - 1./27.*m11*me2
         *RSe456p - 10./27.*m11*m66 - 17./108.*m11*m66*RC1456pe - 13./
         108.*m11*m66*RSe45611p - 13./108.*m11*m66*RSp45611e - 23./270.
         *m11*m66*RS145pe + 7./108.*m11*m66*RS645pe - 11./54.*m11*m66*
         RSe451p );

      M12p6x13R +=  + Abeps(m16,mu2)*pi16 * (  - 11./54.*m11*m66*
         RSp451e + 11./108.*m11*m66*RSe456p + 11./108.*m11*m66*RSp456e
          - 31./108.*m11*m55 - 31./108.*m11*m44 - 11./18.*pow(m11,2) + 
         23./108.*pow(m11,2)*RC1456pe - 1./12.*pow(m11,2)*RSe45611p - 1.
         /12.*pow(m11,2)*RSp45611e - 227./270.*pow(m11,2)*RS145pe + 7./
         108.*pow(m11,2)*RS645pe - 23./54.*pow(m11,2)*RSe451p - 23./54.
         *pow(m11,2)*RSp451e + 7./108.*pow(m11,2)*RSe456p + 7./108.*
         pow(m11,2)*RSp456e );

      M12p6x13R +=  + Abeps(m44,mu2)*pi16 * (  - 13./36.*m11*m44 + 1./9.
         *m11*m44*pow(RS156pe,2) + 13./108.*m11*m44*RS456pe - 1./27.*
         m11*m44*RS456pe*pow(RS156pe,2) + 2./9.*m11*m44*RSe561p*RS156pe
          - 2./27.*m11*m44*RSe561p*RS456pe*RS156pe + 1./9.*m11*m44*pow(
         RSe561p,2) - 1./27.*m11*m44*pow(RSe561p,2)*RS456pe + 2./9.*m11
         *m44*RSp561e*RS156pe - 2./27.*m11*m44*RSp561e*RS456pe*RS156pe
          + 2./9.*m11*m44*RSp561e*RSe561p - 2./27.*m11*m44*RSp561e*
         RSe561p*RS456pe + 1./9.*m11*m44*pow(RSp561e,2) - 1./27.*m11*
         m44*pow(RSp561e,2)*RS456pe );

      M12p6x13R +=  + Abeps(m45,mu2)*pi16 * ( 2./27.*m11*mp2*RSp561e*
         RS146pe - 2./27.*m11*mp2*RSp561e*RS156pe - 2./27.*m11*mp2*
         RSp561e*RSe561p + 2./27.*m11*mp2*RSp561e*RSe461p - 2./27.*m11*
         mp2*pow(RSp561e,2) - 2./27.*m11*mp2*RSp461e*RS146pe + 2./27.*
         m11*mp2*RSp461e*RS156pe + 2./27.*m11*mp2*RSp461e*RSe561p - 2./
         27.*m11*mp2*RSp461e*RSe461p + 4./27.*m11*mp2*RSp461e*RSp561e
          - 2./27.*m11*mp2*pow(RSp461e,2) + 2./27.*m11*me2*RSe561p*
         RS146pe - 2./27.*m11*me2*RSe561p*RS156pe - 2./27.*m11*me2*pow(
         RSe561p,2) - 2./27.*m11*me2*RSe461p*RS146pe + 2./27.*m11*me2*
         RSe461p*RS156pe + 4./27.*m11*me2*RSe461p*RSe561p - 2./27.*m11*
         me2*pow(RSe461p,2) - 2./27.*m11*me2*RSp561e*RSe561p + 2./27.*
         m11*me2*RSp561e*RSe461p + 2./27.*m11*me2*RSp461e*RSe561p - 2./
         27.*m11*me2*RSp461e*RSe461p - 13./36.*m11*m55 + 1./27.*m11*m55
         *pow(RS146pe,2) + 2./27.*m11*m55*RS156pe*RS146pe + 2./27.*m11*
         m55*RSe561p*RS146pe + 2./27.*m11*m55*RSe461p*RS146pe + 2./27.*
         m11*m55*RSe461p*RS156pe );

      M12p6x13R +=  + Abeps(m45,mu2)*pi16 * ( 2./27.*m11*m55*RSe461p*
         RSe561p + 1./27.*m11*m55*pow(RSe461p,2) + 2./27.*m11*m55*
         RSp561e*RS146pe + 2./27.*m11*m55*RSp561e*RSe461p + 2./27.*m11*
         m55*RSp461e*RS146pe + 2./27.*m11*m55*RSp461e*RS156pe + 2./27.*
         m11*m55*RSp461e*RSe561p + 2./27.*m11*m55*RSp461e*RSe461p + 2./
         27.*m11*m55*RSp461e*RSp561e + 1./27.*m11*m55*pow(RSp461e,2) - 
         13./36.*m11*m44 + 2./27.*m11*m44*RS156pe*RS146pe + 1./27.*m11*
         m44*pow(RS156pe,2) + 2./27.*m11*m44*RSe561p*RS146pe + 2./27.*
         m11*m44*RSe561p*RS156pe + 1./27.*m11*m44*pow(RSe561p,2) + 2./
         27.*m11*m44*RSe461p*RS156pe + 2./27.*m11*m44*RSe461p*RSe561p
          + 2./27.*m11*m44*RSp561e*RS146pe + 2./27.*m11*m44*RSp561e*
         RS156pe + 2./27.*m11*m44*RSp561e*RSe561p + 2./27.*m11*m44*
         RSp561e*RSe461p + 1./27.*m11*m44*pow(RSp561e,2) + 2./27.*m11*
         m44*RSp461e*RS156pe + 2./27.*m11*m44*RSp461e*RSe561p + 2./27.*
         m11*m44*RSp461e*RSp561e - 2./27.*pow(m11,2)*pow(RS146pe,2) + 4.
         /27.*pow(m11,2)*RS156pe*RS146pe );

      M12p6x13R +=  + Abeps(m45,mu2)*pi16 * (  - 2./27.*pow(m11,2)*pow(
         RS156pe,2) + 2./27.*pow(m11,2)*RSe561p*RS146pe - 2./27.*pow(
         m11,2)*RSe561p*RS156pe - 2./27.*pow(m11,2)*RSe461p*RS146pe + 2.
         /27.*pow(m11,2)*RSe461p*RS156pe + 2./27.*pow(m11,2)*RSp561e*
         RS146pe - 2./27.*pow(m11,2)*RSp561e*RS156pe - 2./27.*pow(
         m11,2)*RSp461e*RS146pe + 2./27.*pow(m11,2)*RSp461e*RS156pe );

      M12p6x13R +=  + Abeps(m46,mu2)*pi16 * ( 2./27.*m11*mp2*RSp561e*
         RS145pe - 2./27.*m11*mp2*RSp561e*RS156pe - 2./27.*m11*mp2*
         RSp561e*RSe561p + 2./27.*m11*mp2*RSp561e*RSe451p - 2./27.*m11*
         mp2*pow(RSp561e,2) - 2./27.*m11*mp2*RSp451e*RS145pe + 2./27.*
         m11*mp2*RSp451e*RS156pe + 2./27.*m11*mp2*RSp451e*RSe561p - 2./
         27.*m11*mp2*RSp451e*RSe451p + 4./27.*m11*mp2*RSp451e*RSp561e
          - 2./27.*m11*mp2*pow(RSp451e,2) + 2./27.*m11*me2*RSe561p*
         RS145pe - 2./27.*m11*me2*RSe561p*RS156pe - 2./27.*m11*me2*pow(
         RSe561p,2) - 2./27.*m11*me2*RSe451p*RS145pe + 2./27.*m11*me2*
         RSe451p*RS156pe + 4./27.*m11*me2*RSe451p*RSe561p - 2./27.*m11*
         me2*pow(RSe451p,2) - 2./27.*m11*me2*RSp561e*RSe561p + 2./27.*
         m11*me2*RSp561e*RSe451p + 2./27.*m11*me2*RSp451e*RSe561p - 2./
         27.*m11*me2*RSp451e*RSe451p - 13./36.*m11*m66 + 1./27.*m11*m66
         *pow(RS145pe,2) + 2./27.*m11*m66*RS156pe*RS145pe + 2./27.*m11*
         m66*RSe561p*RS145pe + 2./27.*m11*m66*RSe451p*RS145pe + 2./27.*
         m11*m66*RSe451p*RS156pe );

      M12p6x13R +=  + Abeps(m46,mu2)*pi16 * ( 2./27.*m11*m66*RSe451p*
         RSe561p + 1./27.*m11*m66*pow(RSe451p,2) + 2./27.*m11*m66*
         RSp561e*RS145pe + 2./27.*m11*m66*RSp561e*RSe451p + 2./27.*m11*
         m66*RSp451e*RS145pe + 2./27.*m11*m66*RSp451e*RS156pe + 2./27.*
         m11*m66*RSp451e*RSe561p + 2./27.*m11*m66*RSp451e*RSe451p + 2./
         27.*m11*m66*RSp451e*RSp561e + 1./27.*m11*m66*pow(RSp451e,2) - 
         13./36.*m11*m44 + 2./27.*m11*m44*RS156pe*RS145pe + 1./27.*m11*
         m44*pow(RS156pe,2) + 2./27.*m11*m44*RSe561p*RS145pe + 2./27.*
         m11*m44*RSe561p*RS156pe + 1./27.*m11*m44*pow(RSe561p,2) + 2./
         27.*m11*m44*RSe451p*RS156pe + 2./27.*m11*m44*RSe451p*RSe561p
          + 2./27.*m11*m44*RSp561e*RS145pe + 2./27.*m11*m44*RSp561e*
         RS156pe + 2./27.*m11*m44*RSp561e*RSe561p + 2./27.*m11*m44*
         RSp561e*RSe451p + 1./27.*m11*m44*pow(RSp561e,2) + 2./27.*m11*
         m44*RSp451e*RS156pe + 2./27.*m11*m44*RSp451e*RSe561p + 2./27.*
         m11*m44*RSp451e*RSp561e - 2./27.*pow(m11,2)*pow(RS145pe,2) + 4.
         /27.*pow(m11,2)*RS156pe*RS145pe );

      M12p6x13R +=  + Abeps(m46,mu2)*pi16 * (  - 2./27.*pow(m11,2)*pow(
         RS156pe,2) + 2./27.*pow(m11,2)*RSe561p*RS145pe - 2./27.*pow(
         m11,2)*RSe561p*RS156pe - 2./27.*pow(m11,2)*RSe451p*RS145pe + 2.
         /27.*pow(m11,2)*RSe451p*RS156pe + 2./27.*pow(m11,2)*RSp561e*
         RS145pe - 2./27.*pow(m11,2)*RSp561e*RS156pe - 2./27.*pow(
         m11,2)*RSp451e*RS145pe + 2./27.*pow(m11,2)*RSp451e*RS156pe );

      M12p6x13R +=  + Abeps(m55,mu2)*pi16 * (  - 13./36.*m11*m55 + 1./9.
         *m11*m55*pow(RS146pe,2) + 13./108.*m11*m55*RS546pe - 1./27.*
         m11*m55*RS546pe*pow(RS146pe,2) + 2./9.*m11*m55*RSe461p*RS146pe
          - 2./27.*m11*m55*RSe461p*RS546pe*RS146pe + 1./9.*m11*m55*pow(
         RSe461p,2) - 1./27.*m11*m55*pow(RSe461p,2)*RS546pe + 2./9.*m11
         *m55*RSp461e*RS146pe - 2./27.*m11*m55*RSp461e*RS546pe*RS146pe
          + 2./9.*m11*m55*RSp461e*RSe461p - 2./27.*m11*m55*RSp461e*
         RSe461p*RS546pe + 1./9.*m11*m55*pow(RSp461e,2) - 1./27.*m11*
         m55*pow(RSp461e,2)*RS546pe );

      M12p6x13R +=  + Abeps(m56,mu2)*pi16 * ( 2./27.*m11*mp2*RSp461e*
         RS145pe - 2./27.*m11*mp2*RSp461e*RS146pe - 2./27.*m11*mp2*
         RSp461e*RSe461p + 2./27.*m11*mp2*RSp461e*RSe451p - 2./27.*m11*
         mp2*pow(RSp461e,2) - 2./27.*m11*mp2*RSp451e*RS145pe + 2./27.*
         m11*mp2*RSp451e*RS146pe + 2./27.*m11*mp2*RSp451e*RSe461p - 2./
         27.*m11*mp2*RSp451e*RSe451p + 4./27.*m11*mp2*RSp451e*RSp461e
          - 2./27.*m11*mp2*pow(RSp451e,2) + 2./27.*m11*me2*RSe461p*
         RS145pe - 2./27.*m11*me2*RSe461p*RS146pe - 2./27.*m11*me2*pow(
         RSe461p,2) - 2./27.*m11*me2*RSe451p*RS145pe + 2./27.*m11*me2*
         RSe451p*RS146pe + 4./27.*m11*me2*RSe451p*RSe461p - 2./27.*m11*
         me2*pow(RSe451p,2) - 2./27.*m11*me2*RSp461e*RSe461p + 2./27.*
         m11*me2*RSp461e*RSe451p + 2./27.*m11*me2*RSp451e*RSe461p - 2./
         27.*m11*me2*RSp451e*RSe451p - 13./36.*m11*m66 + 1./27.*m11*m66
         *pow(RS145pe,2) + 2./27.*m11*m66*RS146pe*RS145pe + 2./27.*m11*
         m66*RSe461p*RS145pe + 2./27.*m11*m66*RSe451p*RS145pe + 2./27.*
         m11*m66*RSe451p*RS146pe );

      M12p6x13R +=  + Abeps(m56,mu2)*pi16 * ( 2./27.*m11*m66*RSe451p*
         RSe461p + 1./27.*m11*m66*pow(RSe451p,2) + 2./27.*m11*m66*
         RSp461e*RS145pe + 2./27.*m11*m66*RSp461e*RSe451p + 2./27.*m11*
         m66*RSp451e*RS145pe + 2./27.*m11*m66*RSp451e*RS146pe + 2./27.*
         m11*m66*RSp451e*RSe461p + 2./27.*m11*m66*RSp451e*RSe451p + 2./
         27.*m11*m66*RSp451e*RSp461e + 1./27.*m11*m66*pow(RSp451e,2) - 
         13./36.*m11*m55 + 2./27.*m11*m55*RS146pe*RS145pe + 1./27.*m11*
         m55*pow(RS146pe,2) + 2./27.*m11*m55*RSe461p*RS145pe + 2./27.*
         m11*m55*RSe461p*RS146pe + 1./27.*m11*m55*pow(RSe461p,2) + 2./
         27.*m11*m55*RSe451p*RS146pe + 2./27.*m11*m55*RSe451p*RSe461p
          + 2./27.*m11*m55*RSp461e*RS145pe + 2./27.*m11*m55*RSp461e*
         RS146pe + 2./27.*m11*m55*RSp461e*RSe461p + 2./27.*m11*m55*
         RSp461e*RSe451p + 1./27.*m11*m55*pow(RSp461e,2) + 2./27.*m11*
         m55*RSp451e*RS146pe + 2./27.*m11*m55*RSp451e*RSe461p + 2./27.*
         m11*m55*RSp451e*RSp461e - 2./27.*pow(m11,2)*pow(RS145pe,2) + 4.
         /27.*pow(m11,2)*RS146pe*RS145pe );

      M12p6x13R +=  + Abeps(m56,mu2)*pi16 * (  - 2./27.*pow(m11,2)*pow(
         RS146pe,2) + 2./27.*pow(m11,2)*RSe461p*RS145pe - 2./27.*pow(
         m11,2)*RSe461p*RS146pe - 2./27.*pow(m11,2)*RSe451p*RS145pe + 2.
         /27.*pow(m11,2)*RSe451p*RS146pe + 2./27.*pow(m11,2)*RSp461e*
         RS145pe - 2./27.*pow(m11,2)*RSp461e*RS146pe - 2./27.*pow(
         m11,2)*RSp451e*RS145pe + 2./27.*pow(m11,2)*RSp451e*RS146pe );

      M12p6x13R +=  + Abeps(m66,mu2)*pi16 * (  - 13./36.*m11*m66 + 1./9.
         *m11*m66*pow(RS145pe,2) + 13./108.*m11*m66*RS645pe - 1./27.*
         m11*m66*RS645pe*pow(RS145pe,2) + 2./9.*m11*m66*RSe451p*RS145pe
          - 2./27.*m11*m66*RSe451p*RS645pe*RS145pe + 1./9.*m11*m66*pow(
         RSe451p,2) - 1./27.*m11*m66*pow(RSe451p,2)*RS645pe + 2./9.*m11
         *m66*RSp451e*RS145pe - 2./27.*m11*m66*RSp451e*RS645pe*RS145pe
          + 2./9.*m11*m66*RSp451e*RSe451p - 2./27.*m11*m66*RSp451e*
         RSe451p*RS645pe + 1./9.*m11*m66*pow(RSp451e,2) - 1./27.*m11*
         m66*pow(RSp451e,2)*RS645pe );

      M12p6x13R +=  + Abeps(me2,mu2)*pi16 * ( 1./45.*m66*me2*RSe45611p
          - 2./45.*m66*me2*RSe451p + 1./45.*m66*me2*RSe456p + 1./45.*
         pow(m66,2)*RSe451p - 1./45.*pow(m66,2)*RSe456p + 1./45.*m55*
         me2*RSe45611p - 2./45.*m55*me2*RSe461p + 1./45.*m55*me2*
         RSe465p + 1./45.*pow(m55,2)*RSe461p - 1./45.*pow(m55,2)*
         RSe465p + 1./45.*m44*me2*RSe45611p - 2./45.*m44*me2*RSe561p + 
         1./45.*m44*me2*RSe564p + 1./45.*pow(m44,2)*RSe561p - 1./45.*
         pow(m44,2)*RSe564p + 1./9.*m11*RSe45611p*RD1456ep + 1./9.*m11*
         mp2*RSp45611e*RSe45611p + 19./60.*m11*me2*RSe45611p + 1./9.*
         m11*me2*pow(RSe45611p,2) - 2./45.*m11*me2*RSe561p - 2./45.*m11
         *me2*RSe461p - 2./45.*m11*me2*RSe451p + 19./180.*m11*me2*
         RSe564p + 19./180.*m11*me2*RSe465p + 19./180.*m11*me2*RSe456p
          + 17./135.*m11*m66*RSe45611p + 1./27.*m11*m66*pow(RSe561p,2)
          + 1./27.*m11*m66*pow(RSe461p,2) - 1./15.*m11*m66*RSe451p + 1./
         27.*m11*m66*RSe451p*RSe561p + 1./27.*m11*m66*RSe451p*RSe461p
          + 5./27.*m11*m66*pow(RSe451p,2) );

      M12p6x13R +=  + Abeps(me2,mu2)*pi16 * ( 2./27.*m11*m66*RSp561e*
         RSe561p + 1./27.*m11*m66*RSp561e*RSe451p + 2./27.*m11*m66*
         RSp461e*RSe461p + 1./27.*m11*m66*RSp461e*RSe451p + 1./27.*m11*
         m66*RSp451e*RSe561p + 1./27.*m11*m66*RSp451e*RSe461p + 10./27.
         *m11*m66*RSp451e*RSe451p + 2./135.*m11*m66*RSe456p - 1./27.*
         m11*m66*RSe456p*pow(RS145pe,2) - 2./27.*m11*m66*RSe456p*
         RSe451p*RS145pe - 1./27.*m11*m66*RSe456p*pow(RSe451p,2) - 2./
         27.*m11*m66*RSe456p*RSp451e*RS145pe - 2./27.*m11*m66*RSe456p*
         RSp451e*RSe451p - 1./27.*m11*m66*RSe456p*pow(RSp451e,2) + 2./
         27.*m11*m66*mp2*RNep*RSp561e*RSe561p + 1./27.*m11*m66*mp2*RNep
         *RSp561e*RSe451p + 2./27.*m11*m66*mp2*RNep*RSp461e*RSe461p + 1.
         /27.*m11*m66*mp2*RNep*RSp461e*RSe451p + 1./27.*m11*m66*mp2*
         RNep*RSp451e*RSe561p + 1./27.*m11*m66*mp2*RNep*RSp451e*RSe461p
          + 10./27.*m11*m66*mp2*RNep*RSp451e*RSe451p - 2./27.*m11*m66*
         mp2*RNep*RSp456e*RSp451e*RSe451p + 2./27.*m11*m66*mp2*RN1e*
         RSp456e*RSe451p*RS145pe );

      M12p6x13R +=  + Abeps(me2,mu2)*pi16 * (  - 2./27.*m11*m66*me2*
         RNep*RSe456p*RSp451e*RSe451p + 2./9.*m11*m66*me2*RN1e*
         RSe45611p - 1./27.*m11*m66*me2*RN1e*RSe561p*RS145pe - 2./27.*
         m11*m66*me2*RN1e*RSe561p*RS156pe - 1./27.*m11*m66*me2*RN1e*
         RSe461p*RS145pe - 2./27.*m11*m66*me2*RN1e*RSe461p*RS146pe + 1./
         9.*m11*m66*me2*RN1e*RSe451p - 10./27.*m11*m66*me2*RN1e*RSe451p
         *RS145pe - 1./27.*m11*m66*me2*RN1e*RSe451p*RS146pe - 1./27.*
         m11*m66*me2*RN1e*RSe451p*RS156pe + 2./27.*m11*m66*me2*RN1e*
         RSe456p*RSe451p*RS145pe + 1./27.*m11*pow(m66,2)*RNep*RSp561e*
         RSe451p + 1./27.*m11*pow(m66,2)*RNep*RSp461e*RSe451p + 1./27.*
         m11*pow(m66,2)*RNep*RSp451e*RSe561p + 1./27.*m11*pow(m66,2)*
         RNep*RSp451e*RSe461p - 2./81.*m11*pow(m66,2)*RNep*RSp451e*
         RSe451p - 2./27.*m11*pow(m66,2)*RNep*RSp451e*RSe451p*RS645pe
          - 1./27.*m11*pow(m66,2)*RN1e*RSe561p*RS145pe - 1./27.*m11*
         pow(m66,2)*RN1e*RSe461p*RS145pe + 1./9.*m11*pow(m66,2)*RN1e*
         RSe451p );

      M12p6x13R +=  + Abeps(me2,mu2)*pi16 * ( 2./81.*m11*pow(m66,2)*
         RN1e*RSe451p*RS145pe - 1./27.*m11*pow(m66,2)*RN1e*RSe451p*
         RS146pe - 1./27.*m11*pow(m66,2)*RN1e*RSe451p*RS156pe + 2./27.*
         m11*pow(m66,2)*RN1e*RSe451p*RS645pe*RS145pe + 17./135.*m11*m55
         *RSe45611p + 1./27.*m11*m55*pow(RSe561p,2) - 1./15.*m11*m55*
         RSe461p + 1./27.*m11*m55*RSe461p*RSe561p + 5./27.*m11*m55*pow(
         RSe461p,2) + 1./27.*m11*m55*RSe451p*RSe461p + 1./27.*m11*m55*
         pow(RSe451p,2) + 2./27.*m11*m55*RSp561e*RSe561p + 1./27.*m11*
         m55*RSp561e*RSe461p + 1./27.*m11*m55*RSp461e*RSe561p + 10./27.
         *m11*m55*RSp461e*RSe461p + 1./27.*m11*m55*RSp461e*RSe451p + 1./
         27.*m11*m55*RSp451e*RSe461p + 2./27.*m11*m55*RSp451e*RSe451p
          + 2./135.*m11*m55*RSe465p - 1./27.*m11*m55*RSe465p*pow(
         RS146pe,2) - 2./27.*m11*m55*RSe465p*RSe461p*RS146pe - 1./27.*
         m11*m55*RSe465p*pow(RSe461p,2) - 2./27.*m11*m55*RSe465p*
         RSp461e*RS146pe - 2./27.*m11*m55*RSe465p*RSp461e*RSe461p - 1./
         27.*m11*m55*RSe465p*pow(RSp461e,2) );

      M12p6x13R +=  + Abeps(me2,mu2)*pi16 * ( 2./27.*m11*m55*mp2*RNep*
         RSp561e*RSe561p + 1./27.*m11*m55*mp2*RNep*RSp561e*RSe461p + 1./
         27.*m11*m55*mp2*RNep*RSp461e*RSe561p + 10./27.*m11*m55*mp2*
         RNep*RSp461e*RSe461p + 1./27.*m11*m55*mp2*RNep*RSp461e*RSe451p
          + 1./27.*m11*m55*mp2*RNep*RSp451e*RSe461p + 2./27.*m11*m55*
         mp2*RNep*RSp451e*RSe451p - 2./27.*m11*m55*mp2*RNep*RSp465e*
         RSp461e*RSe461p + 2./27.*m11*m55*mp2*RN1e*RSp465e*RSe461p*
         RS146pe - 2./27.*m11*m55*me2*RNep*RSe465p*RSp461e*RSe461p + 2./
         9.*m11*m55*me2*RN1e*RSe45611p - 1./27.*m11*m55*me2*RN1e*
         RSe561p*RS146pe - 2./27.*m11*m55*me2*RN1e*RSe561p*RS156pe + 1./
         9.*m11*m55*me2*RN1e*RSe461p - 1./27.*m11*m55*me2*RN1e*RSe461p*
         RS145pe - 10./27.*m11*m55*me2*RN1e*RSe461p*RS146pe - 1./27.*
         m11*m55*me2*RN1e*RSe461p*RS156pe - 2./27.*m11*m55*me2*RN1e*
         RSe451p*RS145pe - 1./27.*m11*m55*me2*RN1e*RSe451p*RS146pe + 2./
         27.*m11*m55*me2*RN1e*RSe465p*RSe461p*RS146pe - 8./81.*m11*m55*
         m66*RNep*RSp461e*RSe461p );

      M12p6x13R +=  + Abeps(me2,mu2)*pi16 * ( 2./27.*m11*m55*m66*RNep*
         RSp461e*RSe451p + 2./27.*m11*m55*m66*RNep*RSp451e*RSe461p - 8./
         81.*m11*m55*m66*RNep*RSp451e*RSe451p - 2./27.*m11*m55*m66*RN1e
         *RSe461p*RS145pe + 8./81.*m11*m55*m66*RN1e*RSe461p*RS146pe + 8.
         /81.*m11*m55*m66*RN1e*RSe451p*RS145pe - 2./27.*m11*m55*m66*
         RN1e*RSe451p*RS146pe + 1./27.*m11*pow(m55,2)*RNep*RSp561e*
         RSe461p + 1./27.*m11*pow(m55,2)*RNep*RSp461e*RSe561p - 2./81.*
         m11*pow(m55,2)*RNep*RSp461e*RSe461p - 2./27.*m11*pow(m55,2)*
         RNep*RSp461e*RSe461p*RS546pe + 1./27.*m11*pow(m55,2)*RNep*
         RSp461e*RSe451p + 1./27.*m11*pow(m55,2)*RNep*RSp451e*RSe461p
          - 1./27.*m11*pow(m55,2)*RN1e*RSe561p*RS146pe + 1./9.*m11*pow(
         m55,2)*RN1e*RSe461p - 1./27.*m11*pow(m55,2)*RN1e*RSe461p*
         RS145pe + 2./81.*m11*pow(m55,2)*RN1e*RSe461p*RS146pe - 1./27.*
         m11*pow(m55,2)*RN1e*RSe461p*RS156pe + 2./27.*m11*pow(m55,2)*
         RN1e*RSe461p*RS546pe*RS146pe - 1./27.*m11*pow(m55,2)*RN1e*
         RSe451p*RS146pe );

      M12p6x13R +=  + Abeps(me2,mu2)*pi16 * ( 17./135.*m11*m44*
         RSe45611p - 1./15.*m11*m44*RSe561p + 5./27.*m11*m44*pow(
         RSe561p,2) + 1./27.*m11*m44*RSe461p*RSe561p + 1./27.*m11*m44*
         pow(RSe461p,2) + 1./27.*m11*m44*RSe451p*RSe561p + 1./27.*m11*
         m44*pow(RSe451p,2) + 10./27.*m11*m44*RSp561e*RSe561p + 1./27.*
         m11*m44*RSp561e*RSe461p + 1./27.*m11*m44*RSp561e*RSe451p + 1./
         27.*m11*m44*RSp461e*RSe561p + 2./27.*m11*m44*RSp461e*RSe461p
          + 1./27.*m11*m44*RSp451e*RSe561p + 2./27.*m11*m44*RSp451e*
         RSe451p + 2./135.*m11*m44*RSe564p - 1./27.*m11*m44*RSe564p*
         pow(RS156pe,2) - 2./27.*m11*m44*RSe564p*RSe561p*RS156pe - 1./
         27.*m11*m44*RSe564p*pow(RSe561p,2) - 2./27.*m11*m44*RSe564p*
         RSp561e*RS156pe - 2./27.*m11*m44*RSe564p*RSp561e*RSe561p - 1./
         27.*m11*m44*RSe564p*pow(RSp561e,2) + 10./27.*m11*m44*mp2*RNep*
         RSp561e*RSe561p + 1./27.*m11*m44*mp2*RNep*RSp561e*RSe461p + 1./
         27.*m11*m44*mp2*RNep*RSp561e*RSe451p + 1./27.*m11*m44*mp2*RNep
         *RSp461e*RSe561p );

      M12p6x13R +=  + Abeps(me2,mu2)*pi16 * ( 2./27.*m11*m44*mp2*RNep*
         RSp461e*RSe461p + 1./27.*m11*m44*mp2*RNep*RSp451e*RSe561p + 2./
         27.*m11*m44*mp2*RNep*RSp451e*RSe451p - 2./27.*m11*m44*mp2*RNep
         *RSp564e*RSp561e*RSe561p + 2./27.*m11*m44*mp2*RN1e*RSp564e*
         RSe561p*RS156pe - 2./27.*m11*m44*me2*RNep*RSe564p*RSp561e*
         RSe561p + 2./9.*m11*m44*me2*RN1e*RSe45611p + 1./9.*m11*m44*me2
         *RN1e*RSe561p - 1./27.*m11*m44*me2*RN1e*RSe561p*RS145pe - 1./
         27.*m11*m44*me2*RN1e*RSe561p*RS146pe - 10./27.*m11*m44*me2*
         RN1e*RSe561p*RS156pe - 2./27.*m11*m44*me2*RN1e*RSe461p*RS146pe
          - 1./27.*m11*m44*me2*RN1e*RSe461p*RS156pe - 2./27.*m11*m44*
         me2*RN1e*RSe451p*RS145pe - 1./27.*m11*m44*me2*RN1e*RSe451p*
         RS156pe + 2./27.*m11*m44*me2*RN1e*RSe564p*RSe561p*RS156pe - 8./
         81.*m11*m44*m66*RNep*RSp561e*RSe561p + 2./27.*m11*m44*m66*RNep
         *RSp561e*RSe451p + 2./27.*m11*m44*m66*RNep*RSp451e*RSe561p - 8.
         /81.*m11*m44*m66*RNep*RSp451e*RSe451p - 2./27.*m11*m44*m66*
         RN1e*RSe561p*RS145pe );

      M12p6x13R +=  + Abeps(me2,mu2)*pi16 * ( 8./81.*m11*m44*m66*RN1e*
         RSe561p*RS156pe + 8./81.*m11*m44*m66*RN1e*RSe451p*RS145pe - 2./
         27.*m11*m44*m66*RN1e*RSe451p*RS156pe - 8./81.*m11*m44*m55*RNep
         *RSp561e*RSe561p + 2./27.*m11*m44*m55*RNep*RSp561e*RSe461p + 2.
         /27.*m11*m44*m55*RNep*RSp461e*RSe561p - 8./81.*m11*m44*m55*
         RNep*RSp461e*RSe461p - 2./27.*m11*m44*m55*RN1e*RSe561p*RS146pe
          + 8./81.*m11*m44*m55*RN1e*RSe561p*RS156pe + 8./81.*m11*m44*
         m55*RN1e*RSe461p*RS146pe - 2./27.*m11*m44*m55*RN1e*RSe461p*
         RS156pe - 2./81.*m11*pow(m44,2)*RNep*RSp561e*RSe561p - 2./27.*
         m11*pow(m44,2)*RNep*RSp561e*RSe561p*RS456pe + 1./27.*m11*pow(
         m44,2)*RNep*RSp561e*RSe461p + 1./27.*m11*pow(m44,2)*RNep*
         RSp561e*RSe451p + 1./27.*m11*pow(m44,2)*RNep*RSp461e*RSe561p
          + 1./27.*m11*pow(m44,2)*RNep*RSp451e*RSe561p + 1./9.*m11*pow(
         m44,2)*RN1e*RSe561p - 1./27.*m11*pow(m44,2)*RN1e*RSe561p*
         RS145pe - 1./27.*m11*pow(m44,2)*RN1e*RSe561p*RS146pe + 2./81.*
         m11*pow(m44,2)*RN1e*RSe561p*RS156pe );

      M12p6x13R +=  + Abeps(me2,mu2)*pi16 * ( 2./27.*m11*pow(m44,2)*
         RN1e*RSe561p*RS456pe*RS156pe - 1./27.*m11*pow(m44,2)*RN1e*
         RSe461p*RS156pe - 1./27.*m11*pow(m44,2)*RN1e*RSe451p*RS156pe
          + 17./45.*pow(m11,2)*RSe45611p + 1./3.*pow(m11,2)*RSe45611p*
         RC1456pe + 2./9.*pow(m11,2)*pow(RSe45611p,2) + 2./9.*pow(
         m11,2)*RSp45611e*RSe45611p - 4./45.*pow(m11,2)*RSe561p - 4./45.
         *pow(m11,2)*RSe461p - 4./45.*pow(m11,2)*RSe451p - 2./9.*pow(
         m11,2)*RN1e*RSe45611p*RD1456ep - 2./9.*pow(m11,2)*mp2*RN1e*
         RSp45611e*RSe45611p + 2./3.*pow(m11,2)*me2*RN1e*RSe45611p - 2./
         9.*pow(m11,2)*me2*RN1e*pow(RSe45611p,2) + 1./9.*pow(m11,2)*me2
         *RN1e*RSe561p + 1./9.*pow(m11,2)*me2*RN1e*RSe461p + 1./9.*pow(
         m11,2)*me2*RN1e*RSe451p - 8./27.*pow(m11,2)*m66*RN1e*RSe45611p
          + 2./9.*pow(m11,2)*m66*RN1e*RSe451p - 8./27.*pow(m11,2)*m55*
         RN1e*RSe45611p + 2./9.*pow(m11,2)*m55*RN1e*RSe461p - 8./27.*
         pow(m11,2)*m44*RN1e*RSe45611p + 2./9.*pow(m11,2)*m44*RN1e*
         RSe561p );

      M12p6x13R +=  + Abeps(me2,mu2)*pi16 * (  - 2./9.*pow(m11,3)*RN1e*
         RSe45611p + 2./9.*pow(m11,3)*RN1e*RSe45611p*RS1456pee + 2./9.*
         pow(m11,3)*RN1e*RSe45611p*RS1456ppe - 2./9.*pow(m11,3)*RN1e*
         RS145pe*RSe45611p - 2./9.*pow(m11,3)*RN1e*RS146pe*RSe45611p - 
         2./9.*pow(m11,3)*RN1e*RS156pe*RSe45611p + 1./9.*pow(m11,3)*
         RN1e*RSe561p + 1./9.*pow(m11,3)*RN1e*RSe461p + 1./9.*pow(
         m11,3)*RN1e*RSe451p );

      M12p6x13R +=  + Abeps(mp2,mu2)*pi16 * ( 1./45.*m66*mp2*RSp45611e
          - 2./45.*m66*mp2*RSp451e + 1./45.*m66*mp2*RSp456e + 1./45.*
         pow(m66,2)*RSp451e - 1./45.*pow(m66,2)*RSp456e + 1./45.*m55*
         mp2*RSp45611e - 2./45.*m55*mp2*RSp461e + 1./45.*m55*mp2*
         RSp465e + 1./45.*pow(m55,2)*RSp461e - 1./45.*pow(m55,2)*
         RSp465e + 1./45.*m44*mp2*RSp45611e - 2./45.*m44*mp2*RSp561e + 
         1./45.*m44*mp2*RSp564e + 1./45.*pow(m44,2)*RSp561e - 1./45.*
         pow(m44,2)*RSp564e + 1./9.*m11*RSp45611e*RD1456ep + 19./60.*
         m11*mp2*RSp45611e + 1./9.*m11*mp2*pow(RSp45611e,2) - 2./45.*
         m11*mp2*RSp561e - 2./45.*m11*mp2*RSp461e - 2./45.*m11*mp2*
         RSp451e + 19./180.*m11*mp2*RSp564e + 19./180.*m11*mp2*RSp465e
          + 19./180.*m11*mp2*RSp456e + 1./9.*m11*me2*RSp45611e*
         RSe45611p + 17./135.*m11*m66*RSp45611e + 1./27.*m11*m66*pow(
         RSp561e,2) + 1./27.*m11*m66*pow(RSp461e,2) - 1./15.*m11*m66*
         RSp451e + 1./27.*m11*m66*RSp451e*RSp561e + 1./27.*m11*m66*
         RSp451e*RSp461e );

      M12p6x13R +=  + Abeps(mp2,mu2)*pi16 * ( 5./27.*m11*m66*pow(
         RSp451e,2) + 2./135.*m11*m66*RSp456e - 1./27.*m11*m66*RSp456e*
         pow(RS145pe,2) - 2./27.*m11*m66*RSp456e*RSe451p*RS145pe - 1./
         27.*m11*m66*RSp456e*pow(RSe451p,2) - 2./27.*m11*m66*RSp456e*
         RSp451e*RS145pe - 2./27.*m11*m66*RSp456e*RSp451e*RSe451p - 1./
         27.*m11*m66*RSp456e*pow(RSp451e,2) - 2./27.*m11*m66*mp2*RNep*
         RSp561e*RSe561p - 1./27.*m11*m66*mp2*RNep*RSp561e*RSe451p - 2./
         27.*m11*m66*mp2*RNep*RSp461e*RSe461p - 1./27.*m11*m66*mp2*RNep
         *RSp461e*RSe451p - 1./27.*m11*m66*mp2*RNep*RSp451e*RSe561p - 1.
         /27.*m11*m66*mp2*RNep*RSp451e*RSe461p - 10./27.*m11*m66*mp2*
         RNep*RSp451e*RSe451p + 2./27.*m11*m66*mp2*RNep*RSp456e*RSp451e
         *RSe451p + 2./9.*m11*m66*mp2*RN1p*RSp45611e - 1./27.*m11*m66*
         mp2*RN1p*RSp561e*RS145pe - 2./27.*m11*m66*mp2*RN1p*RSp561e*
         RS156pe - 1./27.*m11*m66*mp2*RN1p*RSp461e*RS145pe - 2./27.*m11
         *m66*mp2*RN1p*RSp461e*RS146pe + 1./9.*m11*m66*mp2*RN1p*RSp451e
          - 10./27.*m11*m66*mp2*RN1p*RSp451e*RS145pe );

      M12p6x13R +=  + Abeps(mp2,mu2)*pi16 * (  - 1./27.*m11*m66*mp2*
         RN1p*RSp451e*RS146pe - 1./27.*m11*m66*mp2*RN1p*RSp451e*RS156pe
          + 2./27.*m11*m66*mp2*RN1p*RSp456e*RSp451e*RS145pe + 2./27.*
         m11*m66*me2*RNep*RSe456p*RSp451e*RSe451p + 2./27.*m11*m66*me2*
         RN1p*RSe456p*RSp451e*RS145pe - 1./27.*m11*pow(m66,2)*RNep*
         RSp561e*RSe451p - 1./27.*m11*pow(m66,2)*RNep*RSp461e*RSe451p
          - 1./27.*m11*pow(m66,2)*RNep*RSp451e*RSe561p - 1./27.*m11*
         pow(m66,2)*RNep*RSp451e*RSe461p + 2./81.*m11*pow(m66,2)*RNep*
         RSp451e*RSe451p + 2./27.*m11*pow(m66,2)*RNep*RSp451e*RSe451p*
         RS645pe - 1./27.*m11*pow(m66,2)*RN1p*RSp561e*RS145pe - 1./27.*
         m11*pow(m66,2)*RN1p*RSp461e*RS145pe + 1./9.*m11*pow(m66,2)*
         RN1p*RSp451e + 2./81.*m11*pow(m66,2)*RN1p*RSp451e*RS145pe - 1./
         27.*m11*pow(m66,2)*RN1p*RSp451e*RS146pe - 1./27.*m11*pow(
         m66,2)*RN1p*RSp451e*RS156pe + 2./27.*m11*pow(m66,2)*RN1p*
         RSp451e*RS645pe*RS145pe + 17./135.*m11*m55*RSp45611e + 1./27.*
         m11*m55*pow(RSp561e,2) );

      M12p6x13R +=  + Abeps(mp2,mu2)*pi16 * (  - 1./15.*m11*m55*RSp461e
          + 1./27.*m11*m55*RSp461e*RSp561e + 5./27.*m11*m55*pow(
         RSp461e,2) + 1./27.*m11*m55*RSp451e*RSp461e + 1./27.*m11*m55*
         pow(RSp451e,2) + 2./135.*m11*m55*RSp465e - 1./27.*m11*m55*
         RSp465e*pow(RS146pe,2) - 2./27.*m11*m55*RSp465e*RSe461p*
         RS146pe - 1./27.*m11*m55*RSp465e*pow(RSe461p,2) - 2./27.*m11*
         m55*RSp465e*RSp461e*RS146pe - 2./27.*m11*m55*RSp465e*RSp461e*
         RSe461p - 1./27.*m11*m55*RSp465e*pow(RSp461e,2) - 2./27.*m11*
         m55*mp2*RNep*RSp561e*RSe561p - 1./27.*m11*m55*mp2*RNep*RSp561e
         *RSe461p - 1./27.*m11*m55*mp2*RNep*RSp461e*RSe561p - 10./27.*
         m11*m55*mp2*RNep*RSp461e*RSe461p - 1./27.*m11*m55*mp2*RNep*
         RSp461e*RSe451p - 1./27.*m11*m55*mp2*RNep*RSp451e*RSe461p - 2./
         27.*m11*m55*mp2*RNep*RSp451e*RSe451p + 2./27.*m11*m55*mp2*RNep
         *RSp465e*RSp461e*RSe461p + 2./9.*m11*m55*mp2*RN1p*RSp45611e - 
         1./27.*m11*m55*mp2*RN1p*RSp561e*RS146pe - 2./27.*m11*m55*mp2*
         RN1p*RSp561e*RS156pe );

      M12p6x13R +=  + Abeps(mp2,mu2)*pi16 * ( 1./9.*m11*m55*mp2*RN1p*
         RSp461e - 1./27.*m11*m55*mp2*RN1p*RSp461e*RS145pe - 10./27.*
         m11*m55*mp2*RN1p*RSp461e*RS146pe - 1./27.*m11*m55*mp2*RN1p*
         RSp461e*RS156pe - 2./27.*m11*m55*mp2*RN1p*RSp451e*RS145pe - 1./
         27.*m11*m55*mp2*RN1p*RSp451e*RS146pe + 2./27.*m11*m55*mp2*RN1p
         *RSp465e*RSp461e*RS146pe + 2./27.*m11*m55*me2*RNep*RSe465p*
         RSp461e*RSe461p + 2./27.*m11*m55*me2*RN1p*RSe465p*RSp461e*
         RS146pe + 8./81.*m11*m55*m66*RNep*RSp461e*RSe461p - 2./27.*m11
         *m55*m66*RNep*RSp461e*RSe451p - 2./27.*m11*m55*m66*RNep*
         RSp451e*RSe461p + 8./81.*m11*m55*m66*RNep*RSp451e*RSe451p - 2./
         27.*m11*m55*m66*RN1p*RSp461e*RS145pe + 8./81.*m11*m55*m66*RN1p
         *RSp461e*RS146pe + 8./81.*m11*m55*m66*RN1p*RSp451e*RS145pe - 2.
         /27.*m11*m55*m66*RN1p*RSp451e*RS146pe - 1./27.*m11*pow(m55,2)*
         RNep*RSp561e*RSe461p - 1./27.*m11*pow(m55,2)*RNep*RSp461e*
         RSe561p + 2./81.*m11*pow(m55,2)*RNep*RSp461e*RSe461p + 2./27.*
         m11*pow(m55,2)*RNep*RSp461e*RSe461p*RS546pe );

      M12p6x13R +=  + Abeps(mp2,mu2)*pi16 * (  - 1./27.*m11*pow(m55,2)*
         RNep*RSp461e*RSe451p - 1./27.*m11*pow(m55,2)*RNep*RSp451e*
         RSe461p - 1./27.*m11*pow(m55,2)*RN1p*RSp561e*RS146pe + 1./9.*
         m11*pow(m55,2)*RN1p*RSp461e - 1./27.*m11*pow(m55,2)*RN1p*
         RSp461e*RS145pe + 2./81.*m11*pow(m55,2)*RN1p*RSp461e*RS146pe
          - 1./27.*m11*pow(m55,2)*RN1p*RSp461e*RS156pe + 2./27.*m11*
         pow(m55,2)*RN1p*RSp461e*RS546pe*RS146pe - 1./27.*m11*pow(
         m55,2)*RN1p*RSp451e*RS146pe + 17./135.*m11*m44*RSp45611e - 1./
         15.*m11*m44*RSp561e + 5./27.*m11*m44*pow(RSp561e,2) + 1./27.*
         m11*m44*RSp461e*RSp561e + 1./27.*m11*m44*pow(RSp461e,2) + 1./
         27.*m11*m44*RSp451e*RSp561e + 1./27.*m11*m44*pow(RSp451e,2) + 
         2./135.*m11*m44*RSp564e - 1./27.*m11*m44*RSp564e*pow(
         RS156pe,2) - 2./27.*m11*m44*RSp564e*RSe561p*RS156pe - 1./27.*
         m11*m44*RSp564e*pow(RSe561p,2) - 2./27.*m11*m44*RSp564e*
         RSp561e*RS156pe - 2./27.*m11*m44*RSp564e*RSp561e*RSe561p - 1./
         27.*m11*m44*RSp564e*pow(RSp561e,2) );

      M12p6x13R +=  + Abeps(mp2,mu2)*pi16 * (  - 10./27.*m11*m44*mp2*
         RNep*RSp561e*RSe561p - 1./27.*m11*m44*mp2*RNep*RSp561e*RSe461p
          - 1./27.*m11*m44*mp2*RNep*RSp561e*RSe451p - 1./27.*m11*m44*
         mp2*RNep*RSp461e*RSe561p - 2./27.*m11*m44*mp2*RNep*RSp461e*
         RSe461p - 1./27.*m11*m44*mp2*RNep*RSp451e*RSe561p - 2./27.*m11
         *m44*mp2*RNep*RSp451e*RSe451p + 2./27.*m11*m44*mp2*RNep*
         RSp564e*RSp561e*RSe561p + 2./9.*m11*m44*mp2*RN1p*RSp45611e + 1.
         /9.*m11*m44*mp2*RN1p*RSp561e - 1./27.*m11*m44*mp2*RN1p*RSp561e
         *RS145pe - 1./27.*m11*m44*mp2*RN1p*RSp561e*RS146pe - 10./27.*
         m11*m44*mp2*RN1p*RSp561e*RS156pe - 2./27.*m11*m44*mp2*RN1p*
         RSp461e*RS146pe - 1./27.*m11*m44*mp2*RN1p*RSp461e*RS156pe - 2./
         27.*m11*m44*mp2*RN1p*RSp451e*RS145pe - 1./27.*m11*m44*mp2*RN1p
         *RSp451e*RS156pe + 2./27.*m11*m44*mp2*RN1p*RSp564e*RSp561e*
         RS156pe + 2./27.*m11*m44*me2*RNep*RSe564p*RSp561e*RSe561p + 2./
         27.*m11*m44*me2*RN1p*RSe564p*RSp561e*RS156pe + 8./81.*m11*m44*
         m66*RNep*RSp561e*RSe561p );

      M12p6x13R +=  + Abeps(mp2,mu2)*pi16 * (  - 2./27.*m11*m44*m66*
         RNep*RSp561e*RSe451p - 2./27.*m11*m44*m66*RNep*RSp451e*RSe561p
          + 8./81.*m11*m44*m66*RNep*RSp451e*RSe451p - 2./27.*m11*m44*
         m66*RN1p*RSp561e*RS145pe + 8./81.*m11*m44*m66*RN1p*RSp561e*
         RS156pe + 8./81.*m11*m44*m66*RN1p*RSp451e*RS145pe - 2./27.*m11
         *m44*m66*RN1p*RSp451e*RS156pe + 8./81.*m11*m44*m55*RNep*
         RSp561e*RSe561p - 2./27.*m11*m44*m55*RNep*RSp561e*RSe461p - 2./
         27.*m11*m44*m55*RNep*RSp461e*RSe561p + 8./81.*m11*m44*m55*RNep
         *RSp461e*RSe461p - 2./27.*m11*m44*m55*RN1p*RSp561e*RS146pe + 8.
         /81.*m11*m44*m55*RN1p*RSp561e*RS156pe + 8./81.*m11*m44*m55*
         RN1p*RSp461e*RS146pe - 2./27.*m11*m44*m55*RN1p*RSp461e*RS156pe
          + 2./81.*m11*pow(m44,2)*RNep*RSp561e*RSe561p + 2./27.*m11*
         pow(m44,2)*RNep*RSp561e*RSe561p*RS456pe - 1./27.*m11*pow(
         m44,2)*RNep*RSp561e*RSe461p - 1./27.*m11*pow(m44,2)*RNep*
         RSp561e*RSe451p - 1./27.*m11*pow(m44,2)*RNep*RSp461e*RSe561p
          - 1./27.*m11*pow(m44,2)*RNep*RSp451e*RSe561p );

      M12p6x13R +=  + Abeps(mp2,mu2)*pi16 * ( 1./9.*m11*pow(m44,2)*RN1p
         *RSp561e - 1./27.*m11*pow(m44,2)*RN1p*RSp561e*RS145pe - 1./27.
         *m11*pow(m44,2)*RN1p*RSp561e*RS146pe + 2./81.*m11*pow(m44,2)*
         RN1p*RSp561e*RS156pe + 2./27.*m11*pow(m44,2)*RN1p*RSp561e*
         RS456pe*RS156pe - 1./27.*m11*pow(m44,2)*RN1p*RSp461e*RS156pe
          - 1./27.*m11*pow(m44,2)*RN1p*RSp451e*RS156pe + 17./45.*pow(
         m11,2)*RSp45611e + 1./3.*pow(m11,2)*RSp45611e*RC1456pe + 2./9.
         *pow(m11,2)*RSp45611e*RSe45611p + 2./9.*pow(m11,2)*pow(
         RSp45611e,2) - 4./45.*pow(m11,2)*RSp561e - 4./45.*pow(m11,2)*
         RSp461e - 4./45.*pow(m11,2)*RSp451e - 2./9.*pow(m11,2)*RN1p*
         RSp45611e*RD1456ep + 2./3.*pow(m11,2)*mp2*RN1p*RSp45611e - 2./
         9.*pow(m11,2)*mp2*RN1p*pow(RSp45611e,2) + 1./9.*pow(m11,2)*mp2
         *RN1p*RSp561e + 1./9.*pow(m11,2)*mp2*RN1p*RSp461e + 1./9.*pow(
         m11,2)*mp2*RN1p*RSp451e - 2./9.*pow(m11,2)*me2*RN1p*RSp45611e*
         RSe45611p - 8./27.*pow(m11,2)*m66*RN1p*RSp45611e + 2./9.*pow(
         m11,2)*m66*RN1p*RSp451e );

      M12p6x13R +=  + Abeps(mp2,mu2)*pi16 * (  - 8./27.*pow(m11,2)*m55*
         RN1p*RSp45611e + 2./9.*pow(m11,2)*m55*RN1p*RSp461e - 8./27.*
         pow(m11,2)*m44*RN1p*RSp45611e + 2./9.*pow(m11,2)*m44*RN1p*
         RSp561e - 2./9.*pow(m11,3)*RN1p*RSp45611e + 2./9.*pow(m11,3)*
         RN1p*RSp45611e*RS1456pee + 2./9.*pow(m11,3)*RN1p*RSp45611e*
         RS1456ppe - 2./9.*pow(m11,3)*RN1p*RS145pe*RSp45611e - 2./9.*
         pow(m11,3)*RN1p*RS146pe*RSp45611e - 2./9.*pow(m11,3)*RN1p*
         RS156pe*RSp45611e + 1./9.*pow(m11,3)*RN1p*RSp561e + 1./9.*pow(
         m11,3)*RN1p*RSp461e + 1./9.*pow(m11,3)*RN1p*RSp451e );

      M12p6x13R +=  + Bbeps(m11,mu2)*pi16 * ( 1./9.*m11*pow(RD1456ep,2)
          + 1./9.*m11*mp2*RSp45611e*RD1456ep + 1./9.*m11*me2*RSe45611p*
         RD1456ep - 2./27.*m11*m66*RD1456ep - 1./27.*m11*m66*mp2*
         RSp456e*pow(RS145pe,2) - 1./27.*m11*m66*me2*RSe456p*pow(
         RS145pe,2) - 1./9.*m11*pow(m66,2)*RS145pe - 1./81.*m11*pow(
         m66,2)*pow(RS145pe,2) + 1./27.*m11*pow(m66,2)*RS146pe*RS145pe
          + 1./27.*m11*pow(m66,2)*RS156pe*RS145pe - 1./27.*m11*pow(
         m66,2)*RS645pe*pow(RS145pe,2) - 2./27.*m11*m55*RD1456ep - 1./
         27.*m11*m55*mp2*RSp465e*pow(RS146pe,2) - 1./27.*m11*m55*me2*
         RSe465p*pow(RS146pe,2) - 4./81.*m11*m55*m66*pow(RS145pe,2) + 2.
         /27.*m11*m55*m66*RS146pe*RS145pe - 4./81.*m11*m55*m66*pow(
         RS146pe,2) - 1./9.*m11*pow(m55,2)*RS146pe + 1./27.*m11*pow(
         m55,2)*RS146pe*RS145pe - 1./81.*m11*pow(m55,2)*pow(RS146pe,2)
          + 1./27.*m11*pow(m55,2)*RS156pe*RS146pe - 1./27.*m11*pow(
         m55,2)*RS546pe*pow(RS146pe,2) - 2./27.*m11*m44*RD1456ep - 1./
         27.*m11*m44*mp2*RSp564e*pow(RS156pe,2) );

      M12p6x13R +=  + Bbeps(m11,mu2)*pi16 * (  - 1./27.*m11*m44*me2*
         RSe564p*pow(RS156pe,2) - 4./81.*m11*m44*m66*pow(RS145pe,2) + 2.
         /27.*m11*m44*m66*RS156pe*RS145pe - 4./81.*m11*m44*m66*pow(
         RS156pe,2) - 4./81.*m11*m44*m55*pow(RS146pe,2) + 2./27.*m11*
         m44*m55*RS156pe*RS146pe - 4./81.*m11*m44*m55*pow(RS156pe,2) - 
         1./9.*m11*pow(m44,2)*RS156pe + 1./27.*m11*pow(m44,2)*RS156pe*
         RS145pe + 1./27.*m11*pow(m44,2)*RS156pe*RS146pe - 1./81.*m11*
         pow(m44,2)*pow(RS156pe,2) - 1./27.*m11*pow(m44,2)*RS456pe*pow(
         RS156pe,2) + 1./36.*pow(m11,2)*RD1456ep - 5./9.*pow(m11,2)*
         RS1456pee*RD1456ep - 5./9.*pow(m11,2)*RS1456ppe*RD1456ep + 2./
         9.*pow(m11,2)*RSe45611p*RD1456ep + 2./9.*pow(m11,2)*RSp45611e*
         RD1456ep + 5./9.*pow(m11,2)*RS145pe*RD1456ep + 5./9.*pow(
         m11,2)*RS146pe*RD1456ep + 5./9.*pow(m11,2)*RS156pe*RD1456ep - 
         2./9.*pow(m11,2)*mp2*RSp45611e*RS1456pee - 2./9.*pow(m11,2)*
         mp2*RSp45611e*RS1456ppe + 2./9.*pow(m11,2)*mp2*RS145pe*
         RSp45611e );

      M12p6x13R +=  + Bbeps(m11,mu2)*pi16 * ( 2./9.*pow(m11,2)*mp2*
         RS146pe*RSp45611e + 2./9.*pow(m11,2)*mp2*RS156pe*RSp45611e - 2.
         /9.*pow(m11,2)*me2*RSe45611p*RS1456pee - 2./9.*pow(m11,2)*me2*
         RSe45611p*RS1456ppe + 2./9.*pow(m11,2)*me2*RS145pe*RSe45611p
          + 2./9.*pow(m11,2)*me2*RS146pe*RSe45611p + 2./9.*pow(m11,2)*
         me2*RS156pe*RSe45611p - 2./27.*pow(m11,2)*m66*RS1456pee - 2./
         27.*pow(m11,2)*m66*RS1456ppe - 7./27.*pow(m11,2)*m66*RS145pe
          + 5./27.*pow(m11,2)*m66*pow(RS145pe,2) + 2./27.*pow(m11,2)*
         m66*RS146pe + 1./27.*pow(m11,2)*m66*RS146pe*RS145pe + 1./27.*
         pow(m11,2)*m66*pow(RS146pe,2) + 2./27.*pow(m11,2)*m66*RS156pe
          + 1./27.*pow(m11,2)*m66*RS156pe*RS145pe + 1./27.*pow(m11,2)*
         m66*pow(RS156pe,2) - 2./27.*pow(m11,2)*m55*RS1456pee - 2./27.*
         pow(m11,2)*m55*RS1456ppe + 2./27.*pow(m11,2)*m55*RS145pe + 1./
         27.*pow(m11,2)*m55*pow(RS145pe,2) - 7./27.*pow(m11,2)*m55*
         RS146pe + 1./27.*pow(m11,2)*m55*RS146pe*RS145pe + 5./27.*pow(
         m11,2)*m55*pow(RS146pe,2) );

      M12p6x13R +=  + Bbeps(m11,mu2)*pi16 * ( 2./27.*pow(m11,2)*m55*
         RS156pe + 1./27.*pow(m11,2)*m55*RS156pe*RS146pe + 1./27.*pow(
         m11,2)*m55*pow(RS156pe,2) - 2./27.*pow(m11,2)*m44*RS1456pee - 
         2./27.*pow(m11,2)*m44*RS1456ppe + 2./27.*pow(m11,2)*m44*
         RS145pe + 1./27.*pow(m11,2)*m44*pow(RS145pe,2) + 2./27.*pow(
         m11,2)*m44*RS146pe + 1./27.*pow(m11,2)*m44*pow(RS146pe,2) - 7./
         27.*pow(m11,2)*m44*RS156pe + 1./27.*pow(m11,2)*m44*RS156pe*
         RS145pe + 1./27.*pow(m11,2)*m44*RS156pe*RS146pe + 5./27.*pow(
         m11,2)*m44*pow(RS156pe,2) + pow(m11,3) + 4./9.*pow(m11,3)*
         RS1456pee + 2./9.*pow(m11,3)*pow(RS1456pee,2) + 4./9.*pow(
         m11,3)*RS1456ppe + 4./9.*pow(m11,3)*RS1456ppe*RS1456pee + 2./9.
         *pow(m11,3)*pow(RS1456ppe,2) - 2./3.*pow(m11,3)*RS145pe - 4./9.
         *pow(m11,3)*RS145pe*RS1456pee - 4./9.*pow(m11,3)*RS145pe*
         RS1456ppe + 2./9.*pow(m11,3)*pow(RS145pe,2) - 2./3.*pow(m11,3)
         *RS146pe - 4./9.*pow(m11,3)*RS146pe*RS1456pee - 4./9.*pow(
         m11,3)*RS146pe*RS1456ppe );

      M12p6x13R +=  + Bbeps(m11,mu2)*pi16 * ( 4./9.*pow(m11,3)*RS146pe*
         RS145pe + 2./9.*pow(m11,3)*pow(RS146pe,2) - 2./3.*pow(m11,3)*
         RS156pe - 4./9.*pow(m11,3)*RS156pe*RS1456pee - 4./9.*pow(
         m11,3)*RS156pe*RS1456ppe + 4./9.*pow(m11,3)*RS156pe*RS145pe + 
         4./9.*pow(m11,3)*RS156pe*RS146pe + 2./9.*pow(m11,3)*pow(
         RS156pe,2) );

      M12p6x13R +=  + Bbeps(me2,mu2)*pi16 * (  - 1./27.*m11*m66*mp2*
         RSp456e*pow(RSe451p,2) + 1./27.*m11*m66*me2*pow(RSe561p,2) + 1.
         /27.*m11*m66*me2*pow(RSe461p,2) + 1./27.*m11*m66*me2*RSe451p*
         RSe561p + 1./27.*m11*m66*me2*RSe451p*RSe461p + 5./27.*m11*m66*
         me2*pow(RSe451p,2) - 1./27.*m11*m66*me2*RSe456p*pow(RSe451p,2)
          + 1./27.*m11*pow(m66,2)*RSe451p*RSe561p + 1./27.*m11*pow(
         m66,2)*RSe451p*RSe461p - 1./81.*m11*pow(m66,2)*pow(RSe451p,2)
          - 1./27.*m11*pow(m66,2)*pow(RSe451p,2)*RS645pe - 1./27.*m11*
         m55*mp2*RSp465e*pow(RSe461p,2) + 1./27.*m11*m55*me2*pow(
         RSe561p,2) + 1./27.*m11*m55*me2*RSe461p*RSe561p + 5./27.*m11*
         m55*me2*pow(RSe461p,2) + 1./27.*m11*m55*me2*RSe451p*RSe461p + 
         1./27.*m11*m55*me2*pow(RSe451p,2) - 1./27.*m11*m55*me2*RSe465p
         *pow(RSe461p,2) - 4./81.*m11*m55*m66*pow(RSe461p,2) + 2./27.*
         m11*m55*m66*RSe451p*RSe461p - 4./81.*m11*m55*m66*pow(
         RSe451p,2) + 1./27.*m11*pow(m55,2)*RSe461p*RSe561p - 1./81.*
         m11*pow(m55,2)*pow(RSe461p,2) );

      M12p6x13R +=  + Bbeps(me2,mu2)*pi16 * (  - 1./27.*m11*pow(m55,2)*
         pow(RSe461p,2)*RS546pe + 1./27.*m11*pow(m55,2)*RSe451p*RSe461p
          - 1./27.*m11*m44*mp2*RSp564e*pow(RSe561p,2) + 5./27.*m11*m44*
         me2*pow(RSe561p,2) + 1./27.*m11*m44*me2*RSe461p*RSe561p + 1./
         27.*m11*m44*me2*pow(RSe461p,2) + 1./27.*m11*m44*me2*RSe451p*
         RSe561p + 1./27.*m11*m44*me2*pow(RSe451p,2) - 1./27.*m11*m44*
         me2*RSe564p*pow(RSe561p,2) - 4./81.*m11*m44*m66*pow(RSe561p,2)
          + 2./27.*m11*m44*m66*RSe451p*RSe561p - 4./81.*m11*m44*m66*
         pow(RSe451p,2) - 4./81.*m11*m44*m55*pow(RSe561p,2) + 2./27.*
         m11*m44*m55*RSe461p*RSe561p - 4./81.*m11*m44*m55*pow(
         RSe461p,2) - 1./81.*m11*pow(m44,2)*pow(RSe561p,2) - 1./27.*m11
         *pow(m44,2)*pow(RSe561p,2)*RS456pe + 1./27.*m11*pow(m44,2)*
         RSe461p*RSe561p + 1./27.*m11*pow(m44,2)*RSe451p*RSe561p );

      M12p6x13R +=  + Bbeps(mp2,mu2)*pi16 * ( 1./27.*m11*m66*mp2*pow(
         RSp561e,2) + 1./27.*m11*m66*mp2*pow(RSp461e,2) + 1./27.*m11*
         m66*mp2*RSp451e*RSp561e + 1./27.*m11*m66*mp2*RSp451e*RSp461e
          + 5./27.*m11*m66*mp2*pow(RSp451e,2) - 1./27.*m11*m66*mp2*
         RSp456e*pow(RSp451e,2) - 1./27.*m11*m66*me2*RSe456p*pow(
         RSp451e,2) + 1./27.*m11*pow(m66,2)*RSp451e*RSp561e + 1./27.*
         m11*pow(m66,2)*RSp451e*RSp461e - 1./81.*m11*pow(m66,2)*pow(
         RSp451e,2) - 1./27.*m11*pow(m66,2)*pow(RSp451e,2)*RS645pe + 1./
         27.*m11*m55*mp2*pow(RSp561e,2) + 1./27.*m11*m55*mp2*RSp461e*
         RSp561e + 5./27.*m11*m55*mp2*pow(RSp461e,2) + 1./27.*m11*m55*
         mp2*RSp451e*RSp461e + 1./27.*m11*m55*mp2*pow(RSp451e,2) - 1./
         27.*m11*m55*mp2*RSp465e*pow(RSp461e,2) - 1./27.*m11*m55*me2*
         RSe465p*pow(RSp461e,2) - 4./81.*m11*m55*m66*pow(RSp461e,2) + 2.
         /27.*m11*m55*m66*RSp451e*RSp461e - 4./81.*m11*m55*m66*pow(
         RSp451e,2) + 1./27.*m11*pow(m55,2)*RSp461e*RSp561e - 1./81.*
         m11*pow(m55,2)*pow(RSp461e,2) );

      M12p6x13R +=  + Bbeps(mp2,mu2)*pi16 * (  - 1./27.*m11*pow(m55,2)*
         pow(RSp461e,2)*RS546pe + 1./27.*m11*pow(m55,2)*RSp451e*RSp461e
          + 5./27.*m11*m44*mp2*pow(RSp561e,2) + 1./27.*m11*m44*mp2*
         RSp461e*RSp561e + 1./27.*m11*m44*mp2*pow(RSp461e,2) + 1./27.*
         m11*m44*mp2*RSp451e*RSp561e + 1./27.*m11*m44*mp2*pow(
         RSp451e,2) - 1./27.*m11*m44*mp2*RSp564e*pow(RSp561e,2) - 1./27.
         *m11*m44*me2*RSe564p*pow(RSp561e,2) - 4./81.*m11*m44*m66*pow(
         RSp561e,2) + 2./27.*m11*m44*m66*RSp451e*RSp561e - 4./81.*m11*
         m44*m66*pow(RSp451e,2) - 4./81.*m11*m44*m55*pow(RSp561e,2) + 2.
         /27.*m11*m44*m55*RSp461e*RSp561e - 4./81.*m11*m44*m55*pow(
         RSp461e,2) - 1./81.*m11*pow(m44,2)*pow(RSp561e,2) - 1./27.*m11
         *pow(m44,2)*pow(RSp561e,2)*RS456pe + 1./27.*m11*pow(m44,2)*
         RSp461e*RSp561e + 1./27.*m11*pow(m44,2)*RSp451e*RSp561e );

      M12p6x13R +=  + hh(1,m11,m11,m11,m11,mu2) * ( 1./3.*pow(m11,2) + 
         2./9.*pow(m11,2)*pow(RC1456pe,2) );

      M12p6x13R +=  + hh(1,m11,m11,me2,m11,mu2) * ( 4./9.*pow(m11,2)*
         RSe45611p*RC1456pe );

      M12p6x13R +=  + hh(1,m11,m11,mp2,m11,mu2) * ( 4./9.*pow(m11,2)*
         RSp45611e*RC1456pe );

      M12p6x13R +=  + hh(1,m11,m14,m14,m11,mu2) * (  - 11./36.*m11*
         RD1456ep - 2./9.*m11*m44*RS156pe - 5./12.*pow(m11,2)*RC1456pe
          + 7./18.*pow(m11,2)*RS156pe );

      M12p6x13R +=  + hh(1,m11,m15,m15,m11,mu2) * (  - 11./36.*m11*
         RD1456ep - 2./9.*m11*m55*RS146pe - 5./12.*pow(m11,2)*RC1456pe
          + 7./18.*pow(m11,2)*RS146pe );

      M12p6x13R +=  + hh(1,m11,m16,m16,m11,mu2) * (  - 11./36.*m11*
         RD1456ep - 2./9.*m11*m66*RS145pe - 5./12.*pow(m11,2)*RC1456pe
          + 7./18.*pow(m11,2)*RS145pe );

      M12p6x13R +=  + hh(1,m11,me2,me2,m11,mu2) * ( 2./9.*pow(m11,2)*
         pow(RSe45611p,2) );

      M12p6x13R +=  + hh(1,m11,me2,mp2,m11,mu2) * ( 4./9.*pow(m11,2)*
         RSp45611e*RSe45611p );

      M12p6x13R +=  + hh(1,m11,mp2,mp2,m11,mu2) * ( 2./9.*pow(m11,2)*
         pow(RSp45611e,2) );

      M12p6x13R +=  + hh(1,m14,m14,m44,m11,mu2) * ( 1./4.*m11*m44 - 1./
         12.*m11*m44*RS456pe );

      M12p6x13R +=  + hh(1,m14,m14,me2,m11,mu2) * (  - 1./27.*pow(
         me2,2)*RSe45611p + 2./27.*pow(me2,2)*RSe561p - 1./27.*pow(
         me2,2)*RSe564p - 2./27.*m44*me2*RSe561p + 2./27.*m44*me2*
         RSe564p - 1./27.*pow(m44,2)*RSe564p - 1./108.*m11*me2*
         RSe45611p - 7./54.*m11*me2*RSe561p + 5./36.*m11*me2*RSe564p + 
         2./27.*m11*m44*RSe561p - 2./9.*m11*m44*RSe564p + 8./27.*pow(
         m11,2)*RSe45611p - 4./9.*pow(m11,2)*RSe561p );

      M12p6x13R +=  + hh(1,m14,m14,mp2,m11,mu2) * (  - 1./27.*pow(
         mp2,2)*RSp45611e + 2./27.*pow(mp2,2)*RSp561e - 1./27.*pow(
         mp2,2)*RSp564e - 2./27.*m44*mp2*RSp561e + 2./27.*m44*mp2*
         RSp564e - 1./27.*pow(m44,2)*RSp564e - 1./108.*m11*mp2*
         RSp45611e - 7./54.*m11*mp2*RSp561e + 5./36.*m11*mp2*RSp564e + 
         2./27.*m11*m44*RSp561e - 2./9.*m11*m44*RSp564e + 8./27.*pow(
         m11,2)*RSp45611e - 4./9.*pow(m11,2)*RSp561e );

      M12p6x13R +=  + hh(1,m14,m15,m45,m11,mu2) * ( 1./4.*m11*m55 + 1./
         4.*m11*m44 );

      M12p6x13R +=  + hh(1,m14,m16,m46,m11,mu2) * ( 1./4.*m11*m66 + 1./
         4.*m11*m44 );

      M12p6x13R +=  + hh(1,m15,m15,m55,m11,mu2) * ( 1./4.*m11*m55 - 1./
         12.*m11*m55*RS546pe );

      M12p6x13R +=  + hh(1,m15,m15,me2,m11,mu2) * (  - 1./27.*pow(
         me2,2)*RSe45611p + 2./27.*pow(me2,2)*RSe461p - 1./27.*pow(
         me2,2)*RSe465p - 2./27.*m55*me2*RSe461p + 2./27.*m55*me2*
         RSe465p - 1./27.*pow(m55,2)*RSe465p - 1./108.*m11*me2*
         RSe45611p - 7./54.*m11*me2*RSe461p + 5./36.*m11*me2*RSe465p + 
         2./27.*m11*m55*RSe461p - 2./9.*m11*m55*RSe465p + 8./27.*pow(
         m11,2)*RSe45611p - 4./9.*pow(m11,2)*RSe461p );

      M12p6x13R +=  + hh(1,m15,m15,mp2,m11,mu2) * (  - 1./27.*pow(
         mp2,2)*RSp45611e + 2./27.*pow(mp2,2)*RSp461e - 1./27.*pow(
         mp2,2)*RSp465e - 2./27.*m55*mp2*RSp461e + 2./27.*m55*mp2*
         RSp465e - 1./27.*pow(m55,2)*RSp465e - 1./108.*m11*mp2*
         RSp45611e - 7./54.*m11*mp2*RSp461e + 5./36.*m11*mp2*RSp465e + 
         2./27.*m11*m55*RSp461e - 2./9.*m11*m55*RSp465e + 8./27.*pow(
         m11,2)*RSp45611e - 4./9.*pow(m11,2)*RSp461e );

      M12p6x13R +=  + hh(1,m15,m16,m56,m11,mu2) * ( 1./4.*m11*m66 + 1./
         4.*m11*m55 );

      M12p6x13R +=  + hh(1,m16,m16,m66,m11,mu2) * ( 1./4.*m11*m66 - 1./
         12.*m11*m66*RS645pe );

      M12p6x13R +=  + hh(1,m16,m16,me2,m11,mu2) * (  - 1./27.*pow(
         me2,2)*RSe45611p + 2./27.*pow(me2,2)*RSe451p - 1./27.*pow(
         me2,2)*RSe456p - 2./27.*m66*me2*RSe451p + 2./27.*m66*me2*
         RSe456p - 1./27.*pow(m66,2)*RSe456p - 1./108.*m11*me2*
         RSe45611p - 7./54.*m11*me2*RSe451p + 5./36.*m11*me2*RSe456p + 
         2./27.*m11*m66*RSe451p - 2./9.*m11*m66*RSe456p + 8./27.*pow(
         m11,2)*RSe45611p - 4./9.*pow(m11,2)*RSe451p );

      M12p6x13R +=  + hh(1,m16,m16,mp2,m11,mu2) * (  - 1./27.*pow(
         mp2,2)*RSp45611e + 2./27.*pow(mp2,2)*RSp451e - 1./27.*pow(
         mp2,2)*RSp456e - 2./27.*m66*mp2*RSp451e + 2./27.*m66*mp2*
         RSp456e - 1./27.*pow(m66,2)*RSp456e - 1./108.*m11*mp2*
         RSp45611e - 7./54.*m11*mp2*RSp451e + 5./36.*m11*mp2*RSp456e + 
         2./27.*m11*m66*RSp451e - 2./9.*m11*m66*RSp456e + 8./27.*pow(
         m11,2)*RSp45611e - 4./9.*pow(m11,2)*RSp451e );

      M12p6x13R +=  + hh(2,m11,m11,m11,m11,mu2) * ( 4./9.*pow(m11,2)*
         RD1456ep*RC1456pe );

      M12p6x13R +=  + hh(2,m11,m11,me2,m11,mu2) * ( 4./9.*pow(m11,2)*
         RSe45611p*RD1456ep );

      M12p6x13R +=  + hh(2,m11,m11,mp2,m11,mu2) * ( 4./9.*pow(m11,2)*
         RSp45611e*RD1456ep );

      M12p6x13R +=  + hh(2,m11,m14,m14,m11,mu2) * ( 1./4.*pow(m11,2)*
         RD1456ep );

      M12p6x13R +=  + hh(2,m11,m15,m15,m11,mu2) * ( 1./4.*pow(m11,2)*
         RD1456ep );

      M12p6x13R +=  + hh(2,m11,m16,m16,m11,mu2) * ( 1./4.*pow(m11,2)*
         RD1456ep );

      M12p6x13R +=  + hh(5,m11,m11,m11,m11,mu2) * ( 2./9.*pow(m11,2)*
         pow(RD1456ep,2) );

      M12p6x13R +=  + hh1(1,m11,m14,m14,m11,mu2) * ( 2./9.*m11*RD1456ep
          + 2./9.*m11*m44*RS156pe + 2./3.*pow(m11,2)*RC1456pe - 8./9.*
         pow(m11,2)*RS156pe );

      M12p6x13R +=  + hh1(1,m11,m15,m15,m11,mu2) * ( 2./9.*m11*RD1456ep
          + 2./9.*m11*m55*RS146pe + 2./3.*pow(m11,2)*RC1456pe - 8./9.*
         pow(m11,2)*RS146pe );

      M12p6x13R +=  + hh1(1,m11,m16,m16,m11,mu2) * ( 2./9.*m11*RD1456ep
          + 2./9.*m11*m66*RS145pe + 2./3.*pow(m11,2)*RC1456pe - 8./9.*
         pow(m11,2)*RS145pe );

      M12p6x13R +=  + hh1(1,m14,m14,me2,m11,mu2) * (  - 4./9.*m11*me2*
         RSe45611p + 8./9.*m11*me2*RSe561p - 4./9.*m11*me2*RSe564p - 4./
         9.*m11*m44*RSe561p + 4./9.*m11*m44*RSe564p - 8./9.*pow(m11,2)*
         RSe45611p + 8./9.*pow(m11,2)*RSe561p );

      M12p6x13R +=  + hh1(1,m14,m14,mp2,m11,mu2) * (  - 4./9.*m11*mp2*
         RSp45611e + 8./9.*m11*mp2*RSp561e - 4./9.*m11*mp2*RSp564e - 4./
         9.*m11*m44*RSp561e + 4./9.*m11*m44*RSp564e - 8./9.*pow(m11,2)*
         RSp45611e + 8./9.*pow(m11,2)*RSp561e );

      M12p6x13R +=  + hh1(1,m15,m15,me2,m11,mu2) * (  - 4./9.*m11*me2*
         RSe45611p + 8./9.*m11*me2*RSe461p - 4./9.*m11*me2*RSe465p - 4./
         9.*m11*m55*RSe461p + 4./9.*m11*m55*RSe465p - 8./9.*pow(m11,2)*
         RSe45611p + 8./9.*pow(m11,2)*RSe461p );

      M12p6x13R +=  + hh1(1,m15,m15,mp2,m11,mu2) * (  - 4./9.*m11*mp2*
         RSp45611e + 8./9.*m11*mp2*RSp461e - 4./9.*m11*mp2*RSp465e - 4./
         9.*m11*m55*RSp461e + 4./9.*m11*m55*RSp465e - 8./9.*pow(m11,2)*
         RSp45611e + 8./9.*pow(m11,2)*RSp461e );

      M12p6x13R +=  + hh1(1,m16,m16,me2,m11,mu2) * (  - 4./9.*m11*me2*
         RSe45611p + 8./9.*m11*me2*RSe451p - 4./9.*m11*me2*RSe456p - 4./
         9.*m11*m66*RSe451p + 4./9.*m11*m66*RSe456p - 8./9.*pow(m11,2)*
         RSe45611p + 8./9.*pow(m11,2)*RSe451p );

      M12p6x13R +=  + hh1(1,m16,m16,mp2,m11,mu2) * (  - 4./9.*m11*mp2*
         RSp45611e + 8./9.*m11*mp2*RSp451e - 4./9.*m11*mp2*RSp456e - 4./
         9.*m11*m66*RSp451e + 4./9.*m11*m66*RSp456e - 8./9.*pow(m11,2)*
         RSp45611e + 8./9.*pow(m11,2)*RSp451e );

      M12p6x13R +=  + hh1(3,m14,m11,m14,m11,mu2) * (  - 4./3.*pow(
         m11,2)*RD1456ep );

      M12p6x13R +=  + hh1(3,m15,m11,m15,m11,mu2) * (  - 4./3.*pow(
         m11,2)*RD1456ep );

      M12p6x13R +=  + hh1(3,m16,m11,m16,m11,mu2) * (  - 4./3.*pow(
         m11,2)*RD1456ep );

      M12p6x13R +=  + hh21(1,m11,m14,m14,m11,mu2) * (  - 1./4.*pow(
         m11,2)*RC1456pe + 1./2.*pow(m11,2)*RS156pe );

      M12p6x13R +=  + hh21(1,m11,m15,m15,m11,mu2) * (  - 1./4.*pow(
         m11,2)*RC1456pe + 1./2.*pow(m11,2)*RS146pe );

      M12p6x13R +=  + hh21(1,m11,m16,m16,m11,mu2) * (  - 1./4.*pow(
         m11,2)*RC1456pe + 1./2.*pow(m11,2)*RS145pe );

      M12p6x13R +=  + hh21(1,m44,m14,m14,m11,mu2) * ( 3./4.*pow(m11,2)
          - 1./4.*pow(m11,2)*RS456pe );

      M12p6x13R +=  + hh21(1,m45,m14,m15,m11,mu2) * ( 3./2.*pow(m11,2)
          );

      M12p6x13R +=  + hh21(1,m46,m14,m16,m11,mu2) * ( 3./2.*pow(m11,2)
          );

      M12p6x13R +=  + hh21(1,m55,m15,m15,m11,mu2) * ( 3./4.*pow(m11,2)
          - 1./4.*pow(m11,2)*RS546pe );

      M12p6x13R +=  + hh21(1,m56,m15,m16,m11,mu2) * ( 3./2.*pow(m11,2)
          );

      M12p6x13R +=  + hh21(1,m66,m16,m16,m11,mu2) * ( 3./4.*pow(m11,2)
          - 1./4.*pow(m11,2)*RS645pe );

      M12p6x13R +=  + hh21(1,me2,m14,m14,m11,mu2) * (  - 1./4.*pow(
         m11,2)*RSe45611p + 1./2.*pow(m11,2)*RSe561p - 1./4.*pow(m11,2)
         *RSe564p );

      M12p6x13R +=  + hh21(1,me2,m15,m15,m11,mu2) * (  - 1./4.*pow(
         m11,2)*RSe45611p + 1./2.*pow(m11,2)*RSe461p - 1./4.*pow(m11,2)
         *RSe465p );

      M12p6x13R +=  + hh21(1,me2,m16,m16,m11,mu2) * (  - 1./4.*pow(
         m11,2)*RSe45611p + 1./2.*pow(m11,2)*RSe451p - 1./4.*pow(m11,2)
         *RSe456p );

      M12p6x13R +=  + hh21(1,mp2,m14,m14,m11,mu2) * (  - 1./4.*pow(
         m11,2)*RSp45611e + 1./2.*pow(m11,2)*RSp561e - 1./4.*pow(m11,2)
         *RSp564e );

      M12p6x13R +=  + hh21(1,mp2,m15,m15,m11,mu2) * (  - 1./4.*pow(
         m11,2)*RSp45611e + 1./2.*pow(m11,2)*RSp461e - 1./4.*pow(m11,2)
         *RSp465e );

      M12p6x13R +=  + hh21(1,mp2,m16,m16,m11,mu2) * (  - 1./4.*pow(
         m11,2)*RSp45611e + 1./2.*pow(m11,2)*RSp451e - 1./4.*pow(m11,2)
         *RSp456e );

      M12p6x13R +=  + hh21(2,m11,m14,m14,m11,mu2) * (  - 1./4.*pow(
         m11,2)*RD1456ep );

      M12p6x13R +=  + hh21(2,m11,m15,m15,m11,mu2) * (  - 1./4.*pow(
         m11,2)*RD1456ep );

      M12p6x13R +=  + hh21(2,m11,m16,m16,m11,mu2) * (  - 1./4.*pow(
         m11,2)*RD1456ep );


  return M12p6x13R/pow(f0,4);
}
// 2 valence mass 3 sea mass *****************************************

double mv2s3nf3p4(const quarkmassnf mass, const Linf Liin){
  return mv2s3nf3p4L(mass,Liin)+mv2s3nf3p4R(mass);
}

double mv2s3nf3p4L(const quarkmassnf mass, const Linf Liin){
  double L0r,L1r,L2r,L3r,L4r,L5r,L6r,L7r,L8r,L9r,L10r,L11r,H1r,H2r,mu;
  int nfLi;
  Liin.out(L0r,L1r,L2r,L3r,L4r,L5r,L6r,L7r,L8r,L9r,L10r,L11r,H1r,H2r,mu,nfLi);
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if (mu0 != mu){
    cout << "ERROR: differing mu's in mv2s3nf3p4L\n";}
  if ( (nq != 5) || (B0mq.size() != 5)){
    cout << "ERROR: mv2s3nf3p4L probably called with wrong number"
	 << "of quark masses\n";}
  // two masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m33 = 2.*B0mq[1];
  double m44 = 2.*B0mq[2];
  double m55 = 2.*B0mq[3];
  double m66 = 2.*B0mq[4];

  double M13p4x23L =
       + 8*m33*m66*L6r - 4*m33*m66*L4r + 8*m33*m55*L6r - 4*m33*m55*L4r
          + 8*m33*m44*L6r - 4*m33*m44*L4r + 4*pow(m33,2)*L8r - 2*pow(
         m33,2)*L5r + 8*m11*m66*L6r - 4*m11*m66*L4r + 8*m11*m55*L6r - 4
         *m11*m55*L4r + 8*m11*m44*L6r - 4*m11*m44*L4r + 8*m11*m33*L8r
          - 4*m11*m33*L5r + 4*pow(m11,2)*L8r - 2*pow(m11,2)*L5r;

  return M13p4x23L/pow(f0,2);
}


double mv2s3nf3p4R(const quarkmassnf mass){
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if ( (nq != 5) || (B0mq.size() != 5)){
    cout << "ERROR: mv2s3nf3p4R probably called with wrong number"
	 << "of quark masses\n";}
  // five masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m33 = 2.*B0mq[1];
  double m44 = 2.*B0mq[2];
  double m55 = 2.*B0mq[3];
  double m66 = 2.*B0mq[4];
  double xb1 = (m44+m55+m66)/3.;
  double msqr = sqrt(xb1*xb1-(m44*m55+m44*m66+m55*m66)/3.);
  double mp2 = xb1-msqr; 
  double me2 = xb1+msqr;
  double mu2 = mu0*mu0;
  double RSe45611p = RS(me2,m44,m55,m66,m11,m11,mp2);
  double RSp45611e = RS(mp2,m44,m55,m66,m11,m11,me2);
  double RSe45613p = RS(me2,m44,m55,m66,m11,m33,mp2);
  double RSp45613e = RS(mp2,m44,m55,m66,m11,m33,me2);
  double RSe45633p = RS(me2,m44,m55,m66,m33,m33,mp2);
  double RSp45633e = RS(mp2,m44,m55,m66,m33,m33,me2);
  double RS14563pe = RS(m11,m44,m55,m66,m33,mp2,me2);
  double RS34561pe = RS(m33,m44,m55,m66,m11,mp2,me2);

  double RD1456ep = RD(m11,m44,m55,m66,me2,mp2);
  double RD3456ep = RD(m33,m44,m55,m66,me2,mp2);

  double M13p4x23R =
       + Ab(m11,mu2) * ( 1./18.*RD1456ep - 1./9.*m33*RS14563pe - 2./9.*
         m11*RS14563pe );

      M13p4x23R +=  + Ab(m33,mu2) * ( 1./18.*RD3456ep - 2./9.*m33*
         RS34561pe - 1./9.*m11*RS34561pe );

      M13p4x23R +=  + Ab(me2,mu2) * (  - 1./9.*me2*RSe45613p + 1./18.*
         me2*RSe45633p + 1./18.*me2*RSe45611p - 1./9.*m33*RSe45613p - 1.
         /18.*m33*RSe45633p - 1./9.*m11*RSe45613p - 1./18.*m11*
         RSe45611p );

      M13p4x23R +=  + Ab(mp2,mu2) * (  - 1./9.*mp2*RSp45613e + 1./18.*
         mp2*RSp45633e + 1./18.*mp2*RSp45611e - 1./9.*m33*RSp45613e - 1.
         /18.*m33*RSp45633e - 1./9.*m11*RSp45613e - 1./18.*m11*
         RSp45611e );

  return M13p4x23R/pow(f0,2);
}

double mv2s3nf3p6(const quarkmassnf mass, const Linf Liin, const Ki Kiin){
  return mv2s3nf3p6L(mass,Liin)+mv2s3nf3p6K(mass,Kiin)+mv2s3nf3p6R(mass);
}

double mv2s3nf3p6L(const quarkmassnf mass, const Linf Liin){
  double L0r,L1r,L2r,L3r,L4r,L5r,L6r,L7r,L8r,L9r,L10r,L11r,H1r,H2r,mu;
  int nfLi;
  Liin.out(L0r,L1r,L2r,L3r,L4r,L5r,L6r,L7r,L8r,L9r,L10r,L11r,H1r,H2r,mu,nfLi);
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if (mu0 != mu){
    cout << "ERROR: differing mu's in mv2s3nf3p6L\n";}
  if ( (nq != 5) || (B0mq.size() != 5)){
    cout << "ERROR: mv2s3nf3p6L probably called with wrong number"
	 << "of quark masses\n";}
  // four masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m33 = 2.*B0mq[1];
  double m44 = 2.*B0mq[2];
  double m55 = 2.*B0mq[3];
  double m66 = 2.*B0mq[4];
  double m13 = (m11+m33)/2;
  double m14 = (m11+m44)/2;
  double m15 = (m11+m55)/2;
  double m16 = (m11+m66)/2;
  double m34 = (m33+m44)/2;
  double m35 = (m33+m55)/2;
  double m36 = (m33+m66)/2;
  double m45 = (m44+m55)/2;
  double m46 = (m44+m66)/2;
  double m56 = (m55+m66)/2;

  double xb1 = (m44+m55+m66)/3.;
  double msqr = sqrt(xb1*xb1-(m44*m55+m44*m66+m55*m66)/3.);
  double mp2 = xb1-msqr; 
  double me2 = xb1+msqr;
  double mu2 = mu0*mu0;

  double RN13 = RN(m11,m33);
  double RN1p = RN(m11,mp2);
  double RN1e = RN(m11,me2);
  double RN3p = RN(m33,mp2);
  double RN3e = RN(m33,me2);
  double RNep = RN(me2,mp2);

  double RSp456e = RS(mp2,m44,m55,m66,me2);
  double RSp465e = RS(mp2,m44,m66,m55,me2);
  double RSp564e = RS(mp2,m55,m66,m44,me2);
  double RSe456p = RS(me2,m44,m55,m66,mp2);
  double RSe465p = RS(me2,m44,m66,m55,mp2);
  double RSe564p = RS(me2,m55,m66,m44,mp2);
  double RSp451e = RS(mp2,m44,m55,m11,me2);
  double RSp461e = RS(mp2,m44,m66,m11,me2);
  double RSp561e = RS(mp2,m55,m66,m11,me2);
  double RSe451p = RS(me2,m44,m55,m11,mp2);
  double RSe461p = RS(me2,m44,m66,m11,mp2);
  double RSe561p = RS(me2,m55,m66,m11,mp2);
  double RSp453e = RS(mp2,m44,m55,m33,me2);
  double RSp463e = RS(mp2,m44,m66,m33,me2);
  double RSp563e = RS(mp2,m55,m66,m33,me2);
  double RSe453p = RS(me2,m44,m55,m33,mp2);
  double RSe463p = RS(me2,m44,m66,m33,mp2);
  double RSe563p = RS(me2,m55,m66,m33,mp2);
  double RS456pe = RS(m44,m55,m66,mp2,me2);
  double RS546pe = RS(m55,m44,m66,mp2,me2);
  double RS645pe = RS(m66,m44,m55,mp2,me2);

  double RS156pe = RS(m11,m55,m66,mp2,me2);
  double RS146pe = RS(m11,m44,m66,mp2,me2);
  double RS145pe = RS(m11,m44,m55,mp2,me2);
  double RS356pe = RS(m33,m55,m66,mp2,me2);
  double RS346pe = RS(m33,m44,m66,mp2,me2);
  double RS345pe = RS(m33,m44,m55,mp2,me2);

  double RD1456ep = RD(m11,m44,m55,m66,me2,mp2);
  double RD3456ep = RD(m33,m44,m55,m66,me2,mp2);

  double RS1456ppe = RS(m11,m44,m55,m66,mp2,mp2,me2);
  double RS1456pee = RS(m11,m44,m55,m66,mp2,me2,me2);
  double RSe45611p = RS(me2,m44,m55,m66,m11,m11,mp2);
  double RSp45611e = RS(mp2,m44,m55,m66,m11,m11,me2);
  double RSe45633p = RS(me2,m44,m55,m66,m33,m33,mp2);
  double RSp45633e = RS(mp2,m44,m55,m66,m33,m33,me2);
  double RS3456ppe = RS(m33,m44,m55,m66,mp2,mp2,me2);
  double RS3456pee = RS(m33,m44,m55,m66,mp2,me2,me2);
  double RSp45613e = RS(mp2,m44,m55,m66,m11,m33,me2);
  double RSe45613p = RS(me2,m44,m55,m66,m11,m33,mp2);
  double RS14563pe = RS(m11,m44,m55,m66,m33,mp2,me2);
  double RS34561pe = RS(m33,m44,m55,m66,m11,mp2,me2);

  double M13p6x23L =
       + pi16 * (  - 1./3.*m33*pow(mp2,2)*L3r*RSp45613e - 1./6.*m33*
         pow(mp2,2)*L3r*RSp45633e - 1./6.*m33*pow(mp2,2)*L3r*RSp45611e
          - 1./3.*m33*pow(mp2,2)*L2r*RSp564e - 1./3.*m33*pow(mp2,2)*L2r
         *RSp465e - 1./3.*m33*pow(mp2,2)*L2r*RSp456e - 1./3.*m33*pow(
         mp2,2)*L0r*RSp45633e - 1./3.*m33*pow(mp2,2)*L0r*RSp45611e - 1./
         3.*m33*pow(me2,2)*L3r*RSe45613p - 1./6.*m33*pow(me2,2)*L3r*
         RSe45633p - 1./6.*m33*pow(me2,2)*L3r*RSe45611p - 1./3.*m33*
         pow(me2,2)*L2r*RSe564p - 1./3.*m33*pow(me2,2)*L2r*RSe465p - 1./
         3.*m33*pow(me2,2)*L2r*RSe456p - 1./3.*m33*pow(me2,2)*L0r*
         RSe45633p - 1./3.*m33*pow(me2,2)*L0r*RSe45611p + 1./4.*m33*
         pow(m66,2)*L3r + 2*m33*pow(m66,2)*L2r - 1./3.*m33*pow(m66,2)*
         L2r*RS645pe + 1./2.*m33*pow(m66,2)*L0r + m33*m55*m66*L2r + 1./
         4.*m33*pow(m55,2)*L3r + 2*m33*pow(m55,2)*L2r - 1./3.*m33*pow(
         m55,2)*L2r*RS546pe + 1./2.*m33*pow(m55,2)*L0r + m33*m44*m66*
         L2r + m33*m44*m55*L2r + 1./4.*m33*pow(m44,2)*L3r + 2*m33*pow(
         m44,2)*L2r );

      M13p6x23L +=  + pi16 * (  - 1./3.*m33*pow(m44,2)*L2r*RS456pe + 1./
         2.*m33*pow(m44,2)*L0r - 1./3.*pow(m33,2)*L3r*RD3456ep - 2./3.*
         pow(m33,2)*L0r*RD3456ep + 1./4.*pow(m33,2)*m66*L3r + 1./2.*
         pow(m33,2)*m66*L0r + 1./4.*pow(m33,2)*m55*L3r + 1./2.*pow(
         m33,2)*m55*L0r + 1./4.*pow(m33,2)*m44*L3r + 1./2.*pow(m33,2)*
         m44*L0r + 3./8.*pow(m33,3)*L3r - 1./3.*pow(m33,3)*L3r*
         RS34561pe + 1./6.*pow(m33,3)*L3r*RS3456pee + 1./6.*pow(m33,3)*
         L3r*RS3456ppe - 1./6.*pow(m33,3)*L3r*RS345pe - 1./6.*pow(
         m33,3)*L3r*RS346pe - 1./6.*pow(m33,3)*L3r*RS356pe + 1./4.*pow(
         m33,3)*L2r + 1./2.*pow(m33,3)*L1r + 3./4.*pow(m33,3)*L0r + 1./
         3.*pow(m33,3)*L0r*RS3456pee + 1./3.*pow(m33,3)*L0r*RS3456ppe
          - 1./3.*pow(m33,3)*L0r*RS345pe - 1./3.*pow(m33,3)*L0r*RS346pe
          - 1./3.*pow(m33,3)*L0r*RS356pe - 1./3.*m11*pow(mp2,2)*L3r*
         RSp45613e - 1./6.*m11*pow(mp2,2)*L3r*RSp45633e - 1./6.*m11*
         pow(mp2,2)*L3r*RSp45611e - 1./3.*m11*pow(mp2,2)*L2r*RSp564e - 
         1./3.*m11*pow(mp2,2)*L2r*RSp465e );

      M13p6x23L +=  + pi16 * (  - 1./3.*m11*pow(mp2,2)*L2r*RSp456e - 1./
         3.*m11*pow(mp2,2)*L0r*RSp45633e - 1./3.*m11*pow(mp2,2)*L0r*
         RSp45611e - 1./3.*m11*pow(me2,2)*L3r*RSe45613p - 1./6.*m11*
         pow(me2,2)*L3r*RSe45633p - 1./6.*m11*pow(me2,2)*L3r*RSe45611p
          - 1./3.*m11*pow(me2,2)*L2r*RSe564p - 1./3.*m11*pow(me2,2)*L2r
         *RSe465p - 1./3.*m11*pow(me2,2)*L2r*RSe456p - 1./3.*m11*pow(
         me2,2)*L0r*RSe45633p - 1./3.*m11*pow(me2,2)*L0r*RSe45611p + 1./
         4.*m11*pow(m66,2)*L3r + 2*m11*pow(m66,2)*L2r - 1./3.*m11*pow(
         m66,2)*L2r*RS645pe + 1./2.*m11*pow(m66,2)*L0r + m11*m55*m66*
         L2r + 1./4.*m11*pow(m55,2)*L3r + 2*m11*pow(m55,2)*L2r - 1./3.*
         m11*pow(m55,2)*L2r*RS546pe + 1./2.*m11*pow(m55,2)*L0r + m11*
         m44*m66*L2r + m11*m44*m55*L2r + 1./4.*m11*pow(m44,2)*L3r + 2*
         m11*pow(m44,2)*L2r - 1./3.*m11*pow(m44,2)*L2r*RS456pe + 1./2.*
         m11*pow(m44,2)*L0r - 1./3.*m11*m33*L3r*RD3456ep - 1./3.*m11*
         m33*L3r*RD1456ep - 2./3.*m11*m33*L0r*RD3456ep - 2./3.*m11*m33*
         L0r*RD1456ep );

      M13p6x23L +=  + pi16 * ( 1./2.*m11*m33*m66*L3r + m11*m33*m66*L0r
          + 1./2.*m11*m33*m55*L3r + m11*m33*m55*L0r + 1./2.*m11*m33*m44
         *L3r + m11*m33*m44*L0r + 3./8.*m11*pow(m33,2)*L3r - 1./3.*m11*
         pow(m33,2)*L3r*RS34561pe + 1./6.*m11*pow(m33,2)*L3r*RS3456pee
          + 1./6.*m11*pow(m33,2)*L3r*RS3456ppe - 1./6.*m11*pow(m33,2)*
         L3r*RS345pe - 1./6.*m11*pow(m33,2)*L3r*RS346pe - 1./6.*m11*
         pow(m33,2)*L3r*RS356pe + 3./4.*m11*pow(m33,2)*L2r + 3./2.*m11*
         pow(m33,2)*L1r + 3./4.*m11*pow(m33,2)*L0r + 1./3.*m11*pow(
         m33,2)*L0r*RS3456pee + 1./3.*m11*pow(m33,2)*L0r*RS3456ppe - 1./
         3.*m11*pow(m33,2)*L0r*RS345pe - 1./3.*m11*pow(m33,2)*L0r*
         RS346pe - 1./3.*m11*pow(m33,2)*L0r*RS356pe - 1./3.*pow(m11,2)*
         L3r*RD1456ep - 2./3.*pow(m11,2)*L0r*RD1456ep + 1./4.*pow(
         m11,2)*m66*L3r + 1./2.*pow(m11,2)*m66*L0r + 1./4.*pow(m11,2)*
         m55*L3r + 1./2.*pow(m11,2)*m55*L0r + 1./4.*pow(m11,2)*m44*L3r
          + 1./2.*pow(m11,2)*m44*L0r + 3./8.*pow(m11,2)*m33*L3r - 1./3.
         *pow(m11,2)*m33*L3r*RS14563pe );

      M13p6x23L +=  + pi16 * ( 1./6.*pow(m11,2)*m33*L3r*RS1456pee + 1./
         6.*pow(m11,2)*m33*L3r*RS1456ppe - 1./6.*pow(m11,2)*m33*L3r*
         RS145pe - 1./6.*pow(m11,2)*m33*L3r*RS146pe - 1./6.*pow(m11,2)*
         m33*L3r*RS156pe + 3./4.*pow(m11,2)*m33*L2r + 3./2.*pow(m11,2)*
         m33*L1r + 3./4.*pow(m11,2)*m33*L0r + 1./3.*pow(m11,2)*m33*L0r*
         RS1456pee + 1./3.*pow(m11,2)*m33*L0r*RS1456ppe - 1./3.*pow(
         m11,2)*m33*L0r*RS145pe - 1./3.*pow(m11,2)*m33*L0r*RS146pe - 1./
         3.*pow(m11,2)*m33*L0r*RS156pe + 3./8.*pow(m11,3)*L3r - 1./3.*
         pow(m11,3)*L3r*RS14563pe + 1./6.*pow(m11,3)*L3r*RS1456pee + 1./
         6.*pow(m11,3)*L3r*RS1456ppe - 1./6.*pow(m11,3)*L3r*RS145pe - 1.
         /6.*pow(m11,3)*L3r*RS146pe - 1./6.*pow(m11,3)*L3r*RS156pe + 1./
         4.*pow(m11,3)*L2r + 1./2.*pow(m11,3)*L1r + 3./4.*pow(m11,3)*
         L0r + 1./3.*pow(m11,3)*L0r*RS1456pee + 1./3.*pow(m11,3)*L0r*
         RS1456ppe - 1./3.*pow(m11,3)*L0r*RS145pe - 1./3.*pow(m11,3)*
         L0r*RS146pe - 1./3.*pow(m11,3)*L0r*RS156pe );

      M13p6x23L +=  + Ab(m11,mu2) * (  - 8./9.*m66*L4r*RD1456ep + 16./
         27.*pow(m66,2)*L8r*RS145pe*RSe453p + 16./27.*pow(m66,2)*L8r*
         RS145pe*RSp453e - 8./27.*pow(m66,2)*L8r*pow(RS145pe,2) - 16./
         27.*pow(m66,2)*L8r*RSe451p*RS145pe - 16./27.*pow(m66,2)*L8r*
         RSp451e*RS145pe + 16./27.*pow(m66,2)*L7r*RS145pe*RSe453p + 16./
         27.*pow(m66,2)*L7r*RS145pe*RSp453e - 8./27.*pow(m66,2)*L7r*
         pow(RS145pe,2) - 16./27.*pow(m66,2)*L7r*RSe451p*RS145pe - 16./
         27.*pow(m66,2)*L7r*RSp451e*RS145pe + 16./27.*pow(m66,2)*L6r*
         RS145pe*RSe453p + 16./27.*pow(m66,2)*L6r*RS145pe*RSp453e - 8./
         27.*pow(m66,2)*L6r*pow(RS145pe,2) - 16./27.*pow(m66,2)*L6r*
         RSe451p*RS145pe - 16./27.*pow(m66,2)*L6r*RSp451e*RS145pe + 16./
         27.*pow(m66,2)*mp2*L8r*RN1p*RS145pe*RSp453e - 16./27.*pow(
         m66,2)*mp2*L8r*RN1p*RSp451e*RS145pe + 16./27.*pow(m66,2)*mp2*
         L7r*RN1p*RS145pe*RSp453e - 16./27.*pow(m66,2)*mp2*L7r*RN1p*
         RSp451e*RS145pe + 16./27.*pow(m66,2)*mp2*L6r*RN1p*RS145pe*
         RSp453e );

      M13p6x23L +=  + Ab(m11,mu2) * (  - 16./27.*pow(m66,2)*mp2*L6r*
         RN1p*RSp451e*RS145pe + 16./27.*pow(m66,2)*me2*L8r*RN1e*RS145pe
         *RSe453p - 16./27.*pow(m66,2)*me2*L8r*RN1e*RSe451p*RS145pe + 
         16./27.*pow(m66,2)*me2*L7r*RN1e*RS145pe*RSe453p - 16./27.*pow(
         m66,2)*me2*L7r*RN1e*RSe451p*RS145pe + 16./27.*pow(m66,2)*me2*
         L6r*RN1e*RS145pe*RSe453p - 16./27.*pow(m66,2)*me2*L6r*RN1e*
         RSe451p*RS145pe - 8./9.*m55*L4r*RD1456ep + 16./27.*m55*m66*L7r
         *RS145pe*RSe463p + 16./27.*m55*m66*L7r*RS145pe*RSp463e + 16./
         27.*m55*m66*L7r*RS146pe*RSe453p + 16./27.*m55*m66*L7r*RS146pe*
         RSp453e - 16./27.*m55*m66*L7r*RS146pe*RS145pe - 16./27.*m55*
         m66*L7r*RSe461p*RS145pe - 16./27.*m55*m66*L7r*RSe451p*RS146pe
          - 16./27.*m55*m66*L7r*RSp461e*RS145pe - 16./27.*m55*m66*L7r*
         RSp451e*RS146pe + 16./27.*m55*m66*L6r*RS145pe*RSe453p + 16./27.
         *m55*m66*L6r*RS145pe*RSp453e - 8./27.*m55*m66*L6r*pow(
         RS145pe,2) + 16./27.*m55*m66*L6r*RS146pe*RSe463p + 16./27.*m55
         *m66*L6r*RS146pe*RSp463e );

      M13p6x23L +=  + Ab(m11,mu2) * (  - 8./27.*m55*m66*L6r*pow(
         RS146pe,2) - 16./27.*m55*m66*L6r*RSe461p*RS146pe - 16./27.*m55
         *m66*L6r*RSe451p*RS145pe - 16./27.*m55*m66*L6r*RSp461e*RS146pe
          - 16./27.*m55*m66*L6r*RSp451e*RS145pe + 16./27.*m55*m66*mp2*
         L7r*RN1p*RS145pe*RSp463e + 16./27.*m55*m66*mp2*L7r*RN1p*
         RS146pe*RSp453e - 16./27.*m55*m66*mp2*L7r*RN1p*RSp461e*RS145pe
          - 16./27.*m55*m66*mp2*L7r*RN1p*RSp451e*RS146pe + 16./27.*m55*
         m66*mp2*L6r*RN1p*RS145pe*RSp453e + 16./27.*m55*m66*mp2*L6r*
         RN1p*RS146pe*RSp463e - 16./27.*m55*m66*mp2*L6r*RN1p*RSp461e*
         RS146pe - 16./27.*m55*m66*mp2*L6r*RN1p*RSp451e*RS145pe + 16./
         27.*m55*m66*me2*L7r*RN1e*RS145pe*RSe463p + 16./27.*m55*m66*me2
         *L7r*RN1e*RS146pe*RSe453p - 16./27.*m55*m66*me2*L7r*RN1e*
         RSe461p*RS145pe - 16./27.*m55*m66*me2*L7r*RN1e*RSe451p*RS146pe
          + 16./27.*m55*m66*me2*L6r*RN1e*RS145pe*RSe453p + 16./27.*m55*
         m66*me2*L6r*RN1e*RS146pe*RSe463p - 16./27.*m55*m66*me2*L6r*
         RN1e*RSe461p*RS146pe );

      M13p6x23L +=  + Ab(m11,mu2) * (  - 16./27.*m55*m66*me2*L6r*RN1e*
         RSe451p*RS145pe + 16./27.*pow(m55,2)*L8r*RS146pe*RSe463p + 16./
         27.*pow(m55,2)*L8r*RS146pe*RSp463e - 8./27.*pow(m55,2)*L8r*
         pow(RS146pe,2) - 16./27.*pow(m55,2)*L8r*RSe461p*RS146pe - 16./
         27.*pow(m55,2)*L8r*RSp461e*RS146pe + 16./27.*pow(m55,2)*L7r*
         RS146pe*RSe463p + 16./27.*pow(m55,2)*L7r*RS146pe*RSp463e - 8./
         27.*pow(m55,2)*L7r*pow(RS146pe,2) - 16./27.*pow(m55,2)*L7r*
         RSe461p*RS146pe - 16./27.*pow(m55,2)*L7r*RSp461e*RS146pe + 16./
         27.*pow(m55,2)*L6r*RS146pe*RSe463p + 16./27.*pow(m55,2)*L6r*
         RS146pe*RSp463e - 8./27.*pow(m55,2)*L6r*pow(RS146pe,2) - 16./
         27.*pow(m55,2)*L6r*RSe461p*RS146pe - 16./27.*pow(m55,2)*L6r*
         RSp461e*RS146pe + 16./27.*pow(m55,2)*mp2*L8r*RN1p*RS146pe*
         RSp463e - 16./27.*pow(m55,2)*mp2*L8r*RN1p*RSp461e*RS146pe + 16.
         /27.*pow(m55,2)*mp2*L7r*RN1p*RS146pe*RSp463e - 16./27.*pow(
         m55,2)*mp2*L7r*RN1p*RSp461e*RS146pe + 16./27.*pow(m55,2)*mp2*
         L6r*RN1p*RS146pe*RSp463e );

      M13p6x23L +=  + Ab(m11,mu2) * (  - 16./27.*pow(m55,2)*mp2*L6r*
         RN1p*RSp461e*RS146pe + 16./27.*pow(m55,2)*me2*L8r*RN1e*RS146pe
         *RSe463p - 16./27.*pow(m55,2)*me2*L8r*RN1e*RSe461p*RS146pe + 
         16./27.*pow(m55,2)*me2*L7r*RN1e*RS146pe*RSe463p - 16./27.*pow(
         m55,2)*me2*L7r*RN1e*RSe461p*RS146pe + 16./27.*pow(m55,2)*me2*
         L6r*RN1e*RS146pe*RSe463p - 16./27.*pow(m55,2)*me2*L6r*RN1e*
         RSe461p*RS146pe - 8./9.*m44*L4r*RD1456ep + 16./27.*m44*m66*L7r
         *RS145pe*RSe563p + 16./27.*m44*m66*L7r*RS145pe*RSp563e + 16./
         27.*m44*m66*L7r*RS156pe*RSe453p + 16./27.*m44*m66*L7r*RS156pe*
         RSp453e - 16./27.*m44*m66*L7r*RS156pe*RS145pe - 16./27.*m44*
         m66*L7r*RSe561p*RS145pe - 16./27.*m44*m66*L7r*RSe451p*RS156pe
          - 16./27.*m44*m66*L7r*RSp561e*RS145pe - 16./27.*m44*m66*L7r*
         RSp451e*RS156pe + 16./27.*m44*m66*L6r*RS145pe*RSe453p + 16./27.
         *m44*m66*L6r*RS145pe*RSp453e - 8./27.*m44*m66*L6r*pow(
         RS145pe,2) + 16./27.*m44*m66*L6r*RS156pe*RSe563p + 16./27.*m44
         *m66*L6r*RS156pe*RSp563e );

      M13p6x23L +=  + Ab(m11,mu2) * (  - 8./27.*m44*m66*L6r*pow(
         RS156pe,2) - 16./27.*m44*m66*L6r*RSe561p*RS156pe - 16./27.*m44
         *m66*L6r*RSe451p*RS145pe - 16./27.*m44*m66*L6r*RSp561e*RS156pe
          - 16./27.*m44*m66*L6r*RSp451e*RS145pe + 16./27.*m44*m66*mp2*
         L7r*RN1p*RS145pe*RSp563e + 16./27.*m44*m66*mp2*L7r*RN1p*
         RS156pe*RSp453e - 16./27.*m44*m66*mp2*L7r*RN1p*RSp561e*RS145pe
          - 16./27.*m44*m66*mp2*L7r*RN1p*RSp451e*RS156pe + 16./27.*m44*
         m66*mp2*L6r*RN1p*RS145pe*RSp453e + 16./27.*m44*m66*mp2*L6r*
         RN1p*RS156pe*RSp563e - 16./27.*m44*m66*mp2*L6r*RN1p*RSp561e*
         RS156pe - 16./27.*m44*m66*mp2*L6r*RN1p*RSp451e*RS145pe + 16./
         27.*m44*m66*me2*L7r*RN1e*RS145pe*RSe563p + 16./27.*m44*m66*me2
         *L7r*RN1e*RS156pe*RSe453p - 16./27.*m44*m66*me2*L7r*RN1e*
         RSe561p*RS145pe - 16./27.*m44*m66*me2*L7r*RN1e*RSe451p*RS156pe
          + 16./27.*m44*m66*me2*L6r*RN1e*RS145pe*RSe453p + 16./27.*m44*
         m66*me2*L6r*RN1e*RS156pe*RSe563p - 16./27.*m44*m66*me2*L6r*
         RN1e*RSe561p*RS156pe );

      M13p6x23L +=  + Ab(m11,mu2) * (  - 16./27.*m44*m66*me2*L6r*RN1e*
         RSe451p*RS145pe + 16./27.*m44*m55*L7r*RS146pe*RSe563p + 16./27.
         *m44*m55*L7r*RS146pe*RSp563e + 16./27.*m44*m55*L7r*RS156pe*
         RSe463p + 16./27.*m44*m55*L7r*RS156pe*RSp463e - 16./27.*m44*
         m55*L7r*RS156pe*RS146pe - 16./27.*m44*m55*L7r*RSe561p*RS146pe
          - 16./27.*m44*m55*L7r*RSe461p*RS156pe - 16./27.*m44*m55*L7r*
         RSp561e*RS146pe - 16./27.*m44*m55*L7r*RSp461e*RS156pe + 16./27.
         *m44*m55*L6r*RS146pe*RSe463p + 16./27.*m44*m55*L6r*RS146pe*
         RSp463e - 8./27.*m44*m55*L6r*pow(RS146pe,2) + 16./27.*m44*m55*
         L6r*RS156pe*RSe563p + 16./27.*m44*m55*L6r*RS156pe*RSp563e - 8./
         27.*m44*m55*L6r*pow(RS156pe,2) - 16./27.*m44*m55*L6r*RSe561p*
         RS156pe - 16./27.*m44*m55*L6r*RSe461p*RS146pe - 16./27.*m44*
         m55*L6r*RSp561e*RS156pe - 16./27.*m44*m55*L6r*RSp461e*RS146pe
          + 16./27.*m44*m55*mp2*L7r*RN1p*RS146pe*RSp563e + 16./27.*m44*
         m55*mp2*L7r*RN1p*RS156pe*RSp463e - 16./27.*m44*m55*mp2*L7r*
         RN1p*RSp561e*RS146pe );

      M13p6x23L +=  + Ab(m11,mu2) * (  - 16./27.*m44*m55*mp2*L7r*RN1p*
         RSp461e*RS156pe + 16./27.*m44*m55*mp2*L6r*RN1p*RS146pe*RSp463e
          + 16./27.*m44*m55*mp2*L6r*RN1p*RS156pe*RSp563e - 16./27.*m44*
         m55*mp2*L6r*RN1p*RSp561e*RS156pe - 16./27.*m44*m55*mp2*L6r*
         RN1p*RSp461e*RS146pe + 16./27.*m44*m55*me2*L7r*RN1e*RS146pe*
         RSe563p + 16./27.*m44*m55*me2*L7r*RN1e*RS156pe*RSe463p - 16./
         27.*m44*m55*me2*L7r*RN1e*RSe561p*RS146pe - 16./27.*m44*m55*me2
         *L7r*RN1e*RSe461p*RS156pe + 16./27.*m44*m55*me2*L6r*RN1e*
         RS146pe*RSe463p + 16./27.*m44*m55*me2*L6r*RN1e*RS156pe*RSe563p
          - 16./27.*m44*m55*me2*L6r*RN1e*RSe561p*RS156pe - 16./27.*m44*
         m55*me2*L6r*RN1e*RSe461p*RS146pe + 16./27.*pow(m44,2)*L8r*
         RS156pe*RSe563p + 16./27.*pow(m44,2)*L8r*RS156pe*RSp563e - 8./
         27.*pow(m44,2)*L8r*pow(RS156pe,2) - 16./27.*pow(m44,2)*L8r*
         RSe561p*RS156pe - 16./27.*pow(m44,2)*L8r*RSp561e*RS156pe + 16./
         27.*pow(m44,2)*L7r*RS156pe*RSe563p + 16./27.*pow(m44,2)*L7r*
         RS156pe*RSp563e );

      M13p6x23L +=  + Ab(m11,mu2) * (  - 8./27.*pow(m44,2)*L7r*pow(
         RS156pe,2) - 16./27.*pow(m44,2)*L7r*RSe561p*RS156pe - 16./27.*
         pow(m44,2)*L7r*RSp561e*RS156pe + 16./27.*pow(m44,2)*L6r*
         RS156pe*RSe563p + 16./27.*pow(m44,2)*L6r*RS156pe*RSp563e - 8./
         27.*pow(m44,2)*L6r*pow(RS156pe,2) - 16./27.*pow(m44,2)*L6r*
         RSe561p*RS156pe - 16./27.*pow(m44,2)*L6r*RSp561e*RS156pe + 16./
         27.*pow(m44,2)*mp2*L8r*RN1p*RS156pe*RSp563e - 16./27.*pow(
         m44,2)*mp2*L8r*RN1p*RSp561e*RS156pe + 16./27.*pow(m44,2)*mp2*
         L7r*RN1p*RS156pe*RSp563e - 16./27.*pow(m44,2)*mp2*L7r*RN1p*
         RSp561e*RS156pe + 16./27.*pow(m44,2)*mp2*L6r*RN1p*RS156pe*
         RSp563e - 16./27.*pow(m44,2)*mp2*L6r*RN1p*RSp561e*RS156pe + 16.
         /27.*pow(m44,2)*me2*L8r*RN1e*RS156pe*RSe563p - 16./27.*pow(
         m44,2)*me2*L8r*RN1e*RSe561p*RS156pe + 16./27.*pow(m44,2)*me2*
         L7r*RN1e*RS156pe*RSe563p - 16./27.*pow(m44,2)*me2*L7r*RN1e*
         RSe561p*RS156pe + 16./27.*pow(m44,2)*me2*L6r*RN1e*RS156pe*
         RSe563p );

      M13p6x23L +=  + Ab(m11,mu2) * (  - 16./27.*pow(m44,2)*me2*L6r*
         RN1e*RSe561p*RS156pe + 4./9.*m33*L5r*RD1456ep - 5./3.*m33*L3r*
         RD1456ep - 2./3.*m33*L0r*RD1456ep - 16./9.*m33*m66*L7r*RS145pe
          - 16./9.*m33*m66*L6r*RS14563pe - 8./27.*m33*m66*L5r*RS145pe*
         RSe453p - 8./27.*m33*m66*L5r*RS145pe*RSp453e + 16./9.*m33*m66*
         L4r*RS14563pe + 8./9.*m33*m66*L4r*RSe45613p + 8./9.*m33*m66*
         L4r*RSp45613e - 8./27.*m33*m66*L4r*RS145pe*RSe453p - 8./27.*
         m33*m66*L4r*RS145pe*RSp453e - 8./27.*m33*m66*L4r*RS146pe*
         RSe463p - 8./27.*m33*m66*L4r*RS146pe*RSp463e - 8./27.*m33*m66*
         L4r*RS156pe*RSe563p - 8./27.*m33*m66*L4r*RS156pe*RSp563e - 8./
         27.*m33*m66*mp2*L5r*RN1p*RS145pe*RSp453e + 8./9.*m33*m66*mp2*
         L4r*RN1p*RSp45613e - 8./27.*m33*m66*mp2*L4r*RN1p*RS145pe*
         RSp453e - 8./27.*m33*m66*mp2*L4r*RN1p*RS146pe*RSp463e - 8./27.
         *m33*m66*mp2*L4r*RN1p*RS156pe*RSp563e - 8./27.*m33*m66*me2*L5r
         *RN1e*RS145pe*RSe453p + 8./9.*m33*m66*me2*L4r*RN1e*RSe45613p
          - 8./27.*m33*m66*me2*L4r*RN1e*RS145pe*RSe453p );

      M13p6x23L +=  + Ab(m11,mu2) * (  - 8./27.*m33*m66*me2*L4r*RN1e*
         RS146pe*RSe463p - 8./27.*m33*m66*me2*L4r*RN1e*RS156pe*RSe563p
          + 16./27.*m33*pow(m66,2)*L8r*RN1p*RS145pe*RSp453e + 16./27.*
         m33*pow(m66,2)*L8r*RN1e*RS145pe*RSe453p + 16./27.*m33*pow(
         m66,2)*L8r*RN13*RS145pe*RS345pe + 16./27.*m33*pow(m66,2)*L7r*
         RN1p*RS145pe*RSp453e + 16./27.*m33*pow(m66,2)*L7r*RN1e*RS145pe
         *RSe453p + 16./27.*m33*pow(m66,2)*L7r*RN13*RS145pe*RS345pe + 
         16./27.*m33*pow(m66,2)*L6r*RN1p*RS145pe*RSp453e + 16./27.*m33*
         pow(m66,2)*L6r*RN1e*RS145pe*RSe453p + 16./27.*m33*pow(m66,2)*
         L6r*RN13*RS145pe*RS345pe - 16./9.*m33*m55*L7r*RS146pe - 16./9.
         *m33*m55*L6r*RS14563pe - 8./27.*m33*m55*L5r*RS146pe*RSe463p - 
         8./27.*m33*m55*L5r*RS146pe*RSp463e + 16./9.*m33*m55*L4r*
         RS14563pe + 8./9.*m33*m55*L4r*RSe45613p + 8./9.*m33*m55*L4r*
         RSp45613e - 8./27.*m33*m55*L4r*RS145pe*RSe453p - 8./27.*m33*
         m55*L4r*RS145pe*RSp453e - 8./27.*m33*m55*L4r*RS146pe*RSe463p
          - 8./27.*m33*m55*L4r*RS146pe*RSp463e );

      M13p6x23L +=  + Ab(m11,mu2) * (  - 8./27.*m33*m55*L4r*RS156pe*
         RSe563p - 8./27.*m33*m55*L4r*RS156pe*RSp563e - 8./27.*m33*m55*
         mp2*L5r*RN1p*RS146pe*RSp463e + 8./9.*m33*m55*mp2*L4r*RN1p*
         RSp45613e - 8./27.*m33*m55*mp2*L4r*RN1p*RS145pe*RSp453e - 8./
         27.*m33*m55*mp2*L4r*RN1p*RS146pe*RSp463e - 8./27.*m33*m55*mp2*
         L4r*RN1p*RS156pe*RSp563e - 8./27.*m33*m55*me2*L5r*RN1e*RS146pe
         *RSe463p + 8./9.*m33*m55*me2*L4r*RN1e*RSe45613p - 8./27.*m33*
         m55*me2*L4r*RN1e*RS145pe*RSe453p - 8./27.*m33*m55*me2*L4r*RN1e
         *RS146pe*RSe463p - 8./27.*m33*m55*me2*L4r*RN1e*RS156pe*RSe563p
          + 16./27.*m33*m55*m66*L7r*RN1p*RS145pe*RSp463e + 16./27.*m33*
         m55*m66*L7r*RN1p*RS146pe*RSp453e + 16./27.*m33*m55*m66*L7r*
         RN1e*RS145pe*RSe463p + 16./27.*m33*m55*m66*L7r*RN1e*RS146pe*
         RSe453p + 16./27.*m33*m55*m66*L7r*RN13*RS145pe*RS346pe + 16./
         27.*m33*m55*m66*L7r*RN13*RS146pe*RS345pe + 16./27.*m33*m55*m66
         *L6r*RN1p*RS145pe*RSp453e + 16./27.*m33*m55*m66*L6r*RN1p*
         RS146pe*RSp463e );

      M13p6x23L +=  + Ab(m11,mu2) * ( 16./27.*m33*m55*m66*L6r*RN1e*
         RS145pe*RSe453p + 16./27.*m33*m55*m66*L6r*RN1e*RS146pe*RSe463p
          + 16./27.*m33*m55*m66*L6r*RN13*RS145pe*RS345pe + 16./27.*m33*
         m55*m66*L6r*RN13*RS146pe*RS346pe + 16./27.*m33*pow(m55,2)*L8r*
         RN1p*RS146pe*RSp463e + 16./27.*m33*pow(m55,2)*L8r*RN1e*RS146pe
         *RSe463p + 16./27.*m33*pow(m55,2)*L8r*RN13*RS146pe*RS346pe + 
         16./27.*m33*pow(m55,2)*L7r*RN1p*RS146pe*RSp463e + 16./27.*m33*
         pow(m55,2)*L7r*RN1e*RS146pe*RSe463p + 16./27.*m33*pow(m55,2)*
         L7r*RN13*RS146pe*RS346pe + 16./27.*m33*pow(m55,2)*L6r*RN1p*
         RS146pe*RSp463e + 16./27.*m33*pow(m55,2)*L6r*RN1e*RS146pe*
         RSe463p + 16./27.*m33*pow(m55,2)*L6r*RN13*RS146pe*RS346pe - 16.
         /9.*m33*m44*L7r*RS156pe - 16./9.*m33*m44*L6r*RS14563pe - 8./27.
         *m33*m44*L5r*RS156pe*RSe563p - 8./27.*m33*m44*L5r*RS156pe*
         RSp563e + 16./9.*m33*m44*L4r*RS14563pe + 8./9.*m33*m44*L4r*
         RSe45613p + 8./9.*m33*m44*L4r*RSp45613e - 8./27.*m33*m44*L4r*
         RS145pe*RSe453p );

      M13p6x23L +=  + Ab(m11,mu2) * (  - 8./27.*m33*m44*L4r*RS145pe*
         RSp453e - 8./27.*m33*m44*L4r*RS146pe*RSe463p - 8./27.*m33*m44*
         L4r*RS146pe*RSp463e - 8./27.*m33*m44*L4r*RS156pe*RSe563p - 8./
         27.*m33*m44*L4r*RS156pe*RSp563e - 8./27.*m33*m44*mp2*L5r*RN1p*
         RS156pe*RSp563e + 8./9.*m33*m44*mp2*L4r*RN1p*RSp45613e - 8./27.
         *m33*m44*mp2*L4r*RN1p*RS145pe*RSp453e - 8./27.*m33*m44*mp2*L4r
         *RN1p*RS146pe*RSp463e - 8./27.*m33*m44*mp2*L4r*RN1p*RS156pe*
         RSp563e - 8./27.*m33*m44*me2*L5r*RN1e*RS156pe*RSe563p + 8./9.*
         m33*m44*me2*L4r*RN1e*RSe45613p - 8./27.*m33*m44*me2*L4r*RN1e*
         RS145pe*RSe453p - 8./27.*m33*m44*me2*L4r*RN1e*RS146pe*RSe463p
          - 8./27.*m33*m44*me2*L4r*RN1e*RS156pe*RSe563p + 16./27.*m33*
         m44*m66*L7r*RN1p*RS145pe*RSp563e + 16./27.*m33*m44*m66*L7r*
         RN1p*RS156pe*RSp453e + 16./27.*m33*m44*m66*L7r*RN1e*RS145pe*
         RSe563p + 16./27.*m33*m44*m66*L7r*RN1e*RS156pe*RSe453p + 16./
         27.*m33*m44*m66*L7r*RN13*RS145pe*RS356pe + 16./27.*m33*m44*m66
         *L7r*RN13*RS156pe*RS345pe );

      M13p6x23L +=  + Ab(m11,mu2) * ( 16./27.*m33*m44*m66*L6r*RN1p*
         RS145pe*RSp453e + 16./27.*m33*m44*m66*L6r*RN1p*RS156pe*RSp563e
          + 16./27.*m33*m44*m66*L6r*RN1e*RS145pe*RSe453p + 16./27.*m33*
         m44*m66*L6r*RN1e*RS156pe*RSe563p + 16./27.*m33*m44*m66*L6r*
         RN13*RS145pe*RS345pe + 16./27.*m33*m44*m66*L6r*RN13*RS156pe*
         RS356pe + 16./27.*m33*m44*m55*L7r*RN1p*RS146pe*RSp563e + 16./
         27.*m33*m44*m55*L7r*RN1p*RS156pe*RSp463e + 16./27.*m33*m44*m55
         *L7r*RN1e*RS146pe*RSe563p + 16./27.*m33*m44*m55*L7r*RN1e*
         RS156pe*RSe463p + 16./27.*m33*m44*m55*L7r*RN13*RS146pe*RS356pe
          + 16./27.*m33*m44*m55*L7r*RN13*RS156pe*RS346pe + 16./27.*m33*
         m44*m55*L6r*RN1p*RS146pe*RSp463e + 16./27.*m33*m44*m55*L6r*
         RN1p*RS156pe*RSp563e + 16./27.*m33*m44*m55*L6r*RN1e*RS146pe*
         RSe463p + 16./27.*m33*m44*m55*L6r*RN1e*RS156pe*RSe563p + 16./
         27.*m33*m44*m55*L6r*RN13*RS146pe*RS346pe + 16./27.*m33*m44*m55
         *L6r*RN13*RS156pe*RS356pe + 16./27.*m33*pow(m44,2)*L8r*RN1p*
         RS156pe*RSp563e );

      M13p6x23L +=  + Ab(m11,mu2) * ( 16./27.*m33*pow(m44,2)*L8r*RN1e*
         RS156pe*RSe563p + 16./27.*m33*pow(m44,2)*L8r*RN13*RS156pe*
         RS356pe + 16./27.*m33*pow(m44,2)*L7r*RN1p*RS156pe*RSp563e + 16.
         /27.*m33*pow(m44,2)*L7r*RN1e*RS156pe*RSe563p + 16./27.*m33*
         pow(m44,2)*L7r*RN13*RS156pe*RS356pe + 16./27.*m33*pow(m44,2)*
         L6r*RN1p*RS156pe*RSp563e + 16./27.*m33*pow(m44,2)*L6r*RN1e*
         RS156pe*RSe563p + 16./27.*m33*pow(m44,2)*L6r*RN13*RS156pe*
         RS356pe - 16./9.*pow(m33,2)*L8r*RS14563pe + 4./9.*pow(m33,2)*
         L5r*RS14563pe - 16./9.*pow(m33,2)*m66*L7r*RN13*RS145pe - 16./9.
         *pow(m33,2)*m66*L6r*RN13*RS14563pe - 16./9.*pow(m33,2)*m55*L7r
         *RN13*RS146pe - 16./9.*pow(m33,2)*m55*L6r*RN13*RS14563pe - 16./
         9.*pow(m33,2)*m44*L7r*RN13*RS156pe - 16./9.*pow(m33,2)*m44*L6r
         *RN13*RS14563pe - 16./9.*pow(m33,3)*L8r*RN13*RS14563pe - 5./3.
         *m11*L3r*RD1456ep - 2./3.*m11*L0r*RD1456ep - 16./9.*m11*m66*
         L7r*RSe453p - 16./9.*m11*m66*L7r*RSp453e - 16./9.*m11*m66*L7r*
         RS145pe );

      M13p6x23L +=  + Ab(m11,mu2) * ( 16./9.*m11*m66*L7r*RSe451p + 16./
         9.*m11*m66*L7r*RSp451e - 32./9.*m11*m66*L6r*RS14563pe - 16./9.
         *m11*m66*L6r*RSe45613p - 16./9.*m11*m66*L6r*RSp45613e - 8./9.*
         m11*m66*L6r*RS1456pee - 8./9.*m11*m66*L6r*RS1456ppe + 16./9.*
         m11*m66*L6r*RSe45611p + 16./9.*m11*m66*L6r*RSp45611e + 8./9.*
         m11*m66*L6r*RS145pe + 8./9.*m11*m66*L6r*RS146pe + 8./9.*m11*
         m66*L6r*RS156pe - 8./27.*m11*m66*L5r*RS145pe*RS345pe - 16./27.
         *m11*m66*L5r*RS145pe*RSe453p - 16./27.*m11*m66*L5r*RS145pe*
         RSp453e + 4./27.*m11*m66*L5r*pow(RS145pe,2) + 8./9.*m11*m66*
         L4r*RS34561pe + 40./9.*m11*m66*L4r*RS14563pe + 16./9.*m11*m66*
         L4r*RSe45613p + 16./9.*m11*m66*L4r*RSp45613e + 4./9.*m11*m66*
         L4r*RS1456pee + 4./9.*m11*m66*L4r*RS1456ppe - 4./9.*m11*m66*
         L4r*RS145pe - 8./27.*m11*m66*L4r*RS145pe*RS345pe - 16./27.*m11
         *m66*L4r*RS145pe*RSe453p - 16./27.*m11*m66*L4r*RS145pe*RSp453e
          + 4./27.*m11*m66*L4r*pow(RS145pe,2) - 4./9.*m11*m66*L4r*
         RS146pe );

      M13p6x23L +=  + Ab(m11,mu2) * (  - 8./27.*m11*m66*L4r*RS146pe*
         RS346pe - 16./27.*m11*m66*L4r*RS146pe*RSe463p - 16./27.*m11*
         m66*L4r*RS146pe*RSp463e + 4./27.*m11*m66*L4r*pow(RS146pe,2) - 
         4./9.*m11*m66*L4r*RS156pe - 8./27.*m11*m66*L4r*RS156pe*RS356pe
          - 16./27.*m11*m66*L4r*RS156pe*RSe563p - 16./27.*m11*m66*L4r*
         RS156pe*RSp563e + 4./27.*m11*m66*L4r*pow(RS156pe,2) - 16./9.*
         m11*m66*mp2*L7r*RN1p*RSp453e + 16./9.*m11*m66*mp2*L7r*RN1p*
         RSp451e - 16./9.*m11*m66*mp2*L6r*RN1p*RSp45613e + 16./9.*m11*
         m66*mp2*L6r*RN1p*RSp45611e - 16./27.*m11*m66*mp2*L5r*RN1p*
         RS145pe*RSp453e + 16./9.*m11*m66*mp2*L4r*RN1p*RSp45613e - 16./
         27.*m11*m66*mp2*L4r*RN1p*RS145pe*RSp453e - 16./27.*m11*m66*mp2
         *L4r*RN1p*RS146pe*RSp463e - 16./27.*m11*m66*mp2*L4r*RN1p*
         RS156pe*RSp563e - 16./9.*m11*m66*me2*L7r*RN1e*RSe453p + 16./9.
         *m11*m66*me2*L7r*RN1e*RSe451p - 16./9.*m11*m66*me2*L6r*RN1e*
         RSe45613p + 16./9.*m11*m66*me2*L6r*RN1e*RSe45611p - 16./27.*
         m11*m66*me2*L5r*RN1e*RS145pe*RSe453p );

      M13p6x23L +=  + Ab(m11,mu2) * ( 16./9.*m11*m66*me2*L4r*RN1e*
         RSe45613p - 16./27.*m11*m66*me2*L4r*RN1e*RS145pe*RSe453p - 16./
         27.*m11*m66*me2*L4r*RN1e*RS146pe*RSe463p - 16./27.*m11*m66*me2
         *L4r*RN1e*RS156pe*RSe563p + 16./27.*m11*pow(m66,2)*L8r*RN1p*
         RS145pe*RSp453e + 16./27.*m11*pow(m66,2)*L8r*RN1p*RSp451e*
         RS145pe + 16./27.*m11*pow(m66,2)*L8r*RN1e*RS145pe*RSe453p + 16.
         /27.*m11*pow(m66,2)*L8r*RN1e*RSe451p*RS145pe + 32./27.*m11*
         pow(m66,2)*L8r*RN13*RS145pe*RS345pe + 16./27.*m11*pow(m66,2)*
         L7r*RN1p*RS145pe*RSp453e + 16./27.*m11*pow(m66,2)*L7r*RN1p*
         RSp451e*RS145pe + 16./27.*m11*pow(m66,2)*L7r*RN1e*RS145pe*
         RSe453p + 16./27.*m11*pow(m66,2)*L7r*RN1e*RSe451p*RS145pe + 32.
         /27.*m11*pow(m66,2)*L7r*RN13*RS145pe*RS345pe + 16./27.*m11*
         pow(m66,2)*L6r*RN1p*RS145pe*RSp453e + 16./27.*m11*pow(m66,2)*
         L6r*RN1p*RSp451e*RS145pe + 16./27.*m11*pow(m66,2)*L6r*RN1e*
         RS145pe*RSe453p + 16./27.*m11*pow(m66,2)*L6r*RN1e*RSe451p*
         RS145pe );

      M13p6x23L +=  + Ab(m11,mu2) * ( 32./27.*m11*pow(m66,2)*L6r*RN13*
         RS145pe*RS345pe - 16./9.*m11*m55*L7r*RSe463p - 16./9.*m11*m55*
         L7r*RSp463e - 16./9.*m11*m55*L7r*RS146pe + 16./9.*m11*m55*L7r*
         RSe461p + 16./9.*m11*m55*L7r*RSp461e - 32./9.*m11*m55*L6r*
         RS14563pe - 16./9.*m11*m55*L6r*RSe45613p - 16./9.*m11*m55*L6r*
         RSp45613e - 8./9.*m11*m55*L6r*RS1456pee - 8./9.*m11*m55*L6r*
         RS1456ppe + 16./9.*m11*m55*L6r*RSe45611p + 16./9.*m11*m55*L6r*
         RSp45611e + 8./9.*m11*m55*L6r*RS145pe + 8./9.*m11*m55*L6r*
         RS146pe + 8./9.*m11*m55*L6r*RS156pe - 8./27.*m11*m55*L5r*
         RS146pe*RS346pe - 16./27.*m11*m55*L5r*RS146pe*RSe463p - 16./27.
         *m11*m55*L5r*RS146pe*RSp463e + 4./27.*m11*m55*L5r*pow(
         RS146pe,2) + 8./9.*m11*m55*L4r*RS34561pe + 40./9.*m11*m55*L4r*
         RS14563pe + 16./9.*m11*m55*L4r*RSe45613p + 16./9.*m11*m55*L4r*
         RSp45613e + 4./9.*m11*m55*L4r*RS1456pee + 4./9.*m11*m55*L4r*
         RS1456ppe - 4./9.*m11*m55*L4r*RS145pe - 8./27.*m11*m55*L4r*
         RS145pe*RS345pe );

      M13p6x23L +=  + Ab(m11,mu2) * (  - 16./27.*m11*m55*L4r*RS145pe*
         RSe453p - 16./27.*m11*m55*L4r*RS145pe*RSp453e + 4./27.*m11*m55
         *L4r*pow(RS145pe,2) - 4./9.*m11*m55*L4r*RS146pe - 8./27.*m11*
         m55*L4r*RS146pe*RS346pe - 16./27.*m11*m55*L4r*RS146pe*RSe463p
          - 16./27.*m11*m55*L4r*RS146pe*RSp463e + 4./27.*m11*m55*L4r*
         pow(RS146pe,2) - 4./9.*m11*m55*L4r*RS156pe - 8./27.*m11*m55*
         L4r*RS156pe*RS356pe - 16./27.*m11*m55*L4r*RS156pe*RSe563p - 16.
         /27.*m11*m55*L4r*RS156pe*RSp563e + 4./27.*m11*m55*L4r*pow(
         RS156pe,2) - 16./9.*m11*m55*mp2*L7r*RN1p*RSp463e + 16./9.*m11*
         m55*mp2*L7r*RN1p*RSp461e - 16./9.*m11*m55*mp2*L6r*RN1p*
         RSp45613e + 16./9.*m11*m55*mp2*L6r*RN1p*RSp45611e - 16./27.*
         m11*m55*mp2*L5r*RN1p*RS146pe*RSp463e + 16./9.*m11*m55*mp2*L4r*
         RN1p*RSp45613e - 16./27.*m11*m55*mp2*L4r*RN1p*RS145pe*RSp453e
          - 16./27.*m11*m55*mp2*L4r*RN1p*RS146pe*RSp463e - 16./27.*m11*
         m55*mp2*L4r*RN1p*RS156pe*RSp563e - 16./9.*m11*m55*me2*L7r*RN1e
         *RSe463p );

      M13p6x23L +=  + Ab(m11,mu2) * ( 16./9.*m11*m55*me2*L7r*RN1e*
         RSe461p - 16./9.*m11*m55*me2*L6r*RN1e*RSe45613p + 16./9.*m11*
         m55*me2*L6r*RN1e*RSe45611p - 16./27.*m11*m55*me2*L5r*RN1e*
         RS146pe*RSe463p + 16./9.*m11*m55*me2*L4r*RN1e*RSe45613p - 16./
         27.*m11*m55*me2*L4r*RN1e*RS145pe*RSe453p - 16./27.*m11*m55*me2
         *L4r*RN1e*RS146pe*RSe463p - 16./27.*m11*m55*me2*L4r*RN1e*
         RS156pe*RSe563p + 16./27.*m11*m55*m66*L7r*RN1p*RS145pe*RSp463e
          + 16./27.*m11*m55*m66*L7r*RN1p*RS146pe*RSp453e + 16./27.*m11*
         m55*m66*L7r*RN1p*RSp461e*RS145pe + 16./27.*m11*m55*m66*L7r*
         RN1p*RSp451e*RS146pe + 16./27.*m11*m55*m66*L7r*RN1e*RS145pe*
         RSe463p + 16./27.*m11*m55*m66*L7r*RN1e*RS146pe*RSe453p + 16./
         27.*m11*m55*m66*L7r*RN1e*RSe461p*RS145pe + 16./27.*m11*m55*m66
         *L7r*RN1e*RSe451p*RS146pe + 32./27.*m11*m55*m66*L7r*RN13*
         RS145pe*RS346pe + 32./27.*m11*m55*m66*L7r*RN13*RS146pe*RS345pe
          + 16./27.*m11*m55*m66*L6r*RN1p*RS145pe*RSp453e + 16./27.*m11*
         m55*m66*L6r*RN1p*RS146pe*RSp463e );

      M13p6x23L +=  + Ab(m11,mu2) * ( 16./27.*m11*m55*m66*L6r*RN1p*
         RSp461e*RS146pe + 16./27.*m11*m55*m66*L6r*RN1p*RSp451e*RS145pe
          + 16./27.*m11*m55*m66*L6r*RN1e*RS145pe*RSe453p + 16./27.*m11*
         m55*m66*L6r*RN1e*RS146pe*RSe463p + 16./27.*m11*m55*m66*L6r*
         RN1e*RSe461p*RS146pe + 16./27.*m11*m55*m66*L6r*RN1e*RSe451p*
         RS145pe + 32./27.*m11*m55*m66*L6r*RN13*RS145pe*RS345pe + 32./
         27.*m11*m55*m66*L6r*RN13*RS146pe*RS346pe + 16./27.*m11*pow(
         m55,2)*L8r*RN1p*RS146pe*RSp463e + 16./27.*m11*pow(m55,2)*L8r*
         RN1p*RSp461e*RS146pe + 16./27.*m11*pow(m55,2)*L8r*RN1e*RS146pe
         *RSe463p + 16./27.*m11*pow(m55,2)*L8r*RN1e*RSe461p*RS146pe + 
         32./27.*m11*pow(m55,2)*L8r*RN13*RS146pe*RS346pe + 16./27.*m11*
         pow(m55,2)*L7r*RN1p*RS146pe*RSp463e + 16./27.*m11*pow(m55,2)*
         L7r*RN1p*RSp461e*RS146pe + 16./27.*m11*pow(m55,2)*L7r*RN1e*
         RS146pe*RSe463p + 16./27.*m11*pow(m55,2)*L7r*RN1e*RSe461p*
         RS146pe + 32./27.*m11*pow(m55,2)*L7r*RN13*RS146pe*RS346pe + 16.
         /27.*m11*pow(m55,2)*L6r*RN1p*RS146pe*RSp463e );

      M13p6x23L +=  + Ab(m11,mu2) * ( 16./27.*m11*pow(m55,2)*L6r*RN1p*
         RSp461e*RS146pe + 16./27.*m11*pow(m55,2)*L6r*RN1e*RS146pe*
         RSe463p + 16./27.*m11*pow(m55,2)*L6r*RN1e*RSe461p*RS146pe + 32.
         /27.*m11*pow(m55,2)*L6r*RN13*RS146pe*RS346pe - 16./9.*m11*m44*
         L7r*RSe563p - 16./9.*m11*m44*L7r*RSp563e - 16./9.*m11*m44*L7r*
         RS156pe + 16./9.*m11*m44*L7r*RSe561p + 16./9.*m11*m44*L7r*
         RSp561e - 32./9.*m11*m44*L6r*RS14563pe - 16./9.*m11*m44*L6r*
         RSe45613p - 16./9.*m11*m44*L6r*RSp45613e - 8./9.*m11*m44*L6r*
         RS1456pee - 8./9.*m11*m44*L6r*RS1456ppe + 16./9.*m11*m44*L6r*
         RSe45611p + 16./9.*m11*m44*L6r*RSp45611e + 8./9.*m11*m44*L6r*
         RS145pe + 8./9.*m11*m44*L6r*RS146pe + 8./9.*m11*m44*L6r*
         RS156pe - 8./27.*m11*m44*L5r*RS156pe*RS356pe - 16./27.*m11*m44
         *L5r*RS156pe*RSe563p - 16./27.*m11*m44*L5r*RS156pe*RSp563e + 4.
         /27.*m11*m44*L5r*pow(RS156pe,2) + 8./9.*m11*m44*L4r*RS34561pe
          + 40./9.*m11*m44*L4r*RS14563pe + 16./9.*m11*m44*L4r*RSe45613p
          + 16./9.*m11*m44*L4r*RSp45613e );

      M13p6x23L +=  + Ab(m11,mu2) * ( 4./9.*m11*m44*L4r*RS1456pee + 4./
         9.*m11*m44*L4r*RS1456ppe - 4./9.*m11*m44*L4r*RS145pe - 8./27.*
         m11*m44*L4r*RS145pe*RS345pe - 16./27.*m11*m44*L4r*RS145pe*
         RSe453p - 16./27.*m11*m44*L4r*RS145pe*RSp453e + 4./27.*m11*m44
         *L4r*pow(RS145pe,2) - 4./9.*m11*m44*L4r*RS146pe - 8./27.*m11*
         m44*L4r*RS146pe*RS346pe - 16./27.*m11*m44*L4r*RS146pe*RSe463p
          - 16./27.*m11*m44*L4r*RS146pe*RSp463e + 4./27.*m11*m44*L4r*
         pow(RS146pe,2) - 4./9.*m11*m44*L4r*RS156pe - 8./27.*m11*m44*
         L4r*RS156pe*RS356pe - 16./27.*m11*m44*L4r*RS156pe*RSe563p - 16.
         /27.*m11*m44*L4r*RS156pe*RSp563e + 4./27.*m11*m44*L4r*pow(
         RS156pe,2) - 16./9.*m11*m44*mp2*L7r*RN1p*RSp563e + 16./9.*m11*
         m44*mp2*L7r*RN1p*RSp561e - 16./9.*m11*m44*mp2*L6r*RN1p*
         RSp45613e + 16./9.*m11*m44*mp2*L6r*RN1p*RSp45611e - 16./27.*
         m11*m44*mp2*L5r*RN1p*RS156pe*RSp563e + 16./9.*m11*m44*mp2*L4r*
         RN1p*RSp45613e - 16./27.*m11*m44*mp2*L4r*RN1p*RS145pe*RSp453e
          - 16./27.*m11*m44*mp2*L4r*RN1p*RS146pe*RSp463e );

      M13p6x23L +=  + Ab(m11,mu2) * (  - 16./27.*m11*m44*mp2*L4r*RN1p*
         RS156pe*RSp563e - 16./9.*m11*m44*me2*L7r*RN1e*RSe563p + 16./9.
         *m11*m44*me2*L7r*RN1e*RSe561p - 16./9.*m11*m44*me2*L6r*RN1e*
         RSe45613p + 16./9.*m11*m44*me2*L6r*RN1e*RSe45611p - 16./27.*
         m11*m44*me2*L5r*RN1e*RS156pe*RSe563p + 16./9.*m11*m44*me2*L4r*
         RN1e*RSe45613p - 16./27.*m11*m44*me2*L4r*RN1e*RS145pe*RSe453p
          - 16./27.*m11*m44*me2*L4r*RN1e*RS146pe*RSe463p - 16./27.*m11*
         m44*me2*L4r*RN1e*RS156pe*RSe563p + 16./27.*m11*m44*m66*L7r*
         RN1p*RS145pe*RSp563e + 16./27.*m11*m44*m66*L7r*RN1p*RS156pe*
         RSp453e + 16./27.*m11*m44*m66*L7r*RN1p*RSp561e*RS145pe + 16./
         27.*m11*m44*m66*L7r*RN1p*RSp451e*RS156pe + 16./27.*m11*m44*m66
         *L7r*RN1e*RS145pe*RSe563p + 16./27.*m11*m44*m66*L7r*RN1e*
         RS156pe*RSe453p + 16./27.*m11*m44*m66*L7r*RN1e*RSe561p*RS145pe
          + 16./27.*m11*m44*m66*L7r*RN1e*RSe451p*RS156pe + 32./27.*m11*
         m44*m66*L7r*RN13*RS145pe*RS356pe + 32./27.*m11*m44*m66*L7r*
         RN13*RS156pe*RS345pe );

      M13p6x23L +=  + Ab(m11,mu2) * ( 16./27.*m11*m44*m66*L6r*RN1p*
         RS145pe*RSp453e + 16./27.*m11*m44*m66*L6r*RN1p*RS156pe*RSp563e
          + 16./27.*m11*m44*m66*L6r*RN1p*RSp561e*RS156pe + 16./27.*m11*
         m44*m66*L6r*RN1p*RSp451e*RS145pe + 16./27.*m11*m44*m66*L6r*
         RN1e*RS145pe*RSe453p + 16./27.*m11*m44*m66*L6r*RN1e*RS156pe*
         RSe563p + 16./27.*m11*m44*m66*L6r*RN1e*RSe561p*RS156pe + 16./
         27.*m11*m44*m66*L6r*RN1e*RSe451p*RS145pe + 32./27.*m11*m44*m66
         *L6r*RN13*RS145pe*RS345pe + 32./27.*m11*m44*m66*L6r*RN13*
         RS156pe*RS356pe + 16./27.*m11*m44*m55*L7r*RN1p*RS146pe*RSp563e
          + 16./27.*m11*m44*m55*L7r*RN1p*RS156pe*RSp463e + 16./27.*m11*
         m44*m55*L7r*RN1p*RSp561e*RS146pe + 16./27.*m11*m44*m55*L7r*
         RN1p*RSp461e*RS156pe + 16./27.*m11*m44*m55*L7r*RN1e*RS146pe*
         RSe563p + 16./27.*m11*m44*m55*L7r*RN1e*RS156pe*RSe463p + 16./
         27.*m11*m44*m55*L7r*RN1e*RSe561p*RS146pe + 16./27.*m11*m44*m55
         *L7r*RN1e*RSe461p*RS156pe + 32./27.*m11*m44*m55*L7r*RN13*
         RS146pe*RS356pe );

      M13p6x23L +=  + Ab(m11,mu2) * ( 32./27.*m11*m44*m55*L7r*RN13*
         RS156pe*RS346pe + 16./27.*m11*m44*m55*L6r*RN1p*RS146pe*RSp463e
          + 16./27.*m11*m44*m55*L6r*RN1p*RS156pe*RSp563e + 16./27.*m11*
         m44*m55*L6r*RN1p*RSp561e*RS156pe + 16./27.*m11*m44*m55*L6r*
         RN1p*RSp461e*RS146pe + 16./27.*m11*m44*m55*L6r*RN1e*RS146pe*
         RSe463p + 16./27.*m11*m44*m55*L6r*RN1e*RS156pe*RSe563p + 16./
         27.*m11*m44*m55*L6r*RN1e*RSe561p*RS156pe + 16./27.*m11*m44*m55
         *L6r*RN1e*RSe461p*RS146pe + 32./27.*m11*m44*m55*L6r*RN13*
         RS146pe*RS346pe + 32./27.*m11*m44*m55*L6r*RN13*RS156pe*RS356pe
          + 16./27.*m11*pow(m44,2)*L8r*RN1p*RS156pe*RSp563e + 16./27.*
         m11*pow(m44,2)*L8r*RN1p*RSp561e*RS156pe + 16./27.*m11*pow(
         m44,2)*L8r*RN1e*RS156pe*RSe563p + 16./27.*m11*pow(m44,2)*L8r*
         RN1e*RSe561p*RS156pe + 32./27.*m11*pow(m44,2)*L8r*RN13*RS156pe
         *RS356pe + 16./27.*m11*pow(m44,2)*L7r*RN1p*RS156pe*RSp563e + 
         16./27.*m11*pow(m44,2)*L7r*RN1p*RSp561e*RS156pe + 16./27.*m11*
         pow(m44,2)*L7r*RN1e*RS156pe*RSe563p );

      M13p6x23L +=  + Ab(m11,mu2) * ( 16./27.*m11*pow(m44,2)*L7r*RN1e*
         RSe561p*RS156pe + 32./27.*m11*pow(m44,2)*L7r*RN13*RS156pe*
         RS356pe + 16./27.*m11*pow(m44,2)*L6r*RN1p*RS156pe*RSp563e + 16.
         /27.*m11*pow(m44,2)*L6r*RN1p*RSp561e*RS156pe + 16./27.*m11*
         pow(m44,2)*L6r*RN1e*RS156pe*RSe563p + 16./27.*m11*pow(m44,2)*
         L6r*RN1e*RSe561p*RS156pe + 32./27.*m11*pow(m44,2)*L6r*RN13*
         RS156pe*RS356pe - 16./3.*m11*m33*L8r*RS14563pe + 8./3.*m11*m33
         *L8r*RS1456pee + 8./3.*m11*m33*L8r*RS1456ppe - 8./3.*m11*m33*
         L8r*RS145pe - 8./3.*m11*m33*L8r*RS146pe - 8./3.*m11*m33*L8r*
         RS156pe + 16./3.*m11*m33*L7r + 8./3.*m11*m33*L5r*RS14563pe + 8.
         /9.*m11*m33*L5r*RSe45613p + 8./9.*m11*m33*L5r*RSp45613e - 4./3.
         *m11*m33*L5r*RS1456pee - 4./3.*m11*m33*L5r*RS1456ppe + 4./3.*
         m11*m33*L5r*RS145pe + 4./3.*m11*m33*L5r*RS146pe + 4./3.*m11*
         m33*L5r*RS156pe - 2./3.*m11*m33*L3r*RS14563pe + 5./3.*m11*m33*
         L3r*RS1456pee + 5./3.*m11*m33*L3r*RS1456ppe - 5./3.*m11*m33*
         L3r*RS145pe );

      M13p6x23L +=  + Ab(m11,mu2) * (  - 5./3.*m11*m33*L3r*RS146pe - 5./
         3.*m11*m33*L3r*RS156pe - 8./3.*m11*m33*L0r*RS14563pe + 2./3.*
         m11*m33*L0r*RS1456pee + 2./3.*m11*m33*L0r*RS1456ppe - 2./3.*
         m11*m33*L0r*RS145pe - 2./3.*m11*m33*L0r*RS146pe - 2./3.*m11*
         m33*L0r*RS156pe + 8./9.*m11*m33*mp2*L5r*RN1p*RSp45613e + 8./9.
         *m11*m33*me2*L5r*RN1e*RSe45613p - 16./9.*m11*m33*m66*L7r*RN1p*
         RSp453e - 16./9.*m11*m33*m66*L7r*RN1e*RSe453p - 16./9.*m11*m33
         *m66*L7r*RN13*RS345pe - 32./9.*m11*m33*m66*L7r*RN13*RS145pe - 
         16./9.*m11*m33*m66*L6r*RN1p*RSp45613e - 16./9.*m11*m33*m66*L6r
         *RN1e*RSe45613p - 16./9.*m11*m33*m66*L6r*RN13*RS34561pe - 32./
         9.*m11*m33*m66*L6r*RN13*RS14563pe - 16./27.*m11*m33*m66*L5r*
         RN13*RS145pe*RS345pe + 16./9.*m11*m33*m66*L4r*RN13*RS34561pe
          + 16./9.*m11*m33*m66*L4r*RN13*RS14563pe - 16./27.*m11*m33*m66
         *L4r*RN13*RS145pe*RS345pe - 16./27.*m11*m33*m66*L4r*RN13*
         RS146pe*RS346pe - 16./27.*m11*m33*m66*L4r*RN13*RS156pe*RS356pe
          - 16./9.*m11*m33*m55*L7r*RN1p*RSp463e );

      M13p6x23L +=  + Ab(m11,mu2) * (  - 16./9.*m11*m33*m55*L7r*RN1e*
         RSe463p - 16./9.*m11*m33*m55*L7r*RN13*RS346pe - 32./9.*m11*m33
         *m55*L7r*RN13*RS146pe - 16./9.*m11*m33*m55*L6r*RN1p*RSp45613e
          - 16./9.*m11*m33*m55*L6r*RN1e*RSe45613p - 16./9.*m11*m33*m55*
         L6r*RN13*RS34561pe - 32./9.*m11*m33*m55*L6r*RN13*RS14563pe - 
         16./27.*m11*m33*m55*L5r*RN13*RS146pe*RS346pe + 16./9.*m11*m33*
         m55*L4r*RN13*RS34561pe + 16./9.*m11*m33*m55*L4r*RN13*RS14563pe
          - 16./27.*m11*m33*m55*L4r*RN13*RS145pe*RS345pe - 16./27.*m11*
         m33*m55*L4r*RN13*RS146pe*RS346pe - 16./27.*m11*m33*m55*L4r*
         RN13*RS156pe*RS356pe - 16./9.*m11*m33*m44*L7r*RN1p*RSp563e - 
         16./9.*m11*m33*m44*L7r*RN1e*RSe563p - 16./9.*m11*m33*m44*L7r*
         RN13*RS356pe - 32./9.*m11*m33*m44*L7r*RN13*RS156pe - 16./9.*
         m11*m33*m44*L6r*RN1p*RSp45613e - 16./9.*m11*m33*m44*L6r*RN1e*
         RSe45613p - 16./9.*m11*m33*m44*L6r*RN13*RS34561pe - 32./9.*m11
         *m33*m44*L6r*RN13*RS14563pe - 16./27.*m11*m33*m44*L5r*RN13*
         RS156pe*RS356pe );

      M13p6x23L +=  + Ab(m11,mu2) * ( 16./9.*m11*m33*m44*L4r*RN13*
         RS34561pe + 16./9.*m11*m33*m44*L4r*RN13*RS14563pe - 16./27.*
         m11*m33*m44*L4r*RN13*RS145pe*RS345pe - 16./27.*m11*m33*m44*L4r
         *RN13*RS146pe*RS346pe - 16./27.*m11*m33*m44*L4r*RN13*RS156pe*
         RS356pe - 32./9.*m11*pow(m33,2)*L8r*RN13*RS14563pe + 16./3.*
         m11*pow(m33,2)*L7r*RN13 + 16./9.*m11*pow(m33,2)*L5r*RN13*
         RS14563pe - 32./9.*pow(m11,2)*L8r*RS14563pe - 16./9.*pow(
         m11,2)*L8r*RSe45613p - 16./9.*pow(m11,2)*L8r*RSp45613e + 16./9.
         *pow(m11,2)*L8r*RS1456pee + 16./9.*pow(m11,2)*L8r*RS1456ppe + 
         16./9.*pow(m11,2)*L8r*RSe45611p + 16./9.*pow(m11,2)*L8r*
         RSp45611e - 16./9.*pow(m11,2)*L8r*RS145pe - 16./9.*pow(m11,2)*
         L8r*RS146pe - 16./9.*pow(m11,2)*L8r*RS156pe + 8*pow(m11,2)*L7r
          + 8./9.*pow(m11,2)*L5r*RS34561pe + 28./9.*pow(m11,2)*L5r*
         RS14563pe + 16./9.*pow(m11,2)*L5r*RSe45613p + 16./9.*pow(
         m11,2)*L5r*RSp45613e - 8./9.*pow(m11,2)*L5r*RS1456pee - 8./9.*
         pow(m11,2)*L5r*RS1456ppe );

      M13p6x23L +=  + Ab(m11,mu2) * ( 8./9.*pow(m11,2)*L5r*RS145pe + 8./
         9.*pow(m11,2)*L5r*RS146pe + 8./9.*pow(m11,2)*L5r*RS156pe - 2./
         3.*pow(m11,2)*L3r*RS14563pe + 5./3.*pow(m11,2)*L3r*RS1456pee
          + 5./3.*pow(m11,2)*L3r*RS1456ppe - 5./3.*pow(m11,2)*L3r*
         RS145pe - 5./3.*pow(m11,2)*L3r*RS146pe - 5./3.*pow(m11,2)*L3r*
         RS156pe - 8./3.*pow(m11,2)*L0r*RS14563pe + 2./3.*pow(m11,2)*
         L0r*RS1456pee + 2./3.*pow(m11,2)*L0r*RS1456ppe - 2./3.*pow(
         m11,2)*L0r*RS145pe - 2./3.*pow(m11,2)*L0r*RS146pe - 2./3.*pow(
         m11,2)*L0r*RS156pe - 16./9.*pow(m11,2)*mp2*L8r*RN1p*RSp45613e
          + 16./9.*pow(m11,2)*mp2*L8r*RN1p*RSp45611e + 16./9.*pow(
         m11,2)*mp2*L5r*RN1p*RSp45613e - 16./9.*pow(m11,2)*me2*L8r*RN1e
         *RSe45613p + 16./9.*pow(m11,2)*me2*L8r*RN1e*RSe45611p + 16./9.
         *pow(m11,2)*me2*L5r*RN1e*RSe45613p - 16./9.*pow(m11,2)*m66*L7r
         *RN1p*RSp453e - 16./9.*pow(m11,2)*m66*L7r*RN1p*RSp451e - 16./9.
         *pow(m11,2)*m66*L7r*RN1e*RSe453p - 16./9.*pow(m11,2)*m66*L7r*
         RN1e*RSe451p );

      M13p6x23L +=  + Ab(m11,mu2) * (  - 32./9.*pow(m11,2)*m66*L7r*RN13
         *RS345pe - 16./9.*pow(m11,2)*m66*L6r*RN1p*RSp45613e - 16./9.*
         pow(m11,2)*m66*L6r*RN1p*RSp45611e - 16./9.*pow(m11,2)*m66*L6r*
         RN1e*RSe45613p - 16./9.*pow(m11,2)*m66*L6r*RN1e*RSe45611p - 32.
         /9.*pow(m11,2)*m66*L6r*RN13*RS34561pe - 8./27.*pow(m11,2)*m66*
         L5r*RN13*RS145pe*RS345pe + 8./9.*pow(m11,2)*m66*L4r*RN13*
         RS34561pe + 8./9.*pow(m11,2)*m66*L4r*RN13*RS14563pe - 8./27.*
         pow(m11,2)*m66*L4r*RN13*RS145pe*RS345pe - 8./27.*pow(m11,2)*
         m66*L4r*RN13*RS146pe*RS346pe - 8./27.*pow(m11,2)*m66*L4r*RN13*
         RS156pe*RS356pe - 16./9.*pow(m11,2)*m55*L7r*RN1p*RSp463e - 16./
         9.*pow(m11,2)*m55*L7r*RN1p*RSp461e - 16./9.*pow(m11,2)*m55*L7r
         *RN1e*RSe463p - 16./9.*pow(m11,2)*m55*L7r*RN1e*RSe461p - 32./9.
         *pow(m11,2)*m55*L7r*RN13*RS346pe - 16./9.*pow(m11,2)*m55*L6r*
         RN1p*RSp45613e - 16./9.*pow(m11,2)*m55*L6r*RN1p*RSp45611e - 16.
         /9.*pow(m11,2)*m55*L6r*RN1e*RSe45613p - 16./9.*pow(m11,2)*m55*
         L6r*RN1e*RSe45611p );

      M13p6x23L +=  + Ab(m11,mu2) * (  - 32./9.*pow(m11,2)*m55*L6r*RN13
         *RS34561pe - 8./27.*pow(m11,2)*m55*L5r*RN13*RS146pe*RS346pe + 
         8./9.*pow(m11,2)*m55*L4r*RN13*RS34561pe + 8./9.*pow(m11,2)*m55
         *L4r*RN13*RS14563pe - 8./27.*pow(m11,2)*m55*L4r*RN13*RS145pe*
         RS345pe - 8./27.*pow(m11,2)*m55*L4r*RN13*RS146pe*RS346pe - 8./
         27.*pow(m11,2)*m55*L4r*RN13*RS156pe*RS356pe - 16./9.*pow(
         m11,2)*m44*L7r*RN1p*RSp563e - 16./9.*pow(m11,2)*m44*L7r*RN1p*
         RSp561e - 16./9.*pow(m11,2)*m44*L7r*RN1e*RSe563p - 16./9.*pow(
         m11,2)*m44*L7r*RN1e*RSe561p - 32./9.*pow(m11,2)*m44*L7r*RN13*
         RS356pe - 16./9.*pow(m11,2)*m44*L6r*RN1p*RSp45613e - 16./9.*
         pow(m11,2)*m44*L6r*RN1p*RSp45611e - 16./9.*pow(m11,2)*m44*L6r*
         RN1e*RSe45613p - 16./9.*pow(m11,2)*m44*L6r*RN1e*RSe45611p - 32.
         /9.*pow(m11,2)*m44*L6r*RN13*RS34561pe - 8./27.*pow(m11,2)*m44*
         L5r*RN13*RS156pe*RS356pe + 8./9.*pow(m11,2)*m44*L4r*RN13*
         RS34561pe + 8./9.*pow(m11,2)*m44*L4r*RN13*RS14563pe - 8./27.*
         pow(m11,2)*m44*L4r*RN13*RS145pe*RS345pe );

      M13p6x23L +=  + Ab(m11,mu2) * (  - 8./27.*pow(m11,2)*m44*L4r*RN13
         *RS146pe*RS346pe - 8./27.*pow(m11,2)*m44*L4r*RN13*RS156pe*
         RS356pe - 16./9.*pow(m11,2)*m33*L8r*RN1p*RSp45613e - 16./9.*
         pow(m11,2)*m33*L8r*RN1e*RSe45613p - 16./9.*pow(m11,2)*m33*L8r*
         RN13*RS34561pe + 32./3.*pow(m11,2)*m33*L7r*RN13 + 16./9.*pow(
         m11,2)*m33*L5r*RN13*RS34561pe + 8./9.*pow(m11,2)*m33*L5r*RN13*
         RS14563pe - 16./9.*pow(m11,3)*L8r*RN1p*RSp45613e - 16./9.*pow(
         m11,3)*L8r*RN1p*RSp45611e - 16./9.*pow(m11,3)*L8r*RN1e*
         RSe45613p - 16./9.*pow(m11,3)*L8r*RN1e*RSe45611p - 32./9.*pow(
         m11,3)*L8r*RN13*RS34561pe + 8./9.*pow(m11,3)*L5r*RN13*
         RS34561pe );

      M13p6x23L +=  + Ab(m13,mu2) * ( 8*pow(m33,2)*L6r + 5*pow(m33,2)*
         L2r + 2*pow(m33,2)*L1r + 16*m11*m33*L6r + 10*m11*m33*L2r + 4*
         m11*m33*L1r + 8*pow(m11,2)*L6r + 5*pow(m11,2)*L2r + 2*pow(
         m11,2)*L1r );

      M13p6x23L +=  + Ab(m14,mu2) * ( 4*m33*m44*L8r - 2*m33*m44*L5r + 5.
         /2.*m33*m44*L3r + m33*m44*L0r + 4*m11*m44*L8r - 2*m11*m44*L5r
          + 5./2.*m11*m44*L3r + m11*m44*L0r + 4*m11*m33*L8r - 2*m11*m33
         *L5r + 5./2.*m11*m33*L3r + m11*m33*L0r + 4*pow(m11,2)*L8r - 2*
         pow(m11,2)*L5r + 5./2.*pow(m11,2)*L3r + pow(m11,2)*L0r );

      M13p6x23L +=  + Ab(m15,mu2) * ( 4*m33*m55*L8r - 2*m33*m55*L5r + 5.
         /2.*m33*m55*L3r + m33*m55*L0r + 4*m11*m55*L8r - 2*m11*m55*L5r
          + 5./2.*m11*m55*L3r + m11*m55*L0r + 4*m11*m33*L8r - 2*m11*m33
         *L5r + 5./2.*m11*m33*L3r + m11*m33*L0r + 4*pow(m11,2)*L8r - 2*
         pow(m11,2)*L5r + 5./2.*pow(m11,2)*L3r + pow(m11,2)*L0r );

      M13p6x23L +=  + Ab(m16,mu2) * ( 4*m33*m66*L8r - 2*m33*m66*L5r + 5.
         /2.*m33*m66*L3r + m33*m66*L0r + 4*m11*m66*L8r - 2*m11*m66*L5r
          + 5./2.*m11*m66*L3r + m11*m66*L0r + 4*m11*m33*L8r - 2*m11*m33
         *L5r + 5./2.*m11*m33*L3r + m11*m33*L0r + 4*pow(m11,2)*L8r - 2*
         pow(m11,2)*L5r + 5./2.*pow(m11,2)*L3r + pow(m11,2)*L0r );

      M13p6x23L +=  + Ab(m33,mu2) * (  - 8./9.*m66*L4r*RD3456ep - 8./27.
         *pow(m66,2)*L8r*pow(RS345pe,2) - 16./27.*pow(m66,2)*L8r*
         RSe453p*RS345pe - 16./27.*pow(m66,2)*L8r*RSp453e*RS345pe + 16./
         27.*pow(m66,2)*L8r*RS145pe*RS345pe + 16./27.*pow(m66,2)*L8r*
         RSe451p*RS345pe + 16./27.*pow(m66,2)*L8r*RSp451e*RS345pe - 8./
         27.*pow(m66,2)*L7r*pow(RS345pe,2) - 16./27.*pow(m66,2)*L7r*
         RSe453p*RS345pe - 16./27.*pow(m66,2)*L7r*RSp453e*RS345pe + 16./
         27.*pow(m66,2)*L7r*RS145pe*RS345pe + 16./27.*pow(m66,2)*L7r*
         RSe451p*RS345pe + 16./27.*pow(m66,2)*L7r*RSp451e*RS345pe - 8./
         27.*pow(m66,2)*L6r*pow(RS345pe,2) - 16./27.*pow(m66,2)*L6r*
         RSe453p*RS345pe - 16./27.*pow(m66,2)*L6r*RSp453e*RS345pe + 16./
         27.*pow(m66,2)*L6r*RS145pe*RS345pe + 16./27.*pow(m66,2)*L6r*
         RSe451p*RS345pe + 16./27.*pow(m66,2)*L6r*RSp451e*RS345pe - 16./
         27.*pow(m66,2)*mp2*L8r*RN3p*RSp453e*RS345pe + 16./27.*pow(
         m66,2)*mp2*L8r*RN3p*RSp451e*RS345pe - 16./27.*pow(m66,2)*mp2*
         L7r*RN3p*RSp453e*RS345pe );

      M13p6x23L +=  + Ab(m33,mu2) * ( 16./27.*pow(m66,2)*mp2*L7r*RN3p*
         RSp451e*RS345pe - 16./27.*pow(m66,2)*mp2*L6r*RN3p*RSp453e*
         RS345pe + 16./27.*pow(m66,2)*mp2*L6r*RN3p*RSp451e*RS345pe - 16.
         /27.*pow(m66,2)*me2*L8r*RN3e*RSe453p*RS345pe + 16./27.*pow(
         m66,2)*me2*L8r*RN3e*RSe451p*RS345pe - 16./27.*pow(m66,2)*me2*
         L7r*RN3e*RSe453p*RS345pe + 16./27.*pow(m66,2)*me2*L7r*RN3e*
         RSe451p*RS345pe - 16./27.*pow(m66,2)*me2*L6r*RN3e*RSe453p*
         RS345pe + 16./27.*pow(m66,2)*me2*L6r*RN3e*RSe451p*RS345pe - 8./
         9.*m55*L4r*RD3456ep - 16./27.*m55*m66*L7r*RS346pe*RS345pe - 16.
         /27.*m55*m66*L7r*RSe463p*RS345pe - 16./27.*m55*m66*L7r*RSe453p
         *RS346pe - 16./27.*m55*m66*L7r*RSp463e*RS345pe - 16./27.*m55*
         m66*L7r*RSp453e*RS346pe + 16./27.*m55*m66*L7r*RS145pe*RS346pe
          + 16./27.*m55*m66*L7r*RS146pe*RS345pe + 16./27.*m55*m66*L7r*
         RSe461p*RS345pe + 16./27.*m55*m66*L7r*RSe451p*RS346pe + 16./27.
         *m55*m66*L7r*RSp461e*RS345pe + 16./27.*m55*m66*L7r*RSp451e*
         RS346pe );

      M13p6x23L +=  + Ab(m33,mu2) * (  - 8./27.*m55*m66*L6r*pow(
         RS345pe,2) - 8./27.*m55*m66*L6r*pow(RS346pe,2) - 16./27.*m55*
         m66*L6r*RSe463p*RS346pe - 16./27.*m55*m66*L6r*RSe453p*RS345pe
          - 16./27.*m55*m66*L6r*RSp463e*RS346pe - 16./27.*m55*m66*L6r*
         RSp453e*RS345pe + 16./27.*m55*m66*L6r*RS145pe*RS345pe + 16./27.
         *m55*m66*L6r*RS146pe*RS346pe + 16./27.*m55*m66*L6r*RSe461p*
         RS346pe + 16./27.*m55*m66*L6r*RSe451p*RS345pe + 16./27.*m55*
         m66*L6r*RSp461e*RS346pe + 16./27.*m55*m66*L6r*RSp451e*RS345pe
          - 16./27.*m55*m66*mp2*L7r*RN3p*RSp463e*RS345pe - 16./27.*m55*
         m66*mp2*L7r*RN3p*RSp453e*RS346pe + 16./27.*m55*m66*mp2*L7r*
         RN3p*RSp461e*RS345pe + 16./27.*m55*m66*mp2*L7r*RN3p*RSp451e*
         RS346pe - 16./27.*m55*m66*mp2*L6r*RN3p*RSp463e*RS346pe - 16./
         27.*m55*m66*mp2*L6r*RN3p*RSp453e*RS345pe + 16./27.*m55*m66*mp2
         *L6r*RN3p*RSp461e*RS346pe + 16./27.*m55*m66*mp2*L6r*RN3p*
         RSp451e*RS345pe - 16./27.*m55*m66*me2*L7r*RN3e*RSe463p*RS345pe
          - 16./27.*m55*m66*me2*L7r*RN3e*RSe453p*RS346pe );

      M13p6x23L +=  + Ab(m33,mu2) * ( 16./27.*m55*m66*me2*L7r*RN3e*
         RSe461p*RS345pe + 16./27.*m55*m66*me2*L7r*RN3e*RSe451p*RS346pe
          - 16./27.*m55*m66*me2*L6r*RN3e*RSe463p*RS346pe - 16./27.*m55*
         m66*me2*L6r*RN3e*RSe453p*RS345pe + 16./27.*m55*m66*me2*L6r*
         RN3e*RSe461p*RS346pe + 16./27.*m55*m66*me2*L6r*RN3e*RSe451p*
         RS345pe - 8./27.*pow(m55,2)*L8r*pow(RS346pe,2) - 16./27.*pow(
         m55,2)*L8r*RSe463p*RS346pe - 16./27.*pow(m55,2)*L8r*RSp463e*
         RS346pe + 16./27.*pow(m55,2)*L8r*RS146pe*RS346pe + 16./27.*
         pow(m55,2)*L8r*RSe461p*RS346pe + 16./27.*pow(m55,2)*L8r*
         RSp461e*RS346pe - 8./27.*pow(m55,2)*L7r*pow(RS346pe,2) - 16./
         27.*pow(m55,2)*L7r*RSe463p*RS346pe - 16./27.*pow(m55,2)*L7r*
         RSp463e*RS346pe + 16./27.*pow(m55,2)*L7r*RS146pe*RS346pe + 16./
         27.*pow(m55,2)*L7r*RSe461p*RS346pe + 16./27.*pow(m55,2)*L7r*
         RSp461e*RS346pe - 8./27.*pow(m55,2)*L6r*pow(RS346pe,2) - 16./
         27.*pow(m55,2)*L6r*RSe463p*RS346pe - 16./27.*pow(m55,2)*L6r*
         RSp463e*RS346pe );

      M13p6x23L +=  + Ab(m33,mu2) * ( 16./27.*pow(m55,2)*L6r*RS146pe*
         RS346pe + 16./27.*pow(m55,2)*L6r*RSe461p*RS346pe + 16./27.*
         pow(m55,2)*L6r*RSp461e*RS346pe - 16./27.*pow(m55,2)*mp2*L8r*
         RN3p*RSp463e*RS346pe + 16./27.*pow(m55,2)*mp2*L8r*RN3p*RSp461e
         *RS346pe - 16./27.*pow(m55,2)*mp2*L7r*RN3p*RSp463e*RS346pe + 
         16./27.*pow(m55,2)*mp2*L7r*RN3p*RSp461e*RS346pe - 16./27.*pow(
         m55,2)*mp2*L6r*RN3p*RSp463e*RS346pe + 16./27.*pow(m55,2)*mp2*
         L6r*RN3p*RSp461e*RS346pe - 16./27.*pow(m55,2)*me2*L8r*RN3e*
         RSe463p*RS346pe + 16./27.*pow(m55,2)*me2*L8r*RN3e*RSe461p*
         RS346pe - 16./27.*pow(m55,2)*me2*L7r*RN3e*RSe463p*RS346pe + 16.
         /27.*pow(m55,2)*me2*L7r*RN3e*RSe461p*RS346pe - 16./27.*pow(
         m55,2)*me2*L6r*RN3e*RSe463p*RS346pe + 16./27.*pow(m55,2)*me2*
         L6r*RN3e*RSe461p*RS346pe - 8./9.*m44*L4r*RD3456ep - 16./27.*
         m44*m66*L7r*RS356pe*RS345pe - 16./27.*m44*m66*L7r*RSe563p*
         RS345pe - 16./27.*m44*m66*L7r*RSe453p*RS356pe - 16./27.*m44*
         m66*L7r*RSp563e*RS345pe );

      M13p6x23L +=  + Ab(m33,mu2) * (  - 16./27.*m44*m66*L7r*RSp453e*
         RS356pe + 16./27.*m44*m66*L7r*RS145pe*RS356pe + 16./27.*m44*
         m66*L7r*RS156pe*RS345pe + 16./27.*m44*m66*L7r*RSe561p*RS345pe
          + 16./27.*m44*m66*L7r*RSe451p*RS356pe + 16./27.*m44*m66*L7r*
         RSp561e*RS345pe + 16./27.*m44*m66*L7r*RSp451e*RS356pe - 8./27.
         *m44*m66*L6r*pow(RS345pe,2) - 8./27.*m44*m66*L6r*pow(
         RS356pe,2) - 16./27.*m44*m66*L6r*RSe563p*RS356pe - 16./27.*m44
         *m66*L6r*RSe453p*RS345pe - 16./27.*m44*m66*L6r*RSp563e*RS356pe
          - 16./27.*m44*m66*L6r*RSp453e*RS345pe + 16./27.*m44*m66*L6r*
         RS145pe*RS345pe + 16./27.*m44*m66*L6r*RS156pe*RS356pe + 16./27.
         *m44*m66*L6r*RSe561p*RS356pe + 16./27.*m44*m66*L6r*RSe451p*
         RS345pe + 16./27.*m44*m66*L6r*RSp561e*RS356pe + 16./27.*m44*
         m66*L6r*RSp451e*RS345pe - 16./27.*m44*m66*mp2*L7r*RN3p*RSp563e
         *RS345pe - 16./27.*m44*m66*mp2*L7r*RN3p*RSp453e*RS356pe + 16./
         27.*m44*m66*mp2*L7r*RN3p*RSp561e*RS345pe + 16./27.*m44*m66*mp2
         *L7r*RN3p*RSp451e*RS356pe );

      M13p6x23L +=  + Ab(m33,mu2) * (  - 16./27.*m44*m66*mp2*L6r*RN3p*
         RSp563e*RS356pe - 16./27.*m44*m66*mp2*L6r*RN3p*RSp453e*RS345pe
          + 16./27.*m44*m66*mp2*L6r*RN3p*RSp561e*RS356pe + 16./27.*m44*
         m66*mp2*L6r*RN3p*RSp451e*RS345pe - 16./27.*m44*m66*me2*L7r*
         RN3e*RSe563p*RS345pe - 16./27.*m44*m66*me2*L7r*RN3e*RSe453p*
         RS356pe + 16./27.*m44*m66*me2*L7r*RN3e*RSe561p*RS345pe + 16./
         27.*m44*m66*me2*L7r*RN3e*RSe451p*RS356pe - 16./27.*m44*m66*me2
         *L6r*RN3e*RSe563p*RS356pe - 16./27.*m44*m66*me2*L6r*RN3e*
         RSe453p*RS345pe + 16./27.*m44*m66*me2*L6r*RN3e*RSe561p*RS356pe
          + 16./27.*m44*m66*me2*L6r*RN3e*RSe451p*RS345pe - 16./27.*m44*
         m55*L7r*RS356pe*RS346pe - 16./27.*m44*m55*L7r*RSe563p*RS346pe
          - 16./27.*m44*m55*L7r*RSe463p*RS356pe - 16./27.*m44*m55*L7r*
         RSp563e*RS346pe - 16./27.*m44*m55*L7r*RSp463e*RS356pe + 16./27.
         *m44*m55*L7r*RS146pe*RS356pe + 16./27.*m44*m55*L7r*RS156pe*
         RS346pe + 16./27.*m44*m55*L7r*RSe561p*RS346pe + 16./27.*m44*
         m55*L7r*RSe461p*RS356pe );

      M13p6x23L +=  + Ab(m33,mu2) * ( 16./27.*m44*m55*L7r*RSp561e*
         RS346pe + 16./27.*m44*m55*L7r*RSp461e*RS356pe - 8./27.*m44*m55
         *L6r*pow(RS346pe,2) - 8./27.*m44*m55*L6r*pow(RS356pe,2) - 16./
         27.*m44*m55*L6r*RSe563p*RS356pe - 16./27.*m44*m55*L6r*RSe463p*
         RS346pe - 16./27.*m44*m55*L6r*RSp563e*RS356pe - 16./27.*m44*
         m55*L6r*RSp463e*RS346pe + 16./27.*m44*m55*L6r*RS146pe*RS346pe
          + 16./27.*m44*m55*L6r*RS156pe*RS356pe + 16./27.*m44*m55*L6r*
         RSe561p*RS356pe + 16./27.*m44*m55*L6r*RSe461p*RS346pe + 16./27.
         *m44*m55*L6r*RSp561e*RS356pe + 16./27.*m44*m55*L6r*RSp461e*
         RS346pe - 16./27.*m44*m55*mp2*L7r*RN3p*RSp563e*RS346pe - 16./
         27.*m44*m55*mp2*L7r*RN3p*RSp463e*RS356pe + 16./27.*m44*m55*mp2
         *L7r*RN3p*RSp561e*RS346pe + 16./27.*m44*m55*mp2*L7r*RN3p*
         RSp461e*RS356pe - 16./27.*m44*m55*mp2*L6r*RN3p*RSp563e*RS356pe
          - 16./27.*m44*m55*mp2*L6r*RN3p*RSp463e*RS346pe + 16./27.*m44*
         m55*mp2*L6r*RN3p*RSp561e*RS356pe + 16./27.*m44*m55*mp2*L6r*
         RN3p*RSp461e*RS346pe );

      M13p6x23L +=  + Ab(m33,mu2) * (  - 16./27.*m44*m55*me2*L7r*RN3e*
         RSe563p*RS346pe - 16./27.*m44*m55*me2*L7r*RN3e*RSe463p*RS356pe
          + 16./27.*m44*m55*me2*L7r*RN3e*RSe561p*RS346pe + 16./27.*m44*
         m55*me2*L7r*RN3e*RSe461p*RS356pe - 16./27.*m44*m55*me2*L6r*
         RN3e*RSe563p*RS356pe - 16./27.*m44*m55*me2*L6r*RN3e*RSe463p*
         RS346pe + 16./27.*m44*m55*me2*L6r*RN3e*RSe561p*RS356pe + 16./
         27.*m44*m55*me2*L6r*RN3e*RSe461p*RS346pe - 8./27.*pow(m44,2)*
         L8r*pow(RS356pe,2) - 16./27.*pow(m44,2)*L8r*RSe563p*RS356pe - 
         16./27.*pow(m44,2)*L8r*RSp563e*RS356pe + 16./27.*pow(m44,2)*
         L8r*RS156pe*RS356pe + 16./27.*pow(m44,2)*L8r*RSe561p*RS356pe
          + 16./27.*pow(m44,2)*L8r*RSp561e*RS356pe - 8./27.*pow(m44,2)*
         L7r*pow(RS356pe,2) - 16./27.*pow(m44,2)*L7r*RSe563p*RS356pe - 
         16./27.*pow(m44,2)*L7r*RSp563e*RS356pe + 16./27.*pow(m44,2)*
         L7r*RS156pe*RS356pe + 16./27.*pow(m44,2)*L7r*RSe561p*RS356pe
          + 16./27.*pow(m44,2)*L7r*RSp561e*RS356pe - 8./27.*pow(m44,2)*
         L6r*pow(RS356pe,2) );

      M13p6x23L +=  + Ab(m33,mu2) * (  - 16./27.*pow(m44,2)*L6r*RSe563p
         *RS356pe - 16./27.*pow(m44,2)*L6r*RSp563e*RS356pe + 16./27.*
         pow(m44,2)*L6r*RS156pe*RS356pe + 16./27.*pow(m44,2)*L6r*
         RSe561p*RS356pe + 16./27.*pow(m44,2)*L6r*RSp561e*RS356pe - 16./
         27.*pow(m44,2)*mp2*L8r*RN3p*RSp563e*RS356pe + 16./27.*pow(
         m44,2)*mp2*L8r*RN3p*RSp561e*RS356pe - 16./27.*pow(m44,2)*mp2*
         L7r*RN3p*RSp563e*RS356pe + 16./27.*pow(m44,2)*mp2*L7r*RN3p*
         RSp561e*RS356pe - 16./27.*pow(m44,2)*mp2*L6r*RN3p*RSp563e*
         RS356pe + 16./27.*pow(m44,2)*mp2*L6r*RN3p*RSp561e*RS356pe - 16.
         /27.*pow(m44,2)*me2*L8r*RN3e*RSe563p*RS356pe + 16./27.*pow(
         m44,2)*me2*L8r*RN3e*RSe561p*RS356pe - 16./27.*pow(m44,2)*me2*
         L7r*RN3e*RSe563p*RS356pe + 16./27.*pow(m44,2)*me2*L7r*RN3e*
         RSe561p*RS356pe - 16./27.*pow(m44,2)*me2*L6r*RN3e*RSe563p*
         RS356pe + 16./27.*pow(m44,2)*me2*L6r*RN3e*RSe561p*RS356pe - 5./
         3.*m33*L3r*RD3456ep - 2./3.*m33*L0r*RD3456ep - 16./9.*m33*m66*
         L7r*RS345pe );

      M13p6x23L +=  + Ab(m33,mu2) * ( 16./9.*m33*m66*L7r*RSe453p + 16./
         9.*m33*m66*L7r*RSp453e - 16./9.*m33*m66*L7r*RS145pe - 16./9.*
         m33*m66*L7r*RSe451p - 16./9.*m33*m66*L7r*RSp451e - 32./9.*m33*
         m66*L6r*RS34561pe - 16./9.*m33*m66*L6r*RS14563pe - 16./9.*m33*
         m66*L6r*RSe45613p - 16./9.*m33*m66*L6r*RSp45613e - 8./9.*m33*
         m66*L6r*RS3456pee - 8./9.*m33*m66*L6r*RS3456ppe + 16./9.*m33*
         m66*L6r*RSe45633p + 16./9.*m33*m66*L6r*RSp45633e + 8./9.*m33*
         m66*L6r*RS345pe + 8./9.*m33*m66*L6r*RS346pe + 8./9.*m33*m66*
         L6r*RS356pe + 4./27.*m33*m66*L5r*pow(RS345pe,2) - 16./27.*m33*
         m66*L5r*RS145pe*RS345pe - 16./27.*m33*m66*L5r*RSe451p*RS345pe
          - 16./27.*m33*m66*L5r*RSp451e*RS345pe + 16./3.*m33*m66*L4r*
         RS34561pe + 16./9.*m33*m66*L4r*RS14563pe + 16./9.*m33*m66*L4r*
         RSe45613p + 16./9.*m33*m66*L4r*RSp45613e + 4./9.*m33*m66*L4r*
         RS3456pee + 4./9.*m33*m66*L4r*RS3456ppe - 4./9.*m33*m66*L4r*
         RS345pe + 4./27.*m33*m66*L4r*pow(RS345pe,2) - 4./9.*m33*m66*
         L4r*RS346pe );

      M13p6x23L +=  + Ab(m33,mu2) * ( 4./27.*m33*m66*L4r*pow(RS346pe,2)
          - 4./9.*m33*m66*L4r*RS356pe + 4./27.*m33*m66*L4r*pow(
         RS356pe,2) - 16./27.*m33*m66*L4r*RS145pe*RS345pe - 16./27.*m33
         *m66*L4r*RS146pe*RS346pe - 16./27.*m33*m66*L4r*RS156pe*RS356pe
          - 16./27.*m33*m66*L4r*RSe561p*RS356pe - 16./27.*m33*m66*L4r*
         RSe461p*RS346pe - 16./27.*m33*m66*L4r*RSe451p*RS345pe - 16./27.
         *m33*m66*L4r*RSp561e*RS356pe - 16./27.*m33*m66*L4r*RSp461e*
         RS346pe - 16./27.*m33*m66*L4r*RSp451e*RS345pe + 16./9.*m33*m66
         *mp2*L7r*RN3p*RSp453e - 16./9.*m33*m66*mp2*L7r*RN3p*RSp451e - 
         16./9.*m33*m66*mp2*L6r*RN3p*RSp45613e + 16./9.*m33*m66*mp2*L6r
         *RN3p*RSp45633e - 16./27.*m33*m66*mp2*L5r*RN3p*RSp451e*RS345pe
          + 16./9.*m33*m66*mp2*L4r*RN3p*RSp45613e - 16./27.*m33*m66*mp2
         *L4r*RN3p*RSp561e*RS356pe - 16./27.*m33*m66*mp2*L4r*RN3p*
         RSp461e*RS346pe - 16./27.*m33*m66*mp2*L4r*RN3p*RSp451e*RS345pe
          + 16./9.*m33*m66*me2*L7r*RN3e*RSe453p - 16./9.*m33*m66*me2*
         L7r*RN3e*RSe451p );

      M13p6x23L +=  + Ab(m33,mu2) * (  - 16./9.*m33*m66*me2*L6r*RN3e*
         RSe45613p + 16./9.*m33*m66*me2*L6r*RN3e*RSe45633p - 16./27.*
         m33*m66*me2*L5r*RN3e*RSe451p*RS345pe + 16./9.*m33*m66*me2*L4r*
         RN3e*RSe45613p - 16./27.*m33*m66*me2*L4r*RN3e*RSe561p*RS356pe
          - 16./27.*m33*m66*me2*L4r*RN3e*RSe461p*RS346pe - 16./27.*m33*
         m66*me2*L4r*RN3e*RSe451p*RS345pe + 16./27.*m33*pow(m66,2)*L8r*
         RN3p*RSp453e*RS345pe + 16./27.*m33*pow(m66,2)*L8r*RN3p*RSp451e
         *RS345pe + 16./27.*m33*pow(m66,2)*L8r*RN3e*RSe453p*RS345pe + 
         16./27.*m33*pow(m66,2)*L8r*RN3e*RSe451p*RS345pe - 16./27.*m33*
         pow(m66,2)*L8r*RN13*RS145pe*RS345pe + 16./27.*m33*pow(m66,2)*
         L7r*RN3p*RSp453e*RS345pe + 16./27.*m33*pow(m66,2)*L7r*RN3p*
         RSp451e*RS345pe + 16./27.*m33*pow(m66,2)*L7r*RN3e*RSe453p*
         RS345pe + 16./27.*m33*pow(m66,2)*L7r*RN3e*RSe451p*RS345pe - 16.
         /27.*m33*pow(m66,2)*L7r*RN13*RS145pe*RS345pe + 16./27.*m33*
         pow(m66,2)*L6r*RN3p*RSp453e*RS345pe + 16./27.*m33*pow(m66,2)*
         L6r*RN3p*RSp451e*RS345pe );

      M13p6x23L +=  + Ab(m33,mu2) * ( 16./27.*m33*pow(m66,2)*L6r*RN3e*
         RSe453p*RS345pe + 16./27.*m33*pow(m66,2)*L6r*RN3e*RSe451p*
         RS345pe - 16./27.*m33*pow(m66,2)*L6r*RN13*RS145pe*RS345pe - 16.
         /9.*m33*m55*L7r*RS346pe + 16./9.*m33*m55*L7r*RSe463p + 16./9.*
         m33*m55*L7r*RSp463e - 16./9.*m33*m55*L7r*RS146pe - 16./9.*m33*
         m55*L7r*RSe461p - 16./9.*m33*m55*L7r*RSp461e - 32./9.*m33*m55*
         L6r*RS34561pe - 16./9.*m33*m55*L6r*RS14563pe - 16./9.*m33*m55*
         L6r*RSe45613p - 16./9.*m33*m55*L6r*RSp45613e - 8./9.*m33*m55*
         L6r*RS3456pee - 8./9.*m33*m55*L6r*RS3456ppe + 16./9.*m33*m55*
         L6r*RSe45633p + 16./9.*m33*m55*L6r*RSp45633e + 8./9.*m33*m55*
         L6r*RS345pe + 8./9.*m33*m55*L6r*RS346pe + 8./9.*m33*m55*L6r*
         RS356pe + 4./27.*m33*m55*L5r*pow(RS346pe,2) - 16./27.*m33*m55*
         L5r*RS146pe*RS346pe - 16./27.*m33*m55*L5r*RSe461p*RS346pe - 16.
         /27.*m33*m55*L5r*RSp461e*RS346pe + 16./3.*m33*m55*L4r*
         RS34561pe + 16./9.*m33*m55*L4r*RS14563pe + 16./9.*m33*m55*L4r*
         RSe45613p );

      M13p6x23L +=  + Ab(m33,mu2) * ( 16./9.*m33*m55*L4r*RSp45613e + 4./
         9.*m33*m55*L4r*RS3456pee + 4./9.*m33*m55*L4r*RS3456ppe - 4./9.
         *m33*m55*L4r*RS345pe + 4./27.*m33*m55*L4r*pow(RS345pe,2) - 4./
         9.*m33*m55*L4r*RS346pe + 4./27.*m33*m55*L4r*pow(RS346pe,2) - 4.
         /9.*m33*m55*L4r*RS356pe + 4./27.*m33*m55*L4r*pow(RS356pe,2) - 
         16./27.*m33*m55*L4r*RS145pe*RS345pe - 16./27.*m33*m55*L4r*
         RS146pe*RS346pe - 16./27.*m33*m55*L4r*RS156pe*RS356pe - 16./27.
         *m33*m55*L4r*RSe561p*RS356pe - 16./27.*m33*m55*L4r*RSe461p*
         RS346pe - 16./27.*m33*m55*L4r*RSe451p*RS345pe - 16./27.*m33*
         m55*L4r*RSp561e*RS356pe - 16./27.*m33*m55*L4r*RSp461e*RS346pe
          - 16./27.*m33*m55*L4r*RSp451e*RS345pe + 16./9.*m33*m55*mp2*
         L7r*RN3p*RSp463e - 16./9.*m33*m55*mp2*L7r*RN3p*RSp461e - 16./9.
         *m33*m55*mp2*L6r*RN3p*RSp45613e + 16./9.*m33*m55*mp2*L6r*RN3p*
         RSp45633e - 16./27.*m33*m55*mp2*L5r*RN3p*RSp461e*RS346pe + 16./
         9.*m33*m55*mp2*L4r*RN3p*RSp45613e - 16./27.*m33*m55*mp2*L4r*
         RN3p*RSp561e*RS356pe );

      M13p6x23L +=  + Ab(m33,mu2) * (  - 16./27.*m33*m55*mp2*L4r*RN3p*
         RSp461e*RS346pe - 16./27.*m33*m55*mp2*L4r*RN3p*RSp451e*RS345pe
          + 16./9.*m33*m55*me2*L7r*RN3e*RSe463p - 16./9.*m33*m55*me2*
         L7r*RN3e*RSe461p - 16./9.*m33*m55*me2*L6r*RN3e*RSe45613p + 16./
         9.*m33*m55*me2*L6r*RN3e*RSe45633p - 16./27.*m33*m55*me2*L5r*
         RN3e*RSe461p*RS346pe + 16./9.*m33*m55*me2*L4r*RN3e*RSe45613p
          - 16./27.*m33*m55*me2*L4r*RN3e*RSe561p*RS356pe - 16./27.*m33*
         m55*me2*L4r*RN3e*RSe461p*RS346pe - 16./27.*m33*m55*me2*L4r*
         RN3e*RSe451p*RS345pe + 16./27.*m33*m55*m66*L7r*RN3p*RSp463e*
         RS345pe + 16./27.*m33*m55*m66*L7r*RN3p*RSp453e*RS346pe + 16./
         27.*m33*m55*m66*L7r*RN3p*RSp461e*RS345pe + 16./27.*m33*m55*m66
         *L7r*RN3p*RSp451e*RS346pe + 16./27.*m33*m55*m66*L7r*RN3e*
         RSe463p*RS345pe + 16./27.*m33*m55*m66*L7r*RN3e*RSe453p*RS346pe
          + 16./27.*m33*m55*m66*L7r*RN3e*RSe461p*RS345pe + 16./27.*m33*
         m55*m66*L7r*RN3e*RSe451p*RS346pe - 16./27.*m33*m55*m66*L7r*
         RN13*RS145pe*RS346pe );

      M13p6x23L +=  + Ab(m33,mu2) * (  - 16./27.*m33*m55*m66*L7r*RN13*
         RS146pe*RS345pe + 16./27.*m33*m55*m66*L6r*RN3p*RSp463e*RS346pe
          + 16./27.*m33*m55*m66*L6r*RN3p*RSp453e*RS345pe + 16./27.*m33*
         m55*m66*L6r*RN3p*RSp461e*RS346pe + 16./27.*m33*m55*m66*L6r*
         RN3p*RSp451e*RS345pe + 16./27.*m33*m55*m66*L6r*RN3e*RSe463p*
         RS346pe + 16./27.*m33*m55*m66*L6r*RN3e*RSe453p*RS345pe + 16./
         27.*m33*m55*m66*L6r*RN3e*RSe461p*RS346pe + 16./27.*m33*m55*m66
         *L6r*RN3e*RSe451p*RS345pe - 16./27.*m33*m55*m66*L6r*RN13*
         RS145pe*RS345pe - 16./27.*m33*m55*m66*L6r*RN13*RS146pe*RS346pe
          + 16./27.*m33*pow(m55,2)*L8r*RN3p*RSp463e*RS346pe + 16./27.*
         m33*pow(m55,2)*L8r*RN3p*RSp461e*RS346pe + 16./27.*m33*pow(
         m55,2)*L8r*RN3e*RSe463p*RS346pe + 16./27.*m33*pow(m55,2)*L8r*
         RN3e*RSe461p*RS346pe - 16./27.*m33*pow(m55,2)*L8r*RN13*RS146pe
         *RS346pe + 16./27.*m33*pow(m55,2)*L7r*RN3p*RSp463e*RS346pe + 
         16./27.*m33*pow(m55,2)*L7r*RN3p*RSp461e*RS346pe + 16./27.*m33*
         pow(m55,2)*L7r*RN3e*RSe463p*RS346pe );

      M13p6x23L +=  + Ab(m33,mu2) * ( 16./27.*m33*pow(m55,2)*L7r*RN3e*
         RSe461p*RS346pe - 16./27.*m33*pow(m55,2)*L7r*RN13*RS146pe*
         RS346pe + 16./27.*m33*pow(m55,2)*L6r*RN3p*RSp463e*RS346pe + 16.
         /27.*m33*pow(m55,2)*L6r*RN3p*RSp461e*RS346pe + 16./27.*m33*
         pow(m55,2)*L6r*RN3e*RSe463p*RS346pe + 16./27.*m33*pow(m55,2)*
         L6r*RN3e*RSe461p*RS346pe - 16./27.*m33*pow(m55,2)*L6r*RN13*
         RS146pe*RS346pe - 16./9.*m33*m44*L7r*RS356pe + 16./9.*m33*m44*
         L7r*RSe563p + 16./9.*m33*m44*L7r*RSp563e - 16./9.*m33*m44*L7r*
         RS156pe - 16./9.*m33*m44*L7r*RSe561p - 16./9.*m33*m44*L7r*
         RSp561e - 32./9.*m33*m44*L6r*RS34561pe - 16./9.*m33*m44*L6r*
         RS14563pe - 16./9.*m33*m44*L6r*RSe45613p - 16./9.*m33*m44*L6r*
         RSp45613e - 8./9.*m33*m44*L6r*RS3456pee - 8./9.*m33*m44*L6r*
         RS3456ppe + 16./9.*m33*m44*L6r*RSe45633p + 16./9.*m33*m44*L6r*
         RSp45633e + 8./9.*m33*m44*L6r*RS345pe + 8./9.*m33*m44*L6r*
         RS346pe + 8./9.*m33*m44*L6r*RS356pe + 4./27.*m33*m44*L5r*pow(
         RS356pe,2) );

      M13p6x23L +=  + Ab(m33,mu2) * (  - 16./27.*m33*m44*L5r*RS156pe*
         RS356pe - 16./27.*m33*m44*L5r*RSe561p*RS356pe - 16./27.*m33*
         m44*L5r*RSp561e*RS356pe + 16./3.*m33*m44*L4r*RS34561pe + 16./9.
         *m33*m44*L4r*RS14563pe + 16./9.*m33*m44*L4r*RSe45613p + 16./9.
         *m33*m44*L4r*RSp45613e + 4./9.*m33*m44*L4r*RS3456pee + 4./9.*
         m33*m44*L4r*RS3456ppe - 4./9.*m33*m44*L4r*RS345pe + 4./27.*m33
         *m44*L4r*pow(RS345pe,2) - 4./9.*m33*m44*L4r*RS346pe + 4./27.*
         m33*m44*L4r*pow(RS346pe,2) - 4./9.*m33*m44*L4r*RS356pe + 4./27.
         *m33*m44*L4r*pow(RS356pe,2) - 16./27.*m33*m44*L4r*RS145pe*
         RS345pe - 16./27.*m33*m44*L4r*RS146pe*RS346pe - 16./27.*m33*
         m44*L4r*RS156pe*RS356pe - 16./27.*m33*m44*L4r*RSe561p*RS356pe
          - 16./27.*m33*m44*L4r*RSe461p*RS346pe - 16./27.*m33*m44*L4r*
         RSe451p*RS345pe - 16./27.*m33*m44*L4r*RSp561e*RS356pe - 16./27.
         *m33*m44*L4r*RSp461e*RS346pe - 16./27.*m33*m44*L4r*RSp451e*
         RS345pe + 16./9.*m33*m44*mp2*L7r*RN3p*RSp563e - 16./9.*m33*m44
         *mp2*L7r*RN3p*RSp561e );

      M13p6x23L +=  + Ab(m33,mu2) * (  - 16./9.*m33*m44*mp2*L6r*RN3p*
         RSp45613e + 16./9.*m33*m44*mp2*L6r*RN3p*RSp45633e - 16./27.*
         m33*m44*mp2*L5r*RN3p*RSp561e*RS356pe + 16./9.*m33*m44*mp2*L4r*
         RN3p*RSp45613e - 16./27.*m33*m44*mp2*L4r*RN3p*RSp561e*RS356pe
          - 16./27.*m33*m44*mp2*L4r*RN3p*RSp461e*RS346pe - 16./27.*m33*
         m44*mp2*L4r*RN3p*RSp451e*RS345pe + 16./9.*m33*m44*me2*L7r*RN3e
         *RSe563p - 16./9.*m33*m44*me2*L7r*RN3e*RSe561p - 16./9.*m33*
         m44*me2*L6r*RN3e*RSe45613p + 16./9.*m33*m44*me2*L6r*RN3e*
         RSe45633p - 16./27.*m33*m44*me2*L5r*RN3e*RSe561p*RS356pe + 16./
         9.*m33*m44*me2*L4r*RN3e*RSe45613p - 16./27.*m33*m44*me2*L4r*
         RN3e*RSe561p*RS356pe - 16./27.*m33*m44*me2*L4r*RN3e*RSe461p*
         RS346pe - 16./27.*m33*m44*me2*L4r*RN3e*RSe451p*RS345pe + 16./
         27.*m33*m44*m66*L7r*RN3p*RSp563e*RS345pe + 16./27.*m33*m44*m66
         *L7r*RN3p*RSp453e*RS356pe + 16./27.*m33*m44*m66*L7r*RN3p*
         RSp561e*RS345pe + 16./27.*m33*m44*m66*L7r*RN3p*RSp451e*RS356pe
          + 16./27.*m33*m44*m66*L7r*RN3e*RSe563p*RS345pe );

      M13p6x23L +=  + Ab(m33,mu2) * ( 16./27.*m33*m44*m66*L7r*RN3e*
         RSe453p*RS356pe + 16./27.*m33*m44*m66*L7r*RN3e*RSe561p*RS345pe
          + 16./27.*m33*m44*m66*L7r*RN3e*RSe451p*RS356pe - 16./27.*m33*
         m44*m66*L7r*RN13*RS145pe*RS356pe - 16./27.*m33*m44*m66*L7r*
         RN13*RS156pe*RS345pe + 16./27.*m33*m44*m66*L6r*RN3p*RSp563e*
         RS356pe + 16./27.*m33*m44*m66*L6r*RN3p*RSp453e*RS345pe + 16./
         27.*m33*m44*m66*L6r*RN3p*RSp561e*RS356pe + 16./27.*m33*m44*m66
         *L6r*RN3p*RSp451e*RS345pe + 16./27.*m33*m44*m66*L6r*RN3e*
         RSe563p*RS356pe + 16./27.*m33*m44*m66*L6r*RN3e*RSe453p*RS345pe
          + 16./27.*m33*m44*m66*L6r*RN3e*RSe561p*RS356pe + 16./27.*m33*
         m44*m66*L6r*RN3e*RSe451p*RS345pe - 16./27.*m33*m44*m66*L6r*
         RN13*RS145pe*RS345pe - 16./27.*m33*m44*m66*L6r*RN13*RS156pe*
         RS356pe + 16./27.*m33*m44*m55*L7r*RN3p*RSp563e*RS346pe + 16./
         27.*m33*m44*m55*L7r*RN3p*RSp463e*RS356pe + 16./27.*m33*m44*m55
         *L7r*RN3p*RSp561e*RS346pe + 16./27.*m33*m44*m55*L7r*RN3p*
         RSp461e*RS356pe );

      M13p6x23L +=  + Ab(m33,mu2) * ( 16./27.*m33*m44*m55*L7r*RN3e*
         RSe563p*RS346pe + 16./27.*m33*m44*m55*L7r*RN3e*RSe463p*RS356pe
          + 16./27.*m33*m44*m55*L7r*RN3e*RSe561p*RS346pe + 16./27.*m33*
         m44*m55*L7r*RN3e*RSe461p*RS356pe - 16./27.*m33*m44*m55*L7r*
         RN13*RS146pe*RS356pe - 16./27.*m33*m44*m55*L7r*RN13*RS156pe*
         RS346pe + 16./27.*m33*m44*m55*L6r*RN3p*RSp563e*RS356pe + 16./
         27.*m33*m44*m55*L6r*RN3p*RSp463e*RS346pe + 16./27.*m33*m44*m55
         *L6r*RN3p*RSp561e*RS356pe + 16./27.*m33*m44*m55*L6r*RN3p*
         RSp461e*RS346pe + 16./27.*m33*m44*m55*L6r*RN3e*RSe563p*RS356pe
          + 16./27.*m33*m44*m55*L6r*RN3e*RSe463p*RS346pe + 16./27.*m33*
         m44*m55*L6r*RN3e*RSe561p*RS356pe + 16./27.*m33*m44*m55*L6r*
         RN3e*RSe461p*RS346pe - 16./27.*m33*m44*m55*L6r*RN13*RS146pe*
         RS346pe - 16./27.*m33*m44*m55*L6r*RN13*RS156pe*RS356pe + 16./
         27.*m33*pow(m44,2)*L8r*RN3p*RSp563e*RS356pe + 16./27.*m33*pow(
         m44,2)*L8r*RN3p*RSp561e*RS356pe + 16./27.*m33*pow(m44,2)*L8r*
         RN3e*RSe563p*RS356pe );

      M13p6x23L +=  + Ab(m33,mu2) * ( 16./27.*m33*pow(m44,2)*L8r*RN3e*
         RSe561p*RS356pe - 16./27.*m33*pow(m44,2)*L8r*RN13*RS156pe*
         RS356pe + 16./27.*m33*pow(m44,2)*L7r*RN3p*RSp563e*RS356pe + 16.
         /27.*m33*pow(m44,2)*L7r*RN3p*RSp561e*RS356pe + 16./27.*m33*
         pow(m44,2)*L7r*RN3e*RSe563p*RS356pe + 16./27.*m33*pow(m44,2)*
         L7r*RN3e*RSe561p*RS356pe - 16./27.*m33*pow(m44,2)*L7r*RN13*
         RS156pe*RS356pe + 16./27.*m33*pow(m44,2)*L6r*RN3p*RSp563e*
         RS356pe + 16./27.*m33*pow(m44,2)*L6r*RN3p*RSp561e*RS356pe + 16.
         /27.*m33*pow(m44,2)*L6r*RN3e*RSe563p*RS356pe + 16./27.*m33*
         pow(m44,2)*L6r*RN3e*RSe561p*RS356pe - 16./27.*m33*pow(m44,2)*
         L6r*RN13*RS156pe*RS356pe - 32./9.*pow(m33,2)*L8r*RS34561pe - 
         16./9.*pow(m33,2)*L8r*RS14563pe - 16./9.*pow(m33,2)*L8r*
         RSe45613p - 16./9.*pow(m33,2)*L8r*RSp45613e + 16./9.*pow(
         m33,2)*L8r*RS3456pee + 16./9.*pow(m33,2)*L8r*RS3456ppe + 16./9.
         *pow(m33,2)*L8r*RSe45633p + 16./9.*pow(m33,2)*L8r*RSp45633e - 
         16./9.*pow(m33,2)*L8r*RS345pe );

      M13p6x23L +=  + Ab(m33,mu2) * (  - 16./9.*pow(m33,2)*L8r*RS346pe
          - 16./9.*pow(m33,2)*L8r*RS356pe + 8*pow(m33,2)*L7r + 28./9.*
         pow(m33,2)*L5r*RS34561pe + 16./9.*pow(m33,2)*L5r*RS14563pe + 
         16./9.*pow(m33,2)*L5r*RSe45613p + 16./9.*pow(m33,2)*L5r*
         RSp45613e - 8./9.*pow(m33,2)*L5r*RS3456pee - 8./9.*pow(m33,2)*
         L5r*RS3456ppe + 8./9.*pow(m33,2)*L5r*RS345pe + 8./9.*pow(
         m33,2)*L5r*RS346pe + 8./9.*pow(m33,2)*L5r*RS356pe - 2./3.*pow(
         m33,2)*L3r*RS34561pe + 5./3.*pow(m33,2)*L3r*RS3456pee + 5./3.*
         pow(m33,2)*L3r*RS3456ppe - 5./3.*pow(m33,2)*L3r*RS345pe - 5./3.
         *pow(m33,2)*L3r*RS346pe - 5./3.*pow(m33,2)*L3r*RS356pe - 8./3.
         *pow(m33,2)*L0r*RS34561pe + 2./3.*pow(m33,2)*L0r*RS3456pee + 2.
         /3.*pow(m33,2)*L0r*RS3456ppe - 2./3.*pow(m33,2)*L0r*RS345pe - 
         2./3.*pow(m33,2)*L0r*RS346pe - 2./3.*pow(m33,2)*L0r*RS356pe - 
         16./9.*pow(m33,2)*mp2*L8r*RN3p*RSp45613e + 16./9.*pow(m33,2)*
         mp2*L8r*RN3p*RSp45633e + 16./9.*pow(m33,2)*mp2*L5r*RN3p*
         RSp45613e );

      M13p6x23L +=  + Ab(m33,mu2) * (  - 16./9.*pow(m33,2)*me2*L8r*RN3e
         *RSe45613p + 16./9.*pow(m33,2)*me2*L8r*RN3e*RSe45633p + 16./9.
         *pow(m33,2)*me2*L5r*RN3e*RSe45613p - 16./9.*pow(m33,2)*m66*L7r
         *RN3p*RSp453e - 16./9.*pow(m33,2)*m66*L7r*RN3p*RSp451e - 16./9.
         *pow(m33,2)*m66*L7r*RN3e*RSe453p - 16./9.*pow(m33,2)*m66*L7r*
         RN3e*RSe451p + 16./9.*pow(m33,2)*m66*L7r*RN13*RS145pe - 16./9.
         *pow(m33,2)*m66*L6r*RN3p*RSp45613e - 16./9.*pow(m33,2)*m66*L6r
         *RN3p*RSp45633e - 16./9.*pow(m33,2)*m66*L6r*RN3e*RSe45613p - 
         16./9.*pow(m33,2)*m66*L6r*RN3e*RSe45633p + 16./9.*pow(m33,2)*
         m66*L6r*RN13*RS14563pe - 16./9.*pow(m33,2)*m55*L7r*RN3p*
         RSp463e - 16./9.*pow(m33,2)*m55*L7r*RN3p*RSp461e - 16./9.*pow(
         m33,2)*m55*L7r*RN3e*RSe463p - 16./9.*pow(m33,2)*m55*L7r*RN3e*
         RSe461p + 16./9.*pow(m33,2)*m55*L7r*RN13*RS146pe - 16./9.*pow(
         m33,2)*m55*L6r*RN3p*RSp45613e - 16./9.*pow(m33,2)*m55*L6r*RN3p
         *RSp45633e - 16./9.*pow(m33,2)*m55*L6r*RN3e*RSe45613p - 16./9.
         *pow(m33,2)*m55*L6r*RN3e*RSe45633p );

      M13p6x23L +=  + Ab(m33,mu2) * ( 16./9.*pow(m33,2)*m55*L6r*RN13*
         RS14563pe - 16./9.*pow(m33,2)*m44*L7r*RN3p*RSp563e - 16./9.*
         pow(m33,2)*m44*L7r*RN3p*RSp561e - 16./9.*pow(m33,2)*m44*L7r*
         RN3e*RSe563p - 16./9.*pow(m33,2)*m44*L7r*RN3e*RSe561p + 16./9.
         *pow(m33,2)*m44*L7r*RN13*RS156pe - 16./9.*pow(m33,2)*m44*L6r*
         RN3p*RSp45613e - 16./9.*pow(m33,2)*m44*L6r*RN3p*RSp45633e - 16.
         /9.*pow(m33,2)*m44*L6r*RN3e*RSe45613p - 16./9.*pow(m33,2)*m44*
         L6r*RN3e*RSe45633p + 16./9.*pow(m33,2)*m44*L6r*RN13*RS14563pe
          - 16./9.*pow(m33,3)*L8r*RN3p*RSp45613e - 16./9.*pow(m33,3)*
         L8r*RN3p*RSp45633e - 16./9.*pow(m33,3)*L8r*RN3e*RSe45613p - 16.
         /9.*pow(m33,3)*L8r*RN3e*RSe45633p + 16./9.*pow(m33,3)*L8r*RN13
         *RS14563pe + 4./9.*m11*L5r*RD3456ep - 5./3.*m11*L3r*RD3456ep
          - 2./3.*m11*L0r*RD3456ep - 32./9.*m11*m66*L7r*RS345pe - 32./9.
         *m11*m66*L6r*RS34561pe - 8./27.*m11*m66*L5r*RS145pe*RS345pe - 
         8./27.*m11*m66*L5r*RSe451p*RS345pe - 8./27.*m11*m66*L5r*
         RSp451e*RS345pe );

      M13p6x23L +=  + Ab(m33,mu2) * ( 8./3.*m11*m66*L4r*RS34561pe + 8./
         9.*m11*m66*L4r*RS14563pe + 8./9.*m11*m66*L4r*RSe45613p + 8./9.
         *m11*m66*L4r*RSp45613e - 8./27.*m11*m66*L4r*RS145pe*RS345pe - 
         8./27.*m11*m66*L4r*RS146pe*RS346pe - 8./27.*m11*m66*L4r*
         RS156pe*RS356pe - 8./27.*m11*m66*L4r*RSe561p*RS356pe - 8./27.*
         m11*m66*L4r*RSe461p*RS346pe - 8./27.*m11*m66*L4r*RSe451p*
         RS345pe - 8./27.*m11*m66*L4r*RSp561e*RS356pe - 8./27.*m11*m66*
         L4r*RSp461e*RS346pe - 8./27.*m11*m66*L4r*RSp451e*RS345pe - 8./
         27.*m11*m66*mp2*L5r*RN3p*RSp451e*RS345pe + 8./9.*m11*m66*mp2*
         L4r*RN3p*RSp45613e - 8./27.*m11*m66*mp2*L4r*RN3p*RSp561e*
         RS356pe - 8./27.*m11*m66*mp2*L4r*RN3p*RSp461e*RS346pe - 8./27.
         *m11*m66*mp2*L4r*RN3p*RSp451e*RS345pe - 8./27.*m11*m66*me2*L5r
         *RN3e*RSe451p*RS345pe + 8./9.*m11*m66*me2*L4r*RN3e*RSe45613p
          - 8./27.*m11*m66*me2*L4r*RN3e*RSe561p*RS356pe - 8./27.*m11*
         m66*me2*L4r*RN3e*RSe461p*RS346pe - 8./27.*m11*m66*me2*L4r*RN3e
         *RSe451p*RS345pe );

      M13p6x23L +=  + Ab(m33,mu2) * ( 16./27.*m11*pow(m66,2)*L8r*RN3p*
         RSp451e*RS345pe + 16./27.*m11*pow(m66,2)*L8r*RN3e*RSe451p*
         RS345pe - 32./27.*m11*pow(m66,2)*L8r*RN13*RS145pe*RS345pe + 16.
         /27.*m11*pow(m66,2)*L7r*RN3p*RSp451e*RS345pe + 16./27.*m11*
         pow(m66,2)*L7r*RN3e*RSe451p*RS345pe - 32./27.*m11*pow(m66,2)*
         L7r*RN13*RS145pe*RS345pe + 16./27.*m11*pow(m66,2)*L6r*RN3p*
         RSp451e*RS345pe + 16./27.*m11*pow(m66,2)*L6r*RN3e*RSe451p*
         RS345pe - 32./27.*m11*pow(m66,2)*L6r*RN13*RS145pe*RS345pe - 32.
         /9.*m11*m55*L7r*RS346pe - 32./9.*m11*m55*L6r*RS34561pe - 8./27.
         *m11*m55*L5r*RS146pe*RS346pe - 8./27.*m11*m55*L5r*RSe461p*
         RS346pe - 8./27.*m11*m55*L5r*RSp461e*RS346pe + 8./3.*m11*m55*
         L4r*RS34561pe + 8./9.*m11*m55*L4r*RS14563pe + 8./9.*m11*m55*
         L4r*RSe45613p + 8./9.*m11*m55*L4r*RSp45613e - 8./27.*m11*m55*
         L4r*RS145pe*RS345pe - 8./27.*m11*m55*L4r*RS146pe*RS346pe - 8./
         27.*m11*m55*L4r*RS156pe*RS356pe - 8./27.*m11*m55*L4r*RSe561p*
         RS356pe );

      M13p6x23L +=  + Ab(m33,mu2) * (  - 8./27.*m11*m55*L4r*RSe461p*
         RS346pe - 8./27.*m11*m55*L4r*RSe451p*RS345pe - 8./27.*m11*m55*
         L4r*RSp561e*RS356pe - 8./27.*m11*m55*L4r*RSp461e*RS346pe - 8./
         27.*m11*m55*L4r*RSp451e*RS345pe - 8./27.*m11*m55*mp2*L5r*RN3p*
         RSp461e*RS346pe + 8./9.*m11*m55*mp2*L4r*RN3p*RSp45613e - 8./27.
         *m11*m55*mp2*L4r*RN3p*RSp561e*RS356pe - 8./27.*m11*m55*mp2*L4r
         *RN3p*RSp461e*RS346pe - 8./27.*m11*m55*mp2*L4r*RN3p*RSp451e*
         RS345pe - 8./27.*m11*m55*me2*L5r*RN3e*RSe461p*RS346pe + 8./9.*
         m11*m55*me2*L4r*RN3e*RSe45613p - 8./27.*m11*m55*me2*L4r*RN3e*
         RSe561p*RS356pe - 8./27.*m11*m55*me2*L4r*RN3e*RSe461p*RS346pe
          - 8./27.*m11*m55*me2*L4r*RN3e*RSe451p*RS345pe + 16./27.*m11*
         m55*m66*L7r*RN3p*RSp461e*RS345pe + 16./27.*m11*m55*m66*L7r*
         RN3p*RSp451e*RS346pe + 16./27.*m11*m55*m66*L7r*RN3e*RSe461p*
         RS345pe + 16./27.*m11*m55*m66*L7r*RN3e*RSe451p*RS346pe - 32./
         27.*m11*m55*m66*L7r*RN13*RS145pe*RS346pe - 32./27.*m11*m55*m66
         *L7r*RN13*RS146pe*RS345pe );

      M13p6x23L +=  + Ab(m33,mu2) * ( 16./27.*m11*m55*m66*L6r*RN3p*
         RSp461e*RS346pe + 16./27.*m11*m55*m66*L6r*RN3p*RSp451e*RS345pe
          + 16./27.*m11*m55*m66*L6r*RN3e*RSe461p*RS346pe + 16./27.*m11*
         m55*m66*L6r*RN3e*RSe451p*RS345pe - 32./27.*m11*m55*m66*L6r*
         RN13*RS145pe*RS345pe - 32./27.*m11*m55*m66*L6r*RN13*RS146pe*
         RS346pe + 16./27.*m11*pow(m55,2)*L8r*RN3p*RSp461e*RS346pe + 16.
         /27.*m11*pow(m55,2)*L8r*RN3e*RSe461p*RS346pe - 32./27.*m11*
         pow(m55,2)*L8r*RN13*RS146pe*RS346pe + 16./27.*m11*pow(m55,2)*
         L7r*RN3p*RSp461e*RS346pe + 16./27.*m11*pow(m55,2)*L7r*RN3e*
         RSe461p*RS346pe - 32./27.*m11*pow(m55,2)*L7r*RN13*RS146pe*
         RS346pe + 16./27.*m11*pow(m55,2)*L6r*RN3p*RSp461e*RS346pe + 16.
         /27.*m11*pow(m55,2)*L6r*RN3e*RSe461p*RS346pe - 32./27.*m11*
         pow(m55,2)*L6r*RN13*RS146pe*RS346pe - 32./9.*m11*m44*L7r*
         RS356pe - 32./9.*m11*m44*L6r*RS34561pe - 8./27.*m11*m44*L5r*
         RS156pe*RS356pe - 8./27.*m11*m44*L5r*RSe561p*RS356pe - 8./27.*
         m11*m44*L5r*RSp561e*RS356pe );

      M13p6x23L +=  + Ab(m33,mu2) * ( 8./3.*m11*m44*L4r*RS34561pe + 8./
         9.*m11*m44*L4r*RS14563pe + 8./9.*m11*m44*L4r*RSe45613p + 8./9.
         *m11*m44*L4r*RSp45613e - 8./27.*m11*m44*L4r*RS145pe*RS345pe - 
         8./27.*m11*m44*L4r*RS146pe*RS346pe - 8./27.*m11*m44*L4r*
         RS156pe*RS356pe - 8./27.*m11*m44*L4r*RSe561p*RS356pe - 8./27.*
         m11*m44*L4r*RSe461p*RS346pe - 8./27.*m11*m44*L4r*RSe451p*
         RS345pe - 8./27.*m11*m44*L4r*RSp561e*RS356pe - 8./27.*m11*m44*
         L4r*RSp461e*RS346pe - 8./27.*m11*m44*L4r*RSp451e*RS345pe - 8./
         27.*m11*m44*mp2*L5r*RN3p*RSp561e*RS356pe + 8./9.*m11*m44*mp2*
         L4r*RN3p*RSp45613e - 8./27.*m11*m44*mp2*L4r*RN3p*RSp561e*
         RS356pe - 8./27.*m11*m44*mp2*L4r*RN3p*RSp461e*RS346pe - 8./27.
         *m11*m44*mp2*L4r*RN3p*RSp451e*RS345pe - 8./27.*m11*m44*me2*L5r
         *RN3e*RSe561p*RS356pe + 8./9.*m11*m44*me2*L4r*RN3e*RSe45613p
          - 8./27.*m11*m44*me2*L4r*RN3e*RSe561p*RS356pe - 8./27.*m11*
         m44*me2*L4r*RN3e*RSe461p*RS346pe - 8./27.*m11*m44*me2*L4r*RN3e
         *RSe451p*RS345pe );

      M13p6x23L +=  + Ab(m33,mu2) * ( 16./27.*m11*m44*m66*L7r*RN3p*
         RSp561e*RS345pe + 16./27.*m11*m44*m66*L7r*RN3p*RSp451e*RS356pe
          + 16./27.*m11*m44*m66*L7r*RN3e*RSe561p*RS345pe + 16./27.*m11*
         m44*m66*L7r*RN3e*RSe451p*RS356pe - 32./27.*m11*m44*m66*L7r*
         RN13*RS145pe*RS356pe - 32./27.*m11*m44*m66*L7r*RN13*RS156pe*
         RS345pe + 16./27.*m11*m44*m66*L6r*RN3p*RSp561e*RS356pe + 16./
         27.*m11*m44*m66*L6r*RN3p*RSp451e*RS345pe + 16./27.*m11*m44*m66
         *L6r*RN3e*RSe561p*RS356pe + 16./27.*m11*m44*m66*L6r*RN3e*
         RSe451p*RS345pe - 32./27.*m11*m44*m66*L6r*RN13*RS145pe*RS345pe
          - 32./27.*m11*m44*m66*L6r*RN13*RS156pe*RS356pe + 16./27.*m11*
         m44*m55*L7r*RN3p*RSp561e*RS346pe + 16./27.*m11*m44*m55*L7r*
         RN3p*RSp461e*RS356pe + 16./27.*m11*m44*m55*L7r*RN3e*RSe561p*
         RS346pe + 16./27.*m11*m44*m55*L7r*RN3e*RSe461p*RS356pe - 32./
         27.*m11*m44*m55*L7r*RN13*RS146pe*RS356pe - 32./27.*m11*m44*m55
         *L7r*RN13*RS156pe*RS346pe + 16./27.*m11*m44*m55*L6r*RN3p*
         RSp561e*RS356pe );

      M13p6x23L +=  + Ab(m33,mu2) * ( 16./27.*m11*m44*m55*L6r*RN3p*
         RSp461e*RS346pe + 16./27.*m11*m44*m55*L6r*RN3e*RSe561p*RS356pe
          + 16./27.*m11*m44*m55*L6r*RN3e*RSe461p*RS346pe - 32./27.*m11*
         m44*m55*L6r*RN13*RS146pe*RS346pe - 32./27.*m11*m44*m55*L6r*
         RN13*RS156pe*RS356pe + 16./27.*m11*pow(m44,2)*L8r*RN3p*RSp561e
         *RS356pe + 16./27.*m11*pow(m44,2)*L8r*RN3e*RSe561p*RS356pe - 
         32./27.*m11*pow(m44,2)*L8r*RN13*RS156pe*RS356pe + 16./27.*m11*
         pow(m44,2)*L7r*RN3p*RSp561e*RS356pe + 16./27.*m11*pow(m44,2)*
         L7r*RN3e*RSe561p*RS356pe - 32./27.*m11*pow(m44,2)*L7r*RN13*
         RS156pe*RS356pe + 16./27.*m11*pow(m44,2)*L6r*RN3p*RSp561e*
         RS356pe + 16./27.*m11*pow(m44,2)*L6r*RN3e*RSe561p*RS356pe - 32.
         /27.*m11*pow(m44,2)*L6r*RN13*RS156pe*RS356pe - 16./3.*m11*m33*
         L8r*RS34561pe + 8./3.*m11*m33*L8r*RS3456pee + 8./3.*m11*m33*
         L8r*RS3456ppe - 8./3.*m11*m33*L8r*RS345pe - 8./3.*m11*m33*L8r*
         RS346pe - 8./3.*m11*m33*L8r*RS356pe + 32./3.*m11*m33*L7r + 32./
         9.*m11*m33*L5r*RS34561pe );

      M13p6x23L +=  + Ab(m33,mu2) * ( 8./9.*m11*m33*L5r*RS14563pe + 8./
         9.*m11*m33*L5r*RSe45613p + 8./9.*m11*m33*L5r*RSp45613e - 4./3.
         *m11*m33*L5r*RS3456pee - 4./3.*m11*m33*L5r*RS3456ppe + 4./3.*
         m11*m33*L5r*RS345pe + 4./3.*m11*m33*L5r*RS346pe + 4./3.*m11*
         m33*L5r*RS356pe - 2./3.*m11*m33*L3r*RS34561pe + 5./3.*m11*m33*
         L3r*RS3456pee + 5./3.*m11*m33*L3r*RS3456ppe - 5./3.*m11*m33*
         L3r*RS345pe - 5./3.*m11*m33*L3r*RS346pe - 5./3.*m11*m33*L3r*
         RS356pe - 8./3.*m11*m33*L0r*RS34561pe + 2./3.*m11*m33*L0r*
         RS3456pee + 2./3.*m11*m33*L0r*RS3456ppe - 2./3.*m11*m33*L0r*
         RS345pe - 2./3.*m11*m33*L0r*RS346pe - 2./3.*m11*m33*L0r*
         RS356pe + 8./9.*m11*m33*mp2*L5r*RN3p*RSp45613e + 8./9.*m11*m33
         *me2*L5r*RN3e*RSe45613p - 16./9.*m11*m33*m66*L7r*RN3p*RSp451e
          - 16./9.*m11*m33*m66*L7r*RN3e*RSe451p + 16./9.*m11*m33*m66*
         L7r*RN13*RS345pe + 32./9.*m11*m33*m66*L7r*RN13*RS145pe - 16./9.
         *m11*m33*m66*L6r*RN3p*RSp45613e - 16./9.*m11*m33*m66*L6r*RN3e*
         RSe45613p );

      M13p6x23L +=  + Ab(m33,mu2) * ( 16./9.*m11*m33*m66*L6r*RN13*
         RS34561pe + 32./9.*m11*m33*m66*L6r*RN13*RS14563pe + 16./27.*
         m11*m33*m66*L5r*RN13*RS145pe*RS345pe - 16./9.*m11*m33*m66*L4r*
         RN13*RS34561pe - 16./9.*m11*m33*m66*L4r*RN13*RS14563pe + 16./
         27.*m11*m33*m66*L4r*RN13*RS145pe*RS345pe + 16./27.*m11*m33*m66
         *L4r*RN13*RS146pe*RS346pe + 16./27.*m11*m33*m66*L4r*RN13*
         RS156pe*RS356pe - 16./9.*m11*m33*m55*L7r*RN3p*RSp461e - 16./9.
         *m11*m33*m55*L7r*RN3e*RSe461p + 16./9.*m11*m33*m55*L7r*RN13*
         RS346pe + 32./9.*m11*m33*m55*L7r*RN13*RS146pe - 16./9.*m11*m33
         *m55*L6r*RN3p*RSp45613e - 16./9.*m11*m33*m55*L6r*RN3e*
         RSe45613p + 16./9.*m11*m33*m55*L6r*RN13*RS34561pe + 32./9.*m11
         *m33*m55*L6r*RN13*RS14563pe + 16./27.*m11*m33*m55*L5r*RN13*
         RS146pe*RS346pe - 16./9.*m11*m33*m55*L4r*RN13*RS34561pe - 16./
         9.*m11*m33*m55*L4r*RN13*RS14563pe + 16./27.*m11*m33*m55*L4r*
         RN13*RS145pe*RS345pe + 16./27.*m11*m33*m55*L4r*RN13*RS146pe*
         RS346pe );

      M13p6x23L +=  + Ab(m33,mu2) * ( 16./27.*m11*m33*m55*L4r*RN13*
         RS156pe*RS356pe - 16./9.*m11*m33*m44*L7r*RN3p*RSp561e - 16./9.
         *m11*m33*m44*L7r*RN3e*RSe561p + 16./9.*m11*m33*m44*L7r*RN13*
         RS356pe + 32./9.*m11*m33*m44*L7r*RN13*RS156pe - 16./9.*m11*m33
         *m44*L6r*RN3p*RSp45613e - 16./9.*m11*m33*m44*L6r*RN3e*
         RSe45613p + 16./9.*m11*m33*m44*L6r*RN13*RS34561pe + 32./9.*m11
         *m33*m44*L6r*RN13*RS14563pe + 16./27.*m11*m33*m44*L5r*RN13*
         RS156pe*RS356pe - 16./9.*m11*m33*m44*L4r*RN13*RS34561pe - 16./
         9.*m11*m33*m44*L4r*RN13*RS14563pe + 16./27.*m11*m33*m44*L4r*
         RN13*RS145pe*RS345pe + 16./27.*m11*m33*m44*L4r*RN13*RS146pe*
         RS346pe + 16./27.*m11*m33*m44*L4r*RN13*RS156pe*RS356pe - 16./9.
         *m11*pow(m33,2)*L8r*RN3p*RSp45613e - 16./9.*m11*pow(m33,2)*L8r
         *RN3e*RSe45613p + 32./9.*m11*pow(m33,2)*L8r*RN13*RS14563pe - 
         16./3.*m11*pow(m33,2)*L7r*RN13 - 16./9.*m11*pow(m33,2)*L5r*
         RN13*RS14563pe - 32./9.*pow(m11,2)*L8r*RS34561pe + 4./3.*pow(
         m11,2)*L5r*RS34561pe );

      M13p6x23L +=  + Ab(m33,mu2) * ( 32./9.*pow(m11,2)*m66*L7r*RN13*
         RS345pe + 32./9.*pow(m11,2)*m66*L6r*RN13*RS34561pe + 8./27.*
         pow(m11,2)*m66*L5r*RN13*RS145pe*RS345pe - 8./9.*pow(m11,2)*m66
         *L4r*RN13*RS34561pe - 8./9.*pow(m11,2)*m66*L4r*RN13*RS14563pe
          + 8./27.*pow(m11,2)*m66*L4r*RN13*RS145pe*RS345pe + 8./27.*
         pow(m11,2)*m66*L4r*RN13*RS146pe*RS346pe + 8./27.*pow(m11,2)*
         m66*L4r*RN13*RS156pe*RS356pe + 32./9.*pow(m11,2)*m55*L7r*RN13*
         RS346pe + 32./9.*pow(m11,2)*m55*L6r*RN13*RS34561pe + 8./27.*
         pow(m11,2)*m55*L5r*RN13*RS146pe*RS346pe - 8./9.*pow(m11,2)*m55
         *L4r*RN13*RS34561pe - 8./9.*pow(m11,2)*m55*L4r*RN13*RS14563pe
          + 8./27.*pow(m11,2)*m55*L4r*RN13*RS145pe*RS345pe + 8./27.*
         pow(m11,2)*m55*L4r*RN13*RS146pe*RS346pe + 8./27.*pow(m11,2)*
         m55*L4r*RN13*RS156pe*RS356pe + 32./9.*pow(m11,2)*m44*L7r*RN13*
         RS356pe + 32./9.*pow(m11,2)*m44*L6r*RN13*RS34561pe + 8./27.*
         pow(m11,2)*m44*L5r*RN13*RS156pe*RS356pe - 8./9.*pow(m11,2)*m44
         *L4r*RN13*RS34561pe );

      M13p6x23L +=  + Ab(m33,mu2) * (  - 8./9.*pow(m11,2)*m44*L4r*RN13*
         RS14563pe + 8./27.*pow(m11,2)*m44*L4r*RN13*RS145pe*RS345pe + 8.
         /27.*pow(m11,2)*m44*L4r*RN13*RS146pe*RS346pe + 8./27.*pow(
         m11,2)*m44*L4r*RN13*RS156pe*RS356pe + 16./9.*pow(m11,2)*m33*
         L8r*RN13*RS34561pe - 32./3.*pow(m11,2)*m33*L7r*RN13 - 16./9.*
         pow(m11,2)*m33*L5r*RN13*RS34561pe - 8./9.*pow(m11,2)*m33*L5r*
         RN13*RS14563pe + 32./9.*pow(m11,3)*L8r*RN13*RS34561pe - 8./9.*
         pow(m11,3)*L5r*RN13*RS34561pe );

      M13p6x23L +=  + Ab(m34,mu2) * ( 4*m33*m44*L8r - 2*m33*m44*L5r + 5.
         /2.*m33*m44*L3r + m33*m44*L0r + 4*pow(m33,2)*L8r - 2*pow(
         m33,2)*L5r + 5./2.*pow(m33,2)*L3r + pow(m33,2)*L0r + 4*m11*m44
         *L8r - 2*m11*m44*L5r + 5./2.*m11*m44*L3r + m11*m44*L0r + 4*m11
         *m33*L8r - 2*m11*m33*L5r + 5./2.*m11*m33*L3r + m11*m33*L0r );

      M13p6x23L +=  + Ab(m35,mu2) * ( 4*m33*m55*L8r - 2*m33*m55*L5r + 5.
         /2.*m33*m55*L3r + m33*m55*L0r + 4*pow(m33,2)*L8r - 2*pow(
         m33,2)*L5r + 5./2.*pow(m33,2)*L3r + pow(m33,2)*L0r + 4*m11*m55
         *L8r - 2*m11*m55*L5r + 5./2.*m11*m55*L3r + m11*m55*L0r + 4*m11
         *m33*L8r - 2*m11*m33*L5r + 5./2.*m11*m33*L3r + m11*m33*L0r );

      M13p6x23L +=  + Ab(m36,mu2) * ( 4*m33*m66*L8r - 2*m33*m66*L5r + 5.
         /2.*m33*m66*L3r + m33*m66*L0r + 4*pow(m33,2)*L8r - 2*pow(
         m33,2)*L5r + 5./2.*pow(m33,2)*L3r + pow(m33,2)*L0r + 4*m11*m66
         *L8r - 2*m11*m66*L5r + 5./2.*m11*m66*L3r + m11*m66*L0r + 4*m11
         *m33*L8r - 2*m11*m33*L5r + 5./2.*m11*m33*L3r + m11*m33*L0r );

      M13p6x23L +=  + Ab(m44,mu2) * ( 8*m33*m44*L6r - 8./3.*m33*m44*L6r
         *RS456pe - 8*m33*m44*L4r + 8./3.*m33*m44*L4r*RS456pe + 2*m33*
         m44*L2r - 2./3.*m33*m44*L2r*RS456pe + 8*m33*m44*L1r - 8./3.*
         m33*m44*L1r*RS456pe + 8*m11*m44*L6r - 8./3.*m11*m44*L6r*
         RS456pe - 8*m11*m44*L4r + 8./3.*m11*m44*L4r*RS456pe + 2*m11*
         m44*L2r - 2./3.*m11*m44*L2r*RS456pe + 8*m11*m44*L1r - 8./3.*
         m11*m44*L1r*RS456pe );

      M13p6x23L +=  + Ab(m45,mu2) * ( 8*m33*m55*L6r - 8*m33*m55*L4r + 2
         *m33*m55*L2r + 8*m33*m55*L1r + 8*m33*m44*L6r - 8*m33*m44*L4r
          + 2*m33*m44*L2r + 8*m33*m44*L1r + 8*m11*m55*L6r - 8*m11*m55*
         L4r + 2*m11*m55*L2r + 8*m11*m55*L1r + 8*m11*m44*L6r - 8*m11*
         m44*L4r + 2*m11*m44*L2r + 8*m11*m44*L1r );

      M13p6x23L +=  + Ab(m46,mu2) * ( 8*m33*m66*L6r - 8*m33*m66*L4r + 2
         *m33*m66*L2r + 8*m33*m66*L1r + 8*m33*m44*L6r - 8*m33*m44*L4r
          + 2*m33*m44*L2r + 8*m33*m44*L1r + 8*m11*m66*L6r - 8*m11*m66*
         L4r + 2*m11*m66*L2r + 8*m11*m66*L1r + 8*m11*m44*L6r - 8*m11*
         m44*L4r + 2*m11*m44*L2r + 8*m11*m44*L1r );

      M13p6x23L +=  + Ab(m55,mu2) * ( 8*m33*m55*L6r - 8./3.*m33*m55*L6r
         *RS546pe - 8*m33*m55*L4r + 8./3.*m33*m55*L4r*RS546pe + 2*m33*
         m55*L2r - 2./3.*m33*m55*L2r*RS546pe + 8*m33*m55*L1r - 8./3.*
         m33*m55*L1r*RS546pe + 8*m11*m55*L6r - 8./3.*m11*m55*L6r*
         RS546pe - 8*m11*m55*L4r + 8./3.*m11*m55*L4r*RS546pe + 2*m11*
         m55*L2r - 2./3.*m11*m55*L2r*RS546pe + 8*m11*m55*L1r - 8./3.*
         m11*m55*L1r*RS546pe );

      M13p6x23L +=  + Ab(m56,mu2) * ( 8*m33*m66*L6r - 8*m33*m66*L4r + 2
         *m33*m66*L2r + 8*m33*m66*L1r + 8*m33*m55*L6r - 8*m33*m55*L4r
          + 2*m33*m55*L2r + 8*m33*m55*L1r + 8*m11*m66*L6r - 8*m11*m66*
         L4r + 2*m11*m66*L2r + 8*m11*m66*L1r + 8*m11*m55*L6r - 8*m11*
         m55*L4r + 2*m11*m55*L2r + 8*m11*m55*L1r );

      M13p6x23L +=  + Ab(m66,mu2) * ( 8*m33*m66*L6r - 8./3.*m33*m66*L6r
         *RS645pe - 8*m33*m66*L4r + 8./3.*m33*m66*L4r*RS645pe + 2*m33*
         m66*L2r - 2./3.*m33*m66*L2r*RS645pe + 8*m33*m66*L1r - 8./3.*
         m33*m66*L1r*RS645pe + 8*m11*m66*L6r - 8./3.*m11*m66*L6r*
         RS645pe - 8*m11*m66*L4r + 8./3.*m11*m66*L4r*RS645pe + 2*m11*
         m66*L2r - 2./3.*m11*m66*L2r*RS645pe + 8*m11*m66*L1r - 8./3.*
         m11*m66*L1r*RS645pe );

      M13p6x23L +=  + Ab(me2,mu2) * ( 8./27.*m66*me2*L5r*RSe453p*
         RS345pe + 8./27.*m66*me2*L5r*pow(RSe453p,2) + 8./27.*m66*me2*
         L5r*RSp453e*RSe453p - 8./27.*m66*me2*L5r*RS145pe*RSe453p - 8./
         27.*m66*me2*L5r*RSe451p*RS345pe - 16./27.*m66*me2*L5r*RSe451p*
         RSe453p - 8./27.*m66*me2*L5r*RSe451p*RSp453e + 8./27.*m66*me2*
         L5r*RSe451p*RS145pe + 8./27.*m66*me2*L5r*pow(RSe451p,2) - 8./
         27.*m66*me2*L5r*RSp451e*RSe453p + 8./27.*m66*me2*L5r*RSp451e*
         RSe451p + 16./9.*m66*me2*L4r*RSe45613p - 8./9.*m66*me2*L4r*
         RSe45633p + 8./27.*m66*me2*L4r*RSe563p*RS356pe + 8./27.*m66*
         me2*L4r*pow(RSe563p,2) + 8./27.*m66*me2*L4r*RSe463p*RS346pe + 
         8./27.*m66*me2*L4r*pow(RSe463p,2) + 8./27.*m66*me2*L4r*RSe453p
         *RS345pe + 8./27.*m66*me2*L4r*pow(RSe453p,2) + 8./27.*m66*me2*
         L4r*RSp563e*RSe563p + 8./27.*m66*me2*L4r*RSp463e*RSe463p + 8./
         27.*m66*me2*L4r*RSp453e*RSe453p - 8./9.*m66*me2*L4r*RSe45611p
          - 8./27.*m66*me2*L4r*RS145pe*RSe453p - 8./27.*m66*me2*L4r*
         RS146pe*RSe463p );

      M13p6x23L +=  + Ab(me2,mu2) * (  - 8./27.*m66*me2*L4r*RS156pe*
         RSe563p - 8./27.*m66*me2*L4r*RSe561p*RS356pe - 16./27.*m66*me2
         *L4r*RSe561p*RSe563p - 8./27.*m66*me2*L4r*RSe561p*RSp563e + 8./
         27.*m66*me2*L4r*RSe561p*RS156pe + 8./27.*m66*me2*L4r*pow(
         RSe561p,2) - 8./27.*m66*me2*L4r*RSe461p*RS346pe - 16./27.*m66*
         me2*L4r*RSe461p*RSe463p - 8./27.*m66*me2*L4r*RSe461p*RSp463e
          + 8./27.*m66*me2*L4r*RSe461p*RS146pe + 8./27.*m66*me2*L4r*
         pow(RSe461p,2) - 8./27.*m66*me2*L4r*RSe451p*RS345pe - 16./27.*
         m66*me2*L4r*RSe451p*RSe453p - 8./27.*m66*me2*L4r*RSe451p*
         RSp453e + 8./27.*m66*me2*L4r*RSe451p*RS145pe + 8./27.*m66*me2*
         L4r*pow(RSe451p,2) - 8./27.*m66*me2*L4r*RSp561e*RSe563p + 8./
         27.*m66*me2*L4r*RSp561e*RSe561p - 8./27.*m66*me2*L4r*RSp461e*
         RSe463p + 8./27.*m66*me2*L4r*RSp461e*RSe461p - 8./27.*m66*me2*
         L4r*RSp451e*RSe453p + 8./27.*m66*me2*L4r*RSp451e*RSe451p + 8./
         27.*m66*me2*mp2*L5r*RNep*RSp453e*RSe453p - 8./27.*m66*me2*mp2*
         L5r*RNep*RSe451p*RSp453e );

      M13p6x23L +=  + Ab(me2,mu2) * (  - 8./27.*m66*me2*mp2*L5r*RNep*
         RSp451e*RSe453p + 8./27.*m66*me2*mp2*L5r*RNep*RSp451e*RSe451p
          + 8./27.*m66*me2*mp2*L4r*RNep*RSp563e*RSe563p + 8./27.*m66*
         me2*mp2*L4r*RNep*RSp463e*RSe463p + 8./27.*m66*me2*mp2*L4r*RNep
         *RSp453e*RSe453p - 8./27.*m66*me2*mp2*L4r*RNep*RSe561p*RSp563e
          - 8./27.*m66*me2*mp2*L4r*RNep*RSe461p*RSp463e - 8./27.*m66*
         me2*mp2*L4r*RNep*RSe451p*RSp453e - 8./27.*m66*me2*mp2*L4r*RNep
         *RSp561e*RSe563p + 8./27.*m66*me2*mp2*L4r*RNep*RSp561e*RSe561p
          - 8./27.*m66*me2*mp2*L4r*RNep*RSp461e*RSe463p + 8./27.*m66*
         me2*mp2*L4r*RNep*RSp461e*RSe461p - 8./27.*m66*me2*mp2*L4r*RNep
         *RSp451e*RSe453p + 8./27.*m66*me2*mp2*L4r*RNep*RSp451e*RSe451p
          - 8./27.*pow(m66,2)*L8r*pow(RSe453p,2) - 16./27.*pow(m66,2)*
         L8r*RSp453e*RSe453p + 16./27.*pow(m66,2)*L8r*RSe451p*RSe453p
          + 16./27.*pow(m66,2)*L8r*RSe451p*RSp453e - 8./27.*pow(m66,2)*
         L8r*pow(RSe451p,2) + 16./27.*pow(m66,2)*L8r*RSp451e*RSe453p - 
         16./27.*pow(m66,2)*L8r*RSp451e*RSe451p );

      M13p6x23L +=  + Ab(me2,mu2) * (  - 8./27.*pow(m66,2)*L7r*pow(
         RSe453p,2) - 16./27.*pow(m66,2)*L7r*RSp453e*RSe453p + 16./27.*
         pow(m66,2)*L7r*RSe451p*RSe453p + 16./27.*pow(m66,2)*L7r*
         RSe451p*RSp453e - 8./27.*pow(m66,2)*L7r*pow(RSe451p,2) + 16./
         27.*pow(m66,2)*L7r*RSp451e*RSe453p - 16./27.*pow(m66,2)*L7r*
         RSp451e*RSe451p - 8./27.*pow(m66,2)*L6r*pow(RSe453p,2) - 16./
         27.*pow(m66,2)*L6r*RSp453e*RSe453p + 16./27.*pow(m66,2)*L6r*
         RSe451p*RSe453p + 16./27.*pow(m66,2)*L6r*RSe451p*RSp453e - 8./
         27.*pow(m66,2)*L6r*pow(RSe451p,2) + 16./27.*pow(m66,2)*L6r*
         RSp451e*RSe453p - 16./27.*pow(m66,2)*L6r*RSp451e*RSe451p - 16./
         27.*pow(m66,2)*mp2*L8r*RNep*RSp453e*RSe453p + 16./27.*pow(
         m66,2)*mp2*L8r*RNep*RSe451p*RSp453e + 16./27.*pow(m66,2)*mp2*
         L8r*RNep*RSp451e*RSe453p - 16./27.*pow(m66,2)*mp2*L8r*RNep*
         RSp451e*RSe451p - 16./27.*pow(m66,2)*mp2*L7r*RNep*RSp453e*
         RSe453p + 16./27.*pow(m66,2)*mp2*L7r*RNep*RSe451p*RSp453e + 16.
         /27.*pow(m66,2)*mp2*L7r*RNep*RSp451e*RSe453p );

      M13p6x23L +=  + Ab(me2,mu2) * (  - 16./27.*pow(m66,2)*mp2*L7r*
         RNep*RSp451e*RSe451p - 16./27.*pow(m66,2)*mp2*L6r*RNep*RSp453e
         *RSe453p + 16./27.*pow(m66,2)*mp2*L6r*RNep*RSe451p*RSp453e + 
         16./27.*pow(m66,2)*mp2*L6r*RNep*RSp451e*RSe453p - 16./27.*pow(
         m66,2)*mp2*L6r*RNep*RSp451e*RSe451p + 16./27.*pow(m66,2)*me2*
         L8r*RN3e*RSe453p*RS345pe - 16./27.*pow(m66,2)*me2*L8r*RN3e*
         RSe451p*RS345pe - 16./27.*pow(m66,2)*me2*L8r*RN1e*RS145pe*
         RSe453p + 16./27.*pow(m66,2)*me2*L8r*RN1e*RSe451p*RS145pe + 16.
         /27.*pow(m66,2)*me2*L7r*RN3e*RSe453p*RS345pe - 16./27.*pow(
         m66,2)*me2*L7r*RN3e*RSe451p*RS345pe - 16./27.*pow(m66,2)*me2*
         L7r*RN1e*RS145pe*RSe453p + 16./27.*pow(m66,2)*me2*L7r*RN1e*
         RSe451p*RS145pe + 16./27.*pow(m66,2)*me2*L6r*RN3e*RSe453p*
         RS345pe - 16./27.*pow(m66,2)*me2*L6r*RN3e*RSe451p*RS345pe - 16.
         /27.*pow(m66,2)*me2*L6r*RN1e*RS145pe*RSe453p + 16./27.*pow(
         m66,2)*me2*L6r*RN1e*RSe451p*RS145pe + 8./27.*m55*me2*L5r*
         RSe463p*RS346pe );

      M13p6x23L +=  + Ab(me2,mu2) * ( 8./27.*m55*me2*L5r*pow(RSe463p,2)
          + 8./27.*m55*me2*L5r*RSp463e*RSe463p - 8./27.*m55*me2*L5r*
         RS146pe*RSe463p - 8./27.*m55*me2*L5r*RSe461p*RS346pe - 16./27.
         *m55*me2*L5r*RSe461p*RSe463p - 8./27.*m55*me2*L5r*RSe461p*
         RSp463e + 8./27.*m55*me2*L5r*RSe461p*RS146pe + 8./27.*m55*me2*
         L5r*pow(RSe461p,2) - 8./27.*m55*me2*L5r*RSp461e*RSe463p + 8./
         27.*m55*me2*L5r*RSp461e*RSe461p + 16./9.*m55*me2*L4r*RSe45613p
          - 8./9.*m55*me2*L4r*RSe45633p + 8./27.*m55*me2*L4r*RSe563p*
         RS356pe + 8./27.*m55*me2*L4r*pow(RSe563p,2) + 8./27.*m55*me2*
         L4r*RSe463p*RS346pe + 8./27.*m55*me2*L4r*pow(RSe463p,2) + 8./
         27.*m55*me2*L4r*RSe453p*RS345pe + 8./27.*m55*me2*L4r*pow(
         RSe453p,2) + 8./27.*m55*me2*L4r*RSp563e*RSe563p + 8./27.*m55*
         me2*L4r*RSp463e*RSe463p + 8./27.*m55*me2*L4r*RSp453e*RSe453p
          - 8./9.*m55*me2*L4r*RSe45611p - 8./27.*m55*me2*L4r*RS145pe*
         RSe453p - 8./27.*m55*me2*L4r*RS146pe*RSe463p - 8./27.*m55*me2*
         L4r*RS156pe*RSe563p );

      M13p6x23L +=  + Ab(me2,mu2) * (  - 8./27.*m55*me2*L4r*RSe561p*
         RS356pe - 16./27.*m55*me2*L4r*RSe561p*RSe563p - 8./27.*m55*me2
         *L4r*RSe561p*RSp563e + 8./27.*m55*me2*L4r*RSe561p*RS156pe + 8./
         27.*m55*me2*L4r*pow(RSe561p,2) - 8./27.*m55*me2*L4r*RSe461p*
         RS346pe - 16./27.*m55*me2*L4r*RSe461p*RSe463p - 8./27.*m55*me2
         *L4r*RSe461p*RSp463e + 8./27.*m55*me2*L4r*RSe461p*RS146pe + 8./
         27.*m55*me2*L4r*pow(RSe461p,2) - 8./27.*m55*me2*L4r*RSe451p*
         RS345pe - 16./27.*m55*me2*L4r*RSe451p*RSe453p - 8./27.*m55*me2
         *L4r*RSe451p*RSp453e + 8./27.*m55*me2*L4r*RSe451p*RS145pe + 8./
         27.*m55*me2*L4r*pow(RSe451p,2) - 8./27.*m55*me2*L4r*RSp561e*
         RSe563p + 8./27.*m55*me2*L4r*RSp561e*RSe561p - 8./27.*m55*me2*
         L4r*RSp461e*RSe463p + 8./27.*m55*me2*L4r*RSp461e*RSe461p - 8./
         27.*m55*me2*L4r*RSp451e*RSe453p + 8./27.*m55*me2*L4r*RSp451e*
         RSe451p + 8./27.*m55*me2*mp2*L5r*RNep*RSp463e*RSe463p - 8./27.
         *m55*me2*mp2*L5r*RNep*RSe461p*RSp463e - 8./27.*m55*me2*mp2*L5r
         *RNep*RSp461e*RSe463p );

      M13p6x23L +=  + Ab(me2,mu2) * ( 8./27.*m55*me2*mp2*L5r*RNep*
         RSp461e*RSe461p + 8./27.*m55*me2*mp2*L4r*RNep*RSp563e*RSe563p
          + 8./27.*m55*me2*mp2*L4r*RNep*RSp463e*RSe463p + 8./27.*m55*
         me2*mp2*L4r*RNep*RSp453e*RSe453p - 8./27.*m55*me2*mp2*L4r*RNep
         *RSe561p*RSp563e - 8./27.*m55*me2*mp2*L4r*RNep*RSe461p*RSp463e
          - 8./27.*m55*me2*mp2*L4r*RNep*RSe451p*RSp453e - 8./27.*m55*
         me2*mp2*L4r*RNep*RSp561e*RSe563p + 8./27.*m55*me2*mp2*L4r*RNep
         *RSp561e*RSe561p - 8./27.*m55*me2*mp2*L4r*RNep*RSp461e*RSe463p
          + 8./27.*m55*me2*mp2*L4r*RNep*RSp461e*RSe461p - 8./27.*m55*
         me2*mp2*L4r*RNep*RSp451e*RSe453p + 8./27.*m55*me2*mp2*L4r*RNep
         *RSp451e*RSe451p - 16./27.*m55*m66*L7r*RSe453p*RSe463p - 16./
         27.*m55*m66*L7r*RSp463e*RSe453p - 16./27.*m55*m66*L7r*RSp453e*
         RSe463p + 16./27.*m55*m66*L7r*RSe461p*RSe453p + 16./27.*m55*
         m66*L7r*RSe461p*RSp453e + 16./27.*m55*m66*L7r*RSe451p*RSe463p
          + 16./27.*m55*m66*L7r*RSe451p*RSp463e - 16./27.*m55*m66*L7r*
         RSe451p*RSe461p );

      M13p6x23L +=  + Ab(me2,mu2) * ( 16./27.*m55*m66*L7r*RSp461e*
         RSe453p - 16./27.*m55*m66*L7r*RSp461e*RSe451p + 16./27.*m55*
         m66*L7r*RSp451e*RSe463p - 16./27.*m55*m66*L7r*RSp451e*RSe461p
          - 8./27.*m55*m66*L6r*pow(RSe463p,2) - 8./27.*m55*m66*L6r*pow(
         RSe453p,2) - 16./27.*m55*m66*L6r*RSp463e*RSe463p - 16./27.*m55
         *m66*L6r*RSp453e*RSe453p + 16./27.*m55*m66*L6r*RSe461p*RSe463p
          + 16./27.*m55*m66*L6r*RSe461p*RSp463e - 8./27.*m55*m66*L6r*
         pow(RSe461p,2) + 16./27.*m55*m66*L6r*RSe451p*RSe453p + 16./27.
         *m55*m66*L6r*RSe451p*RSp453e - 8./27.*m55*m66*L6r*pow(
         RSe451p,2) + 16./27.*m55*m66*L6r*RSp461e*RSe463p - 16./27.*m55
         *m66*L6r*RSp461e*RSe461p + 16./27.*m55*m66*L6r*RSp451e*RSe453p
          - 16./27.*m55*m66*L6r*RSp451e*RSe451p - 16./27.*m55*m66*mp2*
         L7r*RNep*RSp463e*RSe453p - 16./27.*m55*m66*mp2*L7r*RNep*
         RSp453e*RSe463p + 16./27.*m55*m66*mp2*L7r*RNep*RSe461p*RSp453e
          + 16./27.*m55*m66*mp2*L7r*RNep*RSe451p*RSp463e + 16./27.*m55*
         m66*mp2*L7r*RNep*RSp461e*RSe453p );

      M13p6x23L +=  + Ab(me2,mu2) * (  - 16./27.*m55*m66*mp2*L7r*RNep*
         RSp461e*RSe451p + 16./27.*m55*m66*mp2*L7r*RNep*RSp451e*RSe463p
          - 16./27.*m55*m66*mp2*L7r*RNep*RSp451e*RSe461p - 16./27.*m55*
         m66*mp2*L6r*RNep*RSp463e*RSe463p - 16./27.*m55*m66*mp2*L6r*
         RNep*RSp453e*RSe453p + 16./27.*m55*m66*mp2*L6r*RNep*RSe461p*
         RSp463e + 16./27.*m55*m66*mp2*L6r*RNep*RSe451p*RSp453e + 16./
         27.*m55*m66*mp2*L6r*RNep*RSp461e*RSe463p - 16./27.*m55*m66*mp2
         *L6r*RNep*RSp461e*RSe461p + 16./27.*m55*m66*mp2*L6r*RNep*
         RSp451e*RSe453p - 16./27.*m55*m66*mp2*L6r*RNep*RSp451e*RSe451p
          + 16./27.*m55*m66*me2*L7r*RN3e*RSe463p*RS345pe + 16./27.*m55*
         m66*me2*L7r*RN3e*RSe453p*RS346pe - 16./27.*m55*m66*me2*L7r*
         RN3e*RSe461p*RS345pe - 16./27.*m55*m66*me2*L7r*RN3e*RSe451p*
         RS346pe - 16./27.*m55*m66*me2*L7r*RN1e*RS145pe*RSe463p - 16./
         27.*m55*m66*me2*L7r*RN1e*RS146pe*RSe453p + 16./27.*m55*m66*me2
         *L7r*RN1e*RSe461p*RS145pe + 16./27.*m55*m66*me2*L7r*RN1e*
         RSe451p*RS146pe );

      M13p6x23L +=  + Ab(me2,mu2) * ( 16./27.*m55*m66*me2*L6r*RN3e*
         RSe463p*RS346pe + 16./27.*m55*m66*me2*L6r*RN3e*RSe453p*RS345pe
          - 16./27.*m55*m66*me2*L6r*RN3e*RSe461p*RS346pe - 16./27.*m55*
         m66*me2*L6r*RN3e*RSe451p*RS345pe - 16./27.*m55*m66*me2*L6r*
         RN1e*RS145pe*RSe453p - 16./27.*m55*m66*me2*L6r*RN1e*RS146pe*
         RSe463p + 16./27.*m55*m66*me2*L6r*RN1e*RSe461p*RS146pe + 16./
         27.*m55*m66*me2*L6r*RN1e*RSe451p*RS145pe - 8./27.*pow(m55,2)*
         L8r*pow(RSe463p,2) - 16./27.*pow(m55,2)*L8r*RSp463e*RSe463p + 
         16./27.*pow(m55,2)*L8r*RSe461p*RSe463p + 16./27.*pow(m55,2)*
         L8r*RSe461p*RSp463e - 8./27.*pow(m55,2)*L8r*pow(RSe461p,2) + 
         16./27.*pow(m55,2)*L8r*RSp461e*RSe463p - 16./27.*pow(m55,2)*
         L8r*RSp461e*RSe461p - 8./27.*pow(m55,2)*L7r*pow(RSe463p,2) - 
         16./27.*pow(m55,2)*L7r*RSp463e*RSe463p + 16./27.*pow(m55,2)*
         L7r*RSe461p*RSe463p + 16./27.*pow(m55,2)*L7r*RSe461p*RSp463e
          - 8./27.*pow(m55,2)*L7r*pow(RSe461p,2) + 16./27.*pow(m55,2)*
         L7r*RSp461e*RSe463p );

      M13p6x23L +=  + Ab(me2,mu2) * (  - 16./27.*pow(m55,2)*L7r*RSp461e
         *RSe461p - 8./27.*pow(m55,2)*L6r*pow(RSe463p,2) - 16./27.*pow(
         m55,2)*L6r*RSp463e*RSe463p + 16./27.*pow(m55,2)*L6r*RSe461p*
         RSe463p + 16./27.*pow(m55,2)*L6r*RSe461p*RSp463e - 8./27.*pow(
         m55,2)*L6r*pow(RSe461p,2) + 16./27.*pow(m55,2)*L6r*RSp461e*
         RSe463p - 16./27.*pow(m55,2)*L6r*RSp461e*RSe461p - 16./27.*
         pow(m55,2)*mp2*L8r*RNep*RSp463e*RSe463p + 16./27.*pow(m55,2)*
         mp2*L8r*RNep*RSe461p*RSp463e + 16./27.*pow(m55,2)*mp2*L8r*RNep
         *RSp461e*RSe463p - 16./27.*pow(m55,2)*mp2*L8r*RNep*RSp461e*
         RSe461p - 16./27.*pow(m55,2)*mp2*L7r*RNep*RSp463e*RSe463p + 16.
         /27.*pow(m55,2)*mp2*L7r*RNep*RSe461p*RSp463e + 16./27.*pow(
         m55,2)*mp2*L7r*RNep*RSp461e*RSe463p - 16./27.*pow(m55,2)*mp2*
         L7r*RNep*RSp461e*RSe461p - 16./27.*pow(m55,2)*mp2*L6r*RNep*
         RSp463e*RSe463p + 16./27.*pow(m55,2)*mp2*L6r*RNep*RSe461p*
         RSp463e + 16./27.*pow(m55,2)*mp2*L6r*RNep*RSp461e*RSe463p - 16.
         /27.*pow(m55,2)*mp2*L6r*RNep*RSp461e*RSe461p );

      M13p6x23L +=  + Ab(me2,mu2) * ( 16./27.*pow(m55,2)*me2*L8r*RN3e*
         RSe463p*RS346pe - 16./27.*pow(m55,2)*me2*L8r*RN3e*RSe461p*
         RS346pe - 16./27.*pow(m55,2)*me2*L8r*RN1e*RS146pe*RSe463p + 16.
         /27.*pow(m55,2)*me2*L8r*RN1e*RSe461p*RS146pe + 16./27.*pow(
         m55,2)*me2*L7r*RN3e*RSe463p*RS346pe - 16./27.*pow(m55,2)*me2*
         L7r*RN3e*RSe461p*RS346pe - 16./27.*pow(m55,2)*me2*L7r*RN1e*
         RS146pe*RSe463p + 16./27.*pow(m55,2)*me2*L7r*RN1e*RSe461p*
         RS146pe + 16./27.*pow(m55,2)*me2*L6r*RN3e*RSe463p*RS346pe - 16.
         /27.*pow(m55,2)*me2*L6r*RN3e*RSe461p*RS346pe - 16./27.*pow(
         m55,2)*me2*L6r*RN1e*RS146pe*RSe463p + 16./27.*pow(m55,2)*me2*
         L6r*RN1e*RSe461p*RS146pe + 8./27.*m44*me2*L5r*RSe563p*RS356pe
          + 8./27.*m44*me2*L5r*pow(RSe563p,2) + 8./27.*m44*me2*L5r*
         RSp563e*RSe563p - 8./27.*m44*me2*L5r*RS156pe*RSe563p - 8./27.*
         m44*me2*L5r*RSe561p*RS356pe - 16./27.*m44*me2*L5r*RSe561p*
         RSe563p - 8./27.*m44*me2*L5r*RSe561p*RSp563e + 8./27.*m44*me2*
         L5r*RSe561p*RS156pe );

      M13p6x23L +=  + Ab(me2,mu2) * ( 8./27.*m44*me2*L5r*pow(RSe561p,2)
          - 8./27.*m44*me2*L5r*RSp561e*RSe563p + 8./27.*m44*me2*L5r*
         RSp561e*RSe561p + 16./9.*m44*me2*L4r*RSe45613p - 8./9.*m44*me2
         *L4r*RSe45633p + 8./27.*m44*me2*L4r*RSe563p*RS356pe + 8./27.*
         m44*me2*L4r*pow(RSe563p,2) + 8./27.*m44*me2*L4r*RSe463p*
         RS346pe + 8./27.*m44*me2*L4r*pow(RSe463p,2) + 8./27.*m44*me2*
         L4r*RSe453p*RS345pe + 8./27.*m44*me2*L4r*pow(RSe453p,2) + 8./
         27.*m44*me2*L4r*RSp563e*RSe563p + 8./27.*m44*me2*L4r*RSp463e*
         RSe463p + 8./27.*m44*me2*L4r*RSp453e*RSe453p - 8./9.*m44*me2*
         L4r*RSe45611p - 8./27.*m44*me2*L4r*RS145pe*RSe453p - 8./27.*
         m44*me2*L4r*RS146pe*RSe463p - 8./27.*m44*me2*L4r*RS156pe*
         RSe563p - 8./27.*m44*me2*L4r*RSe561p*RS356pe - 16./27.*m44*me2
         *L4r*RSe561p*RSe563p - 8./27.*m44*me2*L4r*RSe561p*RSp563e + 8./
         27.*m44*me2*L4r*RSe561p*RS156pe + 8./27.*m44*me2*L4r*pow(
         RSe561p,2) - 8./27.*m44*me2*L4r*RSe461p*RS346pe - 16./27.*m44*
         me2*L4r*RSe461p*RSe463p );

      M13p6x23L +=  + Ab(me2,mu2) * (  - 8./27.*m44*me2*L4r*RSe461p*
         RSp463e + 8./27.*m44*me2*L4r*RSe461p*RS146pe + 8./27.*m44*me2*
         L4r*pow(RSe461p,2) - 8./27.*m44*me2*L4r*RSe451p*RS345pe - 16./
         27.*m44*me2*L4r*RSe451p*RSe453p - 8./27.*m44*me2*L4r*RSe451p*
         RSp453e + 8./27.*m44*me2*L4r*RSe451p*RS145pe + 8./27.*m44*me2*
         L4r*pow(RSe451p,2) - 8./27.*m44*me2*L4r*RSp561e*RSe563p + 8./
         27.*m44*me2*L4r*RSp561e*RSe561p - 8./27.*m44*me2*L4r*RSp461e*
         RSe463p + 8./27.*m44*me2*L4r*RSp461e*RSe461p - 8./27.*m44*me2*
         L4r*RSp451e*RSe453p + 8./27.*m44*me2*L4r*RSp451e*RSe451p + 8./
         27.*m44*me2*mp2*L5r*RNep*RSp563e*RSe563p - 8./27.*m44*me2*mp2*
         L5r*RNep*RSe561p*RSp563e - 8./27.*m44*me2*mp2*L5r*RNep*RSp561e
         *RSe563p + 8./27.*m44*me2*mp2*L5r*RNep*RSp561e*RSe561p + 8./27.
         *m44*me2*mp2*L4r*RNep*RSp563e*RSe563p + 8./27.*m44*me2*mp2*L4r
         *RNep*RSp463e*RSe463p + 8./27.*m44*me2*mp2*L4r*RNep*RSp453e*
         RSe453p - 8./27.*m44*me2*mp2*L4r*RNep*RSe561p*RSp563e - 8./27.
         *m44*me2*mp2*L4r*RNep*RSe461p*RSp463e );

      M13p6x23L +=  + Ab(me2,mu2) * (  - 8./27.*m44*me2*mp2*L4r*RNep*
         RSe451p*RSp453e - 8./27.*m44*me2*mp2*L4r*RNep*RSp561e*RSe563p
          + 8./27.*m44*me2*mp2*L4r*RNep*RSp561e*RSe561p - 8./27.*m44*
         me2*mp2*L4r*RNep*RSp461e*RSe463p + 8./27.*m44*me2*mp2*L4r*RNep
         *RSp461e*RSe461p - 8./27.*m44*me2*mp2*L4r*RNep*RSp451e*RSe453p
          + 8./27.*m44*me2*mp2*L4r*RNep*RSp451e*RSe451p - 16./27.*m44*
         m66*L7r*RSe453p*RSe563p - 16./27.*m44*m66*L7r*RSp563e*RSe453p
          - 16./27.*m44*m66*L7r*RSp453e*RSe563p + 16./27.*m44*m66*L7r*
         RSe561p*RSe453p + 16./27.*m44*m66*L7r*RSe561p*RSp453e + 16./27.
         *m44*m66*L7r*RSe451p*RSe563p + 16./27.*m44*m66*L7r*RSe451p*
         RSp563e - 16./27.*m44*m66*L7r*RSe451p*RSe561p + 16./27.*m44*
         m66*L7r*RSp561e*RSe453p - 16./27.*m44*m66*L7r*RSp561e*RSe451p
          + 16./27.*m44*m66*L7r*RSp451e*RSe563p - 16./27.*m44*m66*L7r*
         RSp451e*RSe561p - 8./27.*m44*m66*L6r*pow(RSe563p,2) - 8./27.*
         m44*m66*L6r*pow(RSe453p,2) - 16./27.*m44*m66*L6r*RSp563e*
         RSe563p );

      M13p6x23L +=  + Ab(me2,mu2) * (  - 16./27.*m44*m66*L6r*RSp453e*
         RSe453p + 16./27.*m44*m66*L6r*RSe561p*RSe563p + 16./27.*m44*
         m66*L6r*RSe561p*RSp563e - 8./27.*m44*m66*L6r*pow(RSe561p,2) + 
         16./27.*m44*m66*L6r*RSe451p*RSe453p + 16./27.*m44*m66*L6r*
         RSe451p*RSp453e - 8./27.*m44*m66*L6r*pow(RSe451p,2) + 16./27.*
         m44*m66*L6r*RSp561e*RSe563p - 16./27.*m44*m66*L6r*RSp561e*
         RSe561p + 16./27.*m44*m66*L6r*RSp451e*RSe453p - 16./27.*m44*
         m66*L6r*RSp451e*RSe451p - 16./27.*m44*m66*mp2*L7r*RNep*RSp563e
         *RSe453p - 16./27.*m44*m66*mp2*L7r*RNep*RSp453e*RSe563p + 16./
         27.*m44*m66*mp2*L7r*RNep*RSe561p*RSp453e + 16./27.*m44*m66*mp2
         *L7r*RNep*RSe451p*RSp563e + 16./27.*m44*m66*mp2*L7r*RNep*
         RSp561e*RSe453p - 16./27.*m44*m66*mp2*L7r*RNep*RSp561e*RSe451p
          + 16./27.*m44*m66*mp2*L7r*RNep*RSp451e*RSe563p - 16./27.*m44*
         m66*mp2*L7r*RNep*RSp451e*RSe561p - 16./27.*m44*m66*mp2*L6r*
         RNep*RSp563e*RSe563p - 16./27.*m44*m66*mp2*L6r*RNep*RSp453e*
         RSe453p );

      M13p6x23L +=  + Ab(me2,mu2) * ( 16./27.*m44*m66*mp2*L6r*RNep*
         RSe561p*RSp563e + 16./27.*m44*m66*mp2*L6r*RNep*RSe451p*RSp453e
          + 16./27.*m44*m66*mp2*L6r*RNep*RSp561e*RSe563p - 16./27.*m44*
         m66*mp2*L6r*RNep*RSp561e*RSe561p + 16./27.*m44*m66*mp2*L6r*
         RNep*RSp451e*RSe453p - 16./27.*m44*m66*mp2*L6r*RNep*RSp451e*
         RSe451p + 16./27.*m44*m66*me2*L7r*RN3e*RSe563p*RS345pe + 16./
         27.*m44*m66*me2*L7r*RN3e*RSe453p*RS356pe - 16./27.*m44*m66*me2
         *L7r*RN3e*RSe561p*RS345pe - 16./27.*m44*m66*me2*L7r*RN3e*
         RSe451p*RS356pe - 16./27.*m44*m66*me2*L7r*RN1e*RS145pe*RSe563p
          - 16./27.*m44*m66*me2*L7r*RN1e*RS156pe*RSe453p + 16./27.*m44*
         m66*me2*L7r*RN1e*RSe561p*RS145pe + 16./27.*m44*m66*me2*L7r*
         RN1e*RSe451p*RS156pe + 16./27.*m44*m66*me2*L6r*RN3e*RSe563p*
         RS356pe + 16./27.*m44*m66*me2*L6r*RN3e*RSe453p*RS345pe - 16./
         27.*m44*m66*me2*L6r*RN3e*RSe561p*RS356pe - 16./27.*m44*m66*me2
         *L6r*RN3e*RSe451p*RS345pe - 16./27.*m44*m66*me2*L6r*RN1e*
         RS145pe*RSe453p );

      M13p6x23L +=  + Ab(me2,mu2) * (  - 16./27.*m44*m66*me2*L6r*RN1e*
         RS156pe*RSe563p + 16./27.*m44*m66*me2*L6r*RN1e*RSe561p*RS156pe
          + 16./27.*m44*m66*me2*L6r*RN1e*RSe451p*RS145pe - 16./27.*m44*
         m55*L7r*RSe463p*RSe563p - 16./27.*m44*m55*L7r*RSp563e*RSe463p
          - 16./27.*m44*m55*L7r*RSp463e*RSe563p + 16./27.*m44*m55*L7r*
         RSe561p*RSe463p + 16./27.*m44*m55*L7r*RSe561p*RSp463e + 16./27.
         *m44*m55*L7r*RSe461p*RSe563p + 16./27.*m44*m55*L7r*RSe461p*
         RSp563e - 16./27.*m44*m55*L7r*RSe461p*RSe561p + 16./27.*m44*
         m55*L7r*RSp561e*RSe463p - 16./27.*m44*m55*L7r*RSp561e*RSe461p
          + 16./27.*m44*m55*L7r*RSp461e*RSe563p - 16./27.*m44*m55*L7r*
         RSp461e*RSe561p - 8./27.*m44*m55*L6r*pow(RSe563p,2) - 8./27.*
         m44*m55*L6r*pow(RSe463p,2) - 16./27.*m44*m55*L6r*RSp563e*
         RSe563p - 16./27.*m44*m55*L6r*RSp463e*RSe463p + 16./27.*m44*
         m55*L6r*RSe561p*RSe563p + 16./27.*m44*m55*L6r*RSe561p*RSp563e
          - 8./27.*m44*m55*L6r*pow(RSe561p,2) + 16./27.*m44*m55*L6r*
         RSe461p*RSe463p );

      M13p6x23L +=  + Ab(me2,mu2) * ( 16./27.*m44*m55*L6r*RSe461p*
         RSp463e - 8./27.*m44*m55*L6r*pow(RSe461p,2) + 16./27.*m44*m55*
         L6r*RSp561e*RSe563p - 16./27.*m44*m55*L6r*RSp561e*RSe561p + 16.
         /27.*m44*m55*L6r*RSp461e*RSe463p - 16./27.*m44*m55*L6r*RSp461e
         *RSe461p - 16./27.*m44*m55*mp2*L7r*RNep*RSp563e*RSe463p - 16./
         27.*m44*m55*mp2*L7r*RNep*RSp463e*RSe563p + 16./27.*m44*m55*mp2
         *L7r*RNep*RSe561p*RSp463e + 16./27.*m44*m55*mp2*L7r*RNep*
         RSe461p*RSp563e + 16./27.*m44*m55*mp2*L7r*RNep*RSp561e*RSe463p
          - 16./27.*m44*m55*mp2*L7r*RNep*RSp561e*RSe461p + 16./27.*m44*
         m55*mp2*L7r*RNep*RSp461e*RSe563p - 16./27.*m44*m55*mp2*L7r*
         RNep*RSp461e*RSe561p - 16./27.*m44*m55*mp2*L6r*RNep*RSp563e*
         RSe563p - 16./27.*m44*m55*mp2*L6r*RNep*RSp463e*RSe463p + 16./
         27.*m44*m55*mp2*L6r*RNep*RSe561p*RSp563e + 16./27.*m44*m55*mp2
         *L6r*RNep*RSe461p*RSp463e + 16./27.*m44*m55*mp2*L6r*RNep*
         RSp561e*RSe563p - 16./27.*m44*m55*mp2*L6r*RNep*RSp561e*RSe561p
          + 16./27.*m44*m55*mp2*L6r*RNep*RSp461e*RSe463p );

      M13p6x23L +=  + Ab(me2,mu2) * (  - 16./27.*m44*m55*mp2*L6r*RNep*
         RSp461e*RSe461p + 16./27.*m44*m55*me2*L7r*RN3e*RSe563p*RS346pe
          + 16./27.*m44*m55*me2*L7r*RN3e*RSe463p*RS356pe - 16./27.*m44*
         m55*me2*L7r*RN3e*RSe561p*RS346pe - 16./27.*m44*m55*me2*L7r*
         RN3e*RSe461p*RS356pe - 16./27.*m44*m55*me2*L7r*RN1e*RS146pe*
         RSe563p - 16./27.*m44*m55*me2*L7r*RN1e*RS156pe*RSe463p + 16./
         27.*m44*m55*me2*L7r*RN1e*RSe561p*RS146pe + 16./27.*m44*m55*me2
         *L7r*RN1e*RSe461p*RS156pe + 16./27.*m44*m55*me2*L6r*RN3e*
         RSe563p*RS356pe + 16./27.*m44*m55*me2*L6r*RN3e*RSe463p*RS346pe
          - 16./27.*m44*m55*me2*L6r*RN3e*RSe561p*RS356pe - 16./27.*m44*
         m55*me2*L6r*RN3e*RSe461p*RS346pe - 16./27.*m44*m55*me2*L6r*
         RN1e*RS146pe*RSe463p - 16./27.*m44*m55*me2*L6r*RN1e*RS156pe*
         RSe563p + 16./27.*m44*m55*me2*L6r*RN1e*RSe561p*RS156pe + 16./
         27.*m44*m55*me2*L6r*RN1e*RSe461p*RS146pe - 8./27.*pow(m44,2)*
         L8r*pow(RSe563p,2) - 16./27.*pow(m44,2)*L8r*RSp563e*RSe563p + 
         16./27.*pow(m44,2)*L8r*RSe561p*RSe563p );

      M13p6x23L +=  + Ab(me2,mu2) * ( 16./27.*pow(m44,2)*L8r*RSe561p*
         RSp563e - 8./27.*pow(m44,2)*L8r*pow(RSe561p,2) + 16./27.*pow(
         m44,2)*L8r*RSp561e*RSe563p - 16./27.*pow(m44,2)*L8r*RSp561e*
         RSe561p - 8./27.*pow(m44,2)*L7r*pow(RSe563p,2) - 16./27.*pow(
         m44,2)*L7r*RSp563e*RSe563p + 16./27.*pow(m44,2)*L7r*RSe561p*
         RSe563p + 16./27.*pow(m44,2)*L7r*RSe561p*RSp563e - 8./27.*pow(
         m44,2)*L7r*pow(RSe561p,2) + 16./27.*pow(m44,2)*L7r*RSp561e*
         RSe563p - 16./27.*pow(m44,2)*L7r*RSp561e*RSe561p - 8./27.*pow(
         m44,2)*L6r*pow(RSe563p,2) - 16./27.*pow(m44,2)*L6r*RSp563e*
         RSe563p + 16./27.*pow(m44,2)*L6r*RSe561p*RSe563p + 16./27.*
         pow(m44,2)*L6r*RSe561p*RSp563e - 8./27.*pow(m44,2)*L6r*pow(
         RSe561p,2) + 16./27.*pow(m44,2)*L6r*RSp561e*RSe563p - 16./27.*
         pow(m44,2)*L6r*RSp561e*RSe561p - 16./27.*pow(m44,2)*mp2*L8r*
         RNep*RSp563e*RSe563p + 16./27.*pow(m44,2)*mp2*L8r*RNep*RSe561p
         *RSp563e + 16./27.*pow(m44,2)*mp2*L8r*RNep*RSp561e*RSe563p - 
         16./27.*pow(m44,2)*mp2*L8r*RNep*RSp561e*RSe561p );

      M13p6x23L +=  + Ab(me2,mu2) * (  - 16./27.*pow(m44,2)*mp2*L7r*
         RNep*RSp563e*RSe563p + 16./27.*pow(m44,2)*mp2*L7r*RNep*RSe561p
         *RSp563e + 16./27.*pow(m44,2)*mp2*L7r*RNep*RSp561e*RSe563p - 
         16./27.*pow(m44,2)*mp2*L7r*RNep*RSp561e*RSe561p - 16./27.*pow(
         m44,2)*mp2*L6r*RNep*RSp563e*RSe563p + 16./27.*pow(m44,2)*mp2*
         L6r*RNep*RSe561p*RSp563e + 16./27.*pow(m44,2)*mp2*L6r*RNep*
         RSp561e*RSe563p - 16./27.*pow(m44,2)*mp2*L6r*RNep*RSp561e*
         RSe561p + 16./27.*pow(m44,2)*me2*L8r*RN3e*RSe563p*RS356pe - 16.
         /27.*pow(m44,2)*me2*L8r*RN3e*RSe561p*RS356pe - 16./27.*pow(
         m44,2)*me2*L8r*RN1e*RS156pe*RSe563p + 16./27.*pow(m44,2)*me2*
         L8r*RN1e*RSe561p*RS156pe + 16./27.*pow(m44,2)*me2*L7r*RN3e*
         RSe563p*RS356pe - 16./27.*pow(m44,2)*me2*L7r*RN3e*RSe561p*
         RS356pe - 16./27.*pow(m44,2)*me2*L7r*RN1e*RS156pe*RSe563p + 16.
         /27.*pow(m44,2)*me2*L7r*RN1e*RSe561p*RS156pe + 16./27.*pow(
         m44,2)*me2*L6r*RN3e*RSe563p*RS356pe - 16./27.*pow(m44,2)*me2*
         L6r*RN3e*RSe561p*RS356pe );

      M13p6x23L +=  + Ab(me2,mu2) * (  - 16./27.*pow(m44,2)*me2*L6r*
         RN1e*RS156pe*RSe563p + 16./27.*pow(m44,2)*me2*L6r*RN1e*RSe561p
         *RS156pe + 8./9.*m33*me2*L5r*RSe45613p + 4./9.*m33*me2*L5r*
         RSe45611p + 4./3.*m33*me2*L4r*RSe564p + 4./3.*m33*me2*L4r*
         RSe465p + 4./3.*m33*me2*L4r*RSe456p - 2./3.*m33*me2*L3r*
         RSe45613p - 5./3.*m33*me2*L3r*RSe45633p - 5./3.*m33*me2*L3r*
         RSe45611p - 2./3.*m33*me2*L2r*RSe564p - 2./3.*m33*me2*L2r*
         RSe465p - 2./3.*m33*me2*L2r*RSe456p - 8./3.*m33*me2*L1r*
         RSe564p - 8./3.*m33*me2*L1r*RSe465p - 8./3.*m33*me2*L1r*
         RSe456p - 8./3.*m33*me2*L0r*RSe45613p - 2./3.*m33*me2*L0r*
         RSe45633p - 2./3.*m33*me2*L0r*RSe45611p - 32./9.*m33*m66*L7r*
         RSe453p - 16./9.*m33*m66*L7r*RSe451p - 16./9.*m33*m66*L6r*
         RSe45613p - 8./9.*m33*m66*L6r*RSe45633p - 8./3.*m33*m66*L6r*
         RSe456p - 4./27.*m33*m66*L5r*pow(RSe453p,2) - 8./27.*m33*m66*
         L5r*RSp453e*RSe453p - 8./27.*m33*m66*L5r*RSe451p*RSe453p - 8./
         27.*m33*m66*L5r*RSe451p*RSp453e );

      M13p6x23L +=  + Ab(me2,mu2) * (  - 8./27.*m33*m66*L5r*RSp451e*
         RSe453p + 8./9.*m33*m66*L4r*RSe45613p + 4./9.*m33*m66*L4r*
         RSe45633p - 4./27.*m33*m66*L4r*pow(RSe563p,2) - 4./27.*m33*m66
         *L4r*pow(RSe463p,2) - 4./27.*m33*m66*L4r*pow(RSe453p,2) - 8./
         27.*m33*m66*L4r*RSp563e*RSe563p - 8./27.*m33*m66*L4r*RSp463e*
         RSe463p - 8./27.*m33*m66*L4r*RSp453e*RSe453p - 8./27.*m33*m66*
         L4r*RSe561p*RSe563p - 8./27.*m33*m66*L4r*RSe561p*RSp563e - 8./
         27.*m33*m66*L4r*RSe461p*RSe463p - 8./27.*m33*m66*L4r*RSe461p*
         RSp463e - 8./27.*m33*m66*L4r*RSe451p*RSe453p - 8./27.*m33*m66*
         L4r*RSe451p*RSp453e - 8./27.*m33*m66*L4r*RSp561e*RSe563p - 8./
         27.*m33*m66*L4r*RSp461e*RSe463p - 8./27.*m33*m66*L4r*RSp451e*
         RSe453p + 4./3.*m33*m66*L4r*RSe456p - 8./27.*m33*m66*mp2*L5r*
         RNep*RSp453e*RSe453p - 8./27.*m33*m66*mp2*L5r*RNep*RSe451p*
         RSp453e - 8./27.*m33*m66*mp2*L5r*RNep*RSp451e*RSe453p - 8./27.
         *m33*m66*mp2*L4r*RNep*RSp563e*RSe563p - 8./27.*m33*m66*mp2*L4r
         *RNep*RSp463e*RSe463p );

      M13p6x23L +=  + Ab(me2,mu2) * (  - 8./27.*m33*m66*mp2*L4r*RNep*
         RSp453e*RSe453p - 8./27.*m33*m66*mp2*L4r*RNep*RSe561p*RSp563e
          - 8./27.*m33*m66*mp2*L4r*RNep*RSe461p*RSp463e - 8./27.*m33*
         m66*mp2*L4r*RNep*RSe451p*RSp453e - 8./27.*m33*m66*mp2*L4r*RNep
         *RSp561e*RSe563p - 8./27.*m33*m66*mp2*L4r*RNep*RSp461e*RSe463p
          - 8./27.*m33*m66*mp2*L4r*RNep*RSp451e*RSe453p - 16./9.*m33*
         m66*me2*L7r*RN3e*RSe453p + 16./9.*m33*m66*me2*L7r*RN3e*RSe451p
          + 16./9.*m33*m66*me2*L6r*RN3e*RSe45613p - 16./9.*m33*m66*me2*
         L6r*RN3e*RSe45633p + 16./27.*m33*m66*me2*L5r*RN3e*RSe451p*
         RS345pe + 8./27.*m33*m66*me2*L5r*RN1e*RS145pe*RSe453p - 16./9.
         *m33*m66*me2*L4r*RN3e*RSe45613p + 16./27.*m33*m66*me2*L4r*RN3e
         *RSe561p*RS356pe + 16./27.*m33*m66*me2*L4r*RN3e*RSe461p*
         RS346pe + 16./27.*m33*m66*me2*L4r*RN3e*RSe451p*RS345pe - 8./9.
         *m33*m66*me2*L4r*RN1e*RSe45613p + 8./27.*m33*m66*me2*L4r*RN1e*
         RS145pe*RSe453p + 8./27.*m33*m66*me2*L4r*RN1e*RS146pe*RSe463p
          + 8./27.*m33*m66*me2*L4r*RN1e*RS156pe*RSe563p );

      M13p6x23L +=  + Ab(me2,mu2) * ( 16./27.*m33*pow(m66,2)*L8r*RNep*
         RSp453e*RSe453p + 16./27.*m33*pow(m66,2)*L8r*RNep*RSe451p*
         RSp453e + 16./27.*m33*pow(m66,2)*L8r*RNep*RSp451e*RSe453p - 16.
         /27.*m33*pow(m66,2)*L8r*RN3e*RSe453p*RS345pe - 16./27.*m33*
         pow(m66,2)*L8r*RN3e*RSe451p*RS345pe - 16./27.*m33*pow(m66,2)*
         L8r*RN1e*RS145pe*RSe453p + 16./27.*m33*pow(m66,2)*L7r*RNep*
         RSp453e*RSe453p + 16./27.*m33*pow(m66,2)*L7r*RNep*RSe451p*
         RSp453e + 16./27.*m33*pow(m66,2)*L7r*RNep*RSp451e*RSe453p - 16.
         /27.*m33*pow(m66,2)*L7r*RN3e*RSe453p*RS345pe - 16./27.*m33*
         pow(m66,2)*L7r*RN3e*RSe451p*RS345pe - 16./27.*m33*pow(m66,2)*
         L7r*RN1e*RS145pe*RSe453p + 16./27.*m33*pow(m66,2)*L6r*RNep*
         RSp453e*RSe453p + 16./27.*m33*pow(m66,2)*L6r*RNep*RSe451p*
         RSp453e + 16./27.*m33*pow(m66,2)*L6r*RNep*RSp451e*RSe453p - 16.
         /27.*m33*pow(m66,2)*L6r*RN3e*RSe453p*RS345pe - 16./27.*m33*
         pow(m66,2)*L6r*RN3e*RSe451p*RS345pe - 16./27.*m33*pow(m66,2)*
         L6r*RN1e*RS145pe*RSe453p );

      M13p6x23L +=  + Ab(me2,mu2) * (  - 32./9.*m33*m55*L7r*RSe463p - 
         16./9.*m33*m55*L7r*RSe461p - 16./9.*m33*m55*L6r*RSe45613p - 8./
         9.*m33*m55*L6r*RSe45633p - 8./3.*m33*m55*L6r*RSe465p - 4./27.*
         m33*m55*L5r*pow(RSe463p,2) - 8./27.*m33*m55*L5r*RSp463e*
         RSe463p - 8./27.*m33*m55*L5r*RSe461p*RSe463p - 8./27.*m33*m55*
         L5r*RSe461p*RSp463e - 8./27.*m33*m55*L5r*RSp461e*RSe463p + 8./
         9.*m33*m55*L4r*RSe45613p + 4./9.*m33*m55*L4r*RSe45633p - 4./27.
         *m33*m55*L4r*pow(RSe563p,2) - 4./27.*m33*m55*L4r*pow(
         RSe463p,2) - 4./27.*m33*m55*L4r*pow(RSe453p,2) - 8./27.*m33*
         m55*L4r*RSp563e*RSe563p - 8./27.*m33*m55*L4r*RSp463e*RSe463p
          - 8./27.*m33*m55*L4r*RSp453e*RSe453p - 8./27.*m33*m55*L4r*
         RSe561p*RSe563p - 8./27.*m33*m55*L4r*RSe561p*RSp563e - 8./27.*
         m33*m55*L4r*RSe461p*RSe463p - 8./27.*m33*m55*L4r*RSe461p*
         RSp463e - 8./27.*m33*m55*L4r*RSe451p*RSe453p - 8./27.*m33*m55*
         L4r*RSe451p*RSp453e - 8./27.*m33*m55*L4r*RSp561e*RSe563p - 8./
         27.*m33*m55*L4r*RSp461e*RSe463p );

      M13p6x23L +=  + Ab(me2,mu2) * (  - 8./27.*m33*m55*L4r*RSp451e*
         RSe453p + 4./3.*m33*m55*L4r*RSe465p - 8./27.*m33*m55*mp2*L5r*
         RNep*RSp463e*RSe463p - 8./27.*m33*m55*mp2*L5r*RNep*RSe461p*
         RSp463e - 8./27.*m33*m55*mp2*L5r*RNep*RSp461e*RSe463p - 8./27.
         *m33*m55*mp2*L4r*RNep*RSp563e*RSe563p - 8./27.*m33*m55*mp2*L4r
         *RNep*RSp463e*RSe463p - 8./27.*m33*m55*mp2*L4r*RNep*RSp453e*
         RSe453p - 8./27.*m33*m55*mp2*L4r*RNep*RSe561p*RSp563e - 8./27.
         *m33*m55*mp2*L4r*RNep*RSe461p*RSp463e - 8./27.*m33*m55*mp2*L4r
         *RNep*RSe451p*RSp453e - 8./27.*m33*m55*mp2*L4r*RNep*RSp561e*
         RSe563p - 8./27.*m33*m55*mp2*L4r*RNep*RSp461e*RSe463p - 8./27.
         *m33*m55*mp2*L4r*RNep*RSp451e*RSe453p - 16./9.*m33*m55*me2*L7r
         *RN3e*RSe463p + 16./9.*m33*m55*me2*L7r*RN3e*RSe461p + 16./9.*
         m33*m55*me2*L6r*RN3e*RSe45613p - 16./9.*m33*m55*me2*L6r*RN3e*
         RSe45633p + 16./27.*m33*m55*me2*L5r*RN3e*RSe461p*RS346pe + 8./
         27.*m33*m55*me2*L5r*RN1e*RS146pe*RSe463p - 16./9.*m33*m55*me2*
         L4r*RN3e*RSe45613p );

      M13p6x23L +=  + Ab(me2,mu2) * ( 16./27.*m33*m55*me2*L4r*RN3e*
         RSe561p*RS356pe + 16./27.*m33*m55*me2*L4r*RN3e*RSe461p*RS346pe
          + 16./27.*m33*m55*me2*L4r*RN3e*RSe451p*RS345pe - 8./9.*m33*
         m55*me2*L4r*RN1e*RSe45613p + 8./27.*m33*m55*me2*L4r*RN1e*
         RS145pe*RSe453p + 8./27.*m33*m55*me2*L4r*RN1e*RS146pe*RSe463p
          + 8./27.*m33*m55*me2*L4r*RN1e*RS156pe*RSe563p + 16./27.*m33*
         m55*m66*L7r*RNep*RSp463e*RSe453p + 16./27.*m33*m55*m66*L7r*
         RNep*RSp453e*RSe463p + 16./27.*m33*m55*m66*L7r*RNep*RSe461p*
         RSp453e + 16./27.*m33*m55*m66*L7r*RNep*RSe451p*RSp463e + 16./
         27.*m33*m55*m66*L7r*RNep*RSp461e*RSe453p + 16./27.*m33*m55*m66
         *L7r*RNep*RSp451e*RSe463p - 16./27.*m33*m55*m66*L7r*RN3e*
         RSe463p*RS345pe - 16./27.*m33*m55*m66*L7r*RN3e*RSe453p*RS346pe
          - 16./27.*m33*m55*m66*L7r*RN3e*RSe461p*RS345pe - 16./27.*m33*
         m55*m66*L7r*RN3e*RSe451p*RS346pe - 16./27.*m33*m55*m66*L7r*
         RN1e*RS145pe*RSe463p - 16./27.*m33*m55*m66*L7r*RN1e*RS146pe*
         RSe453p );

      M13p6x23L +=  + Ab(me2,mu2) * ( 16./27.*m33*m55*m66*L6r*RNep*
         RSp463e*RSe463p + 16./27.*m33*m55*m66*L6r*RNep*RSp453e*RSe453p
          + 16./27.*m33*m55*m66*L6r*RNep*RSe461p*RSp463e + 16./27.*m33*
         m55*m66*L6r*RNep*RSe451p*RSp453e + 16./27.*m33*m55*m66*L6r*
         RNep*RSp461e*RSe463p + 16./27.*m33*m55*m66*L6r*RNep*RSp451e*
         RSe453p - 16./27.*m33*m55*m66*L6r*RN3e*RSe463p*RS346pe - 16./
         27.*m33*m55*m66*L6r*RN3e*RSe453p*RS345pe - 16./27.*m33*m55*m66
         *L6r*RN3e*RSe461p*RS346pe - 16./27.*m33*m55*m66*L6r*RN3e*
         RSe451p*RS345pe - 16./27.*m33*m55*m66*L6r*RN1e*RS145pe*RSe453p
          - 16./27.*m33*m55*m66*L6r*RN1e*RS146pe*RSe463p + 16./27.*m33*
         pow(m55,2)*L8r*RNep*RSp463e*RSe463p + 16./27.*m33*pow(m55,2)*
         L8r*RNep*RSe461p*RSp463e + 16./27.*m33*pow(m55,2)*L8r*RNep*
         RSp461e*RSe463p - 16./27.*m33*pow(m55,2)*L8r*RN3e*RSe463p*
         RS346pe - 16./27.*m33*pow(m55,2)*L8r*RN3e*RSe461p*RS346pe - 16.
         /27.*m33*pow(m55,2)*L8r*RN1e*RS146pe*RSe463p + 16./27.*m33*
         pow(m55,2)*L7r*RNep*RSp463e*RSe463p );

      M13p6x23L +=  + Ab(me2,mu2) * ( 16./27.*m33*pow(m55,2)*L7r*RNep*
         RSe461p*RSp463e + 16./27.*m33*pow(m55,2)*L7r*RNep*RSp461e*
         RSe463p - 16./27.*m33*pow(m55,2)*L7r*RN3e*RSe463p*RS346pe - 16.
         /27.*m33*pow(m55,2)*L7r*RN3e*RSe461p*RS346pe - 16./27.*m33*
         pow(m55,2)*L7r*RN1e*RS146pe*RSe463p + 16./27.*m33*pow(m55,2)*
         L6r*RNep*RSp463e*RSe463p + 16./27.*m33*pow(m55,2)*L6r*RNep*
         RSe461p*RSp463e + 16./27.*m33*pow(m55,2)*L6r*RNep*RSp461e*
         RSe463p - 16./27.*m33*pow(m55,2)*L6r*RN3e*RSe463p*RS346pe - 16.
         /27.*m33*pow(m55,2)*L6r*RN3e*RSe461p*RS346pe - 16./27.*m33*
         pow(m55,2)*L6r*RN1e*RS146pe*RSe463p - 32./9.*m33*m44*L7r*
         RSe563p - 16./9.*m33*m44*L7r*RSe561p - 16./9.*m33*m44*L6r*
         RSe45613p - 8./9.*m33*m44*L6r*RSe45633p - 8./3.*m33*m44*L6r*
         RSe564p - 4./27.*m33*m44*L5r*pow(RSe563p,2) - 8./27.*m33*m44*
         L5r*RSp563e*RSe563p - 8./27.*m33*m44*L5r*RSe561p*RSe563p - 8./
         27.*m33*m44*L5r*RSe561p*RSp563e - 8./27.*m33*m44*L5r*RSp561e*
         RSe563p );

      M13p6x23L +=  + Ab(me2,mu2) * ( 8./9.*m33*m44*L4r*RSe45613p + 4./
         9.*m33*m44*L4r*RSe45633p - 4./27.*m33*m44*L4r*pow(RSe563p,2)
          - 4./27.*m33*m44*L4r*pow(RSe463p,2) - 4./27.*m33*m44*L4r*pow(
         RSe453p,2) - 8./27.*m33*m44*L4r*RSp563e*RSe563p - 8./27.*m33*
         m44*L4r*RSp463e*RSe463p - 8./27.*m33*m44*L4r*RSp453e*RSe453p
          - 8./27.*m33*m44*L4r*RSe561p*RSe563p - 8./27.*m33*m44*L4r*
         RSe561p*RSp563e - 8./27.*m33*m44*L4r*RSe461p*RSe463p - 8./27.*
         m33*m44*L4r*RSe461p*RSp463e - 8./27.*m33*m44*L4r*RSe451p*
         RSe453p - 8./27.*m33*m44*L4r*RSe451p*RSp453e - 8./27.*m33*m44*
         L4r*RSp561e*RSe563p - 8./27.*m33*m44*L4r*RSp461e*RSe463p - 8./
         27.*m33*m44*L4r*RSp451e*RSe453p + 4./3.*m33*m44*L4r*RSe564p - 
         8./27.*m33*m44*mp2*L5r*RNep*RSp563e*RSe563p - 8./27.*m33*m44*
         mp2*L5r*RNep*RSe561p*RSp563e - 8./27.*m33*m44*mp2*L5r*RNep*
         RSp561e*RSe563p - 8./27.*m33*m44*mp2*L4r*RNep*RSp563e*RSe563p
          - 8./27.*m33*m44*mp2*L4r*RNep*RSp463e*RSe463p - 8./27.*m33*
         m44*mp2*L4r*RNep*RSp453e*RSe453p );

      M13p6x23L +=  + Ab(me2,mu2) * (  - 8./27.*m33*m44*mp2*L4r*RNep*
         RSe561p*RSp563e - 8./27.*m33*m44*mp2*L4r*RNep*RSe461p*RSp463e
          - 8./27.*m33*m44*mp2*L4r*RNep*RSe451p*RSp453e - 8./27.*m33*
         m44*mp2*L4r*RNep*RSp561e*RSe563p - 8./27.*m33*m44*mp2*L4r*RNep
         *RSp461e*RSe463p - 8./27.*m33*m44*mp2*L4r*RNep*RSp451e*RSe453p
          - 16./9.*m33*m44*me2*L7r*RN3e*RSe563p + 16./9.*m33*m44*me2*
         L7r*RN3e*RSe561p + 16./9.*m33*m44*me2*L6r*RN3e*RSe45613p - 16./
         9.*m33*m44*me2*L6r*RN3e*RSe45633p + 16./27.*m33*m44*me2*L5r*
         RN3e*RSe561p*RS356pe + 8./27.*m33*m44*me2*L5r*RN1e*RS156pe*
         RSe563p - 16./9.*m33*m44*me2*L4r*RN3e*RSe45613p + 16./27.*m33*
         m44*me2*L4r*RN3e*RSe561p*RS356pe + 16./27.*m33*m44*me2*L4r*
         RN3e*RSe461p*RS346pe + 16./27.*m33*m44*me2*L4r*RN3e*RSe451p*
         RS345pe - 8./9.*m33*m44*me2*L4r*RN1e*RSe45613p + 8./27.*m33*
         m44*me2*L4r*RN1e*RS145pe*RSe453p + 8./27.*m33*m44*me2*L4r*RN1e
         *RS146pe*RSe463p + 8./27.*m33*m44*me2*L4r*RN1e*RS156pe*RSe563p
          + 16./27.*m33*m44*m66*L7r*RNep*RSp563e*RSe453p );

      M13p6x23L +=  + Ab(me2,mu2) * ( 16./27.*m33*m44*m66*L7r*RNep*
         RSp453e*RSe563p + 16./27.*m33*m44*m66*L7r*RNep*RSe561p*RSp453e
          + 16./27.*m33*m44*m66*L7r*RNep*RSe451p*RSp563e + 16./27.*m33*
         m44*m66*L7r*RNep*RSp561e*RSe453p + 16./27.*m33*m44*m66*L7r*
         RNep*RSp451e*RSe563p - 16./27.*m33*m44*m66*L7r*RN3e*RSe563p*
         RS345pe - 16./27.*m33*m44*m66*L7r*RN3e*RSe453p*RS356pe - 16./
         27.*m33*m44*m66*L7r*RN3e*RSe561p*RS345pe - 16./27.*m33*m44*m66
         *L7r*RN3e*RSe451p*RS356pe - 16./27.*m33*m44*m66*L7r*RN1e*
         RS145pe*RSe563p - 16./27.*m33*m44*m66*L7r*RN1e*RS156pe*RSe453p
          + 16./27.*m33*m44*m66*L6r*RNep*RSp563e*RSe563p + 16./27.*m33*
         m44*m66*L6r*RNep*RSp453e*RSe453p + 16./27.*m33*m44*m66*L6r*
         RNep*RSe561p*RSp563e + 16./27.*m33*m44*m66*L6r*RNep*RSe451p*
         RSp453e + 16./27.*m33*m44*m66*L6r*RNep*RSp561e*RSe563p + 16./
         27.*m33*m44*m66*L6r*RNep*RSp451e*RSe453p - 16./27.*m33*m44*m66
         *L6r*RN3e*RSe563p*RS356pe - 16./27.*m33*m44*m66*L6r*RN3e*
         RSe453p*RS345pe );

      M13p6x23L +=  + Ab(me2,mu2) * (  - 16./27.*m33*m44*m66*L6r*RN3e*
         RSe561p*RS356pe - 16./27.*m33*m44*m66*L6r*RN3e*RSe451p*RS345pe
          - 16./27.*m33*m44*m66*L6r*RN1e*RS145pe*RSe453p - 16./27.*m33*
         m44*m66*L6r*RN1e*RS156pe*RSe563p + 16./27.*m33*m44*m55*L7r*
         RNep*RSp563e*RSe463p + 16./27.*m33*m44*m55*L7r*RNep*RSp463e*
         RSe563p + 16./27.*m33*m44*m55*L7r*RNep*RSe561p*RSp463e + 16./
         27.*m33*m44*m55*L7r*RNep*RSe461p*RSp563e + 16./27.*m33*m44*m55
         *L7r*RNep*RSp561e*RSe463p + 16./27.*m33*m44*m55*L7r*RNep*
         RSp461e*RSe563p - 16./27.*m33*m44*m55*L7r*RN3e*RSe563p*RS346pe
          - 16./27.*m33*m44*m55*L7r*RN3e*RSe463p*RS356pe - 16./27.*m33*
         m44*m55*L7r*RN3e*RSe561p*RS346pe - 16./27.*m33*m44*m55*L7r*
         RN3e*RSe461p*RS356pe - 16./27.*m33*m44*m55*L7r*RN1e*RS146pe*
         RSe563p - 16./27.*m33*m44*m55*L7r*RN1e*RS156pe*RSe463p + 16./
         27.*m33*m44*m55*L6r*RNep*RSp563e*RSe563p + 16./27.*m33*m44*m55
         *L6r*RNep*RSp463e*RSe463p + 16./27.*m33*m44*m55*L6r*RNep*
         RSe561p*RSp563e );

      M13p6x23L +=  + Ab(me2,mu2) * ( 16./27.*m33*m44*m55*L6r*RNep*
         RSe461p*RSp463e + 16./27.*m33*m44*m55*L6r*RNep*RSp561e*RSe563p
          + 16./27.*m33*m44*m55*L6r*RNep*RSp461e*RSe463p - 16./27.*m33*
         m44*m55*L6r*RN3e*RSe563p*RS356pe - 16./27.*m33*m44*m55*L6r*
         RN3e*RSe463p*RS346pe - 16./27.*m33*m44*m55*L6r*RN3e*RSe561p*
         RS356pe - 16./27.*m33*m44*m55*L6r*RN3e*RSe461p*RS346pe - 16./
         27.*m33*m44*m55*L6r*RN1e*RS146pe*RSe463p - 16./27.*m33*m44*m55
         *L6r*RN1e*RS156pe*RSe563p + 16./27.*m33*pow(m44,2)*L8r*RNep*
         RSp563e*RSe563p + 16./27.*m33*pow(m44,2)*L8r*RNep*RSe561p*
         RSp563e + 16./27.*m33*pow(m44,2)*L8r*RNep*RSp561e*RSe563p - 16.
         /27.*m33*pow(m44,2)*L8r*RN3e*RSe563p*RS356pe - 16./27.*m33*
         pow(m44,2)*L8r*RN3e*RSe561p*RS356pe - 16./27.*m33*pow(m44,2)*
         L8r*RN1e*RS156pe*RSe563p + 16./27.*m33*pow(m44,2)*L7r*RNep*
         RSp563e*RSe563p + 16./27.*m33*pow(m44,2)*L7r*RNep*RSe561p*
         RSp563e + 16./27.*m33*pow(m44,2)*L7r*RNep*RSp561e*RSe563p - 16.
         /27.*m33*pow(m44,2)*L7r*RN3e*RSe563p*RS356pe );

      M13p6x23L +=  + Ab(me2,mu2) * (  - 16./27.*m33*pow(m44,2)*L7r*
         RN3e*RSe561p*RS356pe - 16./27.*m33*pow(m44,2)*L7r*RN1e*RS156pe
         *RSe563p + 16./27.*m33*pow(m44,2)*L6r*RNep*RSp563e*RSe563p + 
         16./27.*m33*pow(m44,2)*L6r*RNep*RSe561p*RSp563e + 16./27.*m33*
         pow(m44,2)*L6r*RNep*RSp561e*RSe563p - 16./27.*m33*pow(m44,2)*
         L6r*RN3e*RSe563p*RS356pe - 16./27.*m33*pow(m44,2)*L6r*RN3e*
         RSe561p*RS356pe - 16./27.*m33*pow(m44,2)*L6r*RN1e*RS156pe*
         RSe563p - 16./9.*pow(m33,2)*L8r*RSe45613p - 32./9.*pow(m33,2)*
         L8r*RSe45633p + 4./9.*pow(m33,2)*L5r*RSe45613p + 8./9.*pow(
         m33,2)*L5r*RSe45633p + 16./9.*pow(m33,2)*me2*L8r*RN3e*
         RSe45613p - 16./9.*pow(m33,2)*me2*L8r*RN3e*RSe45633p - 16./9.*
         pow(m33,2)*me2*L5r*RN3e*RSe45613p + 16./9.*pow(m33,2)*m66*L7r*
         RN3e*RSe453p + 16./9.*pow(m33,2)*m66*L7r*RN3e*RSe451p + 16./9.
         *pow(m33,2)*m66*L6r*RN3e*RSe45613p + 16./9.*pow(m33,2)*m66*L6r
         *RN3e*RSe45633p + 16./9.*pow(m33,2)*m55*L7r*RN3e*RSe463p + 16./
         9.*pow(m33,2)*m55*L7r*RN3e*RSe461p );

      M13p6x23L +=  + Ab(me2,mu2) * ( 16./9.*pow(m33,2)*m55*L6r*RN3e*
         RSe45613p + 16./9.*pow(m33,2)*m55*L6r*RN3e*RSe45633p + 16./9.*
         pow(m33,2)*m44*L7r*RN3e*RSe563p + 16./9.*pow(m33,2)*m44*L7r*
         RN3e*RSe561p + 16./9.*pow(m33,2)*m44*L6r*RN3e*RSe45613p + 16./
         9.*pow(m33,2)*m44*L6r*RN3e*RSe45633p + 16./9.*pow(m33,3)*L8r*
         RN3e*RSe45613p + 16./9.*pow(m33,3)*L8r*RN3e*RSe45633p + 8./9.*
         m11*me2*L5r*RSe45613p + 4./9.*m11*me2*L5r*RSe45633p + 4./3.*
         m11*me2*L4r*RSe564p + 4./3.*m11*me2*L4r*RSe465p + 4./3.*m11*
         me2*L4r*RSe456p - 2./3.*m11*me2*L3r*RSe45613p - 5./3.*m11*me2*
         L3r*RSe45633p - 5./3.*m11*me2*L3r*RSe45611p - 2./3.*m11*me2*
         L2r*RSe564p - 2./3.*m11*me2*L2r*RSe465p - 2./3.*m11*me2*L2r*
         RSe456p - 8./3.*m11*me2*L1r*RSe564p - 8./3.*m11*me2*L1r*
         RSe465p - 8./3.*m11*me2*L1r*RSe456p - 8./3.*m11*me2*L0r*
         RSe45613p - 2./3.*m11*me2*L0r*RSe45633p - 2./3.*m11*me2*L0r*
         RSe45611p - 16./9.*m11*m66*L7r*RSe453p - 32./9.*m11*m66*L7r*
         RSe451p );

      M13p6x23L +=  + Ab(me2,mu2) * (  - 16./9.*m11*m66*L6r*RSe45613p
          - 8./9.*m11*m66*L6r*RSe45611p - 8./3.*m11*m66*L6r*RSe456p - 8.
         /27.*m11*m66*L5r*RSe451p*RSe453p - 8./27.*m11*m66*L5r*RSe451p*
         RSp453e - 4./27.*m11*m66*L5r*pow(RSe451p,2) - 8./27.*m11*m66*
         L5r*RSp451e*RSe453p - 8./27.*m11*m66*L5r*RSp451e*RSe451p + 8./
         9.*m11*m66*L4r*RSe45613p + 4./9.*m11*m66*L4r*RSe45611p - 8./27.
         *m11*m66*L4r*RSe561p*RSe563p - 8./27.*m11*m66*L4r*RSe561p*
         RSp563e - 4./27.*m11*m66*L4r*pow(RSe561p,2) - 8./27.*m11*m66*
         L4r*RSe461p*RSe463p - 8./27.*m11*m66*L4r*RSe461p*RSp463e - 4./
         27.*m11*m66*L4r*pow(RSe461p,2) - 8./27.*m11*m66*L4r*RSe451p*
         RSe453p - 8./27.*m11*m66*L4r*RSe451p*RSp453e - 4./27.*m11*m66*
         L4r*pow(RSe451p,2) - 8./27.*m11*m66*L4r*RSp561e*RSe563p - 8./
         27.*m11*m66*L4r*RSp561e*RSe561p - 8./27.*m11*m66*L4r*RSp461e*
         RSe463p - 8./27.*m11*m66*L4r*RSp461e*RSe461p - 8./27.*m11*m66*
         L4r*RSp451e*RSe453p - 8./27.*m11*m66*L4r*RSp451e*RSe451p + 4./
         3.*m11*m66*L4r*RSe456p );

      M13p6x23L +=  + Ab(me2,mu2) * (  - 8./27.*m11*m66*mp2*L5r*RNep*
         RSe451p*RSp453e - 8./27.*m11*m66*mp2*L5r*RNep*RSp451e*RSe453p
          - 8./27.*m11*m66*mp2*L5r*RNep*RSp451e*RSe451p - 8./27.*m11*
         m66*mp2*L4r*RNep*RSe561p*RSp563e - 8./27.*m11*m66*mp2*L4r*RNep
         *RSe461p*RSp463e - 8./27.*m11*m66*mp2*L4r*RNep*RSe451p*RSp453e
          - 8./27.*m11*m66*mp2*L4r*RNep*RSp561e*RSe563p - 8./27.*m11*
         m66*mp2*L4r*RNep*RSp561e*RSe561p - 8./27.*m11*m66*mp2*L4r*RNep
         *RSp461e*RSe463p - 8./27.*m11*m66*mp2*L4r*RNep*RSp461e*RSe461p
          - 8./27.*m11*m66*mp2*L4r*RNep*RSp451e*RSe453p - 8./27.*m11*
         m66*mp2*L4r*RNep*RSp451e*RSe451p + 16./9.*m11*m66*me2*L7r*RN1e
         *RSe453p - 16./9.*m11*m66*me2*L7r*RN1e*RSe451p + 16./9.*m11*
         m66*me2*L6r*RN1e*RSe45613p - 16./9.*m11*m66*me2*L6r*RN1e*
         RSe45611p + 8./27.*m11*m66*me2*L5r*RN3e*RSe451p*RS345pe + 16./
         27.*m11*m66*me2*L5r*RN1e*RS145pe*RSe453p - 8./9.*m11*m66*me2*
         L4r*RN3e*RSe45613p + 8./27.*m11*m66*me2*L4r*RN3e*RSe561p*
         RS356pe );

      M13p6x23L +=  + Ab(me2,mu2) * ( 8./27.*m11*m66*me2*L4r*RN3e*
         RSe461p*RS346pe + 8./27.*m11*m66*me2*L4r*RN3e*RSe451p*RS345pe
          - 16./9.*m11*m66*me2*L4r*RN1e*RSe45613p + 16./27.*m11*m66*me2
         *L4r*RN1e*RS145pe*RSe453p + 16./27.*m11*m66*me2*L4r*RN1e*
         RS146pe*RSe463p + 16./27.*m11*m66*me2*L4r*RN1e*RS156pe*RSe563p
          + 16./27.*m11*pow(m66,2)*L8r*RNep*RSe451p*RSp453e + 16./27.*
         m11*pow(m66,2)*L8r*RNep*RSp451e*RSe453p + 16./27.*m11*pow(
         m66,2)*L8r*RNep*RSp451e*RSe451p - 16./27.*m11*pow(m66,2)*L8r*
         RN3e*RSe451p*RS345pe - 16./27.*m11*pow(m66,2)*L8r*RN1e*RS145pe
         *RSe453p - 16./27.*m11*pow(m66,2)*L8r*RN1e*RSe451p*RS145pe + 
         16./27.*m11*pow(m66,2)*L7r*RNep*RSe451p*RSp453e + 16./27.*m11*
         pow(m66,2)*L7r*RNep*RSp451e*RSe453p + 16./27.*m11*pow(m66,2)*
         L7r*RNep*RSp451e*RSe451p - 16./27.*m11*pow(m66,2)*L7r*RN3e*
         RSe451p*RS345pe - 16./27.*m11*pow(m66,2)*L7r*RN1e*RS145pe*
         RSe453p - 16./27.*m11*pow(m66,2)*L7r*RN1e*RSe451p*RS145pe + 16.
         /27.*m11*pow(m66,2)*L6r*RNep*RSe451p*RSp453e );

      M13p6x23L +=  + Ab(me2,mu2) * ( 16./27.*m11*pow(m66,2)*L6r*RNep*
         RSp451e*RSe453p + 16./27.*m11*pow(m66,2)*L6r*RNep*RSp451e*
         RSe451p - 16./27.*m11*pow(m66,2)*L6r*RN3e*RSe451p*RS345pe - 16.
         /27.*m11*pow(m66,2)*L6r*RN1e*RS145pe*RSe453p - 16./27.*m11*
         pow(m66,2)*L6r*RN1e*RSe451p*RS145pe - 16./9.*m11*m55*L7r*
         RSe463p - 32./9.*m11*m55*L7r*RSe461p - 16./9.*m11*m55*L6r*
         RSe45613p - 8./9.*m11*m55*L6r*RSe45611p - 8./3.*m11*m55*L6r*
         RSe465p - 8./27.*m11*m55*L5r*RSe461p*RSe463p - 8./27.*m11*m55*
         L5r*RSe461p*RSp463e - 4./27.*m11*m55*L5r*pow(RSe461p,2) - 8./
         27.*m11*m55*L5r*RSp461e*RSe463p - 8./27.*m11*m55*L5r*RSp461e*
         RSe461p + 8./9.*m11*m55*L4r*RSe45613p + 4./9.*m11*m55*L4r*
         RSe45611p - 8./27.*m11*m55*L4r*RSe561p*RSe563p - 8./27.*m11*
         m55*L4r*RSe561p*RSp563e - 4./27.*m11*m55*L4r*pow(RSe561p,2) - 
         8./27.*m11*m55*L4r*RSe461p*RSe463p - 8./27.*m11*m55*L4r*
         RSe461p*RSp463e - 4./27.*m11*m55*L4r*pow(RSe461p,2) - 8./27.*
         m11*m55*L4r*RSe451p*RSe453p );

      M13p6x23L +=  + Ab(me2,mu2) * (  - 8./27.*m11*m55*L4r*RSe451p*
         RSp453e - 4./27.*m11*m55*L4r*pow(RSe451p,2) - 8./27.*m11*m55*
         L4r*RSp561e*RSe563p - 8./27.*m11*m55*L4r*RSp561e*RSe561p - 8./
         27.*m11*m55*L4r*RSp461e*RSe463p - 8./27.*m11*m55*L4r*RSp461e*
         RSe461p - 8./27.*m11*m55*L4r*RSp451e*RSe453p - 8./27.*m11*m55*
         L4r*RSp451e*RSe451p + 4./3.*m11*m55*L4r*RSe465p - 8./27.*m11*
         m55*mp2*L5r*RNep*RSe461p*RSp463e - 8./27.*m11*m55*mp2*L5r*RNep
         *RSp461e*RSe463p - 8./27.*m11*m55*mp2*L5r*RNep*RSp461e*RSe461p
          - 8./27.*m11*m55*mp2*L4r*RNep*RSe561p*RSp563e - 8./27.*m11*
         m55*mp2*L4r*RNep*RSe461p*RSp463e - 8./27.*m11*m55*mp2*L4r*RNep
         *RSe451p*RSp453e - 8./27.*m11*m55*mp2*L4r*RNep*RSp561e*RSe563p
          - 8./27.*m11*m55*mp2*L4r*RNep*RSp561e*RSe561p - 8./27.*m11*
         m55*mp2*L4r*RNep*RSp461e*RSe463p - 8./27.*m11*m55*mp2*L4r*RNep
         *RSp461e*RSe461p - 8./27.*m11*m55*mp2*L4r*RNep*RSp451e*RSe453p
          - 8./27.*m11*m55*mp2*L4r*RNep*RSp451e*RSe451p + 16./9.*m11*
         m55*me2*L7r*RN1e*RSe463p );

      M13p6x23L +=  + Ab(me2,mu2) * (  - 16./9.*m11*m55*me2*L7r*RN1e*
         RSe461p + 16./9.*m11*m55*me2*L6r*RN1e*RSe45613p - 16./9.*m11*
         m55*me2*L6r*RN1e*RSe45611p + 8./27.*m11*m55*me2*L5r*RN3e*
         RSe461p*RS346pe + 16./27.*m11*m55*me2*L5r*RN1e*RS146pe*RSe463p
          - 8./9.*m11*m55*me2*L4r*RN3e*RSe45613p + 8./27.*m11*m55*me2*
         L4r*RN3e*RSe561p*RS356pe + 8./27.*m11*m55*me2*L4r*RN3e*RSe461p
         *RS346pe + 8./27.*m11*m55*me2*L4r*RN3e*RSe451p*RS345pe - 16./9.
         *m11*m55*me2*L4r*RN1e*RSe45613p + 16./27.*m11*m55*me2*L4r*RN1e
         *RS145pe*RSe453p + 16./27.*m11*m55*me2*L4r*RN1e*RS146pe*
         RSe463p + 16./27.*m11*m55*me2*L4r*RN1e*RS156pe*RSe563p + 16./
         27.*m11*m55*m66*L7r*RNep*RSe461p*RSp453e + 16./27.*m11*m55*m66
         *L7r*RNep*RSe451p*RSp463e + 16./27.*m11*m55*m66*L7r*RNep*
         RSp461e*RSe453p + 16./27.*m11*m55*m66*L7r*RNep*RSp461e*RSe451p
          + 16./27.*m11*m55*m66*L7r*RNep*RSp451e*RSe463p + 16./27.*m11*
         m55*m66*L7r*RNep*RSp451e*RSe461p - 16./27.*m11*m55*m66*L7r*
         RN3e*RSe461p*RS345pe );

      M13p6x23L +=  + Ab(me2,mu2) * (  - 16./27.*m11*m55*m66*L7r*RN3e*
         RSe451p*RS346pe - 16./27.*m11*m55*m66*L7r*RN1e*RS145pe*RSe463p
          - 16./27.*m11*m55*m66*L7r*RN1e*RS146pe*RSe453p - 16./27.*m11*
         m55*m66*L7r*RN1e*RSe461p*RS145pe - 16./27.*m11*m55*m66*L7r*
         RN1e*RSe451p*RS146pe + 16./27.*m11*m55*m66*L6r*RNep*RSe461p*
         RSp463e + 16./27.*m11*m55*m66*L6r*RNep*RSe451p*RSp453e + 16./
         27.*m11*m55*m66*L6r*RNep*RSp461e*RSe463p + 16./27.*m11*m55*m66
         *L6r*RNep*RSp461e*RSe461p + 16./27.*m11*m55*m66*L6r*RNep*
         RSp451e*RSe453p + 16./27.*m11*m55*m66*L6r*RNep*RSp451e*RSe451p
          - 16./27.*m11*m55*m66*L6r*RN3e*RSe461p*RS346pe - 16./27.*m11*
         m55*m66*L6r*RN3e*RSe451p*RS345pe - 16./27.*m11*m55*m66*L6r*
         RN1e*RS145pe*RSe453p - 16./27.*m11*m55*m66*L6r*RN1e*RS146pe*
         RSe463p - 16./27.*m11*m55*m66*L6r*RN1e*RSe461p*RS146pe - 16./
         27.*m11*m55*m66*L6r*RN1e*RSe451p*RS145pe + 16./27.*m11*pow(
         m55,2)*L8r*RNep*RSe461p*RSp463e + 16./27.*m11*pow(m55,2)*L8r*
         RNep*RSp461e*RSe463p );

      M13p6x23L +=  + Ab(me2,mu2) * ( 16./27.*m11*pow(m55,2)*L8r*RNep*
         RSp461e*RSe461p - 16./27.*m11*pow(m55,2)*L8r*RN3e*RSe461p*
         RS346pe - 16./27.*m11*pow(m55,2)*L8r*RN1e*RS146pe*RSe463p - 16.
         /27.*m11*pow(m55,2)*L8r*RN1e*RSe461p*RS146pe + 16./27.*m11*
         pow(m55,2)*L7r*RNep*RSe461p*RSp463e + 16./27.*m11*pow(m55,2)*
         L7r*RNep*RSp461e*RSe463p + 16./27.*m11*pow(m55,2)*L7r*RNep*
         RSp461e*RSe461p - 16./27.*m11*pow(m55,2)*L7r*RN3e*RSe461p*
         RS346pe - 16./27.*m11*pow(m55,2)*L7r*RN1e*RS146pe*RSe463p - 16.
         /27.*m11*pow(m55,2)*L7r*RN1e*RSe461p*RS146pe + 16./27.*m11*
         pow(m55,2)*L6r*RNep*RSe461p*RSp463e + 16./27.*m11*pow(m55,2)*
         L6r*RNep*RSp461e*RSe463p + 16./27.*m11*pow(m55,2)*L6r*RNep*
         RSp461e*RSe461p - 16./27.*m11*pow(m55,2)*L6r*RN3e*RSe461p*
         RS346pe - 16./27.*m11*pow(m55,2)*L6r*RN1e*RS146pe*RSe463p - 16.
         /27.*m11*pow(m55,2)*L6r*RN1e*RSe461p*RS146pe - 16./9.*m11*m44*
         L7r*RSe563p - 32./9.*m11*m44*L7r*RSe561p - 16./9.*m11*m44*L6r*
         RSe45613p );

      M13p6x23L +=  + Ab(me2,mu2) * (  - 8./9.*m11*m44*L6r*RSe45611p - 
         8./3.*m11*m44*L6r*RSe564p - 8./27.*m11*m44*L5r*RSe561p*RSe563p
          - 8./27.*m11*m44*L5r*RSe561p*RSp563e - 4./27.*m11*m44*L5r*
         pow(RSe561p,2) - 8./27.*m11*m44*L5r*RSp561e*RSe563p - 8./27.*
         m11*m44*L5r*RSp561e*RSe561p + 8./9.*m11*m44*L4r*RSe45613p + 4./
         9.*m11*m44*L4r*RSe45611p - 8./27.*m11*m44*L4r*RSe561p*RSe563p
          - 8./27.*m11*m44*L4r*RSe561p*RSp563e - 4./27.*m11*m44*L4r*
         pow(RSe561p,2) - 8./27.*m11*m44*L4r*RSe461p*RSe463p - 8./27.*
         m11*m44*L4r*RSe461p*RSp463e - 4./27.*m11*m44*L4r*pow(
         RSe461p,2) - 8./27.*m11*m44*L4r*RSe451p*RSe453p - 8./27.*m11*
         m44*L4r*RSe451p*RSp453e - 4./27.*m11*m44*L4r*pow(RSe451p,2) - 
         8./27.*m11*m44*L4r*RSp561e*RSe563p - 8./27.*m11*m44*L4r*
         RSp561e*RSe561p - 8./27.*m11*m44*L4r*RSp461e*RSe463p - 8./27.*
         m11*m44*L4r*RSp461e*RSe461p - 8./27.*m11*m44*L4r*RSp451e*
         RSe453p - 8./27.*m11*m44*L4r*RSp451e*RSe451p + 4./3.*m11*m44*
         L4r*RSe564p );

      M13p6x23L +=  + Ab(me2,mu2) * (  - 8./27.*m11*m44*mp2*L5r*RNep*
         RSe561p*RSp563e - 8./27.*m11*m44*mp2*L5r*RNep*RSp561e*RSe563p
          - 8./27.*m11*m44*mp2*L5r*RNep*RSp561e*RSe561p - 8./27.*m11*
         m44*mp2*L4r*RNep*RSe561p*RSp563e - 8./27.*m11*m44*mp2*L4r*RNep
         *RSe461p*RSp463e - 8./27.*m11*m44*mp2*L4r*RNep*RSe451p*RSp453e
          - 8./27.*m11*m44*mp2*L4r*RNep*RSp561e*RSe563p - 8./27.*m11*
         m44*mp2*L4r*RNep*RSp561e*RSe561p - 8./27.*m11*m44*mp2*L4r*RNep
         *RSp461e*RSe463p - 8./27.*m11*m44*mp2*L4r*RNep*RSp461e*RSe461p
          - 8./27.*m11*m44*mp2*L4r*RNep*RSp451e*RSe453p - 8./27.*m11*
         m44*mp2*L4r*RNep*RSp451e*RSe451p + 16./9.*m11*m44*me2*L7r*RN1e
         *RSe563p - 16./9.*m11*m44*me2*L7r*RN1e*RSe561p + 16./9.*m11*
         m44*me2*L6r*RN1e*RSe45613p - 16./9.*m11*m44*me2*L6r*RN1e*
         RSe45611p + 8./27.*m11*m44*me2*L5r*RN3e*RSe561p*RS356pe + 16./
         27.*m11*m44*me2*L5r*RN1e*RS156pe*RSe563p - 8./9.*m11*m44*me2*
         L4r*RN3e*RSe45613p + 8./27.*m11*m44*me2*L4r*RN3e*RSe561p*
         RS356pe );

      M13p6x23L +=  + Ab(me2,mu2) * ( 8./27.*m11*m44*me2*L4r*RN3e*
         RSe461p*RS346pe + 8./27.*m11*m44*me2*L4r*RN3e*RSe451p*RS345pe
          - 16./9.*m11*m44*me2*L4r*RN1e*RSe45613p + 16./27.*m11*m44*me2
         *L4r*RN1e*RS145pe*RSe453p + 16./27.*m11*m44*me2*L4r*RN1e*
         RS146pe*RSe463p + 16./27.*m11*m44*me2*L4r*RN1e*RS156pe*RSe563p
          + 16./27.*m11*m44*m66*L7r*RNep*RSe561p*RSp453e + 16./27.*m11*
         m44*m66*L7r*RNep*RSe451p*RSp563e + 16./27.*m11*m44*m66*L7r*
         RNep*RSp561e*RSe453p + 16./27.*m11*m44*m66*L7r*RNep*RSp561e*
         RSe451p + 16./27.*m11*m44*m66*L7r*RNep*RSp451e*RSe563p + 16./
         27.*m11*m44*m66*L7r*RNep*RSp451e*RSe561p - 16./27.*m11*m44*m66
         *L7r*RN3e*RSe561p*RS345pe - 16./27.*m11*m44*m66*L7r*RN3e*
         RSe451p*RS356pe - 16./27.*m11*m44*m66*L7r*RN1e*RS145pe*RSe563p
          - 16./27.*m11*m44*m66*L7r*RN1e*RS156pe*RSe453p - 16./27.*m11*
         m44*m66*L7r*RN1e*RSe561p*RS145pe - 16./27.*m11*m44*m66*L7r*
         RN1e*RSe451p*RS156pe + 16./27.*m11*m44*m66*L6r*RNep*RSe561p*
         RSp563e );

      M13p6x23L +=  + Ab(me2,mu2) * ( 16./27.*m11*m44*m66*L6r*RNep*
         RSe451p*RSp453e + 16./27.*m11*m44*m66*L6r*RNep*RSp561e*RSe563p
          + 16./27.*m11*m44*m66*L6r*RNep*RSp561e*RSe561p + 16./27.*m11*
         m44*m66*L6r*RNep*RSp451e*RSe453p + 16./27.*m11*m44*m66*L6r*
         RNep*RSp451e*RSe451p - 16./27.*m11*m44*m66*L6r*RN3e*RSe561p*
         RS356pe - 16./27.*m11*m44*m66*L6r*RN3e*RSe451p*RS345pe - 16./
         27.*m11*m44*m66*L6r*RN1e*RS145pe*RSe453p - 16./27.*m11*m44*m66
         *L6r*RN1e*RS156pe*RSe563p - 16./27.*m11*m44*m66*L6r*RN1e*
         RSe561p*RS156pe - 16./27.*m11*m44*m66*L6r*RN1e*RSe451p*RS145pe
          + 16./27.*m11*m44*m55*L7r*RNep*RSe561p*RSp463e + 16./27.*m11*
         m44*m55*L7r*RNep*RSe461p*RSp563e + 16./27.*m11*m44*m55*L7r*
         RNep*RSp561e*RSe463p + 16./27.*m11*m44*m55*L7r*RNep*RSp561e*
         RSe461p + 16./27.*m11*m44*m55*L7r*RNep*RSp461e*RSe563p + 16./
         27.*m11*m44*m55*L7r*RNep*RSp461e*RSe561p - 16./27.*m11*m44*m55
         *L7r*RN3e*RSe561p*RS346pe - 16./27.*m11*m44*m55*L7r*RN3e*
         RSe461p*RS356pe );

      M13p6x23L +=  + Ab(me2,mu2) * (  - 16./27.*m11*m44*m55*L7r*RN1e*
         RS146pe*RSe563p - 16./27.*m11*m44*m55*L7r*RN1e*RS156pe*RSe463p
          - 16./27.*m11*m44*m55*L7r*RN1e*RSe561p*RS146pe - 16./27.*m11*
         m44*m55*L7r*RN1e*RSe461p*RS156pe + 16./27.*m11*m44*m55*L6r*
         RNep*RSe561p*RSp563e + 16./27.*m11*m44*m55*L6r*RNep*RSe461p*
         RSp463e + 16./27.*m11*m44*m55*L6r*RNep*RSp561e*RSe563p + 16./
         27.*m11*m44*m55*L6r*RNep*RSp561e*RSe561p + 16./27.*m11*m44*m55
         *L6r*RNep*RSp461e*RSe463p + 16./27.*m11*m44*m55*L6r*RNep*
         RSp461e*RSe461p - 16./27.*m11*m44*m55*L6r*RN3e*RSe561p*RS356pe
          - 16./27.*m11*m44*m55*L6r*RN3e*RSe461p*RS346pe - 16./27.*m11*
         m44*m55*L6r*RN1e*RS146pe*RSe463p - 16./27.*m11*m44*m55*L6r*
         RN1e*RS156pe*RSe563p - 16./27.*m11*m44*m55*L6r*RN1e*RSe561p*
         RS156pe - 16./27.*m11*m44*m55*L6r*RN1e*RSe461p*RS146pe + 16./
         27.*m11*pow(m44,2)*L8r*RNep*RSe561p*RSp563e + 16./27.*m11*pow(
         m44,2)*L8r*RNep*RSp561e*RSe563p + 16./27.*m11*pow(m44,2)*L8r*
         RNep*RSp561e*RSe561p );

      M13p6x23L +=  + Ab(me2,mu2) * (  - 16./27.*m11*pow(m44,2)*L8r*
         RN3e*RSe561p*RS356pe - 16./27.*m11*pow(m44,2)*L8r*RN1e*RS156pe
         *RSe563p - 16./27.*m11*pow(m44,2)*L8r*RN1e*RSe561p*RS156pe + 
         16./27.*m11*pow(m44,2)*L7r*RNep*RSe561p*RSp563e + 16./27.*m11*
         pow(m44,2)*L7r*RNep*RSp561e*RSe563p + 16./27.*m11*pow(m44,2)*
         L7r*RNep*RSp561e*RSe561p - 16./27.*m11*pow(m44,2)*L7r*RN3e*
         RSe561p*RS356pe - 16./27.*m11*pow(m44,2)*L7r*RN1e*RS156pe*
         RSe563p - 16./27.*m11*pow(m44,2)*L7r*RN1e*RSe561p*RS156pe + 16.
         /27.*m11*pow(m44,2)*L6r*RNep*RSe561p*RSp563e + 16./27.*m11*
         pow(m44,2)*L6r*RNep*RSp561e*RSe563p + 16./27.*m11*pow(m44,2)*
         L6r*RNep*RSp561e*RSe561p - 16./27.*m11*pow(m44,2)*L6r*RN3e*
         RSe561p*RS356pe - 16./27.*m11*pow(m44,2)*L6r*RN1e*RS156pe*
         RSe563p - 16./27.*m11*pow(m44,2)*L6r*RN1e*RSe561p*RS156pe - 16.
         /3.*m11*m33*L8r*RSe45613p - 8./3.*m11*m33*L8r*RSe45633p - 8./3.
         *m11*m33*L8r*RSe45611p + 8./9.*m11*m33*L5r*RSe45613p + 8./9.*
         m11*m33*L5r*RSe45633p );

      M13p6x23L +=  + Ab(me2,mu2) * ( 8./9.*m11*m33*L5r*RSe45611p - 8./
         9.*m11*m33*me2*L5r*RN3e*RSe45613p - 8./9.*m11*m33*me2*L5r*RN1e
         *RSe45613p + 16./9.*m11*m33*m66*L7r*RN3e*RSe451p + 16./9.*m11*
         m33*m66*L7r*RN1e*RSe453p + 16./9.*m11*m33*m66*L6r*RN3e*
         RSe45613p + 16./9.*m11*m33*m66*L6r*RN1e*RSe45613p + 16./9.*m11
         *m33*m55*L7r*RN3e*RSe461p + 16./9.*m11*m33*m55*L7r*RN1e*
         RSe463p + 16./9.*m11*m33*m55*L6r*RN3e*RSe45613p + 16./9.*m11*
         m33*m55*L6r*RN1e*RSe45613p + 16./9.*m11*m33*m44*L7r*RN3e*
         RSe561p + 16./9.*m11*m33*m44*L7r*RN1e*RSe563p + 16./9.*m11*m33
         *m44*L6r*RN3e*RSe45613p + 16./9.*m11*m33*m44*L6r*RN1e*
         RSe45613p + 16./9.*m11*pow(m33,2)*L8r*RN3e*RSe45613p - 16./9.*
         pow(m11,2)*L8r*RSe45613p - 32./9.*pow(m11,2)*L8r*RSe45611p + 4.
         /9.*pow(m11,2)*L5r*RSe45613p + 8./9.*pow(m11,2)*L5r*RSe45611p
          + 16./9.*pow(m11,2)*me2*L8r*RN1e*RSe45613p - 16./9.*pow(
         m11,2)*me2*L8r*RN1e*RSe45611p - 16./9.*pow(m11,2)*me2*L5r*RN1e
         *RSe45613p );

      M13p6x23L +=  + Ab(me2,mu2) * ( 16./9.*pow(m11,2)*m66*L7r*RN1e*
         RSe453p + 16./9.*pow(m11,2)*m66*L7r*RN1e*RSe451p + 16./9.*pow(
         m11,2)*m66*L6r*RN1e*RSe45613p + 16./9.*pow(m11,2)*m66*L6r*RN1e
         *RSe45611p + 16./9.*pow(m11,2)*m55*L7r*RN1e*RSe463p + 16./9.*
         pow(m11,2)*m55*L7r*RN1e*RSe461p + 16./9.*pow(m11,2)*m55*L6r*
         RN1e*RSe45613p + 16./9.*pow(m11,2)*m55*L6r*RN1e*RSe45611p + 16.
         /9.*pow(m11,2)*m44*L7r*RN1e*RSe563p + 16./9.*pow(m11,2)*m44*
         L7r*RN1e*RSe561p + 16./9.*pow(m11,2)*m44*L6r*RN1e*RSe45613p + 
         16./9.*pow(m11,2)*m44*L6r*RN1e*RSe45611p + 16./9.*pow(m11,2)*
         m33*L8r*RN1e*RSe45613p + 16./9.*pow(m11,3)*L8r*RN1e*RSe45613p
          + 16./9.*pow(m11,3)*L8r*RN1e*RSe45611p );

      M13p6x23L +=  + Ab(mp2,mu2) * ( 8./27.*m66*mp2*L5r*RSp453e*
         RS345pe + 8./27.*m66*mp2*L5r*RSp453e*RSe453p + 8./27.*m66*mp2*
         L5r*pow(RSp453e,2) - 8./27.*m66*mp2*L5r*RS145pe*RSp453e - 8./
         27.*m66*mp2*L5r*RSe451p*RSp453e - 8./27.*m66*mp2*L5r*RSp451e*
         RS345pe - 8./27.*m66*mp2*L5r*RSp451e*RSe453p - 16./27.*m66*mp2
         *L5r*RSp451e*RSp453e + 8./27.*m66*mp2*L5r*RSp451e*RS145pe + 8./
         27.*m66*mp2*L5r*RSp451e*RSe451p + 8./27.*m66*mp2*L5r*pow(
         RSp451e,2) + 16./9.*m66*mp2*L4r*RSp45613e - 8./9.*m66*mp2*L4r*
         RSp45633e + 8./27.*m66*mp2*L4r*RSp563e*RS356pe + 8./27.*m66*
         mp2*L4r*RSp563e*RSe563p + 8./27.*m66*mp2*L4r*pow(RSp563e,2) + 
         8./27.*m66*mp2*L4r*RSp463e*RS346pe + 8./27.*m66*mp2*L4r*
         RSp463e*RSe463p + 8./27.*m66*mp2*L4r*pow(RSp463e,2) + 8./27.*
         m66*mp2*L4r*RSp453e*RS345pe + 8./27.*m66*mp2*L4r*RSp453e*
         RSe453p + 8./27.*m66*mp2*L4r*pow(RSp453e,2) - 8./9.*m66*mp2*
         L4r*RSp45611e - 8./27.*m66*mp2*L4r*RS145pe*RSp453e - 8./27.*
         m66*mp2*L4r*RS146pe*RSp463e );

      M13p6x23L +=  + Ab(mp2,mu2) * (  - 8./27.*m66*mp2*L4r*RS156pe*
         RSp563e - 8./27.*m66*mp2*L4r*RSe561p*RSp563e - 8./27.*m66*mp2*
         L4r*RSe461p*RSp463e - 8./27.*m66*mp2*L4r*RSe451p*RSp453e - 8./
         27.*m66*mp2*L4r*RSp561e*RS356pe - 8./27.*m66*mp2*L4r*RSp561e*
         RSe563p - 16./27.*m66*mp2*L4r*RSp561e*RSp563e + 8./27.*m66*mp2
         *L4r*RSp561e*RS156pe + 8./27.*m66*mp2*L4r*RSp561e*RSe561p + 8./
         27.*m66*mp2*L4r*pow(RSp561e,2) - 8./27.*m66*mp2*L4r*RSp461e*
         RS346pe - 8./27.*m66*mp2*L4r*RSp461e*RSe463p - 16./27.*m66*mp2
         *L4r*RSp461e*RSp463e + 8./27.*m66*mp2*L4r*RSp461e*RS146pe + 8./
         27.*m66*mp2*L4r*RSp461e*RSe461p + 8./27.*m66*mp2*L4r*pow(
         RSp461e,2) - 8./27.*m66*mp2*L4r*RSp451e*RS345pe - 8./27.*m66*
         mp2*L4r*RSp451e*RSe453p - 16./27.*m66*mp2*L4r*RSp451e*RSp453e
          + 8./27.*m66*mp2*L4r*RSp451e*RS145pe + 8./27.*m66*mp2*L4r*
         RSp451e*RSe451p + 8./27.*m66*mp2*L4r*pow(RSp451e,2) - 8./27.*
         m66*me2*mp2*L5r*RNep*RSp453e*RSe453p + 8./27.*m66*me2*mp2*L5r*
         RNep*RSe451p*RSp453e );

      M13p6x23L +=  + Ab(mp2,mu2) * ( 8./27.*m66*me2*mp2*L5r*RNep*
         RSp451e*RSe453p - 8./27.*m66*me2*mp2*L5r*RNep*RSp451e*RSe451p
          - 8./27.*m66*me2*mp2*L4r*RNep*RSp563e*RSe563p - 8./27.*m66*
         me2*mp2*L4r*RNep*RSp463e*RSe463p - 8./27.*m66*me2*mp2*L4r*RNep
         *RSp453e*RSe453p + 8./27.*m66*me2*mp2*L4r*RNep*RSe561p*RSp563e
          + 8./27.*m66*me2*mp2*L4r*RNep*RSe461p*RSp463e + 8./27.*m66*
         me2*mp2*L4r*RNep*RSe451p*RSp453e + 8./27.*m66*me2*mp2*L4r*RNep
         *RSp561e*RSe563p - 8./27.*m66*me2*mp2*L4r*RNep*RSp561e*RSe561p
          + 8./27.*m66*me2*mp2*L4r*RNep*RSp461e*RSe463p - 8./27.*m66*
         me2*mp2*L4r*RNep*RSp461e*RSe461p + 8./27.*m66*me2*mp2*L4r*RNep
         *RSp451e*RSe453p - 8./27.*m66*me2*mp2*L4r*RNep*RSp451e*RSe451p
          - 8./27.*pow(m66,2)*L8r*pow(RSp453e,2) + 16./27.*pow(m66,2)*
         L8r*RSp451e*RSp453e - 8./27.*pow(m66,2)*L8r*pow(RSp451e,2) - 8.
         /27.*pow(m66,2)*L7r*pow(RSp453e,2) + 16./27.*pow(m66,2)*L7r*
         RSp451e*RSp453e - 8./27.*pow(m66,2)*L7r*pow(RSp451e,2) - 8./27.
         *pow(m66,2)*L6r*pow(RSp453e,2) );

      M13p6x23L +=  + Ab(mp2,mu2) * ( 16./27.*pow(m66,2)*L6r*RSp451e*
         RSp453e - 8./27.*pow(m66,2)*L6r*pow(RSp451e,2) + 16./27.*pow(
         m66,2)*mp2*L8r*RNep*RSp453e*RSe453p - 16./27.*pow(m66,2)*mp2*
         L8r*RNep*RSe451p*RSp453e - 16./27.*pow(m66,2)*mp2*L8r*RNep*
         RSp451e*RSe453p + 16./27.*pow(m66,2)*mp2*L8r*RNep*RSp451e*
         RSe451p + 16./27.*pow(m66,2)*mp2*L8r*RN3p*RSp453e*RS345pe - 16.
         /27.*pow(m66,2)*mp2*L8r*RN3p*RSp451e*RS345pe - 16./27.*pow(
         m66,2)*mp2*L8r*RN1p*RS145pe*RSp453e + 16./27.*pow(m66,2)*mp2*
         L8r*RN1p*RSp451e*RS145pe + 16./27.*pow(m66,2)*mp2*L7r*RNep*
         RSp453e*RSe453p - 16./27.*pow(m66,2)*mp2*L7r*RNep*RSe451p*
         RSp453e - 16./27.*pow(m66,2)*mp2*L7r*RNep*RSp451e*RSe453p + 16.
         /27.*pow(m66,2)*mp2*L7r*RNep*RSp451e*RSe451p + 16./27.*pow(
         m66,2)*mp2*L7r*RN3p*RSp453e*RS345pe - 16./27.*pow(m66,2)*mp2*
         L7r*RN3p*RSp451e*RS345pe - 16./27.*pow(m66,2)*mp2*L7r*RN1p*
         RS145pe*RSp453e + 16./27.*pow(m66,2)*mp2*L7r*RN1p*RSp451e*
         RS145pe );

      M13p6x23L +=  + Ab(mp2,mu2) * ( 16./27.*pow(m66,2)*mp2*L6r*RNep*
         RSp453e*RSe453p - 16./27.*pow(m66,2)*mp2*L6r*RNep*RSe451p*
         RSp453e - 16./27.*pow(m66,2)*mp2*L6r*RNep*RSp451e*RSe453p + 16.
         /27.*pow(m66,2)*mp2*L6r*RNep*RSp451e*RSe451p + 16./27.*pow(
         m66,2)*mp2*L6r*RN3p*RSp453e*RS345pe - 16./27.*pow(m66,2)*mp2*
         L6r*RN3p*RSp451e*RS345pe - 16./27.*pow(m66,2)*mp2*L6r*RN1p*
         RS145pe*RSp453e + 16./27.*pow(m66,2)*mp2*L6r*RN1p*RSp451e*
         RS145pe + 8./27.*m55*mp2*L5r*RSp463e*RS346pe + 8./27.*m55*mp2*
         L5r*RSp463e*RSe463p + 8./27.*m55*mp2*L5r*pow(RSp463e,2) - 8./
         27.*m55*mp2*L5r*RS146pe*RSp463e - 8./27.*m55*mp2*L5r*RSe461p*
         RSp463e - 8./27.*m55*mp2*L5r*RSp461e*RS346pe - 8./27.*m55*mp2*
         L5r*RSp461e*RSe463p - 16./27.*m55*mp2*L5r*RSp461e*RSp463e + 8./
         27.*m55*mp2*L5r*RSp461e*RS146pe + 8./27.*m55*mp2*L5r*RSp461e*
         RSe461p + 8./27.*m55*mp2*L5r*pow(RSp461e,2) + 16./9.*m55*mp2*
         L4r*RSp45613e - 8./9.*m55*mp2*L4r*RSp45633e + 8./27.*m55*mp2*
         L4r*RSp563e*RS356pe );

      M13p6x23L +=  + Ab(mp2,mu2) * ( 8./27.*m55*mp2*L4r*RSp563e*
         RSe563p + 8./27.*m55*mp2*L4r*pow(RSp563e,2) + 8./27.*m55*mp2*
         L4r*RSp463e*RS346pe + 8./27.*m55*mp2*L4r*RSp463e*RSe463p + 8./
         27.*m55*mp2*L4r*pow(RSp463e,2) + 8./27.*m55*mp2*L4r*RSp453e*
         RS345pe + 8./27.*m55*mp2*L4r*RSp453e*RSe453p + 8./27.*m55*mp2*
         L4r*pow(RSp453e,2) - 8./9.*m55*mp2*L4r*RSp45611e - 8./27.*m55*
         mp2*L4r*RS145pe*RSp453e - 8./27.*m55*mp2*L4r*RS146pe*RSp463e
          - 8./27.*m55*mp2*L4r*RS156pe*RSp563e - 8./27.*m55*mp2*L4r*
         RSe561p*RSp563e - 8./27.*m55*mp2*L4r*RSe461p*RSp463e - 8./27.*
         m55*mp2*L4r*RSe451p*RSp453e - 8./27.*m55*mp2*L4r*RSp561e*
         RS356pe - 8./27.*m55*mp2*L4r*RSp561e*RSe563p - 16./27.*m55*mp2
         *L4r*RSp561e*RSp563e + 8./27.*m55*mp2*L4r*RSp561e*RS156pe + 8./
         27.*m55*mp2*L4r*RSp561e*RSe561p + 8./27.*m55*mp2*L4r*pow(
         RSp561e,2) - 8./27.*m55*mp2*L4r*RSp461e*RS346pe - 8./27.*m55*
         mp2*L4r*RSp461e*RSe463p - 16./27.*m55*mp2*L4r*RSp461e*RSp463e
          + 8./27.*m55*mp2*L4r*RSp461e*RS146pe );

      M13p6x23L +=  + Ab(mp2,mu2) * ( 8./27.*m55*mp2*L4r*RSp461e*
         RSe461p + 8./27.*m55*mp2*L4r*pow(RSp461e,2) - 8./27.*m55*mp2*
         L4r*RSp451e*RS345pe - 8./27.*m55*mp2*L4r*RSp451e*RSe453p - 16./
         27.*m55*mp2*L4r*RSp451e*RSp453e + 8./27.*m55*mp2*L4r*RSp451e*
         RS145pe + 8./27.*m55*mp2*L4r*RSp451e*RSe451p + 8./27.*m55*mp2*
         L4r*pow(RSp451e,2) - 8./27.*m55*me2*mp2*L5r*RNep*RSp463e*
         RSe463p + 8./27.*m55*me2*mp2*L5r*RNep*RSe461p*RSp463e + 8./27.
         *m55*me2*mp2*L5r*RNep*RSp461e*RSe463p - 8./27.*m55*me2*mp2*L5r
         *RNep*RSp461e*RSe461p - 8./27.*m55*me2*mp2*L4r*RNep*RSp563e*
         RSe563p - 8./27.*m55*me2*mp2*L4r*RNep*RSp463e*RSe463p - 8./27.
         *m55*me2*mp2*L4r*RNep*RSp453e*RSe453p + 8./27.*m55*me2*mp2*L4r
         *RNep*RSe561p*RSp563e + 8./27.*m55*me2*mp2*L4r*RNep*RSe461p*
         RSp463e + 8./27.*m55*me2*mp2*L4r*RNep*RSe451p*RSp453e + 8./27.
         *m55*me2*mp2*L4r*RNep*RSp561e*RSe563p - 8./27.*m55*me2*mp2*L4r
         *RNep*RSp561e*RSe561p + 8./27.*m55*me2*mp2*L4r*RNep*RSp461e*
         RSe463p );

      M13p6x23L +=  + Ab(mp2,mu2) * (  - 8./27.*m55*me2*mp2*L4r*RNep*
         RSp461e*RSe461p + 8./27.*m55*me2*mp2*L4r*RNep*RSp451e*RSe453p
          - 8./27.*m55*me2*mp2*L4r*RNep*RSp451e*RSe451p - 16./27.*m55*
         m66*L7r*RSp453e*RSp463e + 16./27.*m55*m66*L7r*RSp461e*RSp453e
          + 16./27.*m55*m66*L7r*RSp451e*RSp463e - 16./27.*m55*m66*L7r*
         RSp451e*RSp461e - 8./27.*m55*m66*L6r*pow(RSp463e,2) - 8./27.*
         m55*m66*L6r*pow(RSp453e,2) + 16./27.*m55*m66*L6r*RSp461e*
         RSp463e - 8./27.*m55*m66*L6r*pow(RSp461e,2) + 16./27.*m55*m66*
         L6r*RSp451e*RSp453e - 8./27.*m55*m66*L6r*pow(RSp451e,2) + 16./
         27.*m55*m66*mp2*L7r*RNep*RSp463e*RSe453p + 16./27.*m55*m66*mp2
         *L7r*RNep*RSp453e*RSe463p - 16./27.*m55*m66*mp2*L7r*RNep*
         RSe461p*RSp453e - 16./27.*m55*m66*mp2*L7r*RNep*RSe451p*RSp463e
          - 16./27.*m55*m66*mp2*L7r*RNep*RSp461e*RSe453p + 16./27.*m55*
         m66*mp2*L7r*RNep*RSp461e*RSe451p - 16./27.*m55*m66*mp2*L7r*
         RNep*RSp451e*RSe463p + 16./27.*m55*m66*mp2*L7r*RNep*RSp451e*
         RSe461p );

      M13p6x23L +=  + Ab(mp2,mu2) * ( 16./27.*m55*m66*mp2*L7r*RN3p*
         RSp463e*RS345pe + 16./27.*m55*m66*mp2*L7r*RN3p*RSp453e*RS346pe
          - 16./27.*m55*m66*mp2*L7r*RN3p*RSp461e*RS345pe - 16./27.*m55*
         m66*mp2*L7r*RN3p*RSp451e*RS346pe - 16./27.*m55*m66*mp2*L7r*
         RN1p*RS145pe*RSp463e - 16./27.*m55*m66*mp2*L7r*RN1p*RS146pe*
         RSp453e + 16./27.*m55*m66*mp2*L7r*RN1p*RSp461e*RS145pe + 16./
         27.*m55*m66*mp2*L7r*RN1p*RSp451e*RS146pe + 16./27.*m55*m66*mp2
         *L6r*RNep*RSp463e*RSe463p + 16./27.*m55*m66*mp2*L6r*RNep*
         RSp453e*RSe453p - 16./27.*m55*m66*mp2*L6r*RNep*RSe461p*RSp463e
          - 16./27.*m55*m66*mp2*L6r*RNep*RSe451p*RSp453e - 16./27.*m55*
         m66*mp2*L6r*RNep*RSp461e*RSe463p + 16./27.*m55*m66*mp2*L6r*
         RNep*RSp461e*RSe461p - 16./27.*m55*m66*mp2*L6r*RNep*RSp451e*
         RSe453p + 16./27.*m55*m66*mp2*L6r*RNep*RSp451e*RSe451p + 16./
         27.*m55*m66*mp2*L6r*RN3p*RSp463e*RS346pe + 16./27.*m55*m66*mp2
         *L6r*RN3p*RSp453e*RS345pe - 16./27.*m55*m66*mp2*L6r*RN3p*
         RSp461e*RS346pe );

      M13p6x23L +=  + Ab(mp2,mu2) * (  - 16./27.*m55*m66*mp2*L6r*RN3p*
         RSp451e*RS345pe - 16./27.*m55*m66*mp2*L6r*RN1p*RS145pe*RSp453e
          - 16./27.*m55*m66*mp2*L6r*RN1p*RS146pe*RSp463e + 16./27.*m55*
         m66*mp2*L6r*RN1p*RSp461e*RS146pe + 16./27.*m55*m66*mp2*L6r*
         RN1p*RSp451e*RS145pe - 8./27.*pow(m55,2)*L8r*pow(RSp463e,2) + 
         16./27.*pow(m55,2)*L8r*RSp461e*RSp463e - 8./27.*pow(m55,2)*L8r
         *pow(RSp461e,2) - 8./27.*pow(m55,2)*L7r*pow(RSp463e,2) + 16./
         27.*pow(m55,2)*L7r*RSp461e*RSp463e - 8./27.*pow(m55,2)*L7r*
         pow(RSp461e,2) - 8./27.*pow(m55,2)*L6r*pow(RSp463e,2) + 16./27.
         *pow(m55,2)*L6r*RSp461e*RSp463e - 8./27.*pow(m55,2)*L6r*pow(
         RSp461e,2) + 16./27.*pow(m55,2)*mp2*L8r*RNep*RSp463e*RSe463p
          - 16./27.*pow(m55,2)*mp2*L8r*RNep*RSe461p*RSp463e - 16./27.*
         pow(m55,2)*mp2*L8r*RNep*RSp461e*RSe463p + 16./27.*pow(m55,2)*
         mp2*L8r*RNep*RSp461e*RSe461p + 16./27.*pow(m55,2)*mp2*L8r*RN3p
         *RSp463e*RS346pe - 16./27.*pow(m55,2)*mp2*L8r*RN3p*RSp461e*
         RS346pe );

      M13p6x23L +=  + Ab(mp2,mu2) * (  - 16./27.*pow(m55,2)*mp2*L8r*
         RN1p*RS146pe*RSp463e + 16./27.*pow(m55,2)*mp2*L8r*RN1p*RSp461e
         *RS146pe + 16./27.*pow(m55,2)*mp2*L7r*RNep*RSp463e*RSe463p - 
         16./27.*pow(m55,2)*mp2*L7r*RNep*RSe461p*RSp463e - 16./27.*pow(
         m55,2)*mp2*L7r*RNep*RSp461e*RSe463p + 16./27.*pow(m55,2)*mp2*
         L7r*RNep*RSp461e*RSe461p + 16./27.*pow(m55,2)*mp2*L7r*RN3p*
         RSp463e*RS346pe - 16./27.*pow(m55,2)*mp2*L7r*RN3p*RSp461e*
         RS346pe - 16./27.*pow(m55,2)*mp2*L7r*RN1p*RS146pe*RSp463e + 16.
         /27.*pow(m55,2)*mp2*L7r*RN1p*RSp461e*RS146pe + 16./27.*pow(
         m55,2)*mp2*L6r*RNep*RSp463e*RSe463p - 16./27.*pow(m55,2)*mp2*
         L6r*RNep*RSe461p*RSp463e - 16./27.*pow(m55,2)*mp2*L6r*RNep*
         RSp461e*RSe463p + 16./27.*pow(m55,2)*mp2*L6r*RNep*RSp461e*
         RSe461p + 16./27.*pow(m55,2)*mp2*L6r*RN3p*RSp463e*RS346pe - 16.
         /27.*pow(m55,2)*mp2*L6r*RN3p*RSp461e*RS346pe - 16./27.*pow(
         m55,2)*mp2*L6r*RN1p*RS146pe*RSp463e + 16./27.*pow(m55,2)*mp2*
         L6r*RN1p*RSp461e*RS146pe );

      M13p6x23L +=  + Ab(mp2,mu2) * ( 8./27.*m44*mp2*L5r*RSp563e*
         RS356pe + 8./27.*m44*mp2*L5r*RSp563e*RSe563p + 8./27.*m44*mp2*
         L5r*pow(RSp563e,2) - 8./27.*m44*mp2*L5r*RS156pe*RSp563e - 8./
         27.*m44*mp2*L5r*RSe561p*RSp563e - 8./27.*m44*mp2*L5r*RSp561e*
         RS356pe - 8./27.*m44*mp2*L5r*RSp561e*RSe563p - 16./27.*m44*mp2
         *L5r*RSp561e*RSp563e + 8./27.*m44*mp2*L5r*RSp561e*RS156pe + 8./
         27.*m44*mp2*L5r*RSp561e*RSe561p + 8./27.*m44*mp2*L5r*pow(
         RSp561e,2) + 16./9.*m44*mp2*L4r*RSp45613e - 8./9.*m44*mp2*L4r*
         RSp45633e + 8./27.*m44*mp2*L4r*RSp563e*RS356pe + 8./27.*m44*
         mp2*L4r*RSp563e*RSe563p + 8./27.*m44*mp2*L4r*pow(RSp563e,2) + 
         8./27.*m44*mp2*L4r*RSp463e*RS346pe + 8./27.*m44*mp2*L4r*
         RSp463e*RSe463p + 8./27.*m44*mp2*L4r*pow(RSp463e,2) + 8./27.*
         m44*mp2*L4r*RSp453e*RS345pe + 8./27.*m44*mp2*L4r*RSp453e*
         RSe453p + 8./27.*m44*mp2*L4r*pow(RSp453e,2) - 8./9.*m44*mp2*
         L4r*RSp45611e - 8./27.*m44*mp2*L4r*RS145pe*RSp453e - 8./27.*
         m44*mp2*L4r*RS146pe*RSp463e );

      M13p6x23L +=  + Ab(mp2,mu2) * (  - 8./27.*m44*mp2*L4r*RS156pe*
         RSp563e - 8./27.*m44*mp2*L4r*RSe561p*RSp563e - 8./27.*m44*mp2*
         L4r*RSe461p*RSp463e - 8./27.*m44*mp2*L4r*RSe451p*RSp453e - 8./
         27.*m44*mp2*L4r*RSp561e*RS356pe - 8./27.*m44*mp2*L4r*RSp561e*
         RSe563p - 16./27.*m44*mp2*L4r*RSp561e*RSp563e + 8./27.*m44*mp2
         *L4r*RSp561e*RS156pe + 8./27.*m44*mp2*L4r*RSp561e*RSe561p + 8./
         27.*m44*mp2*L4r*pow(RSp561e,2) - 8./27.*m44*mp2*L4r*RSp461e*
         RS346pe - 8./27.*m44*mp2*L4r*RSp461e*RSe463p - 16./27.*m44*mp2
         *L4r*RSp461e*RSp463e + 8./27.*m44*mp2*L4r*RSp461e*RS146pe + 8./
         27.*m44*mp2*L4r*RSp461e*RSe461p + 8./27.*m44*mp2*L4r*pow(
         RSp461e,2) - 8./27.*m44*mp2*L4r*RSp451e*RS345pe - 8./27.*m44*
         mp2*L4r*RSp451e*RSe453p - 16./27.*m44*mp2*L4r*RSp451e*RSp453e
          + 8./27.*m44*mp2*L4r*RSp451e*RS145pe + 8./27.*m44*mp2*L4r*
         RSp451e*RSe451p + 8./27.*m44*mp2*L4r*pow(RSp451e,2) - 8./27.*
         m44*me2*mp2*L5r*RNep*RSp563e*RSe563p + 8./27.*m44*me2*mp2*L5r*
         RNep*RSe561p*RSp563e );

      M13p6x23L +=  + Ab(mp2,mu2) * ( 8./27.*m44*me2*mp2*L5r*RNep*
         RSp561e*RSe563p - 8./27.*m44*me2*mp2*L5r*RNep*RSp561e*RSe561p
          - 8./27.*m44*me2*mp2*L4r*RNep*RSp563e*RSe563p - 8./27.*m44*
         me2*mp2*L4r*RNep*RSp463e*RSe463p - 8./27.*m44*me2*mp2*L4r*RNep
         *RSp453e*RSe453p + 8./27.*m44*me2*mp2*L4r*RNep*RSe561p*RSp563e
          + 8./27.*m44*me2*mp2*L4r*RNep*RSe461p*RSp463e + 8./27.*m44*
         me2*mp2*L4r*RNep*RSe451p*RSp453e + 8./27.*m44*me2*mp2*L4r*RNep
         *RSp561e*RSe563p - 8./27.*m44*me2*mp2*L4r*RNep*RSp561e*RSe561p
          + 8./27.*m44*me2*mp2*L4r*RNep*RSp461e*RSe463p - 8./27.*m44*
         me2*mp2*L4r*RNep*RSp461e*RSe461p + 8./27.*m44*me2*mp2*L4r*RNep
         *RSp451e*RSe453p - 8./27.*m44*me2*mp2*L4r*RNep*RSp451e*RSe451p
          - 16./27.*m44*m66*L7r*RSp453e*RSp563e + 16./27.*m44*m66*L7r*
         RSp561e*RSp453e + 16./27.*m44*m66*L7r*RSp451e*RSp563e - 16./27.
         *m44*m66*L7r*RSp451e*RSp561e - 8./27.*m44*m66*L6r*pow(
         RSp563e,2) - 8./27.*m44*m66*L6r*pow(RSp453e,2) + 16./27.*m44*
         m66*L6r*RSp561e*RSp563e );

      M13p6x23L +=  + Ab(mp2,mu2) * (  - 8./27.*m44*m66*L6r*pow(
         RSp561e,2) + 16./27.*m44*m66*L6r*RSp451e*RSp453e - 8./27.*m44*
         m66*L6r*pow(RSp451e,2) + 16./27.*m44*m66*mp2*L7r*RNep*RSp563e*
         RSe453p + 16./27.*m44*m66*mp2*L7r*RNep*RSp453e*RSe563p - 16./
         27.*m44*m66*mp2*L7r*RNep*RSe561p*RSp453e - 16./27.*m44*m66*mp2
         *L7r*RNep*RSe451p*RSp563e - 16./27.*m44*m66*mp2*L7r*RNep*
         RSp561e*RSe453p + 16./27.*m44*m66*mp2*L7r*RNep*RSp561e*RSe451p
          - 16./27.*m44*m66*mp2*L7r*RNep*RSp451e*RSe563p + 16./27.*m44*
         m66*mp2*L7r*RNep*RSp451e*RSe561p + 16./27.*m44*m66*mp2*L7r*
         RN3p*RSp563e*RS345pe + 16./27.*m44*m66*mp2*L7r*RN3p*RSp453e*
         RS356pe - 16./27.*m44*m66*mp2*L7r*RN3p*RSp561e*RS345pe - 16./
         27.*m44*m66*mp2*L7r*RN3p*RSp451e*RS356pe - 16./27.*m44*m66*mp2
         *L7r*RN1p*RS145pe*RSp563e - 16./27.*m44*m66*mp2*L7r*RN1p*
         RS156pe*RSp453e + 16./27.*m44*m66*mp2*L7r*RN1p*RSp561e*RS145pe
          + 16./27.*m44*m66*mp2*L7r*RN1p*RSp451e*RS156pe + 16./27.*m44*
         m66*mp2*L6r*RNep*RSp563e*RSe563p );

      M13p6x23L +=  + Ab(mp2,mu2) * ( 16./27.*m44*m66*mp2*L6r*RNep*
         RSp453e*RSe453p - 16./27.*m44*m66*mp2*L6r*RNep*RSe561p*RSp563e
          - 16./27.*m44*m66*mp2*L6r*RNep*RSe451p*RSp453e - 16./27.*m44*
         m66*mp2*L6r*RNep*RSp561e*RSe563p + 16./27.*m44*m66*mp2*L6r*
         RNep*RSp561e*RSe561p - 16./27.*m44*m66*mp2*L6r*RNep*RSp451e*
         RSe453p + 16./27.*m44*m66*mp2*L6r*RNep*RSp451e*RSe451p + 16./
         27.*m44*m66*mp2*L6r*RN3p*RSp563e*RS356pe + 16./27.*m44*m66*mp2
         *L6r*RN3p*RSp453e*RS345pe - 16./27.*m44*m66*mp2*L6r*RN3p*
         RSp561e*RS356pe - 16./27.*m44*m66*mp2*L6r*RN3p*RSp451e*RS345pe
          - 16./27.*m44*m66*mp2*L6r*RN1p*RS145pe*RSp453e - 16./27.*m44*
         m66*mp2*L6r*RN1p*RS156pe*RSp563e + 16./27.*m44*m66*mp2*L6r*
         RN1p*RSp561e*RS156pe + 16./27.*m44*m66*mp2*L6r*RN1p*RSp451e*
         RS145pe - 16./27.*m44*m55*L7r*RSp463e*RSp563e + 16./27.*m44*
         m55*L7r*RSp561e*RSp463e + 16./27.*m44*m55*L7r*RSp461e*RSp563e
          - 16./27.*m44*m55*L7r*RSp461e*RSp561e - 8./27.*m44*m55*L6r*
         pow(RSp563e,2) );

      M13p6x23L +=  + Ab(mp2,mu2) * (  - 8./27.*m44*m55*L6r*pow(
         RSp463e,2) + 16./27.*m44*m55*L6r*RSp561e*RSp563e - 8./27.*m44*
         m55*L6r*pow(RSp561e,2) + 16./27.*m44*m55*L6r*RSp461e*RSp463e
          - 8./27.*m44*m55*L6r*pow(RSp461e,2) + 16./27.*m44*m55*mp2*L7r
         *RNep*RSp563e*RSe463p + 16./27.*m44*m55*mp2*L7r*RNep*RSp463e*
         RSe563p - 16./27.*m44*m55*mp2*L7r*RNep*RSe561p*RSp463e - 16./
         27.*m44*m55*mp2*L7r*RNep*RSe461p*RSp563e - 16./27.*m44*m55*mp2
         *L7r*RNep*RSp561e*RSe463p + 16./27.*m44*m55*mp2*L7r*RNep*
         RSp561e*RSe461p - 16./27.*m44*m55*mp2*L7r*RNep*RSp461e*RSe563p
          + 16./27.*m44*m55*mp2*L7r*RNep*RSp461e*RSe561p + 16./27.*m44*
         m55*mp2*L7r*RN3p*RSp563e*RS346pe + 16./27.*m44*m55*mp2*L7r*
         RN3p*RSp463e*RS356pe - 16./27.*m44*m55*mp2*L7r*RN3p*RSp561e*
         RS346pe - 16./27.*m44*m55*mp2*L7r*RN3p*RSp461e*RS356pe - 16./
         27.*m44*m55*mp2*L7r*RN1p*RS146pe*RSp563e - 16./27.*m44*m55*mp2
         *L7r*RN1p*RS156pe*RSp463e + 16./27.*m44*m55*mp2*L7r*RN1p*
         RSp561e*RS146pe );

      M13p6x23L +=  + Ab(mp2,mu2) * ( 16./27.*m44*m55*mp2*L7r*RN1p*
         RSp461e*RS156pe + 16./27.*m44*m55*mp2*L6r*RNep*RSp563e*RSe563p
          + 16./27.*m44*m55*mp2*L6r*RNep*RSp463e*RSe463p - 16./27.*m44*
         m55*mp2*L6r*RNep*RSe561p*RSp563e - 16./27.*m44*m55*mp2*L6r*
         RNep*RSe461p*RSp463e - 16./27.*m44*m55*mp2*L6r*RNep*RSp561e*
         RSe563p + 16./27.*m44*m55*mp2*L6r*RNep*RSp561e*RSe561p - 16./
         27.*m44*m55*mp2*L6r*RNep*RSp461e*RSe463p + 16./27.*m44*m55*mp2
         *L6r*RNep*RSp461e*RSe461p + 16./27.*m44*m55*mp2*L6r*RN3p*
         RSp563e*RS356pe + 16./27.*m44*m55*mp2*L6r*RN3p*RSp463e*RS346pe
          - 16./27.*m44*m55*mp2*L6r*RN3p*RSp561e*RS356pe - 16./27.*m44*
         m55*mp2*L6r*RN3p*RSp461e*RS346pe - 16./27.*m44*m55*mp2*L6r*
         RN1p*RS146pe*RSp463e - 16./27.*m44*m55*mp2*L6r*RN1p*RS156pe*
         RSp563e + 16./27.*m44*m55*mp2*L6r*RN1p*RSp561e*RS156pe + 16./
         27.*m44*m55*mp2*L6r*RN1p*RSp461e*RS146pe - 8./27.*pow(m44,2)*
         L8r*pow(RSp563e,2) + 16./27.*pow(m44,2)*L8r*RSp561e*RSp563e - 
         8./27.*pow(m44,2)*L8r*pow(RSp561e,2) );

      M13p6x23L +=  + Ab(mp2,mu2) * (  - 8./27.*pow(m44,2)*L7r*pow(
         RSp563e,2) + 16./27.*pow(m44,2)*L7r*RSp561e*RSp563e - 8./27.*
         pow(m44,2)*L7r*pow(RSp561e,2) - 8./27.*pow(m44,2)*L6r*pow(
         RSp563e,2) + 16./27.*pow(m44,2)*L6r*RSp561e*RSp563e - 8./27.*
         pow(m44,2)*L6r*pow(RSp561e,2) + 16./27.*pow(m44,2)*mp2*L8r*
         RNep*RSp563e*RSe563p - 16./27.*pow(m44,2)*mp2*L8r*RNep*RSe561p
         *RSp563e - 16./27.*pow(m44,2)*mp2*L8r*RNep*RSp561e*RSe563p + 
         16./27.*pow(m44,2)*mp2*L8r*RNep*RSp561e*RSe561p + 16./27.*pow(
         m44,2)*mp2*L8r*RN3p*RSp563e*RS356pe - 16./27.*pow(m44,2)*mp2*
         L8r*RN3p*RSp561e*RS356pe - 16./27.*pow(m44,2)*mp2*L8r*RN1p*
         RS156pe*RSp563e + 16./27.*pow(m44,2)*mp2*L8r*RN1p*RSp561e*
         RS156pe + 16./27.*pow(m44,2)*mp2*L7r*RNep*RSp563e*RSe563p - 16.
         /27.*pow(m44,2)*mp2*L7r*RNep*RSe561p*RSp563e - 16./27.*pow(
         m44,2)*mp2*L7r*RNep*RSp561e*RSe563p + 16./27.*pow(m44,2)*mp2*
         L7r*RNep*RSp561e*RSe561p + 16./27.*pow(m44,2)*mp2*L7r*RN3p*
         RSp563e*RS356pe );

      M13p6x23L +=  + Ab(mp2,mu2) * (  - 16./27.*pow(m44,2)*mp2*L7r*
         RN3p*RSp561e*RS356pe - 16./27.*pow(m44,2)*mp2*L7r*RN1p*RS156pe
         *RSp563e + 16./27.*pow(m44,2)*mp2*L7r*RN1p*RSp561e*RS156pe + 
         16./27.*pow(m44,2)*mp2*L6r*RNep*RSp563e*RSe563p - 16./27.*pow(
         m44,2)*mp2*L6r*RNep*RSe561p*RSp563e - 16./27.*pow(m44,2)*mp2*
         L6r*RNep*RSp561e*RSe563p + 16./27.*pow(m44,2)*mp2*L6r*RNep*
         RSp561e*RSe561p + 16./27.*pow(m44,2)*mp2*L6r*RN3p*RSp563e*
         RS356pe - 16./27.*pow(m44,2)*mp2*L6r*RN3p*RSp561e*RS356pe - 16.
         /27.*pow(m44,2)*mp2*L6r*RN1p*RS156pe*RSp563e + 16./27.*pow(
         m44,2)*mp2*L6r*RN1p*RSp561e*RS156pe + 8./9.*m33*mp2*L5r*
         RSp45613e + 4./9.*m33*mp2*L5r*RSp45611e + 4./3.*m33*mp2*L4r*
         RSp564e + 4./3.*m33*mp2*L4r*RSp465e + 4./3.*m33*mp2*L4r*
         RSp456e - 2./3.*m33*mp2*L3r*RSp45613e - 5./3.*m33*mp2*L3r*
         RSp45633e - 5./3.*m33*mp2*L3r*RSp45611e - 2./3.*m33*mp2*L2r*
         RSp564e - 2./3.*m33*mp2*L2r*RSp465e - 2./3.*m33*mp2*L2r*
         RSp456e );

      M13p6x23L +=  + Ab(mp2,mu2) * (  - 8./3.*m33*mp2*L1r*RSp564e - 8./
         3.*m33*mp2*L1r*RSp465e - 8./3.*m33*mp2*L1r*RSp456e - 8./3.*m33
         *mp2*L0r*RSp45613e - 2./3.*m33*mp2*L0r*RSp45633e - 2./3.*m33*
         mp2*L0r*RSp45611e - 32./9.*m33*m66*L7r*RSp453e - 16./9.*m33*
         m66*L7r*RSp451e - 16./9.*m33*m66*L6r*RSp45613e - 8./9.*m33*m66
         *L6r*RSp45633e - 8./3.*m33*m66*L6r*RSp456e - 4./27.*m33*m66*
         L5r*pow(RSp453e,2) - 8./27.*m33*m66*L5r*RSp451e*RSp453e + 8./9.
         *m33*m66*L4r*RSp45613e + 4./9.*m33*m66*L4r*RSp45633e - 4./27.*
         m33*m66*L4r*pow(RSp563e,2) - 4./27.*m33*m66*L4r*pow(RSp463e,2)
          - 4./27.*m33*m66*L4r*pow(RSp453e,2) - 8./27.*m33*m66*L4r*
         RSp561e*RSp563e - 8./27.*m33*m66*L4r*RSp461e*RSp463e - 8./27.*
         m33*m66*L4r*RSp451e*RSp453e + 4./3.*m33*m66*L4r*RSp456e - 16./
         9.*m33*m66*mp2*L7r*RN3p*RSp453e + 16./9.*m33*m66*mp2*L7r*RN3p*
         RSp451e + 16./9.*m33*m66*mp2*L6r*RN3p*RSp45613e - 16./9.*m33*
         m66*mp2*L6r*RN3p*RSp45633e + 8./27.*m33*m66*mp2*L5r*RNep*
         RSp453e*RSe453p );

      M13p6x23L +=  + Ab(mp2,mu2) * ( 8./27.*m33*m66*mp2*L5r*RNep*
         RSe451p*RSp453e + 8./27.*m33*m66*mp2*L5r*RNep*RSp451e*RSe453p
          + 16./27.*m33*m66*mp2*L5r*RN3p*RSp451e*RS345pe + 8./27.*m33*
         m66*mp2*L5r*RN1p*RS145pe*RSp453e + 8./27.*m33*m66*mp2*L4r*RNep
         *RSp563e*RSe563p + 8./27.*m33*m66*mp2*L4r*RNep*RSp463e*RSe463p
          + 8./27.*m33*m66*mp2*L4r*RNep*RSp453e*RSe453p + 8./27.*m33*
         m66*mp2*L4r*RNep*RSe561p*RSp563e + 8./27.*m33*m66*mp2*L4r*RNep
         *RSe461p*RSp463e + 8./27.*m33*m66*mp2*L4r*RNep*RSe451p*RSp453e
          + 8./27.*m33*m66*mp2*L4r*RNep*RSp561e*RSe563p + 8./27.*m33*
         m66*mp2*L4r*RNep*RSp461e*RSe463p + 8./27.*m33*m66*mp2*L4r*RNep
         *RSp451e*RSe453p - 16./9.*m33*m66*mp2*L4r*RN3p*RSp45613e + 16./
         27.*m33*m66*mp2*L4r*RN3p*RSp561e*RS356pe + 16./27.*m33*m66*mp2
         *L4r*RN3p*RSp461e*RS346pe + 16./27.*m33*m66*mp2*L4r*RN3p*
         RSp451e*RS345pe - 8./9.*m33*m66*mp2*L4r*RN1p*RSp45613e + 8./27.
         *m33*m66*mp2*L4r*RN1p*RS145pe*RSp453e + 8./27.*m33*m66*mp2*L4r
         *RN1p*RS146pe*RSp463e );

      M13p6x23L +=  + Ab(mp2,mu2) * ( 8./27.*m33*m66*mp2*L4r*RN1p*
         RS156pe*RSp563e - 16./27.*m33*pow(m66,2)*L8r*RNep*RSp453e*
         RSe453p - 16./27.*m33*pow(m66,2)*L8r*RNep*RSe451p*RSp453e - 16.
         /27.*m33*pow(m66,2)*L8r*RNep*RSp451e*RSe453p - 16./27.*m33*
         pow(m66,2)*L8r*RN3p*RSp453e*RS345pe - 16./27.*m33*pow(m66,2)*
         L8r*RN3p*RSp451e*RS345pe - 16./27.*m33*pow(m66,2)*L8r*RN1p*
         RS145pe*RSp453e - 16./27.*m33*pow(m66,2)*L7r*RNep*RSp453e*
         RSe453p - 16./27.*m33*pow(m66,2)*L7r*RNep*RSe451p*RSp453e - 16.
         /27.*m33*pow(m66,2)*L7r*RNep*RSp451e*RSe453p - 16./27.*m33*
         pow(m66,2)*L7r*RN3p*RSp453e*RS345pe - 16./27.*m33*pow(m66,2)*
         L7r*RN3p*RSp451e*RS345pe - 16./27.*m33*pow(m66,2)*L7r*RN1p*
         RS145pe*RSp453e - 16./27.*m33*pow(m66,2)*L6r*RNep*RSp453e*
         RSe453p - 16./27.*m33*pow(m66,2)*L6r*RNep*RSe451p*RSp453e - 16.
         /27.*m33*pow(m66,2)*L6r*RNep*RSp451e*RSe453p - 16./27.*m33*
         pow(m66,2)*L6r*RN3p*RSp453e*RS345pe - 16./27.*m33*pow(m66,2)*
         L6r*RN3p*RSp451e*RS345pe );

      M13p6x23L +=  + Ab(mp2,mu2) * (  - 16./27.*m33*pow(m66,2)*L6r*
         RN1p*RS145pe*RSp453e - 32./9.*m33*m55*L7r*RSp463e - 16./9.*m33
         *m55*L7r*RSp461e - 16./9.*m33*m55*L6r*RSp45613e - 8./9.*m33*
         m55*L6r*RSp45633e - 8./3.*m33*m55*L6r*RSp465e - 4./27.*m33*m55
         *L5r*pow(RSp463e,2) - 8./27.*m33*m55*L5r*RSp461e*RSp463e + 8./
         9.*m33*m55*L4r*RSp45613e + 4./9.*m33*m55*L4r*RSp45633e - 4./27.
         *m33*m55*L4r*pow(RSp563e,2) - 4./27.*m33*m55*L4r*pow(
         RSp463e,2) - 4./27.*m33*m55*L4r*pow(RSp453e,2) - 8./27.*m33*
         m55*L4r*RSp561e*RSp563e - 8./27.*m33*m55*L4r*RSp461e*RSp463e
          - 8./27.*m33*m55*L4r*RSp451e*RSp453e + 4./3.*m33*m55*L4r*
         RSp465e - 16./9.*m33*m55*mp2*L7r*RN3p*RSp463e + 16./9.*m33*m55
         *mp2*L7r*RN3p*RSp461e + 16./9.*m33*m55*mp2*L6r*RN3p*RSp45613e
          - 16./9.*m33*m55*mp2*L6r*RN3p*RSp45633e + 8./27.*m33*m55*mp2*
         L5r*RNep*RSp463e*RSe463p + 8./27.*m33*m55*mp2*L5r*RNep*RSe461p
         *RSp463e + 8./27.*m33*m55*mp2*L5r*RNep*RSp461e*RSe463p + 16./
         27.*m33*m55*mp2*L5r*RN3p*RSp461e*RS346pe );

      M13p6x23L +=  + Ab(mp2,mu2) * ( 8./27.*m33*m55*mp2*L5r*RN1p*
         RS146pe*RSp463e + 8./27.*m33*m55*mp2*L4r*RNep*RSp563e*RSe563p
          + 8./27.*m33*m55*mp2*L4r*RNep*RSp463e*RSe463p + 8./27.*m33*
         m55*mp2*L4r*RNep*RSp453e*RSe453p + 8./27.*m33*m55*mp2*L4r*RNep
         *RSe561p*RSp563e + 8./27.*m33*m55*mp2*L4r*RNep*RSe461p*RSp463e
          + 8./27.*m33*m55*mp2*L4r*RNep*RSe451p*RSp453e + 8./27.*m33*
         m55*mp2*L4r*RNep*RSp561e*RSe563p + 8./27.*m33*m55*mp2*L4r*RNep
         *RSp461e*RSe463p + 8./27.*m33*m55*mp2*L4r*RNep*RSp451e*RSe453p
          - 16./9.*m33*m55*mp2*L4r*RN3p*RSp45613e + 16./27.*m33*m55*mp2
         *L4r*RN3p*RSp561e*RS356pe + 16./27.*m33*m55*mp2*L4r*RN3p*
         RSp461e*RS346pe + 16./27.*m33*m55*mp2*L4r*RN3p*RSp451e*RS345pe
          - 8./9.*m33*m55*mp2*L4r*RN1p*RSp45613e + 8./27.*m33*m55*mp2*
         L4r*RN1p*RS145pe*RSp453e + 8./27.*m33*m55*mp2*L4r*RN1p*RS146pe
         *RSp463e + 8./27.*m33*m55*mp2*L4r*RN1p*RS156pe*RSp563e - 16./
         27.*m33*m55*m66*L7r*RNep*RSp463e*RSe453p - 16./27.*m33*m55*m66
         *L7r*RNep*RSp453e*RSe463p );

      M13p6x23L +=  + Ab(mp2,mu2) * (  - 16./27.*m33*m55*m66*L7r*RNep*
         RSe461p*RSp453e - 16./27.*m33*m55*m66*L7r*RNep*RSe451p*RSp463e
          - 16./27.*m33*m55*m66*L7r*RNep*RSp461e*RSe453p - 16./27.*m33*
         m55*m66*L7r*RNep*RSp451e*RSe463p - 16./27.*m33*m55*m66*L7r*
         RN3p*RSp463e*RS345pe - 16./27.*m33*m55*m66*L7r*RN3p*RSp453e*
         RS346pe - 16./27.*m33*m55*m66*L7r*RN3p*RSp461e*RS345pe - 16./
         27.*m33*m55*m66*L7r*RN3p*RSp451e*RS346pe - 16./27.*m33*m55*m66
         *L7r*RN1p*RS145pe*RSp463e - 16./27.*m33*m55*m66*L7r*RN1p*
         RS146pe*RSp453e - 16./27.*m33*m55*m66*L6r*RNep*RSp463e*RSe463p
          - 16./27.*m33*m55*m66*L6r*RNep*RSp453e*RSe453p - 16./27.*m33*
         m55*m66*L6r*RNep*RSe461p*RSp463e - 16./27.*m33*m55*m66*L6r*
         RNep*RSe451p*RSp453e - 16./27.*m33*m55*m66*L6r*RNep*RSp461e*
         RSe463p - 16./27.*m33*m55*m66*L6r*RNep*RSp451e*RSe453p - 16./
         27.*m33*m55*m66*L6r*RN3p*RSp463e*RS346pe - 16./27.*m33*m55*m66
         *L6r*RN3p*RSp453e*RS345pe - 16./27.*m33*m55*m66*L6r*RN3p*
         RSp461e*RS346pe );

      M13p6x23L +=  + Ab(mp2,mu2) * (  - 16./27.*m33*m55*m66*L6r*RN3p*
         RSp451e*RS345pe - 16./27.*m33*m55*m66*L6r*RN1p*RS145pe*RSp453e
          - 16./27.*m33*m55*m66*L6r*RN1p*RS146pe*RSp463e - 16./27.*m33*
         pow(m55,2)*L8r*RNep*RSp463e*RSe463p - 16./27.*m33*pow(m55,2)*
         L8r*RNep*RSe461p*RSp463e - 16./27.*m33*pow(m55,2)*L8r*RNep*
         RSp461e*RSe463p - 16./27.*m33*pow(m55,2)*L8r*RN3p*RSp463e*
         RS346pe - 16./27.*m33*pow(m55,2)*L8r*RN3p*RSp461e*RS346pe - 16.
         /27.*m33*pow(m55,2)*L8r*RN1p*RS146pe*RSp463e - 16./27.*m33*
         pow(m55,2)*L7r*RNep*RSp463e*RSe463p - 16./27.*m33*pow(m55,2)*
         L7r*RNep*RSe461p*RSp463e - 16./27.*m33*pow(m55,2)*L7r*RNep*
         RSp461e*RSe463p - 16./27.*m33*pow(m55,2)*L7r*RN3p*RSp463e*
         RS346pe - 16./27.*m33*pow(m55,2)*L7r*RN3p*RSp461e*RS346pe - 16.
         /27.*m33*pow(m55,2)*L7r*RN1p*RS146pe*RSp463e - 16./27.*m33*
         pow(m55,2)*L6r*RNep*RSp463e*RSe463p - 16./27.*m33*pow(m55,2)*
         L6r*RNep*RSe461p*RSp463e - 16./27.*m33*pow(m55,2)*L6r*RNep*
         RSp461e*RSe463p );

      M13p6x23L +=  + Ab(mp2,mu2) * (  - 16./27.*m33*pow(m55,2)*L6r*
         RN3p*RSp463e*RS346pe - 16./27.*m33*pow(m55,2)*L6r*RN3p*RSp461e
         *RS346pe - 16./27.*m33*pow(m55,2)*L6r*RN1p*RS146pe*RSp463e - 
         32./9.*m33*m44*L7r*RSp563e - 16./9.*m33*m44*L7r*RSp561e - 16./
         9.*m33*m44*L6r*RSp45613e - 8./9.*m33*m44*L6r*RSp45633e - 8./3.
         *m33*m44*L6r*RSp564e - 4./27.*m33*m44*L5r*pow(RSp563e,2) - 8./
         27.*m33*m44*L5r*RSp561e*RSp563e + 8./9.*m33*m44*L4r*RSp45613e
          + 4./9.*m33*m44*L4r*RSp45633e - 4./27.*m33*m44*L4r*pow(
         RSp563e,2) - 4./27.*m33*m44*L4r*pow(RSp463e,2) - 4./27.*m33*
         m44*L4r*pow(RSp453e,2) - 8./27.*m33*m44*L4r*RSp561e*RSp563e - 
         8./27.*m33*m44*L4r*RSp461e*RSp463e - 8./27.*m33*m44*L4r*
         RSp451e*RSp453e + 4./3.*m33*m44*L4r*RSp564e - 16./9.*m33*m44*
         mp2*L7r*RN3p*RSp563e + 16./9.*m33*m44*mp2*L7r*RN3p*RSp561e + 
         16./9.*m33*m44*mp2*L6r*RN3p*RSp45613e - 16./9.*m33*m44*mp2*L6r
         *RN3p*RSp45633e + 8./27.*m33*m44*mp2*L5r*RNep*RSp563e*RSe563p
          + 8./27.*m33*m44*mp2*L5r*RNep*RSe561p*RSp563e );

      M13p6x23L +=  + Ab(mp2,mu2) * ( 8./27.*m33*m44*mp2*L5r*RNep*
         RSp561e*RSe563p + 16./27.*m33*m44*mp2*L5r*RN3p*RSp561e*RS356pe
          + 8./27.*m33*m44*mp2*L5r*RN1p*RS156pe*RSp563e + 8./27.*m33*
         m44*mp2*L4r*RNep*RSp563e*RSe563p + 8./27.*m33*m44*mp2*L4r*RNep
         *RSp463e*RSe463p + 8./27.*m33*m44*mp2*L4r*RNep*RSp453e*RSe453p
          + 8./27.*m33*m44*mp2*L4r*RNep*RSe561p*RSp563e + 8./27.*m33*
         m44*mp2*L4r*RNep*RSe461p*RSp463e + 8./27.*m33*m44*mp2*L4r*RNep
         *RSe451p*RSp453e + 8./27.*m33*m44*mp2*L4r*RNep*RSp561e*RSe563p
          + 8./27.*m33*m44*mp2*L4r*RNep*RSp461e*RSe463p + 8./27.*m33*
         m44*mp2*L4r*RNep*RSp451e*RSe453p - 16./9.*m33*m44*mp2*L4r*RN3p
         *RSp45613e + 16./27.*m33*m44*mp2*L4r*RN3p*RSp561e*RS356pe + 16.
         /27.*m33*m44*mp2*L4r*RN3p*RSp461e*RS346pe + 16./27.*m33*m44*
         mp2*L4r*RN3p*RSp451e*RS345pe - 8./9.*m33*m44*mp2*L4r*RN1p*
         RSp45613e + 8./27.*m33*m44*mp2*L4r*RN1p*RS145pe*RSp453e + 8./
         27.*m33*m44*mp2*L4r*RN1p*RS146pe*RSp463e + 8./27.*m33*m44*mp2*
         L4r*RN1p*RS156pe*RSp563e );

      M13p6x23L +=  + Ab(mp2,mu2) * (  - 16./27.*m33*m44*m66*L7r*RNep*
         RSp563e*RSe453p - 16./27.*m33*m44*m66*L7r*RNep*RSp453e*RSe563p
          - 16./27.*m33*m44*m66*L7r*RNep*RSe561p*RSp453e - 16./27.*m33*
         m44*m66*L7r*RNep*RSe451p*RSp563e - 16./27.*m33*m44*m66*L7r*
         RNep*RSp561e*RSe453p - 16./27.*m33*m44*m66*L7r*RNep*RSp451e*
         RSe563p - 16./27.*m33*m44*m66*L7r*RN3p*RSp563e*RS345pe - 16./
         27.*m33*m44*m66*L7r*RN3p*RSp453e*RS356pe - 16./27.*m33*m44*m66
         *L7r*RN3p*RSp561e*RS345pe - 16./27.*m33*m44*m66*L7r*RN3p*
         RSp451e*RS356pe - 16./27.*m33*m44*m66*L7r*RN1p*RS145pe*RSp563e
          - 16./27.*m33*m44*m66*L7r*RN1p*RS156pe*RSp453e - 16./27.*m33*
         m44*m66*L6r*RNep*RSp563e*RSe563p - 16./27.*m33*m44*m66*L6r*
         RNep*RSp453e*RSe453p - 16./27.*m33*m44*m66*L6r*RNep*RSe561p*
         RSp563e - 16./27.*m33*m44*m66*L6r*RNep*RSe451p*RSp453e - 16./
         27.*m33*m44*m66*L6r*RNep*RSp561e*RSe563p - 16./27.*m33*m44*m66
         *L6r*RNep*RSp451e*RSe453p - 16./27.*m33*m44*m66*L6r*RN3p*
         RSp563e*RS356pe );

      M13p6x23L +=  + Ab(mp2,mu2) * (  - 16./27.*m33*m44*m66*L6r*RN3p*
         RSp453e*RS345pe - 16./27.*m33*m44*m66*L6r*RN3p*RSp561e*RS356pe
          - 16./27.*m33*m44*m66*L6r*RN3p*RSp451e*RS345pe - 16./27.*m33*
         m44*m66*L6r*RN1p*RS145pe*RSp453e - 16./27.*m33*m44*m66*L6r*
         RN1p*RS156pe*RSp563e - 16./27.*m33*m44*m55*L7r*RNep*RSp563e*
         RSe463p - 16./27.*m33*m44*m55*L7r*RNep*RSp463e*RSe563p - 16./
         27.*m33*m44*m55*L7r*RNep*RSe561p*RSp463e - 16./27.*m33*m44*m55
         *L7r*RNep*RSe461p*RSp563e - 16./27.*m33*m44*m55*L7r*RNep*
         RSp561e*RSe463p - 16./27.*m33*m44*m55*L7r*RNep*RSp461e*RSe563p
          - 16./27.*m33*m44*m55*L7r*RN3p*RSp563e*RS346pe - 16./27.*m33*
         m44*m55*L7r*RN3p*RSp463e*RS356pe - 16./27.*m33*m44*m55*L7r*
         RN3p*RSp561e*RS346pe - 16./27.*m33*m44*m55*L7r*RN3p*RSp461e*
         RS356pe - 16./27.*m33*m44*m55*L7r*RN1p*RS146pe*RSp563e - 16./
         27.*m33*m44*m55*L7r*RN1p*RS156pe*RSp463e - 16./27.*m33*m44*m55
         *L6r*RNep*RSp563e*RSe563p - 16./27.*m33*m44*m55*L6r*RNep*
         RSp463e*RSe463p );

      M13p6x23L +=  + Ab(mp2,mu2) * (  - 16./27.*m33*m44*m55*L6r*RNep*
         RSe561p*RSp563e - 16./27.*m33*m44*m55*L6r*RNep*RSe461p*RSp463e
          - 16./27.*m33*m44*m55*L6r*RNep*RSp561e*RSe563p - 16./27.*m33*
         m44*m55*L6r*RNep*RSp461e*RSe463p - 16./27.*m33*m44*m55*L6r*
         RN3p*RSp563e*RS356pe - 16./27.*m33*m44*m55*L6r*RN3p*RSp463e*
         RS346pe - 16./27.*m33*m44*m55*L6r*RN3p*RSp561e*RS356pe - 16./
         27.*m33*m44*m55*L6r*RN3p*RSp461e*RS346pe - 16./27.*m33*m44*m55
         *L6r*RN1p*RS146pe*RSp463e - 16./27.*m33*m44*m55*L6r*RN1p*
         RS156pe*RSp563e - 16./27.*m33*pow(m44,2)*L8r*RNep*RSp563e*
         RSe563p - 16./27.*m33*pow(m44,2)*L8r*RNep*RSe561p*RSp563e - 16.
         /27.*m33*pow(m44,2)*L8r*RNep*RSp561e*RSe563p - 16./27.*m33*
         pow(m44,2)*L8r*RN3p*RSp563e*RS356pe - 16./27.*m33*pow(m44,2)*
         L8r*RN3p*RSp561e*RS356pe - 16./27.*m33*pow(m44,2)*L8r*RN1p*
         RS156pe*RSp563e - 16./27.*m33*pow(m44,2)*L7r*RNep*RSp563e*
         RSe563p - 16./27.*m33*pow(m44,2)*L7r*RNep*RSe561p*RSp563e - 16.
         /27.*m33*pow(m44,2)*L7r*RNep*RSp561e*RSe563p );

      M13p6x23L +=  + Ab(mp2,mu2) * (  - 16./27.*m33*pow(m44,2)*L7r*
         RN3p*RSp563e*RS356pe - 16./27.*m33*pow(m44,2)*L7r*RN3p*RSp561e
         *RS356pe - 16./27.*m33*pow(m44,2)*L7r*RN1p*RS156pe*RSp563e - 
         16./27.*m33*pow(m44,2)*L6r*RNep*RSp563e*RSe563p - 16./27.*m33*
         pow(m44,2)*L6r*RNep*RSe561p*RSp563e - 16./27.*m33*pow(m44,2)*
         L6r*RNep*RSp561e*RSe563p - 16./27.*m33*pow(m44,2)*L6r*RN3p*
         RSp563e*RS356pe - 16./27.*m33*pow(m44,2)*L6r*RN3p*RSp561e*
         RS356pe - 16./27.*m33*pow(m44,2)*L6r*RN1p*RS156pe*RSp563e - 16.
         /9.*pow(m33,2)*L8r*RSp45613e - 32./9.*pow(m33,2)*L8r*RSp45633e
          + 4./9.*pow(m33,2)*L5r*RSp45613e + 8./9.*pow(m33,2)*L5r*
         RSp45633e + 16./9.*pow(m33,2)*mp2*L8r*RN3p*RSp45613e - 16./9.*
         pow(m33,2)*mp2*L8r*RN3p*RSp45633e - 16./9.*pow(m33,2)*mp2*L5r*
         RN3p*RSp45613e + 16./9.*pow(m33,2)*m66*L7r*RN3p*RSp453e + 16./
         9.*pow(m33,2)*m66*L7r*RN3p*RSp451e + 16./9.*pow(m33,2)*m66*L6r
         *RN3p*RSp45613e + 16./9.*pow(m33,2)*m66*L6r*RN3p*RSp45633e + 
         16./9.*pow(m33,2)*m55*L7r*RN3p*RSp463e );

      M13p6x23L +=  + Ab(mp2,mu2) * ( 16./9.*pow(m33,2)*m55*L7r*RN3p*
         RSp461e + 16./9.*pow(m33,2)*m55*L6r*RN3p*RSp45613e + 16./9.*
         pow(m33,2)*m55*L6r*RN3p*RSp45633e + 16./9.*pow(m33,2)*m44*L7r*
         RN3p*RSp563e + 16./9.*pow(m33,2)*m44*L7r*RN3p*RSp561e + 16./9.
         *pow(m33,2)*m44*L6r*RN3p*RSp45613e + 16./9.*pow(m33,2)*m44*L6r
         *RN3p*RSp45633e + 16./9.*pow(m33,3)*L8r*RN3p*RSp45613e + 16./9.
         *pow(m33,3)*L8r*RN3p*RSp45633e + 8./9.*m11*mp2*L5r*RSp45613e
          + 4./9.*m11*mp2*L5r*RSp45633e + 4./3.*m11*mp2*L4r*RSp564e + 4.
         /3.*m11*mp2*L4r*RSp465e + 4./3.*m11*mp2*L4r*RSp456e - 2./3.*
         m11*mp2*L3r*RSp45613e - 5./3.*m11*mp2*L3r*RSp45633e - 5./3.*
         m11*mp2*L3r*RSp45611e - 2./3.*m11*mp2*L2r*RSp564e - 2./3.*m11*
         mp2*L2r*RSp465e - 2./3.*m11*mp2*L2r*RSp456e - 8./3.*m11*mp2*
         L1r*RSp564e - 8./3.*m11*mp2*L1r*RSp465e - 8./3.*m11*mp2*L1r*
         RSp456e - 8./3.*m11*mp2*L0r*RSp45613e - 2./3.*m11*mp2*L0r*
         RSp45633e - 2./3.*m11*mp2*L0r*RSp45611e - 16./9.*m11*m66*L7r*
         RSp453e );

      M13p6x23L +=  + Ab(mp2,mu2) * (  - 32./9.*m11*m66*L7r*RSp451e - 
         16./9.*m11*m66*L6r*RSp45613e - 8./9.*m11*m66*L6r*RSp45611e - 8.
         /3.*m11*m66*L6r*RSp456e - 8./27.*m11*m66*L5r*RSp451e*RSp453e
          - 4./27.*m11*m66*L5r*pow(RSp451e,2) + 8./9.*m11*m66*L4r*
         RSp45613e + 4./9.*m11*m66*L4r*RSp45611e - 8./27.*m11*m66*L4r*
         RSp561e*RSp563e - 4./27.*m11*m66*L4r*pow(RSp561e,2) - 8./27.*
         m11*m66*L4r*RSp461e*RSp463e - 4./27.*m11*m66*L4r*pow(
         RSp461e,2) - 8./27.*m11*m66*L4r*RSp451e*RSp453e - 4./27.*m11*
         m66*L4r*pow(RSp451e,2) + 4./3.*m11*m66*L4r*RSp456e + 16./9.*
         m11*m66*mp2*L7r*RN1p*RSp453e - 16./9.*m11*m66*mp2*L7r*RN1p*
         RSp451e + 16./9.*m11*m66*mp2*L6r*RN1p*RSp45613e - 16./9.*m11*
         m66*mp2*L6r*RN1p*RSp45611e + 8./27.*m11*m66*mp2*L5r*RNep*
         RSe451p*RSp453e + 8./27.*m11*m66*mp2*L5r*RNep*RSp451e*RSe453p
          + 8./27.*m11*m66*mp2*L5r*RNep*RSp451e*RSe451p + 8./27.*m11*
         m66*mp2*L5r*RN3p*RSp451e*RS345pe + 16./27.*m11*m66*mp2*L5r*
         RN1p*RS145pe*RSp453e );

      M13p6x23L +=  + Ab(mp2,mu2) * ( 8./27.*m11*m66*mp2*L4r*RNep*
         RSe561p*RSp563e + 8./27.*m11*m66*mp2*L4r*RNep*RSe461p*RSp463e
          + 8./27.*m11*m66*mp2*L4r*RNep*RSe451p*RSp453e + 8./27.*m11*
         m66*mp2*L4r*RNep*RSp561e*RSe563p + 8./27.*m11*m66*mp2*L4r*RNep
         *RSp561e*RSe561p + 8./27.*m11*m66*mp2*L4r*RNep*RSp461e*RSe463p
          + 8./27.*m11*m66*mp2*L4r*RNep*RSp461e*RSe461p + 8./27.*m11*
         m66*mp2*L4r*RNep*RSp451e*RSe453p + 8./27.*m11*m66*mp2*L4r*RNep
         *RSp451e*RSe451p - 8./9.*m11*m66*mp2*L4r*RN3p*RSp45613e + 8./
         27.*m11*m66*mp2*L4r*RN3p*RSp561e*RS356pe + 8./27.*m11*m66*mp2*
         L4r*RN3p*RSp461e*RS346pe + 8./27.*m11*m66*mp2*L4r*RN3p*RSp451e
         *RS345pe - 16./9.*m11*m66*mp2*L4r*RN1p*RSp45613e + 16./27.*m11
         *m66*mp2*L4r*RN1p*RS145pe*RSp453e + 16./27.*m11*m66*mp2*L4r*
         RN1p*RS146pe*RSp463e + 16./27.*m11*m66*mp2*L4r*RN1p*RS156pe*
         RSp563e - 16./27.*m11*pow(m66,2)*L8r*RNep*RSe451p*RSp453e - 16.
         /27.*m11*pow(m66,2)*L8r*RNep*RSp451e*RSe453p - 16./27.*m11*
         pow(m66,2)*L8r*RNep*RSp451e*RSe451p );

      M13p6x23L +=  + Ab(mp2,mu2) * (  - 16./27.*m11*pow(m66,2)*L8r*
         RN3p*RSp451e*RS345pe - 16./27.*m11*pow(m66,2)*L8r*RN1p*RS145pe
         *RSp453e - 16./27.*m11*pow(m66,2)*L8r*RN1p*RSp451e*RS145pe - 
         16./27.*m11*pow(m66,2)*L7r*RNep*RSe451p*RSp453e - 16./27.*m11*
         pow(m66,2)*L7r*RNep*RSp451e*RSe453p - 16./27.*m11*pow(m66,2)*
         L7r*RNep*RSp451e*RSe451p - 16./27.*m11*pow(m66,2)*L7r*RN3p*
         RSp451e*RS345pe - 16./27.*m11*pow(m66,2)*L7r*RN1p*RS145pe*
         RSp453e - 16./27.*m11*pow(m66,2)*L7r*RN1p*RSp451e*RS145pe - 16.
         /27.*m11*pow(m66,2)*L6r*RNep*RSe451p*RSp453e - 16./27.*m11*
         pow(m66,2)*L6r*RNep*RSp451e*RSe453p - 16./27.*m11*pow(m66,2)*
         L6r*RNep*RSp451e*RSe451p - 16./27.*m11*pow(m66,2)*L6r*RN3p*
         RSp451e*RS345pe - 16./27.*m11*pow(m66,2)*L6r*RN1p*RS145pe*
         RSp453e - 16./27.*m11*pow(m66,2)*L6r*RN1p*RSp451e*RS145pe - 16.
         /9.*m11*m55*L7r*RSp463e - 32./9.*m11*m55*L7r*RSp461e - 16./9.*
         m11*m55*L6r*RSp45613e - 8./9.*m11*m55*L6r*RSp45611e - 8./3.*
         m11*m55*L6r*RSp465e );

      M13p6x23L +=  + Ab(mp2,mu2) * (  - 8./27.*m11*m55*L5r*RSp461e*
         RSp463e - 4./27.*m11*m55*L5r*pow(RSp461e,2) + 8./9.*m11*m55*
         L4r*RSp45613e + 4./9.*m11*m55*L4r*RSp45611e - 8./27.*m11*m55*
         L4r*RSp561e*RSp563e - 4./27.*m11*m55*L4r*pow(RSp561e,2) - 8./
         27.*m11*m55*L4r*RSp461e*RSp463e - 4./27.*m11*m55*L4r*pow(
         RSp461e,2) - 8./27.*m11*m55*L4r*RSp451e*RSp453e - 4./27.*m11*
         m55*L4r*pow(RSp451e,2) + 4./3.*m11*m55*L4r*RSp465e + 16./9.*
         m11*m55*mp2*L7r*RN1p*RSp463e - 16./9.*m11*m55*mp2*L7r*RN1p*
         RSp461e + 16./9.*m11*m55*mp2*L6r*RN1p*RSp45613e - 16./9.*m11*
         m55*mp2*L6r*RN1p*RSp45611e + 8./27.*m11*m55*mp2*L5r*RNep*
         RSe461p*RSp463e + 8./27.*m11*m55*mp2*L5r*RNep*RSp461e*RSe463p
          + 8./27.*m11*m55*mp2*L5r*RNep*RSp461e*RSe461p + 8./27.*m11*
         m55*mp2*L5r*RN3p*RSp461e*RS346pe + 16./27.*m11*m55*mp2*L5r*
         RN1p*RS146pe*RSp463e + 8./27.*m11*m55*mp2*L4r*RNep*RSe561p*
         RSp563e + 8./27.*m11*m55*mp2*L4r*RNep*RSe461p*RSp463e + 8./27.
         *m11*m55*mp2*L4r*RNep*RSe451p*RSp453e );

      M13p6x23L +=  + Ab(mp2,mu2) * ( 8./27.*m11*m55*mp2*L4r*RNep*
         RSp561e*RSe563p + 8./27.*m11*m55*mp2*L4r*RNep*RSp561e*RSe561p
          + 8./27.*m11*m55*mp2*L4r*RNep*RSp461e*RSe463p + 8./27.*m11*
         m55*mp2*L4r*RNep*RSp461e*RSe461p + 8./27.*m11*m55*mp2*L4r*RNep
         *RSp451e*RSe453p + 8./27.*m11*m55*mp2*L4r*RNep*RSp451e*RSe451p
          - 8./9.*m11*m55*mp2*L4r*RN3p*RSp45613e + 8./27.*m11*m55*mp2*
         L4r*RN3p*RSp561e*RS356pe + 8./27.*m11*m55*mp2*L4r*RN3p*RSp461e
         *RS346pe + 8./27.*m11*m55*mp2*L4r*RN3p*RSp451e*RS345pe - 16./9.
         *m11*m55*mp2*L4r*RN1p*RSp45613e + 16./27.*m11*m55*mp2*L4r*RN1p
         *RS145pe*RSp453e + 16./27.*m11*m55*mp2*L4r*RN1p*RS146pe*
         RSp463e + 16./27.*m11*m55*mp2*L4r*RN1p*RS156pe*RSp563e - 16./
         27.*m11*m55*m66*L7r*RNep*RSe461p*RSp453e - 16./27.*m11*m55*m66
         *L7r*RNep*RSe451p*RSp463e - 16./27.*m11*m55*m66*L7r*RNep*
         RSp461e*RSe453p - 16./27.*m11*m55*m66*L7r*RNep*RSp461e*RSe451p
          - 16./27.*m11*m55*m66*L7r*RNep*RSp451e*RSe463p - 16./27.*m11*
         m55*m66*L7r*RNep*RSp451e*RSe461p );

      M13p6x23L +=  + Ab(mp2,mu2) * (  - 16./27.*m11*m55*m66*L7r*RN3p*
         RSp461e*RS345pe - 16./27.*m11*m55*m66*L7r*RN3p*RSp451e*RS346pe
          - 16./27.*m11*m55*m66*L7r*RN1p*RS145pe*RSp463e - 16./27.*m11*
         m55*m66*L7r*RN1p*RS146pe*RSp453e - 16./27.*m11*m55*m66*L7r*
         RN1p*RSp461e*RS145pe - 16./27.*m11*m55*m66*L7r*RN1p*RSp451e*
         RS146pe - 16./27.*m11*m55*m66*L6r*RNep*RSe461p*RSp463e - 16./
         27.*m11*m55*m66*L6r*RNep*RSe451p*RSp453e - 16./27.*m11*m55*m66
         *L6r*RNep*RSp461e*RSe463p - 16./27.*m11*m55*m66*L6r*RNep*
         RSp461e*RSe461p - 16./27.*m11*m55*m66*L6r*RNep*RSp451e*RSe453p
          - 16./27.*m11*m55*m66*L6r*RNep*RSp451e*RSe451p - 16./27.*m11*
         m55*m66*L6r*RN3p*RSp461e*RS346pe - 16./27.*m11*m55*m66*L6r*
         RN3p*RSp451e*RS345pe - 16./27.*m11*m55*m66*L6r*RN1p*RS145pe*
         RSp453e - 16./27.*m11*m55*m66*L6r*RN1p*RS146pe*RSp463e - 16./
         27.*m11*m55*m66*L6r*RN1p*RSp461e*RS146pe - 16./27.*m11*m55*m66
         *L6r*RN1p*RSp451e*RS145pe - 16./27.*m11*pow(m55,2)*L8r*RNep*
         RSe461p*RSp463e );

      M13p6x23L +=  + Ab(mp2,mu2) * (  - 16./27.*m11*pow(m55,2)*L8r*
         RNep*RSp461e*RSe463p - 16./27.*m11*pow(m55,2)*L8r*RNep*RSp461e
         *RSe461p - 16./27.*m11*pow(m55,2)*L8r*RN3p*RSp461e*RS346pe - 
         16./27.*m11*pow(m55,2)*L8r*RN1p*RS146pe*RSp463e - 16./27.*m11*
         pow(m55,2)*L8r*RN1p*RSp461e*RS146pe - 16./27.*m11*pow(m55,2)*
         L7r*RNep*RSe461p*RSp463e - 16./27.*m11*pow(m55,2)*L7r*RNep*
         RSp461e*RSe463p - 16./27.*m11*pow(m55,2)*L7r*RNep*RSp461e*
         RSe461p - 16./27.*m11*pow(m55,2)*L7r*RN3p*RSp461e*RS346pe - 16.
         /27.*m11*pow(m55,2)*L7r*RN1p*RS146pe*RSp463e - 16./27.*m11*
         pow(m55,2)*L7r*RN1p*RSp461e*RS146pe - 16./27.*m11*pow(m55,2)*
         L6r*RNep*RSe461p*RSp463e - 16./27.*m11*pow(m55,2)*L6r*RNep*
         RSp461e*RSe463p - 16./27.*m11*pow(m55,2)*L6r*RNep*RSp461e*
         RSe461p - 16./27.*m11*pow(m55,2)*L6r*RN3p*RSp461e*RS346pe - 16.
         /27.*m11*pow(m55,2)*L6r*RN1p*RS146pe*RSp463e - 16./27.*m11*
         pow(m55,2)*L6r*RN1p*RSp461e*RS146pe - 16./9.*m11*m44*L7r*
         RSp563e );

      M13p6x23L +=  + Ab(mp2,mu2) * (  - 32./9.*m11*m44*L7r*RSp561e - 
         16./9.*m11*m44*L6r*RSp45613e - 8./9.*m11*m44*L6r*RSp45611e - 8.
         /3.*m11*m44*L6r*RSp564e - 8./27.*m11*m44*L5r*RSp561e*RSp563e
          - 4./27.*m11*m44*L5r*pow(RSp561e,2) + 8./9.*m11*m44*L4r*
         RSp45613e + 4./9.*m11*m44*L4r*RSp45611e - 8./27.*m11*m44*L4r*
         RSp561e*RSp563e - 4./27.*m11*m44*L4r*pow(RSp561e,2) - 8./27.*
         m11*m44*L4r*RSp461e*RSp463e - 4./27.*m11*m44*L4r*pow(
         RSp461e,2) - 8./27.*m11*m44*L4r*RSp451e*RSp453e - 4./27.*m11*
         m44*L4r*pow(RSp451e,2) + 4./3.*m11*m44*L4r*RSp564e + 16./9.*
         m11*m44*mp2*L7r*RN1p*RSp563e - 16./9.*m11*m44*mp2*L7r*RN1p*
         RSp561e + 16./9.*m11*m44*mp2*L6r*RN1p*RSp45613e - 16./9.*m11*
         m44*mp2*L6r*RN1p*RSp45611e + 8./27.*m11*m44*mp2*L5r*RNep*
         RSe561p*RSp563e + 8./27.*m11*m44*mp2*L5r*RNep*RSp561e*RSe563p
          + 8./27.*m11*m44*mp2*L5r*RNep*RSp561e*RSe561p + 8./27.*m11*
         m44*mp2*L5r*RN3p*RSp561e*RS356pe + 16./27.*m11*m44*mp2*L5r*
         RN1p*RS156pe*RSp563e );

      M13p6x23L +=  + Ab(mp2,mu2) * ( 8./27.*m11*m44*mp2*L4r*RNep*
         RSe561p*RSp563e + 8./27.*m11*m44*mp2*L4r*RNep*RSe461p*RSp463e
          + 8./27.*m11*m44*mp2*L4r*RNep*RSe451p*RSp453e + 8./27.*m11*
         m44*mp2*L4r*RNep*RSp561e*RSe563p + 8./27.*m11*m44*mp2*L4r*RNep
         *RSp561e*RSe561p + 8./27.*m11*m44*mp2*L4r*RNep*RSp461e*RSe463p
          + 8./27.*m11*m44*mp2*L4r*RNep*RSp461e*RSe461p + 8./27.*m11*
         m44*mp2*L4r*RNep*RSp451e*RSe453p + 8./27.*m11*m44*mp2*L4r*RNep
         *RSp451e*RSe451p - 8./9.*m11*m44*mp2*L4r*RN3p*RSp45613e + 8./
         27.*m11*m44*mp2*L4r*RN3p*RSp561e*RS356pe + 8./27.*m11*m44*mp2*
         L4r*RN3p*RSp461e*RS346pe + 8./27.*m11*m44*mp2*L4r*RN3p*RSp451e
         *RS345pe - 16./9.*m11*m44*mp2*L4r*RN1p*RSp45613e + 16./27.*m11
         *m44*mp2*L4r*RN1p*RS145pe*RSp453e + 16./27.*m11*m44*mp2*L4r*
         RN1p*RS146pe*RSp463e + 16./27.*m11*m44*mp2*L4r*RN1p*RS156pe*
         RSp563e - 16./27.*m11*m44*m66*L7r*RNep*RSe561p*RSp453e - 16./
         27.*m11*m44*m66*L7r*RNep*RSe451p*RSp563e - 16./27.*m11*m44*m66
         *L7r*RNep*RSp561e*RSe453p );

      M13p6x23L +=  + Ab(mp2,mu2) * (  - 16./27.*m11*m44*m66*L7r*RNep*
         RSp561e*RSe451p - 16./27.*m11*m44*m66*L7r*RNep*RSp451e*RSe563p
          - 16./27.*m11*m44*m66*L7r*RNep*RSp451e*RSe561p - 16./27.*m11*
         m44*m66*L7r*RN3p*RSp561e*RS345pe - 16./27.*m11*m44*m66*L7r*
         RN3p*RSp451e*RS356pe - 16./27.*m11*m44*m66*L7r*RN1p*RS145pe*
         RSp563e - 16./27.*m11*m44*m66*L7r*RN1p*RS156pe*RSp453e - 16./
         27.*m11*m44*m66*L7r*RN1p*RSp561e*RS145pe - 16./27.*m11*m44*m66
         *L7r*RN1p*RSp451e*RS156pe - 16./27.*m11*m44*m66*L6r*RNep*
         RSe561p*RSp563e - 16./27.*m11*m44*m66*L6r*RNep*RSe451p*RSp453e
          - 16./27.*m11*m44*m66*L6r*RNep*RSp561e*RSe563p - 16./27.*m11*
         m44*m66*L6r*RNep*RSp561e*RSe561p - 16./27.*m11*m44*m66*L6r*
         RNep*RSp451e*RSe453p - 16./27.*m11*m44*m66*L6r*RNep*RSp451e*
         RSe451p - 16./27.*m11*m44*m66*L6r*RN3p*RSp561e*RS356pe - 16./
         27.*m11*m44*m66*L6r*RN3p*RSp451e*RS345pe - 16./27.*m11*m44*m66
         *L6r*RN1p*RS145pe*RSp453e - 16./27.*m11*m44*m66*L6r*RN1p*
         RS156pe*RSp563e );

      M13p6x23L +=  + Ab(mp2,mu2) * (  - 16./27.*m11*m44*m66*L6r*RN1p*
         RSp561e*RS156pe - 16./27.*m11*m44*m66*L6r*RN1p*RSp451e*RS145pe
          - 16./27.*m11*m44*m55*L7r*RNep*RSe561p*RSp463e - 16./27.*m11*
         m44*m55*L7r*RNep*RSe461p*RSp563e - 16./27.*m11*m44*m55*L7r*
         RNep*RSp561e*RSe463p - 16./27.*m11*m44*m55*L7r*RNep*RSp561e*
         RSe461p - 16./27.*m11*m44*m55*L7r*RNep*RSp461e*RSe563p - 16./
         27.*m11*m44*m55*L7r*RNep*RSp461e*RSe561p - 16./27.*m11*m44*m55
         *L7r*RN3p*RSp561e*RS346pe - 16./27.*m11*m44*m55*L7r*RN3p*
         RSp461e*RS356pe - 16./27.*m11*m44*m55*L7r*RN1p*RS146pe*RSp563e
          - 16./27.*m11*m44*m55*L7r*RN1p*RS156pe*RSp463e - 16./27.*m11*
         m44*m55*L7r*RN1p*RSp561e*RS146pe - 16./27.*m11*m44*m55*L7r*
         RN1p*RSp461e*RS156pe - 16./27.*m11*m44*m55*L6r*RNep*RSe561p*
         RSp563e - 16./27.*m11*m44*m55*L6r*RNep*RSe461p*RSp463e - 16./
         27.*m11*m44*m55*L6r*RNep*RSp561e*RSe563p - 16./27.*m11*m44*m55
         *L6r*RNep*RSp561e*RSe561p - 16./27.*m11*m44*m55*L6r*RNep*
         RSp461e*RSe463p );

      M13p6x23L +=  + Ab(mp2,mu2) * (  - 16./27.*m11*m44*m55*L6r*RNep*
         RSp461e*RSe461p - 16./27.*m11*m44*m55*L6r*RN3p*RSp561e*RS356pe
          - 16./27.*m11*m44*m55*L6r*RN3p*RSp461e*RS346pe - 16./27.*m11*
         m44*m55*L6r*RN1p*RS146pe*RSp463e - 16./27.*m11*m44*m55*L6r*
         RN1p*RS156pe*RSp563e - 16./27.*m11*m44*m55*L6r*RN1p*RSp561e*
         RS156pe - 16./27.*m11*m44*m55*L6r*RN1p*RSp461e*RS146pe - 16./
         27.*m11*pow(m44,2)*L8r*RNep*RSe561p*RSp563e - 16./27.*m11*pow(
         m44,2)*L8r*RNep*RSp561e*RSe563p - 16./27.*m11*pow(m44,2)*L8r*
         RNep*RSp561e*RSe561p - 16./27.*m11*pow(m44,2)*L8r*RN3p*RSp561e
         *RS356pe - 16./27.*m11*pow(m44,2)*L8r*RN1p*RS156pe*RSp563e - 
         16./27.*m11*pow(m44,2)*L8r*RN1p*RSp561e*RS156pe - 16./27.*m11*
         pow(m44,2)*L7r*RNep*RSe561p*RSp563e - 16./27.*m11*pow(m44,2)*
         L7r*RNep*RSp561e*RSe563p - 16./27.*m11*pow(m44,2)*L7r*RNep*
         RSp561e*RSe561p - 16./27.*m11*pow(m44,2)*L7r*RN3p*RSp561e*
         RS356pe - 16./27.*m11*pow(m44,2)*L7r*RN1p*RS156pe*RSp563e - 16.
         /27.*m11*pow(m44,2)*L7r*RN1p*RSp561e*RS156pe );

      M13p6x23L +=  + Ab(mp2,mu2) * (  - 16./27.*m11*pow(m44,2)*L6r*
         RNep*RSe561p*RSp563e - 16./27.*m11*pow(m44,2)*L6r*RNep*RSp561e
         *RSe563p - 16./27.*m11*pow(m44,2)*L6r*RNep*RSp561e*RSe561p - 
         16./27.*m11*pow(m44,2)*L6r*RN3p*RSp561e*RS356pe - 16./27.*m11*
         pow(m44,2)*L6r*RN1p*RS156pe*RSp563e - 16./27.*m11*pow(m44,2)*
         L6r*RN1p*RSp561e*RS156pe - 16./3.*m11*m33*L8r*RSp45613e - 8./3.
         *m11*m33*L8r*RSp45633e - 8./3.*m11*m33*L8r*RSp45611e + 8./9.*
         m11*m33*L5r*RSp45613e + 8./9.*m11*m33*L5r*RSp45633e + 8./9.*
         m11*m33*L5r*RSp45611e - 8./9.*m11*m33*mp2*L5r*RN3p*RSp45613e
          - 8./9.*m11*m33*mp2*L5r*RN1p*RSp45613e + 16./9.*m11*m33*m66*
         L7r*RN3p*RSp451e + 16./9.*m11*m33*m66*L7r*RN1p*RSp453e + 16./9.
         *m11*m33*m66*L6r*RN3p*RSp45613e + 16./9.*m11*m33*m66*L6r*RN1p*
         RSp45613e + 16./9.*m11*m33*m55*L7r*RN3p*RSp461e + 16./9.*m11*
         m33*m55*L7r*RN1p*RSp463e + 16./9.*m11*m33*m55*L6r*RN3p*
         RSp45613e + 16./9.*m11*m33*m55*L6r*RN1p*RSp45613e + 16./9.*m11
         *m33*m44*L7r*RN3p*RSp561e );

      M13p6x23L +=  + Ab(mp2,mu2) * ( 16./9.*m11*m33*m44*L7r*RN1p*
         RSp563e + 16./9.*m11*m33*m44*L6r*RN3p*RSp45613e + 16./9.*m11*
         m33*m44*L6r*RN1p*RSp45613e + 16./9.*m11*pow(m33,2)*L8r*RN3p*
         RSp45613e - 16./9.*pow(m11,2)*L8r*RSp45613e - 32./9.*pow(
         m11,2)*L8r*RSp45611e + 4./9.*pow(m11,2)*L5r*RSp45613e + 8./9.*
         pow(m11,2)*L5r*RSp45611e + 16./9.*pow(m11,2)*mp2*L8r*RN1p*
         RSp45613e - 16./9.*pow(m11,2)*mp2*L8r*RN1p*RSp45611e - 16./9.*
         pow(m11,2)*mp2*L5r*RN1p*RSp45613e + 16./9.*pow(m11,2)*m66*L7r*
         RN1p*RSp453e + 16./9.*pow(m11,2)*m66*L7r*RN1p*RSp451e + 16./9.
         *pow(m11,2)*m66*L6r*RN1p*RSp45613e + 16./9.*pow(m11,2)*m66*L6r
         *RN1p*RSp45611e + 16./9.*pow(m11,2)*m55*L7r*RN1p*RSp463e + 16./
         9.*pow(m11,2)*m55*L7r*RN1p*RSp461e + 16./9.*pow(m11,2)*m55*L6r
         *RN1p*RSp45613e + 16./9.*pow(m11,2)*m55*L6r*RN1p*RSp45611e + 
         16./9.*pow(m11,2)*m44*L7r*RN1p*RSp563e + 16./9.*pow(m11,2)*m44
         *L7r*RN1p*RSp561e + 16./9.*pow(m11,2)*m44*L6r*RN1p*RSp45613e
          + 16./9.*pow(m11,2)*m44*L6r*RN1p*RSp45611e );

      M13p6x23L +=  + Ab(mp2,mu2) * ( 16./9.*pow(m11,2)*m33*L8r*RN1p*
         RSp45613e + 16./9.*pow(m11,3)*L8r*RN1p*RSp45613e + 16./9.*pow(
         m11,3)*L8r*RN1p*RSp45611e );

      M13p6x23L +=  + Bb(m11,mu2) * ( 8./9.*m11*m66*L6r*RD1456ep - 4./9.
         *m11*m66*L4r*RD1456ep + 8./9.*m11*m55*L6r*RD1456ep - 4./9.*m11
         *m55*L4r*RD1456ep + 8./9.*m11*m44*L6r*RD1456ep - 4./9.*m11*m44
         *L4r*RD1456ep - 8./3.*m11*m33*L8r*RD1456ep + 4./3.*m11*m33*L5r
         *RD1456ep - 5./3.*m11*m33*L3r*RD1456ep - 2./3.*m11*m33*L0r*
         RD1456ep - 16./9.*m11*m33*m66*L6r*RS14563pe + 8./9.*m11*m33*
         m66*L4r*RS14563pe - 16./9.*m11*m33*m55*L6r*RS14563pe + 8./9.*
         m11*m33*m55*L4r*RS14563pe - 16./9.*m11*m33*m44*L6r*RS14563pe
          + 8./9.*m11*m33*m44*L4r*RS14563pe - 16./9.*pow(m11,2)*L8r*
         RD1456ep + 8./9.*pow(m11,2)*L5r*RD1456ep - 5./3.*pow(m11,2)*
         L3r*RD1456ep - 2./3.*pow(m11,2)*L0r*RD1456ep - 32./9.*pow(
         m11,2)*m66*L6r*RS14563pe + 16./9.*pow(m11,2)*m66*L4r*RS14563pe
          - 32./9.*pow(m11,2)*m55*L6r*RS14563pe + 16./9.*pow(m11,2)*m55
         *L4r*RS14563pe - 32./9.*pow(m11,2)*m44*L6r*RS14563pe + 16./9.*
         pow(m11,2)*m44*L4r*RS14563pe - 16./9.*pow(m11,2)*m33*L8r*
         RS14563pe );

      M13p6x23L +=  + Bb(m11,mu2) * ( 8./9.*pow(m11,2)*m33*L5r*
         RS14563pe - 32./9.*pow(m11,3)*L8r*RS14563pe + 16./9.*pow(
         m11,3)*L5r*RS14563pe );

      M13p6x23L +=  + Bb(m33,mu2) * ( 8./9.*m33*m66*L6r*RD3456ep - 4./9.
         *m33*m66*L4r*RD3456ep + 8./9.*m33*m55*L6r*RD3456ep - 4./9.*m33
         *m55*L4r*RD3456ep + 8./9.*m33*m44*L6r*RD3456ep - 4./9.*m33*m44
         *L4r*RD3456ep - 16./9.*pow(m33,2)*L8r*RD3456ep + 8./9.*pow(
         m33,2)*L5r*RD3456ep - 5./3.*pow(m33,2)*L3r*RD3456ep - 2./3.*
         pow(m33,2)*L0r*RD3456ep - 32./9.*pow(m33,2)*m66*L6r*RS34561pe
          + 16./9.*pow(m33,2)*m66*L4r*RS34561pe - 32./9.*pow(m33,2)*m55
         *L6r*RS34561pe + 16./9.*pow(m33,2)*m55*L4r*RS34561pe - 32./9.*
         pow(m33,2)*m44*L6r*RS34561pe + 16./9.*pow(m33,2)*m44*L4r*
         RS34561pe - 32./9.*pow(m33,3)*L8r*RS34561pe + 16./9.*pow(
         m33,3)*L5r*RS34561pe - 8./3.*m11*m33*L8r*RD3456ep + 4./3.*m11*
         m33*L5r*RD3456ep - 5./3.*m11*m33*L3r*RD3456ep - 2./3.*m11*m33*
         L0r*RD3456ep - 16./9.*m11*m33*m66*L6r*RS34561pe + 8./9.*m11*
         m33*m66*L4r*RS34561pe - 16./9.*m11*m33*m55*L6r*RS34561pe + 8./
         9.*m11*m33*m55*L4r*RS34561pe - 16./9.*m11*m33*m44*L6r*
         RS34561pe );

      M13p6x23L +=  + Bb(m33,mu2) * ( 8./9.*m11*m33*m44*L4r*RS34561pe
          - 16./9.*m11*pow(m33,2)*L8r*RS34561pe + 8./9.*m11*pow(m33,2)*
         L5r*RS34561pe );

      M13p6x23L +=  + Bb(me2,mu2) * ( 4./27.*m66*pow(me2,2)*L5r*pow(
         RSe453p,2) - 8./27.*m66*pow(me2,2)*L5r*RSe451p*RSe453p + 4./27.
         *m66*pow(me2,2)*L5r*pow(RSe451p,2) + 4./27.*m66*pow(me2,2)*L4r
         *pow(RSe563p,2) + 4./27.*m66*pow(me2,2)*L4r*pow(RSe463p,2) + 4.
         /27.*m66*pow(me2,2)*L4r*pow(RSe453p,2) - 8./27.*m66*pow(me2,2)
         *L4r*RSe561p*RSe563p + 4./27.*m66*pow(me2,2)*L4r*pow(
         RSe561p,2) - 8./27.*m66*pow(me2,2)*L4r*RSe461p*RSe463p + 4./27.
         *m66*pow(me2,2)*L4r*pow(RSe461p,2) - 8./27.*m66*pow(me2,2)*L4r
         *RSe451p*RSe453p + 4./27.*m66*pow(me2,2)*L4r*pow(RSe451p,2) - 
         8./27.*pow(m66,2)*me2*L8r*pow(RSe453p,2) + 16./27.*pow(m66,2)*
         me2*L8r*RSe451p*RSe453p - 8./27.*pow(m66,2)*me2*L8r*pow(
         RSe451p,2) - 8./27.*pow(m66,2)*me2*L7r*pow(RSe453p,2) + 16./27.
         *pow(m66,2)*me2*L7r*RSe451p*RSe453p - 8./27.*pow(m66,2)*me2*
         L7r*pow(RSe451p,2) - 8./27.*pow(m66,2)*me2*L6r*pow(RSe453p,2)
          + 16./27.*pow(m66,2)*me2*L6r*RSe451p*RSe453p - 8./27.*pow(
         m66,2)*me2*L6r*pow(RSe451p,2) );

      M13p6x23L +=  + Bb(me2,mu2) * ( 4./27.*m55*pow(me2,2)*L5r*pow(
         RSe463p,2) - 8./27.*m55*pow(me2,2)*L5r*RSe461p*RSe463p + 4./27.
         *m55*pow(me2,2)*L5r*pow(RSe461p,2) + 4./27.*m55*pow(me2,2)*L4r
         *pow(RSe563p,2) + 4./27.*m55*pow(me2,2)*L4r*pow(RSe463p,2) + 4.
         /27.*m55*pow(me2,2)*L4r*pow(RSe453p,2) - 8./27.*m55*pow(me2,2)
         *L4r*RSe561p*RSe563p + 4./27.*m55*pow(me2,2)*L4r*pow(
         RSe561p,2) - 8./27.*m55*pow(me2,2)*L4r*RSe461p*RSe463p + 4./27.
         *m55*pow(me2,2)*L4r*pow(RSe461p,2) - 8./27.*m55*pow(me2,2)*L4r
         *RSe451p*RSe453p + 4./27.*m55*pow(me2,2)*L4r*pow(RSe451p,2) - 
         16./27.*m55*m66*me2*L7r*RSe453p*RSe463p + 16./27.*m55*m66*me2*
         L7r*RSe461p*RSe453p + 16./27.*m55*m66*me2*L7r*RSe451p*RSe463p
          - 16./27.*m55*m66*me2*L7r*RSe451p*RSe461p - 8./27.*m55*m66*
         me2*L6r*pow(RSe463p,2) - 8./27.*m55*m66*me2*L6r*pow(RSe453p,2)
          + 16./27.*m55*m66*me2*L6r*RSe461p*RSe463p - 8./27.*m55*m66*
         me2*L6r*pow(RSe461p,2) + 16./27.*m55*m66*me2*L6r*RSe451p*
         RSe453p );

      M13p6x23L +=  + Bb(me2,mu2) * (  - 8./27.*m55*m66*me2*L6r*pow(
         RSe451p,2) - 8./27.*pow(m55,2)*me2*L8r*pow(RSe463p,2) + 16./27.
         *pow(m55,2)*me2*L8r*RSe461p*RSe463p - 8./27.*pow(m55,2)*me2*
         L8r*pow(RSe461p,2) - 8./27.*pow(m55,2)*me2*L7r*pow(RSe463p,2)
          + 16./27.*pow(m55,2)*me2*L7r*RSe461p*RSe463p - 8./27.*pow(
         m55,2)*me2*L7r*pow(RSe461p,2) - 8./27.*pow(m55,2)*me2*L6r*pow(
         RSe463p,2) + 16./27.*pow(m55,2)*me2*L6r*RSe461p*RSe463p - 8./
         27.*pow(m55,2)*me2*L6r*pow(RSe461p,2) + 4./27.*m44*pow(me2,2)*
         L5r*pow(RSe563p,2) - 8./27.*m44*pow(me2,2)*L5r*RSe561p*RSe563p
          + 4./27.*m44*pow(me2,2)*L5r*pow(RSe561p,2) + 4./27.*m44*pow(
         me2,2)*L4r*pow(RSe563p,2) + 4./27.*m44*pow(me2,2)*L4r*pow(
         RSe463p,2) + 4./27.*m44*pow(me2,2)*L4r*pow(RSe453p,2) - 8./27.
         *m44*pow(me2,2)*L4r*RSe561p*RSe563p + 4./27.*m44*pow(me2,2)*
         L4r*pow(RSe561p,2) - 8./27.*m44*pow(me2,2)*L4r*RSe461p*RSe463p
          + 4./27.*m44*pow(me2,2)*L4r*pow(RSe461p,2) - 8./27.*m44*pow(
         me2,2)*L4r*RSe451p*RSe453p );

      M13p6x23L +=  + Bb(me2,mu2) * ( 4./27.*m44*pow(me2,2)*L4r*pow(
         RSe451p,2) - 16./27.*m44*m66*me2*L7r*RSe453p*RSe563p + 16./27.
         *m44*m66*me2*L7r*RSe561p*RSe453p + 16./27.*m44*m66*me2*L7r*
         RSe451p*RSe563p - 16./27.*m44*m66*me2*L7r*RSe451p*RSe561p - 8./
         27.*m44*m66*me2*L6r*pow(RSe563p,2) - 8./27.*m44*m66*me2*L6r*
         pow(RSe453p,2) + 16./27.*m44*m66*me2*L6r*RSe561p*RSe563p - 8./
         27.*m44*m66*me2*L6r*pow(RSe561p,2) + 16./27.*m44*m66*me2*L6r*
         RSe451p*RSe453p - 8./27.*m44*m66*me2*L6r*pow(RSe451p,2) - 16./
         27.*m44*m55*me2*L7r*RSe463p*RSe563p + 16./27.*m44*m55*me2*L7r*
         RSe561p*RSe463p + 16./27.*m44*m55*me2*L7r*RSe461p*RSe563p - 16.
         /27.*m44*m55*me2*L7r*RSe461p*RSe561p - 8./27.*m44*m55*me2*L6r*
         pow(RSe563p,2) - 8./27.*m44*m55*me2*L6r*pow(RSe463p,2) + 16./
         27.*m44*m55*me2*L6r*RSe561p*RSe563p - 8./27.*m44*m55*me2*L6r*
         pow(RSe561p,2) + 16./27.*m44*m55*me2*L6r*RSe461p*RSe463p - 8./
         27.*m44*m55*me2*L6r*pow(RSe461p,2) - 8./27.*pow(m44,2)*me2*L8r
         *pow(RSe563p,2) );

      M13p6x23L +=  + Bb(me2,mu2) * ( 16./27.*pow(m44,2)*me2*L8r*
         RSe561p*RSe563p - 8./27.*pow(m44,2)*me2*L8r*pow(RSe561p,2) - 8.
         /27.*pow(m44,2)*me2*L7r*pow(RSe563p,2) + 16./27.*pow(m44,2)*
         me2*L7r*RSe561p*RSe563p - 8./27.*pow(m44,2)*me2*L7r*pow(
         RSe561p,2) - 8./27.*pow(m44,2)*me2*L6r*pow(RSe563p,2) + 16./27.
         *pow(m44,2)*me2*L6r*RSe561p*RSe563p - 8./27.*pow(m44,2)*me2*
         L6r*pow(RSe561p,2) - 4./27.*m33*m66*me2*L5r*pow(RSe453p,2) - 8.
         /27.*m33*m66*me2*L5r*RSe451p*RSe453p - 4./27.*m33*m66*me2*L4r*
         pow(RSe563p,2) - 4./27.*m33*m66*me2*L4r*pow(RSe463p,2) - 4./27.
         *m33*m66*me2*L4r*pow(RSe453p,2) - 8./27.*m33*m66*me2*L4r*
         RSe561p*RSe563p - 8./27.*m33*m66*me2*L4r*RSe461p*RSe463p - 8./
         27.*m33*m66*me2*L4r*RSe451p*RSe453p + 8./27.*m33*pow(m66,2)*
         L8r*pow(RSe453p,2) + 16./27.*m33*pow(m66,2)*L8r*RSe451p*
         RSe453p + 8./27.*m33*pow(m66,2)*L7r*pow(RSe453p,2) + 16./27.*
         m33*pow(m66,2)*L7r*RSe451p*RSe453p + 8./27.*m33*pow(m66,2)*L6r
         *pow(RSe453p,2) );

      M13p6x23L +=  + Bb(me2,mu2) * ( 16./27.*m33*pow(m66,2)*L6r*
         RSe451p*RSe453p - 4./27.*m33*m55*me2*L5r*pow(RSe463p,2) - 8./
         27.*m33*m55*me2*L5r*RSe461p*RSe463p - 4./27.*m33*m55*me2*L4r*
         pow(RSe563p,2) - 4./27.*m33*m55*me2*L4r*pow(RSe463p,2) - 4./27.
         *m33*m55*me2*L4r*pow(RSe453p,2) - 8./27.*m33*m55*me2*L4r*
         RSe561p*RSe563p - 8./27.*m33*m55*me2*L4r*RSe461p*RSe463p - 8./
         27.*m33*m55*me2*L4r*RSe451p*RSe453p + 16./27.*m33*m55*m66*L7r*
         RSe453p*RSe463p + 16./27.*m33*m55*m66*L7r*RSe461p*RSe453p + 16.
         /27.*m33*m55*m66*L7r*RSe451p*RSe463p + 8./27.*m33*m55*m66*L6r*
         pow(RSe463p,2) + 8./27.*m33*m55*m66*L6r*pow(RSe453p,2) + 16./
         27.*m33*m55*m66*L6r*RSe461p*RSe463p + 16./27.*m33*m55*m66*L6r*
         RSe451p*RSe453p + 8./27.*m33*pow(m55,2)*L8r*pow(RSe463p,2) + 
         16./27.*m33*pow(m55,2)*L8r*RSe461p*RSe463p + 8./27.*m33*pow(
         m55,2)*L7r*pow(RSe463p,2) + 16./27.*m33*pow(m55,2)*L7r*RSe461p
         *RSe463p + 8./27.*m33*pow(m55,2)*L6r*pow(RSe463p,2) + 16./27.*
         m33*pow(m55,2)*L6r*RSe461p*RSe463p );

      M13p6x23L +=  + Bb(me2,mu2) * (  - 4./27.*m33*m44*me2*L5r*pow(
         RSe563p,2) - 8./27.*m33*m44*me2*L5r*RSe561p*RSe563p - 4./27.*
         m33*m44*me2*L4r*pow(RSe563p,2) - 4./27.*m33*m44*me2*L4r*pow(
         RSe463p,2) - 4./27.*m33*m44*me2*L4r*pow(RSe453p,2) - 8./27.*
         m33*m44*me2*L4r*RSe561p*RSe563p - 8./27.*m33*m44*me2*L4r*
         RSe461p*RSe463p - 8./27.*m33*m44*me2*L4r*RSe451p*RSe453p + 16./
         27.*m33*m44*m66*L7r*RSe453p*RSe563p + 16./27.*m33*m44*m66*L7r*
         RSe561p*RSe453p + 16./27.*m33*m44*m66*L7r*RSe451p*RSe563p + 8./
         27.*m33*m44*m66*L6r*pow(RSe563p,2) + 8./27.*m33*m44*m66*L6r*
         pow(RSe453p,2) + 16./27.*m33*m44*m66*L6r*RSe561p*RSe563p + 16./
         27.*m33*m44*m66*L6r*RSe451p*RSe453p + 16./27.*m33*m44*m55*L7r*
         RSe463p*RSe563p + 16./27.*m33*m44*m55*L7r*RSe561p*RSe463p + 16.
         /27.*m33*m44*m55*L7r*RSe461p*RSe563p + 8./27.*m33*m44*m55*L6r*
         pow(RSe563p,2) + 8./27.*m33*m44*m55*L6r*pow(RSe463p,2) + 16./
         27.*m33*m44*m55*L6r*RSe561p*RSe563p + 16./27.*m33*m44*m55*L6r*
         RSe461p*RSe463p );

      M13p6x23L +=  + Bb(me2,mu2) * ( 8./27.*m33*pow(m44,2)*L8r*pow(
         RSe563p,2) + 16./27.*m33*pow(m44,2)*L8r*RSe561p*RSe563p + 8./
         27.*m33*pow(m44,2)*L7r*pow(RSe563p,2) + 16./27.*m33*pow(m44,2)
         *L7r*RSe561p*RSe563p + 8./27.*m33*pow(m44,2)*L6r*pow(
         RSe563p,2) + 16./27.*m33*pow(m44,2)*L6r*RSe561p*RSe563p - 8./
         27.*m11*m66*me2*L5r*RSe451p*RSe453p - 4./27.*m11*m66*me2*L5r*
         pow(RSe451p,2) - 8./27.*m11*m66*me2*L4r*RSe561p*RSe563p - 4./
         27.*m11*m66*me2*L4r*pow(RSe561p,2) - 8./27.*m11*m66*me2*L4r*
         RSe461p*RSe463p - 4./27.*m11*m66*me2*L4r*pow(RSe461p,2) - 8./
         27.*m11*m66*me2*L4r*RSe451p*RSe453p - 4./27.*m11*m66*me2*L4r*
         pow(RSe451p,2) + 16./27.*m11*pow(m66,2)*L8r*RSe451p*RSe453p + 
         8./27.*m11*pow(m66,2)*L8r*pow(RSe451p,2) + 16./27.*m11*pow(
         m66,2)*L7r*RSe451p*RSe453p + 8./27.*m11*pow(m66,2)*L7r*pow(
         RSe451p,2) + 16./27.*m11*pow(m66,2)*L6r*RSe451p*RSe453p + 8./
         27.*m11*pow(m66,2)*L6r*pow(RSe451p,2) - 8./27.*m11*m55*me2*L5r
         *RSe461p*RSe463p );

      M13p6x23L +=  + Bb(me2,mu2) * (  - 4./27.*m11*m55*me2*L5r*pow(
         RSe461p,2) - 8./27.*m11*m55*me2*L4r*RSe561p*RSe563p - 4./27.*
         m11*m55*me2*L4r*pow(RSe561p,2) - 8./27.*m11*m55*me2*L4r*
         RSe461p*RSe463p - 4./27.*m11*m55*me2*L4r*pow(RSe461p,2) - 8./
         27.*m11*m55*me2*L4r*RSe451p*RSe453p - 4./27.*m11*m55*me2*L4r*
         pow(RSe451p,2) + 16./27.*m11*m55*m66*L7r*RSe461p*RSe453p + 16./
         27.*m11*m55*m66*L7r*RSe451p*RSe463p + 16./27.*m11*m55*m66*L7r*
         RSe451p*RSe461p + 16./27.*m11*m55*m66*L6r*RSe461p*RSe463p + 8./
         27.*m11*m55*m66*L6r*pow(RSe461p,2) + 16./27.*m11*m55*m66*L6r*
         RSe451p*RSe453p + 8./27.*m11*m55*m66*L6r*pow(RSe451p,2) + 16./
         27.*m11*pow(m55,2)*L8r*RSe461p*RSe463p + 8./27.*m11*pow(m55,2)
         *L8r*pow(RSe461p,2) + 16./27.*m11*pow(m55,2)*L7r*RSe461p*
         RSe463p + 8./27.*m11*pow(m55,2)*L7r*pow(RSe461p,2) + 16./27.*
         m11*pow(m55,2)*L6r*RSe461p*RSe463p + 8./27.*m11*pow(m55,2)*L6r
         *pow(RSe461p,2) - 8./27.*m11*m44*me2*L5r*RSe561p*RSe563p - 4./
         27.*m11*m44*me2*L5r*pow(RSe561p,2) );

      M13p6x23L +=  + Bb(me2,mu2) * (  - 8./27.*m11*m44*me2*L4r*RSe561p
         *RSe563p - 4./27.*m11*m44*me2*L4r*pow(RSe561p,2) - 8./27.*m11*
         m44*me2*L4r*RSe461p*RSe463p - 4./27.*m11*m44*me2*L4r*pow(
         RSe461p,2) - 8./27.*m11*m44*me2*L4r*RSe451p*RSe453p - 4./27.*
         m11*m44*me2*L4r*pow(RSe451p,2) + 16./27.*m11*m44*m66*L7r*
         RSe561p*RSe453p + 16./27.*m11*m44*m66*L7r*RSe451p*RSe563p + 16.
         /27.*m11*m44*m66*L7r*RSe451p*RSe561p + 16./27.*m11*m44*m66*L6r
         *RSe561p*RSe563p + 8./27.*m11*m44*m66*L6r*pow(RSe561p,2) + 16./
         27.*m11*m44*m66*L6r*RSe451p*RSe453p + 8./27.*m11*m44*m66*L6r*
         pow(RSe451p,2) + 16./27.*m11*m44*m55*L7r*RSe561p*RSe463p + 16./
         27.*m11*m44*m55*L7r*RSe461p*RSe563p + 16./27.*m11*m44*m55*L7r*
         RSe461p*RSe561p + 16./27.*m11*m44*m55*L6r*RSe561p*RSe563p + 8./
         27.*m11*m44*m55*L6r*pow(RSe561p,2) + 16./27.*m11*m44*m55*L6r*
         RSe461p*RSe463p + 8./27.*m11*m44*m55*L6r*pow(RSe461p,2) + 16./
         27.*m11*pow(m44,2)*L8r*RSe561p*RSe563p + 8./27.*m11*pow(m44,2)
         *L8r*pow(RSe561p,2) );

      M13p6x23L +=  + Bb(me2,mu2) * ( 16./27.*m11*pow(m44,2)*L7r*
         RSe561p*RSe563p + 8./27.*m11*pow(m44,2)*L7r*pow(RSe561p,2) + 
         16./27.*m11*pow(m44,2)*L6r*RSe561p*RSe563p + 8./27.*m11*pow(
         m44,2)*L6r*pow(RSe561p,2) );

      M13p6x23L +=  + Bb(mp2,mu2) * ( 4./27.*m66*pow(mp2,2)*L5r*pow(
         RSp453e,2) - 8./27.*m66*pow(mp2,2)*L5r*RSp451e*RSp453e + 4./27.
         *m66*pow(mp2,2)*L5r*pow(RSp451e,2) + 4./27.*m66*pow(mp2,2)*L4r
         *pow(RSp563e,2) + 4./27.*m66*pow(mp2,2)*L4r*pow(RSp463e,2) + 4.
         /27.*m66*pow(mp2,2)*L4r*pow(RSp453e,2) - 8./27.*m66*pow(mp2,2)
         *L4r*RSp561e*RSp563e + 4./27.*m66*pow(mp2,2)*L4r*pow(
         RSp561e,2) - 8./27.*m66*pow(mp2,2)*L4r*RSp461e*RSp463e + 4./27.
         *m66*pow(mp2,2)*L4r*pow(RSp461e,2) - 8./27.*m66*pow(mp2,2)*L4r
         *RSp451e*RSp453e + 4./27.*m66*pow(mp2,2)*L4r*pow(RSp451e,2) - 
         8./27.*pow(m66,2)*mp2*L8r*pow(RSp453e,2) + 16./27.*pow(m66,2)*
         mp2*L8r*RSp451e*RSp453e - 8./27.*pow(m66,2)*mp2*L8r*pow(
         RSp451e,2) - 8./27.*pow(m66,2)*mp2*L7r*pow(RSp453e,2) + 16./27.
         *pow(m66,2)*mp2*L7r*RSp451e*RSp453e - 8./27.*pow(m66,2)*mp2*
         L7r*pow(RSp451e,2) - 8./27.*pow(m66,2)*mp2*L6r*pow(RSp453e,2)
          + 16./27.*pow(m66,2)*mp2*L6r*RSp451e*RSp453e - 8./27.*pow(
         m66,2)*mp2*L6r*pow(RSp451e,2) );

      M13p6x23L +=  + Bb(mp2,mu2) * ( 4./27.*m55*pow(mp2,2)*L5r*pow(
         RSp463e,2) - 8./27.*m55*pow(mp2,2)*L5r*RSp461e*RSp463e + 4./27.
         *m55*pow(mp2,2)*L5r*pow(RSp461e,2) + 4./27.*m55*pow(mp2,2)*L4r
         *pow(RSp563e,2) + 4./27.*m55*pow(mp2,2)*L4r*pow(RSp463e,2) + 4.
         /27.*m55*pow(mp2,2)*L4r*pow(RSp453e,2) - 8./27.*m55*pow(mp2,2)
         *L4r*RSp561e*RSp563e + 4./27.*m55*pow(mp2,2)*L4r*pow(
         RSp561e,2) - 8./27.*m55*pow(mp2,2)*L4r*RSp461e*RSp463e + 4./27.
         *m55*pow(mp2,2)*L4r*pow(RSp461e,2) - 8./27.*m55*pow(mp2,2)*L4r
         *RSp451e*RSp453e + 4./27.*m55*pow(mp2,2)*L4r*pow(RSp451e,2) - 
         16./27.*m55*m66*mp2*L7r*RSp453e*RSp463e + 16./27.*m55*m66*mp2*
         L7r*RSp461e*RSp453e + 16./27.*m55*m66*mp2*L7r*RSp451e*RSp463e
          - 16./27.*m55*m66*mp2*L7r*RSp451e*RSp461e - 8./27.*m55*m66*
         mp2*L6r*pow(RSp463e,2) - 8./27.*m55*m66*mp2*L6r*pow(RSp453e,2)
          + 16./27.*m55*m66*mp2*L6r*RSp461e*RSp463e - 8./27.*m55*m66*
         mp2*L6r*pow(RSp461e,2) + 16./27.*m55*m66*mp2*L6r*RSp451e*
         RSp453e );

      M13p6x23L +=  + Bb(mp2,mu2) * (  - 8./27.*m55*m66*mp2*L6r*pow(
         RSp451e,2) - 8./27.*pow(m55,2)*mp2*L8r*pow(RSp463e,2) + 16./27.
         *pow(m55,2)*mp2*L8r*RSp461e*RSp463e - 8./27.*pow(m55,2)*mp2*
         L8r*pow(RSp461e,2) - 8./27.*pow(m55,2)*mp2*L7r*pow(RSp463e,2)
          + 16./27.*pow(m55,2)*mp2*L7r*RSp461e*RSp463e - 8./27.*pow(
         m55,2)*mp2*L7r*pow(RSp461e,2) - 8./27.*pow(m55,2)*mp2*L6r*pow(
         RSp463e,2) + 16./27.*pow(m55,2)*mp2*L6r*RSp461e*RSp463e - 8./
         27.*pow(m55,2)*mp2*L6r*pow(RSp461e,2) + 4./27.*m44*pow(mp2,2)*
         L5r*pow(RSp563e,2) - 8./27.*m44*pow(mp2,2)*L5r*RSp561e*RSp563e
          + 4./27.*m44*pow(mp2,2)*L5r*pow(RSp561e,2) + 4./27.*m44*pow(
         mp2,2)*L4r*pow(RSp563e,2) + 4./27.*m44*pow(mp2,2)*L4r*pow(
         RSp463e,2) + 4./27.*m44*pow(mp2,2)*L4r*pow(RSp453e,2) - 8./27.
         *m44*pow(mp2,2)*L4r*RSp561e*RSp563e + 4./27.*m44*pow(mp2,2)*
         L4r*pow(RSp561e,2) - 8./27.*m44*pow(mp2,2)*L4r*RSp461e*RSp463e
          + 4./27.*m44*pow(mp2,2)*L4r*pow(RSp461e,2) - 8./27.*m44*pow(
         mp2,2)*L4r*RSp451e*RSp453e );

      M13p6x23L +=  + Bb(mp2,mu2) * ( 4./27.*m44*pow(mp2,2)*L4r*pow(
         RSp451e,2) - 16./27.*m44*m66*mp2*L7r*RSp453e*RSp563e + 16./27.
         *m44*m66*mp2*L7r*RSp561e*RSp453e + 16./27.*m44*m66*mp2*L7r*
         RSp451e*RSp563e - 16./27.*m44*m66*mp2*L7r*RSp451e*RSp561e - 8./
         27.*m44*m66*mp2*L6r*pow(RSp563e,2) - 8./27.*m44*m66*mp2*L6r*
         pow(RSp453e,2) + 16./27.*m44*m66*mp2*L6r*RSp561e*RSp563e - 8./
         27.*m44*m66*mp2*L6r*pow(RSp561e,2) + 16./27.*m44*m66*mp2*L6r*
         RSp451e*RSp453e - 8./27.*m44*m66*mp2*L6r*pow(RSp451e,2) - 16./
         27.*m44*m55*mp2*L7r*RSp463e*RSp563e + 16./27.*m44*m55*mp2*L7r*
         RSp561e*RSp463e + 16./27.*m44*m55*mp2*L7r*RSp461e*RSp563e - 16.
         /27.*m44*m55*mp2*L7r*RSp461e*RSp561e - 8./27.*m44*m55*mp2*L6r*
         pow(RSp563e,2) - 8./27.*m44*m55*mp2*L6r*pow(RSp463e,2) + 16./
         27.*m44*m55*mp2*L6r*RSp561e*RSp563e - 8./27.*m44*m55*mp2*L6r*
         pow(RSp561e,2) + 16./27.*m44*m55*mp2*L6r*RSp461e*RSp463e - 8./
         27.*m44*m55*mp2*L6r*pow(RSp461e,2) - 8./27.*pow(m44,2)*mp2*L8r
         *pow(RSp563e,2) );

      M13p6x23L +=  + Bb(mp2,mu2) * ( 16./27.*pow(m44,2)*mp2*L8r*
         RSp561e*RSp563e - 8./27.*pow(m44,2)*mp2*L8r*pow(RSp561e,2) - 8.
         /27.*pow(m44,2)*mp2*L7r*pow(RSp563e,2) + 16./27.*pow(m44,2)*
         mp2*L7r*RSp561e*RSp563e - 8./27.*pow(m44,2)*mp2*L7r*pow(
         RSp561e,2) - 8./27.*pow(m44,2)*mp2*L6r*pow(RSp563e,2) + 16./27.
         *pow(m44,2)*mp2*L6r*RSp561e*RSp563e - 8./27.*pow(m44,2)*mp2*
         L6r*pow(RSp561e,2) - 4./27.*m33*m66*mp2*L5r*pow(RSp453e,2) - 8.
         /27.*m33*m66*mp2*L5r*RSp451e*RSp453e - 4./27.*m33*m66*mp2*L4r*
         pow(RSp563e,2) - 4./27.*m33*m66*mp2*L4r*pow(RSp463e,2) - 4./27.
         *m33*m66*mp2*L4r*pow(RSp453e,2) - 8./27.*m33*m66*mp2*L4r*
         RSp561e*RSp563e - 8./27.*m33*m66*mp2*L4r*RSp461e*RSp463e - 8./
         27.*m33*m66*mp2*L4r*RSp451e*RSp453e + 8./27.*m33*pow(m66,2)*
         L8r*pow(RSp453e,2) + 16./27.*m33*pow(m66,2)*L8r*RSp451e*
         RSp453e + 8./27.*m33*pow(m66,2)*L7r*pow(RSp453e,2) + 16./27.*
         m33*pow(m66,2)*L7r*RSp451e*RSp453e + 8./27.*m33*pow(m66,2)*L6r
         *pow(RSp453e,2) );

      M13p6x23L +=  + Bb(mp2,mu2) * ( 16./27.*m33*pow(m66,2)*L6r*
         RSp451e*RSp453e - 4./27.*m33*m55*mp2*L5r*pow(RSp463e,2) - 8./
         27.*m33*m55*mp2*L5r*RSp461e*RSp463e - 4./27.*m33*m55*mp2*L4r*
         pow(RSp563e,2) - 4./27.*m33*m55*mp2*L4r*pow(RSp463e,2) - 4./27.
         *m33*m55*mp2*L4r*pow(RSp453e,2) - 8./27.*m33*m55*mp2*L4r*
         RSp561e*RSp563e - 8./27.*m33*m55*mp2*L4r*RSp461e*RSp463e - 8./
         27.*m33*m55*mp2*L4r*RSp451e*RSp453e + 16./27.*m33*m55*m66*L7r*
         RSp453e*RSp463e + 16./27.*m33*m55*m66*L7r*RSp461e*RSp453e + 16.
         /27.*m33*m55*m66*L7r*RSp451e*RSp463e + 8./27.*m33*m55*m66*L6r*
         pow(RSp463e,2) + 8./27.*m33*m55*m66*L6r*pow(RSp453e,2) + 16./
         27.*m33*m55*m66*L6r*RSp461e*RSp463e + 16./27.*m33*m55*m66*L6r*
         RSp451e*RSp453e + 8./27.*m33*pow(m55,2)*L8r*pow(RSp463e,2) + 
         16./27.*m33*pow(m55,2)*L8r*RSp461e*RSp463e + 8./27.*m33*pow(
         m55,2)*L7r*pow(RSp463e,2) + 16./27.*m33*pow(m55,2)*L7r*RSp461e
         *RSp463e + 8./27.*m33*pow(m55,2)*L6r*pow(RSp463e,2) + 16./27.*
         m33*pow(m55,2)*L6r*RSp461e*RSp463e );

      M13p6x23L +=  + Bb(mp2,mu2) * (  - 4./27.*m33*m44*mp2*L5r*pow(
         RSp563e,2) - 8./27.*m33*m44*mp2*L5r*RSp561e*RSp563e - 4./27.*
         m33*m44*mp2*L4r*pow(RSp563e,2) - 4./27.*m33*m44*mp2*L4r*pow(
         RSp463e,2) - 4./27.*m33*m44*mp2*L4r*pow(RSp453e,2) - 8./27.*
         m33*m44*mp2*L4r*RSp561e*RSp563e - 8./27.*m33*m44*mp2*L4r*
         RSp461e*RSp463e - 8./27.*m33*m44*mp2*L4r*RSp451e*RSp453e + 16./
         27.*m33*m44*m66*L7r*RSp453e*RSp563e + 16./27.*m33*m44*m66*L7r*
         RSp561e*RSp453e + 16./27.*m33*m44*m66*L7r*RSp451e*RSp563e + 8./
         27.*m33*m44*m66*L6r*pow(RSp563e,2) + 8./27.*m33*m44*m66*L6r*
         pow(RSp453e,2) + 16./27.*m33*m44*m66*L6r*RSp561e*RSp563e + 16./
         27.*m33*m44*m66*L6r*RSp451e*RSp453e + 16./27.*m33*m44*m55*L7r*
         RSp463e*RSp563e + 16./27.*m33*m44*m55*L7r*RSp561e*RSp463e + 16.
         /27.*m33*m44*m55*L7r*RSp461e*RSp563e + 8./27.*m33*m44*m55*L6r*
         pow(RSp563e,2) + 8./27.*m33*m44*m55*L6r*pow(RSp463e,2) + 16./
         27.*m33*m44*m55*L6r*RSp561e*RSp563e + 16./27.*m33*m44*m55*L6r*
         RSp461e*RSp463e );

      M13p6x23L +=  + Bb(mp2,mu2) * ( 8./27.*m33*pow(m44,2)*L8r*pow(
         RSp563e,2) + 16./27.*m33*pow(m44,2)*L8r*RSp561e*RSp563e + 8./
         27.*m33*pow(m44,2)*L7r*pow(RSp563e,2) + 16./27.*m33*pow(m44,2)
         *L7r*RSp561e*RSp563e + 8./27.*m33*pow(m44,2)*L6r*pow(
         RSp563e,2) + 16./27.*m33*pow(m44,2)*L6r*RSp561e*RSp563e - 8./
         27.*m11*m66*mp2*L5r*RSp451e*RSp453e - 4./27.*m11*m66*mp2*L5r*
         pow(RSp451e,2) - 8./27.*m11*m66*mp2*L4r*RSp561e*RSp563e - 4./
         27.*m11*m66*mp2*L4r*pow(RSp561e,2) - 8./27.*m11*m66*mp2*L4r*
         RSp461e*RSp463e - 4./27.*m11*m66*mp2*L4r*pow(RSp461e,2) - 8./
         27.*m11*m66*mp2*L4r*RSp451e*RSp453e - 4./27.*m11*m66*mp2*L4r*
         pow(RSp451e,2) + 16./27.*m11*pow(m66,2)*L8r*RSp451e*RSp453e + 
         8./27.*m11*pow(m66,2)*L8r*pow(RSp451e,2) + 16./27.*m11*pow(
         m66,2)*L7r*RSp451e*RSp453e + 8./27.*m11*pow(m66,2)*L7r*pow(
         RSp451e,2) + 16./27.*m11*pow(m66,2)*L6r*RSp451e*RSp453e + 8./
         27.*m11*pow(m66,2)*L6r*pow(RSp451e,2) - 8./27.*m11*m55*mp2*L5r
         *RSp461e*RSp463e );

      M13p6x23L +=  + Bb(mp2,mu2) * (  - 4./27.*m11*m55*mp2*L5r*pow(
         RSp461e,2) - 8./27.*m11*m55*mp2*L4r*RSp561e*RSp563e - 4./27.*
         m11*m55*mp2*L4r*pow(RSp561e,2) - 8./27.*m11*m55*mp2*L4r*
         RSp461e*RSp463e - 4./27.*m11*m55*mp2*L4r*pow(RSp461e,2) - 8./
         27.*m11*m55*mp2*L4r*RSp451e*RSp453e - 4./27.*m11*m55*mp2*L4r*
         pow(RSp451e,2) + 16./27.*m11*m55*m66*L7r*RSp461e*RSp453e + 16./
         27.*m11*m55*m66*L7r*RSp451e*RSp463e + 16./27.*m11*m55*m66*L7r*
         RSp451e*RSp461e + 16./27.*m11*m55*m66*L6r*RSp461e*RSp463e + 8./
         27.*m11*m55*m66*L6r*pow(RSp461e,2) + 16./27.*m11*m55*m66*L6r*
         RSp451e*RSp453e + 8./27.*m11*m55*m66*L6r*pow(RSp451e,2) + 16./
         27.*m11*pow(m55,2)*L8r*RSp461e*RSp463e + 8./27.*m11*pow(m55,2)
         *L8r*pow(RSp461e,2) + 16./27.*m11*pow(m55,2)*L7r*RSp461e*
         RSp463e + 8./27.*m11*pow(m55,2)*L7r*pow(RSp461e,2) + 16./27.*
         m11*pow(m55,2)*L6r*RSp461e*RSp463e + 8./27.*m11*pow(m55,2)*L6r
         *pow(RSp461e,2) - 8./27.*m11*m44*mp2*L5r*RSp561e*RSp563e - 4./
         27.*m11*m44*mp2*L5r*pow(RSp561e,2) );

      M13p6x23L +=  + Bb(mp2,mu2) * (  - 8./27.*m11*m44*mp2*L4r*RSp561e
         *RSp563e - 4./27.*m11*m44*mp2*L4r*pow(RSp561e,2) - 8./27.*m11*
         m44*mp2*L4r*RSp461e*RSp463e - 4./27.*m11*m44*mp2*L4r*pow(
         RSp461e,2) - 8./27.*m11*m44*mp2*L4r*RSp451e*RSp453e - 4./27.*
         m11*m44*mp2*L4r*pow(RSp451e,2) + 16./27.*m11*m44*m66*L7r*
         RSp561e*RSp453e + 16./27.*m11*m44*m66*L7r*RSp451e*RSp563e + 16.
         /27.*m11*m44*m66*L7r*RSp451e*RSp561e + 16./27.*m11*m44*m66*L6r
         *RSp561e*RSp563e + 8./27.*m11*m44*m66*L6r*pow(RSp561e,2) + 16./
         27.*m11*m44*m66*L6r*RSp451e*RSp453e + 8./27.*m11*m44*m66*L6r*
         pow(RSp451e,2) + 16./27.*m11*m44*m55*L7r*RSp561e*RSp463e + 16./
         27.*m11*m44*m55*L7r*RSp461e*RSp563e + 16./27.*m11*m44*m55*L7r*
         RSp461e*RSp561e + 16./27.*m11*m44*m55*L6r*RSp561e*RSp563e + 8./
         27.*m11*m44*m55*L6r*pow(RSp561e,2) + 16./27.*m11*m44*m55*L6r*
         RSp461e*RSp463e + 8./27.*m11*m44*m55*L6r*pow(RSp461e,2) + 16./
         27.*m11*pow(m44,2)*L8r*RSp561e*RSp563e + 8./27.*m11*pow(m44,2)
         *L8r*pow(RSp561e,2) );

      M13p6x23L +=  + Bb(mp2,mu2) * ( 16./27.*m11*pow(m44,2)*L7r*
         RSp561e*RSp563e + 8./27.*m11*pow(m44,2)*L7r*pow(RSp561e,2) + 
         16./27.*m11*pow(m44,2)*L6r*RSp561e*RSp563e + 8./27.*m11*pow(
         m44,2)*L6r*pow(RSp561e,2) );

      M13p6x23L +=  - 64*m33*pow(m66,2)*L4r*L6r + 32*m33*pow(m66,2)*
         pow(L4r,2) - 128*m33*m55*m66*L4r*L6r + 64*m33*m55*m66*pow(
         L4r,2) - 64*m33*pow(m55,2)*L4r*L6r + 32*m33*pow(m55,2)*pow(
         L4r,2) - 128*m33*m44*m66*L4r*L6r + 64*m33*m44*m66*pow(L4r,2)
          - 128*m33*m44*m55*L4r*L6r + 64*m33*m44*m55*pow(L4r,2) - 64*
         m33*pow(m44,2)*L4r*L6r + 32*m33*pow(m44,2)*pow(L4r,2) - 32*
         pow(m33,2)*m66*L5r*L6r - 32*pow(m33,2)*m66*L4r*L8r + 32*pow(
         m33,2)*m66*L4r*L5r - 32*pow(m33,2)*m55*L5r*L6r - 32*pow(m33,2)
         *m55*L4r*L8r + 32*pow(m33,2)*m55*L4r*L5r - 32*pow(m33,2)*m44*
         L5r*L6r - 32*pow(m33,2)*m44*L4r*L8r + 32*pow(m33,2)*m44*L4r*
         L5r - 16*pow(m33,3)*L5r*L8r + 8*pow(m33,3)*pow(L5r,2) - 64*m11
         *pow(m66,2)*L4r*L6r + 32*m11*pow(m66,2)*pow(L4r,2) - 128*m11*
         m55*m66*L4r*L6r + 64*m11*m55*m66*pow(L4r,2) - 64*m11*pow(
         m55,2)*L4r*L6r + 32*m11*pow(m55,2)*pow(L4r,2) - 128*m11*m44*
         m66*L4r*L6r + 64*m11*m44*m66*pow(L4r,2) - 128*m11*m44*m55*L4r*
         L6r;
      M13p6x23L +=  + 64*m11*m44*m55*pow(L4r,2) - 64*m11*pow(m44,2)*L4r
         *L6r + 32*m11*pow(m44,2)*pow(L4r,2) - 64*m11*m33*m66*L5r*L6r
          - 64*m11*m33*m66*L4r*L8r + 64*m11*m33*m66*L4r*L5r - 64*m11*
         m33*m55*L5r*L6r - 64*m11*m33*m55*L4r*L8r + 64*m11*m33*m55*L4r*
         L5r - 64*m11*m33*m44*L5r*L6r - 64*m11*m33*m44*L4r*L8r + 64*m11
         *m33*m44*L4r*L5r - 48*m11*pow(m33,2)*L5r*L8r + 24*m11*pow(
         m33,2)*pow(L5r,2) - 32*pow(m11,2)*m66*L5r*L6r - 32*pow(m11,2)*
         m66*L4r*L8r + 32*pow(m11,2)*m66*L4r*L5r - 32*pow(m11,2)*m55*
         L5r*L6r - 32*pow(m11,2)*m55*L4r*L8r + 32*pow(m11,2)*m55*L4r*
         L5r - 32*pow(m11,2)*m44*L5r*L6r - 32*pow(m11,2)*m44*L4r*L8r + 
         32*pow(m11,2)*m44*L4r*L5r - 48*pow(m11,2)*m33*L5r*L8r + 24*
         pow(m11,2)*m33*pow(L5r,2) - 16*pow(m11,3)*L5r*L8r + 8*pow(
         m11,3)*pow(L5r,2);

  return M13p6x23L/pow(f0,4);
}

double mv2s3nf3p6K(const quarkmassnf mass, const Ki Kiin){
  double KK[116],mu;
  Kiin.out(KK,mu);
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if (mu0 != mu){
    cout << "ERROR: differing mu's in mv2s3nf3p6L\n";}
  if ( (nq != 5) || (B0mq.size() != 5)){
    cout << "ERROR: mv2s3nf3p6K probably called with wrong number"
	 << "of quark masses\n";}
  // five masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m33 = 2.*B0mq[1];
  double m44 = 2.*B0mq[2];
  double m55 = 2.*B0mq[3];
  double m66 = 2.*B0mq[4];

  double M13p6x23K =
       - 4*KK[17]*pow(m33,3) - 12*KK[17]*m11*pow(m33,2) - 12*KK[17]*
         pow(m11,2)*m33 - 4*KK[17]*pow(m11,3) - 8*KK[18]*pow(m33,2)*m66
          - 8*KK[18]*pow(m33,2)*m55 - 8*KK[18]*pow(m33,2)*m44 - 16*KK[
         18]*m11*m33*m66 - 16*KK[18]*m11*m33*m55 - 16*KK[18]*m11*m33*
         m44 - 8*KK[18]*pow(m11,2)*m66 - 8*KK[18]*pow(m11,2)*m55 - 8*
         KK[18]*pow(m11,2)*m44 - 4*KK[19]*pow(m33,3) - 4*KK[19]*m11*
         pow(m33,2) - 4*KK[19]*pow(m11,2)*m33 - 4*KK[19]*pow(m11,3) - 4
         *KK[20]*pow(m33,2)*m66 - 4*KK[20]*pow(m33,2)*m55 - 4*KK[20]*
         pow(m33,2)*m44 - 8*KK[20]*m11*m33*m66 - 8*KK[20]*m11*m33*m55
          - 8*KK[20]*m11*m33*m44 - 4*KK[20]*pow(m11,2)*m66 - 4*KK[20]*
         pow(m11,2)*m55 - 4*KK[20]*pow(m11,2)*m44 - 8*KK[21]*m33*pow(
         m66,2) - 8*KK[21]*m33*pow(m55,2) - 8*KK[21]*m33*pow(m44,2) - 8
         *KK[21]*m11*pow(m66,2) - 8*KK[21]*m11*pow(m55,2) - 8*KK[21]*
         m11*pow(m44,2) - 8*KK[22]*m33*pow(m66,2) - 16*KK[22]*m33*m55*
         m66;
      M13p6x23K +=  - 8*KK[22]*m33*pow(m55,2) - 16*KK[22]*m33*m44*m66
          - 16*KK[22]*m33*m44*m55 - 8*KK[22]*m33*pow(m44,2) - 8*KK[22]*
         m11*pow(m66,2) - 16*KK[22]*m11*m55*m66 - 8*KK[22]*m11*pow(
         m55,2) - 16*KK[22]*m11*m44*m66 - 16*KK[22]*m11*m44*m55 - 8*KK[
         22]*m11*pow(m44,2) - 8*KK[23]*m11*pow(m33,2) - 8*KK[23]*pow(
         m11,2)*m33 + 12*KK[25]*pow(m33,3) + 12*KK[25]*m11*pow(m33,2)
          + 12*KK[25]*pow(m11,2)*m33 + 12*KK[25]*pow(m11,3) + 8*KK[26]*
         m33*pow(m66,2) + 8*KK[26]*m33*pow(m55,2) + 8*KK[26]*m33*pow(
         m44,2) + 8*KK[26]*pow(m33,2)*m66 + 8*KK[26]*pow(m33,2)*m55 + 8
         *KK[26]*pow(m33,2)*m44 + 8*KK[26]*m11*pow(m66,2) + 8*KK[26]*
         m11*pow(m55,2) + 8*KK[26]*m11*pow(m44,2) + 16*KK[26]*m11*m33*
         m66 + 16*KK[26]*m11*m33*m55 + 16*KK[26]*m11*m33*m44 + 8*KK[26]
         *pow(m11,2)*m66 + 8*KK[26]*pow(m11,2)*m55 + 8*KK[26]*pow(
         m11,2)*m44 + 24*KK[27]*m33*pow(m66,2) + 48*KK[27]*m33*m55*m66
          + 24*KK[27]*m33*pow(m55,2) + 48*KK[27]*m33*m44*m66 + 48*KK[27
         ]*m33*m44*m55;
      M13p6x23K +=  + 24*KK[27]*m33*pow(m44,2) + 24*KK[27]*m11*pow(
         m66,2) + 48*KK[27]*m11*m55*m66 + 24*KK[27]*m11*pow(m55,2) + 48
         *KK[27]*m11*m44*m66 + 48*KK[27]*m11*m44*m55 + 24*KK[27]*m11*
         pow(m44,2) + 4*KK[39]*pow(m33,3) + 12*KK[39]*m11*pow(m33,2) + 
         12*KK[39]*pow(m11,2)*m33 + 4*KK[39]*pow(m11,3) + 8*KK[40]*pow(
         m33,2)*m66 + 8*KK[40]*pow(m33,2)*m55 + 8*KK[40]*pow(m33,2)*m44
          + 16*KK[40]*m11*m33*m66 + 16*KK[40]*m11*m33*m55 + 16*KK[40]*
         m11*m33*m44 + 8*KK[40]*pow(m11,2)*m66 + 8*KK[40]*pow(m11,2)*
         m55 + 8*KK[40]*pow(m11,2)*m44;

  return M13p6x23K/pow(f0,4);
}

double mv2s3nf3p6R(const quarkmassnf mass){
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if ( (nq != 5) || (B0mq.size() != 5)){
    cout << "ERROR: mv2s3nf3p6R probably called with wrong number"
	 << "of quark masses\n";}
  // two masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m33 = 2.*B0mq[1];
  double m44 = 2.*B0mq[2];
  double m55 = 2.*B0mq[3];
  double m66 = 2.*B0mq[4];
  double m13 = (m11+m33)/2;
  double m14 = (m11+m44)/2;
  double m15 = (m11+m55)/2;
  double m16 = (m11+m66)/2;
  double m34 = (m33+m44)/2;
  double m35 = (m33+m55)/2;
  double m36 = (m33+m66)/2;
  double m45 = (m44+m55)/2;
  double m46 = (m44+m66)/2;
  double m56 = (m55+m66)/2;

  double xb1 = (m44+m55+m66)/3.;
  double msqr = sqrt(xb1*xb1-(m44*m55+m44*m66+m55*m66)/3.);
  double mp2 = xb1-msqr; 
  double me2 = xb1+msqr;
  double mu2 = mu0*mu0;

  double RN13 = RN(m11,m33);
  double RN1p = RN(m11,mp2);
  double RN1e = RN(m11,me2);
  double RN3p = RN(m33,mp2);
  double RN3e = RN(m33,me2);
  double RNep = RN(me2,mp2);

  double RSp456e = RS(mp2,m44,m55,m66,me2);
  double RSp465e = RS(mp2,m44,m66,m55,me2);
  double RSp564e = RS(mp2,m55,m66,m44,me2);
  double RSe456p = RS(me2,m44,m55,m66,mp2);
  double RSe465p = RS(me2,m44,m66,m55,mp2);
  double RSe564p = RS(me2,m55,m66,m44,mp2);
  double RSp451e = RS(mp2,m44,m55,m11,me2);
  double RSp461e = RS(mp2,m44,m66,m11,me2);
  double RSp561e = RS(mp2,m55,m66,m11,me2);
  double RSe451p = RS(me2,m44,m55,m11,mp2);
  double RSe461p = RS(me2,m44,m66,m11,mp2);
  double RSe561p = RS(me2,m55,m66,m11,mp2);
  double RSp453e = RS(mp2,m44,m55,m33,me2);
  double RSp463e = RS(mp2,m44,m66,m33,me2);
  double RSp563e = RS(mp2,m55,m66,m33,me2);
  double RSe453p = RS(me2,m44,m55,m33,mp2);
  double RSe463p = RS(me2,m44,m66,m33,mp2);
  double RSe563p = RS(me2,m55,m66,m33,mp2);
  double RS456pe = RS(m44,m55,m66,mp2,me2);
  double RS546pe = RS(m55,m44,m66,mp2,me2);
  double RS645pe = RS(m66,m44,m55,mp2,me2);
  double RS156pe = RS(m11,m55,m66,mp2,me2);
  double RS146pe = RS(m11,m44,m66,mp2,me2);
  double RS145pe = RS(m11,m44,m55,mp2,me2);
  double RS356pe = RS(m33,m55,m66,mp2,me2);
  double RS346pe = RS(m33,m44,m66,mp2,me2);
  double RS345pe = RS(m33,m44,m55,mp2,me2);

  double RD1456ep = RD(m11,m44,m55,m66,me2,mp2);
  double RD3456ep = RD(m33,m44,m55,m66,me2,mp2);
  //  double RC1456pe = RC(m11,m44,m55,m66,mp2,me2);  
  double RSe45611p = RS(me2,m44,m55,m66,m11,m11,mp2);
  double RSp45611e = RS(mp2,m44,m55,m66,m11,m11,me2);
  double RS1456ppe = RS(m11,m44,m55,m66,mp2,mp2,me2);
  double RS1456pee = RS(m11,m44,m55,m66,mp2,me2,me2);
  double RSe45633p = RS(me2,m44,m55,m66,m33,m33,mp2);
  double RSp45633e = RS(mp2,m44,m55,m66,m33,m33,me2);
  double RS3456ppe = RS(m33,m44,m55,m66,mp2,mp2,me2);
  double RS3456pee = RS(m33,m44,m55,m66,mp2,me2,me2);
  double RSe45613p = RS(me2,m44,m55,m66,m11,m33,mp2);
  double RSp45613e = RS(mp2,m44,m55,m66,m11,m33,me2);
  double RS14563pe = RS(m11,m44,m55,m66,m33,mp2,me2);
  double RS34561pe = RS(m33,m44,m55,m66,m11,mp2,me2);

  double M13p6x23R =
       + pi162 * ( 1./576.*m33*pow(RD3456ep,2) + 1./288.*m33*RD1456ep*
         RD3456ep + 1./576.*m33*pow(RD1456ep,2) - 1./144.*m33*mp2*
         RD3456ep*RSp45613e + 1./288.*m33*mp2*RSp45633e*RD3456ep - 1./
         144.*m33*mp2*RD1456ep*RSp45613e + 1./288.*m33*mp2*RD1456ep*
         RSp45633e + 1./288.*m33*mp2*RSp45611e*RD3456ep + 1./288.*m33*
         mp2*RSp45611e*RD1456ep - 7./192.*m33*pow(mp2,2)*RSp45613e + 1./
         144.*m33*pow(mp2,2)*pow(RSp45613e,2) + 1./96.*m33*pow(mp2,2)*
         RSp45633e - 1./144.*m33*pow(mp2,2)*RSp45633e*RSp45613e + 1./
         576.*m33*pow(mp2,2)*pow(RSp45633e,2) + 1./64.*m33*pow(mp2,2)*
         RSp563e + 1./64.*m33*pow(mp2,2)*RSp463e + 1./64.*m33*pow(
         mp2,2)*RSp453e + 1./96.*m33*pow(mp2,2)*RSp45611e - 1./144.*m33
         *pow(mp2,2)*RSp45611e*RSp45613e + 1./288.*m33*pow(mp2,2)*
         RSp45611e*RSp45633e + 1./576.*m33*pow(mp2,2)*pow(RSp45611e,2)
          + 1./64.*m33*pow(mp2,2)*RSp561e + 1./64.*m33*pow(mp2,2)*
         RSp461e + 1./64.*m33*pow(mp2,2)*RSp451e - 1./192.*m33*pow(
         mp2,2)*RSp564e );

      M13p6x23R +=  + pi162 * (  - 1./192.*m33*pow(mp2,2)*RSp465e - 1./
         192.*m33*pow(mp2,2)*RSp456e - 1./144.*m33*me2*RD3456ep*
         RSe45613p + 1./288.*m33*me2*RSe45633p*RD3456ep - 1./144.*m33*
         me2*RD1456ep*RSe45613p + 1./288.*m33*me2*RD1456ep*RSe45633p + 
         1./288.*m33*me2*RSe45611p*RD3456ep + 1./288.*m33*me2*RSe45611p
         *RD1456ep + 1./72.*m33*me2*mp2*RSp45613e*RSe45613p - 1./144.*
         m33*me2*mp2*RSe45633p*RSp45613e - 1./144.*m33*me2*mp2*
         RSp45633e*RSe45613p + 1./288.*m33*me2*mp2*RSp45633e*RSe45633p
          - 1./144.*m33*me2*mp2*RSe45611p*RSp45613e + 1./288.*m33*me2*
         mp2*RSe45611p*RSp45633e - 1./144.*m33*me2*mp2*RSp45611e*
         RSe45613p + 1./288.*m33*me2*mp2*RSp45611e*RSe45633p + 1./288.*
         m33*me2*mp2*RSp45611e*RSe45611p - 7./192.*m33*pow(me2,2)*
         RSe45613p + 1./144.*m33*pow(me2,2)*pow(RSe45613p,2) + 1./96.*
         m33*pow(me2,2)*RSe45633p - 1./144.*m33*pow(me2,2)*RSe45633p*
         RSe45613p + 1./576.*m33*pow(me2,2)*pow(RSe45633p,2) + 1./64.*
         m33*pow(me2,2)*RSe563p );

      M13p6x23R +=  + pi162 * ( 1./64.*m33*pow(me2,2)*RSe463p + 1./64.*
         m33*pow(me2,2)*RSe453p + 1./96.*m33*pow(me2,2)*RSe45611p - 1./
         144.*m33*pow(me2,2)*RSe45611p*RSe45613p + 1./288.*m33*pow(
         me2,2)*RSe45611p*RSe45633p + 1./576.*m33*pow(me2,2)*pow(
         RSe45611p,2) + 1./64.*m33*pow(me2,2)*RSe561p + 1./64.*m33*pow(
         me2,2)*RSe461p + 1./64.*m33*pow(me2,2)*RSe451p - 1./192.*m33*
         pow(me2,2)*RSe564p - 1./192.*m33*pow(me2,2)*RSe465p - 1./192.*
         m33*pow(me2,2)*RSe456p - 5./384.*m33*m66*RD3456ep - 5./384.*
         m33*m66*RD1456ep + 1./96.*m33*m66*mp2*RSp45613e - 5./384.*m33*
         m66*mp2*RSp45633e + 1./64.*m33*m66*mp2*RSp453e - 5./384.*m33*
         m66*mp2*RSp45611e + 1./64.*m33*m66*mp2*RSp451e - 1./64.*m33*
         m66*mp2*RSp456e + 1./96.*m33*m66*me2*RSe45613p - 5./384.*m33*
         m66*me2*RSe45633p + 1./64.*m33*m66*me2*RSe453p - 5./384.*m33*
         m66*me2*RSe45611p + 1./64.*m33*m66*me2*RSe451p - 1./64.*m33*
         m66*me2*RSe456p + 15./128.*m33*pow(m66,2) + 5./384.*m33*pow(
         m66,2)*RS34561pe );

      M13p6x23R +=  + pi162 * ( 5./384.*m33*pow(m66,2)*RS14563pe + 5./
         384.*m33*pow(m66,2)*RSe45613p + 5./384.*m33*pow(m66,2)*
         RSp45613e + 1./128.*m33*pow(m66,2)*RS3456pee + 1./128.*m33*
         pow(m66,2)*RS3456ppe - 1./128.*m33*pow(m66,2)*RSe45633p - 1./
         128.*m33*pow(m66,2)*RSp45633e - 1./192.*m33*pow(m66,2)*RS345pe
          - 1./128.*m33*pow(m66,2)*RS346pe - 1./128.*m33*pow(m66,2)*
         RS356pe + 1./384.*m33*pow(m66,2)*RSe453p + 1./384.*m33*pow(
         m66,2)*RSp453e + 1./128.*m33*pow(m66,2)*RS1456pee + 1./128.*
         m33*pow(m66,2)*RS1456ppe - 1./128.*m33*pow(m66,2)*RSe45611p - 
         1./128.*m33*pow(m66,2)*RSp45611e - 1./192.*m33*pow(m66,2)*
         RS145pe - 1./128.*m33*pow(m66,2)*RS146pe - 1./128.*m33*pow(
         m66,2)*RS156pe - 3./128.*m33*pow(m66,2)*RS645pe + 1./384.*m33*
         pow(m66,2)*RSe451p + 1./384.*m33*pow(m66,2)*RSp451e - 1./384.*
         m33*pow(m66,2)*RSe456p - 1./384.*m33*pow(m66,2)*RSp456e - 5./
         384.*m33*m55*RD3456ep - 5./384.*m33*m55*RD1456ep + 1./96.*m33*
         m55*mp2*RSp45613e );

      M13p6x23R +=  + pi162 * (  - 5./384.*m33*m55*mp2*RSp45633e + 1./
         64.*m33*m55*mp2*RSp463e - 5./384.*m33*m55*mp2*RSp45611e + 1./
         64.*m33*m55*mp2*RSp461e - 1./64.*m33*m55*mp2*RSp465e + 1./96.*
         m33*m55*me2*RSe45613p - 5./384.*m33*m55*me2*RSe45633p + 1./64.
         *m33*m55*me2*RSe463p - 5./384.*m33*m55*me2*RSe45611p + 1./64.*
         m33*m55*me2*RSe461p - 1./64.*m33*m55*me2*RSe465p + 5./64.*m33*
         m55*m66 + 15./128.*m33*pow(m55,2) + 5./384.*m33*pow(m55,2)*
         RS34561pe + 5./384.*m33*pow(m55,2)*RS14563pe + 5./384.*m33*
         pow(m55,2)*RSe45613p + 5./384.*m33*pow(m55,2)*RSp45613e + 1./
         128.*m33*pow(m55,2)*RS3456pee + 1./128.*m33*pow(m55,2)*
         RS3456ppe - 1./128.*m33*pow(m55,2)*RSe45633p - 1./128.*m33*
         pow(m55,2)*RSp45633e - 1./128.*m33*pow(m55,2)*RS345pe - 1./192.
         *m33*pow(m55,2)*RS346pe - 1./128.*m33*pow(m55,2)*RS356pe + 1./
         384.*m33*pow(m55,2)*RSe463p + 1./384.*m33*pow(m55,2)*RSp463e
          + 1./128.*m33*pow(m55,2)*RS1456pee + 1./128.*m33*pow(m55,2)*
         RS1456ppe );

      M13p6x23R +=  + pi162 * (  - 1./128.*m33*pow(m55,2)*RSe45611p - 1.
         /128.*m33*pow(m55,2)*RSp45611e - 1./128.*m33*pow(m55,2)*
         RS145pe - 1./192.*m33*pow(m55,2)*RS146pe - 1./128.*m33*pow(
         m55,2)*RS156pe - 3./128.*m33*pow(m55,2)*RS546pe + 1./384.*m33*
         pow(m55,2)*RSe461p + 1./384.*m33*pow(m55,2)*RSp461e - 1./384.*
         m33*pow(m55,2)*RSe465p - 1./384.*m33*pow(m55,2)*RSp465e - 5./
         384.*m33*m44*RD3456ep - 5./384.*m33*m44*RD1456ep + 1./96.*m33*
         m44*mp2*RSp45613e - 5./384.*m33*m44*mp2*RSp45633e + 1./64.*m33
         *m44*mp2*RSp563e - 5./384.*m33*m44*mp2*RSp45611e + 1./64.*m33*
         m44*mp2*RSp561e - 1./64.*m33*m44*mp2*RSp564e + 1./96.*m33*m44*
         me2*RSe45613p - 5./384.*m33*m44*me2*RSe45633p + 1./64.*m33*m44
         *me2*RSe563p - 5./384.*m33*m44*me2*RSe45611p + 1./64.*m33*m44*
         me2*RSe561p - 1./64.*m33*m44*me2*RSe564p + 5./64.*m33*m44*m66
          + 5./64.*m33*m44*m55 + 15./128.*m33*pow(m44,2) + 5./384.*m33*
         pow(m44,2)*RS34561pe + 5./384.*m33*pow(m44,2)*RS14563pe + 5./
         384.*m33*pow(m44,2)*RSe45613p );

      M13p6x23R +=  + pi162 * ( 5./384.*m33*pow(m44,2)*RSp45613e + 1./
         128.*m33*pow(m44,2)*RS3456pee + 1./128.*m33*pow(m44,2)*
         RS3456ppe - 1./128.*m33*pow(m44,2)*RSe45633p - 1./128.*m33*
         pow(m44,2)*RSp45633e - 1./128.*m33*pow(m44,2)*RS345pe - 1./128.
         *m33*pow(m44,2)*RS346pe - 1./192.*m33*pow(m44,2)*RS356pe + 1./
         384.*m33*pow(m44,2)*RSe563p + 1./384.*m33*pow(m44,2)*RSp563e
          + 1./128.*m33*pow(m44,2)*RS1456pee + 1./128.*m33*pow(m44,2)*
         RS1456ppe - 1./128.*m33*pow(m44,2)*RSe45611p - 1./128.*m33*
         pow(m44,2)*RSp45611e - 1./128.*m33*pow(m44,2)*RS145pe - 1./128.
         *m33*pow(m44,2)*RS146pe - 1./192.*m33*pow(m44,2)*RS156pe - 3./
         128.*m33*pow(m44,2)*RS456pe + 1./384.*m33*pow(m44,2)*RSe561p
          + 1./384.*m33*pow(m44,2)*RSp561e - 1./384.*m33*pow(m44,2)*
         RSe564p - 1./384.*m33*pow(m44,2)*RSp564e + 1./576.*pow(m33,2)*
         RD3456ep - 29./2592.*pow(m33,2)*RD3456ep*RS34561pe - 11./2592.
         *pow(m33,2)*RD3456ep*RS14563pe - 11./2592.*pow(m33,2)*RD3456ep
         *RSe45613p );

      M13p6x23R +=  + pi162 * (  - 11./2592.*pow(m33,2)*RD3456ep*
         RSp45613e - 29./5184.*pow(m33,2)*RS3456pee*RD3456ep - 29./5184.
         *pow(m33,2)*RS3456ppe*RD3456ep + 11./5184.*pow(m33,2)*
         RSe45633p*RD3456ep + 11./5184.*pow(m33,2)*RSp45633e*RD3456ep
          + 29./5184.*pow(m33,2)*RS345pe*RD3456ep + 29./5184.*pow(
         m33,2)*RS346pe*RD3456ep + 29./5184.*pow(m33,2)*RS356pe*
         RD3456ep - 13./1152.*pow(m33,2)*RD1456ep - 29./2592.*pow(
         m33,2)*RD1456ep*RS34561pe - 11./2592.*pow(m33,2)*RD1456ep*
         RS14563pe - 11./2592.*pow(m33,2)*RD1456ep*RSe45613p - 11./2592.
         *pow(m33,2)*RD1456ep*RSp45613e - 29./5184.*pow(m33,2)*RD1456ep
         *RS3456pee - 29./5184.*pow(m33,2)*RD1456ep*RS3456ppe + 11./
         5184.*pow(m33,2)*RD1456ep*RSe45633p + 11./5184.*pow(m33,2)*
         RD1456ep*RSp45633e + 29./5184.*pow(m33,2)*RD1456ep*RS345pe + 
         29./5184.*pow(m33,2)*RD1456ep*RS346pe + 29./5184.*pow(m33,2)*
         RD1456ep*RS356pe - 11./5184.*pow(m33,2)*RS1456pee*RD3456ep - 
         11./5184.*pow(m33,2)*RS1456pee*RD1456ep );

      M13p6x23R +=  + pi162 * (  - 11./5184.*pow(m33,2)*RS1456ppe*
         RD3456ep - 11./5184.*pow(m33,2)*RS1456ppe*RD1456ep + 11./5184.
         *pow(m33,2)*RSe45611p*RD3456ep + 11./5184.*pow(m33,2)*
         RSe45611p*RD1456ep + 11./5184.*pow(m33,2)*RSp45611e*RD3456ep
          + 11./5184.*pow(m33,2)*RSp45611e*RD1456ep + 11./5184.*pow(
         m33,2)*RS145pe*RD3456ep + 11./5184.*pow(m33,2)*RS145pe*
         RD1456ep + 11./5184.*pow(m33,2)*RS146pe*RD3456ep + 11./5184.*
         pow(m33,2)*RS146pe*RD1456ep + 11./5184.*pow(m33,2)*RS156pe*
         RD3456ep + 11./5184.*pow(m33,2)*RS156pe*RD1456ep + 5./288.*
         pow(m33,2)*mp2*RSp45613e + 29./1296.*pow(m33,2)*mp2*RSp45613e*
         RS34561pe + 11./1296.*pow(m33,2)*mp2*RSp45613e*RS14563pe + 11./
         1296.*pow(m33,2)*mp2*RSp45613e*RSe45613p + 11./1296.*pow(
         m33,2)*mp2*pow(RSp45613e,2) + 29./2592.*pow(m33,2)*mp2*
         RS3456pee*RSp45613e + 29./2592.*pow(m33,2)*mp2*RS3456ppe*
         RSp45613e - 11./2592.*pow(m33,2)*mp2*RSe45633p*RSp45613e - 11./
         576.*pow(m33,2)*mp2*RSp45633e );

      M13p6x23R +=  + pi162 * (  - 29./2592.*pow(m33,2)*mp2*RSp45633e*
         RS34561pe - 11./2592.*pow(m33,2)*mp2*RSp45633e*RS14563pe - 11./
         2592.*pow(m33,2)*mp2*RSp45633e*RSe45613p - 11./1296.*pow(
         m33,2)*mp2*RSp45633e*RSp45613e - 29./5184.*pow(m33,2)*mp2*
         RSp45633e*RS3456pee - 29./5184.*pow(m33,2)*mp2*RSp45633e*
         RS3456ppe + 11./5184.*pow(m33,2)*mp2*RSp45633e*RSe45633p + 11./
         5184.*pow(m33,2)*mp2*pow(RSp45633e,2) - 29./2592.*pow(m33,2)*
         mp2*RS345pe*RSp45613e + 29./5184.*pow(m33,2)*mp2*RS345pe*
         RSp45633e - 29./2592.*pow(m33,2)*mp2*RS346pe*RSp45613e + 29./
         5184.*pow(m33,2)*mp2*RS346pe*RSp45633e - 29./2592.*pow(m33,2)*
         mp2*RS356pe*RSp45613e + 29./5184.*pow(m33,2)*mp2*RS356pe*
         RSp45633e + 1./144.*pow(m33,2)*mp2*RSp563e + 1./144.*pow(
         m33,2)*mp2*RSp463e + 1./144.*pow(m33,2)*mp2*RSp453e + 11./2592.
         *pow(m33,2)*mp2*RS1456pee*RSp45613e - 11./5184.*pow(m33,2)*mp2
         *RS1456pee*RSp45633e + 11./2592.*pow(m33,2)*mp2*RS1456ppe*
         RSp45613e );

      M13p6x23R +=  + pi162 * (  - 11./5184.*pow(m33,2)*mp2*RS1456ppe*
         RSp45633e - 11./2592.*pow(m33,2)*mp2*RSe45611p*RSp45613e + 11./
         5184.*pow(m33,2)*mp2*RSe45611p*RSp45633e - 13./1152.*pow(
         m33,2)*mp2*RSp45611e - 29./2592.*pow(m33,2)*mp2*RSp45611e*
         RS34561pe - 11./2592.*pow(m33,2)*mp2*RSp45611e*RS14563pe - 11./
         2592.*pow(m33,2)*mp2*RSp45611e*RSe45613p - 11./1296.*pow(
         m33,2)*mp2*RSp45611e*RSp45613e - 29./5184.*pow(m33,2)*mp2*
         RSp45611e*RS3456pee - 29./5184.*pow(m33,2)*mp2*RSp45611e*
         RS3456ppe + 11./5184.*pow(m33,2)*mp2*RSp45611e*RSe45633p + 11./
         2592.*pow(m33,2)*mp2*RSp45611e*RSp45633e + 29./5184.*pow(
         m33,2)*mp2*RSp45611e*RS345pe + 29./5184.*pow(m33,2)*mp2*
         RSp45611e*RS346pe + 29./5184.*pow(m33,2)*mp2*RSp45611e*RS356pe
          - 11./5184.*pow(m33,2)*mp2*RSp45611e*RS1456pee - 11./5184.*
         pow(m33,2)*mp2*RSp45611e*RS1456ppe + 11./5184.*pow(m33,2)*mp2*
         RSp45611e*RSe45611p + 11./5184.*pow(m33,2)*mp2*pow(
         RSp45611e,2) );

      M13p6x23R +=  + pi162 * (  - 11./2592.*pow(m33,2)*mp2*RS145pe*
         RSp45613e + 11./5184.*pow(m33,2)*mp2*RS145pe*RSp45633e + 11./
         5184.*pow(m33,2)*mp2*RS145pe*RSp45611e - 11./2592.*pow(m33,2)*
         mp2*RS146pe*RSp45613e + 11./5184.*pow(m33,2)*mp2*RS146pe*
         RSp45633e + 11./5184.*pow(m33,2)*mp2*RS146pe*RSp45611e - 11./
         2592.*pow(m33,2)*mp2*RS156pe*RSp45613e + 11./5184.*pow(m33,2)*
         mp2*RS156pe*RSp45633e + 11./5184.*pow(m33,2)*mp2*RS156pe*
         RSp45611e + 1./576.*pow(m33,2)*mp2*RSp561e + 1./576.*pow(
         m33,2)*mp2*RSp461e + 1./576.*pow(m33,2)*mp2*RSp451e - 5./1152.
         *pow(m33,2)*mp2*RSp564e - 5./1152.*pow(m33,2)*mp2*RSp465e - 5./
         1152.*pow(m33,2)*mp2*RSp456e + 5./288.*pow(m33,2)*me2*
         RSe45613p + 29./1296.*pow(m33,2)*me2*RSe45613p*RS34561pe + 11./
         1296.*pow(m33,2)*me2*RSe45613p*RS14563pe + 11./1296.*pow(
         m33,2)*me2*pow(RSe45613p,2) + 11./1296.*pow(m33,2)*me2*
         RSp45613e*RSe45613p + 29./2592.*pow(m33,2)*me2*RS3456pee*
         RSe45613p );

      M13p6x23R +=  + pi162 * ( 29./2592.*pow(m33,2)*me2*RS3456ppe*
         RSe45613p - 11./576.*pow(m33,2)*me2*RSe45633p - 29./2592.*pow(
         m33,2)*me2*RSe45633p*RS34561pe - 11./2592.*pow(m33,2)*me2*
         RSe45633p*RS14563pe - 11./1296.*pow(m33,2)*me2*RSe45633p*
         RSe45613p - 11./2592.*pow(m33,2)*me2*RSe45633p*RSp45613e - 29./
         5184.*pow(m33,2)*me2*RSe45633p*RS3456pee - 29./5184.*pow(
         m33,2)*me2*RSe45633p*RS3456ppe + 11./5184.*pow(m33,2)*me2*pow(
         RSe45633p,2) - 11./2592.*pow(m33,2)*me2*RSp45633e*RSe45613p + 
         11./5184.*pow(m33,2)*me2*RSp45633e*RSe45633p - 29./2592.*pow(
         m33,2)*me2*RS345pe*RSe45613p + 29./5184.*pow(m33,2)*me2*
         RS345pe*RSe45633p - 29./2592.*pow(m33,2)*me2*RS346pe*RSe45613p
          + 29./5184.*pow(m33,2)*me2*RS346pe*RSe45633p - 29./2592.*pow(
         m33,2)*me2*RS356pe*RSe45613p + 29./5184.*pow(m33,2)*me2*
         RS356pe*RSe45633p + 1./144.*pow(m33,2)*me2*RSe563p + 1./144.*
         pow(m33,2)*me2*RSe463p + 1./144.*pow(m33,2)*me2*RSe453p + 11./
         2592.*pow(m33,2)*me2*RS1456pee*RSe45613p );

      M13p6x23R +=  + pi162 * (  - 11./5184.*pow(m33,2)*me2*RS1456pee*
         RSe45633p + 11./2592.*pow(m33,2)*me2*RS1456ppe*RSe45613p - 11./
         5184.*pow(m33,2)*me2*RS1456ppe*RSe45633p - 13./1152.*pow(
         m33,2)*me2*RSe45611p - 29./2592.*pow(m33,2)*me2*RSe45611p*
         RS34561pe - 11./2592.*pow(m33,2)*me2*RSe45611p*RS14563pe - 11./
         1296.*pow(m33,2)*me2*RSe45611p*RSe45613p - 11./2592.*pow(
         m33,2)*me2*RSe45611p*RSp45613e - 29./5184.*pow(m33,2)*me2*
         RSe45611p*RS3456pee - 29./5184.*pow(m33,2)*me2*RSe45611p*
         RS3456ppe + 11./2592.*pow(m33,2)*me2*RSe45611p*RSe45633p + 11./
         5184.*pow(m33,2)*me2*RSe45611p*RSp45633e + 29./5184.*pow(
         m33,2)*me2*RSe45611p*RS345pe + 29./5184.*pow(m33,2)*me2*
         RSe45611p*RS346pe + 29./5184.*pow(m33,2)*me2*RSe45611p*RS356pe
          - 11./5184.*pow(m33,2)*me2*RSe45611p*RS1456pee - 11./5184.*
         pow(m33,2)*me2*RSe45611p*RS1456ppe + 11./5184.*pow(m33,2)*me2*
         pow(RSe45611p,2) - 11./2592.*pow(m33,2)*me2*RSp45611e*
         RSe45613p );

      M13p6x23R +=  + pi162 * ( 11./5184.*pow(m33,2)*me2*RSp45611e*
         RSe45633p + 11./5184.*pow(m33,2)*me2*RSp45611e*RSe45611p - 11./
         2592.*pow(m33,2)*me2*RS145pe*RSe45613p + 11./5184.*pow(m33,2)*
         me2*RS145pe*RSe45633p + 11./5184.*pow(m33,2)*me2*RS145pe*
         RSe45611p - 11./2592.*pow(m33,2)*me2*RS146pe*RSe45613p + 11./
         5184.*pow(m33,2)*me2*RS146pe*RSe45633p + 11./5184.*pow(m33,2)*
         me2*RS146pe*RSe45611p - 11./2592.*pow(m33,2)*me2*RS156pe*
         RSe45613p + 11./5184.*pow(m33,2)*me2*RS156pe*RSe45633p + 11./
         5184.*pow(m33,2)*me2*RS156pe*RSe45611p + 1./576.*pow(m33,2)*
         me2*RSe561p + 1./576.*pow(m33,2)*me2*RSe461p + 1./576.*pow(
         m33,2)*me2*RSe451p - 5./1152.*pow(m33,2)*me2*RSe564p - 5./1152.
         *pow(m33,2)*me2*RSe465p - 5./1152.*pow(m33,2)*me2*RSe456p + 13.
         /384.*pow(m33,2)*m66 + 23./1152.*pow(m33,2)*m66*RS34561pe + 11.
         /1152.*pow(m33,2)*m66*RS14563pe + 11./1152.*pow(m33,2)*m66*
         RSe45613p + 11./1152.*pow(m33,2)*m66*RSp45613e + 77./3456.*
         pow(m33,2)*m66*RS3456pee );

      M13p6x23R +=  + pi162 * ( 77./3456.*pow(m33,2)*m66*RS3456ppe - 1./
         108.*pow(m33,2)*m66*RSe45633p - 1./108.*pow(m33,2)*m66*
         RSp45633e + 1./432.*pow(m33,2)*m66*RS345pe - 77./3456.*pow(
         m33,2)*m66*RS346pe - 77./3456.*pow(m33,2)*m66*RS356pe + 31./
         3456.*pow(m33,2)*m66*RSe453p + 31./3456.*pow(m33,2)*m66*
         RSp453e + 5./3456.*pow(m33,2)*m66*RS1456pee + 5./3456.*pow(
         m33,2)*m66*RS1456ppe - 5./3456.*pow(m33,2)*m66*RSe45611p - 5./
         3456.*pow(m33,2)*m66*RSp45611e - 7./864.*pow(m33,2)*m66*
         RS145pe - 5./3456.*pow(m33,2)*m66*RS146pe - 5./3456.*pow(
         m33,2)*m66*RS156pe - 19./3456.*pow(m33,2)*m66*RS645pe - 23./
         3456.*pow(m33,2)*m66*RSe451p - 23./3456.*pow(m33,2)*m66*
         RSp451e - 1./864.*pow(m33,2)*m66*RSe456p - 1./864.*pow(m33,2)*
         m66*RSp456e + 13./384.*pow(m33,2)*m55 + 23./1152.*pow(m33,2)*
         m55*RS34561pe + 11./1152.*pow(m33,2)*m55*RS14563pe + 11./1152.
         *pow(m33,2)*m55*RSe45613p + 11./1152.*pow(m33,2)*m55*RSp45613e
          + 77./3456.*pow(m33,2)*m55*RS3456pee );

      M13p6x23R +=  + pi162 * ( 77./3456.*pow(m33,2)*m55*RS3456ppe - 1./
         108.*pow(m33,2)*m55*RSe45633p - 1./108.*pow(m33,2)*m55*
         RSp45633e - 77./3456.*pow(m33,2)*m55*RS345pe + 1./432.*pow(
         m33,2)*m55*RS346pe - 77./3456.*pow(m33,2)*m55*RS356pe + 31./
         3456.*pow(m33,2)*m55*RSe463p + 31./3456.*pow(m33,2)*m55*
         RSp463e + 5./3456.*pow(m33,2)*m55*RS1456pee + 5./3456.*pow(
         m33,2)*m55*RS1456ppe - 5./3456.*pow(m33,2)*m55*RSe45611p - 5./
         3456.*pow(m33,2)*m55*RSp45611e - 5./3456.*pow(m33,2)*m55*
         RS145pe - 7./864.*pow(m33,2)*m55*RS146pe - 5./3456.*pow(m33,2)
         *m55*RS156pe - 19./3456.*pow(m33,2)*m55*RS546pe - 23./3456.*
         pow(m33,2)*m55*RSe461p - 23./3456.*pow(m33,2)*m55*RSp461e - 1./
         864.*pow(m33,2)*m55*RSe465p - 1./864.*pow(m33,2)*m55*RSp465e
          + 13./384.*pow(m33,2)*m44 + 23./1152.*pow(m33,2)*m44*
         RS34561pe + 11./1152.*pow(m33,2)*m44*RS14563pe + 11./1152.*
         pow(m33,2)*m44*RSe45613p + 11./1152.*pow(m33,2)*m44*RSp45613e
          + 77./3456.*pow(m33,2)*m44*RS3456pee );

      M13p6x23R +=  + pi162 * ( 77./3456.*pow(m33,2)*m44*RS3456ppe - 1./
         108.*pow(m33,2)*m44*RSe45633p - 1./108.*pow(m33,2)*m44*
         RSp45633e - 77./3456.*pow(m33,2)*m44*RS345pe - 77./3456.*pow(
         m33,2)*m44*RS346pe + 1./432.*pow(m33,2)*m44*RS356pe + 31./3456.
         *pow(m33,2)*m44*RSe563p + 31./3456.*pow(m33,2)*m44*RSp563e + 5.
         /3456.*pow(m33,2)*m44*RS1456pee + 5./3456.*pow(m33,2)*m44*
         RS1456ppe - 5./3456.*pow(m33,2)*m44*RSe45611p - 5./3456.*pow(
         m33,2)*m44*RSp45611e - 5./3456.*pow(m33,2)*m44*RS145pe - 5./
         3456.*pow(m33,2)*m44*RS146pe - 7./864.*pow(m33,2)*m44*RS156pe
          - 19./3456.*pow(m33,2)*m44*RS456pe - 23./3456.*pow(m33,2)*m44
         *RSe561p - 23./3456.*pow(m33,2)*m44*RSp561e - 1./864.*pow(
         m33,2)*m44*RSe564p - 1./864.*pow(m33,2)*m44*RSp564e - 47./512.
         *pow(m33,3) - 55./4608.*pow(m33,3)*RS34561pe + 353./20736.*
         pow(m33,3)*pow(RS34561pe,2) + 11./1536.*pow(m33,3)*RS14563pe
          + 121./10368.*pow(m33,3)*RS14563pe*RS34561pe + 11./6912.*pow(
         m33,3)*pow(RS14563pe,2) );

      M13p6x23R +=  + pi162 * ( 11./1536.*pow(m33,3)*RSe45613p + 121./
         10368.*pow(m33,3)*RSe45613p*RS34561pe + 11./3456.*pow(m33,3)*
         RSe45613p*RS14563pe + 11./6912.*pow(m33,3)*pow(RSe45613p,2) + 
         11./1536.*pow(m33,3)*RSp45613e + 121./10368.*pow(m33,3)*
         RSp45613e*RS34561pe + 11./3456.*pow(m33,3)*RSp45613e*RS14563pe
          + 11./3456.*pow(m33,3)*RSp45613e*RSe45613p + 11./6912.*pow(
         m33,3)*pow(RSp45613e,2) + 73./4608.*pow(m33,3)*RS3456pee + 353.
         /20736.*pow(m33,3)*RS3456pee*RS34561pe + 121./20736.*pow(
         m33,3)*RS3456pee*RS14563pe + 121./20736.*pow(m33,3)*RS3456pee*
         RSe45613p + 121./20736.*pow(m33,3)*RS3456pee*RSp45613e + 353./
         82944.*pow(m33,3)*pow(RS3456pee,2) + 73./4608.*pow(m33,3)*
         RS3456ppe + 353./20736.*pow(m33,3)*RS3456ppe*RS34561pe + 121./
         20736.*pow(m33,3)*RS3456ppe*RS14563pe + 121./20736.*pow(m33,3)
         *RS3456ppe*RSe45613p + 121./20736.*pow(m33,3)*RS3456ppe*
         RSp45613e + 353./41472.*pow(m33,3)*RS3456ppe*RS3456pee + 353./
         82944.*pow(m33,3)*pow(RS3456ppe,2) );

      M13p6x23R +=  + pi162 * (  - 11./1536.*pow(m33,3)*RSe45633p - 121.
         /20736.*pow(m33,3)*RSe45633p*RS34561pe - 11./6912.*pow(m33,3)*
         RSe45633p*RS14563pe - 11./6912.*pow(m33,3)*RSe45633p*RSe45613p
          - 11./6912.*pow(m33,3)*RSe45633p*RSp45613e - 121./41472.*pow(
         m33,3)*RSe45633p*RS3456pee - 121./41472.*pow(m33,3)*RSe45633p*
         RS3456ppe + 11./27648.*pow(m33,3)*pow(RSe45633p,2) - 11./1536.
         *pow(m33,3)*RSp45633e - 121./20736.*pow(m33,3)*RSp45633e*
         RS34561pe - 11./6912.*pow(m33,3)*RSp45633e*RS14563pe - 11./
         6912.*pow(m33,3)*RSp45633e*RSe45613p - 11./6912.*pow(m33,3)*
         RSp45633e*RSp45613e - 121./41472.*pow(m33,3)*RSp45633e*
         RS3456pee - 121./41472.*pow(m33,3)*RSp45633e*RS3456ppe + 11./
         13824.*pow(m33,3)*RSp45633e*RSe45633p + 11./27648.*pow(m33,3)*
         pow(RSp45633e,2) + 7./768.*pow(m33,3)*RS345pe - 353./20736.*
         pow(m33,3)*RS345pe*RS34561pe - 121./20736.*pow(m33,3)*RS345pe*
         RS14563pe - 121./20736.*pow(m33,3)*RS345pe*RSe45613p - 121./
         20736.*pow(m33,3)*RS345pe*RSp45613e );

      M13p6x23R +=  + pi162 * (  - 353./41472.*pow(m33,3)*RS345pe*
         RS3456pee - 353./41472.*pow(m33,3)*RS345pe*RS3456ppe + 121./
         41472.*pow(m33,3)*RS345pe*RSe45633p + 121./41472.*pow(m33,3)*
         RS345pe*RSp45633e + 353./82944.*pow(m33,3)*pow(RS345pe,2) + 7./
         768.*pow(m33,3)*RS346pe - 353./20736.*pow(m33,3)*RS346pe*
         RS34561pe - 121./20736.*pow(m33,3)*RS346pe*RS14563pe - 121./
         20736.*pow(m33,3)*RS346pe*RSe45613p - 121./20736.*pow(m33,3)*
         RS346pe*RSp45613e - 353./41472.*pow(m33,3)*RS346pe*RS3456pee
          - 353./41472.*pow(m33,3)*RS346pe*RS3456ppe + 121./41472.*pow(
         m33,3)*RS346pe*RSe45633p + 121./41472.*pow(m33,3)*RS346pe*
         RSp45633e + 353./41472.*pow(m33,3)*RS346pe*RS345pe + 353./
         82944.*pow(m33,3)*pow(RS346pe,2) + 7./768.*pow(m33,3)*RS356pe
          - 353./20736.*pow(m33,3)*RS356pe*RS34561pe - 121./20736.*pow(
         m33,3)*RS356pe*RS14563pe - 121./20736.*pow(m33,3)*RS356pe*
         RSe45613p - 121./20736.*pow(m33,3)*RS356pe*RSp45613e - 353./
         41472.*pow(m33,3)*RS356pe*RS3456pee );

      M13p6x23R +=  + pi162 * (  - 353./41472.*pow(m33,3)*RS356pe*
         RS3456ppe + 121./41472.*pow(m33,3)*RS356pe*RSe45633p + 121./
         41472.*pow(m33,3)*RS356pe*RSp45633e + 353./41472.*pow(m33,3)*
         RS356pe*RS345pe + 353./41472.*pow(m33,3)*RS356pe*RS346pe + 353.
         /82944.*pow(m33,3)*pow(RS356pe,2) + 11./4608.*pow(m33,3)*
         RSe563p + 11./4608.*pow(m33,3)*RSe463p + 11./4608.*pow(m33,3)*
         RSe453p + 11./4608.*pow(m33,3)*RSp563e + 11./4608.*pow(m33,3)*
         RSp463e + 11./4608.*pow(m33,3)*RSp453e - 7./4608.*pow(m33,3)*
         RS1456pee + 121./20736.*pow(m33,3)*RS1456pee*RS34561pe + 11./
         6912.*pow(m33,3)*RS1456pee*RS14563pe + 11./6912.*pow(m33,3)*
         RS1456pee*RSe45613p + 11./6912.*pow(m33,3)*RS1456pee*RSp45613e
          + 121./41472.*pow(m33,3)*RS1456pee*RS3456pee + 121./41472.*
         pow(m33,3)*RS1456pee*RS3456ppe - 11./13824.*pow(m33,3)*
         RS1456pee*RSe45633p - 11./13824.*pow(m33,3)*RS1456pee*
         RSp45633e - 121./41472.*pow(m33,3)*RS1456pee*RS345pe - 121./
         41472.*pow(m33,3)*RS1456pee*RS346pe );

      M13p6x23R +=  + pi162 * (  - 121./41472.*pow(m33,3)*RS1456pee*
         RS356pe + 11./27648.*pow(m33,3)*pow(RS1456pee,2) - 7./4608.*
         pow(m33,3)*RS1456ppe + 121./20736.*pow(m33,3)*RS1456ppe*
         RS34561pe + 11./6912.*pow(m33,3)*RS1456ppe*RS14563pe + 11./
         6912.*pow(m33,3)*RS1456ppe*RSe45613p + 11./6912.*pow(m33,3)*
         RS1456ppe*RSp45613e + 121./41472.*pow(m33,3)*RS1456ppe*
         RS3456pee + 121./41472.*pow(m33,3)*RS1456ppe*RS3456ppe - 11./
         13824.*pow(m33,3)*RS1456ppe*RSe45633p - 11./13824.*pow(m33,3)*
         RS1456ppe*RSp45633e - 121./41472.*pow(m33,3)*RS1456ppe*RS345pe
          - 121./41472.*pow(m33,3)*RS1456ppe*RS346pe - 121./41472.*pow(
         m33,3)*RS1456ppe*RS356pe + 11./13824.*pow(m33,3)*RS1456ppe*
         RS1456pee + 11./27648.*pow(m33,3)*pow(RS1456ppe,2) + 7./4608.*
         pow(m33,3)*RSe45611p - 121./20736.*pow(m33,3)*RSe45611p*
         RS34561pe - 11./6912.*pow(m33,3)*RSe45611p*RS14563pe - 11./
         6912.*pow(m33,3)*RSe45611p*RSe45613p - 11./6912.*pow(m33,3)*
         RSe45611p*RSp45613e );

      M13p6x23R +=  + pi162 * (  - 121./41472.*pow(m33,3)*RSe45611p*
         RS3456pee - 121./41472.*pow(m33,3)*RSe45611p*RS3456ppe + 11./
         13824.*pow(m33,3)*RSe45611p*RSe45633p + 11./13824.*pow(m33,3)*
         RSe45611p*RSp45633e + 121./41472.*pow(m33,3)*RSe45611p*RS345pe
          + 121./41472.*pow(m33,3)*RSe45611p*RS346pe + 121./41472.*pow(
         m33,3)*RSe45611p*RS356pe - 11./13824.*pow(m33,3)*RSe45611p*
         RS1456pee - 11./13824.*pow(m33,3)*RSe45611p*RS1456ppe + 11./
         27648.*pow(m33,3)*pow(RSe45611p,2) + 7./4608.*pow(m33,3)*
         RSp45611e - 121./20736.*pow(m33,3)*RSp45611e*RS34561pe - 11./
         6912.*pow(m33,3)*RSp45611e*RS14563pe - 11./6912.*pow(m33,3)*
         RSp45611e*RSe45613p - 11./6912.*pow(m33,3)*RSp45611e*RSp45613e
          - 121./41472.*pow(m33,3)*RSp45611e*RS3456pee - 121./41472.*
         pow(m33,3)*RSp45611e*RS3456ppe + 11./13824.*pow(m33,3)*
         RSp45611e*RSe45633p + 11./13824.*pow(m33,3)*RSp45611e*
         RSp45633e + 121./41472.*pow(m33,3)*RSp45611e*RS345pe + 121./
         41472.*pow(m33,3)*RSp45611e*RS346pe );

      M13p6x23R +=  + pi162 * ( 121./41472.*pow(m33,3)*RSp45611e*
         RS356pe - 11./13824.*pow(m33,3)*RSp45611e*RS1456pee - 11./
         13824.*pow(m33,3)*RSp45611e*RS1456ppe + 11./13824.*pow(m33,3)*
         RSp45611e*RSe45611p + 11./27648.*pow(m33,3)*pow(RSp45611e,2)
          - 13./6912.*pow(m33,3)*RS145pe - 121./20736.*pow(m33,3)*
         RS145pe*RS34561pe - 11./6912.*pow(m33,3)*RS145pe*RS14563pe - 
         11./6912.*pow(m33,3)*RS145pe*RSe45613p - 11./6912.*pow(m33,3)*
         RS145pe*RSp45613e - 121./41472.*pow(m33,3)*RS145pe*RS3456pee
          - 121./41472.*pow(m33,3)*RS145pe*RS3456ppe + 11./13824.*pow(
         m33,3)*RS145pe*RSe45633p + 11./13824.*pow(m33,3)*RS145pe*
         RSp45633e + 121./41472.*pow(m33,3)*RS145pe*RS345pe + 121./
         41472.*pow(m33,3)*RS145pe*RS346pe + 121./41472.*pow(m33,3)*
         RS145pe*RS356pe - 11./13824.*pow(m33,3)*RS145pe*RS1456pee - 11.
         /13824.*pow(m33,3)*RS145pe*RS1456ppe + 11./13824.*pow(m33,3)*
         RS145pe*RSe45611p + 11./13824.*pow(m33,3)*RS145pe*RSp45611e + 
         11./27648.*pow(m33,3)*pow(RS145pe,2) );

      M13p6x23R +=  + pi162 * (  - 13./6912.*pow(m33,3)*RS146pe - 121./
         20736.*pow(m33,3)*RS146pe*RS34561pe - 11./6912.*pow(m33,3)*
         RS146pe*RS14563pe - 11./6912.*pow(m33,3)*RS146pe*RSe45613p - 
         11./6912.*pow(m33,3)*RS146pe*RSp45613e - 121./41472.*pow(
         m33,3)*RS146pe*RS3456pee - 121./41472.*pow(m33,3)*RS146pe*
         RS3456ppe + 11./13824.*pow(m33,3)*RS146pe*RSe45633p + 11./
         13824.*pow(m33,3)*RS146pe*RSp45633e + 121./41472.*pow(m33,3)*
         RS146pe*RS345pe + 121./41472.*pow(m33,3)*RS146pe*RS346pe + 121.
         /41472.*pow(m33,3)*RS146pe*RS356pe - 11./13824.*pow(m33,3)*
         RS146pe*RS1456pee - 11./13824.*pow(m33,3)*RS146pe*RS1456ppe + 
         11./13824.*pow(m33,3)*RS146pe*RSe45611p + 11./13824.*pow(
         m33,3)*RS146pe*RSp45611e + 11./13824.*pow(m33,3)*RS146pe*
         RS145pe + 11./27648.*pow(m33,3)*pow(RS146pe,2) - 13./6912.*
         pow(m33,3)*RS156pe - 121./20736.*pow(m33,3)*RS156pe*RS34561pe
          - 11./6912.*pow(m33,3)*RS156pe*RS14563pe - 11./6912.*pow(
         m33,3)*RS156pe*RSe45613p );

      M13p6x23R +=  + pi162 * (  - 11./6912.*pow(m33,3)*RS156pe*
         RSp45613e - 121./41472.*pow(m33,3)*RS156pe*RS3456pee - 121./
         41472.*pow(m33,3)*RS156pe*RS3456ppe + 11./13824.*pow(m33,3)*
         RS156pe*RSe45633p + 11./13824.*pow(m33,3)*RS156pe*RSp45633e + 
         121./41472.*pow(m33,3)*RS156pe*RS345pe + 121./41472.*pow(
         m33,3)*RS156pe*RS346pe + 121./41472.*pow(m33,3)*RS156pe*
         RS356pe - 11./13824.*pow(m33,3)*RS156pe*RS1456pee - 11./13824.
         *pow(m33,3)*RS156pe*RS1456ppe + 11./13824.*pow(m33,3)*RS156pe*
         RSe45611p + 11./13824.*pow(m33,3)*RS156pe*RSp45611e + 11./
         13824.*pow(m33,3)*RS156pe*RS145pe + 11./13824.*pow(m33,3)*
         RS156pe*RS146pe + 11./27648.*pow(m33,3)*pow(RS156pe,2) + 7./
         13824.*pow(m33,3)*RS645pe + 7./13824.*pow(m33,3)*RS546pe + 7./
         13824.*pow(m33,3)*RS456pe - 47./13824.*pow(m33,3)*RSe561p - 47.
         /13824.*pow(m33,3)*RSe461p - 47./13824.*pow(m33,3)*RSe451p - 
         47./13824.*pow(m33,3)*RSp561e - 47./13824.*pow(m33,3)*RSp461e
          - 47./13824.*pow(m33,3)*RSp451e );

      M13p6x23R +=  + pi162 * ( 7./13824.*pow(m33,3)*RSe564p + 7./13824.
         *pow(m33,3)*RSe465p + 7./13824.*pow(m33,3)*RSe456p + 7./13824.
         *pow(m33,3)*RSp564e + 7./13824.*pow(m33,3)*RSp465e + 7./13824.
         *pow(m33,3)*RSp456e + 1./576.*m11*pow(RD3456ep,2) + 1./288.*
         m11*RD1456ep*RD3456ep + 1./576.*m11*pow(RD1456ep,2) - 1./144.*
         m11*mp2*RD3456ep*RSp45613e + 1./288.*m11*mp2*RSp45633e*
         RD3456ep - 1./144.*m11*mp2*RD1456ep*RSp45613e + 1./288.*m11*
         mp2*RD1456ep*RSp45633e + 1./288.*m11*mp2*RSp45611e*RD3456ep + 
         1./288.*m11*mp2*RSp45611e*RD1456ep - 7./192.*m11*pow(mp2,2)*
         RSp45613e + 1./144.*m11*pow(mp2,2)*pow(RSp45613e,2) + 1./96.*
         m11*pow(mp2,2)*RSp45633e - 1./144.*m11*pow(mp2,2)*RSp45633e*
         RSp45613e + 1./576.*m11*pow(mp2,2)*pow(RSp45633e,2) + 1./64.*
         m11*pow(mp2,2)*RSp563e + 1./64.*m11*pow(mp2,2)*RSp463e + 1./64.
         *m11*pow(mp2,2)*RSp453e + 1./96.*m11*pow(mp2,2)*RSp45611e - 1./
         144.*m11*pow(mp2,2)*RSp45611e*RSp45613e + 1./288.*m11*pow(
         mp2,2)*RSp45611e*RSp45633e );

      M13p6x23R +=  + pi162 * ( 1./576.*m11*pow(mp2,2)*pow(RSp45611e,2)
          + 1./64.*m11*pow(mp2,2)*RSp561e + 1./64.*m11*pow(mp2,2)*
         RSp461e + 1./64.*m11*pow(mp2,2)*RSp451e - 1./192.*m11*pow(
         mp2,2)*RSp564e - 1./192.*m11*pow(mp2,2)*RSp465e - 1./192.*m11*
         pow(mp2,2)*RSp456e - 1./144.*m11*me2*RD3456ep*RSe45613p + 1./
         288.*m11*me2*RSe45633p*RD3456ep - 1./144.*m11*me2*RD1456ep*
         RSe45613p + 1./288.*m11*me2*RD1456ep*RSe45633p + 1./288.*m11*
         me2*RSe45611p*RD3456ep + 1./288.*m11*me2*RSe45611p*RD1456ep + 
         1./72.*m11*me2*mp2*RSp45613e*RSe45613p - 1./144.*m11*me2*mp2*
         RSe45633p*RSp45613e - 1./144.*m11*me2*mp2*RSp45633e*RSe45613p
          + 1./288.*m11*me2*mp2*RSp45633e*RSe45633p - 1./144.*m11*me2*
         mp2*RSe45611p*RSp45613e + 1./288.*m11*me2*mp2*RSe45611p*
         RSp45633e - 1./144.*m11*me2*mp2*RSp45611e*RSe45613p + 1./288.*
         m11*me2*mp2*RSp45611e*RSe45633p + 1./288.*m11*me2*mp2*
         RSp45611e*RSe45611p - 7./192.*m11*pow(me2,2)*RSe45613p + 1./
         144.*m11*pow(me2,2)*pow(RSe45613p,2) );

      M13p6x23R +=  + pi162 * ( 1./96.*m11*pow(me2,2)*RSe45633p - 1./
         144.*m11*pow(me2,2)*RSe45633p*RSe45613p + 1./576.*m11*pow(
         me2,2)*pow(RSe45633p,2) + 1./64.*m11*pow(me2,2)*RSe563p + 1./
         64.*m11*pow(me2,2)*RSe463p + 1./64.*m11*pow(me2,2)*RSe453p + 1.
         /96.*m11*pow(me2,2)*RSe45611p - 1./144.*m11*pow(me2,2)*
         RSe45611p*RSe45613p + 1./288.*m11*pow(me2,2)*RSe45611p*
         RSe45633p + 1./576.*m11*pow(me2,2)*pow(RSe45611p,2) + 1./64.*
         m11*pow(me2,2)*RSe561p + 1./64.*m11*pow(me2,2)*RSe461p + 1./64.
         *m11*pow(me2,2)*RSe451p - 1./192.*m11*pow(me2,2)*RSe564p - 1./
         192.*m11*pow(me2,2)*RSe465p - 1./192.*m11*pow(me2,2)*RSe456p
          - 5./384.*m11*m66*RD3456ep - 5./384.*m11*m66*RD1456ep + 1./96.
         *m11*m66*mp2*RSp45613e - 5./384.*m11*m66*mp2*RSp45633e + 1./64.
         *m11*m66*mp2*RSp453e - 5./384.*m11*m66*mp2*RSp45611e + 1./64.*
         m11*m66*mp2*RSp451e - 1./64.*m11*m66*mp2*RSp456e + 1./96.*m11*
         m66*me2*RSe45613p - 5./384.*m11*m66*me2*RSe45633p + 1./64.*m11
         *m66*me2*RSe453p );

      M13p6x23R +=  + pi162 * (  - 5./384.*m11*m66*me2*RSe45611p + 1./
         64.*m11*m66*me2*RSe451p - 1./64.*m11*m66*me2*RSe456p + 15./128.
         *m11*pow(m66,2) + 5./384.*m11*pow(m66,2)*RS34561pe + 5./384.*
         m11*pow(m66,2)*RS14563pe + 5./384.*m11*pow(m66,2)*RSe45613p + 
         5./384.*m11*pow(m66,2)*RSp45613e + 1./128.*m11*pow(m66,2)*
         RS3456pee + 1./128.*m11*pow(m66,2)*RS3456ppe - 1./128.*m11*
         pow(m66,2)*RSe45633p - 1./128.*m11*pow(m66,2)*RSp45633e - 1./
         192.*m11*pow(m66,2)*RS345pe - 1./128.*m11*pow(m66,2)*RS346pe
          - 1./128.*m11*pow(m66,2)*RS356pe + 1./384.*m11*pow(m66,2)*
         RSe453p + 1./384.*m11*pow(m66,2)*RSp453e + 1./128.*m11*pow(
         m66,2)*RS1456pee + 1./128.*m11*pow(m66,2)*RS1456ppe - 1./128.*
         m11*pow(m66,2)*RSe45611p - 1./128.*m11*pow(m66,2)*RSp45611e - 
         1./192.*m11*pow(m66,2)*RS145pe - 1./128.*m11*pow(m66,2)*
         RS146pe - 1./128.*m11*pow(m66,2)*RS156pe - 3./128.*m11*pow(
         m66,2)*RS645pe + 1./384.*m11*pow(m66,2)*RSe451p + 1./384.*m11*
         pow(m66,2)*RSp451e );

      M13p6x23R +=  + pi162 * (  - 1./384.*m11*pow(m66,2)*RSe456p - 1./
         384.*m11*pow(m66,2)*RSp456e - 5./384.*m11*m55*RD3456ep - 5./
         384.*m11*m55*RD1456ep + 1./96.*m11*m55*mp2*RSp45613e - 5./384.
         *m11*m55*mp2*RSp45633e + 1./64.*m11*m55*mp2*RSp463e - 5./384.*
         m11*m55*mp2*RSp45611e + 1./64.*m11*m55*mp2*RSp461e - 1./64.*
         m11*m55*mp2*RSp465e + 1./96.*m11*m55*me2*RSe45613p - 5./384.*
         m11*m55*me2*RSe45633p + 1./64.*m11*m55*me2*RSe463p - 5./384.*
         m11*m55*me2*RSe45611p + 1./64.*m11*m55*me2*RSe461p - 1./64.*
         m11*m55*me2*RSe465p + 5./64.*m11*m55*m66 + 15./128.*m11*pow(
         m55,2) + 5./384.*m11*pow(m55,2)*RS34561pe + 5./384.*m11*pow(
         m55,2)*RS14563pe + 5./384.*m11*pow(m55,2)*RSe45613p + 5./384.*
         m11*pow(m55,2)*RSp45613e + 1./128.*m11*pow(m55,2)*RS3456pee + 
         1./128.*m11*pow(m55,2)*RS3456ppe - 1./128.*m11*pow(m55,2)*
         RSe45633p - 1./128.*m11*pow(m55,2)*RSp45633e - 1./128.*m11*
         pow(m55,2)*RS345pe - 1./192.*m11*pow(m55,2)*RS346pe - 1./128.*
         m11*pow(m55,2)*RS356pe );

      M13p6x23R +=  + pi162 * ( 1./384.*m11*pow(m55,2)*RSe463p + 1./384.
         *m11*pow(m55,2)*RSp463e + 1./128.*m11*pow(m55,2)*RS1456pee + 1.
         /128.*m11*pow(m55,2)*RS1456ppe - 1./128.*m11*pow(m55,2)*
         RSe45611p - 1./128.*m11*pow(m55,2)*RSp45611e - 1./128.*m11*
         pow(m55,2)*RS145pe - 1./192.*m11*pow(m55,2)*RS146pe - 1./128.*
         m11*pow(m55,2)*RS156pe - 3./128.*m11*pow(m55,2)*RS546pe + 1./
         384.*m11*pow(m55,2)*RSe461p + 1./384.*m11*pow(m55,2)*RSp461e
          - 1./384.*m11*pow(m55,2)*RSe465p - 1./384.*m11*pow(m55,2)*
         RSp465e - 5./384.*m11*m44*RD3456ep - 5./384.*m11*m44*RD1456ep
          + 1./96.*m11*m44*mp2*RSp45613e - 5./384.*m11*m44*mp2*
         RSp45633e + 1./64.*m11*m44*mp2*RSp563e - 5./384.*m11*m44*mp2*
         RSp45611e + 1./64.*m11*m44*mp2*RSp561e - 1./64.*m11*m44*mp2*
         RSp564e + 1./96.*m11*m44*me2*RSe45613p - 5./384.*m11*m44*me2*
         RSe45633p + 1./64.*m11*m44*me2*RSe563p - 5./384.*m11*m44*me2*
         RSe45611p + 1./64.*m11*m44*me2*RSe561p - 1./64.*m11*m44*me2*
         RSe564p );

      M13p6x23R +=  + pi162 * ( 5./64.*m11*m44*m66 + 5./64.*m11*m44*m55
          + 15./128.*m11*pow(m44,2) + 5./384.*m11*pow(m44,2)*RS34561pe
          + 5./384.*m11*pow(m44,2)*RS14563pe + 5./384.*m11*pow(m44,2)*
         RSe45613p + 5./384.*m11*pow(m44,2)*RSp45613e + 1./128.*m11*
         pow(m44,2)*RS3456pee + 1./128.*m11*pow(m44,2)*RS3456ppe - 1./
         128.*m11*pow(m44,2)*RSe45633p - 1./128.*m11*pow(m44,2)*
         RSp45633e - 1./128.*m11*pow(m44,2)*RS345pe - 1./128.*m11*pow(
         m44,2)*RS346pe - 1./192.*m11*pow(m44,2)*RS356pe + 1./384.*m11*
         pow(m44,2)*RSe563p + 1./384.*m11*pow(m44,2)*RSp563e + 1./128.*
         m11*pow(m44,2)*RS1456pee + 1./128.*m11*pow(m44,2)*RS1456ppe - 
         1./128.*m11*pow(m44,2)*RSe45611p - 1./128.*m11*pow(m44,2)*
         RSp45611e - 1./128.*m11*pow(m44,2)*RS145pe - 1./128.*m11*pow(
         m44,2)*RS146pe - 1./192.*m11*pow(m44,2)*RS156pe - 3./128.*m11*
         pow(m44,2)*RS456pe + 1./384.*m11*pow(m44,2)*RSe561p + 1./384.*
         m11*pow(m44,2)*RSp561e - 1./384.*m11*pow(m44,2)*RSe564p - 1./
         384.*m11*pow(m44,2)*RSp564e );

      M13p6x23R +=  + pi162 * (  - 11./1152.*m11*m33*RD3456ep - 5./324.
         *m11*m33*RD3456ep*RS34561pe - 5./324.*m11*m33*RD3456ep*
         RS14563pe - 11./1296.*m11*m33*RD3456ep*RSe45613p - 11./1296.*
         m11*m33*RD3456ep*RSp45613e - 5./648.*m11*m33*RS3456pee*
         RD3456ep - 5./648.*m11*m33*RS3456ppe*RD3456ep + 11./2592.*m11*
         m33*RSe45633p*RD3456ep + 11./2592.*m11*m33*RSp45633e*RD3456ep
          + 5./648.*m11*m33*RS345pe*RD3456ep + 5./648.*m11*m33*RS346pe*
         RD3456ep + 5./648.*m11*m33*RS356pe*RD3456ep - 11./1152.*m11*
         m33*RD1456ep - 5./324.*m11*m33*RD1456ep*RS34561pe - 5./324.*
         m11*m33*RD1456ep*RS14563pe - 11./1296.*m11*m33*RD1456ep*
         RSe45613p - 11./1296.*m11*m33*RD1456ep*RSp45613e - 5./648.*m11
         *m33*RD1456ep*RS3456pee - 5./648.*m11*m33*RD1456ep*RS3456ppe
          + 11./2592.*m11*m33*RD1456ep*RSe45633p + 11./2592.*m11*m33*
         RD1456ep*RSp45633e + 5./648.*m11*m33*RD1456ep*RS345pe + 5./648.
         *m11*m33*RD1456ep*RS346pe + 5./648.*m11*m33*RD1456ep*RS356pe
          - 5./648.*m11*m33*RS1456pee*RD3456ep );

      M13p6x23R +=  + pi162 * (  - 5./648.*m11*m33*RS1456pee*RD1456ep
          - 5./648.*m11*m33*RS1456ppe*RD3456ep - 5./648.*m11*m33*
         RS1456ppe*RD1456ep + 11./2592.*m11*m33*RSe45611p*RD3456ep + 11.
         /2592.*m11*m33*RSe45611p*RD1456ep + 11./2592.*m11*m33*
         RSp45611e*RD3456ep + 11./2592.*m11*m33*RSp45611e*RD1456ep + 5./
         648.*m11*m33*RS145pe*RD3456ep + 5./648.*m11*m33*RS145pe*
         RD1456ep + 5./648.*m11*m33*RS146pe*RD3456ep + 5./648.*m11*m33*
         RS146pe*RD1456ep + 5./648.*m11*m33*RS156pe*RD3456ep + 5./648.*
         m11*m33*RS156pe*RD1456ep + 5./144.*m11*m33*mp2*RSp45613e + 5./
         162.*m11*m33*mp2*RSp45613e*RS34561pe + 5./162.*m11*m33*mp2*
         RSp45613e*RS14563pe + 11./648.*m11*m33*mp2*RSp45613e*RSe45613p
          + 11./648.*m11*m33*mp2*pow(RSp45613e,2) + 5./324.*m11*m33*mp2
         *RS3456pee*RSp45613e + 5./324.*m11*m33*mp2*RS3456ppe*RSp45613e
          - 11./1296.*m11*m33*mp2*RSe45633p*RSp45613e - 35./1152.*m11*
         m33*mp2*RSp45633e - 5./324.*m11*m33*mp2*RSp45633e*RS34561pe - 
         5./324.*m11*m33*mp2*RSp45633e*RS14563pe );

      M13p6x23R +=  + pi162 * (  - 11./1296.*m11*m33*mp2*RSp45633e*
         RSe45613p - 11./648.*m11*m33*mp2*RSp45633e*RSp45613e - 5./648.
         *m11*m33*mp2*RSp45633e*RS3456pee - 5./648.*m11*m33*mp2*
         RSp45633e*RS3456ppe + 11./2592.*m11*m33*mp2*RSp45633e*
         RSe45633p + 11./2592.*m11*m33*mp2*pow(RSp45633e,2) - 5./324.*
         m11*m33*mp2*RS345pe*RSp45613e + 5./648.*m11*m33*mp2*RS345pe*
         RSp45633e - 5./324.*m11*m33*mp2*RS346pe*RSp45613e + 5./648.*
         m11*m33*mp2*RS346pe*RSp45633e - 5./324.*m11*m33*mp2*RS356pe*
         RSp45613e + 5./648.*m11*m33*mp2*RS356pe*RSp45633e + 5./576.*
         m11*m33*mp2*RSp563e + 5./576.*m11*m33*mp2*RSp463e + 5./576.*
         m11*m33*mp2*RSp453e + 5./324.*m11*m33*mp2*RS1456pee*RSp45613e
          - 5./648.*m11*m33*mp2*RS1456pee*RSp45633e + 5./324.*m11*m33*
         mp2*RS1456ppe*RSp45613e - 5./648.*m11*m33*mp2*RS1456ppe*
         RSp45633e - 11./1296.*m11*m33*mp2*RSe45611p*RSp45613e + 11./
         2592.*m11*m33*mp2*RSe45611p*RSp45633e - 35./1152.*m11*m33*mp2*
         RSp45611e );

      M13p6x23R +=  + pi162 * (  - 5./324.*m11*m33*mp2*RSp45611e*
         RS34561pe - 5./324.*m11*m33*mp2*RSp45611e*RS14563pe - 11./1296.
         *m11*m33*mp2*RSp45611e*RSe45613p - 11./648.*m11*m33*mp2*
         RSp45611e*RSp45613e - 5./648.*m11*m33*mp2*RSp45611e*RS3456pee
          - 5./648.*m11*m33*mp2*RSp45611e*RS3456ppe + 11./2592.*m11*m33
         *mp2*RSp45611e*RSe45633p + 11./1296.*m11*m33*mp2*RSp45611e*
         RSp45633e + 5./648.*m11*m33*mp2*RSp45611e*RS345pe + 5./648.*
         m11*m33*mp2*RSp45611e*RS346pe + 5./648.*m11*m33*mp2*RSp45611e*
         RS356pe - 5./648.*m11*m33*mp2*RSp45611e*RS1456pee - 5./648.*
         m11*m33*mp2*RSp45611e*RS1456ppe + 11./2592.*m11*m33*mp2*
         RSp45611e*RSe45611p + 11./2592.*m11*m33*mp2*pow(RSp45611e,2)
          - 5./324.*m11*m33*mp2*RS145pe*RSp45613e + 5./648.*m11*m33*mp2
         *RS145pe*RSp45633e + 5./648.*m11*m33*mp2*RS145pe*RSp45611e - 5.
         /324.*m11*m33*mp2*RS146pe*RSp45613e + 5./648.*m11*m33*mp2*
         RS146pe*RSp45633e + 5./648.*m11*m33*mp2*RS146pe*RSp45611e - 5./
         324.*m11*m33*mp2*RS156pe*RSp45613e );

      M13p6x23R +=  + pi162 * ( 5./648.*m11*m33*mp2*RS156pe*RSp45633e
          + 5./648.*m11*m33*mp2*RS156pe*RSp45611e + 5./576.*m11*m33*mp2
         *RSp561e + 5./576.*m11*m33*mp2*RSp461e + 5./576.*m11*m33*mp2*
         RSp451e - 5./576.*m11*m33*mp2*RSp564e - 5./576.*m11*m33*mp2*
         RSp465e - 5./576.*m11*m33*mp2*RSp456e + 5./144.*m11*m33*me2*
         RSe45613p + 5./162.*m11*m33*me2*RSe45613p*RS34561pe + 5./162.*
         m11*m33*me2*RSe45613p*RS14563pe + 11./648.*m11*m33*me2*pow(
         RSe45613p,2) + 11./648.*m11*m33*me2*RSp45613e*RSe45613p + 5./
         324.*m11*m33*me2*RS3456pee*RSe45613p + 5./324.*m11*m33*me2*
         RS3456ppe*RSe45613p - 35./1152.*m11*m33*me2*RSe45633p - 5./324.
         *m11*m33*me2*RSe45633p*RS34561pe - 5./324.*m11*m33*me2*
         RSe45633p*RS14563pe - 11./648.*m11*m33*me2*RSe45633p*RSe45613p
          - 11./1296.*m11*m33*me2*RSe45633p*RSp45613e - 5./648.*m11*m33
         *me2*RSe45633p*RS3456pee - 5./648.*m11*m33*me2*RSe45633p*
         RS3456ppe + 11./2592.*m11*m33*me2*pow(RSe45633p,2) - 11./1296.
         *m11*m33*me2*RSp45633e*RSe45613p );

      M13p6x23R +=  + pi162 * ( 11./2592.*m11*m33*me2*RSp45633e*
         RSe45633p - 5./324.*m11*m33*me2*RS345pe*RSe45613p + 5./648.*
         m11*m33*me2*RS345pe*RSe45633p - 5./324.*m11*m33*me2*RS346pe*
         RSe45613p + 5./648.*m11*m33*me2*RS346pe*RSe45633p - 5./324.*
         m11*m33*me2*RS356pe*RSe45613p + 5./648.*m11*m33*me2*RS356pe*
         RSe45633p + 5./576.*m11*m33*me2*RSe563p + 5./576.*m11*m33*me2*
         RSe463p + 5./576.*m11*m33*me2*RSe453p + 5./324.*m11*m33*me2*
         RS1456pee*RSe45613p - 5./648.*m11*m33*me2*RS1456pee*RSe45633p
          + 5./324.*m11*m33*me2*RS1456ppe*RSe45613p - 5./648.*m11*m33*
         me2*RS1456ppe*RSe45633p - 35./1152.*m11*m33*me2*RSe45611p - 5./
         324.*m11*m33*me2*RSe45611p*RS34561pe - 5./324.*m11*m33*me2*
         RSe45611p*RS14563pe - 11./648.*m11*m33*me2*RSe45611p*RSe45613p
          - 11./1296.*m11*m33*me2*RSe45611p*RSp45613e - 5./648.*m11*m33
         *me2*RSe45611p*RS3456pee - 5./648.*m11*m33*me2*RSe45611p*
         RS3456ppe + 11./1296.*m11*m33*me2*RSe45611p*RSe45633p + 11./
         2592.*m11*m33*me2*RSe45611p*RSp45633e );

      M13p6x23R +=  + pi162 * ( 5./648.*m11*m33*me2*RSe45611p*RS345pe
          + 5./648.*m11*m33*me2*RSe45611p*RS346pe + 5./648.*m11*m33*me2
         *RSe45611p*RS356pe - 5./648.*m11*m33*me2*RSe45611p*RS1456pee
          - 5./648.*m11*m33*me2*RSe45611p*RS1456ppe + 11./2592.*m11*m33
         *me2*pow(RSe45611p,2) - 11./1296.*m11*m33*me2*RSp45611e*
         RSe45613p + 11./2592.*m11*m33*me2*RSp45611e*RSe45633p + 11./
         2592.*m11*m33*me2*RSp45611e*RSe45611p - 5./324.*m11*m33*me2*
         RS145pe*RSe45613p + 5./648.*m11*m33*me2*RS145pe*RSe45633p + 5./
         648.*m11*m33*me2*RS145pe*RSe45611p - 5./324.*m11*m33*me2*
         RS146pe*RSe45613p + 5./648.*m11*m33*me2*RS146pe*RSe45633p + 5./
         648.*m11*m33*me2*RS146pe*RSe45611p - 5./324.*m11*m33*me2*
         RS156pe*RSe45613p + 5./648.*m11*m33*me2*RS156pe*RSe45633p + 5./
         648.*m11*m33*me2*RS156pe*RSe45611p + 5./576.*m11*m33*me2*
         RSe561p + 5./576.*m11*m33*me2*RSe461p + 5./576.*m11*m33*me2*
         RSe451p - 5./576.*m11*m33*me2*RSe564p - 5./576.*m11*m33*me2*
         RSe465p );

      M13p6x23R +=  + pi162 * (  - 5./576.*m11*m33*me2*RSe456p + 13./
         192.*m11*m33*m66 + 17./576.*m11*m33*m66*RS34561pe + 17./576.*
         m11*m33*m66*RS14563pe + 11./576.*m11*m33*m66*RSe45613p + 11./
         576.*m11*m33*m66*RSp45613e + 41./1728.*m11*m33*m66*RS3456pee
          + 41./1728.*m11*m33*m66*RS3456ppe - 37./3456.*m11*m33*m66*
         RSe45633p - 37./3456.*m11*m33*m66*RSp45633e - 5./864.*m11*m33*
         m66*RS345pe - 41./1728.*m11*m33*m66*RS346pe - 41./1728.*m11*
         m33*m66*RS356pe + 1./432.*m11*m33*m66*RSe453p + 1./432.*m11*
         m33*m66*RSp453e + 41./1728.*m11*m33*m66*RS1456pee + 41./1728.*
         m11*m33*m66*RS1456ppe - 37./3456.*m11*m33*m66*RSe45611p - 37./
         3456.*m11*m33*m66*RSp45611e - 5./864.*m11*m33*m66*RS145pe - 41.
         /1728.*m11*m33*m66*RS146pe - 41./1728.*m11*m33*m66*RS156pe - 
         19./1728.*m11*m33*m66*RS645pe + 1./432.*m11*m33*m66*RSe451p + 
         1./432.*m11*m33*m66*RSp451e - 1./432.*m11*m33*m66*RSe456p - 1./
         432.*m11*m33*m66*RSp456e + 13./192.*m11*m33*m55 + 17./576.*m11
         *m33*m55*RS34561pe );

      M13p6x23R +=  + pi162 * ( 17./576.*m11*m33*m55*RS14563pe + 11./
         576.*m11*m33*m55*RSe45613p + 11./576.*m11*m33*m55*RSp45613e + 
         41./1728.*m11*m33*m55*RS3456pee + 41./1728.*m11*m33*m55*
         RS3456ppe - 37./3456.*m11*m33*m55*RSe45633p - 37./3456.*m11*
         m33*m55*RSp45633e - 41./1728.*m11*m33*m55*RS345pe - 5./864.*
         m11*m33*m55*RS346pe - 41./1728.*m11*m33*m55*RS356pe + 1./432.*
         m11*m33*m55*RSe463p + 1./432.*m11*m33*m55*RSp463e + 41./1728.*
         m11*m33*m55*RS1456pee + 41./1728.*m11*m33*m55*RS1456ppe - 37./
         3456.*m11*m33*m55*RSe45611p - 37./3456.*m11*m33*m55*RSp45611e
          - 41./1728.*m11*m33*m55*RS145pe - 5./864.*m11*m33*m55*RS146pe
          - 41./1728.*m11*m33*m55*RS156pe - 19./1728.*m11*m33*m55*
         RS546pe + 1./432.*m11*m33*m55*RSe461p + 1./432.*m11*m33*m55*
         RSp461e - 1./432.*m11*m33*m55*RSe465p - 1./432.*m11*m33*m55*
         RSp465e + 13./192.*m11*m33*m44 + 17./576.*m11*m33*m44*
         RS34561pe + 17./576.*m11*m33*m44*RS14563pe + 11./576.*m11*m33*
         m44*RSe45613p );

      M13p6x23R +=  + pi162 * ( 11./576.*m11*m33*m44*RSp45613e + 41./
         1728.*m11*m33*m44*RS3456pee + 41./1728.*m11*m33*m44*RS3456ppe
          - 37./3456.*m11*m33*m44*RSe45633p - 37./3456.*m11*m33*m44*
         RSp45633e - 41./1728.*m11*m33*m44*RS345pe - 41./1728.*m11*m33*
         m44*RS346pe - 5./864.*m11*m33*m44*RS356pe + 1./432.*m11*m33*
         m44*RSe563p + 1./432.*m11*m33*m44*RSp563e + 41./1728.*m11*m33*
         m44*RS1456pee + 41./1728.*m11*m33*m44*RS1456ppe - 37./3456.*
         m11*m33*m44*RSe45611p - 37./3456.*m11*m33*m44*RSp45611e - 41./
         1728.*m11*m33*m44*RS145pe - 41./1728.*m11*m33*m44*RS146pe - 5./
         864.*m11*m33*m44*RS156pe - 19./1728.*m11*m33*m44*RS456pe + 1./
         432.*m11*m33*m44*RSe561p + 1./432.*m11*m33*m44*RSp561e - 1./
         432.*m11*m33*m44*RSe564p - 1./432.*m11*m33*m44*RSp564e - 1./
         512.*m11*pow(m33,2) + 55./4608.*m11*pow(m33,2)*RS34561pe + 595.
         /20736.*m11*pow(m33,2)*pow(RS34561pe,2) + 143./4608.*m11*pow(
         m33,2)*RS14563pe + 169./3456.*m11*pow(m33,2)*RS14563pe*
         RS34561pe );

      M13p6x23R +=  + pi162 * ( 275./20736.*m11*pow(m33,2)*pow(
         RS14563pe,2) + 7./512.*m11*pow(m33,2)*RSe45613p + 275./10368.*
         m11*pow(m33,2)*RSe45613p*RS34561pe + 187./10368.*m11*pow(
         m33,2)*RSe45613p*RS14563pe + 11./2304.*m11*pow(m33,2)*pow(
         RSe45613p,2) + 7./512.*m11*pow(m33,2)*RSp45613e + 275./10368.*
         m11*pow(m33,2)*RSp45613e*RS34561pe + 187./10368.*m11*pow(
         m33,2)*RSp45613e*RS14563pe + 11./1152.*m11*pow(m33,2)*
         RSp45613e*RSe45613p + 11./2304.*m11*pow(m33,2)*pow(
         RSp45613e,2) + 151./4608.*m11*pow(m33,2)*RS3456pee + 595./
         20736.*m11*pow(m33,2)*RS3456pee*RS34561pe + 169./6912.*m11*
         pow(m33,2)*RS3456pee*RS14563pe + 275./20736.*m11*pow(m33,2)*
         RS3456pee*RSe45613p + 275./20736.*m11*pow(m33,2)*RS3456pee*
         RSp45613e + 595./82944.*m11*pow(m33,2)*pow(RS3456pee,2) + 151./
         4608.*m11*pow(m33,2)*RS3456ppe + 595./20736.*m11*pow(m33,2)*
         RS3456ppe*RS34561pe + 169./6912.*m11*pow(m33,2)*RS3456ppe*
         RS14563pe );

      M13p6x23R +=  + pi162 * ( 275./20736.*m11*pow(m33,2)*RS3456ppe*
         RSe45613p + 275./20736.*m11*pow(m33,2)*RS3456ppe*RSp45613e + 
         595./41472.*m11*pow(m33,2)*RS3456ppe*RS3456pee + 595./82944.*
         m11*pow(m33,2)*pow(RS3456ppe,2) - 59./4608.*m11*pow(m33,2)*
         RSe45633p - 275./20736.*m11*pow(m33,2)*RSe45633p*RS34561pe - 
         187./20736.*m11*pow(m33,2)*RSe45633p*RS14563pe - 11./2304.*m11
         *pow(m33,2)*RSe45633p*RSe45613p - 11./2304.*m11*pow(m33,2)*
         RSe45633p*RSp45613e - 275./41472.*m11*pow(m33,2)*RSe45633p*
         RS3456pee - 275./41472.*m11*pow(m33,2)*RSe45633p*RS3456ppe + 
         11./9216.*m11*pow(m33,2)*pow(RSe45633p,2) - 59./4608.*m11*pow(
         m33,2)*RSp45633e - 275./20736.*m11*pow(m33,2)*RSp45633e*
         RS34561pe - 187./20736.*m11*pow(m33,2)*RSp45633e*RS14563pe - 
         11./2304.*m11*pow(m33,2)*RSp45633e*RSe45613p - 11./2304.*m11*
         pow(m33,2)*RSp45633e*RSp45613e - 275./41472.*m11*pow(m33,2)*
         RSp45633e*RS3456pee - 275./41472.*m11*pow(m33,2)*RSp45633e*
         RS3456ppe );

      M13p6x23R +=  + pi162 * ( 11./4608.*m11*pow(m33,2)*RSp45633e*
         RSe45633p + 11./9216.*m11*pow(m33,2)*pow(RSp45633e,2) - 31./
         6912.*m11*pow(m33,2)*RS345pe - 595./20736.*m11*pow(m33,2)*
         RS345pe*RS34561pe - 169./6912.*m11*pow(m33,2)*RS345pe*
         RS14563pe - 275./20736.*m11*pow(m33,2)*RS345pe*RSe45613p - 275.
         /20736.*m11*pow(m33,2)*RS345pe*RSp45613e - 595./41472.*m11*
         pow(m33,2)*RS345pe*RS3456pee - 595./41472.*m11*pow(m33,2)*
         RS345pe*RS3456ppe + 275./41472.*m11*pow(m33,2)*RS345pe*
         RSe45633p + 275./41472.*m11*pow(m33,2)*RS345pe*RSp45633e + 595.
         /82944.*m11*pow(m33,2)*pow(RS345pe,2) - 31./6912.*m11*pow(
         m33,2)*RS346pe - 595./20736.*m11*pow(m33,2)*RS346pe*RS34561pe
          - 169./6912.*m11*pow(m33,2)*RS346pe*RS14563pe - 275./20736.*
         m11*pow(m33,2)*RS346pe*RSe45613p - 275./20736.*m11*pow(m33,2)*
         RS346pe*RSp45613e - 595./41472.*m11*pow(m33,2)*RS346pe*
         RS3456pee - 595./41472.*m11*pow(m33,2)*RS346pe*RS3456ppe + 275.
         /41472.*m11*pow(m33,2)*RS346pe*RSe45633p );

      M13p6x23R +=  + pi162 * ( 275./41472.*m11*pow(m33,2)*RS346pe*
         RSp45633e + 595./41472.*m11*pow(m33,2)*RS346pe*RS345pe + 595./
         82944.*m11*pow(m33,2)*pow(RS346pe,2) - 31./6912.*m11*pow(
         m33,2)*RS356pe - 595./20736.*m11*pow(m33,2)*RS356pe*RS34561pe
          - 169./6912.*m11*pow(m33,2)*RS356pe*RS14563pe - 275./20736.*
         m11*pow(m33,2)*RS356pe*RSe45613p - 275./20736.*m11*pow(m33,2)*
         RS356pe*RSp45613e - 595./41472.*m11*pow(m33,2)*RS356pe*
         RS3456pee - 595./41472.*m11*pow(m33,2)*RS356pe*RS3456ppe + 275.
         /41472.*m11*pow(m33,2)*RS356pe*RSe45633p + 275./41472.*m11*
         pow(m33,2)*RS356pe*RSp45633e + 595./41472.*m11*pow(m33,2)*
         RS356pe*RS345pe + 595./41472.*m11*pow(m33,2)*RS356pe*RS346pe
          + 595./82944.*m11*pow(m33,2)*pow(RS356pe,2) + 55./13824.*m11*
         pow(m33,2)*RSe563p + 55./13824.*m11*pow(m33,2)*RSe463p + 55./
         13824.*m11*pow(m33,2)*RSe453p + 55./13824.*m11*pow(m33,2)*
         RSp563e + 55./13824.*m11*pow(m33,2)*RSp463e + 55./13824.*m11*
         pow(m33,2)*RSp453e );

      M13p6x23R +=  + pi162 * ( 71./4608.*m11*pow(m33,2)*RS1456pee + 
         169./6912.*m11*pow(m33,2)*RS1456pee*RS34561pe + 275./20736.*
         m11*pow(m33,2)*RS1456pee*RS14563pe + 187./20736.*m11*pow(
         m33,2)*RS1456pee*RSe45613p + 187./20736.*m11*pow(m33,2)*
         RS1456pee*RSp45613e + 169./13824.*m11*pow(m33,2)*RS1456pee*
         RS3456pee + 169./13824.*m11*pow(m33,2)*RS1456pee*RS3456ppe - 
         187./41472.*m11*pow(m33,2)*RS1456pee*RSe45633p - 187./41472.*
         m11*pow(m33,2)*RS1456pee*RSp45633e - 169./13824.*m11*pow(
         m33,2)*RS1456pee*RS345pe - 169./13824.*m11*pow(m33,2)*
         RS1456pee*RS346pe - 169./13824.*m11*pow(m33,2)*RS1456pee*
         RS356pe + 275./82944.*m11*pow(m33,2)*pow(RS1456pee,2) + 71./
         4608.*m11*pow(m33,2)*RS1456ppe + 169./6912.*m11*pow(m33,2)*
         RS1456ppe*RS34561pe + 275./20736.*m11*pow(m33,2)*RS1456ppe*
         RS14563pe + 187./20736.*m11*pow(m33,2)*RS1456ppe*RSe45613p + 
         187./20736.*m11*pow(m33,2)*RS1456ppe*RSp45613e + 169./13824.*
         m11*pow(m33,2)*RS1456ppe*RS3456pee );

      M13p6x23R +=  + pi162 * ( 169./13824.*m11*pow(m33,2)*RS1456ppe*
         RS3456ppe - 187./41472.*m11*pow(m33,2)*RS1456ppe*RSe45633p - 
         187./41472.*m11*pow(m33,2)*RS1456ppe*RSp45633e - 169./13824.*
         m11*pow(m33,2)*RS1456ppe*RS345pe - 169./13824.*m11*pow(m33,2)*
         RS1456ppe*RS346pe - 169./13824.*m11*pow(m33,2)*RS1456ppe*
         RS356pe + 275./41472.*m11*pow(m33,2)*RS1456ppe*RS1456pee + 275.
         /82944.*m11*pow(m33,2)*pow(RS1456ppe,2) - 19./4608.*m11*pow(
         m33,2)*RSe45611p - 275./20736.*m11*pow(m33,2)*RSe45611p*
         RS34561pe - 187./20736.*m11*pow(m33,2)*RSe45611p*RS14563pe - 
         11./2304.*m11*pow(m33,2)*RSe45611p*RSe45613p - 11./2304.*m11*
         pow(m33,2)*RSe45611p*RSp45613e - 275./41472.*m11*pow(m33,2)*
         RSe45611p*RS3456pee - 275./41472.*m11*pow(m33,2)*RSe45611p*
         RS3456ppe + 11./4608.*m11*pow(m33,2)*RSe45611p*RSe45633p + 11./
         4608.*m11*pow(m33,2)*RSe45611p*RSp45633e + 275./41472.*m11*
         pow(m33,2)*RSe45611p*RS345pe + 275./41472.*m11*pow(m33,2)*
         RSe45611p*RS346pe );

      M13p6x23R +=  + pi162 * ( 275./41472.*m11*pow(m33,2)*RSe45611p*
         RS356pe - 187./41472.*m11*pow(m33,2)*RSe45611p*RS1456pee - 187.
         /41472.*m11*pow(m33,2)*RSe45611p*RS1456ppe + 11./9216.*m11*
         pow(m33,2)*pow(RSe45611p,2) - 19./4608.*m11*pow(m33,2)*
         RSp45611e - 275./20736.*m11*pow(m33,2)*RSp45611e*RS34561pe - 
         187./20736.*m11*pow(m33,2)*RSp45611e*RS14563pe - 11./2304.*m11
         *pow(m33,2)*RSp45611e*RSe45613p - 11./2304.*m11*pow(m33,2)*
         RSp45611e*RSp45613e - 275./41472.*m11*pow(m33,2)*RSp45611e*
         RS3456pee - 275./41472.*m11*pow(m33,2)*RSp45611e*RS3456ppe + 
         11./4608.*m11*pow(m33,2)*RSp45611e*RSe45633p + 11./4608.*m11*
         pow(m33,2)*RSp45611e*RSp45633e + 275./41472.*m11*pow(m33,2)*
         RSp45611e*RS345pe + 275./41472.*m11*pow(m33,2)*RSp45611e*
         RS346pe + 275./41472.*m11*pow(m33,2)*RSp45611e*RS356pe - 187./
         41472.*m11*pow(m33,2)*RSp45611e*RS1456pee - 187./41472.*m11*
         pow(m33,2)*RSp45611e*RS1456ppe + 11./4608.*m11*pow(m33,2)*
         RSp45611e*RSe45611p );

      M13p6x23R +=  + pi162 * ( 11./9216.*m11*pow(m33,2)*pow(
         RSp45611e,2) - 107./6912.*m11*pow(m33,2)*RS145pe - 169./6912.*
         m11*pow(m33,2)*RS145pe*RS34561pe - 275./20736.*m11*pow(m33,2)*
         RS145pe*RS14563pe - 187./20736.*m11*pow(m33,2)*RS145pe*
         RSe45613p - 187./20736.*m11*pow(m33,2)*RS145pe*RSp45613e - 169.
         /13824.*m11*pow(m33,2)*RS145pe*RS3456pee - 169./13824.*m11*
         pow(m33,2)*RS145pe*RS3456ppe + 187./41472.*m11*pow(m33,2)*
         RS145pe*RSe45633p + 187./41472.*m11*pow(m33,2)*RS145pe*
         RSp45633e + 169./13824.*m11*pow(m33,2)*RS145pe*RS345pe + 169./
         13824.*m11*pow(m33,2)*RS145pe*RS346pe + 169./13824.*m11*pow(
         m33,2)*RS145pe*RS356pe - 275./41472.*m11*pow(m33,2)*RS145pe*
         RS1456pee - 275./41472.*m11*pow(m33,2)*RS145pe*RS1456ppe + 187.
         /41472.*m11*pow(m33,2)*RS145pe*RSe45611p + 187./41472.*m11*
         pow(m33,2)*RS145pe*RSp45611e + 275./82944.*m11*pow(m33,2)*pow(
         RS145pe,2) - 107./6912.*m11*pow(m33,2)*RS146pe - 169./6912.*
         m11*pow(m33,2)*RS146pe*RS34561pe );

      M13p6x23R +=  + pi162 * (  - 275./20736.*m11*pow(m33,2)*RS146pe*
         RS14563pe - 187./20736.*m11*pow(m33,2)*RS146pe*RSe45613p - 187.
         /20736.*m11*pow(m33,2)*RS146pe*RSp45613e - 169./13824.*m11*
         pow(m33,2)*RS146pe*RS3456pee - 169./13824.*m11*pow(m33,2)*
         RS146pe*RS3456ppe + 187./41472.*m11*pow(m33,2)*RS146pe*
         RSe45633p + 187./41472.*m11*pow(m33,2)*RS146pe*RSp45633e + 169.
         /13824.*m11*pow(m33,2)*RS146pe*RS345pe + 169./13824.*m11*pow(
         m33,2)*RS146pe*RS346pe + 169./13824.*m11*pow(m33,2)*RS146pe*
         RS356pe - 275./41472.*m11*pow(m33,2)*RS146pe*RS1456pee - 275./
         41472.*m11*pow(m33,2)*RS146pe*RS1456ppe + 187./41472.*m11*pow(
         m33,2)*RS146pe*RSe45611p + 187./41472.*m11*pow(m33,2)*RS146pe*
         RSp45611e + 275./41472.*m11*pow(m33,2)*RS146pe*RS145pe + 275./
         82944.*m11*pow(m33,2)*pow(RS146pe,2) - 107./6912.*m11*pow(
         m33,2)*RS156pe - 169./6912.*m11*pow(m33,2)*RS156pe*RS34561pe
          - 275./20736.*m11*pow(m33,2)*RS156pe*RS14563pe - 187./20736.*
         m11*pow(m33,2)*RS156pe*RSe45613p );

      M13p6x23R +=  + pi162 * (  - 187./20736.*m11*pow(m33,2)*RS156pe*
         RSp45613e - 169./13824.*m11*pow(m33,2)*RS156pe*RS3456pee - 169.
         /13824.*m11*pow(m33,2)*RS156pe*RS3456ppe + 187./41472.*m11*
         pow(m33,2)*RS156pe*RSe45633p + 187./41472.*m11*pow(m33,2)*
         RS156pe*RSp45633e + 169./13824.*m11*pow(m33,2)*RS156pe*RS345pe
          + 169./13824.*m11*pow(m33,2)*RS156pe*RS346pe + 169./13824.*
         m11*pow(m33,2)*RS156pe*RS356pe - 275./41472.*m11*pow(m33,2)*
         RS156pe*RS1456pee - 275./41472.*m11*pow(m33,2)*RS156pe*
         RS1456ppe + 187./41472.*m11*pow(m33,2)*RS156pe*RSe45611p + 187.
         /41472.*m11*pow(m33,2)*RS156pe*RSp45611e + 275./41472.*m11*
         pow(m33,2)*RS156pe*RS145pe + 275./41472.*m11*pow(m33,2)*
         RS156pe*RS146pe + 275./82944.*m11*pow(m33,2)*pow(RS156pe,2) - 
         5./4608.*m11*pow(m33,2)*RS645pe - 5./4608.*m11*pow(m33,2)*
         RS546pe - 5./4608.*m11*pow(m33,2)*RS456pe - 25./13824.*m11*
         pow(m33,2)*RSe561p - 25./13824.*m11*pow(m33,2)*RSe461p - 25./
         13824.*m11*pow(m33,2)*RSe451p );

      M13p6x23R +=  + pi162 * (  - 25./13824.*m11*pow(m33,2)*RSp561e - 
         25./13824.*m11*pow(m33,2)*RSp461e - 25./13824.*m11*pow(m33,2)*
         RSp451e - 5./4608.*m11*pow(m33,2)*RSe564p - 5./4608.*m11*pow(
         m33,2)*RSe465p - 5./4608.*m11*pow(m33,2)*RSe456p - 5./4608.*
         m11*pow(m33,2)*RSp564e - 5./4608.*m11*pow(m33,2)*RSp465e - 5./
         4608.*m11*pow(m33,2)*RSp456e - 13./1152.*pow(m11,2)*RD3456ep
          - 11./2592.*pow(m11,2)*RD3456ep*RS34561pe - 29./2592.*pow(
         m11,2)*RD3456ep*RS14563pe - 11./2592.*pow(m11,2)*RD3456ep*
         RSe45613p - 11./2592.*pow(m11,2)*RD3456ep*RSp45613e - 11./5184.
         *pow(m11,2)*RS3456pee*RD3456ep - 11./5184.*pow(m11,2)*
         RS3456ppe*RD3456ep + 11./5184.*pow(m11,2)*RSe45633p*RD3456ep
          + 11./5184.*pow(m11,2)*RSp45633e*RD3456ep + 11./5184.*pow(
         m11,2)*RS345pe*RD3456ep + 11./5184.*pow(m11,2)*RS346pe*
         RD3456ep + 11./5184.*pow(m11,2)*RS356pe*RD3456ep + 1./576.*
         pow(m11,2)*RD1456ep - 11./2592.*pow(m11,2)*RD1456ep*RS34561pe
          - 29./2592.*pow(m11,2)*RD1456ep*RS14563pe );

      M13p6x23R +=  + pi162 * (  - 11./2592.*pow(m11,2)*RD1456ep*
         RSe45613p - 11./2592.*pow(m11,2)*RD1456ep*RSp45613e - 11./5184.
         *pow(m11,2)*RD1456ep*RS3456pee - 11./5184.*pow(m11,2)*RD1456ep
         *RS3456ppe + 11./5184.*pow(m11,2)*RD1456ep*RSe45633p + 11./
         5184.*pow(m11,2)*RD1456ep*RSp45633e + 11./5184.*pow(m11,2)*
         RD1456ep*RS345pe + 11./5184.*pow(m11,2)*RD1456ep*RS346pe + 11./
         5184.*pow(m11,2)*RD1456ep*RS356pe - 29./5184.*pow(m11,2)*
         RS1456pee*RD3456ep - 29./5184.*pow(m11,2)*RS1456pee*RD1456ep
          - 29./5184.*pow(m11,2)*RS1456ppe*RD3456ep - 29./5184.*pow(
         m11,2)*RS1456ppe*RD1456ep + 11./5184.*pow(m11,2)*RSe45611p*
         RD3456ep + 11./5184.*pow(m11,2)*RSe45611p*RD1456ep + 11./5184.
         *pow(m11,2)*RSp45611e*RD3456ep + 11./5184.*pow(m11,2)*
         RSp45611e*RD1456ep + 29./5184.*pow(m11,2)*RS145pe*RD3456ep + 
         29./5184.*pow(m11,2)*RS145pe*RD1456ep + 29./5184.*pow(m11,2)*
         RS146pe*RD3456ep + 29./5184.*pow(m11,2)*RS146pe*RD1456ep + 29./
         5184.*pow(m11,2)*RS156pe*RD3456ep );

      M13p6x23R +=  + pi162 * ( 29./5184.*pow(m11,2)*RS156pe*RD1456ep
          + 5./288.*pow(m11,2)*mp2*RSp45613e + 11./1296.*pow(m11,2)*mp2
         *RSp45613e*RS34561pe + 29./1296.*pow(m11,2)*mp2*RSp45613e*
         RS14563pe + 11./1296.*pow(m11,2)*mp2*RSp45613e*RSe45613p + 11./
         1296.*pow(m11,2)*mp2*pow(RSp45613e,2) + 11./2592.*pow(m11,2)*
         mp2*RS3456pee*RSp45613e + 11./2592.*pow(m11,2)*mp2*RS3456ppe*
         RSp45613e - 11./2592.*pow(m11,2)*mp2*RSe45633p*RSp45613e - 13./
         1152.*pow(m11,2)*mp2*RSp45633e - 11./2592.*pow(m11,2)*mp2*
         RSp45633e*RS34561pe - 29./2592.*pow(m11,2)*mp2*RSp45633e*
         RS14563pe - 11./2592.*pow(m11,2)*mp2*RSp45633e*RSe45613p - 11./
         1296.*pow(m11,2)*mp2*RSp45633e*RSp45613e - 11./5184.*pow(
         m11,2)*mp2*RSp45633e*RS3456pee - 11./5184.*pow(m11,2)*mp2*
         RSp45633e*RS3456ppe + 11./5184.*pow(m11,2)*mp2*RSp45633e*
         RSe45633p + 11./5184.*pow(m11,2)*mp2*pow(RSp45633e,2) - 11./
         2592.*pow(m11,2)*mp2*RS345pe*RSp45613e + 11./5184.*pow(m11,2)*
         mp2*RS345pe*RSp45633e );

      M13p6x23R +=  + pi162 * (  - 11./2592.*pow(m11,2)*mp2*RS346pe*
         RSp45613e + 11./5184.*pow(m11,2)*mp2*RS346pe*RSp45633e - 11./
         2592.*pow(m11,2)*mp2*RS356pe*RSp45613e + 11./5184.*pow(m11,2)*
         mp2*RS356pe*RSp45633e + 1./576.*pow(m11,2)*mp2*RSp563e + 1./
         576.*pow(m11,2)*mp2*RSp463e + 1./576.*pow(m11,2)*mp2*RSp453e
          + 29./2592.*pow(m11,2)*mp2*RS1456pee*RSp45613e - 29./5184.*
         pow(m11,2)*mp2*RS1456pee*RSp45633e + 29./2592.*pow(m11,2)*mp2*
         RS1456ppe*RSp45613e - 29./5184.*pow(m11,2)*mp2*RS1456ppe*
         RSp45633e - 11./2592.*pow(m11,2)*mp2*RSe45611p*RSp45613e + 11./
         5184.*pow(m11,2)*mp2*RSe45611p*RSp45633e - 11./576.*pow(m11,2)
         *mp2*RSp45611e - 11./2592.*pow(m11,2)*mp2*RSp45611e*RS34561pe
          - 29./2592.*pow(m11,2)*mp2*RSp45611e*RS14563pe - 11./2592.*
         pow(m11,2)*mp2*RSp45611e*RSe45613p - 11./1296.*pow(m11,2)*mp2*
         RSp45611e*RSp45613e - 11./5184.*pow(m11,2)*mp2*RSp45611e*
         RS3456pee - 11./5184.*pow(m11,2)*mp2*RSp45611e*RS3456ppe + 11./
         5184.*pow(m11,2)*mp2*RSp45611e*RSe45633p );

      M13p6x23R +=  + pi162 * ( 11./2592.*pow(m11,2)*mp2*RSp45611e*
         RSp45633e + 11./5184.*pow(m11,2)*mp2*RSp45611e*RS345pe + 11./
         5184.*pow(m11,2)*mp2*RSp45611e*RS346pe + 11./5184.*pow(m11,2)*
         mp2*RSp45611e*RS356pe - 29./5184.*pow(m11,2)*mp2*RSp45611e*
         RS1456pee - 29./5184.*pow(m11,2)*mp2*RSp45611e*RS1456ppe + 11./
         5184.*pow(m11,2)*mp2*RSp45611e*RSe45611p + 11./5184.*pow(
         m11,2)*mp2*pow(RSp45611e,2) - 29./2592.*pow(m11,2)*mp2*RS145pe
         *RSp45613e + 29./5184.*pow(m11,2)*mp2*RS145pe*RSp45633e + 29./
         5184.*pow(m11,2)*mp2*RS145pe*RSp45611e - 29./2592.*pow(m11,2)*
         mp2*RS146pe*RSp45613e + 29./5184.*pow(m11,2)*mp2*RS146pe*
         RSp45633e + 29./5184.*pow(m11,2)*mp2*RS146pe*RSp45611e - 29./
         2592.*pow(m11,2)*mp2*RS156pe*RSp45613e + 29./5184.*pow(m11,2)*
         mp2*RS156pe*RSp45633e + 29./5184.*pow(m11,2)*mp2*RS156pe*
         RSp45611e + 1./144.*pow(m11,2)*mp2*RSp561e + 1./144.*pow(
         m11,2)*mp2*RSp461e + 1./144.*pow(m11,2)*mp2*RSp451e - 5./1152.
         *pow(m11,2)*mp2*RSp564e );

      M13p6x23R +=  + pi162 * (  - 5./1152.*pow(m11,2)*mp2*RSp465e - 5./
         1152.*pow(m11,2)*mp2*RSp456e + 5./288.*pow(m11,2)*me2*
         RSe45613p + 11./1296.*pow(m11,2)*me2*RSe45613p*RS34561pe + 29./
         1296.*pow(m11,2)*me2*RSe45613p*RS14563pe + 11./1296.*pow(
         m11,2)*me2*pow(RSe45613p,2) + 11./1296.*pow(m11,2)*me2*
         RSp45613e*RSe45613p + 11./2592.*pow(m11,2)*me2*RS3456pee*
         RSe45613p + 11./2592.*pow(m11,2)*me2*RS3456ppe*RSe45613p - 13./
         1152.*pow(m11,2)*me2*RSe45633p - 11./2592.*pow(m11,2)*me2*
         RSe45633p*RS34561pe - 29./2592.*pow(m11,2)*me2*RSe45633p*
         RS14563pe - 11./1296.*pow(m11,2)*me2*RSe45633p*RSe45613p - 11./
         2592.*pow(m11,2)*me2*RSe45633p*RSp45613e - 11./5184.*pow(
         m11,2)*me2*RSe45633p*RS3456pee - 11./5184.*pow(m11,2)*me2*
         RSe45633p*RS3456ppe + 11./5184.*pow(m11,2)*me2*pow(
         RSe45633p,2) - 11./2592.*pow(m11,2)*me2*RSp45633e*RSe45613p + 
         11./5184.*pow(m11,2)*me2*RSp45633e*RSe45633p - 11./2592.*pow(
         m11,2)*me2*RS345pe*RSe45613p );

      M13p6x23R +=  + pi162 * ( 11./5184.*pow(m11,2)*me2*RS345pe*
         RSe45633p - 11./2592.*pow(m11,2)*me2*RS346pe*RSe45613p + 11./
         5184.*pow(m11,2)*me2*RS346pe*RSe45633p - 11./2592.*pow(m11,2)*
         me2*RS356pe*RSe45613p + 11./5184.*pow(m11,2)*me2*RS356pe*
         RSe45633p + 1./576.*pow(m11,2)*me2*RSe563p + 1./576.*pow(
         m11,2)*me2*RSe463p + 1./576.*pow(m11,2)*me2*RSe453p + 29./2592.
         *pow(m11,2)*me2*RS1456pee*RSe45613p - 29./5184.*pow(m11,2)*me2
         *RS1456pee*RSe45633p + 29./2592.*pow(m11,2)*me2*RS1456ppe*
         RSe45613p - 29./5184.*pow(m11,2)*me2*RS1456ppe*RSe45633p - 11./
         576.*pow(m11,2)*me2*RSe45611p - 11./2592.*pow(m11,2)*me2*
         RSe45611p*RS34561pe - 29./2592.*pow(m11,2)*me2*RSe45611p*
         RS14563pe - 11./1296.*pow(m11,2)*me2*RSe45611p*RSe45613p - 11./
         2592.*pow(m11,2)*me2*RSe45611p*RSp45613e - 11./5184.*pow(
         m11,2)*me2*RSe45611p*RS3456pee - 11./5184.*pow(m11,2)*me2*
         RSe45611p*RS3456ppe + 11./2592.*pow(m11,2)*me2*RSe45611p*
         RSe45633p );

      M13p6x23R +=  + pi162 * ( 11./5184.*pow(m11,2)*me2*RSe45611p*
         RSp45633e + 11./5184.*pow(m11,2)*me2*RSe45611p*RS345pe + 11./
         5184.*pow(m11,2)*me2*RSe45611p*RS346pe + 11./5184.*pow(m11,2)*
         me2*RSe45611p*RS356pe - 29./5184.*pow(m11,2)*me2*RSe45611p*
         RS1456pee - 29./5184.*pow(m11,2)*me2*RSe45611p*RS1456ppe + 11./
         5184.*pow(m11,2)*me2*pow(RSe45611p,2) - 11./2592.*pow(m11,2)*
         me2*RSp45611e*RSe45613p + 11./5184.*pow(m11,2)*me2*RSp45611e*
         RSe45633p + 11./5184.*pow(m11,2)*me2*RSp45611e*RSe45611p - 29./
         2592.*pow(m11,2)*me2*RS145pe*RSe45613p + 29./5184.*pow(m11,2)*
         me2*RS145pe*RSe45633p + 29./5184.*pow(m11,2)*me2*RS145pe*
         RSe45611p - 29./2592.*pow(m11,2)*me2*RS146pe*RSe45613p + 29./
         5184.*pow(m11,2)*me2*RS146pe*RSe45633p + 29./5184.*pow(m11,2)*
         me2*RS146pe*RSe45611p - 29./2592.*pow(m11,2)*me2*RS156pe*
         RSe45613p + 29./5184.*pow(m11,2)*me2*RS156pe*RSe45633p + 29./
         5184.*pow(m11,2)*me2*RS156pe*RSe45611p + 1./144.*pow(m11,2)*
         me2*RSe561p );

      M13p6x23R +=  + pi162 * ( 1./144.*pow(m11,2)*me2*RSe461p + 1./144.
         *pow(m11,2)*me2*RSe451p - 5./1152.*pow(m11,2)*me2*RSe564p - 5./
         1152.*pow(m11,2)*me2*RSe465p - 5./1152.*pow(m11,2)*me2*RSe456p
          + 13./384.*pow(m11,2)*m66 + 11./1152.*pow(m11,2)*m66*
         RS34561pe + 23./1152.*pow(m11,2)*m66*RS14563pe + 11./1152.*
         pow(m11,2)*m66*RSe45613p + 11./1152.*pow(m11,2)*m66*RSp45613e
          + 5./3456.*pow(m11,2)*m66*RS3456pee + 5./3456.*pow(m11,2)*m66
         *RS3456ppe - 5./3456.*pow(m11,2)*m66*RSe45633p - 5./3456.*pow(
         m11,2)*m66*RSp45633e - 7./864.*pow(m11,2)*m66*RS345pe - 5./
         3456.*pow(m11,2)*m66*RS346pe - 5./3456.*pow(m11,2)*m66*RS356pe
          - 23./3456.*pow(m11,2)*m66*RSe453p - 23./3456.*pow(m11,2)*m66
         *RSp453e + 77./3456.*pow(m11,2)*m66*RS1456pee + 77./3456.*pow(
         m11,2)*m66*RS1456ppe - 1./108.*pow(m11,2)*m66*RSe45611p - 1./
         108.*pow(m11,2)*m66*RSp45611e + 1./432.*pow(m11,2)*m66*RS145pe
          - 77./3456.*pow(m11,2)*m66*RS146pe - 77./3456.*pow(m11,2)*m66
         *RS156pe );

      M13p6x23R +=  + pi162 * (  - 19./3456.*pow(m11,2)*m66*RS645pe + 
         31./3456.*pow(m11,2)*m66*RSe451p + 31./3456.*pow(m11,2)*m66*
         RSp451e - 1./864.*pow(m11,2)*m66*RSe456p - 1./864.*pow(m11,2)*
         m66*RSp456e + 13./384.*pow(m11,2)*m55 + 11./1152.*pow(m11,2)*
         m55*RS34561pe + 23./1152.*pow(m11,2)*m55*RS14563pe + 11./1152.
         *pow(m11,2)*m55*RSe45613p + 11./1152.*pow(m11,2)*m55*RSp45613e
          + 5./3456.*pow(m11,2)*m55*RS3456pee + 5./3456.*pow(m11,2)*m55
         *RS3456ppe - 5./3456.*pow(m11,2)*m55*RSe45633p - 5./3456.*pow(
         m11,2)*m55*RSp45633e - 5./3456.*pow(m11,2)*m55*RS345pe - 7./
         864.*pow(m11,2)*m55*RS346pe - 5./3456.*pow(m11,2)*m55*RS356pe
          - 23./3456.*pow(m11,2)*m55*RSe463p - 23./3456.*pow(m11,2)*m55
         *RSp463e + 77./3456.*pow(m11,2)*m55*RS1456pee + 77./3456.*pow(
         m11,2)*m55*RS1456ppe - 1./108.*pow(m11,2)*m55*RSe45611p - 1./
         108.*pow(m11,2)*m55*RSp45611e - 77./3456.*pow(m11,2)*m55*
         RS145pe + 1./432.*pow(m11,2)*m55*RS146pe - 77./3456.*pow(
         m11,2)*m55*RS156pe );

      M13p6x23R +=  + pi162 * (  - 19./3456.*pow(m11,2)*m55*RS546pe + 
         31./3456.*pow(m11,2)*m55*RSe461p + 31./3456.*pow(m11,2)*m55*
         RSp461e - 1./864.*pow(m11,2)*m55*RSe465p - 1./864.*pow(m11,2)*
         m55*RSp465e + 13./384.*pow(m11,2)*m44 + 11./1152.*pow(m11,2)*
         m44*RS34561pe + 23./1152.*pow(m11,2)*m44*RS14563pe + 11./1152.
         *pow(m11,2)*m44*RSe45613p + 11./1152.*pow(m11,2)*m44*RSp45613e
          + 5./3456.*pow(m11,2)*m44*RS3456pee + 5./3456.*pow(m11,2)*m44
         *RS3456ppe - 5./3456.*pow(m11,2)*m44*RSe45633p - 5./3456.*pow(
         m11,2)*m44*RSp45633e - 5./3456.*pow(m11,2)*m44*RS345pe - 5./
         3456.*pow(m11,2)*m44*RS346pe - 7./864.*pow(m11,2)*m44*RS356pe
          - 23./3456.*pow(m11,2)*m44*RSe563p - 23./3456.*pow(m11,2)*m44
         *RSp563e + 77./3456.*pow(m11,2)*m44*RS1456pee + 77./3456.*pow(
         m11,2)*m44*RS1456ppe - 1./108.*pow(m11,2)*m44*RSe45611p - 1./
         108.*pow(m11,2)*m44*RSp45611e - 77./3456.*pow(m11,2)*m44*
         RS145pe - 77./3456.*pow(m11,2)*m44*RS146pe + 1./432.*pow(
         m11,2)*m44*RS156pe );

      M13p6x23R +=  + pi162 * (  - 19./3456.*pow(m11,2)*m44*RS456pe + 
         31./3456.*pow(m11,2)*m44*RSe561p + 31./3456.*pow(m11,2)*m44*
         RSp561e - 1./864.*pow(m11,2)*m44*RSe564p - 1./864.*pow(m11,2)*
         m44*RSp564e - 1./512.*pow(m11,2)*m33 + 143./4608.*pow(m11,2)*
         m33*RS34561pe + 275./20736.*pow(m11,2)*m33*pow(RS34561pe,2) + 
         55./4608.*pow(m11,2)*m33*RS14563pe + 169./3456.*pow(m11,2)*m33
         *RS14563pe*RS34561pe + 595./20736.*pow(m11,2)*m33*pow(
         RS14563pe,2) + 7./512.*pow(m11,2)*m33*RSe45613p + 187./10368.*
         pow(m11,2)*m33*RSe45613p*RS34561pe + 275./10368.*pow(m11,2)*
         m33*RSe45613p*RS14563pe + 11./2304.*pow(m11,2)*m33*pow(
         RSe45613p,2) + 7./512.*pow(m11,2)*m33*RSp45613e + 187./10368.*
         pow(m11,2)*m33*RSp45613e*RS34561pe + 275./10368.*pow(m11,2)*
         m33*RSp45613e*RS14563pe + 11./1152.*pow(m11,2)*m33*RSp45613e*
         RSe45613p + 11./2304.*pow(m11,2)*m33*pow(RSp45613e,2) + 71./
         4608.*pow(m11,2)*m33*RS3456pee + 275./20736.*pow(m11,2)*m33*
         RS3456pee*RS34561pe );

      M13p6x23R +=  + pi162 * ( 169./6912.*pow(m11,2)*m33*RS3456pee*
         RS14563pe + 187./20736.*pow(m11,2)*m33*RS3456pee*RSe45613p + 
         187./20736.*pow(m11,2)*m33*RS3456pee*RSp45613e + 275./82944.*
         pow(m11,2)*m33*pow(RS3456pee,2) + 71./4608.*pow(m11,2)*m33*
         RS3456ppe + 275./20736.*pow(m11,2)*m33*RS3456ppe*RS34561pe + 
         169./6912.*pow(m11,2)*m33*RS3456ppe*RS14563pe + 187./20736.*
         pow(m11,2)*m33*RS3456ppe*RSe45613p + 187./20736.*pow(m11,2)*
         m33*RS3456ppe*RSp45613e + 275./41472.*pow(m11,2)*m33*RS3456ppe
         *RS3456pee + 275./82944.*pow(m11,2)*m33*pow(RS3456ppe,2) - 19./
         4608.*pow(m11,2)*m33*RSe45633p - 187./20736.*pow(m11,2)*m33*
         RSe45633p*RS34561pe - 275./20736.*pow(m11,2)*m33*RSe45633p*
         RS14563pe - 11./2304.*pow(m11,2)*m33*RSe45633p*RSe45613p - 11./
         2304.*pow(m11,2)*m33*RSe45633p*RSp45613e - 187./41472.*pow(
         m11,2)*m33*RSe45633p*RS3456pee - 187./41472.*pow(m11,2)*m33*
         RSe45633p*RS3456ppe + 11./9216.*pow(m11,2)*m33*pow(
         RSe45633p,2) );

      M13p6x23R +=  + pi162 * (  - 19./4608.*pow(m11,2)*m33*RSp45633e
          - 187./20736.*pow(m11,2)*m33*RSp45633e*RS34561pe - 275./20736.
         *pow(m11,2)*m33*RSp45633e*RS14563pe - 11./2304.*pow(m11,2)*m33
         *RSp45633e*RSe45613p - 11./2304.*pow(m11,2)*m33*RSp45633e*
         RSp45613e - 187./41472.*pow(m11,2)*m33*RSp45633e*RS3456pee - 
         187./41472.*pow(m11,2)*m33*RSp45633e*RS3456ppe + 11./4608.*
         pow(m11,2)*m33*RSp45633e*RSe45633p + 11./9216.*pow(m11,2)*m33*
         pow(RSp45633e,2) - 107./6912.*pow(m11,2)*m33*RS345pe - 275./
         20736.*pow(m11,2)*m33*RS345pe*RS34561pe - 169./6912.*pow(
         m11,2)*m33*RS345pe*RS14563pe - 187./20736.*pow(m11,2)*m33*
         RS345pe*RSe45613p - 187./20736.*pow(m11,2)*m33*RS345pe*
         RSp45613e - 275./41472.*pow(m11,2)*m33*RS345pe*RS3456pee - 275.
         /41472.*pow(m11,2)*m33*RS345pe*RS3456ppe + 187./41472.*pow(
         m11,2)*m33*RS345pe*RSe45633p + 187./41472.*pow(m11,2)*m33*
         RS345pe*RSp45633e + 275./82944.*pow(m11,2)*m33*pow(RS345pe,2)
          - 107./6912.*pow(m11,2)*m33*RS346pe );

      M13p6x23R +=  + pi162 * (  - 275./20736.*pow(m11,2)*m33*RS346pe*
         RS34561pe - 169./6912.*pow(m11,2)*m33*RS346pe*RS14563pe - 187./
         20736.*pow(m11,2)*m33*RS346pe*RSe45613p - 187./20736.*pow(
         m11,2)*m33*RS346pe*RSp45613e - 275./41472.*pow(m11,2)*m33*
         RS346pe*RS3456pee - 275./41472.*pow(m11,2)*m33*RS346pe*
         RS3456ppe + 187./41472.*pow(m11,2)*m33*RS346pe*RSe45633p + 187.
         /41472.*pow(m11,2)*m33*RS346pe*RSp45633e + 275./41472.*pow(
         m11,2)*m33*RS346pe*RS345pe + 275./82944.*pow(m11,2)*m33*pow(
         RS346pe,2) - 107./6912.*pow(m11,2)*m33*RS356pe - 275./20736.*
         pow(m11,2)*m33*RS356pe*RS34561pe - 169./6912.*pow(m11,2)*m33*
         RS356pe*RS14563pe - 187./20736.*pow(m11,2)*m33*RS356pe*
         RSe45613p - 187./20736.*pow(m11,2)*m33*RS356pe*RSp45613e - 275.
         /41472.*pow(m11,2)*m33*RS356pe*RS3456pee - 275./41472.*pow(
         m11,2)*m33*RS356pe*RS3456ppe + 187./41472.*pow(m11,2)*m33*
         RS356pe*RSe45633p + 187./41472.*pow(m11,2)*m33*RS356pe*
         RSp45633e );

      M13p6x23R +=  + pi162 * ( 275./41472.*pow(m11,2)*m33*RS356pe*
         RS345pe + 275./41472.*pow(m11,2)*m33*RS356pe*RS346pe + 275./
         82944.*pow(m11,2)*m33*pow(RS356pe,2) - 25./13824.*pow(m11,2)*
         m33*RSe563p - 25./13824.*pow(m11,2)*m33*RSe463p - 25./13824.*
         pow(m11,2)*m33*RSe453p - 25./13824.*pow(m11,2)*m33*RSp563e - 
         25./13824.*pow(m11,2)*m33*RSp463e - 25./13824.*pow(m11,2)*m33*
         RSp453e + 151./4608.*pow(m11,2)*m33*RS1456pee + 169./6912.*
         pow(m11,2)*m33*RS1456pee*RS34561pe + 595./20736.*pow(m11,2)*
         m33*RS1456pee*RS14563pe + 275./20736.*pow(m11,2)*m33*RS1456pee
         *RSe45613p + 275./20736.*pow(m11,2)*m33*RS1456pee*RSp45613e + 
         169./13824.*pow(m11,2)*m33*RS1456pee*RS3456pee + 169./13824.*
         pow(m11,2)*m33*RS1456pee*RS3456ppe - 275./41472.*pow(m11,2)*
         m33*RS1456pee*RSe45633p - 275./41472.*pow(m11,2)*m33*RS1456pee
         *RSp45633e - 169./13824.*pow(m11,2)*m33*RS1456pee*RS345pe - 
         169./13824.*pow(m11,2)*m33*RS1456pee*RS346pe - 169./13824.*
         pow(m11,2)*m33*RS1456pee*RS356pe );

      M13p6x23R +=  + pi162 * ( 595./82944.*pow(m11,2)*m33*pow(
         RS1456pee,2) + 151./4608.*pow(m11,2)*m33*RS1456ppe + 169./6912.
         *pow(m11,2)*m33*RS1456ppe*RS34561pe + 595./20736.*pow(m11,2)*
         m33*RS1456ppe*RS14563pe + 275./20736.*pow(m11,2)*m33*RS1456ppe
         *RSe45613p + 275./20736.*pow(m11,2)*m33*RS1456ppe*RSp45613e + 
         169./13824.*pow(m11,2)*m33*RS1456ppe*RS3456pee + 169./13824.*
         pow(m11,2)*m33*RS1456ppe*RS3456ppe - 275./41472.*pow(m11,2)*
         m33*RS1456ppe*RSe45633p - 275./41472.*pow(m11,2)*m33*RS1456ppe
         *RSp45633e - 169./13824.*pow(m11,2)*m33*RS1456ppe*RS345pe - 
         169./13824.*pow(m11,2)*m33*RS1456ppe*RS346pe - 169./13824.*
         pow(m11,2)*m33*RS1456ppe*RS356pe + 595./41472.*pow(m11,2)*m33*
         RS1456ppe*RS1456pee + 595./82944.*pow(m11,2)*m33*pow(
         RS1456ppe,2) - 59./4608.*pow(m11,2)*m33*RSe45611p - 187./20736.
         *pow(m11,2)*m33*RSe45611p*RS34561pe - 275./20736.*pow(m11,2)*
         m33*RSe45611p*RS14563pe - 11./2304.*pow(m11,2)*m33*RSe45611p*
         RSe45613p );

      M13p6x23R +=  + pi162 * (  - 11./2304.*pow(m11,2)*m33*RSe45611p*
         RSp45613e - 187./41472.*pow(m11,2)*m33*RSe45611p*RS3456pee - 
         187./41472.*pow(m11,2)*m33*RSe45611p*RS3456ppe + 11./4608.*
         pow(m11,2)*m33*RSe45611p*RSe45633p + 11./4608.*pow(m11,2)*m33*
         RSe45611p*RSp45633e + 187./41472.*pow(m11,2)*m33*RSe45611p*
         RS345pe + 187./41472.*pow(m11,2)*m33*RSe45611p*RS346pe + 187./
         41472.*pow(m11,2)*m33*RSe45611p*RS356pe - 275./41472.*pow(
         m11,2)*m33*RSe45611p*RS1456pee - 275./41472.*pow(m11,2)*m33*
         RSe45611p*RS1456ppe + 11./9216.*pow(m11,2)*m33*pow(
         RSe45611p,2) - 59./4608.*pow(m11,2)*m33*RSp45611e - 187./20736.
         *pow(m11,2)*m33*RSp45611e*RS34561pe - 275./20736.*pow(m11,2)*
         m33*RSp45611e*RS14563pe - 11./2304.*pow(m11,2)*m33*RSp45611e*
         RSe45613p - 11./2304.*pow(m11,2)*m33*RSp45611e*RSp45613e - 187.
         /41472.*pow(m11,2)*m33*RSp45611e*RS3456pee - 187./41472.*pow(
         m11,2)*m33*RSp45611e*RS3456ppe + 11./4608.*pow(m11,2)*m33*
         RSp45611e*RSe45633p );

      M13p6x23R +=  + pi162 * ( 11./4608.*pow(m11,2)*m33*RSp45611e*
         RSp45633e + 187./41472.*pow(m11,2)*m33*RSp45611e*RS345pe + 187.
         /41472.*pow(m11,2)*m33*RSp45611e*RS346pe + 187./41472.*pow(
         m11,2)*m33*RSp45611e*RS356pe - 275./41472.*pow(m11,2)*m33*
         RSp45611e*RS1456pee - 275./41472.*pow(m11,2)*m33*RSp45611e*
         RS1456ppe + 11./4608.*pow(m11,2)*m33*RSp45611e*RSe45611p + 11./
         9216.*pow(m11,2)*m33*pow(RSp45611e,2) - 31./6912.*pow(m11,2)*
         m33*RS145pe - 169./6912.*pow(m11,2)*m33*RS145pe*RS34561pe - 
         595./20736.*pow(m11,2)*m33*RS145pe*RS14563pe - 275./20736.*
         pow(m11,2)*m33*RS145pe*RSe45613p - 275./20736.*pow(m11,2)*m33*
         RS145pe*RSp45613e - 169./13824.*pow(m11,2)*m33*RS145pe*
         RS3456pee - 169./13824.*pow(m11,2)*m33*RS145pe*RS3456ppe + 275.
         /41472.*pow(m11,2)*m33*RS145pe*RSe45633p + 275./41472.*pow(
         m11,2)*m33*RS145pe*RSp45633e + 169./13824.*pow(m11,2)*m33*
         RS145pe*RS345pe + 169./13824.*pow(m11,2)*m33*RS145pe*RS346pe
          + 169./13824.*pow(m11,2)*m33*RS145pe*RS356pe );

      M13p6x23R +=  + pi162 * (  - 595./41472.*pow(m11,2)*m33*RS145pe*
         RS1456pee - 595./41472.*pow(m11,2)*m33*RS145pe*RS1456ppe + 275.
         /41472.*pow(m11,2)*m33*RS145pe*RSe45611p + 275./41472.*pow(
         m11,2)*m33*RS145pe*RSp45611e + 595./82944.*pow(m11,2)*m33*pow(
         RS145pe,2) - 31./6912.*pow(m11,2)*m33*RS146pe - 169./6912.*
         pow(m11,2)*m33*RS146pe*RS34561pe - 595./20736.*pow(m11,2)*m33*
         RS146pe*RS14563pe - 275./20736.*pow(m11,2)*m33*RS146pe*
         RSe45613p - 275./20736.*pow(m11,2)*m33*RS146pe*RSp45613e - 169.
         /13824.*pow(m11,2)*m33*RS146pe*RS3456pee - 169./13824.*pow(
         m11,2)*m33*RS146pe*RS3456ppe + 275./41472.*pow(m11,2)*m33*
         RS146pe*RSe45633p + 275./41472.*pow(m11,2)*m33*RS146pe*
         RSp45633e + 169./13824.*pow(m11,2)*m33*RS146pe*RS345pe + 169./
         13824.*pow(m11,2)*m33*RS146pe*RS346pe + 169./13824.*pow(m11,2)
         *m33*RS146pe*RS356pe - 595./41472.*pow(m11,2)*m33*RS146pe*
         RS1456pee - 595./41472.*pow(m11,2)*m33*RS146pe*RS1456ppe + 275.
         /41472.*pow(m11,2)*m33*RS146pe*RSe45611p );

      M13p6x23R +=  + pi162 * ( 275./41472.*pow(m11,2)*m33*RS146pe*
         RSp45611e + 595./41472.*pow(m11,2)*m33*RS146pe*RS145pe + 595./
         82944.*pow(m11,2)*m33*pow(RS146pe,2) - 31./6912.*pow(m11,2)*
         m33*RS156pe - 169./6912.*pow(m11,2)*m33*RS156pe*RS34561pe - 
         595./20736.*pow(m11,2)*m33*RS156pe*RS14563pe - 275./20736.*
         pow(m11,2)*m33*RS156pe*RSe45613p - 275./20736.*pow(m11,2)*m33*
         RS156pe*RSp45613e - 169./13824.*pow(m11,2)*m33*RS156pe*
         RS3456pee - 169./13824.*pow(m11,2)*m33*RS156pe*RS3456ppe + 275.
         /41472.*pow(m11,2)*m33*RS156pe*RSe45633p + 275./41472.*pow(
         m11,2)*m33*RS156pe*RSp45633e + 169./13824.*pow(m11,2)*m33*
         RS156pe*RS345pe + 169./13824.*pow(m11,2)*m33*RS156pe*RS346pe
          + 169./13824.*pow(m11,2)*m33*RS156pe*RS356pe - 595./41472.*
         pow(m11,2)*m33*RS156pe*RS1456pee - 595./41472.*pow(m11,2)*m33*
         RS156pe*RS1456ppe + 275./41472.*pow(m11,2)*m33*RS156pe*
         RSe45611p + 275./41472.*pow(m11,2)*m33*RS156pe*RSp45611e + 595.
         /41472.*pow(m11,2)*m33*RS156pe*RS145pe );

      M13p6x23R +=  + pi162 * ( 595./41472.*pow(m11,2)*m33*RS156pe*
         RS146pe + 595./82944.*pow(m11,2)*m33*pow(RS156pe,2) - 5./4608.
         *pow(m11,2)*m33*RS645pe - 5./4608.*pow(m11,2)*m33*RS546pe - 5./
         4608.*pow(m11,2)*m33*RS456pe + 55./13824.*pow(m11,2)*m33*
         RSe561p + 55./13824.*pow(m11,2)*m33*RSe461p + 55./13824.*pow(
         m11,2)*m33*RSe451p + 55./13824.*pow(m11,2)*m33*RSp561e + 55./
         13824.*pow(m11,2)*m33*RSp461e + 55./13824.*pow(m11,2)*m33*
         RSp451e - 5./4608.*pow(m11,2)*m33*RSe564p - 5./4608.*pow(
         m11,2)*m33*RSe465p - 5./4608.*pow(m11,2)*m33*RSe456p - 5./4608.
         *pow(m11,2)*m33*RSp564e - 5./4608.*pow(m11,2)*m33*RSp465e - 5./
         4608.*pow(m11,2)*m33*RSp456e - 47./512.*pow(m11,3) + 11./1536.
         *pow(m11,3)*RS34561pe + 11./6912.*pow(m11,3)*pow(RS34561pe,2)
          - 55./4608.*pow(m11,3)*RS14563pe + 121./10368.*pow(m11,3)*
         RS14563pe*RS34561pe + 353./20736.*pow(m11,3)*pow(RS14563pe,2)
          + 11./1536.*pow(m11,3)*RSe45613p + 11./3456.*pow(m11,3)*
         RSe45613p*RS34561pe );

      M13p6x23R +=  + pi162 * ( 121./10368.*pow(m11,3)*RSe45613p*
         RS14563pe + 11./6912.*pow(m11,3)*pow(RSe45613p,2) + 11./1536.*
         pow(m11,3)*RSp45613e + 11./3456.*pow(m11,3)*RSp45613e*
         RS34561pe + 121./10368.*pow(m11,3)*RSp45613e*RS14563pe + 11./
         3456.*pow(m11,3)*RSp45613e*RSe45613p + 11./6912.*pow(m11,3)*
         pow(RSp45613e,2) - 7./4608.*pow(m11,3)*RS3456pee + 11./6912.*
         pow(m11,3)*RS3456pee*RS34561pe + 121./20736.*pow(m11,3)*
         RS3456pee*RS14563pe + 11./6912.*pow(m11,3)*RS3456pee*RSe45613p
          + 11./6912.*pow(m11,3)*RS3456pee*RSp45613e + 11./27648.*pow(
         m11,3)*pow(RS3456pee,2) - 7./4608.*pow(m11,3)*RS3456ppe + 11./
         6912.*pow(m11,3)*RS3456ppe*RS34561pe + 121./20736.*pow(m11,3)*
         RS3456ppe*RS14563pe + 11./6912.*pow(m11,3)*RS3456ppe*RSe45613p
          + 11./6912.*pow(m11,3)*RS3456ppe*RSp45613e + 11./13824.*pow(
         m11,3)*RS3456ppe*RS3456pee + 11./27648.*pow(m11,3)*pow(
         RS3456ppe,2) + 7./4608.*pow(m11,3)*RSe45633p - 11./6912.*pow(
         m11,3)*RSe45633p*RS34561pe );

      M13p6x23R +=  + pi162 * (  - 121./20736.*pow(m11,3)*RSe45633p*
         RS14563pe - 11./6912.*pow(m11,3)*RSe45633p*RSe45613p - 11./
         6912.*pow(m11,3)*RSe45633p*RSp45613e - 11./13824.*pow(m11,3)*
         RSe45633p*RS3456pee - 11./13824.*pow(m11,3)*RSe45633p*
         RS3456ppe + 11./27648.*pow(m11,3)*pow(RSe45633p,2) + 7./4608.*
         pow(m11,3)*RSp45633e - 11./6912.*pow(m11,3)*RSp45633e*
         RS34561pe - 121./20736.*pow(m11,3)*RSp45633e*RS14563pe - 11./
         6912.*pow(m11,3)*RSp45633e*RSe45613p - 11./6912.*pow(m11,3)*
         RSp45633e*RSp45613e - 11./13824.*pow(m11,3)*RSp45633e*
         RS3456pee - 11./13824.*pow(m11,3)*RSp45633e*RS3456ppe + 11./
         13824.*pow(m11,3)*RSp45633e*RSe45633p + 11./27648.*pow(m11,3)*
         pow(RSp45633e,2) - 13./6912.*pow(m11,3)*RS345pe - 11./6912.*
         pow(m11,3)*RS345pe*RS34561pe - 121./20736.*pow(m11,3)*RS345pe*
         RS14563pe - 11./6912.*pow(m11,3)*RS345pe*RSe45613p - 11./6912.
         *pow(m11,3)*RS345pe*RSp45613e - 11./13824.*pow(m11,3)*RS345pe*
         RS3456pee );

      M13p6x23R +=  + pi162 * (  - 11./13824.*pow(m11,3)*RS345pe*
         RS3456ppe + 11./13824.*pow(m11,3)*RS345pe*RSe45633p + 11./
         13824.*pow(m11,3)*RS345pe*RSp45633e + 11./27648.*pow(m11,3)*
         pow(RS345pe,2) - 13./6912.*pow(m11,3)*RS346pe - 11./6912.*pow(
         m11,3)*RS346pe*RS34561pe - 121./20736.*pow(m11,3)*RS346pe*
         RS14563pe - 11./6912.*pow(m11,3)*RS346pe*RSe45613p - 11./6912.
         *pow(m11,3)*RS346pe*RSp45613e - 11./13824.*pow(m11,3)*RS346pe*
         RS3456pee - 11./13824.*pow(m11,3)*RS346pe*RS3456ppe + 11./
         13824.*pow(m11,3)*RS346pe*RSe45633p + 11./13824.*pow(m11,3)*
         RS346pe*RSp45633e + 11./13824.*pow(m11,3)*RS346pe*RS345pe + 11.
         /27648.*pow(m11,3)*pow(RS346pe,2) - 13./6912.*pow(m11,3)*
         RS356pe - 11./6912.*pow(m11,3)*RS356pe*RS34561pe - 121./20736.
         *pow(m11,3)*RS356pe*RS14563pe - 11./6912.*pow(m11,3)*RS356pe*
         RSe45613p - 11./6912.*pow(m11,3)*RS356pe*RSp45613e - 11./13824.
         *pow(m11,3)*RS356pe*RS3456pee - 11./13824.*pow(m11,3)*RS356pe*
         RS3456ppe );

      M13p6x23R +=  + pi162 * ( 11./13824.*pow(m11,3)*RS356pe*RSe45633p
          + 11./13824.*pow(m11,3)*RS356pe*RSp45633e + 11./13824.*pow(
         m11,3)*RS356pe*RS345pe + 11./13824.*pow(m11,3)*RS356pe*RS346pe
          + 11./27648.*pow(m11,3)*pow(RS356pe,2) - 47./13824.*pow(
         m11,3)*RSe563p - 47./13824.*pow(m11,3)*RSe463p - 47./13824.*
         pow(m11,3)*RSe453p - 47./13824.*pow(m11,3)*RSp563e - 47./13824.
         *pow(m11,3)*RSp463e - 47./13824.*pow(m11,3)*RSp453e + 73./4608.
         *pow(m11,3)*RS1456pee + 121./20736.*pow(m11,3)*RS1456pee*
         RS34561pe + 353./20736.*pow(m11,3)*RS1456pee*RS14563pe + 121./
         20736.*pow(m11,3)*RS1456pee*RSe45613p + 121./20736.*pow(m11,3)
         *RS1456pee*RSp45613e + 121./41472.*pow(m11,3)*RS1456pee*
         RS3456pee + 121./41472.*pow(m11,3)*RS1456pee*RS3456ppe - 121./
         41472.*pow(m11,3)*RS1456pee*RSe45633p - 121./41472.*pow(m11,3)
         *RS1456pee*RSp45633e - 121./41472.*pow(m11,3)*RS1456pee*
         RS345pe - 121./41472.*pow(m11,3)*RS1456pee*RS346pe - 121./
         41472.*pow(m11,3)*RS1456pee*RS356pe );

      M13p6x23R +=  + pi162 * ( 353./82944.*pow(m11,3)*pow(RS1456pee,2)
          + 73./4608.*pow(m11,3)*RS1456ppe + 121./20736.*pow(m11,3)*
         RS1456ppe*RS34561pe + 353./20736.*pow(m11,3)*RS1456ppe*
         RS14563pe + 121./20736.*pow(m11,3)*RS1456ppe*RSe45613p + 121./
         20736.*pow(m11,3)*RS1456ppe*RSp45613e + 121./41472.*pow(m11,3)
         *RS1456ppe*RS3456pee + 121./41472.*pow(m11,3)*RS1456ppe*
         RS3456ppe - 121./41472.*pow(m11,3)*RS1456ppe*RSe45633p - 121./
         41472.*pow(m11,3)*RS1456ppe*RSp45633e - 121./41472.*pow(m11,3)
         *RS1456ppe*RS345pe - 121./41472.*pow(m11,3)*RS1456ppe*RS346pe
          - 121./41472.*pow(m11,3)*RS1456ppe*RS356pe + 353./41472.*pow(
         m11,3)*RS1456ppe*RS1456pee + 353./82944.*pow(m11,3)*pow(
         RS1456ppe,2) - 11./1536.*pow(m11,3)*RSe45611p - 11./6912.*pow(
         m11,3)*RSe45611p*RS34561pe - 121./20736.*pow(m11,3)*RSe45611p*
         RS14563pe - 11./6912.*pow(m11,3)*RSe45611p*RSe45613p - 11./
         6912.*pow(m11,3)*RSe45611p*RSp45613e - 11./13824.*pow(m11,3)*
         RSe45611p*RS3456pee );

      M13p6x23R +=  + pi162 * (  - 11./13824.*pow(m11,3)*RSe45611p*
         RS3456ppe + 11./13824.*pow(m11,3)*RSe45611p*RSe45633p + 11./
         13824.*pow(m11,3)*RSe45611p*RSp45633e + 11./13824.*pow(m11,3)*
         RSe45611p*RS345pe + 11./13824.*pow(m11,3)*RSe45611p*RS346pe + 
         11./13824.*pow(m11,3)*RSe45611p*RS356pe - 121./41472.*pow(
         m11,3)*RSe45611p*RS1456pee - 121./41472.*pow(m11,3)*RSe45611p*
         RS1456ppe + 11./27648.*pow(m11,3)*pow(RSe45611p,2) - 11./1536.
         *pow(m11,3)*RSp45611e - 11./6912.*pow(m11,3)*RSp45611e*
         RS34561pe - 121./20736.*pow(m11,3)*RSp45611e*RS14563pe - 11./
         6912.*pow(m11,3)*RSp45611e*RSe45613p - 11./6912.*pow(m11,3)*
         RSp45611e*RSp45613e - 11./13824.*pow(m11,3)*RSp45611e*
         RS3456pee - 11./13824.*pow(m11,3)*RSp45611e*RS3456ppe + 11./
         13824.*pow(m11,3)*RSp45611e*RSe45633p + 11./13824.*pow(m11,3)*
         RSp45611e*RSp45633e + 11./13824.*pow(m11,3)*RSp45611e*RS345pe
          + 11./13824.*pow(m11,3)*RSp45611e*RS346pe + 11./13824.*pow(
         m11,3)*RSp45611e*RS356pe );

      M13p6x23R +=  + pi162 * (  - 121./41472.*pow(m11,3)*RSp45611e*
         RS1456pee - 121./41472.*pow(m11,3)*RSp45611e*RS1456ppe + 11./
         13824.*pow(m11,3)*RSp45611e*RSe45611p + 11./27648.*pow(m11,3)*
         pow(RSp45611e,2) + 7./768.*pow(m11,3)*RS145pe - 121./20736.*
         pow(m11,3)*RS145pe*RS34561pe - 353./20736.*pow(m11,3)*RS145pe*
         RS14563pe - 121./20736.*pow(m11,3)*RS145pe*RSe45613p - 121./
         20736.*pow(m11,3)*RS145pe*RSp45613e - 121./41472.*pow(m11,3)*
         RS145pe*RS3456pee - 121./41472.*pow(m11,3)*RS145pe*RS3456ppe
          + 121./41472.*pow(m11,3)*RS145pe*RSe45633p + 121./41472.*pow(
         m11,3)*RS145pe*RSp45633e + 121./41472.*pow(m11,3)*RS145pe*
         RS345pe + 121./41472.*pow(m11,3)*RS145pe*RS346pe + 121./41472.
         *pow(m11,3)*RS145pe*RS356pe - 353./41472.*pow(m11,3)*RS145pe*
         RS1456pee - 353./41472.*pow(m11,3)*RS145pe*RS1456ppe + 121./
         41472.*pow(m11,3)*RS145pe*RSe45611p + 121./41472.*pow(m11,3)*
         RS145pe*RSp45611e + 353./82944.*pow(m11,3)*pow(RS145pe,2) + 7./
         768.*pow(m11,3)*RS146pe );

      M13p6x23R +=  + pi162 * (  - 121./20736.*pow(m11,3)*RS146pe*
         RS34561pe - 353./20736.*pow(m11,3)*RS146pe*RS14563pe - 121./
         20736.*pow(m11,3)*RS146pe*RSe45613p - 121./20736.*pow(m11,3)*
         RS146pe*RSp45613e - 121./41472.*pow(m11,3)*RS146pe*RS3456pee
          - 121./41472.*pow(m11,3)*RS146pe*RS3456ppe + 121./41472.*pow(
         m11,3)*RS146pe*RSe45633p + 121./41472.*pow(m11,3)*RS146pe*
         RSp45633e + 121./41472.*pow(m11,3)*RS146pe*RS345pe + 121./
         41472.*pow(m11,3)*RS146pe*RS346pe + 121./41472.*pow(m11,3)*
         RS146pe*RS356pe - 353./41472.*pow(m11,3)*RS146pe*RS1456pee - 
         353./41472.*pow(m11,3)*RS146pe*RS1456ppe + 121./41472.*pow(
         m11,3)*RS146pe*RSe45611p + 121./41472.*pow(m11,3)*RS146pe*
         RSp45611e + 353./41472.*pow(m11,3)*RS146pe*RS145pe + 353./
         82944.*pow(m11,3)*pow(RS146pe,2) + 7./768.*pow(m11,3)*RS156pe
          - 121./20736.*pow(m11,3)*RS156pe*RS34561pe - 353./20736.*pow(
         m11,3)*RS156pe*RS14563pe - 121./20736.*pow(m11,3)*RS156pe*
         RSe45613p );

      M13p6x23R +=  + pi162 * (  - 121./20736.*pow(m11,3)*RS156pe*
         RSp45613e - 121./41472.*pow(m11,3)*RS156pe*RS3456pee - 121./
         41472.*pow(m11,3)*RS156pe*RS3456ppe + 121./41472.*pow(m11,3)*
         RS156pe*RSe45633p + 121./41472.*pow(m11,3)*RS156pe*RSp45633e
          + 121./41472.*pow(m11,3)*RS156pe*RS345pe + 121./41472.*pow(
         m11,3)*RS156pe*RS346pe + 121./41472.*pow(m11,3)*RS156pe*
         RS356pe - 353./41472.*pow(m11,3)*RS156pe*RS1456pee - 353./
         41472.*pow(m11,3)*RS156pe*RS1456ppe + 121./41472.*pow(m11,3)*
         RS156pe*RSe45611p + 121./41472.*pow(m11,3)*RS156pe*RSp45611e
          + 353./41472.*pow(m11,3)*RS156pe*RS145pe + 353./41472.*pow(
         m11,3)*RS156pe*RS146pe + 353./82944.*pow(m11,3)*pow(RS156pe,2)
          + 7./13824.*pow(m11,3)*RS645pe + 7./13824.*pow(m11,3)*RS546pe
          + 7./13824.*pow(m11,3)*RS456pe + 11./4608.*pow(m11,3)*RSe561p
          + 11./4608.*pow(m11,3)*RSe461p + 11./4608.*pow(m11,3)*RSe451p
          + 11./4608.*pow(m11,3)*RSp561e + 11./4608.*pow(m11,3)*RSp461e
          + 11./4608.*pow(m11,3)*RSp451e );

      M13p6x23R +=  + pi162 * ( 7./13824.*pow(m11,3)*RSe564p + 7./13824.
         *pow(m11,3)*RSe465p + 7./13824.*pow(m11,3)*RSe456p + 7./13824.
         *pow(m11,3)*RSp564e + 7./13824.*pow(m11,3)*RSp465e + 7./13824.
         *pow(m11,3)*RSp456e );

      M13p6x23R +=  + Ab(m11,mu2)*pi16 * (  - 1./72.*m33*RD3456ep*
         RS14563pe + 1./48.*m33*RD1456ep - 1./72.*m33*RD1456ep*
         RS14563pe - 1./144.*m33*RS1456pee*RD3456ep - 1./144.*m33*
         RS1456pee*RD1456ep - 1./144.*m33*RS1456ppe*RD3456ep - 1./144.*
         m33*RS1456ppe*RD1456ep + 1./144.*m33*RS145pe*RD3456ep + 1./144.
         *m33*RS145pe*RD1456ep + 1./144.*m33*RS146pe*RD3456ep + 1./144.
         *m33*RS146pe*RD1456ep + 1./144.*m33*RS156pe*RD3456ep + 1./144.
         *m33*RS156pe*RD1456ep + 1./36.*m33*mp2*RSp45613e*RS14563pe - 1.
         /72.*m33*mp2*RSp45633e*RS14563pe + 1./72.*m33*mp2*RS1456pee*
         RSp45613e - 1./144.*m33*mp2*RS1456pee*RSp45633e + 1./72.*m33*
         mp2*RS1456ppe*RSp45613e - 1./144.*m33*mp2*RS1456ppe*RSp45633e
          - 1./72.*m33*mp2*RSp45611e*RS14563pe - 1./144.*m33*mp2*
         RSp45611e*RS1456pee - 1./144.*m33*mp2*RSp45611e*RS1456ppe - 1./
         72.*m33*mp2*RS145pe*RSp45613e + 1./144.*m33*mp2*RS145pe*
         RSp45633e + 1./144.*m33*mp2*RS145pe*RSp45611e - 1./72.*m33*mp2
         *RS146pe*RSp45613e );

      M13p6x23R +=  + Ab(m11,mu2)*pi16 * ( 1./144.*m33*mp2*RS146pe*
         RSp45633e + 1./144.*m33*mp2*RS146pe*RSp45611e - 1./72.*m33*mp2
         *RS156pe*RSp45613e + 1./144.*m33*mp2*RS156pe*RSp45633e + 1./
         144.*m33*mp2*RS156pe*RSp45611e + 1./36.*m33*me2*RSe45613p*
         RS14563pe - 1./72.*m33*me2*RSe45633p*RS14563pe + 1./72.*m33*
         me2*RS1456pee*RSe45613p - 1./144.*m33*me2*RS1456pee*RSe45633p
          + 1./72.*m33*me2*RS1456ppe*RSe45613p - 1./144.*m33*me2*
         RS1456ppe*RSe45633p - 1./72.*m33*me2*RSe45611p*RS14563pe - 1./
         144.*m33*me2*RSe45611p*RS1456pee - 1./144.*m33*me2*RSe45611p*
         RS1456ppe - 1./72.*m33*me2*RS145pe*RSe45613p + 1./144.*m33*me2
         *RS145pe*RSe45633p + 1./144.*m33*me2*RS145pe*RSe45611p - 1./72.
         *m33*me2*RS146pe*RSe45613p + 1./144.*m33*me2*RS146pe*RSe45633p
          + 1./144.*m33*me2*RS146pe*RSe45611p - 1./72.*m33*me2*RS156pe*
         RSe45613p + 1./144.*m33*me2*RS156pe*RSe45633p + 1./144.*m33*
         me2*RS156pe*RSe45611p + 1./24.*m33*m66*RS14563pe + 1./48.*m33*
         m66*RS1456pee );

      M13p6x23R +=  + Ab(m11,mu2)*pi16 * ( 1./48.*m33*m66*RS1456ppe - 1.
         /48.*m33*m66*RS145pe - 1./48.*m33*m66*RS146pe - 1./48.*m33*m66
         *RS156pe + 1./24.*m33*m55*RS14563pe + 1./48.*m33*m55*RS1456pee
          + 1./48.*m33*m55*RS1456ppe - 1./48.*m33*m55*RS145pe - 1./48.*
         m33*m55*RS146pe - 1./48.*m33*m55*RS156pe + 1./24.*m33*m44*
         RS14563pe + 1./48.*m33*m44*RS1456pee + 1./48.*m33*m44*
         RS1456ppe - 1./48.*m33*m44*RS145pe - 1./48.*m33*m44*RS146pe - 
         1./48.*m33*m44*RS156pe + 1./24.*pow(m33,2)*RS14563pe + 1./27.*
         pow(m33,2)*RS14563pe*RS34561pe + 1./108.*pow(m33,2)*pow(
         RS14563pe,2) + 1./108.*pow(m33,2)*RSe45613p*RS14563pe + 1./108.
         *pow(m33,2)*RSp45613e*RS14563pe + 1./54.*pow(m33,2)*RS3456pee*
         RS14563pe + 1./54.*pow(m33,2)*RS3456ppe*RS14563pe - 1./216.*
         pow(m33,2)*RSe45633p*RS14563pe - 1./216.*pow(m33,2)*RSp45633e*
         RS14563pe - 1./54.*pow(m33,2)*RS345pe*RS14563pe - 1./54.*pow(
         m33,2)*RS346pe*RS14563pe - 1./54.*pow(m33,2)*RS356pe*RS14563pe
          + 1./48.*pow(m33,2)*RS1456pee );

      M13p6x23R +=  + Ab(m11,mu2)*pi16 * ( 1./54.*pow(m33,2)*RS1456pee*
         RS34561pe + 1./108.*pow(m33,2)*RS1456pee*RS14563pe + 1./216.*
         pow(m33,2)*RS1456pee*RSe45613p + 1./216.*pow(m33,2)*RS1456pee*
         RSp45613e + 1./108.*pow(m33,2)*RS1456pee*RS3456pee + 1./108.*
         pow(m33,2)*RS1456pee*RS3456ppe - 1./432.*pow(m33,2)*RS1456pee*
         RSe45633p - 1./432.*pow(m33,2)*RS1456pee*RSp45633e - 1./108.*
         pow(m33,2)*RS1456pee*RS345pe - 1./108.*pow(m33,2)*RS1456pee*
         RS346pe - 1./108.*pow(m33,2)*RS1456pee*RS356pe + 1./432.*pow(
         m33,2)*pow(RS1456pee,2) + 1./48.*pow(m33,2)*RS1456ppe + 1./54.
         *pow(m33,2)*RS1456ppe*RS34561pe + 1./108.*pow(m33,2)*RS1456ppe
         *RS14563pe + 1./216.*pow(m33,2)*RS1456ppe*RSe45613p + 1./216.*
         pow(m33,2)*RS1456ppe*RSp45613e + 1./108.*pow(m33,2)*RS1456ppe*
         RS3456pee + 1./108.*pow(m33,2)*RS1456ppe*RS3456ppe - 1./432.*
         pow(m33,2)*RS1456ppe*RSe45633p - 1./432.*pow(m33,2)*RS1456ppe*
         RSp45633e - 1./108.*pow(m33,2)*RS1456ppe*RS345pe - 1./108.*
         pow(m33,2)*RS1456ppe*RS346pe );

      M13p6x23R +=  + Ab(m11,mu2)*pi16 * (  - 1./108.*pow(m33,2)*
         RS1456ppe*RS356pe + 1./216.*pow(m33,2)*RS1456ppe*RS1456pee + 1.
         /432.*pow(m33,2)*pow(RS1456ppe,2) - 1./216.*pow(m33,2)*
         RSe45611p*RS14563pe - 1./432.*pow(m33,2)*RSe45611p*RS1456pee
          - 1./432.*pow(m33,2)*RSe45611p*RS1456ppe - 1./216.*pow(m33,2)
         *RSp45611e*RS14563pe - 1./432.*pow(m33,2)*RSp45611e*RS1456pee
          - 1./432.*pow(m33,2)*RSp45611e*RS1456ppe - 1./48.*pow(m33,2)*
         RS145pe - 1./54.*pow(m33,2)*RS145pe*RS34561pe - 1./108.*pow(
         m33,2)*RS145pe*RS14563pe - 1./216.*pow(m33,2)*RS145pe*
         RSe45613p - 1./216.*pow(m33,2)*RS145pe*RSp45613e - 1./108.*
         pow(m33,2)*RS145pe*RS3456pee - 1./108.*pow(m33,2)*RS145pe*
         RS3456ppe + 1./432.*pow(m33,2)*RS145pe*RSe45633p + 1./432.*
         pow(m33,2)*RS145pe*RSp45633e + 1./108.*pow(m33,2)*RS145pe*
         RS345pe + 1./108.*pow(m33,2)*RS145pe*RS346pe + 1./108.*pow(
         m33,2)*RS145pe*RS356pe - 1./216.*pow(m33,2)*RS145pe*RS1456pee
          - 1./216.*pow(m33,2)*RS145pe*RS1456ppe );

      M13p6x23R +=  + Ab(m11,mu2)*pi16 * ( 1./432.*pow(m33,2)*RS145pe*
         RSe45611p + 1./432.*pow(m33,2)*RS145pe*RSp45611e + 1./432.*
         pow(m33,2)*pow(RS145pe,2) - 1./48.*pow(m33,2)*RS146pe - 1./54.
         *pow(m33,2)*RS146pe*RS34561pe - 1./108.*pow(m33,2)*RS146pe*
         RS14563pe - 1./216.*pow(m33,2)*RS146pe*RSe45613p - 1./216.*
         pow(m33,2)*RS146pe*RSp45613e - 1./108.*pow(m33,2)*RS146pe*
         RS3456pee - 1./108.*pow(m33,2)*RS146pe*RS3456ppe + 1./432.*
         pow(m33,2)*RS146pe*RSe45633p + 1./432.*pow(m33,2)*RS146pe*
         RSp45633e + 1./108.*pow(m33,2)*RS146pe*RS345pe + 1./108.*pow(
         m33,2)*RS146pe*RS346pe + 1./108.*pow(m33,2)*RS146pe*RS356pe - 
         1./216.*pow(m33,2)*RS146pe*RS1456pee - 1./216.*pow(m33,2)*
         RS146pe*RS1456ppe + 1./432.*pow(m33,2)*RS146pe*RSe45611p + 1./
         432.*pow(m33,2)*RS146pe*RSp45611e + 1./216.*pow(m33,2)*RS146pe
         *RS145pe + 1./432.*pow(m33,2)*pow(RS146pe,2) - 1./48.*pow(
         m33,2)*RS156pe - 1./54.*pow(m33,2)*RS156pe*RS34561pe - 1./108.
         *pow(m33,2)*RS156pe*RS14563pe );

      M13p6x23R +=  + Ab(m11,mu2)*pi16 * (  - 1./216.*pow(m33,2)*
         RS156pe*RSe45613p - 1./216.*pow(m33,2)*RS156pe*RSp45613e - 1./
         108.*pow(m33,2)*RS156pe*RS3456pee - 1./108.*pow(m33,2)*RS156pe
         *RS3456ppe + 1./432.*pow(m33,2)*RS156pe*RSe45633p + 1./432.*
         pow(m33,2)*RS156pe*RSp45633e + 1./108.*pow(m33,2)*RS156pe*
         RS345pe + 1./108.*pow(m33,2)*RS156pe*RS346pe + 1./108.*pow(
         m33,2)*RS156pe*RS356pe - 1./216.*pow(m33,2)*RS156pe*RS1456pee
          - 1./216.*pow(m33,2)*RS156pe*RS1456ppe + 1./432.*pow(m33,2)*
         RS156pe*RSe45611p + 1./432.*pow(m33,2)*RS156pe*RSp45611e + 1./
         216.*pow(m33,2)*RS156pe*RS145pe + 1./216.*pow(m33,2)*RS156pe*
         RS146pe + 1./432.*pow(m33,2)*pow(RS156pe,2) - 1./72.*m11*
         RD3456ep*RS14563pe + 1./48.*m11*RD1456ep - 1./72.*m11*RD1456ep
         *RS14563pe - 1./144.*m11*RS1456pee*RD3456ep - 1./144.*m11*
         RS1456pee*RD1456ep - 1./144.*m11*RS1456ppe*RD3456ep - 1./144.*
         m11*RS1456ppe*RD1456ep + 1./144.*m11*RS145pe*RD3456ep + 1./144.
         *m11*RS145pe*RD1456ep );

      M13p6x23R +=  + Ab(m11,mu2)*pi16 * ( 1./144.*m11*RS146pe*RD3456ep
          + 1./144.*m11*RS146pe*RD1456ep + 1./144.*m11*RS156pe*RD3456ep
          + 1./144.*m11*RS156pe*RD1456ep + 1./36.*m11*mp2*RSp45613e*
         RS14563pe - 1./72.*m11*mp2*RSp45633e*RS14563pe + 1./72.*m11*
         mp2*RS1456pee*RSp45613e - 1./144.*m11*mp2*RS1456pee*RSp45633e
          + 1./72.*m11*mp2*RS1456ppe*RSp45613e - 1./144.*m11*mp2*
         RS1456ppe*RSp45633e - 1./72.*m11*mp2*RSp45611e*RS14563pe - 1./
         144.*m11*mp2*RSp45611e*RS1456pee - 1./144.*m11*mp2*RSp45611e*
         RS1456ppe - 1./72.*m11*mp2*RS145pe*RSp45613e + 1./144.*m11*mp2
         *RS145pe*RSp45633e + 1./144.*m11*mp2*RS145pe*RSp45611e - 1./72.
         *m11*mp2*RS146pe*RSp45613e + 1./144.*m11*mp2*RS146pe*RSp45633e
          + 1./144.*m11*mp2*RS146pe*RSp45611e - 1./72.*m11*mp2*RS156pe*
         RSp45613e + 1./144.*m11*mp2*RS156pe*RSp45633e + 1./144.*m11*
         mp2*RS156pe*RSp45611e + 1./36.*m11*me2*RSe45613p*RS14563pe - 1.
         /72.*m11*me2*RSe45633p*RS14563pe + 1./72.*m11*me2*RS1456pee*
         RSe45613p );

      M13p6x23R +=  + Ab(m11,mu2)*pi16 * (  - 1./144.*m11*me2*RS1456pee
         *RSe45633p + 1./72.*m11*me2*RS1456ppe*RSe45613p - 1./144.*m11*
         me2*RS1456ppe*RSe45633p - 1./72.*m11*me2*RSe45611p*RS14563pe
          - 1./144.*m11*me2*RSe45611p*RS1456pee - 1./144.*m11*me2*
         RSe45611p*RS1456ppe - 1./72.*m11*me2*RS145pe*RSe45613p + 1./
         144.*m11*me2*RS145pe*RSe45633p + 1./144.*m11*me2*RS145pe*
         RSe45611p - 1./72.*m11*me2*RS146pe*RSe45613p + 1./144.*m11*me2
         *RS146pe*RSe45633p + 1./144.*m11*me2*RS146pe*RSe45611p - 1./72.
         *m11*me2*RS156pe*RSe45613p + 1./144.*m11*me2*RS156pe*RSe45633p
          + 1./144.*m11*me2*RS156pe*RSe45611p + 1./24.*m11*m66*
         RS14563pe + 1./48.*m11*m66*RS1456pee + 1./48.*m11*m66*
         RS1456ppe - 1./48.*m11*m66*RS145pe - 1./48.*m11*m66*RS146pe - 
         1./48.*m11*m66*RS156pe + 1./24.*m11*m55*RS14563pe + 1./48.*m11
         *m55*RS1456pee + 1./48.*m11*m55*RS1456ppe - 1./48.*m11*m55*
         RS145pe - 1./48.*m11*m55*RS146pe - 1./48.*m11*m55*RS156pe + 1./
         24.*m11*m44*RS14563pe );

      M13p6x23R +=  + Ab(m11,mu2)*pi16 * ( 1./48.*m11*m44*RS1456pee + 1.
         /48.*m11*m44*RS1456ppe - 1./48.*m11*m44*RS145pe - 1./48.*m11*
         m44*RS146pe - 1./48.*m11*m44*RS156pe - 1./16.*m11*m33 + 1./24.
         *m11*m33*RS14563pe + 5./108.*m11*m33*RS14563pe*RS34561pe + 5./
         108.*m11*m33*pow(RS14563pe,2) + 1./54.*m11*m33*RSe45613p*
         RS14563pe + 1./54.*m11*m33*RSp45613e*RS14563pe + 5./216.*m11*
         m33*RS3456pee*RS14563pe + 5./216.*m11*m33*RS3456ppe*RS14563pe
          - 1./108.*m11*m33*RSe45633p*RS14563pe - 1./108.*m11*m33*
         RSp45633e*RS14563pe - 5./216.*m11*m33*RS345pe*RS14563pe - 5./
         216.*m11*m33*RS346pe*RS14563pe - 5./216.*m11*m33*RS356pe*
         RS14563pe + 1./48.*m11*m33*RS1456pee + 5./216.*m11*m33*
         RS1456pee*RS34561pe + 5./108.*m11*m33*RS1456pee*RS14563pe + 1./
         108.*m11*m33*RS1456pee*RSe45613p + 1./108.*m11*m33*RS1456pee*
         RSp45613e + 5./432.*m11*m33*RS1456pee*RS3456pee + 5./432.*m11*
         m33*RS1456pee*RS3456ppe - 1./216.*m11*m33*RS1456pee*RSe45633p
          - 1./216.*m11*m33*RS1456pee*RSp45633e );

      M13p6x23R +=  + Ab(m11,mu2)*pi16 * (  - 5./432.*m11*m33*RS1456pee
         *RS345pe - 5./432.*m11*m33*RS1456pee*RS346pe - 5./432.*m11*m33
         *RS1456pee*RS356pe + 5./432.*m11*m33*pow(RS1456pee,2) + 1./48.
         *m11*m33*RS1456ppe + 5./216.*m11*m33*RS1456ppe*RS34561pe + 5./
         108.*m11*m33*RS1456ppe*RS14563pe + 1./108.*m11*m33*RS1456ppe*
         RSe45613p + 1./108.*m11*m33*RS1456ppe*RSp45613e + 5./432.*m11*
         m33*RS1456ppe*RS3456pee + 5./432.*m11*m33*RS1456ppe*RS3456ppe
          - 1./216.*m11*m33*RS1456ppe*RSe45633p - 1./216.*m11*m33*
         RS1456ppe*RSp45633e - 5./432.*m11*m33*RS1456ppe*RS345pe - 5./
         432.*m11*m33*RS1456ppe*RS346pe - 5./432.*m11*m33*RS1456ppe*
         RS356pe + 5./216.*m11*m33*RS1456ppe*RS1456pee + 5./432.*m11*
         m33*pow(RS1456ppe,2) - 1./108.*m11*m33*RSe45611p*RS14563pe - 1.
         /216.*m11*m33*RSe45611p*RS1456pee - 1./216.*m11*m33*RSe45611p*
         RS1456ppe - 1./108.*m11*m33*RSp45611e*RS14563pe - 1./216.*m11*
         m33*RSp45611e*RS1456pee - 1./216.*m11*m33*RSp45611e*RS1456ppe
          - 5./216.*m11*m33*RS145pe*RS34561pe );

      M13p6x23R +=  + Ab(m11,mu2)*pi16 * (  - 5./108.*m11*m33*RS145pe*
         RS14563pe - 1./108.*m11*m33*RS145pe*RSe45613p - 1./108.*m11*
         m33*RS145pe*RSp45613e - 5./432.*m11*m33*RS145pe*RS3456pee - 5./
         432.*m11*m33*RS145pe*RS3456ppe + 1./216.*m11*m33*RS145pe*
         RSe45633p + 1./216.*m11*m33*RS145pe*RSp45633e + 5./432.*m11*
         m33*RS145pe*RS345pe + 5./432.*m11*m33*RS145pe*RS346pe + 5./432.
         *m11*m33*RS145pe*RS356pe - 5./216.*m11*m33*RS145pe*RS1456pee
          - 5./216.*m11*m33*RS145pe*RS1456ppe + 1./216.*m11*m33*RS145pe
         *RSe45611p + 1./216.*m11*m33*RS145pe*RSp45611e + 5./432.*m11*
         m33*pow(RS145pe,2) - 5./216.*m11*m33*RS146pe*RS34561pe - 5./
         108.*m11*m33*RS146pe*RS14563pe - 1./108.*m11*m33*RS146pe*
         RSe45613p - 1./108.*m11*m33*RS146pe*RSp45613e - 5./432.*m11*
         m33*RS146pe*RS3456pee - 5./432.*m11*m33*RS146pe*RS3456ppe + 1./
         216.*m11*m33*RS146pe*RSe45633p + 1./216.*m11*m33*RS146pe*
         RSp45633e + 5./432.*m11*m33*RS146pe*RS345pe + 5./432.*m11*m33*
         RS146pe*RS346pe );

      M13p6x23R +=  + Ab(m11,mu2)*pi16 * ( 5./432.*m11*m33*RS146pe*
         RS356pe - 5./216.*m11*m33*RS146pe*RS1456pee - 5./216.*m11*m33*
         RS146pe*RS1456ppe + 1./216.*m11*m33*RS146pe*RSe45611p + 1./216.
         *m11*m33*RS146pe*RSp45611e + 5./216.*m11*m33*RS146pe*RS145pe
          + 5./432.*m11*m33*pow(RS146pe,2) - 5./216.*m11*m33*RS156pe*
         RS34561pe - 5./108.*m11*m33*RS156pe*RS14563pe - 1./108.*m11*
         m33*RS156pe*RSe45613p - 1./108.*m11*m33*RS156pe*RSp45613e - 5./
         432.*m11*m33*RS156pe*RS3456pee - 5./432.*m11*m33*RS156pe*
         RS3456ppe + 1./216.*m11*m33*RS156pe*RSe45633p + 1./216.*m11*
         m33*RS156pe*RSp45633e + 5./432.*m11*m33*RS156pe*RS345pe + 5./
         432.*m11*m33*RS156pe*RS346pe + 5./432.*m11*m33*RS156pe*RS356pe
          - 5./216.*m11*m33*RS156pe*RS1456pee - 5./216.*m11*m33*RS156pe
         *RS1456ppe + 1./216.*m11*m33*RS156pe*RSe45611p + 1./216.*m11*
         m33*RS156pe*RSp45611e + 5./216.*m11*m33*RS156pe*RS145pe + 5./
         216.*m11*m33*RS156pe*RS146pe + 5./432.*m11*m33*pow(RS156pe,2)
          - 1./16.*pow(m11,2) );

      M13p6x23R +=  + Ab(m11,mu2)*pi16 * ( 1./108.*pow(m11,2)*RS14563pe
         *RS34561pe + 1./27.*pow(m11,2)*pow(RS14563pe,2) + 1./108.*pow(
         m11,2)*RSe45613p*RS14563pe + 1./108.*pow(m11,2)*RSp45613e*
         RS14563pe + 1./216.*pow(m11,2)*RS3456pee*RS14563pe + 1./216.*
         pow(m11,2)*RS3456ppe*RS14563pe - 1./216.*pow(m11,2)*RSe45633p*
         RS14563pe - 1./216.*pow(m11,2)*RSp45633e*RS14563pe - 1./216.*
         pow(m11,2)*RS345pe*RS14563pe - 1./216.*pow(m11,2)*RS346pe*
         RS14563pe - 1./216.*pow(m11,2)*RS356pe*RS14563pe + 1./216.*
         pow(m11,2)*RS1456pee*RS34561pe + 1./27.*pow(m11,2)*RS1456pee*
         RS14563pe + 1./216.*pow(m11,2)*RS1456pee*RSe45613p + 1./216.*
         pow(m11,2)*RS1456pee*RSp45613e + 1./432.*pow(m11,2)*RS1456pee*
         RS3456pee + 1./432.*pow(m11,2)*RS1456pee*RS3456ppe - 1./432.*
         pow(m11,2)*RS1456pee*RSe45633p - 1./432.*pow(m11,2)*RS1456pee*
         RSp45633e - 1./432.*pow(m11,2)*RS1456pee*RS345pe - 1./432.*
         pow(m11,2)*RS1456pee*RS346pe - 1./432.*pow(m11,2)*RS1456pee*
         RS356pe );

      M13p6x23R +=  + Ab(m11,mu2)*pi16 * ( 1./108.*pow(m11,2)*pow(
         RS1456pee,2) + 1./216.*pow(m11,2)*RS1456ppe*RS34561pe + 1./27.
         *pow(m11,2)*RS1456ppe*RS14563pe + 1./216.*pow(m11,2)*RS1456ppe
         *RSe45613p + 1./216.*pow(m11,2)*RS1456ppe*RSp45613e + 1./432.*
         pow(m11,2)*RS1456ppe*RS3456pee + 1./432.*pow(m11,2)*RS1456ppe*
         RS3456ppe - 1./432.*pow(m11,2)*RS1456ppe*RSe45633p - 1./432.*
         pow(m11,2)*RS1456ppe*RSp45633e - 1./432.*pow(m11,2)*RS1456ppe*
         RS345pe - 1./432.*pow(m11,2)*RS1456ppe*RS346pe - 1./432.*pow(
         m11,2)*RS1456ppe*RS356pe + 1./54.*pow(m11,2)*RS1456ppe*
         RS1456pee + 1./108.*pow(m11,2)*pow(RS1456ppe,2) - 1./216.*pow(
         m11,2)*RSe45611p*RS14563pe - 1./432.*pow(m11,2)*RSe45611p*
         RS1456pee - 1./432.*pow(m11,2)*RSe45611p*RS1456ppe - 1./216.*
         pow(m11,2)*RSp45611e*RS14563pe - 1./432.*pow(m11,2)*RSp45611e*
         RS1456pee - 1./432.*pow(m11,2)*RSp45611e*RS1456ppe + 1./48.*
         pow(m11,2)*RS145pe - 1./216.*pow(m11,2)*RS145pe*RS34561pe - 1./
         27.*pow(m11,2)*RS145pe*RS14563pe );

      M13p6x23R +=  + Ab(m11,mu2)*pi16 * (  - 1./216.*pow(m11,2)*
         RS145pe*RSe45613p - 1./216.*pow(m11,2)*RS145pe*RSp45613e - 1./
         432.*pow(m11,2)*RS145pe*RS3456pee - 1./432.*pow(m11,2)*RS145pe
         *RS3456ppe + 1./432.*pow(m11,2)*RS145pe*RSe45633p + 1./432.*
         pow(m11,2)*RS145pe*RSp45633e + 1./432.*pow(m11,2)*RS145pe*
         RS345pe + 1./432.*pow(m11,2)*RS145pe*RS346pe + 1./432.*pow(
         m11,2)*RS145pe*RS356pe - 1./54.*pow(m11,2)*RS145pe*RS1456pee
          - 1./54.*pow(m11,2)*RS145pe*RS1456ppe + 1./432.*pow(m11,2)*
         RS145pe*RSe45611p + 1./432.*pow(m11,2)*RS145pe*RSp45611e + 1./
         108.*pow(m11,2)*pow(RS145pe,2) + 1./48.*pow(m11,2)*RS146pe - 1.
         /216.*pow(m11,2)*RS146pe*RS34561pe - 1./27.*pow(m11,2)*RS146pe
         *RS14563pe - 1./216.*pow(m11,2)*RS146pe*RSe45613p - 1./216.*
         pow(m11,2)*RS146pe*RSp45613e - 1./432.*pow(m11,2)*RS146pe*
         RS3456pee - 1./432.*pow(m11,2)*RS146pe*RS3456ppe + 1./432.*
         pow(m11,2)*RS146pe*RSe45633p + 1./432.*pow(m11,2)*RS146pe*
         RSp45633e );

      M13p6x23R +=  + Ab(m11,mu2)*pi16 * ( 1./432.*pow(m11,2)*RS146pe*
         RS345pe + 1./432.*pow(m11,2)*RS146pe*RS346pe + 1./432.*pow(
         m11,2)*RS146pe*RS356pe - 1./54.*pow(m11,2)*RS146pe*RS1456pee
          - 1./54.*pow(m11,2)*RS146pe*RS1456ppe + 1./432.*pow(m11,2)*
         RS146pe*RSe45611p + 1./432.*pow(m11,2)*RS146pe*RSp45611e + 1./
         54.*pow(m11,2)*RS146pe*RS145pe + 1./108.*pow(m11,2)*pow(
         RS146pe,2) + 1./48.*pow(m11,2)*RS156pe - 1./216.*pow(m11,2)*
         RS156pe*RS34561pe - 1./27.*pow(m11,2)*RS156pe*RS14563pe - 1./
         216.*pow(m11,2)*RS156pe*RSe45613p - 1./216.*pow(m11,2)*RS156pe
         *RSp45613e - 1./432.*pow(m11,2)*RS156pe*RS3456pee - 1./432.*
         pow(m11,2)*RS156pe*RS3456ppe + 1./432.*pow(m11,2)*RS156pe*
         RSe45633p + 1./432.*pow(m11,2)*RS156pe*RSp45633e + 1./432.*
         pow(m11,2)*RS156pe*RS345pe + 1./432.*pow(m11,2)*RS156pe*
         RS346pe + 1./432.*pow(m11,2)*RS156pe*RS356pe - 1./54.*pow(
         m11,2)*RS156pe*RS1456pee - 1./54.*pow(m11,2)*RS156pe*RS1456ppe
          + 1./432.*pow(m11,2)*RS156pe*RSe45611p );

      M13p6x23R +=  + Ab(m11,mu2)*pi16 * ( 1./432.*pow(m11,2)*RS156pe*
         RSp45611e + 1./54.*pow(m11,2)*RS156pe*RS145pe + 1./54.*pow(
         m11,2)*RS156pe*RS146pe + 1./108.*pow(m11,2)*pow(RS156pe,2) );

      M13p6x23R +=  + pow(Ab(m11,mu2),2) * (  - 1./270.*RD1456ep*
         RS14563pe - 1./540.*RS1456pee*RD1456ep - 1./540.*RS1456ppe*
         RD1456ep + 1./540.*RS145pe*RD1456ep + 1./540.*RS146pe*RD1456ep
          + 1./540.*RS156pe*RD1456ep + 1./16.*m33 + 13./540.*m33*pow(
         RS14563pe,2) - 1./540.*m33*RS1456pee*RS14563pe + 1./144.*m33*
         pow(RS1456pee,2) - 1./540.*m33*RS1456ppe*RS14563pe + 1./72.*
         m33*RS1456ppe*RS1456pee + 1./144.*m33*pow(RS1456ppe,2) + 1./
         540.*m33*RS145pe*RS14563pe - 1./72.*m33*RS145pe*RS1456pee - 1./
         72.*m33*RS145pe*RS1456ppe + 1./144.*m33*pow(RS145pe,2) + 1./
         540.*m33*RS146pe*RS14563pe - 1./72.*m33*RS146pe*RS1456pee - 1./
         72.*m33*RS146pe*RS1456ppe + 1./72.*m33*RS146pe*RS145pe + 1./
         144.*m33*pow(RS146pe,2) + 1./540.*m33*RS156pe*RS14563pe - 1./
         72.*m33*RS156pe*RS1456pee - 1./72.*m33*RS156pe*RS1456ppe + 1./
         72.*m33*RS156pe*RS145pe + 1./72.*m33*RS156pe*RS146pe + 1./144.
         *m33*pow(RS156pe,2) + 1./16.*m11 + 17./540.*m11*pow(
         RS14563pe,2) );

      M13p6x23R +=  + pow(Ab(m11,mu2),2) * (  - 1./60.*m11*RS1456pee*
         RS14563pe - 1./27.*m11*RS1456pee*RSe45613p - 1./27.*m11*
         RS1456pee*RSp45613e - 5./432.*m11*pow(RS1456pee,2) - 1./60.*
         m11*RS1456ppe*RS14563pe - 1./27.*m11*RS1456ppe*RSe45613p - 1./
         27.*m11*RS1456ppe*RSp45613e - 5./216.*m11*RS1456ppe*RS1456pee
          - 5./432.*m11*pow(RS1456ppe,2) + 1./27.*m11*RSe45611p*
         RS1456pee + 1./27.*m11*RSe45611p*RS1456ppe + 1./27.*m11*
         RSp45611e*RS1456pee + 1./27.*m11*RSp45611e*RS1456ppe + 1./60.*
         m11*RS145pe*RS14563pe + 1./27.*m11*RS145pe*RSe45613p + 1./27.*
         m11*RS145pe*RSp45613e + 5./216.*m11*RS145pe*RS1456pee + 5./216.
         *m11*RS145pe*RS1456ppe - 1./27.*m11*RS145pe*RSe45611p - 1./27.
         *m11*RS145pe*RSp45611e - 5./432.*m11*pow(RS145pe,2) + 1./60.*
         m11*RS146pe*RS14563pe + 1./27.*m11*RS146pe*RSe45613p + 1./27.*
         m11*RS146pe*RSp45613e + 5./216.*m11*RS146pe*RS1456pee + 5./216.
         *m11*RS146pe*RS1456ppe - 1./27.*m11*RS146pe*RSe45611p - 1./27.
         *m11*RS146pe*RSp45611e );

      M13p6x23R +=  + pow(Ab(m11,mu2),2) * (  - 5./216.*m11*RS146pe*
         RS145pe - 5./432.*m11*pow(RS146pe,2) + 1./60.*m11*RS156pe*
         RS14563pe + 1./27.*m11*RS156pe*RSe45613p + 1./27.*m11*RS156pe*
         RSp45613e + 5./216.*m11*RS156pe*RS1456pee + 5./216.*m11*
         RS156pe*RS1456ppe - 1./27.*m11*RS156pe*RSe45611p - 1./27.*m11*
         RS156pe*RSp45611e - 5./216.*m11*RS156pe*RS145pe - 5./216.*m11*
         RS156pe*RS146pe - 5./432.*m11*pow(RS156pe,2) - 1./27.*m11*mp2*
         RN1p*RS1456pee*RSp45613e - 1./27.*m11*mp2*RN1p*RS1456ppe*
         RSp45613e + 1./27.*m11*mp2*RN1p*RSp45611e*RS1456pee + 1./27.*
         m11*mp2*RN1p*RSp45611e*RS1456ppe + 1./27.*m11*mp2*RN1p*RS145pe
         *RSp45613e - 1./27.*m11*mp2*RN1p*RS145pe*RSp45611e + 1./27.*
         m11*mp2*RN1p*RS146pe*RSp45613e - 1./27.*m11*mp2*RN1p*RS146pe*
         RSp45611e + 1./27.*m11*mp2*RN1p*RS156pe*RSp45613e - 1./27.*m11
         *mp2*RN1p*RS156pe*RSp45611e - 1./27.*m11*me2*RN1e*RS1456pee*
         RSe45613p - 1./27.*m11*me2*RN1e*RS1456ppe*RSe45613p + 1./27.*
         m11*me2*RN1e*RSe45611p*RS1456pee );

      M13p6x23R +=  + pow(Ab(m11,mu2),2) * ( 1./27.*m11*me2*RN1e*
         RSe45611p*RS1456ppe + 1./27.*m11*me2*RN1e*RS145pe*RSe45613p - 
         1./27.*m11*me2*RN1e*RS145pe*RSe45611p + 1./27.*m11*me2*RN1e*
         RS146pe*RSe45613p - 1./27.*m11*me2*RN1e*RS146pe*RSe45611p + 1./
         27.*m11*me2*RN1e*RS156pe*RSe45613p - 1./27.*m11*me2*RN1e*
         RS156pe*RSe45611p - 1./27.*m11*m33*RN1p*RS1456pee*RSp45613e - 
         1./27.*m11*m33*RN1p*RS1456ppe*RSp45613e + 1./27.*m11*m33*RN1p*
         RS145pe*RSp45613e + 1./27.*m11*m33*RN1p*RS146pe*RSp45613e + 1./
         27.*m11*m33*RN1p*RS156pe*RSp45613e - 1./27.*m11*m33*RN1e*
         RS1456pee*RSe45613p - 1./27.*m11*m33*RN1e*RS1456ppe*RSe45613p
          + 1./27.*m11*m33*RN1e*RS145pe*RSe45613p + 1./27.*m11*m33*RN1e
         *RS146pe*RSe45613p + 1./27.*m11*m33*RN1e*RS156pe*RSe45613p - 1.
         /27.*m11*m33*RN13*RS1456pee*RS34561pe - 1./27.*m11*m33*RN13*
         RS1456ppe*RS34561pe + 1./27.*m11*m33*RN13*RS145pe*RS34561pe + 
         1./27.*m11*m33*RN13*RS146pe*RS34561pe + 1./27.*m11*m33*RN13*
         RS156pe*RS34561pe );

      M13p6x23R +=  + pow(Ab(m11,mu2),2) * (  - 1./27.*pow(m11,2)*RN1p*
         RS1456pee*RSp45613e - 1./27.*pow(m11,2)*RN1p*RS1456ppe*
         RSp45613e - 1./27.*pow(m11,2)*RN1p*RSp45611e*RS1456pee - 1./27.
         *pow(m11,2)*RN1p*RSp45611e*RS1456ppe + 1./27.*pow(m11,2)*RN1p*
         RS145pe*RSp45613e + 1./27.*pow(m11,2)*RN1p*RS145pe*RSp45611e
          + 1./27.*pow(m11,2)*RN1p*RS146pe*RSp45613e + 1./27.*pow(
         m11,2)*RN1p*RS146pe*RSp45611e + 1./27.*pow(m11,2)*RN1p*RS156pe
         *RSp45613e + 1./27.*pow(m11,2)*RN1p*RS156pe*RSp45611e - 1./27.
         *pow(m11,2)*RN1e*RS1456pee*RSe45613p - 1./27.*pow(m11,2)*RN1e*
         RS1456ppe*RSe45613p - 1./27.*pow(m11,2)*RN1e*RSe45611p*
         RS1456pee - 1./27.*pow(m11,2)*RN1e*RSe45611p*RS1456ppe + 1./27.
         *pow(m11,2)*RN1e*RS145pe*RSe45613p + 1./27.*pow(m11,2)*RN1e*
         RS145pe*RSe45611p + 1./27.*pow(m11,2)*RN1e*RS146pe*RSe45613p
          + 1./27.*pow(m11,2)*RN1e*RS146pe*RSe45611p + 1./27.*pow(
         m11,2)*RN1e*RS156pe*RSe45613p + 1./27.*pow(m11,2)*RN1e*RS156pe
         *RSe45611p );

      M13p6x23R +=  + pow(Ab(m11,mu2),2) * (  - 2./27.*pow(m11,2)*RN13*
         RS1456pee*RS34561pe - 2./27.*pow(m11,2)*RN13*RS1456ppe*
         RS34561pe + 2./27.*pow(m11,2)*RN13*RS145pe*RS34561pe + 2./27.*
         pow(m11,2)*RN13*RS146pe*RS34561pe + 2./27.*pow(m11,2)*RN13*
         RS156pe*RS34561pe );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(m13,mu2) * (  - 1./162.*RD3456ep*
         RS14563pe - 1./54.*RD1456ep*RS34561pe - 2./81.*RD1456ep*
         RS14563pe - 1./54.*RD1456ep*RSe45613p - 1./54.*RD1456ep*
         RSp45613e - 1./108.*RD1456ep*RS3456pee - 1./108.*RD1456ep*
         RS3456ppe + 1./108.*RD1456ep*RSe45633p + 1./108.*RD1456ep*
         RSp45633e + 1./108.*RD1456ep*RS345pe + 1./108.*RD1456ep*
         RS346pe + 1./108.*RD1456ep*RS356pe - 1./324.*RS1456pee*
         RD3456ep - 1./81.*RS1456pee*RD1456ep - 1./324.*RS1456ppe*
         RD3456ep - 1./81.*RS1456ppe*RD1456ep + 1./108.*RSe45611p*
         RD1456ep + 1./108.*RSp45611e*RD1456ep + 1./324.*RS145pe*
         RD3456ep + 1./81.*RS145pe*RD1456ep + 1./324.*RS146pe*RD3456ep
          + 1./81.*RS146pe*RD1456ep + 1./324.*RS156pe*RD3456ep + 1./81.
         *RS156pe*RD1456ep + 1./81.*mp2*RSp45613e*RS14563pe - 1./162.*
         mp2*RSp45633e*RS14563pe + 1./162.*mp2*RS1456pee*RSp45613e - 1./
         324.*mp2*RS1456pee*RSp45633e + 1./162.*mp2*RS1456ppe*RSp45613e
          - 1./324.*mp2*RS1456ppe*RSp45633e );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(m13,mu2) * (  - 1./162.*mp2*
         RSp45611e*RS14563pe - 1./324.*mp2*RSp45611e*RS1456pee - 1./324.
         *mp2*RSp45611e*RS1456ppe - 1./162.*mp2*RS145pe*RSp45613e + 1./
         324.*mp2*RS145pe*RSp45633e + 1./324.*mp2*RS145pe*RSp45611e - 1.
         /162.*mp2*RS146pe*RSp45613e + 1./324.*mp2*RS146pe*RSp45633e + 
         1./324.*mp2*RS146pe*RSp45611e - 1./162.*mp2*RS156pe*RSp45613e
          + 1./324.*mp2*RS156pe*RSp45633e + 1./324.*mp2*RS156pe*
         RSp45611e + 1./81.*me2*RSe45613p*RS14563pe - 1./162.*me2*
         RSe45633p*RS14563pe + 1./162.*me2*RS1456pee*RSe45613p - 1./324.
         *me2*RS1456pee*RSe45633p + 1./162.*me2*RS1456ppe*RSe45613p - 1.
         /324.*me2*RS1456ppe*RSe45633p - 1./162.*me2*RSe45611p*
         RS14563pe - 1./324.*me2*RSe45611p*RS1456pee - 1./324.*me2*
         RSe45611p*RS1456ppe - 1./162.*me2*RS145pe*RSe45613p + 1./324.*
         me2*RS145pe*RSe45633p + 1./324.*me2*RS145pe*RSe45611p - 1./162.
         *me2*RS146pe*RSe45613p + 1./324.*me2*RS146pe*RSe45633p + 1./
         324.*me2*RS146pe*RSe45611p );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(m13,mu2) * (  - 1./162.*me2*
         RS156pe*RSe45613p + 1./324.*me2*RS156pe*RSe45633p + 1./324.*
         me2*RS156pe*RSe45611p - 1./72.*m33 - 1./27.*m33*RS14563pe*
         RS34561pe - 4./81.*m33*pow(RS14563pe,2) - 4./81.*m33*RSe45613p
         *RS14563pe - 4./81.*m33*RSp45613e*RS14563pe - 2./81.*m33*
         RS3456pee*RS14563pe - 2./81.*m33*RS3456ppe*RS14563pe + 5./162.
         *m33*RSe45633p*RS14563pe + 5./162.*m33*RSp45633e*RS14563pe + 2.
         /81.*m33*RS345pe*RS14563pe + 2./81.*m33*RS346pe*RS14563pe + 2./
         81.*m33*RS356pe*RS14563pe - 1./81.*m33*RS1456pee*RS34561pe - 1.
         /27.*m33*RS1456pee*RS14563pe - 1./54.*m33*RS1456pee*RSe45613p
          - 1./54.*m33*RS1456pee*RSp45613e + 1./324.*m33*RS1456pee*
         RSe45633p + 1./324.*m33*RS1456pee*RSp45633e - 5./324.*m33*pow(
         RS1456pee,2) - 1./81.*m33*RS1456ppe*RS34561pe - 1./27.*m33*
         RS1456ppe*RS14563pe - 1./54.*m33*RS1456ppe*RSe45613p - 1./54.*
         m33*RS1456ppe*RSp45613e + 1./324.*m33*RS1456ppe*RSe45633p + 1./
         324.*m33*RS1456ppe*RSp45633e );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(m13,mu2) * (  - 5./162.*m33*
         RS1456ppe*RS1456pee - 5./324.*m33*pow(RS1456ppe,2) + 1./54.*
         m33*RSe45611p*RS14563pe + 5./324.*m33*RSe45611p*RS1456pee + 5./
         324.*m33*RSe45611p*RS1456ppe + 1./54.*m33*RSp45611e*RS14563pe
          + 5./324.*m33*RSp45611e*RS1456pee + 5./324.*m33*RSp45611e*
         RS1456ppe + 1./81.*m33*RS145pe*RS34561pe + 1./27.*m33*RS145pe*
         RS14563pe + 1./54.*m33*RS145pe*RSe45613p + 1./54.*m33*RS145pe*
         RSp45613e - 1./324.*m33*RS145pe*RSe45633p - 1./324.*m33*
         RS145pe*RSp45633e + 5./162.*m33*RS145pe*RS1456pee + 5./162.*
         m33*RS145pe*RS1456ppe - 5./324.*m33*RS145pe*RSe45611p - 5./324.
         *m33*RS145pe*RSp45611e - 5./324.*m33*pow(RS145pe,2) + 1./81.*
         m33*RS146pe*RS34561pe + 1./27.*m33*RS146pe*RS14563pe + 1./54.*
         m33*RS146pe*RSe45613p + 1./54.*m33*RS146pe*RSp45613e - 1./324.
         *m33*RS146pe*RSe45633p - 1./324.*m33*RS146pe*RSp45633e + 5./
         162.*m33*RS146pe*RS1456pee + 5./162.*m33*RS146pe*RS1456ppe - 5.
         /324.*m33*RS146pe*RSe45611p );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(m13,mu2) * (  - 5./324.*m33*
         RS146pe*RSp45611e - 5./162.*m33*RS146pe*RS145pe - 5./324.*m33*
         pow(RS146pe,2) + 1./81.*m33*RS156pe*RS34561pe + 1./27.*m33*
         RS156pe*RS14563pe + 1./54.*m33*RS156pe*RSe45613p + 1./54.*m33*
         RS156pe*RSp45613e - 1./324.*m33*RS156pe*RSe45633p - 1./324.*
         m33*RS156pe*RSp45633e + 5./162.*m33*RS156pe*RS1456pee + 5./162.
         *m33*RS156pe*RS1456ppe - 5./324.*m33*RS156pe*RSe45611p - 5./
         324.*m33*RS156pe*RSp45611e - 5./162.*m33*RS156pe*RS145pe - 5./
         162.*m33*RS156pe*RS146pe - 5./324.*m33*pow(RS156pe,2) + 1./9.*
         pow(m33,2)*RN13 + 1./24.*m11 - 1./81.*m11*RS14563pe*RS34561pe
          - 1./81.*m11*RSe45613p*RS14563pe - 1./81.*m11*RSp45613e*
         RS14563pe - 1./81.*m11*RS1456pee*RS34561pe - 1./81.*m11*
         RS1456pee*RS14563pe - 1./81.*m11*RS1456pee*RSe45613p - 1./81.*
         m11*RS1456pee*RSp45613e + 1./162.*m11*RS1456pee*RS3456pee + 1./
         162.*m11*RS1456pee*RS3456ppe - 1./162.*m11*RS1456pee*RSe45633p
          - 1./162.*m11*RS1456pee*RSp45633e );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(m13,mu2) * (  - 1./162.*m11*
         RS1456pee*RS345pe - 1./162.*m11*RS1456pee*RS346pe - 1./162.*
         m11*RS1456pee*RS356pe - 5./324.*m11*pow(RS1456pee,2) - 1./81.*
         m11*RS1456ppe*RS34561pe - 1./81.*m11*RS1456ppe*RS14563pe - 1./
         81.*m11*RS1456ppe*RSe45613p - 1./81.*m11*RS1456ppe*RSp45613e
          + 1./162.*m11*RS1456ppe*RS3456pee + 1./162.*m11*RS1456ppe*
         RS3456ppe - 1./162.*m11*RS1456ppe*RSe45633p - 1./162.*m11*
         RS1456ppe*RSp45633e - 1./162.*m11*RS1456ppe*RS345pe - 1./162.*
         m11*RS1456ppe*RS346pe - 1./162.*m11*RS1456ppe*RS356pe - 5./162.
         *m11*RS1456ppe*RS1456pee - 5./324.*m11*pow(RS1456ppe,2) + 1./
         81.*m11*RSe45611p*RS14563pe + 1./54.*m11*RSe45611p*RS1456pee
          + 1./54.*m11*RSe45611p*RS1456ppe + 1./81.*m11*RSp45611e*
         RS14563pe + 1./54.*m11*RSp45611e*RS1456pee + 1./54.*m11*
         RSp45611e*RS1456ppe + 1./81.*m11*RS145pe*RS34561pe + 1./81.*
         m11*RS145pe*RS14563pe + 1./81.*m11*RS145pe*RSe45613p + 1./81.*
         m11*RS145pe*RSp45613e );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(m13,mu2) * (  - 1./162.*m11*
         RS145pe*RS3456pee - 1./162.*m11*RS145pe*RS3456ppe + 1./162.*
         m11*RS145pe*RSe45633p + 1./162.*m11*RS145pe*RSp45633e + 1./162.
         *m11*RS145pe*RS345pe + 1./162.*m11*RS145pe*RS346pe + 1./162.*
         m11*RS145pe*RS356pe + 5./162.*m11*RS145pe*RS1456pee + 5./162.*
         m11*RS145pe*RS1456ppe - 1./54.*m11*RS145pe*RSe45611p - 1./54.*
         m11*RS145pe*RSp45611e - 5./324.*m11*pow(RS145pe,2) + 1./81.*
         m11*RS146pe*RS34561pe + 1./81.*m11*RS146pe*RS14563pe + 1./81.*
         m11*RS146pe*RSe45613p + 1./81.*m11*RS146pe*RSp45613e - 1./162.
         *m11*RS146pe*RS3456pee - 1./162.*m11*RS146pe*RS3456ppe + 1./
         162.*m11*RS146pe*RSe45633p + 1./162.*m11*RS146pe*RSp45633e + 1.
         /162.*m11*RS146pe*RS345pe + 1./162.*m11*RS146pe*RS346pe + 1./
         162.*m11*RS146pe*RS356pe + 5./162.*m11*RS146pe*RS1456pee + 5./
         162.*m11*RS146pe*RS1456ppe - 1./54.*m11*RS146pe*RSe45611p - 1./
         54.*m11*RS146pe*RSp45611e - 5./162.*m11*RS146pe*RS145pe - 5./
         324.*m11*pow(RS146pe,2) );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(m13,mu2) * ( 1./81.*m11*RS156pe*
         RS34561pe + 1./81.*m11*RS156pe*RS14563pe + 1./81.*m11*RS156pe*
         RSe45613p + 1./81.*m11*RS156pe*RSp45613e - 1./162.*m11*RS156pe
         *RS3456pee - 1./162.*m11*RS156pe*RS3456ppe + 1./162.*m11*
         RS156pe*RSe45633p + 1./162.*m11*RS156pe*RSp45633e + 1./162.*
         m11*RS156pe*RS345pe + 1./162.*m11*RS156pe*RS346pe + 1./162.*
         m11*RS156pe*RS356pe + 5./162.*m11*RS156pe*RS1456pee + 5./162.*
         m11*RS156pe*RS1456ppe - 1./54.*m11*RS156pe*RSe45611p - 1./54.*
         m11*RS156pe*RSp45611e - 5./162.*m11*RS156pe*RS145pe - 5./162.*
         m11*RS156pe*RS146pe - 5./324.*m11*pow(RS156pe,2) + 5./9.*m11*
         m33*RN13 + 1./3.*pow(m11,2)*RN13 );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(m14,mu2) * (  - 29./540.*RD1456ep
          - 1./27.*m44*RSe563p - 1./27.*m44*RSp563e - 1./270.*m44*
         RS156pe + 1./27.*m44*RSe561p + 1./27.*m44*RSp561e - 1./27.*m44
         *mp2*RN1p*RSp563e + 1./27.*m44*mp2*RN1p*RSp561e - 1./27.*m44*
         me2*RN1e*RSe563p + 1./27.*m44*me2*RN1e*RSe561p + 41./1080.*m33
         *RS14563pe + 1./27.*m33*RSe45613p + 1./27.*m33*RSp45613e - 1./
         27.*m33*RSe563p - 1./27.*m33*RSp563e - 1./24.*m33*RS156pe + 1./
         27.*m33*mp2*RN1p*RSp45613e - 1./27.*m33*mp2*RN1p*RSp563e + 1./
         27.*m33*me2*RN1e*RSe45613p - 1./27.*m33*me2*RN1e*RSe563p - 1./
         27.*m33*m44*RN1p*RSp563e - 1./27.*m33*m44*RN1e*RSe563p - 1./27.
         *m33*m44*RN13*RS356pe + 1./27.*m11*RS34561pe + 109./1080.*m11*
         RS14563pe + 1./27.*m11*RSe45613p + 1./27.*m11*RSp45613e - 1./
         27.*m11*RS356pe - 1./9.*m11*RSe563p - 1./9.*m11*RSp563e + 1./
         27.*m11*RSe45611p + 1./27.*m11*RSp45611e + 19./1080.*m11*
         RS156pe + 1./27.*m11*RSe561p + 1./27.*m11*RSp561e + 1./27.*m11
         *mp2*RN1p*RSp45613e );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(m14,mu2) * (  - 1./9.*m11*mp2*RN1p
         *RSp563e + 1./27.*m11*mp2*RN1p*RSp45611e + 1./27.*m11*mp2*RN1p
         *RSp561e + 1./27.*m11*me2*RN1e*RSe45613p - 1./9.*m11*me2*RN1e*
         RSe563p + 1./27.*m11*me2*RN1e*RSe45611p + 1./27.*m11*me2*RN1e*
         RSe561p - 1./27.*m11*m44*RN1p*RSp563e - 1./27.*m11*m44*RN1p*
         RSp561e - 1./27.*m11*m44*RN1e*RSe563p - 1./27.*m11*m44*RN1e*
         RSe561p - 2./27.*m11*m44*RN13*RS356pe - 1./27.*m11*m33*RN1p*
         RSp45613e - 1./27.*m11*m33*RN1p*RSp563e - 1./27.*m11*m33*RN1e*
         RSe45613p - 1./27.*m11*m33*RN1e*RSe563p + 1./27.*m11*m33*RN13*
         RS34561pe - 1./9.*m11*m33*RN13*RS356pe - 1./27.*pow(m11,2)*
         RN1p*RSp45613e - 1./27.*pow(m11,2)*RN1p*RSp563e - 1./27.*pow(
         m11,2)*RN1p*RSp45611e - 1./27.*pow(m11,2)*RN1p*RSp561e - 1./27.
         *pow(m11,2)*RN1e*RSe45613p - 1./27.*pow(m11,2)*RN1e*RSe563p - 
         1./27.*pow(m11,2)*RN1e*RSe45611p - 1./27.*pow(m11,2)*RN1e*
         RSe561p - 1./27.*pow(m11,2)*RN13*RS34561pe - 1./9.*pow(m11,2)*
         RN13*RS356pe );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(m15,mu2) * (  - 29./540.*RD1456ep
          - 1./27.*m55*RSe463p - 1./27.*m55*RSp463e - 1./270.*m55*
         RS146pe + 1./27.*m55*RSe461p + 1./27.*m55*RSp461e - 1./27.*m55
         *mp2*RN1p*RSp463e + 1./27.*m55*mp2*RN1p*RSp461e - 1./27.*m55*
         me2*RN1e*RSe463p + 1./27.*m55*me2*RN1e*RSe461p + 41./1080.*m33
         *RS14563pe + 1./27.*m33*RSe45613p + 1./27.*m33*RSp45613e - 1./
         27.*m33*RSe463p - 1./27.*m33*RSp463e - 1./24.*m33*RS146pe + 1./
         27.*m33*mp2*RN1p*RSp45613e - 1./27.*m33*mp2*RN1p*RSp463e + 1./
         27.*m33*me2*RN1e*RSe45613p - 1./27.*m33*me2*RN1e*RSe463p - 1./
         27.*m33*m55*RN1p*RSp463e - 1./27.*m33*m55*RN1e*RSe463p - 1./27.
         *m33*m55*RN13*RS346pe + 1./27.*m11*RS34561pe + 109./1080.*m11*
         RS14563pe + 1./27.*m11*RSe45613p + 1./27.*m11*RSp45613e - 1./
         27.*m11*RS346pe - 1./9.*m11*RSe463p - 1./9.*m11*RSp463e + 1./
         27.*m11*RSe45611p + 1./27.*m11*RSp45611e + 19./1080.*m11*
         RS146pe + 1./27.*m11*RSe461p + 1./27.*m11*RSp461e + 1./27.*m11
         *mp2*RN1p*RSp45613e );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(m15,mu2) * (  - 1./9.*m11*mp2*RN1p
         *RSp463e + 1./27.*m11*mp2*RN1p*RSp45611e + 1./27.*m11*mp2*RN1p
         *RSp461e + 1./27.*m11*me2*RN1e*RSe45613p - 1./9.*m11*me2*RN1e*
         RSe463p + 1./27.*m11*me2*RN1e*RSe45611p + 1./27.*m11*me2*RN1e*
         RSe461p - 1./27.*m11*m55*RN1p*RSp463e - 1./27.*m11*m55*RN1p*
         RSp461e - 1./27.*m11*m55*RN1e*RSe463p - 1./27.*m11*m55*RN1e*
         RSe461p - 2./27.*m11*m55*RN13*RS346pe - 1./27.*m11*m33*RN1p*
         RSp45613e - 1./27.*m11*m33*RN1p*RSp463e - 1./27.*m11*m33*RN1e*
         RSe45613p - 1./27.*m11*m33*RN1e*RSe463p + 1./27.*m11*m33*RN13*
         RS34561pe - 1./9.*m11*m33*RN13*RS346pe - 1./27.*pow(m11,2)*
         RN1p*RSp45613e - 1./27.*pow(m11,2)*RN1p*RSp463e - 1./27.*pow(
         m11,2)*RN1p*RSp45611e - 1./27.*pow(m11,2)*RN1p*RSp461e - 1./27.
         *pow(m11,2)*RN1e*RSe45613p - 1./27.*pow(m11,2)*RN1e*RSe463p - 
         1./27.*pow(m11,2)*RN1e*RSe45611p - 1./27.*pow(m11,2)*RN1e*
         RSe461p - 1./27.*pow(m11,2)*RN13*RS34561pe - 1./9.*pow(m11,2)*
         RN13*RS346pe );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(m16,mu2) * (  - 29./540.*RD1456ep
          - 1./27.*m66*RSe453p - 1./27.*m66*RSp453e - 1./270.*m66*
         RS145pe + 1./27.*m66*RSe451p + 1./27.*m66*RSp451e - 1./27.*m66
         *mp2*RN1p*RSp453e + 1./27.*m66*mp2*RN1p*RSp451e - 1./27.*m66*
         me2*RN1e*RSe453p + 1./27.*m66*me2*RN1e*RSe451p + 41./1080.*m33
         *RS14563pe + 1./27.*m33*RSe45613p + 1./27.*m33*RSp45613e - 1./
         27.*m33*RSe453p - 1./27.*m33*RSp453e - 1./24.*m33*RS145pe + 1./
         27.*m33*mp2*RN1p*RSp45613e - 1./27.*m33*mp2*RN1p*RSp453e + 1./
         27.*m33*me2*RN1e*RSe45613p - 1./27.*m33*me2*RN1e*RSe453p - 1./
         27.*m33*m66*RN1p*RSp453e - 1./27.*m33*m66*RN1e*RSe453p - 1./27.
         *m33*m66*RN13*RS345pe + 1./27.*m11*RS34561pe + 109./1080.*m11*
         RS14563pe + 1./27.*m11*RSe45613p + 1./27.*m11*RSp45613e - 1./
         27.*m11*RS345pe - 1./9.*m11*RSe453p - 1./9.*m11*RSp453e + 1./
         27.*m11*RSe45611p + 1./27.*m11*RSp45611e + 19./1080.*m11*
         RS145pe + 1./27.*m11*RSe451p + 1./27.*m11*RSp451e + 1./27.*m11
         *mp2*RN1p*RSp45613e );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(m16,mu2) * (  - 1./9.*m11*mp2*RN1p
         *RSp453e + 1./27.*m11*mp2*RN1p*RSp45611e + 1./27.*m11*mp2*RN1p
         *RSp451e + 1./27.*m11*me2*RN1e*RSe45613p - 1./9.*m11*me2*RN1e*
         RSe453p + 1./27.*m11*me2*RN1e*RSe45611p + 1./27.*m11*me2*RN1e*
         RSe451p - 1./27.*m11*m66*RN1p*RSp453e - 1./27.*m11*m66*RN1p*
         RSp451e - 1./27.*m11*m66*RN1e*RSe453p - 1./27.*m11*m66*RN1e*
         RSe451p - 2./27.*m11*m66*RN13*RS345pe - 1./27.*m11*m33*RN1p*
         RSp45613e - 1./27.*m11*m33*RN1p*RSp453e - 1./27.*m11*m33*RN1e*
         RSe45613p - 1./27.*m11*m33*RN1e*RSe453p + 1./27.*m11*m33*RN13*
         RS34561pe - 1./9.*m11*m33*RN13*RS345pe - 1./27.*pow(m11,2)*
         RN1p*RSp45613e - 1./27.*pow(m11,2)*RN1p*RSp453e - 1./27.*pow(
         m11,2)*RN1p*RSp45611e - 1./27.*pow(m11,2)*RN1p*RSp451e - 1./27.
         *pow(m11,2)*RN1e*RSe45613p - 1./27.*pow(m11,2)*RN1e*RSe453p - 
         1./27.*pow(m11,2)*RN1e*RSe45611p - 1./27.*pow(m11,2)*RN1e*
         RSe451p - 1./27.*pow(m11,2)*RN13*RS34561pe - 1./9.*pow(m11,2)*
         RN13*RS345pe );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(m33,mu2) * (  - 1./270.*RD3456ep*
         RS14563pe - 1./270.*RD1456ep*RS34561pe - 1./540.*RD1456ep*
         RS3456pee - 1./540.*RD1456ep*RS3456ppe + 1./540.*RD1456ep*
         RS345pe + 1./540.*RD1456ep*RS346pe + 1./540.*RD1456ep*RS356pe
          - 1./540.*RS1456pee*RD3456ep - 1./540.*RS1456ppe*RD3456ep + 1.
         /540.*RS145pe*RD3456ep + 1./540.*RS146pe*RD3456ep + 1./540.*
         RS156pe*RD3456ep + 1./18.*m33*RS14563pe*RS34561pe + 1./60.*m33
         *RS3456pee*RS14563pe + 1./60.*m33*RS3456ppe*RS14563pe - 1./60.
         *m33*RS345pe*RS14563pe - 1./60.*m33*RS346pe*RS14563pe - 1./60.
         *m33*RS356pe*RS14563pe + 1./540.*m33*RS1456pee*RS34561pe - 1./
         72.*m33*RS1456pee*RS3456pee - 1./72.*m33*RS1456pee*RS3456ppe
          + 1./72.*m33*RS1456pee*RS345pe + 1./72.*m33*RS1456pee*RS346pe
          + 1./72.*m33*RS1456pee*RS356pe + 1./540.*m33*RS1456ppe*
         RS34561pe - 1./72.*m33*RS1456ppe*RS3456pee - 1./72.*m33*
         RS1456ppe*RS3456ppe + 1./72.*m33*RS1456ppe*RS345pe + 1./72.*
         m33*RS1456ppe*RS346pe );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(m33,mu2) * ( 1./72.*m33*RS1456ppe*
         RS356pe - 1./540.*m33*RS145pe*RS34561pe + 1./72.*m33*RS145pe*
         RS3456pee + 1./72.*m33*RS145pe*RS3456ppe - 1./72.*m33*RS145pe*
         RS345pe - 1./72.*m33*RS145pe*RS346pe - 1./72.*m33*RS145pe*
         RS356pe - 1./540.*m33*RS146pe*RS34561pe + 1./72.*m33*RS146pe*
         RS3456pee + 1./72.*m33*RS146pe*RS3456ppe - 1./72.*m33*RS146pe*
         RS345pe - 1./72.*m33*RS146pe*RS346pe - 1./72.*m33*RS146pe*
         RS356pe - 1./540.*m33*RS156pe*RS34561pe + 1./72.*m33*RS156pe*
         RS3456pee + 1./72.*m33*RS156pe*RS3456ppe - 1./72.*m33*RS156pe*
         RS345pe - 1./72.*m33*RS156pe*RS346pe - 1./72.*m33*RS156pe*
         RS356pe - 1./27.*pow(m33,2)*RN13*RS3456pee*RS14563pe - 1./27.*
         pow(m33,2)*RN13*RS3456ppe*RS14563pe + 1./27.*pow(m33,2)*RN13*
         RS345pe*RS14563pe + 1./27.*pow(m33,2)*RN13*RS346pe*RS14563pe
          + 1./27.*pow(m33,2)*RN13*RS356pe*RS14563pe + 1./18.*m11*
         RS14563pe*RS34561pe + 1./540.*m11*RS3456pee*RS14563pe + 1./540.
         *m11*RS3456ppe*RS14563pe );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(m33,mu2) * (  - 1./540.*m11*
         RS345pe*RS14563pe - 1./540.*m11*RS346pe*RS14563pe - 1./540.*
         m11*RS356pe*RS14563pe - 11./540.*m11*RS1456pee*RS34561pe - 1./
         72.*m11*RS1456pee*RS3456pee - 1./72.*m11*RS1456pee*RS3456ppe
          + 1./72.*m11*RS1456pee*RS345pe + 1./72.*m11*RS1456pee*RS346pe
          + 1./72.*m11*RS1456pee*RS356pe - 11./540.*m11*RS1456ppe*
         RS34561pe - 1./72.*m11*RS1456ppe*RS3456pee - 1./72.*m11*
         RS1456ppe*RS3456ppe + 1./72.*m11*RS1456ppe*RS345pe + 1./72.*
         m11*RS1456ppe*RS346pe + 1./72.*m11*RS1456ppe*RS356pe + 11./540.
         *m11*RS145pe*RS34561pe + 1./72.*m11*RS145pe*RS3456pee + 1./72.
         *m11*RS145pe*RS3456ppe - 1./72.*m11*RS145pe*RS345pe - 1./72.*
         m11*RS145pe*RS346pe - 1./72.*m11*RS145pe*RS356pe + 11./540.*
         m11*RS146pe*RS34561pe + 1./72.*m11*RS146pe*RS3456pee + 1./72.*
         m11*RS146pe*RS3456ppe - 1./72.*m11*RS146pe*RS345pe - 1./72.*
         m11*RS146pe*RS346pe - 1./72.*m11*RS146pe*RS356pe + 11./540.*
         m11*RS156pe*RS34561pe );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(m33,mu2) * ( 1./72.*m11*RS156pe*
         RS3456pee + 1./72.*m11*RS156pe*RS3456ppe - 1./72.*m11*RS156pe*
         RS345pe - 1./72.*m11*RS156pe*RS346pe - 1./72.*m11*RS156pe*
         RS356pe - 2./27.*m11*m33*RN13*RS3456pee*RS14563pe - 2./27.*m11
         *m33*RN13*RS3456ppe*RS14563pe + 2./27.*m11*m33*RN13*RS345pe*
         RS14563pe + 2./27.*m11*m33*RN13*RS346pe*RS14563pe + 2./27.*m11
         *m33*RN13*RS356pe*RS14563pe + 1./27.*m11*m33*RN13*RS1456pee*
         RS34561pe + 1./27.*m11*m33*RN13*RS1456ppe*RS34561pe - 1./27.*
         m11*m33*RN13*RS145pe*RS34561pe - 1./27.*m11*m33*RN13*RS146pe*
         RS34561pe - 1./27.*m11*m33*RN13*RS156pe*RS34561pe + 2./27.*
         pow(m11,2)*RN13*RS1456pee*RS34561pe + 2./27.*pow(m11,2)*RN13*
         RS1456ppe*RS34561pe - 2./27.*pow(m11,2)*RN13*RS145pe*RS34561pe
          - 2./27.*pow(m11,2)*RN13*RS146pe*RS34561pe - 2./27.*pow(
         m11,2)*RN13*RS156pe*RS34561pe );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(m34,mu2) * (  - 1./180.*RD1456ep
          + 17./270.*m44*RS156pe - 19./1080.*m33*RS14563pe - 1./24.*m33
         *RS1456pee - 1./24.*m33*RS1456ppe + 1./24.*m33*RS145pe + 1./24.
         *m33*RS146pe + 5./108.*m33*RS156pe - 1./27.*m33*m44*RN13*
         RS156pe - 1./27.*pow(m33,2)*RN13*RS14563pe - 1./27.*pow(m33,2)
         *RN13*RS156pe + 89./1080.*m11*RS14563pe - 1./24.*m11*RS1456pee
          - 1./24.*m11*RS1456ppe + 1./24.*m11*RS145pe + 1./24.*m11*
         RS146pe - 19./540.*m11*RS156pe - 2./27.*m11*m44*RN13*RS156pe
          - 4./27.*m11*m33*RN13*RS156pe + 1./27.*pow(m11,2)*RN13*
         RS14563pe - 1./27.*pow(m11,2)*RN13*RS156pe );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(m35,mu2) * (  - 1./180.*RD1456ep
          + 17./270.*m55*RS146pe - 19./1080.*m33*RS14563pe - 1./24.*m33
         *RS1456pee - 1./24.*m33*RS1456ppe + 1./24.*m33*RS145pe + 5./
         108.*m33*RS146pe + 1./24.*m33*RS156pe - 1./27.*m33*m55*RN13*
         RS146pe - 1./27.*pow(m33,2)*RN13*RS14563pe - 1./27.*pow(m33,2)
         *RN13*RS146pe + 89./1080.*m11*RS14563pe - 1./24.*m11*RS1456pee
          - 1./24.*m11*RS1456ppe + 1./24.*m11*RS145pe - 19./540.*m11*
         RS146pe + 1./24.*m11*RS156pe - 2./27.*m11*m55*RN13*RS146pe - 4.
         /27.*m11*m33*RN13*RS146pe + 1./27.*pow(m11,2)*RN13*RS14563pe
          - 1./27.*pow(m11,2)*RN13*RS146pe );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(m36,mu2) * (  - 1./180.*RD1456ep
          + 17./270.*m66*RS145pe - 19./1080.*m33*RS14563pe - 1./24.*m33
         *RS1456pee - 1./24.*m33*RS1456ppe + 5./108.*m33*RS145pe + 1./
         24.*m33*RS146pe + 1./24.*m33*RS156pe - 1./27.*m33*m66*RN13*
         RS145pe - 1./27.*pow(m33,2)*RN13*RS14563pe - 1./27.*pow(m33,2)
         *RN13*RS145pe + 89./1080.*m11*RS14563pe - 1./24.*m11*RS1456pee
          - 1./24.*m11*RS1456ppe - 19./540.*m11*RS145pe + 1./24.*m11*
         RS146pe + 1./24.*m11*RS156pe - 2./27.*m11*m66*RN13*RS145pe - 4.
         /27.*m11*m33*RN13*RS145pe + 1./27.*pow(m11,2)*RN13*RS14563pe
          - 1./27.*pow(m11,2)*RN13*RS145pe );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(m44,mu2) * ( 1./27.*m44*RS156pe*
         RSe563p + 1./27.*m44*RS156pe*RSp563e - 1./54.*m44*pow(
         RS156pe,2) - 1./81.*m44*RS456pe*RS156pe*RSe563p - 1./81.*m44*
         RS456pe*RS156pe*RSp563e + 1./162.*m44*RS456pe*pow(RS156pe,2)
          - 1./27.*m44*RSe561p*RS156pe + 1./81.*m44*RSe561p*RS456pe*
         RS156pe - 1./27.*m44*RSp561e*RS156pe + 1./81.*m44*RSp561e*
         RS456pe*RS156pe + 1./27.*m44*mp2*RN1p*RS156pe*RSp563e - 1./81.
         *m44*mp2*RN1p*RS456pe*RS156pe*RSp563e - 1./27.*m44*mp2*RN1p*
         RSp561e*RS156pe + 1./81.*m44*mp2*RN1p*RSp561e*RS456pe*RS156pe
          + 1./27.*m44*me2*RN1e*RS156pe*RSe563p - 1./81.*m44*me2*RN1e*
         RS456pe*RS156pe*RSe563p - 1./27.*m44*me2*RN1e*RSe561p*RS156pe
          + 1./81.*m44*me2*RN1e*RSe561p*RS456pe*RS156pe + 1./27.*m33*
         m44*RN1p*RS156pe*RSp563e - 1./81.*m33*m44*RN1p*RS456pe*RS156pe
         *RSp563e + 1./27.*m33*m44*RN1e*RS156pe*RSe563p - 1./81.*m33*
         m44*RN1e*RS456pe*RS156pe*RSe563p + 1./27.*m33*m44*RN13*RS156pe
         *RS356pe );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(m44,mu2) * (  - 1./81.*m33*m44*
         RN13*RS456pe*RS156pe*RS356pe + 1./27.*m11*m44*RN1p*RS156pe*
         RSp563e - 1./81.*m11*m44*RN1p*RS456pe*RS156pe*RSp563e + 1./27.
         *m11*m44*RN1p*RSp561e*RS156pe - 1./81.*m11*m44*RN1p*RSp561e*
         RS456pe*RS156pe + 1./27.*m11*m44*RN1e*RS156pe*RSe563p - 1./81.
         *m11*m44*RN1e*RS456pe*RS156pe*RSe563p + 1./27.*m11*m44*RN1e*
         RSe561p*RS156pe - 1./81.*m11*m44*RN1e*RSe561p*RS456pe*RS156pe
          + 2./27.*m11*m44*RN13*RS156pe*RS356pe - 2./81.*m11*m44*RN13*
         RS456pe*RS156pe*RS356pe );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(m45,mu2) * ( 1./81.*m55*RS146pe*
         RSe563p + 1./81.*m55*RS146pe*RSe463p + 1./81.*m55*RS146pe*
         RSp563e + 1./81.*m55*RS146pe*RSp463e - 1./162.*m55*pow(
         RS146pe,2) + 1./81.*m55*RS156pe*RSe463p + 1./81.*m55*RS156pe*
         RSp463e - 1./81.*m55*RS156pe*RS146pe - 1./81.*m55*RSe561p*
         RS146pe - 1./81.*m55*RSe461p*RS146pe - 1./81.*m55*RSe461p*
         RS156pe - 1./81.*m55*RSp561e*RS146pe - 1./81.*m55*RSp461e*
         RS146pe - 1./81.*m55*RSp461e*RS156pe + 1./81.*m55*mp2*RN1p*
         RS146pe*RSp563e + 1./81.*m55*mp2*RN1p*RS146pe*RSp463e + 1./81.
         *m55*mp2*RN1p*RS156pe*RSp463e - 1./81.*m55*mp2*RN1p*RSp561e*
         RS146pe - 1./81.*m55*mp2*RN1p*RSp461e*RS146pe - 1./81.*m55*mp2
         *RN1p*RSp461e*RS156pe + 1./81.*m55*me2*RN1e*RS146pe*RSe563p + 
         1./81.*m55*me2*RN1e*RS146pe*RSe463p + 1./81.*m55*me2*RN1e*
         RS156pe*RSe463p - 1./81.*m55*me2*RN1e*RSe561p*RS146pe - 1./81.
         *m55*me2*RN1e*RSe461p*RS146pe - 1./81.*m55*me2*RN1e*RSe461p*
         RS156pe );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(m45,mu2) * ( 1./81.*m44*RS146pe*
         RSe563p + 1./81.*m44*RS146pe*RSp563e + 1./81.*m44*RS156pe*
         RSe563p + 1./81.*m44*RS156pe*RSe463p + 1./81.*m44*RS156pe*
         RSp563e + 1./81.*m44*RS156pe*RSp463e - 1./81.*m44*RS156pe*
         RS146pe - 1./162.*m44*pow(RS156pe,2) - 1./81.*m44*RSe561p*
         RS146pe - 1./81.*m44*RSe561p*RS156pe - 1./81.*m44*RSe461p*
         RS156pe - 1./81.*m44*RSp561e*RS146pe - 1./81.*m44*RSp561e*
         RS156pe - 1./81.*m44*RSp461e*RS156pe + 1./81.*m44*mp2*RN1p*
         RS146pe*RSp563e + 1./81.*m44*mp2*RN1p*RS156pe*RSp563e + 1./81.
         *m44*mp2*RN1p*RS156pe*RSp463e - 1./81.*m44*mp2*RN1p*RSp561e*
         RS146pe - 1./81.*m44*mp2*RN1p*RSp561e*RS156pe - 1./81.*m44*mp2
         *RN1p*RSp461e*RS156pe + 1./81.*m44*me2*RN1e*RS146pe*RSe563p + 
         1./81.*m44*me2*RN1e*RS156pe*RSe563p + 1./81.*m44*me2*RN1e*
         RS156pe*RSe463p - 1./81.*m44*me2*RN1e*RSe561p*RS146pe - 1./81.
         *m44*me2*RN1e*RSe561p*RS156pe - 1./81.*m44*me2*RN1e*RSe461p*
         RS156pe );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(m45,mu2) * ( 1./81.*m33*RS146pe*
         RSe563p - 1./81.*m33*RS146pe*RSe463p + 1./81.*m33*RS146pe*
         RSp563e - 1./81.*m33*RS146pe*RSp463e - 1./81.*m33*RS156pe*
         RSe563p + 1./81.*m33*RS156pe*RSe463p - 1./81.*m33*RS156pe*
         RSp563e + 1./81.*m33*RS156pe*RSp463e + 1./81.*m33*mp2*RN1p*
         RS146pe*RSp563e - 1./81.*m33*mp2*RN1p*RS146pe*RSp463e - 1./81.
         *m33*mp2*RN1p*RS156pe*RSp563e + 1./81.*m33*mp2*RN1p*RS156pe*
         RSp463e + 1./81.*m33*me2*RN1e*RS146pe*RSe563p - 1./81.*m33*me2
         *RN1e*RS146pe*RSe463p - 1./81.*m33*me2*RN1e*RS156pe*RSe563p + 
         1./81.*m33*me2*RN1e*RS156pe*RSe463p + 1./81.*m33*m55*RN1p*
         RS146pe*RSp563e + 1./81.*m33*m55*RN1p*RS146pe*RSp463e + 1./81.
         *m33*m55*RN1p*RS156pe*RSp463e + 1./81.*m33*m55*RN1e*RS146pe*
         RSe563p + 1./81.*m33*m55*RN1e*RS146pe*RSe463p + 1./81.*m33*m55
         *RN1e*RS156pe*RSe463p + 1./81.*m33*m55*RN13*RS146pe*RS346pe + 
         1./81.*m33*m55*RN13*RS146pe*RS356pe + 1./81.*m33*m55*RN13*
         RS156pe*RS346pe );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(m45,mu2) * ( 1./81.*m33*m44*RN1p*
         RS146pe*RSp563e + 1./81.*m33*m44*RN1p*RS156pe*RSp563e + 1./81.
         *m33*m44*RN1p*RS156pe*RSp463e + 1./81.*m33*m44*RN1e*RS146pe*
         RSe563p + 1./81.*m33*m44*RN1e*RS156pe*RSe563p + 1./81.*m33*m44
         *RN1e*RS156pe*RSe463p + 1./81.*m33*m44*RN13*RS146pe*RS356pe + 
         1./81.*m33*m44*RN13*RS156pe*RS346pe + 1./81.*m33*m44*RN13*
         RS156pe*RS356pe - 1./81.*m11*RS146pe*RS346pe + 1./81.*m11*
         RS146pe*RS356pe + 2./81.*m11*RS146pe*RSe563p - 2./81.*m11*
         RS146pe*RSe463p + 2./81.*m11*RS146pe*RSp563e - 2./81.*m11*
         RS146pe*RSp463e + 1./162.*m11*pow(RS146pe,2) + 1./81.*m11*
         RS156pe*RS346pe - 1./81.*m11*RS156pe*RS356pe - 2./81.*m11*
         RS156pe*RSe563p + 2./81.*m11*RS156pe*RSe463p - 2./81.*m11*
         RS156pe*RSp563e + 2./81.*m11*RS156pe*RSp463e - 1./81.*m11*
         RS156pe*RS146pe + 1./162.*m11*pow(RS156pe,2) + 2./81.*m11*mp2*
         RN1p*RS146pe*RSp563e - 2./81.*m11*mp2*RN1p*RS146pe*RSp463e - 2.
         /81.*m11*mp2*RN1p*RS156pe*RSp563e );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(m45,mu2) * ( 2./81.*m11*mp2*RN1p*
         RS156pe*RSp463e + 2./81.*m11*me2*RN1e*RS146pe*RSe563p - 2./81.
         *m11*me2*RN1e*RS146pe*RSe463p - 2./81.*m11*me2*RN1e*RS156pe*
         RSe563p + 2./81.*m11*me2*RN1e*RS156pe*RSe463p + 1./81.*m11*m55
         *RN1p*RS146pe*RSp563e + 1./81.*m11*m55*RN1p*RS146pe*RSp463e + 
         1./81.*m11*m55*RN1p*RS156pe*RSp463e + 1./81.*m11*m55*RN1p*
         RSp561e*RS146pe + 1./81.*m11*m55*RN1p*RSp461e*RS146pe + 1./81.
         *m11*m55*RN1p*RSp461e*RS156pe + 1./81.*m11*m55*RN1e*RS146pe*
         RSe563p + 1./81.*m11*m55*RN1e*RS146pe*RSe463p + 1./81.*m11*m55
         *RN1e*RS156pe*RSe463p + 1./81.*m11*m55*RN1e*RSe561p*RS146pe + 
         1./81.*m11*m55*RN1e*RSe461p*RS146pe + 1./81.*m11*m55*RN1e*
         RSe461p*RS156pe + 2./81.*m11*m55*RN13*RS146pe*RS346pe + 2./81.
         *m11*m55*RN13*RS146pe*RS356pe + 2./81.*m11*m55*RN13*RS156pe*
         RS346pe + 1./81.*m11*m44*RN1p*RS146pe*RSp563e + 1./81.*m11*m44
         *RN1p*RS156pe*RSp563e + 1./81.*m11*m44*RN1p*RS156pe*RSp463e + 
         1./81.*m11*m44*RN1p*RSp561e*RS146pe );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(m45,mu2) * ( 1./81.*m11*m44*RN1p*
         RSp561e*RS156pe + 1./81.*m11*m44*RN1p*RSp461e*RS156pe + 1./81.
         *m11*m44*RN1e*RS146pe*RSe563p + 1./81.*m11*m44*RN1e*RS156pe*
         RSe563p + 1./81.*m11*m44*RN1e*RS156pe*RSe463p + 1./81.*m11*m44
         *RN1e*RSe561p*RS146pe + 1./81.*m11*m44*RN1e*RSe561p*RS156pe + 
         1./81.*m11*m44*RN1e*RSe461p*RS156pe + 2./81.*m11*m44*RN13*
         RS146pe*RS356pe + 2./81.*m11*m44*RN13*RS156pe*RS346pe + 2./81.
         *m11*m44*RN13*RS156pe*RS356pe - 2./81.*m11*m33*RN13*RS146pe*
         RS346pe + 2./81.*m11*m33*RN13*RS146pe*RS356pe + 2./81.*m11*m33
         *RN13*RS156pe*RS346pe - 2./81.*m11*m33*RN13*RS156pe*RS356pe - 
         1./81.*pow(m11,2)*RN13*RS146pe*RS346pe + 1./81.*pow(m11,2)*
         RN13*RS146pe*RS356pe + 1./81.*pow(m11,2)*RN13*RS156pe*RS346pe
          - 1./81.*pow(m11,2)*RN13*RS156pe*RS356pe );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(m46,mu2) * ( 1./81.*m66*RS145pe*
         RSe563p + 1./81.*m66*RS145pe*RSe453p + 1./81.*m66*RS145pe*
         RSp563e + 1./81.*m66*RS145pe*RSp453e - 1./162.*m66*pow(
         RS145pe,2) + 1./81.*m66*RS156pe*RSe453p + 1./81.*m66*RS156pe*
         RSp453e - 1./81.*m66*RS156pe*RS145pe - 1./81.*m66*RSe561p*
         RS145pe - 1./81.*m66*RSe451p*RS145pe - 1./81.*m66*RSe451p*
         RS156pe - 1./81.*m66*RSp561e*RS145pe - 1./81.*m66*RSp451e*
         RS145pe - 1./81.*m66*RSp451e*RS156pe + 1./81.*m66*mp2*RN1p*
         RS145pe*RSp563e + 1./81.*m66*mp2*RN1p*RS145pe*RSp453e + 1./81.
         *m66*mp2*RN1p*RS156pe*RSp453e - 1./81.*m66*mp2*RN1p*RSp561e*
         RS145pe - 1./81.*m66*mp2*RN1p*RSp451e*RS145pe - 1./81.*m66*mp2
         *RN1p*RSp451e*RS156pe + 1./81.*m66*me2*RN1e*RS145pe*RSe563p + 
         1./81.*m66*me2*RN1e*RS145pe*RSe453p + 1./81.*m66*me2*RN1e*
         RS156pe*RSe453p - 1./81.*m66*me2*RN1e*RSe561p*RS145pe - 1./81.
         *m66*me2*RN1e*RSe451p*RS145pe - 1./81.*m66*me2*RN1e*RSe451p*
         RS156pe );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(m46,mu2) * ( 1./81.*m44*RS145pe*
         RSe563p + 1./81.*m44*RS145pe*RSp563e + 1./81.*m44*RS156pe*
         RSe563p + 1./81.*m44*RS156pe*RSe453p + 1./81.*m44*RS156pe*
         RSp563e + 1./81.*m44*RS156pe*RSp453e - 1./81.*m44*RS156pe*
         RS145pe - 1./162.*m44*pow(RS156pe,2) - 1./81.*m44*RSe561p*
         RS145pe - 1./81.*m44*RSe561p*RS156pe - 1./81.*m44*RSe451p*
         RS156pe - 1./81.*m44*RSp561e*RS145pe - 1./81.*m44*RSp561e*
         RS156pe - 1./81.*m44*RSp451e*RS156pe + 1./81.*m44*mp2*RN1p*
         RS145pe*RSp563e + 1./81.*m44*mp2*RN1p*RS156pe*RSp563e + 1./81.
         *m44*mp2*RN1p*RS156pe*RSp453e - 1./81.*m44*mp2*RN1p*RSp561e*
         RS145pe - 1./81.*m44*mp2*RN1p*RSp561e*RS156pe - 1./81.*m44*mp2
         *RN1p*RSp451e*RS156pe + 1./81.*m44*me2*RN1e*RS145pe*RSe563p + 
         1./81.*m44*me2*RN1e*RS156pe*RSe563p + 1./81.*m44*me2*RN1e*
         RS156pe*RSe453p - 1./81.*m44*me2*RN1e*RSe561p*RS145pe - 1./81.
         *m44*me2*RN1e*RSe561p*RS156pe - 1./81.*m44*me2*RN1e*RSe451p*
         RS156pe );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(m46,mu2) * ( 1./81.*m33*RS145pe*
         RSe563p - 1./81.*m33*RS145pe*RSe453p + 1./81.*m33*RS145pe*
         RSp563e - 1./81.*m33*RS145pe*RSp453e - 1./81.*m33*RS156pe*
         RSe563p + 1./81.*m33*RS156pe*RSe453p - 1./81.*m33*RS156pe*
         RSp563e + 1./81.*m33*RS156pe*RSp453e + 1./81.*m33*mp2*RN1p*
         RS145pe*RSp563e - 1./81.*m33*mp2*RN1p*RS145pe*RSp453e - 1./81.
         *m33*mp2*RN1p*RS156pe*RSp563e + 1./81.*m33*mp2*RN1p*RS156pe*
         RSp453e + 1./81.*m33*me2*RN1e*RS145pe*RSe563p - 1./81.*m33*me2
         *RN1e*RS145pe*RSe453p - 1./81.*m33*me2*RN1e*RS156pe*RSe563p + 
         1./81.*m33*me2*RN1e*RS156pe*RSe453p + 1./81.*m33*m66*RN1p*
         RS145pe*RSp563e + 1./81.*m33*m66*RN1p*RS145pe*RSp453e + 1./81.
         *m33*m66*RN1p*RS156pe*RSp453e + 1./81.*m33*m66*RN1e*RS145pe*
         RSe563p + 1./81.*m33*m66*RN1e*RS145pe*RSe453p + 1./81.*m33*m66
         *RN1e*RS156pe*RSe453p + 1./81.*m33*m66*RN13*RS145pe*RS345pe + 
         1./81.*m33*m66*RN13*RS145pe*RS356pe + 1./81.*m33*m66*RN13*
         RS156pe*RS345pe );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(m46,mu2) * ( 1./81.*m33*m44*RN1p*
         RS145pe*RSp563e + 1./81.*m33*m44*RN1p*RS156pe*RSp563e + 1./81.
         *m33*m44*RN1p*RS156pe*RSp453e + 1./81.*m33*m44*RN1e*RS145pe*
         RSe563p + 1./81.*m33*m44*RN1e*RS156pe*RSe563p + 1./81.*m33*m44
         *RN1e*RS156pe*RSe453p + 1./81.*m33*m44*RN13*RS145pe*RS356pe + 
         1./81.*m33*m44*RN13*RS156pe*RS345pe + 1./81.*m33*m44*RN13*
         RS156pe*RS356pe - 1./81.*m11*RS145pe*RS345pe + 1./81.*m11*
         RS145pe*RS356pe + 2./81.*m11*RS145pe*RSe563p - 2./81.*m11*
         RS145pe*RSe453p + 2./81.*m11*RS145pe*RSp563e - 2./81.*m11*
         RS145pe*RSp453e + 1./162.*m11*pow(RS145pe,2) + 1./81.*m11*
         RS156pe*RS345pe - 1./81.*m11*RS156pe*RS356pe - 2./81.*m11*
         RS156pe*RSe563p + 2./81.*m11*RS156pe*RSe453p - 2./81.*m11*
         RS156pe*RSp563e + 2./81.*m11*RS156pe*RSp453e - 1./81.*m11*
         RS156pe*RS145pe + 1./162.*m11*pow(RS156pe,2) + 2./81.*m11*mp2*
         RN1p*RS145pe*RSp563e - 2./81.*m11*mp2*RN1p*RS145pe*RSp453e - 2.
         /81.*m11*mp2*RN1p*RS156pe*RSp563e );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(m46,mu2) * ( 2./81.*m11*mp2*RN1p*
         RS156pe*RSp453e + 2./81.*m11*me2*RN1e*RS145pe*RSe563p - 2./81.
         *m11*me2*RN1e*RS145pe*RSe453p - 2./81.*m11*me2*RN1e*RS156pe*
         RSe563p + 2./81.*m11*me2*RN1e*RS156pe*RSe453p + 1./81.*m11*m66
         *RN1p*RS145pe*RSp563e + 1./81.*m11*m66*RN1p*RS145pe*RSp453e + 
         1./81.*m11*m66*RN1p*RS156pe*RSp453e + 1./81.*m11*m66*RN1p*
         RSp561e*RS145pe + 1./81.*m11*m66*RN1p*RSp451e*RS145pe + 1./81.
         *m11*m66*RN1p*RSp451e*RS156pe + 1./81.*m11*m66*RN1e*RS145pe*
         RSe563p + 1./81.*m11*m66*RN1e*RS145pe*RSe453p + 1./81.*m11*m66
         *RN1e*RS156pe*RSe453p + 1./81.*m11*m66*RN1e*RSe561p*RS145pe + 
         1./81.*m11*m66*RN1e*RSe451p*RS145pe + 1./81.*m11*m66*RN1e*
         RSe451p*RS156pe + 2./81.*m11*m66*RN13*RS145pe*RS345pe + 2./81.
         *m11*m66*RN13*RS145pe*RS356pe + 2./81.*m11*m66*RN13*RS156pe*
         RS345pe + 1./81.*m11*m44*RN1p*RS145pe*RSp563e + 1./81.*m11*m44
         *RN1p*RS156pe*RSp563e + 1./81.*m11*m44*RN1p*RS156pe*RSp453e + 
         1./81.*m11*m44*RN1p*RSp561e*RS145pe );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(m46,mu2) * ( 1./81.*m11*m44*RN1p*
         RSp561e*RS156pe + 1./81.*m11*m44*RN1p*RSp451e*RS156pe + 1./81.
         *m11*m44*RN1e*RS145pe*RSe563p + 1./81.*m11*m44*RN1e*RS156pe*
         RSe563p + 1./81.*m11*m44*RN1e*RS156pe*RSe453p + 1./81.*m11*m44
         *RN1e*RSe561p*RS145pe + 1./81.*m11*m44*RN1e*RSe561p*RS156pe + 
         1./81.*m11*m44*RN1e*RSe451p*RS156pe + 2./81.*m11*m44*RN13*
         RS145pe*RS356pe + 2./81.*m11*m44*RN13*RS156pe*RS345pe + 2./81.
         *m11*m44*RN13*RS156pe*RS356pe - 2./81.*m11*m33*RN13*RS145pe*
         RS345pe + 2./81.*m11*m33*RN13*RS145pe*RS356pe + 2./81.*m11*m33
         *RN13*RS156pe*RS345pe - 2./81.*m11*m33*RN13*RS156pe*RS356pe - 
         1./81.*pow(m11,2)*RN13*RS145pe*RS345pe + 1./81.*pow(m11,2)*
         RN13*RS145pe*RS356pe + 1./81.*pow(m11,2)*RN13*RS156pe*RS345pe
          - 1./81.*pow(m11,2)*RN13*RS156pe*RS356pe );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(m55,mu2) * ( 1./27.*m55*RS146pe*
         RSe463p + 1./27.*m55*RS146pe*RSp463e - 1./54.*m55*pow(
         RS146pe,2) - 1./81.*m55*RS546pe*RS146pe*RSe463p - 1./81.*m55*
         RS546pe*RS146pe*RSp463e + 1./162.*m55*RS546pe*pow(RS146pe,2)
          - 1./27.*m55*RSe461p*RS146pe + 1./81.*m55*RSe461p*RS546pe*
         RS146pe - 1./27.*m55*RSp461e*RS146pe + 1./81.*m55*RSp461e*
         RS546pe*RS146pe + 1./27.*m55*mp2*RN1p*RS146pe*RSp463e - 1./81.
         *m55*mp2*RN1p*RS546pe*RS146pe*RSp463e - 1./27.*m55*mp2*RN1p*
         RSp461e*RS146pe + 1./81.*m55*mp2*RN1p*RSp461e*RS546pe*RS146pe
          + 1./27.*m55*me2*RN1e*RS146pe*RSe463p - 1./81.*m55*me2*RN1e*
         RS546pe*RS146pe*RSe463p - 1./27.*m55*me2*RN1e*RSe461p*RS146pe
          + 1./81.*m55*me2*RN1e*RSe461p*RS546pe*RS146pe + 1./27.*m33*
         m55*RN1p*RS146pe*RSp463e - 1./81.*m33*m55*RN1p*RS546pe*RS146pe
         *RSp463e + 1./27.*m33*m55*RN1e*RS146pe*RSe463p - 1./81.*m33*
         m55*RN1e*RS546pe*RS146pe*RSe463p + 1./27.*m33*m55*RN13*RS146pe
         *RS346pe );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(m55,mu2) * (  - 1./81.*m33*m55*
         RN13*RS546pe*RS146pe*RS346pe + 1./27.*m11*m55*RN1p*RS146pe*
         RSp463e - 1./81.*m11*m55*RN1p*RS546pe*RS146pe*RSp463e + 1./27.
         *m11*m55*RN1p*RSp461e*RS146pe - 1./81.*m11*m55*RN1p*RSp461e*
         RS546pe*RS146pe + 1./27.*m11*m55*RN1e*RS146pe*RSe463p - 1./81.
         *m11*m55*RN1e*RS546pe*RS146pe*RSe463p + 1./27.*m11*m55*RN1e*
         RSe461p*RS146pe - 1./81.*m11*m55*RN1e*RSe461p*RS546pe*RS146pe
          + 2./27.*m11*m55*RN13*RS146pe*RS346pe - 2./81.*m11*m55*RN13*
         RS546pe*RS146pe*RS346pe );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(m56,mu2) * ( 1./81.*m66*RS145pe*
         RSe463p + 1./81.*m66*RS145pe*RSe453p + 1./81.*m66*RS145pe*
         RSp463e + 1./81.*m66*RS145pe*RSp453e - 1./162.*m66*pow(
         RS145pe,2) + 1./81.*m66*RS146pe*RSe453p + 1./81.*m66*RS146pe*
         RSp453e - 1./81.*m66*RS146pe*RS145pe - 1./81.*m66*RSe461p*
         RS145pe - 1./81.*m66*RSe451p*RS145pe - 1./81.*m66*RSe451p*
         RS146pe - 1./81.*m66*RSp461e*RS145pe - 1./81.*m66*RSp451e*
         RS145pe - 1./81.*m66*RSp451e*RS146pe + 1./81.*m66*mp2*RN1p*
         RS145pe*RSp463e + 1./81.*m66*mp2*RN1p*RS145pe*RSp453e + 1./81.
         *m66*mp2*RN1p*RS146pe*RSp453e - 1./81.*m66*mp2*RN1p*RSp461e*
         RS145pe - 1./81.*m66*mp2*RN1p*RSp451e*RS145pe - 1./81.*m66*mp2
         *RN1p*RSp451e*RS146pe + 1./81.*m66*me2*RN1e*RS145pe*RSe463p + 
         1./81.*m66*me2*RN1e*RS145pe*RSe453p + 1./81.*m66*me2*RN1e*
         RS146pe*RSe453p - 1./81.*m66*me2*RN1e*RSe461p*RS145pe - 1./81.
         *m66*me2*RN1e*RSe451p*RS145pe - 1./81.*m66*me2*RN1e*RSe451p*
         RS146pe );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(m56,mu2) * ( 1./81.*m55*RS145pe*
         RSe463p + 1./81.*m55*RS145pe*RSp463e + 1./81.*m55*RS146pe*
         RSe463p + 1./81.*m55*RS146pe*RSe453p + 1./81.*m55*RS146pe*
         RSp463e + 1./81.*m55*RS146pe*RSp453e - 1./81.*m55*RS146pe*
         RS145pe - 1./162.*m55*pow(RS146pe,2) - 1./81.*m55*RSe461p*
         RS145pe - 1./81.*m55*RSe461p*RS146pe - 1./81.*m55*RSe451p*
         RS146pe - 1./81.*m55*RSp461e*RS145pe - 1./81.*m55*RSp461e*
         RS146pe - 1./81.*m55*RSp451e*RS146pe + 1./81.*m55*mp2*RN1p*
         RS145pe*RSp463e + 1./81.*m55*mp2*RN1p*RS146pe*RSp463e + 1./81.
         *m55*mp2*RN1p*RS146pe*RSp453e - 1./81.*m55*mp2*RN1p*RSp461e*
         RS145pe - 1./81.*m55*mp2*RN1p*RSp461e*RS146pe - 1./81.*m55*mp2
         *RN1p*RSp451e*RS146pe + 1./81.*m55*me2*RN1e*RS145pe*RSe463p + 
         1./81.*m55*me2*RN1e*RS146pe*RSe463p + 1./81.*m55*me2*RN1e*
         RS146pe*RSe453p - 1./81.*m55*me2*RN1e*RSe461p*RS145pe - 1./81.
         *m55*me2*RN1e*RSe461p*RS146pe - 1./81.*m55*me2*RN1e*RSe451p*
         RS146pe );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(m56,mu2) * ( 1./81.*m33*RS145pe*
         RSe463p - 1./81.*m33*RS145pe*RSe453p + 1./81.*m33*RS145pe*
         RSp463e - 1./81.*m33*RS145pe*RSp453e - 1./81.*m33*RS146pe*
         RSe463p + 1./81.*m33*RS146pe*RSe453p - 1./81.*m33*RS146pe*
         RSp463e + 1./81.*m33*RS146pe*RSp453e + 1./81.*m33*mp2*RN1p*
         RS145pe*RSp463e - 1./81.*m33*mp2*RN1p*RS145pe*RSp453e - 1./81.
         *m33*mp2*RN1p*RS146pe*RSp463e + 1./81.*m33*mp2*RN1p*RS146pe*
         RSp453e + 1./81.*m33*me2*RN1e*RS145pe*RSe463p - 1./81.*m33*me2
         *RN1e*RS145pe*RSe453p - 1./81.*m33*me2*RN1e*RS146pe*RSe463p + 
         1./81.*m33*me2*RN1e*RS146pe*RSe453p + 1./81.*m33*m66*RN1p*
         RS145pe*RSp463e + 1./81.*m33*m66*RN1p*RS145pe*RSp453e + 1./81.
         *m33*m66*RN1p*RS146pe*RSp453e + 1./81.*m33*m66*RN1e*RS145pe*
         RSe463p + 1./81.*m33*m66*RN1e*RS145pe*RSe453p + 1./81.*m33*m66
         *RN1e*RS146pe*RSe453p + 1./81.*m33*m66*RN13*RS145pe*RS345pe + 
         1./81.*m33*m66*RN13*RS145pe*RS346pe + 1./81.*m33*m66*RN13*
         RS146pe*RS345pe );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(m56,mu2) * ( 1./81.*m33*m55*RN1p*
         RS145pe*RSp463e + 1./81.*m33*m55*RN1p*RS146pe*RSp463e + 1./81.
         *m33*m55*RN1p*RS146pe*RSp453e + 1./81.*m33*m55*RN1e*RS145pe*
         RSe463p + 1./81.*m33*m55*RN1e*RS146pe*RSe463p + 1./81.*m33*m55
         *RN1e*RS146pe*RSe453p + 1./81.*m33*m55*RN13*RS145pe*RS346pe + 
         1./81.*m33*m55*RN13*RS146pe*RS345pe + 1./81.*m33*m55*RN13*
         RS146pe*RS346pe - 1./81.*m11*RS145pe*RS345pe + 1./81.*m11*
         RS145pe*RS346pe + 2./81.*m11*RS145pe*RSe463p - 2./81.*m11*
         RS145pe*RSe453p + 2./81.*m11*RS145pe*RSp463e - 2./81.*m11*
         RS145pe*RSp453e + 1./162.*m11*pow(RS145pe,2) + 1./81.*m11*
         RS146pe*RS345pe - 1./81.*m11*RS146pe*RS346pe - 2./81.*m11*
         RS146pe*RSe463p + 2./81.*m11*RS146pe*RSe453p - 2./81.*m11*
         RS146pe*RSp463e + 2./81.*m11*RS146pe*RSp453e - 1./81.*m11*
         RS146pe*RS145pe + 1./162.*m11*pow(RS146pe,2) + 2./81.*m11*mp2*
         RN1p*RS145pe*RSp463e - 2./81.*m11*mp2*RN1p*RS145pe*RSp453e - 2.
         /81.*m11*mp2*RN1p*RS146pe*RSp463e );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(m56,mu2) * ( 2./81.*m11*mp2*RN1p*
         RS146pe*RSp453e + 2./81.*m11*me2*RN1e*RS145pe*RSe463p - 2./81.
         *m11*me2*RN1e*RS145pe*RSe453p - 2./81.*m11*me2*RN1e*RS146pe*
         RSe463p + 2./81.*m11*me2*RN1e*RS146pe*RSe453p + 1./81.*m11*m66
         *RN1p*RS145pe*RSp463e + 1./81.*m11*m66*RN1p*RS145pe*RSp453e + 
         1./81.*m11*m66*RN1p*RS146pe*RSp453e + 1./81.*m11*m66*RN1p*
         RSp461e*RS145pe + 1./81.*m11*m66*RN1p*RSp451e*RS145pe + 1./81.
         *m11*m66*RN1p*RSp451e*RS146pe + 1./81.*m11*m66*RN1e*RS145pe*
         RSe463p + 1./81.*m11*m66*RN1e*RS145pe*RSe453p + 1./81.*m11*m66
         *RN1e*RS146pe*RSe453p + 1./81.*m11*m66*RN1e*RSe461p*RS145pe + 
         1./81.*m11*m66*RN1e*RSe451p*RS145pe + 1./81.*m11*m66*RN1e*
         RSe451p*RS146pe + 2./81.*m11*m66*RN13*RS145pe*RS345pe + 2./81.
         *m11*m66*RN13*RS145pe*RS346pe + 2./81.*m11*m66*RN13*RS146pe*
         RS345pe + 1./81.*m11*m55*RN1p*RS145pe*RSp463e + 1./81.*m11*m55
         *RN1p*RS146pe*RSp463e + 1./81.*m11*m55*RN1p*RS146pe*RSp453e + 
         1./81.*m11*m55*RN1p*RSp461e*RS145pe );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(m56,mu2) * ( 1./81.*m11*m55*RN1p*
         RSp461e*RS146pe + 1./81.*m11*m55*RN1p*RSp451e*RS146pe + 1./81.
         *m11*m55*RN1e*RS145pe*RSe463p + 1./81.*m11*m55*RN1e*RS146pe*
         RSe463p + 1./81.*m11*m55*RN1e*RS146pe*RSe453p + 1./81.*m11*m55
         *RN1e*RSe461p*RS145pe + 1./81.*m11*m55*RN1e*RSe461p*RS146pe + 
         1./81.*m11*m55*RN1e*RSe451p*RS146pe + 2./81.*m11*m55*RN13*
         RS145pe*RS346pe + 2./81.*m11*m55*RN13*RS146pe*RS345pe + 2./81.
         *m11*m55*RN13*RS146pe*RS346pe - 2./81.*m11*m33*RN13*RS145pe*
         RS345pe + 2./81.*m11*m33*RN13*RS145pe*RS346pe + 2./81.*m11*m33
         *RN13*RS146pe*RS345pe - 2./81.*m11*m33*RN13*RS146pe*RS346pe - 
         1./81.*pow(m11,2)*RN13*RS145pe*RS345pe + 1./81.*pow(m11,2)*
         RN13*RS145pe*RS346pe + 1./81.*pow(m11,2)*RN13*RS146pe*RS345pe
          - 1./81.*pow(m11,2)*RN13*RS146pe*RS346pe );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(m66,mu2) * ( 1./27.*m66*RS145pe*
         RSe453p + 1./27.*m66*RS145pe*RSp453e - 1./54.*m66*pow(
         RS145pe,2) - 1./81.*m66*RS645pe*RS145pe*RSe453p - 1./81.*m66*
         RS645pe*RS145pe*RSp453e + 1./162.*m66*RS645pe*pow(RS145pe,2)
          - 1./27.*m66*RSe451p*RS145pe + 1./81.*m66*RSe451p*RS645pe*
         RS145pe - 1./27.*m66*RSp451e*RS145pe + 1./81.*m66*RSp451e*
         RS645pe*RS145pe + 1./27.*m66*mp2*RN1p*RS145pe*RSp453e - 1./81.
         *m66*mp2*RN1p*RS645pe*RS145pe*RSp453e - 1./27.*m66*mp2*RN1p*
         RSp451e*RS145pe + 1./81.*m66*mp2*RN1p*RSp451e*RS645pe*RS145pe
          + 1./27.*m66*me2*RN1e*RS145pe*RSe453p - 1./81.*m66*me2*RN1e*
         RS645pe*RS145pe*RSe453p - 1./27.*m66*me2*RN1e*RSe451p*RS145pe
          + 1./81.*m66*me2*RN1e*RSe451p*RS645pe*RS145pe + 1./27.*m33*
         m66*RN1p*RS145pe*RSp453e - 1./81.*m33*m66*RN1p*RS645pe*RS145pe
         *RSp453e + 1./27.*m33*m66*RN1e*RS145pe*RSe453p - 1./81.*m33*
         m66*RN1e*RS645pe*RS145pe*RSe453p + 1./27.*m33*m66*RN13*RS145pe
         *RS345pe );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(m66,mu2) * (  - 1./81.*m33*m66*
         RN13*RS645pe*RS145pe*RS345pe + 1./27.*m11*m66*RN1p*RS145pe*
         RSp453e - 1./81.*m11*m66*RN1p*RS645pe*RS145pe*RSp453e + 1./27.
         *m11*m66*RN1p*RSp451e*RS145pe - 1./81.*m11*m66*RN1p*RSp451e*
         RS645pe*RS145pe + 1./27.*m11*m66*RN1e*RS145pe*RSe453p - 1./81.
         *m11*m66*RN1e*RS645pe*RS145pe*RSe453p + 1./27.*m11*m66*RN1e*
         RSe451p*RS145pe - 1./81.*m11*m66*RN1e*RSe451p*RS645pe*RS145pe
          + 2./27.*m11*m66*RN13*RS145pe*RS345pe - 2./81.*m11*m66*RN13*
         RS645pe*RS145pe*RS345pe );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(me2,mu2) * (  - 1./270.*RD1456ep*
         RSe45613p + 1./540.*RD1456ep*RSe45633p + 1./540.*RSe45611p*
         RD1456ep + 1./135.*me2*RSe45613p*RS14563pe - 1./270.*me2*
         RSe45633p*RS14563pe + 1./270.*me2*RS1456pee*RSe45613p - 1./540.
         *me2*RS1456pee*RSe45633p + 1./270.*me2*RS1456ppe*RSe45613p - 1.
         /540.*me2*RS1456ppe*RSe45633p - 1./270.*me2*RSe45611p*
         RS14563pe - 1./540.*me2*RSe45611p*RS1456pee - 1./540.*me2*
         RSe45611p*RS1456ppe - 1./270.*me2*RS145pe*RSe45613p + 1./540.*
         me2*RS145pe*RSe45633p + 1./540.*me2*RS145pe*RSe45611p - 1./270.
         *me2*RS146pe*RSe45613p + 1./540.*me2*RS146pe*RSe45633p + 1./
         540.*me2*RS146pe*RSe45611p - 1./270.*me2*RS156pe*RSe45613p + 1.
         /540.*me2*RS156pe*RSe45633p + 1./540.*me2*RS156pe*RSe45611p - 
         1./81.*m66*RSe456p*RS145pe*RSe453p - 1./81.*m66*RSe456p*
         RS145pe*RSp453e + 1./162.*m66*RSe456p*pow(RS145pe,2) + 1./81.*
         m66*RSe456p*RSe451p*RS145pe + 1./81.*m66*RSe456p*RSp451e*
         RS145pe );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(me2,mu2) * (  - 1./81.*m66*mp2*
         RN1p*RSe456p*RS145pe*RSp453e + 1./81.*m66*mp2*RN1p*RSe456p*
         RSp451e*RS145pe - 1./81.*m66*me2*RN1e*RSe456p*RS145pe*RSe453p
          + 1./81.*m66*me2*RN1e*RSe456p*RSe451p*RS145pe - 1./81.*m55*
         RSe465p*RS146pe*RSe463p - 1./81.*m55*RSe465p*RS146pe*RSp463e
          + 1./162.*m55*RSe465p*pow(RS146pe,2) + 1./81.*m55*RSe465p*
         RSe461p*RS146pe + 1./81.*m55*RSe465p*RSp461e*RS146pe - 1./81.*
         m55*mp2*RN1p*RSe465p*RS146pe*RSp463e + 1./81.*m55*mp2*RN1p*
         RSe465p*RSp461e*RS146pe - 1./81.*m55*me2*RN1e*RSe465p*RS146pe*
         RSe463p + 1./81.*m55*me2*RN1e*RSe465p*RSe461p*RS146pe - 1./81.
         *m44*RSe564p*RS156pe*RSe563p - 1./81.*m44*RSe564p*RS156pe*
         RSp563e + 1./162.*m44*RSe564p*pow(RS156pe,2) + 1./81.*m44*
         RSe564p*RSe561p*RS156pe + 1./81.*m44*RSe564p*RSp561e*RS156pe
          - 1./81.*m44*mp2*RN1p*RSe564p*RS156pe*RSp563e + 1./81.*m44*
         mp2*RN1p*RSe564p*RSp561e*RS156pe - 1./81.*m44*me2*RN1e*RSe564p
         *RS156pe*RSe563p );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(me2,mu2) * ( 1./81.*m44*me2*RN1e*
         RSe564p*RSe561p*RS156pe + 13./270.*m33*RSe45613p*RS14563pe - 7.
         /540.*m33*RSe45633p*RS14563pe - 1./540.*m33*RS1456pee*
         RSe45613p + 17./1080.*m33*RS1456pee*RSe45633p - 1./540.*m33*
         RS1456ppe*RSe45613p + 17./1080.*m33*RS1456ppe*RSe45633p + 1./
         540.*m33*RSe45611p*RS14563pe - 1./72.*m33*RSe45611p*RS1456pee
          - 1./72.*m33*RSe45611p*RS1456ppe + 1./540.*m33*RS145pe*
         RSe45613p - 17./1080.*m33*RS145pe*RSe45633p + 1./72.*m33*
         RS145pe*RSe45611p + 1./540.*m33*RS146pe*RSe45613p - 17./1080.*
         m33*RS146pe*RSe45633p + 1./72.*m33*RS146pe*RSe45611p + 1./540.
         *m33*RS156pe*RSe45613p - 17./1080.*m33*RS156pe*RSe45633p + 1./
         72.*m33*RS156pe*RSe45611p - 1./81.*m33*m66*RN1p*RSe456p*
         RS145pe*RSp453e - 1./81.*m33*m66*RN1e*RSe456p*RS145pe*RSe453p
          - 1./81.*m33*m66*RN13*RSe456p*RS145pe*RS345pe - 1./81.*m33*
         m55*RN1p*RSe465p*RS146pe*RSp463e - 1./81.*m33*m55*RN1e*RSe465p
         *RS146pe*RSe463p );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(me2,mu2) * (  - 1./81.*m33*m55*
         RN13*RSe465p*RS146pe*RS346pe - 1./81.*m33*m44*RN1p*RSe564p*
         RS156pe*RSp563e - 1./81.*m33*m44*RN1e*RSe564p*RS156pe*RSe563p
          - 1./81.*m33*m44*RN13*RSe564p*RS156pe*RS356pe + 1./27.*pow(
         m33,2)*RN13*RSe45633p*RS14563pe + 1./18.*m11*RSe45613p*
         RS14563pe - 1./540.*m11*RSe45633p*RS14563pe + 1./60.*m11*
         RS1456pee*RSe45613p + 1./72.*m11*RS1456pee*RSe45633p + 1./60.*
         m11*RS1456ppe*RSe45613p + 1./72.*m11*RS1456ppe*RSe45633p + 11./
         540.*m11*RSe45611p*RS14563pe + 1./27.*m11*RSe45611p*RSe45613p
          + 1./27.*m11*RSe45611p*RSp45613e - 13./1080.*m11*RSe45611p*
         RS1456pee - 13./1080.*m11*RSe45611p*RS1456ppe - 1./27.*m11*
         pow(RSe45611p,2) - 1./27.*m11*RSp45611e*RSe45611p - 1./60.*m11
         *RS145pe*RSe45613p - 1./72.*m11*RS145pe*RSe45633p + 13./1080.*
         m11*RS145pe*RSe45611p - 1./60.*m11*RS146pe*RSe45613p - 1./72.*
         m11*RS146pe*RSe45633p + 13./1080.*m11*RS146pe*RSe45611p - 1./
         60.*m11*RS156pe*RSe45613p );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(me2,mu2) * (  - 1./72.*m11*RS156pe
         *RSe45633p + 13./1080.*m11*RS156pe*RSe45611p + 1./27.*m11*mp2*
         RN1p*RSe45611p*RSp45613e - 1./27.*m11*mp2*RN1p*RSp45611e*
         RSe45611p + 1./27.*m11*me2*RN1e*RS1456pee*RSe45613p + 1./27.*
         m11*me2*RN1e*RS1456ppe*RSe45613p + 1./27.*m11*me2*RN1e*
         RSe45611p*RSe45613p - 1./27.*m11*me2*RN1e*RSe45611p*RS1456pee
          - 1./27.*m11*me2*RN1e*RSe45611p*RS1456ppe - 1./27.*m11*me2*
         RN1e*pow(RSe45611p,2) - 1./27.*m11*me2*RN1e*RS145pe*RSe45613p
          + 1./27.*m11*me2*RN1e*RS145pe*RSe45611p - 1./27.*m11*me2*RN1e
         *RS146pe*RSe45613p + 1./27.*m11*me2*RN1e*RS146pe*RSe45611p - 1.
         /27.*m11*me2*RN1e*RS156pe*RSe45613p + 1./27.*m11*me2*RN1e*
         RS156pe*RSe45611p - 1./81.*m11*m66*RN1p*RSe456p*RS145pe*
         RSp453e - 1./81.*m11*m66*RN1p*RSe456p*RSp451e*RS145pe - 1./81.
         *m11*m66*RN1e*RSe456p*RS145pe*RSe453p - 1./81.*m11*m66*RN1e*
         RSe456p*RSe451p*RS145pe - 2./81.*m11*m66*RN13*RSe456p*RS145pe*
         RS345pe );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(me2,mu2) * (  - 1./81.*m11*m55*
         RN1p*RSe465p*RS146pe*RSp463e - 1./81.*m11*m55*RN1p*RSe465p*
         RSp461e*RS146pe - 1./81.*m11*m55*RN1e*RSe465p*RS146pe*RSe463p
          - 1./81.*m11*m55*RN1e*RSe465p*RSe461p*RS146pe - 2./81.*m11*
         m55*RN13*RSe465p*RS146pe*RS346pe - 1./81.*m11*m44*RN1p*RSe564p
         *RS156pe*RSp563e - 1./81.*m11*m44*RN1p*RSe564p*RSp561e*RS156pe
          - 1./81.*m11*m44*RN1e*RSe564p*RS156pe*RSe563p - 1./81.*m11*
         m44*RN1e*RSe564p*RSe561p*RS156pe - 2./81.*m11*m44*RN13*RSe564p
         *RS156pe*RS356pe + 1./27.*m11*m33*RN1p*RSe45611p*RSp45613e + 1.
         /27.*m11*m33*RN1e*RS1456pee*RSe45613p + 1./27.*m11*m33*RN1e*
         RS1456ppe*RSe45613p + 1./27.*m11*m33*RN1e*RSe45611p*RSe45613p
          - 1./27.*m11*m33*RN1e*RS145pe*RSe45613p - 1./27.*m11*m33*RN1e
         *RS146pe*RSe45613p - 1./27.*m11*m33*RN1e*RS156pe*RSe45613p + 2.
         /27.*m11*m33*RN13*RSe45633p*RS14563pe + 1./27.*m11*m33*RN13*
         RSe45611p*RS34561pe + 1./27.*pow(m11,2)*RN1p*RSe45611p*
         RSp45613e );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(me2,mu2) * ( 1./27.*pow(m11,2)*
         RN1p*RSp45611e*RSe45611p + 1./27.*pow(m11,2)*RN1e*RS1456pee*
         RSe45613p + 1./27.*pow(m11,2)*RN1e*RS1456ppe*RSe45613p + 1./27.
         *pow(m11,2)*RN1e*RSe45611p*RSe45613p + 1./27.*pow(m11,2)*RN1e*
         RSe45611p*RS1456pee + 1./27.*pow(m11,2)*RN1e*RSe45611p*
         RS1456ppe + 1./27.*pow(m11,2)*RN1e*pow(RSe45611p,2) - 1./27.*
         pow(m11,2)*RN1e*RS145pe*RSe45613p - 1./27.*pow(m11,2)*RN1e*
         RS145pe*RSe45611p - 1./27.*pow(m11,2)*RN1e*RS146pe*RSe45613p
          - 1./27.*pow(m11,2)*RN1e*RS146pe*RSe45611p - 1./27.*pow(
         m11,2)*RN1e*RS156pe*RSe45613p - 1./27.*pow(m11,2)*RN1e*RS156pe
         *RSe45611p + 2./27.*pow(m11,2)*RN13*RSe45611p*RS34561pe );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(mp2,mu2) * (  - 1./270.*RD1456ep*
         RSp45613e + 1./540.*RD1456ep*RSp45633e + 1./540.*RSp45611e*
         RD1456ep + 1./135.*mp2*RSp45613e*RS14563pe - 1./270.*mp2*
         RSp45633e*RS14563pe + 1./270.*mp2*RS1456pee*RSp45613e - 1./540.
         *mp2*RS1456pee*RSp45633e + 1./270.*mp2*RS1456ppe*RSp45613e - 1.
         /540.*mp2*RS1456ppe*RSp45633e - 1./270.*mp2*RSp45611e*
         RS14563pe - 1./540.*mp2*RSp45611e*RS1456pee - 1./540.*mp2*
         RSp45611e*RS1456ppe - 1./270.*mp2*RS145pe*RSp45613e + 1./540.*
         mp2*RS145pe*RSp45633e + 1./540.*mp2*RS145pe*RSp45611e - 1./270.
         *mp2*RS146pe*RSp45613e + 1./540.*mp2*RS146pe*RSp45633e + 1./
         540.*mp2*RS146pe*RSp45611e - 1./270.*mp2*RS156pe*RSp45613e + 1.
         /540.*mp2*RS156pe*RSp45633e + 1./540.*mp2*RS156pe*RSp45611e - 
         1./81.*m66*RSp456e*RS145pe*RSe453p - 1./81.*m66*RSp456e*
         RS145pe*RSp453e + 1./162.*m66*RSp456e*pow(RS145pe,2) + 1./81.*
         m66*RSp456e*RSe451p*RS145pe + 1./81.*m66*RSp456e*RSp451e*
         RS145pe );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(mp2,mu2) * (  - 1./81.*m66*mp2*
         RN1p*RSp456e*RS145pe*RSp453e + 1./81.*m66*mp2*RN1p*RSp456e*
         RSp451e*RS145pe - 1./81.*m66*me2*RN1e*RSp456e*RS145pe*RSe453p
          + 1./81.*m66*me2*RN1e*RSp456e*RSe451p*RS145pe - 1./81.*m55*
         RSp465e*RS146pe*RSe463p - 1./81.*m55*RSp465e*RS146pe*RSp463e
          + 1./162.*m55*RSp465e*pow(RS146pe,2) + 1./81.*m55*RSp465e*
         RSe461p*RS146pe + 1./81.*m55*RSp465e*RSp461e*RS146pe - 1./81.*
         m55*mp2*RN1p*RSp465e*RS146pe*RSp463e + 1./81.*m55*mp2*RN1p*
         RSp465e*RSp461e*RS146pe - 1./81.*m55*me2*RN1e*RSp465e*RS146pe*
         RSe463p + 1./81.*m55*me2*RN1e*RSp465e*RSe461p*RS146pe - 1./81.
         *m44*RSp564e*RS156pe*RSe563p - 1./81.*m44*RSp564e*RS156pe*
         RSp563e + 1./162.*m44*RSp564e*pow(RS156pe,2) + 1./81.*m44*
         RSp564e*RSe561p*RS156pe + 1./81.*m44*RSp564e*RSp561e*RS156pe
          - 1./81.*m44*mp2*RN1p*RSp564e*RS156pe*RSp563e + 1./81.*m44*
         mp2*RN1p*RSp564e*RSp561e*RS156pe - 1./81.*m44*me2*RN1e*RSp564e
         *RS156pe*RSe563p );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(mp2,mu2) * ( 1./81.*m44*me2*RN1e*
         RSp564e*RSe561p*RS156pe + 13./270.*m33*RSp45613e*RS14563pe - 7.
         /540.*m33*RSp45633e*RS14563pe - 1./540.*m33*RS1456pee*
         RSp45613e + 17./1080.*m33*RS1456pee*RSp45633e - 1./540.*m33*
         RS1456ppe*RSp45613e + 17./1080.*m33*RS1456ppe*RSp45633e + 1./
         540.*m33*RSp45611e*RS14563pe - 1./72.*m33*RSp45611e*RS1456pee
          - 1./72.*m33*RSp45611e*RS1456ppe + 1./540.*m33*RS145pe*
         RSp45613e - 17./1080.*m33*RS145pe*RSp45633e + 1./72.*m33*
         RS145pe*RSp45611e + 1./540.*m33*RS146pe*RSp45613e - 17./1080.*
         m33*RS146pe*RSp45633e + 1./72.*m33*RS146pe*RSp45611e + 1./540.
         *m33*RS156pe*RSp45613e - 17./1080.*m33*RS156pe*RSp45633e + 1./
         72.*m33*RS156pe*RSp45611e - 1./81.*m33*m66*RN1p*RSp456e*
         RS145pe*RSp453e - 1./81.*m33*m66*RN1e*RSp456e*RS145pe*RSe453p
          - 1./81.*m33*m66*RN13*RSp456e*RS145pe*RS345pe - 1./81.*m33*
         m55*RN1p*RSp465e*RS146pe*RSp463e - 1./81.*m33*m55*RN1e*RSp465e
         *RS146pe*RSe463p );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(mp2,mu2) * (  - 1./81.*m33*m55*
         RN13*RSp465e*RS146pe*RS346pe - 1./81.*m33*m44*RN1p*RSp564e*
         RS156pe*RSp563e - 1./81.*m33*m44*RN1e*RSp564e*RS156pe*RSe563p
          - 1./81.*m33*m44*RN13*RSp564e*RS156pe*RS356pe + 1./27.*pow(
         m33,2)*RN13*RSp45633e*RS14563pe + 1./18.*m11*RSp45613e*
         RS14563pe - 1./540.*m11*RSp45633e*RS14563pe + 1./60.*m11*
         RS1456pee*RSp45613e + 1./72.*m11*RS1456pee*RSp45633e + 1./60.*
         m11*RS1456ppe*RSp45613e + 1./72.*m11*RS1456ppe*RSp45633e + 11./
         540.*m11*RSp45611e*RS14563pe + 1./27.*m11*RSp45611e*RSe45613p
          + 1./27.*m11*RSp45611e*RSp45613e - 13./1080.*m11*RSp45611e*
         RS1456pee - 13./1080.*m11*RSp45611e*RS1456ppe - 1./27.*m11*
         RSp45611e*RSe45611p - 1./27.*m11*pow(RSp45611e,2) - 1./60.*m11
         *RS145pe*RSp45613e - 1./72.*m11*RS145pe*RSp45633e + 13./1080.*
         m11*RS145pe*RSp45611e - 1./60.*m11*RS146pe*RSp45613e - 1./72.*
         m11*RS146pe*RSp45633e + 13./1080.*m11*RS146pe*RSp45611e - 1./
         60.*m11*RS156pe*RSp45613e );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(mp2,mu2) * (  - 1./72.*m11*RS156pe
         *RSp45633e + 13./1080.*m11*RS156pe*RSp45611e + 1./27.*m11*mp2*
         RN1p*RS1456pee*RSp45613e + 1./27.*m11*mp2*RN1p*RS1456ppe*
         RSp45613e + 1./27.*m11*mp2*RN1p*RSp45611e*RSp45613e - 1./27.*
         m11*mp2*RN1p*RSp45611e*RS1456pee - 1./27.*m11*mp2*RN1p*
         RSp45611e*RS1456ppe - 1./27.*m11*mp2*RN1p*pow(RSp45611e,2) - 1.
         /27.*m11*mp2*RN1p*RS145pe*RSp45613e + 1./27.*m11*mp2*RN1p*
         RS145pe*RSp45611e - 1./27.*m11*mp2*RN1p*RS146pe*RSp45613e + 1./
         27.*m11*mp2*RN1p*RS146pe*RSp45611e - 1./27.*m11*mp2*RN1p*
         RS156pe*RSp45613e + 1./27.*m11*mp2*RN1p*RS156pe*RSp45611e + 1./
         27.*m11*me2*RN1e*RSp45611e*RSe45613p - 1./27.*m11*me2*RN1e*
         RSp45611e*RSe45611p - 1./81.*m11*m66*RN1p*RSp456e*RS145pe*
         RSp453e - 1./81.*m11*m66*RN1p*RSp456e*RSp451e*RS145pe - 1./81.
         *m11*m66*RN1e*RSp456e*RS145pe*RSe453p - 1./81.*m11*m66*RN1e*
         RSp456e*RSe451p*RS145pe - 2./81.*m11*m66*RN13*RSp456e*RS145pe*
         RS345pe );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(mp2,mu2) * (  - 1./81.*m11*m55*
         RN1p*RSp465e*RS146pe*RSp463e - 1./81.*m11*m55*RN1p*RSp465e*
         RSp461e*RS146pe - 1./81.*m11*m55*RN1e*RSp465e*RS146pe*RSe463p
          - 1./81.*m11*m55*RN1e*RSp465e*RSe461p*RS146pe - 2./81.*m11*
         m55*RN13*RSp465e*RS146pe*RS346pe - 1./81.*m11*m44*RN1p*RSp564e
         *RS156pe*RSp563e - 1./81.*m11*m44*RN1p*RSp564e*RSp561e*RS156pe
          - 1./81.*m11*m44*RN1e*RSp564e*RS156pe*RSe563p - 1./81.*m11*
         m44*RN1e*RSp564e*RSe561p*RS156pe - 2./81.*m11*m44*RN13*RSp564e
         *RS156pe*RS356pe + 1./27.*m11*m33*RN1p*RS1456pee*RSp45613e + 1.
         /27.*m11*m33*RN1p*RS1456ppe*RSp45613e + 1./27.*m11*m33*RN1p*
         RSp45611e*RSp45613e - 1./27.*m11*m33*RN1p*RS145pe*RSp45613e - 
         1./27.*m11*m33*RN1p*RS146pe*RSp45613e - 1./27.*m11*m33*RN1p*
         RS156pe*RSp45613e + 1./27.*m11*m33*RN1e*RSp45611e*RSe45613p + 
         2./27.*m11*m33*RN13*RSp45633e*RS14563pe + 1./27.*m11*m33*RN13*
         RSp45611e*RS34561pe + 1./27.*pow(m11,2)*RN1p*RS1456pee*
         RSp45613e );

      M13p6x23R +=  + Ab(m11,mu2)*Ab(mp2,mu2) * ( 1./27.*pow(m11,2)*
         RN1p*RS1456ppe*RSp45613e + 1./27.*pow(m11,2)*RN1p*RSp45611e*
         RSp45613e + 1./27.*pow(m11,2)*RN1p*RSp45611e*RS1456pee + 1./27.
         *pow(m11,2)*RN1p*RSp45611e*RS1456ppe + 1./27.*pow(m11,2)*RN1p*
         pow(RSp45611e,2) - 1./27.*pow(m11,2)*RN1p*RS145pe*RSp45613e - 
         1./27.*pow(m11,2)*RN1p*RS145pe*RSp45611e - 1./27.*pow(m11,2)*
         RN1p*RS146pe*RSp45613e - 1./27.*pow(m11,2)*RN1p*RS146pe*
         RSp45611e - 1./27.*pow(m11,2)*RN1p*RS156pe*RSp45613e - 1./27.*
         pow(m11,2)*RN1p*RS156pe*RSp45611e + 1./27.*pow(m11,2)*RN1e*
         RSp45611e*RSe45613p + 1./27.*pow(m11,2)*RN1e*RSp45611e*
         RSe45611p + 2./27.*pow(m11,2)*RN13*RSp45611e*RS34561pe );

      M13p6x23R +=  + Ab(m11,mu2)*Bb(m11,mu2) * ( 1./540.*pow(
         RD1456ep,2) + 1./540.*m33*RD1456ep*RS14563pe - 1./72.*m33*
         RS1456pee*RD1456ep - 1./72.*m33*RS1456ppe*RD1456ep + 1./72.*
         m33*RS145pe*RD1456ep + 1./72.*m33*RS146pe*RD1456ep + 1./72.*
         m33*RS156pe*RD1456ep + 1./60.*m11*RD1456ep*RS14563pe + 1./27.*
         m11*RD1456ep*RSe45613p + 1./27.*m11*RD1456ep*RSp45613e + 5./
         216.*m11*RS1456pee*RD1456ep + 5./216.*m11*RS1456ppe*RD1456ep
          - 1./27.*m11*RSe45611p*RD1456ep - 1./27.*m11*RSp45611e*
         RD1456ep - 5./216.*m11*RS145pe*RD1456ep - 5./216.*m11*RS146pe*
         RD1456ep - 5./216.*m11*RS156pe*RD1456ep + 1./27.*m11*mp2*RN1p*
         RD1456ep*RSp45613e - 1./27.*m11*mp2*RN1p*RSp45611e*RD1456ep + 
         1./27.*m11*me2*RN1e*RD1456ep*RSe45613p - 1./27.*m11*me2*RN1e*
         RSe45611p*RD1456ep - 1./27.*m11*m33*RS1456pee*RS14563pe - 1./
         27.*m11*m33*RS1456ppe*RS14563pe + 1./27.*m11*m33*RS145pe*
         RS14563pe + 1./27.*m11*m33*RS146pe*RS14563pe + 1./27.*m11*m33*
         RS156pe*RS14563pe );

      M13p6x23R +=  + Ab(m11,mu2)*Bb(m11,mu2) * ( 1./27.*m11*m33*RN1p*
         RD1456ep*RSp45613e + 1./27.*m11*m33*RN1e*RD1456ep*RSe45613p + 
         1./27.*m11*m33*RN13*RD1456ep*RS34561pe - 2./27.*pow(m11,2)*
         RS1456pee*RS14563pe - 2./27.*pow(m11,2)*RS1456ppe*RS14563pe + 
         2./27.*pow(m11,2)*RS145pe*RS14563pe + 2./27.*pow(m11,2)*
         RS146pe*RS14563pe + 2./27.*pow(m11,2)*RS156pe*RS14563pe + 1./
         27.*pow(m11,2)*RN1p*RD1456ep*RSp45613e + 1./27.*pow(m11,2)*
         RN1p*RSp45611e*RD1456ep + 1./27.*pow(m11,2)*RN1e*RD1456ep*
         RSe45613p + 1./27.*pow(m11,2)*RN1e*RSe45611p*RD1456ep + 2./27.
         *pow(m11,2)*RN13*RD1456ep*RS34561pe );

      M13p6x23R +=  + Ab(m11,mu2)*Bb(m33,mu2) * ( 1./540.*RD1456ep*
         RD3456ep - 1./60.*m33*RD3456ep*RS14563pe + 1./72.*m33*
         RS1456pee*RD3456ep + 1./72.*m33*RS1456ppe*RD3456ep - 1./72.*
         m33*RS145pe*RD3456ep - 1./72.*m33*RS146pe*RD3456ep - 1./72.*
         m33*RS156pe*RD3456ep + 1./27.*pow(m33,2)*RN13*RD3456ep*
         RS14563pe - 1./540.*m11*RD3456ep*RS14563pe + 1./72.*m11*
         RS1456pee*RD3456ep + 1./72.*m11*RS1456ppe*RD3456ep - 1./72.*
         m11*RS145pe*RD3456ep - 1./72.*m11*RS146pe*RD3456ep - 1./72.*
         m11*RS156pe*RD3456ep + 2./27.*m11*m33*RN13*RD3456ep*RS14563pe
          );

      M13p6x23R +=  + Ab(m13,mu2)*pi16 * ( 1./144.*pow(m33,2)*pow(
         RS34561pe,2) + 1./72.*pow(m33,2)*RS14563pe*RS34561pe + 1./144.
         *pow(m33,2)*pow(RS14563pe,2) + 1./72.*pow(m33,2)*RSe45613p*
         RS34561pe + 1./72.*pow(m33,2)*RSe45613p*RS14563pe + 1./144.*
         pow(m33,2)*pow(RSe45613p,2) + 1./72.*pow(m33,2)*RSp45613e*
         RS34561pe + 1./72.*pow(m33,2)*RSp45613e*RS14563pe + 1./72.*
         pow(m33,2)*RSp45613e*RSe45613p + 1./144.*pow(m33,2)*pow(
         RSp45613e,2) + 1./144.*pow(m33,2)*RS3456pee*RS34561pe + 1./144.
         *pow(m33,2)*RS3456pee*RS14563pe + 1./144.*pow(m33,2)*RS3456pee
         *RSe45613p + 1./144.*pow(m33,2)*RS3456pee*RSp45613e + 1./576.*
         pow(m33,2)*pow(RS3456pee,2) + 1./144.*pow(m33,2)*RS3456ppe*
         RS34561pe + 1./144.*pow(m33,2)*RS3456ppe*RS14563pe + 1./144.*
         pow(m33,2)*RS3456ppe*RSe45613p + 1./144.*pow(m33,2)*RS3456ppe*
         RSp45613e + 1./288.*pow(m33,2)*RS3456ppe*RS3456pee + 1./576.*
         pow(m33,2)*pow(RS3456ppe,2) - 1./144.*pow(m33,2)*RSe45633p*
         RS34561pe );

      M13p6x23R +=  + Ab(m13,mu2)*pi16 * (  - 1./144.*pow(m33,2)*
         RSe45633p*RS14563pe - 1./144.*pow(m33,2)*RSe45633p*RSe45613p
          - 1./144.*pow(m33,2)*RSe45633p*RSp45613e - 1./288.*pow(m33,2)
         *RSe45633p*RS3456pee - 1./288.*pow(m33,2)*RSe45633p*RS3456ppe
          + 1./576.*pow(m33,2)*pow(RSe45633p,2) - 1./144.*pow(m33,2)*
         RSp45633e*RS34561pe - 1./144.*pow(m33,2)*RSp45633e*RS14563pe
          - 1./144.*pow(m33,2)*RSp45633e*RSe45613p - 1./144.*pow(m33,2)
         *RSp45633e*RSp45613e - 1./288.*pow(m33,2)*RSp45633e*RS3456pee
          - 1./288.*pow(m33,2)*RSp45633e*RS3456ppe + 1./288.*pow(m33,2)
         *RSp45633e*RSe45633p + 1./576.*pow(m33,2)*pow(RSp45633e,2) - 1.
         /144.*pow(m33,2)*RS345pe*RS34561pe - 1./144.*pow(m33,2)*
         RS345pe*RS14563pe - 1./144.*pow(m33,2)*RS345pe*RSe45613p - 1./
         144.*pow(m33,2)*RS345pe*RSp45613e - 1./288.*pow(m33,2)*RS345pe
         *RS3456pee - 1./288.*pow(m33,2)*RS345pe*RS3456ppe + 1./288.*
         pow(m33,2)*RS345pe*RSe45633p + 1./288.*pow(m33,2)*RS345pe*
         RSp45633e );

      M13p6x23R +=  + Ab(m13,mu2)*pi16 * ( 1./576.*pow(m33,2)*pow(
         RS345pe,2) - 1./144.*pow(m33,2)*RS346pe*RS34561pe - 1./144.*
         pow(m33,2)*RS346pe*RS14563pe - 1./144.*pow(m33,2)*RS346pe*
         RSe45613p - 1./144.*pow(m33,2)*RS346pe*RSp45613e - 1./288.*
         pow(m33,2)*RS346pe*RS3456pee - 1./288.*pow(m33,2)*RS346pe*
         RS3456ppe + 1./288.*pow(m33,2)*RS346pe*RSe45633p + 1./288.*
         pow(m33,2)*RS346pe*RSp45633e + 1./288.*pow(m33,2)*RS346pe*
         RS345pe + 1./576.*pow(m33,2)*pow(RS346pe,2) - 1./144.*pow(
         m33,2)*RS356pe*RS34561pe - 1./144.*pow(m33,2)*RS356pe*
         RS14563pe - 1./144.*pow(m33,2)*RS356pe*RSe45613p - 1./144.*
         pow(m33,2)*RS356pe*RSp45613e - 1./288.*pow(m33,2)*RS356pe*
         RS3456pee - 1./288.*pow(m33,2)*RS356pe*RS3456ppe + 1./288.*
         pow(m33,2)*RS356pe*RSe45633p + 1./288.*pow(m33,2)*RS356pe*
         RSp45633e + 1./288.*pow(m33,2)*RS356pe*RS345pe + 1./288.*pow(
         m33,2)*RS356pe*RS346pe + 1./576.*pow(m33,2)*pow(RS356pe,2) + 1.
         /144.*pow(m33,2)*RS1456pee*RS34561pe );

      M13p6x23R +=  + Ab(m13,mu2)*pi16 * ( 1./144.*pow(m33,2)*RS1456pee
         *RS14563pe + 1./144.*pow(m33,2)*RS1456pee*RSe45613p + 1./144.*
         pow(m33,2)*RS1456pee*RSp45613e + 1./288.*pow(m33,2)*RS1456pee*
         RS3456pee + 1./288.*pow(m33,2)*RS1456pee*RS3456ppe - 1./288.*
         pow(m33,2)*RS1456pee*RSe45633p - 1./288.*pow(m33,2)*RS1456pee*
         RSp45633e - 1./288.*pow(m33,2)*RS1456pee*RS345pe - 1./288.*
         pow(m33,2)*RS1456pee*RS346pe - 1./288.*pow(m33,2)*RS1456pee*
         RS356pe + 1./576.*pow(m33,2)*pow(RS1456pee,2) + 1./144.*pow(
         m33,2)*RS1456ppe*RS34561pe + 1./144.*pow(m33,2)*RS1456ppe*
         RS14563pe + 1./144.*pow(m33,2)*RS1456ppe*RSe45613p + 1./144.*
         pow(m33,2)*RS1456ppe*RSp45613e + 1./288.*pow(m33,2)*RS1456ppe*
         RS3456pee + 1./288.*pow(m33,2)*RS1456ppe*RS3456ppe - 1./288.*
         pow(m33,2)*RS1456ppe*RSe45633p - 1./288.*pow(m33,2)*RS1456ppe*
         RSp45633e - 1./288.*pow(m33,2)*RS1456ppe*RS345pe - 1./288.*
         pow(m33,2)*RS1456ppe*RS346pe - 1./288.*pow(m33,2)*RS1456ppe*
         RS356pe );

      M13p6x23R +=  + Ab(m13,mu2)*pi16 * ( 1./288.*pow(m33,2)*RS1456ppe
         *RS1456pee + 1./576.*pow(m33,2)*pow(RS1456ppe,2) - 1./144.*
         pow(m33,2)*RSe45611p*RS34561pe - 1./144.*pow(m33,2)*RSe45611p*
         RS14563pe - 1./144.*pow(m33,2)*RSe45611p*RSe45613p - 1./144.*
         pow(m33,2)*RSe45611p*RSp45613e - 1./288.*pow(m33,2)*RSe45611p*
         RS3456pee - 1./288.*pow(m33,2)*RSe45611p*RS3456ppe + 1./288.*
         pow(m33,2)*RSe45611p*RSe45633p + 1./288.*pow(m33,2)*RSe45611p*
         RSp45633e + 1./288.*pow(m33,2)*RSe45611p*RS345pe + 1./288.*
         pow(m33,2)*RSe45611p*RS346pe + 1./288.*pow(m33,2)*RSe45611p*
         RS356pe - 1./288.*pow(m33,2)*RSe45611p*RS1456pee - 1./288.*
         pow(m33,2)*RSe45611p*RS1456ppe + 1./576.*pow(m33,2)*pow(
         RSe45611p,2) - 1./144.*pow(m33,2)*RSp45611e*RS34561pe - 1./144.
         *pow(m33,2)*RSp45611e*RS14563pe - 1./144.*pow(m33,2)*RSp45611e
         *RSe45613p - 1./144.*pow(m33,2)*RSp45611e*RSp45613e - 1./288.*
         pow(m33,2)*RSp45611e*RS3456pee - 1./288.*pow(m33,2)*RSp45611e*
         RS3456ppe );

      M13p6x23R +=  + Ab(m13,mu2)*pi16 * ( 1./288.*pow(m33,2)*RSp45611e
         *RSe45633p + 1./288.*pow(m33,2)*RSp45611e*RSp45633e + 1./288.*
         pow(m33,2)*RSp45611e*RS345pe + 1./288.*pow(m33,2)*RSp45611e*
         RS346pe + 1./288.*pow(m33,2)*RSp45611e*RS356pe - 1./288.*pow(
         m33,2)*RSp45611e*RS1456pee - 1./288.*pow(m33,2)*RSp45611e*
         RS1456ppe + 1./288.*pow(m33,2)*RSp45611e*RSe45611p + 1./576.*
         pow(m33,2)*pow(RSp45611e,2) - 1./144.*pow(m33,2)*RS145pe*
         RS34561pe - 1./144.*pow(m33,2)*RS145pe*RS14563pe - 1./144.*
         pow(m33,2)*RS145pe*RSe45613p - 1./144.*pow(m33,2)*RS145pe*
         RSp45613e - 1./288.*pow(m33,2)*RS145pe*RS3456pee - 1./288.*
         pow(m33,2)*RS145pe*RS3456ppe + 1./288.*pow(m33,2)*RS145pe*
         RSe45633p + 1./288.*pow(m33,2)*RS145pe*RSp45633e + 1./288.*
         pow(m33,2)*RS145pe*RS345pe + 1./288.*pow(m33,2)*RS145pe*
         RS346pe + 1./288.*pow(m33,2)*RS145pe*RS356pe - 1./288.*pow(
         m33,2)*RS145pe*RS1456pee - 1./288.*pow(m33,2)*RS145pe*
         RS1456ppe );

      M13p6x23R +=  + Ab(m13,mu2)*pi16 * ( 1./288.*pow(m33,2)*RS145pe*
         RSe45611p + 1./288.*pow(m33,2)*RS145pe*RSp45611e + 1./576.*
         pow(m33,2)*pow(RS145pe,2) - 1./144.*pow(m33,2)*RS146pe*
         RS34561pe - 1./144.*pow(m33,2)*RS146pe*RS14563pe - 1./144.*
         pow(m33,2)*RS146pe*RSe45613p - 1./144.*pow(m33,2)*RS146pe*
         RSp45613e - 1./288.*pow(m33,2)*RS146pe*RS3456pee - 1./288.*
         pow(m33,2)*RS146pe*RS3456ppe + 1./288.*pow(m33,2)*RS146pe*
         RSe45633p + 1./288.*pow(m33,2)*RS146pe*RSp45633e + 1./288.*
         pow(m33,2)*RS146pe*RS345pe + 1./288.*pow(m33,2)*RS146pe*
         RS346pe + 1./288.*pow(m33,2)*RS146pe*RS356pe - 1./288.*pow(
         m33,2)*RS146pe*RS1456pee - 1./288.*pow(m33,2)*RS146pe*
         RS1456ppe + 1./288.*pow(m33,2)*RS146pe*RSe45611p + 1./288.*
         pow(m33,2)*RS146pe*RSp45611e + 1./288.*pow(m33,2)*RS146pe*
         RS145pe + 1./576.*pow(m33,2)*pow(RS146pe,2) - 1./144.*pow(
         m33,2)*RS156pe*RS34561pe - 1./144.*pow(m33,2)*RS156pe*
         RS14563pe );

      M13p6x23R +=  + Ab(m13,mu2)*pi16 * (  - 1./144.*pow(m33,2)*
         RS156pe*RSe45613p - 1./144.*pow(m33,2)*RS156pe*RSp45613e - 1./
         288.*pow(m33,2)*RS156pe*RS3456pee - 1./288.*pow(m33,2)*RS156pe
         *RS3456ppe + 1./288.*pow(m33,2)*RS156pe*RSe45633p + 1./288.*
         pow(m33,2)*RS156pe*RSp45633e + 1./288.*pow(m33,2)*RS156pe*
         RS345pe + 1./288.*pow(m33,2)*RS156pe*RS346pe + 1./288.*pow(
         m33,2)*RS156pe*RS356pe - 1./288.*pow(m33,2)*RS156pe*RS1456pee
          - 1./288.*pow(m33,2)*RS156pe*RS1456ppe + 1./288.*pow(m33,2)*
         RS156pe*RSe45611p + 1./288.*pow(m33,2)*RS156pe*RSp45611e + 1./
         288.*pow(m33,2)*RS156pe*RS145pe + 1./288.*pow(m33,2)*RS156pe*
         RS146pe + 1./576.*pow(m33,2)*pow(RS156pe,2) + 1./72.*m11*m33*
         pow(RS34561pe,2) + 1./36.*m11*m33*RS14563pe*RS34561pe + 1./72.
         *m11*m33*pow(RS14563pe,2) + 1./36.*m11*m33*RSe45613p*RS34561pe
          + 1./36.*m11*m33*RSe45613p*RS14563pe + 1./72.*m11*m33*pow(
         RSe45613p,2) + 1./36.*m11*m33*RSp45613e*RS34561pe + 1./36.*m11
         *m33*RSp45613e*RS14563pe );

      M13p6x23R +=  + Ab(m13,mu2)*pi16 * ( 1./36.*m11*m33*RSp45613e*
         RSe45613p + 1./72.*m11*m33*pow(RSp45613e,2) + 1./72.*m11*m33*
         RS3456pee*RS34561pe + 1./72.*m11*m33*RS3456pee*RS14563pe + 1./
         72.*m11*m33*RS3456pee*RSe45613p + 1./72.*m11*m33*RS3456pee*
         RSp45613e + 1./288.*m11*m33*pow(RS3456pee,2) + 1./72.*m11*m33*
         RS3456ppe*RS34561pe + 1./72.*m11*m33*RS3456ppe*RS14563pe + 1./
         72.*m11*m33*RS3456ppe*RSe45613p + 1./72.*m11*m33*RS3456ppe*
         RSp45613e + 1./144.*m11*m33*RS3456ppe*RS3456pee + 1./288.*m11*
         m33*pow(RS3456ppe,2) - 1./72.*m11*m33*RSe45633p*RS34561pe - 1./
         72.*m11*m33*RSe45633p*RS14563pe - 1./72.*m11*m33*RSe45633p*
         RSe45613p - 1./72.*m11*m33*RSe45633p*RSp45613e - 1./144.*m11*
         m33*RSe45633p*RS3456pee - 1./144.*m11*m33*RSe45633p*RS3456ppe
          + 1./288.*m11*m33*pow(RSe45633p,2) - 1./72.*m11*m33*RSp45633e
         *RS34561pe - 1./72.*m11*m33*RSp45633e*RS14563pe - 1./72.*m11*
         m33*RSp45633e*RSe45613p - 1./72.*m11*m33*RSp45633e*RSp45613e
          - 1./144.*m11*m33*RSp45633e*RS3456pee );

      M13p6x23R +=  + Ab(m13,mu2)*pi16 * (  - 1./144.*m11*m33*RSp45633e
         *RS3456ppe + 1./144.*m11*m33*RSp45633e*RSe45633p + 1./288.*m11
         *m33*pow(RSp45633e,2) - 1./72.*m11*m33*RS345pe*RS34561pe - 1./
         72.*m11*m33*RS345pe*RS14563pe - 1./72.*m11*m33*RS345pe*
         RSe45613p - 1./72.*m11*m33*RS345pe*RSp45613e - 1./144.*m11*m33
         *RS345pe*RS3456pee - 1./144.*m11*m33*RS345pe*RS3456ppe + 1./
         144.*m11*m33*RS345pe*RSe45633p + 1./144.*m11*m33*RS345pe*
         RSp45633e + 1./288.*m11*m33*pow(RS345pe,2) - 1./72.*m11*m33*
         RS346pe*RS34561pe - 1./72.*m11*m33*RS346pe*RS14563pe - 1./72.*
         m11*m33*RS346pe*RSe45613p - 1./72.*m11*m33*RS346pe*RSp45613e
          - 1./144.*m11*m33*RS346pe*RS3456pee - 1./144.*m11*m33*RS346pe
         *RS3456ppe + 1./144.*m11*m33*RS346pe*RSe45633p + 1./144.*m11*
         m33*RS346pe*RSp45633e + 1./144.*m11*m33*RS346pe*RS345pe + 1./
         288.*m11*m33*pow(RS346pe,2) - 1./72.*m11*m33*RS356pe*RS34561pe
          - 1./72.*m11*m33*RS356pe*RS14563pe - 1./72.*m11*m33*RS356pe*
         RSe45613p );

      M13p6x23R +=  + Ab(m13,mu2)*pi16 * (  - 1./72.*m11*m33*RS356pe*
         RSp45613e - 1./144.*m11*m33*RS356pe*RS3456pee - 1./144.*m11*
         m33*RS356pe*RS3456ppe + 1./144.*m11*m33*RS356pe*RSe45633p + 1./
         144.*m11*m33*RS356pe*RSp45633e + 1./144.*m11*m33*RS356pe*
         RS345pe + 1./144.*m11*m33*RS356pe*RS346pe + 1./288.*m11*m33*
         pow(RS356pe,2) + 1./72.*m11*m33*RS1456pee*RS34561pe + 1./72.*
         m11*m33*RS1456pee*RS14563pe + 1./72.*m11*m33*RS1456pee*
         RSe45613p + 1./72.*m11*m33*RS1456pee*RSp45613e + 1./144.*m11*
         m33*RS1456pee*RS3456pee + 1./144.*m11*m33*RS1456pee*RS3456ppe
          - 1./144.*m11*m33*RS1456pee*RSe45633p - 1./144.*m11*m33*
         RS1456pee*RSp45633e - 1./144.*m11*m33*RS1456pee*RS345pe - 1./
         144.*m11*m33*RS1456pee*RS346pe - 1./144.*m11*m33*RS1456pee*
         RS356pe + 1./288.*m11*m33*pow(RS1456pee,2) + 1./72.*m11*m33*
         RS1456ppe*RS34561pe + 1./72.*m11*m33*RS1456ppe*RS14563pe + 1./
         72.*m11*m33*RS1456ppe*RSe45613p + 1./72.*m11*m33*RS1456ppe*
         RSp45613e );

      M13p6x23R +=  + Ab(m13,mu2)*pi16 * ( 1./144.*m11*m33*RS1456ppe*
         RS3456pee + 1./144.*m11*m33*RS1456ppe*RS3456ppe - 1./144.*m11*
         m33*RS1456ppe*RSe45633p - 1./144.*m11*m33*RS1456ppe*RSp45633e
          - 1./144.*m11*m33*RS1456ppe*RS345pe - 1./144.*m11*m33*
         RS1456ppe*RS346pe - 1./144.*m11*m33*RS1456ppe*RS356pe + 1./144.
         *m11*m33*RS1456ppe*RS1456pee + 1./288.*m11*m33*pow(
         RS1456ppe,2) - 1./72.*m11*m33*RSe45611p*RS34561pe - 1./72.*m11
         *m33*RSe45611p*RS14563pe - 1./72.*m11*m33*RSe45611p*RSe45613p
          - 1./72.*m11*m33*RSe45611p*RSp45613e - 1./144.*m11*m33*
         RSe45611p*RS3456pee - 1./144.*m11*m33*RSe45611p*RS3456ppe + 1./
         144.*m11*m33*RSe45611p*RSe45633p + 1./144.*m11*m33*RSe45611p*
         RSp45633e + 1./144.*m11*m33*RSe45611p*RS345pe + 1./144.*m11*
         m33*RSe45611p*RS346pe + 1./144.*m11*m33*RSe45611p*RS356pe - 1./
         144.*m11*m33*RSe45611p*RS1456pee - 1./144.*m11*m33*RSe45611p*
         RS1456ppe + 1./288.*m11*m33*pow(RSe45611p,2) - 1./72.*m11*m33*
         RSp45611e*RS34561pe );

      M13p6x23R +=  + Ab(m13,mu2)*pi16 * (  - 1./72.*m11*m33*RSp45611e*
         RS14563pe - 1./72.*m11*m33*RSp45611e*RSe45613p - 1./72.*m11*
         m33*RSp45611e*RSp45613e - 1./144.*m11*m33*RSp45611e*RS3456pee
          - 1./144.*m11*m33*RSp45611e*RS3456ppe + 1./144.*m11*m33*
         RSp45611e*RSe45633p + 1./144.*m11*m33*RSp45611e*RSp45633e + 1./
         144.*m11*m33*RSp45611e*RS345pe + 1./144.*m11*m33*RSp45611e*
         RS346pe + 1./144.*m11*m33*RSp45611e*RS356pe - 1./144.*m11*m33*
         RSp45611e*RS1456pee - 1./144.*m11*m33*RSp45611e*RS1456ppe + 1./
         144.*m11*m33*RSp45611e*RSe45611p + 1./288.*m11*m33*pow(
         RSp45611e,2) - 1./72.*m11*m33*RS145pe*RS34561pe - 1./72.*m11*
         m33*RS145pe*RS14563pe - 1./72.*m11*m33*RS145pe*RSe45613p - 1./
         72.*m11*m33*RS145pe*RSp45613e - 1./144.*m11*m33*RS145pe*
         RS3456pee - 1./144.*m11*m33*RS145pe*RS3456ppe + 1./144.*m11*
         m33*RS145pe*RSe45633p + 1./144.*m11*m33*RS145pe*RSp45633e + 1./
         144.*m11*m33*RS145pe*RS345pe + 1./144.*m11*m33*RS145pe*RS346pe
          + 1./144.*m11*m33*RS145pe*RS356pe );

      M13p6x23R +=  + Ab(m13,mu2)*pi16 * (  - 1./144.*m11*m33*RS145pe*
         RS1456pee - 1./144.*m11*m33*RS145pe*RS1456ppe + 1./144.*m11*
         m33*RS145pe*RSe45611p + 1./144.*m11*m33*RS145pe*RSp45611e + 1./
         288.*m11*m33*pow(RS145pe,2) - 1./72.*m11*m33*RS146pe*RS34561pe
          - 1./72.*m11*m33*RS146pe*RS14563pe - 1./72.*m11*m33*RS146pe*
         RSe45613p - 1./72.*m11*m33*RS146pe*RSp45613e - 1./144.*m11*m33
         *RS146pe*RS3456pee - 1./144.*m11*m33*RS146pe*RS3456ppe + 1./
         144.*m11*m33*RS146pe*RSe45633p + 1./144.*m11*m33*RS146pe*
         RSp45633e + 1./144.*m11*m33*RS146pe*RS345pe + 1./144.*m11*m33*
         RS146pe*RS346pe + 1./144.*m11*m33*RS146pe*RS356pe - 1./144.*
         m11*m33*RS146pe*RS1456pee - 1./144.*m11*m33*RS146pe*RS1456ppe
          + 1./144.*m11*m33*RS146pe*RSe45611p + 1./144.*m11*m33*RS146pe
         *RSp45611e + 1./144.*m11*m33*RS146pe*RS145pe + 1./288.*m11*m33
         *pow(RS146pe,2) - 1./72.*m11*m33*RS156pe*RS34561pe - 1./72.*
         m11*m33*RS156pe*RS14563pe - 1./72.*m11*m33*RS156pe*RSe45613p
          - 1./72.*m11*m33*RS156pe*RSp45613e );

      M13p6x23R +=  + Ab(m13,mu2)*pi16 * (  - 1./144.*m11*m33*RS156pe*
         RS3456pee - 1./144.*m11*m33*RS156pe*RS3456ppe + 1./144.*m11*
         m33*RS156pe*RSe45633p + 1./144.*m11*m33*RS156pe*RSp45633e + 1./
         144.*m11*m33*RS156pe*RS345pe + 1./144.*m11*m33*RS156pe*RS346pe
          + 1./144.*m11*m33*RS156pe*RS356pe - 1./144.*m11*m33*RS156pe*
         RS1456pee - 1./144.*m11*m33*RS156pe*RS1456ppe + 1./144.*m11*
         m33*RS156pe*RSe45611p + 1./144.*m11*m33*RS156pe*RSp45611e + 1./
         144.*m11*m33*RS156pe*RS145pe + 1./144.*m11*m33*RS156pe*RS146pe
          + 1./288.*m11*m33*pow(RS156pe,2) + 1./144.*pow(m11,2)*pow(
         RS34561pe,2) + 1./72.*pow(m11,2)*RS14563pe*RS34561pe + 1./144.
         *pow(m11,2)*pow(RS14563pe,2) + 1./72.*pow(m11,2)*RSe45613p*
         RS34561pe + 1./72.*pow(m11,2)*RSe45613p*RS14563pe + 1./144.*
         pow(m11,2)*pow(RSe45613p,2) + 1./72.*pow(m11,2)*RSp45613e*
         RS34561pe + 1./72.*pow(m11,2)*RSp45613e*RS14563pe + 1./72.*
         pow(m11,2)*RSp45613e*RSe45613p + 1./144.*pow(m11,2)*pow(
         RSp45613e,2) );

      M13p6x23R +=  + Ab(m13,mu2)*pi16 * ( 1./144.*pow(m11,2)*RS3456pee
         *RS34561pe + 1./144.*pow(m11,2)*RS3456pee*RS14563pe + 1./144.*
         pow(m11,2)*RS3456pee*RSe45613p + 1./144.*pow(m11,2)*RS3456pee*
         RSp45613e + 1./576.*pow(m11,2)*pow(RS3456pee,2) + 1./144.*pow(
         m11,2)*RS3456ppe*RS34561pe + 1./144.*pow(m11,2)*RS3456ppe*
         RS14563pe + 1./144.*pow(m11,2)*RS3456ppe*RSe45613p + 1./144.*
         pow(m11,2)*RS3456ppe*RSp45613e + 1./288.*pow(m11,2)*RS3456ppe*
         RS3456pee + 1./576.*pow(m11,2)*pow(RS3456ppe,2) - 1./144.*pow(
         m11,2)*RSe45633p*RS34561pe - 1./144.*pow(m11,2)*RSe45633p*
         RS14563pe - 1./144.*pow(m11,2)*RSe45633p*RSe45613p - 1./144.*
         pow(m11,2)*RSe45633p*RSp45613e - 1./288.*pow(m11,2)*RSe45633p*
         RS3456pee - 1./288.*pow(m11,2)*RSe45633p*RS3456ppe + 1./576.*
         pow(m11,2)*pow(RSe45633p,2) - 1./144.*pow(m11,2)*RSp45633e*
         RS34561pe - 1./144.*pow(m11,2)*RSp45633e*RS14563pe - 1./144.*
         pow(m11,2)*RSp45633e*RSe45613p - 1./144.*pow(m11,2)*RSp45633e*
         RSp45613e );

      M13p6x23R +=  + Ab(m13,mu2)*pi16 * (  - 1./288.*pow(m11,2)*
         RSp45633e*RS3456pee - 1./288.*pow(m11,2)*RSp45633e*RS3456ppe
          + 1./288.*pow(m11,2)*RSp45633e*RSe45633p + 1./576.*pow(m11,2)
         *pow(RSp45633e,2) - 1./144.*pow(m11,2)*RS345pe*RS34561pe - 1./
         144.*pow(m11,2)*RS345pe*RS14563pe - 1./144.*pow(m11,2)*RS345pe
         *RSe45613p - 1./144.*pow(m11,2)*RS345pe*RSp45613e - 1./288.*
         pow(m11,2)*RS345pe*RS3456pee - 1./288.*pow(m11,2)*RS345pe*
         RS3456ppe + 1./288.*pow(m11,2)*RS345pe*RSe45633p + 1./288.*
         pow(m11,2)*RS345pe*RSp45633e + 1./576.*pow(m11,2)*pow(
         RS345pe,2) - 1./144.*pow(m11,2)*RS346pe*RS34561pe - 1./144.*
         pow(m11,2)*RS346pe*RS14563pe - 1./144.*pow(m11,2)*RS346pe*
         RSe45613p - 1./144.*pow(m11,2)*RS346pe*RSp45613e - 1./288.*
         pow(m11,2)*RS346pe*RS3456pee - 1./288.*pow(m11,2)*RS346pe*
         RS3456ppe + 1./288.*pow(m11,2)*RS346pe*RSe45633p + 1./288.*
         pow(m11,2)*RS346pe*RSp45633e + 1./288.*pow(m11,2)*RS346pe*
         RS345pe );

      M13p6x23R +=  + Ab(m13,mu2)*pi16 * ( 1./576.*pow(m11,2)*pow(
         RS346pe,2) - 1./144.*pow(m11,2)*RS356pe*RS34561pe - 1./144.*
         pow(m11,2)*RS356pe*RS14563pe - 1./144.*pow(m11,2)*RS356pe*
         RSe45613p - 1./144.*pow(m11,2)*RS356pe*RSp45613e - 1./288.*
         pow(m11,2)*RS356pe*RS3456pee - 1./288.*pow(m11,2)*RS356pe*
         RS3456ppe + 1./288.*pow(m11,2)*RS356pe*RSe45633p + 1./288.*
         pow(m11,2)*RS356pe*RSp45633e + 1./288.*pow(m11,2)*RS356pe*
         RS345pe + 1./288.*pow(m11,2)*RS356pe*RS346pe + 1./576.*pow(
         m11,2)*pow(RS356pe,2) + 1./144.*pow(m11,2)*RS1456pee*RS34561pe
          + 1./144.*pow(m11,2)*RS1456pee*RS14563pe + 1./144.*pow(m11,2)
         *RS1456pee*RSe45613p + 1./144.*pow(m11,2)*RS1456pee*RSp45613e
          + 1./288.*pow(m11,2)*RS1456pee*RS3456pee + 1./288.*pow(m11,2)
         *RS1456pee*RS3456ppe - 1./288.*pow(m11,2)*RS1456pee*RSe45633p
          - 1./288.*pow(m11,2)*RS1456pee*RSp45633e - 1./288.*pow(m11,2)
         *RS1456pee*RS345pe - 1./288.*pow(m11,2)*RS1456pee*RS346pe - 1./
         288.*pow(m11,2)*RS1456pee*RS356pe );

      M13p6x23R +=  + Ab(m13,mu2)*pi16 * ( 1./576.*pow(m11,2)*pow(
         RS1456pee,2) + 1./144.*pow(m11,2)*RS1456ppe*RS34561pe + 1./144.
         *pow(m11,2)*RS1456ppe*RS14563pe + 1./144.*pow(m11,2)*RS1456ppe
         *RSe45613p + 1./144.*pow(m11,2)*RS1456ppe*RSp45613e + 1./288.*
         pow(m11,2)*RS1456ppe*RS3456pee + 1./288.*pow(m11,2)*RS1456ppe*
         RS3456ppe - 1./288.*pow(m11,2)*RS1456ppe*RSe45633p - 1./288.*
         pow(m11,2)*RS1456ppe*RSp45633e - 1./288.*pow(m11,2)*RS1456ppe*
         RS345pe - 1./288.*pow(m11,2)*RS1456ppe*RS346pe - 1./288.*pow(
         m11,2)*RS1456ppe*RS356pe + 1./288.*pow(m11,2)*RS1456ppe*
         RS1456pee + 1./576.*pow(m11,2)*pow(RS1456ppe,2) - 1./144.*pow(
         m11,2)*RSe45611p*RS34561pe - 1./144.*pow(m11,2)*RSe45611p*
         RS14563pe - 1./144.*pow(m11,2)*RSe45611p*RSe45613p - 1./144.*
         pow(m11,2)*RSe45611p*RSp45613e - 1./288.*pow(m11,2)*RSe45611p*
         RS3456pee - 1./288.*pow(m11,2)*RSe45611p*RS3456ppe + 1./288.*
         pow(m11,2)*RSe45611p*RSe45633p + 1./288.*pow(m11,2)*RSe45611p*
         RSp45633e );

      M13p6x23R +=  + Ab(m13,mu2)*pi16 * ( 1./288.*pow(m11,2)*RSe45611p
         *RS345pe + 1./288.*pow(m11,2)*RSe45611p*RS346pe + 1./288.*pow(
         m11,2)*RSe45611p*RS356pe - 1./288.*pow(m11,2)*RSe45611p*
         RS1456pee - 1./288.*pow(m11,2)*RSe45611p*RS1456ppe + 1./576.*
         pow(m11,2)*pow(RSe45611p,2) - 1./144.*pow(m11,2)*RSp45611e*
         RS34561pe - 1./144.*pow(m11,2)*RSp45611e*RS14563pe - 1./144.*
         pow(m11,2)*RSp45611e*RSe45613p - 1./144.*pow(m11,2)*RSp45611e*
         RSp45613e - 1./288.*pow(m11,2)*RSp45611e*RS3456pee - 1./288.*
         pow(m11,2)*RSp45611e*RS3456ppe + 1./288.*pow(m11,2)*RSp45611e*
         RSe45633p + 1./288.*pow(m11,2)*RSp45611e*RSp45633e + 1./288.*
         pow(m11,2)*RSp45611e*RS345pe + 1./288.*pow(m11,2)*RSp45611e*
         RS346pe + 1./288.*pow(m11,2)*RSp45611e*RS356pe - 1./288.*pow(
         m11,2)*RSp45611e*RS1456pee - 1./288.*pow(m11,2)*RSp45611e*
         RS1456ppe + 1./288.*pow(m11,2)*RSp45611e*RSe45611p + 1./576.*
         pow(m11,2)*pow(RSp45611e,2) - 1./144.*pow(m11,2)*RS145pe*
         RS34561pe );

      M13p6x23R +=  + Ab(m13,mu2)*pi16 * (  - 1./144.*pow(m11,2)*
         RS145pe*RS14563pe - 1./144.*pow(m11,2)*RS145pe*RSe45613p - 1./
         144.*pow(m11,2)*RS145pe*RSp45613e - 1./288.*pow(m11,2)*RS145pe
         *RS3456pee - 1./288.*pow(m11,2)*RS145pe*RS3456ppe + 1./288.*
         pow(m11,2)*RS145pe*RSe45633p + 1./288.*pow(m11,2)*RS145pe*
         RSp45633e + 1./288.*pow(m11,2)*RS145pe*RS345pe + 1./288.*pow(
         m11,2)*RS145pe*RS346pe + 1./288.*pow(m11,2)*RS145pe*RS356pe - 
         1./288.*pow(m11,2)*RS145pe*RS1456pee - 1./288.*pow(m11,2)*
         RS145pe*RS1456ppe + 1./288.*pow(m11,2)*RS145pe*RSe45611p + 1./
         288.*pow(m11,2)*RS145pe*RSp45611e + 1./576.*pow(m11,2)*pow(
         RS145pe,2) - 1./144.*pow(m11,2)*RS146pe*RS34561pe - 1./144.*
         pow(m11,2)*RS146pe*RS14563pe - 1./144.*pow(m11,2)*RS146pe*
         RSe45613p - 1./144.*pow(m11,2)*RS146pe*RSp45613e - 1./288.*
         pow(m11,2)*RS146pe*RS3456pee - 1./288.*pow(m11,2)*RS146pe*
         RS3456ppe + 1./288.*pow(m11,2)*RS146pe*RSe45633p + 1./288.*
         pow(m11,2)*RS146pe*RSp45633e );

      M13p6x23R +=  + Ab(m13,mu2)*pi16 * ( 1./288.*pow(m11,2)*RS146pe*
         RS345pe + 1./288.*pow(m11,2)*RS146pe*RS346pe + 1./288.*pow(
         m11,2)*RS146pe*RS356pe - 1./288.*pow(m11,2)*RS146pe*RS1456pee
          - 1./288.*pow(m11,2)*RS146pe*RS1456ppe + 1./288.*pow(m11,2)*
         RS146pe*RSe45611p + 1./288.*pow(m11,2)*RS146pe*RSp45611e + 1./
         288.*pow(m11,2)*RS146pe*RS145pe + 1./576.*pow(m11,2)*pow(
         RS146pe,2) - 1./144.*pow(m11,2)*RS156pe*RS34561pe - 1./144.*
         pow(m11,2)*RS156pe*RS14563pe - 1./144.*pow(m11,2)*RS156pe*
         RSe45613p - 1./144.*pow(m11,2)*RS156pe*RSp45613e - 1./288.*
         pow(m11,2)*RS156pe*RS3456pee - 1./288.*pow(m11,2)*RS156pe*
         RS3456ppe + 1./288.*pow(m11,2)*RS156pe*RSe45633p + 1./288.*
         pow(m11,2)*RS156pe*RSp45633e + 1./288.*pow(m11,2)*RS156pe*
         RS345pe + 1./288.*pow(m11,2)*RS156pe*RS346pe + 1./288.*pow(
         m11,2)*RS156pe*RS356pe - 1./288.*pow(m11,2)*RS156pe*RS1456pee
          - 1./288.*pow(m11,2)*RS156pe*RS1456ppe + 1./288.*pow(m11,2)*
         RS156pe*RSe45611p );

      M13p6x23R +=  + Ab(m13,mu2)*pi16 * ( 1./288.*pow(m11,2)*RS156pe*
         RSp45611e + 1./288.*pow(m11,2)*RS156pe*RS145pe + 1./288.*pow(
         m11,2)*RS156pe*RS146pe + 1./576.*pow(m11,2)*pow(RS156pe,2) );

      M13p6x23R +=  + pow(Ab(m13,mu2),2) * (  - 1./8.*m33 - 1./8.*m11 )
         ;

      M13p6x23R +=  + Ab(m13,mu2)*Ab(m33,mu2) * (  - 2./81.*RD3456ep*
         RS34561pe - 1./54.*RD3456ep*RS14563pe - 1./54.*RD3456ep*
         RSe45613p - 1./54.*RD3456ep*RSp45613e - 1./81.*RS3456pee*
         RD3456ep - 1./81.*RS3456ppe*RD3456ep + 1./108.*RSe45633p*
         RD3456ep + 1./108.*RSp45633e*RD3456ep + 1./81.*RS345pe*
         RD3456ep + 1./81.*RS346pe*RD3456ep + 1./81.*RS356pe*RD3456ep
          - 1./162.*RD1456ep*RS34561pe - 1./324.*RD1456ep*RS3456pee - 1.
         /324.*RD1456ep*RS3456ppe + 1./324.*RD1456ep*RS345pe + 1./324.*
         RD1456ep*RS346pe + 1./324.*RD1456ep*RS356pe - 1./108.*
         RS1456pee*RD3456ep - 1./108.*RS1456ppe*RD3456ep + 1./108.*
         RSe45611p*RD3456ep + 1./108.*RSp45611e*RD3456ep + 1./108.*
         RS145pe*RD3456ep + 1./108.*RS146pe*RD3456ep + 1./108.*RS156pe*
         RD3456ep + 1./81.*mp2*RSp45613e*RS34561pe + 1./162.*mp2*
         RS3456pee*RSp45613e + 1./162.*mp2*RS3456ppe*RSp45613e - 1./162.
         *mp2*RSp45633e*RS34561pe - 1./324.*mp2*RSp45633e*RS3456pee - 1.
         /324.*mp2*RSp45633e*RS3456ppe );

      M13p6x23R +=  + Ab(m13,mu2)*Ab(m33,mu2) * (  - 1./162.*mp2*
         RS345pe*RSp45613e + 1./324.*mp2*RS345pe*RSp45633e - 1./162.*
         mp2*RS346pe*RSp45613e + 1./324.*mp2*RS346pe*RSp45633e - 1./162.
         *mp2*RS356pe*RSp45613e + 1./324.*mp2*RS356pe*RSp45633e - 1./
         162.*mp2*RSp45611e*RS34561pe - 1./324.*mp2*RSp45611e*RS3456pee
          - 1./324.*mp2*RSp45611e*RS3456ppe + 1./324.*mp2*RSp45611e*
         RS345pe + 1./324.*mp2*RSp45611e*RS346pe + 1./324.*mp2*
         RSp45611e*RS356pe + 1./81.*me2*RSe45613p*RS34561pe + 1./162.*
         me2*RS3456pee*RSe45613p + 1./162.*me2*RS3456ppe*RSe45613p - 1./
         162.*me2*RSe45633p*RS34561pe - 1./324.*me2*RSe45633p*RS3456pee
          - 1./324.*me2*RSe45633p*RS3456ppe - 1./162.*me2*RS345pe*
         RSe45613p + 1./324.*me2*RS345pe*RSe45633p - 1./162.*me2*
         RS346pe*RSe45613p + 1./324.*me2*RS346pe*RSe45633p - 1./162.*
         me2*RS356pe*RSe45613p + 1./324.*me2*RS356pe*RSe45633p - 1./162.
         *me2*RSe45611p*RS34561pe - 1./324.*me2*RSe45611p*RS3456pee - 1.
         /324.*me2*RSe45611p*RS3456ppe );

      M13p6x23R +=  + Ab(m13,mu2)*Ab(m33,mu2) * ( 1./324.*me2*RSe45611p
         *RS345pe + 1./324.*me2*RSe45611p*RS346pe + 1./324.*me2*
         RSe45611p*RS356pe + 19./72.*m33 - 1./81.*m33*RS14563pe*
         RS34561pe - 1./81.*m33*RSe45613p*RS34561pe - 1./81.*m33*
         RSp45613e*RS34561pe - 1./81.*m33*RS3456pee*RS34561pe - 1./81.*
         m33*RS3456pee*RS14563pe - 1./81.*m33*RS3456pee*RSe45613p - 1./
         81.*m33*RS3456pee*RSp45613e - 5./324.*m33*pow(RS3456pee,2) - 1.
         /81.*m33*RS3456ppe*RS34561pe - 1./81.*m33*RS3456ppe*RS14563pe
          - 1./81.*m33*RS3456ppe*RSe45613p - 1./81.*m33*RS3456ppe*
         RSp45613e - 5./162.*m33*RS3456ppe*RS3456pee - 5./324.*m33*pow(
         RS3456ppe,2) + 1./81.*m33*RSe45633p*RS34561pe + 1./54.*m33*
         RSe45633p*RS3456pee + 1./54.*m33*RSe45633p*RS3456ppe + 1./81.*
         m33*RSp45633e*RS34561pe + 1./54.*m33*RSp45633e*RS3456pee + 1./
         54.*m33*RSp45633e*RS3456ppe + 1./81.*m33*RS345pe*RS34561pe + 1.
         /81.*m33*RS345pe*RS14563pe + 1./81.*m33*RS345pe*RSe45613p + 1./
         81.*m33*RS345pe*RSp45613e );

      M13p6x23R +=  + Ab(m13,mu2)*Ab(m33,mu2) * ( 5./162.*m33*RS345pe*
         RS3456pee + 5./162.*m33*RS345pe*RS3456ppe - 1./54.*m33*RS345pe
         *RSe45633p - 1./54.*m33*RS345pe*RSp45633e - 5./324.*m33*pow(
         RS345pe,2) + 1./81.*m33*RS346pe*RS34561pe + 1./81.*m33*RS346pe
         *RS14563pe + 1./81.*m33*RS346pe*RSe45613p + 1./81.*m33*RS346pe
         *RSp45613e + 5./162.*m33*RS346pe*RS3456pee + 5./162.*m33*
         RS346pe*RS3456ppe - 1./54.*m33*RS346pe*RSe45633p - 1./54.*m33*
         RS346pe*RSp45633e - 5./162.*m33*RS346pe*RS345pe - 5./324.*m33*
         pow(RS346pe,2) + 1./81.*m33*RS356pe*RS34561pe + 1./81.*m33*
         RS356pe*RS14563pe + 1./81.*m33*RS356pe*RSe45613p + 1./81.*m33*
         RS356pe*RSp45613e + 5./162.*m33*RS356pe*RS3456pee + 5./162.*
         m33*RS356pe*RS3456ppe - 1./54.*m33*RS356pe*RSe45633p - 1./54.*
         m33*RS356pe*RSp45633e - 5./162.*m33*RS356pe*RS345pe - 5./162.*
         m33*RS356pe*RS346pe - 5./324.*m33*pow(RS356pe,2) + 1./162.*m33
         *RS1456pee*RS3456pee + 1./162.*m33*RS1456pee*RS3456ppe - 1./
         162.*m33*RS1456pee*RS345pe );

      M13p6x23R +=  + Ab(m13,mu2)*Ab(m33,mu2) * (  - 1./162.*m33*
         RS1456pee*RS346pe - 1./162.*m33*RS1456pee*RS356pe + 1./162.*
         m33*RS1456ppe*RS3456pee + 1./162.*m33*RS1456ppe*RS3456ppe - 1./
         162.*m33*RS1456ppe*RS345pe - 1./162.*m33*RS1456ppe*RS346pe - 1.
         /162.*m33*RS1456ppe*RS356pe - 1./162.*m33*RSe45611p*RS3456pee
          - 1./162.*m33*RSe45611p*RS3456ppe + 1./162.*m33*RSe45611p*
         RS345pe + 1./162.*m33*RSe45611p*RS346pe + 1./162.*m33*
         RSe45611p*RS356pe - 1./162.*m33*RSp45611e*RS3456pee - 1./162.*
         m33*RSp45611e*RS3456ppe + 1./162.*m33*RSp45611e*RS345pe + 1./
         162.*m33*RSp45611e*RS346pe + 1./162.*m33*RSp45611e*RS356pe - 1.
         /162.*m33*RS145pe*RS3456pee - 1./162.*m33*RS145pe*RS3456ppe + 
         1./162.*m33*RS145pe*RS345pe + 1./162.*m33*RS145pe*RS346pe + 1./
         162.*m33*RS145pe*RS356pe - 1./162.*m33*RS146pe*RS3456pee - 1./
         162.*m33*RS146pe*RS3456ppe + 1./162.*m33*RS146pe*RS345pe + 1./
         162.*m33*RS146pe*RS346pe + 1./162.*m33*RS146pe*RS356pe - 1./
         162.*m33*RS156pe*RS3456pee );

      M13p6x23R +=  + Ab(m13,mu2)*Ab(m33,mu2) * (  - 1./162.*m33*
         RS156pe*RS3456ppe + 1./162.*m33*RS156pe*RS345pe + 1./162.*m33*
         RS156pe*RS346pe + 1./162.*m33*RS156pe*RS356pe - 1./9.*pow(
         m33,2)*RN13 + 5./24.*m11 - 4./81.*m11*pow(RS34561pe,2) - 1./27.
         *m11*RS14563pe*RS34561pe - 4./81.*m11*RSe45613p*RS34561pe - 4./
         81.*m11*RSp45613e*RS34561pe - 1./27.*m11*RS3456pee*RS34561pe
          - 1./81.*m11*RS3456pee*RS14563pe - 1./54.*m11*RS3456pee*
         RSe45613p - 1./54.*m11*RS3456pee*RSp45613e - 5./324.*m11*pow(
         RS3456pee,2) - 1./27.*m11*RS3456ppe*RS34561pe - 1./81.*m11*
         RS3456ppe*RS14563pe - 1./54.*m11*RS3456ppe*RSe45613p - 1./54.*
         m11*RS3456ppe*RSp45613e - 5./162.*m11*RS3456ppe*RS3456pee - 5./
         324.*m11*pow(RS3456ppe,2) + 1./54.*m11*RSe45633p*RS34561pe + 5.
         /324.*m11*RSe45633p*RS3456pee + 5./324.*m11*RSe45633p*
         RS3456ppe + 1./54.*m11*RSp45633e*RS34561pe + 5./324.*m11*
         RSp45633e*RS3456pee + 5./324.*m11*RSp45633e*RS3456ppe + 1./27.
         *m11*RS345pe*RS34561pe );

      M13p6x23R +=  + Ab(m13,mu2)*Ab(m33,mu2) * ( 1./81.*m11*RS345pe*
         RS14563pe + 1./54.*m11*RS345pe*RSe45613p + 1./54.*m11*RS345pe*
         RSp45613e + 5./162.*m11*RS345pe*RS3456pee + 5./162.*m11*
         RS345pe*RS3456ppe - 5./324.*m11*RS345pe*RSe45633p - 5./324.*
         m11*RS345pe*RSp45633e - 5./324.*m11*pow(RS345pe,2) + 1./27.*
         m11*RS346pe*RS34561pe + 1./81.*m11*RS346pe*RS14563pe + 1./54.*
         m11*RS346pe*RSe45613p + 1./54.*m11*RS346pe*RSp45613e + 5./162.
         *m11*RS346pe*RS3456pee + 5./162.*m11*RS346pe*RS3456ppe - 5./
         324.*m11*RS346pe*RSe45633p - 5./324.*m11*RS346pe*RSp45633e - 5.
         /162.*m11*RS346pe*RS345pe - 5./324.*m11*pow(RS346pe,2) + 1./27.
         *m11*RS356pe*RS34561pe + 1./81.*m11*RS356pe*RS14563pe + 1./54.
         *m11*RS356pe*RSe45613p + 1./54.*m11*RS356pe*RSp45613e + 5./162.
         *m11*RS356pe*RS3456pee + 5./162.*m11*RS356pe*RS3456ppe - 5./
         324.*m11*RS356pe*RSe45633p - 5./324.*m11*RS356pe*RSp45633e - 5.
         /162.*m11*RS356pe*RS345pe - 5./162.*m11*RS356pe*RS346pe - 5./
         324.*m11*pow(RS356pe,2) );

      M13p6x23R +=  + Ab(m13,mu2)*Ab(m33,mu2) * (  - 2./81.*m11*
         RS1456pee*RS34561pe - 2./81.*m11*RS1456ppe*RS34561pe + 5./162.
         *m11*RSe45611p*RS34561pe + 1./324.*m11*RSe45611p*RS3456pee + 1.
         /324.*m11*RSe45611p*RS3456ppe - 1./324.*m11*RSe45611p*RS345pe
          - 1./324.*m11*RSe45611p*RS346pe - 1./324.*m11*RSe45611p*
         RS356pe + 5./162.*m11*RSp45611e*RS34561pe + 1./324.*m11*
         RSp45611e*RS3456pee + 1./324.*m11*RSp45611e*RS3456ppe - 1./324.
         *m11*RSp45611e*RS345pe - 1./324.*m11*RSp45611e*RS346pe - 1./
         324.*m11*RSp45611e*RS356pe + 2./81.*m11*RS145pe*RS34561pe + 2./
         81.*m11*RS146pe*RS34561pe + 2./81.*m11*RS156pe*RS34561pe - 5./
         9.*m11*m33*RN13 - 1./3.*pow(m11,2)*RN13 );

      M13p6x23R +=  + Ab(m13,mu2)*Ab(me2,mu2) * (  - 1./162.*RD3456ep*
         RSe45613p + 1./324.*RSe45633p*RD3456ep - 1./162.*RD1456ep*
         RSe45613p + 1./324.*RD1456ep*RSe45633p + 1./324.*RSe45611p*
         RD3456ep + 1./324.*RSe45611p*RD1456ep + 1./81.*mp2*RSp45613e*
         RSe45613p - 1./162.*mp2*RSe45633p*RSp45613e - 1./162.*mp2*
         RSp45633e*RSe45613p + 1./324.*mp2*RSp45633e*RSe45633p - 1./162.
         *mp2*RSe45611p*RSp45613e + 1./324.*mp2*RSe45611p*RSp45633e - 1.
         /162.*mp2*RSp45611e*RSe45613p + 1./324.*mp2*RSp45611e*
         RSe45633p + 1./324.*mp2*RSp45611e*RSe45611p + 1./27.*me2*
         RSe45613p*RS34561pe + 1./27.*me2*RSe45613p*RS14563pe + 4./81.*
         me2*pow(RSe45613p,2) + 1./27.*me2*RSp45613e*RSe45613p + 1./54.
         *me2*RS3456pee*RSe45613p + 1./54.*me2*RS3456ppe*RSe45613p - 1./
         54.*me2*RSe45633p*RS34561pe - 1./54.*me2*RSe45633p*RS14563pe
          - 4./81.*me2*RSe45633p*RSe45613p - 1./54.*me2*RSe45633p*
         RSp45613e - 1./108.*me2*RSe45633p*RS3456pee - 1./108.*me2*
         RSe45633p*RS3456ppe );

      M13p6x23R +=  + Ab(m13,mu2)*Ab(me2,mu2) * ( 1./81.*me2*pow(
         RSe45633p,2) - 1./54.*me2*RSp45633e*RSe45613p + 1./108.*me2*
         RSp45633e*RSe45633p - 1./54.*me2*RS345pe*RSe45613p + 1./108.*
         me2*RS345pe*RSe45633p - 1./54.*me2*RS346pe*RSe45613p + 1./108.
         *me2*RS346pe*RSe45633p - 1./54.*me2*RS356pe*RSe45613p + 1./108.
         *me2*RS356pe*RSe45633p + 1./54.*me2*RS1456pee*RSe45613p - 1./
         108.*me2*RS1456pee*RSe45633p + 1./54.*me2*RS1456ppe*RSe45613p
          - 1./108.*me2*RS1456ppe*RSe45633p - 1./54.*me2*RSe45611p*
         RS34561pe - 1./54.*me2*RSe45611p*RS14563pe - 4./81.*me2*
         RSe45611p*RSe45613p - 1./54.*me2*RSe45611p*RSp45613e - 1./108.
         *me2*RSe45611p*RS3456pee - 1./108.*me2*RSe45611p*RS3456ppe + 2.
         /81.*me2*RSe45611p*RSe45633p + 1./108.*me2*RSe45611p*RSp45633e
          + 1./108.*me2*RSe45611p*RS345pe + 1./108.*me2*RSe45611p*
         RS346pe + 1./108.*me2*RSe45611p*RS356pe - 1./108.*me2*
         RSe45611p*RS1456pee - 1./108.*me2*RSe45611p*RS1456ppe + 1./81.
         *me2*pow(RSe45611p,2) );

      M13p6x23R +=  + Ab(m13,mu2)*Ab(me2,mu2) * (  - 1./54.*me2*
         RSp45611e*RSe45613p + 1./108.*me2*RSp45611e*RSe45633p + 1./108.
         *me2*RSp45611e*RSe45611p - 1./54.*me2*RS145pe*RSe45613p + 1./
         108.*me2*RS145pe*RSe45633p + 1./108.*me2*RS145pe*RSe45611p - 1.
         /54.*me2*RS146pe*RSe45613p + 1./108.*me2*RS146pe*RSe45633p + 1.
         /108.*me2*RS146pe*RSe45611p - 1./54.*me2*RS156pe*RSe45613p + 1.
         /108.*me2*RS156pe*RSe45633p + 1./108.*me2*RS156pe*RSe45611p - 
         1./27.*m33*RSe45613p*RS34561pe - 4./81.*m33*RSe45613p*
         RS14563pe - 4./81.*m33*pow(RSe45613p,2) - 4./81.*m33*RSp45613e
         *RSe45613p - 2./81.*m33*RS3456pee*RSe45613p - 2./81.*m33*
         RS3456ppe*RSe45613p + 2./81.*m33*RSe45633p*RS34561pe + 5./162.
         *m33*RSe45633p*RS14563pe + 5./81.*m33*RSe45633p*RSe45613p + 5./
         162.*m33*RSe45633p*RSp45613e + 2./81.*m33*RSe45633p*RS3456pee
          + 2./81.*m33*RSe45633p*RS3456ppe - 1./36.*m33*pow(
         RSe45633p,2) + 5./162.*m33*RSp45633e*RSe45613p - 1./36.*m33*
         RSp45633e*RSe45633p );

      M13p6x23R +=  + Ab(m13,mu2)*Ab(me2,mu2) * ( 2./81.*m33*RS345pe*
         RSe45613p - 2./81.*m33*RS345pe*RSe45633p + 2./81.*m33*RS346pe*
         RSe45613p - 2./81.*m33*RS346pe*RSe45633p + 2./81.*m33*RS356pe*
         RSe45613p - 2./81.*m33*RS356pe*RSe45633p - 1./54.*m33*
         RS1456pee*RSe45613p + 1./324.*m33*RS1456pee*RSe45633p - 1./54.
         *m33*RS1456ppe*RSe45613p + 1./324.*m33*RS1456ppe*RSe45633p + 1.
         /81.*m33*RSe45611p*RS34561pe + 1./54.*m33*RSe45611p*RS14563pe
          + 1./27.*m33*RSe45611p*RSe45613p + 1./54.*m33*RSe45611p*
         RSp45613e - 1./162.*m33*RSe45611p*RSe45633p - 1./324.*m33*
         RSe45611p*RSp45633e + 5./324.*m33*RSe45611p*RS1456pee + 5./324.
         *m33*RSe45611p*RS1456ppe - 5./324.*m33*pow(RSe45611p,2) + 1./
         54.*m33*RSp45611e*RSe45613p - 1./324.*m33*RSp45611e*RSe45633p
          - 5./324.*m33*RSp45611e*RSe45611p + 1./54.*m33*RS145pe*
         RSe45613p - 1./324.*m33*RS145pe*RSe45633p - 5./324.*m33*
         RS145pe*RSe45611p + 1./54.*m33*RS146pe*RSe45613p - 1./324.*m33
         *RS146pe*RSe45633p );

      M13p6x23R +=  + Ab(m13,mu2)*Ab(me2,mu2) * (  - 5./324.*m33*
         RS146pe*RSe45611p + 1./54.*m33*RS156pe*RSe45613p - 1./324.*m33
         *RS156pe*RSe45633p - 5./324.*m33*RS156pe*RSe45611p - 4./81.*
         m11*RSe45613p*RS34561pe - 1./27.*m11*RSe45613p*RS14563pe - 4./
         81.*m11*pow(RSe45613p,2) - 4./81.*m11*RSp45613e*RSe45613p - 1./
         54.*m11*RS3456pee*RSe45613p - 1./54.*m11*RS3456ppe*RSe45613p
          + 1./54.*m11*RSe45633p*RS34561pe + 1./81.*m11*RSe45633p*
         RS14563pe + 1./27.*m11*RSe45633p*RSe45613p + 1./54.*m11*
         RSe45633p*RSp45613e + 5./324.*m11*RSe45633p*RS3456pee + 5./324.
         *m11*RSe45633p*RS3456ppe - 5./324.*m11*pow(RSe45633p,2) + 1./
         54.*m11*RSp45633e*RSe45613p - 5./324.*m11*RSp45633e*RSe45633p
          + 1./54.*m11*RS345pe*RSe45613p - 5./324.*m11*RS345pe*
         RSe45633p + 1./54.*m11*RS346pe*RSe45613p - 5./324.*m11*RS346pe
         *RSe45633p + 1./54.*m11*RS356pe*RSe45613p - 5./324.*m11*
         RS356pe*RSe45633p - 2./81.*m11*RS1456pee*RSe45613p - 2./81.*
         m11*RS1456ppe*RSe45613p );

      M13p6x23R +=  + Ab(m13,mu2)*Ab(me2,mu2) * ( 5./162.*m11*RSe45611p
         *RS34561pe + 2./81.*m11*RSe45611p*RS14563pe + 5./81.*m11*
         RSe45611p*RSe45613p + 5./162.*m11*RSe45611p*RSp45613e + 1./324.
         *m11*RSe45611p*RS3456pee + 1./324.*m11*RSe45611p*RS3456ppe - 1.
         /162.*m11*RSe45611p*RSe45633p - 1./324.*m11*RSe45611p*
         RSp45633e - 1./324.*m11*RSe45611p*RS345pe - 1./324.*m11*
         RSe45611p*RS346pe - 1./324.*m11*RSe45611p*RS356pe + 2./81.*m11
         *RSe45611p*RS1456pee + 2./81.*m11*RSe45611p*RS1456ppe - 1./36.
         *m11*pow(RSe45611p,2) + 5./162.*m11*RSp45611e*RSe45613p - 1./
         324.*m11*RSp45611e*RSe45633p - 1./36.*m11*RSp45611e*RSe45611p
          + 2./81.*m11*RS145pe*RSe45613p - 2./81.*m11*RS145pe*RSe45611p
          + 2./81.*m11*RS146pe*RSe45613p - 2./81.*m11*RS146pe*RSe45611p
          + 2./81.*m11*RS156pe*RSe45613p - 2./81.*m11*RS156pe*RSe45611p
          );

      M13p6x23R +=  + Ab(m13,mu2)*Ab(mp2,mu2) * (  - 1./162.*RD3456ep*
         RSp45613e + 1./324.*RSp45633e*RD3456ep - 1./162.*RD1456ep*
         RSp45613e + 1./324.*RD1456ep*RSp45633e + 1./324.*RSp45611e*
         RD3456ep + 1./324.*RSp45611e*RD1456ep + 1./27.*mp2*RSp45613e*
         RS34561pe + 1./27.*mp2*RSp45613e*RS14563pe + 1./27.*mp2*
         RSp45613e*RSe45613p + 4./81.*mp2*pow(RSp45613e,2) + 1./54.*mp2
         *RS3456pee*RSp45613e + 1./54.*mp2*RS3456ppe*RSp45613e - 1./54.
         *mp2*RSe45633p*RSp45613e - 1./54.*mp2*RSp45633e*RS34561pe - 1./
         54.*mp2*RSp45633e*RS14563pe - 1./54.*mp2*RSp45633e*RSe45613p
          - 4./81.*mp2*RSp45633e*RSp45613e - 1./108.*mp2*RSp45633e*
         RS3456pee - 1./108.*mp2*RSp45633e*RS3456ppe + 1./108.*mp2*
         RSp45633e*RSe45633p + 1./81.*mp2*pow(RSp45633e,2) - 1./54.*mp2
         *RS345pe*RSp45613e + 1./108.*mp2*RS345pe*RSp45633e - 1./54.*
         mp2*RS346pe*RSp45613e + 1./108.*mp2*RS346pe*RSp45633e - 1./54.
         *mp2*RS356pe*RSp45613e + 1./108.*mp2*RS356pe*RSp45633e + 1./54.
         *mp2*RS1456pee*RSp45613e );

      M13p6x23R +=  + Ab(m13,mu2)*Ab(mp2,mu2) * (  - 1./108.*mp2*
         RS1456pee*RSp45633e + 1./54.*mp2*RS1456ppe*RSp45613e - 1./108.
         *mp2*RS1456ppe*RSp45633e - 1./54.*mp2*RSe45611p*RSp45613e + 1./
         108.*mp2*RSe45611p*RSp45633e - 1./54.*mp2*RSp45611e*RS34561pe
          - 1./54.*mp2*RSp45611e*RS14563pe - 1./54.*mp2*RSp45611e*
         RSe45613p - 4./81.*mp2*RSp45611e*RSp45613e - 1./108.*mp2*
         RSp45611e*RS3456pee - 1./108.*mp2*RSp45611e*RS3456ppe + 1./108.
         *mp2*RSp45611e*RSe45633p + 2./81.*mp2*RSp45611e*RSp45633e + 1./
         108.*mp2*RSp45611e*RS345pe + 1./108.*mp2*RSp45611e*RS346pe + 1.
         /108.*mp2*RSp45611e*RS356pe - 1./108.*mp2*RSp45611e*RS1456pee
          - 1./108.*mp2*RSp45611e*RS1456ppe + 1./108.*mp2*RSp45611e*
         RSe45611p + 1./81.*mp2*pow(RSp45611e,2) - 1./54.*mp2*RS145pe*
         RSp45613e + 1./108.*mp2*RS145pe*RSp45633e + 1./108.*mp2*
         RS145pe*RSp45611e - 1./54.*mp2*RS146pe*RSp45613e + 1./108.*mp2
         *RS146pe*RSp45633e + 1./108.*mp2*RS146pe*RSp45611e - 1./54.*
         mp2*RS156pe*RSp45613e );

      M13p6x23R +=  + Ab(m13,mu2)*Ab(mp2,mu2) * ( 1./108.*mp2*RS156pe*
         RSp45633e + 1./108.*mp2*RS156pe*RSp45611e + 1./81.*me2*
         RSp45613e*RSe45613p - 1./162.*me2*RSe45633p*RSp45613e - 1./162.
         *me2*RSp45633e*RSe45613p + 1./324.*me2*RSp45633e*RSe45633p - 1.
         /162.*me2*RSe45611p*RSp45613e + 1./324.*me2*RSe45611p*
         RSp45633e - 1./162.*me2*RSp45611e*RSe45613p + 1./324.*me2*
         RSp45611e*RSe45633p + 1./324.*me2*RSp45611e*RSe45611p - 1./27.
         *m33*RSp45613e*RS34561pe - 4./81.*m33*RSp45613e*RS14563pe - 4./
         81.*m33*RSp45613e*RSe45613p - 4./81.*m33*pow(RSp45613e,2) - 2./
         81.*m33*RS3456pee*RSp45613e - 2./81.*m33*RS3456ppe*RSp45613e
          + 5./162.*m33*RSe45633p*RSp45613e + 2./81.*m33*RSp45633e*
         RS34561pe + 5./162.*m33*RSp45633e*RS14563pe + 5./162.*m33*
         RSp45633e*RSe45613p + 5./81.*m33*RSp45633e*RSp45613e + 2./81.*
         m33*RSp45633e*RS3456pee + 2./81.*m33*RSp45633e*RS3456ppe - 1./
         36.*m33*RSp45633e*RSe45633p - 1./36.*m33*pow(RSp45633e,2) + 2./
         81.*m33*RS345pe*RSp45613e );

      M13p6x23R +=  + Ab(m13,mu2)*Ab(mp2,mu2) * (  - 2./81.*m33*RS345pe
         *RSp45633e + 2./81.*m33*RS346pe*RSp45613e - 2./81.*m33*RS346pe
         *RSp45633e + 2./81.*m33*RS356pe*RSp45613e - 2./81.*m33*RS356pe
         *RSp45633e - 1./54.*m33*RS1456pee*RSp45613e + 1./324.*m33*
         RS1456pee*RSp45633e - 1./54.*m33*RS1456ppe*RSp45613e + 1./324.
         *m33*RS1456ppe*RSp45633e + 1./54.*m33*RSe45611p*RSp45613e - 1./
         324.*m33*RSe45611p*RSp45633e + 1./81.*m33*RSp45611e*RS34561pe
          + 1./54.*m33*RSp45611e*RS14563pe + 1./54.*m33*RSp45611e*
         RSe45613p + 1./27.*m33*RSp45611e*RSp45613e - 1./324.*m33*
         RSp45611e*RSe45633p - 1./162.*m33*RSp45611e*RSp45633e + 5./324.
         *m33*RSp45611e*RS1456pee + 5./324.*m33*RSp45611e*RS1456ppe - 5.
         /324.*m33*RSp45611e*RSe45611p - 5./324.*m33*pow(RSp45611e,2)
          + 1./54.*m33*RS145pe*RSp45613e - 1./324.*m33*RS145pe*
         RSp45633e - 5./324.*m33*RS145pe*RSp45611e + 1./54.*m33*RS146pe
         *RSp45613e - 1./324.*m33*RS146pe*RSp45633e - 5./324.*m33*
         RS146pe*RSp45611e );

      M13p6x23R +=  + Ab(m13,mu2)*Ab(mp2,mu2) * ( 1./54.*m33*RS156pe*
         RSp45613e - 1./324.*m33*RS156pe*RSp45633e - 5./324.*m33*
         RS156pe*RSp45611e - 4./81.*m11*RSp45613e*RS34561pe - 1./27.*
         m11*RSp45613e*RS14563pe - 4./81.*m11*RSp45613e*RSe45613p - 4./
         81.*m11*pow(RSp45613e,2) - 1./54.*m11*RS3456pee*RSp45613e - 1./
         54.*m11*RS3456ppe*RSp45613e + 1./54.*m11*RSe45633p*RSp45613e
          + 1./54.*m11*RSp45633e*RS34561pe + 1./81.*m11*RSp45633e*
         RS14563pe + 1./54.*m11*RSp45633e*RSe45613p + 1./27.*m11*
         RSp45633e*RSp45613e + 5./324.*m11*RSp45633e*RS3456pee + 5./324.
         *m11*RSp45633e*RS3456ppe - 5./324.*m11*RSp45633e*RSe45633p - 5.
         /324.*m11*pow(RSp45633e,2) + 1./54.*m11*RS345pe*RSp45613e - 5./
         324.*m11*RS345pe*RSp45633e + 1./54.*m11*RS346pe*RSp45613e - 5./
         324.*m11*RS346pe*RSp45633e + 1./54.*m11*RS356pe*RSp45613e - 5./
         324.*m11*RS356pe*RSp45633e - 2./81.*m11*RS1456pee*RSp45613e - 
         2./81.*m11*RS1456ppe*RSp45613e + 5./162.*m11*RSe45611p*
         RSp45613e );

      M13p6x23R +=  + Ab(m13,mu2)*Ab(mp2,mu2) * (  - 1./324.*m11*
         RSe45611p*RSp45633e + 5./162.*m11*RSp45611e*RS34561pe + 2./81.
         *m11*RSp45611e*RS14563pe + 5./162.*m11*RSp45611e*RSe45613p + 5.
         /81.*m11*RSp45611e*RSp45613e + 1./324.*m11*RSp45611e*RS3456pee
          + 1./324.*m11*RSp45611e*RS3456ppe - 1./324.*m11*RSp45611e*
         RSe45633p - 1./162.*m11*RSp45611e*RSp45633e - 1./324.*m11*
         RSp45611e*RS345pe - 1./324.*m11*RSp45611e*RS346pe - 1./324.*
         m11*RSp45611e*RS356pe + 2./81.*m11*RSp45611e*RS1456pee + 2./81.
         *m11*RSp45611e*RS1456ppe - 1./36.*m11*RSp45611e*RSe45611p - 1./
         36.*m11*pow(RSp45611e,2) + 2./81.*m11*RS145pe*RSp45613e - 2./
         81.*m11*RS145pe*RSp45611e + 2./81.*m11*RS146pe*RSp45613e - 2./
         81.*m11*RS146pe*RSp45611e + 2./81.*m11*RS156pe*RSp45613e - 2./
         81.*m11*RS156pe*RSp45611e );

      M13p6x23R +=  + Ab(m13,mu2)*Bb(m11,mu2) * ( 1./324.*RD1456ep*
         RD3456ep + 1./324.*pow(RD1456ep,2) - 1./162.*mp2*RD1456ep*
         RSp45613e + 1./324.*mp2*RD1456ep*RSp45633e + 1./324.*mp2*
         RSp45611e*RD1456ep - 1./162.*me2*RD1456ep*RSe45613p + 1./324.*
         me2*RD1456ep*RSe45633p + 1./324.*me2*RSe45611p*RD1456ep + 1./
         81.*m33*RD1456ep*RS34561pe + 1./54.*m33*RD1456ep*RS14563pe + 1.
         /54.*m33*RD1456ep*RSe45613p + 1./54.*m33*RD1456ep*RSp45613e - 
         1./324.*m33*RD1456ep*RSe45633p - 1./324.*m33*RD1456ep*
         RSp45633e + 5./324.*m33*RS1456pee*RD1456ep + 5./324.*m33*
         RS1456ppe*RD1456ep - 5./324.*m33*RSe45611p*RD1456ep - 5./324.*
         m33*RSp45611e*RD1456ep - 5./324.*m33*RS145pe*RD1456ep - 5./324.
         *m33*RS146pe*RD1456ep - 5./324.*m33*RS156pe*RD1456ep + 1./81.*
         m11*RD1456ep*RS34561pe + 1./162.*m11*RD1456ep*RS14563pe + 1./
         81.*m11*RD1456ep*RSe45613p + 1./81.*m11*RD1456ep*RSp45613e - 1.
         /162.*m11*RD1456ep*RS3456pee - 1./162.*m11*RD1456ep*RS3456ppe
          + 1./162.*m11*RD1456ep*RSe45633p );

      M13p6x23R +=  + Ab(m13,mu2)*Bb(m11,mu2) * ( 1./162.*m11*RD1456ep*
         RSp45633e + 1./162.*m11*RD1456ep*RS345pe + 1./162.*m11*
         RD1456ep*RS346pe + 1./162.*m11*RD1456ep*RS356pe + 5./324.*m11*
         RS1456pee*RD1456ep + 5./324.*m11*RS1456ppe*RD1456ep - 1./54.*
         m11*RSe45611p*RD1456ep - 1./54.*m11*RSp45611e*RD1456ep - 5./
         324.*m11*RS145pe*RD1456ep - 5./324.*m11*RS146pe*RD1456ep - 5./
         324.*m11*RS156pe*RD1456ep );

      M13p6x23R +=  + Ab(m13,mu2)*Bb(m33,mu2) * ( 1./324.*pow(
         RD3456ep,2) + 1./324.*RD1456ep*RD3456ep - 1./162.*mp2*RD3456ep
         *RSp45613e + 1./324.*mp2*RSp45633e*RD3456ep + 1./324.*mp2*
         RSp45611e*RD3456ep - 1./162.*me2*RD3456ep*RSe45613p + 1./324.*
         me2*RSe45633p*RD3456ep + 1./324.*me2*RSe45611p*RD3456ep + 1./
         162.*m33*RD3456ep*RS34561pe + 1./81.*m33*RD3456ep*RS14563pe + 
         1./81.*m33*RD3456ep*RSe45613p + 1./81.*m33*RD3456ep*RSp45613e
          + 5./324.*m33*RS3456pee*RD3456ep + 5./324.*m33*RS3456ppe*
         RD3456ep - 1./54.*m33*RSe45633p*RD3456ep - 1./54.*m33*
         RSp45633e*RD3456ep - 5./324.*m33*RS345pe*RD3456ep - 5./324.*
         m33*RS346pe*RD3456ep - 5./324.*m33*RS356pe*RD3456ep - 1./162.*
         m33*RS1456pee*RD3456ep - 1./162.*m33*RS1456ppe*RD3456ep + 1./
         162.*m33*RSe45611p*RD3456ep + 1./162.*m33*RSp45611e*RD3456ep
          + 1./162.*m33*RS145pe*RD3456ep + 1./162.*m33*RS146pe*RD3456ep
          + 1./162.*m33*RS156pe*RD3456ep + 1./54.*m11*RD3456ep*
         RS34561pe );

      M13p6x23R +=  + Ab(m13,mu2)*Bb(m33,mu2) * ( 1./81.*m11*RD3456ep*
         RS14563pe + 1./54.*m11*RD3456ep*RSe45613p + 1./54.*m11*
         RD3456ep*RSp45613e + 5./324.*m11*RS3456pee*RD3456ep + 5./324.*
         m11*RS3456ppe*RD3456ep - 5./324.*m11*RSe45633p*RD3456ep - 5./
         324.*m11*RSp45633e*RD3456ep - 5./324.*m11*RS345pe*RD3456ep - 5.
         /324.*m11*RS346pe*RD3456ep - 5./324.*m11*RS356pe*RD3456ep - 1./
         324.*m11*RSe45611p*RD3456ep - 1./324.*m11*RSp45611e*RD3456ep )
         ;

      M13p6x23R +=  + Ab(m14,mu2)*pi16 * (  - 1./48.*m33*RD3456ep - 1./
         48.*m33*mp2*RSp45633e + 1./24.*m33*mp2*RSp563e - 1./48.*m33*
         mp2*RSp564e - 1./48.*m33*me2*RSe45633p + 1./24.*m33*me2*
         RSe563p - 1./48.*m33*me2*RSe564p + 1./16.*m33*m66 + 1./16.*m33
         *m55 + 1./8.*m33*m44 + 1./96.*m33*m44*RS34561pe + 1./96.*m33*
         m44*RS14563pe + 1./96.*m33*m44*RSe45613p + 1./96.*m33*m44*
         RSp45613e + 1./96.*m33*m44*RS3456pee + 1./96.*m33*m44*
         RS3456ppe - 1./96.*m33*m44*RSe45633p - 1./96.*m33*m44*
         RSp45633e - 1./96.*m33*m44*RS345pe - 1./96.*m33*m44*RS346pe + 
         1./96.*m33*m44*RSe563p + 1./96.*m33*m44*RSp563e + 1./96.*m33*
         m44*RS1456pee + 1./96.*m33*m44*RS1456ppe - 1./96.*m33*m44*
         RSe45611p - 1./96.*m33*m44*RSp45611e - 1./96.*m33*m44*RS145pe
          - 1./96.*m33*m44*RS146pe - 1./32.*m33*m44*RS456pe + 1./96.*
         m33*m44*RSe561p + 1./96.*m33*m44*RSp561e - 1./96.*m33*m44*
         RSe564p - 1./96.*m33*m44*RSp564e + 1./16.*pow(m33,2) + 1./36.*
         pow(m33,2)*RS3456pee );

      M13p6x23R +=  + Ab(m14,mu2)*pi16 * ( 1./36.*pow(m33,2)*RS3456ppe
          - 1./144.*pow(m33,2)*RSe45633p - 1./144.*pow(m33,2)*RSp45633e
          - 1./36.*pow(m33,2)*RS345pe - 1./36.*pow(m33,2)*RS346pe + 1./
         36.*pow(m33,2)*RS356pe + 1./72.*pow(m33,2)*RSe563p + 1./72.*
         pow(m33,2)*RSp563e - 1./144.*pow(m33,2)*RS456pe - 1./144.*pow(
         m33,2)*RSe564p - 1./144.*pow(m33,2)*RSp564e - 1./48.*m11*
         RD3456ep - 1./48.*m11*mp2*RSp45633e + 1./24.*m11*mp2*RSp563e
          - 1./48.*m11*mp2*RSp564e - 1./48.*m11*me2*RSe45633p + 1./24.*
         m11*me2*RSe563p - 1./48.*m11*me2*RSe564p + 1./16.*m11*m66 + 1./
         16.*m11*m55 + 1./8.*m11*m44 + 1./96.*m11*m44*RS34561pe + 1./96.
         *m11*m44*RS14563pe + 1./96.*m11*m44*RSe45613p + 1./96.*m11*m44
         *RSp45613e + 1./96.*m11*m44*RS3456pee + 1./96.*m11*m44*
         RS3456ppe - 1./96.*m11*m44*RSe45633p - 1./96.*m11*m44*
         RSp45633e - 1./96.*m11*m44*RS345pe - 1./96.*m11*m44*RS346pe + 
         1./96.*m11*m44*RSe563p + 1./96.*m11*m44*RSp563e + 1./96.*m11*
         m44*RS1456pee );

      M13p6x23R +=  + Ab(m14,mu2)*pi16 * ( 1./96.*m11*m44*RS1456ppe - 1.
         /96.*m11*m44*RSe45611p - 1./96.*m11*m44*RSp45611e - 1./96.*m11
         *m44*RS145pe - 1./96.*m11*m44*RS146pe - 1./32.*m11*m44*RS456pe
          + 1./96.*m11*m44*RSe561p + 1./96.*m11*m44*RSp561e - 1./96.*
         m11*m44*RSe564p - 1./96.*m11*m44*RSp564e + 1./96.*m11*m33*
         RS34561pe + 1./96.*m11*m33*RS14563pe + 1./96.*m11*m33*
         RSe45613p + 1./96.*m11*m33*RSp45613e + 7./288.*m11*m33*
         RS3456pee + 7./288.*m11*m33*RS3456ppe - 1./288.*m11*m33*
         RSe45633p - 1./288.*m11*m33*RSp45633e - 7./288.*m11*m33*
         RS345pe - 7./288.*m11*m33*RS346pe + 1./72.*m11*m33*RS356pe - 1.
         /288.*m11*m33*RSe563p - 1./288.*m11*m33*RSp563e + 1./96.*m11*
         m33*RS1456pee + 1./96.*m11*m33*RS1456ppe - 1./96.*m11*m33*
         RSe45611p - 1./96.*m11*m33*RSp45611e - 1./96.*m11*m33*RS145pe
          - 1./96.*m11*m33*RS146pe - 1./288.*m11*m33*RS456pe + 1./96.*
         m11*m33*RSe561p + 1./96.*m11*m33*RSp561e - 1./288.*m11*m33*
         RSe564p );

      M13p6x23R +=  + Ab(m14,mu2)*pi16 * (  - 1./288.*m11*m33*RSp564e
          - 1./16.*pow(m11,2) + 1./96.*pow(m11,2)*RS34561pe + 1./96.*
         pow(m11,2)*RS14563pe + 1./96.*pow(m11,2)*RSe45613p + 1./96.*
         pow(m11,2)*RSp45613e - 1./288.*pow(m11,2)*RS3456pee - 1./288.*
         pow(m11,2)*RS3456ppe + 1./288.*pow(m11,2)*RSe45633p + 1./288.*
         pow(m11,2)*RSp45633e + 1./288.*pow(m11,2)*RS345pe + 1./288.*
         pow(m11,2)*RS346pe - 1./72.*pow(m11,2)*RS356pe - 5./288.*pow(
         m11,2)*RSe563p - 5./288.*pow(m11,2)*RSp563e + 1./96.*pow(
         m11,2)*RS1456pee + 1./96.*pow(m11,2)*RS1456ppe - 1./96.*pow(
         m11,2)*RSe45611p - 1./96.*pow(m11,2)*RSp45611e - 1./96.*pow(
         m11,2)*RS145pe - 1./96.*pow(m11,2)*RS146pe + 1./288.*pow(
         m11,2)*RS456pe + 1./96.*pow(m11,2)*RSe561p + 1./96.*pow(m11,2)
         *RSp561e + 1./288.*pow(m11,2)*RSe564p + 1./288.*pow(m11,2)*
         RSp564e );

      M13p6x23R +=  + Ab(m14,mu2)*Ab(m33,mu2) * (  - 1./180.*RD3456ep
          + 7./270.*m44*RS356pe + 43./360.*m33*RS34561pe - 1./24.*m33*
         RS3456pee - 1./24.*m33*RS3456ppe + 1./24.*m33*RS345pe + 1./24.
         *m33*RS346pe - 13./180.*m33*RS356pe + 1./27.*m33*m44*RN13*
         RS356pe - 19./1080.*m11*RS34561pe - 1./24.*m11*RS3456pee - 1./
         24.*m11*RS3456ppe + 1./24.*m11*RS345pe + 1./24.*m11*RS346pe - 
         1./36.*m11*RS356pe + 2./27.*m11*m44*RN13*RS356pe - 1./27.*m11*
         m33*RN13*RS34561pe + 1./9.*m11*m33*RN13*RS356pe + 1./27.*pow(
         m11,2)*RN13*RS34561pe + 1./9.*pow(m11,2)*RN13*RS356pe );

      M13p6x23R +=  + Ab(m14,mu2)*Ab(m34,mu2) * (  - 1./108.*RD3456ep
          - 1./108.*RD1456ep - 1./54.*mp2*RSp45613e - 1./108.*mp2*
         RSp45633e + 1./27.*mp2*RSp563e - 1./108.*mp2*RSp45611e + 1./27.
         *mp2*RSp561e - 1./27.*mp2*RSp564e - 1./54.*me2*RSe45613p - 1./
         108.*me2*RSe45633p + 1./27.*me2*RSe563p - 1./108.*me2*
         RSe45611p + 1./27.*me2*RSe561p - 1./27.*me2*RSe564p - 1./54.*
         m44*RS34561pe - 1./54.*m44*RS14563pe - 1./54.*m44*RSe45613p - 
         1./54.*m44*RSp45613e + 1./108.*m44*RS3456pee + 1./108.*m44*
         RS3456ppe - 1./108.*m44*RSe45633p - 1./108.*m44*RSp45633e - 1./
         108.*m44*RS345pe - 1./108.*m44*RS346pe - 1./108.*m44*RS356pe
          + 1./108.*m44*RS1456pee + 1./108.*m44*RS1456ppe - 1./108.*m44
         *RSe45611p - 1./108.*m44*RSp45611e - 1./108.*m44*RS145pe - 1./
         108.*m44*RS146pe - 1./108.*m44*RS156pe + 1./27.*m44*RSe564p + 
         1./27.*m44*RSp564e - 1./8.*m33 - 13./216.*m33*RS34561pe - 1./
         24.*m33*RS14563pe - 1./24.*m33*RSe45613p - 1./24.*m33*
         RSp45613e );

      M13p6x23R +=  + Ab(m14,mu2)*Ab(m34,mu2) * ( 1./108.*m33*RSe45633p
          + 1./108.*m33*RSp45633e + 5./216.*m33*RS356pe - 1./72.*m33*
         RSe563p - 1./72.*m33*RSp563e + 1./108.*m33*RS1456pee + 1./108.
         *m33*RS1456ppe - 1./108.*m33*RSe45611p - 1./108.*m33*RSp45611e
          - 1./108.*m33*RS145pe - 1./108.*m33*RS146pe + 1./72.*m33*
         RS156pe + 7./216.*m33*RS456pe + 5./216.*m33*RSe561p + 5./216.*
         m33*RSp561e + 7./216.*m33*RSe564p + 7./216.*m33*RSp564e - 1./8.
         *m11 - 1./24.*m11*RS34561pe - 13./216.*m11*RS14563pe - 1./24.*
         m11*RSe45613p - 1./24.*m11*RSp45613e + 1./108.*m11*RS3456pee
          + 1./108.*m11*RS3456ppe - 1./108.*m11*RSe45633p - 1./108.*m11
         *RSp45633e - 1./108.*m11*RS345pe - 1./108.*m11*RS346pe + 1./72.
         *m11*RS356pe + 5./216.*m11*RSe563p + 5./216.*m11*RSp563e + 1./
         108.*m11*RSe45611p + 1./108.*m11*RSp45611e + 5./216.*m11*
         RS156pe + 7./216.*m11*RS456pe - 1./72.*m11*RSe561p - 1./72.*
         m11*RSp561e + 7./216.*m11*RSe564p + 7./216.*m11*RSp564e );

      M13p6x23R +=  + Ab(m14,mu2)*Ab(m35,mu2) * (  - 1./8.*m33 - 1./8.*
         m11 );

      M13p6x23R +=  + Ab(m14,mu2)*Ab(m36,mu2) * (  - 1./8.*m33 - 1./8.*
         m11 );

      M13p6x23R +=  + Ab(m14,mu2)*Ab(me2,mu2) * ( 11./135.*me2*
         RSe45613p - 1./180.*me2*RSe45633p - 19./270.*me2*RSe563p - 29./
         540.*me2*RSe45611p + 7./270.*me2*RSe561p + 1./45.*me2*RSe564p
          + 17./270.*m44*RSe563p - 11./270.*m44*RSe561p - 1./45.*m44*
         RSe564p + 1./27.*m44*me2*RN1e*RSe563p - 1./27.*m44*me2*RN1e*
         RSe561p + 1./1080.*m33*RSe45613p + 17./360.*m33*RSe45633p - 7./
         1080.*m33*RSe563p - 1./24.*m33*RSe561p - 1./27.*m33*me2*RN1e*
         RSe45613p + 1./27.*m33*me2*RN1e*RSe563p + 1./27.*m33*m44*RN1e*
         RSe563p - 19./1080.*m11*RSe45613p + 1./24.*m11*RSe45633p + 1./
         216.*m11*RSe563p + 1./60.*m11*RSe45611p - 49./1080.*m11*
         RSe561p - 1./27.*m11*me2*RN1e*RSe45613p + 1./9.*m11*me2*RN1e*
         RSe563p - 1./27.*m11*me2*RN1e*RSe45611p - 1./27.*m11*me2*RN1e*
         RSe561p + 1./27.*m11*m44*RN1e*RSe563p + 1./27.*m11*m44*RN1e*
         RSe561p + 1./27.*m11*m33*RN1e*RSe45613p + 1./27.*m11*m33*RN1e*
         RSe563p + 1./27.*pow(m11,2)*RN1e*RSe45613p + 1./27.*pow(m11,2)
         *RN1e*RSe563p );

      M13p6x23R +=  + Ab(m14,mu2)*Ab(me2,mu2) * ( 1./27.*pow(m11,2)*
         RN1e*RSe45611p + 1./27.*pow(m11,2)*RN1e*RSe561p );

      M13p6x23R +=  + Ab(m14,mu2)*Ab(mp2,mu2) * ( 11./135.*mp2*
         RSp45613e - 1./180.*mp2*RSp45633e - 19./270.*mp2*RSp563e - 29./
         540.*mp2*RSp45611e + 7./270.*mp2*RSp561e + 1./45.*mp2*RSp564e
          + 17./270.*m44*RSp563e - 11./270.*m44*RSp561e - 1./45.*m44*
         RSp564e + 1./27.*m44*mp2*RN1p*RSp563e - 1./27.*m44*mp2*RN1p*
         RSp561e + 1./1080.*m33*RSp45613e + 17./360.*m33*RSp45633e - 7./
         1080.*m33*RSp563e - 1./24.*m33*RSp561e - 1./27.*m33*mp2*RN1p*
         RSp45613e + 1./27.*m33*mp2*RN1p*RSp563e + 1./27.*m33*m44*RN1p*
         RSp563e - 19./1080.*m11*RSp45613e + 1./24.*m11*RSp45633e + 1./
         216.*m11*RSp563e + 1./60.*m11*RSp45611e - 49./1080.*m11*
         RSp561e - 1./27.*m11*mp2*RN1p*RSp45613e + 1./9.*m11*mp2*RN1p*
         RSp563e - 1./27.*m11*mp2*RN1p*RSp45611e - 1./27.*m11*mp2*RN1p*
         RSp561e + 1./27.*m11*m44*RN1p*RSp563e + 1./27.*m11*m44*RN1p*
         RSp561e + 1./27.*m11*m33*RN1p*RSp45613e + 1./27.*m11*m33*RN1p*
         RSp563e + 1./27.*pow(m11,2)*RN1p*RSp45613e + 1./27.*pow(m11,2)
         *RN1p*RSp563e );

      M13p6x23R +=  + Ab(m14,mu2)*Ab(mp2,mu2) * ( 1./27.*pow(m11,2)*
         RN1p*RSp45611e + 1./27.*pow(m11,2)*RN1p*RSp561e );

      M13p6x23R +=  + Ab(m14,mu2)*Bb(m33,mu2) * ( 1./24.*m33*RD3456ep
          + 1./24.*m11*RD3456ep );

      M13p6x23R +=  + Ab(m15,mu2)*pi16 * (  - 1./48.*m33*RD3456ep - 1./
         48.*m33*mp2*RSp45633e + 1./24.*m33*mp2*RSp463e - 1./48.*m33*
         mp2*RSp465e - 1./48.*m33*me2*RSe45633p + 1./24.*m33*me2*
         RSe463p - 1./48.*m33*me2*RSe465p + 1./16.*m33*m66 + 1./8.*m33*
         m55 + 1./96.*m33*m55*RS34561pe + 1./96.*m33*m55*RS14563pe + 1./
         96.*m33*m55*RSe45613p + 1./96.*m33*m55*RSp45613e + 1./96.*m33*
         m55*RS3456pee + 1./96.*m33*m55*RS3456ppe - 1./96.*m33*m55*
         RSe45633p - 1./96.*m33*m55*RSp45633e - 1./96.*m33*m55*RS345pe
          - 1./96.*m33*m55*RS356pe + 1./96.*m33*m55*RSe463p + 1./96.*
         m33*m55*RSp463e + 1./96.*m33*m55*RS1456pee + 1./96.*m33*m55*
         RS1456ppe - 1./96.*m33*m55*RSe45611p - 1./96.*m33*m55*
         RSp45611e - 1./96.*m33*m55*RS145pe - 1./96.*m33*m55*RS156pe - 
         1./32.*m33*m55*RS546pe + 1./96.*m33*m55*RSe461p + 1./96.*m33*
         m55*RSp461e - 1./96.*m33*m55*RSe465p - 1./96.*m33*m55*RSp465e
          + 1./16.*m33*m44 + 1./16.*pow(m33,2) + 1./36.*pow(m33,2)*
         RS3456pee );

      M13p6x23R +=  + Ab(m15,mu2)*pi16 * ( 1./36.*pow(m33,2)*RS3456ppe
          - 1./144.*pow(m33,2)*RSe45633p - 1./144.*pow(m33,2)*RSp45633e
          - 1./36.*pow(m33,2)*RS345pe + 1./36.*pow(m33,2)*RS346pe - 1./
         36.*pow(m33,2)*RS356pe + 1./72.*pow(m33,2)*RSe463p + 1./72.*
         pow(m33,2)*RSp463e - 1./144.*pow(m33,2)*RS546pe - 1./144.*pow(
         m33,2)*RSe465p - 1./144.*pow(m33,2)*RSp465e - 1./48.*m11*
         RD3456ep - 1./48.*m11*mp2*RSp45633e + 1./24.*m11*mp2*RSp463e
          - 1./48.*m11*mp2*RSp465e - 1./48.*m11*me2*RSe45633p + 1./24.*
         m11*me2*RSe463p - 1./48.*m11*me2*RSe465p + 1./16.*m11*m66 + 1./
         8.*m11*m55 + 1./96.*m11*m55*RS34561pe + 1./96.*m11*m55*
         RS14563pe + 1./96.*m11*m55*RSe45613p + 1./96.*m11*m55*
         RSp45613e + 1./96.*m11*m55*RS3456pee + 1./96.*m11*m55*
         RS3456ppe - 1./96.*m11*m55*RSe45633p - 1./96.*m11*m55*
         RSp45633e - 1./96.*m11*m55*RS345pe - 1./96.*m11*m55*RS356pe + 
         1./96.*m11*m55*RSe463p + 1./96.*m11*m55*RSp463e + 1./96.*m11*
         m55*RS1456pee );

      M13p6x23R +=  + Ab(m15,mu2)*pi16 * ( 1./96.*m11*m55*RS1456ppe - 1.
         /96.*m11*m55*RSe45611p - 1./96.*m11*m55*RSp45611e - 1./96.*m11
         *m55*RS145pe - 1./96.*m11*m55*RS156pe - 1./32.*m11*m55*RS546pe
          + 1./96.*m11*m55*RSe461p + 1./96.*m11*m55*RSp461e - 1./96.*
         m11*m55*RSe465p - 1./96.*m11*m55*RSp465e + 1./16.*m11*m44 + 1./
         96.*m11*m33*RS34561pe + 1./96.*m11*m33*RS14563pe + 1./96.*m11*
         m33*RSe45613p + 1./96.*m11*m33*RSp45613e + 7./288.*m11*m33*
         RS3456pee + 7./288.*m11*m33*RS3456ppe - 1./288.*m11*m33*
         RSe45633p - 1./288.*m11*m33*RSp45633e - 7./288.*m11*m33*
         RS345pe + 1./72.*m11*m33*RS346pe - 7./288.*m11*m33*RS356pe - 1.
         /288.*m11*m33*RSe463p - 1./288.*m11*m33*RSp463e + 1./96.*m11*
         m33*RS1456pee + 1./96.*m11*m33*RS1456ppe - 1./96.*m11*m33*
         RSe45611p - 1./96.*m11*m33*RSp45611e - 1./96.*m11*m33*RS145pe
          - 1./96.*m11*m33*RS156pe - 1./288.*m11*m33*RS546pe + 1./96.*
         m11*m33*RSe461p + 1./96.*m11*m33*RSp461e - 1./288.*m11*m33*
         RSe465p );

      M13p6x23R +=  + Ab(m15,mu2)*pi16 * (  - 1./288.*m11*m33*RSp465e
          - 1./16.*pow(m11,2) + 1./96.*pow(m11,2)*RS34561pe + 1./96.*
         pow(m11,2)*RS14563pe + 1./96.*pow(m11,2)*RSe45613p + 1./96.*
         pow(m11,2)*RSp45613e - 1./288.*pow(m11,2)*RS3456pee - 1./288.*
         pow(m11,2)*RS3456ppe + 1./288.*pow(m11,2)*RSe45633p + 1./288.*
         pow(m11,2)*RSp45633e + 1./288.*pow(m11,2)*RS345pe - 1./72.*
         pow(m11,2)*RS346pe + 1./288.*pow(m11,2)*RS356pe - 5./288.*pow(
         m11,2)*RSe463p - 5./288.*pow(m11,2)*RSp463e + 1./96.*pow(
         m11,2)*RS1456pee + 1./96.*pow(m11,2)*RS1456ppe - 1./96.*pow(
         m11,2)*RSe45611p - 1./96.*pow(m11,2)*RSp45611e - 1./96.*pow(
         m11,2)*RS145pe - 1./96.*pow(m11,2)*RS156pe + 1./288.*pow(
         m11,2)*RS546pe + 1./96.*pow(m11,2)*RSe461p + 1./96.*pow(m11,2)
         *RSp461e + 1./288.*pow(m11,2)*RSe465p + 1./288.*pow(m11,2)*
         RSp465e );

      M13p6x23R +=  + Ab(m15,mu2)*Ab(m33,mu2) * (  - 1./180.*RD3456ep
          + 7./270.*m55*RS346pe + 43./360.*m33*RS34561pe - 1./24.*m33*
         RS3456pee - 1./24.*m33*RS3456ppe + 1./24.*m33*RS345pe - 13./
         180.*m33*RS346pe + 1./24.*m33*RS356pe + 1./27.*m33*m55*RN13*
         RS346pe - 19./1080.*m11*RS34561pe - 1./24.*m11*RS3456pee - 1./
         24.*m11*RS3456ppe + 1./24.*m11*RS345pe - 1./36.*m11*RS346pe + 
         1./24.*m11*RS356pe + 2./27.*m11*m55*RN13*RS346pe - 1./27.*m11*
         m33*RN13*RS34561pe + 1./9.*m11*m33*RN13*RS346pe + 1./27.*pow(
         m11,2)*RN13*RS34561pe + 1./9.*pow(m11,2)*RN13*RS346pe );

      M13p6x23R +=  + Ab(m15,mu2)*Ab(m34,mu2) * (  - 1./8.*m33 - 1./8.*
         m11 );

      M13p6x23R +=  + Ab(m15,mu2)*Ab(m35,mu2) * (  - 1./108.*RD3456ep
          - 1./108.*RD1456ep - 1./54.*mp2*RSp45613e - 1./108.*mp2*
         RSp45633e + 1./27.*mp2*RSp463e - 1./108.*mp2*RSp45611e + 1./27.
         *mp2*RSp461e - 1./27.*mp2*RSp465e - 1./54.*me2*RSe45613p - 1./
         108.*me2*RSe45633p + 1./27.*me2*RSe463p - 1./108.*me2*
         RSe45611p + 1./27.*me2*RSe461p - 1./27.*me2*RSe465p - 1./54.*
         m55*RS34561pe - 1./54.*m55*RS14563pe - 1./54.*m55*RSe45613p - 
         1./54.*m55*RSp45613e + 1./108.*m55*RS3456pee + 1./108.*m55*
         RS3456ppe - 1./108.*m55*RSe45633p - 1./108.*m55*RSp45633e - 1./
         108.*m55*RS345pe - 1./108.*m55*RS346pe - 1./108.*m55*RS356pe
          + 1./108.*m55*RS1456pee + 1./108.*m55*RS1456ppe - 1./108.*m55
         *RSe45611p - 1./108.*m55*RSp45611e - 1./108.*m55*RS145pe - 1./
         108.*m55*RS146pe - 1./108.*m55*RS156pe + 1./27.*m55*RSe465p + 
         1./27.*m55*RSp465e - 1./8.*m33 - 13./216.*m33*RS34561pe - 1./
         24.*m33*RS14563pe - 1./24.*m33*RSe45613p - 1./24.*m33*
         RSp45613e );

      M13p6x23R +=  + Ab(m15,mu2)*Ab(m35,mu2) * ( 1./108.*m33*RSe45633p
          + 1./108.*m33*RSp45633e + 5./216.*m33*RS346pe - 1./72.*m33*
         RSe463p - 1./72.*m33*RSp463e + 1./108.*m33*RS1456pee + 1./108.
         *m33*RS1456ppe - 1./108.*m33*RSe45611p - 1./108.*m33*RSp45611e
          - 1./108.*m33*RS145pe + 1./72.*m33*RS146pe - 1./108.*m33*
         RS156pe + 7./216.*m33*RS546pe + 5./216.*m33*RSe461p + 5./216.*
         m33*RSp461e + 7./216.*m33*RSe465p + 7./216.*m33*RSp465e - 1./8.
         *m11 - 1./24.*m11*RS34561pe - 13./216.*m11*RS14563pe - 1./24.*
         m11*RSe45613p - 1./24.*m11*RSp45613e + 1./108.*m11*RS3456pee
          + 1./108.*m11*RS3456ppe - 1./108.*m11*RSe45633p - 1./108.*m11
         *RSp45633e - 1./108.*m11*RS345pe + 1./72.*m11*RS346pe - 1./108.
         *m11*RS356pe + 5./216.*m11*RSe463p + 5./216.*m11*RSp463e + 1./
         108.*m11*RSe45611p + 1./108.*m11*RSp45611e + 5./216.*m11*
         RS146pe + 7./216.*m11*RS546pe - 1./72.*m11*RSe461p - 1./72.*
         m11*RSp461e + 7./216.*m11*RSe465p + 7./216.*m11*RSp465e );

      M13p6x23R +=  + Ab(m15,mu2)*Ab(m36,mu2) * (  - 1./8.*m33 - 1./8.*
         m11 );

      M13p6x23R +=  + Ab(m15,mu2)*Ab(me2,mu2) * ( 11./135.*me2*
         RSe45613p - 1./180.*me2*RSe45633p - 19./270.*me2*RSe463p - 29./
         540.*me2*RSe45611p + 7./270.*me2*RSe461p + 1./45.*me2*RSe465p
          + 17./270.*m55*RSe463p - 11./270.*m55*RSe461p - 1./45.*m55*
         RSe465p + 1./27.*m55*me2*RN1e*RSe463p - 1./27.*m55*me2*RN1e*
         RSe461p + 1./1080.*m33*RSe45613p + 17./360.*m33*RSe45633p - 7./
         1080.*m33*RSe463p - 1./24.*m33*RSe461p - 1./27.*m33*me2*RN1e*
         RSe45613p + 1./27.*m33*me2*RN1e*RSe463p + 1./27.*m33*m55*RN1e*
         RSe463p - 19./1080.*m11*RSe45613p + 1./24.*m11*RSe45633p + 1./
         216.*m11*RSe463p + 1./60.*m11*RSe45611p - 49./1080.*m11*
         RSe461p - 1./27.*m11*me2*RN1e*RSe45613p + 1./9.*m11*me2*RN1e*
         RSe463p - 1./27.*m11*me2*RN1e*RSe45611p - 1./27.*m11*me2*RN1e*
         RSe461p + 1./27.*m11*m55*RN1e*RSe463p + 1./27.*m11*m55*RN1e*
         RSe461p + 1./27.*m11*m33*RN1e*RSe45613p + 1./27.*m11*m33*RN1e*
         RSe463p + 1./27.*pow(m11,2)*RN1e*RSe45613p + 1./27.*pow(m11,2)
         *RN1e*RSe463p );

      M13p6x23R +=  + Ab(m15,mu2)*Ab(me2,mu2) * ( 1./27.*pow(m11,2)*
         RN1e*RSe45611p + 1./27.*pow(m11,2)*RN1e*RSe461p );

      M13p6x23R +=  + Ab(m15,mu2)*Ab(mp2,mu2) * ( 11./135.*mp2*
         RSp45613e - 1./180.*mp2*RSp45633e - 19./270.*mp2*RSp463e - 29./
         540.*mp2*RSp45611e + 7./270.*mp2*RSp461e + 1./45.*mp2*RSp465e
          + 17./270.*m55*RSp463e - 11./270.*m55*RSp461e - 1./45.*m55*
         RSp465e + 1./27.*m55*mp2*RN1p*RSp463e - 1./27.*m55*mp2*RN1p*
         RSp461e + 1./1080.*m33*RSp45613e + 17./360.*m33*RSp45633e - 7./
         1080.*m33*RSp463e - 1./24.*m33*RSp461e - 1./27.*m33*mp2*RN1p*
         RSp45613e + 1./27.*m33*mp2*RN1p*RSp463e + 1./27.*m33*m55*RN1p*
         RSp463e - 19./1080.*m11*RSp45613e + 1./24.*m11*RSp45633e + 1./
         216.*m11*RSp463e + 1./60.*m11*RSp45611e - 49./1080.*m11*
         RSp461e - 1./27.*m11*mp2*RN1p*RSp45613e + 1./9.*m11*mp2*RN1p*
         RSp463e - 1./27.*m11*mp2*RN1p*RSp45611e - 1./27.*m11*mp2*RN1p*
         RSp461e + 1./27.*m11*m55*RN1p*RSp463e + 1./27.*m11*m55*RN1p*
         RSp461e + 1./27.*m11*m33*RN1p*RSp45613e + 1./27.*m11*m33*RN1p*
         RSp463e + 1./27.*pow(m11,2)*RN1p*RSp45613e + 1./27.*pow(m11,2)
         *RN1p*RSp463e );

      M13p6x23R +=  + Ab(m15,mu2)*Ab(mp2,mu2) * ( 1./27.*pow(m11,2)*
         RN1p*RSp45611e + 1./27.*pow(m11,2)*RN1p*RSp461e );

      M13p6x23R +=  + Ab(m15,mu2)*Bb(m33,mu2) * ( 1./24.*m33*RD3456ep
          + 1./24.*m11*RD3456ep );

      M13p6x23R +=  + Ab(m16,mu2)*pi16 * (  - 1./48.*m33*RD3456ep - 1./
         48.*m33*mp2*RSp45633e + 1./24.*m33*mp2*RSp453e - 1./48.*m33*
         mp2*RSp456e - 1./48.*m33*me2*RSe45633p + 1./24.*m33*me2*
         RSe453p - 1./48.*m33*me2*RSe456p + 1./8.*m33*m66 + 1./96.*m33*
         m66*RS34561pe + 1./96.*m33*m66*RS14563pe + 1./96.*m33*m66*
         RSe45613p + 1./96.*m33*m66*RSp45613e + 1./96.*m33*m66*
         RS3456pee + 1./96.*m33*m66*RS3456ppe - 1./96.*m33*m66*
         RSe45633p - 1./96.*m33*m66*RSp45633e - 1./96.*m33*m66*RS346pe
          - 1./96.*m33*m66*RS356pe + 1./96.*m33*m66*RSe453p + 1./96.*
         m33*m66*RSp453e + 1./96.*m33*m66*RS1456pee + 1./96.*m33*m66*
         RS1456ppe - 1./96.*m33*m66*RSe45611p - 1./96.*m33*m66*
         RSp45611e - 1./96.*m33*m66*RS146pe - 1./96.*m33*m66*RS156pe - 
         1./32.*m33*m66*RS645pe + 1./96.*m33*m66*RSe451p + 1./96.*m33*
         m66*RSp451e - 1./96.*m33*m66*RSe456p - 1./96.*m33*m66*RSp456e
          + 1./16.*m33*m55 + 1./16.*m33*m44 + 1./16.*pow(m33,2) + 1./36.
         *pow(m33,2)*RS3456pee );

      M13p6x23R +=  + Ab(m16,mu2)*pi16 * ( 1./36.*pow(m33,2)*RS3456ppe
          - 1./144.*pow(m33,2)*RSe45633p - 1./144.*pow(m33,2)*RSp45633e
          + 1./36.*pow(m33,2)*RS345pe - 1./36.*pow(m33,2)*RS346pe - 1./
         36.*pow(m33,2)*RS356pe + 1./72.*pow(m33,2)*RSe453p + 1./72.*
         pow(m33,2)*RSp453e - 1./144.*pow(m33,2)*RS645pe - 1./144.*pow(
         m33,2)*RSe456p - 1./144.*pow(m33,2)*RSp456e - 1./48.*m11*
         RD3456ep - 1./48.*m11*mp2*RSp45633e + 1./24.*m11*mp2*RSp453e
          - 1./48.*m11*mp2*RSp456e - 1./48.*m11*me2*RSe45633p + 1./24.*
         m11*me2*RSe453p - 1./48.*m11*me2*RSe456p + 1./8.*m11*m66 + 1./
         96.*m11*m66*RS34561pe + 1./96.*m11*m66*RS14563pe + 1./96.*m11*
         m66*RSe45613p + 1./96.*m11*m66*RSp45613e + 1./96.*m11*m66*
         RS3456pee + 1./96.*m11*m66*RS3456ppe - 1./96.*m11*m66*
         RSe45633p - 1./96.*m11*m66*RSp45633e - 1./96.*m11*m66*RS346pe
          - 1./96.*m11*m66*RS356pe + 1./96.*m11*m66*RSe453p + 1./96.*
         m11*m66*RSp453e + 1./96.*m11*m66*RS1456pee + 1./96.*m11*m66*
         RS1456ppe );

      M13p6x23R +=  + Ab(m16,mu2)*pi16 * (  - 1./96.*m11*m66*RSe45611p
          - 1./96.*m11*m66*RSp45611e - 1./96.*m11*m66*RS146pe - 1./96.*
         m11*m66*RS156pe - 1./32.*m11*m66*RS645pe + 1./96.*m11*m66*
         RSe451p + 1./96.*m11*m66*RSp451e - 1./96.*m11*m66*RSe456p - 1./
         96.*m11*m66*RSp456e + 1./16.*m11*m55 + 1./16.*m11*m44 + 1./96.
         *m11*m33*RS34561pe + 1./96.*m11*m33*RS14563pe + 1./96.*m11*m33
         *RSe45613p + 1./96.*m11*m33*RSp45613e + 7./288.*m11*m33*
         RS3456pee + 7./288.*m11*m33*RS3456ppe - 1./288.*m11*m33*
         RSe45633p - 1./288.*m11*m33*RSp45633e + 1./72.*m11*m33*RS345pe
          - 7./288.*m11*m33*RS346pe - 7./288.*m11*m33*RS356pe - 1./288.
         *m11*m33*RSe453p - 1./288.*m11*m33*RSp453e + 1./96.*m11*m33*
         RS1456pee + 1./96.*m11*m33*RS1456ppe - 1./96.*m11*m33*
         RSe45611p - 1./96.*m11*m33*RSp45611e - 1./96.*m11*m33*RS146pe
          - 1./96.*m11*m33*RS156pe - 1./288.*m11*m33*RS645pe + 1./96.*
         m11*m33*RSe451p + 1./96.*m11*m33*RSp451e - 1./288.*m11*m33*
         RSe456p );

      M13p6x23R +=  + Ab(m16,mu2)*pi16 * (  - 1./288.*m11*m33*RSp456e
          - 1./16.*pow(m11,2) + 1./96.*pow(m11,2)*RS34561pe + 1./96.*
         pow(m11,2)*RS14563pe + 1./96.*pow(m11,2)*RSe45613p + 1./96.*
         pow(m11,2)*RSp45613e - 1./288.*pow(m11,2)*RS3456pee - 1./288.*
         pow(m11,2)*RS3456ppe + 1./288.*pow(m11,2)*RSe45633p + 1./288.*
         pow(m11,2)*RSp45633e - 1./72.*pow(m11,2)*RS345pe + 1./288.*
         pow(m11,2)*RS346pe + 1./288.*pow(m11,2)*RS356pe - 5./288.*pow(
         m11,2)*RSe453p - 5./288.*pow(m11,2)*RSp453e + 1./96.*pow(
         m11,2)*RS1456pee + 1./96.*pow(m11,2)*RS1456ppe - 1./96.*pow(
         m11,2)*RSe45611p - 1./96.*pow(m11,2)*RSp45611e - 1./96.*pow(
         m11,2)*RS146pe - 1./96.*pow(m11,2)*RS156pe + 1./288.*pow(
         m11,2)*RS645pe + 1./96.*pow(m11,2)*RSe451p + 1./96.*pow(m11,2)
         *RSp451e + 1./288.*pow(m11,2)*RSe456p + 1./288.*pow(m11,2)*
         RSp456e );

      M13p6x23R +=  + Ab(m16,mu2)*Ab(m33,mu2) * (  - 1./180.*RD3456ep
          + 7./270.*m66*RS345pe + 43./360.*m33*RS34561pe - 1./24.*m33*
         RS3456pee - 1./24.*m33*RS3456ppe - 13./180.*m33*RS345pe + 1./
         24.*m33*RS346pe + 1./24.*m33*RS356pe + 1./27.*m33*m66*RN13*
         RS345pe - 19./1080.*m11*RS34561pe - 1./24.*m11*RS3456pee - 1./
         24.*m11*RS3456ppe - 1./36.*m11*RS345pe + 1./24.*m11*RS346pe + 
         1./24.*m11*RS356pe + 2./27.*m11*m66*RN13*RS345pe - 1./27.*m11*
         m33*RN13*RS34561pe + 1./9.*m11*m33*RN13*RS345pe + 1./27.*pow(
         m11,2)*RN13*RS34561pe + 1./9.*pow(m11,2)*RN13*RS345pe );

      M13p6x23R +=  + Ab(m16,mu2)*Ab(m34,mu2) * (  - 1./8.*m33 - 1./8.*
         m11 );

      M13p6x23R +=  + Ab(m16,mu2)*Ab(m35,mu2) * (  - 1./8.*m33 - 1./8.*
         m11 );

      M13p6x23R +=  + Ab(m16,mu2)*Ab(m36,mu2) * (  - 1./108.*RD3456ep
          - 1./108.*RD1456ep - 1./54.*mp2*RSp45613e - 1./108.*mp2*
         RSp45633e + 1./27.*mp2*RSp453e - 1./108.*mp2*RSp45611e + 1./27.
         *mp2*RSp451e - 1./27.*mp2*RSp456e - 1./54.*me2*RSe45613p - 1./
         108.*me2*RSe45633p + 1./27.*me2*RSe453p - 1./108.*me2*
         RSe45611p + 1./27.*me2*RSe451p - 1./27.*me2*RSe456p - 1./54.*
         m66*RS34561pe - 1./54.*m66*RS14563pe - 1./54.*m66*RSe45613p - 
         1./54.*m66*RSp45613e + 1./108.*m66*RS3456pee + 1./108.*m66*
         RS3456ppe - 1./108.*m66*RSe45633p - 1./108.*m66*RSp45633e - 1./
         108.*m66*RS345pe - 1./108.*m66*RS346pe - 1./108.*m66*RS356pe
          + 1./108.*m66*RS1456pee + 1./108.*m66*RS1456ppe - 1./108.*m66
         *RSe45611p - 1./108.*m66*RSp45611e - 1./108.*m66*RS145pe - 1./
         108.*m66*RS146pe - 1./108.*m66*RS156pe + 1./27.*m66*RSe456p + 
         1./27.*m66*RSp456e - 1./8.*m33 - 13./216.*m33*RS34561pe - 1./
         24.*m33*RS14563pe - 1./24.*m33*RSe45613p - 1./24.*m33*
         RSp45613e );

      M13p6x23R +=  + Ab(m16,mu2)*Ab(m36,mu2) * ( 1./108.*m33*RSe45633p
          + 1./108.*m33*RSp45633e + 5./216.*m33*RS345pe - 1./72.*m33*
         RSe453p - 1./72.*m33*RSp453e + 1./108.*m33*RS1456pee + 1./108.
         *m33*RS1456ppe - 1./108.*m33*RSe45611p - 1./108.*m33*RSp45611e
          + 1./72.*m33*RS145pe - 1./108.*m33*RS146pe - 1./108.*m33*
         RS156pe + 7./216.*m33*RS645pe + 5./216.*m33*RSe451p + 5./216.*
         m33*RSp451e + 7./216.*m33*RSe456p + 7./216.*m33*RSp456e - 1./8.
         *m11 - 1./24.*m11*RS34561pe - 13./216.*m11*RS14563pe - 1./24.*
         m11*RSe45613p - 1./24.*m11*RSp45613e + 1./108.*m11*RS3456pee
          + 1./108.*m11*RS3456ppe - 1./108.*m11*RSe45633p - 1./108.*m11
         *RSp45633e + 1./72.*m11*RS345pe - 1./108.*m11*RS346pe - 1./108.
         *m11*RS356pe + 5./216.*m11*RSe453p + 5./216.*m11*RSp453e + 1./
         108.*m11*RSe45611p + 1./108.*m11*RSp45611e + 5./216.*m11*
         RS145pe + 7./216.*m11*RS645pe - 1./72.*m11*RSe451p - 1./72.*
         m11*RSp451e + 7./216.*m11*RSe456p + 7./216.*m11*RSp456e );

      M13p6x23R +=  + Ab(m16,mu2)*Ab(me2,mu2) * ( 11./135.*me2*
         RSe45613p - 1./180.*me2*RSe45633p - 19./270.*me2*RSe453p - 29./
         540.*me2*RSe45611p + 7./270.*me2*RSe451p + 1./45.*me2*RSe456p
          + 17./270.*m66*RSe453p - 11./270.*m66*RSe451p - 1./45.*m66*
         RSe456p + 1./27.*m66*me2*RN1e*RSe453p - 1./27.*m66*me2*RN1e*
         RSe451p + 1./1080.*m33*RSe45613p + 17./360.*m33*RSe45633p - 7./
         1080.*m33*RSe453p - 1./24.*m33*RSe451p - 1./27.*m33*me2*RN1e*
         RSe45613p + 1./27.*m33*me2*RN1e*RSe453p + 1./27.*m33*m66*RN1e*
         RSe453p - 19./1080.*m11*RSe45613p + 1./24.*m11*RSe45633p + 1./
         216.*m11*RSe453p + 1./60.*m11*RSe45611p - 49./1080.*m11*
         RSe451p - 1./27.*m11*me2*RN1e*RSe45613p + 1./9.*m11*me2*RN1e*
         RSe453p - 1./27.*m11*me2*RN1e*RSe45611p - 1./27.*m11*me2*RN1e*
         RSe451p + 1./27.*m11*m66*RN1e*RSe453p + 1./27.*m11*m66*RN1e*
         RSe451p + 1./27.*m11*m33*RN1e*RSe45613p + 1./27.*m11*m33*RN1e*
         RSe453p + 1./27.*pow(m11,2)*RN1e*RSe45613p + 1./27.*pow(m11,2)
         *RN1e*RSe453p );

      M13p6x23R +=  + Ab(m16,mu2)*Ab(me2,mu2) * ( 1./27.*pow(m11,2)*
         RN1e*RSe45611p + 1./27.*pow(m11,2)*RN1e*RSe451p );

      M13p6x23R +=  + Ab(m16,mu2)*Ab(mp2,mu2) * ( 11./135.*mp2*
         RSp45613e - 1./180.*mp2*RSp45633e - 19./270.*mp2*RSp453e - 29./
         540.*mp2*RSp45611e + 7./270.*mp2*RSp451e + 1./45.*mp2*RSp456e
          + 17./270.*m66*RSp453e - 11./270.*m66*RSp451e - 1./45.*m66*
         RSp456e + 1./27.*m66*mp2*RN1p*RSp453e - 1./27.*m66*mp2*RN1p*
         RSp451e + 1./1080.*m33*RSp45613e + 17./360.*m33*RSp45633e - 7./
         1080.*m33*RSp453e - 1./24.*m33*RSp451e - 1./27.*m33*mp2*RN1p*
         RSp45613e + 1./27.*m33*mp2*RN1p*RSp453e + 1./27.*m33*m66*RN1p*
         RSp453e - 19./1080.*m11*RSp45613e + 1./24.*m11*RSp45633e + 1./
         216.*m11*RSp453e + 1./60.*m11*RSp45611e - 49./1080.*m11*
         RSp451e - 1./27.*m11*mp2*RN1p*RSp45613e + 1./9.*m11*mp2*RN1p*
         RSp453e - 1./27.*m11*mp2*RN1p*RSp45611e - 1./27.*m11*mp2*RN1p*
         RSp451e + 1./27.*m11*m66*RN1p*RSp453e + 1./27.*m11*m66*RN1p*
         RSp451e + 1./27.*m11*m33*RN1p*RSp45613e + 1./27.*m11*m33*RN1p*
         RSp453e + 1./27.*pow(m11,2)*RN1p*RSp45613e + 1./27.*pow(m11,2)
         *RN1p*RSp453e );

      M13p6x23R +=  + Ab(m16,mu2)*Ab(mp2,mu2) * ( 1./27.*pow(m11,2)*
         RN1p*RSp45611e + 1./27.*pow(m11,2)*RN1p*RSp451e );

      M13p6x23R +=  + Ab(m16,mu2)*Bb(m33,mu2) * ( 1./24.*m33*RD3456ep
          + 1./24.*m11*RD3456ep );

      M13p6x23R +=  + Ab(m33,mu2)*pi16 * ( 1./48.*m33*RD3456ep - 1./72.
         *m33*RD3456ep*RS34561pe - 1./144.*m33*RS3456pee*RD3456ep - 1./
         144.*m33*RS3456ppe*RD3456ep + 1./144.*m33*RS345pe*RD3456ep + 1.
         /144.*m33*RS346pe*RD3456ep + 1./144.*m33*RS356pe*RD3456ep - 1./
         72.*m33*RD1456ep*RS34561pe - 1./144.*m33*RD1456ep*RS3456pee - 
         1./144.*m33*RD1456ep*RS3456ppe + 1./144.*m33*RD1456ep*RS345pe
          + 1./144.*m33*RD1456ep*RS346pe + 1./144.*m33*RD1456ep*RS356pe
          + 1./36.*m33*mp2*RSp45613e*RS34561pe + 1./72.*m33*mp2*
         RS3456pee*RSp45613e + 1./72.*m33*mp2*RS3456ppe*RSp45613e - 1./
         72.*m33*mp2*RSp45633e*RS34561pe - 1./144.*m33*mp2*RSp45633e*
         RS3456pee - 1./144.*m33*mp2*RSp45633e*RS3456ppe - 1./72.*m33*
         mp2*RS345pe*RSp45613e + 1./144.*m33*mp2*RS345pe*RSp45633e - 1./
         72.*m33*mp2*RS346pe*RSp45613e + 1./144.*m33*mp2*RS346pe*
         RSp45633e - 1./72.*m33*mp2*RS356pe*RSp45613e + 1./144.*m33*mp2
         *RS356pe*RSp45633e - 1./72.*m33*mp2*RSp45611e*RS34561pe - 1./
         144.*m33*mp2*RSp45611e*RS3456pee );

      M13p6x23R +=  + Ab(m33,mu2)*pi16 * (  - 1./144.*m33*mp2*RSp45611e
         *RS3456ppe + 1./144.*m33*mp2*RSp45611e*RS345pe + 1./144.*m33*
         mp2*RSp45611e*RS346pe + 1./144.*m33*mp2*RSp45611e*RS356pe + 1./
         36.*m33*me2*RSe45613p*RS34561pe + 1./72.*m33*me2*RS3456pee*
         RSe45613p + 1./72.*m33*me2*RS3456ppe*RSe45613p - 1./72.*m33*
         me2*RSe45633p*RS34561pe - 1./144.*m33*me2*RSe45633p*RS3456pee
          - 1./144.*m33*me2*RSe45633p*RS3456ppe - 1./72.*m33*me2*
         RS345pe*RSe45613p + 1./144.*m33*me2*RS345pe*RSe45633p - 1./72.
         *m33*me2*RS346pe*RSe45613p + 1./144.*m33*me2*RS346pe*RSe45633p
          - 1./72.*m33*me2*RS356pe*RSe45613p + 1./144.*m33*me2*RS356pe*
         RSe45633p - 1./72.*m33*me2*RSe45611p*RS34561pe - 1./144.*m33*
         me2*RSe45611p*RS3456pee - 1./144.*m33*me2*RSe45611p*RS3456ppe
          + 1./144.*m33*me2*RSe45611p*RS345pe + 1./144.*m33*me2*
         RSe45611p*RS346pe + 1./144.*m33*me2*RSe45611p*RS356pe + 1./24.
         *m33*m66*RS34561pe + 1./48.*m33*m66*RS3456pee + 1./48.*m33*m66
         *RS3456ppe );

      M13p6x23R +=  + Ab(m33,mu2)*pi16 * (  - 1./48.*m33*m66*RS345pe - 
         1./48.*m33*m66*RS346pe - 1./48.*m33*m66*RS356pe + 1./24.*m33*
         m55*RS34561pe + 1./48.*m33*m55*RS3456pee + 1./48.*m33*m55*
         RS3456ppe - 1./48.*m33*m55*RS345pe - 1./48.*m33*m55*RS346pe - 
         1./48.*m33*m55*RS356pe + 1./24.*m33*m44*RS34561pe + 1./48.*m33
         *m44*RS3456pee + 1./48.*m33*m44*RS3456ppe - 1./48.*m33*m44*
         RS345pe - 1./48.*m33*m44*RS346pe - 1./48.*m33*m44*RS356pe - 1./
         16.*pow(m33,2) + 1./27.*pow(m33,2)*pow(RS34561pe,2) + 1./108.*
         pow(m33,2)*RS14563pe*RS34561pe + 1./108.*pow(m33,2)*RSe45613p*
         RS34561pe + 1./108.*pow(m33,2)*RSp45613e*RS34561pe + 1./27.*
         pow(m33,2)*RS3456pee*RS34561pe + 1./216.*pow(m33,2)*RS3456pee*
         RS14563pe + 1./216.*pow(m33,2)*RS3456pee*RSe45613p + 1./216.*
         pow(m33,2)*RS3456pee*RSp45613e + 1./108.*pow(m33,2)*pow(
         RS3456pee,2) + 1./27.*pow(m33,2)*RS3456ppe*RS34561pe + 1./216.
         *pow(m33,2)*RS3456ppe*RS14563pe + 1./216.*pow(m33,2)*RS3456ppe
         *RSe45613p );

      M13p6x23R +=  + Ab(m33,mu2)*pi16 * ( 1./216.*pow(m33,2)*RS3456ppe
         *RSp45613e + 1./54.*pow(m33,2)*RS3456ppe*RS3456pee + 1./108.*
         pow(m33,2)*pow(RS3456ppe,2) - 1./216.*pow(m33,2)*RSe45633p*
         RS34561pe - 1./432.*pow(m33,2)*RSe45633p*RS3456pee - 1./432.*
         pow(m33,2)*RSe45633p*RS3456ppe - 1./216.*pow(m33,2)*RSp45633e*
         RS34561pe - 1./432.*pow(m33,2)*RSp45633e*RS3456pee - 1./432.*
         pow(m33,2)*RSp45633e*RS3456ppe + 1./48.*pow(m33,2)*RS345pe - 1.
         /27.*pow(m33,2)*RS345pe*RS34561pe - 1./216.*pow(m33,2)*RS345pe
         *RS14563pe - 1./216.*pow(m33,2)*RS345pe*RSe45613p - 1./216.*
         pow(m33,2)*RS345pe*RSp45613e - 1./54.*pow(m33,2)*RS345pe*
         RS3456pee - 1./54.*pow(m33,2)*RS345pe*RS3456ppe + 1./432.*pow(
         m33,2)*RS345pe*RSe45633p + 1./432.*pow(m33,2)*RS345pe*
         RSp45633e + 1./108.*pow(m33,2)*pow(RS345pe,2) + 1./48.*pow(
         m33,2)*RS346pe - 1./27.*pow(m33,2)*RS346pe*RS34561pe - 1./216.
         *pow(m33,2)*RS346pe*RS14563pe - 1./216.*pow(m33,2)*RS346pe*
         RSe45613p );

      M13p6x23R +=  + Ab(m33,mu2)*pi16 * (  - 1./216.*pow(m33,2)*
         RS346pe*RSp45613e - 1./54.*pow(m33,2)*RS346pe*RS3456pee - 1./
         54.*pow(m33,2)*RS346pe*RS3456ppe + 1./432.*pow(m33,2)*RS346pe*
         RSe45633p + 1./432.*pow(m33,2)*RS346pe*RSp45633e + 1./54.*pow(
         m33,2)*RS346pe*RS345pe + 1./108.*pow(m33,2)*pow(RS346pe,2) + 1.
         /48.*pow(m33,2)*RS356pe - 1./27.*pow(m33,2)*RS356pe*RS34561pe
          - 1./216.*pow(m33,2)*RS356pe*RS14563pe - 1./216.*pow(m33,2)*
         RS356pe*RSe45613p - 1./216.*pow(m33,2)*RS356pe*RSp45613e - 1./
         54.*pow(m33,2)*RS356pe*RS3456pee - 1./54.*pow(m33,2)*RS356pe*
         RS3456ppe + 1./432.*pow(m33,2)*RS356pe*RSe45633p + 1./432.*
         pow(m33,2)*RS356pe*RSp45633e + 1./54.*pow(m33,2)*RS356pe*
         RS345pe + 1./54.*pow(m33,2)*RS356pe*RS346pe + 1./108.*pow(
         m33,2)*pow(RS356pe,2) + 1./216.*pow(m33,2)*RS1456pee*RS34561pe
          + 1./432.*pow(m33,2)*RS1456pee*RS3456pee + 1./432.*pow(m33,2)
         *RS1456pee*RS3456ppe - 1./432.*pow(m33,2)*RS1456pee*RS345pe - 
         1./432.*pow(m33,2)*RS1456pee*RS346pe );

      M13p6x23R +=  + Ab(m33,mu2)*pi16 * (  - 1./432.*pow(m33,2)*
         RS1456pee*RS356pe + 1./216.*pow(m33,2)*RS1456ppe*RS34561pe + 1.
         /432.*pow(m33,2)*RS1456ppe*RS3456pee + 1./432.*pow(m33,2)*
         RS1456ppe*RS3456ppe - 1./432.*pow(m33,2)*RS1456ppe*RS345pe - 1.
         /432.*pow(m33,2)*RS1456ppe*RS346pe - 1./432.*pow(m33,2)*
         RS1456ppe*RS356pe - 1./216.*pow(m33,2)*RSe45611p*RS34561pe - 1.
         /432.*pow(m33,2)*RSe45611p*RS3456pee - 1./432.*pow(m33,2)*
         RSe45611p*RS3456ppe + 1./432.*pow(m33,2)*RSe45611p*RS345pe + 1.
         /432.*pow(m33,2)*RSe45611p*RS346pe + 1./432.*pow(m33,2)*
         RSe45611p*RS356pe - 1./216.*pow(m33,2)*RSp45611e*RS34561pe - 1.
         /432.*pow(m33,2)*RSp45611e*RS3456pee - 1./432.*pow(m33,2)*
         RSp45611e*RS3456ppe + 1./432.*pow(m33,2)*RSp45611e*RS345pe + 1.
         /432.*pow(m33,2)*RSp45611e*RS346pe + 1./432.*pow(m33,2)*
         RSp45611e*RS356pe - 1./216.*pow(m33,2)*RS145pe*RS34561pe - 1./
         432.*pow(m33,2)*RS145pe*RS3456pee - 1./432.*pow(m33,2)*RS145pe
         *RS3456ppe );

      M13p6x23R +=  + Ab(m33,mu2)*pi16 * ( 1./432.*pow(m33,2)*RS145pe*
         RS345pe + 1./432.*pow(m33,2)*RS145pe*RS346pe + 1./432.*pow(
         m33,2)*RS145pe*RS356pe - 1./216.*pow(m33,2)*RS146pe*RS34561pe
          - 1./432.*pow(m33,2)*RS146pe*RS3456pee - 1./432.*pow(m33,2)*
         RS146pe*RS3456ppe + 1./432.*pow(m33,2)*RS146pe*RS345pe + 1./
         432.*pow(m33,2)*RS146pe*RS346pe + 1./432.*pow(m33,2)*RS146pe*
         RS356pe - 1./216.*pow(m33,2)*RS156pe*RS34561pe - 1./432.*pow(
         m33,2)*RS156pe*RS3456pee - 1./432.*pow(m33,2)*RS156pe*
         RS3456ppe + 1./432.*pow(m33,2)*RS156pe*RS345pe + 1./432.*pow(
         m33,2)*RS156pe*RS346pe + 1./432.*pow(m33,2)*RS156pe*RS356pe + 
         1./48.*m11*RD3456ep - 1./72.*m11*RD3456ep*RS34561pe - 1./144.*
         m11*RS3456pee*RD3456ep - 1./144.*m11*RS3456ppe*RD3456ep + 1./
         144.*m11*RS345pe*RD3456ep + 1./144.*m11*RS346pe*RD3456ep + 1./
         144.*m11*RS356pe*RD3456ep - 1./72.*m11*RD1456ep*RS34561pe - 1./
         144.*m11*RD1456ep*RS3456pee - 1./144.*m11*RD1456ep*RS3456ppe
          + 1./144.*m11*RD1456ep*RS345pe );

      M13p6x23R +=  + Ab(m33,mu2)*pi16 * ( 1./144.*m11*RD1456ep*RS346pe
          + 1./144.*m11*RD1456ep*RS356pe + 1./36.*m11*mp2*RSp45613e*
         RS34561pe + 1./72.*m11*mp2*RS3456pee*RSp45613e + 1./72.*m11*
         mp2*RS3456ppe*RSp45613e - 1./72.*m11*mp2*RSp45633e*RS34561pe
          - 1./144.*m11*mp2*RSp45633e*RS3456pee - 1./144.*m11*mp2*
         RSp45633e*RS3456ppe - 1./72.*m11*mp2*RS345pe*RSp45613e + 1./
         144.*m11*mp2*RS345pe*RSp45633e - 1./72.*m11*mp2*RS346pe*
         RSp45613e + 1./144.*m11*mp2*RS346pe*RSp45633e - 1./72.*m11*mp2
         *RS356pe*RSp45613e + 1./144.*m11*mp2*RS356pe*RSp45633e - 1./72.
         *m11*mp2*RSp45611e*RS34561pe - 1./144.*m11*mp2*RSp45611e*
         RS3456pee - 1./144.*m11*mp2*RSp45611e*RS3456ppe + 1./144.*m11*
         mp2*RSp45611e*RS345pe + 1./144.*m11*mp2*RSp45611e*RS346pe + 1./
         144.*m11*mp2*RSp45611e*RS356pe + 1./36.*m11*me2*RSe45613p*
         RS34561pe + 1./72.*m11*me2*RS3456pee*RSe45613p + 1./72.*m11*
         me2*RS3456ppe*RSe45613p - 1./72.*m11*me2*RSe45633p*RS34561pe
          - 1./144.*m11*me2*RSe45633p*RS3456pee );

      M13p6x23R +=  + Ab(m33,mu2)*pi16 * (  - 1./144.*m11*me2*RSe45633p
         *RS3456ppe - 1./72.*m11*me2*RS345pe*RSe45613p + 1./144.*m11*
         me2*RS345pe*RSe45633p - 1./72.*m11*me2*RS346pe*RSe45613p + 1./
         144.*m11*me2*RS346pe*RSe45633p - 1./72.*m11*me2*RS356pe*
         RSe45613p + 1./144.*m11*me2*RS356pe*RSe45633p - 1./72.*m11*me2
         *RSe45611p*RS34561pe - 1./144.*m11*me2*RSe45611p*RS3456pee - 1.
         /144.*m11*me2*RSe45611p*RS3456ppe + 1./144.*m11*me2*RSe45611p*
         RS345pe + 1./144.*m11*me2*RSe45611p*RS346pe + 1./144.*m11*me2*
         RSe45611p*RS356pe + 1./24.*m11*m66*RS34561pe + 1./48.*m11*m66*
         RS3456pee + 1./48.*m11*m66*RS3456ppe - 1./48.*m11*m66*RS345pe
          - 1./48.*m11*m66*RS346pe - 1./48.*m11*m66*RS356pe + 1./24.*
         m11*m55*RS34561pe + 1./48.*m11*m55*RS3456pee + 1./48.*m11*m55*
         RS3456ppe - 1./48.*m11*m55*RS345pe - 1./48.*m11*m55*RS346pe - 
         1./48.*m11*m55*RS356pe + 1./24.*m11*m44*RS34561pe + 1./48.*m11
         *m44*RS3456pee + 1./48.*m11*m44*RS3456ppe - 1./48.*m11*m44*
         RS345pe );

      M13p6x23R +=  + Ab(m33,mu2)*pi16 * (  - 1./48.*m11*m44*RS346pe - 
         1./48.*m11*m44*RS356pe - 1./16.*m11*m33 + 1./24.*m11*m33*
         RS34561pe + 5./108.*m11*m33*pow(RS34561pe,2) + 5./108.*m11*m33
         *RS14563pe*RS34561pe + 1./54.*m11*m33*RSe45613p*RS34561pe + 1./
         54.*m11*m33*RSp45613e*RS34561pe + 1./48.*m11*m33*RS3456pee + 5.
         /108.*m11*m33*RS3456pee*RS34561pe + 5./216.*m11*m33*RS3456pee*
         RS14563pe + 1./108.*m11*m33*RS3456pee*RSe45613p + 1./108.*m11*
         m33*RS3456pee*RSp45613e + 5./432.*m11*m33*pow(RS3456pee,2) + 1.
         /48.*m11*m33*RS3456ppe + 5./108.*m11*m33*RS3456ppe*RS34561pe
          + 5./216.*m11*m33*RS3456ppe*RS14563pe + 1./108.*m11*m33*
         RS3456ppe*RSe45613p + 1./108.*m11*m33*RS3456ppe*RSp45613e + 5./
         216.*m11*m33*RS3456ppe*RS3456pee + 5./432.*m11*m33*pow(
         RS3456ppe,2) - 1./108.*m11*m33*RSe45633p*RS34561pe - 1./216.*
         m11*m33*RSe45633p*RS3456pee - 1./216.*m11*m33*RSe45633p*
         RS3456ppe - 1./108.*m11*m33*RSp45633e*RS34561pe - 1./216.*m11*
         m33*RSp45633e*RS3456pee );

      M13p6x23R +=  + Ab(m33,mu2)*pi16 * (  - 1./216.*m11*m33*RSp45633e
         *RS3456ppe - 5./108.*m11*m33*RS345pe*RS34561pe - 5./216.*m11*
         m33*RS345pe*RS14563pe - 1./108.*m11*m33*RS345pe*RSe45613p - 1./
         108.*m11*m33*RS345pe*RSp45613e - 5./216.*m11*m33*RS345pe*
         RS3456pee - 5./216.*m11*m33*RS345pe*RS3456ppe + 1./216.*m11*
         m33*RS345pe*RSe45633p + 1./216.*m11*m33*RS345pe*RSp45633e + 5./
         432.*m11*m33*pow(RS345pe,2) - 5./108.*m11*m33*RS346pe*
         RS34561pe - 5./216.*m11*m33*RS346pe*RS14563pe - 1./108.*m11*
         m33*RS346pe*RSe45613p - 1./108.*m11*m33*RS346pe*RSp45613e - 5./
         216.*m11*m33*RS346pe*RS3456pee - 5./216.*m11*m33*RS346pe*
         RS3456ppe + 1./216.*m11*m33*RS346pe*RSe45633p + 1./216.*m11*
         m33*RS346pe*RSp45633e + 5./216.*m11*m33*RS346pe*RS345pe + 5./
         432.*m11*m33*pow(RS346pe,2) - 5./108.*m11*m33*RS356pe*
         RS34561pe - 5./216.*m11*m33*RS356pe*RS14563pe - 1./108.*m11*
         m33*RS356pe*RSe45613p - 1./108.*m11*m33*RS356pe*RSp45613e - 5./
         216.*m11*m33*RS356pe*RS3456pee );

      M13p6x23R +=  + Ab(m33,mu2)*pi16 * (  - 5./216.*m11*m33*RS356pe*
         RS3456ppe + 1./216.*m11*m33*RS356pe*RSe45633p + 1./216.*m11*
         m33*RS356pe*RSp45633e + 5./216.*m11*m33*RS356pe*RS345pe + 5./
         216.*m11*m33*RS356pe*RS346pe + 5./432.*m11*m33*pow(RS356pe,2)
          + 5./216.*m11*m33*RS1456pee*RS34561pe + 5./432.*m11*m33*
         RS1456pee*RS3456pee + 5./432.*m11*m33*RS1456pee*RS3456ppe - 5./
         432.*m11*m33*RS1456pee*RS345pe - 5./432.*m11*m33*RS1456pee*
         RS346pe - 5./432.*m11*m33*RS1456pee*RS356pe + 5./216.*m11*m33*
         RS1456ppe*RS34561pe + 5./432.*m11*m33*RS1456ppe*RS3456pee + 5./
         432.*m11*m33*RS1456ppe*RS3456ppe - 5./432.*m11*m33*RS1456ppe*
         RS345pe - 5./432.*m11*m33*RS1456ppe*RS346pe - 5./432.*m11*m33*
         RS1456ppe*RS356pe - 1./108.*m11*m33*RSe45611p*RS34561pe - 1./
         216.*m11*m33*RSe45611p*RS3456pee - 1./216.*m11*m33*RSe45611p*
         RS3456ppe + 1./216.*m11*m33*RSe45611p*RS345pe + 1./216.*m11*
         m33*RSe45611p*RS346pe + 1./216.*m11*m33*RSe45611p*RS356pe - 1./
         108.*m11*m33*RSp45611e*RS34561pe );

      M13p6x23R +=  + Ab(m33,mu2)*pi16 * (  - 1./216.*m11*m33*RSp45611e
         *RS3456pee - 1./216.*m11*m33*RSp45611e*RS3456ppe + 1./216.*m11
         *m33*RSp45611e*RS345pe + 1./216.*m11*m33*RSp45611e*RS346pe + 1.
         /216.*m11*m33*RSp45611e*RS356pe - 5./216.*m11*m33*RS145pe*
         RS34561pe - 5./432.*m11*m33*RS145pe*RS3456pee - 5./432.*m11*
         m33*RS145pe*RS3456ppe + 5./432.*m11*m33*RS145pe*RS345pe + 5./
         432.*m11*m33*RS145pe*RS346pe + 5./432.*m11*m33*RS145pe*RS356pe
          - 5./216.*m11*m33*RS146pe*RS34561pe - 5./432.*m11*m33*RS146pe
         *RS3456pee - 5./432.*m11*m33*RS146pe*RS3456ppe + 5./432.*m11*
         m33*RS146pe*RS345pe + 5./432.*m11*m33*RS146pe*RS346pe + 5./432.
         *m11*m33*RS146pe*RS356pe - 5./216.*m11*m33*RS156pe*RS34561pe
          - 5./432.*m11*m33*RS156pe*RS3456pee - 5./432.*m11*m33*RS156pe
         *RS3456ppe + 5./432.*m11*m33*RS156pe*RS345pe + 5./432.*m11*m33
         *RS156pe*RS346pe + 5./432.*m11*m33*RS156pe*RS356pe + 1./24.*
         pow(m11,2)*RS34561pe + 1./108.*pow(m11,2)*pow(RS34561pe,2) + 1.
         /27.*pow(m11,2)*RS14563pe*RS34561pe );

      M13p6x23R +=  + Ab(m33,mu2)*pi16 * ( 1./108.*pow(m11,2)*RSe45613p
         *RS34561pe + 1./108.*pow(m11,2)*RSp45613e*RS34561pe + 1./48.*
         pow(m11,2)*RS3456pee + 1./108.*pow(m11,2)*RS3456pee*RS34561pe
          + 1./54.*pow(m11,2)*RS3456pee*RS14563pe + 1./216.*pow(m11,2)*
         RS3456pee*RSe45613p + 1./216.*pow(m11,2)*RS3456pee*RSp45613e
          + 1./432.*pow(m11,2)*pow(RS3456pee,2) + 1./48.*pow(m11,2)*
         RS3456ppe + 1./108.*pow(m11,2)*RS3456ppe*RS34561pe + 1./54.*
         pow(m11,2)*RS3456ppe*RS14563pe + 1./216.*pow(m11,2)*RS3456ppe*
         RSe45613p + 1./216.*pow(m11,2)*RS3456ppe*RSp45613e + 1./216.*
         pow(m11,2)*RS3456ppe*RS3456pee + 1./432.*pow(m11,2)*pow(
         RS3456ppe,2) - 1./216.*pow(m11,2)*RSe45633p*RS34561pe - 1./432.
         *pow(m11,2)*RSe45633p*RS3456pee - 1./432.*pow(m11,2)*RSe45633p
         *RS3456ppe - 1./216.*pow(m11,2)*RSp45633e*RS34561pe - 1./432.*
         pow(m11,2)*RSp45633e*RS3456pee - 1./432.*pow(m11,2)*RSp45633e*
         RS3456ppe - 1./48.*pow(m11,2)*RS345pe - 1./108.*pow(m11,2)*
         RS345pe*RS34561pe );

      M13p6x23R +=  + Ab(m33,mu2)*pi16 * (  - 1./54.*pow(m11,2)*RS345pe
         *RS14563pe - 1./216.*pow(m11,2)*RS345pe*RSe45613p - 1./216.*
         pow(m11,2)*RS345pe*RSp45613e - 1./216.*pow(m11,2)*RS345pe*
         RS3456pee - 1./216.*pow(m11,2)*RS345pe*RS3456ppe + 1./432.*
         pow(m11,2)*RS345pe*RSe45633p + 1./432.*pow(m11,2)*RS345pe*
         RSp45633e + 1./432.*pow(m11,2)*pow(RS345pe,2) - 1./48.*pow(
         m11,2)*RS346pe - 1./108.*pow(m11,2)*RS346pe*RS34561pe - 1./54.
         *pow(m11,2)*RS346pe*RS14563pe - 1./216.*pow(m11,2)*RS346pe*
         RSe45613p - 1./216.*pow(m11,2)*RS346pe*RSp45613e - 1./216.*
         pow(m11,2)*RS346pe*RS3456pee - 1./216.*pow(m11,2)*RS346pe*
         RS3456ppe + 1./432.*pow(m11,2)*RS346pe*RSe45633p + 1./432.*
         pow(m11,2)*RS346pe*RSp45633e + 1./216.*pow(m11,2)*RS346pe*
         RS345pe + 1./432.*pow(m11,2)*pow(RS346pe,2) - 1./48.*pow(
         m11,2)*RS356pe - 1./108.*pow(m11,2)*RS356pe*RS34561pe - 1./54.
         *pow(m11,2)*RS356pe*RS14563pe - 1./216.*pow(m11,2)*RS356pe*
         RSe45613p );

      M13p6x23R +=  + Ab(m33,mu2)*pi16 * (  - 1./216.*pow(m11,2)*
         RS356pe*RSp45613e - 1./216.*pow(m11,2)*RS356pe*RS3456pee - 1./
         216.*pow(m11,2)*RS356pe*RS3456ppe + 1./432.*pow(m11,2)*RS356pe
         *RSe45633p + 1./432.*pow(m11,2)*RS356pe*RSp45633e + 1./216.*
         pow(m11,2)*RS356pe*RS345pe + 1./216.*pow(m11,2)*RS356pe*
         RS346pe + 1./432.*pow(m11,2)*pow(RS356pe,2) + 1./54.*pow(
         m11,2)*RS1456pee*RS34561pe + 1./108.*pow(m11,2)*RS1456pee*
         RS3456pee + 1./108.*pow(m11,2)*RS1456pee*RS3456ppe - 1./108.*
         pow(m11,2)*RS1456pee*RS345pe - 1./108.*pow(m11,2)*RS1456pee*
         RS346pe - 1./108.*pow(m11,2)*RS1456pee*RS356pe + 1./54.*pow(
         m11,2)*RS1456ppe*RS34561pe + 1./108.*pow(m11,2)*RS1456ppe*
         RS3456pee + 1./108.*pow(m11,2)*RS1456ppe*RS3456ppe - 1./108.*
         pow(m11,2)*RS1456ppe*RS345pe - 1./108.*pow(m11,2)*RS1456ppe*
         RS346pe - 1./108.*pow(m11,2)*RS1456ppe*RS356pe - 1./216.*pow(
         m11,2)*RSe45611p*RS34561pe - 1./432.*pow(m11,2)*RSe45611p*
         RS3456pee );

      M13p6x23R +=  + Ab(m33,mu2)*pi16 * (  - 1./432.*pow(m11,2)*
         RSe45611p*RS3456ppe + 1./432.*pow(m11,2)*RSe45611p*RS345pe + 1.
         /432.*pow(m11,2)*RSe45611p*RS346pe + 1./432.*pow(m11,2)*
         RSe45611p*RS356pe - 1./216.*pow(m11,2)*RSp45611e*RS34561pe - 1.
         /432.*pow(m11,2)*RSp45611e*RS3456pee - 1./432.*pow(m11,2)*
         RSp45611e*RS3456ppe + 1./432.*pow(m11,2)*RSp45611e*RS345pe + 1.
         /432.*pow(m11,2)*RSp45611e*RS346pe + 1./432.*pow(m11,2)*
         RSp45611e*RS356pe - 1./54.*pow(m11,2)*RS145pe*RS34561pe - 1./
         108.*pow(m11,2)*RS145pe*RS3456pee - 1./108.*pow(m11,2)*RS145pe
         *RS3456ppe + 1./108.*pow(m11,2)*RS145pe*RS345pe + 1./108.*pow(
         m11,2)*RS145pe*RS346pe + 1./108.*pow(m11,2)*RS145pe*RS356pe - 
         1./54.*pow(m11,2)*RS146pe*RS34561pe - 1./108.*pow(m11,2)*
         RS146pe*RS3456pee - 1./108.*pow(m11,2)*RS146pe*RS3456ppe + 1./
         108.*pow(m11,2)*RS146pe*RS345pe + 1./108.*pow(m11,2)*RS146pe*
         RS346pe + 1./108.*pow(m11,2)*RS146pe*RS356pe - 1./54.*pow(
         m11,2)*RS156pe*RS34561pe );

      M13p6x23R +=  + Ab(m33,mu2)*pi16 * (  - 1./108.*pow(m11,2)*
         RS156pe*RS3456pee - 1./108.*pow(m11,2)*RS156pe*RS3456ppe + 1./
         108.*pow(m11,2)*RS156pe*RS345pe + 1./108.*pow(m11,2)*RS156pe*
         RS346pe + 1./108.*pow(m11,2)*RS156pe*RS356pe );

      M13p6x23R +=  + pow(Ab(m33,mu2),2) * (  - 1./270.*RD3456ep*
         RS34561pe - 1./540.*RS3456pee*RD3456ep - 1./540.*RS3456ppe*
         RD3456ep + 1./540.*RS345pe*RD3456ep + 1./540.*RS346pe*RD3456ep
          + 1./540.*RS356pe*RD3456ep + 1./16.*m33 + 17./540.*m33*pow(
         RS34561pe,2) - 1./60.*m33*RS3456pee*RS34561pe - 1./27.*m33*
         RS3456pee*RS14563pe - 1./27.*m33*RS3456pee*RSe45613p - 1./27.*
         m33*RS3456pee*RSp45613e - 5./432.*m33*pow(RS3456pee,2) - 1./60.
         *m33*RS3456ppe*RS34561pe - 1./27.*m33*RS3456ppe*RS14563pe - 1./
         27.*m33*RS3456ppe*RSe45613p - 1./27.*m33*RS3456ppe*RSp45613e
          - 5./216.*m33*RS3456ppe*RS3456pee - 5./432.*m33*pow(
         RS3456ppe,2) + 1./27.*m33*RSe45633p*RS3456pee + 1./27.*m33*
         RSe45633p*RS3456ppe + 1./27.*m33*RSp45633e*RS3456pee + 1./27.*
         m33*RSp45633e*RS3456ppe + 1./60.*m33*RS345pe*RS34561pe + 1./27.
         *m33*RS345pe*RS14563pe + 1./27.*m33*RS345pe*RSe45613p + 1./27.
         *m33*RS345pe*RSp45613e + 5./216.*m33*RS345pe*RS3456pee + 5./
         216.*m33*RS345pe*RS3456ppe );

      M13p6x23R +=  + pow(Ab(m33,mu2),2) * (  - 1./27.*m33*RS345pe*
         RSe45633p - 1./27.*m33*RS345pe*RSp45633e - 5./432.*m33*pow(
         RS345pe,2) + 1./60.*m33*RS346pe*RS34561pe + 1./27.*m33*RS346pe
         *RS14563pe + 1./27.*m33*RS346pe*RSe45613p + 1./27.*m33*RS346pe
         *RSp45613e + 5./216.*m33*RS346pe*RS3456pee + 5./216.*m33*
         RS346pe*RS3456ppe - 1./27.*m33*RS346pe*RSe45633p - 1./27.*m33*
         RS346pe*RSp45633e - 5./216.*m33*RS346pe*RS345pe - 5./432.*m33*
         pow(RS346pe,2) + 1./60.*m33*RS356pe*RS34561pe + 1./27.*m33*
         RS356pe*RS14563pe + 1./27.*m33*RS356pe*RSe45613p + 1./27.*m33*
         RS356pe*RSp45613e + 5./216.*m33*RS356pe*RS3456pee + 5./216.*
         m33*RS356pe*RS3456ppe - 1./27.*m33*RS356pe*RSe45633p - 1./27.*
         m33*RS356pe*RSp45633e - 5./216.*m33*RS356pe*RS345pe - 5./216.*
         m33*RS356pe*RS346pe - 5./432.*m33*pow(RS356pe,2) - 1./27.*m33*
         mp2*RN3p*RS3456pee*RSp45613e - 1./27.*m33*mp2*RN3p*RS3456ppe*
         RSp45613e + 1./27.*m33*mp2*RN3p*RSp45633e*RS3456pee + 1./27.*
         m33*mp2*RN3p*RSp45633e*RS3456ppe );

      M13p6x23R +=  + pow(Ab(m33,mu2),2) * ( 1./27.*m33*mp2*RN3p*
         RS345pe*RSp45613e - 1./27.*m33*mp2*RN3p*RS345pe*RSp45633e + 1./
         27.*m33*mp2*RN3p*RS346pe*RSp45613e - 1./27.*m33*mp2*RN3p*
         RS346pe*RSp45633e + 1./27.*m33*mp2*RN3p*RS356pe*RSp45613e - 1./
         27.*m33*mp2*RN3p*RS356pe*RSp45633e - 1./27.*m33*me2*RN3e*
         RS3456pee*RSe45613p - 1./27.*m33*me2*RN3e*RS3456ppe*RSe45613p
          + 1./27.*m33*me2*RN3e*RSe45633p*RS3456pee + 1./27.*m33*me2*
         RN3e*RSe45633p*RS3456ppe + 1./27.*m33*me2*RN3e*RS345pe*
         RSe45613p - 1./27.*m33*me2*RN3e*RS345pe*RSe45633p + 1./27.*m33
         *me2*RN3e*RS346pe*RSe45613p - 1./27.*m33*me2*RN3e*RS346pe*
         RSe45633p + 1./27.*m33*me2*RN3e*RS356pe*RSe45613p - 1./27.*m33
         *me2*RN3e*RS356pe*RSe45633p - 1./27.*pow(m33,2)*RN3p*RS3456pee
         *RSp45613e - 1./27.*pow(m33,2)*RN3p*RS3456ppe*RSp45613e - 1./
         27.*pow(m33,2)*RN3p*RSp45633e*RS3456pee - 1./27.*pow(m33,2)*
         RN3p*RSp45633e*RS3456ppe + 1./27.*pow(m33,2)*RN3p*RS345pe*
         RSp45613e );

      M13p6x23R +=  + pow(Ab(m33,mu2),2) * ( 1./27.*pow(m33,2)*RN3p*
         RS345pe*RSp45633e + 1./27.*pow(m33,2)*RN3p*RS346pe*RSp45613e
          + 1./27.*pow(m33,2)*RN3p*RS346pe*RSp45633e + 1./27.*pow(
         m33,2)*RN3p*RS356pe*RSp45613e + 1./27.*pow(m33,2)*RN3p*RS356pe
         *RSp45633e - 1./27.*pow(m33,2)*RN3e*RS3456pee*RSe45613p - 1./
         27.*pow(m33,2)*RN3e*RS3456ppe*RSe45613p - 1./27.*pow(m33,2)*
         RN3e*RSe45633p*RS3456pee - 1./27.*pow(m33,2)*RN3e*RSe45633p*
         RS3456ppe + 1./27.*pow(m33,2)*RN3e*RS345pe*RSe45613p + 1./27.*
         pow(m33,2)*RN3e*RS345pe*RSe45633p + 1./27.*pow(m33,2)*RN3e*
         RS346pe*RSe45613p + 1./27.*pow(m33,2)*RN3e*RS346pe*RSe45633p
          + 1./27.*pow(m33,2)*RN3e*RS356pe*RSe45613p + 1./27.*pow(
         m33,2)*RN3e*RS356pe*RSe45633p + 1./27.*pow(m33,2)*RN13*
         RS3456pee*RS14563pe + 1./27.*pow(m33,2)*RN13*RS3456ppe*
         RS14563pe - 1./27.*pow(m33,2)*RN13*RS345pe*RS14563pe - 1./27.*
         pow(m33,2)*RN13*RS346pe*RS14563pe - 1./27.*pow(m33,2)*RN13*
         RS356pe*RS14563pe );

      M13p6x23R +=  + pow(Ab(m33,mu2),2) * ( 1./16.*m11 + 13./540.*m11*
         pow(RS34561pe,2) - 1./540.*m11*RS3456pee*RS34561pe + 1./144.*
         m11*pow(RS3456pee,2) - 1./540.*m11*RS3456ppe*RS34561pe + 1./72.
         *m11*RS3456ppe*RS3456pee + 1./144.*m11*pow(RS3456ppe,2) + 1./
         540.*m11*RS345pe*RS34561pe - 1./72.*m11*RS345pe*RS3456pee - 1./
         72.*m11*RS345pe*RS3456ppe + 1./144.*m11*pow(RS345pe,2) + 1./
         540.*m11*RS346pe*RS34561pe - 1./72.*m11*RS346pe*RS3456pee - 1./
         72.*m11*RS346pe*RS3456ppe + 1./72.*m11*RS346pe*RS345pe + 1./
         144.*m11*pow(RS346pe,2) + 1./540.*m11*RS356pe*RS34561pe - 1./
         72.*m11*RS356pe*RS3456pee - 1./72.*m11*RS356pe*RS3456ppe + 1./
         72.*m11*RS356pe*RS345pe + 1./72.*m11*RS356pe*RS346pe + 1./144.
         *m11*pow(RS356pe,2) - 1./27.*m11*m33*RN3p*RS3456pee*RSp45613e
          - 1./27.*m11*m33*RN3p*RS3456ppe*RSp45613e + 1./27.*m11*m33*
         RN3p*RS345pe*RSp45613e + 1./27.*m11*m33*RN3p*RS346pe*RSp45613e
          + 1./27.*m11*m33*RN3p*RS356pe*RSp45613e - 1./27.*m11*m33*RN3e
         *RS3456pee*RSe45613p );

      M13p6x23R +=  + pow(Ab(m33,mu2),2) * (  - 1./27.*m11*m33*RN3e*
         RS3456ppe*RSe45613p + 1./27.*m11*m33*RN3e*RS345pe*RSe45613p + 
         1./27.*m11*m33*RN3e*RS346pe*RSe45613p + 1./27.*m11*m33*RN3e*
         RS356pe*RSe45613p + 2./27.*m11*m33*RN13*RS3456pee*RS14563pe + 
         2./27.*m11*m33*RN13*RS3456ppe*RS14563pe - 2./27.*m11*m33*RN13*
         RS345pe*RS14563pe - 2./27.*m11*m33*RN13*RS346pe*RS14563pe - 2./
         27.*m11*m33*RN13*RS356pe*RS14563pe );

      M13p6x23R +=  + Ab(m33,mu2)*Ab(m34,mu2) * (  - 29./540.*RD3456ep
          - 1./270.*m44*RS356pe + 1./27.*m44*RSe563p + 1./27.*m44*
         RSp563e - 1./27.*m44*RS156pe - 1./27.*m44*RSe561p - 1./27.*m44
         *RSp561e + 1./27.*m44*mp2*RN3p*RSp563e - 1./27.*m44*mp2*RN3p*
         RSp561e + 1./27.*m44*me2*RN3e*RSe563p - 1./27.*m44*me2*RN3e*
         RSe561p + 109./1080.*m33*RS34561pe + 1./27.*m33*RS14563pe + 1./
         27.*m33*RSe45613p + 1./27.*m33*RSp45613e + 1./27.*m33*
         RSe45633p + 1./27.*m33*RSp45633e + 19./1080.*m33*RS356pe + 1./
         27.*m33*RSe563p + 1./27.*m33*RSp563e - 1./9.*m33*RS156pe - 1./
         9.*m33*RSe561p - 1./9.*m33*RSp561e + 1./27.*m33*mp2*RN3p*
         RSp45613e + 1./27.*m33*mp2*RN3p*RSp45633e + 1./27.*m33*mp2*
         RN3p*RSp563e - 1./9.*m33*mp2*RN3p*RSp561e + 1./27.*m33*me2*
         RN3e*RSe45613p + 1./27.*m33*me2*RN3e*RSe45633p + 1./27.*m33*
         me2*RN3e*RSe563p - 1./9.*m33*me2*RN3e*RSe561p - 1./27.*m33*m44
         *RN3p*RSp563e - 1./27.*m33*m44*RN3p*RSp561e - 1./27.*m33*m44*
         RN3e*RSe563p );

      M13p6x23R +=  + Ab(m33,mu2)*Ab(m34,mu2) * (  - 1./27.*m33*m44*
         RN3e*RSe561p + 1./27.*m33*m44*RN13*RS156pe - 1./27.*pow(m33,2)
         *RN3p*RSp45613e - 1./27.*pow(m33,2)*RN3p*RSp45633e - 1./27.*
         pow(m33,2)*RN3p*RSp563e - 1./27.*pow(m33,2)*RN3p*RSp561e - 1./
         27.*pow(m33,2)*RN3e*RSe45613p - 1./27.*pow(m33,2)*RN3e*
         RSe45633p - 1./27.*pow(m33,2)*RN3e*RSe563p - 1./27.*pow(m33,2)
         *RN3e*RSe561p + 1./27.*pow(m33,2)*RN13*RS14563pe + 1./27.*pow(
         m33,2)*RN13*RS156pe + 41./1080.*m11*RS34561pe + 1./27.*m11*
         RS14563pe + 1./27.*m11*RSe45613p + 1./27.*m11*RSp45613e - 1./
         24.*m11*RS356pe - 1./27.*m11*RS156pe - 1./27.*m11*RSe561p - 1./
         27.*m11*RSp561e + 1./27.*m11*mp2*RN3p*RSp45613e - 1./27.*m11*
         mp2*RN3p*RSp561e + 1./27.*m11*me2*RN3e*RSe45613p - 1./27.*m11*
         me2*RN3e*RSe561p - 1./27.*m11*m44*RN3p*RSp561e - 1./27.*m11*
         m44*RN3e*RSe561p + 2./27.*m11*m44*RN13*RS156pe - 1./27.*m11*
         m33*RN3p*RSp45613e - 1./27.*m11*m33*RN3p*RSp561e - 1./27.*m11*
         m33*RN3e*RSe45613p );

      M13p6x23R +=  + Ab(m33,mu2)*Ab(m34,mu2) * (  - 1./27.*m11*m33*
         RN3e*RSe561p + 4./27.*m11*m33*RN13*RS156pe - 1./27.*pow(m11,2)
         *RN13*RS14563pe + 1./27.*pow(m11,2)*RN13*RS156pe );

      M13p6x23R +=  + Ab(m33,mu2)*Ab(m35,mu2) * (  - 29./540.*RD3456ep
          - 1./270.*m55*RS346pe + 1./27.*m55*RSe463p + 1./27.*m55*
         RSp463e - 1./27.*m55*RS146pe - 1./27.*m55*RSe461p - 1./27.*m55
         *RSp461e + 1./27.*m55*mp2*RN3p*RSp463e - 1./27.*m55*mp2*RN3p*
         RSp461e + 1./27.*m55*me2*RN3e*RSe463p - 1./27.*m55*me2*RN3e*
         RSe461p + 109./1080.*m33*RS34561pe + 1./27.*m33*RS14563pe + 1./
         27.*m33*RSe45613p + 1./27.*m33*RSp45613e + 1./27.*m33*
         RSe45633p + 1./27.*m33*RSp45633e + 19./1080.*m33*RS346pe + 1./
         27.*m33*RSe463p + 1./27.*m33*RSp463e - 1./9.*m33*RS146pe - 1./
         9.*m33*RSe461p - 1./9.*m33*RSp461e + 1./27.*m33*mp2*RN3p*
         RSp45613e + 1./27.*m33*mp2*RN3p*RSp45633e + 1./27.*m33*mp2*
         RN3p*RSp463e - 1./9.*m33*mp2*RN3p*RSp461e + 1./27.*m33*me2*
         RN3e*RSe45613p + 1./27.*m33*me2*RN3e*RSe45633p + 1./27.*m33*
         me2*RN3e*RSe463p - 1./9.*m33*me2*RN3e*RSe461p - 1./27.*m33*m55
         *RN3p*RSp463e - 1./27.*m33*m55*RN3p*RSp461e - 1./27.*m33*m55*
         RN3e*RSe463p );

      M13p6x23R +=  + Ab(m33,mu2)*Ab(m35,mu2) * (  - 1./27.*m33*m55*
         RN3e*RSe461p + 1./27.*m33*m55*RN13*RS146pe - 1./27.*pow(m33,2)
         *RN3p*RSp45613e - 1./27.*pow(m33,2)*RN3p*RSp45633e - 1./27.*
         pow(m33,2)*RN3p*RSp463e - 1./27.*pow(m33,2)*RN3p*RSp461e - 1./
         27.*pow(m33,2)*RN3e*RSe45613p - 1./27.*pow(m33,2)*RN3e*
         RSe45633p - 1./27.*pow(m33,2)*RN3e*RSe463p - 1./27.*pow(m33,2)
         *RN3e*RSe461p + 1./27.*pow(m33,2)*RN13*RS14563pe + 1./27.*pow(
         m33,2)*RN13*RS146pe + 41./1080.*m11*RS34561pe + 1./27.*m11*
         RS14563pe + 1./27.*m11*RSe45613p + 1./27.*m11*RSp45613e - 1./
         24.*m11*RS346pe - 1./27.*m11*RS146pe - 1./27.*m11*RSe461p - 1./
         27.*m11*RSp461e + 1./27.*m11*mp2*RN3p*RSp45613e - 1./27.*m11*
         mp2*RN3p*RSp461e + 1./27.*m11*me2*RN3e*RSe45613p - 1./27.*m11*
         me2*RN3e*RSe461p - 1./27.*m11*m55*RN3p*RSp461e - 1./27.*m11*
         m55*RN3e*RSe461p + 2./27.*m11*m55*RN13*RS146pe - 1./27.*m11*
         m33*RN3p*RSp45613e - 1./27.*m11*m33*RN3p*RSp461e - 1./27.*m11*
         m33*RN3e*RSe45613p );

      M13p6x23R +=  + Ab(m33,mu2)*Ab(m35,mu2) * (  - 1./27.*m11*m33*
         RN3e*RSe461p + 4./27.*m11*m33*RN13*RS146pe - 1./27.*pow(m11,2)
         *RN13*RS14563pe + 1./27.*pow(m11,2)*RN13*RS146pe );

      M13p6x23R +=  + Ab(m33,mu2)*Ab(m36,mu2) * (  - 29./540.*RD3456ep
          - 1./270.*m66*RS345pe + 1./27.*m66*RSe453p + 1./27.*m66*
         RSp453e - 1./27.*m66*RS145pe - 1./27.*m66*RSe451p - 1./27.*m66
         *RSp451e + 1./27.*m66*mp2*RN3p*RSp453e - 1./27.*m66*mp2*RN3p*
         RSp451e + 1./27.*m66*me2*RN3e*RSe453p - 1./27.*m66*me2*RN3e*
         RSe451p + 109./1080.*m33*RS34561pe + 1./27.*m33*RS14563pe + 1./
         27.*m33*RSe45613p + 1./27.*m33*RSp45613e + 1./27.*m33*
         RSe45633p + 1./27.*m33*RSp45633e + 19./1080.*m33*RS345pe + 1./
         27.*m33*RSe453p + 1./27.*m33*RSp453e - 1./9.*m33*RS145pe - 1./
         9.*m33*RSe451p - 1./9.*m33*RSp451e + 1./27.*m33*mp2*RN3p*
         RSp45613e + 1./27.*m33*mp2*RN3p*RSp45633e + 1./27.*m33*mp2*
         RN3p*RSp453e - 1./9.*m33*mp2*RN3p*RSp451e + 1./27.*m33*me2*
         RN3e*RSe45613p + 1./27.*m33*me2*RN3e*RSe45633p + 1./27.*m33*
         me2*RN3e*RSe453p - 1./9.*m33*me2*RN3e*RSe451p - 1./27.*m33*m66
         *RN3p*RSp453e - 1./27.*m33*m66*RN3p*RSp451e - 1./27.*m33*m66*
         RN3e*RSe453p );

      M13p6x23R +=  + Ab(m33,mu2)*Ab(m36,mu2) * (  - 1./27.*m33*m66*
         RN3e*RSe451p + 1./27.*m33*m66*RN13*RS145pe - 1./27.*pow(m33,2)
         *RN3p*RSp45613e - 1./27.*pow(m33,2)*RN3p*RSp45633e - 1./27.*
         pow(m33,2)*RN3p*RSp453e - 1./27.*pow(m33,2)*RN3p*RSp451e - 1./
         27.*pow(m33,2)*RN3e*RSe45613p - 1./27.*pow(m33,2)*RN3e*
         RSe45633p - 1./27.*pow(m33,2)*RN3e*RSe453p - 1./27.*pow(m33,2)
         *RN3e*RSe451p + 1./27.*pow(m33,2)*RN13*RS14563pe + 1./27.*pow(
         m33,2)*RN13*RS145pe + 41./1080.*m11*RS34561pe + 1./27.*m11*
         RS14563pe + 1./27.*m11*RSe45613p + 1./27.*m11*RSp45613e - 1./
         24.*m11*RS345pe - 1./27.*m11*RS145pe - 1./27.*m11*RSe451p - 1./
         27.*m11*RSp451e + 1./27.*m11*mp2*RN3p*RSp45613e - 1./27.*m11*
         mp2*RN3p*RSp451e + 1./27.*m11*me2*RN3e*RSe45613p - 1./27.*m11*
         me2*RN3e*RSe451p - 1./27.*m11*m66*RN3p*RSp451e - 1./27.*m11*
         m66*RN3e*RSe451p + 2./27.*m11*m66*RN13*RS145pe - 1./27.*m11*
         m33*RN3p*RSp45613e - 1./27.*m11*m33*RN3p*RSp451e - 1./27.*m11*
         m33*RN3e*RSe45613p );

      M13p6x23R +=  + Ab(m33,mu2)*Ab(m36,mu2) * (  - 1./27.*m11*m33*
         RN3e*RSe451p + 4./27.*m11*m33*RN13*RS145pe - 1./27.*pow(m11,2)
         *RN13*RS14563pe + 1./27.*pow(m11,2)*RN13*RS145pe );

      M13p6x23R +=  + Ab(m33,mu2)*Ab(m44,mu2) * (  - 1./54.*m44*pow(
         RS356pe,2) - 1./27.*m44*RSe563p*RS356pe - 1./27.*m44*RSp563e*
         RS356pe + 1./27.*m44*RS156pe*RS356pe + 1./162.*m44*RS456pe*
         pow(RS356pe,2) + 1./81.*m44*RS456pe*RSe563p*RS356pe + 1./81.*
         m44*RS456pe*RSp563e*RS356pe - 1./81.*m44*RS456pe*RS156pe*
         RS356pe + 1./27.*m44*RSe561p*RS356pe - 1./81.*m44*RSe561p*
         RS456pe*RS356pe + 1./27.*m44*RSp561e*RS356pe - 1./81.*m44*
         RSp561e*RS456pe*RS356pe - 1./27.*m44*mp2*RN3p*RSp563e*RS356pe
          + 1./81.*m44*mp2*RN3p*RS456pe*RSp563e*RS356pe + 1./27.*m44*
         mp2*RN3p*RSp561e*RS356pe - 1./81.*m44*mp2*RN3p*RSp561e*RS456pe
         *RS356pe - 1./27.*m44*me2*RN3e*RSe563p*RS356pe + 1./81.*m44*
         me2*RN3e*RS456pe*RSe563p*RS356pe + 1./27.*m44*me2*RN3e*RSe561p
         *RS356pe - 1./81.*m44*me2*RN3e*RSe561p*RS456pe*RS356pe + 1./27.
         *m33*m44*RN3p*RSp563e*RS356pe - 1./81.*m33*m44*RN3p*RS456pe*
         RSp563e*RS356pe + 1./27.*m33*m44*RN3p*RSp561e*RS356pe - 1./81.
         *m33*m44*RN3p*RSp561e*RS456pe*RS356pe );

      M13p6x23R +=  + Ab(m33,mu2)*Ab(m44,mu2) * ( 1./27.*m33*m44*RN3e*
         RSe563p*RS356pe - 1./81.*m33*m44*RN3e*RS456pe*RSe563p*RS356pe
          + 1./27.*m33*m44*RN3e*RSe561p*RS356pe - 1./81.*m33*m44*RN3e*
         RSe561p*RS456pe*RS356pe - 1./27.*m33*m44*RN13*RS156pe*RS356pe
          + 1./81.*m33*m44*RN13*RS456pe*RS156pe*RS356pe + 1./27.*m11*
         m44*RN3p*RSp561e*RS356pe - 1./81.*m11*m44*RN3p*RSp561e*RS456pe
         *RS356pe + 1./27.*m11*m44*RN3e*RSe561p*RS356pe - 1./81.*m11*
         m44*RN3e*RSe561p*RS456pe*RS356pe - 2./27.*m11*m44*RN13*RS156pe
         *RS356pe + 2./81.*m11*m44*RN13*RS456pe*RS156pe*RS356pe );

      M13p6x23R +=  + Ab(m33,mu2)*Ab(m45,mu2) * (  - 1./162.*m55*pow(
         RS346pe,2) - 1./81.*m55*RS356pe*RS346pe - 1./81.*m55*RSe563p*
         RS346pe - 1./81.*m55*RSe463p*RS346pe - 1./81.*m55*RSe463p*
         RS356pe - 1./81.*m55*RSp563e*RS346pe - 1./81.*m55*RSp463e*
         RS346pe - 1./81.*m55*RSp463e*RS356pe + 1./81.*m55*RS146pe*
         RS346pe + 1./81.*m55*RS146pe*RS356pe + 1./81.*m55*RS156pe*
         RS346pe + 1./81.*m55*RSe561p*RS346pe + 1./81.*m55*RSe461p*
         RS346pe + 1./81.*m55*RSe461p*RS356pe + 1./81.*m55*RSp561e*
         RS346pe + 1./81.*m55*RSp461e*RS346pe + 1./81.*m55*RSp461e*
         RS356pe - 1./81.*m55*mp2*RN3p*RSp563e*RS346pe - 1./81.*m55*mp2
         *RN3p*RSp463e*RS346pe - 1./81.*m55*mp2*RN3p*RSp463e*RS356pe + 
         1./81.*m55*mp2*RN3p*RSp561e*RS346pe + 1./81.*m55*mp2*RN3p*
         RSp461e*RS346pe + 1./81.*m55*mp2*RN3p*RSp461e*RS356pe - 1./81.
         *m55*me2*RN3e*RSe563p*RS346pe - 1./81.*m55*me2*RN3e*RSe463p*
         RS346pe - 1./81.*m55*me2*RN3e*RSe463p*RS356pe + 1./81.*m55*me2
         *RN3e*RSe561p*RS346pe );

      M13p6x23R +=  + Ab(m33,mu2)*Ab(m45,mu2) * ( 1./81.*m55*me2*RN3e*
         RSe461p*RS346pe + 1./81.*m55*me2*RN3e*RSe461p*RS356pe - 1./81.
         *m44*RS356pe*RS346pe - 1./162.*m44*pow(RS356pe,2) - 1./81.*m44
         *RSe563p*RS346pe - 1./81.*m44*RSe563p*RS356pe - 1./81.*m44*
         RSe463p*RS356pe - 1./81.*m44*RSp563e*RS346pe - 1./81.*m44*
         RSp563e*RS356pe - 1./81.*m44*RSp463e*RS356pe + 1./81.*m44*
         RS146pe*RS356pe + 1./81.*m44*RS156pe*RS346pe + 1./81.*m44*
         RS156pe*RS356pe + 1./81.*m44*RSe561p*RS346pe + 1./81.*m44*
         RSe561p*RS356pe + 1./81.*m44*RSe461p*RS356pe + 1./81.*m44*
         RSp561e*RS346pe + 1./81.*m44*RSp561e*RS356pe + 1./81.*m44*
         RSp461e*RS356pe - 1./81.*m44*mp2*RN3p*RSp563e*RS346pe - 1./81.
         *m44*mp2*RN3p*RSp563e*RS356pe - 1./81.*m44*mp2*RN3p*RSp463e*
         RS356pe + 1./81.*m44*mp2*RN3p*RSp561e*RS346pe + 1./81.*m44*mp2
         *RN3p*RSp561e*RS356pe + 1./81.*m44*mp2*RN3p*RSp461e*RS356pe - 
         1./81.*m44*me2*RN3e*RSe563p*RS346pe - 1./81.*m44*me2*RN3e*
         RSe563p*RS356pe );

      M13p6x23R +=  + Ab(m33,mu2)*Ab(m45,mu2) * (  - 1./81.*m44*me2*
         RN3e*RSe463p*RS356pe + 1./81.*m44*me2*RN3e*RSe561p*RS346pe + 1.
         /81.*m44*me2*RN3e*RSe561p*RS356pe + 1./81.*m44*me2*RN3e*
         RSe461p*RS356pe + 1./162.*m33*pow(RS346pe,2) - 1./81.*m33*
         RS356pe*RS346pe + 1./162.*m33*pow(RS356pe,2) - 2./81.*m33*
         RS146pe*RS346pe + 2./81.*m33*RS146pe*RS356pe + 2./81.*m33*
         RS156pe*RS346pe - 2./81.*m33*RS156pe*RS356pe + 2./81.*m33*
         RSe561p*RS346pe - 2./81.*m33*RSe561p*RS356pe - 2./81.*m33*
         RSe461p*RS346pe + 2./81.*m33*RSe461p*RS356pe + 2./81.*m33*
         RSp561e*RS346pe - 2./81.*m33*RSp561e*RS356pe - 2./81.*m33*
         RSp461e*RS346pe + 2./81.*m33*RSp461e*RS356pe + 2./81.*m33*mp2*
         RN3p*RSp561e*RS346pe - 2./81.*m33*mp2*RN3p*RSp561e*RS356pe - 2.
         /81.*m33*mp2*RN3p*RSp461e*RS346pe + 2./81.*m33*mp2*RN3p*
         RSp461e*RS356pe + 2./81.*m33*me2*RN3e*RSe561p*RS346pe - 2./81.
         *m33*me2*RN3e*RSe561p*RS356pe - 2./81.*m33*me2*RN3e*RSe461p*
         RS346pe );

      M13p6x23R +=  + Ab(m33,mu2)*Ab(m45,mu2) * ( 2./81.*m33*me2*RN3e*
         RSe461p*RS356pe + 1./81.*m33*m55*RN3p*RSp563e*RS346pe + 1./81.
         *m33*m55*RN3p*RSp463e*RS346pe + 1./81.*m33*m55*RN3p*RSp463e*
         RS356pe + 1./81.*m33*m55*RN3p*RSp561e*RS346pe + 1./81.*m33*m55
         *RN3p*RSp461e*RS346pe + 1./81.*m33*m55*RN3p*RSp461e*RS356pe + 
         1./81.*m33*m55*RN3e*RSe563p*RS346pe + 1./81.*m33*m55*RN3e*
         RSe463p*RS346pe + 1./81.*m33*m55*RN3e*RSe463p*RS356pe + 1./81.
         *m33*m55*RN3e*RSe561p*RS346pe + 1./81.*m33*m55*RN3e*RSe461p*
         RS346pe + 1./81.*m33*m55*RN3e*RSe461p*RS356pe - 1./81.*m33*m55
         *RN13*RS146pe*RS346pe - 1./81.*m33*m55*RN13*RS146pe*RS356pe - 
         1./81.*m33*m55*RN13*RS156pe*RS346pe + 1./81.*m33*m44*RN3p*
         RSp563e*RS346pe + 1./81.*m33*m44*RN3p*RSp563e*RS356pe + 1./81.
         *m33*m44*RN3p*RSp463e*RS356pe + 1./81.*m33*m44*RN3p*RSp561e*
         RS346pe + 1./81.*m33*m44*RN3p*RSp561e*RS356pe + 1./81.*m33*m44
         *RN3p*RSp461e*RS356pe + 1./81.*m33*m44*RN3e*RSe563p*RS346pe + 
         1./81.*m33*m44*RN3e*RSe563p*RS356pe );

      M13p6x23R +=  + Ab(m33,mu2)*Ab(m45,mu2) * ( 1./81.*m33*m44*RN3e*
         RSe463p*RS356pe + 1./81.*m33*m44*RN3e*RSe561p*RS346pe + 1./81.
         *m33*m44*RN3e*RSe561p*RS356pe + 1./81.*m33*m44*RN3e*RSe461p*
         RS356pe - 1./81.*m33*m44*RN13*RS146pe*RS356pe - 1./81.*m33*m44
         *RN13*RS156pe*RS346pe - 1./81.*m33*m44*RN13*RS156pe*RS356pe - 
         1./81.*m11*RS146pe*RS346pe + 1./81.*m11*RS146pe*RS356pe + 1./
         81.*m11*RS156pe*RS346pe - 1./81.*m11*RS156pe*RS356pe + 1./81.*
         m11*RSe561p*RS346pe - 1./81.*m11*RSe561p*RS356pe - 1./81.*m11*
         RSe461p*RS346pe + 1./81.*m11*RSe461p*RS356pe + 1./81.*m11*
         RSp561e*RS346pe - 1./81.*m11*RSp561e*RS356pe - 1./81.*m11*
         RSp461e*RS346pe + 1./81.*m11*RSp461e*RS356pe + 1./81.*m11*mp2*
         RN3p*RSp561e*RS346pe - 1./81.*m11*mp2*RN3p*RSp561e*RS356pe - 1.
         /81.*m11*mp2*RN3p*RSp461e*RS346pe + 1./81.*m11*mp2*RN3p*
         RSp461e*RS356pe + 1./81.*m11*me2*RN3e*RSe561p*RS346pe - 1./81.
         *m11*me2*RN3e*RSe561p*RS356pe - 1./81.*m11*me2*RN3e*RSe461p*
         RS346pe );

      M13p6x23R +=  + Ab(m33,mu2)*Ab(m45,mu2) * ( 1./81.*m11*me2*RN3e*
         RSe461p*RS356pe + 1./81.*m11*m55*RN3p*RSp561e*RS346pe + 1./81.
         *m11*m55*RN3p*RSp461e*RS346pe + 1./81.*m11*m55*RN3p*RSp461e*
         RS356pe + 1./81.*m11*m55*RN3e*RSe561p*RS346pe + 1./81.*m11*m55
         *RN3e*RSe461p*RS346pe + 1./81.*m11*m55*RN3e*RSe461p*RS356pe - 
         2./81.*m11*m55*RN13*RS146pe*RS346pe - 2./81.*m11*m55*RN13*
         RS146pe*RS356pe - 2./81.*m11*m55*RN13*RS156pe*RS346pe + 1./81.
         *m11*m44*RN3p*RSp561e*RS346pe + 1./81.*m11*m44*RN3p*RSp561e*
         RS356pe + 1./81.*m11*m44*RN3p*RSp461e*RS356pe + 1./81.*m11*m44
         *RN3e*RSe561p*RS346pe + 1./81.*m11*m44*RN3e*RSe561p*RS356pe + 
         1./81.*m11*m44*RN3e*RSe461p*RS356pe - 2./81.*m11*m44*RN13*
         RS146pe*RS356pe - 2./81.*m11*m44*RN13*RS156pe*RS346pe - 2./81.
         *m11*m44*RN13*RS156pe*RS356pe + 2./81.*m11*m33*RN13*RS146pe*
         RS346pe - 2./81.*m11*m33*RN13*RS146pe*RS356pe - 2./81.*m11*m33
         *RN13*RS156pe*RS346pe + 2./81.*m11*m33*RN13*RS156pe*RS356pe + 
         1./81.*pow(m11,2)*RN13*RS146pe*RS346pe );

      M13p6x23R +=  + Ab(m33,mu2)*Ab(m45,mu2) * (  - 1./81.*pow(m11,2)*
         RN13*RS146pe*RS356pe - 1./81.*pow(m11,2)*RN13*RS156pe*RS346pe
          + 1./81.*pow(m11,2)*RN13*RS156pe*RS356pe );

      M13p6x23R +=  + Ab(m33,mu2)*Ab(m46,mu2) * (  - 1./162.*m66*pow(
         RS345pe,2) - 1./81.*m66*RS356pe*RS345pe - 1./81.*m66*RSe563p*
         RS345pe - 1./81.*m66*RSe453p*RS345pe - 1./81.*m66*RSe453p*
         RS356pe - 1./81.*m66*RSp563e*RS345pe - 1./81.*m66*RSp453e*
         RS345pe - 1./81.*m66*RSp453e*RS356pe + 1./81.*m66*RS145pe*
         RS345pe + 1./81.*m66*RS145pe*RS356pe + 1./81.*m66*RS156pe*
         RS345pe + 1./81.*m66*RSe561p*RS345pe + 1./81.*m66*RSe451p*
         RS345pe + 1./81.*m66*RSe451p*RS356pe + 1./81.*m66*RSp561e*
         RS345pe + 1./81.*m66*RSp451e*RS345pe + 1./81.*m66*RSp451e*
         RS356pe - 1./81.*m66*mp2*RN3p*RSp563e*RS345pe - 1./81.*m66*mp2
         *RN3p*RSp453e*RS345pe - 1./81.*m66*mp2*RN3p*RSp453e*RS356pe + 
         1./81.*m66*mp2*RN3p*RSp561e*RS345pe + 1./81.*m66*mp2*RN3p*
         RSp451e*RS345pe + 1./81.*m66*mp2*RN3p*RSp451e*RS356pe - 1./81.
         *m66*me2*RN3e*RSe563p*RS345pe - 1./81.*m66*me2*RN3e*RSe453p*
         RS345pe - 1./81.*m66*me2*RN3e*RSe453p*RS356pe + 1./81.*m66*me2
         *RN3e*RSe561p*RS345pe );

      M13p6x23R +=  + Ab(m33,mu2)*Ab(m46,mu2) * ( 1./81.*m66*me2*RN3e*
         RSe451p*RS345pe + 1./81.*m66*me2*RN3e*RSe451p*RS356pe - 1./81.
         *m44*RS356pe*RS345pe - 1./162.*m44*pow(RS356pe,2) - 1./81.*m44
         *RSe563p*RS345pe - 1./81.*m44*RSe563p*RS356pe - 1./81.*m44*
         RSe453p*RS356pe - 1./81.*m44*RSp563e*RS345pe - 1./81.*m44*
         RSp563e*RS356pe - 1./81.*m44*RSp453e*RS356pe + 1./81.*m44*
         RS145pe*RS356pe + 1./81.*m44*RS156pe*RS345pe + 1./81.*m44*
         RS156pe*RS356pe + 1./81.*m44*RSe561p*RS345pe + 1./81.*m44*
         RSe561p*RS356pe + 1./81.*m44*RSe451p*RS356pe + 1./81.*m44*
         RSp561e*RS345pe + 1./81.*m44*RSp561e*RS356pe + 1./81.*m44*
         RSp451e*RS356pe - 1./81.*m44*mp2*RN3p*RSp563e*RS345pe - 1./81.
         *m44*mp2*RN3p*RSp563e*RS356pe - 1./81.*m44*mp2*RN3p*RSp453e*
         RS356pe + 1./81.*m44*mp2*RN3p*RSp561e*RS345pe + 1./81.*m44*mp2
         *RN3p*RSp561e*RS356pe + 1./81.*m44*mp2*RN3p*RSp451e*RS356pe - 
         1./81.*m44*me2*RN3e*RSe563p*RS345pe - 1./81.*m44*me2*RN3e*
         RSe563p*RS356pe );

      M13p6x23R +=  + Ab(m33,mu2)*Ab(m46,mu2) * (  - 1./81.*m44*me2*
         RN3e*RSe453p*RS356pe + 1./81.*m44*me2*RN3e*RSe561p*RS345pe + 1.
         /81.*m44*me2*RN3e*RSe561p*RS356pe + 1./81.*m44*me2*RN3e*
         RSe451p*RS356pe + 1./162.*m33*pow(RS345pe,2) - 1./81.*m33*
         RS356pe*RS345pe + 1./162.*m33*pow(RS356pe,2) - 2./81.*m33*
         RS145pe*RS345pe + 2./81.*m33*RS145pe*RS356pe + 2./81.*m33*
         RS156pe*RS345pe - 2./81.*m33*RS156pe*RS356pe + 2./81.*m33*
         RSe561p*RS345pe - 2./81.*m33*RSe561p*RS356pe - 2./81.*m33*
         RSe451p*RS345pe + 2./81.*m33*RSe451p*RS356pe + 2./81.*m33*
         RSp561e*RS345pe - 2./81.*m33*RSp561e*RS356pe - 2./81.*m33*
         RSp451e*RS345pe + 2./81.*m33*RSp451e*RS356pe + 2./81.*m33*mp2*
         RN3p*RSp561e*RS345pe - 2./81.*m33*mp2*RN3p*RSp561e*RS356pe - 2.
         /81.*m33*mp2*RN3p*RSp451e*RS345pe + 2./81.*m33*mp2*RN3p*
         RSp451e*RS356pe + 2./81.*m33*me2*RN3e*RSe561p*RS345pe - 2./81.
         *m33*me2*RN3e*RSe561p*RS356pe - 2./81.*m33*me2*RN3e*RSe451p*
         RS345pe );

      M13p6x23R +=  + Ab(m33,mu2)*Ab(m46,mu2) * ( 2./81.*m33*me2*RN3e*
         RSe451p*RS356pe + 1./81.*m33*m66*RN3p*RSp563e*RS345pe + 1./81.
         *m33*m66*RN3p*RSp453e*RS345pe + 1./81.*m33*m66*RN3p*RSp453e*
         RS356pe + 1./81.*m33*m66*RN3p*RSp561e*RS345pe + 1./81.*m33*m66
         *RN3p*RSp451e*RS345pe + 1./81.*m33*m66*RN3p*RSp451e*RS356pe + 
         1./81.*m33*m66*RN3e*RSe563p*RS345pe + 1./81.*m33*m66*RN3e*
         RSe453p*RS345pe + 1./81.*m33*m66*RN3e*RSe453p*RS356pe + 1./81.
         *m33*m66*RN3e*RSe561p*RS345pe + 1./81.*m33*m66*RN3e*RSe451p*
         RS345pe + 1./81.*m33*m66*RN3e*RSe451p*RS356pe - 1./81.*m33*m66
         *RN13*RS145pe*RS345pe - 1./81.*m33*m66*RN13*RS145pe*RS356pe - 
         1./81.*m33*m66*RN13*RS156pe*RS345pe + 1./81.*m33*m44*RN3p*
         RSp563e*RS345pe + 1./81.*m33*m44*RN3p*RSp563e*RS356pe + 1./81.
         *m33*m44*RN3p*RSp453e*RS356pe + 1./81.*m33*m44*RN3p*RSp561e*
         RS345pe + 1./81.*m33*m44*RN3p*RSp561e*RS356pe + 1./81.*m33*m44
         *RN3p*RSp451e*RS356pe + 1./81.*m33*m44*RN3e*RSe563p*RS345pe + 
         1./81.*m33*m44*RN3e*RSe563p*RS356pe );

      M13p6x23R +=  + Ab(m33,mu2)*Ab(m46,mu2) * ( 1./81.*m33*m44*RN3e*
         RSe453p*RS356pe + 1./81.*m33*m44*RN3e*RSe561p*RS345pe + 1./81.
         *m33*m44*RN3e*RSe561p*RS356pe + 1./81.*m33*m44*RN3e*RSe451p*
         RS356pe - 1./81.*m33*m44*RN13*RS145pe*RS356pe - 1./81.*m33*m44
         *RN13*RS156pe*RS345pe - 1./81.*m33*m44*RN13*RS156pe*RS356pe - 
         1./81.*m11*RS145pe*RS345pe + 1./81.*m11*RS145pe*RS356pe + 1./
         81.*m11*RS156pe*RS345pe - 1./81.*m11*RS156pe*RS356pe + 1./81.*
         m11*RSe561p*RS345pe - 1./81.*m11*RSe561p*RS356pe - 1./81.*m11*
         RSe451p*RS345pe + 1./81.*m11*RSe451p*RS356pe + 1./81.*m11*
         RSp561e*RS345pe - 1./81.*m11*RSp561e*RS356pe - 1./81.*m11*
         RSp451e*RS345pe + 1./81.*m11*RSp451e*RS356pe + 1./81.*m11*mp2*
         RN3p*RSp561e*RS345pe - 1./81.*m11*mp2*RN3p*RSp561e*RS356pe - 1.
         /81.*m11*mp2*RN3p*RSp451e*RS345pe + 1./81.*m11*mp2*RN3p*
         RSp451e*RS356pe + 1./81.*m11*me2*RN3e*RSe561p*RS345pe - 1./81.
         *m11*me2*RN3e*RSe561p*RS356pe - 1./81.*m11*me2*RN3e*RSe451p*
         RS345pe );

      M13p6x23R +=  + Ab(m33,mu2)*Ab(m46,mu2) * ( 1./81.*m11*me2*RN3e*
         RSe451p*RS356pe + 1./81.*m11*m66*RN3p*RSp561e*RS345pe + 1./81.
         *m11*m66*RN3p*RSp451e*RS345pe + 1./81.*m11*m66*RN3p*RSp451e*
         RS356pe + 1./81.*m11*m66*RN3e*RSe561p*RS345pe + 1./81.*m11*m66
         *RN3e*RSe451p*RS345pe + 1./81.*m11*m66*RN3e*RSe451p*RS356pe - 
         2./81.*m11*m66*RN13*RS145pe*RS345pe - 2./81.*m11*m66*RN13*
         RS145pe*RS356pe - 2./81.*m11*m66*RN13*RS156pe*RS345pe + 1./81.
         *m11*m44*RN3p*RSp561e*RS345pe + 1./81.*m11*m44*RN3p*RSp561e*
         RS356pe + 1./81.*m11*m44*RN3p*RSp451e*RS356pe + 1./81.*m11*m44
         *RN3e*RSe561p*RS345pe + 1./81.*m11*m44*RN3e*RSe561p*RS356pe + 
         1./81.*m11*m44*RN3e*RSe451p*RS356pe - 2./81.*m11*m44*RN13*
         RS145pe*RS356pe - 2./81.*m11*m44*RN13*RS156pe*RS345pe - 2./81.
         *m11*m44*RN13*RS156pe*RS356pe + 2./81.*m11*m33*RN13*RS145pe*
         RS345pe - 2./81.*m11*m33*RN13*RS145pe*RS356pe - 2./81.*m11*m33
         *RN13*RS156pe*RS345pe + 2./81.*m11*m33*RN13*RS156pe*RS356pe + 
         1./81.*pow(m11,2)*RN13*RS145pe*RS345pe );

      M13p6x23R +=  + Ab(m33,mu2)*Ab(m46,mu2) * (  - 1./81.*pow(m11,2)*
         RN13*RS145pe*RS356pe - 1./81.*pow(m11,2)*RN13*RS156pe*RS345pe
          + 1./81.*pow(m11,2)*RN13*RS156pe*RS356pe );

      M13p6x23R +=  + Ab(m33,mu2)*Ab(m55,mu2) * (  - 1./54.*m55*pow(
         RS346pe,2) - 1./27.*m55*RSe463p*RS346pe - 1./27.*m55*RSp463e*
         RS346pe + 1./27.*m55*RS146pe*RS346pe + 1./162.*m55*RS546pe*
         pow(RS346pe,2) + 1./81.*m55*RS546pe*RSe463p*RS346pe + 1./81.*
         m55*RS546pe*RSp463e*RS346pe - 1./81.*m55*RS546pe*RS146pe*
         RS346pe + 1./27.*m55*RSe461p*RS346pe - 1./81.*m55*RSe461p*
         RS546pe*RS346pe + 1./27.*m55*RSp461e*RS346pe - 1./81.*m55*
         RSp461e*RS546pe*RS346pe - 1./27.*m55*mp2*RN3p*RSp463e*RS346pe
          + 1./81.*m55*mp2*RN3p*RS546pe*RSp463e*RS346pe + 1./27.*m55*
         mp2*RN3p*RSp461e*RS346pe - 1./81.*m55*mp2*RN3p*RSp461e*RS546pe
         *RS346pe - 1./27.*m55*me2*RN3e*RSe463p*RS346pe + 1./81.*m55*
         me2*RN3e*RS546pe*RSe463p*RS346pe + 1./27.*m55*me2*RN3e*RSe461p
         *RS346pe - 1./81.*m55*me2*RN3e*RSe461p*RS546pe*RS346pe + 1./27.
         *m33*m55*RN3p*RSp463e*RS346pe - 1./81.*m33*m55*RN3p*RS546pe*
         RSp463e*RS346pe + 1./27.*m33*m55*RN3p*RSp461e*RS346pe - 1./81.
         *m33*m55*RN3p*RSp461e*RS546pe*RS346pe );

      M13p6x23R +=  + Ab(m33,mu2)*Ab(m55,mu2) * ( 1./27.*m33*m55*RN3e*
         RSe463p*RS346pe - 1./81.*m33*m55*RN3e*RS546pe*RSe463p*RS346pe
          + 1./27.*m33*m55*RN3e*RSe461p*RS346pe - 1./81.*m33*m55*RN3e*
         RSe461p*RS546pe*RS346pe - 1./27.*m33*m55*RN13*RS146pe*RS346pe
          + 1./81.*m33*m55*RN13*RS546pe*RS146pe*RS346pe + 1./27.*m11*
         m55*RN3p*RSp461e*RS346pe - 1./81.*m11*m55*RN3p*RSp461e*RS546pe
         *RS346pe + 1./27.*m11*m55*RN3e*RSe461p*RS346pe - 1./81.*m11*
         m55*RN3e*RSe461p*RS546pe*RS346pe - 2./27.*m11*m55*RN13*RS146pe
         *RS346pe + 2./81.*m11*m55*RN13*RS546pe*RS146pe*RS346pe );

      M13p6x23R +=  + Ab(m33,mu2)*Ab(m56,mu2) * (  - 1./162.*m66*pow(
         RS345pe,2) - 1./81.*m66*RS346pe*RS345pe - 1./81.*m66*RSe463p*
         RS345pe - 1./81.*m66*RSe453p*RS345pe - 1./81.*m66*RSe453p*
         RS346pe - 1./81.*m66*RSp463e*RS345pe - 1./81.*m66*RSp453e*
         RS345pe - 1./81.*m66*RSp453e*RS346pe + 1./81.*m66*RS145pe*
         RS345pe + 1./81.*m66*RS145pe*RS346pe + 1./81.*m66*RS146pe*
         RS345pe + 1./81.*m66*RSe461p*RS345pe + 1./81.*m66*RSe451p*
         RS345pe + 1./81.*m66*RSe451p*RS346pe + 1./81.*m66*RSp461e*
         RS345pe + 1./81.*m66*RSp451e*RS345pe + 1./81.*m66*RSp451e*
         RS346pe - 1./81.*m66*mp2*RN3p*RSp463e*RS345pe - 1./81.*m66*mp2
         *RN3p*RSp453e*RS345pe - 1./81.*m66*mp2*RN3p*RSp453e*RS346pe + 
         1./81.*m66*mp2*RN3p*RSp461e*RS345pe + 1./81.*m66*mp2*RN3p*
         RSp451e*RS345pe + 1./81.*m66*mp2*RN3p*RSp451e*RS346pe - 1./81.
         *m66*me2*RN3e*RSe463p*RS345pe - 1./81.*m66*me2*RN3e*RSe453p*
         RS345pe - 1./81.*m66*me2*RN3e*RSe453p*RS346pe + 1./81.*m66*me2
         *RN3e*RSe461p*RS345pe );

      M13p6x23R +=  + Ab(m33,mu2)*Ab(m56,mu2) * ( 1./81.*m66*me2*RN3e*
         RSe451p*RS345pe + 1./81.*m66*me2*RN3e*RSe451p*RS346pe - 1./81.
         *m55*RS346pe*RS345pe - 1./162.*m55*pow(RS346pe,2) - 1./81.*m55
         *RSe463p*RS345pe - 1./81.*m55*RSe463p*RS346pe - 1./81.*m55*
         RSe453p*RS346pe - 1./81.*m55*RSp463e*RS345pe - 1./81.*m55*
         RSp463e*RS346pe - 1./81.*m55*RSp453e*RS346pe + 1./81.*m55*
         RS145pe*RS346pe + 1./81.*m55*RS146pe*RS345pe + 1./81.*m55*
         RS146pe*RS346pe + 1./81.*m55*RSe461p*RS345pe + 1./81.*m55*
         RSe461p*RS346pe + 1./81.*m55*RSe451p*RS346pe + 1./81.*m55*
         RSp461e*RS345pe + 1./81.*m55*RSp461e*RS346pe + 1./81.*m55*
         RSp451e*RS346pe - 1./81.*m55*mp2*RN3p*RSp463e*RS345pe - 1./81.
         *m55*mp2*RN3p*RSp463e*RS346pe - 1./81.*m55*mp2*RN3p*RSp453e*
         RS346pe + 1./81.*m55*mp2*RN3p*RSp461e*RS345pe + 1./81.*m55*mp2
         *RN3p*RSp461e*RS346pe + 1./81.*m55*mp2*RN3p*RSp451e*RS346pe - 
         1./81.*m55*me2*RN3e*RSe463p*RS345pe - 1./81.*m55*me2*RN3e*
         RSe463p*RS346pe );

      M13p6x23R +=  + Ab(m33,mu2)*Ab(m56,mu2) * (  - 1./81.*m55*me2*
         RN3e*RSe453p*RS346pe + 1./81.*m55*me2*RN3e*RSe461p*RS345pe + 1.
         /81.*m55*me2*RN3e*RSe461p*RS346pe + 1./81.*m55*me2*RN3e*
         RSe451p*RS346pe + 1./162.*m33*pow(RS345pe,2) - 1./81.*m33*
         RS346pe*RS345pe + 1./162.*m33*pow(RS346pe,2) - 2./81.*m33*
         RS145pe*RS345pe + 2./81.*m33*RS145pe*RS346pe + 2./81.*m33*
         RS146pe*RS345pe - 2./81.*m33*RS146pe*RS346pe + 2./81.*m33*
         RSe461p*RS345pe - 2./81.*m33*RSe461p*RS346pe - 2./81.*m33*
         RSe451p*RS345pe + 2./81.*m33*RSe451p*RS346pe + 2./81.*m33*
         RSp461e*RS345pe - 2./81.*m33*RSp461e*RS346pe - 2./81.*m33*
         RSp451e*RS345pe + 2./81.*m33*RSp451e*RS346pe + 2./81.*m33*mp2*
         RN3p*RSp461e*RS345pe - 2./81.*m33*mp2*RN3p*RSp461e*RS346pe - 2.
         /81.*m33*mp2*RN3p*RSp451e*RS345pe + 2./81.*m33*mp2*RN3p*
         RSp451e*RS346pe + 2./81.*m33*me2*RN3e*RSe461p*RS345pe - 2./81.
         *m33*me2*RN3e*RSe461p*RS346pe - 2./81.*m33*me2*RN3e*RSe451p*
         RS345pe );

      M13p6x23R +=  + Ab(m33,mu2)*Ab(m56,mu2) * ( 2./81.*m33*me2*RN3e*
         RSe451p*RS346pe + 1./81.*m33*m66*RN3p*RSp463e*RS345pe + 1./81.
         *m33*m66*RN3p*RSp453e*RS345pe + 1./81.*m33*m66*RN3p*RSp453e*
         RS346pe + 1./81.*m33*m66*RN3p*RSp461e*RS345pe + 1./81.*m33*m66
         *RN3p*RSp451e*RS345pe + 1./81.*m33*m66*RN3p*RSp451e*RS346pe + 
         1./81.*m33*m66*RN3e*RSe463p*RS345pe + 1./81.*m33*m66*RN3e*
         RSe453p*RS345pe + 1./81.*m33*m66*RN3e*RSe453p*RS346pe + 1./81.
         *m33*m66*RN3e*RSe461p*RS345pe + 1./81.*m33*m66*RN3e*RSe451p*
         RS345pe + 1./81.*m33*m66*RN3e*RSe451p*RS346pe - 1./81.*m33*m66
         *RN13*RS145pe*RS345pe - 1./81.*m33*m66*RN13*RS145pe*RS346pe - 
         1./81.*m33*m66*RN13*RS146pe*RS345pe + 1./81.*m33*m55*RN3p*
         RSp463e*RS345pe + 1./81.*m33*m55*RN3p*RSp463e*RS346pe + 1./81.
         *m33*m55*RN3p*RSp453e*RS346pe + 1./81.*m33*m55*RN3p*RSp461e*
         RS345pe + 1./81.*m33*m55*RN3p*RSp461e*RS346pe + 1./81.*m33*m55
         *RN3p*RSp451e*RS346pe + 1./81.*m33*m55*RN3e*RSe463p*RS345pe + 
         1./81.*m33*m55*RN3e*RSe463p*RS346pe );

      M13p6x23R +=  + Ab(m33,mu2)*Ab(m56,mu2) * ( 1./81.*m33*m55*RN3e*
         RSe453p*RS346pe + 1./81.*m33*m55*RN3e*RSe461p*RS345pe + 1./81.
         *m33*m55*RN3e*RSe461p*RS346pe + 1./81.*m33*m55*RN3e*RSe451p*
         RS346pe - 1./81.*m33*m55*RN13*RS145pe*RS346pe - 1./81.*m33*m55
         *RN13*RS146pe*RS345pe - 1./81.*m33*m55*RN13*RS146pe*RS346pe - 
         1./81.*m11*RS145pe*RS345pe + 1./81.*m11*RS145pe*RS346pe + 1./
         81.*m11*RS146pe*RS345pe - 1./81.*m11*RS146pe*RS346pe + 1./81.*
         m11*RSe461p*RS345pe - 1./81.*m11*RSe461p*RS346pe - 1./81.*m11*
         RSe451p*RS345pe + 1./81.*m11*RSe451p*RS346pe + 1./81.*m11*
         RSp461e*RS345pe - 1./81.*m11*RSp461e*RS346pe - 1./81.*m11*
         RSp451e*RS345pe + 1./81.*m11*RSp451e*RS346pe + 1./81.*m11*mp2*
         RN3p*RSp461e*RS345pe - 1./81.*m11*mp2*RN3p*RSp461e*RS346pe - 1.
         /81.*m11*mp2*RN3p*RSp451e*RS345pe + 1./81.*m11*mp2*RN3p*
         RSp451e*RS346pe + 1./81.*m11*me2*RN3e*RSe461p*RS345pe - 1./81.
         *m11*me2*RN3e*RSe461p*RS346pe - 1./81.*m11*me2*RN3e*RSe451p*
         RS345pe );

      M13p6x23R +=  + Ab(m33,mu2)*Ab(m56,mu2) * ( 1./81.*m11*me2*RN3e*
         RSe451p*RS346pe + 1./81.*m11*m66*RN3p*RSp461e*RS345pe + 1./81.
         *m11*m66*RN3p*RSp451e*RS345pe + 1./81.*m11*m66*RN3p*RSp451e*
         RS346pe + 1./81.*m11*m66*RN3e*RSe461p*RS345pe + 1./81.*m11*m66
         *RN3e*RSe451p*RS345pe + 1./81.*m11*m66*RN3e*RSe451p*RS346pe - 
         2./81.*m11*m66*RN13*RS145pe*RS345pe - 2./81.*m11*m66*RN13*
         RS145pe*RS346pe - 2./81.*m11*m66*RN13*RS146pe*RS345pe + 1./81.
         *m11*m55*RN3p*RSp461e*RS345pe + 1./81.*m11*m55*RN3p*RSp461e*
         RS346pe + 1./81.*m11*m55*RN3p*RSp451e*RS346pe + 1./81.*m11*m55
         *RN3e*RSe461p*RS345pe + 1./81.*m11*m55*RN3e*RSe461p*RS346pe + 
         1./81.*m11*m55*RN3e*RSe451p*RS346pe - 2./81.*m11*m55*RN13*
         RS145pe*RS346pe - 2./81.*m11*m55*RN13*RS146pe*RS345pe - 2./81.
         *m11*m55*RN13*RS146pe*RS346pe + 2./81.*m11*m33*RN13*RS145pe*
         RS345pe - 2./81.*m11*m33*RN13*RS145pe*RS346pe - 2./81.*m11*m33
         *RN13*RS146pe*RS345pe + 2./81.*m11*m33*RN13*RS146pe*RS346pe + 
         1./81.*pow(m11,2)*RN13*RS145pe*RS345pe );

      M13p6x23R +=  + Ab(m33,mu2)*Ab(m56,mu2) * (  - 1./81.*pow(m11,2)*
         RN13*RS145pe*RS346pe - 1./81.*pow(m11,2)*RN13*RS146pe*RS345pe
          + 1./81.*pow(m11,2)*RN13*RS146pe*RS346pe );

      M13p6x23R +=  + Ab(m33,mu2)*Ab(m66,mu2) * (  - 1./54.*m66*pow(
         RS345pe,2) - 1./27.*m66*RSe453p*RS345pe - 1./27.*m66*RSp453e*
         RS345pe + 1./27.*m66*RS145pe*RS345pe + 1./162.*m66*RS645pe*
         pow(RS345pe,2) + 1./81.*m66*RS645pe*RSe453p*RS345pe + 1./81.*
         m66*RS645pe*RSp453e*RS345pe - 1./81.*m66*RS645pe*RS145pe*
         RS345pe + 1./27.*m66*RSe451p*RS345pe - 1./81.*m66*RSe451p*
         RS645pe*RS345pe + 1./27.*m66*RSp451e*RS345pe - 1./81.*m66*
         RSp451e*RS645pe*RS345pe - 1./27.*m66*mp2*RN3p*RSp453e*RS345pe
          + 1./81.*m66*mp2*RN3p*RS645pe*RSp453e*RS345pe + 1./27.*m66*
         mp2*RN3p*RSp451e*RS345pe - 1./81.*m66*mp2*RN3p*RSp451e*RS645pe
         *RS345pe - 1./27.*m66*me2*RN3e*RSe453p*RS345pe + 1./81.*m66*
         me2*RN3e*RS645pe*RSe453p*RS345pe + 1./27.*m66*me2*RN3e*RSe451p
         *RS345pe - 1./81.*m66*me2*RN3e*RSe451p*RS645pe*RS345pe + 1./27.
         *m33*m66*RN3p*RSp453e*RS345pe - 1./81.*m33*m66*RN3p*RS645pe*
         RSp453e*RS345pe + 1./27.*m33*m66*RN3p*RSp451e*RS345pe - 1./81.
         *m33*m66*RN3p*RSp451e*RS645pe*RS345pe );

      M13p6x23R +=  + Ab(m33,mu2)*Ab(m66,mu2) * ( 1./27.*m33*m66*RN3e*
         RSe453p*RS345pe - 1./81.*m33*m66*RN3e*RS645pe*RSe453p*RS345pe
          + 1./27.*m33*m66*RN3e*RSe451p*RS345pe - 1./81.*m33*m66*RN3e*
         RSe451p*RS645pe*RS345pe - 1./27.*m33*m66*RN13*RS145pe*RS345pe
          + 1./81.*m33*m66*RN13*RS645pe*RS145pe*RS345pe + 1./27.*m11*
         m66*RN3p*RSp451e*RS345pe - 1./81.*m11*m66*RN3p*RSp451e*RS645pe
         *RS345pe + 1./27.*m11*m66*RN3e*RSe451p*RS345pe - 1./81.*m11*
         m66*RN3e*RSe451p*RS645pe*RS345pe - 2./27.*m11*m66*RN13*RS145pe
         *RS345pe + 2./81.*m11*m66*RN13*RS645pe*RS145pe*RS345pe );

      M13p6x23R +=  + Ab(m33,mu2)*Ab(me2,mu2) * (  - 1./270.*RD3456ep*
         RSe45613p + 1./540.*RSe45633p*RD3456ep + 1./540.*RSe45611p*
         RD3456ep + 1./135.*me2*RSe45613p*RS34561pe + 1./270.*me2*
         RS3456pee*RSe45613p + 1./270.*me2*RS3456ppe*RSe45613p - 1./270.
         *me2*RSe45633p*RS34561pe - 1./540.*me2*RSe45633p*RS3456pee - 1.
         /540.*me2*RSe45633p*RS3456ppe - 1./270.*me2*RS345pe*RSe45613p
          + 1./540.*me2*RS345pe*RSe45633p - 1./270.*me2*RS346pe*
         RSe45613p + 1./540.*me2*RS346pe*RSe45633p - 1./270.*me2*
         RS356pe*RSe45613p + 1./540.*me2*RS356pe*RSe45633p - 1./270.*
         me2*RSe45611p*RS34561pe - 1./540.*me2*RSe45611p*RS3456pee - 1./
         540.*me2*RSe45611p*RS3456ppe + 1./540.*me2*RSe45611p*RS345pe
          + 1./540.*me2*RSe45611p*RS346pe + 1./540.*me2*RSe45611p*
         RS356pe + 1./162.*m66*RSe456p*pow(RS345pe,2) + 1./81.*m66*
         RSe456p*RSe453p*RS345pe + 1./81.*m66*RSe456p*RSp453e*RS345pe
          - 1./81.*m66*RSe456p*RS145pe*RS345pe - 1./81.*m66*RSe456p*
         RSe451p*RS345pe );

      M13p6x23R +=  + Ab(m33,mu2)*Ab(me2,mu2) * (  - 1./81.*m66*RSe456p
         *RSp451e*RS345pe + 1./81.*m66*mp2*RN3p*RSe456p*RSp453e*RS345pe
          - 1./81.*m66*mp2*RN3p*RSe456p*RSp451e*RS345pe + 1./81.*m66*
         me2*RN3e*RSe456p*RSe453p*RS345pe - 1./81.*m66*me2*RN3e*RSe456p
         *RSe451p*RS345pe + 1./162.*m55*RSe465p*pow(RS346pe,2) + 1./81.
         *m55*RSe465p*RSe463p*RS346pe + 1./81.*m55*RSe465p*RSp463e*
         RS346pe - 1./81.*m55*RSe465p*RS146pe*RS346pe - 1./81.*m55*
         RSe465p*RSe461p*RS346pe - 1./81.*m55*RSe465p*RSp461e*RS346pe
          + 1./81.*m55*mp2*RN3p*RSe465p*RSp463e*RS346pe - 1./81.*m55*
         mp2*RN3p*RSe465p*RSp461e*RS346pe + 1./81.*m55*me2*RN3e*RSe465p
         *RSe463p*RS346pe - 1./81.*m55*me2*RN3e*RSe465p*RSe461p*RS346pe
          + 1./162.*m44*RSe564p*pow(RS356pe,2) + 1./81.*m44*RSe564p*
         RSe563p*RS356pe + 1./81.*m44*RSe564p*RSp563e*RS356pe - 1./81.*
         m44*RSe564p*RS156pe*RS356pe - 1./81.*m44*RSe564p*RSe561p*
         RS356pe - 1./81.*m44*RSe564p*RSp561e*RS356pe + 1./81.*m44*mp2*
         RN3p*RSe564p*RSp563e*RS356pe );

      M13p6x23R +=  + Ab(m33,mu2)*Ab(me2,mu2) * (  - 1./81.*m44*mp2*
         RN3p*RSe564p*RSp561e*RS356pe + 1./81.*m44*me2*RN3e*RSe564p*
         RSe563p*RS356pe - 1./81.*m44*me2*RN3e*RSe564p*RSe561p*RS356pe
          + 1./18.*m33*RSe45613p*RS34561pe + 1./60.*m33*RS3456pee*
         RSe45613p + 1./60.*m33*RS3456ppe*RSe45613p + 11./540.*m33*
         RSe45633p*RS34561pe + 1./27.*m33*RSe45633p*RS14563pe + 1./27.*
         m33*RSe45633p*RSe45613p + 1./27.*m33*RSe45633p*RSp45613e - 13./
         1080.*m33*RSe45633p*RS3456pee - 13./1080.*m33*RSe45633p*
         RS3456ppe - 1./27.*m33*pow(RSe45633p,2) - 1./27.*m33*RSp45633e
         *RSe45633p - 1./60.*m33*RS345pe*RSe45613p + 13./1080.*m33*
         RS345pe*RSe45633p - 1./60.*m33*RS346pe*RSe45613p + 13./1080.*
         m33*RS346pe*RSe45633p - 1./60.*m33*RS356pe*RSe45613p + 13./
         1080.*m33*RS356pe*RSe45633p - 1./540.*m33*RSe45611p*RS34561pe
          + 1./72.*m33*RSe45611p*RS3456pee + 1./72.*m33*RSe45611p*
         RS3456ppe - 1./72.*m33*RSe45611p*RS345pe - 1./72.*m33*
         RSe45611p*RS346pe );

      M13p6x23R +=  + Ab(m33,mu2)*Ab(me2,mu2) * (  - 1./72.*m33*
         RSe45611p*RS356pe + 1./27.*m33*mp2*RN3p*RSe45633p*RSp45613e - 
         1./27.*m33*mp2*RN3p*RSp45633e*RSe45633p + 1./27.*m33*me2*RN3e*
         RS3456pee*RSe45613p + 1./27.*m33*me2*RN3e*RS3456ppe*RSe45613p
          + 1./27.*m33*me2*RN3e*RSe45633p*RSe45613p - 1./27.*m33*me2*
         RN3e*RSe45633p*RS3456pee - 1./27.*m33*me2*RN3e*RSe45633p*
         RS3456ppe - 1./27.*m33*me2*RN3e*pow(RSe45633p,2) - 1./27.*m33*
         me2*RN3e*RS345pe*RSe45613p + 1./27.*m33*me2*RN3e*RS345pe*
         RSe45633p - 1./27.*m33*me2*RN3e*RS346pe*RSe45613p + 1./27.*m33
         *me2*RN3e*RS346pe*RSe45633p - 1./27.*m33*me2*RN3e*RS356pe*
         RSe45613p + 1./27.*m33*me2*RN3e*RS356pe*RSe45633p - 1./81.*m33
         *m66*RN3p*RSe456p*RSp453e*RS345pe - 1./81.*m33*m66*RN3p*
         RSe456p*RSp451e*RS345pe - 1./81.*m33*m66*RN3e*RSe456p*RSe453p*
         RS345pe - 1./81.*m33*m66*RN3e*RSe456p*RSe451p*RS345pe + 1./81.
         *m33*m66*RN13*RSe456p*RS145pe*RS345pe - 1./81.*m33*m55*RN3p*
         RSe465p*RSp463e*RS346pe );

      M13p6x23R +=  + Ab(m33,mu2)*Ab(me2,mu2) * (  - 1./81.*m33*m55*
         RN3p*RSe465p*RSp461e*RS346pe - 1./81.*m33*m55*RN3e*RSe465p*
         RSe463p*RS346pe - 1./81.*m33*m55*RN3e*RSe465p*RSe461p*RS346pe
          + 1./81.*m33*m55*RN13*RSe465p*RS146pe*RS346pe - 1./81.*m33*
         m44*RN3p*RSe564p*RSp563e*RS356pe - 1./81.*m33*m44*RN3p*RSe564p
         *RSp561e*RS356pe - 1./81.*m33*m44*RN3e*RSe564p*RSe563p*RS356pe
          - 1./81.*m33*m44*RN3e*RSe564p*RSe561p*RS356pe + 1./81.*m33*
         m44*RN13*RSe564p*RS156pe*RS356pe + 1./27.*pow(m33,2)*RN3p*
         RSe45633p*RSp45613e + 1./27.*pow(m33,2)*RN3p*RSp45633e*
         RSe45633p + 1./27.*pow(m33,2)*RN3e*RS3456pee*RSe45613p + 1./27.
         *pow(m33,2)*RN3e*RS3456ppe*RSe45613p + 1./27.*pow(m33,2)*RN3e*
         RSe45633p*RSe45613p + 1./27.*pow(m33,2)*RN3e*RSe45633p*
         RS3456pee + 1./27.*pow(m33,2)*RN3e*RSe45633p*RS3456ppe + 1./27.
         *pow(m33,2)*RN3e*pow(RSe45633p,2) - 1./27.*pow(m33,2)*RN3e*
         RS345pe*RSe45613p - 1./27.*pow(m33,2)*RN3e*RS345pe*RSe45633p
          - 1./27.*pow(m33,2)*RN3e*RS346pe*RSe45613p );

      M13p6x23R +=  + Ab(m33,mu2)*Ab(me2,mu2) * (  - 1./27.*pow(m33,2)*
         RN3e*RS346pe*RSe45633p - 1./27.*pow(m33,2)*RN3e*RS356pe*
         RSe45613p - 1./27.*pow(m33,2)*RN3e*RS356pe*RSe45633p - 1./27.*
         pow(m33,2)*RN13*RSe45633p*RS14563pe + 13./270.*m11*RSe45613p*
         RS34561pe - 1./540.*m11*RS3456pee*RSe45613p - 1./540.*m11*
         RS3456ppe*RSe45613p + 1./540.*m11*RSe45633p*RS34561pe - 1./72.
         *m11*RSe45633p*RS3456pee - 1./72.*m11*RSe45633p*RS3456ppe + 1./
         540.*m11*RS345pe*RSe45613p + 1./72.*m11*RS345pe*RSe45633p + 1./
         540.*m11*RS346pe*RSe45613p + 1./72.*m11*RS346pe*RSe45633p + 1./
         540.*m11*RS356pe*RSe45613p + 1./72.*m11*RS356pe*RSe45633p + 13.
         /540.*m11*RSe45611p*RS34561pe + 17./1080.*m11*RSe45611p*
         RS3456pee + 17./1080.*m11*RSe45611p*RS3456ppe - 17./1080.*m11*
         RSe45611p*RS345pe - 17./1080.*m11*RSe45611p*RS346pe - 17./1080.
         *m11*RSe45611p*RS356pe - 1./81.*m11*m66*RN3p*RSe456p*RSp451e*
         RS345pe - 1./81.*m11*m66*RN3e*RSe456p*RSe451p*RS345pe + 2./81.
         *m11*m66*RN13*RSe456p*RS145pe*RS345pe );

      M13p6x23R +=  + Ab(m33,mu2)*Ab(me2,mu2) * (  - 1./81.*m11*m55*
         RN3p*RSe465p*RSp461e*RS346pe - 1./81.*m11*m55*RN3e*RSe465p*
         RSe461p*RS346pe + 2./81.*m11*m55*RN13*RSe465p*RS146pe*RS346pe
          - 1./81.*m11*m44*RN3p*RSe564p*RSp561e*RS356pe - 1./81.*m11*
         m44*RN3e*RSe564p*RSe561p*RS356pe + 2./81.*m11*m44*RN13*RSe564p
         *RS156pe*RS356pe + 1./27.*m11*m33*RN3p*RSe45633p*RSp45613e + 1.
         /27.*m11*m33*RN3e*RS3456pee*RSe45613p + 1./27.*m11*m33*RN3e*
         RS3456ppe*RSe45613p + 1./27.*m11*m33*RN3e*RSe45633p*RSe45613p
          - 1./27.*m11*m33*RN3e*RS345pe*RSe45613p - 1./27.*m11*m33*RN3e
         *RS346pe*RSe45613p - 1./27.*m11*m33*RN3e*RS356pe*RSe45613p - 2.
         /27.*m11*m33*RN13*RSe45633p*RS14563pe - 1./27.*m11*m33*RN13*
         RSe45611p*RS34561pe - 2./27.*pow(m11,2)*RN13*RSe45611p*
         RS34561pe );

      M13p6x23R +=  + Ab(m33,mu2)*Ab(mp2,mu2) * (  - 1./270.*RD3456ep*
         RSp45613e + 1./540.*RSp45633e*RD3456ep + 1./540.*RSp45611e*
         RD3456ep + 1./135.*mp2*RSp45613e*RS34561pe + 1./270.*mp2*
         RS3456pee*RSp45613e + 1./270.*mp2*RS3456ppe*RSp45613e - 1./270.
         *mp2*RSp45633e*RS34561pe - 1./540.*mp2*RSp45633e*RS3456pee - 1.
         /540.*mp2*RSp45633e*RS3456ppe - 1./270.*mp2*RS345pe*RSp45613e
          + 1./540.*mp2*RS345pe*RSp45633e - 1./270.*mp2*RS346pe*
         RSp45613e + 1./540.*mp2*RS346pe*RSp45633e - 1./270.*mp2*
         RS356pe*RSp45613e + 1./540.*mp2*RS356pe*RSp45633e - 1./270.*
         mp2*RSp45611e*RS34561pe - 1./540.*mp2*RSp45611e*RS3456pee - 1./
         540.*mp2*RSp45611e*RS3456ppe + 1./540.*mp2*RSp45611e*RS345pe
          + 1./540.*mp2*RSp45611e*RS346pe + 1./540.*mp2*RSp45611e*
         RS356pe + 1./162.*m66*RSp456e*pow(RS345pe,2) + 1./81.*m66*
         RSp456e*RSe453p*RS345pe + 1./81.*m66*RSp456e*RSp453e*RS345pe
          - 1./81.*m66*RSp456e*RS145pe*RS345pe - 1./81.*m66*RSp456e*
         RSe451p*RS345pe );

      M13p6x23R +=  + Ab(m33,mu2)*Ab(mp2,mu2) * (  - 1./81.*m66*RSp456e
         *RSp451e*RS345pe + 1./81.*m66*mp2*RN3p*RSp456e*RSp453e*RS345pe
          - 1./81.*m66*mp2*RN3p*RSp456e*RSp451e*RS345pe + 1./81.*m66*
         me2*RN3e*RSp456e*RSe453p*RS345pe - 1./81.*m66*me2*RN3e*RSp456e
         *RSe451p*RS345pe + 1./162.*m55*RSp465e*pow(RS346pe,2) + 1./81.
         *m55*RSp465e*RSe463p*RS346pe + 1./81.*m55*RSp465e*RSp463e*
         RS346pe - 1./81.*m55*RSp465e*RS146pe*RS346pe - 1./81.*m55*
         RSp465e*RSe461p*RS346pe - 1./81.*m55*RSp465e*RSp461e*RS346pe
          + 1./81.*m55*mp2*RN3p*RSp465e*RSp463e*RS346pe - 1./81.*m55*
         mp2*RN3p*RSp465e*RSp461e*RS346pe + 1./81.*m55*me2*RN3e*RSp465e
         *RSe463p*RS346pe - 1./81.*m55*me2*RN3e*RSp465e*RSe461p*RS346pe
          + 1./162.*m44*RSp564e*pow(RS356pe,2) + 1./81.*m44*RSp564e*
         RSe563p*RS356pe + 1./81.*m44*RSp564e*RSp563e*RS356pe - 1./81.*
         m44*RSp564e*RS156pe*RS356pe - 1./81.*m44*RSp564e*RSe561p*
         RS356pe - 1./81.*m44*RSp564e*RSp561e*RS356pe + 1./81.*m44*mp2*
         RN3p*RSp564e*RSp563e*RS356pe );

      M13p6x23R +=  + Ab(m33,mu2)*Ab(mp2,mu2) * (  - 1./81.*m44*mp2*
         RN3p*RSp564e*RSp561e*RS356pe + 1./81.*m44*me2*RN3e*RSp564e*
         RSe563p*RS356pe - 1./81.*m44*me2*RN3e*RSp564e*RSe561p*RS356pe
          + 1./18.*m33*RSp45613e*RS34561pe + 1./60.*m33*RS3456pee*
         RSp45613e + 1./60.*m33*RS3456ppe*RSp45613e + 11./540.*m33*
         RSp45633e*RS34561pe + 1./27.*m33*RSp45633e*RS14563pe + 1./27.*
         m33*RSp45633e*RSe45613p + 1./27.*m33*RSp45633e*RSp45613e - 13./
         1080.*m33*RSp45633e*RS3456pee - 13./1080.*m33*RSp45633e*
         RS3456ppe - 1./27.*m33*RSp45633e*RSe45633p - 1./27.*m33*pow(
         RSp45633e,2) - 1./60.*m33*RS345pe*RSp45613e + 13./1080.*m33*
         RS345pe*RSp45633e - 1./60.*m33*RS346pe*RSp45613e + 13./1080.*
         m33*RS346pe*RSp45633e - 1./60.*m33*RS356pe*RSp45613e + 13./
         1080.*m33*RS356pe*RSp45633e - 1./540.*m33*RSp45611e*RS34561pe
          + 1./72.*m33*RSp45611e*RS3456pee + 1./72.*m33*RSp45611e*
         RS3456ppe - 1./72.*m33*RSp45611e*RS345pe - 1./72.*m33*
         RSp45611e*RS346pe );

      M13p6x23R +=  + Ab(m33,mu2)*Ab(mp2,mu2) * (  - 1./72.*m33*
         RSp45611e*RS356pe + 1./27.*m33*mp2*RN3p*RS3456pee*RSp45613e + 
         1./27.*m33*mp2*RN3p*RS3456ppe*RSp45613e + 1./27.*m33*mp2*RN3p*
         RSp45633e*RSp45613e - 1./27.*m33*mp2*RN3p*RSp45633e*RS3456pee
          - 1./27.*m33*mp2*RN3p*RSp45633e*RS3456ppe - 1./27.*m33*mp2*
         RN3p*pow(RSp45633e,2) - 1./27.*m33*mp2*RN3p*RS345pe*RSp45613e
          + 1./27.*m33*mp2*RN3p*RS345pe*RSp45633e - 1./27.*m33*mp2*RN3p
         *RS346pe*RSp45613e + 1./27.*m33*mp2*RN3p*RS346pe*RSp45633e - 1.
         /27.*m33*mp2*RN3p*RS356pe*RSp45613e + 1./27.*m33*mp2*RN3p*
         RS356pe*RSp45633e + 1./27.*m33*me2*RN3e*RSp45633e*RSe45613p - 
         1./27.*m33*me2*RN3e*RSp45633e*RSe45633p - 1./81.*m33*m66*RN3p*
         RSp456e*RSp453e*RS345pe - 1./81.*m33*m66*RN3p*RSp456e*RSp451e*
         RS345pe - 1./81.*m33*m66*RN3e*RSp456e*RSe453p*RS345pe - 1./81.
         *m33*m66*RN3e*RSp456e*RSe451p*RS345pe + 1./81.*m33*m66*RN13*
         RSp456e*RS145pe*RS345pe - 1./81.*m33*m55*RN3p*RSp465e*RSp463e*
         RS346pe );

      M13p6x23R +=  + Ab(m33,mu2)*Ab(mp2,mu2) * (  - 1./81.*m33*m55*
         RN3p*RSp465e*RSp461e*RS346pe - 1./81.*m33*m55*RN3e*RSp465e*
         RSe463p*RS346pe - 1./81.*m33*m55*RN3e*RSp465e*RSe461p*RS346pe
          + 1./81.*m33*m55*RN13*RSp465e*RS146pe*RS346pe - 1./81.*m33*
         m44*RN3p*RSp564e*RSp563e*RS356pe - 1./81.*m33*m44*RN3p*RSp564e
         *RSp561e*RS356pe - 1./81.*m33*m44*RN3e*RSp564e*RSe563p*RS356pe
          - 1./81.*m33*m44*RN3e*RSp564e*RSe561p*RS356pe + 1./81.*m33*
         m44*RN13*RSp564e*RS156pe*RS356pe + 1./27.*pow(m33,2)*RN3p*
         RS3456pee*RSp45613e + 1./27.*pow(m33,2)*RN3p*RS3456ppe*
         RSp45613e + 1./27.*pow(m33,2)*RN3p*RSp45633e*RSp45613e + 1./27.
         *pow(m33,2)*RN3p*RSp45633e*RS3456pee + 1./27.*pow(m33,2)*RN3p*
         RSp45633e*RS3456ppe + 1./27.*pow(m33,2)*RN3p*pow(RSp45633e,2)
          - 1./27.*pow(m33,2)*RN3p*RS345pe*RSp45613e - 1./27.*pow(
         m33,2)*RN3p*RS345pe*RSp45633e - 1./27.*pow(m33,2)*RN3p*RS346pe
         *RSp45613e - 1./27.*pow(m33,2)*RN3p*RS346pe*RSp45633e - 1./27.
         *pow(m33,2)*RN3p*RS356pe*RSp45613e );

      M13p6x23R +=  + Ab(m33,mu2)*Ab(mp2,mu2) * (  - 1./27.*pow(m33,2)*
         RN3p*RS356pe*RSp45633e + 1./27.*pow(m33,2)*RN3e*RSp45633e*
         RSe45613p + 1./27.*pow(m33,2)*RN3e*RSp45633e*RSe45633p - 1./27.
         *pow(m33,2)*RN13*RSp45633e*RS14563pe + 13./270.*m11*RSp45613e*
         RS34561pe - 1./540.*m11*RS3456pee*RSp45613e - 1./540.*m11*
         RS3456ppe*RSp45613e + 1./540.*m11*RSp45633e*RS34561pe - 1./72.
         *m11*RSp45633e*RS3456pee - 1./72.*m11*RSp45633e*RS3456ppe + 1./
         540.*m11*RS345pe*RSp45613e + 1./72.*m11*RS345pe*RSp45633e + 1./
         540.*m11*RS346pe*RSp45613e + 1./72.*m11*RS346pe*RSp45633e + 1./
         540.*m11*RS356pe*RSp45613e + 1./72.*m11*RS356pe*RSp45633e + 13.
         /540.*m11*RSp45611e*RS34561pe + 17./1080.*m11*RSp45611e*
         RS3456pee + 17./1080.*m11*RSp45611e*RS3456ppe - 17./1080.*m11*
         RSp45611e*RS345pe - 17./1080.*m11*RSp45611e*RS346pe - 17./1080.
         *m11*RSp45611e*RS356pe - 1./81.*m11*m66*RN3p*RSp456e*RSp451e*
         RS345pe - 1./81.*m11*m66*RN3e*RSp456e*RSe451p*RS345pe + 2./81.
         *m11*m66*RN13*RSp456e*RS145pe*RS345pe );

      M13p6x23R +=  + Ab(m33,mu2)*Ab(mp2,mu2) * (  - 1./81.*m11*m55*
         RN3p*RSp465e*RSp461e*RS346pe - 1./81.*m11*m55*RN3e*RSp465e*
         RSe461p*RS346pe + 2./81.*m11*m55*RN13*RSp465e*RS146pe*RS346pe
          - 1./81.*m11*m44*RN3p*RSp564e*RSp561e*RS356pe - 1./81.*m11*
         m44*RN3e*RSp564e*RSe561p*RS356pe + 2./81.*m11*m44*RN13*RSp564e
         *RS156pe*RS356pe + 1./27.*m11*m33*RN3p*RS3456pee*RSp45613e + 1.
         /27.*m11*m33*RN3p*RS3456ppe*RSp45613e + 1./27.*m11*m33*RN3p*
         RSp45633e*RSp45613e - 1./27.*m11*m33*RN3p*RS345pe*RSp45613e - 
         1./27.*m11*m33*RN3p*RS346pe*RSp45613e - 1./27.*m11*m33*RN3p*
         RS356pe*RSp45613e + 1./27.*m11*m33*RN3e*RSp45633e*RSe45613p - 
         2./27.*m11*m33*RN13*RSp45633e*RS14563pe - 1./27.*m11*m33*RN13*
         RSp45611e*RS34561pe - 2./27.*pow(m11,2)*RN13*RSp45611e*
         RS34561pe );

      M13p6x23R +=  + Ab(m33,mu2)*Bb(m11,mu2) * ( 1./540.*RD1456ep*
         RD3456ep - 1./540.*m33*RD1456ep*RS34561pe + 1./72.*m33*
         RD1456ep*RS3456pee + 1./72.*m33*RD1456ep*RS3456ppe - 1./72.*
         m33*RD1456ep*RS345pe - 1./72.*m33*RD1456ep*RS346pe - 1./72.*
         m33*RD1456ep*RS356pe + 11./540.*m11*RD1456ep*RS34561pe + 1./72.
         *m11*RD1456ep*RS3456pee + 1./72.*m11*RD1456ep*RS3456ppe - 1./
         72.*m11*RD1456ep*RS345pe - 1./72.*m11*RD1456ep*RS346pe - 1./72.
         *m11*RD1456ep*RS356pe - 1./27.*m11*m33*RN13*RD1456ep*RS34561pe
          - 2./27.*pow(m11,2)*RN13*RD1456ep*RS34561pe );

      M13p6x23R +=  + Ab(m33,mu2)*Bb(m33,mu2) * ( 1./540.*pow(
         RD3456ep,2) + 1./60.*m33*RD3456ep*RS34561pe + 1./27.*m33*
         RD3456ep*RS14563pe + 1./27.*m33*RD3456ep*RSe45613p + 1./27.*
         m33*RD3456ep*RSp45613e + 5./216.*m33*RS3456pee*RD3456ep + 5./
         216.*m33*RS3456ppe*RD3456ep - 1./27.*m33*RSe45633p*RD3456ep - 
         1./27.*m33*RSp45633e*RD3456ep - 5./216.*m33*RS345pe*RD3456ep
          - 5./216.*m33*RS346pe*RD3456ep - 5./216.*m33*RS356pe*RD3456ep
          + 1./27.*m33*mp2*RN3p*RD3456ep*RSp45613e - 1./27.*m33*mp2*
         RN3p*RSp45633e*RD3456ep + 1./27.*m33*me2*RN3e*RD3456ep*
         RSe45613p - 1./27.*m33*me2*RN3e*RSe45633p*RD3456ep - 2./27.*
         pow(m33,2)*RS3456pee*RS34561pe - 2./27.*pow(m33,2)*RS3456ppe*
         RS34561pe + 2./27.*pow(m33,2)*RS345pe*RS34561pe + 2./27.*pow(
         m33,2)*RS346pe*RS34561pe + 2./27.*pow(m33,2)*RS356pe*RS34561pe
          + 1./27.*pow(m33,2)*RN3p*RD3456ep*RSp45613e + 1./27.*pow(
         m33,2)*RN3p*RSp45633e*RD3456ep + 1./27.*pow(m33,2)*RN3e*
         RD3456ep*RSe45613p );

      M13p6x23R +=  + Ab(m33,mu2)*Bb(m33,mu2) * ( 1./27.*pow(m33,2)*
         RN3e*RSe45633p*RD3456ep - 1./27.*pow(m33,2)*RN13*RD3456ep*
         RS14563pe + 1./540.*m11*RD3456ep*RS34561pe - 1./72.*m11*
         RS3456pee*RD3456ep - 1./72.*m11*RS3456ppe*RD3456ep + 1./72.*
         m11*RS345pe*RD3456ep + 1./72.*m11*RS346pe*RD3456ep + 1./72.*
         m11*RS356pe*RD3456ep - 1./27.*m11*m33*RS3456pee*RS34561pe - 1./
         27.*m11*m33*RS3456ppe*RS34561pe + 1./27.*m11*m33*RS345pe*
         RS34561pe + 1./27.*m11*m33*RS346pe*RS34561pe + 1./27.*m11*m33*
         RS356pe*RS34561pe + 1./27.*m11*m33*RN3p*RD3456ep*RSp45613e + 1.
         /27.*m11*m33*RN3e*RD3456ep*RSe45613p - 2./27.*m11*m33*RN13*
         RD3456ep*RS14563pe );

      M13p6x23R +=  + Ab(m34,mu2)*pi16 * (  - 1./48.*m33*RD1456ep - 1./
         48.*m33*mp2*RSp45611e + 1./24.*m33*mp2*RSp561e - 1./48.*m33*
         mp2*RSp564e - 1./48.*m33*me2*RSe45611p + 1./24.*m33*me2*
         RSe561p - 1./48.*m33*me2*RSe564p + 1./16.*m33*m66 + 1./16.*m33
         *m55 + 1./8.*m33*m44 + 1./96.*m33*m44*RS34561pe + 1./96.*m33*
         m44*RS14563pe + 1./96.*m33*m44*RSe45613p + 1./96.*m33*m44*
         RSp45613e + 1./96.*m33*m44*RS3456pee + 1./96.*m33*m44*
         RS3456ppe - 1./96.*m33*m44*RSe45633p - 1./96.*m33*m44*
         RSp45633e - 1./96.*m33*m44*RS345pe - 1./96.*m33*m44*RS346pe + 
         1./96.*m33*m44*RSe563p + 1./96.*m33*m44*RSp563e + 1./96.*m33*
         m44*RS1456pee + 1./96.*m33*m44*RS1456ppe - 1./96.*m33*m44*
         RSe45611p - 1./96.*m33*m44*RSp45611e - 1./96.*m33*m44*RS145pe
          - 1./96.*m33*m44*RS146pe - 1./32.*m33*m44*RS456pe + 1./96.*
         m33*m44*RSe561p + 1./96.*m33*m44*RSp561e - 1./96.*m33*m44*
         RSe564p - 1./96.*m33*m44*RSp564e - 1./16.*pow(m33,2) + 1./96.*
         pow(m33,2)*RS34561pe );

      M13p6x23R +=  + Ab(m34,mu2)*pi16 * ( 1./96.*pow(m33,2)*RS14563pe
          + 1./96.*pow(m33,2)*RSe45613p + 1./96.*pow(m33,2)*RSp45613e
          + 1./96.*pow(m33,2)*RS3456pee + 1./96.*pow(m33,2)*RS3456ppe
          - 1./96.*pow(m33,2)*RSe45633p - 1./96.*pow(m33,2)*RSp45633e
          - 1./96.*pow(m33,2)*RS345pe - 1./96.*pow(m33,2)*RS346pe + 1./
         96.*pow(m33,2)*RSe563p + 1./96.*pow(m33,2)*RSp563e - 1./288.*
         pow(m33,2)*RS1456pee - 1./288.*pow(m33,2)*RS1456ppe + 1./288.*
         pow(m33,2)*RSe45611p + 1./288.*pow(m33,2)*RSp45611e + 1./288.*
         pow(m33,2)*RS145pe + 1./288.*pow(m33,2)*RS146pe - 1./72.*pow(
         m33,2)*RS156pe + 1./288.*pow(m33,2)*RS456pe - 5./288.*pow(
         m33,2)*RSe561p - 5./288.*pow(m33,2)*RSp561e + 1./288.*pow(
         m33,2)*RSe564p + 1./288.*pow(m33,2)*RSp564e - 1./48.*m11*
         RD1456ep - 1./48.*m11*mp2*RSp45611e + 1./24.*m11*mp2*RSp561e
          - 1./48.*m11*mp2*RSp564e - 1./48.*m11*me2*RSe45611p + 1./24.*
         m11*me2*RSe561p - 1./48.*m11*me2*RSe564p + 1./16.*m11*m66 + 1./
         16.*m11*m55 );

      M13p6x23R +=  + Ab(m34,mu2)*pi16 * ( 1./8.*m11*m44 + 1./96.*m11*
         m44*RS34561pe + 1./96.*m11*m44*RS14563pe + 1./96.*m11*m44*
         RSe45613p + 1./96.*m11*m44*RSp45613e + 1./96.*m11*m44*
         RS3456pee + 1./96.*m11*m44*RS3456ppe - 1./96.*m11*m44*
         RSe45633p - 1./96.*m11*m44*RSp45633e - 1./96.*m11*m44*RS345pe
          - 1./96.*m11*m44*RS346pe + 1./96.*m11*m44*RSe563p + 1./96.*
         m11*m44*RSp563e + 1./96.*m11*m44*RS1456pee + 1./96.*m11*m44*
         RS1456ppe - 1./96.*m11*m44*RSe45611p - 1./96.*m11*m44*
         RSp45611e - 1./96.*m11*m44*RS145pe - 1./96.*m11*m44*RS146pe - 
         1./32.*m11*m44*RS456pe + 1./96.*m11*m44*RSe561p + 1./96.*m11*
         m44*RSp561e - 1./96.*m11*m44*RSe564p - 1./96.*m11*m44*RSp564e
          + 1./96.*m11*m33*RS34561pe + 1./96.*m11*m33*RS14563pe + 1./96.
         *m11*m33*RSe45613p + 1./96.*m11*m33*RSp45613e + 1./96.*m11*m33
         *RS3456pee + 1./96.*m11*m33*RS3456ppe - 1./96.*m11*m33*
         RSe45633p - 1./96.*m11*m33*RSp45633e - 1./96.*m11*m33*RS345pe
          - 1./96.*m11*m33*RS346pe );

      M13p6x23R +=  + Ab(m34,mu2)*pi16 * ( 1./96.*m11*m33*RSe563p + 1./
         96.*m11*m33*RSp563e + 7./288.*m11*m33*RS1456pee + 7./288.*m11*
         m33*RS1456ppe - 1./288.*m11*m33*RSe45611p - 1./288.*m11*m33*
         RSp45611e - 7./288.*m11*m33*RS145pe - 7./288.*m11*m33*RS146pe
          + 1./72.*m11*m33*RS156pe - 1./288.*m11*m33*RS456pe - 1./288.*
         m11*m33*RSe561p - 1./288.*m11*m33*RSp561e - 1./288.*m11*m33*
         RSe564p - 1./288.*m11*m33*RSp564e + 1./16.*pow(m11,2) + 1./36.
         *pow(m11,2)*RS1456pee + 1./36.*pow(m11,2)*RS1456ppe - 1./144.*
         pow(m11,2)*RSe45611p - 1./144.*pow(m11,2)*RSp45611e - 1./36.*
         pow(m11,2)*RS145pe - 1./36.*pow(m11,2)*RS146pe + 1./36.*pow(
         m11,2)*RS156pe - 1./144.*pow(m11,2)*RS456pe + 1./72.*pow(
         m11,2)*RSe561p + 1./72.*pow(m11,2)*RSp561e - 1./144.*pow(
         m11,2)*RSe564p - 1./144.*pow(m11,2)*RSp564e );

      M13p6x23R +=  + Ab(m34,mu2)*Ab(me2,mu2) * ( 11./135.*me2*
         RSe45613p - 29./540.*me2*RSe45633p + 7./270.*me2*RSe563p - 1./
         180.*me2*RSe45611p - 19./270.*me2*RSe561p + 1./45.*me2*RSe564p
          - 11./270.*m44*RSe563p + 17./270.*m44*RSe561p - 1./45.*m44*
         RSe564p - 1./27.*m44*me2*RN3e*RSe563p + 1./27.*m44*me2*RN3e*
         RSe561p - 19./1080.*m33*RSe45613p + 1./60.*m33*RSe45633p - 49./
         1080.*m33*RSe563p + 1./24.*m33*RSe45611p + 1./216.*m33*RSe561p
          - 1./27.*m33*me2*RN3e*RSe45613p - 1./27.*m33*me2*RN3e*
         RSe45633p - 1./27.*m33*me2*RN3e*RSe563p + 1./9.*m33*me2*RN3e*
         RSe561p + 1./27.*m33*m44*RN3e*RSe563p + 1./27.*m33*m44*RN3e*
         RSe561p + 1./27.*pow(m33,2)*RN3e*RSe45613p + 1./27.*pow(m33,2)
         *RN3e*RSe45633p + 1./27.*pow(m33,2)*RN3e*RSe563p + 1./27.*pow(
         m33,2)*RN3e*RSe561p + 1./1080.*m11*RSe45613p - 1./24.*m11*
         RSe563p + 17./360.*m11*RSe45611p - 7./1080.*m11*RSe561p - 1./
         27.*m11*me2*RN3e*RSe45613p + 1./27.*m11*me2*RN3e*RSe561p + 1./
         27.*m11*m44*RN3e*RSe561p );

      M13p6x23R +=  + Ab(m34,mu2)*Ab(me2,mu2) * ( 1./27.*m11*m33*RN3e*
         RSe45613p + 1./27.*m11*m33*RN3e*RSe561p );

      M13p6x23R +=  + Ab(m34,mu2)*Ab(mp2,mu2) * ( 11./135.*mp2*
         RSp45613e - 29./540.*mp2*RSp45633e + 7./270.*mp2*RSp563e - 1./
         180.*mp2*RSp45611e - 19./270.*mp2*RSp561e + 1./45.*mp2*RSp564e
          - 11./270.*m44*RSp563e + 17./270.*m44*RSp561e - 1./45.*m44*
         RSp564e - 1./27.*m44*mp2*RN3p*RSp563e + 1./27.*m44*mp2*RN3p*
         RSp561e - 19./1080.*m33*RSp45613e + 1./60.*m33*RSp45633e - 49./
         1080.*m33*RSp563e + 1./24.*m33*RSp45611e + 1./216.*m33*RSp561e
          - 1./27.*m33*mp2*RN3p*RSp45613e - 1./27.*m33*mp2*RN3p*
         RSp45633e - 1./27.*m33*mp2*RN3p*RSp563e + 1./9.*m33*mp2*RN3p*
         RSp561e + 1./27.*m33*m44*RN3p*RSp563e + 1./27.*m33*m44*RN3p*
         RSp561e + 1./27.*pow(m33,2)*RN3p*RSp45613e + 1./27.*pow(m33,2)
         *RN3p*RSp45633e + 1./27.*pow(m33,2)*RN3p*RSp563e + 1./27.*pow(
         m33,2)*RN3p*RSp561e + 1./1080.*m11*RSp45613e - 1./24.*m11*
         RSp563e + 17./360.*m11*RSp45611e - 7./1080.*m11*RSp561e - 1./
         27.*m11*mp2*RN3p*RSp45613e + 1./27.*m11*mp2*RN3p*RSp561e + 1./
         27.*m11*m44*RN3p*RSp561e );

      M13p6x23R +=  + Ab(m34,mu2)*Ab(mp2,mu2) * ( 1./27.*m11*m33*RN3p*
         RSp45613e + 1./27.*m11*m33*RN3p*RSp561e );

      M13p6x23R +=  + Ab(m34,mu2)*Bb(m11,mu2) * ( 1./24.*m33*RD1456ep
          + 1./24.*m11*RD1456ep );

      M13p6x23R +=  + Ab(m35,mu2)*pi16 * (  - 1./48.*m33*RD1456ep - 1./
         48.*m33*mp2*RSp45611e + 1./24.*m33*mp2*RSp461e - 1./48.*m33*
         mp2*RSp465e - 1./48.*m33*me2*RSe45611p + 1./24.*m33*me2*
         RSe461p - 1./48.*m33*me2*RSe465p + 1./16.*m33*m66 + 1./8.*m33*
         m55 + 1./96.*m33*m55*RS34561pe + 1./96.*m33*m55*RS14563pe + 1./
         96.*m33*m55*RSe45613p + 1./96.*m33*m55*RSp45613e + 1./96.*m33*
         m55*RS3456pee + 1./96.*m33*m55*RS3456ppe - 1./96.*m33*m55*
         RSe45633p - 1./96.*m33*m55*RSp45633e - 1./96.*m33*m55*RS345pe
          - 1./96.*m33*m55*RS356pe + 1./96.*m33*m55*RSe463p + 1./96.*
         m33*m55*RSp463e + 1./96.*m33*m55*RS1456pee + 1./96.*m33*m55*
         RS1456ppe - 1./96.*m33*m55*RSe45611p - 1./96.*m33*m55*
         RSp45611e - 1./96.*m33*m55*RS145pe - 1./96.*m33*m55*RS156pe - 
         1./32.*m33*m55*RS546pe + 1./96.*m33*m55*RSe461p + 1./96.*m33*
         m55*RSp461e - 1./96.*m33*m55*RSe465p - 1./96.*m33*m55*RSp465e
          + 1./16.*m33*m44 - 1./16.*pow(m33,2) + 1./96.*pow(m33,2)*
         RS34561pe );

      M13p6x23R +=  + Ab(m35,mu2)*pi16 * ( 1./96.*pow(m33,2)*RS14563pe
          + 1./96.*pow(m33,2)*RSe45613p + 1./96.*pow(m33,2)*RSp45613e
          + 1./96.*pow(m33,2)*RS3456pee + 1./96.*pow(m33,2)*RS3456ppe
          - 1./96.*pow(m33,2)*RSe45633p - 1./96.*pow(m33,2)*RSp45633e
          - 1./96.*pow(m33,2)*RS345pe - 1./96.*pow(m33,2)*RS356pe + 1./
         96.*pow(m33,2)*RSe463p + 1./96.*pow(m33,2)*RSp463e - 1./288.*
         pow(m33,2)*RS1456pee - 1./288.*pow(m33,2)*RS1456ppe + 1./288.*
         pow(m33,2)*RSe45611p + 1./288.*pow(m33,2)*RSp45611e + 1./288.*
         pow(m33,2)*RS145pe - 1./72.*pow(m33,2)*RS146pe + 1./288.*pow(
         m33,2)*RS156pe + 1./288.*pow(m33,2)*RS546pe - 5./288.*pow(
         m33,2)*RSe461p - 5./288.*pow(m33,2)*RSp461e + 1./288.*pow(
         m33,2)*RSe465p + 1./288.*pow(m33,2)*RSp465e - 1./48.*m11*
         RD1456ep - 1./48.*m11*mp2*RSp45611e + 1./24.*m11*mp2*RSp461e
          - 1./48.*m11*mp2*RSp465e - 1./48.*m11*me2*RSe45611p + 1./24.*
         m11*me2*RSe461p - 1./48.*m11*me2*RSe465p + 1./16.*m11*m66 + 1./
         8.*m11*m55 );

      M13p6x23R +=  + Ab(m35,mu2)*pi16 * ( 1./96.*m11*m55*RS34561pe + 1.
         /96.*m11*m55*RS14563pe + 1./96.*m11*m55*RSe45613p + 1./96.*m11
         *m55*RSp45613e + 1./96.*m11*m55*RS3456pee + 1./96.*m11*m55*
         RS3456ppe - 1./96.*m11*m55*RSe45633p - 1./96.*m11*m55*
         RSp45633e - 1./96.*m11*m55*RS345pe - 1./96.*m11*m55*RS356pe + 
         1./96.*m11*m55*RSe463p + 1./96.*m11*m55*RSp463e + 1./96.*m11*
         m55*RS1456pee + 1./96.*m11*m55*RS1456ppe - 1./96.*m11*m55*
         RSe45611p - 1./96.*m11*m55*RSp45611e - 1./96.*m11*m55*RS145pe
          - 1./96.*m11*m55*RS156pe - 1./32.*m11*m55*RS546pe + 1./96.*
         m11*m55*RSe461p + 1./96.*m11*m55*RSp461e - 1./96.*m11*m55*
         RSe465p - 1./96.*m11*m55*RSp465e + 1./16.*m11*m44 + 1./96.*m11
         *m33*RS34561pe + 1./96.*m11*m33*RS14563pe + 1./96.*m11*m33*
         RSe45613p + 1./96.*m11*m33*RSp45613e + 1./96.*m11*m33*
         RS3456pee + 1./96.*m11*m33*RS3456ppe - 1./96.*m11*m33*
         RSe45633p - 1./96.*m11*m33*RSp45633e - 1./96.*m11*m33*RS345pe
          - 1./96.*m11*m33*RS356pe );

      M13p6x23R +=  + Ab(m35,mu2)*pi16 * ( 1./96.*m11*m33*RSe463p + 1./
         96.*m11*m33*RSp463e + 7./288.*m11*m33*RS1456pee + 7./288.*m11*
         m33*RS1456ppe - 1./288.*m11*m33*RSe45611p - 1./288.*m11*m33*
         RSp45611e - 7./288.*m11*m33*RS145pe + 1./72.*m11*m33*RS146pe
          - 7./288.*m11*m33*RS156pe - 1./288.*m11*m33*RS546pe - 1./288.
         *m11*m33*RSe461p - 1./288.*m11*m33*RSp461e - 1./288.*m11*m33*
         RSe465p - 1./288.*m11*m33*RSp465e + 1./16.*pow(m11,2) + 1./36.
         *pow(m11,2)*RS1456pee + 1./36.*pow(m11,2)*RS1456ppe - 1./144.*
         pow(m11,2)*RSe45611p - 1./144.*pow(m11,2)*RSp45611e - 1./36.*
         pow(m11,2)*RS145pe + 1./36.*pow(m11,2)*RS146pe - 1./36.*pow(
         m11,2)*RS156pe - 1./144.*pow(m11,2)*RS546pe + 1./72.*pow(
         m11,2)*RSe461p + 1./72.*pow(m11,2)*RSp461e - 1./144.*pow(
         m11,2)*RSe465p - 1./144.*pow(m11,2)*RSp465e );

      M13p6x23R +=  + Ab(m35,mu2)*Ab(me2,mu2) * ( 11./135.*me2*
         RSe45613p - 29./540.*me2*RSe45633p + 7./270.*me2*RSe463p - 1./
         180.*me2*RSe45611p - 19./270.*me2*RSe461p + 1./45.*me2*RSe465p
          - 11./270.*m55*RSe463p + 17./270.*m55*RSe461p - 1./45.*m55*
         RSe465p - 1./27.*m55*me2*RN3e*RSe463p + 1./27.*m55*me2*RN3e*
         RSe461p - 19./1080.*m33*RSe45613p + 1./60.*m33*RSe45633p - 49./
         1080.*m33*RSe463p + 1./24.*m33*RSe45611p + 1./216.*m33*RSe461p
          - 1./27.*m33*me2*RN3e*RSe45613p - 1./27.*m33*me2*RN3e*
         RSe45633p - 1./27.*m33*me2*RN3e*RSe463p + 1./9.*m33*me2*RN3e*
         RSe461p + 1./27.*m33*m55*RN3e*RSe463p + 1./27.*m33*m55*RN3e*
         RSe461p + 1./27.*pow(m33,2)*RN3e*RSe45613p + 1./27.*pow(m33,2)
         *RN3e*RSe45633p + 1./27.*pow(m33,2)*RN3e*RSe463p + 1./27.*pow(
         m33,2)*RN3e*RSe461p + 1./1080.*m11*RSe45613p - 1./24.*m11*
         RSe463p + 17./360.*m11*RSe45611p - 7./1080.*m11*RSe461p - 1./
         27.*m11*me2*RN3e*RSe45613p + 1./27.*m11*me2*RN3e*RSe461p + 1./
         27.*m11*m55*RN3e*RSe461p );

      M13p6x23R +=  + Ab(m35,mu2)*Ab(me2,mu2) * ( 1./27.*m11*m33*RN3e*
         RSe45613p + 1./27.*m11*m33*RN3e*RSe461p );

      M13p6x23R +=  + Ab(m35,mu2)*Ab(mp2,mu2) * ( 11./135.*mp2*
         RSp45613e - 29./540.*mp2*RSp45633e + 7./270.*mp2*RSp463e - 1./
         180.*mp2*RSp45611e - 19./270.*mp2*RSp461e + 1./45.*mp2*RSp465e
          - 11./270.*m55*RSp463e + 17./270.*m55*RSp461e - 1./45.*m55*
         RSp465e - 1./27.*m55*mp2*RN3p*RSp463e + 1./27.*m55*mp2*RN3p*
         RSp461e - 19./1080.*m33*RSp45613e + 1./60.*m33*RSp45633e - 49./
         1080.*m33*RSp463e + 1./24.*m33*RSp45611e + 1./216.*m33*RSp461e
          - 1./27.*m33*mp2*RN3p*RSp45613e - 1./27.*m33*mp2*RN3p*
         RSp45633e - 1./27.*m33*mp2*RN3p*RSp463e + 1./9.*m33*mp2*RN3p*
         RSp461e + 1./27.*m33*m55*RN3p*RSp463e + 1./27.*m33*m55*RN3p*
         RSp461e + 1./27.*pow(m33,2)*RN3p*RSp45613e + 1./27.*pow(m33,2)
         *RN3p*RSp45633e + 1./27.*pow(m33,2)*RN3p*RSp463e + 1./27.*pow(
         m33,2)*RN3p*RSp461e + 1./1080.*m11*RSp45613e - 1./24.*m11*
         RSp463e + 17./360.*m11*RSp45611e - 7./1080.*m11*RSp461e - 1./
         27.*m11*mp2*RN3p*RSp45613e + 1./27.*m11*mp2*RN3p*RSp461e + 1./
         27.*m11*m55*RN3p*RSp461e );

      M13p6x23R +=  + Ab(m35,mu2)*Ab(mp2,mu2) * ( 1./27.*m11*m33*RN3p*
         RSp45613e + 1./27.*m11*m33*RN3p*RSp461e );

      M13p6x23R +=  + Ab(m35,mu2)*Bb(m11,mu2) * ( 1./24.*m33*RD1456ep
          + 1./24.*m11*RD1456ep );

      M13p6x23R +=  + Ab(m36,mu2)*pi16 * (  - 1./48.*m33*RD1456ep - 1./
         48.*m33*mp2*RSp45611e + 1./24.*m33*mp2*RSp451e - 1./48.*m33*
         mp2*RSp456e - 1./48.*m33*me2*RSe45611p + 1./24.*m33*me2*
         RSe451p - 1./48.*m33*me2*RSe456p + 1./8.*m33*m66 + 1./96.*m33*
         m66*RS34561pe + 1./96.*m33*m66*RS14563pe + 1./96.*m33*m66*
         RSe45613p + 1./96.*m33*m66*RSp45613e + 1./96.*m33*m66*
         RS3456pee + 1./96.*m33*m66*RS3456ppe - 1./96.*m33*m66*
         RSe45633p - 1./96.*m33*m66*RSp45633e - 1./96.*m33*m66*RS346pe
          - 1./96.*m33*m66*RS356pe + 1./96.*m33*m66*RSe453p + 1./96.*
         m33*m66*RSp453e + 1./96.*m33*m66*RS1456pee + 1./96.*m33*m66*
         RS1456ppe - 1./96.*m33*m66*RSe45611p - 1./96.*m33*m66*
         RSp45611e - 1./96.*m33*m66*RS146pe - 1./96.*m33*m66*RS156pe - 
         1./32.*m33*m66*RS645pe + 1./96.*m33*m66*RSe451p + 1./96.*m33*
         m66*RSp451e - 1./96.*m33*m66*RSe456p - 1./96.*m33*m66*RSp456e
          + 1./16.*m33*m55 + 1./16.*m33*m44 - 1./16.*pow(m33,2) + 1./96.
         *pow(m33,2)*RS34561pe );

      M13p6x23R +=  + Ab(m36,mu2)*pi16 * ( 1./96.*pow(m33,2)*RS14563pe
          + 1./96.*pow(m33,2)*RSe45613p + 1./96.*pow(m33,2)*RSp45613e
          + 1./96.*pow(m33,2)*RS3456pee + 1./96.*pow(m33,2)*RS3456ppe
          - 1./96.*pow(m33,2)*RSe45633p - 1./96.*pow(m33,2)*RSp45633e
          - 1./96.*pow(m33,2)*RS346pe - 1./96.*pow(m33,2)*RS356pe + 1./
         96.*pow(m33,2)*RSe453p + 1./96.*pow(m33,2)*RSp453e - 1./288.*
         pow(m33,2)*RS1456pee - 1./288.*pow(m33,2)*RS1456ppe + 1./288.*
         pow(m33,2)*RSe45611p + 1./288.*pow(m33,2)*RSp45611e - 1./72.*
         pow(m33,2)*RS145pe + 1./288.*pow(m33,2)*RS146pe + 1./288.*pow(
         m33,2)*RS156pe + 1./288.*pow(m33,2)*RS645pe - 5./288.*pow(
         m33,2)*RSe451p - 5./288.*pow(m33,2)*RSp451e + 1./288.*pow(
         m33,2)*RSe456p + 1./288.*pow(m33,2)*RSp456e - 1./48.*m11*
         RD1456ep - 1./48.*m11*mp2*RSp45611e + 1./24.*m11*mp2*RSp451e
          - 1./48.*m11*mp2*RSp456e - 1./48.*m11*me2*RSe45611p + 1./24.*
         m11*me2*RSe451p - 1./48.*m11*me2*RSe456p + 1./8.*m11*m66 + 1./
         96.*m11*m66*RS34561pe );

      M13p6x23R +=  + Ab(m36,mu2)*pi16 * ( 1./96.*m11*m66*RS14563pe + 1.
         /96.*m11*m66*RSe45613p + 1./96.*m11*m66*RSp45613e + 1./96.*m11
         *m66*RS3456pee + 1./96.*m11*m66*RS3456ppe - 1./96.*m11*m66*
         RSe45633p - 1./96.*m11*m66*RSp45633e - 1./96.*m11*m66*RS346pe
          - 1./96.*m11*m66*RS356pe + 1./96.*m11*m66*RSe453p + 1./96.*
         m11*m66*RSp453e + 1./96.*m11*m66*RS1456pee + 1./96.*m11*m66*
         RS1456ppe - 1./96.*m11*m66*RSe45611p - 1./96.*m11*m66*
         RSp45611e - 1./96.*m11*m66*RS146pe - 1./96.*m11*m66*RS156pe - 
         1./32.*m11*m66*RS645pe + 1./96.*m11*m66*RSe451p + 1./96.*m11*
         m66*RSp451e - 1./96.*m11*m66*RSe456p - 1./96.*m11*m66*RSp456e
          + 1./16.*m11*m55 + 1./16.*m11*m44 + 1./96.*m11*m33*RS34561pe
          + 1./96.*m11*m33*RS14563pe + 1./96.*m11*m33*RSe45613p + 1./96.
         *m11*m33*RSp45613e + 1./96.*m11*m33*RS3456pee + 1./96.*m11*m33
         *RS3456ppe - 1./96.*m11*m33*RSe45633p - 1./96.*m11*m33*
         RSp45633e - 1./96.*m11*m33*RS346pe - 1./96.*m11*m33*RS356pe + 
         1./96.*m11*m33*RSe453p );

      M13p6x23R +=  + Ab(m36,mu2)*pi16 * ( 1./96.*m11*m33*RSp453e + 7./
         288.*m11*m33*RS1456pee + 7./288.*m11*m33*RS1456ppe - 1./288.*
         m11*m33*RSe45611p - 1./288.*m11*m33*RSp45611e + 1./72.*m11*m33
         *RS145pe - 7./288.*m11*m33*RS146pe - 7./288.*m11*m33*RS156pe
          - 1./288.*m11*m33*RS645pe - 1./288.*m11*m33*RSe451p - 1./288.
         *m11*m33*RSp451e - 1./288.*m11*m33*RSe456p - 1./288.*m11*m33*
         RSp456e + 1./16.*pow(m11,2) + 1./36.*pow(m11,2)*RS1456pee + 1./
         36.*pow(m11,2)*RS1456ppe - 1./144.*pow(m11,2)*RSe45611p - 1./
         144.*pow(m11,2)*RSp45611e + 1./36.*pow(m11,2)*RS145pe - 1./36.
         *pow(m11,2)*RS146pe - 1./36.*pow(m11,2)*RS156pe - 1./144.*pow(
         m11,2)*RS645pe + 1./72.*pow(m11,2)*RSe451p + 1./72.*pow(m11,2)
         *RSp451e - 1./144.*pow(m11,2)*RSe456p - 1./144.*pow(m11,2)*
         RSp456e );

      M13p6x23R +=  + Ab(m36,mu2)*Ab(me2,mu2) * ( 11./135.*me2*
         RSe45613p - 29./540.*me2*RSe45633p + 7./270.*me2*RSe453p - 1./
         180.*me2*RSe45611p - 19./270.*me2*RSe451p + 1./45.*me2*RSe456p
          - 11./270.*m66*RSe453p + 17./270.*m66*RSe451p - 1./45.*m66*
         RSe456p - 1./27.*m66*me2*RN3e*RSe453p + 1./27.*m66*me2*RN3e*
         RSe451p - 19./1080.*m33*RSe45613p + 1./60.*m33*RSe45633p - 49./
         1080.*m33*RSe453p + 1./24.*m33*RSe45611p + 1./216.*m33*RSe451p
          - 1./27.*m33*me2*RN3e*RSe45613p - 1./27.*m33*me2*RN3e*
         RSe45633p - 1./27.*m33*me2*RN3e*RSe453p + 1./9.*m33*me2*RN3e*
         RSe451p + 1./27.*m33*m66*RN3e*RSe453p + 1./27.*m33*m66*RN3e*
         RSe451p + 1./27.*pow(m33,2)*RN3e*RSe45613p + 1./27.*pow(m33,2)
         *RN3e*RSe45633p + 1./27.*pow(m33,2)*RN3e*RSe453p + 1./27.*pow(
         m33,2)*RN3e*RSe451p + 1./1080.*m11*RSe45613p - 1./24.*m11*
         RSe453p + 17./360.*m11*RSe45611p - 7./1080.*m11*RSe451p - 1./
         27.*m11*me2*RN3e*RSe45613p + 1./27.*m11*me2*RN3e*RSe451p + 1./
         27.*m11*m66*RN3e*RSe451p );

      M13p6x23R +=  + Ab(m36,mu2)*Ab(me2,mu2) * ( 1./27.*m11*m33*RN3e*
         RSe45613p + 1./27.*m11*m33*RN3e*RSe451p );

      M13p6x23R +=  + Ab(m36,mu2)*Ab(mp2,mu2) * ( 11./135.*mp2*
         RSp45613e - 29./540.*mp2*RSp45633e + 7./270.*mp2*RSp453e - 1./
         180.*mp2*RSp45611e - 19./270.*mp2*RSp451e + 1./45.*mp2*RSp456e
          - 11./270.*m66*RSp453e + 17./270.*m66*RSp451e - 1./45.*m66*
         RSp456e - 1./27.*m66*mp2*RN3p*RSp453e + 1./27.*m66*mp2*RN3p*
         RSp451e - 19./1080.*m33*RSp45613e + 1./60.*m33*RSp45633e - 49./
         1080.*m33*RSp453e + 1./24.*m33*RSp45611e + 1./216.*m33*RSp451e
          - 1./27.*m33*mp2*RN3p*RSp45613e - 1./27.*m33*mp2*RN3p*
         RSp45633e - 1./27.*m33*mp2*RN3p*RSp453e + 1./9.*m33*mp2*RN3p*
         RSp451e + 1./27.*m33*m66*RN3p*RSp453e + 1./27.*m33*m66*RN3p*
         RSp451e + 1./27.*pow(m33,2)*RN3p*RSp45613e + 1./27.*pow(m33,2)
         *RN3p*RSp45633e + 1./27.*pow(m33,2)*RN3p*RSp453e + 1./27.*pow(
         m33,2)*RN3p*RSp451e + 1./1080.*m11*RSp45613e - 1./24.*m11*
         RSp453e + 17./360.*m11*RSp45611e - 7./1080.*m11*RSp451e - 1./
         27.*m11*mp2*RN3p*RSp45613e + 1./27.*m11*mp2*RN3p*RSp451e + 1./
         27.*m11*m66*RN3p*RSp451e );

      M13p6x23R +=  + Ab(m36,mu2)*Ab(mp2,mu2) * ( 1./27.*m11*m33*RN3p*
         RSp45613e + 1./27.*m11*m33*RN3p*RSp451e );

      M13p6x23R +=  + Ab(m36,mu2)*Bb(m11,mu2) * ( 1./24.*m33*RD1456ep
          + 1./24.*m11*RD1456ep );

      M13p6x23R +=  + Ab(m44,mu2)*Ab(me2,mu2) * (  - 1./54.*m44*pow(
         RSe563p,2) - 1./27.*m44*RSp563e*RSe563p + 1./162.*m44*RS456pe*
         pow(RSe563p,2) + 1./81.*m44*RS456pe*RSp563e*RSe563p + 1./27.*
         m44*RSe561p*RSe563p + 1./27.*m44*RSe561p*RSp563e - 1./81.*m44*
         RSe561p*RS456pe*RSe563p - 1./81.*m44*RSe561p*RS456pe*RSp563e
          - 1./54.*m44*pow(RSe561p,2) + 1./162.*m44*pow(RSe561p,2)*
         RS456pe + 1./27.*m44*RSp561e*RSe563p - 1./81.*m44*RSp561e*
         RS456pe*RSe563p - 1./27.*m44*RSp561e*RSe561p + 1./81.*m44*
         RSp561e*RSe561p*RS456pe - 1./27.*m44*mp2*RNep*RSp563e*RSe563p
          + 1./81.*m44*mp2*RNep*RS456pe*RSp563e*RSe563p + 1./27.*m44*
         mp2*RNep*RSe561p*RSp563e - 1./81.*m44*mp2*RNep*RSe561p*RS456pe
         *RSp563e + 1./27.*m44*mp2*RNep*RSp561e*RSe563p - 1./81.*m44*
         mp2*RNep*RSp561e*RS456pe*RSe563p - 1./27.*m44*mp2*RNep*RSp561e
         *RSe561p + 1./81.*m44*mp2*RNep*RSp561e*RSe561p*RS456pe + 1./27.
         *m44*me2*RN3e*RSe563p*RS356pe - 1./81.*m44*me2*RN3e*RS456pe*
         RSe563p*RS356pe );

      M13p6x23R +=  + Ab(m44,mu2)*Ab(me2,mu2) * (  - 1./27.*m44*me2*
         RN3e*RSe561p*RS356pe + 1./81.*m44*me2*RN3e*RSe561p*RS456pe*
         RS356pe - 1./27.*m44*me2*RN1e*RS156pe*RSe563p + 1./81.*m44*me2
         *RN1e*RS456pe*RS156pe*RSe563p + 1./27.*m44*me2*RN1e*RSe561p*
         RS156pe - 1./81.*m44*me2*RN1e*RSe561p*RS456pe*RS156pe + 1./27.
         *m33*m44*RNep*RSp563e*RSe563p - 1./81.*m33*m44*RNep*RS456pe*
         RSp563e*RSe563p + 1./27.*m33*m44*RNep*RSe561p*RSp563e - 1./81.
         *m33*m44*RNep*RSe561p*RS456pe*RSp563e + 1./27.*m33*m44*RNep*
         RSp561e*RSe563p - 1./81.*m33*m44*RNep*RSp561e*RS456pe*RSe563p
          - 1./27.*m33*m44*RN3e*RSe563p*RS356pe + 1./81.*m33*m44*RN3e*
         RS456pe*RSe563p*RS356pe - 1./27.*m33*m44*RN3e*RSe561p*RS356pe
          + 1./81.*m33*m44*RN3e*RSe561p*RS456pe*RS356pe - 1./27.*m33*
         m44*RN1e*RS156pe*RSe563p + 1./81.*m33*m44*RN1e*RS456pe*RS156pe
         *RSe563p + 1./27.*m11*m44*RNep*RSe561p*RSp563e - 1./81.*m11*
         m44*RNep*RSe561p*RS456pe*RSp563e + 1./27.*m11*m44*RNep*RSp561e
         *RSe563p );

      M13p6x23R +=  + Ab(m44,mu2)*Ab(me2,mu2) * (  - 1./81.*m11*m44*
         RNep*RSp561e*RS456pe*RSe563p + 1./27.*m11*m44*RNep*RSp561e*
         RSe561p - 1./81.*m11*m44*RNep*RSp561e*RSe561p*RS456pe - 1./27.
         *m11*m44*RN3e*RSe561p*RS356pe + 1./81.*m11*m44*RN3e*RSe561p*
         RS456pe*RS356pe - 1./27.*m11*m44*RN1e*RS156pe*RSe563p + 1./81.
         *m11*m44*RN1e*RS456pe*RS156pe*RSe563p - 1./27.*m11*m44*RN1e*
         RSe561p*RS156pe + 1./81.*m11*m44*RN1e*RSe561p*RS456pe*RS156pe
          );

      M13p6x23R +=  + Ab(m44,mu2)*Ab(mp2,mu2) * (  - 1./54.*m44*pow(
         RSp563e,2) + 1./162.*m44*RS456pe*pow(RSp563e,2) + 1./27.*m44*
         RSp561e*RSp563e - 1./81.*m44*RSp561e*RS456pe*RSp563e - 1./54.*
         m44*pow(RSp561e,2) + 1./162.*m44*pow(RSp561e,2)*RS456pe + 1./
         27.*m44*mp2*RNep*RSp563e*RSe563p - 1./81.*m44*mp2*RNep*RS456pe
         *RSp563e*RSe563p - 1./27.*m44*mp2*RNep*RSe561p*RSp563e + 1./81.
         *m44*mp2*RNep*RSe561p*RS456pe*RSp563e - 1./27.*m44*mp2*RNep*
         RSp561e*RSe563p + 1./81.*m44*mp2*RNep*RSp561e*RS456pe*RSe563p
          + 1./27.*m44*mp2*RNep*RSp561e*RSe561p - 1./81.*m44*mp2*RNep*
         RSp561e*RSe561p*RS456pe + 1./27.*m44*mp2*RN3p*RSp563e*RS356pe
          - 1./81.*m44*mp2*RN3p*RS456pe*RSp563e*RS356pe - 1./27.*m44*
         mp2*RN3p*RSp561e*RS356pe + 1./81.*m44*mp2*RN3p*RSp561e*RS456pe
         *RS356pe - 1./27.*m44*mp2*RN1p*RS156pe*RSp563e + 1./81.*m44*
         mp2*RN1p*RS456pe*RS156pe*RSp563e + 1./27.*m44*mp2*RN1p*RSp561e
         *RS156pe - 1./81.*m44*mp2*RN1p*RSp561e*RS456pe*RS156pe - 1./27.
         *m33*m44*RNep*RSp563e*RSe563p );

      M13p6x23R +=  + Ab(m44,mu2)*Ab(mp2,mu2) * ( 1./81.*m33*m44*RNep*
         RS456pe*RSp563e*RSe563p - 1./27.*m33*m44*RNep*RSe561p*RSp563e
          + 1./81.*m33*m44*RNep*RSe561p*RS456pe*RSp563e - 1./27.*m33*
         m44*RNep*RSp561e*RSe563p + 1./81.*m33*m44*RNep*RSp561e*RS456pe
         *RSe563p - 1./27.*m33*m44*RN3p*RSp563e*RS356pe + 1./81.*m33*
         m44*RN3p*RS456pe*RSp563e*RS356pe - 1./27.*m33*m44*RN3p*RSp561e
         *RS356pe + 1./81.*m33*m44*RN3p*RSp561e*RS456pe*RS356pe - 1./27.
         *m33*m44*RN1p*RS156pe*RSp563e + 1./81.*m33*m44*RN1p*RS456pe*
         RS156pe*RSp563e - 1./27.*m11*m44*RNep*RSe561p*RSp563e + 1./81.
         *m11*m44*RNep*RSe561p*RS456pe*RSp563e - 1./27.*m11*m44*RNep*
         RSp561e*RSe563p + 1./81.*m11*m44*RNep*RSp561e*RS456pe*RSe563p
          - 1./27.*m11*m44*RNep*RSp561e*RSe561p + 1./81.*m11*m44*RNep*
         RSp561e*RSe561p*RS456pe - 1./27.*m11*m44*RN3p*RSp561e*RS356pe
          + 1./81.*m11*m44*RN3p*RSp561e*RS456pe*RS356pe - 1./27.*m11*
         m44*RN1p*RS156pe*RSp563e + 1./81.*m11*m44*RN1p*RS456pe*RS156pe
         *RSp563e );

      M13p6x23R +=  + Ab(m44,mu2)*Ab(mp2,mu2) * (  - 1./27.*m11*m44*
         RN1p*RSp561e*RS156pe + 1./81.*m11*m44*RN1p*RSp561e*RS456pe*
         RS156pe );

      M13p6x23R +=  + Ab(m44,mu2)*Bb(me2,mu2) * (  - 1./54.*m44*me2*
         pow(RSe563p,2) + 1./162.*m44*me2*RS456pe*pow(RSe563p,2) + 1./
         27.*m44*me2*RSe561p*RSe563p - 1./81.*m44*me2*RSe561p*RS456pe*
         RSe563p - 1./54.*m44*me2*pow(RSe561p,2) + 1./162.*m44*me2*pow(
         RSe561p,2)*RS456pe + 1./54.*m33*m44*pow(RSe563p,2) - 1./162.*
         m33*m44*RS456pe*pow(RSe563p,2) + 1./27.*m33*m44*RSe561p*
         RSe563p - 1./81.*m33*m44*RSe561p*RS456pe*RSe563p + 1./27.*m11*
         m44*RSe561p*RSe563p - 1./81.*m11*m44*RSe561p*RS456pe*RSe563p
          + 1./54.*m11*m44*pow(RSe561p,2) - 1./162.*m11*m44*pow(
         RSe561p,2)*RS456pe );

      M13p6x23R +=  + Ab(m44,mu2)*Bb(mp2,mu2) * (  - 1./54.*m44*mp2*
         pow(RSp563e,2) + 1./162.*m44*mp2*RS456pe*pow(RSp563e,2) + 1./
         27.*m44*mp2*RSp561e*RSp563e - 1./81.*m44*mp2*RSp561e*RS456pe*
         RSp563e - 1./54.*m44*mp2*pow(RSp561e,2) + 1./162.*m44*mp2*pow(
         RSp561e,2)*RS456pe + 1./54.*m33*m44*pow(RSp563e,2) - 1./162.*
         m33*m44*RS456pe*pow(RSp563e,2) + 1./27.*m33*m44*RSp561e*
         RSp563e - 1./81.*m33*m44*RSp561e*RS456pe*RSp563e + 1./27.*m11*
         m44*RSp561e*RSp563e - 1./81.*m11*m44*RSp561e*RS456pe*RSp563e
          + 1./54.*m11*m44*pow(RSp561e,2) - 1./162.*m11*m44*pow(
         RSp561e,2)*RS456pe );

      M13p6x23R +=  + Ab(m45,mu2)*Ab(me2,mu2) * (  - 1./81.*me2*RSe563p
         *RS346pe + 1./81.*me2*RSe563p*RS356pe + 1./81.*me2*pow(
         RSe563p,2) + 1./81.*me2*RSe463p*RS346pe - 1./81.*me2*RSe463p*
         RS356pe - 2./81.*me2*RSe463p*RSe563p + 1./81.*me2*pow(
         RSe463p,2) + 1./81.*me2*RSp563e*RSe563p - 1./81.*me2*RSp563e*
         RSe463p - 1./81.*me2*RSp463e*RSe563p + 1./81.*me2*RSp463e*
         RSe463p + 1./81.*me2*RS146pe*RSe563p - 1./81.*me2*RS146pe*
         RSe463p - 1./81.*me2*RS156pe*RSe563p + 1./81.*me2*RS156pe*
         RSe463p + 1./81.*me2*RSe561p*RS346pe - 1./81.*me2*RSe561p*
         RS356pe - 2./81.*me2*RSe561p*RSe563p + 2./81.*me2*RSe561p*
         RSe463p - 1./81.*me2*RSe561p*RSp563e + 1./81.*me2*RSe561p*
         RSp463e - 1./81.*me2*RSe561p*RS146pe + 1./81.*me2*RSe561p*
         RS156pe + 1./81.*me2*pow(RSe561p,2) - 1./81.*me2*RSe461p*
         RS346pe + 1./81.*me2*RSe461p*RS356pe + 2./81.*me2*RSe461p*
         RSe563p - 2./81.*me2*RSe461p*RSe463p + 1./81.*me2*RSe461p*
         RSp563e );

      M13p6x23R +=  + Ab(m45,mu2)*Ab(me2,mu2) * (  - 1./81.*me2*RSe461p
         *RSp463e + 1./81.*me2*RSe461p*RS146pe - 1./81.*me2*RSe461p*
         RS156pe - 2./81.*me2*RSe461p*RSe561p + 1./81.*me2*pow(
         RSe461p,2) - 1./81.*me2*RSp561e*RSe563p + 1./81.*me2*RSp561e*
         RSe463p + 1./81.*me2*RSp561e*RSe561p - 1./81.*me2*RSp561e*
         RSe461p + 1./81.*me2*RSp461e*RSe563p - 1./81.*me2*RSp461e*
         RSe463p - 1./81.*me2*RSp461e*RSe561p + 1./81.*me2*RSp461e*
         RSe461p + 1./81.*me2*mp2*RNep*RSp563e*RSe563p - 1./81.*me2*mp2
         *RNep*RSp563e*RSe463p - 1./81.*me2*mp2*RNep*RSp463e*RSe563p + 
         1./81.*me2*mp2*RNep*RSp463e*RSe463p - 1./81.*me2*mp2*RNep*
         RSe561p*RSp563e + 1./81.*me2*mp2*RNep*RSe561p*RSp463e + 1./81.
         *me2*mp2*RNep*RSe461p*RSp563e - 1./81.*me2*mp2*RNep*RSe461p*
         RSp463e - 1./81.*me2*mp2*RNep*RSp561e*RSe563p + 1./81.*me2*mp2
         *RNep*RSp561e*RSe463p + 1./81.*me2*mp2*RNep*RSp561e*RSe561p - 
         1./81.*me2*mp2*RNep*RSp561e*RSe461p + 1./81.*me2*mp2*RNep*
         RSp461e*RSe563p );

      M13p6x23R +=  + Ab(m45,mu2)*Ab(me2,mu2) * (  - 1./81.*me2*mp2*
         RNep*RSp461e*RSe463p - 1./81.*me2*mp2*RNep*RSp461e*RSe561p + 1.
         /81.*me2*mp2*RNep*RSp461e*RSe461p - 1./81.*m55*RSe463p*RSe563p
          - 1./162.*m55*pow(RSe463p,2) - 1./81.*m55*RSp563e*RSe463p - 1.
         /81.*m55*RSp463e*RSe563p - 1./81.*m55*RSp463e*RSe463p + 1./81.
         *m55*RSe561p*RSe463p + 1./81.*m55*RSe561p*RSp463e + 1./81.*m55
         *RSe461p*RSe563p + 1./81.*m55*RSe461p*RSe463p + 1./81.*m55*
         RSe461p*RSp563e + 1./81.*m55*RSe461p*RSp463e - 1./81.*m55*
         RSe461p*RSe561p - 1./162.*m55*pow(RSe461p,2) + 1./81.*m55*
         RSp561e*RSe463p - 1./81.*m55*RSp561e*RSe461p + 1./81.*m55*
         RSp461e*RSe563p + 1./81.*m55*RSp461e*RSe463p - 1./81.*m55*
         RSp461e*RSe561p - 1./81.*m55*RSp461e*RSe461p - 1./81.*m55*mp2*
         RNep*RSp563e*RSe463p - 1./81.*m55*mp2*RNep*RSp463e*RSe563p - 1.
         /81.*m55*mp2*RNep*RSp463e*RSe463p + 1./81.*m55*mp2*RNep*
         RSe561p*RSp463e + 1./81.*m55*mp2*RNep*RSe461p*RSp563e + 1./81.
         *m55*mp2*RNep*RSe461p*RSp463e );

      M13p6x23R +=  + Ab(m45,mu2)*Ab(me2,mu2) * ( 1./81.*m55*mp2*RNep*
         RSp561e*RSe463p - 1./81.*m55*mp2*RNep*RSp561e*RSe461p + 1./81.
         *m55*mp2*RNep*RSp461e*RSe563p + 1./81.*m55*mp2*RNep*RSp461e*
         RSe463p - 1./81.*m55*mp2*RNep*RSp461e*RSe561p - 1./81.*m55*mp2
         *RNep*RSp461e*RSe461p + 1./81.*m55*me2*RN3e*RSe563p*RS346pe + 
         1./81.*m55*me2*RN3e*RSe463p*RS346pe + 1./81.*m55*me2*RN3e*
         RSe463p*RS356pe - 1./81.*m55*me2*RN3e*RSe561p*RS346pe - 1./81.
         *m55*me2*RN3e*RSe461p*RS346pe - 1./81.*m55*me2*RN3e*RSe461p*
         RS356pe - 1./81.*m55*me2*RN1e*RS146pe*RSe563p - 1./81.*m55*me2
         *RN1e*RS146pe*RSe463p - 1./81.*m55*me2*RN1e*RS156pe*RSe463p + 
         1./81.*m55*me2*RN1e*RSe561p*RS146pe + 1./81.*m55*me2*RN1e*
         RSe461p*RS146pe + 1./81.*m55*me2*RN1e*RSe461p*RS156pe - 1./162.
         *m44*pow(RSe563p,2) - 1./81.*m44*RSe463p*RSe563p - 1./81.*m44*
         RSp563e*RSe563p - 1./81.*m44*RSp563e*RSe463p - 1./81.*m44*
         RSp463e*RSe563p + 1./81.*m44*RSe561p*RSe563p + 1./81.*m44*
         RSe561p*RSe463p );

      M13p6x23R +=  + Ab(m45,mu2)*Ab(me2,mu2) * ( 1./81.*m44*RSe561p*
         RSp563e + 1./81.*m44*RSe561p*RSp463e - 1./162.*m44*pow(
         RSe561p,2) + 1./81.*m44*RSe461p*RSe563p + 1./81.*m44*RSe461p*
         RSp563e - 1./81.*m44*RSe461p*RSe561p + 1./81.*m44*RSp561e*
         RSe563p + 1./81.*m44*RSp561e*RSe463p - 1./81.*m44*RSp561e*
         RSe561p - 1./81.*m44*RSp561e*RSe461p + 1./81.*m44*RSp461e*
         RSe563p - 1./81.*m44*RSp461e*RSe561p - 1./81.*m44*mp2*RNep*
         RSp563e*RSe563p - 1./81.*m44*mp2*RNep*RSp563e*RSe463p - 1./81.
         *m44*mp2*RNep*RSp463e*RSe563p + 1./81.*m44*mp2*RNep*RSe561p*
         RSp563e + 1./81.*m44*mp2*RNep*RSe561p*RSp463e + 1./81.*m44*mp2
         *RNep*RSe461p*RSp563e + 1./81.*m44*mp2*RNep*RSp561e*RSe563p + 
         1./81.*m44*mp2*RNep*RSp561e*RSe463p - 1./81.*m44*mp2*RNep*
         RSp561e*RSe561p - 1./81.*m44*mp2*RNep*RSp561e*RSe461p + 1./81.
         *m44*mp2*RNep*RSp461e*RSe563p - 1./81.*m44*mp2*RNep*RSp461e*
         RSe561p + 1./81.*m44*me2*RN3e*RSe563p*RS346pe + 1./81.*m44*me2
         *RN3e*RSe563p*RS356pe );

      M13p6x23R +=  + Ab(m45,mu2)*Ab(me2,mu2) * ( 1./81.*m44*me2*RN3e*
         RSe463p*RS356pe - 1./81.*m44*me2*RN3e*RSe561p*RS346pe - 1./81.
         *m44*me2*RN3e*RSe561p*RS356pe - 1./81.*m44*me2*RN3e*RSe461p*
         RS356pe - 1./81.*m44*me2*RN1e*RS146pe*RSe563p - 1./81.*m44*me2
         *RN1e*RS156pe*RSe563p - 1./81.*m44*me2*RN1e*RS156pe*RSe463p + 
         1./81.*m44*me2*RN1e*RSe561p*RS146pe + 1./81.*m44*me2*RN1e*
         RSe561p*RS156pe + 1./81.*m44*me2*RN1e*RSe461p*RS156pe - 1./162.
         *m33*pow(RSe563p,2) + 1./81.*m33*RSe463p*RSe563p - 1./162.*m33
         *pow(RSe463p,2) - 1./81.*m33*RSp563e*RSe563p + 1./81.*m33*
         RSp563e*RSe463p + 1./81.*m33*RSp463e*RSe563p - 1./81.*m33*
         RSp463e*RSe463p - 1./81.*m33*RSe561p*RSe563p + 1./81.*m33*
         RSe561p*RSe463p - 1./81.*m33*RSe561p*RSp563e + 1./81.*m33*
         RSe561p*RSp463e + 1./81.*m33*RSe461p*RSe563p - 1./81.*m33*
         RSe461p*RSe463p + 1./81.*m33*RSe461p*RSp563e - 1./81.*m33*
         RSe461p*RSp463e - 1./81.*m33*RSp561e*RSe563p + 1./81.*m33*
         RSp561e*RSe463p );

      M13p6x23R +=  + Ab(m45,mu2)*Ab(me2,mu2) * ( 1./81.*m33*RSp461e*
         RSe563p - 1./81.*m33*RSp461e*RSe463p - 1./81.*m33*mp2*RNep*
         RSp563e*RSe563p + 1./81.*m33*mp2*RNep*RSp563e*RSe463p + 1./81.
         *m33*mp2*RNep*RSp463e*RSe563p - 1./81.*m33*mp2*RNep*RSp463e*
         RSe463p - 1./81.*m33*mp2*RNep*RSe561p*RSp563e + 1./81.*m33*mp2
         *RNep*RSe561p*RSp463e + 1./81.*m33*mp2*RNep*RSe461p*RSp563e - 
         1./81.*m33*mp2*RNep*RSe461p*RSp463e - 1./81.*m33*mp2*RNep*
         RSp561e*RSe563p + 1./81.*m33*mp2*RNep*RSp561e*RSe463p + 1./81.
         *m33*mp2*RNep*RSp461e*RSe563p - 1./81.*m33*mp2*RNep*RSp461e*
         RSe463p - 2./81.*m33*me2*RN3e*RSe561p*RS346pe + 2./81.*m33*me2
         *RN3e*RSe561p*RS356pe + 2./81.*m33*me2*RN3e*RSe461p*RS346pe - 
         2./81.*m33*me2*RN3e*RSe461p*RS356pe - 1./81.*m33*me2*RN1e*
         RS146pe*RSe563p + 1./81.*m33*me2*RN1e*RS146pe*RSe463p + 1./81.
         *m33*me2*RN1e*RS156pe*RSe563p - 1./81.*m33*me2*RN1e*RS156pe*
         RSe463p + 1./81.*m33*m55*RNep*RSp563e*RSe463p + 1./81.*m33*m55
         *RNep*RSp463e*RSe563p );

      M13p6x23R +=  + Ab(m45,mu2)*Ab(me2,mu2) * ( 1./81.*m33*m55*RNep*
         RSp463e*RSe463p + 1./81.*m33*m55*RNep*RSe561p*RSp463e + 1./81.
         *m33*m55*RNep*RSe461p*RSp563e + 1./81.*m33*m55*RNep*RSe461p*
         RSp463e + 1./81.*m33*m55*RNep*RSp561e*RSe463p + 1./81.*m33*m55
         *RNep*RSp461e*RSe563p + 1./81.*m33*m55*RNep*RSp461e*RSe463p - 
         1./81.*m33*m55*RN3e*RSe563p*RS346pe - 1./81.*m33*m55*RN3e*
         RSe463p*RS346pe - 1./81.*m33*m55*RN3e*RSe463p*RS356pe - 1./81.
         *m33*m55*RN3e*RSe561p*RS346pe - 1./81.*m33*m55*RN3e*RSe461p*
         RS346pe - 1./81.*m33*m55*RN3e*RSe461p*RS356pe - 1./81.*m33*m55
         *RN1e*RS146pe*RSe563p - 1./81.*m33*m55*RN1e*RS146pe*RSe463p - 
         1./81.*m33*m55*RN1e*RS156pe*RSe463p + 1./81.*m33*m44*RNep*
         RSp563e*RSe563p + 1./81.*m33*m44*RNep*RSp563e*RSe463p + 1./81.
         *m33*m44*RNep*RSp463e*RSe563p + 1./81.*m33*m44*RNep*RSe561p*
         RSp563e + 1./81.*m33*m44*RNep*RSe561p*RSp463e + 1./81.*m33*m44
         *RNep*RSe461p*RSp563e + 1./81.*m33*m44*RNep*RSp561e*RSe563p + 
         1./81.*m33*m44*RNep*RSp561e*RSe463p );

      M13p6x23R +=  + Ab(m45,mu2)*Ab(me2,mu2) * ( 1./81.*m33*m44*RNep*
         RSp461e*RSe563p - 1./81.*m33*m44*RN3e*RSe563p*RS346pe - 1./81.
         *m33*m44*RN3e*RSe563p*RS356pe - 1./81.*m33*m44*RN3e*RSe463p*
         RS356pe - 1./81.*m33*m44*RN3e*RSe561p*RS346pe - 1./81.*m33*m44
         *RN3e*RSe561p*RS356pe - 1./81.*m33*m44*RN3e*RSe461p*RS356pe - 
         1./81.*m33*m44*RN1e*RS146pe*RSe563p - 1./81.*m33*m44*RN1e*
         RS156pe*RSe563p - 1./81.*m33*m44*RN1e*RS156pe*RSe463p - 1./81.
         *m11*RSe561p*RSe563p + 1./81.*m11*RSe561p*RSe463p - 1./81.*m11
         *RSe561p*RSp563e + 1./81.*m11*RSe561p*RSp463e - 1./162.*m11*
         pow(RSe561p,2) + 1./81.*m11*RSe461p*RSe563p - 1./81.*m11*
         RSe461p*RSe463p + 1./81.*m11*RSe461p*RSp563e - 1./81.*m11*
         RSe461p*RSp463e + 1./81.*m11*RSe461p*RSe561p - 1./162.*m11*
         pow(RSe461p,2) - 1./81.*m11*RSp561e*RSe563p + 1./81.*m11*
         RSp561e*RSe463p - 1./81.*m11*RSp561e*RSe561p + 1./81.*m11*
         RSp561e*RSe461p + 1./81.*m11*RSp461e*RSe563p - 1./81.*m11*
         RSp461e*RSe463p );

      M13p6x23R +=  + Ab(m45,mu2)*Ab(me2,mu2) * ( 1./81.*m11*RSp461e*
         RSe561p - 1./81.*m11*RSp461e*RSe461p - 1./81.*m11*mp2*RNep*
         RSe561p*RSp563e + 1./81.*m11*mp2*RNep*RSe561p*RSp463e + 1./81.
         *m11*mp2*RNep*RSe461p*RSp563e - 1./81.*m11*mp2*RNep*RSe461p*
         RSp463e - 1./81.*m11*mp2*RNep*RSp561e*RSe563p + 1./81.*m11*mp2
         *RNep*RSp561e*RSe463p - 1./81.*m11*mp2*RNep*RSp561e*RSe561p + 
         1./81.*m11*mp2*RNep*RSp561e*RSe461p + 1./81.*m11*mp2*RNep*
         RSp461e*RSe563p - 1./81.*m11*mp2*RNep*RSp461e*RSe463p + 1./81.
         *m11*mp2*RNep*RSp461e*RSe561p - 1./81.*m11*mp2*RNep*RSp461e*
         RSe461p - 1./81.*m11*me2*RN3e*RSe561p*RS346pe + 1./81.*m11*me2
         *RN3e*RSe561p*RS356pe + 1./81.*m11*me2*RN3e*RSe461p*RS346pe - 
         1./81.*m11*me2*RN3e*RSe461p*RS356pe - 2./81.*m11*me2*RN1e*
         RS146pe*RSe563p + 2./81.*m11*me2*RN1e*RS146pe*RSe463p + 2./81.
         *m11*me2*RN1e*RS156pe*RSe563p - 2./81.*m11*me2*RN1e*RS156pe*
         RSe463p + 1./81.*m11*m55*RNep*RSe561p*RSp463e + 1./81.*m11*m55
         *RNep*RSe461p*RSp563e );

      M13p6x23R +=  + Ab(m45,mu2)*Ab(me2,mu2) * ( 1./81.*m11*m55*RNep*
         RSe461p*RSp463e + 1./81.*m11*m55*RNep*RSp561e*RSe463p + 1./81.
         *m11*m55*RNep*RSp561e*RSe461p + 1./81.*m11*m55*RNep*RSp461e*
         RSe563p + 1./81.*m11*m55*RNep*RSp461e*RSe463p + 1./81.*m11*m55
         *RNep*RSp461e*RSe561p + 1./81.*m11*m55*RNep*RSp461e*RSe461p - 
         1./81.*m11*m55*RN3e*RSe561p*RS346pe - 1./81.*m11*m55*RN3e*
         RSe461p*RS346pe - 1./81.*m11*m55*RN3e*RSe461p*RS356pe - 1./81.
         *m11*m55*RN1e*RS146pe*RSe563p - 1./81.*m11*m55*RN1e*RS146pe*
         RSe463p - 1./81.*m11*m55*RN1e*RS156pe*RSe463p - 1./81.*m11*m55
         *RN1e*RSe561p*RS146pe - 1./81.*m11*m55*RN1e*RSe461p*RS146pe - 
         1./81.*m11*m55*RN1e*RSe461p*RS156pe + 1./81.*m11*m44*RNep*
         RSe561p*RSp563e + 1./81.*m11*m44*RNep*RSe561p*RSp463e + 1./81.
         *m11*m44*RNep*RSe461p*RSp563e + 1./81.*m11*m44*RNep*RSp561e*
         RSe563p + 1./81.*m11*m44*RNep*RSp561e*RSe463p + 1./81.*m11*m44
         *RNep*RSp561e*RSe561p + 1./81.*m11*m44*RNep*RSp561e*RSe461p + 
         1./81.*m11*m44*RNep*RSp461e*RSe563p );

      M13p6x23R +=  + Ab(m45,mu2)*Ab(me2,mu2) * ( 1./81.*m11*m44*RNep*
         RSp461e*RSe561p - 1./81.*m11*m44*RN3e*RSe561p*RS346pe - 1./81.
         *m11*m44*RN3e*RSe561p*RS356pe - 1./81.*m11*m44*RN3e*RSe461p*
         RS356pe - 1./81.*m11*m44*RN1e*RS146pe*RSe563p - 1./81.*m11*m44
         *RN1e*RS156pe*RSe563p - 1./81.*m11*m44*RN1e*RS156pe*RSe463p - 
         1./81.*m11*m44*RN1e*RSe561p*RS146pe - 1./81.*m11*m44*RN1e*
         RSe561p*RS156pe - 1./81.*m11*m44*RN1e*RSe461p*RS156pe );

      M13p6x23R +=  + Ab(m45,mu2)*Ab(mp2,mu2) * (  - 1./81.*mp2*RSp563e
         *RS346pe + 1./81.*mp2*RSp563e*RS356pe + 1./81.*mp2*RSp563e*
         RSe563p - 1./81.*mp2*RSp563e*RSe463p + 1./81.*mp2*pow(
         RSp563e,2) + 1./81.*mp2*RSp463e*RS346pe - 1./81.*mp2*RSp463e*
         RS356pe - 1./81.*mp2*RSp463e*RSe563p + 1./81.*mp2*RSp463e*
         RSe463p - 2./81.*mp2*RSp463e*RSp563e + 1./81.*mp2*pow(
         RSp463e,2) + 1./81.*mp2*RS146pe*RSp563e - 1./81.*mp2*RS146pe*
         RSp463e - 1./81.*mp2*RS156pe*RSp563e + 1./81.*mp2*RS156pe*
         RSp463e - 1./81.*mp2*RSe561p*RSp563e + 1./81.*mp2*RSe561p*
         RSp463e + 1./81.*mp2*RSe461p*RSp563e - 1./81.*mp2*RSe461p*
         RSp463e + 1./81.*mp2*RSp561e*RS346pe - 1./81.*mp2*RSp561e*
         RS356pe - 1./81.*mp2*RSp561e*RSe563p + 1./81.*mp2*RSp561e*
         RSe463p - 2./81.*mp2*RSp561e*RSp563e + 2./81.*mp2*RSp561e*
         RSp463e - 1./81.*mp2*RSp561e*RS146pe + 1./81.*mp2*RSp561e*
         RS156pe + 1./81.*mp2*RSp561e*RSe561p - 1./81.*mp2*RSp561e*
         RSe461p );

      M13p6x23R +=  + Ab(m45,mu2)*Ab(mp2,mu2) * ( 1./81.*mp2*pow(
         RSp561e,2) - 1./81.*mp2*RSp461e*RS346pe + 1./81.*mp2*RSp461e*
         RS356pe + 1./81.*mp2*RSp461e*RSe563p - 1./81.*mp2*RSp461e*
         RSe463p + 2./81.*mp2*RSp461e*RSp563e - 2./81.*mp2*RSp461e*
         RSp463e + 1./81.*mp2*RSp461e*RS146pe - 1./81.*mp2*RSp461e*
         RS156pe - 1./81.*mp2*RSp461e*RSe561p + 1./81.*mp2*RSp461e*
         RSe461p - 2./81.*mp2*RSp461e*RSp561e + 1./81.*mp2*pow(
         RSp461e,2) - 1./81.*me2*mp2*RNep*RSp563e*RSe563p + 1./81.*me2*
         mp2*RNep*RSp563e*RSe463p + 1./81.*me2*mp2*RNep*RSp463e*RSe563p
          - 1./81.*me2*mp2*RNep*RSp463e*RSe463p + 1./81.*me2*mp2*RNep*
         RSe561p*RSp563e - 1./81.*me2*mp2*RNep*RSe561p*RSp463e - 1./81.
         *me2*mp2*RNep*RSe461p*RSp563e + 1./81.*me2*mp2*RNep*RSe461p*
         RSp463e + 1./81.*me2*mp2*RNep*RSp561e*RSe563p - 1./81.*me2*mp2
         *RNep*RSp561e*RSe463p - 1./81.*me2*mp2*RNep*RSp561e*RSe561p + 
         1./81.*me2*mp2*RNep*RSp561e*RSe461p - 1./81.*me2*mp2*RNep*
         RSp461e*RSe563p );

      M13p6x23R +=  + Ab(m45,mu2)*Ab(mp2,mu2) * ( 1./81.*me2*mp2*RNep*
         RSp461e*RSe463p + 1./81.*me2*mp2*RNep*RSp461e*RSe561p - 1./81.
         *me2*mp2*RNep*RSp461e*RSe461p - 1./81.*m55*RSp463e*RSp563e - 1.
         /162.*m55*pow(RSp463e,2) + 1./81.*m55*RSp561e*RSp463e + 1./81.
         *m55*RSp461e*RSp563e + 1./81.*m55*RSp461e*RSp463e - 1./81.*m55
         *RSp461e*RSp561e - 1./162.*m55*pow(RSp461e,2) + 1./81.*m55*mp2
         *RNep*RSp563e*RSe463p + 1./81.*m55*mp2*RNep*RSp463e*RSe563p + 
         1./81.*m55*mp2*RNep*RSp463e*RSe463p - 1./81.*m55*mp2*RNep*
         RSe561p*RSp463e - 1./81.*m55*mp2*RNep*RSe461p*RSp563e - 1./81.
         *m55*mp2*RNep*RSe461p*RSp463e - 1./81.*m55*mp2*RNep*RSp561e*
         RSe463p + 1./81.*m55*mp2*RNep*RSp561e*RSe461p - 1./81.*m55*mp2
         *RNep*RSp461e*RSe563p - 1./81.*m55*mp2*RNep*RSp461e*RSe463p + 
         1./81.*m55*mp2*RNep*RSp461e*RSe561p + 1./81.*m55*mp2*RNep*
         RSp461e*RSe461p + 1./81.*m55*mp2*RN3p*RSp563e*RS346pe + 1./81.
         *m55*mp2*RN3p*RSp463e*RS346pe + 1./81.*m55*mp2*RN3p*RSp463e*
         RS356pe );

      M13p6x23R +=  + Ab(m45,mu2)*Ab(mp2,mu2) * (  - 1./81.*m55*mp2*
         RN3p*RSp561e*RS346pe - 1./81.*m55*mp2*RN3p*RSp461e*RS346pe - 1.
         /81.*m55*mp2*RN3p*RSp461e*RS356pe - 1./81.*m55*mp2*RN1p*
         RS146pe*RSp563e - 1./81.*m55*mp2*RN1p*RS146pe*RSp463e - 1./81.
         *m55*mp2*RN1p*RS156pe*RSp463e + 1./81.*m55*mp2*RN1p*RSp561e*
         RS146pe + 1./81.*m55*mp2*RN1p*RSp461e*RS146pe + 1./81.*m55*mp2
         *RN1p*RSp461e*RS156pe - 1./162.*m44*pow(RSp563e,2) - 1./81.*
         m44*RSp463e*RSp563e + 1./81.*m44*RSp561e*RSp563e + 1./81.*m44*
         RSp561e*RSp463e - 1./162.*m44*pow(RSp561e,2) + 1./81.*m44*
         RSp461e*RSp563e - 1./81.*m44*RSp461e*RSp561e + 1./81.*m44*mp2*
         RNep*RSp563e*RSe563p + 1./81.*m44*mp2*RNep*RSp563e*RSe463p + 1.
         /81.*m44*mp2*RNep*RSp463e*RSe563p - 1./81.*m44*mp2*RNep*
         RSe561p*RSp563e - 1./81.*m44*mp2*RNep*RSe561p*RSp463e - 1./81.
         *m44*mp2*RNep*RSe461p*RSp563e - 1./81.*m44*mp2*RNep*RSp561e*
         RSe563p - 1./81.*m44*mp2*RNep*RSp561e*RSe463p + 1./81.*m44*mp2
         *RNep*RSp561e*RSe561p );

      M13p6x23R +=  + Ab(m45,mu2)*Ab(mp2,mu2) * ( 1./81.*m44*mp2*RNep*
         RSp561e*RSe461p - 1./81.*m44*mp2*RNep*RSp461e*RSe563p + 1./81.
         *m44*mp2*RNep*RSp461e*RSe561p + 1./81.*m44*mp2*RN3p*RSp563e*
         RS346pe + 1./81.*m44*mp2*RN3p*RSp563e*RS356pe + 1./81.*m44*mp2
         *RN3p*RSp463e*RS356pe - 1./81.*m44*mp2*RN3p*RSp561e*RS346pe - 
         1./81.*m44*mp2*RN3p*RSp561e*RS356pe - 1./81.*m44*mp2*RN3p*
         RSp461e*RS356pe - 1./81.*m44*mp2*RN1p*RS146pe*RSp563e - 1./81.
         *m44*mp2*RN1p*RS156pe*RSp563e - 1./81.*m44*mp2*RN1p*RS156pe*
         RSp463e + 1./81.*m44*mp2*RN1p*RSp561e*RS146pe + 1./81.*m44*mp2
         *RN1p*RSp561e*RS156pe + 1./81.*m44*mp2*RN1p*RSp461e*RS156pe - 
         1./162.*m33*pow(RSp563e,2) + 1./81.*m33*RSp463e*RSp563e - 1./
         162.*m33*pow(RSp463e,2) - 1./81.*m33*RSp561e*RSp563e + 1./81.*
         m33*RSp561e*RSp463e + 1./81.*m33*RSp461e*RSp563e - 1./81.*m33*
         RSp461e*RSp463e + 1./81.*m33*mp2*RNep*RSp563e*RSe563p - 1./81.
         *m33*mp2*RNep*RSp563e*RSe463p - 1./81.*m33*mp2*RNep*RSp463e*
         RSe563p );

      M13p6x23R +=  + Ab(m45,mu2)*Ab(mp2,mu2) * ( 1./81.*m33*mp2*RNep*
         RSp463e*RSe463p + 1./81.*m33*mp2*RNep*RSe561p*RSp563e - 1./81.
         *m33*mp2*RNep*RSe561p*RSp463e - 1./81.*m33*mp2*RNep*RSe461p*
         RSp563e + 1./81.*m33*mp2*RNep*RSe461p*RSp463e + 1./81.*m33*mp2
         *RNep*RSp561e*RSe563p - 1./81.*m33*mp2*RNep*RSp561e*RSe463p - 
         1./81.*m33*mp2*RNep*RSp461e*RSe563p + 1./81.*m33*mp2*RNep*
         RSp461e*RSe463p - 2./81.*m33*mp2*RN3p*RSp561e*RS346pe + 2./81.
         *m33*mp2*RN3p*RSp561e*RS356pe + 2./81.*m33*mp2*RN3p*RSp461e*
         RS346pe - 2./81.*m33*mp2*RN3p*RSp461e*RS356pe - 1./81.*m33*mp2
         *RN1p*RS146pe*RSp563e + 1./81.*m33*mp2*RN1p*RS146pe*RSp463e + 
         1./81.*m33*mp2*RN1p*RS156pe*RSp563e - 1./81.*m33*mp2*RN1p*
         RS156pe*RSp463e - 1./81.*m33*m55*RNep*RSp563e*RSe463p - 1./81.
         *m33*m55*RNep*RSp463e*RSe563p - 1./81.*m33*m55*RNep*RSp463e*
         RSe463p - 1./81.*m33*m55*RNep*RSe561p*RSp463e - 1./81.*m33*m55
         *RNep*RSe461p*RSp563e - 1./81.*m33*m55*RNep*RSe461p*RSp463e - 
         1./81.*m33*m55*RNep*RSp561e*RSe463p );

      M13p6x23R +=  + Ab(m45,mu2)*Ab(mp2,mu2) * (  - 1./81.*m33*m55*
         RNep*RSp461e*RSe563p - 1./81.*m33*m55*RNep*RSp461e*RSe463p - 1.
         /81.*m33*m55*RN3p*RSp563e*RS346pe - 1./81.*m33*m55*RN3p*
         RSp463e*RS346pe - 1./81.*m33*m55*RN3p*RSp463e*RS356pe - 1./81.
         *m33*m55*RN3p*RSp561e*RS346pe - 1./81.*m33*m55*RN3p*RSp461e*
         RS346pe - 1./81.*m33*m55*RN3p*RSp461e*RS356pe - 1./81.*m33*m55
         *RN1p*RS146pe*RSp563e - 1./81.*m33*m55*RN1p*RS146pe*RSp463e - 
         1./81.*m33*m55*RN1p*RS156pe*RSp463e - 1./81.*m33*m44*RNep*
         RSp563e*RSe563p - 1./81.*m33*m44*RNep*RSp563e*RSe463p - 1./81.
         *m33*m44*RNep*RSp463e*RSe563p - 1./81.*m33*m44*RNep*RSe561p*
         RSp563e - 1./81.*m33*m44*RNep*RSe561p*RSp463e - 1./81.*m33*m44
         *RNep*RSe461p*RSp563e - 1./81.*m33*m44*RNep*RSp561e*RSe563p - 
         1./81.*m33*m44*RNep*RSp561e*RSe463p - 1./81.*m33*m44*RNep*
         RSp461e*RSe563p - 1./81.*m33*m44*RN3p*RSp563e*RS346pe - 1./81.
         *m33*m44*RN3p*RSp563e*RS356pe - 1./81.*m33*m44*RN3p*RSp463e*
         RS356pe );

      M13p6x23R +=  + Ab(m45,mu2)*Ab(mp2,mu2) * (  - 1./81.*m33*m44*
         RN3p*RSp561e*RS346pe - 1./81.*m33*m44*RN3p*RSp561e*RS356pe - 1.
         /81.*m33*m44*RN3p*RSp461e*RS356pe - 1./81.*m33*m44*RN1p*
         RS146pe*RSp563e - 1./81.*m33*m44*RN1p*RS156pe*RSp563e - 1./81.
         *m33*m44*RN1p*RS156pe*RSp463e - 1./81.*m11*RSp561e*RSp563e + 1.
         /81.*m11*RSp561e*RSp463e - 1./162.*m11*pow(RSp561e,2) + 1./81.
         *m11*RSp461e*RSp563e - 1./81.*m11*RSp461e*RSp463e + 1./81.*m11
         *RSp461e*RSp561e - 1./162.*m11*pow(RSp461e,2) + 1./81.*m11*mp2
         *RNep*RSe561p*RSp563e - 1./81.*m11*mp2*RNep*RSe561p*RSp463e - 
         1./81.*m11*mp2*RNep*RSe461p*RSp563e + 1./81.*m11*mp2*RNep*
         RSe461p*RSp463e + 1./81.*m11*mp2*RNep*RSp561e*RSe563p - 1./81.
         *m11*mp2*RNep*RSp561e*RSe463p + 1./81.*m11*mp2*RNep*RSp561e*
         RSe561p - 1./81.*m11*mp2*RNep*RSp561e*RSe461p - 1./81.*m11*mp2
         *RNep*RSp461e*RSe563p + 1./81.*m11*mp2*RNep*RSp461e*RSe463p - 
         1./81.*m11*mp2*RNep*RSp461e*RSe561p + 1./81.*m11*mp2*RNep*
         RSp461e*RSe461p );

      M13p6x23R +=  + Ab(m45,mu2)*Ab(mp2,mu2) * (  - 1./81.*m11*mp2*
         RN3p*RSp561e*RS346pe + 1./81.*m11*mp2*RN3p*RSp561e*RS356pe + 1.
         /81.*m11*mp2*RN3p*RSp461e*RS346pe - 1./81.*m11*mp2*RN3p*
         RSp461e*RS356pe - 2./81.*m11*mp2*RN1p*RS146pe*RSp563e + 2./81.
         *m11*mp2*RN1p*RS146pe*RSp463e + 2./81.*m11*mp2*RN1p*RS156pe*
         RSp563e - 2./81.*m11*mp2*RN1p*RS156pe*RSp463e - 1./81.*m11*m55
         *RNep*RSe561p*RSp463e - 1./81.*m11*m55*RNep*RSe461p*RSp563e - 
         1./81.*m11*m55*RNep*RSe461p*RSp463e - 1./81.*m11*m55*RNep*
         RSp561e*RSe463p - 1./81.*m11*m55*RNep*RSp561e*RSe461p - 1./81.
         *m11*m55*RNep*RSp461e*RSe563p - 1./81.*m11*m55*RNep*RSp461e*
         RSe463p - 1./81.*m11*m55*RNep*RSp461e*RSe561p - 1./81.*m11*m55
         *RNep*RSp461e*RSe461p - 1./81.*m11*m55*RN3p*RSp561e*RS346pe - 
         1./81.*m11*m55*RN3p*RSp461e*RS346pe - 1./81.*m11*m55*RN3p*
         RSp461e*RS356pe - 1./81.*m11*m55*RN1p*RS146pe*RSp563e - 1./81.
         *m11*m55*RN1p*RS146pe*RSp463e - 1./81.*m11*m55*RN1p*RS156pe*
         RSp463e );

      M13p6x23R +=  + Ab(m45,mu2)*Ab(mp2,mu2) * (  - 1./81.*m11*m55*
         RN1p*RSp561e*RS146pe - 1./81.*m11*m55*RN1p*RSp461e*RS146pe - 1.
         /81.*m11*m55*RN1p*RSp461e*RS156pe - 1./81.*m11*m44*RNep*
         RSe561p*RSp563e - 1./81.*m11*m44*RNep*RSe561p*RSp463e - 1./81.
         *m11*m44*RNep*RSe461p*RSp563e - 1./81.*m11*m44*RNep*RSp561e*
         RSe563p - 1./81.*m11*m44*RNep*RSp561e*RSe463p - 1./81.*m11*m44
         *RNep*RSp561e*RSe561p - 1./81.*m11*m44*RNep*RSp561e*RSe461p - 
         1./81.*m11*m44*RNep*RSp461e*RSe563p - 1./81.*m11*m44*RNep*
         RSp461e*RSe561p - 1./81.*m11*m44*RN3p*RSp561e*RS346pe - 1./81.
         *m11*m44*RN3p*RSp561e*RS356pe - 1./81.*m11*m44*RN3p*RSp461e*
         RS356pe - 1./81.*m11*m44*RN1p*RS146pe*RSp563e - 1./81.*m11*m44
         *RN1p*RS156pe*RSp563e - 1./81.*m11*m44*RN1p*RS156pe*RSp463e - 
         1./81.*m11*m44*RN1p*RSp561e*RS146pe - 1./81.*m11*m44*RN1p*
         RSp561e*RS156pe - 1./81.*m11*m44*RN1p*RSp461e*RS156pe );

      M13p6x23R +=  + Ab(m45,mu2)*Bb(me2,mu2) * ( 1./162.*pow(me2,2)*
         pow(RSe563p,2) - 1./81.*pow(me2,2)*RSe463p*RSe563p + 1./162.*
         pow(me2,2)*pow(RSe463p,2) - 1./81.*pow(me2,2)*RSe561p*RSe563p
          + 1./81.*pow(me2,2)*RSe561p*RSe463p + 1./162.*pow(me2,2)*pow(
         RSe561p,2) + 1./81.*pow(me2,2)*RSe461p*RSe563p - 1./81.*pow(
         me2,2)*RSe461p*RSe463p - 1./81.*pow(me2,2)*RSe461p*RSe561p + 1.
         /162.*pow(me2,2)*pow(RSe461p,2) - 1./81.*m55*me2*RSe463p*
         RSe563p - 1./162.*m55*me2*pow(RSe463p,2) + 1./81.*m55*me2*
         RSe561p*RSe463p + 1./81.*m55*me2*RSe461p*RSe563p + 1./81.*m55*
         me2*RSe461p*RSe463p - 1./81.*m55*me2*RSe461p*RSe561p - 1./162.
         *m55*me2*pow(RSe461p,2) - 1./162.*m44*me2*pow(RSe563p,2) - 1./
         81.*m44*me2*RSe463p*RSe563p + 1./81.*m44*me2*RSe561p*RSe563p
          + 1./81.*m44*me2*RSe561p*RSe463p - 1./162.*m44*me2*pow(
         RSe561p,2) + 1./81.*m44*me2*RSe461p*RSe563p - 1./81.*m44*me2*
         RSe461p*RSe561p - 1./162.*m33*me2*pow(RSe563p,2) + 1./81.*m33*
         me2*RSe463p*RSe563p );

      M13p6x23R +=  + Ab(m45,mu2)*Bb(me2,mu2) * (  - 1./162.*m33*me2*
         pow(RSe463p,2) - 1./81.*m33*me2*RSe561p*RSe563p + 1./81.*m33*
         me2*RSe561p*RSe463p + 1./81.*m33*me2*RSe461p*RSe563p - 1./81.*
         m33*me2*RSe461p*RSe463p + 1./81.*m33*m55*RSe463p*RSe563p + 1./
         162.*m33*m55*pow(RSe463p,2) + 1./81.*m33*m55*RSe561p*RSe463p
          + 1./81.*m33*m55*RSe461p*RSe563p + 1./81.*m33*m55*RSe461p*
         RSe463p + 1./162.*m33*m44*pow(RSe563p,2) + 1./81.*m33*m44*
         RSe463p*RSe563p + 1./81.*m33*m44*RSe561p*RSe563p + 1./81.*m33*
         m44*RSe561p*RSe463p + 1./81.*m33*m44*RSe461p*RSe563p - 1./81.*
         m11*me2*RSe561p*RSe563p + 1./81.*m11*me2*RSe561p*RSe463p - 1./
         162.*m11*me2*pow(RSe561p,2) + 1./81.*m11*me2*RSe461p*RSe563p
          - 1./81.*m11*me2*RSe461p*RSe463p + 1./81.*m11*me2*RSe461p*
         RSe561p - 1./162.*m11*me2*pow(RSe461p,2) + 1./81.*m11*m55*
         RSe561p*RSe463p + 1./81.*m11*m55*RSe461p*RSe563p + 1./81.*m11*
         m55*RSe461p*RSe463p + 1./81.*m11*m55*RSe461p*RSe561p + 1./162.
         *m11*m55*pow(RSe461p,2) );

      M13p6x23R +=  + Ab(m45,mu2)*Bb(me2,mu2) * ( 1./81.*m11*m44*
         RSe561p*RSe563p + 1./81.*m11*m44*RSe561p*RSe463p + 1./162.*m11
         *m44*pow(RSe561p,2) + 1./81.*m11*m44*RSe461p*RSe563p + 1./81.*
         m11*m44*RSe461p*RSe561p );

      M13p6x23R +=  + Ab(m45,mu2)*Bb(mp2,mu2) * ( 1./162.*pow(mp2,2)*
         pow(RSp563e,2) - 1./81.*pow(mp2,2)*RSp463e*RSp563e + 1./162.*
         pow(mp2,2)*pow(RSp463e,2) - 1./81.*pow(mp2,2)*RSp561e*RSp563e
          + 1./81.*pow(mp2,2)*RSp561e*RSp463e + 1./162.*pow(mp2,2)*pow(
         RSp561e,2) + 1./81.*pow(mp2,2)*RSp461e*RSp563e - 1./81.*pow(
         mp2,2)*RSp461e*RSp463e - 1./81.*pow(mp2,2)*RSp461e*RSp561e + 1.
         /162.*pow(mp2,2)*pow(RSp461e,2) - 1./81.*m55*mp2*RSp463e*
         RSp563e - 1./162.*m55*mp2*pow(RSp463e,2) + 1./81.*m55*mp2*
         RSp561e*RSp463e + 1./81.*m55*mp2*RSp461e*RSp563e + 1./81.*m55*
         mp2*RSp461e*RSp463e - 1./81.*m55*mp2*RSp461e*RSp561e - 1./162.
         *m55*mp2*pow(RSp461e,2) - 1./162.*m44*mp2*pow(RSp563e,2) - 1./
         81.*m44*mp2*RSp463e*RSp563e + 1./81.*m44*mp2*RSp561e*RSp563e
          + 1./81.*m44*mp2*RSp561e*RSp463e - 1./162.*m44*mp2*pow(
         RSp561e,2) + 1./81.*m44*mp2*RSp461e*RSp563e - 1./81.*m44*mp2*
         RSp461e*RSp561e - 1./162.*m33*mp2*pow(RSp563e,2) + 1./81.*m33*
         mp2*RSp463e*RSp563e );

      M13p6x23R +=  + Ab(m45,mu2)*Bb(mp2,mu2) * (  - 1./162.*m33*mp2*
         pow(RSp463e,2) - 1./81.*m33*mp2*RSp561e*RSp563e + 1./81.*m33*
         mp2*RSp561e*RSp463e + 1./81.*m33*mp2*RSp461e*RSp563e - 1./81.*
         m33*mp2*RSp461e*RSp463e + 1./81.*m33*m55*RSp463e*RSp563e + 1./
         162.*m33*m55*pow(RSp463e,2) + 1./81.*m33*m55*RSp561e*RSp463e
          + 1./81.*m33*m55*RSp461e*RSp563e + 1./81.*m33*m55*RSp461e*
         RSp463e + 1./162.*m33*m44*pow(RSp563e,2) + 1./81.*m33*m44*
         RSp463e*RSp563e + 1./81.*m33*m44*RSp561e*RSp563e + 1./81.*m33*
         m44*RSp561e*RSp463e + 1./81.*m33*m44*RSp461e*RSp563e - 1./81.*
         m11*mp2*RSp561e*RSp563e + 1./81.*m11*mp2*RSp561e*RSp463e - 1./
         162.*m11*mp2*pow(RSp561e,2) + 1./81.*m11*mp2*RSp461e*RSp563e
          - 1./81.*m11*mp2*RSp461e*RSp463e + 1./81.*m11*mp2*RSp461e*
         RSp561e - 1./162.*m11*mp2*pow(RSp461e,2) + 1./81.*m11*m55*
         RSp561e*RSp463e + 1./81.*m11*m55*RSp461e*RSp563e + 1./81.*m11*
         m55*RSp461e*RSp463e + 1./81.*m11*m55*RSp461e*RSp561e + 1./162.
         *m11*m55*pow(RSp461e,2) );

      M13p6x23R +=  + Ab(m45,mu2)*Bb(mp2,mu2) * ( 1./81.*m11*m44*
         RSp561e*RSp563e + 1./81.*m11*m44*RSp561e*RSp463e + 1./162.*m11
         *m44*pow(RSp561e,2) + 1./81.*m11*m44*RSp461e*RSp563e + 1./81.*
         m11*m44*RSp461e*RSp561e );

      M13p6x23R +=  + Ab(m46,mu2)*Ab(me2,mu2) * (  - 1./81.*me2*RSe563p
         *RS345pe + 1./81.*me2*RSe563p*RS356pe + 1./81.*me2*pow(
         RSe563p,2) + 1./81.*me2*RSe453p*RS345pe - 1./81.*me2*RSe453p*
         RS356pe - 2./81.*me2*RSe453p*RSe563p + 1./81.*me2*pow(
         RSe453p,2) + 1./81.*me2*RSp563e*RSe563p - 1./81.*me2*RSp563e*
         RSe453p - 1./81.*me2*RSp453e*RSe563p + 1./81.*me2*RSp453e*
         RSe453p + 1./81.*me2*RS145pe*RSe563p - 1./81.*me2*RS145pe*
         RSe453p - 1./81.*me2*RS156pe*RSe563p + 1./81.*me2*RS156pe*
         RSe453p + 1./81.*me2*RSe561p*RS345pe - 1./81.*me2*RSe561p*
         RS356pe - 2./81.*me2*RSe561p*RSe563p + 2./81.*me2*RSe561p*
         RSe453p - 1./81.*me2*RSe561p*RSp563e + 1./81.*me2*RSe561p*
         RSp453e - 1./81.*me2*RSe561p*RS145pe + 1./81.*me2*RSe561p*
         RS156pe + 1./81.*me2*pow(RSe561p,2) - 1./81.*me2*RSe451p*
         RS345pe + 1./81.*me2*RSe451p*RS356pe + 2./81.*me2*RSe451p*
         RSe563p - 2./81.*me2*RSe451p*RSe453p + 1./81.*me2*RSe451p*
         RSp563e );

      M13p6x23R +=  + Ab(m46,mu2)*Ab(me2,mu2) * (  - 1./81.*me2*RSe451p
         *RSp453e + 1./81.*me2*RSe451p*RS145pe - 1./81.*me2*RSe451p*
         RS156pe - 2./81.*me2*RSe451p*RSe561p + 1./81.*me2*pow(
         RSe451p,2) - 1./81.*me2*RSp561e*RSe563p + 1./81.*me2*RSp561e*
         RSe453p + 1./81.*me2*RSp561e*RSe561p - 1./81.*me2*RSp561e*
         RSe451p + 1./81.*me2*RSp451e*RSe563p - 1./81.*me2*RSp451e*
         RSe453p - 1./81.*me2*RSp451e*RSe561p + 1./81.*me2*RSp451e*
         RSe451p + 1./81.*me2*mp2*RNep*RSp563e*RSe563p - 1./81.*me2*mp2
         *RNep*RSp563e*RSe453p - 1./81.*me2*mp2*RNep*RSp453e*RSe563p + 
         1./81.*me2*mp2*RNep*RSp453e*RSe453p - 1./81.*me2*mp2*RNep*
         RSe561p*RSp563e + 1./81.*me2*mp2*RNep*RSe561p*RSp453e + 1./81.
         *me2*mp2*RNep*RSe451p*RSp563e - 1./81.*me2*mp2*RNep*RSe451p*
         RSp453e - 1./81.*me2*mp2*RNep*RSp561e*RSe563p + 1./81.*me2*mp2
         *RNep*RSp561e*RSe453p + 1./81.*me2*mp2*RNep*RSp561e*RSe561p - 
         1./81.*me2*mp2*RNep*RSp561e*RSe451p + 1./81.*me2*mp2*RNep*
         RSp451e*RSe563p );

      M13p6x23R +=  + Ab(m46,mu2)*Ab(me2,mu2) * (  - 1./81.*me2*mp2*
         RNep*RSp451e*RSe453p - 1./81.*me2*mp2*RNep*RSp451e*RSe561p + 1.
         /81.*me2*mp2*RNep*RSp451e*RSe451p - 1./81.*m66*RSe453p*RSe563p
          - 1./162.*m66*pow(RSe453p,2) - 1./81.*m66*RSp563e*RSe453p - 1.
         /81.*m66*RSp453e*RSe563p - 1./81.*m66*RSp453e*RSe453p + 1./81.
         *m66*RSe561p*RSe453p + 1./81.*m66*RSe561p*RSp453e + 1./81.*m66
         *RSe451p*RSe563p + 1./81.*m66*RSe451p*RSe453p + 1./81.*m66*
         RSe451p*RSp563e + 1./81.*m66*RSe451p*RSp453e - 1./81.*m66*
         RSe451p*RSe561p - 1./162.*m66*pow(RSe451p,2) + 1./81.*m66*
         RSp561e*RSe453p - 1./81.*m66*RSp561e*RSe451p + 1./81.*m66*
         RSp451e*RSe563p + 1./81.*m66*RSp451e*RSe453p - 1./81.*m66*
         RSp451e*RSe561p - 1./81.*m66*RSp451e*RSe451p - 1./81.*m66*mp2*
         RNep*RSp563e*RSe453p - 1./81.*m66*mp2*RNep*RSp453e*RSe563p - 1.
         /81.*m66*mp2*RNep*RSp453e*RSe453p + 1./81.*m66*mp2*RNep*
         RSe561p*RSp453e + 1./81.*m66*mp2*RNep*RSe451p*RSp563e + 1./81.
         *m66*mp2*RNep*RSe451p*RSp453e );

      M13p6x23R +=  + Ab(m46,mu2)*Ab(me2,mu2) * ( 1./81.*m66*mp2*RNep*
         RSp561e*RSe453p - 1./81.*m66*mp2*RNep*RSp561e*RSe451p + 1./81.
         *m66*mp2*RNep*RSp451e*RSe563p + 1./81.*m66*mp2*RNep*RSp451e*
         RSe453p - 1./81.*m66*mp2*RNep*RSp451e*RSe561p - 1./81.*m66*mp2
         *RNep*RSp451e*RSe451p + 1./81.*m66*me2*RN3e*RSe563p*RS345pe + 
         1./81.*m66*me2*RN3e*RSe453p*RS345pe + 1./81.*m66*me2*RN3e*
         RSe453p*RS356pe - 1./81.*m66*me2*RN3e*RSe561p*RS345pe - 1./81.
         *m66*me2*RN3e*RSe451p*RS345pe - 1./81.*m66*me2*RN3e*RSe451p*
         RS356pe - 1./81.*m66*me2*RN1e*RS145pe*RSe563p - 1./81.*m66*me2
         *RN1e*RS145pe*RSe453p - 1./81.*m66*me2*RN1e*RS156pe*RSe453p + 
         1./81.*m66*me2*RN1e*RSe561p*RS145pe + 1./81.*m66*me2*RN1e*
         RSe451p*RS145pe + 1./81.*m66*me2*RN1e*RSe451p*RS156pe - 1./162.
         *m44*pow(RSe563p,2) - 1./81.*m44*RSe453p*RSe563p - 1./81.*m44*
         RSp563e*RSe563p - 1./81.*m44*RSp563e*RSe453p - 1./81.*m44*
         RSp453e*RSe563p + 1./81.*m44*RSe561p*RSe563p + 1./81.*m44*
         RSe561p*RSe453p );

      M13p6x23R +=  + Ab(m46,mu2)*Ab(me2,mu2) * ( 1./81.*m44*RSe561p*
         RSp563e + 1./81.*m44*RSe561p*RSp453e - 1./162.*m44*pow(
         RSe561p,2) + 1./81.*m44*RSe451p*RSe563p + 1./81.*m44*RSe451p*
         RSp563e - 1./81.*m44*RSe451p*RSe561p + 1./81.*m44*RSp561e*
         RSe563p + 1./81.*m44*RSp561e*RSe453p - 1./81.*m44*RSp561e*
         RSe561p - 1./81.*m44*RSp561e*RSe451p + 1./81.*m44*RSp451e*
         RSe563p - 1./81.*m44*RSp451e*RSe561p - 1./81.*m44*mp2*RNep*
         RSp563e*RSe563p - 1./81.*m44*mp2*RNep*RSp563e*RSe453p - 1./81.
         *m44*mp2*RNep*RSp453e*RSe563p + 1./81.*m44*mp2*RNep*RSe561p*
         RSp563e + 1./81.*m44*mp2*RNep*RSe561p*RSp453e + 1./81.*m44*mp2
         *RNep*RSe451p*RSp563e + 1./81.*m44*mp2*RNep*RSp561e*RSe563p + 
         1./81.*m44*mp2*RNep*RSp561e*RSe453p - 1./81.*m44*mp2*RNep*
         RSp561e*RSe561p - 1./81.*m44*mp2*RNep*RSp561e*RSe451p + 1./81.
         *m44*mp2*RNep*RSp451e*RSe563p - 1./81.*m44*mp2*RNep*RSp451e*
         RSe561p + 1./81.*m44*me2*RN3e*RSe563p*RS345pe + 1./81.*m44*me2
         *RN3e*RSe563p*RS356pe );

      M13p6x23R +=  + Ab(m46,mu2)*Ab(me2,mu2) * ( 1./81.*m44*me2*RN3e*
         RSe453p*RS356pe - 1./81.*m44*me2*RN3e*RSe561p*RS345pe - 1./81.
         *m44*me2*RN3e*RSe561p*RS356pe - 1./81.*m44*me2*RN3e*RSe451p*
         RS356pe - 1./81.*m44*me2*RN1e*RS145pe*RSe563p - 1./81.*m44*me2
         *RN1e*RS156pe*RSe563p - 1./81.*m44*me2*RN1e*RS156pe*RSe453p + 
         1./81.*m44*me2*RN1e*RSe561p*RS145pe + 1./81.*m44*me2*RN1e*
         RSe561p*RS156pe + 1./81.*m44*me2*RN1e*RSe451p*RS156pe - 1./162.
         *m33*pow(RSe563p,2) + 1./81.*m33*RSe453p*RSe563p - 1./162.*m33
         *pow(RSe453p,2) - 1./81.*m33*RSp563e*RSe563p + 1./81.*m33*
         RSp563e*RSe453p + 1./81.*m33*RSp453e*RSe563p - 1./81.*m33*
         RSp453e*RSe453p - 1./81.*m33*RSe561p*RSe563p + 1./81.*m33*
         RSe561p*RSe453p - 1./81.*m33*RSe561p*RSp563e + 1./81.*m33*
         RSe561p*RSp453e + 1./81.*m33*RSe451p*RSe563p - 1./81.*m33*
         RSe451p*RSe453p + 1./81.*m33*RSe451p*RSp563e - 1./81.*m33*
         RSe451p*RSp453e - 1./81.*m33*RSp561e*RSe563p + 1./81.*m33*
         RSp561e*RSe453p );

      M13p6x23R +=  + Ab(m46,mu2)*Ab(me2,mu2) * ( 1./81.*m33*RSp451e*
         RSe563p - 1./81.*m33*RSp451e*RSe453p - 1./81.*m33*mp2*RNep*
         RSp563e*RSe563p + 1./81.*m33*mp2*RNep*RSp563e*RSe453p + 1./81.
         *m33*mp2*RNep*RSp453e*RSe563p - 1./81.*m33*mp2*RNep*RSp453e*
         RSe453p - 1./81.*m33*mp2*RNep*RSe561p*RSp563e + 1./81.*m33*mp2
         *RNep*RSe561p*RSp453e + 1./81.*m33*mp2*RNep*RSe451p*RSp563e - 
         1./81.*m33*mp2*RNep*RSe451p*RSp453e - 1./81.*m33*mp2*RNep*
         RSp561e*RSe563p + 1./81.*m33*mp2*RNep*RSp561e*RSe453p + 1./81.
         *m33*mp2*RNep*RSp451e*RSe563p - 1./81.*m33*mp2*RNep*RSp451e*
         RSe453p - 2./81.*m33*me2*RN3e*RSe561p*RS345pe + 2./81.*m33*me2
         *RN3e*RSe561p*RS356pe + 2./81.*m33*me2*RN3e*RSe451p*RS345pe - 
         2./81.*m33*me2*RN3e*RSe451p*RS356pe - 1./81.*m33*me2*RN1e*
         RS145pe*RSe563p + 1./81.*m33*me2*RN1e*RS145pe*RSe453p + 1./81.
         *m33*me2*RN1e*RS156pe*RSe563p - 1./81.*m33*me2*RN1e*RS156pe*
         RSe453p + 1./81.*m33*m66*RNep*RSp563e*RSe453p + 1./81.*m33*m66
         *RNep*RSp453e*RSe563p );

      M13p6x23R +=  + Ab(m46,mu2)*Ab(me2,mu2) * ( 1./81.*m33*m66*RNep*
         RSp453e*RSe453p + 1./81.*m33*m66*RNep*RSe561p*RSp453e + 1./81.
         *m33*m66*RNep*RSe451p*RSp563e + 1./81.*m33*m66*RNep*RSe451p*
         RSp453e + 1./81.*m33*m66*RNep*RSp561e*RSe453p + 1./81.*m33*m66
         *RNep*RSp451e*RSe563p + 1./81.*m33*m66*RNep*RSp451e*RSe453p - 
         1./81.*m33*m66*RN3e*RSe563p*RS345pe - 1./81.*m33*m66*RN3e*
         RSe453p*RS345pe - 1./81.*m33*m66*RN3e*RSe453p*RS356pe - 1./81.
         *m33*m66*RN3e*RSe561p*RS345pe - 1./81.*m33*m66*RN3e*RSe451p*
         RS345pe - 1./81.*m33*m66*RN3e*RSe451p*RS356pe - 1./81.*m33*m66
         *RN1e*RS145pe*RSe563p - 1./81.*m33*m66*RN1e*RS145pe*RSe453p - 
         1./81.*m33*m66*RN1e*RS156pe*RSe453p + 1./81.*m33*m44*RNep*
         RSp563e*RSe563p + 1./81.*m33*m44*RNep*RSp563e*RSe453p + 1./81.
         *m33*m44*RNep*RSp453e*RSe563p + 1./81.*m33*m44*RNep*RSe561p*
         RSp563e + 1./81.*m33*m44*RNep*RSe561p*RSp453e + 1./81.*m33*m44
         *RNep*RSe451p*RSp563e + 1./81.*m33*m44*RNep*RSp561e*RSe563p + 
         1./81.*m33*m44*RNep*RSp561e*RSe453p );

      M13p6x23R +=  + Ab(m46,mu2)*Ab(me2,mu2) * ( 1./81.*m33*m44*RNep*
         RSp451e*RSe563p - 1./81.*m33*m44*RN3e*RSe563p*RS345pe - 1./81.
         *m33*m44*RN3e*RSe563p*RS356pe - 1./81.*m33*m44*RN3e*RSe453p*
         RS356pe - 1./81.*m33*m44*RN3e*RSe561p*RS345pe - 1./81.*m33*m44
         *RN3e*RSe561p*RS356pe - 1./81.*m33*m44*RN3e*RSe451p*RS356pe - 
         1./81.*m33*m44*RN1e*RS145pe*RSe563p - 1./81.*m33*m44*RN1e*
         RS156pe*RSe563p - 1./81.*m33*m44*RN1e*RS156pe*RSe453p - 1./81.
         *m11*RSe561p*RSe563p + 1./81.*m11*RSe561p*RSe453p - 1./81.*m11
         *RSe561p*RSp563e + 1./81.*m11*RSe561p*RSp453e - 1./162.*m11*
         pow(RSe561p,2) + 1./81.*m11*RSe451p*RSe563p - 1./81.*m11*
         RSe451p*RSe453p + 1./81.*m11*RSe451p*RSp563e - 1./81.*m11*
         RSe451p*RSp453e + 1./81.*m11*RSe451p*RSe561p - 1./162.*m11*
         pow(RSe451p,2) - 1./81.*m11*RSp561e*RSe563p + 1./81.*m11*
         RSp561e*RSe453p - 1./81.*m11*RSp561e*RSe561p + 1./81.*m11*
         RSp561e*RSe451p + 1./81.*m11*RSp451e*RSe563p - 1./81.*m11*
         RSp451e*RSe453p );

      M13p6x23R +=  + Ab(m46,mu2)*Ab(me2,mu2) * ( 1./81.*m11*RSp451e*
         RSe561p - 1./81.*m11*RSp451e*RSe451p - 1./81.*m11*mp2*RNep*
         RSe561p*RSp563e + 1./81.*m11*mp2*RNep*RSe561p*RSp453e + 1./81.
         *m11*mp2*RNep*RSe451p*RSp563e - 1./81.*m11*mp2*RNep*RSe451p*
         RSp453e - 1./81.*m11*mp2*RNep*RSp561e*RSe563p + 1./81.*m11*mp2
         *RNep*RSp561e*RSe453p - 1./81.*m11*mp2*RNep*RSp561e*RSe561p + 
         1./81.*m11*mp2*RNep*RSp561e*RSe451p + 1./81.*m11*mp2*RNep*
         RSp451e*RSe563p - 1./81.*m11*mp2*RNep*RSp451e*RSe453p + 1./81.
         *m11*mp2*RNep*RSp451e*RSe561p - 1./81.*m11*mp2*RNep*RSp451e*
         RSe451p - 1./81.*m11*me2*RN3e*RSe561p*RS345pe + 1./81.*m11*me2
         *RN3e*RSe561p*RS356pe + 1./81.*m11*me2*RN3e*RSe451p*RS345pe - 
         1./81.*m11*me2*RN3e*RSe451p*RS356pe - 2./81.*m11*me2*RN1e*
         RS145pe*RSe563p + 2./81.*m11*me2*RN1e*RS145pe*RSe453p + 2./81.
         *m11*me2*RN1e*RS156pe*RSe563p - 2./81.*m11*me2*RN1e*RS156pe*
         RSe453p + 1./81.*m11*m66*RNep*RSe561p*RSp453e + 1./81.*m11*m66
         *RNep*RSe451p*RSp563e );

      M13p6x23R +=  + Ab(m46,mu2)*Ab(me2,mu2) * ( 1./81.*m11*m66*RNep*
         RSe451p*RSp453e + 1./81.*m11*m66*RNep*RSp561e*RSe453p + 1./81.
         *m11*m66*RNep*RSp561e*RSe451p + 1./81.*m11*m66*RNep*RSp451e*
         RSe563p + 1./81.*m11*m66*RNep*RSp451e*RSe453p + 1./81.*m11*m66
         *RNep*RSp451e*RSe561p + 1./81.*m11*m66*RNep*RSp451e*RSe451p - 
         1./81.*m11*m66*RN3e*RSe561p*RS345pe - 1./81.*m11*m66*RN3e*
         RSe451p*RS345pe - 1./81.*m11*m66*RN3e*RSe451p*RS356pe - 1./81.
         *m11*m66*RN1e*RS145pe*RSe563p - 1./81.*m11*m66*RN1e*RS145pe*
         RSe453p - 1./81.*m11*m66*RN1e*RS156pe*RSe453p - 1./81.*m11*m66
         *RN1e*RSe561p*RS145pe - 1./81.*m11*m66*RN1e*RSe451p*RS145pe - 
         1./81.*m11*m66*RN1e*RSe451p*RS156pe + 1./81.*m11*m44*RNep*
         RSe561p*RSp563e + 1./81.*m11*m44*RNep*RSe561p*RSp453e + 1./81.
         *m11*m44*RNep*RSe451p*RSp563e + 1./81.*m11*m44*RNep*RSp561e*
         RSe563p + 1./81.*m11*m44*RNep*RSp561e*RSe453p + 1./81.*m11*m44
         *RNep*RSp561e*RSe561p + 1./81.*m11*m44*RNep*RSp561e*RSe451p + 
         1./81.*m11*m44*RNep*RSp451e*RSe563p );

      M13p6x23R +=  + Ab(m46,mu2)*Ab(me2,mu2) * ( 1./81.*m11*m44*RNep*
         RSp451e*RSe561p - 1./81.*m11*m44*RN3e*RSe561p*RS345pe - 1./81.
         *m11*m44*RN3e*RSe561p*RS356pe - 1./81.*m11*m44*RN3e*RSe451p*
         RS356pe - 1./81.*m11*m44*RN1e*RS145pe*RSe563p - 1./81.*m11*m44
         *RN1e*RS156pe*RSe563p - 1./81.*m11*m44*RN1e*RS156pe*RSe453p - 
         1./81.*m11*m44*RN1e*RSe561p*RS145pe - 1./81.*m11*m44*RN1e*
         RSe561p*RS156pe - 1./81.*m11*m44*RN1e*RSe451p*RS156pe );

      M13p6x23R +=  + Ab(m46,mu2)*Ab(mp2,mu2) * (  - 1./81.*mp2*RSp563e
         *RS345pe + 1./81.*mp2*RSp563e*RS356pe + 1./81.*mp2*RSp563e*
         RSe563p - 1./81.*mp2*RSp563e*RSe453p + 1./81.*mp2*pow(
         RSp563e,2) + 1./81.*mp2*RSp453e*RS345pe - 1./81.*mp2*RSp453e*
         RS356pe - 1./81.*mp2*RSp453e*RSe563p + 1./81.*mp2*RSp453e*
         RSe453p - 2./81.*mp2*RSp453e*RSp563e + 1./81.*mp2*pow(
         RSp453e,2) + 1./81.*mp2*RS145pe*RSp563e - 1./81.*mp2*RS145pe*
         RSp453e - 1./81.*mp2*RS156pe*RSp563e + 1./81.*mp2*RS156pe*
         RSp453e - 1./81.*mp2*RSe561p*RSp563e + 1./81.*mp2*RSe561p*
         RSp453e + 1./81.*mp2*RSe451p*RSp563e - 1./81.*mp2*RSe451p*
         RSp453e + 1./81.*mp2*RSp561e*RS345pe - 1./81.*mp2*RSp561e*
         RS356pe - 1./81.*mp2*RSp561e*RSe563p + 1./81.*mp2*RSp561e*
         RSe453p - 2./81.*mp2*RSp561e*RSp563e + 2./81.*mp2*RSp561e*
         RSp453e - 1./81.*mp2*RSp561e*RS145pe + 1./81.*mp2*RSp561e*
         RS156pe + 1./81.*mp2*RSp561e*RSe561p - 1./81.*mp2*RSp561e*
         RSe451p );

      M13p6x23R +=  + Ab(m46,mu2)*Ab(mp2,mu2) * ( 1./81.*mp2*pow(
         RSp561e,2) - 1./81.*mp2*RSp451e*RS345pe + 1./81.*mp2*RSp451e*
         RS356pe + 1./81.*mp2*RSp451e*RSe563p - 1./81.*mp2*RSp451e*
         RSe453p + 2./81.*mp2*RSp451e*RSp563e - 2./81.*mp2*RSp451e*
         RSp453e + 1./81.*mp2*RSp451e*RS145pe - 1./81.*mp2*RSp451e*
         RS156pe - 1./81.*mp2*RSp451e*RSe561p + 1./81.*mp2*RSp451e*
         RSe451p - 2./81.*mp2*RSp451e*RSp561e + 1./81.*mp2*pow(
         RSp451e,2) - 1./81.*me2*mp2*RNep*RSp563e*RSe563p + 1./81.*me2*
         mp2*RNep*RSp563e*RSe453p + 1./81.*me2*mp2*RNep*RSp453e*RSe563p
          - 1./81.*me2*mp2*RNep*RSp453e*RSe453p + 1./81.*me2*mp2*RNep*
         RSe561p*RSp563e - 1./81.*me2*mp2*RNep*RSe561p*RSp453e - 1./81.
         *me2*mp2*RNep*RSe451p*RSp563e + 1./81.*me2*mp2*RNep*RSe451p*
         RSp453e + 1./81.*me2*mp2*RNep*RSp561e*RSe563p - 1./81.*me2*mp2
         *RNep*RSp561e*RSe453p - 1./81.*me2*mp2*RNep*RSp561e*RSe561p + 
         1./81.*me2*mp2*RNep*RSp561e*RSe451p - 1./81.*me2*mp2*RNep*
         RSp451e*RSe563p );

      M13p6x23R +=  + Ab(m46,mu2)*Ab(mp2,mu2) * ( 1./81.*me2*mp2*RNep*
         RSp451e*RSe453p + 1./81.*me2*mp2*RNep*RSp451e*RSe561p - 1./81.
         *me2*mp2*RNep*RSp451e*RSe451p - 1./81.*m66*RSp453e*RSp563e - 1.
         /162.*m66*pow(RSp453e,2) + 1./81.*m66*RSp561e*RSp453e + 1./81.
         *m66*RSp451e*RSp563e + 1./81.*m66*RSp451e*RSp453e - 1./81.*m66
         *RSp451e*RSp561e - 1./162.*m66*pow(RSp451e,2) + 1./81.*m66*mp2
         *RNep*RSp563e*RSe453p + 1./81.*m66*mp2*RNep*RSp453e*RSe563p + 
         1./81.*m66*mp2*RNep*RSp453e*RSe453p - 1./81.*m66*mp2*RNep*
         RSe561p*RSp453e - 1./81.*m66*mp2*RNep*RSe451p*RSp563e - 1./81.
         *m66*mp2*RNep*RSe451p*RSp453e - 1./81.*m66*mp2*RNep*RSp561e*
         RSe453p + 1./81.*m66*mp2*RNep*RSp561e*RSe451p - 1./81.*m66*mp2
         *RNep*RSp451e*RSe563p - 1./81.*m66*mp2*RNep*RSp451e*RSe453p + 
         1./81.*m66*mp2*RNep*RSp451e*RSe561p + 1./81.*m66*mp2*RNep*
         RSp451e*RSe451p + 1./81.*m66*mp2*RN3p*RSp563e*RS345pe + 1./81.
         *m66*mp2*RN3p*RSp453e*RS345pe + 1./81.*m66*mp2*RN3p*RSp453e*
         RS356pe );

      M13p6x23R +=  + Ab(m46,mu2)*Ab(mp2,mu2) * (  - 1./81.*m66*mp2*
         RN3p*RSp561e*RS345pe - 1./81.*m66*mp2*RN3p*RSp451e*RS345pe - 1.
         /81.*m66*mp2*RN3p*RSp451e*RS356pe - 1./81.*m66*mp2*RN1p*
         RS145pe*RSp563e - 1./81.*m66*mp2*RN1p*RS145pe*RSp453e - 1./81.
         *m66*mp2*RN1p*RS156pe*RSp453e + 1./81.*m66*mp2*RN1p*RSp561e*
         RS145pe + 1./81.*m66*mp2*RN1p*RSp451e*RS145pe + 1./81.*m66*mp2
         *RN1p*RSp451e*RS156pe - 1./162.*m44*pow(RSp563e,2) - 1./81.*
         m44*RSp453e*RSp563e + 1./81.*m44*RSp561e*RSp563e + 1./81.*m44*
         RSp561e*RSp453e - 1./162.*m44*pow(RSp561e,2) + 1./81.*m44*
         RSp451e*RSp563e - 1./81.*m44*RSp451e*RSp561e + 1./81.*m44*mp2*
         RNep*RSp563e*RSe563p + 1./81.*m44*mp2*RNep*RSp563e*RSe453p + 1.
         /81.*m44*mp2*RNep*RSp453e*RSe563p - 1./81.*m44*mp2*RNep*
         RSe561p*RSp563e - 1./81.*m44*mp2*RNep*RSe561p*RSp453e - 1./81.
         *m44*mp2*RNep*RSe451p*RSp563e - 1./81.*m44*mp2*RNep*RSp561e*
         RSe563p - 1./81.*m44*mp2*RNep*RSp561e*RSe453p + 1./81.*m44*mp2
         *RNep*RSp561e*RSe561p );

      M13p6x23R +=  + Ab(m46,mu2)*Ab(mp2,mu2) * ( 1./81.*m44*mp2*RNep*
         RSp561e*RSe451p - 1./81.*m44*mp2*RNep*RSp451e*RSe563p + 1./81.
         *m44*mp2*RNep*RSp451e*RSe561p + 1./81.*m44*mp2*RN3p*RSp563e*
         RS345pe + 1./81.*m44*mp2*RN3p*RSp563e*RS356pe + 1./81.*m44*mp2
         *RN3p*RSp453e*RS356pe - 1./81.*m44*mp2*RN3p*RSp561e*RS345pe - 
         1./81.*m44*mp2*RN3p*RSp561e*RS356pe - 1./81.*m44*mp2*RN3p*
         RSp451e*RS356pe - 1./81.*m44*mp2*RN1p*RS145pe*RSp563e - 1./81.
         *m44*mp2*RN1p*RS156pe*RSp563e - 1./81.*m44*mp2*RN1p*RS156pe*
         RSp453e + 1./81.*m44*mp2*RN1p*RSp561e*RS145pe + 1./81.*m44*mp2
         *RN1p*RSp561e*RS156pe + 1./81.*m44*mp2*RN1p*RSp451e*RS156pe - 
         1./162.*m33*pow(RSp563e,2) + 1./81.*m33*RSp453e*RSp563e - 1./
         162.*m33*pow(RSp453e,2) - 1./81.*m33*RSp561e*RSp563e + 1./81.*
         m33*RSp561e*RSp453e + 1./81.*m33*RSp451e*RSp563e - 1./81.*m33*
         RSp451e*RSp453e + 1./81.*m33*mp2*RNep*RSp563e*RSe563p - 1./81.
         *m33*mp2*RNep*RSp563e*RSe453p - 1./81.*m33*mp2*RNep*RSp453e*
         RSe563p );

      M13p6x23R +=  + Ab(m46,mu2)*Ab(mp2,mu2) * ( 1./81.*m33*mp2*RNep*
         RSp453e*RSe453p + 1./81.*m33*mp2*RNep*RSe561p*RSp563e - 1./81.
         *m33*mp2*RNep*RSe561p*RSp453e - 1./81.*m33*mp2*RNep*RSe451p*
         RSp563e + 1./81.*m33*mp2*RNep*RSe451p*RSp453e + 1./81.*m33*mp2
         *RNep*RSp561e*RSe563p - 1./81.*m33*mp2*RNep*RSp561e*RSe453p - 
         1./81.*m33*mp2*RNep*RSp451e*RSe563p + 1./81.*m33*mp2*RNep*
         RSp451e*RSe453p - 2./81.*m33*mp2*RN3p*RSp561e*RS345pe + 2./81.
         *m33*mp2*RN3p*RSp561e*RS356pe + 2./81.*m33*mp2*RN3p*RSp451e*
         RS345pe - 2./81.*m33*mp2*RN3p*RSp451e*RS356pe - 1./81.*m33*mp2
         *RN1p*RS145pe*RSp563e + 1./81.*m33*mp2*RN1p*RS145pe*RSp453e + 
         1./81.*m33*mp2*RN1p*RS156pe*RSp563e - 1./81.*m33*mp2*RN1p*
         RS156pe*RSp453e - 1./81.*m33*m66*RNep*RSp563e*RSe453p - 1./81.
         *m33*m66*RNep*RSp453e*RSe563p - 1./81.*m33*m66*RNep*RSp453e*
         RSe453p - 1./81.*m33*m66*RNep*RSe561p*RSp453e - 1./81.*m33*m66
         *RNep*RSe451p*RSp563e - 1./81.*m33*m66*RNep*RSe451p*RSp453e - 
         1./81.*m33*m66*RNep*RSp561e*RSe453p );

      M13p6x23R +=  + Ab(m46,mu2)*Ab(mp2,mu2) * (  - 1./81.*m33*m66*
         RNep*RSp451e*RSe563p - 1./81.*m33*m66*RNep*RSp451e*RSe453p - 1.
         /81.*m33*m66*RN3p*RSp563e*RS345pe - 1./81.*m33*m66*RN3p*
         RSp453e*RS345pe - 1./81.*m33*m66*RN3p*RSp453e*RS356pe - 1./81.
         *m33*m66*RN3p*RSp561e*RS345pe - 1./81.*m33*m66*RN3p*RSp451e*
         RS345pe - 1./81.*m33*m66*RN3p*RSp451e*RS356pe - 1./81.*m33*m66
         *RN1p*RS145pe*RSp563e - 1./81.*m33*m66*RN1p*RS145pe*RSp453e - 
         1./81.*m33*m66*RN1p*RS156pe*RSp453e - 1./81.*m33*m44*RNep*
         RSp563e*RSe563p - 1./81.*m33*m44*RNep*RSp563e*RSe453p - 1./81.
         *m33*m44*RNep*RSp453e*RSe563p - 1./81.*m33*m44*RNep*RSe561p*
         RSp563e - 1./81.*m33*m44*RNep*RSe561p*RSp453e - 1./81.*m33*m44
         *RNep*RSe451p*RSp563e - 1./81.*m33*m44*RNep*RSp561e*RSe563p - 
         1./81.*m33*m44*RNep*RSp561e*RSe453p - 1./81.*m33*m44*RNep*
         RSp451e*RSe563p - 1./81.*m33*m44*RN3p*RSp563e*RS345pe - 1./81.
         *m33*m44*RN3p*RSp563e*RS356pe - 1./81.*m33*m44*RN3p*RSp453e*
         RS356pe );

      M13p6x23R +=  + Ab(m46,mu2)*Ab(mp2,mu2) * (  - 1./81.*m33*m44*
         RN3p*RSp561e*RS345pe - 1./81.*m33*m44*RN3p*RSp561e*RS356pe - 1.
         /81.*m33*m44*RN3p*RSp451e*RS356pe - 1./81.*m33*m44*RN1p*
         RS145pe*RSp563e - 1./81.*m33*m44*RN1p*RS156pe*RSp563e - 1./81.
         *m33*m44*RN1p*RS156pe*RSp453e - 1./81.*m11*RSp561e*RSp563e + 1.
         /81.*m11*RSp561e*RSp453e - 1./162.*m11*pow(RSp561e,2) + 1./81.
         *m11*RSp451e*RSp563e - 1./81.*m11*RSp451e*RSp453e + 1./81.*m11
         *RSp451e*RSp561e - 1./162.*m11*pow(RSp451e,2) + 1./81.*m11*mp2
         *RNep*RSe561p*RSp563e - 1./81.*m11*mp2*RNep*RSe561p*RSp453e - 
         1./81.*m11*mp2*RNep*RSe451p*RSp563e + 1./81.*m11*mp2*RNep*
         RSe451p*RSp453e + 1./81.*m11*mp2*RNep*RSp561e*RSe563p - 1./81.
         *m11*mp2*RNep*RSp561e*RSe453p + 1./81.*m11*mp2*RNep*RSp561e*
         RSe561p - 1./81.*m11*mp2*RNep*RSp561e*RSe451p - 1./81.*m11*mp2
         *RNep*RSp451e*RSe563p + 1./81.*m11*mp2*RNep*RSp451e*RSe453p - 
         1./81.*m11*mp2*RNep*RSp451e*RSe561p + 1./81.*m11*mp2*RNep*
         RSp451e*RSe451p );

      M13p6x23R +=  + Ab(m46,mu2)*Ab(mp2,mu2) * (  - 1./81.*m11*mp2*
         RN3p*RSp561e*RS345pe + 1./81.*m11*mp2*RN3p*RSp561e*RS356pe + 1.
         /81.*m11*mp2*RN3p*RSp451e*RS345pe - 1./81.*m11*mp2*RN3p*
         RSp451e*RS356pe - 2./81.*m11*mp2*RN1p*RS145pe*RSp563e + 2./81.
         *m11*mp2*RN1p*RS145pe*RSp453e + 2./81.*m11*mp2*RN1p*RS156pe*
         RSp563e - 2./81.*m11*mp2*RN1p*RS156pe*RSp453e - 1./81.*m11*m66
         *RNep*RSe561p*RSp453e - 1./81.*m11*m66*RNep*RSe451p*RSp563e - 
         1./81.*m11*m66*RNep*RSe451p*RSp453e - 1./81.*m11*m66*RNep*
         RSp561e*RSe453p - 1./81.*m11*m66*RNep*RSp561e*RSe451p - 1./81.
         *m11*m66*RNep*RSp451e*RSe563p - 1./81.*m11*m66*RNep*RSp451e*
         RSe453p - 1./81.*m11*m66*RNep*RSp451e*RSe561p - 1./81.*m11*m66
         *RNep*RSp451e*RSe451p - 1./81.*m11*m66*RN3p*RSp561e*RS345pe - 
         1./81.*m11*m66*RN3p*RSp451e*RS345pe - 1./81.*m11*m66*RN3p*
         RSp451e*RS356pe - 1./81.*m11*m66*RN1p*RS145pe*RSp563e - 1./81.
         *m11*m66*RN1p*RS145pe*RSp453e - 1./81.*m11*m66*RN1p*RS156pe*
         RSp453e );

      M13p6x23R +=  + Ab(m46,mu2)*Ab(mp2,mu2) * (  - 1./81.*m11*m66*
         RN1p*RSp561e*RS145pe - 1./81.*m11*m66*RN1p*RSp451e*RS145pe - 1.
         /81.*m11*m66*RN1p*RSp451e*RS156pe - 1./81.*m11*m44*RNep*
         RSe561p*RSp563e - 1./81.*m11*m44*RNep*RSe561p*RSp453e - 1./81.
         *m11*m44*RNep*RSe451p*RSp563e - 1./81.*m11*m44*RNep*RSp561e*
         RSe563p - 1./81.*m11*m44*RNep*RSp561e*RSe453p - 1./81.*m11*m44
         *RNep*RSp561e*RSe561p - 1./81.*m11*m44*RNep*RSp561e*RSe451p - 
         1./81.*m11*m44*RNep*RSp451e*RSe563p - 1./81.*m11*m44*RNep*
         RSp451e*RSe561p - 1./81.*m11*m44*RN3p*RSp561e*RS345pe - 1./81.
         *m11*m44*RN3p*RSp561e*RS356pe - 1./81.*m11*m44*RN3p*RSp451e*
         RS356pe - 1./81.*m11*m44*RN1p*RS145pe*RSp563e - 1./81.*m11*m44
         *RN1p*RS156pe*RSp563e - 1./81.*m11*m44*RN1p*RS156pe*RSp453e - 
         1./81.*m11*m44*RN1p*RSp561e*RS145pe - 1./81.*m11*m44*RN1p*
         RSp561e*RS156pe - 1./81.*m11*m44*RN1p*RSp451e*RS156pe );

      M13p6x23R +=  + Ab(m46,mu2)*Bb(me2,mu2) * ( 1./162.*pow(me2,2)*
         pow(RSe563p,2) - 1./81.*pow(me2,2)*RSe453p*RSe563p + 1./162.*
         pow(me2,2)*pow(RSe453p,2) - 1./81.*pow(me2,2)*RSe561p*RSe563p
          + 1./81.*pow(me2,2)*RSe561p*RSe453p + 1./162.*pow(me2,2)*pow(
         RSe561p,2) + 1./81.*pow(me2,2)*RSe451p*RSe563p - 1./81.*pow(
         me2,2)*RSe451p*RSe453p - 1./81.*pow(me2,2)*RSe451p*RSe561p + 1.
         /162.*pow(me2,2)*pow(RSe451p,2) - 1./81.*m66*me2*RSe453p*
         RSe563p - 1./162.*m66*me2*pow(RSe453p,2) + 1./81.*m66*me2*
         RSe561p*RSe453p + 1./81.*m66*me2*RSe451p*RSe563p + 1./81.*m66*
         me2*RSe451p*RSe453p - 1./81.*m66*me2*RSe451p*RSe561p - 1./162.
         *m66*me2*pow(RSe451p,2) - 1./162.*m44*me2*pow(RSe563p,2) - 1./
         81.*m44*me2*RSe453p*RSe563p + 1./81.*m44*me2*RSe561p*RSe563p
          + 1./81.*m44*me2*RSe561p*RSe453p - 1./162.*m44*me2*pow(
         RSe561p,2) + 1./81.*m44*me2*RSe451p*RSe563p - 1./81.*m44*me2*
         RSe451p*RSe561p - 1./162.*m33*me2*pow(RSe563p,2) + 1./81.*m33*
         me2*RSe453p*RSe563p );

      M13p6x23R +=  + Ab(m46,mu2)*Bb(me2,mu2) * (  - 1./162.*m33*me2*
         pow(RSe453p,2) - 1./81.*m33*me2*RSe561p*RSe563p + 1./81.*m33*
         me2*RSe561p*RSe453p + 1./81.*m33*me2*RSe451p*RSe563p - 1./81.*
         m33*me2*RSe451p*RSe453p + 1./81.*m33*m66*RSe453p*RSe563p + 1./
         162.*m33*m66*pow(RSe453p,2) + 1./81.*m33*m66*RSe561p*RSe453p
          + 1./81.*m33*m66*RSe451p*RSe563p + 1./81.*m33*m66*RSe451p*
         RSe453p + 1./162.*m33*m44*pow(RSe563p,2) + 1./81.*m33*m44*
         RSe453p*RSe563p + 1./81.*m33*m44*RSe561p*RSe563p + 1./81.*m33*
         m44*RSe561p*RSe453p + 1./81.*m33*m44*RSe451p*RSe563p - 1./81.*
         m11*me2*RSe561p*RSe563p + 1./81.*m11*me2*RSe561p*RSe453p - 1./
         162.*m11*me2*pow(RSe561p,2) + 1./81.*m11*me2*RSe451p*RSe563p
          - 1./81.*m11*me2*RSe451p*RSe453p + 1./81.*m11*me2*RSe451p*
         RSe561p - 1./162.*m11*me2*pow(RSe451p,2) + 1./81.*m11*m66*
         RSe561p*RSe453p + 1./81.*m11*m66*RSe451p*RSe563p + 1./81.*m11*
         m66*RSe451p*RSe453p + 1./81.*m11*m66*RSe451p*RSe561p + 1./162.
         *m11*m66*pow(RSe451p,2) );

      M13p6x23R +=  + Ab(m46,mu2)*Bb(me2,mu2) * ( 1./81.*m11*m44*
         RSe561p*RSe563p + 1./81.*m11*m44*RSe561p*RSe453p + 1./162.*m11
         *m44*pow(RSe561p,2) + 1./81.*m11*m44*RSe451p*RSe563p + 1./81.*
         m11*m44*RSe451p*RSe561p );

      M13p6x23R +=  + Ab(m46,mu2)*Bb(mp2,mu2) * ( 1./162.*pow(mp2,2)*
         pow(RSp563e,2) - 1./81.*pow(mp2,2)*RSp453e*RSp563e + 1./162.*
         pow(mp2,2)*pow(RSp453e,2) - 1./81.*pow(mp2,2)*RSp561e*RSp563e
          + 1./81.*pow(mp2,2)*RSp561e*RSp453e + 1./162.*pow(mp2,2)*pow(
         RSp561e,2) + 1./81.*pow(mp2,2)*RSp451e*RSp563e - 1./81.*pow(
         mp2,2)*RSp451e*RSp453e - 1./81.*pow(mp2,2)*RSp451e*RSp561e + 1.
         /162.*pow(mp2,2)*pow(RSp451e,2) - 1./81.*m66*mp2*RSp453e*
         RSp563e - 1./162.*m66*mp2*pow(RSp453e,2) + 1./81.*m66*mp2*
         RSp561e*RSp453e + 1./81.*m66*mp2*RSp451e*RSp563e + 1./81.*m66*
         mp2*RSp451e*RSp453e - 1./81.*m66*mp2*RSp451e*RSp561e - 1./162.
         *m66*mp2*pow(RSp451e,2) - 1./162.*m44*mp2*pow(RSp563e,2) - 1./
         81.*m44*mp2*RSp453e*RSp563e + 1./81.*m44*mp2*RSp561e*RSp563e
          + 1./81.*m44*mp2*RSp561e*RSp453e - 1./162.*m44*mp2*pow(
         RSp561e,2) + 1./81.*m44*mp2*RSp451e*RSp563e - 1./81.*m44*mp2*
         RSp451e*RSp561e - 1./162.*m33*mp2*pow(RSp563e,2) + 1./81.*m33*
         mp2*RSp453e*RSp563e );

      M13p6x23R +=  + Ab(m46,mu2)*Bb(mp2,mu2) * (  - 1./162.*m33*mp2*
         pow(RSp453e,2) - 1./81.*m33*mp2*RSp561e*RSp563e + 1./81.*m33*
         mp2*RSp561e*RSp453e + 1./81.*m33*mp2*RSp451e*RSp563e - 1./81.*
         m33*mp2*RSp451e*RSp453e + 1./81.*m33*m66*RSp453e*RSp563e + 1./
         162.*m33*m66*pow(RSp453e,2) + 1./81.*m33*m66*RSp561e*RSp453e
          + 1./81.*m33*m66*RSp451e*RSp563e + 1./81.*m33*m66*RSp451e*
         RSp453e + 1./162.*m33*m44*pow(RSp563e,2) + 1./81.*m33*m44*
         RSp453e*RSp563e + 1./81.*m33*m44*RSp561e*RSp563e + 1./81.*m33*
         m44*RSp561e*RSp453e + 1./81.*m33*m44*RSp451e*RSp563e - 1./81.*
         m11*mp2*RSp561e*RSp563e + 1./81.*m11*mp2*RSp561e*RSp453e - 1./
         162.*m11*mp2*pow(RSp561e,2) + 1./81.*m11*mp2*RSp451e*RSp563e
          - 1./81.*m11*mp2*RSp451e*RSp453e + 1./81.*m11*mp2*RSp451e*
         RSp561e - 1./162.*m11*mp2*pow(RSp451e,2) + 1./81.*m11*m66*
         RSp561e*RSp453e + 1./81.*m11*m66*RSp451e*RSp563e + 1./81.*m11*
         m66*RSp451e*RSp453e + 1./81.*m11*m66*RSp451e*RSp561e + 1./162.
         *m11*m66*pow(RSp451e,2) );

      M13p6x23R +=  + Ab(m46,mu2)*Bb(mp2,mu2) * ( 1./81.*m11*m44*
         RSp561e*RSp563e + 1./81.*m11*m44*RSp561e*RSp453e + 1./162.*m11
         *m44*pow(RSp561e,2) + 1./81.*m11*m44*RSp451e*RSp563e + 1./81.*
         m11*m44*RSp451e*RSp561e );

      M13p6x23R +=  + Ab(m55,mu2)*Ab(me2,mu2) * (  - 1./54.*m55*pow(
         RSe463p,2) - 1./27.*m55*RSp463e*RSe463p + 1./162.*m55*RS546pe*
         pow(RSe463p,2) + 1./81.*m55*RS546pe*RSp463e*RSe463p + 1./27.*
         m55*RSe461p*RSe463p + 1./27.*m55*RSe461p*RSp463e - 1./81.*m55*
         RSe461p*RS546pe*RSe463p - 1./81.*m55*RSe461p*RS546pe*RSp463e
          - 1./54.*m55*pow(RSe461p,2) + 1./162.*m55*pow(RSe461p,2)*
         RS546pe + 1./27.*m55*RSp461e*RSe463p - 1./81.*m55*RSp461e*
         RS546pe*RSe463p - 1./27.*m55*RSp461e*RSe461p + 1./81.*m55*
         RSp461e*RSe461p*RS546pe - 1./27.*m55*mp2*RNep*RSp463e*RSe463p
          + 1./81.*m55*mp2*RNep*RS546pe*RSp463e*RSe463p + 1./27.*m55*
         mp2*RNep*RSe461p*RSp463e - 1./81.*m55*mp2*RNep*RSe461p*RS546pe
         *RSp463e + 1./27.*m55*mp2*RNep*RSp461e*RSe463p - 1./81.*m55*
         mp2*RNep*RSp461e*RS546pe*RSe463p - 1./27.*m55*mp2*RNep*RSp461e
         *RSe461p + 1./81.*m55*mp2*RNep*RSp461e*RSe461p*RS546pe + 1./27.
         *m55*me2*RN3e*RSe463p*RS346pe - 1./81.*m55*me2*RN3e*RS546pe*
         RSe463p*RS346pe );

      M13p6x23R +=  + Ab(m55,mu2)*Ab(me2,mu2) * (  - 1./27.*m55*me2*
         RN3e*RSe461p*RS346pe + 1./81.*m55*me2*RN3e*RSe461p*RS546pe*
         RS346pe - 1./27.*m55*me2*RN1e*RS146pe*RSe463p + 1./81.*m55*me2
         *RN1e*RS546pe*RS146pe*RSe463p + 1./27.*m55*me2*RN1e*RSe461p*
         RS146pe - 1./81.*m55*me2*RN1e*RSe461p*RS546pe*RS146pe + 1./27.
         *m33*m55*RNep*RSp463e*RSe463p - 1./81.*m33*m55*RNep*RS546pe*
         RSp463e*RSe463p + 1./27.*m33*m55*RNep*RSe461p*RSp463e - 1./81.
         *m33*m55*RNep*RSe461p*RS546pe*RSp463e + 1./27.*m33*m55*RNep*
         RSp461e*RSe463p - 1./81.*m33*m55*RNep*RSp461e*RS546pe*RSe463p
          - 1./27.*m33*m55*RN3e*RSe463p*RS346pe + 1./81.*m33*m55*RN3e*
         RS546pe*RSe463p*RS346pe - 1./27.*m33*m55*RN3e*RSe461p*RS346pe
          + 1./81.*m33*m55*RN3e*RSe461p*RS546pe*RS346pe - 1./27.*m33*
         m55*RN1e*RS146pe*RSe463p + 1./81.*m33*m55*RN1e*RS546pe*RS146pe
         *RSe463p + 1./27.*m11*m55*RNep*RSe461p*RSp463e - 1./81.*m11*
         m55*RNep*RSe461p*RS546pe*RSp463e + 1./27.*m11*m55*RNep*RSp461e
         *RSe463p );

      M13p6x23R +=  + Ab(m55,mu2)*Ab(me2,mu2) * (  - 1./81.*m11*m55*
         RNep*RSp461e*RS546pe*RSe463p + 1./27.*m11*m55*RNep*RSp461e*
         RSe461p - 1./81.*m11*m55*RNep*RSp461e*RSe461p*RS546pe - 1./27.
         *m11*m55*RN3e*RSe461p*RS346pe + 1./81.*m11*m55*RN3e*RSe461p*
         RS546pe*RS346pe - 1./27.*m11*m55*RN1e*RS146pe*RSe463p + 1./81.
         *m11*m55*RN1e*RS546pe*RS146pe*RSe463p - 1./27.*m11*m55*RN1e*
         RSe461p*RS146pe + 1./81.*m11*m55*RN1e*RSe461p*RS546pe*RS146pe
          );

      M13p6x23R +=  + Ab(m55,mu2)*Ab(mp2,mu2) * (  - 1./54.*m55*pow(
         RSp463e,2) + 1./162.*m55*RS546pe*pow(RSp463e,2) + 1./27.*m55*
         RSp461e*RSp463e - 1./81.*m55*RSp461e*RS546pe*RSp463e - 1./54.*
         m55*pow(RSp461e,2) + 1./162.*m55*pow(RSp461e,2)*RS546pe + 1./
         27.*m55*mp2*RNep*RSp463e*RSe463p - 1./81.*m55*mp2*RNep*RS546pe
         *RSp463e*RSe463p - 1./27.*m55*mp2*RNep*RSe461p*RSp463e + 1./81.
         *m55*mp2*RNep*RSe461p*RS546pe*RSp463e - 1./27.*m55*mp2*RNep*
         RSp461e*RSe463p + 1./81.*m55*mp2*RNep*RSp461e*RS546pe*RSe463p
          + 1./27.*m55*mp2*RNep*RSp461e*RSe461p - 1./81.*m55*mp2*RNep*
         RSp461e*RSe461p*RS546pe + 1./27.*m55*mp2*RN3p*RSp463e*RS346pe
          - 1./81.*m55*mp2*RN3p*RS546pe*RSp463e*RS346pe - 1./27.*m55*
         mp2*RN3p*RSp461e*RS346pe + 1./81.*m55*mp2*RN3p*RSp461e*RS546pe
         *RS346pe - 1./27.*m55*mp2*RN1p*RS146pe*RSp463e + 1./81.*m55*
         mp2*RN1p*RS546pe*RS146pe*RSp463e + 1./27.*m55*mp2*RN1p*RSp461e
         *RS146pe - 1./81.*m55*mp2*RN1p*RSp461e*RS546pe*RS146pe - 1./27.
         *m33*m55*RNep*RSp463e*RSe463p );

      M13p6x23R +=  + Ab(m55,mu2)*Ab(mp2,mu2) * ( 1./81.*m33*m55*RNep*
         RS546pe*RSp463e*RSe463p - 1./27.*m33*m55*RNep*RSe461p*RSp463e
          + 1./81.*m33*m55*RNep*RSe461p*RS546pe*RSp463e - 1./27.*m33*
         m55*RNep*RSp461e*RSe463p + 1./81.*m33*m55*RNep*RSp461e*RS546pe
         *RSe463p - 1./27.*m33*m55*RN3p*RSp463e*RS346pe + 1./81.*m33*
         m55*RN3p*RS546pe*RSp463e*RS346pe - 1./27.*m33*m55*RN3p*RSp461e
         *RS346pe + 1./81.*m33*m55*RN3p*RSp461e*RS546pe*RS346pe - 1./27.
         *m33*m55*RN1p*RS146pe*RSp463e + 1./81.*m33*m55*RN1p*RS546pe*
         RS146pe*RSp463e - 1./27.*m11*m55*RNep*RSe461p*RSp463e + 1./81.
         *m11*m55*RNep*RSe461p*RS546pe*RSp463e - 1./27.*m11*m55*RNep*
         RSp461e*RSe463p + 1./81.*m11*m55*RNep*RSp461e*RS546pe*RSe463p
          - 1./27.*m11*m55*RNep*RSp461e*RSe461p + 1./81.*m11*m55*RNep*
         RSp461e*RSe461p*RS546pe - 1./27.*m11*m55*RN3p*RSp461e*RS346pe
          + 1./81.*m11*m55*RN3p*RSp461e*RS546pe*RS346pe - 1./27.*m11*
         m55*RN1p*RS146pe*RSp463e + 1./81.*m11*m55*RN1p*RS546pe*RS146pe
         *RSp463e );

      M13p6x23R +=  + Ab(m55,mu2)*Ab(mp2,mu2) * (  - 1./27.*m11*m55*
         RN1p*RSp461e*RS146pe + 1./81.*m11*m55*RN1p*RSp461e*RS546pe*
         RS146pe );

      M13p6x23R +=  + Ab(m55,mu2)*Bb(me2,mu2) * (  - 1./54.*m55*me2*
         pow(RSe463p,2) + 1./162.*m55*me2*RS546pe*pow(RSe463p,2) + 1./
         27.*m55*me2*RSe461p*RSe463p - 1./81.*m55*me2*RSe461p*RS546pe*
         RSe463p - 1./54.*m55*me2*pow(RSe461p,2) + 1./162.*m55*me2*pow(
         RSe461p,2)*RS546pe + 1./54.*m33*m55*pow(RSe463p,2) - 1./162.*
         m33*m55*RS546pe*pow(RSe463p,2) + 1./27.*m33*m55*RSe461p*
         RSe463p - 1./81.*m33*m55*RSe461p*RS546pe*RSe463p + 1./27.*m11*
         m55*RSe461p*RSe463p - 1./81.*m11*m55*RSe461p*RS546pe*RSe463p
          + 1./54.*m11*m55*pow(RSe461p,2) - 1./162.*m11*m55*pow(
         RSe461p,2)*RS546pe );

      M13p6x23R +=  + Ab(m55,mu2)*Bb(mp2,mu2) * (  - 1./54.*m55*mp2*
         pow(RSp463e,2) + 1./162.*m55*mp2*RS546pe*pow(RSp463e,2) + 1./
         27.*m55*mp2*RSp461e*RSp463e - 1./81.*m55*mp2*RSp461e*RS546pe*
         RSp463e - 1./54.*m55*mp2*pow(RSp461e,2) + 1./162.*m55*mp2*pow(
         RSp461e,2)*RS546pe + 1./54.*m33*m55*pow(RSp463e,2) - 1./162.*
         m33*m55*RS546pe*pow(RSp463e,2) + 1./27.*m33*m55*RSp461e*
         RSp463e - 1./81.*m33*m55*RSp461e*RS546pe*RSp463e + 1./27.*m11*
         m55*RSp461e*RSp463e - 1./81.*m11*m55*RSp461e*RS546pe*RSp463e
          + 1./54.*m11*m55*pow(RSp461e,2) - 1./162.*m11*m55*pow(
         RSp461e,2)*RS546pe );

      M13p6x23R +=  + Ab(m56,mu2)*Ab(me2,mu2) * (  - 1./81.*me2*RSe463p
         *RS345pe + 1./81.*me2*RSe463p*RS346pe + 1./81.*me2*pow(
         RSe463p,2) + 1./81.*me2*RSe453p*RS345pe - 1./81.*me2*RSe453p*
         RS346pe - 2./81.*me2*RSe453p*RSe463p + 1./81.*me2*pow(
         RSe453p,2) + 1./81.*me2*RSp463e*RSe463p - 1./81.*me2*RSp463e*
         RSe453p - 1./81.*me2*RSp453e*RSe463p + 1./81.*me2*RSp453e*
         RSe453p + 1./81.*me2*RS145pe*RSe463p - 1./81.*me2*RS145pe*
         RSe453p - 1./81.*me2*RS146pe*RSe463p + 1./81.*me2*RS146pe*
         RSe453p + 1./81.*me2*RSe461p*RS345pe - 1./81.*me2*RSe461p*
         RS346pe - 2./81.*me2*RSe461p*RSe463p + 2./81.*me2*RSe461p*
         RSe453p - 1./81.*me2*RSe461p*RSp463e + 1./81.*me2*RSe461p*
         RSp453e - 1./81.*me2*RSe461p*RS145pe + 1./81.*me2*RSe461p*
         RS146pe + 1./81.*me2*pow(RSe461p,2) - 1./81.*me2*RSe451p*
         RS345pe + 1./81.*me2*RSe451p*RS346pe + 2./81.*me2*RSe451p*
         RSe463p - 2./81.*me2*RSe451p*RSe453p + 1./81.*me2*RSe451p*
         RSp463e );

      M13p6x23R +=  + Ab(m56,mu2)*Ab(me2,mu2) * (  - 1./81.*me2*RSe451p
         *RSp453e + 1./81.*me2*RSe451p*RS145pe - 1./81.*me2*RSe451p*
         RS146pe - 2./81.*me2*RSe451p*RSe461p + 1./81.*me2*pow(
         RSe451p,2) - 1./81.*me2*RSp461e*RSe463p + 1./81.*me2*RSp461e*
         RSe453p + 1./81.*me2*RSp461e*RSe461p - 1./81.*me2*RSp461e*
         RSe451p + 1./81.*me2*RSp451e*RSe463p - 1./81.*me2*RSp451e*
         RSe453p - 1./81.*me2*RSp451e*RSe461p + 1./81.*me2*RSp451e*
         RSe451p + 1./81.*me2*mp2*RNep*RSp463e*RSe463p - 1./81.*me2*mp2
         *RNep*RSp463e*RSe453p - 1./81.*me2*mp2*RNep*RSp453e*RSe463p + 
         1./81.*me2*mp2*RNep*RSp453e*RSe453p - 1./81.*me2*mp2*RNep*
         RSe461p*RSp463e + 1./81.*me2*mp2*RNep*RSe461p*RSp453e + 1./81.
         *me2*mp2*RNep*RSe451p*RSp463e - 1./81.*me2*mp2*RNep*RSe451p*
         RSp453e - 1./81.*me2*mp2*RNep*RSp461e*RSe463p + 1./81.*me2*mp2
         *RNep*RSp461e*RSe453p + 1./81.*me2*mp2*RNep*RSp461e*RSe461p - 
         1./81.*me2*mp2*RNep*RSp461e*RSe451p + 1./81.*me2*mp2*RNep*
         RSp451e*RSe463p );

      M13p6x23R +=  + Ab(m56,mu2)*Ab(me2,mu2) * (  - 1./81.*me2*mp2*
         RNep*RSp451e*RSe453p - 1./81.*me2*mp2*RNep*RSp451e*RSe461p + 1.
         /81.*me2*mp2*RNep*RSp451e*RSe451p - 1./81.*m66*RSe453p*RSe463p
          - 1./162.*m66*pow(RSe453p,2) - 1./81.*m66*RSp463e*RSe453p - 1.
         /81.*m66*RSp453e*RSe463p - 1./81.*m66*RSp453e*RSe453p + 1./81.
         *m66*RSe461p*RSe453p + 1./81.*m66*RSe461p*RSp453e + 1./81.*m66
         *RSe451p*RSe463p + 1./81.*m66*RSe451p*RSe453p + 1./81.*m66*
         RSe451p*RSp463e + 1./81.*m66*RSe451p*RSp453e - 1./81.*m66*
         RSe451p*RSe461p - 1./162.*m66*pow(RSe451p,2) + 1./81.*m66*
         RSp461e*RSe453p - 1./81.*m66*RSp461e*RSe451p + 1./81.*m66*
         RSp451e*RSe463p + 1./81.*m66*RSp451e*RSe453p - 1./81.*m66*
         RSp451e*RSe461p - 1./81.*m66*RSp451e*RSe451p - 1./81.*m66*mp2*
         RNep*RSp463e*RSe453p - 1./81.*m66*mp2*RNep*RSp453e*RSe463p - 1.
         /81.*m66*mp2*RNep*RSp453e*RSe453p + 1./81.*m66*mp2*RNep*
         RSe461p*RSp453e + 1./81.*m66*mp2*RNep*RSe451p*RSp463e + 1./81.
         *m66*mp2*RNep*RSe451p*RSp453e );

      M13p6x23R +=  + Ab(m56,mu2)*Ab(me2,mu2) * ( 1./81.*m66*mp2*RNep*
         RSp461e*RSe453p - 1./81.*m66*mp2*RNep*RSp461e*RSe451p + 1./81.
         *m66*mp2*RNep*RSp451e*RSe463p + 1./81.*m66*mp2*RNep*RSp451e*
         RSe453p - 1./81.*m66*mp2*RNep*RSp451e*RSe461p - 1./81.*m66*mp2
         *RNep*RSp451e*RSe451p + 1./81.*m66*me2*RN3e*RSe463p*RS345pe + 
         1./81.*m66*me2*RN3e*RSe453p*RS345pe + 1./81.*m66*me2*RN3e*
         RSe453p*RS346pe - 1./81.*m66*me2*RN3e*RSe461p*RS345pe - 1./81.
         *m66*me2*RN3e*RSe451p*RS345pe - 1./81.*m66*me2*RN3e*RSe451p*
         RS346pe - 1./81.*m66*me2*RN1e*RS145pe*RSe463p - 1./81.*m66*me2
         *RN1e*RS145pe*RSe453p - 1./81.*m66*me2*RN1e*RS146pe*RSe453p + 
         1./81.*m66*me2*RN1e*RSe461p*RS145pe + 1./81.*m66*me2*RN1e*
         RSe451p*RS145pe + 1./81.*m66*me2*RN1e*RSe451p*RS146pe - 1./162.
         *m55*pow(RSe463p,2) - 1./81.*m55*RSe453p*RSe463p - 1./81.*m55*
         RSp463e*RSe463p - 1./81.*m55*RSp463e*RSe453p - 1./81.*m55*
         RSp453e*RSe463p + 1./81.*m55*RSe461p*RSe463p + 1./81.*m55*
         RSe461p*RSe453p );

      M13p6x23R +=  + Ab(m56,mu2)*Ab(me2,mu2) * ( 1./81.*m55*RSe461p*
         RSp463e + 1./81.*m55*RSe461p*RSp453e - 1./162.*m55*pow(
         RSe461p,2) + 1./81.*m55*RSe451p*RSe463p + 1./81.*m55*RSe451p*
         RSp463e - 1./81.*m55*RSe451p*RSe461p + 1./81.*m55*RSp461e*
         RSe463p + 1./81.*m55*RSp461e*RSe453p - 1./81.*m55*RSp461e*
         RSe461p - 1./81.*m55*RSp461e*RSe451p + 1./81.*m55*RSp451e*
         RSe463p - 1./81.*m55*RSp451e*RSe461p - 1./81.*m55*mp2*RNep*
         RSp463e*RSe463p - 1./81.*m55*mp2*RNep*RSp463e*RSe453p - 1./81.
         *m55*mp2*RNep*RSp453e*RSe463p + 1./81.*m55*mp2*RNep*RSe461p*
         RSp463e + 1./81.*m55*mp2*RNep*RSe461p*RSp453e + 1./81.*m55*mp2
         *RNep*RSe451p*RSp463e + 1./81.*m55*mp2*RNep*RSp461e*RSe463p + 
         1./81.*m55*mp2*RNep*RSp461e*RSe453p - 1./81.*m55*mp2*RNep*
         RSp461e*RSe461p - 1./81.*m55*mp2*RNep*RSp461e*RSe451p + 1./81.
         *m55*mp2*RNep*RSp451e*RSe463p - 1./81.*m55*mp2*RNep*RSp451e*
         RSe461p + 1./81.*m55*me2*RN3e*RSe463p*RS345pe + 1./81.*m55*me2
         *RN3e*RSe463p*RS346pe );

      M13p6x23R +=  + Ab(m56,mu2)*Ab(me2,mu2) * ( 1./81.*m55*me2*RN3e*
         RSe453p*RS346pe - 1./81.*m55*me2*RN3e*RSe461p*RS345pe - 1./81.
         *m55*me2*RN3e*RSe461p*RS346pe - 1./81.*m55*me2*RN3e*RSe451p*
         RS346pe - 1./81.*m55*me2*RN1e*RS145pe*RSe463p - 1./81.*m55*me2
         *RN1e*RS146pe*RSe463p - 1./81.*m55*me2*RN1e*RS146pe*RSe453p + 
         1./81.*m55*me2*RN1e*RSe461p*RS145pe + 1./81.*m55*me2*RN1e*
         RSe461p*RS146pe + 1./81.*m55*me2*RN1e*RSe451p*RS146pe - 1./162.
         *m33*pow(RSe463p,2) + 1./81.*m33*RSe453p*RSe463p - 1./162.*m33
         *pow(RSe453p,2) - 1./81.*m33*RSp463e*RSe463p + 1./81.*m33*
         RSp463e*RSe453p + 1./81.*m33*RSp453e*RSe463p - 1./81.*m33*
         RSp453e*RSe453p - 1./81.*m33*RSe461p*RSe463p + 1./81.*m33*
         RSe461p*RSe453p - 1./81.*m33*RSe461p*RSp463e + 1./81.*m33*
         RSe461p*RSp453e + 1./81.*m33*RSe451p*RSe463p - 1./81.*m33*
         RSe451p*RSe453p + 1./81.*m33*RSe451p*RSp463e - 1./81.*m33*
         RSe451p*RSp453e - 1./81.*m33*RSp461e*RSe463p + 1./81.*m33*
         RSp461e*RSe453p );

      M13p6x23R +=  + Ab(m56,mu2)*Ab(me2,mu2) * ( 1./81.*m33*RSp451e*
         RSe463p - 1./81.*m33*RSp451e*RSe453p - 1./81.*m33*mp2*RNep*
         RSp463e*RSe463p + 1./81.*m33*mp2*RNep*RSp463e*RSe453p + 1./81.
         *m33*mp2*RNep*RSp453e*RSe463p - 1./81.*m33*mp2*RNep*RSp453e*
         RSe453p - 1./81.*m33*mp2*RNep*RSe461p*RSp463e + 1./81.*m33*mp2
         *RNep*RSe461p*RSp453e + 1./81.*m33*mp2*RNep*RSe451p*RSp463e - 
         1./81.*m33*mp2*RNep*RSe451p*RSp453e - 1./81.*m33*mp2*RNep*
         RSp461e*RSe463p + 1./81.*m33*mp2*RNep*RSp461e*RSe453p + 1./81.
         *m33*mp2*RNep*RSp451e*RSe463p - 1./81.*m33*mp2*RNep*RSp451e*
         RSe453p - 2./81.*m33*me2*RN3e*RSe461p*RS345pe + 2./81.*m33*me2
         *RN3e*RSe461p*RS346pe + 2./81.*m33*me2*RN3e*RSe451p*RS345pe - 
         2./81.*m33*me2*RN3e*RSe451p*RS346pe - 1./81.*m33*me2*RN1e*
         RS145pe*RSe463p + 1./81.*m33*me2*RN1e*RS145pe*RSe453p + 1./81.
         *m33*me2*RN1e*RS146pe*RSe463p - 1./81.*m33*me2*RN1e*RS146pe*
         RSe453p + 1./81.*m33*m66*RNep*RSp463e*RSe453p + 1./81.*m33*m66
         *RNep*RSp453e*RSe463p );

      M13p6x23R +=  + Ab(m56,mu2)*Ab(me2,mu2) * ( 1./81.*m33*m66*RNep*
         RSp453e*RSe453p + 1./81.*m33*m66*RNep*RSe461p*RSp453e + 1./81.
         *m33*m66*RNep*RSe451p*RSp463e + 1./81.*m33*m66*RNep*RSe451p*
         RSp453e + 1./81.*m33*m66*RNep*RSp461e*RSe453p + 1./81.*m33*m66
         *RNep*RSp451e*RSe463p + 1./81.*m33*m66*RNep*RSp451e*RSe453p - 
         1./81.*m33*m66*RN3e*RSe463p*RS345pe - 1./81.*m33*m66*RN3e*
         RSe453p*RS345pe - 1./81.*m33*m66*RN3e*RSe453p*RS346pe - 1./81.
         *m33*m66*RN3e*RSe461p*RS345pe - 1./81.*m33*m66*RN3e*RSe451p*
         RS345pe - 1./81.*m33*m66*RN3e*RSe451p*RS346pe - 1./81.*m33*m66
         *RN1e*RS145pe*RSe463p - 1./81.*m33*m66*RN1e*RS145pe*RSe453p - 
         1./81.*m33*m66*RN1e*RS146pe*RSe453p + 1./81.*m33*m55*RNep*
         RSp463e*RSe463p + 1./81.*m33*m55*RNep*RSp463e*RSe453p + 1./81.
         *m33*m55*RNep*RSp453e*RSe463p + 1./81.*m33*m55*RNep*RSe461p*
         RSp463e + 1./81.*m33*m55*RNep*RSe461p*RSp453e + 1./81.*m33*m55
         *RNep*RSe451p*RSp463e + 1./81.*m33*m55*RNep*RSp461e*RSe463p + 
         1./81.*m33*m55*RNep*RSp461e*RSe453p );

      M13p6x23R +=  + Ab(m56,mu2)*Ab(me2,mu2) * ( 1./81.*m33*m55*RNep*
         RSp451e*RSe463p - 1./81.*m33*m55*RN3e*RSe463p*RS345pe - 1./81.
         *m33*m55*RN3e*RSe463p*RS346pe - 1./81.*m33*m55*RN3e*RSe453p*
         RS346pe - 1./81.*m33*m55*RN3e*RSe461p*RS345pe - 1./81.*m33*m55
         *RN3e*RSe461p*RS346pe - 1./81.*m33*m55*RN3e*RSe451p*RS346pe - 
         1./81.*m33*m55*RN1e*RS145pe*RSe463p - 1./81.*m33*m55*RN1e*
         RS146pe*RSe463p - 1./81.*m33*m55*RN1e*RS146pe*RSe453p - 1./81.
         *m11*RSe461p*RSe463p + 1./81.*m11*RSe461p*RSe453p - 1./81.*m11
         *RSe461p*RSp463e + 1./81.*m11*RSe461p*RSp453e - 1./162.*m11*
         pow(RSe461p,2) + 1./81.*m11*RSe451p*RSe463p - 1./81.*m11*
         RSe451p*RSe453p + 1./81.*m11*RSe451p*RSp463e - 1./81.*m11*
         RSe451p*RSp453e + 1./81.*m11*RSe451p*RSe461p - 1./162.*m11*
         pow(RSe451p,2) - 1./81.*m11*RSp461e*RSe463p + 1./81.*m11*
         RSp461e*RSe453p - 1./81.*m11*RSp461e*RSe461p + 1./81.*m11*
         RSp461e*RSe451p + 1./81.*m11*RSp451e*RSe463p - 1./81.*m11*
         RSp451e*RSe453p );

      M13p6x23R +=  + Ab(m56,mu2)*Ab(me2,mu2) * ( 1./81.*m11*RSp451e*
         RSe461p - 1./81.*m11*RSp451e*RSe451p - 1./81.*m11*mp2*RNep*
         RSe461p*RSp463e + 1./81.*m11*mp2*RNep*RSe461p*RSp453e + 1./81.
         *m11*mp2*RNep*RSe451p*RSp463e - 1./81.*m11*mp2*RNep*RSe451p*
         RSp453e - 1./81.*m11*mp2*RNep*RSp461e*RSe463p + 1./81.*m11*mp2
         *RNep*RSp461e*RSe453p - 1./81.*m11*mp2*RNep*RSp461e*RSe461p + 
         1./81.*m11*mp2*RNep*RSp461e*RSe451p + 1./81.*m11*mp2*RNep*
         RSp451e*RSe463p - 1./81.*m11*mp2*RNep*RSp451e*RSe453p + 1./81.
         *m11*mp2*RNep*RSp451e*RSe461p - 1./81.*m11*mp2*RNep*RSp451e*
         RSe451p - 1./81.*m11*me2*RN3e*RSe461p*RS345pe + 1./81.*m11*me2
         *RN3e*RSe461p*RS346pe + 1./81.*m11*me2*RN3e*RSe451p*RS345pe - 
         1./81.*m11*me2*RN3e*RSe451p*RS346pe - 2./81.*m11*me2*RN1e*
         RS145pe*RSe463p + 2./81.*m11*me2*RN1e*RS145pe*RSe453p + 2./81.
         *m11*me2*RN1e*RS146pe*RSe463p - 2./81.*m11*me2*RN1e*RS146pe*
         RSe453p + 1./81.*m11*m66*RNep*RSe461p*RSp453e + 1./81.*m11*m66
         *RNep*RSe451p*RSp463e );

      M13p6x23R +=  + Ab(m56,mu2)*Ab(me2,mu2) * ( 1./81.*m11*m66*RNep*
         RSe451p*RSp453e + 1./81.*m11*m66*RNep*RSp461e*RSe453p + 1./81.
         *m11*m66*RNep*RSp461e*RSe451p + 1./81.*m11*m66*RNep*RSp451e*
         RSe463p + 1./81.*m11*m66*RNep*RSp451e*RSe453p + 1./81.*m11*m66
         *RNep*RSp451e*RSe461p + 1./81.*m11*m66*RNep*RSp451e*RSe451p - 
         1./81.*m11*m66*RN3e*RSe461p*RS345pe - 1./81.*m11*m66*RN3e*
         RSe451p*RS345pe - 1./81.*m11*m66*RN3e*RSe451p*RS346pe - 1./81.
         *m11*m66*RN1e*RS145pe*RSe463p - 1./81.*m11*m66*RN1e*RS145pe*
         RSe453p - 1./81.*m11*m66*RN1e*RS146pe*RSe453p - 1./81.*m11*m66
         *RN1e*RSe461p*RS145pe - 1./81.*m11*m66*RN1e*RSe451p*RS145pe - 
         1./81.*m11*m66*RN1e*RSe451p*RS146pe + 1./81.*m11*m55*RNep*
         RSe461p*RSp463e + 1./81.*m11*m55*RNep*RSe461p*RSp453e + 1./81.
         *m11*m55*RNep*RSe451p*RSp463e + 1./81.*m11*m55*RNep*RSp461e*
         RSe463p + 1./81.*m11*m55*RNep*RSp461e*RSe453p + 1./81.*m11*m55
         *RNep*RSp461e*RSe461p + 1./81.*m11*m55*RNep*RSp461e*RSe451p + 
         1./81.*m11*m55*RNep*RSp451e*RSe463p );

      M13p6x23R +=  + Ab(m56,mu2)*Ab(me2,mu2) * ( 1./81.*m11*m55*RNep*
         RSp451e*RSe461p - 1./81.*m11*m55*RN3e*RSe461p*RS345pe - 1./81.
         *m11*m55*RN3e*RSe461p*RS346pe - 1./81.*m11*m55*RN3e*RSe451p*
         RS346pe - 1./81.*m11*m55*RN1e*RS145pe*RSe463p - 1./81.*m11*m55
         *RN1e*RS146pe*RSe463p - 1./81.*m11*m55*RN1e*RS146pe*RSe453p - 
         1./81.*m11*m55*RN1e*RSe461p*RS145pe - 1./81.*m11*m55*RN1e*
         RSe461p*RS146pe - 1./81.*m11*m55*RN1e*RSe451p*RS146pe );

      M13p6x23R +=  + Ab(m56,mu2)*Ab(mp2,mu2) * (  - 1./81.*mp2*RSp463e
         *RS345pe + 1./81.*mp2*RSp463e*RS346pe + 1./81.*mp2*RSp463e*
         RSe463p - 1./81.*mp2*RSp463e*RSe453p + 1./81.*mp2*pow(
         RSp463e,2) + 1./81.*mp2*RSp453e*RS345pe - 1./81.*mp2*RSp453e*
         RS346pe - 1./81.*mp2*RSp453e*RSe463p + 1./81.*mp2*RSp453e*
         RSe453p - 2./81.*mp2*RSp453e*RSp463e + 1./81.*mp2*pow(
         RSp453e,2) + 1./81.*mp2*RS145pe*RSp463e - 1./81.*mp2*RS145pe*
         RSp453e - 1./81.*mp2*RS146pe*RSp463e + 1./81.*mp2*RS146pe*
         RSp453e - 1./81.*mp2*RSe461p*RSp463e + 1./81.*mp2*RSe461p*
         RSp453e + 1./81.*mp2*RSe451p*RSp463e - 1./81.*mp2*RSe451p*
         RSp453e + 1./81.*mp2*RSp461e*RS345pe - 1./81.*mp2*RSp461e*
         RS346pe - 1./81.*mp2*RSp461e*RSe463p + 1./81.*mp2*RSp461e*
         RSe453p - 2./81.*mp2*RSp461e*RSp463e + 2./81.*mp2*RSp461e*
         RSp453e - 1./81.*mp2*RSp461e*RS145pe + 1./81.*mp2*RSp461e*
         RS146pe + 1./81.*mp2*RSp461e*RSe461p - 1./81.*mp2*RSp461e*
         RSe451p );

      M13p6x23R +=  + Ab(m56,mu2)*Ab(mp2,mu2) * ( 1./81.*mp2*pow(
         RSp461e,2) - 1./81.*mp2*RSp451e*RS345pe + 1./81.*mp2*RSp451e*
         RS346pe + 1./81.*mp2*RSp451e*RSe463p - 1./81.*mp2*RSp451e*
         RSe453p + 2./81.*mp2*RSp451e*RSp463e - 2./81.*mp2*RSp451e*
         RSp453e + 1./81.*mp2*RSp451e*RS145pe - 1./81.*mp2*RSp451e*
         RS146pe - 1./81.*mp2*RSp451e*RSe461p + 1./81.*mp2*RSp451e*
         RSe451p - 2./81.*mp2*RSp451e*RSp461e + 1./81.*mp2*pow(
         RSp451e,2) - 1./81.*me2*mp2*RNep*RSp463e*RSe463p + 1./81.*me2*
         mp2*RNep*RSp463e*RSe453p + 1./81.*me2*mp2*RNep*RSp453e*RSe463p
          - 1./81.*me2*mp2*RNep*RSp453e*RSe453p + 1./81.*me2*mp2*RNep*
         RSe461p*RSp463e - 1./81.*me2*mp2*RNep*RSe461p*RSp453e - 1./81.
         *me2*mp2*RNep*RSe451p*RSp463e + 1./81.*me2*mp2*RNep*RSe451p*
         RSp453e + 1./81.*me2*mp2*RNep*RSp461e*RSe463p - 1./81.*me2*mp2
         *RNep*RSp461e*RSe453p - 1./81.*me2*mp2*RNep*RSp461e*RSe461p + 
         1./81.*me2*mp2*RNep*RSp461e*RSe451p - 1./81.*me2*mp2*RNep*
         RSp451e*RSe463p );

      M13p6x23R +=  + Ab(m56,mu2)*Ab(mp2,mu2) * ( 1./81.*me2*mp2*RNep*
         RSp451e*RSe453p + 1./81.*me2*mp2*RNep*RSp451e*RSe461p - 1./81.
         *me2*mp2*RNep*RSp451e*RSe451p - 1./81.*m66*RSp453e*RSp463e - 1.
         /162.*m66*pow(RSp453e,2) + 1./81.*m66*RSp461e*RSp453e + 1./81.
         *m66*RSp451e*RSp463e + 1./81.*m66*RSp451e*RSp453e - 1./81.*m66
         *RSp451e*RSp461e - 1./162.*m66*pow(RSp451e,2) + 1./81.*m66*mp2
         *RNep*RSp463e*RSe453p + 1./81.*m66*mp2*RNep*RSp453e*RSe463p + 
         1./81.*m66*mp2*RNep*RSp453e*RSe453p - 1./81.*m66*mp2*RNep*
         RSe461p*RSp453e - 1./81.*m66*mp2*RNep*RSe451p*RSp463e - 1./81.
         *m66*mp2*RNep*RSe451p*RSp453e - 1./81.*m66*mp2*RNep*RSp461e*
         RSe453p + 1./81.*m66*mp2*RNep*RSp461e*RSe451p - 1./81.*m66*mp2
         *RNep*RSp451e*RSe463p - 1./81.*m66*mp2*RNep*RSp451e*RSe453p + 
         1./81.*m66*mp2*RNep*RSp451e*RSe461p + 1./81.*m66*mp2*RNep*
         RSp451e*RSe451p + 1./81.*m66*mp2*RN3p*RSp463e*RS345pe + 1./81.
         *m66*mp2*RN3p*RSp453e*RS345pe + 1./81.*m66*mp2*RN3p*RSp453e*
         RS346pe );

      M13p6x23R +=  + Ab(m56,mu2)*Ab(mp2,mu2) * (  - 1./81.*m66*mp2*
         RN3p*RSp461e*RS345pe - 1./81.*m66*mp2*RN3p*RSp451e*RS345pe - 1.
         /81.*m66*mp2*RN3p*RSp451e*RS346pe - 1./81.*m66*mp2*RN1p*
         RS145pe*RSp463e - 1./81.*m66*mp2*RN1p*RS145pe*RSp453e - 1./81.
         *m66*mp2*RN1p*RS146pe*RSp453e + 1./81.*m66*mp2*RN1p*RSp461e*
         RS145pe + 1./81.*m66*mp2*RN1p*RSp451e*RS145pe + 1./81.*m66*mp2
         *RN1p*RSp451e*RS146pe - 1./162.*m55*pow(RSp463e,2) - 1./81.*
         m55*RSp453e*RSp463e + 1./81.*m55*RSp461e*RSp463e + 1./81.*m55*
         RSp461e*RSp453e - 1./162.*m55*pow(RSp461e,2) + 1./81.*m55*
         RSp451e*RSp463e - 1./81.*m55*RSp451e*RSp461e + 1./81.*m55*mp2*
         RNep*RSp463e*RSe463p + 1./81.*m55*mp2*RNep*RSp463e*RSe453p + 1.
         /81.*m55*mp2*RNep*RSp453e*RSe463p - 1./81.*m55*mp2*RNep*
         RSe461p*RSp463e - 1./81.*m55*mp2*RNep*RSe461p*RSp453e - 1./81.
         *m55*mp2*RNep*RSe451p*RSp463e - 1./81.*m55*mp2*RNep*RSp461e*
         RSe463p - 1./81.*m55*mp2*RNep*RSp461e*RSe453p + 1./81.*m55*mp2
         *RNep*RSp461e*RSe461p );

      M13p6x23R +=  + Ab(m56,mu2)*Ab(mp2,mu2) * ( 1./81.*m55*mp2*RNep*
         RSp461e*RSe451p - 1./81.*m55*mp2*RNep*RSp451e*RSe463p + 1./81.
         *m55*mp2*RNep*RSp451e*RSe461p + 1./81.*m55*mp2*RN3p*RSp463e*
         RS345pe + 1./81.*m55*mp2*RN3p*RSp463e*RS346pe + 1./81.*m55*mp2
         *RN3p*RSp453e*RS346pe - 1./81.*m55*mp2*RN3p*RSp461e*RS345pe - 
         1./81.*m55*mp2*RN3p*RSp461e*RS346pe - 1./81.*m55*mp2*RN3p*
         RSp451e*RS346pe - 1./81.*m55*mp2*RN1p*RS145pe*RSp463e - 1./81.
         *m55*mp2*RN1p*RS146pe*RSp463e - 1./81.*m55*mp2*RN1p*RS146pe*
         RSp453e + 1./81.*m55*mp2*RN1p*RSp461e*RS145pe + 1./81.*m55*mp2
         *RN1p*RSp461e*RS146pe + 1./81.*m55*mp2*RN1p*RSp451e*RS146pe - 
         1./162.*m33*pow(RSp463e,2) + 1./81.*m33*RSp453e*RSp463e - 1./
         162.*m33*pow(RSp453e,2) - 1./81.*m33*RSp461e*RSp463e + 1./81.*
         m33*RSp461e*RSp453e + 1./81.*m33*RSp451e*RSp463e - 1./81.*m33*
         RSp451e*RSp453e + 1./81.*m33*mp2*RNep*RSp463e*RSe463p - 1./81.
         *m33*mp2*RNep*RSp463e*RSe453p - 1./81.*m33*mp2*RNep*RSp453e*
         RSe463p );

      M13p6x23R +=  + Ab(m56,mu2)*Ab(mp2,mu2) * ( 1./81.*m33*mp2*RNep*
         RSp453e*RSe453p + 1./81.*m33*mp2*RNep*RSe461p*RSp463e - 1./81.
         *m33*mp2*RNep*RSe461p*RSp453e - 1./81.*m33*mp2*RNep*RSe451p*
         RSp463e + 1./81.*m33*mp2*RNep*RSe451p*RSp453e + 1./81.*m33*mp2
         *RNep*RSp461e*RSe463p - 1./81.*m33*mp2*RNep*RSp461e*RSe453p - 
         1./81.*m33*mp2*RNep*RSp451e*RSe463p + 1./81.*m33*mp2*RNep*
         RSp451e*RSe453p - 2./81.*m33*mp2*RN3p*RSp461e*RS345pe + 2./81.
         *m33*mp2*RN3p*RSp461e*RS346pe + 2./81.*m33*mp2*RN3p*RSp451e*
         RS345pe - 2./81.*m33*mp2*RN3p*RSp451e*RS346pe - 1./81.*m33*mp2
         *RN1p*RS145pe*RSp463e + 1./81.*m33*mp2*RN1p*RS145pe*RSp453e + 
         1./81.*m33*mp2*RN1p*RS146pe*RSp463e - 1./81.*m33*mp2*RN1p*
         RS146pe*RSp453e - 1./81.*m33*m66*RNep*RSp463e*RSe453p - 1./81.
         *m33*m66*RNep*RSp453e*RSe463p - 1./81.*m33*m66*RNep*RSp453e*
         RSe453p - 1./81.*m33*m66*RNep*RSe461p*RSp453e - 1./81.*m33*m66
         *RNep*RSe451p*RSp463e - 1./81.*m33*m66*RNep*RSe451p*RSp453e - 
         1./81.*m33*m66*RNep*RSp461e*RSe453p );

      M13p6x23R +=  + Ab(m56,mu2)*Ab(mp2,mu2) * (  - 1./81.*m33*m66*
         RNep*RSp451e*RSe463p - 1./81.*m33*m66*RNep*RSp451e*RSe453p - 1.
         /81.*m33*m66*RN3p*RSp463e*RS345pe - 1./81.*m33*m66*RN3p*
         RSp453e*RS345pe - 1./81.*m33*m66*RN3p*RSp453e*RS346pe - 1./81.
         *m33*m66*RN3p*RSp461e*RS345pe - 1./81.*m33*m66*RN3p*RSp451e*
         RS345pe - 1./81.*m33*m66*RN3p*RSp451e*RS346pe - 1./81.*m33*m66
         *RN1p*RS145pe*RSp463e - 1./81.*m33*m66*RN1p*RS145pe*RSp453e - 
         1./81.*m33*m66*RN1p*RS146pe*RSp453e - 1./81.*m33*m55*RNep*
         RSp463e*RSe463p - 1./81.*m33*m55*RNep*RSp463e*RSe453p - 1./81.
         *m33*m55*RNep*RSp453e*RSe463p - 1./81.*m33*m55*RNep*RSe461p*
         RSp463e - 1./81.*m33*m55*RNep*RSe461p*RSp453e - 1./81.*m33*m55
         *RNep*RSe451p*RSp463e - 1./81.*m33*m55*RNep*RSp461e*RSe463p - 
         1./81.*m33*m55*RNep*RSp461e*RSe453p - 1./81.*m33*m55*RNep*
         RSp451e*RSe463p - 1./81.*m33*m55*RN3p*RSp463e*RS345pe - 1./81.
         *m33*m55*RN3p*RSp463e*RS346pe - 1./81.*m33*m55*RN3p*RSp453e*
         RS346pe );

      M13p6x23R +=  + Ab(m56,mu2)*Ab(mp2,mu2) * (  - 1./81.*m33*m55*
         RN3p*RSp461e*RS345pe - 1./81.*m33*m55*RN3p*RSp461e*RS346pe - 1.
         /81.*m33*m55*RN3p*RSp451e*RS346pe - 1./81.*m33*m55*RN1p*
         RS145pe*RSp463e - 1./81.*m33*m55*RN1p*RS146pe*RSp463e - 1./81.
         *m33*m55*RN1p*RS146pe*RSp453e - 1./81.*m11*RSp461e*RSp463e + 1.
         /81.*m11*RSp461e*RSp453e - 1./162.*m11*pow(RSp461e,2) + 1./81.
         *m11*RSp451e*RSp463e - 1./81.*m11*RSp451e*RSp453e + 1./81.*m11
         *RSp451e*RSp461e - 1./162.*m11*pow(RSp451e,2) + 1./81.*m11*mp2
         *RNep*RSe461p*RSp463e - 1./81.*m11*mp2*RNep*RSe461p*RSp453e - 
         1./81.*m11*mp2*RNep*RSe451p*RSp463e + 1./81.*m11*mp2*RNep*
         RSe451p*RSp453e + 1./81.*m11*mp2*RNep*RSp461e*RSe463p - 1./81.
         *m11*mp2*RNep*RSp461e*RSe453p + 1./81.*m11*mp2*RNep*RSp461e*
         RSe461p - 1./81.*m11*mp2*RNep*RSp461e*RSe451p - 1./81.*m11*mp2
         *RNep*RSp451e*RSe463p + 1./81.*m11*mp2*RNep*RSp451e*RSe453p - 
         1./81.*m11*mp2*RNep*RSp451e*RSe461p + 1./81.*m11*mp2*RNep*
         RSp451e*RSe451p );

      M13p6x23R +=  + Ab(m56,mu2)*Ab(mp2,mu2) * (  - 1./81.*m11*mp2*
         RN3p*RSp461e*RS345pe + 1./81.*m11*mp2*RN3p*RSp461e*RS346pe + 1.
         /81.*m11*mp2*RN3p*RSp451e*RS345pe - 1./81.*m11*mp2*RN3p*
         RSp451e*RS346pe - 2./81.*m11*mp2*RN1p*RS145pe*RSp463e + 2./81.
         *m11*mp2*RN1p*RS145pe*RSp453e + 2./81.*m11*mp2*RN1p*RS146pe*
         RSp463e - 2./81.*m11*mp2*RN1p*RS146pe*RSp453e - 1./81.*m11*m66
         *RNep*RSe461p*RSp453e - 1./81.*m11*m66*RNep*RSe451p*RSp463e - 
         1./81.*m11*m66*RNep*RSe451p*RSp453e - 1./81.*m11*m66*RNep*
         RSp461e*RSe453p - 1./81.*m11*m66*RNep*RSp461e*RSe451p - 1./81.
         *m11*m66*RNep*RSp451e*RSe463p - 1./81.*m11*m66*RNep*RSp451e*
         RSe453p - 1./81.*m11*m66*RNep*RSp451e*RSe461p - 1./81.*m11*m66
         *RNep*RSp451e*RSe451p - 1./81.*m11*m66*RN3p*RSp461e*RS345pe - 
         1./81.*m11*m66*RN3p*RSp451e*RS345pe - 1./81.*m11*m66*RN3p*
         RSp451e*RS346pe - 1./81.*m11*m66*RN1p*RS145pe*RSp463e - 1./81.
         *m11*m66*RN1p*RS145pe*RSp453e - 1./81.*m11*m66*RN1p*RS146pe*
         RSp453e );

      M13p6x23R +=  + Ab(m56,mu2)*Ab(mp2,mu2) * (  - 1./81.*m11*m66*
         RN1p*RSp461e*RS145pe - 1./81.*m11*m66*RN1p*RSp451e*RS145pe - 1.
         /81.*m11*m66*RN1p*RSp451e*RS146pe - 1./81.*m11*m55*RNep*
         RSe461p*RSp463e - 1./81.*m11*m55*RNep*RSe461p*RSp453e - 1./81.
         *m11*m55*RNep*RSe451p*RSp463e - 1./81.*m11*m55*RNep*RSp461e*
         RSe463p - 1./81.*m11*m55*RNep*RSp461e*RSe453p - 1./81.*m11*m55
         *RNep*RSp461e*RSe461p - 1./81.*m11*m55*RNep*RSp461e*RSe451p - 
         1./81.*m11*m55*RNep*RSp451e*RSe463p - 1./81.*m11*m55*RNep*
         RSp451e*RSe461p - 1./81.*m11*m55*RN3p*RSp461e*RS345pe - 1./81.
         *m11*m55*RN3p*RSp461e*RS346pe - 1./81.*m11*m55*RN3p*RSp451e*
         RS346pe - 1./81.*m11*m55*RN1p*RS145pe*RSp463e - 1./81.*m11*m55
         *RN1p*RS146pe*RSp463e - 1./81.*m11*m55*RN1p*RS146pe*RSp453e - 
         1./81.*m11*m55*RN1p*RSp461e*RS145pe - 1./81.*m11*m55*RN1p*
         RSp461e*RS146pe - 1./81.*m11*m55*RN1p*RSp451e*RS146pe );

      M13p6x23R +=  + Ab(m56,mu2)*Bb(me2,mu2) * ( 1./162.*pow(me2,2)*
         pow(RSe463p,2) - 1./81.*pow(me2,2)*RSe453p*RSe463p + 1./162.*
         pow(me2,2)*pow(RSe453p,2) - 1./81.*pow(me2,2)*RSe461p*RSe463p
          + 1./81.*pow(me2,2)*RSe461p*RSe453p + 1./162.*pow(me2,2)*pow(
         RSe461p,2) + 1./81.*pow(me2,2)*RSe451p*RSe463p - 1./81.*pow(
         me2,2)*RSe451p*RSe453p - 1./81.*pow(me2,2)*RSe451p*RSe461p + 1.
         /162.*pow(me2,2)*pow(RSe451p,2) - 1./81.*m66*me2*RSe453p*
         RSe463p - 1./162.*m66*me2*pow(RSe453p,2) + 1./81.*m66*me2*
         RSe461p*RSe453p + 1./81.*m66*me2*RSe451p*RSe463p + 1./81.*m66*
         me2*RSe451p*RSe453p - 1./81.*m66*me2*RSe451p*RSe461p - 1./162.
         *m66*me2*pow(RSe451p,2) - 1./162.*m55*me2*pow(RSe463p,2) - 1./
         81.*m55*me2*RSe453p*RSe463p + 1./81.*m55*me2*RSe461p*RSe463p
          + 1./81.*m55*me2*RSe461p*RSe453p - 1./162.*m55*me2*pow(
         RSe461p,2) + 1./81.*m55*me2*RSe451p*RSe463p - 1./81.*m55*me2*
         RSe451p*RSe461p - 1./162.*m33*me2*pow(RSe463p,2) + 1./81.*m33*
         me2*RSe453p*RSe463p );

      M13p6x23R +=  + Ab(m56,mu2)*Bb(me2,mu2) * (  - 1./162.*m33*me2*
         pow(RSe453p,2) - 1./81.*m33*me2*RSe461p*RSe463p + 1./81.*m33*
         me2*RSe461p*RSe453p + 1./81.*m33*me2*RSe451p*RSe463p - 1./81.*
         m33*me2*RSe451p*RSe453p + 1./81.*m33*m66*RSe453p*RSe463p + 1./
         162.*m33*m66*pow(RSe453p,2) + 1./81.*m33*m66*RSe461p*RSe453p
          + 1./81.*m33*m66*RSe451p*RSe463p + 1./81.*m33*m66*RSe451p*
         RSe453p + 1./162.*m33*m55*pow(RSe463p,2) + 1./81.*m33*m55*
         RSe453p*RSe463p + 1./81.*m33*m55*RSe461p*RSe463p + 1./81.*m33*
         m55*RSe461p*RSe453p + 1./81.*m33*m55*RSe451p*RSe463p - 1./81.*
         m11*me2*RSe461p*RSe463p + 1./81.*m11*me2*RSe461p*RSe453p - 1./
         162.*m11*me2*pow(RSe461p,2) + 1./81.*m11*me2*RSe451p*RSe463p
          - 1./81.*m11*me2*RSe451p*RSe453p + 1./81.*m11*me2*RSe451p*
         RSe461p - 1./162.*m11*me2*pow(RSe451p,2) + 1./81.*m11*m66*
         RSe461p*RSe453p + 1./81.*m11*m66*RSe451p*RSe463p + 1./81.*m11*
         m66*RSe451p*RSe453p + 1./81.*m11*m66*RSe451p*RSe461p + 1./162.
         *m11*m66*pow(RSe451p,2) );

      M13p6x23R +=  + Ab(m56,mu2)*Bb(me2,mu2) * ( 1./81.*m11*m55*
         RSe461p*RSe463p + 1./81.*m11*m55*RSe461p*RSe453p + 1./162.*m11
         *m55*pow(RSe461p,2) + 1./81.*m11*m55*RSe451p*RSe463p + 1./81.*
         m11*m55*RSe451p*RSe461p );

      M13p6x23R +=  + Ab(m56,mu2)*Bb(mp2,mu2) * ( 1./162.*pow(mp2,2)*
         pow(RSp463e,2) - 1./81.*pow(mp2,2)*RSp453e*RSp463e + 1./162.*
         pow(mp2,2)*pow(RSp453e,2) - 1./81.*pow(mp2,2)*RSp461e*RSp463e
          + 1./81.*pow(mp2,2)*RSp461e*RSp453e + 1./162.*pow(mp2,2)*pow(
         RSp461e,2) + 1./81.*pow(mp2,2)*RSp451e*RSp463e - 1./81.*pow(
         mp2,2)*RSp451e*RSp453e - 1./81.*pow(mp2,2)*RSp451e*RSp461e + 1.
         /162.*pow(mp2,2)*pow(RSp451e,2) - 1./81.*m66*mp2*RSp453e*
         RSp463e - 1./162.*m66*mp2*pow(RSp453e,2) + 1./81.*m66*mp2*
         RSp461e*RSp453e + 1./81.*m66*mp2*RSp451e*RSp463e + 1./81.*m66*
         mp2*RSp451e*RSp453e - 1./81.*m66*mp2*RSp451e*RSp461e - 1./162.
         *m66*mp2*pow(RSp451e,2) - 1./162.*m55*mp2*pow(RSp463e,2) - 1./
         81.*m55*mp2*RSp453e*RSp463e + 1./81.*m55*mp2*RSp461e*RSp463e
          + 1./81.*m55*mp2*RSp461e*RSp453e - 1./162.*m55*mp2*pow(
         RSp461e,2) + 1./81.*m55*mp2*RSp451e*RSp463e - 1./81.*m55*mp2*
         RSp451e*RSp461e - 1./162.*m33*mp2*pow(RSp463e,2) + 1./81.*m33*
         mp2*RSp453e*RSp463e );

      M13p6x23R +=  + Ab(m56,mu2)*Bb(mp2,mu2) * (  - 1./162.*m33*mp2*
         pow(RSp453e,2) - 1./81.*m33*mp2*RSp461e*RSp463e + 1./81.*m33*
         mp2*RSp461e*RSp453e + 1./81.*m33*mp2*RSp451e*RSp463e - 1./81.*
         m33*mp2*RSp451e*RSp453e + 1./81.*m33*m66*RSp453e*RSp463e + 1./
         162.*m33*m66*pow(RSp453e,2) + 1./81.*m33*m66*RSp461e*RSp453e
          + 1./81.*m33*m66*RSp451e*RSp463e + 1./81.*m33*m66*RSp451e*
         RSp453e + 1./162.*m33*m55*pow(RSp463e,2) + 1./81.*m33*m55*
         RSp453e*RSp463e + 1./81.*m33*m55*RSp461e*RSp463e + 1./81.*m33*
         m55*RSp461e*RSp453e + 1./81.*m33*m55*RSp451e*RSp463e - 1./81.*
         m11*mp2*RSp461e*RSp463e + 1./81.*m11*mp2*RSp461e*RSp453e - 1./
         162.*m11*mp2*pow(RSp461e,2) + 1./81.*m11*mp2*RSp451e*RSp463e
          - 1./81.*m11*mp2*RSp451e*RSp453e + 1./81.*m11*mp2*RSp451e*
         RSp461e - 1./162.*m11*mp2*pow(RSp451e,2) + 1./81.*m11*m66*
         RSp461e*RSp453e + 1./81.*m11*m66*RSp451e*RSp463e + 1./81.*m11*
         m66*RSp451e*RSp453e + 1./81.*m11*m66*RSp451e*RSp461e + 1./162.
         *m11*m66*pow(RSp451e,2) );

      M13p6x23R +=  + Ab(m56,mu2)*Bb(mp2,mu2) * ( 1./81.*m11*m55*
         RSp461e*RSp463e + 1./81.*m11*m55*RSp461e*RSp453e + 1./162.*m11
         *m55*pow(RSp461e,2) + 1./81.*m11*m55*RSp451e*RSp463e + 1./81.*
         m11*m55*RSp451e*RSp461e );

      M13p6x23R +=  + Ab(m66,mu2)*Ab(me2,mu2) * (  - 1./54.*m66*pow(
         RSe453p,2) - 1./27.*m66*RSp453e*RSe453p + 1./162.*m66*RS645pe*
         pow(RSe453p,2) + 1./81.*m66*RS645pe*RSp453e*RSe453p + 1./27.*
         m66*RSe451p*RSe453p + 1./27.*m66*RSe451p*RSp453e - 1./81.*m66*
         RSe451p*RS645pe*RSe453p - 1./81.*m66*RSe451p*RS645pe*RSp453e
          - 1./54.*m66*pow(RSe451p,2) + 1./162.*m66*pow(RSe451p,2)*
         RS645pe + 1./27.*m66*RSp451e*RSe453p - 1./81.*m66*RSp451e*
         RS645pe*RSe453p - 1./27.*m66*RSp451e*RSe451p + 1./81.*m66*
         RSp451e*RSe451p*RS645pe - 1./27.*m66*mp2*RNep*RSp453e*RSe453p
          + 1./81.*m66*mp2*RNep*RS645pe*RSp453e*RSe453p + 1./27.*m66*
         mp2*RNep*RSe451p*RSp453e - 1./81.*m66*mp2*RNep*RSe451p*RS645pe
         *RSp453e + 1./27.*m66*mp2*RNep*RSp451e*RSe453p - 1./81.*m66*
         mp2*RNep*RSp451e*RS645pe*RSe453p - 1./27.*m66*mp2*RNep*RSp451e
         *RSe451p + 1./81.*m66*mp2*RNep*RSp451e*RSe451p*RS645pe + 1./27.
         *m66*me2*RN3e*RSe453p*RS345pe - 1./81.*m66*me2*RN3e*RS645pe*
         RSe453p*RS345pe );

      M13p6x23R +=  + Ab(m66,mu2)*Ab(me2,mu2) * (  - 1./27.*m66*me2*
         RN3e*RSe451p*RS345pe + 1./81.*m66*me2*RN3e*RSe451p*RS645pe*
         RS345pe - 1./27.*m66*me2*RN1e*RS145pe*RSe453p + 1./81.*m66*me2
         *RN1e*RS645pe*RS145pe*RSe453p + 1./27.*m66*me2*RN1e*RSe451p*
         RS145pe - 1./81.*m66*me2*RN1e*RSe451p*RS645pe*RS145pe + 1./27.
         *m33*m66*RNep*RSp453e*RSe453p - 1./81.*m33*m66*RNep*RS645pe*
         RSp453e*RSe453p + 1./27.*m33*m66*RNep*RSe451p*RSp453e - 1./81.
         *m33*m66*RNep*RSe451p*RS645pe*RSp453e + 1./27.*m33*m66*RNep*
         RSp451e*RSe453p - 1./81.*m33*m66*RNep*RSp451e*RS645pe*RSe453p
          - 1./27.*m33*m66*RN3e*RSe453p*RS345pe + 1./81.*m33*m66*RN3e*
         RS645pe*RSe453p*RS345pe - 1./27.*m33*m66*RN3e*RSe451p*RS345pe
          + 1./81.*m33*m66*RN3e*RSe451p*RS645pe*RS345pe - 1./27.*m33*
         m66*RN1e*RS145pe*RSe453p + 1./81.*m33*m66*RN1e*RS645pe*RS145pe
         *RSe453p + 1./27.*m11*m66*RNep*RSe451p*RSp453e - 1./81.*m11*
         m66*RNep*RSe451p*RS645pe*RSp453e + 1./27.*m11*m66*RNep*RSp451e
         *RSe453p );

      M13p6x23R +=  + Ab(m66,mu2)*Ab(me2,mu2) * (  - 1./81.*m11*m66*
         RNep*RSp451e*RS645pe*RSe453p + 1./27.*m11*m66*RNep*RSp451e*
         RSe451p - 1./81.*m11*m66*RNep*RSp451e*RSe451p*RS645pe - 1./27.
         *m11*m66*RN3e*RSe451p*RS345pe + 1./81.*m11*m66*RN3e*RSe451p*
         RS645pe*RS345pe - 1./27.*m11*m66*RN1e*RS145pe*RSe453p + 1./81.
         *m11*m66*RN1e*RS645pe*RS145pe*RSe453p - 1./27.*m11*m66*RN1e*
         RSe451p*RS145pe + 1./81.*m11*m66*RN1e*RSe451p*RS645pe*RS145pe
          );

      M13p6x23R +=  + Ab(m66,mu2)*Ab(mp2,mu2) * (  - 1./54.*m66*pow(
         RSp453e,2) + 1./162.*m66*RS645pe*pow(RSp453e,2) + 1./27.*m66*
         RSp451e*RSp453e - 1./81.*m66*RSp451e*RS645pe*RSp453e - 1./54.*
         m66*pow(RSp451e,2) + 1./162.*m66*pow(RSp451e,2)*RS645pe + 1./
         27.*m66*mp2*RNep*RSp453e*RSe453p - 1./81.*m66*mp2*RNep*RS645pe
         *RSp453e*RSe453p - 1./27.*m66*mp2*RNep*RSe451p*RSp453e + 1./81.
         *m66*mp2*RNep*RSe451p*RS645pe*RSp453e - 1./27.*m66*mp2*RNep*
         RSp451e*RSe453p + 1./81.*m66*mp2*RNep*RSp451e*RS645pe*RSe453p
          + 1./27.*m66*mp2*RNep*RSp451e*RSe451p - 1./81.*m66*mp2*RNep*
         RSp451e*RSe451p*RS645pe + 1./27.*m66*mp2*RN3p*RSp453e*RS345pe
          - 1./81.*m66*mp2*RN3p*RS645pe*RSp453e*RS345pe - 1./27.*m66*
         mp2*RN3p*RSp451e*RS345pe + 1./81.*m66*mp2*RN3p*RSp451e*RS645pe
         *RS345pe - 1./27.*m66*mp2*RN1p*RS145pe*RSp453e + 1./81.*m66*
         mp2*RN1p*RS645pe*RS145pe*RSp453e + 1./27.*m66*mp2*RN1p*RSp451e
         *RS145pe - 1./81.*m66*mp2*RN1p*RSp451e*RS645pe*RS145pe - 1./27.
         *m33*m66*RNep*RSp453e*RSe453p );

      M13p6x23R +=  + Ab(m66,mu2)*Ab(mp2,mu2) * ( 1./81.*m33*m66*RNep*
         RS645pe*RSp453e*RSe453p - 1./27.*m33*m66*RNep*RSe451p*RSp453e
          + 1./81.*m33*m66*RNep*RSe451p*RS645pe*RSp453e - 1./27.*m33*
         m66*RNep*RSp451e*RSe453p + 1./81.*m33*m66*RNep*RSp451e*RS645pe
         *RSe453p - 1./27.*m33*m66*RN3p*RSp453e*RS345pe + 1./81.*m33*
         m66*RN3p*RS645pe*RSp453e*RS345pe - 1./27.*m33*m66*RN3p*RSp451e
         *RS345pe + 1./81.*m33*m66*RN3p*RSp451e*RS645pe*RS345pe - 1./27.
         *m33*m66*RN1p*RS145pe*RSp453e + 1./81.*m33*m66*RN1p*RS645pe*
         RS145pe*RSp453e - 1./27.*m11*m66*RNep*RSe451p*RSp453e + 1./81.
         *m11*m66*RNep*RSe451p*RS645pe*RSp453e - 1./27.*m11*m66*RNep*
         RSp451e*RSe453p + 1./81.*m11*m66*RNep*RSp451e*RS645pe*RSe453p
          - 1./27.*m11*m66*RNep*RSp451e*RSe451p + 1./81.*m11*m66*RNep*
         RSp451e*RSe451p*RS645pe - 1./27.*m11*m66*RN3p*RSp451e*RS345pe
          + 1./81.*m11*m66*RN3p*RSp451e*RS645pe*RS345pe - 1./27.*m11*
         m66*RN1p*RS145pe*RSp453e + 1./81.*m11*m66*RN1p*RS645pe*RS145pe
         *RSp453e );

      M13p6x23R +=  + Ab(m66,mu2)*Ab(mp2,mu2) * (  - 1./27.*m11*m66*
         RN1p*RSp451e*RS145pe + 1./81.*m11*m66*RN1p*RSp451e*RS645pe*
         RS145pe );

      M13p6x23R +=  + Ab(m66,mu2)*Bb(me2,mu2) * (  - 1./54.*m66*me2*
         pow(RSe453p,2) + 1./162.*m66*me2*RS645pe*pow(RSe453p,2) + 1./
         27.*m66*me2*RSe451p*RSe453p - 1./81.*m66*me2*RSe451p*RS645pe*
         RSe453p - 1./54.*m66*me2*pow(RSe451p,2) + 1./162.*m66*me2*pow(
         RSe451p,2)*RS645pe + 1./54.*m33*m66*pow(RSe453p,2) - 1./162.*
         m33*m66*RS645pe*pow(RSe453p,2) + 1./27.*m33*m66*RSe451p*
         RSe453p - 1./81.*m33*m66*RSe451p*RS645pe*RSe453p + 1./27.*m11*
         m66*RSe451p*RSe453p - 1./81.*m11*m66*RSe451p*RS645pe*RSe453p
          + 1./54.*m11*m66*pow(RSe451p,2) - 1./162.*m11*m66*pow(
         RSe451p,2)*RS645pe );

      M13p6x23R +=  + Ab(m66,mu2)*Bb(mp2,mu2) * (  - 1./54.*m66*mp2*
         pow(RSp453e,2) + 1./162.*m66*mp2*RS645pe*pow(RSp453e,2) + 1./
         27.*m66*mp2*RSp451e*RSp453e - 1./81.*m66*mp2*RSp451e*RS645pe*
         RSp453e - 1./54.*m66*mp2*pow(RSp451e,2) + 1./162.*m66*mp2*pow(
         RSp451e,2)*RS645pe + 1./54.*m33*m66*pow(RSp453e,2) - 1./162.*
         m33*m66*RS645pe*pow(RSp453e,2) + 1./27.*m33*m66*RSp451e*
         RSp453e - 1./81.*m33*m66*RSp451e*RS645pe*RSp453e + 1./27.*m11*
         m66*RSp451e*RSp453e - 1./81.*m11*m66*RSp451e*RS645pe*RSp453e
          + 1./54.*m11*m66*pow(RSp451e,2) - 1./162.*m11*m66*pow(
         RSp451e,2)*RS645pe );

      M13p6x23R +=  + Ab(me2,mu2)*pi16 * (  - 1./72.*m33*RD3456ep*
         RSe45613p + 1./144.*m33*RSe45633p*RD3456ep - 1./72.*m33*
         RD1456ep*RSe45613p + 1./144.*m33*RD1456ep*RSe45633p + 1./144.*
         m33*RSe45611p*RD3456ep + 1./144.*m33*RSe45611p*RD1456ep + 1./
         36.*m33*mp2*RSp45613e*RSe45613p - 1./72.*m33*mp2*RSe45633p*
         RSp45613e - 1./72.*m33*mp2*RSp45633e*RSe45613p + 1./144.*m33*
         mp2*RSp45633e*RSe45633p - 1./72.*m33*mp2*RSe45611p*RSp45613e
          + 1./144.*m33*mp2*RSe45611p*RSp45633e - 1./72.*m33*mp2*
         RSp45611e*RSe45613p + 1./144.*m33*mp2*RSp45611e*RSe45633p + 1./
         144.*m33*mp2*RSp45611e*RSe45611p - 1./24.*m33*me2*RSe45613p + 
         1./36.*m33*me2*pow(RSe45613p,2) + 1./48.*m33*me2*RSe45633p - 1.
         /36.*m33*me2*RSe45633p*RSe45613p + 1./144.*m33*me2*pow(
         RSe45633p,2) + 1./48.*m33*me2*RSe563p + 1./48.*m33*me2*RSe463p
          + 1./48.*m33*me2*RSe453p + 1./48.*m33*me2*RSe45611p - 1./36.*
         m33*me2*RSe45611p*RSe45613p + 1./72.*m33*me2*RSe45611p*
         RSe45633p );

      M13p6x23R +=  + Ab(me2,mu2)*pi16 * ( 1./144.*m33*me2*pow(
         RSe45611p,2) + 1./48.*m33*me2*RSe561p + 1./48.*m33*me2*RSe461p
          + 1./48.*m33*me2*RSe451p + 1./24.*m33*m66*RSe45613p - 1./48.*
         m33*m66*RSe45633p - 1./48.*m33*m66*RSe45611p + 1./24.*m33*m55*
         RSe45613p - 1./48.*m33*m55*RSe45633p - 1./48.*m33*m55*
         RSe45611p + 1./24.*m33*m44*RSe45613p - 1./48.*m33*m44*
         RSe45633p - 1./48.*m33*m44*RSe45611p + 1./24.*pow(m33,2)*
         RSe45613p + 1./27.*pow(m33,2)*RSe45613p*RS34561pe + 1./108.*
         pow(m33,2)*RSe45613p*RS14563pe + 1./108.*pow(m33,2)*pow(
         RSe45613p,2) + 1./108.*pow(m33,2)*RSp45613e*RSe45613p + 1./54.
         *pow(m33,2)*RS3456pee*RSe45613p + 1./54.*pow(m33,2)*RS3456ppe*
         RSe45613p - 1./48.*pow(m33,2)*RSe45633p - 1./54.*pow(m33,2)*
         RSe45633p*RS34561pe - 1./216.*pow(m33,2)*RSe45633p*RS14563pe
          - 1./108.*pow(m33,2)*RSe45633p*RSe45613p - 1./216.*pow(m33,2)
         *RSe45633p*RSp45613e - 1./108.*pow(m33,2)*RSe45633p*RS3456pee
          - 1./108.*pow(m33,2)*RSe45633p*RS3456ppe );

      M13p6x23R +=  + Ab(me2,mu2)*pi16 * ( 1./432.*pow(m33,2)*pow(
         RSe45633p,2) - 1./216.*pow(m33,2)*RSp45633e*RSe45613p + 1./432.
         *pow(m33,2)*RSp45633e*RSe45633p - 1./54.*pow(m33,2)*RS345pe*
         RSe45613p + 1./108.*pow(m33,2)*RS345pe*RSe45633p - 1./54.*pow(
         m33,2)*RS346pe*RSe45613p + 1./108.*pow(m33,2)*RS346pe*
         RSe45633p - 1./54.*pow(m33,2)*RS356pe*RSe45613p + 1./108.*pow(
         m33,2)*RS356pe*RSe45633p + 1./216.*pow(m33,2)*RS1456pee*
         RSe45613p - 1./432.*pow(m33,2)*RS1456pee*RSe45633p + 1./216.*
         pow(m33,2)*RS1456ppe*RSe45613p - 1./432.*pow(m33,2)*RS1456ppe*
         RSe45633p - 1./48.*pow(m33,2)*RSe45611p - 1./54.*pow(m33,2)*
         RSe45611p*RS34561pe - 1./216.*pow(m33,2)*RSe45611p*RS14563pe
          - 1./108.*pow(m33,2)*RSe45611p*RSe45613p - 1./216.*pow(m33,2)
         *RSe45611p*RSp45613e - 1./108.*pow(m33,2)*RSe45611p*RS3456pee
          - 1./108.*pow(m33,2)*RSe45611p*RS3456ppe + 1./216.*pow(m33,2)
         *RSe45611p*RSe45633p + 1./432.*pow(m33,2)*RSe45611p*RSp45633e
          + 1./108.*pow(m33,2)*RSe45611p*RS345pe );

      M13p6x23R +=  + Ab(me2,mu2)*pi16 * ( 1./108.*pow(m33,2)*RSe45611p
         *RS346pe + 1./108.*pow(m33,2)*RSe45611p*RS356pe - 1./432.*pow(
         m33,2)*RSe45611p*RS1456pee - 1./432.*pow(m33,2)*RSe45611p*
         RS1456ppe + 1./432.*pow(m33,2)*pow(RSe45611p,2) - 1./216.*pow(
         m33,2)*RSp45611e*RSe45613p + 1./432.*pow(m33,2)*RSp45611e*
         RSe45633p + 1./432.*pow(m33,2)*RSp45611e*RSe45611p - 1./216.*
         pow(m33,2)*RS145pe*RSe45613p + 1./432.*pow(m33,2)*RS145pe*
         RSe45633p + 1./432.*pow(m33,2)*RS145pe*RSe45611p - 1./216.*
         pow(m33,2)*RS146pe*RSe45613p + 1./432.*pow(m33,2)*RS146pe*
         RSe45633p + 1./432.*pow(m33,2)*RS146pe*RSe45611p - 1./216.*
         pow(m33,2)*RS156pe*RSe45613p + 1./432.*pow(m33,2)*RS156pe*
         RSe45633p + 1./432.*pow(m33,2)*RS156pe*RSe45611p - 1./72.*m11*
         RD3456ep*RSe45613p + 1./144.*m11*RSe45633p*RD3456ep - 1./72.*
         m11*RD1456ep*RSe45613p + 1./144.*m11*RD1456ep*RSe45633p + 1./
         144.*m11*RSe45611p*RD3456ep + 1./144.*m11*RSe45611p*RD1456ep
          + 1./36.*m11*mp2*RSp45613e*RSe45613p );

      M13p6x23R +=  + Ab(me2,mu2)*pi16 * (  - 1./72.*m11*mp2*RSe45633p*
         RSp45613e - 1./72.*m11*mp2*RSp45633e*RSe45613p + 1./144.*m11*
         mp2*RSp45633e*RSe45633p - 1./72.*m11*mp2*RSe45611p*RSp45613e
          + 1./144.*m11*mp2*RSe45611p*RSp45633e - 1./72.*m11*mp2*
         RSp45611e*RSe45613p + 1./144.*m11*mp2*RSp45611e*RSe45633p + 1./
         144.*m11*mp2*RSp45611e*RSe45611p - 1./24.*m11*me2*RSe45613p + 
         1./36.*m11*me2*pow(RSe45613p,2) + 1./48.*m11*me2*RSe45633p - 1.
         /36.*m11*me2*RSe45633p*RSe45613p + 1./144.*m11*me2*pow(
         RSe45633p,2) + 1./48.*m11*me2*RSe563p + 1./48.*m11*me2*RSe463p
          + 1./48.*m11*me2*RSe453p + 1./48.*m11*me2*RSe45611p - 1./36.*
         m11*me2*RSe45611p*RSe45613p + 1./72.*m11*me2*RSe45611p*
         RSe45633p + 1./144.*m11*me2*pow(RSe45611p,2) + 1./48.*m11*me2*
         RSe561p + 1./48.*m11*me2*RSe461p + 1./48.*m11*me2*RSe451p + 1./
         24.*m11*m66*RSe45613p - 1./48.*m11*m66*RSe45633p - 1./48.*m11*
         m66*RSe45611p + 1./24.*m11*m55*RSe45613p - 1./48.*m11*m55*
         RSe45633p );

      M13p6x23R +=  + Ab(me2,mu2)*pi16 * (  - 1./48.*m11*m55*RSe45611p
          + 1./24.*m11*m44*RSe45613p - 1./48.*m11*m44*RSe45633p - 1./48.
         *m11*m44*RSe45611p + 1./12.*m11*m33*RSe45613p + 5./108.*m11*
         m33*RSe45613p*RS34561pe + 5./108.*m11*m33*RSe45613p*RS14563pe
          + 1./54.*m11*m33*pow(RSe45613p,2) + 1./54.*m11*m33*RSp45613e*
         RSe45613p + 5./216.*m11*m33*RS3456pee*RSe45613p + 5./216.*m11*
         m33*RS3456ppe*RSe45613p - 1./24.*m11*m33*RSe45633p - 5./216.*
         m11*m33*RSe45633p*RS34561pe - 5./216.*m11*m33*RSe45633p*
         RS14563pe - 1./54.*m11*m33*RSe45633p*RSe45613p - 1./108.*m11*
         m33*RSe45633p*RSp45613e - 5./432.*m11*m33*RSe45633p*RS3456pee
          - 5./432.*m11*m33*RSe45633p*RS3456ppe + 1./216.*m11*m33*pow(
         RSe45633p,2) - 1./108.*m11*m33*RSp45633e*RSe45613p + 1./216.*
         m11*m33*RSp45633e*RSe45633p - 5./216.*m11*m33*RS345pe*
         RSe45613p + 5./432.*m11*m33*RS345pe*RSe45633p - 5./216.*m11*
         m33*RS346pe*RSe45613p + 5./432.*m11*m33*RS346pe*RSe45633p - 5./
         216.*m11*m33*RS356pe*RSe45613p );

      M13p6x23R +=  + Ab(me2,mu2)*pi16 * ( 5./432.*m11*m33*RS356pe*
         RSe45633p + 5./216.*m11*m33*RS1456pee*RSe45613p - 5./432.*m11*
         m33*RS1456pee*RSe45633p + 5./216.*m11*m33*RS1456ppe*RSe45613p
          - 5./432.*m11*m33*RS1456ppe*RSe45633p - 1./24.*m11*m33*
         RSe45611p - 5./216.*m11*m33*RSe45611p*RS34561pe - 5./216.*m11*
         m33*RSe45611p*RS14563pe - 1./54.*m11*m33*RSe45611p*RSe45613p
          - 1./108.*m11*m33*RSe45611p*RSp45613e - 5./432.*m11*m33*
         RSe45611p*RS3456pee - 5./432.*m11*m33*RSe45611p*RS3456ppe + 1./
         108.*m11*m33*RSe45611p*RSe45633p + 1./216.*m11*m33*RSe45611p*
         RSp45633e + 5./432.*m11*m33*RSe45611p*RS345pe + 5./432.*m11*
         m33*RSe45611p*RS346pe + 5./432.*m11*m33*RSe45611p*RS356pe - 5./
         432.*m11*m33*RSe45611p*RS1456pee - 5./432.*m11*m33*RSe45611p*
         RS1456ppe + 1./216.*m11*m33*pow(RSe45611p,2) - 1./108.*m11*m33
         *RSp45611e*RSe45613p + 1./216.*m11*m33*RSp45611e*RSe45633p + 1.
         /216.*m11*m33*RSp45611e*RSe45611p - 5./216.*m11*m33*RS145pe*
         RSe45613p );

      M13p6x23R +=  + Ab(me2,mu2)*pi16 * ( 5./432.*m11*m33*RS145pe*
         RSe45633p + 5./432.*m11*m33*RS145pe*RSe45611p - 5./216.*m11*
         m33*RS146pe*RSe45613p + 5./432.*m11*m33*RS146pe*RSe45633p + 5./
         432.*m11*m33*RS146pe*RSe45611p - 5./216.*m11*m33*RS156pe*
         RSe45613p + 5./432.*m11*m33*RS156pe*RSe45633p + 5./432.*m11*
         m33*RS156pe*RSe45611p + 1./24.*pow(m11,2)*RSe45613p + 1./108.*
         pow(m11,2)*RSe45613p*RS34561pe + 1./27.*pow(m11,2)*RSe45613p*
         RS14563pe + 1./108.*pow(m11,2)*pow(RSe45613p,2) + 1./108.*pow(
         m11,2)*RSp45613e*RSe45613p + 1./216.*pow(m11,2)*RS3456pee*
         RSe45613p + 1./216.*pow(m11,2)*RS3456ppe*RSe45613p - 1./48.*
         pow(m11,2)*RSe45633p - 1./216.*pow(m11,2)*RSe45633p*RS34561pe
          - 1./54.*pow(m11,2)*RSe45633p*RS14563pe - 1./108.*pow(m11,2)*
         RSe45633p*RSe45613p - 1./216.*pow(m11,2)*RSe45633p*RSp45613e
          - 1./432.*pow(m11,2)*RSe45633p*RS3456pee - 1./432.*pow(m11,2)
         *RSe45633p*RS3456ppe + 1./432.*pow(m11,2)*pow(RSe45633p,2) - 1.
         /216.*pow(m11,2)*RSp45633e*RSe45613p );

      M13p6x23R +=  + Ab(me2,mu2)*pi16 * ( 1./432.*pow(m11,2)*RSp45633e
         *RSe45633p - 1./216.*pow(m11,2)*RS345pe*RSe45613p + 1./432.*
         pow(m11,2)*RS345pe*RSe45633p - 1./216.*pow(m11,2)*RS346pe*
         RSe45613p + 1./432.*pow(m11,2)*RS346pe*RSe45633p - 1./216.*
         pow(m11,2)*RS356pe*RSe45613p + 1./432.*pow(m11,2)*RS356pe*
         RSe45633p + 1./54.*pow(m11,2)*RS1456pee*RSe45613p - 1./108.*
         pow(m11,2)*RS1456pee*RSe45633p + 1./54.*pow(m11,2)*RS1456ppe*
         RSe45613p - 1./108.*pow(m11,2)*RS1456ppe*RSe45633p - 1./48.*
         pow(m11,2)*RSe45611p - 1./216.*pow(m11,2)*RSe45611p*RS34561pe
          - 1./54.*pow(m11,2)*RSe45611p*RS14563pe - 1./108.*pow(m11,2)*
         RSe45611p*RSe45613p - 1./216.*pow(m11,2)*RSe45611p*RSp45613e
          - 1./432.*pow(m11,2)*RSe45611p*RS3456pee - 1./432.*pow(m11,2)
         *RSe45611p*RS3456ppe + 1./216.*pow(m11,2)*RSe45611p*RSe45633p
          + 1./432.*pow(m11,2)*RSe45611p*RSp45633e + 1./432.*pow(m11,2)
         *RSe45611p*RS345pe + 1./432.*pow(m11,2)*RSe45611p*RS346pe + 1./
         432.*pow(m11,2)*RSe45611p*RS356pe );

      M13p6x23R +=  + Ab(me2,mu2)*pi16 * (  - 1./108.*pow(m11,2)*
         RSe45611p*RS1456pee - 1./108.*pow(m11,2)*RSe45611p*RS1456ppe
          + 1./432.*pow(m11,2)*pow(RSe45611p,2) - 1./216.*pow(m11,2)*
         RSp45611e*RSe45613p + 1./432.*pow(m11,2)*RSp45611e*RSe45633p
          + 1./432.*pow(m11,2)*RSp45611e*RSe45611p - 1./54.*pow(m11,2)*
         RS145pe*RSe45613p + 1./108.*pow(m11,2)*RS145pe*RSe45633p + 1./
         108.*pow(m11,2)*RS145pe*RSe45611p - 1./54.*pow(m11,2)*RS146pe*
         RSe45613p + 1./108.*pow(m11,2)*RS146pe*RSe45633p + 1./108.*
         pow(m11,2)*RS146pe*RSe45611p - 1./54.*pow(m11,2)*RS156pe*
         RSe45613p + 1./108.*pow(m11,2)*RS156pe*RSe45633p + 1./108.*
         pow(m11,2)*RS156pe*RSe45611p );

      M13p6x23R +=  + pow(Ab(me2,mu2),2) * ( 1./135.*me2*pow(
         RSe45613p,2) - 1./135.*me2*RSe45633p*RSe45613p + 1./540.*me2*
         pow(RSe45633p,2) - 1./135.*me2*RSe45611p*RSe45613p + 1./270.*
         me2*RSe45611p*RSe45633p + 1./540.*me2*pow(RSe45611p,2) + 1./
         162.*m66*RSe456p*pow(RSe453p,2) + 1./81.*m66*RSe456p*RSp453e*
         RSe453p - 1./81.*m66*RSe456p*RSe451p*RSe453p - 1./81.*m66*
         RSe456p*RSe451p*RSp453e + 1./162.*m66*RSe456p*pow(RSe451p,2)
          - 1./81.*m66*RSe456p*RSp451e*RSe453p + 1./81.*m66*RSe456p*
         RSp451e*RSe451p + 1./81.*m66*mp2*RNep*RSe456p*RSp453e*RSe453p
          - 1./81.*m66*mp2*RNep*RSe456p*RSe451p*RSp453e - 1./81.*m66*
         mp2*RNep*RSe456p*RSp451e*RSe453p + 1./81.*m66*mp2*RNep*RSe456p
         *RSp451e*RSe451p - 1./81.*m66*me2*RN3e*RSe456p*RSe453p*RS345pe
          + 1./81.*m66*me2*RN3e*RSe456p*RSe451p*RS345pe + 1./81.*m66*
         me2*RN1e*RSe456p*RS145pe*RSe453p - 1./81.*m66*me2*RN1e*RSe456p
         *RSe451p*RS145pe + 1./162.*m55*RSe465p*pow(RSe463p,2) + 1./81.
         *m55*RSe465p*RSp463e*RSe463p );

      M13p6x23R +=  + pow(Ab(me2,mu2),2) * (  - 1./81.*m55*RSe465p*
         RSe461p*RSe463p - 1./81.*m55*RSe465p*RSe461p*RSp463e + 1./162.
         *m55*RSe465p*pow(RSe461p,2) - 1./81.*m55*RSe465p*RSp461e*
         RSe463p + 1./81.*m55*RSe465p*RSp461e*RSe461p + 1./81.*m55*mp2*
         RNep*RSe465p*RSp463e*RSe463p - 1./81.*m55*mp2*RNep*RSe465p*
         RSe461p*RSp463e - 1./81.*m55*mp2*RNep*RSe465p*RSp461e*RSe463p
          + 1./81.*m55*mp2*RNep*RSe465p*RSp461e*RSe461p - 1./81.*m55*
         me2*RN3e*RSe465p*RSe463p*RS346pe + 1./81.*m55*me2*RN3e*RSe465p
         *RSe461p*RS346pe + 1./81.*m55*me2*RN1e*RSe465p*RS146pe*RSe463p
          - 1./81.*m55*me2*RN1e*RSe465p*RSe461p*RS146pe + 1./162.*m44*
         RSe564p*pow(RSe563p,2) + 1./81.*m44*RSe564p*RSp563e*RSe563p - 
         1./81.*m44*RSe564p*RSe561p*RSe563p - 1./81.*m44*RSe564p*
         RSe561p*RSp563e + 1./162.*m44*RSe564p*pow(RSe561p,2) - 1./81.*
         m44*RSe564p*RSp561e*RSe563p + 1./81.*m44*RSe564p*RSp561e*
         RSe561p + 1./81.*m44*mp2*RNep*RSe564p*RSp563e*RSe563p - 1./81.
         *m44*mp2*RNep*RSe564p*RSe561p*RSp563e );

      M13p6x23R +=  + pow(Ab(me2,mu2),2) * (  - 1./81.*m44*mp2*RNep*
         RSe564p*RSp561e*RSe563p + 1./81.*m44*mp2*RNep*RSe564p*RSp561e*
         RSe561p - 1./81.*m44*me2*RN3e*RSe564p*RSe563p*RS356pe + 1./81.
         *m44*me2*RN3e*RSe564p*RSe561p*RS356pe + 1./81.*m44*me2*RN1e*
         RSe564p*RS156pe*RSe563p - 1./81.*m44*me2*RN1e*RSe564p*RSe561p*
         RS156pe + 13./540.*m33*pow(RSe45613p,2) - 7./540.*m33*
         RSe45633p*RSe45613p + 17./720.*m33*pow(RSe45633p,2) + 1./540.*
         m33*RSe45611p*RSe45613p - 17./1080.*m33*RSe45611p*RSe45633p + 
         1./144.*m33*pow(RSe45611p,2) - 1./27.*m33*me2*RN3e*RSe45633p*
         RSe45613p + 1./27.*m33*me2*RN3e*pow(RSe45633p,2) - 1./81.*m33*
         m66*RNep*RSe456p*RSp453e*RSe453p - 1./81.*m33*m66*RNep*RSe456p
         *RSe451p*RSp453e - 1./81.*m33*m66*RNep*RSe456p*RSp451e*RSe453p
          + 1./81.*m33*m66*RN3e*RSe456p*RSe453p*RS345pe + 1./81.*m33*
         m66*RN3e*RSe456p*RSe451p*RS345pe + 1./81.*m33*m66*RN1e*RSe456p
         *RS145pe*RSe453p - 1./81.*m33*m55*RNep*RSe465p*RSp463e*RSe463p
          - 1./81.*m33*m55*RNep*RSe465p*RSe461p*RSp463e );

      M13p6x23R +=  + pow(Ab(me2,mu2),2) * (  - 1./81.*m33*m55*RNep*
         RSe465p*RSp461e*RSe463p + 1./81.*m33*m55*RN3e*RSe465p*RSe463p*
         RS346pe + 1./81.*m33*m55*RN3e*RSe465p*RSe461p*RS346pe + 1./81.
         *m33*m55*RN1e*RSe465p*RS146pe*RSe463p - 1./81.*m33*m44*RNep*
         RSe564p*RSp563e*RSe563p - 1./81.*m33*m44*RNep*RSe564p*RSe561p*
         RSp563e - 1./81.*m33*m44*RNep*RSe564p*RSp561e*RSe563p + 1./81.
         *m33*m44*RN3e*RSe564p*RSe563p*RS356pe + 1./81.*m33*m44*RN3e*
         RSe564p*RSe561p*RS356pe + 1./81.*m33*m44*RN1e*RSe564p*RS156pe*
         RSe563p - 1./27.*pow(m33,2)*RN3e*RSe45633p*RSe45613p - 1./27.*
         pow(m33,2)*RN3e*pow(RSe45633p,2) + 13./540.*m11*pow(
         RSe45613p,2) + 1./540.*m11*RSe45633p*RSe45613p + 1./144.*m11*
         pow(RSe45633p,2) - 7./540.*m11*RSe45611p*RSe45613p - 17./1080.
         *m11*RSe45611p*RSe45633p + 17./720.*m11*pow(RSe45611p,2) - 1./
         27.*m11*me2*RN1e*RSe45611p*RSe45613p + 1./27.*m11*me2*RN1e*
         pow(RSe45611p,2) - 1./81.*m11*m66*RNep*RSe456p*RSe451p*RSp453e
          - 1./81.*m11*m66*RNep*RSe456p*RSp451e*RSe453p );

      M13p6x23R +=  + pow(Ab(me2,mu2),2) * (  - 1./81.*m11*m66*RNep*
         RSe456p*RSp451e*RSe451p + 1./81.*m11*m66*RN3e*RSe456p*RSe451p*
         RS345pe + 1./81.*m11*m66*RN1e*RSe456p*RS145pe*RSe453p + 1./81.
         *m11*m66*RN1e*RSe456p*RSe451p*RS145pe - 1./81.*m11*m55*RNep*
         RSe465p*RSe461p*RSp463e - 1./81.*m11*m55*RNep*RSe465p*RSp461e*
         RSe463p - 1./81.*m11*m55*RNep*RSe465p*RSp461e*RSe461p + 1./81.
         *m11*m55*RN3e*RSe465p*RSe461p*RS346pe + 1./81.*m11*m55*RN1e*
         RSe465p*RS146pe*RSe463p + 1./81.*m11*m55*RN1e*RSe465p*RSe461p*
         RS146pe - 1./81.*m11*m44*RNep*RSe564p*RSe561p*RSp563e - 1./81.
         *m11*m44*RNep*RSe564p*RSp561e*RSe563p - 1./81.*m11*m44*RNep*
         RSe564p*RSp561e*RSe561p + 1./81.*m11*m44*RN3e*RSe564p*RSe561p*
         RS356pe + 1./81.*m11*m44*RN1e*RSe564p*RS156pe*RSe563p + 1./81.
         *m11*m44*RN1e*RSe564p*RSe561p*RS156pe - 1./27.*m11*m33*RN3e*
         RSe45633p*RSe45613p - 1./27.*m11*m33*RN1e*RSe45611p*RSe45613p
          - 1./27.*pow(m11,2)*RN1e*RSe45611p*RSe45613p - 1./27.*pow(
         m11,2)*RN1e*pow(RSe45611p,2) );

      M13p6x23R +=  + Ab(me2,mu2)*Ab(mp2,mu2) * ( 1./135.*mp2*RSp45613e
         *RSe45613p - 1./270.*mp2*RSe45633p*RSp45613e - 1./270.*mp2*
         RSp45633e*RSe45613p + 1./540.*mp2*RSp45633e*RSe45633p - 1./270.
         *mp2*RSe45611p*RSp45613e + 1./540.*mp2*RSe45611p*RSp45633e - 1.
         /270.*mp2*RSp45611e*RSe45613p + 1./540.*mp2*RSp45611e*
         RSe45633p + 1./540.*mp2*RSp45611e*RSe45611p + 1./135.*me2*
         RSp45613e*RSe45613p - 1./270.*me2*RSe45633p*RSp45613e - 1./270.
         *me2*RSp45633e*RSe45613p + 1./540.*me2*RSp45633e*RSe45633p - 1.
         /270.*me2*RSe45611p*RSp45613e + 1./540.*me2*RSe45611p*
         RSp45633e - 1./270.*me2*RSp45611e*RSe45613p + 1./540.*me2*
         RSp45611e*RSe45633p + 1./540.*me2*RSp45611e*RSe45611p + 1./162.
         *m66*RSe456p*pow(RSp453e,2) - 1./81.*m66*RSe456p*RSp451e*
         RSp453e + 1./162.*m66*RSe456p*pow(RSp451e,2) + 1./162.*m66*
         RSp456e*pow(RSe453p,2) + 1./81.*m66*RSp456e*RSp453e*RSe453p - 
         1./81.*m66*RSp456e*RSe451p*RSe453p - 1./81.*m66*RSp456e*
         RSe451p*RSp453e );

      M13p6x23R +=  + Ab(me2,mu2)*Ab(mp2,mu2) * ( 1./162.*m66*RSp456e*
         pow(RSe451p,2) - 1./81.*m66*RSp456e*RSp451e*RSe453p + 1./81.*
         m66*RSp456e*RSp451e*RSe451p - 1./81.*m66*mp2*RNep*RSe456p*
         RSp453e*RSe453p + 1./81.*m66*mp2*RNep*RSe456p*RSe451p*RSp453e
          + 1./81.*m66*mp2*RNep*RSe456p*RSp451e*RSe453p - 1./81.*m66*
         mp2*RNep*RSe456p*RSp451e*RSe451p + 1./81.*m66*mp2*RNep*RSp456e
         *RSp453e*RSe453p - 1./81.*m66*mp2*RNep*RSp456e*RSe451p*RSp453e
          - 1./81.*m66*mp2*RNep*RSp456e*RSp451e*RSe453p + 1./81.*m66*
         mp2*RNep*RSp456e*RSp451e*RSe451p - 1./81.*m66*mp2*RN3p*RSe456p
         *RSp453e*RS345pe + 1./81.*m66*mp2*RN3p*RSe456p*RSp451e*RS345pe
          + 1./81.*m66*mp2*RN1p*RSe456p*RS145pe*RSp453e - 1./81.*m66*
         mp2*RN1p*RSe456p*RSp451e*RS145pe - 1./81.*m66*me2*RN3e*RSp456e
         *RSe453p*RS345pe + 1./81.*m66*me2*RN3e*RSp456e*RSe451p*RS345pe
          + 1./81.*m66*me2*RN1e*RSp456e*RS145pe*RSe453p - 1./81.*m66*
         me2*RN1e*RSp456e*RSe451p*RS145pe + 1./162.*m55*RSe465p*pow(
         RSp463e,2) );

      M13p6x23R +=  + Ab(me2,mu2)*Ab(mp2,mu2) * (  - 1./81.*m55*RSe465p
         *RSp461e*RSp463e + 1./162.*m55*RSe465p*pow(RSp461e,2) + 1./162.
         *m55*RSp465e*pow(RSe463p,2) + 1./81.*m55*RSp465e*RSp463e*
         RSe463p - 1./81.*m55*RSp465e*RSe461p*RSe463p - 1./81.*m55*
         RSp465e*RSe461p*RSp463e + 1./162.*m55*RSp465e*pow(RSe461p,2)
          - 1./81.*m55*RSp465e*RSp461e*RSe463p + 1./81.*m55*RSp465e*
         RSp461e*RSe461p - 1./81.*m55*mp2*RNep*RSe465p*RSp463e*RSe463p
          + 1./81.*m55*mp2*RNep*RSe465p*RSe461p*RSp463e + 1./81.*m55*
         mp2*RNep*RSe465p*RSp461e*RSe463p - 1./81.*m55*mp2*RNep*RSe465p
         *RSp461e*RSe461p + 1./81.*m55*mp2*RNep*RSp465e*RSp463e*RSe463p
          - 1./81.*m55*mp2*RNep*RSp465e*RSe461p*RSp463e - 1./81.*m55*
         mp2*RNep*RSp465e*RSp461e*RSe463p + 1./81.*m55*mp2*RNep*RSp465e
         *RSp461e*RSe461p - 1./81.*m55*mp2*RN3p*RSe465p*RSp463e*RS346pe
          + 1./81.*m55*mp2*RN3p*RSe465p*RSp461e*RS346pe + 1./81.*m55*
         mp2*RN1p*RSe465p*RS146pe*RSp463e - 1./81.*m55*mp2*RN1p*RSe465p
         *RSp461e*RS146pe );

      M13p6x23R +=  + Ab(me2,mu2)*Ab(mp2,mu2) * (  - 1./81.*m55*me2*
         RN3e*RSp465e*RSe463p*RS346pe + 1./81.*m55*me2*RN3e*RSp465e*
         RSe461p*RS346pe + 1./81.*m55*me2*RN1e*RSp465e*RS146pe*RSe463p
          - 1./81.*m55*me2*RN1e*RSp465e*RSe461p*RS146pe + 1./162.*m44*
         RSe564p*pow(RSp563e,2) - 1./81.*m44*RSe564p*RSp561e*RSp563e + 
         1./162.*m44*RSe564p*pow(RSp561e,2) + 1./162.*m44*RSp564e*pow(
         RSe563p,2) + 1./81.*m44*RSp564e*RSp563e*RSe563p - 1./81.*m44*
         RSp564e*RSe561p*RSe563p - 1./81.*m44*RSp564e*RSe561p*RSp563e
          + 1./162.*m44*RSp564e*pow(RSe561p,2) - 1./81.*m44*RSp564e*
         RSp561e*RSe563p + 1./81.*m44*RSp564e*RSp561e*RSe561p - 1./81.*
         m44*mp2*RNep*RSe564p*RSp563e*RSe563p + 1./81.*m44*mp2*RNep*
         RSe564p*RSe561p*RSp563e + 1./81.*m44*mp2*RNep*RSe564p*RSp561e*
         RSe563p - 1./81.*m44*mp2*RNep*RSe564p*RSp561e*RSe561p + 1./81.
         *m44*mp2*RNep*RSp564e*RSp563e*RSe563p - 1./81.*m44*mp2*RNep*
         RSp564e*RSe561p*RSp563e - 1./81.*m44*mp2*RNep*RSp564e*RSp561e*
         RSe563p );

      M13p6x23R +=  + Ab(me2,mu2)*Ab(mp2,mu2) * ( 1./81.*m44*mp2*RNep*
         RSp564e*RSp561e*RSe561p - 1./81.*m44*mp2*RN3p*RSe564p*RSp563e*
         RS356pe + 1./81.*m44*mp2*RN3p*RSe564p*RSp561e*RS356pe + 1./81.
         *m44*mp2*RN1p*RSe564p*RS156pe*RSp563e - 1./81.*m44*mp2*RN1p*
         RSe564p*RSp561e*RS156pe - 1./81.*m44*me2*RN3e*RSp564e*RSe563p*
         RS356pe + 1./81.*m44*me2*RN3e*RSp564e*RSe561p*RS356pe + 1./81.
         *m44*me2*RN1e*RSp564e*RS156pe*RSe563p - 1./81.*m44*me2*RN1e*
         RSp564e*RSe561p*RS156pe + 13./270.*m33*RSp45613e*RSe45613p - 7.
         /540.*m33*RSe45633p*RSp45613e - 7./540.*m33*RSp45633e*
         RSe45613p + 17./360.*m33*RSp45633e*RSe45633p + 1./540.*m33*
         RSe45611p*RSp45613e - 17./1080.*m33*RSe45611p*RSp45633e + 1./
         540.*m33*RSp45611e*RSe45613p - 17./1080.*m33*RSp45611e*
         RSe45633p + 1./72.*m33*RSp45611e*RSe45611p - 1./27.*m33*mp2*
         RN3p*RSe45633p*RSp45613e + 1./27.*m33*mp2*RN3p*RSp45633e*
         RSe45633p - 1./27.*m33*me2*RN3e*RSp45633e*RSe45613p + 1./27.*
         m33*me2*RN3e*RSp45633e*RSe45633p );

      M13p6x23R +=  + Ab(me2,mu2)*Ab(mp2,mu2) * ( 1./81.*m33*m66*RNep*
         RSe456p*RSp453e*RSe453p + 1./81.*m33*m66*RNep*RSe456p*RSe451p*
         RSp453e + 1./81.*m33*m66*RNep*RSe456p*RSp451e*RSe453p - 1./81.
         *m33*m66*RNep*RSp456e*RSp453e*RSe453p - 1./81.*m33*m66*RNep*
         RSp456e*RSe451p*RSp453e - 1./81.*m33*m66*RNep*RSp456e*RSp451e*
         RSe453p + 1./81.*m33*m66*RN3p*RSe456p*RSp453e*RS345pe + 1./81.
         *m33*m66*RN3p*RSe456p*RSp451e*RS345pe + 1./81.*m33*m66*RN3e*
         RSp456e*RSe453p*RS345pe + 1./81.*m33*m66*RN3e*RSp456e*RSe451p*
         RS345pe + 1./81.*m33*m66*RN1p*RSe456p*RS145pe*RSp453e + 1./81.
         *m33*m66*RN1e*RSp456e*RS145pe*RSe453p + 1./81.*m33*m55*RNep*
         RSe465p*RSp463e*RSe463p + 1./81.*m33*m55*RNep*RSe465p*RSe461p*
         RSp463e + 1./81.*m33*m55*RNep*RSe465p*RSp461e*RSe463p - 1./81.
         *m33*m55*RNep*RSp465e*RSp463e*RSe463p - 1./81.*m33*m55*RNep*
         RSp465e*RSe461p*RSp463e - 1./81.*m33*m55*RNep*RSp465e*RSp461e*
         RSe463p + 1./81.*m33*m55*RN3p*RSe465p*RSp463e*RS346pe + 1./81.
         *m33*m55*RN3p*RSe465p*RSp461e*RS346pe );

      M13p6x23R +=  + Ab(me2,mu2)*Ab(mp2,mu2) * ( 1./81.*m33*m55*RN3e*
         RSp465e*RSe463p*RS346pe + 1./81.*m33*m55*RN3e*RSp465e*RSe461p*
         RS346pe + 1./81.*m33*m55*RN1p*RSe465p*RS146pe*RSp463e + 1./81.
         *m33*m55*RN1e*RSp465e*RS146pe*RSe463p + 1./81.*m33*m44*RNep*
         RSe564p*RSp563e*RSe563p + 1./81.*m33*m44*RNep*RSe564p*RSe561p*
         RSp563e + 1./81.*m33*m44*RNep*RSe564p*RSp561e*RSe563p - 1./81.
         *m33*m44*RNep*RSp564e*RSp563e*RSe563p - 1./81.*m33*m44*RNep*
         RSp564e*RSe561p*RSp563e - 1./81.*m33*m44*RNep*RSp564e*RSp561e*
         RSe563p + 1./81.*m33*m44*RN3p*RSe564p*RSp563e*RS356pe + 1./81.
         *m33*m44*RN3p*RSe564p*RSp561e*RS356pe + 1./81.*m33*m44*RN3e*
         RSp564e*RSe563p*RS356pe + 1./81.*m33*m44*RN3e*RSp564e*RSe561p*
         RS356pe + 1./81.*m33*m44*RN1p*RSe564p*RS156pe*RSp563e + 1./81.
         *m33*m44*RN1e*RSp564e*RS156pe*RSe563p - 1./27.*pow(m33,2)*RN3p
         *RSe45633p*RSp45613e - 1./27.*pow(m33,2)*RN3p*RSp45633e*
         RSe45633p - 1./27.*pow(m33,2)*RN3e*RSp45633e*RSe45613p - 1./27.
         *pow(m33,2)*RN3e*RSp45633e*RSe45633p );

      M13p6x23R +=  + Ab(me2,mu2)*Ab(mp2,mu2) * ( 13./270.*m11*
         RSp45613e*RSe45613p + 1./540.*m11*RSe45633p*RSp45613e + 1./540.
         *m11*RSp45633e*RSe45613p + 1./72.*m11*RSp45633e*RSe45633p - 7./
         540.*m11*RSe45611p*RSp45613e - 17./1080.*m11*RSe45611p*
         RSp45633e - 7./540.*m11*RSp45611e*RSe45613p - 17./1080.*m11*
         RSp45611e*RSe45633p + 17./360.*m11*RSp45611e*RSe45611p - 1./27.
         *m11*mp2*RN1p*RSe45611p*RSp45613e + 1./27.*m11*mp2*RN1p*
         RSp45611e*RSe45611p - 1./27.*m11*me2*RN1e*RSp45611e*RSe45613p
          + 1./27.*m11*me2*RN1e*RSp45611e*RSe45611p + 1./81.*m11*m66*
         RNep*RSe456p*RSe451p*RSp453e + 1./81.*m11*m66*RNep*RSe456p*
         RSp451e*RSe453p + 1./81.*m11*m66*RNep*RSe456p*RSp451e*RSe451p
          - 1./81.*m11*m66*RNep*RSp456e*RSe451p*RSp453e - 1./81.*m11*
         m66*RNep*RSp456e*RSp451e*RSe453p - 1./81.*m11*m66*RNep*RSp456e
         *RSp451e*RSe451p + 1./81.*m11*m66*RN3p*RSe456p*RSp451e*RS345pe
          + 1./81.*m11*m66*RN3e*RSp456e*RSe451p*RS345pe + 1./81.*m11*
         m66*RN1p*RSe456p*RS145pe*RSp453e );

      M13p6x23R +=  + Ab(me2,mu2)*Ab(mp2,mu2) * ( 1./81.*m11*m66*RN1p*
         RSe456p*RSp451e*RS145pe + 1./81.*m11*m66*RN1e*RSp456e*RS145pe*
         RSe453p + 1./81.*m11*m66*RN1e*RSp456e*RSe451p*RS145pe + 1./81.
         *m11*m55*RNep*RSe465p*RSe461p*RSp463e + 1./81.*m11*m55*RNep*
         RSe465p*RSp461e*RSe463p + 1./81.*m11*m55*RNep*RSe465p*RSp461e*
         RSe461p - 1./81.*m11*m55*RNep*RSp465e*RSe461p*RSp463e - 1./81.
         *m11*m55*RNep*RSp465e*RSp461e*RSe463p - 1./81.*m11*m55*RNep*
         RSp465e*RSp461e*RSe461p + 1./81.*m11*m55*RN3p*RSe465p*RSp461e*
         RS346pe + 1./81.*m11*m55*RN3e*RSp465e*RSe461p*RS346pe + 1./81.
         *m11*m55*RN1p*RSe465p*RS146pe*RSp463e + 1./81.*m11*m55*RN1p*
         RSe465p*RSp461e*RS146pe + 1./81.*m11*m55*RN1e*RSp465e*RS146pe*
         RSe463p + 1./81.*m11*m55*RN1e*RSp465e*RSe461p*RS146pe + 1./81.
         *m11*m44*RNep*RSe564p*RSe561p*RSp563e + 1./81.*m11*m44*RNep*
         RSe564p*RSp561e*RSe563p + 1./81.*m11*m44*RNep*RSe564p*RSp561e*
         RSe561p - 1./81.*m11*m44*RNep*RSp564e*RSe561p*RSp563e - 1./81.
         *m11*m44*RNep*RSp564e*RSp561e*RSe563p );

      M13p6x23R +=  + Ab(me2,mu2)*Ab(mp2,mu2) * (  - 1./81.*m11*m44*
         RNep*RSp564e*RSp561e*RSe561p + 1./81.*m11*m44*RN3p*RSe564p*
         RSp561e*RS356pe + 1./81.*m11*m44*RN3e*RSp564e*RSe561p*RS356pe
          + 1./81.*m11*m44*RN1p*RSe564p*RS156pe*RSp563e + 1./81.*m11*
         m44*RN1p*RSe564p*RSp561e*RS156pe + 1./81.*m11*m44*RN1e*RSp564e
         *RS156pe*RSe563p + 1./81.*m11*m44*RN1e*RSp564e*RSe561p*RS156pe
          - 1./27.*m11*m33*RN3p*RSe45633p*RSp45613e - 1./27.*m11*m33*
         RN3e*RSp45633e*RSe45613p - 1./27.*m11*m33*RN1p*RSe45611p*
         RSp45613e - 1./27.*m11*m33*RN1e*RSp45611e*RSe45613p - 1./27.*
         pow(m11,2)*RN1p*RSe45611p*RSp45613e - 1./27.*pow(m11,2)*RN1p*
         RSp45611e*RSe45611p - 1./27.*pow(m11,2)*RN1e*RSp45611e*
         RSe45613p - 1./27.*pow(m11,2)*RN1e*RSp45611e*RSe45611p );

      M13p6x23R +=  + Ab(me2,mu2)*Bb(m11,mu2) * (  - 1./270.*me2*
         RD1456ep*RSe45613p + 1./540.*me2*RD1456ep*RSe45633p + 1./540.*
         me2*RSe45611p*RD1456ep + 1./540.*m33*RD1456ep*RSe45613p - 17./
         1080.*m33*RD1456ep*RSe45633p + 1./72.*m33*RSe45611p*RD1456ep
          - 1./60.*m11*RD1456ep*RSe45613p - 1./72.*m11*RD1456ep*
         RSe45633p + 13./1080.*m11*RSe45611p*RD1456ep - 1./27.*m11*me2*
         RN1e*RD1456ep*RSe45613p + 1./27.*m11*me2*RN1e*RSe45611p*
         RD1456ep + 1./27.*m11*m33*RSe45611p*RS14563pe - 1./27.*m11*m33
         *RN1e*RD1456ep*RSe45613p + 2./27.*pow(m11,2)*RSe45611p*
         RS14563pe - 1./27.*pow(m11,2)*RN1e*RD1456ep*RSe45613p - 1./27.
         *pow(m11,2)*RN1e*RSe45611p*RD1456ep );

      M13p6x23R +=  + Ab(me2,mu2)*Bb(m33,mu2) * (  - 1./270.*me2*
         RD3456ep*RSe45613p + 1./540.*me2*RSe45633p*RD3456ep + 1./540.*
         me2*RSe45611p*RD3456ep - 1./60.*m33*RD3456ep*RSe45613p + 13./
         1080.*m33*RSe45633p*RD3456ep - 1./72.*m33*RSe45611p*RD3456ep
          - 1./27.*m33*me2*RN3e*RD3456ep*RSe45613p + 1./27.*m33*me2*
         RN3e*RSe45633p*RD3456ep + 2./27.*pow(m33,2)*RSe45633p*
         RS34561pe - 1./27.*pow(m33,2)*RN3e*RD3456ep*RSe45613p - 1./27.
         *pow(m33,2)*RN3e*RSe45633p*RD3456ep + 1./540.*m11*RD3456ep*
         RSe45613p + 1./72.*m11*RSe45633p*RD3456ep - 17./1080.*m11*
         RSe45611p*RD3456ep + 1./27.*m11*m33*RSe45633p*RS34561pe - 1./
         27.*m11*m33*RN3e*RD3456ep*RSe45613p );

      M13p6x23R +=  + Ab(me2,mu2)*Bb(me2,mu2) * ( 1./162.*m66*me2*
         RSe456p*pow(RSe453p,2) - 1./81.*m66*me2*RSe456p*RSe451p*
         RSe453p + 1./162.*m66*me2*RSe456p*pow(RSe451p,2) + 1./162.*m55
         *me2*RSe465p*pow(RSe463p,2) - 1./81.*m55*me2*RSe465p*RSe461p*
         RSe463p + 1./162.*m55*me2*RSe465p*pow(RSe461p,2) + 1./162.*m44
         *me2*RSe564p*pow(RSe563p,2) - 1./81.*m44*me2*RSe564p*RSe561p*
         RSe563p + 1./162.*m44*me2*RSe564p*pow(RSe561p,2) - 1./162.*m33
         *m66*RSe456p*pow(RSe453p,2) - 1./81.*m33*m66*RSe456p*RSe451p*
         RSe453p - 1./162.*m33*m55*RSe465p*pow(RSe463p,2) - 1./81.*m33*
         m55*RSe465p*RSe461p*RSe463p - 1./162.*m33*m44*RSe564p*pow(
         RSe563p,2) - 1./81.*m33*m44*RSe564p*RSe561p*RSe563p - 1./81.*
         m11*m66*RSe456p*RSe451p*RSe453p - 1./162.*m11*m66*RSe456p*pow(
         RSe451p,2) - 1./81.*m11*m55*RSe465p*RSe461p*RSe463p - 1./162.*
         m11*m55*RSe465p*pow(RSe461p,2) - 1./81.*m11*m44*RSe564p*
         RSe561p*RSe563p - 1./162.*m11*m44*RSe564p*pow(RSe561p,2) );

      M13p6x23R +=  + Ab(me2,mu2)*Bb(mp2,mu2) * ( 1./162.*m66*mp2*
         RSe456p*pow(RSp453e,2) - 1./81.*m66*mp2*RSe456p*RSp451e*
         RSp453e + 1./162.*m66*mp2*RSe456p*pow(RSp451e,2) + 1./162.*m55
         *mp2*RSe465p*pow(RSp463e,2) - 1./81.*m55*mp2*RSe465p*RSp461e*
         RSp463e + 1./162.*m55*mp2*RSe465p*pow(RSp461e,2) + 1./162.*m44
         *mp2*RSe564p*pow(RSp563e,2) - 1./81.*m44*mp2*RSe564p*RSp561e*
         RSp563e + 1./162.*m44*mp2*RSe564p*pow(RSp561e,2) - 1./162.*m33
         *m66*RSe456p*pow(RSp453e,2) - 1./81.*m33*m66*RSe456p*RSp451e*
         RSp453e - 1./162.*m33*m55*RSe465p*pow(RSp463e,2) - 1./81.*m33*
         m55*RSe465p*RSp461e*RSp463e - 1./162.*m33*m44*RSe564p*pow(
         RSp563e,2) - 1./81.*m33*m44*RSe564p*RSp561e*RSp563e - 1./81.*
         m11*m66*RSe456p*RSp451e*RSp453e - 1./162.*m11*m66*RSe456p*pow(
         RSp451e,2) - 1./81.*m11*m55*RSe465p*RSp461e*RSp463e - 1./162.*
         m11*m55*RSe465p*pow(RSp461e,2) - 1./81.*m11*m44*RSe564p*
         RSp561e*RSp563e - 1./162.*m11*m44*RSe564p*pow(RSp561e,2) );

      M13p6x23R +=  + Ab(mp2,mu2)*pi16 * (  - 1./72.*m33*RD3456ep*
         RSp45613e + 1./144.*m33*RSp45633e*RD3456ep - 1./72.*m33*
         RD1456ep*RSp45613e + 1./144.*m33*RD1456ep*RSp45633e + 1./144.*
         m33*RSp45611e*RD3456ep + 1./144.*m33*RSp45611e*RD1456ep - 1./
         24.*m33*mp2*RSp45613e + 1./36.*m33*mp2*pow(RSp45613e,2) + 1./
         48.*m33*mp2*RSp45633e - 1./36.*m33*mp2*RSp45633e*RSp45613e + 1.
         /144.*m33*mp2*pow(RSp45633e,2) + 1./48.*m33*mp2*RSp563e + 1./
         48.*m33*mp2*RSp463e + 1./48.*m33*mp2*RSp453e + 1./48.*m33*mp2*
         RSp45611e - 1./36.*m33*mp2*RSp45611e*RSp45613e + 1./72.*m33*
         mp2*RSp45611e*RSp45633e + 1./144.*m33*mp2*pow(RSp45611e,2) + 1.
         /48.*m33*mp2*RSp561e + 1./48.*m33*mp2*RSp461e + 1./48.*m33*mp2
         *RSp451e + 1./36.*m33*me2*RSp45613e*RSe45613p - 1./72.*m33*me2
         *RSe45633p*RSp45613e - 1./72.*m33*me2*RSp45633e*RSe45613p + 1./
         144.*m33*me2*RSp45633e*RSe45633p - 1./72.*m33*me2*RSe45611p*
         RSp45613e + 1./144.*m33*me2*RSe45611p*RSp45633e - 1./72.*m33*
         me2*RSp45611e*RSe45613p );

      M13p6x23R +=  + Ab(mp2,mu2)*pi16 * ( 1./144.*m33*me2*RSp45611e*
         RSe45633p + 1./144.*m33*me2*RSp45611e*RSe45611p + 1./24.*m33*
         m66*RSp45613e - 1./48.*m33*m66*RSp45633e - 1./48.*m33*m66*
         RSp45611e + 1./24.*m33*m55*RSp45613e - 1./48.*m33*m55*
         RSp45633e - 1./48.*m33*m55*RSp45611e + 1./24.*m33*m44*
         RSp45613e - 1./48.*m33*m44*RSp45633e - 1./48.*m33*m44*
         RSp45611e + 1./24.*pow(m33,2)*RSp45613e + 1./27.*pow(m33,2)*
         RSp45613e*RS34561pe + 1./108.*pow(m33,2)*RSp45613e*RS14563pe
          + 1./108.*pow(m33,2)*RSp45613e*RSe45613p + 1./108.*pow(m33,2)
         *pow(RSp45613e,2) + 1./54.*pow(m33,2)*RS3456pee*RSp45613e + 1./
         54.*pow(m33,2)*RS3456ppe*RSp45613e - 1./216.*pow(m33,2)*
         RSe45633p*RSp45613e - 1./48.*pow(m33,2)*RSp45633e - 1./54.*
         pow(m33,2)*RSp45633e*RS34561pe - 1./216.*pow(m33,2)*RSp45633e*
         RS14563pe - 1./216.*pow(m33,2)*RSp45633e*RSe45613p - 1./108.*
         pow(m33,2)*RSp45633e*RSp45613e - 1./108.*pow(m33,2)*RSp45633e*
         RS3456pee );

      M13p6x23R +=  + Ab(mp2,mu2)*pi16 * (  - 1./108.*pow(m33,2)*
         RSp45633e*RS3456ppe + 1./432.*pow(m33,2)*RSp45633e*RSe45633p
          + 1./432.*pow(m33,2)*pow(RSp45633e,2) - 1./54.*pow(m33,2)*
         RS345pe*RSp45613e + 1./108.*pow(m33,2)*RS345pe*RSp45633e - 1./
         54.*pow(m33,2)*RS346pe*RSp45613e + 1./108.*pow(m33,2)*RS346pe*
         RSp45633e - 1./54.*pow(m33,2)*RS356pe*RSp45613e + 1./108.*pow(
         m33,2)*RS356pe*RSp45633e + 1./216.*pow(m33,2)*RS1456pee*
         RSp45613e - 1./432.*pow(m33,2)*RS1456pee*RSp45633e + 1./216.*
         pow(m33,2)*RS1456ppe*RSp45613e - 1./432.*pow(m33,2)*RS1456ppe*
         RSp45633e - 1./216.*pow(m33,2)*RSe45611p*RSp45613e + 1./432.*
         pow(m33,2)*RSe45611p*RSp45633e - 1./48.*pow(m33,2)*RSp45611e
          - 1./54.*pow(m33,2)*RSp45611e*RS34561pe - 1./216.*pow(m33,2)*
         RSp45611e*RS14563pe - 1./216.*pow(m33,2)*RSp45611e*RSe45613p
          - 1./108.*pow(m33,2)*RSp45611e*RSp45613e - 1./108.*pow(m33,2)
         *RSp45611e*RS3456pee - 1./108.*pow(m33,2)*RSp45611e*RS3456ppe
          + 1./432.*pow(m33,2)*RSp45611e*RSe45633p );

      M13p6x23R +=  + Ab(mp2,mu2)*pi16 * ( 1./216.*pow(m33,2)*RSp45611e
         *RSp45633e + 1./108.*pow(m33,2)*RSp45611e*RS345pe + 1./108.*
         pow(m33,2)*RSp45611e*RS346pe + 1./108.*pow(m33,2)*RSp45611e*
         RS356pe - 1./432.*pow(m33,2)*RSp45611e*RS1456pee - 1./432.*
         pow(m33,2)*RSp45611e*RS1456ppe + 1./432.*pow(m33,2)*RSp45611e*
         RSe45611p + 1./432.*pow(m33,2)*pow(RSp45611e,2) - 1./216.*pow(
         m33,2)*RS145pe*RSp45613e + 1./432.*pow(m33,2)*RS145pe*
         RSp45633e + 1./432.*pow(m33,2)*RS145pe*RSp45611e - 1./216.*
         pow(m33,2)*RS146pe*RSp45613e + 1./432.*pow(m33,2)*RS146pe*
         RSp45633e + 1./432.*pow(m33,2)*RS146pe*RSp45611e - 1./216.*
         pow(m33,2)*RS156pe*RSp45613e + 1./432.*pow(m33,2)*RS156pe*
         RSp45633e + 1./432.*pow(m33,2)*RS156pe*RSp45611e - 1./72.*m11*
         RD3456ep*RSp45613e + 1./144.*m11*RSp45633e*RD3456ep - 1./72.*
         m11*RD1456ep*RSp45613e + 1./144.*m11*RD1456ep*RSp45633e + 1./
         144.*m11*RSp45611e*RD3456ep + 1./144.*m11*RSp45611e*RD1456ep
          - 1./24.*m11*mp2*RSp45613e );

      M13p6x23R +=  + Ab(mp2,mu2)*pi16 * ( 1./36.*m11*mp2*pow(
         RSp45613e,2) + 1./48.*m11*mp2*RSp45633e - 1./36.*m11*mp2*
         RSp45633e*RSp45613e + 1./144.*m11*mp2*pow(RSp45633e,2) + 1./48.
         *m11*mp2*RSp563e + 1./48.*m11*mp2*RSp463e + 1./48.*m11*mp2*
         RSp453e + 1./48.*m11*mp2*RSp45611e - 1./36.*m11*mp2*RSp45611e*
         RSp45613e + 1./72.*m11*mp2*RSp45611e*RSp45633e + 1./144.*m11*
         mp2*pow(RSp45611e,2) + 1./48.*m11*mp2*RSp561e + 1./48.*m11*mp2
         *RSp461e + 1./48.*m11*mp2*RSp451e + 1./36.*m11*me2*RSp45613e*
         RSe45613p - 1./72.*m11*me2*RSe45633p*RSp45613e - 1./72.*m11*
         me2*RSp45633e*RSe45613p + 1./144.*m11*me2*RSp45633e*RSe45633p
          - 1./72.*m11*me2*RSe45611p*RSp45613e + 1./144.*m11*me2*
         RSe45611p*RSp45633e - 1./72.*m11*me2*RSp45611e*RSe45613p + 1./
         144.*m11*me2*RSp45611e*RSe45633p + 1./144.*m11*me2*RSp45611e*
         RSe45611p + 1./24.*m11*m66*RSp45613e - 1./48.*m11*m66*
         RSp45633e - 1./48.*m11*m66*RSp45611e + 1./24.*m11*m55*
         RSp45613e );

      M13p6x23R +=  + Ab(mp2,mu2)*pi16 * (  - 1./48.*m11*m55*RSp45633e
          - 1./48.*m11*m55*RSp45611e + 1./24.*m11*m44*RSp45613e - 1./48.
         *m11*m44*RSp45633e - 1./48.*m11*m44*RSp45611e + 1./12.*m11*m33
         *RSp45613e + 5./108.*m11*m33*RSp45613e*RS34561pe + 5./108.*m11
         *m33*RSp45613e*RS14563pe + 1./54.*m11*m33*RSp45613e*RSe45613p
          + 1./54.*m11*m33*pow(RSp45613e,2) + 5./216.*m11*m33*RS3456pee
         *RSp45613e + 5./216.*m11*m33*RS3456ppe*RSp45613e - 1./108.*m11
         *m33*RSe45633p*RSp45613e - 1./24.*m11*m33*RSp45633e - 5./216.*
         m11*m33*RSp45633e*RS34561pe - 5./216.*m11*m33*RSp45633e*
         RS14563pe - 1./108.*m11*m33*RSp45633e*RSe45613p - 1./54.*m11*
         m33*RSp45633e*RSp45613e - 5./432.*m11*m33*RSp45633e*RS3456pee
          - 5./432.*m11*m33*RSp45633e*RS3456ppe + 1./216.*m11*m33*
         RSp45633e*RSe45633p + 1./216.*m11*m33*pow(RSp45633e,2) - 5./
         216.*m11*m33*RS345pe*RSp45613e + 5./432.*m11*m33*RS345pe*
         RSp45633e - 5./216.*m11*m33*RS346pe*RSp45613e + 5./432.*m11*
         m33*RS346pe*RSp45633e );

      M13p6x23R +=  + Ab(mp2,mu2)*pi16 * (  - 5./216.*m11*m33*RS356pe*
         RSp45613e + 5./432.*m11*m33*RS356pe*RSp45633e + 5./216.*m11*
         m33*RS1456pee*RSp45613e - 5./432.*m11*m33*RS1456pee*RSp45633e
          + 5./216.*m11*m33*RS1456ppe*RSp45613e - 5./432.*m11*m33*
         RS1456ppe*RSp45633e - 1./108.*m11*m33*RSe45611p*RSp45613e + 1./
         216.*m11*m33*RSe45611p*RSp45633e - 1./24.*m11*m33*RSp45611e - 
         5./216.*m11*m33*RSp45611e*RS34561pe - 5./216.*m11*m33*
         RSp45611e*RS14563pe - 1./108.*m11*m33*RSp45611e*RSe45613p - 1./
         54.*m11*m33*RSp45611e*RSp45613e - 5./432.*m11*m33*RSp45611e*
         RS3456pee - 5./432.*m11*m33*RSp45611e*RS3456ppe + 1./216.*m11*
         m33*RSp45611e*RSe45633p + 1./108.*m11*m33*RSp45611e*RSp45633e
          + 5./432.*m11*m33*RSp45611e*RS345pe + 5./432.*m11*m33*
         RSp45611e*RS346pe + 5./432.*m11*m33*RSp45611e*RS356pe - 5./432.
         *m11*m33*RSp45611e*RS1456pee - 5./432.*m11*m33*RSp45611e*
         RS1456ppe + 1./216.*m11*m33*RSp45611e*RSe45611p + 1./216.*m11*
         m33*pow(RSp45611e,2) );

      M13p6x23R +=  + Ab(mp2,mu2)*pi16 * (  - 5./216.*m11*m33*RS145pe*
         RSp45613e + 5./432.*m11*m33*RS145pe*RSp45633e + 5./432.*m11*
         m33*RS145pe*RSp45611e - 5./216.*m11*m33*RS146pe*RSp45613e + 5./
         432.*m11*m33*RS146pe*RSp45633e + 5./432.*m11*m33*RS146pe*
         RSp45611e - 5./216.*m11*m33*RS156pe*RSp45613e + 5./432.*m11*
         m33*RS156pe*RSp45633e + 5./432.*m11*m33*RS156pe*RSp45611e + 1./
         24.*pow(m11,2)*RSp45613e + 1./108.*pow(m11,2)*RSp45613e*
         RS34561pe + 1./27.*pow(m11,2)*RSp45613e*RS14563pe + 1./108.*
         pow(m11,2)*RSp45613e*RSe45613p + 1./108.*pow(m11,2)*pow(
         RSp45613e,2) + 1./216.*pow(m11,2)*RS3456pee*RSp45613e + 1./216.
         *pow(m11,2)*RS3456ppe*RSp45613e - 1./216.*pow(m11,2)*RSe45633p
         *RSp45613e - 1./48.*pow(m11,2)*RSp45633e - 1./216.*pow(m11,2)*
         RSp45633e*RS34561pe - 1./54.*pow(m11,2)*RSp45633e*RS14563pe - 
         1./216.*pow(m11,2)*RSp45633e*RSe45613p - 1./108.*pow(m11,2)*
         RSp45633e*RSp45613e - 1./432.*pow(m11,2)*RSp45633e*RS3456pee
          - 1./432.*pow(m11,2)*RSp45633e*RS3456ppe );

      M13p6x23R +=  + Ab(mp2,mu2)*pi16 * ( 1./432.*pow(m11,2)*RSp45633e
         *RSe45633p + 1./432.*pow(m11,2)*pow(RSp45633e,2) - 1./216.*
         pow(m11,2)*RS345pe*RSp45613e + 1./432.*pow(m11,2)*RS345pe*
         RSp45633e - 1./216.*pow(m11,2)*RS346pe*RSp45613e + 1./432.*
         pow(m11,2)*RS346pe*RSp45633e - 1./216.*pow(m11,2)*RS356pe*
         RSp45613e + 1./432.*pow(m11,2)*RS356pe*RSp45633e + 1./54.*pow(
         m11,2)*RS1456pee*RSp45613e - 1./108.*pow(m11,2)*RS1456pee*
         RSp45633e + 1./54.*pow(m11,2)*RS1456ppe*RSp45613e - 1./108.*
         pow(m11,2)*RS1456ppe*RSp45633e - 1./216.*pow(m11,2)*RSe45611p*
         RSp45613e + 1./432.*pow(m11,2)*RSe45611p*RSp45633e - 1./48.*
         pow(m11,2)*RSp45611e - 1./216.*pow(m11,2)*RSp45611e*RS34561pe
          - 1./54.*pow(m11,2)*RSp45611e*RS14563pe - 1./216.*pow(m11,2)*
         RSp45611e*RSe45613p - 1./108.*pow(m11,2)*RSp45611e*RSp45613e
          - 1./432.*pow(m11,2)*RSp45611e*RS3456pee - 1./432.*pow(m11,2)
         *RSp45611e*RS3456ppe + 1./432.*pow(m11,2)*RSp45611e*RSe45633p
          + 1./216.*pow(m11,2)*RSp45611e*RSp45633e );

      M13p6x23R +=  + Ab(mp2,mu2)*pi16 * ( 1./432.*pow(m11,2)*RSp45611e
         *RS345pe + 1./432.*pow(m11,2)*RSp45611e*RS346pe + 1./432.*pow(
         m11,2)*RSp45611e*RS356pe - 1./108.*pow(m11,2)*RSp45611e*
         RS1456pee - 1./108.*pow(m11,2)*RSp45611e*RS1456ppe + 1./432.*
         pow(m11,2)*RSp45611e*RSe45611p + 1./432.*pow(m11,2)*pow(
         RSp45611e,2) - 1./54.*pow(m11,2)*RS145pe*RSp45613e + 1./108.*
         pow(m11,2)*RS145pe*RSp45633e + 1./108.*pow(m11,2)*RS145pe*
         RSp45611e - 1./54.*pow(m11,2)*RS146pe*RSp45613e + 1./108.*pow(
         m11,2)*RS146pe*RSp45633e + 1./108.*pow(m11,2)*RS146pe*
         RSp45611e - 1./54.*pow(m11,2)*RS156pe*RSp45613e + 1./108.*pow(
         m11,2)*RS156pe*RSp45633e + 1./108.*pow(m11,2)*RS156pe*
         RSp45611e );

      M13p6x23R +=  + pow(Ab(mp2,mu2),2) * ( 1./135.*mp2*pow(
         RSp45613e,2) - 1./135.*mp2*RSp45633e*RSp45613e + 1./540.*mp2*
         pow(RSp45633e,2) - 1./135.*mp2*RSp45611e*RSp45613e + 1./270.*
         mp2*RSp45611e*RSp45633e + 1./540.*mp2*pow(RSp45611e,2) + 1./
         162.*m66*RSp456e*pow(RSp453e,2) - 1./81.*m66*RSp456e*RSp451e*
         RSp453e + 1./162.*m66*RSp456e*pow(RSp451e,2) - 1./81.*m66*mp2*
         RNep*RSp456e*RSp453e*RSe453p + 1./81.*m66*mp2*RNep*RSp456e*
         RSe451p*RSp453e + 1./81.*m66*mp2*RNep*RSp456e*RSp451e*RSe453p
          - 1./81.*m66*mp2*RNep*RSp456e*RSp451e*RSe451p - 1./81.*m66*
         mp2*RN3p*RSp456e*RSp453e*RS345pe + 1./81.*m66*mp2*RN3p*RSp456e
         *RSp451e*RS345pe + 1./81.*m66*mp2*RN1p*RSp456e*RS145pe*RSp453e
          - 1./81.*m66*mp2*RN1p*RSp456e*RSp451e*RS145pe + 1./162.*m55*
         RSp465e*pow(RSp463e,2) - 1./81.*m55*RSp465e*RSp461e*RSp463e + 
         1./162.*m55*RSp465e*pow(RSp461e,2) - 1./81.*m55*mp2*RNep*
         RSp465e*RSp463e*RSe463p + 1./81.*m55*mp2*RNep*RSp465e*RSe461p*
         RSp463e );

      M13p6x23R +=  + pow(Ab(mp2,mu2),2) * ( 1./81.*m55*mp2*RNep*
         RSp465e*RSp461e*RSe463p - 1./81.*m55*mp2*RNep*RSp465e*RSp461e*
         RSe461p - 1./81.*m55*mp2*RN3p*RSp465e*RSp463e*RS346pe + 1./81.
         *m55*mp2*RN3p*RSp465e*RSp461e*RS346pe + 1./81.*m55*mp2*RN1p*
         RSp465e*RS146pe*RSp463e - 1./81.*m55*mp2*RN1p*RSp465e*RSp461e*
         RS146pe + 1./162.*m44*RSp564e*pow(RSp563e,2) - 1./81.*m44*
         RSp564e*RSp561e*RSp563e + 1./162.*m44*RSp564e*pow(RSp561e,2)
          - 1./81.*m44*mp2*RNep*RSp564e*RSp563e*RSe563p + 1./81.*m44*
         mp2*RNep*RSp564e*RSe561p*RSp563e + 1./81.*m44*mp2*RNep*RSp564e
         *RSp561e*RSe563p - 1./81.*m44*mp2*RNep*RSp564e*RSp561e*RSe561p
          - 1./81.*m44*mp2*RN3p*RSp564e*RSp563e*RS356pe + 1./81.*m44*
         mp2*RN3p*RSp564e*RSp561e*RS356pe + 1./81.*m44*mp2*RN1p*RSp564e
         *RS156pe*RSp563e - 1./81.*m44*mp2*RN1p*RSp564e*RSp561e*RS156pe
          + 13./540.*m33*pow(RSp45613e,2) - 7./540.*m33*RSp45633e*
         RSp45613e + 17./720.*m33*pow(RSp45633e,2) + 1./540.*m33*
         RSp45611e*RSp45613e );

      M13p6x23R +=  + pow(Ab(mp2,mu2),2) * (  - 17./1080.*m33*RSp45611e
         *RSp45633e + 1./144.*m33*pow(RSp45611e,2) - 1./27.*m33*mp2*
         RN3p*RSp45633e*RSp45613e + 1./27.*m33*mp2*RN3p*pow(
         RSp45633e,2) + 1./81.*m33*m66*RNep*RSp456e*RSp453e*RSe453p + 1.
         /81.*m33*m66*RNep*RSp456e*RSe451p*RSp453e + 1./81.*m33*m66*
         RNep*RSp456e*RSp451e*RSe453p + 1./81.*m33*m66*RN3p*RSp456e*
         RSp453e*RS345pe + 1./81.*m33*m66*RN3p*RSp456e*RSp451e*RS345pe
          + 1./81.*m33*m66*RN1p*RSp456e*RS145pe*RSp453e + 1./81.*m33*
         m55*RNep*RSp465e*RSp463e*RSe463p + 1./81.*m33*m55*RNep*RSp465e
         *RSe461p*RSp463e + 1./81.*m33*m55*RNep*RSp465e*RSp461e*RSe463p
          + 1./81.*m33*m55*RN3p*RSp465e*RSp463e*RS346pe + 1./81.*m33*
         m55*RN3p*RSp465e*RSp461e*RS346pe + 1./81.*m33*m55*RN1p*RSp465e
         *RS146pe*RSp463e + 1./81.*m33*m44*RNep*RSp564e*RSp563e*RSe563p
          + 1./81.*m33*m44*RNep*RSp564e*RSe561p*RSp563e + 1./81.*m33*
         m44*RNep*RSp564e*RSp561e*RSe563p + 1./81.*m33*m44*RN3p*RSp564e
         *RSp563e*RS356pe );

      M13p6x23R +=  + pow(Ab(mp2,mu2),2) * ( 1./81.*m33*m44*RN3p*
         RSp564e*RSp561e*RS356pe + 1./81.*m33*m44*RN1p*RSp564e*RS156pe*
         RSp563e - 1./27.*pow(m33,2)*RN3p*RSp45633e*RSp45613e - 1./27.*
         pow(m33,2)*RN3p*pow(RSp45633e,2) + 13./540.*m11*pow(
         RSp45613e,2) + 1./540.*m11*RSp45633e*RSp45613e + 1./144.*m11*
         pow(RSp45633e,2) - 7./540.*m11*RSp45611e*RSp45613e - 17./1080.
         *m11*RSp45611e*RSp45633e + 17./720.*m11*pow(RSp45611e,2) - 1./
         27.*m11*mp2*RN1p*RSp45611e*RSp45613e + 1./27.*m11*mp2*RN1p*
         pow(RSp45611e,2) + 1./81.*m11*m66*RNep*RSp456e*RSe451p*RSp453e
          + 1./81.*m11*m66*RNep*RSp456e*RSp451e*RSe453p + 1./81.*m11*
         m66*RNep*RSp456e*RSp451e*RSe451p + 1./81.*m11*m66*RN3p*RSp456e
         *RSp451e*RS345pe + 1./81.*m11*m66*RN1p*RSp456e*RS145pe*RSp453e
          + 1./81.*m11*m66*RN1p*RSp456e*RSp451e*RS145pe + 1./81.*m11*
         m55*RNep*RSp465e*RSe461p*RSp463e + 1./81.*m11*m55*RNep*RSp465e
         *RSp461e*RSe463p + 1./81.*m11*m55*RNep*RSp465e*RSp461e*RSe461p
          + 1./81.*m11*m55*RN3p*RSp465e*RSp461e*RS346pe );

      M13p6x23R +=  + pow(Ab(mp2,mu2),2) * ( 1./81.*m11*m55*RN1p*
         RSp465e*RS146pe*RSp463e + 1./81.*m11*m55*RN1p*RSp465e*RSp461e*
         RS146pe + 1./81.*m11*m44*RNep*RSp564e*RSe561p*RSp563e + 1./81.
         *m11*m44*RNep*RSp564e*RSp561e*RSe563p + 1./81.*m11*m44*RNep*
         RSp564e*RSp561e*RSe561p + 1./81.*m11*m44*RN3p*RSp564e*RSp561e*
         RS356pe + 1./81.*m11*m44*RN1p*RSp564e*RS156pe*RSp563e + 1./81.
         *m11*m44*RN1p*RSp564e*RSp561e*RS156pe - 1./27.*m11*m33*RN3p*
         RSp45633e*RSp45613e - 1./27.*m11*m33*RN1p*RSp45611e*RSp45613e
          - 1./27.*pow(m11,2)*RN1p*RSp45611e*RSp45613e - 1./27.*pow(
         m11,2)*RN1p*pow(RSp45611e,2) );

      M13p6x23R +=  + Ab(mp2,mu2)*Bb(m11,mu2) * (  - 1./270.*mp2*
         RD1456ep*RSp45613e + 1./540.*mp2*RD1456ep*RSp45633e + 1./540.*
         mp2*RSp45611e*RD1456ep + 1./540.*m33*RD1456ep*RSp45613e - 17./
         1080.*m33*RD1456ep*RSp45633e + 1./72.*m33*RSp45611e*RD1456ep
          - 1./60.*m11*RD1456ep*RSp45613e - 1./72.*m11*RD1456ep*
         RSp45633e + 13./1080.*m11*RSp45611e*RD1456ep - 1./27.*m11*mp2*
         RN1p*RD1456ep*RSp45613e + 1./27.*m11*mp2*RN1p*RSp45611e*
         RD1456ep + 1./27.*m11*m33*RSp45611e*RS14563pe - 1./27.*m11*m33
         *RN1p*RD1456ep*RSp45613e + 2./27.*pow(m11,2)*RSp45611e*
         RS14563pe - 1./27.*pow(m11,2)*RN1p*RD1456ep*RSp45613e - 1./27.
         *pow(m11,2)*RN1p*RSp45611e*RD1456ep );

      M13p6x23R +=  + Ab(mp2,mu2)*Bb(m33,mu2) * (  - 1./270.*mp2*
         RD3456ep*RSp45613e + 1./540.*mp2*RSp45633e*RD3456ep + 1./540.*
         mp2*RSp45611e*RD3456ep - 1./60.*m33*RD3456ep*RSp45613e + 13./
         1080.*m33*RSp45633e*RD3456ep - 1./72.*m33*RSp45611e*RD3456ep
          - 1./27.*m33*mp2*RN3p*RD3456ep*RSp45613e + 1./27.*m33*mp2*
         RN3p*RSp45633e*RD3456ep + 2./27.*pow(m33,2)*RSp45633e*
         RS34561pe - 1./27.*pow(m33,2)*RN3p*RD3456ep*RSp45613e - 1./27.
         *pow(m33,2)*RN3p*RSp45633e*RD3456ep + 1./540.*m11*RD3456ep*
         RSp45613e + 1./72.*m11*RSp45633e*RD3456ep - 17./1080.*m11*
         RSp45611e*RD3456ep + 1./27.*m11*m33*RSp45633e*RS34561pe - 1./
         27.*m11*m33*RN3p*RD3456ep*RSp45613e );

      M13p6x23R +=  + Ab(mp2,mu2)*Bb(me2,mu2) * ( 1./162.*m66*me2*
         RSp456e*pow(RSe453p,2) - 1./81.*m66*me2*RSp456e*RSe451p*
         RSe453p + 1./162.*m66*me2*RSp456e*pow(RSe451p,2) + 1./162.*m55
         *me2*RSp465e*pow(RSe463p,2) - 1./81.*m55*me2*RSp465e*RSe461p*
         RSe463p + 1./162.*m55*me2*RSp465e*pow(RSe461p,2) + 1./162.*m44
         *me2*RSp564e*pow(RSe563p,2) - 1./81.*m44*me2*RSp564e*RSe561p*
         RSe563p + 1./162.*m44*me2*RSp564e*pow(RSe561p,2) - 1./162.*m33
         *m66*RSp456e*pow(RSe453p,2) - 1./81.*m33*m66*RSp456e*RSe451p*
         RSe453p - 1./162.*m33*m55*RSp465e*pow(RSe463p,2) - 1./81.*m33*
         m55*RSp465e*RSe461p*RSe463p - 1./162.*m33*m44*RSp564e*pow(
         RSe563p,2) - 1./81.*m33*m44*RSp564e*RSe561p*RSe563p - 1./81.*
         m11*m66*RSp456e*RSe451p*RSe453p - 1./162.*m11*m66*RSp456e*pow(
         RSe451p,2) - 1./81.*m11*m55*RSp465e*RSe461p*RSe463p - 1./162.*
         m11*m55*RSp465e*pow(RSe461p,2) - 1./81.*m11*m44*RSp564e*
         RSe561p*RSe563p - 1./162.*m11*m44*RSp564e*pow(RSe561p,2) );

      M13p6x23R +=  + Ab(mp2,mu2)*Bb(mp2,mu2) * ( 1./162.*m66*mp2*
         RSp456e*pow(RSp453e,2) - 1./81.*m66*mp2*RSp456e*RSp451e*
         RSp453e + 1./162.*m66*mp2*RSp456e*pow(RSp451e,2) + 1./162.*m55
         *mp2*RSp465e*pow(RSp463e,2) - 1./81.*m55*mp2*RSp465e*RSp461e*
         RSp463e + 1./162.*m55*mp2*RSp465e*pow(RSp461e,2) + 1./162.*m44
         *mp2*RSp564e*pow(RSp563e,2) - 1./81.*m44*mp2*RSp564e*RSp561e*
         RSp563e + 1./162.*m44*mp2*RSp564e*pow(RSp561e,2) - 1./162.*m33
         *m66*RSp456e*pow(RSp453e,2) - 1./81.*m33*m66*RSp456e*RSp451e*
         RSp453e - 1./162.*m33*m55*RSp465e*pow(RSp463e,2) - 1./81.*m33*
         m55*RSp465e*RSp461e*RSp463e - 1./162.*m33*m44*RSp564e*pow(
         RSp563e,2) - 1./81.*m33*m44*RSp564e*RSp561e*RSp563e - 1./81.*
         m11*m66*RSp456e*RSp451e*RSp453e - 1./162.*m11*m66*RSp456e*pow(
         RSp451e,2) - 1./81.*m11*m55*RSp465e*RSp461e*RSp463e - 1./162.*
         m11*m55*RSp465e*pow(RSp461e,2) - 1./81.*m11*m44*RSp564e*
         RSp561e*RSp563e - 1./162.*m11*m44*RSp564e*pow(RSp561e,2) );

      M13p6x23R +=  + Bb(m11,mu2)*pi16 * ( 1./144.*m33*RD1456ep*
         RD3456ep + 1./144.*m33*pow(RD1456ep,2) - 1./72.*m33*mp2*
         RD1456ep*RSp45613e + 1./144.*m33*mp2*RD1456ep*RSp45633e + 1./
         144.*m33*mp2*RSp45611e*RD1456ep - 1./72.*m33*me2*RD1456ep*
         RSe45613p + 1./144.*m33*me2*RD1456ep*RSe45633p + 1./144.*m33*
         me2*RSe45611p*RD1456ep - 1./48.*m33*m66*RD1456ep - 1./48.*m33*
         m55*RD1456ep - 1./48.*m33*m44*RD1456ep - 1./48.*pow(m33,2)*
         RD1456ep - 1./54.*pow(m33,2)*RD1456ep*RS34561pe - 1./216.*pow(
         m33,2)*RD1456ep*RS14563pe - 1./216.*pow(m33,2)*RD1456ep*
         RSe45613p - 1./216.*pow(m33,2)*RD1456ep*RSp45613e - 1./108.*
         pow(m33,2)*RD1456ep*RS3456pee - 1./108.*pow(m33,2)*RD1456ep*
         RS3456ppe + 1./432.*pow(m33,2)*RD1456ep*RSe45633p + 1./432.*
         pow(m33,2)*RD1456ep*RSp45633e + 1./108.*pow(m33,2)*RD1456ep*
         RS345pe + 1./108.*pow(m33,2)*RD1456ep*RS346pe + 1./108.*pow(
         m33,2)*RD1456ep*RS356pe - 1./432.*pow(m33,2)*RS1456pee*
         RD1456ep );

      M13p6x23R +=  + Bb(m11,mu2)*pi16 * (  - 1./432.*pow(m33,2)*
         RS1456ppe*RD1456ep + 1./432.*pow(m33,2)*RSe45611p*RD1456ep + 1.
         /432.*pow(m33,2)*RSp45611e*RD1456ep + 1./432.*pow(m33,2)*
         RS145pe*RD1456ep + 1./432.*pow(m33,2)*RS146pe*RD1456ep + 1./
         432.*pow(m33,2)*RS156pe*RD1456ep + 1./144.*m11*RD1456ep*
         RD3456ep + 1./144.*m11*pow(RD1456ep,2) - 1./72.*m11*mp2*
         RD1456ep*RSp45613e + 1./144.*m11*mp2*RD1456ep*RSp45633e + 1./
         144.*m11*mp2*RSp45611e*RD1456ep - 1./72.*m11*me2*RD1456ep*
         RSe45613p + 1./144.*m11*me2*RD1456ep*RSe45633p + 1./144.*m11*
         me2*RSe45611p*RD1456ep - 1./48.*m11*m66*RD1456ep - 1./48.*m11*
         m55*RD1456ep - 1./48.*m11*m44*RD1456ep - 1./48.*m11*m33*
         RD1456ep - 5./216.*m11*m33*RD1456ep*RS34561pe - 5./216.*m11*
         m33*RD1456ep*RS14563pe - 1./108.*m11*m33*RD1456ep*RSe45613p - 
         1./108.*m11*m33*RD1456ep*RSp45613e - 5./432.*m11*m33*RD1456ep*
         RS3456pee - 5./432.*m11*m33*RD1456ep*RS3456ppe + 1./216.*m11*
         m33*RD1456ep*RSe45633p );

      M13p6x23R +=  + Bb(m11,mu2)*pi16 * ( 1./216.*m11*m33*RD1456ep*
         RSp45633e + 5./432.*m11*m33*RD1456ep*RS345pe + 5./432.*m11*m33
         *RD1456ep*RS346pe + 5./432.*m11*m33*RD1456ep*RS356pe - 5./432.
         *m11*m33*RS1456pee*RD1456ep - 5./432.*m11*m33*RS1456ppe*
         RD1456ep + 1./216.*m11*m33*RSe45611p*RD1456ep + 1./216.*m11*
         m33*RSp45611e*RD1456ep + 5./432.*m11*m33*RS145pe*RD1456ep + 5./
         432.*m11*m33*RS146pe*RD1456ep + 5./432.*m11*m33*RS156pe*
         RD1456ep - 1./216.*pow(m11,2)*RD1456ep*RS34561pe - 1./54.*pow(
         m11,2)*RD1456ep*RS14563pe - 1./216.*pow(m11,2)*RD1456ep*
         RSe45613p - 1./216.*pow(m11,2)*RD1456ep*RSp45613e - 1./432.*
         pow(m11,2)*RD1456ep*RS3456pee - 1./432.*pow(m11,2)*RD1456ep*
         RS3456ppe + 1./432.*pow(m11,2)*RD1456ep*RSe45633p + 1./432.*
         pow(m11,2)*RD1456ep*RSp45633e + 1./432.*pow(m11,2)*RD1456ep*
         RS345pe + 1./432.*pow(m11,2)*RD1456ep*RS346pe + 1./432.*pow(
         m11,2)*RD1456ep*RS356pe - 1./108.*pow(m11,2)*RS1456pee*
         RD1456ep );

      M13p6x23R +=  + Bb(m11,mu2)*pi16 * (  - 1./108.*pow(m11,2)*
         RS1456ppe*RD1456ep + 1./432.*pow(m11,2)*RSe45611p*RD1456ep + 1.
         /432.*pow(m11,2)*RSp45611e*RD1456ep + 1./108.*pow(m11,2)*
         RS145pe*RD1456ep + 1./108.*pow(m11,2)*RS146pe*RD1456ep + 1./
         108.*pow(m11,2)*RS156pe*RD1456ep );

      M13p6x23R +=  + pow(Bb(m11,mu2),2) * ( 1./144.*m33*pow(
         RD1456ep,2) - 5./432.*m11*pow(RD1456ep,2) + 1./27.*m11*m33*
         RD1456ep*RS14563pe + 2./27.*pow(m11,2)*RD1456ep*RS14563pe );

      M13p6x23R +=  + Bb(m11,mu2)*Bb(m33,mu2) * (  - 1./72.*m33*
         RD1456ep*RD3456ep - 1./72.*m11*RD1456ep*RD3456ep );

      M13p6x23R +=  + Bb(m33,mu2)*pi16 * ( 1./144.*m33*pow(RD3456ep,2)
          + 1./144.*m33*RD1456ep*RD3456ep - 1./72.*m33*mp2*RD3456ep*
         RSp45613e + 1./144.*m33*mp2*RSp45633e*RD3456ep + 1./144.*m33*
         mp2*RSp45611e*RD3456ep - 1./72.*m33*me2*RD3456ep*RSe45613p + 1.
         /144.*m33*me2*RSe45633p*RD3456ep + 1./144.*m33*me2*RSe45611p*
         RD3456ep - 1./48.*m33*m66*RD3456ep - 1./48.*m33*m55*RD3456ep
          - 1./48.*m33*m44*RD3456ep - 1./54.*pow(m33,2)*RD3456ep*
         RS34561pe - 1./216.*pow(m33,2)*RD3456ep*RS14563pe - 1./216.*
         pow(m33,2)*RD3456ep*RSe45613p - 1./216.*pow(m33,2)*RD3456ep*
         RSp45613e - 1./108.*pow(m33,2)*RS3456pee*RD3456ep - 1./108.*
         pow(m33,2)*RS3456ppe*RD3456ep + 1./432.*pow(m33,2)*RSe45633p*
         RD3456ep + 1./432.*pow(m33,2)*RSp45633e*RD3456ep + 1./108.*
         pow(m33,2)*RS345pe*RD3456ep + 1./108.*pow(m33,2)*RS346pe*
         RD3456ep + 1./108.*pow(m33,2)*RS356pe*RD3456ep - 1./432.*pow(
         m33,2)*RS1456pee*RD3456ep - 1./432.*pow(m33,2)*RS1456ppe*
         RD3456ep );

      M13p6x23R +=  + Bb(m33,mu2)*pi16 * ( 1./432.*pow(m33,2)*RSe45611p
         *RD3456ep + 1./432.*pow(m33,2)*RSp45611e*RD3456ep + 1./432.*
         pow(m33,2)*RS145pe*RD3456ep + 1./432.*pow(m33,2)*RS146pe*
         RD3456ep + 1./432.*pow(m33,2)*RS156pe*RD3456ep + 1./144.*m11*
         pow(RD3456ep,2) + 1./144.*m11*RD1456ep*RD3456ep - 1./72.*m11*
         mp2*RD3456ep*RSp45613e + 1./144.*m11*mp2*RSp45633e*RD3456ep + 
         1./144.*m11*mp2*RSp45611e*RD3456ep - 1./72.*m11*me2*RD3456ep*
         RSe45613p + 1./144.*m11*me2*RSe45633p*RD3456ep + 1./144.*m11*
         me2*RSe45611p*RD3456ep - 1./48.*m11*m66*RD3456ep - 1./48.*m11*
         m55*RD3456ep - 1./48.*m11*m44*RD3456ep - 1./48.*m11*m33*
         RD3456ep - 5./216.*m11*m33*RD3456ep*RS34561pe - 5./216.*m11*
         m33*RD3456ep*RS14563pe - 1./108.*m11*m33*RD3456ep*RSe45613p - 
         1./108.*m11*m33*RD3456ep*RSp45613e - 5./432.*m11*m33*RS3456pee
         *RD3456ep - 5./432.*m11*m33*RS3456ppe*RD3456ep + 1./216.*m11*
         m33*RSe45633p*RD3456ep + 1./216.*m11*m33*RSp45633e*RD3456ep + 
         5./432.*m11*m33*RS345pe*RD3456ep );

      M13p6x23R +=  + Bb(m33,mu2)*pi16 * ( 5./432.*m11*m33*RS346pe*
         RD3456ep + 5./432.*m11*m33*RS356pe*RD3456ep - 5./432.*m11*m33*
         RS1456pee*RD3456ep - 5./432.*m11*m33*RS1456ppe*RD3456ep + 1./
         216.*m11*m33*RSe45611p*RD3456ep + 1./216.*m11*m33*RSp45611e*
         RD3456ep + 5./432.*m11*m33*RS145pe*RD3456ep + 5./432.*m11*m33*
         RS146pe*RD3456ep + 5./432.*m11*m33*RS156pe*RD3456ep - 1./48.*
         pow(m11,2)*RD3456ep - 1./216.*pow(m11,2)*RD3456ep*RS34561pe - 
         1./54.*pow(m11,2)*RD3456ep*RS14563pe - 1./216.*pow(m11,2)*
         RD3456ep*RSe45613p - 1./216.*pow(m11,2)*RD3456ep*RSp45613e - 1.
         /432.*pow(m11,2)*RS3456pee*RD3456ep - 1./432.*pow(m11,2)*
         RS3456ppe*RD3456ep + 1./432.*pow(m11,2)*RSe45633p*RD3456ep + 1.
         /432.*pow(m11,2)*RSp45633e*RD3456ep + 1./432.*pow(m11,2)*
         RS345pe*RD3456ep + 1./432.*pow(m11,2)*RS346pe*RD3456ep + 1./
         432.*pow(m11,2)*RS356pe*RD3456ep - 1./108.*pow(m11,2)*
         RS1456pee*RD3456ep - 1./108.*pow(m11,2)*RS1456ppe*RD3456ep + 1.
         /432.*pow(m11,2)*RSe45611p*RD3456ep );

      M13p6x23R +=  + Bb(m33,mu2)*pi16 * ( 1./432.*pow(m11,2)*RSp45611e
         *RD3456ep + 1./108.*pow(m11,2)*RS145pe*RD3456ep + 1./108.*pow(
         m11,2)*RS146pe*RD3456ep + 1./108.*pow(m11,2)*RS156pe*RD3456ep
          );

      M13p6x23R +=  + pow(Bb(m33,mu2),2) * (  - 5./432.*m33*pow(
         RD3456ep,2) + 2./27.*pow(m33,2)*RD3456ep*RS34561pe + 1./144.*
         m11*pow(RD3456ep,2) + 1./27.*m11*m33*RD3456ep*RS34561pe );

      M13p6x23R +=  + Abeps(m11,mu2)*pi16 * ( 1./540.*RD1456ep*RD3456ep
          + 1./540.*pow(RD1456ep,2) - 1./270.*mp2*RD1456ep*RSp45613e + 
         1./540.*mp2*RD1456ep*RSp45633e + 1./540.*mp2*RSp45611e*
         RD1456ep - 2./405.*pow(mp2,2)*RSp45613e*RS14563pe + 1./405.*
         pow(mp2,2)*RSp45633e*RS14563pe - 1./405.*pow(mp2,2)*RS1456pee*
         RSp45613e + 1./810.*pow(mp2,2)*RS1456pee*RSp45633e - 1./405.*
         pow(mp2,2)*RS1456ppe*RSp45613e + 1./810.*pow(mp2,2)*RS1456ppe*
         RSp45633e + 1./405.*pow(mp2,2)*RSp45611e*RS14563pe + 1./810.*
         pow(mp2,2)*RSp45611e*RS1456pee + 1./810.*pow(mp2,2)*RSp45611e*
         RS1456ppe + 1./405.*pow(mp2,2)*RS145pe*RSp45613e - 1./810.*
         pow(mp2,2)*RS145pe*RSp45633e - 1./810.*pow(mp2,2)*RS145pe*
         RSp45611e + 1./405.*pow(mp2,2)*RS146pe*RSp45613e - 1./810.*
         pow(mp2,2)*RS146pe*RSp45633e - 1./810.*pow(mp2,2)*RS146pe*
         RSp45611e + 1./405.*pow(mp2,2)*RS156pe*RSp45613e - 1./810.*
         pow(mp2,2)*RS156pe*RSp45633e - 1./810.*pow(mp2,2)*RS156pe*
         RSp45611e );

      M13p6x23R +=  + Abeps(m11,mu2)*pi16 * (  - 1./270.*me2*RD1456ep*
         RSe45613p + 1./540.*me2*RD1456ep*RSe45633p + 1./540.*me2*
         RSe45611p*RD1456ep - 2./405.*pow(me2,2)*RSe45613p*RS14563pe + 
         1./405.*pow(me2,2)*RSe45633p*RS14563pe - 1./405.*pow(me2,2)*
         RS1456pee*RSe45613p + 1./810.*pow(me2,2)*RS1456pee*RSe45633p
          - 1./405.*pow(me2,2)*RS1456ppe*RSe45613p + 1./810.*pow(me2,2)
         *RS1456ppe*RSe45633p + 1./405.*pow(me2,2)*RSe45611p*RS14563pe
          + 1./810.*pow(me2,2)*RSe45611p*RS1456pee + 1./810.*pow(me2,2)
         *RSe45611p*RS1456ppe + 1./405.*pow(me2,2)*RS145pe*RSe45613p - 
         1./810.*pow(me2,2)*RS145pe*RSe45633p - 1./810.*pow(me2,2)*
         RS145pe*RSe45611p + 1./405.*pow(me2,2)*RS146pe*RSe45613p - 1./
         810.*pow(me2,2)*RS146pe*RSe45633p - 1./810.*pow(me2,2)*RS146pe
         *RSe45611p + 1./405.*pow(me2,2)*RS156pe*RSe45613p - 1./810.*
         pow(me2,2)*RS156pe*RSe45633p - 1./810.*pow(me2,2)*RS156pe*
         RSe45611p + 7./270.*m66*RD1456ep - 1./81.*m66*mp2*RSp456e*
         RS145pe*RSe453p );

      M13p6x23R +=  + Abeps(m11,mu2)*pi16 * (  - 1./81.*m66*mp2*RSp456e
         *RS145pe*RSp453e + 1./162.*m66*mp2*RSp456e*pow(RS145pe,2) + 1./
         81.*m66*mp2*RSp456e*RSe451p*RS145pe + 1./81.*m66*mp2*RSp456e*
         RSp451e*RS145pe - 1./81.*m66*me2*RSe456p*RS145pe*RSe453p - 1./
         81.*m66*me2*RSe456p*RS145pe*RSp453e + 1./162.*m66*me2*RSe456p*
         pow(RS145pe,2) + 1./81.*m66*me2*RSe456p*RSe451p*RS145pe + 1./
         81.*m66*me2*RSe456p*RSp451e*RS145pe - 1./54.*pow(m66,2)*
         RSe453p - 1./54.*pow(m66,2)*RSp453e + 4./135.*pow(m66,2)*
         RS145pe + 1./162.*pow(m66,2)*RS145pe*RSe563p + 1./162.*pow(
         m66,2)*RS145pe*RSe463p - 1./243.*pow(m66,2)*RS145pe*RSe453p + 
         1./162.*pow(m66,2)*RS145pe*RSp563e + 1./162.*pow(m66,2)*
         RS145pe*RSp463e - 1./243.*pow(m66,2)*RS145pe*RSp453e + 1./486.
         *pow(m66,2)*pow(RS145pe,2) + 1./162.*pow(m66,2)*RS146pe*
         RSe453p + 1./162.*pow(m66,2)*RS146pe*RSp453e - 1./162.*pow(
         m66,2)*RS146pe*RS145pe + 1./162.*pow(m66,2)*RS156pe*RSe453p + 
         1./162.*pow(m66,2)*RS156pe*RSp453e );

      M13p6x23R +=  + Abeps(m11,mu2)*pi16 * (  - 1./162.*pow(m66,2)*
         RS156pe*RS145pe - 1./81.*pow(m66,2)*RS645pe*RS145pe*RSe453p - 
         1./81.*pow(m66,2)*RS645pe*RS145pe*RSp453e + 1./162.*pow(m66,2)
         *RS645pe*pow(RS145pe,2) - 1./162.*pow(m66,2)*RSe561p*RS145pe
          - 1./162.*pow(m66,2)*RSe461p*RS145pe + 1./54.*pow(m66,2)*
         RSe451p + 1./243.*pow(m66,2)*RSe451p*RS145pe - 1./162.*pow(
         m66,2)*RSe451p*RS146pe - 1./162.*pow(m66,2)*RSe451p*RS156pe + 
         1./81.*pow(m66,2)*RSe451p*RS645pe*RS145pe - 1./162.*pow(m66,2)
         *RSp561e*RS145pe - 1./162.*pow(m66,2)*RSp461e*RS145pe + 1./54.
         *pow(m66,2)*RSp451e + 1./243.*pow(m66,2)*RSp451e*RS145pe - 1./
         162.*pow(m66,2)*RSp451e*RS146pe - 1./162.*pow(m66,2)*RSp451e*
         RS156pe + 1./81.*pow(m66,2)*RSp451e*RS645pe*RS145pe + 7./270.*
         m55*RD1456ep - 1./81.*m55*mp2*RSp465e*RS146pe*RSe463p - 1./81.
         *m55*mp2*RSp465e*RS146pe*RSp463e + 1./162.*m55*mp2*RSp465e*
         pow(RS146pe,2) + 1./81.*m55*mp2*RSp465e*RSe461p*RS146pe + 1./
         81.*m55*mp2*RSp465e*RSp461e*RS146pe );

      M13p6x23R +=  + Abeps(m11,mu2)*pi16 * (  - 1./81.*m55*me2*RSe465p
         *RS146pe*RSe463p - 1./81.*m55*me2*RSe465p*RS146pe*RSp463e + 1./
         162.*m55*me2*RSe465p*pow(RS146pe,2) + 1./81.*m55*me2*RSe465p*
         RSe461p*RS146pe + 1./81.*m55*me2*RSe465p*RSp461e*RS146pe + 1./
         81.*m55*m66*RS145pe*RSe463p - 4./243.*m55*m66*RS145pe*RSe453p
          + 1./81.*m55*m66*RS145pe*RSp463e - 4./243.*m55*m66*RS145pe*
         RSp453e + 2./243.*m55*m66*pow(RS145pe,2) - 4./243.*m55*m66*
         RS146pe*RSe463p + 1./81.*m55*m66*RS146pe*RSe453p - 4./243.*m55
         *m66*RS146pe*RSp463e + 1./81.*m55*m66*RS146pe*RSp453e - 1./81.
         *m55*m66*RS146pe*RS145pe + 2./243.*m55*m66*pow(RS146pe,2) - 1./
         81.*m55*m66*RSe461p*RS145pe + 4./243.*m55*m66*RSe461p*RS146pe
          + 4./243.*m55*m66*RSe451p*RS145pe - 1./81.*m55*m66*RSe451p*
         RS146pe - 1./81.*m55*m66*RSp461e*RS145pe + 4./243.*m55*m66*
         RSp461e*RS146pe + 4./243.*m55*m66*RSp451e*RS145pe - 1./81.*m55
         *m66*RSp451e*RS146pe - 1./54.*pow(m55,2)*RSe463p - 1./54.*pow(
         m55,2)*RSp463e );

      M13p6x23R +=  + Abeps(m11,mu2)*pi16 * ( 1./162.*pow(m55,2)*
         RS145pe*RSe463p + 1./162.*pow(m55,2)*RS145pe*RSp463e + 4./135.
         *pow(m55,2)*RS146pe + 1./162.*pow(m55,2)*RS146pe*RSe563p - 1./
         243.*pow(m55,2)*RS146pe*RSe463p + 1./162.*pow(m55,2)*RS146pe*
         RSe453p + 1./162.*pow(m55,2)*RS146pe*RSp563e - 1./243.*pow(
         m55,2)*RS146pe*RSp463e + 1./162.*pow(m55,2)*RS146pe*RSp453e - 
         1./162.*pow(m55,2)*RS146pe*RS145pe + 1./486.*pow(m55,2)*pow(
         RS146pe,2) + 1./162.*pow(m55,2)*RS156pe*RSe463p + 1./162.*pow(
         m55,2)*RS156pe*RSp463e - 1./162.*pow(m55,2)*RS156pe*RS146pe - 
         1./81.*pow(m55,2)*RS546pe*RS146pe*RSe463p - 1./81.*pow(m55,2)*
         RS546pe*RS146pe*RSp463e + 1./162.*pow(m55,2)*RS546pe*pow(
         RS146pe,2) - 1./162.*pow(m55,2)*RSe561p*RS146pe + 1./54.*pow(
         m55,2)*RSe461p - 1./162.*pow(m55,2)*RSe461p*RS145pe + 1./243.*
         pow(m55,2)*RSe461p*RS146pe - 1./162.*pow(m55,2)*RSe461p*
         RS156pe + 1./81.*pow(m55,2)*RSe461p*RS546pe*RS146pe - 1./162.*
         pow(m55,2)*RSe451p*RS146pe );

      M13p6x23R +=  + Abeps(m11,mu2)*pi16 * (  - 1./162.*pow(m55,2)*
         RSp561e*RS146pe + 1./54.*pow(m55,2)*RSp461e - 1./162.*pow(
         m55,2)*RSp461e*RS145pe + 1./243.*pow(m55,2)*RSp461e*RS146pe - 
         1./162.*pow(m55,2)*RSp461e*RS156pe + 1./81.*pow(m55,2)*RSp461e
         *RS546pe*RS146pe - 1./162.*pow(m55,2)*RSp451e*RS146pe + 7./270.
         *m44*RD1456ep - 1./81.*m44*mp2*RSp564e*RS156pe*RSe563p - 1./81.
         *m44*mp2*RSp564e*RS156pe*RSp563e + 1./162.*m44*mp2*RSp564e*
         pow(RS156pe,2) + 1./81.*m44*mp2*RSp564e*RSe561p*RS156pe + 1./
         81.*m44*mp2*RSp564e*RSp561e*RS156pe - 1./81.*m44*me2*RSe564p*
         RS156pe*RSe563p - 1./81.*m44*me2*RSe564p*RS156pe*RSp563e + 1./
         162.*m44*me2*RSe564p*pow(RS156pe,2) + 1./81.*m44*me2*RSe564p*
         RSe561p*RS156pe + 1./81.*m44*me2*RSe564p*RSp561e*RS156pe + 1./
         81.*m44*m66*RS145pe*RSe563p - 4./243.*m44*m66*RS145pe*RSe453p
          + 1./81.*m44*m66*RS145pe*RSp563e - 4./243.*m44*m66*RS145pe*
         RSp453e + 2./243.*m44*m66*pow(RS145pe,2) - 4./243.*m44*m66*
         RS156pe*RSe563p );

      M13p6x23R +=  + Abeps(m11,mu2)*pi16 * ( 1./81.*m44*m66*RS156pe*
         RSe453p - 4./243.*m44*m66*RS156pe*RSp563e + 1./81.*m44*m66*
         RS156pe*RSp453e - 1./81.*m44*m66*RS156pe*RS145pe + 2./243.*m44
         *m66*pow(RS156pe,2) - 1./81.*m44*m66*RSe561p*RS145pe + 4./243.
         *m44*m66*RSe561p*RS156pe + 4./243.*m44*m66*RSe451p*RS145pe - 1.
         /81.*m44*m66*RSe451p*RS156pe - 1./81.*m44*m66*RSp561e*RS145pe
          + 4./243.*m44*m66*RSp561e*RS156pe + 4./243.*m44*m66*RSp451e*
         RS145pe - 1./81.*m44*m66*RSp451e*RS156pe + 1./81.*m44*m55*
         RS146pe*RSe563p - 4./243.*m44*m55*RS146pe*RSe463p + 1./81.*m44
         *m55*RS146pe*RSp563e - 4./243.*m44*m55*RS146pe*RSp463e + 2./
         243.*m44*m55*pow(RS146pe,2) - 4./243.*m44*m55*RS156pe*RSe563p
          + 1./81.*m44*m55*RS156pe*RSe463p - 4./243.*m44*m55*RS156pe*
         RSp563e + 1./81.*m44*m55*RS156pe*RSp463e - 1./81.*m44*m55*
         RS156pe*RS146pe + 2./243.*m44*m55*pow(RS156pe,2) - 1./81.*m44*
         m55*RSe561p*RS146pe + 4./243.*m44*m55*RSe561p*RS156pe + 4./243.
         *m44*m55*RSe461p*RS146pe );

      M13p6x23R +=  + Abeps(m11,mu2)*pi16 * (  - 1./81.*m44*m55*RSe461p
         *RS156pe - 1./81.*m44*m55*RSp561e*RS146pe + 4./243.*m44*m55*
         RSp561e*RS156pe + 4./243.*m44*m55*RSp461e*RS146pe - 1./81.*m44
         *m55*RSp461e*RS156pe - 1./54.*pow(m44,2)*RSe563p - 1./54.*pow(
         m44,2)*RSp563e + 1./162.*pow(m44,2)*RS145pe*RSe563p + 1./162.*
         pow(m44,2)*RS145pe*RSp563e + 1./162.*pow(m44,2)*RS146pe*
         RSe563p + 1./162.*pow(m44,2)*RS146pe*RSp563e + 4./135.*pow(
         m44,2)*RS156pe - 1./243.*pow(m44,2)*RS156pe*RSe563p + 1./162.*
         pow(m44,2)*RS156pe*RSe463p + 1./162.*pow(m44,2)*RS156pe*
         RSe453p - 1./243.*pow(m44,2)*RS156pe*RSp563e + 1./162.*pow(
         m44,2)*RS156pe*RSp463e + 1./162.*pow(m44,2)*RS156pe*RSp453e - 
         1./162.*pow(m44,2)*RS156pe*RS145pe - 1./162.*pow(m44,2)*
         RS156pe*RS146pe + 1./486.*pow(m44,2)*pow(RS156pe,2) - 1./81.*
         pow(m44,2)*RS456pe*RS156pe*RSe563p - 1./81.*pow(m44,2)*RS456pe
         *RS156pe*RSp563e + 1./162.*pow(m44,2)*RS456pe*pow(RS156pe,2)
          + 1./54.*pow(m44,2)*RSe561p );

      M13p6x23R +=  + Abeps(m11,mu2)*pi16 * (  - 1./162.*pow(m44,2)*
         RSe561p*RS145pe - 1./162.*pow(m44,2)*RSe561p*RS146pe + 1./243.
         *pow(m44,2)*RSe561p*RS156pe + 1./81.*pow(m44,2)*RSe561p*
         RS456pe*RS156pe - 1./162.*pow(m44,2)*RSe461p*RS156pe - 1./162.
         *pow(m44,2)*RSe451p*RS156pe + 1./54.*pow(m44,2)*RSp561e - 1./
         162.*pow(m44,2)*RSp561e*RS145pe - 1./162.*pow(m44,2)*RSp561e*
         RS146pe + 1./243.*pow(m44,2)*RSp561e*RS156pe + 1./81.*pow(
         m44,2)*RSp561e*RS456pe*RS156pe - 1./162.*pow(m44,2)*RSp461e*
         RS156pe - 1./162.*pow(m44,2)*RSp451e*RS156pe - 7./405.*m33*
         RD3456ep*RS14563pe + 1./30.*m33*RD1456ep - 7./540.*m33*
         RD1456ep*RS34561pe - 17./1620.*m33*RD1456ep*RS14563pe - 1./108.
         *m33*RD1456ep*RSe45613p - 1./108.*m33*RD1456ep*RSp45613e - 7./
         1080.*m33*RD1456ep*RS3456pee - 7./1080.*m33*RD1456ep*RS3456ppe
          + 1./216.*m33*RD1456ep*RSe45633p + 1./216.*m33*RD1456ep*
         RSp45633e + 7./1080.*m33*RD1456ep*RS345pe + 7./1080.*m33*
         RD1456ep*RS346pe );

      M13p6x23R +=  + Abeps(m11,mu2)*pi16 * ( 7./1080.*m33*RD1456ep*
         RS356pe + 11./810.*m33*RS1456pee*RD3456ep - 13./648.*m33*
         RS1456pee*RD1456ep + 11./810.*m33*RS1456ppe*RD3456ep - 13./648.
         *m33*RS1456ppe*RD1456ep + 1./216.*m33*RSe45611p*RD1456ep + 1./
         216.*m33*RSp45611e*RD1456ep - 11./810.*m33*RS145pe*RD3456ep + 
         13./648.*m33*RS145pe*RD1456ep - 11./810.*m33*RS146pe*RD3456ep
          + 13./648.*m33*RS146pe*RD1456ep - 11./810.*m33*RS156pe*
         RD3456ep + 13./648.*m33*RS156pe*RD1456ep + 17./405.*m33*mp2*
         RSp45613e*RS14563pe - 1./45.*m33*mp2*RSp45633e*RS14563pe - 2./
         405.*m33*mp2*RS1456pee*RSp45613e + 1./90.*m33*mp2*RS1456pee*
         RSp45633e - 2./405.*m33*mp2*RS1456ppe*RSp45613e + 1./90.*m33*
         mp2*RS1456ppe*RSp45633e - 1./810.*m33*mp2*RSp45611e*RS14563pe
          - 5./324.*m33*mp2*RSp45611e*RS1456pee - 5./324.*m33*mp2*
         RSp45611e*RS1456ppe + 2./405.*m33*mp2*RS145pe*RSp45613e - 1./
         90.*m33*mp2*RS145pe*RSp45633e + 5./324.*m33*mp2*RS145pe*
         RSp45611e );

      M13p6x23R +=  + Abeps(m11,mu2)*pi16 * ( 2./405.*m33*mp2*RS146pe*
         RSp45613e - 1./90.*m33*mp2*RS146pe*RSp45633e + 5./324.*m33*mp2
         *RS146pe*RSp45611e + 2./405.*m33*mp2*RS156pe*RSp45613e - 1./90.
         *m33*mp2*RS156pe*RSp45633e + 5./324.*m33*mp2*RS156pe*RSp45611e
          + 17./405.*m33*me2*RSe45613p*RS14563pe - 1./45.*m33*me2*
         RSe45633p*RS14563pe - 2./405.*m33*me2*RS1456pee*RSe45613p + 1./
         90.*m33*me2*RS1456pee*RSe45633p - 2./405.*m33*me2*RS1456ppe*
         RSe45613p + 1./90.*m33*me2*RS1456ppe*RSe45633p - 1./810.*m33*
         me2*RSe45611p*RS14563pe - 5./324.*m33*me2*RSe45611p*RS1456pee
          - 5./324.*m33*me2*RSe45611p*RS1456ppe + 2./405.*m33*me2*
         RS145pe*RSe45613p - 1./90.*m33*me2*RS145pe*RSe45633p + 5./324.
         *m33*me2*RS145pe*RSe45611p + 2./405.*m33*me2*RS146pe*RSe45613p
          - 1./90.*m33*me2*RS146pe*RSe45633p + 5./324.*m33*me2*RS146pe*
         RSe45611p + 2./405.*m33*me2*RS156pe*RSe45613p - 1./90.*m33*me2
         *RS156pe*RSe45633p + 5./324.*m33*me2*RS156pe*RSe45611p - 127./
         3240.*m33*m66*RS14563pe );

      M13p6x23R +=  + Abeps(m11,mu2)*pi16 * (  - 1./27.*m33*m66*
         RSe45613p - 1./27.*m33*m66*RSp45613e - 1./54.*m33*m66*RSe453p
          - 1./54.*m33*m66*RSp453e - 31./1296.*m33*m66*RS1456pee - 31./
         1296.*m33*m66*RS1456ppe + 239./6480.*m33*m66*RS145pe + 1./162.
         *m33*m66*RS145pe*RSe563p + 1./162.*m33*m66*RS145pe*RSe463p + 5.
         /81.*m33*m66*RS145pe*RSe453p + 1./162.*m33*m66*RS145pe*RSp563e
          + 1./162.*m33*m66*RS145pe*RSp463e + 5./81.*m33*m66*RS145pe*
         RSp453e + 31./1296.*m33*m66*RS146pe + 1./81.*m33*m66*RS146pe*
         RSe463p + 1./162.*m33*m66*RS146pe*RSe453p + 1./81.*m33*m66*
         RS146pe*RSp463e + 1./162.*m33*m66*RS146pe*RSp453e + 31./1296.*
         m33*m66*RS156pe + 1./81.*m33*m66*RS156pe*RSe563p + 1./162.*m33
         *m66*RS156pe*RSe453p + 1./81.*m33*m66*RS156pe*RSp563e + 1./162.
         *m33*m66*RS156pe*RSp453e - 127./3240.*m33*m55*RS14563pe - 1./
         27.*m33*m55*RSe45613p - 1./27.*m33*m55*RSp45613e - 1./54.*m33*
         m55*RSe463p - 1./54.*m33*m55*RSp463e - 31./1296.*m33*m55*
         RS1456pee );

      M13p6x23R +=  + Abeps(m11,mu2)*pi16 * (  - 31./1296.*m33*m55*
         RS1456ppe + 31./1296.*m33*m55*RS145pe + 1./162.*m33*m55*
         RS145pe*RSe463p + 1./81.*m33*m55*RS145pe*RSe453p + 1./162.*m33
         *m55*RS145pe*RSp463e + 1./81.*m33*m55*RS145pe*RSp453e + 239./
         6480.*m33*m55*RS146pe + 1./162.*m33*m55*RS146pe*RSe563p + 5./
         81.*m33*m55*RS146pe*RSe463p + 1./162.*m33*m55*RS146pe*RSe453p
          + 1./162.*m33*m55*RS146pe*RSp563e + 5./81.*m33*m55*RS146pe*
         RSp463e + 1./162.*m33*m55*RS146pe*RSp453e + 31./1296.*m33*m55*
         RS156pe + 1./81.*m33*m55*RS156pe*RSe563p + 1./162.*m33*m55*
         RS156pe*RSe463p + 1./81.*m33*m55*RS156pe*RSp563e + 1./162.*m33
         *m55*RS156pe*RSp463e - 127./3240.*m33*m44*RS14563pe - 1./27.*
         m33*m44*RSe45613p - 1./27.*m33*m44*RSp45613e - 1./54.*m33*m44*
         RSe563p - 1./54.*m33*m44*RSp563e - 31./1296.*m33*m44*RS1456pee
          - 31./1296.*m33*m44*RS1456ppe + 31./1296.*m33*m44*RS145pe + 1.
         /162.*m33*m44*RS145pe*RSe563p + 1./81.*m33*m44*RS145pe*RSe453p
          + 1./162.*m33*m44*RS145pe*RSp563e );

      M13p6x23R +=  + Abeps(m11,mu2)*pi16 * ( 1./81.*m33*m44*RS145pe*
         RSp453e + 31./1296.*m33*m44*RS146pe + 1./162.*m33*m44*RS146pe*
         RSe563p + 1./81.*m33*m44*RS146pe*RSe463p + 1./162.*m33*m44*
         RS146pe*RSp563e + 1./81.*m33*m44*RS146pe*RSp463e + 239./6480.*
         m33*m44*RS156pe + 5./81.*m33*m44*RS156pe*RSe563p + 1./162.*m33
         *m44*RS156pe*RSe463p + 1./162.*m33*m44*RS156pe*RSe453p + 5./81.
         *m33*m44*RS156pe*RSp563e + 1./162.*m33*m44*RS156pe*RSp463e + 1.
         /162.*m33*m44*RS156pe*RSp453e - 1./144.*pow(m33,2) + 1./720.*
         pow(m33,2)*RS14563pe + 1./81.*pow(m33,2)*RS14563pe*RS34561pe
          - 2./81.*pow(m33,2)*pow(RS14563pe,2) - 2./81.*pow(m33,2)*
         RSe45613p*RS14563pe - 2./81.*pow(m33,2)*RSp45613e*RS14563pe + 
         7./1620.*pow(m33,2)*RS3456pee*RS14563pe + 7./1620.*pow(m33,2)*
         RS3456ppe*RS14563pe + 5./324.*pow(m33,2)*RSe45633p*RS14563pe
          + 5./324.*pow(m33,2)*RSp45633e*RS14563pe - 7./1620.*pow(
         m33,2)*RS345pe*RS14563pe - 7./1620.*pow(m33,2)*RS346pe*
         RS14563pe );

      M13p6x23R +=  + Abeps(m11,mu2)*pi16 * (  - 7./1620.*pow(m33,2)*
         RS356pe*RS14563pe - 23./432.*pow(m33,2)*RS1456pee - 1./60.*
         pow(m33,2)*RS1456pee*RS34561pe - 1./54.*pow(m33,2)*RS1456pee*
         RS14563pe - 1./108.*pow(m33,2)*RS1456pee*RSe45613p - 1./108.*
         pow(m33,2)*RS1456pee*RSp45613e - 1./72.*pow(m33,2)*RS1456pee*
         RS3456pee - 1./72.*pow(m33,2)*RS1456pee*RS3456ppe + 1./648.*
         pow(m33,2)*RS1456pee*RSe45633p + 1./648.*pow(m33,2)*RS1456pee*
         RSp45633e + 1./72.*pow(m33,2)*RS1456pee*RS345pe + 1./72.*pow(
         m33,2)*RS1456pee*RS346pe + 1./72.*pow(m33,2)*RS1456pee*RS356pe
          - 5./648.*pow(m33,2)*pow(RS1456pee,2) - 23./432.*pow(m33,2)*
         RS1456ppe - 1./60.*pow(m33,2)*RS1456ppe*RS34561pe - 1./54.*
         pow(m33,2)*RS1456ppe*RS14563pe - 1./108.*pow(m33,2)*RS1456ppe*
         RSe45613p - 1./108.*pow(m33,2)*RS1456ppe*RSp45613e - 1./72.*
         pow(m33,2)*RS1456ppe*RS3456pee - 1./72.*pow(m33,2)*RS1456ppe*
         RS3456ppe + 1./648.*pow(m33,2)*RS1456ppe*RSe45633p + 1./648.*
         pow(m33,2)*RS1456ppe*RSp45633e );

      M13p6x23R +=  + Abeps(m11,mu2)*pi16 * ( 1./72.*pow(m33,2)*
         RS1456ppe*RS345pe + 1./72.*pow(m33,2)*RS1456ppe*RS346pe + 1./
         72.*pow(m33,2)*RS1456ppe*RS356pe - 5./324.*pow(m33,2)*
         RS1456ppe*RS1456pee - 5./648.*pow(m33,2)*pow(RS1456ppe,2) + 1./
         108.*pow(m33,2)*RSe45611p*RS14563pe + 5./648.*pow(m33,2)*
         RSe45611p*RS1456pee + 5./648.*pow(m33,2)*RSe45611p*RS1456ppe
          + 1./108.*pow(m33,2)*RSp45611e*RS14563pe + 5./648.*pow(m33,2)
         *RSp45611e*RS1456pee + 5./648.*pow(m33,2)*RSp45611e*RS1456ppe
          + 1./18.*pow(m33,2)*RS145pe + 1./60.*pow(m33,2)*RS145pe*
         RS34561pe + 1./54.*pow(m33,2)*RS145pe*RS14563pe + 1./108.*pow(
         m33,2)*RS145pe*RSe45613p + 1./108.*pow(m33,2)*RS145pe*
         RSp45613e + 1./72.*pow(m33,2)*RS145pe*RS3456pee + 1./72.*pow(
         m33,2)*RS145pe*RS3456ppe - 1./648.*pow(m33,2)*RS145pe*
         RSe45633p - 1./648.*pow(m33,2)*RS145pe*RSp45633e - 1./72.*pow(
         m33,2)*RS145pe*RS345pe - 1./72.*pow(m33,2)*RS145pe*RS346pe - 1.
         /72.*pow(m33,2)*RS145pe*RS356pe );

      M13p6x23R +=  + Abeps(m11,mu2)*pi16 * ( 5./324.*pow(m33,2)*
         RS145pe*RS1456pee + 5./324.*pow(m33,2)*RS145pe*RS1456ppe - 5./
         648.*pow(m33,2)*RS145pe*RSe45611p - 5./648.*pow(m33,2)*RS145pe
         *RSp45611e - 5./648.*pow(m33,2)*pow(RS145pe,2) + 1./18.*pow(
         m33,2)*RS146pe + 1./60.*pow(m33,2)*RS146pe*RS34561pe + 1./54.*
         pow(m33,2)*RS146pe*RS14563pe + 1./108.*pow(m33,2)*RS146pe*
         RSe45613p + 1./108.*pow(m33,2)*RS146pe*RSp45613e + 1./72.*pow(
         m33,2)*RS146pe*RS3456pee + 1./72.*pow(m33,2)*RS146pe*RS3456ppe
          - 1./648.*pow(m33,2)*RS146pe*RSe45633p - 1./648.*pow(m33,2)*
         RS146pe*RSp45633e - 1./72.*pow(m33,2)*RS146pe*RS345pe - 1./72.
         *pow(m33,2)*RS146pe*RS346pe - 1./72.*pow(m33,2)*RS146pe*
         RS356pe + 5./324.*pow(m33,2)*RS146pe*RS1456pee + 5./324.*pow(
         m33,2)*RS146pe*RS1456ppe - 5./648.*pow(m33,2)*RS146pe*
         RSe45611p - 5./648.*pow(m33,2)*RS146pe*RSp45611e - 5./324.*
         pow(m33,2)*RS146pe*RS145pe - 5./648.*pow(m33,2)*pow(RS146pe,2)
          + 1./18.*pow(m33,2)*RS156pe );

      M13p6x23R +=  + Abeps(m11,mu2)*pi16 * ( 1./60.*pow(m33,2)*RS156pe
         *RS34561pe + 1./54.*pow(m33,2)*RS156pe*RS14563pe + 1./108.*
         pow(m33,2)*RS156pe*RSe45613p + 1./108.*pow(m33,2)*RS156pe*
         RSp45613e + 1./72.*pow(m33,2)*RS156pe*RS3456pee + 1./72.*pow(
         m33,2)*RS156pe*RS3456ppe - 1./648.*pow(m33,2)*RS156pe*
         RSe45633p - 1./648.*pow(m33,2)*RS156pe*RSp45633e - 1./72.*pow(
         m33,2)*RS156pe*RS345pe - 1./72.*pow(m33,2)*RS156pe*RS346pe - 1.
         /72.*pow(m33,2)*RS156pe*RS356pe + 5./324.*pow(m33,2)*RS156pe*
         RS1456pee + 5./324.*pow(m33,2)*RS156pe*RS1456ppe - 5./648.*
         pow(m33,2)*RS156pe*RSe45611p - 5./648.*pow(m33,2)*RS156pe*
         RSp45611e - 5./324.*pow(m33,2)*RS156pe*RS145pe - 5./324.*pow(
         m33,2)*RS156pe*RS146pe - 5./648.*pow(m33,2)*pow(RS156pe,2) - 2.
         /405.*m11*RD3456ep*RS14563pe + 2./45.*m11*RD1456ep - 1./108.*
         m11*RD1456ep*RS34561pe + 1./324.*m11*RD1456ep*RS14563pe + 1./
         36.*m11*RD1456ep*RSe45613p + 1./36.*m11*RD1456ep*RSp45613e - 1.
         /216.*m11*RD1456ep*RS3456pee );

      M13p6x23R +=  + Abeps(m11,mu2)*pi16 * (  - 1./216.*m11*RD1456ep*
         RS3456ppe + 1./216.*m11*RD1456ep*RSe45633p + 1./216.*m11*
         RD1456ep*RSp45633e + 1./216.*m11*RD1456ep*RS345pe + 1./216.*
         m11*RD1456ep*RS346pe + 1./216.*m11*RD1456ep*RS356pe + 1./81.*
         m11*RS1456pee*RD3456ep + 53./3240.*m11*RS1456pee*RD1456ep + 1./
         81.*m11*RS1456ppe*RD3456ep + 53./3240.*m11*RS1456ppe*RD1456ep
          - 7./216.*m11*RSe45611p*RD1456ep - 7./216.*m11*RSp45611e*
         RD1456ep - 1./81.*m11*RS145pe*RD3456ep - 53./3240.*m11*RS145pe
         *RD1456ep - 1./81.*m11*RS146pe*RD3456ep - 53./3240.*m11*
         RS146pe*RD1456ep - 1./81.*m11*RS156pe*RD3456ep - 53./3240.*m11
         *RS156pe*RD1456ep + 4./81.*m11*mp2*RSp45613e*RS14563pe - 2./
         405.*m11*mp2*RSp45633e*RS14563pe - 19./810.*m11*mp2*RS1456pee*
         RSp45613e + 1./81.*m11*mp2*RS1456pee*RSp45633e - 19./810.*m11*
         mp2*RS1456ppe*RSp45613e + 1./81.*m11*mp2*RS1456ppe*RSp45633e
          + 1./90.*m11*mp2*RSp45611e*RS14563pe + 1./27.*m11*mp2*
         RSp45611e*RSe45613p );

      M13p6x23R +=  + Abeps(m11,mu2)*pi16 * ( 1./27.*m11*mp2*RSp45611e*
         RSp45613e + 11./540.*m11*mp2*RSp45611e*RS1456pee + 11./540.*
         m11*mp2*RSp45611e*RS1456ppe - 1./27.*m11*mp2*RSp45611e*
         RSe45611p - 1./27.*m11*mp2*pow(RSp45611e,2) + 19./810.*m11*mp2
         *RS145pe*RSp45613e - 1./81.*m11*mp2*RS145pe*RSp45633e - 11./
         540.*m11*mp2*RS145pe*RSp45611e + 19./810.*m11*mp2*RS146pe*
         RSp45613e - 1./81.*m11*mp2*RS146pe*RSp45633e - 11./540.*m11*
         mp2*RS146pe*RSp45611e + 19./810.*m11*mp2*RS156pe*RSp45613e - 1.
         /81.*m11*mp2*RS156pe*RSp45633e - 11./540.*m11*mp2*RS156pe*
         RSp45611e + 4./81.*m11*me2*RSe45613p*RS14563pe - 2./405.*m11*
         me2*RSe45633p*RS14563pe - 19./810.*m11*me2*RS1456pee*RSe45613p
          + 1./81.*m11*me2*RS1456pee*RSe45633p - 19./810.*m11*me2*
         RS1456ppe*RSe45613p + 1./81.*m11*me2*RS1456ppe*RSe45633p + 1./
         90.*m11*me2*RSe45611p*RS14563pe + 1./27.*m11*me2*RSe45611p*
         RSe45613p + 1./27.*m11*me2*RSe45611p*RSp45613e + 11./540.*m11*
         me2*RSe45611p*RS1456pee );

      M13p6x23R +=  + Abeps(m11,mu2)*pi16 * ( 11./540.*m11*me2*
         RSe45611p*RS1456ppe - 1./27.*m11*me2*pow(RSe45611p,2) - 1./27.
         *m11*me2*RSp45611e*RSe45611p + 19./810.*m11*me2*RS145pe*
         RSe45613p - 1./81.*m11*me2*RS145pe*RSe45633p - 11./540.*m11*
         me2*RS145pe*RSe45611p + 19./810.*m11*me2*RS146pe*RSe45613p - 1.
         /81.*m11*me2*RS146pe*RSe45633p - 11./540.*m11*me2*RS146pe*
         RSe45611p + 19./810.*m11*me2*RS156pe*RSe45613p - 1./81.*m11*
         me2*RS156pe*RSe45633p - 11./540.*m11*me2*RS156pe*RSe45611p - 1.
         /27.*m11*m66*RS34561pe - 61./1080.*m11*m66*RS14563pe - 2./81.*
         m11*m66*RSe45613p - 2./81.*m11*m66*RSp45613e - 1./54.*m11*m66*
         RS345pe - 2./27.*m11*m66*RSe453p - 2./27.*m11*m66*RSp453e - 23.
         /1296.*m11*m66*RS1456pee - 23./1296.*m11*m66*RS1456ppe - 4./81.
         *m11*m66*RSe45611p - 4./81.*m11*m66*RSp45611e - 89./6480.*m11*
         m66*RS145pe + 5./81.*m11*m66*RS145pe*RS345pe + 1./162.*m11*m66
         *RS145pe*RS346pe + 1./162.*m11*m66*RS145pe*RS356pe + 1./81.*
         m11*m66*RS145pe*RSe563p );

      M13p6x23R +=  + Abeps(m11,mu2)*pi16 * ( 1./81.*m11*m66*RS145pe*
         RSe463p + 10./81.*m11*m66*RS145pe*RSe453p + 1./81.*m11*m66*
         RS145pe*RSp563e + 1./81.*m11*m66*RS145pe*RSp463e + 10./81.*m11
         *m66*RS145pe*RSp453e - 5./162.*m11*m66*pow(RS145pe,2) + 23./
         1296.*m11*m66*RS146pe + 1./162.*m11*m66*RS146pe*RS345pe + 1./
         81.*m11*m66*RS146pe*RS346pe + 2./81.*m11*m66*RS146pe*RSe463p
          + 1./81.*m11*m66*RS146pe*RSe453p + 2./81.*m11*m66*RS146pe*
         RSp463e + 1./81.*m11*m66*RS146pe*RSp453e - 1./162.*m11*m66*
         RS146pe*RS145pe - 1./162.*m11*m66*pow(RS146pe,2) + 23./1296.*
         m11*m66*RS156pe + 1./162.*m11*m66*RS156pe*RS345pe + 1./81.*m11
         *m66*RS156pe*RS356pe + 2./81.*m11*m66*RS156pe*RSe563p + 1./81.
         *m11*m66*RS156pe*RSe453p + 2./81.*m11*m66*RS156pe*RSp563e + 1./
         81.*m11*m66*RS156pe*RSp453e - 1./162.*m11*m66*RS156pe*RS145pe
          - 1./162.*m11*m66*pow(RS156pe,2) + 1./27.*m11*m66*RSe451p + 1.
         /27.*m11*m66*RSp451e - 1./27.*m11*m55*RS34561pe - 61./1080.*
         m11*m55*RS14563pe );

      M13p6x23R +=  + Abeps(m11,mu2)*pi16 * (  - 2./81.*m11*m55*
         RSe45613p - 2./81.*m11*m55*RSp45613e - 1./54.*m11*m55*RS346pe
          - 2./27.*m11*m55*RSe463p - 2./27.*m11*m55*RSp463e - 23./1296.
         *m11*m55*RS1456pee - 23./1296.*m11*m55*RS1456ppe - 4./81.*m11*
         m55*RSe45611p - 4./81.*m11*m55*RSp45611e + 23./1296.*m11*m55*
         RS145pe + 1./81.*m11*m55*RS145pe*RS345pe + 1./162.*m11*m55*
         RS145pe*RS346pe + 1./81.*m11*m55*RS145pe*RSe463p + 2./81.*m11*
         m55*RS145pe*RSe453p + 1./81.*m11*m55*RS145pe*RSp463e + 2./81.*
         m11*m55*RS145pe*RSp453e - 1./162.*m11*m55*pow(RS145pe,2) - 89./
         6480.*m11*m55*RS146pe + 1./162.*m11*m55*RS146pe*RS345pe + 5./
         81.*m11*m55*RS146pe*RS346pe + 1./162.*m11*m55*RS146pe*RS356pe
          + 1./81.*m11*m55*RS146pe*RSe563p + 10./81.*m11*m55*RS146pe*
         RSe463p + 1./81.*m11*m55*RS146pe*RSe453p + 1./81.*m11*m55*
         RS146pe*RSp563e + 10./81.*m11*m55*RS146pe*RSp463e + 1./81.*m11
         *m55*RS146pe*RSp453e - 1./162.*m11*m55*RS146pe*RS145pe - 5./
         162.*m11*m55*pow(RS146pe,2) );

      M13p6x23R +=  + Abeps(m11,mu2)*pi16 * ( 23./1296.*m11*m55*RS156pe
          + 1./162.*m11*m55*RS156pe*RS346pe + 1./81.*m11*m55*RS156pe*
         RS356pe + 2./81.*m11*m55*RS156pe*RSe563p + 1./81.*m11*m55*
         RS156pe*RSe463p + 2./81.*m11*m55*RS156pe*RSp563e + 1./81.*m11*
         m55*RS156pe*RSp463e - 1./162.*m11*m55*RS156pe*RS146pe - 1./162.
         *m11*m55*pow(RS156pe,2) + 1./27.*m11*m55*RSe461p + 1./27.*m11*
         m55*RSp461e - 1./27.*m11*m44*RS34561pe - 61./1080.*m11*m44*
         RS14563pe - 2./81.*m11*m44*RSe45613p - 2./81.*m11*m44*
         RSp45613e - 1./54.*m11*m44*RS356pe - 2./27.*m11*m44*RSe563p - 
         2./27.*m11*m44*RSp563e - 23./1296.*m11*m44*RS1456pee - 23./
         1296.*m11*m44*RS1456ppe - 4./81.*m11*m44*RSe45611p - 4./81.*
         m11*m44*RSp45611e + 23./1296.*m11*m44*RS145pe + 1./81.*m11*m44
         *RS145pe*RS345pe + 1./162.*m11*m44*RS145pe*RS356pe + 1./81.*
         m11*m44*RS145pe*RSe563p + 2./81.*m11*m44*RS145pe*RSe453p + 1./
         81.*m11*m44*RS145pe*RSp563e + 2./81.*m11*m44*RS145pe*RSp453e
          - 1./162.*m11*m44*pow(RS145pe,2) );

      M13p6x23R +=  + Abeps(m11,mu2)*pi16 * ( 23./1296.*m11*m44*RS146pe
          + 1./81.*m11*m44*RS146pe*RS346pe + 1./162.*m11*m44*RS146pe*
         RS356pe + 1./81.*m11*m44*RS146pe*RSe563p + 2./81.*m11*m44*
         RS146pe*RSe463p + 1./81.*m11*m44*RS146pe*RSp563e + 2./81.*m11*
         m44*RS146pe*RSp463e - 1./162.*m11*m44*pow(RS146pe,2) - 89./
         6480.*m11*m44*RS156pe + 1./162.*m11*m44*RS156pe*RS345pe + 1./
         162.*m11*m44*RS156pe*RS346pe + 5./81.*m11*m44*RS156pe*RS356pe
          + 10./81.*m11*m44*RS156pe*RSe563p + 1./81.*m11*m44*RS156pe*
         RSe463p + 1./81.*m11*m44*RS156pe*RSe453p + 10./81.*m11*m44*
         RS156pe*RSp563e + 1./81.*m11*m44*RS156pe*RSp463e + 1./81.*m11*
         m44*RS156pe*RSp453e - 1./162.*m11*m44*RS156pe*RS145pe - 1./162.
         *m11*m44*RS156pe*RS146pe - 5./162.*m11*m44*pow(RS156pe,2) + 1./
         27.*m11*m44*RSe561p + 1./27.*m11*m44*RSp561e + 5./36.*m11*m33
          + 13./108.*m11*m33*RS14563pe + 1./54.*m11*m33*RS14563pe*
         RS34561pe + 1./90.*m11*m33*pow(RS14563pe,2) - 1./9.*m11*m33*
         RSe45613p );

      M13p6x23R +=  + Abeps(m11,mu2)*pi16 * (  - 5./162.*m11*m33*
         RSe45613p*RS14563pe - 1./9.*m11*m33*RSp45613e - 5./162.*m11*
         m33*RSp45613e*RS14563pe - 17./1620.*m11*m33*RS3456pee*
         RS14563pe - 17./1620.*m11*m33*RS3456ppe*RS14563pe + 5./324.*
         m11*m33*RSe45633p*RS14563pe + 5./324.*m11*m33*RSp45633e*
         RS14563pe + 17./1620.*m11*m33*RS345pe*RS14563pe + 17./1620.*
         m11*m33*RS346pe*RS14563pe + 17./1620.*m11*m33*RS356pe*
         RS14563pe - 1./54.*m11*m33*RSe563p - 1./54.*m11*m33*RSe463p - 
         1./54.*m11*m33*RSe453p - 1./54.*m11*m33*RSp563e - 1./54.*m11*
         m33*RSp463e - 1./54.*m11*m33*RSp453e - 25./432.*m11*m33*
         RS1456pee - 41./540.*m11*m33*RS1456pee*RS34561pe - 29./405.*
         m11*m33*RS1456pee*RS14563pe - 17./324.*m11*m33*RS1456pee*
         RSe45613p - 17./324.*m11*m33*RS1456pee*RSp45613e - 7./648.*m11
         *m33*RS1456pee*RS3456pee - 7./648.*m11*m33*RS1456pee*RS3456ppe
          - 1./648.*m11*m33*RS1456pee*RSe45633p - 1./648.*m11*m33*
         RS1456pee*RSp45633e );

      M13p6x23R +=  + Abeps(m11,mu2)*pi16 * ( 7./648.*m11*m33*RS1456pee
         *RS345pe + 7./648.*m11*m33*RS1456pee*RS346pe + 7./648.*m11*m33
         *RS1456pee*RS356pe - 1./648.*m11*m33*pow(RS1456pee,2) - 25./
         432.*m11*m33*RS1456ppe - 41./540.*m11*m33*RS1456ppe*RS34561pe
          - 29./405.*m11*m33*RS1456ppe*RS14563pe - 17./324.*m11*m33*
         RS1456ppe*RSe45613p - 17./324.*m11*m33*RS1456ppe*RSp45613e - 7.
         /648.*m11*m33*RS1456ppe*RS3456pee - 7./648.*m11*m33*RS1456ppe*
         RS3456ppe - 1./648.*m11*m33*RS1456ppe*RSe45633p - 1./648.*m11*
         m33*RS1456ppe*RSp45633e + 7./648.*m11*m33*RS1456ppe*RS345pe + 
         7./648.*m11*m33*RS1456ppe*RS346pe + 7./648.*m11*m33*RS1456ppe*
         RS356pe - 1./324.*m11*m33*RS1456ppe*RS1456pee - 1./648.*m11*
         m33*pow(RS1456ppe,2) + 5./324.*m11*m33*RSe45611p*RS14563pe + 
         11./648.*m11*m33*RSe45611p*RS1456pee + 11./648.*m11*m33*
         RSe45611p*RS1456ppe + 5./324.*m11*m33*RSp45611e*RS14563pe + 11.
         /648.*m11*m33*RSp45611e*RS1456pee + 11./648.*m11*m33*RSp45611e
         *RS1456ppe );

      M13p6x23R +=  + Abeps(m11,mu2)*pi16 * (  - 1./720.*m11*m33*
         RS145pe + 41./540.*m11*m33*RS145pe*RS34561pe + 29./405.*m11*
         m33*RS145pe*RS14563pe + 17./324.*m11*m33*RS145pe*RSe45613p + 
         17./324.*m11*m33*RS145pe*RSp45613e + 7./648.*m11*m33*RS145pe*
         RS3456pee + 7./648.*m11*m33*RS145pe*RS3456ppe + 1./648.*m11*
         m33*RS145pe*RSe45633p + 1./648.*m11*m33*RS145pe*RSp45633e - 7./
         648.*m11*m33*RS145pe*RS345pe - 7./648.*m11*m33*RS145pe*RS346pe
          - 7./648.*m11*m33*RS145pe*RS356pe + 1./324.*m11*m33*RS145pe*
         RS1456pee + 1./324.*m11*m33*RS145pe*RS1456ppe - 11./648.*m11*
         m33*RS145pe*RSe45611p - 11./648.*m11*m33*RS145pe*RSp45611e - 1.
         /648.*m11*m33*pow(RS145pe,2) - 1./720.*m11*m33*RS146pe + 41./
         540.*m11*m33*RS146pe*RS34561pe + 29./405.*m11*m33*RS146pe*
         RS14563pe + 17./324.*m11*m33*RS146pe*RSe45613p + 17./324.*m11*
         m33*RS146pe*RSp45613e + 7./648.*m11*m33*RS146pe*RS3456pee + 7./
         648.*m11*m33*RS146pe*RS3456ppe + 1./648.*m11*m33*RS146pe*
         RSe45633p );

      M13p6x23R +=  + Abeps(m11,mu2)*pi16 * ( 1./648.*m11*m33*RS146pe*
         RSp45633e - 7./648.*m11*m33*RS146pe*RS345pe - 7./648.*m11*m33*
         RS146pe*RS346pe - 7./648.*m11*m33*RS146pe*RS356pe + 1./324.*
         m11*m33*RS146pe*RS1456pee + 1./324.*m11*m33*RS146pe*RS1456ppe
          - 11./648.*m11*m33*RS146pe*RSe45611p - 11./648.*m11*m33*
         RS146pe*RSp45611e - 1./324.*m11*m33*RS146pe*RS145pe - 1./648.*
         m11*m33*pow(RS146pe,2) - 1./720.*m11*m33*RS156pe + 41./540.*
         m11*m33*RS156pe*RS34561pe + 29./405.*m11*m33*RS156pe*RS14563pe
          + 17./324.*m11*m33*RS156pe*RSe45613p + 17./324.*m11*m33*
         RS156pe*RSp45613e + 7./648.*m11*m33*RS156pe*RS3456pee + 7./648.
         *m11*m33*RS156pe*RS3456ppe + 1./648.*m11*m33*RS156pe*RSe45633p
          + 1./648.*m11*m33*RS156pe*RSp45633e - 7./648.*m11*m33*RS156pe
         *RS345pe - 7./648.*m11*m33*RS156pe*RS346pe - 7./648.*m11*m33*
         RS156pe*RS356pe + 1./324.*m11*m33*RS156pe*RS1456pee + 1./324.*
         m11*m33*RS156pe*RS1456ppe - 11./648.*m11*m33*RS156pe*RSe45611p
          - 11./648.*m11*m33*RS156pe*RSp45611e );

      M13p6x23R +=  + Abeps(m11,mu2)*pi16 * (  - 1./324.*m11*m33*
         RS156pe*RS145pe - 1./324.*m11*m33*RS156pe*RS146pe - 1./648.*
         m11*m33*pow(RS156pe,2) + 7./48.*pow(m11,2) - 1./9.*pow(m11,2)*
         RS34561pe - 223./2160.*pow(m11,2)*RS14563pe - 1./162.*pow(
         m11,2)*RS14563pe*RS34561pe + 31./810.*pow(m11,2)*pow(
         RS14563pe,2) - 5./27.*pow(m11,2)*RSe45613p - 1./162.*pow(
         m11,2)*RSe45613p*RS14563pe - 5./27.*pow(m11,2)*RSp45613e - 1./
         162.*pow(m11,2)*RSp45613e*RS14563pe - 1./54.*pow(m11,2)*
         RS345pe - 1./54.*pow(m11,2)*RS346pe - 1./54.*pow(m11,2)*
         RS356pe - 1./18.*pow(m11,2)*RSe563p - 1./18.*pow(m11,2)*
         RSe463p - 1./18.*pow(m11,2)*RSe453p - 1./18.*pow(m11,2)*
         RSp563e - 1./18.*pow(m11,2)*RSp463e - 1./18.*pow(m11,2)*
         RSp453e - 1./24.*pow(m11,2)*RS1456pee - 13./162.*pow(m11,2)*
         RS1456pee*RS34561pe - 17./135.*pow(m11,2)*RS1456pee*RS14563pe
          - 19./162.*pow(m11,2)*RS1456pee*RSe45613p - 19./162.*pow(
         m11,2)*RS1456pee*RSp45613e );

      M13p6x23R +=  + Abeps(m11,mu2)*pi16 * ( 1./324.*pow(m11,2)*
         RS1456pee*RS3456pee + 1./324.*pow(m11,2)*RS1456pee*RS3456ppe
          - 1./324.*pow(m11,2)*RS1456pee*RSe45633p - 1./324.*pow(m11,2)
         *RS1456pee*RSp45633e - 1./324.*pow(m11,2)*RS1456pee*RS345pe - 
         1./324.*pow(m11,2)*RS1456pee*RS346pe - 1./324.*pow(m11,2)*
         RS1456pee*RS356pe - 5./162.*pow(m11,2)*pow(RS1456pee,2) - 1./
         24.*pow(m11,2)*RS1456ppe - 13./162.*pow(m11,2)*RS1456ppe*
         RS34561pe - 17./135.*pow(m11,2)*RS1456ppe*RS14563pe - 19./162.
         *pow(m11,2)*RS1456ppe*RSe45613p - 19./162.*pow(m11,2)*
         RS1456ppe*RSp45613e + 1./324.*pow(m11,2)*RS1456ppe*RS3456pee
          + 1./324.*pow(m11,2)*RS1456ppe*RS3456ppe - 1./324.*pow(m11,2)
         *RS1456ppe*RSe45633p - 1./324.*pow(m11,2)*RS1456ppe*RSp45633e
          - 1./324.*pow(m11,2)*RS1456ppe*RS345pe - 1./324.*pow(m11,2)*
         RS1456ppe*RS346pe - 1./324.*pow(m11,2)*RS1456ppe*RS356pe - 5./
         81.*pow(m11,2)*RS1456ppe*RS1456pee - 5./162.*pow(m11,2)*pow(
         RS1456ppe,2) );

      M13p6x23R +=  + Abeps(m11,mu2)*pi16 * (  - 1./27.*pow(m11,2)*
         RSe45611p + 1./162.*pow(m11,2)*RSe45611p*RS14563pe + 5./108.*
         pow(m11,2)*RSe45611p*RS1456pee + 5./108.*pow(m11,2)*RSe45611p*
         RS1456ppe - 1./27.*pow(m11,2)*RSp45611e + 1./162.*pow(m11,2)*
         RSp45611e*RS14563pe + 5./108.*pow(m11,2)*RSp45611e*RS1456pee
          + 5./108.*pow(m11,2)*RSp45611e*RS1456ppe + 109./2160.*pow(
         m11,2)*RS145pe + 13./162.*pow(m11,2)*RS145pe*RS34561pe + 17./
         135.*pow(m11,2)*RS145pe*RS14563pe + 19./162.*pow(m11,2)*
         RS145pe*RSe45613p + 19./162.*pow(m11,2)*RS145pe*RSp45613e - 1./
         324.*pow(m11,2)*RS145pe*RS3456pee - 1./324.*pow(m11,2)*RS145pe
         *RS3456ppe + 1./324.*pow(m11,2)*RS145pe*RSe45633p + 1./324.*
         pow(m11,2)*RS145pe*RSp45633e + 1./324.*pow(m11,2)*RS145pe*
         RS345pe + 1./324.*pow(m11,2)*RS145pe*RS346pe + 1./324.*pow(
         m11,2)*RS145pe*RS356pe + 5./81.*pow(m11,2)*RS145pe*RS1456pee
          + 5./81.*pow(m11,2)*RS145pe*RS1456ppe - 5./108.*pow(m11,2)*
         RS145pe*RSe45611p );

      M13p6x23R +=  + Abeps(m11,mu2)*pi16 * (  - 5./108.*pow(m11,2)*
         RS145pe*RSp45611e - 5./162.*pow(m11,2)*pow(RS145pe,2) + 109./
         2160.*pow(m11,2)*RS146pe + 13./162.*pow(m11,2)*RS146pe*
         RS34561pe + 17./135.*pow(m11,2)*RS146pe*RS14563pe + 19./162.*
         pow(m11,2)*RS146pe*RSe45613p + 19./162.*pow(m11,2)*RS146pe*
         RSp45613e - 1./324.*pow(m11,2)*RS146pe*RS3456pee - 1./324.*
         pow(m11,2)*RS146pe*RS3456ppe + 1./324.*pow(m11,2)*RS146pe*
         RSe45633p + 1./324.*pow(m11,2)*RS146pe*RSp45633e + 1./324.*
         pow(m11,2)*RS146pe*RS345pe + 1./324.*pow(m11,2)*RS146pe*
         RS346pe + 1./324.*pow(m11,2)*RS146pe*RS356pe + 5./81.*pow(
         m11,2)*RS146pe*RS1456pee + 5./81.*pow(m11,2)*RS146pe*RS1456ppe
          - 5./108.*pow(m11,2)*RS146pe*RSe45611p - 5./108.*pow(m11,2)*
         RS146pe*RSp45611e - 5./81.*pow(m11,2)*RS146pe*RS145pe - 5./162.
         *pow(m11,2)*pow(RS146pe,2) + 109./2160.*pow(m11,2)*RS156pe + 
         13./162.*pow(m11,2)*RS156pe*RS34561pe + 17./135.*pow(m11,2)*
         RS156pe*RS14563pe );

      M13p6x23R +=  + Abeps(m11,mu2)*pi16 * ( 19./162.*pow(m11,2)*
         RS156pe*RSe45613p + 19./162.*pow(m11,2)*RS156pe*RSp45613e - 1./
         324.*pow(m11,2)*RS156pe*RS3456pee - 1./324.*pow(m11,2)*RS156pe
         *RS3456ppe + 1./324.*pow(m11,2)*RS156pe*RSe45633p + 1./324.*
         pow(m11,2)*RS156pe*RSp45633e + 1./324.*pow(m11,2)*RS156pe*
         RS345pe + 1./324.*pow(m11,2)*RS156pe*RS346pe + 1./324.*pow(
         m11,2)*RS156pe*RS356pe + 5./81.*pow(m11,2)*RS156pe*RS1456pee
          + 5./81.*pow(m11,2)*RS156pe*RS1456ppe - 5./108.*pow(m11,2)*
         RS156pe*RSe45611p - 5./108.*pow(m11,2)*RS156pe*RSp45611e - 5./
         81.*pow(m11,2)*RS156pe*RS145pe - 5./81.*pow(m11,2)*RS156pe*
         RS146pe - 5./162.*pow(m11,2)*pow(RS156pe,2) + 1./54.*pow(
         m11,2)*RSe561p + 1./54.*pow(m11,2)*RSe461p + 1./54.*pow(m11,2)
         *RSe451p + 1./54.*pow(m11,2)*RSp561e + 1./54.*pow(m11,2)*
         RSp461e + 1./54.*pow(m11,2)*RSp451e );

      M13p6x23R +=  + Abeps(m13,mu2)*pi16 * ( 1./324.*pow(RD3456ep,2)
          + 1./162.*RD1456ep*RD3456ep + 1./324.*pow(RD1456ep,2) - 1./81.
         *mp2*RD3456ep*RSp45613e + 1./162.*mp2*RSp45633e*RD3456ep - 1./
         81.*mp2*RD1456ep*RSp45613e + 1./162.*mp2*RD1456ep*RSp45633e + 
         1./162.*mp2*RSp45611e*RD3456ep + 1./162.*mp2*RSp45611e*
         RD1456ep + 1./27.*pow(mp2,2)*RSp45613e*RS34561pe + 1./27.*pow(
         mp2,2)*RSp45613e*RS14563pe + 1./27.*pow(mp2,2)*RSp45613e*
         RSe45613p + 4./81.*pow(mp2,2)*pow(RSp45613e,2) + 1./54.*pow(
         mp2,2)*RS3456pee*RSp45613e + 1./54.*pow(mp2,2)*RS3456ppe*
         RSp45613e - 1./54.*pow(mp2,2)*RSe45633p*RSp45613e - 1./54.*
         pow(mp2,2)*RSp45633e*RS34561pe - 1./54.*pow(mp2,2)*RSp45633e*
         RS14563pe - 1./54.*pow(mp2,2)*RSp45633e*RSe45613p - 4./81.*
         pow(mp2,2)*RSp45633e*RSp45613e - 1./108.*pow(mp2,2)*RSp45633e*
         RS3456pee - 1./108.*pow(mp2,2)*RSp45633e*RS3456ppe + 1./108.*
         pow(mp2,2)*RSp45633e*RSe45633p + 1./81.*pow(mp2,2)*pow(
         RSp45633e,2) );

      M13p6x23R +=  + Abeps(m13,mu2)*pi16 * (  - 1./54.*pow(mp2,2)*
         RS345pe*RSp45613e + 1./108.*pow(mp2,2)*RS345pe*RSp45633e - 1./
         54.*pow(mp2,2)*RS346pe*RSp45613e + 1./108.*pow(mp2,2)*RS346pe*
         RSp45633e - 1./54.*pow(mp2,2)*RS356pe*RSp45613e + 1./108.*pow(
         mp2,2)*RS356pe*RSp45633e + 1./54.*pow(mp2,2)*RS1456pee*
         RSp45613e - 1./108.*pow(mp2,2)*RS1456pee*RSp45633e + 1./54.*
         pow(mp2,2)*RS1456ppe*RSp45613e - 1./108.*pow(mp2,2)*RS1456ppe*
         RSp45633e - 1./54.*pow(mp2,2)*RSe45611p*RSp45613e + 1./108.*
         pow(mp2,2)*RSe45611p*RSp45633e - 1./54.*pow(mp2,2)*RSp45611e*
         RS34561pe - 1./54.*pow(mp2,2)*RSp45611e*RS14563pe - 1./54.*
         pow(mp2,2)*RSp45611e*RSe45613p - 4./81.*pow(mp2,2)*RSp45611e*
         RSp45613e - 1./108.*pow(mp2,2)*RSp45611e*RS3456pee - 1./108.*
         pow(mp2,2)*RSp45611e*RS3456ppe + 1./108.*pow(mp2,2)*RSp45611e*
         RSe45633p + 2./81.*pow(mp2,2)*RSp45611e*RSp45633e + 1./108.*
         pow(mp2,2)*RSp45611e*RS345pe + 1./108.*pow(mp2,2)*RSp45611e*
         RS346pe );

      M13p6x23R +=  + Abeps(m13,mu2)*pi16 * ( 1./108.*pow(mp2,2)*
         RSp45611e*RS356pe - 1./108.*pow(mp2,2)*RSp45611e*RS1456pee - 1.
         /108.*pow(mp2,2)*RSp45611e*RS1456ppe + 1./108.*pow(mp2,2)*
         RSp45611e*RSe45611p + 1./81.*pow(mp2,2)*pow(RSp45611e,2) - 1./
         54.*pow(mp2,2)*RS145pe*RSp45613e + 1./108.*pow(mp2,2)*RS145pe*
         RSp45633e + 1./108.*pow(mp2,2)*RS145pe*RSp45611e - 1./54.*pow(
         mp2,2)*RS146pe*RSp45613e + 1./108.*pow(mp2,2)*RS146pe*
         RSp45633e + 1./108.*pow(mp2,2)*RS146pe*RSp45611e - 1./54.*pow(
         mp2,2)*RS156pe*RSp45613e + 1./108.*pow(mp2,2)*RS156pe*
         RSp45633e + 1./108.*pow(mp2,2)*RS156pe*RSp45611e - 1./81.*me2*
         RD3456ep*RSe45613p + 1./162.*me2*RSe45633p*RD3456ep - 1./81.*
         me2*RD1456ep*RSe45613p + 1./162.*me2*RD1456ep*RSe45633p + 1./
         162.*me2*RSe45611p*RD3456ep + 1./162.*me2*RSe45611p*RD1456ep
          + 2./81.*me2*mp2*RSp45613e*RSe45613p - 1./81.*me2*mp2*
         RSe45633p*RSp45613e - 1./81.*me2*mp2*RSp45633e*RSe45613p + 1./
         162.*me2*mp2*RSp45633e*RSe45633p );

      M13p6x23R +=  + Abeps(m13,mu2)*pi16 * (  - 1./81.*me2*mp2*
         RSe45611p*RSp45613e + 1./162.*me2*mp2*RSe45611p*RSp45633e - 1./
         81.*me2*mp2*RSp45611e*RSe45613p + 1./162.*me2*mp2*RSp45611e*
         RSe45633p + 1./162.*me2*mp2*RSp45611e*RSe45611p + 1./27.*pow(
         me2,2)*RSe45613p*RS34561pe + 1./27.*pow(me2,2)*RSe45613p*
         RS14563pe + 4./81.*pow(me2,2)*pow(RSe45613p,2) + 1./27.*pow(
         me2,2)*RSp45613e*RSe45613p + 1./54.*pow(me2,2)*RS3456pee*
         RSe45613p + 1./54.*pow(me2,2)*RS3456ppe*RSe45613p - 1./54.*
         pow(me2,2)*RSe45633p*RS34561pe - 1./54.*pow(me2,2)*RSe45633p*
         RS14563pe - 4./81.*pow(me2,2)*RSe45633p*RSe45613p - 1./54.*
         pow(me2,2)*RSe45633p*RSp45613e - 1./108.*pow(me2,2)*RSe45633p*
         RS3456pee - 1./108.*pow(me2,2)*RSe45633p*RS3456ppe + 1./81.*
         pow(me2,2)*pow(RSe45633p,2) - 1./54.*pow(me2,2)*RSp45633e*
         RSe45613p + 1./108.*pow(me2,2)*RSp45633e*RSe45633p - 1./54.*
         pow(me2,2)*RS345pe*RSe45613p + 1./108.*pow(me2,2)*RS345pe*
         RSe45633p );

      M13p6x23R +=  + Abeps(m13,mu2)*pi16 * (  - 1./54.*pow(me2,2)*
         RS346pe*RSe45613p + 1./108.*pow(me2,2)*RS346pe*RSe45633p - 1./
         54.*pow(me2,2)*RS356pe*RSe45613p + 1./108.*pow(me2,2)*RS356pe*
         RSe45633p + 1./54.*pow(me2,2)*RS1456pee*RSe45613p - 1./108.*
         pow(me2,2)*RS1456pee*RSe45633p + 1./54.*pow(me2,2)*RS1456ppe*
         RSe45613p - 1./108.*pow(me2,2)*RS1456ppe*RSe45633p - 1./54.*
         pow(me2,2)*RSe45611p*RS34561pe - 1./54.*pow(me2,2)*RSe45611p*
         RS14563pe - 4./81.*pow(me2,2)*RSe45611p*RSe45613p - 1./54.*
         pow(me2,2)*RSe45611p*RSp45613e - 1./108.*pow(me2,2)*RSe45611p*
         RS3456pee - 1./108.*pow(me2,2)*RSe45611p*RS3456ppe + 2./81.*
         pow(me2,2)*RSe45611p*RSe45633p + 1./108.*pow(me2,2)*RSe45611p*
         RSp45633e + 1./108.*pow(me2,2)*RSe45611p*RS345pe + 1./108.*
         pow(me2,2)*RSe45611p*RS346pe + 1./108.*pow(me2,2)*RSe45611p*
         RS356pe - 1./108.*pow(me2,2)*RSe45611p*RS1456pee - 1./108.*
         pow(me2,2)*RSe45611p*RS1456ppe + 1./81.*pow(me2,2)*pow(
         RSe45611p,2) );

      M13p6x23R +=  + Abeps(m13,mu2)*pi16 * (  - 1./54.*pow(me2,2)*
         RSp45611e*RSe45613p + 1./108.*pow(me2,2)*RSp45611e*RSe45633p
          + 1./108.*pow(me2,2)*RSp45611e*RSe45611p - 1./54.*pow(me2,2)*
         RS145pe*RSe45613p + 1./108.*pow(me2,2)*RS145pe*RSe45633p + 1./
         108.*pow(me2,2)*RS145pe*RSe45611p - 1./54.*pow(me2,2)*RS146pe*
         RSe45613p + 1./108.*pow(me2,2)*RS146pe*RSe45633p + 1./108.*
         pow(me2,2)*RS146pe*RSe45611p - 1./54.*pow(me2,2)*RS156pe*
         RSe45613p + 1./108.*pow(me2,2)*RS156pe*RSe45633p + 1./108.*
         pow(me2,2)*RS156pe*RSe45611p - 1./54.*m33*RD3456ep*RS34561pe
          - 1./162.*m33*RD3456ep*RS14563pe - 1./162.*m33*RD3456ep*
         RSe45613p - 1./162.*m33*RD3456ep*RSp45613e + 1./324.*m33*
         RS3456pee*RD3456ep + 1./324.*m33*RS3456ppe*RD3456ep - 1./108.*
         m33*RSe45633p*RD3456ep - 1./108.*m33*RSp45633e*RD3456ep - 1./
         324.*m33*RS345pe*RD3456ep - 1./324.*m33*RS346pe*RD3456ep - 1./
         324.*m33*RS356pe*RD3456ep + 1./162.*m33*RD1456ep*RS34561pe + 1.
         /54.*m33*RD1456ep*RS14563pe );

      M13p6x23R +=  + Abeps(m13,mu2)*pi16 * ( 1./54.*m33*RD1456ep*
         RSe45613p + 1./54.*m33*RD1456ep*RSp45613e - 1./324.*m33*
         RD1456ep*RS3456pee - 1./324.*m33*RD1456ep*RS3456ppe - 1./324.*
         m33*RD1456ep*RSe45633p - 1./324.*m33*RD1456ep*RSp45633e + 1./
         324.*m33*RD1456ep*RS345pe + 1./324.*m33*RD1456ep*RS346pe + 1./
         324.*m33*RD1456ep*RS356pe - 5./324.*m33*RS1456pee*RD3456ep + 5.
         /324.*m33*RS1456pee*RD1456ep - 5./324.*m33*RS1456ppe*RD3456ep
          + 5./324.*m33*RS1456ppe*RD1456ep + 5./324.*m33*RSe45611p*
         RD3456ep - 5./324.*m33*RSe45611p*RD1456ep + 5./324.*m33*
         RSp45611e*RD3456ep - 5./324.*m33*RSp45611e*RD1456ep + 5./324.*
         m33*RS145pe*RD3456ep - 5./324.*m33*RS145pe*RD1456ep + 5./324.*
         m33*RS146pe*RD3456ep - 5./324.*m33*RS146pe*RD1456ep + 5./324.*
         m33*RS156pe*RD3456ep - 5./324.*m33*RS156pe*RD1456ep - 2./81.*
         m33*mp2*RSp45613e*RS34561pe - 4./81.*m33*mp2*RSp45613e*
         RS14563pe - 4./81.*m33*mp2*RSp45613e*RSe45613p - 4./81.*m33*
         mp2*pow(RSp45613e,2) );

      M13p6x23R +=  + Abeps(m13,mu2)*pi16 * (  - 1./54.*m33*mp2*
         RS3456pee*RSp45613e - 1./54.*m33*mp2*RS3456ppe*RSp45613e + 5./
         162.*m33*mp2*RSe45633p*RSp45613e + 1./54.*m33*mp2*RSp45633e*
         RS34561pe + 5./162.*m33*mp2*RSp45633e*RS14563pe + 5./162.*m33*
         mp2*RSp45633e*RSe45613p + 5./81.*m33*mp2*RSp45633e*RSp45613e
          + 7./324.*m33*mp2*RSp45633e*RS3456pee + 7./324.*m33*mp2*
         RSp45633e*RS3456ppe - 1./36.*m33*mp2*RSp45633e*RSe45633p - 1./
         36.*m33*mp2*pow(RSp45633e,2) + 1./54.*m33*mp2*RS345pe*
         RSp45613e - 7./324.*m33*mp2*RS345pe*RSp45633e + 1./54.*m33*mp2
         *RS346pe*RSp45613e - 7./324.*m33*mp2*RS346pe*RSp45633e + 1./54.
         *m33*mp2*RS356pe*RSp45613e - 7./324.*m33*mp2*RS356pe*RSp45633e
          - 1./54.*m33*mp2*RS1456pee*RSp45613e + 1./324.*m33*mp2*
         RS1456pee*RSp45633e - 1./54.*m33*mp2*RS1456ppe*RSp45613e + 1./
         324.*m33*mp2*RS1456ppe*RSp45633e + 1./54.*m33*mp2*RSe45611p*
         RSp45613e - 1./324.*m33*mp2*RSe45611p*RSp45633e + 1./162.*m33*
         mp2*RSp45611e*RS34561pe );

      M13p6x23R +=  + Abeps(m13,mu2)*pi16 * ( 1./54.*m33*mp2*RSp45611e*
         RS14563pe + 1./54.*m33*mp2*RSp45611e*RSe45613p + 1./27.*m33*
         mp2*RSp45611e*RSp45613e - 1./324.*m33*mp2*RSp45611e*RS3456pee
          - 1./324.*m33*mp2*RSp45611e*RS3456ppe - 1./324.*m33*mp2*
         RSp45611e*RSe45633p - 1./162.*m33*mp2*RSp45611e*RSp45633e + 1./
         324.*m33*mp2*RSp45611e*RS345pe + 1./324.*m33*mp2*RSp45611e*
         RS346pe + 1./324.*m33*mp2*RSp45611e*RS356pe + 5./324.*m33*mp2*
         RSp45611e*RS1456pee + 5./324.*m33*mp2*RSp45611e*RS1456ppe - 5./
         324.*m33*mp2*RSp45611e*RSe45611p - 5./324.*m33*mp2*pow(
         RSp45611e,2) + 1./54.*m33*mp2*RS145pe*RSp45613e - 1./324.*m33*
         mp2*RS145pe*RSp45633e - 5./324.*m33*mp2*RS145pe*RSp45611e + 1./
         54.*m33*mp2*RS146pe*RSp45613e - 1./324.*m33*mp2*RS146pe*
         RSp45633e - 5./324.*m33*mp2*RS146pe*RSp45611e + 1./54.*m33*mp2
         *RS156pe*RSp45613e - 1./324.*m33*mp2*RS156pe*RSp45633e - 5./
         324.*m33*mp2*RS156pe*RSp45611e - 2./81.*m33*me2*RSe45613p*
         RS34561pe );

      M13p6x23R +=  + Abeps(m13,mu2)*pi16 * (  - 4./81.*m33*me2*
         RSe45613p*RS14563pe - 4./81.*m33*me2*pow(RSe45613p,2) - 4./81.
         *m33*me2*RSp45613e*RSe45613p - 1./54.*m33*me2*RS3456pee*
         RSe45613p - 1./54.*m33*me2*RS3456ppe*RSe45613p + 1./54.*m33*
         me2*RSe45633p*RS34561pe + 5./162.*m33*me2*RSe45633p*RS14563pe
          + 5./81.*m33*me2*RSe45633p*RSe45613p + 5./162.*m33*me2*
         RSe45633p*RSp45613e + 7./324.*m33*me2*RSe45633p*RS3456pee + 7./
         324.*m33*me2*RSe45633p*RS3456ppe - 1./36.*m33*me2*pow(
         RSe45633p,2) + 5./162.*m33*me2*RSp45633e*RSe45613p - 1./36.*
         m33*me2*RSp45633e*RSe45633p + 1./54.*m33*me2*RS345pe*RSe45613p
          - 7./324.*m33*me2*RS345pe*RSe45633p + 1./54.*m33*me2*RS346pe*
         RSe45613p - 7./324.*m33*me2*RS346pe*RSe45633p + 1./54.*m33*me2
         *RS356pe*RSe45613p - 7./324.*m33*me2*RS356pe*RSe45633p - 1./54.
         *m33*me2*RS1456pee*RSe45613p + 1./324.*m33*me2*RS1456pee*
         RSe45633p - 1./54.*m33*me2*RS1456ppe*RSe45613p + 1./324.*m33*
         me2*RS1456ppe*RSe45633p );

      M13p6x23R +=  + Abeps(m13,mu2)*pi16 * ( 1./162.*m33*me2*RSe45611p
         *RS34561pe + 1./54.*m33*me2*RSe45611p*RS14563pe + 1./27.*m33*
         me2*RSe45611p*RSe45613p + 1./54.*m33*me2*RSe45611p*RSp45613e
          - 1./324.*m33*me2*RSe45611p*RS3456pee - 1./324.*m33*me2*
         RSe45611p*RS3456ppe - 1./162.*m33*me2*RSe45611p*RSe45633p - 1./
         324.*m33*me2*RSe45611p*RSp45633e + 1./324.*m33*me2*RSe45611p*
         RS345pe + 1./324.*m33*me2*RSe45611p*RS346pe + 1./324.*m33*me2*
         RSe45611p*RS356pe + 5./324.*m33*me2*RSe45611p*RS1456pee + 5./
         324.*m33*me2*RSe45611p*RS1456ppe - 5./324.*m33*me2*pow(
         RSe45611p,2) + 1./54.*m33*me2*RSp45611e*RSe45613p - 1./324.*
         m33*me2*RSp45611e*RSe45633p - 5./324.*m33*me2*RSp45611e*
         RSe45611p + 1./54.*m33*me2*RS145pe*RSe45613p - 1./324.*m33*me2
         *RS145pe*RSe45633p - 5./324.*m33*me2*RS145pe*RSe45611p + 1./54.
         *m33*me2*RS146pe*RSe45613p - 1./324.*m33*me2*RS146pe*RSe45633p
          - 5./324.*m33*me2*RS146pe*RSe45611p + 1./54.*m33*me2*RS156pe*
         RSe45613p );

      M13p6x23R +=  + Abeps(m13,mu2)*pi16 * (  - 1./324.*m33*me2*
         RS156pe*RSe45633p - 5./324.*m33*me2*RS156pe*RSe45611p - 31./72.
         *pow(m33,2) - 1./81.*pow(m33,2)*RS14563pe*RS34561pe - 1./81.*
         pow(m33,2)*RSe45613p*RS34561pe - 1./81.*pow(m33,2)*RSp45613e*
         RS34561pe - 1./81.*pow(m33,2)*RS3456pee*RS34561pe - 1./81.*
         pow(m33,2)*RS3456pee*RS14563pe - 1./81.*pow(m33,2)*RS3456pee*
         RSe45613p - 1./81.*pow(m33,2)*RS3456pee*RSp45613e - 5./324.*
         pow(m33,2)*pow(RS3456pee,2) - 1./81.*pow(m33,2)*RS3456ppe*
         RS34561pe - 1./81.*pow(m33,2)*RS3456ppe*RS14563pe - 1./81.*
         pow(m33,2)*RS3456ppe*RSe45613p - 1./81.*pow(m33,2)*RS3456ppe*
         RSp45613e - 5./162.*pow(m33,2)*RS3456ppe*RS3456pee - 5./324.*
         pow(m33,2)*pow(RS3456ppe,2) + 1./81.*pow(m33,2)*RSe45633p*
         RS34561pe + 1./54.*pow(m33,2)*RSe45633p*RS3456pee + 1./54.*
         pow(m33,2)*RSe45633p*RS3456ppe + 1./81.*pow(m33,2)*RSp45633e*
         RS34561pe + 1./54.*pow(m33,2)*RSp45633e*RS3456pee + 1./54.*
         pow(m33,2)*RSp45633e*RS3456ppe );

      M13p6x23R +=  + Abeps(m13,mu2)*pi16 * ( 1./81.*pow(m33,2)*RS345pe
         *RS34561pe + 1./81.*pow(m33,2)*RS345pe*RS14563pe + 1./81.*pow(
         m33,2)*RS345pe*RSe45613p + 1./81.*pow(m33,2)*RS345pe*RSp45613e
          + 5./162.*pow(m33,2)*RS345pe*RS3456pee + 5./162.*pow(m33,2)*
         RS345pe*RS3456ppe - 1./54.*pow(m33,2)*RS345pe*RSe45633p - 1./
         54.*pow(m33,2)*RS345pe*RSp45633e - 5./324.*pow(m33,2)*pow(
         RS345pe,2) + 1./81.*pow(m33,2)*RS346pe*RS34561pe + 1./81.*pow(
         m33,2)*RS346pe*RS14563pe + 1./81.*pow(m33,2)*RS346pe*RSe45613p
          + 1./81.*pow(m33,2)*RS346pe*RSp45613e + 5./162.*pow(m33,2)*
         RS346pe*RS3456pee + 5./162.*pow(m33,2)*RS346pe*RS3456ppe - 1./
         54.*pow(m33,2)*RS346pe*RSe45633p - 1./54.*pow(m33,2)*RS346pe*
         RSp45633e - 5./162.*pow(m33,2)*RS346pe*RS345pe - 5./324.*pow(
         m33,2)*pow(RS346pe,2) + 1./81.*pow(m33,2)*RS356pe*RS34561pe + 
         1./81.*pow(m33,2)*RS356pe*RS14563pe + 1./81.*pow(m33,2)*
         RS356pe*RSe45613p + 1./81.*pow(m33,2)*RS356pe*RSp45613e + 5./
         162.*pow(m33,2)*RS356pe*RS3456pee );

      M13p6x23R +=  + Abeps(m13,mu2)*pi16 * ( 5./162.*pow(m33,2)*
         RS356pe*RS3456ppe - 1./54.*pow(m33,2)*RS356pe*RSe45633p - 1./
         54.*pow(m33,2)*RS356pe*RSp45633e - 5./162.*pow(m33,2)*RS356pe*
         RS345pe - 5./162.*pow(m33,2)*RS356pe*RS346pe - 5./324.*pow(
         m33,2)*pow(RS356pe,2) + 1./162.*pow(m33,2)*RS1456pee*RS3456pee
          + 1./162.*pow(m33,2)*RS1456pee*RS3456ppe - 1./162.*pow(m33,2)
         *RS1456pee*RS345pe - 1./162.*pow(m33,2)*RS1456pee*RS346pe - 1./
         162.*pow(m33,2)*RS1456pee*RS356pe + 1./162.*pow(m33,2)*
         RS1456ppe*RS3456pee + 1./162.*pow(m33,2)*RS1456ppe*RS3456ppe
          - 1./162.*pow(m33,2)*RS1456ppe*RS345pe - 1./162.*pow(m33,2)*
         RS1456ppe*RS346pe - 1./162.*pow(m33,2)*RS1456ppe*RS356pe - 1./
         162.*pow(m33,2)*RSe45611p*RS3456pee - 1./162.*pow(m33,2)*
         RSe45611p*RS3456ppe + 1./162.*pow(m33,2)*RSe45611p*RS345pe + 1.
         /162.*pow(m33,2)*RSe45611p*RS346pe + 1./162.*pow(m33,2)*
         RSe45611p*RS356pe - 1./162.*pow(m33,2)*RSp45611e*RS3456pee - 1.
         /162.*pow(m33,2)*RSp45611e*RS3456ppe );

      M13p6x23R +=  + Abeps(m13,mu2)*pi16 * ( 1./162.*pow(m33,2)*
         RSp45611e*RS345pe + 1./162.*pow(m33,2)*RSp45611e*RS346pe + 1./
         162.*pow(m33,2)*RSp45611e*RS356pe - 1./162.*pow(m33,2)*RS145pe
         *RS3456pee - 1./162.*pow(m33,2)*RS145pe*RS3456ppe + 1./162.*
         pow(m33,2)*RS145pe*RS345pe + 1./162.*pow(m33,2)*RS145pe*
         RS346pe + 1./162.*pow(m33,2)*RS145pe*RS356pe - 1./162.*pow(
         m33,2)*RS146pe*RS3456pee - 1./162.*pow(m33,2)*RS146pe*
         RS3456ppe + 1./162.*pow(m33,2)*RS146pe*RS345pe + 1./162.*pow(
         m33,2)*RS146pe*RS346pe + 1./162.*pow(m33,2)*RS146pe*RS356pe - 
         1./162.*pow(m33,2)*RS156pe*RS3456pee - 1./162.*pow(m33,2)*
         RS156pe*RS3456ppe + 1./162.*pow(m33,2)*RS156pe*RS345pe + 1./
         162.*pow(m33,2)*RS156pe*RS346pe + 1./162.*pow(m33,2)*RS156pe*
         RS356pe + 1./54.*m11*RD3456ep*RS34561pe + 1./162.*m11*RD3456ep
         *RS14563pe + 1./54.*m11*RD3456ep*RSe45613p + 1./54.*m11*
         RD3456ep*RSp45613e + 5./324.*m11*RS3456pee*RD3456ep + 5./324.*
         m11*RS3456ppe*RD3456ep );

      M13p6x23R +=  + Abeps(m13,mu2)*pi16 * (  - 5./324.*m11*RSe45633p*
         RD3456ep - 5./324.*m11*RSp45633e*RD3456ep - 5./324.*m11*
         RS345pe*RD3456ep - 5./324.*m11*RS346pe*RD3456ep - 5./324.*m11*
         RS356pe*RD3456ep - 1./162.*m11*RD1456ep*RS34561pe - 1./54.*m11
         *RD1456ep*RS14563pe - 1./162.*m11*RD1456ep*RSe45613p - 1./162.
         *m11*RD1456ep*RSp45613e - 5./324.*m11*RD1456ep*RS3456pee - 5./
         324.*m11*RD1456ep*RS3456ppe + 5./324.*m11*RD1456ep*RSe45633p
          + 5./324.*m11*RD1456ep*RSp45633e + 5./324.*m11*RD1456ep*
         RS345pe + 5./324.*m11*RD1456ep*RS346pe + 5./324.*m11*RD1456ep*
         RS356pe - 1./324.*m11*RS1456pee*RD3456ep + 1./324.*m11*
         RS1456pee*RD1456ep - 1./324.*m11*RS1456ppe*RD3456ep + 1./324.*
         m11*RS1456ppe*RD1456ep - 1./324.*m11*RSe45611p*RD3456ep - 1./
         108.*m11*RSe45611p*RD1456ep - 1./324.*m11*RSp45611e*RD3456ep
          - 1./108.*m11*RSp45611e*RD1456ep + 1./324.*m11*RS145pe*
         RD3456ep - 1./324.*m11*RS145pe*RD1456ep + 1./324.*m11*RS146pe*
         RD3456ep );

      M13p6x23R +=  + Abeps(m13,mu2)*pi16 * (  - 1./324.*m11*RS146pe*
         RD1456ep + 1./324.*m11*RS156pe*RD3456ep - 1./324.*m11*RS156pe*
         RD1456ep - 4./81.*m11*mp2*RSp45613e*RS34561pe - 2./81.*m11*mp2
         *RSp45613e*RS14563pe - 4./81.*m11*mp2*RSp45613e*RSe45613p - 4./
         81.*m11*mp2*pow(RSp45613e,2) - 1./54.*m11*mp2*RS3456pee*
         RSp45613e - 1./54.*m11*mp2*RS3456ppe*RSp45613e + 1./54.*m11*
         mp2*RSe45633p*RSp45613e + 1./54.*m11*mp2*RSp45633e*RS34561pe
          + 1./162.*m11*mp2*RSp45633e*RS14563pe + 1./54.*m11*mp2*
         RSp45633e*RSe45613p + 1./27.*m11*mp2*RSp45633e*RSp45613e + 5./
         324.*m11*mp2*RSp45633e*RS3456pee + 5./324.*m11*mp2*RSp45633e*
         RS3456ppe - 5./324.*m11*mp2*RSp45633e*RSe45633p - 5./324.*m11*
         mp2*pow(RSp45633e,2) + 1./54.*m11*mp2*RS345pe*RSp45613e - 5./
         324.*m11*mp2*RS345pe*RSp45633e + 1./54.*m11*mp2*RS346pe*
         RSp45613e - 5./324.*m11*mp2*RS346pe*RSp45633e + 1./54.*m11*mp2
         *RS356pe*RSp45613e - 5./324.*m11*mp2*RS356pe*RSp45633e - 1./54.
         *m11*mp2*RS1456pee*RSp45613e );

      M13p6x23R +=  + Abeps(m13,mu2)*pi16 * (  - 1./324.*m11*mp2*
         RS1456pee*RSp45633e - 1./54.*m11*mp2*RS1456ppe*RSp45613e - 1./
         324.*m11*mp2*RS1456ppe*RSp45633e + 5./162.*m11*mp2*RSe45611p*
         RSp45613e - 1./324.*m11*mp2*RSe45611p*RSp45633e + 5./162.*m11*
         mp2*RSp45611e*RS34561pe + 1./54.*m11*mp2*RSp45611e*RS14563pe
          + 5./162.*m11*mp2*RSp45611e*RSe45613p + 5./81.*m11*mp2*
         RSp45611e*RSp45613e + 1./324.*m11*mp2*RSp45611e*RS3456pee + 1./
         324.*m11*mp2*RSp45611e*RS3456ppe - 1./324.*m11*mp2*RSp45611e*
         RSe45633p - 1./162.*m11*mp2*RSp45611e*RSp45633e - 1./324.*m11*
         mp2*RSp45611e*RS345pe - 1./324.*m11*mp2*RSp45611e*RS346pe - 1./
         324.*m11*mp2*RSp45611e*RS356pe + 7./324.*m11*mp2*RSp45611e*
         RS1456pee + 7./324.*m11*mp2*RSp45611e*RS1456ppe - 1./36.*m11*
         mp2*RSp45611e*RSe45611p - 1./36.*m11*mp2*pow(RSp45611e,2) + 1./
         54.*m11*mp2*RS145pe*RSp45613e + 1./324.*m11*mp2*RS145pe*
         RSp45633e - 7./324.*m11*mp2*RS145pe*RSp45611e + 1./54.*m11*mp2
         *RS146pe*RSp45613e );

      M13p6x23R +=  + Abeps(m13,mu2)*pi16 * ( 1./324.*m11*mp2*RS146pe*
         RSp45633e - 7./324.*m11*mp2*RS146pe*RSp45611e + 1./54.*m11*mp2
         *RS156pe*RSp45613e + 1./324.*m11*mp2*RS156pe*RSp45633e - 7./
         324.*m11*mp2*RS156pe*RSp45611e - 4./81.*m11*me2*RSe45613p*
         RS34561pe - 2./81.*m11*me2*RSe45613p*RS14563pe - 4./81.*m11*
         me2*pow(RSe45613p,2) - 4./81.*m11*me2*RSp45613e*RSe45613p - 1./
         54.*m11*me2*RS3456pee*RSe45613p - 1./54.*m11*me2*RS3456ppe*
         RSe45613p + 1./54.*m11*me2*RSe45633p*RS34561pe + 1./162.*m11*
         me2*RSe45633p*RS14563pe + 1./27.*m11*me2*RSe45633p*RSe45613p
          + 1./54.*m11*me2*RSe45633p*RSp45613e + 5./324.*m11*me2*
         RSe45633p*RS3456pee + 5./324.*m11*me2*RSe45633p*RS3456ppe - 5./
         324.*m11*me2*pow(RSe45633p,2) + 1./54.*m11*me2*RSp45633e*
         RSe45613p - 5./324.*m11*me2*RSp45633e*RSe45633p + 1./54.*m11*
         me2*RS345pe*RSe45613p - 5./324.*m11*me2*RS345pe*RSe45633p + 1./
         54.*m11*me2*RS346pe*RSe45613p - 5./324.*m11*me2*RS346pe*
         RSe45633p );

      M13p6x23R +=  + Abeps(m13,mu2)*pi16 * ( 1./54.*m11*me2*RS356pe*
         RSe45613p - 5./324.*m11*me2*RS356pe*RSe45633p - 1./54.*m11*me2
         *RS1456pee*RSe45613p - 1./324.*m11*me2*RS1456pee*RSe45633p - 1.
         /54.*m11*me2*RS1456ppe*RSe45613p - 1./324.*m11*me2*RS1456ppe*
         RSe45633p + 5./162.*m11*me2*RSe45611p*RS34561pe + 1./54.*m11*
         me2*RSe45611p*RS14563pe + 5./81.*m11*me2*RSe45611p*RSe45613p
          + 5./162.*m11*me2*RSe45611p*RSp45613e + 1./324.*m11*me2*
         RSe45611p*RS3456pee + 1./324.*m11*me2*RSe45611p*RS3456ppe - 1./
         162.*m11*me2*RSe45611p*RSe45633p - 1./324.*m11*me2*RSe45611p*
         RSp45633e - 1./324.*m11*me2*RSe45611p*RS345pe - 1./324.*m11*
         me2*RSe45611p*RS346pe - 1./324.*m11*me2*RSe45611p*RS356pe + 7./
         324.*m11*me2*RSe45611p*RS1456pee + 7./324.*m11*me2*RSe45611p*
         RS1456ppe - 1./36.*m11*me2*pow(RSe45611p,2) + 5./162.*m11*me2*
         RSp45611e*RSe45613p - 1./324.*m11*me2*RSp45611e*RSe45633p - 1./
         36.*m11*me2*RSp45611e*RSe45611p + 1./54.*m11*me2*RS145pe*
         RSe45613p );

      M13p6x23R +=  + Abeps(m13,mu2)*pi16 * ( 1./324.*m11*me2*RS145pe*
         RSe45633p - 7./324.*m11*me2*RS145pe*RSe45611p + 1./54.*m11*me2
         *RS146pe*RSe45613p + 1./324.*m11*me2*RS146pe*RSe45633p - 7./
         324.*m11*me2*RS146pe*RSe45611p + 1./54.*m11*me2*RS156pe*
         RSe45613p + 1./324.*m11*me2*RS156pe*RSe45633p - 7./324.*m11*
         me2*RS156pe*RSe45611p - 31./36.*m11*m33 - 4./81.*m11*m33*pow(
         RS34561pe,2) - 2./27.*m11*m33*RS14563pe*RS34561pe - 4./81.*m11
         *m33*pow(RS14563pe,2) - 4./81.*m11*m33*RSe45613p*RS34561pe - 4.
         /81.*m11*m33*RSe45613p*RS14563pe - 4./81.*m11*m33*RSp45613e*
         RS34561pe - 4./81.*m11*m33*RSp45613e*RS14563pe - 1./27.*m11*
         m33*RS3456pee*RS34561pe - 1./27.*m11*m33*RS3456pee*RS14563pe
          - 1./54.*m11*m33*RS3456pee*RSe45613p - 1./54.*m11*m33*
         RS3456pee*RSp45613e - 5./324.*m11*m33*pow(RS3456pee,2) - 1./27.
         *m11*m33*RS3456ppe*RS34561pe - 1./27.*m11*m33*RS3456ppe*
         RS14563pe - 1./54.*m11*m33*RS3456ppe*RSe45613p - 1./54.*m11*
         m33*RS3456ppe*RSp45613e );

      M13p6x23R +=  + Abeps(m13,mu2)*pi16 * (  - 5./162.*m11*m33*
         RS3456ppe*RS3456pee - 5./324.*m11*m33*pow(RS3456ppe,2) + 1./54.
         *m11*m33*RSe45633p*RS34561pe + 5./162.*m11*m33*RSe45633p*
         RS14563pe + 5./324.*m11*m33*RSe45633p*RS3456pee + 5./324.*m11*
         m33*RSe45633p*RS3456ppe + 1./54.*m11*m33*RSp45633e*RS34561pe
          + 5./162.*m11*m33*RSp45633e*RS14563pe + 5./324.*m11*m33*
         RSp45633e*RS3456pee + 5./324.*m11*m33*RSp45633e*RS3456ppe + 1./
         27.*m11*m33*RS345pe*RS34561pe + 1./27.*m11*m33*RS345pe*
         RS14563pe + 1./54.*m11*m33*RS345pe*RSe45613p + 1./54.*m11*m33*
         RS345pe*RSp45613e + 5./162.*m11*m33*RS345pe*RS3456pee + 5./162.
         *m11*m33*RS345pe*RS3456ppe - 5./324.*m11*m33*RS345pe*RSe45633p
          - 5./324.*m11*m33*RS345pe*RSp45633e - 5./324.*m11*m33*pow(
         RS345pe,2) + 1./27.*m11*m33*RS346pe*RS34561pe + 1./27.*m11*m33
         *RS346pe*RS14563pe + 1./54.*m11*m33*RS346pe*RSe45613p + 1./54.
         *m11*m33*RS346pe*RSp45613e + 5./162.*m11*m33*RS346pe*RS3456pee
          + 5./162.*m11*m33*RS346pe*RS3456ppe );

      M13p6x23R +=  + Abeps(m13,mu2)*pi16 * (  - 5./324.*m11*m33*
         RS346pe*RSe45633p - 5./324.*m11*m33*RS346pe*RSp45633e - 5./162.
         *m11*m33*RS346pe*RS345pe - 5./324.*m11*m33*pow(RS346pe,2) + 1./
         27.*m11*m33*RS356pe*RS34561pe + 1./27.*m11*m33*RS356pe*
         RS14563pe + 1./54.*m11*m33*RS356pe*RSe45613p + 1./54.*m11*m33*
         RS356pe*RSp45613e + 5./162.*m11*m33*RS356pe*RS3456pee + 5./162.
         *m11*m33*RS356pe*RS3456ppe - 5./324.*m11*m33*RS356pe*RSe45633p
          - 5./324.*m11*m33*RS356pe*RSp45633e - 5./162.*m11*m33*RS356pe
         *RS345pe - 5./162.*m11*m33*RS356pe*RS346pe - 5./324.*m11*m33*
         pow(RS356pe,2) - 1./27.*m11*m33*RS1456pee*RS34561pe - 1./27.*
         m11*m33*RS1456pee*RS14563pe - 1./54.*m11*m33*RS1456pee*
         RSe45613p - 1./54.*m11*m33*RS1456pee*RSp45613e + 1./324.*m11*
         m33*RS1456pee*RSe45633p + 1./324.*m11*m33*RS1456pee*RSp45633e
          - 5./324.*m11*m33*pow(RS1456pee,2) - 1./27.*m11*m33*RS1456ppe
         *RS34561pe - 1./27.*m11*m33*RS1456ppe*RS14563pe - 1./54.*m11*
         m33*RS1456ppe*RSe45613p );

      M13p6x23R +=  + Abeps(m13,mu2)*pi16 * (  - 1./54.*m11*m33*
         RS1456ppe*RSp45613e + 1./324.*m11*m33*RS1456ppe*RSe45633p + 1./
         324.*m11*m33*RS1456ppe*RSp45633e - 5./162.*m11*m33*RS1456ppe*
         RS1456pee - 5./324.*m11*m33*pow(RS1456ppe,2) + 5./162.*m11*m33
         *RSe45611p*RS34561pe + 1./54.*m11*m33*RSe45611p*RS14563pe + 1./
         324.*m11*m33*RSe45611p*RS3456pee + 1./324.*m11*m33*RSe45611p*
         RS3456ppe - 1./324.*m11*m33*RSe45611p*RS345pe - 1./324.*m11*
         m33*RSe45611p*RS346pe - 1./324.*m11*m33*RSe45611p*RS356pe + 5./
         324.*m11*m33*RSe45611p*RS1456pee + 5./324.*m11*m33*RSe45611p*
         RS1456ppe + 5./162.*m11*m33*RSp45611e*RS34561pe + 1./54.*m11*
         m33*RSp45611e*RS14563pe + 1./324.*m11*m33*RSp45611e*RS3456pee
          + 1./324.*m11*m33*RSp45611e*RS3456ppe - 1./324.*m11*m33*
         RSp45611e*RS345pe - 1./324.*m11*m33*RSp45611e*RS346pe - 1./324.
         *m11*m33*RSp45611e*RS356pe + 5./324.*m11*m33*RSp45611e*
         RS1456pee + 5./324.*m11*m33*RSp45611e*RS1456ppe + 1./27.*m11*
         m33*RS145pe*RS34561pe );

      M13p6x23R +=  + Abeps(m13,mu2)*pi16 * ( 1./27.*m11*m33*RS145pe*
         RS14563pe + 1./54.*m11*m33*RS145pe*RSe45613p + 1./54.*m11*m33*
         RS145pe*RSp45613e - 1./324.*m11*m33*RS145pe*RSe45633p - 1./324.
         *m11*m33*RS145pe*RSp45633e + 5./162.*m11*m33*RS145pe*RS1456pee
          + 5./162.*m11*m33*RS145pe*RS1456ppe - 5./324.*m11*m33*RS145pe
         *RSe45611p - 5./324.*m11*m33*RS145pe*RSp45611e - 5./324.*m11*
         m33*pow(RS145pe,2) + 1./27.*m11*m33*RS146pe*RS34561pe + 1./27.
         *m11*m33*RS146pe*RS14563pe + 1./54.*m11*m33*RS146pe*RSe45613p
          + 1./54.*m11*m33*RS146pe*RSp45613e - 1./324.*m11*m33*RS146pe*
         RSe45633p - 1./324.*m11*m33*RS146pe*RSp45633e + 5./162.*m11*
         m33*RS146pe*RS1456pee + 5./162.*m11*m33*RS146pe*RS1456ppe - 5./
         324.*m11*m33*RS146pe*RSe45611p - 5./324.*m11*m33*RS146pe*
         RSp45611e - 5./162.*m11*m33*RS146pe*RS145pe - 5./324.*m11*m33*
         pow(RS146pe,2) + 1./27.*m11*m33*RS156pe*RS34561pe + 1./27.*m11
         *m33*RS156pe*RS14563pe + 1./54.*m11*m33*RS156pe*RSe45613p + 1./
         54.*m11*m33*RS156pe*RSp45613e );

      M13p6x23R +=  + Abeps(m13,mu2)*pi16 * (  - 1./324.*m11*m33*
         RS156pe*RSe45633p - 1./324.*m11*m33*RS156pe*RSp45633e + 5./162.
         *m11*m33*RS156pe*RS1456pee + 5./162.*m11*m33*RS156pe*RS1456ppe
          - 5./324.*m11*m33*RS156pe*RSe45611p - 5./324.*m11*m33*RS156pe
         *RSp45611e - 5./162.*m11*m33*RS156pe*RS145pe - 5./162.*m11*m33
         *RS156pe*RS146pe - 5./324.*m11*m33*pow(RS156pe,2) - 31./72.*
         pow(m11,2) - 1./81.*pow(m11,2)*RS14563pe*RS34561pe - 1./81.*
         pow(m11,2)*RSe45613p*RS14563pe - 1./81.*pow(m11,2)*RSp45613e*
         RS14563pe - 1./81.*pow(m11,2)*RS1456pee*RS34561pe - 1./81.*
         pow(m11,2)*RS1456pee*RS14563pe - 1./81.*pow(m11,2)*RS1456pee*
         RSe45613p - 1./81.*pow(m11,2)*RS1456pee*RSp45613e + 1./162.*
         pow(m11,2)*RS1456pee*RS3456pee + 1./162.*pow(m11,2)*RS1456pee*
         RS3456ppe - 1./162.*pow(m11,2)*RS1456pee*RSe45633p - 1./162.*
         pow(m11,2)*RS1456pee*RSp45633e - 1./162.*pow(m11,2)*RS1456pee*
         RS345pe - 1./162.*pow(m11,2)*RS1456pee*RS346pe - 1./162.*pow(
         m11,2)*RS1456pee*RS356pe );

      M13p6x23R +=  + Abeps(m13,mu2)*pi16 * (  - 5./324.*pow(m11,2)*
         pow(RS1456pee,2) - 1./81.*pow(m11,2)*RS1456ppe*RS34561pe - 1./
         81.*pow(m11,2)*RS1456ppe*RS14563pe - 1./81.*pow(m11,2)*
         RS1456ppe*RSe45613p - 1./81.*pow(m11,2)*RS1456ppe*RSp45613e + 
         1./162.*pow(m11,2)*RS1456ppe*RS3456pee + 1./162.*pow(m11,2)*
         RS1456ppe*RS3456ppe - 1./162.*pow(m11,2)*RS1456ppe*RSe45633p
          - 1./162.*pow(m11,2)*RS1456ppe*RSp45633e - 1./162.*pow(m11,2)
         *RS1456ppe*RS345pe - 1./162.*pow(m11,2)*RS1456ppe*RS346pe - 1./
         162.*pow(m11,2)*RS1456ppe*RS356pe - 5./162.*pow(m11,2)*
         RS1456ppe*RS1456pee - 5./324.*pow(m11,2)*pow(RS1456ppe,2) + 1./
         81.*pow(m11,2)*RSe45611p*RS14563pe + 1./54.*pow(m11,2)*
         RSe45611p*RS1456pee + 1./54.*pow(m11,2)*RSe45611p*RS1456ppe + 
         1./81.*pow(m11,2)*RSp45611e*RS14563pe + 1./54.*pow(m11,2)*
         RSp45611e*RS1456pee + 1./54.*pow(m11,2)*RSp45611e*RS1456ppe + 
         1./81.*pow(m11,2)*RS145pe*RS34561pe + 1./81.*pow(m11,2)*
         RS145pe*RS14563pe );

      M13p6x23R +=  + Abeps(m13,mu2)*pi16 * ( 1./81.*pow(m11,2)*RS145pe
         *RSe45613p + 1./81.*pow(m11,2)*RS145pe*RSp45613e - 1./162.*
         pow(m11,2)*RS145pe*RS3456pee - 1./162.*pow(m11,2)*RS145pe*
         RS3456ppe + 1./162.*pow(m11,2)*RS145pe*RSe45633p + 1./162.*
         pow(m11,2)*RS145pe*RSp45633e + 1./162.*pow(m11,2)*RS145pe*
         RS345pe + 1./162.*pow(m11,2)*RS145pe*RS346pe + 1./162.*pow(
         m11,2)*RS145pe*RS356pe + 5./162.*pow(m11,2)*RS145pe*RS1456pee
          + 5./162.*pow(m11,2)*RS145pe*RS1456ppe - 1./54.*pow(m11,2)*
         RS145pe*RSe45611p - 1./54.*pow(m11,2)*RS145pe*RSp45611e - 5./
         324.*pow(m11,2)*pow(RS145pe,2) + 1./81.*pow(m11,2)*RS146pe*
         RS34561pe + 1./81.*pow(m11,2)*RS146pe*RS14563pe + 1./81.*pow(
         m11,2)*RS146pe*RSe45613p + 1./81.*pow(m11,2)*RS146pe*RSp45613e
          - 1./162.*pow(m11,2)*RS146pe*RS3456pee - 1./162.*pow(m11,2)*
         RS146pe*RS3456ppe + 1./162.*pow(m11,2)*RS146pe*RSe45633p + 1./
         162.*pow(m11,2)*RS146pe*RSp45633e + 1./162.*pow(m11,2)*RS146pe
         *RS345pe );

      M13p6x23R +=  + Abeps(m13,mu2)*pi16 * ( 1./162.*pow(m11,2)*
         RS146pe*RS346pe + 1./162.*pow(m11,2)*RS146pe*RS356pe + 5./162.
         *pow(m11,2)*RS146pe*RS1456pee + 5./162.*pow(m11,2)*RS146pe*
         RS1456ppe - 1./54.*pow(m11,2)*RS146pe*RSe45611p - 1./54.*pow(
         m11,2)*RS146pe*RSp45611e - 5./162.*pow(m11,2)*RS146pe*RS145pe
          - 5./324.*pow(m11,2)*pow(RS146pe,2) + 1./81.*pow(m11,2)*
         RS156pe*RS34561pe + 1./81.*pow(m11,2)*RS156pe*RS14563pe + 1./
         81.*pow(m11,2)*RS156pe*RSe45613p + 1./81.*pow(m11,2)*RS156pe*
         RSp45613e - 1./162.*pow(m11,2)*RS156pe*RS3456pee - 1./162.*
         pow(m11,2)*RS156pe*RS3456ppe + 1./162.*pow(m11,2)*RS156pe*
         RSe45633p + 1./162.*pow(m11,2)*RS156pe*RSp45633e + 1./162.*
         pow(m11,2)*RS156pe*RS345pe + 1./162.*pow(m11,2)*RS156pe*
         RS346pe + 1./162.*pow(m11,2)*RS156pe*RS356pe + 5./162.*pow(
         m11,2)*RS156pe*RS1456pee + 5./162.*pow(m11,2)*RS156pe*
         RS1456ppe - 1./54.*pow(m11,2)*RS156pe*RSe45611p - 1./54.*pow(
         m11,2)*RS156pe*RSp45611e );

      M13p6x23R +=  + Abeps(m13,mu2)*pi16 * (  - 5./162.*pow(m11,2)*
         RS156pe*RS145pe - 5./162.*pow(m11,2)*RS156pe*RS146pe - 5./324.
         *pow(m11,2)*pow(RS156pe,2) );

      M13p6x23R +=  + Abeps(m14,mu2)*pi16 * ( 17./270.*pow(mp2,2)*
         RSp45613e + 1./270.*pow(mp2,2)*RSp45633e - 19./270.*pow(mp2,2)
         *RSp563e - 2./45.*pow(mp2,2)*RSp45611e + 7./270.*pow(mp2,2)*
         RSp561e + 1./45.*pow(mp2,2)*RSp564e + 17./270.*pow(me2,2)*
         RSe45613p + 1./270.*pow(me2,2)*RSe45633p - 19./270.*pow(me2,2)
         *RSe563p - 2./45.*pow(me2,2)*RSe45611p + 7./270.*pow(me2,2)*
         RSe561p + 1./45.*pow(me2,2)*RSe564p - 1./216.*m44*RD3456ep - 1.
         /216.*m44*RD1456ep - 1./108.*m44*mp2*RSp45613e - 1./216.*m44*
         mp2*RSp45633e + 2./45.*m44*mp2*RSp563e - 1./216.*m44*mp2*
         RSp45611e + 2./135.*m44*mp2*RSp561e - 11./270.*m44*mp2*RSp564e
          - 1./108.*m44*me2*RSe45613p - 1./216.*m44*me2*RSe45633p + 2./
         45.*m44*me2*RSe563p - 1./216.*m44*me2*RSe45611p + 2./135.*m44*
         me2*RSe561p - 11./270.*m44*me2*RSe564p - 1./108.*pow(m44,2)*
         RS34561pe - 1./108.*pow(m44,2)*RS14563pe - 1./108.*pow(m44,2)*
         RSe45613p - 1./108.*pow(m44,2)*RSp45613e + 1./216.*pow(m44,2)*
         RS3456pee );

      M13p6x23R +=  + Abeps(m14,mu2)*pi16 * ( 1./216.*pow(m44,2)*
         RS3456ppe - 1./216.*pow(m44,2)*RSe45633p - 1./216.*pow(m44,2)*
         RSp45633e - 1./216.*pow(m44,2)*RS345pe - 1./216.*pow(m44,2)*
         RS346pe - 1./216.*pow(m44,2)*RS356pe + 1./216.*pow(m44,2)*
         RS1456pee + 1./216.*pow(m44,2)*RS1456ppe - 1./216.*pow(m44,2)*
         RSe45611p - 1./216.*pow(m44,2)*RSp45611e - 1./216.*pow(m44,2)*
         RS145pe - 1./216.*pow(m44,2)*RS146pe - 1./216.*pow(m44,2)*
         RS156pe + 1./54.*pow(m44,2)*RSe564p + 1./54.*pow(m44,2)*
         RSp564e + 11./270.*m33*RD3456ep - 1./216.*m33*RD1456ep + 11./
         1080.*m33*mp2*RSp45613e + 1./30.*m33*mp2*RSp45633e - 1./40.*
         m33*mp2*RSp563e - 1./216.*m33*mp2*RSp45611e - 5./216.*m33*mp2*
         RSp561e - 1./54.*m33*mp2*RSp564e + 11./1080.*m33*me2*RSe45613p
          + 1./30.*m33*me2*RSe45633p - 1./40.*m33*me2*RSe563p - 1./216.
         *m33*me2*RSe45611p - 5./216.*m33*me2*RSe561p - 1./54.*m33*me2*
         RSe564p - 31./432.*m33*m66 - 31./432.*m33*m55 - 5./54.*m33*m44
          - 17./432.*m33*m44*RS34561pe );

      M13p6x23R +=  + Abeps(m14,mu2)*pi16 * (  - 13./432.*m33*m44*
         RS14563pe - 13./432.*m33*m44*RSe45613p - 13./432.*m33*m44*
         RSp45613e + 1./216.*m33*m44*RS3456pee + 1./216.*m33*m44*
         RS3456ppe - 1./216.*m33*m44*RS345pe - 1./216.*m33*m44*RS346pe
          - 1./240.*m33*m44*RS356pe - 19./432.*m33*m44*RSe563p - 19./
         432.*m33*m44*RSp563e + 1./108.*m33*m44*RS1456pee + 1./108.*m33
         *m44*RS1456ppe - 1./108.*m33*m44*RSe45611p - 1./108.*m33*m44*
         RSp45611e - 1./108.*m33*m44*RS145pe - 1./108.*m33*m44*RS146pe
          + 1./432.*m33*m44*RS156pe + 7./432.*m33*m44*RS456pe + 5./432.
         *m33*m44*RSe561p + 5./432.*m33*m44*RSp561e + 5./144.*m33*m44*
         RSe564p + 5./144.*m33*m44*RSp564e - 23./144.*pow(m33,2) + 113./
         2160.*pow(m33,2)*RS34561pe - 1./48.*pow(m33,2)*RS14563pe - 1./
         48.*pow(m33,2)*RSe45613p - 1./48.*pow(m33,2)*RSp45613e - 1./24.
         *pow(m33,2)*RS3456pee - 1./24.*pow(m33,2)*RS3456ppe + 1./216.*
         pow(m33,2)*RSe45633p + 1./216.*pow(m33,2)*RSp45633e + 1./24.*
         pow(m33,2)*RS345pe );

      M13p6x23R +=  + Abeps(m14,mu2)*pi16 * ( 1./24.*pow(m33,2)*RS346pe
          - 131./2160.*pow(m33,2)*RS356pe - 1./144.*pow(m33,2)*RSe563p
          - 1./144.*pow(m33,2)*RSp563e + 1./216.*pow(m33,2)*RS1456pee
          + 1./216.*pow(m33,2)*RS1456ppe - 1./216.*pow(m33,2)*RSe45611p
          - 1./216.*pow(m33,2)*RSp45611e - 1./216.*pow(m33,2)*RS145pe
          - 1./216.*pow(m33,2)*RS146pe + 1./144.*pow(m33,2)*RS156pe + 7.
         /432.*pow(m33,2)*RS456pe + 5./432.*pow(m33,2)*RSe561p + 5./432.
         *pow(m33,2)*RSp561e + 7./432.*pow(m33,2)*RSe564p + 7./432.*
         pow(m33,2)*RSp564e + 1./24.*m11*RD3456ep - 2./45.*m11*RD1456ep
          + 1./1080.*m11*mp2*RSp45613e + 1./24.*m11*mp2*RSp45633e - 23./
         216.*m11*mp2*RSp563e + 2./45.*m11*mp2*RSp45611e - 1./120.*m11*
         mp2*RSp561e + 1./1080.*m11*me2*RSe45613p + 1./24.*m11*me2*
         RSe45633p - 23./216.*m11*me2*RSe563p + 2./45.*m11*me2*
         RSe45611p - 1./120.*m11*me2*RSe561p - 31./432.*m11*m66 - 31./
         432.*m11*m55 - 5./54.*m11*m44 - 1./48.*m11*m44*RS34561pe - 13./
         432.*m11*m44*RS14563pe );

      M13p6x23R +=  + Abeps(m14,mu2)*pi16 * (  - 1./48.*m11*m44*
         RSe45613p - 1./48.*m11*m44*RSp45613e + 1./216.*m11*m44*
         RS3456pee + 1./216.*m11*m44*RS3456ppe - 1./216.*m11*m44*
         RSe45633p - 1./216.*m11*m44*RSp45633e - 1./216.*m11*m44*
         RS345pe - 1./216.*m11*m44*RS346pe - 29./432.*m11*m44*RS356pe
          - 1./16.*m11*m44*RSe563p - 1./16.*m11*m44*RSp563e + 1./216.*
         m11*m44*RSe45611p + 1./216.*m11*m44*RSp45611e + 17./2160.*m11*
         m44*RS156pe + 7./432.*m11*m44*RS456pe - 1./144.*m11*m44*
         RSe561p - 1./144.*m11*m44*RSp561e + 7./432.*m11*m44*RSe564p + 
         7./432.*m11*m44*RSp564e - 11./72.*m11*m33 - 43./2160.*m11*m33*
         RS34561pe - 23./2160.*m11*m33*RS14563pe - 1./48.*m11*m33*
         RSe45613p - 1./48.*m11*m33*RSp45613e - 1./27.*m11*m33*
         RS3456pee - 1./27.*m11*m33*RS3456ppe - 1./216.*m11*m33*
         RSe45633p - 1./216.*m11*m33*RSp45633e + 1./27.*m11*m33*RS345pe
          + 1./27.*m11*m33*RS346pe - 19./144.*m11*m33*RS356pe - 1./16.*
         m11*m33*RSe563p );

      M13p6x23R +=  + Abeps(m14,mu2)*pi16 * (  - 1./16.*m11*m33*RSp563e
          + 1./216.*m11*m33*RSe45611p + 1./216.*m11*m33*RSp45611e - 13./
         432.*m11*m33*RS156pe + 7./432.*m11*m33*RS456pe - 1./144.*m11*
         m33*RSe561p - 1./144.*m11*m33*RSp561e + 7./432.*m11*m33*
         RSe564p + 7./432.*m11*m33*RSp564e + 1./144.*pow(m11,2) + 23./
         360.*pow(m11,2)*RS14563pe - 4./27.*pow(m11,2)*RS356pe - 4./27.
         *pow(m11,2)*RSe563p - 4./27.*pow(m11,2)*RSp563e + 19./1080.*
         pow(m11,2)*RS156pe );

      M13p6x23R +=  + Abeps(m15,mu2)*pi16 * ( 17./270.*pow(mp2,2)*
         RSp45613e + 1./270.*pow(mp2,2)*RSp45633e - 19./270.*pow(mp2,2)
         *RSp463e - 2./45.*pow(mp2,2)*RSp45611e + 7./270.*pow(mp2,2)*
         RSp461e + 1./45.*pow(mp2,2)*RSp465e + 17./270.*pow(me2,2)*
         RSe45613p + 1./270.*pow(me2,2)*RSe45633p - 19./270.*pow(me2,2)
         *RSe463p - 2./45.*pow(me2,2)*RSe45611p + 7./270.*pow(me2,2)*
         RSe461p + 1./45.*pow(me2,2)*RSe465p - 1./216.*m55*RD3456ep - 1.
         /216.*m55*RD1456ep - 1./108.*m55*mp2*RSp45613e - 1./216.*m55*
         mp2*RSp45633e + 2./45.*m55*mp2*RSp463e - 1./216.*m55*mp2*
         RSp45611e + 2./135.*m55*mp2*RSp461e - 11./270.*m55*mp2*RSp465e
          - 1./108.*m55*me2*RSe45613p - 1./216.*m55*me2*RSe45633p + 2./
         45.*m55*me2*RSe463p - 1./216.*m55*me2*RSe45611p + 2./135.*m55*
         me2*RSe461p - 11./270.*m55*me2*RSe465p - 1./108.*pow(m55,2)*
         RS34561pe - 1./108.*pow(m55,2)*RS14563pe - 1./108.*pow(m55,2)*
         RSe45613p - 1./108.*pow(m55,2)*RSp45613e + 1./216.*pow(m55,2)*
         RS3456pee );

      M13p6x23R +=  + Abeps(m15,mu2)*pi16 * ( 1./216.*pow(m55,2)*
         RS3456ppe - 1./216.*pow(m55,2)*RSe45633p - 1./216.*pow(m55,2)*
         RSp45633e - 1./216.*pow(m55,2)*RS345pe - 1./216.*pow(m55,2)*
         RS346pe - 1./216.*pow(m55,2)*RS356pe + 1./216.*pow(m55,2)*
         RS1456pee + 1./216.*pow(m55,2)*RS1456ppe - 1./216.*pow(m55,2)*
         RSe45611p - 1./216.*pow(m55,2)*RSp45611e - 1./216.*pow(m55,2)*
         RS145pe - 1./216.*pow(m55,2)*RS146pe - 1./216.*pow(m55,2)*
         RS156pe + 1./54.*pow(m55,2)*RSe465p + 1./54.*pow(m55,2)*
         RSp465e + 11./270.*m33*RD3456ep - 1./216.*m33*RD1456ep + 11./
         1080.*m33*mp2*RSp45613e + 1./30.*m33*mp2*RSp45633e - 1./40.*
         m33*mp2*RSp463e - 1./216.*m33*mp2*RSp45611e - 5./216.*m33*mp2*
         RSp461e - 1./54.*m33*mp2*RSp465e + 11./1080.*m33*me2*RSe45613p
          + 1./30.*m33*me2*RSe45633p - 1./40.*m33*me2*RSe463p - 1./216.
         *m33*me2*RSe45611p - 5./216.*m33*me2*RSe461p - 1./54.*m33*me2*
         RSe465p - 31./432.*m33*m66 - 5./54.*m33*m55 - 17./432.*m33*m55
         *RS34561pe );

      M13p6x23R +=  + Abeps(m15,mu2)*pi16 * (  - 13./432.*m33*m55*
         RS14563pe - 13./432.*m33*m55*RSe45613p - 13./432.*m33*m55*
         RSp45613e + 1./216.*m33*m55*RS3456pee + 1./216.*m33*m55*
         RS3456ppe - 1./216.*m33*m55*RS345pe - 1./240.*m33*m55*RS346pe
          - 1./216.*m33*m55*RS356pe - 19./432.*m33*m55*RSe463p - 19./
         432.*m33*m55*RSp463e + 1./108.*m33*m55*RS1456pee + 1./108.*m33
         *m55*RS1456ppe - 1./108.*m33*m55*RSe45611p - 1./108.*m33*m55*
         RSp45611e - 1./108.*m33*m55*RS145pe + 1./432.*m33*m55*RS146pe
          - 1./108.*m33*m55*RS156pe + 7./432.*m33*m55*RS546pe + 5./432.
         *m33*m55*RSe461p + 5./432.*m33*m55*RSp461e + 5./144.*m33*m55*
         RSe465p + 5./144.*m33*m55*RSp465e - 31./432.*m33*m44 - 23./144.
         *pow(m33,2) + 113./2160.*pow(m33,2)*RS34561pe - 1./48.*pow(
         m33,2)*RS14563pe - 1./48.*pow(m33,2)*RSe45613p - 1./48.*pow(
         m33,2)*RSp45613e - 1./24.*pow(m33,2)*RS3456pee - 1./24.*pow(
         m33,2)*RS3456ppe + 1./216.*pow(m33,2)*RSe45633p + 1./216.*pow(
         m33,2)*RSp45633e );

      M13p6x23R +=  + Abeps(m15,mu2)*pi16 * ( 1./24.*pow(m33,2)*RS345pe
          - 131./2160.*pow(m33,2)*RS346pe + 1./24.*pow(m33,2)*RS356pe
          - 1./144.*pow(m33,2)*RSe463p - 1./144.*pow(m33,2)*RSp463e + 1.
         /216.*pow(m33,2)*RS1456pee + 1./216.*pow(m33,2)*RS1456ppe - 1./
         216.*pow(m33,2)*RSe45611p - 1./216.*pow(m33,2)*RSp45611e - 1./
         216.*pow(m33,2)*RS145pe + 1./144.*pow(m33,2)*RS146pe - 1./216.
         *pow(m33,2)*RS156pe + 7./432.*pow(m33,2)*RS546pe + 5./432.*
         pow(m33,2)*RSe461p + 5./432.*pow(m33,2)*RSp461e + 7./432.*pow(
         m33,2)*RSe465p + 7./432.*pow(m33,2)*RSp465e + 1./24.*m11*
         RD3456ep - 2./45.*m11*RD1456ep + 1./1080.*m11*mp2*RSp45613e + 
         1./24.*m11*mp2*RSp45633e - 23./216.*m11*mp2*RSp463e + 2./45.*
         m11*mp2*RSp45611e - 1./120.*m11*mp2*RSp461e + 1./1080.*m11*me2
         *RSe45613p + 1./24.*m11*me2*RSe45633p - 23./216.*m11*me2*
         RSe463p + 2./45.*m11*me2*RSe45611p - 1./120.*m11*me2*RSe461p
          - 31./432.*m11*m66 - 5./54.*m11*m55 - 1./48.*m11*m55*
         RS34561pe );

      M13p6x23R +=  + Abeps(m15,mu2)*pi16 * (  - 13./432.*m11*m55*
         RS14563pe - 1./48.*m11*m55*RSe45613p - 1./48.*m11*m55*
         RSp45613e + 1./216.*m11*m55*RS3456pee + 1./216.*m11*m55*
         RS3456ppe - 1./216.*m11*m55*RSe45633p - 1./216.*m11*m55*
         RSp45633e - 1./216.*m11*m55*RS345pe - 29./432.*m11*m55*RS346pe
          - 1./216.*m11*m55*RS356pe - 1./16.*m11*m55*RSe463p - 1./16.*
         m11*m55*RSp463e + 1./216.*m11*m55*RSe45611p + 1./216.*m11*m55*
         RSp45611e + 17./2160.*m11*m55*RS146pe + 7./432.*m11*m55*
         RS546pe - 1./144.*m11*m55*RSe461p - 1./144.*m11*m55*RSp461e + 
         7./432.*m11*m55*RSe465p + 7./432.*m11*m55*RSp465e - 31./432.*
         m11*m44 - 11./72.*m11*m33 - 43./2160.*m11*m33*RS34561pe - 23./
         2160.*m11*m33*RS14563pe - 1./48.*m11*m33*RSe45613p - 1./48.*
         m11*m33*RSp45613e - 1./27.*m11*m33*RS3456pee - 1./27.*m11*m33*
         RS3456ppe - 1./216.*m11*m33*RSe45633p - 1./216.*m11*m33*
         RSp45633e + 1./27.*m11*m33*RS345pe - 19./144.*m11*m33*RS346pe
          + 1./27.*m11*m33*RS356pe );

      M13p6x23R +=  + Abeps(m15,mu2)*pi16 * (  - 1./16.*m11*m33*RSe463p
          - 1./16.*m11*m33*RSp463e + 1./216.*m11*m33*RSe45611p + 1./216.
         *m11*m33*RSp45611e - 13./432.*m11*m33*RS146pe + 7./432.*m11*
         m33*RS546pe - 1./144.*m11*m33*RSe461p - 1./144.*m11*m33*
         RSp461e + 7./432.*m11*m33*RSe465p + 7./432.*m11*m33*RSp465e + 
         1./144.*pow(m11,2) + 23./360.*pow(m11,2)*RS14563pe - 4./27.*
         pow(m11,2)*RS346pe - 4./27.*pow(m11,2)*RSe463p - 4./27.*pow(
         m11,2)*RSp463e + 19./1080.*pow(m11,2)*RS146pe );

      M13p6x23R +=  + Abeps(m16,mu2)*pi16 * ( 17./270.*pow(mp2,2)*
         RSp45613e + 1./270.*pow(mp2,2)*RSp45633e - 19./270.*pow(mp2,2)
         *RSp453e - 2./45.*pow(mp2,2)*RSp45611e + 7./270.*pow(mp2,2)*
         RSp451e + 1./45.*pow(mp2,2)*RSp456e + 17./270.*pow(me2,2)*
         RSe45613p + 1./270.*pow(me2,2)*RSe45633p - 19./270.*pow(me2,2)
         *RSe453p - 2./45.*pow(me2,2)*RSe45611p + 7./270.*pow(me2,2)*
         RSe451p + 1./45.*pow(me2,2)*RSe456p - 1./216.*m66*RD3456ep - 1.
         /216.*m66*RD1456ep - 1./108.*m66*mp2*RSp45613e - 1./216.*m66*
         mp2*RSp45633e + 2./45.*m66*mp2*RSp453e - 1./216.*m66*mp2*
         RSp45611e + 2./135.*m66*mp2*RSp451e - 11./270.*m66*mp2*RSp456e
          - 1./108.*m66*me2*RSe45613p - 1./216.*m66*me2*RSe45633p + 2./
         45.*m66*me2*RSe453p - 1./216.*m66*me2*RSe45611p + 2./135.*m66*
         me2*RSe451p - 11./270.*m66*me2*RSe456p - 1./108.*pow(m66,2)*
         RS34561pe - 1./108.*pow(m66,2)*RS14563pe - 1./108.*pow(m66,2)*
         RSe45613p - 1./108.*pow(m66,2)*RSp45613e + 1./216.*pow(m66,2)*
         RS3456pee );

      M13p6x23R +=  + Abeps(m16,mu2)*pi16 * ( 1./216.*pow(m66,2)*
         RS3456ppe - 1./216.*pow(m66,2)*RSe45633p - 1./216.*pow(m66,2)*
         RSp45633e - 1./216.*pow(m66,2)*RS345pe - 1./216.*pow(m66,2)*
         RS346pe - 1./216.*pow(m66,2)*RS356pe + 1./216.*pow(m66,2)*
         RS1456pee + 1./216.*pow(m66,2)*RS1456ppe - 1./216.*pow(m66,2)*
         RSe45611p - 1./216.*pow(m66,2)*RSp45611e - 1./216.*pow(m66,2)*
         RS145pe - 1./216.*pow(m66,2)*RS146pe - 1./216.*pow(m66,2)*
         RS156pe + 1./54.*pow(m66,2)*RSe456p + 1./54.*pow(m66,2)*
         RSp456e + 11./270.*m33*RD3456ep - 1./216.*m33*RD1456ep + 11./
         1080.*m33*mp2*RSp45613e + 1./30.*m33*mp2*RSp45633e - 1./40.*
         m33*mp2*RSp453e - 1./216.*m33*mp2*RSp45611e - 5./216.*m33*mp2*
         RSp451e - 1./54.*m33*mp2*RSp456e + 11./1080.*m33*me2*RSe45613p
          + 1./30.*m33*me2*RSe45633p - 1./40.*m33*me2*RSe453p - 1./216.
         *m33*me2*RSe45611p - 5./216.*m33*me2*RSe451p - 1./54.*m33*me2*
         RSe456p - 5./54.*m33*m66 - 17./432.*m33*m66*RS34561pe - 13./
         432.*m33*m66*RS14563pe );

      M13p6x23R +=  + Abeps(m16,mu2)*pi16 * (  - 13./432.*m33*m66*
         RSe45613p - 13./432.*m33*m66*RSp45613e + 1./216.*m33*m66*
         RS3456pee + 1./216.*m33*m66*RS3456ppe - 1./240.*m33*m66*
         RS345pe - 1./216.*m33*m66*RS346pe - 1./216.*m33*m66*RS356pe - 
         19./432.*m33*m66*RSe453p - 19./432.*m33*m66*RSp453e + 1./108.*
         m33*m66*RS1456pee + 1./108.*m33*m66*RS1456ppe - 1./108.*m33*
         m66*RSe45611p - 1./108.*m33*m66*RSp45611e + 1./432.*m33*m66*
         RS145pe - 1./108.*m33*m66*RS146pe - 1./108.*m33*m66*RS156pe + 
         7./432.*m33*m66*RS645pe + 5./432.*m33*m66*RSe451p + 5./432.*
         m33*m66*RSp451e + 5./144.*m33*m66*RSe456p + 5./144.*m33*m66*
         RSp456e - 31./432.*m33*m55 - 31./432.*m33*m44 - 23./144.*pow(
         m33,2) + 113./2160.*pow(m33,2)*RS34561pe - 1./48.*pow(m33,2)*
         RS14563pe - 1./48.*pow(m33,2)*RSe45613p - 1./48.*pow(m33,2)*
         RSp45613e - 1./24.*pow(m33,2)*RS3456pee - 1./24.*pow(m33,2)*
         RS3456ppe + 1./216.*pow(m33,2)*RSe45633p + 1./216.*pow(m33,2)*
         RSp45633e );

      M13p6x23R +=  + Abeps(m16,mu2)*pi16 * (  - 131./2160.*pow(m33,2)*
         RS345pe + 1./24.*pow(m33,2)*RS346pe + 1./24.*pow(m33,2)*
         RS356pe - 1./144.*pow(m33,2)*RSe453p - 1./144.*pow(m33,2)*
         RSp453e + 1./216.*pow(m33,2)*RS1456pee + 1./216.*pow(m33,2)*
         RS1456ppe - 1./216.*pow(m33,2)*RSe45611p - 1./216.*pow(m33,2)*
         RSp45611e + 1./144.*pow(m33,2)*RS145pe - 1./216.*pow(m33,2)*
         RS146pe - 1./216.*pow(m33,2)*RS156pe + 7./432.*pow(m33,2)*
         RS645pe + 5./432.*pow(m33,2)*RSe451p + 5./432.*pow(m33,2)*
         RSp451e + 7./432.*pow(m33,2)*RSe456p + 7./432.*pow(m33,2)*
         RSp456e + 1./24.*m11*RD3456ep - 2./45.*m11*RD1456ep + 1./1080.
         *m11*mp2*RSp45613e + 1./24.*m11*mp2*RSp45633e - 23./216.*m11*
         mp2*RSp453e + 2./45.*m11*mp2*RSp45611e - 1./120.*m11*mp2*
         RSp451e + 1./1080.*m11*me2*RSe45613p + 1./24.*m11*me2*
         RSe45633p - 23./216.*m11*me2*RSe453p + 2./45.*m11*me2*
         RSe45611p - 1./120.*m11*me2*RSe451p - 5./54.*m11*m66 - 1./48.*
         m11*m66*RS34561pe );

      M13p6x23R +=  + Abeps(m16,mu2)*pi16 * (  - 13./432.*m11*m66*
         RS14563pe - 1./48.*m11*m66*RSe45613p - 1./48.*m11*m66*
         RSp45613e + 1./216.*m11*m66*RS3456pee + 1./216.*m11*m66*
         RS3456ppe - 1./216.*m11*m66*RSe45633p - 1./216.*m11*m66*
         RSp45633e - 29./432.*m11*m66*RS345pe - 1./216.*m11*m66*RS346pe
          - 1./216.*m11*m66*RS356pe - 1./16.*m11*m66*RSe453p - 1./16.*
         m11*m66*RSp453e + 1./216.*m11*m66*RSe45611p + 1./216.*m11*m66*
         RSp45611e + 17./2160.*m11*m66*RS145pe + 7./432.*m11*m66*
         RS645pe - 1./144.*m11*m66*RSe451p - 1./144.*m11*m66*RSp451e + 
         7./432.*m11*m66*RSe456p + 7./432.*m11*m66*RSp456e - 31./432.*
         m11*m55 - 31./432.*m11*m44 - 11./72.*m11*m33 - 43./2160.*m11*
         m33*RS34561pe - 23./2160.*m11*m33*RS14563pe - 1./48.*m11*m33*
         RSe45613p - 1./48.*m11*m33*RSp45613e - 1./27.*m11*m33*
         RS3456pee - 1./27.*m11*m33*RS3456ppe - 1./216.*m11*m33*
         RSe45633p - 1./216.*m11*m33*RSp45633e - 19./144.*m11*m33*
         RS345pe );

      M13p6x23R +=  + Abeps(m16,mu2)*pi16 * ( 1./27.*m11*m33*RS346pe + 
         1./27.*m11*m33*RS356pe - 1./16.*m11*m33*RSe453p - 1./16.*m11*
         m33*RSp453e + 1./216.*m11*m33*RSe45611p + 1./216.*m11*m33*
         RSp45611e - 13./432.*m11*m33*RS145pe + 7./432.*m11*m33*RS645pe
          - 1./144.*m11*m33*RSe451p - 1./144.*m11*m33*RSp451e + 7./432.
         *m11*m33*RSe456p + 7./432.*m11*m33*RSp456e + 1./144.*pow(
         m11,2) + 23./360.*pow(m11,2)*RS14563pe - 4./27.*pow(m11,2)*
         RS345pe - 4./27.*pow(m11,2)*RSe453p - 4./27.*pow(m11,2)*
         RSp453e + 19./1080.*pow(m11,2)*RS145pe );

      M13p6x23R +=  + Abeps(m33,mu2)*pi16 * ( 1./540.*pow(RD3456ep,2)
          + 1./540.*RD1456ep*RD3456ep - 1./270.*mp2*RD3456ep*RSp45613e
          + 1./540.*mp2*RSp45633e*RD3456ep + 1./540.*mp2*RSp45611e*
         RD3456ep - 2./405.*pow(mp2,2)*RSp45613e*RS34561pe - 1./405.*
         pow(mp2,2)*RS3456pee*RSp45613e - 1./405.*pow(mp2,2)*RS3456ppe*
         RSp45613e + 1./405.*pow(mp2,2)*RSp45633e*RS34561pe + 1./810.*
         pow(mp2,2)*RSp45633e*RS3456pee + 1./810.*pow(mp2,2)*RSp45633e*
         RS3456ppe + 1./405.*pow(mp2,2)*RS345pe*RSp45613e - 1./810.*
         pow(mp2,2)*RS345pe*RSp45633e + 1./405.*pow(mp2,2)*RS346pe*
         RSp45613e - 1./810.*pow(mp2,2)*RS346pe*RSp45633e + 1./405.*
         pow(mp2,2)*RS356pe*RSp45613e - 1./810.*pow(mp2,2)*RS356pe*
         RSp45633e + 1./405.*pow(mp2,2)*RSp45611e*RS34561pe + 1./810.*
         pow(mp2,2)*RSp45611e*RS3456pee + 1./810.*pow(mp2,2)*RSp45611e*
         RS3456ppe - 1./810.*pow(mp2,2)*RSp45611e*RS345pe - 1./810.*
         pow(mp2,2)*RSp45611e*RS346pe - 1./810.*pow(mp2,2)*RSp45611e*
         RS356pe );

      M13p6x23R +=  + Abeps(m33,mu2)*pi16 * (  - 1./270.*me2*RD3456ep*
         RSe45613p + 1./540.*me2*RSe45633p*RD3456ep + 1./540.*me2*
         RSe45611p*RD3456ep - 2./405.*pow(me2,2)*RSe45613p*RS34561pe - 
         1./405.*pow(me2,2)*RS3456pee*RSe45613p - 1./405.*pow(me2,2)*
         RS3456ppe*RSe45613p + 1./405.*pow(me2,2)*RSe45633p*RS34561pe
          + 1./810.*pow(me2,2)*RSe45633p*RS3456pee + 1./810.*pow(me2,2)
         *RSe45633p*RS3456ppe + 1./405.*pow(me2,2)*RS345pe*RSe45613p - 
         1./810.*pow(me2,2)*RS345pe*RSe45633p + 1./405.*pow(me2,2)*
         RS346pe*RSe45613p - 1./810.*pow(me2,2)*RS346pe*RSe45633p + 1./
         405.*pow(me2,2)*RS356pe*RSe45613p - 1./810.*pow(me2,2)*RS356pe
         *RSe45633p + 1./405.*pow(me2,2)*RSe45611p*RS34561pe + 1./810.*
         pow(me2,2)*RSe45611p*RS3456pee + 1./810.*pow(me2,2)*RSe45611p*
         RS3456ppe - 1./810.*pow(me2,2)*RSe45611p*RS345pe - 1./810.*
         pow(me2,2)*RSe45611p*RS346pe - 1./810.*pow(me2,2)*RSe45611p*
         RS356pe + 7./270.*m66*RD3456ep + 1./162.*m66*mp2*RSp456e*pow(
         RS345pe,2) );

      M13p6x23R +=  + Abeps(m33,mu2)*pi16 * ( 1./81.*m66*mp2*RSp456e*
         RSe453p*RS345pe + 1./81.*m66*mp2*RSp456e*RSp453e*RS345pe - 1./
         81.*m66*mp2*RSp456e*RS145pe*RS345pe - 1./81.*m66*mp2*RSp456e*
         RSe451p*RS345pe - 1./81.*m66*mp2*RSp456e*RSp451e*RS345pe + 1./
         162.*m66*me2*RSe456p*pow(RS345pe,2) + 1./81.*m66*me2*RSe456p*
         RSe453p*RS345pe + 1./81.*m66*me2*RSe456p*RSp453e*RS345pe - 1./
         81.*m66*me2*RSe456p*RS145pe*RS345pe - 1./81.*m66*me2*RSe456p*
         RSe451p*RS345pe - 1./81.*m66*me2*RSe456p*RSp451e*RS345pe + 1./
         90.*pow(m66,2)*RS345pe + 1./486.*pow(m66,2)*pow(RS345pe,2) - 1.
         /162.*pow(m66,2)*RS346pe*RS345pe - 1./162.*pow(m66,2)*RS356pe*
         RS345pe - 1./162.*pow(m66,2)*RSe563p*RS345pe - 1./162.*pow(
         m66,2)*RSe463p*RS345pe + 1./54.*pow(m66,2)*RSe453p + 1./243.*
         pow(m66,2)*RSe453p*RS345pe - 1./162.*pow(m66,2)*RSe453p*
         RS346pe - 1./162.*pow(m66,2)*RSe453p*RS356pe - 1./162.*pow(
         m66,2)*RSp563e*RS345pe - 1./162.*pow(m66,2)*RSp463e*RS345pe + 
         1./54.*pow(m66,2)*RSp453e );

      M13p6x23R +=  + Abeps(m33,mu2)*pi16 * ( 1./243.*pow(m66,2)*
         RSp453e*RS345pe - 1./162.*pow(m66,2)*RSp453e*RS346pe - 1./162.
         *pow(m66,2)*RSp453e*RS356pe - 1./54.*pow(m66,2)*RS145pe - 1./
         243.*pow(m66,2)*RS145pe*RS345pe + 1./162.*pow(m66,2)*RS145pe*
         RS346pe + 1./162.*pow(m66,2)*RS145pe*RS356pe + 1./162.*pow(
         m66,2)*RS146pe*RS345pe + 1./162.*pow(m66,2)*RS156pe*RS345pe + 
         1./162.*pow(m66,2)*RS645pe*pow(RS345pe,2) + 1./81.*pow(m66,2)*
         RS645pe*RSe453p*RS345pe + 1./81.*pow(m66,2)*RS645pe*RSp453e*
         RS345pe - 1./81.*pow(m66,2)*RS645pe*RS145pe*RS345pe + 1./162.*
         pow(m66,2)*RSe561p*RS345pe + 1./162.*pow(m66,2)*RSe461p*
         RS345pe - 1./54.*pow(m66,2)*RSe451p - 1./243.*pow(m66,2)*
         RSe451p*RS345pe + 1./162.*pow(m66,2)*RSe451p*RS346pe + 1./162.
         *pow(m66,2)*RSe451p*RS356pe - 1./81.*pow(m66,2)*RSe451p*
         RS645pe*RS345pe + 1./162.*pow(m66,2)*RSp561e*RS345pe + 1./162.
         *pow(m66,2)*RSp461e*RS345pe - 1./54.*pow(m66,2)*RSp451e - 1./
         243.*pow(m66,2)*RSp451e*RS345pe );

      M13p6x23R +=  + Abeps(m33,mu2)*pi16 * ( 1./162.*pow(m66,2)*
         RSp451e*RS346pe + 1./162.*pow(m66,2)*RSp451e*RS356pe - 1./81.*
         pow(m66,2)*RSp451e*RS645pe*RS345pe + 7./270.*m55*RD3456ep + 1./
         162.*m55*mp2*RSp465e*pow(RS346pe,2) + 1./81.*m55*mp2*RSp465e*
         RSe463p*RS346pe + 1./81.*m55*mp2*RSp465e*RSp463e*RS346pe - 1./
         81.*m55*mp2*RSp465e*RS146pe*RS346pe - 1./81.*m55*mp2*RSp465e*
         RSe461p*RS346pe - 1./81.*m55*mp2*RSp465e*RSp461e*RS346pe + 1./
         162.*m55*me2*RSe465p*pow(RS346pe,2) + 1./81.*m55*me2*RSe465p*
         RSe463p*RS346pe + 1./81.*m55*me2*RSe465p*RSp463e*RS346pe - 1./
         81.*m55*me2*RSe465p*RS146pe*RS346pe - 1./81.*m55*me2*RSe465p*
         RSe461p*RS346pe - 1./81.*m55*me2*RSe465p*RSp461e*RS346pe + 2./
         243.*m55*m66*pow(RS345pe,2) - 1./81.*m55*m66*RS346pe*RS345pe
          + 2./243.*m55*m66*pow(RS346pe,2) - 1./81.*m55*m66*RSe463p*
         RS345pe + 4./243.*m55*m66*RSe463p*RS346pe + 4./243.*m55*m66*
         RSe453p*RS345pe - 1./81.*m55*m66*RSe453p*RS346pe - 1./81.*m55*
         m66*RSp463e*RS345pe );

      M13p6x23R +=  + Abeps(m33,mu2)*pi16 * ( 4./243.*m55*m66*RSp463e*
         RS346pe + 4./243.*m55*m66*RSp453e*RS345pe - 1./81.*m55*m66*
         RSp453e*RS346pe - 4./243.*m55*m66*RS145pe*RS345pe + 1./81.*m55
         *m66*RS145pe*RS346pe + 1./81.*m55*m66*RS146pe*RS345pe - 4./243.
         *m55*m66*RS146pe*RS346pe + 1./81.*m55*m66*RSe461p*RS345pe - 4./
         243.*m55*m66*RSe461p*RS346pe - 4./243.*m55*m66*RSe451p*RS345pe
          + 1./81.*m55*m66*RSe451p*RS346pe + 1./81.*m55*m66*RSp461e*
         RS345pe - 4./243.*m55*m66*RSp461e*RS346pe - 4./243.*m55*m66*
         RSp451e*RS345pe + 1./81.*m55*m66*RSp451e*RS346pe + 1./90.*pow(
         m55,2)*RS346pe - 1./162.*pow(m55,2)*RS346pe*RS345pe + 1./486.*
         pow(m55,2)*pow(RS346pe,2) - 1./162.*pow(m55,2)*RS356pe*RS346pe
          - 1./162.*pow(m55,2)*RSe563p*RS346pe + 1./54.*pow(m55,2)*
         RSe463p - 1./162.*pow(m55,2)*RSe463p*RS345pe + 1./243.*pow(
         m55,2)*RSe463p*RS346pe - 1./162.*pow(m55,2)*RSe463p*RS356pe - 
         1./162.*pow(m55,2)*RSe453p*RS346pe - 1./162.*pow(m55,2)*
         RSp563e*RS346pe );

      M13p6x23R +=  + Abeps(m33,mu2)*pi16 * ( 1./54.*pow(m55,2)*RSp463e
          - 1./162.*pow(m55,2)*RSp463e*RS345pe + 1./243.*pow(m55,2)*
         RSp463e*RS346pe - 1./162.*pow(m55,2)*RSp463e*RS356pe - 1./162.
         *pow(m55,2)*RSp453e*RS346pe + 1./162.*pow(m55,2)*RS145pe*
         RS346pe - 1./54.*pow(m55,2)*RS146pe + 1./162.*pow(m55,2)*
         RS146pe*RS345pe - 1./243.*pow(m55,2)*RS146pe*RS346pe + 1./162.
         *pow(m55,2)*RS146pe*RS356pe + 1./162.*pow(m55,2)*RS156pe*
         RS346pe + 1./162.*pow(m55,2)*RS546pe*pow(RS346pe,2) + 1./81.*
         pow(m55,2)*RS546pe*RSe463p*RS346pe + 1./81.*pow(m55,2)*RS546pe
         *RSp463e*RS346pe - 1./81.*pow(m55,2)*RS546pe*RS146pe*RS346pe
          + 1./162.*pow(m55,2)*RSe561p*RS346pe - 1./54.*pow(m55,2)*
         RSe461p + 1./162.*pow(m55,2)*RSe461p*RS345pe - 1./243.*pow(
         m55,2)*RSe461p*RS346pe + 1./162.*pow(m55,2)*RSe461p*RS356pe - 
         1./81.*pow(m55,2)*RSe461p*RS546pe*RS346pe + 1./162.*pow(m55,2)
         *RSe451p*RS346pe + 1./162.*pow(m55,2)*RSp561e*RS346pe - 1./54.
         *pow(m55,2)*RSp461e );

      M13p6x23R +=  + Abeps(m33,mu2)*pi16 * ( 1./162.*pow(m55,2)*
         RSp461e*RS345pe - 1./243.*pow(m55,2)*RSp461e*RS346pe + 1./162.
         *pow(m55,2)*RSp461e*RS356pe - 1./81.*pow(m55,2)*RSp461e*
         RS546pe*RS346pe + 1./162.*pow(m55,2)*RSp451e*RS346pe + 7./270.
         *m44*RD3456ep + 1./162.*m44*mp2*RSp564e*pow(RS356pe,2) + 1./81.
         *m44*mp2*RSp564e*RSe563p*RS356pe + 1./81.*m44*mp2*RSp564e*
         RSp563e*RS356pe - 1./81.*m44*mp2*RSp564e*RS156pe*RS356pe - 1./
         81.*m44*mp2*RSp564e*RSe561p*RS356pe - 1./81.*m44*mp2*RSp564e*
         RSp561e*RS356pe + 1./162.*m44*me2*RSe564p*pow(RS356pe,2) + 1./
         81.*m44*me2*RSe564p*RSe563p*RS356pe + 1./81.*m44*me2*RSe564p*
         RSp563e*RS356pe - 1./81.*m44*me2*RSe564p*RS156pe*RS356pe - 1./
         81.*m44*me2*RSe564p*RSe561p*RS356pe - 1./81.*m44*me2*RSe564p*
         RSp561e*RS356pe + 2./243.*m44*m66*pow(RS345pe,2) - 1./81.*m44*
         m66*RS356pe*RS345pe + 2./243.*m44*m66*pow(RS356pe,2) - 1./81.*
         m44*m66*RSe563p*RS345pe + 4./243.*m44*m66*RSe563p*RS356pe + 4./
         243.*m44*m66*RSe453p*RS345pe );

      M13p6x23R +=  + Abeps(m33,mu2)*pi16 * (  - 1./81.*m44*m66*RSe453p
         *RS356pe - 1./81.*m44*m66*RSp563e*RS345pe + 4./243.*m44*m66*
         RSp563e*RS356pe + 4./243.*m44*m66*RSp453e*RS345pe - 1./81.*m44
         *m66*RSp453e*RS356pe - 4./243.*m44*m66*RS145pe*RS345pe + 1./81.
         *m44*m66*RS145pe*RS356pe + 1./81.*m44*m66*RS156pe*RS345pe - 4./
         243.*m44*m66*RS156pe*RS356pe + 1./81.*m44*m66*RSe561p*RS345pe
          - 4./243.*m44*m66*RSe561p*RS356pe - 4./243.*m44*m66*RSe451p*
         RS345pe + 1./81.*m44*m66*RSe451p*RS356pe + 1./81.*m44*m66*
         RSp561e*RS345pe - 4./243.*m44*m66*RSp561e*RS356pe - 4./243.*
         m44*m66*RSp451e*RS345pe + 1./81.*m44*m66*RSp451e*RS356pe + 2./
         243.*m44*m55*pow(RS346pe,2) - 1./81.*m44*m55*RS356pe*RS346pe
          + 2./243.*m44*m55*pow(RS356pe,2) - 1./81.*m44*m55*RSe563p*
         RS346pe + 4./243.*m44*m55*RSe563p*RS356pe + 4./243.*m44*m55*
         RSe463p*RS346pe - 1./81.*m44*m55*RSe463p*RS356pe - 1./81.*m44*
         m55*RSp563e*RS346pe + 4./243.*m44*m55*RSp563e*RS356pe + 4./243.
         *m44*m55*RSp463e*RS346pe );

      M13p6x23R +=  + Abeps(m33,mu2)*pi16 * (  - 1./81.*m44*m55*RSp463e
         *RS356pe - 4./243.*m44*m55*RS146pe*RS346pe + 1./81.*m44*m55*
         RS146pe*RS356pe + 1./81.*m44*m55*RS156pe*RS346pe - 4./243.*m44
         *m55*RS156pe*RS356pe + 1./81.*m44*m55*RSe561p*RS346pe - 4./243.
         *m44*m55*RSe561p*RS356pe - 4./243.*m44*m55*RSe461p*RS346pe + 1.
         /81.*m44*m55*RSe461p*RS356pe + 1./81.*m44*m55*RSp561e*RS346pe
          - 4./243.*m44*m55*RSp561e*RS356pe - 4./243.*m44*m55*RSp461e*
         RS346pe + 1./81.*m44*m55*RSp461e*RS356pe + 1./90.*pow(m44,2)*
         RS356pe - 1./162.*pow(m44,2)*RS356pe*RS345pe - 1./162.*pow(
         m44,2)*RS356pe*RS346pe + 1./486.*pow(m44,2)*pow(RS356pe,2) + 1.
         /54.*pow(m44,2)*RSe563p - 1./162.*pow(m44,2)*RSe563p*RS345pe
          - 1./162.*pow(m44,2)*RSe563p*RS346pe + 1./243.*pow(m44,2)*
         RSe563p*RS356pe - 1./162.*pow(m44,2)*RSe463p*RS356pe - 1./162.
         *pow(m44,2)*RSe453p*RS356pe + 1./54.*pow(m44,2)*RSp563e - 1./
         162.*pow(m44,2)*RSp563e*RS345pe - 1./162.*pow(m44,2)*RSp563e*
         RS346pe );

      M13p6x23R +=  + Abeps(m33,mu2)*pi16 * ( 1./243.*pow(m44,2)*
         RSp563e*RS356pe - 1./162.*pow(m44,2)*RSp463e*RS356pe - 1./162.
         *pow(m44,2)*RSp453e*RS356pe + 1./162.*pow(m44,2)*RS145pe*
         RS356pe + 1./162.*pow(m44,2)*RS146pe*RS356pe - 1./54.*pow(
         m44,2)*RS156pe + 1./162.*pow(m44,2)*RS156pe*RS345pe + 1./162.*
         pow(m44,2)*RS156pe*RS346pe - 1./243.*pow(m44,2)*RS156pe*
         RS356pe + 1./162.*pow(m44,2)*RS456pe*pow(RS356pe,2) + 1./81.*
         pow(m44,2)*RS456pe*RSe563p*RS356pe + 1./81.*pow(m44,2)*RS456pe
         *RSp563e*RS356pe - 1./81.*pow(m44,2)*RS456pe*RS156pe*RS356pe
          - 1./54.*pow(m44,2)*RSe561p + 1./162.*pow(m44,2)*RSe561p*
         RS345pe + 1./162.*pow(m44,2)*RSe561p*RS346pe - 1./243.*pow(
         m44,2)*RSe561p*RS356pe - 1./81.*pow(m44,2)*RSe561p*RS456pe*
         RS356pe + 1./162.*pow(m44,2)*RSe461p*RS356pe + 1./162.*pow(
         m44,2)*RSe451p*RS356pe - 1./54.*pow(m44,2)*RSp561e + 1./162.*
         pow(m44,2)*RSp561e*RS345pe + 1./162.*pow(m44,2)*RSp561e*
         RS346pe );

      M13p6x23R +=  + Abeps(m33,mu2)*pi16 * (  - 1./243.*pow(m44,2)*
         RSp561e*RS356pe - 1./81.*pow(m44,2)*RSp561e*RS456pe*RS356pe + 
         1./162.*pow(m44,2)*RSp461e*RS356pe + 1./162.*pow(m44,2)*
         RSp451e*RS356pe + 2./45.*m33*RD3456ep + 1./324.*m33*RD3456ep*
         RS34561pe + 1./36.*m33*RD3456ep*RS14563pe + 1./36.*m33*
         RD3456ep*RSe45613p + 1./36.*m33*RD3456ep*RSp45613e + 53./3240.
         *m33*RS3456pee*RD3456ep + 53./3240.*m33*RS3456ppe*RD3456ep - 7.
         /216.*m33*RSe45633p*RD3456ep - 7./216.*m33*RSp45633e*RD3456ep
          - 53./3240.*m33*RS345pe*RD3456ep - 53./3240.*m33*RS346pe*
         RD3456ep - 53./3240.*m33*RS356pe*RD3456ep - 2./405.*m33*
         RD1456ep*RS34561pe + 1./81.*m33*RD1456ep*RS3456pee + 1./81.*
         m33*RD1456ep*RS3456ppe - 1./81.*m33*RD1456ep*RS345pe - 1./81.*
         m33*RD1456ep*RS346pe - 1./81.*m33*RD1456ep*RS356pe - 1./216.*
         m33*RS1456pee*RD3456ep - 1./216.*m33*RS1456ppe*RD3456ep + 1./
         216.*m33*RSe45611p*RD3456ep + 1./216.*m33*RSp45611e*RD3456ep
          + 1./216.*m33*RS145pe*RD3456ep );

      M13p6x23R +=  + Abeps(m33,mu2)*pi16 * ( 1./216.*m33*RS146pe*
         RD3456ep + 1./216.*m33*RS156pe*RD3456ep + 4./81.*m33*mp2*
         RSp45613e*RS34561pe - 19./810.*m33*mp2*RS3456pee*RSp45613e - 
         19./810.*m33*mp2*RS3456ppe*RSp45613e + 1./90.*m33*mp2*
         RSp45633e*RS34561pe + 1./27.*m33*mp2*RSp45633e*RS14563pe + 1./
         27.*m33*mp2*RSp45633e*RSe45613p + 1./27.*m33*mp2*RSp45633e*
         RSp45613e + 11./540.*m33*mp2*RSp45633e*RS3456pee + 11./540.*
         m33*mp2*RSp45633e*RS3456ppe - 1./27.*m33*mp2*RSp45633e*
         RSe45633p - 1./27.*m33*mp2*pow(RSp45633e,2) + 19./810.*m33*mp2
         *RS345pe*RSp45613e - 11./540.*m33*mp2*RS345pe*RSp45633e + 19./
         810.*m33*mp2*RS346pe*RSp45613e - 11./540.*m33*mp2*RS346pe*
         RSp45633e + 19./810.*m33*mp2*RS356pe*RSp45613e - 11./540.*m33*
         mp2*RS356pe*RSp45633e - 2./405.*m33*mp2*RSp45611e*RS34561pe + 
         1./81.*m33*mp2*RSp45611e*RS3456pee + 1./81.*m33*mp2*RSp45611e*
         RS3456ppe - 1./81.*m33*mp2*RSp45611e*RS345pe - 1./81.*m33*mp2*
         RSp45611e*RS346pe );

      M13p6x23R +=  + Abeps(m33,mu2)*pi16 * (  - 1./81.*m33*mp2*
         RSp45611e*RS356pe + 4./81.*m33*me2*RSe45613p*RS34561pe - 19./
         810.*m33*me2*RS3456pee*RSe45613p - 19./810.*m33*me2*RS3456ppe*
         RSe45613p + 1./90.*m33*me2*RSe45633p*RS34561pe + 1./27.*m33*
         me2*RSe45633p*RS14563pe + 1./27.*m33*me2*RSe45633p*RSe45613p
          + 1./27.*m33*me2*RSe45633p*RSp45613e + 11./540.*m33*me2*
         RSe45633p*RS3456pee + 11./540.*m33*me2*RSe45633p*RS3456ppe - 1.
         /27.*m33*me2*pow(RSe45633p,2) - 1./27.*m33*me2*RSp45633e*
         RSe45633p + 19./810.*m33*me2*RS345pe*RSe45613p - 11./540.*m33*
         me2*RS345pe*RSe45633p + 19./810.*m33*me2*RS346pe*RSe45613p - 
         11./540.*m33*me2*RS346pe*RSe45633p + 19./810.*m33*me2*RS356pe*
         RSe45613p - 11./540.*m33*me2*RS356pe*RSe45633p - 2./405.*m33*
         me2*RSe45611p*RS34561pe + 1./81.*m33*me2*RSe45611p*RS3456pee
          + 1./81.*m33*me2*RSe45611p*RS3456ppe - 1./81.*m33*me2*
         RSe45611p*RS345pe - 1./81.*m33*me2*RSe45611p*RS346pe - 1./81.*
         m33*me2*RSe45611p*RS356pe );

      M13p6x23R +=  + Abeps(m33,mu2)*pi16 * (  - 101./1080.*m33*m66*
         RS34561pe - 2./81.*m33*m66*RS14563pe - 2./81.*m33*m66*
         RSe45613p - 2./81.*m33*m66*RSp45613e - 23./1296.*m33*m66*
         RS3456pee - 23./1296.*m33*m66*RS3456ppe - 4./81.*m33*m66*
         RSe45633p - 4./81.*m33*m66*RSp45633e - 209./6480.*m33*m66*
         RS345pe - 5./162.*m33*m66*pow(RS345pe,2) + 23./1296.*m33*m66*
         RS346pe - 1./162.*m33*m66*RS346pe*RS345pe - 1./162.*m33*m66*
         pow(RS346pe,2) + 23./1296.*m33*m66*RS356pe - 1./162.*m33*m66*
         RS356pe*RS345pe - 1./162.*m33*m66*pow(RS356pe,2) + 1./27.*m33*
         m66*RSe453p + 1./27.*m33*m66*RSp453e - 2./27.*m33*m66*RS145pe
          + 10./81.*m33*m66*RS145pe*RS345pe + 1./81.*m33*m66*RS145pe*
         RS346pe + 1./81.*m33*m66*RS145pe*RS356pe + 1./81.*m33*m66*
         RS146pe*RS345pe + 2./81.*m33*m66*RS146pe*RS346pe + 1./81.*m33*
         m66*RS156pe*RS345pe + 2./81.*m33*m66*RS156pe*RS356pe + 1./81.*
         m33*m66*RSe561p*RS345pe + 2./81.*m33*m66*RSe561p*RS356pe + 1./
         81.*m33*m66*RSe461p*RS345pe );

      M13p6x23R +=  + Abeps(m33,mu2)*pi16 * ( 2./81.*m33*m66*RSe461p*
         RS346pe - 2./27.*m33*m66*RSe451p + 10./81.*m33*m66*RSe451p*
         RS345pe + 1./81.*m33*m66*RSe451p*RS346pe + 1./81.*m33*m66*
         RSe451p*RS356pe + 1./81.*m33*m66*RSp561e*RS345pe + 2./81.*m33*
         m66*RSp561e*RS356pe + 1./81.*m33*m66*RSp461e*RS345pe + 2./81.*
         m33*m66*RSp461e*RS346pe - 2./27.*m33*m66*RSp451e + 10./81.*m33
         *m66*RSp451e*RS345pe + 1./81.*m33*m66*RSp451e*RS346pe + 1./81.
         *m33*m66*RSp451e*RS356pe - 101./1080.*m33*m55*RS34561pe - 2./
         81.*m33*m55*RS14563pe - 2./81.*m33*m55*RSe45613p - 2./81.*m33*
         m55*RSp45613e - 23./1296.*m33*m55*RS3456pee - 23./1296.*m33*
         m55*RS3456ppe - 4./81.*m33*m55*RSe45633p - 4./81.*m33*m55*
         RSp45633e + 23./1296.*m33*m55*RS345pe - 1./162.*m33*m55*pow(
         RS345pe,2) - 209./6480.*m33*m55*RS346pe - 1./162.*m33*m55*
         RS346pe*RS345pe - 5./162.*m33*m55*pow(RS346pe,2) + 23./1296.*
         m33*m55*RS356pe - 1./162.*m33*m55*RS356pe*RS346pe - 1./162.*
         m33*m55*pow(RS356pe,2) );

      M13p6x23R +=  + Abeps(m33,mu2)*pi16 * ( 1./27.*m33*m55*RSe463p + 
         1./27.*m33*m55*RSp463e + 2./81.*m33*m55*RS145pe*RS345pe + 1./
         81.*m33*m55*RS145pe*RS346pe - 2./27.*m33*m55*RS146pe + 1./81.*
         m33*m55*RS146pe*RS345pe + 10./81.*m33*m55*RS146pe*RS346pe + 1./
         81.*m33*m55*RS146pe*RS356pe + 1./81.*m33*m55*RS156pe*RS346pe
          + 2./81.*m33*m55*RS156pe*RS356pe + 1./81.*m33*m55*RSe561p*
         RS346pe + 2./81.*m33*m55*RSe561p*RS356pe - 2./27.*m33*m55*
         RSe461p + 1./81.*m33*m55*RSe461p*RS345pe + 10./81.*m33*m55*
         RSe461p*RS346pe + 1./81.*m33*m55*RSe461p*RS356pe + 2./81.*m33*
         m55*RSe451p*RS345pe + 1./81.*m33*m55*RSe451p*RS346pe + 1./81.*
         m33*m55*RSp561e*RS346pe + 2./81.*m33*m55*RSp561e*RS356pe - 2./
         27.*m33*m55*RSp461e + 1./81.*m33*m55*RSp461e*RS345pe + 10./81.
         *m33*m55*RSp461e*RS346pe + 1./81.*m33*m55*RSp461e*RS356pe + 2./
         81.*m33*m55*RSp451e*RS345pe + 1./81.*m33*m55*RSp451e*RS346pe
          - 101./1080.*m33*m44*RS34561pe - 2./81.*m33*m44*RS14563pe - 2.
         /81.*m33*m44*RSe45613p );

      M13p6x23R +=  + Abeps(m33,mu2)*pi16 * (  - 2./81.*m33*m44*
         RSp45613e - 23./1296.*m33*m44*RS3456pee - 23./1296.*m33*m44*
         RS3456ppe - 4./81.*m33*m44*RSe45633p - 4./81.*m33*m44*
         RSp45633e + 23./1296.*m33*m44*RS345pe - 1./162.*m33*m44*pow(
         RS345pe,2) + 23./1296.*m33*m44*RS346pe - 1./162.*m33*m44*pow(
         RS346pe,2) - 209./6480.*m33*m44*RS356pe - 1./162.*m33*m44*
         RS356pe*RS345pe - 1./162.*m33*m44*RS356pe*RS346pe - 5./162.*
         m33*m44*pow(RS356pe,2) + 1./27.*m33*m44*RSe563p + 1./27.*m33*
         m44*RSp563e + 2./81.*m33*m44*RS145pe*RS345pe + 1./81.*m33*m44*
         RS145pe*RS356pe + 2./81.*m33*m44*RS146pe*RS346pe + 1./81.*m33*
         m44*RS146pe*RS356pe - 2./27.*m33*m44*RS156pe + 1./81.*m33*m44*
         RS156pe*RS345pe + 1./81.*m33*m44*RS156pe*RS346pe + 10./81.*m33
         *m44*RS156pe*RS356pe - 2./27.*m33*m44*RSe561p + 1./81.*m33*m44
         *RSe561p*RS345pe + 1./81.*m33*m44*RSe561p*RS346pe + 10./81.*
         m33*m44*RSe561p*RS356pe + 2./81.*m33*m44*RSe461p*RS346pe + 1./
         81.*m33*m44*RSe461p*RS356pe );

      M13p6x23R +=  + Abeps(m33,mu2)*pi16 * ( 2./81.*m33*m44*RSe451p*
         RS345pe + 1./81.*m33*m44*RSe451p*RS356pe - 2./27.*m33*m44*
         RSp561e + 1./81.*m33*m44*RSp561e*RS345pe + 1./81.*m33*m44*
         RSp561e*RS346pe + 10./81.*m33*m44*RSp561e*RS356pe + 2./81.*m33
         *m44*RSp461e*RS346pe + 1./81.*m33*m44*RSp461e*RS356pe + 2./81.
         *m33*m44*RSp451e*RS345pe + 1./81.*m33*m44*RSp451e*RS356pe + 37.
         /144.*pow(m33,2) - 223./2160.*pow(m33,2)*RS34561pe + 31./810.*
         pow(m33,2)*pow(RS34561pe,2) - 5./27.*pow(m33,2)*RS14563pe - 1./
         162.*pow(m33,2)*RS14563pe*RS34561pe - 5./27.*pow(m33,2)*
         RSe45613p - 1./162.*pow(m33,2)*RSe45613p*RS34561pe - 5./27.*
         pow(m33,2)*RSp45613e - 1./162.*pow(m33,2)*RSp45613e*RS34561pe
          - 1./24.*pow(m33,2)*RS3456pee - 17./135.*pow(m33,2)*RS3456pee
         *RS34561pe - 19./162.*pow(m33,2)*RS3456pee*RS14563pe - 19./162.
         *pow(m33,2)*RS3456pee*RSe45613p - 19./162.*pow(m33,2)*
         RS3456pee*RSp45613e - 5./162.*pow(m33,2)*pow(RS3456pee,2) - 1./
         24.*pow(m33,2)*RS3456ppe );

      M13p6x23R +=  + Abeps(m33,mu2)*pi16 * (  - 17./135.*pow(m33,2)*
         RS3456ppe*RS34561pe - 19./162.*pow(m33,2)*RS3456ppe*RS14563pe
          - 19./162.*pow(m33,2)*RS3456ppe*RSe45613p - 19./162.*pow(
         m33,2)*RS3456ppe*RSp45613e - 5./81.*pow(m33,2)*RS3456ppe*
         RS3456pee - 5./162.*pow(m33,2)*pow(RS3456ppe,2) - 1./27.*pow(
         m33,2)*RSe45633p + 1./162.*pow(m33,2)*RSe45633p*RS34561pe + 5./
         108.*pow(m33,2)*RSe45633p*RS3456pee + 5./108.*pow(m33,2)*
         RSe45633p*RS3456ppe - 1./27.*pow(m33,2)*RSp45633e + 1./162.*
         pow(m33,2)*RSp45633e*RS34561pe + 5./108.*pow(m33,2)*RSp45633e*
         RS3456pee + 5./108.*pow(m33,2)*RSp45633e*RS3456ppe + 109./2160.
         *pow(m33,2)*RS345pe + 17./135.*pow(m33,2)*RS345pe*RS34561pe + 
         19./162.*pow(m33,2)*RS345pe*RS14563pe + 19./162.*pow(m33,2)*
         RS345pe*RSe45613p + 19./162.*pow(m33,2)*RS345pe*RSp45613e + 5./
         81.*pow(m33,2)*RS345pe*RS3456pee + 5./81.*pow(m33,2)*RS345pe*
         RS3456ppe - 5./108.*pow(m33,2)*RS345pe*RSe45633p - 5./108.*
         pow(m33,2)*RS345pe*RSp45633e );

      M13p6x23R +=  + Abeps(m33,mu2)*pi16 * (  - 5./162.*pow(m33,2)*
         pow(RS345pe,2) + 109./2160.*pow(m33,2)*RS346pe + 17./135.*pow(
         m33,2)*RS346pe*RS34561pe + 19./162.*pow(m33,2)*RS346pe*
         RS14563pe + 19./162.*pow(m33,2)*RS346pe*RSe45613p + 19./162.*
         pow(m33,2)*RS346pe*RSp45613e + 5./81.*pow(m33,2)*RS346pe*
         RS3456pee + 5./81.*pow(m33,2)*RS346pe*RS3456ppe - 5./108.*pow(
         m33,2)*RS346pe*RSe45633p - 5./108.*pow(m33,2)*RS346pe*
         RSp45633e - 5./81.*pow(m33,2)*RS346pe*RS345pe - 5./162.*pow(
         m33,2)*pow(RS346pe,2) + 109./2160.*pow(m33,2)*RS356pe + 17./
         135.*pow(m33,2)*RS356pe*RS34561pe + 19./162.*pow(m33,2)*
         RS356pe*RS14563pe + 19./162.*pow(m33,2)*RS356pe*RSe45613p + 19.
         /162.*pow(m33,2)*RS356pe*RSp45613e + 5./81.*pow(m33,2)*RS356pe
         *RS3456pee + 5./81.*pow(m33,2)*RS356pe*RS3456ppe - 5./108.*
         pow(m33,2)*RS356pe*RSe45633p - 5./108.*pow(m33,2)*RS356pe*
         RSp45633e - 5./81.*pow(m33,2)*RS356pe*RS345pe - 5./81.*pow(
         m33,2)*RS356pe*RS346pe );

      M13p6x23R +=  + Abeps(m33,mu2)*pi16 * (  - 5./162.*pow(m33,2)*
         pow(RS356pe,2) + 1./54.*pow(m33,2)*RSe563p + 1./54.*pow(m33,2)
         *RSe463p + 1./54.*pow(m33,2)*RSe453p + 1./54.*pow(m33,2)*
         RSp563e + 1./54.*pow(m33,2)*RSp463e + 1./54.*pow(m33,2)*
         RSp453e + 1./324.*pow(m33,2)*RS1456pee*RS3456pee + 1./324.*
         pow(m33,2)*RS1456pee*RS3456ppe - 1./324.*pow(m33,2)*RS1456pee*
         RS345pe - 1./324.*pow(m33,2)*RS1456pee*RS346pe - 1./324.*pow(
         m33,2)*RS1456pee*RS356pe + 1./324.*pow(m33,2)*RS1456ppe*
         RS3456pee + 1./324.*pow(m33,2)*RS1456ppe*RS3456ppe - 1./324.*
         pow(m33,2)*RS1456ppe*RS345pe - 1./324.*pow(m33,2)*RS1456ppe*
         RS346pe - 1./324.*pow(m33,2)*RS1456ppe*RS356pe - 1./324.*pow(
         m33,2)*RSe45611p*RS3456pee - 1./324.*pow(m33,2)*RSe45611p*
         RS3456ppe + 1./324.*pow(m33,2)*RSe45611p*RS345pe + 1./324.*
         pow(m33,2)*RSe45611p*RS346pe + 1./324.*pow(m33,2)*RSe45611p*
         RS356pe - 1./324.*pow(m33,2)*RSp45611e*RS3456pee - 1./324.*
         pow(m33,2)*RSp45611e*RS3456ppe );

      M13p6x23R +=  + Abeps(m33,mu2)*pi16 * ( 1./324.*pow(m33,2)*
         RSp45611e*RS345pe + 1./324.*pow(m33,2)*RSp45611e*RS346pe + 1./
         324.*pow(m33,2)*RSp45611e*RS356pe - 1./18.*pow(m33,2)*RS145pe
          - 1./324.*pow(m33,2)*RS145pe*RS3456pee - 1./324.*pow(m33,2)*
         RS145pe*RS3456ppe + 1./324.*pow(m33,2)*RS145pe*RS345pe + 1./
         324.*pow(m33,2)*RS145pe*RS346pe + 1./324.*pow(m33,2)*RS145pe*
         RS356pe - 1./18.*pow(m33,2)*RS146pe - 1./324.*pow(m33,2)*
         RS146pe*RS3456pee - 1./324.*pow(m33,2)*RS146pe*RS3456ppe + 1./
         324.*pow(m33,2)*RS146pe*RS345pe + 1./324.*pow(m33,2)*RS146pe*
         RS346pe + 1./324.*pow(m33,2)*RS146pe*RS356pe - 1./18.*pow(
         m33,2)*RS156pe - 1./324.*pow(m33,2)*RS156pe*RS3456pee - 1./324.
         *pow(m33,2)*RS156pe*RS3456ppe + 1./324.*pow(m33,2)*RS156pe*
         RS345pe + 1./324.*pow(m33,2)*RS156pe*RS346pe + 1./324.*pow(
         m33,2)*RS156pe*RS356pe - 1./18.*pow(m33,2)*RSe561p - 1./18.*
         pow(m33,2)*RSe461p - 1./18.*pow(m33,2)*RSe451p - 1./18.*pow(
         m33,2)*RSp561e );

      M13p6x23R +=  + Abeps(m33,mu2)*pi16 * (  - 1./18.*pow(m33,2)*
         RSp461e - 1./18.*pow(m33,2)*RSp451e + 1./30.*m11*RD3456ep - 17.
         /1620.*m11*RD3456ep*RS34561pe - 7./540.*m11*RD3456ep*RS14563pe
          - 1./108.*m11*RD3456ep*RSe45613p - 1./108.*m11*RD3456ep*
         RSp45613e - 13./648.*m11*RS3456pee*RD3456ep - 13./648.*m11*
         RS3456ppe*RD3456ep + 1./216.*m11*RSe45633p*RD3456ep + 1./216.*
         m11*RSp45633e*RD3456ep + 13./648.*m11*RS345pe*RD3456ep + 13./
         648.*m11*RS346pe*RD3456ep + 13./648.*m11*RS356pe*RD3456ep + 8./
         405.*m11*RD1456ep*RS34561pe + 11./810.*m11*RD1456ep*RS3456pee
          + 11./810.*m11*RD1456ep*RS3456ppe - 11./810.*m11*RD1456ep*
         RS345pe - 11./810.*m11*RD1456ep*RS346pe - 11./810.*m11*
         RD1456ep*RS356pe - 7./1080.*m11*RS1456pee*RD3456ep - 7./1080.*
         m11*RS1456ppe*RD3456ep + 1./216.*m11*RSe45611p*RD3456ep + 1./
         216.*m11*RSp45611e*RD3456ep + 7./1080.*m11*RS145pe*RD3456ep + 
         7./1080.*m11*RS146pe*RD3456ep + 7./1080.*m11*RS156pe*RD3456ep
          + 17./405.*m11*mp2*RSp45613e*RS34561pe );

      M13p6x23R +=  + Abeps(m33,mu2)*pi16 * (  - 2./405.*m11*mp2*
         RS3456pee*RSp45613e - 2./405.*m11*mp2*RS3456ppe*RSp45613e - 1./
         810.*m11*mp2*RSp45633e*RS34561pe - 5./324.*m11*mp2*RSp45633e*
         RS3456pee - 5./324.*m11*mp2*RSp45633e*RS3456ppe + 2./405.*m11*
         mp2*RS345pe*RSp45613e + 5./324.*m11*mp2*RS345pe*RSp45633e + 2./
         405.*m11*mp2*RS346pe*RSp45613e + 5./324.*m11*mp2*RS346pe*
         RSp45633e + 2./405.*m11*mp2*RS356pe*RSp45613e + 5./324.*m11*
         mp2*RS356pe*RSp45633e + 2./135.*m11*mp2*RSp45611e*RS34561pe + 
         1./90.*m11*mp2*RSp45611e*RS3456pee + 1./90.*m11*mp2*RSp45611e*
         RS3456ppe - 1./90.*m11*mp2*RSp45611e*RS345pe - 1./90.*m11*mp2*
         RSp45611e*RS346pe - 1./90.*m11*mp2*RSp45611e*RS356pe + 17./405.
         *m11*me2*RSe45613p*RS34561pe - 2./405.*m11*me2*RS3456pee*
         RSe45613p - 2./405.*m11*me2*RS3456ppe*RSe45613p - 1./810.*m11*
         me2*RSe45633p*RS34561pe - 5./324.*m11*me2*RSe45633p*RS3456pee
          - 5./324.*m11*me2*RSe45633p*RS3456ppe + 2./405.*m11*me2*
         RS345pe*RSe45613p );

      M13p6x23R +=  + Abeps(m33,mu2)*pi16 * ( 5./324.*m11*me2*RS345pe*
         RSe45633p + 2./405.*m11*me2*RS346pe*RSe45613p + 5./324.*m11*
         me2*RS346pe*RSe45633p + 2./405.*m11*me2*RS356pe*RSe45613p + 5./
         324.*m11*me2*RS356pe*RSe45633p + 2./135.*m11*me2*RSe45611p*
         RS34561pe + 1./90.*m11*me2*RSe45611p*RS3456pee + 1./90.*m11*
         me2*RSe45611p*RS3456ppe - 1./90.*m11*me2*RSe45611p*RS345pe - 1.
         /90.*m11*me2*RSe45611p*RS346pe - 1./90.*m11*me2*RSe45611p*
         RS356pe - 29./1080.*m11*m66*RS34561pe - 1./27.*m11*m66*
         RS14563pe - 1./27.*m11*m66*RSe45613p - 1./27.*m11*m66*
         RSp45613e - 31./1296.*m11*m66*RS3456pee - 31./1296.*m11*m66*
         RS3456ppe - 121./6480.*m11*m66*RS345pe + 31./1296.*m11*m66*
         RS346pe + 31./1296.*m11*m66*RS356pe - 1./54.*m11*m66*RS145pe
          + 5./81.*m11*m66*RS145pe*RS345pe + 1./162.*m11*m66*RS145pe*
         RS346pe + 1./162.*m11*m66*RS145pe*RS356pe + 1./162.*m11*m66*
         RS146pe*RS345pe + 1./81.*m11*m66*RS146pe*RS346pe + 1./162.*m11
         *m66*RS156pe*RS345pe );

      M13p6x23R +=  + Abeps(m33,mu2)*pi16 * ( 1./81.*m11*m66*RS156pe*
         RS356pe + 1./162.*m11*m66*RSe561p*RS345pe + 1./81.*m11*m66*
         RSe561p*RS356pe + 1./162.*m11*m66*RSe461p*RS345pe + 1./81.*m11
         *m66*RSe461p*RS346pe - 1./54.*m11*m66*RSe451p + 5./81.*m11*m66
         *RSe451p*RS345pe + 1./162.*m11*m66*RSe451p*RS346pe + 1./162.*
         m11*m66*RSe451p*RS356pe + 1./162.*m11*m66*RSp561e*RS345pe + 1./
         81.*m11*m66*RSp561e*RS356pe + 1./162.*m11*m66*RSp461e*RS345pe
          + 1./81.*m11*m66*RSp461e*RS346pe - 1./54.*m11*m66*RSp451e + 5.
         /81.*m11*m66*RSp451e*RS345pe + 1./162.*m11*m66*RSp451e*RS346pe
          + 1./162.*m11*m66*RSp451e*RS356pe - 29./1080.*m11*m55*
         RS34561pe - 1./27.*m11*m55*RS14563pe - 1./27.*m11*m55*
         RSe45613p - 1./27.*m11*m55*RSp45613e - 31./1296.*m11*m55*
         RS3456pee - 31./1296.*m11*m55*RS3456ppe + 31./1296.*m11*m55*
         RS345pe - 121./6480.*m11*m55*RS346pe + 31./1296.*m11*m55*
         RS356pe + 1./81.*m11*m55*RS145pe*RS345pe + 1./162.*m11*m55*
         RS145pe*RS346pe );

      M13p6x23R +=  + Abeps(m33,mu2)*pi16 * (  - 1./54.*m11*m55*RS146pe
          + 1./162.*m11*m55*RS146pe*RS345pe + 5./81.*m11*m55*RS146pe*
         RS346pe + 1./162.*m11*m55*RS146pe*RS356pe + 1./162.*m11*m55*
         RS156pe*RS346pe + 1./81.*m11*m55*RS156pe*RS356pe + 1./162.*m11
         *m55*RSe561p*RS346pe + 1./81.*m11*m55*RSe561p*RS356pe - 1./54.
         *m11*m55*RSe461p + 1./162.*m11*m55*RSe461p*RS345pe + 5./81.*
         m11*m55*RSe461p*RS346pe + 1./162.*m11*m55*RSe461p*RS356pe + 1./
         81.*m11*m55*RSe451p*RS345pe + 1./162.*m11*m55*RSe451p*RS346pe
          + 1./162.*m11*m55*RSp561e*RS346pe + 1./81.*m11*m55*RSp561e*
         RS356pe - 1./54.*m11*m55*RSp461e + 1./162.*m11*m55*RSp461e*
         RS345pe + 5./81.*m11*m55*RSp461e*RS346pe + 1./162.*m11*m55*
         RSp461e*RS356pe + 1./81.*m11*m55*RSp451e*RS345pe + 1./162.*m11
         *m55*RSp451e*RS346pe - 29./1080.*m11*m44*RS34561pe - 1./27.*
         m11*m44*RS14563pe - 1./27.*m11*m44*RSe45613p - 1./27.*m11*m44*
         RSp45613e - 31./1296.*m11*m44*RS3456pee - 31./1296.*m11*m44*
         RS3456ppe );

      M13p6x23R +=  + Abeps(m33,mu2)*pi16 * ( 31./1296.*m11*m44*RS345pe
          + 31./1296.*m11*m44*RS346pe - 121./6480.*m11*m44*RS356pe + 1./
         81.*m11*m44*RS145pe*RS345pe + 1./162.*m11*m44*RS145pe*RS356pe
          + 1./81.*m11*m44*RS146pe*RS346pe + 1./162.*m11*m44*RS146pe*
         RS356pe - 1./54.*m11*m44*RS156pe + 1./162.*m11*m44*RS156pe*
         RS345pe + 1./162.*m11*m44*RS156pe*RS346pe + 5./81.*m11*m44*
         RS156pe*RS356pe - 1./54.*m11*m44*RSe561p + 1./162.*m11*m44*
         RSe561p*RS345pe + 1./162.*m11*m44*RSe561p*RS346pe + 5./81.*m11
         *m44*RSe561p*RS356pe + 1./81.*m11*m44*RSe461p*RS346pe + 1./162.
         *m11*m44*RSe461p*RS356pe + 1./81.*m11*m44*RSe451p*RS345pe + 1./
         162.*m11*m44*RSe451p*RS356pe - 1./54.*m11*m44*RSp561e + 1./162.
         *m11*m44*RSp561e*RS345pe + 1./162.*m11*m44*RSp561e*RS346pe + 5.
         /81.*m11*m44*RSp561e*RS356pe + 1./81.*m11*m44*RSp461e*RS346pe
          + 1./162.*m11*m44*RSp461e*RS356pe + 1./81.*m11*m44*RSp451e*
         RS345pe + 1./162.*m11*m44*RSp451e*RS356pe + 13./36.*m11*m33 + 
         1./108.*m11*m33*RS34561pe );

      M13p6x23R +=  + Abeps(m33,mu2)*pi16 * ( 1./90.*m11*m33*pow(
         RS34561pe,2) - 1./9.*m11*m33*RS14563pe + 1./54.*m11*m33*
         RS14563pe*RS34561pe - 1./9.*m11*m33*RSe45613p - 5./162.*m11*
         m33*RSe45613p*RS34561pe - 1./9.*m11*m33*RSp45613e - 5./162.*
         m11*m33*RSp45613e*RS34561pe - 25./432.*m11*m33*RS3456pee - 29./
         405.*m11*m33*RS3456pee*RS34561pe - 41./540.*m11*m33*RS3456pee*
         RS14563pe - 17./324.*m11*m33*RS3456pee*RSe45613p - 17./324.*
         m11*m33*RS3456pee*RSp45613e - 1./648.*m11*m33*pow(RS3456pee,2)
          - 25./432.*m11*m33*RS3456ppe - 29./405.*m11*m33*RS3456ppe*
         RS34561pe - 41./540.*m11*m33*RS3456ppe*RS14563pe - 17./324.*
         m11*m33*RS3456ppe*RSe45613p - 17./324.*m11*m33*RS3456ppe*
         RSp45613e - 1./324.*m11*m33*RS3456ppe*RS3456pee - 1./648.*m11*
         m33*pow(RS3456ppe,2) + 5./324.*m11*m33*RSe45633p*RS34561pe + 
         11./648.*m11*m33*RSe45633p*RS3456pee + 11./648.*m11*m33*
         RSe45633p*RS3456ppe + 5./324.*m11*m33*RSp45633e*RS34561pe + 11.
         /648.*m11*m33*RSp45633e*RS3456pee );

      M13p6x23R +=  + Abeps(m33,mu2)*pi16 * ( 11./648.*m11*m33*
         RSp45633e*RS3456ppe - 43./2160.*m11*m33*RS345pe + 29./405.*m11
         *m33*RS345pe*RS34561pe + 41./540.*m11*m33*RS345pe*RS14563pe + 
         17./324.*m11*m33*RS345pe*RSe45613p + 17./324.*m11*m33*RS345pe*
         RSp45613e + 1./324.*m11*m33*RS345pe*RS3456pee + 1./324.*m11*
         m33*RS345pe*RS3456ppe - 11./648.*m11*m33*RS345pe*RSe45633p - 
         11./648.*m11*m33*RS345pe*RSp45633e - 1./648.*m11*m33*pow(
         RS345pe,2) - 43./2160.*m11*m33*RS346pe + 29./405.*m11*m33*
         RS346pe*RS34561pe + 41./540.*m11*m33*RS346pe*RS14563pe + 17./
         324.*m11*m33*RS346pe*RSe45613p + 17./324.*m11*m33*RS346pe*
         RSp45613e + 1./324.*m11*m33*RS346pe*RS3456pee + 1./324.*m11*
         m33*RS346pe*RS3456ppe - 11./648.*m11*m33*RS346pe*RSe45633p - 
         11./648.*m11*m33*RS346pe*RSp45633e - 1./324.*m11*m33*RS346pe*
         RS345pe - 1./648.*m11*m33*pow(RS346pe,2) - 43./2160.*m11*m33*
         RS356pe + 29./405.*m11*m33*RS356pe*RS34561pe + 41./540.*m11*
         m33*RS356pe*RS14563pe );

      M13p6x23R +=  + Abeps(m33,mu2)*pi16 * ( 17./324.*m11*m33*RS356pe*
         RSe45613p + 17./324.*m11*m33*RS356pe*RSp45613e + 1./324.*m11*
         m33*RS356pe*RS3456pee + 1./324.*m11*m33*RS356pe*RS3456ppe - 11.
         /648.*m11*m33*RS356pe*RSe45633p - 11./648.*m11*m33*RS356pe*
         RSp45633e - 1./324.*m11*m33*RS356pe*RS345pe - 1./324.*m11*m33*
         RS356pe*RS346pe - 1./648.*m11*m33*pow(RS356pe,2) - 17./1620.*
         m11*m33*RS1456pee*RS34561pe - 7./648.*m11*m33*RS1456pee*
         RS3456pee - 7./648.*m11*m33*RS1456pee*RS3456ppe + 7./648.*m11*
         m33*RS1456pee*RS345pe + 7./648.*m11*m33*RS1456pee*RS346pe + 7./
         648.*m11*m33*RS1456pee*RS356pe - 17./1620.*m11*m33*RS1456ppe*
         RS34561pe - 7./648.*m11*m33*RS1456ppe*RS3456pee - 7./648.*m11*
         m33*RS1456ppe*RS3456ppe + 7./648.*m11*m33*RS1456ppe*RS345pe + 
         7./648.*m11*m33*RS1456ppe*RS346pe + 7./648.*m11*m33*RS1456ppe*
         RS356pe + 5./324.*m11*m33*RSe45611p*RS34561pe - 1./648.*m11*
         m33*RSe45611p*RS3456pee - 1./648.*m11*m33*RSe45611p*RS3456ppe
          + 1./648.*m11*m33*RSe45611p*RS345pe );

      M13p6x23R +=  + Abeps(m33,mu2)*pi16 * ( 1./648.*m11*m33*RSe45611p
         *RS346pe + 1./648.*m11*m33*RSe45611p*RS356pe + 5./324.*m11*m33
         *RSp45611e*RS34561pe - 1./648.*m11*m33*RSp45611e*RS3456pee - 1.
         /648.*m11*m33*RSp45611e*RS3456ppe + 1./648.*m11*m33*RSp45611e*
         RS345pe + 1./648.*m11*m33*RSp45611e*RS346pe + 1./648.*m11*m33*
         RSp45611e*RS356pe - 1./54.*m11*m33*RS145pe + 17./1620.*m11*m33
         *RS145pe*RS34561pe + 7./648.*m11*m33*RS145pe*RS3456pee + 7./
         648.*m11*m33*RS145pe*RS3456ppe - 7./648.*m11*m33*RS145pe*
         RS345pe - 7./648.*m11*m33*RS145pe*RS346pe - 7./648.*m11*m33*
         RS145pe*RS356pe - 1./54.*m11*m33*RS146pe + 17./1620.*m11*m33*
         RS146pe*RS34561pe + 7./648.*m11*m33*RS146pe*RS3456pee + 7./648.
         *m11*m33*RS146pe*RS3456ppe - 7./648.*m11*m33*RS146pe*RS345pe
          - 7./648.*m11*m33*RS146pe*RS346pe - 7./648.*m11*m33*RS146pe*
         RS356pe - 1./54.*m11*m33*RS156pe + 17./1620.*m11*m33*RS156pe*
         RS34561pe + 7./648.*m11*m33*RS156pe*RS3456pee + 7./648.*m11*
         m33*RS156pe*RS3456ppe );

      M13p6x23R +=  + Abeps(m33,mu2)*pi16 * (  - 7./648.*m11*m33*
         RS156pe*RS345pe - 7./648.*m11*m33*RS156pe*RS346pe - 7./648.*
         m11*m33*RS156pe*RS356pe - 1./54.*m11*m33*RSe561p - 1./54.*m11*
         m33*RSe461p - 1./54.*m11*m33*RSe451p - 1./54.*m11*m33*RSp561e
          - 1./54.*m11*m33*RSp461e - 1./54.*m11*m33*RSp451e + 5./48.*
         pow(m11,2) - 157./2160.*pow(m11,2)*RS34561pe - 2./81.*pow(
         m11,2)*pow(RS34561pe,2) + 1./81.*pow(m11,2)*RS14563pe*
         RS34561pe - 2./81.*pow(m11,2)*RSe45613p*RS34561pe - 2./81.*
         pow(m11,2)*RSp45613e*RS34561pe - 23./432.*pow(m11,2)*RS3456pee
          - 1./54.*pow(m11,2)*RS3456pee*RS34561pe - 1./60.*pow(m11,2)*
         RS3456pee*RS14563pe - 1./108.*pow(m11,2)*RS3456pee*RSe45613p
          - 1./108.*pow(m11,2)*RS3456pee*RSp45613e - 5./648.*pow(m11,2)
         *pow(RS3456pee,2) - 23./432.*pow(m11,2)*RS3456ppe - 1./54.*
         pow(m11,2)*RS3456ppe*RS34561pe - 1./60.*pow(m11,2)*RS3456ppe*
         RS14563pe - 1./108.*pow(m11,2)*RS3456ppe*RSe45613p - 1./108.*
         pow(m11,2)*RS3456ppe*RSp45613e );

      M13p6x23R +=  + Abeps(m33,mu2)*pi16 * (  - 5./324.*pow(m11,2)*
         RS3456ppe*RS3456pee - 5./648.*pow(m11,2)*pow(RS3456ppe,2) + 1./
         108.*pow(m11,2)*RSe45633p*RS34561pe + 5./648.*pow(m11,2)*
         RSe45633p*RS3456pee + 5./648.*pow(m11,2)*RSe45633p*RS3456ppe
          + 1./108.*pow(m11,2)*RSp45633e*RS34561pe + 5./648.*pow(m11,2)
         *RSp45633e*RS3456pee + 5./648.*pow(m11,2)*RSp45633e*RS3456ppe
          + 1./54.*pow(m11,2)*RS345pe + 1./54.*pow(m11,2)*RS345pe*
         RS34561pe + 1./60.*pow(m11,2)*RS345pe*RS14563pe + 1./108.*pow(
         m11,2)*RS345pe*RSe45613p + 1./108.*pow(m11,2)*RS345pe*
         RSp45613e + 5./324.*pow(m11,2)*RS345pe*RS3456pee + 5./324.*
         pow(m11,2)*RS345pe*RS3456ppe - 5./648.*pow(m11,2)*RS345pe*
         RSe45633p - 5./648.*pow(m11,2)*RS345pe*RSp45633e - 5./648.*
         pow(m11,2)*pow(RS345pe,2) + 1./54.*pow(m11,2)*RS346pe + 1./54.
         *pow(m11,2)*RS346pe*RS34561pe + 1./60.*pow(m11,2)*RS346pe*
         RS14563pe + 1./108.*pow(m11,2)*RS346pe*RSe45613p + 1./108.*
         pow(m11,2)*RS346pe*RSp45613e );

      M13p6x23R +=  + Abeps(m33,mu2)*pi16 * ( 5./324.*pow(m11,2)*
         RS346pe*RS3456pee + 5./324.*pow(m11,2)*RS346pe*RS3456ppe - 5./
         648.*pow(m11,2)*RS346pe*RSe45633p - 5./648.*pow(m11,2)*RS346pe
         *RSp45633e - 5./324.*pow(m11,2)*RS346pe*RS345pe - 5./648.*pow(
         m11,2)*pow(RS346pe,2) + 1./54.*pow(m11,2)*RS356pe + 1./54.*
         pow(m11,2)*RS356pe*RS34561pe + 1./60.*pow(m11,2)*RS356pe*
         RS14563pe + 1./108.*pow(m11,2)*RS356pe*RSe45613p + 1./108.*
         pow(m11,2)*RS356pe*RSp45613e + 5./324.*pow(m11,2)*RS356pe*
         RS3456pee + 5./324.*pow(m11,2)*RS356pe*RS3456ppe - 5./648.*
         pow(m11,2)*RS356pe*RSe45633p - 5./648.*pow(m11,2)*RS356pe*
         RSp45633e - 5./324.*pow(m11,2)*RS356pe*RS345pe - 5./324.*pow(
         m11,2)*RS356pe*RS346pe - 5./648.*pow(m11,2)*pow(RS356pe,2) - 
         53./1620.*pow(m11,2)*RS1456pee*RS34561pe - 1./72.*pow(m11,2)*
         RS1456pee*RS3456pee - 1./72.*pow(m11,2)*RS1456pee*RS3456ppe + 
         1./72.*pow(m11,2)*RS1456pee*RS345pe + 1./72.*pow(m11,2)*
         RS1456pee*RS346pe );

      M13p6x23R +=  + Abeps(m33,mu2)*pi16 * ( 1./72.*pow(m11,2)*
         RS1456pee*RS356pe - 53./1620.*pow(m11,2)*RS1456ppe*RS34561pe
          - 1./72.*pow(m11,2)*RS1456ppe*RS3456pee - 1./72.*pow(m11,2)*
         RS1456ppe*RS3456ppe + 1./72.*pow(m11,2)*RS1456ppe*RS345pe + 1./
         72.*pow(m11,2)*RS1456ppe*RS346pe + 1./72.*pow(m11,2)*RS1456ppe
         *RS356pe + 5./324.*pow(m11,2)*RSe45611p*RS34561pe + 1./648.*
         pow(m11,2)*RSe45611p*RS3456pee + 1./648.*pow(m11,2)*RSe45611p*
         RS3456ppe - 1./648.*pow(m11,2)*RSe45611p*RS345pe - 1./648.*
         pow(m11,2)*RSe45611p*RS346pe - 1./648.*pow(m11,2)*RSe45611p*
         RS356pe + 5./324.*pow(m11,2)*RSp45611e*RS34561pe + 1./648.*
         pow(m11,2)*RSp45611e*RS3456pee + 1./648.*pow(m11,2)*RSp45611e*
         RS3456ppe - 1./648.*pow(m11,2)*RSp45611e*RS345pe - 1./648.*
         pow(m11,2)*RSp45611e*RS346pe - 1./648.*pow(m11,2)*RSp45611e*
         RS356pe + 53./1620.*pow(m11,2)*RS145pe*RS34561pe + 1./72.*pow(
         m11,2)*RS145pe*RS3456pee + 1./72.*pow(m11,2)*RS145pe*RS3456ppe
          - 1./72.*pow(m11,2)*RS145pe*RS345pe );

      M13p6x23R +=  + Abeps(m33,mu2)*pi16 * (  - 1./72.*pow(m11,2)*
         RS145pe*RS346pe - 1./72.*pow(m11,2)*RS145pe*RS356pe + 53./1620.
         *pow(m11,2)*RS146pe*RS34561pe + 1./72.*pow(m11,2)*RS146pe*
         RS3456pee + 1./72.*pow(m11,2)*RS146pe*RS3456ppe - 1./72.*pow(
         m11,2)*RS146pe*RS345pe - 1./72.*pow(m11,2)*RS146pe*RS346pe - 1.
         /72.*pow(m11,2)*RS146pe*RS356pe + 53./1620.*pow(m11,2)*RS156pe
         *RS34561pe + 1./72.*pow(m11,2)*RS156pe*RS3456pee + 1./72.*pow(
         m11,2)*RS156pe*RS3456ppe - 1./72.*pow(m11,2)*RS156pe*RS345pe
          - 1./72.*pow(m11,2)*RS156pe*RS346pe - 1./72.*pow(m11,2)*
         RS156pe*RS356pe );

      M13p6x23R +=  + Abeps(m34,mu2)*pi16 * ( 17./270.*pow(mp2,2)*
         RSp45613e - 2./45.*pow(mp2,2)*RSp45633e + 7./270.*pow(mp2,2)*
         RSp563e + 1./270.*pow(mp2,2)*RSp45611e - 19./270.*pow(mp2,2)*
         RSp561e + 1./45.*pow(mp2,2)*RSp564e + 17./270.*pow(me2,2)*
         RSe45613p - 2./45.*pow(me2,2)*RSe45633p + 7./270.*pow(me2,2)*
         RSe563p + 1./270.*pow(me2,2)*RSe45611p - 19./270.*pow(me2,2)*
         RSe561p + 1./45.*pow(me2,2)*RSe564p - 1./216.*m44*RD3456ep - 1.
         /216.*m44*RD1456ep - 1./108.*m44*mp2*RSp45613e - 1./216.*m44*
         mp2*RSp45633e + 2./135.*m44*mp2*RSp563e - 1./216.*m44*mp2*
         RSp45611e + 2./45.*m44*mp2*RSp561e - 11./270.*m44*mp2*RSp564e
          - 1./108.*m44*me2*RSe45613p - 1./216.*m44*me2*RSe45633p + 2./
         135.*m44*me2*RSe563p - 1./216.*m44*me2*RSe45611p + 2./45.*m44*
         me2*RSe561p - 11./270.*m44*me2*RSe564p - 1./108.*pow(m44,2)*
         RS34561pe - 1./108.*pow(m44,2)*RS14563pe - 1./108.*pow(m44,2)*
         RSe45613p - 1./108.*pow(m44,2)*RSp45613e + 1./216.*pow(m44,2)*
         RS3456pee );

      M13p6x23R +=  + Abeps(m34,mu2)*pi16 * ( 1./216.*pow(m44,2)*
         RS3456ppe - 1./216.*pow(m44,2)*RSe45633p - 1./216.*pow(m44,2)*
         RSp45633e - 1./216.*pow(m44,2)*RS345pe - 1./216.*pow(m44,2)*
         RS346pe - 1./216.*pow(m44,2)*RS356pe + 1./216.*pow(m44,2)*
         RS1456pee + 1./216.*pow(m44,2)*RS1456ppe - 1./216.*pow(m44,2)*
         RSe45611p - 1./216.*pow(m44,2)*RSp45611e - 1./216.*pow(m44,2)*
         RS145pe - 1./216.*pow(m44,2)*RS146pe - 1./216.*pow(m44,2)*
         RS156pe + 1./54.*pow(m44,2)*RSe564p + 1./54.*pow(m44,2)*
         RSp564e - 2./45.*m33*RD3456ep + 1./24.*m33*RD1456ep + 1./1080.
         *m33*mp2*RSp45613e + 2./45.*m33*mp2*RSp45633e - 1./120.*m33*
         mp2*RSp563e + 1./24.*m33*mp2*RSp45611e - 23./216.*m33*mp2*
         RSp561e + 1./1080.*m33*me2*RSe45613p + 2./45.*m33*me2*
         RSe45633p - 1./120.*m33*me2*RSe563p + 1./24.*m33*me2*RSe45611p
          - 23./216.*m33*me2*RSe561p - 31./432.*m33*m66 - 31./432.*m33*
         m55 - 5./54.*m33*m44 - 13./432.*m33*m44*RS34561pe - 1./48.*m33
         *m44*RS14563pe );

      M13p6x23R +=  + Abeps(m34,mu2)*pi16 * (  - 1./48.*m33*m44*
         RSe45613p - 1./48.*m33*m44*RSp45613e + 1./216.*m33*m44*
         RSe45633p + 1./216.*m33*m44*RSp45633e + 17./2160.*m33*m44*
         RS356pe - 1./144.*m33*m44*RSe563p - 1./144.*m33*m44*RSp563e + 
         1./216.*m33*m44*RS1456pee + 1./216.*m33*m44*RS1456ppe - 1./216.
         *m33*m44*RSe45611p - 1./216.*m33*m44*RSp45611e - 1./216.*m33*
         m44*RS145pe - 1./216.*m33*m44*RS146pe - 29./432.*m33*m44*
         RS156pe + 7./432.*m33*m44*RS456pe - 1./16.*m33*m44*RSe561p - 1.
         /16.*m33*m44*RSp561e + 7./432.*m33*m44*RSe564p + 7./432.*m33*
         m44*RSp564e + 1./144.*pow(m33,2) + 23./360.*pow(m33,2)*
         RS34561pe + 19./1080.*pow(m33,2)*RS356pe - 4./27.*pow(m33,2)*
         RS156pe - 4./27.*pow(m33,2)*RSe561p - 4./27.*pow(m33,2)*
         RSp561e - 1./216.*m11*RD3456ep + 11./270.*m11*RD1456ep + 11./
         1080.*m11*mp2*RSp45613e - 1./216.*m11*mp2*RSp45633e - 5./216.*
         m11*mp2*RSp563e + 1./30.*m11*mp2*RSp45611e - 1./40.*m11*mp2*
         RSp561e );

      M13p6x23R +=  + Abeps(m34,mu2)*pi16 * (  - 1./54.*m11*mp2*RSp564e
          + 11./1080.*m11*me2*RSe45613p - 1./216.*m11*me2*RSe45633p - 5.
         /216.*m11*me2*RSe563p + 1./30.*m11*me2*RSe45611p - 1./40.*m11*
         me2*RSe561p - 1./54.*m11*me2*RSe564p - 31./432.*m11*m66 - 31./
         432.*m11*m55 - 5./54.*m11*m44 - 13./432.*m11*m44*RS34561pe - 
         17./432.*m11*m44*RS14563pe - 13./432.*m11*m44*RSe45613p - 13./
         432.*m11*m44*RSp45613e + 1./108.*m11*m44*RS3456pee + 1./108.*
         m11*m44*RS3456ppe - 1./108.*m11*m44*RSe45633p - 1./108.*m11*
         m44*RSp45633e - 1./108.*m11*m44*RS345pe - 1./108.*m11*m44*
         RS346pe + 1./432.*m11*m44*RS356pe + 5./432.*m11*m44*RSe563p + 
         5./432.*m11*m44*RSp563e + 1./216.*m11*m44*RS1456pee + 1./216.*
         m11*m44*RS1456ppe - 1./216.*m11*m44*RS145pe - 1./216.*m11*m44*
         RS146pe - 1./240.*m11*m44*RS156pe + 7./432.*m11*m44*RS456pe - 
         19./432.*m11*m44*RSe561p - 19./432.*m11*m44*RSp561e + 5./144.*
         m11*m44*RSe564p + 5./144.*m11*m44*RSp564e - 11./72.*m11*m33 - 
         23./2160.*m11*m33*RS34561pe );

      M13p6x23R +=  + Abeps(m34,mu2)*pi16 * (  - 43./2160.*m11*m33*
         RS14563pe - 1./48.*m11*m33*RSe45613p - 1./48.*m11*m33*
         RSp45613e + 1./216.*m11*m33*RSe45633p + 1./216.*m11*m33*
         RSp45633e - 13./432.*m11*m33*RS356pe - 1./144.*m11*m33*RSe563p
          - 1./144.*m11*m33*RSp563e - 1./27.*m11*m33*RS1456pee - 1./27.
         *m11*m33*RS1456ppe - 1./216.*m11*m33*RSe45611p - 1./216.*m11*
         m33*RSp45611e + 1./27.*m11*m33*RS145pe + 1./27.*m11*m33*
         RS146pe - 19./144.*m11*m33*RS156pe + 7./432.*m11*m33*RS456pe
          - 1./16.*m11*m33*RSe561p - 1./16.*m11*m33*RSp561e + 7./432.*
         m11*m33*RSe564p + 7./432.*m11*m33*RSp564e - 23./144.*pow(
         m11,2) - 1./48.*pow(m11,2)*RS34561pe + 113./2160.*pow(m11,2)*
         RS14563pe - 1./48.*pow(m11,2)*RSe45613p - 1./48.*pow(m11,2)*
         RSp45613e + 1./216.*pow(m11,2)*RS3456pee + 1./216.*pow(m11,2)*
         RS3456ppe - 1./216.*pow(m11,2)*RSe45633p - 1./216.*pow(m11,2)*
         RSp45633e - 1./216.*pow(m11,2)*RS345pe - 1./216.*pow(m11,2)*
         RS346pe );

      M13p6x23R +=  + Abeps(m34,mu2)*pi16 * ( 1./144.*pow(m11,2)*
         RS356pe + 5./432.*pow(m11,2)*RSe563p + 5./432.*pow(m11,2)*
         RSp563e - 1./24.*pow(m11,2)*RS1456pee - 1./24.*pow(m11,2)*
         RS1456ppe + 1./216.*pow(m11,2)*RSe45611p + 1./216.*pow(m11,2)*
         RSp45611e + 1./24.*pow(m11,2)*RS145pe + 1./24.*pow(m11,2)*
         RS146pe - 131./2160.*pow(m11,2)*RS156pe + 7./432.*pow(m11,2)*
         RS456pe - 1./144.*pow(m11,2)*RSe561p - 1./144.*pow(m11,2)*
         RSp561e + 7./432.*pow(m11,2)*RSe564p + 7./432.*pow(m11,2)*
         RSp564e );

      M13p6x23R +=  + Abeps(m35,mu2)*pi16 * ( 17./270.*pow(mp2,2)*
         RSp45613e - 2./45.*pow(mp2,2)*RSp45633e + 7./270.*pow(mp2,2)*
         RSp463e + 1./270.*pow(mp2,2)*RSp45611e - 19./270.*pow(mp2,2)*
         RSp461e + 1./45.*pow(mp2,2)*RSp465e + 17./270.*pow(me2,2)*
         RSe45613p - 2./45.*pow(me2,2)*RSe45633p + 7./270.*pow(me2,2)*
         RSe463p + 1./270.*pow(me2,2)*RSe45611p - 19./270.*pow(me2,2)*
         RSe461p + 1./45.*pow(me2,2)*RSe465p - 1./216.*m55*RD3456ep - 1.
         /216.*m55*RD1456ep - 1./108.*m55*mp2*RSp45613e - 1./216.*m55*
         mp2*RSp45633e + 2./135.*m55*mp2*RSp463e - 1./216.*m55*mp2*
         RSp45611e + 2./45.*m55*mp2*RSp461e - 11./270.*m55*mp2*RSp465e
          - 1./108.*m55*me2*RSe45613p - 1./216.*m55*me2*RSe45633p + 2./
         135.*m55*me2*RSe463p - 1./216.*m55*me2*RSe45611p + 2./45.*m55*
         me2*RSe461p - 11./270.*m55*me2*RSe465p - 1./108.*pow(m55,2)*
         RS34561pe - 1./108.*pow(m55,2)*RS14563pe - 1./108.*pow(m55,2)*
         RSe45613p - 1./108.*pow(m55,2)*RSp45613e + 1./216.*pow(m55,2)*
         RS3456pee );

      M13p6x23R +=  + Abeps(m35,mu2)*pi16 * ( 1./216.*pow(m55,2)*
         RS3456ppe - 1./216.*pow(m55,2)*RSe45633p - 1./216.*pow(m55,2)*
         RSp45633e - 1./216.*pow(m55,2)*RS345pe - 1./216.*pow(m55,2)*
         RS346pe - 1./216.*pow(m55,2)*RS356pe + 1./216.*pow(m55,2)*
         RS1456pee + 1./216.*pow(m55,2)*RS1456ppe - 1./216.*pow(m55,2)*
         RSe45611p - 1./216.*pow(m55,2)*RSp45611e - 1./216.*pow(m55,2)*
         RS145pe - 1./216.*pow(m55,2)*RS146pe - 1./216.*pow(m55,2)*
         RS156pe + 1./54.*pow(m55,2)*RSe465p + 1./54.*pow(m55,2)*
         RSp465e - 2./45.*m33*RD3456ep + 1./24.*m33*RD1456ep + 1./1080.
         *m33*mp2*RSp45613e + 2./45.*m33*mp2*RSp45633e - 1./120.*m33*
         mp2*RSp463e + 1./24.*m33*mp2*RSp45611e - 23./216.*m33*mp2*
         RSp461e + 1./1080.*m33*me2*RSe45613p + 2./45.*m33*me2*
         RSe45633p - 1./120.*m33*me2*RSe463p + 1./24.*m33*me2*RSe45611p
          - 23./216.*m33*me2*RSe461p - 31./432.*m33*m66 - 5./54.*m33*
         m55 - 13./432.*m33*m55*RS34561pe - 1./48.*m33*m55*RS14563pe - 
         1./48.*m33*m55*RSe45613p );

      M13p6x23R +=  + Abeps(m35,mu2)*pi16 * (  - 1./48.*m33*m55*
         RSp45613e + 1./216.*m33*m55*RSe45633p + 1./216.*m33*m55*
         RSp45633e + 17./2160.*m33*m55*RS346pe - 1./144.*m33*m55*
         RSe463p - 1./144.*m33*m55*RSp463e + 1./216.*m33*m55*RS1456pee
          + 1./216.*m33*m55*RS1456ppe - 1./216.*m33*m55*RSe45611p - 1./
         216.*m33*m55*RSp45611e - 1./216.*m33*m55*RS145pe - 29./432.*
         m33*m55*RS146pe - 1./216.*m33*m55*RS156pe + 7./432.*m33*m55*
         RS546pe - 1./16.*m33*m55*RSe461p - 1./16.*m33*m55*RSp461e + 7./
         432.*m33*m55*RSe465p + 7./432.*m33*m55*RSp465e - 31./432.*m33*
         m44 + 1./144.*pow(m33,2) + 23./360.*pow(m33,2)*RS34561pe + 19./
         1080.*pow(m33,2)*RS346pe - 4./27.*pow(m33,2)*RS146pe - 4./27.*
         pow(m33,2)*RSe461p - 4./27.*pow(m33,2)*RSp461e - 1./216.*m11*
         RD3456ep + 11./270.*m11*RD1456ep + 11./1080.*m11*mp2*RSp45613e
          - 1./216.*m11*mp2*RSp45633e - 5./216.*m11*mp2*RSp463e + 1./30.
         *m11*mp2*RSp45611e - 1./40.*m11*mp2*RSp461e - 1./54.*m11*mp2*
         RSp465e );

      M13p6x23R +=  + Abeps(m35,mu2)*pi16 * ( 11./1080.*m11*me2*
         RSe45613p - 1./216.*m11*me2*RSe45633p - 5./216.*m11*me2*
         RSe463p + 1./30.*m11*me2*RSe45611p - 1./40.*m11*me2*RSe461p - 
         1./54.*m11*me2*RSe465p - 31./432.*m11*m66 - 5./54.*m11*m55 - 
         13./432.*m11*m55*RS34561pe - 17./432.*m11*m55*RS14563pe - 13./
         432.*m11*m55*RSe45613p - 13./432.*m11*m55*RSp45613e + 1./108.*
         m11*m55*RS3456pee + 1./108.*m11*m55*RS3456ppe - 1./108.*m11*
         m55*RSe45633p - 1./108.*m11*m55*RSp45633e - 1./108.*m11*m55*
         RS345pe + 1./432.*m11*m55*RS346pe - 1./108.*m11*m55*RS356pe + 
         5./432.*m11*m55*RSe463p + 5./432.*m11*m55*RSp463e + 1./216.*
         m11*m55*RS1456pee + 1./216.*m11*m55*RS1456ppe - 1./216.*m11*
         m55*RS145pe - 1./240.*m11*m55*RS146pe - 1./216.*m11*m55*
         RS156pe + 7./432.*m11*m55*RS546pe - 19./432.*m11*m55*RSe461p
          - 19./432.*m11*m55*RSp461e + 5./144.*m11*m55*RSe465p + 5./144.
         *m11*m55*RSp465e - 31./432.*m11*m44 - 11./72.*m11*m33 - 23./
         2160.*m11*m33*RS34561pe );

      M13p6x23R +=  + Abeps(m35,mu2)*pi16 * (  - 43./2160.*m11*m33*
         RS14563pe - 1./48.*m11*m33*RSe45613p - 1./48.*m11*m33*
         RSp45613e + 1./216.*m11*m33*RSe45633p + 1./216.*m11*m33*
         RSp45633e - 13./432.*m11*m33*RS346pe - 1./144.*m11*m33*RSe463p
          - 1./144.*m11*m33*RSp463e - 1./27.*m11*m33*RS1456pee - 1./27.
         *m11*m33*RS1456ppe - 1./216.*m11*m33*RSe45611p - 1./216.*m11*
         m33*RSp45611e + 1./27.*m11*m33*RS145pe - 19./144.*m11*m33*
         RS146pe + 1./27.*m11*m33*RS156pe + 7./432.*m11*m33*RS546pe - 1.
         /16.*m11*m33*RSe461p - 1./16.*m11*m33*RSp461e + 7./432.*m11*
         m33*RSe465p + 7./432.*m11*m33*RSp465e - 23./144.*pow(m11,2) - 
         1./48.*pow(m11,2)*RS34561pe + 113./2160.*pow(m11,2)*RS14563pe
          - 1./48.*pow(m11,2)*RSe45613p - 1./48.*pow(m11,2)*RSp45613e
          + 1./216.*pow(m11,2)*RS3456pee + 1./216.*pow(m11,2)*RS3456ppe
          - 1./216.*pow(m11,2)*RSe45633p - 1./216.*pow(m11,2)*RSp45633e
          - 1./216.*pow(m11,2)*RS345pe + 1./144.*pow(m11,2)*RS346pe - 1.
         /216.*pow(m11,2)*RS356pe );

      M13p6x23R +=  + Abeps(m35,mu2)*pi16 * ( 5./432.*pow(m11,2)*
         RSe463p + 5./432.*pow(m11,2)*RSp463e - 1./24.*pow(m11,2)*
         RS1456pee - 1./24.*pow(m11,2)*RS1456ppe + 1./216.*pow(m11,2)*
         RSe45611p + 1./216.*pow(m11,2)*RSp45611e + 1./24.*pow(m11,2)*
         RS145pe - 131./2160.*pow(m11,2)*RS146pe + 1./24.*pow(m11,2)*
         RS156pe + 7./432.*pow(m11,2)*RS546pe - 1./144.*pow(m11,2)*
         RSe461p - 1./144.*pow(m11,2)*RSp461e + 7./432.*pow(m11,2)*
         RSe465p + 7./432.*pow(m11,2)*RSp465e );

      M13p6x23R +=  + Abeps(m36,mu2)*pi16 * ( 17./270.*pow(mp2,2)*
         RSp45613e - 2./45.*pow(mp2,2)*RSp45633e + 7./270.*pow(mp2,2)*
         RSp453e + 1./270.*pow(mp2,2)*RSp45611e - 19./270.*pow(mp2,2)*
         RSp451e + 1./45.*pow(mp2,2)*RSp456e + 17./270.*pow(me2,2)*
         RSe45613p - 2./45.*pow(me2,2)*RSe45633p + 7./270.*pow(me2,2)*
         RSe453p + 1./270.*pow(me2,2)*RSe45611p - 19./270.*pow(me2,2)*
         RSe451p + 1./45.*pow(me2,2)*RSe456p - 1./216.*m66*RD3456ep - 1.
         /216.*m66*RD1456ep - 1./108.*m66*mp2*RSp45613e - 1./216.*m66*
         mp2*RSp45633e + 2./135.*m66*mp2*RSp453e - 1./216.*m66*mp2*
         RSp45611e + 2./45.*m66*mp2*RSp451e - 11./270.*m66*mp2*RSp456e
          - 1./108.*m66*me2*RSe45613p - 1./216.*m66*me2*RSe45633p + 2./
         135.*m66*me2*RSe453p - 1./216.*m66*me2*RSe45611p + 2./45.*m66*
         me2*RSe451p - 11./270.*m66*me2*RSe456p - 1./108.*pow(m66,2)*
         RS34561pe - 1./108.*pow(m66,2)*RS14563pe - 1./108.*pow(m66,2)*
         RSe45613p - 1./108.*pow(m66,2)*RSp45613e + 1./216.*pow(m66,2)*
         RS3456pee );

      M13p6x23R +=  + Abeps(m36,mu2)*pi16 * ( 1./216.*pow(m66,2)*
         RS3456ppe - 1./216.*pow(m66,2)*RSe45633p - 1./216.*pow(m66,2)*
         RSp45633e - 1./216.*pow(m66,2)*RS345pe - 1./216.*pow(m66,2)*
         RS346pe - 1./216.*pow(m66,2)*RS356pe + 1./216.*pow(m66,2)*
         RS1456pee + 1./216.*pow(m66,2)*RS1456ppe - 1./216.*pow(m66,2)*
         RSe45611p - 1./216.*pow(m66,2)*RSp45611e - 1./216.*pow(m66,2)*
         RS145pe - 1./216.*pow(m66,2)*RS146pe - 1./216.*pow(m66,2)*
         RS156pe + 1./54.*pow(m66,2)*RSe456p + 1./54.*pow(m66,2)*
         RSp456e - 2./45.*m33*RD3456ep + 1./24.*m33*RD1456ep + 1./1080.
         *m33*mp2*RSp45613e + 2./45.*m33*mp2*RSp45633e - 1./120.*m33*
         mp2*RSp453e + 1./24.*m33*mp2*RSp45611e - 23./216.*m33*mp2*
         RSp451e + 1./1080.*m33*me2*RSe45613p + 2./45.*m33*me2*
         RSe45633p - 1./120.*m33*me2*RSe453p + 1./24.*m33*me2*RSe45611p
          - 23./216.*m33*me2*RSe451p - 5./54.*m33*m66 - 13./432.*m33*
         m66*RS34561pe - 1./48.*m33*m66*RS14563pe - 1./48.*m33*m66*
         RSe45613p );

      M13p6x23R +=  + Abeps(m36,mu2)*pi16 * (  - 1./48.*m33*m66*
         RSp45613e + 1./216.*m33*m66*RSe45633p + 1./216.*m33*m66*
         RSp45633e + 17./2160.*m33*m66*RS345pe - 1./144.*m33*m66*
         RSe453p - 1./144.*m33*m66*RSp453e + 1./216.*m33*m66*RS1456pee
          + 1./216.*m33*m66*RS1456ppe - 1./216.*m33*m66*RSe45611p - 1./
         216.*m33*m66*RSp45611e - 29./432.*m33*m66*RS145pe - 1./216.*
         m33*m66*RS146pe - 1./216.*m33*m66*RS156pe + 7./432.*m33*m66*
         RS645pe - 1./16.*m33*m66*RSe451p - 1./16.*m33*m66*RSp451e + 7./
         432.*m33*m66*RSe456p + 7./432.*m33*m66*RSp456e - 31./432.*m33*
         m55 - 31./432.*m33*m44 + 1./144.*pow(m33,2) + 23./360.*pow(
         m33,2)*RS34561pe + 19./1080.*pow(m33,2)*RS345pe - 4./27.*pow(
         m33,2)*RS145pe - 4./27.*pow(m33,2)*RSe451p - 4./27.*pow(m33,2)
         *RSp451e - 1./216.*m11*RD3456ep + 11./270.*m11*RD1456ep + 11./
         1080.*m11*mp2*RSp45613e - 1./216.*m11*mp2*RSp45633e - 5./216.*
         m11*mp2*RSp453e + 1./30.*m11*mp2*RSp45611e - 1./40.*m11*mp2*
         RSp451e );

      M13p6x23R +=  + Abeps(m36,mu2)*pi16 * (  - 1./54.*m11*mp2*RSp456e
          + 11./1080.*m11*me2*RSe45613p - 1./216.*m11*me2*RSe45633p - 5.
         /216.*m11*me2*RSe453p + 1./30.*m11*me2*RSe45611p - 1./40.*m11*
         me2*RSe451p - 1./54.*m11*me2*RSe456p - 5./54.*m11*m66 - 13./
         432.*m11*m66*RS34561pe - 17./432.*m11*m66*RS14563pe - 13./432.
         *m11*m66*RSe45613p - 13./432.*m11*m66*RSp45613e + 1./108.*m11*
         m66*RS3456pee + 1./108.*m11*m66*RS3456ppe - 1./108.*m11*m66*
         RSe45633p - 1./108.*m11*m66*RSp45633e + 1./432.*m11*m66*
         RS345pe - 1./108.*m11*m66*RS346pe - 1./108.*m11*m66*RS356pe + 
         5./432.*m11*m66*RSe453p + 5./432.*m11*m66*RSp453e + 1./216.*
         m11*m66*RS1456pee + 1./216.*m11*m66*RS1456ppe - 1./240.*m11*
         m66*RS145pe - 1./216.*m11*m66*RS146pe - 1./216.*m11*m66*
         RS156pe + 7./432.*m11*m66*RS645pe - 19./432.*m11*m66*RSe451p
          - 19./432.*m11*m66*RSp451e + 5./144.*m11*m66*RSe456p + 5./144.
         *m11*m66*RSp456e - 31./432.*m11*m55 - 31./432.*m11*m44 - 11./
         72.*m11*m33 );

      M13p6x23R +=  + Abeps(m36,mu2)*pi16 * (  - 23./2160.*m11*m33*
         RS34561pe - 43./2160.*m11*m33*RS14563pe - 1./48.*m11*m33*
         RSe45613p - 1./48.*m11*m33*RSp45613e + 1./216.*m11*m33*
         RSe45633p + 1./216.*m11*m33*RSp45633e - 13./432.*m11*m33*
         RS345pe - 1./144.*m11*m33*RSe453p - 1./144.*m11*m33*RSp453e - 
         1./27.*m11*m33*RS1456pee - 1./27.*m11*m33*RS1456ppe - 1./216.*
         m11*m33*RSe45611p - 1./216.*m11*m33*RSp45611e - 19./144.*m11*
         m33*RS145pe + 1./27.*m11*m33*RS146pe + 1./27.*m11*m33*RS156pe
          + 7./432.*m11*m33*RS645pe - 1./16.*m11*m33*RSe451p - 1./16.*
         m11*m33*RSp451e + 7./432.*m11*m33*RSe456p + 7./432.*m11*m33*
         RSp456e - 23./144.*pow(m11,2) - 1./48.*pow(m11,2)*RS34561pe + 
         113./2160.*pow(m11,2)*RS14563pe - 1./48.*pow(m11,2)*RSe45613p
          - 1./48.*pow(m11,2)*RSp45613e + 1./216.*pow(m11,2)*RS3456pee
          + 1./216.*pow(m11,2)*RS3456ppe - 1./216.*pow(m11,2)*RSe45633p
          - 1./216.*pow(m11,2)*RSp45633e + 1./144.*pow(m11,2)*RS345pe
          - 1./216.*pow(m11,2)*RS346pe );

      M13p6x23R +=  + Abeps(m36,mu2)*pi16 * (  - 1./216.*pow(m11,2)*
         RS356pe + 5./432.*pow(m11,2)*RSe453p + 5./432.*pow(m11,2)*
         RSp453e - 1./24.*pow(m11,2)*RS1456pee - 1./24.*pow(m11,2)*
         RS1456ppe + 1./216.*pow(m11,2)*RSe45611p + 1./216.*pow(m11,2)*
         RSp45611e - 131./2160.*pow(m11,2)*RS145pe + 1./24.*pow(m11,2)*
         RS146pe + 1./24.*pow(m11,2)*RS156pe + 7./432.*pow(m11,2)*
         RS645pe - 1./144.*pow(m11,2)*RSe451p - 1./144.*pow(m11,2)*
         RSp451e + 7./432.*pow(m11,2)*RSe456p + 7./432.*pow(m11,2)*
         RSp456e );

      M13p6x23R +=  + Abeps(m44,mu2)*pi16 * (  - 1./27.*m44*mp2*RSp563e
         *RS356pe - 1./27.*m44*mp2*RSp563e*RSe563p - 1./27.*m44*mp2*
         pow(RSp563e,2) + 1./27.*m44*mp2*RS156pe*RSp563e + 1./81.*m44*
         mp2*RS456pe*RSp563e*RS356pe + 1./81.*m44*mp2*RS456pe*RSp563e*
         RSe563p + 1./81.*m44*mp2*RS456pe*pow(RSp563e,2) - 1./81.*m44*
         mp2*RS456pe*RS156pe*RSp563e + 1./27.*m44*mp2*RSe561p*RSp563e
          - 1./81.*m44*mp2*RSe561p*RS456pe*RSp563e + 1./27.*m44*mp2*
         RSp561e*RS356pe + 1./27.*m44*mp2*RSp561e*RSe563p + 2./27.*m44*
         mp2*RSp561e*RSp563e - 1./27.*m44*mp2*RSp561e*RS156pe - 1./81.*
         m44*mp2*RSp561e*RS456pe*RS356pe - 1./81.*m44*mp2*RSp561e*
         RS456pe*RSe563p - 2./81.*m44*mp2*RSp561e*RS456pe*RSp563e + 1./
         81.*m44*mp2*RSp561e*RS456pe*RS156pe - 1./27.*m44*mp2*RSp561e*
         RSe561p + 1./81.*m44*mp2*RSp561e*RSe561p*RS456pe - 1./27.*m44*
         mp2*pow(RSp561e,2) + 1./81.*m44*mp2*pow(RSp561e,2)*RS456pe - 1.
         /27.*m44*me2*RSe563p*RS356pe - 1./27.*m44*me2*pow(RSe563p,2)
          - 1./27.*m44*me2*RSp563e*RSe563p );

      M13p6x23R +=  + Abeps(m44,mu2)*pi16 * ( 1./27.*m44*me2*RS156pe*
         RSe563p + 1./81.*m44*me2*RS456pe*RSe563p*RS356pe + 1./81.*m44*
         me2*RS456pe*pow(RSe563p,2) + 1./81.*m44*me2*RS456pe*RSp563e*
         RSe563p - 1./81.*m44*me2*RS456pe*RS156pe*RSe563p + 1./27.*m44*
         me2*RSe561p*RS356pe + 2./27.*m44*me2*RSe561p*RSe563p + 1./27.*
         m44*me2*RSe561p*RSp563e - 1./27.*m44*me2*RSe561p*RS156pe - 1./
         81.*m44*me2*RSe561p*RS456pe*RS356pe - 2./81.*m44*me2*RSe561p*
         RS456pe*RSe563p - 1./81.*m44*me2*RSe561p*RS456pe*RSp563e + 1./
         81.*m44*me2*RSe561p*RS456pe*RS156pe - 1./27.*m44*me2*pow(
         RSe561p,2) + 1./81.*m44*me2*pow(RSe561p,2)*RS456pe + 1./27.*
         m44*me2*RSp561e*RSe563p - 1./81.*m44*me2*RSp561e*RS456pe*
         RSe563p - 1./27.*m44*me2*RSp561e*RSe561p + 1./81.*m44*me2*
         RSp561e*RSe561p*RS456pe - 13./72.*m33*m44 - 1./54.*m33*m44*
         pow(RS356pe,2) + 1./54.*m33*m44*pow(RSe563p,2) + 1./27.*m33*
         m44*RSp563e*RSe563p + 1./54.*m33*m44*pow(RSp563e,2) + 2./27.*
         m33*m44*RS156pe*RS356pe );

      M13p6x23R +=  + Abeps(m44,mu2)*pi16 * ( 1./27.*m33*m44*RS156pe*
         RSe563p + 1./27.*m33*m44*RS156pe*RSp563e + 13./216.*m33*m44*
         RS456pe + 1./162.*m33*m44*RS456pe*pow(RS356pe,2) - 1./162.*m33
         *m44*RS456pe*pow(RSe563p,2) - 1./81.*m33*m44*RS456pe*RSp563e*
         RSe563p - 1./162.*m33*m44*RS456pe*pow(RSp563e,2) - 2./81.*m33*
         m44*RS456pe*RS156pe*RS356pe - 1./81.*m33*m44*RS456pe*RS156pe*
         RSe563p - 1./81.*m33*m44*RS456pe*RS156pe*RSp563e + 2./27.*m33*
         m44*RSe561p*RS356pe + 1./27.*m33*m44*RSe561p*RSe563p + 1./27.*
         m33*m44*RSe561p*RSp563e - 2./81.*m33*m44*RSe561p*RS456pe*
         RS356pe - 1./81.*m33*m44*RSe561p*RS456pe*RSe563p - 1./81.*m33*
         m44*RSe561p*RS456pe*RSp563e + 2./27.*m33*m44*RSp561e*RS356pe
          + 1./27.*m33*m44*RSp561e*RSe563p + 1./27.*m33*m44*RSp561e*
         RSp563e - 2./81.*m33*m44*RSp561e*RS456pe*RS356pe - 1./81.*m33*
         m44*RSp561e*RS456pe*RSe563p - 1./81.*m33*m44*RSp561e*RS456pe*
         RSp563e - 13./72.*m11*m44 + 2./27.*m11*m44*RS156pe*RS356pe + 2.
         /27.*m11*m44*RS156pe*RSe563p );

      M13p6x23R +=  + Abeps(m44,mu2)*pi16 * ( 2./27.*m11*m44*RS156pe*
         RSp563e - 1./54.*m11*m44*pow(RS156pe,2) + 13./216.*m11*m44*
         RS456pe - 2./81.*m11*m44*RS456pe*RS156pe*RS356pe - 2./81.*m11*
         m44*RS456pe*RS156pe*RSe563p - 2./81.*m11*m44*RS456pe*RS156pe*
         RSp563e + 1./162.*m11*m44*RS456pe*pow(RS156pe,2) + 1./27.*m11*
         m44*RSe561p*RS356pe + 1./27.*m11*m44*RSe561p*RSe563p + 1./27.*
         m11*m44*RSe561p*RSp563e - 1./81.*m11*m44*RSe561p*RS456pe*
         RS356pe - 1./81.*m11*m44*RSe561p*RS456pe*RSe563p - 1./81.*m11*
         m44*RSe561p*RS456pe*RSp563e + 1./54.*m11*m44*pow(RSe561p,2) - 
         1./162.*m11*m44*pow(RSe561p,2)*RS456pe + 1./27.*m11*m44*
         RSp561e*RS356pe + 1./27.*m11*m44*RSp561e*RSe563p + 1./27.*m11*
         m44*RSp561e*RSp563e - 1./81.*m11*m44*RSp561e*RS456pe*RS356pe
          - 1./81.*m11*m44*RSp561e*RS456pe*RSe563p - 1./81.*m11*m44*
         RSp561e*RS456pe*RSp563e + 1./27.*m11*m44*RSp561e*RSe561p - 1./
         81.*m11*m44*RSp561e*RSe561p*RS456pe + 1./54.*m11*m44*pow(
         RSp561e,2) );

      M13p6x23R +=  + Abeps(m44,mu2)*pi16 * (  - 1./162.*m11*m44*pow(
         RSp561e,2)*RS456pe );

      M13p6x23R +=  + Abeps(m45,mu2)*pi16 * (  - 1./81.*pow(mp2,2)*
         RSp563e*RS346pe + 1./81.*pow(mp2,2)*RSp563e*RS356pe + 1./81.*
         pow(mp2,2)*RSp563e*RSe563p - 1./81.*pow(mp2,2)*RSp563e*RSe463p
          + 1./54.*pow(mp2,2)*pow(RSp563e,2) + 1./81.*pow(mp2,2)*
         RSp463e*RS346pe - 1./81.*pow(mp2,2)*RSp463e*RS356pe - 1./81.*
         pow(mp2,2)*RSp463e*RSe563p + 1./81.*pow(mp2,2)*RSp463e*RSe463p
          - 1./27.*pow(mp2,2)*RSp463e*RSp563e + 1./54.*pow(mp2,2)*pow(
         RSp463e,2) + 1./81.*pow(mp2,2)*RS146pe*RSp563e - 1./81.*pow(
         mp2,2)*RS146pe*RSp463e - 1./81.*pow(mp2,2)*RS156pe*RSp563e + 1.
         /81.*pow(mp2,2)*RS156pe*RSp463e - 1./81.*pow(mp2,2)*RSe561p*
         RSp563e + 1./81.*pow(mp2,2)*RSe561p*RSp463e + 1./81.*pow(
         mp2,2)*RSe461p*RSp563e - 1./81.*pow(mp2,2)*RSe461p*RSp463e + 1.
         /81.*pow(mp2,2)*RSp561e*RS346pe - 1./81.*pow(mp2,2)*RSp561e*
         RS356pe - 1./81.*pow(mp2,2)*RSp561e*RSe563p + 1./81.*pow(
         mp2,2)*RSp561e*RSe463p - 1./27.*pow(mp2,2)*RSp561e*RSp563e + 1.
         /27.*pow(mp2,2)*RSp561e*RSp463e );

      M13p6x23R +=  + Abeps(m45,mu2)*pi16 * (  - 1./81.*pow(mp2,2)*
         RSp561e*RS146pe + 1./81.*pow(mp2,2)*RSp561e*RS156pe + 1./81.*
         pow(mp2,2)*RSp561e*RSe561p - 1./81.*pow(mp2,2)*RSp561e*RSe461p
          + 1./54.*pow(mp2,2)*pow(RSp561e,2) - 1./81.*pow(mp2,2)*
         RSp461e*RS346pe + 1./81.*pow(mp2,2)*RSp461e*RS356pe + 1./81.*
         pow(mp2,2)*RSp461e*RSe563p - 1./81.*pow(mp2,2)*RSp461e*RSe463p
          + 1./27.*pow(mp2,2)*RSp461e*RSp563e - 1./27.*pow(mp2,2)*
         RSp461e*RSp463e + 1./81.*pow(mp2,2)*RSp461e*RS146pe - 1./81.*
         pow(mp2,2)*RSp461e*RS156pe - 1./81.*pow(mp2,2)*RSp461e*RSe561p
          + 1./81.*pow(mp2,2)*RSp461e*RSe461p - 1./27.*pow(mp2,2)*
         RSp461e*RSp561e + 1./54.*pow(mp2,2)*pow(RSp461e,2) + 1./81.*
         me2*mp2*RSp563e*RSe563p - 1./81.*me2*mp2*RSp563e*RSe463p - 1./
         81.*me2*mp2*RSp463e*RSe563p + 1./81.*me2*mp2*RSp463e*RSe463p
          - 1./81.*me2*mp2*RSe561p*RSp563e + 1./81.*me2*mp2*RSe561p*
         RSp463e + 1./81.*me2*mp2*RSe461p*RSp563e - 1./81.*me2*mp2*
         RSe461p*RSp463e );

      M13p6x23R +=  + Abeps(m45,mu2)*pi16 * (  - 1./81.*me2*mp2*RSp561e
         *RSe563p + 1./81.*me2*mp2*RSp561e*RSe463p + 1./81.*me2*mp2*
         RSp561e*RSe561p - 1./81.*me2*mp2*RSp561e*RSe461p + 1./81.*me2*
         mp2*RSp461e*RSe563p - 1./81.*me2*mp2*RSp461e*RSe463p - 1./81.*
         me2*mp2*RSp461e*RSe561p + 1./81.*me2*mp2*RSp461e*RSe461p - 1./
         81.*pow(me2,2)*RSe563p*RS346pe + 1./81.*pow(me2,2)*RSe563p*
         RS356pe + 1./54.*pow(me2,2)*pow(RSe563p,2) + 1./81.*pow(me2,2)
         *RSe463p*RS346pe - 1./81.*pow(me2,2)*RSe463p*RS356pe - 1./27.*
         pow(me2,2)*RSe463p*RSe563p + 1./54.*pow(me2,2)*pow(RSe463p,2)
          + 1./81.*pow(me2,2)*RSp563e*RSe563p - 1./81.*pow(me2,2)*
         RSp563e*RSe463p - 1./81.*pow(me2,2)*RSp463e*RSe563p + 1./81.*
         pow(me2,2)*RSp463e*RSe463p + 1./81.*pow(me2,2)*RS146pe*RSe563p
          - 1./81.*pow(me2,2)*RS146pe*RSe463p - 1./81.*pow(me2,2)*
         RS156pe*RSe563p + 1./81.*pow(me2,2)*RS156pe*RSe463p + 1./81.*
         pow(me2,2)*RSe561p*RS346pe - 1./81.*pow(me2,2)*RSe561p*RS356pe
          - 1./27.*pow(me2,2)*RSe561p*RSe563p );

      M13p6x23R +=  + Abeps(m45,mu2)*pi16 * ( 1./27.*pow(me2,2)*RSe561p
         *RSe463p - 1./81.*pow(me2,2)*RSe561p*RSp563e + 1./81.*pow(
         me2,2)*RSe561p*RSp463e - 1./81.*pow(me2,2)*RSe561p*RS146pe + 1.
         /81.*pow(me2,2)*RSe561p*RS156pe + 1./54.*pow(me2,2)*pow(
         RSe561p,2) - 1./81.*pow(me2,2)*RSe461p*RS346pe + 1./81.*pow(
         me2,2)*RSe461p*RS356pe + 1./27.*pow(me2,2)*RSe461p*RSe563p - 1.
         /27.*pow(me2,2)*RSe461p*RSe463p + 1./81.*pow(me2,2)*RSe461p*
         RSp563e - 1./81.*pow(me2,2)*RSe461p*RSp463e + 1./81.*pow(
         me2,2)*RSe461p*RS146pe - 1./81.*pow(me2,2)*RSe461p*RS156pe - 1.
         /27.*pow(me2,2)*RSe461p*RSe561p + 1./54.*pow(me2,2)*pow(
         RSe461p,2) - 1./81.*pow(me2,2)*RSp561e*RSe563p + 1./81.*pow(
         me2,2)*RSp561e*RSe463p + 1./81.*pow(me2,2)*RSp561e*RSe561p - 1.
         /81.*pow(me2,2)*RSp561e*RSe461p + 1./81.*pow(me2,2)*RSp461e*
         RSe563p - 1./81.*pow(me2,2)*RSp461e*RSe463p - 1./81.*pow(
         me2,2)*RSp461e*RSe561p + 1./81.*pow(me2,2)*RSp461e*RSe461p - 1.
         /81.*m55*mp2*RSp563e*RS346pe );

      M13p6x23R +=  + Abeps(m45,mu2)*pi16 * (  - 1./81.*m55*mp2*RSp563e
         *RSe463p - 1./81.*m55*mp2*RSp463e*RS346pe - 1./81.*m55*mp2*
         RSp463e*RS356pe - 1./81.*m55*mp2*RSp463e*RSe563p - 1./81.*m55*
         mp2*RSp463e*RSe463p - 2./81.*m55*mp2*RSp463e*RSp563e - 1./81.*
         m55*mp2*pow(RSp463e,2) + 1./81.*m55*mp2*RS146pe*RSp563e + 1./
         81.*m55*mp2*RS146pe*RSp463e + 1./81.*m55*mp2*RS156pe*RSp463e
          + 1./81.*m55*mp2*RSe561p*RSp463e + 1./81.*m55*mp2*RSe461p*
         RSp563e + 1./81.*m55*mp2*RSe461p*RSp463e + 1./81.*m55*mp2*
         RSp561e*RS346pe + 1./81.*m55*mp2*RSp561e*RSe463p + 2./81.*m55*
         mp2*RSp561e*RSp463e - 1./81.*m55*mp2*RSp561e*RS146pe - 1./81.*
         m55*mp2*RSp561e*RSe461p + 1./81.*m55*mp2*RSp461e*RS346pe + 1./
         81.*m55*mp2*RSp461e*RS356pe + 1./81.*m55*mp2*RSp461e*RSe563p
          + 1./81.*m55*mp2*RSp461e*RSe463p + 2./81.*m55*mp2*RSp461e*
         RSp563e + 2./81.*m55*mp2*RSp461e*RSp463e - 1./81.*m55*mp2*
         RSp461e*RS146pe - 1./81.*m55*mp2*RSp461e*RS156pe - 1./81.*m55*
         mp2*RSp461e*RSe561p );

      M13p6x23R +=  + Abeps(m45,mu2)*pi16 * (  - 1./81.*m55*mp2*RSp461e
         *RSe461p - 2./81.*m55*mp2*RSp461e*RSp561e - 1./81.*m55*mp2*
         pow(RSp461e,2) - 1./81.*m55*me2*RSe563p*RS346pe - 1./81.*m55*
         me2*RSe463p*RS346pe - 1./81.*m55*me2*RSe463p*RS356pe - 2./81.*
         m55*me2*RSe463p*RSe563p - 1./81.*m55*me2*pow(RSe463p,2) - 1./
         81.*m55*me2*RSp563e*RSe463p - 1./81.*m55*me2*RSp463e*RSe563p
          - 1./81.*m55*me2*RSp463e*RSe463p + 1./81.*m55*me2*RS146pe*
         RSe563p + 1./81.*m55*me2*RS146pe*RSe463p + 1./81.*m55*me2*
         RS156pe*RSe463p + 1./81.*m55*me2*RSe561p*RS346pe + 2./81.*m55*
         me2*RSe561p*RSe463p + 1./81.*m55*me2*RSe561p*RSp463e - 1./81.*
         m55*me2*RSe561p*RS146pe + 1./81.*m55*me2*RSe461p*RS346pe + 1./
         81.*m55*me2*RSe461p*RS356pe + 2./81.*m55*me2*RSe461p*RSe563p
          + 2./81.*m55*me2*RSe461p*RSe463p + 1./81.*m55*me2*RSe461p*
         RSp563e + 1./81.*m55*me2*RSe461p*RSp463e - 1./81.*m55*me2*
         RSe461p*RS146pe - 1./81.*m55*me2*RSe461p*RS156pe - 2./81.*m55*
         me2*RSe461p*RSe561p );

      M13p6x23R +=  + Abeps(m45,mu2)*pi16 * (  - 1./81.*m55*me2*pow(
         RSe461p,2) + 1./81.*m55*me2*RSp561e*RSe463p - 1./81.*m55*me2*
         RSp561e*RSe461p + 1./81.*m55*me2*RSp461e*RSe563p + 1./81.*m55*
         me2*RSp461e*RSe463p - 1./81.*m55*me2*RSp461e*RSe561p - 1./81.*
         m55*me2*RSp461e*RSe461p - 1./81.*m44*mp2*RSp563e*RS346pe - 1./
         81.*m44*mp2*RSp563e*RS356pe - 1./81.*m44*mp2*RSp563e*RSe563p
          - 1./81.*m44*mp2*RSp563e*RSe463p - 1./81.*m44*mp2*pow(
         RSp563e,2) - 1./81.*m44*mp2*RSp463e*RS356pe - 1./81.*m44*mp2*
         RSp463e*RSe563p - 2./81.*m44*mp2*RSp463e*RSp563e + 1./81.*m44*
         mp2*RS146pe*RSp563e + 1./81.*m44*mp2*RS156pe*RSp563e + 1./81.*
         m44*mp2*RS156pe*RSp463e + 1./81.*m44*mp2*RSe561p*RSp563e + 1./
         81.*m44*mp2*RSe561p*RSp463e + 1./81.*m44*mp2*RSe461p*RSp563e
          + 1./81.*m44*mp2*RSp561e*RS346pe + 1./81.*m44*mp2*RSp561e*
         RS356pe + 1./81.*m44*mp2*RSp561e*RSe563p + 1./81.*m44*mp2*
         RSp561e*RSe463p + 2./81.*m44*mp2*RSp561e*RSp563e + 2./81.*m44*
         mp2*RSp561e*RSp463e );

      M13p6x23R +=  + Abeps(m45,mu2)*pi16 * (  - 1./81.*m44*mp2*RSp561e
         *RS146pe - 1./81.*m44*mp2*RSp561e*RS156pe - 1./81.*m44*mp2*
         RSp561e*RSe561p - 1./81.*m44*mp2*RSp561e*RSe461p - 1./81.*m44*
         mp2*pow(RSp561e,2) + 1./81.*m44*mp2*RSp461e*RS356pe + 1./81.*
         m44*mp2*RSp461e*RSe563p + 2./81.*m44*mp2*RSp461e*RSp563e - 1./
         81.*m44*mp2*RSp461e*RS156pe - 1./81.*m44*mp2*RSp461e*RSe561p
          - 2./81.*m44*mp2*RSp461e*RSp561e - 1./81.*m44*me2*RSe563p*
         RS346pe - 1./81.*m44*me2*RSe563p*RS356pe - 1./81.*m44*me2*pow(
         RSe563p,2) - 1./81.*m44*me2*RSe463p*RS356pe - 2./81.*m44*me2*
         RSe463p*RSe563p - 1./81.*m44*me2*RSp563e*RSe563p - 1./81.*m44*
         me2*RSp563e*RSe463p - 1./81.*m44*me2*RSp463e*RSe563p + 1./81.*
         m44*me2*RS146pe*RSe563p + 1./81.*m44*me2*RS156pe*RSe563p + 1./
         81.*m44*me2*RS156pe*RSe463p + 1./81.*m44*me2*RSe561p*RS346pe
          + 1./81.*m44*me2*RSe561p*RS356pe + 2./81.*m44*me2*RSe561p*
         RSe563p + 2./81.*m44*me2*RSe561p*RSe463p + 1./81.*m44*me2*
         RSe561p*RSp563e );

      M13p6x23R +=  + Abeps(m45,mu2)*pi16 * ( 1./81.*m44*me2*RSe561p*
         RSp463e - 1./81.*m44*me2*RSe561p*RS146pe - 1./81.*m44*me2*
         RSe561p*RS156pe - 1./81.*m44*me2*pow(RSe561p,2) + 1./81.*m44*
         me2*RSe461p*RS356pe + 2./81.*m44*me2*RSe461p*RSe563p + 1./81.*
         m44*me2*RSe461p*RSp563e - 1./81.*m44*me2*RSe461p*RS156pe - 2./
         81.*m44*me2*RSe461p*RSe561p + 1./81.*m44*me2*RSp561e*RSe563p
          + 1./81.*m44*me2*RSp561e*RSe463p - 1./81.*m44*me2*RSp561e*
         RSe561p - 1./81.*m44*me2*RSp561e*RSe461p + 1./81.*m44*me2*
         RSp461e*RSe563p - 1./81.*m44*me2*RSp461e*RSe561p - 1./81.*m33*
         mp2*RSp563e*RSe563p + 1./81.*m33*mp2*RSp563e*RSe463p - 1./81.*
         m33*mp2*pow(RSp563e,2) + 1./81.*m33*mp2*RSp463e*RSe563p - 1./
         81.*m33*mp2*RSp463e*RSe463p + 2./81.*m33*mp2*RSp463e*RSp563e
          - 1./81.*m33*mp2*pow(RSp463e,2) + 1./81.*m33*mp2*RS146pe*
         RSp563e - 1./81.*m33*mp2*RS146pe*RSp463e - 1./81.*m33*mp2*
         RS156pe*RSp563e + 1./81.*m33*mp2*RS156pe*RSp463e - 1./81.*m33*
         mp2*RSe561p*RSp563e );

      M13p6x23R +=  + Abeps(m45,mu2)*pi16 * ( 1./81.*m33*mp2*RSe561p*
         RSp463e + 1./81.*m33*mp2*RSe461p*RSp563e - 1./81.*m33*mp2*
         RSe461p*RSp463e + 2./81.*m33*mp2*RSp561e*RS346pe - 2./81.*m33*
         mp2*RSp561e*RS356pe - 1./81.*m33*mp2*RSp561e*RSe563p + 1./81.*
         m33*mp2*RSp561e*RSe463p - 2./81.*m33*mp2*RSp561e*RSp563e + 2./
         81.*m33*mp2*RSp561e*RSp463e - 2./81.*m33*mp2*RSp461e*RS346pe
          + 2./81.*m33*mp2*RSp461e*RS356pe + 1./81.*m33*mp2*RSp461e*
         RSe563p - 1./81.*m33*mp2*RSp461e*RSe463p + 2./81.*m33*mp2*
         RSp461e*RSp563e - 2./81.*m33*mp2*RSp461e*RSp463e - 1./81.*m33*
         me2*pow(RSe563p,2) + 2./81.*m33*me2*RSe463p*RSe563p - 1./81.*
         m33*me2*pow(RSe463p,2) - 1./81.*m33*me2*RSp563e*RSe563p + 1./
         81.*m33*me2*RSp563e*RSe463p + 1./81.*m33*me2*RSp463e*RSe563p
          - 1./81.*m33*me2*RSp463e*RSe463p + 1./81.*m33*me2*RS146pe*
         RSe563p - 1./81.*m33*me2*RS146pe*RSe463p - 1./81.*m33*me2*
         RS156pe*RSe563p + 1./81.*m33*me2*RS156pe*RSe463p + 2./81.*m33*
         me2*RSe561p*RS346pe );

      M13p6x23R +=  + Abeps(m45,mu2)*pi16 * (  - 2./81.*m33*me2*RSe561p
         *RS356pe - 2./81.*m33*me2*RSe561p*RSe563p + 2./81.*m33*me2*
         RSe561p*RSe463p - 1./81.*m33*me2*RSe561p*RSp563e + 1./81.*m33*
         me2*RSe561p*RSp463e - 2./81.*m33*me2*RSe461p*RS346pe + 2./81.*
         m33*me2*RSe461p*RS356pe + 2./81.*m33*me2*RSe461p*RSe563p - 2./
         81.*m33*me2*RSe461p*RSe463p + 1./81.*m33*me2*RSe461p*RSp563e
          - 1./81.*m33*me2*RSe461p*RSp463e - 1./81.*m33*me2*RSp561e*
         RSe563p + 1./81.*m33*me2*RSp561e*RSe463p + 1./81.*m33*me2*
         RSp461e*RSe563p - 1./81.*m33*me2*RSp461e*RSe463p - 13./72.*m33
         *m55 - 1./162.*m33*m55*pow(RS346pe,2) - 1./81.*m33*m55*RS356pe
         *RS346pe + 1./81.*m33*m55*RSe463p*RSe563p + 1./162.*m33*m55*
         pow(RSe463p,2) + 1./81.*m33*m55*RSp563e*RSe463p + 1./81.*m33*
         m55*RSp463e*RSe563p + 1./81.*m33*m55*RSp463e*RSe463p + 1./81.*
         m33*m55*RSp463e*RSp563e + 1./162.*m33*m55*pow(RSp463e,2) + 2./
         81.*m33*m55*RS146pe*RS346pe + 2./81.*m33*m55*RS146pe*RS356pe
          + 1./81.*m33*m55*RS146pe*RSe563p );

      M13p6x23R +=  + Abeps(m45,mu2)*pi16 * ( 1./81.*m33*m55*RS146pe*
         RSe463p + 1./81.*m33*m55*RS146pe*RSp563e + 1./81.*m33*m55*
         RS146pe*RSp463e + 2./81.*m33*m55*RS156pe*RS346pe + 1./81.*m33*
         m55*RS156pe*RSe463p + 1./81.*m33*m55*RS156pe*RSp463e + 2./81.*
         m33*m55*RSe561p*RS346pe + 1./81.*m33*m55*RSe561p*RSe463p + 1./
         81.*m33*m55*RSe561p*RSp463e + 2./81.*m33*m55*RSe461p*RS346pe
          + 2./81.*m33*m55*RSe461p*RS356pe + 1./81.*m33*m55*RSe461p*
         RSe563p + 1./81.*m33*m55*RSe461p*RSe463p + 1./81.*m33*m55*
         RSe461p*RSp563e + 1./81.*m33*m55*RSe461p*RSp463e + 2./81.*m33*
         m55*RSp561e*RS346pe + 1./81.*m33*m55*RSp561e*RSe463p + 1./81.*
         m33*m55*RSp561e*RSp463e + 2./81.*m33*m55*RSp461e*RS346pe + 2./
         81.*m33*m55*RSp461e*RS356pe + 1./81.*m33*m55*RSp461e*RSe563p
          + 1./81.*m33*m55*RSp461e*RSe463p + 1./81.*m33*m55*RSp461e*
         RSp563e + 1./81.*m33*m55*RSp461e*RSp463e - 13./72.*m33*m44 - 1.
         /81.*m33*m44*RS356pe*RS346pe - 1./162.*m33*m44*pow(RS356pe,2)
          + 1./162.*m33*m44*pow(RSe563p,2) );

      M13p6x23R +=  + Abeps(m45,mu2)*pi16 * ( 1./81.*m33*m44*RSe463p*
         RSe563p + 1./81.*m33*m44*RSp563e*RSe563p + 1./81.*m33*m44*
         RSp563e*RSe463p + 1./162.*m33*m44*pow(RSp563e,2) + 1./81.*m33*
         m44*RSp463e*RSe563p + 1./81.*m33*m44*RSp463e*RSp563e + 2./81.*
         m33*m44*RS146pe*RS356pe + 1./81.*m33*m44*RS146pe*RSe563p + 1./
         81.*m33*m44*RS146pe*RSp563e + 2./81.*m33*m44*RS156pe*RS346pe
          + 2./81.*m33*m44*RS156pe*RS356pe + 1./81.*m33*m44*RS156pe*
         RSe563p + 1./81.*m33*m44*RS156pe*RSe463p + 1./81.*m33*m44*
         RS156pe*RSp563e + 1./81.*m33*m44*RS156pe*RSp463e + 2./81.*m33*
         m44*RSe561p*RS346pe + 2./81.*m33*m44*RSe561p*RS356pe + 1./81.*
         m33*m44*RSe561p*RSe563p + 1./81.*m33*m44*RSe561p*RSe463p + 1./
         81.*m33*m44*RSe561p*RSp563e + 1./81.*m33*m44*RSe561p*RSp463e
          + 2./81.*m33*m44*RSe461p*RS356pe + 1./81.*m33*m44*RSe461p*
         RSe563p + 1./81.*m33*m44*RSe461p*RSp563e + 2./81.*m33*m44*
         RSp561e*RS346pe + 2./81.*m33*m44*RSp561e*RS356pe + 1./81.*m33*
         m44*RSp561e*RSe563p );

      M13p6x23R +=  + Abeps(m45,mu2)*pi16 * ( 1./81.*m33*m44*RSp561e*
         RSe463p + 1./81.*m33*m44*RSp561e*RSp563e + 1./81.*m33*m44*
         RSp561e*RSp463e + 2./81.*m33*m44*RSp461e*RS356pe + 1./81.*m33*
         m44*RSp461e*RSe563p + 1./81.*m33*m44*RSp461e*RSp563e + 1./162.
         *pow(m33,2)*pow(RS346pe,2) - 1./81.*pow(m33,2)*RS356pe*RS346pe
          + 1./162.*pow(m33,2)*pow(RS356pe,2) - 2./81.*pow(m33,2)*
         RS146pe*RS346pe + 2./81.*pow(m33,2)*RS146pe*RS356pe + 2./81.*
         pow(m33,2)*RS156pe*RS346pe - 2./81.*pow(m33,2)*RS156pe*RS356pe
          + 2./81.*pow(m33,2)*RSe561p*RS346pe - 2./81.*pow(m33,2)*
         RSe561p*RS356pe - 2./81.*pow(m33,2)*RSe461p*RS346pe + 2./81.*
         pow(m33,2)*RSe461p*RS356pe + 2./81.*pow(m33,2)*RSp561e*RS346pe
          - 2./81.*pow(m33,2)*RSp561e*RS356pe - 2./81.*pow(m33,2)*
         RSp461e*RS346pe + 2./81.*pow(m33,2)*RSp461e*RS356pe + 2./81.*
         m11*mp2*RS146pe*RSp563e - 2./81.*m11*mp2*RS146pe*RSp463e - 2./
         81.*m11*mp2*RS156pe*RSp563e + 2./81.*m11*mp2*RS156pe*RSp463e
          - 1./81.*m11*mp2*RSe561p*RSp563e );

      M13p6x23R +=  + Abeps(m45,mu2)*pi16 * ( 1./81.*m11*mp2*RSe561p*
         RSp463e + 1./81.*m11*mp2*RSe461p*RSp563e - 1./81.*m11*mp2*
         RSe461p*RSp463e + 1./81.*m11*mp2*RSp561e*RS346pe - 1./81.*m11*
         mp2*RSp561e*RS356pe - 1./81.*m11*mp2*RSp561e*RSe563p + 1./81.*
         m11*mp2*RSp561e*RSe463p - 2./81.*m11*mp2*RSp561e*RSp563e + 2./
         81.*m11*mp2*RSp561e*RSp463e - 1./81.*m11*mp2*RSp561e*RSe561p
          + 1./81.*m11*mp2*RSp561e*RSe461p - 1./81.*m11*mp2*pow(
         RSp561e,2) - 1./81.*m11*mp2*RSp461e*RS346pe + 1./81.*m11*mp2*
         RSp461e*RS356pe + 1./81.*m11*mp2*RSp461e*RSe563p - 1./81.*m11*
         mp2*RSp461e*RSe463p + 2./81.*m11*mp2*RSp461e*RSp563e - 2./81.*
         m11*mp2*RSp461e*RSp463e + 1./81.*m11*mp2*RSp461e*RSe561p - 1./
         81.*m11*mp2*RSp461e*RSe461p + 2./81.*m11*mp2*RSp461e*RSp561e
          - 1./81.*m11*mp2*pow(RSp461e,2) + 2./81.*m11*me2*RS146pe*
         RSe563p - 2./81.*m11*me2*RS146pe*RSe463p - 2./81.*m11*me2*
         RS156pe*RSe563p + 2./81.*m11*me2*RS156pe*RSe463p + 1./81.*m11*
         me2*RSe561p*RS346pe );

      M13p6x23R +=  + Abeps(m45,mu2)*pi16 * (  - 1./81.*m11*me2*RSe561p
         *RS356pe - 2./81.*m11*me2*RSe561p*RSe563p + 2./81.*m11*me2*
         RSe561p*RSe463p - 1./81.*m11*me2*RSe561p*RSp563e + 1./81.*m11*
         me2*RSe561p*RSp463e - 1./81.*m11*me2*pow(RSe561p,2) - 1./81.*
         m11*me2*RSe461p*RS346pe + 1./81.*m11*me2*RSe461p*RS356pe + 2./
         81.*m11*me2*RSe461p*RSe563p - 2./81.*m11*me2*RSe461p*RSe463p
          + 1./81.*m11*me2*RSe461p*RSp563e - 1./81.*m11*me2*RSe461p*
         RSp463e + 2./81.*m11*me2*RSe461p*RSe561p - 1./81.*m11*me2*pow(
         RSe461p,2) - 1./81.*m11*me2*RSp561e*RSe563p + 1./81.*m11*me2*
         RSp561e*RSe463p - 1./81.*m11*me2*RSp561e*RSe561p + 1./81.*m11*
         me2*RSp561e*RSe461p + 1./81.*m11*me2*RSp461e*RSe563p - 1./81.*
         m11*me2*RSp461e*RSe463p + 1./81.*m11*me2*RSp461e*RSe561p - 1./
         81.*m11*me2*RSp461e*RSe461p - 13./72.*m11*m55 + 2./81.*m11*m55
         *RS146pe*RS346pe + 2./81.*m11*m55*RS146pe*RS356pe + 2./81.*m11
         *m55*RS146pe*RSe563p + 2./81.*m11*m55*RS146pe*RSe463p + 2./81.
         *m11*m55*RS146pe*RSp563e );

      M13p6x23R +=  + Abeps(m45,mu2)*pi16 * ( 2./81.*m11*m55*RS146pe*
         RSp463e - 1./162.*m11*m55*pow(RS146pe,2) + 2./81.*m11*m55*
         RS156pe*RS346pe + 2./81.*m11*m55*RS156pe*RSe463p + 2./81.*m11*
         m55*RS156pe*RSp463e - 1./81.*m11*m55*RS156pe*RS146pe + 1./81.*
         m11*m55*RSe561p*RS346pe + 1./81.*m11*m55*RSe561p*RSe463p + 1./
         81.*m11*m55*RSe561p*RSp463e + 1./81.*m11*m55*RSe461p*RS346pe
          + 1./81.*m11*m55*RSe461p*RS356pe + 1./81.*m11*m55*RSe461p*
         RSe563p + 1./81.*m11*m55*RSe461p*RSe463p + 1./81.*m11*m55*
         RSe461p*RSp563e + 1./81.*m11*m55*RSe461p*RSp463e + 1./81.*m11*
         m55*RSe461p*RSe561p + 1./162.*m11*m55*pow(RSe461p,2) + 1./81.*
         m11*m55*RSp561e*RS346pe + 1./81.*m11*m55*RSp561e*RSe463p + 1./
         81.*m11*m55*RSp561e*RSp463e + 1./81.*m11*m55*RSp561e*RSe461p
          + 1./81.*m11*m55*RSp461e*RS346pe + 1./81.*m11*m55*RSp461e*
         RS356pe + 1./81.*m11*m55*RSp461e*RSe563p + 1./81.*m11*m55*
         RSp461e*RSe463p + 1./81.*m11*m55*RSp461e*RSp563e + 1./81.*m11*
         m55*RSp461e*RSp463e );

      M13p6x23R +=  + Abeps(m45,mu2)*pi16 * ( 1./81.*m11*m55*RSp461e*
         RSe561p + 1./81.*m11*m55*RSp461e*RSe461p + 1./81.*m11*m55*
         RSp461e*RSp561e + 1./162.*m11*m55*pow(RSp461e,2) - 13./72.*m11
         *m44 + 2./81.*m11*m44*RS146pe*RS356pe + 2./81.*m11*m44*RS146pe
         *RSe563p + 2./81.*m11*m44*RS146pe*RSp563e + 2./81.*m11*m44*
         RS156pe*RS346pe + 2./81.*m11*m44*RS156pe*RS356pe + 2./81.*m11*
         m44*RS156pe*RSe563p + 2./81.*m11*m44*RS156pe*RSe463p + 2./81.*
         m11*m44*RS156pe*RSp563e + 2./81.*m11*m44*RS156pe*RSp463e - 1./
         81.*m11*m44*RS156pe*RS146pe - 1./162.*m11*m44*pow(RS156pe,2)
          + 1./81.*m11*m44*RSe561p*RS346pe + 1./81.*m11*m44*RSe561p*
         RS356pe + 1./81.*m11*m44*RSe561p*RSe563p + 1./81.*m11*m44*
         RSe561p*RSe463p + 1./81.*m11*m44*RSe561p*RSp563e + 1./81.*m11*
         m44*RSe561p*RSp463e + 1./162.*m11*m44*pow(RSe561p,2) + 1./81.*
         m11*m44*RSe461p*RS356pe + 1./81.*m11*m44*RSe461p*RSe563p + 1./
         81.*m11*m44*RSe461p*RSp563e + 1./81.*m11*m44*RSe461p*RSe561p
          + 1./81.*m11*m44*RSp561e*RS346pe );

      M13p6x23R +=  + Abeps(m45,mu2)*pi16 * ( 1./81.*m11*m44*RSp561e*
         RS356pe + 1./81.*m11*m44*RSp561e*RSe563p + 1./81.*m11*m44*
         RSp561e*RSe463p + 1./81.*m11*m44*RSp561e*RSp563e + 1./81.*m11*
         m44*RSp561e*RSp463e + 1./81.*m11*m44*RSp561e*RSe561p + 1./81.*
         m11*m44*RSp561e*RSe461p + 1./162.*m11*m44*pow(RSp561e,2) + 1./
         81.*m11*m44*RSp461e*RS356pe + 1./81.*m11*m44*RSp461e*RSe563p
          + 1./81.*m11*m44*RSp461e*RSp563e + 1./81.*m11*m44*RSp461e*
         RSe561p + 1./81.*m11*m44*RSp461e*RSp561e - 1./27.*m11*m33*
         RS146pe*RS346pe + 1./27.*m11*m33*RS146pe*RS356pe + 1./81.*m11*
         m33*RS146pe*RSe563p - 1./81.*m11*m33*RS146pe*RSe463p + 1./81.*
         m11*m33*RS146pe*RSp563e - 1./81.*m11*m33*RS146pe*RSp463e + 1./
         27.*m11*m33*RS156pe*RS346pe - 1./27.*m11*m33*RS156pe*RS356pe
          - 1./81.*m11*m33*RS156pe*RSe563p + 1./81.*m11*m33*RS156pe*
         RSe463p - 1./81.*m11*m33*RS156pe*RSp563e + 1./81.*m11*m33*
         RS156pe*RSp463e + 1./81.*m11*m33*RSe561p*RS346pe - 1./81.*m11*
         m33*RSe561p*RS356pe );

      M13p6x23R +=  + Abeps(m45,mu2)*pi16 * (  - 1./81.*m11*m33*RSe461p
         *RS346pe + 1./81.*m11*m33*RSe461p*RS356pe + 1./81.*m11*m33*
         RSp561e*RS346pe - 1./81.*m11*m33*RSp561e*RS356pe - 1./81.*m11*
         m33*RSp461e*RS346pe + 1./81.*m11*m33*RSp461e*RS356pe - 2./81.*
         pow(m11,2)*RS146pe*RS346pe + 2./81.*pow(m11,2)*RS146pe*RS356pe
          + 2./81.*pow(m11,2)*RS146pe*RSe563p - 2./81.*pow(m11,2)*
         RS146pe*RSe463p + 2./81.*pow(m11,2)*RS146pe*RSp563e - 2./81.*
         pow(m11,2)*RS146pe*RSp463e + 1./162.*pow(m11,2)*pow(RS146pe,2)
          + 2./81.*pow(m11,2)*RS156pe*RS346pe - 2./81.*pow(m11,2)*
         RS156pe*RS356pe - 2./81.*pow(m11,2)*RS156pe*RSe563p + 2./81.*
         pow(m11,2)*RS156pe*RSe463p - 2./81.*pow(m11,2)*RS156pe*RSp563e
          + 2./81.*pow(m11,2)*RS156pe*RSp463e - 1./81.*pow(m11,2)*
         RS156pe*RS146pe + 1./162.*pow(m11,2)*pow(RS156pe,2) );

      M13p6x23R +=  + Abeps(m46,mu2)*pi16 * (  - 1./81.*pow(mp2,2)*
         RSp563e*RS345pe + 1./81.*pow(mp2,2)*RSp563e*RS356pe + 1./81.*
         pow(mp2,2)*RSp563e*RSe563p - 1./81.*pow(mp2,2)*RSp563e*RSe453p
          + 1./54.*pow(mp2,2)*pow(RSp563e,2) + 1./81.*pow(mp2,2)*
         RSp453e*RS345pe - 1./81.*pow(mp2,2)*RSp453e*RS356pe - 1./81.*
         pow(mp2,2)*RSp453e*RSe563p + 1./81.*pow(mp2,2)*RSp453e*RSe453p
          - 1./27.*pow(mp2,2)*RSp453e*RSp563e + 1./54.*pow(mp2,2)*pow(
         RSp453e,2) + 1./81.*pow(mp2,2)*RS145pe*RSp563e - 1./81.*pow(
         mp2,2)*RS145pe*RSp453e - 1./81.*pow(mp2,2)*RS156pe*RSp563e + 1.
         /81.*pow(mp2,2)*RS156pe*RSp453e - 1./81.*pow(mp2,2)*RSe561p*
         RSp563e + 1./81.*pow(mp2,2)*RSe561p*RSp453e + 1./81.*pow(
         mp2,2)*RSe451p*RSp563e - 1./81.*pow(mp2,2)*RSe451p*RSp453e + 1.
         /81.*pow(mp2,2)*RSp561e*RS345pe - 1./81.*pow(mp2,2)*RSp561e*
         RS356pe - 1./81.*pow(mp2,2)*RSp561e*RSe563p + 1./81.*pow(
         mp2,2)*RSp561e*RSe453p - 1./27.*pow(mp2,2)*RSp561e*RSp563e + 1.
         /27.*pow(mp2,2)*RSp561e*RSp453e );

      M13p6x23R +=  + Abeps(m46,mu2)*pi16 * (  - 1./81.*pow(mp2,2)*
         RSp561e*RS145pe + 1./81.*pow(mp2,2)*RSp561e*RS156pe + 1./81.*
         pow(mp2,2)*RSp561e*RSe561p - 1./81.*pow(mp2,2)*RSp561e*RSe451p
          + 1./54.*pow(mp2,2)*pow(RSp561e,2) - 1./81.*pow(mp2,2)*
         RSp451e*RS345pe + 1./81.*pow(mp2,2)*RSp451e*RS356pe + 1./81.*
         pow(mp2,2)*RSp451e*RSe563p - 1./81.*pow(mp2,2)*RSp451e*RSe453p
          + 1./27.*pow(mp2,2)*RSp451e*RSp563e - 1./27.*pow(mp2,2)*
         RSp451e*RSp453e + 1./81.*pow(mp2,2)*RSp451e*RS145pe - 1./81.*
         pow(mp2,2)*RSp451e*RS156pe - 1./81.*pow(mp2,2)*RSp451e*RSe561p
          + 1./81.*pow(mp2,2)*RSp451e*RSe451p - 1./27.*pow(mp2,2)*
         RSp451e*RSp561e + 1./54.*pow(mp2,2)*pow(RSp451e,2) + 1./81.*
         me2*mp2*RSp563e*RSe563p - 1./81.*me2*mp2*RSp563e*RSe453p - 1./
         81.*me2*mp2*RSp453e*RSe563p + 1./81.*me2*mp2*RSp453e*RSe453p
          - 1./81.*me2*mp2*RSe561p*RSp563e + 1./81.*me2*mp2*RSe561p*
         RSp453e + 1./81.*me2*mp2*RSe451p*RSp563e - 1./81.*me2*mp2*
         RSe451p*RSp453e );

      M13p6x23R +=  + Abeps(m46,mu2)*pi16 * (  - 1./81.*me2*mp2*RSp561e
         *RSe563p + 1./81.*me2*mp2*RSp561e*RSe453p + 1./81.*me2*mp2*
         RSp561e*RSe561p - 1./81.*me2*mp2*RSp561e*RSe451p + 1./81.*me2*
         mp2*RSp451e*RSe563p - 1./81.*me2*mp2*RSp451e*RSe453p - 1./81.*
         me2*mp2*RSp451e*RSe561p + 1./81.*me2*mp2*RSp451e*RSe451p - 1./
         81.*pow(me2,2)*RSe563p*RS345pe + 1./81.*pow(me2,2)*RSe563p*
         RS356pe + 1./54.*pow(me2,2)*pow(RSe563p,2) + 1./81.*pow(me2,2)
         *RSe453p*RS345pe - 1./81.*pow(me2,2)*RSe453p*RS356pe - 1./27.*
         pow(me2,2)*RSe453p*RSe563p + 1./54.*pow(me2,2)*pow(RSe453p,2)
          + 1./81.*pow(me2,2)*RSp563e*RSe563p - 1./81.*pow(me2,2)*
         RSp563e*RSe453p - 1./81.*pow(me2,2)*RSp453e*RSe563p + 1./81.*
         pow(me2,2)*RSp453e*RSe453p + 1./81.*pow(me2,2)*RS145pe*RSe563p
          - 1./81.*pow(me2,2)*RS145pe*RSe453p - 1./81.*pow(me2,2)*
         RS156pe*RSe563p + 1./81.*pow(me2,2)*RS156pe*RSe453p + 1./81.*
         pow(me2,2)*RSe561p*RS345pe - 1./81.*pow(me2,2)*RSe561p*RS356pe
          - 1./27.*pow(me2,2)*RSe561p*RSe563p );

      M13p6x23R +=  + Abeps(m46,mu2)*pi16 * ( 1./27.*pow(me2,2)*RSe561p
         *RSe453p - 1./81.*pow(me2,2)*RSe561p*RSp563e + 1./81.*pow(
         me2,2)*RSe561p*RSp453e - 1./81.*pow(me2,2)*RSe561p*RS145pe + 1.
         /81.*pow(me2,2)*RSe561p*RS156pe + 1./54.*pow(me2,2)*pow(
         RSe561p,2) - 1./81.*pow(me2,2)*RSe451p*RS345pe + 1./81.*pow(
         me2,2)*RSe451p*RS356pe + 1./27.*pow(me2,2)*RSe451p*RSe563p - 1.
         /27.*pow(me2,2)*RSe451p*RSe453p + 1./81.*pow(me2,2)*RSe451p*
         RSp563e - 1./81.*pow(me2,2)*RSe451p*RSp453e + 1./81.*pow(
         me2,2)*RSe451p*RS145pe - 1./81.*pow(me2,2)*RSe451p*RS156pe - 1.
         /27.*pow(me2,2)*RSe451p*RSe561p + 1./54.*pow(me2,2)*pow(
         RSe451p,2) - 1./81.*pow(me2,2)*RSp561e*RSe563p + 1./81.*pow(
         me2,2)*RSp561e*RSe453p + 1./81.*pow(me2,2)*RSp561e*RSe561p - 1.
         /81.*pow(me2,2)*RSp561e*RSe451p + 1./81.*pow(me2,2)*RSp451e*
         RSe563p - 1./81.*pow(me2,2)*RSp451e*RSe453p - 1./81.*pow(
         me2,2)*RSp451e*RSe561p + 1./81.*pow(me2,2)*RSp451e*RSe451p - 1.
         /81.*m66*mp2*RSp563e*RS345pe );

      M13p6x23R +=  + Abeps(m46,mu2)*pi16 * (  - 1./81.*m66*mp2*RSp563e
         *RSe453p - 1./81.*m66*mp2*RSp453e*RS345pe - 1./81.*m66*mp2*
         RSp453e*RS356pe - 1./81.*m66*mp2*RSp453e*RSe563p - 1./81.*m66*
         mp2*RSp453e*RSe453p - 2./81.*m66*mp2*RSp453e*RSp563e - 1./81.*
         m66*mp2*pow(RSp453e,2) + 1./81.*m66*mp2*RS145pe*RSp563e + 1./
         81.*m66*mp2*RS145pe*RSp453e + 1./81.*m66*mp2*RS156pe*RSp453e
          + 1./81.*m66*mp2*RSe561p*RSp453e + 1./81.*m66*mp2*RSe451p*
         RSp563e + 1./81.*m66*mp2*RSe451p*RSp453e + 1./81.*m66*mp2*
         RSp561e*RS345pe + 1./81.*m66*mp2*RSp561e*RSe453p + 2./81.*m66*
         mp2*RSp561e*RSp453e - 1./81.*m66*mp2*RSp561e*RS145pe - 1./81.*
         m66*mp2*RSp561e*RSe451p + 1./81.*m66*mp2*RSp451e*RS345pe + 1./
         81.*m66*mp2*RSp451e*RS356pe + 1./81.*m66*mp2*RSp451e*RSe563p
          + 1./81.*m66*mp2*RSp451e*RSe453p + 2./81.*m66*mp2*RSp451e*
         RSp563e + 2./81.*m66*mp2*RSp451e*RSp453e - 1./81.*m66*mp2*
         RSp451e*RS145pe - 1./81.*m66*mp2*RSp451e*RS156pe - 1./81.*m66*
         mp2*RSp451e*RSe561p );

      M13p6x23R +=  + Abeps(m46,mu2)*pi16 * (  - 1./81.*m66*mp2*RSp451e
         *RSe451p - 2./81.*m66*mp2*RSp451e*RSp561e - 1./81.*m66*mp2*
         pow(RSp451e,2) - 1./81.*m66*me2*RSe563p*RS345pe - 1./81.*m66*
         me2*RSe453p*RS345pe - 1./81.*m66*me2*RSe453p*RS356pe - 2./81.*
         m66*me2*RSe453p*RSe563p - 1./81.*m66*me2*pow(RSe453p,2) - 1./
         81.*m66*me2*RSp563e*RSe453p - 1./81.*m66*me2*RSp453e*RSe563p
          - 1./81.*m66*me2*RSp453e*RSe453p + 1./81.*m66*me2*RS145pe*
         RSe563p + 1./81.*m66*me2*RS145pe*RSe453p + 1./81.*m66*me2*
         RS156pe*RSe453p + 1./81.*m66*me2*RSe561p*RS345pe + 2./81.*m66*
         me2*RSe561p*RSe453p + 1./81.*m66*me2*RSe561p*RSp453e - 1./81.*
         m66*me2*RSe561p*RS145pe + 1./81.*m66*me2*RSe451p*RS345pe + 1./
         81.*m66*me2*RSe451p*RS356pe + 2./81.*m66*me2*RSe451p*RSe563p
          + 2./81.*m66*me2*RSe451p*RSe453p + 1./81.*m66*me2*RSe451p*
         RSp563e + 1./81.*m66*me2*RSe451p*RSp453e - 1./81.*m66*me2*
         RSe451p*RS145pe - 1./81.*m66*me2*RSe451p*RS156pe - 2./81.*m66*
         me2*RSe451p*RSe561p );

      M13p6x23R +=  + Abeps(m46,mu2)*pi16 * (  - 1./81.*m66*me2*pow(
         RSe451p,2) + 1./81.*m66*me2*RSp561e*RSe453p - 1./81.*m66*me2*
         RSp561e*RSe451p + 1./81.*m66*me2*RSp451e*RSe563p + 1./81.*m66*
         me2*RSp451e*RSe453p - 1./81.*m66*me2*RSp451e*RSe561p - 1./81.*
         m66*me2*RSp451e*RSe451p - 1./81.*m44*mp2*RSp563e*RS345pe - 1./
         81.*m44*mp2*RSp563e*RS356pe - 1./81.*m44*mp2*RSp563e*RSe563p
          - 1./81.*m44*mp2*RSp563e*RSe453p - 1./81.*m44*mp2*pow(
         RSp563e,2) - 1./81.*m44*mp2*RSp453e*RS356pe - 1./81.*m44*mp2*
         RSp453e*RSe563p - 2./81.*m44*mp2*RSp453e*RSp563e + 1./81.*m44*
         mp2*RS145pe*RSp563e + 1./81.*m44*mp2*RS156pe*RSp563e + 1./81.*
         m44*mp2*RS156pe*RSp453e + 1./81.*m44*mp2*RSe561p*RSp563e + 1./
         81.*m44*mp2*RSe561p*RSp453e + 1./81.*m44*mp2*RSe451p*RSp563e
          + 1./81.*m44*mp2*RSp561e*RS345pe + 1./81.*m44*mp2*RSp561e*
         RS356pe + 1./81.*m44*mp2*RSp561e*RSe563p + 1./81.*m44*mp2*
         RSp561e*RSe453p + 2./81.*m44*mp2*RSp561e*RSp563e + 2./81.*m44*
         mp2*RSp561e*RSp453e );

      M13p6x23R +=  + Abeps(m46,mu2)*pi16 * (  - 1./81.*m44*mp2*RSp561e
         *RS145pe - 1./81.*m44*mp2*RSp561e*RS156pe - 1./81.*m44*mp2*
         RSp561e*RSe561p - 1./81.*m44*mp2*RSp561e*RSe451p - 1./81.*m44*
         mp2*pow(RSp561e,2) + 1./81.*m44*mp2*RSp451e*RS356pe + 1./81.*
         m44*mp2*RSp451e*RSe563p + 2./81.*m44*mp2*RSp451e*RSp563e - 1./
         81.*m44*mp2*RSp451e*RS156pe - 1./81.*m44*mp2*RSp451e*RSe561p
          - 2./81.*m44*mp2*RSp451e*RSp561e - 1./81.*m44*me2*RSe563p*
         RS345pe - 1./81.*m44*me2*RSe563p*RS356pe - 1./81.*m44*me2*pow(
         RSe563p,2) - 1./81.*m44*me2*RSe453p*RS356pe - 2./81.*m44*me2*
         RSe453p*RSe563p - 1./81.*m44*me2*RSp563e*RSe563p - 1./81.*m44*
         me2*RSp563e*RSe453p - 1./81.*m44*me2*RSp453e*RSe563p + 1./81.*
         m44*me2*RS145pe*RSe563p + 1./81.*m44*me2*RS156pe*RSe563p + 1./
         81.*m44*me2*RS156pe*RSe453p + 1./81.*m44*me2*RSe561p*RS345pe
          + 1./81.*m44*me2*RSe561p*RS356pe + 2./81.*m44*me2*RSe561p*
         RSe563p + 2./81.*m44*me2*RSe561p*RSe453p + 1./81.*m44*me2*
         RSe561p*RSp563e );

      M13p6x23R +=  + Abeps(m46,mu2)*pi16 * ( 1./81.*m44*me2*RSe561p*
         RSp453e - 1./81.*m44*me2*RSe561p*RS145pe - 1./81.*m44*me2*
         RSe561p*RS156pe - 1./81.*m44*me2*pow(RSe561p,2) + 1./81.*m44*
         me2*RSe451p*RS356pe + 2./81.*m44*me2*RSe451p*RSe563p + 1./81.*
         m44*me2*RSe451p*RSp563e - 1./81.*m44*me2*RSe451p*RS156pe - 2./
         81.*m44*me2*RSe451p*RSe561p + 1./81.*m44*me2*RSp561e*RSe563p
          + 1./81.*m44*me2*RSp561e*RSe453p - 1./81.*m44*me2*RSp561e*
         RSe561p - 1./81.*m44*me2*RSp561e*RSe451p + 1./81.*m44*me2*
         RSp451e*RSe563p - 1./81.*m44*me2*RSp451e*RSe561p - 1./81.*m33*
         mp2*RSp563e*RSe563p + 1./81.*m33*mp2*RSp563e*RSe453p - 1./81.*
         m33*mp2*pow(RSp563e,2) + 1./81.*m33*mp2*RSp453e*RSe563p - 1./
         81.*m33*mp2*RSp453e*RSe453p + 2./81.*m33*mp2*RSp453e*RSp563e
          - 1./81.*m33*mp2*pow(RSp453e,2) + 1./81.*m33*mp2*RS145pe*
         RSp563e - 1./81.*m33*mp2*RS145pe*RSp453e - 1./81.*m33*mp2*
         RS156pe*RSp563e + 1./81.*m33*mp2*RS156pe*RSp453e - 1./81.*m33*
         mp2*RSe561p*RSp563e );

      M13p6x23R +=  + Abeps(m46,mu2)*pi16 * ( 1./81.*m33*mp2*RSe561p*
         RSp453e + 1./81.*m33*mp2*RSe451p*RSp563e - 1./81.*m33*mp2*
         RSe451p*RSp453e + 2./81.*m33*mp2*RSp561e*RS345pe - 2./81.*m33*
         mp2*RSp561e*RS356pe - 1./81.*m33*mp2*RSp561e*RSe563p + 1./81.*
         m33*mp2*RSp561e*RSe453p - 2./81.*m33*mp2*RSp561e*RSp563e + 2./
         81.*m33*mp2*RSp561e*RSp453e - 2./81.*m33*mp2*RSp451e*RS345pe
          + 2./81.*m33*mp2*RSp451e*RS356pe + 1./81.*m33*mp2*RSp451e*
         RSe563p - 1./81.*m33*mp2*RSp451e*RSe453p + 2./81.*m33*mp2*
         RSp451e*RSp563e - 2./81.*m33*mp2*RSp451e*RSp453e - 1./81.*m33*
         me2*pow(RSe563p,2) + 2./81.*m33*me2*RSe453p*RSe563p - 1./81.*
         m33*me2*pow(RSe453p,2) - 1./81.*m33*me2*RSp563e*RSe563p + 1./
         81.*m33*me2*RSp563e*RSe453p + 1./81.*m33*me2*RSp453e*RSe563p
          - 1./81.*m33*me2*RSp453e*RSe453p + 1./81.*m33*me2*RS145pe*
         RSe563p - 1./81.*m33*me2*RS145pe*RSe453p - 1./81.*m33*me2*
         RS156pe*RSe563p + 1./81.*m33*me2*RS156pe*RSe453p + 2./81.*m33*
         me2*RSe561p*RS345pe );

      M13p6x23R +=  + Abeps(m46,mu2)*pi16 * (  - 2./81.*m33*me2*RSe561p
         *RS356pe - 2./81.*m33*me2*RSe561p*RSe563p + 2./81.*m33*me2*
         RSe561p*RSe453p - 1./81.*m33*me2*RSe561p*RSp563e + 1./81.*m33*
         me2*RSe561p*RSp453e - 2./81.*m33*me2*RSe451p*RS345pe + 2./81.*
         m33*me2*RSe451p*RS356pe + 2./81.*m33*me2*RSe451p*RSe563p - 2./
         81.*m33*me2*RSe451p*RSe453p + 1./81.*m33*me2*RSe451p*RSp563e
          - 1./81.*m33*me2*RSe451p*RSp453e - 1./81.*m33*me2*RSp561e*
         RSe563p + 1./81.*m33*me2*RSp561e*RSe453p + 1./81.*m33*me2*
         RSp451e*RSe563p - 1./81.*m33*me2*RSp451e*RSe453p - 13./72.*m33
         *m66 - 1./162.*m33*m66*pow(RS345pe,2) - 1./81.*m33*m66*RS356pe
         *RS345pe + 1./81.*m33*m66*RSe453p*RSe563p + 1./162.*m33*m66*
         pow(RSe453p,2) + 1./81.*m33*m66*RSp563e*RSe453p + 1./81.*m33*
         m66*RSp453e*RSe563p + 1./81.*m33*m66*RSp453e*RSe453p + 1./81.*
         m33*m66*RSp453e*RSp563e + 1./162.*m33*m66*pow(RSp453e,2) + 2./
         81.*m33*m66*RS145pe*RS345pe + 2./81.*m33*m66*RS145pe*RS356pe
          + 1./81.*m33*m66*RS145pe*RSe563p );

      M13p6x23R +=  + Abeps(m46,mu2)*pi16 * ( 1./81.*m33*m66*RS145pe*
         RSe453p + 1./81.*m33*m66*RS145pe*RSp563e + 1./81.*m33*m66*
         RS145pe*RSp453e + 2./81.*m33*m66*RS156pe*RS345pe + 1./81.*m33*
         m66*RS156pe*RSe453p + 1./81.*m33*m66*RS156pe*RSp453e + 2./81.*
         m33*m66*RSe561p*RS345pe + 1./81.*m33*m66*RSe561p*RSe453p + 1./
         81.*m33*m66*RSe561p*RSp453e + 2./81.*m33*m66*RSe451p*RS345pe
          + 2./81.*m33*m66*RSe451p*RS356pe + 1./81.*m33*m66*RSe451p*
         RSe563p + 1./81.*m33*m66*RSe451p*RSe453p + 1./81.*m33*m66*
         RSe451p*RSp563e + 1./81.*m33*m66*RSe451p*RSp453e + 2./81.*m33*
         m66*RSp561e*RS345pe + 1./81.*m33*m66*RSp561e*RSe453p + 1./81.*
         m33*m66*RSp561e*RSp453e + 2./81.*m33*m66*RSp451e*RS345pe + 2./
         81.*m33*m66*RSp451e*RS356pe + 1./81.*m33*m66*RSp451e*RSe563p
          + 1./81.*m33*m66*RSp451e*RSe453p + 1./81.*m33*m66*RSp451e*
         RSp563e + 1./81.*m33*m66*RSp451e*RSp453e - 13./72.*m33*m44 - 1.
         /81.*m33*m44*RS356pe*RS345pe - 1./162.*m33*m44*pow(RS356pe,2)
          + 1./162.*m33*m44*pow(RSe563p,2) );

      M13p6x23R +=  + Abeps(m46,mu2)*pi16 * ( 1./81.*m33*m44*RSe453p*
         RSe563p + 1./81.*m33*m44*RSp563e*RSe563p + 1./81.*m33*m44*
         RSp563e*RSe453p + 1./162.*m33*m44*pow(RSp563e,2) + 1./81.*m33*
         m44*RSp453e*RSe563p + 1./81.*m33*m44*RSp453e*RSp563e + 2./81.*
         m33*m44*RS145pe*RS356pe + 1./81.*m33*m44*RS145pe*RSe563p + 1./
         81.*m33*m44*RS145pe*RSp563e + 2./81.*m33*m44*RS156pe*RS345pe
          + 2./81.*m33*m44*RS156pe*RS356pe + 1./81.*m33*m44*RS156pe*
         RSe563p + 1./81.*m33*m44*RS156pe*RSe453p + 1./81.*m33*m44*
         RS156pe*RSp563e + 1./81.*m33*m44*RS156pe*RSp453e + 2./81.*m33*
         m44*RSe561p*RS345pe + 2./81.*m33*m44*RSe561p*RS356pe + 1./81.*
         m33*m44*RSe561p*RSe563p + 1./81.*m33*m44*RSe561p*RSe453p + 1./
         81.*m33*m44*RSe561p*RSp563e + 1./81.*m33*m44*RSe561p*RSp453e
          + 2./81.*m33*m44*RSe451p*RS356pe + 1./81.*m33*m44*RSe451p*
         RSe563p + 1./81.*m33*m44*RSe451p*RSp563e + 2./81.*m33*m44*
         RSp561e*RS345pe + 2./81.*m33*m44*RSp561e*RS356pe + 1./81.*m33*
         m44*RSp561e*RSe563p );

      M13p6x23R +=  + Abeps(m46,mu2)*pi16 * ( 1./81.*m33*m44*RSp561e*
         RSe453p + 1./81.*m33*m44*RSp561e*RSp563e + 1./81.*m33*m44*
         RSp561e*RSp453e + 2./81.*m33*m44*RSp451e*RS356pe + 1./81.*m33*
         m44*RSp451e*RSe563p + 1./81.*m33*m44*RSp451e*RSp563e + 1./162.
         *pow(m33,2)*pow(RS345pe,2) - 1./81.*pow(m33,2)*RS356pe*RS345pe
          + 1./162.*pow(m33,2)*pow(RS356pe,2) - 2./81.*pow(m33,2)*
         RS145pe*RS345pe + 2./81.*pow(m33,2)*RS145pe*RS356pe + 2./81.*
         pow(m33,2)*RS156pe*RS345pe - 2./81.*pow(m33,2)*RS156pe*RS356pe
          + 2./81.*pow(m33,2)*RSe561p*RS345pe - 2./81.*pow(m33,2)*
         RSe561p*RS356pe - 2./81.*pow(m33,2)*RSe451p*RS345pe + 2./81.*
         pow(m33,2)*RSe451p*RS356pe + 2./81.*pow(m33,2)*RSp561e*RS345pe
          - 2./81.*pow(m33,2)*RSp561e*RS356pe - 2./81.*pow(m33,2)*
         RSp451e*RS345pe + 2./81.*pow(m33,2)*RSp451e*RS356pe + 2./81.*
         m11*mp2*RS145pe*RSp563e - 2./81.*m11*mp2*RS145pe*RSp453e - 2./
         81.*m11*mp2*RS156pe*RSp563e + 2./81.*m11*mp2*RS156pe*RSp453e
          - 1./81.*m11*mp2*RSe561p*RSp563e );

      M13p6x23R +=  + Abeps(m46,mu2)*pi16 * ( 1./81.*m11*mp2*RSe561p*
         RSp453e + 1./81.*m11*mp2*RSe451p*RSp563e - 1./81.*m11*mp2*
         RSe451p*RSp453e + 1./81.*m11*mp2*RSp561e*RS345pe - 1./81.*m11*
         mp2*RSp561e*RS356pe - 1./81.*m11*mp2*RSp561e*RSe563p + 1./81.*
         m11*mp2*RSp561e*RSe453p - 2./81.*m11*mp2*RSp561e*RSp563e + 2./
         81.*m11*mp2*RSp561e*RSp453e - 1./81.*m11*mp2*RSp561e*RSe561p
          + 1./81.*m11*mp2*RSp561e*RSe451p - 1./81.*m11*mp2*pow(
         RSp561e,2) - 1./81.*m11*mp2*RSp451e*RS345pe + 1./81.*m11*mp2*
         RSp451e*RS356pe + 1./81.*m11*mp2*RSp451e*RSe563p - 1./81.*m11*
         mp2*RSp451e*RSe453p + 2./81.*m11*mp2*RSp451e*RSp563e - 2./81.*
         m11*mp2*RSp451e*RSp453e + 1./81.*m11*mp2*RSp451e*RSe561p - 1./
         81.*m11*mp2*RSp451e*RSe451p + 2./81.*m11*mp2*RSp451e*RSp561e
          - 1./81.*m11*mp2*pow(RSp451e,2) + 2./81.*m11*me2*RS145pe*
         RSe563p - 2./81.*m11*me2*RS145pe*RSe453p - 2./81.*m11*me2*
         RS156pe*RSe563p + 2./81.*m11*me2*RS156pe*RSe453p + 1./81.*m11*
         me2*RSe561p*RS345pe );

      M13p6x23R +=  + Abeps(m46,mu2)*pi16 * (  - 1./81.*m11*me2*RSe561p
         *RS356pe - 2./81.*m11*me2*RSe561p*RSe563p + 2./81.*m11*me2*
         RSe561p*RSe453p - 1./81.*m11*me2*RSe561p*RSp563e + 1./81.*m11*
         me2*RSe561p*RSp453e - 1./81.*m11*me2*pow(RSe561p,2) - 1./81.*
         m11*me2*RSe451p*RS345pe + 1./81.*m11*me2*RSe451p*RS356pe + 2./
         81.*m11*me2*RSe451p*RSe563p - 2./81.*m11*me2*RSe451p*RSe453p
          + 1./81.*m11*me2*RSe451p*RSp563e - 1./81.*m11*me2*RSe451p*
         RSp453e + 2./81.*m11*me2*RSe451p*RSe561p - 1./81.*m11*me2*pow(
         RSe451p,2) - 1./81.*m11*me2*RSp561e*RSe563p + 1./81.*m11*me2*
         RSp561e*RSe453p - 1./81.*m11*me2*RSp561e*RSe561p + 1./81.*m11*
         me2*RSp561e*RSe451p + 1./81.*m11*me2*RSp451e*RSe563p - 1./81.*
         m11*me2*RSp451e*RSe453p + 1./81.*m11*me2*RSp451e*RSe561p - 1./
         81.*m11*me2*RSp451e*RSe451p - 13./72.*m11*m66 + 2./81.*m11*m66
         *RS145pe*RS345pe + 2./81.*m11*m66*RS145pe*RS356pe + 2./81.*m11
         *m66*RS145pe*RSe563p + 2./81.*m11*m66*RS145pe*RSe453p + 2./81.
         *m11*m66*RS145pe*RSp563e );

      M13p6x23R +=  + Abeps(m46,mu2)*pi16 * ( 2./81.*m11*m66*RS145pe*
         RSp453e - 1./162.*m11*m66*pow(RS145pe,2) + 2./81.*m11*m66*
         RS156pe*RS345pe + 2./81.*m11*m66*RS156pe*RSe453p + 2./81.*m11*
         m66*RS156pe*RSp453e - 1./81.*m11*m66*RS156pe*RS145pe + 1./81.*
         m11*m66*RSe561p*RS345pe + 1./81.*m11*m66*RSe561p*RSe453p + 1./
         81.*m11*m66*RSe561p*RSp453e + 1./81.*m11*m66*RSe451p*RS345pe
          + 1./81.*m11*m66*RSe451p*RS356pe + 1./81.*m11*m66*RSe451p*
         RSe563p + 1./81.*m11*m66*RSe451p*RSe453p + 1./81.*m11*m66*
         RSe451p*RSp563e + 1./81.*m11*m66*RSe451p*RSp453e + 1./81.*m11*
         m66*RSe451p*RSe561p + 1./162.*m11*m66*pow(RSe451p,2) + 1./81.*
         m11*m66*RSp561e*RS345pe + 1./81.*m11*m66*RSp561e*RSe453p + 1./
         81.*m11*m66*RSp561e*RSp453e + 1./81.*m11*m66*RSp561e*RSe451p
          + 1./81.*m11*m66*RSp451e*RS345pe + 1./81.*m11*m66*RSp451e*
         RS356pe + 1./81.*m11*m66*RSp451e*RSe563p + 1./81.*m11*m66*
         RSp451e*RSe453p + 1./81.*m11*m66*RSp451e*RSp563e + 1./81.*m11*
         m66*RSp451e*RSp453e );

      M13p6x23R +=  + Abeps(m46,mu2)*pi16 * ( 1./81.*m11*m66*RSp451e*
         RSe561p + 1./81.*m11*m66*RSp451e*RSe451p + 1./81.*m11*m66*
         RSp451e*RSp561e + 1./162.*m11*m66*pow(RSp451e,2) - 13./72.*m11
         *m44 + 2./81.*m11*m44*RS145pe*RS356pe + 2./81.*m11*m44*RS145pe
         *RSe563p + 2./81.*m11*m44*RS145pe*RSp563e + 2./81.*m11*m44*
         RS156pe*RS345pe + 2./81.*m11*m44*RS156pe*RS356pe + 2./81.*m11*
         m44*RS156pe*RSe563p + 2./81.*m11*m44*RS156pe*RSe453p + 2./81.*
         m11*m44*RS156pe*RSp563e + 2./81.*m11*m44*RS156pe*RSp453e - 1./
         81.*m11*m44*RS156pe*RS145pe - 1./162.*m11*m44*pow(RS156pe,2)
          + 1./81.*m11*m44*RSe561p*RS345pe + 1./81.*m11*m44*RSe561p*
         RS356pe + 1./81.*m11*m44*RSe561p*RSe563p + 1./81.*m11*m44*
         RSe561p*RSe453p + 1./81.*m11*m44*RSe561p*RSp563e + 1./81.*m11*
         m44*RSe561p*RSp453e + 1./162.*m11*m44*pow(RSe561p,2) + 1./81.*
         m11*m44*RSe451p*RS356pe + 1./81.*m11*m44*RSe451p*RSe563p + 1./
         81.*m11*m44*RSe451p*RSp563e + 1./81.*m11*m44*RSe451p*RSe561p
          + 1./81.*m11*m44*RSp561e*RS345pe );

      M13p6x23R +=  + Abeps(m46,mu2)*pi16 * ( 1./81.*m11*m44*RSp561e*
         RS356pe + 1./81.*m11*m44*RSp561e*RSe563p + 1./81.*m11*m44*
         RSp561e*RSe453p + 1./81.*m11*m44*RSp561e*RSp563e + 1./81.*m11*
         m44*RSp561e*RSp453e + 1./81.*m11*m44*RSp561e*RSe561p + 1./81.*
         m11*m44*RSp561e*RSe451p + 1./162.*m11*m44*pow(RSp561e,2) + 1./
         81.*m11*m44*RSp451e*RS356pe + 1./81.*m11*m44*RSp451e*RSe563p
          + 1./81.*m11*m44*RSp451e*RSp563e + 1./81.*m11*m44*RSp451e*
         RSe561p + 1./81.*m11*m44*RSp451e*RSp561e - 1./27.*m11*m33*
         RS145pe*RS345pe + 1./27.*m11*m33*RS145pe*RS356pe + 1./81.*m11*
         m33*RS145pe*RSe563p - 1./81.*m11*m33*RS145pe*RSe453p + 1./81.*
         m11*m33*RS145pe*RSp563e - 1./81.*m11*m33*RS145pe*RSp453e + 1./
         27.*m11*m33*RS156pe*RS345pe - 1./27.*m11*m33*RS156pe*RS356pe
          - 1./81.*m11*m33*RS156pe*RSe563p + 1./81.*m11*m33*RS156pe*
         RSe453p - 1./81.*m11*m33*RS156pe*RSp563e + 1./81.*m11*m33*
         RS156pe*RSp453e + 1./81.*m11*m33*RSe561p*RS345pe - 1./81.*m11*
         m33*RSe561p*RS356pe );

      M13p6x23R +=  + Abeps(m46,mu2)*pi16 * (  - 1./81.*m11*m33*RSe451p
         *RS345pe + 1./81.*m11*m33*RSe451p*RS356pe + 1./81.*m11*m33*
         RSp561e*RS345pe - 1./81.*m11*m33*RSp561e*RS356pe - 1./81.*m11*
         m33*RSp451e*RS345pe + 1./81.*m11*m33*RSp451e*RS356pe - 2./81.*
         pow(m11,2)*RS145pe*RS345pe + 2./81.*pow(m11,2)*RS145pe*RS356pe
          + 2./81.*pow(m11,2)*RS145pe*RSe563p - 2./81.*pow(m11,2)*
         RS145pe*RSe453p + 2./81.*pow(m11,2)*RS145pe*RSp563e - 2./81.*
         pow(m11,2)*RS145pe*RSp453e + 1./162.*pow(m11,2)*pow(RS145pe,2)
          + 2./81.*pow(m11,2)*RS156pe*RS345pe - 2./81.*pow(m11,2)*
         RS156pe*RS356pe - 2./81.*pow(m11,2)*RS156pe*RSe563p + 2./81.*
         pow(m11,2)*RS156pe*RSe453p - 2./81.*pow(m11,2)*RS156pe*RSp563e
          + 2./81.*pow(m11,2)*RS156pe*RSp453e - 1./81.*pow(m11,2)*
         RS156pe*RS145pe + 1./162.*pow(m11,2)*pow(RS156pe,2) );

      M13p6x23R +=  + Abeps(m55,mu2)*pi16 * (  - 1./27.*m55*mp2*RSp463e
         *RS346pe - 1./27.*m55*mp2*RSp463e*RSe463p - 1./27.*m55*mp2*
         pow(RSp463e,2) + 1./27.*m55*mp2*RS146pe*RSp463e + 1./81.*m55*
         mp2*RS546pe*RSp463e*RS346pe + 1./81.*m55*mp2*RS546pe*RSp463e*
         RSe463p + 1./81.*m55*mp2*RS546pe*pow(RSp463e,2) - 1./81.*m55*
         mp2*RS546pe*RS146pe*RSp463e + 1./27.*m55*mp2*RSe461p*RSp463e
          - 1./81.*m55*mp2*RSe461p*RS546pe*RSp463e + 1./27.*m55*mp2*
         RSp461e*RS346pe + 1./27.*m55*mp2*RSp461e*RSe463p + 2./27.*m55*
         mp2*RSp461e*RSp463e - 1./27.*m55*mp2*RSp461e*RS146pe - 1./81.*
         m55*mp2*RSp461e*RS546pe*RS346pe - 1./81.*m55*mp2*RSp461e*
         RS546pe*RSe463p - 2./81.*m55*mp2*RSp461e*RS546pe*RSp463e + 1./
         81.*m55*mp2*RSp461e*RS546pe*RS146pe - 1./27.*m55*mp2*RSp461e*
         RSe461p + 1./81.*m55*mp2*RSp461e*RSe461p*RS546pe - 1./27.*m55*
         mp2*pow(RSp461e,2) + 1./81.*m55*mp2*pow(RSp461e,2)*RS546pe - 1.
         /27.*m55*me2*RSe463p*RS346pe - 1./27.*m55*me2*pow(RSe463p,2)
          - 1./27.*m55*me2*RSp463e*RSe463p );

      M13p6x23R +=  + Abeps(m55,mu2)*pi16 * ( 1./27.*m55*me2*RS146pe*
         RSe463p + 1./81.*m55*me2*RS546pe*RSe463p*RS346pe + 1./81.*m55*
         me2*RS546pe*pow(RSe463p,2) + 1./81.*m55*me2*RS546pe*RSp463e*
         RSe463p - 1./81.*m55*me2*RS546pe*RS146pe*RSe463p + 1./27.*m55*
         me2*RSe461p*RS346pe + 2./27.*m55*me2*RSe461p*RSe463p + 1./27.*
         m55*me2*RSe461p*RSp463e - 1./27.*m55*me2*RSe461p*RS146pe - 1./
         81.*m55*me2*RSe461p*RS546pe*RS346pe - 2./81.*m55*me2*RSe461p*
         RS546pe*RSe463p - 1./81.*m55*me2*RSe461p*RS546pe*RSp463e + 1./
         81.*m55*me2*RSe461p*RS546pe*RS146pe - 1./27.*m55*me2*pow(
         RSe461p,2) + 1./81.*m55*me2*pow(RSe461p,2)*RS546pe + 1./27.*
         m55*me2*RSp461e*RSe463p - 1./81.*m55*me2*RSp461e*RS546pe*
         RSe463p - 1./27.*m55*me2*RSp461e*RSe461p + 1./81.*m55*me2*
         RSp461e*RSe461p*RS546pe - 13./72.*m33*m55 - 1./54.*m33*m55*
         pow(RS346pe,2) + 1./54.*m33*m55*pow(RSe463p,2) + 1./27.*m33*
         m55*RSp463e*RSe463p + 1./54.*m33*m55*pow(RSp463e,2) + 2./27.*
         m33*m55*RS146pe*RS346pe );

      M13p6x23R +=  + Abeps(m55,mu2)*pi16 * ( 1./27.*m33*m55*RS146pe*
         RSe463p + 1./27.*m33*m55*RS146pe*RSp463e + 13./216.*m33*m55*
         RS546pe + 1./162.*m33*m55*RS546pe*pow(RS346pe,2) - 1./162.*m33
         *m55*RS546pe*pow(RSe463p,2) - 1./81.*m33*m55*RS546pe*RSp463e*
         RSe463p - 1./162.*m33*m55*RS546pe*pow(RSp463e,2) - 2./81.*m33*
         m55*RS546pe*RS146pe*RS346pe - 1./81.*m33*m55*RS546pe*RS146pe*
         RSe463p - 1./81.*m33*m55*RS546pe*RS146pe*RSp463e + 2./27.*m33*
         m55*RSe461p*RS346pe + 1./27.*m33*m55*RSe461p*RSe463p + 1./27.*
         m33*m55*RSe461p*RSp463e - 2./81.*m33*m55*RSe461p*RS546pe*
         RS346pe - 1./81.*m33*m55*RSe461p*RS546pe*RSe463p - 1./81.*m33*
         m55*RSe461p*RS546pe*RSp463e + 2./27.*m33*m55*RSp461e*RS346pe
          + 1./27.*m33*m55*RSp461e*RSe463p + 1./27.*m33*m55*RSp461e*
         RSp463e - 2./81.*m33*m55*RSp461e*RS546pe*RS346pe - 1./81.*m33*
         m55*RSp461e*RS546pe*RSe463p - 1./81.*m33*m55*RSp461e*RS546pe*
         RSp463e - 13./72.*m11*m55 + 2./27.*m11*m55*RS146pe*RS346pe + 2.
         /27.*m11*m55*RS146pe*RSe463p );

      M13p6x23R +=  + Abeps(m55,mu2)*pi16 * ( 2./27.*m11*m55*RS146pe*
         RSp463e - 1./54.*m11*m55*pow(RS146pe,2) + 13./216.*m11*m55*
         RS546pe - 2./81.*m11*m55*RS546pe*RS146pe*RS346pe - 2./81.*m11*
         m55*RS546pe*RS146pe*RSe463p - 2./81.*m11*m55*RS546pe*RS146pe*
         RSp463e + 1./162.*m11*m55*RS546pe*pow(RS146pe,2) + 1./27.*m11*
         m55*RSe461p*RS346pe + 1./27.*m11*m55*RSe461p*RSe463p + 1./27.*
         m11*m55*RSe461p*RSp463e - 1./81.*m11*m55*RSe461p*RS546pe*
         RS346pe - 1./81.*m11*m55*RSe461p*RS546pe*RSe463p - 1./81.*m11*
         m55*RSe461p*RS546pe*RSp463e + 1./54.*m11*m55*pow(RSe461p,2) - 
         1./162.*m11*m55*pow(RSe461p,2)*RS546pe + 1./27.*m11*m55*
         RSp461e*RS346pe + 1./27.*m11*m55*RSp461e*RSe463p + 1./27.*m11*
         m55*RSp461e*RSp463e - 1./81.*m11*m55*RSp461e*RS546pe*RS346pe
          - 1./81.*m11*m55*RSp461e*RS546pe*RSe463p - 1./81.*m11*m55*
         RSp461e*RS546pe*RSp463e + 1./27.*m11*m55*RSp461e*RSe461p - 1./
         81.*m11*m55*RSp461e*RSe461p*RS546pe + 1./54.*m11*m55*pow(
         RSp461e,2) );

      M13p6x23R +=  + Abeps(m55,mu2)*pi16 * (  - 1./162.*m11*m55*pow(
         RSp461e,2)*RS546pe );

      M13p6x23R +=  + Abeps(m56,mu2)*pi16 * (  - 1./81.*pow(mp2,2)*
         RSp463e*RS345pe + 1./81.*pow(mp2,2)*RSp463e*RS346pe + 1./81.*
         pow(mp2,2)*RSp463e*RSe463p - 1./81.*pow(mp2,2)*RSp463e*RSe453p
          + 1./54.*pow(mp2,2)*pow(RSp463e,2) + 1./81.*pow(mp2,2)*
         RSp453e*RS345pe - 1./81.*pow(mp2,2)*RSp453e*RS346pe - 1./81.*
         pow(mp2,2)*RSp453e*RSe463p + 1./81.*pow(mp2,2)*RSp453e*RSe453p
          - 1./27.*pow(mp2,2)*RSp453e*RSp463e + 1./54.*pow(mp2,2)*pow(
         RSp453e,2) + 1./81.*pow(mp2,2)*RS145pe*RSp463e - 1./81.*pow(
         mp2,2)*RS145pe*RSp453e - 1./81.*pow(mp2,2)*RS146pe*RSp463e + 1.
         /81.*pow(mp2,2)*RS146pe*RSp453e - 1./81.*pow(mp2,2)*RSe461p*
         RSp463e + 1./81.*pow(mp2,2)*RSe461p*RSp453e + 1./81.*pow(
         mp2,2)*RSe451p*RSp463e - 1./81.*pow(mp2,2)*RSe451p*RSp453e + 1.
         /81.*pow(mp2,2)*RSp461e*RS345pe - 1./81.*pow(mp2,2)*RSp461e*
         RS346pe - 1./81.*pow(mp2,2)*RSp461e*RSe463p + 1./81.*pow(
         mp2,2)*RSp461e*RSe453p - 1./27.*pow(mp2,2)*RSp461e*RSp463e + 1.
         /27.*pow(mp2,2)*RSp461e*RSp453e );

      M13p6x23R +=  + Abeps(m56,mu2)*pi16 * (  - 1./81.*pow(mp2,2)*
         RSp461e*RS145pe + 1./81.*pow(mp2,2)*RSp461e*RS146pe + 1./81.*
         pow(mp2,2)*RSp461e*RSe461p - 1./81.*pow(mp2,2)*RSp461e*RSe451p
          + 1./54.*pow(mp2,2)*pow(RSp461e,2) - 1./81.*pow(mp2,2)*
         RSp451e*RS345pe + 1./81.*pow(mp2,2)*RSp451e*RS346pe + 1./81.*
         pow(mp2,2)*RSp451e*RSe463p - 1./81.*pow(mp2,2)*RSp451e*RSe453p
          + 1./27.*pow(mp2,2)*RSp451e*RSp463e - 1./27.*pow(mp2,2)*
         RSp451e*RSp453e + 1./81.*pow(mp2,2)*RSp451e*RS145pe - 1./81.*
         pow(mp2,2)*RSp451e*RS146pe - 1./81.*pow(mp2,2)*RSp451e*RSe461p
          + 1./81.*pow(mp2,2)*RSp451e*RSe451p - 1./27.*pow(mp2,2)*
         RSp451e*RSp461e + 1./54.*pow(mp2,2)*pow(RSp451e,2) + 1./81.*
         me2*mp2*RSp463e*RSe463p - 1./81.*me2*mp2*RSp463e*RSe453p - 1./
         81.*me2*mp2*RSp453e*RSe463p + 1./81.*me2*mp2*RSp453e*RSe453p
          - 1./81.*me2*mp2*RSe461p*RSp463e + 1./81.*me2*mp2*RSe461p*
         RSp453e + 1./81.*me2*mp2*RSe451p*RSp463e - 1./81.*me2*mp2*
         RSe451p*RSp453e );

      M13p6x23R +=  + Abeps(m56,mu2)*pi16 * (  - 1./81.*me2*mp2*RSp461e
         *RSe463p + 1./81.*me2*mp2*RSp461e*RSe453p + 1./81.*me2*mp2*
         RSp461e*RSe461p - 1./81.*me2*mp2*RSp461e*RSe451p + 1./81.*me2*
         mp2*RSp451e*RSe463p - 1./81.*me2*mp2*RSp451e*RSe453p - 1./81.*
         me2*mp2*RSp451e*RSe461p + 1./81.*me2*mp2*RSp451e*RSe451p - 1./
         81.*pow(me2,2)*RSe463p*RS345pe + 1./81.*pow(me2,2)*RSe463p*
         RS346pe + 1./54.*pow(me2,2)*pow(RSe463p,2) + 1./81.*pow(me2,2)
         *RSe453p*RS345pe - 1./81.*pow(me2,2)*RSe453p*RS346pe - 1./27.*
         pow(me2,2)*RSe453p*RSe463p + 1./54.*pow(me2,2)*pow(RSe453p,2)
          + 1./81.*pow(me2,2)*RSp463e*RSe463p - 1./81.*pow(me2,2)*
         RSp463e*RSe453p - 1./81.*pow(me2,2)*RSp453e*RSe463p + 1./81.*
         pow(me2,2)*RSp453e*RSe453p + 1./81.*pow(me2,2)*RS145pe*RSe463p
          - 1./81.*pow(me2,2)*RS145pe*RSe453p - 1./81.*pow(me2,2)*
         RS146pe*RSe463p + 1./81.*pow(me2,2)*RS146pe*RSe453p + 1./81.*
         pow(me2,2)*RSe461p*RS345pe - 1./81.*pow(me2,2)*RSe461p*RS346pe
          - 1./27.*pow(me2,2)*RSe461p*RSe463p );

      M13p6x23R +=  + Abeps(m56,mu2)*pi16 * ( 1./27.*pow(me2,2)*RSe461p
         *RSe453p - 1./81.*pow(me2,2)*RSe461p*RSp463e + 1./81.*pow(
         me2,2)*RSe461p*RSp453e - 1./81.*pow(me2,2)*RSe461p*RS145pe + 1.
         /81.*pow(me2,2)*RSe461p*RS146pe + 1./54.*pow(me2,2)*pow(
         RSe461p,2) - 1./81.*pow(me2,2)*RSe451p*RS345pe + 1./81.*pow(
         me2,2)*RSe451p*RS346pe + 1./27.*pow(me2,2)*RSe451p*RSe463p - 1.
         /27.*pow(me2,2)*RSe451p*RSe453p + 1./81.*pow(me2,2)*RSe451p*
         RSp463e - 1./81.*pow(me2,2)*RSe451p*RSp453e + 1./81.*pow(
         me2,2)*RSe451p*RS145pe - 1./81.*pow(me2,2)*RSe451p*RS146pe - 1.
         /27.*pow(me2,2)*RSe451p*RSe461p + 1./54.*pow(me2,2)*pow(
         RSe451p,2) - 1./81.*pow(me2,2)*RSp461e*RSe463p + 1./81.*pow(
         me2,2)*RSp461e*RSe453p + 1./81.*pow(me2,2)*RSp461e*RSe461p - 1.
         /81.*pow(me2,2)*RSp461e*RSe451p + 1./81.*pow(me2,2)*RSp451e*
         RSe463p - 1./81.*pow(me2,2)*RSp451e*RSe453p - 1./81.*pow(
         me2,2)*RSp451e*RSe461p + 1./81.*pow(me2,2)*RSp451e*RSe451p - 1.
         /81.*m66*mp2*RSp463e*RS345pe );

      M13p6x23R +=  + Abeps(m56,mu2)*pi16 * (  - 1./81.*m66*mp2*RSp463e
         *RSe453p - 1./81.*m66*mp2*RSp453e*RS345pe - 1./81.*m66*mp2*
         RSp453e*RS346pe - 1./81.*m66*mp2*RSp453e*RSe463p - 1./81.*m66*
         mp2*RSp453e*RSe453p - 2./81.*m66*mp2*RSp453e*RSp463e - 1./81.*
         m66*mp2*pow(RSp453e,2) + 1./81.*m66*mp2*RS145pe*RSp463e + 1./
         81.*m66*mp2*RS145pe*RSp453e + 1./81.*m66*mp2*RS146pe*RSp453e
          + 1./81.*m66*mp2*RSe461p*RSp453e + 1./81.*m66*mp2*RSe451p*
         RSp463e + 1./81.*m66*mp2*RSe451p*RSp453e + 1./81.*m66*mp2*
         RSp461e*RS345pe + 1./81.*m66*mp2*RSp461e*RSe453p + 2./81.*m66*
         mp2*RSp461e*RSp453e - 1./81.*m66*mp2*RSp461e*RS145pe - 1./81.*
         m66*mp2*RSp461e*RSe451p + 1./81.*m66*mp2*RSp451e*RS345pe + 1./
         81.*m66*mp2*RSp451e*RS346pe + 1./81.*m66*mp2*RSp451e*RSe463p
          + 1./81.*m66*mp2*RSp451e*RSe453p + 2./81.*m66*mp2*RSp451e*
         RSp463e + 2./81.*m66*mp2*RSp451e*RSp453e - 1./81.*m66*mp2*
         RSp451e*RS145pe - 1./81.*m66*mp2*RSp451e*RS146pe - 1./81.*m66*
         mp2*RSp451e*RSe461p );

      M13p6x23R +=  + Abeps(m56,mu2)*pi16 * (  - 1./81.*m66*mp2*RSp451e
         *RSe451p - 2./81.*m66*mp2*RSp451e*RSp461e - 1./81.*m66*mp2*
         pow(RSp451e,2) - 1./81.*m66*me2*RSe463p*RS345pe - 1./81.*m66*
         me2*RSe453p*RS345pe - 1./81.*m66*me2*RSe453p*RS346pe - 2./81.*
         m66*me2*RSe453p*RSe463p - 1./81.*m66*me2*pow(RSe453p,2) - 1./
         81.*m66*me2*RSp463e*RSe453p - 1./81.*m66*me2*RSp453e*RSe463p
          - 1./81.*m66*me2*RSp453e*RSe453p + 1./81.*m66*me2*RS145pe*
         RSe463p + 1./81.*m66*me2*RS145pe*RSe453p + 1./81.*m66*me2*
         RS146pe*RSe453p + 1./81.*m66*me2*RSe461p*RS345pe + 2./81.*m66*
         me2*RSe461p*RSe453p + 1./81.*m66*me2*RSe461p*RSp453e - 1./81.*
         m66*me2*RSe461p*RS145pe + 1./81.*m66*me2*RSe451p*RS345pe + 1./
         81.*m66*me2*RSe451p*RS346pe + 2./81.*m66*me2*RSe451p*RSe463p
          + 2./81.*m66*me2*RSe451p*RSe453p + 1./81.*m66*me2*RSe451p*
         RSp463e + 1./81.*m66*me2*RSe451p*RSp453e - 1./81.*m66*me2*
         RSe451p*RS145pe - 1./81.*m66*me2*RSe451p*RS146pe - 2./81.*m66*
         me2*RSe451p*RSe461p );

      M13p6x23R +=  + Abeps(m56,mu2)*pi16 * (  - 1./81.*m66*me2*pow(
         RSe451p,2) + 1./81.*m66*me2*RSp461e*RSe453p - 1./81.*m66*me2*
         RSp461e*RSe451p + 1./81.*m66*me2*RSp451e*RSe463p + 1./81.*m66*
         me2*RSp451e*RSe453p - 1./81.*m66*me2*RSp451e*RSe461p - 1./81.*
         m66*me2*RSp451e*RSe451p - 1./81.*m55*mp2*RSp463e*RS345pe - 1./
         81.*m55*mp2*RSp463e*RS346pe - 1./81.*m55*mp2*RSp463e*RSe463p
          - 1./81.*m55*mp2*RSp463e*RSe453p - 1./81.*m55*mp2*pow(
         RSp463e,2) - 1./81.*m55*mp2*RSp453e*RS346pe - 1./81.*m55*mp2*
         RSp453e*RSe463p - 2./81.*m55*mp2*RSp453e*RSp463e + 1./81.*m55*
         mp2*RS145pe*RSp463e + 1./81.*m55*mp2*RS146pe*RSp463e + 1./81.*
         m55*mp2*RS146pe*RSp453e + 1./81.*m55*mp2*RSe461p*RSp463e + 1./
         81.*m55*mp2*RSe461p*RSp453e + 1./81.*m55*mp2*RSe451p*RSp463e
          + 1./81.*m55*mp2*RSp461e*RS345pe + 1./81.*m55*mp2*RSp461e*
         RS346pe + 1./81.*m55*mp2*RSp461e*RSe463p + 1./81.*m55*mp2*
         RSp461e*RSe453p + 2./81.*m55*mp2*RSp461e*RSp463e + 2./81.*m55*
         mp2*RSp461e*RSp453e );

      M13p6x23R +=  + Abeps(m56,mu2)*pi16 * (  - 1./81.*m55*mp2*RSp461e
         *RS145pe - 1./81.*m55*mp2*RSp461e*RS146pe - 1./81.*m55*mp2*
         RSp461e*RSe461p - 1./81.*m55*mp2*RSp461e*RSe451p - 1./81.*m55*
         mp2*pow(RSp461e,2) + 1./81.*m55*mp2*RSp451e*RS346pe + 1./81.*
         m55*mp2*RSp451e*RSe463p + 2./81.*m55*mp2*RSp451e*RSp463e - 1./
         81.*m55*mp2*RSp451e*RS146pe - 1./81.*m55*mp2*RSp451e*RSe461p
          - 2./81.*m55*mp2*RSp451e*RSp461e - 1./81.*m55*me2*RSe463p*
         RS345pe - 1./81.*m55*me2*RSe463p*RS346pe - 1./81.*m55*me2*pow(
         RSe463p,2) - 1./81.*m55*me2*RSe453p*RS346pe - 2./81.*m55*me2*
         RSe453p*RSe463p - 1./81.*m55*me2*RSp463e*RSe463p - 1./81.*m55*
         me2*RSp463e*RSe453p - 1./81.*m55*me2*RSp453e*RSe463p + 1./81.*
         m55*me2*RS145pe*RSe463p + 1./81.*m55*me2*RS146pe*RSe463p + 1./
         81.*m55*me2*RS146pe*RSe453p + 1./81.*m55*me2*RSe461p*RS345pe
          + 1./81.*m55*me2*RSe461p*RS346pe + 2./81.*m55*me2*RSe461p*
         RSe463p + 2./81.*m55*me2*RSe461p*RSe453p + 1./81.*m55*me2*
         RSe461p*RSp463e );

      M13p6x23R +=  + Abeps(m56,mu2)*pi16 * ( 1./81.*m55*me2*RSe461p*
         RSp453e - 1./81.*m55*me2*RSe461p*RS145pe - 1./81.*m55*me2*
         RSe461p*RS146pe - 1./81.*m55*me2*pow(RSe461p,2) + 1./81.*m55*
         me2*RSe451p*RS346pe + 2./81.*m55*me2*RSe451p*RSe463p + 1./81.*
         m55*me2*RSe451p*RSp463e - 1./81.*m55*me2*RSe451p*RS146pe - 2./
         81.*m55*me2*RSe451p*RSe461p + 1./81.*m55*me2*RSp461e*RSe463p
          + 1./81.*m55*me2*RSp461e*RSe453p - 1./81.*m55*me2*RSp461e*
         RSe461p - 1./81.*m55*me2*RSp461e*RSe451p + 1./81.*m55*me2*
         RSp451e*RSe463p - 1./81.*m55*me2*RSp451e*RSe461p - 1./81.*m33*
         mp2*RSp463e*RSe463p + 1./81.*m33*mp2*RSp463e*RSe453p - 1./81.*
         m33*mp2*pow(RSp463e,2) + 1./81.*m33*mp2*RSp453e*RSe463p - 1./
         81.*m33*mp2*RSp453e*RSe453p + 2./81.*m33*mp2*RSp453e*RSp463e
          - 1./81.*m33*mp2*pow(RSp453e,2) + 1./81.*m33*mp2*RS145pe*
         RSp463e - 1./81.*m33*mp2*RS145pe*RSp453e - 1./81.*m33*mp2*
         RS146pe*RSp463e + 1./81.*m33*mp2*RS146pe*RSp453e - 1./81.*m33*
         mp2*RSe461p*RSp463e );

      M13p6x23R +=  + Abeps(m56,mu2)*pi16 * ( 1./81.*m33*mp2*RSe461p*
         RSp453e + 1./81.*m33*mp2*RSe451p*RSp463e - 1./81.*m33*mp2*
         RSe451p*RSp453e + 2./81.*m33*mp2*RSp461e*RS345pe - 2./81.*m33*
         mp2*RSp461e*RS346pe - 1./81.*m33*mp2*RSp461e*RSe463p + 1./81.*
         m33*mp2*RSp461e*RSe453p - 2./81.*m33*mp2*RSp461e*RSp463e + 2./
         81.*m33*mp2*RSp461e*RSp453e - 2./81.*m33*mp2*RSp451e*RS345pe
          + 2./81.*m33*mp2*RSp451e*RS346pe + 1./81.*m33*mp2*RSp451e*
         RSe463p - 1./81.*m33*mp2*RSp451e*RSe453p + 2./81.*m33*mp2*
         RSp451e*RSp463e - 2./81.*m33*mp2*RSp451e*RSp453e - 1./81.*m33*
         me2*pow(RSe463p,2) + 2./81.*m33*me2*RSe453p*RSe463p - 1./81.*
         m33*me2*pow(RSe453p,2) - 1./81.*m33*me2*RSp463e*RSe463p + 1./
         81.*m33*me2*RSp463e*RSe453p + 1./81.*m33*me2*RSp453e*RSe463p
          - 1./81.*m33*me2*RSp453e*RSe453p + 1./81.*m33*me2*RS145pe*
         RSe463p - 1./81.*m33*me2*RS145pe*RSe453p - 1./81.*m33*me2*
         RS146pe*RSe463p + 1./81.*m33*me2*RS146pe*RSe453p + 2./81.*m33*
         me2*RSe461p*RS345pe );

      M13p6x23R +=  + Abeps(m56,mu2)*pi16 * (  - 2./81.*m33*me2*RSe461p
         *RS346pe - 2./81.*m33*me2*RSe461p*RSe463p + 2./81.*m33*me2*
         RSe461p*RSe453p - 1./81.*m33*me2*RSe461p*RSp463e + 1./81.*m33*
         me2*RSe461p*RSp453e - 2./81.*m33*me2*RSe451p*RS345pe + 2./81.*
         m33*me2*RSe451p*RS346pe + 2./81.*m33*me2*RSe451p*RSe463p - 2./
         81.*m33*me2*RSe451p*RSe453p + 1./81.*m33*me2*RSe451p*RSp463e
          - 1./81.*m33*me2*RSe451p*RSp453e - 1./81.*m33*me2*RSp461e*
         RSe463p + 1./81.*m33*me2*RSp461e*RSe453p + 1./81.*m33*me2*
         RSp451e*RSe463p - 1./81.*m33*me2*RSp451e*RSe453p - 13./72.*m33
         *m66 - 1./162.*m33*m66*pow(RS345pe,2) - 1./81.*m33*m66*RS346pe
         *RS345pe + 1./81.*m33*m66*RSe453p*RSe463p + 1./162.*m33*m66*
         pow(RSe453p,2) + 1./81.*m33*m66*RSp463e*RSe453p + 1./81.*m33*
         m66*RSp453e*RSe463p + 1./81.*m33*m66*RSp453e*RSe453p + 1./81.*
         m33*m66*RSp453e*RSp463e + 1./162.*m33*m66*pow(RSp453e,2) + 2./
         81.*m33*m66*RS145pe*RS345pe + 2./81.*m33*m66*RS145pe*RS346pe
          + 1./81.*m33*m66*RS145pe*RSe463p );

      M13p6x23R +=  + Abeps(m56,mu2)*pi16 * ( 1./81.*m33*m66*RS145pe*
         RSe453p + 1./81.*m33*m66*RS145pe*RSp463e + 1./81.*m33*m66*
         RS145pe*RSp453e + 2./81.*m33*m66*RS146pe*RS345pe + 1./81.*m33*
         m66*RS146pe*RSe453p + 1./81.*m33*m66*RS146pe*RSp453e + 2./81.*
         m33*m66*RSe461p*RS345pe + 1./81.*m33*m66*RSe461p*RSe453p + 1./
         81.*m33*m66*RSe461p*RSp453e + 2./81.*m33*m66*RSe451p*RS345pe
          + 2./81.*m33*m66*RSe451p*RS346pe + 1./81.*m33*m66*RSe451p*
         RSe463p + 1./81.*m33*m66*RSe451p*RSe453p + 1./81.*m33*m66*
         RSe451p*RSp463e + 1./81.*m33*m66*RSe451p*RSp453e + 2./81.*m33*
         m66*RSp461e*RS345pe + 1./81.*m33*m66*RSp461e*RSe453p + 1./81.*
         m33*m66*RSp461e*RSp453e + 2./81.*m33*m66*RSp451e*RS345pe + 2./
         81.*m33*m66*RSp451e*RS346pe + 1./81.*m33*m66*RSp451e*RSe463p
          + 1./81.*m33*m66*RSp451e*RSe453p + 1./81.*m33*m66*RSp451e*
         RSp463e + 1./81.*m33*m66*RSp451e*RSp453e - 13./72.*m33*m55 - 1.
         /81.*m33*m55*RS346pe*RS345pe - 1./162.*m33*m55*pow(RS346pe,2)
          + 1./162.*m33*m55*pow(RSe463p,2) );

      M13p6x23R +=  + Abeps(m56,mu2)*pi16 * ( 1./81.*m33*m55*RSe453p*
         RSe463p + 1./81.*m33*m55*RSp463e*RSe463p + 1./81.*m33*m55*
         RSp463e*RSe453p + 1./162.*m33*m55*pow(RSp463e,2) + 1./81.*m33*
         m55*RSp453e*RSe463p + 1./81.*m33*m55*RSp453e*RSp463e + 2./81.*
         m33*m55*RS145pe*RS346pe + 1./81.*m33*m55*RS145pe*RSe463p + 1./
         81.*m33*m55*RS145pe*RSp463e + 2./81.*m33*m55*RS146pe*RS345pe
          + 2./81.*m33*m55*RS146pe*RS346pe + 1./81.*m33*m55*RS146pe*
         RSe463p + 1./81.*m33*m55*RS146pe*RSe453p + 1./81.*m33*m55*
         RS146pe*RSp463e + 1./81.*m33*m55*RS146pe*RSp453e + 2./81.*m33*
         m55*RSe461p*RS345pe + 2./81.*m33*m55*RSe461p*RS346pe + 1./81.*
         m33*m55*RSe461p*RSe463p + 1./81.*m33*m55*RSe461p*RSe453p + 1./
         81.*m33*m55*RSe461p*RSp463e + 1./81.*m33*m55*RSe461p*RSp453e
          + 2./81.*m33*m55*RSe451p*RS346pe + 1./81.*m33*m55*RSe451p*
         RSe463p + 1./81.*m33*m55*RSe451p*RSp463e + 2./81.*m33*m55*
         RSp461e*RS345pe + 2./81.*m33*m55*RSp461e*RS346pe + 1./81.*m33*
         m55*RSp461e*RSe463p );

      M13p6x23R +=  + Abeps(m56,mu2)*pi16 * ( 1./81.*m33*m55*RSp461e*
         RSe453p + 1./81.*m33*m55*RSp461e*RSp463e + 1./81.*m33*m55*
         RSp461e*RSp453e + 2./81.*m33*m55*RSp451e*RS346pe + 1./81.*m33*
         m55*RSp451e*RSe463p + 1./81.*m33*m55*RSp451e*RSp463e + 1./162.
         *pow(m33,2)*pow(RS345pe,2) - 1./81.*pow(m33,2)*RS346pe*RS345pe
          + 1./162.*pow(m33,2)*pow(RS346pe,2) - 2./81.*pow(m33,2)*
         RS145pe*RS345pe + 2./81.*pow(m33,2)*RS145pe*RS346pe + 2./81.*
         pow(m33,2)*RS146pe*RS345pe - 2./81.*pow(m33,2)*RS146pe*RS346pe
          + 2./81.*pow(m33,2)*RSe461p*RS345pe - 2./81.*pow(m33,2)*
         RSe461p*RS346pe - 2./81.*pow(m33,2)*RSe451p*RS345pe + 2./81.*
         pow(m33,2)*RSe451p*RS346pe + 2./81.*pow(m33,2)*RSp461e*RS345pe
          - 2./81.*pow(m33,2)*RSp461e*RS346pe - 2./81.*pow(m33,2)*
         RSp451e*RS345pe + 2./81.*pow(m33,2)*RSp451e*RS346pe + 2./81.*
         m11*mp2*RS145pe*RSp463e - 2./81.*m11*mp2*RS145pe*RSp453e - 2./
         81.*m11*mp2*RS146pe*RSp463e + 2./81.*m11*mp2*RS146pe*RSp453e
          - 1./81.*m11*mp2*RSe461p*RSp463e );

      M13p6x23R +=  + Abeps(m56,mu2)*pi16 * ( 1./81.*m11*mp2*RSe461p*
         RSp453e + 1./81.*m11*mp2*RSe451p*RSp463e - 1./81.*m11*mp2*
         RSe451p*RSp453e + 1./81.*m11*mp2*RSp461e*RS345pe - 1./81.*m11*
         mp2*RSp461e*RS346pe - 1./81.*m11*mp2*RSp461e*RSe463p + 1./81.*
         m11*mp2*RSp461e*RSe453p - 2./81.*m11*mp2*RSp461e*RSp463e + 2./
         81.*m11*mp2*RSp461e*RSp453e - 1./81.*m11*mp2*RSp461e*RSe461p
          + 1./81.*m11*mp2*RSp461e*RSe451p - 1./81.*m11*mp2*pow(
         RSp461e,2) - 1./81.*m11*mp2*RSp451e*RS345pe + 1./81.*m11*mp2*
         RSp451e*RS346pe + 1./81.*m11*mp2*RSp451e*RSe463p - 1./81.*m11*
         mp2*RSp451e*RSe453p + 2./81.*m11*mp2*RSp451e*RSp463e - 2./81.*
         m11*mp2*RSp451e*RSp453e + 1./81.*m11*mp2*RSp451e*RSe461p - 1./
         81.*m11*mp2*RSp451e*RSe451p + 2./81.*m11*mp2*RSp451e*RSp461e
          - 1./81.*m11*mp2*pow(RSp451e,2) + 2./81.*m11*me2*RS145pe*
         RSe463p - 2./81.*m11*me2*RS145pe*RSe453p - 2./81.*m11*me2*
         RS146pe*RSe463p + 2./81.*m11*me2*RS146pe*RSe453p + 1./81.*m11*
         me2*RSe461p*RS345pe );

      M13p6x23R +=  + Abeps(m56,mu2)*pi16 * (  - 1./81.*m11*me2*RSe461p
         *RS346pe - 2./81.*m11*me2*RSe461p*RSe463p + 2./81.*m11*me2*
         RSe461p*RSe453p - 1./81.*m11*me2*RSe461p*RSp463e + 1./81.*m11*
         me2*RSe461p*RSp453e - 1./81.*m11*me2*pow(RSe461p,2) - 1./81.*
         m11*me2*RSe451p*RS345pe + 1./81.*m11*me2*RSe451p*RS346pe + 2./
         81.*m11*me2*RSe451p*RSe463p - 2./81.*m11*me2*RSe451p*RSe453p
          + 1./81.*m11*me2*RSe451p*RSp463e - 1./81.*m11*me2*RSe451p*
         RSp453e + 2./81.*m11*me2*RSe451p*RSe461p - 1./81.*m11*me2*pow(
         RSe451p,2) - 1./81.*m11*me2*RSp461e*RSe463p + 1./81.*m11*me2*
         RSp461e*RSe453p - 1./81.*m11*me2*RSp461e*RSe461p + 1./81.*m11*
         me2*RSp461e*RSe451p + 1./81.*m11*me2*RSp451e*RSe463p - 1./81.*
         m11*me2*RSp451e*RSe453p + 1./81.*m11*me2*RSp451e*RSe461p - 1./
         81.*m11*me2*RSp451e*RSe451p - 13./72.*m11*m66 + 2./81.*m11*m66
         *RS145pe*RS345pe + 2./81.*m11*m66*RS145pe*RS346pe + 2./81.*m11
         *m66*RS145pe*RSe463p + 2./81.*m11*m66*RS145pe*RSe453p + 2./81.
         *m11*m66*RS145pe*RSp463e );

      M13p6x23R +=  + Abeps(m56,mu2)*pi16 * ( 2./81.*m11*m66*RS145pe*
         RSp453e - 1./162.*m11*m66*pow(RS145pe,2) + 2./81.*m11*m66*
         RS146pe*RS345pe + 2./81.*m11*m66*RS146pe*RSe453p + 2./81.*m11*
         m66*RS146pe*RSp453e - 1./81.*m11*m66*RS146pe*RS145pe + 1./81.*
         m11*m66*RSe461p*RS345pe + 1./81.*m11*m66*RSe461p*RSe453p + 1./
         81.*m11*m66*RSe461p*RSp453e + 1./81.*m11*m66*RSe451p*RS345pe
          + 1./81.*m11*m66*RSe451p*RS346pe + 1./81.*m11*m66*RSe451p*
         RSe463p + 1./81.*m11*m66*RSe451p*RSe453p + 1./81.*m11*m66*
         RSe451p*RSp463e + 1./81.*m11*m66*RSe451p*RSp453e + 1./81.*m11*
         m66*RSe451p*RSe461p + 1./162.*m11*m66*pow(RSe451p,2) + 1./81.*
         m11*m66*RSp461e*RS345pe + 1./81.*m11*m66*RSp461e*RSe453p + 1./
         81.*m11*m66*RSp461e*RSp453e + 1./81.*m11*m66*RSp461e*RSe451p
          + 1./81.*m11*m66*RSp451e*RS345pe + 1./81.*m11*m66*RSp451e*
         RS346pe + 1./81.*m11*m66*RSp451e*RSe463p + 1./81.*m11*m66*
         RSp451e*RSe453p + 1./81.*m11*m66*RSp451e*RSp463e + 1./81.*m11*
         m66*RSp451e*RSp453e );

      M13p6x23R +=  + Abeps(m56,mu2)*pi16 * ( 1./81.*m11*m66*RSp451e*
         RSe461p + 1./81.*m11*m66*RSp451e*RSe451p + 1./81.*m11*m66*
         RSp451e*RSp461e + 1./162.*m11*m66*pow(RSp451e,2) - 13./72.*m11
         *m55 + 2./81.*m11*m55*RS145pe*RS346pe + 2./81.*m11*m55*RS145pe
         *RSe463p + 2./81.*m11*m55*RS145pe*RSp463e + 2./81.*m11*m55*
         RS146pe*RS345pe + 2./81.*m11*m55*RS146pe*RS346pe + 2./81.*m11*
         m55*RS146pe*RSe463p + 2./81.*m11*m55*RS146pe*RSe453p + 2./81.*
         m11*m55*RS146pe*RSp463e + 2./81.*m11*m55*RS146pe*RSp453e - 1./
         81.*m11*m55*RS146pe*RS145pe - 1./162.*m11*m55*pow(RS146pe,2)
          + 1./81.*m11*m55*RSe461p*RS345pe + 1./81.*m11*m55*RSe461p*
         RS346pe + 1./81.*m11*m55*RSe461p*RSe463p + 1./81.*m11*m55*
         RSe461p*RSe453p + 1./81.*m11*m55*RSe461p*RSp463e + 1./81.*m11*
         m55*RSe461p*RSp453e + 1./162.*m11*m55*pow(RSe461p,2) + 1./81.*
         m11*m55*RSe451p*RS346pe + 1./81.*m11*m55*RSe451p*RSe463p + 1./
         81.*m11*m55*RSe451p*RSp463e + 1./81.*m11*m55*RSe451p*RSe461p
          + 1./81.*m11*m55*RSp461e*RS345pe );

      M13p6x23R +=  + Abeps(m56,mu2)*pi16 * ( 1./81.*m11*m55*RSp461e*
         RS346pe + 1./81.*m11*m55*RSp461e*RSe463p + 1./81.*m11*m55*
         RSp461e*RSe453p + 1./81.*m11*m55*RSp461e*RSp463e + 1./81.*m11*
         m55*RSp461e*RSp453e + 1./81.*m11*m55*RSp461e*RSe461p + 1./81.*
         m11*m55*RSp461e*RSe451p + 1./162.*m11*m55*pow(RSp461e,2) + 1./
         81.*m11*m55*RSp451e*RS346pe + 1./81.*m11*m55*RSp451e*RSe463p
          + 1./81.*m11*m55*RSp451e*RSp463e + 1./81.*m11*m55*RSp451e*
         RSe461p + 1./81.*m11*m55*RSp451e*RSp461e - 1./27.*m11*m33*
         RS145pe*RS345pe + 1./27.*m11*m33*RS145pe*RS346pe + 1./81.*m11*
         m33*RS145pe*RSe463p - 1./81.*m11*m33*RS145pe*RSe453p + 1./81.*
         m11*m33*RS145pe*RSp463e - 1./81.*m11*m33*RS145pe*RSp453e + 1./
         27.*m11*m33*RS146pe*RS345pe - 1./27.*m11*m33*RS146pe*RS346pe
          - 1./81.*m11*m33*RS146pe*RSe463p + 1./81.*m11*m33*RS146pe*
         RSe453p - 1./81.*m11*m33*RS146pe*RSp463e + 1./81.*m11*m33*
         RS146pe*RSp453e + 1./81.*m11*m33*RSe461p*RS345pe - 1./81.*m11*
         m33*RSe461p*RS346pe );

      M13p6x23R +=  + Abeps(m56,mu2)*pi16 * (  - 1./81.*m11*m33*RSe451p
         *RS345pe + 1./81.*m11*m33*RSe451p*RS346pe + 1./81.*m11*m33*
         RSp461e*RS345pe - 1./81.*m11*m33*RSp461e*RS346pe - 1./81.*m11*
         m33*RSp451e*RS345pe + 1./81.*m11*m33*RSp451e*RS346pe - 2./81.*
         pow(m11,2)*RS145pe*RS345pe + 2./81.*pow(m11,2)*RS145pe*RS346pe
          + 2./81.*pow(m11,2)*RS145pe*RSe463p - 2./81.*pow(m11,2)*
         RS145pe*RSe453p + 2./81.*pow(m11,2)*RS145pe*RSp463e - 2./81.*
         pow(m11,2)*RS145pe*RSp453e + 1./162.*pow(m11,2)*pow(RS145pe,2)
          + 2./81.*pow(m11,2)*RS146pe*RS345pe - 2./81.*pow(m11,2)*
         RS146pe*RS346pe - 2./81.*pow(m11,2)*RS146pe*RSe463p + 2./81.*
         pow(m11,2)*RS146pe*RSe453p - 2./81.*pow(m11,2)*RS146pe*RSp463e
          + 2./81.*pow(m11,2)*RS146pe*RSp453e - 1./81.*pow(m11,2)*
         RS146pe*RS145pe + 1./162.*pow(m11,2)*pow(RS146pe,2) );

      M13p6x23R +=  + Abeps(m66,mu2)*pi16 * (  - 1./27.*m66*mp2*RSp453e
         *RS345pe - 1./27.*m66*mp2*RSp453e*RSe453p - 1./27.*m66*mp2*
         pow(RSp453e,2) + 1./27.*m66*mp2*RS145pe*RSp453e + 1./81.*m66*
         mp2*RS645pe*RSp453e*RS345pe + 1./81.*m66*mp2*RS645pe*RSp453e*
         RSe453p + 1./81.*m66*mp2*RS645pe*pow(RSp453e,2) - 1./81.*m66*
         mp2*RS645pe*RS145pe*RSp453e + 1./27.*m66*mp2*RSe451p*RSp453e
          - 1./81.*m66*mp2*RSe451p*RS645pe*RSp453e + 1./27.*m66*mp2*
         RSp451e*RS345pe + 1./27.*m66*mp2*RSp451e*RSe453p + 2./27.*m66*
         mp2*RSp451e*RSp453e - 1./27.*m66*mp2*RSp451e*RS145pe - 1./81.*
         m66*mp2*RSp451e*RS645pe*RS345pe - 1./81.*m66*mp2*RSp451e*
         RS645pe*RSe453p - 2./81.*m66*mp2*RSp451e*RS645pe*RSp453e + 1./
         81.*m66*mp2*RSp451e*RS645pe*RS145pe - 1./27.*m66*mp2*RSp451e*
         RSe451p + 1./81.*m66*mp2*RSp451e*RSe451p*RS645pe - 1./27.*m66*
         mp2*pow(RSp451e,2) + 1./81.*m66*mp2*pow(RSp451e,2)*RS645pe - 1.
         /27.*m66*me2*RSe453p*RS345pe - 1./27.*m66*me2*pow(RSe453p,2)
          - 1./27.*m66*me2*RSp453e*RSe453p );

      M13p6x23R +=  + Abeps(m66,mu2)*pi16 * ( 1./27.*m66*me2*RS145pe*
         RSe453p + 1./81.*m66*me2*RS645pe*RSe453p*RS345pe + 1./81.*m66*
         me2*RS645pe*pow(RSe453p,2) + 1./81.*m66*me2*RS645pe*RSp453e*
         RSe453p - 1./81.*m66*me2*RS645pe*RS145pe*RSe453p + 1./27.*m66*
         me2*RSe451p*RS345pe + 2./27.*m66*me2*RSe451p*RSe453p + 1./27.*
         m66*me2*RSe451p*RSp453e - 1./27.*m66*me2*RSe451p*RS145pe - 1./
         81.*m66*me2*RSe451p*RS645pe*RS345pe - 2./81.*m66*me2*RSe451p*
         RS645pe*RSe453p - 1./81.*m66*me2*RSe451p*RS645pe*RSp453e + 1./
         81.*m66*me2*RSe451p*RS645pe*RS145pe - 1./27.*m66*me2*pow(
         RSe451p,2) + 1./81.*m66*me2*pow(RSe451p,2)*RS645pe + 1./27.*
         m66*me2*RSp451e*RSe453p - 1./81.*m66*me2*RSp451e*RS645pe*
         RSe453p - 1./27.*m66*me2*RSp451e*RSe451p + 1./81.*m66*me2*
         RSp451e*RSe451p*RS645pe - 13./72.*m33*m66 - 1./54.*m33*m66*
         pow(RS345pe,2) + 1./54.*m33*m66*pow(RSe453p,2) + 1./27.*m33*
         m66*RSp453e*RSe453p + 1./54.*m33*m66*pow(RSp453e,2) + 2./27.*
         m33*m66*RS145pe*RS345pe );

      M13p6x23R +=  + Abeps(m66,mu2)*pi16 * ( 1./27.*m33*m66*RS145pe*
         RSe453p + 1./27.*m33*m66*RS145pe*RSp453e + 13./216.*m33*m66*
         RS645pe + 1./162.*m33*m66*RS645pe*pow(RS345pe,2) - 1./162.*m33
         *m66*RS645pe*pow(RSe453p,2) - 1./81.*m33*m66*RS645pe*RSp453e*
         RSe453p - 1./162.*m33*m66*RS645pe*pow(RSp453e,2) - 2./81.*m33*
         m66*RS645pe*RS145pe*RS345pe - 1./81.*m33*m66*RS645pe*RS145pe*
         RSe453p - 1./81.*m33*m66*RS645pe*RS145pe*RSp453e + 2./27.*m33*
         m66*RSe451p*RS345pe + 1./27.*m33*m66*RSe451p*RSe453p + 1./27.*
         m33*m66*RSe451p*RSp453e - 2./81.*m33*m66*RSe451p*RS645pe*
         RS345pe - 1./81.*m33*m66*RSe451p*RS645pe*RSe453p - 1./81.*m33*
         m66*RSe451p*RS645pe*RSp453e + 2./27.*m33*m66*RSp451e*RS345pe
          + 1./27.*m33*m66*RSp451e*RSe453p + 1./27.*m33*m66*RSp451e*
         RSp453e - 2./81.*m33*m66*RSp451e*RS645pe*RS345pe - 1./81.*m33*
         m66*RSp451e*RS645pe*RSe453p - 1./81.*m33*m66*RSp451e*RS645pe*
         RSp453e - 13./72.*m11*m66 + 2./27.*m11*m66*RS145pe*RS345pe + 2.
         /27.*m11*m66*RS145pe*RSe453p );

      M13p6x23R +=  + Abeps(m66,mu2)*pi16 * ( 2./27.*m11*m66*RS145pe*
         RSp453e - 1./54.*m11*m66*pow(RS145pe,2) + 13./216.*m11*m66*
         RS645pe - 2./81.*m11*m66*RS645pe*RS145pe*RS345pe - 2./81.*m11*
         m66*RS645pe*RS145pe*RSe453p - 2./81.*m11*m66*RS645pe*RS145pe*
         RSp453e + 1./162.*m11*m66*RS645pe*pow(RS145pe,2) + 1./27.*m11*
         m66*RSe451p*RS345pe + 1./27.*m11*m66*RSe451p*RSe453p + 1./27.*
         m11*m66*RSe451p*RSp453e - 1./81.*m11*m66*RSe451p*RS645pe*
         RS345pe - 1./81.*m11*m66*RSe451p*RS645pe*RSe453p - 1./81.*m11*
         m66*RSe451p*RS645pe*RSp453e + 1./54.*m11*m66*pow(RSe451p,2) - 
         1./162.*m11*m66*pow(RSe451p,2)*RS645pe + 1./27.*m11*m66*
         RSp451e*RS345pe + 1./27.*m11*m66*RSp451e*RSe453p + 1./27.*m11*
         m66*RSp451e*RSp453e - 1./81.*m11*m66*RSp451e*RS645pe*RS345pe
          - 1./81.*m11*m66*RSp451e*RS645pe*RSe453p - 1./81.*m11*m66*
         RSp451e*RS645pe*RSp453e + 1./27.*m11*m66*RSp451e*RSe451p - 1./
         81.*m11*m66*RSp451e*RSe451p*RS645pe + 1./54.*m11*m66*pow(
         RSp451e,2) );

      M13p6x23R +=  + Abeps(m66,mu2)*pi16 * (  - 1./162.*m11*m66*pow(
         RSp451e,2)*RS645pe );

      M13p6x23R +=  + Abeps(me2,mu2)*pi16 * (  - 2./405.*pow(mp2,2)*
         RSp45613e*RSe45613p + 1./405.*pow(mp2,2)*RSe45633p*RSp45613e
          + 1./405.*pow(mp2,2)*RSp45633e*RSe45613p - 1./810.*pow(mp2,2)
         *RSp45633e*RSe45633p + 1./405.*pow(mp2,2)*RSe45611p*RSp45613e
          - 1./810.*pow(mp2,2)*RSe45611p*RSp45633e + 1./405.*pow(mp2,2)
         *RSp45611e*RSe45613p - 1./810.*pow(mp2,2)*RSp45611e*RSe45633p
          - 1./810.*pow(mp2,2)*RSp45611e*RSe45611p - 1./270.*me2*
         RD3456ep*RSe45613p + 1./540.*me2*RSe45633p*RD3456ep - 1./270.*
         me2*RD1456ep*RSe45613p + 1./540.*me2*RD1456ep*RSe45633p + 1./
         540.*me2*RSe45611p*RD3456ep + 1./540.*me2*RSe45611p*RD1456ep
          + 1./135.*me2*mp2*RSp45613e*RSe45613p - 1./270.*me2*mp2*
         RSe45633p*RSp45613e - 1./270.*me2*mp2*RSp45633e*RSe45613p + 1./
         540.*me2*mp2*RSp45633e*RSe45633p - 1./270.*me2*mp2*RSe45611p*
         RSp45613e + 1./540.*me2*mp2*RSe45611p*RSp45633e - 1./270.*me2*
         mp2*RSp45611e*RSe45613p + 1./540.*me2*mp2*RSp45611e*RSe45633p
          + 1./540.*me2*mp2*RSp45611e*RSe45611p );

      M13p6x23R +=  + Abeps(me2,mu2)*pi16 * ( 1./405.*pow(me2,2)*pow(
         RSe45613p,2) - 1./405.*pow(me2,2)*RSe45633p*RSe45613p + 1./
         1620.*pow(me2,2)*pow(RSe45633p,2) - 1./405.*pow(me2,2)*
         RSe45611p*RSe45613p + 1./810.*pow(me2,2)*RSe45611p*RSe45633p
          + 1./1620.*pow(me2,2)*pow(RSe45611p,2) + 1./81.*m66*mp2*
         RSe456p*RSp453e*RS345pe + 1./81.*m66*mp2*RSe456p*RSp453e*
         RSe453p + 1./81.*m66*mp2*RSe456p*pow(RSp453e,2) - 1./81.*m66*
         mp2*RSe456p*RS145pe*RSp453e - 1./81.*m66*mp2*RSe456p*RSe451p*
         RSp453e - 1./81.*m66*mp2*RSe456p*RSp451e*RS345pe - 1./81.*m66*
         mp2*RSe456p*RSp451e*RSe453p - 2./81.*m66*mp2*RSe456p*RSp451e*
         RSp453e + 1./81.*m66*mp2*RSe456p*RSp451e*RS145pe + 1./81.*m66*
         mp2*RSe456p*RSp451e*RSe451p + 1./81.*m66*mp2*RSe456p*pow(
         RSp451e,2) + 1./162.*m66*mp2*RSp456e*pow(RSe453p,2) + 1./81.*
         m66*mp2*RSp456e*RSp453e*RSe453p - 1./81.*m66*mp2*RSp456e*
         RSe451p*RSe453p - 1./81.*m66*mp2*RSp456e*RSe451p*RSp453e + 1./
         162.*m66*mp2*RSp456e*pow(RSe451p,2) );

      M13p6x23R +=  + Abeps(me2,mu2)*pi16 * (  - 1./81.*m66*mp2*RSp456e
         *RSp451e*RSe453p + 1./81.*m66*mp2*RSp456e*RSp451e*RSe451p - 4./
         135.*m66*me2*RSe45613p + 7./270.*m66*me2*RSe45633p - 1./162.*
         m66*me2*RSe563p*RS345pe - 1./81.*m66*me2*RSe563p*RS356pe - 1./
         81.*m66*me2*pow(RSe563p,2) - 1./162.*m66*me2*RSe463p*RS345pe
          - 1./81.*m66*me2*RSe463p*RS346pe - 1./81.*m66*me2*pow(
         RSe463p,2) - 1./45.*m66*me2*RSe453p - 5./81.*m66*me2*RSe453p*
         RS345pe - 1./162.*m66*me2*RSe453p*RS346pe - 1./162.*m66*me2*
         RSe453p*RS356pe - 1./81.*m66*me2*RSe453p*RSe563p - 1./81.*m66*
         me2*RSe453p*RSe463p - 5./81.*m66*me2*pow(RSe453p,2) - 1./81.*
         m66*me2*RSp563e*RSe563p - 1./162.*m66*me2*RSp563e*RSe453p - 1./
         81.*m66*me2*RSp463e*RSe463p - 1./162.*m66*me2*RSp463e*RSe453p
          - 1./162.*m66*me2*RSp453e*RSe563p - 1./162.*m66*me2*RSp453e*
         RSe463p - 5./81.*m66*me2*RSp453e*RSe453p + 7./270.*m66*me2*
         RSe45611p + 1./162.*m66*me2*RS145pe*RSe563p + 1./162.*m66*me2*
         RS145pe*RSe463p );

      M13p6x23R +=  + Abeps(me2,mu2)*pi16 * ( 5./81.*m66*me2*RS145pe*
         RSe453p + 1./81.*m66*me2*RS146pe*RSe463p + 1./162.*m66*me2*
         RS146pe*RSe453p + 1./81.*m66*me2*RS156pe*RSe563p + 1./162.*m66
         *me2*RS156pe*RSe453p + 1./162.*m66*me2*RSe561p*RS345pe + 1./81.
         *m66*me2*RSe561p*RS356pe + 2./81.*m66*me2*RSe561p*RSe563p + 1./
         81.*m66*me2*RSe561p*RSe453p + 1./81.*m66*me2*RSe561p*RSp563e
          + 1./162.*m66*me2*RSe561p*RSp453e - 1./162.*m66*me2*RSe561p*
         RS145pe - 1./81.*m66*me2*RSe561p*RS156pe - 1./81.*m66*me2*pow(
         RSe561p,2) + 1./162.*m66*me2*RSe461p*RS345pe + 1./81.*m66*me2*
         RSe461p*RS346pe + 2./81.*m66*me2*RSe461p*RSe463p + 1./81.*m66*
         me2*RSe461p*RSe453p + 1./81.*m66*me2*RSe461p*RSp463e + 1./162.
         *m66*me2*RSe461p*RSp453e - 1./162.*m66*me2*RSe461p*RS145pe - 1.
         /81.*m66*me2*RSe461p*RS146pe - 1./81.*m66*me2*pow(RSe461p,2)
          - 1./45.*m66*me2*RSe451p + 5./81.*m66*me2*RSe451p*RS345pe + 1.
         /162.*m66*me2*RSe451p*RS346pe + 1./162.*m66*me2*RSe451p*
         RS356pe );

      M13p6x23R +=  + Abeps(me2,mu2)*pi16 * ( 1./81.*m66*me2*RSe451p*
         RSe563p + 1./81.*m66*me2*RSe451p*RSe463p + 10./81.*m66*me2*
         RSe451p*RSe453p + 1./162.*m66*me2*RSe451p*RSp563e + 1./162.*
         m66*me2*RSe451p*RSp463e + 5./81.*m66*me2*RSe451p*RSp453e - 5./
         81.*m66*me2*RSe451p*RS145pe - 1./162.*m66*me2*RSe451p*RS146pe
          - 1./162.*m66*me2*RSe451p*RS156pe - 1./81.*m66*me2*RSe451p*
         RSe561p - 1./81.*m66*me2*RSe451p*RSe461p - 5./81.*m66*me2*pow(
         RSe451p,2) + 1./81.*m66*me2*RSp561e*RSe563p + 1./162.*m66*me2*
         RSp561e*RSe453p - 1./81.*m66*me2*RSp561e*RSe561p - 1./162.*m66
         *me2*RSp561e*RSe451p + 1./81.*m66*me2*RSp461e*RSe463p + 1./162.
         *m66*me2*RSp461e*RSe453p - 1./81.*m66*me2*RSp461e*RSe461p - 1./
         162.*m66*me2*RSp461e*RSe451p + 1./162.*m66*me2*RSp451e*RSe563p
          + 1./162.*m66*me2*RSp451e*RSe463p + 5./81.*m66*me2*RSp451e*
         RSe453p - 1./162.*m66*me2*RSp451e*RSe561p - 1./162.*m66*me2*
         RSp451e*RSe461p - 5./81.*m66*me2*RSp451e*RSe451p + 1./45.*m66*
         me2*RSe456p );

      M13p6x23R +=  + Abeps(me2,mu2)*pi16 * ( 1./81.*m66*me2*RSe456p*
         RSe453p*RS345pe + 1./54.*m66*me2*RSe456p*pow(RSe453p,2) + 2./
         81.*m66*me2*RSe456p*RSp453e*RSe453p - 1./81.*m66*me2*RSe456p*
         RS145pe*RSe453p - 1./81.*m66*me2*RSe456p*RSe451p*RS345pe - 1./
         27.*m66*me2*RSe456p*RSe451p*RSe453p - 2./81.*m66*me2*RSe456p*
         RSe451p*RSp453e + 1./81.*m66*me2*RSe456p*RSe451p*RS145pe + 1./
         54.*m66*me2*RSe456p*pow(RSe451p,2) - 2./81.*m66*me2*RSe456p*
         RSp451e*RSe453p + 2./81.*m66*me2*RSe456p*RSp451e*RSe451p + 1./
         90.*pow(m66,2)*RSe453p - 1./162.*pow(m66,2)*RSe453p*RSe563p - 
         1./162.*pow(m66,2)*RSe453p*RSe463p + 1./486.*pow(m66,2)*pow(
         RSe453p,2) - 1./162.*pow(m66,2)*RSp563e*RSe453p - 1./162.*pow(
         m66,2)*RSp463e*RSe453p - 1./162.*pow(m66,2)*RSp453e*RSe563p - 
         1./162.*pow(m66,2)*RSp453e*RSe463p + 1./243.*pow(m66,2)*
         RSp453e*RSe453p + 1./162.*pow(m66,2)*RS645pe*pow(RSe453p,2) + 
         1./81.*pow(m66,2)*RS645pe*RSp453e*RSe453p + 1./162.*pow(m66,2)
         *RSe561p*RSe453p );

      M13p6x23R +=  + Abeps(me2,mu2)*pi16 * ( 1./162.*pow(m66,2)*
         RSe561p*RSp453e + 1./162.*pow(m66,2)*RSe461p*RSe453p + 1./162.
         *pow(m66,2)*RSe461p*RSp453e + 1./90.*pow(m66,2)*RSe451p + 1./
         162.*pow(m66,2)*RSe451p*RSe563p + 1./162.*pow(m66,2)*RSe451p*
         RSe463p - 1./243.*pow(m66,2)*RSe451p*RSe453p + 1./162.*pow(
         m66,2)*RSe451p*RSp563e + 1./162.*pow(m66,2)*RSe451p*RSp463e - 
         1./243.*pow(m66,2)*RSe451p*RSp453e - 1./81.*pow(m66,2)*RSe451p
         *RS645pe*RSe453p - 1./81.*pow(m66,2)*RSe451p*RS645pe*RSp453e
          - 1./162.*pow(m66,2)*RSe451p*RSe561p - 1./162.*pow(m66,2)*
         RSe451p*RSe461p + 1./486.*pow(m66,2)*pow(RSe451p,2) + 1./162.*
         pow(m66,2)*pow(RSe451p,2)*RS645pe + 1./162.*pow(m66,2)*RSp561e
         *RSe453p - 1./162.*pow(m66,2)*RSp561e*RSe451p + 1./162.*pow(
         m66,2)*RSp461e*RSe453p - 1./162.*pow(m66,2)*RSp461e*RSe451p + 
         1./162.*pow(m66,2)*RSp451e*RSe563p + 1./162.*pow(m66,2)*
         RSp451e*RSe463p - 1./243.*pow(m66,2)*RSp451e*RSe453p - 1./81.*
         pow(m66,2)*RSp451e*RS645pe*RSe453p );

      M13p6x23R +=  + Abeps(me2,mu2)*pi16 * (  - 1./162.*pow(m66,2)*
         RSp451e*RSe561p - 1./162.*pow(m66,2)*RSp451e*RSe461p + 1./243.
         *pow(m66,2)*RSp451e*RSe451p + 1./81.*pow(m66,2)*RSp451e*
         RSe451p*RS645pe - 1./45.*pow(m66,2)*RSe456p + 1./81.*m55*mp2*
         RSe465p*RSp463e*RS346pe + 1./81.*m55*mp2*RSe465p*RSp463e*
         RSe463p + 1./81.*m55*mp2*RSe465p*pow(RSp463e,2) - 1./81.*m55*
         mp2*RSe465p*RS146pe*RSp463e - 1./81.*m55*mp2*RSe465p*RSe461p*
         RSp463e - 1./81.*m55*mp2*RSe465p*RSp461e*RS346pe - 1./81.*m55*
         mp2*RSe465p*RSp461e*RSe463p - 2./81.*m55*mp2*RSe465p*RSp461e*
         RSp463e + 1./81.*m55*mp2*RSe465p*RSp461e*RS146pe + 1./81.*m55*
         mp2*RSe465p*RSp461e*RSe461p + 1./81.*m55*mp2*RSe465p*pow(
         RSp461e,2) + 1./162.*m55*mp2*RSp465e*pow(RSe463p,2) + 1./81.*
         m55*mp2*RSp465e*RSp463e*RSe463p - 1./81.*m55*mp2*RSp465e*
         RSe461p*RSe463p - 1./81.*m55*mp2*RSp465e*RSe461p*RSp463e + 1./
         162.*m55*mp2*RSp465e*pow(RSe461p,2) - 1./81.*m55*mp2*RSp465e*
         RSp461e*RSe463p );

      M13p6x23R +=  + Abeps(me2,mu2)*pi16 * ( 1./81.*m55*mp2*RSp465e*
         RSp461e*RSe461p - 4./135.*m55*me2*RSe45613p + 7./270.*m55*me2*
         RSe45633p - 1./162.*m55*me2*RSe563p*RS346pe - 1./81.*m55*me2*
         RSe563p*RS356pe - 1./81.*m55*me2*pow(RSe563p,2) - 1./45.*m55*
         me2*RSe463p - 1./162.*m55*me2*RSe463p*RS345pe - 5./81.*m55*me2
         *RSe463p*RS346pe - 1./162.*m55*me2*RSe463p*RS356pe - 1./81.*
         m55*me2*RSe463p*RSe563p - 5./81.*m55*me2*pow(RSe463p,2) - 1./
         81.*m55*me2*RSe453p*RS345pe - 1./162.*m55*me2*RSe453p*RS346pe
          - 1./81.*m55*me2*RSe453p*RSe463p - 1./81.*m55*me2*pow(
         RSe453p,2) - 1./81.*m55*me2*RSp563e*RSe563p - 1./162.*m55*me2*
         RSp563e*RSe463p - 1./162.*m55*me2*RSp463e*RSe563p - 5./81.*m55
         *me2*RSp463e*RSe463p - 1./162.*m55*me2*RSp463e*RSe453p - 1./
         162.*m55*me2*RSp453e*RSe463p - 1./81.*m55*me2*RSp453e*RSe453p
          + 7./270.*m55*me2*RSe45611p + 1./162.*m55*me2*RS145pe*RSe463p
          + 1./81.*m55*me2*RS145pe*RSe453p + 1./162.*m55*me2*RS146pe*
         RSe563p );

      M13p6x23R +=  + Abeps(me2,mu2)*pi16 * ( 5./81.*m55*me2*RS146pe*
         RSe463p + 1./162.*m55*me2*RS146pe*RSe453p + 1./81.*m55*me2*
         RS156pe*RSe563p + 1./162.*m55*me2*RS156pe*RSe463p + 1./162.*
         m55*me2*RSe561p*RS346pe + 1./81.*m55*me2*RSe561p*RS356pe + 2./
         81.*m55*me2*RSe561p*RSe563p + 1./81.*m55*me2*RSe561p*RSe463p
          + 1./81.*m55*me2*RSe561p*RSp563e + 1./162.*m55*me2*RSe561p*
         RSp463e - 1./162.*m55*me2*RSe561p*RS146pe - 1./81.*m55*me2*
         RSe561p*RS156pe - 1./81.*m55*me2*pow(RSe561p,2) - 1./45.*m55*
         me2*RSe461p + 1./162.*m55*me2*RSe461p*RS345pe + 5./81.*m55*me2
         *RSe461p*RS346pe + 1./162.*m55*me2*RSe461p*RS356pe + 1./81.*
         m55*me2*RSe461p*RSe563p + 10./81.*m55*me2*RSe461p*RSe463p + 1./
         81.*m55*me2*RSe461p*RSe453p + 1./162.*m55*me2*RSe461p*RSp563e
          + 5./81.*m55*me2*RSe461p*RSp463e + 1./162.*m55*me2*RSe461p*
         RSp453e - 1./162.*m55*me2*RSe461p*RS145pe - 5./81.*m55*me2*
         RSe461p*RS146pe - 1./162.*m55*me2*RSe461p*RS156pe - 1./81.*m55
         *me2*RSe461p*RSe561p );

      M13p6x23R +=  + Abeps(me2,mu2)*pi16 * (  - 5./81.*m55*me2*pow(
         RSe461p,2) + 1./81.*m55*me2*RSe451p*RS345pe + 1./162.*m55*me2*
         RSe451p*RS346pe + 1./81.*m55*me2*RSe451p*RSe463p + 2./81.*m55*
         me2*RSe451p*RSe453p + 1./162.*m55*me2*RSe451p*RSp463e + 1./81.
         *m55*me2*RSe451p*RSp453e - 1./81.*m55*me2*RSe451p*RS145pe - 1./
         162.*m55*me2*RSe451p*RS146pe - 1./81.*m55*me2*RSe451p*RSe461p
          - 1./81.*m55*me2*pow(RSe451p,2) + 1./81.*m55*me2*RSp561e*
         RSe563p + 1./162.*m55*me2*RSp561e*RSe463p - 1./81.*m55*me2*
         RSp561e*RSe561p - 1./162.*m55*me2*RSp561e*RSe461p + 1./162.*
         m55*me2*RSp461e*RSe563p + 5./81.*m55*me2*RSp461e*RSe463p + 1./
         162.*m55*me2*RSp461e*RSe453p - 1./162.*m55*me2*RSp461e*RSe561p
          - 5./81.*m55*me2*RSp461e*RSe461p - 1./162.*m55*me2*RSp461e*
         RSe451p + 1./162.*m55*me2*RSp451e*RSe463p + 1./81.*m55*me2*
         RSp451e*RSe453p - 1./162.*m55*me2*RSp451e*RSe461p - 1./81.*m55
         *me2*RSp451e*RSe451p + 1./45.*m55*me2*RSe465p + 1./81.*m55*me2
         *RSe465p*RSe463p*RS346pe );

      M13p6x23R +=  + Abeps(me2,mu2)*pi16 * ( 1./54.*m55*me2*RSe465p*
         pow(RSe463p,2) + 2./81.*m55*me2*RSe465p*RSp463e*RSe463p - 1./
         81.*m55*me2*RSe465p*RS146pe*RSe463p - 1./81.*m55*me2*RSe465p*
         RSe461p*RS346pe - 1./27.*m55*me2*RSe465p*RSe461p*RSe463p - 2./
         81.*m55*me2*RSe465p*RSe461p*RSp463e + 1./81.*m55*me2*RSe465p*
         RSe461p*RS146pe + 1./54.*m55*me2*RSe465p*pow(RSe461p,2) - 2./
         81.*m55*me2*RSe465p*RSp461e*RSe463p + 2./81.*m55*me2*RSe465p*
         RSp461e*RSe461p + 2./243.*m55*m66*pow(RSe463p,2) - 1./81.*m55*
         m66*RSe453p*RSe463p + 2./243.*m55*m66*pow(RSe453p,2) + 4./243.
         *m55*m66*RSp463e*RSe463p - 1./81.*m55*m66*RSp463e*RSe453p - 1./
         81.*m55*m66*RSp453e*RSe463p + 4./243.*m55*m66*RSp453e*RSe453p
          - 4./243.*m55*m66*RSe461p*RSe463p + 1./81.*m55*m66*RSe461p*
         RSe453p - 4./243.*m55*m66*RSe461p*RSp463e + 1./81.*m55*m66*
         RSe461p*RSp453e + 2./243.*m55*m66*pow(RSe461p,2) + 1./81.*m55*
         m66*RSe451p*RSe463p - 4./243.*m55*m66*RSe451p*RSe453p + 1./81.
         *m55*m66*RSe451p*RSp463e );

      M13p6x23R +=  + Abeps(me2,mu2)*pi16 * (  - 4./243.*m55*m66*
         RSe451p*RSp453e - 1./81.*m55*m66*RSe451p*RSe461p + 2./243.*m55
         *m66*pow(RSe451p,2) - 4./243.*m55*m66*RSp461e*RSe463p + 1./81.
         *m55*m66*RSp461e*RSe453p + 4./243.*m55*m66*RSp461e*RSe461p - 1.
         /81.*m55*m66*RSp461e*RSe451p + 1./81.*m55*m66*RSp451e*RSe463p
          - 4./243.*m55*m66*RSp451e*RSe453p - 1./81.*m55*m66*RSp451e*
         RSe461p + 4./243.*m55*m66*RSp451e*RSe451p + 1./90.*pow(m55,2)*
         RSe463p - 1./162.*pow(m55,2)*RSe463p*RSe563p + 1./486.*pow(
         m55,2)*pow(RSe463p,2) - 1./162.*pow(m55,2)*RSe453p*RSe463p - 1.
         /162.*pow(m55,2)*RSp563e*RSe463p - 1./162.*pow(m55,2)*RSp463e*
         RSe563p + 1./243.*pow(m55,2)*RSp463e*RSe463p - 1./162.*pow(
         m55,2)*RSp463e*RSe453p - 1./162.*pow(m55,2)*RSp453e*RSe463p + 
         1./162.*pow(m55,2)*RS546pe*pow(RSe463p,2) + 1./81.*pow(m55,2)*
         RS546pe*RSp463e*RSe463p + 1./162.*pow(m55,2)*RSe561p*RSe463p
          + 1./162.*pow(m55,2)*RSe561p*RSp463e + 1./90.*pow(m55,2)*
         RSe461p );

      M13p6x23R +=  + Abeps(me2,mu2)*pi16 * ( 1./162.*pow(m55,2)*
         RSe461p*RSe563p - 1./243.*pow(m55,2)*RSe461p*RSe463p + 1./162.
         *pow(m55,2)*RSe461p*RSe453p + 1./162.*pow(m55,2)*RSe461p*
         RSp563e - 1./243.*pow(m55,2)*RSe461p*RSp463e + 1./162.*pow(
         m55,2)*RSe461p*RSp453e - 1./81.*pow(m55,2)*RSe461p*RS546pe*
         RSe463p - 1./81.*pow(m55,2)*RSe461p*RS546pe*RSp463e - 1./162.*
         pow(m55,2)*RSe461p*RSe561p + 1./486.*pow(m55,2)*pow(RSe461p,2)
          + 1./162.*pow(m55,2)*pow(RSe461p,2)*RS546pe + 1./162.*pow(
         m55,2)*RSe451p*RSe463p + 1./162.*pow(m55,2)*RSe451p*RSp463e - 
         1./162.*pow(m55,2)*RSe451p*RSe461p + 1./162.*pow(m55,2)*
         RSp561e*RSe463p - 1./162.*pow(m55,2)*RSp561e*RSe461p + 1./162.
         *pow(m55,2)*RSp461e*RSe563p - 1./243.*pow(m55,2)*RSp461e*
         RSe463p + 1./162.*pow(m55,2)*RSp461e*RSe453p - 1./81.*pow(
         m55,2)*RSp461e*RS546pe*RSe463p - 1./162.*pow(m55,2)*RSp461e*
         RSe561p + 1./243.*pow(m55,2)*RSp461e*RSe461p + 1./81.*pow(
         m55,2)*RSp461e*RSe461p*RS546pe );

      M13p6x23R +=  + Abeps(me2,mu2)*pi16 * (  - 1./162.*pow(m55,2)*
         RSp461e*RSe451p + 1./162.*pow(m55,2)*RSp451e*RSe463p - 1./162.
         *pow(m55,2)*RSp451e*RSe461p - 1./45.*pow(m55,2)*RSe465p + 1./
         81.*m44*mp2*RSe564p*RSp563e*RS356pe + 1./81.*m44*mp2*RSe564p*
         RSp563e*RSe563p + 1./81.*m44*mp2*RSe564p*pow(RSp563e,2) - 1./
         81.*m44*mp2*RSe564p*RS156pe*RSp563e - 1./81.*m44*mp2*RSe564p*
         RSe561p*RSp563e - 1./81.*m44*mp2*RSe564p*RSp561e*RS356pe - 1./
         81.*m44*mp2*RSe564p*RSp561e*RSe563p - 2./81.*m44*mp2*RSe564p*
         RSp561e*RSp563e + 1./81.*m44*mp2*RSe564p*RSp561e*RS156pe + 1./
         81.*m44*mp2*RSe564p*RSp561e*RSe561p + 1./81.*m44*mp2*RSe564p*
         pow(RSp561e,2) + 1./162.*m44*mp2*RSp564e*pow(RSe563p,2) + 1./
         81.*m44*mp2*RSp564e*RSp563e*RSe563p - 1./81.*m44*mp2*RSp564e*
         RSe561p*RSe563p - 1./81.*m44*mp2*RSp564e*RSe561p*RSp563e + 1./
         162.*m44*mp2*RSp564e*pow(RSe561p,2) - 1./81.*m44*mp2*RSp564e*
         RSp561e*RSe563p + 1./81.*m44*mp2*RSp564e*RSp561e*RSe561p - 4./
         135.*m44*me2*RSe45613p );

      M13p6x23R +=  + Abeps(me2,mu2)*pi16 * ( 7./270.*m44*me2*RSe45633p
          - 1./45.*m44*me2*RSe563p - 1./162.*m44*me2*RSe563p*RS345pe - 
         1./162.*m44*me2*RSe563p*RS346pe - 5./81.*m44*me2*RSe563p*
         RS356pe - 5./81.*m44*me2*pow(RSe563p,2) - 1./81.*m44*me2*
         RSe463p*RS346pe - 1./162.*m44*me2*RSe463p*RS356pe - 1./81.*m44
         *me2*RSe463p*RSe563p - 1./81.*m44*me2*pow(RSe463p,2) - 1./81.*
         m44*me2*RSe453p*RS345pe - 1./162.*m44*me2*RSe453p*RS356pe - 1./
         81.*m44*me2*RSe453p*RSe563p - 1./81.*m44*me2*pow(RSe453p,2) - 
         5./81.*m44*me2*RSp563e*RSe563p - 1./162.*m44*me2*RSp563e*
         RSe463p - 1./162.*m44*me2*RSp563e*RSe453p - 1./162.*m44*me2*
         RSp463e*RSe563p - 1./81.*m44*me2*RSp463e*RSe463p - 1./162.*m44
         *me2*RSp453e*RSe563p - 1./81.*m44*me2*RSp453e*RSe453p + 7./270.
         *m44*me2*RSe45611p + 1./162.*m44*me2*RS145pe*RSe563p + 1./81.*
         m44*me2*RS145pe*RSe453p + 1./162.*m44*me2*RS146pe*RSe563p + 1./
         81.*m44*me2*RS146pe*RSe463p + 5./81.*m44*me2*RS156pe*RSe563p
          + 1./162.*m44*me2*RS156pe*RSe463p );

      M13p6x23R +=  + Abeps(me2,mu2)*pi16 * ( 1./162.*m44*me2*RS156pe*
         RSe453p - 1./45.*m44*me2*RSe561p + 1./162.*m44*me2*RSe561p*
         RS345pe + 1./162.*m44*me2*RSe561p*RS346pe + 5./81.*m44*me2*
         RSe561p*RS356pe + 10./81.*m44*me2*RSe561p*RSe563p + 1./81.*m44
         *me2*RSe561p*RSe463p + 1./81.*m44*me2*RSe561p*RSe453p + 5./81.
         *m44*me2*RSe561p*RSp563e + 1./162.*m44*me2*RSe561p*RSp463e + 1.
         /162.*m44*me2*RSe561p*RSp453e - 1./162.*m44*me2*RSe561p*
         RS145pe - 1./162.*m44*me2*RSe561p*RS146pe - 5./81.*m44*me2*
         RSe561p*RS156pe - 5./81.*m44*me2*pow(RSe561p,2) + 1./81.*m44*
         me2*RSe461p*RS346pe + 1./162.*m44*me2*RSe461p*RS356pe + 1./81.
         *m44*me2*RSe461p*RSe563p + 2./81.*m44*me2*RSe461p*RSe463p + 1./
         162.*m44*me2*RSe461p*RSp563e + 1./81.*m44*me2*RSe461p*RSp463e
          - 1./81.*m44*me2*RSe461p*RS146pe - 1./162.*m44*me2*RSe461p*
         RS156pe - 1./81.*m44*me2*RSe461p*RSe561p - 1./81.*m44*me2*pow(
         RSe461p,2) + 1./81.*m44*me2*RSe451p*RS345pe + 1./162.*m44*me2*
         RSe451p*RS356pe );

      M13p6x23R +=  + Abeps(me2,mu2)*pi16 * ( 1./81.*m44*me2*RSe451p*
         RSe563p + 2./81.*m44*me2*RSe451p*RSe453p + 1./162.*m44*me2*
         RSe451p*RSp563e + 1./81.*m44*me2*RSe451p*RSp453e - 1./81.*m44*
         me2*RSe451p*RS145pe - 1./162.*m44*me2*RSe451p*RS156pe - 1./81.
         *m44*me2*RSe451p*RSe561p - 1./81.*m44*me2*pow(RSe451p,2) + 5./
         81.*m44*me2*RSp561e*RSe563p + 1./162.*m44*me2*RSp561e*RSe463p
          + 1./162.*m44*me2*RSp561e*RSe453p - 5./81.*m44*me2*RSp561e*
         RSe561p - 1./162.*m44*me2*RSp561e*RSe461p - 1./162.*m44*me2*
         RSp561e*RSe451p + 1./162.*m44*me2*RSp461e*RSe563p + 1./81.*m44
         *me2*RSp461e*RSe463p - 1./162.*m44*me2*RSp461e*RSe561p - 1./81.
         *m44*me2*RSp461e*RSe461p + 1./162.*m44*me2*RSp451e*RSe563p + 1.
         /81.*m44*me2*RSp451e*RSe453p - 1./162.*m44*me2*RSp451e*RSe561p
          - 1./81.*m44*me2*RSp451e*RSe451p + 1./45.*m44*me2*RSe564p + 1.
         /81.*m44*me2*RSe564p*RSe563p*RS356pe + 1./54.*m44*me2*RSe564p*
         pow(RSe563p,2) + 2./81.*m44*me2*RSe564p*RSp563e*RSe563p - 1./
         81.*m44*me2*RSe564p*RS156pe*RSe563p );

      M13p6x23R +=  + Abeps(me2,mu2)*pi16 * (  - 1./81.*m44*me2*RSe564p
         *RSe561p*RS356pe - 1./27.*m44*me2*RSe564p*RSe561p*RSe563p - 2./
         81.*m44*me2*RSe564p*RSe561p*RSp563e + 1./81.*m44*me2*RSe564p*
         RSe561p*RS156pe + 1./54.*m44*me2*RSe564p*pow(RSe561p,2) - 2./
         81.*m44*me2*RSe564p*RSp561e*RSe563p + 2./81.*m44*me2*RSe564p*
         RSp561e*RSe561p + 2./243.*m44*m66*pow(RSe563p,2) - 1./81.*m44*
         m66*RSe453p*RSe563p + 2./243.*m44*m66*pow(RSe453p,2) + 4./243.
         *m44*m66*RSp563e*RSe563p - 1./81.*m44*m66*RSp563e*RSe453p - 1./
         81.*m44*m66*RSp453e*RSe563p + 4./243.*m44*m66*RSp453e*RSe453p
          - 4./243.*m44*m66*RSe561p*RSe563p + 1./81.*m44*m66*RSe561p*
         RSe453p - 4./243.*m44*m66*RSe561p*RSp563e + 1./81.*m44*m66*
         RSe561p*RSp453e + 2./243.*m44*m66*pow(RSe561p,2) + 1./81.*m44*
         m66*RSe451p*RSe563p - 4./243.*m44*m66*RSe451p*RSe453p + 1./81.
         *m44*m66*RSe451p*RSp563e - 4./243.*m44*m66*RSe451p*RSp453e - 1.
         /81.*m44*m66*RSe451p*RSe561p + 2./243.*m44*m66*pow(RSe451p,2)
          - 4./243.*m44*m66*RSp561e*RSe563p );

      M13p6x23R +=  + Abeps(me2,mu2)*pi16 * ( 1./81.*m44*m66*RSp561e*
         RSe453p + 4./243.*m44*m66*RSp561e*RSe561p - 1./81.*m44*m66*
         RSp561e*RSe451p + 1./81.*m44*m66*RSp451e*RSe563p - 4./243.*m44
         *m66*RSp451e*RSe453p - 1./81.*m44*m66*RSp451e*RSe561p + 4./243.
         *m44*m66*RSp451e*RSe451p + 2./243.*m44*m55*pow(RSe563p,2) - 1./
         81.*m44*m55*RSe463p*RSe563p + 2./243.*m44*m55*pow(RSe463p,2)
          + 4./243.*m44*m55*RSp563e*RSe563p - 1./81.*m44*m55*RSp563e*
         RSe463p - 1./81.*m44*m55*RSp463e*RSe563p + 4./243.*m44*m55*
         RSp463e*RSe463p - 4./243.*m44*m55*RSe561p*RSe563p + 1./81.*m44
         *m55*RSe561p*RSe463p - 4./243.*m44*m55*RSe561p*RSp563e + 1./81.
         *m44*m55*RSe561p*RSp463e + 2./243.*m44*m55*pow(RSe561p,2) + 1./
         81.*m44*m55*RSe461p*RSe563p - 4./243.*m44*m55*RSe461p*RSe463p
          + 1./81.*m44*m55*RSe461p*RSp563e - 4./243.*m44*m55*RSe461p*
         RSp463e - 1./81.*m44*m55*RSe461p*RSe561p + 2./243.*m44*m55*
         pow(RSe461p,2) - 4./243.*m44*m55*RSp561e*RSe563p + 1./81.*m44*
         m55*RSp561e*RSe463p );

      M13p6x23R +=  + Abeps(me2,mu2)*pi16 * ( 4./243.*m44*m55*RSp561e*
         RSe561p - 1./81.*m44*m55*RSp561e*RSe461p + 1./81.*m44*m55*
         RSp461e*RSe563p - 4./243.*m44*m55*RSp461e*RSe463p - 1./81.*m44
         *m55*RSp461e*RSe561p + 4./243.*m44*m55*RSp461e*RSe461p + 1./90.
         *pow(m44,2)*RSe563p + 1./486.*pow(m44,2)*pow(RSe563p,2) - 1./
         162.*pow(m44,2)*RSe463p*RSe563p - 1./162.*pow(m44,2)*RSe453p*
         RSe563p + 1./243.*pow(m44,2)*RSp563e*RSe563p - 1./162.*pow(
         m44,2)*RSp563e*RSe463p - 1./162.*pow(m44,2)*RSp563e*RSe453p - 
         1./162.*pow(m44,2)*RSp463e*RSe563p - 1./162.*pow(m44,2)*
         RSp453e*RSe563p + 1./162.*pow(m44,2)*RS456pe*pow(RSe563p,2) + 
         1./81.*pow(m44,2)*RS456pe*RSp563e*RSe563p + 1./90.*pow(m44,2)*
         RSe561p - 1./243.*pow(m44,2)*RSe561p*RSe563p + 1./162.*pow(
         m44,2)*RSe561p*RSe463p + 1./162.*pow(m44,2)*RSe561p*RSe453p - 
         1./243.*pow(m44,2)*RSe561p*RSp563e + 1./162.*pow(m44,2)*
         RSe561p*RSp463e + 1./162.*pow(m44,2)*RSe561p*RSp453e - 1./81.*
         pow(m44,2)*RSe561p*RS456pe*RSe563p );

      M13p6x23R +=  + Abeps(me2,mu2)*pi16 * (  - 1./81.*pow(m44,2)*
         RSe561p*RS456pe*RSp563e + 1./486.*pow(m44,2)*pow(RSe561p,2) + 
         1./162.*pow(m44,2)*pow(RSe561p,2)*RS456pe + 1./162.*pow(m44,2)
         *RSe461p*RSe563p + 1./162.*pow(m44,2)*RSe461p*RSp563e - 1./162.
         *pow(m44,2)*RSe461p*RSe561p + 1./162.*pow(m44,2)*RSe451p*
         RSe563p + 1./162.*pow(m44,2)*RSe451p*RSp563e - 1./162.*pow(
         m44,2)*RSe451p*RSe561p - 1./243.*pow(m44,2)*RSp561e*RSe563p + 
         1./162.*pow(m44,2)*RSp561e*RSe463p + 1./162.*pow(m44,2)*
         RSp561e*RSe453p - 1./81.*pow(m44,2)*RSp561e*RS456pe*RSe563p + 
         1./243.*pow(m44,2)*RSp561e*RSe561p + 1./81.*pow(m44,2)*RSp561e
         *RSe561p*RS456pe - 1./162.*pow(m44,2)*RSp561e*RSe461p - 1./162.
         *pow(m44,2)*RSp561e*RSe451p + 1./162.*pow(m44,2)*RSp461e*
         RSe563p - 1./162.*pow(m44,2)*RSp461e*RSe561p + 1./162.*pow(
         m44,2)*RSp451e*RSe563p - 1./162.*pow(m44,2)*RSp451e*RSe561p - 
         1./45.*pow(m44,2)*RSe564p - 7./405.*m33*RD3456ep*RSe45613p + 1.
         /81.*m33*RSe45633p*RD3456ep );

      M13p6x23R +=  + Abeps(me2,mu2)*pi16 * (  - 1./810.*m33*RD1456ep*
         RSe45613p - 23./1620.*m33*RD1456ep*RSe45633p - 11./810.*m33*
         RSe45611p*RD3456ep + 5./324.*m33*RSe45611p*RD1456ep + 17./405.
         *m33*mp2*RSp45613e*RSe45613p + 11./405.*m33*mp2*RSe45633p*
         RSp45613e - 1./45.*m33*mp2*RSp45633e*RSe45613p + 2./135.*m33*
         mp2*RSp45633e*RSe45633p + 2./405.*m33*mp2*RSe45611p*RSp45613e
          - 1./90.*m33*mp2*RSe45611p*RSp45633e - 1./810.*m33*mp2*
         RSp45611e*RSe45613p - 23./1620.*m33*mp2*RSp45611e*RSe45633p + 
         5./324.*m33*mp2*RSp45611e*RSe45611p + 29./360.*m33*me2*
         RSe45613p + 7./270.*m33*me2*RSe45613p*RS34561pe + 1./54.*m33*
         me2*RSe45613p*RS14563pe + 49./810.*m33*me2*pow(RSe45613p,2) + 
         1./54.*m33*me2*RSp45613e*RSe45613p + 7./540.*m33*me2*RS3456pee
         *RSe45613p + 7./540.*m33*me2*RS3456ppe*RSe45613p + 2./45.*m33*
         me2*RSe45633p - 7./540.*m33*me2*RSe45633p*RS34561pe - 1./108.*
         m33*me2*RSe45633p*RS14563pe - 11./810.*m33*me2*RSe45633p*
         RSe45613p );

      M13p6x23R +=  + Abeps(me2,mu2)*pi16 * (  - 1./108.*m33*me2*
         RSe45633p*RSp45613e - 7./1080.*m33*me2*RSe45633p*RS3456pee - 7.
         /1080.*m33*me2*RSe45633p*RS3456ppe + 7./360.*m33*me2*pow(
         RSe45633p,2) - 1./108.*m33*me2*RSp45633e*RSe45613p + 1./216.*
         m33*me2*RSp45633e*RSe45633p - 7./540.*m33*me2*RS345pe*
         RSe45613p + 7./1080.*m33*me2*RS345pe*RSe45633p - 7./540.*m33*
         me2*RS346pe*RSe45613p + 7./1080.*m33*me2*RS346pe*RSe45633p - 7.
         /540.*m33*me2*RS356pe*RSe45613p + 7./1080.*m33*me2*RS356pe*
         RSe45633p + 7./540.*m33*me2*RSe563p + 7./540.*m33*me2*RSe463p
          + 7./540.*m33*me2*RSe453p + 1./108.*m33*me2*RS1456pee*
         RSe45613p - 1./216.*m33*me2*RS1456pee*RSe45633p + 1./108.*m33*
         me2*RS1456ppe*RSe45613p - 1./216.*m33*me2*RS1456ppe*RSe45633p
          + 1./30.*m33*me2*RSe45611p - 7./540.*m33*me2*RSe45611p*
         RS34561pe - 1./108.*m33*me2*RSe45611p*RS14563pe - 2./135.*m33*
         me2*RSe45611p*RSe45613p - 1./108.*m33*me2*RSe45611p*RSp45613e
          - 7./1080.*m33*me2*RSe45611p*RS3456pee );

      M13p6x23R +=  + Abeps(me2,mu2)*pi16 * (  - 7./1080.*m33*me2*
         RSe45611p*RS3456ppe - 13./810.*m33*me2*RSe45611p*RSe45633p + 1.
         /216.*m33*me2*RSe45611p*RSp45633e + 7./1080.*m33*me2*RSe45611p
         *RS345pe + 7./1080.*m33*me2*RSe45611p*RS346pe + 7./1080.*m33*
         me2*RSe45611p*RS356pe - 1./216.*m33*me2*RSe45611p*RS1456pee - 
         1./216.*m33*me2*RSe45611p*RS1456ppe + 13./648.*m33*me2*pow(
         RSe45611p,2) - 1./108.*m33*me2*RSp45611e*RSe45613p + 1./216.*
         m33*me2*RSp45611e*RSe45633p + 1./216.*m33*me2*RSp45611e*
         RSe45611p - 1./108.*m33*me2*RS145pe*RSe45613p + 1./216.*m33*
         me2*RS145pe*RSe45633p + 1./216.*m33*me2*RS145pe*RSe45611p - 1./
         108.*m33*me2*RS146pe*RSe45613p + 1./216.*m33*me2*RS146pe*
         RSe45633p + 1./216.*m33*me2*RS146pe*RSe45611p - 1./108.*m33*
         me2*RS156pe*RSe45613p + 1./216.*m33*me2*RS156pe*RSe45633p + 1./
         216.*m33*me2*RS156pe*RSe45611p - 19./540.*m33*me2*RSe561p - 19.
         /540.*m33*me2*RSe461p - 19./540.*m33*me2*RSe451p + 19./360.*
         m33*me2*RSe564p );

      M13p6x23R +=  + Abeps(me2,mu2)*pi16 * ( 19./360.*m33*me2*RSe465p
          + 19./360.*m33*me2*RSe456p - 7./3240.*m33*m66*RSe45613p + 89./
         2160.*m33*m66*RSe45633p + 1./162.*m33*m66*pow(RSe563p,2) + 1./
         162.*m33*m66*pow(RSe463p,2) - 5./108.*m33*m66*RSe453p + 1./162.
         *m33*m66*RSe453p*RSe563p + 1./162.*m33*m66*RSe453p*RSe463p + 5.
         /162.*m33*m66*pow(RSe453p,2) + 1./81.*m33*m66*RSp563e*RSe563p
          + 1./162.*m33*m66*RSp563e*RSe453p + 1./81.*m33*m66*RSp463e*
         RSe463p + 1./162.*m33*m66*RSp463e*RSe453p + 1./162.*m33*m66*
         RSp453e*RSe563p + 1./162.*m33*m66*RSp453e*RSe463p + 5./81.*m33
         *m66*RSp453e*RSe453p + 31./1296.*m33*m66*RSe45611p + 1./81.*
         m33*m66*RSe561p*RSe563p + 1./162.*m33*m66*RSe561p*RSe453p + 1./
         81.*m33*m66*RSe561p*RSp563e + 1./162.*m33*m66*RSe561p*RSp453e
          + 1./81.*m33*m66*RSe461p*RSe463p + 1./162.*m33*m66*RSe461p*
         RSe453p + 1./81.*m33*m66*RSe461p*RSp463e + 1./162.*m33*m66*
         RSe461p*RSp453e + 7./540.*m33*m66*RSe451p + 1./162.*m33*m66*
         RSe451p*RSe563p );

      M13p6x23R +=  + Abeps(me2,mu2)*pi16 * ( 1./162.*m33*m66*RSe451p*
         RSe463p + 5./81.*m33*m66*RSe451p*RSe453p + 1./162.*m33*m66*
         RSe451p*RSp563e + 1./162.*m33*m66*RSe451p*RSp463e + 5./81.*m33
         *m66*RSe451p*RSp453e + 1./81.*m33*m66*RSp561e*RSe563p + 1./162.
         *m33*m66*RSp561e*RSe453p + 1./81.*m33*m66*RSp461e*RSe463p + 1./
         162.*m33*m66*RSp461e*RSe453p + 1./162.*m33*m66*RSp451e*RSe563p
          + 1./162.*m33*m66*RSp451e*RSe463p + 5./81.*m33*m66*RSp451e*
         RSe453p + 1./135.*m33*m66*RSe456p + 1./162.*m33*m66*RSe456p*
         pow(RS345pe,2) - 1./162.*m33*m66*RSe456p*pow(RSe453p,2) - 1./
         81.*m33*m66*RSe456p*RSp453e*RSe453p - 1./162.*m33*m66*RSe456p*
         pow(RSp453e,2) - 2./81.*m33*m66*RSe456p*RS145pe*RS345pe - 1./
         81.*m33*m66*RSe456p*RS145pe*RSe453p - 1./81.*m33*m66*RSe456p*
         RS145pe*RSp453e - 2./81.*m33*m66*RSe456p*RSe451p*RS345pe - 1./
         81.*m33*m66*RSe456p*RSe451p*RSe453p - 1./81.*m33*m66*RSe456p*
         RSe451p*RSp453e - 2./81.*m33*m66*RSe456p*RSp451e*RS345pe - 1./
         81.*m33*m66*RSe456p*RSp451e*RSe453p );

      M13p6x23R +=  + Abeps(me2,mu2)*pi16 * (  - 1./81.*m33*m66*RSe456p
         *RSp451e*RSp453e - 7./3240.*m33*m55*RSe45613p + 89./2160.*m33*
         m55*RSe45633p + 1./162.*m33*m55*pow(RSe563p,2) - 5./108.*m33*
         m55*RSe463p + 1./162.*m33*m55*RSe463p*RSe563p + 5./162.*m33*
         m55*pow(RSe463p,2) + 1./162.*m33*m55*RSe453p*RSe463p + 1./162.
         *m33*m55*pow(RSe453p,2) + 1./81.*m33*m55*RSp563e*RSe563p + 1./
         162.*m33*m55*RSp563e*RSe463p + 1./162.*m33*m55*RSp463e*RSe563p
          + 5./81.*m33*m55*RSp463e*RSe463p + 1./162.*m33*m55*RSp463e*
         RSe453p + 1./162.*m33*m55*RSp453e*RSe463p + 1./81.*m33*m55*
         RSp453e*RSe453p + 31./1296.*m33*m55*RSe45611p + 1./81.*m33*m55
         *RSe561p*RSe563p + 1./162.*m33*m55*RSe561p*RSe463p + 1./81.*
         m33*m55*RSe561p*RSp563e + 1./162.*m33*m55*RSe561p*RSp463e + 7./
         540.*m33*m55*RSe461p + 1./162.*m33*m55*RSe461p*RSe563p + 5./81.
         *m33*m55*RSe461p*RSe463p + 1./162.*m33*m55*RSe461p*RSe453p + 1.
         /162.*m33*m55*RSe461p*RSp563e + 5./81.*m33*m55*RSe461p*RSp463e
          + 1./162.*m33*m55*RSe461p*RSp453e );

      M13p6x23R +=  + Abeps(me2,mu2)*pi16 * ( 1./162.*m33*m55*RSe451p*
         RSe463p + 1./81.*m33*m55*RSe451p*RSe453p + 1./162.*m33*m55*
         RSe451p*RSp463e + 1./81.*m33*m55*RSe451p*RSp453e + 1./81.*m33*
         m55*RSp561e*RSe563p + 1./162.*m33*m55*RSp561e*RSe463p + 1./162.
         *m33*m55*RSp461e*RSe563p + 5./81.*m33*m55*RSp461e*RSe463p + 1./
         162.*m33*m55*RSp461e*RSe453p + 1./162.*m33*m55*RSp451e*RSe463p
          + 1./81.*m33*m55*RSp451e*RSe453p + 1./135.*m33*m55*RSe465p + 
         1./162.*m33*m55*RSe465p*pow(RS346pe,2) - 1./162.*m33*m55*
         RSe465p*pow(RSe463p,2) - 1./81.*m33*m55*RSe465p*RSp463e*
         RSe463p - 1./162.*m33*m55*RSe465p*pow(RSp463e,2) - 2./81.*m33*
         m55*RSe465p*RS146pe*RS346pe - 1./81.*m33*m55*RSe465p*RS146pe*
         RSe463p - 1./81.*m33*m55*RSe465p*RS146pe*RSp463e - 2./81.*m33*
         m55*RSe465p*RSe461p*RS346pe - 1./81.*m33*m55*RSe465p*RSe461p*
         RSe463p - 1./81.*m33*m55*RSe465p*RSe461p*RSp463e - 2./81.*m33*
         m55*RSe465p*RSp461e*RS346pe - 1./81.*m33*m55*RSe465p*RSp461e*
         RSe463p );

      M13p6x23R +=  + Abeps(me2,mu2)*pi16 * (  - 1./81.*m33*m55*RSe465p
         *RSp461e*RSp463e - 7./3240.*m33*m44*RSe45613p + 89./2160.*m33*
         m44*RSe45633p - 5./108.*m33*m44*RSe563p + 5./162.*m33*m44*pow(
         RSe563p,2) + 1./162.*m33*m44*RSe463p*RSe563p + 1./162.*m33*m44
         *pow(RSe463p,2) + 1./162.*m33*m44*RSe453p*RSe563p + 1./162.*
         m33*m44*pow(RSe453p,2) + 5./81.*m33*m44*RSp563e*RSe563p + 1./
         162.*m33*m44*RSp563e*RSe463p + 1./162.*m33*m44*RSp563e*RSe453p
          + 1./162.*m33*m44*RSp463e*RSe563p + 1./81.*m33*m44*RSp463e*
         RSe463p + 1./162.*m33*m44*RSp453e*RSe563p + 1./81.*m33*m44*
         RSp453e*RSe453p + 31./1296.*m33*m44*RSe45611p + 7./540.*m33*
         m44*RSe561p + 5./81.*m33*m44*RSe561p*RSe563p + 1./162.*m33*m44
         *RSe561p*RSe463p + 1./162.*m33*m44*RSe561p*RSe453p + 5./81.*
         m33*m44*RSe561p*RSp563e + 1./162.*m33*m44*RSe561p*RSp463e + 1./
         162.*m33*m44*RSe561p*RSp453e + 1./162.*m33*m44*RSe461p*RSe563p
          + 1./81.*m33*m44*RSe461p*RSe463p + 1./162.*m33*m44*RSe461p*
         RSp563e );

      M13p6x23R +=  + Abeps(me2,mu2)*pi16 * ( 1./81.*m33*m44*RSe461p*
         RSp463e + 1./162.*m33*m44*RSe451p*RSe563p + 1./81.*m33*m44*
         RSe451p*RSe453p + 1./162.*m33*m44*RSe451p*RSp563e + 1./81.*m33
         *m44*RSe451p*RSp453e + 5./81.*m33*m44*RSp561e*RSe563p + 1./162.
         *m33*m44*RSp561e*RSe463p + 1./162.*m33*m44*RSp561e*RSe453p + 1.
         /162.*m33*m44*RSp461e*RSe563p + 1./81.*m33*m44*RSp461e*RSe463p
          + 1./162.*m33*m44*RSp451e*RSe563p + 1./81.*m33*m44*RSp451e*
         RSe453p + 1./135.*m33*m44*RSe564p + 1./162.*m33*m44*RSe564p*
         pow(RS356pe,2) - 1./162.*m33*m44*RSe564p*pow(RSe563p,2) - 1./
         81.*m33*m44*RSe564p*RSp563e*RSe563p - 1./162.*m33*m44*RSe564p*
         pow(RSp563e,2) - 2./81.*m33*m44*RSe564p*RS156pe*RS356pe - 1./
         81.*m33*m44*RSe564p*RS156pe*RSe563p - 1./81.*m33*m44*RSe564p*
         RS156pe*RSp563e - 2./81.*m33*m44*RSe564p*RSe561p*RS356pe - 1./
         81.*m33*m44*RSe564p*RSe561p*RSe563p - 1./81.*m33*m44*RSe564p*
         RSe561p*RSp563e - 2./81.*m33*m44*RSe564p*RSp561e*RS356pe - 1./
         81.*m33*m44*RSe564p*RSp561e*RSe563p );

      M13p6x23R +=  + Abeps(me2,mu2)*pi16 * (  - 1./81.*m33*m44*RSe564p
         *RSp561e*RSp563e + 1./720.*pow(m33,2)*RSe45613p + 1./81.*pow(
         m33,2)*RSe45613p*RS34561pe - 2./81.*pow(m33,2)*RSe45613p*
         RS14563pe - 2./81.*pow(m33,2)*pow(RSe45613p,2) - 2./81.*pow(
         m33,2)*RSp45613e*RSe45613p + 7./1620.*pow(m33,2)*RS3456pee*
         RSe45613p + 7./1620.*pow(m33,2)*RS3456ppe*RSe45613p + 37./1080.
         *pow(m33,2)*RSe45633p + 193./1620.*pow(m33,2)*RSe45633p*
         RS34561pe + 29./324.*pow(m33,2)*RSe45633p*RS14563pe + 17./162.
         *pow(m33,2)*RSe45633p*RSe45613p + 29./324.*pow(m33,2)*
         RSe45633p*RSp45613e + 1./3240.*pow(m33,2)*RSe45633p*RS3456pee
          + 1./3240.*pow(m33,2)*RSe45633p*RS3456ppe - 1./72.*pow(m33,2)
         *pow(RSe45633p,2) + 5./324.*pow(m33,2)*RSp45633e*RSe45613p - 1.
         /72.*pow(m33,2)*RSp45633e*RSe45633p - 7./1620.*pow(m33,2)*
         RS345pe*RSe45613p - 1./3240.*pow(m33,2)*RS345pe*RSe45633p - 7./
         1620.*pow(m33,2)*RS346pe*RSe45613p - 1./3240.*pow(m33,2)*
         RS346pe*RSe45633p );

      M13p6x23R +=  + Abeps(me2,mu2)*pi16 * (  - 7./1620.*pow(m33,2)*
         RS356pe*RSe45613p - 1./3240.*pow(m33,2)*RS356pe*RSe45633p - 49.
         /2160.*pow(m33,2)*RSe563p - 49./2160.*pow(m33,2)*RSe463p - 49./
         2160.*pow(m33,2)*RSe453p - 1./108.*pow(m33,2)*RS1456pee*
         RSe45613p + 1./648.*pow(m33,2)*RS1456pee*RSe45633p - 1./108.*
         pow(m33,2)*RS1456ppe*RSe45613p + 1./648.*pow(m33,2)*RS1456ppe*
         RSe45633p + 23./432.*pow(m33,2)*RSe45611p + 1./60.*pow(m33,2)*
         RSe45611p*RS34561pe + 1./108.*pow(m33,2)*RSe45611p*RS14563pe
          + 1./54.*pow(m33,2)*RSe45611p*RSe45613p + 1./108.*pow(m33,2)*
         RSe45611p*RSp45613e + 1./72.*pow(m33,2)*RSe45611p*RS3456pee + 
         1./72.*pow(m33,2)*RSe45611p*RS3456ppe - 1./324.*pow(m33,2)*
         RSe45611p*RSe45633p - 1./648.*pow(m33,2)*RSe45611p*RSp45633e
          - 1./72.*pow(m33,2)*RSe45611p*RS345pe - 1./72.*pow(m33,2)*
         RSe45611p*RS346pe - 1./72.*pow(m33,2)*RSe45611p*RS356pe + 5./
         648.*pow(m33,2)*RSe45611p*RS1456pee + 5./648.*pow(m33,2)*
         RSe45611p*RS1456ppe );

      M13p6x23R +=  + Abeps(me2,mu2)*pi16 * (  - 5./648.*pow(m33,2)*
         pow(RSe45611p,2) + 1./108.*pow(m33,2)*RSp45611e*RSe45613p - 1./
         648.*pow(m33,2)*RSp45611e*RSe45633p - 5./648.*pow(m33,2)*
         RSp45611e*RSe45611p + 1./108.*pow(m33,2)*RS145pe*RSe45613p - 1.
         /648.*pow(m33,2)*RS145pe*RSe45633p - 5./648.*pow(m33,2)*
         RS145pe*RSe45611p + 1./108.*pow(m33,2)*RS146pe*RSe45613p - 1./
         648.*pow(m33,2)*RS146pe*RSe45633p - 5./648.*pow(m33,2)*RS146pe
         *RSe45611p + 1./108.*pow(m33,2)*RS156pe*RSe45613p - 1./648.*
         pow(m33,2)*RS156pe*RSe45633p - 5./648.*pow(m33,2)*RS156pe*
         RSe45611p + 1./432.*pow(m33,2)*RSe561p + 1./432.*pow(m33,2)*
         RSe461p + 1./432.*pow(m33,2)*RSe451p - 1./810.*m11*RD3456ep*
         RSe45613p + 5./324.*m11*RSe45633p*RD3456ep - 7./405.*m11*
         RD1456ep*RSe45613p - 11./810.*m11*RD1456ep*RSe45633p - 23./
         1620.*m11*RSe45611p*RD3456ep + 1./81.*m11*RSe45611p*RD1456ep
          + 17./405.*m11*mp2*RSp45613e*RSe45613p + 2./405.*m11*mp2*
         RSe45633p*RSp45613e );

      M13p6x23R +=  + Abeps(me2,mu2)*pi16 * (  - 1./810.*m11*mp2*
         RSp45633e*RSe45613p + 5./324.*m11*mp2*RSp45633e*RSe45633p + 11.
         /405.*m11*mp2*RSe45611p*RSp45613e - 23./1620.*m11*mp2*
         RSe45611p*RSp45633e - 1./45.*m11*mp2*RSp45611e*RSe45613p - 1./
         90.*m11*mp2*RSp45611e*RSe45633p + 2./135.*m11*mp2*RSp45611e*
         RSe45611p + 29./360.*m11*me2*RSe45613p + 1./54.*m11*me2*
         RSe45613p*RS34561pe + 7./270.*m11*me2*RSe45613p*RS14563pe + 49.
         /810.*m11*me2*pow(RSe45613p,2) + 1./54.*m11*me2*RSp45613e*
         RSe45613p + 1./108.*m11*me2*RS3456pee*RSe45613p + 1./108.*m11*
         me2*RS3456ppe*RSe45613p + 1./30.*m11*me2*RSe45633p - 1./108.*
         m11*me2*RSe45633p*RS34561pe - 7./540.*m11*me2*RSe45633p*
         RS14563pe - 2./135.*m11*me2*RSe45633p*RSe45613p - 1./108.*m11*
         me2*RSe45633p*RSp45613e - 1./216.*m11*me2*RSe45633p*RS3456pee
          - 1./216.*m11*me2*RSe45633p*RS3456ppe + 13./648.*m11*me2*pow(
         RSe45633p,2) - 1./108.*m11*me2*RSp45633e*RSe45613p + 1./216.*
         m11*me2*RSp45633e*RSe45633p );

      M13p6x23R +=  + Abeps(me2,mu2)*pi16 * (  - 1./108.*m11*me2*
         RS345pe*RSe45613p + 1./216.*m11*me2*RS345pe*RSe45633p - 1./108.
         *m11*me2*RS346pe*RSe45613p + 1./216.*m11*me2*RS346pe*RSe45633p
          - 1./108.*m11*me2*RS356pe*RSe45613p + 1./216.*m11*me2*RS356pe
         *RSe45633p - 19./540.*m11*me2*RSe563p - 19./540.*m11*me2*
         RSe463p - 19./540.*m11*me2*RSe453p + 7./540.*m11*me2*RS1456pee
         *RSe45613p - 7./1080.*m11*me2*RS1456pee*RSe45633p + 7./540.*
         m11*me2*RS1456ppe*RSe45613p - 7./1080.*m11*me2*RS1456ppe*
         RSe45633p + 2./45.*m11*me2*RSe45611p - 1./108.*m11*me2*
         RSe45611p*RS34561pe - 7./540.*m11*me2*RSe45611p*RS14563pe - 11.
         /810.*m11*me2*RSe45611p*RSe45613p - 1./108.*m11*me2*RSe45611p*
         RSp45613e - 1./216.*m11*me2*RSe45611p*RS3456pee - 1./216.*m11*
         me2*RSe45611p*RS3456ppe - 13./810.*m11*me2*RSe45611p*RSe45633p
          + 1./216.*m11*me2*RSe45611p*RSp45633e + 1./216.*m11*me2*
         RSe45611p*RS345pe + 1./216.*m11*me2*RSe45611p*RS346pe + 1./216.
         *m11*me2*RSe45611p*RS356pe );

      M13p6x23R +=  + Abeps(me2,mu2)*pi16 * (  - 7./1080.*m11*me2*
         RSe45611p*RS1456pee - 7./1080.*m11*me2*RSe45611p*RS1456ppe + 7.
         /360.*m11*me2*pow(RSe45611p,2) - 1./108.*m11*me2*RSp45611e*
         RSe45613p + 1./216.*m11*me2*RSp45611e*RSe45633p + 1./216.*m11*
         me2*RSp45611e*RSe45611p - 7./540.*m11*me2*RS145pe*RSe45613p + 
         7./1080.*m11*me2*RS145pe*RSe45633p + 7./1080.*m11*me2*RS145pe*
         RSe45611p - 7./540.*m11*me2*RS146pe*RSe45613p + 7./1080.*m11*
         me2*RS146pe*RSe45633p + 7./1080.*m11*me2*RS146pe*RSe45611p - 7.
         /540.*m11*me2*RS156pe*RSe45613p + 7./1080.*m11*me2*RS156pe*
         RSe45633p + 7./1080.*m11*me2*RS156pe*RSe45611p + 7./540.*m11*
         me2*RSe561p + 7./540.*m11*me2*RSe461p + 7./540.*m11*me2*
         RSe451p + 19./360.*m11*me2*RSe564p + 19./360.*m11*me2*RSe465p
          + 19./360.*m11*me2*RSe456p - 7./3240.*m11*m66*RSe45613p + 31./
         1296.*m11*m66*RSe45633p + 7./540.*m11*m66*RSe453p + 89./2160.*
         m11*m66*RSe45611p + 1./81.*m11*m66*RSe561p*RSe563p + 1./162.*
         m11*m66*RSe561p*RSe453p );

      M13p6x23R +=  + Abeps(me2,mu2)*pi16 * ( 1./81.*m11*m66*RSe561p*
         RSp563e + 1./162.*m11*m66*RSe561p*RSp453e + 1./162.*m11*m66*
         pow(RSe561p,2) + 1./81.*m11*m66*RSe461p*RSe463p + 1./162.*m11*
         m66*RSe461p*RSe453p + 1./81.*m11*m66*RSe461p*RSp463e + 1./162.
         *m11*m66*RSe461p*RSp453e + 1./162.*m11*m66*pow(RSe461p,2) - 5./
         108.*m11*m66*RSe451p + 1./162.*m11*m66*RSe451p*RSe563p + 1./
         162.*m11*m66*RSe451p*RSe463p + 5./81.*m11*m66*RSe451p*RSe453p
          + 1./162.*m11*m66*RSe451p*RSp563e + 1./162.*m11*m66*RSe451p*
         RSp463e + 5./81.*m11*m66*RSe451p*RSp453e + 1./162.*m11*m66*
         RSe451p*RSe561p + 1./162.*m11*m66*RSe451p*RSe461p + 5./162.*
         m11*m66*pow(RSe451p,2) + 1./81.*m11*m66*RSp561e*RSe563p + 1./
         162.*m11*m66*RSp561e*RSe453p + 1./81.*m11*m66*RSp561e*RSe561p
          + 1./162.*m11*m66*RSp561e*RSe451p + 1./81.*m11*m66*RSp461e*
         RSe463p + 1./162.*m11*m66*RSp461e*RSe453p + 1./81.*m11*m66*
         RSp461e*RSe461p + 1./162.*m11*m66*RSp461e*RSe451p + 1./162.*
         m11*m66*RSp451e*RSe563p );

      M13p6x23R +=  + Abeps(me2,mu2)*pi16 * ( 1./162.*m11*m66*RSp451e*
         RSe463p + 5./81.*m11*m66*RSp451e*RSe453p + 1./162.*m11*m66*
         RSp451e*RSe561p + 1./162.*m11*m66*RSp451e*RSe461p + 5./81.*m11
         *m66*RSp451e*RSe451p + 1./135.*m11*m66*RSe456p - 2./81.*m11*
         m66*RSe456p*RS145pe*RS345pe - 2./81.*m11*m66*RSe456p*RS145pe*
         RSe453p - 2./81.*m11*m66*RSe456p*RS145pe*RSp453e + 1./162.*m11
         *m66*RSe456p*pow(RS145pe,2) - 1./81.*m11*m66*RSe456p*RSe451p*
         RS345pe - 1./81.*m11*m66*RSe456p*RSe451p*RSe453p - 1./81.*m11*
         m66*RSe456p*RSe451p*RSp453e - 1./162.*m11*m66*RSe456p*pow(
         RSe451p,2) - 1./81.*m11*m66*RSe456p*RSp451e*RS345pe - 1./81.*
         m11*m66*RSe456p*RSp451e*RSe453p - 1./81.*m11*m66*RSe456p*
         RSp451e*RSp453e - 1./81.*m11*m66*RSe456p*RSp451e*RSe451p - 1./
         162.*m11*m66*RSe456p*pow(RSp451e,2) - 7./3240.*m11*m55*
         RSe45613p + 31./1296.*m11*m55*RSe45633p + 7./540.*m11*m55*
         RSe463p + 89./2160.*m11*m55*RSe45611p + 1./81.*m11*m55*RSe561p
         *RSe563p );

      M13p6x23R +=  + Abeps(me2,mu2)*pi16 * ( 1./162.*m11*m55*RSe561p*
         RSe463p + 1./81.*m11*m55*RSe561p*RSp563e + 1./162.*m11*m55*
         RSe561p*RSp463e + 1./162.*m11*m55*pow(RSe561p,2) - 5./108.*m11
         *m55*RSe461p + 1./162.*m11*m55*RSe461p*RSe563p + 5./81.*m11*
         m55*RSe461p*RSe463p + 1./162.*m11*m55*RSe461p*RSe453p + 1./162.
         *m11*m55*RSe461p*RSp563e + 5./81.*m11*m55*RSe461p*RSp463e + 1./
         162.*m11*m55*RSe461p*RSp453e + 1./162.*m11*m55*RSe461p*RSe561p
          + 5./162.*m11*m55*pow(RSe461p,2) + 1./162.*m11*m55*RSe451p*
         RSe463p + 1./81.*m11*m55*RSe451p*RSe453p + 1./162.*m11*m55*
         RSe451p*RSp463e + 1./81.*m11*m55*RSe451p*RSp453e + 1./162.*m11
         *m55*RSe451p*RSe461p + 1./162.*m11*m55*pow(RSe451p,2) + 1./81.
         *m11*m55*RSp561e*RSe563p + 1./162.*m11*m55*RSp561e*RSe463p + 1.
         /81.*m11*m55*RSp561e*RSe561p + 1./162.*m11*m55*RSp561e*RSe461p
          + 1./162.*m11*m55*RSp461e*RSe563p + 5./81.*m11*m55*RSp461e*
         RSe463p + 1./162.*m11*m55*RSp461e*RSe453p + 1./162.*m11*m55*
         RSp461e*RSe561p );

      M13p6x23R +=  + Abeps(me2,mu2)*pi16 * ( 5./81.*m11*m55*RSp461e*
         RSe461p + 1./162.*m11*m55*RSp461e*RSe451p + 1./162.*m11*m55*
         RSp451e*RSe463p + 1./81.*m11*m55*RSp451e*RSe453p + 1./162.*m11
         *m55*RSp451e*RSe461p + 1./81.*m11*m55*RSp451e*RSe451p + 1./135.
         *m11*m55*RSe465p - 2./81.*m11*m55*RSe465p*RS146pe*RS346pe - 2./
         81.*m11*m55*RSe465p*RS146pe*RSe463p - 2./81.*m11*m55*RSe465p*
         RS146pe*RSp463e + 1./162.*m11*m55*RSe465p*pow(RS146pe,2) - 1./
         81.*m11*m55*RSe465p*RSe461p*RS346pe - 1./81.*m11*m55*RSe465p*
         RSe461p*RSe463p - 1./81.*m11*m55*RSe465p*RSe461p*RSp463e - 1./
         162.*m11*m55*RSe465p*pow(RSe461p,2) - 1./81.*m11*m55*RSe465p*
         RSp461e*RS346pe - 1./81.*m11*m55*RSe465p*RSp461e*RSe463p - 1./
         81.*m11*m55*RSe465p*RSp461e*RSp463e - 1./81.*m11*m55*RSe465p*
         RSp461e*RSe461p - 1./162.*m11*m55*RSe465p*pow(RSp461e,2) - 7./
         3240.*m11*m44*RSe45613p + 31./1296.*m11*m44*RSe45633p + 7./540.
         *m11*m44*RSe563p + 89./2160.*m11*m44*RSe45611p - 5./108.*m11*
         m44*RSe561p );

      M13p6x23R +=  + Abeps(me2,mu2)*pi16 * ( 5./81.*m11*m44*RSe561p*
         RSe563p + 1./162.*m11*m44*RSe561p*RSe463p + 1./162.*m11*m44*
         RSe561p*RSe453p + 5./81.*m11*m44*RSe561p*RSp563e + 1./162.*m11
         *m44*RSe561p*RSp463e + 1./162.*m11*m44*RSe561p*RSp453e + 5./
         162.*m11*m44*pow(RSe561p,2) + 1./162.*m11*m44*RSe461p*RSe563p
          + 1./81.*m11*m44*RSe461p*RSe463p + 1./162.*m11*m44*RSe461p*
         RSp563e + 1./81.*m11*m44*RSe461p*RSp463e + 1./162.*m11*m44*
         RSe461p*RSe561p + 1./162.*m11*m44*pow(RSe461p,2) + 1./162.*m11
         *m44*RSe451p*RSe563p + 1./81.*m11*m44*RSe451p*RSe453p + 1./162.
         *m11*m44*RSe451p*RSp563e + 1./81.*m11*m44*RSe451p*RSp453e + 1./
         162.*m11*m44*RSe451p*RSe561p + 1./162.*m11*m44*pow(RSe451p,2)
          + 5./81.*m11*m44*RSp561e*RSe563p + 1./162.*m11*m44*RSp561e*
         RSe463p + 1./162.*m11*m44*RSp561e*RSe453p + 5./81.*m11*m44*
         RSp561e*RSe561p + 1./162.*m11*m44*RSp561e*RSe461p + 1./162.*
         m11*m44*RSp561e*RSe451p + 1./162.*m11*m44*RSp461e*RSe563p + 1./
         81.*m11*m44*RSp461e*RSe463p );

      M13p6x23R +=  + Abeps(me2,mu2)*pi16 * ( 1./162.*m11*m44*RSp461e*
         RSe561p + 1./81.*m11*m44*RSp461e*RSe461p + 1./162.*m11*m44*
         RSp451e*RSe563p + 1./81.*m11*m44*RSp451e*RSe453p + 1./162.*m11
         *m44*RSp451e*RSe561p + 1./81.*m11*m44*RSp451e*RSe451p + 1./135.
         *m11*m44*RSe564p - 2./81.*m11*m44*RSe564p*RS156pe*RS356pe - 2./
         81.*m11*m44*RSe564p*RS156pe*RSe563p - 2./81.*m11*m44*RSe564p*
         RS156pe*RSp563e + 1./162.*m11*m44*RSe564p*pow(RS156pe,2) - 1./
         81.*m11*m44*RSe564p*RSe561p*RS356pe - 1./81.*m11*m44*RSe564p*
         RSe561p*RSe563p - 1./81.*m11*m44*RSe564p*RSe561p*RSp563e - 1./
         162.*m11*m44*RSe564p*pow(RSe561p,2) - 1./81.*m11*m44*RSe564p*
         RSp561e*RS356pe - 1./81.*m11*m44*RSe564p*RSp561e*RSe563p - 1./
         81.*m11*m44*RSe564p*RSp561e*RSp563e - 1./81.*m11*m44*RSe564p*
         RSp561e*RSe561p - 1./162.*m11*m44*RSe564p*pow(RSp561e,2) + 163.
         /1080.*m11*m33*RSe45613p - 1./135.*m11*m33*RSe45613p*RS34561pe
          - 1./135.*m11*m33*RSe45613p*RS14563pe - 4./81.*m11*m33*pow(
         RSe45613p,2) );

      M13p6x23R +=  + Abeps(me2,mu2)*pi16 * (  - 4./81.*m11*m33*
         RSp45613e*RSe45613p - 19./810.*m11*m33*RS3456pee*RSe45613p - 
         19./810.*m11*m33*RS3456ppe*RSe45613p + 53./2160.*m11*m33*
         RSe45633p + 1./15.*m11*m33*RSe45633p*RS34561pe + 4./45.*m11*
         m33*RSe45633p*RS14563pe + 7./81.*m11*m33*RSe45633p*RSe45613p
          + 5./81.*m11*m33*RSe45633p*RSp45613e + 1./162.*m11*m33*
         RSe45633p*RS3456pee + 1./162.*m11*m33*RSe45633p*RS3456ppe - 7./
         324.*m11*m33*pow(RSe45633p,2) + 2./81.*m11*m33*RSp45633e*
         RSe45613p - 7./324.*m11*m33*RSp45633e*RSe45633p + 19./810.*m11
         *m33*RS345pe*RSe45613p - 1./162.*m11*m33*RS345pe*RSe45633p + 
         19./810.*m11*m33*RS346pe*RSe45613p - 1./162.*m11*m33*RS346pe*
         RSe45633p + 19./810.*m11*m33*RS356pe*RSe45613p - 1./162.*m11*
         m33*RS356pe*RSe45633p - 13./540.*m11*m33*RSe563p - 13./540.*
         m11*m33*RSe463p - 13./540.*m11*m33*RSe453p - 19./810.*m11*m33*
         RS1456pee*RSe45613p + 7./405.*m11*m33*RS1456pee*RSe45633p - 19.
         /810.*m11*m33*RS1456ppe*RSe45613p );

      M13p6x23R +=  + Abeps(me2,mu2)*pi16 * ( 7./405.*m11*m33*RS1456ppe
         *RSe45633p + 53./2160.*m11*m33*RSe45611p + 4./45.*m11*m33*
         RSe45611p*RS34561pe + 1./15.*m11*m33*RSe45611p*RS14563pe + 7./
         81.*m11*m33*RSe45611p*RSe45613p + 5./81.*m11*m33*RSe45611p*
         RSp45613e + 7./405.*m11*m33*RSe45611p*RS3456pee + 7./405.*m11*
         m33*RSe45611p*RS3456ppe - 1./162.*m11*m33*RSe45611p*RSe45633p
          - 1./324.*m11*m33*RSe45611p*RSp45633e - 7./405.*m11*m33*
         RSe45611p*RS345pe - 7./405.*m11*m33*RSe45611p*RS346pe - 7./405.
         *m11*m33*RSe45611p*RS356pe + 1./162.*m11*m33*RSe45611p*
         RS1456pee + 1./162.*m11*m33*RSe45611p*RS1456ppe - 7./324.*m11*
         m33*pow(RSe45611p,2) + 2./81.*m11*m33*RSp45611e*RSe45613p - 1./
         324.*m11*m33*RSp45611e*RSe45633p - 7./324.*m11*m33*RSp45611e*
         RSe45611p + 19./810.*m11*m33*RS145pe*RSe45613p - 7./405.*m11*
         m33*RS145pe*RSe45633p - 1./162.*m11*m33*RS145pe*RSe45611p + 19.
         /810.*m11*m33*RS146pe*RSe45613p - 7./405.*m11*m33*RS146pe*
         RSe45633p );

      M13p6x23R +=  + Abeps(me2,mu2)*pi16 * (  - 1./162.*m11*m33*
         RS146pe*RSe45611p + 19./810.*m11*m33*RS156pe*RSe45613p - 7./
         405.*m11*m33*RS156pe*RSe45633p - 1./162.*m11*m33*RS156pe*
         RSe45611p - 13./540.*m11*m33*RSe561p - 13./540.*m11*m33*
         RSe461p - 13./540.*m11*m33*RSe451p + 1./720.*pow(m11,2)*
         RSe45613p - 2./81.*pow(m11,2)*RSe45613p*RS34561pe + 1./81.*
         pow(m11,2)*RSe45613p*RS14563pe - 2./81.*pow(m11,2)*pow(
         RSe45613p,2) - 2./81.*pow(m11,2)*RSp45613e*RSe45613p - 1./108.
         *pow(m11,2)*RS3456pee*RSe45613p - 1./108.*pow(m11,2)*RS3456ppe
         *RSe45613p + 23./432.*pow(m11,2)*RSe45633p + 1./108.*pow(
         m11,2)*RSe45633p*RS34561pe + 1./60.*pow(m11,2)*RSe45633p*
         RS14563pe + 1./54.*pow(m11,2)*RSe45633p*RSe45613p + 1./108.*
         pow(m11,2)*RSe45633p*RSp45613e + 5./648.*pow(m11,2)*RSe45633p*
         RS3456pee + 5./648.*pow(m11,2)*RSe45633p*RS3456ppe - 5./648.*
         pow(m11,2)*pow(RSe45633p,2) + 1./108.*pow(m11,2)*RSp45633e*
         RSe45613p );

      M13p6x23R +=  + Abeps(me2,mu2)*pi16 * (  - 5./648.*pow(m11,2)*
         RSp45633e*RSe45633p + 1./108.*pow(m11,2)*RS345pe*RSe45613p - 5.
         /648.*pow(m11,2)*RS345pe*RSe45633p + 1./108.*pow(m11,2)*
         RS346pe*RSe45613p - 5./648.*pow(m11,2)*RS346pe*RSe45633p + 1./
         108.*pow(m11,2)*RS356pe*RSe45613p - 5./648.*pow(m11,2)*RS356pe
         *RSe45633p + 1./432.*pow(m11,2)*RSe563p + 1./432.*pow(m11,2)*
         RSe463p + 1./432.*pow(m11,2)*RSe453p + 7./1620.*pow(m11,2)*
         RS1456pee*RSe45613p + 1./72.*pow(m11,2)*RS1456pee*RSe45633p + 
         7./1620.*pow(m11,2)*RS1456ppe*RSe45613p + 1./72.*pow(m11,2)*
         RS1456ppe*RSe45633p + 37./1080.*pow(m11,2)*RSe45611p + 29./324.
         *pow(m11,2)*RSe45611p*RS34561pe + 193./1620.*pow(m11,2)*
         RSe45611p*RS14563pe + 17./162.*pow(m11,2)*RSe45611p*RSe45613p
          + 29./324.*pow(m11,2)*RSe45611p*RSp45613e + 1./648.*pow(
         m11,2)*RSe45611p*RS3456pee + 1./648.*pow(m11,2)*RSe45611p*
         RS3456ppe - 1./324.*pow(m11,2)*RSe45611p*RSe45633p - 1./648.*
         pow(m11,2)*RSe45611p*RSp45633e );

      M13p6x23R +=  + Abeps(me2,mu2)*pi16 * (  - 1./648.*pow(m11,2)*
         RSe45611p*RS345pe - 1./648.*pow(m11,2)*RSe45611p*RS346pe - 1./
         648.*pow(m11,2)*RSe45611p*RS356pe + 1./3240.*pow(m11,2)*
         RSe45611p*RS1456pee + 1./3240.*pow(m11,2)*RSe45611p*RS1456ppe
          - 1./72.*pow(m11,2)*pow(RSe45611p,2) + 5./324.*pow(m11,2)*
         RSp45611e*RSe45613p - 1./648.*pow(m11,2)*RSp45611e*RSe45633p
          - 1./72.*pow(m11,2)*RSp45611e*RSe45611p - 7./1620.*pow(m11,2)
         *RS145pe*RSe45613p - 1./72.*pow(m11,2)*RS145pe*RSe45633p - 1./
         3240.*pow(m11,2)*RS145pe*RSe45611p - 7./1620.*pow(m11,2)*
         RS146pe*RSe45613p - 1./72.*pow(m11,2)*RS146pe*RSe45633p - 1./
         3240.*pow(m11,2)*RS146pe*RSe45611p - 7./1620.*pow(m11,2)*
         RS156pe*RSe45613p - 1./72.*pow(m11,2)*RS156pe*RSe45633p - 1./
         3240.*pow(m11,2)*RS156pe*RSe45611p - 49./2160.*pow(m11,2)*
         RSe561p - 49./2160.*pow(m11,2)*RSe461p - 49./2160.*pow(m11,2)*
         RSe451p );

      M13p6x23R +=  + Abeps(mp2,mu2)*pi16 * (  - 1./270.*mp2*RD3456ep*
         RSp45613e + 1./540.*mp2*RSp45633e*RD3456ep - 1./270.*mp2*
         RD1456ep*RSp45613e + 1./540.*mp2*RD1456ep*RSp45633e + 1./540.*
         mp2*RSp45611e*RD3456ep + 1./540.*mp2*RSp45611e*RD1456ep + 1./
         405.*pow(mp2,2)*pow(RSp45613e,2) - 1./405.*pow(mp2,2)*
         RSp45633e*RSp45613e + 1./1620.*pow(mp2,2)*pow(RSp45633e,2) - 1.
         /405.*pow(mp2,2)*RSp45611e*RSp45613e + 1./810.*pow(mp2,2)*
         RSp45611e*RSp45633e + 1./1620.*pow(mp2,2)*pow(RSp45611e,2) + 1.
         /135.*me2*mp2*RSp45613e*RSe45613p - 1./270.*me2*mp2*RSe45633p*
         RSp45613e - 1./270.*me2*mp2*RSp45633e*RSe45613p + 1./540.*me2*
         mp2*RSp45633e*RSe45633p - 1./270.*me2*mp2*RSe45611p*RSp45613e
          + 1./540.*me2*mp2*RSe45611p*RSp45633e - 1./270.*me2*mp2*
         RSp45611e*RSe45613p + 1./540.*me2*mp2*RSp45611e*RSe45633p + 1./
         540.*me2*mp2*RSp45611e*RSe45611p - 2./405.*pow(me2,2)*
         RSp45613e*RSe45613p + 1./405.*pow(me2,2)*RSe45633p*RSp45613e
          + 1./405.*pow(me2,2)*RSp45633e*RSe45613p );

      M13p6x23R +=  + Abeps(mp2,mu2)*pi16 * (  - 1./810.*pow(me2,2)*
         RSp45633e*RSe45633p + 1./405.*pow(me2,2)*RSe45611p*RSp45613e
          - 1./810.*pow(me2,2)*RSe45611p*RSp45633e + 1./405.*pow(me2,2)
         *RSp45611e*RSe45613p - 1./810.*pow(me2,2)*RSp45611e*RSe45633p
          - 1./810.*pow(me2,2)*RSp45611e*RSe45611p - 4./135.*m66*mp2*
         RSp45613e + 7./270.*m66*mp2*RSp45633e - 1./162.*m66*mp2*
         RSp563e*RS345pe - 1./81.*m66*mp2*RSp563e*RS356pe - 1./81.*m66*
         mp2*RSp563e*RSe563p - 1./162.*m66*mp2*RSp563e*RSe453p - 1./81.
         *m66*mp2*pow(RSp563e,2) - 1./162.*m66*mp2*RSp463e*RS345pe - 1./
         81.*m66*mp2*RSp463e*RS346pe - 1./81.*m66*mp2*RSp463e*RSe463p
          - 1./162.*m66*mp2*RSp463e*RSe453p - 1./81.*m66*mp2*pow(
         RSp463e,2) - 1./45.*m66*mp2*RSp453e - 5./81.*m66*mp2*RSp453e*
         RS345pe - 1./162.*m66*mp2*RSp453e*RS346pe - 1./162.*m66*mp2*
         RSp453e*RS356pe - 1./162.*m66*mp2*RSp453e*RSe563p - 1./162.*
         m66*mp2*RSp453e*RSe463p - 5./81.*m66*mp2*RSp453e*RSe453p - 1./
         81.*m66*mp2*RSp453e*RSp563e );

      M13p6x23R +=  + Abeps(mp2,mu2)*pi16 * (  - 1./81.*m66*mp2*RSp453e
         *RSp463e - 5./81.*m66*mp2*pow(RSp453e,2) + 7./270.*m66*mp2*
         RSp45611e + 1./162.*m66*mp2*RS145pe*RSp563e + 1./162.*m66*mp2*
         RS145pe*RSp463e + 5./81.*m66*mp2*RS145pe*RSp453e + 1./81.*m66*
         mp2*RS146pe*RSp463e + 1./162.*m66*mp2*RS146pe*RSp453e + 1./81.
         *m66*mp2*RS156pe*RSp563e + 1./162.*m66*mp2*RS156pe*RSp453e + 1.
         /81.*m66*mp2*RSe561p*RSp563e + 1./162.*m66*mp2*RSe561p*RSp453e
          + 1./81.*m66*mp2*RSe461p*RSp463e + 1./162.*m66*mp2*RSe461p*
         RSp453e + 1./162.*m66*mp2*RSe451p*RSp563e + 1./162.*m66*mp2*
         RSe451p*RSp463e + 5./81.*m66*mp2*RSe451p*RSp453e + 1./162.*m66
         *mp2*RSp561e*RS345pe + 1./81.*m66*mp2*RSp561e*RS356pe + 1./81.
         *m66*mp2*RSp561e*RSe563p + 1./162.*m66*mp2*RSp561e*RSe453p + 2.
         /81.*m66*mp2*RSp561e*RSp563e + 1./81.*m66*mp2*RSp561e*RSp453e
          - 1./162.*m66*mp2*RSp561e*RS145pe - 1./81.*m66*mp2*RSp561e*
         RS156pe - 1./81.*m66*mp2*RSp561e*RSe561p - 1./162.*m66*mp2*
         RSp561e*RSe451p );

      M13p6x23R +=  + Abeps(mp2,mu2)*pi16 * (  - 1./81.*m66*mp2*pow(
         RSp561e,2) + 1./162.*m66*mp2*RSp461e*RS345pe + 1./81.*m66*mp2*
         RSp461e*RS346pe + 1./81.*m66*mp2*RSp461e*RSe463p + 1./162.*m66
         *mp2*RSp461e*RSe453p + 2./81.*m66*mp2*RSp461e*RSp463e + 1./81.
         *m66*mp2*RSp461e*RSp453e - 1./162.*m66*mp2*RSp461e*RS145pe - 1.
         /81.*m66*mp2*RSp461e*RS146pe - 1./81.*m66*mp2*RSp461e*RSe461p
          - 1./162.*m66*mp2*RSp461e*RSe451p - 1./81.*m66*mp2*pow(
         RSp461e,2) - 1./45.*m66*mp2*RSp451e + 5./81.*m66*mp2*RSp451e*
         RS345pe + 1./162.*m66*mp2*RSp451e*RS346pe + 1./162.*m66*mp2*
         RSp451e*RS356pe + 1./162.*m66*mp2*RSp451e*RSe563p + 1./162.*
         m66*mp2*RSp451e*RSe463p + 5./81.*m66*mp2*RSp451e*RSe453p + 1./
         81.*m66*mp2*RSp451e*RSp563e + 1./81.*m66*mp2*RSp451e*RSp463e
          + 10./81.*m66*mp2*RSp451e*RSp453e - 5./81.*m66*mp2*RSp451e*
         RS145pe - 1./162.*m66*mp2*RSp451e*RS146pe - 1./162.*m66*mp2*
         RSp451e*RS156pe - 1./162.*m66*mp2*RSp451e*RSe561p - 1./162.*
         m66*mp2*RSp451e*RSe461p );

      M13p6x23R +=  + Abeps(mp2,mu2)*pi16 * (  - 5./81.*m66*mp2*RSp451e
         *RSe451p - 1./81.*m66*mp2*RSp451e*RSp561e - 1./81.*m66*mp2*
         RSp451e*RSp461e - 5./81.*m66*mp2*pow(RSp451e,2) + 1./45.*m66*
         mp2*RSp456e + 1./81.*m66*mp2*RSp456e*RSp453e*RS345pe + 1./81.*
         m66*mp2*RSp456e*RSp453e*RSe453p + 1./54.*m66*mp2*RSp456e*pow(
         RSp453e,2) - 1./81.*m66*mp2*RSp456e*RS145pe*RSp453e - 1./81.*
         m66*mp2*RSp456e*RSe451p*RSp453e - 1./81.*m66*mp2*RSp456e*
         RSp451e*RS345pe - 1./81.*m66*mp2*RSp456e*RSp451e*RSe453p - 1./
         27.*m66*mp2*RSp456e*RSp451e*RSp453e + 1./81.*m66*mp2*RSp456e*
         RSp451e*RS145pe + 1./81.*m66*mp2*RSp456e*RSp451e*RSe451p + 1./
         54.*m66*mp2*RSp456e*pow(RSp451e,2) + 1./162.*m66*me2*RSe456p*
         pow(RSp453e,2) - 1./81.*m66*me2*RSe456p*RSp451e*RSp453e + 1./
         162.*m66*me2*RSe456p*pow(RSp451e,2) + 1./81.*m66*me2*RSp456e*
         RSe453p*RS345pe + 1./81.*m66*me2*RSp456e*pow(RSe453p,2) + 1./
         81.*m66*me2*RSp456e*RSp453e*RSe453p - 1./81.*m66*me2*RSp456e*
         RS145pe*RSe453p );

      M13p6x23R +=  + Abeps(mp2,mu2)*pi16 * (  - 1./81.*m66*me2*RSp456e
         *RSe451p*RS345pe - 2./81.*m66*me2*RSp456e*RSe451p*RSe453p - 1./
         81.*m66*me2*RSp456e*RSe451p*RSp453e + 1./81.*m66*me2*RSp456e*
         RSe451p*RS145pe + 1./81.*m66*me2*RSp456e*pow(RSe451p,2) - 1./
         81.*m66*me2*RSp456e*RSp451e*RSe453p + 1./81.*m66*me2*RSp456e*
         RSp451e*RSe451p + 1./90.*pow(m66,2)*RSp453e - 1./162.*pow(
         m66,2)*RSp453e*RSp563e - 1./162.*pow(m66,2)*RSp453e*RSp463e + 
         1./486.*pow(m66,2)*pow(RSp453e,2) + 1./162.*pow(m66,2)*RS645pe
         *pow(RSp453e,2) + 1./162.*pow(m66,2)*RSp561e*RSp453e + 1./162.
         *pow(m66,2)*RSp461e*RSp453e + 1./90.*pow(m66,2)*RSp451e + 1./
         162.*pow(m66,2)*RSp451e*RSp563e + 1./162.*pow(m66,2)*RSp451e*
         RSp463e - 1./243.*pow(m66,2)*RSp451e*RSp453e - 1./81.*pow(
         m66,2)*RSp451e*RS645pe*RSp453e - 1./162.*pow(m66,2)*RSp451e*
         RSp561e - 1./162.*pow(m66,2)*RSp451e*RSp461e + 1./486.*pow(
         m66,2)*pow(RSp451e,2) + 1./162.*pow(m66,2)*pow(RSp451e,2)*
         RS645pe );

      M13p6x23R +=  + Abeps(mp2,mu2)*pi16 * (  - 1./45.*pow(m66,2)*
         RSp456e - 4./135.*m55*mp2*RSp45613e + 7./270.*m55*mp2*
         RSp45633e - 1./162.*m55*mp2*RSp563e*RS346pe - 1./81.*m55*mp2*
         RSp563e*RS356pe - 1./81.*m55*mp2*RSp563e*RSe563p - 1./162.*m55
         *mp2*RSp563e*RSe463p - 1./81.*m55*mp2*pow(RSp563e,2) - 1./45.*
         m55*mp2*RSp463e - 1./162.*m55*mp2*RSp463e*RS345pe - 5./81.*m55
         *mp2*RSp463e*RS346pe - 1./162.*m55*mp2*RSp463e*RS356pe - 1./
         162.*m55*mp2*RSp463e*RSe563p - 5./81.*m55*mp2*RSp463e*RSe463p
          - 1./162.*m55*mp2*RSp463e*RSe453p - 1./81.*m55*mp2*RSp463e*
         RSp563e - 5./81.*m55*mp2*pow(RSp463e,2) - 1./81.*m55*mp2*
         RSp453e*RS345pe - 1./162.*m55*mp2*RSp453e*RS346pe - 1./162.*
         m55*mp2*RSp453e*RSe463p - 1./81.*m55*mp2*RSp453e*RSe453p - 1./
         81.*m55*mp2*RSp453e*RSp463e - 1./81.*m55*mp2*pow(RSp453e,2) + 
         7./270.*m55*mp2*RSp45611e + 1./162.*m55*mp2*RS145pe*RSp463e + 
         1./81.*m55*mp2*RS145pe*RSp453e + 1./162.*m55*mp2*RS146pe*
         RSp563e );

      M13p6x23R +=  + Abeps(mp2,mu2)*pi16 * ( 5./81.*m55*mp2*RS146pe*
         RSp463e + 1./162.*m55*mp2*RS146pe*RSp453e + 1./81.*m55*mp2*
         RS156pe*RSp563e + 1./162.*m55*mp2*RS156pe*RSp463e + 1./81.*m55
         *mp2*RSe561p*RSp563e + 1./162.*m55*mp2*RSe561p*RSp463e + 1./
         162.*m55*mp2*RSe461p*RSp563e + 5./81.*m55*mp2*RSe461p*RSp463e
          + 1./162.*m55*mp2*RSe461p*RSp453e + 1./162.*m55*mp2*RSe451p*
         RSp463e + 1./81.*m55*mp2*RSe451p*RSp453e + 1./162.*m55*mp2*
         RSp561e*RS346pe + 1./81.*m55*mp2*RSp561e*RS356pe + 1./81.*m55*
         mp2*RSp561e*RSe563p + 1./162.*m55*mp2*RSp561e*RSe463p + 2./81.
         *m55*mp2*RSp561e*RSp563e + 1./81.*m55*mp2*RSp561e*RSp463e - 1./
         162.*m55*mp2*RSp561e*RS146pe - 1./81.*m55*mp2*RSp561e*RS156pe
          - 1./81.*m55*mp2*RSp561e*RSe561p - 1./162.*m55*mp2*RSp561e*
         RSe461p - 1./81.*m55*mp2*pow(RSp561e,2) - 1./45.*m55*mp2*
         RSp461e + 1./162.*m55*mp2*RSp461e*RS345pe + 5./81.*m55*mp2*
         RSp461e*RS346pe + 1./162.*m55*mp2*RSp461e*RS356pe + 1./162.*
         m55*mp2*RSp461e*RSe563p );

      M13p6x23R +=  + Abeps(mp2,mu2)*pi16 * ( 5./81.*m55*mp2*RSp461e*
         RSe463p + 1./162.*m55*mp2*RSp461e*RSe453p + 1./81.*m55*mp2*
         RSp461e*RSp563e + 10./81.*m55*mp2*RSp461e*RSp463e + 1./81.*m55
         *mp2*RSp461e*RSp453e - 1./162.*m55*mp2*RSp461e*RS145pe - 5./81.
         *m55*mp2*RSp461e*RS146pe - 1./162.*m55*mp2*RSp461e*RS156pe - 1.
         /162.*m55*mp2*RSp461e*RSe561p - 5./81.*m55*mp2*RSp461e*RSe461p
          - 1./162.*m55*mp2*RSp461e*RSe451p - 1./81.*m55*mp2*RSp461e*
         RSp561e - 5./81.*m55*mp2*pow(RSp461e,2) + 1./81.*m55*mp2*
         RSp451e*RS345pe + 1./162.*m55*mp2*RSp451e*RS346pe + 1./162.*
         m55*mp2*RSp451e*RSe463p + 1./81.*m55*mp2*RSp451e*RSe453p + 1./
         81.*m55*mp2*RSp451e*RSp463e + 2./81.*m55*mp2*RSp451e*RSp453e
          - 1./81.*m55*mp2*RSp451e*RS145pe - 1./162.*m55*mp2*RSp451e*
         RS146pe - 1./162.*m55*mp2*RSp451e*RSe461p - 1./81.*m55*mp2*
         RSp451e*RSe451p - 1./81.*m55*mp2*RSp451e*RSp461e - 1./81.*m55*
         mp2*pow(RSp451e,2) + 1./45.*m55*mp2*RSp465e + 1./81.*m55*mp2*
         RSp465e*RSp463e*RS346pe );

      M13p6x23R +=  + Abeps(mp2,mu2)*pi16 * ( 1./81.*m55*mp2*RSp465e*
         RSp463e*RSe463p + 1./54.*m55*mp2*RSp465e*pow(RSp463e,2) - 1./
         81.*m55*mp2*RSp465e*RS146pe*RSp463e - 1./81.*m55*mp2*RSp465e*
         RSe461p*RSp463e - 1./81.*m55*mp2*RSp465e*RSp461e*RS346pe - 1./
         81.*m55*mp2*RSp465e*RSp461e*RSe463p - 1./27.*m55*mp2*RSp465e*
         RSp461e*RSp463e + 1./81.*m55*mp2*RSp465e*RSp461e*RS146pe + 1./
         81.*m55*mp2*RSp465e*RSp461e*RSe461p + 1./54.*m55*mp2*RSp465e*
         pow(RSp461e,2) + 1./162.*m55*me2*RSe465p*pow(RSp463e,2) - 1./
         81.*m55*me2*RSe465p*RSp461e*RSp463e + 1./162.*m55*me2*RSe465p*
         pow(RSp461e,2) + 1./81.*m55*me2*RSp465e*RSe463p*RS346pe + 1./
         81.*m55*me2*RSp465e*pow(RSe463p,2) + 1./81.*m55*me2*RSp465e*
         RSp463e*RSe463p - 1./81.*m55*me2*RSp465e*RS146pe*RSe463p - 1./
         81.*m55*me2*RSp465e*RSe461p*RS346pe - 2./81.*m55*me2*RSp465e*
         RSe461p*RSe463p - 1./81.*m55*me2*RSp465e*RSe461p*RSp463e + 1./
         81.*m55*me2*RSp465e*RSe461p*RS146pe + 1./81.*m55*me2*RSp465e*
         pow(RSe461p,2) );

      M13p6x23R +=  + Abeps(mp2,mu2)*pi16 * (  - 1./81.*m55*me2*RSp465e
         *RSp461e*RSe463p + 1./81.*m55*me2*RSp465e*RSp461e*RSe461p + 2./
         243.*m55*m66*pow(RSp463e,2) - 1./81.*m55*m66*RSp453e*RSp463e
          + 2./243.*m55*m66*pow(RSp453e,2) - 4./243.*m55*m66*RSp461e*
         RSp463e + 1./81.*m55*m66*RSp461e*RSp453e + 2./243.*m55*m66*
         pow(RSp461e,2) + 1./81.*m55*m66*RSp451e*RSp463e - 4./243.*m55*
         m66*RSp451e*RSp453e - 1./81.*m55*m66*RSp451e*RSp461e + 2./243.
         *m55*m66*pow(RSp451e,2) + 1./90.*pow(m55,2)*RSp463e - 1./162.*
         pow(m55,2)*RSp463e*RSp563e + 1./486.*pow(m55,2)*pow(RSp463e,2)
          - 1./162.*pow(m55,2)*RSp453e*RSp463e + 1./162.*pow(m55,2)*
         RS546pe*pow(RSp463e,2) + 1./162.*pow(m55,2)*RSp561e*RSp463e + 
         1./90.*pow(m55,2)*RSp461e + 1./162.*pow(m55,2)*RSp461e*RSp563e
          - 1./243.*pow(m55,2)*RSp461e*RSp463e + 1./162.*pow(m55,2)*
         RSp461e*RSp453e - 1./81.*pow(m55,2)*RSp461e*RS546pe*RSp463e - 
         1./162.*pow(m55,2)*RSp461e*RSp561e + 1./486.*pow(m55,2)*pow(
         RSp461e,2) );

      M13p6x23R +=  + Abeps(mp2,mu2)*pi16 * ( 1./162.*pow(m55,2)*pow(
         RSp461e,2)*RS546pe + 1./162.*pow(m55,2)*RSp451e*RSp463e - 1./
         162.*pow(m55,2)*RSp451e*RSp461e - 1./45.*pow(m55,2)*RSp465e - 
         4./135.*m44*mp2*RSp45613e + 7./270.*m44*mp2*RSp45633e - 1./45.
         *m44*mp2*RSp563e - 1./162.*m44*mp2*RSp563e*RS345pe - 1./162.*
         m44*mp2*RSp563e*RS346pe - 5./81.*m44*mp2*RSp563e*RS356pe - 5./
         81.*m44*mp2*RSp563e*RSe563p - 1./162.*m44*mp2*RSp563e*RSe463p
          - 1./162.*m44*mp2*RSp563e*RSe453p - 5./81.*m44*mp2*pow(
         RSp563e,2) - 1./81.*m44*mp2*RSp463e*RS346pe - 1./162.*m44*mp2*
         RSp463e*RS356pe - 1./162.*m44*mp2*RSp463e*RSe563p - 1./81.*m44
         *mp2*RSp463e*RSe463p - 1./81.*m44*mp2*RSp463e*RSp563e - 1./81.
         *m44*mp2*pow(RSp463e,2) - 1./81.*m44*mp2*RSp453e*RS345pe - 1./
         162.*m44*mp2*RSp453e*RS356pe - 1./162.*m44*mp2*RSp453e*RSe563p
          - 1./81.*m44*mp2*RSp453e*RSe453p - 1./81.*m44*mp2*RSp453e*
         RSp563e - 1./81.*m44*mp2*pow(RSp453e,2) + 7./270.*m44*mp2*
         RSp45611e );

      M13p6x23R +=  + Abeps(mp2,mu2)*pi16 * ( 1./162.*m44*mp2*RS145pe*
         RSp563e + 1./81.*m44*mp2*RS145pe*RSp453e + 1./162.*m44*mp2*
         RS146pe*RSp563e + 1./81.*m44*mp2*RS146pe*RSp463e + 5./81.*m44*
         mp2*RS156pe*RSp563e + 1./162.*m44*mp2*RS156pe*RSp463e + 1./162.
         *m44*mp2*RS156pe*RSp453e + 5./81.*m44*mp2*RSe561p*RSp563e + 1./
         162.*m44*mp2*RSe561p*RSp463e + 1./162.*m44*mp2*RSe561p*RSp453e
          + 1./162.*m44*mp2*RSe461p*RSp563e + 1./81.*m44*mp2*RSe461p*
         RSp463e + 1./162.*m44*mp2*RSe451p*RSp563e + 1./81.*m44*mp2*
         RSe451p*RSp453e - 1./45.*m44*mp2*RSp561e + 1./162.*m44*mp2*
         RSp561e*RS345pe + 1./162.*m44*mp2*RSp561e*RS346pe + 5./81.*m44
         *mp2*RSp561e*RS356pe + 5./81.*m44*mp2*RSp561e*RSe563p + 1./162.
         *m44*mp2*RSp561e*RSe463p + 1./162.*m44*mp2*RSp561e*RSe453p + 
         10./81.*m44*mp2*RSp561e*RSp563e + 1./81.*m44*mp2*RSp561e*
         RSp463e + 1./81.*m44*mp2*RSp561e*RSp453e - 1./162.*m44*mp2*
         RSp561e*RS145pe - 1./162.*m44*mp2*RSp561e*RS146pe - 5./81.*m44
         *mp2*RSp561e*RS156pe );

      M13p6x23R +=  + Abeps(mp2,mu2)*pi16 * (  - 5./81.*m44*mp2*RSp561e
         *RSe561p - 1./162.*m44*mp2*RSp561e*RSe461p - 1./162.*m44*mp2*
         RSp561e*RSe451p - 5./81.*m44*mp2*pow(RSp561e,2) + 1./81.*m44*
         mp2*RSp461e*RS346pe + 1./162.*m44*mp2*RSp461e*RS356pe + 1./162.
         *m44*mp2*RSp461e*RSe563p + 1./81.*m44*mp2*RSp461e*RSe463p + 1./
         81.*m44*mp2*RSp461e*RSp563e + 2./81.*m44*mp2*RSp461e*RSp463e
          - 1./81.*m44*mp2*RSp461e*RS146pe - 1./162.*m44*mp2*RSp461e*
         RS156pe - 1./162.*m44*mp2*RSp461e*RSe561p - 1./81.*m44*mp2*
         RSp461e*RSe461p - 1./81.*m44*mp2*RSp461e*RSp561e - 1./81.*m44*
         mp2*pow(RSp461e,2) + 1./81.*m44*mp2*RSp451e*RS345pe + 1./162.*
         m44*mp2*RSp451e*RS356pe + 1./162.*m44*mp2*RSp451e*RSe563p + 1./
         81.*m44*mp2*RSp451e*RSe453p + 1./81.*m44*mp2*RSp451e*RSp563e
          + 2./81.*m44*mp2*RSp451e*RSp453e - 1./81.*m44*mp2*RSp451e*
         RS145pe - 1./162.*m44*mp2*RSp451e*RS156pe - 1./162.*m44*mp2*
         RSp451e*RSe561p - 1./81.*m44*mp2*RSp451e*RSe451p - 1./81.*m44*
         mp2*RSp451e*RSp561e );

      M13p6x23R +=  + Abeps(mp2,mu2)*pi16 * (  - 1./81.*m44*mp2*pow(
         RSp451e,2) + 1./45.*m44*mp2*RSp564e + 1./81.*m44*mp2*RSp564e*
         RSp563e*RS356pe + 1./81.*m44*mp2*RSp564e*RSp563e*RSe563p + 1./
         54.*m44*mp2*RSp564e*pow(RSp563e,2) - 1./81.*m44*mp2*RSp564e*
         RS156pe*RSp563e - 1./81.*m44*mp2*RSp564e*RSe561p*RSp563e - 1./
         81.*m44*mp2*RSp564e*RSp561e*RS356pe - 1./81.*m44*mp2*RSp564e*
         RSp561e*RSe563p - 1./27.*m44*mp2*RSp564e*RSp561e*RSp563e + 1./
         81.*m44*mp2*RSp564e*RSp561e*RS156pe + 1./81.*m44*mp2*RSp564e*
         RSp561e*RSe561p + 1./54.*m44*mp2*RSp564e*pow(RSp561e,2) + 1./
         162.*m44*me2*RSe564p*pow(RSp563e,2) - 1./81.*m44*me2*RSe564p*
         RSp561e*RSp563e + 1./162.*m44*me2*RSe564p*pow(RSp561e,2) + 1./
         81.*m44*me2*RSp564e*RSe563p*RS356pe + 1./81.*m44*me2*RSp564e*
         pow(RSe563p,2) + 1./81.*m44*me2*RSp564e*RSp563e*RSe563p - 1./
         81.*m44*me2*RSp564e*RS156pe*RSe563p - 1./81.*m44*me2*RSp564e*
         RSe561p*RS356pe - 2./81.*m44*me2*RSp564e*RSe561p*RSe563p - 1./
         81.*m44*me2*RSp564e*RSe561p*RSp563e );

      M13p6x23R +=  + Abeps(mp2,mu2)*pi16 * ( 1./81.*m44*me2*RSp564e*
         RSe561p*RS156pe + 1./81.*m44*me2*RSp564e*pow(RSe561p,2) - 1./
         81.*m44*me2*RSp564e*RSp561e*RSe563p + 1./81.*m44*me2*RSp564e*
         RSp561e*RSe561p + 2./243.*m44*m66*pow(RSp563e,2) - 1./81.*m44*
         m66*RSp453e*RSp563e + 2./243.*m44*m66*pow(RSp453e,2) - 4./243.
         *m44*m66*RSp561e*RSp563e + 1./81.*m44*m66*RSp561e*RSp453e + 2./
         243.*m44*m66*pow(RSp561e,2) + 1./81.*m44*m66*RSp451e*RSp563e
          - 4./243.*m44*m66*RSp451e*RSp453e - 1./81.*m44*m66*RSp451e*
         RSp561e + 2./243.*m44*m66*pow(RSp451e,2) + 2./243.*m44*m55*
         pow(RSp563e,2) - 1./81.*m44*m55*RSp463e*RSp563e + 2./243.*m44*
         m55*pow(RSp463e,2) - 4./243.*m44*m55*RSp561e*RSp563e + 1./81.*
         m44*m55*RSp561e*RSp463e + 2./243.*m44*m55*pow(RSp561e,2) + 1./
         81.*m44*m55*RSp461e*RSp563e - 4./243.*m44*m55*RSp461e*RSp463e
          - 1./81.*m44*m55*RSp461e*RSp561e + 2./243.*m44*m55*pow(
         RSp461e,2) + 1./90.*pow(m44,2)*RSp563e + 1./486.*pow(m44,2)*
         pow(RSp563e,2) );

      M13p6x23R +=  + Abeps(mp2,mu2)*pi16 * (  - 1./162.*pow(m44,2)*
         RSp463e*RSp563e - 1./162.*pow(m44,2)*RSp453e*RSp563e + 1./162.
         *pow(m44,2)*RS456pe*pow(RSp563e,2) + 1./90.*pow(m44,2)*RSp561e
          - 1./243.*pow(m44,2)*RSp561e*RSp563e + 1./162.*pow(m44,2)*
         RSp561e*RSp463e + 1./162.*pow(m44,2)*RSp561e*RSp453e - 1./81.*
         pow(m44,2)*RSp561e*RS456pe*RSp563e + 1./486.*pow(m44,2)*pow(
         RSp561e,2) + 1./162.*pow(m44,2)*pow(RSp561e,2)*RS456pe + 1./
         162.*pow(m44,2)*RSp461e*RSp563e - 1./162.*pow(m44,2)*RSp461e*
         RSp561e + 1./162.*pow(m44,2)*RSp451e*RSp563e - 1./162.*pow(
         m44,2)*RSp451e*RSp561e - 1./45.*pow(m44,2)*RSp564e - 7./405.*
         m33*RD3456ep*RSp45613e + 1./81.*m33*RSp45633e*RD3456ep - 1./
         810.*m33*RD1456ep*RSp45613e - 23./1620.*m33*RD1456ep*RSp45633e
          - 11./810.*m33*RSp45611e*RD3456ep + 5./324.*m33*RSp45611e*
         RD1456ep + 29./360.*m33*mp2*RSp45613e + 7./270.*m33*mp2*
         RSp45613e*RS34561pe + 1./54.*m33*mp2*RSp45613e*RS14563pe + 1./
         54.*m33*mp2*RSp45613e*RSe45613p );

      M13p6x23R +=  + Abeps(mp2,mu2)*pi16 * ( 49./810.*m33*mp2*pow(
         RSp45613e,2) + 7./540.*m33*mp2*RS3456pee*RSp45613e + 7./540.*
         m33*mp2*RS3456ppe*RSp45613e - 1./108.*m33*mp2*RSe45633p*
         RSp45613e + 2./45.*m33*mp2*RSp45633e - 7./540.*m33*mp2*
         RSp45633e*RS34561pe - 1./108.*m33*mp2*RSp45633e*RS14563pe - 1./
         108.*m33*mp2*RSp45633e*RSe45613p - 11./810.*m33*mp2*RSp45633e*
         RSp45613e - 7./1080.*m33*mp2*RSp45633e*RS3456pee - 7./1080.*
         m33*mp2*RSp45633e*RS3456ppe + 1./216.*m33*mp2*RSp45633e*
         RSe45633p + 7./360.*m33*mp2*pow(RSp45633e,2) - 7./540.*m33*mp2
         *RS345pe*RSp45613e + 7./1080.*m33*mp2*RS345pe*RSp45633e - 7./
         540.*m33*mp2*RS346pe*RSp45613e + 7./1080.*m33*mp2*RS346pe*
         RSp45633e - 7./540.*m33*mp2*RS356pe*RSp45613e + 7./1080.*m33*
         mp2*RS356pe*RSp45633e + 7./540.*m33*mp2*RSp563e + 7./540.*m33*
         mp2*RSp463e + 7./540.*m33*mp2*RSp453e + 1./108.*m33*mp2*
         RS1456pee*RSp45613e - 1./216.*m33*mp2*RS1456pee*RSp45633e + 1./
         108.*m33*mp2*RS1456ppe*RSp45613e );

      M13p6x23R +=  + Abeps(mp2,mu2)*pi16 * (  - 1./216.*m33*mp2*
         RS1456ppe*RSp45633e - 1./108.*m33*mp2*RSe45611p*RSp45613e + 1./
         216.*m33*mp2*RSe45611p*RSp45633e + 1./30.*m33*mp2*RSp45611e - 
         7./540.*m33*mp2*RSp45611e*RS34561pe - 1./108.*m33*mp2*
         RSp45611e*RS14563pe - 1./108.*m33*mp2*RSp45611e*RSe45613p - 2./
         135.*m33*mp2*RSp45611e*RSp45613e - 7./1080.*m33*mp2*RSp45611e*
         RS3456pee - 7./1080.*m33*mp2*RSp45611e*RS3456ppe + 1./216.*m33
         *mp2*RSp45611e*RSe45633p - 13./810.*m33*mp2*RSp45611e*
         RSp45633e + 7./1080.*m33*mp2*RSp45611e*RS345pe + 7./1080.*m33*
         mp2*RSp45611e*RS346pe + 7./1080.*m33*mp2*RSp45611e*RS356pe - 1.
         /216.*m33*mp2*RSp45611e*RS1456pee - 1./216.*m33*mp2*RSp45611e*
         RS1456ppe + 1./216.*m33*mp2*RSp45611e*RSe45611p + 13./648.*m33
         *mp2*pow(RSp45611e,2) - 1./108.*m33*mp2*RS145pe*RSp45613e + 1./
         216.*m33*mp2*RS145pe*RSp45633e + 1./216.*m33*mp2*RS145pe*
         RSp45611e - 1./108.*m33*mp2*RS146pe*RSp45613e + 1./216.*m33*
         mp2*RS146pe*RSp45633e );

      M13p6x23R +=  + Abeps(mp2,mu2)*pi16 * ( 1./216.*m33*mp2*RS146pe*
         RSp45611e - 1./108.*m33*mp2*RS156pe*RSp45613e + 1./216.*m33*
         mp2*RS156pe*RSp45633e + 1./216.*m33*mp2*RS156pe*RSp45611e - 19.
         /540.*m33*mp2*RSp561e - 19./540.*m33*mp2*RSp461e - 19./540.*
         m33*mp2*RSp451e + 19./360.*m33*mp2*RSp564e + 19./360.*m33*mp2*
         RSp465e + 19./360.*m33*mp2*RSp456e + 17./405.*m33*me2*
         RSp45613e*RSe45613p - 1./45.*m33*me2*RSe45633p*RSp45613e + 11./
         405.*m33*me2*RSp45633e*RSe45613p + 2./135.*m33*me2*RSp45633e*
         RSe45633p - 1./810.*m33*me2*RSe45611p*RSp45613e - 23./1620.*
         m33*me2*RSe45611p*RSp45633e + 2./405.*m33*me2*RSp45611e*
         RSe45613p - 1./90.*m33*me2*RSp45611e*RSe45633p + 5./324.*m33*
         me2*RSp45611e*RSe45611p - 7./3240.*m33*m66*RSp45613e + 89./
         2160.*m33*m66*RSp45633e + 1./162.*m33*m66*pow(RSp563e,2) + 1./
         162.*m33*m66*pow(RSp463e,2) - 5./108.*m33*m66*RSp453e + 1./162.
         *m33*m66*RSp453e*RSp563e + 1./162.*m33*m66*RSp453e*RSp463e + 5.
         /162.*m33*m66*pow(RSp453e,2) );

      M13p6x23R +=  + Abeps(mp2,mu2)*pi16 * ( 31./1296.*m33*m66*
         RSp45611e + 1./81.*m33*m66*RSp561e*RSp563e + 1./162.*m33*m66*
         RSp561e*RSp453e + 1./81.*m33*m66*RSp461e*RSp463e + 1./162.*m33
         *m66*RSp461e*RSp453e + 7./540.*m33*m66*RSp451e + 1./162.*m33*
         m66*RSp451e*RSp563e + 1./162.*m33*m66*RSp451e*RSp463e + 5./81.
         *m33*m66*RSp451e*RSp453e + 1./135.*m33*m66*RSp456e + 1./162.*
         m33*m66*RSp456e*pow(RS345pe,2) - 1./162.*m33*m66*RSp456e*pow(
         RSe453p,2) - 1./81.*m33*m66*RSp456e*RSp453e*RSe453p - 1./162.*
         m33*m66*RSp456e*pow(RSp453e,2) - 2./81.*m33*m66*RSp456e*
         RS145pe*RS345pe - 1./81.*m33*m66*RSp456e*RS145pe*RSe453p - 1./
         81.*m33*m66*RSp456e*RS145pe*RSp453e - 2./81.*m33*m66*RSp456e*
         RSe451p*RS345pe - 1./81.*m33*m66*RSp456e*RSe451p*RSe453p - 1./
         81.*m33*m66*RSp456e*RSe451p*RSp453e - 2./81.*m33*m66*RSp456e*
         RSp451e*RS345pe - 1./81.*m33*m66*RSp456e*RSp451e*RSe453p - 1./
         81.*m33*m66*RSp456e*RSp451e*RSp453e - 7./3240.*m33*m55*
         RSp45613e );

      M13p6x23R +=  + Abeps(mp2,mu2)*pi16 * ( 89./2160.*m33*m55*
         RSp45633e + 1./162.*m33*m55*pow(RSp563e,2) - 5./108.*m33*m55*
         RSp463e + 1./162.*m33*m55*RSp463e*RSp563e + 5./162.*m33*m55*
         pow(RSp463e,2) + 1./162.*m33*m55*RSp453e*RSp463e + 1./162.*m33
         *m55*pow(RSp453e,2) + 31./1296.*m33*m55*RSp45611e + 1./81.*m33
         *m55*RSp561e*RSp563e + 1./162.*m33*m55*RSp561e*RSp463e + 7./
         540.*m33*m55*RSp461e + 1./162.*m33*m55*RSp461e*RSp563e + 5./81.
         *m33*m55*RSp461e*RSp463e + 1./162.*m33*m55*RSp461e*RSp453e + 1.
         /162.*m33*m55*RSp451e*RSp463e + 1./81.*m33*m55*RSp451e*RSp453e
          + 1./135.*m33*m55*RSp465e + 1./162.*m33*m55*RSp465e*pow(
         RS346pe,2) - 1./162.*m33*m55*RSp465e*pow(RSe463p,2) - 1./81.*
         m33*m55*RSp465e*RSp463e*RSe463p - 1./162.*m33*m55*RSp465e*pow(
         RSp463e,2) - 2./81.*m33*m55*RSp465e*RS146pe*RS346pe - 1./81.*
         m33*m55*RSp465e*RS146pe*RSe463p - 1./81.*m33*m55*RSp465e*
         RS146pe*RSp463e - 2./81.*m33*m55*RSp465e*RSe461p*RS346pe - 1./
         81.*m33*m55*RSp465e*RSe461p*RSe463p );

      M13p6x23R +=  + Abeps(mp2,mu2)*pi16 * (  - 1./81.*m33*m55*RSp465e
         *RSe461p*RSp463e - 2./81.*m33*m55*RSp465e*RSp461e*RS346pe - 1./
         81.*m33*m55*RSp465e*RSp461e*RSe463p - 1./81.*m33*m55*RSp465e*
         RSp461e*RSp463e - 7./3240.*m33*m44*RSp45613e + 89./2160.*m33*
         m44*RSp45633e - 5./108.*m33*m44*RSp563e + 5./162.*m33*m44*pow(
         RSp563e,2) + 1./162.*m33*m44*RSp463e*RSp563e + 1./162.*m33*m44
         *pow(RSp463e,2) + 1./162.*m33*m44*RSp453e*RSp563e + 1./162.*
         m33*m44*pow(RSp453e,2) + 31./1296.*m33*m44*RSp45611e + 7./540.
         *m33*m44*RSp561e + 5./81.*m33*m44*RSp561e*RSp563e + 1./162.*
         m33*m44*RSp561e*RSp463e + 1./162.*m33*m44*RSp561e*RSp453e + 1./
         162.*m33*m44*RSp461e*RSp563e + 1./81.*m33*m44*RSp461e*RSp463e
          + 1./162.*m33*m44*RSp451e*RSp563e + 1./81.*m33*m44*RSp451e*
         RSp453e + 1./135.*m33*m44*RSp564e + 1./162.*m33*m44*RSp564e*
         pow(RS356pe,2) - 1./162.*m33*m44*RSp564e*pow(RSe563p,2) - 1./
         81.*m33*m44*RSp564e*RSp563e*RSe563p - 1./162.*m33*m44*RSp564e*
         pow(RSp563e,2) );

      M13p6x23R +=  + Abeps(mp2,mu2)*pi16 * (  - 2./81.*m33*m44*RSp564e
         *RS156pe*RS356pe - 1./81.*m33*m44*RSp564e*RS156pe*RSe563p - 1./
         81.*m33*m44*RSp564e*RS156pe*RSp563e - 2./81.*m33*m44*RSp564e*
         RSe561p*RS356pe - 1./81.*m33*m44*RSp564e*RSe561p*RSe563p - 1./
         81.*m33*m44*RSp564e*RSe561p*RSp563e - 2./81.*m33*m44*RSp564e*
         RSp561e*RS356pe - 1./81.*m33*m44*RSp564e*RSp561e*RSe563p - 1./
         81.*m33*m44*RSp564e*RSp561e*RSp563e + 1./720.*pow(m33,2)*
         RSp45613e + 1./81.*pow(m33,2)*RSp45613e*RS34561pe - 2./81.*
         pow(m33,2)*RSp45613e*RS14563pe - 2./81.*pow(m33,2)*RSp45613e*
         RSe45613p - 2./81.*pow(m33,2)*pow(RSp45613e,2) + 7./1620.*pow(
         m33,2)*RS3456pee*RSp45613e + 7./1620.*pow(m33,2)*RS3456ppe*
         RSp45613e + 5./324.*pow(m33,2)*RSe45633p*RSp45613e + 37./1080.
         *pow(m33,2)*RSp45633e + 193./1620.*pow(m33,2)*RSp45633e*
         RS34561pe + 29./324.*pow(m33,2)*RSp45633e*RS14563pe + 29./324.
         *pow(m33,2)*RSp45633e*RSe45613p + 17./162.*pow(m33,2)*
         RSp45633e*RSp45613e );

      M13p6x23R +=  + Abeps(mp2,mu2)*pi16 * ( 1./3240.*pow(m33,2)*
         RSp45633e*RS3456pee + 1./3240.*pow(m33,2)*RSp45633e*RS3456ppe
          - 1./72.*pow(m33,2)*RSp45633e*RSe45633p - 1./72.*pow(m33,2)*
         pow(RSp45633e,2) - 7./1620.*pow(m33,2)*RS345pe*RSp45613e - 1./
         3240.*pow(m33,2)*RS345pe*RSp45633e - 7./1620.*pow(m33,2)*
         RS346pe*RSp45613e - 1./3240.*pow(m33,2)*RS346pe*RSp45633e - 7./
         1620.*pow(m33,2)*RS356pe*RSp45613e - 1./3240.*pow(m33,2)*
         RS356pe*RSp45633e - 49./2160.*pow(m33,2)*RSp563e - 49./2160.*
         pow(m33,2)*RSp463e - 49./2160.*pow(m33,2)*RSp453e - 1./108.*
         pow(m33,2)*RS1456pee*RSp45613e + 1./648.*pow(m33,2)*RS1456pee*
         RSp45633e - 1./108.*pow(m33,2)*RS1456ppe*RSp45613e + 1./648.*
         pow(m33,2)*RS1456ppe*RSp45633e + 1./108.*pow(m33,2)*RSe45611p*
         RSp45613e - 1./648.*pow(m33,2)*RSe45611p*RSp45633e + 23./432.*
         pow(m33,2)*RSp45611e + 1./60.*pow(m33,2)*RSp45611e*RS34561pe
          + 1./108.*pow(m33,2)*RSp45611e*RS14563pe + 1./108.*pow(m33,2)
         *RSp45611e*RSe45613p );

      M13p6x23R +=  + Abeps(mp2,mu2)*pi16 * ( 1./54.*pow(m33,2)*
         RSp45611e*RSp45613e + 1./72.*pow(m33,2)*RSp45611e*RS3456pee + 
         1./72.*pow(m33,2)*RSp45611e*RS3456ppe - 1./648.*pow(m33,2)*
         RSp45611e*RSe45633p - 1./324.*pow(m33,2)*RSp45611e*RSp45633e
          - 1./72.*pow(m33,2)*RSp45611e*RS345pe - 1./72.*pow(m33,2)*
         RSp45611e*RS346pe - 1./72.*pow(m33,2)*RSp45611e*RS356pe + 5./
         648.*pow(m33,2)*RSp45611e*RS1456pee + 5./648.*pow(m33,2)*
         RSp45611e*RS1456ppe - 5./648.*pow(m33,2)*RSp45611e*RSe45611p
          - 5./648.*pow(m33,2)*pow(RSp45611e,2) + 1./108.*pow(m33,2)*
         RS145pe*RSp45613e - 1./648.*pow(m33,2)*RS145pe*RSp45633e - 5./
         648.*pow(m33,2)*RS145pe*RSp45611e + 1./108.*pow(m33,2)*RS146pe
         *RSp45613e - 1./648.*pow(m33,2)*RS146pe*RSp45633e - 5./648.*
         pow(m33,2)*RS146pe*RSp45611e + 1./108.*pow(m33,2)*RS156pe*
         RSp45613e - 1./648.*pow(m33,2)*RS156pe*RSp45633e - 5./648.*
         pow(m33,2)*RS156pe*RSp45611e + 1./432.*pow(m33,2)*RSp561e + 1./
         432.*pow(m33,2)*RSp461e );

      M13p6x23R +=  + Abeps(mp2,mu2)*pi16 * ( 1./432.*pow(m33,2)*
         RSp451e - 1./810.*m11*RD3456ep*RSp45613e + 5./324.*m11*
         RSp45633e*RD3456ep - 7./405.*m11*RD1456ep*RSp45613e - 11./810.
         *m11*RD1456ep*RSp45633e - 23./1620.*m11*RSp45611e*RD3456ep + 1.
         /81.*m11*RSp45611e*RD1456ep + 29./360.*m11*mp2*RSp45613e + 1./
         54.*m11*mp2*RSp45613e*RS34561pe + 7./270.*m11*mp2*RSp45613e*
         RS14563pe + 1./54.*m11*mp2*RSp45613e*RSe45613p + 49./810.*m11*
         mp2*pow(RSp45613e,2) + 1./108.*m11*mp2*RS3456pee*RSp45613e + 1.
         /108.*m11*mp2*RS3456ppe*RSp45613e - 1./108.*m11*mp2*RSe45633p*
         RSp45613e + 1./30.*m11*mp2*RSp45633e - 1./108.*m11*mp2*
         RSp45633e*RS34561pe - 7./540.*m11*mp2*RSp45633e*RS14563pe - 1./
         108.*m11*mp2*RSp45633e*RSe45613p - 2./135.*m11*mp2*RSp45633e*
         RSp45613e - 1./216.*m11*mp2*RSp45633e*RS3456pee - 1./216.*m11*
         mp2*RSp45633e*RS3456ppe + 1./216.*m11*mp2*RSp45633e*RSe45633p
          + 13./648.*m11*mp2*pow(RSp45633e,2) - 1./108.*m11*mp2*RS345pe
         *RSp45613e );

      M13p6x23R +=  + Abeps(mp2,mu2)*pi16 * ( 1./216.*m11*mp2*RS345pe*
         RSp45633e - 1./108.*m11*mp2*RS346pe*RSp45613e + 1./216.*m11*
         mp2*RS346pe*RSp45633e - 1./108.*m11*mp2*RS356pe*RSp45613e + 1./
         216.*m11*mp2*RS356pe*RSp45633e - 19./540.*m11*mp2*RSp563e - 19.
         /540.*m11*mp2*RSp463e - 19./540.*m11*mp2*RSp453e + 7./540.*m11
         *mp2*RS1456pee*RSp45613e - 7./1080.*m11*mp2*RS1456pee*
         RSp45633e + 7./540.*m11*mp2*RS1456ppe*RSp45613e - 7./1080.*m11
         *mp2*RS1456ppe*RSp45633e - 1./108.*m11*mp2*RSe45611p*RSp45613e
          + 1./216.*m11*mp2*RSe45611p*RSp45633e + 2./45.*m11*mp2*
         RSp45611e - 1./108.*m11*mp2*RSp45611e*RS34561pe - 7./540.*m11*
         mp2*RSp45611e*RS14563pe - 1./108.*m11*mp2*RSp45611e*RSe45613p
          - 11./810.*m11*mp2*RSp45611e*RSp45613e - 1./216.*m11*mp2*
         RSp45611e*RS3456pee - 1./216.*m11*mp2*RSp45611e*RS3456ppe + 1./
         216.*m11*mp2*RSp45611e*RSe45633p - 13./810.*m11*mp2*RSp45611e*
         RSp45633e + 1./216.*m11*mp2*RSp45611e*RS345pe + 1./216.*m11*
         mp2*RSp45611e*RS346pe );

      M13p6x23R +=  + Abeps(mp2,mu2)*pi16 * ( 1./216.*m11*mp2*RSp45611e
         *RS356pe - 7./1080.*m11*mp2*RSp45611e*RS1456pee - 7./1080.*m11
         *mp2*RSp45611e*RS1456ppe + 1./216.*m11*mp2*RSp45611e*RSe45611p
          + 7./360.*m11*mp2*pow(RSp45611e,2) - 7./540.*m11*mp2*RS145pe*
         RSp45613e + 7./1080.*m11*mp2*RS145pe*RSp45633e + 7./1080.*m11*
         mp2*RS145pe*RSp45611e - 7./540.*m11*mp2*RS146pe*RSp45613e + 7./
         1080.*m11*mp2*RS146pe*RSp45633e + 7./1080.*m11*mp2*RS146pe*
         RSp45611e - 7./540.*m11*mp2*RS156pe*RSp45613e + 7./1080.*m11*
         mp2*RS156pe*RSp45633e + 7./1080.*m11*mp2*RS156pe*RSp45611e + 7.
         /540.*m11*mp2*RSp561e + 7./540.*m11*mp2*RSp461e + 7./540.*m11*
         mp2*RSp451e + 19./360.*m11*mp2*RSp564e + 19./360.*m11*mp2*
         RSp465e + 19./360.*m11*mp2*RSp456e + 17./405.*m11*me2*
         RSp45613e*RSe45613p - 1./810.*m11*me2*RSe45633p*RSp45613e + 2./
         405.*m11*me2*RSp45633e*RSe45613p + 5./324.*m11*me2*RSp45633e*
         RSe45633p - 1./45.*m11*me2*RSe45611p*RSp45613e - 1./90.*m11*
         me2*RSe45611p*RSp45633e );

      M13p6x23R +=  + Abeps(mp2,mu2)*pi16 * ( 11./405.*m11*me2*
         RSp45611e*RSe45613p - 23./1620.*m11*me2*RSp45611e*RSe45633p + 
         2./135.*m11*me2*RSp45611e*RSe45611p - 7./3240.*m11*m66*
         RSp45613e + 31./1296.*m11*m66*RSp45633e + 7./540.*m11*m66*
         RSp453e + 89./2160.*m11*m66*RSp45611e + 1./81.*m11*m66*RSp561e
         *RSp563e + 1./162.*m11*m66*RSp561e*RSp453e + 1./162.*m11*m66*
         pow(RSp561e,2) + 1./81.*m11*m66*RSp461e*RSp463e + 1./162.*m11*
         m66*RSp461e*RSp453e + 1./162.*m11*m66*pow(RSp461e,2) - 5./108.
         *m11*m66*RSp451e + 1./162.*m11*m66*RSp451e*RSp563e + 1./162.*
         m11*m66*RSp451e*RSp463e + 5./81.*m11*m66*RSp451e*RSp453e + 1./
         162.*m11*m66*RSp451e*RSp561e + 1./162.*m11*m66*RSp451e*RSp461e
          + 5./162.*m11*m66*pow(RSp451e,2) + 1./135.*m11*m66*RSp456e - 
         2./81.*m11*m66*RSp456e*RS145pe*RS345pe - 2./81.*m11*m66*
         RSp456e*RS145pe*RSe453p - 2./81.*m11*m66*RSp456e*RS145pe*
         RSp453e + 1./162.*m11*m66*RSp456e*pow(RS145pe,2) - 1./81.*m11*
         m66*RSp456e*RSe451p*RS345pe );

      M13p6x23R +=  + Abeps(mp2,mu2)*pi16 * (  - 1./81.*m11*m66*RSp456e
         *RSe451p*RSe453p - 1./81.*m11*m66*RSp456e*RSe451p*RSp453e - 1./
         162.*m11*m66*RSp456e*pow(RSe451p,2) - 1./81.*m11*m66*RSp456e*
         RSp451e*RS345pe - 1./81.*m11*m66*RSp456e*RSp451e*RSe453p - 1./
         81.*m11*m66*RSp456e*RSp451e*RSp453e - 1./81.*m11*m66*RSp456e*
         RSp451e*RSe451p - 1./162.*m11*m66*RSp456e*pow(RSp451e,2) - 7./
         3240.*m11*m55*RSp45613e + 31./1296.*m11*m55*RSp45633e + 7./540.
         *m11*m55*RSp463e + 89./2160.*m11*m55*RSp45611e + 1./81.*m11*
         m55*RSp561e*RSp563e + 1./162.*m11*m55*RSp561e*RSp463e + 1./162.
         *m11*m55*pow(RSp561e,2) - 5./108.*m11*m55*RSp461e + 1./162.*
         m11*m55*RSp461e*RSp563e + 5./81.*m11*m55*RSp461e*RSp463e + 1./
         162.*m11*m55*RSp461e*RSp453e + 1./162.*m11*m55*RSp461e*RSp561e
          + 5./162.*m11*m55*pow(RSp461e,2) + 1./162.*m11*m55*RSp451e*
         RSp463e + 1./81.*m11*m55*RSp451e*RSp453e + 1./162.*m11*m55*
         RSp451e*RSp461e + 1./162.*m11*m55*pow(RSp451e,2) + 1./135.*m11
         *m55*RSp465e );

      M13p6x23R +=  + Abeps(mp2,mu2)*pi16 * (  - 2./81.*m11*m55*RSp465e
         *RS146pe*RS346pe - 2./81.*m11*m55*RSp465e*RS146pe*RSe463p - 2./
         81.*m11*m55*RSp465e*RS146pe*RSp463e + 1./162.*m11*m55*RSp465e*
         pow(RS146pe,2) - 1./81.*m11*m55*RSp465e*RSe461p*RS346pe - 1./
         81.*m11*m55*RSp465e*RSe461p*RSe463p - 1./81.*m11*m55*RSp465e*
         RSe461p*RSp463e - 1./162.*m11*m55*RSp465e*pow(RSe461p,2) - 1./
         81.*m11*m55*RSp465e*RSp461e*RS346pe - 1./81.*m11*m55*RSp465e*
         RSp461e*RSe463p - 1./81.*m11*m55*RSp465e*RSp461e*RSp463e - 1./
         81.*m11*m55*RSp465e*RSp461e*RSe461p - 1./162.*m11*m55*RSp465e*
         pow(RSp461e,2) - 7./3240.*m11*m44*RSp45613e + 31./1296.*m11*
         m44*RSp45633e + 7./540.*m11*m44*RSp563e + 89./2160.*m11*m44*
         RSp45611e - 5./108.*m11*m44*RSp561e + 5./81.*m11*m44*RSp561e*
         RSp563e + 1./162.*m11*m44*RSp561e*RSp463e + 1./162.*m11*m44*
         RSp561e*RSp453e + 5./162.*m11*m44*pow(RSp561e,2) + 1./162.*m11
         *m44*RSp461e*RSp563e + 1./81.*m11*m44*RSp461e*RSp463e + 1./162.
         *m11*m44*RSp461e*RSp561e );

      M13p6x23R +=  + Abeps(mp2,mu2)*pi16 * ( 1./162.*m11*m44*pow(
         RSp461e,2) + 1./162.*m11*m44*RSp451e*RSp563e + 1./81.*m11*m44*
         RSp451e*RSp453e + 1./162.*m11*m44*RSp451e*RSp561e + 1./162.*
         m11*m44*pow(RSp451e,2) + 1./135.*m11*m44*RSp564e - 2./81.*m11*
         m44*RSp564e*RS156pe*RS356pe - 2./81.*m11*m44*RSp564e*RS156pe*
         RSe563p - 2./81.*m11*m44*RSp564e*RS156pe*RSp563e + 1./162.*m11
         *m44*RSp564e*pow(RS156pe,2) - 1./81.*m11*m44*RSp564e*RSe561p*
         RS356pe - 1./81.*m11*m44*RSp564e*RSe561p*RSe563p - 1./81.*m11*
         m44*RSp564e*RSe561p*RSp563e - 1./162.*m11*m44*RSp564e*pow(
         RSe561p,2) - 1./81.*m11*m44*RSp564e*RSp561e*RS356pe - 1./81.*
         m11*m44*RSp564e*RSp561e*RSe563p - 1./81.*m11*m44*RSp564e*
         RSp561e*RSp563e - 1./81.*m11*m44*RSp564e*RSp561e*RSe561p - 1./
         162.*m11*m44*RSp564e*pow(RSp561e,2) + 163./1080.*m11*m33*
         RSp45613e - 1./135.*m11*m33*RSp45613e*RS34561pe - 1./135.*m11*
         m33*RSp45613e*RS14563pe - 4./81.*m11*m33*RSp45613e*RSe45613p
          - 4./81.*m11*m33*pow(RSp45613e,2) );

      M13p6x23R +=  + Abeps(mp2,mu2)*pi16 * (  - 19./810.*m11*m33*
         RS3456pee*RSp45613e - 19./810.*m11*m33*RS3456ppe*RSp45613e + 2.
         /81.*m11*m33*RSe45633p*RSp45613e + 53./2160.*m11*m33*RSp45633e
          + 1./15.*m11*m33*RSp45633e*RS34561pe + 4./45.*m11*m33*
         RSp45633e*RS14563pe + 5./81.*m11*m33*RSp45633e*RSe45613p + 7./
         81.*m11*m33*RSp45633e*RSp45613e + 1./162.*m11*m33*RSp45633e*
         RS3456pee + 1./162.*m11*m33*RSp45633e*RS3456ppe - 7./324.*m11*
         m33*RSp45633e*RSe45633p - 7./324.*m11*m33*pow(RSp45633e,2) + 
         19./810.*m11*m33*RS345pe*RSp45613e - 1./162.*m11*m33*RS345pe*
         RSp45633e + 19./810.*m11*m33*RS346pe*RSp45613e - 1./162.*m11*
         m33*RS346pe*RSp45633e + 19./810.*m11*m33*RS356pe*RSp45613e - 1.
         /162.*m11*m33*RS356pe*RSp45633e - 13./540.*m11*m33*RSp563e - 
         13./540.*m11*m33*RSp463e - 13./540.*m11*m33*RSp453e - 19./810.
         *m11*m33*RS1456pee*RSp45613e + 7./405.*m11*m33*RS1456pee*
         RSp45633e - 19./810.*m11*m33*RS1456ppe*RSp45613e + 7./405.*m11
         *m33*RS1456ppe*RSp45633e );

      M13p6x23R +=  + Abeps(mp2,mu2)*pi16 * ( 2./81.*m11*m33*RSe45611p*
         RSp45613e - 1./324.*m11*m33*RSe45611p*RSp45633e + 53./2160.*
         m11*m33*RSp45611e + 4./45.*m11*m33*RSp45611e*RS34561pe + 1./15.
         *m11*m33*RSp45611e*RS14563pe + 5./81.*m11*m33*RSp45611e*
         RSe45613p + 7./81.*m11*m33*RSp45611e*RSp45613e + 7./405.*m11*
         m33*RSp45611e*RS3456pee + 7./405.*m11*m33*RSp45611e*RS3456ppe
          - 1./324.*m11*m33*RSp45611e*RSe45633p - 1./162.*m11*m33*
         RSp45611e*RSp45633e - 7./405.*m11*m33*RSp45611e*RS345pe - 7./
         405.*m11*m33*RSp45611e*RS346pe - 7./405.*m11*m33*RSp45611e*
         RS356pe + 1./162.*m11*m33*RSp45611e*RS1456pee + 1./162.*m11*
         m33*RSp45611e*RS1456ppe - 7./324.*m11*m33*RSp45611e*RSe45611p
          - 7./324.*m11*m33*pow(RSp45611e,2) + 19./810.*m11*m33*RS145pe
         *RSp45613e - 7./405.*m11*m33*RS145pe*RSp45633e - 1./162.*m11*
         m33*RS145pe*RSp45611e + 19./810.*m11*m33*RS146pe*RSp45613e - 7.
         /405.*m11*m33*RS146pe*RSp45633e - 1./162.*m11*m33*RS146pe*
         RSp45611e );

      M13p6x23R +=  + Abeps(mp2,mu2)*pi16 * ( 19./810.*m11*m33*RS156pe*
         RSp45613e - 7./405.*m11*m33*RS156pe*RSp45633e - 1./162.*m11*
         m33*RS156pe*RSp45611e - 13./540.*m11*m33*RSp561e - 13./540.*
         m11*m33*RSp461e - 13./540.*m11*m33*RSp451e + 1./720.*pow(
         m11,2)*RSp45613e - 2./81.*pow(m11,2)*RSp45613e*RS34561pe + 1./
         81.*pow(m11,2)*RSp45613e*RS14563pe - 2./81.*pow(m11,2)*
         RSp45613e*RSe45613p - 2./81.*pow(m11,2)*pow(RSp45613e,2) - 1./
         108.*pow(m11,2)*RS3456pee*RSp45613e - 1./108.*pow(m11,2)*
         RS3456ppe*RSp45613e + 1./108.*pow(m11,2)*RSe45633p*RSp45613e
          + 23./432.*pow(m11,2)*RSp45633e + 1./108.*pow(m11,2)*
         RSp45633e*RS34561pe + 1./60.*pow(m11,2)*RSp45633e*RS14563pe + 
         1./108.*pow(m11,2)*RSp45633e*RSe45613p + 1./54.*pow(m11,2)*
         RSp45633e*RSp45613e + 5./648.*pow(m11,2)*RSp45633e*RS3456pee
          + 5./648.*pow(m11,2)*RSp45633e*RS3456ppe - 5./648.*pow(m11,2)
         *RSp45633e*RSe45633p - 5./648.*pow(m11,2)*pow(RSp45633e,2) + 1.
         /108.*pow(m11,2)*RS345pe*RSp45613e );

      M13p6x23R +=  + Abeps(mp2,mu2)*pi16 * (  - 5./648.*pow(m11,2)*
         RS345pe*RSp45633e + 1./108.*pow(m11,2)*RS346pe*RSp45613e - 5./
         648.*pow(m11,2)*RS346pe*RSp45633e + 1./108.*pow(m11,2)*RS356pe
         *RSp45613e - 5./648.*pow(m11,2)*RS356pe*RSp45633e + 1./432.*
         pow(m11,2)*RSp563e + 1./432.*pow(m11,2)*RSp463e + 1./432.*pow(
         m11,2)*RSp453e + 7./1620.*pow(m11,2)*RS1456pee*RSp45613e + 1./
         72.*pow(m11,2)*RS1456pee*RSp45633e + 7./1620.*pow(m11,2)*
         RS1456ppe*RSp45613e + 1./72.*pow(m11,2)*RS1456ppe*RSp45633e + 
         5./324.*pow(m11,2)*RSe45611p*RSp45613e - 1./648.*pow(m11,2)*
         RSe45611p*RSp45633e + 37./1080.*pow(m11,2)*RSp45611e + 29./324.
         *pow(m11,2)*RSp45611e*RS34561pe + 193./1620.*pow(m11,2)*
         RSp45611e*RS14563pe + 29./324.*pow(m11,2)*RSp45611e*RSe45613p
          + 17./162.*pow(m11,2)*RSp45611e*RSp45613e + 1./648.*pow(
         m11,2)*RSp45611e*RS3456pee + 1./648.*pow(m11,2)*RSp45611e*
         RS3456ppe - 1./648.*pow(m11,2)*RSp45611e*RSe45633p - 1./324.*
         pow(m11,2)*RSp45611e*RSp45633e );

      M13p6x23R +=  + Abeps(mp2,mu2)*pi16 * (  - 1./648.*pow(m11,2)*
         RSp45611e*RS345pe - 1./648.*pow(m11,2)*RSp45611e*RS346pe - 1./
         648.*pow(m11,2)*RSp45611e*RS356pe + 1./3240.*pow(m11,2)*
         RSp45611e*RS1456pee + 1./3240.*pow(m11,2)*RSp45611e*RS1456ppe
          - 1./72.*pow(m11,2)*RSp45611e*RSe45611p - 1./72.*pow(m11,2)*
         pow(RSp45611e,2) - 7./1620.*pow(m11,2)*RS145pe*RSp45613e - 1./
         72.*pow(m11,2)*RS145pe*RSp45633e - 1./3240.*pow(m11,2)*RS145pe
         *RSp45611e - 7./1620.*pow(m11,2)*RS146pe*RSp45613e - 1./72.*
         pow(m11,2)*RS146pe*RSp45633e - 1./3240.*pow(m11,2)*RS146pe*
         RSp45611e - 7./1620.*pow(m11,2)*RS156pe*RSp45613e - 1./72.*
         pow(m11,2)*RS156pe*RSp45633e - 1./3240.*pow(m11,2)*RS156pe*
         RSp45611e - 49./2160.*pow(m11,2)*RSp561e - 49./2160.*pow(
         m11,2)*RSp461e - 49./2160.*pow(m11,2)*RSp451e );

      M13p6x23R +=  + Bbeps(m11,mu2)*pi16 * ( 1./405.*pow(mp2,2)*
         RD1456ep*RSp45613e - 1./810.*pow(mp2,2)*RD1456ep*RSp45633e - 1.
         /810.*pow(mp2,2)*RSp45611e*RD1456ep + 1./405.*pow(me2,2)*
         RD1456ep*RSe45613p - 1./810.*pow(me2,2)*RD1456ep*RSe45633p - 1.
         /810.*pow(me2,2)*RSe45611p*RD1456ep - 11./810.*m33*RD1456ep*
         RD3456ep + 5./324.*m33*pow(RD1456ep,2) + 2./405.*m33*mp2*
         RD1456ep*RSp45613e - 1./90.*m33*mp2*RD1456ep*RSp45633e + 5./
         324.*m33*mp2*RSp45611e*RD1456ep + 2./405.*m33*me2*RD1456ep*
         RSe45613p - 1./90.*m33*me2*RD1456ep*RSe45633p + 5./324.*m33*
         me2*RSe45611p*RD1456ep + 31./1296.*m33*m66*RD1456ep + 31./1296.
         *m33*m55*RD1456ep + 31./1296.*m33*m44*RD1456ep + 23./432.*pow(
         m33,2)*RD1456ep + 1./60.*pow(m33,2)*RD1456ep*RS34561pe + 1./
         108.*pow(m33,2)*RD1456ep*RS14563pe + 1./108.*pow(m33,2)*
         RD1456ep*RSe45613p + 1./108.*pow(m33,2)*RD1456ep*RSp45613e + 1.
         /72.*pow(m33,2)*RD1456ep*RS3456pee + 1./72.*pow(m33,2)*
         RD1456ep*RS3456ppe );

      M13p6x23R +=  + Bbeps(m11,mu2)*pi16 * (  - 1./648.*pow(m33,2)*
         RD1456ep*RSe45633p - 1./648.*pow(m33,2)*RD1456ep*RSp45633e - 1.
         /72.*pow(m33,2)*RD1456ep*RS345pe - 1./72.*pow(m33,2)*RD1456ep*
         RS346pe - 1./72.*pow(m33,2)*RD1456ep*RS356pe + 5./648.*pow(
         m33,2)*RS1456pee*RD1456ep + 5./648.*pow(m33,2)*RS1456ppe*
         RD1456ep - 5./648.*pow(m33,2)*RSe45611p*RD1456ep - 5./648.*
         pow(m33,2)*RSp45611e*RD1456ep - 5./648.*pow(m33,2)*RS145pe*
         RD1456ep - 5./648.*pow(m33,2)*RS146pe*RD1456ep - 5./648.*pow(
         m33,2)*RS156pe*RD1456ep - 1./81.*m11*RD1456ep*RD3456ep - 37./
         1620.*m11*pow(RD1456ep,2) + 19./810.*m11*mp2*RD1456ep*
         RSp45613e - 1./81.*m11*mp2*RD1456ep*RSp45633e - 11./540.*m11*
         mp2*RSp45611e*RD1456ep + 19./810.*m11*me2*RD1456ep*RSe45613p
          - 1./81.*m11*me2*RD1456ep*RSe45633p - 11./540.*m11*me2*
         RSe45611p*RD1456ep + 23./1296.*m11*m66*RD1456ep + 23./1296.*
         m11*m55*RD1456ep + 23./1296.*m11*m44*RD1456ep + 25./432.*m11*
         m33*RD1456ep );

      M13p6x23R +=  + Bbeps(m11,mu2)*pi16 * ( 41./540.*m11*m33*RD1456ep
         *RS34561pe + 17./180.*m11*m33*RD1456ep*RS14563pe + 17./324.*
         m11*m33*RD1456ep*RSe45613p + 17./324.*m11*m33*RD1456ep*
         RSp45613e + 7./648.*m11*m33*RD1456ep*RS3456pee + 7./648.*m11*
         m33*RD1456ep*RS3456ppe + 1./648.*m11*m33*RD1456ep*RSe45633p + 
         1./648.*m11*m33*RD1456ep*RSp45633e - 7./648.*m11*m33*RD1456ep*
         RS345pe - 7./648.*m11*m33*RD1456ep*RS346pe - 7./648.*m11*m33*
         RD1456ep*RS356pe + 1./648.*m11*m33*RS1456pee*RD1456ep + 1./648.
         *m11*m33*RS1456ppe*RD1456ep - 11./648.*m11*m33*RSe45611p*
         RD1456ep - 11./648.*m11*m33*RSp45611e*RD1456ep - 1./648.*m11*
         m33*RS145pe*RD1456ep - 1./648.*m11*m33*RS146pe*RD1456ep - 1./
         648.*m11*m33*RS156pe*RD1456ep + 1./27.*m11*m33*mp2*RSp45611e*
         RS14563pe + 1./27.*m11*m33*me2*RSe45611p*RS14563pe + 1./81.*
         m11*m33*m66*RS14563pe + 1./81.*m11*m33*m55*RS14563pe + 1./81.*
         m11*m33*m44*RS14563pe + 1./24.*pow(m11,2)*RD1456ep + 13./162.*
         pow(m11,2)*RD1456ep*RS34561pe );

      M13p6x23R +=  + Bbeps(m11,mu2)*pi16 * ( 73./405.*pow(m11,2)*
         RD1456ep*RS14563pe + 13./162.*pow(m11,2)*RD1456ep*RSe45613p + 
         13./162.*pow(m11,2)*RD1456ep*RSp45613e - 1./324.*pow(m11,2)*
         RD1456ep*RS3456pee - 1./324.*pow(m11,2)*RD1456ep*RS3456ppe + 1.
         /324.*pow(m11,2)*RD1456ep*RSe45633p + 1./324.*pow(m11,2)*
         RD1456ep*RSp45633e + 1./324.*pow(m11,2)*RD1456ep*RS345pe + 1./
         324.*pow(m11,2)*RD1456ep*RS346pe + 1./324.*pow(m11,2)*RD1456ep
         *RS356pe + 5./162.*pow(m11,2)*RS1456pee*RD1456ep + 5./162.*
         pow(m11,2)*RS1456ppe*RD1456ep - 1./108.*pow(m11,2)*RSe45611p*
         RD1456ep - 1./108.*pow(m11,2)*RSp45611e*RD1456ep - 5./162.*
         pow(m11,2)*RS145pe*RD1456ep - 5./162.*pow(m11,2)*RS146pe*
         RD1456ep - 5./162.*pow(m11,2)*RS156pe*RD1456ep + 2./27.*pow(
         m11,2)*mp2*RSp45611e*RS14563pe + 2./27.*pow(m11,2)*me2*
         RSe45611p*RS14563pe + 2./81.*pow(m11,2)*m66*RS14563pe + 2./81.
         *pow(m11,2)*m55*RS14563pe + 2./81.*pow(m11,2)*m44*RS14563pe - 
         2./27.*pow(m11,2)*m33*RS14563pe );

      M13p6x23R +=  + Bbeps(m11,mu2)*pi16 * (  - 1./27.*pow(m11,2)*m33*
         RS1456pee*RS14563pe - 1./27.*pow(m11,2)*m33*RS1456ppe*
         RS14563pe + 1./27.*pow(m11,2)*m33*RS145pe*RS14563pe + 1./27.*
         pow(m11,2)*m33*RS146pe*RS14563pe + 1./27.*pow(m11,2)*m33*
         RS156pe*RS14563pe - 4./27.*pow(m11,3)*RS14563pe - 2./27.*pow(
         m11,3)*RS1456pee*RS14563pe - 2./27.*pow(m11,3)*RS1456ppe*
         RS14563pe + 2./27.*pow(m11,3)*RS145pe*RS14563pe + 2./27.*pow(
         m11,3)*RS146pe*RS14563pe + 2./27.*pow(m11,3)*RS156pe*RS14563pe
          );

      M13p6x23R +=  + Bbeps(m33,mu2)*pi16 * ( 1./405.*pow(mp2,2)*
         RD3456ep*RSp45613e - 1./810.*pow(mp2,2)*RSp45633e*RD3456ep - 1.
         /810.*pow(mp2,2)*RSp45611e*RD3456ep + 1./405.*pow(me2,2)*
         RD3456ep*RSe45613p - 1./810.*pow(me2,2)*RSe45633p*RD3456ep - 1.
         /810.*pow(me2,2)*RSe45611p*RD3456ep - 37./1620.*m33*pow(
         RD3456ep,2) - 1./81.*m33*RD1456ep*RD3456ep + 19./810.*m33*mp2*
         RD3456ep*RSp45613e - 11./540.*m33*mp2*RSp45633e*RD3456ep - 1./
         81.*m33*mp2*RSp45611e*RD3456ep + 19./810.*m33*me2*RD3456ep*
         RSe45613p - 11./540.*m33*me2*RSe45633p*RD3456ep - 1./81.*m33*
         me2*RSe45611p*RD3456ep + 23./1296.*m33*m66*RD3456ep + 23./1296.
         *m33*m55*RD3456ep + 23./1296.*m33*m44*RD3456ep + 1./24.*pow(
         m33,2)*RD3456ep + 73./405.*pow(m33,2)*RD3456ep*RS34561pe + 13./
         162.*pow(m33,2)*RD3456ep*RS14563pe + 13./162.*pow(m33,2)*
         RD3456ep*RSe45613p + 13./162.*pow(m33,2)*RD3456ep*RSp45613e + 
         5./162.*pow(m33,2)*RS3456pee*RD3456ep + 5./162.*pow(m33,2)*
         RS3456ppe*RD3456ep );

      M13p6x23R +=  + Bbeps(m33,mu2)*pi16 * (  - 1./108.*pow(m33,2)*
         RSe45633p*RD3456ep - 1./108.*pow(m33,2)*RSp45633e*RD3456ep - 5.
         /162.*pow(m33,2)*RS345pe*RD3456ep - 5./162.*pow(m33,2)*RS346pe
         *RD3456ep - 5./162.*pow(m33,2)*RS356pe*RD3456ep - 1./324.*pow(
         m33,2)*RS1456pee*RD3456ep - 1./324.*pow(m33,2)*RS1456ppe*
         RD3456ep + 1./324.*pow(m33,2)*RSe45611p*RD3456ep + 1./324.*
         pow(m33,2)*RSp45611e*RD3456ep + 1./324.*pow(m33,2)*RS145pe*
         RD3456ep + 1./324.*pow(m33,2)*RS146pe*RD3456ep + 1./324.*pow(
         m33,2)*RS156pe*RD3456ep + 2./27.*pow(m33,2)*mp2*RSp45633e*
         RS34561pe + 2./27.*pow(m33,2)*me2*RSe45633p*RS34561pe + 2./81.
         *pow(m33,2)*m66*RS34561pe + 2./81.*pow(m33,2)*m55*RS34561pe + 
         2./81.*pow(m33,2)*m44*RS34561pe - 4./27.*pow(m33,3)*RS34561pe
          - 2./27.*pow(m33,3)*RS3456pee*RS34561pe - 2./27.*pow(m33,3)*
         RS3456ppe*RS34561pe + 2./27.*pow(m33,3)*RS345pe*RS34561pe + 2./
         27.*pow(m33,3)*RS346pe*RS34561pe + 2./27.*pow(m33,3)*RS356pe*
         RS34561pe );

      M13p6x23R +=  + Bbeps(m33,mu2)*pi16 * ( 5./324.*m11*pow(
         RD3456ep,2) - 11./810.*m11*RD1456ep*RD3456ep + 2./405.*m11*mp2
         *RD3456ep*RSp45613e + 5./324.*m11*mp2*RSp45633e*RD3456ep - 1./
         90.*m11*mp2*RSp45611e*RD3456ep + 2./405.*m11*me2*RD3456ep*
         RSe45613p + 5./324.*m11*me2*RSe45633p*RD3456ep - 1./90.*m11*
         me2*RSe45611p*RD3456ep + 31./1296.*m11*m66*RD3456ep + 31./1296.
         *m11*m55*RD3456ep + 31./1296.*m11*m44*RD3456ep + 25./432.*m11*
         m33*RD3456ep + 17./180.*m11*m33*RD3456ep*RS34561pe + 41./540.*
         m11*m33*RD3456ep*RS14563pe + 17./324.*m11*m33*RD3456ep*
         RSe45613p + 17./324.*m11*m33*RD3456ep*RSp45613e + 1./648.*m11*
         m33*RS3456pee*RD3456ep + 1./648.*m11*m33*RS3456ppe*RD3456ep - 
         11./648.*m11*m33*RSe45633p*RD3456ep - 11./648.*m11*m33*
         RSp45633e*RD3456ep - 1./648.*m11*m33*RS345pe*RD3456ep - 1./648.
         *m11*m33*RS346pe*RD3456ep - 1./648.*m11*m33*RS356pe*RD3456ep
          + 7./648.*m11*m33*RS1456pee*RD3456ep + 7./648.*m11*m33*
         RS1456ppe*RD3456ep );

      M13p6x23R +=  + Bbeps(m33,mu2)*pi16 * ( 1./648.*m11*m33*RSe45611p
         *RD3456ep + 1./648.*m11*m33*RSp45611e*RD3456ep - 7./648.*m11*
         m33*RS145pe*RD3456ep - 7./648.*m11*m33*RS146pe*RD3456ep - 7./
         648.*m11*m33*RS156pe*RD3456ep + 1./27.*m11*m33*mp2*RSp45633e*
         RS34561pe + 1./27.*m11*m33*me2*RSe45633p*RS34561pe + 1./81.*
         m11*m33*m66*RS34561pe + 1./81.*m11*m33*m55*RS34561pe + 1./81.*
         m11*m33*m44*RS34561pe - 2./27.*m11*pow(m33,2)*RS34561pe - 1./
         27.*m11*pow(m33,2)*RS3456pee*RS34561pe - 1./27.*m11*pow(m33,2)
         *RS3456ppe*RS34561pe + 1./27.*m11*pow(m33,2)*RS345pe*RS34561pe
          + 1./27.*m11*pow(m33,2)*RS346pe*RS34561pe + 1./27.*m11*pow(
         m33,2)*RS356pe*RS34561pe + 23./432.*pow(m11,2)*RD3456ep + 1./
         108.*pow(m11,2)*RD3456ep*RS34561pe + 1./60.*pow(m11,2)*
         RD3456ep*RS14563pe + 1./108.*pow(m11,2)*RD3456ep*RSe45613p + 1.
         /108.*pow(m11,2)*RD3456ep*RSp45613e + 5./648.*pow(m11,2)*
         RS3456pee*RD3456ep + 5./648.*pow(m11,2)*RS3456ppe*RD3456ep - 5.
         /648.*pow(m11,2)*RSe45633p*RD3456ep );

      M13p6x23R +=  + Bbeps(m33,mu2)*pi16 * (  - 5./648.*pow(m11,2)*
         RSp45633e*RD3456ep - 5./648.*pow(m11,2)*RS345pe*RD3456ep - 5./
         648.*pow(m11,2)*RS346pe*RD3456ep - 5./648.*pow(m11,2)*RS356pe*
         RD3456ep + 1./72.*pow(m11,2)*RS1456pee*RD3456ep + 1./72.*pow(
         m11,2)*RS1456ppe*RD3456ep - 1./648.*pow(m11,2)*RSe45611p*
         RD3456ep - 1./648.*pow(m11,2)*RSp45611e*RD3456ep - 1./72.*pow(
         m11,2)*RS145pe*RD3456ep - 1./72.*pow(m11,2)*RS146pe*RD3456ep
          - 1./72.*pow(m11,2)*RS156pe*RD3456ep );

      M13p6x23R +=  + Bbeps(me2,mu2)*pi16 * ( 1./162.*m66*me2*mp2*
         RSp456e*pow(RSe453p,2) - 1./81.*m66*me2*mp2*RSp456e*RSe451p*
         RSe453p + 1./162.*m66*me2*mp2*RSp456e*pow(RSe451p,2) - 1./162.
         *m66*pow(me2,2)*pow(RSe563p,2) - 1./162.*m66*pow(me2,2)*pow(
         RSe463p,2) - 1./162.*m66*pow(me2,2)*RSe453p*RSe563p - 1./162.*
         m66*pow(me2,2)*RSe453p*RSe463p - 5./162.*m66*pow(me2,2)*pow(
         RSe453p,2) + 1./81.*m66*pow(me2,2)*RSe561p*RSe563p + 1./162.*
         m66*pow(me2,2)*RSe561p*RSe453p - 1./162.*m66*pow(me2,2)*pow(
         RSe561p,2) + 1./81.*m66*pow(me2,2)*RSe461p*RSe463p + 1./162.*
         m66*pow(me2,2)*RSe461p*RSe453p - 1./162.*m66*pow(me2,2)*pow(
         RSe461p,2) + 1./162.*m66*pow(me2,2)*RSe451p*RSe563p + 1./162.*
         m66*pow(me2,2)*RSe451p*RSe463p + 5./81.*m66*pow(me2,2)*RSe451p
         *RSe453p - 1./162.*m66*pow(me2,2)*RSe451p*RSe561p - 1./162.*
         m66*pow(me2,2)*RSe451p*RSe461p - 5./162.*m66*pow(me2,2)*pow(
         RSe451p,2) + 1./162.*m66*pow(me2,2)*RSe456p*pow(RSe453p,2) - 1.
         /81.*m66*pow(me2,2)*RSe456p*RSe451p*RSe453p );

      M13p6x23R +=  + Bbeps(me2,mu2)*pi16 * ( 1./162.*m66*pow(me2,2)*
         RSe456p*pow(RSe451p,2) - 1./162.*pow(m66,2)*me2*RSe453p*
         RSe563p - 1./162.*pow(m66,2)*me2*RSe453p*RSe463p + 1./486.*
         pow(m66,2)*me2*pow(RSe453p,2) + 1./162.*pow(m66,2)*me2*RS645pe
         *pow(RSe453p,2) + 1./162.*pow(m66,2)*me2*RSe561p*RSe453p + 1./
         162.*pow(m66,2)*me2*RSe461p*RSe453p + 1./162.*pow(m66,2)*me2*
         RSe451p*RSe563p + 1./162.*pow(m66,2)*me2*RSe451p*RSe463p - 1./
         243.*pow(m66,2)*me2*RSe451p*RSe453p - 1./81.*pow(m66,2)*me2*
         RSe451p*RS645pe*RSe453p - 1./162.*pow(m66,2)*me2*RSe451p*
         RSe561p - 1./162.*pow(m66,2)*me2*RSe451p*RSe461p + 1./486.*
         pow(m66,2)*me2*pow(RSe451p,2) + 1./162.*pow(m66,2)*me2*pow(
         RSe451p,2)*RS645pe + 1./162.*m55*me2*mp2*RSp465e*pow(
         RSe463p,2) - 1./81.*m55*me2*mp2*RSp465e*RSe461p*RSe463p + 1./
         162.*m55*me2*mp2*RSp465e*pow(RSe461p,2) - 1./162.*m55*pow(
         me2,2)*pow(RSe563p,2) - 1./162.*m55*pow(me2,2)*RSe463p*RSe563p
          - 5./162.*m55*pow(me2,2)*pow(RSe463p,2) );

      M13p6x23R +=  + Bbeps(me2,mu2)*pi16 * (  - 1./162.*m55*pow(me2,2)
         *RSe453p*RSe463p - 1./162.*m55*pow(me2,2)*pow(RSe453p,2) + 1./
         81.*m55*pow(me2,2)*RSe561p*RSe563p + 1./162.*m55*pow(me2,2)*
         RSe561p*RSe463p - 1./162.*m55*pow(me2,2)*pow(RSe561p,2) + 1./
         162.*m55*pow(me2,2)*RSe461p*RSe563p + 5./81.*m55*pow(me2,2)*
         RSe461p*RSe463p + 1./162.*m55*pow(me2,2)*RSe461p*RSe453p - 1./
         162.*m55*pow(me2,2)*RSe461p*RSe561p - 5./162.*m55*pow(me2,2)*
         pow(RSe461p,2) + 1./162.*m55*pow(me2,2)*RSe451p*RSe463p + 1./
         81.*m55*pow(me2,2)*RSe451p*RSe453p - 1./162.*m55*pow(me2,2)*
         RSe451p*RSe461p - 1./162.*m55*pow(me2,2)*pow(RSe451p,2) + 1./
         162.*m55*pow(me2,2)*RSe465p*pow(RSe463p,2) - 1./81.*m55*pow(
         me2,2)*RSe465p*RSe461p*RSe463p + 1./162.*m55*pow(me2,2)*
         RSe465p*pow(RSe461p,2) + 2./243.*m55*m66*me2*pow(RSe463p,2) - 
         1./81.*m55*m66*me2*RSe453p*RSe463p + 2./243.*m55*m66*me2*pow(
         RSe453p,2) - 4./243.*m55*m66*me2*RSe461p*RSe463p + 1./81.*m55*
         m66*me2*RSe461p*RSe453p );

      M13p6x23R +=  + Bbeps(me2,mu2)*pi16 * ( 2./243.*m55*m66*me2*pow(
         RSe461p,2) + 1./81.*m55*m66*me2*RSe451p*RSe463p - 4./243.*m55*
         m66*me2*RSe451p*RSe453p - 1./81.*m55*m66*me2*RSe451p*RSe461p
          + 2./243.*m55*m66*me2*pow(RSe451p,2) - 1./162.*pow(m55,2)*me2
         *RSe463p*RSe563p + 1./486.*pow(m55,2)*me2*pow(RSe463p,2) - 1./
         162.*pow(m55,2)*me2*RSe453p*RSe463p + 1./162.*pow(m55,2)*me2*
         RS546pe*pow(RSe463p,2) + 1./162.*pow(m55,2)*me2*RSe561p*
         RSe463p + 1./162.*pow(m55,2)*me2*RSe461p*RSe563p - 1./243.*
         pow(m55,2)*me2*RSe461p*RSe463p + 1./162.*pow(m55,2)*me2*
         RSe461p*RSe453p - 1./81.*pow(m55,2)*me2*RSe461p*RS546pe*
         RSe463p - 1./162.*pow(m55,2)*me2*RSe461p*RSe561p + 1./486.*
         pow(m55,2)*me2*pow(RSe461p,2) + 1./162.*pow(m55,2)*me2*pow(
         RSe461p,2)*RS546pe + 1./162.*pow(m55,2)*me2*RSe451p*RSe463p - 
         1./162.*pow(m55,2)*me2*RSe451p*RSe461p + 1./162.*m44*me2*mp2*
         RSp564e*pow(RSe563p,2) - 1./81.*m44*me2*mp2*RSp564e*RSe561p*
         RSe563p );

      M13p6x23R +=  + Bbeps(me2,mu2)*pi16 * ( 1./162.*m44*me2*mp2*
         RSp564e*pow(RSe561p,2) - 5./162.*m44*pow(me2,2)*pow(RSe563p,2)
          - 1./162.*m44*pow(me2,2)*RSe463p*RSe563p - 1./162.*m44*pow(
         me2,2)*pow(RSe463p,2) - 1./162.*m44*pow(me2,2)*RSe453p*RSe563p
          - 1./162.*m44*pow(me2,2)*pow(RSe453p,2) + 5./81.*m44*pow(
         me2,2)*RSe561p*RSe563p + 1./162.*m44*pow(me2,2)*RSe561p*
         RSe463p + 1./162.*m44*pow(me2,2)*RSe561p*RSe453p - 5./162.*m44
         *pow(me2,2)*pow(RSe561p,2) + 1./162.*m44*pow(me2,2)*RSe461p*
         RSe563p + 1./81.*m44*pow(me2,2)*RSe461p*RSe463p - 1./162.*m44*
         pow(me2,2)*RSe461p*RSe561p - 1./162.*m44*pow(me2,2)*pow(
         RSe461p,2) + 1./162.*m44*pow(me2,2)*RSe451p*RSe563p + 1./81.*
         m44*pow(me2,2)*RSe451p*RSe453p - 1./162.*m44*pow(me2,2)*
         RSe451p*RSe561p - 1./162.*m44*pow(me2,2)*pow(RSe451p,2) + 1./
         162.*m44*pow(me2,2)*RSe564p*pow(RSe563p,2) - 1./81.*m44*pow(
         me2,2)*RSe564p*RSe561p*RSe563p + 1./162.*m44*pow(me2,2)*
         RSe564p*pow(RSe561p,2) );

      M13p6x23R +=  + Bbeps(me2,mu2)*pi16 * ( 2./243.*m44*m66*me2*pow(
         RSe563p,2) - 1./81.*m44*m66*me2*RSe453p*RSe563p + 2./243.*m44*
         m66*me2*pow(RSe453p,2) - 4./243.*m44*m66*me2*RSe561p*RSe563p
          + 1./81.*m44*m66*me2*RSe561p*RSe453p + 2./243.*m44*m66*me2*
         pow(RSe561p,2) + 1./81.*m44*m66*me2*RSe451p*RSe563p - 4./243.*
         m44*m66*me2*RSe451p*RSe453p - 1./81.*m44*m66*me2*RSe451p*
         RSe561p + 2./243.*m44*m66*me2*pow(RSe451p,2) + 2./243.*m44*m55
         *me2*pow(RSe563p,2) - 1./81.*m44*m55*me2*RSe463p*RSe563p + 2./
         243.*m44*m55*me2*pow(RSe463p,2) - 4./243.*m44*m55*me2*RSe561p*
         RSe563p + 1./81.*m44*m55*me2*RSe561p*RSe463p + 2./243.*m44*m55
         *me2*pow(RSe561p,2) + 1./81.*m44*m55*me2*RSe461p*RSe563p - 4./
         243.*m44*m55*me2*RSe461p*RSe463p - 1./81.*m44*m55*me2*RSe461p*
         RSe561p + 2./243.*m44*m55*me2*pow(RSe461p,2) + 1./486.*pow(
         m44,2)*me2*pow(RSe563p,2) - 1./162.*pow(m44,2)*me2*RSe463p*
         RSe563p - 1./162.*pow(m44,2)*me2*RSe453p*RSe563p + 1./162.*
         pow(m44,2)*me2*RS456pe*pow(RSe563p,2) );

      M13p6x23R +=  + Bbeps(me2,mu2)*pi16 * (  - 1./243.*pow(m44,2)*me2
         *RSe561p*RSe563p + 1./162.*pow(m44,2)*me2*RSe561p*RSe463p + 1./
         162.*pow(m44,2)*me2*RSe561p*RSe453p - 1./81.*pow(m44,2)*me2*
         RSe561p*RS456pe*RSe563p + 1./486.*pow(m44,2)*me2*pow(
         RSe561p,2) + 1./162.*pow(m44,2)*me2*pow(RSe561p,2)*RS456pe + 1.
         /162.*pow(m44,2)*me2*RSe461p*RSe563p - 1./162.*pow(m44,2)*me2*
         RSe461p*RSe561p + 1./162.*pow(m44,2)*me2*RSe451p*RSe563p - 1./
         162.*pow(m44,2)*me2*RSe451p*RSe561p - 1./162.*m33*m66*mp2*
         RSp456e*pow(RSe453p,2) - 1./81.*m33*m66*mp2*RSp456e*RSe451p*
         RSe453p + 1./162.*m33*m66*me2*pow(RSe563p,2) + 1./162.*m33*m66
         *me2*pow(RSe463p,2) + 1./162.*m33*m66*me2*RSe453p*RSe563p + 1./
         162.*m33*m66*me2*RSe453p*RSe463p + 5./162.*m33*m66*me2*pow(
         RSe453p,2) + 1./81.*m33*m66*me2*RSe561p*RSe563p + 1./162.*m33*
         m66*me2*RSe561p*RSe453p + 1./81.*m33*m66*me2*RSe461p*RSe463p
          + 1./162.*m33*m66*me2*RSe461p*RSe453p + 1./162.*m33*m66*me2*
         RSe451p*RSe563p );

      M13p6x23R +=  + Bbeps(me2,mu2)*pi16 * ( 1./162.*m33*m66*me2*
         RSe451p*RSe463p + 5./81.*m33*m66*me2*RSe451p*RSe453p - 1./162.
         *m33*m66*me2*RSe456p*pow(RSe453p,2) - 1./81.*m33*m66*me2*
         RSe456p*RSe451p*RSe453p + 1./162.*m33*pow(m66,2)*RSe453p*
         RSe563p + 1./162.*m33*pow(m66,2)*RSe453p*RSe463p - 1./486.*m33
         *pow(m66,2)*pow(RSe453p,2) - 1./162.*m33*pow(m66,2)*RS645pe*
         pow(RSe453p,2) + 1./162.*m33*pow(m66,2)*RSe561p*RSe453p + 1./
         162.*m33*pow(m66,2)*RSe461p*RSe453p + 1./162.*m33*pow(m66,2)*
         RSe451p*RSe563p + 1./162.*m33*pow(m66,2)*RSe451p*RSe463p - 1./
         243.*m33*pow(m66,2)*RSe451p*RSe453p - 1./81.*m33*pow(m66,2)*
         RSe451p*RS645pe*RSe453p - 1./162.*m33*m55*mp2*RSp465e*pow(
         RSe463p,2) - 1./81.*m33*m55*mp2*RSp465e*RSe461p*RSe463p + 1./
         162.*m33*m55*me2*pow(RSe563p,2) + 1./162.*m33*m55*me2*RSe463p*
         RSe563p + 5./162.*m33*m55*me2*pow(RSe463p,2) + 1./162.*m33*m55
         *me2*RSe453p*RSe463p + 1./162.*m33*m55*me2*pow(RSe453p,2) + 1./
         81.*m33*m55*me2*RSe561p*RSe563p );

      M13p6x23R +=  + Bbeps(me2,mu2)*pi16 * ( 1./162.*m33*m55*me2*
         RSe561p*RSe463p + 1./162.*m33*m55*me2*RSe461p*RSe563p + 5./81.
         *m33*m55*me2*RSe461p*RSe463p + 1./162.*m33*m55*me2*RSe461p*
         RSe453p + 1./162.*m33*m55*me2*RSe451p*RSe463p + 1./81.*m33*m55
         *me2*RSe451p*RSe453p - 1./162.*m33*m55*me2*RSe465p*pow(
         RSe463p,2) - 1./81.*m33*m55*me2*RSe465p*RSe461p*RSe463p - 2./
         243.*m33*m55*m66*pow(RSe463p,2) + 1./81.*m33*m55*m66*RSe453p*
         RSe463p - 2./243.*m33*m55*m66*pow(RSe453p,2) - 4./243.*m33*m55
         *m66*RSe461p*RSe463p + 1./81.*m33*m55*m66*RSe461p*RSe453p + 1./
         81.*m33*m55*m66*RSe451p*RSe463p - 4./243.*m33*m55*m66*RSe451p*
         RSe453p + 1./162.*m33*pow(m55,2)*RSe463p*RSe563p - 1./486.*m33
         *pow(m55,2)*pow(RSe463p,2) + 1./162.*m33*pow(m55,2)*RSe453p*
         RSe463p - 1./162.*m33*pow(m55,2)*RS546pe*pow(RSe463p,2) + 1./
         162.*m33*pow(m55,2)*RSe561p*RSe463p + 1./162.*m33*pow(m55,2)*
         RSe461p*RSe563p - 1./243.*m33*pow(m55,2)*RSe461p*RSe463p + 1./
         162.*m33*pow(m55,2)*RSe461p*RSe453p );

      M13p6x23R +=  + Bbeps(me2,mu2)*pi16 * (  - 1./81.*m33*pow(m55,2)*
         RSe461p*RS546pe*RSe463p + 1./162.*m33*pow(m55,2)*RSe451p*
         RSe463p - 1./162.*m33*m44*mp2*RSp564e*pow(RSe563p,2) - 1./81.*
         m33*m44*mp2*RSp564e*RSe561p*RSe563p + 5./162.*m33*m44*me2*pow(
         RSe563p,2) + 1./162.*m33*m44*me2*RSe463p*RSe563p + 1./162.*m33
         *m44*me2*pow(RSe463p,2) + 1./162.*m33*m44*me2*RSe453p*RSe563p
          + 1./162.*m33*m44*me2*pow(RSe453p,2) + 5./81.*m33*m44*me2*
         RSe561p*RSe563p + 1./162.*m33*m44*me2*RSe561p*RSe463p + 1./162.
         *m33*m44*me2*RSe561p*RSe453p + 1./162.*m33*m44*me2*RSe461p*
         RSe563p + 1./81.*m33*m44*me2*RSe461p*RSe463p + 1./162.*m33*m44
         *me2*RSe451p*RSe563p + 1./81.*m33*m44*me2*RSe451p*RSe453p - 1./
         162.*m33*m44*me2*RSe564p*pow(RSe563p,2) - 1./81.*m33*m44*me2*
         RSe564p*RSe561p*RSe563p - 2./243.*m33*m44*m66*pow(RSe563p,2)
          + 1./81.*m33*m44*m66*RSe453p*RSe563p - 2./243.*m33*m44*m66*
         pow(RSe453p,2) - 4./243.*m33*m44*m66*RSe561p*RSe563p + 1./81.*
         m33*m44*m66*RSe561p*RSe453p );

      M13p6x23R +=  + Bbeps(me2,mu2)*pi16 * ( 1./81.*m33*m44*m66*
         RSe451p*RSe563p - 4./243.*m33*m44*m66*RSe451p*RSe453p - 2./243.
         *m33*m44*m55*pow(RSe563p,2) + 1./81.*m33*m44*m55*RSe463p*
         RSe563p - 2./243.*m33*m44*m55*pow(RSe463p,2) - 4./243.*m33*m44
         *m55*RSe561p*RSe563p + 1./81.*m33*m44*m55*RSe561p*RSe463p + 1./
         81.*m33*m44*m55*RSe461p*RSe563p - 4./243.*m33*m44*m55*RSe461p*
         RSe463p - 1./486.*m33*pow(m44,2)*pow(RSe563p,2) + 1./162.*m33*
         pow(m44,2)*RSe463p*RSe563p + 1./162.*m33*pow(m44,2)*RSe453p*
         RSe563p - 1./162.*m33*pow(m44,2)*RS456pe*pow(RSe563p,2) - 1./
         243.*m33*pow(m44,2)*RSe561p*RSe563p + 1./162.*m33*pow(m44,2)*
         RSe561p*RSe463p + 1./162.*m33*pow(m44,2)*RSe561p*RSe453p - 1./
         81.*m33*pow(m44,2)*RSe561p*RS456pe*RSe563p + 1./162.*m33*pow(
         m44,2)*RSe461p*RSe563p + 1./162.*m33*pow(m44,2)*RSe451p*
         RSe563p - 1./81.*m11*m66*mp2*RSp456e*RSe451p*RSe453p - 1./162.
         *m11*m66*mp2*RSp456e*pow(RSe451p,2) + 1./81.*m11*m66*me2*
         RSe561p*RSe563p );

      M13p6x23R +=  + Bbeps(me2,mu2)*pi16 * ( 1./162.*m11*m66*me2*
         RSe561p*RSe453p + 1./162.*m11*m66*me2*pow(RSe561p,2) + 1./81.*
         m11*m66*me2*RSe461p*RSe463p + 1./162.*m11*m66*me2*RSe461p*
         RSe453p + 1./162.*m11*m66*me2*pow(RSe461p,2) + 1./162.*m11*m66
         *me2*RSe451p*RSe563p + 1./162.*m11*m66*me2*RSe451p*RSe463p + 5.
         /81.*m11*m66*me2*RSe451p*RSe453p + 1./162.*m11*m66*me2*RSe451p
         *RSe561p + 1./162.*m11*m66*me2*RSe451p*RSe461p + 5./162.*m11*
         m66*me2*pow(RSe451p,2) - 1./81.*m11*m66*me2*RSe456p*RSe451p*
         RSe453p - 1./162.*m11*m66*me2*RSe456p*pow(RSe451p,2) + 1./162.
         *m11*pow(m66,2)*RSe561p*RSe453p + 1./162.*m11*pow(m66,2)*
         RSe461p*RSe453p + 1./162.*m11*pow(m66,2)*RSe451p*RSe563p + 1./
         162.*m11*pow(m66,2)*RSe451p*RSe463p - 1./243.*m11*pow(m66,2)*
         RSe451p*RSe453p - 1./81.*m11*pow(m66,2)*RSe451p*RS645pe*
         RSe453p + 1./162.*m11*pow(m66,2)*RSe451p*RSe561p + 1./162.*m11
         *pow(m66,2)*RSe451p*RSe461p - 1./486.*m11*pow(m66,2)*pow(
         RSe451p,2) );

      M13p6x23R +=  + Bbeps(me2,mu2)*pi16 * (  - 1./162.*m11*pow(m66,2)
         *pow(RSe451p,2)*RS645pe - 1./81.*m11*m55*mp2*RSp465e*RSe461p*
         RSe463p - 1./162.*m11*m55*mp2*RSp465e*pow(RSe461p,2) + 1./81.*
         m11*m55*me2*RSe561p*RSe563p + 1./162.*m11*m55*me2*RSe561p*
         RSe463p + 1./162.*m11*m55*me2*pow(RSe561p,2) + 1./162.*m11*m55
         *me2*RSe461p*RSe563p + 5./81.*m11*m55*me2*RSe461p*RSe463p + 1./
         162.*m11*m55*me2*RSe461p*RSe453p + 1./162.*m11*m55*me2*RSe461p
         *RSe561p + 5./162.*m11*m55*me2*pow(RSe461p,2) + 1./162.*m11*
         m55*me2*RSe451p*RSe463p + 1./81.*m11*m55*me2*RSe451p*RSe453p
          + 1./162.*m11*m55*me2*RSe451p*RSe461p + 1./162.*m11*m55*me2*
         pow(RSe451p,2) - 1./81.*m11*m55*me2*RSe465p*RSe461p*RSe463p - 
         1./162.*m11*m55*me2*RSe465p*pow(RSe461p,2) - 4./243.*m11*m55*
         m66*RSe461p*RSe463p + 1./81.*m11*m55*m66*RSe461p*RSe453p - 2./
         243.*m11*m55*m66*pow(RSe461p,2) + 1./81.*m11*m55*m66*RSe451p*
         RSe463p - 4./243.*m11*m55*m66*RSe451p*RSe453p + 1./81.*m11*m55
         *m66*RSe451p*RSe461p );

      M13p6x23R +=  + Bbeps(me2,mu2)*pi16 * (  - 2./243.*m11*m55*m66*
         pow(RSe451p,2) + 1./162.*m11*pow(m55,2)*RSe561p*RSe463p + 1./
         162.*m11*pow(m55,2)*RSe461p*RSe563p - 1./243.*m11*pow(m55,2)*
         RSe461p*RSe463p + 1./162.*m11*pow(m55,2)*RSe461p*RSe453p - 1./
         81.*m11*pow(m55,2)*RSe461p*RS546pe*RSe463p + 1./162.*m11*pow(
         m55,2)*RSe461p*RSe561p - 1./486.*m11*pow(m55,2)*pow(RSe461p,2)
          - 1./162.*m11*pow(m55,2)*pow(RSe461p,2)*RS546pe + 1./162.*m11
         *pow(m55,2)*RSe451p*RSe463p + 1./162.*m11*pow(m55,2)*RSe451p*
         RSe461p - 1./81.*m11*m44*mp2*RSp564e*RSe561p*RSe563p - 1./162.
         *m11*m44*mp2*RSp564e*pow(RSe561p,2) + 5./81.*m11*m44*me2*
         RSe561p*RSe563p + 1./162.*m11*m44*me2*RSe561p*RSe463p + 1./162.
         *m11*m44*me2*RSe561p*RSe453p + 5./162.*m11*m44*me2*pow(
         RSe561p,2) + 1./162.*m11*m44*me2*RSe461p*RSe563p + 1./81.*m11*
         m44*me2*RSe461p*RSe463p + 1./162.*m11*m44*me2*RSe461p*RSe561p
          + 1./162.*m11*m44*me2*pow(RSe461p,2) + 1./162.*m11*m44*me2*
         RSe451p*RSe563p );

      M13p6x23R +=  + Bbeps(me2,mu2)*pi16 * ( 1./81.*m11*m44*me2*
         RSe451p*RSe453p + 1./162.*m11*m44*me2*RSe451p*RSe561p + 1./162.
         *m11*m44*me2*pow(RSe451p,2) - 1./81.*m11*m44*me2*RSe564p*
         RSe561p*RSe563p - 1./162.*m11*m44*me2*RSe564p*pow(RSe561p,2)
          - 4./243.*m11*m44*m66*RSe561p*RSe563p + 1./81.*m11*m44*m66*
         RSe561p*RSe453p - 2./243.*m11*m44*m66*pow(RSe561p,2) + 1./81.*
         m11*m44*m66*RSe451p*RSe563p - 4./243.*m11*m44*m66*RSe451p*
         RSe453p + 1./81.*m11*m44*m66*RSe451p*RSe561p - 2./243.*m11*m44
         *m66*pow(RSe451p,2) - 4./243.*m11*m44*m55*RSe561p*RSe563p + 1./
         81.*m11*m44*m55*RSe561p*RSe463p - 2./243.*m11*m44*m55*pow(
         RSe561p,2) + 1./81.*m11*m44*m55*RSe461p*RSe563p - 4./243.*m11*
         m44*m55*RSe461p*RSe463p + 1./81.*m11*m44*m55*RSe461p*RSe561p
          - 2./243.*m11*m44*m55*pow(RSe461p,2) - 1./243.*m11*pow(m44,2)
         *RSe561p*RSe563p + 1./162.*m11*pow(m44,2)*RSe561p*RSe463p + 1./
         162.*m11*pow(m44,2)*RSe561p*RSe453p - 1./81.*m11*pow(m44,2)*
         RSe561p*RS456pe*RSe563p );

      M13p6x23R +=  + Bbeps(me2,mu2)*pi16 * (  - 1./486.*m11*pow(m44,2)
         *pow(RSe561p,2) - 1./162.*m11*pow(m44,2)*pow(RSe561p,2)*
         RS456pe + 1./162.*m11*pow(m44,2)*RSe461p*RSe563p + 1./162.*m11
         *pow(m44,2)*RSe461p*RSe561p + 1./162.*m11*pow(m44,2)*RSe451p*
         RSe563p + 1./162.*m11*pow(m44,2)*RSe451p*RSe561p );

      M13p6x23R +=  + Bbeps(mp2,mu2)*pi16 * (  - 1./162.*m66*pow(mp2,2)
         *pow(RSp563e,2) - 1./162.*m66*pow(mp2,2)*pow(RSp463e,2) - 1./
         162.*m66*pow(mp2,2)*RSp453e*RSp563e - 1./162.*m66*pow(mp2,2)*
         RSp453e*RSp463e - 5./162.*m66*pow(mp2,2)*pow(RSp453e,2) + 1./
         81.*m66*pow(mp2,2)*RSp561e*RSp563e + 1./162.*m66*pow(mp2,2)*
         RSp561e*RSp453e - 1./162.*m66*pow(mp2,2)*pow(RSp561e,2) + 1./
         81.*m66*pow(mp2,2)*RSp461e*RSp463e + 1./162.*m66*pow(mp2,2)*
         RSp461e*RSp453e - 1./162.*m66*pow(mp2,2)*pow(RSp461e,2) + 1./
         162.*m66*pow(mp2,2)*RSp451e*RSp563e + 1./162.*m66*pow(mp2,2)*
         RSp451e*RSp463e + 5./81.*m66*pow(mp2,2)*RSp451e*RSp453e - 1./
         162.*m66*pow(mp2,2)*RSp451e*RSp561e - 1./162.*m66*pow(mp2,2)*
         RSp451e*RSp461e - 5./162.*m66*pow(mp2,2)*pow(RSp451e,2) + 1./
         162.*m66*pow(mp2,2)*RSp456e*pow(RSp453e,2) - 1./81.*m66*pow(
         mp2,2)*RSp456e*RSp451e*RSp453e + 1./162.*m66*pow(mp2,2)*
         RSp456e*pow(RSp451e,2) + 1./162.*m66*me2*mp2*RSe456p*pow(
         RSp453e,2) );

      M13p6x23R +=  + Bbeps(mp2,mu2)*pi16 * (  - 1./81.*m66*me2*mp2*
         RSe456p*RSp451e*RSp453e + 1./162.*m66*me2*mp2*RSe456p*pow(
         RSp451e,2) - 1./162.*pow(m66,2)*mp2*RSp453e*RSp563e - 1./162.*
         pow(m66,2)*mp2*RSp453e*RSp463e + 1./486.*pow(m66,2)*mp2*pow(
         RSp453e,2) + 1./162.*pow(m66,2)*mp2*RS645pe*pow(RSp453e,2) + 1.
         /162.*pow(m66,2)*mp2*RSp561e*RSp453e + 1./162.*pow(m66,2)*mp2*
         RSp461e*RSp453e + 1./162.*pow(m66,2)*mp2*RSp451e*RSp563e + 1./
         162.*pow(m66,2)*mp2*RSp451e*RSp463e - 1./243.*pow(m66,2)*mp2*
         RSp451e*RSp453e - 1./81.*pow(m66,2)*mp2*RSp451e*RS645pe*
         RSp453e - 1./162.*pow(m66,2)*mp2*RSp451e*RSp561e - 1./162.*
         pow(m66,2)*mp2*RSp451e*RSp461e + 1./486.*pow(m66,2)*mp2*pow(
         RSp451e,2) + 1./162.*pow(m66,2)*mp2*pow(RSp451e,2)*RS645pe - 1.
         /162.*m55*pow(mp2,2)*pow(RSp563e,2) - 1./162.*m55*pow(mp2,2)*
         RSp463e*RSp563e - 5./162.*m55*pow(mp2,2)*pow(RSp463e,2) - 1./
         162.*m55*pow(mp2,2)*RSp453e*RSp463e - 1./162.*m55*pow(mp2,2)*
         pow(RSp453e,2) );

      M13p6x23R +=  + Bbeps(mp2,mu2)*pi16 * ( 1./81.*m55*pow(mp2,2)*
         RSp561e*RSp563e + 1./162.*m55*pow(mp2,2)*RSp561e*RSp463e - 1./
         162.*m55*pow(mp2,2)*pow(RSp561e,2) + 1./162.*m55*pow(mp2,2)*
         RSp461e*RSp563e + 5./81.*m55*pow(mp2,2)*RSp461e*RSp463e + 1./
         162.*m55*pow(mp2,2)*RSp461e*RSp453e - 1./162.*m55*pow(mp2,2)*
         RSp461e*RSp561e - 5./162.*m55*pow(mp2,2)*pow(RSp461e,2) + 1./
         162.*m55*pow(mp2,2)*RSp451e*RSp463e + 1./81.*m55*pow(mp2,2)*
         RSp451e*RSp453e - 1./162.*m55*pow(mp2,2)*RSp451e*RSp461e - 1./
         162.*m55*pow(mp2,2)*pow(RSp451e,2) + 1./162.*m55*pow(mp2,2)*
         RSp465e*pow(RSp463e,2) - 1./81.*m55*pow(mp2,2)*RSp465e*RSp461e
         *RSp463e + 1./162.*m55*pow(mp2,2)*RSp465e*pow(RSp461e,2) + 1./
         162.*m55*me2*mp2*RSe465p*pow(RSp463e,2) - 1./81.*m55*me2*mp2*
         RSe465p*RSp461e*RSp463e + 1./162.*m55*me2*mp2*RSe465p*pow(
         RSp461e,2) + 2./243.*m55*m66*mp2*pow(RSp463e,2) - 1./81.*m55*
         m66*mp2*RSp453e*RSp463e + 2./243.*m55*m66*mp2*pow(RSp453e,2)
          - 4./243.*m55*m66*mp2*RSp461e*RSp463e );

      M13p6x23R +=  + Bbeps(mp2,mu2)*pi16 * ( 1./81.*m55*m66*mp2*
         RSp461e*RSp453e + 2./243.*m55*m66*mp2*pow(RSp461e,2) + 1./81.*
         m55*m66*mp2*RSp451e*RSp463e - 4./243.*m55*m66*mp2*RSp451e*
         RSp453e - 1./81.*m55*m66*mp2*RSp451e*RSp461e + 2./243.*m55*m66
         *mp2*pow(RSp451e,2) - 1./162.*pow(m55,2)*mp2*RSp463e*RSp563e
          + 1./486.*pow(m55,2)*mp2*pow(RSp463e,2) - 1./162.*pow(m55,2)*
         mp2*RSp453e*RSp463e + 1./162.*pow(m55,2)*mp2*RS546pe*pow(
         RSp463e,2) + 1./162.*pow(m55,2)*mp2*RSp561e*RSp463e + 1./162.*
         pow(m55,2)*mp2*RSp461e*RSp563e - 1./243.*pow(m55,2)*mp2*
         RSp461e*RSp463e + 1./162.*pow(m55,2)*mp2*RSp461e*RSp453e - 1./
         81.*pow(m55,2)*mp2*RSp461e*RS546pe*RSp463e - 1./162.*pow(
         m55,2)*mp2*RSp461e*RSp561e + 1./486.*pow(m55,2)*mp2*pow(
         RSp461e,2) + 1./162.*pow(m55,2)*mp2*pow(RSp461e,2)*RS546pe + 1.
         /162.*pow(m55,2)*mp2*RSp451e*RSp463e - 1./162.*pow(m55,2)*mp2*
         RSp451e*RSp461e - 5./162.*m44*pow(mp2,2)*pow(RSp563e,2) - 1./
         162.*m44*pow(mp2,2)*RSp463e*RSp563e );

      M13p6x23R +=  + Bbeps(mp2,mu2)*pi16 * (  - 1./162.*m44*pow(mp2,2)
         *pow(RSp463e,2) - 1./162.*m44*pow(mp2,2)*RSp453e*RSp563e - 1./
         162.*m44*pow(mp2,2)*pow(RSp453e,2) + 5./81.*m44*pow(mp2,2)*
         RSp561e*RSp563e + 1./162.*m44*pow(mp2,2)*RSp561e*RSp463e + 1./
         162.*m44*pow(mp2,2)*RSp561e*RSp453e - 5./162.*m44*pow(mp2,2)*
         pow(RSp561e,2) + 1./162.*m44*pow(mp2,2)*RSp461e*RSp563e + 1./
         81.*m44*pow(mp2,2)*RSp461e*RSp463e - 1./162.*m44*pow(mp2,2)*
         RSp461e*RSp561e - 1./162.*m44*pow(mp2,2)*pow(RSp461e,2) + 1./
         162.*m44*pow(mp2,2)*RSp451e*RSp563e + 1./81.*m44*pow(mp2,2)*
         RSp451e*RSp453e - 1./162.*m44*pow(mp2,2)*RSp451e*RSp561e - 1./
         162.*m44*pow(mp2,2)*pow(RSp451e,2) + 1./162.*m44*pow(mp2,2)*
         RSp564e*pow(RSp563e,2) - 1./81.*m44*pow(mp2,2)*RSp564e*RSp561e
         *RSp563e + 1./162.*m44*pow(mp2,2)*RSp564e*pow(RSp561e,2) + 1./
         162.*m44*me2*mp2*RSe564p*pow(RSp563e,2) - 1./81.*m44*me2*mp2*
         RSe564p*RSp561e*RSp563e + 1./162.*m44*me2*mp2*RSe564p*pow(
         RSp561e,2) );

      M13p6x23R +=  + Bbeps(mp2,mu2)*pi16 * ( 2./243.*m44*m66*mp2*pow(
         RSp563e,2) - 1./81.*m44*m66*mp2*RSp453e*RSp563e + 2./243.*m44*
         m66*mp2*pow(RSp453e,2) - 4./243.*m44*m66*mp2*RSp561e*RSp563e
          + 1./81.*m44*m66*mp2*RSp561e*RSp453e + 2./243.*m44*m66*mp2*
         pow(RSp561e,2) + 1./81.*m44*m66*mp2*RSp451e*RSp563e - 4./243.*
         m44*m66*mp2*RSp451e*RSp453e - 1./81.*m44*m66*mp2*RSp451e*
         RSp561e + 2./243.*m44*m66*mp2*pow(RSp451e,2) + 2./243.*m44*m55
         *mp2*pow(RSp563e,2) - 1./81.*m44*m55*mp2*RSp463e*RSp563e + 2./
         243.*m44*m55*mp2*pow(RSp463e,2) - 4./243.*m44*m55*mp2*RSp561e*
         RSp563e + 1./81.*m44*m55*mp2*RSp561e*RSp463e + 2./243.*m44*m55
         *mp2*pow(RSp561e,2) + 1./81.*m44*m55*mp2*RSp461e*RSp563e - 4./
         243.*m44*m55*mp2*RSp461e*RSp463e - 1./81.*m44*m55*mp2*RSp461e*
         RSp561e + 2./243.*m44*m55*mp2*pow(RSp461e,2) + 1./486.*pow(
         m44,2)*mp2*pow(RSp563e,2) - 1./162.*pow(m44,2)*mp2*RSp463e*
         RSp563e - 1./162.*pow(m44,2)*mp2*RSp453e*RSp563e + 1./162.*
         pow(m44,2)*mp2*RS456pe*pow(RSp563e,2) );

      M13p6x23R +=  + Bbeps(mp2,mu2)*pi16 * (  - 1./243.*pow(m44,2)*mp2
         *RSp561e*RSp563e + 1./162.*pow(m44,2)*mp2*RSp561e*RSp463e + 1./
         162.*pow(m44,2)*mp2*RSp561e*RSp453e - 1./81.*pow(m44,2)*mp2*
         RSp561e*RS456pe*RSp563e + 1./486.*pow(m44,2)*mp2*pow(
         RSp561e,2) + 1./162.*pow(m44,2)*mp2*pow(RSp561e,2)*RS456pe + 1.
         /162.*pow(m44,2)*mp2*RSp461e*RSp563e - 1./162.*pow(m44,2)*mp2*
         RSp461e*RSp561e + 1./162.*pow(m44,2)*mp2*RSp451e*RSp563e - 1./
         162.*pow(m44,2)*mp2*RSp451e*RSp561e + 1./162.*m33*m66*mp2*pow(
         RSp563e,2) + 1./162.*m33*m66*mp2*pow(RSp463e,2) + 1./162.*m33*
         m66*mp2*RSp453e*RSp563e + 1./162.*m33*m66*mp2*RSp453e*RSp463e
          + 5./162.*m33*m66*mp2*pow(RSp453e,2) + 1./81.*m33*m66*mp2*
         RSp561e*RSp563e + 1./162.*m33*m66*mp2*RSp561e*RSp453e + 1./81.
         *m33*m66*mp2*RSp461e*RSp463e + 1./162.*m33*m66*mp2*RSp461e*
         RSp453e + 1./162.*m33*m66*mp2*RSp451e*RSp563e + 1./162.*m33*
         m66*mp2*RSp451e*RSp463e + 5./81.*m33*m66*mp2*RSp451e*RSp453e
          - 1./162.*m33*m66*mp2*RSp456e*pow(RSp453e,2) );

      M13p6x23R +=  + Bbeps(mp2,mu2)*pi16 * (  - 1./81.*m33*m66*mp2*
         RSp456e*RSp451e*RSp453e - 1./162.*m33*m66*me2*RSe456p*pow(
         RSp453e,2) - 1./81.*m33*m66*me2*RSe456p*RSp451e*RSp453e + 1./
         162.*m33*pow(m66,2)*RSp453e*RSp563e + 1./162.*m33*pow(m66,2)*
         RSp453e*RSp463e - 1./486.*m33*pow(m66,2)*pow(RSp453e,2) - 1./
         162.*m33*pow(m66,2)*RS645pe*pow(RSp453e,2) + 1./162.*m33*pow(
         m66,2)*RSp561e*RSp453e + 1./162.*m33*pow(m66,2)*RSp461e*
         RSp453e + 1./162.*m33*pow(m66,2)*RSp451e*RSp563e + 1./162.*m33
         *pow(m66,2)*RSp451e*RSp463e - 1./243.*m33*pow(m66,2)*RSp451e*
         RSp453e - 1./81.*m33*pow(m66,2)*RSp451e*RS645pe*RSp453e + 1./
         162.*m33*m55*mp2*pow(RSp563e,2) + 1./162.*m33*m55*mp2*RSp463e*
         RSp563e + 5./162.*m33*m55*mp2*pow(RSp463e,2) + 1./162.*m33*m55
         *mp2*RSp453e*RSp463e + 1./162.*m33*m55*mp2*pow(RSp453e,2) + 1./
         81.*m33*m55*mp2*RSp561e*RSp563e + 1./162.*m33*m55*mp2*RSp561e*
         RSp463e + 1./162.*m33*m55*mp2*RSp461e*RSp563e + 5./81.*m33*m55
         *mp2*RSp461e*RSp463e );

      M13p6x23R +=  + Bbeps(mp2,mu2)*pi16 * ( 1./162.*m33*m55*mp2*
         RSp461e*RSp453e + 1./162.*m33*m55*mp2*RSp451e*RSp463e + 1./81.
         *m33*m55*mp2*RSp451e*RSp453e - 1./162.*m33*m55*mp2*RSp465e*
         pow(RSp463e,2) - 1./81.*m33*m55*mp2*RSp465e*RSp461e*RSp463e - 
         1./162.*m33*m55*me2*RSe465p*pow(RSp463e,2) - 1./81.*m33*m55*
         me2*RSe465p*RSp461e*RSp463e - 2./243.*m33*m55*m66*pow(
         RSp463e,2) + 1./81.*m33*m55*m66*RSp453e*RSp463e - 2./243.*m33*
         m55*m66*pow(RSp453e,2) - 4./243.*m33*m55*m66*RSp461e*RSp463e
          + 1./81.*m33*m55*m66*RSp461e*RSp453e + 1./81.*m33*m55*m66*
         RSp451e*RSp463e - 4./243.*m33*m55*m66*RSp451e*RSp453e + 1./162.
         *m33*pow(m55,2)*RSp463e*RSp563e - 1./486.*m33*pow(m55,2)*pow(
         RSp463e,2) + 1./162.*m33*pow(m55,2)*RSp453e*RSp463e - 1./162.*
         m33*pow(m55,2)*RS546pe*pow(RSp463e,2) + 1./162.*m33*pow(m55,2)
         *RSp561e*RSp463e + 1./162.*m33*pow(m55,2)*RSp461e*RSp563e - 1./
         243.*m33*pow(m55,2)*RSp461e*RSp463e + 1./162.*m33*pow(m55,2)*
         RSp461e*RSp453e );

      M13p6x23R +=  + Bbeps(mp2,mu2)*pi16 * (  - 1./81.*m33*pow(m55,2)*
         RSp461e*RS546pe*RSp463e + 1./162.*m33*pow(m55,2)*RSp451e*
         RSp463e + 5./162.*m33*m44*mp2*pow(RSp563e,2) + 1./162.*m33*m44
         *mp2*RSp463e*RSp563e + 1./162.*m33*m44*mp2*pow(RSp463e,2) + 1./
         162.*m33*m44*mp2*RSp453e*RSp563e + 1./162.*m33*m44*mp2*pow(
         RSp453e,2) + 5./81.*m33*m44*mp2*RSp561e*RSp563e + 1./162.*m33*
         m44*mp2*RSp561e*RSp463e + 1./162.*m33*m44*mp2*RSp561e*RSp453e
          + 1./162.*m33*m44*mp2*RSp461e*RSp563e + 1./81.*m33*m44*mp2*
         RSp461e*RSp463e + 1./162.*m33*m44*mp2*RSp451e*RSp563e + 1./81.
         *m33*m44*mp2*RSp451e*RSp453e - 1./162.*m33*m44*mp2*RSp564e*
         pow(RSp563e,2) - 1./81.*m33*m44*mp2*RSp564e*RSp561e*RSp563e - 
         1./162.*m33*m44*me2*RSe564p*pow(RSp563e,2) - 1./81.*m33*m44*
         me2*RSe564p*RSp561e*RSp563e - 2./243.*m33*m44*m66*pow(
         RSp563e,2) + 1./81.*m33*m44*m66*RSp453e*RSp563e - 2./243.*m33*
         m44*m66*pow(RSp453e,2) - 4./243.*m33*m44*m66*RSp561e*RSp563e
          + 1./81.*m33*m44*m66*RSp561e*RSp453e );

      M13p6x23R +=  + Bbeps(mp2,mu2)*pi16 * ( 1./81.*m33*m44*m66*
         RSp451e*RSp563e - 4./243.*m33*m44*m66*RSp451e*RSp453e - 2./243.
         *m33*m44*m55*pow(RSp563e,2) + 1./81.*m33*m44*m55*RSp463e*
         RSp563e - 2./243.*m33*m44*m55*pow(RSp463e,2) - 4./243.*m33*m44
         *m55*RSp561e*RSp563e + 1./81.*m33*m44*m55*RSp561e*RSp463e + 1./
         81.*m33*m44*m55*RSp461e*RSp563e - 4./243.*m33*m44*m55*RSp461e*
         RSp463e - 1./486.*m33*pow(m44,2)*pow(RSp563e,2) + 1./162.*m33*
         pow(m44,2)*RSp463e*RSp563e + 1./162.*m33*pow(m44,2)*RSp453e*
         RSp563e - 1./162.*m33*pow(m44,2)*RS456pe*pow(RSp563e,2) - 1./
         243.*m33*pow(m44,2)*RSp561e*RSp563e + 1./162.*m33*pow(m44,2)*
         RSp561e*RSp463e + 1./162.*m33*pow(m44,2)*RSp561e*RSp453e - 1./
         81.*m33*pow(m44,2)*RSp561e*RS456pe*RSp563e + 1./162.*m33*pow(
         m44,2)*RSp461e*RSp563e + 1./162.*m33*pow(m44,2)*RSp451e*
         RSp563e + 1./81.*m11*m66*mp2*RSp561e*RSp563e + 1./162.*m11*m66
         *mp2*RSp561e*RSp453e + 1./162.*m11*m66*mp2*pow(RSp561e,2) + 1./
         81.*m11*m66*mp2*RSp461e*RSp463e );

      M13p6x23R +=  + Bbeps(mp2,mu2)*pi16 * ( 1./162.*m11*m66*mp2*
         RSp461e*RSp453e + 1./162.*m11*m66*mp2*pow(RSp461e,2) + 1./162.
         *m11*m66*mp2*RSp451e*RSp563e + 1./162.*m11*m66*mp2*RSp451e*
         RSp463e + 5./81.*m11*m66*mp2*RSp451e*RSp453e + 1./162.*m11*m66
         *mp2*RSp451e*RSp561e + 1./162.*m11*m66*mp2*RSp451e*RSp461e + 5.
         /162.*m11*m66*mp2*pow(RSp451e,2) - 1./81.*m11*m66*mp2*RSp456e*
         RSp451e*RSp453e - 1./162.*m11*m66*mp2*RSp456e*pow(RSp451e,2)
          - 1./81.*m11*m66*me2*RSe456p*RSp451e*RSp453e - 1./162.*m11*
         m66*me2*RSe456p*pow(RSp451e,2) + 1./162.*m11*pow(m66,2)*
         RSp561e*RSp453e + 1./162.*m11*pow(m66,2)*RSp461e*RSp453e + 1./
         162.*m11*pow(m66,2)*RSp451e*RSp563e + 1./162.*m11*pow(m66,2)*
         RSp451e*RSp463e - 1./243.*m11*pow(m66,2)*RSp451e*RSp453e - 1./
         81.*m11*pow(m66,2)*RSp451e*RS645pe*RSp453e + 1./162.*m11*pow(
         m66,2)*RSp451e*RSp561e + 1./162.*m11*pow(m66,2)*RSp451e*
         RSp461e - 1./486.*m11*pow(m66,2)*pow(RSp451e,2) - 1./162.*m11*
         pow(m66,2)*pow(RSp451e,2)*RS645pe );

      M13p6x23R +=  + Bbeps(mp2,mu2)*pi16 * ( 1./81.*m11*m55*mp2*
         RSp561e*RSp563e + 1./162.*m11*m55*mp2*RSp561e*RSp463e + 1./162.
         *m11*m55*mp2*pow(RSp561e,2) + 1./162.*m11*m55*mp2*RSp461e*
         RSp563e + 5./81.*m11*m55*mp2*RSp461e*RSp463e + 1./162.*m11*m55
         *mp2*RSp461e*RSp453e + 1./162.*m11*m55*mp2*RSp461e*RSp561e + 5.
         /162.*m11*m55*mp2*pow(RSp461e,2) + 1./162.*m11*m55*mp2*RSp451e
         *RSp463e + 1./81.*m11*m55*mp2*RSp451e*RSp453e + 1./162.*m11*
         m55*mp2*RSp451e*RSp461e + 1./162.*m11*m55*mp2*pow(RSp451e,2)
          - 1./81.*m11*m55*mp2*RSp465e*RSp461e*RSp463e - 1./162.*m11*
         m55*mp2*RSp465e*pow(RSp461e,2) - 1./81.*m11*m55*me2*RSe465p*
         RSp461e*RSp463e - 1./162.*m11*m55*me2*RSe465p*pow(RSp461e,2)
          - 4./243.*m11*m55*m66*RSp461e*RSp463e + 1./81.*m11*m55*m66*
         RSp461e*RSp453e - 2./243.*m11*m55*m66*pow(RSp461e,2) + 1./81.*
         m11*m55*m66*RSp451e*RSp463e - 4./243.*m11*m55*m66*RSp451e*
         RSp453e + 1./81.*m11*m55*m66*RSp451e*RSp461e - 2./243.*m11*m55
         *m66*pow(RSp451e,2) );

      M13p6x23R +=  + Bbeps(mp2,mu2)*pi16 * ( 1./162.*m11*pow(m55,2)*
         RSp561e*RSp463e + 1./162.*m11*pow(m55,2)*RSp461e*RSp563e - 1./
         243.*m11*pow(m55,2)*RSp461e*RSp463e + 1./162.*m11*pow(m55,2)*
         RSp461e*RSp453e - 1./81.*m11*pow(m55,2)*RSp461e*RS546pe*
         RSp463e + 1./162.*m11*pow(m55,2)*RSp461e*RSp561e - 1./486.*m11
         *pow(m55,2)*pow(RSp461e,2) - 1./162.*m11*pow(m55,2)*pow(
         RSp461e,2)*RS546pe + 1./162.*m11*pow(m55,2)*RSp451e*RSp463e + 
         1./162.*m11*pow(m55,2)*RSp451e*RSp461e + 5./81.*m11*m44*mp2*
         RSp561e*RSp563e + 1./162.*m11*m44*mp2*RSp561e*RSp463e + 1./162.
         *m11*m44*mp2*RSp561e*RSp453e + 5./162.*m11*m44*mp2*pow(
         RSp561e,2) + 1./162.*m11*m44*mp2*RSp461e*RSp563e + 1./81.*m11*
         m44*mp2*RSp461e*RSp463e + 1./162.*m11*m44*mp2*RSp461e*RSp561e
          + 1./162.*m11*m44*mp2*pow(RSp461e,2) + 1./162.*m11*m44*mp2*
         RSp451e*RSp563e + 1./81.*m11*m44*mp2*RSp451e*RSp453e + 1./162.
         *m11*m44*mp2*RSp451e*RSp561e + 1./162.*m11*m44*mp2*pow(
         RSp451e,2) );

      M13p6x23R +=  + Bbeps(mp2,mu2)*pi16 * (  - 1./81.*m11*m44*mp2*
         RSp564e*RSp561e*RSp563e - 1./162.*m11*m44*mp2*RSp564e*pow(
         RSp561e,2) - 1./81.*m11*m44*me2*RSe564p*RSp561e*RSp563e - 1./
         162.*m11*m44*me2*RSe564p*pow(RSp561e,2) - 4./243.*m11*m44*m66*
         RSp561e*RSp563e + 1./81.*m11*m44*m66*RSp561e*RSp453e - 2./243.
         *m11*m44*m66*pow(RSp561e,2) + 1./81.*m11*m44*m66*RSp451e*
         RSp563e - 4./243.*m11*m44*m66*RSp451e*RSp453e + 1./81.*m11*m44
         *m66*RSp451e*RSp561e - 2./243.*m11*m44*m66*pow(RSp451e,2) - 4./
         243.*m11*m44*m55*RSp561e*RSp563e + 1./81.*m11*m44*m55*RSp561e*
         RSp463e - 2./243.*m11*m44*m55*pow(RSp561e,2) + 1./81.*m11*m44*
         m55*RSp461e*RSp563e - 4./243.*m11*m44*m55*RSp461e*RSp463e + 1./
         81.*m11*m44*m55*RSp461e*RSp561e - 2./243.*m11*m44*m55*pow(
         RSp461e,2) - 1./243.*m11*pow(m44,2)*RSp561e*RSp563e + 1./162.*
         m11*pow(m44,2)*RSp561e*RSp463e + 1./162.*m11*pow(m44,2)*
         RSp561e*RSp453e - 1./81.*m11*pow(m44,2)*RSp561e*RS456pe*
         RSp563e );

      M13p6x23R +=  + Bbeps(mp2,mu2)*pi16 * (  - 1./486.*m11*pow(m44,2)
         *pow(RSp561e,2) - 1./162.*m11*pow(m44,2)*pow(RSp561e,2)*
         RS456pe + 1./162.*m11*pow(m44,2)*RSp461e*RSp563e + 1./162.*m11
         *pow(m44,2)*RSp461e*RSp561e + 1./162.*m11*pow(m44,2)*RSp451e*
         RSp563e + 1./162.*m11*pow(m44,2)*RSp451e*RSp561e );

      M13p6x23R +=  + hh(1,m11,m11,m13,m13,mu2) * ( 1./324.*pow(
         RD1456ep,2) + 2./81.*m33*RD1456ep*RS14563pe + 1./54.*m33*
         RS1456pee*RD1456ep + 1./54.*m33*RS1456ppe*RD1456ep - 1./54.*
         m33*RS145pe*RD1456ep - 1./54.*m33*RS146pe*RD1456ep - 1./54.*
         m33*RS156pe*RD1456ep + 5./32.*pow(m33,2) + 41./648.*pow(m33,2)
         *pow(RS14563pe,2) + 7./216.*pow(m33,2)*RS1456pee*RS14563pe + 5.
         /288.*pow(m33,2)*pow(RS1456pee,2) + 7./216.*pow(m33,2)*
         RS1456ppe*RS14563pe + 5./144.*pow(m33,2)*RS1456ppe*RS1456pee
          + 5./288.*pow(m33,2)*pow(RS1456ppe,2) - 7./216.*pow(m33,2)*
         RS145pe*RS14563pe - 5./144.*pow(m33,2)*RS145pe*RS1456pee - 5./
         144.*pow(m33,2)*RS145pe*RS1456ppe + 5./288.*pow(m33,2)*pow(
         RS145pe,2) - 7./216.*pow(m33,2)*RS146pe*RS14563pe - 5./144.*
         pow(m33,2)*RS146pe*RS1456pee - 5./144.*pow(m33,2)*RS146pe*
         RS1456ppe + 5./144.*pow(m33,2)*RS146pe*RS145pe + 5./288.*pow(
         m33,2)*pow(RS146pe,2) - 7./216.*pow(m33,2)*RS156pe*RS14563pe
          - 5./144.*pow(m33,2)*RS156pe*RS1456pee );

      M13p6x23R +=  + hh(1,m11,m11,m13,m13,mu2) * (  - 5./144.*pow(
         m33,2)*RS156pe*RS1456ppe + 5./144.*pow(m33,2)*RS156pe*RS145pe
          + 5./144.*pow(m33,2)*RS156pe*RS146pe + 5./288.*pow(m33,2)*
         pow(RS156pe,2) + 1./81.*m11*RD1456ep*RS14563pe + 1./54.*m11*
         RS1456pee*RD1456ep + 1./54.*m11*RS1456ppe*RD1456ep - 1./54.*
         m11*RS145pe*RD1456ep - 1./54.*m11*RS146pe*RD1456ep - 1./54.*
         m11*RS156pe*RD1456ep + 3./16.*m11*m33 + 25./324.*m11*m33*pow(
         RS14563pe,2) + 1./36.*m11*m33*RS1456pee*RS14563pe + 5./144.*
         m11*m33*pow(RS1456pee,2) + 1./36.*m11*m33*RS1456ppe*RS14563pe
          + 5./72.*m11*m33*RS1456ppe*RS1456pee + 5./144.*m11*m33*pow(
         RS1456ppe,2) - 1./36.*m11*m33*RS145pe*RS14563pe - 5./72.*m11*
         m33*RS145pe*RS1456pee - 5./72.*m11*m33*RS145pe*RS1456ppe + 5./
         144.*m11*m33*pow(RS145pe,2) - 1./36.*m11*m33*RS146pe*RS14563pe
          - 5./72.*m11*m33*RS146pe*RS1456pee - 5./72.*m11*m33*RS146pe*
         RS1456ppe + 5./72.*m11*m33*RS146pe*RS145pe + 5./144.*m11*m33*
         pow(RS146pe,2) );

      M13p6x23R +=  + hh(1,m11,m11,m13,m13,mu2) * (  - 1./36.*m11*m33*
         RS156pe*RS14563pe - 5./72.*m11*m33*RS156pe*RS1456pee - 5./72.*
         m11*m33*RS156pe*RS1456ppe + 5./72.*m11*m33*RS156pe*RS145pe + 5.
         /72.*m11*m33*RS156pe*RS146pe + 5./144.*m11*m33*pow(RS156pe,2)
          + 1./32.*pow(m11,2) + 17./648.*pow(m11,2)*pow(RS14563pe,2) - 
         1./216.*pow(m11,2)*RS1456pee*RS14563pe + 5./288.*pow(m11,2)*
         pow(RS1456pee,2) - 1./216.*pow(m11,2)*RS1456ppe*RS14563pe + 5./
         144.*pow(m11,2)*RS1456ppe*RS1456pee + 5./288.*pow(m11,2)*pow(
         RS1456ppe,2) + 1./216.*pow(m11,2)*RS145pe*RS14563pe - 5./144.*
         pow(m11,2)*RS145pe*RS1456pee - 5./144.*pow(m11,2)*RS145pe*
         RS1456ppe + 5./288.*pow(m11,2)*pow(RS145pe,2) + 1./216.*pow(
         m11,2)*RS146pe*RS14563pe - 5./144.*pow(m11,2)*RS146pe*
         RS1456pee - 5./144.*pow(m11,2)*RS146pe*RS1456ppe + 5./144.*
         pow(m11,2)*RS146pe*RS145pe + 5./288.*pow(m11,2)*pow(RS146pe,2)
          + 1./216.*pow(m11,2)*RS156pe*RS14563pe - 5./144.*pow(m11,2)*
         RS156pe*RS1456pee );

      M13p6x23R +=  + hh(1,m11,m11,m13,m13,mu2) * (  - 5./144.*pow(
         m11,2)*RS156pe*RS1456ppe + 5./144.*pow(m11,2)*RS156pe*RS145pe
          + 5./144.*pow(m11,2)*RS156pe*RS146pe + 5./288.*pow(m11,2)*
         pow(RS156pe,2) );

      M13p6x23R +=  + hh(1,m11,m13,m33,m13,mu2) * ( 1./162.*RD1456ep*
         RD3456ep + 2./81.*m33*RD3456ep*RS14563pe + 1./81.*m33*RD1456ep
         *RS34561pe + 25./324.*pow(m33,2)*RS14563pe*RS34561pe + 1./72.*
         pow(m33,2)*RS3456pee*RS14563pe + 1./72.*pow(m33,2)*RS3456ppe*
         RS14563pe - 1./72.*pow(m33,2)*RS345pe*RS14563pe - 1./72.*pow(
         m33,2)*RS346pe*RS14563pe - 1./72.*pow(m33,2)*RS356pe*RS14563pe
          + 1./72.*pow(m33,2)*RS1456pee*RS34561pe + 1./144.*pow(m33,2)*
         RS1456pee*RS3456pee + 1./144.*pow(m33,2)*RS1456pee*RS3456ppe
          - 1./144.*pow(m33,2)*RS1456pee*RS345pe - 1./144.*pow(m33,2)*
         RS1456pee*RS346pe - 1./144.*pow(m33,2)*RS1456pee*RS356pe + 1./
         72.*pow(m33,2)*RS1456ppe*RS34561pe + 1./144.*pow(m33,2)*
         RS1456ppe*RS3456pee + 1./144.*pow(m33,2)*RS1456ppe*RS3456ppe
          - 1./144.*pow(m33,2)*RS1456ppe*RS345pe - 1./144.*pow(m33,2)*
         RS1456ppe*RS346pe - 1./144.*pow(m33,2)*RS1456ppe*RS356pe - 1./
         72.*pow(m33,2)*RS145pe*RS34561pe - 1./144.*pow(m33,2)*RS145pe*
         RS3456pee );

      M13p6x23R +=  + hh(1,m11,m13,m33,m13,mu2) * (  - 1./144.*pow(
         m33,2)*RS145pe*RS3456ppe + 1./144.*pow(m33,2)*RS145pe*RS345pe
          + 1./144.*pow(m33,2)*RS145pe*RS346pe + 1./144.*pow(m33,2)*
         RS145pe*RS356pe - 1./72.*pow(m33,2)*RS146pe*RS34561pe - 1./144.
         *pow(m33,2)*RS146pe*RS3456pee - 1./144.*pow(m33,2)*RS146pe*
         RS3456ppe + 1./144.*pow(m33,2)*RS146pe*RS345pe + 1./144.*pow(
         m33,2)*RS146pe*RS346pe + 1./144.*pow(m33,2)*RS146pe*RS356pe - 
         1./72.*pow(m33,2)*RS156pe*RS34561pe - 1./144.*pow(m33,2)*
         RS156pe*RS3456pee - 1./144.*pow(m33,2)*RS156pe*RS3456ppe + 1./
         144.*pow(m33,2)*RS156pe*RS345pe + 1./144.*pow(m33,2)*RS156pe*
         RS346pe + 1./144.*pow(m33,2)*RS156pe*RS356pe + 1./81.*m11*
         RD3456ep*RS14563pe + 2./81.*m11*RD1456ep*RS34561pe + 29./162.*
         m11*m33*RS14563pe*RS34561pe + 1./36.*m11*m33*RS3456pee*
         RS14563pe + 1./36.*m11*m33*RS3456ppe*RS14563pe - 1./36.*m11*
         m33*RS345pe*RS14563pe - 1./36.*m11*m33*RS346pe*RS14563pe - 1./
         36.*m11*m33*RS356pe*RS14563pe );

      M13p6x23R +=  + hh(1,m11,m13,m33,m13,mu2) * ( 1./36.*m11*m33*
         RS1456pee*RS34561pe + 1./72.*m11*m33*RS1456pee*RS3456pee + 1./
         72.*m11*m33*RS1456pee*RS3456ppe - 1./72.*m11*m33*RS1456pee*
         RS345pe - 1./72.*m11*m33*RS1456pee*RS346pe - 1./72.*m11*m33*
         RS1456pee*RS356pe + 1./36.*m11*m33*RS1456ppe*RS34561pe + 1./72.
         *m11*m33*RS1456ppe*RS3456pee + 1./72.*m11*m33*RS1456ppe*
         RS3456ppe - 1./72.*m11*m33*RS1456ppe*RS345pe - 1./72.*m11*m33*
         RS1456ppe*RS346pe - 1./72.*m11*m33*RS1456ppe*RS356pe - 1./36.*
         m11*m33*RS145pe*RS34561pe - 1./72.*m11*m33*RS145pe*RS3456pee
          - 1./72.*m11*m33*RS145pe*RS3456ppe + 1./72.*m11*m33*RS145pe*
         RS345pe + 1./72.*m11*m33*RS145pe*RS346pe + 1./72.*m11*m33*
         RS145pe*RS356pe - 1./36.*m11*m33*RS146pe*RS34561pe - 1./72.*
         m11*m33*RS146pe*RS3456pee - 1./72.*m11*m33*RS146pe*RS3456ppe
          + 1./72.*m11*m33*RS146pe*RS345pe + 1./72.*m11*m33*RS146pe*
         RS346pe + 1./72.*m11*m33*RS146pe*RS356pe - 1./36.*m11*m33*
         RS156pe*RS34561pe );

      M13p6x23R +=  + hh(1,m11,m13,m33,m13,mu2) * (  - 1./72.*m11*m33*
         RS156pe*RS3456pee - 1./72.*m11*m33*RS156pe*RS3456ppe + 1./72.*
         m11*m33*RS156pe*RS345pe + 1./72.*m11*m33*RS156pe*RS346pe + 1./
         72.*m11*m33*RS156pe*RS356pe + 25./324.*pow(m11,2)*RS14563pe*
         RS34561pe + 1./72.*pow(m11,2)*RS3456pee*RS14563pe + 1./72.*
         pow(m11,2)*RS3456ppe*RS14563pe - 1./72.*pow(m11,2)*RS345pe*
         RS14563pe - 1./72.*pow(m11,2)*RS346pe*RS14563pe - 1./72.*pow(
         m11,2)*RS356pe*RS14563pe + 1./72.*pow(m11,2)*RS1456pee*
         RS34561pe + 1./144.*pow(m11,2)*RS1456pee*RS3456pee + 1./144.*
         pow(m11,2)*RS1456pee*RS3456ppe - 1./144.*pow(m11,2)*RS1456pee*
         RS345pe - 1./144.*pow(m11,2)*RS1456pee*RS346pe - 1./144.*pow(
         m11,2)*RS1456pee*RS356pe + 1./72.*pow(m11,2)*RS1456ppe*
         RS34561pe + 1./144.*pow(m11,2)*RS1456ppe*RS3456pee + 1./144.*
         pow(m11,2)*RS1456ppe*RS3456ppe - 1./144.*pow(m11,2)*RS1456ppe*
         RS345pe - 1./144.*pow(m11,2)*RS1456ppe*RS346pe - 1./144.*pow(
         m11,2)*RS1456ppe*RS356pe );

      M13p6x23R +=  + hh(1,m11,m13,m33,m13,mu2) * (  - 1./72.*pow(
         m11,2)*RS145pe*RS34561pe - 1./144.*pow(m11,2)*RS145pe*
         RS3456pee - 1./144.*pow(m11,2)*RS145pe*RS3456ppe + 1./144.*
         pow(m11,2)*RS145pe*RS345pe + 1./144.*pow(m11,2)*RS145pe*
         RS346pe + 1./144.*pow(m11,2)*RS145pe*RS356pe - 1./72.*pow(
         m11,2)*RS146pe*RS34561pe - 1./144.*pow(m11,2)*RS146pe*
         RS3456pee - 1./144.*pow(m11,2)*RS146pe*RS3456ppe + 1./144.*
         pow(m11,2)*RS146pe*RS345pe + 1./144.*pow(m11,2)*RS146pe*
         RS346pe + 1./144.*pow(m11,2)*RS146pe*RS356pe - 1./72.*pow(
         m11,2)*RS156pe*RS34561pe - 1./144.*pow(m11,2)*RS156pe*
         RS3456pee - 1./144.*pow(m11,2)*RS156pe*RS3456ppe + 1./144.*
         pow(m11,2)*RS156pe*RS345pe + 1./144.*pow(m11,2)*RS156pe*
         RS346pe + 1./144.*pow(m11,2)*RS156pe*RS356pe );

      M13p6x23R +=  + hh(1,m11,m13,me2,m13,mu2) * (  - 1./81.*me2*
         RD1456ep*RSe45613p + 1./162.*me2*RD1456ep*RSe45633p + 1./162.*
         me2*RSe45611p*RD1456ep + 1./81.*pow(me2,2)*RSe45613p*RS14563pe
          - 1./162.*pow(me2,2)*RSe45633p*RS14563pe + 1./162.*pow(me2,2)
         *RS1456pee*RSe45613p - 1./324.*pow(me2,2)*RS1456pee*RSe45633p
          + 1./162.*pow(me2,2)*RS1456ppe*RSe45613p - 1./324.*pow(me2,2)
         *RS1456ppe*RSe45633p - 1./162.*pow(me2,2)*RSe45611p*RS14563pe
          - 1./324.*pow(me2,2)*RSe45611p*RS1456pee - 1./324.*pow(me2,2)
         *RSe45611p*RS1456ppe - 1./162.*pow(me2,2)*RS145pe*RSe45613p + 
         1./324.*pow(me2,2)*RS145pe*RSe45633p + 1./324.*pow(me2,2)*
         RS145pe*RSe45611p - 1./162.*pow(me2,2)*RS146pe*RSe45613p + 1./
         324.*pow(me2,2)*RS146pe*RSe45633p + 1./324.*pow(me2,2)*RS146pe
         *RSe45611p - 1./162.*pow(me2,2)*RS156pe*RSe45613p + 1./324.*
         pow(me2,2)*RS156pe*RSe45633p + 1./324.*pow(me2,2)*RS156pe*
         RSe45611p + 2./81.*m33*RD1456ep*RSe45613p - 1./162.*m33*
         RD1456ep*RSe45633p );

      M13p6x23R +=  + hh(1,m11,m13,me2,m13,mu2) * (  - 1./54.*m33*
         RSe45611p*RD1456ep - 5./81.*m33*me2*RSe45613p*RS14563pe + 1./
         27.*m33*me2*RSe45633p*RS14563pe - 2./81.*m33*me2*RS1456pee*
         RSe45613p + 1./162.*m33*me2*RS1456pee*RSe45633p - 2./81.*m33*
         me2*RS1456ppe*RSe45613p + 1./162.*m33*me2*RS1456ppe*RSe45633p
          + 2./81.*m33*me2*RSe45611p*RS14563pe + 1./54.*m33*me2*
         RSe45611p*RS1456pee + 1./54.*m33*me2*RSe45611p*RS1456ppe + 2./
         81.*m33*me2*RS145pe*RSe45613p - 1./162.*m33*me2*RS145pe*
         RSe45633p - 1./54.*m33*me2*RS145pe*RSe45611p + 2./81.*m33*me2*
         RS146pe*RSe45613p - 1./162.*m33*me2*RS146pe*RSe45633p - 1./54.
         *m33*me2*RS146pe*RSe45611p + 2./81.*m33*me2*RS156pe*RSe45613p
          - 1./162.*m33*me2*RS156pe*RSe45633p - 1./54.*m33*me2*RS156pe*
         RSe45611p + 41./324.*pow(m33,2)*RSe45613p*RS14563pe - 29./648.
         *pow(m33,2)*RSe45633p*RS14563pe + 7./216.*pow(m33,2)*RS1456pee
         *RSe45613p - 13./1296.*pow(m33,2)*RS1456pee*RSe45633p + 7./216.
         *pow(m33,2)*RS1456ppe*RSe45613p );

      M13p6x23R +=  + hh(1,m11,m13,me2,m13,mu2) * (  - 13./1296.*pow(
         m33,2)*RS1456ppe*RSe45633p - 7./216.*pow(m33,2)*RSe45611p*
         RS14563pe - 5./144.*pow(m33,2)*RSe45611p*RS1456pee - 5./144.*
         pow(m33,2)*RSe45611p*RS1456ppe - 7./216.*pow(m33,2)*RS145pe*
         RSe45613p + 13./1296.*pow(m33,2)*RS145pe*RSe45633p + 5./144.*
         pow(m33,2)*RS145pe*RSe45611p - 7./216.*pow(m33,2)*RS146pe*
         RSe45613p + 13./1296.*pow(m33,2)*RS146pe*RSe45633p + 5./144.*
         pow(m33,2)*RS146pe*RSe45611p - 7./216.*pow(m33,2)*RS156pe*
         RSe45613p + 13./1296.*pow(m33,2)*RS156pe*RSe45633p + 5./144.*
         pow(m33,2)*RS156pe*RSe45611p + 2./81.*m11*RD1456ep*RSe45613p
          - 2./81.*m11*RSe45611p*RD1456ep - 1./27.*m11*me2*RSe45613p*
         RS14563pe + 1./81.*m11*me2*RSe45633p*RS14563pe - 2./81.*m11*
         me2*RS1456pee*RSe45613p - 2./81.*m11*me2*RS1456ppe*RSe45613p
          + 2./81.*m11*me2*RSe45611p*RS14563pe + 2./81.*m11*me2*
         RSe45611p*RS1456pee + 2./81.*m11*me2*RSe45611p*RS1456ppe + 2./
         81.*m11*me2*RS145pe*RSe45613p );

      M13p6x23R +=  + hh(1,m11,m13,me2,m13,mu2) * (  - 2./81.*m11*me2*
         RS145pe*RSe45611p + 2./81.*m11*me2*RS146pe*RSe45613p - 2./81.*
         m11*me2*RS146pe*RSe45611p + 2./81.*m11*me2*RS156pe*RSe45613p
          - 2./81.*m11*me2*RS156pe*RSe45611p + 35./162.*m11*m33*
         RSe45613p*RS14563pe - 13./324.*m11*m33*RSe45633p*RS14563pe + 
         17./324.*m11*m33*RS1456pee*RSe45613p - 1./72.*m11*m33*
         RS1456pee*RSe45633p + 17./324.*m11*m33*RS1456ppe*RSe45613p - 1.
         /72.*m11*m33*RS1456ppe*RSe45633p - 17./324.*m11*m33*RSe45611p*
         RS14563pe - 19./216.*m11*m33*RSe45611p*RS1456pee - 19./216.*
         m11*m33*RSe45611p*RS1456ppe - 17./324.*m11*m33*RS145pe*
         RSe45613p + 1./72.*m11*m33*RS145pe*RSe45633p + 19./216.*m11*
         m33*RS145pe*RSe45611p - 17./324.*m11*m33*RS146pe*RSe45613p + 1.
         /72.*m11*m33*RS146pe*RSe45633p + 19./216.*m11*m33*RS146pe*
         RSe45611p - 17./324.*m11*m33*RS156pe*RSe45613p + 1./72.*m11*
         m33*RS156pe*RSe45633p + 19./216.*m11*m33*RS156pe*RSe45611p + 
         25./324.*pow(m11,2)*RSe45613p*RS14563pe );

      M13p6x23R +=  + hh(1,m11,m13,me2,m13,mu2) * (  - 1./72.*pow(
         m11,2)*RSe45633p*RS14563pe + 1./72.*pow(m11,2)*RS1456pee*
         RSe45613p - 1./144.*pow(m11,2)*RS1456pee*RSe45633p + 1./72.*
         pow(m11,2)*RS1456ppe*RSe45613p - 1./144.*pow(m11,2)*RS1456ppe*
         RSe45633p - 1./72.*pow(m11,2)*RSe45611p*RS14563pe - 73./1296.*
         pow(m11,2)*RSe45611p*RS1456pee - 73./1296.*pow(m11,2)*
         RSe45611p*RS1456ppe - 1./72.*pow(m11,2)*RS145pe*RSe45613p + 1./
         144.*pow(m11,2)*RS145pe*RSe45633p + 73./1296.*pow(m11,2)*
         RS145pe*RSe45611p - 1./72.*pow(m11,2)*RS146pe*RSe45613p + 1./
         144.*pow(m11,2)*RS146pe*RSe45633p + 73./1296.*pow(m11,2)*
         RS146pe*RSe45611p - 1./72.*pow(m11,2)*RS156pe*RSe45613p + 1./
         144.*pow(m11,2)*RS156pe*RSe45633p + 73./1296.*pow(m11,2)*
         RS156pe*RSe45611p );

      M13p6x23R +=  + hh(1,m11,m13,mp2,m13,mu2) * (  - 1./81.*mp2*
         RD1456ep*RSp45613e + 1./162.*mp2*RD1456ep*RSp45633e + 1./162.*
         mp2*RSp45611e*RD1456ep + 1./81.*pow(mp2,2)*RSp45613e*RS14563pe
          - 1./162.*pow(mp2,2)*RSp45633e*RS14563pe + 1./162.*pow(mp2,2)
         *RS1456pee*RSp45613e - 1./324.*pow(mp2,2)*RS1456pee*RSp45633e
          + 1./162.*pow(mp2,2)*RS1456ppe*RSp45613e - 1./324.*pow(mp2,2)
         *RS1456ppe*RSp45633e - 1./162.*pow(mp2,2)*RSp45611e*RS14563pe
          - 1./324.*pow(mp2,2)*RSp45611e*RS1456pee - 1./324.*pow(mp2,2)
         *RSp45611e*RS1456ppe - 1./162.*pow(mp2,2)*RS145pe*RSp45613e + 
         1./324.*pow(mp2,2)*RS145pe*RSp45633e + 1./324.*pow(mp2,2)*
         RS145pe*RSp45611e - 1./162.*pow(mp2,2)*RS146pe*RSp45613e + 1./
         324.*pow(mp2,2)*RS146pe*RSp45633e + 1./324.*pow(mp2,2)*RS146pe
         *RSp45611e - 1./162.*pow(mp2,2)*RS156pe*RSp45613e + 1./324.*
         pow(mp2,2)*RS156pe*RSp45633e + 1./324.*pow(mp2,2)*RS156pe*
         RSp45611e + 2./81.*m33*RD1456ep*RSp45613e - 1./162.*m33*
         RD1456ep*RSp45633e );

      M13p6x23R +=  + hh(1,m11,m13,mp2,m13,mu2) * (  - 1./54.*m33*
         RSp45611e*RD1456ep - 5./81.*m33*mp2*RSp45613e*RS14563pe + 1./
         27.*m33*mp2*RSp45633e*RS14563pe - 2./81.*m33*mp2*RS1456pee*
         RSp45613e + 1./162.*m33*mp2*RS1456pee*RSp45633e - 2./81.*m33*
         mp2*RS1456ppe*RSp45613e + 1./162.*m33*mp2*RS1456ppe*RSp45633e
          + 2./81.*m33*mp2*RSp45611e*RS14563pe + 1./54.*m33*mp2*
         RSp45611e*RS1456pee + 1./54.*m33*mp2*RSp45611e*RS1456ppe + 2./
         81.*m33*mp2*RS145pe*RSp45613e - 1./162.*m33*mp2*RS145pe*
         RSp45633e - 1./54.*m33*mp2*RS145pe*RSp45611e + 2./81.*m33*mp2*
         RS146pe*RSp45613e - 1./162.*m33*mp2*RS146pe*RSp45633e - 1./54.
         *m33*mp2*RS146pe*RSp45611e + 2./81.*m33*mp2*RS156pe*RSp45613e
          - 1./162.*m33*mp2*RS156pe*RSp45633e - 1./54.*m33*mp2*RS156pe*
         RSp45611e + 41./324.*pow(m33,2)*RSp45613e*RS14563pe - 29./648.
         *pow(m33,2)*RSp45633e*RS14563pe + 7./216.*pow(m33,2)*RS1456pee
         *RSp45613e - 13./1296.*pow(m33,2)*RS1456pee*RSp45633e + 7./216.
         *pow(m33,2)*RS1456ppe*RSp45613e );

      M13p6x23R +=  + hh(1,m11,m13,mp2,m13,mu2) * (  - 13./1296.*pow(
         m33,2)*RS1456ppe*RSp45633e - 7./216.*pow(m33,2)*RSp45611e*
         RS14563pe - 5./144.*pow(m33,2)*RSp45611e*RS1456pee - 5./144.*
         pow(m33,2)*RSp45611e*RS1456ppe - 7./216.*pow(m33,2)*RS145pe*
         RSp45613e + 13./1296.*pow(m33,2)*RS145pe*RSp45633e + 5./144.*
         pow(m33,2)*RS145pe*RSp45611e - 7./216.*pow(m33,2)*RS146pe*
         RSp45613e + 13./1296.*pow(m33,2)*RS146pe*RSp45633e + 5./144.*
         pow(m33,2)*RS146pe*RSp45611e - 7./216.*pow(m33,2)*RS156pe*
         RSp45613e + 13./1296.*pow(m33,2)*RS156pe*RSp45633e + 5./144.*
         pow(m33,2)*RS156pe*RSp45611e + 2./81.*m11*RD1456ep*RSp45613e
          - 2./81.*m11*RSp45611e*RD1456ep - 1./27.*m11*mp2*RSp45613e*
         RS14563pe + 1./81.*m11*mp2*RSp45633e*RS14563pe - 2./81.*m11*
         mp2*RS1456pee*RSp45613e - 2./81.*m11*mp2*RS1456ppe*RSp45613e
          + 2./81.*m11*mp2*RSp45611e*RS14563pe + 2./81.*m11*mp2*
         RSp45611e*RS1456pee + 2./81.*m11*mp2*RSp45611e*RS1456ppe + 2./
         81.*m11*mp2*RS145pe*RSp45613e );

      M13p6x23R +=  + hh(1,m11,m13,mp2,m13,mu2) * (  - 2./81.*m11*mp2*
         RS145pe*RSp45611e + 2./81.*m11*mp2*RS146pe*RSp45613e - 2./81.*
         m11*mp2*RS146pe*RSp45611e + 2./81.*m11*mp2*RS156pe*RSp45613e
          - 2./81.*m11*mp2*RS156pe*RSp45611e + 35./162.*m11*m33*
         RSp45613e*RS14563pe - 13./324.*m11*m33*RSp45633e*RS14563pe + 
         17./324.*m11*m33*RS1456pee*RSp45613e - 1./72.*m11*m33*
         RS1456pee*RSp45633e + 17./324.*m11*m33*RS1456ppe*RSp45613e - 1.
         /72.*m11*m33*RS1456ppe*RSp45633e - 17./324.*m11*m33*RSp45611e*
         RS14563pe - 19./216.*m11*m33*RSp45611e*RS1456pee - 19./216.*
         m11*m33*RSp45611e*RS1456ppe - 17./324.*m11*m33*RS145pe*
         RSp45613e + 1./72.*m11*m33*RS145pe*RSp45633e + 19./216.*m11*
         m33*RS145pe*RSp45611e - 17./324.*m11*m33*RS146pe*RSp45613e + 1.
         /72.*m11*m33*RS146pe*RSp45633e + 19./216.*m11*m33*RS146pe*
         RSp45611e - 17./324.*m11*m33*RS156pe*RSp45613e + 1./72.*m11*
         m33*RS156pe*RSp45633e + 19./216.*m11*m33*RS156pe*RSp45611e + 
         25./324.*pow(m11,2)*RSp45613e*RS14563pe );

      M13p6x23R +=  + hh(1,m11,m13,mp2,m13,mu2) * (  - 1./72.*pow(
         m11,2)*RSp45633e*RS14563pe + 1./72.*pow(m11,2)*RS1456pee*
         RSp45613e - 1./144.*pow(m11,2)*RS1456pee*RSp45633e + 1./72.*
         pow(m11,2)*RS1456ppe*RSp45613e - 1./144.*pow(m11,2)*RS1456ppe*
         RSp45633e - 1./72.*pow(m11,2)*RSp45611e*RS14563pe - 73./1296.*
         pow(m11,2)*RSp45611e*RS1456pee - 73./1296.*pow(m11,2)*
         RSp45611e*RS1456ppe - 1./72.*pow(m11,2)*RS145pe*RSp45613e + 1./
         144.*pow(m11,2)*RS145pe*RSp45633e + 73./1296.*pow(m11,2)*
         RS145pe*RSp45611e - 1./72.*pow(m11,2)*RS146pe*RSp45613e + 1./
         144.*pow(m11,2)*RS146pe*RSp45633e + 73./1296.*pow(m11,2)*
         RS146pe*RSp45611e - 1./72.*pow(m11,2)*RS156pe*RSp45613e + 1./
         144.*pow(m11,2)*RS156pe*RSp45633e + 73./1296.*pow(m11,2)*
         RS156pe*RSp45611e );

      M13p6x23R +=  + hh(1,m11,m14,m34,m13,mu2) * ( 1./24.*m33*m44*
         RS14563pe + 1./48.*m33*m44*RS1456pee + 1./48.*m33*m44*
         RS1456ppe - 1./48.*m33*m44*RS145pe - 1./48.*m33*m44*RS146pe - 
         1./48.*m33*m44*RS156pe - 1./16.*pow(m33,2)*RS14563pe + 1./16.*
         pow(m33,2)*RS156pe + 1./24.*m11*m44*RS14563pe + 1./48.*m11*m44
         *RS1456pee + 1./48.*m11*m44*RS1456ppe - 1./48.*m11*m44*RS145pe
          - 1./48.*m11*m44*RS146pe - 1./48.*m11*m44*RS156pe - 1./6.*m11
         *m33*RS14563pe + 1./48.*m11*m33*RS1456pee + 1./48.*m11*m33*
         RS1456ppe - 1./48.*m11*m33*RS145pe - 1./48.*m11*m33*RS146pe + 
         3./16.*m11*m33*RS156pe - 5./48.*pow(m11,2)*RS14563pe + 1./48.*
         pow(m11,2)*RS1456pee + 1./48.*pow(m11,2)*RS1456ppe - 1./48.*
         pow(m11,2)*RS145pe - 1./48.*pow(m11,2)*RS146pe + 1./8.*pow(
         m11,2)*RS156pe );

      M13p6x23R +=  + hh(1,m11,m15,m35,m13,mu2) * ( 1./24.*m33*m55*
         RS14563pe + 1./48.*m33*m55*RS1456pee + 1./48.*m33*m55*
         RS1456ppe - 1./48.*m33*m55*RS145pe - 1./48.*m33*m55*RS146pe - 
         1./48.*m33*m55*RS156pe - 1./16.*pow(m33,2)*RS14563pe + 1./16.*
         pow(m33,2)*RS146pe + 1./24.*m11*m55*RS14563pe + 1./48.*m11*m55
         *RS1456pee + 1./48.*m11*m55*RS1456ppe - 1./48.*m11*m55*RS145pe
          - 1./48.*m11*m55*RS146pe - 1./48.*m11*m55*RS156pe - 1./6.*m11
         *m33*RS14563pe + 1./48.*m11*m33*RS1456pee + 1./48.*m11*m33*
         RS1456ppe - 1./48.*m11*m33*RS145pe + 3./16.*m11*m33*RS146pe - 
         1./48.*m11*m33*RS156pe - 5./48.*pow(m11,2)*RS14563pe + 1./48.*
         pow(m11,2)*RS1456pee + 1./48.*pow(m11,2)*RS1456ppe - 1./48.*
         pow(m11,2)*RS145pe + 1./8.*pow(m11,2)*RS146pe - 1./48.*pow(
         m11,2)*RS156pe );

      M13p6x23R +=  + hh(1,m11,m16,m36,m13,mu2) * ( 1./24.*m33*m66*
         RS14563pe + 1./48.*m33*m66*RS1456pee + 1./48.*m33*m66*
         RS1456ppe - 1./48.*m33*m66*RS145pe - 1./48.*m33*m66*RS146pe - 
         1./48.*m33*m66*RS156pe - 1./16.*pow(m33,2)*RS14563pe + 1./16.*
         pow(m33,2)*RS145pe + 1./24.*m11*m66*RS14563pe + 1./48.*m11*m66
         *RS1456pee + 1./48.*m11*m66*RS1456ppe - 1./48.*m11*m66*RS145pe
          - 1./48.*m11*m66*RS146pe - 1./48.*m11*m66*RS156pe - 1./6.*m11
         *m33*RS14563pe + 1./48.*m11*m33*RS1456pee + 1./48.*m11*m33*
         RS1456ppe + 3./16.*m11*m33*RS145pe - 1./48.*m11*m33*RS146pe - 
         1./48.*m11*m33*RS156pe - 5./48.*pow(m11,2)*RS14563pe + 1./48.*
         pow(m11,2)*RS1456pee + 1./48.*pow(m11,2)*RS1456ppe + 1./8.*
         pow(m11,2)*RS145pe - 1./48.*pow(m11,2)*RS146pe - 1./48.*pow(
         m11,2)*RS156pe );

      M13p6x23R +=  + hh(1,m13,m13,m13,m13,mu2) * ( 1./16.*pow(m33,2)
          + 1./8.*m11*m33 + 1./16.*pow(m11,2) );

      M13p6x23R +=  + hh(1,m13,m33,m33,m13,mu2) * ( 1./324.*pow(
         RD3456ep,2) + 1./81.*m33*RD3456ep*RS34561pe + 1./54.*m33*
         RS3456pee*RD3456ep + 1./54.*m33*RS3456ppe*RD3456ep - 1./54.*
         m33*RS345pe*RD3456ep - 1./54.*m33*RS346pe*RD3456ep - 1./54.*
         m33*RS356pe*RD3456ep + 1./32.*pow(m33,2) + 17./648.*pow(m33,2)
         *pow(RS34561pe,2) - 1./216.*pow(m33,2)*RS3456pee*RS34561pe + 5.
         /288.*pow(m33,2)*pow(RS3456pee,2) - 1./216.*pow(m33,2)*
         RS3456ppe*RS34561pe + 5./144.*pow(m33,2)*RS3456ppe*RS3456pee
          + 5./288.*pow(m33,2)*pow(RS3456ppe,2) + 1./216.*pow(m33,2)*
         RS345pe*RS34561pe - 5./144.*pow(m33,2)*RS345pe*RS3456pee - 5./
         144.*pow(m33,2)*RS345pe*RS3456ppe + 5./288.*pow(m33,2)*pow(
         RS345pe,2) + 1./216.*pow(m33,2)*RS346pe*RS34561pe - 5./144.*
         pow(m33,2)*RS346pe*RS3456pee - 5./144.*pow(m33,2)*RS346pe*
         RS3456ppe + 5./144.*pow(m33,2)*RS346pe*RS345pe + 5./288.*pow(
         m33,2)*pow(RS346pe,2) + 1./216.*pow(m33,2)*RS356pe*RS34561pe
          - 5./144.*pow(m33,2)*RS356pe*RS3456pee );

      M13p6x23R +=  + hh(1,m13,m33,m33,m13,mu2) * (  - 5./144.*pow(
         m33,2)*RS356pe*RS3456ppe + 5./144.*pow(m33,2)*RS356pe*RS345pe
          + 5./144.*pow(m33,2)*RS356pe*RS346pe + 5./288.*pow(m33,2)*
         pow(RS356pe,2) + 2./81.*m11*RD3456ep*RS34561pe + 1./54.*m11*
         RS3456pee*RD3456ep + 1./54.*m11*RS3456ppe*RD3456ep - 1./54.*
         m11*RS345pe*RD3456ep - 1./54.*m11*RS346pe*RD3456ep - 1./54.*
         m11*RS356pe*RD3456ep + 3./16.*m11*m33 + 25./324.*m11*m33*pow(
         RS34561pe,2) + 1./36.*m11*m33*RS3456pee*RS34561pe + 5./144.*
         m11*m33*pow(RS3456pee,2) + 1./36.*m11*m33*RS3456ppe*RS34561pe
          + 5./72.*m11*m33*RS3456ppe*RS3456pee + 5./144.*m11*m33*pow(
         RS3456ppe,2) - 1./36.*m11*m33*RS345pe*RS34561pe - 5./72.*m11*
         m33*RS345pe*RS3456pee - 5./72.*m11*m33*RS345pe*RS3456ppe + 5./
         144.*m11*m33*pow(RS345pe,2) - 1./36.*m11*m33*RS346pe*RS34561pe
          - 5./72.*m11*m33*RS346pe*RS3456pee - 5./72.*m11*m33*RS346pe*
         RS3456ppe + 5./72.*m11*m33*RS346pe*RS345pe + 5./144.*m11*m33*
         pow(RS346pe,2) );

      M13p6x23R +=  + hh(1,m13,m33,m33,m13,mu2) * (  - 1./36.*m11*m33*
         RS356pe*RS34561pe - 5./72.*m11*m33*RS356pe*RS3456pee - 5./72.*
         m11*m33*RS356pe*RS3456ppe + 5./72.*m11*m33*RS356pe*RS345pe + 5.
         /72.*m11*m33*RS356pe*RS346pe + 5./144.*m11*m33*pow(RS356pe,2)
          + 5./32.*pow(m11,2) + 41./648.*pow(m11,2)*pow(RS34561pe,2) + 
         7./216.*pow(m11,2)*RS3456pee*RS34561pe + 5./288.*pow(m11,2)*
         pow(RS3456pee,2) + 7./216.*pow(m11,2)*RS3456ppe*RS34561pe + 5./
         144.*pow(m11,2)*RS3456ppe*RS3456pee + 5./288.*pow(m11,2)*pow(
         RS3456ppe,2) - 7./216.*pow(m11,2)*RS345pe*RS34561pe - 5./144.*
         pow(m11,2)*RS345pe*RS3456pee - 5./144.*pow(m11,2)*RS345pe*
         RS3456ppe + 5./288.*pow(m11,2)*pow(RS345pe,2) - 7./216.*pow(
         m11,2)*RS346pe*RS34561pe - 5./144.*pow(m11,2)*RS346pe*
         RS3456pee - 5./144.*pow(m11,2)*RS346pe*RS3456ppe + 5./144.*
         pow(m11,2)*RS346pe*RS345pe + 5./288.*pow(m11,2)*pow(RS346pe,2)
          - 7./216.*pow(m11,2)*RS356pe*RS34561pe - 5./144.*pow(m11,2)*
         RS356pe*RS3456pee );

      M13p6x23R +=  + hh(1,m13,m33,m33,m13,mu2) * (  - 5./144.*pow(
         m11,2)*RS356pe*RS3456ppe + 5./144.*pow(m11,2)*RS356pe*RS345pe
          + 5./144.*pow(m11,2)*RS356pe*RS346pe + 5./288.*pow(m11,2)*
         pow(RS356pe,2) );

      M13p6x23R +=  + hh(1,m13,m33,me2,m13,mu2) * (  - 1./81.*me2*
         RD3456ep*RSe45613p + 1./162.*me2*RSe45633p*RD3456ep + 1./162.*
         me2*RSe45611p*RD3456ep + 1./81.*pow(me2,2)*RSe45613p*RS34561pe
          + 1./162.*pow(me2,2)*RS3456pee*RSe45613p + 1./162.*pow(me2,2)
         *RS3456ppe*RSe45613p - 1./162.*pow(me2,2)*RSe45633p*RS34561pe
          - 1./324.*pow(me2,2)*RSe45633p*RS3456pee - 1./324.*pow(me2,2)
         *RSe45633p*RS3456ppe - 1./162.*pow(me2,2)*RS345pe*RSe45613p + 
         1./324.*pow(me2,2)*RS345pe*RSe45633p - 1./162.*pow(me2,2)*
         RS346pe*RSe45613p + 1./324.*pow(me2,2)*RS346pe*RSe45633p - 1./
         162.*pow(me2,2)*RS356pe*RSe45613p + 1./324.*pow(me2,2)*RS356pe
         *RSe45633p - 1./162.*pow(me2,2)*RSe45611p*RS34561pe - 1./324.*
         pow(me2,2)*RSe45611p*RS3456pee - 1./324.*pow(me2,2)*RSe45611p*
         RS3456ppe + 1./324.*pow(me2,2)*RSe45611p*RS345pe + 1./324.*
         pow(me2,2)*RSe45611p*RS346pe + 1./324.*pow(me2,2)*RSe45611p*
         RS356pe + 2./81.*m33*RD3456ep*RSe45613p - 2./81.*m33*RSe45633p
         *RD3456ep );

      M13p6x23R +=  + hh(1,m13,m33,me2,m13,mu2) * (  - 1./27.*m33*me2*
         RSe45613p*RS34561pe - 2./81.*m33*me2*RS3456pee*RSe45613p - 2./
         81.*m33*me2*RS3456ppe*RSe45613p + 2./81.*m33*me2*RSe45633p*
         RS34561pe + 2./81.*m33*me2*RSe45633p*RS3456pee + 2./81.*m33*
         me2*RSe45633p*RS3456ppe + 2./81.*m33*me2*RS345pe*RSe45613p - 2.
         /81.*m33*me2*RS345pe*RSe45633p + 2./81.*m33*me2*RS346pe*
         RSe45613p - 2./81.*m33*me2*RS346pe*RSe45633p + 2./81.*m33*me2*
         RS356pe*RSe45613p - 2./81.*m33*me2*RS356pe*RSe45633p + 1./81.*
         m33*me2*RSe45611p*RS34561pe + 25./324.*pow(m33,2)*RSe45613p*
         RS34561pe + 1./72.*pow(m33,2)*RS3456pee*RSe45613p + 1./72.*
         pow(m33,2)*RS3456ppe*RSe45613p - 1./72.*pow(m33,2)*RSe45633p*
         RS34561pe - 73./1296.*pow(m33,2)*RSe45633p*RS3456pee - 73./
         1296.*pow(m33,2)*RSe45633p*RS3456ppe - 1./72.*pow(m33,2)*
         RS345pe*RSe45613p + 73./1296.*pow(m33,2)*RS345pe*RSe45633p - 1.
         /72.*pow(m33,2)*RS346pe*RSe45613p + 73./1296.*pow(m33,2)*
         RS346pe*RSe45633p );

      M13p6x23R +=  + hh(1,m13,m33,me2,m13,mu2) * (  - 1./72.*pow(
         m33,2)*RS356pe*RSe45613p + 73./1296.*pow(m33,2)*RS356pe*
         RSe45633p - 1./72.*pow(m33,2)*RSe45611p*RS34561pe - 1./144.*
         pow(m33,2)*RSe45611p*RS3456pee - 1./144.*pow(m33,2)*RSe45611p*
         RS3456ppe + 1./144.*pow(m33,2)*RSe45611p*RS345pe + 1./144.*
         pow(m33,2)*RSe45611p*RS346pe + 1./144.*pow(m33,2)*RSe45611p*
         RS356pe + 2./81.*m11*RD3456ep*RSe45613p - 1./54.*m11*RSe45633p
         *RD3456ep - 1./162.*m11*RSe45611p*RD3456ep - 5./81.*m11*me2*
         RSe45613p*RS34561pe - 2./81.*m11*me2*RS3456pee*RSe45613p - 2./
         81.*m11*me2*RS3456ppe*RSe45613p + 2./81.*m11*me2*RSe45633p*
         RS34561pe + 1./54.*m11*me2*RSe45633p*RS3456pee + 1./54.*m11*
         me2*RSe45633p*RS3456ppe + 2./81.*m11*me2*RS345pe*RSe45613p - 1.
         /54.*m11*me2*RS345pe*RSe45633p + 2./81.*m11*me2*RS346pe*
         RSe45613p - 1./54.*m11*me2*RS346pe*RSe45633p + 2./81.*m11*me2*
         RS356pe*RSe45613p - 1./54.*m11*me2*RS356pe*RSe45633p + 1./27.*
         m11*me2*RSe45611p*RS34561pe );

      M13p6x23R +=  + hh(1,m13,m33,me2,m13,mu2) * ( 1./162.*m11*me2*
         RSe45611p*RS3456pee + 1./162.*m11*me2*RSe45611p*RS3456ppe - 1./
         162.*m11*me2*RSe45611p*RS345pe - 1./162.*m11*me2*RSe45611p*
         RS346pe - 1./162.*m11*me2*RSe45611p*RS356pe + 35./162.*m11*m33
         *RSe45613p*RS34561pe + 17./324.*m11*m33*RS3456pee*RSe45613p + 
         17./324.*m11*m33*RS3456ppe*RSe45613p - 17./324.*m11*m33*
         RSe45633p*RS34561pe - 19./216.*m11*m33*RSe45633p*RS3456pee - 
         19./216.*m11*m33*RSe45633p*RS3456ppe - 17./324.*m11*m33*
         RS345pe*RSe45613p + 19./216.*m11*m33*RS345pe*RSe45633p - 17./
         324.*m11*m33*RS346pe*RSe45613p + 19./216.*m11*m33*RS346pe*
         RSe45633p - 17./324.*m11*m33*RS356pe*RSe45613p + 19./216.*m11*
         m33*RS356pe*RSe45633p - 13./324.*m11*m33*RSe45611p*RS34561pe
          - 1./72.*m11*m33*RSe45611p*RS3456pee - 1./72.*m11*m33*
         RSe45611p*RS3456ppe + 1./72.*m11*m33*RSe45611p*RS345pe + 1./72.
         *m11*m33*RSe45611p*RS346pe + 1./72.*m11*m33*RSe45611p*RS356pe
          + 41./324.*pow(m11,2)*RSe45613p*RS34561pe );

      M13p6x23R +=  + hh(1,m13,m33,me2,m13,mu2) * ( 7./216.*pow(m11,2)*
         RS3456pee*RSe45613p + 7./216.*pow(m11,2)*RS3456ppe*RSe45613p
          - 7./216.*pow(m11,2)*RSe45633p*RS34561pe - 5./144.*pow(m11,2)
         *RSe45633p*RS3456pee - 5./144.*pow(m11,2)*RSe45633p*RS3456ppe
          - 7./216.*pow(m11,2)*RS345pe*RSe45613p + 5./144.*pow(m11,2)*
         RS345pe*RSe45633p - 7./216.*pow(m11,2)*RS346pe*RSe45613p + 5./
         144.*pow(m11,2)*RS346pe*RSe45633p - 7./216.*pow(m11,2)*RS356pe
         *RSe45613p + 5./144.*pow(m11,2)*RS356pe*RSe45633p - 29./648.*
         pow(m11,2)*RSe45611p*RS34561pe - 13./1296.*pow(m11,2)*
         RSe45611p*RS3456pee - 13./1296.*pow(m11,2)*RSe45611p*RS3456ppe
          + 13./1296.*pow(m11,2)*RSe45611p*RS345pe + 13./1296.*pow(
         m11,2)*RSe45611p*RS346pe + 13./1296.*pow(m11,2)*RSe45611p*
         RS356pe );

      M13p6x23R +=  + hh(1,m13,m33,mp2,m13,mu2) * (  - 1./81.*mp2*
         RD3456ep*RSp45613e + 1./162.*mp2*RSp45633e*RD3456ep + 1./162.*
         mp2*RSp45611e*RD3456ep + 1./81.*pow(mp2,2)*RSp45613e*RS34561pe
          + 1./162.*pow(mp2,2)*RS3456pee*RSp45613e + 1./162.*pow(mp2,2)
         *RS3456ppe*RSp45613e - 1./162.*pow(mp2,2)*RSp45633e*RS34561pe
          - 1./324.*pow(mp2,2)*RSp45633e*RS3456pee - 1./324.*pow(mp2,2)
         *RSp45633e*RS3456ppe - 1./162.*pow(mp2,2)*RS345pe*RSp45613e + 
         1./324.*pow(mp2,2)*RS345pe*RSp45633e - 1./162.*pow(mp2,2)*
         RS346pe*RSp45613e + 1./324.*pow(mp2,2)*RS346pe*RSp45633e - 1./
         162.*pow(mp2,2)*RS356pe*RSp45613e + 1./324.*pow(mp2,2)*RS356pe
         *RSp45633e - 1./162.*pow(mp2,2)*RSp45611e*RS34561pe - 1./324.*
         pow(mp2,2)*RSp45611e*RS3456pee - 1./324.*pow(mp2,2)*RSp45611e*
         RS3456ppe + 1./324.*pow(mp2,2)*RSp45611e*RS345pe + 1./324.*
         pow(mp2,2)*RSp45611e*RS346pe + 1./324.*pow(mp2,2)*RSp45611e*
         RS356pe + 2./81.*m33*RD3456ep*RSp45613e - 2./81.*m33*RSp45633e
         *RD3456ep );

      M13p6x23R +=  + hh(1,m13,m33,mp2,m13,mu2) * (  - 1./27.*m33*mp2*
         RSp45613e*RS34561pe - 2./81.*m33*mp2*RS3456pee*RSp45613e - 2./
         81.*m33*mp2*RS3456ppe*RSp45613e + 2./81.*m33*mp2*RSp45633e*
         RS34561pe + 2./81.*m33*mp2*RSp45633e*RS3456pee + 2./81.*m33*
         mp2*RSp45633e*RS3456ppe + 2./81.*m33*mp2*RS345pe*RSp45613e - 2.
         /81.*m33*mp2*RS345pe*RSp45633e + 2./81.*m33*mp2*RS346pe*
         RSp45613e - 2./81.*m33*mp2*RS346pe*RSp45633e + 2./81.*m33*mp2*
         RS356pe*RSp45613e - 2./81.*m33*mp2*RS356pe*RSp45633e + 1./81.*
         m33*mp2*RSp45611e*RS34561pe + 25./324.*pow(m33,2)*RSp45613e*
         RS34561pe + 1./72.*pow(m33,2)*RS3456pee*RSp45613e + 1./72.*
         pow(m33,2)*RS3456ppe*RSp45613e - 1./72.*pow(m33,2)*RSp45633e*
         RS34561pe - 73./1296.*pow(m33,2)*RSp45633e*RS3456pee - 73./
         1296.*pow(m33,2)*RSp45633e*RS3456ppe - 1./72.*pow(m33,2)*
         RS345pe*RSp45613e + 73./1296.*pow(m33,2)*RS345pe*RSp45633e - 1.
         /72.*pow(m33,2)*RS346pe*RSp45613e + 73./1296.*pow(m33,2)*
         RS346pe*RSp45633e );

      M13p6x23R +=  + hh(1,m13,m33,mp2,m13,mu2) * (  - 1./72.*pow(
         m33,2)*RS356pe*RSp45613e + 73./1296.*pow(m33,2)*RS356pe*
         RSp45633e - 1./72.*pow(m33,2)*RSp45611e*RS34561pe - 1./144.*
         pow(m33,2)*RSp45611e*RS3456pee - 1./144.*pow(m33,2)*RSp45611e*
         RS3456ppe + 1./144.*pow(m33,2)*RSp45611e*RS345pe + 1./144.*
         pow(m33,2)*RSp45611e*RS346pe + 1./144.*pow(m33,2)*RSp45611e*
         RS356pe + 2./81.*m11*RD3456ep*RSp45613e - 1./54.*m11*RSp45633e
         *RD3456ep - 1./162.*m11*RSp45611e*RD3456ep - 5./81.*m11*mp2*
         RSp45613e*RS34561pe - 2./81.*m11*mp2*RS3456pee*RSp45613e - 2./
         81.*m11*mp2*RS3456ppe*RSp45613e + 2./81.*m11*mp2*RSp45633e*
         RS34561pe + 1./54.*m11*mp2*RSp45633e*RS3456pee + 1./54.*m11*
         mp2*RSp45633e*RS3456ppe + 2./81.*m11*mp2*RS345pe*RSp45613e - 1.
         /54.*m11*mp2*RS345pe*RSp45633e + 2./81.*m11*mp2*RS346pe*
         RSp45613e - 1./54.*m11*mp2*RS346pe*RSp45633e + 2./81.*m11*mp2*
         RS356pe*RSp45613e - 1./54.*m11*mp2*RS356pe*RSp45633e + 1./27.*
         m11*mp2*RSp45611e*RS34561pe );

      M13p6x23R +=  + hh(1,m13,m33,mp2,m13,mu2) * ( 1./162.*m11*mp2*
         RSp45611e*RS3456pee + 1./162.*m11*mp2*RSp45611e*RS3456ppe - 1./
         162.*m11*mp2*RSp45611e*RS345pe - 1./162.*m11*mp2*RSp45611e*
         RS346pe - 1./162.*m11*mp2*RSp45611e*RS356pe + 35./162.*m11*m33
         *RSp45613e*RS34561pe + 17./324.*m11*m33*RS3456pee*RSp45613e + 
         17./324.*m11*m33*RS3456ppe*RSp45613e - 17./324.*m11*m33*
         RSp45633e*RS34561pe - 19./216.*m11*m33*RSp45633e*RS3456pee - 
         19./216.*m11*m33*RSp45633e*RS3456ppe - 17./324.*m11*m33*
         RS345pe*RSp45613e + 19./216.*m11*m33*RS345pe*RSp45633e - 17./
         324.*m11*m33*RS346pe*RSp45613e + 19./216.*m11*m33*RS346pe*
         RSp45633e - 17./324.*m11*m33*RS356pe*RSp45613e + 19./216.*m11*
         m33*RS356pe*RSp45633e - 13./324.*m11*m33*RSp45611e*RS34561pe
          - 1./72.*m11*m33*RSp45611e*RS3456pee - 1./72.*m11*m33*
         RSp45611e*RS3456ppe + 1./72.*m11*m33*RSp45611e*RS345pe + 1./72.
         *m11*m33*RSp45611e*RS346pe + 1./72.*m11*m33*RSp45611e*RS356pe
          + 41./324.*pow(m11,2)*RSp45613e*RS34561pe );

      M13p6x23R +=  + hh(1,m13,m33,mp2,m13,mu2) * ( 7./216.*pow(m11,2)*
         RS3456pee*RSp45613e + 7./216.*pow(m11,2)*RS3456ppe*RSp45613e
          - 7./216.*pow(m11,2)*RSp45633e*RS34561pe - 5./144.*pow(m11,2)
         *RSp45633e*RS3456pee - 5./144.*pow(m11,2)*RSp45633e*RS3456ppe
          - 7./216.*pow(m11,2)*RS345pe*RSp45613e + 5./144.*pow(m11,2)*
         RS345pe*RSp45633e - 7./216.*pow(m11,2)*RS346pe*RSp45613e + 5./
         144.*pow(m11,2)*RS346pe*RSp45633e - 7./216.*pow(m11,2)*RS356pe
         *RSp45613e + 5./144.*pow(m11,2)*RS356pe*RSp45633e - 29./648.*
         pow(m11,2)*RSp45611e*RS34561pe - 13./1296.*pow(m11,2)*
         RSp45611e*RS3456pee - 13./1296.*pow(m11,2)*RSp45611e*RS3456ppe
          + 13./1296.*pow(m11,2)*RSp45611e*RS345pe + 13./1296.*pow(
         m11,2)*RSp45611e*RS346pe + 13./1296.*pow(m11,2)*RSp45611e*
         RS356pe );

      M13p6x23R +=  + hh(1,m13,me2,me2,m13,mu2) * ( 2./81.*pow(me2,2)*
         pow(RSe45613p,2) - 2./81.*pow(me2,2)*RSe45633p*RSe45613p + 1./
         162.*pow(me2,2)*pow(RSe45633p,2) - 2./81.*pow(me2,2)*RSe45611p
         *RSe45613p + 1./81.*pow(me2,2)*RSe45611p*RSe45633p + 1./162.*
         pow(me2,2)*pow(RSe45611p,2) - 5./81.*m33*me2*pow(RSe45613p,2)
          + 2./27.*m33*me2*RSe45633p*RSe45613p - 5./162.*m33*me2*pow(
         RSe45633p,2) + 4./81.*m33*me2*RSe45611p*RSe45613p - 1./81.*m33
         *me2*RSe45611p*RSe45633p - 1./54.*m33*me2*pow(RSe45611p,2) + 
         41./648.*pow(m33,2)*pow(RSe45613p,2) - 29./648.*pow(m33,2)*
         RSe45633p*RSe45613p + 109./2592.*pow(m33,2)*pow(RSe45633p,2)
          - 7./216.*pow(m33,2)*RSe45611p*RSe45613p + 13./1296.*pow(
         m33,2)*RSe45611p*RSe45633p + 5./288.*pow(m33,2)*pow(
         RSe45611p,2) - 5./81.*m11*me2*pow(RSe45613p,2) + 4./81.*m11*
         me2*RSe45633p*RSe45613p - 1./54.*m11*me2*pow(RSe45633p,2) + 2./
         27.*m11*me2*RSe45611p*RSe45613p - 1./81.*m11*me2*RSe45611p*
         RSe45633p );

      M13p6x23R +=  + hh(1,m13,me2,me2,m13,mu2) * (  - 5./162.*m11*me2*
         pow(RSe45611p,2) + 5./36.*m11*m33*pow(RSe45613p,2) - 25./324.*
         m11*m33*RSe45633p*RSe45613p + 23./432.*m11*m33*pow(
         RSe45633p,2) - 25./324.*m11*m33*RSe45611p*RSe45613p + 13./648.
         *m11*m33*RSe45611p*RSe45633p + 23./432.*m11*m33*pow(
         RSe45611p,2) + 41./648.*pow(m11,2)*pow(RSe45613p,2) - 7./216.*
         pow(m11,2)*RSe45633p*RSe45613p + 5./288.*pow(m11,2)*pow(
         RSe45633p,2) - 29./648.*pow(m11,2)*RSe45611p*RSe45613p + 13./
         1296.*pow(m11,2)*RSe45611p*RSe45633p + 109./2592.*pow(m11,2)*
         pow(RSe45611p,2) );

      M13p6x23R +=  + hh(1,m13,me2,mp2,m13,mu2) * ( 1./81.*pow(mp2,2)*
         RSp45613e*RSe45613p - 1./162.*pow(mp2,2)*RSe45633p*RSp45613e
          - 1./162.*pow(mp2,2)*RSp45633e*RSe45613p + 1./324.*pow(mp2,2)
         *RSp45633e*RSe45633p - 1./162.*pow(mp2,2)*RSe45611p*RSp45613e
          + 1./324.*pow(mp2,2)*RSe45611p*RSp45633e - 1./162.*pow(mp2,2)
         *RSp45611e*RSe45613p + 1./324.*pow(mp2,2)*RSp45611e*RSe45633p
          + 1./324.*pow(mp2,2)*RSp45611e*RSe45611p + 2./81.*me2*mp2*
         RSp45613e*RSe45613p - 1./81.*me2*mp2*RSe45633p*RSp45613e - 1./
         81.*me2*mp2*RSp45633e*RSe45613p + 1./162.*me2*mp2*RSp45633e*
         RSe45633p - 1./81.*me2*mp2*RSe45611p*RSp45613e + 1./162.*me2*
         mp2*RSe45611p*RSp45633e - 1./81.*me2*mp2*RSp45611e*RSe45613p
          + 1./162.*me2*mp2*RSp45611e*RSe45633p + 1./162.*me2*mp2*
         RSp45611e*RSe45611p + 1./81.*pow(me2,2)*RSp45613e*RSe45613p - 
         1./162.*pow(me2,2)*RSe45633p*RSp45613e - 1./162.*pow(me2,2)*
         RSp45633e*RSe45613p + 1./324.*pow(me2,2)*RSp45633e*RSe45633p
          - 1./162.*pow(me2,2)*RSe45611p*RSp45613e );

      M13p6x23R +=  + hh(1,m13,me2,mp2,m13,mu2) * ( 1./324.*pow(me2,2)*
         RSe45611p*RSp45633e - 1./162.*pow(me2,2)*RSp45611e*RSe45613p
          + 1./324.*pow(me2,2)*RSp45611e*RSe45633p + 1./324.*pow(me2,2)
         *RSp45611e*RSe45611p - 5./81.*m33*mp2*RSp45613e*RSe45613p + 1./
         27.*m33*mp2*RSe45633p*RSp45613e + 1./27.*m33*mp2*RSp45633e*
         RSe45613p - 5./162.*m33*mp2*RSp45633e*RSe45633p + 2./81.*m33*
         mp2*RSe45611p*RSp45613e - 1./162.*m33*mp2*RSe45611p*RSp45633e
          + 2./81.*m33*mp2*RSp45611e*RSe45613p - 1./162.*m33*mp2*
         RSp45611e*RSe45633p - 1./54.*m33*mp2*RSp45611e*RSe45611p - 5./
         81.*m33*me2*RSp45613e*RSe45613p + 1./27.*m33*me2*RSe45633p*
         RSp45613e + 1./27.*m33*me2*RSp45633e*RSe45613p - 5./162.*m33*
         me2*RSp45633e*RSe45633p + 2./81.*m33*me2*RSe45611p*RSp45613e
          - 1./162.*m33*me2*RSe45611p*RSp45633e + 2./81.*m33*me2*
         RSp45611e*RSe45613p - 1./162.*m33*me2*RSp45611e*RSe45633p - 1./
         54.*m33*me2*RSp45611e*RSe45611p + 41./324.*pow(m33,2)*
         RSp45613e*RSe45613p );

      M13p6x23R +=  + hh(1,m13,me2,mp2,m13,mu2) * (  - 29./648.*pow(
         m33,2)*RSe45633p*RSp45613e - 29./648.*pow(m33,2)*RSp45633e*
         RSe45613p + 109./1296.*pow(m33,2)*RSp45633e*RSe45633p - 7./216.
         *pow(m33,2)*RSe45611p*RSp45613e + 13./1296.*pow(m33,2)*
         RSe45611p*RSp45633e - 7./216.*pow(m33,2)*RSp45611e*RSe45613p
          + 13./1296.*pow(m33,2)*RSp45611e*RSe45633p + 5./144.*pow(
         m33,2)*RSp45611e*RSe45611p - 5./81.*m11*mp2*RSp45613e*
         RSe45613p + 2./81.*m11*mp2*RSe45633p*RSp45613e + 2./81.*m11*
         mp2*RSp45633e*RSe45613p - 1./54.*m11*mp2*RSp45633e*RSe45633p
          + 1./27.*m11*mp2*RSe45611p*RSp45613e - 1./162.*m11*mp2*
         RSe45611p*RSp45633e + 1./27.*m11*mp2*RSp45611e*RSe45613p - 1./
         162.*m11*mp2*RSp45611e*RSe45633p - 5./162.*m11*mp2*RSp45611e*
         RSe45611p - 5./81.*m11*me2*RSp45613e*RSe45613p + 2./81.*m11*
         me2*RSe45633p*RSp45613e + 2./81.*m11*me2*RSp45633e*RSe45613p
          - 1./54.*m11*me2*RSp45633e*RSe45633p + 1./27.*m11*me2*
         RSe45611p*RSp45613e );

      M13p6x23R +=  + hh(1,m13,me2,mp2,m13,mu2) * (  - 1./162.*m11*me2*
         RSe45611p*RSp45633e + 1./27.*m11*me2*RSp45611e*RSe45613p - 1./
         162.*m11*me2*RSp45611e*RSe45633p - 5./162.*m11*me2*RSp45611e*
         RSe45611p + 5./18.*m11*m33*RSp45613e*RSe45613p - 25./324.*m11*
         m33*RSe45633p*RSp45613e - 25./324.*m11*m33*RSp45633e*RSe45613p
          + 23./216.*m11*m33*RSp45633e*RSe45633p - 25./324.*m11*m33*
         RSe45611p*RSp45613e + 13./648.*m11*m33*RSe45611p*RSp45633e - 
         25./324.*m11*m33*RSp45611e*RSe45613p + 13./648.*m11*m33*
         RSp45611e*RSe45633p + 23./216.*m11*m33*RSp45611e*RSe45611p + 
         41./324.*pow(m11,2)*RSp45613e*RSe45613p - 7./216.*pow(m11,2)*
         RSe45633p*RSp45613e - 7./216.*pow(m11,2)*RSp45633e*RSe45613p
          + 5./144.*pow(m11,2)*RSp45633e*RSe45633p - 29./648.*pow(
         m11,2)*RSe45611p*RSp45613e + 13./1296.*pow(m11,2)*RSe45611p*
         RSp45633e - 29./648.*pow(m11,2)*RSp45611e*RSe45613p + 13./1296.
         *pow(m11,2)*RSp45611e*RSe45633p + 109./1296.*pow(m11,2)*
         RSp45611e*RSe45611p );

      M13p6x23R +=  + hh(1,m13,mp2,mp2,m13,mu2) * ( 2./81.*pow(mp2,2)*
         pow(RSp45613e,2) - 2./81.*pow(mp2,2)*RSp45633e*RSp45613e + 1./
         162.*pow(mp2,2)*pow(RSp45633e,2) - 2./81.*pow(mp2,2)*RSp45611e
         *RSp45613e + 1./81.*pow(mp2,2)*RSp45611e*RSp45633e + 1./162.*
         pow(mp2,2)*pow(RSp45611e,2) - 5./81.*m33*mp2*pow(RSp45613e,2)
          + 2./27.*m33*mp2*RSp45633e*RSp45613e - 5./162.*m33*mp2*pow(
         RSp45633e,2) + 4./81.*m33*mp2*RSp45611e*RSp45613e - 1./81.*m33
         *mp2*RSp45611e*RSp45633e - 1./54.*m33*mp2*pow(RSp45611e,2) + 
         41./648.*pow(m33,2)*pow(RSp45613e,2) - 29./648.*pow(m33,2)*
         RSp45633e*RSp45613e + 109./2592.*pow(m33,2)*pow(RSp45633e,2)
          - 7./216.*pow(m33,2)*RSp45611e*RSp45613e + 13./1296.*pow(
         m33,2)*RSp45611e*RSp45633e + 5./288.*pow(m33,2)*pow(
         RSp45611e,2) - 5./81.*m11*mp2*pow(RSp45613e,2) + 4./81.*m11*
         mp2*RSp45633e*RSp45613e - 1./54.*m11*mp2*pow(RSp45633e,2) + 2./
         27.*m11*mp2*RSp45611e*RSp45613e - 1./81.*m11*mp2*RSp45611e*
         RSp45633e );

      M13p6x23R +=  + hh(1,m13,mp2,mp2,m13,mu2) * (  - 5./162.*m11*mp2*
         pow(RSp45611e,2) + 5./36.*m11*m33*pow(RSp45613e,2) - 25./324.*
         m11*m33*RSp45633e*RSp45613e + 23./432.*m11*m33*pow(
         RSp45633e,2) - 25./324.*m11*m33*RSp45611e*RSp45613e + 13./648.
         *m11*m33*RSp45611e*RSp45633e + 23./432.*m11*m33*pow(
         RSp45611e,2) + 41./648.*pow(m11,2)*pow(RSp45613e,2) - 7./216.*
         pow(m11,2)*RSp45633e*RSp45613e + 5./288.*pow(m11,2)*pow(
         RSp45633e,2) - 29./648.*pow(m11,2)*RSp45611e*RSp45613e + 13./
         1296.*pow(m11,2)*RSp45611e*RSp45633e + 109./2592.*pow(m11,2)*
         pow(RSp45611e,2) );

      M13p6x23R +=  + hh(1,m14,m33,m34,m13,mu2) * ( 1./24.*m33*m44*
         RS34561pe + 1./48.*m33*m44*RS3456pee + 1./48.*m33*m44*
         RS3456ppe - 1./48.*m33*m44*RS345pe - 1./48.*m33*m44*RS346pe - 
         1./48.*m33*m44*RS356pe + 1./16.*pow(m33,2)*RS34561pe + 1./48.*
         pow(m33,2)*RS3456pee + 1./48.*pow(m33,2)*RS3456ppe - 1./48.*
         pow(m33,2)*RS345pe - 1./48.*pow(m33,2)*RS346pe - 1./24.*pow(
         m33,2)*RS356pe + 1./24.*m11*m44*RS34561pe + 1./48.*m11*m44*
         RS3456pee + 1./48.*m11*m44*RS3456ppe - 1./48.*m11*m44*RS345pe
          - 1./48.*m11*m44*RS346pe - 1./48.*m11*m44*RS356pe + 1./6.*m11
         *m33*RS34561pe + 1./48.*m11*m33*RS3456pee + 1./48.*m11*m33*
         RS3456ppe - 1./48.*m11*m33*RS345pe - 1./48.*m11*m33*RS346pe - 
         7./48.*m11*m33*RS356pe + 5./48.*pow(m11,2)*RS34561pe - 5./48.*
         pow(m11,2)*RS356pe );

      M13p6x23R +=  + hh(1,m14,m34,m44,m13,mu2) * ( 1./8.*m33*m44 - 1./
         24.*m33*m44*RS456pe + 1./8.*m11*m44 - 1./24.*m11*m44*RS456pe )
         ;

      M13p6x23R +=  + hh(1,m14,m34,me2,m13,mu2) * (  - 1./54.*pow(
         me2,2)*RSe45613p - 1./108.*pow(me2,2)*RSe45633p + 1./27.*pow(
         me2,2)*RSe563p - 1./108.*pow(me2,2)*RSe45611p + 1./27.*pow(
         me2,2)*RSe561p - 1./27.*pow(me2,2)*RSe564p - 1./27.*m44*me2*
         RSe563p - 1./27.*m44*me2*RSe561p + 2./27.*m44*me2*RSe564p - 1./
         27.*pow(m44,2)*RSe564p - 17./216.*m33*me2*RSe45613p - 1./27.*
         m33*me2*RSe45633p + 25./216.*m33*me2*RSe563p + 1./24.*m33*me2*
         RSe561p - 1./24.*m33*me2*RSe564p + 1./24.*m33*m44*RSe45613p - 
         1./48.*m33*m44*RSe45633p - 2./27.*m33*m44*RSe563p - 1./48.*m33
         *m44*RSe45611p - 1./16.*pow(m33,2)*RSe45613p + 11./432.*pow(
         m33,2)*RSe45633p - 1./16.*pow(m33,2)*RSe563p + 1./16.*pow(
         m33,2)*RSe561p - 17./216.*m11*me2*RSe45613p - 1./18.*m11*me2*
         RSe45633p + 11./72.*m11*me2*RSe563p + 1./54.*m11*me2*RSe45611p
          + 1./216.*m11*me2*RSe561p - 1./24.*m11*me2*RSe564p + 1./24.*
         m11*m44*RSe45613p - 1./48.*m11*m44*RSe45633p - 1./9.*m11*m44*
         RSe563p );

      M13p6x23R +=  + hh(1,m14,m34,me2,m13,mu2) * (  - 1./48.*m11*m44*
         RSe45611p + 1./27.*m11*m44*RSe561p - 19./216.*m11*m33*
         RSe45613p + 5./144.*m11*m33*RSe45633p - 1./6.*m11*m33*RSe563p
          - 1./48.*m11*m33*RSe45611p + 1./6.*m11*m33*RSe561p - 1./144.*
         pow(m11,2)*RSe45613p - 5./48.*pow(m11,2)*RSe563p - 13./432.*
         pow(m11,2)*RSe45611p + 5./48.*pow(m11,2)*RSe561p );

      M13p6x23R +=  + hh(1,m14,m34,mp2,m13,mu2) * (  - 1./54.*pow(
         mp2,2)*RSp45613e - 1./108.*pow(mp2,2)*RSp45633e + 1./27.*pow(
         mp2,2)*RSp563e - 1./108.*pow(mp2,2)*RSp45611e + 1./27.*pow(
         mp2,2)*RSp561e - 1./27.*pow(mp2,2)*RSp564e - 1./27.*m44*mp2*
         RSp563e - 1./27.*m44*mp2*RSp561e + 2./27.*m44*mp2*RSp564e - 1./
         27.*pow(m44,2)*RSp564e - 17./216.*m33*mp2*RSp45613e - 1./27.*
         m33*mp2*RSp45633e + 25./216.*m33*mp2*RSp563e + 1./24.*m33*mp2*
         RSp561e - 1./24.*m33*mp2*RSp564e + 1./24.*m33*m44*RSp45613e - 
         1./48.*m33*m44*RSp45633e - 2./27.*m33*m44*RSp563e - 1./48.*m33
         *m44*RSp45611e - 1./16.*pow(m33,2)*RSp45613e + 11./432.*pow(
         m33,2)*RSp45633e - 1./16.*pow(m33,2)*RSp563e + 1./16.*pow(
         m33,2)*RSp561e - 17./216.*m11*mp2*RSp45613e - 1./18.*m11*mp2*
         RSp45633e + 11./72.*m11*mp2*RSp563e + 1./54.*m11*mp2*RSp45611e
          + 1./216.*m11*mp2*RSp561e - 1./24.*m11*mp2*RSp564e + 1./24.*
         m11*m44*RSp45613e - 1./48.*m11*m44*RSp45633e - 1./9.*m11*m44*
         RSp563e );

      M13p6x23R +=  + hh(1,m14,m34,mp2,m13,mu2) * (  - 1./48.*m11*m44*
         RSp45611e + 1./27.*m11*m44*RSp561e - 19./216.*m11*m33*
         RSp45613e + 5./144.*m11*m33*RSp45633e - 1./6.*m11*m33*RSp563e
          - 1./48.*m11*m33*RSp45611e + 1./6.*m11*m33*RSp561e - 1./144.*
         pow(m11,2)*RSp45613e - 5./48.*pow(m11,2)*RSp563e - 13./432.*
         pow(m11,2)*RSp45611e + 5./48.*pow(m11,2)*RSp561e );

      M13p6x23R +=  + hh(1,m14,m35,m45,m13,mu2) * ( 1./16.*m33*m55 + 1./
         16.*m33*m44 + 1./16.*m11*m55 + 1./16.*m11*m44 );

      M13p6x23R +=  + hh(1,m14,m36,m46,m13,mu2) * ( 1./16.*m33*m66 + 1./
         16.*m33*m44 + 1./16.*m11*m66 + 1./16.*m11*m44 );

      M13p6x23R +=  + hh(1,m15,m33,m35,m13,mu2) * ( 1./24.*m33*m55*
         RS34561pe + 1./48.*m33*m55*RS3456pee + 1./48.*m33*m55*
         RS3456ppe - 1./48.*m33*m55*RS345pe - 1./48.*m33*m55*RS346pe - 
         1./48.*m33*m55*RS356pe + 1./16.*pow(m33,2)*RS34561pe + 1./48.*
         pow(m33,2)*RS3456pee + 1./48.*pow(m33,2)*RS3456ppe - 1./48.*
         pow(m33,2)*RS345pe - 1./24.*pow(m33,2)*RS346pe - 1./48.*pow(
         m33,2)*RS356pe + 1./24.*m11*m55*RS34561pe + 1./48.*m11*m55*
         RS3456pee + 1./48.*m11*m55*RS3456ppe - 1./48.*m11*m55*RS345pe
          - 1./48.*m11*m55*RS346pe - 1./48.*m11*m55*RS356pe + 1./6.*m11
         *m33*RS34561pe + 1./48.*m11*m33*RS3456pee + 1./48.*m11*m33*
         RS3456ppe - 1./48.*m11*m33*RS345pe - 7./48.*m11*m33*RS346pe - 
         1./48.*m11*m33*RS356pe + 5./48.*pow(m11,2)*RS34561pe - 5./48.*
         pow(m11,2)*RS346pe );

      M13p6x23R +=  + hh(1,m15,m34,m45,m13,mu2) * ( 1./16.*m33*m55 + 1./
         16.*m33*m44 + 1./16.*m11*m55 + 1./16.*m11*m44 );

      M13p6x23R +=  + hh(1,m15,m35,m55,m13,mu2) * ( 1./8.*m33*m55 - 1./
         24.*m33*m55*RS546pe + 1./8.*m11*m55 - 1./24.*m11*m55*RS546pe )
         ;

      M13p6x23R +=  + hh(1,m15,m35,me2,m13,mu2) * (  - 1./54.*pow(
         me2,2)*RSe45613p - 1./108.*pow(me2,2)*RSe45633p + 1./27.*pow(
         me2,2)*RSe463p - 1./108.*pow(me2,2)*RSe45611p + 1./27.*pow(
         me2,2)*RSe461p - 1./27.*pow(me2,2)*RSe465p - 1./27.*m55*me2*
         RSe463p - 1./27.*m55*me2*RSe461p + 2./27.*m55*me2*RSe465p - 1./
         27.*pow(m55,2)*RSe465p - 17./216.*m33*me2*RSe45613p - 1./27.*
         m33*me2*RSe45633p + 25./216.*m33*me2*RSe463p + 1./24.*m33*me2*
         RSe461p - 1./24.*m33*me2*RSe465p + 1./24.*m33*m55*RSe45613p - 
         1./48.*m33*m55*RSe45633p - 2./27.*m33*m55*RSe463p - 1./48.*m33
         *m55*RSe45611p - 1./16.*pow(m33,2)*RSe45613p + 11./432.*pow(
         m33,2)*RSe45633p - 1./16.*pow(m33,2)*RSe463p + 1./16.*pow(
         m33,2)*RSe461p - 17./216.*m11*me2*RSe45613p - 1./18.*m11*me2*
         RSe45633p + 11./72.*m11*me2*RSe463p + 1./54.*m11*me2*RSe45611p
          + 1./216.*m11*me2*RSe461p - 1./24.*m11*me2*RSe465p + 1./24.*
         m11*m55*RSe45613p - 1./48.*m11*m55*RSe45633p - 1./9.*m11*m55*
         RSe463p );

      M13p6x23R +=  + hh(1,m15,m35,me2,m13,mu2) * (  - 1./48.*m11*m55*
         RSe45611p + 1./27.*m11*m55*RSe461p - 19./216.*m11*m33*
         RSe45613p + 5./144.*m11*m33*RSe45633p - 1./6.*m11*m33*RSe463p
          - 1./48.*m11*m33*RSe45611p + 1./6.*m11*m33*RSe461p - 1./144.*
         pow(m11,2)*RSe45613p - 5./48.*pow(m11,2)*RSe463p - 13./432.*
         pow(m11,2)*RSe45611p + 5./48.*pow(m11,2)*RSe461p );

      M13p6x23R +=  + hh(1,m15,m35,mp2,m13,mu2) * (  - 1./54.*pow(
         mp2,2)*RSp45613e - 1./108.*pow(mp2,2)*RSp45633e + 1./27.*pow(
         mp2,2)*RSp463e - 1./108.*pow(mp2,2)*RSp45611e + 1./27.*pow(
         mp2,2)*RSp461e - 1./27.*pow(mp2,2)*RSp465e - 1./27.*m55*mp2*
         RSp463e - 1./27.*m55*mp2*RSp461e + 2./27.*m55*mp2*RSp465e - 1./
         27.*pow(m55,2)*RSp465e - 17./216.*m33*mp2*RSp45613e - 1./27.*
         m33*mp2*RSp45633e + 25./216.*m33*mp2*RSp463e + 1./24.*m33*mp2*
         RSp461e - 1./24.*m33*mp2*RSp465e + 1./24.*m33*m55*RSp45613e - 
         1./48.*m33*m55*RSp45633e - 2./27.*m33*m55*RSp463e - 1./48.*m33
         *m55*RSp45611e - 1./16.*pow(m33,2)*RSp45613e + 11./432.*pow(
         m33,2)*RSp45633e - 1./16.*pow(m33,2)*RSp463e + 1./16.*pow(
         m33,2)*RSp461e - 17./216.*m11*mp2*RSp45613e - 1./18.*m11*mp2*
         RSp45633e + 11./72.*m11*mp2*RSp463e + 1./54.*m11*mp2*RSp45611e
          + 1./216.*m11*mp2*RSp461e - 1./24.*m11*mp2*RSp465e + 1./24.*
         m11*m55*RSp45613e - 1./48.*m11*m55*RSp45633e - 1./9.*m11*m55*
         RSp463e );

      M13p6x23R +=  + hh(1,m15,m35,mp2,m13,mu2) * (  - 1./48.*m11*m55*
         RSp45611e + 1./27.*m11*m55*RSp461e - 19./216.*m11*m33*
         RSp45613e + 5./144.*m11*m33*RSp45633e - 1./6.*m11*m33*RSp463e
          - 1./48.*m11*m33*RSp45611e + 1./6.*m11*m33*RSp461e - 1./144.*
         pow(m11,2)*RSp45613e - 5./48.*pow(m11,2)*RSp463e - 13./432.*
         pow(m11,2)*RSp45611e + 5./48.*pow(m11,2)*RSp461e );

      M13p6x23R +=  + hh(1,m15,m36,m56,m13,mu2) * ( 1./16.*m33*m66 + 1./
         16.*m33*m55 + 1./16.*m11*m66 + 1./16.*m11*m55 );

      M13p6x23R +=  + hh(1,m16,m33,m36,m13,mu2) * ( 1./24.*m33*m66*
         RS34561pe + 1./48.*m33*m66*RS3456pee + 1./48.*m33*m66*
         RS3456ppe - 1./48.*m33*m66*RS345pe - 1./48.*m33*m66*RS346pe - 
         1./48.*m33*m66*RS356pe + 1./16.*pow(m33,2)*RS34561pe + 1./48.*
         pow(m33,2)*RS3456pee + 1./48.*pow(m33,2)*RS3456ppe - 1./24.*
         pow(m33,2)*RS345pe - 1./48.*pow(m33,2)*RS346pe - 1./48.*pow(
         m33,2)*RS356pe + 1./24.*m11*m66*RS34561pe + 1./48.*m11*m66*
         RS3456pee + 1./48.*m11*m66*RS3456ppe - 1./48.*m11*m66*RS345pe
          - 1./48.*m11*m66*RS346pe - 1./48.*m11*m66*RS356pe + 1./6.*m11
         *m33*RS34561pe + 1./48.*m11*m33*RS3456pee + 1./48.*m11*m33*
         RS3456ppe - 7./48.*m11*m33*RS345pe - 1./48.*m11*m33*RS346pe - 
         1./48.*m11*m33*RS356pe + 5./48.*pow(m11,2)*RS34561pe - 5./48.*
         pow(m11,2)*RS345pe );

      M13p6x23R +=  + hh(1,m16,m34,m46,m13,mu2) * ( 1./16.*m33*m66 + 1./
         16.*m33*m44 + 1./16.*m11*m66 + 1./16.*m11*m44 );

      M13p6x23R +=  + hh(1,m16,m35,m56,m13,mu2) * ( 1./16.*m33*m66 + 1./
         16.*m33*m55 + 1./16.*m11*m66 + 1./16.*m11*m55 );

      M13p6x23R +=  + hh(1,m16,m36,m66,m13,mu2) * ( 1./8.*m33*m66 - 1./
         24.*m33*m66*RS645pe + 1./8.*m11*m66 - 1./24.*m11*m66*RS645pe )
         ;

      M13p6x23R +=  + hh(1,m16,m36,me2,m13,mu2) * (  - 1./54.*pow(
         me2,2)*RSe45613p - 1./108.*pow(me2,2)*RSe45633p + 1./27.*pow(
         me2,2)*RSe453p - 1./108.*pow(me2,2)*RSe45611p + 1./27.*pow(
         me2,2)*RSe451p - 1./27.*pow(me2,2)*RSe456p - 1./27.*m66*me2*
         RSe453p - 1./27.*m66*me2*RSe451p + 2./27.*m66*me2*RSe456p - 1./
         27.*pow(m66,2)*RSe456p - 17./216.*m33*me2*RSe45613p - 1./27.*
         m33*me2*RSe45633p + 25./216.*m33*me2*RSe453p + 1./24.*m33*me2*
         RSe451p - 1./24.*m33*me2*RSe456p + 1./24.*m33*m66*RSe45613p - 
         1./48.*m33*m66*RSe45633p - 2./27.*m33*m66*RSe453p - 1./48.*m33
         *m66*RSe45611p - 1./16.*pow(m33,2)*RSe45613p + 11./432.*pow(
         m33,2)*RSe45633p - 1./16.*pow(m33,2)*RSe453p + 1./16.*pow(
         m33,2)*RSe451p - 17./216.*m11*me2*RSe45613p - 1./18.*m11*me2*
         RSe45633p + 11./72.*m11*me2*RSe453p + 1./54.*m11*me2*RSe45611p
          + 1./216.*m11*me2*RSe451p - 1./24.*m11*me2*RSe456p + 1./24.*
         m11*m66*RSe45613p - 1./48.*m11*m66*RSe45633p - 1./9.*m11*m66*
         RSe453p );

      M13p6x23R +=  + hh(1,m16,m36,me2,m13,mu2) * (  - 1./48.*m11*m66*
         RSe45611p + 1./27.*m11*m66*RSe451p - 19./216.*m11*m33*
         RSe45613p + 5./144.*m11*m33*RSe45633p - 1./6.*m11*m33*RSe453p
          - 1./48.*m11*m33*RSe45611p + 1./6.*m11*m33*RSe451p - 1./144.*
         pow(m11,2)*RSe45613p - 5./48.*pow(m11,2)*RSe453p - 13./432.*
         pow(m11,2)*RSe45611p + 5./48.*pow(m11,2)*RSe451p );

      M13p6x23R +=  + hh(1,m16,m36,mp2,m13,mu2) * (  - 1./54.*pow(
         mp2,2)*RSp45613e - 1./108.*pow(mp2,2)*RSp45633e + 1./27.*pow(
         mp2,2)*RSp453e - 1./108.*pow(mp2,2)*RSp45611e + 1./27.*pow(
         mp2,2)*RSp451e - 1./27.*pow(mp2,2)*RSp456e - 1./27.*m66*mp2*
         RSp453e - 1./27.*m66*mp2*RSp451e + 2./27.*m66*mp2*RSp456e - 1./
         27.*pow(m66,2)*RSp456e - 17./216.*m33*mp2*RSp45613e - 1./27.*
         m33*mp2*RSp45633e + 25./216.*m33*mp2*RSp453e + 1./24.*m33*mp2*
         RSp451e - 1./24.*m33*mp2*RSp456e + 1./24.*m33*m66*RSp45613e - 
         1./48.*m33*m66*RSp45633e - 2./27.*m33*m66*RSp453e - 1./48.*m33
         *m66*RSp45611e - 1./16.*pow(m33,2)*RSp45613e + 11./432.*pow(
         m33,2)*RSp45633e - 1./16.*pow(m33,2)*RSp453e + 1./16.*pow(
         m33,2)*RSp451e - 17./216.*m11*mp2*RSp45613e - 1./18.*m11*mp2*
         RSp45633e + 11./72.*m11*mp2*RSp453e + 1./54.*m11*mp2*RSp45611e
          + 1./216.*m11*mp2*RSp451e - 1./24.*m11*mp2*RSp456e + 1./24.*
         m11*m66*RSp45613e - 1./48.*m11*m66*RSp45633e - 1./9.*m11*m66*
         RSp453e );

      M13p6x23R +=  + hh(1,m16,m36,mp2,m13,mu2) * (  - 1./48.*m11*m66*
         RSp45611e + 1./27.*m11*m66*RSp451e - 19./216.*m11*m33*
         RSp45613e + 5./144.*m11*m33*RSp45633e - 1./6.*m11*m33*RSp453e
          - 1./48.*m11*m33*RSp45611e + 1./6.*m11*m33*RSp451e - 1./144.*
         pow(m11,2)*RSp45613e - 5./48.*pow(m11,2)*RSp453e - 13./432.*
         pow(m11,2)*RSp45611e + 5./48.*pow(m11,2)*RSp451e );

      M13p6x23R +=  + hh(2,m11,m11,m13,m13,mu2) * (  - 1./54.*m33*pow(
         RD1456ep,2) - 7./216.*pow(m33,2)*RD1456ep*RS14563pe - 5./144.*
         pow(m33,2)*RS1456pee*RD1456ep - 5./144.*pow(m33,2)*RS1456ppe*
         RD1456ep + 5./144.*pow(m33,2)*RS145pe*RD1456ep + 5./144.*pow(
         m33,2)*RS146pe*RD1456ep + 5./144.*pow(m33,2)*RS156pe*RD1456ep
          - 1./54.*m11*pow(RD1456ep,2) - 1./36.*m11*m33*RD1456ep*
         RS14563pe - 5./72.*m11*m33*RS1456pee*RD1456ep - 5./72.*m11*m33
         *RS1456ppe*RD1456ep + 5./72.*m11*m33*RS145pe*RD1456ep + 5./72.
         *m11*m33*RS146pe*RD1456ep + 5./72.*m11*m33*RS156pe*RD1456ep + 
         1./216.*pow(m11,2)*RD1456ep*RS14563pe - 5./144.*pow(m11,2)*
         RS1456pee*RD1456ep - 5./144.*pow(m11,2)*RS1456ppe*RD1456ep + 5.
         /144.*pow(m11,2)*RS145pe*RD1456ep + 5./144.*pow(m11,2)*RS146pe
         *RD1456ep + 5./144.*pow(m11,2)*RS156pe*RD1456ep );

      M13p6x23R +=  + hh(2,m11,m13,m33,m13,mu2) * (  - 1./72.*pow(
         m33,2)*RD1456ep*RS34561pe - 1./144.*pow(m33,2)*RD1456ep*
         RS3456pee - 1./144.*pow(m33,2)*RD1456ep*RS3456ppe + 1./144.*
         pow(m33,2)*RD1456ep*RS345pe + 1./144.*pow(m33,2)*RD1456ep*
         RS346pe + 1./144.*pow(m33,2)*RD1456ep*RS356pe - 1./36.*m11*m33
         *RD1456ep*RS34561pe - 1./72.*m11*m33*RD1456ep*RS3456pee - 1./
         72.*m11*m33*RD1456ep*RS3456ppe + 1./72.*m11*m33*RD1456ep*
         RS345pe + 1./72.*m11*m33*RD1456ep*RS346pe + 1./72.*m11*m33*
         RD1456ep*RS356pe - 1./72.*pow(m11,2)*RD1456ep*RS34561pe - 1./
         144.*pow(m11,2)*RD1456ep*RS3456pee - 1./144.*pow(m11,2)*
         RD1456ep*RS3456ppe + 1./144.*pow(m11,2)*RD1456ep*RS345pe + 1./
         144.*pow(m11,2)*RD1456ep*RS346pe + 1./144.*pow(m11,2)*RD1456ep
         *RS356pe );

      M13p6x23R +=  + hh(2,m11,m13,me2,m13,mu2) * (  - 1./162.*pow(
         me2,2)*RD1456ep*RSe45613p + 1./324.*pow(me2,2)*RD1456ep*
         RSe45633p + 1./324.*pow(me2,2)*RSe45611p*RD1456ep + 2./81.*m33
         *me2*RD1456ep*RSe45613p - 1./162.*m33*me2*RD1456ep*RSe45633p
          - 1./54.*m33*me2*RSe45611p*RD1456ep - 7./216.*pow(m33,2)*
         RD1456ep*RSe45613p + 13./1296.*pow(m33,2)*RD1456ep*RSe45633p
          + 5./144.*pow(m33,2)*RSe45611p*RD1456ep + 2./81.*m11*me2*
         RD1456ep*RSe45613p - 2./81.*m11*me2*RSe45611p*RD1456ep - 17./
         324.*m11*m33*RD1456ep*RSe45613p + 1./72.*m11*m33*RD1456ep*
         RSe45633p + 19./216.*m11*m33*RSe45611p*RD1456ep - 1./72.*pow(
         m11,2)*RD1456ep*RSe45613p + 1./144.*pow(m11,2)*RD1456ep*
         RSe45633p + 73./1296.*pow(m11,2)*RSe45611p*RD1456ep );

      M13p6x23R +=  + hh(2,m11,m13,mp2,m13,mu2) * (  - 1./162.*pow(
         mp2,2)*RD1456ep*RSp45613e + 1./324.*pow(mp2,2)*RD1456ep*
         RSp45633e + 1./324.*pow(mp2,2)*RSp45611e*RD1456ep + 2./81.*m33
         *mp2*RD1456ep*RSp45613e - 1./162.*m33*mp2*RD1456ep*RSp45633e
          - 1./54.*m33*mp2*RSp45611e*RD1456ep - 7./216.*pow(m33,2)*
         RD1456ep*RSp45613e + 13./1296.*pow(m33,2)*RD1456ep*RSp45633e
          + 5./144.*pow(m33,2)*RSp45611e*RD1456ep + 2./81.*m11*mp2*
         RD1456ep*RSp45613e - 2./81.*m11*mp2*RSp45611e*RD1456ep - 17./
         324.*m11*m33*RD1456ep*RSp45613e + 1./72.*m11*m33*RD1456ep*
         RSp45633e + 19./216.*m11*m33*RSp45611e*RD1456ep - 1./72.*pow(
         m11,2)*RD1456ep*RSp45613e + 1./144.*pow(m11,2)*RD1456ep*
         RSp45633e + 73./1296.*pow(m11,2)*RSp45611e*RD1456ep );

      M13p6x23R +=  + hh(2,m11,m14,m34,m13,mu2) * (  - 1./48.*m33*m44*
         RD1456ep - 1./48.*m11*m44*RD1456ep - 1./48.*m11*m33*RD1456ep
          - 1./48.*pow(m11,2)*RD1456ep );

      M13p6x23R +=  + hh(2,m11,m15,m35,m13,mu2) * (  - 1./48.*m33*m55*
         RD1456ep - 1./48.*m11*m55*RD1456ep - 1./48.*m11*m33*RD1456ep
          - 1./48.*pow(m11,2)*RD1456ep );

      M13p6x23R +=  + hh(2,m11,m16,m36,m13,mu2) * (  - 1./48.*m33*m66*
         RD1456ep - 1./48.*m11*m66*RD1456ep - 1./48.*m11*m33*RD1456ep
          - 1./48.*pow(m11,2)*RD1456ep );

      M13p6x23R +=  + hh(2,m33,m11,m13,m13,mu2) * (  - 1./72.*pow(
         m33,2)*RD3456ep*RS14563pe - 1./144.*pow(m33,2)*RS1456pee*
         RD3456ep - 1./144.*pow(m33,2)*RS1456ppe*RD3456ep + 1./144.*
         pow(m33,2)*RS145pe*RD3456ep + 1./144.*pow(m33,2)*RS146pe*
         RD3456ep + 1./144.*pow(m33,2)*RS156pe*RD3456ep - 1./36.*m11*
         m33*RD3456ep*RS14563pe - 1./72.*m11*m33*RS1456pee*RD3456ep - 1.
         /72.*m11*m33*RS1456ppe*RD3456ep + 1./72.*m11*m33*RS145pe*
         RD3456ep + 1./72.*m11*m33*RS146pe*RD3456ep + 1./72.*m11*m33*
         RS156pe*RD3456ep - 1./72.*pow(m11,2)*RD3456ep*RS14563pe - 1./
         144.*pow(m11,2)*RS1456pee*RD3456ep - 1./144.*pow(m11,2)*
         RS1456ppe*RD3456ep + 1./144.*pow(m11,2)*RS145pe*RD3456ep + 1./
         144.*pow(m11,2)*RS146pe*RD3456ep + 1./144.*pow(m11,2)*RS156pe*
         RD3456ep );

      M13p6x23R +=  + hh(2,m33,m13,m33,m13,mu2) * (  - 1./54.*m33*pow(
         RD3456ep,2) + 1./216.*pow(m33,2)*RD3456ep*RS34561pe - 5./144.*
         pow(m33,2)*RS3456pee*RD3456ep - 5./144.*pow(m33,2)*RS3456ppe*
         RD3456ep + 5./144.*pow(m33,2)*RS345pe*RD3456ep + 5./144.*pow(
         m33,2)*RS346pe*RD3456ep + 5./144.*pow(m33,2)*RS356pe*RD3456ep
          - 1./54.*m11*pow(RD3456ep,2) - 1./36.*m11*m33*RD3456ep*
         RS34561pe - 5./72.*m11*m33*RS3456pee*RD3456ep - 5./72.*m11*m33
         *RS3456ppe*RD3456ep + 5./72.*m11*m33*RS345pe*RD3456ep + 5./72.
         *m11*m33*RS346pe*RD3456ep + 5./72.*m11*m33*RS356pe*RD3456ep - 
         7./216.*pow(m11,2)*RD3456ep*RS34561pe - 5./144.*pow(m11,2)*
         RS3456pee*RD3456ep - 5./144.*pow(m11,2)*RS3456ppe*RD3456ep + 5.
         /144.*pow(m11,2)*RS345pe*RD3456ep + 5./144.*pow(m11,2)*RS346pe
         *RD3456ep + 5./144.*pow(m11,2)*RS356pe*RD3456ep );

      M13p6x23R +=  + hh(2,m33,m13,me2,m13,mu2) * (  - 1./162.*pow(
         me2,2)*RD3456ep*RSe45613p + 1./324.*pow(me2,2)*RSe45633p*
         RD3456ep + 1./324.*pow(me2,2)*RSe45611p*RD3456ep + 2./81.*m33*
         me2*RD3456ep*RSe45613p - 2./81.*m33*me2*RSe45633p*RD3456ep - 1.
         /72.*pow(m33,2)*RD3456ep*RSe45613p + 73./1296.*pow(m33,2)*
         RSe45633p*RD3456ep + 1./144.*pow(m33,2)*RSe45611p*RD3456ep + 2.
         /81.*m11*me2*RD3456ep*RSe45613p - 1./54.*m11*me2*RSe45633p*
         RD3456ep - 1./162.*m11*me2*RSe45611p*RD3456ep - 17./324.*m11*
         m33*RD3456ep*RSe45613p + 19./216.*m11*m33*RSe45633p*RD3456ep
          + 1./72.*m11*m33*RSe45611p*RD3456ep - 7./216.*pow(m11,2)*
         RD3456ep*RSe45613p + 5./144.*pow(m11,2)*RSe45633p*RD3456ep + 
         13./1296.*pow(m11,2)*RSe45611p*RD3456ep );

      M13p6x23R +=  + hh(2,m33,m13,mp2,m13,mu2) * (  - 1./162.*pow(
         mp2,2)*RD3456ep*RSp45613e + 1./324.*pow(mp2,2)*RSp45633e*
         RD3456ep + 1./324.*pow(mp2,2)*RSp45611e*RD3456ep + 2./81.*m33*
         mp2*RD3456ep*RSp45613e - 2./81.*m33*mp2*RSp45633e*RD3456ep - 1.
         /72.*pow(m33,2)*RD3456ep*RSp45613e + 73./1296.*pow(m33,2)*
         RSp45633e*RD3456ep + 1./144.*pow(m33,2)*RSp45611e*RD3456ep + 2.
         /81.*m11*mp2*RD3456ep*RSp45613e - 1./54.*m11*mp2*RSp45633e*
         RD3456ep - 1./162.*m11*mp2*RSp45611e*RD3456ep - 17./324.*m11*
         m33*RD3456ep*RSp45613e + 19./216.*m11*m33*RSp45633e*RD3456ep
          + 1./72.*m11*m33*RSp45611e*RD3456ep - 7./216.*pow(m11,2)*
         RD3456ep*RSp45613e + 5./144.*pow(m11,2)*RSp45633e*RD3456ep + 
         13./1296.*pow(m11,2)*RSp45611e*RD3456ep );

      M13p6x23R +=  + hh(2,m33,m14,m34,m13,mu2) * (  - 1./48.*m33*m44*
         RD3456ep - 1./48.*pow(m33,2)*RD3456ep - 1./48.*m11*m44*
         RD3456ep - 1./48.*m11*m33*RD3456ep );

      M13p6x23R +=  + hh(2,m33,m15,m35,m13,mu2) * (  - 1./48.*m33*m55*
         RD3456ep - 1./48.*pow(m33,2)*RD3456ep - 1./48.*m11*m55*
         RD3456ep - 1./48.*m11*m33*RD3456ep );

      M13p6x23R +=  + hh(2,m33,m16,m36,m13,mu2) * (  - 1./48.*m33*m66*
         RD3456ep - 1./48.*pow(m33,2)*RD3456ep - 1./48.*m11*m66*
         RD3456ep - 1./48.*m11*m33*RD3456ep );

      M13p6x23R +=  + hh(5,m11,m11,m13,m13,mu2) * ( 5./288.*pow(m33,2)*
         pow(RD1456ep,2) + 5./144.*m11*m33*pow(RD1456ep,2) + 5./288.*
         pow(m11,2)*pow(RD1456ep,2) );

      M13p6x23R +=  + hh(5,m11,m33,m13,m13,mu2) * ( 1./144.*pow(m33,2)*
         RD1456ep*RD3456ep + 1./72.*m11*m33*RD1456ep*RD3456ep + 1./144.
         *pow(m11,2)*RD1456ep*RD3456ep );

      M13p6x23R +=  + hh(5,m33,m33,m13,m13,mu2) * ( 5./288.*pow(m33,2)*
         pow(RD3456ep,2) + 5./144.*m11*m33*pow(RD3456ep,2) + 5./288.*
         pow(m11,2)*pow(RD3456ep,2) );

      M13p6x23R +=  + hh1(1,m11,m14,m34,m13,mu2) * ( 1./6.*pow(m33,2)*
         RS14563pe - 1./6.*pow(m33,2)*RS156pe + 1./3.*m11*m33*RS14563pe
          - 1./3.*m11*m33*RS156pe + 1./6.*pow(m11,2)*RS14563pe - 1./6.*
         pow(m11,2)*RS156pe );

      M13p6x23R +=  + hh1(1,m11,m15,m35,m13,mu2) * ( 1./6.*pow(m33,2)*
         RS14563pe - 1./6.*pow(m33,2)*RS146pe + 1./3.*m11*m33*RS14563pe
          - 1./3.*m11*m33*RS146pe + 1./6.*pow(m11,2)*RS14563pe - 1./6.*
         pow(m11,2)*RS146pe );

      M13p6x23R +=  + hh1(1,m11,m16,m36,m13,mu2) * ( 1./6.*pow(m33,2)*
         RS14563pe - 1./6.*pow(m33,2)*RS145pe + 1./3.*m11*m33*RS14563pe
          - 1./3.*m11*m33*RS145pe + 1./6.*pow(m11,2)*RS14563pe - 1./6.*
         pow(m11,2)*RS145pe );

      M13p6x23R +=  + hh1(1,m13,m11,m11,m13,mu2) * (  - 1./27.*m33*
         RD1456ep*RS14563pe - 1./54.*m33*RS1456pee*RD1456ep - 1./54.*
         m33*RS1456ppe*RD1456ep + 1./54.*m33*RS145pe*RD1456ep + 1./54.*
         m33*RS146pe*RD1456ep + 1./54.*m33*RS156pe*RD1456ep - 1./4.*
         pow(m33,2) - 5./54.*pow(m33,2)*pow(RS14563pe,2) - 2./27.*pow(
         m33,2)*RS1456pee*RS14563pe - 1./36.*pow(m33,2)*pow(
         RS1456pee,2) - 2./27.*pow(m33,2)*RS1456ppe*RS14563pe - 1./18.*
         pow(m33,2)*RS1456ppe*RS1456pee - 1./36.*pow(m33,2)*pow(
         RS1456ppe,2) + 2./27.*pow(m33,2)*RS145pe*RS14563pe + 1./18.*
         pow(m33,2)*RS145pe*RS1456pee + 1./18.*pow(m33,2)*RS145pe*
         RS1456ppe - 1./36.*pow(m33,2)*pow(RS145pe,2) + 2./27.*pow(
         m33,2)*RS146pe*RS14563pe + 1./18.*pow(m33,2)*RS146pe*RS1456pee
          + 1./18.*pow(m33,2)*RS146pe*RS1456ppe - 1./18.*pow(m33,2)*
         RS146pe*RS145pe - 1./36.*pow(m33,2)*pow(RS146pe,2) + 2./27.*
         pow(m33,2)*RS156pe*RS14563pe + 1./18.*pow(m33,2)*RS156pe*
         RS1456pee );

      M13p6x23R +=  + hh1(1,m13,m11,m11,m13,mu2) * ( 1./18.*pow(m33,2)*
         RS156pe*RS1456ppe - 1./18.*pow(m33,2)*RS156pe*RS145pe - 1./18.
         *pow(m33,2)*RS156pe*RS146pe - 1./36.*pow(m33,2)*pow(RS156pe,2)
          - 1./27.*m11*RD1456ep*RS14563pe - 1./54.*m11*RS1456pee*
         RD1456ep - 1./54.*m11*RS1456ppe*RD1456ep + 1./54.*m11*RS145pe*
         RD1456ep + 1./54.*m11*RS146pe*RD1456ep + 1./54.*m11*RS156pe*
         RD1456ep - 1./2.*m11*m33 - 1./9.*m11*m33*pow(RS14563pe,2) - 1./
         9.*m11*m33*RS1456pee*RS14563pe - 1./18.*m11*m33*pow(
         RS1456pee,2) - 1./9.*m11*m33*RS1456ppe*RS14563pe - 1./9.*m11*
         m33*RS1456ppe*RS1456pee - 1./18.*m11*m33*pow(RS1456ppe,2) + 1./
         9.*m11*m33*RS145pe*RS14563pe + 1./9.*m11*m33*RS145pe*RS1456pee
          + 1./9.*m11*m33*RS145pe*RS1456ppe - 1./18.*m11*m33*pow(
         RS145pe,2) + 1./9.*m11*m33*RS146pe*RS14563pe + 1./9.*m11*m33*
         RS146pe*RS1456pee + 1./9.*m11*m33*RS146pe*RS1456ppe - 1./9.*
         m11*m33*RS146pe*RS145pe - 1./18.*m11*m33*pow(RS146pe,2) + 1./9.
         *m11*m33*RS156pe*RS14563pe );

      M13p6x23R +=  + hh1(1,m13,m11,m11,m13,mu2) * ( 1./9.*m11*m33*
         RS156pe*RS1456pee + 1./9.*m11*m33*RS156pe*RS1456ppe - 1./9.*
         m11*m33*RS156pe*RS145pe - 1./9.*m11*m33*RS156pe*RS146pe - 1./
         18.*m11*m33*pow(RS156pe,2) - 1./4.*pow(m11,2) - 1./54.*pow(
         m11,2)*pow(RS14563pe,2) - 1./27.*pow(m11,2)*RS1456pee*
         RS14563pe - 1./36.*pow(m11,2)*pow(RS1456pee,2) - 1./27.*pow(
         m11,2)*RS1456ppe*RS14563pe - 1./18.*pow(m11,2)*RS1456ppe*
         RS1456pee - 1./36.*pow(m11,2)*pow(RS1456ppe,2) + 1./27.*pow(
         m11,2)*RS145pe*RS14563pe + 1./18.*pow(m11,2)*RS145pe*RS1456pee
          + 1./18.*pow(m11,2)*RS145pe*RS1456ppe - 1./36.*pow(m11,2)*
         pow(RS145pe,2) + 1./27.*pow(m11,2)*RS146pe*RS14563pe + 1./18.*
         pow(m11,2)*RS146pe*RS1456pee + 1./18.*pow(m11,2)*RS146pe*
         RS1456ppe - 1./18.*pow(m11,2)*RS146pe*RS145pe - 1./36.*pow(
         m11,2)*pow(RS146pe,2) + 1./27.*pow(m11,2)*RS156pe*RS14563pe + 
         1./18.*pow(m11,2)*RS156pe*RS1456pee + 1./18.*pow(m11,2)*
         RS156pe*RS1456ppe );

      M13p6x23R +=  + hh1(1,m13,m11,m11,m13,mu2) * (  - 1./18.*pow(
         m11,2)*RS156pe*RS145pe - 1./18.*pow(m11,2)*RS156pe*RS146pe - 1.
         /36.*pow(m11,2)*pow(RS156pe,2) );

      M13p6x23R +=  + hh1(1,m13,m11,m33,m13,mu2) * (  - 1./27.*m33*
         RD3456ep*RS14563pe - 1./27.*m33*RD1456ep*RS34561pe - 1./54.*
         m33*RD1456ep*RS3456pee - 1./54.*m33*RD1456ep*RS3456ppe + 1./54.
         *m33*RD1456ep*RS345pe + 1./54.*m33*RD1456ep*RS346pe + 1./54.*
         m33*RD1456ep*RS356pe - 1./54.*m33*RS1456pee*RD3456ep - 1./54.*
         m33*RS1456ppe*RD3456ep + 1./54.*m33*RS145pe*RD3456ep + 1./54.*
         m33*RS146pe*RD3456ep + 1./54.*m33*RS156pe*RD3456ep - 1./9.*
         pow(m33,2)*RS14563pe*RS34561pe - 2./27.*pow(m33,2)*RS3456pee*
         RS14563pe - 2./27.*pow(m33,2)*RS3456ppe*RS14563pe + 2./27.*
         pow(m33,2)*RS345pe*RS14563pe + 2./27.*pow(m33,2)*RS346pe*
         RS14563pe + 2./27.*pow(m33,2)*RS356pe*RS14563pe - 1./27.*pow(
         m33,2)*RS1456pee*RS34561pe - 1./27.*pow(m33,2)*RS1456ppe*
         RS34561pe + 1./27.*pow(m33,2)*RS145pe*RS34561pe + 1./27.*pow(
         m33,2)*RS146pe*RS34561pe + 1./27.*pow(m33,2)*RS156pe*RS34561pe
          - 1./27.*m11*RD3456ep*RS14563pe - 1./27.*m11*RD1456ep*
         RS34561pe );

      M13p6x23R +=  + hh1(1,m13,m11,m33,m13,mu2) * (  - 1./54.*m11*
         RD1456ep*RS3456pee - 1./54.*m11*RD1456ep*RS3456ppe + 1./54.*
         m11*RD1456ep*RS345pe + 1./54.*m11*RD1456ep*RS346pe + 1./54.*
         m11*RD1456ep*RS356pe - 1./54.*m11*RS1456pee*RD3456ep - 1./54.*
         m11*RS1456ppe*RD3456ep + 1./54.*m11*RS145pe*RD3456ep + 1./54.*
         m11*RS146pe*RD3456ep + 1./54.*m11*RS156pe*RD3456ep - 2./9.*m11
         *m33*RS14563pe*RS34561pe - 1./9.*m11*m33*RS3456pee*RS14563pe
          - 1./9.*m11*m33*RS3456ppe*RS14563pe + 1./9.*m11*m33*RS345pe*
         RS14563pe + 1./9.*m11*m33*RS346pe*RS14563pe + 1./9.*m11*m33*
         RS356pe*RS14563pe - 1./9.*m11*m33*RS1456pee*RS34561pe - 1./9.*
         m11*m33*RS1456ppe*RS34561pe + 1./9.*m11*m33*RS145pe*RS34561pe
          + 1./9.*m11*m33*RS146pe*RS34561pe + 1./9.*m11*m33*RS156pe*
         RS34561pe - 1./9.*pow(m11,2)*RS14563pe*RS34561pe - 1./27.*pow(
         m11,2)*RS3456pee*RS14563pe - 1./27.*pow(m11,2)*RS3456ppe*
         RS14563pe + 1./27.*pow(m11,2)*RS345pe*RS14563pe + 1./27.*pow(
         m11,2)*RS346pe*RS14563pe );

      M13p6x23R +=  + hh1(1,m13,m11,m33,m13,mu2) * ( 1./27.*pow(m11,2)*
         RS356pe*RS14563pe - 2./27.*pow(m11,2)*RS1456pee*RS34561pe - 2./
         27.*pow(m11,2)*RS1456ppe*RS34561pe + 2./27.*pow(m11,2)*RS145pe
         *RS34561pe + 2./27.*pow(m11,2)*RS146pe*RS34561pe + 2./27.*pow(
         m11,2)*RS156pe*RS34561pe );

      M13p6x23R +=  + hh1(1,m13,m11,me2,m13,mu2) * (  - 1./27.*m33*
         RD1456ep*RSe45613p + 1./54.*m33*RD1456ep*RSe45633p + 1./54.*
         m33*RSe45611p*RD1456ep + 2./27.*m33*me2*RSe45613p*RS14563pe - 
         1./27.*m33*me2*RSe45633p*RS14563pe + 1./27.*m33*me2*RS1456pee*
         RSe45613p - 1./54.*m33*me2*RS1456pee*RSe45633p + 1./27.*m33*
         me2*RS1456ppe*RSe45613p - 1./54.*m33*me2*RS1456ppe*RSe45633p
          - 1./27.*m33*me2*RSe45611p*RS14563pe - 1./54.*m33*me2*
         RSe45611p*RS1456pee - 1./54.*m33*me2*RSe45611p*RS1456ppe - 1./
         27.*m33*me2*RS145pe*RSe45613p + 1./54.*m33*me2*RS145pe*
         RSe45633p + 1./54.*m33*me2*RS145pe*RSe45611p - 1./27.*m33*me2*
         RS146pe*RSe45613p + 1./54.*m33*me2*RS146pe*RSe45633p + 1./54.*
         m33*me2*RS146pe*RSe45611p - 1./27.*m33*me2*RS156pe*RSe45613p
          + 1./54.*m33*me2*RS156pe*RSe45633p + 1./54.*m33*me2*RS156pe*
         RSe45611p - 5./27.*pow(m33,2)*RSe45613p*RS14563pe + 1./9.*pow(
         m33,2)*RSe45633p*RS14563pe - 2./27.*pow(m33,2)*RS1456pee*
         RSe45613p );

      M13p6x23R +=  + hh1(1,m13,m11,me2,m13,mu2) * ( 1./54.*pow(m33,2)*
         RS1456pee*RSe45633p - 2./27.*pow(m33,2)*RS1456ppe*RSe45613p + 
         1./54.*pow(m33,2)*RS1456ppe*RSe45633p + 2./27.*pow(m33,2)*
         RSe45611p*RS14563pe + 1./18.*pow(m33,2)*RSe45611p*RS1456pee + 
         1./18.*pow(m33,2)*RSe45611p*RS1456ppe + 2./27.*pow(m33,2)*
         RS145pe*RSe45613p - 1./54.*pow(m33,2)*RS145pe*RSe45633p - 1./
         18.*pow(m33,2)*RS145pe*RSe45611p + 2./27.*pow(m33,2)*RS146pe*
         RSe45613p - 1./54.*pow(m33,2)*RS146pe*RSe45633p - 1./18.*pow(
         m33,2)*RS146pe*RSe45611p + 2./27.*pow(m33,2)*RS156pe*RSe45613p
          - 1./54.*pow(m33,2)*RS156pe*RSe45633p - 1./18.*pow(m33,2)*
         RS156pe*RSe45611p - 1./27.*m11*RD1456ep*RSe45613p + 1./54.*m11
         *RD1456ep*RSe45633p + 1./54.*m11*RSe45611p*RD1456ep + 2./27.*
         m11*me2*RSe45613p*RS14563pe - 1./27.*m11*me2*RSe45633p*
         RS14563pe + 1./27.*m11*me2*RS1456pee*RSe45613p - 1./54.*m11*
         me2*RS1456pee*RSe45633p + 1./27.*m11*me2*RS1456ppe*RSe45613p
          - 1./54.*m11*me2*RS1456ppe*RSe45633p );

      M13p6x23R +=  + hh1(1,m13,m11,me2,m13,mu2) * (  - 1./27.*m11*me2*
         RSe45611p*RS14563pe - 1./54.*m11*me2*RSe45611p*RS1456pee - 1./
         54.*m11*me2*RSe45611p*RS1456ppe - 1./27.*m11*me2*RS145pe*
         RSe45613p + 1./54.*m11*me2*RS145pe*RSe45633p + 1./54.*m11*me2*
         RS145pe*RSe45611p - 1./27.*m11*me2*RS146pe*RSe45613p + 1./54.*
         m11*me2*RS146pe*RSe45633p + 1./54.*m11*me2*RS146pe*RSe45611p
          - 1./27.*m11*me2*RS156pe*RSe45613p + 1./54.*m11*me2*RS156pe*
         RSe45633p + 1./54.*m11*me2*RS156pe*RSe45611p - 8./27.*m11*m33*
         RSe45613p*RS14563pe + 4./27.*m11*m33*RSe45633p*RS14563pe - 4./
         27.*m11*m33*RS1456pee*RSe45613p + 1./54.*m11*m33*RS1456pee*
         RSe45633p - 4./27.*m11*m33*RS1456ppe*RSe45613p + 1./54.*m11*
         m33*RS1456ppe*RSe45633p + 4./27.*m11*m33*RSe45611p*RS14563pe
          + 7./54.*m11*m33*RSe45611p*RS1456pee + 7./54.*m11*m33*
         RSe45611p*RS1456ppe + 4./27.*m11*m33*RS145pe*RSe45613p - 1./54.
         *m11*m33*RS145pe*RSe45633p - 7./54.*m11*m33*RS145pe*RSe45611p
          + 4./27.*m11*m33*RS146pe*RSe45613p );

      M13p6x23R +=  + hh1(1,m13,m11,me2,m13,mu2) * (  - 1./54.*m11*m33*
         RS146pe*RSe45633p - 7./54.*m11*m33*RS146pe*RSe45611p + 4./27.*
         m11*m33*RS156pe*RSe45613p - 1./54.*m11*m33*RS156pe*RSe45633p
          - 7./54.*m11*m33*RS156pe*RSe45611p - 1./9.*pow(m11,2)*
         RSe45613p*RS14563pe + 1./27.*pow(m11,2)*RSe45633p*RS14563pe - 
         2./27.*pow(m11,2)*RS1456pee*RSe45613p - 2./27.*pow(m11,2)*
         RS1456ppe*RSe45613p + 2./27.*pow(m11,2)*RSe45611p*RS14563pe + 
         2./27.*pow(m11,2)*RSe45611p*RS1456pee + 2./27.*pow(m11,2)*
         RSe45611p*RS1456ppe + 2./27.*pow(m11,2)*RS145pe*RSe45613p - 2./
         27.*pow(m11,2)*RS145pe*RSe45611p + 2./27.*pow(m11,2)*RS146pe*
         RSe45613p - 2./27.*pow(m11,2)*RS146pe*RSe45611p + 2./27.*pow(
         m11,2)*RS156pe*RSe45613p - 2./27.*pow(m11,2)*RS156pe*RSe45611p
          );

      M13p6x23R +=  + hh1(1,m13,m11,mp2,m13,mu2) * (  - 1./27.*m33*
         RD1456ep*RSp45613e + 1./54.*m33*RD1456ep*RSp45633e + 1./54.*
         m33*RSp45611e*RD1456ep + 2./27.*m33*mp2*RSp45613e*RS14563pe - 
         1./27.*m33*mp2*RSp45633e*RS14563pe + 1./27.*m33*mp2*RS1456pee*
         RSp45613e - 1./54.*m33*mp2*RS1456pee*RSp45633e + 1./27.*m33*
         mp2*RS1456ppe*RSp45613e - 1./54.*m33*mp2*RS1456ppe*RSp45633e
          - 1./27.*m33*mp2*RSp45611e*RS14563pe - 1./54.*m33*mp2*
         RSp45611e*RS1456pee - 1./54.*m33*mp2*RSp45611e*RS1456ppe - 1./
         27.*m33*mp2*RS145pe*RSp45613e + 1./54.*m33*mp2*RS145pe*
         RSp45633e + 1./54.*m33*mp2*RS145pe*RSp45611e - 1./27.*m33*mp2*
         RS146pe*RSp45613e + 1./54.*m33*mp2*RS146pe*RSp45633e + 1./54.*
         m33*mp2*RS146pe*RSp45611e - 1./27.*m33*mp2*RS156pe*RSp45613e
          + 1./54.*m33*mp2*RS156pe*RSp45633e + 1./54.*m33*mp2*RS156pe*
         RSp45611e - 5./27.*pow(m33,2)*RSp45613e*RS14563pe + 1./9.*pow(
         m33,2)*RSp45633e*RS14563pe - 2./27.*pow(m33,2)*RS1456pee*
         RSp45613e );

      M13p6x23R +=  + hh1(1,m13,m11,mp2,m13,mu2) * ( 1./54.*pow(m33,2)*
         RS1456pee*RSp45633e - 2./27.*pow(m33,2)*RS1456ppe*RSp45613e + 
         1./54.*pow(m33,2)*RS1456ppe*RSp45633e + 2./27.*pow(m33,2)*
         RSp45611e*RS14563pe + 1./18.*pow(m33,2)*RSp45611e*RS1456pee + 
         1./18.*pow(m33,2)*RSp45611e*RS1456ppe + 2./27.*pow(m33,2)*
         RS145pe*RSp45613e - 1./54.*pow(m33,2)*RS145pe*RSp45633e - 1./
         18.*pow(m33,2)*RS145pe*RSp45611e + 2./27.*pow(m33,2)*RS146pe*
         RSp45613e - 1./54.*pow(m33,2)*RS146pe*RSp45633e - 1./18.*pow(
         m33,2)*RS146pe*RSp45611e + 2./27.*pow(m33,2)*RS156pe*RSp45613e
          - 1./54.*pow(m33,2)*RS156pe*RSp45633e - 1./18.*pow(m33,2)*
         RS156pe*RSp45611e - 1./27.*m11*RD1456ep*RSp45613e + 1./54.*m11
         *RD1456ep*RSp45633e + 1./54.*m11*RSp45611e*RD1456ep + 2./27.*
         m11*mp2*RSp45613e*RS14563pe - 1./27.*m11*mp2*RSp45633e*
         RS14563pe + 1./27.*m11*mp2*RS1456pee*RSp45613e - 1./54.*m11*
         mp2*RS1456pee*RSp45633e + 1./27.*m11*mp2*RS1456ppe*RSp45613e
          - 1./54.*m11*mp2*RS1456ppe*RSp45633e );

      M13p6x23R +=  + hh1(1,m13,m11,mp2,m13,mu2) * (  - 1./27.*m11*mp2*
         RSp45611e*RS14563pe - 1./54.*m11*mp2*RSp45611e*RS1456pee - 1./
         54.*m11*mp2*RSp45611e*RS1456ppe - 1./27.*m11*mp2*RS145pe*
         RSp45613e + 1./54.*m11*mp2*RS145pe*RSp45633e + 1./54.*m11*mp2*
         RS145pe*RSp45611e - 1./27.*m11*mp2*RS146pe*RSp45613e + 1./54.*
         m11*mp2*RS146pe*RSp45633e + 1./54.*m11*mp2*RS146pe*RSp45611e
          - 1./27.*m11*mp2*RS156pe*RSp45613e + 1./54.*m11*mp2*RS156pe*
         RSp45633e + 1./54.*m11*mp2*RS156pe*RSp45611e - 8./27.*m11*m33*
         RSp45613e*RS14563pe + 4./27.*m11*m33*RSp45633e*RS14563pe - 4./
         27.*m11*m33*RS1456pee*RSp45613e + 1./54.*m11*m33*RS1456pee*
         RSp45633e - 4./27.*m11*m33*RS1456ppe*RSp45613e + 1./54.*m11*
         m33*RS1456ppe*RSp45633e + 4./27.*m11*m33*RSp45611e*RS14563pe
          + 7./54.*m11*m33*RSp45611e*RS1456pee + 7./54.*m11*m33*
         RSp45611e*RS1456ppe + 4./27.*m11*m33*RS145pe*RSp45613e - 1./54.
         *m11*m33*RS145pe*RSp45633e - 7./54.*m11*m33*RS145pe*RSp45611e
          + 4./27.*m11*m33*RS146pe*RSp45613e );

      M13p6x23R +=  + hh1(1,m13,m11,mp2,m13,mu2) * (  - 1./54.*m11*m33*
         RS146pe*RSp45633e - 7./54.*m11*m33*RS146pe*RSp45611e + 4./27.*
         m11*m33*RS156pe*RSp45613e - 1./54.*m11*m33*RS156pe*RSp45633e
          - 7./54.*m11*m33*RS156pe*RSp45611e - 1./9.*pow(m11,2)*
         RSp45613e*RS14563pe + 1./27.*pow(m11,2)*RSp45633e*RS14563pe - 
         2./27.*pow(m11,2)*RS1456pee*RSp45613e - 2./27.*pow(m11,2)*
         RS1456ppe*RSp45613e + 2./27.*pow(m11,2)*RSp45611e*RS14563pe + 
         2./27.*pow(m11,2)*RSp45611e*RS1456pee + 2./27.*pow(m11,2)*
         RSp45611e*RS1456ppe + 2./27.*pow(m11,2)*RS145pe*RSp45613e - 2./
         27.*pow(m11,2)*RS145pe*RSp45611e + 2./27.*pow(m11,2)*RS146pe*
         RSp45613e - 2./27.*pow(m11,2)*RS146pe*RSp45611e + 2./27.*pow(
         m11,2)*RS156pe*RSp45613e - 2./27.*pow(m11,2)*RS156pe*RSp45611e
          );

      M13p6x23R +=  + hh1(1,m13,m33,m33,m13,mu2) * (  - 1./27.*m33*
         RD3456ep*RS34561pe - 1./54.*m33*RS3456pee*RD3456ep - 1./54.*
         m33*RS3456ppe*RD3456ep + 1./54.*m33*RS345pe*RD3456ep + 1./54.*
         m33*RS346pe*RD3456ep + 1./54.*m33*RS356pe*RD3456ep - 1./4.*
         pow(m33,2) - 1./54.*pow(m33,2)*pow(RS34561pe,2) - 1./27.*pow(
         m33,2)*RS3456pee*RS34561pe - 1./36.*pow(m33,2)*pow(
         RS3456pee,2) - 1./27.*pow(m33,2)*RS3456ppe*RS34561pe - 1./18.*
         pow(m33,2)*RS3456ppe*RS3456pee - 1./36.*pow(m33,2)*pow(
         RS3456ppe,2) + 1./27.*pow(m33,2)*RS345pe*RS34561pe + 1./18.*
         pow(m33,2)*RS345pe*RS3456pee + 1./18.*pow(m33,2)*RS345pe*
         RS3456ppe - 1./36.*pow(m33,2)*pow(RS345pe,2) + 1./27.*pow(
         m33,2)*RS346pe*RS34561pe + 1./18.*pow(m33,2)*RS346pe*RS3456pee
          + 1./18.*pow(m33,2)*RS346pe*RS3456ppe - 1./18.*pow(m33,2)*
         RS346pe*RS345pe - 1./36.*pow(m33,2)*pow(RS346pe,2) + 1./27.*
         pow(m33,2)*RS356pe*RS34561pe + 1./18.*pow(m33,2)*RS356pe*
         RS3456pee );

      M13p6x23R +=  + hh1(1,m13,m33,m33,m13,mu2) * ( 1./18.*pow(m33,2)*
         RS356pe*RS3456ppe - 1./18.*pow(m33,2)*RS356pe*RS345pe - 1./18.
         *pow(m33,2)*RS356pe*RS346pe - 1./36.*pow(m33,2)*pow(RS356pe,2)
          - 1./27.*m11*RD3456ep*RS34561pe - 1./54.*m11*RS3456pee*
         RD3456ep - 1./54.*m11*RS3456ppe*RD3456ep + 1./54.*m11*RS345pe*
         RD3456ep + 1./54.*m11*RS346pe*RD3456ep + 1./54.*m11*RS356pe*
         RD3456ep - 1./2.*m11*m33 - 1./9.*m11*m33*pow(RS34561pe,2) - 1./
         9.*m11*m33*RS3456pee*RS34561pe - 1./18.*m11*m33*pow(
         RS3456pee,2) - 1./9.*m11*m33*RS3456ppe*RS34561pe - 1./9.*m11*
         m33*RS3456ppe*RS3456pee - 1./18.*m11*m33*pow(RS3456ppe,2) + 1./
         9.*m11*m33*RS345pe*RS34561pe + 1./9.*m11*m33*RS345pe*RS3456pee
          + 1./9.*m11*m33*RS345pe*RS3456ppe - 1./18.*m11*m33*pow(
         RS345pe,2) + 1./9.*m11*m33*RS346pe*RS34561pe + 1./9.*m11*m33*
         RS346pe*RS3456pee + 1./9.*m11*m33*RS346pe*RS3456ppe - 1./9.*
         m11*m33*RS346pe*RS345pe - 1./18.*m11*m33*pow(RS346pe,2) + 1./9.
         *m11*m33*RS356pe*RS34561pe );

      M13p6x23R +=  + hh1(1,m13,m33,m33,m13,mu2) * ( 1./9.*m11*m33*
         RS356pe*RS3456pee + 1./9.*m11*m33*RS356pe*RS3456ppe - 1./9.*
         m11*m33*RS356pe*RS345pe - 1./9.*m11*m33*RS356pe*RS346pe - 1./
         18.*m11*m33*pow(RS356pe,2) - 1./4.*pow(m11,2) - 5./54.*pow(
         m11,2)*pow(RS34561pe,2) - 2./27.*pow(m11,2)*RS3456pee*
         RS34561pe - 1./36.*pow(m11,2)*pow(RS3456pee,2) - 2./27.*pow(
         m11,2)*RS3456ppe*RS34561pe - 1./18.*pow(m11,2)*RS3456ppe*
         RS3456pee - 1./36.*pow(m11,2)*pow(RS3456ppe,2) + 2./27.*pow(
         m11,2)*RS345pe*RS34561pe + 1./18.*pow(m11,2)*RS345pe*RS3456pee
          + 1./18.*pow(m11,2)*RS345pe*RS3456ppe - 1./36.*pow(m11,2)*
         pow(RS345pe,2) + 2./27.*pow(m11,2)*RS346pe*RS34561pe + 1./18.*
         pow(m11,2)*RS346pe*RS3456pee + 1./18.*pow(m11,2)*RS346pe*
         RS3456ppe - 1./18.*pow(m11,2)*RS346pe*RS345pe - 1./36.*pow(
         m11,2)*pow(RS346pe,2) + 2./27.*pow(m11,2)*RS356pe*RS34561pe + 
         1./18.*pow(m11,2)*RS356pe*RS3456pee + 1./18.*pow(m11,2)*
         RS356pe*RS3456ppe );

      M13p6x23R +=  + hh1(1,m13,m33,m33,m13,mu2) * (  - 1./18.*pow(
         m11,2)*RS356pe*RS345pe - 1./18.*pow(m11,2)*RS356pe*RS346pe - 1.
         /36.*pow(m11,2)*pow(RS356pe,2) );

      M13p6x23R +=  + hh1(1,m13,m33,me2,m13,mu2) * (  - 1./27.*m33*
         RD3456ep*RSe45613p + 1./54.*m33*RSe45633p*RD3456ep + 1./54.*
         m33*RSe45611p*RD3456ep + 2./27.*m33*me2*RSe45613p*RS34561pe + 
         1./27.*m33*me2*RS3456pee*RSe45613p + 1./27.*m33*me2*RS3456ppe*
         RSe45613p - 1./27.*m33*me2*RSe45633p*RS34561pe - 1./54.*m33*
         me2*RSe45633p*RS3456pee - 1./54.*m33*me2*RSe45633p*RS3456ppe
          - 1./27.*m33*me2*RS345pe*RSe45613p + 1./54.*m33*me2*RS345pe*
         RSe45633p - 1./27.*m33*me2*RS346pe*RSe45613p + 1./54.*m33*me2*
         RS346pe*RSe45633p - 1./27.*m33*me2*RS356pe*RSe45613p + 1./54.*
         m33*me2*RS356pe*RSe45633p - 1./27.*m33*me2*RSe45611p*RS34561pe
          - 1./54.*m33*me2*RSe45611p*RS3456pee - 1./54.*m33*me2*
         RSe45611p*RS3456ppe + 1./54.*m33*me2*RSe45611p*RS345pe + 1./54.
         *m33*me2*RSe45611p*RS346pe + 1./54.*m33*me2*RSe45611p*RS356pe
          - 1./9.*pow(m33,2)*RSe45613p*RS34561pe - 2./27.*pow(m33,2)*
         RS3456pee*RSe45613p - 2./27.*pow(m33,2)*RS3456ppe*RSe45613p + 
         2./27.*pow(m33,2)*RSe45633p*RS34561pe );

      M13p6x23R +=  + hh1(1,m13,m33,me2,m13,mu2) * ( 2./27.*pow(m33,2)*
         RSe45633p*RS3456pee + 2./27.*pow(m33,2)*RSe45633p*RS3456ppe + 
         2./27.*pow(m33,2)*RS345pe*RSe45613p - 2./27.*pow(m33,2)*
         RS345pe*RSe45633p + 2./27.*pow(m33,2)*RS346pe*RSe45613p - 2./
         27.*pow(m33,2)*RS346pe*RSe45633p + 2./27.*pow(m33,2)*RS356pe*
         RSe45613p - 2./27.*pow(m33,2)*RS356pe*RSe45633p + 1./27.*pow(
         m33,2)*RSe45611p*RS34561pe - 1./27.*m11*RD3456ep*RSe45613p + 1.
         /54.*m11*RSe45633p*RD3456ep + 1./54.*m11*RSe45611p*RD3456ep + 
         2./27.*m11*me2*RSe45613p*RS34561pe + 1./27.*m11*me2*RS3456pee*
         RSe45613p + 1./27.*m11*me2*RS3456ppe*RSe45613p - 1./27.*m11*
         me2*RSe45633p*RS34561pe - 1./54.*m11*me2*RSe45633p*RS3456pee
          - 1./54.*m11*me2*RSe45633p*RS3456ppe - 1./27.*m11*me2*RS345pe
         *RSe45613p + 1./54.*m11*me2*RS345pe*RSe45633p - 1./27.*m11*me2
         *RS346pe*RSe45613p + 1./54.*m11*me2*RS346pe*RSe45633p - 1./27.
         *m11*me2*RS356pe*RSe45613p + 1./54.*m11*me2*RS356pe*RSe45633p
          - 1./27.*m11*me2*RSe45611p*RS34561pe );

      M13p6x23R +=  + hh1(1,m13,m33,me2,m13,mu2) * (  - 1./54.*m11*me2*
         RSe45611p*RS3456pee - 1./54.*m11*me2*RSe45611p*RS3456ppe + 1./
         54.*m11*me2*RSe45611p*RS345pe + 1./54.*m11*me2*RSe45611p*
         RS346pe + 1./54.*m11*me2*RSe45611p*RS356pe - 8./27.*m11*m33*
         RSe45613p*RS34561pe - 4./27.*m11*m33*RS3456pee*RSe45613p - 4./
         27.*m11*m33*RS3456ppe*RSe45613p + 4./27.*m11*m33*RSe45633p*
         RS34561pe + 7./54.*m11*m33*RSe45633p*RS3456pee + 7./54.*m11*
         m33*RSe45633p*RS3456ppe + 4./27.*m11*m33*RS345pe*RSe45613p - 7.
         /54.*m11*m33*RS345pe*RSe45633p + 4./27.*m11*m33*RS346pe*
         RSe45613p - 7./54.*m11*m33*RS346pe*RSe45633p + 4./27.*m11*m33*
         RS356pe*RSe45613p - 7./54.*m11*m33*RS356pe*RSe45633p + 4./27.*
         m11*m33*RSe45611p*RS34561pe + 1./54.*m11*m33*RSe45611p*
         RS3456pee + 1./54.*m11*m33*RSe45611p*RS3456ppe - 1./54.*m11*
         m33*RSe45611p*RS345pe - 1./54.*m11*m33*RSe45611p*RS346pe - 1./
         54.*m11*m33*RSe45611p*RS356pe - 5./27.*pow(m11,2)*RSe45613p*
         RS34561pe );

      M13p6x23R +=  + hh1(1,m13,m33,me2,m13,mu2) * (  - 2./27.*pow(
         m11,2)*RS3456pee*RSe45613p - 2./27.*pow(m11,2)*RS3456ppe*
         RSe45613p + 2./27.*pow(m11,2)*RSe45633p*RS34561pe + 1./18.*
         pow(m11,2)*RSe45633p*RS3456pee + 1./18.*pow(m11,2)*RSe45633p*
         RS3456ppe + 2./27.*pow(m11,2)*RS345pe*RSe45613p - 1./18.*pow(
         m11,2)*RS345pe*RSe45633p + 2./27.*pow(m11,2)*RS346pe*RSe45613p
          - 1./18.*pow(m11,2)*RS346pe*RSe45633p + 2./27.*pow(m11,2)*
         RS356pe*RSe45613p - 1./18.*pow(m11,2)*RS356pe*RSe45633p + 1./9.
         *pow(m11,2)*RSe45611p*RS34561pe + 1./54.*pow(m11,2)*RSe45611p*
         RS3456pee + 1./54.*pow(m11,2)*RSe45611p*RS3456ppe - 1./54.*
         pow(m11,2)*RSe45611p*RS345pe - 1./54.*pow(m11,2)*RSe45611p*
         RS346pe - 1./54.*pow(m11,2)*RSe45611p*RS356pe );

      M13p6x23R +=  + hh1(1,m13,m33,mp2,m13,mu2) * (  - 1./27.*m33*
         RD3456ep*RSp45613e + 1./54.*m33*RSp45633e*RD3456ep + 1./54.*
         m33*RSp45611e*RD3456ep + 2./27.*m33*mp2*RSp45613e*RS34561pe + 
         1./27.*m33*mp2*RS3456pee*RSp45613e + 1./27.*m33*mp2*RS3456ppe*
         RSp45613e - 1./27.*m33*mp2*RSp45633e*RS34561pe - 1./54.*m33*
         mp2*RSp45633e*RS3456pee - 1./54.*m33*mp2*RSp45633e*RS3456ppe
          - 1./27.*m33*mp2*RS345pe*RSp45613e + 1./54.*m33*mp2*RS345pe*
         RSp45633e - 1./27.*m33*mp2*RS346pe*RSp45613e + 1./54.*m33*mp2*
         RS346pe*RSp45633e - 1./27.*m33*mp2*RS356pe*RSp45613e + 1./54.*
         m33*mp2*RS356pe*RSp45633e - 1./27.*m33*mp2*RSp45611e*RS34561pe
          - 1./54.*m33*mp2*RSp45611e*RS3456pee - 1./54.*m33*mp2*
         RSp45611e*RS3456ppe + 1./54.*m33*mp2*RSp45611e*RS345pe + 1./54.
         *m33*mp2*RSp45611e*RS346pe + 1./54.*m33*mp2*RSp45611e*RS356pe
          - 1./9.*pow(m33,2)*RSp45613e*RS34561pe - 2./27.*pow(m33,2)*
         RS3456pee*RSp45613e - 2./27.*pow(m33,2)*RS3456ppe*RSp45613e + 
         2./27.*pow(m33,2)*RSp45633e*RS34561pe );

      M13p6x23R +=  + hh1(1,m13,m33,mp2,m13,mu2) * ( 2./27.*pow(m33,2)*
         RSp45633e*RS3456pee + 2./27.*pow(m33,2)*RSp45633e*RS3456ppe + 
         2./27.*pow(m33,2)*RS345pe*RSp45613e - 2./27.*pow(m33,2)*
         RS345pe*RSp45633e + 2./27.*pow(m33,2)*RS346pe*RSp45613e - 2./
         27.*pow(m33,2)*RS346pe*RSp45633e + 2./27.*pow(m33,2)*RS356pe*
         RSp45613e - 2./27.*pow(m33,2)*RS356pe*RSp45633e + 1./27.*pow(
         m33,2)*RSp45611e*RS34561pe - 1./27.*m11*RD3456ep*RSp45613e + 1.
         /54.*m11*RSp45633e*RD3456ep + 1./54.*m11*RSp45611e*RD3456ep + 
         2./27.*m11*mp2*RSp45613e*RS34561pe + 1./27.*m11*mp2*RS3456pee*
         RSp45613e + 1./27.*m11*mp2*RS3456ppe*RSp45613e - 1./27.*m11*
         mp2*RSp45633e*RS34561pe - 1./54.*m11*mp2*RSp45633e*RS3456pee
          - 1./54.*m11*mp2*RSp45633e*RS3456ppe - 1./27.*m11*mp2*RS345pe
         *RSp45613e + 1./54.*m11*mp2*RS345pe*RSp45633e - 1./27.*m11*mp2
         *RS346pe*RSp45613e + 1./54.*m11*mp2*RS346pe*RSp45633e - 1./27.
         *m11*mp2*RS356pe*RSp45613e + 1./54.*m11*mp2*RS356pe*RSp45633e
          - 1./27.*m11*mp2*RSp45611e*RS34561pe );

      M13p6x23R +=  + hh1(1,m13,m33,mp2,m13,mu2) * (  - 1./54.*m11*mp2*
         RSp45611e*RS3456pee - 1./54.*m11*mp2*RSp45611e*RS3456ppe + 1./
         54.*m11*mp2*RSp45611e*RS345pe + 1./54.*m11*mp2*RSp45611e*
         RS346pe + 1./54.*m11*mp2*RSp45611e*RS356pe - 8./27.*m11*m33*
         RSp45613e*RS34561pe - 4./27.*m11*m33*RS3456pee*RSp45613e - 4./
         27.*m11*m33*RS3456ppe*RSp45613e + 4./27.*m11*m33*RSp45633e*
         RS34561pe + 7./54.*m11*m33*RSp45633e*RS3456pee + 7./54.*m11*
         m33*RSp45633e*RS3456ppe + 4./27.*m11*m33*RS345pe*RSp45613e - 7.
         /54.*m11*m33*RS345pe*RSp45633e + 4./27.*m11*m33*RS346pe*
         RSp45613e - 7./54.*m11*m33*RS346pe*RSp45633e + 4./27.*m11*m33*
         RS356pe*RSp45613e - 7./54.*m11*m33*RS356pe*RSp45633e + 4./27.*
         m11*m33*RSp45611e*RS34561pe + 1./54.*m11*m33*RSp45611e*
         RS3456pee + 1./54.*m11*m33*RSp45611e*RS3456ppe - 1./54.*m11*
         m33*RSp45611e*RS345pe - 1./54.*m11*m33*RSp45611e*RS346pe - 1./
         54.*m11*m33*RSp45611e*RS356pe - 5./27.*pow(m11,2)*RSp45613e*
         RS34561pe );

      M13p6x23R +=  + hh1(1,m13,m33,mp2,m13,mu2) * (  - 2./27.*pow(
         m11,2)*RS3456pee*RSp45613e - 2./27.*pow(m11,2)*RS3456ppe*
         RSp45613e + 2./27.*pow(m11,2)*RSp45633e*RS34561pe + 1./18.*
         pow(m11,2)*RSp45633e*RS3456pee + 1./18.*pow(m11,2)*RSp45633e*
         RS3456ppe + 2./27.*pow(m11,2)*RS345pe*RSp45613e - 1./18.*pow(
         m11,2)*RS345pe*RSp45633e + 2./27.*pow(m11,2)*RS346pe*RSp45613e
          - 1./18.*pow(m11,2)*RS346pe*RSp45633e + 2./27.*pow(m11,2)*
         RS356pe*RSp45613e - 1./18.*pow(m11,2)*RS356pe*RSp45633e + 1./9.
         *pow(m11,2)*RSp45611e*RS34561pe + 1./54.*pow(m11,2)*RSp45611e*
         RS3456pee + 1./54.*pow(m11,2)*RSp45611e*RS3456ppe - 1./54.*
         pow(m11,2)*RSp45611e*RS345pe - 1./54.*pow(m11,2)*RSp45611e*
         RS346pe - 1./54.*pow(m11,2)*RSp45611e*RS356pe );

      M13p6x23R +=  + hh1(1,m13,me2,me2,m13,mu2) * ( 2./27.*m33*me2*
         pow(RSe45613p,2) - 2./27.*m33*me2*RSe45633p*RSe45613p + 1./54.
         *m33*me2*pow(RSe45633p,2) - 2./27.*m33*me2*RSe45611p*RSe45613p
          + 1./27.*m33*me2*RSe45611p*RSe45633p + 1./54.*m33*me2*pow(
         RSe45611p,2) - 5./54.*pow(m33,2)*pow(RSe45613p,2) + 1./9.*pow(
         m33,2)*RSe45633p*RSe45613p - 5./108.*pow(m33,2)*pow(
         RSe45633p,2) + 2./27.*pow(m33,2)*RSe45611p*RSe45613p - 1./54.*
         pow(m33,2)*RSe45611p*RSe45633p - 1./36.*pow(m33,2)*pow(
         RSe45611p,2) + 2./27.*m11*me2*pow(RSe45613p,2) - 2./27.*m11*
         me2*RSe45633p*RSe45613p + 1./54.*m11*me2*pow(RSe45633p,2) - 2./
         27.*m11*me2*RSe45611p*RSe45613p + 1./27.*m11*me2*RSe45611p*
         RSe45633p + 1./54.*m11*me2*pow(RSe45611p,2) - 5./27.*m11*m33*
         pow(RSe45613p,2) + 5./27.*m11*m33*RSe45633p*RSe45613p - 2./27.
         *m11*m33*pow(RSe45633p,2) + 5./27.*m11*m33*RSe45611p*RSe45613p
          - 1./27.*m11*m33*RSe45611p*RSe45633p - 2./27.*m11*m33*pow(
         RSe45611p,2) );

      M13p6x23R +=  + hh1(1,m13,me2,me2,m13,mu2) * (  - 5./54.*pow(
         m11,2)*pow(RSe45613p,2) + 2./27.*pow(m11,2)*RSe45633p*
         RSe45613p - 1./36.*pow(m11,2)*pow(RSe45633p,2) + 1./9.*pow(
         m11,2)*RSe45611p*RSe45613p - 1./54.*pow(m11,2)*RSe45611p*
         RSe45633p - 5./108.*pow(m11,2)*pow(RSe45611p,2) );

      M13p6x23R +=  + hh1(1,m13,me2,mp2,m13,mu2) * ( 2./27.*m33*mp2*
         RSp45613e*RSe45613p - 1./27.*m33*mp2*RSe45633p*RSp45613e - 1./
         27.*m33*mp2*RSp45633e*RSe45613p + 1./54.*m33*mp2*RSp45633e*
         RSe45633p - 1./27.*m33*mp2*RSe45611p*RSp45613e + 1./54.*m33*
         mp2*RSe45611p*RSp45633e - 1./27.*m33*mp2*RSp45611e*RSe45613p
          + 1./54.*m33*mp2*RSp45611e*RSe45633p + 1./54.*m33*mp2*
         RSp45611e*RSe45611p + 2./27.*m33*me2*RSp45613e*RSe45613p - 1./
         27.*m33*me2*RSe45633p*RSp45613e - 1./27.*m33*me2*RSp45633e*
         RSe45613p + 1./54.*m33*me2*RSp45633e*RSe45633p - 1./27.*m33*
         me2*RSe45611p*RSp45613e + 1./54.*m33*me2*RSe45611p*RSp45633e
          - 1./27.*m33*me2*RSp45611e*RSe45613p + 1./54.*m33*me2*
         RSp45611e*RSe45633p + 1./54.*m33*me2*RSp45611e*RSe45611p - 5./
         27.*pow(m33,2)*RSp45613e*RSe45613p + 1./9.*pow(m33,2)*
         RSe45633p*RSp45613e + 1./9.*pow(m33,2)*RSp45633e*RSe45613p - 5.
         /54.*pow(m33,2)*RSp45633e*RSe45633p + 2./27.*pow(m33,2)*
         RSe45611p*RSp45613e );

      M13p6x23R +=  + hh1(1,m13,me2,mp2,m13,mu2) * (  - 1./54.*pow(
         m33,2)*RSe45611p*RSp45633e + 2./27.*pow(m33,2)*RSp45611e*
         RSe45613p - 1./54.*pow(m33,2)*RSp45611e*RSe45633p - 1./18.*
         pow(m33,2)*RSp45611e*RSe45611p + 2./27.*m11*mp2*RSp45613e*
         RSe45613p - 1./27.*m11*mp2*RSe45633p*RSp45613e - 1./27.*m11*
         mp2*RSp45633e*RSe45613p + 1./54.*m11*mp2*RSp45633e*RSe45633p
          - 1./27.*m11*mp2*RSe45611p*RSp45613e + 1./54.*m11*mp2*
         RSe45611p*RSp45633e - 1./27.*m11*mp2*RSp45611e*RSe45613p + 1./
         54.*m11*mp2*RSp45611e*RSe45633p + 1./54.*m11*mp2*RSp45611e*
         RSe45611p + 2./27.*m11*me2*RSp45613e*RSe45613p - 1./27.*m11*
         me2*RSe45633p*RSp45613e - 1./27.*m11*me2*RSp45633e*RSe45613p
          + 1./54.*m11*me2*RSp45633e*RSe45633p - 1./27.*m11*me2*
         RSe45611p*RSp45613e + 1./54.*m11*me2*RSe45611p*RSp45633e - 1./
         27.*m11*me2*RSp45611e*RSe45613p + 1./54.*m11*me2*RSp45611e*
         RSe45633p + 1./54.*m11*me2*RSp45611e*RSe45611p - 10./27.*m11*
         m33*RSp45613e*RSe45613p );

      M13p6x23R +=  + hh1(1,m13,me2,mp2,m13,mu2) * ( 5./27.*m11*m33*
         RSe45633p*RSp45613e + 5./27.*m11*m33*RSp45633e*RSe45613p - 4./
         27.*m11*m33*RSp45633e*RSe45633p + 5./27.*m11*m33*RSe45611p*
         RSp45613e - 1./27.*m11*m33*RSe45611p*RSp45633e + 5./27.*m11*
         m33*RSp45611e*RSe45613p - 1./27.*m11*m33*RSp45611e*RSe45633p
          - 4./27.*m11*m33*RSp45611e*RSe45611p - 5./27.*pow(m11,2)*
         RSp45613e*RSe45613p + 2./27.*pow(m11,2)*RSe45633p*RSp45613e + 
         2./27.*pow(m11,2)*RSp45633e*RSe45613p - 1./18.*pow(m11,2)*
         RSp45633e*RSe45633p + 1./9.*pow(m11,2)*RSe45611p*RSp45613e - 1.
         /54.*pow(m11,2)*RSe45611p*RSp45633e + 1./9.*pow(m11,2)*
         RSp45611e*RSe45613p - 1./54.*pow(m11,2)*RSp45611e*RSe45633p - 
         5./54.*pow(m11,2)*RSp45611e*RSe45611p );

      M13p6x23R +=  + hh1(1,m13,mp2,mp2,m13,mu2) * ( 2./27.*m33*mp2*
         pow(RSp45613e,2) - 2./27.*m33*mp2*RSp45633e*RSp45613e + 1./54.
         *m33*mp2*pow(RSp45633e,2) - 2./27.*m33*mp2*RSp45611e*RSp45613e
          + 1./27.*m33*mp2*RSp45611e*RSp45633e + 1./54.*m33*mp2*pow(
         RSp45611e,2) - 5./54.*pow(m33,2)*pow(RSp45613e,2) + 1./9.*pow(
         m33,2)*RSp45633e*RSp45613e - 5./108.*pow(m33,2)*pow(
         RSp45633e,2) + 2./27.*pow(m33,2)*RSp45611e*RSp45613e - 1./54.*
         pow(m33,2)*RSp45611e*RSp45633e - 1./36.*pow(m33,2)*pow(
         RSp45611e,2) + 2./27.*m11*mp2*pow(RSp45613e,2) - 2./27.*m11*
         mp2*RSp45633e*RSp45613e + 1./54.*m11*mp2*pow(RSp45633e,2) - 2./
         27.*m11*mp2*RSp45611e*RSp45613e + 1./27.*m11*mp2*RSp45611e*
         RSp45633e + 1./54.*m11*mp2*pow(RSp45611e,2) - 5./27.*m11*m33*
         pow(RSp45613e,2) + 5./27.*m11*m33*RSp45633e*RSp45613e - 2./27.
         *m11*m33*pow(RSp45633e,2) + 5./27.*m11*m33*RSp45611e*RSp45613e
          - 1./27.*m11*m33*RSp45611e*RSp45633e - 2./27.*m11*m33*pow(
         RSp45611e,2) );

      M13p6x23R +=  + hh1(1,m13,mp2,mp2,m13,mu2) * (  - 5./54.*pow(
         m11,2)*pow(RSp45613e,2) + 2./27.*pow(m11,2)*RSp45633e*
         RSp45613e - 1./36.*pow(m11,2)*pow(RSp45633e,2) + 1./9.*pow(
         m11,2)*RSp45611e*RSp45613e - 1./54.*pow(m11,2)*RSp45611e*
         RSp45633e - 5./108.*pow(m11,2)*pow(RSp45611e,2) );

      M13p6x23R +=  + hh1(1,m14,m11,m34,m13,mu2) * (  - 1./18.*m33*
         RD1456ep - 1./9.*m33*m44*RS156pe + 1./18.*pow(m33,2)*RS14563pe
          - 1./6.*pow(m33,2)*RS156pe - 1./18.*m11*RD1456ep - 1./9.*m11*
         m44*RS156pe - 2./9.*m11*m33*RS156pe - 1./18.*pow(m11,2)*
         RS14563pe - 1./18.*pow(m11,2)*RS156pe );

      M13p6x23R +=  + hh1(1,m14,m33,m34,m13,mu2) * (  - 1./6.*pow(
         m33,2)*RS34561pe + 1./6.*pow(m33,2)*RS356pe - 1./3.*m11*m33*
         RS34561pe + 1./3.*m11*m33*RS356pe - 1./6.*pow(m11,2)*RS34561pe
          + 1./6.*pow(m11,2)*RS356pe );

      M13p6x23R +=  + hh1(1,m14,m34,me2,m13,mu2) * ( 1./18.*m33*me2*
         RSe45633p - 1./9.*m33*me2*RSe563p - 1./18.*m33*me2*RSe45611p
          + 1./9.*m33*me2*RSe561p + 1./9.*m33*m44*RSe563p - 1./9.*m33*
         m44*RSe561p + 1./18.*pow(m33,2)*RSe45613p - 1./18.*pow(m33,2)*
         RSe45633p + 1./6.*pow(m33,2)*RSe563p - 1./6.*pow(m33,2)*
         RSe561p + 1./18.*m11*me2*RSe45633p - 1./9.*m11*me2*RSe563p - 1.
         /18.*m11*me2*RSe45611p + 1./9.*m11*me2*RSe561p + 1./9.*m11*m44
         *RSe563p - 1./9.*m11*m44*RSe561p - 1./18.*m11*m33*RSe45633p + 
         1./3.*m11*m33*RSe563p + 1./18.*m11*m33*RSe45611p - 1./3.*m11*
         m33*RSe561p - 1./18.*pow(m11,2)*RSe45613p + 1./6.*pow(m11,2)*
         RSe563p + 1./18.*pow(m11,2)*RSe45611p - 1./6.*pow(m11,2)*
         RSe561p );

      M13p6x23R +=  + hh1(1,m14,m34,mp2,m13,mu2) * ( 1./18.*m33*mp2*
         RSp45633e - 1./9.*m33*mp2*RSp563e - 1./18.*m33*mp2*RSp45611e
          + 1./9.*m33*mp2*RSp561e + 1./9.*m33*m44*RSp563e - 1./9.*m33*
         m44*RSp561e + 1./18.*pow(m33,2)*RSp45613e - 1./18.*pow(m33,2)*
         RSp45633e + 1./6.*pow(m33,2)*RSp563e - 1./6.*pow(m33,2)*
         RSp561e + 1./18.*m11*mp2*RSp45633e - 1./9.*m11*mp2*RSp563e - 1.
         /18.*m11*mp2*RSp45611e + 1./9.*m11*mp2*RSp561e + 1./9.*m11*m44
         *RSp563e - 1./9.*m11*m44*RSp561e - 1./18.*m11*m33*RSp45633e + 
         1./3.*m11*m33*RSp563e + 1./18.*m11*m33*RSp45611e - 1./3.*m11*
         m33*RSp561e - 1./18.*pow(m11,2)*RSp45613e + 1./6.*pow(m11,2)*
         RSp563e + 1./18.*pow(m11,2)*RSp45611e - 1./6.*pow(m11,2)*
         RSp561e );

      M13p6x23R +=  + hh1(1,m15,m11,m35,m13,mu2) * (  - 1./18.*m33*
         RD1456ep - 1./9.*m33*m55*RS146pe + 1./18.*pow(m33,2)*RS14563pe
          - 1./6.*pow(m33,2)*RS146pe - 1./18.*m11*RD1456ep - 1./9.*m11*
         m55*RS146pe - 2./9.*m11*m33*RS146pe - 1./18.*pow(m11,2)*
         RS14563pe - 1./18.*pow(m11,2)*RS146pe );

      M13p6x23R +=  + hh1(1,m15,m33,m35,m13,mu2) * (  - 1./6.*pow(
         m33,2)*RS34561pe + 1./6.*pow(m33,2)*RS346pe - 1./3.*m11*m33*
         RS34561pe + 1./3.*m11*m33*RS346pe - 1./6.*pow(m11,2)*RS34561pe
          + 1./6.*pow(m11,2)*RS346pe );

      M13p6x23R +=  + hh1(1,m15,m35,me2,m13,mu2) * ( 1./18.*m33*me2*
         RSe45633p - 1./9.*m33*me2*RSe463p - 1./18.*m33*me2*RSe45611p
          + 1./9.*m33*me2*RSe461p + 1./9.*m33*m55*RSe463p - 1./9.*m33*
         m55*RSe461p + 1./18.*pow(m33,2)*RSe45613p - 1./18.*pow(m33,2)*
         RSe45633p + 1./6.*pow(m33,2)*RSe463p - 1./6.*pow(m33,2)*
         RSe461p + 1./18.*m11*me2*RSe45633p - 1./9.*m11*me2*RSe463p - 1.
         /18.*m11*me2*RSe45611p + 1./9.*m11*me2*RSe461p + 1./9.*m11*m55
         *RSe463p - 1./9.*m11*m55*RSe461p - 1./18.*m11*m33*RSe45633p + 
         1./3.*m11*m33*RSe463p + 1./18.*m11*m33*RSe45611p - 1./3.*m11*
         m33*RSe461p - 1./18.*pow(m11,2)*RSe45613p + 1./6.*pow(m11,2)*
         RSe463p + 1./18.*pow(m11,2)*RSe45611p - 1./6.*pow(m11,2)*
         RSe461p );

      M13p6x23R +=  + hh1(1,m15,m35,mp2,m13,mu2) * ( 1./18.*m33*mp2*
         RSp45633e - 1./9.*m33*mp2*RSp463e - 1./18.*m33*mp2*RSp45611e
          + 1./9.*m33*mp2*RSp461e + 1./9.*m33*m55*RSp463e - 1./9.*m33*
         m55*RSp461e + 1./18.*pow(m33,2)*RSp45613e - 1./18.*pow(m33,2)*
         RSp45633e + 1./6.*pow(m33,2)*RSp463e - 1./6.*pow(m33,2)*
         RSp461e + 1./18.*m11*mp2*RSp45633e - 1./9.*m11*mp2*RSp463e - 1.
         /18.*m11*mp2*RSp45611e + 1./9.*m11*mp2*RSp461e + 1./9.*m11*m55
         *RSp463e - 1./9.*m11*m55*RSp461e - 1./18.*m11*m33*RSp45633e + 
         1./3.*m11*m33*RSp463e + 1./18.*m11*m33*RSp45611e - 1./3.*m11*
         m33*RSp461e - 1./18.*pow(m11,2)*RSp45613e + 1./6.*pow(m11,2)*
         RSp463e + 1./18.*pow(m11,2)*RSp45611e - 1./6.*pow(m11,2)*
         RSp461e );

      M13p6x23R +=  + hh1(1,m16,m11,m36,m13,mu2) * (  - 1./18.*m33*
         RD1456ep - 1./9.*m33*m66*RS145pe + 1./18.*pow(m33,2)*RS14563pe
          - 1./6.*pow(m33,2)*RS145pe - 1./18.*m11*RD1456ep - 1./9.*m11*
         m66*RS145pe - 2./9.*m11*m33*RS145pe - 1./18.*pow(m11,2)*
         RS14563pe - 1./18.*pow(m11,2)*RS145pe );

      M13p6x23R +=  + hh1(1,m16,m33,m36,m13,mu2) * (  - 1./6.*pow(
         m33,2)*RS34561pe + 1./6.*pow(m33,2)*RS345pe - 1./3.*m11*m33*
         RS34561pe + 1./3.*m11*m33*RS345pe - 1./6.*pow(m11,2)*RS34561pe
          + 1./6.*pow(m11,2)*RS345pe );

      M13p6x23R +=  + hh1(1,m16,m36,me2,m13,mu2) * ( 1./18.*m33*me2*
         RSe45633p - 1./9.*m33*me2*RSe453p - 1./18.*m33*me2*RSe45611p
          + 1./9.*m33*me2*RSe451p + 1./9.*m33*m66*RSe453p - 1./9.*m33*
         m66*RSe451p + 1./18.*pow(m33,2)*RSe45613p - 1./18.*pow(m33,2)*
         RSe45633p + 1./6.*pow(m33,2)*RSe453p - 1./6.*pow(m33,2)*
         RSe451p + 1./18.*m11*me2*RSe45633p - 1./9.*m11*me2*RSe453p - 1.
         /18.*m11*me2*RSe45611p + 1./9.*m11*me2*RSe451p + 1./9.*m11*m66
         *RSe453p - 1./9.*m11*m66*RSe451p - 1./18.*m11*m33*RSe45633p + 
         1./3.*m11*m33*RSe453p + 1./18.*m11*m33*RSe45611p - 1./3.*m11*
         m33*RSe451p - 1./18.*pow(m11,2)*RSe45613p + 1./6.*pow(m11,2)*
         RSe453p + 1./18.*pow(m11,2)*RSe45611p - 1./6.*pow(m11,2)*
         RSe451p );

      M13p6x23R +=  + hh1(1,m16,m36,mp2,m13,mu2) * ( 1./18.*m33*mp2*
         RSp45633e - 1./9.*m33*mp2*RSp453e - 1./18.*m33*mp2*RSp45611e
          + 1./9.*m33*mp2*RSp451e + 1./9.*m33*m66*RSp453e - 1./9.*m33*
         m66*RSp451e + 1./18.*pow(m33,2)*RSp45613e - 1./18.*pow(m33,2)*
         RSp45633e + 1./6.*pow(m33,2)*RSp453e - 1./6.*pow(m33,2)*
         RSp451e + 1./18.*m11*mp2*RSp45633e - 1./9.*m11*mp2*RSp453e - 1.
         /18.*m11*mp2*RSp45611e + 1./9.*m11*mp2*RSp451e + 1./9.*m11*m66
         *RSp453e - 1./9.*m11*m66*RSp451e - 1./18.*m11*m33*RSp45633e + 
         1./3.*m11*m33*RSp453e + 1./18.*m11*m33*RSp45611e - 1./3.*m11*
         m33*RSp451e - 1./18.*pow(m11,2)*RSp45613e + 1./6.*pow(m11,2)*
         RSp453e + 1./18.*pow(m11,2)*RSp45611e - 1./6.*pow(m11,2)*
         RSp451e );

      M13p6x23R +=  + hh1(1,m34,m14,m33,m13,mu2) * (  - 1./18.*m33*
         RD3456ep - 1./9.*m33*m44*RS356pe - 2./9.*pow(m33,2)*RS34561pe
          + 1./9.*pow(m33,2)*RS356pe - 1./18.*m11*RD3456ep - 1./9.*m11*
         m44*RS356pe - 1./3.*m11*m33*RS34561pe + 1./9.*m11*m33*RS356pe
          - 1./9.*pow(m11,2)*RS34561pe );

      M13p6x23R +=  + hh1(1,m35,m15,m33,m13,mu2) * (  - 1./18.*m33*
         RD3456ep - 1./9.*m33*m55*RS346pe - 2./9.*pow(m33,2)*RS34561pe
          + 1./9.*pow(m33,2)*RS346pe - 1./18.*m11*RD3456ep - 1./9.*m11*
         m55*RS346pe - 1./3.*m11*m33*RS34561pe + 1./9.*m11*m33*RS346pe
          - 1./9.*pow(m11,2)*RS34561pe );

      M13p6x23R +=  + hh1(1,m36,m16,m33,m13,mu2) * (  - 1./18.*m33*
         RD3456ep - 1./9.*m33*m66*RS345pe - 2./9.*pow(m33,2)*RS34561pe
          + 1./9.*pow(m33,2)*RS345pe - 1./18.*m11*RD3456ep - 1./9.*m11*
         m66*RS345pe - 1./3.*m11*m33*RS34561pe + 1./9.*m11*m33*RS345pe
          - 1./9.*pow(m11,2)*RS34561pe );

      M13p6x23R +=  + hh1(1,me2,m14,m34,m13,mu2) * ( 1./18.*m33*me2*
         RSe45613p + 1./18.*m33*me2*RSe45633p - 1./6.*m33*me2*RSe563p
          - 1./18.*m33*me2*RSe561p + 1./9.*m33*me2*RSe564p + 1./9.*m33*
         m44*RSe563p - 1./9.*m33*m44*RSe564p + 1./6.*pow(m33,2)*
         RSe45613p - 1./18.*pow(m33,2)*RSe45633p + 1./18.*pow(m33,2)*
         RSe563p - 1./6.*pow(m33,2)*RSe561p + 1./18.*m11*me2*RSe45613p
          + 1./18.*m11*me2*RSe45633p - 1./6.*m11*me2*RSe563p - 1./18.*
         m11*me2*RSe561p + 1./9.*m11*me2*RSe564p + 1./9.*m11*m44*
         RSe563p - 1./9.*m11*m44*RSe564p + 5./18.*m11*m33*RSe45613p - 1.
         /18.*m11*m33*RSe45633p + 1./18.*m11*m33*RSe563p - 5./18.*m11*
         m33*RSe561p + 1./9.*pow(m11,2)*RSe45613p - 1./9.*pow(m11,2)*
         RSe561p );

      M13p6x23R +=  + hh1(1,me2,m15,m35,m13,mu2) * ( 1./18.*m33*me2*
         RSe45613p + 1./18.*m33*me2*RSe45633p - 1./6.*m33*me2*RSe463p
          - 1./18.*m33*me2*RSe461p + 1./9.*m33*me2*RSe465p + 1./9.*m33*
         m55*RSe463p - 1./9.*m33*m55*RSe465p + 1./6.*pow(m33,2)*
         RSe45613p - 1./18.*pow(m33,2)*RSe45633p + 1./18.*pow(m33,2)*
         RSe463p - 1./6.*pow(m33,2)*RSe461p + 1./18.*m11*me2*RSe45613p
          + 1./18.*m11*me2*RSe45633p - 1./6.*m11*me2*RSe463p - 1./18.*
         m11*me2*RSe461p + 1./9.*m11*me2*RSe465p + 1./9.*m11*m55*
         RSe463p - 1./9.*m11*m55*RSe465p + 5./18.*m11*m33*RSe45613p - 1.
         /18.*m11*m33*RSe45633p + 1./18.*m11*m33*RSe463p - 5./18.*m11*
         m33*RSe461p + 1./9.*pow(m11,2)*RSe45613p - 1./9.*pow(m11,2)*
         RSe461p );

      M13p6x23R +=  + hh1(1,me2,m16,m36,m13,mu2) * ( 1./18.*m33*me2*
         RSe45613p + 1./18.*m33*me2*RSe45633p - 1./6.*m33*me2*RSe453p
          - 1./18.*m33*me2*RSe451p + 1./9.*m33*me2*RSe456p + 1./9.*m33*
         m66*RSe453p - 1./9.*m33*m66*RSe456p + 1./6.*pow(m33,2)*
         RSe45613p - 1./18.*pow(m33,2)*RSe45633p + 1./18.*pow(m33,2)*
         RSe453p - 1./6.*pow(m33,2)*RSe451p + 1./18.*m11*me2*RSe45613p
          + 1./18.*m11*me2*RSe45633p - 1./6.*m11*me2*RSe453p - 1./18.*
         m11*me2*RSe451p + 1./9.*m11*me2*RSe456p + 1./9.*m11*m66*
         RSe453p - 1./9.*m11*m66*RSe456p + 5./18.*m11*m33*RSe45613p - 1.
         /18.*m11*m33*RSe45633p + 1./18.*m11*m33*RSe453p - 5./18.*m11*
         m33*RSe451p + 1./9.*pow(m11,2)*RSe45613p - 1./9.*pow(m11,2)*
         RSe451p );

      M13p6x23R +=  + hh1(1,mp2,m14,m34,m13,mu2) * ( 1./18.*m33*mp2*
         RSp45613e + 1./18.*m33*mp2*RSp45633e - 1./6.*m33*mp2*RSp563e
          - 1./18.*m33*mp2*RSp561e + 1./9.*m33*mp2*RSp564e + 1./9.*m33*
         m44*RSp563e - 1./9.*m33*m44*RSp564e + 1./6.*pow(m33,2)*
         RSp45613e - 1./18.*pow(m33,2)*RSp45633e + 1./18.*pow(m33,2)*
         RSp563e - 1./6.*pow(m33,2)*RSp561e + 1./18.*m11*mp2*RSp45613e
          + 1./18.*m11*mp2*RSp45633e - 1./6.*m11*mp2*RSp563e - 1./18.*
         m11*mp2*RSp561e + 1./9.*m11*mp2*RSp564e + 1./9.*m11*m44*
         RSp563e - 1./9.*m11*m44*RSp564e + 5./18.*m11*m33*RSp45613e - 1.
         /18.*m11*m33*RSp45633e + 1./18.*m11*m33*RSp563e - 5./18.*m11*
         m33*RSp561e + 1./9.*pow(m11,2)*RSp45613e - 1./9.*pow(m11,2)*
         RSp561e );

      M13p6x23R +=  + hh1(1,mp2,m15,m35,m13,mu2) * ( 1./18.*m33*mp2*
         RSp45613e + 1./18.*m33*mp2*RSp45633e - 1./6.*m33*mp2*RSp463e
          - 1./18.*m33*mp2*RSp461e + 1./9.*m33*mp2*RSp465e + 1./9.*m33*
         m55*RSp463e - 1./9.*m33*m55*RSp465e + 1./6.*pow(m33,2)*
         RSp45613e - 1./18.*pow(m33,2)*RSp45633e + 1./18.*pow(m33,2)*
         RSp463e - 1./6.*pow(m33,2)*RSp461e + 1./18.*m11*mp2*RSp45613e
          + 1./18.*m11*mp2*RSp45633e - 1./6.*m11*mp2*RSp463e - 1./18.*
         m11*mp2*RSp461e + 1./9.*m11*mp2*RSp465e + 1./9.*m11*m55*
         RSp463e - 1./9.*m11*m55*RSp465e + 5./18.*m11*m33*RSp45613e - 1.
         /18.*m11*m33*RSp45633e + 1./18.*m11*m33*RSp463e - 5./18.*m11*
         m33*RSp461e + 1./9.*pow(m11,2)*RSp45613e - 1./9.*pow(m11,2)*
         RSp461e );

      M13p6x23R +=  + hh1(1,mp2,m16,m36,m13,mu2) * ( 1./18.*m33*mp2*
         RSp45613e + 1./18.*m33*mp2*RSp45633e - 1./6.*m33*mp2*RSp453e
          - 1./18.*m33*mp2*RSp451e + 1./9.*m33*mp2*RSp456e + 1./9.*m33*
         m66*RSp453e - 1./9.*m33*m66*RSp456e + 1./6.*pow(m33,2)*
         RSp45613e - 1./18.*pow(m33,2)*RSp45633e + 1./18.*pow(m33,2)*
         RSp453e - 1./6.*pow(m33,2)*RSp451e + 1./18.*m11*mp2*RSp45613e
          + 1./18.*m11*mp2*RSp45633e - 1./6.*m11*mp2*RSp453e - 1./18.*
         m11*mp2*RSp451e + 1./9.*m11*mp2*RSp456e + 1./9.*m11*m66*
         RSp453e - 1./9.*m11*m66*RSp456e + 5./18.*m11*m33*RSp45613e - 1.
         /18.*m11*m33*RSp45633e + 1./18.*m11*m33*RSp453e - 5./18.*m11*
         m33*RSp451e + 1./9.*pow(m11,2)*RSp45613e - 1./9.*pow(m11,2)*
         RSp451e );

      M13p6x23R +=  + hh1(3,m13,m11,m11,m13,mu2) * ( 1./54.*m33*pow(
         RD1456ep,2) + 2./27.*pow(m33,2)*RD1456ep*RS14563pe + 1./18.*
         pow(m33,2)*RS1456pee*RD1456ep + 1./18.*pow(m33,2)*RS1456ppe*
         RD1456ep - 1./18.*pow(m33,2)*RS145pe*RD1456ep - 1./18.*pow(
         m33,2)*RS146pe*RD1456ep - 1./18.*pow(m33,2)*RS156pe*RD1456ep
          + 1./54.*m11*pow(RD1456ep,2) + 1./9.*m11*m33*RD1456ep*
         RS14563pe + 1./9.*m11*m33*RS1456pee*RD1456ep + 1./9.*m11*m33*
         RS1456ppe*RD1456ep - 1./9.*m11*m33*RS145pe*RD1456ep - 1./9.*
         m11*m33*RS146pe*RD1456ep - 1./9.*m11*m33*RS156pe*RD1456ep + 1./
         27.*pow(m11,2)*RD1456ep*RS14563pe + 1./18.*pow(m11,2)*
         RS1456pee*RD1456ep + 1./18.*pow(m11,2)*RS1456ppe*RD1456ep - 1./
         18.*pow(m11,2)*RS145pe*RD1456ep - 1./18.*pow(m11,2)*RS146pe*
         RD1456ep - 1./18.*pow(m11,2)*RS156pe*RD1456ep );

      M13p6x23R +=  + hh1(3,m13,m11,m33,m13,mu2) * ( 1./54.*m33*
         RD1456ep*RD3456ep + 1./27.*pow(m33,2)*RD1456ep*RS34561pe + 1./
         54.*m11*RD1456ep*RD3456ep + 1./9.*m11*m33*RD1456ep*RS34561pe
          + 2./27.*pow(m11,2)*RD1456ep*RS34561pe );

      M13p6x23R +=  + hh1(3,m13,m11,me2,m13,mu2) * (  - 1./27.*m33*me2*
         RD1456ep*RSe45613p + 1./54.*m33*me2*RD1456ep*RSe45633p + 1./54.
         *m33*me2*RSe45611p*RD1456ep + 2./27.*pow(m33,2)*RD1456ep*
         RSe45613p - 1./54.*pow(m33,2)*RD1456ep*RSe45633p - 1./18.*pow(
         m33,2)*RSe45611p*RD1456ep - 1./27.*m11*me2*RD1456ep*RSe45613p
          + 1./54.*m11*me2*RD1456ep*RSe45633p + 1./54.*m11*me2*
         RSe45611p*RD1456ep + 4./27.*m11*m33*RD1456ep*RSe45613p - 1./54.
         *m11*m33*RD1456ep*RSe45633p - 7./54.*m11*m33*RSe45611p*
         RD1456ep + 2./27.*pow(m11,2)*RD1456ep*RSe45613p - 2./27.*pow(
         m11,2)*RSe45611p*RD1456ep );

      M13p6x23R +=  + hh1(3,m13,m11,mp2,m13,mu2) * (  - 1./27.*m33*mp2*
         RD1456ep*RSp45613e + 1./54.*m33*mp2*RD1456ep*RSp45633e + 1./54.
         *m33*mp2*RSp45611e*RD1456ep + 2./27.*pow(m33,2)*RD1456ep*
         RSp45613e - 1./54.*pow(m33,2)*RD1456ep*RSp45633e - 1./18.*pow(
         m33,2)*RSp45611e*RD1456ep - 1./27.*m11*mp2*RD1456ep*RSp45613e
          + 1./54.*m11*mp2*RD1456ep*RSp45633e + 1./54.*m11*mp2*
         RSp45611e*RD1456ep + 4./27.*m11*m33*RD1456ep*RSp45613e - 1./54.
         *m11*m33*RD1456ep*RSp45633e - 7./54.*m11*m33*RSp45611e*
         RD1456ep + 2./27.*pow(m11,2)*RD1456ep*RSp45613e - 2./27.*pow(
         m11,2)*RSp45611e*RD1456ep );

      M13p6x23R +=  + hh1(3,m13,m33,m11,m13,mu2) * ( 1./54.*m33*
         RD1456ep*RD3456ep + 2./27.*pow(m33,2)*RD3456ep*RS14563pe + 1./
         54.*m11*RD1456ep*RD3456ep + 1./9.*m11*m33*RD3456ep*RS14563pe
          + 1./27.*pow(m11,2)*RD3456ep*RS14563pe );

      M13p6x23R +=  + hh1(3,m13,m33,m33,m13,mu2) * ( 1./54.*m33*pow(
         RD3456ep,2) + 1./27.*pow(m33,2)*RD3456ep*RS34561pe + 1./18.*
         pow(m33,2)*RS3456pee*RD3456ep + 1./18.*pow(m33,2)*RS3456ppe*
         RD3456ep - 1./18.*pow(m33,2)*RS345pe*RD3456ep - 1./18.*pow(
         m33,2)*RS346pe*RD3456ep - 1./18.*pow(m33,2)*RS356pe*RD3456ep
          + 1./54.*m11*pow(RD3456ep,2) + 1./9.*m11*m33*RD3456ep*
         RS34561pe + 1./9.*m11*m33*RS3456pee*RD3456ep + 1./9.*m11*m33*
         RS3456ppe*RD3456ep - 1./9.*m11*m33*RS345pe*RD3456ep - 1./9.*
         m11*m33*RS346pe*RD3456ep - 1./9.*m11*m33*RS356pe*RD3456ep + 2./
         27.*pow(m11,2)*RD3456ep*RS34561pe + 1./18.*pow(m11,2)*
         RS3456pee*RD3456ep + 1./18.*pow(m11,2)*RS3456ppe*RD3456ep - 1./
         18.*pow(m11,2)*RS345pe*RD3456ep - 1./18.*pow(m11,2)*RS346pe*
         RD3456ep - 1./18.*pow(m11,2)*RS356pe*RD3456ep );

      M13p6x23R +=  + hh1(3,m13,m33,me2,m13,mu2) * (  - 1./27.*m33*me2*
         RD3456ep*RSe45613p + 1./54.*m33*me2*RSe45633p*RD3456ep + 1./54.
         *m33*me2*RSe45611p*RD3456ep + 2./27.*pow(m33,2)*RD3456ep*
         RSe45613p - 2./27.*pow(m33,2)*RSe45633p*RD3456ep - 1./27.*m11*
         me2*RD3456ep*RSe45613p + 1./54.*m11*me2*RSe45633p*RD3456ep + 1.
         /54.*m11*me2*RSe45611p*RD3456ep + 4./27.*m11*m33*RD3456ep*
         RSe45613p - 7./54.*m11*m33*RSe45633p*RD3456ep - 1./54.*m11*m33
         *RSe45611p*RD3456ep + 2./27.*pow(m11,2)*RD3456ep*RSe45613p - 1.
         /18.*pow(m11,2)*RSe45633p*RD3456ep - 1./54.*pow(m11,2)*
         RSe45611p*RD3456ep );

      M13p6x23R +=  + hh1(3,m13,m33,mp2,m13,mu2) * (  - 1./27.*m33*mp2*
         RD3456ep*RSp45613e + 1./54.*m33*mp2*RSp45633e*RD3456ep + 1./54.
         *m33*mp2*RSp45611e*RD3456ep + 2./27.*pow(m33,2)*RD3456ep*
         RSp45613e - 2./27.*pow(m33,2)*RSp45633e*RD3456ep - 1./27.*m11*
         mp2*RD3456ep*RSp45613e + 1./54.*m11*mp2*RSp45633e*RD3456ep + 1.
         /54.*m11*mp2*RSp45611e*RD3456ep + 4./27.*m11*m33*RD3456ep*
         RSp45613e - 7./54.*m11*m33*RSp45633e*RD3456ep - 1./54.*m11*m33
         *RSp45611e*RD3456ep + 2./27.*pow(m11,2)*RD3456ep*RSp45613e - 1.
         /18.*pow(m11,2)*RSp45633e*RD3456ep - 1./54.*pow(m11,2)*
         RSp45611e*RD3456ep );

      M13p6x23R +=  + hh1(7,m13,m11,m11,m13,mu2) * (  - 1./36.*pow(
         m33,2)*pow(RD1456ep,2) - 1./18.*m11*m33*pow(RD1456ep,2) - 1./
         36.*pow(m11,2)*pow(RD1456ep,2) );

      M13p6x23R +=  + hh1(7,m13,m33,m33,m13,mu2) * (  - 1./36.*pow(
         m33,2)*pow(RD3456ep,2) - 1./18.*m11*m33*pow(RD3456ep,2) - 1./
         36.*pow(m11,2)*pow(RD3456ep,2) );

      M13p6x23R +=  + hh21(1,m11,m11,m13,m13,mu2) * (  - 3./16.*pow(
         m33,2) - 3./8.*m11*m33 - 3./16.*pow(m11,2) );

      M13p6x23R +=  + hh21(1,m11,m14,m34,m13,mu2) * (  - 1./16.*pow(
         m33,2)*RS14563pe + 1./16.*pow(m33,2)*RS156pe - 1./8.*m11*m33*
         RS14563pe + 1./8.*m11*m33*RS156pe - 1./16.*pow(m11,2)*
         RS14563pe + 1./16.*pow(m11,2)*RS156pe );

      M13p6x23R +=  + hh21(1,m11,m15,m35,m13,mu2) * (  - 1./16.*pow(
         m33,2)*RS14563pe + 1./16.*pow(m33,2)*RS146pe - 1./8.*m11*m33*
         RS14563pe + 1./8.*m11*m33*RS146pe - 1./16.*pow(m11,2)*
         RS14563pe + 1./16.*pow(m11,2)*RS146pe );

      M13p6x23R +=  + hh21(1,m11,m16,m36,m13,mu2) * (  - 1./16.*pow(
         m33,2)*RS14563pe + 1./16.*pow(m33,2)*RS145pe - 1./8.*m11*m33*
         RS14563pe + 1./8.*m11*m33*RS145pe - 1./16.*pow(m11,2)*
         RS14563pe + 1./16.*pow(m11,2)*RS145pe );

      M13p6x23R +=  + hh21(1,m13,m11,m11,m13,mu2) * ( 3./32.*pow(m33,2)
          + 1./24.*pow(m33,2)*pow(RS14563pe,2) + 1./24.*pow(m33,2)*
         RS1456pee*RS14563pe + 1./96.*pow(m33,2)*pow(RS1456pee,2) + 1./
         24.*pow(m33,2)*RS1456ppe*RS14563pe + 1./48.*pow(m33,2)*
         RS1456ppe*RS1456pee + 1./96.*pow(m33,2)*pow(RS1456ppe,2) - 1./
         24.*pow(m33,2)*RS145pe*RS14563pe - 1./48.*pow(m33,2)*RS145pe*
         RS1456pee - 1./48.*pow(m33,2)*RS145pe*RS1456ppe + 1./96.*pow(
         m33,2)*pow(RS145pe,2) - 1./24.*pow(m33,2)*RS146pe*RS14563pe - 
         1./48.*pow(m33,2)*RS146pe*RS1456pee - 1./48.*pow(m33,2)*
         RS146pe*RS1456ppe + 1./48.*pow(m33,2)*RS146pe*RS145pe + 1./96.
         *pow(m33,2)*pow(RS146pe,2) - 1./24.*pow(m33,2)*RS156pe*
         RS14563pe - 1./48.*pow(m33,2)*RS156pe*RS1456pee - 1./48.*pow(
         m33,2)*RS156pe*RS1456ppe + 1./48.*pow(m33,2)*RS156pe*RS145pe
          + 1./48.*pow(m33,2)*RS156pe*RS146pe + 1./96.*pow(m33,2)*pow(
         RS156pe,2) + 3./16.*m11*m33 + 1./12.*m11*m33*pow(RS14563pe,2)
          + 1./12.*m11*m33*RS1456pee*RS14563pe );

      M13p6x23R +=  + hh21(1,m13,m11,m11,m13,mu2) * ( 1./48.*m11*m33*
         pow(RS1456pee,2) + 1./12.*m11*m33*RS1456ppe*RS14563pe + 1./24.
         *m11*m33*RS1456ppe*RS1456pee + 1./48.*m11*m33*pow(RS1456ppe,2)
          - 1./12.*m11*m33*RS145pe*RS14563pe - 1./24.*m11*m33*RS145pe*
         RS1456pee - 1./24.*m11*m33*RS145pe*RS1456ppe + 1./48.*m11*m33*
         pow(RS145pe,2) - 1./12.*m11*m33*RS146pe*RS14563pe - 1./24.*m11
         *m33*RS146pe*RS1456pee - 1./24.*m11*m33*RS146pe*RS1456ppe + 1./
         24.*m11*m33*RS146pe*RS145pe + 1./48.*m11*m33*pow(RS146pe,2) - 
         1./12.*m11*m33*RS156pe*RS14563pe - 1./24.*m11*m33*RS156pe*
         RS1456pee - 1./24.*m11*m33*RS156pe*RS1456ppe + 1./24.*m11*m33*
         RS156pe*RS145pe + 1./24.*m11*m33*RS156pe*RS146pe + 1./48.*m11*
         m33*pow(RS156pe,2) + 3./32.*pow(m11,2) + 1./24.*pow(m11,2)*
         pow(RS14563pe,2) + 1./24.*pow(m11,2)*RS1456pee*RS14563pe + 1./
         96.*pow(m11,2)*pow(RS1456pee,2) + 1./24.*pow(m11,2)*RS1456ppe*
         RS14563pe + 1./48.*pow(m11,2)*RS1456ppe*RS1456pee + 1./96.*
         pow(m11,2)*pow(RS1456ppe,2) );

      M13p6x23R +=  + hh21(1,m13,m11,m11,m13,mu2) * (  - 1./24.*pow(
         m11,2)*RS145pe*RS14563pe - 1./48.*pow(m11,2)*RS145pe*RS1456pee
          - 1./48.*pow(m11,2)*RS145pe*RS1456ppe + 1./96.*pow(m11,2)*
         pow(RS145pe,2) - 1./24.*pow(m11,2)*RS146pe*RS14563pe - 1./48.*
         pow(m11,2)*RS146pe*RS1456pee - 1./48.*pow(m11,2)*RS146pe*
         RS1456ppe + 1./48.*pow(m11,2)*RS146pe*RS145pe + 1./96.*pow(
         m11,2)*pow(RS146pe,2) - 1./24.*pow(m11,2)*RS156pe*RS14563pe - 
         1./48.*pow(m11,2)*RS156pe*RS1456pee - 1./48.*pow(m11,2)*
         RS156pe*RS1456ppe + 1./48.*pow(m11,2)*RS156pe*RS145pe + 1./48.
         *pow(m11,2)*RS156pe*RS146pe + 1./96.*pow(m11,2)*pow(RS156pe,2)
          );

      M13p6x23R +=  + hh21(1,m13,m11,m33,m13,mu2) * ( 1./12.*pow(m33,2)
         *RS14563pe*RS34561pe + 1./24.*pow(m33,2)*RS3456pee*RS14563pe
          + 1./24.*pow(m33,2)*RS3456ppe*RS14563pe - 1./24.*pow(m33,2)*
         RS345pe*RS14563pe - 1./24.*pow(m33,2)*RS346pe*RS14563pe - 1./
         24.*pow(m33,2)*RS356pe*RS14563pe + 1./24.*pow(m33,2)*RS1456pee
         *RS34561pe + 1./48.*pow(m33,2)*RS1456pee*RS3456pee + 1./48.*
         pow(m33,2)*RS1456pee*RS3456ppe - 1./48.*pow(m33,2)*RS1456pee*
         RS345pe - 1./48.*pow(m33,2)*RS1456pee*RS346pe - 1./48.*pow(
         m33,2)*RS1456pee*RS356pe + 1./24.*pow(m33,2)*RS1456ppe*
         RS34561pe + 1./48.*pow(m33,2)*RS1456ppe*RS3456pee + 1./48.*
         pow(m33,2)*RS1456ppe*RS3456ppe - 1./48.*pow(m33,2)*RS1456ppe*
         RS345pe - 1./48.*pow(m33,2)*RS1456ppe*RS346pe - 1./48.*pow(
         m33,2)*RS1456ppe*RS356pe - 1./24.*pow(m33,2)*RS145pe*RS34561pe
          - 1./48.*pow(m33,2)*RS145pe*RS3456pee - 1./48.*pow(m33,2)*
         RS145pe*RS3456ppe + 1./48.*pow(m33,2)*RS145pe*RS345pe + 1./48.
         *pow(m33,2)*RS145pe*RS346pe );

      M13p6x23R +=  + hh21(1,m13,m11,m33,m13,mu2) * ( 1./48.*pow(m33,2)
         *RS145pe*RS356pe - 1./24.*pow(m33,2)*RS146pe*RS34561pe - 1./48.
         *pow(m33,2)*RS146pe*RS3456pee - 1./48.*pow(m33,2)*RS146pe*
         RS3456ppe + 1./48.*pow(m33,2)*RS146pe*RS345pe + 1./48.*pow(
         m33,2)*RS146pe*RS346pe + 1./48.*pow(m33,2)*RS146pe*RS356pe - 1.
         /24.*pow(m33,2)*RS156pe*RS34561pe - 1./48.*pow(m33,2)*RS156pe*
         RS3456pee - 1./48.*pow(m33,2)*RS156pe*RS3456ppe + 1./48.*pow(
         m33,2)*RS156pe*RS345pe + 1./48.*pow(m33,2)*RS156pe*RS346pe + 1.
         /48.*pow(m33,2)*RS156pe*RS356pe + 1./6.*m11*m33*RS14563pe*
         RS34561pe + 1./12.*m11*m33*RS3456pee*RS14563pe + 1./12.*m11*
         m33*RS3456ppe*RS14563pe - 1./12.*m11*m33*RS345pe*RS14563pe - 1.
         /12.*m11*m33*RS346pe*RS14563pe - 1./12.*m11*m33*RS356pe*
         RS14563pe + 1./12.*m11*m33*RS1456pee*RS34561pe + 1./24.*m11*
         m33*RS1456pee*RS3456pee + 1./24.*m11*m33*RS1456pee*RS3456ppe
          - 1./24.*m11*m33*RS1456pee*RS345pe - 1./24.*m11*m33*RS1456pee
         *RS346pe );

      M13p6x23R +=  + hh21(1,m13,m11,m33,m13,mu2) * (  - 1./24.*m11*m33
         *RS1456pee*RS356pe + 1./12.*m11*m33*RS1456ppe*RS34561pe + 1./
         24.*m11*m33*RS1456ppe*RS3456pee + 1./24.*m11*m33*RS1456ppe*
         RS3456ppe - 1./24.*m11*m33*RS1456ppe*RS345pe - 1./24.*m11*m33*
         RS1456ppe*RS346pe - 1./24.*m11*m33*RS1456ppe*RS356pe - 1./12.*
         m11*m33*RS145pe*RS34561pe - 1./24.*m11*m33*RS145pe*RS3456pee
          - 1./24.*m11*m33*RS145pe*RS3456ppe + 1./24.*m11*m33*RS145pe*
         RS345pe + 1./24.*m11*m33*RS145pe*RS346pe + 1./24.*m11*m33*
         RS145pe*RS356pe - 1./12.*m11*m33*RS146pe*RS34561pe - 1./24.*
         m11*m33*RS146pe*RS3456pee - 1./24.*m11*m33*RS146pe*RS3456ppe
          + 1./24.*m11*m33*RS146pe*RS345pe + 1./24.*m11*m33*RS146pe*
         RS346pe + 1./24.*m11*m33*RS146pe*RS356pe - 1./12.*m11*m33*
         RS156pe*RS34561pe - 1./24.*m11*m33*RS156pe*RS3456pee - 1./24.*
         m11*m33*RS156pe*RS3456ppe + 1./24.*m11*m33*RS156pe*RS345pe + 1.
         /24.*m11*m33*RS156pe*RS346pe + 1./24.*m11*m33*RS156pe*RS356pe
          + 1./12.*pow(m11,2)*RS14563pe*RS34561pe );

      M13p6x23R +=  + hh21(1,m13,m11,m33,m13,mu2) * ( 1./24.*pow(m11,2)
         *RS3456pee*RS14563pe + 1./24.*pow(m11,2)*RS3456ppe*RS14563pe
          - 1./24.*pow(m11,2)*RS345pe*RS14563pe - 1./24.*pow(m11,2)*
         RS346pe*RS14563pe - 1./24.*pow(m11,2)*RS356pe*RS14563pe + 1./
         24.*pow(m11,2)*RS1456pee*RS34561pe + 1./48.*pow(m11,2)*
         RS1456pee*RS3456pee + 1./48.*pow(m11,2)*RS1456pee*RS3456ppe - 
         1./48.*pow(m11,2)*RS1456pee*RS345pe - 1./48.*pow(m11,2)*
         RS1456pee*RS346pe - 1./48.*pow(m11,2)*RS1456pee*RS356pe + 1./
         24.*pow(m11,2)*RS1456ppe*RS34561pe + 1./48.*pow(m11,2)*
         RS1456ppe*RS3456pee + 1./48.*pow(m11,2)*RS1456ppe*RS3456ppe - 
         1./48.*pow(m11,2)*RS1456ppe*RS345pe - 1./48.*pow(m11,2)*
         RS1456ppe*RS346pe - 1./48.*pow(m11,2)*RS1456ppe*RS356pe - 1./
         24.*pow(m11,2)*RS145pe*RS34561pe - 1./48.*pow(m11,2)*RS145pe*
         RS3456pee - 1./48.*pow(m11,2)*RS145pe*RS3456ppe + 1./48.*pow(
         m11,2)*RS145pe*RS345pe + 1./48.*pow(m11,2)*RS145pe*RS346pe + 1.
         /48.*pow(m11,2)*RS145pe*RS356pe );

      M13p6x23R +=  + hh21(1,m13,m11,m33,m13,mu2) * (  - 1./24.*pow(
         m11,2)*RS146pe*RS34561pe - 1./48.*pow(m11,2)*RS146pe*RS3456pee
          - 1./48.*pow(m11,2)*RS146pe*RS3456ppe + 1./48.*pow(m11,2)*
         RS146pe*RS345pe + 1./48.*pow(m11,2)*RS146pe*RS346pe + 1./48.*
         pow(m11,2)*RS146pe*RS356pe - 1./24.*pow(m11,2)*RS156pe*
         RS34561pe - 1./48.*pow(m11,2)*RS156pe*RS3456pee - 1./48.*pow(
         m11,2)*RS156pe*RS3456ppe + 1./48.*pow(m11,2)*RS156pe*RS345pe
          + 1./48.*pow(m11,2)*RS156pe*RS346pe + 1./48.*pow(m11,2)*
         RS156pe*RS356pe );

      M13p6x23R +=  + hh21(1,m13,m11,me2,m13,mu2) * ( 1./12.*pow(m33,2)
         *RSe45613p*RS14563pe - 1./24.*pow(m33,2)*RSe45633p*RS14563pe
          + 1./24.*pow(m33,2)*RS1456pee*RSe45613p - 1./48.*pow(m33,2)*
         RS1456pee*RSe45633p + 1./24.*pow(m33,2)*RS1456ppe*RSe45613p - 
         1./48.*pow(m33,2)*RS1456ppe*RSe45633p - 1./24.*pow(m33,2)*
         RSe45611p*RS14563pe - 1./48.*pow(m33,2)*RSe45611p*RS1456pee - 
         1./48.*pow(m33,2)*RSe45611p*RS1456ppe - 1./24.*pow(m33,2)*
         RS145pe*RSe45613p + 1./48.*pow(m33,2)*RS145pe*RSe45633p + 1./
         48.*pow(m33,2)*RS145pe*RSe45611p - 1./24.*pow(m33,2)*RS146pe*
         RSe45613p + 1./48.*pow(m33,2)*RS146pe*RSe45633p + 1./48.*pow(
         m33,2)*RS146pe*RSe45611p - 1./24.*pow(m33,2)*RS156pe*RSe45613p
          + 1./48.*pow(m33,2)*RS156pe*RSe45633p + 1./48.*pow(m33,2)*
         RS156pe*RSe45611p + 1./6.*m11*m33*RSe45613p*RS14563pe - 1./12.
         *m11*m33*RSe45633p*RS14563pe + 1./12.*m11*m33*RS1456pee*
         RSe45613p - 1./24.*m11*m33*RS1456pee*RSe45633p + 1./12.*m11*
         m33*RS1456ppe*RSe45613p );

      M13p6x23R +=  + hh21(1,m13,m11,me2,m13,mu2) * (  - 1./24.*m11*m33
         *RS1456ppe*RSe45633p - 1./12.*m11*m33*RSe45611p*RS14563pe - 1./
         24.*m11*m33*RSe45611p*RS1456pee - 1./24.*m11*m33*RSe45611p*
         RS1456ppe - 1./12.*m11*m33*RS145pe*RSe45613p + 1./24.*m11*m33*
         RS145pe*RSe45633p + 1./24.*m11*m33*RS145pe*RSe45611p - 1./12.*
         m11*m33*RS146pe*RSe45613p + 1./24.*m11*m33*RS146pe*RSe45633p
          + 1./24.*m11*m33*RS146pe*RSe45611p - 1./12.*m11*m33*RS156pe*
         RSe45613p + 1./24.*m11*m33*RS156pe*RSe45633p + 1./24.*m11*m33*
         RS156pe*RSe45611p + 1./12.*pow(m11,2)*RSe45613p*RS14563pe - 1./
         24.*pow(m11,2)*RSe45633p*RS14563pe + 1./24.*pow(m11,2)*
         RS1456pee*RSe45613p - 1./48.*pow(m11,2)*RS1456pee*RSe45633p + 
         1./24.*pow(m11,2)*RS1456ppe*RSe45613p - 1./48.*pow(m11,2)*
         RS1456ppe*RSe45633p - 1./24.*pow(m11,2)*RSe45611p*RS14563pe - 
         1./48.*pow(m11,2)*RSe45611p*RS1456pee - 1./48.*pow(m11,2)*
         RSe45611p*RS1456ppe - 1./24.*pow(m11,2)*RS145pe*RSe45613p + 1./
         48.*pow(m11,2)*RS145pe*RSe45633p );

      M13p6x23R +=  + hh21(1,m13,m11,me2,m13,mu2) * ( 1./48.*pow(m11,2)
         *RS145pe*RSe45611p - 1./24.*pow(m11,2)*RS146pe*RSe45613p + 1./
         48.*pow(m11,2)*RS146pe*RSe45633p + 1./48.*pow(m11,2)*RS146pe*
         RSe45611p - 1./24.*pow(m11,2)*RS156pe*RSe45613p + 1./48.*pow(
         m11,2)*RS156pe*RSe45633p + 1./48.*pow(m11,2)*RS156pe*RSe45611p
          );

      M13p6x23R +=  + hh21(1,m13,m11,mp2,m13,mu2) * ( 1./12.*pow(m33,2)
         *RSp45613e*RS14563pe - 1./24.*pow(m33,2)*RSp45633e*RS14563pe
          + 1./24.*pow(m33,2)*RS1456pee*RSp45613e - 1./48.*pow(m33,2)*
         RS1456pee*RSp45633e + 1./24.*pow(m33,2)*RS1456ppe*RSp45613e - 
         1./48.*pow(m33,2)*RS1456ppe*RSp45633e - 1./24.*pow(m33,2)*
         RSp45611e*RS14563pe - 1./48.*pow(m33,2)*RSp45611e*RS1456pee - 
         1./48.*pow(m33,2)*RSp45611e*RS1456ppe - 1./24.*pow(m33,2)*
         RS145pe*RSp45613e + 1./48.*pow(m33,2)*RS145pe*RSp45633e + 1./
         48.*pow(m33,2)*RS145pe*RSp45611e - 1./24.*pow(m33,2)*RS146pe*
         RSp45613e + 1./48.*pow(m33,2)*RS146pe*RSp45633e + 1./48.*pow(
         m33,2)*RS146pe*RSp45611e - 1./24.*pow(m33,2)*RS156pe*RSp45613e
          + 1./48.*pow(m33,2)*RS156pe*RSp45633e + 1./48.*pow(m33,2)*
         RS156pe*RSp45611e + 1./6.*m11*m33*RSp45613e*RS14563pe - 1./12.
         *m11*m33*RSp45633e*RS14563pe + 1./12.*m11*m33*RS1456pee*
         RSp45613e - 1./24.*m11*m33*RS1456pee*RSp45633e + 1./12.*m11*
         m33*RS1456ppe*RSp45613e );

      M13p6x23R +=  + hh21(1,m13,m11,mp2,m13,mu2) * (  - 1./24.*m11*m33
         *RS1456ppe*RSp45633e - 1./12.*m11*m33*RSp45611e*RS14563pe - 1./
         24.*m11*m33*RSp45611e*RS1456pee - 1./24.*m11*m33*RSp45611e*
         RS1456ppe - 1./12.*m11*m33*RS145pe*RSp45613e + 1./24.*m11*m33*
         RS145pe*RSp45633e + 1./24.*m11*m33*RS145pe*RSp45611e - 1./12.*
         m11*m33*RS146pe*RSp45613e + 1./24.*m11*m33*RS146pe*RSp45633e
          + 1./24.*m11*m33*RS146pe*RSp45611e - 1./12.*m11*m33*RS156pe*
         RSp45613e + 1./24.*m11*m33*RS156pe*RSp45633e + 1./24.*m11*m33*
         RS156pe*RSp45611e + 1./12.*pow(m11,2)*RSp45613e*RS14563pe - 1./
         24.*pow(m11,2)*RSp45633e*RS14563pe + 1./24.*pow(m11,2)*
         RS1456pee*RSp45613e - 1./48.*pow(m11,2)*RS1456pee*RSp45633e + 
         1./24.*pow(m11,2)*RS1456ppe*RSp45613e - 1./48.*pow(m11,2)*
         RS1456ppe*RSp45633e - 1./24.*pow(m11,2)*RSp45611e*RS14563pe - 
         1./48.*pow(m11,2)*RSp45611e*RS1456pee - 1./48.*pow(m11,2)*
         RSp45611e*RS1456ppe - 1./24.*pow(m11,2)*RS145pe*RSp45613e + 1./
         48.*pow(m11,2)*RS145pe*RSp45633e );

      M13p6x23R +=  + hh21(1,m13,m11,mp2,m13,mu2) * ( 1./48.*pow(m11,2)
         *RS145pe*RSp45611e - 1./24.*pow(m11,2)*RS146pe*RSp45613e + 1./
         48.*pow(m11,2)*RS146pe*RSp45633e + 1./48.*pow(m11,2)*RS146pe*
         RSp45611e - 1./24.*pow(m11,2)*RS156pe*RSp45613e + 1./48.*pow(
         m11,2)*RS156pe*RSp45633e + 1./48.*pow(m11,2)*RS156pe*RSp45611e
          );

      M13p6x23R +=  + hh21(1,m13,m13,m13,m13,mu2) * ( 3./16.*pow(m33,2)
          + 3./8.*m11*m33 + 3./16.*pow(m11,2) );

      M13p6x23R +=  + hh21(1,m13,m33,m33,m13,mu2) * ( 3./32.*pow(m33,2)
          + 1./24.*pow(m33,2)*pow(RS34561pe,2) + 1./24.*pow(m33,2)*
         RS3456pee*RS34561pe + 1./96.*pow(m33,2)*pow(RS3456pee,2) + 1./
         24.*pow(m33,2)*RS3456ppe*RS34561pe + 1./48.*pow(m33,2)*
         RS3456ppe*RS3456pee + 1./96.*pow(m33,2)*pow(RS3456ppe,2) - 1./
         24.*pow(m33,2)*RS345pe*RS34561pe - 1./48.*pow(m33,2)*RS345pe*
         RS3456pee - 1./48.*pow(m33,2)*RS345pe*RS3456ppe + 1./96.*pow(
         m33,2)*pow(RS345pe,2) - 1./24.*pow(m33,2)*RS346pe*RS34561pe - 
         1./48.*pow(m33,2)*RS346pe*RS3456pee - 1./48.*pow(m33,2)*
         RS346pe*RS3456ppe + 1./48.*pow(m33,2)*RS346pe*RS345pe + 1./96.
         *pow(m33,2)*pow(RS346pe,2) - 1./24.*pow(m33,2)*RS356pe*
         RS34561pe - 1./48.*pow(m33,2)*RS356pe*RS3456pee - 1./48.*pow(
         m33,2)*RS356pe*RS3456ppe + 1./48.*pow(m33,2)*RS356pe*RS345pe
          + 1./48.*pow(m33,2)*RS356pe*RS346pe + 1./96.*pow(m33,2)*pow(
         RS356pe,2) + 3./16.*m11*m33 + 1./12.*m11*m33*pow(RS34561pe,2)
          + 1./12.*m11*m33*RS3456pee*RS34561pe );

      M13p6x23R +=  + hh21(1,m13,m33,m33,m13,mu2) * ( 1./48.*m11*m33*
         pow(RS3456pee,2) + 1./12.*m11*m33*RS3456ppe*RS34561pe + 1./24.
         *m11*m33*RS3456ppe*RS3456pee + 1./48.*m11*m33*pow(RS3456ppe,2)
          - 1./12.*m11*m33*RS345pe*RS34561pe - 1./24.*m11*m33*RS345pe*
         RS3456pee - 1./24.*m11*m33*RS345pe*RS3456ppe + 1./48.*m11*m33*
         pow(RS345pe,2) - 1./12.*m11*m33*RS346pe*RS34561pe - 1./24.*m11
         *m33*RS346pe*RS3456pee - 1./24.*m11*m33*RS346pe*RS3456ppe + 1./
         24.*m11*m33*RS346pe*RS345pe + 1./48.*m11*m33*pow(RS346pe,2) - 
         1./12.*m11*m33*RS356pe*RS34561pe - 1./24.*m11*m33*RS356pe*
         RS3456pee - 1./24.*m11*m33*RS356pe*RS3456ppe + 1./24.*m11*m33*
         RS356pe*RS345pe + 1./24.*m11*m33*RS356pe*RS346pe + 1./48.*m11*
         m33*pow(RS356pe,2) + 3./32.*pow(m11,2) + 1./24.*pow(m11,2)*
         pow(RS34561pe,2) + 1./24.*pow(m11,2)*RS3456pee*RS34561pe + 1./
         96.*pow(m11,2)*pow(RS3456pee,2) + 1./24.*pow(m11,2)*RS3456ppe*
         RS34561pe + 1./48.*pow(m11,2)*RS3456ppe*RS3456pee + 1./96.*
         pow(m11,2)*pow(RS3456ppe,2) );

      M13p6x23R +=  + hh21(1,m13,m33,m33,m13,mu2) * (  - 1./24.*pow(
         m11,2)*RS345pe*RS34561pe - 1./48.*pow(m11,2)*RS345pe*RS3456pee
          - 1./48.*pow(m11,2)*RS345pe*RS3456ppe + 1./96.*pow(m11,2)*
         pow(RS345pe,2) - 1./24.*pow(m11,2)*RS346pe*RS34561pe - 1./48.*
         pow(m11,2)*RS346pe*RS3456pee - 1./48.*pow(m11,2)*RS346pe*
         RS3456ppe + 1./48.*pow(m11,2)*RS346pe*RS345pe + 1./96.*pow(
         m11,2)*pow(RS346pe,2) - 1./24.*pow(m11,2)*RS356pe*RS34561pe - 
         1./48.*pow(m11,2)*RS356pe*RS3456pee - 1./48.*pow(m11,2)*
         RS356pe*RS3456ppe + 1./48.*pow(m11,2)*RS356pe*RS345pe + 1./48.
         *pow(m11,2)*RS356pe*RS346pe + 1./96.*pow(m11,2)*pow(RS356pe,2)
          );

      M13p6x23R +=  + hh21(1,m13,m33,me2,m13,mu2) * ( 1./12.*pow(m33,2)
         *RSe45613p*RS34561pe + 1./24.*pow(m33,2)*RS3456pee*RSe45613p
          + 1./24.*pow(m33,2)*RS3456ppe*RSe45613p - 1./24.*pow(m33,2)*
         RSe45633p*RS34561pe - 1./48.*pow(m33,2)*RSe45633p*RS3456pee - 
         1./48.*pow(m33,2)*RSe45633p*RS3456ppe - 1./24.*pow(m33,2)*
         RS345pe*RSe45613p + 1./48.*pow(m33,2)*RS345pe*RSe45633p - 1./
         24.*pow(m33,2)*RS346pe*RSe45613p + 1./48.*pow(m33,2)*RS346pe*
         RSe45633p - 1./24.*pow(m33,2)*RS356pe*RSe45613p + 1./48.*pow(
         m33,2)*RS356pe*RSe45633p - 1./24.*pow(m33,2)*RSe45611p*
         RS34561pe - 1./48.*pow(m33,2)*RSe45611p*RS3456pee - 1./48.*
         pow(m33,2)*RSe45611p*RS3456ppe + 1./48.*pow(m33,2)*RSe45611p*
         RS345pe + 1./48.*pow(m33,2)*RSe45611p*RS346pe + 1./48.*pow(
         m33,2)*RSe45611p*RS356pe + 1./6.*m11*m33*RSe45613p*RS34561pe
          + 1./12.*m11*m33*RS3456pee*RSe45613p + 1./12.*m11*m33*
         RS3456ppe*RSe45613p - 1./12.*m11*m33*RSe45633p*RS34561pe - 1./
         24.*m11*m33*RSe45633p*RS3456pee );

      M13p6x23R +=  + hh21(1,m13,m33,me2,m13,mu2) * (  - 1./24.*m11*m33
         *RSe45633p*RS3456ppe - 1./12.*m11*m33*RS345pe*RSe45613p + 1./
         24.*m11*m33*RS345pe*RSe45633p - 1./12.*m11*m33*RS346pe*
         RSe45613p + 1./24.*m11*m33*RS346pe*RSe45633p - 1./12.*m11*m33*
         RS356pe*RSe45613p + 1./24.*m11*m33*RS356pe*RSe45633p - 1./12.*
         m11*m33*RSe45611p*RS34561pe - 1./24.*m11*m33*RSe45611p*
         RS3456pee - 1./24.*m11*m33*RSe45611p*RS3456ppe + 1./24.*m11*
         m33*RSe45611p*RS345pe + 1./24.*m11*m33*RSe45611p*RS346pe + 1./
         24.*m11*m33*RSe45611p*RS356pe + 1./12.*pow(m11,2)*RSe45613p*
         RS34561pe + 1./24.*pow(m11,2)*RS3456pee*RSe45613p + 1./24.*
         pow(m11,2)*RS3456ppe*RSe45613p - 1./24.*pow(m11,2)*RSe45633p*
         RS34561pe - 1./48.*pow(m11,2)*RSe45633p*RS3456pee - 1./48.*
         pow(m11,2)*RSe45633p*RS3456ppe - 1./24.*pow(m11,2)*RS345pe*
         RSe45613p + 1./48.*pow(m11,2)*RS345pe*RSe45633p - 1./24.*pow(
         m11,2)*RS346pe*RSe45613p + 1./48.*pow(m11,2)*RS346pe*RSe45633p
          - 1./24.*pow(m11,2)*RS356pe*RSe45613p );

      M13p6x23R +=  + hh21(1,m13,m33,me2,m13,mu2) * ( 1./48.*pow(m11,2)
         *RS356pe*RSe45633p - 1./24.*pow(m11,2)*RSe45611p*RS34561pe - 1.
         /48.*pow(m11,2)*RSe45611p*RS3456pee - 1./48.*pow(m11,2)*
         RSe45611p*RS3456ppe + 1./48.*pow(m11,2)*RSe45611p*RS345pe + 1./
         48.*pow(m11,2)*RSe45611p*RS346pe + 1./48.*pow(m11,2)*RSe45611p
         *RS356pe );

      M13p6x23R +=  + hh21(1,m13,m33,mp2,m13,mu2) * ( 1./12.*pow(m33,2)
         *RSp45613e*RS34561pe + 1./24.*pow(m33,2)*RS3456pee*RSp45613e
          + 1./24.*pow(m33,2)*RS3456ppe*RSp45613e - 1./24.*pow(m33,2)*
         RSp45633e*RS34561pe - 1./48.*pow(m33,2)*RSp45633e*RS3456pee - 
         1./48.*pow(m33,2)*RSp45633e*RS3456ppe - 1./24.*pow(m33,2)*
         RS345pe*RSp45613e + 1./48.*pow(m33,2)*RS345pe*RSp45633e - 1./
         24.*pow(m33,2)*RS346pe*RSp45613e + 1./48.*pow(m33,2)*RS346pe*
         RSp45633e - 1./24.*pow(m33,2)*RS356pe*RSp45613e + 1./48.*pow(
         m33,2)*RS356pe*RSp45633e - 1./24.*pow(m33,2)*RSp45611e*
         RS34561pe - 1./48.*pow(m33,2)*RSp45611e*RS3456pee - 1./48.*
         pow(m33,2)*RSp45611e*RS3456ppe + 1./48.*pow(m33,2)*RSp45611e*
         RS345pe + 1./48.*pow(m33,2)*RSp45611e*RS346pe + 1./48.*pow(
         m33,2)*RSp45611e*RS356pe + 1./6.*m11*m33*RSp45613e*RS34561pe
          + 1./12.*m11*m33*RS3456pee*RSp45613e + 1./12.*m11*m33*
         RS3456ppe*RSp45613e - 1./12.*m11*m33*RSp45633e*RS34561pe - 1./
         24.*m11*m33*RSp45633e*RS3456pee );

      M13p6x23R +=  + hh21(1,m13,m33,mp2,m13,mu2) * (  - 1./24.*m11*m33
         *RSp45633e*RS3456ppe - 1./12.*m11*m33*RS345pe*RSp45613e + 1./
         24.*m11*m33*RS345pe*RSp45633e - 1./12.*m11*m33*RS346pe*
         RSp45613e + 1./24.*m11*m33*RS346pe*RSp45633e - 1./12.*m11*m33*
         RS356pe*RSp45613e + 1./24.*m11*m33*RS356pe*RSp45633e - 1./12.*
         m11*m33*RSp45611e*RS34561pe - 1./24.*m11*m33*RSp45611e*
         RS3456pee - 1./24.*m11*m33*RSp45611e*RS3456ppe + 1./24.*m11*
         m33*RSp45611e*RS345pe + 1./24.*m11*m33*RSp45611e*RS346pe + 1./
         24.*m11*m33*RSp45611e*RS356pe + 1./12.*pow(m11,2)*RSp45613e*
         RS34561pe + 1./24.*pow(m11,2)*RS3456pee*RSp45613e + 1./24.*
         pow(m11,2)*RS3456ppe*RSp45613e - 1./24.*pow(m11,2)*RSp45633e*
         RS34561pe - 1./48.*pow(m11,2)*RSp45633e*RS3456pee - 1./48.*
         pow(m11,2)*RSp45633e*RS3456ppe - 1./24.*pow(m11,2)*RS345pe*
         RSp45613e + 1./48.*pow(m11,2)*RS345pe*RSp45633e - 1./24.*pow(
         m11,2)*RS346pe*RSp45613e + 1./48.*pow(m11,2)*RS346pe*RSp45633e
          - 1./24.*pow(m11,2)*RS356pe*RSp45613e );

      M13p6x23R +=  + hh21(1,m13,m33,mp2,m13,mu2) * ( 1./48.*pow(m11,2)
         *RS356pe*RSp45633e - 1./24.*pow(m11,2)*RSp45611e*RS34561pe - 1.
         /48.*pow(m11,2)*RSp45611e*RS3456pee - 1./48.*pow(m11,2)*
         RSp45611e*RS3456ppe + 1./48.*pow(m11,2)*RSp45611e*RS345pe + 1./
         48.*pow(m11,2)*RSp45611e*RS346pe + 1./48.*pow(m11,2)*RSp45611e
         *RS356pe );

      M13p6x23R +=  + hh21(1,m13,me2,me2,m13,mu2) * ( 1./24.*pow(m33,2)
         *pow(RSe45613p,2) - 1./24.*pow(m33,2)*RSe45633p*RSe45613p + 1./
         96.*pow(m33,2)*pow(RSe45633p,2) - 1./24.*pow(m33,2)*RSe45611p*
         RSe45613p + 1./48.*pow(m33,2)*RSe45611p*RSe45633p + 1./96.*
         pow(m33,2)*pow(RSe45611p,2) + 1./12.*m11*m33*pow(RSe45613p,2)
          - 1./12.*m11*m33*RSe45633p*RSe45613p + 1./48.*m11*m33*pow(
         RSe45633p,2) - 1./12.*m11*m33*RSe45611p*RSe45613p + 1./24.*m11
         *m33*RSe45611p*RSe45633p + 1./48.*m11*m33*pow(RSe45611p,2) + 1.
         /24.*pow(m11,2)*pow(RSe45613p,2) - 1./24.*pow(m11,2)*RSe45633p
         *RSe45613p + 1./96.*pow(m11,2)*pow(RSe45633p,2) - 1./24.*pow(
         m11,2)*RSe45611p*RSe45613p + 1./48.*pow(m11,2)*RSe45611p*
         RSe45633p + 1./96.*pow(m11,2)*pow(RSe45611p,2) );

      M13p6x23R +=  + hh21(1,m13,me2,mp2,m13,mu2) * ( 1./12.*pow(m33,2)
         *RSp45613e*RSe45613p - 1./24.*pow(m33,2)*RSe45633p*RSp45613e
          - 1./24.*pow(m33,2)*RSp45633e*RSe45613p + 1./48.*pow(m33,2)*
         RSp45633e*RSe45633p - 1./24.*pow(m33,2)*RSe45611p*RSp45613e + 
         1./48.*pow(m33,2)*RSe45611p*RSp45633e - 1./24.*pow(m33,2)*
         RSp45611e*RSe45613p + 1./48.*pow(m33,2)*RSp45611e*RSe45633p + 
         1./48.*pow(m33,2)*RSp45611e*RSe45611p + 1./6.*m11*m33*
         RSp45613e*RSe45613p - 1./12.*m11*m33*RSe45633p*RSp45613e - 1./
         12.*m11*m33*RSp45633e*RSe45613p + 1./24.*m11*m33*RSp45633e*
         RSe45633p - 1./12.*m11*m33*RSe45611p*RSp45613e + 1./24.*m11*
         m33*RSe45611p*RSp45633e - 1./12.*m11*m33*RSp45611e*RSe45613p
          + 1./24.*m11*m33*RSp45611e*RSe45633p + 1./24.*m11*m33*
         RSp45611e*RSe45611p + 1./12.*pow(m11,2)*RSp45613e*RSe45613p - 
         1./24.*pow(m11,2)*RSe45633p*RSp45613e - 1./24.*pow(m11,2)*
         RSp45633e*RSe45613p + 1./48.*pow(m11,2)*RSp45633e*RSe45633p - 
         1./24.*pow(m11,2)*RSe45611p*RSp45613e );

      M13p6x23R +=  + hh21(1,m13,me2,mp2,m13,mu2) * ( 1./48.*pow(m11,2)
         *RSe45611p*RSp45633e - 1./24.*pow(m11,2)*RSp45611e*RSe45613p
          + 1./48.*pow(m11,2)*RSp45611e*RSe45633p + 1./48.*pow(m11,2)*
         RSp45611e*RSe45611p );

      M13p6x23R +=  + hh21(1,m13,mp2,mp2,m13,mu2) * ( 1./24.*pow(m33,2)
         *pow(RSp45613e,2) - 1./24.*pow(m33,2)*RSp45633e*RSp45613e + 1./
         96.*pow(m33,2)*pow(RSp45633e,2) - 1./24.*pow(m33,2)*RSp45611e*
         RSp45613e + 1./48.*pow(m33,2)*RSp45611e*RSp45633e + 1./96.*
         pow(m33,2)*pow(RSp45611e,2) + 1./12.*m11*m33*pow(RSp45613e,2)
          - 1./12.*m11*m33*RSp45633e*RSp45613e + 1./48.*m11*m33*pow(
         RSp45633e,2) - 1./12.*m11*m33*RSp45611e*RSp45613e + 1./24.*m11
         *m33*RSp45611e*RSp45633e + 1./48.*m11*m33*pow(RSp45611e,2) + 1.
         /24.*pow(m11,2)*pow(RSp45613e,2) - 1./24.*pow(m11,2)*RSp45633e
         *RSp45613e + 1./96.*pow(m11,2)*pow(RSp45633e,2) - 1./24.*pow(
         m11,2)*RSp45611e*RSp45613e + 1./48.*pow(m11,2)*RSp45611e*
         RSp45633e + 1./96.*pow(m11,2)*pow(RSp45611e,2) );

      M13p6x23R +=  + hh21(1,m14,m11,m34,m13,mu2) * ( 1./16.*pow(m33,2)
         *RS14563pe + 1./16.*pow(m33,2)*RS1456pee + 1./16.*pow(m33,2)*
         RS1456ppe - 1./16.*pow(m33,2)*RS145pe - 1./16.*pow(m33,2)*
         RS146pe + 1./8.*m11*m33*RS14563pe + 1./8.*m11*m33*RS1456pee + 
         1./8.*m11*m33*RS1456ppe - 1./8.*m11*m33*RS145pe - 1./8.*m11*
         m33*RS146pe + 1./16.*pow(m11,2)*RS14563pe + 1./16.*pow(m11,2)*
         RS1456pee + 1./16.*pow(m11,2)*RS1456ppe - 1./16.*pow(m11,2)*
         RS145pe - 1./16.*pow(m11,2)*RS146pe );

      M13p6x23R +=  + hh21(1,m14,m33,m34,m13,mu2) * ( 1./16.*pow(m33,2)
         *RS34561pe - 1./16.*pow(m33,2)*RS356pe + 1./8.*m11*m33*
         RS34561pe - 1./8.*m11*m33*RS356pe + 1./16.*pow(m11,2)*
         RS34561pe - 1./16.*pow(m11,2)*RS356pe );

      M13p6x23R +=  + hh21(1,m14,m34,me2,m13,mu2) * ( 1./16.*pow(m33,2)
         *RSe45613p - 1./16.*pow(m33,2)*RSe563p - 1./16.*pow(m33,2)*
         RSe45611p + 1./16.*pow(m33,2)*RSe561p + 1./8.*m11*m33*
         RSe45613p - 1./8.*m11*m33*RSe563p - 1./8.*m11*m33*RSe45611p + 
         1./8.*m11*m33*RSe561p + 1./16.*pow(m11,2)*RSe45613p - 1./16.*
         pow(m11,2)*RSe563p - 1./16.*pow(m11,2)*RSe45611p + 1./16.*pow(
         m11,2)*RSe561p );

      M13p6x23R +=  + hh21(1,m14,m34,mp2,m13,mu2) * ( 1./16.*pow(m33,2)
         *RSp45613e - 1./16.*pow(m33,2)*RSp563e - 1./16.*pow(m33,2)*
         RSp45611e + 1./16.*pow(m33,2)*RSp561e + 1./8.*m11*m33*
         RSp45613e - 1./8.*m11*m33*RSp563e - 1./8.*m11*m33*RSp45611e + 
         1./8.*m11*m33*RSp561e + 1./16.*pow(m11,2)*RSp45613e - 1./16.*
         pow(m11,2)*RSp563e - 1./16.*pow(m11,2)*RSp45611e + 1./16.*pow(
         m11,2)*RSp561e );

      M13p6x23R +=  + hh21(1,m15,m11,m35,m13,mu2) * ( 1./16.*pow(m33,2)
         *RS14563pe + 1./16.*pow(m33,2)*RS1456pee + 1./16.*pow(m33,2)*
         RS1456ppe - 1./16.*pow(m33,2)*RS145pe - 1./16.*pow(m33,2)*
         RS156pe + 1./8.*m11*m33*RS14563pe + 1./8.*m11*m33*RS1456pee + 
         1./8.*m11*m33*RS1456ppe - 1./8.*m11*m33*RS145pe - 1./8.*m11*
         m33*RS156pe + 1./16.*pow(m11,2)*RS14563pe + 1./16.*pow(m11,2)*
         RS1456pee + 1./16.*pow(m11,2)*RS1456ppe - 1./16.*pow(m11,2)*
         RS145pe - 1./16.*pow(m11,2)*RS156pe );

      M13p6x23R +=  + hh21(1,m15,m33,m35,m13,mu2) * ( 1./16.*pow(m33,2)
         *RS34561pe - 1./16.*pow(m33,2)*RS346pe + 1./8.*m11*m33*
         RS34561pe - 1./8.*m11*m33*RS346pe + 1./16.*pow(m11,2)*
         RS34561pe - 1./16.*pow(m11,2)*RS346pe );

      M13p6x23R +=  + hh21(1,m15,m35,me2,m13,mu2) * ( 1./16.*pow(m33,2)
         *RSe45613p - 1./16.*pow(m33,2)*RSe463p - 1./16.*pow(m33,2)*
         RSe45611p + 1./16.*pow(m33,2)*RSe461p + 1./8.*m11*m33*
         RSe45613p - 1./8.*m11*m33*RSe463p - 1./8.*m11*m33*RSe45611p + 
         1./8.*m11*m33*RSe461p + 1./16.*pow(m11,2)*RSe45613p - 1./16.*
         pow(m11,2)*RSe463p - 1./16.*pow(m11,2)*RSe45611p + 1./16.*pow(
         m11,2)*RSe461p );

      M13p6x23R +=  + hh21(1,m15,m35,mp2,m13,mu2) * ( 1./16.*pow(m33,2)
         *RSp45613e - 1./16.*pow(m33,2)*RSp463e - 1./16.*pow(m33,2)*
         RSp45611e + 1./16.*pow(m33,2)*RSp461e + 1./8.*m11*m33*
         RSp45613e - 1./8.*m11*m33*RSp463e - 1./8.*m11*m33*RSp45611e + 
         1./8.*m11*m33*RSp461e + 1./16.*pow(m11,2)*RSp45613e - 1./16.*
         pow(m11,2)*RSp463e - 1./16.*pow(m11,2)*RSp45611e + 1./16.*pow(
         m11,2)*RSp461e );

      M13p6x23R +=  + hh21(1,m16,m11,m36,m13,mu2) * ( 1./16.*pow(m33,2)
         *RS14563pe + 1./16.*pow(m33,2)*RS1456pee + 1./16.*pow(m33,2)*
         RS1456ppe - 1./16.*pow(m33,2)*RS146pe - 1./16.*pow(m33,2)*
         RS156pe + 1./8.*m11*m33*RS14563pe + 1./8.*m11*m33*RS1456pee + 
         1./8.*m11*m33*RS1456ppe - 1./8.*m11*m33*RS146pe - 1./8.*m11*
         m33*RS156pe + 1./16.*pow(m11,2)*RS14563pe + 1./16.*pow(m11,2)*
         RS1456pee + 1./16.*pow(m11,2)*RS1456ppe - 1./16.*pow(m11,2)*
         RS146pe - 1./16.*pow(m11,2)*RS156pe );

      M13p6x23R +=  + hh21(1,m16,m33,m36,m13,mu2) * ( 1./16.*pow(m33,2)
         *RS34561pe - 1./16.*pow(m33,2)*RS345pe + 1./8.*m11*m33*
         RS34561pe - 1./8.*m11*m33*RS345pe + 1./16.*pow(m11,2)*
         RS34561pe - 1./16.*pow(m11,2)*RS345pe );

      M13p6x23R +=  + hh21(1,m16,m36,me2,m13,mu2) * ( 1./16.*pow(m33,2)
         *RSe45613p - 1./16.*pow(m33,2)*RSe453p - 1./16.*pow(m33,2)*
         RSe45611p + 1./16.*pow(m33,2)*RSe451p + 1./8.*m11*m33*
         RSe45613p - 1./8.*m11*m33*RSe453p - 1./8.*m11*m33*RSe45611p + 
         1./8.*m11*m33*RSe451p + 1./16.*pow(m11,2)*RSe45613p - 1./16.*
         pow(m11,2)*RSe453p - 1./16.*pow(m11,2)*RSe45611p + 1./16.*pow(
         m11,2)*RSe451p );

      M13p6x23R +=  + hh21(1,m16,m36,mp2,m13,mu2) * ( 1./16.*pow(m33,2)
         *RSp45613e - 1./16.*pow(m33,2)*RSp453e - 1./16.*pow(m33,2)*
         RSp45611e + 1./16.*pow(m33,2)*RSp451e + 1./8.*m11*m33*
         RSp45613e - 1./8.*m11*m33*RSp453e - 1./8.*m11*m33*RSp45611e + 
         1./8.*m11*m33*RSp451e + 1./16.*pow(m11,2)*RSp45613e - 1./16.*
         pow(m11,2)*RSp453e - 1./16.*pow(m11,2)*RSp45611e + 1./16.*pow(
         m11,2)*RSp451e );

      M13p6x23R +=  + hh21(1,m33,m13,m33,m13,mu2) * (  - 3./16.*pow(
         m33,2) - 3./8.*m11*m33 - 3./16.*pow(m11,2) );

      M13p6x23R +=  + hh21(1,m33,m14,m34,m13,mu2) * (  - 1./16.*pow(
         m33,2)*RS34561pe + 1./16.*pow(m33,2)*RS356pe - 1./8.*m11*m33*
         RS34561pe + 1./8.*m11*m33*RS356pe - 1./16.*pow(m11,2)*
         RS34561pe + 1./16.*pow(m11,2)*RS356pe );

      M13p6x23R +=  + hh21(1,m33,m15,m35,m13,mu2) * (  - 1./16.*pow(
         m33,2)*RS34561pe + 1./16.*pow(m33,2)*RS346pe - 1./8.*m11*m33*
         RS34561pe + 1./8.*m11*m33*RS346pe - 1./16.*pow(m11,2)*
         RS34561pe + 1./16.*pow(m11,2)*RS346pe );

      M13p6x23R +=  + hh21(1,m33,m16,m36,m13,mu2) * (  - 1./16.*pow(
         m33,2)*RS34561pe + 1./16.*pow(m33,2)*RS345pe - 1./8.*m11*m33*
         RS34561pe + 1./8.*m11*m33*RS345pe - 1./16.*pow(m11,2)*
         RS34561pe + 1./16.*pow(m11,2)*RS345pe );

      M13p6x23R +=  + hh21(1,m34,m11,m14,m13,mu2) * ( 1./16.*pow(m33,2)
         *RS14563pe - 1./16.*pow(m33,2)*RS156pe + 1./8.*m11*m33*
         RS14563pe - 1./8.*m11*m33*RS156pe + 1./16.*pow(m11,2)*
         RS14563pe - 1./16.*pow(m11,2)*RS156pe );

      M13p6x23R +=  + hh21(1,m34,m14,m33,m13,mu2) * ( 1./16.*pow(m33,2)
         *RS34561pe + 1./16.*pow(m33,2)*RS3456pee + 1./16.*pow(m33,2)*
         RS3456ppe - 1./16.*pow(m33,2)*RS345pe - 1./16.*pow(m33,2)*
         RS346pe + 1./8.*m11*m33*RS34561pe + 1./8.*m11*m33*RS3456pee + 
         1./8.*m11*m33*RS3456ppe - 1./8.*m11*m33*RS345pe - 1./8.*m11*
         m33*RS346pe + 1./16.*pow(m11,2)*RS34561pe + 1./16.*pow(m11,2)*
         RS3456pee + 1./16.*pow(m11,2)*RS3456ppe - 1./16.*pow(m11,2)*
         RS345pe - 1./16.*pow(m11,2)*RS346pe );

      M13p6x23R +=  + hh21(1,m34,m14,me2,m13,mu2) * ( 1./16.*pow(m33,2)
         *RSe45613p - 1./16.*pow(m33,2)*RSe45633p + 1./16.*pow(m33,2)*
         RSe563p - 1./16.*pow(m33,2)*RSe561p + 1./8.*m11*m33*RSe45613p
          - 1./8.*m11*m33*RSe45633p + 1./8.*m11*m33*RSe563p - 1./8.*m11
         *m33*RSe561p + 1./16.*pow(m11,2)*RSe45613p - 1./16.*pow(m11,2)
         *RSe45633p + 1./16.*pow(m11,2)*RSe563p - 1./16.*pow(m11,2)*
         RSe561p );

      M13p6x23R +=  + hh21(1,m34,m14,mp2,m13,mu2) * ( 1./16.*pow(m33,2)
         *RSp45613e - 1./16.*pow(m33,2)*RSp45633e + 1./16.*pow(m33,2)*
         RSp563e - 1./16.*pow(m33,2)*RSp561e + 1./8.*m11*m33*RSp45613e
          - 1./8.*m11*m33*RSp45633e + 1./8.*m11*m33*RSp563e - 1./8.*m11
         *m33*RSp561e + 1./16.*pow(m11,2)*RSp45613e - 1./16.*pow(m11,2)
         *RSp45633e + 1./16.*pow(m11,2)*RSp563e - 1./16.*pow(m11,2)*
         RSp561e );

      M13p6x23R +=  + hh21(1,m35,m11,m15,m13,mu2) * ( 1./16.*pow(m33,2)
         *RS14563pe - 1./16.*pow(m33,2)*RS146pe + 1./8.*m11*m33*
         RS14563pe - 1./8.*m11*m33*RS146pe + 1./16.*pow(m11,2)*
         RS14563pe - 1./16.*pow(m11,2)*RS146pe );

      M13p6x23R +=  + hh21(1,m35,m15,m33,m13,mu2) * ( 1./16.*pow(m33,2)
         *RS34561pe + 1./16.*pow(m33,2)*RS3456pee + 1./16.*pow(m33,2)*
         RS3456ppe - 1./16.*pow(m33,2)*RS345pe - 1./16.*pow(m33,2)*
         RS356pe + 1./8.*m11*m33*RS34561pe + 1./8.*m11*m33*RS3456pee + 
         1./8.*m11*m33*RS3456ppe - 1./8.*m11*m33*RS345pe - 1./8.*m11*
         m33*RS356pe + 1./16.*pow(m11,2)*RS34561pe + 1./16.*pow(m11,2)*
         RS3456pee + 1./16.*pow(m11,2)*RS3456ppe - 1./16.*pow(m11,2)*
         RS345pe - 1./16.*pow(m11,2)*RS356pe );

      M13p6x23R +=  + hh21(1,m35,m15,me2,m13,mu2) * ( 1./16.*pow(m33,2)
         *RSe45613p - 1./16.*pow(m33,2)*RSe45633p + 1./16.*pow(m33,2)*
         RSe463p - 1./16.*pow(m33,2)*RSe461p + 1./8.*m11*m33*RSe45613p
          - 1./8.*m11*m33*RSe45633p + 1./8.*m11*m33*RSe463p - 1./8.*m11
         *m33*RSe461p + 1./16.*pow(m11,2)*RSe45613p - 1./16.*pow(m11,2)
         *RSe45633p + 1./16.*pow(m11,2)*RSe463p - 1./16.*pow(m11,2)*
         RSe461p );

      M13p6x23R +=  + hh21(1,m35,m15,mp2,m13,mu2) * ( 1./16.*pow(m33,2)
         *RSp45613e - 1./16.*pow(m33,2)*RSp45633e + 1./16.*pow(m33,2)*
         RSp463e - 1./16.*pow(m33,2)*RSp461e + 1./8.*m11*m33*RSp45613e
          - 1./8.*m11*m33*RSp45633e + 1./8.*m11*m33*RSp463e - 1./8.*m11
         *m33*RSp461e + 1./16.*pow(m11,2)*RSp45613e - 1./16.*pow(m11,2)
         *RSp45633e + 1./16.*pow(m11,2)*RSp463e - 1./16.*pow(m11,2)*
         RSp461e );

      M13p6x23R +=  + hh21(1,m36,m11,m16,m13,mu2) * ( 1./16.*pow(m33,2)
         *RS14563pe - 1./16.*pow(m33,2)*RS145pe + 1./8.*m11*m33*
         RS14563pe - 1./8.*m11*m33*RS145pe + 1./16.*pow(m11,2)*
         RS14563pe - 1./16.*pow(m11,2)*RS145pe );

      M13p6x23R +=  + hh21(1,m36,m16,m33,m13,mu2) * ( 1./16.*pow(m33,2)
         *RS34561pe + 1./16.*pow(m33,2)*RS3456pee + 1./16.*pow(m33,2)*
         RS3456ppe - 1./16.*pow(m33,2)*RS346pe - 1./16.*pow(m33,2)*
         RS356pe + 1./8.*m11*m33*RS34561pe + 1./8.*m11*m33*RS3456pee + 
         1./8.*m11*m33*RS3456ppe - 1./8.*m11*m33*RS346pe - 1./8.*m11*
         m33*RS356pe + 1./16.*pow(m11,2)*RS34561pe + 1./16.*pow(m11,2)*
         RS3456pee + 1./16.*pow(m11,2)*RS3456ppe - 1./16.*pow(m11,2)*
         RS346pe - 1./16.*pow(m11,2)*RS356pe );

      M13p6x23R +=  + hh21(1,m36,m16,me2,m13,mu2) * ( 1./16.*pow(m33,2)
         *RSe45613p - 1./16.*pow(m33,2)*RSe45633p + 1./16.*pow(m33,2)*
         RSe453p - 1./16.*pow(m33,2)*RSe451p + 1./8.*m11*m33*RSe45613p
          - 1./8.*m11*m33*RSe45633p + 1./8.*m11*m33*RSe453p - 1./8.*m11
         *m33*RSe451p + 1./16.*pow(m11,2)*RSe45613p - 1./16.*pow(m11,2)
         *RSe45633p + 1./16.*pow(m11,2)*RSe453p - 1./16.*pow(m11,2)*
         RSe451p );

      M13p6x23R +=  + hh21(1,m36,m16,mp2,m13,mu2) * ( 1./16.*pow(m33,2)
         *RSp45613e - 1./16.*pow(m33,2)*RSp45633e + 1./16.*pow(m33,2)*
         RSp453e - 1./16.*pow(m33,2)*RSp451e + 1./8.*m11*m33*RSp45613e
          - 1./8.*m11*m33*RSp45633e + 1./8.*m11*m33*RSp453e - 1./8.*m11
         *m33*RSp451e + 1./16.*pow(m11,2)*RSp45613e - 1./16.*pow(m11,2)
         *RSp45633e + 1./16.*pow(m11,2)*RSp453e - 1./16.*pow(m11,2)*
         RSp451e );

      M13p6x23R +=  + hh21(1,m44,m14,m34,m13,mu2) * ( 3./16.*pow(m33,2)
          - 1./16.*pow(m33,2)*RS456pe + 3./8.*m11*m33 - 1./8.*m11*m33*
         RS456pe + 3./16.*pow(m11,2) - 1./16.*pow(m11,2)*RS456pe );

      M13p6x23R +=  + hh21(1,m45,m14,m35,m13,mu2) * ( 3./16.*pow(m33,2)
          + 3./8.*m11*m33 + 3./16.*pow(m11,2) );

      M13p6x23R +=  + hh21(1,m45,m15,m34,m13,mu2) * ( 3./16.*pow(m33,2)
          + 3./8.*m11*m33 + 3./16.*pow(m11,2) );

      M13p6x23R +=  + hh21(1,m46,m14,m36,m13,mu2) * ( 3./16.*pow(m33,2)
          + 3./8.*m11*m33 + 3./16.*pow(m11,2) );

      M13p6x23R +=  + hh21(1,m46,m16,m34,m13,mu2) * ( 3./16.*pow(m33,2)
          + 3./8.*m11*m33 + 3./16.*pow(m11,2) );

      M13p6x23R +=  + hh21(1,m55,m15,m35,m13,mu2) * ( 3./16.*pow(m33,2)
          - 1./16.*pow(m33,2)*RS546pe + 3./8.*m11*m33 - 1./8.*m11*m33*
         RS546pe + 3./16.*pow(m11,2) - 1./16.*pow(m11,2)*RS546pe );

      M13p6x23R +=  + hh21(1,m56,m15,m36,m13,mu2) * ( 3./16.*pow(m33,2)
          + 3./8.*m11*m33 + 3./16.*pow(m11,2) );

      M13p6x23R +=  + hh21(1,m56,m16,m35,m13,mu2) * ( 3./16.*pow(m33,2)
          + 3./8.*m11*m33 + 3./16.*pow(m11,2) );

      M13p6x23R +=  + hh21(1,m66,m16,m36,m13,mu2) * ( 3./16.*pow(m33,2)
          - 1./16.*pow(m33,2)*RS645pe + 3./8.*m11*m33 - 1./8.*m11*m33*
         RS645pe + 3./16.*pow(m11,2) - 1./16.*pow(m11,2)*RS645pe );

      M13p6x23R +=  + hh21(1,me2,m14,m34,m13,mu2) * (  - 1./16.*pow(
         m33,2)*RSe45613p + 1./16.*pow(m33,2)*RSe563p + 1./16.*pow(
         m33,2)*RSe561p - 1./16.*pow(m33,2)*RSe564p - 1./8.*m11*m33*
         RSe45613p + 1./8.*m11*m33*RSe563p + 1./8.*m11*m33*RSe561p - 1./
         8.*m11*m33*RSe564p - 1./16.*pow(m11,2)*RSe45613p + 1./16.*pow(
         m11,2)*RSe563p + 1./16.*pow(m11,2)*RSe561p - 1./16.*pow(m11,2)
         *RSe564p );

      M13p6x23R +=  + hh21(1,me2,m15,m35,m13,mu2) * (  - 1./16.*pow(
         m33,2)*RSe45613p + 1./16.*pow(m33,2)*RSe463p + 1./16.*pow(
         m33,2)*RSe461p - 1./16.*pow(m33,2)*RSe465p - 1./8.*m11*m33*
         RSe45613p + 1./8.*m11*m33*RSe463p + 1./8.*m11*m33*RSe461p - 1./
         8.*m11*m33*RSe465p - 1./16.*pow(m11,2)*RSe45613p + 1./16.*pow(
         m11,2)*RSe463p + 1./16.*pow(m11,2)*RSe461p - 1./16.*pow(m11,2)
         *RSe465p );

      M13p6x23R +=  + hh21(1,me2,m16,m36,m13,mu2) * (  - 1./16.*pow(
         m33,2)*RSe45613p + 1./16.*pow(m33,2)*RSe453p + 1./16.*pow(
         m33,2)*RSe451p - 1./16.*pow(m33,2)*RSe456p - 1./8.*m11*m33*
         RSe45613p + 1./8.*m11*m33*RSe453p + 1./8.*m11*m33*RSe451p - 1./
         8.*m11*m33*RSe456p - 1./16.*pow(m11,2)*RSe45613p + 1./16.*pow(
         m11,2)*RSe453p + 1./16.*pow(m11,2)*RSe451p - 1./16.*pow(m11,2)
         *RSe456p );

      M13p6x23R +=  + hh21(1,mp2,m14,m34,m13,mu2) * (  - 1./16.*pow(
         m33,2)*RSp45613e + 1./16.*pow(m33,2)*RSp563e + 1./16.*pow(
         m33,2)*RSp561e - 1./16.*pow(m33,2)*RSp564e - 1./8.*m11*m33*
         RSp45613e + 1./8.*m11*m33*RSp563e + 1./8.*m11*m33*RSp561e - 1./
         8.*m11*m33*RSp564e - 1./16.*pow(m11,2)*RSp45613e + 1./16.*pow(
         m11,2)*RSp563e + 1./16.*pow(m11,2)*RSp561e - 1./16.*pow(m11,2)
         *RSp564e );

      M13p6x23R +=  + hh21(1,mp2,m15,m35,m13,mu2) * (  - 1./16.*pow(
         m33,2)*RSp45613e + 1./16.*pow(m33,2)*RSp463e + 1./16.*pow(
         m33,2)*RSp461e - 1./16.*pow(m33,2)*RSp465e - 1./8.*m11*m33*
         RSp45613e + 1./8.*m11*m33*RSp463e + 1./8.*m11*m33*RSp461e - 1./
         8.*m11*m33*RSp465e - 1./16.*pow(m11,2)*RSp45613e + 1./16.*pow(
         m11,2)*RSp463e + 1./16.*pow(m11,2)*RSp461e - 1./16.*pow(m11,2)
         *RSp465e );

      M13p6x23R +=  + hh21(1,mp2,m16,m36,m13,mu2) * (  - 1./16.*pow(
         m33,2)*RSp45613e + 1./16.*pow(m33,2)*RSp453e + 1./16.*pow(
         m33,2)*RSp451e - 1./16.*pow(m33,2)*RSp456e - 1./8.*m11*m33*
         RSp45613e + 1./8.*m11*m33*RSp453e + 1./8.*m11*m33*RSp451e - 1./
         8.*m11*m33*RSp456e - 1./16.*pow(m11,2)*RSp45613e + 1./16.*pow(
         m11,2)*RSp453e + 1./16.*pow(m11,2)*RSp451e - 1./16.*pow(m11,2)
         *RSp456e );

      M13p6x23R +=  + hh21(3,m13,m11,m11,m13,mu2) * (  - 1./24.*pow(
         m33,2)*RD1456ep*RS14563pe - 1./48.*pow(m33,2)*RS1456pee*
         RD1456ep - 1./48.*pow(m33,2)*RS1456ppe*RD1456ep + 1./48.*pow(
         m33,2)*RS145pe*RD1456ep + 1./48.*pow(m33,2)*RS146pe*RD1456ep
          + 1./48.*pow(m33,2)*RS156pe*RD1456ep - 1./12.*m11*m33*
         RD1456ep*RS14563pe - 1./24.*m11*m33*RS1456pee*RD1456ep - 1./24.
         *m11*m33*RS1456ppe*RD1456ep + 1./24.*m11*m33*RS145pe*RD1456ep
          + 1./24.*m11*m33*RS146pe*RD1456ep + 1./24.*m11*m33*RS156pe*
         RD1456ep - 1./24.*pow(m11,2)*RD1456ep*RS14563pe - 1./48.*pow(
         m11,2)*RS1456pee*RD1456ep - 1./48.*pow(m11,2)*RS1456ppe*
         RD1456ep + 1./48.*pow(m11,2)*RS145pe*RD1456ep + 1./48.*pow(
         m11,2)*RS146pe*RD1456ep + 1./48.*pow(m11,2)*RS156pe*RD1456ep )
         ;

      M13p6x23R +=  + hh21(3,m13,m11,m33,m13,mu2) * (  - 1./24.*pow(
         m33,2)*RD1456ep*RS34561pe - 1./48.*pow(m33,2)*RD1456ep*
         RS3456pee - 1./48.*pow(m33,2)*RD1456ep*RS3456ppe + 1./48.*pow(
         m33,2)*RD1456ep*RS345pe + 1./48.*pow(m33,2)*RD1456ep*RS346pe
          + 1./48.*pow(m33,2)*RD1456ep*RS356pe - 1./12.*m11*m33*
         RD1456ep*RS34561pe - 1./24.*m11*m33*RD1456ep*RS3456pee - 1./24.
         *m11*m33*RD1456ep*RS3456ppe + 1./24.*m11*m33*RD1456ep*RS345pe
          + 1./24.*m11*m33*RD1456ep*RS346pe + 1./24.*m11*m33*RD1456ep*
         RS356pe - 1./24.*pow(m11,2)*RD1456ep*RS34561pe - 1./48.*pow(
         m11,2)*RD1456ep*RS3456pee - 1./48.*pow(m11,2)*RD1456ep*
         RS3456ppe + 1./48.*pow(m11,2)*RD1456ep*RS345pe + 1./48.*pow(
         m11,2)*RD1456ep*RS346pe + 1./48.*pow(m11,2)*RD1456ep*RS356pe )
         ;

      M13p6x23R +=  + hh21(3,m13,m11,me2,m13,mu2) * (  - 1./24.*pow(
         m33,2)*RD1456ep*RSe45613p + 1./48.*pow(m33,2)*RD1456ep*
         RSe45633p + 1./48.*pow(m33,2)*RSe45611p*RD1456ep - 1./12.*m11*
         m33*RD1456ep*RSe45613p + 1./24.*m11*m33*RD1456ep*RSe45633p + 1.
         /24.*m11*m33*RSe45611p*RD1456ep - 1./24.*pow(m11,2)*RD1456ep*
         RSe45613p + 1./48.*pow(m11,2)*RD1456ep*RSe45633p + 1./48.*pow(
         m11,2)*RSe45611p*RD1456ep );

      M13p6x23R +=  + hh21(3,m13,m11,mp2,m13,mu2) * (  - 1./24.*pow(
         m33,2)*RD1456ep*RSp45613e + 1./48.*pow(m33,2)*RD1456ep*
         RSp45633e + 1./48.*pow(m33,2)*RSp45611e*RD1456ep - 1./12.*m11*
         m33*RD1456ep*RSp45613e + 1./24.*m11*m33*RD1456ep*RSp45633e + 1.
         /24.*m11*m33*RSp45611e*RD1456ep - 1./24.*pow(m11,2)*RD1456ep*
         RSp45613e + 1./48.*pow(m11,2)*RD1456ep*RSp45633e + 1./48.*pow(
         m11,2)*RSp45611e*RD1456ep );

      M13p6x23R +=  + hh21(3,m13,m33,m11,m13,mu2) * (  - 1./24.*pow(
         m33,2)*RD3456ep*RS14563pe - 1./48.*pow(m33,2)*RS1456pee*
         RD3456ep - 1./48.*pow(m33,2)*RS1456ppe*RD3456ep + 1./48.*pow(
         m33,2)*RS145pe*RD3456ep + 1./48.*pow(m33,2)*RS146pe*RD3456ep
          + 1./48.*pow(m33,2)*RS156pe*RD3456ep - 1./12.*m11*m33*
         RD3456ep*RS14563pe - 1./24.*m11*m33*RS1456pee*RD3456ep - 1./24.
         *m11*m33*RS1456ppe*RD3456ep + 1./24.*m11*m33*RS145pe*RD3456ep
          + 1./24.*m11*m33*RS146pe*RD3456ep + 1./24.*m11*m33*RS156pe*
         RD3456ep - 1./24.*pow(m11,2)*RD3456ep*RS14563pe - 1./48.*pow(
         m11,2)*RS1456pee*RD3456ep - 1./48.*pow(m11,2)*RS1456ppe*
         RD3456ep + 1./48.*pow(m11,2)*RS145pe*RD3456ep + 1./48.*pow(
         m11,2)*RS146pe*RD3456ep + 1./48.*pow(m11,2)*RS156pe*RD3456ep )
         ;

      M13p6x23R +=  + hh21(3,m13,m33,m33,m13,mu2) * (  - 1./24.*pow(
         m33,2)*RD3456ep*RS34561pe - 1./48.*pow(m33,2)*RS3456pee*
         RD3456ep - 1./48.*pow(m33,2)*RS3456ppe*RD3456ep + 1./48.*pow(
         m33,2)*RS345pe*RD3456ep + 1./48.*pow(m33,2)*RS346pe*RD3456ep
          + 1./48.*pow(m33,2)*RS356pe*RD3456ep - 1./12.*m11*m33*
         RD3456ep*RS34561pe - 1./24.*m11*m33*RS3456pee*RD3456ep - 1./24.
         *m11*m33*RS3456ppe*RD3456ep + 1./24.*m11*m33*RS345pe*RD3456ep
          + 1./24.*m11*m33*RS346pe*RD3456ep + 1./24.*m11*m33*RS356pe*
         RD3456ep - 1./24.*pow(m11,2)*RD3456ep*RS34561pe - 1./48.*pow(
         m11,2)*RS3456pee*RD3456ep - 1./48.*pow(m11,2)*RS3456ppe*
         RD3456ep + 1./48.*pow(m11,2)*RS345pe*RD3456ep + 1./48.*pow(
         m11,2)*RS346pe*RD3456ep + 1./48.*pow(m11,2)*RS356pe*RD3456ep )
         ;

      M13p6x23R +=  + hh21(3,m13,m33,me2,m13,mu2) * (  - 1./24.*pow(
         m33,2)*RD3456ep*RSe45613p + 1./48.*pow(m33,2)*RSe45633p*
         RD3456ep + 1./48.*pow(m33,2)*RSe45611p*RD3456ep - 1./12.*m11*
         m33*RD3456ep*RSe45613p + 1./24.*m11*m33*RSe45633p*RD3456ep + 1.
         /24.*m11*m33*RSe45611p*RD3456ep - 1./24.*pow(m11,2)*RD3456ep*
         RSe45613p + 1./48.*pow(m11,2)*RSe45633p*RD3456ep + 1./48.*pow(
         m11,2)*RSe45611p*RD3456ep );

      M13p6x23R +=  + hh21(3,m13,m33,mp2,m13,mu2) * (  - 1./24.*pow(
         m33,2)*RD3456ep*RSp45613e + 1./48.*pow(m33,2)*RSp45633e*
         RD3456ep + 1./48.*pow(m33,2)*RSp45611e*RD3456ep - 1./12.*m11*
         m33*RD3456ep*RSp45613e + 1./24.*m11*m33*RSp45633e*RD3456ep + 1.
         /24.*m11*m33*RSp45611e*RD3456ep - 1./24.*pow(m11,2)*RD3456ep*
         RSp45613e + 1./48.*pow(m11,2)*RSp45633e*RD3456ep + 1./48.*pow(
         m11,2)*RSp45611e*RD3456ep );

      M13p6x23R +=  + hh21(3,m14,m11,m34,m13,mu2) * (  - 1./16.*pow(
         m33,2)*RD1456ep - 1./8.*m11*m33*RD1456ep - 1./16.*pow(m11,2)*
         RD1456ep );

      M13p6x23R +=  + hh21(3,m15,m11,m35,m13,mu2) * (  - 1./16.*pow(
         m33,2)*RD1456ep - 1./8.*m11*m33*RD1456ep - 1./16.*pow(m11,2)*
         RD1456ep );

      M13p6x23R +=  + hh21(3,m16,m11,m36,m13,mu2) * (  - 1./16.*pow(
         m33,2)*RD1456ep - 1./8.*m11*m33*RD1456ep - 1./16.*pow(m11,2)*
         RD1456ep );

      M13p6x23R +=  + hh21(3,m34,m33,m14,m13,mu2) * (  - 1./16.*pow(
         m33,2)*RD3456ep - 1./8.*m11*m33*RD3456ep - 1./16.*pow(m11,2)*
         RD3456ep );

      M13p6x23R +=  + hh21(3,m35,m33,m15,m13,mu2) * (  - 1./16.*pow(
         m33,2)*RD3456ep - 1./8.*m11*m33*RD3456ep - 1./16.*pow(m11,2)*
         RD3456ep );

      M13p6x23R +=  + hh21(3,m36,m33,m16,m13,mu2) * (  - 1./16.*pow(
         m33,2)*RD3456ep - 1./8.*m11*m33*RD3456ep - 1./16.*pow(m11,2)*
         RD3456ep );

      M13p6x23R +=  + hh21(7,m13,m11,m11,m13,mu2) * ( 1./96.*pow(m33,2)
         *pow(RD1456ep,2) + 1./48.*m11*m33*pow(RD1456ep,2) + 1./96.*
         pow(m11,2)*pow(RD1456ep,2) );

      M13p6x23R +=  + hh21(7,m13,m11,m33,m13,mu2) * ( 1./48.*pow(m33,2)
         *RD1456ep*RD3456ep + 1./24.*m11*m33*RD1456ep*RD3456ep + 1./48.
         *pow(m11,2)*RD1456ep*RD3456ep );

      M13p6x23R +=  + hh21(7,m13,m33,m33,m13,mu2) * ( 1./96.*pow(m33,2)
         *pow(RD3456ep,2) + 1./48.*m11*m33*pow(RD3456ep,2) + 1./96.*
         pow(m11,2)*pow(RD3456ep,2) );

  return M13p6x23R/pow(f0,4);
}

//*********************************************************************
// decay constants
// 1 valence mass 1 sea mass *****************************************

double fv1s1nf3p4(const quarkmassnf mass, const Linf Liin){
  return fv1s1nf3p4L(mass,Liin)+fv1s1nf3p4R(mass);
}

double fv1s1nf3p4L(const quarkmassnf mass, const Linf Liin){
  double L0r,L1r,L2r,L3r,L4r,L5r,L6r,L7r,L8r,L9r,L10r,L11r,H1r,H2r,mu;
  int nfLi;
  Liin.out(L0r,L1r,L2r,L3r,L4r,L5r,L6r,L7r,L8r,L9r,L10r,L11r,H1r,H2r,mu,nfLi);
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if (mu0 != mu){
    cout << "ERROR: differing mu's in fv1s1nf3p4L\n";}
  if ( (nq != 2) || (B0mq.size() != 2)){
    cout << "ERROR: fv1s1nf3p4L probably called with wrong number"
	 << "of quark masses\n";}
  // two masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m44 = 2.*B0mq[1];
  double F12p4x11L =
       + 12*m44*L4r + 4*m11*L5r;

  return F12p4x11L/pow(f0,2);
}


double fv1s1nf3p4R(const quarkmassnf mass){
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if ( (nq != 2) || (B0mq.size() != 2)){
    cout << "ERROR: fv2s1nf3p4R probably called with wrong number"
	 << "of quark masses\n";}
  // two masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m44 = 2.*B0mq[1];
  double m14 = (m11+m44)/2.;
  double mu2 = mu0*mu0;
  double F12p4x11R =
       + Ab(m14,mu2) * ( 3./2. );

  return F12p4x11R/pow(f0,2);
}

double fv1s1nf3p6(const quarkmassnf mass, const Linf Liin, const Ki Kiin){
  return fv1s1nf3p6L(mass,Liin)+fv1s1nf3p6K(mass,Kiin)+fv1s1nf3p6R(mass);
}

double fv1s1nf3p6L(const quarkmassnf mass, const Linf Liin){
  double L0r,L1r,L2r,L3r,L4r,L5r,L6r,L7r,L8r,L9r,L10r,L11r,H1r,H2r,mu;
  int nfLi;
  Liin.out(L0r,L1r,L2r,L3r,L4r,L5r,L6r,L7r,L8r,L9r,L10r,L11r,H1r,H2r,mu,nfLi);
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if (mu0 != mu){
    cout << "ERROR: differing mu's in fv1s1nf3p6L\n";}
  if ( (nq != 2) || (B0mq.size() != 2)){
    cout << "ERROR: fv1s1nf3p6L probably called with wrong number"
	 << "of quark masses\n";}
  // two masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m44 = 2.*B0mq[1];
  double m14 = (m11+m44)/2.;
  double mu2 = mu0*mu0;
  double F12p6x11L =
       + pi16 * (  - 3./4.*pow(m44,2)*L3r - 8*pow(m44,2)*L2r - 3./2.*
         pow(m44,2)*L0r - 17./6.*m11*m44*L3r - 13./3.*m11*m44*L0r + 5./
         4.*pow(m11,2)*L3r - pow(m11,2)*L2r - 2*pow(m11,2)*L1r + 1./2.*
         pow(m11,2)*L0r );

      F12p6x11L +=  + Ab(m11,mu2) * (  - 4*m44*L3r - 4*m44*L0r - 4./3.*
         m11*L5r + 8*m11*L3r - 10*m11*L2r - 4*m11*L1r + 8*m11*L0r );

      F12p6x11L +=  + Ab(m14,mu2) * (  - 18*m44*L4r - 30*m14*L3r - 12*
         m14*L0r + 6*m11*L5r );

      F12p6x11L +=  + Ab(m44,mu2) * ( 32*m44*L4r - 16*m44*L2r - 64*m44*
         L1r );

      F12p6x11L +=  + Bb(m11,mu2) * ( 4./3.*m11*m44*L5r - 4*m11*m44*L3r
          - 4*m11*m44*L0r - 4./3.*pow(m11,2)*L5r + 4*pow(m11,2)*L3r + 4
         *pow(m11,2)*L0r );

      F12p6x11L +=  + Bb(m14,mu2) * ( 72*m14*m44*L6r - 36*m14*m44*L4r
          + 24*pow(m14,2)*L8r - 12*pow(m14,2)*L5r );

      F12p6x11L +=  - 72*pow(m44,2)*pow(L4r,2) - 48*m11*m44*L4r*L5r - 8
         *pow(m11,2)*pow(L5r,2);

  return F12p6x11L/pow(f0,4);
}

double fv1s1nf3p6K(const quarkmassnf mass, const Ki Kiin){
  double KK[116],mu;
  Kiin.out(KK,mu);
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if (mu0 != mu){
    cout << "ERROR: differing mu's in fv1s1nf3p6L\n";}
  if ( (nq != 2) || (B0mq.size() != 2)){
    cout << "ERROR: fv1s1nf3p6K probably called with wrong number"
	 << "of quark masses\n";}
  // two masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m44 = 2.*B0mq[1];
  double F12p6x11K =
       + 8*KK[19]*pow(m11,2) + 24*KK[20]*m11*m44 + 24*KK[21]*pow(m44,2)
          + 72*KK[22]*pow(m44,2) + 8*KK[23]*pow(m11,2);

  return F12p6x11K/pow(f0,4);
}

double fv1s1nf3p6R(const quarkmassnf mass){
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if ( (nq != 2) || (B0mq.size() != 2)){
    cout << "ERROR: fv1s1nf3p6R probably called with wrong number"
	 << "of quark masses\n";}
  // two masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m44 = 2.*B0mq[1];
  double m14 = (m11+m44)/2.;
  double mu2 = mu0*mu0;

  double F12p6x11R =
       + pi162 * (  - 9./128.*pow(m44,2) + 9./8.*m14*m44 - 3./8.*m11*
         m44 - 3./64.*pow(m11,2) + 1./12.*pi2*pow(m44,2) + 3./16.*pi2*
         m14*m44 + 1./48.*pi2*m11*m44 - 1./96.*pi2*pow(m11,2) );

      F12p6x11R +=  + pow(Ab(m11,mu2),2) * ( 1./16. );

      F12p6x11R +=  + Ab(m11,mu2)*Bb(m11,mu2) * ( 1./8.*m44 - 1./8.*m11
          );

      F12p6x11R +=  + Ab(m11,mu2)*Bb(m14,mu2) * (  - 1./2.*m14 );

      F12p6x11R +=  + Ab(m14,mu2)*pi16 * (  - 3./2.*m44 - 3./8.*m11 );

      F12p6x11R +=  + pow(Ab(m14,mu2),2) * ( 9./8.*pow(m14,-1)*m44 );

      F12p6x11R +=  + pow(Ab(m44,mu2),2) * ( 1./2. );

      F12p6x11R +=  + hh(1,m11,m14,m14,m11,mu2) * (  - 1./8.*m44 );

      F12p6x11R +=  + hh(1,m14,m14,m44,m11,mu2) * (  - m44 );

      F12p6x11R +=  + hh(2,m11,m14,m14,m11,mu2) * (  - 1./8.*m11*m44 + 
         1./8.*pow(m11,2) );

      F12p6x11R +=  + hhd(1,m11,m11,m11,m11,mu2) * ( 5./18.*pow(m11,2)
          );

      F12p6x11R +=  + hhd(1,m11,m14,m14,m11,mu2) * ( 1./8.*m11*m44 - 1./
         2.*pow(m11,2) );

      F12p6x11R +=  + hhd(1,m14,m14,m44,m11,mu2) * ( m11*m44 );

      F12p6x11R +=  + hhd(2,m11,m11,m11,m11,mu2) * (  - 2./9.*pow(
         m11,2)*m44 + 2./9.*pow(m11,3) );

      F12p6x11R +=  + hhd(2,m11,m14,m14,m11,mu2) * (  - 3./8.*pow(
         m11,2)*m44 + 3./8.*pow(m11,3) );

      F12p6x11R +=  + hhd(5,m11,m11,m11,m11,mu2) * ( 1./9.*pow(m11,2)*
         pow(m44,2) - 2./9.*pow(m11,3)*m44 + 1./9.*pow(m11,4) );

      F12p6x11R +=  + hh1d(3,m14,m11,m14,m11,mu2) * ( 2*pow(m11,2)*m44
          - 2*pow(m11,3) );

      F12p6x11R +=  + hh21d(1,m11,m14,m14,m11,mu2) * ( 3./8.*pow(m11,2)
          );

      F12p6x11R +=  + hh21d(1,m44,m14,m14,m11,mu2) * ( 3*pow(m11,2) );

      F12p6x11R +=  + hh21d(2,m11,m14,m14,m11,mu2) * ( 3./8.*pow(m11,2)
         *m44 - 3./8.*pow(m11,3) );

  return F12p6x11R/pow(f0,4);
}
// 2 valence mass 1 sea mass *****************************************

double fv2s1nf3p4(const quarkmassnf mass, const Linf Liin){
  return fv2s1nf3p4L(mass,Liin)+fv2s1nf3p4R(mass);
}

double fv2s1nf3p4L(const quarkmassnf mass, const Linf Liin){
  double L0r,L1r,L2r,L3r,L4r,L5r,L6r,L7r,L8r,L9r,L10r,L11r,H1r,H2r,mu;
  int nfLi;
  Liin.out(L0r,L1r,L2r,L3r,L4r,L5r,L6r,L7r,L8r,L9r,L10r,L11r,H1r,H2r,mu,nfLi);
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if (mu0 != mu){
    cout << "ERROR: differing mu's in fv1s1nf3p4L\n";}
  if ( (nq != 3) || (B0mq.size() != 3)){
    cout << "ERROR: fv1s1nf3p4L probably called with wrong number"
	 << "of quark masses\n";}
  // three masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m33 = 2.*B0mq[1];
  double m44 = 2.*B0mq[2];
  double F13p4x21L =
       + 12*m44*L4r + 2*m33*L5r + 2*m11*L5r;

  return F13p4x21L/pow(f0,2);
}

double fv2s1nf3p4R(const quarkmassnf mass){
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if ( (nq != 3) || (B0mq.size() != 3)){
    cout << "ERROR: fv1s1nf3p4R probably called with wrong number"
	 << "of quark masses\n";}
  // three masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m33 = 2.*B0mq[1];
  double m44 = 2.*B0mq[2];
  double m14 = (m11+m44)/2.;
  double m34 = (m33+m44)/2.;
  double RN13 = RN(m11,m33);
  double mu2 = mu0*mu0;
  double F13p4x21R =
       + Ab(m11,mu2) * ( 1./12. - 1./6.*m44*RN13 + 1./6.*m33*RN13 );

      F13p4x21R +=  + Ab(m14,mu2) * ( 3./4. );

      F13p4x21R +=  + Ab(m33,mu2) * (  - 1./12. + 1./6.*m44*RN13 - 1./6.
         *m33*RN13 );

      F13p4x21R +=  + Ab(m34,mu2) * ( 3./4. );

      F13p4x21R +=  + Bb(m11,mu2) * ( 1./12.*m44 - 1./12.*m11 );

      F13p4x21R +=  + Bb(m33,mu2) * ( 1./12.*m44 - 1./12.*m33 );

  return F13p4x21R/pow(f0,2);
}

double fv2s1nf3p6(const quarkmassnf mass, const Linf Liin, const Ki Kiin){
  return fv2s1nf3p6L(mass,Liin)+fv2s1nf3p6K(mass,Kiin)+fv2s1nf3p6R(mass);
}

double fv2s1nf3p6L(const quarkmassnf mass, const Linf Liin){
  double L0r,L1r,L2r,L3r,L4r,L5r,L6r,L7r,L8r,L9r,L10r,L11r,H1r,H2r,mu;
  int nfLi;
  Liin.out(L0r,L1r,L2r,L3r,L4r,L5r,L6r,L7r,L8r,L9r,L10r,L11r,H1r,H2r,mu,nfLi);
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if (mu0 != mu){
    cout << "ERROR: differing mu's in fv2s1nf3p6L\n";}
  if ( (nq != 3) || (B0mq.size() != 3)){
    cout << "ERROR: fv2s1nf3p6L probably called with wrong number"
	 << "of quark masses\n";}
  // three masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m33 = 2.*B0mq[1];
  double m44 = 2.*B0mq[2];
  double m13 = (m11+m33)/2.;
  double m14 = (m11+m44)/2.;
  double m34 = (m33+m44)/2.;
  double RN13 = RN(m11,m33);
  double mu2 = mu0*mu0;
  double F13p6x21L =
       + pi16 * (  - 3./4.*pow(m44,2)*L3r - 8*pow(m44,2)*L2r - 3./2.*
         pow(m44,2)*L0r - 17./12.*m33*m44*L3r - 13./6.*m33*m44*L0r + 11.
         /24.*pow(m33,2)*L3r - 1./4.*pow(m33,2)*L2r - 1./2.*pow(m33,2)*
         L1r + 1./4.*pow(m33,2)*L0r - 17./12.*m11*m44*L3r - 13./6.*m11*
         m44*L0r + 1./3.*m11*m33*L3r - 1./2.*m11*m33*L2r - m11*m33*L1r
          + 11./24.*pow(m11,2)*L3r - 1./4.*pow(m11,2)*L2r - 1./2.*pow(
         m11,2)*L1r + 1./4.*pow(m11,2)*L0r );

      F13p6x21L +=  + Ab(m11,mu2) * ( 8./3.*m44*L8r + 8*m44*L7r - 1./3.
         *m44*L5r + m44*L4r - 5./3.*m44*L3r - 2./3.*m44*L0r - 8./3.*
         pow(m44,2)*L8r*RN13 - 8*pow(m44,2)*L7r*RN13 - 8./3.*m33*L8r - 
         8*m33*L7r - 1./6.*m33*L5r + 16./3.*m33*m44*L8r*RN13 + 16*m33*
         m44*L7r*RN13 - 8./3.*pow(m33,2)*L8r*RN13 - 8*pow(m33,2)*L7r*
         RN13 + 1./6.*m11*L5r + 10./3.*m11*L3r + 4./3.*m11*L0r - 2*RS(
         m11,m44,m33)*m44*L4r - 2./3.*RS(m11,m44,m33)*m11*L5r + 2./3.*
         RS(m11,m44,m33)*m11*L3r + 8./3.*RS(m11,m44,m33)*m11*L0r );

      F13p6x21L +=  + Ab(m13,mu2) * (  - 5*m33*L2r - 2*m33*L1r - 5*m11*
         L2r - 2*m11*L1r );

      F13p6x21L +=  + Ab(m14,mu2) * (  - 9*m44*L4r - 15./2.*m44*L3r - 3
         *m44*L0r + 3./2.*m33*L5r + 3./2.*m11*L5r - 15./2.*m11*L3r - 3*
         m11*L0r );

      F13p6x21L +=  + Ab(m33,mu2) * (  - 8./3.*m44*L8r - 8*m44*L7r + 1./
         3.*m44*L5r - m44*L4r - 7./3.*m44*L3r - 10./3.*m44*L0r + 8./3.*
         pow(m44,2)*L8r*RN13 + 8*pow(m44,2)*L7r*RN13 + 8./3.*m33*L8r + 
         8*m33*L7r - 1./2.*m33*L5r + 4*m33*L3r + 4*m33*L0r - 16./3.*m33
         *m44*L8r*RN13 - 16*m33*m44*L7r*RN13 + 8./3.*pow(m33,2)*L8r*
         RN13 + 8*pow(m33,2)*L7r*RN13 - 5./6.*m11*L5r + 2./3.*m11*L3r
          + 8./3.*m11*L0r + 2*RS(m11,m44,m33)*m44*L4r + 2./3.*RS(m11,
         m44,m33)*m11*L5r - 2./3.*RS(m11,m44,m33)*m11*L3r - 8./3.*RS(
         m11,m44,m33)*m11*L0r );

      F13p6x21L +=  + Ab(m34,mu2) * (  - 9*m44*L4r - 15./2.*m44*L3r - 3
         *m44*L0r + 3./2.*m33*L5r - 15./2.*m33*L3r - 3*m33*L0r + 3./2.*
         m11*L5r );

      F13p6x21L +=  + Ab(m44,mu2) * ( 32*m44*L4r - 16*m44*L2r - 64*m44*
         L1r );

      F13p6x21L +=  + Bb(m11,mu2) * ( 4./3.*pow(m44,2)*L8r + 4*pow(
         m44,2)*L7r + 4*pow(m44,2)*L6r - 3*pow(m44,2)*L4r + 1./6.*m33*
         m44*L5r - 8*m11*m44*L7r - 8*m11*m44*L6r - 7./6.*m11*m44*L5r + 
         5*m11*m44*L4r - 5./3.*m11*m44*L3r - 2./3.*m11*m44*L0r - 1./6.*
         m11*m33*L5r - 8./3.*pow(m11,2)*L8r + 4*pow(m11,2)*L7r + 11./6.
         *pow(m11,2)*L5r + 5./3.*pow(m11,2)*L3r + 2./3.*pow(m11,2)*L0r
          + 8*RS(m11,m44,m33)*m11*m44*L6r - 4*RS(m11,m44,m33)*m11*m44*
         L4r + 8./3.*RS(m11,m44,m33)*pow(m11,2)*L8r - 4./3.*RS(m11,m44,
         m33)*pow(m11,2)*L5r );

      F13p6x21L +=  + Bb(m14,mu2) * ( 3*pow(m44,2)*L8r + 18*pow(m44,2)*
         L6r - 3./2.*pow(m44,2)*L5r - 9*pow(m44,2)*L4r + 6*m11*m44*L8r
          + 18*m11*m44*L6r - 3*m11*m44*L5r - 9*m11*m44*L4r + 3*pow(
         m11,2)*L8r - 3./2.*pow(m11,2)*L5r );

      F13p6x21L +=  + Bb(m33,mu2) * ( 4./3.*pow(m44,2)*L8r + 4*pow(
         m44,2)*L7r - 4*pow(m44,2)*L6r + pow(m44,2)*L4r - 8./3.*m33*m44
         *L8r - 8*m33*m44*L7r + 1./6.*m33*m44*L5r + m33*m44*L4r - 5./3.
         *m33*m44*L3r - 2./3.*m33*m44*L0r + 4*pow(m33,2)*L7r + 1./2.*
         pow(m33,2)*L5r + 5./3.*pow(m33,2)*L3r + 2./3.*pow(m33,2)*L0r
          - 8./3.*m11*m44*L8r + 8*m11*m44*L6r + 3./2.*m11*m44*L5r - 4*
         m11*m44*L4r + 8./3.*m11*m33*L8r - 3./2.*m11*m33*L5r + 8./3.*
         pow(m11,2)*L8r - 4./3.*pow(m11,2)*L5r - 8*RS(m11,m44,m33)*m11*
         m44*L6r + 4*RS(m11,m44,m33)*m11*m44*L4r - 8./3.*RS(m11,m44,m33
         )*pow(m11,2)*L8r + 4./3.*RS(m11,m44,m33)*pow(m11,2)*L5r );

      F13p6x21L +=  + Bb(m34,mu2) * ( 3*pow(m44,2)*L8r + 18*pow(m44,2)*
         L6r - 3./2.*pow(m44,2)*L5r - 9*pow(m44,2)*L4r + 6*m33*m44*L8r
          + 18*m33*m44*L6r - 3*m33*m44*L5r - 9*m33*m44*L4r + 3*pow(
         m33,2)*L8r - 3./2.*pow(m33,2)*L5r );

      F13p6x21L +=  + Cb(m11,mu2) * ( 8*m11*pow(m44,2)*L6r - 4*m11*pow(
         m44,2)*L4r + 8./3.*pow(m11,2)*m44*L8r - 8*pow(m11,2)*m44*L6r
          - 4./3.*pow(m11,2)*m44*L5r + 4*pow(m11,2)*m44*L4r - 8./3.*
         pow(m11,3)*L8r + 4./3.*pow(m11,3)*L5r );

      F13p6x21L +=  + Cb(m33,mu2) * ( 8*m33*pow(m44,2)*L6r - 4*m33*pow(
         m44,2)*L4r + 8./3.*pow(m33,2)*m44*L8r - 8*pow(m33,2)*m44*L6r
          - 4./3.*pow(m33,2)*m44*L5r + 4*pow(m33,2)*m44*L4r - 8./3.*
         pow(m33,3)*L8r + 4./3.*pow(m33,3)*L5r );

      F13p6x21L +=  - 72*pow(m44,2)*pow(L4r,2) - 24*m33*m44*L4r*L5r - 2
         *pow(m33,2)*pow(L5r,2) - 24*m11*m44*L4r*L5r - 4*m11*m33*pow(
         L5r,2) - 2*pow(m11,2)*pow(L5r,2);

  return F13p6x21L/pow(f0,4);
}

double fv2s1nf3p6K(const quarkmassnf mass, const Ki Kiin){
  double KK[116],mu;
  Kiin.out(KK,mu);
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if (mu0 != mu){
    cout << "ERROR: differing mu's in fv2s1nf3p6L\n";}
  if ( (nq != 3) || (B0mq.size() != 3)){
    cout << "ERROR: fv2s1nf3p6K probably called with wrong number"
	 << "of quark masses\n";}
  // two masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m33 = 2.*B0mq[1];
  double m44 = 2.*B0mq[2];
  double F13p6x21K =
       + KK[19] * ( 4*pow(m33,2) + 4*pow(m11,2) );

      F13p6x21K +=  + KK[20] * ( 12*m33*m44 + 12*m11*m44 );

      F13p6x21K +=  + KK[21] * ( 24*pow(m44,2) );

      F13p6x21K +=  + KK[22] * ( 72*pow(m44,2) );

      F13p6x21K +=  + KK[23] * ( 8*m11*m33 );

  return F13p6x21K/pow(f0,4);
}

double fv2s1nf3p6R(const quarkmassnf mass){
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if ( (nq != 3) || (B0mq.size() != 3)){
    cout << "ERROR: fv2s1nf3p6R probably called with wrong number"
	 << "of quark masses\n";}
  // three masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m33 = 2.*B0mq[1];
  double m44 = 2.*B0mq[2];
  double m13 = (m11+m33)/2.;
  double m14 = (m11+m44)/2.;
  double m34 = (m33+m44)/2.;
  double RN13 = RN(m11,m33);
  double RD14 = m11-m44;
  double RD34 = m33-m44;
  double mu2 = mu0*mu0;

  double F13p6x21R =
       + pi162 * (  - 1./128.*pow(m44,2) + 9./16.*m34*m44 - 3./16.*m33*
         m44 - 79./1536.*pow(m33,2) + 9./16.*m14*m44 + 1./16.*m13*m33
          - 1./4.*m11*m44 - 53./768.*m11*m33 + 1./16.*m11*m13 - 79./
         1536.*pow(m11,2) + 3./32.*pi2*pow(m44,2) + 3./32.*pi2*m34*m44
          + 1./96.*pi2*m33*m44 - 1./64.*pi2*pow(m33,2) + 3./32.*pi2*m14
         *m44 + 1./96.*pi2*m13*m33 + 1./96.*pi2*m11*m13 - 1./64.*pi2*
         pow(m11,2) - 1./16.*RS(m11,m44,m33)*m33*m44 + 1./16.*RS(m11,
         m44,m33)*m11*m44 - 1./96.*RS(m11,m44,m33)*pi2*m33*m44 + 1./96.
         *RS(m11,m44,m33)*pi2*m11*m44 );

      F13p6x21R +=  + Ab(m11,mu2)*pi16 * ( 1./48.*m44 + 1./32.*m33 + 5./
         96.*m11 - 1./8.*RS(m11,m44,m33)*m44 - 1./12.*RS(m11,m44,m33)*
         m11 );

      F13p6x21R +=  + pow(Ab(m11,mu2),2) * (  - 1./48. - 1./32.*pow(
         m11,-2)*pow(m44,2) - 1./18.*m33*RN13 + 1./24.*RS(m11,m44,m33)
          + 1./16.*RS(m11,m44,m33)*pow(m11,-1)*m44 + 1./18.*RS(m11,m44,
         m33)*m33*RN13 + 1./72.*pow(RS(m11,m44,m33),2) );

      F13p6x21R +=  + Ab(m11,mu2)*Ab(m13,mu2) * (  - 3./8. - 1./2.*m33*
         RN13 );

      F13p6x21R +=  + Ab(m11,mu2)*Ab(m14,mu2) * ( 13./48. + 1./6.*m44*
         RN13 + 1./3.*m33*RN13 - 1./12.*RS(m11,m44,m33) );

      F13p6x21R +=  + Ab(m11,mu2)*Ab(m33,mu2) * ( 5./48. - 1./18.*m44*
         RN13 + 1./9.*m33*RN13 - 1./12.*RS(m11,m44,m33) - 1./9.*RS(m11,
         m44,m33)*m33*RN13 - 1./36.*pow(RS(m11,m44,m33),2) );

      F13p6x21R +=  + Ab(m11,mu2)*Ab(m34,mu2) * (  - 1./48. + 1./3.*m44
         *RN13 + 1./6.*m33*RN13 + 1./12.*RS(m11,m44,m33) );

      F13p6x21R +=  + Ab(m11,mu2)*Ab(m44,mu2) * (  - 4./9.*m44*RN13 );

      F13p6x21R +=  + Ab(m11,mu2)*Bb(m11,mu2) * ( 1./16.*pow(m11,-1)*
         pow(m44,2) + 7./144.*m44 - 1./18.*m33 + 1./18.*m33*m44*RN13 - 
         1./18.*pow(m33,2)*RN13 + 7./36.*m11 - 1./24.*RS(m11,m44,m33)*
         m44 + 1./18.*RS(m11,m44,m33)*m33 - 1./18.*RS(m11,m44,m33)*m33*
         m44*RN13 + 1./18.*RS(m11,m44,m33)*pow(m33,2)*RN13 - 1./72.*RS(
         m11,m44,m33)*m11 );

      F13p6x21R +=  + Ab(m11,mu2)*Bb(m14,mu2) * (  - 1./8.*m44 - 1./8.*
         m11 );

      F13p6x21R +=  + Ab(m11,mu2)*Bb(m33,mu2) * (  - 19./144.*m44 + 1./
         18.*pow(m44,2)*RN13 + 17./144.*m33 - 1./6.*m33*m44*RN13 + 1./9.
         *pow(m33,2)*RN13 + 5./72.*m11 + 5./72.*RS(m11,m44,m33)*m44 - 1.
         /18.*RS(m11,m44,m33)*m33 + 1./18.*RS(m11,m44,m33)*m33*m44*RN13
          - 1./18.*RS(m11,m44,m33)*pow(m33,2)*RN13 - 5./72.*RS(m11,m44,
         m33)*m11 );

      F13p6x21R +=  + Ab(m11,mu2)*Cb(m11,mu2) * (  - 1./18.*m11*m44 + 1.
         /18.*pow(m11,2) );

      F13p6x21R +=  + pow(Ab(m13,mu2),2) * ( 1./8. + 1./16.*pow(m13,-1)
         *m33 + 1./16.*m11*pow(m13,-1) );

      F13p6x21R +=  + Ab(m13,mu2)*Ab(m33,mu2) * ( 1./8. + 1./2.*m33*
         RN13 );

      F13p6x21R +=  + Ab(m14,mu2)*pi16 * (  - 3./4.*m44 - 9./32.*m33 + 
         3./32.*m11 );

      F13p6x21R +=  + pow(Ab(m14,mu2),2) * (  - 9./32. + 9./16.*pow(
         m14,-1)*m44 );

      F13p6x21R +=  + Ab(m14,mu2)*Ab(m33,mu2) * (  - 13./48. - 1./6.*
         m44*RN13 - 1./3.*m33*RN13 + 1./12.*RS(m11,m44,m33) );

      F13p6x21R +=  + Ab(m14,mu2)*Ab(m34,mu2) * ( 9./16. );

      F13p6x21R +=  + Ab(m14,mu2)*Bb(m11,mu2) * (  - 5./16.*m44 - 3./16.
         *m11 );

      F13p6x21R +=  + Ab(m14,mu2)*Bb(m33,mu2) * ( 1./16.*m44 - 1./16.*
         m33 );

      F13p6x21R +=  + Ab(m33,mu2)*pi16 * (  - 1./48.*m44 - 1./32.*m33
          - 5./96.*m11 + 1./8.*RS(m11,m44,m33)*m44 + 1./12.*RS(m11,m44,
         m33)*m11 );

      F13p6x21R +=  + pow(Ab(m33,mu2),2) * (  - 1./48. - 1./32.*pow(
         m33,-2)*pow(m44,2) + 1./16.*pow(m33,-1)*m44 + 1./18.*m44*RN13
          - 1./18.*m33*RN13 + 1./24.*RS(m11,m44,m33) - 1./16.*RS(m11,
         m44,m33)*pow(m33,-1)*m44 + 1./18.*RS(m11,m44,m33)*m33*RN13 + 1.
         /72.*pow(RS(m11,m44,m33),2) );

      F13p6x21R +=  + Ab(m33,mu2)*Ab(m34,mu2) * ( 1./48. - 1./3.*m44*
         RN13 - 1./6.*m33*RN13 - 1./12.*RS(m11,m44,m33) );

      F13p6x21R +=  + Ab(m33,mu2)*Ab(m44,mu2) * ( 4./9.*m44*RN13 );

      F13p6x21R +=  + Ab(m33,mu2)*Bb(m11,mu2) * (  - 7./144.*m44 + 1./
         18.*m33 - 1./18.*m33*m44*RN13 + 1./18.*pow(m33,2)*RN13 + 7./
         144.*m11 + 1./24.*RS(m11,m44,m33)*m44 - 1./18.*RS(m11,m44,m33)
         *m33 + 1./18.*RS(m11,m44,m33)*m33*m44*RN13 - 1./18.*RS(m11,m44
         ,m33)*pow(m33,2)*RN13 - 1./24.*RS(m11,m44,m33)*m11 );

      F13p6x21R +=  + Ab(m33,mu2)*Bb(m33,mu2) * ( 1./16.*pow(m33,-1)*
         pow(m44,2) + 3./16.*m44 - 1./18.*pow(m44,2)*RN13 + 5./72.*m33
          + 1./6.*m33*m44*RN13 - 1./9.*pow(m33,2)*RN13 - 1./8.*m11 - 5./
         72.*RS(m11,m44,m33)*m44 + 1./18.*RS(m11,m44,m33)*m33 - 1./18.*
         RS(m11,m44,m33)*m33*m44*RN13 + 1./18.*RS(m11,m44,m33)*pow(
         m33,2)*RN13 + 1./8.*RS(m11,m44,m33)*m11 );

      F13p6x21R +=  + Ab(m33,mu2)*Bb(m34,mu2) * (  - 1./8.*m44 - 1./8.*
         m33 );

      F13p6x21R +=  + Ab(m33,mu2)*Cb(m33,mu2) * (  - 1./18.*m33*m44 + 1.
         /18.*pow(m33,2) );

      F13p6x21R +=  + Ab(m34,mu2)*pi16 * (  - 3./4.*m44 + 3./32.*m33 - 
         9./32.*m11 );

      F13p6x21R +=  + pow(Ab(m34,mu2),2) * (  - 9./32. + 9./16.*pow(
         m34,-1)*m44 );

      F13p6x21R +=  + Ab(m34,mu2)*Bb(m11,mu2) * ( 1./16.*m44 - 1./16.*
         m11 );

      F13p6x21R +=  + Ab(m34,mu2)*Bb(m33,mu2) * (  - 5./16.*m44 - 3./16.
         *m33 );

      F13p6x21R +=  + pow(Ab(m44,mu2),2) * ( 1./2. );

      F13p6x21R +=  + Ab(m44,mu2)*Bb(m11,mu2) * ( 2./9.*m44 );

      F13p6x21R +=  + Ab(m44,mu2)*Bb(m33,mu2) * ( 2./9.*m44 );

      F13p6x21R +=  + Bb(m11,mu2)*pi16 * (  - 1./16.*pow(m44,2) - 1./32.
         *m33*m44 + 5./96.*m11*m44 + 1./32.*m11*m33 + 1./96.*pow(m11,2)
          );

      F13p6x21R +=  + pow(Bb(m11,mu2),2) * ( 1./288.*pow(m44,2) - 1./16.
         *m11*m44 + 17./288.*pow(m11,2) + 1./18.*RS(m11,m44,m33)*m11*
         m44 - 1./18.*RS(m11,m44,m33)*pow(m11,2) );

      F13p6x21R +=  + Bb(m11,mu2)*Bb(m33,mu2) * ( 1./144.*pow(m44,2) - 
         1./144.*m33*m44 - 1./144.*m11*m44 + 1./144.*m11*m33 );

      F13p6x21R +=  + Bb(m11,mu2)*Cb(m11,mu2) * ( 1./18.*m11*pow(m44,2)
          - 1./9.*pow(m11,2)*m44 + 1./18.*pow(m11,3) );

      F13p6x21R +=  + Bb(m33,mu2)*pi16 * (  - 1./16.*pow(m44,2) + 5./96.
         *m33*m44 + 1./96.*pow(m33,2) - 1./32.*m11*m44 + 1./32.*m11*m33
          );

      F13p6x21R +=  + pow(Bb(m33,mu2),2) * (  - 5./96.*pow(m44,2) + 7./
         144.*m33*m44 + 1./288.*pow(m33,2) + 1./9.*m11*m44 - 1./18.*m11
         *m33 - 1./18.*pow(m11,2) - 1./18.*RS(m11,m44,m33)*m11*m44 + 1./
         18.*RS(m11,m44,m33)*pow(m11,2) );

      F13p6x21R +=  + Bb(m33,mu2)*Cb(m33,mu2) * ( 1./18.*m33*pow(m44,2)
          - 1./9.*pow(m33,2)*m44 + 1./18.*pow(m33,3) );

      F13p6x21R +=  + hh(1,m11,m11,m13,m13,mu2) * ( 1./72.*m44 - 5./144.
         *m33 + 11./144.*m11 - 1./72.*RS(m11,m44,m33)*m44 + 1./24.*RS(
         m11,m44,m33)*m11 - 1./36.*pow(RS(m11,m44,m33),2)*m11 );

      F13p6x21R +=  + hh(1,m11,m13,m33,m13,mu2) * (  - 1./36.*m44 + 1./
         144.*m33 + 5./144.*m11 + 1./36.*RS(m11,m44,m33)*m44 - 1./12.*
         RS(m11,m44,m33)*m11 + 1./18.*pow(RS(m11,m44,m33),2)*m11 );

      F13p6x21R +=  + hh(1,m11,m14,m34,m13,mu2) * ( 1./16.*m33 - 1./16.
         *m11 - 1./8.*RS(m11,m44,m33)*m44 );

      F13p6x21R +=  + hh(1,m13,m13,m13,m13,mu2) * (  - 1./8.*m13 );

      F13p6x21R +=  + hh(1,m13,m33,m33,m13,mu2) * ( 1./72.*m44 + 13./
         144.*m33 - 7./144.*m11 - 1./72.*RS(m11,m44,m33)*m44 + 1./24.*
         RS(m11,m44,m33)*m11 - 1./36.*pow(RS(m11,m44,m33),2)*m11 );

      F13p6x21R +=  + hh(1,m14,m33,m34,m13,mu2) * (  - 1./8.*m44 - 1./
         16.*m33 + 1./16.*m11 + 1./8.*RS(m11,m44,m33)*m44 );

      F13p6x21R +=  + hh(1,m14,m34,m44,m13,mu2) * (  - m44 );

      F13p6x21R +=  + hh(2,m11,m11,m13,m13,mu2) * (  - 1./72.*m13*RD14
          + 1./36.*RS(m11,m44,m33)*m13*RD14 );

      F13p6x21R +=  + hh(2,m11,m13,m33,m13,mu2) * ( 1./72.*m13*RD14 - 1.
         /36.*RS(m11,m44,m33)*m13*RD14 );

      F13p6x21R +=  + hh(2,m11,m14,m34,m13,mu2) * ( 1./8.*m14*RD14 );

      F13p6x21R +=  + hh(2,m33,m11,m13,m13,mu2) * ( 1./72.*m13*RD34 - 1.
         /36.*RS(m33,m44,m11)*m13*RD34 );

      F13p6x21R +=  + hh(2,m33,m13,m33,m13,mu2) * (  - 1./72.*m13*RD34
          + 1./36.*RS(m33,m44,m11)*m13*RD34 );

      F13p6x21R +=  + hh(2,m33,m14,m34,m13,mu2) * ( 1./8.*m34*RD34 );

      F13p6x21R +=  + hh(5,m11,m11,m13,m13,mu2) * (  - 1./144.*m13*pow(
         RD14,2) );

      F13p6x21R +=  + hh(5,m11,m33,m13,m13,mu2) * (  - 1./72.*m13*RD14*
         RD34 );

      F13p6x21R +=  + hh(5,m33,m33,m13,m13,mu2) * (  - 1./144.*m13*pow(
         RD34,2) );

      F13p6x21R +=  + hhd(1,m11,m11,m13,m13,mu2) * ( 1./48.*pow(m44,2)
          - 1./144.*m33*m44 + 25./288.*pow(m33,2) - 1./16.*m11*m44 + 17.
         /144.*m11*m33 + 19./288.*pow(m11,2) + 1./12.*RS(m11,m44,m33)*
         m11*m44 - 1./9.*RS(m11,m44,m33)*pow(m11,2) + 1./12.*pow(RS(m11
         ,m44,m33),2)*pow(m11,2) );

      F13p6x21R +=  + hhd(1,m11,m13,m33,m13,mu2) * (  - 1./24.*pow(
         m44,2) + 23./288.*pow(m33,2) - 1./6.*pow(m13,2) + 1./12.*m11*
         m44 + 23./144.*m11*m33 + 11./288.*pow(m11,2) - 1./6.*RS(m11,
         m44,m33)*m11*m44 + 1./6.*RS(m11,m44,m33)*pow(m11,2) - 1./6.*
         pow(RS(m11,m44,m33),2)*pow(m11,2) - 1./6.*RS(m33,m44,m11)*pow(
         m13,2) );

      F13p6x21R +=  + hhd(1,m11,m14,m34,m13,mu2) * ( 1./16.*pow(m44,2)
          + 1./8.*m33*m44 - 5./32.*pow(m33,2) + 5./16.*m11*m44 - 3./8.*
         m11*m33 - 15./32.*pow(m11,2) + 1./8.*RS(m11,m44,m33)*m11*m44
          + 1./2.*RS(m11,m44,m33)*pow(m11,2) + RS(m33,m44,m11)*pow(
         m13,2) );

      F13p6x21R +=  + hhd(1,m13,m13,m13,m13,mu2) * ( 1./8.*pow(m13,2) )
         ;

      F13p6x21R +=  + hhd(1,m13,m33,m33,m13,mu2) * ( 1./48.*pow(m44,2)
          - 5./144.*m33*m44 + 11./288.*pow(m33,2) - 7./48.*m11*m44 + 25.
         /144.*m11*m33 + 65./288.*pow(m11,2) + 1./12.*RS(m11,m44,m33)*
         m11*m44 - 2./9.*RS(m11,m44,m33)*pow(m11,2) + 1./12.*pow(RS(m11
         ,m44,m33),2)*pow(m11,2) );

      F13p6x21R +=  + hhd(1,m14,m33,m34,m13,mu2) * (  - 1./16.*pow(
         m44,2) - 1./16.*m33*m44 + 1./32.*pow(m33,2) - 1./4.*m11*m44 + 
         1./8.*m11*m33 + 11./32.*pow(m11,2) - 1./8.*RS(m11,m44,m33)*m11
         *m44 - 1./2.*RS(m11,m44,m33)*pow(m11,2) );

      F13p6x21R +=  + hhd(1,m14,m34,m44,m13,mu2) * ( m13*m44 );

      F13p6x21R +=  + hhd(2,m11,m11,m13,m13,mu2) * ( 1./24.*pow(m13,2)*
         RD14 + 1./36.*RS(m33,m44,m11)*pow(m13,2)*RD14 );

      F13p6x21R +=  + hhd(2,m11,m13,m33,m13,mu2) * (  - 1./72.*pow(
         m13,2)*RD14 + 1./36.*RS(m11,m44,m33)*pow(m13,2)*RD14 );

      F13p6x21R +=  + hhd(2,m11,m14,m34,m13,mu2) * (  - 1./8.*m13*m14*
         RD14 );

      F13p6x21R +=  + hhd(2,m33,m11,m13,m13,mu2) * (  - 1./72.*pow(
         m13,2)*RD34 + 1./36.*RS(m33,m44,m11)*pow(m13,2)*RD34 );

      F13p6x21R +=  + hhd(2,m33,m13,m33,m13,mu2) * ( 5./72.*pow(m13,2)*
         RD34 - 1./36.*RS(m33,m44,m11)*pow(m13,2)*RD34 );

      F13p6x21R +=  + hhd(2,m33,m14,m34,m13,mu2) * (  - 1./8.*m13*m34*
         RD34 );

      F13p6x21R +=  + hhd(5,m11,m11,m13,m13,mu2) * ( 5./144.*pow(m13,2)
         *pow(RD14,2) );

      F13p6x21R +=  + hhd(5,m11,m33,m13,m13,mu2) * ( 1./72.*pow(m13,2)*
         RD14*RD34 );

      F13p6x21R +=  + hhd(5,m33,m33,m13,m13,mu2) * ( 5./144.*pow(m13,2)
         *pow(RD34,2) );

      F13p6x21R +=  + hh1d(1,m11,m14,m34,m13,mu2) * (  - RS(m33,m44,m11
         )*pow(m13,2) );

      F13p6x21R +=  + hh1d(1,m13,m11,m11,m13,mu2) * (  - 5./9.*pow(
         m13,2) + 1./9.*RS(m11,m44,m33)*RS(m33,m44,m11)*pow(m13,2) );

      F13p6x21R +=  + hh1d(1,m13,m11,m33,m13,mu2) * ( 1./9.*pow(m13,2)
          - 2./9.*RS(m11,m44,m33)*RS(m33,m44,m11)*pow(m13,2) );

      F13p6x21R +=  + hh1d(1,m13,m33,m33,m13,mu2) * (  - 5./9.*pow(
         m13,2) + 1./9.*RS(m11,m44,m33)*RS(m33,m44,m11)*pow(m13,2) );

      F13p6x21R +=  + hh1d(1,m14,m11,m34,m13,mu2) * (  - RS(m11,m44,m33
         )*pow(m13,2) - RS(m33,m44,m11)*pow(m13,2) );

      F13p6x21R +=  + hh1d(1,m14,m33,m34,m13,mu2) * ( RS(m11,m44,m33)*
         pow(m13,2) );

      F13p6x21R +=  + hh1d(1,m34,m14,m33,m13,mu2) * (  - RS(m33,m44,m11
         )*pow(m13,2) );

      F13p6x21R +=  + hh1d(3,m13,m11,m11,m13,mu2) * (  - 1./9.*RS(m33,
         m44,m11)*pow(m13,2)*RD14 );

      F13p6x21R +=  + hh1d(3,m13,m11,m33,m13,mu2) * ( 1./9.*RS(m33,m44,
         m11)*pow(m13,2)*RD14 );

      F13p6x21R +=  + hh1d(3,m13,m33,m11,m13,mu2) * (  - 1./9.*RS(m33,
         m44,m11)*pow(m13,2)*RD14 );

      F13p6x21R +=  + hh1d(3,m13,m33,m33,m13,mu2) * (  - 1./9.*RS(m11,
         m44,m33)*pow(m13,2)*RD34 );

      F13p6x21R +=  + hh1d(7,m13,m11,m11,m13,mu2) * (  - 1./18.*pow(
         m13,2)*pow(RD14,2) );

      F13p6x21R +=  + hh1d(7,m13,m33,m33,m13,mu2) * (  - 1./18.*pow(
         m13,2)*pow(RD34,2) );

      F13p6x21R +=  + hh21d(1,m11,m11,m13,m13,mu2) * (  - 3./8.*pow(
         m13,2) );

      F13p6x21R +=  + hh21d(1,m11,m14,m34,m13,mu2) * ( 3./8.*RS(m33,m44
         ,m11)*pow(m13,2) );

      F13p6x21R +=  + hh21d(1,m13,m11,m11,m13,mu2) * ( 5./24.*pow(
         m13,2) - 1./12.*RS(m11,m44,m33)*RS(m33,m44,m11)*pow(m13,2) );

      F13p6x21R +=  + hh21d(1,m13,m11,m33,m13,mu2) * (  - 1./24.*pow(
         m13,2) + 1./6.*RS(m11,m44,m33)*RS(m33,m44,m11)*pow(m13,2) );

      F13p6x21R +=  + hh21d(1,m13,m13,m13,m13,mu2) * ( 3./8.*pow(m13,2)
          );

      F13p6x21R +=  + hh21d(1,m13,m33,m33,m13,mu2) * ( 5./24.*pow(
         m13,2) - 1./12.*RS(m11,m44,m33)*RS(m33,m44,m11)*pow(m13,2) );

      F13p6x21R +=  + hh21d(1,m14,m11,m34,m13,mu2) * ( 3./8.*RS(m11,m44
         ,m33)*pow(m13,2) );

      F13p6x21R +=  + hh21d(1,m14,m33,m34,m13,mu2) * (  - 3./8.*RS(m11,
         m44,m33)*pow(m13,2) );

      F13p6x21R +=  + hh21d(1,m33,m13,m33,m13,mu2) * (  - 3./8.*pow(
         m13,2) );

      F13p6x21R +=  + hh21d(1,m33,m14,m34,m13,mu2) * ( 3./8.*RS(m11,m44
         ,m33)*pow(m13,2) );

      F13p6x21R +=  + hh21d(1,m34,m11,m14,m13,mu2) * (  - 3./8.*RS(m33,
         m44,m11)*pow(m13,2) );

      F13p6x21R +=  + hh21d(1,m34,m14,m33,m13,mu2) * ( 3./8.*RS(m33,m44
         ,m11)*pow(m13,2) );

      F13p6x21R +=  + hh21d(1,m44,m14,m34,m13,mu2) * ( 3*pow(m13,2) );

      F13p6x21R +=  + hh21d(3,m13,m11,m11,m13,mu2) * ( 1./12.*RS(m11,
         m44,m33)*pow(m13,2)*m44 - 1./12.*RS(m11,m44,m33)*pow(m13,3) );

      F13p6x21R +=  + hh21d(3,m13,m11,m33,m13,mu2) * (  - 1./12.*RS(m11
         ,m44,m33)*pow(m13,2)*m44 + 1./12.*RS(m11,m44,m33)*pow(m13,3) )
         ;

      F13p6x21R +=  + hh21d(3,m13,m33,m11,m13,mu2) * (  - 1./12.*RS(m33
         ,m44,m11)*pow(m13,2)*m44 + 1./12.*RS(m33,m44,m11)*pow(m13,3) )
         ;

      F13p6x21R +=  + hh21d(3,m13,m33,m33,m13,mu2) * ( 1./12.*RS(m33,
         m44,m11)*pow(m13,2)*m44 - 1./12.*RS(m33,m44,m11)*pow(m13,3) );

      F13p6x21R +=  + hh21d(3,m14,m11,m34,m13,mu2) * (  - 3./8.*pow(
         m13,2)*RD14 );

      F13p6x21R +=  + hh21d(3,m34,m33,m14,m13,mu2) * (  - 3./8.*pow(
         m13,2)*RD34 );

      F13p6x21R +=  + hh21d(7,m13,m11,m11,m13,mu2) * ( 1./48.*pow(
         m13,2)*pow(RD14,2) );

      F13p6x21R +=  + hh21d(7,m13,m11,m33,m13,mu2) * ( 1./24.*pow(
         m13,2)*RD14*RD34 );

      F13p6x21R +=  + hh21d(7,m13,m33,m33,m13,mu2) * ( 1./48.*pow(
         m13,2)*pow(RD34,2) );

  return F13p6x21R/pow(f0,4);
}

// 1 valence mass 2 sea mass *****************************************

double fv1s2nf3p4(const quarkmassnf mass, const Linf Liin){
  return fv1s2nf3p4L(mass,Liin)+fv1s2nf3p4R(mass);
}

double fv1s2nf3p4L(const quarkmassnf mass, const Linf Liin){
  double L0r,L1r,L2r,L3r,L4r,L5r,L6r,L7r,L8r,L9r,L10r,L11r,H1r,H2r,mu;
  int nfLi;
  Liin.out(L0r,L1r,L2r,L3r,L4r,L5r,L6r,L7r,L8r,L9r,L10r,L11r,H1r,H2r,mu,nfLi);
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if (mu0 != mu){
    cout << "ERROR: differing mu's in fv1s2nf3p4L\n";}
  if ( (nq != 3) || (B0mq.size() != 3)){
    cout << "ERROR: fv1s2nf3p4L probably called with wrong number"
	 << "of quark masses\n";}
  // three masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m44 = 2.*B0mq[1];
  double m66 = 2.*B0mq[2];
  double F12p4x12L =
       + 4*m66*L4r + 8*m44*L4r + 4*m11*L5r;

  return F12p4x12L/pow(f0,2);
}

double fv1s2nf3p4R(const quarkmassnf mass){
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if ( (nq != 3) || (B0mq.size() != 3)){
    cout << "ERROR: fv1s2nf3p4R probably called with wrong number"
	 << "of quark masses\n";}
  // three masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m44 = 2.*B0mq[1];
  double m66 = 2.*B0mq[2];
  double m14 = (m11+m44)/2.;
  double m16 = (m11+m66)/2.;
  double mu2 = mu0*mu0;

  double F12p4x12R =
       + Ab(m14,mu2) * ( 1 );

      F12p4x12R +=  + Ab(m16,mu2) * ( 1./2. );

  return F12p4x12R/pow(f0,2);
}

double fv1s2nf3p6(const quarkmassnf mass, const Linf Liin, const Ki Kiin){
  return fv1s2nf3p6L(mass,Liin)+fv1s2nf3p6K(mass,Kiin)+fv1s2nf3p6R(mass);
}

double fv1s2nf3p6L(const quarkmassnf mass, const Linf Liin){
  double L0r,L1r,L2r,L3r,L4r,L5r,L6r,L7r,L8r,L9r,L10r,L11r,H1r,H2r,mu;
  int nfLi;
  Liin.out(L0r,L1r,L2r,L3r,L4r,L5r,L6r,L7r,L8r,L9r,L10r,L11r,H1r,H2r,mu,nfLi);
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if (mu0 != mu){
    cout << "ERROR: differing mu's in fv1s2nf3p6L\n";}
  if ( (nq != 3) || (B0mq.size() != 3)){
    cout << "ERROR: fv1s2nf3p6L probably called with wrong number"
	 << "of quark masses\n";}
  // three masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m44 = 2.*B0mq[1];
  double m66 = 2.*B0mq[2];
  double m14 = (m11+m44)/2.;
  double m16 = (m11+m66)/2.;
  double m46 = (m44+m66)/2.;
  double me2 = (m44+2.*m66)/3.;
  double mu2 = mu0*mu0;
  double RC146e = RC(m11,m44,m66,me2);
  double RSe61 = RS(me2,m66,m11);
  double RSe6611 = pow(RSe61,2);
  double RD146e = RD(m11,m44,m66,me2);

  double F12p6x12L =
       + pi16 * (  - 43./108.*pow(m66,2)*L3r - 13./9.*pow(m66,2)*L2r - 
         35./54.*pow(m66,2)*L0r + 8./27.*m44*m66*L3r - 22./9.*m44*m66*
         L2r + 8./27.*m44*m66*L0r - 35./54.*pow(m44,2)*L3r - 37./9.*
         pow(m44,2)*L2r - 31./27.*pow(m44,2)*L0r - 17./18.*m11*m66*L3r
          - 13./9.*m11*m66*L0r - 17./9.*m11*m44*L3r - 26./9.*m11*m44*
         L0r + 5./4.*pow(m11,2)*L3r - pow(m11,2)*L2r - 2*pow(m11,2)*L1r
          + 1./2.*pow(m11,2)*L0r );

      F12p6x12L +=  + Ab(m11,mu2) * ( 4*L3r*RD146e + 4*L0r*RD146e - 4./
         3.*m11*L5r*RC146e + 4*m11*L3r*RC146e - 10*m11*L2r - 4*m11*L1r
          + 4*m11*L0r*RC146e );

      F12p6x12L +=  + Ab(m14,mu2) * (  - 4*m66*L4r - 8*m44*L4r - 20*m14
         *L3r - 8*m14*L0r + 4*m11*L5r );

      F12p6x12L +=  + Ab(m16,mu2) * (  - 2*m66*L4r - 4*m44*L4r - 10*m16
         *L3r - 4*m16*L0r + 2*m11*L5r );

      F12p6x12L +=  + Ab(m44,mu2) * ( 12*m44*L4r - 6*m44*L2r - 24*m44*
         L1r );

      F12p6x12L +=  + Ab(m46,mu2) * ( 16*m46*L4r - 8*m46*L2r - 32*m46*
         L1r );

      F12p6x12L +=  + Ab(me2,mu2) * (  - 8*me2*L3r*RSe6611 - 2*me2*L2r
          - 8*me2*L1r - 8*me2*L0r*RSe6611 + 8./3.*m66*L4r + 4./3.*m44*
         L4r + 8./3.*m11*L5r*RSe6611 );

      F12p6x12L +=  + Bb(m11,mu2) * (  - 4./3.*m11*L5r*RD146e + 4*m11*
         L3r*RD146e + 4*m11*L0r*RD146e );

      F12p6x12L +=  + Bb(m14,mu2) * ( 16*m14*m66*L6r - 8*m14*m66*L4r + 
         32*m14*m44*L6r - 16*m14*m44*L4r + 16*pow(m14,2)*L8r - 8*pow(
         m14,2)*L5r );

      F12p6x12L +=  + Bb(m16,mu2) * ( 8*m16*m66*L6r - 4*m16*m66*L4r + 
         16*m16*m44*L6r - 8*m16*m44*L4r + 8*pow(m16,2)*L8r - 4*pow(
         m16,2)*L5r );

      F12p6x12L +=  - 8*pow(m66,2)*pow(L4r,2) - 32*m44*m66*pow(L4r,2)
          - 32*pow(m44,2)*pow(L4r,2) - 16*m11*m66*L4r*L5r - 32*m11*m44*
         L4r*L5r - 8*pow(m11,2)*pow(L5r,2);

  return F12p6x12L/pow(f0,4);
}

double fv1s2nf3p6K(const quarkmassnf mass, const Ki Kiin){
  double KK[116],mu;
  Kiin.out(KK,mu);
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if (mu0 != mu){
    cout << "ERROR: differing mu's in fv1s2nf3p6K\n";}
  if ( (nq != 3) || (B0mq.size() != 3)){
    cout << "ERROR: fv1s2nf3p6K probably called with wrong number"
	 << "of quark masses\n";}
  // three masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m44 = 2.*B0mq[1];
  double m66 = 2.*B0mq[2];
  double F12p6x12K =
       + KK[19] * ( 8*pow(m11,2) );

      F12p6x12K +=  + KK[20] * ( 8*m11*m66 + 16*m11*m44 );

      F12p6x12K +=  + KK[21] * ( 8*pow(m66,2) + 16*pow(m44,2) );

      F12p6x12K +=  + KK[22] * ( 8*pow(m66,2) + 32*m44*m66 + 32*pow(
         m44,2) );

      F12p6x12K +=  + KK[23] * ( 8*pow(m11,2) );

  return F12p6x12K/pow(f0,4);
}

double fv1s2nf3p6R(const quarkmassnf mass){
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if ( (nq != 3) || (B0mq.size() != 3)){
    cout << "ERROR: fv1s2nf3p6R probably called with wrong number"
	 << "of quark masses\n";}
  // three masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m44 = 2.*B0mq[1];
  double m66 = 2.*B0mq[2];
  double m14 = (m11+m44)/2.;
  double m16 = (m11+m66)/2.;
  double m46 = (m44+m66)/2.;
  double me2 = (m44+2.*m66)/3.;
  double mu2 = mu0*mu0;
  double RC146e = RC(m11,m44,m66,me2);
  double RSe61 = RS(me2,m66,m11);
  double RSe6611 = pow(RSe61,2);
  double RS14e = RS(m11,m44,me2);
  double RS16e = RS(m11,m66,me2);
  double RS144ee = RS14e*RS14e;
  double RS166ee = RS16e*RS16e;
  double RD146e = RD(m11,m44,m66,me2);

  double F12p6x12R =
       + pi162 * ( 1./16.*pow(me2,2)*RC146e - 43./384.*pow(m66,2) + 1./
         4.*pow(m46,2) - 5./32.*m44*m66 - 11./96.*pow(m44,2) + 1./4.*
         m16*m66 + 1./8.*m16*m44 + 1./8.*m14*m66 + 5./8.*m14*m44 - 1./8.
         *m11*RD146e - 1./6.*m11*m66 - 1./3.*m11*m44 + 9./64.*pow(
         m11,2) - 1./16.*pow(m11,2)*RC146e + 1./96.*pi2*pow(me2,2)*
         RC146e + 1./24.*pi2*pow(m46,2) + 1./32.*pi2*pow(m44,2) + 1./24.
         *pi2*m16*m66 + 1./48.*pi2*m16*m44 + 1./48.*pi2*m14*m66 + 5./48.
         *pi2*m14*m44 - 1./48.*pi2*m11*RD146e + 1./48.*pi2*pow(m11,2)
          - 1./96.*pi2*pow(m11,2)*RC146e );

      F12p6x12R +=  + pow(Ab(m11,mu2),2) * ( 1./8. - 1./16.*RC146e );

      F12p6x12R +=  + Ab(m11,mu2)*Bb(m11,mu2) * (  - 1./8.*RD146e );

      F12p6x12R +=  + Ab(m11,mu2)*Bb(m14,mu2) * (  - 1./3.*m14*RS16e );

      F12p6x12R +=  + Ab(m11,mu2)*Bb(m16,mu2) * (  - 1./6.*m16*RS14e );

      F12p6x12R +=  + Ab(m14,mu2)*pi16 * (  - 1./4.*m66 - 3./4.*m44 - 1.
         /4.*m11 );

      F12p6x12R +=  + pow(Ab(m14,mu2),2) * ( 1./8.*pow(m14,-1)*m66 + 5./
         8.*pow(m14,-1)*m44 );

      F12p6x12R +=  + Ab(m16,mu2)*pi16 * (  - 1./4.*m66 - 1./4.*m44 - 1.
         /8.*m11 );

      F12p6x12R +=  + pow(Ab(m16,mu2),2) * ( 1./4.*pow(m16,-1)*m66 + 1./
         8.*pow(m16,-1)*m44 );

      F12p6x12R +=  + pow(Ab(m44,mu2),2) * ( 3./16. );

      F12p6x12R +=  + pow(Ab(m46,mu2),2) * ( 1./4. );

      F12p6x12R +=  + pow(Ab(me2,mu2),2) * ( 1./16.*RC146e );

      F12p6x12R +=  + Ab(me2,mu2)*Bb(m14,mu2) * (  - 1./3.*m14*RSe61 );

      F12p6x12R +=  + Ab(me2,mu2)*Bb(m16,mu2) * ( 1./3.*m16*RSe61 );

      F12p6x12R +=  + hh(1,m11,m14,m14,m11,mu2) * ( 1./12.*RD146e - 1./
         6.*m11*RS16e + 1./12.*m11*RC146e );

      F12p6x12R +=  + hh(1,m11,m16,m16,m11,mu2) * ( 1./24.*RD146e - 1./
         12.*m11*RS14e + 1./24.*m11*RC146e );

      F12p6x12R +=  + hh(1,m14,m14,m44,m11,mu2) * (  - 3./8.*m44 );

      F12p6x12R +=  + hh(1,m14,m14,me2,m11,mu2) * (  - 1./24.*me2*
         RS144ee );

      F12p6x12R +=  + hh(1,m14,m16,m46,m11,mu2) * (  - 1./4.*m66 - 1./4.
         *m44 );

      F12p6x12R +=  + hh(1,m16,m16,me2,m11,mu2) * (  - 1./12.*me2*
         RS166ee );

      F12p6x12R +=  + hh(2,m11,m14,m14,m11,mu2) * ( 1./12.*m11*RD146e )
         ;

      F12p6x12R +=  + hh(2,m11,m16,m16,m11,mu2) * ( 1./24.*m11*RD146e )
         ;

      F12p6x12R +=  + hhd(1,m11,m11,m11,m11,mu2) * ( 1./6.*pow(m11,2)
          + 1./9.*pow(m11,2)*pow(RC146e,2) );

      F12p6x12R +=  + hhd(1,m11,m11,me2,m11,mu2) * (  - 4./9.*pow(
         m11,2)*RC146e*RSe6611 );

      F12p6x12R +=  + hhd(1,m11,m14,m14,m11,mu2) * (  - 1./12.*m11*
         RD146e - 1./2.*pow(m11,2)*RS16e - 2./3.*pow(m11,2)*RS14e*RSe61
          + 1./4.*pow(m11,2)*RC146e );

      F12p6x12R +=  + hhd(1,m11,m16,m16,m11,mu2) * (  - 1./24.*m11*
         RD146e + 2./3.*pow(m11,2)*RS16e*RSe61 - 1./4.*pow(m11,2)*RS14e
          + 1./8.*pow(m11,2)*RC146e );

      F12p6x12R +=  + hhd(1,m11,me2,me2,m11,mu2) * ( 4./9.*pow(m11,2)*
         pow(RSe6611,2) );

      F12p6x12R +=  + hhd(1,m14,m14,m44,m11,mu2) * ( 3./8.*m11*m44 );

      F12p6x12R +=  + hhd(1,m14,m14,me2,m11,mu2) * ( 1./24.*m11*me2*
         RS144ee + 2./3.*pow(m11,2)*RSe6611 - 2./3.*pow(m11,2)*RS14e*
         RSe61 );

      F12p6x12R +=  + hhd(1,m14,m16,m46,m11,mu2) * ( 1./4.*m11*m66 + 1./
         4.*m11*m44 );

      F12p6x12R +=  + hhd(1,m16,m16,me2,m11,mu2) * ( 1./12.*m11*me2*
         RS166ee + 1./3.*pow(m11,2)*RSe6611 + 2./3.*pow(m11,2)*RS16e*
         RSe61 );

      F12p6x12R +=  + hhd(2,m11,m11,m11,m11,mu2) * ( 2./9.*pow(m11,2)*
         RD146e*RC146e );

      F12p6x12R +=  + hhd(2,m11,m11,me2,m11,mu2) * (  - 4./9.*pow(
         m11,2)*RD146e*RSe6611 );

      F12p6x12R +=  + hhd(2,m11,m14,m14,m11,mu2) * ( 1./4.*pow(m11,2)*
         RD146e );

      F12p6x12R +=  + hhd(2,m11,m16,m16,m11,mu2) * ( 1./8.*pow(m11,2)*
         RD146e );

      F12p6x12R +=  + hhd(5,m11,m11,m11,m11,mu2) * ( 1./9.*pow(m11,2)*
         pow(RD146e,2) );

      F12p6x12R +=  + hh1d(1,m11,m14,m14,m11,mu2) * ( 2./3.*pow(m11,2)*
         RS14e*RSe61 );

      F12p6x12R +=  + hh1d(1,m11,m16,m16,m11,mu2) * (  - 2./3.*pow(
         m11,2)*RS16e*RSe61 );

      F12p6x12R +=  + hh1d(1,m14,m14,me2,m11,mu2) * ( 4./3.*pow(m11,2)*
         RS14e*RSe61 );

      F12p6x12R +=  + hh1d(1,m16,m16,me2,m11,mu2) * (  - 4./3.*pow(
         m11,2)*RS16e*RSe61 );

      F12p6x12R +=  + hh1d(3,m14,m11,m14,m11,mu2) * (  - 4./3.*pow(
         m11,2)*RD146e );

      F12p6x12R +=  + hh1d(3,m16,m11,m16,m11,mu2) * (  - 2./3.*pow(
         m11,2)*RD146e );

      F12p6x12R +=  + hh21d(1,m11,m14,m14,m11,mu2) * ( 1./2.*pow(m11,2)
         *RS16e - 1./4.*pow(m11,2)*RC146e );

      F12p6x12R +=  + hh21d(1,m11,m16,m16,m11,mu2) * ( 1./4.*pow(m11,2)
         *RS14e - 1./8.*pow(m11,2)*RC146e );

      F12p6x12R +=  + hh21d(1,m44,m14,m14,m11,mu2) * ( 9./8.*pow(m11,2)
          );

      F12p6x12R +=  + hh21d(1,m46,m14,m16,m11,mu2) * ( 3./2.*pow(m11,2)
          );

      F12p6x12R +=  + hh21d(1,me2,m14,m14,m11,mu2) * ( 1./8.*pow(m11,2)
         *RS144ee );

      F12p6x12R +=  + hh21d(1,me2,m16,m16,m11,mu2) * ( 1./4.*pow(m11,2)
         *RS166ee );

      F12p6x12R +=  + hh21d(2,m11,m14,m14,m11,mu2) * (  - 1./4.*pow(
         m11,2)*RD146e );

      F12p6x12R +=  + hh21d(2,m11,m16,m16,m11,mu2) * (  - 1./8.*pow(
         m11,2)*RD146e );

  return F12p6x12R/pow(f0,4);
}

// 2 valence mass 2 sea mass *****************************************

double fv2s2nf3p4(const quarkmassnf mass, const Linf Liin){
  return fv2s2nf3p4L(mass,Liin)+fv2s2nf3p4R(mass);
}

double fv2s2nf3p4L(const quarkmassnf mass, const Linf Liin){
  double L0r,L1r,L2r,L3r,L4r,L5r,L6r,L7r,L8r,L9r,L10r,L11r,H1r,H2r,mu;
  int nfLi;
  Liin.out(L0r,L1r,L2r,L3r,L4r,L5r,L6r,L7r,L8r,L9r,L10r,L11r,H1r,H2r,mu,nfLi);
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if (mu0 != mu){
    cout << "ERROR: differing mu's in fv2s2nf3p4L\n";}
  if ( (nq != 4) || (B0mq.size() != 4)){
    cout << "ERROR: fv2s2nf3p4L probably called with wrong number"
	 << "of quark masses\n";}
  // four masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m33 = 2.*B0mq[1];
  double m44 = 2.*B0mq[2];
  double m66 = 2.*B0mq[3];
  double F13p4x22L =
       + 4*m66*L4r + 8*m44*L4r + 2*m33*L5r + 2*m11*L5r;

  return F13p4x22L/pow(f0,2);
}


double fv2s2nf3p4R(const quarkmassnf mass){
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if ( (nq != 4) || (B0mq.size() != 4)){
    cout << "ERROR: fv2s1nf3p4R probably called with wrong number"
	 << "of quark masses\n";}
  // four masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m33 = 2.*B0mq[1];
  double m44 = 2.*B0mq[2];
  double m66 = 2.*B0mq[3];
  double me2 = (m44+2.*m66)/3.;

  double m14 = (m11+m44)/2;
  double m16 = (m11+m66)/2;
  double m34 = (m33+m44)/2;
  double m36 = (m33+m66)/2;

  double RC146e = RC(m11,m44,m66,me2);
  double RC346e = RC(m33,m44,m66,me2);
  double RD146e = RD(m11,m44,m66,me2);
  double RD346e = RD(m33,m44,m66,me2);
  double RS1463e = RS(m11,m44,m66,m33,me2);
  double RS3461e = RS(m33,m44,m66,m11,me2);
  double RSe6611 = RS(me2,m66,m66,m11,m11);
  double RSe6613 = RS(me2,m66,m66,m11,m33);
  double RSe6633 = RS(me2,m66,m66,m33,m33);
  double mu2 = mu0*mu0;
  double F13p4x22R =
       + Ab(m11,mu2) * ( 1./6.*RS1463e - 1./12.*RC146e );

      F13p4x22R +=  + Ab(m14,mu2) * ( 1./2. );

      F13p4x22R +=  + Ab(m16,mu2) * ( 1./4. );

      F13p4x22R +=  + Ab(m33,mu2) * ( 1./6.*RS3461e - 1./12.*RC346e );

      F13p4x22R +=  + Ab(m34,mu2) * ( 1./2. );

      F13p4x22R +=  + Ab(m36,mu2) * ( 1./4. );

      F13p4x22R +=  + Ab(me2,mu2) * (  - 1./3.*RSe6613 + 1./6.*RSe6633
          + 1./6.*RSe6611 );

      F13p4x22R +=  + Bb(m11,mu2) * (  - 1./12.*RD146e );

      F13p4x22R +=  + Bb(m33,mu2) * (  - 1./12.*RD346e );

  return F13p4x22R/pow(f0,2);
}

double fv2s2nf3p6(const quarkmassnf mass, const Linf Liin, const Ki Kiin){
  return fv2s2nf3p6L(mass,Liin)+fv2s2nf3p6K(mass,Kiin)+fv2s2nf3p6R(mass);
}

double fv2s2nf3p6L(const quarkmassnf mass, const Linf Liin){
  double L0r,L1r,L2r,L3r,L4r,L5r,L6r,L7r,L8r,L9r,L10r,L11r,H1r,H2r,mu;
  int nfLi;
  Liin.out(L0r,L1r,L2r,L3r,L4r,L5r,L6r,L7r,L8r,L9r,L10r,L11r,H1r,H2r,mu,nfLi);
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if (mu0 != mu){
    cout << "ERROR: differing mu's in fv2s2nf3p6L\n";}
  if ( (nq != 4) || (B0mq.size() != 4)){
    cout << "ERROR: fv2s2nf3p6L probably called with wrong number"
	 << "of quark masses\n";}
  // four masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m33 = 2.*B0mq[1];
  double m44 = 2.*B0mq[2];
  double m66 = 2.*B0mq[3];
  double m13 = (m11+m33)/2;
  double m14 = (m11+m44)/2;
  double m16 = (m11+m66)/2;
  double m34 = (m33+m44)/2;
  double m36 = (m33+m66)/2;
  double m46 = (m44+m66)/2;
  double me2 = (m44+2.*m66)/3.;

  double RN13 = RN(m11,m33);
  double RN1e = RN(m11,me2);
  double RN3e = RN(m33,me2);

  double RD13 = m11-m33;
  double RD46 = m44-m66;

  double RD146e = RD(m11,m44,m66,me2);
  double RD346e = RD(m33,m44,m66,me2);

  double RC146e = RC(m11,m44,m66,me2);
  double RC346e = RC(m33,m44,m66,me2);

  double RS14e = RS(m11,m44,me2);
  double RS16e = RS(m11,m66,me2);
  double RS34e = RS(m33,m44,me2);
  double RS36e = RS(m33,m66,me2);
  double RSe61 = RS(me2,m66,m11);
  double RSe63 = RS(me2,m66,m33);

  double RS1463e = RS(m11,m44,m66,m33,me2);
  double RS3461e = RS(m33,m44,m66,m11,me2);
  double RSe6611 = RS(me2,m66,m66,m11,m11);
  double RSe6613 = RS(me2,m66,m66,m11,m33);
  double RSe6633 = RS(me2,m66,m66,m33,m33);

  double mu2 = mu0*mu0;

  double F13p6x22L =
       + pi16 * (  - 2./3.*pow(me2,2)*L3r*RSe6613 - 1./3.*pow(me2,2)*
         L3r*RSe6633 - 1./3.*pow(me2,2)*L3r*RSe6611 - pow(me2,2)*L2r - 
         2./3.*pow(me2,2)*L0r*RSe6633 - 2./3.*pow(me2,2)*L0r*RSe6611 - 
         1./4.*pow(m66,2)*L3r - pow(m66,2)*L2r - 1./2.*pow(m66,2)*L0r
          - 2*m44*m66*L2r - 1./2.*pow(m44,2)*L3r - 4*pow(m44,2)*L2r - 
         pow(m44,2)*L0r + 1./3.*m33*L3r*RD346e + 2./3.*m33*L0r*RD346e
          - 1./4.*m33*m66*L3r - 1./2.*m33*m66*L0r - 1./2.*m33*m44*L3r
          - m33*m44*L0r - 3./8.*pow(m33,2)*L3r + 1./3.*pow(m33,2)*L3r*
         RS3461e + 1./6.*pow(m33,2)*L3r*RC346e - 1./4.*pow(m33,2)*L2r
          - 1./2.*pow(m33,2)*L1r - 3./4.*pow(m33,2)*L0r + 1./3.*pow(
         m33,2)*L0r*RC346e + 1./3.*m11*L3r*RD146e + 2./3.*m11*L0r*
         RD146e - 1./4.*m11*m66*L3r - 1./2.*m11*m66*L0r - 1./2.*m11*m44
         *L3r - m11*m44*L0r - 1./2.*m11*m33*L2r - m11*m33*L1r - 3./8.*
         pow(m11,2)*L3r + 1./3.*pow(m11,2)*L3r*RS1463e + 1./6.*pow(
         m11,2)*L3r*RC146e - 1./4.*pow(m11,2)*L2r - 1./2.*pow(m11,2)*
         L1r );

      F13p6x22L +=  + pi16 * (  - 3./4.*pow(m11,2)*L0r + 1./3.*pow(
         m11,2)*L0r*RC146e );

      F13p6x22L +=  + Ab(m11,mu2) * ( 5./3.*L3r*RD146e + 2./3.*L0r*
         RD146e - 8./9.*m66*L5r*RS14e*RSe63 + 8./9.*m66*L5r*RS14e*RSe61
          - 2./9.*m66*L5r*pow(RS14e,2) - 2./3.*m66*L4r*RS1463e + 8./3.*
         m66*L4r*RSe6613 - 8./3.*m66*L4r*RSe6611 + 8./9.*m66*L4r*RS16e*
         RSe63 - 8./9.*m66*L4r*RS16e*RSe61 - 4./9.*m66*L4r*pow(RS16e,2)
          - 8./9.*m66*L4r*RS14e*RSe63 + 8./9.*m66*L4r*RS14e*RSe61 - 2./
         9.*m66*L4r*pow(RS14e,2) + m66*L4r*RC146e - 8./9.*m66*me2*L5r*
         RN1e*RS14e*RSe63 + 8./9.*m66*me2*L5r*RN1e*RS14e*RSe61 + 8./3.*
         m66*me2*L4r*RN1e*RSe6613 - 8./3.*m66*me2*L4r*RN1e*RSe6611 + 8./
         9.*m66*me2*L4r*RN1e*RS16e*RSe63 - 8./9.*m66*me2*L4r*RN1e*RS16e
         *RSe61 - 8./9.*m66*me2*L4r*RN1e*RS14e*RSe63 + 8./9.*m66*me2*
         L4r*RN1e*RS14e*RSe61 + 16./9.*pow(m66,2)*L8r*RN1e*RS14e*RSe63
          - 16./9.*pow(m66,2)*L8r*RN1e*RS14e*RSe61 - 8./9.*pow(m66,2)*
         L8r*RN13*RS14e*RS34e + 16./9.*pow(m66,2)*L7r*RN1e*RS14e*RSe63
          - 16./9.*pow(m66,2)*L7r*RN1e*RS14e*RSe61 - 8./9.*pow(m66,2)*
         L7r*RN13*RS14e*RS34e );

      F13p6x22L +=  + Ab(m11,mu2) * ( 16./9.*pow(m66,2)*L6r*RN1e*RS14e*
         RSe63 - 16./9.*pow(m66,2)*L6r*RN1e*RS14e*RSe61 - 8./9.*pow(
         m66,2)*L6r*RN13*RS14e*RS34e + 8./9.*m44*L5r*RS16e*RSe63 - 8./9.
         *m44*L5r*RS16e*RSe61 - 4./9.*m44*L5r*pow(RS16e,2) - 4./3.*m44*
         L4r*RS1463e + 16./3.*m44*L4r*RSe6613 - 16./3.*m44*L4r*RSe6611
          + 16./9.*m44*L4r*RS16e*RSe63 - 16./9.*m44*L4r*RS16e*RSe61 - 8.
         /9.*m44*L4r*pow(RS16e,2) - 16./9.*m44*L4r*RS14e*RSe63 + 16./9.
         *m44*L4r*RS14e*RSe61 - 4./9.*m44*L4r*pow(RS14e,2) + 2*m44*L4r*
         RC146e + 8./9.*m44*me2*L5r*RN1e*RS16e*RSe63 - 8./9.*m44*me2*
         L5r*RN1e*RS16e*RSe61 + 16./3.*m44*me2*L4r*RN1e*RSe6613 - 16./3.
         *m44*me2*L4r*RN1e*RSe6611 + 16./9.*m44*me2*L4r*RN1e*RS16e*
         RSe63 - 16./9.*m44*me2*L4r*RN1e*RS16e*RSe61 - 16./9.*m44*me2*
         L4r*RN1e*RS14e*RSe63 + 16./9.*m44*me2*L4r*RN1e*RS14e*RSe61 + 
         32./9.*m44*m66*L7r*RN1e*RS16e*RSe63 - 32./9.*m44*m66*L7r*RN1e*
         RS16e*RSe61 - 16./9.*m44*m66*L7r*RN1e*RS14e*RSe63 + 16./9.*m44
         *m66*L7r*RN1e*RS14e*RSe61 );

      F13p6x22L +=  + Ab(m11,mu2) * (  - 16./9.*m44*m66*L7r*RN13*RS16e*
         RS34e - 16./9.*m44*m66*L7r*RN13*RS14e*RS36e - 16./9.*m44*m66*
         L6r*RN1e*RS16e*RSe63 + 16./9.*m44*m66*L6r*RN1e*RS16e*RSe61 + 
         32./9.*m44*m66*L6r*RN1e*RS14e*RSe63 - 32./9.*m44*m66*L6r*RN1e*
         RS14e*RSe61 - 16./9.*m44*m66*L6r*RN13*RS16e*RS36e - 16./9.*m44
         *m66*L6r*RN13*RS14e*RS34e - 16./9.*pow(m44,2)*L8r*RN1e*RS16e*
         RSe63 + 16./9.*pow(m44,2)*L8r*RN1e*RS16e*RSe61 - 16./9.*pow(
         m44,2)*L8r*RN13*RS16e*RS36e - 32./9.*pow(m44,2)*L7r*RN1e*RS16e
         *RSe63 + 32./9.*pow(m44,2)*L7r*RN1e*RS16e*RSe61 - 32./9.*pow(
         m44,2)*L7r*RN13*RS16e*RS36e - 32./9.*pow(m44,2)*L6r*RN1e*RS16e
         *RSe63 + 32./9.*pow(m44,2)*L6r*RN1e*RS16e*RSe61 - 32./9.*pow(
         m44,2)*L6r*RN13*RS16e*RS36e + 1./3.*m33*L5r*RS1463e - 1./6.*
         m33*L5r*RC146e + 8./3.*m33*m66*L7r*RN13*RS14e + 8./3.*m33*m66*
         L6r*RN13*RS1463e + 16./3.*m33*m44*L7r*RN13*RS16e + 16./3.*m33*
         m44*L6r*RN13*RS1463e + 8./3.*pow(m33,2)*L8r*RN13*RS1463e - m11
         *L5r*RS1463e );

      F13p6x22L +=  + Ab(m11,mu2) * ( 8./3.*m11*L5r*RSe6613 - 8./3.*m11
         *L5r*RSe6611 + 1./2.*m11*L5r*RC146e + 2./3.*m11*L3r*RS1463e + 
         5./3.*m11*L3r*RC146e + 8./3.*m11*L0r*RS1463e + 2./3.*m11*L0r*
         RC146e + 8./3.*m11*me2*L5r*RN1e*RSe6613 - 8./3.*m11*me2*L5r*
         RN1e*RSe6611 - 16./3.*m11*m66*L7r*RN1e*RSe63 + 16./3.*m11*m66*
         L7r*RN1e*RSe61 + 8./3.*m11*m66*L7r*RN13*RS34e - 16./3.*m11*m66
         *L6r*RN1e*RSe6613 + 16./3.*m11*m66*L6r*RN1e*RSe6611 + 8./3.*
         m11*m66*L6r*RN13*RS3461e + 4./9.*m11*m66*L5r*RN13*RS14e*RS34e
          - 4./3.*m11*m66*L4r*RN13*RS3461e - 4./3.*m11*m66*L4r*RN13*
         RS1463e + 8./9.*m11*m66*L4r*RN13*RS16e*RS36e + 4./9.*m11*m66*
         L4r*RN13*RS14e*RS34e + 16./3.*m11*m44*L7r*RN1e*RSe63 - 16./3.*
         m11*m44*L7r*RN1e*RSe61 + 16./3.*m11*m44*L7r*RN13*RS36e - 32./3.
         *m11*m44*L6r*RN1e*RSe6613 + 32./3.*m11*m44*L6r*RN1e*RSe6611 + 
         16./3.*m11*m44*L6r*RN13*RS3461e + 8./9.*m11*m44*L5r*RN13*RS16e
         *RS36e - 8./3.*m11*m44*L4r*RN13*RS3461e - 8./3.*m11*m44*L4r*
         RN13*RS1463e );

      F13p6x22L +=  + Ab(m11,mu2) * ( 16./9.*m11*m44*L4r*RN13*RS16e*
         RS36e + 8./9.*m11*m44*L4r*RN13*RS14e*RS34e - 8*m11*m33*L7r*
         RN13 - 4./3.*m11*m33*L5r*RN13*RS1463e - 16./3.*pow(m11,2)*L8r*
         RN1e*RSe6613 + 16./3.*pow(m11,2)*L8r*RN1e*RSe6611 + 8./3.*pow(
         m11,2)*L8r*RN13*RS3461e - 4./3.*pow(m11,2)*L5r*RN13*RS3461e );

      F13p6x22L +=  + Ab(m13,mu2) * (  - 5*m33*L2r - 2*m33*L1r - 5*m11*
         L2r - 2*m11*L1r );

      F13p6x22L +=  + Ab(m14,mu2) * (  - 2*m66*L4r - 4*m44*L4r - 5*m44*
         L3r - 2*m44*L0r + m33*L5r + m11*L5r - 5*m11*L3r - 2*m11*L0r );

      F13p6x22L +=  + Ab(m16,mu2) * (  - m66*L4r - 5./2.*m66*L3r - m66*
         L0r - 2*m44*L4r + 1./2.*m33*L5r + 1./2.*m11*L5r - 5./2.*m11*
         L3r - m11*L0r );

      F13p6x22L +=  + Ab(m33,mu2) * ( 5./3.*L3r*RD346e + 2./3.*L0r*
         RD346e + 8./9.*m66*L5r*RS34e*RSe63 - 2./9.*m66*L5r*pow(
         RS34e,2) - 8./9.*m66*L5r*RSe61*RS34e + 4./9.*m66*L5r*RS14e*
         RS34e - 2*m66*L4r*RS3461e - 4./3.*m66*L4r*RS1463e + 8./3.*m66*
         L4r*RSe6613 - 8./3.*m66*L4r*RSe6633 - 8./9.*m66*L4r*RS36e*
         RSe63 - 4./9.*m66*L4r*pow(RS36e,2) + 8./9.*m66*L4r*RS34e*RSe63
          - 2./9.*m66*L4r*pow(RS34e,2) + m66*L4r*RC346e + 8./9.*m66*L4r
         *RSe61*RS36e - 8./9.*m66*L4r*RSe61*RS34e + 8./9.*m66*L4r*RS16e
         *RS36e + 4./9.*m66*L4r*RS14e*RS34e + 8./9.*m66*me2*L5r*RN3e*
         RS34e*RSe63 - 8./9.*m66*me2*L5r*RN3e*RSe61*RS34e + 8./3.*m66*
         me2*L4r*RN3e*RSe6613 - 8./3.*m66*me2*L4r*RN3e*RSe6633 - 8./9.*
         m66*me2*L4r*RN3e*RS36e*RSe63 + 8./9.*m66*me2*L4r*RN3e*RS34e*
         RSe63 + 8./9.*m66*me2*L4r*RN3e*RSe61*RS36e - 8./9.*m66*me2*L4r
         *RN3e*RSe61*RS34e - 16./9.*pow(m66,2)*L8r*RN3e*RS34e*RSe63 + 
         16./9.*pow(m66,2)*L8r*RN3e*RSe61*RS34e + 8./9.*pow(m66,2)*L8r*
         RN13*RS14e*RS34e );

      F13p6x22L +=  + Ab(m33,mu2) * (  - 16./9.*pow(m66,2)*L7r*RN3e*
         RS34e*RSe63 + 16./9.*pow(m66,2)*L7r*RN3e*RSe61*RS34e + 8./9.*
         pow(m66,2)*L7r*RN13*RS14e*RS34e - 16./9.*pow(m66,2)*L6r*RN3e*
         RS34e*RSe63 + 16./9.*pow(m66,2)*L6r*RN3e*RSe61*RS34e + 8./9.*
         pow(m66,2)*L6r*RN13*RS14e*RS34e - 8./9.*m44*L5r*RS36e*RSe63 - 
         4./9.*m44*L5r*pow(RS36e,2) + 8./9.*m44*L5r*RSe61*RS36e + 8./9.
         *m44*L5r*RS16e*RS36e - 4*m44*L4r*RS3461e - 8./3.*m44*L4r*
         RS1463e + 16./3.*m44*L4r*RSe6613 - 16./3.*m44*L4r*RSe6633 - 16.
         /9.*m44*L4r*RS36e*RSe63 - 8./9.*m44*L4r*pow(RS36e,2) + 16./9.*
         m44*L4r*RS34e*RSe63 - 4./9.*m44*L4r*pow(RS34e,2) + 2*m44*L4r*
         RC346e + 16./9.*m44*L4r*RSe61*RS36e - 16./9.*m44*L4r*RSe61*
         RS34e + 16./9.*m44*L4r*RS16e*RS36e + 8./9.*m44*L4r*RS14e*RS34e
          - 8./9.*m44*me2*L5r*RN3e*RS36e*RSe63 + 8./9.*m44*me2*L5r*RN3e
         *RSe61*RS36e + 16./3.*m44*me2*L4r*RN3e*RSe6613 - 16./3.*m44*
         me2*L4r*RN3e*RSe6633 - 16./9.*m44*me2*L4r*RN3e*RS36e*RSe63 + 
         16./9.*m44*me2*L4r*RN3e*RS34e*RSe63 );

      F13p6x22L +=  + Ab(m33,mu2) * ( 16./9.*m44*me2*L4r*RN3e*RSe61*
         RS36e - 16./9.*m44*me2*L4r*RN3e*RSe61*RS34e - 32./9.*m44*m66*
         L7r*RN3e*RS36e*RSe63 + 16./9.*m44*m66*L7r*RN3e*RS34e*RSe63 + 
         32./9.*m44*m66*L7r*RN3e*RSe61*RS36e - 16./9.*m44*m66*L7r*RN3e*
         RSe61*RS34e + 16./9.*m44*m66*L7r*RN13*RS16e*RS34e + 16./9.*m44
         *m66*L7r*RN13*RS14e*RS36e + 16./9.*m44*m66*L6r*RN3e*RS36e*
         RSe63 - 32./9.*m44*m66*L6r*RN3e*RS34e*RSe63 - 16./9.*m44*m66*
         L6r*RN3e*RSe61*RS36e + 32./9.*m44*m66*L6r*RN3e*RSe61*RS34e + 
         16./9.*m44*m66*L6r*RN13*RS16e*RS36e + 16./9.*m44*m66*L6r*RN13*
         RS14e*RS34e + 16./9.*pow(m44,2)*L8r*RN3e*RS36e*RSe63 - 16./9.*
         pow(m44,2)*L8r*RN3e*RSe61*RS36e + 16./9.*pow(m44,2)*L8r*RN13*
         RS16e*RS36e + 32./9.*pow(m44,2)*L7r*RN3e*RS36e*RSe63 - 32./9.*
         pow(m44,2)*L7r*RN3e*RSe61*RS36e + 32./9.*pow(m44,2)*L7r*RN13*
         RS16e*RS36e + 32./9.*pow(m44,2)*L6r*RN3e*RS36e*RSe63 - 32./9.*
         pow(m44,2)*L6r*RN3e*RSe61*RS36e + 32./9.*pow(m44,2)*L6r*RN13*
         RS16e*RS36e );

      F13p6x22L +=  + Ab(m33,mu2) * (  - m33*L5r*RS3461e - 4./3.*m33*
         L5r*RS1463e + 8./3.*m33*L5r*RSe6613 - 8./3.*m33*L5r*RSe6633 + 
         1./2.*m33*L5r*RC346e + 2./3.*m33*L3r*RS3461e + 5./3.*m33*L3r*
         RC346e + 8./3.*m33*L0r*RS3461e + 2./3.*m33*L0r*RC346e + 8./3.*
         m33*me2*L5r*RN3e*RSe6613 - 8./3.*m33*me2*L5r*RN3e*RSe6633 + 16.
         /3.*m33*m66*L7r*RN3e*RSe63 - 16./3.*m33*m66*L7r*RN3e*RSe61 - 8.
         /3.*m33*m66*L7r*RN13*RS14e - 16./3.*m33*m66*L6r*RN3e*RSe6613
          + 16./3.*m33*m66*L6r*RN3e*RSe6633 - 8./3.*m33*m66*L6r*RN13*
         RS1463e - 16./3.*m33*m44*L7r*RN3e*RSe63 + 16./3.*m33*m44*L7r*
         RN3e*RSe61 - 16./3.*m33*m44*L7r*RN13*RS16e - 32./3.*m33*m44*
         L6r*RN3e*RSe6613 + 32./3.*m33*m44*L6r*RN3e*RSe6633 - 16./3.*
         m33*m44*L6r*RN13*RS1463e - 16./3.*pow(m33,2)*L8r*RN3e*RSe6613
          + 16./3.*pow(m33,2)*L8r*RN3e*RSe6633 - 8./3.*pow(m33,2)*L8r*
         RN13*RS1463e - m11*L5r*RS3461e - 1./6.*m11*L5r*RC346e - 8./3.*
         m11*m66*L7r*RN13*RS34e - 8./3.*m11*m66*L6r*RN13*RS3461e - 4./9.
         *m11*m66*L5r*RN13*RS14e*RS34e );

      F13p6x22L +=  + Ab(m33,mu2) * ( 4./3.*m11*m66*L4r*RN13*RS3461e + 
         4./3.*m11*m66*L4r*RN13*RS1463e - 8./9.*m11*m66*L4r*RN13*RS16e*
         RS36e - 4./9.*m11*m66*L4r*RN13*RS14e*RS34e - 16./3.*m11*m44*
         L7r*RN13*RS36e - 16./3.*m11*m44*L6r*RN13*RS3461e - 8./9.*m11*
         m44*L5r*RN13*RS16e*RS36e + 8./3.*m11*m44*L4r*RN13*RS3461e + 8./
         3.*m11*m44*L4r*RN13*RS1463e - 16./9.*m11*m44*L4r*RN13*RS16e*
         RS36e - 8./9.*m11*m44*L4r*RN13*RS14e*RS34e + 8*m11*m33*L7r*
         RN13 + 4./3.*m11*m33*L5r*RN13*RS1463e - 8./3.*pow(m11,2)*L8r*
         RN13*RS3461e + 4./3.*pow(m11,2)*L5r*RN13*RS3461e );

      F13p6x22L +=  + Ab(m34,mu2) * (  - 2*m66*L4r - 4*m44*L4r - 5*m44*
         L3r - 2*m44*L0r + m33*L5r - 5*m33*L3r - 2*m33*L0r + m11*L5r );

      F13p6x22L +=  + Ab(m36,mu2) * (  - m66*L4r - 5./2.*m66*L3r - m66*
         L0r - 2*m44*L4r + 1./2.*m33*L5r - 5./2.*m33*L3r - m33*L0r + 1./
         2.*m11*L5r );

      F13p6x22L +=  + Ab(m44,mu2) * ( 12*m44*L4r - 6*m44*L2r - 24*m44*
         L1r );

      F13p6x22L +=  + Ab(m46,mu2) * ( 8*m66*L4r - 4*m66*L2r - 16*m66*
         L1r + 8*m44*L4r - 4*m44*L2r - 16*m44*L1r );

      F13p6x22L +=  + Ab(me2,mu2) * ( 4*me2*L4r - 4./3.*me2*L3r*RSe6613
          - 10./3.*me2*L3r*RSe6633 - 10./3.*me2*L3r*RSe6611 - 2*me2*L2r
          - 8*me2*L1r - 16./3.*me2*L0r*RSe6613 - 4./3.*me2*L0r*RSe6633
          - 4./3.*me2*L0r*RSe6611 - 2./27.*m66*L4r*pow(RN1e,2)*pow(
         RN3e,2)*pow(RD13,2)*pow(RD46,2) - 8./9.*m66*me2*L5r*RN3e*RS34e
         *RSe63 + 8./9.*m66*me2*L5r*RN3e*RSe61*RS34e + 8./9.*m66*me2*
         L5r*RN1e*RS14e*RSe63 - 8./9.*m66*me2*L5r*RN1e*RS14e*RSe61 - 8./
         3.*m66*me2*L4r*RN3e*RSe6613 + 8./3.*m66*me2*L4r*RN3e*RSe6633
          + 8./9.*m66*me2*L4r*RN3e*RS36e*RSe63 - 8./9.*m66*me2*L4r*RN3e
         *RS34e*RSe63 - 8./9.*m66*me2*L4r*RN3e*RSe61*RS36e + 8./9.*m66*
         me2*L4r*RN3e*RSe61*RS34e - 8./3.*m66*me2*L4r*RN1e*RSe6613 + 8./
         3.*m66*me2*L4r*RN1e*RSe6611 - 8./9.*m66*me2*L4r*RN1e*RS16e*
         RSe63 + 8./9.*m66*me2*L4r*RN1e*RS16e*RSe61 + 8./9.*m66*me2*L4r
         *RN1e*RS14e*RSe63 - 8./9.*m66*me2*L4r*RN1e*RS14e*RSe61 + 16./9.
         *pow(m66,2)*L8r*RN3e*RS34e*RSe63 - 16./9.*pow(m66,2)*L8r*RN3e*
         RSe61*RS34e );

      F13p6x22L +=  + Ab(me2,mu2) * (  - 16./9.*pow(m66,2)*L8r*RN1e*
         RS14e*RSe63 + 16./9.*pow(m66,2)*L8r*RN1e*RS14e*RSe61 + 16./9.*
         pow(m66,2)*L7r*RN3e*RS34e*RSe63 - 16./9.*pow(m66,2)*L7r*RN3e*
         RSe61*RS34e - 16./9.*pow(m66,2)*L7r*RN1e*RS14e*RSe63 + 16./9.*
         pow(m66,2)*L7r*RN1e*RS14e*RSe61 + 16./9.*pow(m66,2)*L6r*RN3e*
         RS34e*RSe63 - 16./9.*pow(m66,2)*L6r*RN3e*RSe61*RS34e - 16./9.*
         pow(m66,2)*L6r*RN1e*RS14e*RSe63 + 16./9.*pow(m66,2)*L6r*RN1e*
         RS14e*RSe61 - 4./27.*m44*L4r*pow(RN1e,2)*pow(RN3e,2)*pow(
         RD13,2)*pow(RD46,2) + 8./9.*m44*me2*L5r*RN3e*RS36e*RSe63 - 8./
         9.*m44*me2*L5r*RN3e*RSe61*RS36e - 8./9.*m44*me2*L5r*RN1e*RS16e
         *RSe63 + 8./9.*m44*me2*L5r*RN1e*RS16e*RSe61 - 16./3.*m44*me2*
         L4r*RN3e*RSe6613 + 16./3.*m44*me2*L4r*RN3e*RSe6633 + 16./9.*
         m44*me2*L4r*RN3e*RS36e*RSe63 - 16./9.*m44*me2*L4r*RN3e*RS34e*
         RSe63 - 16./9.*m44*me2*L4r*RN3e*RSe61*RS36e + 16./9.*m44*me2*
         L4r*RN3e*RSe61*RS34e - 16./3.*m44*me2*L4r*RN1e*RSe6613 + 16./3.
         *m44*me2*L4r*RN1e*RSe6611 );

      F13p6x22L +=  + Ab(me2,mu2) * (  - 16./9.*m44*me2*L4r*RN1e*RS16e*
         RSe63 + 16./9.*m44*me2*L4r*RN1e*RS16e*RSe61 + 16./9.*m44*me2*
         L4r*RN1e*RS14e*RSe63 - 16./9.*m44*me2*L4r*RN1e*RS14e*RSe61 + 
         32./9.*m44*m66*L7r*RN3e*RS36e*RSe63 - 16./9.*m44*m66*L7r*RN3e*
         RS34e*RSe63 - 32./9.*m44*m66*L7r*RN3e*RSe61*RS36e + 16./9.*m44
         *m66*L7r*RN3e*RSe61*RS34e - 32./9.*m44*m66*L7r*RN1e*RS16e*
         RSe63 + 32./9.*m44*m66*L7r*RN1e*RS16e*RSe61 + 16./9.*m44*m66*
         L7r*RN1e*RS14e*RSe63 - 16./9.*m44*m66*L7r*RN1e*RS14e*RSe61 - 
         16./9.*m44*m66*L6r*RN3e*RS36e*RSe63 + 32./9.*m44*m66*L6r*RN3e*
         RS34e*RSe63 + 16./9.*m44*m66*L6r*RN3e*RSe61*RS36e - 32./9.*m44
         *m66*L6r*RN3e*RSe61*RS34e + 16./9.*m44*m66*L6r*RN1e*RS16e*
         RSe63 - 16./9.*m44*m66*L6r*RN1e*RS16e*RSe61 - 32./9.*m44*m66*
         L6r*RN1e*RS14e*RSe63 + 32./9.*m44*m66*L6r*RN1e*RS14e*RSe61 - 
         16./9.*pow(m44,2)*L8r*RN3e*RS36e*RSe63 + 16./9.*pow(m44,2)*L8r
         *RN3e*RSe61*RS36e + 16./9.*pow(m44,2)*L8r*RN1e*RS16e*RSe63 - 
         16./9.*pow(m44,2)*L8r*RN1e*RS16e*RSe61 );

      F13p6x22L +=  + Ab(me2,mu2) * (  - 32./9.*pow(m44,2)*L7r*RN3e*
         RS36e*RSe63 + 32./9.*pow(m44,2)*L7r*RN3e*RSe61*RS36e + 32./9.*
         pow(m44,2)*L7r*RN1e*RS16e*RSe63 - 32./9.*pow(m44,2)*L7r*RN1e*
         RS16e*RSe61 - 32./9.*pow(m44,2)*L6r*RN3e*RS36e*RSe63 + 32./9.*
         pow(m44,2)*L6r*RN3e*RSe61*RS36e + 32./9.*pow(m44,2)*L6r*RN1e*
         RS16e*RSe63 - 32./9.*pow(m44,2)*L6r*RN1e*RS16e*RSe61 - 8./3.*
         m33*me2*L5r*RN3e*RSe6613 + 8./3.*m33*me2*L5r*RN3e*RSe6633 - 16.
         /3.*m33*m66*L7r*RN3e*RSe63 + 16./3.*m33*m66*L7r*RN3e*RSe61 + 
         16./3.*m33*m66*L6r*RN3e*RSe6613 - 16./3.*m33*m66*L6r*RN3e*
         RSe6633 + 16./3.*m33*m44*L7r*RN3e*RSe63 - 16./3.*m33*m44*L7r*
         RN3e*RSe61 + 32./3.*m33*m44*L6r*RN3e*RSe6613 - 32./3.*m33*m44*
         L6r*RN3e*RSe6633 + 16./3.*pow(m33,2)*L8r*RN3e*RSe6613 - 16./3.
         *pow(m33,2)*L8r*RN3e*RSe6633 + 8./27.*m13*pow(me2,2)*L5r*pow(
         RN1e,2)*pow(RN3e,2)*pow(RD46,2) - 16./27.*pow(m13,2)*me2*L5r*
         pow(RN1e,2)*pow(RN3e,2)*pow(RD46,2) + 8./27.*pow(m13,3)*L5r*
         pow(RN1e,2)*pow(RN3e,2)*pow(RD46,2) );

      F13p6x22L +=  + Ab(me2,mu2) * (  - 8./3.*m11*me2*L5r*RN1e*RSe6613
          + 8./3.*m11*me2*L5r*RN1e*RSe6611 + 16./3.*m11*m66*L7r*RN1e*
         RSe63 - 16./3.*m11*m66*L7r*RN1e*RSe61 + 16./3.*m11*m66*L6r*
         RN1e*RSe6613 - 16./3.*m11*m66*L6r*RN1e*RSe6611 - 16./3.*m11*
         m44*L7r*RN1e*RSe63 + 16./3.*m11*m44*L7r*RN1e*RSe61 + 32./3.*
         m11*m44*L6r*RN1e*RSe6613 - 32./3.*m11*m44*L6r*RN1e*RSe6611 + 
         16./3.*pow(m11,2)*L8r*RN1e*RSe6613 - 16./3.*pow(m11,2)*L8r*
         RN1e*RSe6611 );

      F13p6x22L +=  + Bb(m11,mu2) * ( m66*L4r*RD146e + 4./9.*pow(m66,2)
         *L8r*pow(RS14e,2) + 4./9.*pow(m66,2)*L7r*pow(RS14e,2) + 4./9.*
         pow(m66,2)*L6r*pow(RS14e,2) + 2*m44*L4r*RD146e + 16./9.*m44*
         m66*L7r*RS14e*RS16e + 8./9.*m44*m66*L6r*pow(RS16e,2) + 8./9.*
         m44*m66*L6r*pow(RS14e,2) + 8./9.*pow(m44,2)*L8r*pow(RS16e,2)
          + 16./9.*pow(m44,2)*L7r*pow(RS16e,2) + 16./9.*pow(m44,2)*L6r*
         pow(RS16e,2) - 1./6.*m33*L5r*RD146e + 1./2.*m11*L5r*RD146e + 5.
         /3.*m11*L3r*RD146e + 2./3.*m11*L0r*RD146e - 8./3.*m11*m66*L7r*
         RS14e + 8./3.*m11*m66*L6r*RS1463e - 8./3.*m11*m66*L6r*RC146e
          - 2./9.*m11*m66*L5r*pow(RS14e,2) - 4./3.*m11*m66*L4r*RS1463e
          - 4./9.*m11*m66*L4r*pow(RS16e,2) - 2./9.*m11*m66*L4r*pow(
         RS14e,2) + 4./3.*m11*m66*L4r*RC146e - 16./3.*m11*m44*L7r*RS16e
          + 16./3.*m11*m44*L6r*RS1463e - 16./3.*m11*m44*L6r*RC146e - 4./
         9.*m11*m44*L5r*pow(RS16e,2) - 8./3.*m11*m44*L4r*RS1463e - 8./9.
         *m11*m44*L4r*pow(RS16e,2) - 4./9.*m11*m44*L4r*pow(RS14e,2) + 8.
         /3.*m11*m44*L4r*RC146e );

      F13p6x22L +=  + Bb(m11,mu2) * ( 8./3.*pow(m11,2)*L8r*RS1463e - 8./
         3.*pow(m11,2)*L8r*RC146e + 4*pow(m11,2)*L7r - 4./3.*pow(m11,2)
         *L5r*RS1463e + 4./3.*pow(m11,2)*L5r*RC146e );

      F13p6x22L +=  + Bb(m14,mu2) * ( 4*m44*m66*L6r - 2*m44*m66*L4r + 2
         *pow(m44,2)*L8r + 8*pow(m44,2)*L6r - pow(m44,2)*L5r - 4*pow(
         m44,2)*L4r + 4*m11*m66*L6r - 2*m11*m66*L4r + 4*m11*m44*L8r + 8
         *m11*m44*L6r - 2*m11*m44*L5r - 4*m11*m44*L4r + 2*pow(m11,2)*
         L8r - pow(m11,2)*L5r );

      F13p6x22L +=  + Bb(m16,mu2) * ( pow(m66,2)*L8r + 2*pow(m66,2)*L6r
          - 1./2.*pow(m66,2)*L5r - pow(m66,2)*L4r + 4*m44*m66*L6r - 2*
         m44*m66*L4r + 2*m11*m66*L8r + 2*m11*m66*L6r - m11*m66*L5r - 
         m11*m66*L4r + 4*m11*m44*L6r - 2*m11*m44*L4r + pow(m11,2)*L8r
          - 1./2.*pow(m11,2)*L5r );

      F13p6x22L +=  + Bb(m33,mu2) * ( m66*L4r*RD346e + 4./9.*pow(m66,2)
         *L8r*pow(RS34e,2) + 4./9.*pow(m66,2)*L7r*pow(RS34e,2) + 4./9.*
         pow(m66,2)*L6r*pow(RS34e,2) + 2*m44*L4r*RD346e + 16./9.*m44*
         m66*L7r*RS34e*RS36e + 8./9.*m44*m66*L6r*pow(RS36e,2) + 8./9.*
         m44*m66*L6r*pow(RS34e,2) + 8./9.*pow(m44,2)*L8r*pow(RS36e,2)
          + 16./9.*pow(m44,2)*L7r*pow(RS36e,2) + 16./9.*pow(m44,2)*L6r*
         pow(RS36e,2) + 1./2.*m33*L5r*RD346e + 5./3.*m33*L3r*RD346e + 2.
         /3.*m33*L0r*RD346e - 8./3.*m33*m66*L7r*RS34e + 8./3.*m33*m66*
         L6r*RS3461e - 8./3.*m33*m66*L6r*RC346e - 2./9.*m33*m66*L5r*
         pow(RS34e,2) - 4./3.*m33*m66*L4r*RS3461e - 4./9.*m33*m66*L4r*
         pow(RS36e,2) - 2./9.*m33*m66*L4r*pow(RS34e,2) + 4./3.*m33*m66*
         L4r*RC346e - 16./3.*m33*m44*L7r*RS36e + 16./3.*m33*m44*L6r*
         RS3461e - 16./3.*m33*m44*L6r*RC346e - 4./9.*m33*m44*L5r*pow(
         RS36e,2) - 8./3.*m33*m44*L4r*RS3461e - 8./9.*m33*m44*L4r*pow(
         RS36e,2) - 4./9.*m33*m44*L4r*pow(RS34e,2) + 8./3.*m33*m44*L4r*
         RC346e );

      F13p6x22L +=  + Bb(m33,mu2) * ( 8./3.*pow(m33,2)*L8r*RS3461e - 8./
         3.*pow(m33,2)*L8r*RC346e + 4*pow(m33,2)*L7r - 4./3.*pow(m33,2)
         *L5r*RS3461e + 4./3.*pow(m33,2)*L5r*RC346e - 1./6.*m11*L5r*
         RD346e );

      F13p6x22L +=  + Bb(m34,mu2) * ( 4*m44*m66*L6r - 2*m44*m66*L4r + 2
         *pow(m44,2)*L8r + 8*pow(m44,2)*L6r - pow(m44,2)*L5r - 4*pow(
         m44,2)*L4r + 4*m33*m66*L6r - 2*m33*m66*L4r + 4*m33*m44*L8r + 8
         *m33*m44*L6r - 2*m33*m44*L5r - 4*m33*m44*L4r + 2*pow(m33,2)*
         L8r - pow(m33,2)*L5r );

      F13p6x22L +=  + Bb(m36,mu2) * ( pow(m66,2)*L8r + 2*pow(m66,2)*L6r
          - 1./2.*pow(m66,2)*L5r - pow(m66,2)*L4r + 4*m44*m66*L6r - 2*
         m44*m66*L4r + 2*m33*m66*L8r + 2*m33*m66*L6r - m33*m66*L5r - 
         m33*m66*L4r + 4*m33*m44*L6r - 2*m33*m44*L4r + pow(m33,2)*L8r
          - 1./2.*pow(m33,2)*L5r );

      F13p6x22L +=  + Bb(me2,mu2) * (  - 8./9.*m66*me2*L5r*pow(RSe63,2)
          + 16./9.*m66*me2*L5r*RSe61*RSe63 - 8./9.*m66*me2*L5r*pow(
         RSe61,2) - 4./3.*m66*me2*L4r*pow(RSe63,2) + 8./3.*m66*me2*L4r*
         RSe61*RSe63 - 4./3.*m66*me2*L4r*pow(RSe61,2) + 16./9.*pow(
         m66,2)*L8r*pow(RSe63,2) - 32./9.*pow(m66,2)*L8r*RSe61*RSe63 + 
         16./9.*pow(m66,2)*L8r*pow(RSe61,2) + 16./9.*pow(m66,2)*L7r*
         pow(RSe63,2) - 32./9.*pow(m66,2)*L7r*RSe61*RSe63 + 16./9.*pow(
         m66,2)*L7r*pow(RSe61,2) + 16./9.*pow(m66,2)*L6r*pow(RSe63,2)
          - 32./9.*pow(m66,2)*L6r*RSe61*RSe63 + 16./9.*pow(m66,2)*L6r*
         pow(RSe61,2) - 4./9.*m44*me2*L5r*pow(RSe63,2) + 8./9.*m44*me2*
         L5r*RSe61*RSe63 - 4./9.*m44*me2*L5r*pow(RSe61,2) - 8./3.*m44*
         me2*L4r*pow(RSe63,2) + 16./3.*m44*me2*L4r*RSe61*RSe63 - 8./3.*
         m44*me2*L4r*pow(RSe61,2) - 32./9.*m44*m66*L7r*pow(RSe63,2) + 
         64./9.*m44*m66*L7r*RSe61*RSe63 - 32./9.*m44*m66*L7r*pow(
         RSe61,2) + 40./9.*m44*m66*L6r*pow(RSe63,2) - 80./9.*m44*m66*
         L6r*RSe61*RSe63 );

      F13p6x22L +=  + Bb(me2,mu2) * ( 40./9.*m44*m66*L6r*pow(RSe61,2)
          + 8./9.*pow(m44,2)*L8r*pow(RSe63,2) - 16./9.*pow(m44,2)*L8r*
         RSe61*RSe63 + 8./9.*pow(m44,2)*L8r*pow(RSe61,2) + 16./9.*pow(
         m44,2)*L7r*pow(RSe63,2) - 32./9.*pow(m44,2)*L7r*RSe61*RSe63 + 
         16./9.*pow(m44,2)*L7r*pow(RSe61,2) + 16./9.*pow(m44,2)*L6r*
         pow(RSe63,2) - 32./9.*pow(m44,2)*L6r*RSe61*RSe63 + 16./9.*pow(
         m44,2)*L6r*pow(RSe61,2) );

      F13p6x22L +=  + Cb(m11,mu2) * (  - 8./3.*m11*m66*L6r*RD146e + 4./
         3.*m11*m66*L4r*RD146e - 16./3.*m11*m44*L6r*RD146e + 8./3.*m11*
         m44*L4r*RD146e - 8./3.*pow(m11,2)*L8r*RD146e + 4./3.*pow(
         m11,2)*L5r*RD146e );

      F13p6x22L +=  + Cb(m33,mu2) * (  - 8./3.*m33*m66*L6r*RD346e + 4./
         3.*m33*m66*L4r*RD346e - 16./3.*m33*m44*L6r*RD346e + 8./3.*m33*
         m44*L4r*RD346e - 8./3.*pow(m33,2)*L8r*RD346e + 4./3.*pow(
         m33,2)*L5r*RD346e );

      F13p6x22L +=  - 8*pow(m66,2)*pow(L4r,2) - 32*m44*m66*pow(L4r,2)
          - 32*pow(m44,2)*pow(L4r,2) - 8*m33*m66*L4r*L5r - 16*m33*m44*
         L4r*L5r - 2*pow(m33,2)*pow(L5r,2) - 8*m11*m66*L4r*L5r - 16*m11
         *m44*L4r*L5r - 4*m11*m33*pow(L5r,2) - 2*pow(m11,2)*pow(L5r,2);

  return F13p6x22L/pow(f0,4);
}

double fv2s2nf3p6K(const quarkmassnf mass, const Ki Kiin){
  double KK[116],mu;
  Kiin.out(KK,mu);
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if (mu0 != mu){
    cout << "ERROR: differing mu's in fv2s2nf3p6L\n";}
  if ( (nq != 4) || (B0mq.size() != 4)){
    cout << "ERROR: fv2s2nf3p6K probably called with wrong number"
	 << "of quark masses\n";}
  // four masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m33 = 2.*B0mq[1];
  double m44 = 2.*B0mq[2];
  double m66 = 2.*B0mq[3];
  double F13p6x22K =
       + KK[19] * ( 4*pow(m33,2) + 4*pow(m11,2) );

      F13p6x22K +=  + KK[20] * ( 4*m33*m66 + 8*m33*m44 + 4*m11*m66 + 8*
         m11*m44 );

      F13p6x22K +=  + KK[21] * ( 8*pow(m66,2) + 16*pow(m44,2) );

      F13p6x22K +=  + KK[22] * ( 8*pow(m66,2) + 32*m44*m66 + 32*pow(
         m44,2) );

      F13p6x22K +=  + KK[23] * ( 8*m11*m33 );

  return F13p6x22K/pow(f0,4);
}

double fv2s2nf3p6R(const quarkmassnf mass){
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if ( (nq != 4) || (B0mq.size() != 4)){
    cout << "ERROR: fv2s2nf3p6R probably called with wrong number"
	 << "of quark masses\n";}
  // four masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m33 = 2.*B0mq[1];
  double m44 = 2.*B0mq[2];
  double m66 = 2.*B0mq[3];
  double m13 = (m11+m33)/2;
  double m14 = (m11+m44)/2;
  double m16 = (m11+m66)/2;
  double m34 = (m33+m44)/2;
  double m36 = (m33+m66)/2;
  double m46 = (m44+m66)/2;
  double me2 = (m44+2.*m66)/3.;

  double RN13 = RN(m11,m33);
  double RN1e = RN(m11,me2);
  double RN3e = RN(m33,me2);

  double RD13 = m11-m33;
  double RD14 = m11-m44;
  double RD16 = m11-m66;
  double RD34 = m33-m44;
  double RD36 = m33-m66;
  double RD46 = m44-m66;

  double RD146e = RD(m11,m44,m66,me2);
  double RD346e = RD(m33,m44,m66,me2);

  double RC146e = RC(m11,m44,m66,me2);
  double RC346e = RC(m33,m44,m66,me2);

  double RS14e = RS(m11,m44,me2);
  double RS16e = RS(m11,m66,me2);
  double RS34e = RS(m33,m44,me2);
  double RS36e = RS(m33,m66,me2);
  double RSe61 = RS(me2,m66,m11);
  double RSe63 = RS(me2,m66,m33);

  double RS1463e = RS(m11,m44,m66,m33,me2);
  double RS3461e = RS(m33,m44,m66,m11,me2);
  double RSe6611 = RS(me2,m66,m66,m11,m11);
  double RSe6613 = RS(me2,m66,m66,m11,m33);
  double RSe6633 = RS(me2,m66,m66,m33,m33);

  double mu2 = mu0*mu0;

  double 
   F13p6x22R =
       + pi162 * ( 1./72.*me2*RD346e*RSe6613 - 1./144.*me2*RD346e*
         RSe6633 - 1./144.*me2*RSe6611*RD346e + 1./72.*me2*RD146e*
         RSe6613 - 1./144.*me2*RD146e*RSe6633 - 1./144.*me2*RD146e*
         RSe6611 + 7./48.*pow(me2,2) + 1./8.*pow(me2,2)*RSe6613 + 1./18.
         *pow(me2,2)*pow(RSe6613,2) + 1./12.*pow(me2,2)*RSe6633 - 1./18.
         *pow(me2,2)*RSe6633*RSe6613 + 1./72.*pow(me2,2)*pow(RSe6633,2)
          + 1./12.*pow(me2,2)*RSe6611 - 1./18.*pow(me2,2)*RSe6611*
         RSe6613 + 1./36.*pow(me2,2)*RSe6611*RSe6633 + 1./72.*pow(
         me2,2)*pow(RSe6611,2) - 1./96.*m66*RN3e*RD34*RD36 - 1./96.*m66
         *RN1e*RD14*RD16 - 1./18.*m66*me2 + 1./72.*m66*me2*RSe6613 + 1./
         36.*m66*me2*RSe63 + 13./108.*m66*me2*pow(RSe63,2) - 1./144.*
         m66*me2*RSe6633 - 1./27.*m66*me2*RS34e*RSe63 + 1./108.*m66*me2
         *pow(RS34e,2) + 1./36.*m66*me2*RSe61 - 13./54.*m66*me2*RSe61*
         RSe63 + 1./27.*m66*me2*RSe61*RS34e + 13./108.*m66*me2*pow(
         RSe61,2) - 1./144.*m66*me2*RSe6611 + 1./27.*m66*me2*RS14e*
         RSe63 );

      F13p6x22R +=  + pi162 * (  - 1./54.*m66*me2*RS14e*RS34e - 1./27.*
         m66*me2*RS14e*RSe61 + 1./108.*m66*me2*pow(RS14e,2) - 43./384.*
         pow(m66,2) - 1./6.*m46*me2*pow(RSe63,2) - 1./18.*m46*me2*RS36e
         *RSe63 + 1./18.*m46*me2*RS34e*RSe63 + 1./3.*m46*me2*RSe61*
         RSe63 + 1./18.*m46*me2*RSe61*RS36e - 1./18.*m46*me2*RSe61*
         RS34e - 1./6.*m46*me2*pow(RSe61,2) + 1./18.*m46*me2*RS16e*
         RSe63 - 1./18.*m46*me2*RS16e*RSe61 - 1./18.*m46*me2*RS14e*
         RSe63 + 1./18.*m46*me2*RS14e*RSe61 + 1./8.*m46*m66 - 1./27.*
         m46*m66*RS36e*RSe63 - 1./54.*m46*m66*RS34e*RSe63 + 1./54.*m46*
         m66*RS34e*RS36e + 1./108.*m46*m66*pow(RS34e,2) + 1./27.*m46*
         m66*RSe61*RS36e + 1./54.*m46*m66*RSe61*RS34e + 1./27.*m46*m66*
         RS16e*RSe63 - 1./54.*m46*m66*RS16e*RS34e - 1./27.*m46*m66*
         RS16e*RSe61 + 1./54.*m46*m66*RS14e*RSe63 - 1./54.*m46*m66*
         RS14e*RS36e - 1./54.*m46*m66*RS14e*RS34e - 1./54.*m46*m66*
         RS14e*RSe61 + 1./54.*m46*m66*RS14e*RS16e + 1./108.*m46*m66*
         pow(RS14e,2) );

      F13p6x22R +=  + pi162 * (  - 1./48.*m44*RN3e*RD34*RD36 - 1./48.*
         m44*RN1e*RD14*RD16 - 1./36.*m44*me2 + 1./36.*m44*me2*RSe6613
          - 1./36.*m44*me2*RSe63 + 19./216.*m44*me2*pow(RSe63,2) - 1./
         72.*m44*me2*RSe6633 + 1./108.*m44*me2*RS36e*RSe63 + 1./216.*
         m44*me2*pow(RS36e,2) - 1./36.*m44*me2*RSe61 - 19./108.*m44*me2
         *RSe61*RSe63 - 1./108.*m44*me2*RSe61*RS36e + 19./216.*m44*me2*
         pow(RSe61,2) - 1./72.*m44*me2*RSe6611 - 1./108.*m44*me2*RS16e*
         RSe63 - 1./108.*m44*me2*RS16e*RS36e + 1./108.*m44*me2*RS16e*
         RSe61 + 1./216.*m44*me2*pow(RS16e,2) - 5./32.*m44*m66 + 1./8.*
         m44*m46 - 1./36.*m44*m46*pow(RSe63,2) - 1./54.*m44*m46*RS36e*
         RSe63 + 1./108.*m44*m46*pow(RS36e,2) + 1./54.*m44*m46*RS34e*
         RSe63 + 1./54.*m44*m46*RS34e*RS36e + 1./18.*m44*m46*RSe61*
         RSe63 + 1./54.*m44*m46*RSe61*RS36e - 1./54.*m44*m46*RSe61*
         RS34e - 1./36.*m44*m46*pow(RSe61,2) + 1./54.*m44*m46*RS16e*
         RSe63 - 1./54.*m44*m46*RS16e*RS36e - 1./54.*m44*m46*RS16e*
         RS34e );

      F13p6x22R +=  + pi162 * (  - 1./54.*m44*m46*RS16e*RSe61 + 1./108.
         *m44*m46*pow(RS16e,2) - 1./54.*m44*m46*RS14e*RSe63 - 1./54.*
         m44*m46*RS14e*RS36e + 1./54.*m44*m46*RS14e*RSe61 + 1./54.*m44*
         m46*RS14e*RS16e - 11./96.*pow(m44,2) + 1./24.*pow(m44,2)*pow(
         RSe63,2) + 1./12.*pow(m44,2)*RS36e*RSe63 + 1./24.*pow(m44,2)*
         pow(RS36e,2) - 1./12.*pow(m44,2)*RSe61*RSe63 - 1./12.*pow(
         m44,2)*RSe61*RS36e + 1./24.*pow(m44,2)*pow(RSe61,2) - 1./12.*
         pow(m44,2)*RS16e*RSe63 - 1./12.*pow(m44,2)*RS16e*RS36e + 1./12.
         *pow(m44,2)*RS16e*RSe61 + 1./24.*pow(m44,2)*pow(RS16e,2) + 7./
         288.*m36*RD346e - 1./96.*m36*RD146e + 1./36.*m36*me2*RSe6613
          + 5./72.*m36*me2*RSe63 - 7./144.*m36*me2*RSe6633 - 5./72.*m36
         *me2*RSe61 + 1./48.*m36*me2*RSe6611 + 113./864.*m36*m66 + 7./
         864.*m36*m66*RS3461e + 7./864.*m36*m66*RS1463e - 7./432.*m36*
         m66*RSe6613 + 29./432.*m36*m66*RSe63 + 1./432.*m36*m66*RSe6633
          - 29./864.*m36*m66*RS34e - 1./864.*m36*m66*RC346e - 19./432.*
         m36*m66*RSe61 );

      F13p6x22R +=  + pi162 * ( 1./432.*m36*m66*RSe6611 + 19./864.*m36*
         m66*RS14e - 1./864.*m36*m66*RC146e + 1./16.*m36*m44 + 7./144.*
         m34*RD346e - 1./48.*m34*RD146e + 1./18.*m34*me2*RSe6613 - 5./
         72.*m34*me2*RSe63 - 7./72.*m34*me2*RSe6633 + 5./72.*m34*me2*
         RSe61 + 1./24.*m34*me2*RSe6611 + 1./16.*m34*m66 + 35./108.*m34
         *m44 + 7./432.*m34*m44*RS3461e + 7./432.*m34*m44*RS1463e - 7./
         216.*m34*m44*RSe6613 - 29./432.*m34*m44*RSe63 + 1./216.*m34*
         m44*RSe6633 - 29./432.*m34*m44*RS36e - 1./432.*m34*m44*RC346e
          + 19./432.*m34*m44*RSe61 + 1./216.*m34*m44*RSe6611 + 19./432.
         *m34*m44*RS16e - 1./432.*m34*m44*RC146e - 1./24.*m33*RD346e - 
         1./144.*m33*RD346e*RS3461e + 1./288.*m33*RD346e*RC346e - 1./
         144.*m33*RD146e*RS3461e + 1./288.*m33*RD146e*RC346e - 1./32.*
         m33*RN3e*RD34*RD36 + 1./48.*m33*me2*RSe6613 - 19./324.*m33*me2
         *RSe6613*RS3461e - 1./324.*m33*me2*RSe6613*RS1463e + 1./162.*
         m33*me2*pow(RSe6613,2) - 5./48.*m33*me2*RSe6633 + 55./648.*m33
         *me2*RSe6633*RS3461e );

      F13p6x22R +=  + pi162 * ( 37./648.*m33*me2*RSe6633*RS1463e - 19./
         162.*m33*me2*RSe6633*RSe6613 + 73./648.*m33*me2*pow(RSe6633,2)
          + 19./648.*m33*me2*RC346e*RSe6613 - 91./1296.*m33*me2*RC346e*
         RSe6633 + 19./648.*m33*me2*RSe6611*RS3461e + 1./648.*m33*me2*
         RSe6611*RS1463e - 1./162.*m33*me2*RSe6611*RSe6613 + 1./324.*
         m33*me2*RSe6611*RSe6633 - 19./1296.*m33*me2*RSe6611*RC346e + 1.
         /648.*m33*me2*pow(RSe6611,2) + 1./648.*m33*me2*RC146e*RSe6613
          - 1./1296.*m33*me2*RC146e*RSe6633 - 1./1296.*m33*me2*RC146e*
         RSe6611 - 1./12.*m33*m66 + 7./144.*m33*m66*RS3461e + 1./36.*
         m33*m66*RS1463e - 1./18.*m33*m66*RSe6613 + 1./36.*m33*m66*
         RSe63 + 1./18.*m33*m66*RSe6633 + 1./108.*m33*m66*pow(RS36e,2)
          - 1./72.*m33*m66*RS34e - 1./12.*m33*m66*RS34e*RSe63 + 1./108.
         *m33*m66*RS34e*RS36e + 5./216.*m33*m66*pow(RS34e,2) - 7./288.*
         m33*m66*RC346e - 1./36.*m33*m66*RSe61 + 1./12.*m33*m66*RSe61*
         RS34e - 1./54.*m33*m66*RS16e*RS36e - 1./108.*m33*m66*RS16e*
         RS34e );

      F13p6x22R +=  + pi162 * ( 1./72.*m33*m66*RS14e - 1./108.*m33*m66*
         RS14e*RS36e - 5./108.*m33*m66*RS14e*RS34e - 1./18.*m33*m46*
         RS36e*RSe63 - 1./54.*m33*m46*pow(RS36e,2) + 1./18.*m33*m46*
         RS34e*RSe63 + 1./27.*m33*m46*RS34e*RS36e - 1./54.*m33*m46*pow(
         RS34e,2) + 1./18.*m33*m46*RSe61*RS36e - 1./18.*m33*m46*RSe61*
         RS34e + 1./54.*m33*m46*RS16e*RS36e - 1./54.*m33*m46*RS16e*
         RS34e - 1./54.*m33*m46*RS14e*RS36e + 1./54.*m33*m46*RS14e*
         RS34e - 1./6.*m33*m44 + 7./72.*m33*m44*RS3461e + 1./18.*m33*
         m44*RS1463e - 1./9.*m33*m44*RSe6613 - 1./36.*m33*m44*RSe63 + 1.
         /9.*m33*m44*RSe6633 - 1./36.*m33*m44*RS36e + 1./9.*m33*m44*
         RS36e*RSe63 + 7./108.*m33*m44*pow(RS36e,2) - 1./36.*m33*m44*
         RS34e*RSe63 + 1./108.*m33*m44*RS34e*RS36e + 1./108.*m33*m44*
         pow(RS34e,2) - 7./144.*m33*m44*RC346e + 1./36.*m33*m44*RSe61
          - 1./9.*m33*m44*RSe61*RS36e + 1./36.*m33*m44*RSe61*RS34e + 1./
         36.*m33*m44*RS16e - 7./54.*m33*m44*RS16e*RS36e - 1./108.*m33*
         m44*RS16e*RS34e );

      F13p6x22R +=  + pi162 * (  - 1./108.*m33*m44*RS14e*RS36e - 1./54.
         *m33*m44*RS14e*RS34e - 1./72.*m33*m36*RS3461e + 1./9.*m33*m36*
         RSe63 - 13./144.*m33*m36*RS34e + 7./288.*m33*m36*RC346e - 1./9.
         *m33*m36*RSe61 + 1./18.*m33*m36*RS14e - 1./36.*m33*m34*RS3461e
          - 1./9.*m33*m34*RSe63 - 13./72.*m33*m34*RS36e + 7./144.*m33*
         m34*RC346e + 1./9.*m33*m34*RSe61 + 1./9.*m33*m34*RS16e + 91./
         512.*pow(m33,2) + 1./96.*pow(m33,2)*RS3461e + 5./324.*pow(
         m33,2)*pow(RS3461e,2) + 1./12.*pow(m33,2)*RS1463e + 1./648.*
         pow(m33,2)*RS1463e*RS3461e - 1./6.*pow(m33,2)*RSe6613 - 1./324.
         *pow(m33,2)*RSe6613*RS3461e + 1./6.*pow(m33,2)*RSe6633 + 1./
         648.*pow(m33,2)*RSe6633*RS3461e - 13./144.*pow(m33,2)*RS36e - 
         13./288.*pow(m33,2)*RS34e - 7./96.*pow(m33,2)*RC346e - 7./162.
         *pow(m33,2)*RC346e*RS3461e - 37./1296.*pow(m33,2)*RC346e*
         RS1463e + 37./648.*pow(m33,2)*RC346e*RSe6613 - 73./1296.*pow(
         m33,2)*RC346e*RSe6633 + 41./1296.*pow(m33,2)*pow(RC346e,2) + 1.
         /648.*pow(m33,2)*RSe6611*RS3461e );

      F13p6x22R +=  + pi162 * (  - 1./1296.*pow(m33,2)*RSe6611*RC346e
          + 1./36.*pow(m33,2)*RS16e + 1./72.*pow(m33,2)*RS14e - 1./1296.
         *pow(m33,2)*RC146e*RS3461e + 1./2592.*pow(m33,2)*RC146e*RC346e
          - 1./96.*m16*RD346e + 7./288.*m16*RD146e + 1./36.*m16*me2*
         RSe6613 - 5./72.*m16*me2*RSe63 + 1./48.*m16*me2*RSe6633 + 5./
         72.*m16*me2*RSe61 - 7./144.*m16*me2*RSe6611 + 113./864.*m16*
         m66 + 7./864.*m16*m66*RS3461e + 7./864.*m16*m66*RS1463e - 7./
         432.*m16*m66*RSe6613 - 19./432.*m16*m66*RSe63 + 1./432.*m16*
         m66*RSe6633 + 19./864.*m16*m66*RS34e - 1./864.*m16*m66*RC346e
          + 29./432.*m16*m66*RSe61 + 1./432.*m16*m66*RSe6611 - 29./864.
         *m16*m66*RS14e - 1./864.*m16*m66*RC146e + 1./16.*m16*m44 + 5./
         864.*m16*m33 - 5./864.*m16*m33*RS3461e + 7./864.*m16*m33*
         RS1463e - 7./432.*m16*m33*RSe6613 + 5./432.*m16*m33*RSe63 + 1./
         432.*m16*m33*RSe6633 + 25./864.*m16*m33*RS34e - 5./432.*m16*
         m33*RC346e + 5./432.*m16*m33*RSe61 + 1./432.*m16*m33*RSe6611
          - 5./864.*m16*m33*RS14e );

      F13p6x22R +=  + pi162 * (  - 1./864.*m16*m33*RC146e - 1./48.*m14*
         RD346e + 7./144.*m14*RD146e + 1./18.*m14*me2*RSe6613 + 5./72.*
         m14*me2*RSe63 + 1./24.*m14*me2*RSe6633 - 5./72.*m14*me2*RSe61
          - 7./72.*m14*me2*RSe6611 + 1./16.*m14*m66 + 35./108.*m14*m44
          + 7./432.*m14*m44*RS3461e + 7./432.*m14*m44*RS1463e - 7./216.
         *m14*m44*RSe6613 + 19./432.*m14*m44*RSe63 + 1./216.*m14*m44*
         RSe6633 + 19./432.*m14*m44*RS36e - 1./432.*m14*m44*RC346e - 29.
         /432.*m14*m44*RSe61 + 1./216.*m14*m44*RSe6611 - 29./432.*m14*
         m44*RS16e - 1./432.*m14*m44*RC146e + 5./432.*m14*m33 - 5./432.
         *m14*m33*RS3461e + 7./432.*m14*m33*RS1463e - 7./216.*m14*m33*
         RSe6613 - 5./432.*m14*m33*RSe63 + 1./216.*m14*m33*RSe6633 + 25.
         /432.*m14*m33*RS36e - 5./216.*m14*m33*RC346e - 5./432.*m14*m33
         *RSe61 + 1./216.*m14*m33*RSe6611 - 5./432.*m14*m33*RS16e - 1./
         432.*m14*m33*RC146e - 1./648.*m13*RD346e*RS3461e - 1./648.*m13
         *RD346e*RS1463e + 1./324.*m13*RD346e*RSe6613 - 1./648.*m13*
         RD346e*RSe6633 );

      F13p6x22R +=  + pi162 * ( 1./1296.*m13*RD346e*RC346e - 1./648.*
         m13*RSe6611*RD346e + 1./1296.*m13*RC146e*RD346e - 1./648.*m13*
         RD146e*RS3461e - 1./648.*m13*RD146e*RS1463e + 1./324.*m13*
         RD146e*RSe6613 - 1./648.*m13*RD146e*RSe6633 + 1./1296.*m13*
         RD146e*RC346e - 1./648.*m13*RD146e*RSe6611 + 1./1296.*m13*
         RD146e*RC146e - 1./162.*m13*me2*RSe6613*RS3461e - 1./162.*m13*
         me2*RSe6613*RS1463e + 1./81.*m13*me2*pow(RSe6613,2) + 1./324.*
         m13*me2*RSe6633*RS3461e + 1./324.*m13*me2*RSe6633*RS1463e - 1./
         81.*m13*me2*RSe6633*RSe6613 + 1./324.*m13*me2*pow(RSe6633,2)
          + 1./324.*m13*me2*RC346e*RSe6613 - 1./648.*m13*me2*RC346e*
         RSe6633 + 1./324.*m13*me2*RSe6611*RS3461e + 1./324.*m13*me2*
         RSe6611*RS1463e - 1./81.*m13*me2*RSe6611*RSe6613 + 1./162.*m13
         *me2*RSe6611*RSe6633 - 1./648.*m13*me2*RSe6611*RC346e + 1./324.
         *m13*me2*pow(RSe6611,2) + 1./324.*m13*me2*RC146e*RSe6613 - 1./
         648.*m13*me2*RC146e*RSe6633 - 1./648.*m13*me2*RC146e*RSe6611
          + 1./16.*m13*m33 );

      F13p6x22R +=  + pi162 * ( 1./324.*m13*m33*pow(RS3461e,2) + 1./324.
         *m13*m33*RS1463e*RS3461e - 1./162.*m13*m33*RSe6613*RS3461e + 1.
         /324.*m13*m33*RSe6633*RS3461e - 1./324.*m13*m33*RC346e*RS3461e
          - 1./648.*m13*m33*RC346e*RS1463e + 1./324.*m13*m33*RC346e*
         RSe6613 - 1./648.*m13*m33*RC346e*RSe6633 + 1./1296.*m13*m33*
         pow(RC346e,2) + 1./324.*m13*m33*RSe6611*RS3461e - 1./648.*m13*
         m33*RSe6611*RC346e - 1./648.*m13*m33*RC146e*RS3461e + 1./1296.
         *m13*m33*RC146e*RC346e - 1./144.*m11*RD346e*RS1463e + 1./288.*
         m11*RC146e*RD346e - 1./24.*m11*RD146e - 1./144.*m11*RD146e*
         RS1463e + 1./288.*m11*RD146e*RC146e - 1./32.*m11*RN1e*RD14*
         RD16 + 1./48.*m11*me2*RSe6613 - 1./324.*m11*me2*RSe6613*
         RS3461e - 19./324.*m11*me2*RSe6613*RS1463e + 1./162.*m11*me2*
         pow(RSe6613,2) + 1./648.*m11*me2*RSe6633*RS3461e + 19./648.*
         m11*me2*RSe6633*RS1463e - 1./162.*m11*me2*RSe6633*RSe6613 + 1./
         648.*m11*me2*pow(RSe6633,2) + 1./648.*m11*me2*RC346e*RSe6613
          - 1./1296.*m11*me2*RC346e*RSe6633 );

      F13p6x22R +=  + pi162 * (  - 5./48.*m11*me2*RSe6611 + 37./648.*
         m11*me2*RSe6611*RS3461e + 55./648.*m11*me2*RSe6611*RS1463e - 
         19./162.*m11*me2*RSe6611*RSe6613 + 1./324.*m11*me2*RSe6611*
         RSe6633 - 1./1296.*m11*me2*RSe6611*RC346e + 73./648.*m11*me2*
         pow(RSe6611,2) + 19./648.*m11*me2*RC146e*RSe6613 - 19./1296.*
         m11*me2*RC146e*RSe6633 - 91./1296.*m11*me2*RC146e*RSe6611 - 1./
         12.*m11*m66 + 1./48.*m11*m66*RS1463e - 1./18.*m11*m66*RSe6613
          - 1./36.*m11*m66*RSe63 + 1./36.*m11*m66*RSe61 + 1./18.*m11*
         m66*RSe6611 + 1./108.*m11*m66*pow(RS16e,2) - 1./36.*m11*m66*
         RS14e + 1./12.*m11*m66*RS14e*RSe63 - 1./12.*m11*m66*RS14e*
         RSe61 + 1./108.*m11*m66*RS14e*RS16e + 5./216.*m11*m66*pow(
         RS14e,2) - 7./288.*m11*m66*RC146e + 1./18.*m11*m46*RS16e*RSe63
          + 1./54.*m11*m46*RS16e*RS36e - 1./54.*m11*m46*RS16e*RS34e - 1.
         /18.*m11*m46*RS16e*RSe61 - 1./54.*m11*m46*pow(RS16e,2) - 1./18.
         *m11*m46*RS14e*RSe63 - 1./54.*m11*m46*RS14e*RS36e + 1./54.*m11
         *m46*RS14e*RS34e );

      F13p6x22R +=  + pi162 * ( 1./18.*m11*m46*RS14e*RSe61 + 1./27.*m11
         *m46*RS14e*RS16e - 1./54.*m11*m46*pow(RS14e,2) - 1./6.*m11*m44
          + 1./24.*m11*m44*RS1463e - 1./9.*m11*m44*RSe6613 + 1./36.*m11
         *m44*RSe63 - 1./36.*m11*m44*RSe61 + 1./9.*m11*m44*RSe6611 - 1./
         18.*m11*m44*RS16e - 1./9.*m11*m44*RS16e*RSe63 + 1./9.*m11*m44*
         RS16e*RSe61 + 7./108.*m11*m44*pow(RS16e,2) + 1./36.*m11*m44*
         RS14e*RSe63 - 1./36.*m11*m44*RS14e*RSe61 + 1./108.*m11*m44*
         RS14e*RS16e + 1./108.*m11*m44*pow(RS14e,2) - 7./144.*m11*m44*
         RC146e + 5./864.*m11*m36 + 7./864.*m11*m36*RS3461e - 5./864.*
         m11*m36*RS1463e - 7./432.*m11*m36*RSe6613 + 5./432.*m11*m36*
         RSe63 + 1./432.*m11*m36*RSe6633 - 5./864.*m11*m36*RS34e - 1./
         864.*m11*m36*RC346e + 5./432.*m11*m36*RSe61 + 1./432.*m11*m36*
         RSe6611 + 25./864.*m11*m36*RS14e - 5./432.*m11*m36*RC146e + 5./
         432.*m11*m34 + 7./432.*m11*m34*RS3461e - 5./432.*m11*m34*
         RS1463e - 7./216.*m11*m34*RSe6613 - 5./432.*m11*m34*RSe63 + 1./
         216.*m11*m34*RSe6633 );

      F13p6x22R +=  + pi162 * (  - 5./432.*m11*m34*RS36e - 1./432.*m11*
         m34*RC346e - 5./432.*m11*m34*RSe61 + 1./216.*m11*m34*RSe6611
          + 25./432.*m11*m34*RS16e - 5./216.*m11*m34*RC146e - 101./768.
         *m11*m33 + 7./96.*m11*m33*RS3461e + 1./648.*m11*m33*pow(
         RS3461e,2) - 1./96.*m11*m33*RS1463e + 5./162.*m11*m33*RS1463e*
         RS3461e + 1./648.*m11*m33*pow(RS1463e,2) - 1./324.*m11*m33*
         RSe6613*RS3461e - 1./324.*m11*m33*RSe6613*RS1463e + 1./648.*
         m11*m33*RSe6633*RS3461e + 1./648.*m11*m33*RSe6633*RS1463e + 5./
         144.*m11*m33*RS36e + 5./288.*m11*m33*RS34e - 1./648.*m11*m33*
         RC346e*RS3461e - 5./324.*m11*m33*RC346e*RS1463e + 1./648.*m11*
         m33*RC346e*RSe6613 - 1./1296.*m11*m33*RC346e*RSe6633 + 1./2592.
         *m11*m33*pow(RC346e,2) + 1./648.*m11*m33*RSe6611*RS3461e + 1./
         648.*m11*m33*RSe6611*RS1463e - 1./1296.*m11*m33*RSe6611*RC346e
          + 1./144.*m11*m33*RS16e + 1./288.*m11*m33*RS14e - 5./324.*m11
         *m33*RC146e*RS3461e - 1./648.*m11*m33*RC146e*RS1463e + 1./648.
         *m11*m33*RC146e*RSe6613 );

      F13p6x22R +=  + pi162 * (  - 1./1296.*m11*m33*RC146e*RSe6633 + 5./
         648.*m11*m33*RC146e*RC346e - 1./1296.*m11*m33*RC146e*RSe6611
          + 1./2592.*m11*m33*pow(RC146e,2) - 1./72.*m11*m16*RS1463e - 1.
         /9.*m11*m16*RSe63 + 1./18.*m11*m16*RS34e + 1./9.*m11*m16*RSe61
          - 13./144.*m11*m16*RS14e + 7./288.*m11*m16*RC146e - 1./36.*
         m11*m14*RS1463e + 1./9.*m11*m14*RSe63 + 1./9.*m11*m14*RS36e - 
         1./9.*m11*m14*RSe61 - 13./72.*m11*m14*RS16e + 7./144.*m11*m14*
         RC146e + 1./16.*m11*m13 + 1./324.*m11*m13*RS1463e*RS3461e + 1./
         324.*m11*m13*pow(RS1463e,2) - 1./162.*m11*m13*RSe6613*RS1463e
          + 1./324.*m11*m13*RSe6633*RS1463e - 1./648.*m11*m13*RC346e*
         RS1463e + 1./324.*m11*m13*RSe6611*RS1463e - 1./648.*m11*m13*
         RC146e*RS3461e - 1./324.*m11*m13*RC146e*RS1463e + 1./324.*m11*
         m13*RC146e*RSe6613 - 1./648.*m11*m13*RC146e*RSe6633 + 1./1296.
         *m11*m13*RC146e*RC346e - 1./648.*m11*m13*RC146e*RSe6611 + 1./
         1296.*m11*m13*pow(RC146e,2) + 337./1536.*pow(m11,2) + 1./96.*
         pow(m11,2)*RS1463e );

      F13p6x22R +=  + pi162 * ( 1./648.*pow(m11,2)*RS1463e*RS3461e + 5./
         324.*pow(m11,2)*pow(RS1463e,2) - 1./6.*pow(m11,2)*RSe6613 - 1./
         324.*pow(m11,2)*RSe6613*RS1463e + 1./648.*pow(m11,2)*RSe6633*
         RS1463e - 1./1296.*pow(m11,2)*RC346e*RS1463e + 1./6.*pow(
         m11,2)*RSe6611 + 1./648.*pow(m11,2)*RSe6611*RS1463e - 13./144.
         *pow(m11,2)*RS16e - 13./288.*pow(m11,2)*RS14e - 7./96.*pow(
         m11,2)*RC146e - 37./1296.*pow(m11,2)*RC146e*RS3461e - 7./162.*
         pow(m11,2)*RC146e*RS1463e + 37./648.*pow(m11,2)*RC146e*RSe6613
          - 1./1296.*pow(m11,2)*RC146e*RSe6633 + 1./2592.*pow(m11,2)*
         RC146e*RC346e - 73./1296.*pow(m11,2)*RC146e*RSe6611 + 41./1296.
         *pow(m11,2)*pow(RC146e,2) + 1./432.*pi2*me2*RD346e*RSe6613 - 1.
         /864.*pi2*me2*RD346e*RSe6633 - 1./864.*pi2*me2*RSe6611*RD346e
          + 1./432.*pi2*me2*RD146e*RSe6613 - 1./864.*pi2*me2*RD146e*
         RSe6633 - 1./864.*pi2*me2*RD146e*RSe6611 + 7./288.*pi2*pow(
         me2,2) + 1./48.*pi2*pow(me2,2)*RSe6613 + 1./108.*pi2*pow(
         me2,2)*pow(RSe6613,2) );

      F13p6x22R +=  + pi162 * ( 1./72.*pi2*pow(me2,2)*RSe6633 - 1./108.
         *pi2*pow(me2,2)*RSe6633*RSe6613 + 1./432.*pi2*pow(me2,2)*pow(
         RSe6633,2) + 1./72.*pi2*pow(me2,2)*RSe6611 - 1./108.*pi2*pow(
         me2,2)*RSe6611*RSe6613 + 1./216.*pi2*pow(me2,2)*RSe6611*
         RSe6633 + 1./432.*pi2*pow(me2,2)*pow(RSe6611,2) - 1./576.*pi2*
         m66*RN3e*RD34*RD36 - 1./576.*pi2*m66*RN1e*RD14*RD16 - 1./108.*
         pi2*m66*me2 + 1./432.*pi2*m66*me2*RSe6613 + 1./216.*pi2*m66*
         me2*RSe63 + 13./648.*pi2*m66*me2*pow(RSe63,2) - 1./864.*pi2*
         m66*me2*RSe6633 - 1./162.*pi2*m66*me2*RS34e*RSe63 + 1./648.*
         pi2*m66*me2*pow(RS34e,2) + 1./216.*pi2*m66*me2*RSe61 - 13./324.
         *pi2*m66*me2*RSe61*RSe63 + 1./162.*pi2*m66*me2*RSe61*RS34e + 
         13./648.*pi2*m66*me2*pow(RSe61,2) - 1./864.*pi2*m66*me2*
         RSe6611 + 1./162.*pi2*m66*me2*RS14e*RSe63 - 1./324.*pi2*m66*
         me2*RS14e*RS34e - 1./162.*pi2*m66*me2*RS14e*RSe61 + 1./648.*
         pi2*m66*me2*pow(RS14e,2) - 1./36.*pi2*m46*me2*pow(RSe63,2) - 1.
         /108.*pi2*m46*me2*RS36e*RSe63 );

      F13p6x22R +=  + pi162 * ( 1./108.*pi2*m46*me2*RS34e*RSe63 + 1./18.
         *pi2*m46*me2*RSe61*RSe63 + 1./108.*pi2*m46*me2*RSe61*RS36e - 1.
         /108.*pi2*m46*me2*RSe61*RS34e - 1./36.*pi2*m46*me2*pow(
         RSe61,2) + 1./108.*pi2*m46*me2*RS16e*RSe63 - 1./108.*pi2*m46*
         me2*RS16e*RSe61 - 1./108.*pi2*m46*me2*RS14e*RSe63 + 1./108.*
         pi2*m46*me2*RS14e*RSe61 + 1./48.*pi2*m46*m66 - 1./162.*pi2*m46
         *m66*RS36e*RSe63 - 1./324.*pi2*m46*m66*RS34e*RSe63 + 1./324.*
         pi2*m46*m66*RS34e*RS36e + 1./648.*pi2*m46*m66*pow(RS34e,2) + 1.
         /162.*pi2*m46*m66*RSe61*RS36e + 1./324.*pi2*m46*m66*RSe61*
         RS34e + 1./162.*pi2*m46*m66*RS16e*RSe63 - 1./324.*pi2*m46*m66*
         RS16e*RS34e - 1./162.*pi2*m46*m66*RS16e*RSe61 + 1./324.*pi2*
         m46*m66*RS14e*RSe63 - 1./324.*pi2*m46*m66*RS14e*RS36e - 1./324.
         *pi2*m46*m66*RS14e*RS34e - 1./324.*pi2*m46*m66*RS14e*RSe61 + 1.
         /324.*pi2*m46*m66*RS14e*RS16e + 1./648.*pi2*m46*m66*pow(
         RS14e,2) - 1./288.*pi2*m44*RN3e*RD34*RD36 - 1./288.*pi2*m44*
         RN1e*RD14*RD16 );

      F13p6x22R +=  + pi162 * (  - 1./216.*pi2*m44*me2 + 1./216.*pi2*
         m44*me2*RSe6613 - 1./216.*pi2*m44*me2*RSe63 + 19./1296.*pi2*
         m44*me2*pow(RSe63,2) - 1./432.*pi2*m44*me2*RSe6633 + 1./648.*
         pi2*m44*me2*RS36e*RSe63 + 1./1296.*pi2*m44*me2*pow(RS36e,2) - 
         1./216.*pi2*m44*me2*RSe61 - 19./648.*pi2*m44*me2*RSe61*RSe63
          - 1./648.*pi2*m44*me2*RSe61*RS36e + 19./1296.*pi2*m44*me2*
         pow(RSe61,2) - 1./432.*pi2*m44*me2*RSe6611 - 1./648.*pi2*m44*
         me2*RS16e*RSe63 - 1./648.*pi2*m44*me2*RS16e*RS36e + 1./648.*
         pi2*m44*me2*RS16e*RSe61 + 1./1296.*pi2*m44*me2*pow(RS16e,2) + 
         1./48.*pi2*m44*m46 - 1./216.*pi2*m44*m46*pow(RSe63,2) - 1./324.
         *pi2*m44*m46*RS36e*RSe63 + 1./648.*pi2*m44*m46*pow(RS36e,2) + 
         1./324.*pi2*m44*m46*RS34e*RSe63 + 1./324.*pi2*m44*m46*RS34e*
         RS36e + 1./108.*pi2*m44*m46*RSe61*RSe63 + 1./324.*pi2*m44*m46*
         RSe61*RS36e - 1./324.*pi2*m44*m46*RSe61*RS34e - 1./216.*pi2*
         m44*m46*pow(RSe61,2) + 1./324.*pi2*m44*m46*RS16e*RSe63 - 1./
         324.*pi2*m44*m46*RS16e*RS36e );

      F13p6x22R +=  + pi162 * (  - 1./324.*pi2*m44*m46*RS16e*RS34e - 1./
         324.*pi2*m44*m46*RS16e*RSe61 + 1./648.*pi2*m44*m46*pow(
         RS16e,2) - 1./324.*pi2*m44*m46*RS14e*RSe63 - 1./324.*pi2*m44*
         m46*RS14e*RS36e + 1./324.*pi2*m44*m46*RS14e*RSe61 + 1./324.*
         pi2*m44*m46*RS14e*RS16e + 1./32.*pi2*pow(m44,2) + 1./144.*pi2*
         pow(m44,2)*pow(RSe63,2) + 1./72.*pi2*pow(m44,2)*RS36e*RSe63 + 
         1./144.*pi2*pow(m44,2)*pow(RS36e,2) - 1./72.*pi2*pow(m44,2)*
         RSe61*RSe63 - 1./72.*pi2*pow(m44,2)*RSe61*RS36e + 1./144.*pi2*
         pow(m44,2)*pow(RSe61,2) - 1./72.*pi2*pow(m44,2)*RS16e*RSe63 - 
         1./72.*pi2*pow(m44,2)*RS16e*RS36e + 1./72.*pi2*pow(m44,2)*
         RS16e*RSe61 + 1./144.*pi2*pow(m44,2)*pow(RS16e,2) + 7./1728.*
         pi2*m36*RD346e - 1./576.*pi2*m36*RD146e + 1./216.*pi2*m36*me2*
         RSe6613 + 5./432.*pi2*m36*me2*RSe63 - 7./864.*pi2*m36*me2*
         RSe6633 - 5./432.*pi2*m36*me2*RSe61 + 1./288.*pi2*m36*me2*
         RSe6611 + 113./5184.*pi2*m36*m66 + 7./5184.*pi2*m36*m66*
         RS3461e );

      F13p6x22R +=  + pi162 * ( 7./5184.*pi2*m36*m66*RS1463e - 7./2592.
         *pi2*m36*m66*RSe6613 + 29./2592.*pi2*m36*m66*RSe63 + 1./2592.*
         pi2*m36*m66*RSe6633 - 29./5184.*pi2*m36*m66*RS34e - 1./5184.*
         pi2*m36*m66*RC346e - 19./2592.*pi2*m36*m66*RSe61 + 1./2592.*
         pi2*m36*m66*RSe6611 + 19./5184.*pi2*m36*m66*RS14e - 1./5184.*
         pi2*m36*m66*RC146e + 1./96.*pi2*m36*m44 + 7./864.*pi2*m34*
         RD346e - 1./288.*pi2*m34*RD146e + 1./108.*pi2*m34*me2*RSe6613
          - 5./432.*pi2*m34*me2*RSe63 - 7./432.*pi2*m34*me2*RSe6633 + 5.
         /432.*pi2*m34*me2*RSe61 + 1./144.*pi2*m34*me2*RSe6611 + 1./96.
         *pi2*m34*m66 + 35./648.*pi2*m34*m44 + 7./2592.*pi2*m34*m44*
         RS3461e + 7./2592.*pi2*m34*m44*RS1463e - 7./1296.*pi2*m34*m44*
         RSe6613 - 29./2592.*pi2*m34*m44*RSe63 + 1./1296.*pi2*m34*m44*
         RSe6633 - 29./2592.*pi2*m34*m44*RS36e - 1./2592.*pi2*m34*m44*
         RC346e + 19./2592.*pi2*m34*m44*RSe61 + 1./1296.*pi2*m34*m44*
         RSe6611 + 19./2592.*pi2*m34*m44*RS16e - 1./2592.*pi2*m34*m44*
         RC146e );

      F13p6x22R +=  + pi162 * (  - 1./144.*pi2*m33*RD346e - 1./864.*pi2
         *m33*RD346e*RS3461e + 1./1728.*pi2*m33*RD346e*RC346e - 1./864.
         *pi2*m33*RD146e*RS3461e + 1./1728.*pi2*m33*RD146e*RC346e - 1./
         192.*pi2*m33*RN3e*RD34*RD36 + 1./288.*pi2*m33*me2*RSe6613 - 19.
         /1944.*pi2*m33*me2*RSe6613*RS3461e - 1./1944.*pi2*m33*me2*
         RSe6613*RS1463e + 1./972.*pi2*m33*me2*pow(RSe6613,2) - 5./288.
         *pi2*m33*me2*RSe6633 + 55./3888.*pi2*m33*me2*RSe6633*RS3461e
          + 37./3888.*pi2*m33*me2*RSe6633*RS1463e - 19./972.*pi2*m33*
         me2*RSe6633*RSe6613 + 73./3888.*pi2*m33*me2*pow(RSe6633,2) + 
         19./3888.*pi2*m33*me2*RC346e*RSe6613 - 91./7776.*pi2*m33*me2*
         RC346e*RSe6633 + 19./3888.*pi2*m33*me2*RSe6611*RS3461e + 1./
         3888.*pi2*m33*me2*RSe6611*RS1463e - 1./972.*pi2*m33*me2*
         RSe6611*RSe6613 + 1./1944.*pi2*m33*me2*RSe6611*RSe6633 - 19./
         7776.*pi2*m33*me2*RSe6611*RC346e + 1./3888.*pi2*m33*me2*pow(
         RSe6611,2) + 1./3888.*pi2*m33*me2*RC146e*RSe6613 - 1./7776.*
         pi2*m33*me2*RC146e*RSe6633 );

      F13p6x22R +=  + pi162 * (  - 1./7776.*pi2*m33*me2*RC146e*RSe6611
          + 7./864.*pi2*m33*m66*RS3461e + 1./216.*pi2*m33*m66*RS1463e
          - 1./108.*pi2*m33*m66*RSe6613 + 1./216.*pi2*m33*m66*RSe63 + 1.
         /108.*pi2*m33*m66*RSe6633 + 1./648.*pi2*m33*m66*pow(RS36e,2)
          - 1./432.*pi2*m33*m66*RS34e - 1./72.*pi2*m33*m66*RS34e*RSe63
          + 1./648.*pi2*m33*m66*RS34e*RS36e + 5./1296.*pi2*m33*m66*pow(
         RS34e,2) - 7./1728.*pi2*m33*m66*RC346e - 1./216.*pi2*m33*m66*
         RSe61 + 1./72.*pi2*m33*m66*RSe61*RS34e - 1./324.*pi2*m33*m66*
         RS16e*RS36e - 1./648.*pi2*m33*m66*RS16e*RS34e + 1./432.*pi2*
         m33*m66*RS14e - 1./648.*pi2*m33*m66*RS14e*RS36e - 5./648.*pi2*
         m33*m66*RS14e*RS34e - 1./108.*pi2*m33*m46*RS36e*RSe63 - 1./324.
         *pi2*m33*m46*pow(RS36e,2) + 1./108.*pi2*m33*m46*RS34e*RSe63 + 
         1./162.*pi2*m33*m46*RS34e*RS36e - 1./324.*pi2*m33*m46*pow(
         RS34e,2) + 1./108.*pi2*m33*m46*RSe61*RS36e - 1./108.*pi2*m33*
         m46*RSe61*RS34e + 1./324.*pi2*m33*m46*RS16e*RS36e - 1./324.*
         pi2*m33*m46*RS16e*RS34e );

      F13p6x22R +=  + pi162 * (  - 1./324.*pi2*m33*m46*RS14e*RS36e + 1./
         324.*pi2*m33*m46*RS14e*RS34e + 7./432.*pi2*m33*m44*RS3461e + 1.
         /108.*pi2*m33*m44*RS1463e - 1./54.*pi2*m33*m44*RSe6613 - 1./
         216.*pi2*m33*m44*RSe63 + 1./54.*pi2*m33*m44*RSe6633 - 1./216.*
         pi2*m33*m44*RS36e + 1./54.*pi2*m33*m44*RS36e*RSe63 + 7./648.*
         pi2*m33*m44*pow(RS36e,2) - 1./216.*pi2*m33*m44*RS34e*RSe63 + 1.
         /648.*pi2*m33*m44*RS34e*RS36e + 1./648.*pi2*m33*m44*pow(
         RS34e,2) - 7./864.*pi2*m33*m44*RC346e + 1./216.*pi2*m33*m44*
         RSe61 - 1./54.*pi2*m33*m44*RSe61*RS36e + 1./216.*pi2*m33*m44*
         RSe61*RS34e + 1./216.*pi2*m33*m44*RS16e - 7./324.*pi2*m33*m44*
         RS16e*RS36e - 1./648.*pi2*m33*m44*RS16e*RS34e - 1./648.*pi2*
         m33*m44*RS14e*RS36e - 1./324.*pi2*m33*m44*RS14e*RS34e - 1./432.
         *pi2*m33*m36*RS3461e + 1./54.*pi2*m33*m36*RSe63 - 13./864.*pi2
         *m33*m36*RS34e + 7./1728.*pi2*m33*m36*RC346e - 1./54.*pi2*m33*
         m36*RSe61 + 1./108.*pi2*m33*m36*RS14e - 1./216.*pi2*m33*m34*
         RS3461e );

      F13p6x22R +=  + pi162 * (  - 1./54.*pi2*m33*m34*RSe63 - 13./432.*
         pi2*m33*m34*RS36e + 7./864.*pi2*m33*m34*RC346e + 1./54.*pi2*
         m33*m34*RSe61 + 1./54.*pi2*m33*m34*RS16e + 13./576.*pi2*pow(
         m33,2) + 1./576.*pi2*pow(m33,2)*RS3461e + 5./1944.*pi2*pow(
         m33,2)*pow(RS3461e,2) + 1./72.*pi2*pow(m33,2)*RS1463e + 1./
         3888.*pi2*pow(m33,2)*RS1463e*RS3461e - 1./36.*pi2*pow(m33,2)*
         RSe6613 - 1./1944.*pi2*pow(m33,2)*RSe6613*RS3461e + 1./36.*pi2
         *pow(m33,2)*RSe6633 + 1./3888.*pi2*pow(m33,2)*RSe6633*RS3461e
          - 13./864.*pi2*pow(m33,2)*RS36e - 13./1728.*pi2*pow(m33,2)*
         RS34e - 7./576.*pi2*pow(m33,2)*RC346e - 7./972.*pi2*pow(m33,2)
         *RC346e*RS3461e - 37./7776.*pi2*pow(m33,2)*RC346e*RS1463e + 37.
         /3888.*pi2*pow(m33,2)*RC346e*RSe6613 - 73./7776.*pi2*pow(
         m33,2)*RC346e*RSe6633 + 41./7776.*pi2*pow(m33,2)*pow(RC346e,2)
          + 1./3888.*pi2*pow(m33,2)*RSe6611*RS3461e - 1./7776.*pi2*pow(
         m33,2)*RSe6611*RC346e + 1./216.*pi2*pow(m33,2)*RS16e + 1./432.
         *pi2*pow(m33,2)*RS14e );

      F13p6x22R +=  + pi162 * (  - 1./7776.*pi2*pow(m33,2)*RC146e*
         RS3461e + 1./15552.*pi2*pow(m33,2)*RC146e*RC346e - 1./576.*pi2
         *m16*RD346e + 7./1728.*pi2*m16*RD146e + 1./216.*pi2*m16*me2*
         RSe6613 - 5./432.*pi2*m16*me2*RSe63 + 1./288.*pi2*m16*me2*
         RSe6633 + 5./432.*pi2*m16*me2*RSe61 - 7./864.*pi2*m16*me2*
         RSe6611 + 113./5184.*pi2*m16*m66 + 7./5184.*pi2*m16*m66*
         RS3461e + 7./5184.*pi2*m16*m66*RS1463e - 7./2592.*pi2*m16*m66*
         RSe6613 - 19./2592.*pi2*m16*m66*RSe63 + 1./2592.*pi2*m16*m66*
         RSe6633 + 19./5184.*pi2*m16*m66*RS34e - 1./5184.*pi2*m16*m66*
         RC346e + 29./2592.*pi2*m16*m66*RSe61 + 1./2592.*pi2*m16*m66*
         RSe6611 - 29./5184.*pi2*m16*m66*RS14e - 1./5184.*pi2*m16*m66*
         RC146e + 1./96.*pi2*m16*m44 + 5./5184.*pi2*m16*m33 - 5./5184.*
         pi2*m16*m33*RS3461e + 7./5184.*pi2*m16*m33*RS1463e - 7./2592.*
         pi2*m16*m33*RSe6613 + 5./2592.*pi2*m16*m33*RSe63 + 1./2592.*
         pi2*m16*m33*RSe6633 + 25./5184.*pi2*m16*m33*RS34e - 5./2592.*
         pi2*m16*m33*RC346e );

      F13p6x22R +=  + pi162 * ( 5./2592.*pi2*m16*m33*RSe61 + 1./2592.*
         pi2*m16*m33*RSe6611 - 5./5184.*pi2*m16*m33*RS14e - 1./5184.*
         pi2*m16*m33*RC146e - 1./288.*pi2*m14*RD346e + 7./864.*pi2*m14*
         RD146e + 1./108.*pi2*m14*me2*RSe6613 + 5./432.*pi2*m14*me2*
         RSe63 + 1./144.*pi2*m14*me2*RSe6633 - 5./432.*pi2*m14*me2*
         RSe61 - 7./432.*pi2*m14*me2*RSe6611 + 1./96.*pi2*m14*m66 + 35./
         648.*pi2*m14*m44 + 7./2592.*pi2*m14*m44*RS3461e + 7./2592.*pi2
         *m14*m44*RS1463e - 7./1296.*pi2*m14*m44*RSe6613 + 19./2592.*
         pi2*m14*m44*RSe63 + 1./1296.*pi2*m14*m44*RSe6633 + 19./2592.*
         pi2*m14*m44*RS36e - 1./2592.*pi2*m14*m44*RC346e - 29./2592.*
         pi2*m14*m44*RSe61 + 1./1296.*pi2*m14*m44*RSe6611 - 29./2592.*
         pi2*m14*m44*RS16e - 1./2592.*pi2*m14*m44*RC146e + 5./2592.*pi2
         *m14*m33 - 5./2592.*pi2*m14*m33*RS3461e + 7./2592.*pi2*m14*m33
         *RS1463e - 7./1296.*pi2*m14*m33*RSe6613 - 5./2592.*pi2*m14*m33
         *RSe63 + 1./1296.*pi2*m14*m33*RSe6633 + 25./2592.*pi2*m14*m33*
         RS36e );

      F13p6x22R +=  + pi162 * (  - 5./1296.*pi2*m14*m33*RC346e - 5./
         2592.*pi2*m14*m33*RSe61 + 1./1296.*pi2*m14*m33*RSe6611 - 5./
         2592.*pi2*m14*m33*RS16e - 1./2592.*pi2*m14*m33*RC146e - 1./
         3888.*pi2*m13*RD346e*RS3461e - 1./3888.*pi2*m13*RD346e*RS1463e
          + 1./1944.*pi2*m13*RD346e*RSe6613 - 1./3888.*pi2*m13*RD346e*
         RSe6633 + 1./7776.*pi2*m13*RD346e*RC346e - 1./3888.*pi2*m13*
         RSe6611*RD346e + 1./7776.*pi2*m13*RC146e*RD346e - 1./3888.*pi2
         *m13*RD146e*RS3461e - 1./3888.*pi2*m13*RD146e*RS1463e + 1./
         1944.*pi2*m13*RD146e*RSe6613 - 1./3888.*pi2*m13*RD146e*RSe6633
          + 1./7776.*pi2*m13*RD146e*RC346e - 1./3888.*pi2*m13*RD146e*
         RSe6611 + 1./7776.*pi2*m13*RD146e*RC146e - 1./972.*pi2*m13*me2
         *RSe6613*RS3461e - 1./972.*pi2*m13*me2*RSe6613*RS1463e + 1./
         486.*pi2*m13*me2*pow(RSe6613,2) + 1./1944.*pi2*m13*me2*RSe6633
         *RS3461e + 1./1944.*pi2*m13*me2*RSe6633*RS1463e - 1./486.*pi2*
         m13*me2*RSe6633*RSe6613 + 1./1944.*pi2*m13*me2*pow(RSe6633,2)
          + 1./1944.*pi2*m13*me2*RC346e*RSe6613 );

      F13p6x22R +=  + pi162 * (  - 1./3888.*pi2*m13*me2*RC346e*RSe6633
          + 1./1944.*pi2*m13*me2*RSe6611*RS3461e + 1./1944.*pi2*m13*me2
         *RSe6611*RS1463e - 1./486.*pi2*m13*me2*RSe6611*RSe6613 + 1./
         972.*pi2*m13*me2*RSe6611*RSe6633 - 1./3888.*pi2*m13*me2*
         RSe6611*RC346e + 1./1944.*pi2*m13*me2*pow(RSe6611,2) + 1./1944.
         *pi2*m13*me2*RC146e*RSe6613 - 1./3888.*pi2*m13*me2*RC146e*
         RSe6633 - 1./3888.*pi2*m13*me2*RC146e*RSe6611 + 1./96.*pi2*m13
         *m33 + 1./1944.*pi2*m13*m33*pow(RS3461e,2) + 1./1944.*pi2*m13*
         m33*RS1463e*RS3461e - 1./972.*pi2*m13*m33*RSe6613*RS3461e + 1./
         1944.*pi2*m13*m33*RSe6633*RS3461e - 1./1944.*pi2*m13*m33*
         RC346e*RS3461e - 1./3888.*pi2*m13*m33*RC346e*RS1463e + 1./1944.
         *pi2*m13*m33*RC346e*RSe6613 - 1./3888.*pi2*m13*m33*RC346e*
         RSe6633 + 1./7776.*pi2*m13*m33*pow(RC346e,2) + 1./1944.*pi2*
         m13*m33*RSe6611*RS3461e - 1./3888.*pi2*m13*m33*RSe6611*RC346e
          - 1./3888.*pi2*m13*m33*RC146e*RS3461e + 1./7776.*pi2*m13*m33*
         RC146e*RC346e );

      F13p6x22R +=  + pi162 * (  - 1./864.*pi2*m11*RD346e*RS1463e + 1./
         1728.*pi2*m11*RC146e*RD346e - 1./144.*pi2*m11*RD146e - 1./864.
         *pi2*m11*RD146e*RS1463e + 1./1728.*pi2*m11*RD146e*RC146e - 1./
         192.*pi2*m11*RN1e*RD14*RD16 + 1./288.*pi2*m11*me2*RSe6613 - 1./
         1944.*pi2*m11*me2*RSe6613*RS3461e - 19./1944.*pi2*m11*me2*
         RSe6613*RS1463e + 1./972.*pi2*m11*me2*pow(RSe6613,2) + 1./3888.
         *pi2*m11*me2*RSe6633*RS3461e + 19./3888.*pi2*m11*me2*RSe6633*
         RS1463e - 1./972.*pi2*m11*me2*RSe6633*RSe6613 + 1./3888.*pi2*
         m11*me2*pow(RSe6633,2) + 1./3888.*pi2*m11*me2*RC346e*RSe6613
          - 1./7776.*pi2*m11*me2*RC346e*RSe6633 - 5./288.*pi2*m11*me2*
         RSe6611 + 37./3888.*pi2*m11*me2*RSe6611*RS3461e + 55./3888.*
         pi2*m11*me2*RSe6611*RS1463e - 19./972.*pi2*m11*me2*RSe6611*
         RSe6613 + 1./1944.*pi2*m11*me2*RSe6611*RSe6633 - 1./7776.*pi2*
         m11*me2*RSe6611*RC346e + 73./3888.*pi2*m11*me2*pow(RSe6611,2)
          + 19./3888.*pi2*m11*me2*RC146e*RSe6613 - 19./7776.*pi2*m11*
         me2*RC146e*RSe6633 );

      F13p6x22R +=  + pi162 * (  - 91./7776.*pi2*m11*me2*RC146e*RSe6611
          + 1./288.*pi2*m11*m66*RS1463e - 1./108.*pi2*m11*m66*RSe6613
          - 1./216.*pi2*m11*m66*RSe63 + 1./216.*pi2*m11*m66*RSe61 + 1./
         108.*pi2*m11*m66*RSe6611 + 1./648.*pi2*m11*m66*pow(RS16e,2) - 
         1./216.*pi2*m11*m66*RS14e + 1./72.*pi2*m11*m66*RS14e*RSe63 - 1.
         /72.*pi2*m11*m66*RS14e*RSe61 + 1./648.*pi2*m11*m66*RS14e*RS16e
          + 5./1296.*pi2*m11*m66*pow(RS14e,2) - 7./1728.*pi2*m11*m66*
         RC146e + 1./108.*pi2*m11*m46*RS16e*RSe63 + 1./324.*pi2*m11*m46
         *RS16e*RS36e - 1./324.*pi2*m11*m46*RS16e*RS34e - 1./108.*pi2*
         m11*m46*RS16e*RSe61 - 1./324.*pi2*m11*m46*pow(RS16e,2) - 1./
         108.*pi2*m11*m46*RS14e*RSe63 - 1./324.*pi2*m11*m46*RS14e*RS36e
          + 1./324.*pi2*m11*m46*RS14e*RS34e + 1./108.*pi2*m11*m46*RS14e
         *RSe61 + 1./162.*pi2*m11*m46*RS14e*RS16e - 1./324.*pi2*m11*m46
         *pow(RS14e,2) + 1./144.*pi2*m11*m44*RS1463e - 1./54.*pi2*m11*
         m44*RSe6613 + 1./216.*pi2*m11*m44*RSe63 - 1./216.*pi2*m11*m44*
         RSe61 );

      F13p6x22R +=  + pi162 * ( 1./54.*pi2*m11*m44*RSe6611 - 1./108.*
         pi2*m11*m44*RS16e - 1./54.*pi2*m11*m44*RS16e*RSe63 + 1./54.*
         pi2*m11*m44*RS16e*RSe61 + 7./648.*pi2*m11*m44*pow(RS16e,2) + 1.
         /216.*pi2*m11*m44*RS14e*RSe63 - 1./216.*pi2*m11*m44*RS14e*
         RSe61 + 1./648.*pi2*m11*m44*RS14e*RS16e + 1./648.*pi2*m11*m44*
         pow(RS14e,2) - 7./864.*pi2*m11*m44*RC146e + 5./5184.*pi2*m11*
         m36 + 7./5184.*pi2*m11*m36*RS3461e - 5./5184.*pi2*m11*m36*
         RS1463e - 7./2592.*pi2*m11*m36*RSe6613 + 5./2592.*pi2*m11*m36*
         RSe63 + 1./2592.*pi2*m11*m36*RSe6633 - 5./5184.*pi2*m11*m36*
         RS34e - 1./5184.*pi2*m11*m36*RC346e + 5./2592.*pi2*m11*m36*
         RSe61 + 1./2592.*pi2*m11*m36*RSe6611 + 25./5184.*pi2*m11*m36*
         RS14e - 5./2592.*pi2*m11*m36*RC146e + 5./2592.*pi2*m11*m34 + 7.
         /2592.*pi2*m11*m34*RS3461e - 5./2592.*pi2*m11*m34*RS1463e - 7./
         1296.*pi2*m11*m34*RSe6613 - 5./2592.*pi2*m11*m34*RSe63 + 1./
         1296.*pi2*m11*m34*RSe6633 - 5./2592.*pi2*m11*m34*RS36e - 1./
         2592.*pi2*m11*m34*RC346e );

      F13p6x22R +=  + pi162 * (  - 5./2592.*pi2*m11*m34*RSe61 + 1./1296.
         *pi2*m11*m34*RSe6611 + 25./2592.*pi2*m11*m34*RS16e - 5./1296.*
         pi2*m11*m34*RC146e - 1./96.*pi2*m11*m33 + 7./576.*pi2*m11*m33*
         RS3461e + 1./3888.*pi2*m11*m33*pow(RS3461e,2) - 1./576.*pi2*
         m11*m33*RS1463e + 5./972.*pi2*m11*m33*RS1463e*RS3461e + 1./
         3888.*pi2*m11*m33*pow(RS1463e,2) - 1./1944.*pi2*m11*m33*
         RSe6613*RS3461e - 1./1944.*pi2*m11*m33*RSe6613*RS1463e + 1./
         3888.*pi2*m11*m33*RSe6633*RS3461e + 1./3888.*pi2*m11*m33*
         RSe6633*RS1463e + 5./864.*pi2*m11*m33*RS36e + 5./1728.*pi2*m11
         *m33*RS34e - 1./3888.*pi2*m11*m33*RC346e*RS3461e - 5./1944.*
         pi2*m11*m33*RC346e*RS1463e + 1./3888.*pi2*m11*m33*RC346e*
         RSe6613 - 1./7776.*pi2*m11*m33*RC346e*RSe6633 + 1./15552.*pi2*
         m11*m33*pow(RC346e,2) + 1./3888.*pi2*m11*m33*RSe6611*RS3461e
          + 1./3888.*pi2*m11*m33*RSe6611*RS1463e - 1./7776.*pi2*m11*m33
         *RSe6611*RC346e + 1./864.*pi2*m11*m33*RS16e + 1./1728.*pi2*m11
         *m33*RS14e );

      F13p6x22R +=  + pi162 * (  - 5./1944.*pi2*m11*m33*RC146e*RS3461e
          - 1./3888.*pi2*m11*m33*RC146e*RS1463e + 1./3888.*pi2*m11*m33*
         RC146e*RSe6613 - 1./7776.*pi2*m11*m33*RC146e*RSe6633 + 5./3888.
         *pi2*m11*m33*RC146e*RC346e - 1./7776.*pi2*m11*m33*RC146e*
         RSe6611 + 1./15552.*pi2*m11*m33*pow(RC146e,2) - 1./432.*pi2*
         m11*m16*RS1463e - 1./54.*pi2*m11*m16*RSe63 + 1./108.*pi2*m11*
         m16*RS34e + 1./54.*pi2*m11*m16*RSe61 - 13./864.*pi2*m11*m16*
         RS14e + 7./1728.*pi2*m11*m16*RC146e - 1./216.*pi2*m11*m14*
         RS1463e + 1./54.*pi2*m11*m14*RSe63 + 1./54.*pi2*m11*m14*RS36e
          - 1./54.*pi2*m11*m14*RSe61 - 13./432.*pi2*m11*m14*RS16e + 7./
         864.*pi2*m11*m14*RC146e + 1./96.*pi2*m11*m13 + 1./1944.*pi2*
         m11*m13*RS1463e*RS3461e + 1./1944.*pi2*m11*m13*pow(RS1463e,2)
          - 1./972.*pi2*m11*m13*RSe6613*RS1463e + 1./1944.*pi2*m11*m13*
         RSe6633*RS1463e - 1./3888.*pi2*m11*m13*RC346e*RS1463e + 1./
         1944.*pi2*m11*m13*RSe6611*RS1463e - 1./3888.*pi2*m11*m13*
         RC146e*RS3461e );

      F13p6x22R +=  + pi162 * (  - 1./1944.*pi2*m11*m13*RC146e*RS1463e
          + 1./1944.*pi2*m11*m13*RC146e*RSe6613 - 1./3888.*pi2*m11*m13*
         RC146e*RSe6633 + 1./7776.*pi2*m11*m13*RC146e*RC346e - 1./3888.
         *pi2*m11*m13*RC146e*RSe6611 + 1./7776.*pi2*m11*m13*pow(
         RC146e,2) + 17./576.*pi2*pow(m11,2) + 1./576.*pi2*pow(m11,2)*
         RS1463e + 1./3888.*pi2*pow(m11,2)*RS1463e*RS3461e + 5./1944.*
         pi2*pow(m11,2)*pow(RS1463e,2) - 1./36.*pi2*pow(m11,2)*RSe6613
          - 1./1944.*pi2*pow(m11,2)*RSe6613*RS1463e + 1./3888.*pi2*pow(
         m11,2)*RSe6633*RS1463e - 1./7776.*pi2*pow(m11,2)*RC346e*
         RS1463e + 1./36.*pi2*pow(m11,2)*RSe6611 + 1./3888.*pi2*pow(
         m11,2)*RSe6611*RS1463e - 13./864.*pi2*pow(m11,2)*RS16e - 13./
         1728.*pi2*pow(m11,2)*RS14e - 7./576.*pi2*pow(m11,2)*RC146e - 
         37./7776.*pi2*pow(m11,2)*RC146e*RS3461e - 7./972.*pi2*pow(
         m11,2)*RC146e*RS1463e + 37./3888.*pi2*pow(m11,2)*RC146e*
         RSe6613 - 1./7776.*pi2*pow(m11,2)*RC146e*RSe6633 + 1./15552.*
         pi2*pow(m11,2)*RC146e*RC346e );

      F13p6x22R +=  + pi162 * (  - 73./7776.*pi2*pow(m11,2)*RC146e*
         RSe6611 + 41./7776.*pi2*pow(m11,2)*pow(RC146e,2) );

      F13p6x22R +=  + Ab(m11,mu2)*pi16 * ( 1./72.*RD346e*RS1463e - 1./
         144.*RC146e*RD346e - 1./48.*RD146e + 1./72.*RD146e*RS1463e - 1.
         /144.*RD146e*RC146e + 1./18.*me2*RSe6613*RS1463e - 1./36.*me2*
         RSe6633*RS1463e - 1./36.*me2*RSe6611*RS1463e - 1./36.*me2*
         RC146e*RSe6613 + 1./72.*me2*RC146e*RSe6633 + 1./72.*me2*RC146e
         *RSe6611 - 1./24.*m66*RS1463e + 1./48.*m66*RC146e - 1./12.*m44
         *RS1463e + 1./24.*m44*RC146e - 1./16.*m33*RS1463e - 1./24.*m33
         *RS1463e*RS3461e - 1./72.*m33*pow(RS1463e,2) + 1./36.*m33*
         RSe6613*RS1463e - 1./72.*m33*RSe6633*RS1463e + 1./48.*m33*
         RC346e*RS1463e - 1./72.*m33*RSe6611*RS1463e + 1./32.*m33*
         RC146e + 1./48.*m33*RC146e*RS3461e + 1./72.*m33*RC146e*RS1463e
          - 1./72.*m33*RC146e*RSe6613 + 1./144.*m33*RC146e*RSe6633 - 1./
         96.*m33*RC146e*RC346e + 1./144.*m33*RC146e*RSe6611 - 1./288.*
         m33*pow(RC146e,2) + 1./16.*m11 - 1./48.*m11*RS1463e - 1./72.*
         m11*RS1463e*RS3461e - 1./24.*m11*pow(RS1463e,2) + 1./36.*m11*
         RSe6613*RS1463e );

      F13p6x22R +=  + Ab(m11,mu2)*pi16 * (  - 1./72.*m11*RSe6633*
         RS1463e + 1./144.*m11*RC346e*RS1463e - 1./72.*m11*RSe6611*
         RS1463e - 1./24.*m11*RS16e - 1./48.*m11*RS14e + 1./96.*m11*
         RC146e + 1./144.*m11*RC146e*RS3461e + 1./24.*m11*RC146e*
         RS1463e - 1./72.*m11*RC146e*RSe6613 + 1./144.*m11*RC146e*
         RSe6633 - 1./288.*m11*RC146e*RC346e + 1./144.*m11*RC146e*
         RSe6611 - 1./96.*m11*pow(RC146e,2) );

      F13p6x22R +=  + pow(Ab(m11,mu2),2) * ( 23./96. + 1./96.*pow(
         m11,-2)*m66*RN1e*RD14*RD16 + 1./48.*pow(m11,-2)*m44*RN1e*RD14*
         RD16 - 1./144.*pow(m11,-1)*RD346e*RS1463e + 1./288.*pow(
         m11,-1)*RC146e*RD346e - 1./24.*pow(m11,-1)*RD146e - 1./144.*
         pow(m11,-1)*RD146e*RS1463e + 1./288.*pow(m11,-1)*RD146e*RC146e
          + 1./32.*pow(m11,-1)*RN1e*RD14*RD16 - 1./36.*pow(m11,-1)*me2*
         RSe6613*RS1463e + 1./72.*pow(m11,-1)*me2*RSe6633*RS1463e + 1./
         72.*pow(m11,-1)*me2*RSe6611*RS1463e + 1./72.*pow(m11,-1)*me2*
         RC146e*RSe6613 - 1./144.*pow(m11,-1)*me2*RC146e*RSe6633 - 1./
         144.*pow(m11,-1)*me2*RC146e*RSe6611 + 1./48.*pow(m11,-1)*m66*
         RS1463e - 1./18.*pow(m11,-1)*m66*RSe6613 - 1./36.*pow(m11,-1)*
         m66*RSe63 + 1./36.*pow(m11,-1)*m66*RSe61 + 1./18.*pow(m11,-1)*
         m66*RSe6611 + 1./108.*pow(m11,-1)*m66*pow(RS16e,2) - 1./36.*
         pow(m11,-1)*m66*RS14e + 1./12.*pow(m11,-1)*m66*RS14e*RSe63 - 1.
         /12.*pow(m11,-1)*m66*RS14e*RSe61 + 1./108.*pow(m11,-1)*m66*
         RS14e*RS16e );

      F13p6x22R +=  + pow(Ab(m11,mu2),2) * ( 5./216.*pow(m11,-1)*m66*
         pow(RS14e,2) - 7./288.*pow(m11,-1)*m66*RC146e + 1./24.*pow(
         m11,-1)*m44*RS1463e - 1./9.*pow(m11,-1)*m44*RSe6613 + 1./36.*
         pow(m11,-1)*m44*RSe63 - 1./36.*pow(m11,-1)*m44*RSe61 + 1./9.*
         pow(m11,-1)*m44*RSe6611 - 1./18.*pow(m11,-1)*m44*RS16e - 1./9.
         *pow(m11,-1)*m44*RS16e*RSe63 + 1./9.*pow(m11,-1)*m44*RS16e*
         RSe61 + 7./108.*pow(m11,-1)*m44*pow(RS16e,2) + 1./36.*pow(
         m11,-1)*m44*RS14e*RSe63 - 1./36.*pow(m11,-1)*m44*RS14e*RSe61
          + 1./108.*pow(m11,-1)*m44*RS14e*RS16e + 1./108.*pow(m11,-1)*
         m44*pow(RS14e,2) - 7./144.*pow(m11,-1)*m44*RC146e - 1./96.*
         pow(m11,-1)*m33 - 1./96.*pow(m11,-1)*m33*RS1463e + 5./324.*
         pow(m11,-1)*m33*RS1463e*RS3461e + 1./648.*pow(m11,-1)*m33*pow(
         RS1463e,2) - 1./324.*pow(m11,-1)*m33*RSe6613*RS1463e + 1./648.
         *pow(m11,-1)*m33*RSe6633*RS1463e - 5./648.*pow(m11,-1)*m33*
         RC346e*RS1463e + 1./648.*pow(m11,-1)*m33*RSe6611*RS1463e + 1./
         144.*pow(m11,-1)*m33*RS16e );

      F13p6x22R +=  + pow(Ab(m11,mu2),2) * ( 1./288.*pow(m11,-1)*m33*
         RS14e - 5./648.*pow(m11,-1)*m33*RC146e*RS3461e - 1./648.*pow(
         m11,-1)*m33*RC146e*RS1463e + 1./648.*pow(m11,-1)*m33*RC146e*
         RSe6613 - 1./1296.*pow(m11,-1)*m33*RC146e*RSe6633 + 5./1296.*
         pow(m11,-1)*m33*RC146e*RC346e - 1./1296.*pow(m11,-1)*m33*
         RC146e*RSe6611 + 1./2592.*pow(m11,-1)*m33*pow(RC146e,2) + 1./
         96.*RS1463e + 1./648.*RS1463e*RS3461e + 19./648.*pow(
         RS1463e,2) - 1./6.*RSe6613 - 1./324.*RSe6613*RS1463e + 1./648.
         *RSe6633*RS1463e - 1./1296.*RC346e*RS1463e + 1./6.*RSe6611 + 1.
         /648.*RSe6611*RS1463e - 13./144.*RS16e - 13./288.*RS14e - 7./
         96.*RC146e - 37./1296.*RC146e*RS3461e - 37./648.*RC146e*
         RS1463e + 37./648.*RC146e*RSe6613 - 1./1296.*RC146e*RSe6633 + 
         1./2592.*RC146e*RC346e - 73./1296.*RC146e*RSe6611 + 91./2592.*
         pow(RC146e,2) + 1./9.*m11*RN1e*RC146e*RSe6613 - 1./9.*m11*RN1e
         *RC146e*RSe6611 - 1./18.*m11*RN13*RC146e*RS3461e );

      F13p6x22R +=  + Ab(m11,mu2)*Ab(m13,mu2) * (  - 1./24. + 1./162.*
         RS1463e*RS3461e + 1./162.*pow(RS1463e,2) - 1./81.*RSe6613*
         RS1463e + 1./162.*RSe6633*RS1463e - 1./324.*RC346e*RS1463e + 1.
         /162.*RSe6611*RS1463e - 1./324.*RC146e*RS3461e - 1./162.*
         RC146e*RS1463e + 1./162.*RC146e*RSe6613 - 1./324.*RC146e*
         RSe6633 + 1./648.*RC146e*RC346e - 1./324.*RC146e*RSe6611 + 1./
         648.*pow(RC146e,2) - 1./6.*m33*RN13 - 1./3.*m11*RN13 );

      F13p6x22R +=  + Ab(m11,mu2)*Ab(m14,mu2) * (  - 1./18.*RS1463e + 2.
         /9.*RSe6613 + 1./9.*RSe63 - 1./9.*RSe61 - 2./9.*RSe6611 - 5./
         36.*RS16e + 7./72.*RC146e + 2./9.*me2*RN1e*RSe6613 + 1./9.*me2
         *RN1e*RSe63 - 1./9.*me2*RN1e*RSe61 - 2./9.*me2*RN1e*RSe6611 + 
         1./9.*m44*RN1e*RSe63 - 1./9.*m44*RN1e*RSe61 + 1./9.*m44*RN13*
         RS36e - 2./9.*m11*RN1e*RSe6613 + 1./9.*m11*RN1e*RSe63 - 1./9.*
         m11*RN1e*RSe61 + 2./9.*m11*RN1e*RSe6611 + 2./9.*m11*RN13*RS36e
          );

      F13p6x22R +=  + Ab(m11,mu2)*Ab(m16,mu2) * (  - 1./36.*RS1463e + 1.
         /9.*RSe6613 - 1./9.*RSe63 + 1./9.*RSe61 - 1./9.*RSe6611 - 5./
         72.*RS14e + 7./144.*RC146e + 1./9.*me2*RN1e*RSe6613 - 1./9.*
         me2*RN1e*RSe63 + 1./9.*me2*RN1e*RSe61 - 1./9.*me2*RN1e*RSe6611
          - 1./9.*m66*RN1e*RSe63 + 1./9.*m66*RN1e*RSe61 + 1./18.*m66*
         RN13*RS34e - 1./9.*m11*RN1e*RSe6613 - 1./9.*m11*RN1e*RSe63 + 1.
         /9.*m11*RN1e*RSe61 + 1./9.*m11*RN1e*RSe6611 + 1./9.*m11*RN13*
         RS34e );

      F13p6x22R +=  + Ab(m11,mu2)*Ab(m33,mu2) * ( 1./36.*RS1463e*
         RS3461e - 1./72.*RC346e*RS1463e - 1./72.*RC146e*RS3461e + 1./
         144.*RC146e*RC346e - 1./18.*m33*RN13*RC346e*RS1463e + 1./18.*
         m11*RN13*RC146e*RS3461e );

      F13p6x22R +=  + Ab(m11,mu2)*Ab(m34,mu2) * ( 1./18.*RS1463e + 1./
         36.*RS16e - 1./24.*RC146e + 1./9.*m44*RN13*RS16e + 1./9.*m33*
         RN13*RS1463e + 1./9.*m33*RN13*RS16e - 1./9.*m11*RN13*RS1463e
          + 1./9.*m11*RN13*RS16e );

      F13p6x22R +=  + Ab(m11,mu2)*Ab(m36,mu2) * ( 1./36.*RS1463e + 1./
         72.*RS14e - 1./48.*RC146e + 1./18.*m66*RN13*RS14e + 1./18.*m33
         *RN13*RS1463e + 1./18.*m33*RN13*RS14e - 1./18.*m11*RN13*
         RS1463e + 1./18.*m11*RN13*RS14e );

      F13p6x22R +=  + Ab(m11,mu2)*Ab(m44,mu2) * (  - 1./6.*m44*RN1e*
         RS16e*RSe63 + 1./6.*m44*RN1e*RS16e*RSe61 - 1./6.*m44*RN13*
         RS16e*RS36e );

      F13p6x22R +=  + Ab(m11,mu2)*Ab(m46,mu2) * ( 1./9.*RS16e*RSe63 - 1.
         /9.*RS16e*RSe61 - 1./54.*pow(RS16e,2) - 1./9.*RS14e*RSe63 + 1./
         9.*RS14e*RSe61 + 1./27.*RS14e*RS16e - 1./54.*pow(RS14e,2) + 1./
         9.*me2*RN1e*RS16e*RSe63 - 1./9.*me2*RN1e*RS16e*RSe61 - 1./9.*
         me2*RN1e*RS14e*RSe63 + 1./9.*me2*RN1e*RS14e*RSe61 + 2./27.*m66
         *RN1e*RS16e*RSe63 - 2./27.*m66*RN1e*RS16e*RSe61 + 1./27.*m66*
         RN1e*RS14e*RSe63 - 1./27.*m66*RN1e*RS14e*RSe61 - 1./27.*m66*
         RN13*RS16e*RS34e - 1./27.*m66*RN13*RS14e*RS36e - 1./27.*m66*
         RN13*RS14e*RS34e + 1./27.*m44*RN1e*RS16e*RSe63 - 1./27.*m44*
         RN1e*RS16e*RSe61 - 1./27.*m44*RN1e*RS14e*RSe63 + 1./27.*m44*
         RN1e*RS14e*RSe61 - 1./27.*m44*RN13*RS16e*RS36e - 1./27.*m44*
         RN13*RS16e*RS34e - 1./27.*m44*RN13*RS14e*RS36e + 1./27.*m11*
         RN13*RS16e*RS36e - 1./27.*m11*RN13*RS16e*RS34e - 1./27.*m11*
         RN13*RS14e*RS36e + 1./27.*m11*RN13*RS14e*RS34e );

      F13p6x22R +=  + Ab(m11,mu2)*Ab(me2,mu2) * (  - 1./18.*RSe6613*
         RS1463e + 1./36.*RSe6633*RS1463e + 1./36.*RSe6611*RS1463e + 1./
         36.*RC146e*RSe6613 - 1./72.*RC146e*RSe6633 - 1./72.*RC146e*
         RSe6611 + 2./27.*m66*RN1e*RS14e*RSe63 - 2./27.*m66*RN1e*RS14e*
         RSe61 - 1./27.*m66*RN13*RS14e*RS34e - 1./54.*m44*RN1e*RS16e*
         RSe63 + 1./54.*m44*RN1e*RS16e*RSe61 - 1./54.*m44*RN13*RS16e*
         RS36e + 1./9.*m33*RN13*RSe6633*RS1463e - 2./9.*m11*RN1e*
         RSe6611*RSe6613 + 2./9.*m11*RN1e*pow(RSe6611,2) - 1./9.*m11*
         RN1e*RC146e*RSe6613 + 1./9.*m11*RN1e*RC146e*RSe6611 + 1./9.*
         m11*RN13*RSe6611*RS3461e );

      F13p6x22R +=  + Ab(m11,mu2)*Bb(m11,mu2) * (  - 1./48.*pow(m11,-1)
         *m66*RN1e*RD14*RD16 - 1./24.*pow(m11,-1)*m44*RN1e*RD14*RD16 - 
         1./72.*RD146e*RS1463e + 1./144.*RD146e*RC146e - 1./16.*RN1e*
         RD14*RD16 + 1./4.*m11 - 1./18.*m11*RC146e*RS1463e + 1./18.*m11
         *pow(RC146e,2) + 1./9.*m11*RN1e*RD146e*RSe6613 - 1./9.*m11*
         RN1e*RD146e*RSe6611 - 1./18.*m11*RN13*RD146e*RS3461e );

      F13p6x22R +=  + Ab(m11,mu2)*Bb(m14,mu2) * ( 1./36.*RD146e - 1./18.
         *m44*RS16e - 1./9.*m11*RS16e );

      F13p6x22R +=  + Ab(m11,mu2)*Bb(m16,mu2) * ( 1./72.*RD146e - 1./36.
         *m66*RS14e - 1./18.*m11*RS14e );

      F13p6x22R +=  + Ab(m11,mu2)*Bb(m33,mu2) * (  - 1./72.*RD346e*
         RS1463e + 1./144.*RC146e*RD346e - 1./18.*m33*RN13*RD346e*
         RS1463e );

      F13p6x22R +=  + Ab(m11,mu2)*Cb(m11,mu2) * ( 1./18.*m11*RD146e*
         RC146e );

      F13p6x22R +=  + Ab(m13,mu2)*pi16 * (  - 1./144.*m33*pow(
         RS3461e,2) - 1./72.*m33*RS1463e*RS3461e - 1./144.*m33*pow(
         RS1463e,2) + 1./36.*m33*RSe6613*RS3461e + 1./36.*m33*RSe6613*
         RS1463e - 1./36.*m33*pow(RSe6613,2) - 1./72.*m33*RSe6633*
         RS3461e - 1./72.*m33*RSe6633*RS1463e + 1./36.*m33*RSe6633*
         RSe6613 - 1./144.*m33*pow(RSe6633,2) + 1./144.*m33*RC346e*
         RS3461e + 1./144.*m33*RC346e*RS1463e - 1./72.*m33*RC346e*
         RSe6613 + 1./144.*m33*RC346e*RSe6633 - 1./576.*m33*pow(
         RC346e,2) - 1./72.*m33*RSe6611*RS3461e - 1./72.*m33*RSe6611*
         RS1463e + 1./36.*m33*RSe6611*RSe6613 - 1./72.*m33*RSe6611*
         RSe6633 + 1./144.*m33*RSe6611*RC346e - 1./144.*m33*pow(
         RSe6611,2) + 1./144.*m33*RC146e*RS3461e + 1./144.*m33*RC146e*
         RS1463e - 1./72.*m33*RC146e*RSe6613 + 1./144.*m33*RC146e*
         RSe6633 - 1./288.*m33*RC146e*RC346e + 1./144.*m33*RC146e*
         RSe6611 - 1./576.*m33*pow(RC146e,2) - 1./144.*m11*pow(
         RS3461e,2) );

      F13p6x22R +=  + Ab(m13,mu2)*pi16 * (  - 1./72.*m11*RS1463e*
         RS3461e - 1./144.*m11*pow(RS1463e,2) + 1./36.*m11*RSe6613*
         RS3461e + 1./36.*m11*RSe6613*RS1463e - 1./36.*m11*pow(
         RSe6613,2) - 1./72.*m11*RSe6633*RS3461e - 1./72.*m11*RSe6633*
         RS1463e + 1./36.*m11*RSe6633*RSe6613 - 1./144.*m11*pow(
         RSe6633,2) + 1./144.*m11*RC346e*RS3461e + 1./144.*m11*RC346e*
         RS1463e - 1./72.*m11*RC346e*RSe6613 + 1./144.*m11*RC346e*
         RSe6633 - 1./576.*m11*pow(RC346e,2) - 1./72.*m11*RSe6611*
         RS3461e - 1./72.*m11*RSe6611*RS1463e + 1./36.*m11*RSe6611*
         RSe6613 - 1./72.*m11*RSe6611*RSe6633 + 1./144.*m11*RSe6611*
         RC346e - 1./144.*m11*pow(RSe6611,2) + 1./144.*m11*RC146e*
         RS3461e + 1./144.*m11*RC146e*RS1463e - 1./72.*m11*RC146e*
         RSe6613 + 1./144.*m11*RC146e*RSe6633 - 1./288.*m11*RC146e*
         RC346e + 1./144.*m11*RC146e*RSe6611 - 1./576.*m11*pow(
         RC146e,2) );

      F13p6x22R +=  + pow(Ab(m13,mu2),2) * ( 1./8. - 1./648.*pow(
         m13,-1)*RD346e*RS3461e - 1./648.*pow(m13,-1)*RD346e*RS1463e + 
         1./324.*pow(m13,-1)*RD346e*RSe6613 - 1./648.*pow(m13,-1)*
         RD346e*RSe6633 + 1./1296.*pow(m13,-1)*RD346e*RC346e - 1./648.*
         pow(m13,-1)*RSe6611*RD346e + 1./1296.*pow(m13,-1)*RC146e*
         RD346e - 1./648.*pow(m13,-1)*RD146e*RS3461e - 1./648.*pow(
         m13,-1)*RD146e*RS1463e + 1./324.*pow(m13,-1)*RD146e*RSe6613 - 
         1./648.*pow(m13,-1)*RD146e*RSe6633 + 1./1296.*pow(m13,-1)*
         RD146e*RC346e - 1./648.*pow(m13,-1)*RD146e*RSe6611 + 1./1296.*
         pow(m13,-1)*RD146e*RC146e - 1./162.*pow(m13,-1)*me2*RSe6613*
         RS3461e - 1./162.*pow(m13,-1)*me2*RSe6613*RS1463e + 1./81.*
         pow(m13,-1)*me2*pow(RSe6613,2) + 1./324.*pow(m13,-1)*me2*
         RSe6633*RS3461e + 1./324.*pow(m13,-1)*me2*RSe6633*RS1463e - 1./
         81.*pow(m13,-1)*me2*RSe6633*RSe6613 + 1./324.*pow(m13,-1)*me2*
         pow(RSe6633,2) + 1./324.*pow(m13,-1)*me2*RC346e*RSe6613 - 1./
         648.*pow(m13,-1)*me2*RC346e*RSe6633 );

      F13p6x22R +=  + pow(Ab(m13,mu2),2) * ( 1./324.*pow(m13,-1)*me2*
         RSe6611*RS3461e + 1./324.*pow(m13,-1)*me2*RSe6611*RS1463e - 1./
         81.*pow(m13,-1)*me2*RSe6611*RSe6613 + 1./162.*pow(m13,-1)*me2*
         RSe6611*RSe6633 - 1./648.*pow(m13,-1)*me2*RSe6611*RC346e + 1./
         324.*pow(m13,-1)*me2*pow(RSe6611,2) + 1./324.*pow(m13,-1)*me2*
         RC146e*RSe6613 - 1./648.*pow(m13,-1)*me2*RC146e*RSe6633 - 1./
         648.*pow(m13,-1)*me2*RC146e*RSe6611 + 1./16.*pow(m13,-1)*m33
          + 1./324.*pow(m13,-1)*m33*pow(RS3461e,2) + 1./324.*pow(
         m13,-1)*m33*RS1463e*RS3461e - 1./162.*pow(m13,-1)*m33*RSe6613*
         RS3461e + 1./324.*pow(m13,-1)*m33*RSe6633*RS3461e - 1./324.*
         pow(m13,-1)*m33*RC346e*RS3461e - 1./648.*pow(m13,-1)*m33*
         RC346e*RS1463e + 1./324.*pow(m13,-1)*m33*RC346e*RSe6613 - 1./
         648.*pow(m13,-1)*m33*RC346e*RSe6633 + 1./1296.*pow(m13,-1)*m33
         *pow(RC346e,2) + 1./324.*pow(m13,-1)*m33*RSe6611*RS3461e - 1./
         648.*pow(m13,-1)*m33*RSe6611*RC346e - 1./648.*pow(m13,-1)*m33*
         RC146e*RS3461e );

      F13p6x22R +=  + pow(Ab(m13,mu2),2) * ( 1./1296.*pow(m13,-1)*m33*
         RC146e*RC346e + 1./16.*m11*pow(m13,-1) + 1./324.*m11*pow(
         m13,-1)*RS1463e*RS3461e + 1./324.*m11*pow(m13,-1)*pow(
         RS1463e,2) - 1./162.*m11*pow(m13,-1)*RSe6613*RS1463e + 1./324.
         *m11*pow(m13,-1)*RSe6633*RS1463e - 1./648.*m11*pow(m13,-1)*
         RC346e*RS1463e + 1./324.*m11*pow(m13,-1)*RSe6611*RS1463e - 1./
         648.*m11*pow(m13,-1)*RC146e*RS3461e - 1./324.*m11*pow(m13,-1)*
         RC146e*RS1463e + 1./324.*m11*pow(m13,-1)*RC146e*RSe6613 - 1./
         648.*m11*pow(m13,-1)*RC146e*RSe6633 + 1./1296.*m11*pow(m13,-1)
         *RC146e*RC346e - 1./648.*m11*pow(m13,-1)*RC146e*RSe6611 + 1./
         1296.*m11*pow(m13,-1)*pow(RC146e,2) );

      F13p6x22R +=  + Ab(m13,mu2)*Ab(m33,mu2) * (  - 5./24. + 1./162.*
         pow(RS3461e,2) + 1./162.*RS1463e*RS3461e - 1./81.*RSe6613*
         RS3461e + 1./162.*RSe6633*RS3461e - 1./162.*RC346e*RS3461e - 1.
         /324.*RC346e*RS1463e + 1./162.*RC346e*RSe6613 - 1./324.*RC346e
         *RSe6633 + 1./648.*pow(RC346e,2) + 1./162.*RSe6611*RS3461e - 1.
         /324.*RSe6611*RC346e - 1./324.*RC146e*RS3461e + 1./648.*RC146e
         *RC346e + 1./6.*m33*RN13 + 1./3.*m11*RN13 );

      F13p6x22R +=  + Ab(m13,mu2)*Ab(me2,mu2) * (  - 1./81.*RSe6613*
         RS3461e - 1./81.*RSe6613*RS1463e + 2./81.*pow(RSe6613,2) + 1./
         162.*RSe6633*RS3461e + 1./162.*RSe6633*RS1463e - 2./81.*
         RSe6633*RSe6613 + 1./162.*pow(RSe6633,2) + 1./162.*RC346e*
         RSe6613 - 1./324.*RC346e*RSe6633 + 1./162.*RSe6611*RS3461e + 1.
         /162.*RSe6611*RS1463e - 2./81.*RSe6611*RSe6613 + 1./81.*
         RSe6611*RSe6633 - 1./324.*RSe6611*RC346e + 1./162.*pow(
         RSe6611,2) + 1./162.*RC146e*RSe6613 - 1./324.*RC146e*RSe6633
          - 1./324.*RC146e*RSe6611 );

      F13p6x22R +=  + Ab(m13,mu2)*Bb(m11,mu2) * (  - 1./324.*RD146e*
         RS3461e - 1./324.*RD146e*RS1463e + 1./162.*RD146e*RSe6613 - 1./
         324.*RD146e*RSe6633 + 1./648.*RD146e*RC346e - 1./324.*RD146e*
         RSe6611 + 1./648.*RD146e*RC146e );

      F13p6x22R +=  + Ab(m13,mu2)*Bb(m33,mu2) * (  - 1./324.*RD346e*
         RS3461e - 1./324.*RD346e*RS1463e + 1./162.*RD346e*RSe6613 - 1./
         324.*RD346e*RSe6633 + 1./648.*RD346e*RC346e - 1./324.*RSe6611*
         RD346e + 1./648.*RC146e*RD346e );

      F13p6x22R +=  + Ab(m14,mu2)*pi16 * ( 1./24.*RD346e - 1./48.*me2
          - 1./12.*me2*RSe63 - 1./12.*me2*RSe6633 - 1./8.*m66 - 7./24.*
         m44 - 1./48.*m44*RS3461e - 1./48.*m44*RS1463e + 1./24.*m44*
         RSe6613 - 1./48.*m44*RSe63 - 1./24.*m44*RSe6633 - 1./48.*m44*
         RS36e + 1./48.*m44*RC346e - 1./48.*m44*RSe61 - 1./24.*m44*
         RSe6611 - 1./48.*m44*RS16e + 1./48.*m44*RC146e - 1./6.*m33 - 1.
         /24.*m33*RSe63 - 1./24.*m33*RSe6633 - 1./8.*m33*RS36e + 1./16.
         *m33*RC346e + 1./16.*m11 - 1./48.*m11*RS3461e - 1./48.*m11*
         RS1463e + 1./24.*m11*RSe6613 + 1./48.*m11*RSe63 + 1./48.*m11*
         RS36e - 1./48.*m11*RSe61 - 1./24.*m11*RSe6611 - 1./48.*m11*
         RS16e + 1./48.*m11*RC146e );

      F13p6x22R +=  + pow(Ab(m14,mu2),2) * (  - 1./8. - 1./48.*pow(
         m14,-1)*RD346e + 7./144.*pow(m14,-1)*RD146e + 1./18.*pow(
         m14,-1)*me2*RSe6613 + 5./72.*pow(m14,-1)*me2*RSe63 + 1./24.*
         pow(m14,-1)*me2*RSe6633 - 5./72.*pow(m14,-1)*me2*RSe61 - 7./72.
         *pow(m14,-1)*me2*RSe6611 + 1./16.*pow(m14,-1)*m66 + 35./108.*
         pow(m14,-1)*m44 + 7./432.*pow(m14,-1)*m44*RS3461e + 7./432.*
         pow(m14,-1)*m44*RS1463e - 7./216.*pow(m14,-1)*m44*RSe6613 + 19.
         /432.*pow(m14,-1)*m44*RSe63 + 1./216.*pow(m14,-1)*m44*RSe6633
          + 19./432.*pow(m14,-1)*m44*RS36e - 1./432.*pow(m14,-1)*m44*
         RC346e - 29./432.*pow(m14,-1)*m44*RSe61 + 1./216.*pow(m14,-1)*
         m44*RSe6611 - 29./432.*pow(m14,-1)*m44*RS16e - 1./432.*pow(
         m14,-1)*m44*RC146e + 5./432.*pow(m14,-1)*m33 - 5./432.*pow(
         m14,-1)*m33*RS3461e + 7./432.*pow(m14,-1)*m33*RS1463e - 7./216.
         *pow(m14,-1)*m33*RSe6613 - 5./432.*pow(m14,-1)*m33*RSe63 + 1./
         216.*pow(m14,-1)*m33*RSe6633 + 25./432.*pow(m14,-1)*m33*RS36e
          - 5./216.*pow(m14,-1)*m33*RC346e );

      F13p6x22R +=  + pow(Ab(m14,mu2),2) * (  - 5./432.*pow(m14,-1)*m33
         *RSe61 + 1./216.*pow(m14,-1)*m33*RSe6611 - 5./432.*pow(m14,-1)
         *m33*RS16e - 1./432.*pow(m14,-1)*m33*RC146e - 1./36.*m11*pow(
         m14,-1)*RS1463e + 1./9.*m11*pow(m14,-1)*RSe63 + 1./9.*m11*pow(
         m14,-1)*RS36e - 1./9.*m11*pow(m14,-1)*RSe61 - 13./72.*m11*pow(
         m14,-1)*RS16e + 7./144.*m11*pow(m14,-1)*RC146e );

      F13p6x22R +=  + Ab(m14,mu2)*Ab(m16,mu2) * (  - 1./8. );

      F13p6x22R +=  + Ab(m14,mu2)*Ab(m33,mu2) * (  - 1./18.*RS3461e + 5.
         /36.*RS36e - 1./24.*RC346e - 1./9.*m44*RN13*RS36e - 2./9.*m11*
         RN13*RS36e );

      F13p6x22R +=  + Ab(m14,mu2)*Ab(m34,mu2) * ( 8./27. + 7./108.*
         RS3461e + 7./108.*RS1463e - 7./54.*RSe6613 - 5./108.*RSe63 + 1.
         /54.*RSe6633 - 5./108.*RS36e - 1./108.*RC346e - 5./108.*RSe61
          + 1./54.*RSe6611 - 5./108.*RS16e - 1./108.*RC146e );

      F13p6x22R +=  + Ab(m14,mu2)*Ab(m36,mu2) * ( 1./8. );

      F13p6x22R +=  + Ab(m14,mu2)*Ab(me2,mu2) * (  - 1./9.*RSe6613 + 1./
         36.*RSe63 + 1./12.*RSe6633 - 1./36.*RSe61 + 1./36.*RSe6611 - 2.
         /9.*me2*RN1e*RSe6613 - 1./9.*me2*RN1e*RSe63 + 1./9.*me2*RN1e*
         RSe61 + 2./9.*me2*RN1e*RSe6611 - 1./9.*m44*RN1e*RSe63 + 1./9.*
         m44*RN1e*RSe61 + 2./9.*m11*RN1e*RSe6613 - 1./9.*m11*RN1e*RSe63
          + 1./9.*m11*RN1e*RSe61 - 2./9.*m11*RN1e*RSe6611 );

      F13p6x22R +=  + Ab(m14,mu2)*Bb(m11,mu2) * ( 7./72.*RD146e - 1./9.
         *m44*RS16e - 2./9.*m11*RS16e );

      F13p6x22R +=  + Ab(m14,mu2)*Bb(m33,mu2) * (  - 1./24.*RD346e );

      F13p6x22R +=  + Ab(m16,mu2)*pi16 * ( 1./48.*RD346e - 1./24.*me2
          + 1./12.*me2*RSe63 - 1./24.*me2*RSe6633 - 5./96.*m66 - 1./96.
         *m66*RS3461e - 1./96.*m66*RS1463e + 1./48.*m66*RSe6613 + 1./48.
         *m66*RSe63 - 1./48.*m66*RSe6633 - 1./96.*m66*RS34e + 1./96.*
         m66*RC346e + 1./48.*m66*RSe61 - 1./48.*m66*RSe6611 - 1./96.*
         m66*RS14e + 1./96.*m66*RC146e - 1./8.*m44 - 1./12.*m33 + 1./24.
         *m33*RSe63 - 1./48.*m33*RSe6633 - 1./16.*m33*RS34e + 1./32.*
         m33*RC346e + 1./32.*m11 - 1./96.*m11*RS3461e - 1./96.*m11*
         RS1463e + 1./48.*m11*RSe6613 - 1./48.*m11*RSe63 + 1./96.*m11*
         RS34e + 1./48.*m11*RSe61 - 1./48.*m11*RSe6611 - 1./96.*m11*
         RS14e + 1./96.*m11*RC146e );

      F13p6x22R +=  + pow(Ab(m16,mu2),2) * (  - 1./32. - 1./96.*pow(
         m16,-1)*RD346e + 7./288.*pow(m16,-1)*RD146e + 1./36.*pow(
         m16,-1)*me2*RSe6613 - 5./72.*pow(m16,-1)*me2*RSe63 + 1./48.*
         pow(m16,-1)*me2*RSe6633 + 5./72.*pow(m16,-1)*me2*RSe61 - 7./
         144.*pow(m16,-1)*me2*RSe6611 + 113./864.*pow(m16,-1)*m66 + 7./
         864.*pow(m16,-1)*m66*RS3461e + 7./864.*pow(m16,-1)*m66*RS1463e
          - 7./432.*pow(m16,-1)*m66*RSe6613 - 19./432.*pow(m16,-1)*m66*
         RSe63 + 1./432.*pow(m16,-1)*m66*RSe6633 + 19./864.*pow(m16,-1)
         *m66*RS34e - 1./864.*pow(m16,-1)*m66*RC346e + 29./432.*pow(
         m16,-1)*m66*RSe61 + 1./432.*pow(m16,-1)*m66*RSe6611 - 29./864.
         *pow(m16,-1)*m66*RS14e - 1./864.*pow(m16,-1)*m66*RC146e + 1./
         16.*pow(m16,-1)*m44 + 5./864.*pow(m16,-1)*m33 - 5./864.*pow(
         m16,-1)*m33*RS3461e + 7./864.*pow(m16,-1)*m33*RS1463e - 7./432.
         *pow(m16,-1)*m33*RSe6613 + 5./432.*pow(m16,-1)*m33*RSe63 + 1./
         432.*pow(m16,-1)*m33*RSe6633 + 25./864.*pow(m16,-1)*m33*RS34e
          - 5./432.*pow(m16,-1)*m33*RC346e );

      F13p6x22R +=  + pow(Ab(m16,mu2),2) * ( 5./432.*pow(m16,-1)*m33*
         RSe61 + 1./432.*pow(m16,-1)*m33*RSe6611 - 5./864.*pow(m16,-1)*
         m33*RS14e - 1./864.*pow(m16,-1)*m33*RC146e - 1./72.*m11*pow(
         m16,-1)*RS1463e - 1./9.*m11*pow(m16,-1)*RSe63 + 1./18.*m11*
         pow(m16,-1)*RS34e + 1./9.*m11*pow(m16,-1)*RSe61 - 13./144.*m11
         *pow(m16,-1)*RS14e + 7./288.*m11*pow(m16,-1)*RC146e );

      F13p6x22R +=  + Ab(m16,mu2)*Ab(m33,mu2) * (  - 1./36.*RS3461e + 5.
         /72.*RS34e - 1./48.*RC346e - 1./18.*m66*RN13*RS34e - 1./9.*m11
         *RN13*RS34e );

      F13p6x22R +=  + Ab(m16,mu2)*Ab(m34,mu2) * ( 1./8. );

      F13p6x22R +=  + Ab(m16,mu2)*Ab(m36,mu2) * ( 37./432. + 7./216.*
         RS3461e + 7./216.*RS1463e - 7./108.*RSe6613 + 5./108.*RSe63 + 
         1./108.*RSe6633 - 5./216.*RS34e - 1./216.*RC346e + 5./108.*
         RSe61 + 1./108.*RSe6611 - 5./216.*RS14e - 1./216.*RC146e );

      F13p6x22R +=  + Ab(m16,mu2)*Ab(me2,mu2) * (  - 1./18.*RSe6613 - 1.
         /36.*RSe63 + 1./24.*RSe6633 + 1./36.*RSe61 + 1./72.*RSe6611 - 
         1./9.*me2*RN1e*RSe6613 + 1./9.*me2*RN1e*RSe63 - 1./9.*me2*RN1e
         *RSe61 + 1./9.*me2*RN1e*RSe6611 + 1./9.*m66*RN1e*RSe63 - 1./9.
         *m66*RN1e*RSe61 + 1./9.*m11*RN1e*RSe6613 + 1./9.*m11*RN1e*
         RSe63 - 1./9.*m11*RN1e*RSe61 - 1./9.*m11*RN1e*RSe6611 );

      F13p6x22R +=  + Ab(m16,mu2)*Bb(m11,mu2) * ( 7./144.*RD146e - 1./
         18.*m66*RS14e - 1./9.*m11*RS14e );

      F13p6x22R +=  + Ab(m16,mu2)*Bb(m33,mu2) * (  - 1./48.*RD346e );

      F13p6x22R +=  + Ab(m33,mu2)*pi16 * (  - 1./48.*RD346e + 1./72.*
         RD346e*RS3461e - 1./144.*RD346e*RC346e + 1./72.*RD146e*RS3461e
          - 1./144.*RD146e*RC346e + 1./18.*me2*RSe6613*RS3461e - 1./36.
         *me2*RSe6633*RS3461e - 1./36.*me2*RC346e*RSe6613 + 1./72.*me2*
         RC346e*RSe6633 - 1./36.*me2*RSe6611*RS3461e + 1./72.*me2*
         RSe6611*RC346e - 1./24.*m66*RS3461e + 1./48.*m66*RC346e - 1./
         12.*m44*RS3461e + 1./24.*m44*RC346e + 1./16.*m33 - 1./48.*m33*
         RS3461e - 1./24.*m33*pow(RS3461e,2) - 1./72.*m33*RS1463e*
         RS3461e + 1./36.*m33*RSe6613*RS3461e - 1./72.*m33*RSe6633*
         RS3461e - 1./24.*m33*RS36e - 1./48.*m33*RS34e + 1./96.*m33*
         RC346e + 1./24.*m33*RC346e*RS3461e + 1./144.*m33*RC346e*
         RS1463e - 1./72.*m33*RC346e*RSe6613 + 1./144.*m33*RC346e*
         RSe6633 - 1./96.*m33*pow(RC346e,2) - 1./72.*m33*RSe6611*
         RS3461e + 1./144.*m33*RSe6611*RC346e + 1./144.*m33*RC146e*
         RS3461e - 1./288.*m33*RC146e*RC346e - 1./16.*m11*RS3461e - 1./
         72.*m11*pow(RS3461e,2) );

      F13p6x22R +=  + Ab(m33,mu2)*pi16 * (  - 1./24.*m11*RS1463e*
         RS3461e + 1./36.*m11*RSe6613*RS3461e - 1./72.*m11*RSe6633*
         RS3461e + 1./32.*m11*RC346e + 1./72.*m11*RC346e*RS3461e + 1./
         48.*m11*RC346e*RS1463e - 1./72.*m11*RC346e*RSe6613 + 1./144.*
         m11*RC346e*RSe6633 - 1./288.*m11*pow(RC346e,2) - 1./72.*m11*
         RSe6611*RS3461e + 1./144.*m11*RSe6611*RC346e + 1./48.*m11*
         RC146e*RS3461e - 1./96.*m11*RC146e*RC346e );

      F13p6x22R +=  + pow(Ab(m33,mu2),2) * ( 19./96. + 1./96.*pow(
         m33,-2)*m66*RN3e*RD34*RD36 + 1./48.*pow(m33,-2)*m44*RN3e*RD34*
         RD36 - 1./24.*pow(m33,-1)*RD346e - 1./144.*pow(m33,-1)*RD346e*
         RS3461e + 1./288.*pow(m33,-1)*RD346e*RC346e - 1./144.*pow(
         m33,-1)*RD146e*RS3461e + 1./288.*pow(m33,-1)*RD146e*RC346e + 1.
         /32.*pow(m33,-1)*RN3e*RD34*RD36 - 1./36.*pow(m33,-1)*me2*
         RSe6613*RS3461e + 1./72.*pow(m33,-1)*me2*RSe6633*RS3461e + 1./
         72.*pow(m33,-1)*me2*RC346e*RSe6613 - 1./144.*pow(m33,-1)*me2*
         RC346e*RSe6633 + 1./72.*pow(m33,-1)*me2*RSe6611*RS3461e - 1./
         144.*pow(m33,-1)*me2*RSe6611*RC346e + 7./144.*pow(m33,-1)*m66*
         RS3461e + 1./36.*pow(m33,-1)*m66*RS1463e - 1./18.*pow(m33,-1)*
         m66*RSe6613 + 1./36.*pow(m33,-1)*m66*RSe63 + 1./18.*pow(
         m33,-1)*m66*RSe6633 + 1./108.*pow(m33,-1)*m66*pow(RS36e,2) - 1.
         /72.*pow(m33,-1)*m66*RS34e - 1./12.*pow(m33,-1)*m66*RS34e*
         RSe63 + 1./108.*pow(m33,-1)*m66*RS34e*RS36e + 5./216.*pow(
         m33,-1)*m66*pow(RS34e,2) );

      F13p6x22R +=  + pow(Ab(m33,mu2),2) * (  - 7./288.*pow(m33,-1)*m66
         *RC346e - 1./36.*pow(m33,-1)*m66*RSe61 + 1./12.*pow(m33,-1)*
         m66*RSe61*RS34e - 1./54.*pow(m33,-1)*m66*RS16e*RS36e - 1./108.
         *pow(m33,-1)*m66*RS16e*RS34e + 1./72.*pow(m33,-1)*m66*RS14e - 
         1./108.*pow(m33,-1)*m66*RS14e*RS36e - 5./108.*pow(m33,-1)*m66*
         RS14e*RS34e + 7./72.*pow(m33,-1)*m44*RS3461e + 1./18.*pow(
         m33,-1)*m44*RS1463e - 1./9.*pow(m33,-1)*m44*RSe6613 - 1./36.*
         pow(m33,-1)*m44*RSe63 + 1./9.*pow(m33,-1)*m44*RSe6633 - 1./36.
         *pow(m33,-1)*m44*RS36e + 1./9.*pow(m33,-1)*m44*RS36e*RSe63 + 7.
         /108.*pow(m33,-1)*m44*pow(RS36e,2) - 1./36.*pow(m33,-1)*m44*
         RS34e*RSe63 + 1./108.*pow(m33,-1)*m44*RS34e*RS36e + 1./108.*
         pow(m33,-1)*m44*pow(RS34e,2) - 7./144.*pow(m33,-1)*m44*RC346e
          + 1./36.*pow(m33,-1)*m44*RSe61 - 1./9.*pow(m33,-1)*m44*RSe61*
         RS36e + 1./36.*pow(m33,-1)*m44*RSe61*RS34e + 1./36.*pow(
         m33,-1)*m44*RS16e - 7./54.*pow(m33,-1)*m44*RS16e*RS36e - 1./
         108.*pow(m33,-1)*m44*RS16e*RS34e );

      F13p6x22R +=  + pow(Ab(m33,mu2),2) * (  - 1./108.*pow(m33,-1)*m44
         *RS14e*RS36e - 1./54.*pow(m33,-1)*m44*RS14e*RS34e + 1./96.*
         RS3461e + 19./648.*pow(RS3461e,2) + 1./12.*RS1463e + 1./648.*
         RS1463e*RS3461e - 1./6.*RSe6613 - 1./324.*RSe6613*RS3461e + 1./
         6.*RSe6633 + 1./648.*RSe6633*RS3461e - 13./144.*RS36e - 13./
         288.*RS34e - 7./96.*RC346e - 37./648.*RC346e*RS3461e - 37./
         1296.*RC346e*RS1463e + 37./648.*RC346e*RSe6613 - 73./1296.*
         RC346e*RSe6633 + 91./2592.*pow(RC346e,2) + 1./648.*RSe6611*
         RS3461e - 1./1296.*RSe6611*RC346e + 1./36.*RS16e + 1./72.*
         RS14e - 1./1296.*RC146e*RS3461e + 1./2592.*RC146e*RC346e + 1./
         9.*m33*RN3e*RC346e*RSe6613 - 1./9.*m33*RN3e*RC346e*RSe6633 + 1.
         /18.*m33*RN13*RC346e*RS1463e - 5./96.*m11*pow(m33,-1) + 7./96.
         *m11*pow(m33,-1)*RS3461e + 1./648.*m11*pow(m33,-1)*pow(
         RS3461e,2) + 5./324.*m11*pow(m33,-1)*RS1463e*RS3461e - 1./324.
         *m11*pow(m33,-1)*RSe6613*RS3461e + 1./648.*m11*pow(m33,-1)*
         RSe6633*RS3461e );

      F13p6x22R +=  + pow(Ab(m33,mu2),2) * ( 5./144.*m11*pow(m33,-1)*
         RS36e + 5./288.*m11*pow(m33,-1)*RS34e - 1./648.*m11*pow(
         m33,-1)*RC346e*RS3461e - 5./648.*m11*pow(m33,-1)*RC346e*
         RS1463e + 1./648.*m11*pow(m33,-1)*RC346e*RSe6613 - 1./1296.*
         m11*pow(m33,-1)*RC346e*RSe6633 + 1./2592.*m11*pow(m33,-1)*pow(
         RC346e,2) + 1./648.*m11*pow(m33,-1)*RSe6611*RS3461e - 1./1296.
         *m11*pow(m33,-1)*RSe6611*RC346e - 5./648.*m11*pow(m33,-1)*
         RC146e*RS3461e + 5./1296.*m11*pow(m33,-1)*RC146e*RC346e );

      F13p6x22R +=  + Ab(m33,mu2)*Ab(m34,mu2) * (  - 1./18.*RS3461e - 1.
         /9.*RS1463e + 2./9.*RSe6613 - 1./9.*RSe63 - 2./9.*RSe6633 - 5./
         36.*RS36e + 7./72.*RC346e + 1./9.*RSe61 + 1./9.*RS16e + 2./9.*
         me2*RN3e*RSe6613 - 1./9.*me2*RN3e*RSe63 - 2./9.*me2*RN3e*
         RSe6633 + 1./9.*me2*RN3e*RSe61 - 1./9.*m44*RN3e*RSe63 + 1./9.*
         m44*RN3e*RSe61 - 1./9.*m44*RN13*RS16e - 2./9.*m33*RN3e*RSe6613
          - 1./9.*m33*RN3e*RSe63 + 2./9.*m33*RN3e*RSe6633 + 1./9.*m33*
         RN3e*RSe61 - 1./9.*m33*RN13*RS1463e - 1./9.*m33*RN13*RS16e + 1.
         /9.*m11*RN13*RS1463e - 1./9.*m11*RN13*RS16e );

      F13p6x22R +=  + Ab(m33,mu2)*Ab(m36,mu2) * (  - 1./36.*RS3461e - 1.
         /18.*RS1463e + 1./9.*RSe6613 + 1./9.*RSe63 - 1./9.*RSe6633 - 5.
         /72.*RS34e + 7./144.*RC346e - 1./9.*RSe61 + 1./18.*RS14e + 1./
         9.*me2*RN3e*RSe6613 + 1./9.*me2*RN3e*RSe63 - 1./9.*me2*RN3e*
         RSe6633 - 1./9.*me2*RN3e*RSe61 + 1./9.*m66*RN3e*RSe63 - 1./9.*
         m66*RN3e*RSe61 - 1./18.*m66*RN13*RS14e - 1./9.*m33*RN3e*
         RSe6613 + 1./9.*m33*RN3e*RSe63 + 1./9.*m33*RN3e*RSe6633 - 1./9.
         *m33*RN3e*RSe61 - 1./18.*m33*RN13*RS1463e - 1./18.*m33*RN13*
         RS14e + 1./18.*m11*RN13*RS1463e - 1./18.*m11*RN13*RS14e );

      F13p6x22R +=  + Ab(m33,mu2)*Ab(m44,mu2) * ( 1./6.*m44*RN3e*RS36e*
         RSe63 - 1./6.*m44*RN3e*RSe61*RS36e + 1./6.*m44*RN13*RS16e*
         RS36e );

      F13p6x22R +=  + Ab(m33,mu2)*Ab(m46,mu2) * (  - 1./9.*RS36e*RSe63
          - 1./54.*pow(RS36e,2) + 1./9.*RS34e*RSe63 + 1./27.*RS34e*
         RS36e - 1./54.*pow(RS34e,2) + 1./9.*RSe61*RS36e - 1./9.*RSe61*
         RS34e + 1./27.*RS16e*RS36e - 1./27.*RS16e*RS34e - 1./27.*RS14e
         *RS36e + 1./27.*RS14e*RS34e - 1./9.*me2*RN3e*RS36e*RSe63 + 1./
         9.*me2*RN3e*RS34e*RSe63 + 1./9.*me2*RN3e*RSe61*RS36e - 1./9.*
         me2*RN3e*RSe61*RS34e - 2./27.*m66*RN3e*RS36e*RSe63 - 1./27.*
         m66*RN3e*RS34e*RSe63 + 2./27.*m66*RN3e*RSe61*RS36e + 1./27.*
         m66*RN3e*RSe61*RS34e + 1./27.*m66*RN13*RS16e*RS34e + 1./27.*
         m66*RN13*RS14e*RS36e + 1./27.*m66*RN13*RS14e*RS34e - 1./27.*
         m44*RN3e*RS36e*RSe63 + 1./27.*m44*RN3e*RS34e*RSe63 + 1./27.*
         m44*RN3e*RSe61*RS36e - 1./27.*m44*RN3e*RSe61*RS34e + 1./27.*
         m44*RN13*RS16e*RS36e + 1./27.*m44*RN13*RS16e*RS34e + 1./27.*
         m44*RN13*RS14e*RS36e - 1./27.*m11*RN13*RS16e*RS36e + 1./27.*
         m11*RN13*RS16e*RS34e + 1./27.*m11*RN13*RS14e*RS36e - 1./27.*
         m11*RN13*RS14e*RS34e );

      F13p6x22R +=  + Ab(m33,mu2)*Ab(me2,mu2) * (  - 1./18.*RSe6613*
         RS3461e + 1./36.*RSe6633*RS3461e + 1./36.*RC346e*RSe6613 - 1./
         72.*RC346e*RSe6633 + 1./36.*RSe6611*RS3461e - 1./72.*RSe6611*
         RC346e - 2./27.*m66*RN3e*RS34e*RSe63 + 2./27.*m66*RN3e*RSe61*
         RS34e + 1./27.*m66*RN13*RS14e*RS34e + 1./54.*m44*RN3e*RS36e*
         RSe63 - 1./54.*m44*RN3e*RSe61*RS36e + 1./54.*m44*RN13*RS16e*
         RS36e - 2./9.*m33*RN3e*RSe6633*RSe6613 + 2./9.*m33*RN3e*pow(
         RSe6633,2) - 1./9.*m33*RN3e*RC346e*RSe6613 + 1./9.*m33*RN3e*
         RC346e*RSe6633 - 1./9.*m33*RN13*RSe6633*RS1463e - 1./9.*m11*
         RN13*RSe6611*RS3461e );

      F13p6x22R +=  + Ab(m33,mu2)*Bb(m11,mu2) * (  - 1./72.*RD146e*
         RS3461e + 1./144.*RD146e*RC346e + 1./18.*m11*RN13*RD146e*
         RS3461e );

      F13p6x22R +=  + Ab(m33,mu2)*Bb(m33,mu2) * (  - 1./48.*pow(m33,-1)
         *m66*RN3e*RD34*RD36 - 1./24.*pow(m33,-1)*m44*RN3e*RD34*RD36 - 
         1./72.*RD346e*RS3461e + 1./144.*RD346e*RC346e - 1./16.*RN3e*
         RD34*RD36 + 1./4.*m33 - 1./18.*m33*RC346e*RS3461e + 1./18.*m33
         *pow(RC346e,2) + 1./9.*m33*RN3e*RD346e*RSe6613 - 1./9.*m33*
         RN3e*RD346e*RSe6633 + 1./18.*m33*RN13*RD346e*RS1463e );

      F13p6x22R +=  + Ab(m33,mu2)*Bb(m34,mu2) * ( 1./36.*RD346e - 1./18.
         *m44*RS36e - 1./9.*m33*RS36e );

      F13p6x22R +=  + Ab(m33,mu2)*Bb(m36,mu2) * ( 1./72.*RD346e - 1./36.
         *m66*RS34e - 1./18.*m33*RS34e );

      F13p6x22R +=  + Ab(m33,mu2)*Cb(m33,mu2) * ( 1./18.*m33*RD346e*
         RC346e );

      F13p6x22R +=  + Ab(m34,mu2)*pi16 * ( 1./24.*RD146e - 1./48.*me2
          - 1./12.*me2*RSe61 - 1./12.*me2*RSe6611 - 1./8.*m66 - 7./24.*
         m44 - 1./48.*m44*RS3461e - 1./48.*m44*RS1463e + 1./24.*m44*
         RSe6613 - 1./48.*m44*RSe63 - 1./24.*m44*RSe6633 - 1./48.*m44*
         RS36e + 1./48.*m44*RC346e - 1./48.*m44*RSe61 - 1./24.*m44*
         RSe6611 - 1./48.*m44*RS16e + 1./48.*m44*RC146e + 1./16.*m33 - 
         1./48.*m33*RS3461e - 1./48.*m33*RS1463e + 1./24.*m33*RSe6613
          - 1./48.*m33*RSe63 - 1./24.*m33*RSe6633 - 1./48.*m33*RS36e + 
         1./48.*m33*RC346e + 1./48.*m33*RSe61 + 1./48.*m33*RS16e - 1./6.
         *m11 - 1./24.*m11*RSe61 - 1./24.*m11*RSe6611 - 1./8.*m11*RS16e
          + 1./16.*m11*RC146e );

      F13p6x22R +=  + pow(Ab(m34,mu2),2) * (  - 1./8. + 7./144.*pow(
         m34,-1)*RD346e - 1./48.*pow(m34,-1)*RD146e + 1./18.*pow(
         m34,-1)*me2*RSe6613 - 5./72.*pow(m34,-1)*me2*RSe63 - 7./72.*
         pow(m34,-1)*me2*RSe6633 + 5./72.*pow(m34,-1)*me2*RSe61 + 1./24.
         *pow(m34,-1)*me2*RSe6611 + 1./16.*pow(m34,-1)*m66 + 35./108.*
         pow(m34,-1)*m44 + 7./432.*pow(m34,-1)*m44*RS3461e + 7./432.*
         pow(m34,-1)*m44*RS1463e - 7./216.*pow(m34,-1)*m44*RSe6613 - 29.
         /432.*pow(m34,-1)*m44*RSe63 + 1./216.*pow(m34,-1)*m44*RSe6633
          - 29./432.*pow(m34,-1)*m44*RS36e - 1./432.*pow(m34,-1)*m44*
         RC346e + 19./432.*pow(m34,-1)*m44*RSe61 + 1./216.*pow(m34,-1)*
         m44*RSe6611 + 19./432.*pow(m34,-1)*m44*RS16e - 1./432.*pow(
         m34,-1)*m44*RC146e - 1./36.*m33*pow(m34,-1)*RS3461e - 1./9.*
         m33*pow(m34,-1)*RSe63 - 13./72.*m33*pow(m34,-1)*RS36e + 7./144.
         *m33*pow(m34,-1)*RC346e + 1./9.*m33*pow(m34,-1)*RSe61 + 1./9.*
         m33*pow(m34,-1)*RS16e + 5./432.*m11*pow(m34,-1) + 7./432.*m11*
         pow(m34,-1)*RS3461e );

      F13p6x22R +=  + pow(Ab(m34,mu2),2) * (  - 5./432.*m11*pow(m34,-1)
         *RS1463e - 7./216.*m11*pow(m34,-1)*RSe6613 - 5./432.*m11*pow(
         m34,-1)*RSe63 + 1./216.*m11*pow(m34,-1)*RSe6633 - 5./432.*m11*
         pow(m34,-1)*RS36e - 1./432.*m11*pow(m34,-1)*RC346e - 5./432.*
         m11*pow(m34,-1)*RSe61 + 1./216.*m11*pow(m34,-1)*RSe6611 + 25./
         432.*m11*pow(m34,-1)*RS16e - 5./216.*m11*pow(m34,-1)*RC146e );

      F13p6x22R +=  + Ab(m34,mu2)*Ab(m36,mu2) * (  - 1./8. );

      F13p6x22R +=  + Ab(m34,mu2)*Ab(me2,mu2) * (  - 1./9.*RSe6613 - 1./
         36.*RSe63 + 1./36.*RSe6633 + 1./36.*RSe61 + 1./12.*RSe6611 - 2.
         /9.*me2*RN3e*RSe6613 + 1./9.*me2*RN3e*RSe63 + 2./9.*me2*RN3e*
         RSe6633 - 1./9.*me2*RN3e*RSe61 + 1./9.*m44*RN3e*RSe63 - 1./9.*
         m44*RN3e*RSe61 + 2./9.*m33*RN3e*RSe6613 + 1./9.*m33*RN3e*RSe63
          - 2./9.*m33*RN3e*RSe6633 - 1./9.*m33*RN3e*RSe61 );

      F13p6x22R +=  + Ab(m34,mu2)*Bb(m11,mu2) * (  - 1./24.*RD146e );

      F13p6x22R +=  + Ab(m34,mu2)*Bb(m33,mu2) * ( 7./72.*RD346e - 1./9.
         *m44*RS36e - 2./9.*m33*RS36e );

      F13p6x22R +=  + Ab(m36,mu2)*pi16 * ( 1./48.*RD146e - 1./24.*me2
          + 1./12.*me2*RSe61 - 1./24.*me2*RSe6611 - 5./96.*m66 - 1./96.
         *m66*RS3461e - 1./96.*m66*RS1463e + 1./48.*m66*RSe6613 + 1./48.
         *m66*RSe63 - 1./48.*m66*RSe6633 - 1./96.*m66*RS34e + 1./96.*
         m66*RC346e + 1./48.*m66*RSe61 - 1./48.*m66*RSe6611 - 1./96.*
         m66*RS14e + 1./96.*m66*RC146e - 1./8.*m44 + 1./32.*m33 - 1./96.
         *m33*RS3461e - 1./96.*m33*RS1463e + 1./48.*m33*RSe6613 + 1./48.
         *m33*RSe63 - 1./48.*m33*RSe6633 - 1./96.*m33*RS34e + 1./96.*
         m33*RC346e - 1./48.*m33*RSe61 + 1./96.*m33*RS14e - 1./12.*m11
          + 1./24.*m11*RSe61 - 1./48.*m11*RSe6611 - 1./16.*m11*RS14e + 
         1./32.*m11*RC146e );

      F13p6x22R +=  + pow(Ab(m36,mu2),2) * (  - 1./32. + 7./288.*pow(
         m36,-1)*RD346e - 1./96.*pow(m36,-1)*RD146e + 1./36.*pow(
         m36,-1)*me2*RSe6613 + 5./72.*pow(m36,-1)*me2*RSe63 - 7./144.*
         pow(m36,-1)*me2*RSe6633 - 5./72.*pow(m36,-1)*me2*RSe61 + 1./48.
         *pow(m36,-1)*me2*RSe6611 + 113./864.*pow(m36,-1)*m66 + 7./864.
         *pow(m36,-1)*m66*RS3461e + 7./864.*pow(m36,-1)*m66*RS1463e - 7.
         /432.*pow(m36,-1)*m66*RSe6613 + 29./432.*pow(m36,-1)*m66*RSe63
          + 1./432.*pow(m36,-1)*m66*RSe6633 - 29./864.*pow(m36,-1)*m66*
         RS34e - 1./864.*pow(m36,-1)*m66*RC346e - 19./432.*pow(m36,-1)*
         m66*RSe61 + 1./432.*pow(m36,-1)*m66*RSe6611 + 19./864.*pow(
         m36,-1)*m66*RS14e - 1./864.*pow(m36,-1)*m66*RC146e + 1./16.*
         pow(m36,-1)*m44 - 1./72.*m33*pow(m36,-1)*RS3461e + 1./9.*m33*
         pow(m36,-1)*RSe63 - 13./144.*m33*pow(m36,-1)*RS34e + 7./288.*
         m33*pow(m36,-1)*RC346e - 1./9.*m33*pow(m36,-1)*RSe61 + 1./18.*
         m33*pow(m36,-1)*RS14e + 5./864.*m11*pow(m36,-1) + 7./864.*m11*
         pow(m36,-1)*RS3461e );

      F13p6x22R +=  + pow(Ab(m36,mu2),2) * (  - 5./864.*m11*pow(m36,-1)
         *RS1463e - 7./432.*m11*pow(m36,-1)*RSe6613 + 5./432.*m11*pow(
         m36,-1)*RSe63 + 1./432.*m11*pow(m36,-1)*RSe6633 - 5./864.*m11*
         pow(m36,-1)*RS34e - 1./864.*m11*pow(m36,-1)*RC346e + 5./432.*
         m11*pow(m36,-1)*RSe61 + 1./432.*m11*pow(m36,-1)*RSe6611 + 25./
         864.*m11*pow(m36,-1)*RS14e - 5./432.*m11*pow(m36,-1)*RC146e );

      F13p6x22R +=  + Ab(m36,mu2)*Ab(me2,mu2) * (  - 1./18.*RSe6613 + 1.
         /36.*RSe63 + 1./72.*RSe6633 - 1./36.*RSe61 + 1./24.*RSe6611 - 
         1./9.*me2*RN3e*RSe6613 - 1./9.*me2*RN3e*RSe63 + 1./9.*me2*RN3e
         *RSe6633 + 1./9.*me2*RN3e*RSe61 - 1./9.*m66*RN3e*RSe63 + 1./9.
         *m66*RN3e*RSe61 + 1./9.*m33*RN3e*RSe6613 - 1./9.*m33*RN3e*
         RSe63 - 1./9.*m33*RN3e*RSe6633 + 1./9.*m33*RN3e*RSe61 );

      F13p6x22R +=  + Ab(m36,mu2)*Bb(m11,mu2) * (  - 1./48.*RD146e );

      F13p6x22R +=  + Ab(m36,mu2)*Bb(m33,mu2) * ( 7./144.*RD346e - 1./
         18.*m66*RS34e - 1./9.*m33*RS34e );

      F13p6x22R +=  + pow(Ab(m44,mu2),2) * ( 3./16. + 1./24.*pow(
         RSe63,2) + 1./12.*RS36e*RSe63 + 1./24.*pow(RS36e,2) - 1./12.*
         RSe61*RSe63 - 1./12.*RSe61*RS36e + 1./24.*pow(RSe61,2) - 1./12.
         *RS16e*RSe63 - 1./12.*RS16e*RS36e + 1./12.*RS16e*RSe61 + 1./24.
         *pow(RS16e,2) );

      F13p6x22R +=  + Ab(m44,mu2)*Ab(me2,mu2) * (  - 1./6.*m44*RN3e*
         RS36e*RSe63 + 1./6.*m44*RN3e*RSe61*RS36e + 1./6.*m44*RN1e*
         RS16e*RSe63 - 1./6.*m44*RN1e*RS16e*RSe61 );

      F13p6x22R +=  + Ab(m44,mu2)*Bb(m11,mu2) * ( 1./12.*m44*pow(
         RS16e,2) );

      F13p6x22R +=  + Ab(m44,mu2)*Bb(m33,mu2) * ( 1./12.*m44*pow(
         RS36e,2) );

      F13p6x22R +=  + Ab(m44,mu2)*Bb(me2,mu2) * ( 1./12.*m44*pow(
         RSe63,2) - 1./6.*m44*RSe61*RSe63 + 1./12.*m44*pow(RSe61,2) );

      F13p6x22R +=  + pow(Ab(m46,mu2),2) * (  - 1./6.*pow(m46,-1)*me2*
         pow(RSe63,2) - 1./18.*pow(m46,-1)*me2*RS36e*RSe63 + 1./18.*
         pow(m46,-1)*me2*RS34e*RSe63 + 1./3.*pow(m46,-1)*me2*RSe61*
         RSe63 + 1./18.*pow(m46,-1)*me2*RSe61*RS36e - 1./18.*pow(
         m46,-1)*me2*RSe61*RS34e - 1./6.*pow(m46,-1)*me2*pow(RSe61,2)
          + 1./18.*pow(m46,-1)*me2*RS16e*RSe63 - 1./18.*pow(m46,-1)*me2
         *RS16e*RSe61 - 1./18.*pow(m46,-1)*me2*RS14e*RSe63 + 1./18.*
         pow(m46,-1)*me2*RS14e*RSe61 + 1./8.*pow(m46,-1)*m66 - 1./27.*
         pow(m46,-1)*m66*RS36e*RSe63 - 1./54.*pow(m46,-1)*m66*RS34e*
         RSe63 + 1./54.*pow(m46,-1)*m66*RS34e*RS36e + 1./108.*pow(
         m46,-1)*m66*pow(RS34e,2) + 1./27.*pow(m46,-1)*m66*RSe61*RS36e
          + 1./54.*pow(m46,-1)*m66*RSe61*RS34e + 1./27.*pow(m46,-1)*m66
         *RS16e*RSe63 - 1./54.*pow(m46,-1)*m66*RS16e*RS34e - 1./27.*
         pow(m46,-1)*m66*RS16e*RSe61 + 1./54.*pow(m46,-1)*m66*RS14e*
         RSe63 - 1./54.*pow(m46,-1)*m66*RS14e*RS36e - 1./54.*pow(
         m46,-1)*m66*RS14e*RS34e );

      F13p6x22R +=  + pow(Ab(m46,mu2),2) * (  - 1./54.*pow(m46,-1)*m66*
         RS14e*RSe61 + 1./54.*pow(m46,-1)*m66*RS14e*RS16e + 1./108.*
         pow(m46,-1)*m66*pow(RS14e,2) + 1./8.*m44*pow(m46,-1) - 1./36.*
         m44*pow(m46,-1)*pow(RSe63,2) - 1./54.*m44*pow(m46,-1)*RS36e*
         RSe63 + 1./108.*m44*pow(m46,-1)*pow(RS36e,2) + 1./54.*m44*pow(
         m46,-1)*RS34e*RSe63 + 1./54.*m44*pow(m46,-1)*RS34e*RS36e + 1./
         18.*m44*pow(m46,-1)*RSe61*RSe63 + 1./54.*m44*pow(m46,-1)*RSe61
         *RS36e - 1./54.*m44*pow(m46,-1)*RSe61*RS34e - 1./36.*m44*pow(
         m46,-1)*pow(RSe61,2) + 1./54.*m44*pow(m46,-1)*RS16e*RSe63 - 1./
         54.*m44*pow(m46,-1)*RS16e*RS36e - 1./54.*m44*pow(m46,-1)*RS16e
         *RS34e - 1./54.*m44*pow(m46,-1)*RS16e*RSe61 + 1./108.*m44*pow(
         m46,-1)*pow(RS16e,2) - 1./54.*m44*pow(m46,-1)*RS14e*RSe63 - 1./
         54.*m44*pow(m46,-1)*RS14e*RS36e + 1./54.*m44*pow(m46,-1)*RS14e
         *RSe61 + 1./54.*m44*pow(m46,-1)*RS14e*RS16e - 1./18.*m33*pow(
         m46,-1)*RS36e*RSe63 - 1./54.*m33*pow(m46,-1)*pow(RS36e,2) + 1./
         18.*m33*pow(m46,-1)*RS34e*RSe63 );

      F13p6x22R +=  + pow(Ab(m46,mu2),2) * ( 1./27.*m33*pow(m46,-1)*
         RS34e*RS36e - 1./54.*m33*pow(m46,-1)*pow(RS34e,2) + 1./18.*m33
         *pow(m46,-1)*RSe61*RS36e - 1./18.*m33*pow(m46,-1)*RSe61*RS34e
          + 1./54.*m33*pow(m46,-1)*RS16e*RS36e - 1./54.*m33*pow(m46,-1)
         *RS16e*RS34e - 1./54.*m33*pow(m46,-1)*RS14e*RS36e + 1./54.*m33
         *pow(m46,-1)*RS14e*RS34e + 1./18.*m11*pow(m46,-1)*RS16e*RSe63
          + 1./54.*m11*pow(m46,-1)*RS16e*RS36e - 1./54.*m11*pow(m46,-1)
         *RS16e*RS34e - 1./18.*m11*pow(m46,-1)*RS16e*RSe61 - 1./54.*m11
         *pow(m46,-1)*pow(RS16e,2) - 1./18.*m11*pow(m46,-1)*RS14e*RSe63
          - 1./54.*m11*pow(m46,-1)*RS14e*RS36e + 1./54.*m11*pow(m46,-1)
         *RS14e*RS34e + 1./18.*m11*pow(m46,-1)*RS14e*RSe61 + 1./27.*m11
         *pow(m46,-1)*RS14e*RS16e - 1./54.*m11*pow(m46,-1)*pow(RS14e,2)
          );

      F13p6x22R +=  + Ab(m46,mu2)*Ab(me2,mu2) * (  - 1./6.*pow(RSe63,2)
          + 1./3.*RSe61*RSe63 - 1./6.*pow(RSe61,2) + 1./9.*me2*RN3e*
         RS36e*RSe63 - 1./9.*me2*RN3e*RS34e*RSe63 - 1./9.*me2*RN3e*
         RSe61*RS36e + 1./9.*me2*RN3e*RSe61*RS34e - 1./9.*me2*RN1e*
         RS16e*RSe63 + 1./9.*me2*RN1e*RS16e*RSe61 + 1./9.*me2*RN1e*
         RS14e*RSe63 - 1./9.*me2*RN1e*RS14e*RSe61 + 2./27.*m66*RN3e*
         RS36e*RSe63 + 1./27.*m66*RN3e*RS34e*RSe63 - 2./27.*m66*RN3e*
         RSe61*RS36e - 1./27.*m66*RN3e*RSe61*RS34e - 2./27.*m66*RN1e*
         RS16e*RSe63 + 2./27.*m66*RN1e*RS16e*RSe61 - 1./27.*m66*RN1e*
         RS14e*RSe63 + 1./27.*m66*RN1e*RS14e*RSe61 + 1./27.*m44*RN3e*
         RS36e*RSe63 - 1./27.*m44*RN3e*RS34e*RSe63 - 1./27.*m44*RN3e*
         RSe61*RS36e + 1./27.*m44*RN3e*RSe61*RS34e - 1./27.*m44*RN1e*
         RS16e*RSe63 + 1./27.*m44*RN1e*RS16e*RSe61 + 1./27.*m44*RN1e*
         RS14e*RSe63 - 1./27.*m44*RN1e*RS14e*RSe61 );

      F13p6x22R +=  + Ab(m46,mu2)*Bb(m11,mu2) * ( 1./27.*m66*RS14e*
         RS16e + 1./54.*m66*pow(RS14e,2) + 1./54.*m44*pow(RS16e,2) + 1./
         27.*m44*RS14e*RS16e - 1./54.*m11*pow(RS16e,2) + 1./27.*m11*
         RS14e*RS16e - 1./54.*m11*pow(RS14e,2) );

      F13p6x22R +=  + Ab(m46,mu2)*Bb(m33,mu2) * ( 1./27.*m66*RS34e*
         RS36e + 1./54.*m66*pow(RS34e,2) + 1./54.*m44*pow(RS36e,2) + 1./
         27.*m44*RS34e*RS36e - 1./54.*m33*pow(RS36e,2) + 1./27.*m33*
         RS34e*RS36e - 1./54.*m33*pow(RS34e,2) );

      F13p6x22R +=  + Ab(m46,mu2)*Bb(me2,mu2) * (  - 1./6.*me2*pow(
         RSe63,2) + 1./3.*me2*RSe61*RSe63 - 1./6.*me2*pow(RSe61,2) - 1./
         18.*m44*pow(RSe63,2) + 1./9.*m44*RSe61*RSe63 - 1./18.*m44*pow(
         RSe61,2) );

      F13p6x22R +=  + Ab(me2,mu2)*pi16 * (  - 1./648.*m66*pow(RN1e,2)*
         pow(RN3e,2)*pow(RD13,2)*pow(RD46,2) - 5./648.*m44*pow(RN1e,2)*
         pow(RN3e,2)*pow(RD13,2)*pow(RD46,2) - 1./144.*m33*pow(RN1e,2)*
         pow(RN3e,2)*pow(RD13,2)*pow(RD46,2) - 1./144.*m11*pow(RN1e,2)*
         pow(RN3e,2)*pow(RD13,2)*pow(RD46,2) );

      F13p6x22R +=  + pow(Ab(me2,mu2),2) * ( 7./48. + 1./72.*pow(
         me2,-1)*RD346e*RSe6613 - 1./144.*pow(me2,-1)*RD346e*RSe6633 - 
         1./144.*pow(me2,-1)*RSe6611*RD346e + 1./72.*pow(me2,-1)*RD146e
         *RSe6613 - 1./144.*pow(me2,-1)*RD146e*RSe6633 - 1./144.*pow(
         me2,-1)*RD146e*RSe6611 + 1./8.*RSe6613 + 1./18.*pow(RSe6613,2)
          + 1./12.*RSe6633 - 1./18.*RSe6633*RSe6613 + 1./72.*pow(
         RSe6633,2) + 1./12.*RSe6611 - 1./18.*RSe6611*RSe6613 + 1./36.*
         RSe6611*RSe6633 + 1./72.*pow(RSe6611,2) + 1./5832.*pow(RN1e,4)
         *pow(RN3e,4)*pow(RD13,4)*pow(RD46,4) - 1./18.*m66*pow(me2,-1)
          + 1./72.*m66*pow(me2,-1)*RSe6613 + 1./36.*m66*pow(me2,-1)*
         RSe63 + 13./108.*m66*pow(me2,-1)*pow(RSe63,2) - 1./144.*m66*
         pow(me2,-1)*RSe6633 - 1./27.*m66*pow(me2,-1)*RS34e*RSe63 + 1./
         108.*m66*pow(me2,-1)*pow(RS34e,2) + 1./36.*m66*pow(me2,-1)*
         RSe61 - 13./54.*m66*pow(me2,-1)*RSe61*RSe63 + 1./27.*m66*pow(
         me2,-1)*RSe61*RS34e + 13./108.*m66*pow(me2,-1)*pow(RSe61,2) - 
         1./144.*m66*pow(me2,-1)*RSe6611 );

      F13p6x22R +=  + pow(Ab(me2,mu2),2) * ( 1./27.*m66*pow(me2,-1)*
         RS14e*RSe63 - 1./54.*m66*pow(me2,-1)*RS14e*RS34e - 1./27.*m66*
         pow(me2,-1)*RS14e*RSe61 + 1./108.*m66*pow(me2,-1)*pow(RS14e,2)
          + 2./27.*m66*RN3e*RS34e*RSe63 - 2./27.*m66*RN3e*RSe61*RS34e
          - 2./27.*m66*RN1e*RS14e*RSe63 + 2./27.*m66*RN1e*RS14e*RSe61
          - 1./36.*m44*pow(me2,-1) + 1./36.*m44*pow(me2,-1)*RSe6613 - 1.
         /36.*m44*pow(me2,-1)*RSe63 + 19./216.*m44*pow(me2,-1)*pow(
         RSe63,2) - 1./72.*m44*pow(me2,-1)*RSe6633 + 1./108.*m44*pow(
         me2,-1)*RS36e*RSe63 + 1./216.*m44*pow(me2,-1)*pow(RS36e,2) - 1.
         /36.*m44*pow(me2,-1)*RSe61 - 19./108.*m44*pow(me2,-1)*RSe61*
         RSe63 - 1./108.*m44*pow(me2,-1)*RSe61*RS36e + 19./216.*m44*
         pow(me2,-1)*pow(RSe61,2) - 1./72.*m44*pow(me2,-1)*RSe6611 - 1./
         108.*m44*pow(me2,-1)*RS16e*RSe63 - 1./108.*m44*pow(me2,-1)*
         RS16e*RS36e + 1./108.*m44*pow(me2,-1)*RS16e*RSe61 + 1./216.*
         m44*pow(me2,-1)*pow(RS16e,2) - 1./54.*m44*RN3e*RS36e*RSe63 + 1.
         /54.*m44*RN3e*RSe61*RS36e );

      F13p6x22R +=  + pow(Ab(me2,mu2),2) * ( 1./54.*m44*RN1e*RS16e*
         RSe63 - 1./54.*m44*RN1e*RS16e*RSe61 + 1./48.*m33*pow(me2,-1)*
         RSe6613 - 5./162.*m33*pow(me2,-1)*RSe6613*RS3461e - 1./324.*
         m33*pow(me2,-1)*RSe6613*RS1463e + 1./162.*m33*pow(me2,-1)*pow(
         RSe6613,2) - 5./48.*m33*pow(me2,-1)*RSe6633 + 23./324.*m33*
         pow(me2,-1)*RSe6633*RS3461e + 37./648.*m33*pow(me2,-1)*RSe6633
         *RS1463e - 19./162.*m33*pow(me2,-1)*RSe6633*RSe6613 + 73./648.
         *m33*pow(me2,-1)*pow(RSe6633,2) + 5./324.*m33*pow(me2,-1)*
         RC346e*RSe6613 - 41./648.*m33*pow(me2,-1)*RC346e*RSe6633 + 5./
         324.*m33*pow(me2,-1)*RSe6611*RS3461e + 1./648.*m33*pow(me2,-1)
         *RSe6611*RS1463e - 1./162.*m33*pow(me2,-1)*RSe6611*RSe6613 + 1.
         /324.*m33*pow(me2,-1)*RSe6611*RSe6633 - 5./648.*m33*pow(
         me2,-1)*RSe6611*RC346e + 1./648.*m33*pow(me2,-1)*pow(
         RSe6611,2) + 1./648.*m33*pow(me2,-1)*RC146e*RSe6613 - 1./1296.
         *m33*pow(me2,-1)*RC146e*RSe6633 - 1./1296.*m33*pow(me2,-1)*
         RC146e*RSe6611 );

      F13p6x22R +=  + pow(Ab(me2,mu2),2) * ( 2./9.*m33*RN3e*RSe6633*
         RSe6613 - 2./9.*m33*RN3e*pow(RSe6633,2) + 1./48.*m11*pow(
         me2,-1)*RSe6613 - 1./324.*m11*pow(me2,-1)*RSe6613*RS3461e - 5./
         162.*m11*pow(me2,-1)*RSe6613*RS1463e + 1./162.*m11*pow(me2,-1)
         *pow(RSe6613,2) + 1./648.*m11*pow(me2,-1)*RSe6633*RS3461e + 5./
         324.*m11*pow(me2,-1)*RSe6633*RS1463e - 1./162.*m11*pow(me2,-1)
         *RSe6633*RSe6613 + 1./648.*m11*pow(me2,-1)*pow(RSe6633,2) + 1./
         648.*m11*pow(me2,-1)*RC346e*RSe6613 - 1./1296.*m11*pow(me2,-1)
         *RC346e*RSe6633 - 5./48.*m11*pow(me2,-1)*RSe6611 + 37./648.*
         m11*pow(me2,-1)*RSe6611*RS3461e + 23./324.*m11*pow(me2,-1)*
         RSe6611*RS1463e - 19./162.*m11*pow(me2,-1)*RSe6611*RSe6613 + 1.
         /324.*m11*pow(me2,-1)*RSe6611*RSe6633 - 1./1296.*m11*pow(
         me2,-1)*RSe6611*RC346e + 73./648.*m11*pow(me2,-1)*pow(
         RSe6611,2) + 5./324.*m11*pow(me2,-1)*RC146e*RSe6613 - 5./648.*
         m11*pow(me2,-1)*RC146e*RSe6633 - 41./648.*m11*pow(me2,-1)*
         RC146e*RSe6611 );

      F13p6x22R +=  + pow(Ab(me2,mu2),2) * ( 2./9.*m11*RN1e*RSe6611*
         RSe6613 - 2./9.*m11*RN1e*pow(RSe6611,2) );

      F13p6x22R +=  + Ab(me2,mu2)*Bb(m11,mu2) * ( 1./36.*RD146e*RSe6613
          - 1./72.*RD146e*RSe6633 - 1./72.*RD146e*RSe6611 + 1./54.*m66*
         pow(RS14e,2) + 1./108.*m44*pow(RS16e,2) + 1./9.*m11*RSe6611*
         RS1463e - 1./9.*m11*RC146e*RSe6611 - 1./9.*m11*RN1e*RD146e*
         RSe6613 + 1./9.*m11*RN1e*RD146e*RSe6611 );

      F13p6x22R +=  + Ab(me2,mu2)*Bb(m14,mu2) * (  - 1./72.*me2 - 1./18.
         *me2*RSe61 - 1./18.*me2*RSe6611 + 1./72.*m44 - 1./18.*m44*
         RSe61 - 1./18.*m11*RSe61 + 1./18.*m11*RSe6611 );

      F13p6x22R +=  + Ab(me2,mu2)*Bb(m16,mu2) * (  - 1./36.*me2 + 1./18.
         *me2*RSe61 - 1./36.*me2*RSe6611 + 1./36.*m66 + 1./18.*m66*
         RSe61 + 1./18.*m11*RSe61 + 1./36.*m11*RSe6611 );

      F13p6x22R +=  + Ab(me2,mu2)*Bb(m33,mu2) * ( 1./36.*RD346e*RSe6613
          - 1./72.*RD346e*RSe6633 - 1./72.*RSe6611*RD346e + 1./54.*m66*
         pow(RS34e,2) + 1./108.*m44*pow(RS36e,2) + 1./9.*m33*RSe6633*
         RS3461e - 1./9.*m33*RC346e*RSe6633 - 1./9.*m33*RN3e*RD346e*
         RSe6613 + 1./9.*m33*RN3e*RD346e*RSe6633 );

      F13p6x22R +=  + Ab(me2,mu2)*Bb(m34,mu2) * (  - 1./72.*me2 - 1./18.
         *me2*RSe63 - 1./18.*me2*RSe6633 + 1./72.*m44 - 1./18.*m44*
         RSe63 - 1./18.*m33*RSe63 + 1./18.*m33*RSe6633 );

      F13p6x22R +=  + Ab(me2,mu2)*Bb(m36,mu2) * (  - 1./36.*me2 + 1./18.
         *me2*RSe63 - 1./36.*me2*RSe6633 + 1./36.*m66 + 1./18.*m66*
         RSe63 + 1./18.*m33*RSe63 + 1./36.*m33*RSe6633 );

      F13p6x22R +=  + Ab(me2,mu2)*Bb(me2,mu2) * ( 2./27.*m66*pow(
         RSe63,2) - 4./27.*m66*RSe61*RSe63 + 2./27.*m66*pow(RSe61,2) + 
         1./108.*m44*pow(RSe63,2) - 1./54.*m44*RSe61*RSe63 + 1./108.*
         m44*pow(RSe61,2) );

      F13p6x22R +=  + Ab(me2,mu2)*Cb(m11,mu2) * (  - 1./9.*m11*RD146e*
         RSe6611 );

      F13p6x22R +=  + Ab(me2,mu2)*Cb(m33,mu2) * (  - 1./9.*m33*RD346e*
         RSe6633 );

      F13p6x22R +=  + Bb(m11,mu2)*pi16 * (  - 1./144.*RD146e*RD346e - 1.
         /144.*pow(RD146e,2) - 1./36.*me2*RD146e*RSe6613 + 1./72.*me2*
         RD146e*RSe6633 + 1./72.*me2*RD146e*RSe6611 + 1./48.*m66*RD146e
          + 1./24.*m44*RD146e + 1./32.*m33*RD146e + 1./48.*m33*RD146e*
         RS3461e + 1./144.*m33*RD146e*RS1463e - 1./72.*m33*RD146e*
         RSe6613 + 1./144.*m33*RD146e*RSe6633 - 1./96.*m33*RD146e*
         RC346e + 1./144.*m33*RD146e*RSe6611 - 1./288.*m33*RD146e*
         RC146e + 1./96.*m11*RD146e + 1./144.*m11*RD146e*RS3461e + 1./
         48.*m11*RD146e*RS1463e - 1./72.*m11*RD146e*RSe6613 + 1./144.*
         m11*RD146e*RSe6633 - 1./288.*m11*RD146e*RC346e + 1./144.*m11*
         RD146e*RSe6611 - 1./96.*m11*RD146e*RC146e );

      F13p6x22R +=  + pow(Bb(m11,mu2),2) * ( 1./288.*pow(RD146e,2) - 1./
         18.*m11*RD146e*RS1463e + 1./18.*m11*RD146e*RC146e );

      F13p6x22R +=  + Bb(m11,mu2)*Bb(m33,mu2) * ( 1./144.*RD146e*RD346e
          );

      F13p6x22R +=  + Bb(m11,mu2)*Cb(m11,mu2) * ( 1./18.*m11*pow(
         RD146e,2) );

      F13p6x22R +=  + Bb(m33,mu2)*pi16 * (  - 1./144.*pow(RD346e,2) - 1.
         /144.*RD146e*RD346e - 1./36.*me2*RD346e*RSe6613 + 1./72.*me2*
         RD346e*RSe6633 + 1./72.*me2*RSe6611*RD346e + 1./48.*m66*RD346e
          + 1./24.*m44*RD346e + 1./96.*m33*RD346e + 1./48.*m33*RD346e*
         RS3461e + 1./144.*m33*RD346e*RS1463e - 1./72.*m33*RD346e*
         RSe6613 + 1./144.*m33*RD346e*RSe6633 - 1./96.*m33*RD346e*
         RC346e + 1./144.*m33*RSe6611*RD346e - 1./288.*m33*RC146e*
         RD346e + 1./32.*m11*RD346e + 1./144.*m11*RD346e*RS3461e + 1./
         48.*m11*RD346e*RS1463e - 1./72.*m11*RD346e*RSe6613 + 1./144.*
         m11*RD346e*RSe6633 - 1./288.*m11*RD346e*RC346e + 1./144.*m11*
         RSe6611*RD346e - 1./96.*m11*RC146e*RD346e );

      F13p6x22R +=  + pow(Bb(m33,mu2),2) * ( 1./288.*pow(RD346e,2) - 1./
         18.*m33*RD346e*RS3461e + 1./18.*m33*RD346e*RC346e );

      F13p6x22R +=  + Bb(m33,mu2)*Cb(m33,mu2) * ( 1./18.*m33*pow(
         RD346e,2) );

      F13p6x22R +=  + hh(1,m11,m11,m13,m13,mu2) * (  - 1./32.*m33 - 1./
         72.*m33*pow(RS1463e,2) + 1./72.*m33*RC146e*RS1463e - 1./288.*
         m33*pow(RC146e,2) + 3./32.*m11 - 1./72.*m11*pow(RS1463e,2) + 1.
         /72.*m11*RC146e*RS1463e - 1./288.*m11*pow(RC146e,2) );

      F13p6x22R +=  + hh(1,m11,m13,m33,m13,mu2) * (  - 1./36.*m33*
         RS1463e*RS3461e + 1./72.*m33*RC346e*RS1463e + 1./72.*m33*
         RC146e*RS3461e - 1./144.*m33*RC146e*RC346e - 1./36.*m11*
         RS1463e*RS3461e + 1./72.*m11*RC346e*RS1463e + 1./72.*m11*
         RC146e*RS3461e - 1./144.*m11*RC146e*RC346e );

      F13p6x22R +=  + hh(1,m11,m13,me2,m13,mu2) * ( 1./18.*m33*RSe6613*
         RS1463e - 1./36.*m33*RSe6633*RS1463e - 1./36.*m33*RSe6611*
         RS1463e - 1./36.*m33*RC146e*RSe6613 + 1./72.*m33*RC146e*
         RSe6633 + 1./72.*m33*RC146e*RSe6611 + 1./18.*m11*RSe6613*
         RS1463e - 1./36.*m11*RSe6633*RS1463e - 1./36.*m11*RSe6611*
         RS1463e - 1./36.*m11*RC146e*RSe6613 + 1./72.*m11*RC146e*
         RSe6633 + 1./72.*m11*RC146e*RSe6611 );

      F13p6x22R +=  + hh(1,m11,m14,m34,m13,mu2) * (  - 1./12.*m44*
         RS1463e + 1./24.*m44*RC146e - 1./24.*m33*RS1463e + 1./24.*m33*
         RS16e + 1./24.*m11*RS1463e - 1./8.*m11*RS16e + 1./24.*m11*
         RC146e );

      F13p6x22R +=  + hh(1,m11,m16,m36,m13,mu2) * (  - 1./24.*m66*
         RS1463e + 1./48.*m66*RC146e - 1./48.*m33*RS1463e + 1./48.*m33*
         RS14e + 1./48.*m11*RS1463e - 1./16.*m11*RS14e + 1./48.*m11*
         RC146e );

      F13p6x22R +=  + hh(1,m13,m13,m13,m13,mu2) * (  - 1./16.*m33 - 1./
         16.*m11 );

      F13p6x22R +=  + hh(1,m13,m33,m33,m13,mu2) * ( 3./32.*m33 - 1./72.
         *m33*pow(RS3461e,2) + 1./72.*m33*RC346e*RS3461e - 1./288.*m33*
         pow(RC346e,2) - 1./32.*m11 - 1./72.*m11*pow(RS3461e,2) + 1./72.
         *m11*RC346e*RS3461e - 1./288.*m11*pow(RC346e,2) );

      F13p6x22R +=  + hh(1,m13,m33,me2,m13,mu2) * ( 1./18.*m33*RSe6613*
         RS3461e - 1./36.*m33*RSe6633*RS3461e - 1./36.*m33*RC346e*
         RSe6613 + 1./72.*m33*RC346e*RSe6633 - 1./36.*m33*RSe6611*
         RS3461e + 1./72.*m33*RSe6611*RC346e + 1./18.*m11*RSe6613*
         RS3461e - 1./36.*m11*RSe6633*RS3461e - 1./36.*m11*RC346e*
         RSe6613 + 1./72.*m11*RC346e*RSe6633 - 1./36.*m11*RSe6611*
         RS3461e + 1./72.*m11*RSe6611*RC346e );

      F13p6x22R +=  + hh(1,m13,me2,me2,m13,mu2) * (  - 1./2916.*m13*
         pow(RN1e,4)*pow(RN3e,4)*pow(RD13,4)*pow(RD46,4) );

      F13p6x22R +=  + hh(1,m14,m33,m34,m13,mu2) * (  - 1./12.*m44*
         RS3461e + 1./24.*m44*RC346e + 1./24.*m33*RS3461e - 1./8.*m33*
         RS36e + 1./24.*m33*RC346e - 1./24.*m11*RS3461e + 1./24.*m11*
         RS36e );

      F13p6x22R +=  + hh(1,m14,m34,m44,m13,mu2) * (  - 3./8.*m44 );

      F13p6x22R +=  + hh(1,m14,m34,me2,m13,mu2) * (  - 1./24.*me2 - 1./
         6.*me2*RSe6613 - 1./12.*me2*RSe63 - 1./12.*me2*RSe61 + 1./6.*
         m44*RSe6613 - 1./12.*m44*RSe6633 - 1./12.*m44*RSe6611 + 1./12.
         *m33*RSe6613 - 1./24.*m33*RSe63 - 1./12.*m33*RSe6633 + 1./24.*
         m33*RSe61 + 1./12.*m11*RSe6613 + 1./24.*m11*RSe63 - 1./24.*m11
         *RSe61 - 1./12.*m11*RSe6611 );

      F13p6x22R +=  + hh(1,m14,m36,m46,m13,mu2) * (  - 1./8.*m66 - 1./8.
         *m44 );

      F13p6x22R +=  + hh(1,m16,m33,m36,m13,mu2) * (  - 1./24.*m66*
         RS3461e + 1./48.*m66*RC346e + 1./48.*m33*RS3461e - 1./16.*m33*
         RS34e + 1./48.*m33*RC346e - 1./48.*m11*RS3461e + 1./48.*m11*
         RS34e );

      F13p6x22R +=  + hh(1,m16,m34,m46,m13,mu2) * (  - 1./8.*m66 - 1./8.
         *m44 );

      F13p6x22R +=  + hh(1,m16,m36,me2,m13,mu2) * (  - 1./12.*me2 - 1./
         12.*me2*RSe6613 + 1./12.*me2*RSe63 + 1./12.*me2*RSe61 + 1./12.
         *m66*RSe6613 - 1./24.*m66*RSe6633 - 1./24.*m66*RSe6611 + 1./24.
         *m33*RSe6613 + 1./24.*m33*RSe63 - 1./24.*m33*RSe6633 - 1./24.*
         m33*RSe61 + 1./24.*m11*RSe6613 - 1./24.*m11*RSe63 + 1./24.*m11
         *RSe61 - 1./24.*m11*RSe6611 );

      F13p6x22R +=  + hh(2,m11,m11,m13,m13,mu2) * ( 1./72.*m33*RD146e*
         RS1463e - 1./144.*m33*RD146e*RC146e + 1./72.*m11*RD146e*
         RS1463e - 1./144.*m11*RD146e*RC146e );

      F13p6x22R +=  + hh(2,m11,m13,m33,m13,mu2) * ( 1./72.*m33*RD146e*
         RS3461e - 1./144.*m33*RD146e*RC346e + 1./72.*m11*RD146e*
         RS3461e - 1./144.*m11*RD146e*RC346e );

      F13p6x22R +=  + hh(2,m11,m13,me2,m13,mu2) * (  - 1./36.*m33*
         RD146e*RSe6613 + 1./72.*m33*RD146e*RSe6633 + 1./72.*m33*RD146e
         *RSe6611 - 1./36.*m11*RD146e*RSe6613 + 1./72.*m11*RD146e*
         RSe6633 + 1./72.*m11*RD146e*RSe6611 );

      F13p6x22R +=  + hh(2,m11,m14,m34,m13,mu2) * ( 1./24.*m44*RD146e
          + 1./24.*m11*RD146e );

      F13p6x22R +=  + hh(2,m11,m16,m36,m13,mu2) * ( 1./48.*m66*RD146e
          + 1./48.*m11*RD146e );

      F13p6x22R +=  + hh(2,m33,m11,m13,m13,mu2) * ( 1./72.*m33*RD346e*
         RS1463e - 1./144.*m33*RC146e*RD346e + 1./72.*m11*RD346e*
         RS1463e - 1./144.*m11*RC146e*RD346e );

      F13p6x22R +=  + hh(2,m33,m13,m33,m13,mu2) * ( 1./72.*m33*RD346e*
         RS3461e - 1./144.*m33*RD346e*RC346e + 1./72.*m11*RD346e*
         RS3461e - 1./144.*m11*RD346e*RC346e );

      F13p6x22R +=  + hh(2,m33,m13,me2,m13,mu2) * (  - 1./36.*m33*
         RD346e*RSe6613 + 1./72.*m33*RD346e*RSe6633 + 1./72.*m33*
         RSe6611*RD346e - 1./36.*m11*RD346e*RSe6613 + 1./72.*m11*RD346e
         *RSe6633 + 1./72.*m11*RSe6611*RD346e );

      F13p6x22R +=  + hh(2,m33,m14,m34,m13,mu2) * ( 1./24.*m44*RD346e
          + 1./24.*m33*RD346e );

      F13p6x22R +=  + hh(2,m33,m16,m36,m13,mu2) * ( 1./48.*m66*RD346e
          + 1./48.*m33*RD346e );

      F13p6x22R +=  + hh(5,m11,m11,m13,m13,mu2) * (  - 1./288.*m33*pow(
         RD146e,2) - 1./288.*m11*pow(RD146e,2) );

      F13p6x22R +=  + hh(5,m11,m33,m13,m13,mu2) * (  - 1./144.*m33*
         RD146e*RD346e - 1./144.*m11*RD146e*RD346e );

      F13p6x22R +=  + hh(5,m33,m33,m13,m13,mu2) * (  - 1./288.*m33*pow(
         RD346e,2) - 1./288.*m11*pow(RD346e,2) );

      F13p6x22R +=  + hhd(1,m11,m11,m13,m13,mu2) * ( 1./648.*pow(
         RD146e,2) + 1./81.*m33*RD146e*RS1463e - 1./108.*m33*RD146e*
         RC146e + 5./64.*pow(m33,2) + 41./1296.*pow(m33,2)*pow(
         RS1463e,2) - 7./432.*pow(m33,2)*RC146e*RS1463e + 5./576.*pow(
         m33,2)*pow(RC146e,2) + 1./162.*m11*RD146e*RS1463e - 1./108.*
         m11*RD146e*RC146e + 3./32.*m11*m33 + 25./648.*m11*m33*pow(
         RS1463e,2) - 1./72.*m11*m33*RC146e*RS1463e + 5./288.*m11*m33*
         pow(RC146e,2) + 1./64.*pow(m11,2) + 17./1296.*pow(m11,2)*pow(
         RS1463e,2) + 1./432.*pow(m11,2)*RC146e*RS1463e + 5./576.*pow(
         m11,2)*pow(RC146e,2) );

      F13p6x22R +=  + hhd(1,m11,m13,m33,m13,mu2) * ( 1./324.*RD146e*
         RD346e + 1./81.*m33*RD346e*RS1463e + 1./162.*m33*RD146e*
         RS3461e + 25./648.*pow(m33,2)*RS1463e*RS3461e - 1./144.*pow(
         m33,2)*RC346e*RS1463e - 1./144.*pow(m33,2)*RC146e*RS3461e + 1./
         288.*pow(m33,2)*RC146e*RC346e + 1./162.*m11*RD346e*RS1463e + 1.
         /81.*m11*RD146e*RS3461e + 29./324.*m11*m33*RS1463e*RS3461e - 1.
         /72.*m11*m33*RC346e*RS1463e - 1./72.*m11*m33*RC146e*RS3461e + 
         1./144.*m11*m33*RC146e*RC346e + 25./648.*pow(m11,2)*RS1463e*
         RS3461e - 1./144.*pow(m11,2)*RC346e*RS1463e - 1./144.*pow(
         m11,2)*RC146e*RS3461e + 1./288.*pow(m11,2)*RC146e*RC346e );

      F13p6x22R +=  + hhd(1,m11,m13,me2,m13,mu2) * ( 1./81.*me2*RD146e*
         RSe6613 - 1./162.*me2*RD146e*RSe6633 - 1./162.*me2*RD146e*
         RSe6611 - 1./81.*pow(me2,2)*RSe6613*RS1463e + 1./162.*pow(
         me2,2)*RSe6633*RS1463e + 1./162.*pow(me2,2)*RSe6611*RS1463e + 
         1./162.*pow(me2,2)*RC146e*RSe6613 - 1./324.*pow(me2,2)*RC146e*
         RSe6633 - 1./324.*pow(me2,2)*RC146e*RSe6611 - 2./81.*m33*
         RD146e*RSe6613 + 1./162.*m33*RD146e*RSe6633 + 1./54.*m33*
         RD146e*RSe6611 + 5./81.*m33*me2*RSe6613*RS1463e - 1./27.*m33*
         me2*RSe6633*RS1463e - 2./81.*m33*me2*RSe6611*RS1463e - 2./81.*
         m33*me2*RC146e*RSe6613 + 1./162.*m33*me2*RC146e*RSe6633 + 1./
         54.*m33*me2*RC146e*RSe6611 - 41./324.*pow(m33,2)*RSe6613*
         RS1463e + 29./648.*pow(m33,2)*RSe6633*RS1463e + 7./216.*pow(
         m33,2)*RSe6611*RS1463e + 7./216.*pow(m33,2)*RC146e*RSe6613 - 
         13./1296.*pow(m33,2)*RC146e*RSe6633 - 5./144.*pow(m33,2)*
         RC146e*RSe6611 - 2./81.*m11*RD146e*RSe6613 + 2./81.*m11*RD146e
         *RSe6611 );

      F13p6x22R +=  + hhd(1,m11,m13,me2,m13,mu2) * ( 1./27.*m11*me2*
         RSe6613*RS1463e - 1./81.*m11*me2*RSe6633*RS1463e - 2./81.*m11*
         me2*RSe6611*RS1463e - 2./81.*m11*me2*RC146e*RSe6613 + 2./81.*
         m11*me2*RC146e*RSe6611 - 35./162.*m11*m33*RSe6613*RS1463e + 13.
         /324.*m11*m33*RSe6633*RS1463e + 17./324.*m11*m33*RSe6611*
         RS1463e + 17./324.*m11*m33*RC146e*RSe6613 - 1./72.*m11*m33*
         RC146e*RSe6633 - 19./216.*m11*m33*RC146e*RSe6611 - 25./324.*
         pow(m11,2)*RSe6613*RS1463e + 1./72.*pow(m11,2)*RSe6633*RS1463e
          + 1./72.*pow(m11,2)*RSe6611*RS1463e + 1./72.*pow(m11,2)*
         RC146e*RSe6613 - 1./144.*pow(m11,2)*RC146e*RSe6633 - 73./1296.
         *pow(m11,2)*RC146e*RSe6611 );

      F13p6x22R +=  + hhd(1,m11,m14,m34,m13,mu2) * ( 1./24.*m33*m44*
         RS1463e - 1./48.*m33*m44*RC146e - 1./16.*pow(m33,2)*RS1463e + 
         1./16.*pow(m33,2)*RS16e + 1./24.*m11*m44*RS1463e - 1./48.*m11*
         m44*RC146e - 1./6.*m11*m33*RS1463e + 5./24.*m11*m33*RS16e - 1./
         48.*m11*m33*RC146e - 5./48.*pow(m11,2)*RS1463e + 7./48.*pow(
         m11,2)*RS16e - 1./48.*pow(m11,2)*RC146e );

      F13p6x22R +=  + hhd(1,m11,m16,m36,m13,mu2) * ( 1./48.*m33*m66*
         RS1463e - 1./96.*m33*m66*RC146e - 1./32.*pow(m33,2)*RS1463e + 
         1./32.*pow(m33,2)*RS14e + 1./48.*m11*m66*RS1463e - 1./96.*m11*
         m66*RC146e - 1./12.*m11*m33*RS1463e + 5./48.*m11*m33*RS14e - 1.
         /96.*m11*m33*RC146e - 5./96.*pow(m11,2)*RS1463e + 7./96.*pow(
         m11,2)*RS14e - 1./96.*pow(m11,2)*RC146e );

      F13p6x22R +=  + hhd(1,m13,m13,m13,m13,mu2) * ( 1./32.*pow(m33,2)
          + 1./16.*m11*m33 + 1./32.*pow(m11,2) );

      F13p6x22R +=  + hhd(1,m13,m33,m33,m13,mu2) * ( 1./648.*pow(
         RD346e,2) + 1./162.*m33*RD346e*RS3461e - 1./108.*m33*RD346e*
         RC346e + 1./64.*pow(m33,2) + 17./1296.*pow(m33,2)*pow(
         RS3461e,2) + 1./432.*pow(m33,2)*RC346e*RS3461e + 5./576.*pow(
         m33,2)*pow(RC346e,2) + 1./81.*m11*RD346e*RS3461e - 1./108.*m11
         *RD346e*RC346e + 3./32.*m11*m33 + 25./648.*m11*m33*pow(
         RS3461e,2) - 1./72.*m11*m33*RC346e*RS3461e + 5./288.*m11*m33*
         pow(RC346e,2) + 5./64.*pow(m11,2) + 41./1296.*pow(m11,2)*pow(
         RS3461e,2) - 7./432.*pow(m11,2)*RC346e*RS3461e + 5./576.*pow(
         m11,2)*pow(RC346e,2) );

      F13p6x22R +=  + hhd(1,m13,m33,me2,m13,mu2) * ( 1./81.*me2*RD346e*
         RSe6613 - 1./162.*me2*RD346e*RSe6633 - 1./162.*me2*RSe6611*
         RD346e - 1./81.*pow(me2,2)*RSe6613*RS3461e + 1./162.*pow(
         me2,2)*RSe6633*RS3461e + 1./162.*pow(me2,2)*RC346e*RSe6613 - 1.
         /324.*pow(me2,2)*RC346e*RSe6633 + 1./162.*pow(me2,2)*RSe6611*
         RS3461e - 1./324.*pow(me2,2)*RSe6611*RC346e - 2./81.*m33*
         RD346e*RSe6613 + 2./81.*m33*RD346e*RSe6633 + 1./27.*m33*me2*
         RSe6613*RS3461e - 2./81.*m33*me2*RSe6633*RS3461e - 2./81.*m33*
         me2*RC346e*RSe6613 + 2./81.*m33*me2*RC346e*RSe6633 - 1./81.*
         m33*me2*RSe6611*RS3461e - 25./324.*pow(m33,2)*RSe6613*RS3461e
          + 1./72.*pow(m33,2)*RSe6633*RS3461e + 1./72.*pow(m33,2)*
         RC346e*RSe6613 - 73./1296.*pow(m33,2)*RC346e*RSe6633 + 1./72.*
         pow(m33,2)*RSe6611*RS3461e - 1./144.*pow(m33,2)*RSe6611*RC346e
          - 2./81.*m11*RD346e*RSe6613 + 1./54.*m11*RD346e*RSe6633 + 1./
         162.*m11*RSe6611*RD346e + 5./81.*m11*me2*RSe6613*RS3461e - 2./
         81.*m11*me2*RSe6633*RS3461e );

      F13p6x22R +=  + hhd(1,m13,m33,me2,m13,mu2) * (  - 2./81.*m11*me2*
         RC346e*RSe6613 + 1./54.*m11*me2*RC346e*RSe6633 - 1./27.*m11*
         me2*RSe6611*RS3461e + 1./162.*m11*me2*RSe6611*RC346e - 35./162.
         *m11*m33*RSe6613*RS3461e + 17./324.*m11*m33*RSe6633*RS3461e + 
         17./324.*m11*m33*RC346e*RSe6613 - 19./216.*m11*m33*RC346e*
         RSe6633 + 13./324.*m11*m33*RSe6611*RS3461e - 1./72.*m11*m33*
         RSe6611*RC346e - 41./324.*pow(m11,2)*RSe6613*RS3461e + 7./216.
         *pow(m11,2)*RSe6633*RS3461e + 7./216.*pow(m11,2)*RC346e*
         RSe6613 - 5./144.*pow(m11,2)*RC346e*RSe6633 + 29./648.*pow(
         m11,2)*RSe6611*RS3461e - 13./1296.*pow(m11,2)*RSe6611*RC346e )
         ;

      F13p6x22R +=  + hhd(1,m13,me2,me2,m13,mu2) * ( 4./81.*pow(me2,2)*
         pow(RSe6613,2) - 4./81.*pow(me2,2)*RSe6633*RSe6613 + 1./81.*
         pow(me2,2)*pow(RSe6633,2) - 4./81.*pow(me2,2)*RSe6611*RSe6613
          + 2./81.*pow(me2,2)*RSe6611*RSe6633 + 1./81.*pow(me2,2)*pow(
         RSe6611,2) - 10./81.*m33*me2*pow(RSe6613,2) + 4./27.*m33*me2*
         RSe6633*RSe6613 - 5./81.*m33*me2*pow(RSe6633,2) + 8./81.*m33*
         me2*RSe6611*RSe6613 - 2./81.*m33*me2*RSe6611*RSe6633 - 1./27.*
         m33*me2*pow(RSe6611,2) + 41./324.*pow(m33,2)*pow(RSe6613,2) - 
         29./324.*pow(m33,2)*RSe6633*RSe6613 + 109./1296.*pow(m33,2)*
         pow(RSe6633,2) - 7./108.*pow(m33,2)*RSe6611*RSe6613 + 13./648.
         *pow(m33,2)*RSe6611*RSe6633 + 5./144.*pow(m33,2)*pow(
         RSe6611,2) - 10./81.*m11*me2*pow(RSe6613,2) + 8./81.*m11*me2*
         RSe6633*RSe6613 - 1./27.*m11*me2*pow(RSe6633,2) + 4./27.*m11*
         me2*RSe6611*RSe6613 - 2./81.*m11*me2*RSe6611*RSe6633 - 5./81.*
         m11*me2*pow(RSe6611,2) + 5./18.*m11*m33*pow(RSe6613,2) - 25./
         162.*m11*m33*RSe6633*RSe6613 );

      F13p6x22R +=  + hhd(1,m13,me2,me2,m13,mu2) * ( 23./216.*m11*m33*
         pow(RSe6633,2) - 25./162.*m11*m33*RSe6611*RSe6613 + 13./324.*
         m11*m33*RSe6611*RSe6633 + 23./216.*m11*m33*pow(RSe6611,2) + 41.
         /324.*pow(m11,2)*pow(RSe6613,2) - 7./108.*pow(m11,2)*RSe6633*
         RSe6613 + 5./144.*pow(m11,2)*pow(RSe6633,2) - 29./324.*pow(
         m11,2)*RSe6611*RSe6613 + 13./648.*pow(m11,2)*RSe6611*RSe6633
          + 109./1296.*pow(m11,2)*pow(RSe6611,2) );

      F13p6x22R +=  + hhd(1,m14,m33,m34,m13,mu2) * ( 1./24.*m33*m44*
         RS3461e - 1./48.*m33*m44*RC346e + 1./16.*pow(m33,2)*RS3461e - 
         1./48.*pow(m33,2)*RS36e - 1./48.*pow(m33,2)*RC346e + 1./24.*
         m11*m44*RS3461e - 1./48.*m11*m44*RC346e + 1./6.*m11*m33*
         RS3461e - 1./8.*m11*m33*RS36e - 1./48.*m11*m33*RC346e + 5./48.
         *pow(m11,2)*RS3461e - 5./48.*pow(m11,2)*RS36e );

      F13p6x22R +=  + hhd(1,m14,m34,m44,m13,mu2) * ( 3./16.*m33*m44 + 3.
         /16.*m11*m44 );

      F13p6x22R +=  + hhd(1,m14,m34,me2,m13,mu2) * ( 1./54.*pow(me2,2)
          + 1./27.*pow(me2,2)*RSe6613 + 1./27.*pow(me2,2)*RSe63 + 1./54.
         *pow(me2,2)*RSe6633 + 1./27.*pow(me2,2)*RSe61 + 1./54.*pow(
         me2,2)*RSe6611 - 1./27.*m44*me2 - 1./27.*m44*me2*RSe63 - 1./27.
         *m44*me2*RSe61 + 1./54.*pow(m44,2) + 1./48.*m33*me2 + 17./108.
         *m33*me2*RSe6613 + 25./216.*m33*me2*RSe63 + 2./27.*m33*me2*
         RSe6633 + 1./24.*m33*me2*RSe61 - 1./12.*m33*m44*RSe6613 - 2./
         27.*m33*m44*RSe63 + 1./24.*m33*m44*RSe6633 + 1./24.*m33*m44*
         RSe6611 + 1./8.*pow(m33,2)*RSe6613 - 1./16.*pow(m33,2)*RSe63
          - 11./216.*pow(m33,2)*RSe6633 + 1./16.*pow(m33,2)*RSe61 + 1./
         48.*m11*me2 + 17./108.*m11*me2*RSe6613 + 11./72.*m11*me2*RSe63
          + 1./9.*m11*me2*RSe6633 + 1./216.*m11*me2*RSe61 - 1./27.*m11*
         me2*RSe6611 - 1./12.*m11*m44*RSe6613 - 1./9.*m11*m44*RSe63 + 1.
         /24.*m11*m44*RSe6633 + 1./27.*m11*m44*RSe61 + 1./24.*m11*m44*
         RSe6611 + 19./108.*m11*m33*RSe6613 - 1./6.*m11*m33*RSe63 - 5./
         72.*m11*m33*RSe6633 );

      F13p6x22R +=  + hhd(1,m14,m34,me2,m13,mu2) * ( 1./6.*m11*m33*
         RSe61 + 1./24.*m11*m33*RSe6611 + 1./72.*pow(m11,2)*RSe6613 - 5.
         /48.*pow(m11,2)*RSe63 + 5./48.*pow(m11,2)*RSe61 + 13./216.*
         pow(m11,2)*RSe6611 );

      F13p6x22R +=  + hhd(1,m14,m36,m46,m13,mu2) * ( 1./16.*m33*m66 + 1.
         /16.*m33*m44 + 1./16.*m11*m66 + 1./16.*m11*m44 );

      F13p6x22R +=  + hhd(1,m16,m33,m36,m13,mu2) * ( 1./48.*m33*m66*
         RS3461e - 1./96.*m33*m66*RC346e + 1./32.*pow(m33,2)*RS3461e - 
         1./96.*pow(m33,2)*RS34e - 1./96.*pow(m33,2)*RC346e + 1./48.*
         m11*m66*RS3461e - 1./96.*m11*m66*RC346e + 1./12.*m11*m33*
         RS3461e - 1./16.*m11*m33*RS34e - 1./96.*m11*m33*RC346e + 5./96.
         *pow(m11,2)*RS3461e - 5./96.*pow(m11,2)*RS34e );

      F13p6x22R +=  + hhd(1,m16,m34,m46,m13,mu2) * ( 1./16.*m33*m66 + 1.
         /16.*m33*m44 + 1./16.*m11*m66 + 1./16.*m11*m44 );

      F13p6x22R +=  + hhd(1,m16,m36,me2,m13,mu2) * ( 1./27.*pow(me2,2)
          + 1./54.*pow(me2,2)*RSe6613 - 1./27.*pow(me2,2)*RSe63 + 1./
         108.*pow(me2,2)*RSe6633 - 1./27.*pow(me2,2)*RSe61 + 1./108.*
         pow(me2,2)*RSe6611 - 2./27.*m66*me2 + 1./27.*m66*me2*RSe63 + 1.
         /27.*m66*me2*RSe61 + 1./27.*pow(m66,2) + 1./24.*m33*me2 + 17./
         216.*m33*me2*RSe6613 - 25./216.*m33*me2*RSe63 + 1./27.*m33*me2
         *RSe6633 - 1./24.*m33*me2*RSe61 - 1./24.*m33*m66*RSe6613 + 2./
         27.*m33*m66*RSe63 + 1./48.*m33*m66*RSe6633 + 1./48.*m33*m66*
         RSe6611 + 1./16.*pow(m33,2)*RSe6613 + 1./16.*pow(m33,2)*RSe63
          - 11./432.*pow(m33,2)*RSe6633 - 1./16.*pow(m33,2)*RSe61 + 1./
         24.*m11*me2 + 17./216.*m11*me2*RSe6613 - 11./72.*m11*me2*RSe63
          + 1./18.*m11*me2*RSe6633 - 1./216.*m11*me2*RSe61 - 1./54.*m11
         *me2*RSe6611 - 1./24.*m11*m66*RSe6613 + 1./9.*m11*m66*RSe63 + 
         1./48.*m11*m66*RSe6633 - 1./27.*m11*m66*RSe61 + 1./48.*m11*m66
         *RSe6611 + 19./216.*m11*m33*RSe6613 + 1./6.*m11*m33*RSe63 - 5./
         144.*m11*m33*RSe6633 );

      F13p6x22R +=  + hhd(1,m16,m36,me2,m13,mu2) * (  - 1./6.*m11*m33*
         RSe61 + 1./48.*m11*m33*RSe6611 + 1./144.*pow(m11,2)*RSe6613 + 
         5./48.*pow(m11,2)*RSe63 - 5./48.*pow(m11,2)*RSe61 + 13./432.*
         pow(m11,2)*RSe6611 );

      F13p6x22R +=  + hhd(2,m11,m11,m13,m13,mu2) * (  - 1./108.*m33*
         pow(RD146e,2) - 7./432.*pow(m33,2)*RD146e*RS1463e + 5./288.*
         pow(m33,2)*RD146e*RC146e - 1./108.*m11*pow(RD146e,2) - 1./72.*
         m11*m33*RD146e*RS1463e + 5./144.*m11*m33*RD146e*RC146e + 1./
         432.*pow(m11,2)*RD146e*RS1463e + 5./288.*pow(m11,2)*RD146e*
         RC146e );

      F13p6x22R +=  + hhd(2,m11,m13,m33,m13,mu2) * (  - 1./144.*pow(
         m33,2)*RD146e*RS3461e + 1./288.*pow(m33,2)*RD146e*RC346e - 1./
         72.*m11*m33*RD146e*RS3461e + 1./144.*m11*m33*RD146e*RC346e - 1.
         /144.*pow(m11,2)*RD146e*RS3461e + 1./288.*pow(m11,2)*RD146e*
         RC346e );

      F13p6x22R +=  + hhd(2,m11,m13,me2,m13,mu2) * ( 1./162.*pow(me2,2)
         *RD146e*RSe6613 - 1./324.*pow(me2,2)*RD146e*RSe6633 - 1./324.*
         pow(me2,2)*RD146e*RSe6611 - 2./81.*m33*me2*RD146e*RSe6613 + 1./
         162.*m33*me2*RD146e*RSe6633 + 1./54.*m33*me2*RD146e*RSe6611 + 
         7./216.*pow(m33,2)*RD146e*RSe6613 - 13./1296.*pow(m33,2)*
         RD146e*RSe6633 - 5./144.*pow(m33,2)*RD146e*RSe6611 - 2./81.*
         m11*me2*RD146e*RSe6613 + 2./81.*m11*me2*RD146e*RSe6611 + 17./
         324.*m11*m33*RD146e*RSe6613 - 1./72.*m11*m33*RD146e*RSe6633 - 
         19./216.*m11*m33*RD146e*RSe6611 + 1./72.*pow(m11,2)*RD146e*
         RSe6613 - 1./144.*pow(m11,2)*RD146e*RSe6633 - 73./1296.*pow(
         m11,2)*RD146e*RSe6611 );

      F13p6x22R +=  + hhd(2,m11,m14,m34,m13,mu2) * (  - 1./48.*m33*m44*
         RD146e - 1./48.*m11*m44*RD146e - 1./48.*m11*m33*RD146e - 1./48.
         *pow(m11,2)*RD146e );

      F13p6x22R +=  + hhd(2,m11,m16,m36,m13,mu2) * (  - 1./96.*m33*m66*
         RD146e - 1./96.*m11*m66*RD146e - 1./96.*m11*m33*RD146e - 1./96.
         *pow(m11,2)*RD146e );

      F13p6x22R +=  + hhd(2,m33,m11,m13,m13,mu2) * (  - 1./144.*pow(
         m33,2)*RD346e*RS1463e + 1./288.*pow(m33,2)*RC146e*RD346e - 1./
         72.*m11*m33*RD346e*RS1463e + 1./144.*m11*m33*RC146e*RD346e - 1.
         /144.*pow(m11,2)*RD346e*RS1463e + 1./288.*pow(m11,2)*RC146e*
         RD346e );

      F13p6x22R +=  + hhd(2,m33,m13,m33,m13,mu2) * (  - 1./108.*m33*
         pow(RD346e,2) + 1./432.*pow(m33,2)*RD346e*RS3461e + 5./288.*
         pow(m33,2)*RD346e*RC346e - 1./108.*m11*pow(RD346e,2) - 1./72.*
         m11*m33*RD346e*RS3461e + 5./144.*m11*m33*RD346e*RC346e - 7./
         432.*pow(m11,2)*RD346e*RS3461e + 5./288.*pow(m11,2)*RD346e*
         RC346e );

      F13p6x22R +=  + hhd(2,m33,m13,me2,m13,mu2) * ( 1./162.*pow(me2,2)
         *RD346e*RSe6613 - 1./324.*pow(me2,2)*RD346e*RSe6633 - 1./324.*
         pow(me2,2)*RSe6611*RD346e - 2./81.*m33*me2*RD346e*RSe6613 + 2./
         81.*m33*me2*RD346e*RSe6633 + 1./72.*pow(m33,2)*RD346e*RSe6613
          - 73./1296.*pow(m33,2)*RD346e*RSe6633 - 1./144.*pow(m33,2)*
         RSe6611*RD346e - 2./81.*m11*me2*RD346e*RSe6613 + 1./54.*m11*
         me2*RD346e*RSe6633 + 1./162.*m11*me2*RSe6611*RD346e + 17./324.
         *m11*m33*RD346e*RSe6613 - 19./216.*m11*m33*RD346e*RSe6633 - 1./
         72.*m11*m33*RSe6611*RD346e + 7./216.*pow(m11,2)*RD346e*RSe6613
          - 5./144.*pow(m11,2)*RD346e*RSe6633 - 13./1296.*pow(m11,2)*
         RSe6611*RD346e );

      F13p6x22R +=  + hhd(2,m33,m14,m34,m13,mu2) * (  - 1./48.*m33*m44*
         RD346e - 1./48.*pow(m33,2)*RD346e - 1./48.*m11*m44*RD346e - 1./
         48.*m11*m33*RD346e );

      F13p6x22R +=  + hhd(2,m33,m16,m36,m13,mu2) * (  - 1./96.*m33*m66*
         RD346e - 1./96.*pow(m33,2)*RD346e - 1./96.*m11*m66*RD346e - 1./
         96.*m11*m33*RD346e );

      F13p6x22R +=  + hhd(5,m11,m11,m13,m13,mu2) * ( 5./576.*pow(m33,2)
         *pow(RD146e,2) + 5./288.*m11*m33*pow(RD146e,2) + 5./576.*pow(
         m11,2)*pow(RD146e,2) );

      F13p6x22R +=  + hhd(5,m11,m33,m13,m13,mu2) * ( 1./288.*pow(m33,2)
         *RD146e*RD346e + 1./144.*m11*m33*RD146e*RD346e + 1./288.*pow(
         m11,2)*RD146e*RD346e );

      F13p6x22R +=  + hhd(5,m33,m33,m13,m13,mu2) * ( 5./576.*pow(m33,2)
         *pow(RD346e,2) + 5./288.*m11*m33*pow(RD346e,2) + 5./576.*pow(
         m11,2)*pow(RD346e,2) );

      F13p6x22R +=  + hh1d(1,m11,m14,m34,m13,mu2) * ( 1./6.*pow(m33,2)*
         RS1463e - 1./6.*pow(m33,2)*RS16e + 1./3.*m11*m33*RS1463e - 1./
         3.*m11*m33*RS16e + 1./6.*pow(m11,2)*RS1463e - 1./6.*pow(m11,2)
         *RS16e );

      F13p6x22R +=  + hh1d(1,m11,m16,m36,m13,mu2) * ( 1./12.*pow(m33,2)
         *RS1463e - 1./12.*pow(m33,2)*RS14e + 1./6.*m11*m33*RS1463e - 1.
         /6.*m11*m33*RS14e + 1./12.*pow(m11,2)*RS1463e - 1./12.*pow(
         m11,2)*RS14e );

      F13p6x22R +=  + hh1d(1,m13,m11,m11,m13,mu2) * (  - 1./54.*m33*
         RD146e*RS1463e + 1./108.*m33*RD146e*RC146e - 1./8.*pow(m33,2)
          - 5./108.*pow(m33,2)*pow(RS1463e,2) + 1./27.*pow(m33,2)*
         RC146e*RS1463e - 1./72.*pow(m33,2)*pow(RC146e,2) - 1./54.*m11*
         RD146e*RS1463e + 1./108.*m11*RD146e*RC146e - 1./4.*m11*m33 - 1.
         /18.*m11*m33*pow(RS1463e,2) + 1./18.*m11*m33*RC146e*RS1463e - 
         1./36.*m11*m33*pow(RC146e,2) - 1./8.*pow(m11,2) - 1./108.*pow(
         m11,2)*pow(RS1463e,2) + 1./54.*pow(m11,2)*RC146e*RS1463e - 1./
         72.*pow(m11,2)*pow(RC146e,2) );

      F13p6x22R +=  + hh1d(1,m13,m11,m33,m13,mu2) * (  - 1./54.*m33*
         RD346e*RS1463e + 1./108.*m33*RC146e*RD346e - 1./54.*m33*RD146e
         *RS3461e + 1./108.*m33*RD146e*RC346e - 1./18.*pow(m33,2)*
         RS1463e*RS3461e + 1./27.*pow(m33,2)*RC346e*RS1463e + 1./54.*
         pow(m33,2)*RC146e*RS3461e - 1./54.*m11*RD346e*RS1463e + 1./108.
         *m11*RC146e*RD346e - 1./54.*m11*RD146e*RS3461e + 1./108.*m11*
         RD146e*RC346e - 1./9.*m11*m33*RS1463e*RS3461e + 1./18.*m11*m33
         *RC346e*RS1463e + 1./18.*m11*m33*RC146e*RS3461e - 1./18.*pow(
         m11,2)*RS1463e*RS3461e + 1./54.*pow(m11,2)*RC346e*RS1463e + 1./
         27.*pow(m11,2)*RC146e*RS3461e );

      F13p6x22R +=  + hh1d(1,m13,m11,me2,m13,mu2) * ( 1./27.*m33*RD146e
         *RSe6613 - 1./54.*m33*RD146e*RSe6633 - 1./54.*m33*RD146e*
         RSe6611 - 2./27.*m33*me2*RSe6613*RS1463e + 1./27.*m33*me2*
         RSe6633*RS1463e + 1./27.*m33*me2*RSe6611*RS1463e + 1./27.*m33*
         me2*RC146e*RSe6613 - 1./54.*m33*me2*RC146e*RSe6633 - 1./54.*
         m33*me2*RC146e*RSe6611 + 5./27.*pow(m33,2)*RSe6613*RS1463e - 1.
         /9.*pow(m33,2)*RSe6633*RS1463e - 2./27.*pow(m33,2)*RSe6611*
         RS1463e - 2./27.*pow(m33,2)*RC146e*RSe6613 + 1./54.*pow(m33,2)
         *RC146e*RSe6633 + 1./18.*pow(m33,2)*RC146e*RSe6611 + 1./27.*
         m11*RD146e*RSe6613 - 1./54.*m11*RD146e*RSe6633 - 1./54.*m11*
         RD146e*RSe6611 - 2./27.*m11*me2*RSe6613*RS1463e + 1./27.*m11*
         me2*RSe6633*RS1463e + 1./27.*m11*me2*RSe6611*RS1463e + 1./27.*
         m11*me2*RC146e*RSe6613 - 1./54.*m11*me2*RC146e*RSe6633 - 1./54.
         *m11*me2*RC146e*RSe6611 + 8./27.*m11*m33*RSe6613*RS1463e - 4./
         27.*m11*m33*RSe6633*RS1463e - 4./27.*m11*m33*RSe6611*RS1463e
          - 4./27.*m11*m33*RC146e*RSe6613 );

      F13p6x22R +=  + hh1d(1,m13,m11,me2,m13,mu2) * ( 1./54.*m11*m33*
         RC146e*RSe6633 + 7./54.*m11*m33*RC146e*RSe6611 + 1./9.*pow(
         m11,2)*RSe6613*RS1463e - 1./27.*pow(m11,2)*RSe6633*RS1463e - 2.
         /27.*pow(m11,2)*RSe6611*RS1463e - 2./27.*pow(m11,2)*RC146e*
         RSe6613 + 2./27.*pow(m11,2)*RC146e*RSe6611 );

      F13p6x22R +=  + hh1d(1,m13,m33,m33,m13,mu2) * (  - 1./54.*m33*
         RD346e*RS3461e + 1./108.*m33*RD346e*RC346e - 1./8.*pow(m33,2)
          - 1./108.*pow(m33,2)*pow(RS3461e,2) + 1./54.*pow(m33,2)*
         RC346e*RS3461e - 1./72.*pow(m33,2)*pow(RC346e,2) - 1./54.*m11*
         RD346e*RS3461e + 1./108.*m11*RD346e*RC346e - 1./4.*m11*m33 - 1.
         /18.*m11*m33*pow(RS3461e,2) + 1./18.*m11*m33*RC346e*RS3461e - 
         1./36.*m11*m33*pow(RC346e,2) - 1./8.*pow(m11,2) - 5./108.*pow(
         m11,2)*pow(RS3461e,2) + 1./27.*pow(m11,2)*RC346e*RS3461e - 1./
         72.*pow(m11,2)*pow(RC346e,2) );

      F13p6x22R +=  + hh1d(1,m13,m33,me2,m13,mu2) * ( 1./27.*m33*RD346e
         *RSe6613 - 1./54.*m33*RD346e*RSe6633 - 1./54.*m33*RSe6611*
         RD346e - 2./27.*m33*me2*RSe6613*RS3461e + 1./27.*m33*me2*
         RSe6633*RS3461e + 1./27.*m33*me2*RC346e*RSe6613 - 1./54.*m33*
         me2*RC346e*RSe6633 + 1./27.*m33*me2*RSe6611*RS3461e - 1./54.*
         m33*me2*RSe6611*RC346e + 1./9.*pow(m33,2)*RSe6613*RS3461e - 2./
         27.*pow(m33,2)*RSe6633*RS3461e - 2./27.*pow(m33,2)*RC346e*
         RSe6613 + 2./27.*pow(m33,2)*RC346e*RSe6633 - 1./27.*pow(m33,2)
         *RSe6611*RS3461e + 1./27.*m11*RD346e*RSe6613 - 1./54.*m11*
         RD346e*RSe6633 - 1./54.*m11*RSe6611*RD346e - 2./27.*m11*me2*
         RSe6613*RS3461e + 1./27.*m11*me2*RSe6633*RS3461e + 1./27.*m11*
         me2*RC346e*RSe6613 - 1./54.*m11*me2*RC346e*RSe6633 + 1./27.*
         m11*me2*RSe6611*RS3461e - 1./54.*m11*me2*RSe6611*RC346e + 8./
         27.*m11*m33*RSe6613*RS3461e - 4./27.*m11*m33*RSe6633*RS3461e
          - 4./27.*m11*m33*RC346e*RSe6613 + 7./54.*m11*m33*RC346e*
         RSe6633 );

      F13p6x22R +=  + hh1d(1,m13,m33,me2,m13,mu2) * (  - 4./27.*m11*m33
         *RSe6611*RS3461e + 1./54.*m11*m33*RSe6611*RC346e + 5./27.*pow(
         m11,2)*RSe6613*RS3461e - 2./27.*pow(m11,2)*RSe6633*RS3461e - 2.
         /27.*pow(m11,2)*RC346e*RSe6613 + 1./18.*pow(m11,2)*RC346e*
         RSe6633 - 1./9.*pow(m11,2)*RSe6611*RS3461e + 1./54.*pow(m11,2)
         *RSe6611*RC346e );

      F13p6x22R +=  + hh1d(1,m13,me2,me2,m13,mu2) * ( 4./27.*m33*me2*
         pow(RSe6613,2) - 4./27.*m33*me2*RSe6633*RSe6613 + 1./27.*m33*
         me2*pow(RSe6633,2) - 4./27.*m33*me2*RSe6611*RSe6613 + 2./27.*
         m33*me2*RSe6611*RSe6633 + 1./27.*m33*me2*pow(RSe6611,2) - 5./
         27.*pow(m33,2)*pow(RSe6613,2) + 2./9.*pow(m33,2)*RSe6633*
         RSe6613 - 5./54.*pow(m33,2)*pow(RSe6633,2) + 4./27.*pow(m33,2)
         *RSe6611*RSe6613 - 1./27.*pow(m33,2)*RSe6611*RSe6633 - 1./18.*
         pow(m33,2)*pow(RSe6611,2) + 4./27.*m11*me2*pow(RSe6613,2) - 4./
         27.*m11*me2*RSe6633*RSe6613 + 1./27.*m11*me2*pow(RSe6633,2) - 
         4./27.*m11*me2*RSe6611*RSe6613 + 2./27.*m11*me2*RSe6611*
         RSe6633 + 1./27.*m11*me2*pow(RSe6611,2) - 10./27.*m11*m33*pow(
         RSe6613,2) + 10./27.*m11*m33*RSe6633*RSe6613 - 4./27.*m11*m33*
         pow(RSe6633,2) + 10./27.*m11*m33*RSe6611*RSe6613 - 2./27.*m11*
         m33*RSe6611*RSe6633 - 4./27.*m11*m33*pow(RSe6611,2) - 5./27.*
         pow(m11,2)*pow(RSe6613,2) + 4./27.*pow(m11,2)*RSe6633*RSe6613
          - 1./18.*pow(m11,2)*pow(RSe6633,2) );

      F13p6x22R +=  + hh1d(1,m13,me2,me2,m13,mu2) * ( 2./9.*pow(m11,2)*
         RSe6611*RSe6613 - 1./27.*pow(m11,2)*RSe6611*RSe6633 - 5./54.*
         pow(m11,2)*pow(RSe6611,2) );

      F13p6x22R +=  + hh1d(1,m14,m11,m34,m13,mu2) * (  - 1./18.*m33*
         RD146e - 1./9.*m33*m44*RS16e + 1./18.*pow(m33,2)*RS1463e - 1./
         6.*pow(m33,2)*RS16e - 1./18.*m11*RD146e - 1./9.*m11*m44*RS16e
          - 2./9.*m11*m33*RS16e - 1./18.*pow(m11,2)*RS1463e - 1./18.*
         pow(m11,2)*RS16e );

      F13p6x22R +=  + hh1d(1,m14,m33,m34,m13,mu2) * (  - 1./6.*pow(
         m33,2)*RS3461e + 1./6.*pow(m33,2)*RS36e - 1./3.*m11*m33*
         RS3461e + 1./3.*m11*m33*RS36e - 1./6.*pow(m11,2)*RS3461e + 1./
         6.*pow(m11,2)*RS36e );

      F13p6x22R +=  + hh1d(1,m14,m34,me2,m13,mu2) * (  - 1./9.*m33*me2*
         RSe63 - 1./9.*m33*me2*RSe6633 + 1./9.*m33*me2*RSe61 + 1./9.*
         m33*me2*RSe6611 + 1./9.*m33*m44*RSe63 - 1./9.*m33*m44*RSe61 - 
         1./9.*pow(m33,2)*RSe6613 + 1./6.*pow(m33,2)*RSe63 + 1./9.*pow(
         m33,2)*RSe6633 - 1./6.*pow(m33,2)*RSe61 - 1./9.*m11*me2*RSe63
          - 1./9.*m11*me2*RSe6633 + 1./9.*m11*me2*RSe61 + 1./9.*m11*me2
         *RSe6611 + 1./9.*m11*m44*RSe63 - 1./9.*m11*m44*RSe61 + 1./3.*
         m11*m33*RSe63 + 1./9.*m11*m33*RSe6633 - 1./3.*m11*m33*RSe61 - 
         1./9.*m11*m33*RSe6611 + 1./9.*pow(m11,2)*RSe6613 + 1./6.*pow(
         m11,2)*RSe63 - 1./6.*pow(m11,2)*RSe61 - 1./9.*pow(m11,2)*
         RSe6611 );

      F13p6x22R +=  + hh1d(1,m16,m11,m36,m13,mu2) * (  - 1./36.*m33*
         RD146e - 1./18.*m33*m66*RS14e + 1./36.*pow(m33,2)*RS1463e - 1./
         12.*pow(m33,2)*RS14e - 1./36.*m11*RD146e - 1./18.*m11*m66*
         RS14e - 1./9.*m11*m33*RS14e - 1./36.*pow(m11,2)*RS1463e - 1./
         36.*pow(m11,2)*RS14e );

      F13p6x22R +=  + hh1d(1,m16,m33,m36,m13,mu2) * (  - 1./12.*pow(
         m33,2)*RS3461e + 1./12.*pow(m33,2)*RS34e - 1./6.*m11*m33*
         RS3461e + 1./6.*m11*m33*RS34e - 1./12.*pow(m11,2)*RS3461e + 1./
         12.*pow(m11,2)*RS34e );

      F13p6x22R +=  + hh1d(1,m16,m36,me2,m13,mu2) * ( 1./9.*m33*me2*
         RSe63 - 1./18.*m33*me2*RSe6633 - 1./9.*m33*me2*RSe61 + 1./18.*
         m33*me2*RSe6611 - 1./9.*m33*m66*RSe63 + 1./9.*m33*m66*RSe61 - 
         1./18.*pow(m33,2)*RSe6613 - 1./6.*pow(m33,2)*RSe63 + 1./18.*
         pow(m33,2)*RSe6633 + 1./6.*pow(m33,2)*RSe61 + 1./9.*m11*me2*
         RSe63 - 1./18.*m11*me2*RSe6633 - 1./9.*m11*me2*RSe61 + 1./18.*
         m11*me2*RSe6611 - 1./9.*m11*m66*RSe63 + 1./9.*m11*m66*RSe61 - 
         1./3.*m11*m33*RSe63 + 1./18.*m11*m33*RSe6633 + 1./3.*m11*m33*
         RSe61 - 1./18.*m11*m33*RSe6611 + 1./18.*pow(m11,2)*RSe6613 - 1.
         /6.*pow(m11,2)*RSe63 + 1./6.*pow(m11,2)*RSe61 - 1./18.*pow(
         m11,2)*RSe6611 );

      F13p6x22R +=  + hh1d(1,m34,m14,m33,m13,mu2) * (  - 1./18.*m33*
         RD346e - 1./9.*m33*m44*RS36e - 2./9.*pow(m33,2)*RS3461e + 1./9.
         *pow(m33,2)*RS36e - 1./18.*m11*RD346e - 1./9.*m11*m44*RS36e - 
         1./3.*m11*m33*RS3461e + 1./9.*m11*m33*RS36e - 1./9.*pow(m11,2)
         *RS3461e );

      F13p6x22R +=  + hh1d(1,m36,m16,m33,m13,mu2) * (  - 1./36.*m33*
         RD346e - 1./18.*m33*m66*RS34e - 1./9.*pow(m33,2)*RS3461e + 1./
         18.*pow(m33,2)*RS34e - 1./36.*m11*RD346e - 1./18.*m11*m66*
         RS34e - 1./6.*m11*m33*RS3461e + 1./18.*m11*m33*RS34e - 1./18.*
         pow(m11,2)*RS3461e );

      F13p6x22R +=  + hh1d(1,me2,m14,m34,m13,mu2) * (  - 1./18.*m33*me2
          - 1./9.*m33*me2*RSe6613 - 1./6.*m33*me2*RSe63 - 1./9.*m33*me2
         *RSe6633 - 1./18.*m33*me2*RSe61 + 1./18.*m33*m44 + 1./9.*m33*
         m44*RSe63 - 1./3.*pow(m33,2)*RSe6613 + 1./18.*pow(m33,2)*RSe63
          + 1./9.*pow(m33,2)*RSe6633 - 1./6.*pow(m33,2)*RSe61 - 1./18.*
         m11*me2 - 1./9.*m11*me2*RSe6613 - 1./6.*m11*me2*RSe63 - 1./9.*
         m11*me2*RSe6633 - 1./18.*m11*me2*RSe61 + 1./18.*m11*m44 + 1./9.
         *m11*m44*RSe63 - 5./9.*m11*m33*RSe6613 + 1./18.*m11*m33*RSe63
          + 1./9.*m11*m33*RSe6633 - 5./18.*m11*m33*RSe61 - 2./9.*pow(
         m11,2)*RSe6613 - 1./9.*pow(m11,2)*RSe61 );

      F13p6x22R +=  + hh1d(1,me2,m16,m36,m13,mu2) * (  - 1./9.*m33*me2
          - 1./18.*m33*me2*RSe6613 + 1./6.*m33*me2*RSe63 - 1./18.*m33*
         me2*RSe6633 + 1./18.*m33*me2*RSe61 + 1./9.*m33*m66 - 1./9.*m33
         *m66*RSe63 - 1./6.*pow(m33,2)*RSe6613 - 1./18.*pow(m33,2)*
         RSe63 + 1./18.*pow(m33,2)*RSe6633 + 1./6.*pow(m33,2)*RSe61 - 1.
         /9.*m11*me2 - 1./18.*m11*me2*RSe6613 + 1./6.*m11*me2*RSe63 - 1.
         /18.*m11*me2*RSe6633 + 1./18.*m11*me2*RSe61 + 1./9.*m11*m66 - 
         1./9.*m11*m66*RSe63 - 5./18.*m11*m33*RSe6613 - 1./18.*m11*m33*
         RSe63 + 1./18.*m11*m33*RSe6633 + 5./18.*m11*m33*RSe61 - 1./9.*
         pow(m11,2)*RSe6613 + 1./9.*pow(m11,2)*RSe61 );

      F13p6x22R +=  + hh1d(3,m13,m11,m11,m13,mu2) * ( 1./108.*m33*pow(
         RD146e,2) + 1./27.*pow(m33,2)*RD146e*RS1463e - 1./36.*pow(
         m33,2)*RD146e*RC146e + 1./108.*m11*pow(RD146e,2) + 1./18.*m11*
         m33*RD146e*RS1463e - 1./18.*m11*m33*RD146e*RC146e + 1./54.*
         pow(m11,2)*RD146e*RS1463e - 1./36.*pow(m11,2)*RD146e*RC146e );

      F13p6x22R +=  + hh1d(3,m13,m11,m33,m13,mu2) * ( 1./108.*m33*
         RD146e*RD346e + 1./54.*pow(m33,2)*RD146e*RS3461e + 1./108.*m11
         *RD146e*RD346e + 1./18.*m11*m33*RD146e*RS3461e + 1./27.*pow(
         m11,2)*RD146e*RS3461e );

      F13p6x22R +=  + hh1d(3,m13,m11,me2,m13,mu2) * ( 1./27.*m33*me2*
         RD146e*RSe6613 - 1./54.*m33*me2*RD146e*RSe6633 - 1./54.*m33*
         me2*RD146e*RSe6611 - 2./27.*pow(m33,2)*RD146e*RSe6613 + 1./54.
         *pow(m33,2)*RD146e*RSe6633 + 1./18.*pow(m33,2)*RD146e*RSe6611
          + 1./27.*m11*me2*RD146e*RSe6613 - 1./54.*m11*me2*RD146e*
         RSe6633 - 1./54.*m11*me2*RD146e*RSe6611 - 4./27.*m11*m33*
         RD146e*RSe6613 + 1./54.*m11*m33*RD146e*RSe6633 + 7./54.*m11*
         m33*RD146e*RSe6611 - 2./27.*pow(m11,2)*RD146e*RSe6613 + 2./27.
         *pow(m11,2)*RD146e*RSe6611 );

      F13p6x22R +=  + hh1d(3,m13,m33,m11,m13,mu2) * ( 1./108.*m33*
         RD146e*RD346e + 1./27.*pow(m33,2)*RD346e*RS1463e + 1./108.*m11
         *RD146e*RD346e + 1./18.*m11*m33*RD346e*RS1463e + 1./54.*pow(
         m11,2)*RD346e*RS1463e );

      F13p6x22R +=  + hh1d(3,m13,m33,m33,m13,mu2) * ( 1./108.*m33*pow(
         RD346e,2) + 1./54.*pow(m33,2)*RD346e*RS3461e - 1./36.*pow(
         m33,2)*RD346e*RC346e + 1./108.*m11*pow(RD346e,2) + 1./18.*m11*
         m33*RD346e*RS3461e - 1./18.*m11*m33*RD346e*RC346e + 1./27.*
         pow(m11,2)*RD346e*RS3461e - 1./36.*pow(m11,2)*RD346e*RC346e );

      F13p6x22R +=  + hh1d(3,m13,m33,me2,m13,mu2) * ( 1./27.*m33*me2*
         RD346e*RSe6613 - 1./54.*m33*me2*RD346e*RSe6633 - 1./54.*m33*
         me2*RSe6611*RD346e - 2./27.*pow(m33,2)*RD346e*RSe6613 + 2./27.
         *pow(m33,2)*RD346e*RSe6633 + 1./27.*m11*me2*RD346e*RSe6613 - 1.
         /54.*m11*me2*RD346e*RSe6633 - 1./54.*m11*me2*RSe6611*RD346e - 
         4./27.*m11*m33*RD346e*RSe6613 + 7./54.*m11*m33*RD346e*RSe6633
          + 1./54.*m11*m33*RSe6611*RD346e - 2./27.*pow(m11,2)*RD346e*
         RSe6613 + 1./18.*pow(m11,2)*RD346e*RSe6633 + 1./54.*pow(m11,2)
         *RSe6611*RD346e );

      F13p6x22R +=  + hh1d(7,m13,m11,m11,m13,mu2) * (  - 1./72.*pow(
         m33,2)*pow(RD146e,2) - 1./36.*m11*m33*pow(RD146e,2) - 1./72.*
         pow(m11,2)*pow(RD146e,2) );

      F13p6x22R +=  + hh1d(7,m13,m33,m33,m13,mu2) * (  - 1./72.*pow(
         m33,2)*pow(RD346e,2) - 1./36.*m11*m33*pow(RD346e,2) - 1./72.*
         pow(m11,2)*pow(RD346e,2) );

      F13p6x22R +=  + hh21d(1,m11,m11,m13,m13,mu2) * (  - 3./8.*pow(
         m13,2) );

      F13p6x22R +=  + hh21d(1,m11,m14,m34,m13,mu2) * (  - 1./4.*pow(
         m13,2)*RS1463e + 1./4.*pow(m13,2)*RS16e );

      F13p6x22R +=  + hh21d(1,m11,m16,m36,m13,mu2) * (  - 1./8.*pow(
         m13,2)*RS1463e + 1./8.*pow(m13,2)*RS14e );

      F13p6x22R +=  + hh21d(1,m13,m11,m11,m13,mu2) * ( 3./16.*pow(
         m13,2) + 1./12.*pow(m13,2)*pow(RS1463e,2) - 1./12.*pow(m13,2)*
         RC146e*RS1463e + 1./48.*pow(m13,2)*pow(RC146e,2) );

      F13p6x22R +=  + hh21d(1,m13,m11,m33,m13,mu2) * ( 1./6.*pow(m13,2)
         *RS1463e*RS3461e - 1./12.*pow(m13,2)*RC346e*RS1463e - 1./12.*
         pow(m13,2)*RC146e*RS3461e + 1./24.*pow(m13,2)*RC146e*RC346e );

      F13p6x22R +=  + hh21d(1,m13,m11,me2,m13,mu2) * (  - 1./3.*pow(
         m13,2)*RSe6613*RS1463e + 1./6.*pow(m13,2)*RSe6633*RS1463e + 1./
         6.*pow(m13,2)*RSe6611*RS1463e + 1./6.*pow(m13,2)*RC146e*
         RSe6613 - 1./12.*pow(m13,2)*RC146e*RSe6633 - 1./12.*pow(m13,2)
         *RC146e*RSe6611 );

      F13p6x22R +=  + hh21d(1,m13,m13,m13,m13,mu2) * ( 3./8.*pow(m13,2)
          );

      F13p6x22R +=  + hh21d(1,m13,m33,m33,m13,mu2) * ( 3./16.*pow(
         m13,2) + 1./12.*pow(m13,2)*pow(RS3461e,2) - 1./12.*pow(m13,2)*
         RC346e*RS3461e + 1./48.*pow(m13,2)*pow(RC346e,2) );

      F13p6x22R +=  + hh21d(1,m13,m33,me2,m13,mu2) * (  - 1./3.*pow(
         m13,2)*RSe6613*RS3461e + 1./6.*pow(m13,2)*RSe6633*RS3461e + 1./
         6.*pow(m13,2)*RC346e*RSe6613 - 1./12.*pow(m13,2)*RC346e*
         RSe6633 + 1./6.*pow(m13,2)*RSe6611*RS3461e - 1./12.*pow(m13,2)
         *RSe6611*RC346e );

      F13p6x22R +=  + hh21d(1,m13,me2,me2,m13,mu2) * ( 1./3.*pow(m13,2)
         *pow(RSe6613,2) - 1./3.*pow(m13,2)*RSe6633*RSe6613 + 1./12.*
         pow(m13,2)*pow(RSe6633,2) - 1./3.*pow(m13,2)*RSe6611*RSe6613
          + 1./6.*pow(m13,2)*RSe6611*RSe6633 + 1./12.*pow(m13,2)*pow(
         RSe6611,2) );

      F13p6x22R +=  + hh21d(1,m14,m11,m34,m13,mu2) * ( 1./4.*pow(m13,2)
         *RS1463e + 1./4.*pow(m13,2)*RS16e - 1./4.*pow(m13,2)*RC146e );

      F13p6x22R +=  + hh21d(1,m14,m33,m34,m13,mu2) * ( 1./4.*pow(m13,2)
         *RS3461e - 1./4.*pow(m13,2)*RS36e );

      F13p6x22R +=  + hh21d(1,m14,m34,me2,m13,mu2) * (  - 1./2.*pow(
         m13,2)*RSe6613 - 1./4.*pow(m13,2)*RSe63 + 1./4.*pow(m13,2)*
         RSe61 + 1./2.*pow(m13,2)*RSe6611 );

      F13p6x22R +=  + hh21d(1,m16,m11,m36,m13,mu2) * ( 1./8.*pow(m13,2)
         *RS1463e + 1./8.*pow(m13,2)*RS14e - 1./8.*pow(m13,2)*RC146e );

      F13p6x22R +=  + hh21d(1,m16,m33,m36,m13,mu2) * ( 1./8.*pow(m13,2)
         *RS3461e - 1./8.*pow(m13,2)*RS34e );

      F13p6x22R +=  + hh21d(1,m16,m36,me2,m13,mu2) * (  - 1./4.*pow(
         m13,2)*RSe6613 + 1./4.*pow(m13,2)*RSe63 - 1./4.*pow(m13,2)*
         RSe61 + 1./4.*pow(m13,2)*RSe6611 );

      F13p6x22R +=  + hh21d(1,m33,m13,m33,m13,mu2) * (  - 3./8.*pow(
         m13,2) );

      F13p6x22R +=  + hh21d(1,m33,m14,m34,m13,mu2) * (  - 1./4.*pow(
         m13,2)*RS3461e + 1./4.*pow(m13,2)*RS36e );

      F13p6x22R +=  + hh21d(1,m33,m16,m36,m13,mu2) * (  - 1./8.*pow(
         m13,2)*RS3461e + 1./8.*pow(m13,2)*RS34e );

      F13p6x22R +=  + hh21d(1,m34,m11,m14,m13,mu2) * ( 1./4.*pow(m13,2)
         *RS1463e - 1./4.*pow(m13,2)*RS16e );

      F13p6x22R +=  + hh21d(1,m34,m14,m33,m13,mu2) * ( 1./4.*pow(m13,2)
         *RS3461e + 1./4.*pow(m13,2)*RS36e - 1./4.*pow(m13,2)*RC346e );

      F13p6x22R +=  + hh21d(1,m34,m14,me2,m13,mu2) * (  - 1./2.*pow(
         m13,2)*RSe6613 + 1./4.*pow(m13,2)*RSe63 + 1./2.*pow(m13,2)*
         RSe6633 - 1./4.*pow(m13,2)*RSe61 );

      F13p6x22R +=  + hh21d(1,m36,m11,m16,m13,mu2) * ( 1./8.*pow(m13,2)
         *RS1463e - 1./8.*pow(m13,2)*RS14e );

      F13p6x22R +=  + hh21d(1,m36,m16,m33,m13,mu2) * ( 1./8.*pow(m13,2)
         *RS3461e + 1./8.*pow(m13,2)*RS34e - 1./8.*pow(m13,2)*RC346e );

      F13p6x22R +=  + hh21d(1,m36,m16,me2,m13,mu2) * (  - 1./4.*pow(
         m13,2)*RSe6613 - 1./4.*pow(m13,2)*RSe63 + 1./4.*pow(m13,2)*
         RSe6633 + 1./4.*pow(m13,2)*RSe61 );

      F13p6x22R +=  + hh21d(1,m44,m14,m34,m13,mu2) * ( 9./8.*pow(m13,2)
          );

      F13p6x22R +=  + hh21d(1,m46,m14,m36,m13,mu2) * ( 3./4.*pow(m13,2)
          );

      F13p6x22R +=  + hh21d(1,m46,m16,m34,m13,mu2) * ( 3./4.*pow(m13,2)
          );

      F13p6x22R +=  + hh21d(1,me2,m14,m34,m13,mu2) * ( 1./8.*pow(m13,2)
          + 1./2.*pow(m13,2)*RSe6613 + 1./4.*pow(m13,2)*RSe63 + 1./4.*
         pow(m13,2)*RSe61 );

      F13p6x22R +=  + hh21d(1,me2,m16,m36,m13,mu2) * ( 1./4.*pow(m13,2)
          + 1./4.*pow(m13,2)*RSe6613 - 1./4.*pow(m13,2)*RSe63 - 1./4.*
         pow(m13,2)*RSe61 );

      F13p6x22R +=  + hh21d(3,m13,m11,m11,m13,mu2) * (  - 1./12.*pow(
         m13,2)*RD146e*RS1463e + 1./24.*pow(m13,2)*RD146e*RC146e );

      F13p6x22R +=  + hh21d(3,m13,m11,m33,m13,mu2) * (  - 1./12.*pow(
         m13,2)*RD146e*RS3461e + 1./24.*pow(m13,2)*RD146e*RC346e );

      F13p6x22R +=  + hh21d(3,m13,m11,me2,m13,mu2) * ( 1./6.*pow(m13,2)
         *RD146e*RSe6613 - 1./12.*pow(m13,2)*RD146e*RSe6633 - 1./12.*
         pow(m13,2)*RD146e*RSe6611 );

      F13p6x22R +=  + hh21d(3,m13,m33,m11,m13,mu2) * (  - 1./12.*pow(
         m13,2)*RD346e*RS1463e + 1./24.*pow(m13,2)*RC146e*RD346e );

      F13p6x22R +=  + hh21d(3,m13,m33,m33,m13,mu2) * (  - 1./12.*pow(
         m13,2)*RD346e*RS3461e + 1./24.*pow(m13,2)*RD346e*RC346e );

      F13p6x22R +=  + hh21d(3,m13,m33,me2,m13,mu2) * ( 1./6.*pow(m13,2)
         *RD346e*RSe6613 - 1./12.*pow(m13,2)*RD346e*RSe6633 - 1./12.*
         pow(m13,2)*RSe6611*RD346e );

      F13p6x22R +=  + hh21d(3,m14,m11,m34,m13,mu2) * (  - 1./4.*pow(
         m13,2)*RD146e );

      F13p6x22R +=  + hh21d(3,m16,m11,m36,m13,mu2) * (  - 1./8.*pow(
         m13,2)*RD146e );

      F13p6x22R +=  + hh21d(3,m34,m33,m14,m13,mu2) * (  - 1./4.*pow(
         m13,2)*RD346e );

      F13p6x22R +=  + hh21d(3,m36,m33,m16,m13,mu2) * (  - 1./8.*pow(
         m13,2)*RD346e );

      F13p6x22R +=  + hh21d(7,m13,m11,m11,m13,mu2) * ( 1./48.*pow(
         m13,2)*pow(RD146e,2) );

      F13p6x22R +=  + hh21d(7,m13,m11,m33,m13,mu2) * ( 1./24.*pow(
         m13,2)*RD146e*RD346e );

      F13p6x22R +=  + hh21d(7,m13,m33,m33,m13,mu2) * ( 1./48.*pow(
         m13,2)*pow(RD346e,2) );

  return F13p6x22R/pow(f0,4);
}

// 1 valence mass 3 sea mass *****************************************
double fv1s3nf3p4(const quarkmassnf mass, const Linf Liin){
  return fv1s3nf3p4L(mass,Liin)+fv1s3nf3p4R(mass);
}

double fv1s3nf3p4L(const quarkmassnf mass, const Linf Liin){
  double L0r,L1r,L2r,L3r,L4r,L5r,L6r,L7r,L8r,L9r,L10r,L11r,H1r,H2r,mu;
  int nfLi;
  Liin.out(L0r,L1r,L2r,L3r,L4r,L5r,L6r,L7r,L8r,L9r,L10r,L11r,H1r,H2r,mu,nfLi);
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if (mu0 != mu){
    cout << "ERROR: differing mu's in fv1s3nf3p4L\n";}
  if ( (nq != 4) || (B0mq.size() != 4)){
    cout << "ERROR: fv1s3nf3p4L probably called with wrong number"
	 << "of quark masses\n";}
  // four masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m44 = 2.*B0mq[1];
  double m55 = 2.*B0mq[2];
  double m66 = 2.*B0mq[3];
  double F12p4x13L =
       + 4*m66*L4r + 4*m55*L4r + 4*m44*L4r + 4*m11*L5r;

  return F12p4x13L/pow(f0,2);
}


double fv1s3nf3p4R(const quarkmassnf mass){
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if ( (nq != 4) || (B0mq.size() != 4)){
    cout << "ERROR: fv2s3nf3p4R probably called with wrong number"
	 << "of quark masses\n";}
  // four masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m44 = 2.*B0mq[1];
  double m55 = 2.*B0mq[2];
  double m66 = 2.*B0mq[3];
  double m14 = (m11+m44)/2;
  double m15 = (m11+m55)/2;
  double m16 = (m11+m66)/2;

  double mu2 = mu0*mu0;

  double F12p4x13R =
       + Ab(m14,mu2) * ( 1./2. );

      F12p4x13R +=  + Ab(m15,mu2) * ( 1./2. );

      F12p4x13R +=  + Ab(m16,mu2) * ( 1./2. );

  return F12p4x13R/pow(f0,2);
}

double fv1s3nf3p6(const quarkmassnf mass, const Linf Liin, const Ki Kiin){
  return fv1s3nf3p6L(mass,Liin)+fv1s3nf3p6K(mass,Kiin)+fv1s3nf3p6R(mass);
}

double fv1s3nf3p6L(const quarkmassnf mass, const Linf Liin){
  double L0r,L1r,L2r,L3r,L4r,L5r,L6r,L7r,L8r,L9r,L10r,L11r,H1r,H2r,mu;
  int nfLi;
  Liin.out(L0r,L1r,L2r,L3r,L4r,L5r,L6r,L7r,L8r,L9r,L10r,L11r,H1r,H2r,mu,nfLi);
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if (mu0 != mu){
    cout << "ERROR: differing mu's in fv1s3nf3p6L\n";}
  if ( (nq != 4) || (B0mq.size() != 4)){
    cout << "ERROR: fv1s3nf3p6L probably called with wrong number"
	 << "of quark masses\n";}
  // four masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m44 = 2.*B0mq[1];
  double m55 = 2.*B0mq[2];
  double m66 = 2.*B0mq[3];
  double m14 = (m11+m44)/2;
  double m15 = (m11+m55)/2;
  double m16 = (m11+m66)/2;
  double m45 = (m44+m55)/2;
  double m46 = (m44+m66)/2;
  double m56 = (m55+m66)/2;

  double xb1 = (m44+m55+m66)/3.;
  double msqr = sqrt(xb1*xb1-(m44*m55+m44*m66+m55*m66)/3.);
  double mp2 = xb1-msqr; 
  double me2 = xb1+msqr;
  double mu2 = mu0*mu0;


  double RS456pe = RS(m44,m55,m66,mp2,me2);
  double RS546pe = RS(m55,m44,m66,mp2,me2);
  double RS645pe = RS(m66,m44,m55,mp2,me2);

  double RD1456ep = RD(m11,m44,m55,m66,me2,mp2);
  double RC1456pe = RC(m11,m44,m55,m66,mp2,me2);  
  double RSe45611p = RS(me2,m44,m55,m66,m11,m11,mp2);
  double RSp45611e = RS(mp2,m44,m55,m66,m11,m11,me2);

  double F12p6x13L =
       + pi16 * ( 4./9.*me2*mp2*L3r - 11./3.*me2*mp2*L2r + 4./9.*me2*
         mp2*L0r - 43./108.*pow(m66,2)*L3r - 13./9.*pow(m66,2)*L2r - 35.
         /54.*pow(m66,2)*L0r - 43./108.*pow(m55,2)*L3r - 13./9.*pow(
         m55,2)*L2r - 35./54.*pow(m55,2)*L0r - 43./108.*pow(m44,2)*L3r
          - 13./9.*pow(m44,2)*L2r - 35./54.*pow(m44,2)*L0r - 17./18.*
         m11*m66*L3r - 13./9.*m11*m66*L0r - 17./18.*m11*m55*L3r - 13./9.
         *m11*m55*L0r - 17./18.*m11*m44*L3r - 13./9.*m11*m44*L0r + 5./4.
         *pow(m11,2)*L3r - pow(m11,2)*L2r - 2*pow(m11,2)*L1r + 1./2.*
         pow(m11,2)*L0r );

      F12p6x13L +=  + Ab(m11,mu2) * ( 4*L3r*RD1456ep + 4*L0r*RD1456ep
          - 4./3.*m11*L5r*RC1456pe + 4*m11*L3r*RC1456pe - 10*m11*L2r - 
         4*m11*L1r + 4*m11*L0r*RC1456pe );

      F12p6x13L +=  + Ab(m14,mu2) * (  - 2*m66*L4r - 2*m55*L4r - 2*m44*
         L4r - 5*m44*L3r - 2*m44*L0r + 2*m11*L5r - 5*m11*L3r - 2*m11*
         L0r );

      F12p6x13L +=  + Ab(m15,mu2) * (  - 2*m66*L4r - 2*m55*L4r - 5*m55*
         L3r - 2*m55*L0r - 2*m44*L4r + 2*m11*L5r - 5*m11*L3r - 2*m11*
         L0r );

      F12p6x13L +=  + Ab(m16,mu2) * (  - 2*m66*L4r - 5*m66*L3r - 2*m66*
         L0r - 2*m55*L4r - 2*m44*L4r + 2*m11*L5r - 5*m11*L3r - 2*m11*
         L0r );

      F12p6x13L +=  + Ab(m44,mu2) * ( 4*m44*L4r - 4./3.*m44*L4r*RS456pe
          - 2*m44*L2r + 2./3.*m44*L2r*RS456pe - 8*m44*L1r + 8./3.*m44*
         L1r*RS456pe );

      F12p6x13L +=  + Ab(m45,mu2) * ( 4*m55*L4r - 2*m55*L2r - 8*m55*L1r
          + 4*m44*L4r - 2*m44*L2r - 8*m44*L1r );

      F12p6x13L +=  + Ab(m46,mu2) * ( 4*m66*L4r - 2*m66*L2r - 8*m66*L1r
          + 4*m44*L4r - 2*m44*L2r - 8*m44*L1r );

      F12p6x13L +=  + Ab(m55,mu2) * ( 4*m55*L4r - 4./3.*m55*L4r*RS546pe
          - 2*m55*L2r + 2./3.*m55*L2r*RS546pe - 8*m55*L1r + 8./3.*m55*
         L1r*RS546pe );

      F12p6x13L +=  + Ab(m56,mu2) * ( 4*m66*L4r - 2*m66*L2r - 8*m66*L1r
          + 4*m55*L4r - 2*m55*L2r - 8*m55*L1r );

      F12p6x13L +=  + Ab(m66,mu2) * ( 4*m66*L4r - 4./3.*m66*L4r*RS645pe
          - 2*m66*L2r + 2./3.*m66*L2r*RS645pe - 8*m66*L1r + 8./3.*m66*
         L1r*RS645pe );

      F12p6x13L +=  + Ab(me2,mu2) * ( 4*me2*L3r*RSe45611p + 4*me2*L0r*
         RSe45611p - 4./3.*m11*L5r*RSe45611p + 2./3.*RS(me2,m44,m55,m66
         ,mp2,m44,m44)*me2*L2r + 8./3.*RS(me2,m44,m55,m66,mp2,m44,m44)*
         me2*L1r - 4./3.*RS(me2,m44,m55,m66,mp2,m44,m44)*m44*L4r + 2./3.
         *RS(me2,m44,m55,m66,mp2,m55,m55)*me2*L2r + 8./3.*RS(me2,m44,
         m55,m66,mp2,m55,m55)*me2*L1r - 4./3.*RS(me2,m44,m55,m66,mp2,
         m55,m55)*m55*L4r + 2./3.*RS(me2,m44,m55,m66,mp2,m66,m66)*me2*
         L2r + 8./3.*RS(me2,m44,m55,m66,mp2,m66,m66)*me2*L1r - 4./3.*
         RS(me2,m44,m55,m66,mp2,m66,m66)*m66*L4r );

      F12p6x13L +=  + Ab(mp2,mu2) * ( 4*mp2*L3r*RSp45611e + 4*mp2*L0r*
         RSp45611e - 4./3.*m11*L5r*RSp45611e + 2./3.*RS(mp2,m44,m55,m66
         ,me2,m44,m44)*mp2*L2r + 8./3.*RS(mp2,m44,m55,m66,me2,m44,m44)*
         mp2*L1r - 4./3.*RS(mp2,m44,m55,m66,me2,m44,m44)*m44*L4r + 2./3.
         *RS(mp2,m44,m55,m66,me2,m55,m55)*mp2*L2r + 8./3.*RS(mp2,m44,
         m55,m66,me2,m55,m55)*mp2*L1r - 4./3.*RS(mp2,m44,m55,m66,me2,
         m55,m55)*m55*L4r + 2./3.*RS(mp2,m44,m55,m66,me2,m66,m66)*mp2*
         L2r + 8./3.*RS(mp2,m44,m55,m66,me2,m66,m66)*mp2*L1r - 4./3.*
         RS(mp2,m44,m55,m66,me2,m66,m66)*m66*L4r );

      F12p6x13L +=  + Bb(m11,mu2) * (  - 4./3.*m11*L5r*RD1456ep + 4*m11
         *L3r*RD1456ep + 4*m11*L0r*RD1456ep );

      F12p6x13L +=  + Bb(m14,mu2) * ( 4*m44*m66*L6r - 2*m44*m66*L4r + 4
         *m44*m55*L6r - 2*m44*m55*L4r + 2*pow(m44,2)*L8r + 4*pow(m44,2)
         *L6r - pow(m44,2)*L5r - 2*pow(m44,2)*L4r + 4*m11*m66*L6r - 2*
         m11*m66*L4r + 4*m11*m55*L6r - 2*m11*m55*L4r + 4*m11*m44*L8r + 
         4*m11*m44*L6r - 2*m11*m44*L5r - 2*m11*m44*L4r + 2*pow(m11,2)*
         L8r - pow(m11,2)*L5r );

      F12p6x13L +=  + Bb(m15,mu2) * ( 4*m55*m66*L6r - 2*m55*m66*L4r + 2
         *pow(m55,2)*L8r + 4*pow(m55,2)*L6r - pow(m55,2)*L5r - 2*pow(
         m55,2)*L4r + 4*m44*m55*L6r - 2*m44*m55*L4r + 4*m11*m66*L6r - 2
         *m11*m66*L4r + 4*m11*m55*L8r + 4*m11*m55*L6r - 2*m11*m55*L5r
          - 2*m11*m55*L4r + 4*m11*m44*L6r - 2*m11*m44*L4r + 2*pow(
         m11,2)*L8r - pow(m11,2)*L5r );

      F12p6x13L +=  + Bb(m16,mu2) * ( 2*pow(m66,2)*L8r + 4*pow(m66,2)*
         L6r - pow(m66,2)*L5r - 2*pow(m66,2)*L4r + 4*m55*m66*L6r - 2*
         m55*m66*L4r + 4*m44*m66*L6r - 2*m44*m66*L4r + 4*m11*m66*L8r + 
         4*m11*m66*L6r - 2*m11*m66*L5r - 2*m11*m66*L4r + 4*m11*m55*L6r
          - 2*m11*m55*L4r + 4*m11*m44*L6r - 2*m11*m44*L4r + 2*pow(
         m11,2)*L8r - pow(m11,2)*L5r );

      F12p6x13L +=  - 8*pow(m66,2)*pow(L4r,2) - 16*m55*m66*pow(L4r,2)
          - 8*pow(m55,2)*pow(L4r,2) - 16*m44*m66*pow(L4r,2) - 16*m44*
         m55*pow(L4r,2) - 8*pow(m44,2)*pow(L4r,2) - 16*m11*m66*L4r*L5r
          - 16*m11*m55*L4r*L5r - 16*m11*m44*L4r*L5r - 8*pow(m11,2)*pow(
         L5r,2);

  return F12p6x13L/pow(f0,4);
}

double fv1s3nf3p6K(const quarkmassnf mass, const Ki Kiin){
  double KK[116],mu;
  Kiin.out(KK,mu);
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if (mu0 != mu){
    cout << "ERROR: differing mu's in fv1s3nf3p6L\n";}
  if ( (nq != 4) || (B0mq.size() != 4)){
    cout << "ERROR: fv1s3nf3p6K probably called with wrong number"
	 << "of quark masses\n";}
  // four masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m44 = 2.*B0mq[1];
  double m55 = 2.*B0mq[2];
  double m66 = 2.*B0mq[3];

  double F12p6x13K =
       + KK[19] * ( 8*pow(m11,2) );

      F12p6x13K +=  + KK[20] * ( 8*m11*m66 + 8*m11*m55 + 8*m11*m44 );

      F12p6x13K +=  + KK[21] * ( 8*pow(m66,2) + 8*pow(m55,2) + 8*pow(
         m44,2) );

      F12p6x13K +=  + KK[22] * ( 8*pow(m66,2) + 16*m55*m66 + 8*pow(
         m55,2) + 16*m44*m66 + 16*m44*m55 + 8*pow(m44,2) );

      F12p6x13K +=  + KK[23] * ( 8*pow(m11,2) );

  return F12p6x13K/pow(f0,4);
}

double fv1s3nf3p6R(const quarkmassnf mass){
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if ( (nq != 4) || (B0mq.size() != 4)){
    cout << "ERROR: fv1s3nf3p6R probably called with wrong number"
	 << "of quark masses\n";}
  // four masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m44 = 2.*B0mq[1];
  double m55 = 2.*B0mq[2];
  double m66 = 2.*B0mq[3];
  double m14 = (m11+m44)/2;
  double m15 = (m11+m55)/2;
  double m16 = (m11+m66)/2;
  double m45 = (m44+m55)/2;
  double m46 = (m44+m66)/2;
  double m56 = (m55+m66)/2;

  double xb1 = (m44+m55+m66)/3.;
  double msqr = sqrt(xb1*xb1-(m44*m55+m44*m66+m55*m66)/3.);
  double mp2 = xb1-msqr; 
  double me2 = xb1+msqr;
  double mu2 = mu0*mu0;

//  double RN1p = RN(m11,mp2);
//  double RN1e = RN(m11,me2);
//  double RNep = RN(me2,mp2);

  double RS14e = RS(m11,m44,me2);
  double RS16e = RS(m11,m66,me2);
  double RS144ee = RS14e*RS14e;
  double RS166ee = RS16e*RS16e;
  double RSe6611 = RS(me2,m66,m66,m11,m11);
  double RSp456e = RS(mp2,m44,m55,m66,me2);
  double RSp465e = RS(mp2,m44,m66,m55,me2);
  double RSp564e = RS(mp2,m55,m66,m44,me2);
  double RSe456p = RS(me2,m44,m55,m66,mp2);
  double RSe465p = RS(me2,m44,m66,m55,mp2);
  double RSe564p = RS(me2,m55,m66,m44,mp2);
  double RSp451e = RS(mp2,m44,m55,m11,me2);
  double RSp461e = RS(mp2,m44,m66,m11,me2);
  double RSp561e = RS(mp2,m55,m66,m11,me2);
  double RSe451p = RS(me2,m44,m55,m11,mp2);
  double RSe461p = RS(me2,m44,m66,m11,mp2);
  double RSe561p = RS(me2,m55,m66,m11,mp2);
  double RS456pe = RS(m44,m55,m66,mp2,me2);
  double RS546pe = RS(m55,m44,m66,mp2,me2);
  double RS645pe = RS(m66,m44,m55,mp2,me2);
  double RS156pe = RS(m11,m55,m66,mp2,me2);
  double RS146pe = RS(m11,m44,m66,mp2,me2);
  double RS145pe = RS(m11,m44,m55,mp2,me2);

  double RD1456ep = RD(m11,m44,m55,m66,me2,mp2);
  double RC1456pe = RC(m11,m44,m55,m66,mp2,me2);  
  double RSe45611p = RS(me2,m44,m55,m66,m11,m11,mp2);
  double RSp45611e = RS(mp2,m44,m55,m66,m11,m11,me2);
//  double RS1456ppe = RS(m11,m44,m55,m66,mp2,mp2,me2);
//  double RS1456pee = RS(m11,m44,m55,m66,mp2,me2,me2);

  double F12p6x13R =
       + pi162 * (  - 15./64.*me2*mp2 - 43./384.*pow(m66,2) - 43./384.*
         pow(m55,2) - 43./384.*pow(m44,2) - 1./6.*m11*m66 - 1./6.*m11*
         m55 - 1./6.*m11*m44 + 1./64.*pow(m11,2) );

      F12p6x13R +=  + Ab(m11,mu2)*Bb(m14,mu2) * (  - 1./6.*m14*RS156pe
          );

      F12p6x13R +=  + Ab(m11,mu2)*Bb(m15,mu2) * (  - 1./6.*m15*RS146pe
          );

      F12p6x13R +=  + Ab(m11,mu2)*Bb(m16,mu2) * (  - 1./6.*m16*RS145pe
          );

      F12p6x13R +=  + Ab(m14,mu2)*pi16 * (  - 1./8.*m66 - 1./8.*m55 - 1.
         /8.*m44 - 1./4.*m14 );

      F12p6x13R +=  + Ab(m15,mu2)*pi16 * (  - 1./8.*m66 - 1./8.*m55 - 1.
         /8.*m44 - 1./4.*m15 );

      F12p6x13R +=  + Ab(m16,mu2)*pi16 * (  - 1./8.*m66 - 1./8.*m55 - 1.
         /8.*m44 - 1./4.*m16 );

      F12p6x13R +=  + Ab(me2,mu2)*Bb(m14,mu2) * (  - 1./6.*m14*RSe561p
          );

      F12p6x13R +=  + Ab(me2,mu2)*Bb(m15,mu2) * (  - 1./6.*m15*RSe461p
          );

      F12p6x13R +=  + Ab(me2,mu2)*Bb(m16,mu2) * (  - 1./6.*m16*RSe451p
          );

      F12p6x13R +=  + Ab(mp2,mu2)*Bb(m14,mu2) * (  - 1./6.*m14*RSp561e
          );

      F12p6x13R +=  + Ab(mp2,mu2)*Bb(m15,mu2) * (  - 1./6.*m15*RSp461e
          );

      F12p6x13R +=  + Ab(mp2,mu2)*Bb(m16,mu2) * (  - 1./6.*m16*RSp451e
          );

      F12p6x13R +=  + Abeps(m11,mu2)*pi16 * (  - 1./8.*RD1456ep + 1./4.
         *m11 - 1./8.*m11*RC1456pe );

      F12p6x13R +=  + Abeps(m14,mu2)*pi16 * ( 1./8.*m66 + 1./8.*m55 + 1.
         /2.*m44 );

      F12p6x13R +=  + Abeps(m15,mu2)*pi16 * ( 1./8.*m66 + 1./2.*m55 + 1.
         /8.*m44 );

      F12p6x13R +=  + Abeps(m16,mu2)*pi16 * ( 1./2.*m66 + 1./8.*m55 + 1.
         /8.*m44 );

      F12p6x13R +=  + Abeps(m44,mu2)*pi16 * ( 1./8.*m44 - 1./24.*m44*
         RS456pe );

      F12p6x13R +=  + Abeps(m45,mu2)*pi16 * ( 1./8.*m55 + 1./8.*m44 );

      F12p6x13R +=  + Abeps(m46,mu2)*pi16 * ( 1./8.*m66 + 1./8.*m44 );

      F12p6x13R +=  + Abeps(m55,mu2)*pi16 * ( 1./8.*m55 - 1./24.*m55*
         RS546pe );

      F12p6x13R +=  + Abeps(m56,mu2)*pi16 * ( 1./8.*m66 + 1./8.*m55 );

      F12p6x13R +=  + Abeps(m66,mu2)*pi16 * ( 1./8.*m66 - 1./24.*m66*
         RS645pe );

      F12p6x13R +=  + Abeps(me2,mu2)*pi16 * (  - 1./8.*me2*RSe45611p - 
         7./72.*RS(me2,m44,m55,m66,mp2,m44,m44)*me2 + 1./18.*RS(me2,m44
         ,m55,m66,mp2,m44,m44)*m44 - 7./72.*RS(me2,m44,m55,m66,mp2,m55,
         m55)*me2 + 1./18.*RS(me2,m44,m55,m66,mp2,m55,m55)*m55 - 7./72.
         *RS(me2,m44,m55,m66,mp2,m66,m66)*me2 + 1./18.*RS(me2,m44,m55,
         m66,mp2,m66,m66)*m66 );

      F12p6x13R +=  + Abeps(mp2,mu2)*pi16 * (  - 1./8.*mp2*RSp45611e - 
         7./72.*RS(mp2,m44,m55,m66,me2,m44,m44)*mp2 + 1./18.*RS(mp2,m44
         ,m55,m66,me2,m44,m44)*m44 - 7./72.*RS(mp2,m44,m55,m66,me2,m55,
         m55)*mp2 + 1./18.*RS(mp2,m44,m55,m66,me2,m55,m55)*m55 - 7./72.
         *RS(mp2,m44,m55,m66,me2,m66,m66)*mp2 + 1./18.*RS(mp2,m44,m55,
         m66,me2,m66,m66)*m66 );

      F12p6x13R +=  + Bbeps(m11,mu2)*pi16 * (  - 1./8.*m11*RD1456ep );

      F12p6x13R +=  + hh(1,m11,m14,m14,m11,mu2) * ( 1./24.*RD1456ep + 1.
         /24.*m11*RC1456pe - 1./12.*m11*RS156pe );

      F12p6x13R +=  + hh(1,m11,m15,m15,m11,mu2) * ( 1./24.*RD1456ep + 1.
         /24.*m11*RC1456pe - 1./12.*m11*RS146pe );

      F12p6x13R +=  + hh(1,m11,m16,m16,m11,mu2) * ( 1./24.*RD1456ep + 1.
         /24.*m11*RC1456pe - 1./12.*m11*RS145pe );

      F12p6x13R +=  + hh(1,m14,m14,m44,m11,mu2) * (  - 1./8.*m44 + 1./
         24.*m44*RS456pe );

      F12p6x13R +=  + hh(1,m14,m14,me2,m11,mu2) * ( 1./24.*me2*
         RSe45611p - 1./12.*me2*RSe561p + 1./24.*me2*RSe564p );

      F12p6x13R +=  + hh(1,m14,m14,mp2,m11,mu2) * ( 1./24.*mp2*
         RSp45611e - 1./12.*mp2*RSp561e + 1./24.*mp2*RSp564e );

      F12p6x13R +=  + hh(1,m14,m15,m45,m11,mu2) * (  - 1./8.*m55 - 1./8.
         *m44 );

      F12p6x13R +=  + hh(1,m14,m16,m46,m11,mu2) * (  - 1./8.*m66 - 1./8.
         *m44 );

      F12p6x13R +=  + hh(1,m15,m15,m55,m11,mu2) * (  - 1./8.*m55 + 1./
         24.*m55*RS546pe );

      F12p6x13R +=  + hh(1,m15,m15,me2,m11,mu2) * ( 1./24.*me2*
         RSe45611p - 1./12.*me2*RSe461p + 1./24.*me2*RSe465p );

      F12p6x13R +=  + hh(1,m15,m15,mp2,m11,mu2) * ( 1./24.*mp2*
         RSp45611e - 1./12.*mp2*RSp461e + 1./24.*mp2*RSp465e );

      F12p6x13R +=  + hh(1,m15,m16,m56,m11,mu2) * (  - 1./8.*m66 - 1./8.
         *m55 );

      F12p6x13R +=  + hh(1,m16,m16,m66,m11,mu2) * (  - 1./8.*m66 + 1./
         24.*m66*RS645pe );

      F12p6x13R +=  + hh(1,m16,m16,me2,m11,mu2) * ( 1./24.*me2*
         RSe45611p - 1./12.*me2*RSe451p + 1./24.*me2*RSe456p );

      F12p6x13R +=  + hh(1,m16,m16,mp2,m11,mu2) * ( 1./24.*mp2*
         RSp45611e - 1./12.*mp2*RSp451e + 1./24.*mp2*RSp456e );

      F12p6x13R +=  + hh(2,m11,m14,m14,m11,mu2) * ( 1./24.*m11*RD1456ep
          );

      F12p6x13R +=  + hh(2,m11,m15,m15,m11,mu2) * ( 1./24.*m11*RD1456ep
          );

      F12p6x13R +=  + hh(2,m11,m16,m16,m11,mu2) * ( 1./24.*m11*RD1456ep
          );

      F12p6x13R +=  + hhd(1,m11,m11,m11,m11,mu2) * ( 1./6.*pow(m11,2)
          + 1./9.*pow(m11,2)*pow(RC1456pe,2) );

      F12p6x13R +=  + hhd(1,m11,m11,me2,m11,mu2) * ( 2./9.*pow(m11,2)*
         RSe45611p*RC1456pe );

      F12p6x13R +=  + hhd(1,m11,m11,mp2,m11,mu2) * ( 2./9.*pow(m11,2)*
         RSp45611e*RC1456pe );

      F12p6x13R +=  + hhd(1,m11,m14,m14,m11,mu2) * (  - 1./24.*m11*
         RD1456ep + 1./8.*pow(m11,2)*RC1456pe - 1./4.*pow(m11,2)*
         RS156pe - 1./3.*pow(m11,2)*RS14e*RSe561p - 1./3.*RS(m11,m44,
         mp2)*pow(m11,2)*RSp561e );

      F12p6x13R +=  + hhd(1,m11,m15,m15,m11,mu2) * (  - 1./24.*m11*
         RD1456ep + 1./8.*pow(m11,2)*RC1456pe - 1./4.*pow(m11,2)*
         RS146pe - 1./3.*RS(m11,m55,me2)*pow(m11,2)*RSe461p - 1./3.*RS(
         m11,m55,mp2)*pow(m11,2)*RSp461e );

      F12p6x13R +=  + hhd(1,m11,m16,m16,m11,mu2) * (  - 1./24.*m11*
         RD1456ep + 1./8.*pow(m11,2)*RC1456pe - 1./4.*pow(m11,2)*
         RS145pe - 1./3.*pow(m11,2)*RS16e*RSe451p - 1./3.*RS(m11,m66,
         mp2)*pow(m11,2)*RSp451e );

      F12p6x13R +=  + hhd(1,m11,me2,me2,m11,mu2) * ( 1./9.*pow(m11,2)*
         pow(RSe45611p,2) );

      F12p6x13R +=  + hhd(1,m11,me2,mp2,m11,mu2) * ( 2./9.*pow(m11,2)*
         RSp45611e*RSe45611p );

      F12p6x13R +=  + hhd(1,m11,mp2,mp2,m11,mu2) * ( 1./9.*pow(m11,2)*
         pow(RSp45611e,2) );

      F12p6x13R +=  + hhd(1,m14,m14,m44,m11,mu2) * ( 1./8.*m11*m44 - 1./
         24.*m11*m44*RS456pe );

      F12p6x13R +=  + hhd(1,m14,m14,me2,m11,mu2) * (  - 1./24.*m11*me2*
         RS144ee*RSe564p - 1./3.*pow(m11,2)*RS14e*RSe561p - 1./6.*RS(
         me2,m44,m44,m11,m11)*pow(m11,2)*RSe564p );

      F12p6x13R +=  + hhd(1,m14,m14,mp2,m11,mu2) * (  - 1./24.*RS(m11,
         m44,m44,mp2,mp2)*m11*mp2*RSp564e - 1./3.*RS(m11,m44,mp2)*pow(
         m11,2)*RSp561e - 1./6.*RS(mp2,m44,m44,m11,m11)*pow(m11,2)*
         RSp564e );

      F12p6x13R +=  + hhd(1,m14,m15,m45,m11,mu2) * ( 1./8.*m11*m55 + 1./
         8.*m11*m44 );

      F12p6x13R +=  + hhd(1,m14,m16,m46,m11,mu2) * ( 1./8.*m11*m66 + 1./
         8.*m11*m44 );

      F12p6x13R +=  + hhd(1,m15,m15,m55,m11,mu2) * ( 1./8.*m11*m55 - 1./
         24.*m11*m55*RS546pe );

      F12p6x13R +=  + hhd(1,m15,m15,me2,m11,mu2) * (  - 1./24.*RS(m11,
         m55,m55,me2,me2)*m11*me2*RSe465p - 1./3.*RS(m11,m55,me2)*pow(
         m11,2)*RSe461p - 1./6.*RS(me2,m55,m55,m11,m11)*pow(m11,2)*
         RSe465p );

      F12p6x13R +=  + hhd(1,m15,m15,mp2,m11,mu2) * (  - 1./24.*RS(m11,
         m55,m55,mp2,mp2)*m11*mp2*RSp465e - 1./3.*RS(m11,m55,mp2)*pow(
         m11,2)*RSp461e - 1./6.*RS(mp2,m55,m55,m11,m11)*pow(m11,2)*
         RSp465e );

      F12p6x13R +=  + hhd(1,m15,m16,m56,m11,mu2) * ( 1./8.*m11*m66 + 1./
         8.*m11*m55 );

      F12p6x13R +=  + hhd(1,m16,m16,m66,m11,mu2) * ( 1./8.*m11*m66 - 1./
         24.*m11*m66*RS645pe );

      F12p6x13R +=  + hhd(1,m16,m16,me2,m11,mu2) * (  - 1./24.*m11*me2*
         RS166ee*RSe456p - 1./6.*pow(m11,2)*RSe6611*RSe456p - 1./3.*
         pow(m11,2)*RS16e*RSe451p );

      F12p6x13R +=  + hhd(1,m16,m16,mp2,m11,mu2) * (  - 1./24.*RS(m11,
         m66,m66,mp2,mp2)*m11*mp2*RSp456e - 1./3.*RS(m11,m66,mp2)*pow(
         m11,2)*RSp451e - 1./6.*RS(mp2,m66,m66,m11,m11)*pow(m11,2)*
         RSp456e );

      F12p6x13R +=  + hhd(2,m11,m11,m11,m11,mu2) * ( 2./9.*pow(m11,2)*
         RD1456ep*RC1456pe );

      F12p6x13R +=  + hhd(2,m11,m11,me2,m11,mu2) * ( 2./9.*pow(m11,2)*
         RSe45611p*RD1456ep );

      F12p6x13R +=  + hhd(2,m11,m11,mp2,m11,mu2) * ( 2./9.*pow(m11,2)*
         RSp45611e*RD1456ep );

      F12p6x13R +=  + hhd(2,m11,m14,m14,m11,mu2) * ( 1./8.*pow(m11,2)*
         RD1456ep );

      F12p6x13R +=  + hhd(2,m11,m15,m15,m11,mu2) * ( 1./8.*pow(m11,2)*
         RD1456ep );

      F12p6x13R +=  + hhd(2,m11,m16,m16,m11,mu2) * ( 1./8.*pow(m11,2)*
         RD1456ep );

      F12p6x13R +=  + hhd(5,m11,m11,m11,m11,mu2) * ( 1./9.*pow(m11,2)*
         pow(RD1456ep,2) );

      F12p6x13R +=  + hh1d(1,m11,m14,m14,m11,mu2) * ( 1./3.*pow(m11,2)*
         RS14e*RSe561p + 1./3.*RS(m11,m44,mp2)*pow(m11,2)*RSp561e );

      F12p6x13R +=  + hh1d(1,m11,m15,m15,m11,mu2) * ( 1./3.*RS(m11,m55,
         me2)*pow(m11,2)*RSe461p + 1./3.*RS(m11,m55,mp2)*pow(m11,2)*
         RSp461e );

      F12p6x13R +=  + hh1d(1,m11,m16,m16,m11,mu2) * ( 1./3.*pow(m11,2)*
         RS16e*RSe451p + 1./3.*RS(m11,m66,mp2)*pow(m11,2)*RSp451e );

      F12p6x13R +=  + hh1d(1,m14,m14,me2,m11,mu2) * ( 2./3.*pow(m11,2)*
         RS14e*RSe561p );

      F12p6x13R +=  + hh1d(1,m14,m14,mp2,m11,mu2) * ( 2./3.*RS(m11,m44,
         mp2)*pow(m11,2)*RSp561e );

      F12p6x13R +=  + hh1d(1,m15,m15,me2,m11,mu2) * ( 2./3.*RS(m11,m55,
         me2)*pow(m11,2)*RSe461p );

      F12p6x13R +=  + hh1d(1,m15,m15,mp2,m11,mu2) * ( 2./3.*RS(m11,m55,
         mp2)*pow(m11,2)*RSp461e );

      F12p6x13R +=  + hh1d(1,m16,m16,me2,m11,mu2) * ( 2./3.*pow(m11,2)*
         RS16e*RSe451p );

      F12p6x13R +=  + hh1d(1,m16,m16,mp2,m11,mu2) * ( 2./3.*RS(m11,m66,
         mp2)*pow(m11,2)*RSp451e );

      F12p6x13R +=  + hh1d(3,m14,m11,m14,m11,mu2) * (  - 2./3.*pow(
         m11,2)*RD1456ep );

      F12p6x13R +=  + hh1d(3,m15,m11,m15,m11,mu2) * (  - 2./3.*pow(
         m11,2)*RD1456ep );

      F12p6x13R +=  + hh1d(3,m16,m11,m16,m11,mu2) * (  - 2./3.*pow(
         m11,2)*RD1456ep );

      F12p6x13R +=  + hh21d(1,m11,m14,m14,m11,mu2) * (  - 1./8.*pow(
         m11,2)*RC1456pe + 1./4.*pow(m11,2)*RS156pe );

      F12p6x13R +=  + hh21d(1,m11,m15,m15,m11,mu2) * (  - 1./8.*pow(
         m11,2)*RC1456pe + 1./4.*pow(m11,2)*RS146pe );

      F12p6x13R +=  + hh21d(1,m11,m16,m16,m11,mu2) * (  - 1./8.*pow(
         m11,2)*RC1456pe + 1./4.*pow(m11,2)*RS145pe );

      F12p6x13R +=  + hh21d(1,m44,m14,m14,m11,mu2) * ( 3./8.*pow(m11,2)
          - 1./8.*pow(m11,2)*RS456pe );

      F12p6x13R +=  + hh21d(1,m45,m14,m15,m11,mu2) * ( 3./4.*pow(m11,2)
          );

      F12p6x13R +=  + hh21d(1,m46,m14,m16,m11,mu2) * ( 3./4.*pow(m11,2)
          );

      F12p6x13R +=  + hh21d(1,m55,m15,m15,m11,mu2) * ( 3./8.*pow(m11,2)
          - 1./8.*pow(m11,2)*RS546pe );

      F12p6x13R +=  + hh21d(1,m56,m15,m16,m11,mu2) * ( 3./4.*pow(m11,2)
          );

      F12p6x13R +=  + hh21d(1,m66,m16,m16,m11,mu2) * ( 3./8.*pow(m11,2)
          - 1./8.*pow(m11,2)*RS645pe );

      F12p6x13R +=  + hh21d(1,me2,m14,m14,m11,mu2) * (  - 1./8.*pow(
         m11,2)*RSe45611p + 1./4.*pow(m11,2)*RSe561p - 1./8.*pow(m11,2)
         *RSe564p );

      F12p6x13R +=  + hh21d(1,me2,m15,m15,m11,mu2) * (  - 1./8.*pow(
         m11,2)*RSe45611p + 1./4.*pow(m11,2)*RSe461p - 1./8.*pow(m11,2)
         *RSe465p );

      F12p6x13R +=  + hh21d(1,me2,m16,m16,m11,mu2) * (  - 1./8.*pow(
         m11,2)*RSe45611p + 1./4.*pow(m11,2)*RSe451p - 1./8.*pow(m11,2)
         *RSe456p );

      F12p6x13R +=  + hh21d(1,mp2,m14,m14,m11,mu2) * (  - 1./8.*pow(
         m11,2)*RSp45611e + 1./4.*pow(m11,2)*RSp561e - 1./8.*pow(m11,2)
         *RSp564e );

      F12p6x13R +=  + hh21d(1,mp2,m15,m15,m11,mu2) * (  - 1./8.*pow(
         m11,2)*RSp45611e + 1./4.*pow(m11,2)*RSp461e - 1./8.*pow(m11,2)
         *RSp465e );

      F12p6x13R +=  + hh21d(1,mp2,m16,m16,m11,mu2) * (  - 1./8.*pow(
         m11,2)*RSp45611e + 1./4.*pow(m11,2)*RSp451e - 1./8.*pow(m11,2)
         *RSp456e );

      F12p6x13R +=  + hh21d(2,m11,m14,m14,m11,mu2) * (  - 1./8.*pow(
         m11,2)*RD1456ep );

      F12p6x13R +=  + hh21d(2,m11,m15,m15,m11,mu2) * (  - 1./8.*pow(
         m11,2)*RD1456ep );

      F12p6x13R +=  + hh21d(2,m11,m16,m16,m11,mu2) * (  - 1./8.*pow(
         m11,2)*RD1456ep );

  return F12p6x13R/pow(f0,4);
}
// 2 valence mass 3 sea mass *****************************************

double fv2s3nf3p4(const quarkmassnf mass, const Linf Liin){
  return fv2s3nf3p4L(mass,Liin)+fv2s3nf3p4R(mass);
}

double fv2s3nf3p4L(const quarkmassnf mass, const Linf Liin){
  double L0r,L1r,L2r,L3r,L4r,L5r,L6r,L7r,L8r,L9r,L10r,L11r,H1r,H2r,mu;
  int nfLi;
  Liin.out(L0r,L1r,L2r,L3r,L4r,L5r,L6r,L7r,L8r,L9r,L10r,L11r,H1r,H2r,mu,nfLi);
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if (mu0 != mu){
    cout << "ERROR: differing mu's in fv2s3nf3p4L\n";}
  if ( (nq != 5) || (B0mq.size() != 5)){
    cout << "ERROR: fv2s3nf3p4L probably called with wrong number"
	 << "of quark masses\n";}
  // two masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m33 = 2.*B0mq[1];
  double m44 = 2.*B0mq[2];
  double m55 = 2.*B0mq[3];
  double m66 = 2.*B0mq[4];

  double F13p4x23L =
       + 4*m66*L4r + 4*m55*L4r + 4*m44*L4r + 2*m33*L5r + 2*m11*L5r;

  return F13p4x23L/pow(f0,2);
}


double fv2s3nf3p4R(const quarkmassnf mass){
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if ( (nq != 5) || (B0mq.size() != 5)){
    cout << "ERROR: fv2s3nf3p4R probably called with wrong number"
	 << "of quark masses\n";}
  // five masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m33 = 2.*B0mq[1];
  double m44 = 2.*B0mq[2];
  double m55 = 2.*B0mq[3];
  double m66 = 2.*B0mq[4];
  double m14 = (m11+m44)/2;
  double m15 = (m11+m55)/2;
  double m16 = (m11+m66)/2;
  double m34 = (m33+m44)/2;
  double m35 = (m33+m55)/2;
  double m36 = (m33+m66)/2;

  double xb1 = (m44+m55+m66)/3.;
  double msqr = sqrt(xb1*xb1-(m44*m55+m44*m66+m55*m66)/3.);
  double mp2 = xb1-msqr; 
  double me2 = xb1+msqr;
  double mu2 = mu0*mu0;
  double RSe45611p = RS(me2,m44,m55,m66,m11,m11,mp2);
  double RSp45611e = RS(mp2,m44,m55,m66,m11,m11,me2);
  double RSe45613p = RS(me2,m44,m55,m66,m11,m33,mp2);
  double RSp45613e = RS(mp2,m44,m55,m66,m11,m33,me2);
  double RSe45633p = RS(me2,m44,m55,m66,m33,m33,mp2);
  double RSp45633e = RS(mp2,m44,m55,m66,m33,m33,me2);
  double RS14563pe = RS(m11,m44,m55,m66,m33,mp2,me2);
  double RS34561pe = RS(m33,m44,m55,m66,m11,mp2,me2);

  double RC1456pe = RC(m11,m44,m55,m66,mp2,me2);
  double RC3456pe = RC(m33,m44,m55,m66,mp2,me2);
  double RD1456ep = RD(m11,m44,m55,m66,me2,mp2);
  double RD3456ep = RD(m33,m44,m55,m66,me2,mp2);

  double F13p4x23R =
       + Ab(m11,mu2) * ( 1./6.*RS14563pe - 1./12.*RC1456pe );

      F13p4x23R +=  + Ab(m14,mu2) * ( 1./4. );

      F13p4x23R +=  + Ab(m15,mu2) * ( 1./4. );

      F13p4x23R +=  + Ab(m16,mu2) * ( 1./4. );

      F13p4x23R +=  + Ab(m33,mu2) * ( 1./6.*RS34561pe - 1./12.*RC3456pe
          );

      F13p4x23R +=  + Ab(m34,mu2) * ( 1./4. );

      F13p4x23R +=  + Ab(m35,mu2) * ( 1./4. );

      F13p4x23R +=  + Ab(m36,mu2) * ( 1./4. );

      F13p4x23R +=  + Ab(me2,mu2) * ( 1./6.*RSe45613p - 1./12.*
         RSe45633p - 1./12.*RSe45611p );

      F13p4x23R +=  + Ab(mp2,mu2) * ( 1./6.*RSp45613e - 1./12.*
         RSp45633e - 1./12.*RSp45611e );

      F13p4x23R +=  + Bb(m11,mu2) * (  - 1./12.*RD1456ep );

      F13p4x23R +=  + Bb(m33,mu2) * (  - 1./12.*RD3456ep );

  return F13p4x23R/pow(f0,2);
}

double fv2s3nf3p6(const quarkmassnf mass, const Linf Liin, const Ki Kiin){
  return fv2s3nf3p6L(mass,Liin)+fv2s3nf3p6K(mass,Kiin)+fv2s3nf3p6R(mass);
}

double fv2s3nf3p6L(const quarkmassnf mass, const Linf Liin){
  double L0r,L1r,L2r,L3r,L4r,L5r,L6r,L7r,L8r,L9r,L10r,L11r,H1r,H2r,mu;
  int nfLi;
  Liin.out(L0r,L1r,L2r,L3r,L4r,L5r,L6r,L7r,L8r,L9r,L10r,L11r,H1r,H2r,mu,nfLi);
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if (mu0 != mu){
    cout << "ERROR: differing mu's in fv2s3nf3p6L\n";}
  if ( (nq != 5) || (B0mq.size() != 5)){
    cout << "ERROR: fv2s3nf3p6L probably called with wrong number"
	 << "of quark masses\n";}
  // four masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m33 = 2.*B0mq[1];
  double m44 = 2.*B0mq[2];
  double m55 = 2.*B0mq[3];
  double m66 = 2.*B0mq[4];
  double m13 = (m11+m33)/2;
  double m14 = (m11+m44)/2;
  double m15 = (m11+m55)/2;
  double m16 = (m11+m66)/2;
  double m34 = (m33+m44)/2;
  double m35 = (m33+m55)/2;
  double m36 = (m33+m66)/2;
  double m45 = (m44+m55)/2;
  double m46 = (m44+m66)/2;
  double m56 = (m55+m66)/2;

  double xb1 = (m44+m55+m66)/3.;
  double msqr = sqrt(xb1*xb1-(m44*m55+m44*m66+m55*m66)/3.);
  double mp2 = xb1-msqr; 
  double me2 = xb1+msqr;
  double mu2 = mu0*mu0;

  double RN13 = RN(m11,m33);
  double RN1p = RN(m11,mp2);
  double RN1e = RN(m11,me2);
  double RN3p = RN(m33,mp2);
  double RN3e = RN(m33,me2);
  double RNep = RN(me2,mp2);

  double RSp456e = RS(mp2,m44,m55,m66,me2);
  double RSp465e = RS(mp2,m44,m66,m55,me2);
  double RSp564e = RS(mp2,m55,m66,m44,me2);
  double RSe456p = RS(me2,m44,m55,m66,mp2);
  double RSe465p = RS(me2,m44,m66,m55,mp2);
  double RSe564p = RS(me2,m55,m66,m44,mp2);
  double RSp451e = RS(mp2,m44,m55,m11,me2);
  double RSp461e = RS(mp2,m44,m66,m11,me2);
  double RSp561e = RS(mp2,m55,m66,m11,me2);
  double RSe451p = RS(me2,m44,m55,m11,mp2);
  double RSe461p = RS(me2,m44,m66,m11,mp2);
  double RSe561p = RS(me2,m55,m66,m11,mp2);
  double RSp453e = RS(mp2,m44,m55,m33,me2);
  double RSp463e = RS(mp2,m44,m66,m33,me2);
  double RSp563e = RS(mp2,m55,m66,m33,me2);
  double RSe453p = RS(me2,m44,m55,m33,mp2);
  double RSe463p = RS(me2,m44,m66,m33,mp2);
  double RSe563p = RS(me2,m55,m66,m33,mp2);
  double RS456pe = RS(m44,m55,m66,mp2,me2);
  double RS546pe = RS(m55,m44,m66,mp2,me2);
  double RS645pe = RS(m66,m44,m55,mp2,me2);

  double RS156pe = RS(m11,m55,m66,mp2,me2);
  double RS146pe = RS(m11,m44,m66,mp2,me2);
  double RS145pe = RS(m11,m44,m55,mp2,me2);
  double RS356pe = RS(m33,m55,m66,mp2,me2);
  double RS346pe = RS(m33,m44,m66,mp2,me2);
  double RS345pe = RS(m33,m44,m55,mp2,me2);

  double RD1456ep = RD(m11,m44,m55,m66,me2,mp2);
  double RD3456ep = RD(m33,m44,m55,m66,me2,mp2);

  double RS1456ppe = RS(m11,m44,m55,m66,mp2,mp2,me2);
  double RS1456pee = RS(m11,m44,m55,m66,mp2,me2,me2);
  double RSe45611p = RS(me2,m44,m55,m66,m11,m11,mp2);
  double RSp45611e = RS(mp2,m44,m55,m66,m11,m11,me2);
  double RSe45633p = RS(me2,m44,m55,m66,m33,m33,mp2);
  double RSp45633e = RS(mp2,m44,m55,m66,m33,m33,me2);
  double RS3456ppe = RS(m33,m44,m55,m66,mp2,mp2,me2);
  double RS3456pee = RS(m33,m44,m55,m66,mp2,me2,me2);
  double RSp45613e = RS(mp2,m44,m55,m66,m11,m33,me2);
  double RSe45613p = RS(me2,m44,m55,m66,m11,m33,mp2);
  double RS14563pe = RS(m11,m44,m55,m66,m33,mp2,me2);
  double RS34561pe = RS(m33,m44,m55,m66,m11,mp2,me2);

  double F13p6x23L =
       + pi16 * ( 1./3.*pow(mp2,2)*L3r*RSp45613e + 1./6.*pow(mp2,2)*L3r
         *RSp45633e + 1./6.*pow(mp2,2)*L3r*RSp45611e + 1./3.*pow(mp2,2)
         *L2r*RSp564e + 1./3.*pow(mp2,2)*L2r*RSp465e + 1./3.*pow(mp2,2)
         *L2r*RSp456e + 1./3.*pow(mp2,2)*L0r*RSp45633e + 1./3.*pow(
         mp2,2)*L0r*RSp45611e + 1./3.*pow(me2,2)*L3r*RSe45613p + 1./6.*
         pow(me2,2)*L3r*RSe45633p + 1./6.*pow(me2,2)*L3r*RSe45611p + 1./
         3.*pow(me2,2)*L2r*RSe564p + 1./3.*pow(me2,2)*L2r*RSe465p + 1./
         3.*pow(me2,2)*L2r*RSe456p + 1./3.*pow(me2,2)*L0r*RSe45633p + 1.
         /3.*pow(me2,2)*L0r*RSe45611p - 1./4.*pow(m66,2)*L3r - 2*pow(
         m66,2)*L2r + 1./3.*pow(m66,2)*L2r*RS645pe - 1./2.*pow(m66,2)*
         L0r - m55*m66*L2r - 1./4.*pow(m55,2)*L3r - 2*pow(m55,2)*L2r + 
         1./3.*pow(m55,2)*L2r*RS546pe - 1./2.*pow(m55,2)*L0r - m44*m66*
         L2r - m44*m55*L2r - 1./4.*pow(m44,2)*L3r - 2*pow(m44,2)*L2r + 
         1./3.*pow(m44,2)*L2r*RS456pe - 1./2.*pow(m44,2)*L0r + 1./3.*
         m33*L3r*RD3456ep + 2./3.*m33*L0r*RD3456ep - 1./4.*m33*m66*L3r
          - 1./2.*m33*m66*L0r );

      F13p6x23L +=  + pi16 * (  - 1./4.*m33*m55*L3r - 1./2.*m33*m55*L0r
          - 1./4.*m33*m44*L3r - 1./2.*m33*m44*L0r - 3./8.*pow(m33,2)*
         L3r + 1./3.*pow(m33,2)*L3r*RS34561pe - 1./6.*pow(m33,2)*L3r*
         RS3456pee - 1./6.*pow(m33,2)*L3r*RS3456ppe + 1./6.*pow(m33,2)*
         L3r*RS345pe + 1./6.*pow(m33,2)*L3r*RS346pe + 1./6.*pow(m33,2)*
         L3r*RS356pe - 1./4.*pow(m33,2)*L2r - 1./2.*pow(m33,2)*L1r - 3./
         4.*pow(m33,2)*L0r - 1./3.*pow(m33,2)*L0r*RS3456pee - 1./3.*
         pow(m33,2)*L0r*RS3456ppe + 1./3.*pow(m33,2)*L0r*RS345pe + 1./3.
         *pow(m33,2)*L0r*RS346pe + 1./3.*pow(m33,2)*L0r*RS356pe + 1./3.
         *m11*L3r*RD1456ep + 2./3.*m11*L0r*RD1456ep - 1./4.*m11*m66*L3r
          - 1./2.*m11*m66*L0r - 1./4.*m11*m55*L3r - 1./2.*m11*m55*L0r
          - 1./4.*m11*m44*L3r - 1./2.*m11*m44*L0r - 1./2.*m11*m33*L2r
          - m11*m33*L1r - 3./8.*pow(m11,2)*L3r + 1./3.*pow(m11,2)*L3r*
         RS14563pe - 1./6.*pow(m11,2)*L3r*RS1456pee - 1./6.*pow(m11,2)*
         L3r*RS1456ppe + 1./6.*pow(m11,2)*L3r*RS145pe + 1./6.*pow(
         m11,2)*L3r*RS146pe );

      F13p6x23L +=  + pi16 * ( 1./6.*pow(m11,2)*L3r*RS156pe - 1./4.*
         pow(m11,2)*L2r - 1./2.*pow(m11,2)*L1r - 3./4.*pow(m11,2)*L0r
          - 1./3.*pow(m11,2)*L0r*RS1456pee - 1./3.*pow(m11,2)*L0r*
         RS1456ppe + 1./3.*pow(m11,2)*L0r*RS145pe + 1./3.*pow(m11,2)*
         L0r*RS146pe + 1./3.*pow(m11,2)*L0r*RS156pe );

      F13p6x23L +=  + Ab(m11,mu2) * ( 5./3.*L3r*RD1456ep + 2./3.*L0r*
         RD1456ep + 4./9.*m66*L5r*RS145pe*RSe453p + 4./9.*m66*L5r*
         RS145pe*RSp453e - 2./9.*m66*L5r*pow(RS145pe,2) - 4./9.*m66*L5r
         *RSe451p*RS145pe - 4./9.*m66*L5r*RSp451e*RS145pe - 2./3.*m66*
         L4r*RS14563pe - 4./3.*m66*L4r*RSe45613p - 4./3.*m66*L4r*
         RSp45613e - m66*L4r*RS1456pee - m66*L4r*RS1456ppe + 4./3.*m66*
         L4r*RSe45611p + 4./3.*m66*L4r*RSp45611e + m66*L4r*RS145pe + 4./
         9.*m66*L4r*RS145pe*RSe453p + 4./9.*m66*L4r*RS145pe*RSp453e - 2.
         /9.*m66*L4r*pow(RS145pe,2) + m66*L4r*RS146pe + 4./9.*m66*L4r*
         RS146pe*RSe463p + 4./9.*m66*L4r*RS146pe*RSp463e - 2./9.*m66*
         L4r*pow(RS146pe,2) + m66*L4r*RS156pe + 4./9.*m66*L4r*RS156pe*
         RSe563p + 4./9.*m66*L4r*RS156pe*RSp563e - 2./9.*m66*L4r*pow(
         RS156pe,2) - 4./9.*m66*L4r*RSe561p*RS156pe - 4./9.*m66*L4r*
         RSe461p*RS146pe - 4./9.*m66*L4r*RSe451p*RS145pe - 4./9.*m66*
         L4r*RSp561e*RS156pe - 4./9.*m66*L4r*RSp461e*RS146pe - 4./9.*
         m66*L4r*RSp451e*RS145pe );

      F13p6x23L +=  + Ab(m11,mu2) * ( 4./9.*m66*mp2*L5r*RN1p*RS145pe*
         RSp453e - 4./9.*m66*mp2*L5r*RN1p*RSp451e*RS145pe - 4./3.*m66*
         mp2*L4r*RN1p*RSp45613e + 4./3.*m66*mp2*L4r*RN1p*RSp45611e + 4./
         9.*m66*mp2*L4r*RN1p*RS145pe*RSp453e + 4./9.*m66*mp2*L4r*RN1p*
         RS146pe*RSp463e + 4./9.*m66*mp2*L4r*RN1p*RS156pe*RSp563e - 4./
         9.*m66*mp2*L4r*RN1p*RSp561e*RS156pe - 4./9.*m66*mp2*L4r*RN1p*
         RSp461e*RS146pe - 4./9.*m66*mp2*L4r*RN1p*RSp451e*RS145pe + 4./
         9.*m66*me2*L5r*RN1e*RS145pe*RSe453p - 4./9.*m66*me2*L5r*RN1e*
         RSe451p*RS145pe - 4./3.*m66*me2*L4r*RN1e*RSe45613p + 4./3.*m66
         *me2*L4r*RN1e*RSe45611p + 4./9.*m66*me2*L4r*RN1e*RS145pe*
         RSe453p + 4./9.*m66*me2*L4r*RN1e*RS146pe*RSe463p + 4./9.*m66*
         me2*L4r*RN1e*RS156pe*RSe563p - 4./9.*m66*me2*L4r*RN1e*RSe561p*
         RS156pe - 4./9.*m66*me2*L4r*RN1e*RSe461p*RS146pe - 4./9.*m66*
         me2*L4r*RN1e*RSe451p*RS145pe - 8./9.*pow(m66,2)*L8r*RN1p*
         RS145pe*RSp453e + 8./9.*pow(m66,2)*L8r*RN1p*RSp451e*RS145pe - 
         8./9.*pow(m66,2)*L8r*RN1e*RS145pe*RSe453p );

      F13p6x23L +=  + Ab(m11,mu2) * ( 8./9.*pow(m66,2)*L8r*RN1e*RSe451p
         *RS145pe - 8./9.*pow(m66,2)*L8r*RN13*RS145pe*RS345pe - 8./9.*
         pow(m66,2)*L7r*RN1p*RS145pe*RSp453e + 8./9.*pow(m66,2)*L7r*
         RN1p*RSp451e*RS145pe - 8./9.*pow(m66,2)*L7r*RN1e*RS145pe*
         RSe453p + 8./9.*pow(m66,2)*L7r*RN1e*RSe451p*RS145pe - 8./9.*
         pow(m66,2)*L7r*RN13*RS145pe*RS345pe - 8./9.*pow(m66,2)*L6r*
         RN1p*RS145pe*RSp453e + 8./9.*pow(m66,2)*L6r*RN1p*RSp451e*
         RS145pe - 8./9.*pow(m66,2)*L6r*RN1e*RS145pe*RSe453p + 8./9.*
         pow(m66,2)*L6r*RN1e*RSe451p*RS145pe - 8./9.*pow(m66,2)*L6r*
         RN13*RS145pe*RS345pe + 4./9.*m55*L5r*RS146pe*RSe463p + 4./9.*
         m55*L5r*RS146pe*RSp463e - 2./9.*m55*L5r*pow(RS146pe,2) - 4./9.
         *m55*L5r*RSe461p*RS146pe - 4./9.*m55*L5r*RSp461e*RS146pe - 2./
         3.*m55*L4r*RS14563pe - 4./3.*m55*L4r*RSe45613p - 4./3.*m55*L4r
         *RSp45613e - m55*L4r*RS1456pee - m55*L4r*RS1456ppe + 4./3.*m55
         *L4r*RSe45611p + 4./3.*m55*L4r*RSp45611e + m55*L4r*RS145pe + 4.
         /9.*m55*L4r*RS145pe*RSe453p );

      F13p6x23L +=  + Ab(m11,mu2) * ( 4./9.*m55*L4r*RS145pe*RSp453e - 2.
         /9.*m55*L4r*pow(RS145pe,2) + m55*L4r*RS146pe + 4./9.*m55*L4r*
         RS146pe*RSe463p + 4./9.*m55*L4r*RS146pe*RSp463e - 2./9.*m55*
         L4r*pow(RS146pe,2) + m55*L4r*RS156pe + 4./9.*m55*L4r*RS156pe*
         RSe563p + 4./9.*m55*L4r*RS156pe*RSp563e - 2./9.*m55*L4r*pow(
         RS156pe,2) - 4./9.*m55*L4r*RSe561p*RS156pe - 4./9.*m55*L4r*
         RSe461p*RS146pe - 4./9.*m55*L4r*RSe451p*RS145pe - 4./9.*m55*
         L4r*RSp561e*RS156pe - 4./9.*m55*L4r*RSp461e*RS146pe - 4./9.*
         m55*L4r*RSp451e*RS145pe + 4./9.*m55*mp2*L5r*RN1p*RS146pe*
         RSp463e - 4./9.*m55*mp2*L5r*RN1p*RSp461e*RS146pe - 4./3.*m55*
         mp2*L4r*RN1p*RSp45613e + 4./3.*m55*mp2*L4r*RN1p*RSp45611e + 4./
         9.*m55*mp2*L4r*RN1p*RS145pe*RSp453e + 4./9.*m55*mp2*L4r*RN1p*
         RS146pe*RSp463e + 4./9.*m55*mp2*L4r*RN1p*RS156pe*RSp563e - 4./
         9.*m55*mp2*L4r*RN1p*RSp561e*RS156pe - 4./9.*m55*mp2*L4r*RN1p*
         RSp461e*RS146pe - 4./9.*m55*mp2*L4r*RN1p*RSp451e*RS145pe + 4./
         9.*m55*me2*L5r*RN1e*RS146pe*RSe463p );

      F13p6x23L +=  + Ab(m11,mu2) * (  - 4./9.*m55*me2*L5r*RN1e*RSe461p
         *RS146pe - 4./3.*m55*me2*L4r*RN1e*RSe45613p + 4./3.*m55*me2*
         L4r*RN1e*RSe45611p + 4./9.*m55*me2*L4r*RN1e*RS145pe*RSe453p + 
         4./9.*m55*me2*L4r*RN1e*RS146pe*RSe463p + 4./9.*m55*me2*L4r*
         RN1e*RS156pe*RSe563p - 4./9.*m55*me2*L4r*RN1e*RSe561p*RS156pe
          - 4./9.*m55*me2*L4r*RN1e*RSe461p*RS146pe - 4./9.*m55*me2*L4r*
         RN1e*RSe451p*RS145pe - 8./9.*m55*m66*L7r*RN1p*RS145pe*RSp463e
          - 8./9.*m55*m66*L7r*RN1p*RS146pe*RSp453e + 8./9.*m55*m66*L7r*
         RN1p*RSp461e*RS145pe + 8./9.*m55*m66*L7r*RN1p*RSp451e*RS146pe
          - 8./9.*m55*m66*L7r*RN1e*RS145pe*RSe463p - 8./9.*m55*m66*L7r*
         RN1e*RS146pe*RSe453p + 8./9.*m55*m66*L7r*RN1e*RSe461p*RS145pe
          + 8./9.*m55*m66*L7r*RN1e*RSe451p*RS146pe - 8./9.*m55*m66*L7r*
         RN13*RS145pe*RS346pe - 8./9.*m55*m66*L7r*RN13*RS146pe*RS345pe
          - 8./9.*m55*m66*L6r*RN1p*RS145pe*RSp453e - 8./9.*m55*m66*L6r*
         RN1p*RS146pe*RSp463e + 8./9.*m55*m66*L6r*RN1p*RSp461e*RS146pe
          + 8./9.*m55*m66*L6r*RN1p*RSp451e*RS145pe );

      F13p6x23L +=  + Ab(m11,mu2) * (  - 8./9.*m55*m66*L6r*RN1e*RS145pe
         *RSe453p - 8./9.*m55*m66*L6r*RN1e*RS146pe*RSe463p + 8./9.*m55*
         m66*L6r*RN1e*RSe461p*RS146pe + 8./9.*m55*m66*L6r*RN1e*RSe451p*
         RS145pe - 8./9.*m55*m66*L6r*RN13*RS145pe*RS345pe - 8./9.*m55*
         m66*L6r*RN13*RS146pe*RS346pe - 8./9.*pow(m55,2)*L8r*RN1p*
         RS146pe*RSp463e + 8./9.*pow(m55,2)*L8r*RN1p*RSp461e*RS146pe - 
         8./9.*pow(m55,2)*L8r*RN1e*RS146pe*RSe463p + 8./9.*pow(m55,2)*
         L8r*RN1e*RSe461p*RS146pe - 8./9.*pow(m55,2)*L8r*RN13*RS146pe*
         RS346pe - 8./9.*pow(m55,2)*L7r*RN1p*RS146pe*RSp463e + 8./9.*
         pow(m55,2)*L7r*RN1p*RSp461e*RS146pe - 8./9.*pow(m55,2)*L7r*
         RN1e*RS146pe*RSe463p + 8./9.*pow(m55,2)*L7r*RN1e*RSe461p*
         RS146pe - 8./9.*pow(m55,2)*L7r*RN13*RS146pe*RS346pe - 8./9.*
         pow(m55,2)*L6r*RN1p*RS146pe*RSp463e + 8./9.*pow(m55,2)*L6r*
         RN1p*RSp461e*RS146pe - 8./9.*pow(m55,2)*L6r*RN1e*RS146pe*
         RSe463p + 8./9.*pow(m55,2)*L6r*RN1e*RSe461p*RS146pe - 8./9.*
         pow(m55,2)*L6r*RN13*RS146pe*RS346pe );

      F13p6x23L +=  + Ab(m11,mu2) * ( 4./9.*m44*L5r*RS156pe*RSe563p + 4.
         /9.*m44*L5r*RS156pe*RSp563e - 2./9.*m44*L5r*pow(RS156pe,2) - 4.
         /9.*m44*L5r*RSe561p*RS156pe - 4./9.*m44*L5r*RSp561e*RS156pe - 
         2./3.*m44*L4r*RS14563pe - 4./3.*m44*L4r*RSe45613p - 4./3.*m44*
         L4r*RSp45613e - m44*L4r*RS1456pee - m44*L4r*RS1456ppe + 4./3.*
         m44*L4r*RSe45611p + 4./3.*m44*L4r*RSp45611e + m44*L4r*RS145pe
          + 4./9.*m44*L4r*RS145pe*RSe453p + 4./9.*m44*L4r*RS145pe*
         RSp453e - 2./9.*m44*L4r*pow(RS145pe,2) + m44*L4r*RS146pe + 4./
         9.*m44*L4r*RS146pe*RSe463p + 4./9.*m44*L4r*RS146pe*RSp463e - 2.
         /9.*m44*L4r*pow(RS146pe,2) + m44*L4r*RS156pe + 4./9.*m44*L4r*
         RS156pe*RSe563p + 4./9.*m44*L4r*RS156pe*RSp563e - 2./9.*m44*
         L4r*pow(RS156pe,2) - 4./9.*m44*L4r*RSe561p*RS156pe - 4./9.*m44
         *L4r*RSe461p*RS146pe - 4./9.*m44*L4r*RSe451p*RS145pe - 4./9.*
         m44*L4r*RSp561e*RS156pe - 4./9.*m44*L4r*RSp461e*RS146pe - 4./9.
         *m44*L4r*RSp451e*RS145pe + 4./9.*m44*mp2*L5r*RN1p*RS156pe*
         RSp563e );

      F13p6x23L +=  + Ab(m11,mu2) * (  - 4./9.*m44*mp2*L5r*RN1p*RSp561e
         *RS156pe - 4./3.*m44*mp2*L4r*RN1p*RSp45613e + 4./3.*m44*mp2*
         L4r*RN1p*RSp45611e + 4./9.*m44*mp2*L4r*RN1p*RS145pe*RSp453e + 
         4./9.*m44*mp2*L4r*RN1p*RS146pe*RSp463e + 4./9.*m44*mp2*L4r*
         RN1p*RS156pe*RSp563e - 4./9.*m44*mp2*L4r*RN1p*RSp561e*RS156pe
          - 4./9.*m44*mp2*L4r*RN1p*RSp461e*RS146pe - 4./9.*m44*mp2*L4r*
         RN1p*RSp451e*RS145pe + 4./9.*m44*me2*L5r*RN1e*RS156pe*RSe563p
          - 4./9.*m44*me2*L5r*RN1e*RSe561p*RS156pe - 4./3.*m44*me2*L4r*
         RN1e*RSe45613p + 4./3.*m44*me2*L4r*RN1e*RSe45611p + 4./9.*m44*
         me2*L4r*RN1e*RS145pe*RSe453p + 4./9.*m44*me2*L4r*RN1e*RS146pe*
         RSe463p + 4./9.*m44*me2*L4r*RN1e*RS156pe*RSe563p - 4./9.*m44*
         me2*L4r*RN1e*RSe561p*RS156pe - 4./9.*m44*me2*L4r*RN1e*RSe461p*
         RS146pe - 4./9.*m44*me2*L4r*RN1e*RSe451p*RS145pe - 8./9.*m44*
         m66*L7r*RN1p*RS145pe*RSp563e - 8./9.*m44*m66*L7r*RN1p*RS156pe*
         RSp453e + 8./9.*m44*m66*L7r*RN1p*RSp561e*RS145pe + 8./9.*m44*
         m66*L7r*RN1p*RSp451e*RS156pe );

      F13p6x23L +=  + Ab(m11,mu2) * (  - 8./9.*m44*m66*L7r*RN1e*RS145pe
         *RSe563p - 8./9.*m44*m66*L7r*RN1e*RS156pe*RSe453p + 8./9.*m44*
         m66*L7r*RN1e*RSe561p*RS145pe + 8./9.*m44*m66*L7r*RN1e*RSe451p*
         RS156pe - 8./9.*m44*m66*L7r*RN13*RS145pe*RS356pe - 8./9.*m44*
         m66*L7r*RN13*RS156pe*RS345pe - 8./9.*m44*m66*L6r*RN1p*RS145pe*
         RSp453e - 8./9.*m44*m66*L6r*RN1p*RS156pe*RSp563e + 8./9.*m44*
         m66*L6r*RN1p*RSp561e*RS156pe + 8./9.*m44*m66*L6r*RN1p*RSp451e*
         RS145pe - 8./9.*m44*m66*L6r*RN1e*RS145pe*RSe453p - 8./9.*m44*
         m66*L6r*RN1e*RS156pe*RSe563p + 8./9.*m44*m66*L6r*RN1e*RSe561p*
         RS156pe + 8./9.*m44*m66*L6r*RN1e*RSe451p*RS145pe - 8./9.*m44*
         m66*L6r*RN13*RS145pe*RS345pe - 8./9.*m44*m66*L6r*RN13*RS156pe*
         RS356pe - 8./9.*m44*m55*L7r*RN1p*RS146pe*RSp563e - 8./9.*m44*
         m55*L7r*RN1p*RS156pe*RSp463e + 8./9.*m44*m55*L7r*RN1p*RSp561e*
         RS146pe + 8./9.*m44*m55*L7r*RN1p*RSp461e*RS156pe - 8./9.*m44*
         m55*L7r*RN1e*RS146pe*RSe563p - 8./9.*m44*m55*L7r*RN1e*RS156pe*
         RSe463p );

      F13p6x23L +=  + Ab(m11,mu2) * ( 8./9.*m44*m55*L7r*RN1e*RSe561p*
         RS146pe + 8./9.*m44*m55*L7r*RN1e*RSe461p*RS156pe - 8./9.*m44*
         m55*L7r*RN13*RS146pe*RS356pe - 8./9.*m44*m55*L7r*RN13*RS156pe*
         RS346pe - 8./9.*m44*m55*L6r*RN1p*RS146pe*RSp463e - 8./9.*m44*
         m55*L6r*RN1p*RS156pe*RSp563e + 8./9.*m44*m55*L6r*RN1p*RSp561e*
         RS156pe + 8./9.*m44*m55*L6r*RN1p*RSp461e*RS146pe - 8./9.*m44*
         m55*L6r*RN1e*RS146pe*RSe463p - 8./9.*m44*m55*L6r*RN1e*RS156pe*
         RSe563p + 8./9.*m44*m55*L6r*RN1e*RSe561p*RS156pe + 8./9.*m44*
         m55*L6r*RN1e*RSe461p*RS146pe - 8./9.*m44*m55*L6r*RN13*RS146pe*
         RS346pe - 8./9.*m44*m55*L6r*RN13*RS156pe*RS356pe - 8./9.*pow(
         m44,2)*L8r*RN1p*RS156pe*RSp563e + 8./9.*pow(m44,2)*L8r*RN1p*
         RSp561e*RS156pe - 8./9.*pow(m44,2)*L8r*RN1e*RS156pe*RSe563p + 
         8./9.*pow(m44,2)*L8r*RN1e*RSe561p*RS156pe - 8./9.*pow(m44,2)*
         L8r*RN13*RS156pe*RS356pe - 8./9.*pow(m44,2)*L7r*RN1p*RS156pe*
         RSp563e + 8./9.*pow(m44,2)*L7r*RN1p*RSp561e*RS156pe - 8./9.*
         pow(m44,2)*L7r*RN1e*RS156pe*RSe563p );

      F13p6x23L +=  + Ab(m11,mu2) * ( 8./9.*pow(m44,2)*L7r*RN1e*RSe561p
         *RS156pe - 8./9.*pow(m44,2)*L7r*RN13*RS156pe*RS356pe - 8./9.*
         pow(m44,2)*L6r*RN1p*RS156pe*RSp563e + 8./9.*pow(m44,2)*L6r*
         RN1p*RSp561e*RS156pe - 8./9.*pow(m44,2)*L6r*RN1e*RS156pe*
         RSe563p + 8./9.*pow(m44,2)*L6r*RN1e*RSe561p*RS156pe - 8./9.*
         pow(m44,2)*L6r*RN13*RS156pe*RS356pe + 1./3.*m33*L5r*RS14563pe
          + 1./6.*m33*L5r*RS1456pee + 1./6.*m33*L5r*RS1456ppe - 1./6.*
         m33*L5r*RS145pe - 1./6.*m33*L5r*RS146pe - 1./6.*m33*L5r*
         RS156pe + 8./3.*m33*m66*L7r*RN13*RS145pe + 8./3.*m33*m66*L6r*
         RN13*RS14563pe + 8./3.*m33*m55*L7r*RN13*RS146pe + 8./3.*m33*
         m55*L6r*RN13*RS14563pe + 8./3.*m33*m44*L7r*RN13*RS156pe + 8./3.
         *m33*m44*L6r*RN13*RS14563pe + 8./3.*pow(m33,2)*L8r*RN13*
         RS14563pe - m11*L5r*RS14563pe - 4./3.*m11*L5r*RSe45613p - 4./3.
         *m11*L5r*RSp45613e - 1./2.*m11*L5r*RS1456pee - 1./2.*m11*L5r*
         RS1456ppe + 4./3.*m11*L5r*RSe45611p + 4./3.*m11*L5r*RSp45611e
          + 1./2.*m11*L5r*RS145pe );

      F13p6x23L +=  + Ab(m11,mu2) * ( 1./2.*m11*L5r*RS146pe + 1./2.*m11
         *L5r*RS156pe + 2./3.*m11*L3r*RS14563pe - 5./3.*m11*L3r*
         RS1456pee - 5./3.*m11*L3r*RS1456ppe + 5./3.*m11*L3r*RS145pe + 
         5./3.*m11*L3r*RS146pe + 5./3.*m11*L3r*RS156pe + 8./3.*m11*L0r*
         RS14563pe - 2./3.*m11*L0r*RS1456pee - 2./3.*m11*L0r*RS1456ppe
          + 2./3.*m11*L0r*RS145pe + 2./3.*m11*L0r*RS146pe + 2./3.*m11*
         L0r*RS156pe - 4./3.*m11*mp2*L5r*RN1p*RSp45613e + 4./3.*m11*mp2
         *L5r*RN1p*RSp45611e - 4./3.*m11*me2*L5r*RN1e*RSe45613p + 4./3.
         *m11*me2*L5r*RN1e*RSe45611p + 8./3.*m11*m66*L7r*RN1p*RSp453e
          - 8./3.*m11*m66*L7r*RN1p*RSp451e + 8./3.*m11*m66*L7r*RN1e*
         RSe453p - 8./3.*m11*m66*L7r*RN1e*RSe451p + 8./3.*m11*m66*L7r*
         RN13*RS345pe + 8./3.*m11*m66*L6r*RN1p*RSp45613e - 8./3.*m11*
         m66*L6r*RN1p*RSp45611e + 8./3.*m11*m66*L6r*RN1e*RSe45613p - 8./
         3.*m11*m66*L6r*RN1e*RSe45611p + 8./3.*m11*m66*L6r*RN13*
         RS34561pe + 4./9.*m11*m66*L5r*RN13*RS145pe*RS345pe - 4./3.*m11
         *m66*L4r*RN13*RS34561pe );

      F13p6x23L +=  + Ab(m11,mu2) * (  - 4./3.*m11*m66*L4r*RN13*
         RS14563pe + 4./9.*m11*m66*L4r*RN13*RS145pe*RS345pe + 4./9.*m11
         *m66*L4r*RN13*RS146pe*RS346pe + 4./9.*m11*m66*L4r*RN13*RS156pe
         *RS356pe + 8./3.*m11*m55*L7r*RN1p*RSp463e - 8./3.*m11*m55*L7r*
         RN1p*RSp461e + 8./3.*m11*m55*L7r*RN1e*RSe463p - 8./3.*m11*m55*
         L7r*RN1e*RSe461p + 8./3.*m11*m55*L7r*RN13*RS346pe + 8./3.*m11*
         m55*L6r*RN1p*RSp45613e - 8./3.*m11*m55*L6r*RN1p*RSp45611e + 8./
         3.*m11*m55*L6r*RN1e*RSe45613p - 8./3.*m11*m55*L6r*RN1e*
         RSe45611p + 8./3.*m11*m55*L6r*RN13*RS34561pe + 4./9.*m11*m55*
         L5r*RN13*RS146pe*RS346pe - 4./3.*m11*m55*L4r*RN13*RS34561pe - 
         4./3.*m11*m55*L4r*RN13*RS14563pe + 4./9.*m11*m55*L4r*RN13*
         RS145pe*RS345pe + 4./9.*m11*m55*L4r*RN13*RS146pe*RS346pe + 4./
         9.*m11*m55*L4r*RN13*RS156pe*RS356pe + 8./3.*m11*m44*L7r*RN1p*
         RSp563e - 8./3.*m11*m44*L7r*RN1p*RSp561e + 8./3.*m11*m44*L7r*
         RN1e*RSe563p - 8./3.*m11*m44*L7r*RN1e*RSe561p + 8./3.*m11*m44*
         L7r*RN13*RS356pe );

      F13p6x23L +=  + Ab(m11,mu2) * ( 8./3.*m11*m44*L6r*RN1p*RSp45613e
          - 8./3.*m11*m44*L6r*RN1p*RSp45611e + 8./3.*m11*m44*L6r*RN1e*
         RSe45613p - 8./3.*m11*m44*L6r*RN1e*RSe45611p + 8./3.*m11*m44*
         L6r*RN13*RS34561pe + 4./9.*m11*m44*L5r*RN13*RS156pe*RS356pe - 
         4./3.*m11*m44*L4r*RN13*RS34561pe - 4./3.*m11*m44*L4r*RN13*
         RS14563pe + 4./9.*m11*m44*L4r*RN13*RS145pe*RS345pe + 4./9.*m11
         *m44*L4r*RN13*RS146pe*RS346pe + 4./9.*m11*m44*L4r*RN13*RS156pe
         *RS356pe - 8*m11*m33*L7r*RN13 - 4./3.*m11*m33*L5r*RN13*
         RS14563pe + 8./3.*pow(m11,2)*L8r*RN1p*RSp45613e - 8./3.*pow(
         m11,2)*L8r*RN1p*RSp45611e + 8./3.*pow(m11,2)*L8r*RN1e*
         RSe45613p - 8./3.*pow(m11,2)*L8r*RN1e*RSe45611p + 8./3.*pow(
         m11,2)*L8r*RN13*RS34561pe - 4./3.*pow(m11,2)*L5r*RN13*
         RS34561pe );

      F13p6x23L +=  + Ab(m13,mu2) * (  - 5*m33*L2r - 2*m33*L1r - 5*m11*
         L2r - 2*m11*L1r );

      F13p6x23L +=  + Ab(m14,mu2) * (  - m66*L4r - m55*L4r - m44*L4r - 
         5./2.*m44*L3r - m44*L0r + 1./2.*m33*L5r + 1./2.*m11*L5r - 5./2.
         *m11*L3r - m11*L0r );

      F13p6x23L +=  + Ab(m15,mu2) * (  - m66*L4r - m55*L4r - 5./2.*m55*
         L3r - m55*L0r - m44*L4r + 1./2.*m33*L5r + 1./2.*m11*L5r - 5./2.
         *m11*L3r - m11*L0r );

      F13p6x23L +=  + Ab(m16,mu2) * (  - m66*L4r - 5./2.*m66*L3r - m66*
         L0r - m55*L4r - m44*L4r + 1./2.*m33*L5r + 1./2.*m11*L5r - 5./2.
         *m11*L3r - m11*L0r );

      F13p6x23L +=  + Ab(m33,mu2) * ( 5./3.*L3r*RD3456ep + 2./3.*L0r*
         RD3456ep - 2./9.*m66*L5r*pow(RS345pe,2) - 4./9.*m66*L5r*
         RSe453p*RS345pe - 4./9.*m66*L5r*RSp453e*RS345pe + 4./9.*m66*
         L5r*RS145pe*RS345pe + 4./9.*m66*L5r*RSe451p*RS345pe + 4./9.*
         m66*L5r*RSp451e*RS345pe - 2*m66*L4r*RS34561pe - 4./3.*m66*L4r*
         RS14563pe - 4./3.*m66*L4r*RSe45613p - 4./3.*m66*L4r*RSp45613e
          - m66*L4r*RS3456pee - m66*L4r*RS3456ppe + 4./3.*m66*L4r*
         RSe45633p + 4./3.*m66*L4r*RSp45633e + m66*L4r*RS345pe - 2./9.*
         m66*L4r*pow(RS345pe,2) + m66*L4r*RS346pe - 2./9.*m66*L4r*pow(
         RS346pe,2) + m66*L4r*RS356pe - 2./9.*m66*L4r*pow(RS356pe,2) - 
         4./9.*m66*L4r*RSe563p*RS356pe - 4./9.*m66*L4r*RSe463p*RS346pe
          - 4./9.*m66*L4r*RSe453p*RS345pe - 4./9.*m66*L4r*RSp563e*
         RS356pe - 4./9.*m66*L4r*RSp463e*RS346pe - 4./9.*m66*L4r*
         RSp453e*RS345pe + 4./9.*m66*L4r*RS145pe*RS345pe + 4./9.*m66*
         L4r*RS146pe*RS346pe + 4./9.*m66*L4r*RS156pe*RS356pe + 4./9.*
         m66*L4r*RSe561p*RS356pe );

      F13p6x23L +=  + Ab(m33,mu2) * ( 4./9.*m66*L4r*RSe461p*RS346pe + 4.
         /9.*m66*L4r*RSe451p*RS345pe + 4./9.*m66*L4r*RSp561e*RS356pe + 
         4./9.*m66*L4r*RSp461e*RS346pe + 4./9.*m66*L4r*RSp451e*RS345pe
          - 4./9.*m66*mp2*L5r*RN3p*RSp453e*RS345pe + 4./9.*m66*mp2*L5r*
         RN3p*RSp451e*RS345pe - 4./3.*m66*mp2*L4r*RN3p*RSp45613e + 4./3.
         *m66*mp2*L4r*RN3p*RSp45633e - 4./9.*m66*mp2*L4r*RN3p*RSp563e*
         RS356pe - 4./9.*m66*mp2*L4r*RN3p*RSp463e*RS346pe - 4./9.*m66*
         mp2*L4r*RN3p*RSp453e*RS345pe + 4./9.*m66*mp2*L4r*RN3p*RSp561e*
         RS356pe + 4./9.*m66*mp2*L4r*RN3p*RSp461e*RS346pe + 4./9.*m66*
         mp2*L4r*RN3p*RSp451e*RS345pe - 4./9.*m66*me2*L5r*RN3e*RSe453p*
         RS345pe + 4./9.*m66*me2*L5r*RN3e*RSe451p*RS345pe - 4./3.*m66*
         me2*L4r*RN3e*RSe45613p + 4./3.*m66*me2*L4r*RN3e*RSe45633p - 4./
         9.*m66*me2*L4r*RN3e*RSe563p*RS356pe - 4./9.*m66*me2*L4r*RN3e*
         RSe463p*RS346pe - 4./9.*m66*me2*L4r*RN3e*RSe453p*RS345pe + 4./
         9.*m66*me2*L4r*RN3e*RSe561p*RS356pe + 4./9.*m66*me2*L4r*RN3e*
         RSe461p*RS346pe );

      F13p6x23L +=  + Ab(m33,mu2) * ( 4./9.*m66*me2*L4r*RN3e*RSe451p*
         RS345pe + 8./9.*pow(m66,2)*L8r*RN3p*RSp453e*RS345pe - 8./9.*
         pow(m66,2)*L8r*RN3p*RSp451e*RS345pe + 8./9.*pow(m66,2)*L8r*
         RN3e*RSe453p*RS345pe - 8./9.*pow(m66,2)*L8r*RN3e*RSe451p*
         RS345pe + 8./9.*pow(m66,2)*L8r*RN13*RS145pe*RS345pe + 8./9.*
         pow(m66,2)*L7r*RN3p*RSp453e*RS345pe - 8./9.*pow(m66,2)*L7r*
         RN3p*RSp451e*RS345pe + 8./9.*pow(m66,2)*L7r*RN3e*RSe453p*
         RS345pe - 8./9.*pow(m66,2)*L7r*RN3e*RSe451p*RS345pe + 8./9.*
         pow(m66,2)*L7r*RN13*RS145pe*RS345pe + 8./9.*pow(m66,2)*L6r*
         RN3p*RSp453e*RS345pe - 8./9.*pow(m66,2)*L6r*RN3p*RSp451e*
         RS345pe + 8./9.*pow(m66,2)*L6r*RN3e*RSe453p*RS345pe - 8./9.*
         pow(m66,2)*L6r*RN3e*RSe451p*RS345pe + 8./9.*pow(m66,2)*L6r*
         RN13*RS145pe*RS345pe - 2./9.*m55*L5r*pow(RS346pe,2) - 4./9.*
         m55*L5r*RSe463p*RS346pe - 4./9.*m55*L5r*RSp463e*RS346pe + 4./9.
         *m55*L5r*RS146pe*RS346pe + 4./9.*m55*L5r*RSe461p*RS346pe + 4./
         9.*m55*L5r*RSp461e*RS346pe );

      F13p6x23L +=  + Ab(m33,mu2) * (  - 2*m55*L4r*RS34561pe - 4./3.*
         m55*L4r*RS14563pe - 4./3.*m55*L4r*RSe45613p - 4./3.*m55*L4r*
         RSp45613e - m55*L4r*RS3456pee - m55*L4r*RS3456ppe + 4./3.*m55*
         L4r*RSe45633p + 4./3.*m55*L4r*RSp45633e + m55*L4r*RS345pe - 2./
         9.*m55*L4r*pow(RS345pe,2) + m55*L4r*RS346pe - 2./9.*m55*L4r*
         pow(RS346pe,2) + m55*L4r*RS356pe - 2./9.*m55*L4r*pow(
         RS356pe,2) - 4./9.*m55*L4r*RSe563p*RS356pe - 4./9.*m55*L4r*
         RSe463p*RS346pe - 4./9.*m55*L4r*RSe453p*RS345pe - 4./9.*m55*
         L4r*RSp563e*RS356pe - 4./9.*m55*L4r*RSp463e*RS346pe - 4./9.*
         m55*L4r*RSp453e*RS345pe + 4./9.*m55*L4r*RS145pe*RS345pe + 4./9.
         *m55*L4r*RS146pe*RS346pe + 4./9.*m55*L4r*RS156pe*RS356pe + 4./
         9.*m55*L4r*RSe561p*RS356pe + 4./9.*m55*L4r*RSe461p*RS346pe + 4.
         /9.*m55*L4r*RSe451p*RS345pe + 4./9.*m55*L4r*RSp561e*RS356pe + 
         4./9.*m55*L4r*RSp461e*RS346pe + 4./9.*m55*L4r*RSp451e*RS345pe
          - 4./9.*m55*mp2*L5r*RN3p*RSp463e*RS346pe + 4./9.*m55*mp2*L5r*
         RN3p*RSp461e*RS346pe );

      F13p6x23L +=  + Ab(m33,mu2) * (  - 4./3.*m55*mp2*L4r*RN3p*
         RSp45613e + 4./3.*m55*mp2*L4r*RN3p*RSp45633e - 4./9.*m55*mp2*
         L4r*RN3p*RSp563e*RS356pe - 4./9.*m55*mp2*L4r*RN3p*RSp463e*
         RS346pe - 4./9.*m55*mp2*L4r*RN3p*RSp453e*RS345pe + 4./9.*m55*
         mp2*L4r*RN3p*RSp561e*RS356pe + 4./9.*m55*mp2*L4r*RN3p*RSp461e*
         RS346pe + 4./9.*m55*mp2*L4r*RN3p*RSp451e*RS345pe - 4./9.*m55*
         me2*L5r*RN3e*RSe463p*RS346pe + 4./9.*m55*me2*L5r*RN3e*RSe461p*
         RS346pe - 4./3.*m55*me2*L4r*RN3e*RSe45613p + 4./3.*m55*me2*L4r
         *RN3e*RSe45633p - 4./9.*m55*me2*L4r*RN3e*RSe563p*RS356pe - 4./
         9.*m55*me2*L4r*RN3e*RSe463p*RS346pe - 4./9.*m55*me2*L4r*RN3e*
         RSe453p*RS345pe + 4./9.*m55*me2*L4r*RN3e*RSe561p*RS356pe + 4./
         9.*m55*me2*L4r*RN3e*RSe461p*RS346pe + 4./9.*m55*me2*L4r*RN3e*
         RSe451p*RS345pe + 8./9.*m55*m66*L7r*RN3p*RSp463e*RS345pe + 8./
         9.*m55*m66*L7r*RN3p*RSp453e*RS346pe - 8./9.*m55*m66*L7r*RN3p*
         RSp461e*RS345pe - 8./9.*m55*m66*L7r*RN3p*RSp451e*RS346pe + 8./
         9.*m55*m66*L7r*RN3e*RSe463p*RS345pe );

      F13p6x23L +=  + Ab(m33,mu2) * ( 8./9.*m55*m66*L7r*RN3e*RSe453p*
         RS346pe - 8./9.*m55*m66*L7r*RN3e*RSe461p*RS345pe - 8./9.*m55*
         m66*L7r*RN3e*RSe451p*RS346pe + 8./9.*m55*m66*L7r*RN13*RS145pe*
         RS346pe + 8./9.*m55*m66*L7r*RN13*RS146pe*RS345pe + 8./9.*m55*
         m66*L6r*RN3p*RSp463e*RS346pe + 8./9.*m55*m66*L6r*RN3p*RSp453e*
         RS345pe - 8./9.*m55*m66*L6r*RN3p*RSp461e*RS346pe - 8./9.*m55*
         m66*L6r*RN3p*RSp451e*RS345pe + 8./9.*m55*m66*L6r*RN3e*RSe463p*
         RS346pe + 8./9.*m55*m66*L6r*RN3e*RSe453p*RS345pe - 8./9.*m55*
         m66*L6r*RN3e*RSe461p*RS346pe - 8./9.*m55*m66*L6r*RN3e*RSe451p*
         RS345pe + 8./9.*m55*m66*L6r*RN13*RS145pe*RS345pe + 8./9.*m55*
         m66*L6r*RN13*RS146pe*RS346pe + 8./9.*pow(m55,2)*L8r*RN3p*
         RSp463e*RS346pe - 8./9.*pow(m55,2)*L8r*RN3p*RSp461e*RS346pe + 
         8./9.*pow(m55,2)*L8r*RN3e*RSe463p*RS346pe - 8./9.*pow(m55,2)*
         L8r*RN3e*RSe461p*RS346pe + 8./9.*pow(m55,2)*L8r*RN13*RS146pe*
         RS346pe + 8./9.*pow(m55,2)*L7r*RN3p*RSp463e*RS346pe - 8./9.*
         pow(m55,2)*L7r*RN3p*RSp461e*RS346pe );

      F13p6x23L +=  + Ab(m33,mu2) * ( 8./9.*pow(m55,2)*L7r*RN3e*RSe463p
         *RS346pe - 8./9.*pow(m55,2)*L7r*RN3e*RSe461p*RS346pe + 8./9.*
         pow(m55,2)*L7r*RN13*RS146pe*RS346pe + 8./9.*pow(m55,2)*L6r*
         RN3p*RSp463e*RS346pe - 8./9.*pow(m55,2)*L6r*RN3p*RSp461e*
         RS346pe + 8./9.*pow(m55,2)*L6r*RN3e*RSe463p*RS346pe - 8./9.*
         pow(m55,2)*L6r*RN3e*RSe461p*RS346pe + 8./9.*pow(m55,2)*L6r*
         RN13*RS146pe*RS346pe - 2./9.*m44*L5r*pow(RS356pe,2) - 4./9.*
         m44*L5r*RSe563p*RS356pe - 4./9.*m44*L5r*RSp563e*RS356pe + 4./9.
         *m44*L5r*RS156pe*RS356pe + 4./9.*m44*L5r*RSe561p*RS356pe + 4./
         9.*m44*L5r*RSp561e*RS356pe - 2*m44*L4r*RS34561pe - 4./3.*m44*
         L4r*RS14563pe - 4./3.*m44*L4r*RSe45613p - 4./3.*m44*L4r*
         RSp45613e - m44*L4r*RS3456pee - m44*L4r*RS3456ppe + 4./3.*m44*
         L4r*RSe45633p + 4./3.*m44*L4r*RSp45633e + m44*L4r*RS345pe - 2./
         9.*m44*L4r*pow(RS345pe,2) + m44*L4r*RS346pe - 2./9.*m44*L4r*
         pow(RS346pe,2) + m44*L4r*RS356pe - 2./9.*m44*L4r*pow(
         RS356pe,2) );

      F13p6x23L +=  + Ab(m33,mu2) * (  - 4./9.*m44*L4r*RSe563p*RS356pe
          - 4./9.*m44*L4r*RSe463p*RS346pe - 4./9.*m44*L4r*RSe453p*
         RS345pe - 4./9.*m44*L4r*RSp563e*RS356pe - 4./9.*m44*L4r*
         RSp463e*RS346pe - 4./9.*m44*L4r*RSp453e*RS345pe + 4./9.*m44*
         L4r*RS145pe*RS345pe + 4./9.*m44*L4r*RS146pe*RS346pe + 4./9.*
         m44*L4r*RS156pe*RS356pe + 4./9.*m44*L4r*RSe561p*RS356pe + 4./9.
         *m44*L4r*RSe461p*RS346pe + 4./9.*m44*L4r*RSe451p*RS345pe + 4./
         9.*m44*L4r*RSp561e*RS356pe + 4./9.*m44*L4r*RSp461e*RS346pe + 4.
         /9.*m44*L4r*RSp451e*RS345pe - 4./9.*m44*mp2*L5r*RN3p*RSp563e*
         RS356pe + 4./9.*m44*mp2*L5r*RN3p*RSp561e*RS356pe - 4./3.*m44*
         mp2*L4r*RN3p*RSp45613e + 4./3.*m44*mp2*L4r*RN3p*RSp45633e - 4./
         9.*m44*mp2*L4r*RN3p*RSp563e*RS356pe - 4./9.*m44*mp2*L4r*RN3p*
         RSp463e*RS346pe - 4./9.*m44*mp2*L4r*RN3p*RSp453e*RS345pe + 4./
         9.*m44*mp2*L4r*RN3p*RSp561e*RS356pe + 4./9.*m44*mp2*L4r*RN3p*
         RSp461e*RS346pe + 4./9.*m44*mp2*L4r*RN3p*RSp451e*RS345pe - 4./
         9.*m44*me2*L5r*RN3e*RSe563p*RS356pe );

      F13p6x23L +=  + Ab(m33,mu2) * ( 4./9.*m44*me2*L5r*RN3e*RSe561p*
         RS356pe - 4./3.*m44*me2*L4r*RN3e*RSe45613p + 4./3.*m44*me2*L4r
         *RN3e*RSe45633p - 4./9.*m44*me2*L4r*RN3e*RSe563p*RS356pe - 4./
         9.*m44*me2*L4r*RN3e*RSe463p*RS346pe - 4./9.*m44*me2*L4r*RN3e*
         RSe453p*RS345pe + 4./9.*m44*me2*L4r*RN3e*RSe561p*RS356pe + 4./
         9.*m44*me2*L4r*RN3e*RSe461p*RS346pe + 4./9.*m44*me2*L4r*RN3e*
         RSe451p*RS345pe + 8./9.*m44*m66*L7r*RN3p*RSp563e*RS345pe + 8./
         9.*m44*m66*L7r*RN3p*RSp453e*RS356pe - 8./9.*m44*m66*L7r*RN3p*
         RSp561e*RS345pe - 8./9.*m44*m66*L7r*RN3p*RSp451e*RS356pe + 8./
         9.*m44*m66*L7r*RN3e*RSe563p*RS345pe + 8./9.*m44*m66*L7r*RN3e*
         RSe453p*RS356pe - 8./9.*m44*m66*L7r*RN3e*RSe561p*RS345pe - 8./
         9.*m44*m66*L7r*RN3e*RSe451p*RS356pe + 8./9.*m44*m66*L7r*RN13*
         RS145pe*RS356pe + 8./9.*m44*m66*L7r*RN13*RS156pe*RS345pe + 8./
         9.*m44*m66*L6r*RN3p*RSp563e*RS356pe + 8./9.*m44*m66*L6r*RN3p*
         RSp453e*RS345pe - 8./9.*m44*m66*L6r*RN3p*RSp561e*RS356pe - 8./
         9.*m44*m66*L6r*RN3p*RSp451e*RS345pe );

      F13p6x23L +=  + Ab(m33,mu2) * ( 8./9.*m44*m66*L6r*RN3e*RSe563p*
         RS356pe + 8./9.*m44*m66*L6r*RN3e*RSe453p*RS345pe - 8./9.*m44*
         m66*L6r*RN3e*RSe561p*RS356pe - 8./9.*m44*m66*L6r*RN3e*RSe451p*
         RS345pe + 8./9.*m44*m66*L6r*RN13*RS145pe*RS345pe + 8./9.*m44*
         m66*L6r*RN13*RS156pe*RS356pe + 8./9.*m44*m55*L7r*RN3p*RSp563e*
         RS346pe + 8./9.*m44*m55*L7r*RN3p*RSp463e*RS356pe - 8./9.*m44*
         m55*L7r*RN3p*RSp561e*RS346pe - 8./9.*m44*m55*L7r*RN3p*RSp461e*
         RS356pe + 8./9.*m44*m55*L7r*RN3e*RSe563p*RS346pe + 8./9.*m44*
         m55*L7r*RN3e*RSe463p*RS356pe - 8./9.*m44*m55*L7r*RN3e*RSe561p*
         RS346pe - 8./9.*m44*m55*L7r*RN3e*RSe461p*RS356pe + 8./9.*m44*
         m55*L7r*RN13*RS146pe*RS356pe + 8./9.*m44*m55*L7r*RN13*RS156pe*
         RS346pe + 8./9.*m44*m55*L6r*RN3p*RSp563e*RS356pe + 8./9.*m44*
         m55*L6r*RN3p*RSp463e*RS346pe - 8./9.*m44*m55*L6r*RN3p*RSp561e*
         RS356pe - 8./9.*m44*m55*L6r*RN3p*RSp461e*RS346pe + 8./9.*m44*
         m55*L6r*RN3e*RSe563p*RS356pe + 8./9.*m44*m55*L6r*RN3e*RSe463p*
         RS346pe );

      F13p6x23L +=  + Ab(m33,mu2) * (  - 8./9.*m44*m55*L6r*RN3e*RSe561p
         *RS356pe - 8./9.*m44*m55*L6r*RN3e*RSe461p*RS346pe + 8./9.*m44*
         m55*L6r*RN13*RS146pe*RS346pe + 8./9.*m44*m55*L6r*RN13*RS156pe*
         RS356pe + 8./9.*pow(m44,2)*L8r*RN3p*RSp563e*RS356pe - 8./9.*
         pow(m44,2)*L8r*RN3p*RSp561e*RS356pe + 8./9.*pow(m44,2)*L8r*
         RN3e*RSe563p*RS356pe - 8./9.*pow(m44,2)*L8r*RN3e*RSe561p*
         RS356pe + 8./9.*pow(m44,2)*L8r*RN13*RS156pe*RS356pe + 8./9.*
         pow(m44,2)*L7r*RN3p*RSp563e*RS356pe - 8./9.*pow(m44,2)*L7r*
         RN3p*RSp561e*RS356pe + 8./9.*pow(m44,2)*L7r*RN3e*RSe563p*
         RS356pe - 8./9.*pow(m44,2)*L7r*RN3e*RSe561p*RS356pe + 8./9.*
         pow(m44,2)*L7r*RN13*RS156pe*RS356pe + 8./9.*pow(m44,2)*L6r*
         RN3p*RSp563e*RS356pe - 8./9.*pow(m44,2)*L6r*RN3p*RSp561e*
         RS356pe + 8./9.*pow(m44,2)*L6r*RN3e*RSe563p*RS356pe - 8./9.*
         pow(m44,2)*L6r*RN3e*RSe561p*RS356pe + 8./9.*pow(m44,2)*L6r*
         RN13*RS156pe*RS356pe - m33*L5r*RS34561pe - 4./3.*m33*L5r*
         RS14563pe );

      F13p6x23L +=  + Ab(m33,mu2) * (  - 4./3.*m33*L5r*RSe45613p - 4./3.
         *m33*L5r*RSp45613e - 1./2.*m33*L5r*RS3456pee - 1./2.*m33*L5r*
         RS3456ppe + 4./3.*m33*L5r*RSe45633p + 4./3.*m33*L5r*RSp45633e
          + 1./2.*m33*L5r*RS345pe + 1./2.*m33*L5r*RS346pe + 1./2.*m33*
         L5r*RS356pe + 2./3.*m33*L3r*RS34561pe - 5./3.*m33*L3r*
         RS3456pee - 5./3.*m33*L3r*RS3456ppe + 5./3.*m33*L3r*RS345pe + 
         5./3.*m33*L3r*RS346pe + 5./3.*m33*L3r*RS356pe + 8./3.*m33*L0r*
         RS34561pe - 2./3.*m33*L0r*RS3456pee - 2./3.*m33*L0r*RS3456ppe
          + 2./3.*m33*L0r*RS345pe + 2./3.*m33*L0r*RS346pe + 2./3.*m33*
         L0r*RS356pe - 4./3.*m33*mp2*L5r*RN3p*RSp45613e + 4./3.*m33*mp2
         *L5r*RN3p*RSp45633e - 4./3.*m33*me2*L5r*RN3e*RSe45613p + 4./3.
         *m33*me2*L5r*RN3e*RSe45633p - 8./3.*m33*m66*L7r*RN3p*RSp453e
          + 8./3.*m33*m66*L7r*RN3p*RSp451e - 8./3.*m33*m66*L7r*RN3e*
         RSe453p + 8./3.*m33*m66*L7r*RN3e*RSe451p - 8./3.*m33*m66*L7r*
         RN13*RS145pe + 8./3.*m33*m66*L6r*RN3p*RSp45613e - 8./3.*m33*
         m66*L6r*RN3p*RSp45633e );

      F13p6x23L +=  + Ab(m33,mu2) * ( 8./3.*m33*m66*L6r*RN3e*RSe45613p
          - 8./3.*m33*m66*L6r*RN3e*RSe45633p - 8./3.*m33*m66*L6r*RN13*
         RS14563pe - 8./3.*m33*m55*L7r*RN3p*RSp463e + 8./3.*m33*m55*L7r
         *RN3p*RSp461e - 8./3.*m33*m55*L7r*RN3e*RSe463p + 8./3.*m33*m55
         *L7r*RN3e*RSe461p - 8./3.*m33*m55*L7r*RN13*RS146pe + 8./3.*m33
         *m55*L6r*RN3p*RSp45613e - 8./3.*m33*m55*L6r*RN3p*RSp45633e + 8.
         /3.*m33*m55*L6r*RN3e*RSe45613p - 8./3.*m33*m55*L6r*RN3e*
         RSe45633p - 8./3.*m33*m55*L6r*RN13*RS14563pe - 8./3.*m33*m44*
         L7r*RN3p*RSp563e + 8./3.*m33*m44*L7r*RN3p*RSp561e - 8./3.*m33*
         m44*L7r*RN3e*RSe563p + 8./3.*m33*m44*L7r*RN3e*RSe561p - 8./3.*
         m33*m44*L7r*RN13*RS156pe + 8./3.*m33*m44*L6r*RN3p*RSp45613e - 
         8./3.*m33*m44*L6r*RN3p*RSp45633e + 8./3.*m33*m44*L6r*RN3e*
         RSe45613p - 8./3.*m33*m44*L6r*RN3e*RSe45633p - 8./3.*m33*m44*
         L6r*RN13*RS14563pe + 8./3.*pow(m33,2)*L8r*RN3p*RSp45613e - 8./
         3.*pow(m33,2)*L8r*RN3p*RSp45633e + 8./3.*pow(m33,2)*L8r*RN3e*
         RSe45613p );

      F13p6x23L +=  + Ab(m33,mu2) * (  - 8./3.*pow(m33,2)*L8r*RN3e*
         RSe45633p - 8./3.*pow(m33,2)*L8r*RN13*RS14563pe - m11*L5r*
         RS34561pe + 1./6.*m11*L5r*RS3456pee + 1./6.*m11*L5r*RS3456ppe
          - 1./6.*m11*L5r*RS345pe - 1./6.*m11*L5r*RS346pe - 1./6.*m11*
         L5r*RS356pe - 8./3.*m11*m66*L7r*RN13*RS345pe - 8./3.*m11*m66*
         L6r*RN13*RS34561pe - 4./9.*m11*m66*L5r*RN13*RS145pe*RS345pe + 
         4./3.*m11*m66*L4r*RN13*RS34561pe + 4./3.*m11*m66*L4r*RN13*
         RS14563pe - 4./9.*m11*m66*L4r*RN13*RS145pe*RS345pe - 4./9.*m11
         *m66*L4r*RN13*RS146pe*RS346pe - 4./9.*m11*m66*L4r*RN13*RS156pe
         *RS356pe - 8./3.*m11*m55*L7r*RN13*RS346pe - 8./3.*m11*m55*L6r*
         RN13*RS34561pe - 4./9.*m11*m55*L5r*RN13*RS146pe*RS346pe + 4./3.
         *m11*m55*L4r*RN13*RS34561pe + 4./3.*m11*m55*L4r*RN13*RS14563pe
          - 4./9.*m11*m55*L4r*RN13*RS145pe*RS345pe - 4./9.*m11*m55*L4r*
         RN13*RS146pe*RS346pe - 4./9.*m11*m55*L4r*RN13*RS156pe*RS356pe
          - 8./3.*m11*m44*L7r*RN13*RS356pe - 8./3.*m11*m44*L6r*RN13*
         RS34561pe );

      F13p6x23L +=  + Ab(m33,mu2) * (  - 4./9.*m11*m44*L5r*RN13*RS156pe
         *RS356pe + 4./3.*m11*m44*L4r*RN13*RS34561pe + 4./3.*m11*m44*
         L4r*RN13*RS14563pe - 4./9.*m11*m44*L4r*RN13*RS145pe*RS345pe - 
         4./9.*m11*m44*L4r*RN13*RS146pe*RS346pe - 4./9.*m11*m44*L4r*
         RN13*RS156pe*RS356pe + 8*m11*m33*L7r*RN13 + 4./3.*m11*m33*L5r*
         RN13*RS14563pe - 8./3.*pow(m11,2)*L8r*RN13*RS34561pe + 4./3.*
         pow(m11,2)*L5r*RN13*RS34561pe );

      F13p6x23L +=  + Ab(m34,mu2) * (  - m66*L4r - m55*L4r - m44*L4r - 
         5./2.*m44*L3r - m44*L0r + 1./2.*m33*L5r - 5./2.*m33*L3r - m33*
         L0r + 1./2.*m11*L5r );

      F13p6x23L +=  + Ab(m35,mu2) * (  - m66*L4r - m55*L4r - 5./2.*m55*
         L3r - m55*L0r - m44*L4r + 1./2.*m33*L5r - 5./2.*m33*L3r - m33*
         L0r + 1./2.*m11*L5r );

      F13p6x23L +=  + Ab(m36,mu2) * (  - m66*L4r - 5./2.*m66*L3r - m66*
         L0r - m55*L4r - m44*L4r + 1./2.*m33*L5r - 5./2.*m33*L3r - m33*
         L0r + 1./2.*m11*L5r );

      F13p6x23L +=  + Ab(m44,mu2) * ( 4*m44*L4r - 4./3.*m44*L4r*RS456pe
          - 2*m44*L2r + 2./3.*m44*L2r*RS456pe - 8*m44*L1r + 8./3.*m44*
         L1r*RS456pe );

      F13p6x23L +=  + Ab(m45,mu2) * ( 4*m55*L4r - 2*m55*L2r - 8*m55*L1r
          + 4*m44*L4r - 2*m44*L2r - 8*m44*L1r );

      F13p6x23L +=  + Ab(m46,mu2) * ( 4*m66*L4r - 2*m66*L2r - 8*m66*L1r
          + 4*m44*L4r - 2*m44*L2r - 8*m44*L1r );

      F13p6x23L +=  + Ab(m55,mu2) * ( 4*m55*L4r - 4./3.*m55*L4r*RS546pe
          - 2*m55*L2r + 2./3.*m55*L2r*RS546pe - 8*m55*L1r + 8./3.*m55*
         L1r*RS546pe );

      F13p6x23L +=  + Ab(m56,mu2) * ( 4*m66*L4r - 2*m66*L2r - 8*m66*L1r
          + 4*m55*L4r - 2*m55*L2r - 8*m55*L1r );

      F13p6x23L +=  + Ab(m66,mu2) * ( 4*m66*L4r - 4./3.*m66*L4r*RS645pe
          - 2*m66*L2r + 2./3.*m66*L2r*RS645pe - 8*m66*L1r + 8./3.*m66*
         L1r*RS645pe );

      F13p6x23L +=  + Ab(me2,mu2) * ( 2./3.*me2*L3r*RSe45613p + 5./3.*
         me2*L3r*RSe45633p + 5./3.*me2*L3r*RSe45611p + 2./3.*me2*L2r*
         RSe564p + 2./3.*me2*L2r*RSe465p + 2./3.*me2*L2r*RSe456p + 8./3.
         *me2*L1r*RSe564p + 8./3.*me2*L1r*RSe465p + 8./3.*me2*L1r*
         RSe456p + 8./3.*me2*L0r*RSe45613p + 2./3.*me2*L0r*RSe45633p + 
         2./3.*me2*L0r*RSe45611p - 2./9.*m66*L5r*pow(RSe453p,2) - 4./9.
         *m66*L5r*RSp453e*RSe453p + 4./9.*m66*L5r*RSe451p*RSe453p + 4./
         9.*m66*L5r*RSe451p*RSp453e - 2./9.*m66*L5r*pow(RSe451p,2) + 4./
         9.*m66*L5r*RSp451e*RSe453p - 4./9.*m66*L5r*RSp451e*RSe451p + 2.
         /3.*m66*L4r*RSe45613p - 1./3.*m66*L4r*RSe45633p - 2./9.*m66*
         L4r*pow(RSe563p,2) - 2./9.*m66*L4r*pow(RSe463p,2) - 2./9.*m66*
         L4r*pow(RSe453p,2) - 4./9.*m66*L4r*RSp563e*RSe563p - 4./9.*m66
         *L4r*RSp463e*RSe463p - 4./9.*m66*L4r*RSp453e*RSe453p - 1./3.*
         m66*L4r*RSe45611p + 4./9.*m66*L4r*RSe561p*RSe563p + 4./9.*m66*
         L4r*RSe561p*RSp563e - 2./9.*m66*L4r*pow(RSe561p,2) + 4./9.*m66
         *L4r*RSe461p*RSe463p );

      F13p6x23L +=  + Ab(me2,mu2) * ( 4./9.*m66*L4r*RSe461p*RSp463e - 2.
         /9.*m66*L4r*pow(RSe461p,2) + 4./9.*m66*L4r*RSe451p*RSe453p + 4.
         /9.*m66*L4r*RSe451p*RSp453e - 2./9.*m66*L4r*pow(RSe451p,2) + 4.
         /9.*m66*L4r*RSp561e*RSe563p - 4./9.*m66*L4r*RSp561e*RSe561p + 
         4./9.*m66*L4r*RSp461e*RSe463p - 4./9.*m66*L4r*RSp461e*RSe461p
          + 4./9.*m66*L4r*RSp451e*RSe453p - 4./9.*m66*L4r*RSp451e*
         RSe451p - 4./3.*m66*L4r*RSe456p - 4./9.*m66*mp2*L5r*RNep*
         RSp453e*RSe453p + 4./9.*m66*mp2*L5r*RNep*RSe451p*RSp453e + 4./
         9.*m66*mp2*L5r*RNep*RSp451e*RSe453p - 4./9.*m66*mp2*L5r*RNep*
         RSp451e*RSe451p - 4./9.*m66*mp2*L4r*RNep*RSp563e*RSe563p - 4./
         9.*m66*mp2*L4r*RNep*RSp463e*RSe463p - 4./9.*m66*mp2*L4r*RNep*
         RSp453e*RSe453p + 4./9.*m66*mp2*L4r*RNep*RSe561p*RSp563e + 4./
         9.*m66*mp2*L4r*RNep*RSe461p*RSp463e + 4./9.*m66*mp2*L4r*RNep*
         RSe451p*RSp453e + 4./9.*m66*mp2*L4r*RNep*RSp561e*RSe563p - 4./
         9.*m66*mp2*L4r*RNep*RSp561e*RSe561p + 4./9.*m66*mp2*L4r*RNep*
         RSp461e*RSe463p );

      F13p6x23L +=  + Ab(me2,mu2) * (  - 4./9.*m66*mp2*L4r*RNep*RSp461e
         *RSe461p + 4./9.*m66*mp2*L4r*RNep*RSp451e*RSe453p - 4./9.*m66*
         mp2*L4r*RNep*RSp451e*RSe451p + 4./9.*m66*me2*L5r*RN3e*RSe453p*
         RS345pe - 4./9.*m66*me2*L5r*RN3e*RSe451p*RS345pe - 4./9.*m66*
         me2*L5r*RN1e*RS145pe*RSe453p + 4./9.*m66*me2*L5r*RN1e*RSe451p*
         RS145pe + 4./3.*m66*me2*L4r*RN3e*RSe45613p - 4./3.*m66*me2*L4r
         *RN3e*RSe45633p + 4./9.*m66*me2*L4r*RN3e*RSe563p*RS356pe + 4./
         9.*m66*me2*L4r*RN3e*RSe463p*RS346pe + 4./9.*m66*me2*L4r*RN3e*
         RSe453p*RS345pe - 4./9.*m66*me2*L4r*RN3e*RSe561p*RS356pe - 4./
         9.*m66*me2*L4r*RN3e*RSe461p*RS346pe - 4./9.*m66*me2*L4r*RN3e*
         RSe451p*RS345pe + 4./3.*m66*me2*L4r*RN1e*RSe45613p - 4./3.*m66
         *me2*L4r*RN1e*RSe45611p - 4./9.*m66*me2*L4r*RN1e*RS145pe*
         RSe453p - 4./9.*m66*me2*L4r*RN1e*RS146pe*RSe463p - 4./9.*m66*
         me2*L4r*RN1e*RS156pe*RSe563p + 4./9.*m66*me2*L4r*RN1e*RSe561p*
         RS156pe + 4./9.*m66*me2*L4r*RN1e*RSe461p*RS146pe + 4./9.*m66*
         me2*L4r*RN1e*RSe451p*RS145pe );

      F13p6x23L +=  + Ab(me2,mu2) * ( 8./9.*pow(m66,2)*L8r*RNep*RSp453e
         *RSe453p - 8./9.*pow(m66,2)*L8r*RNep*RSe451p*RSp453e - 8./9.*
         pow(m66,2)*L8r*RNep*RSp451e*RSe453p + 8./9.*pow(m66,2)*L8r*
         RNep*RSp451e*RSe451p - 8./9.*pow(m66,2)*L8r*RN3e*RSe453p*
         RS345pe + 8./9.*pow(m66,2)*L8r*RN3e*RSe451p*RS345pe + 8./9.*
         pow(m66,2)*L8r*RN1e*RS145pe*RSe453p - 8./9.*pow(m66,2)*L8r*
         RN1e*RSe451p*RS145pe + 8./9.*pow(m66,2)*L7r*RNep*RSp453e*
         RSe453p - 8./9.*pow(m66,2)*L7r*RNep*RSe451p*RSp453e - 8./9.*
         pow(m66,2)*L7r*RNep*RSp451e*RSe453p + 8./9.*pow(m66,2)*L7r*
         RNep*RSp451e*RSe451p - 8./9.*pow(m66,2)*L7r*RN3e*RSe453p*
         RS345pe + 8./9.*pow(m66,2)*L7r*RN3e*RSe451p*RS345pe + 8./9.*
         pow(m66,2)*L7r*RN1e*RS145pe*RSe453p - 8./9.*pow(m66,2)*L7r*
         RN1e*RSe451p*RS145pe + 8./9.*pow(m66,2)*L6r*RNep*RSp453e*
         RSe453p - 8./9.*pow(m66,2)*L6r*RNep*RSe451p*RSp453e - 8./9.*
         pow(m66,2)*L6r*RNep*RSp451e*RSe453p + 8./9.*pow(m66,2)*L6r*
         RNep*RSp451e*RSe451p );

      F13p6x23L +=  + Ab(me2,mu2) * (  - 8./9.*pow(m66,2)*L6r*RN3e*
         RSe453p*RS345pe + 8./9.*pow(m66,2)*L6r*RN3e*RSe451p*RS345pe + 
         8./9.*pow(m66,2)*L6r*RN1e*RS145pe*RSe453p - 8./9.*pow(m66,2)*
         L6r*RN1e*RSe451p*RS145pe - 2./9.*m55*L5r*pow(RSe463p,2) - 4./9.
         *m55*L5r*RSp463e*RSe463p + 4./9.*m55*L5r*RSe461p*RSe463p + 4./
         9.*m55*L5r*RSe461p*RSp463e - 2./9.*m55*L5r*pow(RSe461p,2) + 4./
         9.*m55*L5r*RSp461e*RSe463p - 4./9.*m55*L5r*RSp461e*RSe461p + 2.
         /3.*m55*L4r*RSe45613p - 1./3.*m55*L4r*RSe45633p - 2./9.*m55*
         L4r*pow(RSe563p,2) - 2./9.*m55*L4r*pow(RSe463p,2) - 2./9.*m55*
         L4r*pow(RSe453p,2) - 4./9.*m55*L4r*RSp563e*RSe563p - 4./9.*m55
         *L4r*RSp463e*RSe463p - 4./9.*m55*L4r*RSp453e*RSe453p - 1./3.*
         m55*L4r*RSe45611p + 4./9.*m55*L4r*RSe561p*RSe563p + 4./9.*m55*
         L4r*RSe561p*RSp563e - 2./9.*m55*L4r*pow(RSe561p,2) + 4./9.*m55
         *L4r*RSe461p*RSe463p + 4./9.*m55*L4r*RSe461p*RSp463e - 2./9.*
         m55*L4r*pow(RSe461p,2) + 4./9.*m55*L4r*RSe451p*RSe453p + 4./9.
         *m55*L4r*RSe451p*RSp453e );

      F13p6x23L +=  + Ab(me2,mu2) * (  - 2./9.*m55*L4r*pow(RSe451p,2)
          + 4./9.*m55*L4r*RSp561e*RSe563p - 4./9.*m55*L4r*RSp561e*
         RSe561p + 4./9.*m55*L4r*RSp461e*RSe463p - 4./9.*m55*L4r*
         RSp461e*RSe461p + 4./9.*m55*L4r*RSp451e*RSe453p - 4./9.*m55*
         L4r*RSp451e*RSe451p - 4./3.*m55*L4r*RSe465p - 4./9.*m55*mp2*
         L5r*RNep*RSp463e*RSe463p + 4./9.*m55*mp2*L5r*RNep*RSe461p*
         RSp463e + 4./9.*m55*mp2*L5r*RNep*RSp461e*RSe463p - 4./9.*m55*
         mp2*L5r*RNep*RSp461e*RSe461p - 4./9.*m55*mp2*L4r*RNep*RSp563e*
         RSe563p - 4./9.*m55*mp2*L4r*RNep*RSp463e*RSe463p - 4./9.*m55*
         mp2*L4r*RNep*RSp453e*RSe453p + 4./9.*m55*mp2*L4r*RNep*RSe561p*
         RSp563e + 4./9.*m55*mp2*L4r*RNep*RSe461p*RSp463e + 4./9.*m55*
         mp2*L4r*RNep*RSe451p*RSp453e + 4./9.*m55*mp2*L4r*RNep*RSp561e*
         RSe563p - 4./9.*m55*mp2*L4r*RNep*RSp561e*RSe561p + 4./9.*m55*
         mp2*L4r*RNep*RSp461e*RSe463p - 4./9.*m55*mp2*L4r*RNep*RSp461e*
         RSe461p + 4./9.*m55*mp2*L4r*RNep*RSp451e*RSe453p - 4./9.*m55*
         mp2*L4r*RNep*RSp451e*RSe451p );

      F13p6x23L +=  + Ab(me2,mu2) * ( 4./9.*m55*me2*L5r*RN3e*RSe463p*
         RS346pe - 4./9.*m55*me2*L5r*RN3e*RSe461p*RS346pe - 4./9.*m55*
         me2*L5r*RN1e*RS146pe*RSe463p + 4./9.*m55*me2*L5r*RN1e*RSe461p*
         RS146pe + 4./3.*m55*me2*L4r*RN3e*RSe45613p - 4./3.*m55*me2*L4r
         *RN3e*RSe45633p + 4./9.*m55*me2*L4r*RN3e*RSe563p*RS356pe + 4./
         9.*m55*me2*L4r*RN3e*RSe463p*RS346pe + 4./9.*m55*me2*L4r*RN3e*
         RSe453p*RS345pe - 4./9.*m55*me2*L4r*RN3e*RSe561p*RS356pe - 4./
         9.*m55*me2*L4r*RN3e*RSe461p*RS346pe - 4./9.*m55*me2*L4r*RN3e*
         RSe451p*RS345pe + 4./3.*m55*me2*L4r*RN1e*RSe45613p - 4./3.*m55
         *me2*L4r*RN1e*RSe45611p - 4./9.*m55*me2*L4r*RN1e*RS145pe*
         RSe453p - 4./9.*m55*me2*L4r*RN1e*RS146pe*RSe463p - 4./9.*m55*
         me2*L4r*RN1e*RS156pe*RSe563p + 4./9.*m55*me2*L4r*RN1e*RSe561p*
         RS156pe + 4./9.*m55*me2*L4r*RN1e*RSe461p*RS146pe + 4./9.*m55*
         me2*L4r*RN1e*RSe451p*RS145pe + 8./9.*m55*m66*L7r*RNep*RSp463e*
         RSe453p + 8./9.*m55*m66*L7r*RNep*RSp453e*RSe463p - 8./9.*m55*
         m66*L7r*RNep*RSe461p*RSp453e );

      F13p6x23L +=  + Ab(me2,mu2) * (  - 8./9.*m55*m66*L7r*RNep*RSe451p
         *RSp463e - 8./9.*m55*m66*L7r*RNep*RSp461e*RSe453p + 8./9.*m55*
         m66*L7r*RNep*RSp461e*RSe451p - 8./9.*m55*m66*L7r*RNep*RSp451e*
         RSe463p + 8./9.*m55*m66*L7r*RNep*RSp451e*RSe461p - 8./9.*m55*
         m66*L7r*RN3e*RSe463p*RS345pe - 8./9.*m55*m66*L7r*RN3e*RSe453p*
         RS346pe + 8./9.*m55*m66*L7r*RN3e*RSe461p*RS345pe + 8./9.*m55*
         m66*L7r*RN3e*RSe451p*RS346pe + 8./9.*m55*m66*L7r*RN1e*RS145pe*
         RSe463p + 8./9.*m55*m66*L7r*RN1e*RS146pe*RSe453p - 8./9.*m55*
         m66*L7r*RN1e*RSe461p*RS145pe - 8./9.*m55*m66*L7r*RN1e*RSe451p*
         RS146pe + 8./9.*m55*m66*L6r*RNep*RSp463e*RSe463p + 8./9.*m55*
         m66*L6r*RNep*RSp453e*RSe453p - 8./9.*m55*m66*L6r*RNep*RSe461p*
         RSp463e - 8./9.*m55*m66*L6r*RNep*RSe451p*RSp453e - 8./9.*m55*
         m66*L6r*RNep*RSp461e*RSe463p + 8./9.*m55*m66*L6r*RNep*RSp461e*
         RSe461p - 8./9.*m55*m66*L6r*RNep*RSp451e*RSe453p + 8./9.*m55*
         m66*L6r*RNep*RSp451e*RSe451p - 8./9.*m55*m66*L6r*RN3e*RSe463p*
         RS346pe );

      F13p6x23L +=  + Ab(me2,mu2) * (  - 8./9.*m55*m66*L6r*RN3e*RSe453p
         *RS345pe + 8./9.*m55*m66*L6r*RN3e*RSe461p*RS346pe + 8./9.*m55*
         m66*L6r*RN3e*RSe451p*RS345pe + 8./9.*m55*m66*L6r*RN1e*RS145pe*
         RSe453p + 8./9.*m55*m66*L6r*RN1e*RS146pe*RSe463p - 8./9.*m55*
         m66*L6r*RN1e*RSe461p*RS146pe - 8./9.*m55*m66*L6r*RN1e*RSe451p*
         RS145pe + 8./9.*pow(m55,2)*L8r*RNep*RSp463e*RSe463p - 8./9.*
         pow(m55,2)*L8r*RNep*RSe461p*RSp463e - 8./9.*pow(m55,2)*L8r*
         RNep*RSp461e*RSe463p + 8./9.*pow(m55,2)*L8r*RNep*RSp461e*
         RSe461p - 8./9.*pow(m55,2)*L8r*RN3e*RSe463p*RS346pe + 8./9.*
         pow(m55,2)*L8r*RN3e*RSe461p*RS346pe + 8./9.*pow(m55,2)*L8r*
         RN1e*RS146pe*RSe463p - 8./9.*pow(m55,2)*L8r*RN1e*RSe461p*
         RS146pe + 8./9.*pow(m55,2)*L7r*RNep*RSp463e*RSe463p - 8./9.*
         pow(m55,2)*L7r*RNep*RSe461p*RSp463e - 8./9.*pow(m55,2)*L7r*
         RNep*RSp461e*RSe463p + 8./9.*pow(m55,2)*L7r*RNep*RSp461e*
         RSe461p - 8./9.*pow(m55,2)*L7r*RN3e*RSe463p*RS346pe + 8./9.*
         pow(m55,2)*L7r*RN3e*RSe461p*RS346pe );

      F13p6x23L +=  + Ab(me2,mu2) * ( 8./9.*pow(m55,2)*L7r*RN1e*RS146pe
         *RSe463p - 8./9.*pow(m55,2)*L7r*RN1e*RSe461p*RS146pe + 8./9.*
         pow(m55,2)*L6r*RNep*RSp463e*RSe463p - 8./9.*pow(m55,2)*L6r*
         RNep*RSe461p*RSp463e - 8./9.*pow(m55,2)*L6r*RNep*RSp461e*
         RSe463p + 8./9.*pow(m55,2)*L6r*RNep*RSp461e*RSe461p - 8./9.*
         pow(m55,2)*L6r*RN3e*RSe463p*RS346pe + 8./9.*pow(m55,2)*L6r*
         RN3e*RSe461p*RS346pe + 8./9.*pow(m55,2)*L6r*RN1e*RS146pe*
         RSe463p - 8./9.*pow(m55,2)*L6r*RN1e*RSe461p*RS146pe - 2./9.*
         m44*L5r*pow(RSe563p,2) - 4./9.*m44*L5r*RSp563e*RSe563p + 4./9.
         *m44*L5r*RSe561p*RSe563p + 4./9.*m44*L5r*RSe561p*RSp563e - 2./
         9.*m44*L5r*pow(RSe561p,2) + 4./9.*m44*L5r*RSp561e*RSe563p - 4./
         9.*m44*L5r*RSp561e*RSe561p + 2./3.*m44*L4r*RSe45613p - 1./3.*
         m44*L4r*RSe45633p - 2./9.*m44*L4r*pow(RSe563p,2) - 2./9.*m44*
         L4r*pow(RSe463p,2) - 2./9.*m44*L4r*pow(RSe453p,2) - 4./9.*m44*
         L4r*RSp563e*RSe563p - 4./9.*m44*L4r*RSp463e*RSe463p - 4./9.*
         m44*L4r*RSp453e*RSe453p );

      F13p6x23L +=  + Ab(me2,mu2) * (  - 1./3.*m44*L4r*RSe45611p + 4./9.
         *m44*L4r*RSe561p*RSe563p + 4./9.*m44*L4r*RSe561p*RSp563e - 2./
         9.*m44*L4r*pow(RSe561p,2) + 4./9.*m44*L4r*RSe461p*RSe463p + 4./
         9.*m44*L4r*RSe461p*RSp463e - 2./9.*m44*L4r*pow(RSe461p,2) + 4./
         9.*m44*L4r*RSe451p*RSe453p + 4./9.*m44*L4r*RSe451p*RSp453e - 2.
         /9.*m44*L4r*pow(RSe451p,2) + 4./9.*m44*L4r*RSp561e*RSe563p - 4.
         /9.*m44*L4r*RSp561e*RSe561p + 4./9.*m44*L4r*RSp461e*RSe463p - 
         4./9.*m44*L4r*RSp461e*RSe461p + 4./9.*m44*L4r*RSp451e*RSe453p
          - 4./9.*m44*L4r*RSp451e*RSe451p - 4./3.*m44*L4r*RSe564p - 4./
         9.*m44*mp2*L5r*RNep*RSp563e*RSe563p + 4./9.*m44*mp2*L5r*RNep*
         RSe561p*RSp563e + 4./9.*m44*mp2*L5r*RNep*RSp561e*RSe563p - 4./
         9.*m44*mp2*L5r*RNep*RSp561e*RSe561p - 4./9.*m44*mp2*L4r*RNep*
         RSp563e*RSe563p - 4./9.*m44*mp2*L4r*RNep*RSp463e*RSe463p - 4./
         9.*m44*mp2*L4r*RNep*RSp453e*RSe453p + 4./9.*m44*mp2*L4r*RNep*
         RSe561p*RSp563e + 4./9.*m44*mp2*L4r*RNep*RSe461p*RSp463e + 4./
         9.*m44*mp2*L4r*RNep*RSe451p*RSp453e );

      F13p6x23L +=  + Ab(me2,mu2) * ( 4./9.*m44*mp2*L4r*RNep*RSp561e*
         RSe563p - 4./9.*m44*mp2*L4r*RNep*RSp561e*RSe561p + 4./9.*m44*
         mp2*L4r*RNep*RSp461e*RSe463p - 4./9.*m44*mp2*L4r*RNep*RSp461e*
         RSe461p + 4./9.*m44*mp2*L4r*RNep*RSp451e*RSe453p - 4./9.*m44*
         mp2*L4r*RNep*RSp451e*RSe451p + 4./9.*m44*me2*L5r*RN3e*RSe563p*
         RS356pe - 4./9.*m44*me2*L5r*RN3e*RSe561p*RS356pe - 4./9.*m44*
         me2*L5r*RN1e*RS156pe*RSe563p + 4./9.*m44*me2*L5r*RN1e*RSe561p*
         RS156pe + 4./3.*m44*me2*L4r*RN3e*RSe45613p - 4./3.*m44*me2*L4r
         *RN3e*RSe45633p + 4./9.*m44*me2*L4r*RN3e*RSe563p*RS356pe + 4./
         9.*m44*me2*L4r*RN3e*RSe463p*RS346pe + 4./9.*m44*me2*L4r*RN3e*
         RSe453p*RS345pe - 4./9.*m44*me2*L4r*RN3e*RSe561p*RS356pe - 4./
         9.*m44*me2*L4r*RN3e*RSe461p*RS346pe - 4./9.*m44*me2*L4r*RN3e*
         RSe451p*RS345pe + 4./3.*m44*me2*L4r*RN1e*RSe45613p - 4./3.*m44
         *me2*L4r*RN1e*RSe45611p - 4./9.*m44*me2*L4r*RN1e*RS145pe*
         RSe453p - 4./9.*m44*me2*L4r*RN1e*RS146pe*RSe463p - 4./9.*m44*
         me2*L4r*RN1e*RS156pe*RSe563p );

      F13p6x23L +=  + Ab(me2,mu2) * ( 4./9.*m44*me2*L4r*RN1e*RSe561p*
         RS156pe + 4./9.*m44*me2*L4r*RN1e*RSe461p*RS146pe + 4./9.*m44*
         me2*L4r*RN1e*RSe451p*RS145pe + 8./9.*m44*m66*L7r*RNep*RSp563e*
         RSe453p + 8./9.*m44*m66*L7r*RNep*RSp453e*RSe563p - 8./9.*m44*
         m66*L7r*RNep*RSe561p*RSp453e - 8./9.*m44*m66*L7r*RNep*RSe451p*
         RSp563e - 8./9.*m44*m66*L7r*RNep*RSp561e*RSe453p + 8./9.*m44*
         m66*L7r*RNep*RSp561e*RSe451p - 8./9.*m44*m66*L7r*RNep*RSp451e*
         RSe563p + 8./9.*m44*m66*L7r*RNep*RSp451e*RSe561p - 8./9.*m44*
         m66*L7r*RN3e*RSe563p*RS345pe - 8./9.*m44*m66*L7r*RN3e*RSe453p*
         RS356pe + 8./9.*m44*m66*L7r*RN3e*RSe561p*RS345pe + 8./9.*m44*
         m66*L7r*RN3e*RSe451p*RS356pe + 8./9.*m44*m66*L7r*RN1e*RS145pe*
         RSe563p + 8./9.*m44*m66*L7r*RN1e*RS156pe*RSe453p - 8./9.*m44*
         m66*L7r*RN1e*RSe561p*RS145pe - 8./9.*m44*m66*L7r*RN1e*RSe451p*
         RS156pe + 8./9.*m44*m66*L6r*RNep*RSp563e*RSe563p + 8./9.*m44*
         m66*L6r*RNep*RSp453e*RSe453p - 8./9.*m44*m66*L6r*RNep*RSe561p*
         RSp563e );

      F13p6x23L +=  + Ab(me2,mu2) * (  - 8./9.*m44*m66*L6r*RNep*RSe451p
         *RSp453e - 8./9.*m44*m66*L6r*RNep*RSp561e*RSe563p + 8./9.*m44*
         m66*L6r*RNep*RSp561e*RSe561p - 8./9.*m44*m66*L6r*RNep*RSp451e*
         RSe453p + 8./9.*m44*m66*L6r*RNep*RSp451e*RSe451p - 8./9.*m44*
         m66*L6r*RN3e*RSe563p*RS356pe - 8./9.*m44*m66*L6r*RN3e*RSe453p*
         RS345pe + 8./9.*m44*m66*L6r*RN3e*RSe561p*RS356pe + 8./9.*m44*
         m66*L6r*RN3e*RSe451p*RS345pe + 8./9.*m44*m66*L6r*RN1e*RS145pe*
         RSe453p + 8./9.*m44*m66*L6r*RN1e*RS156pe*RSe563p - 8./9.*m44*
         m66*L6r*RN1e*RSe561p*RS156pe - 8./9.*m44*m66*L6r*RN1e*RSe451p*
         RS145pe + 8./9.*m44*m55*L7r*RNep*RSp563e*RSe463p + 8./9.*m44*
         m55*L7r*RNep*RSp463e*RSe563p - 8./9.*m44*m55*L7r*RNep*RSe561p*
         RSp463e - 8./9.*m44*m55*L7r*RNep*RSe461p*RSp563e - 8./9.*m44*
         m55*L7r*RNep*RSp561e*RSe463p + 8./9.*m44*m55*L7r*RNep*RSp561e*
         RSe461p - 8./9.*m44*m55*L7r*RNep*RSp461e*RSe563p + 8./9.*m44*
         m55*L7r*RNep*RSp461e*RSe561p - 8./9.*m44*m55*L7r*RN3e*RSe563p*
         RS346pe );

      F13p6x23L +=  + Ab(me2,mu2) * (  - 8./9.*m44*m55*L7r*RN3e*RSe463p
         *RS356pe + 8./9.*m44*m55*L7r*RN3e*RSe561p*RS346pe + 8./9.*m44*
         m55*L7r*RN3e*RSe461p*RS356pe + 8./9.*m44*m55*L7r*RN1e*RS146pe*
         RSe563p + 8./9.*m44*m55*L7r*RN1e*RS156pe*RSe463p - 8./9.*m44*
         m55*L7r*RN1e*RSe561p*RS146pe - 8./9.*m44*m55*L7r*RN1e*RSe461p*
         RS156pe + 8./9.*m44*m55*L6r*RNep*RSp563e*RSe563p + 8./9.*m44*
         m55*L6r*RNep*RSp463e*RSe463p - 8./9.*m44*m55*L6r*RNep*RSe561p*
         RSp563e - 8./9.*m44*m55*L6r*RNep*RSe461p*RSp463e - 8./9.*m44*
         m55*L6r*RNep*RSp561e*RSe563p + 8./9.*m44*m55*L6r*RNep*RSp561e*
         RSe561p - 8./9.*m44*m55*L6r*RNep*RSp461e*RSe463p + 8./9.*m44*
         m55*L6r*RNep*RSp461e*RSe461p - 8./9.*m44*m55*L6r*RN3e*RSe563p*
         RS356pe - 8./9.*m44*m55*L6r*RN3e*RSe463p*RS346pe + 8./9.*m44*
         m55*L6r*RN3e*RSe561p*RS356pe + 8./9.*m44*m55*L6r*RN3e*RSe461p*
         RS346pe + 8./9.*m44*m55*L6r*RN1e*RS146pe*RSe463p + 8./9.*m44*
         m55*L6r*RN1e*RS156pe*RSe563p - 8./9.*m44*m55*L6r*RN1e*RSe561p*
         RS156pe );

      F13p6x23L +=  + Ab(me2,mu2) * (  - 8./9.*m44*m55*L6r*RN1e*RSe461p
         *RS146pe + 8./9.*pow(m44,2)*L8r*RNep*RSp563e*RSe563p - 8./9.*
         pow(m44,2)*L8r*RNep*RSe561p*RSp563e - 8./9.*pow(m44,2)*L8r*
         RNep*RSp561e*RSe563p + 8./9.*pow(m44,2)*L8r*RNep*RSp561e*
         RSe561p - 8./9.*pow(m44,2)*L8r*RN3e*RSe563p*RS356pe + 8./9.*
         pow(m44,2)*L8r*RN3e*RSe561p*RS356pe + 8./9.*pow(m44,2)*L8r*
         RN1e*RS156pe*RSe563p - 8./9.*pow(m44,2)*L8r*RN1e*RSe561p*
         RS156pe + 8./9.*pow(m44,2)*L7r*RNep*RSp563e*RSe563p - 8./9.*
         pow(m44,2)*L7r*RNep*RSe561p*RSp563e - 8./9.*pow(m44,2)*L7r*
         RNep*RSp561e*RSe563p + 8./9.*pow(m44,2)*L7r*RNep*RSp561e*
         RSe561p - 8./9.*pow(m44,2)*L7r*RN3e*RSe563p*RS356pe + 8./9.*
         pow(m44,2)*L7r*RN3e*RSe561p*RS356pe + 8./9.*pow(m44,2)*L7r*
         RN1e*RS156pe*RSe563p - 8./9.*pow(m44,2)*L7r*RN1e*RSe561p*
         RS156pe + 8./9.*pow(m44,2)*L6r*RNep*RSp563e*RSe563p - 8./9.*
         pow(m44,2)*L6r*RNep*RSe561p*RSp563e - 8./9.*pow(m44,2)*L6r*
         RNep*RSp561e*RSe563p );

      F13p6x23L +=  + Ab(me2,mu2) * ( 8./9.*pow(m44,2)*L6r*RNep*RSp561e
         *RSe561p - 8./9.*pow(m44,2)*L6r*RN3e*RSe563p*RS356pe + 8./9.*
         pow(m44,2)*L6r*RN3e*RSe561p*RS356pe + 8./9.*pow(m44,2)*L6r*
         RN1e*RS156pe*RSe563p - 8./9.*pow(m44,2)*L6r*RN1e*RSe561p*
         RS156pe + 1./3.*m33*L5r*RSe45613p - 5./6.*m33*L5r*RSe45633p - 
         1./6.*m33*L5r*RSe45611p + 4./3.*m33*me2*L5r*RN3e*RSe45613p - 4.
         /3.*m33*me2*L5r*RN3e*RSe45633p + 8./3.*m33*m66*L7r*RN3e*
         RSe453p - 8./3.*m33*m66*L7r*RN3e*RSe451p - 8./3.*m33*m66*L6r*
         RN3e*RSe45613p + 8./3.*m33*m66*L6r*RN3e*RSe45633p + 8./3.*m33*
         m55*L7r*RN3e*RSe463p - 8./3.*m33*m55*L7r*RN3e*RSe461p - 8./3.*
         m33*m55*L6r*RN3e*RSe45613p + 8./3.*m33*m55*L6r*RN3e*RSe45633p
          + 8./3.*m33*m44*L7r*RN3e*RSe563p - 8./3.*m33*m44*L7r*RN3e*
         RSe561p - 8./3.*m33*m44*L6r*RN3e*RSe45613p + 8./3.*m33*m44*L6r
         *RN3e*RSe45633p - 8./3.*pow(m33,2)*L8r*RN3e*RSe45613p + 8./3.*
         pow(m33,2)*L8r*RN3e*RSe45633p + 1./3.*m11*L5r*RSe45613p - 1./6.
         *m11*L5r*RSe45633p );

      F13p6x23L +=  + Ab(me2,mu2) * (  - 5./6.*m11*L5r*RSe45611p + 4./3.
         *m11*me2*L5r*RN1e*RSe45613p - 4./3.*m11*me2*L5r*RN1e*RSe45611p
          - 8./3.*m11*m66*L7r*RN1e*RSe453p + 8./3.*m11*m66*L7r*RN1e*
         RSe451p - 8./3.*m11*m66*L6r*RN1e*RSe45613p + 8./3.*m11*m66*L6r
         *RN1e*RSe45611p - 8./3.*m11*m55*L7r*RN1e*RSe463p + 8./3.*m11*
         m55*L7r*RN1e*RSe461p - 8./3.*m11*m55*L6r*RN1e*RSe45613p + 8./3.
         *m11*m55*L6r*RN1e*RSe45611p - 8./3.*m11*m44*L7r*RN1e*RSe563p
          + 8./3.*m11*m44*L7r*RN1e*RSe561p - 8./3.*m11*m44*L6r*RN1e*
         RSe45613p + 8./3.*m11*m44*L6r*RN1e*RSe45611p - 8./3.*pow(
         m11,2)*L8r*RN1e*RSe45613p + 8./3.*pow(m11,2)*L8r*RN1e*
         RSe45611p );

      F13p6x23L +=  + Ab(mp2,mu2) * ( 2./3.*mp2*L3r*RSp45613e + 5./3.*
         mp2*L3r*RSp45633e + 5./3.*mp2*L3r*RSp45611e + 2./3.*mp2*L2r*
         RSp564e + 2./3.*mp2*L2r*RSp465e + 2./3.*mp2*L2r*RSp456e + 8./3.
         *mp2*L1r*RSp564e + 8./3.*mp2*L1r*RSp465e + 8./3.*mp2*L1r*
         RSp456e + 8./3.*mp2*L0r*RSp45613e + 2./3.*mp2*L0r*RSp45633e + 
         2./3.*mp2*L0r*RSp45611e - 2./9.*m66*L5r*pow(RSp453e,2) + 4./9.
         *m66*L5r*RSp451e*RSp453e - 2./9.*m66*L5r*pow(RSp451e,2) + 2./3.
         *m66*L4r*RSp45613e - 1./3.*m66*L4r*RSp45633e - 2./9.*m66*L4r*
         pow(RSp563e,2) - 2./9.*m66*L4r*pow(RSp463e,2) - 2./9.*m66*L4r*
         pow(RSp453e,2) - 1./3.*m66*L4r*RSp45611e + 4./9.*m66*L4r*
         RSp561e*RSp563e - 2./9.*m66*L4r*pow(RSp561e,2) + 4./9.*m66*L4r
         *RSp461e*RSp463e - 2./9.*m66*L4r*pow(RSp461e,2) + 4./9.*m66*
         L4r*RSp451e*RSp453e - 2./9.*m66*L4r*pow(RSp451e,2) - 4./3.*m66
         *L4r*RSp456e + 4./9.*m66*mp2*L5r*RNep*RSp453e*RSe453p - 4./9.*
         m66*mp2*L5r*RNep*RSe451p*RSp453e - 4./9.*m66*mp2*L5r*RNep*
         RSp451e*RSe453p );

      F13p6x23L +=  + Ab(mp2,mu2) * ( 4./9.*m66*mp2*L5r*RNep*RSp451e*
         RSe451p + 4./9.*m66*mp2*L5r*RN3p*RSp453e*RS345pe - 4./9.*m66*
         mp2*L5r*RN3p*RSp451e*RS345pe - 4./9.*m66*mp2*L5r*RN1p*RS145pe*
         RSp453e + 4./9.*m66*mp2*L5r*RN1p*RSp451e*RS145pe + 4./9.*m66*
         mp2*L4r*RNep*RSp563e*RSe563p + 4./9.*m66*mp2*L4r*RNep*RSp463e*
         RSe463p + 4./9.*m66*mp2*L4r*RNep*RSp453e*RSe453p - 4./9.*m66*
         mp2*L4r*RNep*RSe561p*RSp563e - 4./9.*m66*mp2*L4r*RNep*RSe461p*
         RSp463e - 4./9.*m66*mp2*L4r*RNep*RSe451p*RSp453e - 4./9.*m66*
         mp2*L4r*RNep*RSp561e*RSe563p + 4./9.*m66*mp2*L4r*RNep*RSp561e*
         RSe561p - 4./9.*m66*mp2*L4r*RNep*RSp461e*RSe463p + 4./9.*m66*
         mp2*L4r*RNep*RSp461e*RSe461p - 4./9.*m66*mp2*L4r*RNep*RSp451e*
         RSe453p + 4./9.*m66*mp2*L4r*RNep*RSp451e*RSe451p + 4./3.*m66*
         mp2*L4r*RN3p*RSp45613e - 4./3.*m66*mp2*L4r*RN3p*RSp45633e + 4./
         9.*m66*mp2*L4r*RN3p*RSp563e*RS356pe + 4./9.*m66*mp2*L4r*RN3p*
         RSp463e*RS346pe + 4./9.*m66*mp2*L4r*RN3p*RSp453e*RS345pe - 4./
         9.*m66*mp2*L4r*RN3p*RSp561e*RS356pe );

      F13p6x23L +=  + Ab(mp2,mu2) * (  - 4./9.*m66*mp2*L4r*RN3p*RSp461e
         *RS346pe - 4./9.*m66*mp2*L4r*RN3p*RSp451e*RS345pe + 4./3.*m66*
         mp2*L4r*RN1p*RSp45613e - 4./3.*m66*mp2*L4r*RN1p*RSp45611e - 4./
         9.*m66*mp2*L4r*RN1p*RS145pe*RSp453e - 4./9.*m66*mp2*L4r*RN1p*
         RS146pe*RSp463e - 4./9.*m66*mp2*L4r*RN1p*RS156pe*RSp563e + 4./
         9.*m66*mp2*L4r*RN1p*RSp561e*RS156pe + 4./9.*m66*mp2*L4r*RN1p*
         RSp461e*RS146pe + 4./9.*m66*mp2*L4r*RN1p*RSp451e*RS145pe - 8./
         9.*pow(m66,2)*L8r*RNep*RSp453e*RSe453p + 8./9.*pow(m66,2)*L8r*
         RNep*RSe451p*RSp453e + 8./9.*pow(m66,2)*L8r*RNep*RSp451e*
         RSe453p - 8./9.*pow(m66,2)*L8r*RNep*RSp451e*RSe451p - 8./9.*
         pow(m66,2)*L8r*RN3p*RSp453e*RS345pe + 8./9.*pow(m66,2)*L8r*
         RN3p*RSp451e*RS345pe + 8./9.*pow(m66,2)*L8r*RN1p*RS145pe*
         RSp453e - 8./9.*pow(m66,2)*L8r*RN1p*RSp451e*RS145pe - 8./9.*
         pow(m66,2)*L7r*RNep*RSp453e*RSe453p + 8./9.*pow(m66,2)*L7r*
         RNep*RSe451p*RSp453e + 8./9.*pow(m66,2)*L7r*RNep*RSp451e*
         RSe453p );

      F13p6x23L +=  + Ab(mp2,mu2) * (  - 8./9.*pow(m66,2)*L7r*RNep*
         RSp451e*RSe451p - 8./9.*pow(m66,2)*L7r*RN3p*RSp453e*RS345pe + 
         8./9.*pow(m66,2)*L7r*RN3p*RSp451e*RS345pe + 8./9.*pow(m66,2)*
         L7r*RN1p*RS145pe*RSp453e - 8./9.*pow(m66,2)*L7r*RN1p*RSp451e*
         RS145pe - 8./9.*pow(m66,2)*L6r*RNep*RSp453e*RSe453p + 8./9.*
         pow(m66,2)*L6r*RNep*RSe451p*RSp453e + 8./9.*pow(m66,2)*L6r*
         RNep*RSp451e*RSe453p - 8./9.*pow(m66,2)*L6r*RNep*RSp451e*
         RSe451p - 8./9.*pow(m66,2)*L6r*RN3p*RSp453e*RS345pe + 8./9.*
         pow(m66,2)*L6r*RN3p*RSp451e*RS345pe + 8./9.*pow(m66,2)*L6r*
         RN1p*RS145pe*RSp453e - 8./9.*pow(m66,2)*L6r*RN1p*RSp451e*
         RS145pe - 2./9.*m55*L5r*pow(RSp463e,2) + 4./9.*m55*L5r*RSp461e
         *RSp463e - 2./9.*m55*L5r*pow(RSp461e,2) + 2./3.*m55*L4r*
         RSp45613e - 1./3.*m55*L4r*RSp45633e - 2./9.*m55*L4r*pow(
         RSp563e,2) - 2./9.*m55*L4r*pow(RSp463e,2) - 2./9.*m55*L4r*pow(
         RSp453e,2) - 1./3.*m55*L4r*RSp45611e + 4./9.*m55*L4r*RSp561e*
         RSp563e );

      F13p6x23L +=  + Ab(mp2,mu2) * (  - 2./9.*m55*L4r*pow(RSp561e,2)
          + 4./9.*m55*L4r*RSp461e*RSp463e - 2./9.*m55*L4r*pow(
         RSp461e,2) + 4./9.*m55*L4r*RSp451e*RSp453e - 2./9.*m55*L4r*
         pow(RSp451e,2) - 4./3.*m55*L4r*RSp465e + 4./9.*m55*mp2*L5r*
         RNep*RSp463e*RSe463p - 4./9.*m55*mp2*L5r*RNep*RSe461p*RSp463e
          - 4./9.*m55*mp2*L5r*RNep*RSp461e*RSe463p + 4./9.*m55*mp2*L5r*
         RNep*RSp461e*RSe461p + 4./9.*m55*mp2*L5r*RN3p*RSp463e*RS346pe
          - 4./9.*m55*mp2*L5r*RN3p*RSp461e*RS346pe - 4./9.*m55*mp2*L5r*
         RN1p*RS146pe*RSp463e + 4./9.*m55*mp2*L5r*RN1p*RSp461e*RS146pe
          + 4./9.*m55*mp2*L4r*RNep*RSp563e*RSe563p + 4./9.*m55*mp2*L4r*
         RNep*RSp463e*RSe463p + 4./9.*m55*mp2*L4r*RNep*RSp453e*RSe453p
          - 4./9.*m55*mp2*L4r*RNep*RSe561p*RSp563e - 4./9.*m55*mp2*L4r*
         RNep*RSe461p*RSp463e - 4./9.*m55*mp2*L4r*RNep*RSe451p*RSp453e
          - 4./9.*m55*mp2*L4r*RNep*RSp561e*RSe563p + 4./9.*m55*mp2*L4r*
         RNep*RSp561e*RSe561p - 4./9.*m55*mp2*L4r*RNep*RSp461e*RSe463p
          + 4./9.*m55*mp2*L4r*RNep*RSp461e*RSe461p );

      F13p6x23L +=  + Ab(mp2,mu2) * (  - 4./9.*m55*mp2*L4r*RNep*RSp451e
         *RSe453p + 4./9.*m55*mp2*L4r*RNep*RSp451e*RSe451p + 4./3.*m55*
         mp2*L4r*RN3p*RSp45613e - 4./3.*m55*mp2*L4r*RN3p*RSp45633e + 4./
         9.*m55*mp2*L4r*RN3p*RSp563e*RS356pe + 4./9.*m55*mp2*L4r*RN3p*
         RSp463e*RS346pe + 4./9.*m55*mp2*L4r*RN3p*RSp453e*RS345pe - 4./
         9.*m55*mp2*L4r*RN3p*RSp561e*RS356pe - 4./9.*m55*mp2*L4r*RN3p*
         RSp461e*RS346pe - 4./9.*m55*mp2*L4r*RN3p*RSp451e*RS345pe + 4./
         3.*m55*mp2*L4r*RN1p*RSp45613e - 4./3.*m55*mp2*L4r*RN1p*
         RSp45611e - 4./9.*m55*mp2*L4r*RN1p*RS145pe*RSp453e - 4./9.*m55
         *mp2*L4r*RN1p*RS146pe*RSp463e - 4./9.*m55*mp2*L4r*RN1p*RS156pe
         *RSp563e + 4./9.*m55*mp2*L4r*RN1p*RSp561e*RS156pe + 4./9.*m55*
         mp2*L4r*RN1p*RSp461e*RS146pe + 4./9.*m55*mp2*L4r*RN1p*RSp451e*
         RS145pe - 8./9.*m55*m66*L7r*RNep*RSp463e*RSe453p - 8./9.*m55*
         m66*L7r*RNep*RSp453e*RSe463p + 8./9.*m55*m66*L7r*RNep*RSe461p*
         RSp453e + 8./9.*m55*m66*L7r*RNep*RSe451p*RSp463e + 8./9.*m55*
         m66*L7r*RNep*RSp461e*RSe453p );

      F13p6x23L +=  + Ab(mp2,mu2) * (  - 8./9.*m55*m66*L7r*RNep*RSp461e
         *RSe451p + 8./9.*m55*m66*L7r*RNep*RSp451e*RSe463p - 8./9.*m55*
         m66*L7r*RNep*RSp451e*RSe461p - 8./9.*m55*m66*L7r*RN3p*RSp463e*
         RS345pe - 8./9.*m55*m66*L7r*RN3p*RSp453e*RS346pe + 8./9.*m55*
         m66*L7r*RN3p*RSp461e*RS345pe + 8./9.*m55*m66*L7r*RN3p*RSp451e*
         RS346pe + 8./9.*m55*m66*L7r*RN1p*RS145pe*RSp463e + 8./9.*m55*
         m66*L7r*RN1p*RS146pe*RSp453e - 8./9.*m55*m66*L7r*RN1p*RSp461e*
         RS145pe - 8./9.*m55*m66*L7r*RN1p*RSp451e*RS146pe - 8./9.*m55*
         m66*L6r*RNep*RSp463e*RSe463p - 8./9.*m55*m66*L6r*RNep*RSp453e*
         RSe453p + 8./9.*m55*m66*L6r*RNep*RSe461p*RSp463e + 8./9.*m55*
         m66*L6r*RNep*RSe451p*RSp453e + 8./9.*m55*m66*L6r*RNep*RSp461e*
         RSe463p - 8./9.*m55*m66*L6r*RNep*RSp461e*RSe461p + 8./9.*m55*
         m66*L6r*RNep*RSp451e*RSe453p - 8./9.*m55*m66*L6r*RNep*RSp451e*
         RSe451p - 8./9.*m55*m66*L6r*RN3p*RSp463e*RS346pe - 8./9.*m55*
         m66*L6r*RN3p*RSp453e*RS345pe + 8./9.*m55*m66*L6r*RN3p*RSp461e*
         RS346pe );

      F13p6x23L +=  + Ab(mp2,mu2) * ( 8./9.*m55*m66*L6r*RN3p*RSp451e*
         RS345pe + 8./9.*m55*m66*L6r*RN1p*RS145pe*RSp453e + 8./9.*m55*
         m66*L6r*RN1p*RS146pe*RSp463e - 8./9.*m55*m66*L6r*RN1p*RSp461e*
         RS146pe - 8./9.*m55*m66*L6r*RN1p*RSp451e*RS145pe - 8./9.*pow(
         m55,2)*L8r*RNep*RSp463e*RSe463p + 8./9.*pow(m55,2)*L8r*RNep*
         RSe461p*RSp463e + 8./9.*pow(m55,2)*L8r*RNep*RSp461e*RSe463p - 
         8./9.*pow(m55,2)*L8r*RNep*RSp461e*RSe461p - 8./9.*pow(m55,2)*
         L8r*RN3p*RSp463e*RS346pe + 8./9.*pow(m55,2)*L8r*RN3p*RSp461e*
         RS346pe + 8./9.*pow(m55,2)*L8r*RN1p*RS146pe*RSp463e - 8./9.*
         pow(m55,2)*L8r*RN1p*RSp461e*RS146pe - 8./9.*pow(m55,2)*L7r*
         RNep*RSp463e*RSe463p + 8./9.*pow(m55,2)*L7r*RNep*RSe461p*
         RSp463e + 8./9.*pow(m55,2)*L7r*RNep*RSp461e*RSe463p - 8./9.*
         pow(m55,2)*L7r*RNep*RSp461e*RSe461p - 8./9.*pow(m55,2)*L7r*
         RN3p*RSp463e*RS346pe + 8./9.*pow(m55,2)*L7r*RN3p*RSp461e*
         RS346pe + 8./9.*pow(m55,2)*L7r*RN1p*RS146pe*RSp463e - 8./9.*
         pow(m55,2)*L7r*RN1p*RSp461e*RS146pe );

      F13p6x23L +=  + Ab(mp2,mu2) * (  - 8./9.*pow(m55,2)*L6r*RNep*
         RSp463e*RSe463p + 8./9.*pow(m55,2)*L6r*RNep*RSe461p*RSp463e + 
         8./9.*pow(m55,2)*L6r*RNep*RSp461e*RSe463p - 8./9.*pow(m55,2)*
         L6r*RNep*RSp461e*RSe461p - 8./9.*pow(m55,2)*L6r*RN3p*RSp463e*
         RS346pe + 8./9.*pow(m55,2)*L6r*RN3p*RSp461e*RS346pe + 8./9.*
         pow(m55,2)*L6r*RN1p*RS146pe*RSp463e - 8./9.*pow(m55,2)*L6r*
         RN1p*RSp461e*RS146pe - 2./9.*m44*L5r*pow(RSp563e,2) + 4./9.*
         m44*L5r*RSp561e*RSp563e - 2./9.*m44*L5r*pow(RSp561e,2) + 2./3.
         *m44*L4r*RSp45613e - 1./3.*m44*L4r*RSp45633e - 2./9.*m44*L4r*
         pow(RSp563e,2) - 2./9.*m44*L4r*pow(RSp463e,2) - 2./9.*m44*L4r*
         pow(RSp453e,2) - 1./3.*m44*L4r*RSp45611e + 4./9.*m44*L4r*
         RSp561e*RSp563e - 2./9.*m44*L4r*pow(RSp561e,2) + 4./9.*m44*L4r
         *RSp461e*RSp463e - 2./9.*m44*L4r*pow(RSp461e,2) + 4./9.*m44*
         L4r*RSp451e*RSp453e - 2./9.*m44*L4r*pow(RSp451e,2) - 4./3.*m44
         *L4r*RSp564e + 4./9.*m44*mp2*L5r*RNep*RSp563e*RSe563p - 4./9.*
         m44*mp2*L5r*RNep*RSe561p*RSp563e );

      F13p6x23L +=  + Ab(mp2,mu2) * (  - 4./9.*m44*mp2*L5r*RNep*RSp561e
         *RSe563p + 4./9.*m44*mp2*L5r*RNep*RSp561e*RSe561p + 4./9.*m44*
         mp2*L5r*RN3p*RSp563e*RS356pe - 4./9.*m44*mp2*L5r*RN3p*RSp561e*
         RS356pe - 4./9.*m44*mp2*L5r*RN1p*RS156pe*RSp563e + 4./9.*m44*
         mp2*L5r*RN1p*RSp561e*RS156pe + 4./9.*m44*mp2*L4r*RNep*RSp563e*
         RSe563p + 4./9.*m44*mp2*L4r*RNep*RSp463e*RSe463p + 4./9.*m44*
         mp2*L4r*RNep*RSp453e*RSe453p - 4./9.*m44*mp2*L4r*RNep*RSe561p*
         RSp563e - 4./9.*m44*mp2*L4r*RNep*RSe461p*RSp463e - 4./9.*m44*
         mp2*L4r*RNep*RSe451p*RSp453e - 4./9.*m44*mp2*L4r*RNep*RSp561e*
         RSe563p + 4./9.*m44*mp2*L4r*RNep*RSp561e*RSe561p - 4./9.*m44*
         mp2*L4r*RNep*RSp461e*RSe463p + 4./9.*m44*mp2*L4r*RNep*RSp461e*
         RSe461p - 4./9.*m44*mp2*L4r*RNep*RSp451e*RSe453p + 4./9.*m44*
         mp2*L4r*RNep*RSp451e*RSe451p + 4./3.*m44*mp2*L4r*RN3p*
         RSp45613e - 4./3.*m44*mp2*L4r*RN3p*RSp45633e + 4./9.*m44*mp2*
         L4r*RN3p*RSp563e*RS356pe + 4./9.*m44*mp2*L4r*RN3p*RSp463e*
         RS346pe );

      F13p6x23L +=  + Ab(mp2,mu2) * ( 4./9.*m44*mp2*L4r*RN3p*RSp453e*
         RS345pe - 4./9.*m44*mp2*L4r*RN3p*RSp561e*RS356pe - 4./9.*m44*
         mp2*L4r*RN3p*RSp461e*RS346pe - 4./9.*m44*mp2*L4r*RN3p*RSp451e*
         RS345pe + 4./3.*m44*mp2*L4r*RN1p*RSp45613e - 4./3.*m44*mp2*L4r
         *RN1p*RSp45611e - 4./9.*m44*mp2*L4r*RN1p*RS145pe*RSp453e - 4./
         9.*m44*mp2*L4r*RN1p*RS146pe*RSp463e - 4./9.*m44*mp2*L4r*RN1p*
         RS156pe*RSp563e + 4./9.*m44*mp2*L4r*RN1p*RSp561e*RS156pe + 4./
         9.*m44*mp2*L4r*RN1p*RSp461e*RS146pe + 4./9.*m44*mp2*L4r*RN1p*
         RSp451e*RS145pe - 8./9.*m44*m66*L7r*RNep*RSp563e*RSe453p - 8./
         9.*m44*m66*L7r*RNep*RSp453e*RSe563p + 8./9.*m44*m66*L7r*RNep*
         RSe561p*RSp453e + 8./9.*m44*m66*L7r*RNep*RSe451p*RSp563e + 8./
         9.*m44*m66*L7r*RNep*RSp561e*RSe453p - 8./9.*m44*m66*L7r*RNep*
         RSp561e*RSe451p + 8./9.*m44*m66*L7r*RNep*RSp451e*RSe563p - 8./
         9.*m44*m66*L7r*RNep*RSp451e*RSe561p - 8./9.*m44*m66*L7r*RN3p*
         RSp563e*RS345pe - 8./9.*m44*m66*L7r*RN3p*RSp453e*RS356pe + 8./
         9.*m44*m66*L7r*RN3p*RSp561e*RS345pe );

      F13p6x23L +=  + Ab(mp2,mu2) * ( 8./9.*m44*m66*L7r*RN3p*RSp451e*
         RS356pe + 8./9.*m44*m66*L7r*RN1p*RS145pe*RSp563e + 8./9.*m44*
         m66*L7r*RN1p*RS156pe*RSp453e - 8./9.*m44*m66*L7r*RN1p*RSp561e*
         RS145pe - 8./9.*m44*m66*L7r*RN1p*RSp451e*RS156pe - 8./9.*m44*
         m66*L6r*RNep*RSp563e*RSe563p - 8./9.*m44*m66*L6r*RNep*RSp453e*
         RSe453p + 8./9.*m44*m66*L6r*RNep*RSe561p*RSp563e + 8./9.*m44*
         m66*L6r*RNep*RSe451p*RSp453e + 8./9.*m44*m66*L6r*RNep*RSp561e*
         RSe563p - 8./9.*m44*m66*L6r*RNep*RSp561e*RSe561p + 8./9.*m44*
         m66*L6r*RNep*RSp451e*RSe453p - 8./9.*m44*m66*L6r*RNep*RSp451e*
         RSe451p - 8./9.*m44*m66*L6r*RN3p*RSp563e*RS356pe - 8./9.*m44*
         m66*L6r*RN3p*RSp453e*RS345pe + 8./9.*m44*m66*L6r*RN3p*RSp561e*
         RS356pe + 8./9.*m44*m66*L6r*RN3p*RSp451e*RS345pe + 8./9.*m44*
         m66*L6r*RN1p*RS145pe*RSp453e + 8./9.*m44*m66*L6r*RN1p*RS156pe*
         RSp563e - 8./9.*m44*m66*L6r*RN1p*RSp561e*RS156pe - 8./9.*m44*
         m66*L6r*RN1p*RSp451e*RS145pe - 8./9.*m44*m55*L7r*RNep*RSp563e*
         RSe463p );

      F13p6x23L +=  + Ab(mp2,mu2) * (  - 8./9.*m44*m55*L7r*RNep*RSp463e
         *RSe563p + 8./9.*m44*m55*L7r*RNep*RSe561p*RSp463e + 8./9.*m44*
         m55*L7r*RNep*RSe461p*RSp563e + 8./9.*m44*m55*L7r*RNep*RSp561e*
         RSe463p - 8./9.*m44*m55*L7r*RNep*RSp561e*RSe461p + 8./9.*m44*
         m55*L7r*RNep*RSp461e*RSe563p - 8./9.*m44*m55*L7r*RNep*RSp461e*
         RSe561p - 8./9.*m44*m55*L7r*RN3p*RSp563e*RS346pe - 8./9.*m44*
         m55*L7r*RN3p*RSp463e*RS356pe + 8./9.*m44*m55*L7r*RN3p*RSp561e*
         RS346pe + 8./9.*m44*m55*L7r*RN3p*RSp461e*RS356pe + 8./9.*m44*
         m55*L7r*RN1p*RS146pe*RSp563e + 8./9.*m44*m55*L7r*RN1p*RS156pe*
         RSp463e - 8./9.*m44*m55*L7r*RN1p*RSp561e*RS146pe - 8./9.*m44*
         m55*L7r*RN1p*RSp461e*RS156pe - 8./9.*m44*m55*L6r*RNep*RSp563e*
         RSe563p - 8./9.*m44*m55*L6r*RNep*RSp463e*RSe463p + 8./9.*m44*
         m55*L6r*RNep*RSe561p*RSp563e + 8./9.*m44*m55*L6r*RNep*RSe461p*
         RSp463e + 8./9.*m44*m55*L6r*RNep*RSp561e*RSe563p - 8./9.*m44*
         m55*L6r*RNep*RSp561e*RSe561p + 8./9.*m44*m55*L6r*RNep*RSp461e*
         RSe463p );

      F13p6x23L +=  + Ab(mp2,mu2) * (  - 8./9.*m44*m55*L6r*RNep*RSp461e
         *RSe461p - 8./9.*m44*m55*L6r*RN3p*RSp563e*RS356pe - 8./9.*m44*
         m55*L6r*RN3p*RSp463e*RS346pe + 8./9.*m44*m55*L6r*RN3p*RSp561e*
         RS356pe + 8./9.*m44*m55*L6r*RN3p*RSp461e*RS346pe + 8./9.*m44*
         m55*L6r*RN1p*RS146pe*RSp463e + 8./9.*m44*m55*L6r*RN1p*RS156pe*
         RSp563e - 8./9.*m44*m55*L6r*RN1p*RSp561e*RS156pe - 8./9.*m44*
         m55*L6r*RN1p*RSp461e*RS146pe - 8./9.*pow(m44,2)*L8r*RNep*
         RSp563e*RSe563p + 8./9.*pow(m44,2)*L8r*RNep*RSe561p*RSp563e + 
         8./9.*pow(m44,2)*L8r*RNep*RSp561e*RSe563p - 8./9.*pow(m44,2)*
         L8r*RNep*RSp561e*RSe561p - 8./9.*pow(m44,2)*L8r*RN3p*RSp563e*
         RS356pe + 8./9.*pow(m44,2)*L8r*RN3p*RSp561e*RS356pe + 8./9.*
         pow(m44,2)*L8r*RN1p*RS156pe*RSp563e - 8./9.*pow(m44,2)*L8r*
         RN1p*RSp561e*RS156pe - 8./9.*pow(m44,2)*L7r*RNep*RSp563e*
         RSe563p + 8./9.*pow(m44,2)*L7r*RNep*RSe561p*RSp563e + 8./9.*
         pow(m44,2)*L7r*RNep*RSp561e*RSe563p - 8./9.*pow(m44,2)*L7r*
         RNep*RSp561e*RSe561p );

      F13p6x23L +=  + Ab(mp2,mu2) * (  - 8./9.*pow(m44,2)*L7r*RN3p*
         RSp563e*RS356pe + 8./9.*pow(m44,2)*L7r*RN3p*RSp561e*RS356pe + 
         8./9.*pow(m44,2)*L7r*RN1p*RS156pe*RSp563e - 8./9.*pow(m44,2)*
         L7r*RN1p*RSp561e*RS156pe - 8./9.*pow(m44,2)*L6r*RNep*RSp563e*
         RSe563p + 8./9.*pow(m44,2)*L6r*RNep*RSe561p*RSp563e + 8./9.*
         pow(m44,2)*L6r*RNep*RSp561e*RSe563p - 8./9.*pow(m44,2)*L6r*
         RNep*RSp561e*RSe561p - 8./9.*pow(m44,2)*L6r*RN3p*RSp563e*
         RS356pe + 8./9.*pow(m44,2)*L6r*RN3p*RSp561e*RS356pe + 8./9.*
         pow(m44,2)*L6r*RN1p*RS156pe*RSp563e - 8./9.*pow(m44,2)*L6r*
         RN1p*RSp561e*RS156pe + 1./3.*m33*L5r*RSp45613e - 5./6.*m33*L5r
         *RSp45633e - 1./6.*m33*L5r*RSp45611e + 4./3.*m33*mp2*L5r*RN3p*
         RSp45613e - 4./3.*m33*mp2*L5r*RN3p*RSp45633e + 8./3.*m33*m66*
         L7r*RN3p*RSp453e - 8./3.*m33*m66*L7r*RN3p*RSp451e - 8./3.*m33*
         m66*L6r*RN3p*RSp45613e + 8./3.*m33*m66*L6r*RN3p*RSp45633e + 8./
         3.*m33*m55*L7r*RN3p*RSp463e - 8./3.*m33*m55*L7r*RN3p*RSp461e
          - 8./3.*m33*m55*L6r*RN3p*RSp45613e );

      F13p6x23L +=  + Ab(mp2,mu2) * ( 8./3.*m33*m55*L6r*RN3p*RSp45633e
          + 8./3.*m33*m44*L7r*RN3p*RSp563e - 8./3.*m33*m44*L7r*RN3p*
         RSp561e - 8./3.*m33*m44*L6r*RN3p*RSp45613e + 8./3.*m33*m44*L6r
         *RN3p*RSp45633e - 8./3.*pow(m33,2)*L8r*RN3p*RSp45613e + 8./3.*
         pow(m33,2)*L8r*RN3p*RSp45633e + 1./3.*m11*L5r*RSp45613e - 1./6.
         *m11*L5r*RSp45633e - 5./6.*m11*L5r*RSp45611e + 4./3.*m11*mp2*
         L5r*RN1p*RSp45613e - 4./3.*m11*mp2*L5r*RN1p*RSp45611e - 8./3.*
         m11*m66*L7r*RN1p*RSp453e + 8./3.*m11*m66*L7r*RN1p*RSp451e - 8./
         3.*m11*m66*L6r*RN1p*RSp45613e + 8./3.*m11*m66*L6r*RN1p*
         RSp45611e - 8./3.*m11*m55*L7r*RN1p*RSp463e + 8./3.*m11*m55*L7r
         *RN1p*RSp461e - 8./3.*m11*m55*L6r*RN1p*RSp45613e + 8./3.*m11*
         m55*L6r*RN1p*RSp45611e - 8./3.*m11*m44*L7r*RN1p*RSp563e + 8./3.
         *m11*m44*L7r*RN1p*RSp561e - 8./3.*m11*m44*L6r*RN1p*RSp45613e
          + 8./3.*m11*m44*L6r*RN1p*RSp45611e - 8./3.*pow(m11,2)*L8r*
         RN1p*RSp45613e + 8./3.*pow(m11,2)*L8r*RN1p*RSp45611e );

      F13p6x23L +=  + Bb(m11,mu2) * ( m66*L4r*RD1456ep + 4./9.*pow(
         m66,2)*L8r*pow(RS145pe,2) + 4./9.*pow(m66,2)*L7r*pow(
         RS145pe,2) + 4./9.*pow(m66,2)*L6r*pow(RS145pe,2) + m55*L4r*
         RD1456ep + 8./9.*m55*m66*L7r*RS146pe*RS145pe + 4./9.*m55*m66*
         L6r*pow(RS145pe,2) + 4./9.*m55*m66*L6r*pow(RS146pe,2) + 4./9.*
         pow(m55,2)*L8r*pow(RS146pe,2) + 4./9.*pow(m55,2)*L7r*pow(
         RS146pe,2) + 4./9.*pow(m55,2)*L6r*pow(RS146pe,2) + m44*L4r*
         RD1456ep + 8./9.*m44*m66*L7r*RS156pe*RS145pe + 4./9.*m44*m66*
         L6r*pow(RS145pe,2) + 4./9.*m44*m66*L6r*pow(RS156pe,2) + 8./9.*
         m44*m55*L7r*RS156pe*RS146pe + 4./9.*m44*m55*L6r*pow(RS146pe,2)
          + 4./9.*m44*m55*L6r*pow(RS156pe,2) + 4./9.*pow(m44,2)*L8r*
         pow(RS156pe,2) + 4./9.*pow(m44,2)*L7r*pow(RS156pe,2) + 4./9.*
         pow(m44,2)*L6r*pow(RS156pe,2) - 1./6.*m33*L5r*RD1456ep + 1./2.
         *m11*L5r*RD1456ep + 5./3.*m11*L3r*RD1456ep + 2./3.*m11*L0r*
         RD1456ep - 8./3.*m11*m66*L7r*RS145pe + 8./3.*m11*m66*L6r*
         RS14563pe );

      F13p6x23L +=  + Bb(m11,mu2) * ( 8./3.*m11*m66*L6r*RS1456pee + 8./
         3.*m11*m66*L6r*RS1456ppe - 8./3.*m11*m66*L6r*RS145pe - 8./3.*
         m11*m66*L6r*RS146pe - 8./3.*m11*m66*L6r*RS156pe - 2./9.*m11*
         m66*L5r*pow(RS145pe,2) - 4./3.*m11*m66*L4r*RS14563pe - 4./3.*
         m11*m66*L4r*RS1456pee - 4./3.*m11*m66*L4r*RS1456ppe + 4./3.*
         m11*m66*L4r*RS145pe - 2./9.*m11*m66*L4r*pow(RS145pe,2) + 4./3.
         *m11*m66*L4r*RS146pe - 2./9.*m11*m66*L4r*pow(RS146pe,2) + 4./3.
         *m11*m66*L4r*RS156pe - 2./9.*m11*m66*L4r*pow(RS156pe,2) - 8./3.
         *m11*m55*L7r*RS146pe + 8./3.*m11*m55*L6r*RS14563pe + 8./3.*m11
         *m55*L6r*RS1456pee + 8./3.*m11*m55*L6r*RS1456ppe - 8./3.*m11*
         m55*L6r*RS145pe - 8./3.*m11*m55*L6r*RS146pe - 8./3.*m11*m55*
         L6r*RS156pe - 2./9.*m11*m55*L5r*pow(RS146pe,2) - 4./3.*m11*m55
         *L4r*RS14563pe - 4./3.*m11*m55*L4r*RS1456pee - 4./3.*m11*m55*
         L4r*RS1456ppe + 4./3.*m11*m55*L4r*RS145pe - 2./9.*m11*m55*L4r*
         pow(RS145pe,2) + 4./3.*m11*m55*L4r*RS146pe - 2./9.*m11*m55*L4r
         *pow(RS146pe,2) );

      F13p6x23L +=  + Bb(m11,mu2) * ( 4./3.*m11*m55*L4r*RS156pe - 2./9.
         *m11*m55*L4r*pow(RS156pe,2) - 8./3.*m11*m44*L7r*RS156pe + 8./3.
         *m11*m44*L6r*RS14563pe + 8./3.*m11*m44*L6r*RS1456pee + 8./3.*
         m11*m44*L6r*RS1456ppe - 8./3.*m11*m44*L6r*RS145pe - 8./3.*m11*
         m44*L6r*RS146pe - 8./3.*m11*m44*L6r*RS156pe - 2./9.*m11*m44*
         L5r*pow(RS156pe,2) - 4./3.*m11*m44*L4r*RS14563pe - 4./3.*m11*
         m44*L4r*RS1456pee - 4./3.*m11*m44*L4r*RS1456ppe + 4./3.*m11*
         m44*L4r*RS145pe - 2./9.*m11*m44*L4r*pow(RS145pe,2) + 4./3.*m11
         *m44*L4r*RS146pe - 2./9.*m11*m44*L4r*pow(RS146pe,2) + 4./3.*
         m11*m44*L4r*RS156pe - 2./9.*m11*m44*L4r*pow(RS156pe,2) + 8./3.
         *pow(m11,2)*L8r*RS14563pe + 8./3.*pow(m11,2)*L8r*RS1456pee + 8.
         /3.*pow(m11,2)*L8r*RS1456ppe - 8./3.*pow(m11,2)*L8r*RS145pe - 
         8./3.*pow(m11,2)*L8r*RS146pe - 8./3.*pow(m11,2)*L8r*RS156pe + 
         4*pow(m11,2)*L7r - 4./3.*pow(m11,2)*L5r*RS14563pe - 4./3.*pow(
         m11,2)*L5r*RS1456pee - 4./3.*pow(m11,2)*L5r*RS1456ppe + 4./3.*
         pow(m11,2)*L5r*RS145pe );

      F13p6x23L +=  + Bb(m11,mu2) * ( 4./3.*pow(m11,2)*L5r*RS146pe + 4./
         3.*pow(m11,2)*L5r*RS156pe );

      F13p6x23L +=  + Bb(m14,mu2) * ( 2*m44*m66*L6r - m44*m66*L4r + 2*
         m44*m55*L6r - m44*m55*L4r + pow(m44,2)*L8r + 2*pow(m44,2)*L6r
          - 1./2.*pow(m44,2)*L5r - pow(m44,2)*L4r + 2*m11*m66*L6r - m11
         *m66*L4r + 2*m11*m55*L6r - m11*m55*L4r + 2*m11*m44*L8r + 2*m11
         *m44*L6r - m11*m44*L5r - m11*m44*L4r + pow(m11,2)*L8r - 1./2.*
         pow(m11,2)*L5r );

      F13p6x23L +=  + Bb(m15,mu2) * ( 2*m55*m66*L6r - m55*m66*L4r + 
         pow(m55,2)*L8r + 2*pow(m55,2)*L6r - 1./2.*pow(m55,2)*L5r - 
         pow(m55,2)*L4r + 2*m44*m55*L6r - m44*m55*L4r + 2*m11*m66*L6r
          - m11*m66*L4r + 2*m11*m55*L8r + 2*m11*m55*L6r - m11*m55*L5r
          - m11*m55*L4r + 2*m11*m44*L6r - m11*m44*L4r + pow(m11,2)*L8r
          - 1./2.*pow(m11,2)*L5r );

      F13p6x23L +=  + Bb(m16,mu2) * ( pow(m66,2)*L8r + 2*pow(m66,2)*L6r
          - 1./2.*pow(m66,2)*L5r - pow(m66,2)*L4r + 2*m55*m66*L6r - m55
         *m66*L4r + 2*m44*m66*L6r - m44*m66*L4r + 2*m11*m66*L8r + 2*m11
         *m66*L6r - m11*m66*L5r - m11*m66*L4r + 2*m11*m55*L6r - m11*m55
         *L4r + 2*m11*m44*L6r - m11*m44*L4r + pow(m11,2)*L8r - 1./2.*
         pow(m11,2)*L5r );

      F13p6x23L +=  + Bb(m33,mu2) * ( m66*L4r*RD3456ep + 4./9.*pow(
         m66,2)*L8r*pow(RS345pe,2) + 4./9.*pow(m66,2)*L7r*pow(
         RS345pe,2) + 4./9.*pow(m66,2)*L6r*pow(RS345pe,2) + m55*L4r*
         RD3456ep + 8./9.*m55*m66*L7r*RS346pe*RS345pe + 4./9.*m55*m66*
         L6r*pow(RS345pe,2) + 4./9.*m55*m66*L6r*pow(RS346pe,2) + 4./9.*
         pow(m55,2)*L8r*pow(RS346pe,2) + 4./9.*pow(m55,2)*L7r*pow(
         RS346pe,2) + 4./9.*pow(m55,2)*L6r*pow(RS346pe,2) + m44*L4r*
         RD3456ep + 8./9.*m44*m66*L7r*RS356pe*RS345pe + 4./9.*m44*m66*
         L6r*pow(RS345pe,2) + 4./9.*m44*m66*L6r*pow(RS356pe,2) + 8./9.*
         m44*m55*L7r*RS356pe*RS346pe + 4./9.*m44*m55*L6r*pow(RS346pe,2)
          + 4./9.*m44*m55*L6r*pow(RS356pe,2) + 4./9.*pow(m44,2)*L8r*
         pow(RS356pe,2) + 4./9.*pow(m44,2)*L7r*pow(RS356pe,2) + 4./9.*
         pow(m44,2)*L6r*pow(RS356pe,2) + 1./2.*m33*L5r*RD3456ep + 5./3.
         *m33*L3r*RD3456ep + 2./3.*m33*L0r*RD3456ep - 8./3.*m33*m66*L7r
         *RS345pe + 8./3.*m33*m66*L6r*RS34561pe + 8./3.*m33*m66*L6r*
         RS3456pee );

      F13p6x23L +=  + Bb(m33,mu2) * ( 8./3.*m33*m66*L6r*RS3456ppe - 8./
         3.*m33*m66*L6r*RS345pe - 8./3.*m33*m66*L6r*RS346pe - 8./3.*m33
         *m66*L6r*RS356pe - 2./9.*m33*m66*L5r*pow(RS345pe,2) - 4./3.*
         m33*m66*L4r*RS34561pe - 4./3.*m33*m66*L4r*RS3456pee - 4./3.*
         m33*m66*L4r*RS3456ppe + 4./3.*m33*m66*L4r*RS345pe - 2./9.*m33*
         m66*L4r*pow(RS345pe,2) + 4./3.*m33*m66*L4r*RS346pe - 2./9.*m33
         *m66*L4r*pow(RS346pe,2) + 4./3.*m33*m66*L4r*RS356pe - 2./9.*
         m33*m66*L4r*pow(RS356pe,2) - 8./3.*m33*m55*L7r*RS346pe + 8./3.
         *m33*m55*L6r*RS34561pe + 8./3.*m33*m55*L6r*RS3456pee + 8./3.*
         m33*m55*L6r*RS3456ppe - 8./3.*m33*m55*L6r*RS345pe - 8./3.*m33*
         m55*L6r*RS346pe - 8./3.*m33*m55*L6r*RS356pe - 2./9.*m33*m55*
         L5r*pow(RS346pe,2) - 4./3.*m33*m55*L4r*RS34561pe - 4./3.*m33*
         m55*L4r*RS3456pee - 4./3.*m33*m55*L4r*RS3456ppe + 4./3.*m33*
         m55*L4r*RS345pe - 2./9.*m33*m55*L4r*pow(RS345pe,2) + 4./3.*m33
         *m55*L4r*RS346pe - 2./9.*m33*m55*L4r*pow(RS346pe,2) + 4./3.*
         m33*m55*L4r*RS356pe );

      F13p6x23L +=  + Bb(m33,mu2) * (  - 2./9.*m33*m55*L4r*pow(
         RS356pe,2) - 8./3.*m33*m44*L7r*RS356pe + 8./3.*m33*m44*L6r*
         RS34561pe + 8./3.*m33*m44*L6r*RS3456pee + 8./3.*m33*m44*L6r*
         RS3456ppe - 8./3.*m33*m44*L6r*RS345pe - 8./3.*m33*m44*L6r*
         RS346pe - 8./3.*m33*m44*L6r*RS356pe - 2./9.*m33*m44*L5r*pow(
         RS356pe,2) - 4./3.*m33*m44*L4r*RS34561pe - 4./3.*m33*m44*L4r*
         RS3456pee - 4./3.*m33*m44*L4r*RS3456ppe + 4./3.*m33*m44*L4r*
         RS345pe - 2./9.*m33*m44*L4r*pow(RS345pe,2) + 4./3.*m33*m44*L4r
         *RS346pe - 2./9.*m33*m44*L4r*pow(RS346pe,2) + 4./3.*m33*m44*
         L4r*RS356pe - 2./9.*m33*m44*L4r*pow(RS356pe,2) + 8./3.*pow(
         m33,2)*L8r*RS34561pe + 8./3.*pow(m33,2)*L8r*RS3456pee + 8./3.*
         pow(m33,2)*L8r*RS3456ppe - 8./3.*pow(m33,2)*L8r*RS345pe - 8./3.
         *pow(m33,2)*L8r*RS346pe - 8./3.*pow(m33,2)*L8r*RS356pe + 4*
         pow(m33,2)*L7r - 4./3.*pow(m33,2)*L5r*RS34561pe - 4./3.*pow(
         m33,2)*L5r*RS3456pee - 4./3.*pow(m33,2)*L5r*RS3456ppe + 4./3.*
         pow(m33,2)*L5r*RS345pe );

      F13p6x23L +=  + Bb(m33,mu2) * ( 4./3.*pow(m33,2)*L5r*RS346pe + 4./
         3.*pow(m33,2)*L5r*RS356pe - 1./6.*m11*L5r*RD3456ep );

      F13p6x23L +=  + Bb(m34,mu2) * ( 2*m44*m66*L6r - m44*m66*L4r + 2*
         m44*m55*L6r - m44*m55*L4r + pow(m44,2)*L8r + 2*pow(m44,2)*L6r
          - 1./2.*pow(m44,2)*L5r - pow(m44,2)*L4r + 2*m33*m66*L6r - m33
         *m66*L4r + 2*m33*m55*L6r - m33*m55*L4r + 2*m33*m44*L8r + 2*m33
         *m44*L6r - m33*m44*L5r - m33*m44*L4r + pow(m33,2)*L8r - 1./2.*
         pow(m33,2)*L5r );

      F13p6x23L +=  + Bb(m35,mu2) * ( 2*m55*m66*L6r - m55*m66*L4r + 
         pow(m55,2)*L8r + 2*pow(m55,2)*L6r - 1./2.*pow(m55,2)*L5r - 
         pow(m55,2)*L4r + 2*m44*m55*L6r - m44*m55*L4r + 2*m33*m66*L6r
          - m33*m66*L4r + 2*m33*m55*L8r + 2*m33*m55*L6r - m33*m55*L5r
          - m33*m55*L4r + 2*m33*m44*L6r - m33*m44*L4r + pow(m33,2)*L8r
          - 1./2.*pow(m33,2)*L5r );

      F13p6x23L +=  + Bb(m36,mu2) * ( pow(m66,2)*L8r + 2*pow(m66,2)*L6r
          - 1./2.*pow(m66,2)*L5r - pow(m66,2)*L4r + 2*m55*m66*L6r - m55
         *m66*L4r + 2*m44*m66*L6r - m44*m66*L4r + 2*m33*m66*L8r + 2*m33
         *m66*L6r - m33*m66*L5r - m33*m66*L4r + 2*m33*m55*L6r - m33*m55
         *L4r + 2*m33*m44*L6r - m33*m44*L4r + pow(m33,2)*L8r - 1./2.*
         pow(m33,2)*L5r );

      F13p6x23L +=  + Bb(me2,mu2) * (  - 2./9.*m66*me2*L5r*pow(
         RSe453p,2) + 4./9.*m66*me2*L5r*RSe451p*RSe453p - 2./9.*m66*me2
         *L5r*pow(RSe451p,2) - 2./9.*m66*me2*L4r*pow(RSe563p,2) - 2./9.
         *m66*me2*L4r*pow(RSe463p,2) - 2./9.*m66*me2*L4r*pow(RSe453p,2)
          + 4./9.*m66*me2*L4r*RSe561p*RSe563p - 2./9.*m66*me2*L4r*pow(
         RSe561p,2) + 4./9.*m66*me2*L4r*RSe461p*RSe463p - 2./9.*m66*me2
         *L4r*pow(RSe461p,2) + 4./9.*m66*me2*L4r*RSe451p*RSe453p - 2./9.
         *m66*me2*L4r*pow(RSe451p,2) + 4./9.*pow(m66,2)*L8r*pow(
         RSe453p,2) - 8./9.*pow(m66,2)*L8r*RSe451p*RSe453p + 4./9.*pow(
         m66,2)*L8r*pow(RSe451p,2) + 4./9.*pow(m66,2)*L7r*pow(
         RSe453p,2) - 8./9.*pow(m66,2)*L7r*RSe451p*RSe453p + 4./9.*pow(
         m66,2)*L7r*pow(RSe451p,2) + 4./9.*pow(m66,2)*L6r*pow(
         RSe453p,2) - 8./9.*pow(m66,2)*L6r*RSe451p*RSe453p + 4./9.*pow(
         m66,2)*L6r*pow(RSe451p,2) - 2./9.*m55*me2*L5r*pow(RSe463p,2)
          + 4./9.*m55*me2*L5r*RSe461p*RSe463p - 2./9.*m55*me2*L5r*pow(
         RSe461p,2) );

      F13p6x23L +=  + Bb(me2,mu2) * (  - 2./9.*m55*me2*L4r*pow(
         RSe563p,2) - 2./9.*m55*me2*L4r*pow(RSe463p,2) - 2./9.*m55*me2*
         L4r*pow(RSe453p,2) + 4./9.*m55*me2*L4r*RSe561p*RSe563p - 2./9.
         *m55*me2*L4r*pow(RSe561p,2) + 4./9.*m55*me2*L4r*RSe461p*
         RSe463p - 2./9.*m55*me2*L4r*pow(RSe461p,2) + 4./9.*m55*me2*L4r
         *RSe451p*RSe453p - 2./9.*m55*me2*L4r*pow(RSe451p,2) + 8./9.*
         m55*m66*L7r*RSe453p*RSe463p - 8./9.*m55*m66*L7r*RSe461p*
         RSe453p - 8./9.*m55*m66*L7r*RSe451p*RSe463p + 8./9.*m55*m66*
         L7r*RSe451p*RSe461p + 4./9.*m55*m66*L6r*pow(RSe463p,2) + 4./9.
         *m55*m66*L6r*pow(RSe453p,2) - 8./9.*m55*m66*L6r*RSe461p*
         RSe463p + 4./9.*m55*m66*L6r*pow(RSe461p,2) - 8./9.*m55*m66*L6r
         *RSe451p*RSe453p + 4./9.*m55*m66*L6r*pow(RSe451p,2) + 4./9.*
         pow(m55,2)*L8r*pow(RSe463p,2) - 8./9.*pow(m55,2)*L8r*RSe461p*
         RSe463p + 4./9.*pow(m55,2)*L8r*pow(RSe461p,2) + 4./9.*pow(
         m55,2)*L7r*pow(RSe463p,2) - 8./9.*pow(m55,2)*L7r*RSe461p*
         RSe463p );

      F13p6x23L +=  + Bb(me2,mu2) * ( 4./9.*pow(m55,2)*L7r*pow(
         RSe461p,2) + 4./9.*pow(m55,2)*L6r*pow(RSe463p,2) - 8./9.*pow(
         m55,2)*L6r*RSe461p*RSe463p + 4./9.*pow(m55,2)*L6r*pow(
         RSe461p,2) - 2./9.*m44*me2*L5r*pow(RSe563p,2) + 4./9.*m44*me2*
         L5r*RSe561p*RSe563p - 2./9.*m44*me2*L5r*pow(RSe561p,2) - 2./9.
         *m44*me2*L4r*pow(RSe563p,2) - 2./9.*m44*me2*L4r*pow(RSe463p,2)
          - 2./9.*m44*me2*L4r*pow(RSe453p,2) + 4./9.*m44*me2*L4r*
         RSe561p*RSe563p - 2./9.*m44*me2*L4r*pow(RSe561p,2) + 4./9.*m44
         *me2*L4r*RSe461p*RSe463p - 2./9.*m44*me2*L4r*pow(RSe461p,2) + 
         4./9.*m44*me2*L4r*RSe451p*RSe453p - 2./9.*m44*me2*L4r*pow(
         RSe451p,2) + 8./9.*m44*m66*L7r*RSe453p*RSe563p - 8./9.*m44*m66
         *L7r*RSe561p*RSe453p - 8./9.*m44*m66*L7r*RSe451p*RSe563p + 8./
         9.*m44*m66*L7r*RSe451p*RSe561p + 4./9.*m44*m66*L6r*pow(
         RSe563p,2) + 4./9.*m44*m66*L6r*pow(RSe453p,2) - 8./9.*m44*m66*
         L6r*RSe561p*RSe563p + 4./9.*m44*m66*L6r*pow(RSe561p,2) - 8./9.
         *m44*m66*L6r*RSe451p*RSe453p );

      F13p6x23L +=  + Bb(me2,mu2) * ( 4./9.*m44*m66*L6r*pow(RSe451p,2)
          + 8./9.*m44*m55*L7r*RSe463p*RSe563p - 8./9.*m44*m55*L7r*
         RSe561p*RSe463p - 8./9.*m44*m55*L7r*RSe461p*RSe563p + 8./9.*
         m44*m55*L7r*RSe461p*RSe561p + 4./9.*m44*m55*L6r*pow(RSe563p,2)
          + 4./9.*m44*m55*L6r*pow(RSe463p,2) - 8./9.*m44*m55*L6r*
         RSe561p*RSe563p + 4./9.*m44*m55*L6r*pow(RSe561p,2) - 8./9.*m44
         *m55*L6r*RSe461p*RSe463p + 4./9.*m44*m55*L6r*pow(RSe461p,2) + 
         4./9.*pow(m44,2)*L8r*pow(RSe563p,2) - 8./9.*pow(m44,2)*L8r*
         RSe561p*RSe563p + 4./9.*pow(m44,2)*L8r*pow(RSe561p,2) + 4./9.*
         pow(m44,2)*L7r*pow(RSe563p,2) - 8./9.*pow(m44,2)*L7r*RSe561p*
         RSe563p + 4./9.*pow(m44,2)*L7r*pow(RSe561p,2) + 4./9.*pow(
         m44,2)*L6r*pow(RSe563p,2) - 8./9.*pow(m44,2)*L6r*RSe561p*
         RSe563p + 4./9.*pow(m44,2)*L6r*pow(RSe561p,2) );

      F13p6x23L +=  + Bb(mp2,mu2) * (  - 2./9.*m66*mp2*L5r*pow(
         RSp453e,2) + 4./9.*m66*mp2*L5r*RSp451e*RSp453e - 2./9.*m66*mp2
         *L5r*pow(RSp451e,2) - 2./9.*m66*mp2*L4r*pow(RSp563e,2) - 2./9.
         *m66*mp2*L4r*pow(RSp463e,2) - 2./9.*m66*mp2*L4r*pow(RSp453e,2)
          + 4./9.*m66*mp2*L4r*RSp561e*RSp563e - 2./9.*m66*mp2*L4r*pow(
         RSp561e,2) + 4./9.*m66*mp2*L4r*RSp461e*RSp463e - 2./9.*m66*mp2
         *L4r*pow(RSp461e,2) + 4./9.*m66*mp2*L4r*RSp451e*RSp453e - 2./9.
         *m66*mp2*L4r*pow(RSp451e,2) + 4./9.*pow(m66,2)*L8r*pow(
         RSp453e,2) - 8./9.*pow(m66,2)*L8r*RSp451e*RSp453e + 4./9.*pow(
         m66,2)*L8r*pow(RSp451e,2) + 4./9.*pow(m66,2)*L7r*pow(
         RSp453e,2) - 8./9.*pow(m66,2)*L7r*RSp451e*RSp453e + 4./9.*pow(
         m66,2)*L7r*pow(RSp451e,2) + 4./9.*pow(m66,2)*L6r*pow(
         RSp453e,2) - 8./9.*pow(m66,2)*L6r*RSp451e*RSp453e + 4./9.*pow(
         m66,2)*L6r*pow(RSp451e,2) - 2./9.*m55*mp2*L5r*pow(RSp463e,2)
          + 4./9.*m55*mp2*L5r*RSp461e*RSp463e - 2./9.*m55*mp2*L5r*pow(
         RSp461e,2) );

      F13p6x23L +=  + Bb(mp2,mu2) * (  - 2./9.*m55*mp2*L4r*pow(
         RSp563e,2) - 2./9.*m55*mp2*L4r*pow(RSp463e,2) - 2./9.*m55*mp2*
         L4r*pow(RSp453e,2) + 4./9.*m55*mp2*L4r*RSp561e*RSp563e - 2./9.
         *m55*mp2*L4r*pow(RSp561e,2) + 4./9.*m55*mp2*L4r*RSp461e*
         RSp463e - 2./9.*m55*mp2*L4r*pow(RSp461e,2) + 4./9.*m55*mp2*L4r
         *RSp451e*RSp453e - 2./9.*m55*mp2*L4r*pow(RSp451e,2) + 8./9.*
         m55*m66*L7r*RSp453e*RSp463e - 8./9.*m55*m66*L7r*RSp461e*
         RSp453e - 8./9.*m55*m66*L7r*RSp451e*RSp463e + 8./9.*m55*m66*
         L7r*RSp451e*RSp461e + 4./9.*m55*m66*L6r*pow(RSp463e,2) + 4./9.
         *m55*m66*L6r*pow(RSp453e,2) - 8./9.*m55*m66*L6r*RSp461e*
         RSp463e + 4./9.*m55*m66*L6r*pow(RSp461e,2) - 8./9.*m55*m66*L6r
         *RSp451e*RSp453e + 4./9.*m55*m66*L6r*pow(RSp451e,2) + 4./9.*
         pow(m55,2)*L8r*pow(RSp463e,2) - 8./9.*pow(m55,2)*L8r*RSp461e*
         RSp463e + 4./9.*pow(m55,2)*L8r*pow(RSp461e,2) + 4./9.*pow(
         m55,2)*L7r*pow(RSp463e,2) - 8./9.*pow(m55,2)*L7r*RSp461e*
         RSp463e );

      F13p6x23L +=  + Bb(mp2,mu2) * ( 4./9.*pow(m55,2)*L7r*pow(
         RSp461e,2) + 4./9.*pow(m55,2)*L6r*pow(RSp463e,2) - 8./9.*pow(
         m55,2)*L6r*RSp461e*RSp463e + 4./9.*pow(m55,2)*L6r*pow(
         RSp461e,2) - 2./9.*m44*mp2*L5r*pow(RSp563e,2) + 4./9.*m44*mp2*
         L5r*RSp561e*RSp563e - 2./9.*m44*mp2*L5r*pow(RSp561e,2) - 2./9.
         *m44*mp2*L4r*pow(RSp563e,2) - 2./9.*m44*mp2*L4r*pow(RSp463e,2)
          - 2./9.*m44*mp2*L4r*pow(RSp453e,2) + 4./9.*m44*mp2*L4r*
         RSp561e*RSp563e - 2./9.*m44*mp2*L4r*pow(RSp561e,2) + 4./9.*m44
         *mp2*L4r*RSp461e*RSp463e - 2./9.*m44*mp2*L4r*pow(RSp461e,2) + 
         4./9.*m44*mp2*L4r*RSp451e*RSp453e - 2./9.*m44*mp2*L4r*pow(
         RSp451e,2) + 8./9.*m44*m66*L7r*RSp453e*RSp563e - 8./9.*m44*m66
         *L7r*RSp561e*RSp453e - 8./9.*m44*m66*L7r*RSp451e*RSp563e + 8./
         9.*m44*m66*L7r*RSp451e*RSp561e + 4./9.*m44*m66*L6r*pow(
         RSp563e,2) + 4./9.*m44*m66*L6r*pow(RSp453e,2) - 8./9.*m44*m66*
         L6r*RSp561e*RSp563e + 4./9.*m44*m66*L6r*pow(RSp561e,2) - 8./9.
         *m44*m66*L6r*RSp451e*RSp453e );

      F13p6x23L +=  + Bb(mp2,mu2) * ( 4./9.*m44*m66*L6r*pow(RSp451e,2)
          + 8./9.*m44*m55*L7r*RSp463e*RSp563e - 8./9.*m44*m55*L7r*
         RSp561e*RSp463e - 8./9.*m44*m55*L7r*RSp461e*RSp563e + 8./9.*
         m44*m55*L7r*RSp461e*RSp561e + 4./9.*m44*m55*L6r*pow(RSp563e,2)
          + 4./9.*m44*m55*L6r*pow(RSp463e,2) - 8./9.*m44*m55*L6r*
         RSp561e*RSp563e + 4./9.*m44*m55*L6r*pow(RSp561e,2) - 8./9.*m44
         *m55*L6r*RSp461e*RSp463e + 4./9.*m44*m55*L6r*pow(RSp461e,2) + 
         4./9.*pow(m44,2)*L8r*pow(RSp563e,2) - 8./9.*pow(m44,2)*L8r*
         RSp561e*RSp563e + 4./9.*pow(m44,2)*L8r*pow(RSp561e,2) + 4./9.*
         pow(m44,2)*L7r*pow(RSp563e,2) - 8./9.*pow(m44,2)*L7r*RSp561e*
         RSp563e + 4./9.*pow(m44,2)*L7r*pow(RSp561e,2) + 4./9.*pow(
         m44,2)*L6r*pow(RSp563e,2) - 8./9.*pow(m44,2)*L6r*RSp561e*
         RSp563e + 4./9.*pow(m44,2)*L6r*pow(RSp561e,2) );

      F13p6x23L +=  + Cb(m11,mu2) * (  - 8./3.*m11*m66*L6r*RD1456ep + 4.
         /3.*m11*m66*L4r*RD1456ep - 8./3.*m11*m55*L6r*RD1456ep + 4./3.*
         m11*m55*L4r*RD1456ep - 8./3.*m11*m44*L6r*RD1456ep + 4./3.*m11*
         m44*L4r*RD1456ep - 8./3.*pow(m11,2)*L8r*RD1456ep + 4./3.*pow(
         m11,2)*L5r*RD1456ep );

      F13p6x23L +=  + Cb(m33,mu2) * (  - 8./3.*m33*m66*L6r*RD3456ep + 4.
         /3.*m33*m66*L4r*RD3456ep - 8./3.*m33*m55*L6r*RD3456ep + 4./3.*
         m33*m55*L4r*RD3456ep - 8./3.*m33*m44*L6r*RD3456ep + 4./3.*m33*
         m44*L4r*RD3456ep - 8./3.*pow(m33,2)*L8r*RD3456ep + 4./3.*pow(
         m33,2)*L5r*RD3456ep );

      F13p6x23L +=  - 8*pow(m66,2)*pow(L4r,2) - 16*m55*m66*pow(L4r,2)
          - 8*pow(m55,2)*pow(L4r,2) - 16*m44*m66*pow(L4r,2) - 16*m44*
         m55*pow(L4r,2) - 8*pow(m44,2)*pow(L4r,2) - 8*m33*m66*L4r*L5r
          - 8*m33*m55*L4r*L5r - 8*m33*m44*L4r*L5r - 2*pow(m33,2)*pow(
         L5r,2) - 8*m11*m66*L4r*L5r - 8*m11*m55*L4r*L5r - 8*m11*m44*L4r
         *L5r - 4*m11*m33*pow(L5r,2) - 2*pow(m11,2)*pow(L5r,2);

  return F13p6x23L/pow(f0,4);
}

double fv2s3nf3p6K(const quarkmassnf mass, const Ki Kiin){
  double KK[116],mu;
  Kiin.out(KK,mu);
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if (mu0 != mu){
    cout << "ERROR: differing mu's in fv2s3nf3p6L\n";}
  if ( (nq != 5) || (B0mq.size() != 5)){
    cout << "ERROR: fv2s3nf3p6K probably called with wrong number"
	 << "of quark masses\n";}
  // five masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m33 = 2.*B0mq[1];
  double m44 = 2.*B0mq[2];
  double m55 = 2.*B0mq[3];
  double m66 = 2.*B0mq[4];

  double F13p6x23K =
       + KK[19] * ( 4*pow(m33,2) + 4*pow(m11,2) );

      F13p6x23K +=  + KK[20] * ( 4*m33*m66 + 4*m33*m55 + 4*m33*m44 + 4*
         m11*m66 + 4*m11*m55 + 4*m11*m44 );

      F13p6x23K +=  + KK[21] * ( 8*pow(m66,2) + 8*pow(m55,2) + 8*pow(
         m44,2) );

      F13p6x23K +=  + KK[22] * ( 8*pow(m66,2) + 16*m55*m66 + 8*pow(
         m55,2) + 16*m44*m66 + 16*m44*m55 + 8*pow(m44,2) );

      F13p6x23K +=  + KK[23] * ( 8*m11*m33 );

  return F13p6x23K/pow(f0,4);
}

double fv2s3nf3p6R(const quarkmassnf mass){
  vector<double> B0mq;
  double f0,mu0;
  int nq;
  mass.out(B0mq,f0,mu0,nq);
  if ( (nq != 5) || (B0mq.size() != 5)){
    cout << "ERROR: fv2s3nf3p6R probably called with wrong number"
	 << "of quark masses\n";}
  // two masses: valence and sea quark mass
  double m11 = 2.*B0mq[0];
  double m33 = 2.*B0mq[1];
  double m44 = 2.*B0mq[2];
  double m55 = 2.*B0mq[3];
  double m66 = 2.*B0mq[4];
  double m13 = (m11+m33)/2;
  double m14 = (m11+m44)/2;
  double m15 = (m11+m55)/2;
  double m16 = (m11+m66)/2;
  double m34 = (m33+m44)/2;
  double m35 = (m33+m55)/2;
  double m36 = (m33+m66)/2;
  double m45 = (m44+m55)/2;
  double m46 = (m44+m66)/2;
  double m56 = (m55+m66)/2;

  double xb1 = (m44+m55+m66)/3.;
  double msqr = sqrt(xb1*xb1-(m44*m55+m44*m66+m55*m66)/3.);
  double mp2 = xb1-msqr; 
  double me2 = xb1+msqr;
  double mu2 = mu0*mu0;

  double RN13 = RN(m11,m33);
  double RN1p = RN(m11,mp2);
  double RN1e = RN(m11,me2);
  double RN3p = RN(m33,mp2);
  double RN3e = RN(m33,me2);
  double RNep = RN(me2,mp2);

  double RSp456e = RS(mp2,m44,m55,m66,me2);
  double RSp465e = RS(mp2,m44,m66,m55,me2);
  double RSp564e = RS(mp2,m55,m66,m44,me2);
  double RSe456p = RS(me2,m44,m55,m66,mp2);
  double RSe465p = RS(me2,m44,m66,m55,mp2);
  double RSe564p = RS(me2,m55,m66,m44,mp2);
  double RSp451e = RS(mp2,m44,m55,m11,me2);
  double RSp461e = RS(mp2,m44,m66,m11,me2);
  double RSp561e = RS(mp2,m55,m66,m11,me2);
  double RSe451p = RS(me2,m44,m55,m11,mp2);
  double RSe461p = RS(me2,m44,m66,m11,mp2);
  double RSe561p = RS(me2,m55,m66,m11,mp2);
  double RSp453e = RS(mp2,m44,m55,m33,me2);
  double RSp463e = RS(mp2,m44,m66,m33,me2);
  double RSp563e = RS(mp2,m55,m66,m33,me2);
  double RSe453p = RS(me2,m44,m55,m33,mp2);
  double RSe463p = RS(me2,m44,m66,m33,mp2);
  double RSe563p = RS(me2,m55,m66,m33,mp2);
  double RS456pe = RS(m44,m55,m66,mp2,me2);
  double RS546pe = RS(m55,m44,m66,mp2,me2);
  double RS645pe = RS(m66,m44,m55,mp2,me2);
  double RS156pe = RS(m11,m55,m66,mp2,me2);
  double RS146pe = RS(m11,m44,m66,mp2,me2);
  double RS145pe = RS(m11,m44,m55,mp2,me2);
  double RS356pe = RS(m33,m55,m66,mp2,me2);
  double RS346pe = RS(m33,m44,m66,mp2,me2);
  double RS345pe = RS(m33,m44,m55,mp2,me2);

  double RD1456ep = RD(m11,m44,m55,m66,me2,mp2);
  double RD3456ep = RD(m33,m44,m55,m66,me2,mp2);
  //  double RC1456pe = RC(m11,m44,m55,m66,mp2,me2);  
  double RSe45611p = RS(me2,m44,m55,m66,m11,m11,mp2);
  double RSp45611e = RS(mp2,m44,m55,m66,m11,m11,me2);
  double RS1456ppe = RS(m11,m44,m55,m66,mp2,mp2,me2);
  double RS1456pee = RS(m11,m44,m55,m66,mp2,me2,me2);
  double RSe45633p = RS(me2,m44,m55,m66,m33,m33,mp2);
  double RSp45633e = RS(mp2,m44,m55,m66,m33,m33,me2);
  double RS3456ppe = RS(m33,m44,m55,m66,mp2,mp2,me2);
  double RS3456pee = RS(m33,m44,m55,m66,mp2,me2,me2);
  double RSe45613p = RS(me2,m44,m55,m66,m11,m33,mp2);
  double RSp45613e = RS(mp2,m44,m55,m66,m11,m33,me2);
  double RS14563pe = RS(m11,m44,m55,m66,m33,mp2,me2);
  double RS34561pe = RS(m33,m44,m55,m66,m11,mp2,me2);

  double F13p6x23R =
       + pi162 * (  - 1./576.*pow(RD3456ep,2) - 1./288.*RD1456ep*
         RD3456ep - 1./576.*pow(RD1456ep,2) + 1./144.*mp2*RD3456ep*
         RSp45613e - 1./288.*mp2*RSp45633e*RD3456ep + 1./144.*mp2*
         RD1456ep*RSp45613e - 1./288.*mp2*RD1456ep*RSp45633e - 1./288.*
         mp2*RSp45611e*RD3456ep - 1./288.*mp2*RSp45611e*RD1456ep + 7./
         192.*pow(mp2,2)*RSp45613e - 1./144.*pow(mp2,2)*pow(
         RSp45613e,2) - 1./96.*pow(mp2,2)*RSp45633e + 1./144.*pow(
         mp2,2)*RSp45633e*RSp45613e - 1./576.*pow(mp2,2)*pow(
         RSp45633e,2) - 1./64.*pow(mp2,2)*RSp563e - 1./64.*pow(mp2,2)*
         RSp463e - 1./64.*pow(mp2,2)*RSp453e - 1./96.*pow(mp2,2)*
         RSp45611e + 1./144.*pow(mp2,2)*RSp45611e*RSp45613e - 1./288.*
         pow(mp2,2)*RSp45611e*RSp45633e - 1./576.*pow(mp2,2)*pow(
         RSp45611e,2) - 1./64.*pow(mp2,2)*RSp561e - 1./64.*pow(mp2,2)*
         RSp461e - 1./64.*pow(mp2,2)*RSp451e + 1./192.*pow(mp2,2)*
         RSp564e + 1./192.*pow(mp2,2)*RSp465e + 1./192.*pow(mp2,2)*
         RSp456e );

      F13p6x23R +=  + pi162 * ( 1./144.*me2*RD3456ep*RSe45613p - 1./288.
         *me2*RSe45633p*RD3456ep + 1./144.*me2*RD1456ep*RSe45613p - 1./
         288.*me2*RD1456ep*RSe45633p - 1./288.*me2*RSe45611p*RD3456ep
          - 1./288.*me2*RSe45611p*RD1456ep - 1./72.*me2*mp2*RSp45613e*
         RSe45613p + 1./144.*me2*mp2*RSe45633p*RSp45613e + 1./144.*me2*
         mp2*RSp45633e*RSe45613p - 1./288.*me2*mp2*RSp45633e*RSe45633p
          + 1./144.*me2*mp2*RSe45611p*RSp45613e - 1./288.*me2*mp2*
         RSe45611p*RSp45633e + 1./144.*me2*mp2*RSp45611e*RSe45613p - 1./
         288.*me2*mp2*RSp45611e*RSe45633p - 1./288.*me2*mp2*RSp45611e*
         RSe45611p + 7./192.*pow(me2,2)*RSe45613p - 1./144.*pow(me2,2)*
         pow(RSe45613p,2) - 1./96.*pow(me2,2)*RSe45633p + 1./144.*pow(
         me2,2)*RSe45633p*RSe45613p - 1./576.*pow(me2,2)*pow(
         RSe45633p,2) - 1./64.*pow(me2,2)*RSe563p - 1./64.*pow(me2,2)*
         RSe463p - 1./64.*pow(me2,2)*RSe453p - 1./96.*pow(me2,2)*
         RSe45611p + 1./144.*pow(me2,2)*RSe45611p*RSe45613p - 1./288.*
         pow(me2,2)*RSe45611p*RSe45633p );

      F13p6x23R +=  + pi162 * (  - 1./576.*pow(me2,2)*pow(RSe45611p,2)
          - 1./64.*pow(me2,2)*RSe561p - 1./64.*pow(me2,2)*RSe461p - 1./
         64.*pow(me2,2)*RSe451p + 1./192.*pow(me2,2)*RSe564p + 1./192.*
         pow(me2,2)*RSe465p + 1./192.*pow(me2,2)*RSe456p + 5./384.*m66*
         RD3456ep + 5./384.*m66*RD1456ep - 1./96.*m66*mp2*RSp45613e + 5.
         /384.*m66*mp2*RSp45633e - 1./64.*m66*mp2*RSp453e + 5./384.*m66
         *mp2*RSp45611e - 1./64.*m66*mp2*RSp451e + 1./64.*m66*mp2*
         RSp456e - 1./96.*m66*me2*RSe45613p + 5./384.*m66*me2*RSe45633p
          - 1./64.*m66*me2*RSe453p + 5./384.*m66*me2*RSe45611p - 1./64.
         *m66*me2*RSe451p + 1./64.*m66*me2*RSe456p - 15./128.*pow(
         m66,2) - 5./384.*pow(m66,2)*RS34561pe - 5./384.*pow(m66,2)*
         RS14563pe - 5./384.*pow(m66,2)*RSe45613p - 5./384.*pow(m66,2)*
         RSp45613e - 1./128.*pow(m66,2)*RS3456pee - 1./128.*pow(m66,2)*
         RS3456ppe + 1./128.*pow(m66,2)*RSe45633p + 1./128.*pow(m66,2)*
         RSp45633e + 1./192.*pow(m66,2)*RS345pe + 1./128.*pow(m66,2)*
         RS346pe );

      F13p6x23R +=  + pi162 * ( 1./128.*pow(m66,2)*RS356pe - 1./384.*
         pow(m66,2)*RSe453p - 1./384.*pow(m66,2)*RSp453e - 1./128.*pow(
         m66,2)*RS1456pee - 1./128.*pow(m66,2)*RS1456ppe + 1./128.*pow(
         m66,2)*RSe45611p + 1./128.*pow(m66,2)*RSp45611e + 1./192.*pow(
         m66,2)*RS145pe + 1./128.*pow(m66,2)*RS146pe + 1./128.*pow(
         m66,2)*RS156pe + 3./128.*pow(m66,2)*RS645pe - 1./384.*pow(
         m66,2)*RSe451p - 1./384.*pow(m66,2)*RSp451e + 1./384.*pow(
         m66,2)*RSe456p + 1./384.*pow(m66,2)*RSp456e + 5./384.*m55*
         RD3456ep + 5./384.*m55*RD1456ep - 1./96.*m55*mp2*RSp45613e + 5.
         /384.*m55*mp2*RSp45633e - 1./64.*m55*mp2*RSp463e + 5./384.*m55
         *mp2*RSp45611e - 1./64.*m55*mp2*RSp461e + 1./64.*m55*mp2*
         RSp465e - 1./96.*m55*me2*RSe45613p + 5./384.*m55*me2*RSe45633p
          - 1./64.*m55*me2*RSe463p + 5./384.*m55*me2*RSe45611p - 1./64.
         *m55*me2*RSe461p + 1./64.*m55*me2*RSe465p - 5./64.*m55*m66 - 
         15./128.*pow(m55,2) - 5./384.*pow(m55,2)*RS34561pe - 5./384.*
         pow(m55,2)*RS14563pe );

      F13p6x23R +=  + pi162 * (  - 5./384.*pow(m55,2)*RSe45613p - 5./
         384.*pow(m55,2)*RSp45613e - 1./128.*pow(m55,2)*RS3456pee - 1./
         128.*pow(m55,2)*RS3456ppe + 1./128.*pow(m55,2)*RSe45633p + 1./
         128.*pow(m55,2)*RSp45633e + 1./128.*pow(m55,2)*RS345pe + 1./
         192.*pow(m55,2)*RS346pe + 1./128.*pow(m55,2)*RS356pe - 1./384.
         *pow(m55,2)*RSe463p - 1./384.*pow(m55,2)*RSp463e - 1./128.*
         pow(m55,2)*RS1456pee - 1./128.*pow(m55,2)*RS1456ppe + 1./128.*
         pow(m55,2)*RSe45611p + 1./128.*pow(m55,2)*RSp45611e + 1./128.*
         pow(m55,2)*RS145pe + 1./192.*pow(m55,2)*RS146pe + 1./128.*pow(
         m55,2)*RS156pe + 3./128.*pow(m55,2)*RS546pe - 1./384.*pow(
         m55,2)*RSe461p - 1./384.*pow(m55,2)*RSp461e + 1./384.*pow(
         m55,2)*RSe465p + 1./384.*pow(m55,2)*RSp465e + 5./384.*m44*
         RD3456ep + 5./384.*m44*RD1456ep - 1./96.*m44*mp2*RSp45613e + 5.
         /384.*m44*mp2*RSp45633e - 1./64.*m44*mp2*RSp563e + 5./384.*m44
         *mp2*RSp45611e - 1./64.*m44*mp2*RSp561e + 1./64.*m44*mp2*
         RSp564e );

      F13p6x23R +=  + pi162 * (  - 1./96.*m44*me2*RSe45613p + 5./384.*
         m44*me2*RSe45633p - 1./64.*m44*me2*RSe563p + 5./384.*m44*me2*
         RSe45611p - 1./64.*m44*me2*RSe561p + 1./64.*m44*me2*RSe564p - 
         5./64.*m44*m66 - 5./64.*m44*m55 - 15./128.*pow(m44,2) - 5./384.
         *pow(m44,2)*RS34561pe - 5./384.*pow(m44,2)*RS14563pe - 5./384.
         *pow(m44,2)*RSe45613p - 5./384.*pow(m44,2)*RSp45613e - 1./128.
         *pow(m44,2)*RS3456pee - 1./128.*pow(m44,2)*RS3456ppe + 1./128.
         *pow(m44,2)*RSe45633p + 1./128.*pow(m44,2)*RSp45633e + 1./128.
         *pow(m44,2)*RS345pe + 1./128.*pow(m44,2)*RS346pe + 1./192.*
         pow(m44,2)*RS356pe - 1./384.*pow(m44,2)*RSe563p - 1./384.*pow(
         m44,2)*RSp563e - 1./128.*pow(m44,2)*RS1456pee - 1./128.*pow(
         m44,2)*RS1456ppe + 1./128.*pow(m44,2)*RSe45611p + 1./128.*pow(
         m44,2)*RSp45611e + 1./128.*pow(m44,2)*RS145pe + 1./128.*pow(
         m44,2)*RS146pe + 1./192.*pow(m44,2)*RS156pe + 3./128.*pow(
         m44,2)*RS456pe - 1./384.*pow(m44,2)*RSe561p - 1./384.*pow(
         m44,2)*RSp561e );

      F13p6x23R +=  + pi162 * ( 1./384.*pow(m44,2)*RSe564p + 1./384.*
         pow(m44,2)*RSp564e + 1./384.*m33*RD3456ep + 7./576.*m33*
         RD3456ep*RS34561pe + 1./192.*m33*RD3456ep*RS14563pe + 1./192.*
         m33*RD3456ep*RSe45613p + 1./192.*m33*RD3456ep*RSp45613e + 7./
         1152.*m33*RS3456pee*RD3456ep + 7./1152.*m33*RS3456ppe*RD3456ep
          - 1./384.*m33*RSe45633p*RD3456ep - 1./384.*m33*RSp45633e*
         RD3456ep - 7./1152.*m33*RS345pe*RD3456ep - 7./1152.*m33*
         RS346pe*RD3456ep - 7./1152.*m33*RS356pe*RD3456ep + 1./64.*m33*
         RD1456ep + 7./576.*m33*RD1456ep*RS34561pe + 1./192.*m33*
         RD1456ep*RS14563pe + 1./192.*m33*RD1456ep*RSe45613p + 1./192.*
         m33*RD1456ep*RSp45613e + 7./1152.*m33*RD1456ep*RS3456pee + 7./
         1152.*m33*RD1456ep*RS3456ppe - 1./384.*m33*RD1456ep*RSe45633p
          - 1./384.*m33*RD1456ep*RSp45633e - 7./1152.*m33*RD1456ep*
         RS345pe - 7./1152.*m33*RD1456ep*RS346pe - 7./1152.*m33*
         RD1456ep*RS356pe + 1./384.*m33*RS1456pee*RD3456ep + 1./384.*
         m33*RS1456pee*RD1456ep );

      F13p6x23R +=  + pi162 * ( 1./384.*m33*RS1456ppe*RD3456ep + 1./384.
         *m33*RS1456ppe*RD1456ep - 1./384.*m33*RSe45611p*RD3456ep - 1./
         384.*m33*RSe45611p*RD1456ep - 1./384.*m33*RSp45611e*RD3456ep
          - 1./384.*m33*RSp45611e*RD1456ep - 1./384.*m33*RS145pe*
         RD3456ep - 1./384.*m33*RS145pe*RD1456ep - 1./384.*m33*RS146pe*
         RD3456ep - 1./384.*m33*RS146pe*RD1456ep - 1./384.*m33*RS156pe*
         RD3456ep - 1./384.*m33*RS156pe*RD1456ep - 1./64.*m33*mp2*
         RSp45613e - 7./288.*m33*mp2*RSp45613e*RS34561pe - 1./96.*m33*
         mp2*RSp45613e*RS14563pe - 1./96.*m33*mp2*RSp45613e*RSe45613p
          - 1./96.*m33*mp2*pow(RSp45613e,2) - 7./576.*m33*mp2*RS3456pee
         *RSp45613e - 7./576.*m33*mp2*RS3456ppe*RSp45613e + 1./192.*m33
         *mp2*RSe45633p*RSp45613e + 3./128.*m33*mp2*RSp45633e + 7./576.
         *m33*mp2*RSp45633e*RS34561pe + 1./192.*m33*mp2*RSp45633e*
         RS14563pe + 1./192.*m33*mp2*RSp45633e*RSe45613p + 1./96.*m33*
         mp2*RSp45633e*RSp45613e + 7./1152.*m33*mp2*RSp45633e*RS3456pee
          + 7./1152.*m33*mp2*RSp45633e*RS3456ppe );

      F13p6x23R +=  + pi162 * (  - 1./384.*m33*mp2*RSp45633e*RSe45633p
          - 1./384.*m33*mp2*pow(RSp45633e,2) + 7./576.*m33*mp2*RS345pe*
         RSp45613e - 7./1152.*m33*mp2*RS345pe*RSp45633e + 7./576.*m33*
         mp2*RS346pe*RSp45613e - 7./1152.*m33*mp2*RS346pe*RSp45633e + 7.
         /576.*m33*mp2*RS356pe*RSp45613e - 7./1152.*m33*mp2*RS356pe*
         RSp45633e - 1./96.*m33*mp2*RSp563e - 1./96.*m33*mp2*RSp463e - 
         1./96.*m33*mp2*RSp453e - 1./192.*m33*mp2*RS1456pee*RSp45613e
          + 1./384.*m33*mp2*RS1456pee*RSp45633e - 1./192.*m33*mp2*
         RS1456ppe*RSp45613e + 1./384.*m33*mp2*RS1456ppe*RSp45633e + 1./
         192.*m33*mp2*RSe45611p*RSp45613e - 1./384.*m33*mp2*RSe45611p*
         RSp45633e + 1./64.*m33*mp2*RSp45611e + 7./576.*m33*mp2*
         RSp45611e*RS34561pe + 1./192.*m33*mp2*RSp45611e*RS14563pe + 1./
         192.*m33*mp2*RSp45611e*RSe45613p + 1./96.*m33*mp2*RSp45611e*
         RSp45613e + 7./1152.*m33*mp2*RSp45611e*RS3456pee + 7./1152.*
         m33*mp2*RSp45611e*RS3456ppe - 1./384.*m33*mp2*RSp45611e*
         RSe45633p );

      F13p6x23R +=  + pi162 * (  - 1./192.*m33*mp2*RSp45611e*RSp45633e
          - 7./1152.*m33*mp2*RSp45611e*RS345pe - 7./1152.*m33*mp2*
         RSp45611e*RS346pe - 7./1152.*m33*mp2*RSp45611e*RS356pe + 1./
         384.*m33*mp2*RSp45611e*RS1456pee + 1./384.*m33*mp2*RSp45611e*
         RS1456ppe - 1./384.*m33*mp2*RSp45611e*RSe45611p - 1./384.*m33*
         mp2*pow(RSp45611e,2) + 1./192.*m33*mp2*RS145pe*RSp45613e - 1./
         384.*m33*mp2*RS145pe*RSp45633e - 1./384.*m33*mp2*RS145pe*
         RSp45611e + 1./192.*m33*mp2*RS146pe*RSp45613e - 1./384.*m33*
         mp2*RS146pe*RSp45633e - 1./384.*m33*mp2*RS146pe*RSp45611e + 1./
         192.*m33*mp2*RS156pe*RSp45613e - 1./384.*m33*mp2*RS156pe*
         RSp45633e - 1./384.*m33*mp2*RS156pe*RSp45611e - 1./192.*m33*
         mp2*RSp561e - 1./192.*m33*mp2*RSp461e - 1./192.*m33*mp2*
         RSp451e + 1./128.*m33*mp2*RSp564e + 1./128.*m33*mp2*RSp465e + 
         1./128.*m33*mp2*RSp456e - 1./64.*m33*me2*RSe45613p - 7./288.*
         m33*me2*RSe45613p*RS34561pe - 1./96.*m33*me2*RSe45613p*
         RS14563pe );

      F13p6x23R +=  + pi162 * (  - 1./96.*m33*me2*pow(RSe45613p,2) - 1./
         96.*m33*me2*RSp45613e*RSe45613p - 7./576.*m33*me2*RS3456pee*
         RSe45613p - 7./576.*m33*me2*RS3456ppe*RSe45613p + 3./128.*m33*
         me2*RSe45633p + 7./576.*m33*me2*RSe45633p*RS34561pe + 1./192.*
         m33*me2*RSe45633p*RS14563pe + 1./96.*m33*me2*RSe45633p*
         RSe45613p + 1./192.*m33*me2*RSe45633p*RSp45613e + 7./1152.*m33
         *me2*RSe45633p*RS3456pee + 7./1152.*m33*me2*RSe45633p*
         RS3456ppe - 1./384.*m33*me2*pow(RSe45633p,2) + 1./192.*m33*me2
         *RSp45633e*RSe45613p - 1./384.*m33*me2*RSp45633e*RSe45633p + 7.
         /576.*m33*me2*RS345pe*RSe45613p - 7./1152.*m33*me2*RS345pe*
         RSe45633p + 7./576.*m33*me2*RS346pe*RSe45613p - 7./1152.*m33*
         me2*RS346pe*RSe45633p + 7./576.*m33*me2*RS356pe*RSe45613p - 7./
         1152.*m33*me2*RS356pe*RSe45633p - 1./96.*m33*me2*RSe563p - 1./
         96.*m33*me2*RSe463p - 1./96.*m33*me2*RSe453p - 1./192.*m33*me2
         *RS1456pee*RSe45613p + 1./384.*m33*me2*RS1456pee*RSe45633p - 1.
         /192.*m33*me2*RS1456ppe*RSe45613p );

      F13p6x23R +=  + pi162 * ( 1./384.*m33*me2*RS1456ppe*RSe45633p + 1.
         /64.*m33*me2*RSe45611p + 7./576.*m33*me2*RSe45611p*RS34561pe
          + 1./192.*m33*me2*RSe45611p*RS14563pe + 1./96.*m33*me2*
         RSe45611p*RSe45613p + 1./192.*m33*me2*RSe45611p*RSp45613e + 7./
         1152.*m33*me2*RSe45611p*RS3456pee + 7./1152.*m33*me2*RSe45611p
         *RS3456ppe - 1./192.*m33*me2*RSe45611p*RSe45633p - 1./384.*m33
         *me2*RSe45611p*RSp45633e - 7./1152.*m33*me2*RSe45611p*RS345pe
          - 7./1152.*m33*me2*RSe45611p*RS346pe - 7./1152.*m33*me2*
         RSe45611p*RS356pe + 1./384.*m33*me2*RSe45611p*RS1456pee + 1./
         384.*m33*me2*RSe45611p*RS1456ppe - 1./384.*m33*me2*pow(
         RSe45611p,2) + 1./192.*m33*me2*RSp45611e*RSe45613p - 1./384.*
         m33*me2*RSp45611e*RSe45633p - 1./384.*m33*me2*RSp45611e*
         RSe45611p + 1./192.*m33*me2*RS145pe*RSe45613p - 1./384.*m33*
         me2*RS145pe*RSe45633p - 1./384.*m33*me2*RS145pe*RSe45611p + 1./
         192.*m33*me2*RS146pe*RSe45613p - 1./384.*m33*me2*RS146pe*
         RSe45633p );

      F13p6x23R +=  + pi162 * (  - 1./384.*m33*me2*RS146pe*RSe45611p + 
         1./192.*m33*me2*RS156pe*RSe45613p - 1./384.*m33*me2*RS156pe*
         RSe45633p - 1./384.*m33*me2*RS156pe*RSe45611p - 1./192.*m33*
         me2*RSe561p - 1./192.*m33*me2*RSe461p - 1./192.*m33*me2*
         RSe451p + 1./128.*m33*me2*RSe564p + 1./128.*m33*me2*RSe465p + 
         1./128.*m33*me2*RSe456p - 5./64.*m33*m66 - 3./128.*m33*m66*
         RS34561pe - 5./384.*m33*m66*RS14563pe - 5./384.*m33*m66*
         RSe45613p - 5./384.*m33*m66*RSp45613e - 19./768.*m33*m66*
         RS3456pee - 19./768.*m33*m66*RS3456ppe + 3./256.*m33*m66*
         RSe45633p + 3./256.*m33*m66*RSp45633e - 1./768.*m33*m66*
         RS345pe + 19./768.*m33*m66*RS346pe + 19./768.*m33*m66*RS356pe
          - 1./96.*m33*m66*RSe453p - 1./96.*m33*m66*RSp453e - 1./256.*
         m33*m66*RS1456pee - 1./256.*m33*m66*RS1456ppe + 1./256.*m33*
         m66*RSe45611p + 1./256.*m33*m66*RSp45611e + 7./768.*m33*m66*
         RS145pe + 1./256.*m33*m66*RS146pe + 1./256.*m33*m66*RS156pe + 
         1./96.*m33*m66*RS645pe );

      F13p6x23R +=  + pi162 * ( 1./192.*m33*m66*RSe451p + 1./192.*m33*
         m66*RSp451e + 1./384.*m33*m66*RSe456p + 1./384.*m33*m66*
         RSp456e - 5./64.*m33*m55 - 3./128.*m33*m55*RS34561pe - 5./384.
         *m33*m55*RS14563pe - 5./384.*m33*m55*RSe45613p - 5./384.*m33*
         m55*RSp45613e - 19./768.*m33*m55*RS3456pee - 19./768.*m33*m55*
         RS3456ppe + 3./256.*m33*m55*RSe45633p + 3./256.*m33*m55*
         RSp45633e + 19./768.*m33*m55*RS345pe - 1./768.*m33*m55*RS346pe
          + 19./768.*m33*m55*RS356pe - 1./96.*m33*m55*RSe463p - 1./96.*
         m33*m55*RSp463e - 1./256.*m33*m55*RS1456pee - 1./256.*m33*m55*
         RS1456ppe + 1./256.*m33*m55*RSe45611p + 1./256.*m33*m55*
         RSp45611e + 1./256.*m33*m55*RS145pe + 7./768.*m33*m55*RS146pe
          + 1./256.*m33*m55*RS156pe + 1./96.*m33*m55*RS546pe + 1./192.*
         m33*m55*RSe461p + 1./192.*m33*m55*RSp461e + 1./384.*m33*m55*
         RSe465p + 1./384.*m33*m55*RSp465e - 5./64.*m33*m44 - 3./128.*
         m33*m44*RS34561pe - 5./384.*m33*m44*RS14563pe - 5./384.*m33*
         m44*RSe45613p );

      F13p6x23R +=  + pi162 * (  - 5./384.*m33*m44*RSp45613e - 19./768.
         *m33*m44*RS3456pee - 19./768.*m33*m44*RS3456ppe + 3./256.*m33*
         m44*RSe45633p + 3./256.*m33*m44*RSp45633e + 19./768.*m33*m44*
         RS345pe + 19./768.*m33*m44*RS346pe - 1./768.*m33*m44*RS356pe
          - 1./96.*m33*m44*RSe563p - 1./96.*m33*m44*RSp563e - 1./256.*
         m33*m44*RS1456pee - 1./256.*m33*m44*RS1456ppe + 1./256.*m33*
         m44*RSe45611p + 1./256.*m33*m44*RSp45611e + 1./256.*m33*m44*
         RS145pe + 1./256.*m33*m44*RS146pe + 7./768.*m33*m44*RS156pe + 
         1./96.*m33*m44*RS456pe + 1./192.*m33*m44*RSe561p + 1./192.*m33
         *m44*RSp561e + 1./384.*m33*m44*RSe564p + 1./384.*m33*m44*
         RSp564e + 43./512.*pow(m33,2) + 5./512.*pow(m33,2)*RS34561pe
          - 137./6912.*pow(m33,2)*pow(RS34561pe,2) - 17./1536.*pow(
         m33,2)*RS14563pe - 53./3456.*pow(m33,2)*RS14563pe*RS34561pe - 
         17./6912.*pow(m33,2)*pow(RS14563pe,2) - 17./1536.*pow(m33,2)*
         RSe45613p - 53./3456.*pow(m33,2)*RSe45613p*RS34561pe - 17./
         3456.*pow(m33,2)*RSe45613p*RS14563pe );

      F13p6x23R +=  + pi162 * (  - 17./6912.*pow(m33,2)*pow(
         RSe45613p,2) - 17./1536.*pow(m33,2)*RSp45613e - 53./3456.*pow(
         m33,2)*RSp45613e*RS34561pe - 17./3456.*pow(m33,2)*RSp45613e*
         RS14563pe - 17./3456.*pow(m33,2)*RSp45613e*RSe45613p - 17./
         6912.*pow(m33,2)*pow(RSp45613e,2) - 37./1536.*pow(m33,2)*
         RS3456pee - 137./6912.*pow(m33,2)*RS3456pee*RS34561pe - 53./
         6912.*pow(m33,2)*RS3456pee*RS14563pe - 53./6912.*pow(m33,2)*
         RS3456pee*RSe45613p - 53./6912.*pow(m33,2)*RS3456pee*RSp45613e
          - 137./27648.*pow(m33,2)*pow(RS3456pee,2) - 37./1536.*pow(
         m33,2)*RS3456ppe - 137./6912.*pow(m33,2)*RS3456ppe*RS34561pe
          - 53./6912.*pow(m33,2)*RS3456ppe*RS14563pe - 53./6912.*pow(
         m33,2)*RS3456ppe*RSe45613p - 53./6912.*pow(m33,2)*RS3456ppe*
         RSp45613e - 137./13824.*pow(m33,2)*RS3456ppe*RS3456pee - 137./
         27648.*pow(m33,2)*pow(RS3456ppe,2) + 17./1536.*pow(m33,2)*
         RSe45633p + 53./6912.*pow(m33,2)*RSe45633p*RS34561pe + 17./
         6912.*pow(m33,2)*RSe45633p*RS14563pe );

      F13p6x23R +=  + pi162 * ( 17./6912.*pow(m33,2)*RSe45633p*
         RSe45613p + 17./6912.*pow(m33,2)*RSe45633p*RSp45613e + 53./
         13824.*pow(m33,2)*RSe45633p*RS3456pee + 53./13824.*pow(m33,2)*
         RSe45633p*RS3456ppe - 17./27648.*pow(m33,2)*pow(RSe45633p,2)
          + 17./1536.*pow(m33,2)*RSp45633e + 53./6912.*pow(m33,2)*
         RSp45633e*RS34561pe + 17./6912.*pow(m33,2)*RSp45633e*RS14563pe
          + 17./6912.*pow(m33,2)*RSp45633e*RSe45613p + 17./6912.*pow(
         m33,2)*RSp45633e*RSp45613e + 53./13824.*pow(m33,2)*RSp45633e*
         RS3456pee + 53./13824.*pow(m33,2)*RSp45633e*RS3456ppe - 17./
         13824.*pow(m33,2)*RSp45633e*RSe45633p - 17./27648.*pow(m33,2)*
         pow(RSp45633e,2) - 13./2304.*pow(m33,2)*RS345pe + 137./6912.*
         pow(m33,2)*RS345pe*RS34561pe + 53./6912.*pow(m33,2)*RS345pe*
         RS14563pe + 53./6912.*pow(m33,2)*RS345pe*RSe45613p + 53./6912.
         *pow(m33,2)*RS345pe*RSp45613e + 137./13824.*pow(m33,2)*RS345pe
         *RS3456pee + 137./13824.*pow(m33,2)*RS345pe*RS3456ppe - 53./
         13824.*pow(m33,2)*RS345pe*RSe45633p );

      F13p6x23R +=  + pi162 * (  - 53./13824.*pow(m33,2)*RS345pe*
         RSp45633e - 137./27648.*pow(m33,2)*pow(RS345pe,2) - 13./2304.*
         pow(m33,2)*RS346pe + 137./6912.*pow(m33,2)*RS346pe*RS34561pe
          + 53./6912.*pow(m33,2)*RS346pe*RS14563pe + 53./6912.*pow(
         m33,2)*RS346pe*RSe45613p + 53./6912.*pow(m33,2)*RS346pe*
         RSp45613e + 137./13824.*pow(m33,2)*RS346pe*RS3456pee + 137./
         13824.*pow(m33,2)*RS346pe*RS3456ppe - 53./13824.*pow(m33,2)*
         RS346pe*RSe45633p - 53./13824.*pow(m33,2)*RS346pe*RSp45633e - 
         137./13824.*pow(m33,2)*RS346pe*RS345pe - 137./27648.*pow(
         m33,2)*pow(RS346pe,2) - 13./2304.*pow(m33,2)*RS356pe + 137./
         6912.*pow(m33,2)*RS356pe*RS34561pe + 53./6912.*pow(m33,2)*
         RS356pe*RS14563pe + 53./6912.*pow(m33,2)*RS356pe*RSe45613p + 
         53./6912.*pow(m33,2)*RS356pe*RSp45613e + 137./13824.*pow(
         m33,2)*RS356pe*RS3456pee + 137./13824.*pow(m33,2)*RS356pe*
         RS3456ppe - 53./13824.*pow(m33,2)*RS356pe*RSe45633p - 53./
         13824.*pow(m33,2)*RS356pe*RSp45633e );

      F13p6x23R +=  + pi162 * (  - 137./13824.*pow(m33,2)*RS356pe*
         RS345pe - 137./13824.*pow(m33,2)*RS356pe*RS346pe - 137./27648.
         *pow(m33,2)*pow(RS356pe,2) - 17./4608.*pow(m33,2)*RSe563p - 17.
         /4608.*pow(m33,2)*RSe463p - 17./4608.*pow(m33,2)*RSe453p - 17./
         4608.*pow(m33,2)*RSp563e - 17./4608.*pow(m33,2)*RSp463e - 17./
         4608.*pow(m33,2)*RSp453e + 1./1536.*pow(m33,2)*RS1456pee - 53./
         6912.*pow(m33,2)*RS1456pee*RS34561pe - 17./6912.*pow(m33,2)*
         RS1456pee*RS14563pe - 17./6912.*pow(m33,2)*RS1456pee*RSe45613p
          - 17./6912.*pow(m33,2)*RS1456pee*RSp45613e - 53./13824.*pow(
         m33,2)*RS1456pee*RS3456pee - 53./13824.*pow(m33,2)*RS1456pee*
         RS3456ppe + 17./13824.*pow(m33,2)*RS1456pee*RSe45633p + 17./
         13824.*pow(m33,2)*RS1456pee*RSp45633e + 53./13824.*pow(m33,2)*
         RS1456pee*RS345pe + 53./13824.*pow(m33,2)*RS1456pee*RS346pe + 
         53./13824.*pow(m33,2)*RS1456pee*RS356pe - 17./27648.*pow(
         m33,2)*pow(RS1456pee,2) + 1./1536.*pow(m33,2)*RS1456ppe - 53./
         6912.*pow(m33,2)*RS1456ppe*RS34561pe );

      F13p6x23R +=  + pi162 * (  - 17./6912.*pow(m33,2)*RS1456ppe*
         RS14563pe - 17./6912.*pow(m33,2)*RS1456ppe*RSe45613p - 17./
         6912.*pow(m33,2)*RS1456ppe*RSp45613e - 53./13824.*pow(m33,2)*
         RS1456ppe*RS3456pee - 53./13824.*pow(m33,2)*RS1456ppe*
         RS3456ppe + 17./13824.*pow(m33,2)*RS1456ppe*RSe45633p + 17./
         13824.*pow(m33,2)*RS1456ppe*RSp45633e + 53./13824.*pow(m33,2)*
         RS1456ppe*RS345pe + 53./13824.*pow(m33,2)*RS1456ppe*RS346pe + 
         53./13824.*pow(m33,2)*RS1456ppe*RS356pe - 17./13824.*pow(
         m33,2)*RS1456ppe*RS1456pee - 17./27648.*pow(m33,2)*pow(
         RS1456ppe,2) - 1./1536.*pow(m33,2)*RSe45611p + 53./6912.*pow(
         m33,2)*RSe45611p*RS34561pe + 17./6912.*pow(m33,2)*RSe45611p*
         RS14563pe + 17./6912.*pow(m33,2)*RSe45611p*RSe45613p + 17./
         6912.*pow(m33,2)*RSe45611p*RSp45613e + 53./13824.*pow(m33,2)*
         RSe45611p*RS3456pee + 53./13824.*pow(m33,2)*RSe45611p*
         RS3456ppe - 17./13824.*pow(m33,2)*RSe45611p*RSe45633p - 17./
         13824.*pow(m33,2)*RSe45611p*RSp45633e );

      F13p6x23R +=  + pi162 * (  - 53./13824.*pow(m33,2)*RSe45611p*
         RS345pe - 53./13824.*pow(m33,2)*RSe45611p*RS346pe - 53./13824.
         *pow(m33,2)*RSe45611p*RS356pe + 17./13824.*pow(m33,2)*
         RSe45611p*RS1456pee + 17./13824.*pow(m33,2)*RSe45611p*
         RS1456ppe - 17./27648.*pow(m33,2)*pow(RSe45611p,2) - 1./1536.*
         pow(m33,2)*RSp45611e + 53./6912.*pow(m33,2)*RSp45611e*
         RS34561pe + 17./6912.*pow(m33,2)*RSp45611e*RS14563pe + 17./
         6912.*pow(m33,2)*RSp45611e*RSe45613p + 17./6912.*pow(m33,2)*
         RSp45611e*RSp45613e + 53./13824.*pow(m33,2)*RSp45611e*
         RS3456pee + 53./13824.*pow(m33,2)*RSp45611e*RS3456ppe - 17./
         13824.*pow(m33,2)*RSp45611e*RSe45633p - 17./13824.*pow(m33,2)*
         RSp45611e*RSp45633e - 53./13824.*pow(m33,2)*RSp45611e*RS345pe
          - 53./13824.*pow(m33,2)*RSp45611e*RS346pe - 53./13824.*pow(
         m33,2)*RSp45611e*RS356pe + 17./13824.*pow(m33,2)*RSp45611e*
         RS1456pee + 17./13824.*pow(m33,2)*RSp45611e*RS1456ppe - 17./
         13824.*pow(m33,2)*RSp45611e*RSe45611p );

      F13p6x23R +=  + pi162 * (  - 17./27648.*pow(m33,2)*pow(
         RSp45611e,2) + 1./288.*pow(m33,2)*RS145pe + 53./6912.*pow(
         m33,2)*RS145pe*RS34561pe + 17./6912.*pow(m33,2)*RS145pe*
         RS14563pe + 17./6912.*pow(m33,2)*RS145pe*RSe45613p + 17./6912.
         *pow(m33,2)*RS145pe*RSp45613e + 53./13824.*pow(m33,2)*RS145pe*
         RS3456pee + 53./13824.*pow(m33,2)*RS145pe*RS3456ppe - 17./
         13824.*pow(m33,2)*RS145pe*RSe45633p - 17./13824.*pow(m33,2)*
         RS145pe*RSp45633e - 53./13824.*pow(m33,2)*RS145pe*RS345pe - 53.
         /13824.*pow(m33,2)*RS145pe*RS346pe - 53./13824.*pow(m33,2)*
         RS145pe*RS356pe + 17./13824.*pow(m33,2)*RS145pe*RS1456pee + 17.
         /13824.*pow(m33,2)*RS145pe*RS1456ppe - 17./13824.*pow(m33,2)*
         RS145pe*RSe45611p - 17./13824.*pow(m33,2)*RS145pe*RSp45611e - 
         17./27648.*pow(m33,2)*pow(RS145pe,2) + 1./288.*pow(m33,2)*
         RS146pe + 53./6912.*pow(m33,2)*RS146pe*RS34561pe + 17./6912.*
         pow(m33,2)*RS146pe*RS14563pe + 17./6912.*pow(m33,2)*RS146pe*
         RSe45613p );

      F13p6x23R +=  + pi162 * ( 17./6912.*pow(m33,2)*RS146pe*RSp45613e
          + 53./13824.*pow(m33,2)*RS146pe*RS3456pee + 53./13824.*pow(
         m33,2)*RS146pe*RS3456ppe - 17./13824.*pow(m33,2)*RS146pe*
         RSe45633p - 17./13824.*pow(m33,2)*RS146pe*RSp45633e - 53./
         13824.*pow(m33,2)*RS146pe*RS345pe - 53./13824.*pow(m33,2)*
         RS146pe*RS346pe - 53./13824.*pow(m33,2)*RS146pe*RS356pe + 17./
         13824.*pow(m33,2)*RS146pe*RS1456pee + 17./13824.*pow(m33,2)*
         RS146pe*RS1456ppe - 17./13824.*pow(m33,2)*RS146pe*RSe45611p - 
         17./13824.*pow(m33,2)*RS146pe*RSp45611e - 17./13824.*pow(
         m33,2)*RS146pe*RS145pe - 17./27648.*pow(m33,2)*pow(RS146pe,2)
          + 1./288.*pow(m33,2)*RS156pe + 53./6912.*pow(m33,2)*RS156pe*
         RS34561pe + 17./6912.*pow(m33,2)*RS156pe*RS14563pe + 17./6912.
         *pow(m33,2)*RS156pe*RSe45613p + 17./6912.*pow(m33,2)*RS156pe*
         RSp45613e + 53./13824.*pow(m33,2)*RS156pe*RS3456pee + 53./
         13824.*pow(m33,2)*RS156pe*RS3456ppe - 17./13824.*pow(m33,2)*
         RS156pe*RSe45633p );

      F13p6x23R +=  + pi162 * (  - 17./13824.*pow(m33,2)*RS156pe*
         RSp45633e - 53./13824.*pow(m33,2)*RS156pe*RS345pe - 53./13824.
         *pow(m33,2)*RS156pe*RS346pe - 53./13824.*pow(m33,2)*RS156pe*
         RS356pe + 17./13824.*pow(m33,2)*RS156pe*RS1456pee + 17./13824.
         *pow(m33,2)*RS156pe*RS1456ppe - 17./13824.*pow(m33,2)*RS156pe*
         RSe45611p - 17./13824.*pow(m33,2)*RS156pe*RSp45611e - 17./
         13824.*pow(m33,2)*RS156pe*RS145pe - 17./13824.*pow(m33,2)*
         RS156pe*RS146pe - 17./27648.*pow(m33,2)*pow(RS156pe,2) - 1./
         4608.*pow(m33,2)*RS645pe - 1./4608.*pow(m33,2)*RS546pe - 1./
         4608.*pow(m33,2)*RS456pe + 19./4608.*pow(m33,2)*RSe561p + 19./
         4608.*pow(m33,2)*RSe461p + 19./4608.*pow(m33,2)*RSe451p + 19./
         4608.*pow(m33,2)*RSp561e + 19./4608.*pow(m33,2)*RSp461e + 19./
         4608.*pow(m33,2)*RSp451e - 1./4608.*pow(m33,2)*RSe564p - 1./
         4608.*pow(m33,2)*RSe465p - 1./4608.*pow(m33,2)*RSe456p - 1./
         4608.*pow(m33,2)*RSp564e - 1./4608.*pow(m33,2)*RSp465e - 1./
         4608.*pow(m33,2)*RSp456e );

      F13p6x23R +=  + pi162 * ( 1./64.*m11*RD3456ep + 1./192.*m11*
         RD3456ep*RS34561pe + 7./576.*m11*RD3456ep*RS14563pe + 1./192.*
         m11*RD3456ep*RSe45613p + 1./192.*m11*RD3456ep*RSp45613e + 1./
         384.*m11*RS3456pee*RD3456ep + 1./384.*m11*RS3456ppe*RD3456ep
          - 1./384.*m11*RSe45633p*RD3456ep - 1./384.*m11*RSp45633e*
         RD3456ep - 1./384.*m11*RS345pe*RD3456ep - 1./384.*m11*RS346pe*
         RD3456ep - 1./384.*m11*RS356pe*RD3456ep + 1./384.*m11*RD1456ep
          + 1./192.*m11*RD1456ep*RS34561pe + 7./576.*m11*RD1456ep*
         RS14563pe + 1./192.*m11*RD1456ep*RSe45613p + 1./192.*m11*
         RD1456ep*RSp45613e + 1./384.*m11*RD1456ep*RS3456pee + 1./384.*
         m11*RD1456ep*RS3456ppe - 1./384.*m11*RD1456ep*RSe45633p - 1./
         384.*m11*RD1456ep*RSp45633e - 1./384.*m11*RD1456ep*RS345pe - 1.
         /384.*m11*RD1456ep*RS346pe - 1./384.*m11*RD1456ep*RS356pe + 7./
         1152.*m11*RS1456pee*RD3456ep + 7./1152.*m11*RS1456pee*RD1456ep
          + 7./1152.*m11*RS1456ppe*RD3456ep + 7./1152.*m11*RS1456ppe*
         RD1456ep );

      F13p6x23R +=  + pi162 * (  - 1./384.*m11*RSe45611p*RD3456ep - 1./
         384.*m11*RSe45611p*RD1456ep - 1./384.*m11*RSp45611e*RD3456ep
          - 1./384.*m11*RSp45611e*RD1456ep - 7./1152.*m11*RS145pe*
         RD3456ep - 7./1152.*m11*RS145pe*RD1456ep - 7./1152.*m11*
         RS146pe*RD3456ep - 7./1152.*m11*RS146pe*RD1456ep - 7./1152.*
         m11*RS156pe*RD3456ep - 7./1152.*m11*RS156pe*RD1456ep - 1./64.*
         m11*mp2*RSp45613e - 1./96.*m11*mp2*RSp45613e*RS34561pe - 7./
         288.*m11*mp2*RSp45613e*RS14563pe - 1./96.*m11*mp2*RSp45613e*
         RSe45613p - 1./96.*m11*mp2*pow(RSp45613e,2) - 1./192.*m11*mp2*
         RS3456pee*RSp45613e - 1./192.*m11*mp2*RS3456ppe*RSp45613e + 1./
         192.*m11*mp2*RSe45633p*RSp45613e + 1./64.*m11*mp2*RSp45633e + 
         1./192.*m11*mp2*RSp45633e*RS34561pe + 7./576.*m11*mp2*
         RSp45633e*RS14563pe + 1./192.*m11*mp2*RSp45633e*RSe45613p + 1./
         96.*m11*mp2*RSp45633e*RSp45613e + 1./384.*m11*mp2*RSp45633e*
         RS3456pee + 1./384.*m11*mp2*RSp45633e*RS3456ppe - 1./384.*m11*
         mp2*RSp45633e*RSe45633p );

      F13p6x23R +=  + pi162 * (  - 1./384.*m11*mp2*pow(RSp45633e,2) + 1.
         /192.*m11*mp2*RS345pe*RSp45613e - 1./384.*m11*mp2*RS345pe*
         RSp45633e + 1./192.*m11*mp2*RS346pe*RSp45613e - 1./384.*m11*
         mp2*RS346pe*RSp45633e + 1./192.*m11*mp2*RS356pe*RSp45613e - 1./
         384.*m11*mp2*RS356pe*RSp45633e - 1./192.*m11*mp2*RSp563e - 1./
         192.*m11*mp2*RSp463e - 1./192.*m11*mp2*RSp453e - 7./576.*m11*
         mp2*RS1456pee*RSp45613e + 7./1152.*m11*mp2*RS1456pee*RSp45633e
          - 7./576.*m11*mp2*RS1456ppe*RSp45613e + 7./1152.*m11*mp2*
         RS1456ppe*RSp45633e + 1./192.*m11*mp2*RSe45611p*RSp45613e - 1./
         384.*m11*mp2*RSe45611p*RSp45633e + 3./128.*m11*mp2*RSp45611e
          + 1./192.*m11*mp2*RSp45611e*RS34561pe + 7./576.*m11*mp2*
         RSp45611e*RS14563pe + 1./192.*m11*mp2*RSp45611e*RSe45613p + 1./
         96.*m11*mp2*RSp45611e*RSp45613e + 1./384.*m11*mp2*RSp45611e*
         RS3456pee + 1./384.*m11*mp2*RSp45611e*RS3456ppe - 1./384.*m11*
         mp2*RSp45611e*RSe45633p - 1./192.*m11*mp2*RSp45611e*RSp45633e
          - 1./384.*m11*mp2*RSp45611e*RS345pe );

      F13p6x23R +=  + pi162 * (  - 1./384.*m11*mp2*RSp45611e*RS346pe - 
         1./384.*m11*mp2*RSp45611e*RS356pe + 7./1152.*m11*mp2*RSp45611e
         *RS1456pee + 7./1152.*m11*mp2*RSp45611e*RS1456ppe - 1./384.*
         m11*mp2*RSp45611e*RSe45611p - 1./384.*m11*mp2*pow(RSp45611e,2)
          + 7./576.*m11*mp2*RS145pe*RSp45613e - 7./1152.*m11*mp2*
         RS145pe*RSp45633e - 7./1152.*m11*mp2*RS145pe*RSp45611e + 7./
         576.*m11*mp2*RS146pe*RSp45613e - 7./1152.*m11*mp2*RS146pe*
         RSp45633e - 7./1152.*m11*mp2*RS146pe*RSp45611e + 7./576.*m11*
         mp2*RS156pe*RSp45613e - 7./1152.*m11*mp2*RS156pe*RSp45633e - 7.
         /1152.*m11*mp2*RS156pe*RSp45611e - 1./96.*m11*mp2*RSp561e - 1./
         96.*m11*mp2*RSp461e - 1./96.*m11*mp2*RSp451e + 1./128.*m11*mp2
         *RSp564e + 1./128.*m11*mp2*RSp465e + 1./128.*m11*mp2*RSp456e
          - 1./64.*m11*me2*RSe45613p - 1./96.*m11*me2*RSe45613p*
         RS34561pe - 7./288.*m11*me2*RSe45613p*RS14563pe - 1./96.*m11*
         me2*pow(RSe45613p,2) - 1./96.*m11*me2*RSp45613e*RSe45613p - 1./
         192.*m11*me2*RS3456pee*RSe45613p );

      F13p6x23R +=  + pi162 * (  - 1./192.*m11*me2*RS3456ppe*RSe45613p
          + 1./64.*m11*me2*RSe45633p + 1./192.*m11*me2*RSe45633p*
         RS34561pe + 7./576.*m11*me2*RSe45633p*RS14563pe + 1./96.*m11*
         me2*RSe45633p*RSe45613p + 1./192.*m11*me2*RSe45633p*RSp45613e
          + 1./384.*m11*me2*RSe45633p*RS3456pee + 1./384.*m11*me2*
         RSe45633p*RS3456ppe - 1./384.*m11*me2*pow(RSe45633p,2) + 1./
         192.*m11*me2*RSp45633e*RSe45613p - 1./384.*m11*me2*RSp45633e*
         RSe45633p + 1./192.*m11*me2*RS345pe*RSe45613p - 1./384.*m11*
         me2*RS345pe*RSe45633p + 1./192.*m11*me2*RS346pe*RSe45613p - 1./
         384.*m11*me2*RS346pe*RSe45633p + 1./192.*m11*me2*RS356pe*
         RSe45613p - 1./384.*m11*me2*RS356pe*RSe45633p - 1./192.*m11*
         me2*RSe563p - 1./192.*m11*me2*RSe463p - 1./192.*m11*me2*
         RSe453p - 7./576.*m11*me2*RS1456pee*RSe45613p + 7./1152.*m11*
         me2*RS1456pee*RSe45633p - 7./576.*m11*me2*RS1456ppe*RSe45613p
          + 7./1152.*m11*me2*RS1456ppe*RSe45633p + 3./128.*m11*me2*
         RSe45611p );

      F13p6x23R +=  + pi162 * ( 1./192.*m11*me2*RSe45611p*RS34561pe + 7.
         /576.*m11*me2*RSe45611p*RS14563pe + 1./96.*m11*me2*RSe45611p*
         RSe45613p + 1./192.*m11*me2*RSe45611p*RSp45613e + 1./384.*m11*
         me2*RSe45611p*RS3456pee + 1./384.*m11*me2*RSe45611p*RS3456ppe
          - 1./192.*m11*me2*RSe45611p*RSe45633p - 1./384.*m11*me2*
         RSe45611p*RSp45633e - 1./384.*m11*me2*RSe45611p*RS345pe - 1./
         384.*m11*me2*RSe45611p*RS346pe - 1./384.*m11*me2*RSe45611p*
         RS356pe + 7./1152.*m11*me2*RSe45611p*RS1456pee + 7./1152.*m11*
         me2*RSe45611p*RS1456ppe - 1./384.*m11*me2*pow(RSe45611p,2) + 1.
         /192.*m11*me2*RSp45611e*RSe45613p - 1./384.*m11*me2*RSp45611e*
         RSe45633p - 1./384.*m11*me2*RSp45611e*RSe45611p + 7./576.*m11*
         me2*RS145pe*RSe45613p - 7./1152.*m11*me2*RS145pe*RSe45633p - 7.
         /1152.*m11*me2*RS145pe*RSe45611p + 7./576.*m11*me2*RS146pe*
         RSe45613p - 7./1152.*m11*me2*RS146pe*RSe45633p - 7./1152.*m11*
         me2*RS146pe*RSe45611p + 7./576.*m11*me2*RS156pe*RSe45613p - 7./
         1152.*m11*me2*RS156pe*RSe45633p );

      F13p6x23R +=  + pi162 * (  - 7./1152.*m11*me2*RS156pe*RSe45611p
          - 1./96.*m11*me2*RSe561p - 1./96.*m11*me2*RSe461p - 1./96.*
         m11*me2*RSe451p + 1./128.*m11*me2*RSe564p + 1./128.*m11*me2*
         RSe465p + 1./128.*m11*me2*RSe456p - 5./64.*m11*m66 - 5./384.*
         m11*m66*RS34561pe - 3./128.*m11*m66*RS14563pe - 5./384.*m11*
         m66*RSe45613p - 5./384.*m11*m66*RSp45613e - 1./256.*m11*m66*
         RS3456pee - 1./256.*m11*m66*RS3456ppe + 1./256.*m11*m66*
         RSe45633p + 1./256.*m11*m66*RSp45633e + 7./768.*m11*m66*
         RS345pe + 1./256.*m11*m66*RS346pe + 1./256.*m11*m66*RS356pe + 
         1./192.*m11*m66*RSe453p + 1./192.*m11*m66*RSp453e - 19./768.*
         m11*m66*RS1456pee - 19./768.*m11*m66*RS1456ppe + 3./256.*m11*
         m66*RSe45611p + 3./256.*m11*m66*RSp45611e - 1./768.*m11*m66*
         RS145pe + 19./768.*m11*m66*RS146pe + 19./768.*m11*m66*RS156pe
          + 1./96.*m11*m66*RS645pe - 1./96.*m11*m66*RSe451p - 1./96.*
         m11*m66*RSp451e + 1./384.*m11*m66*RSe456p + 1./384.*m11*m66*
         RSp456e );

      F13p6x23R +=  + pi162 * (  - 5./64.*m11*m55 - 5./384.*m11*m55*
         RS34561pe - 3./128.*m11*m55*RS14563pe - 5./384.*m11*m55*
         RSe45613p - 5./384.*m11*m55*RSp45613e - 1./256.*m11*m55*
         RS3456pee - 1./256.*m11*m55*RS3456ppe + 1./256.*m11*m55*
         RSe45633p + 1./256.*m11*m55*RSp45633e + 1./256.*m11*m55*
         RS345pe + 7./768.*m11*m55*RS346pe + 1./256.*m11*m55*RS356pe + 
         1./192.*m11*m55*RSe463p + 1./192.*m11*m55*RSp463e - 19./768.*
         m11*m55*RS1456pee - 19./768.*m11*m55*RS1456ppe + 3./256.*m11*
         m55*RSe45611p + 3./256.*m11*m55*RSp45611e + 19./768.*m11*m55*
         RS145pe - 1./768.*m11*m55*RS146pe + 19./768.*m11*m55*RS156pe
          + 1./96.*m11*m55*RS546pe - 1./96.*m11*m55*RSe461p - 1./96.*
         m11*m55*RSp461e + 1./384.*m11*m55*RSe465p + 1./384.*m11*m55*
         RSp465e - 5./64.*m11*m44 - 5./384.*m11*m44*RS34561pe - 3./128.
         *m11*m44*RS14563pe - 5./384.*m11*m44*RSe45613p - 5./384.*m11*
         m44*RSp45613e - 1./256.*m11*m44*RS3456pee - 1./256.*m11*m44*
         RS3456ppe );

      F13p6x23R +=  + pi162 * ( 1./256.*m11*m44*RSe45633p + 1./256.*m11
         *m44*RSp45633e + 1./256.*m11*m44*RS345pe + 1./256.*m11*m44*
         RS346pe + 7./768.*m11*m44*RS356pe + 1./192.*m11*m44*RSe563p + 
         1./192.*m11*m44*RSp563e - 19./768.*m11*m44*RS1456pee - 19./768.
         *m11*m44*RS1456ppe + 3./256.*m11*m44*RSe45611p + 3./256.*m11*
         m44*RSp45611e + 19./768.*m11*m44*RS145pe + 19./768.*m11*m44*
         RS146pe - 1./768.*m11*m44*RS156pe + 1./96.*m11*m44*RS456pe - 1.
         /96.*m11*m44*RSe561p - 1./96.*m11*m44*RSp561e + 1./384.*m11*
         m44*RSe564p + 1./384.*m11*m44*RSp564e - 27./256.*m11*m33 - 23./
         768.*m11*m33*RS34561pe - 53./3456.*m11*m33*pow(RS34561pe,2) - 
         23./768.*m11*m33*RS14563pe - 77./1728.*m11*m33*RS14563pe*
         RS34561pe - 53./3456.*m11*m33*pow(RS14563pe,2) - 11./768.*m11*
         m33*RSe45613p - 35./1728.*m11*m33*RSe45613p*RS34561pe - 35./
         1728.*m11*m33*RSe45613p*RS14563pe - 17./3456.*m11*m33*pow(
         RSe45613p,2) - 11./768.*m11*m33*RSp45613e - 35./1728.*m11*m33*
         RSp45613e*RS34561pe );

      F13p6x23R +=  + pi162 * (  - 35./1728.*m11*m33*RSp45613e*
         RS14563pe - 17./1728.*m11*m33*RSp45613e*RSe45613p - 17./3456.*
         m11*m33*pow(RSp45613e,2) - 5./192.*m11*m33*RS3456pee - 53./
         3456.*m11*m33*RS3456pee*RS34561pe - 77./3456.*m11*m33*
         RS3456pee*RS14563pe - 35./3456.*m11*m33*RS3456pee*RSe45613p - 
         35./3456.*m11*m33*RS3456pee*RSp45613e - 53./13824.*m11*m33*
         pow(RS3456pee,2) - 5./192.*m11*m33*RS3456ppe - 53./3456.*m11*
         m33*RS3456ppe*RS34561pe - 77./3456.*m11*m33*RS3456ppe*
         RS14563pe - 35./3456.*m11*m33*RS3456ppe*RSe45613p - 35./3456.*
         m11*m33*RS3456ppe*RSp45613e - 53./6912.*m11*m33*RS3456ppe*
         RS3456pee - 53./13824.*m11*m33*pow(RS3456ppe,2) + 1./96.*m11*
         m33*RSe45633p + 35./3456.*m11*m33*RSe45633p*RS34561pe + 35./
         3456.*m11*m33*RSe45633p*RS14563pe + 17./3456.*m11*m33*
         RSe45633p*RSe45613p + 17./3456.*m11*m33*RSe45633p*RSp45613e + 
         35./6912.*m11*m33*RSe45633p*RS3456pee + 35./6912.*m11*m33*
         RSe45633p*RS3456ppe );

      F13p6x23R +=  + pi162 * (  - 17./13824.*m11*m33*pow(RSe45633p,2)
          + 1./96.*m11*m33*RSp45633e + 35./3456.*m11*m33*RSp45633e*
         RS34561pe + 35./3456.*m11*m33*RSp45633e*RS14563pe + 17./3456.*
         m11*m33*RSp45633e*RSe45613p + 17./3456.*m11*m33*RSp45633e*
         RSp45613e + 35./6912.*m11*m33*RSp45633e*RS3456pee + 35./6912.*
         m11*m33*RSp45633e*RS3456ppe - 17./6912.*m11*m33*RSp45633e*
         RSe45633p - 17./13824.*m11*m33*pow(RSp45633e,2) + 43./2304.*
         m11*m33*RS345pe + 53./3456.*m11*m33*RS345pe*RS34561pe + 77./
         3456.*m11*m33*RS345pe*RS14563pe + 35./3456.*m11*m33*RS345pe*
         RSe45613p + 35./3456.*m11*m33*RS345pe*RSp45613e + 53./6912.*
         m11*m33*RS345pe*RS3456pee + 53./6912.*m11*m33*RS345pe*
         RS3456ppe - 35./6912.*m11*m33*RS345pe*RSe45633p - 35./6912.*
         m11*m33*RS345pe*RSp45633e - 53./13824.*m11*m33*pow(RS345pe,2)
          + 43./2304.*m11*m33*RS346pe + 53./3456.*m11*m33*RS346pe*
         RS34561pe + 77./3456.*m11*m33*RS346pe*RS14563pe + 35./3456.*
         m11*m33*RS346pe*RSe45613p );

      F13p6x23R +=  + pi162 * ( 35./3456.*m11*m33*RS346pe*RSp45613e + 
         53./6912.*m11*m33*RS346pe*RS3456pee + 53./6912.*m11*m33*
         RS346pe*RS3456ppe - 35./6912.*m11*m33*RS346pe*RSe45633p - 35./
         6912.*m11*m33*RS346pe*RSp45633e - 53./6912.*m11*m33*RS346pe*
         RS345pe - 53./13824.*m11*m33*pow(RS346pe,2) + 43./2304.*m11*
         m33*RS356pe + 53./3456.*m11*m33*RS356pe*RS34561pe + 77./3456.*
         m11*m33*RS356pe*RS14563pe + 35./3456.*m11*m33*RS356pe*
         RSe45613p + 35./3456.*m11*m33*RS356pe*RSp45613e + 53./6912.*
         m11*m33*RS356pe*RS3456pee + 53./6912.*m11*m33*RS356pe*
         RS3456ppe - 35./6912.*m11*m33*RS356pe*RSe45633p - 35./6912.*
         m11*m33*RS356pe*RSp45633e - 53./6912.*m11*m33*RS356pe*RS345pe
          - 53./6912.*m11*m33*RS356pe*RS346pe - 53./13824.*m11*m33*pow(
         RS356pe,2) - 5./2304.*m11*m33*RSe563p - 5./2304.*m11*m33*
         RSe463p - 5./2304.*m11*m33*RSe453p - 5./2304.*m11*m33*RSp563e
          - 5./2304.*m11*m33*RSp463e - 5./2304.*m11*m33*RSp453e - 5./
         192.*m11*m33*RS1456pee );

      F13p6x23R +=  + pi162 * (  - 77./3456.*m11*m33*RS1456pee*
         RS34561pe - 53./3456.*m11*m33*RS1456pee*RS14563pe - 35./3456.*
         m11*m33*RS1456pee*RSe45613p - 35./3456.*m11*m33*RS1456pee*
         RSp45613e - 77./6912.*m11*m33*RS1456pee*RS3456pee - 77./6912.*
         m11*m33*RS1456pee*RS3456ppe + 35./6912.*m11*m33*RS1456pee*
         RSe45633p + 35./6912.*m11*m33*RS1456pee*RSp45633e + 77./6912.*
         m11*m33*RS1456pee*RS345pe + 77./6912.*m11*m33*RS1456pee*
         RS346pe + 77./6912.*m11*m33*RS1456pee*RS356pe - 53./13824.*m11
         *m33*pow(RS1456pee,2) - 5./192.*m11*m33*RS1456ppe - 77./3456.*
         m11*m33*RS1456ppe*RS34561pe - 53./3456.*m11*m33*RS1456ppe*
         RS14563pe - 35./3456.*m11*m33*RS1456ppe*RSe45613p - 35./3456.*
         m11*m33*RS1456ppe*RSp45613e - 77./6912.*m11*m33*RS1456ppe*
         RS3456pee - 77./6912.*m11*m33*RS1456ppe*RS3456ppe + 35./6912.*
         m11*m33*RS1456ppe*RSe45633p + 35./6912.*m11*m33*RS1456ppe*
         RSp45633e + 77./6912.*m11*m33*RS1456ppe*RS345pe + 77./6912.*
         m11*m33*RS1456ppe*RS346pe );

      F13p6x23R +=  + pi162 * ( 77./6912.*m11*m33*RS1456ppe*RS356pe - 
         53./6912.*m11*m33*RS1456ppe*RS1456pee - 53./13824.*m11*m33*
         pow(RS1456ppe,2) + 1./96.*m11*m33*RSe45611p + 35./3456.*m11*
         m33*RSe45611p*RS34561pe + 35./3456.*m11*m33*RSe45611p*
         RS14563pe + 17./3456.*m11*m33*RSe45611p*RSe45613p + 17./3456.*
         m11*m33*RSe45611p*RSp45613e + 35./6912.*m11*m33*RSe45611p*
         RS3456pee + 35./6912.*m11*m33*RSe45611p*RS3456ppe - 17./6912.*
         m11*m33*RSe45611p*RSe45633p - 17./6912.*m11*m33*RSe45611p*
         RSp45633e - 35./6912.*m11*m33*RSe45611p*RS345pe - 35./6912.*
         m11*m33*RSe45611p*RS346pe - 35./6912.*m11*m33*RSe45611p*
         RS356pe + 35./6912.*m11*m33*RSe45611p*RS1456pee + 35./6912.*
         m11*m33*RSe45611p*RS1456ppe - 17./13824.*m11*m33*pow(
         RSe45611p,2) + 1./96.*m11*m33*RSp45611e + 35./3456.*m11*m33*
         RSp45611e*RS34561pe + 35./3456.*m11*m33*RSp45611e*RS14563pe + 
         17./3456.*m11*m33*RSp45611e*RSe45613p + 17./3456.*m11*m33*
         RSp45611e*RSp45613e );

      F13p6x23R +=  + pi162 * ( 35./6912.*m11*m33*RSp45611e*RS3456pee
          + 35./6912.*m11*m33*RSp45611e*RS3456ppe - 17./6912.*m11*m33*
         RSp45611e*RSe45633p - 17./6912.*m11*m33*RSp45611e*RSp45633e - 
         35./6912.*m11*m33*RSp45611e*RS345pe - 35./6912.*m11*m33*
         RSp45611e*RS346pe - 35./6912.*m11*m33*RSp45611e*RS356pe + 35./
         6912.*m11*m33*RSp45611e*RS1456pee + 35./6912.*m11*m33*
         RSp45611e*RS1456ppe - 17./6912.*m11*m33*RSp45611e*RSe45611p - 
         17./13824.*m11*m33*pow(RSp45611e,2) + 43./2304.*m11*m33*
         RS145pe + 77./3456.*m11*m33*RS145pe*RS34561pe + 53./3456.*m11*
         m33*RS145pe*RS14563pe + 35./3456.*m11*m33*RS145pe*RSe45613p + 
         35./3456.*m11*m33*RS145pe*RSp45613e + 77./6912.*m11*m33*
         RS145pe*RS3456pee + 77./6912.*m11*m33*RS145pe*RS3456ppe - 35./
         6912.*m11*m33*RS145pe*RSe45633p - 35./6912.*m11*m33*RS145pe*
         RSp45633e - 77./6912.*m11*m33*RS145pe*RS345pe - 77./6912.*m11*
         m33*RS145pe*RS346pe - 77./6912.*m11*m33*RS145pe*RS356pe + 53./
         6912.*m11*m33*RS145pe*RS1456pee );

      F13p6x23R +=  + pi162 * ( 53./6912.*m11*m33*RS145pe*RS1456ppe - 
         35./6912.*m11*m33*RS145pe*RSe45611p - 35./6912.*m11*m33*
         RS145pe*RSp45611e - 53./13824.*m11*m33*pow(RS145pe,2) + 43./
         2304.*m11*m33*RS146pe + 77./3456.*m11*m33*RS146pe*RS34561pe + 
         53./3456.*m11*m33*RS146pe*RS14563pe + 35./3456.*m11*m33*
         RS146pe*RSe45613p + 35./3456.*m11*m33*RS146pe*RSp45613e + 77./
         6912.*m11*m33*RS146pe*RS3456pee + 77./6912.*m11*m33*RS146pe*
         RS3456ppe - 35./6912.*m11*m33*RS146pe*RSe45633p - 35./6912.*
         m11*m33*RS146pe*RSp45633e - 77./6912.*m11*m33*RS146pe*RS345pe
          - 77./6912.*m11*m33*RS146pe*RS346pe - 77./6912.*m11*m33*
         RS146pe*RS356pe + 53./6912.*m11*m33*RS146pe*RS1456pee + 53./
         6912.*m11*m33*RS146pe*RS1456ppe - 35./6912.*m11*m33*RS146pe*
         RSe45611p - 35./6912.*m11*m33*RS146pe*RSp45611e - 53./6912.*
         m11*m33*RS146pe*RS145pe - 53./13824.*m11*m33*pow(RS146pe,2) + 
         43./2304.*m11*m33*RS156pe + 77./3456.*m11*m33*RS156pe*
         RS34561pe );

      F13p6x23R +=  + pi162 * ( 53./3456.*m11*m33*RS156pe*RS14563pe + 
         35./3456.*m11*m33*RS156pe*RSe45613p + 35./3456.*m11*m33*
         RS156pe*RSp45613e + 77./6912.*m11*m33*RS156pe*RS3456pee + 77./
         6912.*m11*m33*RS156pe*RS3456ppe - 35./6912.*m11*m33*RS156pe*
         RSe45633p - 35./6912.*m11*m33*RS156pe*RSp45633e - 77./6912.*
         m11*m33*RS156pe*RS345pe - 77./6912.*m11*m33*RS156pe*RS346pe - 
         77./6912.*m11*m33*RS156pe*RS356pe + 53./6912.*m11*m33*RS156pe*
         RS1456pee + 53./6912.*m11*m33*RS156pe*RS1456ppe - 35./6912.*
         m11*m33*RS156pe*RSe45611p - 35./6912.*m11*m33*RS156pe*
         RSp45611e - 53./6912.*m11*m33*RS156pe*RS145pe - 53./6912.*m11*
         m33*RS156pe*RS146pe - 53./13824.*m11*m33*pow(RS156pe,2) + 5./
         2304.*m11*m33*RS645pe + 5./2304.*m11*m33*RS546pe + 5./2304.*
         m11*m33*RS456pe - 5./2304.*m11*m33*RSe561p - 5./2304.*m11*m33*
         RSe461p - 5./2304.*m11*m33*RSe451p - 5./2304.*m11*m33*RSp561e
          - 5./2304.*m11*m33*RSp461e - 5./2304.*m11*m33*RSp451e + 5./
         2304.*m11*m33*RSe564p );

      F13p6x23R +=  + pi162 * ( 5./2304.*m11*m33*RSe465p + 5./2304.*m11
         *m33*RSe456p + 5./2304.*m11*m33*RSp564e + 5./2304.*m11*m33*
         RSp465e + 5./2304.*m11*m33*RSp456e + 43./512.*pow(m11,2) - 17./
         1536.*pow(m11,2)*RS34561pe - 17./6912.*pow(m11,2)*pow(
         RS34561pe,2) + 5./512.*pow(m11,2)*RS14563pe - 53./3456.*pow(
         m11,2)*RS14563pe*RS34561pe - 137./6912.*pow(m11,2)*pow(
         RS14563pe,2) - 17./1536.*pow(m11,2)*RSe45613p - 17./3456.*pow(
         m11,2)*RSe45613p*RS34561pe - 53./3456.*pow(m11,2)*RSe45613p*
         RS14563pe - 17./6912.*pow(m11,2)*pow(RSe45613p,2) - 17./1536.*
         pow(m11,2)*RSp45613e - 17./3456.*pow(m11,2)*RSp45613e*
         RS34561pe - 53./3456.*pow(m11,2)*RSp45613e*RS14563pe - 17./
         3456.*pow(m11,2)*RSp45613e*RSe45613p - 17./6912.*pow(m11,2)*
         pow(RSp45613e,2) + 1./1536.*pow(m11,2)*RS3456pee - 17./6912.*
         pow(m11,2)*RS3456pee*RS34561pe - 53./6912.*pow(m11,2)*
         RS3456pee*RS14563pe - 17./6912.*pow(m11,2)*RS3456pee*RSe45613p
          - 17./6912.*pow(m11,2)*RS3456pee*RSp45613e );

      F13p6x23R +=  + pi162 * (  - 17./27648.*pow(m11,2)*pow(
         RS3456pee,2) + 1./1536.*pow(m11,2)*RS3456ppe - 17./6912.*pow(
         m11,2)*RS3456ppe*RS34561pe - 53./6912.*pow(m11,2)*RS3456ppe*
         RS14563pe - 17./6912.*pow(m11,2)*RS3456ppe*RSe45613p - 17./
         6912.*pow(m11,2)*RS3456ppe*RSp45613e - 17./13824.*pow(m11,2)*
         RS3456ppe*RS3456pee - 17./27648.*pow(m11,2)*pow(RS3456ppe,2)
          - 1./1536.*pow(m11,2)*RSe45633p + 17./6912.*pow(m11,2)*
         RSe45633p*RS34561pe + 53./6912.*pow(m11,2)*RSe45633p*RS14563pe
          + 17./6912.*pow(m11,2)*RSe45633p*RSe45613p + 17./6912.*pow(
         m11,2)*RSe45633p*RSp45613e + 17./13824.*pow(m11,2)*RSe45633p*
         RS3456pee + 17./13824.*pow(m11,2)*RSe45633p*RS3456ppe - 17./
         27648.*pow(m11,2)*pow(RSe45633p,2) - 1./1536.*pow(m11,2)*
         RSp45633e + 17./6912.*pow(m11,2)*RSp45633e*RS34561pe + 53./
         6912.*pow(m11,2)*RSp45633e*RS14563pe + 17./6912.*pow(m11,2)*
         RSp45633e*RSe45613p + 17./6912.*pow(m11,2)*RSp45633e*RSp45613e
          + 17./13824.*pow(m11,2)*RSp45633e*RS3456pee );

      F13p6x23R +=  + pi162 * ( 17./13824.*pow(m11,2)*RSp45633e*
         RS3456ppe - 17./13824.*pow(m11,2)*RSp45633e*RSe45633p - 17./
         27648.*pow(m11,2)*pow(RSp45633e,2) + 1./288.*pow(m11,2)*
         RS345pe + 17./6912.*pow(m11,2)*RS345pe*RS34561pe + 53./6912.*
         pow(m11,2)*RS345pe*RS14563pe + 17./6912.*pow(m11,2)*RS345pe*
         RSe45613p + 17./6912.*pow(m11,2)*RS345pe*RSp45613e + 17./13824.
         *pow(m11,2)*RS345pe*RS3456pee + 17./13824.*pow(m11,2)*RS345pe*
         RS3456ppe - 17./13824.*pow(m11,2)*RS345pe*RSe45633p - 17./
         13824.*pow(m11,2)*RS345pe*RSp45633e - 17./27648.*pow(m11,2)*
         pow(RS345pe,2) + 1./288.*pow(m11,2)*RS346pe + 17./6912.*pow(
         m11,2)*RS346pe*RS34561pe + 53./6912.*pow(m11,2)*RS346pe*
         RS14563pe + 17./6912.*pow(m11,2)*RS346pe*RSe45613p + 17./6912.
         *pow(m11,2)*RS346pe*RSp45613e + 17./13824.*pow(m11,2)*RS346pe*
         RS3456pee + 17./13824.*pow(m11,2)*RS346pe*RS3456ppe - 17./
         13824.*pow(m11,2)*RS346pe*RSe45633p - 17./13824.*pow(m11,2)*
         RS346pe*RSp45633e );

      F13p6x23R +=  + pi162 * (  - 17./13824.*pow(m11,2)*RS346pe*
         RS345pe - 17./27648.*pow(m11,2)*pow(RS346pe,2) + 1./288.*pow(
         m11,2)*RS356pe + 17./6912.*pow(m11,2)*RS356pe*RS34561pe + 53./
         6912.*pow(m11,2)*RS356pe*RS14563pe + 17./6912.*pow(m11,2)*
         RS356pe*RSe45613p + 17./6912.*pow(m11,2)*RS356pe*RSp45613e + 
         17./13824.*pow(m11,2)*RS356pe*RS3456pee + 17./13824.*pow(
         m11,2)*RS356pe*RS3456ppe - 17./13824.*pow(m11,2)*RS356pe*
         RSe45633p - 17./13824.*pow(m11,2)*RS356pe*RSp45633e - 17./
         13824.*pow(m11,2)*RS356pe*RS345pe - 17./13824.*pow(m11,2)*
         RS356pe*RS346pe - 17./27648.*pow(m11,2)*pow(RS356pe,2) + 19./
         4608.*pow(m11,2)*RSe563p + 19./4608.*pow(m11,2)*RSe463p + 19./
         4608.*pow(m11,2)*RSe453p + 19./4608.*pow(m11,2)*RSp563e + 19./
         4608.*pow(m11,2)*RSp463e + 19./4608.*pow(m11,2)*RSp453e - 37./
         1536.*pow(m11,2)*RS1456pee - 53./6912.*pow(m11,2)*RS1456pee*
         RS34561pe - 137./6912.*pow(m11,2)*RS1456pee*RS14563pe - 53./
         6912.*pow(m11,2)*RS1456pee*RSe45613p );

      F13p6x23R +=  + pi162 * (  - 53./6912.*pow(m11,2)*RS1456pee*
         RSp45613e - 53./13824.*pow(m11,2)*RS1456pee*RS3456pee - 53./
         13824.*pow(m11,2)*RS1456pee*RS3456ppe + 53./13824.*pow(m11,2)*
         RS1456pee*RSe45633p + 53./13824.*pow(m11,2)*RS1456pee*
         RSp45633e + 53./13824.*pow(m11,2)*RS1456pee*RS345pe + 53./
         13824.*pow(m11,2)*RS1456pee*RS346pe + 53./13824.*pow(m11,2)*
         RS1456pee*RS356pe - 137./27648.*pow(m11,2)*pow(RS1456pee,2) - 
         37./1536.*pow(m11,2)*RS1456ppe - 53./6912.*pow(m11,2)*
         RS1456ppe*RS34561pe - 137./6912.*pow(m11,2)*RS1456ppe*
         RS14563pe - 53./6912.*pow(m11,2)*RS1456ppe*RSe45613p - 53./
         6912.*pow(m11,2)*RS1456ppe*RSp45613e - 53./13824.*pow(m11,2)*
         RS1456ppe*RS3456pee - 53./13824.*pow(m11,2)*RS1456ppe*
         RS3456ppe + 53./13824.*pow(m11,2)*RS1456ppe*RSe45633p + 53./
         13824.*pow(m11,2)*RS1456ppe*RSp45633e + 53./13824.*pow(m11,2)*
         RS1456ppe*RS345pe + 53./13824.*pow(m11,2)*RS1456ppe*RS346pe + 
         53./13824.*pow(m11,2)*RS1456ppe*RS356pe );

      F13p6x23R +=  + pi162 * (  - 137./13824.*pow(m11,2)*RS1456ppe*
         RS1456pee - 137./27648.*pow(m11,2)*pow(RS1456ppe,2) + 17./1536.
         *pow(m11,2)*RSe45611p + 17./6912.*pow(m11,2)*RSe45611p*
         RS34561pe + 53./6912.*pow(m11,2)*RSe45611p*RS14563pe + 17./
         6912.*pow(m11,2)*RSe45611p*RSe45613p + 17./6912.*pow(m11,2)*
         RSe45611p*RSp45613e + 17./13824.*pow(m11,2)*RSe45611p*
         RS3456pee + 17./13824.*pow(m11,2)*RSe45611p*RS3456ppe - 17./
         13824.*pow(m11,2)*RSe45611p*RSe45633p - 17./13824.*pow(m11,2)*
         RSe45611p*RSp45633e - 17./13824.*pow(m11,2)*RSe45611p*RS345pe
          - 17./13824.*pow(m11,2)*RSe45611p*RS346pe - 17./13824.*pow(
         m11,2)*RSe45611p*RS356pe + 53./13824.*pow(m11,2)*RSe45611p*
         RS1456pee + 53./13824.*pow(m11,2)*RSe45611p*RS1456ppe - 17./
         27648.*pow(m11,2)*pow(RSe45611p,2) + 17./1536.*pow(m11,2)*
         RSp45611e + 17./6912.*pow(m11,2)*RSp45611e*RS34561pe + 53./
         6912.*pow(m11,2)*RSp45611e*RS14563pe + 17./6912.*pow(m11,2)*
         RSp45611e*RSe45613p );

      F13p6x23R +=  + pi162 * ( 17./6912.*pow(m11,2)*RSp45611e*
         RSp45613e + 17./13824.*pow(m11,2)*RSp45611e*RS3456pee + 17./
         13824.*pow(m11,2)*RSp45611e*RS3456ppe - 17./13824.*pow(m11,2)*
         RSp45611e*RSe45633p - 17./13824.*pow(m11,2)*RSp45611e*
         RSp45633e - 17./13824.*pow(m11,2)*RSp45611e*RS345pe - 17./
         13824.*pow(m11,2)*RSp45611e*RS346pe - 17./13824.*pow(m11,2)*
         RSp45611e*RS356pe + 53./13824.*pow(m11,2)*RSp45611e*RS1456pee
          + 53./13824.*pow(m11,2)*RSp45611e*RS1456ppe - 17./13824.*pow(
         m11,2)*RSp45611e*RSe45611p - 17./27648.*pow(m11,2)*pow(
         RSp45611e,2) - 13./2304.*pow(m11,2)*RS145pe + 53./6912.*pow(
         m11,2)*RS145pe*RS34561pe + 137./6912.*pow(m11,2)*RS145pe*
         RS14563pe + 53./6912.*pow(m11,2)*RS145pe*RSe45613p + 53./6912.
         *pow(m11,2)*RS145pe*RSp45613e + 53./13824.*pow(m11,2)*RS145pe*
         RS3456pee + 53./13824.*pow(m11,2)*RS145pe*RS3456ppe - 53./
         13824.*pow(m11,2)*RS145pe*RSe45633p - 53./13824.*pow(m11,2)*
         RS145pe*RSp45633e );

      F13p6x23R +=  + pi162 * (  - 53./13824.*pow(m11,2)*RS145pe*
         RS345pe - 53./13824.*pow(m11,2)*RS145pe*RS346pe - 53./13824.*
         pow(m11,2)*RS145pe*RS356pe + 137./13824.*pow(m11,2)*RS145pe*
         RS1456pee + 137./13824.*pow(m11,2)*RS145pe*RS1456ppe - 53./
         13824.*pow(m11,2)*RS145pe*RSe45611p - 53./13824.*pow(m11,2)*
         RS145pe*RSp45611e - 137./27648.*pow(m11,2)*pow(RS145pe,2) - 13.
         /2304.*pow(m11,2)*RS146pe + 53./6912.*pow(m11,2)*RS146pe*
         RS34561pe + 137./6912.*pow(m11,2)*RS146pe*RS14563pe + 53./6912.
         *pow(m11,2)*RS146pe*RSe45613p + 53./6912.*pow(m11,2)*RS146pe*
         RSp45613e + 53./13824.*pow(m11,2)*RS146pe*RS3456pee + 53./
         13824.*pow(m11,2)*RS146pe*RS3456ppe - 53./13824.*pow(m11,2)*
         RS146pe*RSe45633p - 53./13824.*pow(m11,2)*RS146pe*RSp45633e - 
         53./13824.*pow(m11,2)*RS146pe*RS345pe - 53./13824.*pow(m11,2)*
         RS146pe*RS346pe - 53./13824.*pow(m11,2)*RS146pe*RS356pe + 137./
         13824.*pow(m11,2)*RS146pe*RS1456pee + 137./13824.*pow(m11,2)*
         RS146pe*RS1456ppe );

      F13p6x23R +=  + pi162 * (  - 53./13824.*pow(m11,2)*RS146pe*
         RSe45611p - 53./13824.*pow(m11,2)*RS146pe*RSp45611e - 137./
         13824.*pow(m11,2)*RS146pe*RS145pe - 137./27648.*pow(m11,2)*
         pow(RS146pe,2) - 13./2304.*pow(m11,2)*RS156pe + 53./6912.*pow(
         m11,2)*RS156pe*RS34561pe + 137./6912.*pow(m11,2)*RS156pe*
         RS14563pe + 53./6912.*pow(m11,2)*RS156pe*RSe45613p + 53./6912.
         *pow(m11,2)*RS156pe*RSp45613e + 53./13824.*pow(m11,2)*RS156pe*
         RS3456pee + 53./13824.*pow(m11,2)*RS156pe*RS3456ppe - 53./
         13824.*pow(m11,2)*RS156pe*RSe45633p - 53./13824.*pow(m11,2)*
         RS156pe*RSp45633e - 53./13824.*pow(m11,2)*RS156pe*RS345pe - 53.
         /13824.*pow(m11,2)*RS156pe*RS346pe - 53./13824.*pow(m11,2)*
         RS156pe*RS356pe + 137./13824.*pow(m11,2)*RS156pe*RS1456pee + 
         137./13824.*pow(m11,2)*RS156pe*RS1456ppe - 53./13824.*pow(
         m11,2)*RS156pe*RSe45611p - 53./13824.*pow(m11,2)*RS156pe*
         RSp45611e - 137./13824.*pow(m11,2)*RS156pe*RS145pe - 137./
         13824.*pow(m11,2)*RS156pe*RS146pe );

      F13p6x23R +=  + pi162 * (  - 137./27648.*pow(m11,2)*pow(
         RS156pe,2) - 1./4608.*pow(m11,2)*RS645pe - 1./4608.*pow(m11,2)
         *RS546pe - 1./4608.*pow(m11,2)*RS456pe - 17./4608.*pow(m11,2)*
         RSe561p - 17./4608.*pow(m11,2)*RSe461p - 17./4608.*pow(m11,2)*
         RSe451p - 17./4608.*pow(m11,2)*RSp561e - 17./4608.*pow(m11,2)*
         RSp461e - 17./4608.*pow(m11,2)*RSp451e - 1./4608.*pow(m11,2)*
         RSe564p - 1./4608.*pow(m11,2)*RSe465p - 1./4608.*pow(m11,2)*
         RSe456p - 1./4608.*pow(m11,2)*RSp564e - 1./4608.*pow(m11,2)*
         RSp465e - 1./4608.*pow(m11,2)*RSp456e );

      F13p6x23R +=  + Ab(m11,mu2)*pi16 * ( 1./72.*RD3456ep*RS14563pe - 
         1./48.*RD1456ep + 1./72.*RD1456ep*RS14563pe + 1./144.*
         RS1456pee*RD3456ep + 1./144.*RS1456pee*RD1456ep + 1./144.*
         RS1456ppe*RD3456ep + 1./144.*RS1456ppe*RD1456ep - 1./144.*
         RS145pe*RD3456ep - 1./144.*RS145pe*RD1456ep - 1./144.*RS146pe*
         RD3456ep - 1./144.*RS146pe*RD1456ep - 1./144.*RS156pe*RD3456ep
          - 1./144.*RS156pe*RD1456ep - 1./36.*mp2*RSp45613e*RS14563pe
          + 1./72.*mp2*RSp45633e*RS14563pe - 1./72.*mp2*RS1456pee*
         RSp45613e + 1./144.*mp2*RS1456pee*RSp45633e - 1./72.*mp2*
         RS1456ppe*RSp45613e + 1./144.*mp2*RS1456ppe*RSp45633e + 1./72.
         *mp2*RSp45611e*RS14563pe + 1./144.*mp2*RSp45611e*RS1456pee + 1.
         /144.*mp2*RSp45611e*RS1456ppe + 1./72.*mp2*RS145pe*RSp45613e
          - 1./144.*mp2*RS145pe*RSp45633e - 1./144.*mp2*RS145pe*
         RSp45611e + 1./72.*mp2*RS146pe*RSp45613e - 1./144.*mp2*RS146pe
         *RSp45633e - 1./144.*mp2*RS146pe*RSp45611e + 1./72.*mp2*
         RS156pe*RSp45613e );

      F13p6x23R +=  + Ab(m11,mu2)*pi16 * (  - 1./144.*mp2*RS156pe*
         RSp45633e - 1./144.*mp2*RS156pe*RSp45611e - 1./36.*me2*
         RSe45613p*RS14563pe + 1./72.*me2*RSe45633p*RS14563pe - 1./72.*
         me2*RS1456pee*RSe45613p + 1./144.*me2*RS1456pee*RSe45633p - 1./
         72.*me2*RS1456ppe*RSe45613p + 1./144.*me2*RS1456ppe*RSe45633p
          + 1./72.*me2*RSe45611p*RS14563pe + 1./144.*me2*RSe45611p*
         RS1456pee + 1./144.*me2*RSe45611p*RS1456ppe + 1./72.*me2*
         RS145pe*RSe45613p - 1./144.*me2*RS145pe*RSe45633p - 1./144.*
         me2*RS145pe*RSe45611p + 1./72.*me2*RS146pe*RSe45613p - 1./144.
         *me2*RS146pe*RSe45633p - 1./144.*me2*RS146pe*RSe45611p + 1./72.
         *me2*RS156pe*RSe45613p - 1./144.*me2*RS156pe*RSe45633p - 1./
         144.*me2*RS156pe*RSe45611p - 1./24.*m66*RS14563pe - 1./48.*m66
         *RS1456pee - 1./48.*m66*RS1456ppe + 1./48.*m66*RS145pe + 1./48.
         *m66*RS146pe + 1./48.*m66*RS156pe - 1./24.*m55*RS14563pe - 1./
         48.*m55*RS1456pee - 1./48.*m55*RS1456ppe + 1./48.*m55*RS145pe
          + 1./48.*m55*RS146pe );

      F13p6x23R +=  + Ab(m11,mu2)*pi16 * ( 1./48.*m55*RS156pe - 1./24.*
         m44*RS14563pe - 1./48.*m44*RS1456pee - 1./48.*m44*RS1456ppe + 
         1./48.*m44*RS145pe + 1./48.*m44*RS146pe + 1./48.*m44*RS156pe
          - 1./16.*m33*RS14563pe - 1./24.*m33*RS14563pe*RS34561pe - 1./
         72.*m33*pow(RS14563pe,2) - 1./72.*m33*RSe45613p*RS14563pe - 1./
         72.*m33*RSp45613e*RS14563pe - 1./48.*m33*RS3456pee*RS14563pe
          - 1./48.*m33*RS3456ppe*RS14563pe + 1./144.*m33*RSe45633p*
         RS14563pe + 1./144.*m33*RSp45633e*RS14563pe + 1./48.*m33*
         RS345pe*RS14563pe + 1./48.*m33*RS346pe*RS14563pe + 1./48.*m33*
         RS356pe*RS14563pe - 1./32.*m33*RS1456pee - 1./48.*m33*
         RS1456pee*RS34561pe - 1./72.*m33*RS1456pee*RS14563pe - 1./144.
         *m33*RS1456pee*RSe45613p - 1./144.*m33*RS1456pee*RSp45613e - 1.
         /96.*m33*RS1456pee*RS3456pee - 1./96.*m33*RS1456pee*RS3456ppe
          + 1./288.*m33*RS1456pee*RSe45633p + 1./288.*m33*RS1456pee*
         RSp45633e + 1./96.*m33*RS1456pee*RS345pe + 1./96.*m33*
         RS1456pee*RS346pe );

      F13p6x23R +=  + Ab(m11,mu2)*pi16 * ( 1./96.*m33*RS1456pee*RS356pe
          - 1./288.*m33*pow(RS1456pee,2) - 1./32.*m33*RS1456ppe - 1./48.
         *m33*RS1456ppe*RS34561pe - 1./72.*m33*RS1456ppe*RS14563pe - 1./
         144.*m33*RS1456ppe*RSe45613p - 1./144.*m33*RS1456ppe*RSp45613e
          - 1./96.*m33*RS1456ppe*RS3456pee - 1./96.*m33*RS1456ppe*
         RS3456ppe + 1./288.*m33*RS1456ppe*RSe45633p + 1./288.*m33*
         RS1456ppe*RSp45633e + 1./96.*m33*RS1456ppe*RS345pe + 1./96.*
         m33*RS1456ppe*RS346pe + 1./96.*m33*RS1456ppe*RS356pe - 1./144.
         *m33*RS1456ppe*RS1456pee - 1./288.*m33*pow(RS1456ppe,2) + 1./
         144.*m33*RSe45611p*RS14563pe + 1./288.*m33*RSe45611p*RS1456pee
          + 1./288.*m33*RSe45611p*RS1456ppe + 1./144.*m33*RSp45611e*
         RS14563pe + 1./288.*m33*RSp45611e*RS1456pee + 1./288.*m33*
         RSp45611e*RS1456ppe + 1./32.*m33*RS145pe + 1./48.*m33*RS145pe*
         RS34561pe + 1./72.*m33*RS145pe*RS14563pe + 1./144.*m33*RS145pe
         *RSe45613p + 1./144.*m33*RS145pe*RSp45613e + 1./96.*m33*
         RS145pe*RS3456pee );

      F13p6x23R +=  + Ab(m11,mu2)*pi16 * ( 1./96.*m33*RS145pe*RS3456ppe
          - 1./288.*m33*RS145pe*RSe45633p - 1./288.*m33*RS145pe*
         RSp45633e - 1./96.*m33*RS145pe*RS345pe - 1./96.*m33*RS145pe*
         RS346pe - 1./96.*m33*RS145pe*RS356pe + 1./144.*m33*RS145pe*
         RS1456pee + 1./144.*m33*RS145pe*RS1456ppe - 1./288.*m33*
         RS145pe*RSe45611p - 1./288.*m33*RS145pe*RSp45611e - 1./288.*
         m33*pow(RS145pe,2) + 1./32.*m33*RS146pe + 1./48.*m33*RS146pe*
         RS34561pe + 1./72.*m33*RS146pe*RS14563pe + 1./144.*m33*RS146pe
         *RSe45613p + 1./144.*m33*RS146pe*RSp45613e + 1./96.*m33*
         RS146pe*RS3456pee + 1./96.*m33*RS146pe*RS3456ppe - 1./288.*m33
         *RS146pe*RSe45633p - 1./288.*m33*RS146pe*RSp45633e - 1./96.*
         m33*RS146pe*RS345pe - 1./96.*m33*RS146pe*RS346pe - 1./96.*m33*
         RS146pe*RS356pe + 1./144.*m33*RS146pe*RS1456pee + 1./144.*m33*
         RS146pe*RS1456ppe - 1./288.*m33*RS146pe*RSe45611p - 1./288.*
         m33*RS146pe*RSp45611e - 1./144.*m33*RS146pe*RS145pe - 1./288.*
         m33*pow(RS146pe,2) );

      F13p6x23R +=  + Ab(m11,mu2)*pi16 * ( 1./32.*m33*RS156pe + 1./48.*
         m33*RS156pe*RS34561pe + 1./72.*m33*RS156pe*RS14563pe + 1./144.
         *m33*RS156pe*RSe45613p + 1./144.*m33*RS156pe*RSp45613e + 1./96.
         *m33*RS156pe*RS3456pee + 1./96.*m33*RS156pe*RS3456ppe - 1./288.
         *m33*RS156pe*RSe45633p - 1./288.*m33*RS156pe*RSp45633e - 1./96.
         *m33*RS156pe*RS345pe - 1./96.*m33*RS156pe*RS346pe - 1./96.*m33
         *RS156pe*RS356pe + 1./144.*m33*RS156pe*RS1456pee + 1./144.*m33
         *RS156pe*RS1456ppe - 1./288.*m33*RS156pe*RSe45611p - 1./288.*
         m33*RS156pe*RSp45611e - 1./144.*m33*RS156pe*RS145pe - 1./144.*
         m33*RS156pe*RS146pe - 1./288.*m33*pow(RS156pe,2) + 1./16.*m11
          - 1./48.*m11*RS14563pe - 1./72.*m11*RS14563pe*RS34561pe - 1./
         24.*m11*pow(RS14563pe,2) - 1./72.*m11*RSe45613p*RS14563pe - 1./
         72.*m11*RSp45613e*RS14563pe - 1./144.*m11*RS3456pee*RS14563pe
          - 1./144.*m11*RS3456ppe*RS14563pe + 1./144.*m11*RSe45633p*
         RS14563pe + 1./144.*m11*RSp45633e*RS14563pe + 1./144.*m11*
         RS345pe*RS14563pe );

      F13p6x23R +=  + Ab(m11,mu2)*pi16 * ( 1./144.*m11*RS346pe*
         RS14563pe + 1./144.*m11*RS356pe*RS14563pe - 1./96.*m11*
         RS1456pee - 1./144.*m11*RS1456pee*RS34561pe - 1./24.*m11*
         RS1456pee*RS14563pe - 1./144.*m11*RS1456pee*RSe45613p - 1./144.
         *m11*RS1456pee*RSp45613e - 1./288.*m11*RS1456pee*RS3456pee - 1.
         /288.*m11*RS1456pee*RS3456ppe + 1./288.*m11*RS1456pee*
         RSe45633p + 1./288.*m11*RS1456pee*RSp45633e + 1./288.*m11*
         RS1456pee*RS345pe + 1./288.*m11*RS1456pee*RS346pe + 1./288.*
         m11*RS1456pee*RS356pe - 1./96.*m11*pow(RS1456pee,2) - 1./96.*
         m11*RS1456ppe - 1./144.*m11*RS1456ppe*RS34561pe - 1./24.*m11*
         RS1456ppe*RS14563pe - 1./144.*m11*RS1456ppe*RSe45613p - 1./144.
         *m11*RS1456ppe*RSp45613e - 1./288.*m11*RS1456ppe*RS3456pee - 1.
         /288.*m11*RS1456ppe*RS3456ppe + 1./288.*m11*RS1456ppe*
         RSe45633p + 1./288.*m11*RS1456ppe*RSp45633e + 1./288.*m11*
         RS1456ppe*RS345pe + 1./288.*m11*RS1456ppe*RS346pe + 1./288.*
         m11*RS1456ppe*RS356pe );

      F13p6x23R +=  + Ab(m11,mu2)*pi16 * (  - 1./48.*m11*RS1456ppe*
         RS1456pee - 1./96.*m11*pow(RS1456ppe,2) + 1./144.*m11*
         RSe45611p*RS14563pe + 1./288.*m11*RSe45611p*RS1456pee + 1./288.
         *m11*RSe45611p*RS1456ppe + 1./144.*m11*RSp45611e*RS14563pe + 1.
         /288.*m11*RSp45611e*RS1456pee + 1./288.*m11*RSp45611e*
         RS1456ppe - 1./96.*m11*RS145pe + 1./144.*m11*RS145pe*RS34561pe
          + 1./24.*m11*RS145pe*RS14563pe + 1./144.*m11*RS145pe*
         RSe45613p + 1./144.*m11*RS145pe*RSp45613e + 1./288.*m11*
         RS145pe*RS3456pee + 1./288.*m11*RS145pe*RS3456ppe - 1./288.*
         m11*RS145pe*RSe45633p - 1./288.*m11*RS145pe*RSp45633e - 1./288.
         *m11*RS145pe*RS345pe - 1./288.*m11*RS145pe*RS346pe - 1./288.*
         m11*RS145pe*RS356pe + 1./48.*m11*RS145pe*RS1456pee + 1./48.*
         m11*RS145pe*RS1456ppe - 1./288.*m11*RS145pe*RSe45611p - 1./288.
         *m11*RS145pe*RSp45611e - 1./96.*m11*pow(RS145pe,2) - 1./96.*
         m11*RS146pe + 1./144.*m11*RS146pe*RS34561pe + 1./24.*m11*
         RS146pe*RS14563pe );

      F13p6x23R +=  + Ab(m11,mu2)*pi16 * ( 1./144.*m11*RS146pe*
         RSe45613p + 1./144.*m11*RS146pe*RSp45613e + 1./288.*m11*
         RS146pe*RS3456pee + 1./288.*m11*RS146pe*RS3456ppe - 1./288.*
         m11*RS146pe*RSe45633p - 1./288.*m11*RS146pe*RSp45633e - 1./288.
         *m11*RS146pe*RS345pe - 1./288.*m11*RS146pe*RS346pe - 1./288.*
         m11*RS146pe*RS356pe + 1./48.*m11*RS146pe*RS1456pee + 1./48.*
         m11*RS146pe*RS1456ppe - 1./288.*m11*RS146pe*RSe45611p - 1./288.
         *m11*RS146pe*RSp45611e - 1./48.*m11*RS146pe*RS145pe - 1./96.*
         m11*pow(RS146pe,2) - 1./96.*m11*RS156pe + 1./144.*m11*RS156pe*
         RS34561pe + 1./24.*m11*RS156pe*RS14563pe + 1./144.*m11*RS156pe
         *RSe45613p + 1./144.*m11*RS156pe*RSp45613e + 1./288.*m11*
         RS156pe*RS3456pee + 1./288.*m11*RS156pe*RS3456ppe - 1./288.*
         m11*RS156pe*RSe45633p - 1./288.*m11*RS156pe*RSp45633e - 1./288.
         *m11*RS156pe*RS345pe - 1./288.*m11*RS156pe*RS346pe - 1./288.*
         m11*RS156pe*RS356pe + 1./48.*m11*RS156pe*RS1456pee + 1./48.*
         m11*RS156pe*RS1456ppe );

      F13p6x23R +=  + Ab(m11,mu2)*pi16 * (  - 1./288.*m11*RS156pe*
         RSe45611p - 1./288.*m11*RS156pe*RSp45611e - 1./48.*m11*RS156pe
         *RS145pe - 1./48.*m11*RS156pe*RS146pe - 1./96.*m11*pow(
         RS156pe,2) );

      F13p6x23R +=  + pow(Ab(m11,mu2),2) * ( 1./16. + 1./72.*pow(
         RS14563pe,2) + 1./72.*RS1456pee*RS14563pe + 1./288.*pow(
         RS1456pee,2) + 1./72.*RS1456ppe*RS14563pe + 1./144.*RS1456ppe*
         RS1456pee + 1./288.*pow(RS1456ppe,2) - 1./72.*RS145pe*
         RS14563pe - 1./144.*RS145pe*RS1456pee - 1./144.*RS145pe*
         RS1456ppe + 1./288.*pow(RS145pe,2) - 1./72.*RS146pe*RS14563pe
          - 1./144.*RS146pe*RS1456pee - 1./144.*RS146pe*RS1456ppe + 1./
         144.*RS146pe*RS145pe + 1./288.*pow(RS146pe,2) - 1./72.*RS156pe
         *RS14563pe - 1./144.*RS156pe*RS1456pee - 1./144.*RS156pe*
         RS1456ppe + 1./144.*RS156pe*RS145pe + 1./144.*RS156pe*RS146pe
          + 1./288.*pow(RS156pe,2) + 1./18.*m11*RN1p*RS1456pee*
         RSp45613e + 1./18.*m11*RN1p*RS1456ppe*RSp45613e - 1./18.*m11*
         RN1p*RSp45611e*RS1456pee - 1./18.*m11*RN1p*RSp45611e*RS1456ppe
          - 1./18.*m11*RN1p*RS145pe*RSp45613e + 1./18.*m11*RN1p*RS145pe
         *RSp45611e - 1./18.*m11*RN1p*RS146pe*RSp45613e + 1./18.*m11*
         RN1p*RS146pe*RSp45611e );

      F13p6x23R +=  + pow(Ab(m11,mu2),2) * (  - 1./18.*m11*RN1p*RS156pe
         *RSp45613e + 1./18.*m11*RN1p*RS156pe*RSp45611e + 1./18.*m11*
         RN1e*RS1456pee*RSe45613p + 1./18.*m11*RN1e*RS1456ppe*RSe45613p
          - 1./18.*m11*RN1e*RSe45611p*RS1456pee - 1./18.*m11*RN1e*
         RSe45611p*RS1456ppe - 1./18.*m11*RN1e*RS145pe*RSe45613p + 1./
         18.*m11*RN1e*RS145pe*RSe45611p - 1./18.*m11*RN1e*RS146pe*
         RSe45613p + 1./18.*m11*RN1e*RS146pe*RSe45611p - 1./18.*m11*
         RN1e*RS156pe*RSe45613p + 1./18.*m11*RN1e*RS156pe*RSe45611p + 1.
         /18.*m11*RN13*RS1456pee*RS34561pe + 1./18.*m11*RN13*RS1456ppe*
         RS34561pe - 1./18.*m11*RN13*RS145pe*RS34561pe - 1./18.*m11*
         RN13*RS146pe*RS34561pe - 1./18.*m11*RN13*RS156pe*RS34561pe );

      F13p6x23R +=  + Ab(m11,mu2)*Ab(m13,mu2) * (  - 1./24. + 1./162.*
         RS14563pe*RS34561pe + 1./162.*pow(RS14563pe,2) + 1./162.*
         RSe45613p*RS14563pe + 1./162.*RSp45613e*RS14563pe + 1./324.*
         RS3456pee*RS14563pe + 1./324.*RS3456ppe*RS14563pe - 1./324.*
         RSe45633p*RS14563pe - 1./324.*RSp45633e*RS14563pe - 1./324.*
         RS345pe*RS14563pe - 1./324.*RS346pe*RS14563pe - 1./324.*
         RS356pe*RS14563pe + 1./324.*RS1456pee*RS34561pe + 1./162.*
         RS1456pee*RS14563pe + 1./324.*RS1456pee*RSe45613p + 1./324.*
         RS1456pee*RSp45613e + 1./648.*RS1456pee*RS3456pee + 1./648.*
         RS1456pee*RS3456ppe - 1./648.*RS1456pee*RSe45633p - 1./648.*
         RS1456pee*RSp45633e - 1./648.*RS1456pee*RS345pe - 1./648.*
         RS1456pee*RS346pe - 1./648.*RS1456pee*RS356pe + 1./648.*pow(
         RS1456pee,2) + 1./324.*RS1456ppe*RS34561pe + 1./162.*RS1456ppe
         *RS14563pe + 1./324.*RS1456ppe*RSe45613p + 1./324.*RS1456ppe*
         RSp45613e + 1./648.*RS1456ppe*RS3456pee + 1./648.*RS1456ppe*
         RS3456ppe );

      F13p6x23R +=  + Ab(m11,mu2)*Ab(m13,mu2) * (  - 1./648.*RS1456ppe*
         RSe45633p - 1./648.*RS1456ppe*RSp45633e - 1./648.*RS1456ppe*
         RS345pe - 1./648.*RS1456ppe*RS346pe - 1./648.*RS1456ppe*
         RS356pe + 1./324.*RS1456ppe*RS1456pee + 1./648.*pow(
         RS1456ppe,2) - 1./324.*RSe45611p*RS14563pe - 1./648.*RSe45611p
         *RS1456pee - 1./648.*RSe45611p*RS1456ppe - 1./324.*RSp45611e*
         RS14563pe - 1./648.*RSp45611e*RS1456pee - 1./648.*RSp45611e*
         RS1456ppe - 1./324.*RS145pe*RS34561pe - 1./162.*RS145pe*
         RS14563pe - 1./324.*RS145pe*RSe45613p - 1./324.*RS145pe*
         RSp45613e - 1./648.*RS145pe*RS3456pee - 1./648.*RS145pe*
         RS3456ppe + 1./648.*RS145pe*RSe45633p + 1./648.*RS145pe*
         RSp45633e + 1./648.*RS145pe*RS345pe + 1./648.*RS145pe*RS346pe
          + 1./648.*RS145pe*RS356pe - 1./324.*RS145pe*RS1456pee - 1./
         324.*RS145pe*RS1456ppe + 1./648.*RS145pe*RSe45611p + 1./648.*
         RS145pe*RSp45611e + 1./648.*pow(RS145pe,2) - 1./324.*RS146pe*
         RS34561pe );

      F13p6x23R +=  + Ab(m11,mu2)*Ab(m13,mu2) * (  - 1./162.*RS146pe*
         RS14563pe - 1./324.*RS146pe*RSe45613p - 1./324.*RS146pe*
         RSp45613e - 1./648.*RS146pe*RS3456pee - 1./648.*RS146pe*
         RS3456ppe + 1./648.*RS146pe*RSe45633p + 1./648.*RS146pe*
         RSp45633e + 1./648.*RS146pe*RS345pe + 1./648.*RS146pe*RS346pe
          + 1./648.*RS146pe*RS356pe - 1./324.*RS146pe*RS1456pee - 1./
         324.*RS146pe*RS1456ppe + 1./648.*RS146pe*RSe45611p + 1./648.*
         RS146pe*RSp45611e + 1./324.*RS146pe*RS145pe + 1./648.*pow(
         RS146pe,2) - 1./324.*RS156pe*RS34561pe - 1./162.*RS156pe*
         RS14563pe - 1./324.*RS156pe*RSe45613p - 1./324.*RS156pe*
         RSp45613e - 1./648.*RS156pe*RS3456pee - 1./648.*RS156pe*
         RS3456ppe + 1./648.*RS156pe*RSe45633p + 1./648.*RS156pe*
         RSp45633e + 1./648.*RS156pe*RS345pe + 1./648.*RS156pe*RS346pe
          + 1./648.*RS156pe*RS356pe - 1./324.*RS156pe*RS1456pee - 1./
         324.*RS156pe*RS1456ppe + 1./648.*RS156pe*RSe45611p + 1./648.*
         RS156pe*RSp45611e );

      F13p6x23R +=  + Ab(m11,mu2)*Ab(m13,mu2) * ( 1./324.*RS156pe*
         RS145pe + 1./324.*RS156pe*RS146pe + 1./648.*pow(RS156pe,2) - 1.
         /6.*m33*RN13 - 1./3.*m11*RN13 );

      F13p6x23R +=  + Ab(m11,mu2)*Ab(m14,mu2) * (  - 1./36.*RS14563pe
          - 1./18.*RSe45613p - 1./18.*RSp45613e + 1./18.*RSe563p + 1./
         18.*RSp563e - 7./144.*RS1456pee - 7./144.*RS1456ppe + 1./18.*
         RSe45611p + 1./18.*RSp45611e + 7./144.*RS145pe + 7./144.*
         RS146pe - 1./48.*RS156pe - 1./18.*RSe561p - 1./18.*RSp561e - 1.
         /18.*mp2*RN1p*RSp45613e + 1./18.*mp2*RN1p*RSp563e + 1./18.*mp2
         *RN1p*RSp45611e - 1./18.*mp2*RN1p*RSp561e - 1./18.*me2*RN1e*
         RSe45613p + 1./18.*me2*RN1e*RSe563p + 1./18.*me2*RN1e*
         RSe45611p - 1./18.*me2*RN1e*RSe561p + 1./18.*m44*RN1p*RSp563e
          - 1./18.*m44*RN1p*RSp561e + 1./18.*m44*RN1e*RSe563p - 1./18.*
         m44*RN1e*RSe561p + 1./18.*m44*RN13*RS356pe + 1./18.*m11*RN1p*
         RSp45613e + 1./18.*m11*RN1p*RSp563e - 1./18.*m11*RN1p*
         RSp45611e - 1./18.*m11*RN1p*RSp561e + 1./18.*m11*RN1e*
         RSe45613p + 1./18.*m11*RN1e*RSe563p - 1./18.*m11*RN1e*
         RSe45611p - 1./18.*m11*RN1e*RSe561p + 1./9.*m11*RN13*RS356pe )
         ;

      F13p6x23R +=  + Ab(m11,mu2)*Ab(m15,mu2) * (  - 1./36.*RS14563pe
          - 1./18.*RSe45613p - 1./18.*RSp45613e + 1./18.*RSe463p + 1./
         18.*RSp463e - 7./144.*RS1456pee - 7./144.*RS1456ppe + 1./18.*
         RSe45611p + 1./18.*RSp45611e + 7./144.*RS145pe - 1./48.*
         RS146pe + 7./144.*RS156pe - 1./18.*RSe461p - 1./18.*RSp461e - 
         1./18.*mp2*RN1p*RSp45613e + 1./18.*mp2*RN1p*RSp463e + 1./18.*
         mp2*RN1p*RSp45611e - 1./18.*mp2*RN1p*RSp461e - 1./18.*me2*RN1e
         *RSe45613p + 1./18.*me2*RN1e*RSe463p + 1./18.*me2*RN1e*
         RSe45611p - 1./18.*me2*RN1e*RSe461p + 1./18.*m55*RN1p*RSp463e
          - 1./18.*m55*RN1p*RSp461e + 1./18.*m55*RN1e*RSe463p - 1./18.*
         m55*RN1e*RSe461p + 1./18.*m55*RN13*RS346pe + 1./18.*m11*RN1p*
         RSp45613e + 1./18.*m11*RN1p*RSp463e - 1./18.*m11*RN1p*
         RSp45611e - 1./18.*m11*RN1p*RSp461e + 1./18.*m11*RN1e*
         RSe45613p + 1./18.*m11*RN1e*RSe463p - 1./18.*m11*RN1e*
         RSe45611p - 1./18.*m11*RN1e*RSe461p + 1./9.*m11*RN13*RS346pe )
         ;

      F13p6x23R +=  + Ab(m11,mu2)*Ab(m16,mu2) * (  - 1./36.*RS14563pe
          - 1./18.*RSe45613p - 1./18.*RSp45613e + 1./18.*RSe453p + 1./
         18.*RSp453e - 7./144.*RS1456pee - 7./144.*RS1456ppe + 1./18.*
         RSe45611p + 1./18.*RSp45611e - 1./48.*RS145pe + 7./144.*
         RS146pe + 7./144.*RS156pe - 1./18.*RSe451p - 1./18.*RSp451e - 
         1./18.*mp2*RN1p*RSp45613e + 1./18.*mp2*RN1p*RSp453e + 1./18.*
         mp2*RN1p*RSp45611e - 1./18.*mp2*RN1p*RSp451e - 1./18.*me2*RN1e
         *RSe45613p + 1./18.*me2*RN1e*RSe453p + 1./18.*me2*RN1e*
         RSe45611p - 1./18.*me2*RN1e*RSe451p + 1./18.*m66*RN1p*RSp453e
          - 1./18.*m66*RN1p*RSp451e + 1./18.*m66*RN1e*RSe453p - 1./18.*
         m66*RN1e*RSe451p + 1./18.*m66*RN13*RS345pe + 1./18.*m11*RN1p*
         RSp45613e + 1./18.*m11*RN1p*RSp453e - 1./18.*m11*RN1p*
         RSp45611e - 1./18.*m11*RN1p*RSp451e + 1./18.*m11*RN1e*
         RSe45613p + 1./18.*m11*RN1e*RSe453p - 1./18.*m11*RN1e*
         RSe45611p - 1./18.*m11*RN1e*RSe451p + 1./9.*m11*RN13*RS345pe )
         ;

      F13p6x23R +=  + Ab(m11,mu2)*Ab(m33,mu2) * ( 1./36.*RS14563pe*
         RS34561pe + 1./72.*RS3456pee*RS14563pe + 1./72.*RS3456ppe*
         RS14563pe - 1./72.*RS345pe*RS14563pe - 1./72.*RS346pe*
         RS14563pe - 1./72.*RS356pe*RS14563pe + 1./72.*RS1456pee*
         RS34561pe + 1./144.*RS1456pee*RS3456pee + 1./144.*RS1456pee*
         RS3456ppe - 1./144.*RS1456pee*RS345pe - 1./144.*RS1456pee*
         RS346pe - 1./144.*RS1456pee*RS356pe + 1./72.*RS1456ppe*
         RS34561pe + 1./144.*RS1456ppe*RS3456pee + 1./144.*RS1456ppe*
         RS3456ppe - 1./144.*RS1456ppe*RS345pe - 1./144.*RS1456ppe*
         RS346pe - 1./144.*RS1456ppe*RS356pe - 1./72.*RS145pe*RS34561pe
          - 1./144.*RS145pe*RS3456pee - 1./144.*RS145pe*RS3456ppe + 1./
         144.*RS145pe*RS345pe + 1./144.*RS145pe*RS346pe + 1./144.*
         RS145pe*RS356pe - 1./72.*RS146pe*RS34561pe - 1./144.*RS146pe*
         RS3456pee - 1./144.*RS146pe*RS3456ppe + 1./144.*RS146pe*
         RS345pe + 1./144.*RS146pe*RS346pe + 1./144.*RS146pe*RS356pe - 
         1./72.*RS156pe*RS34561pe );

      F13p6x23R +=  + Ab(m11,mu2)*Ab(m33,mu2) * (  - 1./144.*RS156pe*
         RS3456pee - 1./144.*RS156pe*RS3456ppe + 1./144.*RS156pe*
         RS345pe + 1./144.*RS156pe*RS346pe + 1./144.*RS156pe*RS356pe + 
         1./18.*m33*RN13*RS3456pee*RS14563pe + 1./18.*m33*RN13*
         RS3456ppe*RS14563pe - 1./18.*m33*RN13*RS345pe*RS14563pe - 1./
         18.*m33*RN13*RS346pe*RS14563pe - 1./18.*m33*RN13*RS356pe*
         RS14563pe - 1./18.*m11*RN13*RS1456pee*RS34561pe - 1./18.*m11*
         RN13*RS1456ppe*RS34561pe + 1./18.*m11*RN13*RS145pe*RS34561pe
          + 1./18.*m11*RN13*RS146pe*RS34561pe + 1./18.*m11*RN13*RS156pe
         *RS34561pe );

      F13p6x23R +=  + Ab(m11,mu2)*Ab(m34,mu2) * ( 1./36.*RS14563pe + 1./
         48.*RS1456pee + 1./48.*RS1456ppe - 1./48.*RS145pe - 1./48.*
         RS146pe - 1./144.*RS156pe + 1./18.*m44*RN13*RS156pe + 1./18.*
         m33*RN13*RS14563pe + 1./18.*m33*RN13*RS156pe - 1./18.*m11*RN13
         *RS14563pe + 1./18.*m11*RN13*RS156pe );

      F13p6x23R +=  + Ab(m11,mu2)*Ab(m35,mu2) * ( 1./36.*RS14563pe + 1./
         48.*RS1456pee + 1./48.*RS1456ppe - 1./48.*RS145pe - 1./144.*
         RS146pe - 1./48.*RS156pe + 1./18.*m55*RN13*RS146pe + 1./18.*
         m33*RN13*RS14563pe + 1./18.*m33*RN13*RS146pe - 1./18.*m11*RN13
         *RS14563pe + 1./18.*m11*RN13*RS146pe );

      F13p6x23R +=  + Ab(m11,mu2)*Ab(m36,mu2) * ( 1./36.*RS14563pe + 1./
         48.*RS1456pee + 1./48.*RS1456ppe - 1./144.*RS145pe - 1./48.*
         RS146pe - 1./48.*RS156pe + 1./18.*m66*RN13*RS145pe + 1./18.*
         m33*RN13*RS14563pe + 1./18.*m33*RN13*RS145pe - 1./18.*m11*RN13
         *RS14563pe + 1./18.*m11*RN13*RS145pe );

      F13p6x23R +=  + Ab(m11,mu2)*Ab(m44,mu2) * (  - 1./18.*m44*RN1p*
         RS156pe*RSp563e + 1./54.*m44*RN1p*RS456pe*RS156pe*RSp563e + 1./
         18.*m44*RN1p*RSp561e*RS156pe - 1./54.*m44*RN1p*RSp561e*RS456pe
         *RS156pe - 1./18.*m44*RN1e*RS156pe*RSe563p + 1./54.*m44*RN1e*
         RS456pe*RS156pe*RSe563p + 1./18.*m44*RN1e*RSe561p*RS156pe - 1./
         54.*m44*RN1e*RSe561p*RS456pe*RS156pe - 1./18.*m44*RN13*RS156pe
         *RS356pe + 1./54.*m44*RN13*RS456pe*RS156pe*RS356pe );

      F13p6x23R +=  + Ab(m11,mu2)*Ab(m45,mu2) * (  - 1./54.*RS146pe*
         RSe563p + 1./54.*RS146pe*RSe463p - 1./54.*RS146pe*RSp563e + 1./
         54.*RS146pe*RSp463e - 1./108.*pow(RS146pe,2) + 1./54.*RS156pe*
         RSe563p - 1./54.*RS156pe*RSe463p + 1./54.*RS156pe*RSp563e - 1./
         54.*RS156pe*RSp463e + 1./54.*RS156pe*RS146pe - 1./108.*pow(
         RS156pe,2) + 1./54.*RSe561p*RS146pe - 1./54.*RSe561p*RS156pe
          - 1./54.*RSe461p*RS146pe + 1./54.*RSe461p*RS156pe + 1./54.*
         RSp561e*RS146pe - 1./54.*RSp561e*RS156pe - 1./54.*RSp461e*
         RS146pe + 1./54.*RSp461e*RS156pe - 1./54.*mp2*RN1p*RS146pe*
         RSp563e + 1./54.*mp2*RN1p*RS146pe*RSp463e + 1./54.*mp2*RN1p*
         RS156pe*RSp563e - 1./54.*mp2*RN1p*RS156pe*RSp463e + 1./54.*mp2
         *RN1p*RSp561e*RS146pe - 1./54.*mp2*RN1p*RSp561e*RS156pe - 1./
         54.*mp2*RN1p*RSp461e*RS146pe + 1./54.*mp2*RN1p*RSp461e*RS156pe
          - 1./54.*me2*RN1e*RS146pe*RSe563p + 1./54.*me2*RN1e*RS146pe*
         RSe463p + 1./54.*me2*RN1e*RS156pe*RSe563p - 1./54.*me2*RN1e*
         RS156pe*RSe463p );

      F13p6x23R +=  + Ab(m11,mu2)*Ab(m45,mu2) * ( 1./54.*me2*RN1e*
         RSe561p*RS146pe - 1./54.*me2*RN1e*RSe561p*RS156pe - 1./54.*me2
         *RN1e*RSe461p*RS146pe + 1./54.*me2*RN1e*RSe461p*RS156pe - 1./
         54.*m55*RN1p*RS146pe*RSp563e - 1./54.*m55*RN1p*RS146pe*RSp463e
          - 1./54.*m55*RN1p*RS156pe*RSp463e + 1./54.*m55*RN1p*RSp561e*
         RS146pe + 1./54.*m55*RN1p*RSp461e*RS146pe + 1./54.*m55*RN1p*
         RSp461e*RS156pe - 1./54.*m55*RN1e*RS146pe*RSe563p - 1./54.*m55
         *RN1e*RS146pe*RSe463p - 1./54.*m55*RN1e*RS156pe*RSe463p + 1./
         54.*m55*RN1e*RSe561p*RS146pe + 1./54.*m55*RN1e*RSe461p*RS146pe
          + 1./54.*m55*RN1e*RSe461p*RS156pe - 1./54.*m55*RN13*RS146pe*
         RS346pe - 1./54.*m55*RN13*RS146pe*RS356pe - 1./54.*m55*RN13*
         RS156pe*RS346pe - 1./54.*m44*RN1p*RS146pe*RSp563e - 1./54.*m44
         *RN1p*RS156pe*RSp563e - 1./54.*m44*RN1p*RS156pe*RSp463e + 1./
         54.*m44*RN1p*RSp561e*RS146pe + 1./54.*m44*RN1p*RSp561e*RS156pe
          + 1./54.*m44*RN1p*RSp461e*RS156pe - 1./54.*m44*RN1e*RS146pe*
         RSe563p );

      F13p6x23R +=  + Ab(m11,mu2)*Ab(m45,mu2) * (  - 1./54.*m44*RN1e*
         RS156pe*RSe563p - 1./54.*m44*RN1e*RS156pe*RSe463p + 1./54.*m44
         *RN1e*RSe561p*RS146pe + 1./54.*m44*RN1e*RSe561p*RS156pe + 1./
         54.*m44*RN1e*RSe461p*RS156pe - 1./54.*m44*RN13*RS146pe*RS356pe
          - 1./54.*m44*RN13*RS156pe*RS346pe - 1./54.*m44*RN13*RS156pe*
         RS356pe + 1./54.*m11*RN13*RS146pe*RS346pe - 1./54.*m11*RN13*
         RS146pe*RS356pe - 1./54.*m11*RN13*RS156pe*RS346pe + 1./54.*m11
         *RN13*RS156pe*RS356pe );

      F13p6x23R +=  + Ab(m11,mu2)*Ab(m46,mu2) * (  - 1./54.*RS145pe*
         RSe563p + 1./54.*RS145pe*RSe453p - 1./54.*RS145pe*RSp563e + 1./
         54.*RS145pe*RSp453e - 1./108.*pow(RS145pe,2) + 1./54.*RS156pe*
         RSe563p - 1./54.*RS156pe*RSe453p + 1./54.*RS156pe*RSp563e - 1./
         54.*RS156pe*RSp453e + 1./54.*RS156pe*RS145pe - 1./108.*pow(
         RS156pe,2) + 1./54.*RSe561p*RS145pe - 1./54.*RSe561p*RS156pe
          - 1./54.*RSe451p*RS145pe + 1./54.*RSe451p*RS156pe + 1./54.*
         RSp561e*RS145pe - 1./54.*RSp561e*RS156pe - 1./54.*RSp451e*
         RS145pe + 1./54.*RSp451e*RS156pe - 1./54.*mp2*RN1p*RS145pe*
         RSp563e + 1./54.*mp2*RN1p*RS145pe*RSp453e + 1./54.*mp2*RN1p*
         RS156pe*RSp563e - 1./54.*mp2*RN1p*RS156pe*RSp453e + 1./54.*mp2
         *RN1p*RSp561e*RS145pe - 1./54.*mp2*RN1p*RSp561e*RS156pe - 1./
         54.*mp2*RN1p*RSp451e*RS145pe + 1./54.*mp2*RN1p*RSp451e*RS156pe
          - 1./54.*me2*RN1e*RS145pe*RSe563p + 1./54.*me2*RN1e*RS145pe*
         RSe453p + 1./54.*me2*RN1e*RS156pe*RSe563p - 1./54.*me2*RN1e*
         RS156pe*RSe453p );

      F13p6x23R +=  + Ab(m11,mu2)*Ab(m46,mu2) * ( 1./54.*me2*RN1e*
         RSe561p*RS145pe - 1./54.*me2*RN1e*RSe561p*RS156pe - 1./54.*me2
         *RN1e*RSe451p*RS145pe + 1./54.*me2*RN1e*RSe451p*RS156pe - 1./
         54.*m66*RN1p*RS145pe*RSp563e - 1./54.*m66*RN1p*RS145pe*RSp453e
          - 1./54.*m66*RN1p*RS156pe*RSp453e + 1./54.*m66*RN1p*RSp561e*
         RS145pe + 1./54.*m66*RN1p*RSp451e*RS145pe + 1./54.*m66*RN1p*
         RSp451e*RS156pe - 1./54.*m66*RN1e*RS145pe*RSe563p - 1./54.*m66
         *RN1e*RS145pe*RSe453p - 1./54.*m66*RN1e*RS156pe*RSe453p + 1./
         54.*m66*RN1e*RSe561p*RS145pe + 1./54.*m66*RN1e*RSe451p*RS145pe
          + 1./54.*m66*RN1e*RSe451p*RS156pe - 1./54.*m66*RN13*RS145pe*
         RS345pe - 1./54.*m66*RN13*RS145pe*RS356pe - 1./54.*m66*RN13*
         RS156pe*RS345pe - 1./54.*m44*RN1p*RS145pe*RSp563e - 1./54.*m44
         *RN1p*RS156pe*RSp563e - 1./54.*m44*RN1p*RS156pe*RSp453e + 1./
         54.*m44*RN1p*RSp561e*RS145pe + 1./54.*m44*RN1p*RSp561e*RS156pe
          + 1./54.*m44*RN1p*RSp451e*RS156pe - 1./54.*m44*RN1e*RS145pe*
         RSe563p );

      F13p6x23R +=  + Ab(m11,mu2)*Ab(m46,mu2) * (  - 1./54.*m44*RN1e*
         RS156pe*RSe563p - 1./54.*m44*RN1e*RS156pe*RSe453p + 1./54.*m44
         *RN1e*RSe561p*RS145pe + 1./54.*m44*RN1e*RSe561p*RS156pe + 1./
         54.*m44*RN1e*RSe451p*RS156pe - 1./54.*m44*RN13*RS145pe*RS356pe
          - 1./54.*m44*RN13*RS156pe*RS345pe - 1./54.*m44*RN13*RS156pe*
         RS356pe + 1./54.*m11*RN13*RS145pe*RS345pe - 1./54.*m11*RN13*
         RS145pe*RS356pe - 1./54.*m11*RN13*RS156pe*RS345pe + 1./54.*m11
         *RN13*RS156pe*RS356pe );

      F13p6x23R +=  + Ab(m11,mu2)*Ab(m55,mu2) * (  - 1./18.*m55*RN1p*
         RS146pe*RSp463e + 1./54.*m55*RN1p*RS546pe*RS146pe*RSp463e + 1./
         18.*m55*RN1p*RSp461e*RS146pe - 1./54.*m55*RN1p*RSp461e*RS546pe
         *RS146pe - 1./18.*m55*RN1e*RS146pe*RSe463p + 1./54.*m55*RN1e*
         RS546pe*RS146pe*RSe463p + 1./18.*m55*RN1e*RSe461p*RS146pe - 1./
         54.*m55*RN1e*RSe461p*RS546pe*RS146pe - 1./18.*m55*RN13*RS146pe
         *RS346pe + 1./54.*m55*RN13*RS546pe*RS146pe*RS346pe );

      F13p6x23R +=  + Ab(m11,mu2)*Ab(m56,mu2) * (  - 1./54.*RS145pe*
         RSe463p + 1./54.*RS145pe*RSe453p - 1./54.*RS145pe*RSp463e + 1./
         54.*RS145pe*RSp453e - 1./108.*pow(RS145pe,2) + 1./54.*RS146pe*
         RSe463p - 1./54.*RS146pe*RSe453p + 1./54.*RS146pe*RSp463e - 1./
         54.*RS146pe*RSp453e + 1./54.*RS146pe*RS145pe - 1./108.*pow(
         RS146pe,2) + 1./54.*RSe461p*RS145pe - 1./54.*RSe461p*RS146pe
          - 1./54.*RSe451p*RS145pe + 1./54.*RSe451p*RS146pe + 1./54.*
         RSp461e*RS145pe - 1./54.*RSp461e*RS146pe - 1./54.*RSp451e*
         RS145pe + 1./54.*RSp451e*RS146pe - 1./54.*mp2*RN1p*RS145pe*
         RSp463e + 1./54.*mp2*RN1p*RS145pe*RSp453e + 1./54.*mp2*RN1p*
         RS146pe*RSp463e - 1./54.*mp2*RN1p*RS146pe*RSp453e + 1./54.*mp2
         *RN1p*RSp461e*RS145pe - 1./54.*mp2*RN1p*RSp461e*RS146pe - 1./
         54.*mp2*RN1p*RSp451e*RS145pe + 1./54.*mp2*RN1p*RSp451e*RS146pe
          - 1./54.*me2*RN1e*RS145pe*RSe463p + 1./54.*me2*RN1e*RS145pe*
         RSe453p + 1./54.*me2*RN1e*RS146pe*RSe463p - 1./54.*me2*RN1e*
         RS146pe*RSe453p );

      F13p6x23R +=  + Ab(m11,mu2)*Ab(m56,mu2) * ( 1./54.*me2*RN1e*
         RSe461p*RS145pe - 1./54.*me2*RN1e*RSe461p*RS146pe - 1./54.*me2
         *RN1e*RSe451p*RS145pe + 1./54.*me2*RN1e*RSe451p*RS146pe - 1./
         54.*m66*RN1p*RS145pe*RSp463e - 1./54.*m66*RN1p*RS145pe*RSp453e
          - 1./54.*m66*RN1p*RS146pe*RSp453e + 1./54.*m66*RN1p*RSp461e*
         RS145pe + 1./54.*m66*RN1p*RSp451e*RS145pe + 1./54.*m66*RN1p*
         RSp451e*RS146pe - 1./54.*m66*RN1e*RS145pe*RSe463p - 1./54.*m66
         *RN1e*RS145pe*RSe453p - 1./54.*m66*RN1e*RS146pe*RSe453p + 1./
         54.*m66*RN1e*RSe461p*RS145pe + 1./54.*m66*RN1e*RSe451p*RS145pe
          + 1./54.*m66*RN1e*RSe451p*RS146pe - 1./54.*m66*RN13*RS145pe*
         RS345pe - 1./54.*m66*RN13*RS145pe*RS346pe - 1./54.*m66*RN13*
         RS146pe*RS345pe - 1./54.*m55*RN1p*RS145pe*RSp463e - 1./54.*m55
         *RN1p*RS146pe*RSp463e - 1./54.*m55*RN1p*RS146pe*RSp453e + 1./
         54.*m55*RN1p*RSp461e*RS145pe + 1./54.*m55*RN1p*RSp461e*RS146pe
          + 1./54.*m55*RN1p*RSp451e*RS146pe - 1./54.*m55*RN1e*RS145pe*
         RSe463p );

      F13p6x23R +=  + Ab(m11,mu2)*Ab(m56,mu2) * (  - 1./54.*m55*RN1e*
         RS146pe*RSe463p - 1./54.*m55*RN1e*RS146pe*RSe453p + 1./54.*m55
         *RN1e*RSe461p*RS145pe + 1./54.*m55*RN1e*RSe461p*RS146pe + 1./
         54.*m55*RN1e*RSe451p*RS146pe - 1./54.*m55*RN13*RS145pe*RS346pe
          - 1./54.*m55*RN13*RS146pe*RS345pe - 1./54.*m55*RN13*RS146pe*
         RS346pe + 1./54.*m11*RN13*RS145pe*RS345pe - 1./54.*m11*RN13*
         RS145pe*RS346pe - 1./54.*m11*RN13*RS146pe*RS345pe + 1./54.*m11
         *RN13*RS146pe*RS346pe );

      F13p6x23R +=  + Ab(m11,mu2)*Ab(m66,mu2) * (  - 1./18.*m66*RN1p*
         RS145pe*RSp453e + 1./54.*m66*RN1p*RS645pe*RS145pe*RSp453e + 1./
         18.*m66*RN1p*RSp451e*RS145pe - 1./54.*m66*RN1p*RSp451e*RS645pe
         *RS145pe - 1./18.*m66*RN1e*RS145pe*RSe453p + 1./54.*m66*RN1e*
         RS645pe*RS145pe*RSe453p + 1./18.*m66*RN1e*RSe451p*RS145pe - 1./
         54.*m66*RN1e*RSe451p*RS645pe*RS145pe - 1./18.*m66*RN13*RS145pe
         *RS345pe + 1./54.*m66*RN13*RS645pe*RS145pe*RS345pe );

      F13p6x23R +=  + Ab(m11,mu2)*Ab(me2,mu2) * ( 1./36.*RSe45613p*
         RS14563pe - 1./72.*RSe45633p*RS14563pe + 1./72.*RS1456pee*
         RSe45613p - 1./144.*RS1456pee*RSe45633p + 1./72.*RS1456ppe*
         RSe45613p - 1./144.*RS1456ppe*RSe45633p - 1./72.*RSe45611p*
         RS14563pe - 1./144.*RSe45611p*RS1456pee - 1./144.*RSe45611p*
         RS1456ppe - 1./72.*RS145pe*RSe45613p + 1./144.*RS145pe*
         RSe45633p + 1./144.*RS145pe*RSe45611p - 1./72.*RS146pe*
         RSe45613p + 1./144.*RS146pe*RSe45633p + 1./144.*RS146pe*
         RSe45611p - 1./72.*RS156pe*RSe45613p + 1./144.*RS156pe*
         RSe45633p + 1./144.*RS156pe*RSe45611p + 1./54.*m66*RN1p*
         RSe456p*RS145pe*RSp453e - 1./54.*m66*RN1p*RSe456p*RSp451e*
         RS145pe + 1./54.*m66*RN1e*RSe456p*RS145pe*RSe453p - 1./54.*m66
         *RN1e*RSe456p*RSe451p*RS145pe + 1./54.*m66*RN13*RSe456p*
         RS145pe*RS345pe + 1./54.*m55*RN1p*RSe465p*RS146pe*RSp463e - 1./
         54.*m55*RN1p*RSe465p*RSp461e*RS146pe + 1./54.*m55*RN1e*RSe465p
         *RS146pe*RSe463p );

      F13p6x23R +=  + Ab(m11,mu2)*Ab(me2,mu2) * (  - 1./54.*m55*RN1e*
         RSe465p*RSe461p*RS146pe + 1./54.*m55*RN13*RSe465p*RS146pe*
         RS346pe + 1./54.*m44*RN1p*RSe564p*RS156pe*RSp563e - 1./54.*m44
         *RN1p*RSe564p*RSp561e*RS156pe + 1./54.*m44*RN1e*RSe564p*
         RS156pe*RSe563p - 1./54.*m44*RN1e*RSe564p*RSe561p*RS156pe + 1./
         54.*m44*RN13*RSe564p*RS156pe*RS356pe - 1./18.*m33*RN13*
         RSe45633p*RS14563pe - 1./18.*m11*RN1p*RSe45611p*RSp45613e + 1./
         18.*m11*RN1p*RSp45611e*RSe45611p - 1./18.*m11*RN1e*RS1456pee*
         RSe45613p - 1./18.*m11*RN1e*RS1456ppe*RSe45613p - 1./18.*m11*
         RN1e*RSe45611p*RSe45613p + 1./18.*m11*RN1e*RSe45611p*RS1456pee
          + 1./18.*m11*RN1e*RSe45611p*RS1456ppe + 1./18.*m11*RN1e*pow(
         RSe45611p,2) + 1./18.*m11*RN1e*RS145pe*RSe45613p - 1./18.*m11*
         RN1e*RS145pe*RSe45611p + 1./18.*m11*RN1e*RS146pe*RSe45613p - 1.
         /18.*m11*RN1e*RS146pe*RSe45611p + 1./18.*m11*RN1e*RS156pe*
         RSe45613p - 1./18.*m11*RN1e*RS156pe*RSe45611p - 1./18.*m11*
         RN13*RSe45611p*RS34561pe );

      F13p6x23R +=  + Ab(m11,mu2)*Ab(mp2,mu2) * ( 1./36.*RSp45613e*
         RS14563pe - 1./72.*RSp45633e*RS14563pe + 1./72.*RS1456pee*
         RSp45613e - 1./144.*RS1456pee*RSp45633e + 1./72.*RS1456ppe*
         RSp45613e - 1./144.*RS1456ppe*RSp45633e - 1./72.*RSp45611e*
         RS14563pe - 1./144.*RSp45611e*RS1456pee - 1./144.*RSp45611e*
         RS1456ppe - 1./72.*RS145pe*RSp45613e + 1./144.*RS145pe*
         RSp45633e + 1./144.*RS145pe*RSp45611e - 1./72.*RS146pe*
         RSp45613e + 1./144.*RS146pe*RSp45633e + 1./144.*RS146pe*
         RSp45611e - 1./72.*RS156pe*RSp45613e + 1./144.*RS156pe*
         RSp45633e + 1./144.*RS156pe*RSp45611e + 1./54.*m66*RN1p*
         RSp456e*RS145pe*RSp453e - 1./54.*m66*RN1p*RSp456e*RSp451e*
         RS145pe + 1./54.*m66*RN1e*RSp456e*RS145pe*RSe453p - 1./54.*m66
         *RN1e*RSp456e*RSe451p*RS145pe + 1./54.*m66*RN13*RSp456e*
         RS145pe*RS345pe + 1./54.*m55*RN1p*RSp465e*RS146pe*RSp463e - 1./
         54.*m55*RN1p*RSp465e*RSp461e*RS146pe + 1./54.*m55*RN1e*RSp465e
         *RS146pe*RSe463p );

      F13p6x23R +=  + Ab(m11,mu2)*Ab(mp2,mu2) * (  - 1./54.*m55*RN1e*
         RSp465e*RSe461p*RS146pe + 1./54.*m55*RN13*RSp465e*RS146pe*
         RS346pe + 1./54.*m44*RN1p*RSp564e*RS156pe*RSp563e - 1./54.*m44
         *RN1p*RSp564e*RSp561e*RS156pe + 1./54.*m44*RN1e*RSp564e*
         RS156pe*RSe563p - 1./54.*m44*RN1e*RSp564e*RSe561p*RS156pe + 1./
         54.*m44*RN13*RSp564e*RS156pe*RS356pe - 1./18.*m33*RN13*
         RSp45633e*RS14563pe - 1./18.*m11*RN1p*RS1456pee*RSp45613e - 1./
         18.*m11*RN1p*RS1456ppe*RSp45613e - 1./18.*m11*RN1p*RSp45611e*
         RSp45613e + 1./18.*m11*RN1p*RSp45611e*RS1456pee + 1./18.*m11*
         RN1p*RSp45611e*RS1456ppe + 1./18.*m11*RN1p*pow(RSp45611e,2) + 
         1./18.*m11*RN1p*RS145pe*RSp45613e - 1./18.*m11*RN1p*RS145pe*
         RSp45611e + 1./18.*m11*RN1p*RS146pe*RSp45613e - 1./18.*m11*
         RN1p*RS146pe*RSp45611e + 1./18.*m11*RN1p*RS156pe*RSp45613e - 1.
         /18.*m11*RN1p*RS156pe*RSp45611e - 1./18.*m11*RN1e*RSp45611e*
         RSe45613p + 1./18.*m11*RN1e*RSp45611e*RSe45611p - 1./18.*m11*
         RN13*RSp45611e*RS34561pe );

      F13p6x23R +=  + Ab(m11,mu2)*Bb(m11,mu2) * (  - 1./72.*RD1456ep*
         RS14563pe - 1./144.*RS1456pee*RD1456ep - 1./144.*RS1456ppe*
         RD1456ep + 1./144.*RS145pe*RD1456ep + 1./144.*RS146pe*RD1456ep
          + 1./144.*RS156pe*RD1456ep + 1./4.*m11 + 1./18.*m11*RS1456pee
         *RS14563pe + 1./18.*m11*pow(RS1456pee,2) + 1./18.*m11*
         RS1456ppe*RS14563pe + 1./9.*m11*RS1456ppe*RS1456pee + 1./18.*
         m11*pow(RS1456ppe,2) - 1./18.*m11*RS145pe*RS14563pe - 1./9.*
         m11*RS145pe*RS1456pee - 1./9.*m11*RS145pe*RS1456ppe + 1./18.*
         m11*pow(RS145pe,2) - 1./18.*m11*RS146pe*RS14563pe - 1./9.*m11*
         RS146pe*RS1456pee - 1./9.*m11*RS146pe*RS1456ppe + 1./9.*m11*
         RS146pe*RS145pe + 1./18.*m11*pow(RS146pe,2) - 1./18.*m11*
         RS156pe*RS14563pe - 1./9.*m11*RS156pe*RS1456pee - 1./9.*m11*
         RS156pe*RS1456ppe + 1./9.*m11*RS156pe*RS145pe + 1./9.*m11*
         RS156pe*RS146pe + 1./18.*m11*pow(RS156pe,2) - 1./18.*m11*RN1p*
         RD1456ep*RSp45613e + 1./18.*m11*RN1p*RSp45611e*RD1456ep - 1./
         18.*m11*RN1e*RD1456ep*RSe45613p );

      F13p6x23R +=  + Ab(m11,mu2)*Bb(m11,mu2) * ( 1./18.*m11*RN1e*
         RSe45611p*RD1456ep - 1./18.*m11*RN13*RD1456ep*RS34561pe );

      F13p6x23R +=  + Ab(m11,mu2)*Bb(m14,mu2) * ( 1./72.*RD1456ep - 1./
         36.*m44*RS156pe - 1./18.*m11*RS156pe );

      F13p6x23R +=  + Ab(m11,mu2)*Bb(m15,mu2) * ( 1./72.*RD1456ep - 1./
         36.*m55*RS146pe - 1./18.*m11*RS146pe );

      F13p6x23R +=  + Ab(m11,mu2)*Bb(m16,mu2) * ( 1./72.*RD1456ep - 1./
         36.*m66*RS145pe - 1./18.*m11*RS145pe );

      F13p6x23R +=  + Ab(m11,mu2)*Bb(m33,mu2) * (  - 1./72.*RD3456ep*
         RS14563pe - 1./144.*RS1456pee*RD3456ep - 1./144.*RS1456ppe*
         RD3456ep + 1./144.*RS145pe*RD3456ep + 1./144.*RS146pe*RD3456ep
          + 1./144.*RS156pe*RD3456ep - 1./18.*m33*RN13*RD3456ep*
         RS14563pe );

      F13p6x23R +=  + Ab(m11,mu2)*Cb(m11,mu2) * (  - 1./18.*m11*
         RS1456pee*RD1456ep - 1./18.*m11*RS1456ppe*RD1456ep + 1./18.*
         m11*RS145pe*RD1456ep + 1./18.*m11*RS146pe*RD1456ep + 1./18.*
         m11*RS156pe*RD1456ep );

      F13p6x23R +=  + Ab(m13,mu2)*pi16 * (  - 1./144.*m33*pow(
         RS34561pe,2) - 1./72.*m33*RS14563pe*RS34561pe - 1./144.*m33*
         pow(RS14563pe,2) - 1./72.*m33*RSe45613p*RS34561pe - 1./72.*m33
         *RSe45613p*RS14563pe - 1./144.*m33*pow(RSe45613p,2) - 1./72.*
         m33*RSp45613e*RS34561pe - 1./72.*m33*RSp45613e*RS14563pe - 1./
         72.*m33*RSp45613e*RSe45613p - 1./144.*m33*pow(RSp45613e,2) - 1.
         /144.*m33*RS3456pee*RS34561pe - 1./144.*m33*RS3456pee*
         RS14563pe - 1./144.*m33*RS3456pee*RSe45613p - 1./144.*m33*
         RS3456pee*RSp45613e - 1./576.*m33*pow(RS3456pee,2) - 1./144.*
         m33*RS3456ppe*RS34561pe - 1./144.*m33*RS3456ppe*RS14563pe - 1./
         144.*m33*RS3456ppe*RSe45613p - 1./144.*m33*RS3456ppe*RSp45613e
          - 1./288.*m33*RS3456ppe*RS3456pee - 1./576.*m33*pow(
         RS3456ppe,2) + 1./144.*m33*RSe45633p*RS34561pe + 1./144.*m33*
         RSe45633p*RS14563pe + 1./144.*m33*RSe45633p*RSe45613p + 1./144.
         *m33*RSe45633p*RSp45613e + 1./288.*m33*RSe45633p*RS3456pee + 1.
         /288.*m33*RSe45633p*RS3456ppe );

      F13p6x23R +=  + Ab(m13,mu2)*pi16 * (  - 1./576.*m33*pow(
         RSe45633p,2) + 1./144.*m33*RSp45633e*RS34561pe + 1./144.*m33*
         RSp45633e*RS14563pe + 1./144.*m33*RSp45633e*RSe45613p + 1./144.
         *m33*RSp45633e*RSp45613e + 1./288.*m33*RSp45633e*RS3456pee + 1.
         /288.*m33*RSp45633e*RS3456ppe - 1./288.*m33*RSp45633e*
         RSe45633p - 1./576.*m33*pow(RSp45633e,2) + 1./144.*m33*RS345pe
         *RS34561pe + 1./144.*m33*RS345pe*RS14563pe + 1./144.*m33*
         RS345pe*RSe45613p + 1./144.*m33*RS345pe*RSp45613e + 1./288.*
         m33*RS345pe*RS3456pee + 1./288.*m33*RS345pe*RS3456ppe - 1./288.
         *m33*RS345pe*RSe45633p - 1./288.*m33*RS345pe*RSp45633e - 1./
         576.*m33*pow(RS345pe,2) + 1./144.*m33*RS346pe*RS34561pe + 1./
         144.*m33*RS346pe*RS14563pe + 1./144.*m33*RS346pe*RSe45613p + 1.
         /144.*m33*RS346pe*RSp45613e + 1./288.*m33*RS346pe*RS3456pee + 
         1./288.*m33*RS346pe*RS3456ppe - 1./288.*m33*RS346pe*RSe45633p
          - 1./288.*m33*RS346pe*RSp45633e - 1./288.*m33*RS346pe*RS345pe
          - 1./576.*m33*pow(RS346pe,2) );

      F13p6x23R +=  + Ab(m13,mu2)*pi16 * ( 1./144.*m33*RS356pe*
         RS34561pe + 1./144.*m33*RS356pe*RS14563pe + 1./144.*m33*
         RS356pe*RSe45613p + 1./144.*m33*RS356pe*RSp45613e + 1./288.*
         m33*RS356pe*RS3456pee + 1./288.*m33*RS356pe*RS3456ppe - 1./288.
         *m33*RS356pe*RSe45633p - 1./288.*m33*RS356pe*RSp45633e - 1./
         288.*m33*RS356pe*RS345pe - 1./288.*m33*RS356pe*RS346pe - 1./
         576.*m33*pow(RS356pe,2) - 1./144.*m33*RS1456pee*RS34561pe - 1./
         144.*m33*RS1456pee*RS14563pe - 1./144.*m33*RS1456pee*RSe45613p
          - 1./144.*m33*RS1456pee*RSp45613e - 1./288.*m33*RS1456pee*
         RS3456pee - 1./288.*m33*RS1456pee*RS3456ppe + 1./288.*m33*
         RS1456pee*RSe45633p + 1./288.*m33*RS1456pee*RSp45633e + 1./288.
         *m33*RS1456pee*RS345pe + 1./288.*m33*RS1456pee*RS346pe + 1./
         288.*m33*RS1456pee*RS356pe - 1./576.*m33*pow(RS1456pee,2) - 1./
         144.*m33*RS1456ppe*RS34561pe - 1./144.*m33*RS1456ppe*RS14563pe
          - 1./144.*m33*RS1456ppe*RSe45613p - 1./144.*m33*RS1456ppe*
         RSp45613e );

      F13p6x23R +=  + Ab(m13,mu2)*pi16 * (  - 1./288.*m33*RS1456ppe*
         RS3456pee - 1./288.*m33*RS1456ppe*RS3456ppe + 1./288.*m33*
         RS1456ppe*RSe45633p + 1./288.*m33*RS1456ppe*RSp45633e + 1./288.
         *m33*RS1456ppe*RS345pe + 1./288.*m33*RS1456ppe*RS346pe + 1./
         288.*m33*RS1456ppe*RS356pe - 1./288.*m33*RS1456ppe*RS1456pee
          - 1./576.*m33*pow(RS1456ppe,2) + 1./144.*m33*RSe45611p*
         RS34561pe + 1./144.*m33*RSe45611p*RS14563pe + 1./144.*m33*
         RSe45611p*RSe45613p + 1./144.*m33*RSe45611p*RSp45613e + 1./288.
         *m33*RSe45611p*RS3456pee + 1./288.*m33*RSe45611p*RS3456ppe - 1.
         /288.*m33*RSe45611p*RSe45633p - 1./288.*m33*RSe45611p*
         RSp45633e - 1./288.*m33*RSe45611p*RS345pe - 1./288.*m33*
         RSe45611p*RS346pe - 1./288.*m33*RSe45611p*RS356pe + 1./288.*
         m33*RSe45611p*RS1456pee + 1./288.*m33*RSe45611p*RS1456ppe - 1./
         576.*m33*pow(RSe45611p,2) + 1./144.*m33*RSp45611e*RS34561pe + 
         1./144.*m33*RSp45611e*RS14563pe + 1./144.*m33*RSp45611e*
         RSe45613p );

      F13p6x23R +=  + Ab(m13,mu2)*pi16 * ( 1./144.*m33*RSp45611e*
         RSp45613e + 1./288.*m33*RSp45611e*RS3456pee + 1./288.*m33*
         RSp45611e*RS3456ppe - 1./288.*m33*RSp45611e*RSe45633p - 1./288.
         *m33*RSp45611e*RSp45633e - 1./288.*m33*RSp45611e*RS345pe - 1./
         288.*m33*RSp45611e*RS346pe - 1./288.*m33*RSp45611e*RS356pe + 1.
         /288.*m33*RSp45611e*RS1456pee + 1./288.*m33*RSp45611e*
         RS1456ppe - 1./288.*m33*RSp45611e*RSe45611p - 1./576.*m33*pow(
         RSp45611e,2) + 1./144.*m33*RS145pe*RS34561pe + 1./144.*m33*
         RS145pe*RS14563pe + 1./144.*m33*RS145pe*RSe45613p + 1./144.*
         m33*RS145pe*RSp45613e + 1./288.*m33*RS145pe*RS3456pee + 1./288.
         *m33*RS145pe*RS3456ppe - 1./288.*m33*RS145pe*RSe45633p - 1./
         288.*m33*RS145pe*RSp45633e - 1./288.*m33*RS145pe*RS345pe - 1./
         288.*m33*RS145pe*RS346pe - 1./288.*m33*RS145pe*RS356pe + 1./
         288.*m33*RS145pe*RS1456pee + 1./288.*m33*RS145pe*RS1456ppe - 1.
         /288.*m33*RS145pe*RSe45611p - 1./288.*m33*RS145pe*RSp45611e - 
         1./576.*m33*pow(RS145pe,2) );

      F13p6x23R +=  + Ab(m13,mu2)*pi16 * ( 1./144.*m33*RS146pe*
         RS34561pe + 1./144.*m33*RS146pe*RS14563pe + 1./144.*m33*
         RS146pe*RSe45613p + 1./144.*m33*RS146pe*RSp45613e + 1./288.*
         m33*RS146pe*RS3456pee + 1./288.*m33*RS146pe*RS3456ppe - 1./288.
         *m33*RS146pe*RSe45633p - 1./288.*m33*RS146pe*RSp45633e - 1./
         288.*m33*RS146pe*RS345pe - 1./288.*m33*RS146pe*RS346pe - 1./
         288.*m33*RS146pe*RS356pe + 1./288.*m33*RS146pe*RS1456pee + 1./
         288.*m33*RS146pe*RS1456ppe - 1./288.*m33*RS146pe*RSe45611p - 1.
         /288.*m33*RS146pe*RSp45611e - 1./288.*m33*RS146pe*RS145pe - 1./
         576.*m33*pow(RS146pe,2) + 1./144.*m33*RS156pe*RS34561pe + 1./
         144.*m33*RS156pe*RS14563pe + 1./144.*m33*RS156pe*RSe45613p + 1.
         /144.*m33*RS156pe*RSp45613e + 1./288.*m33*RS156pe*RS3456pee + 
         1./288.*m33*RS156pe*RS3456ppe - 1./288.*m33*RS156pe*RSe45633p
          - 1./288.*m33*RS156pe*RSp45633e - 1./288.*m33*RS156pe*RS345pe
          - 1./288.*m33*RS156pe*RS346pe - 1./288.*m33*RS156pe*RS356pe
          + 1./288.*m33*RS156pe*RS1456pee );

      F13p6x23R +=  + Ab(m13,mu2)*pi16 * ( 1./288.*m33*RS156pe*
         RS1456ppe - 1./288.*m33*RS156pe*RSe45611p - 1./288.*m33*
         RS156pe*RSp45611e - 1./288.*m33*RS156pe*RS145pe - 1./288.*m33*
         RS156pe*RS146pe - 1./576.*m33*pow(RS156pe,2) - 1./144.*m11*
         pow(RS34561pe,2) - 1./72.*m11*RS14563pe*RS34561pe - 1./144.*
         m11*pow(RS14563pe,2) - 1./72.*m11*RSe45613p*RS34561pe - 1./72.
         *m11*RSe45613p*RS14563pe - 1./144.*m11*pow(RSe45613p,2) - 1./
         72.*m11*RSp45613e*RS34561pe - 1./72.*m11*RSp45613e*RS14563pe
          - 1./72.*m11*RSp45613e*RSe45613p - 1./144.*m11*pow(
         RSp45613e,2) - 1./144.*m11*RS3456pee*RS34561pe - 1./144.*m11*
         RS3456pee*RS14563pe - 1./144.*m11*RS3456pee*RSe45613p - 1./144.
         *m11*RS3456pee*RSp45613e - 1./576.*m11*pow(RS3456pee,2) - 1./
         144.*m11*RS3456ppe*RS34561pe - 1./144.*m11*RS3456ppe*RS14563pe
          - 1./144.*m11*RS3456ppe*RSe45613p - 1./144.*m11*RS3456ppe*
         RSp45613e - 1./288.*m11*RS3456ppe*RS3456pee - 1./576.*m11*pow(
         RS3456ppe,2) );

      F13p6x23R +=  + Ab(m13,mu2)*pi16 * ( 1./144.*m11*RSe45633p*
         RS34561pe + 1./144.*m11*RSe45633p*RS14563pe + 1./144.*m11*
         RSe45633p*RSe45613p + 1./144.*m11*RSe45633p*RSp45613e + 1./288.
         *m11*RSe45633p*RS3456pee + 1./288.*m11*RSe45633p*RS3456ppe - 1.
         /576.*m11*pow(RSe45633p,2) + 1./144.*m11*RSp45633e*RS34561pe
          + 1./144.*m11*RSp45633e*RS14563pe + 1./144.*m11*RSp45633e*
         RSe45613p + 1./144.*m11*RSp45633e*RSp45613e + 1./288.*m11*
         RSp45633e*RS3456pee + 1./288.*m11*RSp45633e*RS3456ppe - 1./288.
         *m11*RSp45633e*RSe45633p - 1./576.*m11*pow(RSp45633e,2) + 1./
         144.*m11*RS345pe*RS34561pe + 1./144.*m11*RS345pe*RS14563pe + 1.
         /144.*m11*RS345pe*RSe45613p + 1./144.*m11*RS345pe*RSp45613e + 
         1./288.*m11*RS345pe*RS3456pee + 1./288.*m11*RS345pe*RS3456ppe
          - 1./288.*m11*RS345pe*RSe45633p - 1./288.*m11*RS345pe*
         RSp45633e - 1./576.*m11*pow(RS345pe,2) + 1./144.*m11*RS346pe*
         RS34561pe + 1./144.*m11*RS346pe*RS14563pe + 1./144.*m11*
         RS346pe*RSe45613p );

      F13p6x23R +=  + Ab(m13,mu2)*pi16 * ( 1./144.*m11*RS346pe*
         RSp45613e + 1./288.*m11*RS346pe*RS3456pee + 1./288.*m11*
         RS346pe*RS3456ppe - 1./288.*m11*RS346pe*RSe45633p - 1./288.*
         m11*RS346pe*RSp45633e - 1./288.*m11*RS346pe*RS345pe - 1./576.*
         m11*pow(RS346pe,2) + 1./144.*m11*RS356pe*RS34561pe + 1./144.*
         m11*RS356pe*RS14563pe + 1./144.*m11*RS356pe*RSe45613p + 1./144.
         *m11*RS356pe*RSp45613e + 1./288.*m11*RS356pe*RS3456pee + 1./
         288.*m11*RS356pe*RS3456ppe - 1./288.*m11*RS356pe*RSe45633p - 1.
         /288.*m11*RS356pe*RSp45633e - 1./288.*m11*RS356pe*RS345pe - 1./
         288.*m11*RS356pe*RS346pe - 1./576.*m11*pow(RS356pe,2) - 1./144.
         *m11*RS1456pee*RS34561pe - 1./144.*m11*RS1456pee*RS14563pe - 1.
         /144.*m11*RS1456pee*RSe45613p - 1./144.*m11*RS1456pee*
         RSp45613e - 1./288.*m11*RS1456pee*RS3456pee - 1./288.*m11*
         RS1456pee*RS3456ppe + 1./288.*m11*RS1456pee*RSe45633p + 1./288.
         *m11*RS1456pee*RSp45633e + 1./288.*m11*RS1456pee*RS345pe + 1./
         288.*m11*RS1456pee*RS346pe );

      F13p6x23R +=  + Ab(m13,mu2)*pi16 * ( 1./288.*m11*RS1456pee*
         RS356pe - 1./576.*m11*pow(RS1456pee,2) - 1./144.*m11*RS1456ppe
         *RS34561pe - 1./144.*m11*RS1456ppe*RS14563pe - 1./144.*m11*
         RS1456ppe*RSe45613p - 1./144.*m11*RS1456ppe*RSp45613e - 1./288.
         *m11*RS1456ppe*RS3456pee - 1./288.*m11*RS1456ppe*RS3456ppe + 1.
         /288.*m11*RS1456ppe*RSe45633p + 1./288.*m11*RS1456ppe*
         RSp45633e + 1./288.*m11*RS1456ppe*RS345pe + 1./288.*m11*
         RS1456ppe*RS346pe + 1./288.*m11*RS1456ppe*RS356pe - 1./288.*
         m11*RS1456ppe*RS1456pee - 1./576.*m11*pow(RS1456ppe,2) + 1./
         144.*m11*RSe45611p*RS34561pe + 1./144.*m11*RSe45611p*RS14563pe
          + 1./144.*m11*RSe45611p*RSe45613p + 1./144.*m11*RSe45611p*
         RSp45613e + 1./288.*m11*RSe45611p*RS3456pee + 1./288.*m11*
         RSe45611p*RS3456ppe - 1./288.*m11*RSe45611p*RSe45633p - 1./288.
         *m11*RSe45611p*RSp45633e - 1./288.*m11*RSe45611p*RS345pe - 1./
         288.*m11*RSe45611p*RS346pe - 1./288.*m11*RSe45611p*RS356pe + 1.
         /288.*m11*RSe45611p*RS1456pee );

      F13p6x23R +=  + Ab(m13,mu2)*pi16 * ( 1./288.*m11*RSe45611p*
         RS1456ppe - 1./576.*m11*pow(RSe45611p,2) + 1./144.*m11*
         RSp45611e*RS34561pe + 1./144.*m11*RSp45611e*RS14563pe + 1./144.
         *m11*RSp45611e*RSe45613p + 1./144.*m11*RSp45611e*RSp45613e + 1.
         /288.*m11*RSp45611e*RS3456pee + 1./288.*m11*RSp45611e*
         RS3456ppe - 1./288.*m11*RSp45611e*RSe45633p - 1./288.*m11*
         RSp45611e*RSp45633e - 1./288.*m11*RSp45611e*RS345pe - 1./288.*
         m11*RSp45611e*RS346pe - 1./288.*m11*RSp45611e*RS356pe + 1./288.
         *m11*RSp45611e*RS1456pee + 1./288.*m11*RSp45611e*RS1456ppe - 1.
         /288.*m11*RSp45611e*RSe45611p - 1./576.*m11*pow(RSp45611e,2)
          + 1./144.*m11*RS145pe*RS34561pe + 1./144.*m11*RS145pe*
         RS14563pe + 1./144.*m11*RS145pe*RSe45613p + 1./144.*m11*
         RS145pe*RSp45613e + 1./288.*m11*RS145pe*RS3456pee + 1./288.*
         m11*RS145pe*RS3456ppe - 1./288.*m11*RS145pe*RSe45633p - 1./288.
         *m11*RS145pe*RSp45633e - 1./288.*m11*RS145pe*RS345pe - 1./288.
         *m11*RS145pe*RS346pe );

      F13p6x23R +=  + Ab(m13,mu2)*pi16 * (  - 1./288.*m11*RS145pe*
         RS356pe + 1./288.*m11*RS145pe*RS1456pee + 1./288.*m11*RS145pe*
         RS1456ppe - 1./288.*m11*RS145pe*RSe45611p - 1./288.*m11*
         RS145pe*RSp45611e - 1./576.*m11*pow(RS145pe,2) + 1./144.*m11*
         RS146pe*RS34561pe + 1./144.*m11*RS146pe*RS14563pe + 1./144.*
         m11*RS146pe*RSe45613p + 1./144.*m11*RS146pe*RSp45613e + 1./288.
         *m11*RS146pe*RS3456pee + 1./288.*m11*RS146pe*RS3456ppe - 1./
         288.*m11*RS146pe*RSe45633p - 1./288.*m11*RS146pe*RSp45633e - 1.
         /288.*m11*RS146pe*RS345pe - 1./288.*m11*RS146pe*RS346pe - 1./
         288.*m11*RS146pe*RS356pe + 1./288.*m11*RS146pe*RS1456pee + 1./
         288.*m11*RS146pe*RS1456ppe - 1./288.*m11*RS146pe*RSe45611p - 1.
         /288.*m11*RS146pe*RSp45611e - 1./288.*m11*RS146pe*RS145pe - 1./
         576.*m11*pow(RS146pe,2) + 1./144.*m11*RS156pe*RS34561pe + 1./
         144.*m11*RS156pe*RS14563pe + 1./144.*m11*RS156pe*RSe45613p + 1.
         /144.*m11*RS156pe*RSp45613e + 1./288.*m11*RS156pe*RS3456pee + 
         1./288.*m11*RS156pe*RS3456ppe );

      F13p6x23R +=  + Ab(m13,mu2)*pi16 * (  - 1./288.*m11*RS156pe*
         RSe45633p - 1./288.*m11*RS156pe*RSp45633e - 1./288.*m11*
         RS156pe*RS345pe - 1./288.*m11*RS156pe*RS346pe - 1./288.*m11*
         RS156pe*RS356pe + 1./288.*m11*RS156pe*RS1456pee + 1./288.*m11*
         RS156pe*RS1456ppe - 1./288.*m11*RS156pe*RSe45611p - 1./288.*
         m11*RS156pe*RSp45611e - 1./288.*m11*RS156pe*RS145pe - 1./288.*
         m11*RS156pe*RS146pe - 1./576.*m11*pow(RS156pe,2) );

      F13p6x23R +=  + pow(Ab(m13,mu2),2) * ( 1./8. );

      F13p6x23R +=  + Ab(m13,mu2)*Ab(m33,mu2) * (  - 5./24. + 1./162.*
         pow(RS34561pe,2) + 1./162.*RS14563pe*RS34561pe + 1./162.*
         RSe45613p*RS34561pe + 1./162.*RSp45613e*RS34561pe + 1./162.*
         RS3456pee*RS34561pe + 1./324.*RS3456pee*RS14563pe + 1./324.*
         RS3456pee*RSe45613p + 1./324.*RS3456pee*RSp45613e + 1./648.*
         pow(RS3456pee,2) + 1./162.*RS3456ppe*RS34561pe + 1./324.*
         RS3456ppe*RS14563pe + 1./324.*RS3456ppe*RSe45613p + 1./324.*
         RS3456ppe*RSp45613e + 1./324.*RS3456ppe*RS3456pee + 1./648.*
         pow(RS3456ppe,2) - 1./324.*RSe45633p*RS34561pe - 1./648.*
         RSe45633p*RS3456pee - 1./648.*RSe45633p*RS3456ppe - 1./324.*
         RSp45633e*RS34561pe - 1./648.*RSp45633e*RS3456pee - 1./648.*
         RSp45633e*RS3456ppe - 1./162.*RS345pe*RS34561pe - 1./324.*
         RS345pe*RS14563pe - 1./324.*RS345pe*RSe45613p - 1./324.*
         RS345pe*RSp45613e - 1./324.*RS345pe*RS3456pee - 1./324.*
         RS345pe*RS3456ppe + 1./648.*RS345pe*RSe45633p + 1./648.*
         RS345pe*RSp45633e );

      F13p6x23R +=  + Ab(m13,mu2)*Ab(m33,mu2) * ( 1./648.*pow(
         RS345pe,2) - 1./162.*RS346pe*RS34561pe - 1./324.*RS346pe*
         RS14563pe - 1./324.*RS346pe*RSe45613p - 1./324.*RS346pe*
         RSp45613e - 1./324.*RS346pe*RS3456pee - 1./324.*RS346pe*
         RS3456ppe + 1./648.*RS346pe*RSe45633p + 1./648.*RS346pe*
         RSp45633e + 1./324.*RS346pe*RS345pe + 1./648.*pow(RS346pe,2)
          - 1./162.*RS356pe*RS34561pe - 1./324.*RS356pe*RS14563pe - 1./
         324.*RS356pe*RSe45613p - 1./324.*RS356pe*RSp45613e - 1./324.*
         RS356pe*RS3456pee - 1./324.*RS356pe*RS3456ppe + 1./648.*
         RS356pe*RSe45633p + 1./648.*RS356pe*RSp45633e + 1./324.*
         RS356pe*RS345pe + 1./324.*RS356pe*RS346pe + 1./648.*pow(
         RS356pe,2) + 1./324.*RS1456pee*RS34561pe + 1./648.*RS1456pee*
         RS3456pee + 1./648.*RS1456pee*RS3456ppe - 1./648.*RS1456pee*
         RS345pe - 1./648.*RS1456pee*RS346pe - 1./648.*RS1456pee*
         RS356pe + 1./324.*RS1456ppe*RS34561pe + 1./648.*RS1456ppe*
         RS3456pee );

      F13p6x23R +=  + Ab(m13,mu2)*Ab(m33,mu2) * ( 1./648.*RS1456ppe*
         RS3456ppe - 1./648.*RS1456ppe*RS345pe - 1./648.*RS1456ppe*
         RS346pe - 1./648.*RS1456ppe*RS356pe - 1./324.*RSe45611p*
         RS34561pe - 1./648.*RSe45611p*RS3456pee - 1./648.*RSe45611p*
         RS3456ppe + 1./648.*RSe45611p*RS345pe + 1./648.*RSe45611p*
         RS346pe + 1./648.*RSe45611p*RS356pe - 1./324.*RSp45611e*
         RS34561pe - 1./648.*RSp45611e*RS3456pee - 1./648.*RSp45611e*
         RS3456ppe + 1./648.*RSp45611e*RS345pe + 1./648.*RSp45611e*
         RS346pe + 1./648.*RSp45611e*RS356pe - 1./324.*RS145pe*
         RS34561pe - 1./648.*RS145pe*RS3456pee - 1./648.*RS145pe*
         RS3456ppe + 1./648.*RS145pe*RS345pe + 1./648.*RS145pe*RS346pe
          + 1./648.*RS145pe*RS356pe - 1./324.*RS146pe*RS34561pe - 1./
         648.*RS146pe*RS3456pee - 1./648.*RS146pe*RS3456ppe + 1./648.*
         RS146pe*RS345pe + 1./648.*RS146pe*RS346pe + 1./648.*RS146pe*
         RS356pe - 1./324.*RS156pe*RS34561pe - 1./648.*RS156pe*
         RS3456pee );

      F13p6x23R +=  + Ab(m13,mu2)*Ab(m33,mu2) * (  - 1./648.*RS156pe*
         RS3456ppe + 1./648.*RS156pe*RS345pe + 1./648.*RS156pe*RS346pe
          + 1./648.*RS156pe*RS356pe + 1./6.*m33*RN13 + 1./3.*m11*RN13 )
         ;

      F13p6x23R +=  + Ab(m13,mu2)*Ab(me2,mu2) * ( 1./162.*RSe45613p*
         RS34561pe + 1./162.*RSe45613p*RS14563pe + 1./162.*pow(
         RSe45613p,2) + 1./162.*RSp45613e*RSe45613p + 1./324.*RS3456pee
         *RSe45613p + 1./324.*RS3456ppe*RSe45613p - 1./324.*RSe45633p*
         RS34561pe - 1./324.*RSe45633p*RS14563pe - 1./162.*RSe45633p*
         RSe45613p - 1./324.*RSe45633p*RSp45613e - 1./648.*RSe45633p*
         RS3456pee - 1./648.*RSe45633p*RS3456ppe + 1./648.*pow(
         RSe45633p,2) - 1./324.*RSp45633e*RSe45613p + 1./648.*RSp45633e
         *RSe45633p - 1./324.*RS345pe*RSe45613p + 1./648.*RS345pe*
         RSe45633p - 1./324.*RS346pe*RSe45613p + 1./648.*RS346pe*
         RSe45633p - 1./324.*RS356pe*RSe45613p + 1./648.*RS356pe*
         RSe45633p + 1./324.*RS1456pee*RSe45613p - 1./648.*RS1456pee*
         RSe45633p + 1./324.*RS1456ppe*RSe45613p - 1./648.*RS1456ppe*
         RSe45633p - 1./324.*RSe45611p*RS34561pe - 1./324.*RSe45611p*
         RS14563pe - 1./162.*RSe45611p*RSe45613p - 1./324.*RSe45611p*
         RSp45613e );

      F13p6x23R +=  + Ab(m13,mu2)*Ab(me2,mu2) * (  - 1./648.*RSe45611p*
         RS3456pee - 1./648.*RSe45611p*RS3456ppe + 1./324.*RSe45611p*
         RSe45633p + 1./648.*RSe45611p*RSp45633e + 1./648.*RSe45611p*
         RS345pe + 1./648.*RSe45611p*RS346pe + 1./648.*RSe45611p*
         RS356pe - 1./648.*RSe45611p*RS1456pee - 1./648.*RSe45611p*
         RS1456ppe + 1./648.*pow(RSe45611p,2) - 1./324.*RSp45611e*
         RSe45613p + 1./648.*RSp45611e*RSe45633p + 1./648.*RSp45611e*
         RSe45611p - 1./324.*RS145pe*RSe45613p + 1./648.*RS145pe*
         RSe45633p + 1./648.*RS145pe*RSe45611p - 1./324.*RS146pe*
         RSe45613p + 1./648.*RS146pe*RSe45633p + 1./648.*RS146pe*
         RSe45611p - 1./324.*RS156pe*RSe45613p + 1./648.*RS156pe*
         RSe45633p + 1./648.*RS156pe*RSe45611p );

      F13p6x23R +=  + Ab(m13,mu2)*Ab(mp2,mu2) * ( 1./162.*RSp45613e*
         RS34561pe + 1./162.*RSp45613e*RS14563pe + 1./162.*RSp45613e*
         RSe45613p + 1./162.*pow(RSp45613e,2) + 1./324.*RS3456pee*
         RSp45613e + 1./324.*RS3456ppe*RSp45613e - 1./324.*RSe45633p*
         RSp45613e - 1./324.*RSp45633e*RS34561pe - 1./324.*RSp45633e*
         RS14563pe - 1./324.*RSp45633e*RSe45613p - 1./162.*RSp45633e*
         RSp45613e - 1./648.*RSp45633e*RS3456pee - 1./648.*RSp45633e*
         RS3456ppe + 1./648.*RSp45633e*RSe45633p + 1./648.*pow(
         RSp45633e,2) - 1./324.*RS345pe*RSp45613e + 1./648.*RS345pe*
         RSp45633e - 1./324.*RS346pe*RSp45613e + 1./648.*RS346pe*
         RSp45633e - 1./324.*RS356pe*RSp45613e + 1./648.*RS356pe*
         RSp45633e + 1./324.*RS1456pee*RSp45613e - 1./648.*RS1456pee*
         RSp45633e + 1./324.*RS1456ppe*RSp45613e - 1./648.*RS1456ppe*
         RSp45633e - 1./324.*RSe45611p*RSp45613e + 1./648.*RSe45611p*
         RSp45633e - 1./324.*RSp45611e*RS34561pe - 1./324.*RSp45611e*
         RS14563pe );

      F13p6x23R +=  + Ab(m13,mu2)*Ab(mp2,mu2) * (  - 1./324.*RSp45611e*
         RSe45613p - 1./162.*RSp45611e*RSp45613e - 1./648.*RSp45611e*
         RS3456pee - 1./648.*RSp45611e*RS3456ppe + 1./648.*RSp45611e*
         RSe45633p + 1./324.*RSp45611e*RSp45633e + 1./648.*RSp45611e*
         RS345pe + 1./648.*RSp45611e*RS346pe + 1./648.*RSp45611e*
         RS356pe - 1./648.*RSp45611e*RS1456pee - 1./648.*RSp45611e*
         RS1456ppe + 1./648.*RSp45611e*RSe45611p + 1./648.*pow(
         RSp45611e,2) - 1./324.*RS145pe*RSp45613e + 1./648.*RS145pe*
         RSp45633e + 1./648.*RS145pe*RSp45611e - 1./324.*RS146pe*
         RSp45613e + 1./648.*RS146pe*RSp45633e + 1./648.*RS146pe*
         RSp45611e - 1./324.*RS156pe*RSp45613e + 1./648.*RS156pe*
         RSp45633e + 1./648.*RS156pe*RSp45611e );

      F13p6x23R +=  + Ab(m13,mu2)*Bb(m11,mu2) * (  - 1./324.*RD1456ep*
         RS34561pe - 1./324.*RD1456ep*RS14563pe - 1./324.*RD1456ep*
         RSe45613p - 1./324.*RD1456ep*RSp45613e - 1./648.*RD1456ep*
         RS3456pee - 1./648.*RD1456ep*RS3456ppe + 1./648.*RD1456ep*
         RSe45633p + 1./648.*RD1456ep*RSp45633e + 1./648.*RD1456ep*
         RS345pe + 1./648.*RD1456ep*RS346pe + 1./648.*RD1456ep*RS356pe
          - 1./648.*RS1456pee*RD1456ep - 1./648.*RS1456ppe*RD1456ep + 1.
         /648.*RSe45611p*RD1456ep + 1./648.*RSp45611e*RD1456ep + 1./648.
         *RS145pe*RD1456ep + 1./648.*RS146pe*RD1456ep + 1./648.*RS156pe
         *RD1456ep );

      F13p6x23R +=  + Ab(m13,mu2)*Bb(m33,mu2) * (  - 1./324.*RD3456ep*
         RS34561pe - 1./324.*RD3456ep*RS14563pe - 1./324.*RD3456ep*
         RSe45613p - 1./324.*RD3456ep*RSp45613e - 1./648.*RS3456pee*
         RD3456ep - 1./648.*RS3456ppe*RD3456ep + 1./648.*RSe45633p*
         RD3456ep + 1./648.*RSp45633e*RD3456ep + 1./648.*RS345pe*
         RD3456ep + 1./648.*RS346pe*RD3456ep + 1./648.*RS356pe*RD3456ep
          - 1./648.*RS1456pee*RD3456ep - 1./648.*RS1456ppe*RD3456ep + 1.
         /648.*RSe45611p*RD3456ep + 1./648.*RSp45611e*RD3456ep + 1./648.
         *RS145pe*RD3456ep + 1./648.*RS146pe*RD3456ep + 1./648.*RS156pe
         *RD3456ep );

      F13p6x23R +=  + Ab(m14,mu2)*pi16 * ( 1./48.*RD3456ep + 1./48.*mp2
         *RSp45633e - 1./24.*mp2*RSp563e + 1./48.*mp2*RSp564e + 1./48.*
         me2*RSe45633p - 1./24.*me2*RSe563p + 1./48.*me2*RSe564p - 1./
         16.*m66 - 1./16.*m55 - 1./8.*m44 - 1./96.*m44*RS34561pe - 1./
         96.*m44*RS14563pe - 1./96.*m44*RSe45613p - 1./96.*m44*
         RSp45613e - 1./96.*m44*RS3456pee - 1./96.*m44*RS3456ppe + 1./
         96.*m44*RSe45633p + 1./96.*m44*RSp45633e + 1./96.*m44*RS345pe
          + 1./96.*m44*RS346pe - 1./96.*m44*RSe563p - 1./96.*m44*
         RSp563e - 1./96.*m44*RS1456pee - 1./96.*m44*RS1456ppe + 1./96.
         *m44*RSe45611p + 1./96.*m44*RSp45611e + 1./96.*m44*RS145pe + 1.
         /96.*m44*RS146pe + 1./32.*m44*RS456pe - 1./96.*m44*RSe561p - 1.
         /96.*m44*RSp561e + 1./96.*m44*RSe564p + 1./96.*m44*RSp564e - 3.
         /32.*m33 - 1./32.*m33*RS3456pee - 1./32.*m33*RS3456ppe + 1./96.
         *m33*RSe45633p + 1./96.*m33*RSp45633e + 1./32.*m33*RS345pe + 1.
         /32.*m33*RS346pe - 1./32.*m33*RS356pe - 1./48.*m33*RSe563p - 1.
         /48.*m33*RSp563e );

      F13p6x23R +=  + Ab(m14,mu2)*pi16 * ( 1./96.*m33*RS456pe + 1./96.*
         m33*RSe564p + 1./96.*m33*RSp564e + 1./32.*m11 - 1./96.*m11*
         RS34561pe - 1./96.*m11*RS14563pe - 1./96.*m11*RSe45613p - 1./
         96.*m11*RSp45613e + 1./96.*m11*RS356pe + 1./96.*m11*RSe563p + 
         1./96.*m11*RSp563e - 1./96.*m11*RS1456pee - 1./96.*m11*
         RS1456ppe + 1./96.*m11*RSe45611p + 1./96.*m11*RSp45611e + 1./
         96.*m11*RS145pe + 1./96.*m11*RS146pe - 1./96.*m11*RSe561p - 1./
         96.*m11*RSp561e );

      F13p6x23R +=  + pow(Ab(m14,mu2),2) * (  - 1./32. );

      F13p6x23R +=  + Ab(m14,mu2)*Ab(m15,mu2) * (  - 1./16. );

      F13p6x23R +=  + Ab(m14,mu2)*Ab(m16,mu2) * (  - 1./16. );

      F13p6x23R +=  + Ab(m14,mu2)*Ab(m33,mu2) * (  - 1./36.*RS34561pe
          + 1./48.*RS3456pee + 1./48.*RS3456ppe - 1./48.*RS345pe - 1./
         48.*RS346pe + 7./144.*RS356pe - 1./18.*m44*RN13*RS356pe - 1./9.
         *m11*RN13*RS356pe );

      F13p6x23R +=  + Ab(m14,mu2)*Ab(m34,mu2) * ( 1./16. + 7./216.*
         RS34561pe + 7./216.*RS14563pe + 7./216.*RSe45613p + 7./216.*
         RSp45613e + 1./216.*RS3456pee + 1./216.*RS3456ppe - 1./216.*
         RSe45633p - 1./216.*RSp45633e - 1./216.*RS345pe - 1./216.*
         RS346pe - 1./36.*RS356pe - 5./216.*RSe563p - 5./216.*RSp563e
          + 1./216.*RS1456pee + 1./216.*RS1456ppe - 1./216.*RSe45611p
          - 1./216.*RSp45611e - 1./216.*RS145pe - 1./216.*RS146pe - 1./
         36.*RS156pe + 5./216.*RS456pe - 5./216.*RSe561p - 5./216.*
         RSp561e + 5./216.*RSe564p + 5./216.*RSp564e );

      F13p6x23R +=  + Ab(m14,mu2)*Ab(m35,mu2) * ( 1./16. );

      F13p6x23R +=  + Ab(m14,mu2)*Ab(m36,mu2) * ( 1./16. );

      F13p6x23R +=  + Ab(m14,mu2)*Ab(me2,mu2) * ( 1./36.*RSe45613p - 1./
         48.*RSe45633p + 1./72.*RSe563p - 1./144.*RSe45611p - 1./72.*
         RSe561p + 1./18.*me2*RN1e*RSe45613p - 1./18.*me2*RN1e*RSe563p
          - 1./18.*me2*RN1e*RSe45611p + 1./18.*me2*RN1e*RSe561p - 1./18.
         *m44*RN1e*RSe563p + 1./18.*m44*RN1e*RSe561p - 1./18.*m11*RN1e*
         RSe45613p - 1./18.*m11*RN1e*RSe563p + 1./18.*m11*RN1e*
         RSe45611p + 1./18.*m11*RN1e*RSe561p );

      F13p6x23R +=  + Ab(m14,mu2)*Ab(mp2,mu2) * ( 1./36.*RSp45613e - 1./
         48.*RSp45633e + 1./72.*RSp563e - 1./144.*RSp45611e - 1./72.*
         RSp561e + 1./18.*mp2*RN1p*RSp45613e - 1./18.*mp2*RN1p*RSp563e
          - 1./18.*mp2*RN1p*RSp45611e + 1./18.*mp2*RN1p*RSp561e - 1./18.
         *m44*RN1p*RSp563e + 1./18.*m44*RN1p*RSp561e - 1./18.*m11*RN1p*
         RSp45613e - 1./18.*m11*RN1p*RSp563e + 1./18.*m11*RN1p*
         RSp45611e + 1./18.*m11*RN1p*RSp561e );

      F13p6x23R +=  + Ab(m14,mu2)*Bb(m11,mu2) * ( 7./144.*RD1456ep - 1./
         18.*m44*RS156pe - 1./9.*m11*RS156pe );

      F13p6x23R +=  + Ab(m14,mu2)*Bb(m33,mu2) * (  - 1./48.*RD3456ep );

      F13p6x23R +=  + Ab(m15,mu2)*pi16 * ( 1./48.*RD3456ep + 1./48.*mp2
         *RSp45633e - 1./24.*mp2*RSp463e + 1./48.*mp2*RSp465e + 1./48.*
         me2*RSe45633p - 1./24.*me2*RSe463p + 1./48.*me2*RSe465p - 1./
         16.*m66 - 1./8.*m55 - 1./96.*m55*RS34561pe - 1./96.*m55*
         RS14563pe - 1./96.*m55*RSe45613p - 1./96.*m55*RSp45613e - 1./
         96.*m55*RS3456pee - 1./96.*m55*RS3456ppe + 1./96.*m55*
         RSe45633p + 1./96.*m55*RSp45633e + 1./96.*m55*RS345pe + 1./96.
         *m55*RS356pe - 1./96.*m55*RSe463p - 1./96.*m55*RSp463e - 1./96.
         *m55*RS1456pee - 1./96.*m55*RS1456ppe + 1./96.*m55*RSe45611p
          + 1./96.*m55*RSp45611e + 1./96.*m55*RS145pe + 1./96.*m55*
         RS156pe + 1./32.*m55*RS546pe - 1./96.*m55*RSe461p - 1./96.*m55
         *RSp461e + 1./96.*m55*RSe465p + 1./96.*m55*RSp465e - 1./16.*
         m44 - 3./32.*m33 - 1./32.*m33*RS3456pee - 1./32.*m33*RS3456ppe
          + 1./96.*m33*RSe45633p + 1./96.*m33*RSp45633e + 1./32.*m33*
         RS345pe - 1./32.*m33*RS346pe + 1./32.*m33*RS356pe - 1./48.*m33
         *RSe463p );

      F13p6x23R +=  + Ab(m15,mu2)*pi16 * (  - 1./48.*m33*RSp463e + 1./
         96.*m33*RS546pe + 1./96.*m33*RSe465p + 1./96.*m33*RSp465e + 1./
         32.*m11 - 1./96.*m11*RS34561pe - 1./96.*m11*RS14563pe - 1./96.
         *m11*RSe45613p - 1./96.*m11*RSp45613e + 1./96.*m11*RS346pe + 1.
         /96.*m11*RSe463p + 1./96.*m11*RSp463e - 1./96.*m11*RS1456pee
          - 1./96.*m11*RS1456ppe + 1./96.*m11*RSe45611p + 1./96.*m11*
         RSp45611e + 1./96.*m11*RS145pe + 1./96.*m11*RS156pe - 1./96.*
         m11*RSe461p - 1./96.*m11*RSp461e );

      F13p6x23R +=  + pow(Ab(m15,mu2),2) * (  - 1./32. );

      F13p6x23R +=  + Ab(m15,mu2)*Ab(m16,mu2) * (  - 1./16. );

      F13p6x23R +=  + Ab(m15,mu2)*Ab(m33,mu2) * (  - 1./36.*RS34561pe
          + 1./48.*RS3456pee + 1./48.*RS3456ppe - 1./48.*RS345pe + 7./
         144.*RS346pe - 1./48.*RS356pe - 1./18.*m55*RN13*RS346pe - 1./9.
         *m11*RN13*RS346pe );

      F13p6x23R +=  + Ab(m15,mu2)*Ab(m34,mu2) * ( 1./16. );

      F13p6x23R +=  + Ab(m15,mu2)*Ab(m35,mu2) * ( 1./16. + 7./216.*
         RS34561pe + 7./216.*RS14563pe + 7./216.*RSe45613p + 7./216.*
         RSp45613e + 1./216.*RS3456pee + 1./216.*RS3456ppe - 1./216.*
         RSe45633p - 1./216.*RSp45633e - 1./216.*RS345pe - 1./36.*
         RS346pe - 1./216.*RS356pe - 5./216.*RSe463p - 5./216.*RSp463e
          + 1./216.*RS1456pee + 1./216.*RS1456ppe - 1./216.*RSe45611p
          - 1./216.*RSp45611e - 1./216.*RS145pe - 1./36.*RS146pe - 1./
         216.*RS156pe + 5./216.*RS546pe - 5./216.*RSe461p - 5./216.*
         RSp461e + 5./216.*RSe465p + 5./216.*RSp465e );

      F13p6x23R +=  + Ab(m15,mu2)*Ab(m36,mu2) * ( 1./16. );

      F13p6x23R +=  + Ab(m15,mu2)*Ab(me2,mu2) * ( 1./36.*RSe45613p - 1./
         48.*RSe45633p + 1./72.*RSe463p - 1./144.*RSe45611p - 1./72.*
         RSe461p + 1./18.*me2*RN1e*RSe45613p - 1./18.*me2*RN1e*RSe463p
          - 1./18.*me2*RN1e*RSe45611p + 1./18.*me2*RN1e*RSe461p - 1./18.
         *m55*RN1e*RSe463p + 1./18.*m55*RN1e*RSe461p - 1./18.*m11*RN1e*
         RSe45613p - 1./18.*m11*RN1e*RSe463p + 1./18.*m11*RN1e*
         RSe45611p + 1./18.*m11*RN1e*RSe461p );

      F13p6x23R +=  + Ab(m15,mu2)*Ab(mp2,mu2) * ( 1./36.*RSp45613e - 1./
         48.*RSp45633e + 1./72.*RSp463e - 1./144.*RSp45611e - 1./72.*
         RSp461e + 1./18.*mp2*RN1p*RSp45613e - 1./18.*mp2*RN1p*RSp463e
          - 1./18.*mp2*RN1p*RSp45611e + 1./18.*mp2*RN1p*RSp461e - 1./18.
         *m55*RN1p*RSp463e + 1./18.*m55*RN1p*RSp461e - 1./18.*m11*RN1p*
         RSp45613e - 1./18.*m11*RN1p*RSp463e + 1./18.*m11*RN1p*
         RSp45611e + 1./18.*m11*RN1p*RSp461e );

      F13p6x23R +=  + Ab(m15,mu2)*Bb(m11,mu2) * ( 7./144.*RD1456ep - 1./
         18.*m55*RS146pe - 1./9.*m11*RS146pe );

      F13p6x23R +=  + Ab(m15,mu2)*Bb(m33,mu2) * (  - 1./48.*RD3456ep );

      F13p6x23R +=  + Ab(m16,mu2)*pi16 * ( 1./48.*RD3456ep + 1./48.*mp2
         *RSp45633e - 1./24.*mp2*RSp453e + 1./48.*mp2*RSp456e + 1./48.*
         me2*RSe45633p - 1./24.*me2*RSe453p + 1./48.*me2*RSe456p - 1./8.
         *m66 - 1./96.*m66*RS34561pe - 1./96.*m66*RS14563pe - 1./96.*
         m66*RSe45613p - 1./96.*m66*RSp45613e - 1./96.*m66*RS3456pee - 
         1./96.*m66*RS3456ppe + 1./96.*m66*RSe45633p + 1./96.*m66*
         RSp45633e + 1./96.*m66*RS346pe + 1./96.*m66*RS356pe - 1./96.*
         m66*RSe453p - 1./96.*m66*RSp453e - 1./96.*m66*RS1456pee - 1./
         96.*m66*RS1456ppe + 1./96.*m66*RSe45611p + 1./96.*m66*
         RSp45611e + 1./96.*m66*RS146pe + 1./96.*m66*RS156pe + 1./32.*
         m66*RS645pe - 1./96.*m66*RSe451p - 1./96.*m66*RSp451e + 1./96.
         *m66*RSe456p + 1./96.*m66*RSp456e - 1./16.*m55 - 1./16.*m44 - 
         3./32.*m33 - 1./32.*m33*RS3456pee - 1./32.*m33*RS3456ppe + 1./
         96.*m33*RSe45633p + 1./96.*m33*RSp45633e - 1./32.*m33*RS345pe
          + 1./32.*m33*RS346pe + 1./32.*m33*RS356pe - 1./48.*m33*
         RSe453p );

      F13p6x23R +=  + Ab(m16,mu2)*pi16 * (  - 1./48.*m33*RSp453e + 1./
         96.*m33*RS645pe + 1./96.*m33*RSe456p + 1./96.*m33*RSp456e + 1./
         32.*m11 - 1./96.*m11*RS34561pe - 1./96.*m11*RS14563pe - 1./96.
         *m11*RSe45613p - 1./96.*m11*RSp45613e + 1./96.*m11*RS345pe + 1.
         /96.*m11*RSe453p + 1./96.*m11*RSp453e - 1./96.*m11*RS1456pee
          - 1./96.*m11*RS1456ppe + 1./96.*m11*RSe45611p + 1./96.*m11*
         RSp45611e + 1./96.*m11*RS146pe + 1./96.*m11*RS156pe - 1./96.*
         m11*RSe451p - 1./96.*m11*RSp451e );

      F13p6x23R +=  + pow(Ab(m16,mu2),2) * (  - 1./32. );

      F13p6x23R +=  + Ab(m16,mu2)*Ab(m33,mu2) * (  - 1./36.*RS34561pe
          + 1./48.*RS3456pee + 1./48.*RS3456ppe + 7./144.*RS345pe - 1./
         48.*RS346pe - 1./48.*RS356pe - 1./18.*m66*RN13*RS345pe - 1./9.
         *m11*RN13*RS345pe );

      F13p6x23R +=  + Ab(m16,mu2)*Ab(m34,mu2) * ( 1./16. );

      F13p6x23R +=  + Ab(m16,mu2)*Ab(m35,mu2) * ( 1./16. );

      F13p6x23R +=  + Ab(m16,mu2)*Ab(m36,mu2) * ( 1./16. + 7./216.*
         RS34561pe + 7./216.*RS14563pe + 7./216.*RSe45613p + 7./216.*
         RSp45613e + 1./216.*RS3456pee + 1./216.*RS3456ppe - 1./216.*
         RSe45633p - 1./216.*RSp45633e - 1./36.*RS345pe - 1./216.*
         RS346pe - 1./216.*RS356pe - 5./216.*RSe453p - 5./216.*RSp453e
          + 1./216.*RS1456pee + 1./216.*RS1456ppe - 1./216.*RSe45611p
          - 1./216.*RSp45611e - 1./36.*RS145pe - 1./216.*RS146pe - 1./
         216.*RS156pe + 5./216.*RS645pe - 5./216.*RSe451p - 5./216.*
         RSp451e + 5./216.*RSe456p + 5./216.*RSp456e );

      F13p6x23R +=  + Ab(m16,mu2)*Ab(me2,mu2) * ( 1./36.*RSe45613p - 1./
         48.*RSe45633p + 1./72.*RSe453p - 1./144.*RSe45611p - 1./72.*
         RSe451p + 1./18.*me2*RN1e*RSe45613p - 1./18.*me2*RN1e*RSe453p
          - 1./18.*me2*RN1e*RSe45611p + 1./18.*me2*RN1e*RSe451p - 1./18.
         *m66*RN1e*RSe453p + 1./18.*m66*RN1e*RSe451p - 1./18.*m11*RN1e*
         RSe45613p - 1./18.*m11*RN1e*RSe453p + 1./18.*m11*RN1e*
         RSe45611p + 1./18.*m11*RN1e*RSe451p );

      F13p6x23R +=  + Ab(m16,mu2)*Ab(mp2,mu2) * ( 1./36.*RSp45613e - 1./
         48.*RSp45633e + 1./72.*RSp453e - 1./144.*RSp45611e - 1./72.*
         RSp451e + 1./18.*mp2*RN1p*RSp45613e - 1./18.*mp2*RN1p*RSp453e
          - 1./18.*mp2*RN1p*RSp45611e + 1./18.*mp2*RN1p*RSp451e - 1./18.
         *m66*RN1p*RSp453e + 1./18.*m66*RN1p*RSp451e - 1./18.*m11*RN1p*
         RSp45613e - 1./18.*m11*RN1p*RSp453e + 1./18.*m11*RN1p*
         RSp45611e + 1./18.*m11*RN1p*RSp451e );

      F13p6x23R +=  + Ab(m16,mu2)*Bb(m11,mu2) * ( 7./144.*RD1456ep - 1./
         18.*m66*RS145pe - 1./9.*m11*RS145pe );

      F13p6x23R +=  + Ab(m16,mu2)*Bb(m33,mu2) * (  - 1./48.*RD3456ep );

      F13p6x23R +=  + Ab(m33,mu2)*pi16 * (  - 1./48.*RD3456ep + 1./72.*
         RD3456ep*RS34561pe + 1./144.*RS3456pee*RD3456ep + 1./144.*
         RS3456ppe*RD3456ep - 1./144.*RS345pe*RD3456ep - 1./144.*
         RS346pe*RD3456ep - 1./144.*RS356pe*RD3456ep + 1./72.*RD1456ep*
         RS34561pe + 1./144.*RD1456ep*RS3456pee + 1./144.*RD1456ep*
         RS3456ppe - 1./144.*RD1456ep*RS345pe - 1./144.*RD1456ep*
         RS346pe - 1./144.*RD1456ep*RS356pe - 1./36.*mp2*RSp45613e*
         RS34561pe - 1./72.*mp2*RS3456pee*RSp45613e - 1./72.*mp2*
         RS3456ppe*RSp45613e + 1./72.*mp2*RSp45633e*RS34561pe + 1./144.
         *mp2*RSp45633e*RS3456pee + 1./144.*mp2*RSp45633e*RS3456ppe + 1.
         /72.*mp2*RS345pe*RSp45613e - 1./144.*mp2*RS345pe*RSp45633e + 1.
         /72.*mp2*RS346pe*RSp45613e - 1./144.*mp2*RS346pe*RSp45633e + 1.
         /72.*mp2*RS356pe*RSp45613e - 1./144.*mp2*RS356pe*RSp45633e + 1.
         /72.*mp2*RSp45611e*RS34561pe + 1./144.*mp2*RSp45611e*RS3456pee
          + 1./144.*mp2*RSp45611e*RS3456ppe - 1./144.*mp2*RSp45611e*
         RS345pe );

      F13p6x23R +=  + Ab(m33,mu2)*pi16 * (  - 1./144.*mp2*RSp45611e*
         RS346pe - 1./144.*mp2*RSp45611e*RS356pe - 1./36.*me2*RSe45613p
         *RS34561pe - 1./72.*me2*RS3456pee*RSe45613p - 1./72.*me2*
         RS3456ppe*RSe45613p + 1./72.*me2*RSe45633p*RS34561pe + 1./144.
         *me2*RSe45633p*RS3456pee + 1./144.*me2*RSe45633p*RS3456ppe + 1.
         /72.*me2*RS345pe*RSe45613p - 1./144.*me2*RS345pe*RSe45633p + 1.
         /72.*me2*RS346pe*RSe45613p - 1./144.*me2*RS346pe*RSe45633p + 1.
         /72.*me2*RS356pe*RSe45613p - 1./144.*me2*RS356pe*RSe45633p + 1.
         /72.*me2*RSe45611p*RS34561pe + 1./144.*me2*RSe45611p*RS3456pee
          + 1./144.*me2*RSe45611p*RS3456ppe - 1./144.*me2*RSe45611p*
         RS345pe - 1./144.*me2*RSe45611p*RS346pe - 1./144.*me2*
         RSe45611p*RS356pe - 1./24.*m66*RS34561pe - 1./48.*m66*
         RS3456pee - 1./48.*m66*RS3456ppe + 1./48.*m66*RS345pe + 1./48.
         *m66*RS346pe + 1./48.*m66*RS356pe - 1./24.*m55*RS34561pe - 1./
         48.*m55*RS3456pee - 1./48.*m55*RS3456ppe + 1./48.*m55*RS345pe
          + 1./48.*m55*RS346pe );

      F13p6x23R +=  + Ab(m33,mu2)*pi16 * ( 1./48.*m55*RS356pe - 1./24.*
         m44*RS34561pe - 1./48.*m44*RS3456pee - 1./48.*m44*RS3456ppe + 
         1./48.*m44*RS345pe + 1./48.*m44*RS346pe + 1./48.*m44*RS356pe
          + 1./16.*m33 - 1./48.*m33*RS34561pe - 1./24.*m33*pow(
         RS34561pe,2) - 1./72.*m33*RS14563pe*RS34561pe - 1./72.*m33*
         RSe45613p*RS34561pe - 1./72.*m33*RSp45613e*RS34561pe - 1./96.*
         m33*RS3456pee - 1./24.*m33*RS3456pee*RS34561pe - 1./144.*m33*
         RS3456pee*RS14563pe - 1./144.*m33*RS3456pee*RSe45613p - 1./144.
         *m33*RS3456pee*RSp45613e - 1./96.*m33*pow(RS3456pee,2) - 1./96.
         *m33*RS3456ppe - 1./24.*m33*RS3456ppe*RS34561pe - 1./144.*m33*
         RS3456ppe*RS14563pe - 1./144.*m33*RS3456ppe*RSe45613p - 1./144.
         *m33*RS3456ppe*RSp45613e - 1./48.*m33*RS3456ppe*RS3456pee - 1./
         96.*m33*pow(RS3456ppe,2) + 1./144.*m33*RSe45633p*RS34561pe + 1.
         /288.*m33*RSe45633p*RS3456pee + 1./288.*m33*RSe45633p*
         RS3456ppe + 1./144.*m33*RSp45633e*RS34561pe + 1./288.*m33*
         RSp45633e*RS3456pee );

      F13p6x23R +=  + Ab(m33,mu2)*pi16 * ( 1./288.*m33*RSp45633e*
         RS3456ppe - 1./96.*m33*RS345pe + 1./24.*m33*RS345pe*RS34561pe
          + 1./144.*m33*RS345pe*RS14563pe + 1./144.*m33*RS345pe*
         RSe45613p + 1./144.*m33*RS345pe*RSp45613e + 1./48.*m33*RS345pe
         *RS3456pee + 1./48.*m33*RS345pe*RS3456ppe - 1./288.*m33*
         RS345pe*RSe45633p - 1./288.*m33*RS345pe*RSp45633e - 1./96.*m33
         *pow(RS345pe,2) - 1./96.*m33*RS346pe + 1./24.*m33*RS346pe*
         RS34561pe + 1./144.*m33*RS346pe*RS14563pe + 1./144.*m33*
         RS346pe*RSe45613p + 1./144.*m33*RS346pe*RSp45613e + 1./48.*m33
         *RS346pe*RS3456pee + 1./48.*m33*RS346pe*RS3456ppe - 1./288.*
         m33*RS346pe*RSe45633p - 1./288.*m33*RS346pe*RSp45633e - 1./48.
         *m33*RS346pe*RS345pe - 1./96.*m33*pow(RS346pe,2) - 1./96.*m33*
         RS356pe + 1./24.*m33*RS356pe*RS34561pe + 1./144.*m33*RS356pe*
         RS14563pe + 1./144.*m33*RS356pe*RSe45613p + 1./144.*m33*
         RS356pe*RSp45613e + 1./48.*m33*RS356pe*RS3456pee + 1./48.*m33*
         RS356pe*RS3456ppe );

      F13p6x23R +=  + Ab(m33,mu2)*pi16 * (  - 1./288.*m33*RS356pe*
         RSe45633p - 1./288.*m33*RS356pe*RSp45633e - 1./48.*m33*RS356pe
         *RS345pe - 1./48.*m33*RS356pe*RS346pe - 1./96.*m33*pow(
         RS356pe,2) - 1./144.*m33*RS1456pee*RS34561pe - 1./288.*m33*
         RS1456pee*RS3456pee - 1./288.*m33*RS1456pee*RS3456ppe + 1./288.
         *m33*RS1456pee*RS345pe + 1./288.*m33*RS1456pee*RS346pe + 1./
         288.*m33*RS1456pee*RS356pe - 1./144.*m33*RS1456ppe*RS34561pe
          - 1./288.*m33*RS1456ppe*RS3456pee - 1./288.*m33*RS1456ppe*
         RS3456ppe + 1./288.*m33*RS1456ppe*RS345pe + 1./288.*m33*
         RS1456ppe*RS346pe + 1./288.*m33*RS1456ppe*RS356pe + 1./144.*
         m33*RSe45611p*RS34561pe + 1./288.*m33*RSe45611p*RS3456pee + 1./
         288.*m33*RSe45611p*RS3456ppe - 1./288.*m33*RSe45611p*RS345pe
          - 1./288.*m33*RSe45611p*RS346pe - 1./288.*m33*RSe45611p*
         RS356pe + 1./144.*m33*RSp45611e*RS34561pe + 1./288.*m33*
         RSp45611e*RS3456pee + 1./288.*m33*RSp45611e*RS3456ppe - 1./288.
         *m33*RSp45611e*RS345pe );

      F13p6x23R +=  + Ab(m33,mu2)*pi16 * (  - 1./288.*m33*RSp45611e*
         RS346pe - 1./288.*m33*RSp45611e*RS356pe + 1./144.*m33*RS145pe*
         RS34561pe + 1./288.*m33*RS145pe*RS3456pee + 1./288.*m33*
         RS145pe*RS3456ppe - 1./288.*m33*RS145pe*RS345pe - 1./288.*m33*
         RS145pe*RS346pe - 1./288.*m33*RS145pe*RS356pe + 1./144.*m33*
         RS146pe*RS34561pe + 1./288.*m33*RS146pe*RS3456pee + 1./288.*
         m33*RS146pe*RS3456ppe - 1./288.*m33*RS146pe*RS345pe - 1./288.*
         m33*RS146pe*RS346pe - 1./288.*m33*RS146pe*RS356pe + 1./144.*
         m33*RS156pe*RS34561pe + 1./288.*m33*RS156pe*RS3456pee + 1./288.
         *m33*RS156pe*RS3456ppe - 1./288.*m33*RS156pe*RS345pe - 1./288.
         *m33*RS156pe*RS346pe - 1./288.*m33*RS156pe*RS356pe - 1./16.*
         m11*RS34561pe - 1./72.*m11*pow(RS34561pe,2) - 1./24.*m11*
         RS14563pe*RS34561pe - 1./72.*m11*RSe45613p*RS34561pe - 1./72.*
         m11*RSp45613e*RS34561pe - 1./32.*m11*RS3456pee - 1./72.*m11*
         RS3456pee*RS34561pe - 1./48.*m11*RS3456pee*RS14563pe - 1./144.
         *m11*RS3456pee*RSe45613p );

      F13p6x23R +=  + Ab(m33,mu2)*pi16 * (  - 1./144.*m11*RS3456pee*
         RSp45613e - 1./288.*m11*pow(RS3456pee,2) - 1./32.*m11*
         RS3456ppe - 1./72.*m11*RS3456ppe*RS34561pe - 1./48.*m11*
         RS3456ppe*RS14563pe - 1./144.*m11*RS3456ppe*RSe45613p - 1./144.
         *m11*RS3456ppe*RSp45613e - 1./144.*m11*RS3456ppe*RS3456pee - 1.
         /288.*m11*pow(RS3456ppe,2) + 1./144.*m11*RSe45633p*RS34561pe
          + 1./288.*m11*RSe45633p*RS3456pee + 1./288.*m11*RSe45633p*
         RS3456ppe + 1./144.*m11*RSp45633e*RS34561pe + 1./288.*m11*
         RSp45633e*RS3456pee + 1./288.*m11*RSp45633e*RS3456ppe + 1./32.
         *m11*RS345pe + 1./72.*m11*RS345pe*RS34561pe + 1./48.*m11*
         RS345pe*RS14563pe + 1./144.*m11*RS345pe*RSe45613p + 1./144.*
         m11*RS345pe*RSp45613e + 1./144.*m11*RS345pe*RS3456pee + 1./144.
         *m11*RS345pe*RS3456ppe - 1./288.*m11*RS345pe*RSe45633p - 1./
         288.*m11*RS345pe*RSp45633e - 1./288.*m11*pow(RS345pe,2) + 1./
         32.*m11*RS346pe + 1./72.*m11*RS346pe*RS34561pe + 1./48.*m11*
         RS346pe*RS14563pe );

      F13p6x23R +=  + Ab(m33,mu2)*pi16 * ( 1./144.*m11*RS346pe*
         RSe45613p + 1./144.*m11*RS346pe*RSp45613e + 1./144.*m11*
         RS346pe*RS3456pee + 1./144.*m11*RS346pe*RS3456ppe - 1./288.*
         m11*RS346pe*RSe45633p - 1./288.*m11*RS346pe*RSp45633e - 1./144.
         *m11*RS346pe*RS345pe - 1./288.*m11*pow(RS346pe,2) + 1./32.*m11
         *RS356pe + 1./72.*m11*RS356pe*RS34561pe + 1./48.*m11*RS356pe*
         RS14563pe + 1./144.*m11*RS356pe*RSe45613p + 1./144.*m11*
         RS356pe*RSp45613e + 1./144.*m11*RS356pe*RS3456pee + 1./144.*
         m11*RS356pe*RS3456ppe - 1./288.*m11*RS356pe*RSe45633p - 1./288.
         *m11*RS356pe*RSp45633e - 1./144.*m11*RS356pe*RS345pe - 1./144.
         *m11*RS356pe*RS346pe - 1./288.*m11*pow(RS356pe,2) - 1./48.*m11
         *RS1456pee*RS34561pe - 1./96.*m11*RS1456pee*RS3456pee - 1./96.
         *m11*RS1456pee*RS3456ppe + 1./96.*m11*RS1456pee*RS345pe + 1./
         96.*m11*RS1456pee*RS346pe + 1./96.*m11*RS1456pee*RS356pe - 1./
         48.*m11*RS1456ppe*RS34561pe - 1./96.*m11*RS1456ppe*RS3456pee
          - 1./96.*m11*RS1456ppe*RS3456ppe );

      F13p6x23R +=  + Ab(m33,mu2)*pi16 * ( 1./96.*m11*RS1456ppe*RS345pe
          + 1./96.*m11*RS1456ppe*RS346pe + 1./96.*m11*RS1456ppe*RS356pe
          + 1./144.*m11*RSe45611p*RS34561pe + 1./288.*m11*RSe45611p*
         RS3456pee + 1./288.*m11*RSe45611p*RS3456ppe - 1./288.*m11*
         RSe45611p*RS345pe - 1./288.*m11*RSe45611p*RS346pe - 1./288.*
         m11*RSe45611p*RS356pe + 1./144.*m11*RSp45611e*RS34561pe + 1./
         288.*m11*RSp45611e*RS3456pee + 1./288.*m11*RSp45611e*RS3456ppe
          - 1./288.*m11*RSp45611e*RS345pe - 1./288.*m11*RSp45611e*
         RS346pe - 1./288.*m11*RSp45611e*RS356pe + 1./48.*m11*RS145pe*
         RS34561pe + 1./96.*m11*RS145pe*RS3456pee + 1./96.*m11*RS145pe*
         RS3456ppe - 1./96.*m11*RS145pe*RS345pe - 1./96.*m11*RS145pe*
         RS346pe - 1./96.*m11*RS145pe*RS356pe + 1./48.*m11*RS146pe*
         RS34561pe + 1./96.*m11*RS146pe*RS3456pee + 1./96.*m11*RS146pe*
         RS3456ppe - 1./96.*m11*RS146pe*RS345pe - 1./96.*m11*RS146pe*
         RS346pe - 1./96.*m11*RS146pe*RS356pe + 1./48.*m11*RS156pe*
         RS34561pe );

      F13p6x23R +=  + Ab(m33,mu2)*pi16 * ( 1./96.*m11*RS156pe*RS3456pee
          + 1./96.*m11*RS156pe*RS3456ppe - 1./96.*m11*RS156pe*RS345pe
          - 1./96.*m11*RS156pe*RS346pe - 1./96.*m11*RS156pe*RS356pe );

      F13p6x23R +=  + pow(Ab(m33,mu2),2) * ( 1./16. + 1./72.*pow(
         RS34561pe,2) + 1./72.*RS3456pee*RS34561pe + 1./288.*pow(
         RS3456pee,2) + 1./72.*RS3456ppe*RS34561pe + 1./144.*RS3456ppe*
         RS3456pee + 1./288.*pow(RS3456ppe,2) - 1./72.*RS345pe*
         RS34561pe - 1./144.*RS345pe*RS3456pee - 1./144.*RS345pe*
         RS3456ppe + 1./288.*pow(RS345pe,2) - 1./72.*RS346pe*RS34561pe
          - 1./144.*RS346pe*RS3456pee - 1./144.*RS346pe*RS3456ppe + 1./
         144.*RS346pe*RS345pe + 1./288.*pow(RS346pe,2) - 1./72.*RS356pe
         *RS34561pe - 1./144.*RS356pe*RS3456pee - 1./144.*RS356pe*
         RS3456ppe + 1./144.*RS356pe*RS345pe + 1./144.*RS356pe*RS346pe
          + 1./288.*pow(RS356pe,2) + 1./18.*m33*RN3p*RS3456pee*
         RSp45613e + 1./18.*m33*RN3p*RS3456ppe*RSp45613e - 1./18.*m33*
         RN3p*RSp45633e*RS3456pee - 1./18.*m33*RN3p*RSp45633e*RS3456ppe
          - 1./18.*m33*RN3p*RS345pe*RSp45613e + 1./18.*m33*RN3p*RS345pe
         *RSp45633e - 1./18.*m33*RN3p*RS346pe*RSp45613e + 1./18.*m33*
         RN3p*RS346pe*RSp45633e );

      F13p6x23R +=  + pow(Ab(m33,mu2),2) * (  - 1./18.*m33*RN3p*RS356pe
         *RSp45613e + 1./18.*m33*RN3p*RS356pe*RSp45633e + 1./18.*m33*
         RN3e*RS3456pee*RSe45613p + 1./18.*m33*RN3e*RS3456ppe*RSe45613p
          - 1./18.*m33*RN3e*RSe45633p*RS3456pee - 1./18.*m33*RN3e*
         RSe45633p*RS3456ppe - 1./18.*m33*RN3e*RS345pe*RSe45613p + 1./
         18.*m33*RN3e*RS345pe*RSe45633p - 1./18.*m33*RN3e*RS346pe*
         RSe45613p + 1./18.*m33*RN3e*RS346pe*RSe45633p - 1./18.*m33*
         RN3e*RS356pe*RSe45613p + 1./18.*m33*RN3e*RS356pe*RSe45633p - 1.
         /18.*m33*RN13*RS3456pee*RS14563pe - 1./18.*m33*RN13*RS3456ppe*
         RS14563pe + 1./18.*m33*RN13*RS345pe*RS14563pe + 1./18.*m33*
         RN13*RS346pe*RS14563pe + 1./18.*m33*RN13*RS356pe*RS14563pe );

      F13p6x23R +=  + Ab(m33,mu2)*Ab(m34,mu2) * (  - 1./36.*RS34561pe
          - 1./18.*RS14563pe - 1./18.*RSe45613p - 1./18.*RSp45613e - 7./
         144.*RS3456pee - 7./144.*RS3456ppe + 1./18.*RSe45633p + 1./18.
         *RSp45633e + 7./144.*RS345pe + 7./144.*RS346pe - 1./48.*
         RS356pe - 1./18.*RSe563p - 1./18.*RSp563e + 1./18.*RS156pe + 1.
         /18.*RSe561p + 1./18.*RSp561e - 1./18.*mp2*RN3p*RSp45613e + 1./
         18.*mp2*RN3p*RSp45633e - 1./18.*mp2*RN3p*RSp563e + 1./18.*mp2*
         RN3p*RSp561e - 1./18.*me2*RN3e*RSe45613p + 1./18.*me2*RN3e*
         RSe45633p - 1./18.*me2*RN3e*RSe563p + 1./18.*me2*RN3e*RSe561p
          - 1./18.*m44*RN3p*RSp563e + 1./18.*m44*RN3p*RSp561e - 1./18.*
         m44*RN3e*RSe563p + 1./18.*m44*RN3e*RSe561p - 1./18.*m44*RN13*
         RS156pe + 1./18.*m33*RN3p*RSp45613e - 1./18.*m33*RN3p*
         RSp45633e - 1./18.*m33*RN3p*RSp563e + 1./18.*m33*RN3p*RSp561e
          + 1./18.*m33*RN3e*RSe45613p - 1./18.*m33*RN3e*RSe45633p - 1./
         18.*m33*RN3e*RSe563p + 1./18.*m33*RN3e*RSe561p - 1./18.*m33*
         RN13*RS14563pe );

      F13p6x23R +=  + Ab(m33,mu2)*Ab(m34,mu2) * (  - 1./18.*m33*RN13*
         RS156pe + 1./18.*m11*RN13*RS14563pe - 1./18.*m11*RN13*RS156pe
          );

      F13p6x23R +=  + Ab(m33,mu2)*Ab(m35,mu2) * (  - 1./36.*RS34561pe
          - 1./18.*RS14563pe - 1./18.*RSe45613p - 1./18.*RSp45613e - 7./
         144.*RS3456pee - 7./144.*RS3456ppe + 1./18.*RSe45633p + 1./18.
         *RSp45633e + 7./144.*RS345pe - 1./48.*RS346pe + 7./144.*
         RS356pe - 1./18.*RSe463p - 1./18.*RSp463e + 1./18.*RS146pe + 1.
         /18.*RSe461p + 1./18.*RSp461e - 1./18.*mp2*RN3p*RSp45613e + 1./
         18.*mp2*RN3p*RSp45633e - 1./18.*mp2*RN3p*RSp463e + 1./18.*mp2*
         RN3p*RSp461e - 1./18.*me2*RN3e*RSe45613p + 1./18.*me2*RN3e*
         RSe45633p - 1./18.*me2*RN3e*RSe463p + 1./18.*me2*RN3e*RSe461p
          - 1./18.*m55*RN3p*RSp463e + 1./18.*m55*RN3p*RSp461e - 1./18.*
         m55*RN3e*RSe463p + 1./18.*m55*RN3e*RSe461p - 1./18.*m55*RN13*
         RS146pe + 1./18.*m33*RN3p*RSp45613e - 1./18.*m33*RN3p*
         RSp45633e - 1./18.*m33*RN3p*RSp463e + 1./18.*m33*RN3p*RSp461e
          + 1./18.*m33*RN3e*RSe45613p - 1./18.*m33*RN3e*RSe45633p - 1./
         18.*m33*RN3e*RSe463p + 1./18.*m33*RN3e*RSe461p - 1./18.*m33*
         RN13*RS14563pe );

      F13p6x23R +=  + Ab(m33,mu2)*Ab(m35,mu2) * (  - 1./18.*m33*RN13*
         RS146pe + 1./18.*m11*RN13*RS14563pe - 1./18.*m11*RN13*RS146pe
          );

      F13p6x23R +=  + Ab(m33,mu2)*Ab(m36,mu2) * (  - 1./36.*RS34561pe
          - 1./18.*RS14563pe - 1./18.*RSe45613p - 1./18.*RSp45613e - 7./
         144.*RS3456pee - 7./144.*RS3456ppe + 1./18.*RSe45633p + 1./18.
         *RSp45633e - 1./48.*RS345pe + 7./144.*RS346pe + 7./144.*
         RS356pe - 1./18.*RSe453p - 1./18.*RSp453e + 1./18.*RS145pe + 1.
         /18.*RSe451p + 1./18.*RSp451e - 1./18.*mp2*RN3p*RSp45613e + 1./
         18.*mp2*RN3p*RSp45633e - 1./18.*mp2*RN3p*RSp453e + 1./18.*mp2*
         RN3p*RSp451e - 1./18.*me2*RN3e*RSe45613p + 1./18.*me2*RN3e*
         RSe45633p - 1./18.*me2*RN3e*RSe453p + 1./18.*me2*RN3e*RSe451p
          - 1./18.*m66*RN3p*RSp453e + 1./18.*m66*RN3p*RSp451e - 1./18.*
         m66*RN3e*RSe453p + 1./18.*m66*RN3e*RSe451p - 1./18.*m66*RN13*
         RS145pe + 1./18.*m33*RN3p*RSp45613e - 1./18.*m33*RN3p*
         RSp45633e - 1./18.*m33*RN3p*RSp453e + 1./18.*m33*RN3p*RSp451e
          + 1./18.*m33*RN3e*RSe45613p - 1./18.*m33*RN3e*RSe45633p - 1./
         18.*m33*RN3e*RSe453p + 1./18.*m33*RN3e*RSe451p - 1./18.*m33*
         RN13*RS14563pe );

      F13p6x23R +=  + Ab(m33,mu2)*Ab(m36,mu2) * (  - 1./18.*m33*RN13*
         RS145pe + 1./18.*m11*RN13*RS14563pe - 1./18.*m11*RN13*RS145pe
          );

      F13p6x23R +=  + Ab(m33,mu2)*Ab(m44,mu2) * ( 1./18.*m44*RN3p*
         RSp563e*RS356pe - 1./54.*m44*RN3p*RS456pe*RSp563e*RS356pe - 1./
         18.*m44*RN3p*RSp561e*RS356pe + 1./54.*m44*RN3p*RSp561e*RS456pe
         *RS356pe + 1./18.*m44*RN3e*RSe563p*RS356pe - 1./54.*m44*RN3e*
         RS456pe*RSe563p*RS356pe - 1./18.*m44*RN3e*RSe561p*RS356pe + 1./
         54.*m44*RN3e*RSe561p*RS456pe*RS356pe + 1./18.*m44*RN13*RS156pe
         *RS356pe - 1./54.*m44*RN13*RS456pe*RS156pe*RS356pe );

      F13p6x23R +=  + Ab(m33,mu2)*Ab(m45,mu2) * (  - 1./108.*pow(
         RS346pe,2) + 1./54.*RS356pe*RS346pe - 1./108.*pow(RS356pe,2)
          + 1./54.*RSe563p*RS346pe - 1./54.*RSe563p*RS356pe - 1./54.*
         RSe463p*RS346pe + 1./54.*RSe463p*RS356pe + 1./54.*RSp563e*
         RS346pe - 1./54.*RSp563e*RS356pe - 1./54.*RSp463e*RS346pe + 1./
         54.*RSp463e*RS356pe + 1./54.*RS146pe*RS346pe - 1./54.*RS146pe*
         RS356pe - 1./54.*RS156pe*RS346pe + 1./54.*RS156pe*RS356pe - 1./
         54.*RSe561p*RS346pe + 1./54.*RSe561p*RS356pe + 1./54.*RSe461p*
         RS346pe - 1./54.*RSe461p*RS356pe - 1./54.*RSp561e*RS346pe + 1./
         54.*RSp561e*RS356pe + 1./54.*RSp461e*RS346pe - 1./54.*RSp461e*
         RS356pe + 1./54.*mp2*RN3p*RSp563e*RS346pe - 1./54.*mp2*RN3p*
         RSp563e*RS356pe - 1./54.*mp2*RN3p*RSp463e*RS346pe + 1./54.*mp2
         *RN3p*RSp463e*RS356pe - 1./54.*mp2*RN3p*RSp561e*RS346pe + 1./
         54.*mp2*RN3p*RSp561e*RS356pe + 1./54.*mp2*RN3p*RSp461e*RS346pe
          - 1./54.*mp2*RN3p*RSp461e*RS356pe + 1./54.*me2*RN3e*RSe563p*
         RS346pe );

      F13p6x23R +=  + Ab(m33,mu2)*Ab(m45,mu2) * (  - 1./54.*me2*RN3e*
         RSe563p*RS356pe - 1./54.*me2*RN3e*RSe463p*RS346pe + 1./54.*me2
         *RN3e*RSe463p*RS356pe - 1./54.*me2*RN3e*RSe561p*RS346pe + 1./
         54.*me2*RN3e*RSe561p*RS356pe + 1./54.*me2*RN3e*RSe461p*RS346pe
          - 1./54.*me2*RN3e*RSe461p*RS356pe + 1./54.*m55*RN3p*RSp563e*
         RS346pe + 1./54.*m55*RN3p*RSp463e*RS346pe + 1./54.*m55*RN3p*
         RSp463e*RS356pe - 1./54.*m55*RN3p*RSp561e*RS346pe - 1./54.*m55
         *RN3p*RSp461e*RS346pe - 1./54.*m55*RN3p*RSp461e*RS356pe + 1./
         54.*m55*RN3e*RSe563p*RS346pe + 1./54.*m55*RN3e*RSe463p*RS346pe
          + 1./54.*m55*RN3e*RSe463p*RS356pe - 1./54.*m55*RN3e*RSe561p*
         RS346pe - 1./54.*m55*RN3e*RSe461p*RS346pe - 1./54.*m55*RN3e*
         RSe461p*RS356pe + 1./54.*m55*RN13*RS146pe*RS346pe + 1./54.*m55
         *RN13*RS146pe*RS356pe + 1./54.*m55*RN13*RS156pe*RS346pe + 1./
         54.*m44*RN3p*RSp563e*RS346pe + 1./54.*m44*RN3p*RSp563e*RS356pe
          + 1./54.*m44*RN3p*RSp463e*RS356pe - 1./54.*m44*RN3p*RSp561e*
         RS346pe );

      F13p6x23R +=  + Ab(m33,mu2)*Ab(m45,mu2) * (  - 1./54.*m44*RN3p*
         RSp561e*RS356pe - 1./54.*m44*RN3p*RSp461e*RS356pe + 1./54.*m44
         *RN3e*RSe563p*RS346pe + 1./54.*m44*RN3e*RSe563p*RS356pe + 1./
         54.*m44*RN3e*RSe463p*RS356pe - 1./54.*m44*RN3e*RSe561p*RS346pe
          - 1./54.*m44*RN3e*RSe561p*RS356pe - 1./54.*m44*RN3e*RSe461p*
         RS356pe + 1./54.*m44*RN13*RS146pe*RS356pe + 1./54.*m44*RN13*
         RS156pe*RS346pe + 1./54.*m44*RN13*RS156pe*RS356pe - 1./54.*m11
         *RN13*RS146pe*RS346pe + 1./54.*m11*RN13*RS146pe*RS356pe + 1./
         54.*m11*RN13*RS156pe*RS346pe - 1./54.*m11*RN13*RS156pe*RS356pe
          );

      F13p6x23R +=  + Ab(m33,mu2)*Ab(m46,mu2) * (  - 1./108.*pow(
         RS345pe,2) + 1./54.*RS356pe*RS345pe - 1./108.*pow(RS356pe,2)
          + 1./54.*RSe563p*RS345pe - 1./54.*RSe563p*RS356pe - 1./54.*
         RSe453p*RS345pe + 1./54.*RSe453p*RS356pe + 1./54.*RSp563e*
         RS345pe - 1./54.*RSp563e*RS356pe - 1./54.*RSp453e*RS345pe + 1./
         54.*RSp453e*RS356pe + 1./54.*RS145pe*RS345pe - 1./54.*RS145pe*
         RS356pe - 1./54.*RS156pe*RS345pe + 1./54.*RS156pe*RS356pe - 1./
         54.*RSe561p*RS345pe + 1./54.*RSe561p*RS356pe + 1./54.*RSe451p*
         RS345pe - 1./54.*RSe451p*RS356pe - 1./54.*RSp561e*RS345pe + 1./
         54.*RSp561e*RS356pe + 1./54.*RSp451e*RS345pe - 1./54.*RSp451e*
         RS356pe + 1./54.*mp2*RN3p*RSp563e*RS345pe - 1./54.*mp2*RN3p*
         RSp563e*RS356pe - 1./54.*mp2*RN3p*RSp453e*RS345pe + 1./54.*mp2
         *RN3p*RSp453e*RS356pe - 1./54.*mp2*RN3p*RSp561e*RS345pe + 1./
         54.*mp2*RN3p*RSp561e*RS356pe + 1./54.*mp2*RN3p*RSp451e*RS345pe
          - 1./54.*mp2*RN3p*RSp451e*RS356pe + 1./54.*me2*RN3e*RSe563p*
         RS345pe );

      F13p6x23R +=  + Ab(m33,mu2)*Ab(m46,mu2) * (  - 1./54.*me2*RN3e*
         RSe563p*RS356pe - 1./54.*me2*RN3e*RSe453p*RS345pe + 1./54.*me2
         *RN3e*RSe453p*RS356pe - 1./54.*me2*RN3e*RSe561p*RS345pe + 1./
         54.*me2*RN3e*RSe561p*RS356pe + 1./54.*me2*RN3e*RSe451p*RS345pe
          - 1./54.*me2*RN3e*RSe451p*RS356pe + 1./54.*m66*RN3p*RSp563e*
         RS345pe + 1./54.*m66*RN3p*RSp453e*RS345pe + 1./54.*m66*RN3p*
         RSp453e*RS356pe - 1./54.*m66*RN3p*RSp561e*RS345pe - 1./54.*m66
         *RN3p*RSp451e*RS345pe - 1./54.*m66*RN3p*RSp451e*RS356pe + 1./
         54.*m66*RN3e*RSe563p*RS345pe + 1./54.*m66*RN3e*RSe453p*RS345pe
          + 1./54.*m66*RN3e*RSe453p*RS356pe - 1./54.*m66*RN3e*RSe561p*
         RS345pe - 1./54.*m66*RN3e*RSe451p*RS345pe - 1./54.*m66*RN3e*
         RSe451p*RS356pe + 1./54.*m66*RN13*RS145pe*RS345pe + 1./54.*m66
         *RN13*RS145pe*RS356pe + 1./54.*m66*RN13*RS156pe*RS345pe + 1./
         54.*m44*RN3p*RSp563e*RS345pe + 1./54.*m44*RN3p*RSp563e*RS356pe
          + 1./54.*m44*RN3p*RSp453e*RS356pe - 1./54.*m44*RN3p*RSp561e*
         RS345pe );

      F13p6x23R +=  + Ab(m33,mu2)*Ab(m46,mu2) * (  - 1./54.*m44*RN3p*
         RSp561e*RS356pe - 1./54.*m44*RN3p*RSp451e*RS356pe + 1./54.*m44
         *RN3e*RSe563p*RS345pe + 1./54.*m44*RN3e*RSe563p*RS356pe + 1./
         54.*m44*RN3e*RSe453p*RS356pe - 1./54.*m44*RN3e*RSe561p*RS345pe
          - 1./54.*m44*RN3e*RSe561p*RS356pe - 1./54.*m44*RN3e*RSe451p*
         RS356pe + 1./54.*m44*RN13*RS145pe*RS356pe + 1./54.*m44*RN13*
         RS156pe*RS345pe + 1./54.*m44*RN13*RS156pe*RS356pe - 1./54.*m11
         *RN13*RS145pe*RS345pe + 1./54.*m11*RN13*RS145pe*RS356pe + 1./
         54.*m11*RN13*RS156pe*RS345pe - 1./54.*m11*RN13*RS156pe*RS356pe
          );

      F13p6x23R +=  + Ab(m33,mu2)*Ab(m55,mu2) * ( 1./18.*m55*RN3p*
         RSp463e*RS346pe - 1./54.*m55*RN3p*RS546pe*RSp463e*RS346pe - 1./
         18.*m55*RN3p*RSp461e*RS346pe + 1./54.*m55*RN3p*RSp461e*RS546pe
         *RS346pe + 1./18.*m55*RN3e*RSe463p*RS346pe - 1./54.*m55*RN3e*
         RS546pe*RSe463p*RS346pe - 1./18.*m55*RN3e*RSe461p*RS346pe + 1./
         54.*m55*RN3e*RSe461p*RS546pe*RS346pe + 1./18.*m55*RN13*RS146pe
         *RS346pe - 1./54.*m55*RN13*RS546pe*RS146pe*RS346pe );

      F13p6x23R +=  + Ab(m33,mu2)*Ab(m56,mu2) * (  - 1./108.*pow(
         RS345pe,2) + 1./54.*RS346pe*RS345pe - 1./108.*pow(RS346pe,2)
          + 1./54.*RSe463p*RS345pe - 1./54.*RSe463p*RS346pe - 1./54.*
         RSe453p*RS345pe + 1./54.*RSe453p*RS346pe + 1./54.*RSp463e*
         RS345pe - 1./54.*RSp463e*RS346pe - 1./54.*RSp453e*RS345pe + 1./
         54.*RSp453e*RS346pe + 1./54.*RS145pe*RS345pe - 1./54.*RS145pe*
         RS346pe - 1./54.*RS146pe*RS345pe + 1./54.*RS146pe*RS346pe - 1./
         54.*RSe461p*RS345pe + 1./54.*RSe461p*RS346pe + 1./54.*RSe451p*
         RS345pe - 1./54.*RSe451p*RS346pe - 1./54.*RSp461e*RS345pe + 1./
         54.*RSp461e*RS346pe + 1./54.*RSp451e*RS345pe - 1./54.*RSp451e*
         RS346pe + 1./54.*mp2*RN3p*RSp463e*RS345pe - 1./54.*mp2*RN3p*
         RSp463e*RS346pe - 1./54.*mp2*RN3p*RSp453e*RS345pe + 1./54.*mp2
         *RN3p*RSp453e*RS346pe - 1./54.*mp2*RN3p*RSp461e*RS345pe + 1./
         54.*mp2*RN3p*RSp461e*RS346pe + 1./54.*mp2*RN3p*RSp451e*RS345pe
          - 1./54.*mp2*RN3p*RSp451e*RS346pe + 1./54.*me2*RN3e*RSe463p*
         RS345pe );

      F13p6x23R +=  + Ab(m33,mu2)*Ab(m56,mu2) * (  - 1./54.*me2*RN3e*
         RSe463p*RS346pe - 1./54.*me2*RN3e*RSe453p*RS345pe + 1./54.*me2
         *RN3e*RSe453p*RS346pe - 1./54.*me2*RN3e*RSe461p*RS345pe + 1./
         54.*me2*RN3e*RSe461p*RS346pe + 1./54.*me2*RN3e*RSe451p*RS345pe
          - 1./54.*me2*RN3e*RSe451p*RS346pe + 1./54.*m66*RN3p*RSp463e*
         RS345pe + 1./54.*m66*RN3p*RSp453e*RS345pe + 1./54.*m66*RN3p*
         RSp453e*RS346pe - 1./54.*m66*RN3p*RSp461e*RS345pe - 1./54.*m66
         *RN3p*RSp451e*RS345pe - 1./54.*m66*RN3p*RSp451e*RS346pe + 1./
         54.*m66*RN3e*RSe463p*RS345pe + 1./54.*m66*RN3e*RSe453p*RS345pe
          + 1./54.*m66*RN3e*RSe453p*RS346pe - 1./54.*m66*RN3e*RSe461p*
         RS345pe - 1./54.*m66*RN3e*RSe451p*RS345pe - 1./54.*m66*RN3e*
         RSe451p*RS346pe + 1./54.*m66*RN13*RS145pe*RS345pe + 1./54.*m66
         *RN13*RS145pe*RS346pe + 1./54.*m66*RN13*RS146pe*RS345pe + 1./
         54.*m55*RN3p*RSp463e*RS345pe + 1./54.*m55*RN3p*RSp463e*RS346pe
          + 1./54.*m55*RN3p*RSp453e*RS346pe - 1./54.*m55*RN3p*RSp461e*
         RS345pe );

      F13p6x23R +=  + Ab(m33,mu2)*Ab(m56,mu2) * (  - 1./54.*m55*RN3p*
         RSp461e*RS346pe - 1./54.*m55*RN3p*RSp451e*RS346pe + 1./54.*m55
         *RN3e*RSe463p*RS345pe + 1./54.*m55*RN3e*RSe463p*RS346pe + 1./
         54.*m55*RN3e*RSe453p*RS346pe - 1./54.*m55*RN3e*RSe461p*RS345pe
          - 1./54.*m55*RN3e*RSe461p*RS346pe - 1./54.*m55*RN3e*RSe451p*
         RS346pe + 1./54.*m55*RN13*RS145pe*RS346pe + 1./54.*m55*RN13*
         RS146pe*RS345pe + 1./54.*m55*RN13*RS146pe*RS346pe - 1./54.*m11
         *RN13*RS145pe*RS345pe + 1./54.*m11*RN13*RS145pe*RS346pe + 1./
         54.*m11*RN13*RS146pe*RS345pe - 1./54.*m11*RN13*RS146pe*RS346pe
          );

      F13p6x23R +=  + Ab(m33,mu2)*Ab(m66,mu2) * ( 1./18.*m66*RN3p*
         RSp453e*RS345pe - 1./54.*m66*RN3p*RS645pe*RSp453e*RS345pe - 1./
         18.*m66*RN3p*RSp451e*RS345pe + 1./54.*m66*RN3p*RSp451e*RS645pe
         *RS345pe + 1./18.*m66*RN3e*RSe453p*RS345pe - 1./54.*m66*RN3e*
         RS645pe*RSe453p*RS345pe - 1./18.*m66*RN3e*RSe451p*RS345pe + 1./
         54.*m66*RN3e*RSe451p*RS645pe*RS345pe + 1./18.*m66*RN13*RS145pe
         *RS345pe - 1./54.*m66*RN13*RS645pe*RS145pe*RS345pe );

      F13p6x23R +=  + Ab(m33,mu2)*Ab(me2,mu2) * ( 1./36.*RSe45613p*
         RS34561pe + 1./72.*RS3456pee*RSe45613p + 1./72.*RS3456ppe*
         RSe45613p - 1./72.*RSe45633p*RS34561pe - 1./144.*RSe45633p*
         RS3456pee - 1./144.*RSe45633p*RS3456ppe - 1./72.*RS345pe*
         RSe45613p + 1./144.*RS345pe*RSe45633p - 1./72.*RS346pe*
         RSe45613p + 1./144.*RS346pe*RSe45633p - 1./72.*RS356pe*
         RSe45613p + 1./144.*RS356pe*RSe45633p - 1./72.*RSe45611p*
         RS34561pe - 1./144.*RSe45611p*RS3456pee - 1./144.*RSe45611p*
         RS3456ppe + 1./144.*RSe45611p*RS345pe + 1./144.*RSe45611p*
         RS346pe + 1./144.*RSe45611p*RS356pe - 1./54.*m66*RN3p*RSe456p*
         RSp453e*RS345pe + 1./54.*m66*RN3p*RSe456p*RSp451e*RS345pe - 1./
         54.*m66*RN3e*RSe456p*RSe453p*RS345pe + 1./54.*m66*RN3e*RSe456p
         *RSe451p*RS345pe - 1./54.*m66*RN13*RSe456p*RS145pe*RS345pe - 1.
         /54.*m55*RN3p*RSe465p*RSp463e*RS346pe + 1./54.*m55*RN3p*
         RSe465p*RSp461e*RS346pe - 1./54.*m55*RN3e*RSe465p*RSe463p*
         RS346pe );

      F13p6x23R +=  + Ab(m33,mu2)*Ab(me2,mu2) * ( 1./54.*m55*RN3e*
         RSe465p*RSe461p*RS346pe - 1./54.*m55*RN13*RSe465p*RS146pe*
         RS346pe - 1./54.*m44*RN3p*RSe564p*RSp563e*RS356pe + 1./54.*m44
         *RN3p*RSe564p*RSp561e*RS356pe - 1./54.*m44*RN3e*RSe564p*
         RSe563p*RS356pe + 1./54.*m44*RN3e*RSe564p*RSe561p*RS356pe - 1./
         54.*m44*RN13*RSe564p*RS156pe*RS356pe - 1./18.*m33*RN3p*
         RSe45633p*RSp45613e + 1./18.*m33*RN3p*RSp45633e*RSe45633p - 1./
         18.*m33*RN3e*RS3456pee*RSe45613p - 1./18.*m33*RN3e*RS3456ppe*
         RSe45613p - 1./18.*m33*RN3e*RSe45633p*RSe45613p + 1./18.*m33*
         RN3e*RSe45633p*RS3456pee + 1./18.*m33*RN3e*RSe45633p*RS3456ppe
          + 1./18.*m33*RN3e*pow(RSe45633p,2) + 1./18.*m33*RN3e*RS345pe*
         RSe45613p - 1./18.*m33*RN3e*RS345pe*RSe45633p + 1./18.*m33*
         RN3e*RS346pe*RSe45613p - 1./18.*m33*RN3e*RS346pe*RSe45633p + 1.
         /18.*m33*RN3e*RS356pe*RSe45613p - 1./18.*m33*RN3e*RS356pe*
         RSe45633p + 1./18.*m33*RN13*RSe45633p*RS14563pe + 1./18.*m11*
         RN13*RSe45611p*RS34561pe );

      F13p6x23R +=  + Ab(m33,mu2)*Ab(mp2,mu2) * ( 1./36.*RSp45613e*
         RS34561pe + 1./72.*RS3456pee*RSp45613e + 1./72.*RS3456ppe*
         RSp45613e - 1./72.*RSp45633e*RS34561pe - 1./144.*RSp45633e*
         RS3456pee - 1./144.*RSp45633e*RS3456ppe - 1./72.*RS345pe*
         RSp45613e + 1./144.*RS345pe*RSp45633e - 1./72.*RS346pe*
         RSp45613e + 1./144.*RS346pe*RSp45633e - 1./72.*RS356pe*
         RSp45613e + 1./144.*RS356pe*RSp45633e - 1./72.*RSp45611e*
         RS34561pe - 1./144.*RSp45611e*RS3456pee - 1./144.*RSp45611e*
         RS3456ppe + 1./144.*RSp45611e*RS345pe + 1./144.*RSp45611e*
         RS346pe + 1./144.*RSp45611e*RS356pe - 1./54.*m66*RN3p*RSp456e*
         RSp453e*RS345pe + 1./54.*m66*RN3p*RSp456e*RSp451e*RS345pe - 1./
         54.*m66*RN3e*RSp456e*RSe453p*RS345pe + 1./54.*m66*RN3e*RSp456e
         *RSe451p*RS345pe - 1./54.*m66*RN13*RSp456e*RS145pe*RS345pe - 1.
         /54.*m55*RN3p*RSp465e*RSp463e*RS346pe + 1./54.*m55*RN3p*
         RSp465e*RSp461e*RS346pe - 1./54.*m55*RN3e*RSp465e*RSe463p*
         RS346pe );

      F13p6x23R +=  + Ab(m33,mu2)*Ab(mp2,mu2) * ( 1./54.*m55*RN3e*
         RSp465e*RSe461p*RS346pe - 1./54.*m55*RN13*RSp465e*RS146pe*
         RS346pe - 1./54.*m44*RN3p*RSp564e*RSp563e*RS356pe + 1./54.*m44
         *RN3p*RSp564e*RSp561e*RS356pe - 1./54.*m44*RN3e*RSp564e*
         RSe563p*RS356pe + 1./54.*m44*RN3e*RSp564e*RSe561p*RS356pe - 1./
         54.*m44*RN13*RSp564e*RS156pe*RS356pe - 1./18.*m33*RN3p*
         RS3456pee*RSp45613e - 1./18.*m33*RN3p*RS3456ppe*RSp45613e - 1./
         18.*m33*RN3p*RSp45633e*RSp45613e + 1./18.*m33*RN3p*RSp45633e*
         RS3456pee + 1./18.*m33*RN3p*RSp45633e*RS3456ppe + 1./18.*m33*
         RN3p*pow(RSp45633e,2) + 1./18.*m33*RN3p*RS345pe*RSp45613e - 1./
         18.*m33*RN3p*RS345pe*RSp45633e + 1./18.*m33*RN3p*RS346pe*
         RSp45613e - 1./18.*m33*RN3p*RS346pe*RSp45633e + 1./18.*m33*
         RN3p*RS356pe*RSp45613e - 1./18.*m33*RN3p*RS356pe*RSp45633e - 1.
         /18.*m33*RN3e*RSp45633e*RSe45613p + 1./18.*m33*RN3e*RSp45633e*
         RSe45633p + 1./18.*m33*RN13*RSp45633e*RS14563pe + 1./18.*m11*
         RN13*RSp45611e*RS34561pe );

      F13p6x23R +=  + Ab(m33,mu2)*Bb(m11,mu2) * (  - 1./72.*RD1456ep*
         RS34561pe - 1./144.*RD1456ep*RS3456pee - 1./144.*RD1456ep*
         RS3456ppe + 1./144.*RD1456ep*RS345pe + 1./144.*RD1456ep*
         RS346pe + 1./144.*RD1456ep*RS356pe + 1./18.*m11*RN13*RD1456ep*
         RS34561pe );

      F13p6x23R +=  + Ab(m33,mu2)*Bb(m33,mu2) * (  - 1./72.*RD3456ep*
         RS34561pe - 1./144.*RS3456pee*RD3456ep - 1./144.*RS3456ppe*
         RD3456ep + 1./144.*RS345pe*RD3456ep + 1./144.*RS346pe*RD3456ep
          + 1./144.*RS356pe*RD3456ep + 1./4.*m33 + 1./18.*m33*RS3456pee
         *RS34561pe + 1./18.*m33*pow(RS3456pee,2) + 1./18.*m33*
         RS3456ppe*RS34561pe + 1./9.*m33*RS3456ppe*RS3456pee + 1./18.*
         m33*pow(RS3456ppe,2) - 1./18.*m33*RS345pe*RS34561pe - 1./9.*
         m33*RS345pe*RS3456pee - 1./9.*m33*RS345pe*RS3456ppe + 1./18.*
         m33*pow(RS345pe,2) - 1./18.*m33*RS346pe*RS34561pe - 1./9.*m33*
         RS346pe*RS3456pee - 1./9.*m33*RS346pe*RS3456ppe + 1./9.*m33*
         RS346pe*RS345pe + 1./18.*m33*pow(RS346pe,2) - 1./18.*m33*
         RS356pe*RS34561pe - 1./9.*m33*RS356pe*RS3456pee - 1./9.*m33*
         RS356pe*RS3456ppe + 1./9.*m33*RS356pe*RS345pe + 1./9.*m33*
         RS356pe*RS346pe + 1./18.*m33*pow(RS356pe,2) - 1./18.*m33*RN3p*
         RD3456ep*RSp45613e + 1./18.*m33*RN3p*RSp45633e*RD3456ep - 1./
         18.*m33*RN3e*RD3456ep*RSe45613p );

      F13p6x23R +=  + Ab(m33,mu2)*Bb(m33,mu2) * ( 1./18.*m33*RN3e*
         RSe45633p*RD3456ep + 1./18.*m33*RN13*RD3456ep*RS14563pe );

      F13p6x23R +=  + Ab(m33,mu2)*Bb(m34,mu2) * ( 1./72.*RD3456ep - 1./
         36.*m44*RS356pe - 1./18.*m33*RS356pe );

      F13p6x23R +=  + Ab(m33,mu2)*Bb(m35,mu2) * ( 1./72.*RD3456ep - 1./
         36.*m55*RS346pe - 1./18.*m33*RS346pe );

      F13p6x23R +=  + Ab(m33,mu2)*Bb(m36,mu2) * ( 1./72.*RD3456ep - 1./
         36.*m66*RS345pe - 1./18.*m33*RS345pe );

      F13p6x23R +=  + Ab(m33,mu2)*Cb(m33,mu2) * (  - 1./18.*m33*
         RS3456pee*RD3456ep - 1./18.*m33*RS3456ppe*RD3456ep + 1./18.*
         m33*RS345pe*RD3456ep + 1./18.*m33*RS346pe*RD3456ep + 1./18.*
         m33*RS356pe*RD3456ep );

      F13p6x23R +=  + Ab(m34,mu2)*pi16 * ( 1./48.*RD1456ep + 1./48.*mp2
         *RSp45611e - 1./24.*mp2*RSp561e + 1./48.*mp2*RSp564e + 1./48.*
         me2*RSe45611p - 1./24.*me2*RSe561p + 1./48.*me2*RSe564p - 1./
         16.*m66 - 1./16.*m55 - 1./8.*m44 - 1./96.*m44*RS34561pe - 1./
         96.*m44*RS14563pe - 1./96.*m44*RSe45613p - 1./96.*m44*
         RSp45613e - 1./96.*m44*RS3456pee - 1./96.*m44*RS3456ppe + 1./
         96.*m44*RSe45633p + 1./96.*m44*RSp45633e + 1./96.*m44*RS345pe
          + 1./96.*m44*RS346pe - 1./96.*m44*RSe563p - 1./96.*m44*
         RSp563e - 1./96.*m44*RS1456pee - 1./96.*m44*RS1456ppe + 1./96.
         *m44*RSe45611p + 1./96.*m44*RSp45611e + 1./96.*m44*RS145pe + 1.
         /96.*m44*RS146pe + 1./32.*m44*RS456pe - 1./96.*m44*RSe561p - 1.
         /96.*m44*RSp561e + 1./96.*m44*RSe564p + 1./96.*m44*RSp564e + 1.
         /32.*m33 - 1./96.*m33*RS34561pe - 1./96.*m33*RS14563pe - 1./96.
         *m33*RSe45613p - 1./96.*m33*RSp45613e - 1./96.*m33*RS3456pee
          - 1./96.*m33*RS3456ppe + 1./96.*m33*RSe45633p + 1./96.*m33*
         RSp45633e );

      F13p6x23R +=  + Ab(m34,mu2)*pi16 * ( 1./96.*m33*RS345pe + 1./96.*
         m33*RS346pe - 1./96.*m33*RSe563p - 1./96.*m33*RSp563e + 1./96.
         *m33*RS156pe + 1./96.*m33*RSe561p + 1./96.*m33*RSp561e - 3./32.
         *m11 - 1./32.*m11*RS1456pee - 1./32.*m11*RS1456ppe + 1./96.*
         m11*RSe45611p + 1./96.*m11*RSp45611e + 1./32.*m11*RS145pe + 1./
         32.*m11*RS146pe - 1./32.*m11*RS156pe + 1./96.*m11*RS456pe - 1./
         48.*m11*RSe561p - 1./48.*m11*RSp561e + 1./96.*m11*RSe564p + 1./
         96.*m11*RSp564e );

      F13p6x23R +=  + pow(Ab(m34,mu2),2) * (  - 1./32. );

      F13p6x23R +=  + Ab(m34,mu2)*Ab(m35,mu2) * (  - 1./16. );

      F13p6x23R +=  + Ab(m34,mu2)*Ab(m36,mu2) * (  - 1./16. );

      F13p6x23R +=  + Ab(m34,mu2)*Ab(me2,mu2) * ( 1./36.*RSe45613p - 1./
         144.*RSe45633p - 1./72.*RSe563p - 1./48.*RSe45611p + 1./72.*
         RSe561p + 1./18.*me2*RN3e*RSe45613p - 1./18.*me2*RN3e*
         RSe45633p + 1./18.*me2*RN3e*RSe563p - 1./18.*me2*RN3e*RSe561p
          + 1./18.*m44*RN3e*RSe563p - 1./18.*m44*RN3e*RSe561p - 1./18.*
         m33*RN3e*RSe45613p + 1./18.*m33*RN3e*RSe45633p + 1./18.*m33*
         RN3e*RSe563p - 1./18.*m33*RN3e*RSe561p );

      F13p6x23R +=  + Ab(m34,mu2)*Ab(mp2,mu2) * ( 1./36.*RSp45613e - 1./
         144.*RSp45633e - 1./72.*RSp563e - 1./48.*RSp45611e + 1./72.*
         RSp561e + 1./18.*mp2*RN3p*RSp45613e - 1./18.*mp2*RN3p*
         RSp45633e + 1./18.*mp2*RN3p*RSp563e - 1./18.*mp2*RN3p*RSp561e
          + 1./18.*m44*RN3p*RSp563e - 1./18.*m44*RN3p*RSp561e - 1./18.*
         m33*RN3p*RSp45613e + 1./18.*m33*RN3p*RSp45633e + 1./18.*m33*
         RN3p*RSp563e - 1./18.*m33*RN3p*RSp561e );

      F13p6x23R +=  + Ab(m34,mu2)*Bb(m11,mu2) * (  - 1./48.*RD1456ep );

      F13p6x23R +=  + Ab(m34,mu2)*Bb(m33,mu2) * ( 7./144.*RD3456ep - 1./
         18.*m44*RS356pe - 1./9.*m33*RS356pe );

      F13p6x23R +=  + Ab(m35,mu2)*pi16 * ( 1./48.*RD1456ep + 1./48.*mp2
         *RSp45611e - 1./24.*mp2*RSp461e + 1./48.*mp2*RSp465e + 1./48.*
         me2*RSe45611p - 1./24.*me2*RSe461p + 1./48.*me2*RSe465p - 1./
         16.*m66 - 1./8.*m55 - 1./96.*m55*RS34561pe - 1./96.*m55*
         RS14563pe - 1./96.*m55*RSe45613p - 1./96.*m55*RSp45613e - 1./
         96.*m55*RS3456pee - 1./96.*m55*RS3456ppe + 1./96.*m55*
         RSe45633p + 1./96.*m55*RSp45633e + 1./96.*m55*RS345pe + 1./96.
         *m55*RS356pe - 1./96.*m55*RSe463p - 1./96.*m55*RSp463e - 1./96.
         *m55*RS1456pee - 1./96.*m55*RS1456ppe + 1./96.*m55*RSe45611p
          + 1./96.*m55*RSp45611e + 1./96.*m55*RS145pe + 1./96.*m55*
         RS156pe + 1./32.*m55*RS546pe - 1./96.*m55*RSe461p - 1./96.*m55
         *RSp461e + 1./96.*m55*RSe465p + 1./96.*m55*RSp465e - 1./16.*
         m44 + 1./32.*m33 - 1./96.*m33*RS34561pe - 1./96.*m33*RS14563pe
          - 1./96.*m33*RSe45613p - 1./96.*m33*RSp45613e - 1./96.*m33*
         RS3456pee - 1./96.*m33*RS3456ppe + 1./96.*m33*RSe45633p + 1./
         96.*m33*RSp45633e );

      F13p6x23R +=  + Ab(m35,mu2)*pi16 * ( 1./96.*m33*RS345pe + 1./96.*
         m33*RS356pe - 1./96.*m33*RSe463p - 1./96.*m33*RSp463e + 1./96.
         *m33*RS146pe + 1./96.*m33*RSe461p + 1./96.*m33*RSp461e - 3./32.
         *m11 - 1./32.*m11*RS1456pee - 1./32.*m11*RS1456ppe + 1./96.*
         m11*RSe45611p + 1./96.*m11*RSp45611e + 1./32.*m11*RS145pe - 1./
         32.*m11*RS146pe + 1./32.*m11*RS156pe + 1./96.*m11*RS546pe - 1./
         48.*m11*RSe461p - 1./48.*m11*RSp461e + 1./96.*m11*RSe465p + 1./
         96.*m11*RSp465e );

      F13p6x23R +=  + pow(Ab(m35,mu2),2) * (  - 1./32. );

      F13p6x23R +=  + Ab(m35,mu2)*Ab(m36,mu2) * (  - 1./16. );

      F13p6x23R +=  + Ab(m35,mu2)*Ab(me2,mu2) * ( 1./36.*RSe45613p - 1./
         144.*RSe45633p - 1./72.*RSe463p - 1./48.*RSe45611p + 1./72.*
         RSe461p + 1./18.*me2*RN3e*RSe45613p - 1./18.*me2*RN3e*
         RSe45633p + 1./18.*me2*RN3e*RSe463p - 1./18.*me2*RN3e*RSe461p
          + 1./18.*m55*RN3e*RSe463p - 1./18.*m55*RN3e*RSe461p - 1./18.*
         m33*RN3e*RSe45613p + 1./18.*m33*RN3e*RSe45633p + 1./18.*m33*
         RN3e*RSe463p - 1./18.*m33*RN3e*RSe461p );

      F13p6x23R +=  + Ab(m35,mu2)*Ab(mp2,mu2) * ( 1./36.*RSp45613e - 1./
         144.*RSp45633e - 1./72.*RSp463e - 1./48.*RSp45611e + 1./72.*
         RSp461e + 1./18.*mp2*RN3p*RSp45613e - 1./18.*mp2*RN3p*
         RSp45633e + 1./18.*mp2*RN3p*RSp463e - 1./18.*mp2*RN3p*RSp461e
          + 1./18.*m55*RN3p*RSp463e - 1./18.*m55*RN3p*RSp461e - 1./18.*
         m33*RN3p*RSp45613e + 1./18.*m33*RN3p*RSp45633e + 1./18.*m33*
         RN3p*RSp463e - 1./18.*m33*RN3p*RSp461e );

      F13p6x23R +=  + Ab(m35,mu2)*Bb(m11,mu2) * (  - 1./48.*RD1456ep );

      F13p6x23R +=  + Ab(m35,mu2)*Bb(m33,mu2) * ( 7./144.*RD3456ep - 1./
         18.*m55*RS346pe - 1./9.*m33*RS346pe );

      F13p6x23R +=  + Ab(m36,mu2)*pi16 * ( 1./48.*RD1456ep + 1./48.*mp2
         *RSp45611e - 1./24.*mp2*RSp451e + 1./48.*mp2*RSp456e + 1./48.*
         me2*RSe45611p - 1./24.*me2*RSe451p + 1./48.*me2*RSe456p - 1./8.
         *m66 - 1./96.*m66*RS34561pe - 1./96.*m66*RS14563pe - 1./96.*
         m66*RSe45613p - 1./96.*m66*RSp45613e - 1./96.*m66*RS3456pee - 
         1./96.*m66*RS3456ppe + 1./96.*m66*RSe45633p + 1./96.*m66*
         RSp45633e + 1./96.*m66*RS346pe + 1./96.*m66*RS356pe - 1./96.*
         m66*RSe453p - 1./96.*m66*RSp453e - 1./96.*m66*RS1456pee - 1./
         96.*m66*RS1456ppe + 1./96.*m66*RSe45611p + 1./96.*m66*
         RSp45611e + 1./96.*m66*RS146pe + 1./96.*m66*RS156pe + 1./32.*
         m66*RS645pe - 1./96.*m66*RSe451p - 1./96.*m66*RSp451e + 1./96.
         *m66*RSe456p + 1./96.*m66*RSp456e - 1./16.*m55 - 1./16.*m44 + 
         1./32.*m33 - 1./96.*m33*RS34561pe - 1./96.*m33*RS14563pe - 1./
         96.*m33*RSe45613p - 1./96.*m33*RSp45613e - 1./96.*m33*
         RS3456pee - 1./96.*m33*RS3456ppe + 1./96.*m33*RSe45633p + 1./
         96.*m33*RSp45633e );

      F13p6x23R +=  + Ab(m36,mu2)*pi16 * ( 1./96.*m33*RS346pe + 1./96.*
         m33*RS356pe - 1./96.*m33*RSe453p - 1./96.*m33*RSp453e + 1./96.
         *m33*RS145pe + 1./96.*m33*RSe451p + 1./96.*m33*RSp451e - 3./32.
         *m11 - 1./32.*m11*RS1456pee - 1./32.*m11*RS1456ppe + 1./96.*
         m11*RSe45611p + 1./96.*m11*RSp45611e - 1./32.*m11*RS145pe + 1./
         32.*m11*RS146pe + 1./32.*m11*RS156pe + 1./96.*m11*RS645pe - 1./
         48.*m11*RSe451p - 1./48.*m11*RSp451e + 1./96.*m11*RSe456p + 1./
         96.*m11*RSp456e );

      F13p6x23R +=  + pow(Ab(m36,mu2),2) * (  - 1./32. );

      F13p6x23R +=  + Ab(m36,mu2)*Ab(me2,mu2) * ( 1./36.*RSe45613p - 1./
         144.*RSe45633p - 1./72.*RSe453p - 1./48.*RSe45611p + 1./72.*
         RSe451p + 1./18.*me2*RN3e*RSe45613p - 1./18.*me2*RN3e*
         RSe45633p + 1./18.*me2*RN3e*RSe453p - 1./18.*me2*RN3e*RSe451p
          + 1./18.*m66*RN3e*RSe453p - 1./18.*m66*RN3e*RSe451p - 1./18.*
         m33*RN3e*RSe45613p + 1./18.*m33*RN3e*RSe45633p + 1./18.*m33*
         RN3e*RSe453p - 1./18.*m33*RN3e*RSe451p );

      F13p6x23R +=  + Ab(m36,mu2)*Ab(mp2,mu2) * ( 1./36.*RSp45613e - 1./
         144.*RSp45633e - 1./72.*RSp453e - 1./48.*RSp45611e + 1./72.*
         RSp451e + 1./18.*mp2*RN3p*RSp45613e - 1./18.*mp2*RN3p*
         RSp45633e + 1./18.*mp2*RN3p*RSp453e - 1./18.*mp2*RN3p*RSp451e
          + 1./18.*m66*RN3p*RSp453e - 1./18.*m66*RN3p*RSp451e - 1./18.*
         m33*RN3p*RSp45613e + 1./18.*m33*RN3p*RSp45633e + 1./18.*m33*
         RN3p*RSp453e - 1./18.*m33*RN3p*RSp451e );

      F13p6x23R +=  + Ab(m36,mu2)*Bb(m11,mu2) * (  - 1./48.*RD1456ep );

      F13p6x23R +=  + Ab(m36,mu2)*Bb(m33,mu2) * ( 7./144.*RD3456ep - 1./
         18.*m66*RS345pe - 1./9.*m33*RS345pe );

      F13p6x23R +=  + Ab(m44,mu2)*Ab(me2,mu2) * ( 1./18.*m44*RNep*
         RSp563e*RSe563p - 1./54.*m44*RNep*RS456pe*RSp563e*RSe563p - 1./
         18.*m44*RNep*RSe561p*RSp563e + 1./54.*m44*RNep*RSe561p*RS456pe
         *RSp563e - 1./18.*m44*RNep*RSp561e*RSe563p + 1./54.*m44*RNep*
         RSp561e*RS456pe*RSe563p + 1./18.*m44*RNep*RSp561e*RSe561p - 1./
         54.*m44*RNep*RSp561e*RSe561p*RS456pe - 1./18.*m44*RN3e*RSe563p
         *RS356pe + 1./54.*m44*RN3e*RS456pe*RSe563p*RS356pe + 1./18.*
         m44*RN3e*RSe561p*RS356pe - 1./54.*m44*RN3e*RSe561p*RS456pe*
         RS356pe + 1./18.*m44*RN1e*RS156pe*RSe563p - 1./54.*m44*RN1e*
         RS456pe*RS156pe*RSe563p - 1./18.*m44*RN1e*RSe561p*RS156pe + 1./
         54.*m44*RN1e*RSe561p*RS456pe*RS156pe );

      F13p6x23R +=  + Ab(m44,mu2)*Ab(mp2,mu2) * (  - 1./18.*m44*RNep*
         RSp563e*RSe563p + 1./54.*m44*RNep*RS456pe*RSp563e*RSe563p + 1./
         18.*m44*RNep*RSe561p*RSp563e - 1./54.*m44*RNep*RSe561p*RS456pe
         *RSp563e + 1./18.*m44*RNep*RSp561e*RSe563p - 1./54.*m44*RNep*
         RSp561e*RS456pe*RSe563p - 1./18.*m44*RNep*RSp561e*RSe561p + 1./
         54.*m44*RNep*RSp561e*RSe561p*RS456pe - 1./18.*m44*RN3p*RSp563e
         *RS356pe + 1./54.*m44*RN3p*RS456pe*RSp563e*RS356pe + 1./18.*
         m44*RN3p*RSp561e*RS356pe - 1./54.*m44*RN3p*RSp561e*RS456pe*
         RS356pe + 1./18.*m44*RN1p*RS156pe*RSp563e - 1./54.*m44*RN1p*
         RS456pe*RS156pe*RSp563e - 1./18.*m44*RN1p*RSp561e*RS156pe + 1./
         54.*m44*RN1p*RSp561e*RS456pe*RS156pe );

      F13p6x23R +=  + Ab(m44,mu2)*Bb(m11,mu2) * ( 1./36.*m44*pow(
         RS156pe,2) - 1./108.*m44*RS456pe*pow(RS156pe,2) );

      F13p6x23R +=  + Ab(m44,mu2)*Bb(m33,mu2) * ( 1./36.*m44*pow(
         RS356pe,2) - 1./108.*m44*RS456pe*pow(RS356pe,2) );

      F13p6x23R +=  + Ab(m44,mu2)*Bb(me2,mu2) * ( 1./36.*m44*pow(
         RSe563p,2) - 1./108.*m44*RS456pe*pow(RSe563p,2) - 1./18.*m44*
         RSe561p*RSe563p + 1./54.*m44*RSe561p*RS456pe*RSe563p + 1./36.*
         m44*pow(RSe561p,2) - 1./108.*m44*pow(RSe561p,2)*RS456pe );

      F13p6x23R +=  + Ab(m44,mu2)*Bb(mp2,mu2) * ( 1./36.*m44*pow(
         RSp563e,2) - 1./108.*m44*RS456pe*pow(RSp563e,2) - 1./18.*m44*
         RSp561e*RSp563e + 1./54.*m44*RSp561e*RS456pe*RSp563e + 1./36.*
         m44*pow(RSp561e,2) - 1./108.*m44*pow(RSp561e,2)*RS456pe );

      F13p6x23R +=  + Ab(m45,mu2)*Ab(me2,mu2) * (  - 1./108.*pow(
         RSe563p,2) + 1./54.*RSe463p*RSe563p - 1./108.*pow(RSe463p,2)
          - 1./54.*RSp563e*RSe563p + 1./54.*RSp563e*RSe463p + 1./54.*
         RSp463e*RSe563p - 1./54.*RSp463e*RSe463p + 1./54.*RSe561p*
         RSe563p - 1./54.*RSe561p*RSe463p + 1./54.*RSe561p*RSp563e - 1./
         54.*RSe561p*RSp463e - 1./108.*pow(RSe561p,2) - 1./54.*RSe461p*
         RSe563p + 1./54.*RSe461p*RSe463p - 1./54.*RSe461p*RSp563e + 1./
         54.*RSe461p*RSp463e + 1./54.*RSe461p*RSe561p - 1./108.*pow(
         RSe461p,2) + 1./54.*RSp561e*RSe563p - 1./54.*RSp561e*RSe463p
          - 1./54.*RSp561e*RSe561p + 1./54.*RSp561e*RSe461p - 1./54.*
         RSp461e*RSe563p + 1./54.*RSp461e*RSe463p + 1./54.*RSp461e*
         RSe561p - 1./54.*RSp461e*RSe461p - 1./54.*mp2*RNep*RSp563e*
         RSe563p + 1./54.*mp2*RNep*RSp563e*RSe463p + 1./54.*mp2*RNep*
         RSp463e*RSe563p - 1./54.*mp2*RNep*RSp463e*RSe463p + 1./54.*mp2
         *RNep*RSe561p*RSp563e - 1./54.*mp2*RNep*RSe561p*RSp463e - 1./
         54.*mp2*RNep*RSe461p*RSp563e );

      F13p6x23R +=  + Ab(m45,mu2)*Ab(me2,mu2) * ( 1./54.*mp2*RNep*
         RSe461p*RSp463e + 1./54.*mp2*RNep*RSp561e*RSe563p - 1./54.*mp2
         *RNep*RSp561e*RSe463p - 1./54.*mp2*RNep*RSp561e*RSe561p + 1./
         54.*mp2*RNep*RSp561e*RSe461p - 1./54.*mp2*RNep*RSp461e*RSe563p
          + 1./54.*mp2*RNep*RSp461e*RSe463p + 1./54.*mp2*RNep*RSp461e*
         RSe561p - 1./54.*mp2*RNep*RSp461e*RSe461p - 1./54.*me2*RN3e*
         RSe563p*RS346pe + 1./54.*me2*RN3e*RSe563p*RS356pe + 1./54.*me2
         *RN3e*RSe463p*RS346pe - 1./54.*me2*RN3e*RSe463p*RS356pe + 1./
         54.*me2*RN3e*RSe561p*RS346pe - 1./54.*me2*RN3e*RSe561p*RS356pe
          - 1./54.*me2*RN3e*RSe461p*RS346pe + 1./54.*me2*RN3e*RSe461p*
         RS356pe + 1./54.*me2*RN1e*RS146pe*RSe563p - 1./54.*me2*RN1e*
         RS146pe*RSe463p - 1./54.*me2*RN1e*RS156pe*RSe563p + 1./54.*me2
         *RN1e*RS156pe*RSe463p - 1./54.*me2*RN1e*RSe561p*RS146pe + 1./
         54.*me2*RN1e*RSe561p*RS156pe + 1./54.*me2*RN1e*RSe461p*RS146pe
          - 1./54.*me2*RN1e*RSe461p*RS156pe + 1./54.*m55*RNep*RSp563e*
         RSe463p );

      F13p6x23R +=  + Ab(m45,mu2)*Ab(me2,mu2) * ( 1./54.*m55*RNep*
         RSp463e*RSe563p + 1./54.*m55*RNep*RSp463e*RSe463p - 1./54.*m55
         *RNep*RSe561p*RSp463e - 1./54.*m55*RNep*RSe461p*RSp563e - 1./
         54.*m55*RNep*RSe461p*RSp463e - 1./54.*m55*RNep*RSp561e*RSe463p
          + 1./54.*m55*RNep*RSp561e*RSe461p - 1./54.*m55*RNep*RSp461e*
         RSe563p - 1./54.*m55*RNep*RSp461e*RSe463p + 1./54.*m55*RNep*
         RSp461e*RSe561p + 1./54.*m55*RNep*RSp461e*RSe461p - 1./54.*m55
         *RN3e*RSe563p*RS346pe - 1./54.*m55*RN3e*RSe463p*RS346pe - 1./
         54.*m55*RN3e*RSe463p*RS356pe + 1./54.*m55*RN3e*RSe561p*RS346pe
          + 1./54.*m55*RN3e*RSe461p*RS346pe + 1./54.*m55*RN3e*RSe461p*
         RS356pe + 1./54.*m55*RN1e*RS146pe*RSe563p + 1./54.*m55*RN1e*
         RS146pe*RSe463p + 1./54.*m55*RN1e*RS156pe*RSe463p - 1./54.*m55
         *RN1e*RSe561p*RS146pe - 1./54.*m55*RN1e*RSe461p*RS146pe - 1./
         54.*m55*RN1e*RSe461p*RS156pe + 1./54.*m44*RNep*RSp563e*RSe563p
          + 1./54.*m44*RNep*RSp563e*RSe463p + 1./54.*m44*RNep*RSp463e*
         RSe563p );

      F13p6x23R +=  + Ab(m45,mu2)*Ab(me2,mu2) * (  - 1./54.*m44*RNep*
         RSe561p*RSp563e - 1./54.*m44*RNep*RSe561p*RSp463e - 1./54.*m44
         *RNep*RSe461p*RSp563e - 1./54.*m44*RNep*RSp561e*RSe563p - 1./
         54.*m44*RNep*RSp561e*RSe463p + 1./54.*m44*RNep*RSp561e*RSe561p
          + 1./54.*m44*RNep*RSp561e*RSe461p - 1./54.*m44*RNep*RSp461e*
         RSe563p + 1./54.*m44*RNep*RSp461e*RSe561p - 1./54.*m44*RN3e*
         RSe563p*RS346pe - 1./54.*m44*RN3e*RSe563p*RS356pe - 1./54.*m44
         *RN3e*RSe463p*RS356pe + 1./54.*m44*RN3e*RSe561p*RS346pe + 1./
         54.*m44*RN3e*RSe561p*RS356pe + 1./54.*m44*RN3e*RSe461p*RS356pe
          + 1./54.*m44*RN1e*RS146pe*RSe563p + 1./54.*m44*RN1e*RS156pe*
         RSe563p + 1./54.*m44*RN1e*RS156pe*RSe463p - 1./54.*m44*RN1e*
         RSe561p*RS146pe - 1./54.*m44*RN1e*RSe561p*RS156pe - 1./54.*m44
         *RN1e*RSe461p*RS156pe );

      F13p6x23R +=  + Ab(m45,mu2)*Ab(mp2,mu2) * (  - 1./108.*pow(
         RSp563e,2) + 1./54.*RSp463e*RSp563e - 1./108.*pow(RSp463e,2)
          + 1./54.*RSp561e*RSp563e - 1./54.*RSp561e*RSp463e - 1./108.*
         pow(RSp561e,2) - 1./54.*RSp461e*RSp563e + 1./54.*RSp461e*
         RSp463e + 1./54.*RSp461e*RSp561e - 1./108.*pow(RSp461e,2) + 1./
         54.*mp2*RNep*RSp563e*RSe563p - 1./54.*mp2*RNep*RSp563e*RSe463p
          - 1./54.*mp2*RNep*RSp463e*RSe563p + 1./54.*mp2*RNep*RSp463e*
         RSe463p - 1./54.*mp2*RNep*RSe561p*RSp563e + 1./54.*mp2*RNep*
         RSe561p*RSp463e + 1./54.*mp2*RNep*RSe461p*RSp563e - 1./54.*mp2
         *RNep*RSe461p*RSp463e - 1./54.*mp2*RNep*RSp561e*RSe563p + 1./
         54.*mp2*RNep*RSp561e*RSe463p + 1./54.*mp2*RNep*RSp561e*RSe561p
          - 1./54.*mp2*RNep*RSp561e*RSe461p + 1./54.*mp2*RNep*RSp461e*
         RSe563p - 1./54.*mp2*RNep*RSp461e*RSe463p - 1./54.*mp2*RNep*
         RSp461e*RSe561p + 1./54.*mp2*RNep*RSp461e*RSe461p - 1./54.*mp2
         *RN3p*RSp563e*RS346pe + 1./54.*mp2*RN3p*RSp563e*RS356pe + 1./
         54.*mp2*RN3p*RSp463e*RS346pe );

      F13p6x23R +=  + Ab(m45,mu2)*Ab(mp2,mu2) * (  - 1./54.*mp2*RN3p*
         RSp463e*RS356pe + 1./54.*mp2*RN3p*RSp561e*RS346pe - 1./54.*mp2
         *RN3p*RSp561e*RS356pe - 1./54.*mp2*RN3p*RSp461e*RS346pe + 1./
         54.*mp2*RN3p*RSp461e*RS356pe + 1./54.*mp2*RN1p*RS146pe*RSp563e
          - 1./54.*mp2*RN1p*RS146pe*RSp463e - 1./54.*mp2*RN1p*RS156pe*
         RSp563e + 1./54.*mp2*RN1p*RS156pe*RSp463e - 1./54.*mp2*RN1p*
         RSp561e*RS146pe + 1./54.*mp2*RN1p*RSp561e*RS156pe + 1./54.*mp2
         *RN1p*RSp461e*RS146pe - 1./54.*mp2*RN1p*RSp461e*RS156pe - 1./
         54.*m55*RNep*RSp563e*RSe463p - 1./54.*m55*RNep*RSp463e*RSe563p
          - 1./54.*m55*RNep*RSp463e*RSe463p + 1./54.*m55*RNep*RSe561p*
         RSp463e + 1./54.*m55*RNep*RSe461p*RSp563e + 1./54.*m55*RNep*
         RSe461p*RSp463e + 1./54.*m55*RNep*RSp561e*RSe463p - 1./54.*m55
         *RNep*RSp561e*RSe461p + 1./54.*m55*RNep*RSp461e*RSe563p + 1./
         54.*m55*RNep*RSp461e*RSe463p - 1./54.*m55*RNep*RSp461e*RSe561p
          - 1./54.*m55*RNep*RSp461e*RSe461p - 1./54.*m55*RN3p*RSp563e*
         RS346pe );

      F13p6x23R +=  + Ab(m45,mu2)*Ab(mp2,mu2) * (  - 1./54.*m55*RN3p*
         RSp463e*RS346pe - 1./54.*m55*RN3p*RSp463e*RS356pe + 1./54.*m55
         *RN3p*RSp561e*RS346pe + 1./54.*m55*RN3p*RSp461e*RS346pe + 1./
         54.*m55*RN3p*RSp461e*RS356pe + 1./54.*m55*RN1p*RS146pe*RSp563e
          + 1./54.*m55*RN1p*RS146pe*RSp463e + 1./54.*m55*RN1p*RS156pe*
         RSp463e - 1./54.*m55*RN1p*RSp561e*RS146pe - 1./54.*m55*RN1p*
         RSp461e*RS146pe - 1./54.*m55*RN1p*RSp461e*RS156pe - 1./54.*m44
         *RNep*RSp563e*RSe563p - 1./54.*m44*RNep*RSp563e*RSe463p - 1./
         54.*m44*RNep*RSp463e*RSe563p + 1./54.*m44*RNep*RSe561p*RSp563e
          + 1./54.*m44*RNep*RSe561p*RSp463e + 1./54.*m44*RNep*RSe461p*
         RSp563e + 1./54.*m44*RNep*RSp561e*RSe563p + 1./54.*m44*RNep*
         RSp561e*RSe463p - 1./54.*m44*RNep*RSp561e*RSe561p - 1./54.*m44
         *RNep*RSp561e*RSe461p + 1./54.*m44*RNep*RSp461e*RSe563p - 1./
         54.*m44*RNep*RSp461e*RSe561p - 1./54.*m44*RN3p*RSp563e*RS346pe
          - 1./54.*m44*RN3p*RSp563e*RS356pe - 1./54.*m44*RN3p*RSp463e*
         RS356pe );

      F13p6x23R +=  + Ab(m45,mu2)*Ab(mp2,mu2) * ( 1./54.*m44*RN3p*
         RSp561e*RS346pe + 1./54.*m44*RN3p*RSp561e*RS356pe + 1./54.*m44
         *RN3p*RSp461e*RS356pe + 1./54.*m44*RN1p*RS146pe*RSp563e + 1./
         54.*m44*RN1p*RS156pe*RSp563e + 1./54.*m44*RN1p*RS156pe*RSp463e
          - 1./54.*m44*RN1p*RSp561e*RS146pe - 1./54.*m44*RN1p*RSp561e*
         RS156pe - 1./54.*m44*RN1p*RSp461e*RS156pe );

      F13p6x23R +=  + Ab(m45,mu2)*Bb(m11,mu2) * ( 1./108.*m55*pow(
         RS146pe,2) + 1./54.*m55*RS156pe*RS146pe + 1./54.*m44*RS156pe*
         RS146pe + 1./108.*m44*pow(RS156pe,2) - 1./108.*m11*pow(
         RS146pe,2) + 1./54.*m11*RS156pe*RS146pe - 1./108.*m11*pow(
         RS156pe,2) );

      F13p6x23R +=  + Ab(m45,mu2)*Bb(m33,mu2) * ( 1./108.*m55*pow(
         RS346pe,2) + 1./54.*m55*RS356pe*RS346pe + 1./54.*m44*RS356pe*
         RS346pe + 1./108.*m44*pow(RS356pe,2) - 1./108.*m33*pow(
         RS346pe,2) + 1./54.*m33*RS356pe*RS346pe - 1./108.*m33*pow(
         RS356pe,2) );

      F13p6x23R +=  + Ab(m45,mu2)*Bb(me2,mu2) * (  - 1./108.*me2*pow(
         RSe563p,2) + 1./54.*me2*RSe463p*RSe563p - 1./108.*me2*pow(
         RSe463p,2) + 1./54.*me2*RSe561p*RSe563p - 1./54.*me2*RSe561p*
         RSe463p - 1./108.*me2*pow(RSe561p,2) - 1./54.*me2*RSe461p*
         RSe563p + 1./54.*me2*RSe461p*RSe463p + 1./54.*me2*RSe461p*
         RSe561p - 1./108.*me2*pow(RSe461p,2) + 1./54.*m55*RSe463p*
         RSe563p + 1./108.*m55*pow(RSe463p,2) - 1./54.*m55*RSe561p*
         RSe463p - 1./54.*m55*RSe461p*RSe563p - 1./54.*m55*RSe461p*
         RSe463p + 1./54.*m55*RSe461p*RSe561p + 1./108.*m55*pow(
         RSe461p,2) + 1./108.*m44*pow(RSe563p,2) + 1./54.*m44*RSe463p*
         RSe563p - 1./54.*m44*RSe561p*RSe563p - 1./54.*m44*RSe561p*
         RSe463p + 1./108.*m44*pow(RSe561p,2) - 1./54.*m44*RSe461p*
         RSe563p + 1./54.*m44*RSe461p*RSe561p );

      F13p6x23R +=  + Ab(m45,mu2)*Bb(mp2,mu2) * (  - 1./108.*mp2*pow(
         RSp563e,2) + 1./54.*mp2*RSp463e*RSp563e - 1./108.*mp2*pow(
         RSp463e,2) + 1./54.*mp2*RSp561e*RSp563e - 1./54.*mp2*RSp561e*
         RSp463e - 1./108.*mp2*pow(RSp561e,2) - 1./54.*mp2*RSp461e*
         RSp563e + 1./54.*mp2*RSp461e*RSp463e + 1./54.*mp2*RSp461e*
         RSp561e - 1./108.*mp2*pow(RSp461e,2) + 1./54.*m55*RSp463e*
         RSp563e + 1./108.*m55*pow(RSp463e,2) - 1./54.*m55*RSp561e*
         RSp463e - 1./54.*m55*RSp461e*RSp563e - 1./54.*m55*RSp461e*
         RSp463e + 1./54.*m55*RSp461e*RSp561e + 1./108.*m55*pow(
         RSp461e,2) + 1./108.*m44*pow(RSp563e,2) + 1./54.*m44*RSp463e*
         RSp563e - 1./54.*m44*RSp561e*RSp563e - 1./54.*m44*RSp561e*
         RSp463e + 1./108.*m44*pow(RSp561e,2) - 1./54.*m44*RSp461e*
         RSp563e + 1./54.*m44*RSp461e*RSp561e );

      F13p6x23R +=  + Ab(m46,mu2)*Ab(me2,mu2) * (  - 1./108.*pow(
         RSe563p,2) + 1./54.*RSe453p*RSe563p - 1./108.*pow(RSe453p,2)
          - 1./54.*RSp563e*RSe563p + 1./54.*RSp563e*RSe453p + 1./54.*
         RSp453e*RSe563p - 1./54.*RSp453e*RSe453p + 1./54.*RSe561p*
         RSe563p - 1./54.*RSe561p*RSe453p + 1./54.*RSe561p*RSp563e - 1./
         54.*RSe561p*RSp453e - 1./108.*pow(RSe561p,2) - 1./54.*RSe451p*
         RSe563p + 1./54.*RSe451p*RSe453p - 1./54.*RSe451p*RSp563e + 1./
         54.*RSe451p*RSp453e + 1./54.*RSe451p*RSe561p - 1./108.*pow(
         RSe451p,2) + 1./54.*RSp561e*RSe563p - 1./54.*RSp561e*RSe453p
          - 1./54.*RSp561e*RSe561p + 1./54.*RSp561e*RSe451p - 1./54.*
         RSp451e*RSe563p + 1./54.*RSp451e*RSe453p + 1./54.*RSp451e*
         RSe561p - 1./54.*RSp451e*RSe451p - 1./54.*mp2*RNep*RSp563e*
         RSe563p + 1./54.*mp2*RNep*RSp563e*RSe453p + 1./54.*mp2*RNep*
         RSp453e*RSe563p - 1./54.*mp2*RNep*RSp453e*RSe453p + 1./54.*mp2
         *RNep*RSe561p*RSp563e - 1./54.*mp2*RNep*RSe561p*RSp453e - 1./
         54.*mp2*RNep*RSe451p*RSp563e );

      F13p6x23R +=  + Ab(m46,mu2)*Ab(me2,mu2) * ( 1./54.*mp2*RNep*
         RSe451p*RSp453e + 1./54.*mp2*RNep*RSp561e*RSe563p - 1./54.*mp2
         *RNep*RSp561e*RSe453p - 1./54.*mp2*RNep*RSp561e*RSe561p + 1./
         54.*mp2*RNep*RSp561e*RSe451p - 1./54.*mp2*RNep*RSp451e*RSe563p
          + 1./54.*mp2*RNep*RSp451e*RSe453p + 1./54.*mp2*RNep*RSp451e*
         RSe561p - 1./54.*mp2*RNep*RSp451e*RSe451p - 1./54.*me2*RN3e*
         RSe563p*RS345pe + 1./54.*me2*RN3e*RSe563p*RS356pe + 1./54.*me2
         *RN3e*RSe453p*RS345pe - 1./54.*me2*RN3e*RSe453p*RS356pe + 1./
         54.*me2*RN3e*RSe561p*RS345pe - 1./54.*me2*RN3e*RSe561p*RS356pe
          - 1./54.*me2*RN3e*RSe451p*RS345pe + 1./54.*me2*RN3e*RSe451p*
         RS356pe + 1./54.*me2*RN1e*RS145pe*RSe563p - 1./54.*me2*RN1e*
         RS145pe*RSe453p - 1./54.*me2*RN1e*RS156pe*RSe563p + 1./54.*me2
         *RN1e*RS156pe*RSe453p - 1./54.*me2*RN1e*RSe561p*RS145pe + 1./
         54.*me2*RN1e*RSe561p*RS156pe + 1./54.*me2*RN1e*RSe451p*RS145pe
          - 1./54.*me2*RN1e*RSe451p*RS156pe + 1./54.*m66*RNep*RSp563e*
         RSe453p );

      F13p6x23R +=  + Ab(m46,mu2)*Ab(me2,mu2) * ( 1./54.*m66*RNep*
         RSp453e*RSe563p + 1./54.*m66*RNep*RSp453e*RSe453p - 1./54.*m66
         *RNep*RSe561p*RSp453e - 1./54.*m66*RNep*RSe451p*RSp563e - 1./
         54.*m66*RNep*RSe451p*RSp453e - 1./54.*m66*RNep*RSp561e*RSe453p
          + 1./54.*m66*RNep*RSp561e*RSe451p - 1./54.*m66*RNep*RSp451e*
         RSe563p - 1./54.*m66*RNep*RSp451e*RSe453p + 1./54.*m66*RNep*
         RSp451e*RSe561p + 1./54.*m66*RNep*RSp451e*RSe451p - 1./54.*m66
         *RN3e*RSe563p*RS345pe - 1./54.*m66*RN3e*RSe453p*RS345pe - 1./
         54.*m66*RN3e*RSe453p*RS356pe + 1./54.*m66*RN3e*RSe561p*RS345pe
          + 1./54.*m66*RN3e*RSe451p*RS345pe + 1./54.*m66*RN3e*RSe451p*
         RS356pe + 1./54.*m66*RN1e*RS145pe*RSe563p + 1./54.*m66*RN1e*
         RS145pe*RSe453p + 1./54.*m66*RN1e*RS156pe*RSe453p - 1./54.*m66
         *RN1e*RSe561p*RS145pe - 1./54.*m66*RN1e*RSe451p*RS145pe - 1./
         54.*m66*RN1e*RSe451p*RS156pe + 1./54.*m44*RNep*RSp563e*RSe563p
          + 1./54.*m44*RNep*RSp563e*RSe453p + 1./54.*m44*RNep*RSp453e*
         RSe563p );

      F13p6x23R +=  + Ab(m46,mu2)*Ab(me2,mu2) * (  - 1./54.*m44*RNep*
         RSe561p*RSp563e - 1./54.*m44*RNep*RSe561p*RSp453e - 1./54.*m44
         *RNep*RSe451p*RSp563e - 1./54.*m44*RNep*RSp561e*RSe563p - 1./
         54.*m44*RNep*RSp561e*RSe453p + 1./54.*m44*RNep*RSp561e*RSe561p
          + 1./54.*m44*RNep*RSp561e*RSe451p - 1./54.*m44*RNep*RSp451e*
         RSe563p + 1./54.*m44*RNep*RSp451e*RSe561p - 1./54.*m44*RN3e*
         RSe563p*RS345pe - 1./54.*m44*RN3e*RSe563p*RS356pe - 1./54.*m44
         *RN3e*RSe453p*RS356pe + 1./54.*m44*RN3e*RSe561p*RS345pe + 1./
         54.*m44*RN3e*RSe561p*RS356pe + 1./54.*m44*RN3e*RSe451p*RS356pe
          + 1./54.*m44*RN1e*RS145pe*RSe563p + 1./54.*m44*RN1e*RS156pe*
         RSe563p + 1./54.*m44*RN1e*RS156pe*RSe453p - 1./54.*m44*RN1e*
         RSe561p*RS145pe - 1./54.*m44*RN1e*RSe561p*RS156pe - 1./54.*m44
         *RN1e*RSe451p*RS156pe );

      F13p6x23R +=  + Ab(m46,mu2)*Ab(mp2,mu2) * (  - 1./108.*pow(
         RSp563e,2) + 1./54.*RSp453e*RSp563e - 1./108.*pow(RSp453e,2)
          + 1./54.*RSp561e*RSp563e - 1./54.*RSp561e*RSp453e - 1./108.*
         pow(RSp561e,2) - 1./54.*RSp451e*RSp563e + 1./54.*RSp451e*
         RSp453e + 1./54.*RSp451e*RSp561e - 1./108.*pow(RSp451e,2) + 1./
         54.*mp2*RNep*RSp563e*RSe563p - 1./54.*mp2*RNep*RSp563e*RSe453p
          - 1./54.*mp2*RNep*RSp453e*RSe563p + 1./54.*mp2*RNep*RSp453e*
         RSe453p - 1./54.*mp2*RNep*RSe561p*RSp563e + 1./54.*mp2*RNep*
         RSe561p*RSp453e + 1./54.*mp2*RNep*RSe451p*RSp563e - 1./54.*mp2
         *RNep*RSe451p*RSp453e - 1./54.*mp2*RNep*RSp561e*RSe563p + 1./
         54.*mp2*RNep*RSp561e*RSe453p + 1./54.*mp2*RNep*RSp561e*RSe561p
          - 1./54.*mp2*RNep*RSp561e*RSe451p + 1./54.*mp2*RNep*RSp451e*
         RSe563p - 1./54.*mp2*RNep*RSp451e*RSe453p - 1./54.*mp2*RNep*
         RSp451e*RSe561p + 1./54.*mp2*RNep*RSp451e*RSe451p - 1./54.*mp2
         *RN3p*RSp563e*RS345pe + 1./54.*mp2*RN3p*RSp563e*RS356pe + 1./
         54.*mp2*RN3p*RSp453e*RS345pe );

      F13p6x23R +=  + Ab(m46,mu2)*Ab(mp2,mu2) * (  - 1./54.*mp2*RN3p*
         RSp453e*RS356pe + 1./54.*mp2*RN3p*RSp561e*RS345pe - 1./54.*mp2
         *RN3p*RSp561e*RS356pe - 1./54.*mp2*RN3p*RSp451e*RS345pe + 1./
         54.*mp2*RN3p*RSp451e*RS356pe + 1./54.*mp2*RN1p*RS145pe*RSp563e
          - 1./54.*mp2*RN1p*RS145pe*RSp453e - 1./54.*mp2*RN1p*RS156pe*
         RSp563e + 1./54.*mp2*RN1p*RS156pe*RSp453e - 1./54.*mp2*RN1p*
         RSp561e*RS145pe + 1./54.*mp2*RN1p*RSp561e*RS156pe + 1./54.*mp2
         *RN1p*RSp451e*RS145pe - 1./54.*mp2*RN1p*RSp451e*RS156pe - 1./
         54.*m66*RNep*RSp563e*RSe453p - 1./54.*m66*RNep*RSp453e*RSe563p
          - 1./54.*m66*RNep*RSp453e*RSe453p + 1./54.*m66*RNep*RSe561p*
         RSp453e + 1./54.*m66*RNep*RSe451p*RSp563e + 1./54.*m66*RNep*
         RSe451p*RSp453e + 1./54.*m66*RNep*RSp561e*RSe453p - 1./54.*m66
         *RNep*RSp561e*RSe451p + 1./54.*m66*RNep*RSp451e*RSe563p + 1./
         54.*m66*RNep*RSp451e*RSe453p - 1./54.*m66*RNep*RSp451e*RSe561p
          - 1./54.*m66*RNep*RSp451e*RSe451p - 1./54.*m66*RN3p*RSp563e*
         RS345pe );

      F13p6x23R +=  + Ab(m46,mu2)*Ab(mp2,mu2) * (  - 1./54.*m66*RN3p*
         RSp453e*RS345pe - 1./54.*m66*RN3p*RSp453e*RS356pe + 1./54.*m66
         *RN3p*RSp561e*RS345pe + 1./54.*m66*RN3p*RSp451e*RS345pe + 1./
         54.*m66*RN3p*RSp451e*RS356pe + 1./54.*m66*RN1p*RS145pe*RSp563e
          + 1./54.*m66*RN1p*RS145pe*RSp453e + 1./54.*m66*RN1p*RS156pe*
         RSp453e - 1./54.*m66*RN1p*RSp561e*RS145pe - 1./54.*m66*RN1p*
         RSp451e*RS145pe - 1./54.*m66*RN1p*RSp451e*RS156pe - 1./54.*m44
         *RNep*RSp563e*RSe563p - 1./54.*m44*RNep*RSp563e*RSe453p - 1./
         54.*m44*RNep*RSp453e*RSe563p + 1./54.*m44*RNep*RSe561p*RSp563e
          + 1./54.*m44*RNep*RSe561p*RSp453e + 1./54.*m44*RNep*RSe451p*
         RSp563e + 1./54.*m44*RNep*RSp561e*RSe563p + 1./54.*m44*RNep*
         RSp561e*RSe453p - 1./54.*m44*RNep*RSp561e*RSe561p - 1./54.*m44
         *RNep*RSp561e*RSe451p + 1./54.*m44*RNep*RSp451e*RSe563p - 1./
         54.*m44*RNep*RSp451e*RSe561p - 1./54.*m44*RN3p*RSp563e*RS345pe
          - 1./54.*m44*RN3p*RSp563e*RS356pe - 1./54.*m44*RN3p*RSp453e*
         RS356pe );

      F13p6x23R +=  + Ab(m46,mu2)*Ab(mp2,mu2) * ( 1./54.*m44*RN3p*
         RSp561e*RS345pe + 1./54.*m44*RN3p*RSp561e*RS356pe + 1./54.*m44
         *RN3p*RSp451e*RS356pe + 1./54.*m44*RN1p*RS145pe*RSp563e + 1./
         54.*m44*RN1p*RS156pe*RSp563e + 1./54.*m44*RN1p*RS156pe*RSp453e
          - 1./54.*m44*RN1p*RSp561e*RS145pe - 1./54.*m44*RN1p*RSp561e*
         RS156pe - 1./54.*m44*RN1p*RSp451e*RS156pe );

      F13p6x23R +=  + Ab(m46,mu2)*Bb(m11,mu2) * ( 1./108.*m66*pow(
         RS145pe,2) + 1./54.*m66*RS156pe*RS145pe + 1./54.*m44*RS156pe*
         RS145pe + 1./108.*m44*pow(RS156pe,2) - 1./108.*m11*pow(
         RS145pe,2) + 1./54.*m11*RS156pe*RS145pe - 1./108.*m11*pow(
         RS156pe,2) );

      F13p6x23R +=  + Ab(m46,mu2)*Bb(m33,mu2) * ( 1./108.*m66*pow(
         RS345pe,2) + 1./54.*m66*RS356pe*RS345pe + 1./54.*m44*RS356pe*
         RS345pe + 1./108.*m44*pow(RS356pe,2) - 1./108.*m33*pow(
         RS345pe,2) + 1./54.*m33*RS356pe*RS345pe - 1./108.*m33*pow(
         RS356pe,2) );

      F13p6x23R +=  + Ab(m46,mu2)*Bb(me2,mu2) * (  - 1./108.*me2*pow(
         RSe563p,2) + 1./54.*me2*RSe453p*RSe563p - 1./108.*me2*pow(
         RSe453p,2) + 1./54.*me2*RSe561p*RSe563p - 1./54.*me2*RSe561p*
         RSe453p - 1./108.*me2*pow(RSe561p,2) - 1./54.*me2*RSe451p*
         RSe563p + 1./54.*me2*RSe451p*RSe453p + 1./54.*me2*RSe451p*
         RSe561p - 1./108.*me2*pow(RSe451p,2) + 1./54.*m66*RSe453p*
         RSe563p + 1./108.*m66*pow(RSe453p,2) - 1./54.*m66*RSe561p*
         RSe453p - 1./54.*m66*RSe451p*RSe563p - 1./54.*m66*RSe451p*
         RSe453p + 1./54.*m66*RSe451p*RSe561p + 1./108.*m66*pow(
         RSe451p,2) + 1./108.*m44*pow(RSe563p,2) + 1./54.*m44*RSe453p*
         RSe563p - 1./54.*m44*RSe561p*RSe563p - 1./54.*m44*RSe561p*
         RSe453p + 1./108.*m44*pow(RSe561p,2) - 1./54.*m44*RSe451p*
         RSe563p + 1./54.*m44*RSe451p*RSe561p );

      F13p6x23R +=  + Ab(m46,mu2)*Bb(mp2,mu2) * (  - 1./108.*mp2*pow(
         RSp563e,2) + 1./54.*mp2*RSp453e*RSp563e - 1./108.*mp2*pow(
         RSp453e,2) + 1./54.*mp2*RSp561e*RSp563e - 1./54.*mp2*RSp561e*
         RSp453e - 1./108.*mp2*pow(RSp561e,2) - 1./54.*mp2*RSp451e*
         RSp563e + 1./54.*mp2*RSp451e*RSp453e + 1./54.*mp2*RSp451e*
         RSp561e - 1./108.*mp2*pow(RSp451e,2) + 1./54.*m66*RSp453e*
         RSp563e + 1./108.*m66*pow(RSp453e,2) - 1./54.*m66*RSp561e*
         RSp453e - 1./54.*m66*RSp451e*RSp563e - 1./54.*m66*RSp451e*
         RSp453e + 1./54.*m66*RSp451e*RSp561e + 1./108.*m66*pow(
         RSp451e,2) + 1./108.*m44*pow(RSp563e,2) + 1./54.*m44*RSp453e*
         RSp563e - 1./54.*m44*RSp561e*RSp563e - 1./54.*m44*RSp561e*
         RSp453e + 1./108.*m44*pow(RSp561e,2) - 1./54.*m44*RSp451e*
         RSp563e + 1./54.*m44*RSp451e*RSp561e );

      F13p6x23R +=  + Ab(m55,mu2)*Ab(me2,mu2) * ( 1./18.*m55*RNep*
         RSp463e*RSe463p - 1./54.*m55*RNep*RS546pe*RSp463e*RSe463p - 1./
         18.*m55*RNep*RSe461p*RSp463e + 1./54.*m55*RNep*RSe461p*RS546pe
         *RSp463e - 1./18.*m55*RNep*RSp461e*RSe463p + 1./54.*m55*RNep*
         RSp461e*RS546pe*RSe463p + 1./18.*m55*RNep*RSp461e*RSe461p - 1./
         54.*m55*RNep*RSp461e*RSe461p*RS546pe - 1./18.*m55*RN3e*RSe463p
         *RS346pe + 1./54.*m55*RN3e*RS546pe*RSe463p*RS346pe + 1./18.*
         m55*RN3e*RSe461p*RS346pe - 1./54.*m55*RN3e*RSe461p*RS546pe*
         RS346pe + 1./18.*m55*RN1e*RS146pe*RSe463p - 1./54.*m55*RN1e*
         RS546pe*RS146pe*RSe463p - 1./18.*m55*RN1e*RSe461p*RS146pe + 1./
         54.*m55*RN1e*RSe461p*RS546pe*RS146pe );

      F13p6x23R +=  + Ab(m55,mu2)*Ab(mp2,mu2) * (  - 1./18.*m55*RNep*
         RSp463e*RSe463p + 1./54.*m55*RNep*RS546pe*RSp463e*RSe463p + 1./
         18.*m55*RNep*RSe461p*RSp463e - 1./54.*m55*RNep*RSe461p*RS546pe
         *RSp463e + 1./18.*m55*RNep*RSp461e*RSe463p - 1./54.*m55*RNep*
         RSp461e*RS546pe*RSe463p - 1./18.*m55*RNep*RSp461e*RSe461p + 1./
         54.*m55*RNep*RSp461e*RSe461p*RS546pe - 1./18.*m55*RN3p*RSp463e
         *RS346pe + 1./54.*m55*RN3p*RS546pe*RSp463e*RS346pe + 1./18.*
         m55*RN3p*RSp461e*RS346pe - 1./54.*m55*RN3p*RSp461e*RS546pe*
         RS346pe + 1./18.*m55*RN1p*RS146pe*RSp463e - 1./54.*m55*RN1p*
         RS546pe*RS146pe*RSp463e - 1./18.*m55*RN1p*RSp461e*RS146pe + 1./
         54.*m55*RN1p*RSp461e*RS546pe*RS146pe );

      F13p6x23R +=  + Ab(m55,mu2)*Bb(m11,mu2) * ( 1./36.*m55*pow(
         RS146pe,2) - 1./108.*m55*RS546pe*pow(RS146pe,2) );

      F13p6x23R +=  + Ab(m55,mu2)*Bb(m33,mu2) * ( 1./36.*m55*pow(
         RS346pe,2) - 1./108.*m55*RS546pe*pow(RS346pe,2) );

      F13p6x23R +=  + Ab(m55,mu2)*Bb(me2,mu2) * ( 1./36.*m55*pow(
         RSe463p,2) - 1./108.*m55*RS546pe*pow(RSe463p,2) - 1./18.*m55*
         RSe461p*RSe463p + 1./54.*m55*RSe461p*RS546pe*RSe463p + 1./36.*
         m55*pow(RSe461p,2) - 1./108.*m55*pow(RSe461p,2)*RS546pe );

      F13p6x23R +=  + Ab(m55,mu2)*Bb(mp2,mu2) * ( 1./36.*m55*pow(
         RSp463e,2) - 1./108.*m55*RS546pe*pow(RSp463e,2) - 1./18.*m55*
         RSp461e*RSp463e + 1./54.*m55*RSp461e*RS546pe*RSp463e + 1./36.*
         m55*pow(RSp461e,2) - 1./108.*m55*pow(RSp461e,2)*RS546pe );

      F13p6x23R +=  + Ab(m56,mu2)*Ab(me2,mu2) * (  - 1./108.*pow(
         RSe463p,2) + 1./54.*RSe453p*RSe463p - 1./108.*pow(RSe453p,2)
          - 1./54.*RSp463e*RSe463p + 1./54.*RSp463e*RSe453p + 1./54.*
         RSp453e*RSe463p - 1./54.*RSp453e*RSe453p + 1./54.*RSe461p*
         RSe463p - 1./54.*RSe461p*RSe453p + 1./54.*RSe461p*RSp463e - 1./
         54.*RSe461p*RSp453e - 1./108.*pow(RSe461p,2) - 1./54.*RSe451p*
         RSe463p + 1./54.*RSe451p*RSe453p - 1./54.*RSe451p*RSp463e + 1./
         54.*RSe451p*RSp453e + 1./54.*RSe451p*RSe461p - 1./108.*pow(
         RSe451p,2) + 1./54.*RSp461e*RSe463p - 1./54.*RSp461e*RSe453p
          - 1./54.*RSp461e*RSe461p + 1./54.*RSp461e*RSe451p - 1./54.*
         RSp451e*RSe463p + 1./54.*RSp451e*RSe453p + 1./54.*RSp451e*
         RSe461p - 1./54.*RSp451e*RSe451p - 1./54.*mp2*RNep*RSp463e*
         RSe463p + 1./54.*mp2*RNep*RSp463e*RSe453p + 1./54.*mp2*RNep*
         RSp453e*RSe463p - 1./54.*mp2*RNep*RSp453e*RSe453p + 1./54.*mp2
         *RNep*RSe461p*RSp463e - 1./54.*mp2*RNep*RSe461p*RSp453e - 1./
         54.*mp2*RNep*RSe451p*RSp463e );

      F13p6x23R +=  + Ab(m56,mu2)*Ab(me2,mu2) * ( 1./54.*mp2*RNep*
         RSe451p*RSp453e + 1./54.*mp2*RNep*RSp461e*RSe463p - 1./54.*mp2
         *RNep*RSp461e*RSe453p - 1./54.*mp2*RNep*RSp461e*RSe461p + 1./
         54.*mp2*RNep*RSp461e*RSe451p - 1./54.*mp2*RNep*RSp451e*RSe463p
          + 1./54.*mp2*RNep*RSp451e*RSe453p + 1./54.*mp2*RNep*RSp451e*
         RSe461p - 1./54.*mp2*RNep*RSp451e*RSe451p - 1./54.*me2*RN3e*
         RSe463p*RS345pe + 1./54.*me2*RN3e*RSe463p*RS346pe + 1./54.*me2
         *RN3e*RSe453p*RS345pe - 1./54.*me2*RN3e*RSe453p*RS346pe + 1./
         54.*me2*RN3e*RSe461p*RS345pe - 1./54.*me2*RN3e*RSe461p*RS346pe
          - 1./54.*me2*RN3e*RSe451p*RS345pe + 1./54.*me2*RN3e*RSe451p*
         RS346pe + 1./54.*me2*RN1e*RS145pe*RSe463p - 1./54.*me2*RN1e*
         RS145pe*RSe453p - 1./54.*me2*RN1e*RS146pe*RSe463p + 1./54.*me2
         *RN1e*RS146pe*RSe453p - 1./54.*me2*RN1e*RSe461p*RS145pe + 1./
         54.*me2*RN1e*RSe461p*RS146pe + 1./54.*me2*RN1e*RSe451p*RS145pe
          - 1./54.*me2*RN1e*RSe451p*RS146pe + 1./54.*m66*RNep*RSp463e*
         RSe453p );

      F13p6x23R +=  + Ab(m56,mu2)*Ab(me2,mu2) * ( 1./54.*m66*RNep*
         RSp453e*RSe463p + 1./54.*m66*RNep*RSp453e*RSe453p - 1./54.*m66
         *RNep*RSe461p*RSp453e - 1./54.*m66*RNep*RSe451p*RSp463e - 1./
         54.*m66*RNep*RSe451p*RSp453e - 1./54.*m66*RNep*RSp461e*RSe453p
          + 1./54.*m66*RNep*RSp461e*RSe451p - 1./54.*m66*RNep*RSp451e*
         RSe463p - 1./54.*m66*RNep*RSp451e*RSe453p + 1./54.*m66*RNep*
         RSp451e*RSe461p + 1./54.*m66*RNep*RSp451e*RSe451p - 1./54.*m66
         *RN3e*RSe463p*RS345pe - 1./54.*m66*RN3e*RSe453p*RS345pe - 1./
         54.*m66*RN3e*RSe453p*RS346pe + 1./54.*m66*RN3e*RSe461p*RS345pe
          + 1./54.*m66*RN3e*RSe451p*RS345pe + 1./54.*m66*RN3e*RSe451p*
         RS346pe + 1./54.*m66*RN1e*RS145pe*RSe463p + 1./54.*m66*RN1e*
         RS145pe*RSe453p + 1./54.*m66*RN1e*RS146pe*RSe453p - 1./54.*m66
         *RN1e*RSe461p*RS145pe - 1./54.*m66*RN1e*RSe451p*RS145pe - 1./
         54.*m66*RN1e*RSe451p*RS146pe + 1./54.*m55*RNep*RSp463e*RSe463p
          + 1./54.*m55*RNep*RSp463e*RSe453p + 1./54.*m55*RNep*RSp453e*
         RSe463p );

      F13p6x23R +=  + Ab(m56,mu2)*Ab(me2,mu2) * (  - 1./54.*m55*RNep*
         RSe461p*RSp463e - 1./54.*m55*RNep*RSe461p*RSp453e - 1./54.*m55
         *RNep*RSe451p*RSp463e - 1./54.*m55*RNep*RSp461e*RSe463p - 1./
         54.*m55*RNep*RSp461e*RSe453p + 1./54.*m55*RNep*RSp461e*RSe461p
          + 1./54.*m55*RNep*RSp461e*RSe451p - 1./54.*m55*RNep*RSp451e*
         RSe463p + 1./54.*m55*RNep*RSp451e*RSe461p - 1./54.*m55*RN3e*
         RSe463p*RS345pe - 1./54.*m55*RN3e*RSe463p*RS346pe - 1./54.*m55
         *RN3e*RSe453p*RS346pe + 1./54.*m55*RN3e*RSe461p*RS345pe + 1./
         54.*m55*RN3e*RSe461p*RS346pe + 1./54.*m55*RN3e*RSe451p*RS346pe
          + 1./54.*m55*RN1e*RS145pe*RSe463p + 1./54.*m55*RN1e*RS146pe*
         RSe463p + 1./54.*m55*RN1e*RS146pe*RSe453p - 1./54.*m55*RN1e*
         RSe461p*RS145pe - 1./54.*m55*RN1e*RSe461p*RS146pe - 1./54.*m55
         *RN1e*RSe451p*RS146pe );

      F13p6x23R +=  + Ab(m56,mu2)*Ab(mp2,mu2) * (  - 1./108.*pow(
         RSp463e,2) + 1./54.*RSp453e*RSp463e - 1./108.*pow(RSp453e,2)
          + 1./54.*RSp461e*RSp463e - 1./54.*RSp461e*RSp453e - 1./108.*
         pow(RSp461e,2) - 1./54.*RSp451e*RSp463e + 1./54.*RSp451e*
         RSp453e + 1./54.*RSp451e*RSp461e - 1./108.*pow(RSp451e,2) + 1./
         54.*mp2*RNep*RSp463e*RSe463p - 1./54.*mp2*RNep*RSp463e*RSe453p
          - 1./54.*mp2*RNep*RSp453e*RSe463p + 1./54.*mp2*RNep*RSp453e*
         RSe453p - 1./54.*mp2*RNep*RSe461p*RSp463e + 1./54.*mp2*RNep*
         RSe461p*RSp453e + 1./54.*mp2*RNep*RSe451p*RSp463e - 1./54.*mp2
         *RNep*RSe451p*RSp453e - 1./54.*mp2*RNep*RSp461e*RSe463p + 1./
         54.*mp2*RNep*RSp461e*RSe453p + 1./54.*mp2*RNep*RSp461e*RSe461p
          - 1./54.*mp2*RNep*RSp461e*RSe451p + 1./54.*mp2*RNep*RSp451e*
         RSe463p - 1./54.*mp2*RNep*RSp451e*RSe453p - 1./54.*mp2*RNep*
         RSp451e*RSe461p + 1./54.*mp2*RNep*RSp451e*RSe451p - 1./54.*mp2
         *RN3p*RSp463e*RS345pe + 1./54.*mp2*RN3p*RSp463e*RS346pe + 1./
         54.*mp2*RN3p*RSp453e*RS345pe );

      F13p6x23R +=  + Ab(m56,mu2)*Ab(mp2,mu2) * (  - 1./54.*mp2*RN3p*
         RSp453e*RS346pe + 1./54.*mp2*RN3p*RSp461e*RS345pe - 1./54.*mp2
         *RN3p*RSp461e*RS346pe - 1./54.*mp2*RN3p*RSp451e*RS345pe + 1./
         54.*mp2*RN3p*RSp451e*RS346pe + 1./54.*mp2*RN1p*RS145pe*RSp463e
          - 1./54.*mp2*RN1p*RS145pe*RSp453e - 1./54.*mp2*RN1p*RS146pe*
         RSp463e + 1./54.*mp2*RN1p*RS146pe*RSp453e - 1./54.*mp2*RN1p*
         RSp461e*RS145pe + 1./54.*mp2*RN1p*RSp461e*RS146pe + 1./54.*mp2
         *RN1p*RSp451e*RS145pe - 1./54.*mp2*RN1p*RSp451e*RS146pe - 1./
         54.*m66*RNep*RSp463e*RSe453p - 1./54.*m66*RNep*RSp453e*RSe463p
          - 1./54.*m66*RNep*RSp453e*RSe453p + 1./54.*m66*RNep*RSe461p*
         RSp453e + 1./54.*m66*RNep*RSe451p*RSp463e + 1./54.*m66*RNep*
         RSe451p*RSp453e + 1./54.*m66*RNep*RSp461e*RSe453p - 1./54.*m66
         *RNep*RSp461e*RSe451p + 1./54.*m66*RNep*RSp451e*RSe463p + 1./
         54.*m66*RNep*RSp451e*RSe453p - 1./54.*m66*RNep*RSp451e*RSe461p
          - 1./54.*m66*RNep*RSp451e*RSe451p - 1./54.*m66*RN3p*RSp463e*
         RS345pe );

      F13p6x23R +=  + Ab(m56,mu2)*Ab(mp2,mu2) * (  - 1./54.*m66*RN3p*
         RSp453e*RS345pe - 1./54.*m66*RN3p*RSp453e*RS346pe + 1./54.*m66
         *RN3p*RSp461e*RS345pe + 1./54.*m66*RN3p*RSp451e*RS345pe + 1./
         54.*m66*RN3p*RSp451e*RS346pe + 1./54.*m66*RN1p*RS145pe*RSp463e
          + 1./54.*m66*RN1p*RS145pe*RSp453e + 1./54.*m66*RN1p*RS146pe*
         RSp453e - 1./54.*m66*RN1p*RSp461e*RS145pe - 1./54.*m66*RN1p*
         RSp451e*RS145pe - 1./54.*m66*RN1p*RSp451e*RS146pe - 1./54.*m55
         *RNep*RSp463e*RSe463p - 1./54.*m55*RNep*RSp463e*RSe453p - 1./
         54.*m55*RNep*RSp453e*RSe463p + 1./54.*m55*RNep*RSe461p*RSp463e
          + 1./54.*m55*RNep*RSe461p*RSp453e + 1./54.*m55*RNep*RSe451p*
         RSp463e + 1./54.*m55*RNep*RSp461e*RSe463p + 1./54.*m55*RNep*
         RSp461e*RSe453p - 1./54.*m55*RNep*RSp461e*RSe461p - 1./54.*m55
         *RNep*RSp461e*RSe451p + 1./54.*m55*RNep*RSp451e*RSe463p - 1./
         54.*m55*RNep*RSp451e*RSe461p - 1./54.*m55*RN3p*RSp463e*RS345pe
          - 1./54.*m55*RN3p*RSp463e*RS346pe - 1./54.*m55*RN3p*RSp453e*
         RS346pe );

      F13p6x23R +=  + Ab(m56,mu2)*Ab(mp2,mu2) * ( 1./54.*m55*RN3p*
         RSp461e*RS345pe + 1./54.*m55*RN3p*RSp461e*RS346pe + 1./54.*m55
         *RN3p*RSp451e*RS346pe + 1./54.*m55*RN1p*RS145pe*RSp463e + 1./
         54.*m55*RN1p*RS146pe*RSp463e + 1./54.*m55*RN1p*RS146pe*RSp453e
          - 1./54.*m55*RN1p*RSp461e*RS145pe - 1./54.*m55*RN1p*RSp461e*
         RS146pe - 1./54.*m55*RN1p*RSp451e*RS146pe );

      F13p6x23R +=  + Ab(m56,mu2)*Bb(m11,mu2) * ( 1./108.*m66*pow(
         RS145pe,2) + 1./54.*m66*RS146pe*RS145pe + 1./54.*m55*RS146pe*
         RS145pe + 1./108.*m55*pow(RS146pe,2) - 1./108.*m11*pow(
         RS145pe,2) + 1./54.*m11*RS146pe*RS145pe - 1./108.*m11*pow(
         RS146pe,2) );

      F13p6x23R +=  + Ab(m56,mu2)*Bb(m33,mu2) * ( 1./108.*m66*pow(
         RS345pe,2) + 1./54.*m66*RS346pe*RS345pe + 1./54.*m55*RS346pe*
         RS345pe + 1./108.*m55*pow(RS346pe,2) - 1./108.*m33*pow(
         RS345pe,2) + 1./54.*m33*RS346pe*RS345pe - 1./108.*m33*pow(
         RS346pe,2) );

      F13p6x23R +=  + Ab(m56,mu2)*Bb(me2,mu2) * (  - 1./108.*me2*pow(
         RSe463p,2) + 1./54.*me2*RSe453p*RSe463p - 1./108.*me2*pow(
         RSe453p,2) + 1./54.*me2*RSe461p*RSe463p - 1./54.*me2*RSe461p*
         RSe453p - 1./108.*me2*pow(RSe461p,2) - 1./54.*me2*RSe451p*
         RSe463p + 1./54.*me2*RSe451p*RSe453p + 1./54.*me2*RSe451p*
         RSe461p - 1./108.*me2*pow(RSe451p,2) + 1./54.*m66*RSe453p*
         RSe463p + 1./108.*m66*pow(RSe453p,2) - 1./54.*m66*RSe461p*
         RSe453p - 1./54.*m66*RSe451p*RSe463p - 1./54.*m66*RSe451p*
         RSe453p + 1./54.*m66*RSe451p*RSe461p + 1./108.*m66*pow(
         RSe451p,2) + 1./108.*m55*pow(RSe463p,2) + 1./54.*m55*RSe453p*
         RSe463p - 1./54.*m55*RSe461p*RSe463p - 1./54.*m55*RSe461p*
         RSe453p + 1./108.*m55*pow(RSe461p,2) - 1./54.*m55*RSe451p*
         RSe463p + 1./54.*m55*RSe451p*RSe461p );

      F13p6x23R +=  + Ab(m56,mu2)*Bb(mp2,mu2) * (  - 1./108.*mp2*pow(
         RSp463e,2) + 1./54.*mp2*RSp453e*RSp463e - 1./108.*mp2*pow(
         RSp453e,2) + 1./54.*mp2*RSp461e*RSp463e - 1./54.*mp2*RSp461e*
         RSp453e - 1./108.*mp2*pow(RSp461e,2) - 1./54.*mp2*RSp451e*
         RSp463e + 1./54.*mp2*RSp451e*RSp453e + 1./54.*mp2*RSp451e*
         RSp461e - 1./108.*mp2*pow(RSp451e,2) + 1./54.*m66*RSp453e*
         RSp463e + 1./108.*m66*pow(RSp453e,2) - 1./54.*m66*RSp461e*
         RSp453e - 1./54.*m66*RSp451e*RSp463e - 1./54.*m66*RSp451e*
         RSp453e + 1./54.*m66*RSp451e*RSp461e + 1./108.*m66*pow(
         RSp451e,2) + 1./108.*m55*pow(RSp463e,2) + 1./54.*m55*RSp453e*
         RSp463e - 1./54.*m55*RSp461e*RSp463e - 1./54.*m55*RSp461e*
         RSp453e + 1./108.*m55*pow(RSp461e,2) - 1./54.*m55*RSp451e*
         RSp463e + 1./54.*m55*RSp451e*RSp461e );

      F13p6x23R +=  + Ab(m66,mu2)*Ab(me2,mu2) * ( 1./18.*m66*RNep*
         RSp453e*RSe453p - 1./54.*m66*RNep*RS645pe*RSp453e*RSe453p - 1./
         18.*m66*RNep*RSe451p*RSp453e + 1./54.*m66*RNep*RSe451p*RS645pe
         *RSp453e - 1./18.*m66*RNep*RSp451e*RSe453p + 1./54.*m66*RNep*
         RSp451e*RS645pe*RSe453p + 1./18.*m66*RNep*RSp451e*RSe451p - 1./
         54.*m66*RNep*RSp451e*RSe451p*RS645pe - 1./18.*m66*RN3e*RSe453p
         *RS345pe + 1./54.*m66*RN3e*RS645pe*RSe453p*RS345pe + 1./18.*
         m66*RN3e*RSe451p*RS345pe - 1./54.*m66*RN3e*RSe451p*RS645pe*
         RS345pe + 1./18.*m66*RN1e*RS145pe*RSe453p - 1./54.*m66*RN1e*
         RS645pe*RS145pe*RSe453p - 1./18.*m66*RN1e*RSe451p*RS145pe + 1./
         54.*m66*RN1e*RSe451p*RS645pe*RS145pe );

      F13p6x23R +=  + Ab(m66,mu2)*Ab(mp2,mu2) * (  - 1./18.*m66*RNep*
         RSp453e*RSe453p + 1./54.*m66*RNep*RS645pe*RSp453e*RSe453p + 1./
         18.*m66*RNep*RSe451p*RSp453e - 1./54.*m66*RNep*RSe451p*RS645pe
         *RSp453e + 1./18.*m66*RNep*RSp451e*RSe453p - 1./54.*m66*RNep*
         RSp451e*RS645pe*RSe453p - 1./18.*m66*RNep*RSp451e*RSe451p + 1./
         54.*m66*RNep*RSp451e*RSe451p*RS645pe - 1./18.*m66*RN3p*RSp453e
         *RS345pe + 1./54.*m66*RN3p*RS645pe*RSp453e*RS345pe + 1./18.*
         m66*RN3p*RSp451e*RS345pe - 1./54.*m66*RN3p*RSp451e*RS645pe*
         RS345pe + 1./18.*m66*RN1p*RS145pe*RSp453e - 1./54.*m66*RN1p*
         RS645pe*RS145pe*RSp453e - 1./18.*m66*RN1p*RSp451e*RS145pe + 1./
         54.*m66*RN1p*RSp451e*RS645pe*RS145pe );

      F13p6x23R +=  + Ab(m66,mu2)*Bb(m11,mu2) * ( 1./36.*m66*pow(
         RS145pe,2) - 1./108.*m66*RS645pe*pow(RS145pe,2) );

      F13p6x23R +=  + Ab(m66,mu2)*Bb(m33,mu2) * ( 1./36.*m66*pow(
         RS345pe,2) - 1./108.*m66*RS645pe*pow(RS345pe,2) );

      F13p6x23R +=  + Ab(m66,mu2)*Bb(me2,mu2) * ( 1./36.*m66*pow(
         RSe453p,2) - 1./108.*m66*RS645pe*pow(RSe453p,2) - 1./18.*m66*
         RSe451p*RSe453p + 1./54.*m66*RSe451p*RS645pe*RSe453p + 1./36.*
         m66*pow(RSe451p,2) - 1./108.*m66*pow(RSe451p,2)*RS645pe );

      F13p6x23R +=  + Ab(m66,mu2)*Bb(mp2,mu2) * ( 1./36.*m66*pow(
         RSp453e,2) - 1./108.*m66*RS645pe*pow(RSp453e,2) - 1./18.*m66*
         RSp451e*RSp453e + 1./54.*m66*RSp451e*RS645pe*RSp453e + 1./36.*
         m66*pow(RSp451e,2) - 1./108.*m66*pow(RSp451e,2)*RS645pe );

      F13p6x23R +=  + Ab(me2,mu2)*pi16 * ( 1./72.*RD3456ep*RSe45613p - 
         1./144.*RSe45633p*RD3456ep + 1./72.*RD1456ep*RSe45613p - 1./
         144.*RD1456ep*RSe45633p - 1./144.*RSe45611p*RD3456ep - 1./144.
         *RSe45611p*RD1456ep - 1./36.*mp2*RSp45613e*RSe45613p + 1./72.*
         mp2*RSe45633p*RSp45613e + 1./72.*mp2*RSp45633e*RSe45613p - 1./
         144.*mp2*RSp45633e*RSe45633p + 1./72.*mp2*RSe45611p*RSp45613e
          - 1./144.*mp2*RSe45611p*RSp45633e + 1./72.*mp2*RSp45611e*
         RSe45613p - 1./144.*mp2*RSp45611e*RSe45633p - 1./144.*mp2*
         RSp45611e*RSe45611p + 1./24.*me2*RSe45613p - 1./36.*me2*pow(
         RSe45613p,2) - 1./48.*me2*RSe45633p + 1./36.*me2*RSe45633p*
         RSe45613p - 1./144.*me2*pow(RSe45633p,2) - 1./48.*me2*RSe563p
          - 1./48.*me2*RSe463p - 1./48.*me2*RSe453p - 1./48.*me2*
         RSe45611p + 1./36.*me2*RSe45611p*RSe45613p - 1./72.*me2*
         RSe45611p*RSe45633p - 1./144.*me2*pow(RSe45611p,2) - 1./48.*
         me2*RSe561p - 1./48.*me2*RSe461p - 1./48.*me2*RSe451p - 1./24.
         *m66*RSe45613p );

      F13p6x23R +=  + Ab(me2,mu2)*pi16 * ( 1./48.*m66*RSe45633p + 1./48.
         *m66*RSe45611p - 1./24.*m55*RSe45613p + 1./48.*m55*RSe45633p
          + 1./48.*m55*RSe45611p - 1./24.*m44*RSe45613p + 1./48.*m44*
         RSe45633p + 1./48.*m44*RSe45611p - 1./16.*m33*RSe45613p - 1./
         24.*m33*RSe45613p*RS34561pe - 1./72.*m33*RSe45613p*RS14563pe
          - 1./72.*m33*pow(RSe45613p,2) - 1./72.*m33*RSp45613e*
         RSe45613p - 1./48.*m33*RS3456pee*RSe45613p - 1./48.*m33*
         RS3456ppe*RSe45613p + 1./32.*m33*RSe45633p + 1./48.*m33*
         RSe45633p*RS34561pe + 1./144.*m33*RSe45633p*RS14563pe + 1./72.
         *m33*RSe45633p*RSe45613p + 1./144.*m33*RSe45633p*RSp45613e + 1.
         /96.*m33*RSe45633p*RS3456pee + 1./96.*m33*RSe45633p*RS3456ppe
          - 1./288.*m33*pow(RSe45633p,2) + 1./144.*m33*RSp45633e*
         RSe45613p - 1./288.*m33*RSp45633e*RSe45633p + 1./48.*m33*
         RS345pe*RSe45613p - 1./96.*m33*RS345pe*RSe45633p + 1./48.*m33*
         RS346pe*RSe45613p - 1./96.*m33*RS346pe*RSe45633p + 1./48.*m33*
         RS356pe*RSe45613p );

      F13p6x23R +=  + Ab(me2,mu2)*pi16 * (  - 1./96.*m33*RS356pe*
         RSe45633p - 1./144.*m33*RS1456pee*RSe45613p + 1./288.*m33*
         RS1456pee*RSe45633p - 1./144.*m33*RS1456ppe*RSe45613p + 1./288.
         *m33*RS1456ppe*RSe45633p + 1./32.*m33*RSe45611p + 1./48.*m33*
         RSe45611p*RS34561pe + 1./144.*m33*RSe45611p*RS14563pe + 1./72.
         *m33*RSe45611p*RSe45613p + 1./144.*m33*RSe45611p*RSp45613e + 1.
         /96.*m33*RSe45611p*RS3456pee + 1./96.*m33*RSe45611p*RS3456ppe
          - 1./144.*m33*RSe45611p*RSe45633p - 1./288.*m33*RSe45611p*
         RSp45633e - 1./96.*m33*RSe45611p*RS345pe - 1./96.*m33*
         RSe45611p*RS346pe - 1./96.*m33*RSe45611p*RS356pe + 1./288.*m33
         *RSe45611p*RS1456pee + 1./288.*m33*RSe45611p*RS1456ppe - 1./
         288.*m33*pow(RSe45611p,2) + 1./144.*m33*RSp45611e*RSe45613p - 
         1./288.*m33*RSp45611e*RSe45633p - 1./288.*m33*RSp45611e*
         RSe45611p + 1./144.*m33*RS145pe*RSe45613p - 1./288.*m33*
         RS145pe*RSe45633p - 1./288.*m33*RS145pe*RSe45611p + 1./144.*
         m33*RS146pe*RSe45613p );

      F13p6x23R +=  + Ab(me2,mu2)*pi16 * (  - 1./288.*m33*RS146pe*
         RSe45633p - 1./288.*m33*RS146pe*RSe45611p + 1./144.*m33*
         RS156pe*RSe45613p - 1./288.*m33*RS156pe*RSe45633p - 1./288.*
         m33*RS156pe*RSe45611p - 1./16.*m11*RSe45613p - 1./72.*m11*
         RSe45613p*RS34561pe - 1./24.*m11*RSe45613p*RS14563pe - 1./72.*
         m11*pow(RSe45613p,2) - 1./72.*m11*RSp45613e*RSe45613p - 1./144.
         *m11*RS3456pee*RSe45613p - 1./144.*m11*RS3456ppe*RSe45613p + 1.
         /32.*m11*RSe45633p + 1./144.*m11*RSe45633p*RS34561pe + 1./48.*
         m11*RSe45633p*RS14563pe + 1./72.*m11*RSe45633p*RSe45613p + 1./
         144.*m11*RSe45633p*RSp45613e + 1./288.*m11*RSe45633p*RS3456pee
          + 1./288.*m11*RSe45633p*RS3456ppe - 1./288.*m11*pow(
         RSe45633p,2) + 1./144.*m11*RSp45633e*RSe45613p - 1./288.*m11*
         RSp45633e*RSe45633p + 1./144.*m11*RS345pe*RSe45613p - 1./288.*
         m11*RS345pe*RSe45633p + 1./144.*m11*RS346pe*RSe45613p - 1./288.
         *m11*RS346pe*RSe45633p + 1./144.*m11*RS356pe*RSe45613p - 1./
         288.*m11*RS356pe*RSe45633p );

      F13p6x23R +=  + Ab(me2,mu2)*pi16 * (  - 1./48.*m11*RS1456pee*
         RSe45613p + 1./96.*m11*RS1456pee*RSe45633p - 1./48.*m11*
         RS1456ppe*RSe45613p + 1./96.*m11*RS1456ppe*RSe45633p + 1./32.*
         m11*RSe45611p + 1./144.*m11*RSe45611p*RS34561pe + 1./48.*m11*
         RSe45611p*RS14563pe + 1./72.*m11*RSe45611p*RSe45613p + 1./144.
         *m11*RSe45611p*RSp45613e + 1./288.*m11*RSe45611p*RS3456pee + 1.
         /288.*m11*RSe45611p*RS3456ppe - 1./144.*m11*RSe45611p*
         RSe45633p - 1./288.*m11*RSe45611p*RSp45633e - 1./288.*m11*
         RSe45611p*RS345pe - 1./288.*m11*RSe45611p*RS346pe - 1./288.*
         m11*RSe45611p*RS356pe + 1./96.*m11*RSe45611p*RS1456pee + 1./96.
         *m11*RSe45611p*RS1456ppe - 1./288.*m11*pow(RSe45611p,2) + 1./
         144.*m11*RSp45611e*RSe45613p - 1./288.*m11*RSp45611e*RSe45633p
          - 1./288.*m11*RSp45611e*RSe45611p + 1./48.*m11*RS145pe*
         RSe45613p - 1./96.*m11*RS145pe*RSe45633p - 1./96.*m11*RS145pe*
         RSe45611p + 1./48.*m11*RS146pe*RSe45613p - 1./96.*m11*RS146pe*
         RSe45633p );

      F13p6x23R +=  + Ab(me2,mu2)*pi16 * (  - 1./96.*m11*RS146pe*
         RSe45611p + 1./48.*m11*RS156pe*RSe45613p - 1./96.*m11*RS156pe*
         RSe45633p - 1./96.*m11*RS156pe*RSe45611p );

      F13p6x23R +=  + pow(Ab(me2,mu2),2) * ( 1./72.*pow(RSe45613p,2) - 
         1./72.*RSe45633p*RSe45613p + 1./288.*pow(RSe45633p,2) - 1./72.
         *RSe45611p*RSe45613p + 1./144.*RSe45611p*RSe45633p + 1./288.*
         pow(RSe45611p,2) - 1./54.*m66*RNep*RSe456p*RSp453e*RSe453p + 1.
         /54.*m66*RNep*RSe456p*RSe451p*RSp453e + 1./54.*m66*RNep*
         RSe456p*RSp451e*RSe453p - 1./54.*m66*RNep*RSe456p*RSp451e*
         RSe451p + 1./54.*m66*RN3e*RSe456p*RSe453p*RS345pe - 1./54.*m66
         *RN3e*RSe456p*RSe451p*RS345pe - 1./54.*m66*RN1e*RSe456p*
         RS145pe*RSe453p + 1./54.*m66*RN1e*RSe456p*RSe451p*RS145pe - 1./
         54.*m55*RNep*RSe465p*RSp463e*RSe463p + 1./54.*m55*RNep*RSe465p
         *RSe461p*RSp463e + 1./54.*m55*RNep*RSe465p*RSp461e*RSe463p - 1.
         /54.*m55*RNep*RSe465p*RSp461e*RSe461p + 1./54.*m55*RN3e*
         RSe465p*RSe463p*RS346pe - 1./54.*m55*RN3e*RSe465p*RSe461p*
         RS346pe - 1./54.*m55*RN1e*RSe465p*RS146pe*RSe463p + 1./54.*m55
         *RN1e*RSe465p*RSe461p*RS146pe - 1./54.*m44*RNep*RSe564p*
         RSp563e*RSe563p );

      F13p6x23R +=  + pow(Ab(me2,mu2),2) * ( 1./54.*m44*RNep*RSe564p*
         RSe561p*RSp563e + 1./54.*m44*RNep*RSe564p*RSp561e*RSe563p - 1./
         54.*m44*RNep*RSe564p*RSp561e*RSe561p + 1./54.*m44*RN3e*RSe564p
         *RSe563p*RS356pe - 1./54.*m44*RN3e*RSe564p*RSe561p*RS356pe - 1.
         /54.*m44*RN1e*RSe564p*RS156pe*RSe563p + 1./54.*m44*RN1e*
         RSe564p*RSe561p*RS156pe + 1./18.*m33*RN3e*RSe45633p*RSe45613p
          - 1./18.*m33*RN3e*pow(RSe45633p,2) + 1./18.*m11*RN1e*
         RSe45611p*RSe45613p - 1./18.*m11*RN1e*pow(RSe45611p,2) );

      F13p6x23R +=  + Ab(me2,mu2)*Ab(mp2,mu2) * ( 1./36.*RSp45613e*
         RSe45613p - 1./72.*RSe45633p*RSp45613e - 1./72.*RSp45633e*
         RSe45613p + 1./144.*RSp45633e*RSe45633p - 1./72.*RSe45611p*
         RSp45613e + 1./144.*RSe45611p*RSp45633e - 1./72.*RSp45611e*
         RSe45613p + 1./144.*RSp45611e*RSe45633p + 1./144.*RSp45611e*
         RSe45611p + 1./54.*m66*RNep*RSe456p*RSp453e*RSe453p - 1./54.*
         m66*RNep*RSe456p*RSe451p*RSp453e - 1./54.*m66*RNep*RSe456p*
         RSp451e*RSe453p + 1./54.*m66*RNep*RSe456p*RSp451e*RSe451p - 1./
         54.*m66*RNep*RSp456e*RSp453e*RSe453p + 1./54.*m66*RNep*RSp456e
         *RSe451p*RSp453e + 1./54.*m66*RNep*RSp456e*RSp451e*RSe453p - 1.
         /54.*m66*RNep*RSp456e*RSp451e*RSe451p + 1./54.*m66*RN3p*
         RSe456p*RSp453e*RS345pe - 1./54.*m66*RN3p*RSe456p*RSp451e*
         RS345pe + 1./54.*m66*RN3e*RSp456e*RSe453p*RS345pe - 1./54.*m66
         *RN3e*RSp456e*RSe451p*RS345pe - 1./54.*m66*RN1p*RSe456p*
         RS145pe*RSp453e + 1./54.*m66*RN1p*RSe456p*RSp451e*RS145pe - 1./
         54.*m66*RN1e*RSp456e*RS145pe*RSe453p );

      F13p6x23R +=  + Ab(me2,mu2)*Ab(mp2,mu2) * ( 1./54.*m66*RN1e*
         RSp456e*RSe451p*RS145pe + 1./54.*m55*RNep*RSe465p*RSp463e*
         RSe463p - 1./54.*m55*RNep*RSe465p*RSe461p*RSp463e - 1./54.*m55
         *RNep*RSe465p*RSp461e*RSe463p + 1./54.*m55*RNep*RSe465p*
         RSp461e*RSe461p - 1./54.*m55*RNep*RSp465e*RSp463e*RSe463p + 1./
         54.*m55*RNep*RSp465e*RSe461p*RSp463e + 1./54.*m55*RNep*RSp465e
         *RSp461e*RSe463p - 1./54.*m55*RNep*RSp465e*RSp461e*RSe461p + 1.
         /54.*m55*RN3p*RSe465p*RSp463e*RS346pe - 1./54.*m55*RN3p*
         RSe465p*RSp461e*RS346pe + 1./54.*m55*RN3e*RSp465e*RSe463p*
         RS346pe - 1./54.*m55*RN3e*RSp465e*RSe461p*RS346pe - 1./54.*m55
         *RN1p*RSe465p*RS146pe*RSp463e + 1./54.*m55*RN1p*RSe465p*
         RSp461e*RS146pe - 1./54.*m55*RN1e*RSp465e*RS146pe*RSe463p + 1./
         54.*m55*RN1e*RSp465e*RSe461p*RS146pe + 1./54.*m44*RNep*RSe564p
         *RSp563e*RSe563p - 1./54.*m44*RNep*RSe564p*RSe561p*RSp563e - 1.
         /54.*m44*RNep*RSe564p*RSp561e*RSe563p + 1./54.*m44*RNep*
         RSe564p*RSp561e*RSe561p );

      F13p6x23R +=  + Ab(me2,mu2)*Ab(mp2,mu2) * (  - 1./54.*m44*RNep*
         RSp564e*RSp563e*RSe563p + 1./54.*m44*RNep*RSp564e*RSe561p*
         RSp563e + 1./54.*m44*RNep*RSp564e*RSp561e*RSe563p - 1./54.*m44
         *RNep*RSp564e*RSp561e*RSe561p + 1./54.*m44*RN3p*RSe564p*
         RSp563e*RS356pe - 1./54.*m44*RN3p*RSe564p*RSp561e*RS356pe + 1./
         54.*m44*RN3e*RSp564e*RSe563p*RS356pe - 1./54.*m44*RN3e*RSp564e
         *RSe561p*RS356pe - 1./54.*m44*RN1p*RSe564p*RS156pe*RSp563e + 1.
         /54.*m44*RN1p*RSe564p*RSp561e*RS156pe - 1./54.*m44*RN1e*
         RSp564e*RS156pe*RSe563p + 1./54.*m44*RN1e*RSp564e*RSe561p*
         RS156pe + 1./18.*m33*RN3p*RSe45633p*RSp45613e - 1./18.*m33*
         RN3p*RSp45633e*RSe45633p + 1./18.*m33*RN3e*RSp45633e*RSe45613p
          - 1./18.*m33*RN3e*RSp45633e*RSe45633p + 1./18.*m11*RN1p*
         RSe45611p*RSp45613e - 1./18.*m11*RN1p*RSp45611e*RSe45611p + 1./
         18.*m11*RN1e*RSp45611e*RSe45613p - 1./18.*m11*RN1e*RSp45611e*
         RSe45611p );

      F13p6x23R +=  + Ab(me2,mu2)*Bb(m11,mu2) * (  - 1./72.*RD1456ep*
         RSe45613p + 1./144.*RD1456ep*RSe45633p + 1./144.*RSe45611p*
         RD1456ep - 1./108.*m66*RSe456p*pow(RS145pe,2) - 1./108.*m55*
         RSe465p*pow(RS146pe,2) - 1./108.*m44*RSe564p*pow(RS156pe,2) - 
         1./18.*m11*RSe45611p*RS14563pe - 1./18.*m11*RSe45611p*
         RS1456pee - 1./18.*m11*RSe45611p*RS1456ppe + 1./18.*m11*
         RS145pe*RSe45611p + 1./18.*m11*RS146pe*RSe45611p + 1./18.*m11*
         RS156pe*RSe45611p + 1./18.*m11*RN1e*RD1456ep*RSe45613p - 1./18.
         *m11*RN1e*RSe45611p*RD1456ep );

      F13p6x23R +=  + Ab(me2,mu2)*Bb(m14,mu2) * ( 1./72.*me2*RSe45611p
          - 1./36.*me2*RSe561p + 1./72.*me2*RSe564p - 1./36.*m44*
         RSe561p - 1./72.*m44*RSe564p - 1./72.*m11*RSe45611p - 1./36.*
         m11*RSe561p );

      F13p6x23R +=  + Ab(me2,mu2)*Bb(m15,mu2) * ( 1./72.*me2*RSe45611p
          - 1./36.*me2*RSe461p + 1./72.*me2*RSe465p - 1./36.*m55*
         RSe461p - 1./72.*m55*RSe465p - 1./72.*m11*RSe45611p - 1./36.*
         m11*RSe461p );

      F13p6x23R +=  + Ab(me2,mu2)*Bb(m16,mu2) * ( 1./72.*me2*RSe45611p
          - 1./36.*me2*RSe451p + 1./72.*me2*RSe456p - 1./36.*m66*
         RSe451p - 1./72.*m66*RSe456p - 1./72.*m11*RSe45611p - 1./36.*
         m11*RSe451p );

      F13p6x23R +=  + Ab(me2,mu2)*Bb(m33,mu2) * (  - 1./72.*RD3456ep*
         RSe45613p + 1./144.*RSe45633p*RD3456ep + 1./144.*RSe45611p*
         RD3456ep - 1./108.*m66*RSe456p*pow(RS345pe,2) - 1./108.*m55*
         RSe465p*pow(RS346pe,2) - 1./108.*m44*RSe564p*pow(RS356pe,2) - 
         1./18.*m33*RSe45633p*RS34561pe - 1./18.*m33*RSe45633p*
         RS3456pee - 1./18.*m33*RSe45633p*RS3456ppe + 1./18.*m33*
         RS345pe*RSe45633p + 1./18.*m33*RS346pe*RSe45633p + 1./18.*m33*
         RS356pe*RSe45633p + 1./18.*m33*RN3e*RD3456ep*RSe45613p - 1./18.
         *m33*RN3e*RSe45633p*RD3456ep );

      F13p6x23R +=  + Ab(me2,mu2)*Bb(m34,mu2) * ( 1./72.*me2*RSe45633p
          - 1./36.*me2*RSe563p + 1./72.*me2*RSe564p - 1./36.*m44*
         RSe563p - 1./72.*m44*RSe564p - 1./72.*m33*RSe45633p - 1./36.*
         m33*RSe563p );

      F13p6x23R +=  + Ab(me2,mu2)*Bb(m35,mu2) * ( 1./72.*me2*RSe45633p
          - 1./36.*me2*RSe463p + 1./72.*me2*RSe465p - 1./36.*m55*
         RSe463p - 1./72.*m55*RSe465p - 1./72.*m33*RSe45633p - 1./36.*
         m33*RSe463p );

      F13p6x23R +=  + Ab(me2,mu2)*Bb(m36,mu2) * ( 1./72.*me2*RSe45633p
          - 1./36.*me2*RSe453p + 1./72.*me2*RSe456p - 1./36.*m66*
         RSe453p - 1./72.*m66*RSe456p - 1./72.*m33*RSe45633p - 1./36.*
         m33*RSe453p );

      F13p6x23R +=  + Ab(me2,mu2)*Bb(me2,mu2) * (  - 1./108.*m66*
         RSe456p*pow(RSe453p,2) + 1./54.*m66*RSe456p*RSe451p*RSe453p - 
         1./108.*m66*RSe456p*pow(RSe451p,2) - 1./108.*m55*RSe465p*pow(
         RSe463p,2) + 1./54.*m55*RSe465p*RSe461p*RSe463p - 1./108.*m55*
         RSe465p*pow(RSe461p,2) - 1./108.*m44*RSe564p*pow(RSe563p,2) + 
         1./54.*m44*RSe564p*RSe561p*RSe563p - 1./108.*m44*RSe564p*pow(
         RSe561p,2) );

      F13p6x23R +=  + Ab(me2,mu2)*Bb(mp2,mu2) * (  - 1./108.*m66*
         RSe456p*pow(RSp453e,2) + 1./54.*m66*RSe456p*RSp451e*RSp453e - 
         1./108.*m66*RSe456p*pow(RSp451e,2) - 1./108.*m55*RSe465p*pow(
         RSp463e,2) + 1./54.*m55*RSe465p*RSp461e*RSp463e - 1./108.*m55*
         RSe465p*pow(RSp461e,2) - 1./108.*m44*RSe564p*pow(RSp563e,2) + 
         1./54.*m44*RSe564p*RSp561e*RSp563e - 1./108.*m44*RSe564p*pow(
         RSp561e,2) );

      F13p6x23R +=  + Ab(me2,mu2)*Cb(m11,mu2) * ( 1./18.*m11*RSe45611p*
         RD1456ep );

      F13p6x23R +=  + Ab(me2,mu2)*Cb(m33,mu2) * ( 1./18.*m33*RSe45633p*
         RD3456ep );

      F13p6x23R +=  + Ab(mp2,mu2)*pi16 * ( 1./72.*RD3456ep*RSp45613e - 
         1./144.*RSp45633e*RD3456ep + 1./72.*RD1456ep*RSp45613e - 1./
         144.*RD1456ep*RSp45633e - 1./144.*RSp45611e*RD3456ep - 1./144.
         *RSp45611e*RD1456ep + 1./24.*mp2*RSp45613e - 1./36.*mp2*pow(
         RSp45613e,2) - 1./48.*mp2*RSp45633e + 1./36.*mp2*RSp45633e*
         RSp45613e - 1./144.*mp2*pow(RSp45633e,2) - 1./48.*mp2*RSp563e
          - 1./48.*mp2*RSp463e - 1./48.*mp2*RSp453e - 1./48.*mp2*
         RSp45611e + 1./36.*mp2*RSp45611e*RSp45613e - 1./72.*mp2*
         RSp45611e*RSp45633e - 1./144.*mp2*pow(RSp45611e,2) - 1./48.*
         mp2*RSp561e - 1./48.*mp2*RSp461e - 1./48.*mp2*RSp451e - 1./36.
         *me2*RSp45613e*RSe45613p + 1./72.*me2*RSe45633p*RSp45613e + 1./
         72.*me2*RSp45633e*RSe45613p - 1./144.*me2*RSp45633e*RSe45633p
          + 1./72.*me2*RSe45611p*RSp45613e - 1./144.*me2*RSe45611p*
         RSp45633e + 1./72.*me2*RSp45611e*RSe45613p - 1./144.*me2*
         RSp45611e*RSe45633p - 1./144.*me2*RSp45611e*RSe45611p - 1./24.
         *m66*RSp45613e );

      F13p6x23R +=  + Ab(mp2,mu2)*pi16 * ( 1./48.*m66*RSp45633e + 1./48.
         *m66*RSp45611e - 1./24.*m55*RSp45613e + 1./48.*m55*RSp45633e
          + 1./48.*m55*RSp45611e - 1./24.*m44*RSp45613e + 1./48.*m44*
         RSp45633e + 1./48.*m44*RSp45611e - 1./16.*m33*RSp45613e - 1./
         24.*m33*RSp45613e*RS34561pe - 1./72.*m33*RSp45613e*RS14563pe
          - 1./72.*m33*RSp45613e*RSe45613p - 1./72.*m33*pow(
         RSp45613e,2) - 1./48.*m33*RS3456pee*RSp45613e - 1./48.*m33*
         RS3456ppe*RSp45613e + 1./144.*m33*RSe45633p*RSp45613e + 1./32.
         *m33*RSp45633e + 1./48.*m33*RSp45633e*RS34561pe + 1./144.*m33*
         RSp45633e*RS14563pe + 1./144.*m33*RSp45633e*RSe45613p + 1./72.
         *m33*RSp45633e*RSp45613e + 1./96.*m33*RSp45633e*RS3456pee + 1./
         96.*m33*RSp45633e*RS3456ppe - 1./288.*m33*RSp45633e*RSe45633p
          - 1./288.*m33*pow(RSp45633e,2) + 1./48.*m33*RS345pe*RSp45613e
          - 1./96.*m33*RS345pe*RSp45633e + 1./48.*m33*RS346pe*RSp45613e
          - 1./96.*m33*RS346pe*RSp45633e + 1./48.*m33*RS356pe*RSp45613e
          - 1./96.*m33*RS356pe*RSp45633e );

      F13p6x23R +=  + Ab(mp2,mu2)*pi16 * (  - 1./144.*m33*RS1456pee*
         RSp45613e + 1./288.*m33*RS1456pee*RSp45633e - 1./144.*m33*
         RS1456ppe*RSp45613e + 1./288.*m33*RS1456ppe*RSp45633e + 1./144.
         *m33*RSe45611p*RSp45613e - 1./288.*m33*RSe45611p*RSp45633e + 1.
         /32.*m33*RSp45611e + 1./48.*m33*RSp45611e*RS34561pe + 1./144.*
         m33*RSp45611e*RS14563pe + 1./144.*m33*RSp45611e*RSe45613p + 1./
         72.*m33*RSp45611e*RSp45613e + 1./96.*m33*RSp45611e*RS3456pee
          + 1./96.*m33*RSp45611e*RS3456ppe - 1./288.*m33*RSp45611e*
         RSe45633p - 1./144.*m33*RSp45611e*RSp45633e - 1./96.*m33*
         RSp45611e*RS345pe - 1./96.*m33*RSp45611e*RS346pe - 1./96.*m33*
         RSp45611e*RS356pe + 1./288.*m33*RSp45611e*RS1456pee + 1./288.*
         m33*RSp45611e*RS1456ppe - 1./288.*m33*RSp45611e*RSe45611p - 1./
         288.*m33*pow(RSp45611e,2) + 1./144.*m33*RS145pe*RSp45613e - 1./
         288.*m33*RS145pe*RSp45633e - 1./288.*m33*RS145pe*RSp45611e + 1.
         /144.*m33*RS146pe*RSp45613e - 1./288.*m33*RS146pe*RSp45633e - 
         1./288.*m33*RS146pe*RSp45611e );

      F13p6x23R +=  + Ab(mp2,mu2)*pi16 * ( 1./144.*m33*RS156pe*
         RSp45613e - 1./288.*m33*RS156pe*RSp45633e - 1./288.*m33*
         RS156pe*RSp45611e - 1./16.*m11*RSp45613e - 1./72.*m11*
         RSp45613e*RS34561pe - 1./24.*m11*RSp45613e*RS14563pe - 1./72.*
         m11*RSp45613e*RSe45613p - 1./72.*m11*pow(RSp45613e,2) - 1./144.
         *m11*RS3456pee*RSp45613e - 1./144.*m11*RS3456ppe*RSp45613e + 1.
         /144.*m11*RSe45633p*RSp45613e + 1./32.*m11*RSp45633e + 1./144.
         *m11*RSp45633e*RS34561pe + 1./48.*m11*RSp45633e*RS14563pe + 1./
         144.*m11*RSp45633e*RSe45613p + 1./72.*m11*RSp45633e*RSp45613e
          + 1./288.*m11*RSp45633e*RS3456pee + 1./288.*m11*RSp45633e*
         RS3456ppe - 1./288.*m11*RSp45633e*RSe45633p - 1./288.*m11*pow(
         RSp45633e,2) + 1./144.*m11*RS345pe*RSp45613e - 1./288.*m11*
         RS345pe*RSp45633e + 1./144.*m11*RS346pe*RSp45613e - 1./288.*
         m11*RS346pe*RSp45633e + 1./144.*m11*RS356pe*RSp45613e - 1./288.
         *m11*RS356pe*RSp45633e - 1./48.*m11*RS1456pee*RSp45613e + 1./
         96.*m11*RS1456pee*RSp45633e );

      F13p6x23R +=  + Ab(mp2,mu2)*pi16 * (  - 1./48.*m11*RS1456ppe*
         RSp45613e + 1./96.*m11*RS1456ppe*RSp45633e + 1./144.*m11*
         RSe45611p*RSp45613e - 1./288.*m11*RSe45611p*RSp45633e + 1./32.
         *m11*RSp45611e + 1./144.*m11*RSp45611e*RS34561pe + 1./48.*m11*
         RSp45611e*RS14563pe + 1./144.*m11*RSp45611e*RSe45613p + 1./72.
         *m11*RSp45611e*RSp45613e + 1./288.*m11*RSp45611e*RS3456pee + 1.
         /288.*m11*RSp45611e*RS3456ppe - 1./288.*m11*RSp45611e*
         RSe45633p - 1./144.*m11*RSp45611e*RSp45633e - 1./288.*m11*
         RSp45611e*RS345pe - 1./288.*m11*RSp45611e*RS346pe - 1./288.*
         m11*RSp45611e*RS356pe + 1./96.*m11*RSp45611e*RS1456pee + 1./96.
         *m11*RSp45611e*RS1456ppe - 1./288.*m11*RSp45611e*RSe45611p - 1.
         /288.*m11*pow(RSp45611e,2) + 1./48.*m11*RS145pe*RSp45613e - 1./
         96.*m11*RS145pe*RSp45633e - 1./96.*m11*RS145pe*RSp45611e + 1./
         48.*m11*RS146pe*RSp45613e - 1./96.*m11*RS146pe*RSp45633e - 1./
         96.*m11*RS146pe*RSp45611e + 1./48.*m11*RS156pe*RSp45613e - 1./
         96.*m11*RS156pe*RSp45633e );

      F13p6x23R +=  + Ab(mp2,mu2)*pi16 * (  - 1./96.*m11*RS156pe*
         RSp45611e );

      F13p6x23R +=  + pow(Ab(mp2,mu2),2) * ( 1./72.*pow(RSp45613e,2) - 
         1./72.*RSp45633e*RSp45613e + 1./288.*pow(RSp45633e,2) - 1./72.
         *RSp45611e*RSp45613e + 1./144.*RSp45611e*RSp45633e + 1./288.*
         pow(RSp45611e,2) + 1./54.*m66*RNep*RSp456e*RSp453e*RSe453p - 1.
         /54.*m66*RNep*RSp456e*RSe451p*RSp453e - 1./54.*m66*RNep*
         RSp456e*RSp451e*RSe453p + 1./54.*m66*RNep*RSp456e*RSp451e*
         RSe451p + 1./54.*m66*RN3p*RSp456e*RSp453e*RS345pe - 1./54.*m66
         *RN3p*RSp456e*RSp451e*RS345pe - 1./54.*m66*RN1p*RSp456e*
         RS145pe*RSp453e + 1./54.*m66*RN1p*RSp456e*RSp451e*RS145pe + 1./
         54.*m55*RNep*RSp465e*RSp463e*RSe463p - 1./54.*m55*RNep*RSp465e
         *RSe461p*RSp463e - 1./54.*m55*RNep*RSp465e*RSp461e*RSe463p + 1.
         /54.*m55*RNep*RSp465e*RSp461e*RSe461p + 1./54.*m55*RN3p*
         RSp465e*RSp463e*RS346pe - 1./54.*m55*RN3p*RSp465e*RSp461e*
         RS346pe - 1./54.*m55*RN1p*RSp465e*RS146pe*RSp463e + 1./54.*m55
         *RN1p*RSp465e*RSp461e*RS146pe + 1./54.*m44*RNep*RSp564e*
         RSp563e*RSe563p );

      F13p6x23R +=  + pow(Ab(mp2,mu2),2) * (  - 1./54.*m44*RNep*RSp564e
         *RSe561p*RSp563e - 1./54.*m44*RNep*RSp564e*RSp561e*RSe563p + 1.
         /54.*m44*RNep*RSp564e*RSp561e*RSe561p + 1./54.*m44*RN3p*
         RSp564e*RSp563e*RS356pe - 1./54.*m44*RN3p*RSp564e*RSp561e*
         RS356pe - 1./54.*m44*RN1p*RSp564e*RS156pe*RSp563e + 1./54.*m44
         *RN1p*RSp564e*RSp561e*RS156pe + 1./18.*m33*RN3p*RSp45633e*
         RSp45613e - 1./18.*m33*RN3p*pow(RSp45633e,2) + 1./18.*m11*RN1p
         *RSp45611e*RSp45613e - 1./18.*m11*RN1p*pow(RSp45611e,2) );

      F13p6x23R +=  + Ab(mp2,mu2)*Bb(m11,mu2) * (  - 1./72.*RD1456ep*
         RSp45613e + 1./144.*RD1456ep*RSp45633e + 1./144.*RSp45611e*
         RD1456ep - 1./108.*m66*RSp456e*pow(RS145pe,2) - 1./108.*m55*
         RSp465e*pow(RS146pe,2) - 1./108.*m44*RSp564e*pow(RS156pe,2) - 
         1./18.*m11*RSp45611e*RS14563pe - 1./18.*m11*RSp45611e*
         RS1456pee - 1./18.*m11*RSp45611e*RS1456ppe + 1./18.*m11*
         RS145pe*RSp45611e + 1./18.*m11*RS146pe*RSp45611e + 1./18.*m11*
         RS156pe*RSp45611e + 1./18.*m11*RN1p*RD1456ep*RSp45613e - 1./18.
         *m11*RN1p*RSp45611e*RD1456ep );

      F13p6x23R +=  + Ab(mp2,mu2)*Bb(m14,mu2) * ( 1./72.*mp2*RSp45611e
          - 1./36.*mp2*RSp561e + 1./72.*mp2*RSp564e - 1./36.*m44*
         RSp561e - 1./72.*m44*RSp564e - 1./72.*m11*RSp45611e - 1./36.*
         m11*RSp561e );

      F13p6x23R +=  + Ab(mp2,mu2)*Bb(m15,mu2) * ( 1./72.*mp2*RSp45611e
          - 1./36.*mp2*RSp461e + 1./72.*mp2*RSp465e - 1./36.*m55*
         RSp461e - 1./72.*m55*RSp465e - 1./72.*m11*RSp45611e - 1./36.*
         m11*RSp461e );

      F13p6x23R +=  + Ab(mp2,mu2)*Bb(m16,mu2) * ( 1./72.*mp2*RSp45611e
          - 1./36.*mp2*RSp451e + 1./72.*mp2*RSp456e - 1./36.*m66*
         RSp451e - 1./72.*m66*RSp456e - 1./72.*m11*RSp45611e - 1./36.*
         m11*RSp451e );

      F13p6x23R +=  + Ab(mp2,mu2)*Bb(m33,mu2) * (  - 1./72.*RD3456ep*
         RSp45613e + 1./144.*RSp45633e*RD3456ep + 1./144.*RSp45611e*
         RD3456ep - 1./108.*m66*RSp456e*pow(RS345pe,2) - 1./108.*m55*
         RSp465e*pow(RS346pe,2) - 1./108.*m44*RSp564e*pow(RS356pe,2) - 
         1./18.*m33*RSp45633e*RS34561pe - 1./18.*m33*RSp45633e*
         RS3456pee - 1./18.*m33*RSp45633e*RS3456ppe + 1./18.*m33*
         RS345pe*RSp45633e + 1./18.*m33*RS346pe*RSp45633e + 1./18.*m33*
         RS356pe*RSp45633e + 1./18.*m33*RN3p*RD3456ep*RSp45613e - 1./18.
         *m33*RN3p*RSp45633e*RD3456ep );

      F13p6x23R +=  + Ab(mp2,mu2)*Bb(m34,mu2) * ( 1./72.*mp2*RSp45633e
          - 1./36.*mp2*RSp563e + 1./72.*mp2*RSp564e - 1./36.*m44*
         RSp563e - 1./72.*m44*RSp564e - 1./72.*m33*RSp45633e - 1./36.*
         m33*RSp563e );

      F13p6x23R +=  + Ab(mp2,mu2)*Bb(m35,mu2) * ( 1./72.*mp2*RSp45633e
          - 1./36.*mp2*RSp463e + 1./72.*mp2*RSp465e - 1./36.*m55*
         RSp463e - 1./72.*m55*RSp465e - 1./72.*m33*RSp45633e - 1./36.*
         m33*RSp463e );

      F13p6x23R +=  + Ab(mp2,mu2)*Bb(m36,mu2) * ( 1./72.*mp2*RSp45633e
          - 1./36.*mp2*RSp453e + 1./72.*mp2*RSp456e - 1./36.*m66*
         RSp453e - 1./72.*m66*RSp456e - 1./72.*m33*RSp45633e - 1./36.*
         m33*RSp453e );

      F13p6x23R +=  + Ab(mp2,mu2)*Bb(me2,mu2) * (  - 1./108.*m66*
         RSp456e*pow(RSe453p,2) + 1./54.*m66*RSp456e*RSe451p*RSe453p - 
         1./108.*m66*RSp456e*pow(RSe451p,2) - 1./108.*m55*RSp465e*pow(
         RSe463p,2) + 1./54.*m55*RSp465e*RSe461p*RSe463p - 1./108.*m55*
         RSp465e*pow(RSe461p,2) - 1./108.*m44*RSp564e*pow(RSe563p,2) + 
         1./54.*m44*RSp564e*RSe561p*RSe563p - 1./108.*m44*RSp564e*pow(
         RSe561p,2) );

      F13p6x23R +=  + Ab(mp2,mu2)*Bb(mp2,mu2) * (  - 1./108.*m66*
         RSp456e*pow(RSp453e,2) + 1./54.*m66*RSp456e*RSp451e*RSp453e - 
         1./108.*m66*RSp456e*pow(RSp451e,2) - 1./108.*m55*RSp465e*pow(
         RSp463e,2) + 1./54.*m55*RSp465e*RSp461e*RSp463e - 1./108.*m55*
         RSp465e*pow(RSp461e,2) - 1./108.*m44*RSp564e*pow(RSp563e,2) + 
         1./54.*m44*RSp564e*RSp561e*RSp563e - 1./108.*m44*RSp564e*pow(
         RSp561e,2) );

      F13p6x23R +=  + Ab(mp2,mu2)*Cb(m11,mu2) * ( 1./18.*m11*RSp45611e*
         RD1456ep );

      F13p6x23R +=  + Ab(mp2,mu2)*Cb(m33,mu2) * ( 1./18.*m33*RSp45633e*
         RD3456ep );

      F13p6x23R +=  + Bb(m11,mu2)*pi16 * (  - 1./144.*RD1456ep*RD3456ep
          - 1./144.*pow(RD1456ep,2) + 1./72.*mp2*RD1456ep*RSp45613e - 1.
         /144.*mp2*RD1456ep*RSp45633e - 1./144.*mp2*RSp45611e*RD1456ep
          + 1./72.*me2*RD1456ep*RSe45613p - 1./144.*me2*RD1456ep*
         RSe45633p - 1./144.*me2*RSe45611p*RD1456ep + 1./48.*m66*
         RD1456ep + 1./48.*m55*RD1456ep + 1./48.*m44*RD1456ep + 1./32.*
         m33*RD1456ep + 1./48.*m33*RD1456ep*RS34561pe + 1./144.*m33*
         RD1456ep*RS14563pe + 1./144.*m33*RD1456ep*RSe45613p + 1./144.*
         m33*RD1456ep*RSp45613e + 1./96.*m33*RD1456ep*RS3456pee + 1./96.
         *m33*RD1456ep*RS3456ppe - 1./288.*m33*RD1456ep*RSe45633p - 1./
         288.*m33*RD1456ep*RSp45633e - 1./96.*m33*RD1456ep*RS345pe - 1./
         96.*m33*RD1456ep*RS346pe - 1./96.*m33*RD1456ep*RS356pe + 1./
         288.*m33*RS1456pee*RD1456ep + 1./288.*m33*RS1456ppe*RD1456ep
          - 1./288.*m33*RSe45611p*RD1456ep - 1./288.*m33*RSp45611e*
         RD1456ep - 1./288.*m33*RS145pe*RD1456ep - 1./288.*m33*RS146pe*
         RD1456ep );

      F13p6x23R +=  + Bb(m11,mu2)*pi16 * (  - 1./288.*m33*RS156pe*
         RD1456ep + 1./96.*m11*RD1456ep + 1./144.*m11*RD1456ep*
         RS34561pe + 1./48.*m11*RD1456ep*RS14563pe + 1./144.*m11*
         RD1456ep*RSe45613p + 1./144.*m11*RD1456ep*RSp45613e + 1./288.*
         m11*RD1456ep*RS3456pee + 1./288.*m11*RD1456ep*RS3456ppe - 1./
         288.*m11*RD1456ep*RSe45633p - 1./288.*m11*RD1456ep*RSp45633e
          - 1./288.*m11*RD1456ep*RS345pe - 1./288.*m11*RD1456ep*RS346pe
          - 1./288.*m11*RD1456ep*RS356pe + 1./96.*m11*RS1456pee*
         RD1456ep + 1./96.*m11*RS1456ppe*RD1456ep - 1./288.*m11*
         RSe45611p*RD1456ep - 1./288.*m11*RSp45611e*RD1456ep - 1./96.*
         m11*RS145pe*RD1456ep - 1./96.*m11*RS146pe*RD1456ep - 1./96.*
         m11*RS156pe*RD1456ep );

      F13p6x23R +=  + pow(Bb(m11,mu2),2) * ( 1./288.*pow(RD1456ep,2) - 
         1./18.*m11*RD1456ep*RS14563pe - 1./18.*m11*RS1456pee*RD1456ep
          - 1./18.*m11*RS1456ppe*RD1456ep + 1./18.*m11*RS145pe*RD1456ep
          + 1./18.*m11*RS146pe*RD1456ep + 1./18.*m11*RS156pe*RD1456ep )
         ;

      F13p6x23R +=  + Bb(m11,mu2)*Bb(m33,mu2) * ( 1./144.*RD1456ep*
         RD3456ep );

      F13p6x23R +=  + Bb(m11,mu2)*Cb(m11,mu2) * ( 1./18.*m11*pow(
         RD1456ep,2) );

      F13p6x23R +=  + Bb(m33,mu2)*pi16 * (  - 1./144.*pow(RD3456ep,2)
          - 1./144.*RD1456ep*RD3456ep + 1./72.*mp2*RD3456ep*RSp45613e
          - 1./144.*mp2*RSp45633e*RD3456ep - 1./144.*mp2*RSp45611e*
         RD3456ep + 1./72.*me2*RD3456ep*RSe45613p - 1./144.*me2*
         RSe45633p*RD3456ep - 1./144.*me2*RSe45611p*RD3456ep + 1./48.*
         m66*RD3456ep + 1./48.*m55*RD3456ep + 1./48.*m44*RD3456ep + 1./
         96.*m33*RD3456ep + 1./48.*m33*RD3456ep*RS34561pe + 1./144.*m33
         *RD3456ep*RS14563pe + 1./144.*m33*RD3456ep*RSe45613p + 1./144.
         *m33*RD3456ep*RSp45613e + 1./96.*m33*RS3456pee*RD3456ep + 1./
         96.*m33*RS3456ppe*RD3456ep - 1./288.*m33*RSe45633p*RD3456ep - 
         1./288.*m33*RSp45633e*RD3456ep - 1./96.*m33*RS345pe*RD3456ep
          - 1./96.*m33*RS346pe*RD3456ep - 1./96.*m33*RS356pe*RD3456ep
          + 1./288.*m33*RS1456pee*RD3456ep + 1./288.*m33*RS1456ppe*
         RD3456ep - 1./288.*m33*RSe45611p*RD3456ep - 1./288.*m33*
         RSp45611e*RD3456ep - 1./288.*m33*RS145pe*RD3456ep - 1./288.*
         m33*RS146pe*RD3456ep );

      F13p6x23R +=  + Bb(m33,mu2)*pi16 * (  - 1./288.*m33*RS156pe*
         RD3456ep + 1./32.*m11*RD3456ep + 1./144.*m11*RD3456ep*
         RS34561pe + 1./48.*m11*RD3456ep*RS14563pe + 1./144.*m11*
         RD3456ep*RSe45613p + 1./144.*m11*RD3456ep*RSp45613e + 1./288.*
         m11*RS3456pee*RD3456ep + 1./288.*m11*RS3456ppe*RD3456ep - 1./
         288.*m11*RSe45633p*RD3456ep - 1./288.*m11*RSp45633e*RD3456ep
          - 1./288.*m11*RS345pe*RD3456ep - 1./288.*m11*RS346pe*RD3456ep
          - 1./288.*m11*RS356pe*RD3456ep + 1./96.*m11*RS1456pee*
         RD3456ep + 1./96.*m11*RS1456ppe*RD3456ep - 1./288.*m11*
         RSe45611p*RD3456ep - 1./288.*m11*RSp45611e*RD3456ep - 1./96.*
         m11*RS145pe*RD3456ep - 1./96.*m11*RS146pe*RD3456ep - 1./96.*
         m11*RS156pe*RD3456ep );

      F13p6x23R +=  + pow(Bb(m33,mu2),2) * ( 1./288.*pow(RD3456ep,2) - 
         1./18.*m33*RD3456ep*RS34561pe - 1./18.*m33*RS3456pee*RD3456ep
          - 1./18.*m33*RS3456ppe*RD3456ep + 1./18.*m33*RS345pe*RD3456ep
          + 1./18.*m33*RS346pe*RD3456ep + 1./18.*m33*RS356pe*RD3456ep )
         ;

      F13p6x23R +=  + Bb(m33,mu2)*Cb(m33,mu2) * ( 1./18.*m33*pow(
         RD3456ep,2) );

      F13p6x23R +=  + Abeps(m11,mu2)*pi16 * (  - 1./72.*RD3456ep*
         RS14563pe - 1./12.*RD1456ep - 1./72.*RD1456ep*RS14563pe - 1./
         144.*RS1456pee*RD3456ep - 1./144.*RS1456pee*RD1456ep - 1./144.
         *RS1456ppe*RD3456ep - 1./144.*RS1456ppe*RD1456ep + 1./144.*
         RS145pe*RD3456ep + 1./144.*RS145pe*RD1456ep + 1./144.*RS146pe*
         RD3456ep + 1./144.*RS146pe*RD1456ep + 1./144.*RS156pe*RD3456ep
          + 1./144.*RS156pe*RD1456ep + 1./36.*mp2*RSp45613e*RS14563pe
          - 1./72.*mp2*RSp45633e*RS14563pe + 1./72.*mp2*RS1456pee*
         RSp45613e - 1./144.*mp2*RS1456pee*RSp45633e + 1./72.*mp2*
         RS1456ppe*RSp45613e - 1./144.*mp2*RS1456ppe*RSp45633e - 1./72.
         *mp2*RSp45611e*RS14563pe - 1./144.*mp2*RSp45611e*RS1456pee - 1.
         /144.*mp2*RSp45611e*RS1456ppe - 1./72.*mp2*RS145pe*RSp45613e
          + 1./144.*mp2*RS145pe*RSp45633e + 1./144.*mp2*RS145pe*
         RSp45611e - 1./72.*mp2*RS146pe*RSp45613e + 1./144.*mp2*RS146pe
         *RSp45633e + 1./144.*mp2*RS146pe*RSp45611e - 1./72.*mp2*
         RS156pe*RSp45613e );

      F13p6x23R +=  + Abeps(m11,mu2)*pi16 * ( 1./144.*mp2*RS156pe*
         RSp45633e + 1./144.*mp2*RS156pe*RSp45611e + 1./36.*me2*
         RSe45613p*RS14563pe - 1./72.*me2*RSe45633p*RS14563pe + 1./72.*
         me2*RS1456pee*RSe45613p - 1./144.*me2*RS1456pee*RSe45633p + 1./
         72.*me2*RS1456ppe*RSe45613p - 1./144.*me2*RS1456ppe*RSe45633p
          - 1./72.*me2*RSe45611p*RS14563pe - 1./144.*me2*RSe45611p*
         RS1456pee - 1./144.*me2*RSe45611p*RS1456ppe - 1./72.*me2*
         RS145pe*RSe45613p + 1./144.*me2*RS145pe*RSe45633p + 1./144.*
         me2*RS145pe*RSe45611p - 1./72.*me2*RS146pe*RSe45613p + 1./144.
         *me2*RS146pe*RSe45633p + 1./144.*me2*RS146pe*RSe45611p - 1./72.
         *me2*RS156pe*RSe45613p + 1./144.*me2*RS156pe*RSe45633p + 1./
         144.*me2*RS156pe*RSe45611p + 1./24.*m66*RS14563pe + 1./18.*m66
         *RSe45613p + 1./18.*m66*RSp45613e + 1./36.*m66*RSe453p + 1./36.
         *m66*RSp453e + 7./144.*m66*RS1456pee + 7./144.*m66*RS1456ppe
          - 1./18.*m66*RSe45611p - 1./18.*m66*RSp45611e - 5./48.*m66*
         RS145pe );

      F13p6x23R +=  + Abeps(m11,mu2)*pi16 * (  - 1./108.*m66*RS145pe*
         RSe563p - 1./108.*m66*RS145pe*RSe463p - 5./54.*m66*RS145pe*
         RSe453p - 1./108.*m66*RS145pe*RSp563e - 1./108.*m66*RS145pe*
         RSp463e - 5./54.*m66*RS145pe*RSp453e + 5./108.*m66*pow(
         RS145pe,2) - 7./144.*m66*RS146pe - 1./54.*m66*RS146pe*RSe463p
          - 1./108.*m66*RS146pe*RSe453p - 1./54.*m66*RS146pe*RSp463e - 
         1./108.*m66*RS146pe*RSp453e + 1./108.*m66*RS146pe*RS145pe + 1./
         108.*m66*pow(RS146pe,2) - 7./144.*m66*RS156pe - 1./54.*m66*
         RS156pe*RSe563p - 1./108.*m66*RS156pe*RSe453p - 1./54.*m66*
         RS156pe*RSp563e - 1./108.*m66*RS156pe*RSp453e + 1./108.*m66*
         RS156pe*RS145pe + 1./108.*m66*pow(RS156pe,2) + 1./108.*m66*
         RSe561p*RS145pe + 1./54.*m66*RSe561p*RS156pe + 1./108.*m66*
         RSe461p*RS145pe + 1./54.*m66*RSe461p*RS146pe - 1./36.*m66*
         RSe451p + 5./54.*m66*RSe451p*RS145pe + 1./108.*m66*RSe451p*
         RS146pe + 1./108.*m66*RSe451p*RS156pe + 1./108.*m66*RSp561e*
         RS145pe );

      F13p6x23R +=  + Abeps(m11,mu2)*pi16 * ( 1./54.*m66*RSp561e*
         RS156pe + 1./108.*m66*RSp461e*RS145pe + 1./54.*m66*RSp461e*
         RS146pe - 1./36.*m66*RSp451e + 5./54.*m66*RSp451e*RS145pe + 1./
         108.*m66*RSp451e*RS146pe + 1./108.*m66*RSp451e*RS156pe + 1./24.
         *m55*RS14563pe + 1./18.*m55*RSe45613p + 1./18.*m55*RSp45613e
          + 1./36.*m55*RSe463p + 1./36.*m55*RSp463e + 7./144.*m55*
         RS1456pee + 7./144.*m55*RS1456ppe - 1./18.*m55*RSe45611p - 1./
         18.*m55*RSp45611e - 7./144.*m55*RS145pe - 1./108.*m55*RS145pe*
         RSe463p - 1./54.*m55*RS145pe*RSe453p - 1./108.*m55*RS145pe*
         RSp463e - 1./54.*m55*RS145pe*RSp453e + 1./108.*m55*pow(
         RS145pe,2) - 5./48.*m55*RS146pe - 1./108.*m55*RS146pe*RSe563p
          - 5./54.*m55*RS146pe*RSe463p - 1./108.*m55*RS146pe*RSe453p - 
         1./108.*m55*RS146pe*RSp563e - 5./54.*m55*RS146pe*RSp463e - 1./
         108.*m55*RS146pe*RSp453e + 1./108.*m55*RS146pe*RS145pe + 5./
         108.*m55*pow(RS146pe,2) - 7./144.*m55*RS156pe - 1./54.*m55*
         RS156pe*RSe563p );

      F13p6x23R +=  + Abeps(m11,mu2)*pi16 * (  - 1./108.*m55*RS156pe*
         RSe463p - 1./54.*m55*RS156pe*RSp563e - 1./108.*m55*RS156pe*
         RSp463e + 1./108.*m55*RS156pe*RS146pe + 1./108.*m55*pow(
         RS156pe,2) + 1./108.*m55*RSe561p*RS146pe + 1./54.*m55*RSe561p*
         RS156pe - 1./36.*m55*RSe461p + 1./108.*m55*RSe461p*RS145pe + 5.
         /54.*m55*RSe461p*RS146pe + 1./108.*m55*RSe461p*RS156pe + 1./54.
         *m55*RSe451p*RS145pe + 1./108.*m55*RSe451p*RS146pe + 1./108.*
         m55*RSp561e*RS146pe + 1./54.*m55*RSp561e*RS156pe - 1./36.*m55*
         RSp461e + 1./108.*m55*RSp461e*RS145pe + 5./54.*m55*RSp461e*
         RS146pe + 1./108.*m55*RSp461e*RS156pe + 1./54.*m55*RSp451e*
         RS145pe + 1./108.*m55*RSp451e*RS146pe + 1./24.*m44*RS14563pe
          + 1./18.*m44*RSe45613p + 1./18.*m44*RSp45613e + 1./36.*m44*
         RSe563p + 1./36.*m44*RSp563e + 7./144.*m44*RS1456pee + 7./144.
         *m44*RS1456ppe - 1./18.*m44*RSe45611p - 1./18.*m44*RSp45611e
          - 7./144.*m44*RS145pe - 1./108.*m44*RS145pe*RSe563p - 1./54.*
         m44*RS145pe*RSe453p );

      F13p6x23R +=  + Abeps(m11,mu2)*pi16 * (  - 1./108.*m44*RS145pe*
         RSp563e - 1./54.*m44*RS145pe*RSp453e + 1./108.*m44*pow(
         RS145pe,2) - 7./144.*m44*RS146pe - 1./108.*m44*RS146pe*RSe563p
          - 1./54.*m44*RS146pe*RSe463p - 1./108.*m44*RS146pe*RSp563e - 
         1./54.*m44*RS146pe*RSp463e + 1./108.*m44*pow(RS146pe,2) - 5./
         48.*m44*RS156pe - 5./54.*m44*RS156pe*RSe563p - 1./108.*m44*
         RS156pe*RSe463p - 1./108.*m44*RS156pe*RSe453p - 5./54.*m44*
         RS156pe*RSp563e - 1./108.*m44*RS156pe*RSp463e - 1./108.*m44*
         RS156pe*RSp453e + 1./108.*m44*RS156pe*RS145pe + 1./108.*m44*
         RS156pe*RS146pe + 5./108.*m44*pow(RS156pe,2) - 1./36.*m44*
         RSe561p + 1./108.*m44*RSe561p*RS145pe + 1./108.*m44*RSe561p*
         RS146pe + 5./54.*m44*RSe561p*RS156pe + 1./54.*m44*RSe461p*
         RS146pe + 1./108.*m44*RSe461p*RS156pe + 1./54.*m44*RSe451p*
         RS145pe + 1./108.*m44*RSe451p*RS156pe - 1./36.*m44*RSp561e + 1.
         /108.*m44*RSp561e*RS145pe + 1./108.*m44*RSp561e*RS146pe + 5./
         54.*m44*RSp561e*RS156pe );

      F13p6x23R +=  + Abeps(m11,mu2)*pi16 * ( 1./54.*m44*RSp461e*
         RS146pe + 1./108.*m44*RSp461e*RS156pe + 1./54.*m44*RSp451e*
         RS145pe + 1./108.*m44*RSp451e*RS156pe - 1./48.*m33 - 1./48.*
         m33*RS14563pe + 5./162.*m33*RS14563pe*RS34561pe + 1./324.*m33*
         pow(RS14563pe,2) + 1./324.*m33*RSe45613p*RS14563pe + 1./324.*
         m33*RSp45613e*RS14563pe + 5./324.*m33*RS3456pee*RS14563pe + 5./
         324.*m33*RS3456ppe*RS14563pe - 1./648.*m33*RSe45633p*RS14563pe
          - 1./648.*m33*RSp45633e*RS14563pe - 5./324.*m33*RS345pe*
         RS14563pe - 5./324.*m33*RS346pe*RS14563pe - 5./324.*m33*
         RS356pe*RS14563pe + 5./324.*m33*RS1456pee*RS34561pe + 1./324.*
         m33*RS1456pee*RS14563pe + 1./648.*m33*RS1456pee*RSe45613p + 1./
         648.*m33*RS1456pee*RSp45613e + 5./648.*m33*RS1456pee*RS3456pee
          + 5./648.*m33*RS1456pee*RS3456ppe - 1./1296.*m33*RS1456pee*
         RSe45633p - 1./1296.*m33*RS1456pee*RSp45633e - 5./648.*m33*
         RS1456pee*RS345pe - 5./648.*m33*RS1456pee*RS346pe - 5./648.*
         m33*RS1456pee*RS356pe );

      F13p6x23R +=  + Abeps(m11,mu2)*pi16 * ( 1./1296.*m33*pow(
         RS1456pee,2) + 5./324.*m33*RS1456ppe*RS34561pe + 1./324.*m33*
         RS1456ppe*RS14563pe + 1./648.*m33*RS1456ppe*RSe45613p + 1./648.
         *m33*RS1456ppe*RSp45613e + 5./648.*m33*RS1456ppe*RS3456pee + 5.
         /648.*m33*RS1456ppe*RS3456ppe - 1./1296.*m33*RS1456ppe*
         RSe45633p - 1./1296.*m33*RS1456ppe*RSp45633e - 5./648.*m33*
         RS1456ppe*RS345pe - 5./648.*m33*RS1456ppe*RS346pe - 5./648.*
         m33*RS1456ppe*RS356pe + 1./648.*m33*RS1456ppe*RS1456pee + 1./
         1296.*m33*pow(RS1456ppe,2) - 1./648.*m33*RSe45611p*RS14563pe
          - 1./1296.*m33*RSe45611p*RS1456pee - 1./1296.*m33*RSe45611p*
         RS1456ppe - 1./648.*m33*RSp45611e*RS14563pe - 1./1296.*m33*
         RSp45611e*RS1456pee - 1./1296.*m33*RSp45611e*RS1456ppe + 1./
         144.*m33*RS145pe - 5./324.*m33*RS145pe*RS34561pe - 1./324.*m33
         *RS145pe*RS14563pe - 1./648.*m33*RS145pe*RSe45613p - 1./648.*
         m33*RS145pe*RSp45613e - 5./648.*m33*RS145pe*RS3456pee - 5./648.
         *m33*RS145pe*RS3456ppe );

      F13p6x23R +=  + Abeps(m11,mu2)*pi16 * ( 1./1296.*m33*RS145pe*
         RSe45633p + 1./1296.*m33*RS145pe*RSp45633e + 5./648.*m33*
         RS145pe*RS345pe + 5./648.*m33*RS145pe*RS346pe + 5./648.*m33*
         RS145pe*RS356pe - 1./648.*m33*RS145pe*RS1456pee - 1./648.*m33*
         RS145pe*RS1456ppe + 1./1296.*m33*RS145pe*RSe45611p + 1./1296.*
         m33*RS145pe*RSp45611e + 1./1296.*m33*pow(RS145pe,2) + 1./144.*
         m33*RS146pe - 5./324.*m33*RS146pe*RS34561pe - 1./324.*m33*
         RS146pe*RS14563pe - 1./648.*m33*RS146pe*RSe45613p - 1./648.*
         m33*RS146pe*RSp45613e - 5./648.*m33*RS146pe*RS3456pee - 5./648.
         *m33*RS146pe*RS3456ppe + 1./1296.*m33*RS146pe*RSe45633p + 1./
         1296.*m33*RS146pe*RSp45633e + 5./648.*m33*RS146pe*RS345pe + 5./
         648.*m33*RS146pe*RS346pe + 5./648.*m33*RS146pe*RS356pe - 1./
         648.*m33*RS146pe*RS1456pee - 1./648.*m33*RS146pe*RS1456ppe + 1.
         /1296.*m33*RS146pe*RSe45611p + 1./1296.*m33*RS146pe*RSp45611e
          + 1./648.*m33*RS146pe*RS145pe + 1./1296.*m33*pow(RS146pe,2)
          + 1./144.*m33*RS156pe );

      F13p6x23R +=  + Abeps(m11,mu2)*pi16 * (  - 5./324.*m33*RS156pe*
         RS34561pe - 1./324.*m33*RS156pe*RS14563pe - 1./648.*m33*
         RS156pe*RSe45613p - 1./648.*m33*RS156pe*RSp45613e - 5./648.*
         m33*RS156pe*RS3456pee - 5./648.*m33*RS156pe*RS3456ppe + 1./
         1296.*m33*RS156pe*RSe45633p + 1./1296.*m33*RS156pe*RSp45633e
          + 5./648.*m33*RS156pe*RS345pe + 5./648.*m33*RS156pe*RS346pe
          + 5./648.*m33*RS156pe*RS356pe - 1./648.*m33*RS156pe*RS1456pee
          - 1./648.*m33*RS156pe*RS1456ppe + 1./1296.*m33*RS156pe*
         RSe45611p + 1./1296.*m33*RS156pe*RSp45611e + 1./648.*m33*
         RS156pe*RS145pe + 1./648.*m33*RS156pe*RS146pe + 1./1296.*m33*
         pow(RS156pe,2) + 17./48.*m11 + 1./48.*m11*RS14563pe + 1./324.*
         m11*RS14563pe*RS34561pe + 5./162.*m11*pow(RS14563pe,2) + 1./6.
         *m11*RSe45613p + 1./324.*m11*RSe45613p*RS14563pe + 1./6.*m11*
         RSp45613e + 1./324.*m11*RSp45613e*RS14563pe + 1./648.*m11*
         RS3456pee*RS14563pe + 1./648.*m11*RS3456ppe*RS14563pe - 1./648.
         *m11*RSe45633p*RS14563pe );

      F13p6x23R +=  + Abeps(m11,mu2)*pi16 * (  - 1./648.*m11*RSp45633e*
         RS14563pe - 1./648.*m11*RS345pe*RS14563pe - 1./648.*m11*
         RS346pe*RS14563pe - 1./648.*m11*RS356pe*RS14563pe + 1./36.*m11
         *RSe563p + 1./36.*m11*RSe463p + 1./36.*m11*RSe453p + 1./36.*
         m11*RSp563e + 1./36.*m11*RSp463e + 1./36.*m11*RSp453e + 7./48.
         *m11*RS1456pee + 37./648.*m11*RS1456pee*RS34561pe + 7./81.*m11
         *RS1456pee*RS14563pe + 37./648.*m11*RS1456pee*RSe45613p + 37./
         648.*m11*RS1456pee*RSp45613e + 1./1296.*m11*RS1456pee*
         RS3456pee + 1./1296.*m11*RS1456pee*RS3456ppe - 1./1296.*m11*
         RS1456pee*RSe45633p - 1./1296.*m11*RS1456pee*RSp45633e - 1./
         1296.*m11*RS1456pee*RS345pe - 1./1296.*m11*RS1456pee*RS346pe
          - 1./1296.*m11*RS1456pee*RS356pe + 41./648.*m11*pow(
         RS1456pee,2) + 7./48.*m11*RS1456ppe + 37./648.*m11*RS1456ppe*
         RS34561pe + 7./81.*m11*RS1456ppe*RS14563pe + 37./648.*m11*
         RS1456ppe*RSe45613p + 37./648.*m11*RS1456ppe*RSp45613e + 1./
         1296.*m11*RS1456ppe*RS3456pee );

      F13p6x23R +=  + Abeps(m11,mu2)*pi16 * ( 1./1296.*m11*RS1456ppe*
         RS3456ppe - 1./1296.*m11*RS1456ppe*RSe45633p - 1./1296.*m11*
         RS1456ppe*RSp45633e - 1./1296.*m11*RS1456ppe*RS345pe - 1./1296.
         *m11*RS1456ppe*RS346pe - 1./1296.*m11*RS1456ppe*RS356pe + 41./
         324.*m11*RS1456ppe*RS1456pee + 41./648.*m11*pow(RS1456ppe,2)
          - 1./6.*m11*RSe45611p - 1./648.*m11*RSe45611p*RS14563pe - 73./
         1296.*m11*RSe45611p*RS1456pee - 73./1296.*m11*RSe45611p*
         RS1456ppe - 1./6.*m11*RSp45611e - 1./648.*m11*RSp45611e*
         RS14563pe - 73./1296.*m11*RSp45611e*RS1456pee - 73./1296.*m11*
         RSp45611e*RS1456ppe - 17./72.*m11*RS145pe - 37./648.*m11*
         RS145pe*RS34561pe - 7./81.*m11*RS145pe*RS14563pe - 37./648.*
         m11*RS145pe*RSe45613p - 37./648.*m11*RS145pe*RSp45613e - 1./
         1296.*m11*RS145pe*RS3456pee - 1./1296.*m11*RS145pe*RS3456ppe
          + 1./1296.*m11*RS145pe*RSe45633p + 1./1296.*m11*RS145pe*
         RSp45633e + 1./1296.*m11*RS145pe*RS345pe + 1./1296.*m11*
         RS145pe*RS346pe );

      F13p6x23R +=  + Abeps(m11,mu2)*pi16 * ( 1./1296.*m11*RS145pe*
         RS356pe - 41./324.*m11*RS145pe*RS1456pee - 41./324.*m11*
         RS145pe*RS1456ppe + 73./1296.*m11*RS145pe*RSe45611p + 73./1296.
         *m11*RS145pe*RSp45611e + 41./648.*m11*pow(RS145pe,2) - 17./72.
         *m11*RS146pe - 37./648.*m11*RS146pe*RS34561pe - 7./81.*m11*
         RS146pe*RS14563pe - 37./648.*m11*RS146pe*RSe45613p - 37./648.*
         m11*RS146pe*RSp45613e - 1./1296.*m11*RS146pe*RS3456pee - 1./
         1296.*m11*RS146pe*RS3456ppe + 1./1296.*m11*RS146pe*RSe45633p
          + 1./1296.*m11*RS146pe*RSp45633e + 1./1296.*m11*RS146pe*
         RS345pe + 1./1296.*m11*RS146pe*RS346pe + 1./1296.*m11*RS146pe*
         RS356pe - 41./324.*m11*RS146pe*RS1456pee - 41./324.*m11*
         RS146pe*RS1456ppe + 73./1296.*m11*RS146pe*RSe45611p + 73./1296.
         *m11*RS146pe*RSp45611e + 41./324.*m11*RS146pe*RS145pe + 41./
         648.*m11*pow(RS146pe,2) - 17./72.*m11*RS156pe - 37./648.*m11*
         RS156pe*RS34561pe - 7./81.*m11*RS156pe*RS14563pe - 37./648.*
         m11*RS156pe*RSe45613p );

      F13p6x23R +=  + Abeps(m11,mu2)*pi16 * (  - 37./648.*m11*RS156pe*
         RSp45613e - 1./1296.*m11*RS156pe*RS3456pee - 1./1296.*m11*
         RS156pe*RS3456ppe + 1./1296.*m11*RS156pe*RSe45633p + 1./1296.*
         m11*RS156pe*RSp45633e + 1./1296.*m11*RS156pe*RS345pe + 1./1296.
         *m11*RS156pe*RS346pe + 1./1296.*m11*RS156pe*RS356pe - 41./324.
         *m11*RS156pe*RS1456pee - 41./324.*m11*RS156pe*RS1456ppe + 73./
         1296.*m11*RS156pe*RSe45611p + 73./1296.*m11*RS156pe*RSp45611e
          + 41./324.*m11*RS156pe*RS145pe + 41./324.*m11*RS156pe*RS146pe
          + 41./648.*m11*pow(RS156pe,2) - 1./36.*m11*RSe561p - 1./36.*
         m11*RSe461p - 1./36.*m11*RSe451p - 1./36.*m11*RSp561e - 1./36.
         *m11*RSp461e - 1./36.*m11*RSp451e );

      F13p6x23R +=  + Abeps(m13,mu2)*pi16 * (  - 1./324.*RD3456ep*
         RS34561pe - 1./324.*RD3456ep*RS14563pe - 1./324.*RD3456ep*
         RSe45613p - 1./324.*RD3456ep*RSp45613e - 1./648.*RS3456pee*
         RD3456ep - 1./648.*RS3456ppe*RD3456ep + 1./648.*RSe45633p*
         RD3456ep + 1./648.*RSp45633e*RD3456ep + 1./648.*RS345pe*
         RD3456ep + 1./648.*RS346pe*RD3456ep + 1./648.*RS356pe*RD3456ep
          - 1./324.*RD1456ep*RS34561pe - 1./324.*RD1456ep*RS14563pe - 1.
         /324.*RD1456ep*RSe45613p - 1./324.*RD1456ep*RSp45613e - 1./648.
         *RD1456ep*RS3456pee - 1./648.*RD1456ep*RS3456ppe + 1./648.*
         RD1456ep*RSe45633p + 1./648.*RD1456ep*RSp45633e + 1./648.*
         RD1456ep*RS345pe + 1./648.*RD1456ep*RS346pe + 1./648.*RD1456ep
         *RS356pe - 1./648.*RS1456pee*RD3456ep - 1./648.*RS1456pee*
         RD1456ep - 1./648.*RS1456ppe*RD3456ep - 1./648.*RS1456ppe*
         RD1456ep + 1./648.*RSe45611p*RD3456ep + 1./648.*RSe45611p*
         RD1456ep + 1./648.*RSp45611e*RD3456ep + 1./648.*RSp45611e*
         RD1456ep );

      F13p6x23R +=  + Abeps(m13,mu2)*pi16 * ( 1./648.*RS145pe*RD3456ep
          + 1./648.*RS145pe*RD1456ep + 1./648.*RS146pe*RD3456ep + 1./
         648.*RS146pe*RD1456ep + 1./648.*RS156pe*RD3456ep + 1./648.*
         RS156pe*RD1456ep + 1./162.*mp2*RSp45613e*RS34561pe + 1./162.*
         mp2*RSp45613e*RS14563pe + 1./162.*mp2*RSp45613e*RSe45613p + 1./
         162.*mp2*pow(RSp45613e,2) + 1./324.*mp2*RS3456pee*RSp45613e + 
         1./324.*mp2*RS3456ppe*RSp45613e - 1./324.*mp2*RSe45633p*
         RSp45613e - 1./324.*mp2*RSp45633e*RS34561pe - 1./324.*mp2*
         RSp45633e*RS14563pe - 1./324.*mp2*RSp45633e*RSe45613p - 1./162.
         *mp2*RSp45633e*RSp45613e - 1./648.*mp2*RSp45633e*RS3456pee - 1.
         /648.*mp2*RSp45633e*RS3456ppe + 1./648.*mp2*RSp45633e*
         RSe45633p + 1./648.*mp2*pow(RSp45633e,2) - 1./324.*mp2*RS345pe
         *RSp45613e + 1./648.*mp2*RS345pe*RSp45633e - 1./324.*mp2*
         RS346pe*RSp45613e + 1./648.*mp2*RS346pe*RSp45633e - 1./324.*
         mp2*RS356pe*RSp45613e + 1./648.*mp2*RS356pe*RSp45633e + 1./324.
         *mp2*RS1456pee*RSp45613e );

      F13p6x23R +=  + Abeps(m13,mu2)*pi16 * (  - 1./648.*mp2*RS1456pee*
         RSp45633e + 1./324.*mp2*RS1456ppe*RSp45613e - 1./648.*mp2*
         RS1456ppe*RSp45633e - 1./324.*mp2*RSe45611p*RSp45613e + 1./648.
         *mp2*RSe45611p*RSp45633e - 1./324.*mp2*RSp45611e*RS34561pe - 1.
         /324.*mp2*RSp45611e*RS14563pe - 1./324.*mp2*RSp45611e*
         RSe45613p - 1./162.*mp2*RSp45611e*RSp45613e - 1./648.*mp2*
         RSp45611e*RS3456pee - 1./648.*mp2*RSp45611e*RS3456ppe + 1./648.
         *mp2*RSp45611e*RSe45633p + 1./324.*mp2*RSp45611e*RSp45633e + 1.
         /648.*mp2*RSp45611e*RS345pe + 1./648.*mp2*RSp45611e*RS346pe + 
         1./648.*mp2*RSp45611e*RS356pe - 1./648.*mp2*RSp45611e*
         RS1456pee - 1./648.*mp2*RSp45611e*RS1456ppe + 1./648.*mp2*
         RSp45611e*RSe45611p + 1./648.*mp2*pow(RSp45611e,2) - 1./324.*
         mp2*RS145pe*RSp45613e + 1./648.*mp2*RS145pe*RSp45633e + 1./648.
         *mp2*RS145pe*RSp45611e - 1./324.*mp2*RS146pe*RSp45613e + 1./
         648.*mp2*RS146pe*RSp45633e + 1./648.*mp2*RS146pe*RSp45611e - 1.
         /324.*mp2*RS156pe*RSp45613e );

      F13p6x23R +=  + Abeps(m13,mu2)*pi16 * ( 1./648.*mp2*RS156pe*
         RSp45633e + 1./648.*mp2*RS156pe*RSp45611e + 1./162.*me2*
         RSe45613p*RS34561pe + 1./162.*me2*RSe45613p*RS14563pe + 1./162.
         *me2*pow(RSe45613p,2) + 1./162.*me2*RSp45613e*RSe45613p + 1./
         324.*me2*RS3456pee*RSe45613p + 1./324.*me2*RS3456ppe*RSe45613p
          - 1./324.*me2*RSe45633p*RS34561pe - 1./324.*me2*RSe45633p*
         RS14563pe - 1./162.*me2*RSe45633p*RSe45613p - 1./324.*me2*
         RSe45633p*RSp45613e - 1./648.*me2*RSe45633p*RS3456pee - 1./648.
         *me2*RSe45633p*RS3456ppe + 1./648.*me2*pow(RSe45633p,2) - 1./
         324.*me2*RSp45633e*RSe45613p + 1./648.*me2*RSp45633e*RSe45633p
          - 1./324.*me2*RS345pe*RSe45613p + 1./648.*me2*RS345pe*
         RSe45633p - 1./324.*me2*RS346pe*RSe45613p + 1./648.*me2*
         RS346pe*RSe45633p - 1./324.*me2*RS356pe*RSe45613p + 1./648.*
         me2*RS356pe*RSe45633p + 1./324.*me2*RS1456pee*RSe45613p - 1./
         648.*me2*RS1456pee*RSe45633p + 1./324.*me2*RS1456ppe*RSe45613p
          - 1./648.*me2*RS1456ppe*RSe45633p );

      F13p6x23R +=  + Abeps(m13,mu2)*pi16 * (  - 1./324.*me2*RSe45611p*
         RS34561pe - 1./324.*me2*RSe45611p*RS14563pe - 1./162.*me2*
         RSe45611p*RSe45613p - 1./324.*me2*RSe45611p*RSp45613e - 1./648.
         *me2*RSe45611p*RS3456pee - 1./648.*me2*RSe45611p*RS3456ppe + 1.
         /324.*me2*RSe45611p*RSe45633p + 1./648.*me2*RSe45611p*
         RSp45633e + 1./648.*me2*RSe45611p*RS345pe + 1./648.*me2*
         RSe45611p*RS346pe + 1./648.*me2*RSe45611p*RS356pe - 1./648.*
         me2*RSe45611p*RS1456pee - 1./648.*me2*RSe45611p*RS1456ppe + 1./
         648.*me2*pow(RSe45611p,2) - 1./324.*me2*RSp45611e*RSe45613p + 
         1./648.*me2*RSp45611e*RSe45633p + 1./648.*me2*RSp45611e*
         RSe45611p - 1./324.*me2*RS145pe*RSe45613p + 1./648.*me2*
         RS145pe*RSe45633p + 1./648.*me2*RS145pe*RSe45611p - 1./324.*
         me2*RS146pe*RSe45613p + 1./648.*me2*RS146pe*RSe45633p + 1./648.
         *me2*RS146pe*RSe45611p - 1./324.*me2*RS156pe*RSe45613p + 1./
         648.*me2*RS156pe*RSe45633p + 1./648.*me2*RS156pe*RSe45611p + 1.
         /8.*m33 );

      F13p6x23R +=  + Abeps(m13,mu2)*pi16 * ( 1./162.*m33*pow(
         RS34561pe,2) + 1./162.*m33*RS14563pe*RS34561pe + 1./162.*m33*
         RSe45613p*RS34561pe + 1./162.*m33*RSp45613e*RS34561pe + 1./162.
         *m33*RS3456pee*RS34561pe + 1./324.*m33*RS3456pee*RS14563pe + 1.
         /324.*m33*RS3456pee*RSe45613p + 1./324.*m33*RS3456pee*
         RSp45613e + 1./648.*m33*pow(RS3456pee,2) + 1./162.*m33*
         RS3456ppe*RS34561pe + 1./324.*m33*RS3456ppe*RS14563pe + 1./324.
         *m33*RS3456ppe*RSe45613p + 1./324.*m33*RS3456ppe*RSp45613e + 1.
         /324.*m33*RS3456ppe*RS3456pee + 1./648.*m33*pow(RS3456ppe,2)
          - 1./324.*m33*RSe45633p*RS34561pe - 1./648.*m33*RSe45633p*
         RS3456pee - 1./648.*m33*RSe45633p*RS3456ppe - 1./324.*m33*
         RSp45633e*RS34561pe - 1./648.*m33*RSp45633e*RS3456pee - 1./648.
         *m33*RSp45633e*RS3456ppe - 1./162.*m33*RS345pe*RS34561pe - 1./
         324.*m33*RS345pe*RS14563pe - 1./324.*m33*RS345pe*RSe45613p - 1.
         /324.*m33*RS345pe*RSp45613e - 1./324.*m33*RS345pe*RS3456pee - 
         1./324.*m33*RS345pe*RS3456ppe );

      F13p6x23R +=  + Abeps(m13,mu2)*pi16 * ( 1./648.*m33*RS345pe*
         RSe45633p + 1./648.*m33*RS345pe*RSp45633e + 1./648.*m33*pow(
         RS345pe,2) - 1./162.*m33*RS346pe*RS34561pe - 1./324.*m33*
         RS346pe*RS14563pe - 1./324.*m33*RS346pe*RSe45613p - 1./324.*
         m33*RS346pe*RSp45613e - 1./324.*m33*RS346pe*RS3456pee - 1./324.
         *m33*RS346pe*RS3456ppe + 1./648.*m33*RS346pe*RSe45633p + 1./
         648.*m33*RS346pe*RSp45633e + 1./324.*m33*RS346pe*RS345pe + 1./
         648.*m33*pow(RS346pe,2) - 1./162.*m33*RS356pe*RS34561pe - 1./
         324.*m33*RS356pe*RS14563pe - 1./324.*m33*RS356pe*RSe45613p - 1.
         /324.*m33*RS356pe*RSp45613e - 1./324.*m33*RS356pe*RS3456pee - 
         1./324.*m33*RS356pe*RS3456ppe + 1./648.*m33*RS356pe*RSe45633p
          + 1./648.*m33*RS356pe*RSp45633e + 1./324.*m33*RS356pe*RS345pe
          + 1./324.*m33*RS356pe*RS346pe + 1./648.*m33*pow(RS356pe,2) + 
         1./324.*m33*RS1456pee*RS34561pe + 1./648.*m33*RS1456pee*
         RS3456pee + 1./648.*m33*RS1456pee*RS3456ppe - 1./648.*m33*
         RS1456pee*RS345pe );

      F13p6x23R +=  + Abeps(m13,mu2)*pi16 * (  - 1./648.*m33*RS1456pee*
         RS346pe - 1./648.*m33*RS1456pee*RS356pe + 1./324.*m33*
         RS1456ppe*RS34561pe + 1./648.*m33*RS1456ppe*RS3456pee + 1./648.
         *m33*RS1456ppe*RS3456ppe - 1./648.*m33*RS1456ppe*RS345pe - 1./
         648.*m33*RS1456ppe*RS346pe - 1./648.*m33*RS1456ppe*RS356pe - 1.
         /324.*m33*RSe45611p*RS34561pe - 1./648.*m33*RSe45611p*
         RS3456pee - 1./648.*m33*RSe45611p*RS3456ppe + 1./648.*m33*
         RSe45611p*RS345pe + 1./648.*m33*RSe45611p*RS346pe + 1./648.*
         m33*RSe45611p*RS356pe - 1./324.*m33*RSp45611e*RS34561pe - 1./
         648.*m33*RSp45611e*RS3456pee - 1./648.*m33*RSp45611e*RS3456ppe
          + 1./648.*m33*RSp45611e*RS345pe + 1./648.*m33*RSp45611e*
         RS346pe + 1./648.*m33*RSp45611e*RS356pe - 1./324.*m33*RS145pe*
         RS34561pe - 1./648.*m33*RS145pe*RS3456pee - 1./648.*m33*
         RS145pe*RS3456ppe + 1./648.*m33*RS145pe*RS345pe + 1./648.*m33*
         RS145pe*RS346pe + 1./648.*m33*RS145pe*RS356pe - 1./324.*m33*
         RS146pe*RS34561pe );

      F13p6x23R +=  + Abeps(m13,mu2)*pi16 * (  - 1./648.*m33*RS146pe*
         RS3456pee - 1./648.*m33*RS146pe*RS3456ppe + 1./648.*m33*
         RS146pe*RS345pe + 1./648.*m33*RS146pe*RS346pe + 1./648.*m33*
         RS146pe*RS356pe - 1./324.*m33*RS156pe*RS34561pe - 1./648.*m33*
         RS156pe*RS3456pee - 1./648.*m33*RS156pe*RS3456ppe + 1./648.*
         m33*RS156pe*RS345pe + 1./648.*m33*RS156pe*RS346pe + 1./648.*
         m33*RS156pe*RS356pe + 1./8.*m11 + 1./162.*m11*RS14563pe*
         RS34561pe + 1./162.*m11*pow(RS14563pe,2) + 1./162.*m11*
         RSe45613p*RS14563pe + 1./162.*m11*RSp45613e*RS14563pe + 1./324.
         *m11*RS3456pee*RS14563pe + 1./324.*m11*RS3456ppe*RS14563pe - 1.
         /324.*m11*RSe45633p*RS14563pe - 1./324.*m11*RSp45633e*
         RS14563pe - 1./324.*m11*RS345pe*RS14563pe - 1./324.*m11*
         RS346pe*RS14563pe - 1./324.*m11*RS356pe*RS14563pe + 1./324.*
         m11*RS1456pee*RS34561pe + 1./162.*m11*RS1456pee*RS14563pe + 1./
         324.*m11*RS1456pee*RSe45613p + 1./324.*m11*RS1456pee*RSp45613e
          + 1./648.*m11*RS1456pee*RS3456pee );

      F13p6x23R +=  + Abeps(m13,mu2)*pi16 * ( 1./648.*m11*RS1456pee*
         RS3456ppe - 1./648.*m11*RS1456pee*RSe45633p - 1./648.*m11*
         RS1456pee*RSp45633e - 1./648.*m11*RS1456pee*RS345pe - 1./648.*
         m11*RS1456pee*RS346pe - 1./648.*m11*RS1456pee*RS356pe + 1./648.
         *m11*pow(RS1456pee,2) + 1./324.*m11*RS1456ppe*RS34561pe + 1./
         162.*m11*RS1456ppe*RS14563pe + 1./324.*m11*RS1456ppe*RSe45613p
          + 1./324.*m11*RS1456ppe*RSp45613e + 1./648.*m11*RS1456ppe*
         RS3456pee + 1./648.*m11*RS1456ppe*RS3456ppe - 1./648.*m11*
         RS1456ppe*RSe45633p - 1./648.*m11*RS1456ppe*RSp45633e - 1./648.
         *m11*RS1456ppe*RS345pe - 1./648.*m11*RS1456ppe*RS346pe - 1./
         648.*m11*RS1456ppe*RS356pe + 1./324.*m11*RS1456ppe*RS1456pee
          + 1./648.*m11*pow(RS1456ppe,2) - 1./324.*m11*RSe45611p*
         RS14563pe - 1./648.*m11*RSe45611p*RS1456pee - 1./648.*m11*
         RSe45611p*RS1456ppe - 1./324.*m11*RSp45611e*RS14563pe - 1./648.
         *m11*RSp45611e*RS1456pee - 1./648.*m11*RSp45611e*RS1456ppe - 1.
         /324.*m11*RS145pe*RS34561pe );

      F13p6x23R +=  + Abeps(m13,mu2)*pi16 * (  - 1./162.*m11*RS145pe*
         RS14563pe - 1./324.*m11*RS145pe*RSe45613p - 1./324.*m11*
         RS145pe*RSp45613e - 1./648.*m11*RS145pe*RS3456pee - 1./648.*
         m11*RS145pe*RS3456ppe + 1./648.*m11*RS145pe*RSe45633p + 1./648.
         *m11*RS145pe*RSp45633e + 1./648.*m11*RS145pe*RS345pe + 1./648.
         *m11*RS145pe*RS346pe + 1./648.*m11*RS145pe*RS356pe - 1./324.*
         m11*RS145pe*RS1456pee - 1./324.*m11*RS145pe*RS1456ppe + 1./648.
         *m11*RS145pe*RSe45611p + 1./648.*m11*RS145pe*RSp45611e + 1./
         648.*m11*pow(RS145pe,2) - 1./324.*m11*RS146pe*RS34561pe - 1./
         162.*m11*RS146pe*RS14563pe - 1./324.*m11*RS146pe*RSe45613p - 1.
         /324.*m11*RS146pe*RSp45613e - 1./648.*m11*RS146pe*RS3456pee - 
         1./648.*m11*RS146pe*RS3456ppe + 1./648.*m11*RS146pe*RSe45633p
          + 1./648.*m11*RS146pe*RSp45633e + 1./648.*m11*RS146pe*RS345pe
          + 1./648.*m11*RS146pe*RS346pe + 1./648.*m11*RS146pe*RS356pe
          - 1./324.*m11*RS146pe*RS1456pee - 1./324.*m11*RS146pe*
         RS1456ppe );

      F13p6x23R +=  + Abeps(m13,mu2)*pi16 * ( 1./648.*m11*RS146pe*
         RSe45611p + 1./648.*m11*RS146pe*RSp45611e + 1./324.*m11*
         RS146pe*RS145pe + 1./648.*m11*pow(RS146pe,2) - 1./324.*m11*
         RS156pe*RS34561pe - 1./162.*m11*RS156pe*RS14563pe - 1./324.*
         m11*RS156pe*RSe45613p - 1./324.*m11*RS156pe*RSp45613e - 1./648.
         *m11*RS156pe*RS3456pee - 1./648.*m11*RS156pe*RS3456ppe + 1./
         648.*m11*RS156pe*RSe45633p + 1./648.*m11*RS156pe*RSp45633e + 1.
         /648.*m11*RS156pe*RS345pe + 1./648.*m11*RS156pe*RS346pe + 1./
         648.*m11*RS156pe*RS356pe - 1./324.*m11*RS156pe*RS1456pee - 1./
         324.*m11*RS156pe*RS1456ppe + 1./648.*m11*RS156pe*RSe45611p + 1.
         /648.*m11*RS156pe*RSp45611e + 1./324.*m11*RS156pe*RS145pe + 1./
         324.*m11*RS156pe*RS146pe + 1./648.*m11*pow(RS156pe,2) );

      F13p6x23R +=  + Abeps(m14,mu2)*pi16 * (  - 1./48.*RD3456ep + 7./
         144.*RD1456ep - 1./36.*mp2*RSp45613e - 1./48.*mp2*RSp45633e + 
         5./72.*mp2*RSp563e + 7./144.*mp2*RSp45611e - 5./72.*mp2*
         RSp561e - 1./36.*me2*RSe45613p - 1./48.*me2*RSe45633p + 5./72.
         *me2*RSe563p + 7./144.*me2*RSe45611p - 5./72.*me2*RSe561p + 1./
         16.*m66 + 1./16.*m55 + 1./4.*m44 + 7./432.*m44*RS34561pe + 7./
         432.*m44*RS14563pe + 7./432.*m44*RSe45613p + 7./432.*m44*
         RSp45613e + 1./432.*m44*RS3456pee + 1./432.*m44*RS3456ppe - 1./
         432.*m44*RSe45633p - 1./432.*m44*RSp45633e - 1./432.*m44*
         RS345pe - 1./432.*m44*RS346pe + 1./24.*m44*RS356pe + 19./432.*
         m44*RSe563p + 19./432.*m44*RSp563e + 1./432.*m44*RS1456pee + 1.
         /432.*m44*RS1456ppe - 1./432.*m44*RSe45611p - 1./432.*m44*
         RSp45611e - 1./432.*m44*RS145pe - 1./432.*m44*RS146pe - 5./72.
         *m44*RS156pe + 5./432.*m44*RS456pe - 29./432.*m44*RSe561p - 29.
         /432.*m44*RSp561e + 5./432.*m44*RSe564p + 5./432.*m44*RSp564e
          - 5./432.*m33*RS34561pe );

      F13p6x23R +=  + Abeps(m14,mu2)*pi16 * ( 7./432.*m33*RS14563pe + 7.
         /432.*m33*RSe45613p + 7./432.*m33*RSp45613e + 5./216.*m33*
         RS3456pee + 5./216.*m33*RS3456ppe - 1./432.*m33*RSe45633p - 1./
         432.*m33*RSp45633e - 5./216.*m33*RS345pe - 5./216.*m33*RS346pe
          + 5./144.*m33*RS356pe - 5./432.*m33*RSe563p - 5./432.*m33*
         RSp563e + 1./432.*m33*RS1456pee + 1./432.*m33*RS1456ppe - 1./
         432.*m33*RSe45611p - 1./432.*m33*RSp45611e - 1./432.*m33*
         RS145pe - 1./432.*m33*RS146pe - 1./72.*m33*RS156pe + 5./432.*
         m33*RS456pe - 5./432.*m33*RSe561p - 5./432.*m33*RSp561e + 5./
         432.*m33*RSe564p + 5./432.*m33*RSp564e - 1./36.*m11*RS14563pe
          + 1./9.*m11*RS356pe + 1./9.*m11*RSe563p + 1./9.*m11*RSp563e
          - 7./144.*m11*RS1456pee - 7./144.*m11*RS1456ppe + 7./144.*m11
         *RS145pe + 7./144.*m11*RS146pe - 19./144.*m11*RS156pe - 1./9.*
         m11*RSe561p - 1./9.*m11*RSp561e );

      F13p6x23R +=  + Abeps(m15,mu2)*pi16 * (  - 1./48.*RD3456ep + 7./
         144.*RD1456ep - 1./36.*mp2*RSp45613e - 1./48.*mp2*RSp45633e + 
         5./72.*mp2*RSp463e + 7./144.*mp2*RSp45611e - 5./72.*mp2*
         RSp461e - 1./36.*me2*RSe45613p - 1./48.*me2*RSe45633p + 5./72.
         *me2*RSe463p + 7./144.*me2*RSe45611p - 5./72.*me2*RSe461p + 1./
         16.*m66 + 1./4.*m55 + 7./432.*m55*RS34561pe + 7./432.*m55*
         RS14563pe + 7./432.*m55*RSe45613p + 7./432.*m55*RSp45613e + 1./
         432.*m55*RS3456pee + 1./432.*m55*RS3456ppe - 1./432.*m55*
         RSe45633p - 1./432.*m55*RSp45633e - 1./432.*m55*RS345pe + 1./
         24.*m55*RS346pe - 1./432.*m55*RS356pe + 19./432.*m55*RSe463p
          + 19./432.*m55*RSp463e + 1./432.*m55*RS1456pee + 1./432.*m55*
         RS1456ppe - 1./432.*m55*RSe45611p - 1./432.*m55*RSp45611e - 1./
         432.*m55*RS145pe - 5./72.*m55*RS146pe - 1./432.*m55*RS156pe + 
         5./432.*m55*RS546pe - 29./432.*m55*RSe461p - 29./432.*m55*
         RSp461e + 5./432.*m55*RSe465p + 5./432.*m55*RSp465e + 1./16.*
         m44 );

      F13p6x23R +=  + Abeps(m15,mu2)*pi16 * (  - 5./432.*m33*RS34561pe
          + 7./432.*m33*RS14563pe + 7./432.*m33*RSe45613p + 7./432.*m33
         *RSp45613e + 5./216.*m33*RS3456pee + 5./216.*m33*RS3456ppe - 1.
         /432.*m33*RSe45633p - 1./432.*m33*RSp45633e - 5./216.*m33*
         RS345pe + 5./144.*m33*RS346pe - 5./216.*m33*RS356pe - 5./432.*
         m33*RSe463p - 5./432.*m33*RSp463e + 1./432.*m33*RS1456pee + 1./
         432.*m33*RS1456ppe - 1./432.*m33*RSe45611p - 1./432.*m33*
         RSp45611e - 1./432.*m33*RS145pe - 1./72.*m33*RS146pe - 1./432.
         *m33*RS156pe + 5./432.*m33*RS546pe - 5./432.*m33*RSe461p - 5./
         432.*m33*RSp461e + 5./432.*m33*RSe465p + 5./432.*m33*RSp465e
          - 1./36.*m11*RS14563pe + 1./9.*m11*RS346pe + 1./9.*m11*
         RSe463p + 1./9.*m11*RSp463e - 7./144.*m11*RS1456pee - 7./144.*
         m11*RS1456ppe + 7./144.*m11*RS145pe - 19./144.*m11*RS146pe + 7.
         /144.*m11*RS156pe - 1./9.*m11*RSe461p - 1./9.*m11*RSp461e );

      F13p6x23R +=  + Abeps(m16,mu2)*pi16 * (  - 1./48.*RD3456ep + 7./
         144.*RD1456ep - 1./36.*mp2*RSp45613e - 1./48.*mp2*RSp45633e + 
         5./72.*mp2*RSp453e + 7./144.*mp2*RSp45611e - 5./72.*mp2*
         RSp451e - 1./36.*me2*RSe45613p - 1./48.*me2*RSe45633p + 5./72.
         *me2*RSe453p + 7./144.*me2*RSe45611p - 5./72.*me2*RSe451p + 1./
         4.*m66 + 7./432.*m66*RS34561pe + 7./432.*m66*RS14563pe + 7./
         432.*m66*RSe45613p + 7./432.*m66*RSp45613e + 1./432.*m66*
         RS3456pee + 1./432.*m66*RS3456ppe - 1./432.*m66*RSe45633p - 1./
         432.*m66*RSp45633e + 1./24.*m66*RS345pe - 1./432.*m66*RS346pe
          - 1./432.*m66*RS356pe + 19./432.*m66*RSe453p + 19./432.*m66*
         RSp453e + 1./432.*m66*RS1456pee + 1./432.*m66*RS1456ppe - 1./
         432.*m66*RSe45611p - 1./432.*m66*RSp45611e - 5./72.*m66*
         RS145pe - 1./432.*m66*RS146pe - 1./432.*m66*RS156pe + 5./432.*
         m66*RS645pe - 29./432.*m66*RSe451p - 29./432.*m66*RSp451e + 5./
         432.*m66*RSe456p + 5./432.*m66*RSp456e + 1./16.*m55 + 1./16.*
         m44 );

      F13p6x23R +=  + Abeps(m16,mu2)*pi16 * (  - 5./432.*m33*RS34561pe
          + 7./432.*m33*RS14563pe + 7./432.*m33*RSe45613p + 7./432.*m33
         *RSp45613e + 5./216.*m33*RS3456pee + 5./216.*m33*RS3456ppe - 1.
         /432.*m33*RSe45633p - 1./432.*m33*RSp45633e + 5./144.*m33*
         RS345pe - 5./216.*m33*RS346pe - 5./216.*m33*RS356pe - 5./432.*
         m33*RSe453p - 5./432.*m33*RSp453e + 1./432.*m33*RS1456pee + 1./
         432.*m33*RS1456ppe - 1./432.*m33*RSe45611p - 1./432.*m33*
         RSp45611e - 1./72.*m33*RS145pe - 1./432.*m33*RS146pe - 1./432.
         *m33*RS156pe + 5./432.*m33*RS645pe - 5./432.*m33*RSe451p - 5./
         432.*m33*RSp451e + 5./432.*m33*RSe456p + 5./432.*m33*RSp456e
          - 1./36.*m11*RS14563pe + 1./9.*m11*RS345pe + 1./9.*m11*
         RSe453p + 1./9.*m11*RSp453e - 7./144.*m11*RS1456pee - 7./144.*
         m11*RS1456ppe - 19./144.*m11*RS145pe + 7./144.*m11*RS146pe + 7.
         /144.*m11*RS156pe - 1./9.*m11*RSe451p - 1./9.*m11*RSp451e );

      F13p6x23R +=  + Abeps(m33,mu2)*pi16 * (  - 1./12.*RD3456ep - 1./
         72.*RD3456ep*RS34561pe - 1./144.*RS3456pee*RD3456ep - 1./144.*
         RS3456ppe*RD3456ep + 1./144.*RS345pe*RD3456ep + 1./144.*
         RS346pe*RD3456ep + 1./144.*RS356pe*RD3456ep - 1./72.*RD1456ep*
         RS34561pe - 1./144.*RD1456ep*RS3456pee - 1./144.*RD1456ep*
         RS3456ppe + 1./144.*RD1456ep*RS345pe + 1./144.*RD1456ep*
         RS346pe + 1./144.*RD1456ep*RS356pe + 1./36.*mp2*RSp45613e*
         RS34561pe + 1./72.*mp2*RS3456pee*RSp45613e + 1./72.*mp2*
         RS3456ppe*RSp45613e - 1./72.*mp2*RSp45633e*RS34561pe - 1./144.
         *mp2*RSp45633e*RS3456pee - 1./144.*mp2*RSp45633e*RS3456ppe - 1.
         /72.*mp2*RS345pe*RSp45613e + 1./144.*mp2*RS345pe*RSp45633e - 1.
         /72.*mp2*RS346pe*RSp45613e + 1./144.*mp2*RS346pe*RSp45633e - 1.
         /72.*mp2*RS356pe*RSp45613e + 1./144.*mp2*RS356pe*RSp45633e - 1.
         /72.*mp2*RSp45611e*RS34561pe - 1./144.*mp2*RSp45611e*RS3456pee
          - 1./144.*mp2*RSp45611e*RS3456ppe + 1./144.*mp2*RSp45611e*
         RS345pe );

      F13p6x23R +=  + Abeps(m33,mu2)*pi16 * ( 1./144.*mp2*RSp45611e*
         RS346pe + 1./144.*mp2*RSp45611e*RS356pe + 1./36.*me2*RSe45613p
         *RS34561pe + 1./72.*me2*RS3456pee*RSe45613p + 1./72.*me2*
         RS3456ppe*RSe45613p - 1./72.*me2*RSe45633p*RS34561pe - 1./144.
         *me2*RSe45633p*RS3456pee - 1./144.*me2*RSe45633p*RS3456ppe - 1.
         /72.*me2*RS345pe*RSe45613p + 1./144.*me2*RS345pe*RSe45633p - 1.
         /72.*me2*RS346pe*RSe45613p + 1./144.*me2*RS346pe*RSe45633p - 1.
         /72.*me2*RS356pe*RSe45613p + 1./144.*me2*RS356pe*RSe45633p - 1.
         /72.*me2*RSe45611p*RS34561pe - 1./144.*me2*RSe45611p*RS3456pee
          - 1./144.*me2*RSe45611p*RS3456ppe + 1./144.*me2*RSe45611p*
         RS345pe + 1./144.*me2*RSe45611p*RS346pe + 1./144.*me2*
         RSe45611p*RS356pe + 7./72.*m66*RS34561pe + 1./18.*m66*
         RS14563pe + 1./18.*m66*RSe45613p + 1./18.*m66*RSp45613e + 7./
         144.*m66*RS3456pee + 7./144.*m66*RS3456ppe - 1./18.*m66*
         RSe45633p - 1./18.*m66*RSp45633e - 11./144.*m66*RS345pe + 5./
         108.*m66*pow(RS345pe,2) );

      F13p6x23R +=  + Abeps(m33,mu2)*pi16 * (  - 7./144.*m66*RS346pe + 
         1./108.*m66*RS346pe*RS345pe + 1./108.*m66*pow(RS346pe,2) - 7./
         144.*m66*RS356pe + 1./108.*m66*RS356pe*RS345pe + 1./108.*m66*
         pow(RS356pe,2) + 1./108.*m66*RSe563p*RS345pe + 1./54.*m66*
         RSe563p*RS356pe + 1./108.*m66*RSe463p*RS345pe + 1./54.*m66*
         RSe463p*RS346pe - 1./36.*m66*RSe453p + 5./54.*m66*RSe453p*
         RS345pe + 1./108.*m66*RSe453p*RS346pe + 1./108.*m66*RSe453p*
         RS356pe + 1./108.*m66*RSp563e*RS345pe + 1./54.*m66*RSp563e*
         RS356pe + 1./108.*m66*RSp463e*RS345pe + 1./54.*m66*RSp463e*
         RS346pe - 1./36.*m66*RSp453e + 5./54.*m66*RSp453e*RS345pe + 1./
         108.*m66*RSp453e*RS346pe + 1./108.*m66*RSp453e*RS356pe + 1./36.
         *m66*RS145pe - 5./54.*m66*RS145pe*RS345pe - 1./108.*m66*
         RS145pe*RS346pe - 1./108.*m66*RS145pe*RS356pe - 1./108.*m66*
         RS146pe*RS345pe - 1./54.*m66*RS146pe*RS346pe - 1./108.*m66*
         RS156pe*RS345pe - 1./54.*m66*RS156pe*RS356pe - 1./108.*m66*
         RSe561p*RS345pe );

      F13p6x23R +=  + Abeps(m33,mu2)*pi16 * (  - 1./54.*m66*RSe561p*
         RS356pe - 1./108.*m66*RSe461p*RS345pe - 1./54.*m66*RSe461p*
         RS346pe + 1./36.*m66*RSe451p - 5./54.*m66*RSe451p*RS345pe - 1./
         108.*m66*RSe451p*RS346pe - 1./108.*m66*RSe451p*RS356pe - 1./
         108.*m66*RSp561e*RS345pe - 1./54.*m66*RSp561e*RS356pe - 1./108.
         *m66*RSp461e*RS345pe - 1./54.*m66*RSp461e*RS346pe + 1./36.*m66
         *RSp451e - 5./54.*m66*RSp451e*RS345pe - 1./108.*m66*RSp451e*
         RS346pe - 1./108.*m66*RSp451e*RS356pe + 7./72.*m55*RS34561pe
          + 1./18.*m55*RS14563pe + 1./18.*m55*RSe45613p + 1./18.*m55*
         RSp45613e + 7./144.*m55*RS3456pee + 7./144.*m55*RS3456ppe - 1./
         18.*m55*RSe45633p - 1./18.*m55*RSp45633e - 7./144.*m55*RS345pe
          + 1./108.*m55*pow(RS345pe,2) - 11./144.*m55*RS346pe + 1./108.
         *m55*RS346pe*RS345pe + 5./108.*m55*pow(RS346pe,2) - 7./144.*
         m55*RS356pe + 1./108.*m55*RS356pe*RS346pe + 1./108.*m55*pow(
         RS356pe,2) + 1./108.*m55*RSe563p*RS346pe + 1./54.*m55*RSe563p*
         RS356pe );

      F13p6x23R +=  + Abeps(m33,mu2)*pi16 * (  - 1./36.*m55*RSe463p + 1.
         /108.*m55*RSe463p*RS345pe + 5./54.*m55*RSe463p*RS346pe + 1./
         108.*m55*RSe463p*RS356pe + 1./54.*m55*RSe453p*RS345pe + 1./108.
         *m55*RSe453p*RS346pe + 1./108.*m55*RSp563e*RS346pe + 1./54.*
         m55*RSp563e*RS356pe - 1./36.*m55*RSp463e + 1./108.*m55*RSp463e
         *RS345pe + 5./54.*m55*RSp463e*RS346pe + 1./108.*m55*RSp463e*
         RS356pe + 1./54.*m55*RSp453e*RS345pe + 1./108.*m55*RSp453e*
         RS346pe - 1./54.*m55*RS145pe*RS345pe - 1./108.*m55*RS145pe*
         RS346pe + 1./36.*m55*RS146pe - 1./108.*m55*RS146pe*RS345pe - 5.
         /54.*m55*RS146pe*RS346pe - 1./108.*m55*RS146pe*RS356pe - 1./
         108.*m55*RS156pe*RS346pe - 1./54.*m55*RS156pe*RS356pe - 1./108.
         *m55*RSe561p*RS346pe - 1./54.*m55*RSe561p*RS356pe + 1./36.*m55
         *RSe461p - 1./108.*m55*RSe461p*RS345pe - 5./54.*m55*RSe461p*
         RS346pe - 1./108.*m55*RSe461p*RS356pe - 1./54.*m55*RSe451p*
         RS345pe - 1./108.*m55*RSe451p*RS346pe - 1./108.*m55*RSp561e*
         RS346pe );

      F13p6x23R +=  + Abeps(m33,mu2)*pi16 * (  - 1./54.*m55*RSp561e*
         RS356pe + 1./36.*m55*RSp461e - 1./108.*m55*RSp461e*RS345pe - 5.
         /54.*m55*RSp461e*RS346pe - 1./108.*m55*RSp461e*RS356pe - 1./54.
         *m55*RSp451e*RS345pe - 1./108.*m55*RSp451e*RS346pe + 7./72.*
         m44*RS34561pe + 1./18.*m44*RS14563pe + 1./18.*m44*RSe45613p + 
         1./18.*m44*RSp45613e + 7./144.*m44*RS3456pee + 7./144.*m44*
         RS3456ppe - 1./18.*m44*RSe45633p - 1./18.*m44*RSp45633e - 7./
         144.*m44*RS345pe + 1./108.*m44*pow(RS345pe,2) - 7./144.*m44*
         RS346pe + 1./108.*m44*pow(RS346pe,2) - 11./144.*m44*RS356pe + 
         1./108.*m44*RS356pe*RS345pe + 1./108.*m44*RS356pe*RS346pe + 5./
         108.*m44*pow(RS356pe,2) - 1./36.*m44*RSe563p + 1./108.*m44*
         RSe563p*RS345pe + 1./108.*m44*RSe563p*RS346pe + 5./54.*m44*
         RSe563p*RS356pe + 1./54.*m44*RSe463p*RS346pe + 1./108.*m44*
         RSe463p*RS356pe + 1./54.*m44*RSe453p*RS345pe + 1./108.*m44*
         RSe453p*RS356pe - 1./36.*m44*RSp563e + 1./108.*m44*RSp563e*
         RS345pe );

      F13p6x23R +=  + Abeps(m33,mu2)*pi16 * ( 1./108.*m44*RSp563e*
         RS346pe + 5./54.*m44*RSp563e*RS356pe + 1./54.*m44*RSp463e*
         RS346pe + 1./108.*m44*RSp463e*RS356pe + 1./54.*m44*RSp453e*
         RS345pe + 1./108.*m44*RSp453e*RS356pe - 1./54.*m44*RS145pe*
         RS345pe - 1./108.*m44*RS145pe*RS356pe - 1./54.*m44*RS146pe*
         RS346pe - 1./108.*m44*RS146pe*RS356pe + 1./36.*m44*RS156pe - 1.
         /108.*m44*RS156pe*RS345pe - 1./108.*m44*RS156pe*RS346pe - 5./
         54.*m44*RS156pe*RS356pe + 1./36.*m44*RSe561p - 1./108.*m44*
         RSe561p*RS345pe - 1./108.*m44*RSe561p*RS346pe - 5./54.*m44*
         RSe561p*RS356pe - 1./54.*m44*RSe461p*RS346pe - 1./108.*m44*
         RSe461p*RS356pe - 1./54.*m44*RSe451p*RS345pe - 1./108.*m44*
         RSe451p*RS356pe + 1./36.*m44*RSp561e - 1./108.*m44*RSp561e*
         RS345pe - 1./108.*m44*RSp561e*RS346pe - 5./54.*m44*RSp561e*
         RS356pe - 1./54.*m44*RSp461e*RS346pe - 1./108.*m44*RSp461e*
         RS356pe - 1./54.*m44*RSp451e*RS345pe - 1./108.*m44*RSp451e*
         RS356pe );

      F13p6x23R +=  + Abeps(m33,mu2)*pi16 * ( 13./48.*m33 + 1./48.*m33*
         RS34561pe + 5./162.*m33*pow(RS34561pe,2) + 1./6.*m33*RS14563pe
          + 1./324.*m33*RS14563pe*RS34561pe + 1./6.*m33*RSe45613p + 1./
         324.*m33*RSe45613p*RS34561pe + 1./6.*m33*RSp45613e + 1./324.*
         m33*RSp45613e*RS34561pe + 7./48.*m33*RS3456pee + 7./81.*m33*
         RS3456pee*RS34561pe + 37./648.*m33*RS3456pee*RS14563pe + 37./
         648.*m33*RS3456pee*RSe45613p + 37./648.*m33*RS3456pee*
         RSp45613e + 41./648.*m33*pow(RS3456pee,2) + 7./48.*m33*
         RS3456ppe + 7./81.*m33*RS3456ppe*RS34561pe + 37./648.*m33*
         RS3456ppe*RS14563pe + 37./648.*m33*RS3456ppe*RSe45613p + 37./
         648.*m33*RS3456ppe*RSp45613e + 41./324.*m33*RS3456ppe*
         RS3456pee + 41./648.*m33*pow(RS3456ppe,2) - 1./6.*m33*
         RSe45633p - 1./648.*m33*RSe45633p*RS34561pe - 73./1296.*m33*
         RSe45633p*RS3456pee - 73./1296.*m33*RSe45633p*RS3456ppe - 1./6.
         *m33*RSp45633e - 1./648.*m33*RSp45633e*RS34561pe - 73./1296.*
         m33*RSp45633e*RS3456pee );

      F13p6x23R +=  + Abeps(m33,mu2)*pi16 * (  - 73./1296.*m33*
         RSp45633e*RS3456ppe - 17./72.*m33*RS345pe - 7./81.*m33*RS345pe
         *RS34561pe - 37./648.*m33*RS345pe*RS14563pe - 37./648.*m33*
         RS345pe*RSe45613p - 37./648.*m33*RS345pe*RSp45613e - 41./324.*
         m33*RS345pe*RS3456pee - 41./324.*m33*RS345pe*RS3456ppe + 73./
         1296.*m33*RS345pe*RSe45633p + 73./1296.*m33*RS345pe*RSp45633e
          + 41./648.*m33*pow(RS345pe,2) - 17./72.*m33*RS346pe - 7./81.*
         m33*RS346pe*RS34561pe - 37./648.*m33*RS346pe*RS14563pe - 37./
         648.*m33*RS346pe*RSe45613p - 37./648.*m33*RS346pe*RSp45613e - 
         41./324.*m33*RS346pe*RS3456pee - 41./324.*m33*RS346pe*
         RS3456ppe + 73./1296.*m33*RS346pe*RSe45633p + 73./1296.*m33*
         RS346pe*RSp45633e + 41./324.*m33*RS346pe*RS345pe + 41./648.*
         m33*pow(RS346pe,2) - 17./72.*m33*RS356pe - 7./81.*m33*RS356pe*
         RS34561pe - 37./648.*m33*RS356pe*RS14563pe - 37./648.*m33*
         RS356pe*RSe45613p - 37./648.*m33*RS356pe*RSp45613e - 41./324.*
         m33*RS356pe*RS3456pee );

      F13p6x23R +=  + Abeps(m33,mu2)*pi16 * (  - 41./324.*m33*RS356pe*
         RS3456ppe + 73./1296.*m33*RS356pe*RSe45633p + 73./1296.*m33*
         RS356pe*RSp45633e + 41./324.*m33*RS356pe*RS345pe + 41./324.*
         m33*RS356pe*RS346pe + 41./648.*m33*pow(RS356pe,2) - 1./36.*m33
         *RSe563p - 1./36.*m33*RSe463p - 1./36.*m33*RSe453p - 1./36.*
         m33*RSp563e - 1./36.*m33*RSp463e - 1./36.*m33*RSp453e + 1./648.
         *m33*RS1456pee*RS34561pe + 1./1296.*m33*RS1456pee*RS3456pee + 
         1./1296.*m33*RS1456pee*RS3456ppe - 1./1296.*m33*RS1456pee*
         RS345pe - 1./1296.*m33*RS1456pee*RS346pe - 1./1296.*m33*
         RS1456pee*RS356pe + 1./648.*m33*RS1456ppe*RS34561pe + 1./1296.
         *m33*RS1456ppe*RS3456pee + 1./1296.*m33*RS1456ppe*RS3456ppe - 
         1./1296.*m33*RS1456ppe*RS345pe - 1./1296.*m33*RS1456ppe*
         RS346pe - 1./1296.*m33*RS1456ppe*RS356pe - 1./648.*m33*
         RSe45611p*RS34561pe - 1./1296.*m33*RSe45611p*RS3456pee - 1./
         1296.*m33*RSe45611p*RS3456ppe + 1./1296.*m33*RSe45611p*RS345pe
          + 1./1296.*m33*RSe45611p*RS346pe );

      F13p6x23R +=  + Abeps(m33,mu2)*pi16 * ( 1./1296.*m33*RSe45611p*
         RS356pe - 1./648.*m33*RSp45611e*RS34561pe - 1./1296.*m33*
         RSp45611e*RS3456pee - 1./1296.*m33*RSp45611e*RS3456ppe + 1./
         1296.*m33*RSp45611e*RS345pe + 1./1296.*m33*RSp45611e*RS346pe
          + 1./1296.*m33*RSp45611e*RS356pe + 1./36.*m33*RS145pe - 1./
         648.*m33*RS145pe*RS34561pe - 1./1296.*m33*RS145pe*RS3456pee - 
         1./1296.*m33*RS145pe*RS3456ppe + 1./1296.*m33*RS145pe*RS345pe
          + 1./1296.*m33*RS145pe*RS346pe + 1./1296.*m33*RS145pe*RS356pe
          + 1./36.*m33*RS146pe - 1./648.*m33*RS146pe*RS34561pe - 1./
         1296.*m33*RS146pe*RS3456pee - 1./1296.*m33*RS146pe*RS3456ppe
          + 1./1296.*m33*RS146pe*RS345pe + 1./1296.*m33*RS146pe*RS346pe
          + 1./1296.*m33*RS146pe*RS356pe + 1./36.*m33*RS156pe - 1./648.
         *m33*RS156pe*RS34561pe - 1./1296.*m33*RS156pe*RS3456pee - 1./
         1296.*m33*RS156pe*RS3456ppe + 1./1296.*m33*RS156pe*RS345pe + 1.
         /1296.*m33*RS156pe*RS346pe + 1./1296.*m33*RS156pe*RS356pe + 1./
         36.*m33*RSe561p );

      F13p6x23R +=  + Abeps(m33,mu2)*pi16 * ( 1./36.*m33*RSe461p + 1./
         36.*m33*RSe451p + 1./36.*m33*RSp561e + 1./36.*m33*RSp461e + 1./
         36.*m33*RSp451e - 5./48.*m11 + 7./48.*m11*RS34561pe + 1./324.*
         m11*pow(RS34561pe,2) + 5./162.*m11*RS14563pe*RS34561pe + 1./
         324.*m11*RSe45613p*RS34561pe + 1./324.*m11*RSp45613e*RS34561pe
          + 1./324.*m11*RS3456pee*RS34561pe + 5./324.*m11*RS3456pee*
         RS14563pe + 1./648.*m11*RS3456pee*RSe45613p + 1./648.*m11*
         RS3456pee*RSp45613e + 1./1296.*m11*pow(RS3456pee,2) + 1./324.*
         m11*RS3456ppe*RS34561pe + 5./324.*m11*RS3456ppe*RS14563pe + 1./
         648.*m11*RS3456ppe*RSe45613p + 1./648.*m11*RS3456ppe*RSp45613e
          + 1./648.*m11*RS3456ppe*RS3456pee + 1./1296.*m11*pow(
         RS3456ppe,2) - 1./648.*m11*RSe45633p*RS34561pe - 1./1296.*m11*
         RSe45633p*RS3456pee - 1./1296.*m11*RSe45633p*RS3456ppe - 1./
         648.*m11*RSp45633e*RS34561pe - 1./1296.*m11*RSp45633e*
         RS3456pee - 1./1296.*m11*RSp45633e*RS3456ppe + 5./144.*m11*
         RS345pe );

      F13p6x23R +=  + Abeps(m33,mu2)*pi16 * (  - 1./324.*m11*RS345pe*
         RS34561pe - 5./324.*m11*RS345pe*RS14563pe - 1./648.*m11*
         RS345pe*RSe45613p - 1./648.*m11*RS345pe*RSp45613e - 1./648.*
         m11*RS345pe*RS3456pee - 1./648.*m11*RS345pe*RS3456ppe + 1./
         1296.*m11*RS345pe*RSe45633p + 1./1296.*m11*RS345pe*RSp45633e
          + 1./1296.*m11*pow(RS345pe,2) + 5./144.*m11*RS346pe - 1./324.
         *m11*RS346pe*RS34561pe - 5./324.*m11*RS346pe*RS14563pe - 1./
         648.*m11*RS346pe*RSe45613p - 1./648.*m11*RS346pe*RSp45613e - 1.
         /648.*m11*RS346pe*RS3456pee - 1./648.*m11*RS346pe*RS3456ppe + 
         1./1296.*m11*RS346pe*RSe45633p + 1./1296.*m11*RS346pe*
         RSp45633e + 1./648.*m11*RS346pe*RS345pe + 1./1296.*m11*pow(
         RS346pe,2) + 5./144.*m11*RS356pe - 1./324.*m11*RS356pe*
         RS34561pe - 5./324.*m11*RS356pe*RS14563pe - 1./648.*m11*
         RS356pe*RSe45613p - 1./648.*m11*RS356pe*RSp45613e - 1./648.*
         m11*RS356pe*RS3456pee - 1./648.*m11*RS356pe*RS3456ppe + 1./
         1296.*m11*RS356pe*RSe45633p );

      F13p6x23R +=  + Abeps(m33,mu2)*pi16 * ( 1./1296.*m11*RS356pe*
         RSp45633e + 1./648.*m11*RS356pe*RS345pe + 1./648.*m11*RS356pe*
         RS346pe + 1./1296.*m11*pow(RS356pe,2) + 5./324.*m11*RS1456pee*
         RS34561pe + 5./648.*m11*RS1456pee*RS3456pee + 5./648.*m11*
         RS1456pee*RS3456ppe - 5./648.*m11*RS1456pee*RS345pe - 5./648.*
         m11*RS1456pee*RS346pe - 5./648.*m11*RS1456pee*RS356pe + 5./324.
         *m11*RS1456ppe*RS34561pe + 5./648.*m11*RS1456ppe*RS3456pee + 5.
         /648.*m11*RS1456ppe*RS3456ppe - 5./648.*m11*RS1456ppe*RS345pe
          - 5./648.*m11*RS1456ppe*RS346pe - 5./648.*m11*RS1456ppe*
         RS356pe - 1./648.*m11*RSe45611p*RS34561pe - 1./1296.*m11*
         RSe45611p*RS3456pee - 1./1296.*m11*RSe45611p*RS3456ppe + 1./
         1296.*m11*RSe45611p*RS345pe + 1./1296.*m11*RSe45611p*RS346pe
          + 1./1296.*m11*RSe45611p*RS356pe - 1./648.*m11*RSp45611e*
         RS34561pe - 1./1296.*m11*RSp45611e*RS3456pee - 1./1296.*m11*
         RSp45611e*RS3456ppe + 1./1296.*m11*RSp45611e*RS345pe + 1./1296.
         *m11*RSp45611e*RS346pe );

      F13p6x23R +=  + Abeps(m33,mu2)*pi16 * ( 1./1296.*m11*RSp45611e*
         RS356pe - 5./324.*m11*RS145pe*RS34561pe - 5./648.*m11*RS145pe*
         RS3456pee - 5./648.*m11*RS145pe*RS3456ppe + 5./648.*m11*
         RS145pe*RS345pe + 5./648.*m11*RS145pe*RS346pe + 5./648.*m11*
         RS145pe*RS356pe - 5./324.*m11*RS146pe*RS34561pe - 5./648.*m11*
         RS146pe*RS3456pee - 5./648.*m11*RS146pe*RS3456ppe + 5./648.*
         m11*RS146pe*RS345pe + 5./648.*m11*RS146pe*RS346pe + 5./648.*
         m11*RS146pe*RS356pe - 5./324.*m11*RS156pe*RS34561pe - 5./648.*
         m11*RS156pe*RS3456pee - 5./648.*m11*RS156pe*RS3456ppe + 5./648.
         *m11*RS156pe*RS345pe + 5./648.*m11*RS156pe*RS346pe + 5./648.*
         m11*RS156pe*RS356pe );

      F13p6x23R +=  + Abeps(m34,mu2)*pi16 * ( 7./144.*RD3456ep - 1./48.
         *RD1456ep - 1./36.*mp2*RSp45613e + 7./144.*mp2*RSp45633e - 5./
         72.*mp2*RSp563e - 1./48.*mp2*RSp45611e + 5./72.*mp2*RSp561e - 
         1./36.*me2*RSe45613p + 7./144.*me2*RSe45633p - 5./72.*me2*
         RSe563p - 1./48.*me2*RSe45611p + 5./72.*me2*RSe561p + 1./16.*
         m66 + 1./16.*m55 + 1./4.*m44 + 7./432.*m44*RS34561pe + 7./432.
         *m44*RS14563pe + 7./432.*m44*RSe45613p + 7./432.*m44*RSp45613e
          + 1./432.*m44*RS3456pee + 1./432.*m44*RS3456ppe - 1./432.*m44
         *RSe45633p - 1./432.*m44*RSp45633e - 1./432.*m44*RS345pe - 1./
         432.*m44*RS346pe - 5./72.*m44*RS356pe - 29./432.*m44*RSe563p
          - 29./432.*m44*RSp563e + 1./432.*m44*RS1456pee + 1./432.*m44*
         RS1456ppe - 1./432.*m44*RSe45611p - 1./432.*m44*RSp45611e - 1./
         432.*m44*RS145pe - 1./432.*m44*RS146pe + 1./24.*m44*RS156pe + 
         5./432.*m44*RS456pe + 19./432.*m44*RSe561p + 19./432.*m44*
         RSp561e + 5./432.*m44*RSe564p + 5./432.*m44*RSp564e - 1./36.*
         m33*RS34561pe );

      F13p6x23R +=  + Abeps(m34,mu2)*pi16 * (  - 7./144.*m33*RS3456pee
          - 7./144.*m33*RS3456ppe + 7./144.*m33*RS345pe + 7./144.*m33*
         RS346pe - 19./144.*m33*RS356pe - 1./9.*m33*RSe563p - 1./9.*m33
         *RSp563e + 1./9.*m33*RS156pe + 1./9.*m33*RSe561p + 1./9.*m33*
         RSp561e + 7./432.*m11*RS34561pe - 5./432.*m11*RS14563pe + 7./
         432.*m11*RSe45613p + 7./432.*m11*RSp45613e + 1./432.*m11*
         RS3456pee + 1./432.*m11*RS3456ppe - 1./432.*m11*RSe45633p - 1./
         432.*m11*RSp45633e - 1./432.*m11*RS345pe - 1./432.*m11*RS346pe
          - 1./72.*m11*RS356pe - 5./432.*m11*RSe563p - 5./432.*m11*
         RSp563e + 5./216.*m11*RS1456pee + 5./216.*m11*RS1456ppe - 1./
         432.*m11*RSe45611p - 1./432.*m11*RSp45611e - 5./216.*m11*
         RS145pe - 5./216.*m11*RS146pe + 5./144.*m11*RS156pe + 5./432.*
         m11*RS456pe - 5./432.*m11*RSe561p - 5./432.*m11*RSp561e + 5./
         432.*m11*RSe564p + 5./432.*m11*RSp564e );

      F13p6x23R +=  + Abeps(m35,mu2)*pi16 * ( 7./144.*RD3456ep - 1./48.
         *RD1456ep - 1./36.*mp2*RSp45613e + 7./144.*mp2*RSp45633e - 5./
         72.*mp2*RSp463e - 1./48.*mp2*RSp45611e + 5./72.*mp2*RSp461e - 
         1./36.*me2*RSe45613p + 7./144.*me2*RSe45633p - 5./72.*me2*
         RSe463p - 1./48.*me2*RSe45611p + 5./72.*me2*RSe461p + 1./16.*
         m66 + 1./4.*m55 + 7./432.*m55*RS34561pe + 7./432.*m55*
         RS14563pe + 7./432.*m55*RSe45613p + 7./432.*m55*RSp45613e + 1./
         432.*m55*RS3456pee + 1./432.*m55*RS3456ppe - 1./432.*m55*
         RSe45633p - 1./432.*m55*RSp45633e - 1./432.*m55*RS345pe - 5./
         72.*m55*RS346pe - 1./432.*m55*RS356pe - 29./432.*m55*RSe463p
          - 29./432.*m55*RSp463e + 1./432.*m55*RS1456pee + 1./432.*m55*
         RS1456ppe - 1./432.*m55*RSe45611p - 1./432.*m55*RSp45611e - 1./
         432.*m55*RS145pe + 1./24.*m55*RS146pe - 1./432.*m55*RS156pe + 
         5./432.*m55*RS546pe + 19./432.*m55*RSe461p + 19./432.*m55*
         RSp461e + 5./432.*m55*RSe465p + 5./432.*m55*RSp465e + 1./16.*
         m44 );

      F13p6x23R +=  + Abeps(m35,mu2)*pi16 * (  - 1./36.*m33*RS34561pe
          - 7./144.*m33*RS3456pee - 7./144.*m33*RS3456ppe + 7./144.*m33
         *RS345pe - 19./144.*m33*RS346pe + 7./144.*m33*RS356pe - 1./9.*
         m33*RSe463p - 1./9.*m33*RSp463e + 1./9.*m33*RS146pe + 1./9.*
         m33*RSe461p + 1./9.*m33*RSp461e + 7./432.*m11*RS34561pe - 5./
         432.*m11*RS14563pe + 7./432.*m11*RSe45613p + 7./432.*m11*
         RSp45613e + 1./432.*m11*RS3456pee + 1./432.*m11*RS3456ppe - 1./
         432.*m11*RSe45633p - 1./432.*m11*RSp45633e - 1./432.*m11*
         RS345pe - 1./72.*m11*RS346pe - 1./432.*m11*RS356pe - 5./432.*
         m11*RSe463p - 5./432.*m11*RSp463e + 5./216.*m11*RS1456pee + 5./
         216.*m11*RS1456ppe - 1./432.*m11*RSe45611p - 1./432.*m11*
         RSp45611e - 5./216.*m11*RS145pe + 5./144.*m11*RS146pe - 5./216.
         *m11*RS156pe + 5./432.*m11*RS546pe - 5./432.*m11*RSe461p - 5./
         432.*m11*RSp461e + 5./432.*m11*RSe465p + 5./432.*m11*RSp465e )
         ;

      F13p6x23R +=  + Abeps(m36,mu2)*pi16 * ( 7./144.*RD3456ep - 1./48.
         *RD1456ep - 1./36.*mp2*RSp45613e + 7./144.*mp2*RSp45633e - 5./
         72.*mp2*RSp453e - 1./48.*mp2*RSp45611e + 5./72.*mp2*RSp451e - 
         1./36.*me2*RSe45613p + 7./144.*me2*RSe45633p - 5./72.*me2*
         RSe453p - 1./48.*me2*RSe45611p + 5./72.*me2*RSe451p + 1./4.*
         m66 + 7./432.*m66*RS34561pe + 7./432.*m66*RS14563pe + 7./432.*
         m66*RSe45613p + 7./432.*m66*RSp45613e + 1./432.*m66*RS3456pee
          + 1./432.*m66*RS3456ppe - 1./432.*m66*RSe45633p - 1./432.*m66
         *RSp45633e - 5./72.*m66*RS345pe - 1./432.*m66*RS346pe - 1./432.
         *m66*RS356pe - 29./432.*m66*RSe453p - 29./432.*m66*RSp453e + 1.
         /432.*m66*RS1456pee + 1./432.*m66*RS1456ppe - 1./432.*m66*
         RSe45611p - 1./432.*m66*RSp45611e + 1./24.*m66*RS145pe - 1./
         432.*m66*RS146pe - 1./432.*m66*RS156pe + 5./432.*m66*RS645pe
          + 19./432.*m66*RSe451p + 19./432.*m66*RSp451e + 5./432.*m66*
         RSe456p + 5./432.*m66*RSp456e + 1./16.*m55 + 1./16.*m44 - 1./
         36.*m33*RS34561pe );

      F13p6x23R +=  + Abeps(m36,mu2)*pi16 * (  - 7./144.*m33*RS3456pee
          - 7./144.*m33*RS3456ppe - 19./144.*m33*RS345pe + 7./144.*m33*
         RS346pe + 7./144.*m33*RS356pe - 1./9.*m33*RSe453p - 1./9.*m33*
         RSp453e + 1./9.*m33*RS145pe + 1./9.*m33*RSe451p + 1./9.*m33*
         RSp451e + 7./432.*m11*RS34561pe - 5./432.*m11*RS14563pe + 7./
         432.*m11*RSe45613p + 7./432.*m11*RSp45613e + 1./432.*m11*
         RS3456pee + 1./432.*m11*RS3456ppe - 1./432.*m11*RSe45633p - 1./
         432.*m11*RSp45633e - 1./72.*m11*RS345pe - 1./432.*m11*RS346pe
          - 1./432.*m11*RS356pe - 5./432.*m11*RSe453p - 5./432.*m11*
         RSp453e + 5./216.*m11*RS1456pee + 5./216.*m11*RS1456ppe - 1./
         432.*m11*RSe45611p - 1./432.*m11*RSp45611e + 5./144.*m11*
         RS145pe - 5./216.*m11*RS146pe - 5./216.*m11*RS156pe + 5./432.*
         m11*RS645pe - 5./432.*m11*RSe451p - 5./432.*m11*RSp451e + 5./
         432.*m11*RSe456p + 5./432.*m11*RSp456e );

      F13p6x23R +=  + Abeps(m44,mu2)*pi16 * ( 1./8.*m44 + 1./36.*m44*
         pow(RS356pe,2) + 1./18.*m44*RSe563p*RS356pe + 1./36.*m44*pow(
         RSe563p,2) + 1./18.*m44*RSp563e*RS356pe + 1./18.*m44*RSp563e*
         RSe563p + 1./36.*m44*pow(RSp563e,2) - 1./18.*m44*RS156pe*
         RS356pe - 1./18.*m44*RS156pe*RSe563p - 1./18.*m44*RS156pe*
         RSp563e + 1./36.*m44*pow(RS156pe,2) - 1./24.*m44*RS456pe - 1./
         108.*m44*RS456pe*pow(RS356pe,2) - 1./54.*m44*RS456pe*RSe563p*
         RS356pe - 1./108.*m44*RS456pe*pow(RSe563p,2) - 1./54.*m44*
         RS456pe*RSp563e*RS356pe - 1./54.*m44*RS456pe*RSp563e*RSe563p
          - 1./108.*m44*RS456pe*pow(RSp563e,2) + 1./54.*m44*RS456pe*
         RS156pe*RS356pe + 1./54.*m44*RS456pe*RS156pe*RSe563p + 1./54.*
         m44*RS456pe*RS156pe*RSp563e - 1./108.*m44*RS456pe*pow(
         RS156pe,2) - 1./18.*m44*RSe561p*RS356pe - 1./18.*m44*RSe561p*
         RSe563p - 1./18.*m44*RSe561p*RSp563e + 1./18.*m44*RSe561p*
         RS156pe + 1./54.*m44*RSe561p*RS456pe*RS356pe + 1./54.*m44*
         RSe561p*RS456pe*RSe563p );

      F13p6x23R +=  + Abeps(m44,mu2)*pi16 * ( 1./54.*m44*RSe561p*
         RS456pe*RSp563e - 1./54.*m44*RSe561p*RS456pe*RS156pe + 1./36.*
         m44*pow(RSe561p,2) - 1./108.*m44*pow(RSe561p,2)*RS456pe - 1./
         18.*m44*RSp561e*RS356pe - 1./18.*m44*RSp561e*RSe563p - 1./18.*
         m44*RSp561e*RSp563e + 1./18.*m44*RSp561e*RS156pe + 1./54.*m44*
         RSp561e*RS456pe*RS356pe + 1./54.*m44*RSp561e*RS456pe*RSe563p
          + 1./54.*m44*RSp561e*RS456pe*RSp563e - 1./54.*m44*RSp561e*
         RS456pe*RS156pe + 1./18.*m44*RSp561e*RSe561p - 1./54.*m44*
         RSp561e*RSe561p*RS456pe + 1./36.*m44*pow(RSp561e,2) - 1./108.*
         m44*pow(RSp561e,2)*RS456pe );

      F13p6x23R +=  + Abeps(m45,mu2)*pi16 * ( 1./54.*mp2*RSp563e*
         RS346pe - 1./54.*mp2*RSp563e*RS356pe - 1./54.*mp2*RSp563e*
         RSe563p + 1./54.*mp2*RSp563e*RSe463p - 1./54.*mp2*pow(
         RSp563e,2) - 1./54.*mp2*RSp463e*RS346pe + 1./54.*mp2*RSp463e*
         RS356pe + 1./54.*mp2*RSp463e*RSe563p - 1./54.*mp2*RSp463e*
         RSe463p + 1./27.*mp2*RSp463e*RSp563e - 1./54.*mp2*pow(
         RSp463e,2) - 1./54.*mp2*RS146pe*RSp563e + 1./54.*mp2*RS146pe*
         RSp463e + 1./54.*mp2*RS156pe*RSp563e - 1./54.*mp2*RS156pe*
         RSp463e + 1./54.*mp2*RSe561p*RSp563e - 1./54.*mp2*RSe561p*
         RSp463e - 1./54.*mp2*RSe461p*RSp563e + 1./54.*mp2*RSe461p*
         RSp463e - 1./54.*mp2*RSp561e*RS346pe + 1./54.*mp2*RSp561e*
         RS356pe + 1./54.*mp2*RSp561e*RSe563p - 1./54.*mp2*RSp561e*
         RSe463p + 1./27.*mp2*RSp561e*RSp563e - 1./27.*mp2*RSp561e*
         RSp463e + 1./54.*mp2*RSp561e*RS146pe - 1./54.*mp2*RSp561e*
         RS156pe - 1./54.*mp2*RSp561e*RSe561p + 1./54.*mp2*RSp561e*
         RSe461p );

      F13p6x23R +=  + Abeps(m45,mu2)*pi16 * (  - 1./54.*mp2*pow(
         RSp561e,2) + 1./54.*mp2*RSp461e*RS346pe - 1./54.*mp2*RSp461e*
         RS356pe - 1./54.*mp2*RSp461e*RSe563p + 1./54.*mp2*RSp461e*
         RSe463p - 1./27.*mp2*RSp461e*RSp563e + 1./27.*mp2*RSp461e*
         RSp463e - 1./54.*mp2*RSp461e*RS146pe + 1./54.*mp2*RSp461e*
         RS156pe + 1./54.*mp2*RSp461e*RSe561p - 1./54.*mp2*RSp461e*
         RSe461p + 1./27.*mp2*RSp461e*RSp561e - 1./54.*mp2*pow(
         RSp461e,2) + 1./54.*me2*RSe563p*RS346pe - 1./54.*me2*RSe563p*
         RS356pe - 1./54.*me2*pow(RSe563p,2) - 1./54.*me2*RSe463p*
         RS346pe + 1./54.*me2*RSe463p*RS356pe + 1./27.*me2*RSe463p*
         RSe563p - 1./54.*me2*pow(RSe463p,2) - 1./54.*me2*RSp563e*
         RSe563p + 1./54.*me2*RSp563e*RSe463p + 1./54.*me2*RSp463e*
         RSe563p - 1./54.*me2*RSp463e*RSe463p - 1./54.*me2*RS146pe*
         RSe563p + 1./54.*me2*RS146pe*RSe463p + 1./54.*me2*RS156pe*
         RSe563p - 1./54.*me2*RS156pe*RSe463p - 1./54.*me2*RSe561p*
         RS346pe );

      F13p6x23R +=  + Abeps(m45,mu2)*pi16 * ( 1./54.*me2*RSe561p*
         RS356pe + 1./27.*me2*RSe561p*RSe563p - 1./27.*me2*RSe561p*
         RSe463p + 1./54.*me2*RSe561p*RSp563e - 1./54.*me2*RSe561p*
         RSp463e + 1./54.*me2*RSe561p*RS146pe - 1./54.*me2*RSe561p*
         RS156pe - 1./54.*me2*pow(RSe561p,2) + 1./54.*me2*RSe461p*
         RS346pe - 1./54.*me2*RSe461p*RS356pe - 1./27.*me2*RSe461p*
         RSe563p + 1./27.*me2*RSe461p*RSe463p - 1./54.*me2*RSe461p*
         RSp563e + 1./54.*me2*RSe461p*RSp463e - 1./54.*me2*RSe461p*
         RS146pe + 1./54.*me2*RSe461p*RS156pe + 1./27.*me2*RSe461p*
         RSe561p - 1./54.*me2*pow(RSe461p,2) + 1./54.*me2*RSp561e*
         RSe563p - 1./54.*me2*RSp561e*RSe463p - 1./54.*me2*RSp561e*
         RSe561p + 1./54.*me2*RSp561e*RSe461p - 1./54.*me2*RSp461e*
         RSe563p + 1./54.*me2*RSp461e*RSe463p + 1./54.*me2*RSp461e*
         RSe561p - 1./54.*me2*RSp461e*RSe461p + 1./8.*m55 + 1./108.*m55
         *pow(RS346pe,2) + 1./54.*m55*RS356pe*RS346pe + 1./54.*m55*
         RSe563p*RS346pe );

      F13p6x23R +=  + Abeps(m45,mu2)*pi16 * ( 1./54.*m55*RSe463p*
         RS346pe + 1./54.*m55*RSe463p*RS356pe + 1./54.*m55*RSe463p*
         RSe563p + 1./108.*m55*pow(RSe463p,2) + 1./54.*m55*RSp563e*
         RS346pe + 1./54.*m55*RSp563e*RSe463p + 1./54.*m55*RSp463e*
         RS346pe + 1./54.*m55*RSp463e*RS356pe + 1./54.*m55*RSp463e*
         RSe563p + 1./54.*m55*RSp463e*RSe463p + 1./54.*m55*RSp463e*
         RSp563e + 1./108.*m55*pow(RSp463e,2) - 1./54.*m55*RS146pe*
         RS346pe - 1./54.*m55*RS146pe*RS356pe - 1./54.*m55*RS146pe*
         RSe563p - 1./54.*m55*RS146pe*RSe463p - 1./54.*m55*RS146pe*
         RSp563e - 1./54.*m55*RS146pe*RSp463e + 1./108.*m55*pow(
         RS146pe,2) - 1./54.*m55*RS156pe*RS346pe - 1./54.*m55*RS156pe*
         RSe463p - 1./54.*m55*RS156pe*RSp463e + 1./54.*m55*RS156pe*
         RS146pe - 1./54.*m55*RSe561p*RS346pe - 1./54.*m55*RSe561p*
         RSe463p - 1./54.*m55*RSe561p*RSp463e + 1./54.*m55*RSe561p*
         RS146pe - 1./54.*m55*RSe461p*RS346pe - 1./54.*m55*RSe461p*
         RS356pe );

      F13p6x23R +=  + Abeps(m45,mu2)*pi16 * (  - 1./54.*m55*RSe461p*
         RSe563p - 1./54.*m55*RSe461p*RSe463p - 1./54.*m55*RSe461p*
         RSp563e - 1./54.*m55*RSe461p*RSp463e + 1./54.*m55*RSe461p*
         RS146pe + 1./54.*m55*RSe461p*RS156pe + 1./54.*m55*RSe461p*
         RSe561p + 1./108.*m55*pow(RSe461p,2) - 1./54.*m55*RSp561e*
         RS346pe - 1./54.*m55*RSp561e*RSe463p - 1./54.*m55*RSp561e*
         RSp463e + 1./54.*m55*RSp561e*RS146pe + 1./54.*m55*RSp561e*
         RSe461p - 1./54.*m55*RSp461e*RS346pe - 1./54.*m55*RSp461e*
         RS356pe - 1./54.*m55*RSp461e*RSe563p - 1./54.*m55*RSp461e*
         RSe463p - 1./54.*m55*RSp461e*RSp563e - 1./54.*m55*RSp461e*
         RSp463e + 1./54.*m55*RSp461e*RS146pe + 1./54.*m55*RSp461e*
         RS156pe + 1./54.*m55*RSp461e*RSe561p + 1./54.*m55*RSp461e*
         RSe461p + 1./54.*m55*RSp461e*RSp561e + 1./108.*m55*pow(
         RSp461e,2) + 1./8.*m44 + 1./54.*m44*RS356pe*RS346pe + 1./108.*
         m44*pow(RS356pe,2) + 1./54.*m44*RSe563p*RS346pe + 1./54.*m44*
         RSe563p*RS356pe );

      F13p6x23R +=  + Abeps(m45,mu2)*pi16 * ( 1./108.*m44*pow(
         RSe563p,2) + 1./54.*m44*RSe463p*RS356pe + 1./54.*m44*RSe463p*
         RSe563p + 1./54.*m44*RSp563e*RS346pe + 1./54.*m44*RSp563e*
         RS356pe + 1./54.*m44*RSp563e*RSe563p + 1./54.*m44*RSp563e*
         RSe463p + 1./108.*m44*pow(RSp563e,2) + 1./54.*m44*RSp463e*
         RS356pe + 1./54.*m44*RSp463e*RSe563p + 1./54.*m44*RSp463e*
         RSp563e - 1./54.*m44*RS146pe*RS356pe - 1./54.*m44*RS146pe*
         RSe563p - 1./54.*m44*RS146pe*RSp563e - 1./54.*m44*RS156pe*
         RS346pe - 1./54.*m44*RS156pe*RS356pe - 1./54.*m44*RS156pe*
         RSe563p - 1./54.*m44*RS156pe*RSe463p - 1./54.*m44*RS156pe*
         RSp563e - 1./54.*m44*RS156pe*RSp463e + 1./54.*m44*RS156pe*
         RS146pe + 1./108.*m44*pow(RS156pe,2) - 1./54.*m44*RSe561p*
         RS346pe - 1./54.*m44*RSe561p*RS356pe - 1./54.*m44*RSe561p*
         RSe563p - 1./54.*m44*RSe561p*RSe463p - 1./54.*m44*RSe561p*
         RSp563e - 1./54.*m44*RSe561p*RSp463e + 1./54.*m44*RSe561p*
         RS146pe );

      F13p6x23R +=  + Abeps(m45,mu2)*pi16 * ( 1./54.*m44*RSe561p*
         RS156pe + 1./108.*m44*pow(RSe561p,2) - 1./54.*m44*RSe461p*
         RS356pe - 1./54.*m44*RSe461p*RSe563p - 1./54.*m44*RSe461p*
         RSp563e + 1./54.*m44*RSe461p*RS156pe + 1./54.*m44*RSe461p*
         RSe561p - 1./54.*m44*RSp561e*RS346pe - 1./54.*m44*RSp561e*
         RS356pe - 1./54.*m44*RSp561e*RSe563p - 1./54.*m44*RSp561e*
         RSe463p - 1./54.*m44*RSp561e*RSp563e - 1./54.*m44*RSp561e*
         RSp463e + 1./54.*m44*RSp561e*RS146pe + 1./54.*m44*RSp561e*
         RS156pe + 1./54.*m44*RSp561e*RSe561p + 1./54.*m44*RSp561e*
         RSe461p + 1./108.*m44*pow(RSp561e,2) - 1./54.*m44*RSp461e*
         RS356pe - 1./54.*m44*RSp461e*RSe563p - 1./54.*m44*RSp461e*
         RSp563e + 1./54.*m44*RSp461e*RS156pe + 1./54.*m44*RSp461e*
         RSe561p + 1./54.*m44*RSp461e*RSp561e - 1./54.*m33*pow(
         RS346pe,2) + 1./27.*m33*RS356pe*RS346pe - 1./54.*m33*pow(
         RS356pe,2) + 1./54.*m33*RSe563p*RS346pe - 1./54.*m33*RSe563p*
         RS356pe );

      F13p6x23R +=  + Abeps(m45,mu2)*pi16 * (  - 1./54.*m33*RSe463p*
         RS346pe + 1./54.*m33*RSe463p*RS356pe + 1./54.*m33*RSp563e*
         RS346pe - 1./54.*m33*RSp563e*RS356pe - 1./54.*m33*RSp463e*
         RS346pe + 1./54.*m33*RSp463e*RS356pe + 1./54.*m33*RS146pe*
         RS346pe - 1./54.*m33*RS146pe*RS356pe - 1./54.*m33*RS156pe*
         RS346pe + 1./54.*m33*RS156pe*RS356pe - 1./54.*m33*RSe561p*
         RS346pe + 1./54.*m33*RSe561p*RS356pe + 1./54.*m33*RSe461p*
         RS346pe - 1./54.*m33*RSe461p*RS356pe - 1./54.*m33*RSp561e*
         RS346pe + 1./54.*m33*RSp561e*RS356pe + 1./54.*m33*RSp461e*
         RS346pe - 1./54.*m33*RSp461e*RS356pe + 1./54.*m11*RS146pe*
         RS346pe - 1./54.*m11*RS146pe*RS356pe - 1./54.*m11*RS146pe*
         RSe563p + 1./54.*m11*RS146pe*RSe463p - 1./54.*m11*RS146pe*
         RSp563e + 1./54.*m11*RS146pe*RSp463e - 1./54.*m11*pow(
         RS146pe,2) - 1./54.*m11*RS156pe*RS346pe + 1./54.*m11*RS156pe*
         RS356pe + 1./54.*m11*RS156pe*RSe563p - 1./54.*m11*RS156pe*
         RSe463p );

      F13p6x23R +=  + Abeps(m45,mu2)*pi16 * ( 1./54.*m11*RS156pe*
         RSp563e - 1./54.*m11*RS156pe*RSp463e + 1./27.*m11*RS156pe*
         RS146pe - 1./54.*m11*pow(RS156pe,2) + 1./54.*m11*RSe561p*
         RS146pe - 1./54.*m11*RSe561p*RS156pe - 1./54.*m11*RSe461p*
         RS146pe + 1./54.*m11*RSe461p*RS156pe + 1./54.*m11*RSp561e*
         RS146pe - 1./54.*m11*RSp561e*RS156pe - 1./54.*m11*RSp461e*
         RS146pe + 1./54.*m11*RSp461e*RS156pe );

      F13p6x23R +=  + Abeps(m46,mu2)*pi16 * ( 1./54.*mp2*RSp563e*
         RS345pe - 1./54.*mp2*RSp563e*RS356pe - 1./54.*mp2*RSp563e*
         RSe563p + 1./54.*mp2*RSp563e*RSe453p - 1./54.*mp2*pow(
         RSp563e,2) - 1./54.*mp2*RSp453e*RS345pe + 1./54.*mp2*RSp453e*
         RS356pe + 1./54.*mp2*RSp453e*RSe563p - 1./54.*mp2*RSp453e*
         RSe453p + 1./27.*mp2*RSp453e*RSp563e - 1./54.*mp2*pow(
         RSp453e,2) - 1./54.*mp2*RS145pe*RSp563e + 1./54.*mp2*RS145pe*
         RSp453e + 1./54.*mp2*RS156pe*RSp563e - 1./54.*mp2*RS156pe*
         RSp453e + 1./54.*mp2*RSe561p*RSp563e - 1./54.*mp2*RSe561p*
         RSp453e - 1./54.*mp2*RSe451p*RSp563e + 1./54.*mp2*RSe451p*
         RSp453e - 1./54.*mp2*RSp561e*RS345pe + 1./54.*mp2*RSp561e*
         RS356pe + 1./54.*mp2*RSp561e*RSe563p - 1./54.*mp2*RSp561e*
         RSe453p + 1./27.*mp2*RSp561e*RSp563e - 1./27.*mp2*RSp561e*
         RSp453e + 1./54.*mp2*RSp561e*RS145pe - 1./54.*mp2*RSp561e*
         RS156pe - 1./54.*mp2*RSp561e*RSe561p + 1./54.*mp2*RSp561e*
         RSe451p );

      F13p6x23R +=  + Abeps(m46,mu2)*pi16 * (  - 1./54.*mp2*pow(
         RSp561e,2) + 1./54.*mp2*RSp451e*RS345pe - 1./54.*mp2*RSp451e*
         RS356pe - 1./54.*mp2*RSp451e*RSe563p + 1./54.*mp2*RSp451e*
         RSe453p - 1./27.*mp2*RSp451e*RSp563e + 1./27.*mp2*RSp451e*
         RSp453e - 1./54.*mp2*RSp451e*RS145pe + 1./54.*mp2*RSp451e*
         RS156pe + 1./54.*mp2*RSp451e*RSe561p - 1./54.*mp2*RSp451e*
         RSe451p + 1./27.*mp2*RSp451e*RSp561e - 1./54.*mp2*pow(
         RSp451e,2) + 1./54.*me2*RSe563p*RS345pe - 1./54.*me2*RSe563p*
         RS356pe - 1./54.*me2*pow(RSe563p,2) - 1./54.*me2*RSe453p*
         RS345pe + 1./54.*me2*RSe453p*RS356pe + 1./27.*me2*RSe453p*
         RSe563p - 1./54.*me2*pow(RSe453p,2) - 1./54.*me2*RSp563e*
         RSe563p + 1./54.*me2*RSp563e*RSe453p + 1./54.*me2*RSp453e*
         RSe563p - 1./54.*me2*RSp453e*RSe453p - 1./54.*me2*RS145pe*
         RSe563p + 1./54.*me2*RS145pe*RSe453p + 1./54.*me2*RS156pe*
         RSe563p - 1./54.*me2*RS156pe*RSe453p - 1./54.*me2*RSe561p*
         RS345pe );

      F13p6x23R +=  + Abeps(m46,mu2)*pi16 * ( 1./54.*me2*RSe561p*
         RS356pe + 1./27.*me2*RSe561p*RSe563p - 1./27.*me2*RSe561p*
         RSe453p + 1./54.*me2*RSe561p*RSp563e - 1./54.*me2*RSe561p*
         RSp453e + 1./54.*me2*RSe561p*RS145pe - 1./54.*me2*RSe561p*
         RS156pe - 1./54.*me2*pow(RSe561p,2) + 1./54.*me2*RSe451p*
         RS345pe - 1./54.*me2*RSe451p*RS356pe - 1./27.*me2*RSe451p*
         RSe563p + 1./27.*me2*RSe451p*RSe453p - 1./54.*me2*RSe451p*
         RSp563e + 1./54.*me2*RSe451p*RSp453e - 1./54.*me2*RSe451p*
         RS145pe + 1./54.*me2*RSe451p*RS156pe + 1./27.*me2*RSe451p*
         RSe561p - 1./54.*me2*pow(RSe451p,2) + 1./54.*me2*RSp561e*
         RSe563p - 1./54.*me2*RSp561e*RSe453p - 1./54.*me2*RSp561e*
         RSe561p + 1./54.*me2*RSp561e*RSe451p - 1./54.*me2*RSp451e*
         RSe563p + 1./54.*me2*RSp451e*RSe453p + 1./54.*me2*RSp451e*
         RSe561p - 1./54.*me2*RSp451e*RSe451p + 1./8.*m66 + 1./108.*m66
         *pow(RS345pe,2) + 1./54.*m66*RS356pe*RS345pe + 1./54.*m66*
         RSe563p*RS345pe );

      F13p6x23R +=  + Abeps(m46,mu2)*pi16 * ( 1./54.*m66*RSe453p*
         RS345pe + 1./54.*m66*RSe453p*RS356pe + 1./54.*m66*RSe453p*
         RSe563p + 1./108.*m66*pow(RSe453p,2) + 1./54.*m66*RSp563e*
         RS345pe + 1./54.*m66*RSp563e*RSe453p + 1./54.*m66*RSp453e*
         RS345pe + 1./54.*m66*RSp453e*RS356pe + 1./54.*m66*RSp453e*
         RSe563p + 1./54.*m66*RSp453e*RSe453p + 1./54.*m66*RSp453e*
         RSp563e + 1./108.*m66*pow(RSp453e,2) - 1./54.*m66*RS145pe*
         RS345pe - 1./54.*m66*RS145pe*RS356pe - 1./54.*m66*RS145pe*
         RSe563p - 1./54.*m66*RS145pe*RSe453p - 1./54.*m66*RS145pe*
         RSp563e - 1./54.*m66*RS145pe*RSp453e + 1./108.*m66*pow(
         RS145pe,2) - 1./54.*m66*RS156pe*RS345pe - 1./54.*m66*RS156pe*
         RSe453p - 1./54.*m66*RS156pe*RSp453e + 1./54.*m66*RS156pe*
         RS145pe - 1./54.*m66*RSe561p*RS345pe - 1./54.*m66*RSe561p*
         RSe453p - 1./54.*m66*RSe561p*RSp453e + 1./54.*m66*RSe561p*
         RS145pe - 1./54.*m66*RSe451p*RS345pe - 1./54.*m66*RSe451p*
         RS356pe );

      F13p6x23R +=  + Abeps(m46,mu2)*pi16 * (  - 1./54.*m66*RSe451p*
         RSe563p - 1./54.*m66*RSe451p*RSe453p - 1./54.*m66*RSe451p*
         RSp563e - 1./54.*m66*RSe451p*RSp453e + 1./54.*m66*RSe451p*
         RS145pe + 1./54.*m66*RSe451p*RS156pe + 1./54.*m66*RSe451p*
         RSe561p + 1./108.*m66*pow(RSe451p,2) - 1./54.*m66*RSp561e*
         RS345pe - 1./54.*m66*RSp561e*RSe453p - 1./54.*m66*RSp561e*
         RSp453e + 1./54.*m66*RSp561e*RS145pe + 1./54.*m66*RSp561e*
         RSe451p - 1./54.*m66*RSp451e*RS345pe - 1./54.*m66*RSp451e*
         RS356pe - 1./54.*m66*RSp451e*RSe563p - 1./54.*m66*RSp451e*
         RSe453p - 1./54.*m66*RSp451e*RSp563e - 1./54.*m66*RSp451e*
         RSp453e + 1./54.*m66*RSp451e*RS145pe + 1./54.*m66*RSp451e*
         RS156pe + 1./54.*m66*RSp451e*RSe561p + 1./54.*m66*RSp451e*
         RSe451p + 1./54.*m66*RSp451e*RSp561e + 1./108.*m66*pow(
         RSp451e,2) + 1./8.*m44 + 1./54.*m44*RS356pe*RS345pe + 1./108.*
         m44*pow(RS356pe,2) + 1./54.*m44*RSe563p*RS345pe + 1./54.*m44*
         RSe563p*RS356pe );

      F13p6x23R +=  + Abeps(m46,mu2)*pi16 * ( 1./108.*m44*pow(
         RSe563p,2) + 1./54.*m44*RSe453p*RS356pe + 1./54.*m44*RSe453p*
         RSe563p + 1./54.*m44*RSp563e*RS345pe + 1./54.*m44*RSp563e*
         RS356pe + 1./54.*m44*RSp563e*RSe563p + 1./54.*m44*RSp563e*
         RSe453p + 1./108.*m44*pow(RSp563e,2) + 1./54.*m44*RSp453e*
         RS356pe + 1./54.*m44*RSp453e*RSe563p + 1./54.*m44*RSp453e*
         RSp563e - 1./54.*m44*RS145pe*RS356pe - 1./54.*m44*RS145pe*
         RSe563p - 1./54.*m44*RS145pe*RSp563e - 1./54.*m44*RS156pe*
         RS345pe - 1./54.*m44*RS156pe*RS356pe - 1./54.*m44*RS156pe*
         RSe563p - 1./54.*m44*RS156pe*RSe453p - 1./54.*m44*RS156pe*
         RSp563e - 1./54.*m44*RS156pe*RSp453e + 1./54.*m44*RS156pe*
         RS145pe + 1./108.*m44*pow(RS156pe,2) - 1./54.*m44*RSe561p*
         RS345pe - 1./54.*m44*RSe561p*RS356pe - 1./54.*m44*RSe561p*
         RSe563p - 1./54.*m44*RSe561p*RSe453p - 1./54.*m44*RSe561p*
         RSp563e - 1./54.*m44*RSe561p*RSp453e + 1./54.*m44*RSe561p*
         RS145pe );

      F13p6x23R +=  + Abeps(m46,mu2)*pi16 * ( 1./54.*m44*RSe561p*
         RS156pe + 1./108.*m44*pow(RSe561p,2) - 1./54.*m44*RSe451p*
         RS356pe - 1./54.*m44*RSe451p*RSe563p - 1./54.*m44*RSe451p*
         RSp563e + 1./54.*m44*RSe451p*RS156pe + 1./54.*m44*RSe451p*
         RSe561p - 1./54.*m44*RSp561e*RS345pe - 1./54.*m44*RSp561e*
         RS356pe - 1./54.*m44*RSp561e*RSe563p - 1./54.*m44*RSp561e*
         RSe453p - 1./54.*m44*RSp561e*RSp563e - 1./54.*m44*RSp561e*
         RSp453e + 1./54.*m44*RSp561e*RS145pe + 1./54.*m44*RSp561e*
         RS156pe + 1./54.*m44*RSp561e*RSe561p + 1./54.*m44*RSp561e*
         RSe451p + 1./108.*m44*pow(RSp561e,2) - 1./54.*m44*RSp451e*
         RS356pe - 1./54.*m44*RSp451e*RSe563p - 1./54.*m44*RSp451e*
         RSp563e + 1./54.*m44*RSp451e*RS156pe + 1./54.*m44*RSp451e*
         RSe561p + 1./54.*m44*RSp451e*RSp561e - 1./54.*m33*pow(
         RS345pe,2) + 1./27.*m33*RS356pe*RS345pe - 1./54.*m33*pow(
         RS356pe,2) + 1./54.*m33*RSe563p*RS345pe - 1./54.*m33*RSe563p*
         RS356pe );

      F13p6x23R +=  + Abeps(m46,mu2)*pi16 * (  - 1./54.*m33*RSe453p*
         RS345pe + 1./54.*m33*RSe453p*RS356pe + 1./54.*m33*RSp563e*
         RS345pe - 1./54.*m33*RSp563e*RS356pe - 1./54.*m33*RSp453e*
         RS345pe + 1./54.*m33*RSp453e*RS356pe + 1./54.*m33*RS145pe*
         RS345pe - 1./54.*m33*RS145pe*RS356pe - 1./54.*m33*RS156pe*
         RS345pe + 1./54.*m33*RS156pe*RS356pe - 1./54.*m33*RSe561p*
         RS345pe + 1./54.*m33*RSe561p*RS356pe + 1./54.*m33*RSe451p*
         RS345pe - 1./54.*m33*RSe451p*RS356pe - 1./54.*m33*RSp561e*
         RS345pe + 1./54.*m33*RSp561e*RS356pe + 1./54.*m33*RSp451e*
         RS345pe - 1./54.*m33*RSp451e*RS356pe + 1./54.*m11*RS145pe*
         RS345pe - 1./54.*m11*RS145pe*RS356pe - 1./54.*m11*RS145pe*
         RSe563p + 1./54.*m11*RS145pe*RSe453p - 1./54.*m11*RS145pe*
         RSp563e + 1./54.*m11*RS145pe*RSp453e - 1./54.*m11*pow(
         RS145pe,2) - 1./54.*m11*RS156pe*RS345pe + 1./54.*m11*RS156pe*
         RS356pe + 1./54.*m11*RS156pe*RSe563p - 1./54.*m11*RS156pe*
         RSe453p );

      F13p6x23R +=  + Abeps(m46,mu2)*pi16 * ( 1./54.*m11*RS156pe*
         RSp563e - 1./54.*m11*RS156pe*RSp453e + 1./27.*m11*RS156pe*
         RS145pe - 1./54.*m11*pow(RS156pe,2) + 1./54.*m11*RSe561p*
         RS145pe - 1./54.*m11*RSe561p*RS156pe - 1./54.*m11*RSe451p*
         RS145pe + 1./54.*m11*RSe451p*RS156pe + 1./54.*m11*RSp561e*
         RS145pe - 1./54.*m11*RSp561e*RS156pe - 1./54.*m11*RSp451e*
         RS145pe + 1./54.*m11*RSp451e*RS156pe );

      F13p6x23R +=  + Abeps(m55,mu2)*pi16 * ( 1./8.*m55 + 1./36.*m55*
         pow(RS346pe,2) + 1./18.*m55*RSe463p*RS346pe + 1./36.*m55*pow(
         RSe463p,2) + 1./18.*m55*RSp463e*RS346pe + 1./18.*m55*RSp463e*
         RSe463p + 1./36.*m55*pow(RSp463e,2) - 1./18.*m55*RS146pe*
         RS346pe - 1./18.*m55*RS146pe*RSe463p - 1./18.*m55*RS146pe*
         RSp463e + 1./36.*m55*pow(RS146pe,2) - 1./24.*m55*RS546pe - 1./
         108.*m55*RS546pe*pow(RS346pe,2) - 1./54.*m55*RS546pe*RSe463p*
         RS346pe - 1./108.*m55*RS546pe*pow(RSe463p,2) - 1./54.*m55*
         RS546pe*RSp463e*RS346pe - 1./54.*m55*RS546pe*RSp463e*RSe463p
          - 1./108.*m55*RS546pe*pow(RSp463e,2) + 1./54.*m55*RS546pe*
         RS146pe*RS346pe + 1./54.*m55*RS546pe*RS146pe*RSe463p + 1./54.*
         m55*RS546pe*RS146pe*RSp463e - 1./108.*m55*RS546pe*pow(
         RS146pe,2) - 1./18.*m55*RSe461p*RS346pe - 1./18.*m55*RSe461p*
         RSe463p - 1./18.*m55*RSe461p*RSp463e + 1./18.*m55*RSe461p*
         RS146pe + 1./54.*m55*RSe461p*RS546pe*RS346pe + 1./54.*m55*
         RSe461p*RS546pe*RSe463p );

      F13p6x23R +=  + Abeps(m55,mu2)*pi16 * ( 1./54.*m55*RSe461p*
         RS546pe*RSp463e - 1./54.*m55*RSe461p*RS546pe*RS146pe + 1./36.*
         m55*pow(RSe461p,2) - 1./108.*m55*pow(RSe461p,2)*RS546pe - 1./
         18.*m55*RSp461e*RS346pe - 1./18.*m55*RSp461e*RSe463p - 1./18.*
         m55*RSp461e*RSp463e + 1./18.*m55*RSp461e*RS146pe + 1./54.*m55*
         RSp461e*RS546pe*RS346pe + 1./54.*m55*RSp461e*RS546pe*RSe463p
          + 1./54.*m55*RSp461e*RS546pe*RSp463e - 1./54.*m55*RSp461e*
         RS546pe*RS146pe + 1./18.*m55*RSp461e*RSe461p - 1./54.*m55*
         RSp461e*RSe461p*RS546pe + 1./36.*m55*pow(RSp461e,2) - 1./108.*
         m55*pow(RSp461e,2)*RS546pe );

      F13p6x23R +=  + Abeps(m56,mu2)*pi16 * ( 1./54.*mp2*RSp463e*
         RS345pe - 1./54.*mp2*RSp463e*RS346pe - 1./54.*mp2*RSp463e*
         RSe463p + 1./54.*mp2*RSp463e*RSe453p - 1./54.*mp2*pow(
         RSp463e,2) - 1./54.*mp2*RSp453e*RS345pe + 1./54.*mp2*RSp453e*
         RS346pe + 1./54.*mp2*RSp453e*RSe463p - 1./54.*mp2*RSp453e*
         RSe453p + 1./27.*mp2*RSp453e*RSp463e - 1./54.*mp2*pow(
         RSp453e,2) - 1./54.*mp2*RS145pe*RSp463e + 1./54.*mp2*RS145pe*
         RSp453e + 1./54.*mp2*RS146pe*RSp463e - 1./54.*mp2*RS146pe*
         RSp453e + 1./54.*mp2*RSe461p*RSp463e - 1./54.*mp2*RSe461p*
         RSp453e - 1./54.*mp2*RSe451p*RSp463e + 1./54.*mp2*RSe451p*
         RSp453e - 1./54.*mp2*RSp461e*RS345pe + 1./54.*mp2*RSp461e*
         RS346pe + 1./54.*mp2*RSp461e*RSe463p - 1./54.*mp2*RSp461e*
         RSe453p + 1./27.*mp2*RSp461e*RSp463e - 1./27.*mp2*RSp461e*
         RSp453e + 1./54.*mp2*RSp461e*RS145pe - 1./54.*mp2*RSp461e*
         RS146pe - 1./54.*mp2*RSp461e*RSe461p + 1./54.*mp2*RSp461e*
         RSe451p );

      F13p6x23R +=  + Abeps(m56,mu2)*pi16 * (  - 1./54.*mp2*pow(
         RSp461e,2) + 1./54.*mp2*RSp451e*RS345pe - 1./54.*mp2*RSp451e*
         RS346pe - 1./54.*mp2*RSp451e*RSe463p + 1./54.*mp2*RSp451e*
         RSe453p - 1./27.*mp2*RSp451e*RSp463e + 1./27.*mp2*RSp451e*
         RSp453e - 1./54.*mp2*RSp451e*RS145pe + 1./54.*mp2*RSp451e*
         RS146pe + 1./54.*mp2*RSp451e*RSe461p - 1./54.*mp2*RSp451e*
         RSe451p + 1./27.*mp2*RSp451e*RSp461e - 1./54.*mp2*pow(
         RSp451e,2) + 1./54.*me2*RSe463p*RS345pe - 1./54.*me2*RSe463p*
         RS346pe - 1./54.*me2*pow(RSe463p,2) - 1./54.*me2*RSe453p*
         RS345pe + 1./54.*me2*RSe453p*RS346pe + 1./27.*me2*RSe453p*
         RSe463p - 1./54.*me2*pow(RSe453p,2) - 1./54.*me2*RSp463e*
         RSe463p + 1./54.*me2*RSp463e*RSe453p + 1./54.*me2*RSp453e*
         RSe463p - 1./54.*me2*RSp453e*RSe453p - 1./54.*me2*RS145pe*
         RSe463p + 1./54.*me2*RS145pe*RSe453p + 1./54.*me2*RS146pe*
         RSe463p - 1./54.*me2*RS146pe*RSe453p - 1./54.*me2*RSe461p*
         RS345pe );

      F13p6x23R +=  + Abeps(m56,mu2)*pi16 * ( 1./54.*me2*RSe461p*
         RS346pe + 1./27.*me2*RSe461p*RSe463p - 1./27.*me2*RSe461p*
         RSe453p + 1./54.*me2*RSe461p*RSp463e - 1./54.*me2*RSe461p*
         RSp453e + 1./54.*me2*RSe461p*RS145pe - 1./54.*me2*RSe461p*
         RS146pe - 1./54.*me2*pow(RSe461p,2) + 1./54.*me2*RSe451p*
         RS345pe - 1./54.*me2*RSe451p*RS346pe - 1./27.*me2*RSe451p*
         RSe463p + 1./27.*me2*RSe451p*RSe453p - 1./54.*me2*RSe451p*
         RSp463e + 1./54.*me2*RSe451p*RSp453e - 1./54.*me2*RSe451p*
         RS145pe + 1./54.*me2*RSe451p*RS146pe + 1./27.*me2*RSe451p*
         RSe461p - 1./54.*me2*pow(RSe451p,2) + 1./54.*me2*RSp461e*
         RSe463p - 1./54.*me2*RSp461e*RSe453p - 1./54.*me2*RSp461e*
         RSe461p + 1./54.*me2*RSp461e*RSe451p - 1./54.*me2*RSp451e*
         RSe463p + 1./54.*me2*RSp451e*RSe453p + 1./54.*me2*RSp451e*
         RSe461p - 1./54.*me2*RSp451e*RSe451p + 1./8.*m66 + 1./108.*m66
         *pow(RS345pe,2) + 1./54.*m66*RS346pe*RS345pe + 1./54.*m66*
         RSe463p*RS345pe );

      F13p6x23R +=  + Abeps(m56,mu2)*pi16 * ( 1./54.*m66*RSe453p*
         RS345pe + 1./54.*m66*RSe453p*RS346pe + 1./54.*m66*RSe453p*
         RSe463p + 1./108.*m66*pow(RSe453p,2) + 1./54.*m66*RSp463e*
         RS345pe + 1./54.*m66*RSp463e*RSe453p + 1./54.*m66*RSp453e*
         RS345pe + 1./54.*m66*RSp453e*RS346pe + 1./54.*m66*RSp453e*
         RSe463p + 1./54.*m66*RSp453e*RSe453p + 1./54.*m66*RSp453e*
         RSp463e + 1./108.*m66*pow(RSp453e,2) - 1./54.*m66*RS145pe*
         RS345pe - 1./54.*m66*RS145pe*RS346pe - 1./54.*m66*RS145pe*
         RSe463p - 1./54.*m66*RS145pe*RSe453p - 1./54.*m66*RS145pe*
         RSp463e - 1./54.*m66*RS145pe*RSp453e + 1./108.*m66*pow(
         RS145pe,2) - 1./54.*m66*RS146pe*RS345pe - 1./54.*m66*RS146pe*
         RSe453p - 1./54.*m66*RS146pe*RSp453e + 1./54.*m66*RS146pe*
         RS145pe - 1./54.*m66*RSe461p*RS345pe - 1./54.*m66*RSe461p*
         RSe453p - 1./54.*m66*RSe461p*RSp453e + 1./54.*m66*RSe461p*
         RS145pe - 1./54.*m66*RSe451p*RS345pe - 1./54.*m66*RSe451p*
         RS346pe );

      F13p6x23R +=  + Abeps(m56,mu2)*pi16 * (  - 1./54.*m66*RSe451p*
         RSe463p - 1./54.*m66*RSe451p*RSe453p - 1./54.*m66*RSe451p*
         RSp463e - 1./54.*m66*RSe451p*RSp453e + 1./54.*m66*RSe451p*
         RS145pe + 1./54.*m66*RSe451p*RS146pe + 1./54.*m66*RSe451p*
         RSe461p + 1./108.*m66*pow(RSe451p,2) - 1./54.*m66*RSp461e*
         RS345pe - 1./54.*m66*RSp461e*RSe453p - 1./54.*m66*RSp461e*
         RSp453e + 1./54.*m66*RSp461e*RS145pe + 1./54.*m66*RSp461e*
         RSe451p - 1./54.*m66*RSp451e*RS345pe - 1./54.*m66*RSp451e*
         RS346pe - 1./54.*m66*RSp451e*RSe463p - 1./54.*m66*RSp451e*
         RSe453p - 1./54.*m66*RSp451e*RSp463e - 1./54.*m66*RSp451e*
         RSp453e + 1./54.*m66*RSp451e*RS145pe + 1./54.*m66*RSp451e*
         RS146pe + 1./54.*m66*RSp451e*RSe461p + 1./54.*m66*RSp451e*
         RSe451p + 1./54.*m66*RSp451e*RSp461e + 1./108.*m66*pow(
         RSp451e,2) + 1./8.*m55 + 1./54.*m55*RS346pe*RS345pe + 1./108.*
         m55*pow(RS346pe,2) + 1./54.*m55*RSe463p*RS345pe + 1./54.*m55*
         RSe463p*RS346pe );

      F13p6x23R +=  + Abeps(m56,mu2)*pi16 * ( 1./108.*m55*pow(
         RSe463p,2) + 1./54.*m55*RSe453p*RS346pe + 1./54.*m55*RSe453p*
         RSe463p + 1./54.*m55*RSp463e*RS345pe + 1./54.*m55*RSp463e*
         RS346pe + 1./54.*m55*RSp463e*RSe463p + 1./54.*m55*RSp463e*
         RSe453p + 1./108.*m55*pow(RSp463e,2) + 1./54.*m55*RSp453e*
         RS346pe + 1./54.*m55*RSp453e*RSe463p + 1./54.*m55*RSp453e*
         RSp463e - 1./54.*m55*RS145pe*RS346pe - 1./54.*m55*RS145pe*
         RSe463p - 1./54.*m55*RS145pe*RSp463e - 1./54.*m55*RS146pe*
         RS345pe - 1./54.*m55*RS146pe*RS346pe - 1./54.*m55*RS146pe*
         RSe463p - 1./54.*m55*RS146pe*RSe453p - 1./54.*m55*RS146pe*
         RSp463e - 1./54.*m55*RS146pe*RSp453e + 1./54.*m55*RS146pe*
         RS145pe + 1./108.*m55*pow(RS146pe,2) - 1./54.*m55*RSe461p*
         RS345pe - 1./54.*m55*RSe461p*RS346pe - 1./54.*m55*RSe461p*
         RSe463p - 1./54.*m55*RSe461p*RSe453p - 1./54.*m55*RSe461p*
         RSp463e - 1./54.*m55*RSe461p*RSp453e + 1./54.*m55*RSe461p*
         RS145pe );

      F13p6x23R +=  + Abeps(m56,mu2)*pi16 * ( 1./54.*m55*RSe461p*
         RS146pe + 1./108.*m55*pow(RSe461p,2) - 1./54.*m55*RSe451p*
         RS346pe - 1./54.*m55*RSe451p*RSe463p - 1./54.*m55*RSe451p*
         RSp463e + 1./54.*m55*RSe451p*RS146pe + 1./54.*m55*RSe451p*
         RSe461p - 1./54.*m55*RSp461e*RS345pe - 1./54.*m55*RSp461e*
         RS346pe - 1./54.*m55*RSp461e*RSe463p - 1./54.*m55*RSp461e*
         RSe453p - 1./54.*m55*RSp461e*RSp463e - 1./54.*m55*RSp461e*
         RSp453e + 1./54.*m55*RSp461e*RS145pe + 1./54.*m55*RSp461e*
         RS146pe + 1./54.*m55*RSp461e*RSe461p + 1./54.*m55*RSp461e*
         RSe451p + 1./108.*m55*pow(RSp461e,2) - 1./54.*m55*RSp451e*
         RS346pe - 1./54.*m55*RSp451e*RSe463p - 1./54.*m55*RSp451e*
         RSp463e + 1./54.*m55*RSp451e*RS146pe + 1./54.*m55*RSp451e*
         RSe461p + 1./54.*m55*RSp451e*RSp461e - 1./54.*m33*pow(
         RS345pe,2) + 1./27.*m33*RS346pe*RS345pe - 1./54.*m33*pow(
         RS346pe,2) + 1./54.*m33*RSe463p*RS345pe - 1./54.*m33*RSe463p*
         RS346pe );

      F13p6x23R +=  + Abeps(m56,mu2)*pi16 * (  - 1./54.*m33*RSe453p*
         RS345pe + 1./54.*m33*RSe453p*RS346pe + 1./54.*m33*RSp463e*
         RS345pe - 1./54.*m33*RSp463e*RS346pe - 1./54.*m33*RSp453e*
         RS345pe + 1./54.*m33*RSp453e*RS346pe + 1./54.*m33*RS145pe*
         RS345pe - 1./54.*m33*RS145pe*RS346pe - 1./54.*m33*RS146pe*
         RS345pe + 1./54.*m33*RS146pe*RS346pe - 1./54.*m33*RSe461p*
         RS345pe + 1./54.*m33*RSe461p*RS346pe + 1./54.*m33*RSe451p*
         RS345pe - 1./54.*m33*RSe451p*RS346pe - 1./54.*m33*RSp461e*
         RS345pe + 1./54.*m33*RSp461e*RS346pe + 1./54.*m33*RSp451e*
         RS345pe - 1./54.*m33*RSp451e*RS346pe + 1./54.*m11*RS145pe*
         RS345pe - 1./54.*m11*RS145pe*RS346pe - 1./54.*m11*RS145pe*
         RSe463p + 1./54.*m11*RS145pe*RSe453p - 1./54.*m11*RS145pe*
         RSp463e + 1./54.*m11*RS145pe*RSp453e - 1./54.*m11*pow(
         RS145pe,2) - 1./54.*m11*RS146pe*RS345pe + 1./54.*m11*RS146pe*
         RS346pe + 1./54.*m11*RS146pe*RSe463p - 1./54.*m11*RS146pe*
         RSe453p );

      F13p6x23R +=  + Abeps(m56,mu2)*pi16 * ( 1./54.*m11*RS146pe*
         RSp463e - 1./54.*m11*RS146pe*RSp453e + 1./27.*m11*RS146pe*
         RS145pe - 1./54.*m11*pow(RS146pe,2) + 1./54.*m11*RSe461p*
         RS145pe - 1./54.*m11*RSe461p*RS146pe - 1./54.*m11*RSe451p*
         RS145pe + 1./54.*m11*RSe451p*RS146pe + 1./54.*m11*RSp461e*
         RS145pe - 1./54.*m11*RSp461e*RS146pe - 1./54.*m11*RSp451e*
         RS145pe + 1./54.*m11*RSp451e*RS146pe );

      F13p6x23R +=  + Abeps(m66,mu2)*pi16 * ( 1./8.*m66 + 1./36.*m66*
         pow(RS345pe,2) + 1./18.*m66*RSe453p*RS345pe + 1./36.*m66*pow(
         RSe453p,2) + 1./18.*m66*RSp453e*RS345pe + 1./18.*m66*RSp453e*
         RSe453p + 1./36.*m66*pow(RSp453e,2) - 1./18.*m66*RS145pe*
         RS345pe - 1./18.*m66*RS145pe*RSe453p - 1./18.*m66*RS145pe*
         RSp453e + 1./36.*m66*pow(RS145pe,2) - 1./24.*m66*RS645pe - 1./
         108.*m66*RS645pe*pow(RS345pe,2) - 1./54.*m66*RS645pe*RSe453p*
         RS345pe - 1./108.*m66*RS645pe*pow(RSe453p,2) - 1./54.*m66*
         RS645pe*RSp453e*RS345pe - 1./54.*m66*RS645pe*RSp453e*RSe453p
          - 1./108.*m66*RS645pe*pow(RSp453e,2) + 1./54.*m66*RS645pe*
         RS145pe*RS345pe + 1./54.*m66*RS645pe*RS145pe*RSe453p + 1./54.*
         m66*RS645pe*RS145pe*RSp453e - 1./108.*m66*RS645pe*pow(
         RS145pe,2) - 1./18.*m66*RSe451p*RS345pe - 1./18.*m66*RSe451p*
         RSe453p - 1./18.*m66*RSe451p*RSp453e + 1./18.*m66*RSe451p*
         RS145pe + 1./54.*m66*RSe451p*RS645pe*RS345pe + 1./54.*m66*
         RSe451p*RS645pe*RSe453p );

      F13p6x23R +=  + Abeps(m66,mu2)*pi16 * ( 1./54.*m66*RSe451p*
         RS645pe*RSp453e - 1./54.*m66*RSe451p*RS645pe*RS145pe + 1./36.*
         m66*pow(RSe451p,2) - 1./108.*m66*pow(RSe451p,2)*RS645pe - 1./
         18.*m66*RSp451e*RS345pe - 1./18.*m66*RSp451e*RSe453p - 1./18.*
         m66*RSp451e*RSp453e + 1./18.*m66*RSp451e*RS145pe + 1./54.*m66*
         RSp451e*RS645pe*RS345pe + 1./54.*m66*RSp451e*RS645pe*RSe453p
          + 1./54.*m66*RSp451e*RS645pe*RSp453e - 1./54.*m66*RSp451e*
         RS645pe*RS145pe + 1./18.*m66*RSp451e*RSe451p - 1./54.*m66*
         RSp451e*RSe451p*RS645pe + 1./36.*m66*pow(RSp451e,2) - 1./108.*
         m66*pow(RSp451e,2)*RS645pe );

      F13p6x23R +=  + Abeps(me2,mu2)*pi16 * (  - 1./72.*RD3456ep*
         RSe45613p + 1./144.*RSe45633p*RD3456ep - 1./72.*RD1456ep*
         RSe45613p + 1./144.*RD1456ep*RSe45633p + 1./144.*RSe45611p*
         RD3456ep + 1./144.*RSe45611p*RD1456ep + 1./36.*mp2*RSp45613e*
         RSe45613p - 1./72.*mp2*RSe45633p*RSp45613e - 1./72.*mp2*
         RSp45633e*RSe45613p + 1./144.*mp2*RSp45633e*RSe45633p - 1./72.
         *mp2*RSe45611p*RSp45613e + 1./144.*mp2*RSe45611p*RSp45633e - 1.
         /72.*mp2*RSp45611e*RSe45613p + 1./144.*mp2*RSp45611e*RSe45633p
          + 1./144.*mp2*RSp45611e*RSe45611p - 1./8.*me2*RSe45613p + 1./
         36.*me2*pow(RSe45613p,2) - 1./12.*me2*RSe45633p - 1./36.*me2*
         RSe45633p*RSe45613p + 1./144.*me2*pow(RSe45633p,2) - 1./36.*
         me2*RSe563p - 1./36.*me2*RSe463p - 1./36.*me2*RSe453p - 1./12.
         *me2*RSe45611p - 1./36.*me2*RSe45611p*RSe45613p + 1./72.*me2*
         RSe45611p*RSe45633p + 1./144.*me2*pow(RSe45611p,2) - 1./36.*
         me2*RSe561p - 1./36.*me2*RSe461p - 1./36.*me2*RSe451p - 7./72.
         *me2*RSe564p );

      F13p6x23R +=  + Abeps(me2,mu2)*pi16 * (  - 7./72.*me2*RSe465p - 7.
         /72.*me2*RSe456p - 1./72.*m66*RSe45613p + 1./144.*m66*
         RSe45633p + 1./108.*m66*pow(RSe563p,2) + 1./108.*m66*pow(
         RSe463p,2) - 1./36.*m66*RSe453p + 1./108.*m66*RSe453p*RSe563p
          + 1./108.*m66*RSe453p*RSe463p + 5./108.*m66*pow(RSe453p,2) + 
         1./54.*m66*RSp563e*RSe563p + 1./108.*m66*RSp563e*RSe453p + 1./
         54.*m66*RSp463e*RSe463p + 1./108.*m66*RSp463e*RSe453p + 1./108.
         *m66*RSp453e*RSe563p + 1./108.*m66*RSp453e*RSe463p + 5./54.*
         m66*RSp453e*RSe453p + 1./144.*m66*RSe45611p - 1./54.*m66*
         RSe561p*RSe563p - 1./108.*m66*RSe561p*RSe453p - 1./54.*m66*
         RSe561p*RSp563e - 1./108.*m66*RSe561p*RSp453e + 1./108.*m66*
         pow(RSe561p,2) - 1./54.*m66*RSe461p*RSe463p - 1./108.*m66*
         RSe461p*RSe453p - 1./54.*m66*RSe461p*RSp463e - 1./108.*m66*
         RSe461p*RSp453e + 1./108.*m66*pow(RSe461p,2) - 1./36.*m66*
         RSe451p - 1./108.*m66*RSe451p*RSe563p - 1./108.*m66*RSe451p*
         RSe463p );

      F13p6x23R +=  + Abeps(me2,mu2)*pi16 * (  - 5./54.*m66*RSe451p*
         RSe453p - 1./108.*m66*RSe451p*RSp563e - 1./108.*m66*RSe451p*
         RSp463e - 5./54.*m66*RSe451p*RSp453e + 1./108.*m66*RSe451p*
         RSe561p + 1./108.*m66*RSe451p*RSe461p + 5./108.*m66*pow(
         RSe451p,2) - 1./54.*m66*RSp561e*RSe563p - 1./108.*m66*RSp561e*
         RSe453p + 1./54.*m66*RSp561e*RSe561p + 1./108.*m66*RSp561e*
         RSe451p - 1./54.*m66*RSp461e*RSe463p - 1./108.*m66*RSp461e*
         RSe453p + 1./54.*m66*RSp461e*RSe461p + 1./108.*m66*RSp461e*
         RSe451p - 1./108.*m66*RSp451e*RSe563p - 1./108.*m66*RSp451e*
         RSe463p - 5./54.*m66*RSp451e*RSe453p + 1./108.*m66*RSp451e*
         RSe561p + 1./108.*m66*RSp451e*RSe461p + 5./54.*m66*RSp451e*
         RSe451p + 1./18.*m66*RSe456p - 1./108.*m66*RSe456p*pow(
         RS345pe,2) - 1./54.*m66*RSe456p*RSe453p*RS345pe - 1./108.*m66*
         RSe456p*pow(RSe453p,2) - 1./54.*m66*RSe456p*RSp453e*RS345pe - 
         1./54.*m66*RSe456p*RSp453e*RSe453p - 1./108.*m66*RSe456p*pow(
         RSp453e,2) );

      F13p6x23R +=  + Abeps(me2,mu2)*pi16 * ( 1./54.*m66*RSe456p*
         RS145pe*RS345pe + 1./54.*m66*RSe456p*RS145pe*RSe453p + 1./54.*
         m66*RSe456p*RS145pe*RSp453e - 1./108.*m66*RSe456p*pow(
         RS145pe,2) + 1./54.*m66*RSe456p*RSe451p*RS345pe + 1./54.*m66*
         RSe456p*RSe451p*RSe453p + 1./54.*m66*RSe456p*RSe451p*RSp453e
          - 1./54.*m66*RSe456p*RSe451p*RS145pe - 1./108.*m66*RSe456p*
         pow(RSe451p,2) + 1./54.*m66*RSe456p*RSp451e*RS345pe + 1./54.*
         m66*RSe456p*RSp451e*RSe453p + 1./54.*m66*RSe456p*RSp451e*
         RSp453e - 1./54.*m66*RSe456p*RSp451e*RS145pe - 1./54.*m66*
         RSe456p*RSp451e*RSe451p - 1./108.*m66*RSe456p*pow(RSp451e,2)
          - 1./72.*m55*RSe45613p + 1./144.*m55*RSe45633p + 1./108.*m55*
         pow(RSe563p,2) - 1./36.*m55*RSe463p + 1./108.*m55*RSe463p*
         RSe563p + 5./108.*m55*pow(RSe463p,2) + 1./108.*m55*RSe453p*
         RSe463p + 1./108.*m55*pow(RSe453p,2) + 1./54.*m55*RSp563e*
         RSe563p + 1./108.*m55*RSp563e*RSe463p + 1./108.*m55*RSp463e*
         RSe563p );

      F13p6x23R +=  + Abeps(me2,mu2)*pi16 * ( 5./54.*m55*RSp463e*
         RSe463p + 1./108.*m55*RSp463e*RSe453p + 1./108.*m55*RSp453e*
         RSe463p + 1./54.*m55*RSp453e*RSe453p + 1./144.*m55*RSe45611p
          - 1./54.*m55*RSe561p*RSe563p - 1./108.*m55*RSe561p*RSe463p - 
         1./54.*m55*RSe561p*RSp563e - 1./108.*m55*RSe561p*RSp463e + 1./
         108.*m55*pow(RSe561p,2) - 1./36.*m55*RSe461p - 1./108.*m55*
         RSe461p*RSe563p - 5./54.*m55*RSe461p*RSe463p - 1./108.*m55*
         RSe461p*RSe453p - 1./108.*m55*RSe461p*RSp563e - 5./54.*m55*
         RSe461p*RSp463e - 1./108.*m55*RSe461p*RSp453e + 1./108.*m55*
         RSe461p*RSe561p + 5./108.*m55*pow(RSe461p,2) - 1./108.*m55*
         RSe451p*RSe463p - 1./54.*m55*RSe451p*RSe453p - 1./108.*m55*
         RSe451p*RSp463e - 1./54.*m55*RSe451p*RSp453e + 1./108.*m55*
         RSe451p*RSe461p + 1./108.*m55*pow(RSe451p,2) - 1./54.*m55*
         RSp561e*RSe563p - 1./108.*m55*RSp561e*RSe463p + 1./54.*m55*
         RSp561e*RSe561p + 1./108.*m55*RSp561e*RSe461p - 1./108.*m55*
         RSp461e*RSe563p );

      F13p6x23R +=  + Abeps(me2,mu2)*pi16 * (  - 5./54.*m55*RSp461e*
         RSe463p - 1./108.*m55*RSp461e*RSe453p + 1./108.*m55*RSp461e*
         RSe561p + 5./54.*m55*RSp461e*RSe461p + 1./108.*m55*RSp461e*
         RSe451p - 1./108.*m55*RSp451e*RSe463p - 1./54.*m55*RSp451e*
         RSe453p + 1./108.*m55*RSp451e*RSe461p + 1./54.*m55*RSp451e*
         RSe451p + 1./18.*m55*RSe465p - 1./108.*m55*RSe465p*pow(
         RS346pe,2) - 1./54.*m55*RSe465p*RSe463p*RS346pe - 1./108.*m55*
         RSe465p*pow(RSe463p,2) - 1./54.*m55*RSe465p*RSp463e*RS346pe - 
         1./54.*m55*RSe465p*RSp463e*RSe463p - 1./108.*m55*RSe465p*pow(
         RSp463e,2) + 1./54.*m55*RSe465p*RS146pe*RS346pe + 1./54.*m55*
         RSe465p*RS146pe*RSe463p + 1./54.*m55*RSe465p*RS146pe*RSp463e
          - 1./108.*m55*RSe465p*pow(RS146pe,2) + 1./54.*m55*RSe465p*
         RSe461p*RS346pe + 1./54.*m55*RSe465p*RSe461p*RSe463p + 1./54.*
         m55*RSe465p*RSe461p*RSp463e - 1./54.*m55*RSe465p*RSe461p*
         RS146pe - 1./108.*m55*RSe465p*pow(RSe461p,2) + 1./54.*m55*
         RSe465p*RSp461e*RS346pe );

      F13p6x23R +=  + Abeps(me2,mu2)*pi16 * ( 1./54.*m55*RSe465p*
         RSp461e*RSe463p + 1./54.*m55*RSe465p*RSp461e*RSp463e - 1./54.*
         m55*RSe465p*RSp461e*RS146pe - 1./54.*m55*RSe465p*RSp461e*
         RSe461p - 1./108.*m55*RSe465p*pow(RSp461e,2) - 1./72.*m44*
         RSe45613p + 1./144.*m44*RSe45633p - 1./36.*m44*RSe563p + 5./
         108.*m44*pow(RSe563p,2) + 1./108.*m44*RSe463p*RSe563p + 1./108.
         *m44*pow(RSe463p,2) + 1./108.*m44*RSe453p*RSe563p + 1./108.*
         m44*pow(RSe453p,2) + 5./54.*m44*RSp563e*RSe563p + 1./108.*m44*
         RSp563e*RSe463p + 1./108.*m44*RSp563e*RSe453p + 1./108.*m44*
         RSp463e*RSe563p + 1./54.*m44*RSp463e*RSe463p + 1./108.*m44*
         RSp453e*RSe563p + 1./54.*m44*RSp453e*RSe453p + 1./144.*m44*
         RSe45611p - 1./36.*m44*RSe561p - 5./54.*m44*RSe561p*RSe563p - 
         1./108.*m44*RSe561p*RSe463p - 1./108.*m44*RSe561p*RSe453p - 5./
         54.*m44*RSe561p*RSp563e - 1./108.*m44*RSe561p*RSp463e - 1./108.
         *m44*RSe561p*RSp453e + 5./108.*m44*pow(RSe561p,2) - 1./108.*
         m44*RSe461p*RSe563p );

      F13p6x23R +=  + Abeps(me2,mu2)*pi16 * (  - 1./54.*m44*RSe461p*
         RSe463p - 1./108.*m44*RSe461p*RSp563e - 1./54.*m44*RSe461p*
         RSp463e + 1./108.*m44*RSe461p*RSe561p + 1./108.*m44*pow(
         RSe461p,2) - 1./108.*m44*RSe451p*RSe563p - 1./54.*m44*RSe451p*
         RSe453p - 1./108.*m44*RSe451p*RSp563e - 1./54.*m44*RSe451p*
         RSp453e + 1./108.*m44*RSe451p*RSe561p + 1./108.*m44*pow(
         RSe451p,2) - 5./54.*m44*RSp561e*RSe563p - 1./108.*m44*RSp561e*
         RSe463p - 1./108.*m44*RSp561e*RSe453p + 5./54.*m44*RSp561e*
         RSe561p + 1./108.*m44*RSp561e*RSe461p + 1./108.*m44*RSp561e*
         RSe451p - 1./108.*m44*RSp461e*RSe563p - 1./54.*m44*RSp461e*
         RSe463p + 1./108.*m44*RSp461e*RSe561p + 1./54.*m44*RSp461e*
         RSe461p - 1./108.*m44*RSp451e*RSe563p - 1./54.*m44*RSp451e*
         RSe453p + 1./108.*m44*RSp451e*RSe561p + 1./54.*m44*RSp451e*
         RSe451p + 1./18.*m44*RSe564p - 1./108.*m44*RSe564p*pow(
         RS356pe,2) - 1./54.*m44*RSe564p*RSe563p*RS356pe - 1./108.*m44*
         RSe564p*pow(RSe563p,2) );

      F13p6x23R +=  + Abeps(me2,mu2)*pi16 * (  - 1./54.*m44*RSe564p*
         RSp563e*RS356pe - 1./54.*m44*RSe564p*RSp563e*RSe563p - 1./108.
         *m44*RSe564p*pow(RSp563e,2) + 1./54.*m44*RSe564p*RS156pe*
         RS356pe + 1./54.*m44*RSe564p*RS156pe*RSe563p + 1./54.*m44*
         RSe564p*RS156pe*RSp563e - 1./108.*m44*RSe564p*pow(RS156pe,2)
          + 1./54.*m44*RSe564p*RSe561p*RS356pe + 1./54.*m44*RSe564p*
         RSe561p*RSe563p + 1./54.*m44*RSe564p*RSe561p*RSp563e - 1./54.*
         m44*RSe564p*RSe561p*RS156pe - 1./108.*m44*RSe564p*pow(
         RSe561p,2) + 1./54.*m44*RSe564p*RSp561e*RS356pe + 1./54.*m44*
         RSe564p*RSp561e*RSe563p + 1./54.*m44*RSe564p*RSp561e*RSp563e
          - 1./54.*m44*RSe564p*RSp561e*RS156pe - 1./54.*m44*RSe564p*
         RSp561e*RSe561p - 1./108.*m44*RSe564p*pow(RSp561e,2) - 1./48.*
         m33*RSe45613p + 5./162.*m33*RSe45613p*RS34561pe + 1./324.*m33*
         RSe45613p*RS14563pe + 1./324.*m33*pow(RSe45613p,2) + 1./324.*
         m33*RSp45613e*RSe45613p + 5./324.*m33*RS3456pee*RSe45613p + 5./
         324.*m33*RS3456ppe*RSe45613p );

      F13p6x23R +=  + Abeps(me2,mu2)*pi16 * ( 5./48.*m33*RSe45633p - 23.
         /324.*m33*RSe45633p*RS34561pe - 37./648.*m33*RSe45633p*
         RS14563pe - 19./324.*m33*RSe45633p*RSe45613p - 37./648.*m33*
         RSe45633p*RSp45613e - 41./648.*m33*RSe45633p*RS3456pee - 41./
         648.*m33*RSe45633p*RS3456ppe + 73./1296.*m33*pow(RSe45633p,2)
          - 1./648.*m33*RSp45633e*RSe45613p + 73./1296.*m33*RSp45633e*
         RSe45633p - 5./324.*m33*RS345pe*RSe45613p + 41./648.*m33*
         RS345pe*RSe45633p - 5./324.*m33*RS346pe*RSe45613p + 41./648.*
         m33*RS346pe*RSe45633p - 5./324.*m33*RS356pe*RSe45613p + 41./
         648.*m33*RS356pe*RSe45633p - 5./144.*m33*RSe563p - 5./144.*m33
         *RSe463p - 5./144.*m33*RSe453p + 1./648.*m33*RS1456pee*
         RSe45613p - 1./1296.*m33*RS1456pee*RSe45633p + 1./648.*m33*
         RS1456ppe*RSe45613p - 1./1296.*m33*RS1456ppe*RSe45633p - 5./
         324.*m33*RSe45611p*RS34561pe - 1./648.*m33*RSe45611p*RS14563pe
          - 1./324.*m33*RSe45611p*RSe45613p - 1./648.*m33*RSe45611p*
         RSp45613e );

      F13p6x23R +=  + Abeps(me2,mu2)*pi16 * (  - 5./648.*m33*RSe45611p*
         RS3456pee - 5./648.*m33*RSe45611p*RS3456ppe + 1./648.*m33*
         RSe45611p*RSe45633p + 1./1296.*m33*RSe45611p*RSp45633e + 5./
         648.*m33*RSe45611p*RS345pe + 5./648.*m33*RSe45611p*RS346pe + 5.
         /648.*m33*RSe45611p*RS356pe - 1./1296.*m33*RSe45611p*RS1456pee
          - 1./1296.*m33*RSe45611p*RS1456ppe + 1./1296.*m33*pow(
         RSe45611p,2) - 1./648.*m33*RSp45611e*RSe45613p + 1./1296.*m33*
         RSp45611e*RSe45633p + 1./1296.*m33*RSp45611e*RSe45611p - 1./
         648.*m33*RS145pe*RSe45613p + 1./1296.*m33*RS145pe*RSe45633p + 
         1./1296.*m33*RS145pe*RSe45611p - 1./648.*m33*RS146pe*RSe45613p
          + 1./1296.*m33*RS146pe*RSe45633p + 1./1296.*m33*RS146pe*
         RSe45611p - 1./648.*m33*RS156pe*RSe45613p + 1./1296.*m33*
         RS156pe*RSe45633p + 1./1296.*m33*RS156pe*RSe45611p + 1./144.*
         m33*RSe561p + 1./144.*m33*RSe461p + 1./144.*m33*RSe451p - 1./
         48.*m11*RSe45613p + 1./324.*m11*RSe45613p*RS34561pe + 5./162.*
         m11*RSe45613p*RS14563pe );

      F13p6x23R +=  + Abeps(me2,mu2)*pi16 * ( 1./324.*m11*pow(
         RSe45613p,2) + 1./324.*m11*RSp45613e*RSe45613p + 1./648.*m11*
         RS3456pee*RSe45613p + 1./648.*m11*RS3456ppe*RSe45613p - 1./648.
         *m11*RSe45633p*RS34561pe - 5./324.*m11*RSe45633p*RS14563pe - 1.
         /324.*m11*RSe45633p*RSe45613p - 1./648.*m11*RSe45633p*
         RSp45613e - 1./1296.*m11*RSe45633p*RS3456pee - 1./1296.*m11*
         RSe45633p*RS3456ppe + 1./1296.*m11*pow(RSe45633p,2) - 1./648.*
         m11*RSp45633e*RSe45613p + 1./1296.*m11*RSp45633e*RSe45633p - 1.
         /648.*m11*RS345pe*RSe45613p + 1./1296.*m11*RS345pe*RSe45633p
          - 1./648.*m11*RS346pe*RSe45613p + 1./1296.*m11*RS346pe*
         RSe45633p - 1./648.*m11*RS356pe*RSe45613p + 1./1296.*m11*
         RS356pe*RSe45633p + 1./144.*m11*RSe563p + 1./144.*m11*RSe463p
          + 1./144.*m11*RSe453p + 5./324.*m11*RS1456pee*RSe45613p - 5./
         648.*m11*RS1456pee*RSe45633p + 5./324.*m11*RS1456ppe*RSe45613p
          - 5./648.*m11*RS1456ppe*RSe45633p + 5./48.*m11*RSe45611p - 37.
         /648.*m11*RSe45611p*RS34561pe );

      F13p6x23R +=  + Abeps(me2,mu2)*pi16 * (  - 23./324.*m11*RSe45611p
         *RS14563pe - 19./324.*m11*RSe45611p*RSe45613p - 37./648.*m11*
         RSe45611p*RSp45613e - 1./1296.*m11*RSe45611p*RS3456pee - 1./
         1296.*m11*RSe45611p*RS3456ppe + 1./648.*m11*RSe45611p*
         RSe45633p + 1./1296.*m11*RSe45611p*RSp45633e + 1./1296.*m11*
         RSe45611p*RS345pe + 1./1296.*m11*RSe45611p*RS346pe + 1./1296.*
         m11*RSe45611p*RS356pe - 41./648.*m11*RSe45611p*RS1456pee - 41./
         648.*m11*RSe45611p*RS1456ppe + 73./1296.*m11*pow(RSe45611p,2)
          - 1./648.*m11*RSp45611e*RSe45613p + 1./1296.*m11*RSp45611e*
         RSe45633p + 73./1296.*m11*RSp45611e*RSe45611p - 5./324.*m11*
         RS145pe*RSe45613p + 5./648.*m11*RS145pe*RSe45633p + 41./648.*
         m11*RS145pe*RSe45611p - 5./324.*m11*RS146pe*RSe45613p + 5./648.
         *m11*RS146pe*RSe45633p + 41./648.*m11*RS146pe*RSe45611p - 5./
         324.*m11*RS156pe*RSe45613p + 5./648.*m11*RS156pe*RSe45633p + 
         41./648.*m11*RS156pe*RSe45611p - 5./144.*m11*RSe561p - 5./144.
         *m11*RSe461p );

      F13p6x23R +=  + Abeps(me2,mu2)*pi16 * (  - 5./144.*m11*RSe451p );

      F13p6x23R +=  + Abeps(mp2,mu2)*pi16 * (  - 1./72.*RD3456ep*
         RSp45613e + 1./144.*RSp45633e*RD3456ep - 1./72.*RD1456ep*
         RSp45613e + 1./144.*RD1456ep*RSp45633e + 1./144.*RSp45611e*
         RD3456ep + 1./144.*RSp45611e*RD1456ep - 1./8.*mp2*RSp45613e + 
         1./36.*mp2*pow(RSp45613e,2) - 1./12.*mp2*RSp45633e - 1./36.*
         mp2*RSp45633e*RSp45613e + 1./144.*mp2*pow(RSp45633e,2) - 1./36.
         *mp2*RSp563e - 1./36.*mp2*RSp463e - 1./36.*mp2*RSp453e - 1./12.
         *mp2*RSp45611e - 1./36.*mp2*RSp45611e*RSp45613e + 1./72.*mp2*
         RSp45611e*RSp45633e + 1./144.*mp2*pow(RSp45611e,2) - 1./36.*
         mp2*RSp561e - 1./36.*mp2*RSp461e - 1./36.*mp2*RSp451e - 7./72.
         *mp2*RSp564e - 7./72.*mp2*RSp465e - 7./72.*mp2*RSp456e + 1./36.
         *me2*RSp45613e*RSe45613p - 1./72.*me2*RSe45633p*RSp45613e - 1./
         72.*me2*RSp45633e*RSe45613p + 1./144.*me2*RSp45633e*RSe45633p
          - 1./72.*me2*RSe45611p*RSp45613e + 1./144.*me2*RSe45611p*
         RSp45633e - 1./72.*me2*RSp45611e*RSe45613p + 1./144.*me2*
         RSp45611e*RSe45633p );

      F13p6x23R +=  + Abeps(mp2,mu2)*pi16 * ( 1./144.*me2*RSp45611e*
         RSe45611p - 1./72.*m66*RSp45613e + 1./144.*m66*RSp45633e + 1./
         108.*m66*pow(RSp563e,2) + 1./108.*m66*pow(RSp463e,2) - 1./36.*
         m66*RSp453e + 1./108.*m66*RSp453e*RSp563e + 1./108.*m66*
         RSp453e*RSp463e + 5./108.*m66*pow(RSp453e,2) + 1./144.*m66*
         RSp45611e - 1./54.*m66*RSp561e*RSp563e - 1./108.*m66*RSp561e*
         RSp453e + 1./108.*m66*pow(RSp561e,2) - 1./54.*m66*RSp461e*
         RSp463e - 1./108.*m66*RSp461e*RSp453e + 1./108.*m66*pow(
         RSp461e,2) - 1./36.*m66*RSp451e - 1./108.*m66*RSp451e*RSp563e
          - 1./108.*m66*RSp451e*RSp463e - 5./54.*m66*RSp451e*RSp453e + 
         1./108.*m66*RSp451e*RSp561e + 1./108.*m66*RSp451e*RSp461e + 5./
         108.*m66*pow(RSp451e,2) + 1./18.*m66*RSp456e - 1./108.*m66*
         RSp456e*pow(RS345pe,2) - 1./54.*m66*RSp456e*RSe453p*RS345pe - 
         1./108.*m66*RSp456e*pow(RSe453p,2) - 1./54.*m66*RSp456e*
         RSp453e*RS345pe - 1./54.*m66*RSp456e*RSp453e*RSe453p - 1./108.
         *m66*RSp456e*pow(RSp453e,2) );

      F13p6x23R +=  + Abeps(mp2,mu2)*pi16 * ( 1./54.*m66*RSp456e*
         RS145pe*RS345pe + 1./54.*m66*RSp456e*RS145pe*RSe453p + 1./54.*
         m66*RSp456e*RS145pe*RSp453e - 1./108.*m66*RSp456e*pow(
         RS145pe,2) + 1./54.*m66*RSp456e*RSe451p*RS345pe + 1./54.*m66*
         RSp456e*RSe451p*RSe453p + 1./54.*m66*RSp456e*RSe451p*RSp453e
          - 1./54.*m66*RSp456e*RSe451p*RS145pe - 1./108.*m66*RSp456e*
         pow(RSe451p,2) + 1./54.*m66*RSp456e*RSp451e*RS345pe + 1./54.*
         m66*RSp456e*RSp451e*RSe453p + 1./54.*m66*RSp456e*RSp451e*
         RSp453e - 1./54.*m66*RSp456e*RSp451e*RS145pe - 1./54.*m66*
         RSp456e*RSp451e*RSe451p - 1./108.*m66*RSp456e*pow(RSp451e,2)
          - 1./72.*m55*RSp45613e + 1./144.*m55*RSp45633e + 1./108.*m55*
         pow(RSp563e,2) - 1./36.*m55*RSp463e + 1./108.*m55*RSp463e*
         RSp563e + 5./108.*m55*pow(RSp463e,2) + 1./108.*m55*RSp453e*
         RSp463e + 1./108.*m55*pow(RSp453e,2) + 1./144.*m55*RSp45611e
          - 1./54.*m55*RSp561e*RSp563e - 1./108.*m55*RSp561e*RSp463e + 
         1./108.*m55*pow(RSp561e,2) );

      F13p6x23R +=  + Abeps(mp2,mu2)*pi16 * (  - 1./36.*m55*RSp461e - 1.
         /108.*m55*RSp461e*RSp563e - 5./54.*m55*RSp461e*RSp463e - 1./
         108.*m55*RSp461e*RSp453e + 1./108.*m55*RSp461e*RSp561e + 5./
         108.*m55*pow(RSp461e,2) - 1./108.*m55*RSp451e*RSp463e - 1./54.
         *m55*RSp451e*RSp453e + 1./108.*m55*RSp451e*RSp461e + 1./108.*
         m55*pow(RSp451e,2) + 1./18.*m55*RSp465e - 1./108.*m55*RSp465e*
         pow(RS346pe,2) - 1./54.*m55*RSp465e*RSe463p*RS346pe - 1./108.*
         m55*RSp465e*pow(RSe463p,2) - 1./54.*m55*RSp465e*RSp463e*
         RS346pe - 1./54.*m55*RSp465e*RSp463e*RSe463p - 1./108.*m55*
         RSp465e*pow(RSp463e,2) + 1./54.*m55*RSp465e*RS146pe*RS346pe + 
         1./54.*m55*RSp465e*RS146pe*RSe463p + 1./54.*m55*RSp465e*
         RS146pe*RSp463e - 1./108.*m55*RSp465e*pow(RS146pe,2) + 1./54.*
         m55*RSp465e*RSe461p*RS346pe + 1./54.*m55*RSp465e*RSe461p*
         RSe463p + 1./54.*m55*RSp465e*RSe461p*RSp463e - 1./54.*m55*
         RSp465e*RSe461p*RS146pe - 1./108.*m55*RSp465e*pow(RSe461p,2)
          + 1./54.*m55*RSp465e*RSp461e*RS346pe );

      F13p6x23R +=  + Abeps(mp2,mu2)*pi16 * ( 1./54.*m55*RSp465e*
         RSp461e*RSe463p + 1./54.*m55*RSp465e*RSp461e*RSp463e - 1./54.*
         m55*RSp465e*RSp461e*RS146pe - 1./54.*m55*RSp465e*RSp461e*
         RSe461p - 1./108.*m55*RSp465e*pow(RSp461e,2) - 1./72.*m44*
         RSp45613e + 1./144.*m44*RSp45633e - 1./36.*m44*RSp563e + 5./
         108.*m44*pow(RSp563e,2) + 1./108.*m44*RSp463e*RSp563e + 1./108.
         *m44*pow(RSp463e,2) + 1./108.*m44*RSp453e*RSp563e + 1./108.*
         m44*pow(RSp453e,2) + 1./144.*m44*RSp45611e - 1./36.*m44*
         RSp561e - 5./54.*m44*RSp561e*RSp563e - 1./108.*m44*RSp561e*
         RSp463e - 1./108.*m44*RSp561e*RSp453e + 5./108.*m44*pow(
         RSp561e,2) - 1./108.*m44*RSp461e*RSp563e - 1./54.*m44*RSp461e*
         RSp463e + 1./108.*m44*RSp461e*RSp561e + 1./108.*m44*pow(
         RSp461e,2) - 1./108.*m44*RSp451e*RSp563e - 1./54.*m44*RSp451e*
         RSp453e + 1./108.*m44*RSp451e*RSp561e + 1./108.*m44*pow(
         RSp451e,2) + 1./18.*m44*RSp564e - 1./108.*m44*RSp564e*pow(
         RS356pe,2) );

      F13p6x23R +=  + Abeps(mp2,mu2)*pi16 * (  - 1./54.*m44*RSp564e*
         RSe563p*RS356pe - 1./108.*m44*RSp564e*pow(RSe563p,2) - 1./54.*
         m44*RSp564e*RSp563e*RS356pe - 1./54.*m44*RSp564e*RSp563e*
         RSe563p - 1./108.*m44*RSp564e*pow(RSp563e,2) + 1./54.*m44*
         RSp564e*RS156pe*RS356pe + 1./54.*m44*RSp564e*RS156pe*RSe563p
          + 1./54.*m44*RSp564e*RS156pe*RSp563e - 1./108.*m44*RSp564e*
         pow(RS156pe,2) + 1./54.*m44*RSp564e*RSe561p*RS356pe + 1./54.*
         m44*RSp564e*RSe561p*RSe563p + 1./54.*m44*RSp564e*RSe561p*
         RSp563e - 1./54.*m44*RSp564e*RSe561p*RS156pe - 1./108.*m44*
         RSp564e*pow(RSe561p,2) + 1./54.*m44*RSp564e*RSp561e*RS356pe + 
         1./54.*m44*RSp564e*RSp561e*RSe563p + 1./54.*m44*RSp564e*
         RSp561e*RSp563e - 1./54.*m44*RSp564e*RSp561e*RS156pe - 1./54.*
         m44*RSp564e*RSp561e*RSe561p - 1./108.*m44*RSp564e*pow(
         RSp561e,2) - 1./48.*m33*RSp45613e + 5./162.*m33*RSp45613e*
         RS34561pe + 1./324.*m33*RSp45613e*RS14563pe + 1./324.*m33*
         RSp45613e*RSe45613p );

      F13p6x23R +=  + Abeps(mp2,mu2)*pi16 * ( 1./324.*m33*pow(
         RSp45613e,2) + 5./324.*m33*RS3456pee*RSp45613e + 5./324.*m33*
         RS3456ppe*RSp45613e - 1./648.*m33*RSe45633p*RSp45613e + 5./48.
         *m33*RSp45633e - 23./324.*m33*RSp45633e*RS34561pe - 37./648.*
         m33*RSp45633e*RS14563pe - 37./648.*m33*RSp45633e*RSe45613p - 
         19./324.*m33*RSp45633e*RSp45613e - 41./648.*m33*RSp45633e*
         RS3456pee - 41./648.*m33*RSp45633e*RS3456ppe + 73./1296.*m33*
         RSp45633e*RSe45633p + 73./1296.*m33*pow(RSp45633e,2) - 5./324.
         *m33*RS345pe*RSp45613e + 41./648.*m33*RS345pe*RSp45633e - 5./
         324.*m33*RS346pe*RSp45613e + 41./648.*m33*RS346pe*RSp45633e - 
         5./324.*m33*RS356pe*RSp45613e + 41./648.*m33*RS356pe*RSp45633e
          - 5./144.*m33*RSp563e - 5./144.*m33*RSp463e - 5./144.*m33*
         RSp453e + 1./648.*m33*RS1456pee*RSp45613e - 1./1296.*m33*
         RS1456pee*RSp45633e + 1./648.*m33*RS1456ppe*RSp45613e - 1./
         1296.*m33*RS1456ppe*RSp45633e - 1./648.*m33*RSe45611p*
         RSp45613e );

      F13p6x23R +=  + Abeps(mp2,mu2)*pi16 * ( 1./1296.*m33*RSe45611p*
         RSp45633e - 5./324.*m33*RSp45611e*RS34561pe - 1./648.*m33*
         RSp45611e*RS14563pe - 1./648.*m33*RSp45611e*RSe45613p - 1./324.
         *m33*RSp45611e*RSp45613e - 5./648.*m33*RSp45611e*RS3456pee - 5.
         /648.*m33*RSp45611e*RS3456ppe + 1./1296.*m33*RSp45611e*
         RSe45633p + 1./648.*m33*RSp45611e*RSp45633e + 5./648.*m33*
         RSp45611e*RS345pe + 5./648.*m33*RSp45611e*RS346pe + 5./648.*
         m33*RSp45611e*RS356pe - 1./1296.*m33*RSp45611e*RS1456pee - 1./
         1296.*m33*RSp45611e*RS1456ppe + 1./1296.*m33*RSp45611e*
         RSe45611p + 1./1296.*m33*pow(RSp45611e,2) - 1./648.*m33*
         RS145pe*RSp45613e + 1./1296.*m33*RS145pe*RSp45633e + 1./1296.*
         m33*RS145pe*RSp45611e - 1./648.*m33*RS146pe*RSp45613e + 1./
         1296.*m33*RS146pe*RSp45633e + 1./1296.*m33*RS146pe*RSp45611e
          - 1./648.*m33*RS156pe*RSp45613e + 1./1296.*m33*RS156pe*
         RSp45633e + 1./1296.*m33*RS156pe*RSp45611e + 1./144.*m33*
         RSp561e );

      F13p6x23R +=  + Abeps(mp2,mu2)*pi16 * ( 1./144.*m33*RSp461e + 1./
         144.*m33*RSp451e - 1./48.*m11*RSp45613e + 1./324.*m11*
         RSp45613e*RS34561pe + 5./162.*m11*RSp45613e*RS14563pe + 1./324.
         *m11*RSp45613e*RSe45613p + 1./324.*m11*pow(RSp45613e,2) + 1./
         648.*m11*RS3456pee*RSp45613e + 1./648.*m11*RS3456ppe*RSp45613e
          - 1./648.*m11*RSe45633p*RSp45613e - 1./648.*m11*RSp45633e*
         RS34561pe - 5./324.*m11*RSp45633e*RS14563pe - 1./648.*m11*
         RSp45633e*RSe45613p - 1./324.*m11*RSp45633e*RSp45613e - 1./
         1296.*m11*RSp45633e*RS3456pee - 1./1296.*m11*RSp45633e*
         RS3456ppe + 1./1296.*m11*RSp45633e*RSe45633p + 1./1296.*m11*
         pow(RSp45633e,2) - 1./648.*m11*RS345pe*RSp45613e + 1./1296.*
         m11*RS345pe*RSp45633e - 1./648.*m11*RS346pe*RSp45613e + 1./
         1296.*m11*RS346pe*RSp45633e - 1./648.*m11*RS356pe*RSp45613e + 
         1./1296.*m11*RS356pe*RSp45633e + 1./144.*m11*RSp563e + 1./144.
         *m11*RSp463e + 1./144.*m11*RSp453e + 5./324.*m11*RS1456pee*
         RSp45613e );

      F13p6x23R +=  + Abeps(mp2,mu2)*pi16 * (  - 5./648.*m11*RS1456pee*
         RSp45633e + 5./324.*m11*RS1456ppe*RSp45613e - 5./648.*m11*
         RS1456ppe*RSp45633e - 1./648.*m11*RSe45611p*RSp45613e + 1./
         1296.*m11*RSe45611p*RSp45633e + 5./48.*m11*RSp45611e - 37./648.
         *m11*RSp45611e*RS34561pe - 23./324.*m11*RSp45611e*RS14563pe - 
         37./648.*m11*RSp45611e*RSe45613p - 19./324.*m11*RSp45611e*
         RSp45613e - 1./1296.*m11*RSp45611e*RS3456pee - 1./1296.*m11*
         RSp45611e*RS3456ppe + 1./1296.*m11*RSp45611e*RSe45633p + 1./
         648.*m11*RSp45611e*RSp45633e + 1./1296.*m11*RSp45611e*RS345pe
          + 1./1296.*m11*RSp45611e*RS346pe + 1./1296.*m11*RSp45611e*
         RS356pe - 41./648.*m11*RSp45611e*RS1456pee - 41./648.*m11*
         RSp45611e*RS1456ppe + 73./1296.*m11*RSp45611e*RSe45611p + 73./
         1296.*m11*pow(RSp45611e,2) - 5./324.*m11*RS145pe*RSp45613e + 5.
         /648.*m11*RS145pe*RSp45633e + 41./648.*m11*RS145pe*RSp45611e
          - 5./324.*m11*RS146pe*RSp45613e + 5./648.*m11*RS146pe*
         RSp45633e );

      F13p6x23R +=  + Abeps(mp2,mu2)*pi16 * ( 41./648.*m11*RS146pe*
         RSp45611e - 5./324.*m11*RS156pe*RSp45613e + 5./648.*m11*
         RS156pe*RSp45633e + 41./648.*m11*RS156pe*RSp45611e - 5./144.*
         m11*RSp561e - 5./144.*m11*RSp461e - 5./144.*m11*RSp451e );

      F13p6x23R +=  + Bbeps(m11,mu2)*pi16 * ( 1./144.*RD1456ep*RD3456ep
          + 1./144.*pow(RD1456ep,2) - 1./72.*mp2*RD1456ep*RSp45613e + 1.
         /144.*mp2*RD1456ep*RSp45633e + 1./144.*mp2*RSp45611e*RD1456ep
          - 1./72.*me2*RD1456ep*RSe45613p + 1./144.*me2*RD1456ep*
         RSe45633p + 1./144.*me2*RSe45611p*RD1456ep - 7./144.*m66*
         RD1456ep - 1./108.*m66*mp2*RSp456e*pow(RS145pe,2) - 1./108.*
         m66*me2*RSe456p*pow(RS145pe,2) - 1./36.*pow(m66,2)*RS145pe - 1.
         /324.*pow(m66,2)*pow(RS145pe,2) + 1./108.*pow(m66,2)*RS146pe*
         RS145pe + 1./108.*pow(m66,2)*RS156pe*RS145pe - 1./108.*pow(
         m66,2)*RS645pe*pow(RS145pe,2) - 7./144.*m55*RD1456ep - 1./108.
         *m55*mp2*RSp465e*pow(RS146pe,2) - 1./108.*m55*me2*RSe465p*pow(
         RS146pe,2) - 1./81.*m55*m66*pow(RS145pe,2) + 1./54.*m55*m66*
         RS146pe*RS145pe - 1./81.*m55*m66*pow(RS146pe,2) - 1./36.*pow(
         m55,2)*RS146pe + 1./108.*pow(m55,2)*RS146pe*RS145pe - 1./324.*
         pow(m55,2)*pow(RS146pe,2) + 1./108.*pow(m55,2)*RS156pe*RS146pe
          - 1./108.*pow(m55,2)*RS546pe*pow(RS146pe,2) );

      F13p6x23R +=  + Bbeps(m11,mu2)*pi16 * (  - 7./144.*m44*RD1456ep
          - 1./108.*m44*mp2*RSp564e*pow(RS156pe,2) - 1./108.*m44*me2*
         RSe564p*pow(RS156pe,2) - 1./81.*m44*m66*pow(RS145pe,2) + 1./54.
         *m44*m66*RS156pe*RS145pe - 1./81.*m44*m66*pow(RS156pe,2) - 1./
         81.*m44*m55*pow(RS146pe,2) + 1./54.*m44*m55*RS156pe*RS146pe - 
         1./81.*m44*m55*pow(RS156pe,2) - 1./36.*pow(m44,2)*RS156pe + 1./
         108.*pow(m44,2)*RS156pe*RS145pe + 1./108.*pow(m44,2)*RS156pe*
         RS146pe - 1./324.*pow(m44,2)*pow(RS156pe,2) - 1./108.*pow(
         m44,2)*RS456pe*pow(RS156pe,2) - 5./324.*m33*RD1456ep*RS34561pe
          - 1./648.*m33*RD1456ep*RS14563pe - 1./648.*m33*RD1456ep*
         RSe45613p - 1./648.*m33*RD1456ep*RSp45613e - 5./648.*m33*
         RD1456ep*RS3456pee - 5./648.*m33*RD1456ep*RS3456ppe + 1./1296.
         *m33*RD1456ep*RSe45633p + 1./1296.*m33*RD1456ep*RSp45633e + 5./
         648.*m33*RD1456ep*RS345pe + 5./648.*m33*RD1456ep*RS346pe + 5./
         648.*m33*RD1456ep*RS356pe - 1./1296.*m33*RS1456pee*RD1456ep - 
         1./1296.*m33*RS1456ppe*RD1456ep );

      F13p6x23R +=  + Bbeps(m11,mu2)*pi16 * ( 1./1296.*m33*RSe45611p*
         RD1456ep + 1./1296.*m33*RSp45611e*RD1456ep + 1./1296.*m33*
         RS145pe*RD1456ep + 1./1296.*m33*RS146pe*RD1456ep + 1./1296.*
         m33*RS156pe*RD1456ep - 7./48.*m11*RD1456ep - 37./648.*m11*
         RD1456ep*RS34561pe - 41./324.*m11*RD1456ep*RS14563pe - 37./648.
         *m11*RD1456ep*RSe45613p - 37./648.*m11*RD1456ep*RSp45613e - 1./
         1296.*m11*RD1456ep*RS3456pee - 1./1296.*m11*RD1456ep*RS3456ppe
          + 1./1296.*m11*RD1456ep*RSe45633p + 1./1296.*m11*RD1456ep*
         RSp45633e + 1./1296.*m11*RD1456ep*RS345pe + 1./1296.*m11*
         RD1456ep*RS346pe + 1./1296.*m11*RD1456ep*RS356pe - 77./648.*
         m11*RS1456pee*RD1456ep - 77./648.*m11*RS1456ppe*RD1456ep + 73./
         1296.*m11*RSe45611p*RD1456ep + 73./1296.*m11*RSp45611e*
         RD1456ep + 77./648.*m11*RS145pe*RD1456ep + 77./648.*m11*
         RS146pe*RD1456ep + 77./648.*m11*RS156pe*RD1456ep - 1./18.*m11*
         mp2*RSp45611e*RS14563pe - 1./18.*m11*mp2*RSp45611e*RS1456pee
          - 1./18.*m11*mp2*RSp45611e*RS1456ppe );

      F13p6x23R +=  + Bbeps(m11,mu2)*pi16 * ( 1./18.*m11*mp2*RS145pe*
         RSp45611e + 1./18.*m11*mp2*RS146pe*RSp45611e + 1./18.*m11*mp2*
         RS156pe*RSp45611e - 1./18.*m11*me2*RSe45611p*RS14563pe - 1./18.
         *m11*me2*RSe45611p*RS1456pee - 1./18.*m11*me2*RSe45611p*
         RS1456ppe + 1./18.*m11*me2*RS145pe*RSe45611p + 1./18.*m11*me2*
         RS146pe*RSe45611p + 1./18.*m11*me2*RS156pe*RSe45611p - 1./54.*
         m11*m66*RS14563pe - 1./54.*m11*m66*RS1456pee - 1./54.*m11*m66*
         RS1456ppe - 7./108.*m11*m66*RS145pe + 5./108.*m11*m66*pow(
         RS145pe,2) + 1./54.*m11*m66*RS146pe + 1./108.*m11*m66*RS146pe*
         RS145pe + 1./108.*m11*m66*pow(RS146pe,2) + 1./54.*m11*m66*
         RS156pe + 1./108.*m11*m66*RS156pe*RS145pe + 1./108.*m11*m66*
         pow(RS156pe,2) - 1./54.*m11*m55*RS14563pe - 1./54.*m11*m55*
         RS1456pee - 1./54.*m11*m55*RS1456ppe + 1./54.*m11*m55*RS145pe
          + 1./108.*m11*m55*pow(RS145pe,2) - 7./108.*m11*m55*RS146pe + 
         1./108.*m11*m55*RS146pe*RS145pe + 5./108.*m11*m55*pow(
         RS146pe,2) );

      F13p6x23R +=  + Bbeps(m11,mu2)*pi16 * ( 1./54.*m11*m55*RS156pe + 
         1./108.*m11*m55*RS156pe*RS146pe + 1./108.*m11*m55*pow(
         RS156pe,2) - 1./54.*m11*m44*RS14563pe - 1./54.*m11*m44*
         RS1456pee - 1./54.*m11*m44*RS1456ppe + 1./54.*m11*m44*RS145pe
          + 1./108.*m11*m44*pow(RS145pe,2) + 1./54.*m11*m44*RS146pe + 1.
         /108.*m11*m44*pow(RS146pe,2) - 7./108.*m11*m44*RS156pe + 1./
         108.*m11*m44*RS156pe*RS145pe + 1./108.*m11*m44*RS156pe*RS146pe
          + 5./108.*m11*m44*pow(RS156pe,2) + 1./4.*pow(m11,2) + 1./9.*
         pow(m11,2)*RS14563pe + 1./9.*pow(m11,2)*RS1456pee + 1./18.*
         pow(m11,2)*RS1456pee*RS14563pe + 1./18.*pow(m11,2)*pow(
         RS1456pee,2) + 1./9.*pow(m11,2)*RS1456ppe + 1./18.*pow(m11,2)*
         RS1456ppe*RS14563pe + 1./9.*pow(m11,2)*RS1456ppe*RS1456pee + 1.
         /18.*pow(m11,2)*pow(RS1456ppe,2) - 1./6.*pow(m11,2)*RS145pe - 
         1./18.*pow(m11,2)*RS145pe*RS14563pe - 1./9.*pow(m11,2)*RS145pe
         *RS1456pee - 1./9.*pow(m11,2)*RS145pe*RS1456ppe + 1./18.*pow(
         m11,2)*pow(RS145pe,2) );

      F13p6x23R +=  + Bbeps(m11,mu2)*pi16 * (  - 1./6.*pow(m11,2)*
         RS146pe - 1./18.*pow(m11,2)*RS146pe*RS14563pe - 1./9.*pow(
         m11,2)*RS146pe*RS1456pee - 1./9.*pow(m11,2)*RS146pe*RS1456ppe
          + 1./9.*pow(m11,2)*RS146pe*RS145pe + 1./18.*pow(m11,2)*pow(
         RS146pe,2) - 1./6.*pow(m11,2)*RS156pe - 1./18.*pow(m11,2)*
         RS156pe*RS14563pe - 1./9.*pow(m11,2)*RS156pe*RS1456pee - 1./9.
         *pow(m11,2)*RS156pe*RS1456ppe + 1./9.*pow(m11,2)*RS156pe*
         RS145pe + 1./9.*pow(m11,2)*RS156pe*RS146pe + 1./18.*pow(m11,2)
         *pow(RS156pe,2) );

      F13p6x23R +=  + Bbeps(m14,mu2)*pi16 * ( 1./72.*pow(mp2,2)*
         RSp45611e - 1./36.*pow(mp2,2)*RSp561e + 1./72.*pow(mp2,2)*
         RSp564e + 1./72.*pow(me2,2)*RSe45611p - 1./36.*pow(me2,2)*
         RSe561p + 1./72.*pow(me2,2)*RSe564p - 1./36.*m44*mp2*RSp561e
          - 1./72.*m44*mp2*RSp564e - 1./36.*m44*me2*RSe561p - 1./72.*
         m44*me2*RSe564p - 1./72.*m44*m66 - 1./72.*m44*m55 + 1./36.*
         pow(m44,2) + 1./72.*m11*RD1456ep - 1./72.*m11*mp2*RSp45611e - 
         1./36.*m11*mp2*RSp561e - 1./72.*m11*me2*RSe45611p - 1./36.*m11
         *me2*RSe561p - 1./72.*m11*m66 - 1./72.*m11*m55 + 5./72.*m11*
         m44 - 1./36.*m11*m44*RS156pe + 1./24.*pow(m11,2) - 1./18.*pow(
         m11,2)*RS156pe );

      F13p6x23R +=  + Bbeps(m15,mu2)*pi16 * ( 1./72.*pow(mp2,2)*
         RSp45611e - 1./36.*pow(mp2,2)*RSp461e + 1./72.*pow(mp2,2)*
         RSp465e + 1./72.*pow(me2,2)*RSe45611p - 1./36.*pow(me2,2)*
         RSe461p + 1./72.*pow(me2,2)*RSe465p - 1./36.*m55*mp2*RSp461e
          - 1./72.*m55*mp2*RSp465e - 1./36.*m55*me2*RSe461p - 1./72.*
         m55*me2*RSe465p - 1./72.*m55*m66 + 1./36.*pow(m55,2) - 1./72.*
         m44*m55 + 1./72.*m11*RD1456ep - 1./72.*m11*mp2*RSp45611e - 1./
         36.*m11*mp2*RSp461e - 1./72.*m11*me2*RSe45611p - 1./36.*m11*
         me2*RSe461p - 1./72.*m11*m66 + 5./72.*m11*m55 - 1./36.*m11*m55
         *RS146pe - 1./72.*m11*m44 + 1./24.*pow(m11,2) - 1./18.*pow(
         m11,2)*RS146pe );

      F13p6x23R +=  + Bbeps(m16,mu2)*pi16 * ( 1./72.*pow(mp2,2)*
         RSp45611e - 1./36.*pow(mp2,2)*RSp451e + 1./72.*pow(mp2,2)*
         RSp456e + 1./72.*pow(me2,2)*RSe45611p - 1./36.*pow(me2,2)*
         RSe451p + 1./72.*pow(me2,2)*RSe456p - 1./36.*m66*mp2*RSp451e
          - 1./72.*m66*mp2*RSp456e - 1./36.*m66*me2*RSe451p - 1./72.*
         m66*me2*RSe456p + 1./36.*pow(m66,2) - 1./72.*m55*m66 - 1./72.*
         m44*m66 + 1./72.*m11*RD1456ep - 1./72.*m11*mp2*RSp45611e - 1./
         36.*m11*mp2*RSp451e - 1./72.*m11*me2*RSe45611p - 1./36.*m11*
         me2*RSe451p + 5./72.*m11*m66 - 1./36.*m11*m66*RS145pe - 1./72.
         *m11*m55 - 1./72.*m11*m44 + 1./24.*pow(m11,2) - 1./18.*pow(
         m11,2)*RS145pe );

      F13p6x23R +=  + Bbeps(m33,mu2)*pi16 * ( 1./144.*pow(RD3456ep,2)
          + 1./144.*RD1456ep*RD3456ep - 1./72.*mp2*RD3456ep*RSp45613e
          + 1./144.*mp2*RSp45633e*RD3456ep + 1./144.*mp2*RSp45611e*
         RD3456ep - 1./72.*me2*RD3456ep*RSe45613p + 1./144.*me2*
         RSe45633p*RD3456ep + 1./144.*me2*RSe45611p*RD3456ep - 7./144.*
         m66*RD3456ep - 1./108.*m66*mp2*RSp456e*pow(RS345pe,2) - 1./108.
         *m66*me2*RSe456p*pow(RS345pe,2) - 1./36.*pow(m66,2)*RS345pe - 
         1./324.*pow(m66,2)*pow(RS345pe,2) + 1./108.*pow(m66,2)*RS346pe
         *RS345pe + 1./108.*pow(m66,2)*RS356pe*RS345pe - 1./108.*pow(
         m66,2)*RS645pe*pow(RS345pe,2) - 7./144.*m55*RD3456ep - 1./108.
         *m55*mp2*RSp465e*pow(RS346pe,2) - 1./108.*m55*me2*RSe465p*pow(
         RS346pe,2) - 1./81.*m55*m66*pow(RS345pe,2) + 1./54.*m55*m66*
         RS346pe*RS345pe - 1./81.*m55*m66*pow(RS346pe,2) - 1./36.*pow(
         m55,2)*RS346pe + 1./108.*pow(m55,2)*RS346pe*RS345pe - 1./324.*
         pow(m55,2)*pow(RS346pe,2) + 1./108.*pow(m55,2)*RS356pe*RS346pe
          - 1./108.*pow(m55,2)*RS546pe*pow(RS346pe,2) );

      F13p6x23R +=  + Bbeps(m33,mu2)*pi16 * (  - 7./144.*m44*RD3456ep
          - 1./108.*m44*mp2*RSp564e*pow(RS356pe,2) - 1./108.*m44*me2*
         RSe564p*pow(RS356pe,2) - 1./81.*m44*m66*pow(RS345pe,2) + 1./54.
         *m44*m66*RS356pe*RS345pe - 1./81.*m44*m66*pow(RS356pe,2) - 1./
         81.*m44*m55*pow(RS346pe,2) + 1./54.*m44*m55*RS356pe*RS346pe - 
         1./81.*m44*m55*pow(RS356pe,2) - 1./36.*pow(m44,2)*RS356pe + 1./
         108.*pow(m44,2)*RS356pe*RS345pe + 1./108.*pow(m44,2)*RS356pe*
         RS346pe - 1./324.*pow(m44,2)*pow(RS356pe,2) - 1./108.*pow(
         m44,2)*RS456pe*pow(RS356pe,2) - 7./48.*m33*RD3456ep - 41./324.
         *m33*RD3456ep*RS34561pe - 37./648.*m33*RD3456ep*RS14563pe - 37.
         /648.*m33*RD3456ep*RSe45613p - 37./648.*m33*RD3456ep*RSp45613e
          - 77./648.*m33*RS3456pee*RD3456ep - 77./648.*m33*RS3456ppe*
         RD3456ep + 73./1296.*m33*RSe45633p*RD3456ep + 73./1296.*m33*
         RSp45633e*RD3456ep + 77./648.*m33*RS345pe*RD3456ep + 77./648.*
         m33*RS346pe*RD3456ep + 77./648.*m33*RS356pe*RD3456ep - 1./1296.
         *m33*RS1456pee*RD3456ep );

      F13p6x23R +=  + Bbeps(m33,mu2)*pi16 * (  - 1./1296.*m33*RS1456ppe
         *RD3456ep + 1./1296.*m33*RSe45611p*RD3456ep + 1./1296.*m33*
         RSp45611e*RD3456ep + 1./1296.*m33*RS145pe*RD3456ep + 1./1296.*
         m33*RS146pe*RD3456ep + 1./1296.*m33*RS156pe*RD3456ep - 1./18.*
         m33*mp2*RSp45633e*RS34561pe - 1./18.*m33*mp2*RSp45633e*
         RS3456pee - 1./18.*m33*mp2*RSp45633e*RS3456ppe + 1./18.*m33*
         mp2*RS345pe*RSp45633e + 1./18.*m33*mp2*RS346pe*RSp45633e + 1./
         18.*m33*mp2*RS356pe*RSp45633e - 1./18.*m33*me2*RSe45633p*
         RS34561pe - 1./18.*m33*me2*RSe45633p*RS3456pee - 1./18.*m33*
         me2*RSe45633p*RS3456ppe + 1./18.*m33*me2*RS345pe*RSe45633p + 1.
         /18.*m33*me2*RS346pe*RSe45633p + 1./18.*m33*me2*RS356pe*
         RSe45633p - 1./54.*m33*m66*RS34561pe - 1./54.*m33*m66*
         RS3456pee - 1./54.*m33*m66*RS3456ppe - 7./108.*m33*m66*RS345pe
          + 5./108.*m33*m66*pow(RS345pe,2) + 1./54.*m33*m66*RS346pe + 1.
         /108.*m33*m66*RS346pe*RS345pe + 1./108.*m33*m66*pow(RS346pe,2)
          + 1./54.*m33*m66*RS356pe );

      F13p6x23R +=  + Bbeps(m33,mu2)*pi16 * ( 1./108.*m33*m66*RS356pe*
         RS345pe + 1./108.*m33*m66*pow(RS356pe,2) - 1./54.*m33*m55*
         RS34561pe - 1./54.*m33*m55*RS3456pee - 1./54.*m33*m55*
         RS3456ppe + 1./54.*m33*m55*RS345pe + 1./108.*m33*m55*pow(
         RS345pe,2) - 7./108.*m33*m55*RS346pe + 1./108.*m33*m55*RS346pe
         *RS345pe + 5./108.*m33*m55*pow(RS346pe,2) + 1./54.*m33*m55*
         RS356pe + 1./108.*m33*m55*RS356pe*RS346pe + 1./108.*m33*m55*
         pow(RS356pe,2) - 1./54.*m33*m44*RS34561pe - 1./54.*m33*m44*
         RS3456pee - 1./54.*m33*m44*RS3456ppe + 1./54.*m33*m44*RS345pe
          + 1./108.*m33*m44*pow(RS345pe,2) + 1./54.*m33*m44*RS346pe + 1.
         /108.*m33*m44*pow(RS346pe,2) - 7./108.*m33*m44*RS356pe + 1./
         108.*m33*m44*RS356pe*RS345pe + 1./108.*m33*m44*RS356pe*RS346pe
          + 5./108.*m33*m44*pow(RS356pe,2) + 1./4.*pow(m33,2) + 1./9.*
         pow(m33,2)*RS34561pe + 1./9.*pow(m33,2)*RS3456pee + 1./18.*
         pow(m33,2)*RS3456pee*RS34561pe + 1./18.*pow(m33,2)*pow(
         RS3456pee,2) );

      F13p6x23R +=  + Bbeps(m33,mu2)*pi16 * ( 1./9.*pow(m33,2)*
         RS3456ppe + 1./18.*pow(m33,2)*RS3456ppe*RS34561pe + 1./9.*pow(
         m33,2)*RS3456ppe*RS3456pee + 1./18.*pow(m33,2)*pow(
         RS3456ppe,2) - 1./6.*pow(m33,2)*RS345pe - 1./18.*pow(m33,2)*
         RS345pe*RS34561pe - 1./9.*pow(m33,2)*RS345pe*RS3456pee - 1./9.
         *pow(m33,2)*RS345pe*RS3456ppe + 1./18.*pow(m33,2)*pow(
         RS345pe,2) - 1./6.*pow(m33,2)*RS346pe - 1./18.*pow(m33,2)*
         RS346pe*RS34561pe - 1./9.*pow(m33,2)*RS346pe*RS3456pee - 1./9.
         *pow(m33,2)*RS346pe*RS3456ppe + 1./9.*pow(m33,2)*RS346pe*
         RS345pe + 1./18.*pow(m33,2)*pow(RS346pe,2) - 1./6.*pow(m33,2)*
         RS356pe - 1./18.*pow(m33,2)*RS356pe*RS34561pe - 1./9.*pow(
         m33,2)*RS356pe*RS3456pee - 1./9.*pow(m33,2)*RS356pe*RS3456ppe
          + 1./9.*pow(m33,2)*RS356pe*RS345pe + 1./9.*pow(m33,2)*RS356pe
         *RS346pe + 1./18.*pow(m33,2)*pow(RS356pe,2) - 1./648.*m11*
         RD3456ep*RS34561pe - 5./324.*m11*RD3456ep*RS14563pe - 1./648.*
         m11*RD3456ep*RSe45613p );

      F13p6x23R +=  + Bbeps(m33,mu2)*pi16 * (  - 1./648.*m11*RD3456ep*
         RSp45613e - 1./1296.*m11*RS3456pee*RD3456ep - 1./1296.*m11*
         RS3456ppe*RD3456ep + 1./1296.*m11*RSe45633p*RD3456ep + 1./1296.
         *m11*RSp45633e*RD3456ep + 1./1296.*m11*RS345pe*RD3456ep + 1./
         1296.*m11*RS346pe*RD3456ep + 1./1296.*m11*RS356pe*RD3456ep - 5.
         /648.*m11*RS1456pee*RD3456ep - 5./648.*m11*RS1456ppe*RD3456ep
          + 1./1296.*m11*RSe45611p*RD3456ep + 1./1296.*m11*RSp45611e*
         RD3456ep + 5./648.*m11*RS145pe*RD3456ep + 5./648.*m11*RS146pe*
         RD3456ep + 5./648.*m11*RS156pe*RD3456ep );

      F13p6x23R +=  + Bbeps(m34,mu2)*pi16 * ( 1./72.*pow(mp2,2)*
         RSp45633e - 1./36.*pow(mp2,2)*RSp563e + 1./72.*pow(mp2,2)*
         RSp564e + 1./72.*pow(me2,2)*RSe45633p - 1./36.*pow(me2,2)*
         RSe563p + 1./72.*pow(me2,2)*RSe564p - 1./36.*m44*mp2*RSp563e
          - 1./72.*m44*mp2*RSp564e - 1./36.*m44*me2*RSe563p - 1./72.*
         m44*me2*RSe564p - 1./72.*m44*m66 - 1./72.*m44*m55 + 1./36.*
         pow(m44,2) + 1./72.*m33*RD3456ep - 1./72.*m33*mp2*RSp45633e - 
         1./36.*m33*mp2*RSp563e - 1./72.*m33*me2*RSe45633p - 1./36.*m33
         *me2*RSe563p - 1./72.*m33*m66 - 1./72.*m33*m55 + 5./72.*m33*
         m44 - 1./36.*m33*m44*RS356pe + 1./24.*pow(m33,2) - 1./18.*pow(
         m33,2)*RS356pe );

      F13p6x23R +=  + Bbeps(m35,mu2)*pi16 * ( 1./72.*pow(mp2,2)*
         RSp45633e - 1./36.*pow(mp2,2)*RSp463e + 1./72.*pow(mp2,2)*
         RSp465e + 1./72.*pow(me2,2)*RSe45633p - 1./36.*pow(me2,2)*
         RSe463p + 1./72.*pow(me2,2)*RSe465p - 1./36.*m55*mp2*RSp463e
          - 1./72.*m55*mp2*RSp465e - 1./36.*m55*me2*RSe463p - 1./72.*
         m55*me2*RSe465p - 1./72.*m55*m66 + 1./36.*pow(m55,2) - 1./72.*
         m44*m55 + 1./72.*m33*RD3456ep - 1./72.*m33*mp2*RSp45633e - 1./
         36.*m33*mp2*RSp463e - 1./72.*m33*me2*RSe45633p - 1./36.*m33*
         me2*RSe463p - 1./72.*m33*m66 + 5./72.*m33*m55 - 1./36.*m33*m55
         *RS346pe - 1./72.*m33*m44 + 1./24.*pow(m33,2) - 1./18.*pow(
         m33,2)*RS346pe );

      F13p6x23R +=  + Bbeps(m36,mu2)*pi16 * ( 1./72.*pow(mp2,2)*
         RSp45633e - 1./36.*pow(mp2,2)*RSp453e + 1./72.*pow(mp2,2)*
         RSp456e + 1./72.*pow(me2,2)*RSe45633p - 1./36.*pow(me2,2)*
         RSe453p + 1./72.*pow(me2,2)*RSe456p - 1./36.*m66*mp2*RSp453e
          - 1./72.*m66*mp2*RSp456e - 1./36.*m66*me2*RSe453p - 1./72.*
         m66*me2*RSe456p + 1./36.*pow(m66,2) - 1./72.*m55*m66 - 1./72.*
         m44*m66 + 1./72.*m33*RD3456ep - 1./72.*m33*mp2*RSp45633e - 1./
         36.*m33*mp2*RSp453e - 1./72.*m33*me2*RSe45633p - 1./36.*m33*
         me2*RSe453p + 5./72.*m33*m66 - 1./36.*m33*m66*RS345pe - 1./72.
         *m33*m55 - 1./72.*m33*m44 + 1./24.*pow(m33,2) - 1./18.*pow(
         m33,2)*RS345pe );

      F13p6x23R +=  + Bbeps(me2,mu2)*pi16 * (  - 1./108.*m66*mp2*
         RSp456e*pow(RSe453p,2) + 1./54.*m66*mp2*RSp456e*RSe451p*
         RSe453p - 1./108.*m66*mp2*RSp456e*pow(RSe451p,2) + 1./108.*m66
         *me2*pow(RSe563p,2) + 1./108.*m66*me2*pow(RSe463p,2) + 1./108.
         *m66*me2*RSe453p*RSe563p + 1./108.*m66*me2*RSe453p*RSe463p + 5.
         /108.*m66*me2*pow(RSe453p,2) - 1./54.*m66*me2*RSe561p*RSe563p
          - 1./108.*m66*me2*RSe561p*RSe453p + 1./108.*m66*me2*pow(
         RSe561p,2) - 1./54.*m66*me2*RSe461p*RSe463p - 1./108.*m66*me2*
         RSe461p*RSe453p + 1./108.*m66*me2*pow(RSe461p,2) - 1./108.*m66
         *me2*RSe451p*RSe563p - 1./108.*m66*me2*RSe451p*RSe463p - 5./54.
         *m66*me2*RSe451p*RSe453p + 1./108.*m66*me2*RSe451p*RSe561p + 1.
         /108.*m66*me2*RSe451p*RSe461p + 5./108.*m66*me2*pow(RSe451p,2)
          - 1./108.*m66*me2*RSe456p*pow(RSe453p,2) + 1./54.*m66*me2*
         RSe456p*RSe451p*RSe453p - 1./108.*m66*me2*RSe456p*pow(
         RSe451p,2) + 1./108.*pow(m66,2)*RSe453p*RSe563p + 1./108.*pow(
         m66,2)*RSe453p*RSe463p );

      F13p6x23R +=  + Bbeps(me2,mu2)*pi16 * (  - 1./324.*pow(m66,2)*
         pow(RSe453p,2) - 1./108.*pow(m66,2)*RS645pe*pow(RSe453p,2) - 1.
         /108.*pow(m66,2)*RSe561p*RSe453p - 1./108.*pow(m66,2)*RSe461p*
         RSe453p - 1./108.*pow(m66,2)*RSe451p*RSe563p - 1./108.*pow(
         m66,2)*RSe451p*RSe463p + 1./162.*pow(m66,2)*RSe451p*RSe453p + 
         1./54.*pow(m66,2)*RSe451p*RS645pe*RSe453p + 1./108.*pow(m66,2)
         *RSe451p*RSe561p + 1./108.*pow(m66,2)*RSe451p*RSe461p - 1./324.
         *pow(m66,2)*pow(RSe451p,2) - 1./108.*pow(m66,2)*pow(RSe451p,2)
         *RS645pe - 1./108.*m55*mp2*RSp465e*pow(RSe463p,2) + 1./54.*m55
         *mp2*RSp465e*RSe461p*RSe463p - 1./108.*m55*mp2*RSp465e*pow(
         RSe461p,2) + 1./108.*m55*me2*pow(RSe563p,2) + 1./108.*m55*me2*
         RSe463p*RSe563p + 5./108.*m55*me2*pow(RSe463p,2) + 1./108.*m55
         *me2*RSe453p*RSe463p + 1./108.*m55*me2*pow(RSe453p,2) - 1./54.
         *m55*me2*RSe561p*RSe563p - 1./108.*m55*me2*RSe561p*RSe463p + 1.
         /108.*m55*me2*pow(RSe561p,2) - 1./108.*m55*me2*RSe461p*RSe563p
          - 5./54.*m55*me2*RSe461p*RSe463p );

      F13p6x23R +=  + Bbeps(me2,mu2)*pi16 * (  - 1./108.*m55*me2*
         RSe461p*RSe453p + 1./108.*m55*me2*RSe461p*RSe561p + 5./108.*
         m55*me2*pow(RSe461p,2) - 1./108.*m55*me2*RSe451p*RSe463p - 1./
         54.*m55*me2*RSe451p*RSe453p + 1./108.*m55*me2*RSe451p*RSe461p
          + 1./108.*m55*me2*pow(RSe451p,2) - 1./108.*m55*me2*RSe465p*
         pow(RSe463p,2) + 1./54.*m55*me2*RSe465p*RSe461p*RSe463p - 1./
         108.*m55*me2*RSe465p*pow(RSe461p,2) - 1./81.*m55*m66*pow(
         RSe463p,2) + 1./54.*m55*m66*RSe453p*RSe463p - 1./81.*m55*m66*
         pow(RSe453p,2) + 2./81.*m55*m66*RSe461p*RSe463p - 1./54.*m55*
         m66*RSe461p*RSe453p - 1./81.*m55*m66*pow(RSe461p,2) - 1./54.*
         m55*m66*RSe451p*RSe463p + 2./81.*m55*m66*RSe451p*RSe453p + 1./
         54.*m55*m66*RSe451p*RSe461p - 1./81.*m55*m66*pow(RSe451p,2) + 
         1./108.*pow(m55,2)*RSe463p*RSe563p - 1./324.*pow(m55,2)*pow(
         RSe463p,2) + 1./108.*pow(m55,2)*RSe453p*RSe463p - 1./108.*pow(
         m55,2)*RS546pe*pow(RSe463p,2) - 1./108.*pow(m55,2)*RSe561p*
         RSe463p );

      F13p6x23R +=  + Bbeps(me2,mu2)*pi16 * (  - 1./108.*pow(m55,2)*
         RSe461p*RSe563p + 1./162.*pow(m55,2)*RSe461p*RSe463p - 1./108.
         *pow(m55,2)*RSe461p*RSe453p + 1./54.*pow(m55,2)*RSe461p*
         RS546pe*RSe463p + 1./108.*pow(m55,2)*RSe461p*RSe561p - 1./324.
         *pow(m55,2)*pow(RSe461p,2) - 1./108.*pow(m55,2)*pow(RSe461p,2)
         *RS546pe - 1./108.*pow(m55,2)*RSe451p*RSe463p + 1./108.*pow(
         m55,2)*RSe451p*RSe461p - 1./108.*m44*mp2*RSp564e*pow(
         RSe563p,2) + 1./54.*m44*mp2*RSp564e*RSe561p*RSe563p - 1./108.*
         m44*mp2*RSp564e*pow(RSe561p,2) + 5./108.*m44*me2*pow(
         RSe563p,2) + 1./108.*m44*me2*RSe463p*RSe563p + 1./108.*m44*me2
         *pow(RSe463p,2) + 1./108.*m44*me2*RSe453p*RSe563p + 1./108.*
         m44*me2*pow(RSe453p,2) - 5./54.*m44*me2*RSe561p*RSe563p - 1./
         108.*m44*me2*RSe561p*RSe463p - 1./108.*m44*me2*RSe561p*RSe453p
          + 5./108.*m44*me2*pow(RSe561p,2) - 1./108.*m44*me2*RSe461p*
         RSe563p - 1./54.*m44*me2*RSe461p*RSe463p + 1./108.*m44*me2*
         RSe461p*RSe561p );

      F13p6x23R +=  + Bbeps(me2,mu2)*pi16 * ( 1./108.*m44*me2*pow(
         RSe461p,2) - 1./108.*m44*me2*RSe451p*RSe563p - 1./54.*m44*me2*
         RSe451p*RSe453p + 1./108.*m44*me2*RSe451p*RSe561p + 1./108.*
         m44*me2*pow(RSe451p,2) - 1./108.*m44*me2*RSe564p*pow(
         RSe563p,2) + 1./54.*m44*me2*RSe564p*RSe561p*RSe563p - 1./108.*
         m44*me2*RSe564p*pow(RSe561p,2) - 1./81.*m44*m66*pow(RSe563p,2)
          + 1./54.*m44*m66*RSe453p*RSe563p - 1./81.*m44*m66*pow(
         RSe453p,2) + 2./81.*m44*m66*RSe561p*RSe563p - 1./54.*m44*m66*
         RSe561p*RSe453p - 1./81.*m44*m66*pow(RSe561p,2) - 1./54.*m44*
         m66*RSe451p*RSe563p + 2./81.*m44*m66*RSe451p*RSe453p + 1./54.*
         m44*m66*RSe451p*RSe561p - 1./81.*m44*m66*pow(RSe451p,2) - 1./
         81.*m44*m55*pow(RSe563p,2) + 1./54.*m44*m55*RSe463p*RSe563p - 
         1./81.*m44*m55*pow(RSe463p,2) + 2./81.*m44*m55*RSe561p*RSe563p
          - 1./54.*m44*m55*RSe561p*RSe463p - 1./81.*m44*m55*pow(
         RSe561p,2) - 1./54.*m44*m55*RSe461p*RSe563p + 2./81.*m44*m55*
         RSe461p*RSe463p );

      F13p6x23R +=  + Bbeps(me2,mu2)*pi16 * ( 1./54.*m44*m55*RSe461p*
         RSe561p - 1./81.*m44*m55*pow(RSe461p,2) - 1./324.*pow(m44,2)*
         pow(RSe563p,2) + 1./108.*pow(m44,2)*RSe463p*RSe563p + 1./108.*
         pow(m44,2)*RSe453p*RSe563p - 1./108.*pow(m44,2)*RS456pe*pow(
         RSe563p,2) + 1./162.*pow(m44,2)*RSe561p*RSe563p - 1./108.*pow(
         m44,2)*RSe561p*RSe463p - 1./108.*pow(m44,2)*RSe561p*RSe453p + 
         1./54.*pow(m44,2)*RSe561p*RS456pe*RSe563p - 1./324.*pow(m44,2)
         *pow(RSe561p,2) - 1./108.*pow(m44,2)*pow(RSe561p,2)*RS456pe - 
         1./108.*pow(m44,2)*RSe461p*RSe563p + 1./108.*pow(m44,2)*
         RSe461p*RSe561p - 1./108.*pow(m44,2)*RSe451p*RSe563p + 1./108.
         *pow(m44,2)*RSe451p*RSe561p );

      F13p6x23R +=  + Bbeps(mp2,mu2)*pi16 * ( 1./108.*m66*mp2*pow(
         RSp563e,2) + 1./108.*m66*mp2*pow(RSp463e,2) + 1./108.*m66*mp2*
         RSp453e*RSp563e + 1./108.*m66*mp2*RSp453e*RSp463e + 5./108.*
         m66*mp2*pow(RSp453e,2) - 1./54.*m66*mp2*RSp561e*RSp563e - 1./
         108.*m66*mp2*RSp561e*RSp453e + 1./108.*m66*mp2*pow(RSp561e,2)
          - 1./54.*m66*mp2*RSp461e*RSp463e - 1./108.*m66*mp2*RSp461e*
         RSp453e + 1./108.*m66*mp2*pow(RSp461e,2) - 1./108.*m66*mp2*
         RSp451e*RSp563e - 1./108.*m66*mp2*RSp451e*RSp463e - 5./54.*m66
         *mp2*RSp451e*RSp453e + 1./108.*m66*mp2*RSp451e*RSp561e + 1./
         108.*m66*mp2*RSp451e*RSp461e + 5./108.*m66*mp2*pow(RSp451e,2)
          - 1./108.*m66*mp2*RSp456e*pow(RSp453e,2) + 1./54.*m66*mp2*
         RSp456e*RSp451e*RSp453e - 1./108.*m66*mp2*RSp456e*pow(
         RSp451e,2) - 1./108.*m66*me2*RSe456p*pow(RSp453e,2) + 1./54.*
         m66*me2*RSe456p*RSp451e*RSp453e - 1./108.*m66*me2*RSe456p*pow(
         RSp451e,2) + 1./108.*pow(m66,2)*RSp453e*RSp563e + 1./108.*pow(
         m66,2)*RSp453e*RSp463e );

      F13p6x23R +=  + Bbeps(mp2,mu2)*pi16 * (  - 1./324.*pow(m66,2)*
         pow(RSp453e,2) - 1./108.*pow(m66,2)*RS645pe*pow(RSp453e,2) - 1.
         /108.*pow(m66,2)*RSp561e*RSp453e - 1./108.*pow(m66,2)*RSp461e*
         RSp453e - 1./108.*pow(m66,2)*RSp451e*RSp563e - 1./108.*pow(
         m66,2)*RSp451e*RSp463e + 1./162.*pow(m66,2)*RSp451e*RSp453e + 
         1./54.*pow(m66,2)*RSp451e*RS645pe*RSp453e + 1./108.*pow(m66,2)
         *RSp451e*RSp561e + 1./108.*pow(m66,2)*RSp451e*RSp461e - 1./324.
         *pow(m66,2)*pow(RSp451e,2) - 1./108.*pow(m66,2)*pow(RSp451e,2)
         *RS645pe + 1./108.*m55*mp2*pow(RSp563e,2) + 1./108.*m55*mp2*
         RSp463e*RSp563e + 5./108.*m55*mp2*pow(RSp463e,2) + 1./108.*m55
         *mp2*RSp453e*RSp463e + 1./108.*m55*mp2*pow(RSp453e,2) - 1./54.
         *m55*mp2*RSp561e*RSp563e - 1./108.*m55*mp2*RSp561e*RSp463e + 1.
         /108.*m55*mp2*pow(RSp561e,2) - 1./108.*m55*mp2*RSp461e*RSp563e
          - 5./54.*m55*mp2*RSp461e*RSp463e - 1./108.*m55*mp2*RSp461e*
         RSp453e + 1./108.*m55*mp2*RSp461e*RSp561e + 5./108.*m55*mp2*
         pow(RSp461e,2) );

      F13p6x23R +=  + Bbeps(mp2,mu2)*pi16 * (  - 1./108.*m55*mp2*
         RSp451e*RSp463e - 1./54.*m55*mp2*RSp451e*RSp453e + 1./108.*m55
         *mp2*RSp451e*RSp461e + 1./108.*m55*mp2*pow(RSp451e,2) - 1./108.
         *m55*mp2*RSp465e*pow(RSp463e,2) + 1./54.*m55*mp2*RSp465e*
         RSp461e*RSp463e - 1./108.*m55*mp2*RSp465e*pow(RSp461e,2) - 1./
         108.*m55*me2*RSe465p*pow(RSp463e,2) + 1./54.*m55*me2*RSe465p*
         RSp461e*RSp463e - 1./108.*m55*me2*RSe465p*pow(RSp461e,2) - 1./
         81.*m55*m66*pow(RSp463e,2) + 1./54.*m55*m66*RSp453e*RSp463e - 
         1./81.*m55*m66*pow(RSp453e,2) + 2./81.*m55*m66*RSp461e*RSp463e
          - 1./54.*m55*m66*RSp461e*RSp453e - 1./81.*m55*m66*pow(
         RSp461e,2) - 1./54.*m55*m66*RSp451e*RSp463e + 2./81.*m55*m66*
         RSp451e*RSp453e + 1./54.*m55*m66*RSp451e*RSp461e - 1./81.*m55*
         m66*pow(RSp451e,2) + 1./108.*pow(m55,2)*RSp463e*RSp563e - 1./
         324.*pow(m55,2)*pow(RSp463e,2) + 1./108.*pow(m55,2)*RSp453e*
         RSp463e - 1./108.*pow(m55,2)*RS546pe*pow(RSp463e,2) - 1./108.*
         pow(m55,2)*RSp561e*RSp463e );

      F13p6x23R +=  + Bbeps(mp2,mu2)*pi16 * (  - 1./108.*pow(m55,2)*
         RSp461e*RSp563e + 1./162.*pow(m55,2)*RSp461e*RSp463e - 1./108.
         *pow(m55,2)*RSp461e*RSp453e + 1./54.*pow(m55,2)*RSp461e*
         RS546pe*RSp463e + 1./108.*pow(m55,2)*RSp461e*RSp561e - 1./324.
         *pow(m55,2)*pow(RSp461e,2) - 1./108.*pow(m55,2)*pow(RSp461e,2)
         *RS546pe - 1./108.*pow(m55,2)*RSp451e*RSp463e + 1./108.*pow(
         m55,2)*RSp451e*RSp461e + 5./108.*m44*mp2*pow(RSp563e,2) + 1./
         108.*m44*mp2*RSp463e*RSp563e + 1./108.*m44*mp2*pow(RSp463e,2)
          + 1./108.*m44*mp2*RSp453e*RSp563e + 1./108.*m44*mp2*pow(
         RSp453e,2) - 5./54.*m44*mp2*RSp561e*RSp563e - 1./108.*m44*mp2*
         RSp561e*RSp463e - 1./108.*m44*mp2*RSp561e*RSp453e + 5./108.*
         m44*mp2*pow(RSp561e,2) - 1./108.*m44*mp2*RSp461e*RSp563e - 1./
         54.*m44*mp2*RSp461e*RSp463e + 1./108.*m44*mp2*RSp461e*RSp561e
          + 1./108.*m44*mp2*pow(RSp461e,2) - 1./108.*m44*mp2*RSp451e*
         RSp563e - 1./54.*m44*mp2*RSp451e*RSp453e + 1./108.*m44*mp2*
         RSp451e*RSp561e );

      F13p6x23R +=  + Bbeps(mp2,mu2)*pi16 * ( 1./108.*m44*mp2*pow(
         RSp451e,2) - 1./108.*m44*mp2*RSp564e*pow(RSp563e,2) + 1./54.*
         m44*mp2*RSp564e*RSp561e*RSp563e - 1./108.*m44*mp2*RSp564e*pow(
         RSp561e,2) - 1./108.*m44*me2*RSe564p*pow(RSp563e,2) + 1./54.*
         m44*me2*RSe564p*RSp561e*RSp563e - 1./108.*m44*me2*RSe564p*pow(
         RSp561e,2) - 1./81.*m44*m66*pow(RSp563e,2) + 1./54.*m44*m66*
         RSp453e*RSp563e - 1./81.*m44*m66*pow(RSp453e,2) + 2./81.*m44*
         m66*RSp561e*RSp563e - 1./54.*m44*m66*RSp561e*RSp453e - 1./81.*
         m44*m66*pow(RSp561e,2) - 1./54.*m44*m66*RSp451e*RSp563e + 2./
         81.*m44*m66*RSp451e*RSp453e + 1./54.*m44*m66*RSp451e*RSp561e
          - 1./81.*m44*m66*pow(RSp451e,2) - 1./81.*m44*m55*pow(
         RSp563e,2) + 1./54.*m44*m55*RSp463e*RSp563e - 1./81.*m44*m55*
         pow(RSp463e,2) + 2./81.*m44*m55*RSp561e*RSp563e - 1./54.*m44*
         m55*RSp561e*RSp463e - 1./81.*m44*m55*pow(RSp561e,2) - 1./54.*
         m44*m55*RSp461e*RSp563e + 2./81.*m44*m55*RSp461e*RSp463e + 1./
         54.*m44*m55*RSp461e*RSp561e );

      F13p6x23R +=  + Bbeps(mp2,mu2)*pi16 * (  - 1./81.*m44*m55*pow(
         RSp461e,2) - 1./324.*pow(m44,2)*pow(RSp563e,2) + 1./108.*pow(
         m44,2)*RSp463e*RSp563e + 1./108.*pow(m44,2)*RSp453e*RSp563e - 
         1./108.*pow(m44,2)*RS456pe*pow(RSp563e,2) + 1./162.*pow(m44,2)
         *RSp561e*RSp563e - 1./108.*pow(m44,2)*RSp561e*RSp463e - 1./108.
         *pow(m44,2)*RSp561e*RSp453e + 1./54.*pow(m44,2)*RSp561e*
         RS456pe*RSp563e - 1./324.*pow(m44,2)*pow(RSp561e,2) - 1./108.*
         pow(m44,2)*pow(RSp561e,2)*RS456pe - 1./108.*pow(m44,2)*RSp461e
         *RSp563e + 1./108.*pow(m44,2)*RSp461e*RSp561e - 1./108.*pow(
         m44,2)*RSp451e*RSp563e + 1./108.*pow(m44,2)*RSp451e*RSp561e );

      F13p6x23R +=  + hh(1,m11,m11,m13,m13,mu2) * (  - 1./32.*m33 - 1./
         72.*m33*pow(RS14563pe,2) - 1./72.*m33*RS1456pee*RS14563pe - 1./
         288.*m33*pow(RS1456pee,2) - 1./72.*m33*RS1456ppe*RS14563pe - 1.
         /144.*m33*RS1456ppe*RS1456pee - 1./288.*m33*pow(RS1456ppe,2)
          + 1./72.*m33*RS145pe*RS14563pe + 1./144.*m33*RS145pe*
         RS1456pee + 1./144.*m33*RS145pe*RS1456ppe - 1./288.*m33*pow(
         RS145pe,2) + 1./72.*m33*RS146pe*RS14563pe + 1./144.*m33*
         RS146pe*RS1456pee + 1./144.*m33*RS146pe*RS1456ppe - 1./144.*
         m33*RS146pe*RS145pe - 1./288.*m33*pow(RS146pe,2) + 1./72.*m33*
         RS156pe*RS14563pe + 1./144.*m33*RS156pe*RS1456pee + 1./144.*
         m33*RS156pe*RS1456ppe - 1./144.*m33*RS156pe*RS145pe - 1./144.*
         m33*RS156pe*RS146pe - 1./288.*m33*pow(RS156pe,2) + 3./32.*m11
          - 1./72.*m11*pow(RS14563pe,2) - 1./72.*m11*RS1456pee*
         RS14563pe - 1./288.*m11*pow(RS1456pee,2) - 1./72.*m11*
         RS1456ppe*RS14563pe - 1./144.*m11*RS1456ppe*RS1456pee - 1./288.
         *m11*pow(RS1456ppe,2) );

      F13p6x23R +=  + hh(1,m11,m11,m13,m13,mu2) * ( 1./72.*m11*RS145pe*
         RS14563pe + 1./144.*m11*RS145pe*RS1456pee + 1./144.*m11*
         RS145pe*RS1456ppe - 1./288.*m11*pow(RS145pe,2) + 1./72.*m11*
         RS146pe*RS14563pe + 1./144.*m11*RS146pe*RS1456pee + 1./144.*
         m11*RS146pe*RS1456ppe - 1./144.*m11*RS146pe*RS145pe - 1./288.*
         m11*pow(RS146pe,2) + 1./72.*m11*RS156pe*RS14563pe + 1./144.*
         m11*RS156pe*RS1456pee + 1./144.*m11*RS156pe*RS1456ppe - 1./144.
         *m11*RS156pe*RS145pe - 1./144.*m11*RS156pe*RS146pe - 1./288.*
         m11*pow(RS156pe,2) );

      F13p6x23R +=  + hh(1,m11,m13,m33,m13,mu2) * (  - 1./36.*m33*
         RS14563pe*RS34561pe - 1./72.*m33*RS3456pee*RS14563pe - 1./72.*
         m33*RS3456ppe*RS14563pe + 1./72.*m33*RS345pe*RS14563pe + 1./72.
         *m33*RS346pe*RS14563pe + 1./72.*m33*RS356pe*RS14563pe - 1./72.
         *m33*RS1456pee*RS34561pe - 1./144.*m33*RS1456pee*RS3456pee - 1.
         /144.*m33*RS1456pee*RS3456ppe + 1./144.*m33*RS1456pee*RS345pe
          + 1./144.*m33*RS1456pee*RS346pe + 1./144.*m33*RS1456pee*
         RS356pe - 1./72.*m33*RS1456ppe*RS34561pe - 1./144.*m33*
         RS1456ppe*RS3456pee - 1./144.*m33*RS1456ppe*RS3456ppe + 1./144.
         *m33*RS1456ppe*RS345pe + 1./144.*m33*RS1456ppe*RS346pe + 1./
         144.*m33*RS1456ppe*RS356pe + 1./72.*m33*RS145pe*RS34561pe + 1./
         144.*m33*RS145pe*RS3456pee + 1./144.*m33*RS145pe*RS3456ppe - 1.
         /144.*m33*RS145pe*RS345pe - 1./144.*m33*RS145pe*RS346pe - 1./
         144.*m33*RS145pe*RS356pe + 1./72.*m33*RS146pe*RS34561pe + 1./
         144.*m33*RS146pe*RS3456pee + 1./144.*m33*RS146pe*RS3456ppe - 1.
         /144.*m33*RS146pe*RS345pe );

      F13p6x23R +=  + hh(1,m11,m13,m33,m13,mu2) * (  - 1./144.*m33*
         RS146pe*RS346pe - 1./144.*m33*RS146pe*RS356pe + 1./72.*m33*
         RS156pe*RS34561pe + 1./144.*m33*RS156pe*RS3456pee + 1./144.*
         m33*RS156pe*RS3456ppe - 1./144.*m33*RS156pe*RS345pe - 1./144.*
         m33*RS156pe*RS346pe - 1./144.*m33*RS156pe*RS356pe - 1./36.*m11
         *RS14563pe*RS34561pe - 1./72.*m11*RS3456pee*RS14563pe - 1./72.
         *m11*RS3456ppe*RS14563pe + 1./72.*m11*RS345pe*RS14563pe + 1./
         72.*m11*RS346pe*RS14563pe + 1./72.*m11*RS356pe*RS14563pe - 1./
         72.*m11*RS1456pee*RS34561pe - 1./144.*m11*RS1456pee*RS3456pee
          - 1./144.*m11*RS1456pee*RS3456ppe + 1./144.*m11*RS1456pee*
         RS345pe + 1./144.*m11*RS1456pee*RS346pe + 1./144.*m11*
         RS1456pee*RS356pe - 1./72.*m11*RS1456ppe*RS34561pe - 1./144.*
         m11*RS1456ppe*RS3456pee - 1./144.*m11*RS1456ppe*RS3456ppe + 1./
         144.*m11*RS1456ppe*RS345pe + 1./144.*m11*RS1456ppe*RS346pe + 1.
         /144.*m11*RS1456ppe*RS356pe + 1./72.*m11*RS145pe*RS34561pe + 1.
         /144.*m11*RS145pe*RS3456pee );

      F13p6x23R +=  + hh(1,m11,m13,m33,m13,mu2) * ( 1./144.*m11*RS145pe
         *RS3456ppe - 1./144.*m11*RS145pe*RS345pe - 1./144.*m11*RS145pe
         *RS346pe - 1./144.*m11*RS145pe*RS356pe + 1./72.*m11*RS146pe*
         RS34561pe + 1./144.*m11*RS146pe*RS3456pee + 1./144.*m11*
         RS146pe*RS3456ppe - 1./144.*m11*RS146pe*RS345pe - 1./144.*m11*
         RS146pe*RS346pe - 1./144.*m11*RS146pe*RS356pe + 1./72.*m11*
         RS156pe*RS34561pe + 1./144.*m11*RS156pe*RS3456pee + 1./144.*
         m11*RS156pe*RS3456ppe - 1./144.*m11*RS156pe*RS345pe - 1./144.*
         m11*RS156pe*RS346pe - 1./144.*m11*RS156pe*RS356pe );

      F13p6x23R +=  + hh(1,m11,m13,me2,m13,mu2) * (  - 1./36.*m33*
         RSe45613p*RS14563pe + 1./72.*m33*RSe45633p*RS14563pe - 1./72.*
         m33*RS1456pee*RSe45613p + 1./144.*m33*RS1456pee*RSe45633p - 1./
         72.*m33*RS1456ppe*RSe45613p + 1./144.*m33*RS1456ppe*RSe45633p
          + 1./72.*m33*RSe45611p*RS14563pe + 1./144.*m33*RSe45611p*
         RS1456pee + 1./144.*m33*RSe45611p*RS1456ppe + 1./72.*m33*
         RS145pe*RSe45613p - 1./144.*m33*RS145pe*RSe45633p - 1./144.*
         m33*RS145pe*RSe45611p + 1./72.*m33*RS146pe*RSe45613p - 1./144.
         *m33*RS146pe*RSe45633p - 1./144.*m33*RS146pe*RSe45611p + 1./72.
         *m33*RS156pe*RSe45613p - 1./144.*m33*RS156pe*RSe45633p - 1./
         144.*m33*RS156pe*RSe45611p - 1./36.*m11*RSe45613p*RS14563pe + 
         1./72.*m11*RSe45633p*RS14563pe - 1./72.*m11*RS1456pee*
         RSe45613p + 1./144.*m11*RS1456pee*RSe45633p - 1./72.*m11*
         RS1456ppe*RSe45613p + 1./144.*m11*RS1456ppe*RSe45633p + 1./72.
         *m11*RSe45611p*RS14563pe + 1./144.*m11*RSe45611p*RS1456pee + 1.
         /144.*m11*RSe45611p*RS1456ppe );

      F13p6x23R +=  + hh(1,m11,m13,me2,m13,mu2) * ( 1./72.*m11*RS145pe*
         RSe45613p - 1./144.*m11*RS145pe*RSe45633p - 1./144.*m11*
         RS145pe*RSe45611p + 1./72.*m11*RS146pe*RSe45613p - 1./144.*m11
         *RS146pe*RSe45633p - 1./144.*m11*RS146pe*RSe45611p + 1./72.*
         m11*RS156pe*RSe45613p - 1./144.*m11*RS156pe*RSe45633p - 1./144.
         *m11*RS156pe*RSe45611p );

      F13p6x23R +=  + hh(1,m11,m13,mp2,m13,mu2) * (  - 1./36.*m33*
         RSp45613e*RS14563pe + 1./72.*m33*RSp45633e*RS14563pe - 1./72.*
         m33*RS1456pee*RSp45613e + 1./144.*m33*RS1456pee*RSp45633e - 1./
         72.*m33*RS1456ppe*RSp45613e + 1./144.*m33*RS1456ppe*RSp45633e
          + 1./72.*m33*RSp45611e*RS14563pe + 1./144.*m33*RSp45611e*
         RS1456pee + 1./144.*m33*RSp45611e*RS1456ppe + 1./72.*m33*
         RS145pe*RSp45613e - 1./144.*m33*RS145pe*RSp45633e - 1./144.*
         m33*RS145pe*RSp45611e + 1./72.*m33*RS146pe*RSp45613e - 1./144.
         *m33*RS146pe*RSp45633e - 1./144.*m33*RS146pe*RSp45611e + 1./72.
         *m33*RS156pe*RSp45613e - 1./144.*m33*RS156pe*RSp45633e - 1./
         144.*m33*RS156pe*RSp45611e - 1./36.*m11*RSp45613e*RS14563pe + 
         1./72.*m11*RSp45633e*RS14563pe - 1./72.*m11*RS1456pee*
         RSp45613e + 1./144.*m11*RS1456pee*RSp45633e - 1./72.*m11*
         RS1456ppe*RSp45613e + 1./144.*m11*RS1456ppe*RSp45633e + 1./72.
         *m11*RSp45611e*RS14563pe + 1./144.*m11*RSp45611e*RS1456pee + 1.
         /144.*m11*RSp45611e*RS1456ppe );

      F13p6x23R +=  + hh(1,m11,m13,mp2,m13,mu2) * ( 1./72.*m11*RS145pe*
         RSp45613e - 1./144.*m11*RS145pe*RSp45633e - 1./144.*m11*
         RS145pe*RSp45611e + 1./72.*m11*RS146pe*RSp45613e - 1./144.*m11
         *RS146pe*RSp45633e - 1./144.*m11*RS146pe*RSp45611e + 1./72.*
         m11*RS156pe*RSp45613e - 1./144.*m11*RS156pe*RSp45633e - 1./144.
         *m11*RS156pe*RSp45611e );

      F13p6x23R +=  + hh(1,m11,m14,m34,m13,mu2) * (  - 1./24.*m44*
         RS14563pe - 1./48.*m44*RS1456pee - 1./48.*m44*RS1456ppe + 1./
         48.*m44*RS145pe + 1./48.*m44*RS146pe + 1./48.*m44*RS156pe - 1./
         48.*m33*RS14563pe + 1./48.*m33*RS156pe + 1./48.*m11*RS14563pe
          - 1./48.*m11*RS1456pee - 1./48.*m11*RS1456ppe + 1./48.*m11*
         RS145pe + 1./48.*m11*RS146pe - 1./24.*m11*RS156pe );

      F13p6x23R +=  + hh(1,m11,m15,m35,m13,mu2) * (  - 1./24.*m55*
         RS14563pe - 1./48.*m55*RS1456pee - 1./48.*m55*RS1456ppe + 1./
         48.*m55*RS145pe + 1./48.*m55*RS146pe + 1./48.*m55*RS156pe - 1./
         48.*m33*RS14563pe + 1./48.*m33*RS146pe + 1./48.*m11*RS14563pe
          - 1./48.*m11*RS1456pee - 1./48.*m11*RS1456ppe + 1./48.*m11*
         RS145pe - 1./24.*m11*RS146pe + 1./48.*m11*RS156pe );

      F13p6x23R +=  + hh(1,m11,m16,m36,m13,mu2) * (  - 1./24.*m66*
         RS14563pe - 1./48.*m66*RS1456pee - 1./48.*m66*RS1456ppe + 1./
         48.*m66*RS145pe + 1./48.*m66*RS146pe + 1./48.*m66*RS156pe - 1./
         48.*m33*RS14563pe + 1./48.*m33*RS145pe + 1./48.*m11*RS14563pe
          - 1./48.*m11*RS1456pee - 1./48.*m11*RS1456ppe - 1./24.*m11*
         RS145pe + 1./48.*m11*RS146pe + 1./48.*m11*RS156pe );

      F13p6x23R +=  + hh(1,m13,m13,m13,m13,mu2) * (  - 1./16.*m33 - 1./
         16.*m11 );

      F13p6x23R +=  + hh(1,m13,m33,m33,m13,mu2) * ( 3./32.*m33 - 1./72.
         *m33*pow(RS34561pe,2) - 1./72.*m33*RS3456pee*RS34561pe - 1./
         288.*m33*pow(RS3456pee,2) - 1./72.*m33*RS3456ppe*RS34561pe - 1.
         /144.*m33*RS3456ppe*RS3456pee - 1./288.*m33*pow(RS3456ppe,2)
          + 1./72.*m33*RS345pe*RS34561pe + 1./144.*m33*RS345pe*
         RS3456pee + 1./144.*m33*RS345pe*RS3456ppe - 1./288.*m33*pow(
         RS345pe,2) + 1./72.*m33*RS346pe*RS34561pe + 1./144.*m33*
         RS346pe*RS3456pee + 1./144.*m33*RS346pe*RS3456ppe - 1./144.*
         m33*RS346pe*RS345pe - 1./288.*m33*pow(RS346pe,2) + 1./72.*m33*
         RS356pe*RS34561pe + 1./144.*m33*RS356pe*RS3456pee + 1./144.*
         m33*RS356pe*RS3456ppe - 1./144.*m33*RS356pe*RS345pe - 1./144.*
         m33*RS356pe*RS346pe - 1./288.*m33*pow(RS356pe,2) - 1./32.*m11
          - 1./72.*m11*pow(RS34561pe,2) - 1./72.*m11*RS3456pee*
         RS34561pe - 1./288.*m11*pow(RS3456pee,2) - 1./72.*m11*
         RS3456ppe*RS34561pe - 1./144.*m11*RS3456ppe*RS3456pee - 1./288.
         *m11*pow(RS3456ppe,2) );

      F13p6x23R +=  + hh(1,m13,m33,m33,m13,mu2) * ( 1./72.*m11*RS345pe*
         RS34561pe + 1./144.*m11*RS345pe*RS3456pee + 1./144.*m11*
         RS345pe*RS3456ppe - 1./288.*m11*pow(RS345pe,2) + 1./72.*m11*
         RS346pe*RS34561pe + 1./144.*m11*RS346pe*RS3456pee + 1./144.*
         m11*RS346pe*RS3456ppe - 1./144.*m11*RS346pe*RS345pe - 1./288.*
         m11*pow(RS346pe,2) + 1./72.*m11*RS356pe*RS34561pe + 1./144.*
         m11*RS356pe*RS3456pee + 1./144.*m11*RS356pe*RS3456ppe - 1./144.
         *m11*RS356pe*RS345pe - 1./144.*m11*RS356pe*RS346pe - 1./288.*
         m11*pow(RS356pe,2) );

      F13p6x23R +=  + hh(1,m13,m33,me2,m13,mu2) * (  - 1./36.*m33*
         RSe45613p*RS34561pe - 1./72.*m33*RS3456pee*RSe45613p - 1./72.*
         m33*RS3456ppe*RSe45613p + 1./72.*m33*RSe45633p*RS34561pe + 1./
         144.*m33*RSe45633p*RS3456pee + 1./144.*m33*RSe45633p*RS3456ppe
          + 1./72.*m33*RS345pe*RSe45613p - 1./144.*m33*RS345pe*
         RSe45633p + 1./72.*m33*RS346pe*RSe45613p - 1./144.*m33*RS346pe
         *RSe45633p + 1./72.*m33*RS356pe*RSe45613p - 1./144.*m33*
         RS356pe*RSe45633p + 1./72.*m33*RSe45611p*RS34561pe + 1./144.*
         m33*RSe45611p*RS3456pee + 1./144.*m33*RSe45611p*RS3456ppe - 1./
         144.*m33*RSe45611p*RS345pe - 1./144.*m33*RSe45611p*RS346pe - 1.
         /144.*m33*RSe45611p*RS356pe - 1./36.*m11*RSe45613p*RS34561pe
          - 1./72.*m11*RS3456pee*RSe45613p - 1./72.*m11*RS3456ppe*
         RSe45613p + 1./72.*m11*RSe45633p*RS34561pe + 1./144.*m11*
         RSe45633p*RS3456pee + 1./144.*m11*RSe45633p*RS3456ppe + 1./72.
         *m11*RS345pe*RSe45613p - 1./144.*m11*RS345pe*RSe45633p + 1./72.
         *m11*RS346pe*RSe45613p );

      F13p6x23R +=  + hh(1,m13,m33,me2,m13,mu2) * (  - 1./144.*m11*
         RS346pe*RSe45633p + 1./72.*m11*RS356pe*RSe45613p - 1./144.*m11
         *RS356pe*RSe45633p + 1./72.*m11*RSe45611p*RS34561pe + 1./144.*
         m11*RSe45611p*RS3456pee + 1./144.*m11*RSe45611p*RS3456ppe - 1./
         144.*m11*RSe45611p*RS345pe - 1./144.*m11*RSe45611p*RS346pe - 1.
         /144.*m11*RSe45611p*RS356pe );

      F13p6x23R +=  + hh(1,m13,m33,mp2,m13,mu2) * (  - 1./36.*m33*
         RSp45613e*RS34561pe - 1./72.*m33*RS3456pee*RSp45613e - 1./72.*
         m33*RS3456ppe*RSp45613e + 1./72.*m33*RSp45633e*RS34561pe + 1./
         144.*m33*RSp45633e*RS3456pee + 1./144.*m33*RSp45633e*RS3456ppe
          + 1./72.*m33*RS345pe*RSp45613e - 1./144.*m33*RS345pe*
         RSp45633e + 1./72.*m33*RS346pe*RSp45613e - 1./144.*m33*RS346pe
         *RSp45633e + 1./72.*m33*RS356pe*RSp45613e - 1./144.*m33*
         RS356pe*RSp45633e + 1./72.*m33*RSp45611e*RS34561pe + 1./144.*
         m33*RSp45611e*RS3456pee + 1./144.*m33*RSp45611e*RS3456ppe - 1./
         144.*m33*RSp45611e*RS345pe - 1./144.*m33*RSp45611e*RS346pe - 1.
         /144.*m33*RSp45611e*RS356pe - 1./36.*m11*RSp45613e*RS34561pe
          - 1./72.*m11*RS3456pee*RSp45613e - 1./72.*m11*RS3456ppe*
         RSp45613e + 1./72.*m11*RSp45633e*RS34561pe + 1./144.*m11*
         RSp45633e*RS3456pee + 1./144.*m11*RSp45633e*RS3456ppe + 1./72.
         *m11*RS345pe*RSp45613e - 1./144.*m11*RS345pe*RSp45633e + 1./72.
         *m11*RS346pe*RSp45613e );

      F13p6x23R +=  + hh(1,m13,m33,mp2,m13,mu2) * (  - 1./144.*m11*
         RS346pe*RSp45633e + 1./72.*m11*RS356pe*RSp45613e - 1./144.*m11
         *RS356pe*RSp45633e + 1./72.*m11*RSp45611e*RS34561pe + 1./144.*
         m11*RSp45611e*RS3456pee + 1./144.*m11*RSp45611e*RS3456ppe - 1./
         144.*m11*RSp45611e*RS345pe - 1./144.*m11*RSp45611e*RS346pe - 1.
         /144.*m11*RSp45611e*RS356pe );

      F13p6x23R +=  + hh(1,m13,me2,me2,m13,mu2) * (  - 1./72.*m33*pow(
         RSe45613p,2) + 1./72.*m33*RSe45633p*RSe45613p - 1./288.*m33*
         pow(RSe45633p,2) + 1./72.*m33*RSe45611p*RSe45613p - 1./144.*
         m33*RSe45611p*RSe45633p - 1./288.*m33*pow(RSe45611p,2) - 1./72.
         *m11*pow(RSe45613p,2) + 1./72.*m11*RSe45633p*RSe45613p - 1./
         288.*m11*pow(RSe45633p,2) + 1./72.*m11*RSe45611p*RSe45613p - 1.
         /144.*m11*RSe45611p*RSe45633p - 1./288.*m11*pow(RSe45611p,2) )
         ;

      F13p6x23R +=  + hh(1,m13,me2,mp2,m13,mu2) * (  - 1./36.*m33*
         RSp45613e*RSe45613p + 1./72.*m33*RSe45633p*RSp45613e + 1./72.*
         m33*RSp45633e*RSe45613p - 1./144.*m33*RSp45633e*RSe45633p + 1./
         72.*m33*RSe45611p*RSp45613e - 1./144.*m33*RSe45611p*RSp45633e
          + 1./72.*m33*RSp45611e*RSe45613p - 1./144.*m33*RSp45611e*
         RSe45633p - 1./144.*m33*RSp45611e*RSe45611p - 1./36.*m11*
         RSp45613e*RSe45613p + 1./72.*m11*RSe45633p*RSp45613e + 1./72.*
         m11*RSp45633e*RSe45613p - 1./144.*m11*RSp45633e*RSe45633p + 1./
         72.*m11*RSe45611p*RSp45613e - 1./144.*m11*RSe45611p*RSp45633e
          + 1./72.*m11*RSp45611e*RSe45613p - 1./144.*m11*RSp45611e*
         RSe45633p - 1./144.*m11*RSp45611e*RSe45611p );

      F13p6x23R +=  + hh(1,m13,mp2,mp2,m13,mu2) * (  - 1./72.*m33*pow(
         RSp45613e,2) + 1./72.*m33*RSp45633e*RSp45613e - 1./288.*m33*
         pow(RSp45633e,2) + 1./72.*m33*RSp45611e*RSp45613e - 1./144.*
         m33*RSp45611e*RSp45633e - 1./288.*m33*pow(RSp45611e,2) - 1./72.
         *m11*pow(RSp45613e,2) + 1./72.*m11*RSp45633e*RSp45613e - 1./
         288.*m11*pow(RSp45633e,2) + 1./72.*m11*RSp45611e*RSp45613e - 1.
         /144.*m11*RSp45611e*RSp45633e - 1./288.*m11*pow(RSp45611e,2) )
         ;

      F13p6x23R +=  + hh(1,m14,m33,m34,m13,mu2) * (  - 1./24.*m44*
         RS34561pe - 1./48.*m44*RS3456pee - 1./48.*m44*RS3456ppe + 1./
         48.*m44*RS345pe + 1./48.*m44*RS346pe + 1./48.*m44*RS356pe + 1./
         48.*m33*RS34561pe - 1./48.*m33*RS3456pee - 1./48.*m33*
         RS3456ppe + 1./48.*m33*RS345pe + 1./48.*m33*RS346pe - 1./24.*
         m33*RS356pe - 1./48.*m11*RS34561pe + 1./48.*m11*RS356pe );

      F13p6x23R +=  + hh(1,m14,m34,m44,m13,mu2) * (  - 1./8.*m44 + 1./
         24.*m44*RS456pe );

      F13p6x23R +=  + hh(1,m14,m34,me2,m13,mu2) * ( 1./24.*me2*
         RSe45613p - 1./24.*me2*RSe563p - 1./24.*me2*RSe561p + 1./24.*
         me2*RSe564p - 1./24.*m44*RSe45613p + 1./48.*m44*RSe45633p + 1./
         48.*m44*RSe45611p - 1./48.*m33*RSe45613p + 1./48.*m33*
         RSe45633p - 1./48.*m33*RSe563p + 1./48.*m33*RSe561p - 1./48.*
         m11*RSe45613p + 1./48.*m11*RSe563p + 1./48.*m11*RSe45611p - 1./
         48.*m11*RSe561p );

      F13p6x23R +=  + hh(1,m14,m34,mp2,m13,mu2) * ( 1./24.*mp2*
         RSp45613e - 1./24.*mp2*RSp563e - 1./24.*mp2*RSp561e + 1./24.*
         mp2*RSp564e - 1./24.*m44*RSp45613e + 1./48.*m44*RSp45633e + 1./
         48.*m44*RSp45611e - 1./48.*m33*RSp45613e + 1./48.*m33*
         RSp45633e - 1./48.*m33*RSp563e + 1./48.*m33*RSp561e - 1./48.*
         m11*RSp45613e + 1./48.*m11*RSp563e + 1./48.*m11*RSp45611e - 1./
         48.*m11*RSp561e );

      F13p6x23R +=  + hh(1,m14,m35,m45,m13,mu2) * (  - 1./16.*m55 - 1./
         16.*m44 );

      F13p6x23R +=  + hh(1,m14,m36,m46,m13,mu2) * (  - 1./16.*m66 - 1./
         16.*m44 );

      F13p6x23R +=  + hh(1,m15,m33,m35,m13,mu2) * (  - 1./24.*m55*
         RS34561pe - 1./48.*m55*RS3456pee - 1./48.*m55*RS3456ppe + 1./
         48.*m55*RS345pe + 1./48.*m55*RS346pe + 1./48.*m55*RS356pe + 1./
         48.*m33*RS34561pe - 1./48.*m33*RS3456pee - 1./48.*m33*
         RS3456ppe + 1./48.*m33*RS345pe - 1./24.*m33*RS346pe + 1./48.*
         m33*RS356pe - 1./48.*m11*RS34561pe + 1./48.*m11*RS346pe );

      F13p6x23R +=  + hh(1,m15,m34,m45,m13,mu2) * (  - 1./16.*m55 - 1./
         16.*m44 );

      F13p6x23R +=  + hh(1,m15,m35,m55,m13,mu2) * (  - 1./8.*m55 + 1./
         24.*m55*RS546pe );

      F13p6x23R +=  + hh(1,m15,m35,me2,m13,mu2) * ( 1./24.*me2*
         RSe45613p - 1./24.*me2*RSe463p - 1./24.*me2*RSe461p + 1./24.*
         me2*RSe465p - 1./24.*m55*RSe45613p + 1./48.*m55*RSe45633p + 1./
         48.*m55*RSe45611p - 1./48.*m33*RSe45613p + 1./48.*m33*
         RSe45633p - 1./48.*m33*RSe463p + 1./48.*m33*RSe461p - 1./48.*
         m11*RSe45613p + 1./48.*m11*RSe463p + 1./48.*m11*RSe45611p - 1./
         48.*m11*RSe461p );

      F13p6x23R +=  + hh(1,m15,m35,mp2,m13,mu2) * ( 1./24.*mp2*
         RSp45613e - 1./24.*mp2*RSp463e - 1./24.*mp2*RSp461e + 1./24.*
         mp2*RSp465e - 1./24.*m55*RSp45613e + 1./48.*m55*RSp45633e + 1./
         48.*m55*RSp45611e - 1./48.*m33*RSp45613e + 1./48.*m33*
         RSp45633e - 1./48.*m33*RSp463e + 1./48.*m33*RSp461e - 1./48.*
         m11*RSp45613e + 1./48.*m11*RSp463e + 1./48.*m11*RSp45611e - 1./
         48.*m11*RSp461e );

      F13p6x23R +=  + hh(1,m15,m36,m56,m13,mu2) * (  - 1./16.*m66 - 1./
         16.*m55 );

      F13p6x23R +=  + hh(1,m16,m33,m36,m13,mu2) * (  - 1./24.*m66*
         RS34561pe - 1./48.*m66*RS3456pee - 1./48.*m66*RS3456ppe + 1./
         48.*m66*RS345pe + 1./48.*m66*RS346pe + 1./48.*m66*RS356pe + 1./
         48.*m33*RS34561pe - 1./48.*m33*RS3456pee - 1./48.*m33*
         RS3456ppe - 1./24.*m33*RS345pe + 1./48.*m33*RS346pe + 1./48.*
         m33*RS356pe - 1./48.*m11*RS34561pe + 1./48.*m11*RS345pe );

      F13p6x23R +=  + hh(1,m16,m34,m46,m13,mu2) * (  - 1./16.*m66 - 1./
         16.*m44 );

      F13p6x23R +=  + hh(1,m16,m35,m56,m13,mu2) * (  - 1./16.*m66 - 1./
         16.*m55 );

      F13p6x23R +=  + hh(1,m16,m36,m66,m13,mu2) * (  - 1./8.*m66 + 1./
         24.*m66*RS645pe );

      F13p6x23R +=  + hh(1,m16,m36,me2,m13,mu2) * ( 1./24.*me2*
         RSe45613p - 1./24.*me2*RSe453p - 1./24.*me2*RSe451p + 1./24.*
         me2*RSe456p - 1./24.*m66*RSe45613p + 1./48.*m66*RSe45633p + 1./
         48.*m66*RSe45611p - 1./48.*m33*RSe45613p + 1./48.*m33*
         RSe45633p - 1./48.*m33*RSe453p + 1./48.*m33*RSe451p - 1./48.*
         m11*RSe45613p + 1./48.*m11*RSe453p + 1./48.*m11*RSe45611p - 1./
         48.*m11*RSe451p );

      F13p6x23R +=  + hh(1,m16,m36,mp2,m13,mu2) * ( 1./24.*mp2*
         RSp45613e - 1./24.*mp2*RSp453e - 1./24.*mp2*RSp451e + 1./24.*
         mp2*RSp456e - 1./24.*m66*RSp45613e + 1./48.*m66*RSp45633e + 1./
         48.*m66*RSp45611e - 1./48.*m33*RSp45613e + 1./48.*m33*
         RSp45633e - 1./48.*m33*RSp453e + 1./48.*m33*RSp451e - 1./48.*
         m11*RSp45613e + 1./48.*m11*RSp453e + 1./48.*m11*RSp45611e - 1./
         48.*m11*RSp451e );

      F13p6x23R +=  + hh(2,m11,m11,m13,m13,mu2) * ( 1./72.*m33*RD1456ep
         *RS14563pe + 1./144.*m33*RS1456pee*RD1456ep + 1./144.*m33*
         RS1456ppe*RD1456ep - 1./144.*m33*RS145pe*RD1456ep - 1./144.*
         m33*RS146pe*RD1456ep - 1./144.*m33*RS156pe*RD1456ep + 1./72.*
         m11*RD1456ep*RS14563pe + 1./144.*m11*RS1456pee*RD1456ep + 1./
         144.*m11*RS1456ppe*RD1456ep - 1./144.*m11*RS145pe*RD1456ep - 1.
         /144.*m11*RS146pe*RD1456ep - 1./144.*m11*RS156pe*RD1456ep );

      F13p6x23R +=  + hh(2,m11,m13,m33,m13,mu2) * ( 1./72.*m33*RD1456ep
         *RS34561pe + 1./144.*m33*RD1456ep*RS3456pee + 1./144.*m33*
         RD1456ep*RS3456ppe - 1./144.*m33*RD1456ep*RS345pe - 1./144.*
         m33*RD1456ep*RS346pe - 1./144.*m33*RD1456ep*RS356pe + 1./72.*
         m11*RD1456ep*RS34561pe + 1./144.*m11*RD1456ep*RS3456pee + 1./
         144.*m11*RD1456ep*RS3456ppe - 1./144.*m11*RD1456ep*RS345pe - 1.
         /144.*m11*RD1456ep*RS346pe - 1./144.*m11*RD1456ep*RS356pe );

      F13p6x23R +=  + hh(2,m11,m13,me2,m13,mu2) * ( 1./72.*m33*RD1456ep
         *RSe45613p - 1./144.*m33*RD1456ep*RSe45633p - 1./144.*m33*
         RSe45611p*RD1456ep + 1./72.*m11*RD1456ep*RSe45613p - 1./144.*
         m11*RD1456ep*RSe45633p - 1./144.*m11*RSe45611p*RD1456ep );

      F13p6x23R +=  + hh(2,m11,m13,mp2,m13,mu2) * ( 1./72.*m33*RD1456ep
         *RSp45613e - 1./144.*m33*RD1456ep*RSp45633e - 1./144.*m33*
         RSp45611e*RD1456ep + 1./72.*m11*RD1456ep*RSp45613e - 1./144.*
         m11*RD1456ep*RSp45633e - 1./144.*m11*RSp45611e*RD1456ep );

      F13p6x23R +=  + hh(2,m11,m14,m34,m13,mu2) * ( 1./48.*m44*RD1456ep
          + 1./48.*m11*RD1456ep );

      F13p6x23R +=  + hh(2,m11,m15,m35,m13,mu2) * ( 1./48.*m55*RD1456ep
          + 1./48.*m11*RD1456ep );

      F13p6x23R +=  + hh(2,m11,m16,m36,m13,mu2) * ( 1./48.*m66*RD1456ep
          + 1./48.*m11*RD1456ep );

      F13p6x23R +=  + hh(2,m33,m11,m13,m13,mu2) * ( 1./72.*m33*RD3456ep
         *RS14563pe + 1./144.*m33*RS1456pee*RD3456ep + 1./144.*m33*
         RS1456ppe*RD3456ep - 1./144.*m33*RS145pe*RD3456ep - 1./144.*
         m33*RS146pe*RD3456ep - 1./144.*m33*RS156pe*RD3456ep + 1./72.*
         m11*RD3456ep*RS14563pe + 1./144.*m11*RS1456pee*RD3456ep + 1./
         144.*m11*RS1456ppe*RD3456ep - 1./144.*m11*RS145pe*RD3456ep - 1.
         /144.*m11*RS146pe*RD3456ep - 1./144.*m11*RS156pe*RD3456ep );

      F13p6x23R +=  + hh(2,m33,m13,m33,m13,mu2) * ( 1./72.*m33*RD3456ep
         *RS34561pe + 1./144.*m33*RS3456pee*RD3456ep + 1./144.*m33*
         RS3456ppe*RD3456ep - 1./144.*m33*RS345pe*RD3456ep - 1./144.*
         m33*RS346pe*RD3456ep - 1./144.*m33*RS356pe*RD3456ep + 1./72.*
         m11*RD3456ep*RS34561pe + 1./144.*m11*RS3456pee*RD3456ep + 1./
         144.*m11*RS3456ppe*RD3456ep - 1./144.*m11*RS345pe*RD3456ep - 1.
         /144.*m11*RS346pe*RD3456ep - 1./144.*m11*RS356pe*RD3456ep );

      F13p6x23R +=  + hh(2,m33,m13,me2,m13,mu2) * ( 1./72.*m33*RD3456ep
         *RSe45613p - 1./144.*m33*RSe45633p*RD3456ep - 1./144.*m33*
         RSe45611p*RD3456ep + 1./72.*m11*RD3456ep*RSe45613p - 1./144.*
         m11*RSe45633p*RD3456ep - 1./144.*m11*RSe45611p*RD3456ep );

      F13p6x23R +=  + hh(2,m33,m13,mp2,m13,mu2) * ( 1./72.*m33*RD3456ep
         *RSp45613e - 1./144.*m33*RSp45633e*RD3456ep - 1./144.*m33*
         RSp45611e*RD3456ep + 1./72.*m11*RD3456ep*RSp45613e - 1./144.*
         m11*RSp45633e*RD3456ep - 1./144.*m11*RSp45611e*RD3456ep );

      F13p6x23R +=  + hh(2,m33,m14,m34,m13,mu2) * ( 1./48.*m44*RD3456ep
          + 1./48.*m33*RD3456ep );

      F13p6x23R +=  + hh(2,m33,m15,m35,m13,mu2) * ( 1./48.*m55*RD3456ep
          + 1./48.*m33*RD3456ep );

      F13p6x23R +=  + hh(2,m33,m16,m36,m13,mu2) * ( 1./48.*m66*RD3456ep
          + 1./48.*m33*RD3456ep );

      F13p6x23R +=  + hh(5,m11,m11,m13,m13,mu2) * (  - 1./288.*m33*pow(
         RD1456ep,2) - 1./288.*m11*pow(RD1456ep,2) );

      F13p6x23R +=  + hh(5,m11,m33,m13,m13,mu2) * (  - 1./144.*m33*
         RD1456ep*RD3456ep - 1./144.*m11*RD1456ep*RD3456ep );

      F13p6x23R +=  + hh(5,m33,m33,m13,m13,mu2) * (  - 1./288.*m33*pow(
         RD3456ep,2) - 1./288.*m11*pow(RD3456ep,2) );

      F13p6x23R +=  + hhd(1,m11,m11,m13,m13,mu2) * ( 1./648.*pow(
         RD1456ep,2) + 1./81.*m33*RD1456ep*RS14563pe + 1./108.*m33*
         RS1456pee*RD1456ep + 1./108.*m33*RS1456ppe*RD1456ep - 1./108.*
         m33*RS145pe*RD1456ep - 1./108.*m33*RS146pe*RD1456ep - 1./108.*
         m33*RS156pe*RD1456ep + 5./64.*pow(m33,2) + 41./1296.*pow(
         m33,2)*pow(RS14563pe,2) + 7./432.*pow(m33,2)*RS1456pee*
         RS14563pe + 5./576.*pow(m33,2)*pow(RS1456pee,2) + 7./432.*pow(
         m33,2)*RS1456ppe*RS14563pe + 5./288.*pow(m33,2)*RS1456ppe*
         RS1456pee + 5./576.*pow(m33,2)*pow(RS1456ppe,2) - 7./432.*pow(
         m33,2)*RS145pe*RS14563pe - 5./288.*pow(m33,2)*RS145pe*
         RS1456pee - 5./288.*pow(m33,2)*RS145pe*RS1456ppe + 5./576.*
         pow(m33,2)*pow(RS145pe,2) - 7./432.*pow(m33,2)*RS146pe*
         RS14563pe - 5./288.*pow(m33,2)*RS146pe*RS1456pee - 5./288.*
         pow(m33,2)*RS146pe*RS1456ppe + 5./288.*pow(m33,2)*RS146pe*
         RS145pe + 5./576.*pow(m33,2)*pow(RS146pe,2) - 7./432.*pow(
         m33,2)*RS156pe*RS14563pe );

      F13p6x23R +=  + hhd(1,m11,m11,m13,m13,mu2) * (  - 5./288.*pow(
         m33,2)*RS156pe*RS1456pee - 5./288.*pow(m33,2)*RS156pe*
         RS1456ppe + 5./288.*pow(m33,2)*RS156pe*RS145pe + 5./288.*pow(
         m33,2)*RS156pe*RS146pe + 5./576.*pow(m33,2)*pow(RS156pe,2) + 1.
         /162.*m11*RD1456ep*RS14563pe + 1./108.*m11*RS1456pee*RD1456ep
          + 1./108.*m11*RS1456ppe*RD1456ep - 1./108.*m11*RS145pe*
         RD1456ep - 1./108.*m11*RS146pe*RD1456ep - 1./108.*m11*RS156pe*
         RD1456ep + 3./32.*m11*m33 + 25./648.*m11*m33*pow(RS14563pe,2)
          + 1./72.*m11*m33*RS1456pee*RS14563pe + 5./288.*m11*m33*pow(
         RS1456pee,2) + 1./72.*m11*m33*RS1456ppe*RS14563pe + 5./144.*
         m11*m33*RS1456ppe*RS1456pee + 5./288.*m11*m33*pow(RS1456ppe,2)
          - 1./72.*m11*m33*RS145pe*RS14563pe - 5./144.*m11*m33*RS145pe*
         RS1456pee - 5./144.*m11*m33*RS145pe*RS1456ppe + 5./288.*m11*
         m33*pow(RS145pe,2) - 1./72.*m11*m33*RS146pe*RS14563pe - 5./144.
         *m11*m33*RS146pe*RS1456pee - 5./144.*m11*m33*RS146pe*RS1456ppe
          + 5./144.*m11*m33*RS146pe*RS145pe );

      F13p6x23R +=  + hhd(1,m11,m11,m13,m13,mu2) * ( 5./288.*m11*m33*
         pow(RS146pe,2) - 1./72.*m11*m33*RS156pe*RS14563pe - 5./144.*
         m11*m33*RS156pe*RS1456pee - 5./144.*m11*m33*RS156pe*RS1456ppe
          + 5./144.*m11*m33*RS156pe*RS145pe + 5./144.*m11*m33*RS156pe*
         RS146pe + 5./288.*m11*m33*pow(RS156pe,2) + 1./64.*pow(m11,2)
          + 17./1296.*pow(m11,2)*pow(RS14563pe,2) - 1./432.*pow(m11,2)*
         RS1456pee*RS14563pe + 5./576.*pow(m11,2)*pow(RS1456pee,2) - 1./
         432.*pow(m11,2)*RS1456ppe*RS14563pe + 5./288.*pow(m11,2)*
         RS1456ppe*RS1456pee + 5./576.*pow(m11,2)*pow(RS1456ppe,2) + 1./
         432.*pow(m11,2)*RS145pe*RS14563pe - 5./288.*pow(m11,2)*RS145pe
         *RS1456pee - 5./288.*pow(m11,2)*RS145pe*RS1456ppe + 5./576.*
         pow(m11,2)*pow(RS145pe,2) + 1./432.*pow(m11,2)*RS146pe*
         RS14563pe - 5./288.*pow(m11,2)*RS146pe*RS1456pee - 5./288.*
         pow(m11,2)*RS146pe*RS1456ppe + 5./288.*pow(m11,2)*RS146pe*
         RS145pe + 5./576.*pow(m11,2)*pow(RS146pe,2) + 1./432.*pow(
         m11,2)*RS156pe*RS14563pe );

      F13p6x23R +=  + hhd(1,m11,m11,m13,m13,mu2) * (  - 5./288.*pow(
         m11,2)*RS156pe*RS1456pee - 5./288.*pow(m11,2)*RS156pe*
         RS1456ppe + 5./288.*pow(m11,2)*RS156pe*RS145pe + 5./288.*pow(
         m11,2)*RS156pe*RS146pe + 5./576.*pow(m11,2)*pow(RS156pe,2) );

      F13p6x23R +=  + hhd(1,m11,m13,m33,m13,mu2) * ( 1./324.*RD1456ep*
         RD3456ep + 1./81.*m33*RD3456ep*RS14563pe + 1./162.*m33*
         RD1456ep*RS34561pe + 25./648.*pow(m33,2)*RS14563pe*RS34561pe
          + 1./144.*pow(m33,2)*RS3456pee*RS14563pe + 1./144.*pow(m33,2)
         *RS3456ppe*RS14563pe - 1./144.*pow(m33,2)*RS345pe*RS14563pe - 
         1./144.*pow(m33,2)*RS346pe*RS14563pe - 1./144.*pow(m33,2)*
         RS356pe*RS14563pe + 1./144.*pow(m33,2)*RS1456pee*RS34561pe + 1.
         /288.*pow(m33,2)*RS1456pee*RS3456pee + 1./288.*pow(m33,2)*
         RS1456pee*RS3456ppe - 1./288.*pow(m33,2)*RS1456pee*RS345pe - 1.
         /288.*pow(m33,2)*RS1456pee*RS346pe - 1./288.*pow(m33,2)*
         RS1456pee*RS356pe + 1./144.*pow(m33,2)*RS1456ppe*RS34561pe + 1.
         /288.*pow(m33,2)*RS1456ppe*RS3456pee + 1./288.*pow(m33,2)*
         RS1456ppe*RS3456ppe - 1./288.*pow(m33,2)*RS1456ppe*RS345pe - 1.
         /288.*pow(m33,2)*RS1456ppe*RS346pe - 1./288.*pow(m33,2)*
         RS1456ppe*RS356pe - 1./144.*pow(m33,2)*RS145pe*RS34561pe - 1./
         288.*pow(m33,2)*RS145pe*RS3456pee );

      F13p6x23R +=  + hhd(1,m11,m13,m33,m13,mu2) * (  - 1./288.*pow(
         m33,2)*RS145pe*RS3456ppe + 1./288.*pow(m33,2)*RS145pe*RS345pe
          + 1./288.*pow(m33,2)*RS145pe*RS346pe + 1./288.*pow(m33,2)*
         RS145pe*RS356pe - 1./144.*pow(m33,2)*RS146pe*RS34561pe - 1./
         288.*pow(m33,2)*RS146pe*RS3456pee - 1./288.*pow(m33,2)*RS146pe
         *RS3456ppe + 1./288.*pow(m33,2)*RS146pe*RS345pe + 1./288.*pow(
         m33,2)*RS146pe*RS346pe + 1./288.*pow(m33,2)*RS146pe*RS356pe - 
         1./144.*pow(m33,2)*RS156pe*RS34561pe - 1./288.*pow(m33,2)*
         RS156pe*RS3456pee - 1./288.*pow(m33,2)*RS156pe*RS3456ppe + 1./
         288.*pow(m33,2)*RS156pe*RS345pe + 1./288.*pow(m33,2)*RS156pe*
         RS346pe + 1./288.*pow(m33,2)*RS156pe*RS356pe + 1./162.*m11*
         RD3456ep*RS14563pe + 1./81.*m11*RD1456ep*RS34561pe + 29./324.*
         m11*m33*RS14563pe*RS34561pe + 1./72.*m11*m33*RS3456pee*
         RS14563pe + 1./72.*m11*m33*RS3456ppe*RS14563pe - 1./72.*m11*
         m33*RS345pe*RS14563pe - 1./72.*m11*m33*RS346pe*RS14563pe - 1./
         72.*m11*m33*RS356pe*RS14563pe );

      F13p6x23R +=  + hhd(1,m11,m13,m33,m13,mu2) * ( 1./72.*m11*m33*
         RS1456pee*RS34561pe + 1./144.*m11*m33*RS1456pee*RS3456pee + 1./
         144.*m11*m33*RS1456pee*RS3456ppe - 1./144.*m11*m33*RS1456pee*
         RS345pe - 1./144.*m11*m33*RS1456pee*RS346pe - 1./144.*m11*m33*
         RS1456pee*RS356pe + 1./72.*m11*m33*RS1456ppe*RS34561pe + 1./
         144.*m11*m33*RS1456ppe*RS3456pee + 1./144.*m11*m33*RS1456ppe*
         RS3456ppe - 1./144.*m11*m33*RS1456ppe*RS345pe - 1./144.*m11*
         m33*RS1456ppe*RS346pe - 1./144.*m11*m33*RS1456ppe*RS356pe - 1./
         72.*m11*m33*RS145pe*RS34561pe - 1./144.*m11*m33*RS145pe*
         RS3456pee - 1./144.*m11*m33*RS145pe*RS3456ppe + 1./144.*m11*
         m33*RS145pe*RS345pe + 1./144.*m11*m33*RS145pe*RS346pe + 1./144.
         *m11*m33*RS145pe*RS356pe - 1./72.*m11*m33*RS146pe*RS34561pe - 
         1./144.*m11*m33*RS146pe*RS3456pee - 1./144.*m11*m33*RS146pe*
         RS3456ppe + 1./144.*m11*m33*RS146pe*RS345pe + 1./144.*m11*m33*
         RS146pe*RS346pe + 1./144.*m11*m33*RS146pe*RS356pe - 1./72.*m11
         *m33*RS156pe*RS34561pe );

      F13p6x23R +=  + hhd(1,m11,m13,m33,m13,mu2) * (  - 1./144.*m11*m33
         *RS156pe*RS3456pee - 1./144.*m11*m33*RS156pe*RS3456ppe + 1./
         144.*m11*m33*RS156pe*RS345pe + 1./144.*m11*m33*RS156pe*RS346pe
          + 1./144.*m11*m33*RS156pe*RS356pe + 25./648.*pow(m11,2)*
         RS14563pe*RS34561pe + 1./144.*pow(m11,2)*RS3456pee*RS14563pe
          + 1./144.*pow(m11,2)*RS3456ppe*RS14563pe - 1./144.*pow(m11,2)
         *RS345pe*RS14563pe - 1./144.*pow(m11,2)*RS346pe*RS14563pe - 1./
         144.*pow(m11,2)*RS356pe*RS14563pe + 1./144.*pow(m11,2)*
         RS1456pee*RS34561pe + 1./288.*pow(m11,2)*RS1456pee*RS3456pee
          + 1./288.*pow(m11,2)*RS1456pee*RS3456ppe - 1./288.*pow(m11,2)
         *RS1456pee*RS345pe - 1./288.*pow(m11,2)*RS1456pee*RS346pe - 1./
         288.*pow(m11,2)*RS1456pee*RS356pe + 1./144.*pow(m11,2)*
         RS1456ppe*RS34561pe + 1./288.*pow(m11,2)*RS1456ppe*RS3456pee
          + 1./288.*pow(m11,2)*RS1456ppe*RS3456ppe - 1./288.*pow(m11,2)
         *RS1456ppe*RS345pe - 1./288.*pow(m11,2)*RS1456ppe*RS346pe - 1./
         288.*pow(m11,2)*RS1456ppe*RS356pe );

      F13p6x23R +=  + hhd(1,m11,m13,m33,m13,mu2) * (  - 1./144.*pow(
         m11,2)*RS145pe*RS34561pe - 1./288.*pow(m11,2)*RS145pe*
         RS3456pee - 1./288.*pow(m11,2)*RS145pe*RS3456ppe + 1./288.*
         pow(m11,2)*RS145pe*RS345pe + 1./288.*pow(m11,2)*RS145pe*
         RS346pe + 1./288.*pow(m11,2)*RS145pe*RS356pe - 1./144.*pow(
         m11,2)*RS146pe*RS34561pe - 1./288.*pow(m11,2)*RS146pe*
         RS3456pee - 1./288.*pow(m11,2)*RS146pe*RS3456ppe + 1./288.*
         pow(m11,2)*RS146pe*RS345pe + 1./288.*pow(m11,2)*RS146pe*
         RS346pe + 1./288.*pow(m11,2)*RS146pe*RS356pe - 1./144.*pow(
         m11,2)*RS156pe*RS34561pe - 1./288.*pow(m11,2)*RS156pe*
         RS3456pee - 1./288.*pow(m11,2)*RS156pe*RS3456ppe + 1./288.*
         pow(m11,2)*RS156pe*RS345pe + 1./288.*pow(m11,2)*RS156pe*
         RS346pe + 1./288.*pow(m11,2)*RS156pe*RS356pe );

      F13p6x23R +=  + hhd(1,m11,m13,me2,m13,mu2) * (  - 1./162.*me2*
         RD1456ep*RSe45613p + 1./324.*me2*RD1456ep*RSe45633p + 1./324.*
         me2*RSe45611p*RD1456ep + 1./162.*pow(me2,2)*RSe45613p*
         RS14563pe - 1./324.*pow(me2,2)*RSe45633p*RS14563pe + 1./324.*
         pow(me2,2)*RS1456pee*RSe45613p - 1./648.*pow(me2,2)*RS1456pee*
         RSe45633p + 1./324.*pow(me2,2)*RS1456ppe*RSe45613p - 1./648.*
         pow(me2,2)*RS1456ppe*RSe45633p - 1./324.*pow(me2,2)*RSe45611p*
         RS14563pe - 1./648.*pow(me2,2)*RSe45611p*RS1456pee - 1./648.*
         pow(me2,2)*RSe45611p*RS1456ppe - 1./324.*pow(me2,2)*RS145pe*
         RSe45613p + 1./648.*pow(me2,2)*RS145pe*RSe45633p + 1./648.*
         pow(me2,2)*RS145pe*RSe45611p - 1./324.*pow(me2,2)*RS146pe*
         RSe45613p + 1./648.*pow(me2,2)*RS146pe*RSe45633p + 1./648.*
         pow(me2,2)*RS146pe*RSe45611p - 1./324.*pow(me2,2)*RS156pe*
         RSe45613p + 1./648.*pow(me2,2)*RS156pe*RSe45633p + 1./648.*
         pow(me2,2)*RS156pe*RSe45611p + 1./81.*m33*RD1456ep*RSe45613p
          - 1./324.*m33*RD1456ep*RSe45633p );

      F13p6x23R +=  + hhd(1,m11,m13,me2,m13,mu2) * (  - 1./108.*m33*
         RSe45611p*RD1456ep - 5./162.*m33*me2*RSe45613p*RS14563pe + 1./
         54.*m33*me2*RSe45633p*RS14563pe - 1./81.*m33*me2*RS1456pee*
         RSe45613p + 1./324.*m33*me2*RS1456pee*RSe45633p - 1./81.*m33*
         me2*RS1456ppe*RSe45613p + 1./324.*m33*me2*RS1456ppe*RSe45633p
          + 1./81.*m33*me2*RSe45611p*RS14563pe + 1./108.*m33*me2*
         RSe45611p*RS1456pee + 1./108.*m33*me2*RSe45611p*RS1456ppe + 1./
         81.*m33*me2*RS145pe*RSe45613p - 1./324.*m33*me2*RS145pe*
         RSe45633p - 1./108.*m33*me2*RS145pe*RSe45611p + 1./81.*m33*me2
         *RS146pe*RSe45613p - 1./324.*m33*me2*RS146pe*RSe45633p - 1./
         108.*m33*me2*RS146pe*RSe45611p + 1./81.*m33*me2*RS156pe*
         RSe45613p - 1./324.*m33*me2*RS156pe*RSe45633p - 1./108.*m33*
         me2*RS156pe*RSe45611p + 41./648.*pow(m33,2)*RSe45613p*
         RS14563pe - 29./1296.*pow(m33,2)*RSe45633p*RS14563pe + 7./432.
         *pow(m33,2)*RS1456pee*RSe45613p - 13./2592.*pow(m33,2)*
         RS1456pee*RSe45633p );

      F13p6x23R +=  + hhd(1,m11,m13,me2,m13,mu2) * ( 7./432.*pow(m33,2)
         *RS1456ppe*RSe45613p - 13./2592.*pow(m33,2)*RS1456ppe*
         RSe45633p - 7./432.*pow(m33,2)*RSe45611p*RS14563pe - 5./288.*
         pow(m33,2)*RSe45611p*RS1456pee - 5./288.*pow(m33,2)*RSe45611p*
         RS1456ppe - 7./432.*pow(m33,2)*RS145pe*RSe45613p + 13./2592.*
         pow(m33,2)*RS145pe*RSe45633p + 5./288.*pow(m33,2)*RS145pe*
         RSe45611p - 7./432.*pow(m33,2)*RS146pe*RSe45613p + 13./2592.*
         pow(m33,2)*RS146pe*RSe45633p + 5./288.*pow(m33,2)*RS146pe*
         RSe45611p - 7./432.*pow(m33,2)*RS156pe*RSe45613p + 13./2592.*
         pow(m33,2)*RS156pe*RSe45633p + 5./288.*pow(m33,2)*RS156pe*
         RSe45611p + 1./81.*m11*RD1456ep*RSe45613p - 1./81.*m11*
         RSe45611p*RD1456ep - 1./54.*m11*me2*RSe45613p*RS14563pe + 1./
         162.*m11*me2*RSe45633p*RS14563pe - 1./81.*m11*me2*RS1456pee*
         RSe45613p - 1./81.*m11*me2*RS1456ppe*RSe45613p + 1./81.*m11*
         me2*RSe45611p*RS14563pe + 1./81.*m11*me2*RSe45611p*RS1456pee
          + 1./81.*m11*me2*RSe45611p*RS1456ppe );

      F13p6x23R +=  + hhd(1,m11,m13,me2,m13,mu2) * ( 1./81.*m11*me2*
         RS145pe*RSe45613p - 1./81.*m11*me2*RS145pe*RSe45611p + 1./81.*
         m11*me2*RS146pe*RSe45613p - 1./81.*m11*me2*RS146pe*RSe45611p
          + 1./81.*m11*me2*RS156pe*RSe45613p - 1./81.*m11*me2*RS156pe*
         RSe45611p + 35./324.*m11*m33*RSe45613p*RS14563pe - 13./648.*
         m11*m33*RSe45633p*RS14563pe + 17./648.*m11*m33*RS1456pee*
         RSe45613p - 1./144.*m11*m33*RS1456pee*RSe45633p + 17./648.*m11
         *m33*RS1456ppe*RSe45613p - 1./144.*m11*m33*RS1456ppe*RSe45633p
          - 17./648.*m11*m33*RSe45611p*RS14563pe - 19./432.*m11*m33*
         RSe45611p*RS1456pee - 19./432.*m11*m33*RSe45611p*RS1456ppe - 
         17./648.*m11*m33*RS145pe*RSe45613p + 1./144.*m11*m33*RS145pe*
         RSe45633p + 19./432.*m11*m33*RS145pe*RSe45611p - 17./648.*m11*
         m33*RS146pe*RSe45613p + 1./144.*m11*m33*RS146pe*RSe45633p + 19.
         /432.*m11*m33*RS146pe*RSe45611p - 17./648.*m11*m33*RS156pe*
         RSe45613p + 1./144.*m11*m33*RS156pe*RSe45633p + 19./432.*m11*
         m33*RS156pe*RSe45611p );

      F13p6x23R +=  + hhd(1,m11,m13,me2,m13,mu2) * ( 25./648.*pow(
         m11,2)*RSe45613p*RS14563pe - 1./144.*pow(m11,2)*RSe45633p*
         RS14563pe + 1./144.*pow(m11,2)*RS1456pee*RSe45613p - 1./288.*
         pow(m11,2)*RS1456pee*RSe45633p + 1./144.*pow(m11,2)*RS1456ppe*
         RSe45613p - 1./288.*pow(m11,2)*RS1456ppe*RSe45633p - 1./144.*
         pow(m11,2)*RSe45611p*RS14563pe - 73./2592.*pow(m11,2)*
         RSe45611p*RS1456pee - 73./2592.*pow(m11,2)*RSe45611p*RS1456ppe
          - 1./144.*pow(m11,2)*RS145pe*RSe45613p + 1./288.*pow(m11,2)*
         RS145pe*RSe45633p + 73./2592.*pow(m11,2)*RS145pe*RSe45611p - 1.
         /144.*pow(m11,2)*RS146pe*RSe45613p + 1./288.*pow(m11,2)*
         RS146pe*RSe45633p + 73./2592.*pow(m11,2)*RS146pe*RSe45611p - 1.
         /144.*pow(m11,2)*RS156pe*RSe45613p + 1./288.*pow(m11,2)*
         RS156pe*RSe45633p + 73./2592.*pow(m11,2)*RS156pe*RSe45611p );

      F13p6x23R +=  + hhd(1,m11,m13,mp2,m13,mu2) * (  - 1./162.*mp2*
         RD1456ep*RSp45613e + 1./324.*mp2*RD1456ep*RSp45633e + 1./324.*
         mp2*RSp45611e*RD1456ep + 1./162.*pow(mp2,2)*RSp45613e*
         RS14563pe - 1./324.*pow(mp2,2)*RSp45633e*RS14563pe + 1./324.*
         pow(mp2,2)*RS1456pee*RSp45613e - 1./648.*pow(mp2,2)*RS1456pee*
         RSp45633e + 1./324.*pow(mp2,2)*RS1456ppe*RSp45613e - 1./648.*
         pow(mp2,2)*RS1456ppe*RSp45633e - 1./324.*pow(mp2,2)*RSp45611e*
         RS14563pe - 1./648.*pow(mp2,2)*RSp45611e*RS1456pee - 1./648.*
         pow(mp2,2)*RSp45611e*RS1456ppe - 1./324.*pow(mp2,2)*RS145pe*
         RSp45613e + 1./648.*pow(mp2,2)*RS145pe*RSp45633e + 1./648.*
         pow(mp2,2)*RS145pe*RSp45611e - 1./324.*pow(mp2,2)*RS146pe*
         RSp45613e + 1./648.*pow(mp2,2)*RS146pe*RSp45633e + 1./648.*
         pow(mp2,2)*RS146pe*RSp45611e - 1./324.*pow(mp2,2)*RS156pe*
         RSp45613e + 1./648.*pow(mp2,2)*RS156pe*RSp45633e + 1./648.*
         pow(mp2,2)*RS156pe*RSp45611e + 1./81.*m33*RD1456ep*RSp45613e
          - 1./324.*m33*RD1456ep*RSp45633e );

      F13p6x23R +=  + hhd(1,m11,m13,mp2,m13,mu2) * (  - 1./108.*m33*
         RSp45611e*RD1456ep - 5./162.*m33*mp2*RSp45613e*RS14563pe + 1./
         54.*m33*mp2*RSp45633e*RS14563pe - 1./81.*m33*mp2*RS1456pee*
         RSp45613e + 1./324.*m33*mp2*RS1456pee*RSp45633e - 1./81.*m33*
         mp2*RS1456ppe*RSp45613e + 1./324.*m33*mp2*RS1456ppe*RSp45633e
          + 1./81.*m33*mp2*RSp45611e*RS14563pe + 1./108.*m33*mp2*
         RSp45611e*RS1456pee + 1./108.*m33*mp2*RSp45611e*RS1456ppe + 1./
         81.*m33*mp2*RS145pe*RSp45613e - 1./324.*m33*mp2*RS145pe*
         RSp45633e - 1./108.*m33*mp2*RS145pe*RSp45611e + 1./81.*m33*mp2
         *RS146pe*RSp45613e - 1./324.*m33*mp2*RS146pe*RSp45633e - 1./
         108.*m33*mp2*RS146pe*RSp45611e + 1./81.*m33*mp2*RS156pe*
         RSp45613e - 1./324.*m33*mp2*RS156pe*RSp45633e - 1./108.*m33*
         mp2*RS156pe*RSp45611e + 41./648.*pow(m33,2)*RSp45613e*
         RS14563pe - 29./1296.*pow(m33,2)*RSp45633e*RS14563pe + 7./432.
         *pow(m33,2)*RS1456pee*RSp45613e - 13./2592.*pow(m33,2)*
         RS1456pee*RSp45633e );

      F13p6x23R +=  + hhd(1,m11,m13,mp2,m13,mu2) * ( 7./432.*pow(m33,2)
         *RS1456ppe*RSp45613e - 13./2592.*pow(m33,2)*RS1456ppe*
         RSp45633e - 7./432.*pow(m33,2)*RSp45611e*RS14563pe - 5./288.*
         pow(m33,2)*RSp45611e*RS1456pee - 5./288.*pow(m33,2)*RSp45611e*
         RS1456ppe - 7./432.*pow(m33,2)*RS145pe*RSp45613e + 13./2592.*
         pow(m33,2)*RS145pe*RSp45633e + 5./288.*pow(m33,2)*RS145pe*
         RSp45611e - 7./432.*pow(m33,2)*RS146pe*RSp45613e + 13./2592.*
         pow(m33,2)*RS146pe*RSp45633e + 5./288.*pow(m33,2)*RS146pe*
         RSp45611e - 7./432.*pow(m33,2)*RS156pe*RSp45613e + 13./2592.*
         pow(m33,2)*RS156pe*RSp45633e + 5./288.*pow(m33,2)*RS156pe*
         RSp45611e + 1./81.*m11*RD1456ep*RSp45613e - 1./81.*m11*
         RSp45611e*RD1456ep - 1./54.*m11*mp2*RSp45613e*RS14563pe + 1./
         162.*m11*mp2*RSp45633e*RS14563pe - 1./81.*m11*mp2*RS1456pee*
         RSp45613e - 1./81.*m11*mp2*RS1456ppe*RSp45613e + 1./81.*m11*
         mp2*RSp45611e*RS14563pe + 1./81.*m11*mp2*RSp45611e*RS1456pee
          + 1./81.*m11*mp2*RSp45611e*RS1456ppe );

      F13p6x23R +=  + hhd(1,m11,m13,mp2,m13,mu2) * ( 1./81.*m11*mp2*
         RS145pe*RSp45613e - 1./81.*m11*mp2*RS145pe*RSp45611e + 1./81.*
         m11*mp2*RS146pe*RSp45613e - 1./81.*m11*mp2*RS146pe*RSp45611e
          + 1./81.*m11*mp2*RS156pe*RSp45613e - 1./81.*m11*mp2*RS156pe*
         RSp45611e + 35./324.*m11*m33*RSp45613e*RS14563pe - 13./648.*
         m11*m33*RSp45633e*RS14563pe + 17./648.*m11*m33*RS1456pee*
         RSp45613e - 1./144.*m11*m33*RS1456pee*RSp45633e + 17./648.*m11
         *m33*RS1456ppe*RSp45613e - 1./144.*m11*m33*RS1456ppe*RSp45633e
          - 17./648.*m11*m33*RSp45611e*RS14563pe - 19./432.*m11*m33*
         RSp45611e*RS1456pee - 19./432.*m11*m33*RSp45611e*RS1456ppe - 
         17./648.*m11*m33*RS145pe*RSp45613e + 1./144.*m11*m33*RS145pe*
         RSp45633e + 19./432.*m11*m33*RS145pe*RSp45611e - 17./648.*m11*
         m33*RS146pe*RSp45613e + 1./144.*m11*m33*RS146pe*RSp45633e + 19.
         /432.*m11*m33*RS146pe*RSp45611e - 17./648.*m11*m33*RS156pe*
         RSp45613e + 1./144.*m11*m33*RS156pe*RSp45633e + 19./432.*m11*
         m33*RS156pe*RSp45611e );

      F13p6x23R +=  + hhd(1,m11,m13,mp2,m13,mu2) * ( 25./648.*pow(
         m11,2)*RSp45613e*RS14563pe - 1./144.*pow(m11,2)*RSp45633e*
         RS14563pe + 1./144.*pow(m11,2)*RS1456pee*RSp45613e - 1./288.*
         pow(m11,2)*RS1456pee*RSp45633e + 1./144.*pow(m11,2)*RS1456ppe*
         RSp45613e - 1./288.*pow(m11,2)*RS1456ppe*RSp45633e - 1./144.*
         pow(m11,2)*RSp45611e*RS14563pe - 73./2592.*pow(m11,2)*
         RSp45611e*RS1456pee - 73./2592.*pow(m11,2)*RSp45611e*RS1456ppe
          - 1./144.*pow(m11,2)*RS145pe*RSp45613e + 1./288.*pow(m11,2)*
         RS145pe*RSp45633e + 73./2592.*pow(m11,2)*RS145pe*RSp45611e - 1.
         /144.*pow(m11,2)*RS146pe*RSp45613e + 1./288.*pow(m11,2)*
         RS146pe*RSp45633e + 73./2592.*pow(m11,2)*RS146pe*RSp45611e - 1.
         /144.*pow(m11,2)*RS156pe*RSp45613e + 1./288.*pow(m11,2)*
         RS156pe*RSp45633e + 73./2592.*pow(m11,2)*RS156pe*RSp45611e );

      F13p6x23R +=  + hhd(1,m11,m14,m34,m13,mu2) * ( 1./48.*m33*m44*
         RS14563pe + 1./96.*m33*m44*RS1456pee + 1./96.*m33*m44*
         RS1456ppe - 1./96.*m33*m44*RS145pe - 1./96.*m33*m44*RS146pe - 
         1./96.*m33*m44*RS156pe - 1./32.*pow(m33,2)*RS14563pe + 1./32.*
         pow(m33,2)*RS156pe + 1./48.*m11*m44*RS14563pe + 1./96.*m11*m44
         *RS1456pee + 1./96.*m11*m44*RS1456ppe - 1./96.*m11*m44*RS145pe
          - 1./96.*m11*m44*RS146pe - 1./96.*m11*m44*RS156pe - 1./12.*
         m11*m33*RS14563pe + 1./96.*m11*m33*RS1456pee + 1./96.*m11*m33*
         RS1456ppe - 1./96.*m11*m33*RS145pe - 1./96.*m11*m33*RS146pe + 
         3./32.*m11*m33*RS156pe - 5./96.*pow(m11,2)*RS14563pe + 1./96.*
         pow(m11,2)*RS1456pee + 1./96.*pow(m11,2)*RS1456ppe - 1./96.*
         pow(m11,2)*RS145pe - 1./96.*pow(m11,2)*RS146pe + 1./16.*pow(
         m11,2)*RS156pe );

      F13p6x23R +=  + hhd(1,m11,m15,m35,m13,mu2) * ( 1./48.*m33*m55*
         RS14563pe + 1./96.*m33*m55*RS1456pee + 1./96.*m33*m55*
         RS1456ppe - 1./96.*m33*m55*RS145pe - 1./96.*m33*m55*RS146pe - 
         1./96.*m33*m55*RS156pe - 1./32.*pow(m33,2)*RS14563pe + 1./32.*
         pow(m33,2)*RS146pe + 1./48.*m11*m55*RS14563pe + 1./96.*m11*m55
         *RS1456pee + 1./96.*m11*m55*RS1456ppe - 1./96.*m11*m55*RS145pe
          - 1./96.*m11*m55*RS146pe - 1./96.*m11*m55*RS156pe - 1./12.*
         m11*m33*RS14563pe + 1./96.*m11*m33*RS1456pee + 1./96.*m11*m33*
         RS1456ppe - 1./96.*m11*m33*RS145pe + 3./32.*m11*m33*RS146pe - 
         1./96.*m11*m33*RS156pe - 5./96.*pow(m11,2)*RS14563pe + 1./96.*
         pow(m11,2)*RS1456pee + 1./96.*pow(m11,2)*RS1456ppe - 1./96.*
         pow(m11,2)*RS145pe + 1./16.*pow(m11,2)*RS146pe - 1./96.*pow(
         m11,2)*RS156pe );

      F13p6x23R +=  + hhd(1,m11,m16,m36,m13,mu2) * ( 1./48.*m33*m66*
         RS14563pe + 1./96.*m33*m66*RS1456pee + 1./96.*m33*m66*
         RS1456ppe - 1./96.*m33*m66*RS145pe - 1./96.*m33*m66*RS146pe - 
         1./96.*m33*m66*RS156pe - 1./32.*pow(m33,2)*RS14563pe + 1./32.*
         pow(m33,2)*RS145pe + 1./48.*m11*m66*RS14563pe + 1./96.*m11*m66
         *RS1456pee + 1./96.*m11*m66*RS1456ppe - 1./96.*m11*m66*RS145pe
          - 1./96.*m11*m66*RS146pe - 1./96.*m11*m66*RS156pe - 1./12.*
         m11*m33*RS14563pe + 1./96.*m11*m33*RS1456pee + 1./96.*m11*m33*
         RS1456ppe + 3./32.*m11*m33*RS145pe - 1./96.*m11*m33*RS146pe - 
         1./96.*m11*m33*RS156pe - 5./96.*pow(m11,2)*RS14563pe + 1./96.*
         pow(m11,2)*RS1456pee + 1./96.*pow(m11,2)*RS1456ppe + 1./16.*
         pow(m11,2)*RS145pe - 1./96.*pow(m11,2)*RS146pe - 1./96.*pow(
         m11,2)*RS156pe );

      F13p6x23R +=  + hhd(1,m13,m13,m13,m13,mu2) * ( 1./32.*pow(m33,2)
          + 1./16.*m11*m33 + 1./32.*pow(m11,2) );

      F13p6x23R +=  + hhd(1,m13,m33,m33,m13,mu2) * ( 1./648.*pow(
         RD3456ep,2) + 1./162.*m33*RD3456ep*RS34561pe + 1./108.*m33*
         RS3456pee*RD3456ep + 1./108.*m33*RS3456ppe*RD3456ep - 1./108.*
         m33*RS345pe*RD3456ep - 1./108.*m33*RS346pe*RD3456ep - 1./108.*
         m33*RS356pe*RD3456ep + 1./64.*pow(m33,2) + 17./1296.*pow(
         m33,2)*pow(RS34561pe,2) - 1./432.*pow(m33,2)*RS3456pee*
         RS34561pe + 5./576.*pow(m33,2)*pow(RS3456pee,2) - 1./432.*pow(
         m33,2)*RS3456ppe*RS34561pe + 5./288.*pow(m33,2)*RS3456ppe*
         RS3456pee + 5./576.*pow(m33,2)*pow(RS3456ppe,2) + 1./432.*pow(
         m33,2)*RS345pe*RS34561pe - 5./288.*pow(m33,2)*RS345pe*
         RS3456pee - 5./288.*pow(m33,2)*RS345pe*RS3456ppe + 5./576.*
         pow(m33,2)*pow(RS345pe,2) + 1./432.*pow(m33,2)*RS346pe*
         RS34561pe - 5./288.*pow(m33,2)*RS346pe*RS3456pee - 5./288.*
         pow(m33,2)*RS346pe*RS3456ppe + 5./288.*pow(m33,2)*RS346pe*
         RS345pe + 5./576.*pow(m33,2)*pow(RS346pe,2) + 1./432.*pow(
         m33,2)*RS356pe*RS34561pe );

      F13p6x23R +=  + hhd(1,m13,m33,m33,m13,mu2) * (  - 5./288.*pow(
         m33,2)*RS356pe*RS3456pee - 5./288.*pow(m33,2)*RS356pe*
         RS3456ppe + 5./288.*pow(m33,2)*RS356pe*RS345pe + 5./288.*pow(
         m33,2)*RS356pe*RS346pe + 5./576.*pow(m33,2)*pow(RS356pe,2) + 1.
         /81.*m11*RD3456ep*RS34561pe + 1./108.*m11*RS3456pee*RD3456ep
          + 1./108.*m11*RS3456ppe*RD3456ep - 1./108.*m11*RS345pe*
         RD3456ep - 1./108.*m11*RS346pe*RD3456ep - 1./108.*m11*RS356pe*
         RD3456ep + 3./32.*m11*m33 + 25./648.*m11*m33*pow(RS34561pe,2)
          + 1./72.*m11*m33*RS3456pee*RS34561pe + 5./288.*m11*m33*pow(
         RS3456pee,2) + 1./72.*m11*m33*RS3456ppe*RS34561pe + 5./144.*
         m11*m33*RS3456ppe*RS3456pee + 5./288.*m11*m33*pow(RS3456ppe,2)
          - 1./72.*m11*m33*RS345pe*RS34561pe - 5./144.*m11*m33*RS345pe*
         RS3456pee - 5./144.*m11*m33*RS345pe*RS3456ppe + 5./288.*m11*
         m33*pow(RS345pe,2) - 1./72.*m11*m33*RS346pe*RS34561pe - 5./144.
         *m11*m33*RS346pe*RS3456pee - 5./144.*m11*m33*RS346pe*RS3456ppe
          + 5./144.*m11*m33*RS346pe*RS345pe );

      F13p6x23R +=  + hhd(1,m13,m33,m33,m13,mu2) * ( 5./288.*m11*m33*
         pow(RS346pe,2) - 1./72.*m11*m33*RS356pe*RS34561pe - 5./144.*
         m11*m33*RS356pe*RS3456pee - 5./144.*m11*m33*RS356pe*RS3456ppe
          + 5./144.*m11*m33*RS356pe*RS345pe + 5./144.*m11*m33*RS356pe*
         RS346pe + 5./288.*m11*m33*pow(RS356pe,2) + 5./64.*pow(m11,2)
          + 41./1296.*pow(m11,2)*pow(RS34561pe,2) + 7./432.*pow(m11,2)*
         RS3456pee*RS34561pe + 5./576.*pow(m11,2)*pow(RS3456pee,2) + 7./
         432.*pow(m11,2)*RS3456ppe*RS34561pe + 5./288.*pow(m11,2)*
         RS3456ppe*RS3456pee + 5./576.*pow(m11,2)*pow(RS3456ppe,2) - 7./
         432.*pow(m11,2)*RS345pe*RS34561pe - 5./288.*pow(m11,2)*RS345pe
         *RS3456pee - 5./288.*pow(m11,2)*RS345pe*RS3456ppe + 5./576.*
         pow(m11,2)*pow(RS345pe,2) - 7./432.*pow(m11,2)*RS346pe*
         RS34561pe - 5./288.*pow(m11,2)*RS346pe*RS3456pee - 5./288.*
         pow(m11,2)*RS346pe*RS3456ppe + 5./288.*pow(m11,2)*RS346pe*
         RS345pe + 5./576.*pow(m11,2)*pow(RS346pe,2) - 7./432.*pow(
         m11,2)*RS356pe*RS34561pe );

      F13p6x23R +=  + hhd(1,m13,m33,m33,m13,mu2) * (  - 5./288.*pow(
         m11,2)*RS356pe*RS3456pee - 5./288.*pow(m11,2)*RS356pe*
         RS3456ppe + 5./288.*pow(m11,2)*RS356pe*RS345pe + 5./288.*pow(
         m11,2)*RS356pe*RS346pe + 5./576.*pow(m11,2)*pow(RS356pe,2) );

      F13p6x23R +=  + hhd(1,m13,m33,me2,m13,mu2) * (  - 1./162.*me2*
         RD3456ep*RSe45613p + 1./324.*me2*RSe45633p*RD3456ep + 1./324.*
         me2*RSe45611p*RD3456ep + 1./162.*pow(me2,2)*RSe45613p*
         RS34561pe + 1./324.*pow(me2,2)*RS3456pee*RSe45613p + 1./324.*
         pow(me2,2)*RS3456ppe*RSe45613p - 1./324.*pow(me2,2)*RSe45633p*
         RS34561pe - 1./648.*pow(me2,2)*RSe45633p*RS3456pee - 1./648.*
         pow(me2,2)*RSe45633p*RS3456ppe - 1./324.*pow(me2,2)*RS345pe*
         RSe45613p + 1./648.*pow(me2,2)*RS345pe*RSe45633p - 1./324.*
         pow(me2,2)*RS346pe*RSe45613p + 1./648.*pow(me2,2)*RS346pe*
         RSe45633p - 1./324.*pow(me2,2)*RS356pe*RSe45613p + 1./648.*
         pow(me2,2)*RS356pe*RSe45633p - 1./324.*pow(me2,2)*RSe45611p*
         RS34561pe - 1./648.*pow(me2,2)*RSe45611p*RS3456pee - 1./648.*
         pow(me2,2)*RSe45611p*RS3456ppe + 1./648.*pow(me2,2)*RSe45611p*
         RS345pe + 1./648.*pow(me2,2)*RSe45611p*RS346pe + 1./648.*pow(
         me2,2)*RSe45611p*RS356pe + 1./81.*m33*RD3456ep*RSe45613p - 1./
         81.*m33*RSe45633p*RD3456ep );

      F13p6x23R +=  + hhd(1,m13,m33,me2,m13,mu2) * (  - 1./54.*m33*me2*
         RSe45613p*RS34561pe - 1./81.*m33*me2*RS3456pee*RSe45613p - 1./
         81.*m33*me2*RS3456ppe*RSe45613p + 1./81.*m33*me2*RSe45633p*
         RS34561pe + 1./81.*m33*me2*RSe45633p*RS3456pee + 1./81.*m33*
         me2*RSe45633p*RS3456ppe + 1./81.*m33*me2*RS345pe*RSe45613p - 1.
         /81.*m33*me2*RS345pe*RSe45633p + 1./81.*m33*me2*RS346pe*
         RSe45613p - 1./81.*m33*me2*RS346pe*RSe45633p + 1./81.*m33*me2*
         RS356pe*RSe45613p - 1./81.*m33*me2*RS356pe*RSe45633p + 1./162.
         *m33*me2*RSe45611p*RS34561pe + 25./648.*pow(m33,2)*RSe45613p*
         RS34561pe + 1./144.*pow(m33,2)*RS3456pee*RSe45613p + 1./144.*
         pow(m33,2)*RS3456ppe*RSe45613p - 1./144.*pow(m33,2)*RSe45633p*
         RS34561pe - 73./2592.*pow(m33,2)*RSe45633p*RS3456pee - 73./
         2592.*pow(m33,2)*RSe45633p*RS3456ppe - 1./144.*pow(m33,2)*
         RS345pe*RSe45613p + 73./2592.*pow(m33,2)*RS345pe*RSe45633p - 1.
         /144.*pow(m33,2)*RS346pe*RSe45613p + 73./2592.*pow(m33,2)*
         RS346pe*RSe45633p );

      F13p6x23R +=  + hhd(1,m13,m33,me2,m13,mu2) * (  - 1./144.*pow(
         m33,2)*RS356pe*RSe45613p + 73./2592.*pow(m33,2)*RS356pe*
         RSe45633p - 1./144.*pow(m33,2)*RSe45611p*RS34561pe - 1./288.*
         pow(m33,2)*RSe45611p*RS3456pee - 1./288.*pow(m33,2)*RSe45611p*
         RS3456ppe + 1./288.*pow(m33,2)*RSe45611p*RS345pe + 1./288.*
         pow(m33,2)*RSe45611p*RS346pe + 1./288.*pow(m33,2)*RSe45611p*
         RS356pe + 1./81.*m11*RD3456ep*RSe45613p - 1./108.*m11*
         RSe45633p*RD3456ep - 1./324.*m11*RSe45611p*RD3456ep - 5./162.*
         m11*me2*RSe45613p*RS34561pe - 1./81.*m11*me2*RS3456pee*
         RSe45613p - 1./81.*m11*me2*RS3456ppe*RSe45613p + 1./81.*m11*
         me2*RSe45633p*RS34561pe + 1./108.*m11*me2*RSe45633p*RS3456pee
          + 1./108.*m11*me2*RSe45633p*RS3456ppe + 1./81.*m11*me2*
         RS345pe*RSe45613p - 1./108.*m11*me2*RS345pe*RSe45633p + 1./81.
         *m11*me2*RS346pe*RSe45613p - 1./108.*m11*me2*RS346pe*RSe45633p
          + 1./81.*m11*me2*RS356pe*RSe45613p - 1./108.*m11*me2*RS356pe*
         RSe45633p );

      F13p6x23R +=  + hhd(1,m13,m33,me2,m13,mu2) * ( 1./54.*m11*me2*
         RSe45611p*RS34561pe + 1./324.*m11*me2*RSe45611p*RS3456pee + 1./
         324.*m11*me2*RSe45611p*RS3456ppe - 1./324.*m11*me2*RSe45611p*
         RS345pe - 1./324.*m11*me2*RSe45611p*RS346pe - 1./324.*m11*me2*
         RSe45611p*RS356pe + 35./324.*m11*m33*RSe45613p*RS34561pe + 17./
         648.*m11*m33*RS3456pee*RSe45613p + 17./648.*m11*m33*RS3456ppe*
         RSe45613p - 17./648.*m11*m33*RSe45633p*RS34561pe - 19./432.*
         m11*m33*RSe45633p*RS3456pee - 19./432.*m11*m33*RSe45633p*
         RS3456ppe - 17./648.*m11*m33*RS345pe*RSe45613p + 19./432.*m11*
         m33*RS345pe*RSe45633p - 17./648.*m11*m33*RS346pe*RSe45613p + 
         19./432.*m11*m33*RS346pe*RSe45633p - 17./648.*m11*m33*RS356pe*
         RSe45613p + 19./432.*m11*m33*RS356pe*RSe45633p - 13./648.*m11*
         m33*RSe45611p*RS34561pe - 1./144.*m11*m33*RSe45611p*RS3456pee
          - 1./144.*m11*m33*RSe45611p*RS3456ppe + 1./144.*m11*m33*
         RSe45611p*RS345pe + 1./144.*m11*m33*RSe45611p*RS346pe + 1./144.
         *m11*m33*RSe45611p*RS356pe );

      F13p6x23R +=  + hhd(1,m13,m33,me2,m13,mu2) * ( 41./648.*pow(
         m11,2)*RSe45613p*RS34561pe + 7./432.*pow(m11,2)*RS3456pee*
         RSe45613p + 7./432.*pow(m11,2)*RS3456ppe*RSe45613p - 7./432.*
         pow(m11,2)*RSe45633p*RS34561pe - 5./288.*pow(m11,2)*RSe45633p*
         RS3456pee - 5./288.*pow(m11,2)*RSe45633p*RS3456ppe - 7./432.*
         pow(m11,2)*RS345pe*RSe45613p + 5./288.*pow(m11,2)*RS345pe*
         RSe45633p - 7./432.*pow(m11,2)*RS346pe*RSe45613p + 5./288.*
         pow(m11,2)*RS346pe*RSe45633p - 7./432.*pow(m11,2)*RS356pe*
         RSe45613p + 5./288.*pow(m11,2)*RS356pe*RSe45633p - 29./1296.*
         pow(m11,2)*RSe45611p*RS34561pe - 13./2592.*pow(m11,2)*
         RSe45611p*RS3456pee - 13./2592.*pow(m11,2)*RSe45611p*RS3456ppe
          + 13./2592.*pow(m11,2)*RSe45611p*RS345pe + 13./2592.*pow(
         m11,2)*RSe45611p*RS346pe + 13./2592.*pow(m11,2)*RSe45611p*
         RS356pe );

      F13p6x23R +=  + hhd(1,m13,m33,mp2,m13,mu2) * (  - 1./162.*mp2*
         RD3456ep*RSp45613e + 1./324.*mp2*RSp45633e*RD3456ep + 1./324.*
         mp2*RSp45611e*RD3456ep + 1./162.*pow(mp2,2)*RSp45613e*
         RS34561pe + 1./324.*pow(mp2,2)*RS3456pee*RSp45613e + 1./324.*
         pow(mp2,2)*RS3456ppe*RSp45613e - 1./324.*pow(mp2,2)*RSp45633e*
         RS34561pe - 1./648.*pow(mp2,2)*RSp45633e*RS3456pee - 1./648.*
         pow(mp2,2)*RSp45633e*RS3456ppe - 1./324.*pow(mp2,2)*RS345pe*
         RSp45613e + 1./648.*pow(mp2,2)*RS345pe*RSp45633e - 1./324.*
         pow(mp2,2)*RS346pe*RSp45613e + 1./648.*pow(mp2,2)*RS346pe*
         RSp45633e - 1./324.*pow(mp2,2)*RS356pe*RSp45613e + 1./648.*
         pow(mp2,2)*RS356pe*RSp45633e - 1./324.*pow(mp2,2)*RSp45611e*
         RS34561pe - 1./648.*pow(mp2,2)*RSp45611e*RS3456pee - 1./648.*
         pow(mp2,2)*RSp45611e*RS3456ppe + 1./648.*pow(mp2,2)*RSp45611e*
         RS345pe + 1./648.*pow(mp2,2)*RSp45611e*RS346pe + 1./648.*pow(
         mp2,2)*RSp45611e*RS356pe + 1./81.*m33*RD3456ep*RSp45613e - 1./
         81.*m33*RSp45633e*RD3456ep );

      F13p6x23R +=  + hhd(1,m13,m33,mp2,m13,mu2) * (  - 1./54.*m33*mp2*
         RSp45613e*RS34561pe - 1./81.*m33*mp2*RS3456pee*RSp45613e - 1./
         81.*m33*mp2*RS3456ppe*RSp45613e + 1./81.*m33*mp2*RSp45633e*
         RS34561pe + 1./81.*m33*mp2*RSp45633e*RS3456pee + 1./81.*m33*
         mp2*RSp45633e*RS3456ppe + 1./81.*m33*mp2*RS345pe*RSp45613e - 1.
         /81.*m33*mp2*RS345pe*RSp45633e + 1./81.*m33*mp2*RS346pe*
         RSp45613e - 1./81.*m33*mp2*RS346pe*RSp45633e + 1./81.*m33*mp2*
         RS356pe*RSp45613e - 1./81.*m33*mp2*RS356pe*RSp45633e + 1./162.
         *m33*mp2*RSp45611e*RS34561pe + 25./648.*pow(m33,2)*RSp45613e*
         RS34561pe + 1./144.*pow(m33,2)*RS3456pee*RSp45613e + 1./144.*
         pow(m33,2)*RS3456ppe*RSp45613e - 1./144.*pow(m33,2)*RSp45633e*
         RS34561pe - 73./2592.*pow(m33,2)*RSp45633e*RS3456pee - 73./
         2592.*pow(m33,2)*RSp45633e*RS3456ppe - 1./144.*pow(m33,2)*
         RS345pe*RSp45613e + 73./2592.*pow(m33,2)*RS345pe*RSp45633e - 1.
         /144.*pow(m33,2)*RS346pe*RSp45613e + 73./2592.*pow(m33,2)*
         RS346pe*RSp45633e );

      F13p6x23R +=  + hhd(1,m13,m33,mp2,m13,mu2) * (  - 1./144.*pow(
         m33,2)*RS356pe*RSp45613e + 73./2592.*pow(m33,2)*RS356pe*
         RSp45633e - 1./144.*pow(m33,2)*RSp45611e*RS34561pe - 1./288.*
         pow(m33,2)*RSp45611e*RS3456pee - 1./288.*pow(m33,2)*RSp45611e*
         RS3456ppe + 1./288.*pow(m33,2)*RSp45611e*RS345pe + 1./288.*
         pow(m33,2)*RSp45611e*RS346pe + 1./288.*pow(m33,2)*RSp45611e*
         RS356pe + 1./81.*m11*RD3456ep*RSp45613e - 1./108.*m11*
         RSp45633e*RD3456ep - 1./324.*m11*RSp45611e*RD3456ep - 5./162.*
         m11*mp2*RSp45613e*RS34561pe - 1./81.*m11*mp2*RS3456pee*
         RSp45613e - 1./81.*m11*mp2*RS3456ppe*RSp45613e + 1./81.*m11*
         mp2*RSp45633e*RS34561pe + 1./108.*m11*mp2*RSp45633e*RS3456pee
          + 1./108.*m11*mp2*RSp45633e*RS3456ppe + 1./81.*m11*mp2*
         RS345pe*RSp45613e - 1./108.*m11*mp2*RS345pe*RSp45633e + 1./81.
         *m11*mp2*RS346pe*RSp45613e - 1./108.*m11*mp2*RS346pe*RSp45633e
          + 1./81.*m11*mp2*RS356pe*RSp45613e - 1./108.*m11*mp2*RS356pe*
         RSp45633e );

      F13p6x23R +=  + hhd(1,m13,m33,mp2,m13,mu2) * ( 1./54.*m11*mp2*
         RSp45611e*RS34561pe + 1./324.*m11*mp2*RSp45611e*RS3456pee + 1./
         324.*m11*mp2*RSp45611e*RS3456ppe - 1./324.*m11*mp2*RSp45611e*
         RS345pe - 1./324.*m11*mp2*RSp45611e*RS346pe - 1./324.*m11*mp2*
         RSp45611e*RS356pe + 35./324.*m11*m33*RSp45613e*RS34561pe + 17./
         648.*m11*m33*RS3456pee*RSp45613e + 17./648.*m11*m33*RS3456ppe*
         RSp45613e - 17./648.*m11*m33*RSp45633e*RS34561pe - 19./432.*
         m11*m33*RSp45633e*RS3456pee - 19./432.*m11*m33*RSp45633e*
         RS3456ppe - 17./648.*m11*m33*RS345pe*RSp45613e + 19./432.*m11*
         m33*RS345pe*RSp45633e - 17./648.*m11*m33*RS346pe*RSp45613e + 
         19./432.*m11*m33*RS346pe*RSp45633e - 17./648.*m11*m33*RS356pe*
         RSp45613e + 19./432.*m11*m33*RS356pe*RSp45633e - 13./648.*m11*
         m33*RSp45611e*RS34561pe - 1./144.*m11*m33*RSp45611e*RS3456pee
          - 1./144.*m11*m33*RSp45611e*RS3456ppe + 1./144.*m11*m33*
         RSp45611e*RS345pe + 1./144.*m11*m33*RSp45611e*RS346pe + 1./144.
         *m11*m33*RSp45611e*RS356pe );

      F13p6x23R +=  + hhd(1,m13,m33,mp2,m13,mu2) * ( 41./648.*pow(
         m11,2)*RSp45613e*RS34561pe + 7./432.*pow(m11,2)*RS3456pee*
         RSp45613e + 7./432.*pow(m11,2)*RS3456ppe*RSp45613e - 7./432.*
         pow(m11,2)*RSp45633e*RS34561pe - 5./288.*pow(m11,2)*RSp45633e*
         RS3456pee - 5./288.*pow(m11,2)*RSp45633e*RS3456ppe - 7./432.*
         pow(m11,2)*RS345pe*RSp45613e + 5./288.*pow(m11,2)*RS345pe*
         RSp45633e - 7./432.*pow(m11,2)*RS346pe*RSp45613e + 5./288.*
         pow(m11,2)*RS346pe*RSp45633e - 7./432.*pow(m11,2)*RS356pe*
         RSp45613e + 5./288.*pow(m11,2)*RS356pe*RSp45633e - 29./1296.*
         pow(m11,2)*RSp45611e*RS34561pe - 13./2592.*pow(m11,2)*
         RSp45611e*RS3456pee - 13./2592.*pow(m11,2)*RSp45611e*RS3456ppe
          + 13./2592.*pow(m11,2)*RSp45611e*RS345pe + 13./2592.*pow(
         m11,2)*RSp45611e*RS346pe + 13./2592.*pow(m11,2)*RSp45611e*
         RS356pe );

      F13p6x23R +=  + hhd(1,m13,me2,me2,m13,mu2) * ( 1./81.*pow(me2,2)*
         pow(RSe45613p,2) - 1./81.*pow(me2,2)*RSe45633p*RSe45613p + 1./
         324.*pow(me2,2)*pow(RSe45633p,2) - 1./81.*pow(me2,2)*RSe45611p
         *RSe45613p + 1./162.*pow(me2,2)*RSe45611p*RSe45633p + 1./324.*
         pow(me2,2)*pow(RSe45611p,2) - 5./162.*m33*me2*pow(RSe45613p,2)
          + 1./27.*m33*me2*RSe45633p*RSe45613p - 5./324.*m33*me2*pow(
         RSe45633p,2) + 2./81.*m33*me2*RSe45611p*RSe45613p - 1./162.*
         m33*me2*RSe45611p*RSe45633p - 1./108.*m33*me2*pow(RSe45611p,2)
          + 41./1296.*pow(m33,2)*pow(RSe45613p,2) - 29./1296.*pow(
         m33,2)*RSe45633p*RSe45613p + 109./5184.*pow(m33,2)*pow(
         RSe45633p,2) - 7./432.*pow(m33,2)*RSe45611p*RSe45613p + 13./
         2592.*pow(m33,2)*RSe45611p*RSe45633p + 5./576.*pow(m33,2)*pow(
         RSe45611p,2) - 5./162.*m11*me2*pow(RSe45613p,2) + 2./81.*m11*
         me2*RSe45633p*RSe45613p - 1./108.*m11*me2*pow(RSe45633p,2) + 1.
         /27.*m11*me2*RSe45611p*RSe45613p - 1./162.*m11*me2*RSe45611p*
         RSe45633p );

      F13p6x23R +=  + hhd(1,m13,me2,me2,m13,mu2) * (  - 5./324.*m11*me2
         *pow(RSe45611p,2) + 5./72.*m11*m33*pow(RSe45613p,2) - 25./648.
         *m11*m33*RSe45633p*RSe45613p + 23./864.*m11*m33*pow(
         RSe45633p,2) - 25./648.*m11*m33*RSe45611p*RSe45613p + 13./1296.
         *m11*m33*RSe45611p*RSe45633p + 23./864.*m11*m33*pow(
         RSe45611p,2) + 41./1296.*pow(m11,2)*pow(RSe45613p,2) - 7./432.
         *pow(m11,2)*RSe45633p*RSe45613p + 5./576.*pow(m11,2)*pow(
         RSe45633p,2) - 29./1296.*pow(m11,2)*RSe45611p*RSe45613p + 13./
         2592.*pow(m11,2)*RSe45611p*RSe45633p + 109./5184.*pow(m11,2)*
         pow(RSe45611p,2) );

      F13p6x23R +=  + hhd(1,m13,me2,mp2,m13,mu2) * ( 1./162.*pow(mp2,2)
         *RSp45613e*RSe45613p - 1./324.*pow(mp2,2)*RSe45633p*RSp45613e
          - 1./324.*pow(mp2,2)*RSp45633e*RSe45613p + 1./648.*pow(mp2,2)
         *RSp45633e*RSe45633p - 1./324.*pow(mp2,2)*RSe45611p*RSp45613e
          + 1./648.*pow(mp2,2)*RSe45611p*RSp45633e - 1./324.*pow(mp2,2)
         *RSp45611e*RSe45613p + 1./648.*pow(mp2,2)*RSp45611e*RSe45633p
          + 1./648.*pow(mp2,2)*RSp45611e*RSe45611p + 1./81.*me2*mp2*
         RSp45613e*RSe45613p - 1./162.*me2*mp2*RSe45633p*RSp45613e - 1./
         162.*me2*mp2*RSp45633e*RSe45613p + 1./324.*me2*mp2*RSp45633e*
         RSe45633p - 1./162.*me2*mp2*RSe45611p*RSp45613e + 1./324.*me2*
         mp2*RSe45611p*RSp45633e - 1./162.*me2*mp2*RSp45611e*RSe45613p
          + 1./324.*me2*mp2*RSp45611e*RSe45633p + 1./324.*me2*mp2*
         RSp45611e*RSe45611p + 1./162.*pow(me2,2)*RSp45613e*RSe45613p
          - 1./324.*pow(me2,2)*RSe45633p*RSp45613e - 1./324.*pow(me2,2)
         *RSp45633e*RSe45613p + 1./648.*pow(me2,2)*RSp45633e*RSe45633p
          - 1./324.*pow(me2,2)*RSe45611p*RSp45613e );

      F13p6x23R +=  + hhd(1,m13,me2,mp2,m13,mu2) * ( 1./648.*pow(me2,2)
         *RSe45611p*RSp45633e - 1./324.*pow(me2,2)*RSp45611e*RSe45613p
          + 1./648.*pow(me2,2)*RSp45611e*RSe45633p + 1./648.*pow(me2,2)
         *RSp45611e*RSe45611p - 5./162.*m33*mp2*RSp45613e*RSe45613p + 1.
         /54.*m33*mp2*RSe45633p*RSp45613e + 1./54.*m33*mp2*RSp45633e*
         RSe45613p - 5./324.*m33*mp2*RSp45633e*RSe45633p + 1./81.*m33*
         mp2*RSe45611p*RSp45613e - 1./324.*m33*mp2*RSe45611p*RSp45633e
          + 1./81.*m33*mp2*RSp45611e*RSe45613p - 1./324.*m33*mp2*
         RSp45611e*RSe45633p - 1./108.*m33*mp2*RSp45611e*RSe45611p - 5./
         162.*m33*me2*RSp45613e*RSe45613p + 1./54.*m33*me2*RSe45633p*
         RSp45613e + 1./54.*m33*me2*RSp45633e*RSe45613p - 5./324.*m33*
         me2*RSp45633e*RSe45633p + 1./81.*m33*me2*RSe45611p*RSp45613e
          - 1./324.*m33*me2*RSe45611p*RSp45633e + 1./81.*m33*me2*
         RSp45611e*RSe45613p - 1./324.*m33*me2*RSp45611e*RSe45633p - 1./
         108.*m33*me2*RSp45611e*RSe45611p + 41./648.*pow(m33,2)*
         RSp45613e*RSe45613p );

      F13p6x23R +=  + hhd(1,m13,me2,mp2,m13,mu2) * (  - 29./1296.*pow(
         m33,2)*RSe45633p*RSp45613e - 29./1296.*pow(m33,2)*RSp45633e*
         RSe45613p + 109./2592.*pow(m33,2)*RSp45633e*RSe45633p - 7./432.
         *pow(m33,2)*RSe45611p*RSp45613e + 13./2592.*pow(m33,2)*
         RSe45611p*RSp45633e - 7./432.*pow(m33,2)*RSp45611e*RSe45613p
          + 13./2592.*pow(m33,2)*RSp45611e*RSe45633p + 5./288.*pow(
         m33,2)*RSp45611e*RSe45611p - 5./162.*m11*mp2*RSp45613e*
         RSe45613p + 1./81.*m11*mp2*RSe45633p*RSp45613e + 1./81.*m11*
         mp2*RSp45633e*RSe45613p - 1./108.*m11*mp2*RSp45633e*RSe45633p
          + 1./54.*m11*mp2*RSe45611p*RSp45613e - 1./324.*m11*mp2*
         RSe45611p*RSp45633e + 1./54.*m11*mp2*RSp45611e*RSe45613p - 1./
         324.*m11*mp2*RSp45611e*RSe45633p - 5./324.*m11*mp2*RSp45611e*
         RSe45611p - 5./162.*m11*me2*RSp45613e*RSe45613p + 1./81.*m11*
         me2*RSe45633p*RSp45613e + 1./81.*m11*me2*RSp45633e*RSe45613p
          - 1./108.*m11*me2*RSp45633e*RSe45633p + 1./54.*m11*me2*
         RSe45611p*RSp45613e );

      F13p6x23R +=  + hhd(1,m13,me2,mp2,m13,mu2) * (  - 1./324.*m11*me2
         *RSe45611p*RSp45633e + 1./54.*m11*me2*RSp45611e*RSe45613p - 1./
         324.*m11*me2*RSp45611e*RSe45633p - 5./324.*m11*me2*RSp45611e*
         RSe45611p + 5./36.*m11*m33*RSp45613e*RSe45613p - 25./648.*m11*
         m33*RSe45633p*RSp45613e - 25./648.*m11*m33*RSp45633e*RSe45613p
          + 23./432.*m11*m33*RSp45633e*RSe45633p - 25./648.*m11*m33*
         RSe45611p*RSp45613e + 13./1296.*m11*m33*RSe45611p*RSp45633e - 
         25./648.*m11*m33*RSp45611e*RSe45613p + 13./1296.*m11*m33*
         RSp45611e*RSe45633p + 23./432.*m11*m33*RSp45611e*RSe45611p + 
         41./648.*pow(m11,2)*RSp45613e*RSe45613p - 7./432.*pow(m11,2)*
         RSe45633p*RSp45613e - 7./432.*pow(m11,2)*RSp45633e*RSe45613p
          + 5./288.*pow(m11,2)*RSp45633e*RSe45633p - 29./1296.*pow(
         m11,2)*RSe45611p*RSp45613e + 13./2592.*pow(m11,2)*RSe45611p*
         RSp45633e - 29./1296.*pow(m11,2)*RSp45611e*RSe45613p + 13./
         2592.*pow(m11,2)*RSp45611e*RSe45633p + 109./2592.*pow(m11,2)*
         RSp45611e*RSe45611p );

      F13p6x23R +=  + hhd(1,m13,mp2,mp2,m13,mu2) * ( 1./81.*pow(mp2,2)*
         pow(RSp45613e,2) - 1./81.*pow(mp2,2)*RSp45633e*RSp45613e + 1./
         324.*pow(mp2,2)*pow(RSp45633e,2) - 1./81.*pow(mp2,2)*RSp45611e
         *RSp45613e + 1./162.*pow(mp2,2)*RSp45611e*RSp45633e + 1./324.*
         pow(mp2,2)*pow(RSp45611e,2) - 5./162.*m33*mp2*pow(RSp45613e,2)
          + 1./27.*m33*mp2*RSp45633e*RSp45613e - 5./324.*m33*mp2*pow(
         RSp45633e,2) + 2./81.*m33*mp2*RSp45611e*RSp45613e - 1./162.*
         m33*mp2*RSp45611e*RSp45633e - 1./108.*m33*mp2*pow(RSp45611e,2)
          + 41./1296.*pow(m33,2)*pow(RSp45613e,2) - 29./1296.*pow(
         m33,2)*RSp45633e*RSp45613e + 109./5184.*pow(m33,2)*pow(
         RSp45633e,2) - 7./432.*pow(m33,2)*RSp45611e*RSp45613e + 13./
         2592.*pow(m33,2)*RSp45611e*RSp45633e + 5./576.*pow(m33,2)*pow(
         RSp45611e,2) - 5./162.*m11*mp2*pow(RSp45613e,2) + 2./81.*m11*
         mp2*RSp45633e*RSp45613e - 1./108.*m11*mp2*pow(RSp45633e,2) + 1.
         /27.*m11*mp2*RSp45611e*RSp45613e - 1./162.*m11*mp2*RSp45611e*
         RSp45633e );

      F13p6x23R +=  + hhd(1,m13,mp2,mp2,m13,mu2) * (  - 5./324.*m11*mp2
         *pow(RSp45611e,2) + 5./72.*m11*m33*pow(RSp45613e,2) - 25./648.
         *m11*m33*RSp45633e*RSp45613e + 23./864.*m11*m33*pow(
         RSp45633e,2) - 25./648.*m11*m33*RSp45611e*RSp45613e + 13./1296.
         *m11*m33*RSp45611e*RSp45633e + 23./864.*m11*m33*pow(
         RSp45611e,2) + 41./1296.*pow(m11,2)*pow(RSp45613e,2) - 7./432.
         *pow(m11,2)*RSp45633e*RSp45613e + 5./576.*pow(m11,2)*pow(
         RSp45633e,2) - 29./1296.*pow(m11,2)*RSp45611e*RSp45613e + 13./
         2592.*pow(m11,2)*RSp45611e*RSp45633e + 109./5184.*pow(m11,2)*
         pow(RSp45611e,2) );

      F13p6x23R +=  + hhd(1,m14,m33,m34,m13,mu2) * ( 1./48.*m33*m44*
         RS34561pe + 1./96.*m33*m44*RS3456pee + 1./96.*m33*m44*
         RS3456ppe - 1./96.*m33*m44*RS345pe - 1./96.*m33*m44*RS346pe - 
         1./96.*m33*m44*RS356pe + 1./32.*pow(m33,2)*RS34561pe + 1./96.*
         pow(m33,2)*RS3456pee + 1./96.*pow(m33,2)*RS3456ppe - 1./96.*
         pow(m33,2)*RS345pe - 1./96.*pow(m33,2)*RS346pe - 1./48.*pow(
         m33,2)*RS356pe + 1./48.*m11*m44*RS34561pe + 1./96.*m11*m44*
         RS3456pee + 1./96.*m11*m44*RS3456ppe - 1./96.*m11*m44*RS345pe
          - 1./96.*m11*m44*RS346pe - 1./96.*m11*m44*RS356pe + 1./12.*
         m11*m33*RS34561pe + 1./96.*m11*m33*RS3456pee + 1./96.*m11*m33*
         RS3456ppe - 1./96.*m11*m33*RS345pe - 1./96.*m11*m33*RS346pe - 
         7./96.*m11*m33*RS356pe + 5./96.*pow(m11,2)*RS34561pe - 5./96.*
         pow(m11,2)*RS356pe );

      F13p6x23R +=  + hhd(1,m14,m34,m44,m13,mu2) * ( 1./16.*m33*m44 - 1.
         /48.*m33*m44*RS456pe + 1./16.*m11*m44 - 1./48.*m11*m44*RS456pe
          );

      F13p6x23R +=  + hhd(1,m14,m34,me2,m13,mu2) * (  - 1./108.*pow(
         me2,2)*RSe45613p - 1./216.*pow(me2,2)*RSe45633p + 1./54.*pow(
         me2,2)*RSe563p - 1./216.*pow(me2,2)*RSe45611p + 1./54.*pow(
         me2,2)*RSe561p - 1./54.*pow(me2,2)*RSe564p - 1./54.*m44*me2*
         RSe563p - 1./54.*m44*me2*RSe561p + 1./27.*m44*me2*RSe564p - 1./
         54.*pow(m44,2)*RSe564p - 17./432.*m33*me2*RSe45613p - 1./54.*
         m33*me2*RSe45633p + 25./432.*m33*me2*RSe563p + 1./48.*m33*me2*
         RSe561p - 1./48.*m33*me2*RSe564p + 1./48.*m33*m44*RSe45613p - 
         1./96.*m33*m44*RSe45633p - 1./27.*m33*m44*RSe563p - 1./96.*m33
         *m44*RSe45611p - 1./32.*pow(m33,2)*RSe45613p + 11./864.*pow(
         m33,2)*RSe45633p - 1./32.*pow(m33,2)*RSe563p + 1./32.*pow(
         m33,2)*RSe561p - 17./432.*m11*me2*RSe45613p - 1./36.*m11*me2*
         RSe45633p + 11./144.*m11*me2*RSe563p + 1./108.*m11*me2*
         RSe45611p + 1./432.*m11*me2*RSe561p - 1./48.*m11*me2*RSe564p
          + 1./48.*m11*m44*RSe45613p - 1./96.*m11*m44*RSe45633p - 1./18.
         *m11*m44*RSe563p );

      F13p6x23R +=  + hhd(1,m14,m34,me2,m13,mu2) * (  - 1./96.*m11*m44*
         RSe45611p + 1./54.*m11*m44*RSe561p - 19./432.*m11*m33*
         RSe45613p + 5./288.*m11*m33*RSe45633p - 1./12.*m11*m33*RSe563p
          - 1./96.*m11*m33*RSe45611p + 1./12.*m11*m33*RSe561p - 1./288.
         *pow(m11,2)*RSe45613p - 5./96.*pow(m11,2)*RSe563p - 13./864.*
         pow(m11,2)*RSe45611p + 5./96.*pow(m11,2)*RSe561p );

      F13p6x23R +=  + hhd(1,m14,m34,mp2,m13,mu2) * (  - 1./108.*pow(
         mp2,2)*RSp45613e - 1./216.*pow(mp2,2)*RSp45633e + 1./54.*pow(
         mp2,2)*RSp563e - 1./216.*pow(mp2,2)*RSp45611e + 1./54.*pow(
         mp2,2)*RSp561e - 1./54.*pow(mp2,2)*RSp564e - 1./54.*m44*mp2*
         RSp563e - 1./54.*m44*mp2*RSp561e + 1./27.*m44*mp2*RSp564e - 1./
         54.*pow(m44,2)*RSp564e - 17./432.*m33*mp2*RSp45613e - 1./54.*
         m33*mp2*RSp45633e + 25./432.*m33*mp2*RSp563e + 1./48.*m33*mp2*
         RSp561e - 1./48.*m33*mp2*RSp564e + 1./48.*m33*m44*RSp45613e - 
         1./96.*m33*m44*RSp45633e - 1./27.*m33*m44*RSp563e - 1./96.*m33
         *m44*RSp45611e - 1./32.*pow(m33,2)*RSp45613e + 11./864.*pow(
         m33,2)*RSp45633e - 1./32.*pow(m33,2)*RSp563e + 1./32.*pow(
         m33,2)*RSp561e - 17./432.*m11*mp2*RSp45613e - 1./36.*m11*mp2*
         RSp45633e + 11./144.*m11*mp2*RSp563e + 1./108.*m11*mp2*
         RSp45611e + 1./432.*m11*mp2*RSp561e - 1./48.*m11*mp2*RSp564e
          + 1./48.*m11*m44*RSp45613e - 1./96.*m11*m44*RSp45633e - 1./18.
         *m11*m44*RSp563e );

      F13p6x23R +=  + hhd(1,m14,m34,mp2,m13,mu2) * (  - 1./96.*m11*m44*
         RSp45611e + 1./54.*m11*m44*RSp561e - 19./432.*m11*m33*
         RSp45613e + 5./288.*m11*m33*RSp45633e - 1./12.*m11*m33*RSp563e
          - 1./96.*m11*m33*RSp45611e + 1./12.*m11*m33*RSp561e - 1./288.
         *pow(m11,2)*RSp45613e - 5./96.*pow(m11,2)*RSp563e - 13./864.*
         pow(m11,2)*RSp45611e + 5./96.*pow(m11,2)*RSp561e );

      F13p6x23R +=  + hhd(1,m14,m35,m45,m13,mu2) * ( 1./32.*m33*m55 + 1.
         /32.*m33*m44 + 1./32.*m11*m55 + 1./32.*m11*m44 );

      F13p6x23R +=  + hhd(1,m14,m36,m46,m13,mu2) * ( 1./32.*m33*m66 + 1.
         /32.*m33*m44 + 1./32.*m11*m66 + 1./32.*m11*m44 );

      F13p6x23R +=  + hhd(1,m15,m33,m35,m13,mu2) * ( 1./48.*m33*m55*
         RS34561pe + 1./96.*m33*m55*RS3456pee + 1./96.*m33*m55*
         RS3456ppe - 1./96.*m33*m55*RS345pe - 1./96.*m33*m55*RS346pe - 
         1./96.*m33*m55*RS356pe + 1./32.*pow(m33,2)*RS34561pe + 1./96.*
         pow(m33,2)*RS3456pee + 1./96.*pow(m33,2)*RS3456ppe - 1./96.*
         pow(m33,2)*RS345pe - 1./48.*pow(m33,2)*RS346pe - 1./96.*pow(
         m33,2)*RS356pe + 1./48.*m11*m55*RS34561pe + 1./96.*m11*m55*
         RS3456pee + 1./96.*m11*m55*RS3456ppe - 1./96.*m11*m55*RS345pe
          - 1./96.*m11*m55*RS346pe - 1./96.*m11*m55*RS356pe + 1./12.*
         m11*m33*RS34561pe + 1./96.*m11*m33*RS3456pee + 1./96.*m11*m33*
         RS3456ppe - 1./96.*m11*m33*RS345pe - 7./96.*m11*m33*RS346pe - 
         1./96.*m11*m33*RS356pe + 5./96.*pow(m11,2)*RS34561pe - 5./96.*
         pow(m11,2)*RS346pe );

      F13p6x23R +=  + hhd(1,m15,m34,m45,m13,mu2) * ( 1./32.*m33*m55 + 1.
         /32.*m33*m44 + 1./32.*m11*m55 + 1./32.*m11*m44 );

      F13p6x23R +=  + hhd(1,m15,m35,m55,m13,mu2) * ( 1./16.*m33*m55 - 1.
         /48.*m33*m55*RS546pe + 1./16.*m11*m55 - 1./48.*m11*m55*RS546pe
          );

      F13p6x23R +=  + hhd(1,m15,m35,me2,m13,mu2) * (  - 1./108.*pow(
         me2,2)*RSe45613p - 1./216.*pow(me2,2)*RSe45633p + 1./54.*pow(
         me2,2)*RSe463p - 1./216.*pow(me2,2)*RSe45611p + 1./54.*pow(
         me2,2)*RSe461p - 1./54.*pow(me2,2)*RSe465p - 1./54.*m55*me2*
         RSe463p - 1./54.*m55*me2*RSe461p + 1./27.*m55*me2*RSe465p - 1./
         54.*pow(m55,2)*RSe465p - 17./432.*m33*me2*RSe45613p - 1./54.*
         m33*me2*RSe45633p + 25./432.*m33*me2*RSe463p + 1./48.*m33*me2*
         RSe461p - 1./48.*m33*me2*RSe465p + 1./48.*m33*m55*RSe45613p - 
         1./96.*m33*m55*RSe45633p - 1./27.*m33*m55*RSe463p - 1./96.*m33
         *m55*RSe45611p - 1./32.*pow(m33,2)*RSe45613p + 11./864.*pow(
         m33,2)*RSe45633p - 1./32.*pow(m33,2)*RSe463p + 1./32.*pow(
         m33,2)*RSe461p - 17./432.*m11*me2*RSe45613p - 1./36.*m11*me2*
         RSe45633p + 11./144.*m11*me2*RSe463p + 1./108.*m11*me2*
         RSe45611p + 1./432.*m11*me2*RSe461p - 1./48.*m11*me2*RSe465p
          + 1./48.*m11*m55*RSe45613p - 1./96.*m11*m55*RSe45633p - 1./18.
         *m11*m55*RSe463p );

      F13p6x23R +=  + hhd(1,m15,m35,me2,m13,mu2) * (  - 1./96.*m11*m55*
         RSe45611p + 1./54.*m11*m55*RSe461p - 19./432.*m11*m33*
         RSe45613p + 5./288.*m11*m33*RSe45633p - 1./12.*m11*m33*RSe463p
          - 1./96.*m11*m33*RSe45611p + 1./12.*m11*m33*RSe461p - 1./288.
         *pow(m11,2)*RSe45613p - 5./96.*pow(m11,2)*RSe463p - 13./864.*
         pow(m11,2)*RSe45611p + 5./96.*pow(m11,2)*RSe461p );

      F13p6x23R +=  + hhd(1,m15,m35,mp2,m13,mu2) * (  - 1./108.*pow(
         mp2,2)*RSp45613e - 1./216.*pow(mp2,2)*RSp45633e + 1./54.*pow(
         mp2,2)*RSp463e - 1./216.*pow(mp2,2)*RSp45611e + 1./54.*pow(
         mp2,2)*RSp461e - 1./54.*pow(mp2,2)*RSp465e - 1./54.*m55*mp2*
         RSp463e - 1./54.*m55*mp2*RSp461e + 1./27.*m55*mp2*RSp465e - 1./
         54.*pow(m55,2)*RSp465e - 17./432.*m33*mp2*RSp45613e - 1./54.*
         m33*mp2*RSp45633e + 25./432.*m33*mp2*RSp463e + 1./48.*m33*mp2*
         RSp461e - 1./48.*m33*mp2*RSp465e + 1./48.*m33*m55*RSp45613e - 
         1./96.*m33*m55*RSp45633e - 1./27.*m33*m55*RSp463e - 1./96.*m33
         *m55*RSp45611e - 1./32.*pow(m33,2)*RSp45613e + 11./864.*pow(
         m33,2)*RSp45633e - 1./32.*pow(m33,2)*RSp463e + 1./32.*pow(
         m33,2)*RSp461e - 17./432.*m11*mp2*RSp45613e - 1./36.*m11*mp2*
         RSp45633e + 11./144.*m11*mp2*RSp463e + 1./108.*m11*mp2*
         RSp45611e + 1./432.*m11*mp2*RSp461e - 1./48.*m11*mp2*RSp465e
          + 1./48.*m11*m55*RSp45613e - 1./96.*m11*m55*RSp45633e - 1./18.
         *m11*m55*RSp463e );

      F13p6x23R +=  + hhd(1,m15,m35,mp2,m13,mu2) * (  - 1./96.*m11*m55*
         RSp45611e + 1./54.*m11*m55*RSp461e - 19./432.*m11*m33*
         RSp45613e + 5./288.*m11*m33*RSp45633e - 1./12.*m11*m33*RSp463e
          - 1./96.*m11*m33*RSp45611e + 1./12.*m11*m33*RSp461e - 1./288.
         *pow(m11,2)*RSp45613e - 5./96.*pow(m11,2)*RSp463e - 13./864.*
         pow(m11,2)*RSp45611e + 5./96.*pow(m11,2)*RSp461e );

      F13p6x23R +=  + hhd(1,m15,m36,m56,m13,mu2) * ( 1./32.*m33*m66 + 1.
         /32.*m33*m55 + 1./32.*m11*m66 + 1./32.*m11*m55 );

      F13p6x23R +=  + hhd(1,m16,m33,m36,m13,mu2) * ( 1./48.*m33*m66*
         RS34561pe + 1./96.*m33*m66*RS3456pee + 1./96.*m33*m66*
         RS3456ppe - 1./96.*m33*m66*RS345pe - 1./96.*m33*m66*RS346pe - 
         1./96.*m33*m66*RS356pe + 1./32.*pow(m33,2)*RS34561pe + 1./96.*
         pow(m33,2)*RS3456pee + 1./96.*pow(m33,2)*RS3456ppe - 1./48.*
         pow(m33,2)*RS345pe - 1./96.*pow(m33,2)*RS346pe - 1./96.*pow(
         m33,2)*RS356pe + 1./48.*m11*m66*RS34561pe + 1./96.*m11*m66*
         RS3456pee + 1./96.*m11*m66*RS3456ppe - 1./96.*m11*m66*RS345pe
          - 1./96.*m11*m66*RS346pe - 1./96.*m11*m66*RS356pe + 1./12.*
         m11*m33*RS34561pe + 1./96.*m11*m33*RS3456pee + 1./96.*m11*m33*
         RS3456ppe - 7./96.*m11*m33*RS345pe - 1./96.*m11*m33*RS346pe - 
         1./96.*m11*m33*RS356pe + 5./96.*pow(m11,2)*RS34561pe - 5./96.*
         pow(m11,2)*RS345pe );

      F13p6x23R +=  + hhd(1,m16,m34,m46,m13,mu2) * ( 1./32.*m33*m66 + 1.
         /32.*m33*m44 + 1./32.*m11*m66 + 1./32.*m11*m44 );

      F13p6x23R +=  + hhd(1,m16,m35,m56,m13,mu2) * ( 1./32.*m33*m66 + 1.
         /32.*m33*m55 + 1./32.*m11*m66 + 1./32.*m11*m55 );

      F13p6x23R +=  + hhd(1,m16,m36,m66,m13,mu2) * ( 1./16.*m33*m66 - 1.
         /48.*m33*m66*RS645pe + 1./16.*m11*m66 - 1./48.*m11*m66*RS645pe
          );

      F13p6x23R +=  + hhd(1,m16,m36,me2,m13,mu2) * (  - 1./108.*pow(
         me2,2)*RSe45613p - 1./216.*pow(me2,2)*RSe45633p + 1./54.*pow(
         me2,2)*RSe453p - 1./216.*pow(me2,2)*RSe45611p + 1./54.*pow(
         me2,2)*RSe451p - 1./54.*pow(me2,2)*RSe456p - 1./54.*m66*me2*
         RSe453p - 1./54.*m66*me2*RSe451p + 1./27.*m66*me2*RSe456p - 1./
         54.*pow(m66,2)*RSe456p - 17./432.*m33*me2*RSe45613p - 1./54.*
         m33*me2*RSe45633p + 25./432.*m33*me2*RSe453p + 1./48.*m33*me2*
         RSe451p - 1./48.*m33*me2*RSe456p + 1./48.*m33*m66*RSe45613p - 
         1./96.*m33*m66*RSe45633p - 1./27.*m33*m66*RSe453p - 1./96.*m33
         *m66*RSe45611p - 1./32.*pow(m33,2)*RSe45613p + 11./864.*pow(
         m33,2)*RSe45633p - 1./32.*pow(m33,2)*RSe453p + 1./32.*pow(
         m33,2)*RSe451p - 17./432.*m11*me2*RSe45613p - 1./36.*m11*me2*
         RSe45633p + 11./144.*m11*me2*RSe453p + 1./108.*m11*me2*
         RSe45611p + 1./432.*m11*me2*RSe451p - 1./48.*m11*me2*RSe456p
          + 1./48.*m11*m66*RSe45613p - 1./96.*m11*m66*RSe45633p - 1./18.
         *m11*m66*RSe453p );

      F13p6x23R +=  + hhd(1,m16,m36,me2,m13,mu2) * (  - 1./96.*m11*m66*
         RSe45611p + 1./54.*m11*m66*RSe451p - 19./432.*m11*m33*
         RSe45613p + 5./288.*m11*m33*RSe45633p - 1./12.*m11*m33*RSe453p
          - 1./96.*m11*m33*RSe45611p + 1./12.*m11*m33*RSe451p - 1./288.
         *pow(m11,2)*RSe45613p - 5./96.*pow(m11,2)*RSe453p - 13./864.*
         pow(m11,2)*RSe45611p + 5./96.*pow(m11,2)*RSe451p );

      F13p6x23R +=  + hhd(1,m16,m36,mp2,m13,mu2) * (  - 1./108.*pow(
         mp2,2)*RSp45613e - 1./216.*pow(mp2,2)*RSp45633e + 1./54.*pow(
         mp2,2)*RSp453e - 1./216.*pow(mp2,2)*RSp45611e + 1./54.*pow(
         mp2,2)*RSp451e - 1./54.*pow(mp2,2)*RSp456e - 1./54.*m66*mp2*
         RSp453e - 1./54.*m66*mp2*RSp451e + 1./27.*m66*mp2*RSp456e - 1./
         54.*pow(m66,2)*RSp456e - 17./432.*m33*mp2*RSp45613e - 1./54.*
         m33*mp2*RSp45633e + 25./432.*m33*mp2*RSp453e + 1./48.*m33*mp2*
         RSp451e - 1./48.*m33*mp2*RSp456e + 1./48.*m33*m66*RSp45613e - 
         1./96.*m33*m66*RSp45633e - 1./27.*m33*m66*RSp453e - 1./96.*m33
         *m66*RSp45611e - 1./32.*pow(m33,2)*RSp45613e + 11./864.*pow(
         m33,2)*RSp45633e - 1./32.*pow(m33,2)*RSp453e + 1./32.*pow(
         m33,2)*RSp451e - 17./432.*m11*mp2*RSp45613e - 1./36.*m11*mp2*
         RSp45633e + 11./144.*m11*mp2*RSp453e + 1./108.*m11*mp2*
         RSp45611e + 1./432.*m11*mp2*RSp451e - 1./48.*m11*mp2*RSp456e
          + 1./48.*m11*m66*RSp45613e - 1./96.*m11*m66*RSp45633e - 1./18.
         *m11*m66*RSp453e );

      F13p6x23R +=  + hhd(1,m16,m36,mp2,m13,mu2) * (  - 1./96.*m11*m66*
         RSp45611e + 1./54.*m11*m66*RSp451e - 19./432.*m11*m33*
         RSp45613e + 5./288.*m11*m33*RSp45633e - 1./12.*m11*m33*RSp453e
          - 1./96.*m11*m33*RSp45611e + 1./12.*m11*m33*RSp451e - 1./288.
         *pow(m11,2)*RSp45613e - 5./96.*pow(m11,2)*RSp453e - 13./864.*
         pow(m11,2)*RSp45611e + 5./96.*pow(m11,2)*RSp451e );

      F13p6x23R +=  + hhd(2,m11,m11,m13,m13,mu2) * (  - 1./108.*m33*
         pow(RD1456ep,2) - 7./432.*pow(m33,2)*RD1456ep*RS14563pe - 5./
         288.*pow(m33,2)*RS1456pee*RD1456ep - 5./288.*pow(m33,2)*
         RS1456ppe*RD1456ep + 5./288.*pow(m33,2)*RS145pe*RD1456ep + 5./
         288.*pow(m33,2)*RS146pe*RD1456ep + 5./288.*pow(m33,2)*RS156pe*
         RD1456ep - 1./108.*m11*pow(RD1456ep,2) - 1./72.*m11*m33*
         RD1456ep*RS14563pe - 5./144.*m11*m33*RS1456pee*RD1456ep - 5./
         144.*m11*m33*RS1456ppe*RD1456ep + 5./144.*m11*m33*RS145pe*
         RD1456ep + 5./144.*m11*m33*RS146pe*RD1456ep + 5./144.*m11*m33*
         RS156pe*RD1456ep + 1./432.*pow(m11,2)*RD1456ep*RS14563pe - 5./
         288.*pow(m11,2)*RS1456pee*RD1456ep - 5./288.*pow(m11,2)*
         RS1456ppe*RD1456ep + 5./288.*pow(m11,2)*RS145pe*RD1456ep + 5./
         288.*pow(m11,2)*RS146pe*RD1456ep + 5./288.*pow(m11,2)*RS156pe*
         RD1456ep );

      F13p6x23R +=  + hhd(2,m11,m13,m33,m13,mu2) * (  - 1./144.*pow(
         m33,2)*RD1456ep*RS34561pe - 1./288.*pow(m33,2)*RD1456ep*
         RS3456pee - 1./288.*pow(m33,2)*RD1456ep*RS3456ppe + 1./288.*
         pow(m33,2)*RD1456ep*RS345pe + 1./288.*pow(m33,2)*RD1456ep*
         RS346pe + 1./288.*pow(m33,2)*RD1456ep*RS356pe - 1./72.*m11*m33
         *RD1456ep*RS34561pe - 1./144.*m11*m33*RD1456ep*RS3456pee - 1./
         144.*m11*m33*RD1456ep*RS3456ppe + 1./144.*m11*m33*RD1456ep*
         RS345pe + 1./144.*m11*m33*RD1456ep*RS346pe + 1./144.*m11*m33*
         RD1456ep*RS356pe - 1./144.*pow(m11,2)*RD1456ep*RS34561pe - 1./
         288.*pow(m11,2)*RD1456ep*RS3456pee - 1./288.*pow(m11,2)*
         RD1456ep*RS3456ppe + 1./288.*pow(m11,2)*RD1456ep*RS345pe + 1./
         288.*pow(m11,2)*RD1456ep*RS346pe + 1./288.*pow(m11,2)*RD1456ep
         *RS356pe );

      F13p6x23R +=  + hhd(2,m11,m13,me2,m13,mu2) * (  - 1./324.*pow(
         me2,2)*RD1456ep*RSe45613p + 1./648.*pow(me2,2)*RD1456ep*
         RSe45633p + 1./648.*pow(me2,2)*RSe45611p*RD1456ep + 1./81.*m33
         *me2*RD1456ep*RSe45613p - 1./324.*m33*me2*RD1456ep*RSe45633p
          - 1./108.*m33*me2*RSe45611p*RD1456ep - 7./432.*pow(m33,2)*
         RD1456ep*RSe45613p + 13./2592.*pow(m33,2)*RD1456ep*RSe45633p
          + 5./288.*pow(m33,2)*RSe45611p*RD1456ep + 1./81.*m11*me2*
         RD1456ep*RSe45613p - 1./81.*m11*me2*RSe45611p*RD1456ep - 17./
         648.*m11*m33*RD1456ep*RSe45613p + 1./144.*m11*m33*RD1456ep*
         RSe45633p + 19./432.*m11*m33*RSe45611p*RD1456ep - 1./144.*pow(
         m11,2)*RD1456ep*RSe45613p + 1./288.*pow(m11,2)*RD1456ep*
         RSe45633p + 73./2592.*pow(m11,2)*RSe45611p*RD1456ep );

      F13p6x23R +=  + hhd(2,m11,m13,mp2,m13,mu2) * (  - 1./324.*pow(
         mp2,2)*RD1456ep*RSp45613e + 1./648.*pow(mp2,2)*RD1456ep*
         RSp45633e + 1./648.*pow(mp2,2)*RSp45611e*RD1456ep + 1./81.*m33
         *mp2*RD1456ep*RSp45613e - 1./324.*m33*mp2*RD1456ep*RSp45633e
          - 1./108.*m33*mp2*RSp45611e*RD1456ep - 7./432.*pow(m33,2)*
         RD1456ep*RSp45613e + 13./2592.*pow(m33,2)*RD1456ep*RSp45633e
          + 5./288.*pow(m33,2)*RSp45611e*RD1456ep + 1./81.*m11*mp2*
         RD1456ep*RSp45613e - 1./81.*m11*mp2*RSp45611e*RD1456ep - 17./
         648.*m11*m33*RD1456ep*RSp45613e + 1./144.*m11*m33*RD1456ep*
         RSp45633e + 19./432.*m11*m33*RSp45611e*RD1456ep - 1./144.*pow(
         m11,2)*RD1456ep*RSp45613e + 1./288.*pow(m11,2)*RD1456ep*
         RSp45633e + 73./2592.*pow(m11,2)*RSp45611e*RD1456ep );

      F13p6x23R +=  + hhd(2,m11,m14,m34,m13,mu2) * (  - 1./96.*m33*m44*
         RD1456ep - 1./96.*m11*m44*RD1456ep - 1./96.*m11*m33*RD1456ep
          - 1./96.*pow(m11,2)*RD1456ep );

      F13p6x23R +=  + hhd(2,m11,m15,m35,m13,mu2) * (  - 1./96.*m33*m55*
         RD1456ep - 1./96.*m11*m55*RD1456ep - 1./96.*m11*m33*RD1456ep
          - 1./96.*pow(m11,2)*RD1456ep );

      F13p6x23R +=  + hhd(2,m11,m16,m36,m13,mu2) * (  - 1./96.*m33*m66*
         RD1456ep - 1./96.*m11*m66*RD1456ep - 1./96.*m11*m33*RD1456ep
          - 1./96.*pow(m11,2)*RD1456ep );

      F13p6x23R +=  + hhd(2,m33,m11,m13,m13,mu2) * (  - 1./144.*pow(
         m33,2)*RD3456ep*RS14563pe - 1./288.*pow(m33,2)*RS1456pee*
         RD3456ep - 1./288.*pow(m33,2)*RS1456ppe*RD3456ep + 1./288.*
         pow(m33,2)*RS145pe*RD3456ep + 1./288.*pow(m33,2)*RS146pe*
         RD3456ep + 1./288.*pow(m33,2)*RS156pe*RD3456ep - 1./72.*m11*
         m33*RD3456ep*RS14563pe - 1./144.*m11*m33*RS1456pee*RD3456ep - 
         1./144.*m11*m33*RS1456ppe*RD3456ep + 1./144.*m11*m33*RS145pe*
         RD3456ep + 1./144.*m11*m33*RS146pe*RD3456ep + 1./144.*m11*m33*
         RS156pe*RD3456ep - 1./144.*pow(m11,2)*RD3456ep*RS14563pe - 1./
         288.*pow(m11,2)*RS1456pee*RD3456ep - 1./288.*pow(m11,2)*
         RS1456ppe*RD3456ep + 1./288.*pow(m11,2)*RS145pe*RD3456ep + 1./
         288.*pow(m11,2)*RS146pe*RD3456ep + 1./288.*pow(m11,2)*RS156pe*
         RD3456ep );

      F13p6x23R +=  + hhd(2,m33,m13,m33,m13,mu2) * (  - 1./108.*m33*
         pow(RD3456ep,2) + 1./432.*pow(m33,2)*RD3456ep*RS34561pe - 5./
         288.*pow(m33,2)*RS3456pee*RD3456ep - 5./288.*pow(m33,2)*
         RS3456ppe*RD3456ep + 5./288.*pow(m33,2)*RS345pe*RD3456ep + 5./
         288.*pow(m33,2)*RS346pe*RD3456ep + 5./288.*pow(m33,2)*RS356pe*
         RD3456ep - 1./108.*m11*pow(RD3456ep,2) - 1./72.*m11*m33*
         RD3456ep*RS34561pe - 5./144.*m11*m33*RS3456pee*RD3456ep - 5./
         144.*m11*m33*RS3456ppe*RD3456ep + 5./144.*m11*m33*RS345pe*
         RD3456ep + 5./144.*m11*m33*RS346pe*RD3456ep + 5./144.*m11*m33*
         RS356pe*RD3456ep - 7./432.*pow(m11,2)*RD3456ep*RS34561pe - 5./
         288.*pow(m11,2)*RS3456pee*RD3456ep - 5./288.*pow(m11,2)*
         RS3456ppe*RD3456ep + 5./288.*pow(m11,2)*RS345pe*RD3456ep + 5./
         288.*pow(m11,2)*RS346pe*RD3456ep + 5./288.*pow(m11,2)*RS356pe*
         RD3456ep );

      F13p6x23R +=  + hhd(2,m33,m13,me2,m13,mu2) * (  - 1./324.*pow(
         me2,2)*RD3456ep*RSe45613p + 1./648.*pow(me2,2)*RSe45633p*
         RD3456ep + 1./648.*pow(me2,2)*RSe45611p*RD3456ep + 1./81.*m33*
         me2*RD3456ep*RSe45613p - 1./81.*m33*me2*RSe45633p*RD3456ep - 1.
         /144.*pow(m33,2)*RD3456ep*RSe45613p + 73./2592.*pow(m33,2)*
         RSe45633p*RD3456ep + 1./288.*pow(m33,2)*RSe45611p*RD3456ep + 1.
         /81.*m11*me2*RD3456ep*RSe45613p - 1./108.*m11*me2*RSe45633p*
         RD3456ep - 1./324.*m11*me2*RSe45611p*RD3456ep - 17./648.*m11*
         m33*RD3456ep*RSe45613p + 19./432.*m11*m33*RSe45633p*RD3456ep
          + 1./144.*m11*m33*RSe45611p*RD3456ep - 7./432.*pow(m11,2)*
         RD3456ep*RSe45613p + 5./288.*pow(m11,2)*RSe45633p*RD3456ep + 
         13./2592.*pow(m11,2)*RSe45611p*RD3456ep );

      F13p6x23R +=  + hhd(2,m33,m13,mp2,m13,mu2) * (  - 1./324.*pow(
         mp2,2)*RD3456ep*RSp45613e + 1./648.*pow(mp2,2)*RSp45633e*
         RD3456ep + 1./648.*pow(mp2,2)*RSp45611e*RD3456ep + 1./81.*m33*
         mp2*RD3456ep*RSp45613e - 1./81.*m33*mp2*RSp45633e*RD3456ep - 1.
         /144.*pow(m33,2)*RD3456ep*RSp45613e + 73./2592.*pow(m33,2)*
         RSp45633e*RD3456ep + 1./288.*pow(m33,2)*RSp45611e*RD3456ep + 1.
         /81.*m11*mp2*RD3456ep*RSp45613e - 1./108.*m11*mp2*RSp45633e*
         RD3456ep - 1./324.*m11*mp2*RSp45611e*RD3456ep - 17./648.*m11*
         m33*RD3456ep*RSp45613e + 19./432.*m11*m33*RSp45633e*RD3456ep
          + 1./144.*m11*m33*RSp45611e*RD3456ep - 7./432.*pow(m11,2)*
         RD3456ep*RSp45613e + 5./288.*pow(m11,2)*RSp45633e*RD3456ep + 
         13./2592.*pow(m11,2)*RSp45611e*RD3456ep );

      F13p6x23R +=  + hhd(2,m33,m14,m34,m13,mu2) * (  - 1./96.*m33*m44*
         RD3456ep - 1./96.*pow(m33,2)*RD3456ep - 1./96.*m11*m44*
         RD3456ep - 1./96.*m11*m33*RD3456ep );

      F13p6x23R +=  + hhd(2,m33,m15,m35,m13,mu2) * (  - 1./96.*m33*m55*
         RD3456ep - 1./96.*pow(m33,2)*RD3456ep - 1./96.*m11*m55*
         RD3456ep - 1./96.*m11*m33*RD3456ep );

      F13p6x23R +=  + hhd(2,m33,m16,m36,m13,mu2) * (  - 1./96.*m33*m66*
         RD3456ep - 1./96.*pow(m33,2)*RD3456ep - 1./96.*m11*m66*
         RD3456ep - 1./96.*m11*m33*RD3456ep );

      F13p6x23R +=  + hhd(5,m11,m11,m13,m13,mu2) * ( 5./576.*pow(m33,2)
         *pow(RD1456ep,2) + 5./288.*m11*m33*pow(RD1456ep,2) + 5./576.*
         pow(m11,2)*pow(RD1456ep,2) );

      F13p6x23R +=  + hhd(5,m11,m33,m13,m13,mu2) * ( 1./288.*pow(m33,2)
         *RD1456ep*RD3456ep + 1./144.*m11*m33*RD1456ep*RD3456ep + 1./
         288.*pow(m11,2)*RD1456ep*RD3456ep );

      F13p6x23R +=  + hhd(5,m33,m33,m13,m13,mu2) * ( 5./576.*pow(m33,2)
         *pow(RD3456ep,2) + 5./288.*m11*m33*pow(RD3456ep,2) + 5./576.*
         pow(m11,2)*pow(RD3456ep,2) );

      F13p6x23R +=  + hh1d(1,m11,m14,m34,m13,mu2) * ( 1./12.*pow(m33,2)
         *RS14563pe - 1./12.*pow(m33,2)*RS156pe + 1./6.*m11*m33*
         RS14563pe - 1./6.*m11*m33*RS156pe + 1./12.*pow(m11,2)*
         RS14563pe - 1./12.*pow(m11,2)*RS156pe );

      F13p6x23R +=  + hh1d(1,m11,m15,m35,m13,mu2) * ( 1./12.*pow(m33,2)
         *RS14563pe - 1./12.*pow(m33,2)*RS146pe + 1./6.*m11*m33*
         RS14563pe - 1./6.*m11*m33*RS146pe + 1./12.*pow(m11,2)*
         RS14563pe - 1./12.*pow(m11,2)*RS146pe );

      F13p6x23R +=  + hh1d(1,m11,m16,m36,m13,mu2) * ( 1./12.*pow(m33,2)
         *RS14563pe - 1./12.*pow(m33,2)*RS145pe + 1./6.*m11*m33*
         RS14563pe - 1./6.*m11*m33*RS145pe + 1./12.*pow(m11,2)*
         RS14563pe - 1./12.*pow(m11,2)*RS145pe );

      F13p6x23R +=  + hh1d(1,m13,m11,m11,m13,mu2) * (  - 1./54.*m33*
         RD1456ep*RS14563pe - 1./108.*m33*RS1456pee*RD1456ep - 1./108.*
         m33*RS1456ppe*RD1456ep + 1./108.*m33*RS145pe*RD1456ep + 1./108.
         *m33*RS146pe*RD1456ep + 1./108.*m33*RS156pe*RD1456ep - 1./8.*
         pow(m33,2) - 5./108.*pow(m33,2)*pow(RS14563pe,2) - 1./27.*pow(
         m33,2)*RS1456pee*RS14563pe - 1./72.*pow(m33,2)*pow(
         RS1456pee,2) - 1./27.*pow(m33,2)*RS1456ppe*RS14563pe - 1./36.*
         pow(m33,2)*RS1456ppe*RS1456pee - 1./72.*pow(m33,2)*pow(
         RS1456ppe,2) + 1./27.*pow(m33,2)*RS145pe*RS14563pe + 1./36.*
         pow(m33,2)*RS145pe*RS1456pee + 1./36.*pow(m33,2)*RS145pe*
         RS1456ppe - 1./72.*pow(m33,2)*pow(RS145pe,2) + 1./27.*pow(
         m33,2)*RS146pe*RS14563pe + 1./36.*pow(m33,2)*RS146pe*RS1456pee
          + 1./36.*pow(m33,2)*RS146pe*RS1456ppe - 1./36.*pow(m33,2)*
         RS146pe*RS145pe - 1./72.*pow(m33,2)*pow(RS146pe,2) + 1./27.*
         pow(m33,2)*RS156pe*RS14563pe + 1./36.*pow(m33,2)*RS156pe*
         RS1456pee );

      F13p6x23R +=  + hh1d(1,m13,m11,m11,m13,mu2) * ( 1./36.*pow(m33,2)
         *RS156pe*RS1456ppe - 1./36.*pow(m33,2)*RS156pe*RS145pe - 1./36.
         *pow(m33,2)*RS156pe*RS146pe - 1./72.*pow(m33,2)*pow(RS156pe,2)
          - 1./54.*m11*RD1456ep*RS14563pe - 1./108.*m11*RS1456pee*
         RD1456ep - 1./108.*m11*RS1456ppe*RD1456ep + 1./108.*m11*
         RS145pe*RD1456ep + 1./108.*m11*RS146pe*RD1456ep + 1./108.*m11*
         RS156pe*RD1456ep - 1./4.*m11*m33 - 1./18.*m11*m33*pow(
         RS14563pe,2) - 1./18.*m11*m33*RS1456pee*RS14563pe - 1./36.*m11
         *m33*pow(RS1456pee,2) - 1./18.*m11*m33*RS1456ppe*RS14563pe - 1.
         /18.*m11*m33*RS1456ppe*RS1456pee - 1./36.*m11*m33*pow(
         RS1456ppe,2) + 1./18.*m11*m33*RS145pe*RS14563pe + 1./18.*m11*
         m33*RS145pe*RS1456pee + 1./18.*m11*m33*RS145pe*RS1456ppe - 1./
         36.*m11*m33*pow(RS145pe,2) + 1./18.*m11*m33*RS146pe*RS14563pe
          + 1./18.*m11*m33*RS146pe*RS1456pee + 1./18.*m11*m33*RS146pe*
         RS1456ppe - 1./18.*m11*m33*RS146pe*RS145pe - 1./36.*m11*m33*
         pow(RS146pe,2) );

      F13p6x23R +=  + hh1d(1,m13,m11,m11,m13,mu2) * ( 1./18.*m11*m33*
         RS156pe*RS14563pe + 1./18.*m11*m33*RS156pe*RS1456pee + 1./18.*
         m11*m33*RS156pe*RS1456ppe - 1./18.*m11*m33*RS156pe*RS145pe - 1.
         /18.*m11*m33*RS156pe*RS146pe - 1./36.*m11*m33*pow(RS156pe,2)
          - 1./8.*pow(m11,2) - 1./108.*pow(m11,2)*pow(RS14563pe,2) - 1./
         54.*pow(m11,2)*RS1456pee*RS14563pe - 1./72.*pow(m11,2)*pow(
         RS1456pee,2) - 1./54.*pow(m11,2)*RS1456ppe*RS14563pe - 1./36.*
         pow(m11,2)*RS1456ppe*RS1456pee - 1./72.*pow(m11,2)*pow(
         RS1456ppe,2) + 1./54.*pow(m11,2)*RS145pe*RS14563pe + 1./36.*
         pow(m11,2)*RS145pe*RS1456pee + 1./36.*pow(m11,2)*RS145pe*
         RS1456ppe - 1./72.*pow(m11,2)*pow(RS145pe,2) + 1./54.*pow(
         m11,2)*RS146pe*RS14563pe + 1./36.*pow(m11,2)*RS146pe*RS1456pee
          + 1./36.*pow(m11,2)*RS146pe*RS1456ppe - 1./36.*pow(m11,2)*
         RS146pe*RS145pe - 1./72.*pow(m11,2)*pow(RS146pe,2) + 1./54.*
         pow(m11,2)*RS156pe*RS14563pe + 1./36.*pow(m11,2)*RS156pe*
         RS1456pee );

      F13p6x23R +=  + hh1d(1,m13,m11,m11,m13,mu2) * ( 1./36.*pow(m11,2)
         *RS156pe*RS1456ppe - 1./36.*pow(m11,2)*RS156pe*RS145pe - 1./36.
         *pow(m11,2)*RS156pe*RS146pe - 1./72.*pow(m11,2)*pow(RS156pe,2)
          );

      F13p6x23R +=  + hh1d(1,m13,m11,m33,m13,mu2) * (  - 1./54.*m33*
         RD3456ep*RS14563pe - 1./54.*m33*RD1456ep*RS34561pe - 1./108.*
         m33*RD1456ep*RS3456pee - 1./108.*m33*RD1456ep*RS3456ppe + 1./
         108.*m33*RD1456ep*RS345pe + 1./108.*m33*RD1456ep*RS346pe + 1./
         108.*m33*RD1456ep*RS356pe - 1./108.*m33*RS1456pee*RD3456ep - 1.
         /108.*m33*RS1456ppe*RD3456ep + 1./108.*m33*RS145pe*RD3456ep + 
         1./108.*m33*RS146pe*RD3456ep + 1./108.*m33*RS156pe*RD3456ep - 
         1./18.*pow(m33,2)*RS14563pe*RS34561pe - 1./27.*pow(m33,2)*
         RS3456pee*RS14563pe - 1./27.*pow(m33,2)*RS3456ppe*RS14563pe + 
         1./27.*pow(m33,2)*RS345pe*RS14563pe + 1./27.*pow(m33,2)*
         RS346pe*RS14563pe + 1./27.*pow(m33,2)*RS356pe*RS14563pe - 1./
         54.*pow(m33,2)*RS1456pee*RS34561pe - 1./54.*pow(m33,2)*
         RS1456ppe*RS34561pe + 1./54.*pow(m33,2)*RS145pe*RS34561pe + 1./
         54.*pow(m33,2)*RS146pe*RS34561pe + 1./54.*pow(m33,2)*RS156pe*
         RS34561pe - 1./54.*m11*RD3456ep*RS14563pe - 1./54.*m11*
         RD1456ep*RS34561pe );

      F13p6x23R +=  + hh1d(1,m13,m11,m33,m13,mu2) * (  - 1./108.*m11*
         RD1456ep*RS3456pee - 1./108.*m11*RD1456ep*RS3456ppe + 1./108.*
         m11*RD1456ep*RS345pe + 1./108.*m11*RD1456ep*RS346pe + 1./108.*
         m11*RD1456ep*RS356pe - 1./108.*m11*RS1456pee*RD3456ep - 1./108.
         *m11*RS1456ppe*RD3456ep + 1./108.*m11*RS145pe*RD3456ep + 1./
         108.*m11*RS146pe*RD3456ep + 1./108.*m11*RS156pe*RD3456ep - 1./
         9.*m11*m33*RS14563pe*RS34561pe - 1./18.*m11*m33*RS3456pee*
         RS14563pe - 1./18.*m11*m33*RS3456ppe*RS14563pe + 1./18.*m11*
         m33*RS345pe*RS14563pe + 1./18.*m11*m33*RS346pe*RS14563pe + 1./
         18.*m11*m33*RS356pe*RS14563pe - 1./18.*m11*m33*RS1456pee*
         RS34561pe - 1./18.*m11*m33*RS1456ppe*RS34561pe + 1./18.*m11*
         m33*RS145pe*RS34561pe + 1./18.*m11*m33*RS146pe*RS34561pe + 1./
         18.*m11*m33*RS156pe*RS34561pe - 1./18.*pow(m11,2)*RS14563pe*
         RS34561pe - 1./54.*pow(m11,2)*RS3456pee*RS14563pe - 1./54.*
         pow(m11,2)*RS3456ppe*RS14563pe + 1./54.*pow(m11,2)*RS345pe*
         RS14563pe );

      F13p6x23R +=  + hh1d(1,m13,m11,m33,m13,mu2) * ( 1./54.*pow(m11,2)
         *RS346pe*RS14563pe + 1./54.*pow(m11,2)*RS356pe*RS14563pe - 1./
         27.*pow(m11,2)*RS1456pee*RS34561pe - 1./27.*pow(m11,2)*
         RS1456ppe*RS34561pe + 1./27.*pow(m11,2)*RS145pe*RS34561pe + 1./
         27.*pow(m11,2)*RS146pe*RS34561pe + 1./27.*pow(m11,2)*RS156pe*
         RS34561pe );

      F13p6x23R +=  + hh1d(1,m13,m11,me2,m13,mu2) * (  - 1./54.*m33*
         RD1456ep*RSe45613p + 1./108.*m33*RD1456ep*RSe45633p + 1./108.*
         m33*RSe45611p*RD1456ep + 1./27.*m33*me2*RSe45613p*RS14563pe - 
         1./54.*m33*me2*RSe45633p*RS14563pe + 1./54.*m33*me2*RS1456pee*
         RSe45613p - 1./108.*m33*me2*RS1456pee*RSe45633p + 1./54.*m33*
         me2*RS1456ppe*RSe45613p - 1./108.*m33*me2*RS1456ppe*RSe45633p
          - 1./54.*m33*me2*RSe45611p*RS14563pe - 1./108.*m33*me2*
         RSe45611p*RS1456pee - 1./108.*m33*me2*RSe45611p*RS1456ppe - 1./
         54.*m33*me2*RS145pe*RSe45613p + 1./108.*m33*me2*RS145pe*
         RSe45633p + 1./108.*m33*me2*RS145pe*RSe45611p - 1./54.*m33*me2
         *RS146pe*RSe45613p + 1./108.*m33*me2*RS146pe*RSe45633p + 1./
         108.*m33*me2*RS146pe*RSe45611p - 1./54.*m33*me2*RS156pe*
         RSe45613p + 1./108.*m33*me2*RS156pe*RSe45633p + 1./108.*m33*
         me2*RS156pe*RSe45611p - 5./54.*pow(m33,2)*RSe45613p*RS14563pe
          + 1./18.*pow(m33,2)*RSe45633p*RS14563pe - 1./27.*pow(m33,2)*
         RS1456pee*RSe45613p );

      F13p6x23R +=  + hh1d(1,m13,m11,me2,m13,mu2) * ( 1./108.*pow(
         m33,2)*RS1456pee*RSe45633p - 1./27.*pow(m33,2)*RS1456ppe*
         RSe45613p + 1./108.*pow(m33,2)*RS1456ppe*RSe45633p + 1./27.*
         pow(m33,2)*RSe45611p*RS14563pe + 1./36.*pow(m33,2)*RSe45611p*
         RS1456pee + 1./36.*pow(m33,2)*RSe45611p*RS1456ppe + 1./27.*
         pow(m33,2)*RS145pe*RSe45613p - 1./108.*pow(m33,2)*RS145pe*
         RSe45633p - 1./36.*pow(m33,2)*RS145pe*RSe45611p + 1./27.*pow(
         m33,2)*RS146pe*RSe45613p - 1./108.*pow(m33,2)*RS146pe*
         RSe45633p - 1./36.*pow(m33,2)*RS146pe*RSe45611p + 1./27.*pow(
         m33,2)*RS156pe*RSe45613p - 1./108.*pow(m33,2)*RS156pe*
         RSe45633p - 1./36.*pow(m33,2)*RS156pe*RSe45611p - 1./54.*m11*
         RD1456ep*RSe45613p + 1./108.*m11*RD1456ep*RSe45633p + 1./108.*
         m11*RSe45611p*RD1456ep + 1./27.*m11*me2*RSe45613p*RS14563pe - 
         1./54.*m11*me2*RSe45633p*RS14563pe + 1./54.*m11*me2*RS1456pee*
         RSe45613p - 1./108.*m11*me2*RS1456pee*RSe45633p + 1./54.*m11*
         me2*RS1456ppe*RSe45613p );

      F13p6x23R +=  + hh1d(1,m13,m11,me2,m13,mu2) * (  - 1./108.*m11*
         me2*RS1456ppe*RSe45633p - 1./54.*m11*me2*RSe45611p*RS14563pe
          - 1./108.*m11*me2*RSe45611p*RS1456pee - 1./108.*m11*me2*
         RSe45611p*RS1456ppe - 1./54.*m11*me2*RS145pe*RSe45613p + 1./
         108.*m11*me2*RS145pe*RSe45633p + 1./108.*m11*me2*RS145pe*
         RSe45611p - 1./54.*m11*me2*RS146pe*RSe45613p + 1./108.*m11*me2
         *RS146pe*RSe45633p + 1./108.*m11*me2*RS146pe*RSe45611p - 1./54.
         *m11*me2*RS156pe*RSe45613p + 1./108.*m11*me2*RS156pe*RSe45633p
          + 1./108.*m11*me2*RS156pe*RSe45611p - 4./27.*m11*m33*
         RSe45613p*RS14563pe + 2./27.*m11*m33*RSe45633p*RS14563pe - 2./
         27.*m11*m33*RS1456pee*RSe45613p + 1./108.*m11*m33*RS1456pee*
         RSe45633p - 2./27.*m11*m33*RS1456ppe*RSe45613p + 1./108.*m11*
         m33*RS1456ppe*RSe45633p + 2./27.*m11*m33*RSe45611p*RS14563pe
          + 7./108.*m11*m33*RSe45611p*RS1456pee + 7./108.*m11*m33*
         RSe45611p*RS1456ppe + 2./27.*m11*m33*RS145pe*RSe45613p - 1./
         108.*m11*m33*RS145pe*RSe45633p );

      F13p6x23R +=  + hh1d(1,m13,m11,me2,m13,mu2) * (  - 7./108.*m11*
         m33*RS145pe*RSe45611p + 2./27.*m11*m33*RS146pe*RSe45613p - 1./
         108.*m11*m33*RS146pe*RSe45633p - 7./108.*m11*m33*RS146pe*
         RSe45611p + 2./27.*m11*m33*RS156pe*RSe45613p - 1./108.*m11*m33
         *RS156pe*RSe45633p - 7./108.*m11*m33*RS156pe*RSe45611p - 1./18.
         *pow(m11,2)*RSe45613p*RS14563pe + 1./54.*pow(m11,2)*RSe45633p*
         RS14563pe - 1./27.*pow(m11,2)*RS1456pee*RSe45613p - 1./27.*
         pow(m11,2)*RS1456ppe*RSe45613p + 1./27.*pow(m11,2)*RSe45611p*
         RS14563pe + 1./27.*pow(m11,2)*RSe45611p*RS1456pee + 1./27.*
         pow(m11,2)*RSe45611p*RS1456ppe + 1./27.*pow(m11,2)*RS145pe*
         RSe45613p - 1./27.*pow(m11,2)*RS145pe*RSe45611p + 1./27.*pow(
         m11,2)*RS146pe*RSe45613p - 1./27.*pow(m11,2)*RS146pe*RSe45611p
          + 1./27.*pow(m11,2)*RS156pe*RSe45613p - 1./27.*pow(m11,2)*
         RS156pe*RSe45611p );

      F13p6x23R +=  + hh1d(1,m13,m11,mp2,m13,mu2) * (  - 1./54.*m33*
         RD1456ep*RSp45613e + 1./108.*m33*RD1456ep*RSp45633e + 1./108.*
         m33*RSp45611e*RD1456ep + 1./27.*m33*mp2*RSp45613e*RS14563pe - 
         1./54.*m33*mp2*RSp45633e*RS14563pe + 1./54.*m33*mp2*RS1456pee*
         RSp45613e - 1./108.*m33*mp2*RS1456pee*RSp45633e + 1./54.*m33*
         mp2*RS1456ppe*RSp45613e - 1./108.*m33*mp2*RS1456ppe*RSp45633e
          - 1./54.*m33*mp2*RSp45611e*RS14563pe - 1./108.*m33*mp2*
         RSp45611e*RS1456pee - 1./108.*m33*mp2*RSp45611e*RS1456ppe - 1./
         54.*m33*mp2*RS145pe*RSp45613e + 1./108.*m33*mp2*RS145pe*
         RSp45633e + 1./108.*m33*mp2*RS145pe*RSp45611e - 1./54.*m33*mp2
         *RS146pe*RSp45613e + 1./108.*m33*mp2*RS146pe*RSp45633e + 1./
         108.*m33*mp2*RS146pe*RSp45611e - 1./54.*m33*mp2*RS156pe*
         RSp45613e + 1./108.*m33*mp2*RS156pe*RSp45633e + 1./108.*m33*
         mp2*RS156pe*RSp45611e - 5./54.*pow(m33,2)*RSp45613e*RS14563pe
          + 1./18.*pow(m33,2)*RSp45633e*RS14563pe - 1./27.*pow(m33,2)*
         RS1456pee*RSp45613e );

      F13p6x23R +=  + hh1d(1,m13,m11,mp2,m13,mu2) * ( 1./108.*pow(
         m33,2)*RS1456pee*RSp45633e - 1./27.*pow(m33,2)*RS1456ppe*
         RSp45613e + 1./108.*pow(m33,2)*RS1456ppe*RSp45633e + 1./27.*
         pow(m33,2)*RSp45611e*RS14563pe + 1./36.*pow(m33,2)*RSp45611e*
         RS1456pee + 1./36.*pow(m33,2)*RSp45611e*RS1456ppe + 1./27.*
         pow(m33,2)*RS145pe*RSp45613e - 1./108.*pow(m33,2)*RS145pe*
         RSp45633e - 1./36.*pow(m33,2)*RS145pe*RSp45611e + 1./27.*pow(
         m33,2)*RS146pe*RSp45613e - 1./108.*pow(m33,2)*RS146pe*
         RSp45633e - 1./36.*pow(m33,2)*RS146pe*RSp45611e + 1./27.*pow(
         m33,2)*RS156pe*RSp45613e - 1./108.*pow(m33,2)*RS156pe*
         RSp45633e - 1./36.*pow(m33,2)*RS156pe*RSp45611e - 1./54.*m11*
         RD1456ep*RSp45613e + 1./108.*m11*RD1456ep*RSp45633e + 1./108.*
         m11*RSp45611e*RD1456ep + 1./27.*m11*mp2*RSp45613e*RS14563pe - 
         1./54.*m11*mp2*RSp45633e*RS14563pe + 1./54.*m11*mp2*RS1456pee*
         RSp45613e - 1./108.*m11*mp2*RS1456pee*RSp45633e + 1./54.*m11*
         mp2*RS1456ppe*RSp45613e );

      F13p6x23R +=  + hh1d(1,m13,m11,mp2,m13,mu2) * (  - 1./108.*m11*
         mp2*RS1456ppe*RSp45633e - 1./54.*m11*mp2*RSp45611e*RS14563pe
          - 1./108.*m11*mp2*RSp45611e*RS1456pee - 1./108.*m11*mp2*
         RSp45611e*RS1456ppe - 1./54.*m11*mp2*RS145pe*RSp45613e + 1./
         108.*m11*mp2*RS145pe*RSp45633e + 1./108.*m11*mp2*RS145pe*
         RSp45611e - 1./54.*m11*mp2*RS146pe*RSp45613e + 1./108.*m11*mp2
         *RS146pe*RSp45633e + 1./108.*m11*mp2*RS146pe*RSp45611e - 1./54.
         *m11*mp2*RS156pe*RSp45613e + 1./108.*m11*mp2*RS156pe*RSp45633e
          + 1./108.*m11*mp2*RS156pe*RSp45611e - 4./27.*m11*m33*
         RSp45613e*RS14563pe + 2./27.*m11*m33*RSp45633e*RS14563pe - 2./
         27.*m11*m33*RS1456pee*RSp45613e + 1./108.*m11*m33*RS1456pee*
         RSp45633e - 2./27.*m11*m33*RS1456ppe*RSp45613e + 1./108.*m11*
         m33*RS1456ppe*RSp45633e + 2./27.*m11*m33*RSp45611e*RS14563pe
          + 7./108.*m11*m33*RSp45611e*RS1456pee + 7./108.*m11*m33*
         RSp45611e*RS1456ppe + 2./27.*m11*m33*RS145pe*RSp45613e - 1./
         108.*m11*m33*RS145pe*RSp45633e );

      F13p6x23R +=  + hh1d(1,m13,m11,mp2,m13,mu2) * (  - 7./108.*m11*
         m33*RS145pe*RSp45611e + 2./27.*m11*m33*RS146pe*RSp45613e - 1./
         108.*m11*m33*RS146pe*RSp45633e - 7./108.*m11*m33*RS146pe*
         RSp45611e + 2./27.*m11*m33*RS156pe*RSp45613e - 1./108.*m11*m33
         *RS156pe*RSp45633e - 7./108.*m11*m33*RS156pe*RSp45611e - 1./18.
         *pow(m11,2)*RSp45613e*RS14563pe + 1./54.*pow(m11,2)*RSp45633e*
         RS14563pe - 1./27.*pow(m11,2)*RS1456pee*RSp45613e - 1./27.*
         pow(m11,2)*RS1456ppe*RSp45613e + 1./27.*pow(m11,2)*RSp45611e*
         RS14563pe + 1./27.*pow(m11,2)*RSp45611e*RS1456pee + 1./27.*
         pow(m11,2)*RSp45611e*RS1456ppe + 1./27.*pow(m11,2)*RS145pe*
         RSp45613e - 1./27.*pow(m11,2)*RS145pe*RSp45611e + 1./27.*pow(
         m11,2)*RS146pe*RSp45613e - 1./27.*pow(m11,2)*RS146pe*RSp45611e
          + 1./27.*pow(m11,2)*RS156pe*RSp45613e - 1./27.*pow(m11,2)*
         RS156pe*RSp45611e );

      F13p6x23R +=  + hh1d(1,m13,m33,m33,m13,mu2) * (  - 1./54.*m33*
         RD3456ep*RS34561pe - 1./108.*m33*RS3456pee*RD3456ep - 1./108.*
         m33*RS3456ppe*RD3456ep + 1./108.*m33*RS345pe*RD3456ep + 1./108.
         *m33*RS346pe*RD3456ep + 1./108.*m33*RS356pe*RD3456ep - 1./8.*
         pow(m33,2) - 1./108.*pow(m33,2)*pow(RS34561pe,2) - 1./54.*pow(
         m33,2)*RS3456pee*RS34561pe - 1./72.*pow(m33,2)*pow(
         RS3456pee,2) - 1./54.*pow(m33,2)*RS3456ppe*RS34561pe - 1./36.*
         pow(m33,2)*RS3456ppe*RS3456pee - 1./72.*pow(m33,2)*pow(
         RS3456ppe,2) + 1./54.*pow(m33,2)*RS345pe*RS34561pe + 1./36.*
         pow(m33,2)*RS345pe*RS3456pee + 1./36.*pow(m33,2)*RS345pe*
         RS3456ppe - 1./72.*pow(m33,2)*pow(RS345pe,2) + 1./54.*pow(
         m33,2)*RS346pe*RS34561pe + 1./36.*pow(m33,2)*RS346pe*RS3456pee
          + 1./36.*pow(m33,2)*RS346pe*RS3456ppe - 1./36.*pow(m33,2)*
         RS346pe*RS345pe - 1./72.*pow(m33,2)*pow(RS346pe,2) + 1./54.*
         pow(m33,2)*RS356pe*RS34561pe + 1./36.*pow(m33,2)*RS356pe*
         RS3456pee );

      F13p6x23R +=  + hh1d(1,m13,m33,m33,m13,mu2) * ( 1./36.*pow(m33,2)
         *RS356pe*RS3456ppe - 1./36.*pow(m33,2)*RS356pe*RS345pe - 1./36.
         *pow(m33,2)*RS356pe*RS346pe - 1./72.*pow(m33,2)*pow(RS356pe,2)
          - 1./54.*m11*RD3456ep*RS34561pe - 1./108.*m11*RS3456pee*
         RD3456ep - 1./108.*m11*RS3456ppe*RD3456ep + 1./108.*m11*
         RS345pe*RD3456ep + 1./108.*m11*RS346pe*RD3456ep + 1./108.*m11*
         RS356pe*RD3456ep - 1./4.*m11*m33 - 1./18.*m11*m33*pow(
         RS34561pe,2) - 1./18.*m11*m33*RS3456pee*RS34561pe - 1./36.*m11
         *m33*pow(RS3456pee,2) - 1./18.*m11*m33*RS3456ppe*RS34561pe - 1.
         /18.*m11*m33*RS3456ppe*RS3456pee - 1./36.*m11*m33*pow(
         RS3456ppe,2) + 1./18.*m11*m33*RS345pe*RS34561pe + 1./18.*m11*
         m33*RS345pe*RS3456pee + 1./18.*m11*m33*RS345pe*RS3456ppe - 1./
         36.*m11*m33*pow(RS345pe,2) + 1./18.*m11*m33*RS346pe*RS34561pe
          + 1./18.*m11*m33*RS346pe*RS3456pee + 1./18.*m11*m33*RS346pe*
         RS3456ppe - 1./18.*m11*m33*RS346pe*RS345pe - 1./36.*m11*m33*
         pow(RS346pe,2) );

      F13p6x23R +=  + hh1d(1,m13,m33,m33,m13,mu2) * ( 1./18.*m11*m33*
         RS356pe*RS34561pe + 1./18.*m11*m33*RS356pe*RS3456pee + 1./18.*
         m11*m33*RS356pe*RS3456ppe - 1./18.*m11*m33*RS356pe*RS345pe - 1.
         /18.*m11*m33*RS356pe*RS346pe - 1./36.*m11*m33*pow(RS356pe,2)
          - 1./8.*pow(m11,2) - 5./108.*pow(m11,2)*pow(RS34561pe,2) - 1./
         27.*pow(m11,2)*RS3456pee*RS34561pe - 1./72.*pow(m11,2)*pow(
         RS3456pee,2) - 1./27.*pow(m11,2)*RS3456ppe*RS34561pe - 1./36.*
         pow(m11,2)*RS3456ppe*RS3456pee - 1./72.*pow(m11,2)*pow(
         RS3456ppe,2) + 1./27.*pow(m11,2)*RS345pe*RS34561pe + 1./36.*
         pow(m11,2)*RS345pe*RS3456pee + 1./36.*pow(m11,2)*RS345pe*
         RS3456ppe - 1./72.*pow(m11,2)*pow(RS345pe,2) + 1./27.*pow(
         m11,2)*RS346pe*RS34561pe + 1./36.*pow(m11,2)*RS346pe*RS3456pee
          + 1./36.*pow(m11,2)*RS346pe*RS3456ppe - 1./36.*pow(m11,2)*
         RS346pe*RS345pe - 1./72.*pow(m11,2)*pow(RS346pe,2) + 1./27.*
         pow(m11,2)*RS356pe*RS34561pe + 1./36.*pow(m11,2)*RS356pe*
         RS3456pee );

      F13p6x23R +=  + hh1d(1,m13,m33,m33,m13,mu2) * ( 1./36.*pow(m11,2)
         *RS356pe*RS3456ppe - 1./36.*pow(m11,2)*RS356pe*RS345pe - 1./36.
         *pow(m11,2)*RS356pe*RS346pe - 1./72.*pow(m11,2)*pow(RS356pe,2)
          );

      F13p6x23R +=  + hh1d(1,m13,m33,me2,m13,mu2) * (  - 1./54.*m33*
         RD3456ep*RSe45613p + 1./108.*m33*RSe45633p*RD3456ep + 1./108.*
         m33*RSe45611p*RD3456ep + 1./27.*m33*me2*RSe45613p*RS34561pe + 
         1./54.*m33*me2*RS3456pee*RSe45613p + 1./54.*m33*me2*RS3456ppe*
         RSe45613p - 1./54.*m33*me2*RSe45633p*RS34561pe - 1./108.*m33*
         me2*RSe45633p*RS3456pee - 1./108.*m33*me2*RSe45633p*RS3456ppe
          - 1./54.*m33*me2*RS345pe*RSe45613p + 1./108.*m33*me2*RS345pe*
         RSe45633p - 1./54.*m33*me2*RS346pe*RSe45613p + 1./108.*m33*me2
         *RS346pe*RSe45633p - 1./54.*m33*me2*RS356pe*RSe45613p + 1./108.
         *m33*me2*RS356pe*RSe45633p - 1./54.*m33*me2*RSe45611p*
         RS34561pe - 1./108.*m33*me2*RSe45611p*RS3456pee - 1./108.*m33*
         me2*RSe45611p*RS3456ppe + 1./108.*m33*me2*RSe45611p*RS345pe + 
         1./108.*m33*me2*RSe45611p*RS346pe + 1./108.*m33*me2*RSe45611p*
         RS356pe - 1./18.*pow(m33,2)*RSe45613p*RS34561pe - 1./27.*pow(
         m33,2)*RS3456pee*RSe45613p - 1./27.*pow(m33,2)*RS3456ppe*
         RSe45613p );

      F13p6x23R +=  + hh1d(1,m13,m33,me2,m13,mu2) * ( 1./27.*pow(m33,2)
         *RSe45633p*RS34561pe + 1./27.*pow(m33,2)*RSe45633p*RS3456pee
          + 1./27.*pow(m33,2)*RSe45633p*RS3456ppe + 1./27.*pow(m33,2)*
         RS345pe*RSe45613p - 1./27.*pow(m33,2)*RS345pe*RSe45633p + 1./
         27.*pow(m33,2)*RS346pe*RSe45613p - 1./27.*pow(m33,2)*RS346pe*
         RSe45633p + 1./27.*pow(m33,2)*RS356pe*RSe45613p - 1./27.*pow(
         m33,2)*RS356pe*RSe45633p + 1./54.*pow(m33,2)*RSe45611p*
         RS34561pe - 1./54.*m11*RD3456ep*RSe45613p + 1./108.*m11*
         RSe45633p*RD3456ep + 1./108.*m11*RSe45611p*RD3456ep + 1./27.*
         m11*me2*RSe45613p*RS34561pe + 1./54.*m11*me2*RS3456pee*
         RSe45613p + 1./54.*m11*me2*RS3456ppe*RSe45613p - 1./54.*m11*
         me2*RSe45633p*RS34561pe - 1./108.*m11*me2*RSe45633p*RS3456pee
          - 1./108.*m11*me2*RSe45633p*RS3456ppe - 1./54.*m11*me2*
         RS345pe*RSe45613p + 1./108.*m11*me2*RS345pe*RSe45633p - 1./54.
         *m11*me2*RS346pe*RSe45613p + 1./108.*m11*me2*RS346pe*RSe45633p
          - 1./54.*m11*me2*RS356pe*RSe45613p );

      F13p6x23R +=  + hh1d(1,m13,m33,me2,m13,mu2) * ( 1./108.*m11*me2*
         RS356pe*RSe45633p - 1./54.*m11*me2*RSe45611p*RS34561pe - 1./
         108.*m11*me2*RSe45611p*RS3456pee - 1./108.*m11*me2*RSe45611p*
         RS3456ppe + 1./108.*m11*me2*RSe45611p*RS345pe + 1./108.*m11*
         me2*RSe45611p*RS346pe + 1./108.*m11*me2*RSe45611p*RS356pe - 4./
         27.*m11*m33*RSe45613p*RS34561pe - 2./27.*m11*m33*RS3456pee*
         RSe45613p - 2./27.*m11*m33*RS3456ppe*RSe45613p + 2./27.*m11*
         m33*RSe45633p*RS34561pe + 7./108.*m11*m33*RSe45633p*RS3456pee
          + 7./108.*m11*m33*RSe45633p*RS3456ppe + 2./27.*m11*m33*
         RS345pe*RSe45613p - 7./108.*m11*m33*RS345pe*RSe45633p + 2./27.
         *m11*m33*RS346pe*RSe45613p - 7./108.*m11*m33*RS346pe*RSe45633p
          + 2./27.*m11*m33*RS356pe*RSe45613p - 7./108.*m11*m33*RS356pe*
         RSe45633p + 2./27.*m11*m33*RSe45611p*RS34561pe + 1./108.*m11*
         m33*RSe45611p*RS3456pee + 1./108.*m11*m33*RSe45611p*RS3456ppe
          - 1./108.*m11*m33*RSe45611p*RS345pe - 1./108.*m11*m33*
         RSe45611p*RS346pe );

      F13p6x23R +=  + hh1d(1,m13,m33,me2,m13,mu2) * (  - 1./108.*m11*
         m33*RSe45611p*RS356pe - 5./54.*pow(m11,2)*RSe45613p*RS34561pe
          - 1./27.*pow(m11,2)*RS3456pee*RSe45613p - 1./27.*pow(m11,2)*
         RS3456ppe*RSe45613p + 1./27.*pow(m11,2)*RSe45633p*RS34561pe + 
         1./36.*pow(m11,2)*RSe45633p*RS3456pee + 1./36.*pow(m11,2)*
         RSe45633p*RS3456ppe + 1./27.*pow(m11,2)*RS345pe*RSe45613p - 1./
         36.*pow(m11,2)*RS345pe*RSe45633p + 1./27.*pow(m11,2)*RS346pe*
         RSe45613p - 1./36.*pow(m11,2)*RS346pe*RSe45633p + 1./27.*pow(
         m11,2)*RS356pe*RSe45613p - 1./36.*pow(m11,2)*RS356pe*RSe45633p
          + 1./18.*pow(m11,2)*RSe45611p*RS34561pe + 1./108.*pow(m11,2)*
         RSe45611p*RS3456pee + 1./108.*pow(m11,2)*RSe45611p*RS3456ppe
          - 1./108.*pow(m11,2)*RSe45611p*RS345pe - 1./108.*pow(m11,2)*
         RSe45611p*RS346pe - 1./108.*pow(m11,2)*RSe45611p*RS356pe );

      F13p6x23R +=  + hh1d(1,m13,m33,mp2,m13,mu2) * (  - 1./54.*m33*
         RD3456ep*RSp45613e + 1./108.*m33*RSp45633e*RD3456ep + 1./108.*
         m33*RSp45611e*RD3456ep + 1./27.*m33*mp2*RSp45613e*RS34561pe + 
         1./54.*m33*mp2*RS3456pee*RSp45613e + 1./54.*m33*mp2*RS3456ppe*
         RSp45613e - 1./54.*m33*mp2*RSp45633e*RS34561pe - 1./108.*m33*
         mp2*RSp45633e*RS3456pee - 1./108.*m33*mp2*RSp45633e*RS3456ppe
          - 1./54.*m33*mp2*RS345pe*RSp45613e + 1./108.*m33*mp2*RS345pe*
         RSp45633e - 1./54.*m33*mp2*RS346pe*RSp45613e + 1./108.*m33*mp2
         *RS346pe*RSp45633e - 1./54.*m33*mp2*RS356pe*RSp45613e + 1./108.
         *m33*mp2*RS356pe*RSp45633e - 1./54.*m33*mp2*RSp45611e*
         RS34561pe - 1./108.*m33*mp2*RSp45611e*RS3456pee - 1./108.*m33*
         mp2*RSp45611e*RS3456ppe + 1./108.*m33*mp2*RSp45611e*RS345pe + 
         1./108.*m33*mp2*RSp45611e*RS346pe + 1./108.*m33*mp2*RSp45611e*
         RS356pe - 1./18.*pow(m33,2)*RSp45613e*RS34561pe - 1./27.*pow(
         m33,2)*RS3456pee*RSp45613e - 1./27.*pow(m33,2)*RS3456ppe*
         RSp45613e );

      F13p6x23R +=  + hh1d(1,m13,m33,mp2,m13,mu2) * ( 1./27.*pow(m33,2)
         *RSp45633e*RS34561pe + 1./27.*pow(m33,2)*RSp45633e*RS3456pee
          + 1./27.*pow(m33,2)*RSp45633e*RS3456ppe + 1./27.*pow(m33,2)*
         RS345pe*RSp45613e - 1./27.*pow(m33,2)*RS345pe*RSp45633e + 1./
         27.*pow(m33,2)*RS346pe*RSp45613e - 1./27.*pow(m33,2)*RS346pe*
         RSp45633e + 1./27.*pow(m33,2)*RS356pe*RSp45613e - 1./27.*pow(
         m33,2)*RS356pe*RSp45633e + 1./54.*pow(m33,2)*RSp45611e*
         RS34561pe - 1./54.*m11*RD3456ep*RSp45613e + 1./108.*m11*
         RSp45633e*RD3456ep + 1./108.*m11*RSp45611e*RD3456ep + 1./27.*
         m11*mp2*RSp45613e*RS34561pe + 1./54.*m11*mp2*RS3456pee*
         RSp45613e + 1./54.*m11*mp2*RS3456ppe*RSp45613e - 1./54.*m11*
         mp2*RSp45633e*RS34561pe - 1./108.*m11*mp2*RSp45633e*RS3456pee
          - 1./108.*m11*mp2*RSp45633e*RS3456ppe - 1./54.*m11*mp2*
         RS345pe*RSp45613e + 1./108.*m11*mp2*RS345pe*RSp45633e - 1./54.
         *m11*mp2*RS346pe*RSp45613e + 1./108.*m11*mp2*RS346pe*RSp45633e
          - 1./54.*m11*mp2*RS356pe*RSp45613e );

      F13p6x23R +=  + hh1d(1,m13,m33,mp2,m13,mu2) * ( 1./108.*m11*mp2*
         RS356pe*RSp45633e - 1./54.*m11*mp2*RSp45611e*RS34561pe - 1./
         108.*m11*mp2*RSp45611e*RS3456pee - 1./108.*m11*mp2*RSp45611e*
         RS3456ppe + 1./108.*m11*mp2*RSp45611e*RS345pe + 1./108.*m11*
         mp2*RSp45611e*RS346pe + 1./108.*m11*mp2*RSp45611e*RS356pe - 4./
         27.*m11*m33*RSp45613e*RS34561pe - 2./27.*m11*m33*RS3456pee*
         RSp45613e - 2./27.*m11*m33*RS3456ppe*RSp45613e + 2./27.*m11*
         m33*RSp45633e*RS34561pe + 7./108.*m11*m33*RSp45633e*RS3456pee
          + 7./108.*m11*m33*RSp45633e*RS3456ppe + 2./27.*m11*m33*
         RS345pe*RSp45613e - 7./108.*m11*m33*RS345pe*RSp45633e + 2./27.
         *m11*m33*RS346pe*RSp45613e - 7./108.*m11*m33*RS346pe*RSp45633e
          + 2./27.*m11*m33*RS356pe*RSp45613e - 7./108.*m11*m33*RS356pe*
         RSp45633e + 2./27.*m11*m33*RSp45611e*RS34561pe + 1./108.*m11*
         m33*RSp45611e*RS3456pee + 1./108.*m11*m33*RSp45611e*RS3456ppe
          - 1./108.*m11*m33*RSp45611e*RS345pe - 1./108.*m11*m33*
         RSp45611e*RS346pe );

      F13p6x23R +=  + hh1d(1,m13,m33,mp2,m13,mu2) * (  - 1./108.*m11*
         m33*RSp45611e*RS356pe - 5./54.*pow(m11,2)*RSp45613e*RS34561pe
          - 1./27.*pow(m11,2)*RS3456pee*RSp45613e - 1./27.*pow(m11,2)*
         RS3456ppe*RSp45613e + 1./27.*pow(m11,2)*RSp45633e*RS34561pe + 
         1./36.*pow(m11,2)*RSp45633e*RS3456pee + 1./36.*pow(m11,2)*
         RSp45633e*RS3456ppe + 1./27.*pow(m11,2)*RS345pe*RSp45613e - 1./
         36.*pow(m11,2)*RS345pe*RSp45633e + 1./27.*pow(m11,2)*RS346pe*
         RSp45613e - 1./36.*pow(m11,2)*RS346pe*RSp45633e + 1./27.*pow(
         m11,2)*RS356pe*RSp45613e - 1./36.*pow(m11,2)*RS356pe*RSp45633e
          + 1./18.*pow(m11,2)*RSp45611e*RS34561pe + 1./108.*pow(m11,2)*
         RSp45611e*RS3456pee + 1./108.*pow(m11,2)*RSp45611e*RS3456ppe
          - 1./108.*pow(m11,2)*RSp45611e*RS345pe - 1./108.*pow(m11,2)*
         RSp45611e*RS346pe - 1./108.*pow(m11,2)*RSp45611e*RS356pe );

      F13p6x23R +=  + hh1d(1,m13,me2,me2,m13,mu2) * ( 1./27.*m33*me2*
         pow(RSe45613p,2) - 1./27.*m33*me2*RSe45633p*RSe45613p + 1./108.
         *m33*me2*pow(RSe45633p,2) - 1./27.*m33*me2*RSe45611p*RSe45613p
          + 1./54.*m33*me2*RSe45611p*RSe45633p + 1./108.*m33*me2*pow(
         RSe45611p,2) - 5./108.*pow(m33,2)*pow(RSe45613p,2) + 1./18.*
         pow(m33,2)*RSe45633p*RSe45613p - 5./216.*pow(m33,2)*pow(
         RSe45633p,2) + 1./27.*pow(m33,2)*RSe45611p*RSe45613p - 1./108.
         *pow(m33,2)*RSe45611p*RSe45633p - 1./72.*pow(m33,2)*pow(
         RSe45611p,2) + 1./27.*m11*me2*pow(RSe45613p,2) - 1./27.*m11*
         me2*RSe45633p*RSe45613p + 1./108.*m11*me2*pow(RSe45633p,2) - 1.
         /27.*m11*me2*RSe45611p*RSe45613p + 1./54.*m11*me2*RSe45611p*
         RSe45633p + 1./108.*m11*me2*pow(RSe45611p,2) - 5./54.*m11*m33*
         pow(RSe45613p,2) + 5./54.*m11*m33*RSe45633p*RSe45613p - 1./27.
         *m11*m33*pow(RSe45633p,2) + 5./54.*m11*m33*RSe45611p*RSe45613p
          - 1./54.*m11*m33*RSe45611p*RSe45633p - 1./27.*m11*m33*pow(
         RSe45611p,2) );

      F13p6x23R +=  + hh1d(1,m13,me2,me2,m13,mu2) * (  - 5./108.*pow(
         m11,2)*pow(RSe45613p,2) + 1./27.*pow(m11,2)*RSe45633p*
         RSe45613p - 1./72.*pow(m11,2)*pow(RSe45633p,2) + 1./18.*pow(
         m11,2)*RSe45611p*RSe45613p - 1./108.*pow(m11,2)*RSe45611p*
         RSe45633p - 5./216.*pow(m11,2)*pow(RSe45611p,2) );

      F13p6x23R +=  + hh1d(1,m13,me2,mp2,m13,mu2) * ( 1./27.*m33*mp2*
         RSp45613e*RSe45613p - 1./54.*m33*mp2*RSe45633p*RSp45613e - 1./
         54.*m33*mp2*RSp45633e*RSe45613p + 1./108.*m33*mp2*RSp45633e*
         RSe45633p - 1./54.*m33*mp2*RSe45611p*RSp45613e + 1./108.*m33*
         mp2*RSe45611p*RSp45633e - 1./54.*m33*mp2*RSp45611e*RSe45613p
          + 1./108.*m33*mp2*RSp45611e*RSe45633p + 1./108.*m33*mp2*
         RSp45611e*RSe45611p + 1./27.*m33*me2*RSp45613e*RSe45613p - 1./
         54.*m33*me2*RSe45633p*RSp45613e - 1./54.*m33*me2*RSp45633e*
         RSe45613p + 1./108.*m33*me2*RSp45633e*RSe45633p - 1./54.*m33*
         me2*RSe45611p*RSp45613e + 1./108.*m33*me2*RSe45611p*RSp45633e
          - 1./54.*m33*me2*RSp45611e*RSe45613p + 1./108.*m33*me2*
         RSp45611e*RSe45633p + 1./108.*m33*me2*RSp45611e*RSe45611p - 5./
         54.*pow(m33,2)*RSp45613e*RSe45613p + 1./18.*pow(m33,2)*
         RSe45633p*RSp45613e + 1./18.*pow(m33,2)*RSp45633e*RSe45613p - 
         5./108.*pow(m33,2)*RSp45633e*RSe45633p + 1./27.*pow(m33,2)*
         RSe45611p*RSp45613e );

      F13p6x23R +=  + hh1d(1,m13,me2,mp2,m13,mu2) * (  - 1./108.*pow(
         m33,2)*RSe45611p*RSp45633e + 1./27.*pow(m33,2)*RSp45611e*
         RSe45613p - 1./108.*pow(m33,2)*RSp45611e*RSe45633p - 1./36.*
         pow(m33,2)*RSp45611e*RSe45611p + 1./27.*m11*mp2*RSp45613e*
         RSe45613p - 1./54.*m11*mp2*RSe45633p*RSp45613e - 1./54.*m11*
         mp2*RSp45633e*RSe45613p + 1./108.*m11*mp2*RSp45633e*RSe45633p
          - 1./54.*m11*mp2*RSe45611p*RSp45613e + 1./108.*m11*mp2*
         RSe45611p*RSp45633e - 1./54.*m11*mp2*RSp45611e*RSe45613p + 1./
         108.*m11*mp2*RSp45611e*RSe45633p + 1./108.*m11*mp2*RSp45611e*
         RSe45611p + 1./27.*m11*me2*RSp45613e*RSe45613p - 1./54.*m11*
         me2*RSe45633p*RSp45613e - 1./54.*m11*me2*RSp45633e*RSe45613p
          + 1./108.*m11*me2*RSp45633e*RSe45633p - 1./54.*m11*me2*
         RSe45611p*RSp45613e + 1./108.*m11*me2*RSe45611p*RSp45633e - 1./
         54.*m11*me2*RSp45611e*RSe45613p + 1./108.*m11*me2*RSp45611e*
         RSe45633p + 1./108.*m11*me2*RSp45611e*RSe45611p - 5./27.*m11*
         m33*RSp45613e*RSe45613p );

      F13p6x23R +=  + hh1d(1,m13,me2,mp2,m13,mu2) * ( 5./54.*m11*m33*
         RSe45633p*RSp45613e + 5./54.*m11*m33*RSp45633e*RSe45613p - 2./
         27.*m11*m33*RSp45633e*RSe45633p + 5./54.*m11*m33*RSe45611p*
         RSp45613e - 1./54.*m11*m33*RSe45611p*RSp45633e + 5./54.*m11*
         m33*RSp45611e*RSe45613p - 1./54.*m11*m33*RSp45611e*RSe45633p
          - 2./27.*m11*m33*RSp45611e*RSe45611p - 5./54.*pow(m11,2)*
         RSp45613e*RSe45613p + 1./27.*pow(m11,2)*RSe45633p*RSp45613e + 
         1./27.*pow(m11,2)*RSp45633e*RSe45613p - 1./36.*pow(m11,2)*
         RSp45633e*RSe45633p + 1./18.*pow(m11,2)*RSe45611p*RSp45613e - 
         1./108.*pow(m11,2)*RSe45611p*RSp45633e + 1./18.*pow(m11,2)*
         RSp45611e*RSe45613p - 1./108.*pow(m11,2)*RSp45611e*RSe45633p
          - 5./108.*pow(m11,2)*RSp45611e*RSe45611p );

      F13p6x23R +=  + hh1d(1,m13,mp2,mp2,m13,mu2) * ( 1./27.*m33*mp2*
         pow(RSp45613e,2) - 1./27.*m33*mp2*RSp45633e*RSp45613e + 1./108.
         *m33*mp2*pow(RSp45633e,2) - 1./27.*m33*mp2*RSp45611e*RSp45613e
          + 1./54.*m33*mp2*RSp45611e*RSp45633e + 1./108.*m33*mp2*pow(
         RSp45611e,2) - 5./108.*pow(m33,2)*pow(RSp45613e,2) + 1./18.*
         pow(m33,2)*RSp45633e*RSp45613e - 5./216.*pow(m33,2)*pow(
         RSp45633e,2) + 1./27.*pow(m33,2)*RSp45611e*RSp45613e - 1./108.
         *pow(m33,2)*RSp45611e*RSp45633e - 1./72.*pow(m33,2)*pow(
         RSp45611e,2) + 1./27.*m11*mp2*pow(RSp45613e,2) - 1./27.*m11*
         mp2*RSp45633e*RSp45613e + 1./108.*m11*mp2*pow(RSp45633e,2) - 1.
         /27.*m11*mp2*RSp45611e*RSp45613e + 1./54.*m11*mp2*RSp45611e*
         RSp45633e + 1./108.*m11*mp2*pow(RSp45611e,2) - 5./54.*m11*m33*
         pow(RSp45613e,2) + 5./54.*m11*m33*RSp45633e*RSp45613e - 1./27.
         *m11*m33*pow(RSp45633e,2) + 5./54.*m11*m33*RSp45611e*RSp45613e
          - 1./54.*m11*m33*RSp45611e*RSp45633e - 1./27.*m11*m33*pow(
         RSp45611e,2) );

      F13p6x23R +=  + hh1d(1,m13,mp2,mp2,m13,mu2) * (  - 5./108.*pow(
         m11,2)*pow(RSp45613e,2) + 1./27.*pow(m11,2)*RSp45633e*
         RSp45613e - 1./72.*pow(m11,2)*pow(RSp45633e,2) + 1./18.*pow(
         m11,2)*RSp45611e*RSp45613e - 1./108.*pow(m11,2)*RSp45611e*
         RSp45633e - 5./216.*pow(m11,2)*pow(RSp45611e,2) );

      F13p6x23R +=  + hh1d(1,m14,m11,m34,m13,mu2) * (  - 1./36.*m33*
         RD1456ep - 1./18.*m33*m44*RS156pe + 1./36.*pow(m33,2)*
         RS14563pe - 1./12.*pow(m33,2)*RS156pe - 1./36.*m11*RD1456ep - 
         1./18.*m11*m44*RS156pe - 1./9.*m11*m33*RS156pe - 1./36.*pow(
         m11,2)*RS14563pe - 1./36.*pow(m11,2)*RS156pe );

      F13p6x23R +=  + hh1d(1,m14,m33,m34,m13,mu2) * (  - 1./12.*pow(
         m33,2)*RS34561pe + 1./12.*pow(m33,2)*RS356pe - 1./6.*m11*m33*
         RS34561pe + 1./6.*m11*m33*RS356pe - 1./12.*pow(m11,2)*
         RS34561pe + 1./12.*pow(m11,2)*RS356pe );

      F13p6x23R +=  + hh1d(1,m14,m34,me2,m13,mu2) * ( 1./36.*m33*me2*
         RSe45633p - 1./18.*m33*me2*RSe563p - 1./36.*m33*me2*RSe45611p
          + 1./18.*m33*me2*RSe561p + 1./18.*m33*m44*RSe563p - 1./18.*
         m33*m44*RSe561p + 1./36.*pow(m33,2)*RSe45613p - 1./36.*pow(
         m33,2)*RSe45633p + 1./12.*pow(m33,2)*RSe563p - 1./12.*pow(
         m33,2)*RSe561p + 1./36.*m11*me2*RSe45633p - 1./18.*m11*me2*
         RSe563p - 1./36.*m11*me2*RSe45611p + 1./18.*m11*me2*RSe561p + 
         1./18.*m11*m44*RSe563p - 1./18.*m11*m44*RSe561p - 1./36.*m11*
         m33*RSe45633p + 1./6.*m11*m33*RSe563p + 1./36.*m11*m33*
         RSe45611p - 1./6.*m11*m33*RSe561p - 1./36.*pow(m11,2)*
         RSe45613p + 1./12.*pow(m11,2)*RSe563p + 1./36.*pow(m11,2)*
         RSe45611p - 1./12.*pow(m11,2)*RSe561p );

      F13p6x23R +=  + hh1d(1,m14,m34,mp2,m13,mu2) * ( 1./36.*m33*mp2*
         RSp45633e - 1./18.*m33*mp2*RSp563e - 1./36.*m33*mp2*RSp45611e
          + 1./18.*m33*mp2*RSp561e + 1./18.*m33*m44*RSp563e - 1./18.*
         m33*m44*RSp561e + 1./36.*pow(m33,2)*RSp45613e - 1./36.*pow(
         m33,2)*RSp45633e + 1./12.*pow(m33,2)*RSp563e - 1./12.*pow(
         m33,2)*RSp561e + 1./36.*m11*mp2*RSp45633e - 1./18.*m11*mp2*
         RSp563e - 1./36.*m11*mp2*RSp45611e + 1./18.*m11*mp2*RSp561e + 
         1./18.*m11*m44*RSp563e - 1./18.*m11*m44*RSp561e - 1./36.*m11*
         m33*RSp45633e + 1./6.*m11*m33*RSp563e + 1./36.*m11*m33*
         RSp45611e - 1./6.*m11*m33*RSp561e - 1./36.*pow(m11,2)*
         RSp45613e + 1./12.*pow(m11,2)*RSp563e + 1./36.*pow(m11,2)*
         RSp45611e - 1./12.*pow(m11,2)*RSp561e );

      F13p6x23R +=  + hh1d(1,m15,m11,m35,m13,mu2) * (  - 1./36.*m33*
         RD1456ep - 1./18.*m33*m55*RS146pe + 1./36.*pow(m33,2)*
         RS14563pe - 1./12.*pow(m33,2)*RS146pe - 1./36.*m11*RD1456ep - 
         1./18.*m11*m55*RS146pe - 1./9.*m11*m33*RS146pe - 1./36.*pow(
         m11,2)*RS14563pe - 1./36.*pow(m11,2)*RS146pe );

      F13p6x23R +=  + hh1d(1,m15,m33,m35,m13,mu2) * (  - 1./12.*pow(
         m33,2)*RS34561pe + 1./12.*pow(m33,2)*RS346pe - 1./6.*m11*m33*
         RS34561pe + 1./6.*m11*m33*RS346pe - 1./12.*pow(m11,2)*
         RS34561pe + 1./12.*pow(m11,2)*RS346pe );

      F13p6x23R +=  + hh1d(1,m15,m35,me2,m13,mu2) * ( 1./36.*m33*me2*
         RSe45633p - 1./18.*m33*me2*RSe463p - 1./36.*m33*me2*RSe45611p
          + 1./18.*m33*me2*RSe461p + 1./18.*m33*m55*RSe463p - 1./18.*
         m33*m55*RSe461p + 1./36.*pow(m33,2)*RSe45613p - 1./36.*pow(
         m33,2)*RSe45633p + 1./12.*pow(m33,2)*RSe463p - 1./12.*pow(
         m33,2)*RSe461p + 1./36.*m11*me2*RSe45633p - 1./18.*m11*me2*
         RSe463p - 1./36.*m11*me2*RSe45611p + 1./18.*m11*me2*RSe461p + 
         1./18.*m11*m55*RSe463p - 1./18.*m11*m55*RSe461p - 1./36.*m11*
         m33*RSe45633p + 1./6.*m11*m33*RSe463p + 1./36.*m11*m33*
         RSe45611p - 1./6.*m11*m33*RSe461p - 1./36.*pow(m11,2)*
         RSe45613p + 1./12.*pow(m11,2)*RSe463p + 1./36.*pow(m11,2)*
         RSe45611p - 1./12.*pow(m11,2)*RSe461p );

      F13p6x23R +=  + hh1d(1,m15,m35,mp2,m13,mu2) * ( 1./36.*m33*mp2*
         RSp45633e - 1./18.*m33*mp2*RSp463e - 1./36.*m33*mp2*RSp45611e
          + 1./18.*m33*mp2*RSp461e + 1./18.*m33*m55*RSp463e - 1./18.*
         m33*m55*RSp461e + 1./36.*pow(m33,2)*RSp45613e - 1./36.*pow(
         m33,2)*RSp45633e + 1./12.*pow(m33,2)*RSp463e - 1./12.*pow(
         m33,2)*RSp461e + 1./36.*m11*mp2*RSp45633e - 1./18.*m11*mp2*
         RSp463e - 1./36.*m11*mp2*RSp45611e + 1./18.*m11*mp2*RSp461e + 
         1./18.*m11*m55*RSp463e - 1./18.*m11*m55*RSp461e - 1./36.*m11*
         m33*RSp45633e + 1./6.*m11*m33*RSp463e + 1./36.*m11*m33*
         RSp45611e - 1./6.*m11*m33*RSp461e - 1./36.*pow(m11,2)*
         RSp45613e + 1./12.*pow(m11,2)*RSp463e + 1./36.*pow(m11,2)*
         RSp45611e - 1./12.*pow(m11,2)*RSp461e );

      F13p6x23R +=  + hh1d(1,m16,m11,m36,m13,mu2) * (  - 1./36.*m33*
         RD1456ep - 1./18.*m33*m66*RS145pe + 1./36.*pow(m33,2)*
         RS14563pe - 1./12.*pow(m33,2)*RS145pe - 1./36.*m11*RD1456ep - 
         1./18.*m11*m66*RS145pe - 1./9.*m11*m33*RS145pe - 1./36.*pow(
         m11,2)*RS14563pe - 1./36.*pow(m11,2)*RS145pe );

      F13p6x23R +=  + hh1d(1,m16,m33,m36,m13,mu2) * (  - 1./12.*pow(
         m33,2)*RS34561pe + 1./12.*pow(m33,2)*RS345pe - 1./6.*m11*m33*
         RS34561pe + 1./6.*m11*m33*RS345pe - 1./12.*pow(m11,2)*
         RS34561pe + 1./12.*pow(m11,2)*RS345pe );

      F13p6x23R +=  + hh1d(1,m16,m36,me2,m13,mu2) * ( 1./36.*m33*me2*
         RSe45633p - 1./18.*m33*me2*RSe453p - 1./36.*m33*me2*RSe45611p
          + 1./18.*m33*me2*RSe451p + 1./18.*m33*m66*RSe453p - 1./18.*
         m33*m66*RSe451p + 1./36.*pow(m33,2)*RSe45613p - 1./36.*pow(
         m33,2)*RSe45633p + 1./12.*pow(m33,2)*RSe453p - 1./12.*pow(
         m33,2)*RSe451p + 1./36.*m11*me2*RSe45633p - 1./18.*m11*me2*
         RSe453p - 1./36.*m11*me2*RSe45611p + 1./18.*m11*me2*RSe451p + 
         1./18.*m11*m66*RSe453p - 1./18.*m11*m66*RSe451p - 1./36.*m11*
         m33*RSe45633p + 1./6.*m11*m33*RSe453p + 1./36.*m11*m33*
         RSe45611p - 1./6.*m11*m33*RSe451p - 1./36.*pow(m11,2)*
         RSe45613p + 1./12.*pow(m11,2)*RSe453p + 1./36.*pow(m11,2)*
         RSe45611p - 1./12.*pow(m11,2)*RSe451p );

      F13p6x23R +=  + hh1d(1,m16,m36,mp2,m13,mu2) * ( 1./36.*m33*mp2*
         RSp45633e - 1./18.*m33*mp2*RSp453e - 1./36.*m33*mp2*RSp45611e
          + 1./18.*m33*mp2*RSp451e + 1./18.*m33*m66*RSp453e - 1./18.*
         m33*m66*RSp451e + 1./36.*pow(m33,2)*RSp45613e - 1./36.*pow(
         m33,2)*RSp45633e + 1./12.*pow(m33,2)*RSp453e - 1./12.*pow(
         m33,2)*RSp451e + 1./36.*m11*mp2*RSp45633e - 1./18.*m11*mp2*
         RSp453e - 1./36.*m11*mp2*RSp45611e + 1./18.*m11*mp2*RSp451e + 
         1./18.*m11*m66*RSp453e - 1./18.*m11*m66*RSp451e - 1./36.*m11*
         m33*RSp45633e + 1./6.*m11*m33*RSp453e + 1./36.*m11*m33*
         RSp45611e - 1./6.*m11*m33*RSp451e - 1./36.*pow(m11,2)*
         RSp45613e + 1./12.*pow(m11,2)*RSp453e + 1./36.*pow(m11,2)*
         RSp45611e - 1./12.*pow(m11,2)*RSp451e );

      F13p6x23R +=  + hh1d(1,m34,m14,m33,m13,mu2) * (  - 1./36.*m33*
         RD3456ep - 1./18.*m33*m44*RS356pe - 1./9.*pow(m33,2)*RS34561pe
          + 1./18.*pow(m33,2)*RS356pe - 1./36.*m11*RD3456ep - 1./18.*
         m11*m44*RS356pe - 1./6.*m11*m33*RS34561pe + 1./18.*m11*m33*
         RS356pe - 1./18.*pow(m11,2)*RS34561pe );

      F13p6x23R +=  + hh1d(1,m35,m15,m33,m13,mu2) * (  - 1./36.*m33*
         RD3456ep - 1./18.*m33*m55*RS346pe - 1./9.*pow(m33,2)*RS34561pe
          + 1./18.*pow(m33,2)*RS346pe - 1./36.*m11*RD3456ep - 1./18.*
         m11*m55*RS346pe - 1./6.*m11*m33*RS34561pe + 1./18.*m11*m33*
         RS346pe - 1./18.*pow(m11,2)*RS34561pe );

      F13p6x23R +=  + hh1d(1,m36,m16,m33,m13,mu2) * (  - 1./36.*m33*
         RD3456ep - 1./18.*m33*m66*RS345pe - 1./9.*pow(m33,2)*RS34561pe
          + 1./18.*pow(m33,2)*RS345pe - 1./36.*m11*RD3456ep - 1./18.*
         m11*m66*RS345pe - 1./6.*m11*m33*RS34561pe + 1./18.*m11*m33*
         RS345pe - 1./18.*pow(m11,2)*RS34561pe );

      F13p6x23R +=  + hh1d(1,me2,m14,m34,m13,mu2) * ( 1./36.*m33*me2*
         RSe45613p + 1./36.*m33*me2*RSe45633p - 1./12.*m33*me2*RSe563p
          - 1./36.*m33*me2*RSe561p + 1./18.*m33*me2*RSe564p + 1./18.*
         m33*m44*RSe563p - 1./18.*m33*m44*RSe564p + 1./12.*pow(m33,2)*
         RSe45613p - 1./36.*pow(m33,2)*RSe45633p + 1./36.*pow(m33,2)*
         RSe563p - 1./12.*pow(m33,2)*RSe561p + 1./36.*m11*me2*RSe45613p
          + 1./36.*m11*me2*RSe45633p - 1./12.*m11*me2*RSe563p - 1./36.*
         m11*me2*RSe561p + 1./18.*m11*me2*RSe564p + 1./18.*m11*m44*
         RSe563p - 1./18.*m11*m44*RSe564p + 5./36.*m11*m33*RSe45613p - 
         1./36.*m11*m33*RSe45633p + 1./36.*m11*m33*RSe563p - 5./36.*m11
         *m33*RSe561p + 1./18.*pow(m11,2)*RSe45613p - 1./18.*pow(m11,2)
         *RSe561p );

      F13p6x23R +=  + hh1d(1,me2,m15,m35,m13,mu2) * ( 1./36.*m33*me2*
         RSe45613p + 1./36.*m33*me2*RSe45633p - 1./12.*m33*me2*RSe463p
          - 1./36.*m33*me2*RSe461p + 1./18.*m33*me2*RSe465p + 1./18.*
         m33*m55*RSe463p - 1./18.*m33*m55*RSe465p + 1./12.*pow(m33,2)*
         RSe45613p - 1./36.*pow(m33,2)*RSe45633p + 1./36.*pow(m33,2)*
         RSe463p - 1./12.*pow(m33,2)*RSe461p + 1./36.*m11*me2*RSe45613p
          + 1./36.*m11*me2*RSe45633p - 1./12.*m11*me2*RSe463p - 1./36.*
         m11*me2*RSe461p + 1./18.*m11*me2*RSe465p + 1./18.*m11*m55*
         RSe463p - 1./18.*m11*m55*RSe465p + 5./36.*m11*m33*RSe45613p - 
         1./36.*m11*m33*RSe45633p + 1./36.*m11*m33*RSe463p - 5./36.*m11
         *m33*RSe461p + 1./18.*pow(m11,2)*RSe45613p - 1./18.*pow(m11,2)
         *RSe461p );

      F13p6x23R +=  + hh1d(1,me2,m16,m36,m13,mu2) * ( 1./36.*m33*me2*
         RSe45613p + 1./36.*m33*me2*RSe45633p - 1./12.*m33*me2*RSe453p
          - 1./36.*m33*me2*RSe451p + 1./18.*m33*me2*RSe456p + 1./18.*
         m33*m66*RSe453p - 1./18.*m33*m66*RSe456p + 1./12.*pow(m33,2)*
         RSe45613p - 1./36.*pow(m33,2)*RSe45633p + 1./36.*pow(m33,2)*
         RSe453p - 1./12.*pow(m33,2)*RSe451p + 1./36.*m11*me2*RSe45613p
          + 1./36.*m11*me2*RSe45633p - 1./12.*m11*me2*RSe453p - 1./36.*
         m11*me2*RSe451p + 1./18.*m11*me2*RSe456p + 1./18.*m11*m66*
         RSe453p - 1./18.*m11*m66*RSe456p + 5./36.*m11*m33*RSe45613p - 
         1./36.*m11*m33*RSe45633p + 1./36.*m11*m33*RSe453p - 5./36.*m11
         *m33*RSe451p + 1./18.*pow(m11,2)*RSe45613p - 1./18.*pow(m11,2)
         *RSe451p );

      F13p6x23R +=  + hh1d(1,mp2,m14,m34,m13,mu2) * ( 1./36.*m33*mp2*
         RSp45613e + 1./36.*m33*mp2*RSp45633e - 1./12.*m33*mp2*RSp563e
          - 1./36.*m33*mp2*RSp561e + 1./18.*m33*mp2*RSp564e + 1./18.*
         m33*m44*RSp563e - 1./18.*m33*m44*RSp564e + 1./12.*pow(m33,2)*
         RSp45613e - 1./36.*pow(m33,2)*RSp45633e + 1./36.*pow(m33,2)*
         RSp563e - 1./12.*pow(m33,2)*RSp561e + 1./36.*m11*mp2*RSp45613e
          + 1./36.*m11*mp2*RSp45633e - 1./12.*m11*mp2*RSp563e - 1./36.*
         m11*mp2*RSp561e + 1./18.*m11*mp2*RSp564e + 1./18.*m11*m44*
         RSp563e - 1./18.*m11*m44*RSp564e + 5./36.*m11*m33*RSp45613e - 
         1./36.*m11*m33*RSp45633e + 1./36.*m11*m33*RSp563e - 5./36.*m11
         *m33*RSp561e + 1./18.*pow(m11,2)*RSp45613e - 1./18.*pow(m11,2)
         *RSp561e );

      F13p6x23R +=  + hh1d(1,mp2,m15,m35,m13,mu2) * ( 1./36.*m33*mp2*
         RSp45613e + 1./36.*m33*mp2*RSp45633e - 1./12.*m33*mp2*RSp463e
          - 1./36.*m33*mp2*RSp461e + 1./18.*m33*mp2*RSp465e + 1./18.*
         m33*m55*RSp463e - 1./18.*m33*m55*RSp465e + 1./12.*pow(m33,2)*
         RSp45613e - 1./36.*pow(m33,2)*RSp45633e + 1./36.*pow(m33,2)*
         RSp463e - 1./12.*pow(m33,2)*RSp461e + 1./36.*m11*mp2*RSp45613e
          + 1./36.*m11*mp2*RSp45633e - 1./12.*m11*mp2*RSp463e - 1./36.*
         m11*mp2*RSp461e + 1./18.*m11*mp2*RSp465e + 1./18.*m11*m55*
         RSp463e - 1./18.*m11*m55*RSp465e + 5./36.*m11*m33*RSp45613e - 
         1./36.*m11*m33*RSp45633e + 1./36.*m11*m33*RSp463e - 5./36.*m11
         *m33*RSp461e + 1./18.*pow(m11,2)*RSp45613e - 1./18.*pow(m11,2)
         *RSp461e );

      F13p6x23R +=  + hh1d(1,mp2,m16,m36,m13,mu2) * ( 1./36.*m33*mp2*
         RSp45613e + 1./36.*m33*mp2*RSp45633e - 1./12.*m33*mp2*RSp453e
          - 1./36.*m33*mp2*RSp451e + 1./18.*m33*mp2*RSp456e + 1./18.*
         m33*m66*RSp453e - 1./18.*m33*m66*RSp456e + 1./12.*pow(m33,2)*
         RSp45613e - 1./36.*pow(m33,2)*RSp45633e + 1./36.*pow(m33,2)*
         RSp453e - 1./12.*pow(m33,2)*RSp451e + 1./36.*m11*mp2*RSp45613e
          + 1./36.*m11*mp2*RSp45633e - 1./12.*m11*mp2*RSp453e - 1./36.*
         m11*mp2*RSp451e + 1./18.*m11*mp2*RSp456e + 1./18.*m11*m66*
         RSp453e - 1./18.*m11*m66*RSp456e + 5./36.*m11*m33*RSp45613e - 
         1./36.*m11*m33*RSp45633e + 1./36.*m11*m33*RSp453e - 5./36.*m11
         *m33*RSp451e + 1./18.*pow(m11,2)*RSp45613e - 1./18.*pow(m11,2)
         *RSp451e );

      F13p6x23R +=  + hh1d(3,m13,m11,m11,m13,mu2) * ( 1./108.*m33*pow(
         RD1456ep,2) + 1./27.*pow(m33,2)*RD1456ep*RS14563pe + 1./36.*
         pow(m33,2)*RS1456pee*RD1456ep + 1./36.*pow(m33,2)*RS1456ppe*
         RD1456ep - 1./36.*pow(m33,2)*RS145pe*RD1456ep - 1./36.*pow(
         m33,2)*RS146pe*RD1456ep - 1./36.*pow(m33,2)*RS156pe*RD1456ep
          + 1./108.*m11*pow(RD1456ep,2) + 1./18.*m11*m33*RD1456ep*
         RS14563pe + 1./18.*m11*m33*RS1456pee*RD1456ep + 1./18.*m11*m33
         *RS1456ppe*RD1456ep - 1./18.*m11*m33*RS145pe*RD1456ep - 1./18.
         *m11*m33*RS146pe*RD1456ep - 1./18.*m11*m33*RS156pe*RD1456ep + 
         1./54.*pow(m11,2)*RD1456ep*RS14563pe + 1./36.*pow(m11,2)*
         RS1456pee*RD1456ep + 1./36.*pow(m11,2)*RS1456ppe*RD1456ep - 1./
         36.*pow(m11,2)*RS145pe*RD1456ep - 1./36.*pow(m11,2)*RS146pe*
         RD1456ep - 1./36.*pow(m11,2)*RS156pe*RD1456ep );

      F13p6x23R +=  + hh1d(3,m13,m11,m33,m13,mu2) * ( 1./108.*m33*
         RD1456ep*RD3456ep + 1./54.*pow(m33,2)*RD1456ep*RS34561pe + 1./
         108.*m11*RD1456ep*RD3456ep + 1./18.*m11*m33*RD1456ep*RS34561pe
          + 1./27.*pow(m11,2)*RD1456ep*RS34561pe );

      F13p6x23R +=  + hh1d(3,m13,m11,me2,m13,mu2) * (  - 1./54.*m33*me2
         *RD1456ep*RSe45613p + 1./108.*m33*me2*RD1456ep*RSe45633p + 1./
         108.*m33*me2*RSe45611p*RD1456ep + 1./27.*pow(m33,2)*RD1456ep*
         RSe45613p - 1./108.*pow(m33,2)*RD1456ep*RSe45633p - 1./36.*
         pow(m33,2)*RSe45611p*RD1456ep - 1./54.*m11*me2*RD1456ep*
         RSe45613p + 1./108.*m11*me2*RD1456ep*RSe45633p + 1./108.*m11*
         me2*RSe45611p*RD1456ep + 2./27.*m11*m33*RD1456ep*RSe45613p - 1.
         /108.*m11*m33*RD1456ep*RSe45633p - 7./108.*m11*m33*RSe45611p*
         RD1456ep + 1./27.*pow(m11,2)*RD1456ep*RSe45613p - 1./27.*pow(
         m11,2)*RSe45611p*RD1456ep );

      F13p6x23R +=  + hh1d(3,m13,m11,mp2,m13,mu2) * (  - 1./54.*m33*mp2
         *RD1456ep*RSp45613e + 1./108.*m33*mp2*RD1456ep*RSp45633e + 1./
         108.*m33*mp2*RSp45611e*RD1456ep + 1./27.*pow(m33,2)*RD1456ep*
         RSp45613e - 1./108.*pow(m33,2)*RD1456ep*RSp45633e - 1./36.*
         pow(m33,2)*RSp45611e*RD1456ep - 1./54.*m11*mp2*RD1456ep*
         RSp45613e + 1./108.*m11*mp2*RD1456ep*RSp45633e + 1./108.*m11*
         mp2*RSp45611e*RD1456ep + 2./27.*m11*m33*RD1456ep*RSp45613e - 1.
         /108.*m11*m33*RD1456ep*RSp45633e - 7./108.*m11*m33*RSp45611e*
         RD1456ep + 1./27.*pow(m11,2)*RD1456ep*RSp45613e - 1./27.*pow(
         m11,2)*RSp45611e*RD1456ep );

      F13p6x23R +=  + hh1d(3,m13,m33,m11,m13,mu2) * ( 1./108.*m33*
         RD1456ep*RD3456ep + 1./27.*pow(m33,2)*RD3456ep*RS14563pe + 1./
         108.*m11*RD1456ep*RD3456ep + 1./18.*m11*m33*RD3456ep*RS14563pe
          + 1./54.*pow(m11,2)*RD3456ep*RS14563pe );

      F13p6x23R +=  + hh1d(3,m13,m33,m33,m13,mu2) * ( 1./108.*m33*pow(
         RD3456ep,2) + 1./54.*pow(m33,2)*RD3456ep*RS34561pe + 1./36.*
         pow(m33,2)*RS3456pee*RD3456ep + 1./36.*pow(m33,2)*RS3456ppe*
         RD3456ep - 1./36.*pow(m33,2)*RS345pe*RD3456ep - 1./36.*pow(
         m33,2)*RS346pe*RD3456ep - 1./36.*pow(m33,2)*RS356pe*RD3456ep
          + 1./108.*m11*pow(RD3456ep,2) + 1./18.*m11*m33*RD3456ep*
         RS34561pe + 1./18.*m11*m33*RS3456pee*RD3456ep + 1./18.*m11*m33
         *RS3456ppe*RD3456ep - 1./18.*m11*m33*RS345pe*RD3456ep - 1./18.
         *m11*m33*RS346pe*RD3456ep - 1./18.*m11*m33*RS356pe*RD3456ep + 
         1./27.*pow(m11,2)*RD3456ep*RS34561pe + 1./36.*pow(m11,2)*
         RS3456pee*RD3456ep + 1./36.*pow(m11,2)*RS3456ppe*RD3456ep - 1./
         36.*pow(m11,2)*RS345pe*RD3456ep - 1./36.*pow(m11,2)*RS346pe*
         RD3456ep - 1./36.*pow(m11,2)*RS356pe*RD3456ep );

      F13p6x23R +=  + hh1d(3,m13,m33,me2,m13,mu2) * (  - 1./54.*m33*me2
         *RD3456ep*RSe45613p + 1./108.*m33*me2*RSe45633p*RD3456ep + 1./
         108.*m33*me2*RSe45611p*RD3456ep + 1./27.*pow(m33,2)*RD3456ep*
         RSe45613p - 1./27.*pow(m33,2)*RSe45633p*RD3456ep - 1./54.*m11*
         me2*RD3456ep*RSe45613p + 1./108.*m11*me2*RSe45633p*RD3456ep + 
         1./108.*m11*me2*RSe45611p*RD3456ep + 2./27.*m11*m33*RD3456ep*
         RSe45613p - 7./108.*m11*m33*RSe45633p*RD3456ep - 1./108.*m11*
         m33*RSe45611p*RD3456ep + 1./27.*pow(m11,2)*RD3456ep*RSe45613p
          - 1./36.*pow(m11,2)*RSe45633p*RD3456ep - 1./108.*pow(m11,2)*
         RSe45611p*RD3456ep );

      F13p6x23R +=  + hh1d(3,m13,m33,mp2,m13,mu2) * (  - 1./54.*m33*mp2
         *RD3456ep*RSp45613e + 1./108.*m33*mp2*RSp45633e*RD3456ep + 1./
         108.*m33*mp2*RSp45611e*RD3456ep + 1./27.*pow(m33,2)*RD3456ep*
         RSp45613e - 1./27.*pow(m33,2)*RSp45633e*RD3456ep - 1./54.*m11*
         mp2*RD3456ep*RSp45613e + 1./108.*m11*mp2*RSp45633e*RD3456ep + 
         1./108.*m11*mp2*RSp45611e*RD3456ep + 2./27.*m11*m33*RD3456ep*
         RSp45613e - 7./108.*m11*m33*RSp45633e*RD3456ep - 1./108.*m11*
         m33*RSp45611e*RD3456ep + 1./27.*pow(m11,2)*RD3456ep*RSp45613e
          - 1./36.*pow(m11,2)*RSp45633e*RD3456ep - 1./108.*pow(m11,2)*
         RSp45611e*RD3456ep );

      F13p6x23R +=  + hh1d(7,m13,m11,m11,m13,mu2) * (  - 1./72.*pow(
         m33,2)*pow(RD1456ep,2) - 1./36.*m11*m33*pow(RD1456ep,2) - 1./
         72.*pow(m11,2)*pow(RD1456ep,2) );

      F13p6x23R +=  + hh1d(7,m13,m33,m33,m13,mu2) * (  - 1./72.*pow(
         m33,2)*pow(RD3456ep,2) - 1./36.*m11*m33*pow(RD3456ep,2) - 1./
         72.*pow(m11,2)*pow(RD3456ep,2) );

      F13p6x23R +=  + hh21d(1,m11,m11,m13,m13,mu2) * (  - 3./32.*pow(
         m33,2) - 3./16.*m11*m33 - 3./32.*pow(m11,2) );

      F13p6x23R +=  + hh21d(1,m11,m14,m34,m13,mu2) * (  - 1./32.*pow(
         m33,2)*RS14563pe + 1./32.*pow(m33,2)*RS156pe - 1./16.*m11*m33*
         RS14563pe + 1./16.*m11*m33*RS156pe - 1./32.*pow(m11,2)*
         RS14563pe + 1./32.*pow(m11,2)*RS156pe );

      F13p6x23R +=  + hh21d(1,m11,m15,m35,m13,mu2) * (  - 1./32.*pow(
         m33,2)*RS14563pe + 1./32.*pow(m33,2)*RS146pe - 1./16.*m11*m33*
         RS14563pe + 1./16.*m11*m33*RS146pe - 1./32.*pow(m11,2)*
         RS14563pe + 1./32.*pow(m11,2)*RS146pe );

      F13p6x23R +=  + hh21d(1,m11,m16,m36,m13,mu2) * (  - 1./32.*pow(
         m33,2)*RS14563pe + 1./32.*pow(m33,2)*RS145pe - 1./16.*m11*m33*
         RS14563pe + 1./16.*m11*m33*RS145pe - 1./32.*pow(m11,2)*
         RS14563pe + 1./32.*pow(m11,2)*RS145pe );

      F13p6x23R +=  + hh21d(1,m13,m11,m11,m13,mu2) * ( 3./64.*pow(
         m33,2) + 1./48.*pow(m33,2)*pow(RS14563pe,2) + 1./48.*pow(
         m33,2)*RS1456pee*RS14563pe + 1./192.*pow(m33,2)*pow(
         RS1456pee,2) + 1./48.*pow(m33,2)*RS1456ppe*RS14563pe + 1./96.*
         pow(m33,2)*RS1456ppe*RS1456pee + 1./192.*pow(m33,2)*pow(
         RS1456ppe,2) - 1./48.*pow(m33,2)*RS145pe*RS14563pe - 1./96.*
         pow(m33,2)*RS145pe*RS1456pee - 1./96.*pow(m33,2)*RS145pe*
         RS1456ppe + 1./192.*pow(m33,2)*pow(RS145pe,2) - 1./48.*pow(
         m33,2)*RS146pe*RS14563pe - 1./96.*pow(m33,2)*RS146pe*RS1456pee
          - 1./96.*pow(m33,2)*RS146pe*RS1456ppe + 1./96.*pow(m33,2)*
         RS146pe*RS145pe + 1./192.*pow(m33,2)*pow(RS146pe,2) - 1./48.*
         pow(m33,2)*RS156pe*RS14563pe - 1./96.*pow(m33,2)*RS156pe*
         RS1456pee - 1./96.*pow(m33,2)*RS156pe*RS1456ppe + 1./96.*pow(
         m33,2)*RS156pe*RS145pe + 1./96.*pow(m33,2)*RS156pe*RS146pe + 1.
         /192.*pow(m33,2)*pow(RS156pe,2) + 3./32.*m11*m33 + 1./24.*m11*
         m33*pow(RS14563pe,2) );

      F13p6x23R +=  + hh21d(1,m13,m11,m11,m13,mu2) * ( 1./24.*m11*m33*
         RS1456pee*RS14563pe + 1./96.*m11*m33*pow(RS1456pee,2) + 1./24.
         *m11*m33*RS1456ppe*RS14563pe + 1./48.*m11*m33*RS1456ppe*
         RS1456pee + 1./96.*m11*m33*pow(RS1456ppe,2) - 1./24.*m11*m33*
         RS145pe*RS14563pe - 1./48.*m11*m33*RS145pe*RS1456pee - 1./48.*
         m11*m33*RS145pe*RS1456ppe + 1./96.*m11*m33*pow(RS145pe,2) - 1./
         24.*m11*m33*RS146pe*RS14563pe - 1./48.*m11*m33*RS146pe*
         RS1456pee - 1./48.*m11*m33*RS146pe*RS1456ppe + 1./48.*m11*m33*
         RS146pe*RS145pe + 1./96.*m11*m33*pow(RS146pe,2) - 1./24.*m11*
         m33*RS156pe*RS14563pe - 1./48.*m11*m33*RS156pe*RS1456pee - 1./
         48.*m11*m33*RS156pe*RS1456ppe + 1./48.*m11*m33*RS156pe*RS145pe
          + 1./48.*m11*m33*RS156pe*RS146pe + 1./96.*m11*m33*pow(
         RS156pe,2) + 3./64.*pow(m11,2) + 1./48.*pow(m11,2)*pow(
         RS14563pe,2) + 1./48.*pow(m11,2)*RS1456pee*RS14563pe + 1./192.
         *pow(m11,2)*pow(RS1456pee,2) + 1./48.*pow(m11,2)*RS1456ppe*
         RS14563pe );

      F13p6x23R +=  + hh21d(1,m13,m11,m11,m13,mu2) * ( 1./96.*pow(
         m11,2)*RS1456ppe*RS1456pee + 1./192.*pow(m11,2)*pow(
         RS1456ppe,2) - 1./48.*pow(m11,2)*RS145pe*RS14563pe - 1./96.*
         pow(m11,2)*RS145pe*RS1456pee - 1./96.*pow(m11,2)*RS145pe*
         RS1456ppe + 1./192.*pow(m11,2)*pow(RS145pe,2) - 1./48.*pow(
         m11,2)*RS146pe*RS14563pe - 1./96.*pow(m11,2)*RS146pe*RS1456pee
          - 1./96.*pow(m11,2)*RS146pe*RS1456ppe + 1./96.*pow(m11,2)*
         RS146pe*RS145pe + 1./192.*pow(m11,2)*pow(RS146pe,2) - 1./48.*
         pow(m11,2)*RS156pe*RS14563pe - 1./96.*pow(m11,2)*RS156pe*
         RS1456pee - 1./96.*pow(m11,2)*RS156pe*RS1456ppe + 1./96.*pow(
         m11,2)*RS156pe*RS145pe + 1./96.*pow(m11,2)*RS156pe*RS146pe + 1.
         /192.*pow(m11,2)*pow(RS156pe,2) );

      F13p6x23R +=  + hh21d(1,m13,m11,m33,m13,mu2) * ( 1./24.*pow(
         m33,2)*RS14563pe*RS34561pe + 1./48.*pow(m33,2)*RS3456pee*
         RS14563pe + 1./48.*pow(m33,2)*RS3456ppe*RS14563pe - 1./48.*
         pow(m33,2)*RS345pe*RS14563pe - 1./48.*pow(m33,2)*RS346pe*
         RS14563pe - 1./48.*pow(m33,2)*RS356pe*RS14563pe + 1./48.*pow(
         m33,2)*RS1456pee*RS34561pe + 1./96.*pow(m33,2)*RS1456pee*
         RS3456pee + 1./96.*pow(m33,2)*RS1456pee*RS3456ppe - 1./96.*
         pow(m33,2)*RS1456pee*RS345pe - 1./96.*pow(m33,2)*RS1456pee*
         RS346pe - 1./96.*pow(m33,2)*RS1456pee*RS356pe + 1./48.*pow(
         m33,2)*RS1456ppe*RS34561pe + 1./96.*pow(m33,2)*RS1456ppe*
         RS3456pee + 1./96.*pow(m33,2)*RS1456ppe*RS3456ppe - 1./96.*
         pow(m33,2)*RS1456ppe*RS345pe - 1./96.*pow(m33,2)*RS1456ppe*
         RS346pe - 1./96.*pow(m33,2)*RS1456ppe*RS356pe - 1./48.*pow(
         m33,2)*RS145pe*RS34561pe - 1./96.*pow(m33,2)*RS145pe*RS3456pee
          - 1./96.*pow(m33,2)*RS145pe*RS3456ppe + 1./96.*pow(m33,2)*
         RS145pe*RS345pe );

      F13p6x23R +=  + hh21d(1,m13,m11,m33,m13,mu2) * ( 1./96.*pow(
         m33,2)*RS145pe*RS346pe + 1./96.*pow(m33,2)*RS145pe*RS356pe - 1.
         /48.*pow(m33,2)*RS146pe*RS34561pe - 1./96.*pow(m33,2)*RS146pe*
         RS3456pee - 1./96.*pow(m33,2)*RS146pe*RS3456ppe + 1./96.*pow(
         m33,2)*RS146pe*RS345pe + 1./96.*pow(m33,2)*RS146pe*RS346pe + 1.
         /96.*pow(m33,2)*RS146pe*RS356pe - 1./48.*pow(m33,2)*RS156pe*
         RS34561pe - 1./96.*pow(m33,2)*RS156pe*RS3456pee - 1./96.*pow(
         m33,2)*RS156pe*RS3456ppe + 1./96.*pow(m33,2)*RS156pe*RS345pe
          + 1./96.*pow(m33,2)*RS156pe*RS346pe + 1./96.*pow(m33,2)*
         RS156pe*RS356pe + 1./12.*m11*m33*RS14563pe*RS34561pe + 1./24.*
         m11*m33*RS3456pee*RS14563pe + 1./24.*m11*m33*RS3456ppe*
         RS14563pe - 1./24.*m11*m33*RS345pe*RS14563pe - 1./24.*m11*m33*
         RS346pe*RS14563pe - 1./24.*m11*m33*RS356pe*RS14563pe + 1./24.*
         m11*m33*RS1456pee*RS34561pe + 1./48.*m11*m33*RS1456pee*
         RS3456pee + 1./48.*m11*m33*RS1456pee*RS3456ppe - 1./48.*m11*
         m33*RS1456pee*RS345pe );

      F13p6x23R +=  + hh21d(1,m13,m11,m33,m13,mu2) * (  - 1./48.*m11*
         m33*RS1456pee*RS346pe - 1./48.*m11*m33*RS1456pee*RS356pe + 1./
         24.*m11*m33*RS1456ppe*RS34561pe + 1./48.*m11*m33*RS1456ppe*
         RS3456pee + 1./48.*m11*m33*RS1456ppe*RS3456ppe - 1./48.*m11*
         m33*RS1456ppe*RS345pe - 1./48.*m11*m33*RS1456ppe*RS346pe - 1./
         48.*m11*m33*RS1456ppe*RS356pe - 1./24.*m11*m33*RS145pe*
         RS34561pe - 1./48.*m11*m33*RS145pe*RS3456pee - 1./48.*m11*m33*
         RS145pe*RS3456ppe + 1./48.*m11*m33*RS145pe*RS345pe + 1./48.*
         m11*m33*RS145pe*RS346pe + 1./48.*m11*m33*RS145pe*RS356pe - 1./
         24.*m11*m33*RS146pe*RS34561pe - 1./48.*m11*m33*RS146pe*
         RS3456pee - 1./48.*m11*m33*RS146pe*RS3456ppe + 1./48.*m11*m33*
         RS146pe*RS345pe + 1./48.*m11*m33*RS146pe*RS346pe + 1./48.*m11*
         m33*RS146pe*RS356pe - 1./24.*m11*m33*RS156pe*RS34561pe - 1./48.
         *m11*m33*RS156pe*RS3456pee - 1./48.*m11*m33*RS156pe*RS3456ppe
          + 1./48.*m11*m33*RS156pe*RS345pe + 1./48.*m11*m33*RS156pe*
         RS346pe );

      F13p6x23R +=  + hh21d(1,m13,m11,m33,m13,mu2) * ( 1./48.*m11*m33*
         RS156pe*RS356pe + 1./24.*pow(m11,2)*RS14563pe*RS34561pe + 1./
         48.*pow(m11,2)*RS3456pee*RS14563pe + 1./48.*pow(m11,2)*
         RS3456ppe*RS14563pe - 1./48.*pow(m11,2)*RS345pe*RS14563pe - 1./
         48.*pow(m11,2)*RS346pe*RS14563pe - 1./48.*pow(m11,2)*RS356pe*
         RS14563pe + 1./48.*pow(m11,2)*RS1456pee*RS34561pe + 1./96.*
         pow(m11,2)*RS1456pee*RS3456pee + 1./96.*pow(m11,2)*RS1456pee*
         RS3456ppe - 1./96.*pow(m11,2)*RS1456pee*RS345pe - 1./96.*pow(
         m11,2)*RS1456pee*RS346pe - 1./96.*pow(m11,2)*RS1456pee*RS356pe
          + 1./48.*pow(m11,2)*RS1456ppe*RS34561pe + 1./96.*pow(m11,2)*
         RS1456ppe*RS3456pee + 1./96.*pow(m11,2)*RS1456ppe*RS3456ppe - 
         1./96.*pow(m11,2)*RS1456ppe*RS345pe - 1./96.*pow(m11,2)*
         RS1456ppe*RS346pe - 1./96.*pow(m11,2)*RS1456ppe*RS356pe - 1./
         48.*pow(m11,2)*RS145pe*RS34561pe - 1./96.*pow(m11,2)*RS145pe*
         RS3456pee - 1./96.*pow(m11,2)*RS145pe*RS3456ppe + 1./96.*pow(
         m11,2)*RS145pe*RS345pe );

      F13p6x23R +=  + hh21d(1,m13,m11,m33,m13,mu2) * ( 1./96.*pow(
         m11,2)*RS145pe*RS346pe + 1./96.*pow(m11,2)*RS145pe*RS356pe - 1.
         /48.*pow(m11,2)*RS146pe*RS34561pe - 1./96.*pow(m11,2)*RS146pe*
         RS3456pee - 1./96.*pow(m11,2)*RS146pe*RS3456ppe + 1./96.*pow(
         m11,2)*RS146pe*RS345pe + 1./96.*pow(m11,2)*RS146pe*RS346pe + 1.
         /96.*pow(m11,2)*RS146pe*RS356pe - 1./48.*pow(m11,2)*RS156pe*
         RS34561pe - 1./96.*pow(m11,2)*RS156pe*RS3456pee - 1./96.*pow(
         m11,2)*RS156pe*RS3456ppe + 1./96.*pow(m11,2)*RS156pe*RS345pe
          + 1./96.*pow(m11,2)*RS156pe*RS346pe + 1./96.*pow(m11,2)*
         RS156pe*RS356pe );

      F13p6x23R +=  + hh21d(1,m13,m11,me2,m13,mu2) * ( 1./24.*pow(
         m33,2)*RSe45613p*RS14563pe - 1./48.*pow(m33,2)*RSe45633p*
         RS14563pe + 1./48.*pow(m33,2)*RS1456pee*RSe45613p - 1./96.*
         pow(m33,2)*RS1456pee*RSe45633p + 1./48.*pow(m33,2)*RS1456ppe*
         RSe45613p - 1./96.*pow(m33,2)*RS1456ppe*RSe45633p - 1./48.*
         pow(m33,2)*RSe45611p*RS14563pe - 1./96.*pow(m33,2)*RSe45611p*
         RS1456pee - 1./96.*pow(m33,2)*RSe45611p*RS1456ppe - 1./48.*
         pow(m33,2)*RS145pe*RSe45613p + 1./96.*pow(m33,2)*RS145pe*
         RSe45633p + 1./96.*pow(m33,2)*RS145pe*RSe45611p - 1./48.*pow(
         m33,2)*RS146pe*RSe45613p + 1./96.*pow(m33,2)*RS146pe*RSe45633p
          + 1./96.*pow(m33,2)*RS146pe*RSe45611p - 1./48.*pow(m33,2)*
         RS156pe*RSe45613p + 1./96.*pow(m33,2)*RS156pe*RSe45633p + 1./
         96.*pow(m33,2)*RS156pe*RSe45611p + 1./12.*m11*m33*RSe45613p*
         RS14563pe - 1./24.*m11*m33*RSe45633p*RS14563pe + 1./24.*m11*
         m33*RS1456pee*RSe45613p - 1./48.*m11*m33*RS1456pee*RSe45633p
          + 1./24.*m11*m33*RS1456ppe*RSe45613p );

      F13p6x23R +=  + hh21d(1,m13,m11,me2,m13,mu2) * (  - 1./48.*m11*
         m33*RS1456ppe*RSe45633p - 1./24.*m11*m33*RSe45611p*RS14563pe
          - 1./48.*m11*m33*RSe45611p*RS1456pee - 1./48.*m11*m33*
         RSe45611p*RS1456ppe - 1./24.*m11*m33*RS145pe*RSe45613p + 1./48.
         *m11*m33*RS145pe*RSe45633p + 1./48.*m11*m33*RS145pe*RSe45611p
          - 1./24.*m11*m33*RS146pe*RSe45613p + 1./48.*m11*m33*RS146pe*
         RSe45633p + 1./48.*m11*m33*RS146pe*RSe45611p - 1./24.*m11*m33*
         RS156pe*RSe45613p + 1./48.*m11*m33*RS156pe*RSe45633p + 1./48.*
         m11*m33*RS156pe*RSe45611p + 1./24.*pow(m11,2)*RSe45613p*
         RS14563pe - 1./48.*pow(m11,2)*RSe45633p*RS14563pe + 1./48.*
         pow(m11,2)*RS1456pee*RSe45613p - 1./96.*pow(m11,2)*RS1456pee*
         RSe45633p + 1./48.*pow(m11,2)*RS1456ppe*RSe45613p - 1./96.*
         pow(m11,2)*RS1456ppe*RSe45633p - 1./48.*pow(m11,2)*RSe45611p*
         RS14563pe - 1./96.*pow(m11,2)*RSe45611p*RS1456pee - 1./96.*
         pow(m11,2)*RSe45611p*RS1456ppe - 1./48.*pow(m11,2)*RS145pe*
         RSe45613p );

      F13p6x23R +=  + hh21d(1,m13,m11,me2,m13,mu2) * ( 1./96.*pow(
         m11,2)*RS145pe*RSe45633p + 1./96.*pow(m11,2)*RS145pe*RSe45611p
          - 1./48.*pow(m11,2)*RS146pe*RSe45613p + 1./96.*pow(m11,2)*
         RS146pe*RSe45633p + 1./96.*pow(m11,2)*RS146pe*RSe45611p - 1./
         48.*pow(m11,2)*RS156pe*RSe45613p + 1./96.*pow(m11,2)*RS156pe*
         RSe45633p + 1./96.*pow(m11,2)*RS156pe*RSe45611p );

      F13p6x23R +=  + hh21d(1,m13,m11,mp2,m13,mu2) * ( 1./24.*pow(
         m33,2)*RSp45613e*RS14563pe - 1./48.*pow(m33,2)*RSp45633e*
         RS14563pe + 1./48.*pow(m33,2)*RS1456pee*RSp45613e - 1./96.*
         pow(m33,2)*RS1456pee*RSp45633e + 1./48.*pow(m33,2)*RS1456ppe*
         RSp45613e - 1./96.*pow(m33,2)*RS1456ppe*RSp45633e - 1./48.*
         pow(m33,2)*RSp45611e*RS14563pe - 1./96.*pow(m33,2)*RSp45611e*
         RS1456pee - 1./96.*pow(m33,2)*RSp45611e*RS1456ppe - 1./48.*
         pow(m33,2)*RS145pe*RSp45613e + 1./96.*pow(m33,2)*RS145pe*
         RSp45633e + 1./96.*pow(m33,2)*RS145pe*RSp45611e - 1./48.*pow(
         m33,2)*RS146pe*RSp45613e + 1./96.*pow(m33,2)*RS146pe*RSp45633e
          + 1./96.*pow(m33,2)*RS146pe*RSp45611e - 1./48.*pow(m33,2)*
         RS156pe*RSp45613e + 1./96.*pow(m33,2)*RS156pe*RSp45633e + 1./
         96.*pow(m33,2)*RS156pe*RSp45611e + 1./12.*m11*m33*RSp45613e*
         RS14563pe - 1./24.*m11*m33*RSp45633e*RS14563pe + 1./24.*m11*
         m33*RS1456pee*RSp45613e - 1./48.*m11*m33*RS1456pee*RSp45633e
          + 1./24.*m11*m33*RS1456ppe*RSp45613e );

      F13p6x23R +=  + hh21d(1,m13,m11,mp2,m13,mu2) * (  - 1./48.*m11*
         m33*RS1456ppe*RSp45633e - 1./24.*m11*m33*RSp45611e*RS14563pe
          - 1./48.*m11*m33*RSp45611e*RS1456pee - 1./48.*m11*m33*
         RSp45611e*RS1456ppe - 1./24.*m11*m33*RS145pe*RSp45613e + 1./48.
         *m11*m33*RS145pe*RSp45633e + 1./48.*m11*m33*RS145pe*RSp45611e
          - 1./24.*m11*m33*RS146pe*RSp45613e + 1./48.*m11*m33*RS146pe*
         RSp45633e + 1./48.*m11*m33*RS146pe*RSp45611e - 1./24.*m11*m33*
         RS156pe*RSp45613e + 1./48.*m11*m33*RS156pe*RSp45633e + 1./48.*
         m11*m33*RS156pe*RSp45611e + 1./24.*pow(m11,2)*RSp45613e*
         RS14563pe - 1./48.*pow(m11,2)*RSp45633e*RS14563pe + 1./48.*
         pow(m11,2)*RS1456pee*RSp45613e - 1./96.*pow(m11,2)*RS1456pee*
         RSp45633e + 1./48.*pow(m11,2)*RS1456ppe*RSp45613e - 1./96.*
         pow(m11,2)*RS1456ppe*RSp45633e - 1./48.*pow(m11,2)*RSp45611e*
         RS14563pe - 1./96.*pow(m11,2)*RSp45611e*RS1456pee - 1./96.*
         pow(m11,2)*RSp45611e*RS1456ppe - 1./48.*pow(m11,2)*RS145pe*
         RSp45613e );

      F13p6x23R +=  + hh21d(1,m13,m11,mp2,m13,mu2) * ( 1./96.*pow(
         m11,2)*RS145pe*RSp45633e + 1./96.*pow(m11,2)*RS145pe*RSp45611e
          - 1./48.*pow(m11,2)*RS146pe*RSp45613e + 1./96.*pow(m11,2)*
         RS146pe*RSp45633e + 1./96.*pow(m11,2)*RS146pe*RSp45611e - 1./
         48.*pow(m11,2)*RS156pe*RSp45613e + 1./96.*pow(m11,2)*RS156pe*
         RSp45633e + 1./96.*pow(m11,2)*RS156pe*RSp45611e );

      F13p6x23R +=  + hh21d(1,m13,m13,m13,m13,mu2) * ( 3./32.*pow(
         m33,2) + 3./16.*m11*m33 + 3./32.*pow(m11,2) );

      F13p6x23R +=  + hh21d(1,m13,m33,m33,m13,mu2) * ( 3./64.*pow(
         m33,2) + 1./48.*pow(m33,2)*pow(RS34561pe,2) + 1./48.*pow(
         m33,2)*RS3456pee*RS34561pe + 1./192.*pow(m33,2)*pow(
         RS3456pee,2) + 1./48.*pow(m33,2)*RS3456ppe*RS34561pe + 1./96.*
         pow(m33,2)*RS3456ppe*RS3456pee + 1./192.*pow(m33,2)*pow(
         RS3456ppe,2) - 1./48.*pow(m33,2)*RS345pe*RS34561pe - 1./96.*
         pow(m33,2)*RS345pe*RS3456pee - 1./96.*pow(m33,2)*RS345pe*
         RS3456ppe + 1./192.*pow(m33,2)*pow(RS345pe,2) - 1./48.*pow(
         m33,2)*RS346pe*RS34561pe - 1./96.*pow(m33,2)*RS346pe*RS3456pee
          - 1./96.*pow(m33,2)*RS346pe*RS3456ppe + 1./96.*pow(m33,2)*
         RS346pe*RS345pe + 1./192.*pow(m33,2)*pow(RS346pe,2) - 1./48.*
         pow(m33,2)*RS356pe*RS34561pe - 1./96.*pow(m33,2)*RS356pe*
         RS3456pee - 1./96.*pow(m33,2)*RS356pe*RS3456ppe + 1./96.*pow(
         m33,2)*RS356pe*RS345pe + 1./96.*pow(m33,2)*RS356pe*RS346pe + 1.
         /192.*pow(m33,2)*pow(RS356pe,2) + 3./32.*m11*m33 + 1./24.*m11*
         m33*pow(RS34561pe,2) );

      F13p6x23R +=  + hh21d(1,m13,m33,m33,m13,mu2) * ( 1./24.*m11*m33*
         RS3456pee*RS34561pe + 1./96.*m11*m33*pow(RS3456pee,2) + 1./24.
         *m11*m33*RS3456ppe*RS34561pe + 1./48.*m11*m33*RS3456ppe*
         RS3456pee + 1./96.*m11*m33*pow(RS3456ppe,2) - 1./24.*m11*m33*
         RS345pe*RS34561pe - 1./48.*m11*m33*RS345pe*RS3456pee - 1./48.*
         m11*m33*RS345pe*RS3456ppe + 1./96.*m11*m33*pow(RS345pe,2) - 1./
         24.*m11*m33*RS346pe*RS34561pe - 1./48.*m11*m33*RS346pe*
         RS3456pee - 1./48.*m11*m33*RS346pe*RS3456ppe + 1./48.*m11*m33*
         RS346pe*RS345pe + 1./96.*m11*m33*pow(RS346pe,2) - 1./24.*m11*
         m33*RS356pe*RS34561pe - 1./48.*m11*m33*RS356pe*RS3456pee - 1./
         48.*m11*m33*RS356pe*RS3456ppe + 1./48.*m11*m33*RS356pe*RS345pe
          + 1./48.*m11*m33*RS356pe*RS346pe + 1./96.*m11*m33*pow(
         RS356pe,2) + 3./64.*pow(m11,2) + 1./48.*pow(m11,2)*pow(
         RS34561pe,2) + 1./48.*pow(m11,2)*RS3456pee*RS34561pe + 1./192.
         *pow(m11,2)*pow(RS3456pee,2) + 1./48.*pow(m11,2)*RS3456ppe*
         RS34561pe );

      F13p6x23R +=  + hh21d(1,m13,m33,m33,m13,mu2) * ( 1./96.*pow(
         m11,2)*RS3456ppe*RS3456pee + 1./192.*pow(m11,2)*pow(
         RS3456ppe,2) - 1./48.*pow(m11,2)*RS345pe*RS34561pe - 1./96.*
         pow(m11,2)*RS345pe*RS3456pee - 1./96.*pow(m11,2)*RS345pe*
         RS3456ppe + 1./192.*pow(m11,2)*pow(RS345pe,2) - 1./48.*pow(
         m11,2)*RS346pe*RS34561pe - 1./96.*pow(m11,2)*RS346pe*RS3456pee
          - 1./96.*pow(m11,2)*RS346pe*RS3456ppe + 1./96.*pow(m11,2)*
         RS346pe*RS345pe + 1./192.*pow(m11,2)*pow(RS346pe,2) - 1./48.*
         pow(m11,2)*RS356pe*RS34561pe - 1./96.*pow(m11,2)*RS356pe*
         RS3456pee - 1./96.*pow(m11,2)*RS356pe*RS3456ppe + 1./96.*pow(
         m11,2)*RS356pe*RS345pe + 1./96.*pow(m11,2)*RS356pe*RS346pe + 1.
         /192.*pow(m11,2)*pow(RS356pe,2) );

      F13p6x23R +=  + hh21d(1,m13,m33,me2,m13,mu2) * ( 1./24.*pow(
         m33,2)*RSe45613p*RS34561pe + 1./48.*pow(m33,2)*RS3456pee*
         RSe45613p + 1./48.*pow(m33,2)*RS3456ppe*RSe45613p - 1./48.*
         pow(m33,2)*RSe45633p*RS34561pe - 1./96.*pow(m33,2)*RSe45633p*
         RS3456pee - 1./96.*pow(m33,2)*RSe45633p*RS3456ppe - 1./48.*
         pow(m33,2)*RS345pe*RSe45613p + 1./96.*pow(m33,2)*RS345pe*
         RSe45633p - 1./48.*pow(m33,2)*RS346pe*RSe45613p + 1./96.*pow(
         m33,2)*RS346pe*RSe45633p - 1./48.*pow(m33,2)*RS356pe*RSe45613p
          + 1./96.*pow(m33,2)*RS356pe*RSe45633p - 1./48.*pow(m33,2)*
         RSe45611p*RS34561pe - 1./96.*pow(m33,2)*RSe45611p*RS3456pee - 
         1./96.*pow(m33,2)*RSe45611p*RS3456ppe + 1./96.*pow(m33,2)*
         RSe45611p*RS345pe + 1./96.*pow(m33,2)*RSe45611p*RS346pe + 1./
         96.*pow(m33,2)*RSe45611p*RS356pe + 1./12.*m11*m33*RSe45613p*
         RS34561pe + 1./24.*m11*m33*RS3456pee*RSe45613p + 1./24.*m11*
         m33*RS3456ppe*RSe45613p - 1./24.*m11*m33*RSe45633p*RS34561pe
          - 1./48.*m11*m33*RSe45633p*RS3456pee );

      F13p6x23R +=  + hh21d(1,m13,m33,me2,m13,mu2) * (  - 1./48.*m11*
         m33*RSe45633p*RS3456ppe - 1./24.*m11*m33*RS345pe*RSe45613p + 1.
         /48.*m11*m33*RS345pe*RSe45633p - 1./24.*m11*m33*RS346pe*
         RSe45613p + 1./48.*m11*m33*RS346pe*RSe45633p - 1./24.*m11*m33*
         RS356pe*RSe45613p + 1./48.*m11*m33*RS356pe*RSe45633p - 1./24.*
         m11*m33*RSe45611p*RS34561pe - 1./48.*m11*m33*RSe45611p*
         RS3456pee - 1./48.*m11*m33*RSe45611p*RS3456ppe + 1./48.*m11*
         m33*RSe45611p*RS345pe + 1./48.*m11*m33*RSe45611p*RS346pe + 1./
         48.*m11*m33*RSe45611p*RS356pe + 1./24.*pow(m11,2)*RSe45613p*
         RS34561pe + 1./48.*pow(m11,2)*RS3456pee*RSe45613p + 1./48.*
         pow(m11,2)*RS3456ppe*RSe45613p - 1./48.*pow(m11,2)*RSe45633p*
         RS34561pe - 1./96.*pow(m11,2)*RSe45633p*RS3456pee - 1./96.*
         pow(m11,2)*RSe45633p*RS3456ppe - 1./48.*pow(m11,2)*RS345pe*
         RSe45613p + 1./96.*pow(m11,2)*RS345pe*RSe45633p - 1./48.*pow(
         m11,2)*RS346pe*RSe45613p + 1./96.*pow(m11,2)*RS346pe*RSe45633p
          - 1./48.*pow(m11,2)*RS356pe*RSe45613p );

      F13p6x23R +=  + hh21d(1,m13,m33,me2,m13,mu2) * ( 1./96.*pow(
         m11,2)*RS356pe*RSe45633p - 1./48.*pow(m11,2)*RSe45611p*
         RS34561pe - 1./96.*pow(m11,2)*RSe45611p*RS3456pee - 1./96.*
         pow(m11,2)*RSe45611p*RS3456ppe + 1./96.*pow(m11,2)*RSe45611p*
         RS345pe + 1./96.*pow(m11,2)*RSe45611p*RS346pe + 1./96.*pow(
         m11,2)*RSe45611p*RS356pe );

      F13p6x23R +=  + hh21d(1,m13,m33,mp2,m13,mu2) * ( 1./24.*pow(
         m33,2)*RSp45613e*RS34561pe + 1./48.*pow(m33,2)*RS3456pee*
         RSp45613e + 1./48.*pow(m33,2)*RS3456ppe*RSp45613e - 1./48.*
         pow(m33,2)*RSp45633e*RS34561pe - 1./96.*pow(m33,2)*RSp45633e*
         RS3456pee - 1./96.*pow(m33,2)*RSp45633e*RS3456ppe - 1./48.*
         pow(m33,2)*RS345pe*RSp45613e + 1./96.*pow(m33,2)*RS345pe*
         RSp45633e - 1./48.*pow(m33,2)*RS346pe*RSp45613e + 1./96.*pow(
         m33,2)*RS346pe*RSp45633e - 1./48.*pow(m33,2)*RS356pe*RSp45613e
          + 1./96.*pow(m33,2)*RS356pe*RSp45633e - 1./48.*pow(m33,2)*
         RSp45611e*RS34561pe - 1./96.*pow(m33,2)*RSp45611e*RS3456pee - 
         1./96.*pow(m33,2)*RSp45611e*RS3456ppe + 1./96.*pow(m33,2)*
         RSp45611e*RS345pe + 1./96.*pow(m33,2)*RSp45611e*RS346pe + 1./
         96.*pow(m33,2)*RSp45611e*RS356pe + 1./12.*m11*m33*RSp45613e*
         RS34561pe + 1./24.*m11*m33*RS3456pee*RSp45613e + 1./24.*m11*
         m33*RS3456ppe*RSp45613e - 1./24.*m11*m33*RSp45633e*RS34561pe
          - 1./48.*m11*m33*RSp45633e*RS3456pee );

      F13p6x23R +=  + hh21d(1,m13,m33,mp2,m13,mu2) * (  - 1./48.*m11*
         m33*RSp45633e*RS3456ppe - 1./24.*m11*m33*RS345pe*RSp45613e + 1.
         /48.*m11*m33*RS345pe*RSp45633e - 1./24.*m11*m33*RS346pe*
         RSp45613e + 1./48.*m11*m33*RS346pe*RSp45633e - 1./24.*m11*m33*
         RS356pe*RSp45613e + 1./48.*m11*m33*RS356pe*RSp45633e - 1./24.*
         m11*m33*RSp45611e*RS34561pe - 1./48.*m11*m33*RSp45611e*
         RS3456pee - 1./48.*m11*m33*RSp45611e*RS3456ppe + 1./48.*m11*
         m33*RSp45611e*RS345pe + 1./48.*m11*m33*RSp45611e*RS346pe + 1./
         48.*m11*m33*RSp45611e*RS356pe + 1./24.*pow(m11,2)*RSp45613e*
         RS34561pe + 1./48.*pow(m11,2)*RS3456pee*RSp45613e + 1./48.*
         pow(m11,2)*RS3456ppe*RSp45613e - 1./48.*pow(m11,2)*RSp45633e*
         RS34561pe - 1./96.*pow(m11,2)*RSp45633e*RS3456pee - 1./96.*
         pow(m11,2)*RSp45633e*RS3456ppe - 1./48.*pow(m11,2)*RS345pe*
         RSp45613e + 1./96.*pow(m11,2)*RS345pe*RSp45633e - 1./48.*pow(
         m11,2)*RS346pe*RSp45613e + 1./96.*pow(m11,2)*RS346pe*RSp45633e
          - 1./48.*pow(m11,2)*RS356pe*RSp45613e );

      F13p6x23R +=  + hh21d(1,m13,m33,mp2,m13,mu2) * ( 1./96.*pow(
         m11,2)*RS356pe*RSp45633e - 1./48.*pow(m11,2)*RSp45611e*
         RS34561pe - 1./96.*pow(m11,2)*RSp45611e*RS3456pee - 1./96.*
         pow(m11,2)*RSp45611e*RS3456ppe + 1./96.*pow(m11,2)*RSp45611e*
         RS345pe + 1./96.*pow(m11,2)*RSp45611e*RS346pe + 1./96.*pow(
         m11,2)*RSp45611e*RS356pe );

      F13p6x23R +=  + hh21d(1,m13,me2,me2,m13,mu2) * ( 1./48.*pow(
         m33,2)*pow(RSe45613p,2) - 1./48.*pow(m33,2)*RSe45633p*
         RSe45613p + 1./192.*pow(m33,2)*pow(RSe45633p,2) - 1./48.*pow(
         m33,2)*RSe45611p*RSe45613p + 1./96.*pow(m33,2)*RSe45611p*
         RSe45633p + 1./192.*pow(m33,2)*pow(RSe45611p,2) + 1./24.*m11*
         m33*pow(RSe45613p,2) - 1./24.*m11*m33*RSe45633p*RSe45613p + 1./
         96.*m11*m33*pow(RSe45633p,2) - 1./24.*m11*m33*RSe45611p*
         RSe45613p + 1./48.*m11*m33*RSe45611p*RSe45633p + 1./96.*m11*
         m33*pow(RSe45611p,2) + 1./48.*pow(m11,2)*pow(RSe45613p,2) - 1./
         48.*pow(m11,2)*RSe45633p*RSe45613p + 1./192.*pow(m11,2)*pow(
         RSe45633p,2) - 1./48.*pow(m11,2)*RSe45611p*RSe45613p + 1./96.*
         pow(m11,2)*RSe45611p*RSe45633p + 1./192.*pow(m11,2)*pow(
         RSe45611p,2) );

      F13p6x23R +=  + hh21d(1,m13,me2,mp2,m13,mu2) * ( 1./24.*pow(
         m33,2)*RSp45613e*RSe45613p - 1./48.*pow(m33,2)*RSe45633p*
         RSp45613e - 1./48.*pow(m33,2)*RSp45633e*RSe45613p + 1./96.*
         pow(m33,2)*RSp45633e*RSe45633p - 1./48.*pow(m33,2)*RSe45611p*
         RSp45613e + 1./96.*pow(m33,2)*RSe45611p*RSp45633e - 1./48.*
         pow(m33,2)*RSp45611e*RSe45613p + 1./96.*pow(m33,2)*RSp45611e*
         RSe45633p + 1./96.*pow(m33,2)*RSp45611e*RSe45611p + 1./12.*m11
         *m33*RSp45613e*RSe45613p - 1./24.*m11*m33*RSe45633p*RSp45613e
          - 1./24.*m11*m33*RSp45633e*RSe45613p + 1./48.*m11*m33*
         RSp45633e*RSe45633p - 1./24.*m11*m33*RSe45611p*RSp45613e + 1./
         48.*m11*m33*RSe45611p*RSp45633e - 1./24.*m11*m33*RSp45611e*
         RSe45613p + 1./48.*m11*m33*RSp45611e*RSe45633p + 1./48.*m11*
         m33*RSp45611e*RSe45611p + 1./24.*pow(m11,2)*RSp45613e*
         RSe45613p - 1./48.*pow(m11,2)*RSe45633p*RSp45613e - 1./48.*
         pow(m11,2)*RSp45633e*RSe45613p + 1./96.*pow(m11,2)*RSp45633e*
         RSe45633p );

      F13p6x23R +=  + hh21d(1,m13,me2,mp2,m13,mu2) * (  - 1./48.*pow(
         m11,2)*RSe45611p*RSp45613e + 1./96.*pow(m11,2)*RSe45611p*
         RSp45633e - 1./48.*pow(m11,2)*RSp45611e*RSe45613p + 1./96.*
         pow(m11,2)*RSp45611e*RSe45633p + 1./96.*pow(m11,2)*RSp45611e*
         RSe45611p );

      F13p6x23R +=  + hh21d(1,m13,mp2,mp2,m13,mu2) * ( 1./48.*pow(
         m33,2)*pow(RSp45613e,2) - 1./48.*pow(m33,2)*RSp45633e*
         RSp45613e + 1./192.*pow(m33,2)*pow(RSp45633e,2) - 1./48.*pow(
         m33,2)*RSp45611e*RSp45613e + 1./96.*pow(m33,2)*RSp45611e*
         RSp45633e + 1./192.*pow(m33,2)*pow(RSp45611e,2) + 1./24.*m11*
         m33*pow(RSp45613e,2) - 1./24.*m11*m33*RSp45633e*RSp45613e + 1./
         96.*m11*m33*pow(RSp45633e,2) - 1./24.*m11*m33*RSp45611e*
         RSp45613e + 1./48.*m11*m33*RSp45611e*RSp45633e + 1./96.*m11*
         m33*pow(RSp45611e,2) + 1./48.*pow(m11,2)*pow(RSp45613e,2) - 1./
         48.*pow(m11,2)*RSp45633e*RSp45613e + 1./192.*pow(m11,2)*pow(
         RSp45633e,2) - 1./48.*pow(m11,2)*RSp45611e*RSp45613e + 1./96.*
         pow(m11,2)*RSp45611e*RSp45633e + 1./192.*pow(m11,2)*pow(
         RSp45611e,2) );

      F13p6x23R +=  + hh21d(1,m14,m11,m34,m13,mu2) * ( 1./32.*pow(
         m33,2)*RS14563pe + 1./32.*pow(m33,2)*RS1456pee + 1./32.*pow(
         m33,2)*RS1456ppe - 1./32.*pow(m33,2)*RS145pe - 1./32.*pow(
         m33,2)*RS146pe + 1./16.*m11*m33*RS14563pe + 1./16.*m11*m33*
         RS1456pee + 1./16.*m11*m33*RS1456ppe - 1./16.*m11*m33*RS145pe
          - 1./16.*m11*m33*RS146pe + 1./32.*pow(m11,2)*RS14563pe + 1./
         32.*pow(m11,2)*RS1456pee + 1./32.*pow(m11,2)*RS1456ppe - 1./32.
         *pow(m11,2)*RS145pe - 1./32.*pow(m11,2)*RS146pe );

      F13p6x23R +=  + hh21d(1,m14,m33,m34,m13,mu2) * ( 1./32.*pow(
         m33,2)*RS34561pe - 1./32.*pow(m33,2)*RS356pe + 1./16.*m11*m33*
         RS34561pe - 1./16.*m11*m33*RS356pe + 1./32.*pow(m11,2)*
         RS34561pe - 1./32.*pow(m11,2)*RS356pe );

      F13p6x23R +=  + hh21d(1,m14,m34,me2,m13,mu2) * ( 1./32.*pow(
         m33,2)*RSe45613p - 1./32.*pow(m33,2)*RSe563p - 1./32.*pow(
         m33,2)*RSe45611p + 1./32.*pow(m33,2)*RSe561p + 1./16.*m11*m33*
         RSe45613p - 1./16.*m11*m33*RSe563p - 1./16.*m11*m33*RSe45611p
          + 1./16.*m11*m33*RSe561p + 1./32.*pow(m11,2)*RSe45613p - 1./
         32.*pow(m11,2)*RSe563p - 1./32.*pow(m11,2)*RSe45611p + 1./32.*
         pow(m11,2)*RSe561p );

      F13p6x23R +=  + hh21d(1,m14,m34,mp2,m13,mu2) * ( 1./32.*pow(
         m33,2)*RSp45613e - 1./32.*pow(m33,2)*RSp563e - 1./32.*pow(
         m33,2)*RSp45611e + 1./32.*pow(m33,2)*RSp561e + 1./16.*m11*m33*
         RSp45613e - 1./16.*m11*m33*RSp563e - 1./16.*m11*m33*RSp45611e
          + 1./16.*m11*m33*RSp561e + 1./32.*pow(m11,2)*RSp45613e - 1./
         32.*pow(m11,2)*RSp563e - 1./32.*pow(m11,2)*RSp45611e + 1./32.*
         pow(m11,2)*RSp561e );

      F13p6x23R +=  + hh21d(1,m15,m11,m35,m13,mu2) * ( 1./32.*pow(
         m33,2)*RS14563pe + 1./32.*pow(m33,2)*RS1456pee + 1./32.*pow(
         m33,2)*RS1456ppe - 1./32.*pow(m33,2)*RS145pe - 1./32.*pow(
         m33,2)*RS156pe + 1./16.*m11*m33*RS14563pe + 1./16.*m11*m33*
         RS1456pee + 1./16.*m11*m33*RS1456ppe - 1./16.*m11*m33*RS145pe
          - 1./16.*m11*m33*RS156pe + 1./32.*pow(m11,2)*RS14563pe + 1./
         32.*pow(m11,2)*RS1456pee + 1./32.*pow(m11,2)*RS1456ppe - 1./32.
         *pow(m11,2)*RS145pe - 1./32.*pow(m11,2)*RS156pe );

      F13p6x23R +=  + hh21d(1,m15,m33,m35,m13,mu2) * ( 1./32.*pow(
         m33,2)*RS34561pe - 1./32.*pow(m33,2)*RS346pe + 1./16.*m11*m33*
         RS34561pe - 1./16.*m11*m33*RS346pe + 1./32.*pow(m11,2)*
         RS34561pe - 1./32.*pow(m11,2)*RS346pe );

      F13p6x23R +=  + hh21d(1,m15,m35,me2,m13,mu2) * ( 1./32.*pow(
         m33,2)*RSe45613p - 1./32.*pow(m33,2)*RSe463p - 1./32.*pow(
         m33,2)*RSe45611p + 1./32.*pow(m33,2)*RSe461p + 1./16.*m11*m33*
         RSe45613p - 1./16.*m11*m33*RSe463p - 1./16.*m11*m33*RSe45611p
          + 1./16.*m11*m33*RSe461p + 1./32.*pow(m11,2)*RSe45613p - 1./
         32.*pow(m11,2)*RSe463p - 1./32.*pow(m11,2)*RSe45611p + 1./32.*
         pow(m11,2)*RSe461p );

      F13p6x23R +=  + hh21d(1,m15,m35,mp2,m13,mu2) * ( 1./32.*pow(
         m33,2)*RSp45613e - 1./32.*pow(m33,2)*RSp463e - 1./32.*pow(
         m33,2)*RSp45611e + 1./32.*pow(m33,2)*RSp461e + 1./16.*m11*m33*
         RSp45613e - 1./16.*m11*m33*RSp463e - 1./16.*m11*m33*RSp45611e
          + 1./16.*m11*m33*RSp461e + 1./32.*pow(m11,2)*RSp45613e - 1./
         32.*pow(m11,2)*RSp463e - 1./32.*pow(m11,2)*RSp45611e + 1./32.*
         pow(m11,2)*RSp461e );

      F13p6x23R +=  + hh21d(1,m16,m11,m36,m13,mu2) * ( 1./32.*pow(
         m33,2)*RS14563pe + 1./32.*pow(m33,2)*RS1456pee + 1./32.*pow(
         m33,2)*RS1456ppe - 1./32.*pow(m33,2)*RS146pe - 1./32.*pow(
         m33,2)*RS156pe + 1./16.*m11*m33*RS14563pe + 1./16.*m11*m33*
         RS1456pee + 1./16.*m11*m33*RS1456ppe - 1./16.*m11*m33*RS146pe
          - 1./16.*m11*m33*RS156pe + 1./32.*pow(m11,2)*RS14563pe + 1./
         32.*pow(m11,2)*RS1456pee + 1./32.*pow(m11,2)*RS1456ppe - 1./32.
         *pow(m11,2)*RS146pe - 1./32.*pow(m11,2)*RS156pe );

      F13p6x23R +=  + hh21d(1,m16,m33,m36,m13,mu2) * ( 1./32.*pow(
         m33,2)*RS34561pe - 1./32.*pow(m33,2)*RS345pe + 1./16.*m11*m33*
         RS34561pe - 1./16.*m11*m33*RS345pe + 1./32.*pow(m11,2)*
         RS34561pe - 1./32.*pow(m11,2)*RS345pe );

      F13p6x23R +=  + hh21d(1,m16,m36,me2,m13,mu2) * ( 1./32.*pow(
         m33,2)*RSe45613p - 1./32.*pow(m33,2)*RSe453p - 1./32.*pow(
         m33,2)*RSe45611p + 1./32.*pow(m33,2)*RSe451p + 1./16.*m11*m33*
         RSe45613p - 1./16.*m11*m33*RSe453p - 1./16.*m11*m33*RSe45611p
          + 1./16.*m11*m33*RSe451p + 1./32.*pow(m11,2)*RSe45613p - 1./
         32.*pow(m11,2)*RSe453p - 1./32.*pow(m11,2)*RSe45611p + 1./32.*
         pow(m11,2)*RSe451p );

      F13p6x23R +=  + hh21d(1,m16,m36,mp2,m13,mu2) * ( 1./32.*pow(
         m33,2)*RSp45613e - 1./32.*pow(m33,2)*RSp453e - 1./32.*pow(
         m33,2)*RSp45611e + 1./32.*pow(m33,2)*RSp451e + 1./16.*m11*m33*
         RSp45613e - 1./16.*m11*m33*RSp453e - 1./16.*m11*m33*RSp45611e
          + 1./16.*m11*m33*RSp451e + 1./32.*pow(m11,2)*RSp45613e - 1./
         32.*pow(m11,2)*RSp453e - 1./32.*pow(m11,2)*RSp45611e + 1./32.*
         pow(m11,2)*RSp451e );

      F13p6x23R +=  + hh21d(1,m33,m13,m33,m13,mu2) * (  - 3./32.*pow(
         m33,2) - 3./16.*m11*m33 - 3./32.*pow(m11,2) );

      F13p6x23R +=  + hh21d(1,m33,m14,m34,m13,mu2) * (  - 1./32.*pow(
         m33,2)*RS34561pe + 1./32.*pow(m33,2)*RS356pe - 1./16.*m11*m33*
         RS34561pe + 1./16.*m11*m33*RS356pe - 1./32.*pow(m11,2)*
         RS34561pe + 1./32.*pow(m11,2)*RS356pe );

      F13p6x23R +=  + hh21d(1,m33,m15,m35,m13,mu2) * (  - 1./32.*pow(
         m33,2)*RS34561pe + 1./32.*pow(m33,2)*RS346pe - 1./16.*m11*m33*
         RS34561pe + 1./16.*m11*m33*RS346pe - 1./32.*pow(m11,2)*
         RS34561pe + 1./32.*pow(m11,2)*RS346pe );

      F13p6x23R +=  + hh21d(1,m33,m16,m36,m13,mu2) * (  - 1./32.*pow(
         m33,2)*RS34561pe + 1./32.*pow(m33,2)*RS345pe - 1./16.*m11*m33*
         RS34561pe + 1./16.*m11*m33*RS345pe - 1./32.*pow(m11,2)*
         RS34561pe + 1./32.*pow(m11,2)*RS345pe );

      F13p6x23R +=  + hh21d(1,m34,m11,m14,m13,mu2) * ( 1./32.*pow(
         m33,2)*RS14563pe - 1./32.*pow(m33,2)*RS156pe + 1./16.*m11*m33*
         RS14563pe - 1./16.*m11*m33*RS156pe + 1./32.*pow(m11,2)*
         RS14563pe - 1./32.*pow(m11,2)*RS156pe );

      F13p6x23R +=  + hh21d(1,m34,m14,m33,m13,mu2) * ( 1./32.*pow(
         m33,2)*RS34561pe + 1./32.*pow(m33,2)*RS3456pee + 1./32.*pow(
         m33,2)*RS3456ppe - 1./32.*pow(m33,2)*RS345pe - 1./32.*pow(
         m33,2)*RS346pe + 1./16.*m11*m33*RS34561pe + 1./16.*m11*m33*
         RS3456pee + 1./16.*m11*m33*RS3456ppe - 1./16.*m11*m33*RS345pe
          - 1./16.*m11*m33*RS346pe + 1./32.*pow(m11,2)*RS34561pe + 1./
         32.*pow(m11,2)*RS3456pee + 1./32.*pow(m11,2)*RS3456ppe - 1./32.
         *pow(m11,2)*RS345pe - 1./32.*pow(m11,2)*RS346pe );

      F13p6x23R +=  + hh21d(1,m34,m14,me2,m13,mu2) * ( 1./32.*pow(
         m33,2)*RSe45613p - 1./32.*pow(m33,2)*RSe45633p + 1./32.*pow(
         m33,2)*RSe563p - 1./32.*pow(m33,2)*RSe561p + 1./16.*m11*m33*
         RSe45613p - 1./16.*m11*m33*RSe45633p + 1./16.*m11*m33*RSe563p
          - 1./16.*m11*m33*RSe561p + 1./32.*pow(m11,2)*RSe45613p - 1./
         32.*pow(m11,2)*RSe45633p + 1./32.*pow(m11,2)*RSe563p - 1./32.*
         pow(m11,2)*RSe561p );

      F13p6x23R +=  + hh21d(1,m34,m14,mp2,m13,mu2) * ( 1./32.*pow(
         m33,2)*RSp45613e - 1./32.*pow(m33,2)*RSp45633e + 1./32.*pow(
         m33,2)*RSp563e - 1./32.*pow(m33,2)*RSp561e + 1./16.*m11*m33*
         RSp45613e - 1./16.*m11*m33*RSp45633e + 1./16.*m11*m33*RSp563e
          - 1./16.*m11*m33*RSp561e + 1./32.*pow(m11,2)*RSp45613e - 1./
         32.*pow(m11,2)*RSp45633e + 1./32.*pow(m11,2)*RSp563e - 1./32.*
         pow(m11,2)*RSp561e );

      F13p6x23R +=  + hh21d(1,m35,m11,m15,m13,mu2) * ( 1./32.*pow(
         m33,2)*RS14563pe - 1./32.*pow(m33,2)*RS146pe + 1./16.*m11*m33*
         RS14563pe - 1./16.*m11*m33*RS146pe + 1./32.*pow(m11,2)*
         RS14563pe - 1./32.*pow(m11,2)*RS146pe );

      F13p6x23R +=  + hh21d(1,m35,m15,m33,m13,mu2) * ( 1./32.*pow(
         m33,2)*RS34561pe + 1./32.*pow(m33,2)*RS3456pee + 1./32.*pow(
         m33,2)*RS3456ppe - 1./32.*pow(m33,2)*RS345pe - 1./32.*pow(
         m33,2)*RS356pe + 1./16.*m11*m33*RS34561pe + 1./16.*m11*m33*
         RS3456pee + 1./16.*m11*m33*RS3456ppe - 1./16.*m11*m33*RS345pe
          - 1./16.*m11*m33*RS356pe + 1./32.*pow(m11,2)*RS34561pe + 1./
         32.*pow(m11,2)*RS3456pee + 1./32.*pow(m11,2)*RS3456ppe - 1./32.
         *pow(m11,2)*RS345pe - 1./32.*pow(m11,2)*RS356pe );

      F13p6x23R +=  + hh21d(1,m35,m15,me2,m13,mu2) * ( 1./32.*pow(
         m33,2)*RSe45613p - 1./32.*pow(m33,2)*RSe45633p + 1./32.*pow(
         m33,2)*RSe463p - 1./32.*pow(m33,2)*RSe461p + 1./16.*m11*m33*
         RSe45613p - 1./16.*m11*m33*RSe45633p + 1./16.*m11*m33*RSe463p
          - 1./16.*m11*m33*RSe461p + 1./32.*pow(m11,2)*RSe45613p - 1./
         32.*pow(m11,2)*RSe45633p + 1./32.*pow(m11,2)*RSe463p - 1./32.*
         pow(m11,2)*RSe461p );

      F13p6x23R +=  + hh21d(1,m35,m15,mp2,m13,mu2) * ( 1./32.*pow(
         m33,2)*RSp45613e - 1./32.*pow(m33,2)*RSp45633e + 1./32.*pow(
         m33,2)*RSp463e - 1./32.*pow(m33,2)*RSp461e + 1./16.*m11*m33*
         RSp45613e - 1./16.*m11*m33*RSp45633e + 1./16.*m11*m33*RSp463e
          - 1./16.*m11*m33*RSp461e + 1./32.*pow(m11,2)*RSp45613e - 1./
         32.*pow(m11,2)*RSp45633e + 1./32.*pow(m11,2)*RSp463e - 1./32.*
         pow(m11,2)*RSp461e );

      F13p6x23R +=  + hh21d(1,m36,m11,m16,m13,mu2) * ( 1./32.*pow(
         m33,2)*RS14563pe - 1./32.*pow(m33,2)*RS145pe + 1./16.*m11*m33*
         RS14563pe - 1./16.*m11*m33*RS145pe + 1./32.*pow(m11,2)*
         RS14563pe - 1./32.*pow(m11,2)*RS145pe );

      F13p6x23R +=  + hh21d(1,m36,m16,m33,m13,mu2) * ( 1./32.*pow(
         m33,2)*RS34561pe + 1./32.*pow(m33,2)*RS3456pee + 1./32.*pow(
         m33,2)*RS3456ppe - 1./32.*pow(m33,2)*RS346pe - 1./32.*pow(
         m33,2)*RS356pe + 1./16.*m11*m33*RS34561pe + 1./16.*m11*m33*
         RS3456pee + 1./16.*m11*m33*RS3456ppe - 1./16.*m11*m33*RS346pe
          - 1./16.*m11*m33*RS356pe + 1./32.*pow(m11,2)*RS34561pe + 1./
         32.*pow(m11,2)*RS3456pee + 1./32.*pow(m11,2)*RS3456ppe - 1./32.
         *pow(m11,2)*RS346pe - 1./32.*pow(m11,2)*RS356pe );

      F13p6x23R +=  + hh21d(1,m36,m16,me2,m13,mu2) * ( 1./32.*pow(
         m33,2)*RSe45613p - 1./32.*pow(m33,2)*RSe45633p + 1./32.*pow(
         m33,2)*RSe453p - 1./32.*pow(m33,2)*RSe451p + 1./16.*m11*m33*
         RSe45613p - 1./16.*m11*m33*RSe45633p + 1./16.*m11*m33*RSe453p
          - 1./16.*m11*m33*RSe451p + 1./32.*pow(m11,2)*RSe45613p - 1./
         32.*pow(m11,2)*RSe45633p + 1./32.*pow(m11,2)*RSe453p - 1./32.*
         pow(m11,2)*RSe451p );

      F13p6x23R +=  + hh21d(1,m36,m16,mp2,m13,mu2) * ( 1./32.*pow(
         m33,2)*RSp45613e - 1./32.*pow(m33,2)*RSp45633e + 1./32.*pow(
         m33,2)*RSp453e - 1./32.*pow(m33,2)*RSp451e + 1./16.*m11*m33*
         RSp45613e - 1./16.*m11*m33*RSp45633e + 1./16.*m11*m33*RSp453e
          - 1./16.*m11*m33*RSp451e + 1./32.*pow(m11,2)*RSp45613e - 1./
         32.*pow(m11,2)*RSp45633e + 1./32.*pow(m11,2)*RSp453e - 1./32.*
         pow(m11,2)*RSp451e );

      F13p6x23R +=  + hh21d(1,m44,m14,m34,m13,mu2) * ( 3./32.*pow(
         m33,2) - 1./32.*pow(m33,2)*RS456pe + 3./16.*m11*m33 - 1./16.*
         m11*m33*RS456pe + 3./32.*pow(m11,2) - 1./32.*pow(m11,2)*
         RS456pe );

      F13p6x23R +=  + hh21d(1,m45,m14,m35,m13,mu2) * ( 3./32.*pow(
         m33,2) + 3./16.*m11*m33 + 3./32.*pow(m11,2) );

      F13p6x23R +=  + hh21d(1,m45,m15,m34,m13,mu2) * ( 3./32.*pow(
         m33,2) + 3./16.*m11*m33 + 3./32.*pow(m11,2) );

      F13p6x23R +=  + hh21d(1,m46,m14,m36,m13,mu2) * ( 3./32.*pow(
         m33,2) + 3./16.*m11*m33 + 3./32.*pow(m11,2) );

      F13p6x23R +=  + hh21d(1,m46,m16,m34,m13,mu2) * ( 3./32.*pow(
         m33,2) + 3./16.*m11*m33 + 3./32.*pow(m11,2) );

      F13p6x23R +=  + hh21d(1,m55,m15,m35,m13,mu2) * ( 3./32.*pow(
         m33,2) - 1./32.*pow(m33,2)*RS546pe + 3./16.*m11*m33 - 1./16.*
         m11*m33*RS546pe + 3./32.*pow(m11,2) - 1./32.*pow(m11,2)*
         RS546pe );

      F13p6x23R +=  + hh21d(1,m56,m15,m36,m13,mu2) * ( 3./32.*pow(
         m33,2) + 3./16.*m11*m33 + 3./32.*pow(m11,2) );

      F13p6x23R +=  + hh21d(1,m56,m16,m35,m13,mu2) * ( 3./32.*pow(
         m33,2) + 3./16.*m11*m33 + 3./32.*pow(m11,2) );

      F13p6x23R +=  + hh21d(1,m66,m16,m36,m13,mu2) * ( 3./32.*pow(
         m33,2) - 1./32.*pow(m33,2)*RS645pe + 3./16.*m11*m33 - 1./16.*
         m11*m33*RS645pe + 3./32.*pow(m11,2) - 1./32.*pow(m11,2)*
         RS645pe );

      F13p6x23R +=  + hh21d(1,me2,m14,m34,m13,mu2) * (  - 1./32.*pow(
         m33,2)*RSe45613p + 1./32.*pow(m33,2)*RSe563p + 1./32.*pow(
         m33,2)*RSe561p - 1./32.*pow(m33,2)*RSe564p - 1./16.*m11*m33*
         RSe45613p + 1./16.*m11*m33*RSe563p + 1./16.*m11*m33*RSe561p - 
         1./16.*m11*m33*RSe564p - 1./32.*pow(m11,2)*RSe45613p + 1./32.*
         pow(m11,2)*RSe563p + 1./32.*pow(m11,2)*RSe561p - 1./32.*pow(
         m11,2)*RSe564p );

      F13p6x23R +=  + hh21d(1,me2,m15,m35,m13,mu2) * (  - 1./32.*pow(
         m33,2)*RSe45613p + 1./32.*pow(m33,2)*RSe463p + 1./32.*pow(
         m33,2)*RSe461p - 1./32.*pow(m33,2)*RSe465p - 1./16.*m11*m33*
         RSe45613p + 1./16.*m11*m33*RSe463p + 1./16.*m11*m33*RSe461p - 
         1./16.*m11*m33*RSe465p - 1./32.*pow(m11,2)*RSe45613p + 1./32.*
         pow(m11,2)*RSe463p + 1./32.*pow(m11,2)*RSe461p - 1./32.*pow(
         m11,2)*RSe465p );

      F13p6x23R +=  + hh21d(1,me2,m16,m36,m13,mu2) * (  - 1./32.*pow(
         m33,2)*RSe45613p + 1./32.*pow(m33,2)*RSe453p + 1./32.*pow(
         m33,2)*RSe451p - 1./32.*pow(m33,2)*RSe456p - 1./16.*m11*m33*
         RSe45613p + 1./16.*m11*m33*RSe453p + 1./16.*m11*m33*RSe451p - 
         1./16.*m11*m33*RSe456p - 1./32.*pow(m11,2)*RSe45613p + 1./32.*
         pow(m11,2)*RSe453p + 1./32.*pow(m11,2)*RSe451p - 1./32.*pow(
         m11,2)*RSe456p );

      F13p6x23R +=  + hh21d(1,mp2,m14,m34,m13,mu2) * (  - 1./32.*pow(
         m33,2)*RSp45613e + 1./32.*pow(m33,2)*RSp563e + 1./32.*pow(
         m33,2)*RSp561e - 1./32.*pow(m33,2)*RSp564e - 1./16.*m11*m33*
         RSp45613e + 1./16.*m11*m33*RSp563e + 1./16.*m11*m33*RSp561e - 
         1./16.*m11*m33*RSp564e - 1./32.*pow(m11,2)*RSp45613e + 1./32.*
         pow(m11,2)*RSp563e + 1./32.*pow(m11,2)*RSp561e - 1./32.*pow(
         m11,2)*RSp564e );

      F13p6x23R +=  + hh21d(1,mp2,m15,m35,m13,mu2) * (  - 1./32.*pow(
         m33,2)*RSp45613e + 1./32.*pow(m33,2)*RSp463e + 1./32.*pow(
         m33,2)*RSp461e - 1./32.*pow(m33,2)*RSp465e - 1./16.*m11*m33*
         RSp45613e + 1./16.*m11*m33*RSp463e + 1./16.*m11*m33*RSp461e - 
         1./16.*m11*m33*RSp465e - 1./32.*pow(m11,2)*RSp45613e + 1./32.*
         pow(m11,2)*RSp463e + 1./32.*pow(m11,2)*RSp461e - 1./32.*pow(
         m11,2)*RSp465e );

      F13p6x23R +=  + hh21d(1,mp2,m16,m36,m13,mu2) * (  - 1./32.*pow(
         m33,2)*RSp45613e + 1./32.*pow(m33,2)*RSp453e + 1./32.*pow(
         m33,2)*RSp451e - 1./32.*pow(m33,2)*RSp456e - 1./16.*m11*m33*
         RSp45613e + 1./16.*m11*m33*RSp453e + 1./16.*m11*m33*RSp451e - 
         1./16.*m11*m33*RSp456e - 1./32.*pow(m11,2)*RSp45613e + 1./32.*
         pow(m11,2)*RSp453e + 1./32.*pow(m11,2)*RSp451e - 1./32.*pow(
         m11,2)*RSp456e );

      F13p6x23R +=  + hh21d(3,m13,m11,m11,m13,mu2) * (  - 1./48.*pow(
         m33,2)*RD1456ep*RS14563pe - 1./96.*pow(m33,2)*RS1456pee*
         RD1456ep - 1./96.*pow(m33,2)*RS1456ppe*RD1456ep + 1./96.*pow(
         m33,2)*RS145pe*RD1456ep + 1./96.*pow(m33,2)*RS146pe*RD1456ep
          + 1./96.*pow(m33,2)*RS156pe*RD1456ep - 1./24.*m11*m33*
         RD1456ep*RS14563pe - 1./48.*m11*m33*RS1456pee*RD1456ep - 1./48.
         *m11*m33*RS1456ppe*RD1456ep + 1./48.*m11*m33*RS145pe*RD1456ep
          + 1./48.*m11*m33*RS146pe*RD1456ep + 1./48.*m11*m33*RS156pe*
         RD1456ep - 1./48.*pow(m11,2)*RD1456ep*RS14563pe - 1./96.*pow(
         m11,2)*RS1456pee*RD1456ep - 1./96.*pow(m11,2)*RS1456ppe*
         RD1456ep + 1./96.*pow(m11,2)*RS145pe*RD1456ep + 1./96.*pow(
         m11,2)*RS146pe*RD1456ep + 1./96.*pow(m11,2)*RS156pe*RD1456ep )
         ;

      F13p6x23R +=  + hh21d(3,m13,m11,m33,m13,mu2) * (  - 1./48.*pow(
         m33,2)*RD1456ep*RS34561pe - 1./96.*pow(m33,2)*RD1456ep*
         RS3456pee - 1./96.*pow(m33,2)*RD1456ep*RS3456ppe + 1./96.*pow(
         m33,2)*RD1456ep*RS345pe + 1./96.*pow(m33,2)*RD1456ep*RS346pe
          + 1./96.*pow(m33,2)*RD1456ep*RS356pe - 1./24.*m11*m33*
         RD1456ep*RS34561pe - 1./48.*m11*m33*RD1456ep*RS3456pee - 1./48.
         *m11*m33*RD1456ep*RS3456ppe + 1./48.*m11*m33*RD1456ep*RS345pe
          + 1./48.*m11*m33*RD1456ep*RS346pe + 1./48.*m11*m33*RD1456ep*
         RS356pe - 1./48.*pow(m11,2)*RD1456ep*RS34561pe - 1./96.*pow(
         m11,2)*RD1456ep*RS3456pee - 1./96.*pow(m11,2)*RD1456ep*
         RS3456ppe + 1./96.*pow(m11,2)*RD1456ep*RS345pe + 1./96.*pow(
         m11,2)*RD1456ep*RS346pe + 1./96.*pow(m11,2)*RD1456ep*RS356pe )
         ;

      F13p6x23R +=  + hh21d(3,m13,m11,me2,m13,mu2) * (  - 1./48.*pow(
         m33,2)*RD1456ep*RSe45613p + 1./96.*pow(m33,2)*RD1456ep*
         RSe45633p + 1./96.*pow(m33,2)*RSe45611p*RD1456ep - 1./24.*m11*
         m33*RD1456ep*RSe45613p + 1./48.*m11*m33*RD1456ep*RSe45633p + 1.
         /48.*m11*m33*RSe45611p*RD1456ep - 1./48.*pow(m11,2)*RD1456ep*
         RSe45613p + 1./96.*pow(m11,2)*RD1456ep*RSe45633p + 1./96.*pow(
         m11,2)*RSe45611p*RD1456ep );

      F13p6x23R +=  + hh21d(3,m13,m11,mp2,m13,mu2) * (  - 1./48.*pow(
         m33,2)*RD1456ep*RSp45613e + 1./96.*pow(m33,2)*RD1456ep*
         RSp45633e + 1./96.*pow(m33,2)*RSp45611e*RD1456ep - 1./24.*m11*
         m33*RD1456ep*RSp45613e + 1./48.*m11*m33*RD1456ep*RSp45633e + 1.
         /48.*m11*m33*RSp45611e*RD1456ep - 1./48.*pow(m11,2)*RD1456ep*
         RSp45613e + 1./96.*pow(m11,2)*RD1456ep*RSp45633e + 1./96.*pow(
         m11,2)*RSp45611e*RD1456ep );

      F13p6x23R +=  + hh21d(3,m13,m33,m11,m13,mu2) * (  - 1./48.*pow(
         m33,2)*RD3456ep*RS14563pe - 1./96.*pow(m33,2)*RS1456pee*
         RD3456ep - 1./96.*pow(m33,2)*RS1456ppe*RD3456ep + 1./96.*pow(
         m33,2)*RS145pe*RD3456ep + 1./96.*pow(m33,2)*RS146pe*RD3456ep
          + 1./96.*pow(m33,2)*RS156pe*RD3456ep - 1./24.*m11*m33*
         RD3456ep*RS14563pe - 1./48.*m11*m33*RS1456pee*RD3456ep - 1./48.
         *m11*m33*RS1456ppe*RD3456ep + 1./48.*m11*m33*RS145pe*RD3456ep
          + 1./48.*m11*m33*RS146pe*RD3456ep + 1./48.*m11*m33*RS156pe*
         RD3456ep - 1./48.*pow(m11,2)*RD3456ep*RS14563pe - 1./96.*pow(
         m11,2)*RS1456pee*RD3456ep - 1./96.*pow(m11,2)*RS1456ppe*
         RD3456ep + 1./96.*pow(m11,2)*RS145pe*RD3456ep + 1./96.*pow(
         m11,2)*RS146pe*RD3456ep + 1./96.*pow(m11,2)*RS156pe*RD3456ep )
         ;

      F13p6x23R +=  + hh21d(3,m13,m33,m33,m13,mu2) * (  - 1./48.*pow(
         m33,2)*RD3456ep*RS34561pe - 1./96.*pow(m33,2)*RS3456pee*
         RD3456ep - 1./96.*pow(m33,2)*RS3456ppe*RD3456ep + 1./96.*pow(
         m33,2)*RS345pe*RD3456ep + 1./96.*pow(m33,2)*RS346pe*RD3456ep
          + 1./96.*pow(m33,2)*RS356pe*RD3456ep - 1./24.*m11*m33*
         RD3456ep*RS34561pe - 1./48.*m11*m33*RS3456pee*RD3456ep - 1./48.
         *m11*m33*RS3456ppe*RD3456ep + 1./48.*m11*m33*RS345pe*RD3456ep
          + 1./48.*m11*m33*RS346pe*RD3456ep + 1./48.*m11*m33*RS356pe*
         RD3456ep - 1./48.*pow(m11,2)*RD3456ep*RS34561pe - 1./96.*pow(
         m11,2)*RS3456pee*RD3456ep - 1./96.*pow(m11,2)*RS3456ppe*
         RD3456ep + 1./96.*pow(m11,2)*RS345pe*RD3456ep + 1./96.*pow(
         m11,2)*RS346pe*RD3456ep + 1./96.*pow(m11,2)*RS356pe*RD3456ep )
         ;

      F13p6x23R +=  + hh21d(3,m13,m33,me2,m13,mu2) * (  - 1./48.*pow(
         m33,2)*RD3456ep*RSe45613p + 1./96.*pow(m33,2)*RSe45633p*
         RD3456ep + 1./96.*pow(m33,2)*RSe45611p*RD3456ep - 1./24.*m11*
         m33*RD3456ep*RSe45613p + 1./48.*m11*m33*RSe45633p*RD3456ep + 1.
         /48.*m11*m33*RSe45611p*RD3456ep - 1./48.*pow(m11,2)*RD3456ep*
         RSe45613p + 1./96.*pow(m11,2)*RSe45633p*RD3456ep + 1./96.*pow(
         m11,2)*RSe45611p*RD3456ep );

      F13p6x23R +=  + hh21d(3,m13,m33,mp2,m13,mu2) * (  - 1./48.*pow(
         m33,2)*RD3456ep*RSp45613e + 1./96.*pow(m33,2)*RSp45633e*
         RD3456ep + 1./96.*pow(m33,2)*RSp45611e*RD3456ep - 1./24.*m11*
         m33*RD3456ep*RSp45613e + 1./48.*m11*m33*RSp45633e*RD3456ep + 1.
         /48.*m11*m33*RSp45611e*RD3456ep - 1./48.*pow(m11,2)*RD3456ep*
         RSp45613e + 1./96.*pow(m11,2)*RSp45633e*RD3456ep + 1./96.*pow(
         m11,2)*RSp45611e*RD3456ep );

      F13p6x23R +=  + hh21d(3,m14,m11,m34,m13,mu2) * (  - 1./32.*pow(
         m33,2)*RD1456ep - 1./16.*m11*m33*RD1456ep - 1./32.*pow(m11,2)*
         RD1456ep );

      F13p6x23R +=  + hh21d(3,m15,m11,m35,m13,mu2) * (  - 1./32.*pow(
         m33,2)*RD1456ep - 1./16.*m11*m33*RD1456ep - 1./32.*pow(m11,2)*
         RD1456ep );

      F13p6x23R +=  + hh21d(3,m16,m11,m36,m13,mu2) * (  - 1./32.*pow(
         m33,2)*RD1456ep - 1./16.*m11*m33*RD1456ep - 1./32.*pow(m11,2)*
         RD1456ep );

      F13p6x23R +=  + hh21d(3,m34,m33,m14,m13,mu2) * (  - 1./32.*pow(
         m33,2)*RD3456ep - 1./16.*m11*m33*RD3456ep - 1./32.*pow(m11,2)*
         RD3456ep );

      F13p6x23R +=  + hh21d(3,m35,m33,m15,m13,mu2) * (  - 1./32.*pow(
         m33,2)*RD3456ep - 1./16.*m11*m33*RD3456ep - 1./32.*pow(m11,2)*
         RD3456ep );

      F13p6x23R +=  + hh21d(3,m36,m33,m16,m13,mu2) * (  - 1./32.*pow(
         m33,2)*RD3456ep - 1./16.*m11*m33*RD3456ep - 1./32.*pow(m11,2)*
         RD3456ep );

      F13p6x23R +=  + hh21d(7,m13,m11,m11,m13,mu2) * ( 1./192.*pow(
         m33,2)*pow(RD1456ep,2) + 1./96.*m11*m33*pow(RD1456ep,2) + 1./
         192.*pow(m11,2)*pow(RD1456ep,2) );

      F13p6x23R +=  + hh21d(7,m13,m11,m33,m13,mu2) * ( 1./96.*pow(
         m33,2)*RD1456ep*RD3456ep + 1./48.*m11*m33*RD1456ep*RD3456ep + 
         1./96.*pow(m11,2)*RD1456ep*RD3456ep );

      F13p6x23R +=  + hh21d(7,m13,m33,m33,m13,mu2) * ( 1./192.*pow(
         m33,2)*pow(RD3456ep,2) + 1./96.*m11*m33*pow(RD3456ep,2) + 1./
         192.*pow(m11,2)*pow(RD3456ep,2) );

  return F13p6x23R/pow(f0,4);
}
